/* Generated by Cython 3.0.11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "extra_compile_args": [
            "-O2"
        ],
        "name": "uvloop.loop",
        "sources": [
            "uvloop/loop.pyx"
        ]
    },
    "module_name": "uvloop.loop"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_11" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000BF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__uvloop__loop
#define __PYX_HAVE_API__uvloop__loop
/* Early includes */
#include <stdint.h>
#include <sys/types.h>
#include <unistd.h>
#include "arpa/inet.h"
#include "sys/socket.h"
#include "sys/un.h"
#include "unistd.h"
#include "pthread.h"
#include "includes/compat.h"
#include "includes/fork_handler.h"
#include "uv.h"

    // Number of seconds to wait for SSL handshake to complete
    // The default timeout matches that of Nginx.
    #define SSL_HANDSHAKE_TIMEOUT 60.0

    // Number of seconds to wait for SSL shutdown to complete
    // The default timeout mimics lingering_time
    #define SSL_SHUTDOWN_TIMEOUT 30.0
    
#include "includes/debug.h"
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stddef.h>
#include "pythread.h"

    enum {__PREALLOCED_BUFS = 4};
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "uvloop/loop.pyx",
  "uvloop/errors.pyx",
  "uvloop/handles/handle.pyx",
  "uvloop/handles/tcp.pyx",
  "uvloop/handles/process.pyx",
  "uvloop/dns.pyx",
  "<stringsource>",
  "uvloop/includes/flowcontrol.pxd",
  ".eggs/Cython-3.0.11-py3.12-linux-x86_64.egg/Cython/Includes/cpython/contextvars.pxd",
  "uvloop/loop.pxd",
  "uvloop/cbhandles.pyx",
  "uvloop/pseudosock.pyx",
  "uvloop/lru.pyx",
  "uvloop/handles/async_.pyx",
  "uvloop/handles/idle.pyx",
  "uvloop/handles/check.pyx",
  "uvloop/handles/timer.pyx",
  "uvloop/handles/poll.pyx",
  "uvloop/handles/basetransport.pyx",
  "uvloop/handles/basetransport.pxd",
  "uvloop/handles/stream.pyx",
  "uvloop/handles/streamserver.pyx",
  "uvloop/handles/pipe.pyx",
  "uvloop/handles/fsevent.pyx",
  "uvloop/request.pyx",
  "uvloop/sslproto.pyx",
  "uvloop/handles/udp.pyx",
  "uvloop/server.pyx",
  ".eggs/Cython-3.0.11-py3.12-linux-x86_64.egg/Cython/Includes/cpython/type.pxd",
  ".eggs/Cython-3.0.11-py3.12-linux-x86_64.egg/Cython/Includes/cpython/bool.pxd",
  ".eggs/Cython-3.0.11-py3.12-linux-x86_64.egg/Cython/Includes/cpython/complex.pxd",
  "uvloop/includes/stdlib.pxi",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_6uvloop_4loop_Loop;
struct __pyx_obj_6uvloop_4loop_Handle;
struct __pyx_obj_6uvloop_4loop_TimerHandle;
struct __pyx_obj_6uvloop_4loop_UVHandle;
struct __pyx_obj_6uvloop_4loop_UVSocketHandle;
struct __pyx_obj_6uvloop_4loop_UVAsync;
struct __pyx_obj_6uvloop_4loop_UVIdle;
struct __pyx_obj_6uvloop_4loop_UVCheck;
struct __pyx_obj_6uvloop_4loop_UVTimer;
struct __pyx_obj_6uvloop_4loop_UVPoll;
struct __pyx_obj_6uvloop_4loop_UVBaseTransport;
struct __pyx_obj_6uvloop_4loop_UVStream;
struct __pyx_obj_6uvloop_4loop_UVStreamServer;
struct __pyx_obj_6uvloop_4loop_TCPServer;
struct __pyx_obj_6uvloop_4loop_TCPTransport;
struct __pyx_obj_6uvloop_4loop_UnixServer;
struct __pyx_obj_6uvloop_4loop_UnixTransport;
struct __pyx_obj_6uvloop_4loop_ReadUnixTransport;
struct __pyx_obj_6uvloop_4loop_WriteUnixTransport;
struct __pyx_obj_6uvloop_4loop_UVProcess;
struct __pyx_obj_6uvloop_4loop_UVProcessTransport;
struct __pyx_obj_6uvloop_4loop_UVFSEvent;
struct __pyx_obj_6uvloop_4loop_UVRequest;
struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport;
struct __pyx_obj_6uvloop_4loop_SSLProtocol;
struct __pyx_obj_6uvloop_4loop_UDPTransport;
struct __pyx_obj_6uvloop_4loop_Server;
struct __pyx_obj_6uvloop_4loop_PseudoSocket;
struct __pyx_obj_6uvloop_4loop_LruCache;
struct __pyx_obj_6uvloop_4loop__StreamWriteContext;
struct __pyx_obj_6uvloop_4loop__TCPConnectRequest;
struct __pyx_obj_6uvloop_4loop__PipeConnectRequest;
struct __pyx_obj_6uvloop_4loop_SockAddrHolder;
struct __pyx_obj_6uvloop_4loop_AddrInfo;
struct __pyx_obj_6uvloop_4loop_AddrInfoRequest;
struct __pyx_obj_6uvloop_4loop_NameInfoRequest;
struct __pyx_obj_6uvloop_4loop__UDPSendContext;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever;
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled;

/* "uv.pxd":14
 * # setting UV_HANDLE_READABLE on pipe_t to workaround this limitation
 * # temporarily, please see also #317.
 * cdef enum:             # <<<<<<<<<<<<<<
 *     UV_INTERNAL_HANDLE_READABLE = 0x00004000
 * 
 */
enum  {
  __pyx_e_6uvloop_8includes_2uv_UV_INTERNAL_HANDLE_READABLE = 0x00004000
};
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};
struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error;
struct __pyx_opt_args_6uvloop_4loop_15UVBaseTransport__fatal_error;
struct __pyx_opt_args_6uvloop_4loop_9TCPServer_bind;
struct __pyx_opt_args_6uvloop_4loop_12TCPTransport_bind;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__get_extra_info;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__start_shutdown;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_flush;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_shutdown;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__call_eof_received;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits;
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error;
struct __pyx_opt_args_6uvloop_4loop_12UDPTransport__on_sent;
struct __pyx_opt_args_6uvloop_4loop_6UVPoll__fatal_error;
struct __pyx_opt_args_6uvloop_4loop_14UVStreamServer__fatal_error;

/* "uvloop/includes/consts.pxi":1
 * cdef enum:             # <<<<<<<<<<<<<<
 *     UV_STREAM_RECV_BUF_SIZE = 256000  # 250kb
 * 
 */
enum  {

  /* "uvloop/includes/consts.pxi":18
 * 
 *     LOG_THRESHOLD_FOR_CONNLOST_WRITES = 5
 *     SSL_READ_MAX_SIZE = 256 * 1024             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_e_6uvloop_4loop_UV_STREAM_RECV_BUF_SIZE = 0x3E800,
  __pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER = 64,
  __pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER_SSL_READ = 0x100,
  __pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER_SSL_WRITE = 0x200,
  __pyx_e_6uvloop_4loop_DEFAULT_FREELIST_SIZE = 0xFA,
  __pyx_e_6uvloop_4loop_DNS_PYADDR_TO_SOCKADDR_CACHE_SIZE = 0x800,
  __pyx_e_6uvloop_4loop_DEBUG_STACK_DEPTH = 10,
  __pyx_e_6uvloop_4loop___PROCESS_DEBUG_SLEEP_AFTER_FORK = 1,
  __pyx_e_6uvloop_4loop_LOG_THRESHOLD_FOR_CONNLOST_WRITES = 5,
  __pyx_e_6uvloop_4loop_SSL_READ_MAX_SIZE = (0x100 * 0x400)
};

/* "uvloop/sslproto.pxd":1
 * cdef enum SSLProtocolState:             # <<<<<<<<<<<<<<
 *     UNWRAPPED = 0
 *     DO_HANDSHAKE = 1
 */
enum __pyx_t_6uvloop_4loop_SSLProtocolState {
  __pyx_e_6uvloop_4loop_UNWRAPPED = 0,
  __pyx_e_6uvloop_4loop_DO_HANDSHAKE = 1,
  __pyx_e_6uvloop_4loop_WRAPPED = 2,
  __pyx_e_6uvloop_4loop_FLUSHING = 3,
  __pyx_e_6uvloop_4loop_SHUTDOWN = 4
};

/* "uvloop/sslproto.pxd":9
 * 
 * 
 * cdef enum AppProtocolState:             # <<<<<<<<<<<<<<
 *     # This tracks the state of app protocol (https://git.io/fj59P):
 *     #
 */
enum __pyx_t_6uvloop_4loop_AppProtocolState {
  __pyx_e_6uvloop_4loop_STATE_INIT = 0,
  __pyx_e_6uvloop_4loop_STATE_CON_MADE = 1,
  __pyx_e_6uvloop_4loop_STATE_EOF = 2,
  __pyx_e_6uvloop_4loop_STATE_CON_LOST = 3
};

/* "uvloop/loop.pxd":26
 * cdef class UVBaseTransport(UVSocketHandle)
 * 
 * ctypedef object (*method_t)(object)             # <<<<<<<<<<<<<<
 * ctypedef object (*method1_t)(object, object)
 * ctypedef object (*method2_t)(object, object, object)
 */
typedef PyObject *(*__pyx_t_6uvloop_4loop_method_t)(PyObject *);

/* "uvloop/loop.pxd":27
 * 
 * ctypedef object (*method_t)(object)
 * ctypedef object (*method1_t)(object, object)             # <<<<<<<<<<<<<<
 * ctypedef object (*method2_t)(object, object, object)
 * ctypedef object (*method3_t)(object, object, object, object)
 */
typedef PyObject *(*__pyx_t_6uvloop_4loop_method1_t)(PyObject *, PyObject *);

/* "uvloop/loop.pxd":28
 * ctypedef object (*method_t)(object)
 * ctypedef object (*method1_t)(object, object)
 * ctypedef object (*method2_t)(object, object, object)             # <<<<<<<<<<<<<<
 * ctypedef object (*method3_t)(object, object, object, object)
 * 
 */
typedef PyObject *(*__pyx_t_6uvloop_4loop_method2_t)(PyObject *, PyObject *, PyObject *);

/* "uvloop/loop.pxd":29
 * ctypedef object (*method1_t)(object, object)
 * ctypedef object (*method2_t)(object, object, object)
 * ctypedef object (*method3_t)(object, object, object, object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef PyObject *(*__pyx_t_6uvloop_4loop_method3_t)(PyObject *, PyObject *, PyObject *, PyObject *);

/* "uvloop/handles/handle.pxd":24
 * 
 *     cdef _error(self, exc, throw)
 *     cdef _fatal_error(self, exc, throw, reason=?)             # <<<<<<<<<<<<<<
 * 
 *     cdef _warn_unclosed(self)
 */
struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error {
  int __pyx_n;
  PyObject *reason;
};

/* "uvloop/handles/basetransport.pxd":35
 * 
 *     # Overloads of UVHandle methods:
 *     cdef _fatal_error(self, exc, throw, reason=?)             # <<<<<<<<<<<<<<
 *     cdef _close(self)
 * 
 */
struct __pyx_opt_args_6uvloop_4loop_15UVBaseTransport__fatal_error {
  int __pyx_n;
  PyObject *reason;
};

/* "uvloop/handles/tcp.pxd":2
 * cdef class TCPServer(UVStreamServer):
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=*)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
struct __pyx_opt_args_6uvloop_4loop_9TCPServer_bind {
  int __pyx_n;
  unsigned int flags;
};

/* "uvloop/handles/tcp.pxd":20
 *         system.sockaddr_storage __sockname
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=*)             # <<<<<<<<<<<<<<
 *     cdef connect(self, system.sockaddr* addr)
 *     cdef _set_nodelay(self)
 */
struct __pyx_opt_args_6uvloop_4loop_12TCPTransport_bind {
  int __pyx_n;
  unsigned int flags;
};

/* "uvloop/sslproto.pxd":88
 * 
 *     cdef _set_app_protocol(self, app_protocol)
 *     cdef _wakeup_waiter(self, exc=*)             # <<<<<<<<<<<<<<
 *     cdef _get_extra_info(self, name, default=*)
 *     cdef _set_state(self, SSLProtocolState new_state)
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter {
  int __pyx_n;
  PyObject *exc;
};

/* "uvloop/sslproto.pxd":89
 *     cdef _set_app_protocol(self, app_protocol)
 *     cdef _wakeup_waiter(self, exc=*)
 *     cdef _get_extra_info(self, name, default=*)             # <<<<<<<<<<<<<<
 *     cdef _set_state(self, SSLProtocolState new_state)
 * 
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__get_extra_info {
  int __pyx_n;
  PyObject *__pyx_default;
};

/* "uvloop/sslproto.pxd":101
 *     # Shutdown flow
 * 
 *     cdef _start_shutdown(self, object context=*)             # <<<<<<<<<<<<<<
 *     cdef _check_shutdown_timeout(self)
 *     cdef _do_read_into_void(self, object context)
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__start_shutdown {
  int __pyx_n;
  PyObject *context;
};

/* "uvloop/sslproto.pxd":104
 *     cdef _check_shutdown_timeout(self)
 *     cdef _do_read_into_void(self, object context)
 *     cdef _do_flush(self, object context=*)             # <<<<<<<<<<<<<<
 *     cdef _do_shutdown(self, object context=*)
 *     cdef _on_shutdown_complete(self, shutdown_exc)
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_flush {
  int __pyx_n;
  PyObject *context;
};

/* "uvloop/sslproto.pxd":105
 *     cdef _do_read_into_void(self, object context)
 *     cdef _do_flush(self, object context=*)
 *     cdef _do_shutdown(self, object context=*)             # <<<<<<<<<<<<<<
 *     cdef _on_shutdown_complete(self, shutdown_exc)
 *     cdef _abort(self, exc)
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_shutdown {
  int __pyx_n;
  PyObject *context;
};

/* "uvloop/sslproto.pxd":120
 *     cdef _do_read__buffered(self)
 *     cdef _do_read__copied(self)
 *     cdef _call_eof_received(self, object context=*)             # <<<<<<<<<<<<<<
 * 
 *     # Flow control for writes from APP socket
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__call_eof_received {
  int __pyx_n;
  PyObject *context;
};

/* "uvloop/sslproto.pxd":124
 *     # Flow control for writes from APP socket
 * 
 *     cdef _control_app_writing(self, object context=*)             # <<<<<<<<<<<<<<
 *     cdef size_t _get_write_buffer_size(self)
 *     cdef _set_write_buffer_limits(self, high=*, low=*)
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing {
  int __pyx_n;
  PyObject *context;
};

/* "uvloop/sslproto.pxd":126
 *     cdef _control_app_writing(self, object context=*)
 *     cdef size_t _get_write_buffer_size(self)
 *     cdef _set_write_buffer_limits(self, high=*, low=*)             # <<<<<<<<<<<<<<
 * 
 *     # Flow control for reads to APP socket
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits {
  int __pyx_n;
  PyObject *high;
  PyObject *low;
};

/* "uvloop/sslproto.pxd":136
 * 
 *     cdef _control_ssl_reading(self)
 *     cdef _set_read_buffer_limits(self, high=*, low=*)             # <<<<<<<<<<<<<<
 *     cdef size_t _get_read_buffer_size(self)
 *     cdef _fatal_error(self, exc, message=*)
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits {
  int __pyx_n;
  PyObject *high;
  PyObject *low;
};

/* "uvloop/sslproto.pxd":138
 *     cdef _set_read_buffer_limits(self, high=*, low=*)
 *     cdef size_t _get_read_buffer_size(self)
 *     cdef _fatal_error(self, exc, message=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error {
  int __pyx_n;
  PyObject *message;
};

/* "uvloop/handles/udp.pxd":22
 * 
 *     cdef _on_receive(self, bytes data, object exc, object addr)
 *     cdef _on_sent(self, object exc, object context=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_6uvloop_4loop_12UDPTransport__on_sent {
  int __pyx_n;
  PyObject *context;
};

/* "uvloop/handles/poll.pyx":173
 *         UVHandle._close(<UVHandle>self)
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         try:
 *             if self.reading_handle is not None:
 */
struct __pyx_opt_args_6uvloop_4loop_6UVPoll__fatal_error {
  int __pyx_n;
  PyObject *reason;
};

/* "uvloop/handles/streamserver.pyx":92
 *         client._accept(<UVStream>self)
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         # Overload UVHandle._fatal_error
 * 
 */
struct __pyx_opt_args_6uvloop_4loop_14UVStreamServer__fatal_error {
  int __pyx_n;
  PyObject *reason;
};

/* "uvloop/loop.pxd":32
 * 
 * 
 * cdef class Loop:             # <<<<<<<<<<<<<<
 *     cdef:
 *         uv.uv_loop_t *uvloop
 */
struct __pyx_obj_6uvloop_4loop_Loop {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_Loop *__pyx_vtab;
  uv_loop_t *uvloop;
  int _coroutine_debug_set;
  int _coroutine_origin_tracking_saved_depth;
  PyObject *slow_callback_duration;
  int _closed;
  int _debug;
  int _running;
  int _stopping;
  uint64_t _thread_id;
  PyObject *_task_factory;
  PyObject *_exception_handler;
  PyObject *_default_executor;
  PyObject *_ready;
  PyObject *_queued_streams;
  PyObject *_executing_streams;
  Py_ssize_t _ready_len;
  PyObject *_servers;
  PyObject *_transports;
  PyObject *_processes;
  PyObject *_fd_to_reader_fileobj;
  PyObject *_fd_to_writer_fileobj;
  PyObject *_unix_server_sockets;
  PyObject *_signals;
  PyObject *_signal_handlers;
  PyObject *_ssock;
  PyObject *_csock;
  int _listening_signals;
  int _old_signal_wakeup_id;
  PyObject *_timers;
  PyObject *_polls;
  struct __pyx_obj_6uvloop_4loop_UVProcess *active_process_handler;
  struct __pyx_obj_6uvloop_4loop_UVAsync *handler_async;
  struct __pyx_obj_6uvloop_4loop_UVIdle *handler_idle;
  struct __pyx_obj_6uvloop_4loop_UVCheck *handler_check__exec_writes;
  PyObject *_last_error;
  PyObject *__weakref__;
  PyObject *_asyncgens;
  int _asyncgens_shutdown_called;
  int _executor_shutdown_called;
  char _recv_buffer[__pyx_e_6uvloop_4loop_UV_STREAM_RECV_BUF_SIZE];
  int _recv_buffer_in_use;
  int _debug_cc;
  PyObject *_debug_handles_total;
  PyObject *_debug_handles_closed;
  PyObject *_debug_handles_current;
  uint64_t _debug_uv_handles_total;
  uint64_t _debug_uv_handles_freed;
  uint64_t _debug_cb_handles_total;
  uint64_t _debug_cb_handles_count;
  uint64_t _debug_cb_timer_handles_total;
  uint64_t _debug_cb_timer_handles_count;
  uint64_t _debug_stream_shutdown_errors_total;
  uint64_t _debug_stream_listen_errors_total;
  uint64_t _debug_stream_read_cb_total;
  uint64_t _debug_stream_read_cb_errors_total;
  uint64_t _debug_stream_read_eof_total;
  uint64_t _debug_stream_read_eof_cb_errors_total;
  uint64_t _debug_stream_read_errors_total;
  uint64_t _debug_stream_write_tries;
  uint64_t _debug_stream_write_errors_total;
  uint64_t _debug_stream_write_ctx_total;
  uint64_t _debug_stream_write_ctx_cnt;
  uint64_t _debug_stream_write_cb_errors_total;
  uint64_t _poll_read_events_total;
  uint64_t _poll_read_cb_errors_total;
  uint64_t _poll_write_events_total;
  uint64_t _poll_write_cb_errors_total;
  uint64_t _sock_try_write_total;
  uint64_t _debug_exception_handler_cnt;
};


/* "uvloop/cbhandles.pxd":1
 * cdef class Handle:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Loop loop
 */
struct __pyx_obj_6uvloop_4loop_Handle {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_Handle *__pyx_vtab;
  struct __pyx_obj_6uvloop_4loop_Loop *loop;
  PyObject *context;
  int _cancelled;
  PyObject *meth_name;
  int cb_type;
  void *callback;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *__weakref__;
  PyObject *_source_traceback;
};


/* "uvloop/cbhandles.pxd":25
 * 
 * 
 * cdef class TimerHandle:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object callback
 */
struct __pyx_obj_6uvloop_4loop_TimerHandle {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_TimerHandle *__pyx_vtab;
  PyObject *callback;
  PyObject *args;
  int _cancelled;
  struct __pyx_obj_6uvloop_4loop_UVTimer *timer;
  struct __pyx_obj_6uvloop_4loop_Loop *loop;
  PyObject *context;
  PyObject *_debug_info;
  PyObject *__weakref__;
  PyObject *_when;
};


/* "uvloop/loop.pxd":17
 * 
 * 
 * cdef class UVHandle             # <<<<<<<<<<<<<<
 * cdef class UVSocketHandle(UVHandle)
 * 
 */
struct __pyx_obj_6uvloop_4loop_UVHandle {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *__pyx_vtab;
  uv_handle_t *_handle;
  struct __pyx_obj_6uvloop_4loop_Loop *_loop;
  PyObject *_source_traceback;
  int _closed;
  int _inited;
  PyObject *context;
  int _has_handle;
};


/* "uvloop/loop.pxd":18
 * 
 * cdef class UVHandle
 * cdef class UVSocketHandle(UVHandle)             # <<<<<<<<<<<<<<
 * 
 * cdef class UVAsync(UVHandle)
 */
struct __pyx_obj_6uvloop_4loop_UVSocketHandle {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *_fileobj;
  PyObject *_UVSocketHandle__cached_socket;
};


/* "uvloop/loop.pxd":20
 * cdef class UVSocketHandle(UVHandle)
 * 
 * cdef class UVAsync(UVHandle)             # <<<<<<<<<<<<<<
 * cdef class UVTimer(UVHandle)
 * cdef class UVIdle(UVHandle)
 */
struct __pyx_obj_6uvloop_4loop_UVAsync {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  __pyx_t_6uvloop_4loop_method_t callback;
  PyObject *ctx;
};


/* "uvloop/loop.pxd":22
 * cdef class UVAsync(UVHandle)
 * cdef class UVTimer(UVHandle)
 * cdef class UVIdle(UVHandle)             # <<<<<<<<<<<<<<
 * 
 * cdef class UVBaseTransport(UVSocketHandle)
 */
struct __pyx_obj_6uvloop_4loop_UVIdle {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  struct __pyx_obj_6uvloop_4loop_Handle *h;
  int running;
};


/* "uvloop/handles/check.pxd":1
 * cdef class UVCheck(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Handle h
 */
struct __pyx_obj_6uvloop_4loop_UVCheck {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  struct __pyx_obj_6uvloop_4loop_Handle *h;
  int running;
};


/* "uvloop/loop.pxd":21
 * 
 * cdef class UVAsync(UVHandle)
 * cdef class UVTimer(UVHandle)             # <<<<<<<<<<<<<<
 * cdef class UVIdle(UVHandle)
 * 
 */
struct __pyx_obj_6uvloop_4loop_UVTimer {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  __pyx_t_6uvloop_4loop_method_t callback;
  PyObject *ctx;
  int running;
  uint64_t timeout;
  uint64_t start_t;
};


/* "uvloop/handles/poll.pxd":1
 * cdef class UVPoll(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int fd
 */
struct __pyx_obj_6uvloop_4loop_UVPoll {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  int fd;
  struct __pyx_obj_6uvloop_4loop_Handle *reading_handle;
  struct __pyx_obj_6uvloop_4loop_Handle *writing_handle;
};


/* "uvloop/loop.pxd":24
 * cdef class UVIdle(UVHandle)
 * 
 * cdef class UVBaseTransport(UVSocketHandle)             # <<<<<<<<<<<<<<
 * 
 * ctypedef object (*method_t)(object)
 */
struct __pyx_obj_6uvloop_4loop_UVBaseTransport {
  struct __pyx_obj_6uvloop_4loop_UVSocketHandle __pyx_base;
  int _closing;
  int _protocol_connected;
  int _protocol_paused;
  PyObject *_protocol_data_received;
  size_t _high_water;
  size_t _low_water;
  PyObject *_protocol;
  struct __pyx_obj_6uvloop_4loop_Server *_server;
  PyObject *_waiter;
  PyObject *_extra_info;
  uint32_t _conn_lost;
  PyObject *__weakref__;
};


/* "uvloop/handles/stream.pxd":1
 * cdef class UVStream(UVBaseTransport):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uv.uv_shutdown_t _shutdown_req
 */
struct __pyx_obj_6uvloop_4loop_UVStream {
  struct __pyx_obj_6uvloop_4loop_UVBaseTransport __pyx_base;
  uv_shutdown_t _shutdown_req;
  int _UVStream__shutting_down;
  int _UVStream__reading;
  int _UVStream__read_error_close;
  int _UVStream__buffered;
  PyObject *_protocol_get_buffer;
  PyObject *_protocol_buffer_updated;
  int _eof;
  PyObject *_buffer;
  size_t _buffer_size;
  Py_buffer _read_pybuf;
  int _read_pybuf_acquired;
};


/* "uvloop/handles/streamserver.pxd":1
 * cdef class UVStreamServer(UVSocketHandle):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int backlog
 */
struct __pyx_obj_6uvloop_4loop_UVStreamServer {
  struct __pyx_obj_6uvloop_4loop_UVSocketHandle __pyx_base;
  int backlog;
  PyObject *ssl;
  PyObject *ssl_handshake_timeout;
  PyObject *ssl_shutdown_timeout;
  PyObject *protocol_factory;
  int opened;
  struct __pyx_obj_6uvloop_4loop_Server *_server;
};


/* "uvloop/handles/tcp.pxd":1
 * cdef class TCPServer(UVStreamServer):             # <<<<<<<<<<<<<<
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=*)
 * 
 */
struct __pyx_obj_6uvloop_4loop_TCPServer {
  struct __pyx_obj_6uvloop_4loop_UVStreamServer __pyx_base;
};


/* "uvloop/handles/tcp.pxd":13
 * 
 * 
 * cdef class TCPTransport(UVStream):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint __peername_set
 */
struct __pyx_obj_6uvloop_4loop_TCPTransport {
  struct __pyx_obj_6uvloop_4loop_UVStream __pyx_base;
  int _TCPTransport__peername_set;
  int _TCPTransport__sockname_set;
  struct sockaddr_storage _TCPTransport__peername;
  struct sockaddr_storage _TCPTransport__sockname;
};


/* "uvloop/handles/pipe.pxd":1
 * cdef class UnixServer(UVStreamServer):             # <<<<<<<<<<<<<<
 * 
 *     cdef bind(self, str path)
 */
struct __pyx_obj_6uvloop_4loop_UnixServer {
  struct __pyx_obj_6uvloop_4loop_UVStreamServer __pyx_base;
};


/* "uvloop/handles/pipe.pxd":13
 * 
 * 
 * cdef class UnixTransport(UVStream):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
struct __pyx_obj_6uvloop_4loop_UnixTransport {
  struct __pyx_obj_6uvloop_4loop_UVStream __pyx_base;
};


/* "uvloop/handles/pipe.pxd":22
 * 
 * 
 * cdef class ReadUnixTransport(UVStream):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
struct __pyx_obj_6uvloop_4loop_ReadUnixTransport {
  struct __pyx_obj_6uvloop_4loop_UVStream __pyx_base;
};


/* "uvloop/handles/pipe.pxd":29
 * 
 * 
 * cdef class WriteUnixTransport(UVStream):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
struct __pyx_obj_6uvloop_4loop_WriteUnixTransport {
  struct __pyx_obj_6uvloop_4loop_UVStream __pyx_base;
};


/* "uvloop/handles/process.pxd":1
 * cdef class UVProcess(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _returncode
 */
struct __pyx_obj_6uvloop_4loop_UVProcess {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *_returncode;
  PyObject *_pid;
  PyObject *_errpipe_read;
  PyObject *_errpipe_write;
  PyObject *_preexec_fn;
  int _restore_signals;
  PyObject *_fds_to_close;
  uv_process_options_t options;
  uv_stdio_container_t iocnt[3];
  PyObject *_UVProcess__env;
  char **uv_opt_env;
  PyObject *_UVProcess__args;
  char **uv_opt_args;
  char *uv_opt_file;
  PyObject *_UVProcess__cwd;
};


/* "uvloop/handles/process.pxd":45
 * 
 * 
 * cdef class UVProcessTransport(UVProcess):             # <<<<<<<<<<<<<<
 *     cdef:
 *         list _exit_waiters
 */
struct __pyx_obj_6uvloop_4loop_UVProcessTransport {
  struct __pyx_obj_6uvloop_4loop_UVProcess __pyx_base;
  PyObject *_exit_waiters;
  PyObject *_init_futs;
  int _stdio_ready;
  PyObject *_pending_calls;
  PyObject *_protocol;
  int _finished;
  struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *_stdin;
  struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *_stdout;
  struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *_stderr;
  PyObject *stdin_proto;
  PyObject *stdout_proto;
  PyObject *stderr_proto;
};


/* "uvloop/handles/fsevent.pxd":1
 * cdef class UVFSEvent(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object callback
 */
struct __pyx_obj_6uvloop_4loop_UVFSEvent {
  struct __pyx_obj_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *callback;
  int running;
};


/* "uvloop/request.pxd":1
 * cdef class UVRequest:             # <<<<<<<<<<<<<<
 *     cdef:
 *         uv.uv_req_t *request
 */
struct __pyx_obj_6uvloop_4loop_UVRequest {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_UVRequest *__pyx_vtab;
  uv_req_t *request;
  int done;
  struct __pyx_obj_6uvloop_4loop_Loop *loop;
};


/* "uvloop/sslproto.pxd":25
 * 
 * 
 * cdef class _SSLProtocolTransport:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Loop _loop
 */
struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_Loop *_loop;
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *_ssl_protocol;
  int _closed;
  PyObject *context;
};


/* "uvloop/sslproto.pxd":33
 * 
 * 
 * cdef class SSLProtocol:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint _server_side
 */
struct __pyx_obj_6uvloop_4loop_SSLProtocol {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *__pyx_vtab;
  int _server_side;
  PyObject *_server_hostname;
  PyObject *_sslcontext;
  PyObject *_extra;
  PyObject *_write_backlog;
  size_t _write_buffer_size;
  PyObject *_waiter;
  struct __pyx_obj_6uvloop_4loop_Loop *_loop;
  struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *_app_transport;
  int _app_transport_created;
  PyObject *_transport;
  PyObject *_ssl_handshake_timeout;
  PyObject *_ssl_shutdown_timeout;
  PyObject *_sslobj;
  PyObject *_sslobj_read;
  PyObject *_sslobj_write;
  PyObject *_incoming;
  PyObject *_incoming_write;
  PyObject *_outgoing;
  PyObject *_outgoing_read;
  char *_ssl_buffer;
  size_t _ssl_buffer_len;
  PyObject *_ssl_buffer_view;
  enum __pyx_t_6uvloop_4loop_SSLProtocolState _state;
  size_t _conn_lost;
  enum __pyx_t_6uvloop_4loop_AppProtocolState _app_state;
  int _ssl_writing_paused;
  int _app_reading_paused;
  size_t _incoming_high_water;
  size_t _incoming_low_water;
  int _ssl_reading_paused;
  int _app_writing_paused;
  size_t _outgoing_high_water;
  size_t _outgoing_low_water;
  PyObject *_app_protocol;
  int _app_protocol_is_buffer;
  PyObject *_app_protocol_get_buffer;
  PyObject *_app_protocol_buffer_updated;
  PyObject *_handshake_start_time;
  PyObject *_handshake_timeout_handle;
  PyObject *_shutdown_timeout_handle;
};


/* "uvloop/handles/udp.pxd":1
 * cdef class UDPTransport(UVBaseTransport):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint __receiving
 */
struct __pyx_obj_6uvloop_4loop_UDPTransport {
  struct __pyx_obj_6uvloop_4loop_UVBaseTransport __pyx_base;
  int _UDPTransport__receiving;
  int _family;
  PyObject *_address;
};


/* "uvloop/server.pxd":1
 * cdef class Server:             # <<<<<<<<<<<<<<
 *     cdef:
 *         list _servers
 */
struct __pyx_obj_6uvloop_4loop_Server {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_Server *__pyx_vtab;
  PyObject *_servers;
  PyObject *_waiters;
  int _active_count;
  struct __pyx_obj_6uvloop_4loop_Loop *_loop;
  int _serving;
  PyObject *_serving_forever_fut;
  PyObject *__weakref__;
};


/* "uvloop/pseudosock.pyx":1
 * cdef class PseudoSocket:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int _family
 */
struct __pyx_obj_6uvloop_4loop_PseudoSocket {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *__pyx_vtab;
  int _family;
  int _type;
  int _proto;
  int _fd;
  PyObject *_peername;
  PyObject *_sockname;
};


/* "uvloop/lru.pyx":5
 * 
 * @cython.final
 * cdef class LruCache:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_6uvloop_4loop_LruCache {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_LruCache *__pyx_vtab;
  PyObject *_dict;
  int _maxsize;
  PyObject *_dict_move_to_end;
  PyObject *_dict_get;
};


/* "uvloop/handles/stream.pyx":10
 * @cython.no_gc_clear
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class _StreamWriteContext:             # <<<<<<<<<<<<<<
 *     # used to hold additional write request information for uv_write
 * 
 */
struct __pyx_obj_6uvloop_4loop__StreamWriteContext {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *__pyx_vtab;
  uv_write_t req;
  PyObject *buffers;
  uv_buf_t uv_bufs_sml[__PREALLOCED_BUFS];
  Py_buffer py_bufs_sml[__PREALLOCED_BUFS];
  int py_bufs_sml_inuse;
  uv_buf_t *uv_bufs;
  Py_buffer *py_bufs;
  size_t py_bufs_len;
  uv_buf_t *uv_bufs_start;
  size_t uv_bufs_len;
  struct __pyx_obj_6uvloop_4loop_UVStream *stream;
  int closed;
};


/* "uvloop/handles/tcp.pyx":185
 * 
 * 
 * cdef class _TCPConnectRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         TCPTransport transport
 */
struct __pyx_obj_6uvloop_4loop__TCPConnectRequest {
  struct __pyx_obj_6uvloop_4loop_UVRequest __pyx_base;
  struct __pyx_obj_6uvloop_4loop_TCPTransport *transport;
  uv_connect_t _req_data;
};


/* "uvloop/handles/pipe.pyx":209
 * 
 * 
 * cdef class _PipeConnectRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         UnixTransport transport
 */
struct __pyx_obj_6uvloop_4loop__PipeConnectRequest {
  struct __pyx_obj_6uvloop_4loop_UVRequest __pyx_base;
  struct __pyx_obj_6uvloop_4loop_UnixTransport *transport;
  uv_connect_t _req_data;
};


/* "uvloop/dns.pyx":72
 * 
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class SockAddrHolder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int family
 */
struct __pyx_obj_6uvloop_4loop_SockAddrHolder {
  PyObject_HEAD
  int family;
  struct sockaddr_storage addr;
  Py_ssize_t addr_size;
};


/* "uvloop/dns.pyx":289
 * 
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class AddrInfo:             # <<<<<<<<<<<<<<
 *     cdef:
 *         system.addrinfo *data
 */
struct __pyx_obj_6uvloop_4loop_AddrInfo {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop_AddrInfo *__pyx_vtab;
  struct addrinfo *data;
};


/* "uvloop/dns.pyx":333
 * 
 * 
 * cdef class AddrInfoRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         system.addrinfo hints
 */
struct __pyx_obj_6uvloop_4loop_AddrInfoRequest {
  struct __pyx_obj_6uvloop_4loop_UVRequest __pyx_base;
  struct addrinfo hints;
  PyObject *callback;
  uv_getaddrinfo_t _req_data;
};


/* "uvloop/dns.pyx":396
 * 
 * 
 * cdef class NameInfoRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object callback
 */
struct __pyx_obj_6uvloop_4loop_NameInfoRequest {
  struct __pyx_obj_6uvloop_4loop_UVRequest __pyx_base;
  PyObject *callback;
  uv_getnameinfo_t _req_data;
};


/* "uvloop/handles/udp.pyx":3
 * @cython.no_gc_clear
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class _UDPSendContext:             # <<<<<<<<<<<<<<
 *     # used to hold additional write request information for uv_write
 * 
 */
struct __pyx_obj_6uvloop_4loop__UDPSendContext {
  PyObject_HEAD
  struct __pyx_vtabstruct_6uvloop_4loop__UDPSendContext *__pyx_vtab;
  uv_udp_send_t req;
  uv_buf_t uv_buf;
  Py_buffer py_buf;
  struct __pyx_obj_6uvloop_4loop_UDPTransport *udp;
  int closed;
};


/* "uvloop/loop.pyx":894
 *         return poll.is_writing()
 * 
 *     cdef _getaddrinfo(self, object host, object port,             # <<<<<<<<<<<<<<
 *                       int family, int type,
 *                       int proto, int flags,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo {
  PyObject_HEAD
  PyObject *__pyx_v_fut;
  int __pyx_v_unpack;
};


/* "uvloop/loop.pyx":936
 *         return fut
 * 
 *     cdef _getnameinfo(self, system.sockaddr *addr, int flags):             # <<<<<<<<<<<<<<
 *         cdef NameInfoRequest nr
 *         fut = self._new_future()
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo {
  PyObject_HEAD
  PyObject *__pyx_v_fut;
};


/* "uvloop/loop.pyx":1251
 * 
 *     property print_debug_info:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 return lambda: self._print_debug_info()
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
};


/* "uvloop/loop.pyx":1474
 *         return self._task_factory
 * 
 *     def run_until_complete(self, future):             # <<<<<<<<<<<<<<
 *         """Run until the Future is done.
 * 
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
};


/* "uvloop/loop.pyx":1520
 *         return future.result()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getaddrinfo(self, object host, object port, *,
 *                           int family=0, int type=0, int proto=0, int flags=0):
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo {
  PyObject_HEAD
  PyObject *__pyx_v_addr;
  int __pyx_v_family;
  int __pyx_v_flags;
  PyObject *__pyx_v_host;
  PyObject *__pyx_v_port;
  int __pyx_v_proto;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  int __pyx_v_type;
};


/* "uvloop/loop.pyx":1532
 *             host, port, family, type, proto, flags, 1)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getnameinfo(self, sockaddr, int flags=0):
 *         cdef:
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo {
  PyObject_HEAD
  struct addrinfo *__pyx_v_ai;
  struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_ai_cnt;
  int __pyx_v_flags;
  PyObject *__pyx_v_flowinfo;
  PyObject *__pyx_v_scope_id;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  struct sockaddr_in6 *__pyx_v_sin6;
  Py_ssize_t __pyx_v_sl;
  PyObject *__pyx_v_sockaddr;
};


/* "uvloop/loop.pyx":1587
 *         return await self._getnameinfo(ai.ai_addr, flags)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls {
  PyObject_HEAD
  PyObject *__pyx_v_app_transport;
  PyObject *__pyx_v_conmade_cb;
  PyObject *__pyx_v_context;
  PyObject *__pyx_v_protocol;
  PyObject *__pyx_v_resume_cb;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_server_hostname;
  PyObject *__pyx_v_server_side;
  PyObject *__pyx_v_ssl_handshake_timeout;
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_ssl_protocol;
  PyObject *__pyx_v_ssl_shutdown_timeout;
  PyObject *__pyx_v_sslcontext;
  PyObject *__pyx_v_transport;
  PyObject *__pyx_v_waiter;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "uvloop/loop.pyx":1643
 *         return app_transport
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server {
  PyObject_HEAD
  struct addrinfo *__pyx_v_addrinfo;
  PyObject *__pyx_v_backlog;
  int __pyx_v_completed;
  PyObject *__pyx_v_err;
  int __pyx_v_family;
  int __pyx_v_flags;
  PyObject *__pyx_v_fs;
  PyObject *__pyx_7genexpr__pyx_v_host;
  PyObject *__pyx_v_host;
  PyObject *__pyx_v_hosts;
  PyObject *__pyx_v_info;
  PyObject *__pyx_v_infos;
  PyObject *__pyx_v_port;
  PyObject *__pyx_v_protocol_factory;
  PyObject *__pyx_v_pyaddr;
  PyObject *__pyx_v_reuse_address;
  PyObject *__pyx_v_reuse_port;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server;
  PyObject *__pyx_v_sock;
  PyObject *__pyx_v_ssl;
  PyObject *__pyx_v_ssl_handshake_timeout;
  PyObject *__pyx_v_ssl_shutdown_timeout;
  PyObject *__pyx_v_start_serving;
  struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_tcp;
};


/* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection {
  PyObject_HEAD
  PyObject *__pyx_v_addr;
  struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_ai_local;
  struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_ai_remote;
  PyObject *__pyx_v_app_protocol;
  PyObject *__pyx_v_app_transport;
  PyObject *__pyx_v_context;
  PyObject *__pyx_v_exc;
  PyObject *__pyx_v_exceptions;
  PyObject *__pyx_v_f1;
  PyObject *__pyx_v_f2;
  PyObject *__pyx_v_family;
  PyObject *__pyx_v_flags;
  PyObject *__pyx_v_fs;
  PyObject *__pyx_v_genexpr;
  PyObject *__pyx_v_host;
  struct addrinfo *__pyx_v_lai;
  struct sockaddr_storage __pyx_v_lai_addr_static;
  struct addrinfo *__pyx_v_lai_iter;
  struct addrinfo __pyx_v_lai_static;
  PyObject *__pyx_v_local_addr;
  PyObject *__pyx_v_model;
  PyObject *__pyx_v_port;
  PyObject *__pyx_v_proto;
  PyObject *__pyx_v_protocol;
  PyObject *__pyx_v_protocol_factory;
  struct addrinfo *__pyx_v_rai;
  struct sockaddr_storage __pyx_v_rai_addr_static;
  struct addrinfo *__pyx_v_rai_iter;
  struct addrinfo __pyx_v_rai_static;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_server_hostname;
  PyObject *__pyx_v_sock;
  PyObject *__pyx_v_ssl;
  PyObject *__pyx_v_ssl_handshake_timeout;
  PyObject *__pyx_v_ssl_shutdown_timeout;
  PyObject *__pyx_v_ssl_waiter;
  PyObject *__pyx_v_sslcontext;
  struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_tr;
  PyObject *__pyx_v_waiter;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "uvloop/loop.pyx":2042
 *                 # If they all have the same str(), raise one.
 *                 model = str(exceptions[0])
 *                 if all(str(exc) == model for exc in exceptions):             # <<<<<<<<<<<<<<
 *                     raise exceptions[0]
 *                 # Raise a combined exception so the user can see all
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_exc;
};


/* "uvloop/loop.pyx":2047
 *                 # the various error messages.
 *                 raise OSError('Multiple exceptions: {}'.format(
 *                     ', '.join(str(exc) for exc in exceptions)))             # <<<<<<<<<<<<<<
 *         else:
 *             if sock is None:
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_exc;
};


/* "uvloop/loop.pyx":2093
 *             return tr, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server {
  PyObject_HEAD
  PyObject *__pyx_v_backlog;
  PyObject *__pyx_v_cleanup_socket;
  PyObject *__pyx_v_err;
  PyObject *__pyx_v_exc;
  PyObject *__pyx_v_msg;
  PyObject *__pyx_v_orig_path;
  PyObject *__pyx_v_path;
  struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_pipe;
  PyObject *__pyx_v_protocol_factory;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server;
  PyObject *__pyx_v_sock;
  PyObject *__pyx_v_ssl;
  PyObject *__pyx_v_ssl_handshake_timeout;
  PyObject *__pyx_v_ssl_shutdown_timeout;
  PyObject *__pyx_v_start_serving;
};


/* "uvloop/loop.pyx":2235
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection {
  PyObject_HEAD
  PyObject *__pyx_v_app_protocol;
  PyObject *__pyx_v_app_transport;
  PyObject *__pyx_v_context;
  PyObject *__pyx_v_path;
  PyObject *__pyx_v_protocol;
  PyObject *__pyx_v_protocol_factory;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_server_hostname;
  PyObject *__pyx_v_sock;
  PyObject *__pyx_v_ssl;
  PyObject *__pyx_v_ssl_handshake_timeout;
  PyObject *__pyx_v_ssl_shutdown_timeout;
  PyObject *__pyx_v_ssl_waiter;
  PyObject *__pyx_v_sslcontext;
  struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_tr;
  PyObject *__pyx_v_waiter;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "uvloop/loop.pyx":2475
 *         self._remove_writer(fileobj)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv(self, sock, n):
 *         """Receive data from the socket.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv {
  PyObject_HEAD
  PyObject *__pyx_v_fut;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle;
  PyObject *__pyx_v_n;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2503
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv_into(self, sock, buf):
 *         """Receive data from the socket.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into {
  PyObject_HEAD
  PyObject *__pyx_v_buf;
  PyObject *__pyx_v_fut;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2530
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendall(self, sock, data):
 *         """Send data to the socket.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall {
  PyObject_HEAD
  PyObject *__pyx_v_data;
  PyObject *__pyx_v_fut;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle;
  Py_ssize_t __pyx_v_n;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2584
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_accept(self, sock):
 *         """Accept a connection.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept {
  PyObject_HEAD
  PyObject *__pyx_v_fut;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2613
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_connect(self, sock, address):
 *         """Connect to a remote socket at address.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect {
  PyObject_HEAD
  PyObject *__pyx_v__;
  PyObject *__pyx_v_address;
  PyObject *__pyx_v_addrs;
  PyObject *__pyx_v_fut;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2637
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom(self, sock, bufsize):
 *         raise NotImplementedError
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom {
  PyObject_HEAD
  PyObject *__pyx_v_bufsize;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2641
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom_into(self, sock, buf, nbytes=0):
 *         raise NotImplementedError
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into {
  PyObject_HEAD
  PyObject *__pyx_v_buf;
  PyObject *__pyx_v_nbytes;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2645
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendto(self, sock, data, address):
 *         raise NotImplementedError
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto {
  PyObject_HEAD
  PyObject *__pyx_v_address;
  PyObject *__pyx_v_data;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
};


/* "uvloop/loop.pyx":2649
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket {
  PyObject_HEAD
  PyObject *__pyx_v_app_protocol;
  PyObject *__pyx_v_app_transport;
  PyObject *__pyx_v_context;
  PyObject *__pyx_v_protocol;
  PyObject *__pyx_v_protocol_factory;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
  PyObject *__pyx_v_ssl;
  PyObject *__pyx_v_ssl_handshake_timeout;
  PyObject *__pyx_v_ssl_shutdown_timeout;
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_transport;
  PyObject *__pyx_v_transport_waiter;
  PyObject *__pyx_v_waiter;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "uvloop/loop.pyx":2752
 *         self._default_executor = executor
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run {
  PyObject_HEAD
  PyObject *__pyx_v_args;
  PyObject *__pyx_v_bufsize;
  PyObject *__pyx_v_close_fds;
  PyObject *__pyx_v_creationflags;
  PyObject *__pyx_v_cwd;
  int __pyx_v_debug_flags;
  PyObject *__pyx_v_env;
  PyObject *__pyx_v_executable;
  PyObject *__pyx_v_kwargs;
  PyObject *__pyx_v_pass_fds;
  PyObject *__pyx_v_preexec_fn;
  struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_proc;
  PyObject *__pyx_v_protocol;
  PyObject *__pyx_v_protocol_factory;
  PyObject *__pyx_v_restore_signals;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_shell;
  PyObject *__pyx_v_start_new_session;
  PyObject *__pyx_v_startupinfo;
  PyObject *__pyx_v_stderr;
  PyObject *__pyx_v_stdin;
  PyObject *__pyx_v_stdout;
  PyObject *__pyx_v_universal_newlines;
  PyObject *__pyx_v_waiter;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "uvloop/loop.pyx":2817
 *         return proc, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell {
  PyObject_HEAD
  PyObject *__pyx_v_args;
  PyObject *__pyx_v_cmd;
  PyObject *__pyx_v_kwargs;
  PyObject *__pyx_v_protocol_factory;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_shell;
};


/* "uvloop/loop.pyx":2832
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec {
  PyObject_HEAD
  PyObject *__pyx_v_args;
  PyObject *__pyx_v_kwargs;
  PyObject *__pyx_v_program;
  PyObject *__pyx_v_protocol_factory;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_shell;
};


/* "uvloop/loop.pyx":2844
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_read_pipe(self, proto_factory, pipe):
 *         """Register read pipe in event loop. Set the pipe to non-blocking mode.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe {
  PyObject_HEAD
  PyObject *__pyx_v_pipe;
  PyObject *__pyx_v_proto;
  PyObject *__pyx_v_proto_factory;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_transp;
  PyObject *__pyx_v_waiter;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "uvloop/loop.pyx":2871
 *         return transp, proto
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_write_pipe(self, proto_factory, pipe):
 *         """Register write pipe in event loop.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe {
  PyObject_HEAD
  PyObject *__pyx_v_pipe;
  PyObject *__pyx_v_proto;
  PyObject *__pyx_v_proto_factory;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_transp;
  PyObject *__pyx_v_waiter;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint {
  PyObject_HEAD
  PyObject *__pyx_v_allow_broadcast;
  PyObject *__pyx_v_ctx;
  PyObject *__pyx_v_ex;
  PyObject *__pyx_v_excs;
  PyObject *__pyx_v_family;
  PyObject *__pyx_v_flags;
  PyObject *__pyx_v_genexpr;
  PyObject *__pyx_v_lads;
  struct addrinfo *__pyx_v_lai;
  PyObject *__pyx_v_local_addr;
  PyObject *__pyx_v_opts;
  PyObject *__pyx_v_problems;
  PyObject *__pyx_v_proto;
  PyObject *__pyx_v_protocol;
  PyObject *__pyx_v_protocol_factory;
  PyObject *__pyx_v_rads;
  struct addrinfo *__pyx_v_rai;
  PyObject *__pyx_v_remote_addr;
  PyObject *__pyx_v_reuse_address;
  PyObject *__pyx_v_reuse_port;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_sock;
  struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_udp;
  PyObject *__pyx_v_waiter;
};


/* "uvloop/loop.pyx":3049
 *                             allow_broadcast=allow_broadcast)
 *                 problems = ', '.join(
 *                     '{}={}'.format(k, v) for k, v in opts.items() if v)             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'socket modifier keyword arguments can not be used '
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
};


/* "uvloop/loop.pyx":3198
 *         self._asyncgens.add(agen)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_asyncgens(self):
 *         """Shutdown all active asynchronous generators."""
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens {
  PyObject_HEAD
  PyObject *__pyx_8genexpr4__pyx_v_ag;
  PyObject *__pyx_v_agen;
  PyObject *__pyx_v_closing_agens;
  PyObject *__pyx_v_result;
  PyObject *__pyx_v_results;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_shutdown_coro;
};


/* "uvloop/loop.pyx":3223
 *                 })
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_default_executor(self, timeout=None):
 *         """Schedule the shutdown of the default executor.
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor {
  PyObject_HEAD
  PyObject *__pyx_v_future;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self;
  PyObject *__pyx_v_thread;
  PyObject *__pyx_v_timeout;
};


/* "uvloop/sslproto.pyx":462
 *     # Handshake flow
 * 
 *     cdef _start_handshake(self):             # <<<<<<<<<<<<<<
 *         if self._loop.get_debug():
 *             aio_logger.debug("%r starts SSL handshake", self)
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self;
};


/* "uvloop/sslproto.pyx":559
 *     # Shutdown flow
 * 
 *     cdef _start_shutdown(self, object context=None):             # <<<<<<<<<<<<<<
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             return
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self;
};


/* "uvloop/server.pyx":51
 *     # Public API
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aenter__(self):
 *         return self
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self;
};


/* "uvloop/server.pyx":55
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aexit__(self, *exc):
 *         self.close()
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ {
  PyObject_HEAD
  PyObject *__pyx_v_exc;
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self;
};


/* "uvloop/server.pyx":66
 *         return self._loop
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def wait_closed(self):
 *         # Do not remove `self._servers is None` below
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self;
  PyObject *__pyx_v_waiter;
};


/* "uvloop/server.pyx":99
 *         return self._serving
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_serving(self):
 *         self._start_serving()
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self;
};


/* "uvloop/server.pyx":103
 *         self._start_serving()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever {
  PyObject_HEAD
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5;
};


/* "uvloop/loop.pyx":3422
 * # Helpers for tests
 * 
 * @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 * async def _test_coroutine_1():
 *     return 42
 */
struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 {
  PyObject_HEAD
};


/* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled")
 * cdef object __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(object (*f)(Loop, bint) ):             # <<<<<<<<<<<<<<
 *     def wrap(Loop self, bint enabled):
 *         """wrap(self: 'Loop', enabled: bool)"""
 */
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled {
  PyObject_HEAD
  PyObject *(*__pyx_v_f)(struct __pyx_obj_6uvloop_4loop_Loop *, int);
};



/* "uvloop/handles/handle.pyx":1
 * cdef class UVHandle:             # <<<<<<<<<<<<<<
 *     """A base class for all libuv handles.
 * 
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVHandle {
  PyObject *(*_start_init)(struct __pyx_obj_6uvloop_4loop_UVHandle *, struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_abort_init)(struct __pyx_obj_6uvloop_4loop_UVHandle *);
  PyObject *(*_finish_init)(struct __pyx_obj_6uvloop_4loop_UVHandle *);
  int (*_is_alive)(struct __pyx_obj_6uvloop_4loop_UVHandle *);
  PyObject *(*_ensure_alive)(struct __pyx_obj_6uvloop_4loop_UVHandle *);
  PyObject *(*_error)(struct __pyx_obj_6uvloop_4loop_UVHandle *, PyObject *, PyObject *);
  PyObject *(*_fatal_error)(struct __pyx_obj_6uvloop_4loop_UVHandle *, PyObject *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error *__pyx_optional_args);
  PyObject *(*_warn_unclosed)(struct __pyx_obj_6uvloop_4loop_UVHandle *);
  PyObject *(*_free)(struct __pyx_obj_6uvloop_4loop_UVHandle *);
  PyObject *(*_close)(struct __pyx_obj_6uvloop_4loop_UVHandle *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *__pyx_vtabptr_6uvloop_4loop_UVHandle;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__start_init(struct __pyx_obj_6uvloop_4loop_UVHandle *, struct __pyx_obj_6uvloop_4loop_Loop *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__abort_init(struct __pyx_obj_6uvloop_4loop_UVHandle *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__finish_init(struct __pyx_obj_6uvloop_4loop_UVHandle *);
static CYTHON_INLINE int __pyx_f_6uvloop_4loop_8UVHandle__is_alive(struct __pyx_obj_6uvloop_4loop_UVHandle *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__ensure_alive(struct __pyx_obj_6uvloop_4loop_UVHandle *);


/* "uvloop/handles/handle.pyx":221
 * 
 * 
 * cdef class UVSocketHandle(UVHandle):             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_fileno)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *);
  PyObject *(*_new_socket)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *);
  PyObject *(*_get_socket)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *);
  PyObject *(*_attach_fileobj)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, PyObject *);
  PyObject *(*_open)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle *__pyx_vtabptr_6uvloop_4loop_UVSocketHandle;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__get_socket(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__attach_fileobj(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, PyObject *);


/* "uvloop/handles/async_.pyx":2
 * @cython.no_gc_clear
 * cdef class UVAsync(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef _init(self, Loop loop, method_t callback, object ctx):
 *         cdef int err
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVAsync {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVAsync *, struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *);
  PyObject *(*send)(struct __pyx_obj_6uvloop_4loop_UVAsync *);
  struct __pyx_obj_6uvloop_4loop_UVAsync *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *__pyx_vtabptr_6uvloop_4loop_UVAsync;


/* "uvloop/handles/timer.pyx":2
 * @cython.no_gc_clear
 * cdef class UVTimer(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef _init(self, Loop loop, method_t callback, object ctx,
 *                uint64_t timeout):
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVTimer {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVTimer *, struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *, uint64_t);
  PyObject *(*stop)(struct __pyx_obj_6uvloop_4loop_UVTimer *);
  PyObject *(*start)(struct __pyx_obj_6uvloop_4loop_UVTimer *);
  PyObject *(*get_when)(struct __pyx_obj_6uvloop_4loop_UVTimer *);
  struct __pyx_obj_6uvloop_4loop_UVTimer *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *, uint64_t);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *__pyx_vtabptr_6uvloop_4loop_UVTimer;


/* "uvloop/handles/idle.pyx":2
 * @cython.no_gc_clear
 * cdef class UVIdle(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef _init(self, Loop loop, Handle h):
 *         cdef int err
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVIdle {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVIdle *, struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*stop)(struct __pyx_obj_6uvloop_4loop_UVIdle *);
  PyObject *(*start)(struct __pyx_obj_6uvloop_4loop_UVIdle *);
  struct __pyx_obj_6uvloop_4loop_UVIdle *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *__pyx_vtabptr_6uvloop_4loop_UVIdle;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVIdle_stop(struct __pyx_obj_6uvloop_4loop_UVIdle *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVIdle_start(struct __pyx_obj_6uvloop_4loop_UVIdle *);


/* "uvloop/handles/basetransport.pyx":1
 * cdef class UVBaseTransport(UVSocketHandle):             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport {
  struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle __pyx_base;
  PyObject *(*_maybe_pause_protocol)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_maybe_resume_protocol)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_schedule_call_connection_made)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_schedule_call_connection_lost)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *);
  PyObject *(*_wakeup_waiter)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_call_connection_made)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_call_connection_lost)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *);
  PyObject *(*_set_server)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, struct __pyx_obj_6uvloop_4loop_Server *);
  PyObject *(*_set_waiter)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *);
  PyObject *(*_set_protocol)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *);
  PyObject *(*_clear_protocol)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_init_protocol)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_add_extra_info)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *, PyObject *);
  size_t (*_get_write_buffer_size)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  int (*_is_reading)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_start_reading)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_stop_reading)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *__pyx_vtabptr_6uvloop_4loop_UVBaseTransport;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_pause_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_resume_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_made(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_lost(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_server(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, struct __pyx_obj_6uvloop_4loop_Server *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_waiter(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__init_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__add_extra_info(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *, PyObject *);


/* "uvloop/loop.pyx":128
 * 
 * @cython.no_gc_clear
 * cdef class Loop:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         cdef int err
 */

struct __pyx_vtabstruct_6uvloop_4loop_Loop {
  PyObject *(*_init_debug_fields)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_on_wake)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_on_idle)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_Loop__run)(struct __pyx_obj_6uvloop_4loop_Loop *, uv_run_mode);
  PyObject *(*_run)(struct __pyx_obj_6uvloop_4loop_Loop *, uv_run_mode);
  PyObject *(*_close)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_stop)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  uint64_t (*_time)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_queue_write)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_exec_queued_writes)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_call_soon)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_append_ready_handle)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*_call_soon_handle)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*_call_later)(struct __pyx_obj_6uvloop_4loop_Loop *, uint64_t, PyObject *, PyObject *, PyObject *);
  void (*_handle_exception)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_is_main_thread)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_new_future)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_check_signal)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_check_closed)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_check_thread)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_getaddrinfo)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, int, int, int, int, int);
  PyObject *(*_getnameinfo)(struct __pyx_obj_6uvloop_4loop_Loop *, struct sockaddr *, int);
  PyObject *(*_track_transport)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVBaseTransport *);
  PyObject *(*_fileobj_to_fd)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_ensure_fd_no_transport)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_track_process)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVProcess *);
  PyObject *(*_untrack_process)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVProcess *);
  PyObject *(*_add_reader)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*_has_reader)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_remove_reader)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_add_writer)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*_has_writer)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_remove_writer)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_sock_recv)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_sock_recv_into)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_sock_sendall)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_sock_accept)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *);
  PyObject *(*_sock_connect)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *);
  PyObject *(*_sock_connect_cb)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_sock_set_reuseport)(struct __pyx_obj_6uvloop_4loop_Loop *, int);
  PyObject *(*_setup_or_resume_signals)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_shutdown_signals)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_pause_signals)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_handle_signal)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_read_from_self)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_ceval_process_signals)(struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_invoke_signals)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
  PyObject *(*_set_coroutine_debug)(struct __pyx_obj_6uvloop_4loop_Loop *, int);
  PyObject *(*_print_debug_info)(struct __pyx_obj_6uvloop_4loop_Loop *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_Loop *__pyx_vtabptr_6uvloop_4loop_Loop;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__queue_write(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__call_soon(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__append_ready_handle(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__call_soon_handle(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__is_main_thread(struct __pyx_obj_6uvloop_4loop_Loop *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__new_future(struct __pyx_obj_6uvloop_4loop_Loop *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_signal(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_closed(struct __pyx_obj_6uvloop_4loop_Loop *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_thread(struct __pyx_obj_6uvloop_4loop_Loop *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__ceval_process_signals(struct __pyx_obj_6uvloop_4loop_Loop *);


/* "uvloop/cbhandles.pyx":3
 * @cython.no_gc_clear
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class Handle:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self._cancelled = 0
 */

struct __pyx_vtabstruct_6uvloop_4loop_Handle {
  PyObject *(*_set_loop)(struct __pyx_obj_6uvloop_4loop_Handle *, struct __pyx_obj_6uvloop_4loop_Loop *);
  PyObject *(*_set_context)(struct __pyx_obj_6uvloop_4loop_Handle *, PyObject *);
  PyObject *(*_run)(struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*_cancel)(struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*_format_handle)(struct __pyx_obj_6uvloop_4loop_Handle *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_Handle *__pyx_vtabptr_6uvloop_4loop_Handle;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__set_loop(struct __pyx_obj_6uvloop_4loop_Handle *, struct __pyx_obj_6uvloop_4loop_Loop *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__set_context(struct __pyx_obj_6uvloop_4loop_Handle *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__run(struct __pyx_obj_6uvloop_4loop_Handle *);


/* "uvloop/cbhandles.pyx":167
 * @cython.no_gc_clear
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class TimerHandle:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Loop loop, object callback, object args,
 *                   uint64_t delay, object context):
 */

struct __pyx_vtabstruct_6uvloop_4loop_TimerHandle {
  PyObject *(*_run)(struct __pyx_obj_6uvloop_4loop_TimerHandle *);
  PyObject *(*_cancel)(struct __pyx_obj_6uvloop_4loop_TimerHandle *);
  PyObject *(*_clear)(struct __pyx_obj_6uvloop_4loop_TimerHandle *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_TimerHandle *__pyx_vtabptr_6uvloop_4loop_TimerHandle;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_11TimerHandle__clear(struct __pyx_obj_6uvloop_4loop_TimerHandle *);


/* "uvloop/handles/check.pyx":2
 * @cython.no_gc_clear
 * cdef class UVCheck(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef _init(self, Loop loop, Handle h):
 *         cdef int err
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVCheck {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVCheck *, struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*stop)(struct __pyx_obj_6uvloop_4loop_UVCheck *);
  PyObject *(*start)(struct __pyx_obj_6uvloop_4loop_UVCheck *);
  struct __pyx_obj_6uvloop_4loop_UVCheck *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *__pyx_vtabptr_6uvloop_4loop_UVCheck;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_7UVCheck_stop(struct __pyx_obj_6uvloop_4loop_UVCheck *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_7UVCheck_start(struct __pyx_obj_6uvloop_4loop_UVCheck *);


/* "uvloop/handles/poll.pyx":2
 * @cython.no_gc_clear
 * cdef class UVPoll(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef _init(self, Loop loop, int fd):
 *         cdef int err
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVPoll {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVPoll *, struct __pyx_obj_6uvloop_4loop_Loop *, int);
  PyObject *(*_poll_start)(struct __pyx_obj_6uvloop_4loop_UVPoll *, int);
  PyObject *(*_poll_stop)(struct __pyx_obj_6uvloop_4loop_UVPoll *);
  int (*is_active)(struct __pyx_obj_6uvloop_4loop_UVPoll *);
  PyObject *(*is_reading)(struct __pyx_obj_6uvloop_4loop_UVPoll *);
  PyObject *(*is_writing)(struct __pyx_obj_6uvloop_4loop_UVPoll *);
  PyObject *(*start_reading)(struct __pyx_obj_6uvloop_4loop_UVPoll *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*start_writing)(struct __pyx_obj_6uvloop_4loop_UVPoll *, struct __pyx_obj_6uvloop_4loop_Handle *);
  PyObject *(*stop_reading)(struct __pyx_obj_6uvloop_4loop_UVPoll *);
  PyObject *(*stop_writing)(struct __pyx_obj_6uvloop_4loop_UVPoll *);
  PyObject *(*stop)(struct __pyx_obj_6uvloop_4loop_UVPoll *);
  struct __pyx_obj_6uvloop_4loop_UVPoll *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, int);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *__pyx_vtabptr_6uvloop_4loop_UVPoll;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVPoll__poll_start(struct __pyx_obj_6uvloop_4loop_UVPoll *, int);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVPoll__poll_stop(struct __pyx_obj_6uvloop_4loop_UVPoll *);


/* "uvloop/handles/stream.pyx":213
 * 
 * @cython.no_gc_clear
 * cdef class UVStream(UVBaseTransport):             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVStream {
  struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *);
  PyObject *(*_shutdown)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_accept)(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_close_on_read_error)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_UVStream__reading_started)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_UVStream__reading_stopped)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_buffer_write)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *);
  PyObject *(*_initiate_write)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_exec_write)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_try_write)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *);
  PyObject *(*_on_accept)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_on_eof)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_on_write)(struct __pyx_obj_6uvloop_4loop_UVStream *);
  PyObject *(*_on_connect)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVStream *__pyx_vtabptr_6uvloop_4loop_UVStream;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__init(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__shutdown(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__accept(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__close_on_read_error(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_started(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_stopped(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__buffer_write(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__initiate_write(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__exec_write(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__try_write(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_accept(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_eof(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_write(struct __pyx_obj_6uvloop_4loop_UVStream *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_connect(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *);


/* "uvloop/handles/streamserver.pyx":2
 * @cython.no_gc_clear
 * cdef class UVStreamServer(UVSocketHandle):             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer {
  struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_mark_as_open)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *);
  PyObject *(*listen)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *);
  PyObject *(*_on_listen)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *);
  struct __pyx_obj_6uvloop_4loop_UVStream *(*_make_new_transport)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *, PyObject *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *__pyx_vtabptr_6uvloop_4loop_UVStreamServer;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__init(struct __pyx_obj_6uvloop_4loop_UVStreamServer *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__mark_as_open(struct __pyx_obj_6uvloop_4loop_UVStreamServer *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer_listen(struct __pyx_obj_6uvloop_4loop_UVStreamServer *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__on_listen(struct __pyx_obj_6uvloop_4loop_UVStreamServer *);


/* "uvloop/handles/tcp.pyx":56
 * 
 * @cython.no_gc_clear
 * cdef class TCPServer(UVStreamServer):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6uvloop_4loop_TCPServer {
  struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer __pyx_base;
  PyObject *(*bind)(struct __pyx_obj_6uvloop_4loop_TCPServer *, struct sockaddr *, struct __pyx_opt_args_6uvloop_4loop_9TCPServer_bind *__pyx_optional_args);
  struct __pyx_obj_6uvloop_4loop_TCPServer *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, unsigned int, PyObject *, PyObject *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *__pyx_vtabptr_6uvloop_4loop_TCPServer;


/* "uvloop/handles/tcp.pyx":103
 * 
 * @cython.no_gc_clear
 * cdef class TCPTransport(UVStream):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport {
  struct __pyx_vtabstruct_6uvloop_4loop_UVStream __pyx_base;
  PyObject *(*bind)(struct __pyx_obj_6uvloop_4loop_TCPTransport *, struct sockaddr *, struct __pyx_opt_args_6uvloop_4loop_12TCPTransport_bind *__pyx_optional_args);
  PyObject *(*connect)(struct __pyx_obj_6uvloop_4loop_TCPTransport *, struct sockaddr *);
  PyObject *(*_set_nodelay)(struct __pyx_obj_6uvloop_4loop_TCPTransport *);
  struct __pyx_obj_6uvloop_4loop_TCPTransport *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *__pyx_vtabptr_6uvloop_4loop_TCPTransport;


/* "uvloop/handles/pipe.pyx":40
 * 
 * @cython.no_gc_clear
 * cdef class UnixServer(UVStreamServer):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6uvloop_4loop_UnixServer {
  struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer __pyx_base;
  PyObject *(*bind)(struct __pyx_obj_6uvloop_4loop_UnixServer *, PyObject *);
  struct __pyx_obj_6uvloop_4loop_UnixServer *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *__pyx_vtabptr_6uvloop_4loop_UnixServer;


/* "uvloop/handles/pipe.pyx":106
 * 
 * @cython.no_gc_clear
 * cdef class UnixTransport(UVStream):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport {
  struct __pyx_vtabstruct_6uvloop_4loop_UVStream __pyx_base;
  struct __pyx_obj_6uvloop_4loop_UnixTransport *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *);
  PyObject *(*connect)(struct __pyx_obj_6uvloop_4loop_UnixTransport *, char *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *__pyx_vtabptr_6uvloop_4loop_UnixTransport;


/* "uvloop/handles/pipe.pyx":131
 * 
 * @cython.no_gc_clear
 * cdef class ReadUnixTransport(UVStream):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport {
  struct __pyx_vtabstruct_6uvloop_4loop_UVStream __pyx_base;
  struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *__pyx_vtabptr_6uvloop_4loop_ReadUnixTransport;


/* "uvloop/handles/pipe.pyx":176
 * 
 * @cython.no_gc_clear
 * cdef class WriteUnixTransport(UVStream):             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport {
  struct __pyx_vtabstruct_6uvloop_4loop_UVStream __pyx_base;
  struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *__pyx_vtabptr_6uvloop_4loop_WriteUnixTransport;


/* "uvloop/handles/process.pyx":2
 * @cython.no_gc_clear
 * cdef class UVProcess(UVHandle):             # <<<<<<<<<<<<<<
 *     """Abstract class; wrapper over uv_process_t handle."""
 * 
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVProcess {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_close_process_handle)(struct __pyx_obj_6uvloop_4loop_UVProcess *);
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVProcess *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_after_fork)(struct __pyx_obj_6uvloop_4loop_UVProcess *);
  char **(*_UVProcess__to_cstring_array)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *);
  PyObject *(*_init_args)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *);
  PyObject *(*_init_env)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *);
  PyObject *(*_init_files)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_init_options)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, int);
  PyObject *(*_close_after_spawn)(struct __pyx_obj_6uvloop_4loop_UVProcess *, int);
  PyObject *(*_on_exit)(struct __pyx_obj_6uvloop_4loop_UVProcess *, int64_t, int);
  PyObject *(*_kill)(struct __pyx_obj_6uvloop_4loop_UVProcess *, int);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *__pyx_vtabptr_6uvloop_4loop_UVProcess;


/* "uvloop/handles/process.pyx":367
 * 
 * @cython.no_gc_clear
 * cdef class UVProcessTransport(UVProcess):             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self._exit_waiters = []
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport {
  struct __pyx_vtabstruct_6uvloop_4loop_UVProcess __pyx_base;
  PyObject *(*_file_redirect_stdio)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, int);
  PyObject *(*_file_devnull)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *);
  PyObject *(*_file_inpipe)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *);
  PyObject *(*_file_outpipe)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *);
  PyObject *(*_check_proc)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *);
  PyObject *(*_pipe_connection_lost)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, int, PyObject *);
  PyObject *(*_pipe_data_received)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, int, PyObject *);
  PyObject *(*_call_connection_made)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, PyObject *);
  PyObject *(*_try_finish)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *);
  struct __pyx_obj_6uvloop_4loop_UVProcessTransport *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *__pyx_vtabptr_6uvloop_4loop_UVProcessTransport;


/* "uvloop/handles/fsevent.pyx":11
 * 
 * @cython.no_gc_clear
 * cdef class UVFSEvent(UVHandle):             # <<<<<<<<<<<<<<
 *     cdef _init(self, Loop loop, object callback, object context):
 *         cdef int err
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent {
  struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UVFSEvent *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *);
  PyObject *(*start)(struct __pyx_obj_6uvloop_4loop_UVFSEvent *, char *, int);
  PyObject *(*stop)(struct __pyx_obj_6uvloop_4loop_UVFSEvent *);
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *(*new)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *__pyx_vtabptr_6uvloop_4loop_UVFSEvent;


/* "uvloop/request.pyx":1
 * cdef class UVRequest:             # <<<<<<<<<<<<<<
 *     """A base class for all libuv requests (uv_getaddrinfo_t, etc).
 * 
 */

struct __pyx_vtabstruct_6uvloop_4loop_UVRequest {
  PyObject *(*on_done)(struct __pyx_obj_6uvloop_4loop_UVRequest *);
  PyObject *(*cancel)(struct __pyx_obj_6uvloop_4loop_UVRequest *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UVRequest *__pyx_vtabptr_6uvloop_4loop_UVRequest;


/* "uvloop/sslproto.pyx":195
 * 
 * 
 * cdef class SSLProtocol:             # <<<<<<<<<<<<<<
 *     """SSL protocol.
 * 
 */

struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol {
  PyObject *(*_set_app_protocol)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *);
  PyObject *(*_wakeup_waiter)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter *__pyx_optional_args);
  PyObject *(*_get_extra_info)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__get_extra_info *__pyx_optional_args);
  PyObject *(*_set_state)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, enum __pyx_t_6uvloop_4loop_SSLProtocolState);
  PyObject *(*_start_handshake)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_check_handshake_timeout)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_do_handshake)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_on_handshake_complete)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *);
  PyObject *(*_start_shutdown)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__start_shutdown *__pyx_optional_args);
  PyObject *(*_check_shutdown_timeout)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_do_read_into_void)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *);
  PyObject *(*_do_flush)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_flush *__pyx_optional_args);
  PyObject *(*_do_shutdown)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_shutdown *__pyx_optional_args);
  PyObject *(*_on_shutdown_complete)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *);
  PyObject *(*_abort)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *);
  PyObject *(*_write_appdata)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *, PyObject *);
  PyObject *(*_do_write)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_process_outgoing)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_do_read)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_do_read__buffered)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_do_read__copied)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_call_eof_received)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__call_eof_received *__pyx_optional_args);
  PyObject *(*_control_app_writing)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing *__pyx_optional_args);
  size_t (*_get_write_buffer_size)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_set_write_buffer_limits)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits *__pyx_optional_args);
  PyObject *(*_pause_reading)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_resume_reading)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *);
  PyObject *(*_control_ssl_reading)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_set_read_buffer_limits)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits *__pyx_optional_args);
  size_t (*_get_read_buffer_size)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *);
  PyObject *(*_fatal_error)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error *__pyx_optional_args);
};
static struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *__pyx_vtabptr_6uvloop_4loop_SSLProtocol;


/* "uvloop/handles/udp.pyx":54
 * 
 * @cython.no_gc_clear
 * cdef class UDPTransport(UVBaseTransport):             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self._family = uv.AF_UNSPEC
 */

struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport {
  struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport __pyx_base;
  PyObject *(*_init)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct __pyx_obj_6uvloop_4loop_Loop *, unsigned int);
  PyObject *(*_set_address)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct addrinfo *);
  PyObject *(*_connect)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct sockaddr *, size_t);
  PyObject *(*_bind)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct sockaddr *);
  PyObject *(*open)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, int, int);
  PyObject *(*_set_broadcast)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, int);
  PyObject *(*_UDPTransport__receiving_started)(struct __pyx_obj_6uvloop_4loop_UDPTransport *);
  PyObject *(*_UDPTransport__receiving_stopped)(struct __pyx_obj_6uvloop_4loop_UDPTransport *);
  PyObject *(*_send)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *, PyObject *);
  PyObject *(*_on_receive)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_on_sent)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_12UDPTransport__on_sent *__pyx_optional_args);
};
static struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *__pyx_vtabptr_6uvloop_4loop_UDPTransport;
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_started(struct __pyx_obj_6uvloop_4loop_UDPTransport *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_stopped(struct __pyx_obj_6uvloop_4loop_UDPTransport *);


/* "uvloop/server.pyx":4
 * 
 * 
 * cdef class Server:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Loop loop):
 *         self._loop = loop
 */

struct __pyx_vtabstruct_6uvloop_4loop_Server {
  PyObject *(*_add_server)(struct __pyx_obj_6uvloop_4loop_Server *, struct __pyx_obj_6uvloop_4loop_UVStreamServer *);
  PyObject *(*_start_serving)(struct __pyx_obj_6uvloop_4loop_Server *);
  PyObject *(*_wakeup)(struct __pyx_obj_6uvloop_4loop_Server *);
  PyObject *(*_attach)(struct __pyx_obj_6uvloop_4loop_Server *);
  PyObject *(*_detach)(struct __pyx_obj_6uvloop_4loop_Server *);
  PyObject *(*_ref)(struct __pyx_obj_6uvloop_4loop_Server *);
  PyObject *(*_unref)(struct __pyx_obj_6uvloop_4loop_Server *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_Server *__pyx_vtabptr_6uvloop_4loop_Server;


/* "uvloop/pseudosock.pyx":1
 * cdef class PseudoSocket:             # <<<<<<<<<<<<<<
 *     cdef:
 *         int _family
 */

struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket {
  PyObject *(*_na)(struct __pyx_obj_6uvloop_4loop_PseudoSocket *, PyObject *);
  PyObject *(*_make_sock)(struct __pyx_obj_6uvloop_4loop_PseudoSocket *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *__pyx_vtabptr_6uvloop_4loop_PseudoSocket;


/* "uvloop/lru.pyx":5
 * 
 * @cython.final
 * cdef class LruCache:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_6uvloop_4loop_LruCache {
  PyObject *(*get)(struct __pyx_obj_6uvloop_4loop_LruCache *, PyObject *, PyObject *);
  PyObject *(*needs_cleanup)(struct __pyx_obj_6uvloop_4loop_LruCache *);
  PyObject *(*cleanup_one)(struct __pyx_obj_6uvloop_4loop_LruCache *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_LruCache *__pyx_vtabptr_6uvloop_4loop_LruCache;
static PyObject *__pyx_f_6uvloop_4loop_8LruCache_get(struct __pyx_obj_6uvloop_4loop_LruCache *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8LruCache_needs_cleanup(struct __pyx_obj_6uvloop_4loop_LruCache *);
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8LruCache_cleanup_one(struct __pyx_obj_6uvloop_4loop_LruCache *);


/* "uvloop/handles/stream.pyx":10
 * @cython.no_gc_clear
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class _StreamWriteContext:             # <<<<<<<<<<<<<<
 *     # used to hold additional write request information for uv_write
 * 
 */

struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext {
  PyObject *(*free_bufs)(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *);
  PyObject *(*close)(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *);
  PyObject *(*advance_uv_buf)(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *, size_t);
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *(*new)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *__pyx_vtabptr_6uvloop_4loop__StreamWriteContext;


/* "uvloop/handles/tcp.pyx":185
 * 
 * 
 * cdef class _TCPConnectRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         TCPTransport transport
 */

struct __pyx_vtabstruct_6uvloop_4loop__TCPConnectRequest {
  struct __pyx_vtabstruct_6uvloop_4loop_UVRequest __pyx_base;
  PyObject *(*connect)(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *, struct sockaddr *);
};
static struct __pyx_vtabstruct_6uvloop_4loop__TCPConnectRequest *__pyx_vtabptr_6uvloop_4loop__TCPConnectRequest;


/* "uvloop/handles/pipe.pyx":209
 * 
 * 
 * cdef class _PipeConnectRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         UnixTransport transport
 */

struct __pyx_vtabstruct_6uvloop_4loop__PipeConnectRequest {
  struct __pyx_vtabstruct_6uvloop_4loop_UVRequest __pyx_base;
  PyObject *(*connect)(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *, char *);
};
static struct __pyx_vtabstruct_6uvloop_4loop__PipeConnectRequest *__pyx_vtabptr_6uvloop_4loop__PipeConnectRequest;


/* "uvloop/dns.pyx":289
 * 
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class AddrInfo:             # <<<<<<<<<<<<<<
 *     cdef:
 *         system.addrinfo *data
 */

struct __pyx_vtabstruct_6uvloop_4loop_AddrInfo {
  void (*set_data)(struct __pyx_obj_6uvloop_4loop_AddrInfo *, struct addrinfo *);
  PyObject *(*unpack)(struct __pyx_obj_6uvloop_4loop_AddrInfo *);
  int (*isinstance)(PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop_AddrInfo *__pyx_vtabptr_6uvloop_4loop_AddrInfo;


/* "uvloop/dns.pyx":333
 * 
 * 
 * cdef class AddrInfoRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         system.addrinfo hints
 */

struct __pyx_vtabstruct_6uvloop_4loop_AddrInfoRequest {
  struct __pyx_vtabstruct_6uvloop_4loop_UVRequest __pyx_base;
};
static struct __pyx_vtabstruct_6uvloop_4loop_AddrInfoRequest *__pyx_vtabptr_6uvloop_4loop_AddrInfoRequest;


/* "uvloop/dns.pyx":396
 * 
 * 
 * cdef class NameInfoRequest(UVRequest):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object callback
 */

struct __pyx_vtabstruct_6uvloop_4loop_NameInfoRequest {
  struct __pyx_vtabstruct_6uvloop_4loop_UVRequest __pyx_base;
  PyObject *(*query)(struct __pyx_obj_6uvloop_4loop_NameInfoRequest *, struct sockaddr *, int);
};
static struct __pyx_vtabstruct_6uvloop_4loop_NameInfoRequest *__pyx_vtabptr_6uvloop_4loop_NameInfoRequest;


/* "uvloop/handles/udp.pyx":3
 * @cython.no_gc_clear
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class _UDPSendContext:             # <<<<<<<<<<<<<<
 *     # used to hold additional write request information for uv_write
 * 
 */

struct __pyx_vtabstruct_6uvloop_4loop__UDPSendContext {
  PyObject *(*close)(struct __pyx_obj_6uvloop_4loop__UDPSendContext *);
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *(*new)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *);
};
static struct __pyx_vtabstruct_6uvloop_4loop__UDPSendContext *__pyx_vtabptr_6uvloop_4loop__UDPSendContext;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AndObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAnd(op1, op2) : PyNumber_And(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* py_set_discard.proto */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* py_dict_clear.proto */
#define __Pyx_PyDict_Clear(d) (PyDict_Clear(d), 0)

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* RaiseClosureNameError.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_MultiplyObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* PyMethodNew2Arg.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyMethod_New2Arg PyMethod_New
#else
#define __Pyx_PyMethod_New2Arg(func, self) PyMethod_New(func, self, (PyObject*)Py_TYPE(self))
#endif

/* CallableCheck.proto */
#if CYTHON_USE_TYPE_SLOTS && PY_MAJOR_VERSION >= 3
#define __Pyx_PyCallable_Check(obj)   (Py_TYPE(obj)->tp_call != NULL)
#else
#define __Pyx_PyCallable_Check(obj)   PyCallable_Check(obj)
#endif

/* ReturnWithStopIteration.proto */
#define __Pyx_ReturnWithStopIteration(value)\
    if (value == Py_None) PyErr_SetNone(PyExc_StopIteration); else __Pyx__ReturnWithStopIteration(value)
static void __Pyx__ReturnWithStopIteration(PyObject* value);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Coroutine.proto */
#define __Pyx_Coroutine_USED
#define __Pyx_Coroutine_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineType)
#define __Pyx_Coroutine_Check(obj) __Pyx_Coroutine_CheckExact(obj)
#define __Pyx_CoroutineAwait_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineAwaitType)
#define __Pyx_Coroutine_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_CoroutineType, body, code, closure, name, qualname, module_name)
static int __pyx_Coroutine_init(PyObject *module);
static PyObject *__Pyx__Coroutine_await(PyObject *coroutine);
typedef struct {
    PyObject_HEAD
    PyObject *coroutine;
} __pyx_CoroutineAwaitObject;
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg);
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args);

/* GetAwaitIter.proto */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o);
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *o);

/* CoroutineYieldFrom.proto */
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* MergeKeywords.proto */
static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right)
    #endif
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_ConcatInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_ConcatInPlaceSafe(left, right) ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_ConcatInPlace(left, right))

/* py_set_remove.proto */
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key);

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* RaiseKeywordRequired.proto */
static void __Pyx_RaiseKeywordRequired(const char* func_name, PyObject* kw_name);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(string);
    size = PyBytes_GET_SIZE(string);
#else
    if (PyBytes_AsStringAndSize(string, &as_c_string, &size) < 0) {
        return NULL;
    }
#endif
    return __Pyx_decode_c_bytes(
        as_c_string, size,
        start, stop, encoding, errors, decode_func);
}

/* DelItemInt.proto */
#define __Pyx_DelItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_DelItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_DelItem_Generic(o, to_py_func(i))))
static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j);
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               int is_list, int wraparound);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_11
#define __PYX_HAVE_RT_ImportType_proto_3_0_11
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_11 {
   __Pyx_ImportType_CheckSize_Error_3_0_11 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_11 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_11 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* PatchAsyncIO.proto */
static PyObject* __Pyx_patch_asyncio(PyObject* module);

/* pybytes_as_double.proto */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From___pyx_anon_enum(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__uv_fs_event(enum uv_fs_event value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uv_os_fd_t(uv_os_fd_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_short(unsigned short value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_volatile__int(volatile int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_6uvloop_4loop_SSLProtocolState(enum __pyx_t_6uvloop_4loop_SSLProtocolState value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* IterableCoroutine.proto */
#define __Pyx_IterableCoroutine_USED
#undef __Pyx_Coroutine_Check
#define __Pyx_Coroutine_Check(obj) (__Pyx_Coroutine_CheckExact(obj) || __Pyx_IS_TYPE(obj, __pyx_IterableCoroutineType))
#define __Pyx_IterableCoroutine_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_IterableCoroutineType, body, code, closure, name, qualname, module_name)
static int __pyx_IterableCoroutine_init(PyObject *module);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__is_main_thread(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__init_debug_fields(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__setup_or_resume_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__pause_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__shutdown_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__ceval_process_signals(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__read_from_self(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__invoke_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__handle_signal(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__on_wake(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__on_idle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__stop(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_exc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__Loop__run(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, uv_run_mode __pyx_v_mode); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__run(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, uv_run_mode __pyx_v_mode); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__close(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static uint64_t __pyx_f_6uvloop_4loop_4Loop__time(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__queue_write(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_stream); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__exec_queued_writes(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__call_soon(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, PyObject *__pyx_v_context); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__append_ready_handle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__call_soon_handle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__call_later(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, uint64_t __pyx_v_delay, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, PyObject *__pyx_v_context); /* proto*/
static void __pyx_f_6uvloop_4loop_4Loop__handle_exception(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_ex); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_signal(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_closed(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_thread(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__new_future(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__track_transport(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_transport); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__track_process(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_proc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__untrack_process(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_proc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__fileobj_to_fd(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__ensure_fd_no_transport(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__add_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__remove_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__has_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__add_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__remove_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__has_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__getaddrinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_flags, int __pyx_v_unpack); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__getnameinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct sockaddr *__pyx_v_addr, int __pyx_v_flags); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_recv(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_n); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_recv_into(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_buf); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_sendall(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_data); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_accept(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_connect(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_address); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_connect_cb(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_address); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_set_reuseport(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, int __pyx_v_fd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__set_coroutine_debug(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, int __pyx_v_enabled); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_4Loop__print_debug_info(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__set_loop(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__set_context(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self, PyObject *__pyx_v_context); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__run(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Handle__cancel(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Handle__format_handle(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11TimerHandle__cancel(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_11TimerHandle__clear(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11TimerHandle__run(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12PseudoSocket__na(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_what); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12PseudoSocket__make_sock(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8LruCache_get(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8LruCache_needs_cleanup(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8LruCache_cleanup_one(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__free(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__warn_unclosed(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__abort_init(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__finish_init(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__start_init(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop); /* proto*/
static CYTHON_INLINE int __pyx_f_6uvloop_4loop_8UVHandle__is_alive(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__ensure_alive(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__fatal_error(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__error(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__close(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__fileno(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__new_socket(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__get_socket(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__attach_fileobj(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self, PyObject *__pyx_v_file); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__close(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__open(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self, CYTHON_UNUSED int __pyx_v_sockfd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_7UVAsync__init(struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_7UVAsync_send(struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_f_6uvloop_4loop_7UVAsync_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVIdle__init(struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVIdle_stop(struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVIdle_start(struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_f_6uvloop_4loop_6UVIdle_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_7UVCheck__init(struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_7UVCheck_stop(struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_7UVCheck_start(struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_f_6uvloop_4loop_7UVCheck_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_7UVTimer__init(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx, uint64_t __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_7UVTimer_stop(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_7UVTimer_start(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_7UVTimer_get_when(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_f_6uvloop_4loop_7UVTimer_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx, uint64_t __pyx_v_timeout); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll__init(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, int __pyx_v_fd); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_f_6uvloop_4loop_6UVPoll_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, int __pyx_v_fd); /* proto*/
static int __pyx_f_6uvloop_4loop_6UVPoll_is_active(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVPoll__poll_start(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, int __pyx_v_flags); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVPoll__poll_stop(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_is_reading(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_is_writing(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_start_reading(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_callback); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_start_writing(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_callback); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_stop_reading(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_stop_writing(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_stop(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll__close(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6UVPoll__fatal_error(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc, CYTHON_UNUSED PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_6UVPoll__fatal_error *__pyx_optional_args); /* proto*/
static size_t __pyx_f_6uvloop_4loop_15UVBaseTransport__get_write_buffer_size(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_made(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_lost(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__fatal_error(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_15UVBaseTransport__fatal_error *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_pause_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_resume_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__wakeup_waiter(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__call_connection_made(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__call_connection_lost(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_server(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_waiter(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_waiter); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_protocol); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__clear_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__init_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__add_extra_info(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_obj); /* proto*/
static int __pyx_f_6uvloop_4loop_15UVBaseTransport__is_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__start_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__stop_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_19_StreamWriteContext_free_bufs(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_19_StreamWriteContext_close(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_19_StreamWriteContext_advance_uv_buf(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self, size_t __pyx_v_sent); /* proto*/
static struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_f_6uvloop_4loop_19_StreamWriteContext_new(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_stream, PyObject *__pyx_v_buffers); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVStream__set_protocol(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_protocol); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVStream__clear_protocol(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__shutdown(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__accept(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_server); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__close_on_read_error(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static int __pyx_f_6uvloop_4loop_8UVStream__is_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVStream__start_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_started(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_stopped(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVStream__stop_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__try_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__buffer_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__initiate_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__exec_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static size_t __pyx_f_6uvloop_4loop_8UVStream__get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8UVStream__close(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_accept(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_eof(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__init(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_connect(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_exc); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__init(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol_factory, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer_listen(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__on_listen(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__fatal_error(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_14UVStreamServer__fatal_error *__pyx_optional_args); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__mark_as_open(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_f_6uvloop_4loop_14UVStreamServer__make_new_transport(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_protocol, CYTHON_UNUSED PyObject *__pyx_v_waiter, CYTHON_UNUSED PyObject *__pyx_v_context); /* proto*/
static struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_f_6uvloop_4loop_9TCPServer_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol_factory, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, unsigned int __pyx_v_flags, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9TCPServer__new_socket(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9TCPServer__open(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, int __pyx_v_sockfd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9TCPServer_bind(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, struct sockaddr *__pyx_v_addr, struct __pyx_opt_args_6uvloop_4loop_9TCPServer_bind *__pyx_optional_args); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_f_6uvloop_4loop_9TCPServer__make_new_transport(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, PyObject *__pyx_v_protocol, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context); /* proto*/
static struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_f_6uvloop_4loop_12TCPTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__set_nodelay(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__call_connection_made(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__new_socket(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport_bind(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr, struct __pyx_opt_args_6uvloop_4loop_12TCPTransport_bind *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__open(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, int __pyx_v_sockfd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport_connect(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18_TCPConnectRequest_connect(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self, struct sockaddr *__pyx_v_addr); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_f_6uvloop_4loop_10UnixServer_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol_factory, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_10UnixServer__new_socket(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_10UnixServer__open(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, int __pyx_v_sockfd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_10UnixServer_bind(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, PyObject *__pyx_v_path); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_f_6uvloop_4loop_10UnixServer__make_new_transport(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, PyObject *__pyx_v_protocol, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_10UnixServer__close(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_f_6uvloop_4loop_13UnixTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_13UnixTransport__new_socket(struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_13UnixTransport__open(struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self, int __pyx_v_sockfd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_13UnixTransport_connect(struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self, char *__pyx_v_addr); /* proto*/
static struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_f_6uvloop_4loop_17ReadUnixTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_17ReadUnixTransport__new_socket(struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_17ReadUnixTransport__open(struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, int __pyx_v_sockfd); /* proto*/
static struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_f_6uvloop_4loop_18WriteUnixTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18WriteUnixTransport__new_socket(struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18WriteUnixTransport__open(struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self, int __pyx_v_sockfd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_19_PipeConnectRequest_connect(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self, char *__pyx_v_addr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__close_process_handle(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_args, PyObject *__pyx_v_env, PyObject *__pyx_v_cwd, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr, PyObject *__pyx_v_pass_fds, PyObject *__pyx_v_debug_flags, PyObject *__pyx_v_preexec_fn, PyObject *__pyx_v_restore_signals); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__after_fork(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__close_after_spawn(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, int __pyx_v_fd); /* proto*/
static char **__pyx_f_6uvloop_4loop_9UVProcess__UVProcess__to_cstring_array(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_arr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_options(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_env, PyObject *__pyx_v_cwd, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr, int __pyx_v_force_fork); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_args(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_env(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_env); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_files(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__stdin, CYTHON_UNUSED PyObject *__pyx_v__stdout, CYTHON_UNUSED PyObject *__pyx_v__stderr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__kill(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, int __pyx_v_signum); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__on_exit(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, int64_t __pyx_v_exit_status, int __pyx_v_term_signal); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__close(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__on_exit(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int64_t __pyx_v_exit_status, int __pyx_v_term_signal); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__check_proc(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__pipe_connection_lost(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_fd, PyObject *__pyx_v_exc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__pipe_data_received(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_fd, PyObject *__pyx_v_data); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_redirect_stdio(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_fd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_devnull(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_outpipe(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_inpipe(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__init_files(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__call_connection_made(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_waiter); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__try_finish(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_f_6uvloop_4loop_18UVProcessTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, PyObject *__pyx_v_args, PyObject *__pyx_v_env, PyObject *__pyx_v_cwd, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr, PyObject *__pyx_v_pass_fds, PyObject *__pyx_v_waiter, PyObject *__pyx_v_debug_flags, PyObject *__pyx_v_preexec_fn, PyObject *__pyx_v_restore_signals); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent__init(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent_start(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self, char *__pyx_v_path, int __pyx_v_flags); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent_stop(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent__close(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_f_6uvloop_4loop_9UVFSEvent_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVRequest_on_done(struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_9UVRequest_cancel(struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self); /* proto*/
static void __pyx_f_6uvloop_4loop_8AddrInfo_set_data(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self, struct addrinfo *__pyx_v_data); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_8AddrInfo_unpack(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self); /* proto*/
static int __pyx_f_6uvloop_4loop_8AddrInfo_isinstance(PyObject *__pyx_v_other); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15NameInfoRequest_query(struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self, struct sockaddr *__pyx_v_addr, int __pyx_v_flags); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_app_protocol(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_app_protocol); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__wakeup_waiter(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__get_extra_info(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_name, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__get_extra_info *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_state(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, enum __pyx_t_6uvloop_4loop_SSLProtocolState __pyx_v_new_state); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__start_handshake(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__check_handshake_timeout(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_handshake(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__on_handshake_complete(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_handshake_exc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__start_shutdown(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__start_shutdown *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__check_shutdown_timeout(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read_into_void(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_flush(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_flush *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_shutdown(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_shutdown *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__on_shutdown_complete(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_shutdown_exc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__abort(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_exc); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__write_appdata(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_list_of_data, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_write(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__process_outgoing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read__buffered(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read__copied(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__call_eof_received(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__call_eof_received *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__control_app_writing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing *__pyx_optional_args); /* proto*/
static size_t __pyx_f_6uvloop_4loop_11SSLProtocol__get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__pause_reading(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__resume_reading(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__control_ssl_reading(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits *__pyx_optional_args); /* proto*/
static size_t __pyx_f_6uvloop_4loop_11SSLProtocol__get_read_buffer_size(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__fatal_error(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_exc, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_15_UDPSendContext_close(struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self); /* proto*/
static struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_f_6uvloop_4loop_15_UDPSendContext_new(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_udp, PyObject *__pyx_v_data); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__init(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, unsigned int __pyx_v_family); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__set_address(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct addrinfo *__pyx_v_addr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__connect(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr, CYTHON_UNUSED size_t __pyx_v_addr_len); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport_open(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, int __pyx_v_family, int __pyx_v_sockfd); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__bind(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__set_broadcast(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, int __pyx_v_on); /* proto*/
static size_t __pyx_f_6uvloop_4loop_12UDPTransport__get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto*/
static int __pyx_f_6uvloop_4loop_12UDPTransport__is_reading(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__start_reading(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__stop_reading(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_started(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_stopped(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__new_socket(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__send(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_addr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__on_receive(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_exc, PyObject *__pyx_v_addr); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__on_sent(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_exc, struct __pyx_opt_args_6uvloop_4loop_12UDPTransport__on_sent *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Server__add_server(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_srv); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Server__start_serving(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Server__wakeup(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Server__attach(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Server__detach(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Server__ref(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6uvloop_4loop_6Server__unref(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto*/

/* Module declarations from "uvloop.includes" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "posix.types" */

/* Module declarations from "posix.unistd" */

/* Module declarations from "uvloop.includes.system" */

/* Module declarations from "uvloop.includes.uv" */

/* Module declarations from "cython" */

/* Module declarations from "uvloop.includes.debug" */

/* Module declarations from "uvloop.includes.python" */

/* Module declarations from "uvloop.includes.flowcontrol" */
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_8includes_11flowcontrol_add_flowcontrol_defaults(PyObject *, PyObject *, int); /*proto*/

/* Module declarations from "libc.string" */

/* Module declarations from "libc" */

/* Module declarations from "libc.errno" */

/* Module declarations from "cpython.version" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "uvloop.loop" */
static PyObject *__pyx_v_6uvloop_4loop_aio_get_event_loop = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_CancelledError = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_InvalidStateError = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_TimeoutError = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_Future = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_Task = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_ensure_future = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_gather = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_wait = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_wrap_future = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_logger = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_iscoroutine = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_iscoroutinefunction = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_BaseProtocol = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_Protocol = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_isfuture = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_get_running_loop = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_set_running_loop = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_debug_wrapper = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_AbstractChildWatcher = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_Transport = 0;
static PyObject *__pyx_v_6uvloop_4loop_aio_FlowControlMixin = 0;
static PyObject *__pyx_v_6uvloop_4loop_col_deque = 0;
static PyObject *__pyx_v_6uvloop_4loop_col_Iterable = 0;
static PyObject *__pyx_v_6uvloop_4loop_col_Counter = 0;
static PyObject *__pyx_v_6uvloop_4loop_col_OrderedDict = 0;
static PyObject *__pyx_v_6uvloop_4loop_cc_ThreadPoolExecutor = 0;
static PyObject *__pyx_v_6uvloop_4loop_cc_Future = 0;
static PyObject *__pyx_v_6uvloop_4loop_errno_EBADF = 0;
static PyObject *__pyx_v_6uvloop_4loop_errno_EINVAL = 0;
static PyObject *__pyx_v_6uvloop_4loop_ft_partial = 0;
static PyObject *__pyx_v_6uvloop_4loop_gc_disable = 0;
static PyObject *__pyx_v_6uvloop_4loop_iter_chain = 0;
static PyObject *__pyx_v_6uvloop_4loop_inspect_isgenerator = 0;
static int __pyx_v_6uvloop_4loop_has_IPV6_V6ONLY;
static int __pyx_v_6uvloop_4loop_IPV6_V6ONLY;
static int __pyx_v_6uvloop_4loop_has_SO_REUSEPORT;
static int __pyx_v_6uvloop_4loop_SO_REUSEPORT;
static int __pyx_v_6uvloop_4loop_SO_BROADCAST;
static int __pyx_v_6uvloop_4loop_SOCK_NONBLOCK;
static int __pyx_v_6uvloop_4loop_socket_AI_CANONNAME;
static PyObject *__pyx_v_6uvloop_4loop_socket_gaierror = 0;
static PyObject *__pyx_v_6uvloop_4loop_socket_error = 0;
static PyObject *__pyx_v_6uvloop_4loop_socket_timeout = 0;
static PyObject *__pyx_v_6uvloop_4loop_socket_socket = 0;
static PyObject *__pyx_v_6uvloop_4loop_socket_socketpair = 0;
static PyObject *__pyx_v_6uvloop_4loop_socket_getservbyname = 0;
static PyObject *__pyx_v_6uvloop_4loop_socket_AddressFamily = 0;
static PyObject *__pyx_v_6uvloop_4loop_socket_SocketKind = 0;
static int __pyx_v_6uvloop_4loop_socket_EAI_ADDRFAMILY;
static int __pyx_v_6uvloop_4loop_socket_EAI_AGAIN;
static int __pyx_v_6uvloop_4loop_socket_EAI_BADFLAGS;
static int __pyx_v_6uvloop_4loop_socket_EAI_BADHINTS;
static int __pyx_v_6uvloop_4loop_socket_EAI_CANCELED;
static int __pyx_v_6uvloop_4loop_socket_EAI_FAIL;
static int __pyx_v_6uvloop_4loop_socket_EAI_FAMILY;
static int __pyx_v_6uvloop_4loop_socket_EAI_MEMORY;
static int __pyx_v_6uvloop_4loop_socket_EAI_NODATA;
static int __pyx_v_6uvloop_4loop_socket_EAI_NONAME;
static int __pyx_v_6uvloop_4loop_socket_EAI_OVERFLOW;
static int __pyx_v_6uvloop_4loop_socket_EAI_PROTOCOL;
static int __pyx_v_6uvloop_4loop_socket_EAI_SERVICE;
static int __pyx_v_6uvloop_4loop_socket_EAI_SOCKTYPE;
static PyObject *__pyx_v_6uvloop_4loop_os_name = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_environ = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_dup = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_set_inheritable = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_get_inheritable = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_close = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_open = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_devnull = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_O_RDWR = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_pipe = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_read = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_remove = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_stat = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_unlink = 0;
static PyObject *__pyx_v_6uvloop_4loop_os_fspath = 0;
static PyObject *__pyx_v_6uvloop_4loop_stat_S_ISSOCK = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_ignore_environment = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_dev_mode = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_exc_info = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_set_coroutine_wrapper = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_get_coroutine_wrapper = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_getframe = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_version_info = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_getfilesystemencoding = 0;
static PyObject *__pyx_v_6uvloop_4loop_sys_platform = 0;
static PyObject *__pyx_v_6uvloop_4loop_ssl_SSLContext = 0;
static PyObject *__pyx_v_6uvloop_4loop_ssl_MemoryBIO = 0;
static PyObject *__pyx_v_6uvloop_4loop_ssl_create_default_context = 0;
static PyObject *__pyx_v_6uvloop_4loop_ssl_SSLError = 0;
static PyObject *__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors = 0;
static PyObject *__pyx_v_6uvloop_4loop_ssl_SSLZeroReturnError = 0;
static PyObject *__pyx_v_6uvloop_4loop_ssl_CertificateError = 0;
static int __pyx_v_6uvloop_4loop_ssl_SSL_ERROR_WANT_READ;
static int __pyx_v_6uvloop_4loop_ssl_SSL_ERROR_WANT_WRITE;
static int __pyx_v_6uvloop_4loop_ssl_SSL_ERROR_SYSCALL;
static PyObject *__pyx_v_6uvloop_4loop_threading_Thread = 0;
static PyObject *__pyx_v_6uvloop_4loop_threading_main_thread = 0;
static int __pyx_v_6uvloop_4loop_subprocess_PIPE;
static int __pyx_v_6uvloop_4loop_subprocess_STDOUT;
static int __pyx_v_6uvloop_4loop_subprocess_DEVNULL;
static PyObject *__pyx_v_6uvloop_4loop_subprocess_SubprocessError = 0;
static int __pyx_v_6uvloop_4loop_signal_NSIG;
static PyObject *__pyx_v_6uvloop_4loop_signal_signal = 0;
static PyObject *__pyx_v_6uvloop_4loop_signal_siginterrupt = 0;
static PyObject *__pyx_v_6uvloop_4loop_signal_set_wakeup_fd = 0;
static PyObject *__pyx_v_6uvloop_4loop_signal_default_int_handler = 0;
static PyObject *__pyx_v_6uvloop_4loop_signal_SIG_DFL = 0;
static PyObject *__pyx_v_6uvloop_4loop_time_sleep = 0;
static PyObject *__pyx_v_6uvloop_4loop_time_monotonic = 0;
static PyObject *__pyx_v_6uvloop_4loop_tb_StackSummary = 0;
static PyObject *__pyx_v_6uvloop_4loop_tb_walk_stack = 0;
static PyObject *__pyx_v_6uvloop_4loop_tb_format_list = 0;
static PyObject *__pyx_v_6uvloop_4loop_warnings_warn = 0;
static PyObject *__pyx_v_6uvloop_4loop_weakref_WeakValueDictionary = 0;
static PyObject *__pyx_v_6uvloop_4loop_weakref_WeakSet = 0;
static PyObject *__pyx_v_6uvloop_4loop_py_inf = 0;
static int __pyx_v_6uvloop_4loop_PY39;
static int __pyx_v_6uvloop_4loop_PY311;
static int __pyx_v_6uvloop_4loop_PY313;
static uint64_t __pyx_v_6uvloop_4loop_MAX_SLEEP;
static PyObject *__pyx_v_6uvloop_4loop__LRU_MARKER = 0;
static struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_6uvloop_4loop_sockaddrs = 0;
static volatile int __pyx_v_6uvloop_4loop___atfork_installed;
static volatile int __pyx_v_6uvloop_4loop___forking;
static struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_6uvloop_4loop___forking_loop = 0;
static volatile int __pyx_v_6uvloop_4loop___mem_installed;
static PyObject *__pyx_f_6uvloop_4loop___strerr(int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___convert_python_error(int); /*proto*/
static int __pyx_f_6uvloop_4loop___convert_socket_error(int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_convert_error(int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop__is_sock_stream(PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop__is_sock_dgram(PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_isfuture(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_socket_inc_io_ref(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_socket_dec_io_ref(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_run_in_context(PyObject *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_run_in_context1(PyObject *, PyObject *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_run_in_context2(PyObject *, PyObject *, PyObject *, PyObject *); /*proto*/
static void __pyx_f_6uvloop_4loop___loop_alloc_buffer(uv_handle_t *, size_t, uv_buf_t *); /*proto*/
static CYTHON_INLINE void __pyx_f_6uvloop_4loop___loop_free_buffer(struct __pyx_obj_6uvloop_4loop_Loop *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_format_callback_name(PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_new_Handle(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, __pyx_t_6uvloop_4loop_method_t, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle1(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, __pyx_t_6uvloop_4loop_method1_t, PyObject *, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle2(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, __pyx_t_6uvloop_4loop_method2_t, PyObject *, PyObject *, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle3(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, __pyx_t_6uvloop_4loop_method3_t, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop_extract_stack(void); /*proto*/
static CYTHON_INLINE int __pyx_f_6uvloop_4loop___ensure_handle_data(uv_handle_t *, char const *); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_close_handle_cb(uv_handle_t *); /*proto*/
static void __pyx_f_6uvloop_4loop___close_all_handles(struct __pyx_obj_6uvloop_4loop_Loop *); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_walk_close_all_handles_cb(uv_handle_t *, void *); /*proto*/
static void __pyx_f_6uvloop_4loop___uvasync_callback(uv_async_t *); /*proto*/
static void __pyx_f_6uvloop_4loop_cb_idle_callback(uv_idle_t *); /*proto*/
static void __pyx_f_6uvloop_4loop_cb_check_callback(uv_check_t *); /*proto*/
static void __pyx_f_6uvloop_4loop___uvtimer_callback(uv_timer_t *); /*proto*/
static void __pyx_f_6uvloop_4loop___on_uvpoll_event(uv_poll_t *, int, int); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_stream_on_shutdown(uv_shutdown_t *, int); /*proto*/
static CYTHON_INLINE int __pyx_f_6uvloop_4loop___uv_stream_on_read_common(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_Loop *, Py_ssize_t); /*proto*/
static CYTHON_INLINE void __pyx_f_6uvloop_4loop___uv_stream_on_read_impl(uv_stream_t *, Py_ssize_t, uv_buf_t const *); /*proto*/
static CYTHON_INLINE void __pyx_f_6uvloop_4loop___uv_stream_on_write_impl(uv_write_t *, int); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_stream_on_read(uv_stream_t *, Py_ssize_t, uv_buf_t const *); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_stream_on_write(uv_write_t *, int); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_stream_buffered_alloc(uv_handle_t *, size_t, uv_buf_t *); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_stream_buffered_on_read(uv_stream_t *, Py_ssize_t, uv_buf_t const *); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_streamserver_on_listen(uv_stream_t *, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___tcp_init_uv_handle(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_Loop *, unsigned int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___tcp_bind(struct __pyx_obj_6uvloop_4loop_UVStream *, struct sockaddr *, unsigned int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___tcp_open(struct __pyx_obj_6uvloop_4loop_UVStream *, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___tcp_get_socket(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *); /*proto*/
static void __pyx_f_6uvloop_4loop___tcp_connect_callback(uv_connect_t *, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___pipe_init_uv_handle(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_Loop *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___pipe_open(struct __pyx_obj_6uvloop_4loop_UVStream *, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___pipe_get_socket(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *); /*proto*/
static void __pyx_f_6uvloop_4loop___pipe_connect_callback(uv_connect_t *, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___process_convert_fileno(PyObject *); /*proto*/
static void __pyx_f_6uvloop_4loop___uvprocess_on_exit_callback(uv_process_t *, int64_t, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___socketpair(void); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_close_process_handle_cb(uv_handle_t *); /*proto*/
static void __pyx_f_6uvloop_4loop___uvfsevent_callback(uv_fs_event_t *, char const *, int, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___port_to_int(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(struct sockaddr const *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___convert_pyaddr_to_sockaddr(int, PyObject *, struct sockaddr *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___static_getaddrinfo(PyObject *, PyObject *, int, int, int, struct sockaddr *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___static_getaddrinfo_pyaddr(PyObject *, PyObject *, int, int, int, int); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop__intenum_converter(PyObject *, PyObject *); /*proto*/
static void __pyx_f_6uvloop_4loop___on_addrinfo_resolved(uv_getaddrinfo_t *, int, struct addrinfo *); /*proto*/
static void __pyx_f_6uvloop_4loop___on_nameinfo_resolved(uv_getnameinfo_t *, int, char const *, char const *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop__create_transport_context(PyObject *, PyObject *); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_udp_on_receive(uv_udp_t *, Py_ssize_t, uv_buf_t const *, struct sockaddr const *, unsigned int); /*proto*/
static void __pyx_f_6uvloop_4loop___uv_udp_on_send(uv_udp_send_t *, int); /*proto*/
static void __pyx_f_6uvloop_4loop___get_fork_handler(void); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___install_atfork(void); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___install_pymem(void); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop__set_signal_wakeup_fd(PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___pyx_unpickle_PseudoSocket__set_state(struct __pyx_obj_6uvloop_4loop_PseudoSocket *, PyObject *); /*proto*/
static PyObject *__pyx_f_6uvloop_4loop___pyx_unpickle_LruCache__set_state(struct __pyx_obj_6uvloop_4loop_LruCache *, PyObject *); /*proto*/
static PyObject *__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, int)); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "uvloop.loop"
extern int __pyx_module_is_main_uvloop__loop;
int __pyx_module_is_main_uvloop__loop = 0;

/* Implementation of "uvloop.loop" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_OSError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_BaseException;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_KeyboardInterrupt;
static PyObject *__pyx_builtin_SystemExit;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_OverflowError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_RuntimeWarning;
static PyObject *__pyx_builtin_DeprecationWarning;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_UnicodeEncodeError;
/* #### Code section: string_decls ### */
static const char __pyx_k_1[] = "1";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "-c";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_ag[] = "ag";
static const char __pyx_k_ai[] = "ai";
static const char __pyx_k_ex[] = "ex";
static const char __pyx_k_f1[] = "f1";
static const char __pyx_k_f2[] = "f2";
static const char __pyx_k_fd[] = "fd";
static const char __pyx_k_fs[] = "fs";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_sl[] = "sl";
static const char __pyx_k_tb[] = "tb";
static const char __pyx_k_tr[] = "tr";
static const char __pyx_k_wb[] = "wb";
static const char __pyx_k__10[] = "";
static const char __pyx_k__33[] = "-------------------------------+---------+---------";
static const char __pyx_k__61[] = ": ";
static const char __pyx_k__69[] = ", ";
static const char __pyx_k__75[] = "\000";
static const char __pyx_k__78[] = "{}: {}";
static const char __pyx_k__79[] = "\n";
static const char __pyx_k_abc[] = "abc";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_buf[] = "buf";
static const char __pyx_k_cmd[] = "cmd";
static const char __pyx_k_ctx[] = "ctx";
static const char __pyx_k_cwd[] = "cwd";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_dup[] = "dup";
static const char __pyx_k_env[] = "env";
static const char __pyx_k_err[] = "err";
static const char __pyx_k_exc[] = "exc";
static const char __pyx_k_fut[] = "fut";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_lai[] = "lai";
static const char __pyx_k_log[] = "log";
static const char __pyx_k_low[] = "low";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_r_s[] = "%r: %s";
static const char __pyx_k_rai[] = "rai";
static const char __pyx_k_run[] = "run";
static const char __pyx_k_sig[] = "sig";
static const char __pyx_k_ssl[] = "ssl";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_tcp[] = "tcp";
static const char __pyx_k_udp[] = "udp";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_Loop[] = "Loop";
static const char __pyx_k_NSIG[] = "NSIG";
static const char __pyx_k_PIPE[] = "PIPE";
static const char __pyx_k_Task[] = "Task";
static const char __pyx_k__115[] = "{}={}";
static const char __pyx_k__128[] = "<";
static const char __pyx_k__129[] = " ";
static const char __pyx_k__130[] = ">";
static const char __pyx_k__140[] = " (";
static const char __pyx_k__141[] = ")";
static const char __pyx_k__161[] = ":";
static const char __pyx_k__170[] = "=";
static const char __pyx_k__196[] = ".";
static const char __pyx_k__198[] = "*";
static const char __pyx_k__263[] = "_";
static const char __pyx_k__542[] = "?";
static const char __pyx_k_addr[] = "addr";
static const char __pyx_k_agen[] = "agen";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bind[] = "bind";
static const char __pyx_k_bufs[] = "bufs";
static const char __pyx_k_cast[] = "cast";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_coro[] = "coro";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_done[] = "done";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_excs[] = "excs";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_fd_2[] = " fd=";
static const char __pyx_k_flag[] = "flag";
static const char __pyx_k_func[] = "func";
static const char __pyx_k_high[] = "high (";
static const char __pyx_k_host[] = "host";
static const char __pyx_k_idna[] = "idna";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_kill[] = "kill";
static const char __pyx_k_lads[] = "lads";
static const char __pyx_k_last[] = "last";
static const char __pyx_k_loop[] = "loop";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_nexc[] = "nexc";
static const char __pyx_k_noop[] = "_noop";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_opts[] = "opts";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_pipe[] = "pipe";
static const char __pyx_k_port[] = "port";
static const char __pyx_k_proc[] = "proc";
static const char __pyx_k_rads[] = "rads";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_recv[] = "recv";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sin6[] = "sin6";
static const char __pyx_k_sock[] = "sock";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_stat[] = "stat";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_task[] = "task";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_wait[] = "wait";
static const char __pyx_k_want[] = "want";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_when[] = "when";
static const char __pyx_k_wrap[] = "wrap";
static const char __pyx_k_abort[] = "abort";
static const char __pyx_k_addrs[] = "addrs";
static const char __pyx_k_aexit[] = "__aexit__";
static const char __pyx_k_ascii[] = "ascii";
static const char __pyx_k_await[] = "__await__";
static const char __pyx_k_cause[] = "__cause__";
static const char __pyx_k_chain[] = "chain";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_delay[] = "delay";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_errno[] = "errno";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_frame[] = "frame";
static const char __pyx_k_hosts[] = "hosts";
static const char __pyx_k_ident[] = "ident";
static const char __pyx_k_infos[] = "infos";
static const char __pyx_k_ioctl[] = "ioctl";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_limit[] = "limit";
static const char __pyx_k_lower[] = "lower";
static const char __pyx_k_model[] = "model";
static const char __pyx_k_posix[] = "posix";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_proto[] = "proto";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_share[] = "share";
static const char __pyx_k_shell[] = "shell";
static const char __pyx_k_sleep[] = "sleep";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_stdin[] = "stdin";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_unset[] = "_unset";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_CHANGE[] = "CHANGE";
static const char __pyx_k_Future[] = "Future";
static const char __pyx_k_Handle[] = "Handle";
static const char __pyx_k_O_RDWR[] = "O_RDWR";
static const char __pyx_k_RENAME[] = "RENAME";
static const char __pyx_k_STDOUT[] = "STDOUT";
static const char __pyx_k_Server[] = "Server";
static const char __pyx_k_Thread[] = "Thread";
static const char __pyx_k_UVIdle[] = "UVIdle";
static const char __pyx_k_UVPoll[] = "UVPoll";
static const char __pyx_k_accept[] = "accept";
static const char __pyx_k_aclose[] = "aclose";
static const char __pyx_k_aenter[] = "__aenter__";
static const char __pyx_k_ai_cnt[] = "ai_cnt";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_bin_sh[] = "/bin/sh";
static const char __pyx_k_cancel[] = "cancel";
static const char __pyx_k_cipher[] = "cipher";
static const char __pyx_k_cygwin[] = "cygwin";
static const char __pyx_k_darwin[] = "darwin";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_detach[] = "detach";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_family[] = "family";
static const char __pyx_k_fileno[] = "fileno";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_fspath[] = "fspath";
static const char __pyx_k_future[] = "future";
static const char __pyx_k_gather[] = "gather";
static const char __pyx_k_handle[] = "handle";
static const char __pyx_k_high_2[] = "high";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_listen[] = "listen";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_nbytes[] = "nbytes";
static const char __pyx_k_noop_2[] = "noop";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_pipe_2[] = " pipe=";
static const char __pyx_k_pyaddr[] = "pyaddr";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_rstrip[] = "rstrip";
static const char __pyx_k_self_2[] = "__self__";
static const char __pyx_k_sendto[] = "sendto";
static const char __pyx_k_server[] = "server ";
static const char __pyx_k_signal[] = "signal";
static const char __pyx_k_signum[] = "signum";
static const char __pyx_k_socket[] = "socket ";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_st_ino[] = "st_ino";
static const char __pyx_k_stderr[] = "stderr";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_submit[] = "submit";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_thread[] = "thread";
static const char __pyx_k_transp[] = "transp";
static const char __pyx_k_unlink[] = "unlink";
static const char __pyx_k_unwrap[] = "unwrap";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_wait_2[] = "_wait";
static const char __pyx_k_waiter[] = "waiter";
static const char __pyx_k_Counter[] = "Counter";
static const char __pyx_k_DEVNULL[] = "DEVNULL";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_OSError[] = "OSError";
static const char __pyx_k_SIG_DFL[] = "SIG_DFL";
static const char __pyx_k_UVAsync[] = "UVAsync";
static const char __pyx_k_UVCheck[] = "UVCheck";
static const char __pyx_k_UVTimer[] = "UVTimer";
static const char __pyx_k_WeakSet[] = "WeakSet";
static const char __pyx_k_address[] = "address";
static const char __pyx_k_asyncio[] = "asyncio";
static const char __pyx_k_backlog[] = "backlog";
static const char __pyx_k_bufsize[] = "bufsize";
static const char __pyx_k_call_at[] = "call_at";
static const char __pyx_k_connect[] = "connect";
static const char __pyx_k_context[] = "context";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_devnull[] = "devnull";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_discard[] = "discard";
static const char __pyx_k_done_cb[] = "done_cb";
static const char __pyx_k_enabled[] = "enabled";
static const char __pyx_k_environ[] = "environ";
static const char __pyx_k_extract[] = "extract";
static const char __pyx_k_factory[] = "factory";
static const char __pyx_k_fileobj[] = "fileobj";
static const char __pyx_k_futures[] = "futures";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_get_pid[] = "get_pid";
static const char __pyx_k_handler[] = "handler";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_io_refs[] = "_io_refs";
static const char __pyx_k_laddr_s[] = ", laddr=%s";
static const char __pyx_k_maxsize[] = "maxsize";
static const char __pyx_k_message[] = "message";
static const char __pyx_k_p_bytes[] = "p_bytes";
static const char __pyx_k_partial[] = "partial";
static const char __pyx_k_pending[] = "pending";
static const char __pyx_k_popitem[] = "popitem";
static const char __pyx_k_popleft[] = "popleft";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_program[] = "program";
static const char __pyx_k_raddr_s[] = ", raddr=%s";
static const char __pyx_k_recvmsg[] = "recvmsg";
static const char __pyx_k_results[] = "results";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_seconds[] = " seconds.";
static const char __pyx_k_sendall[] = "sendall";
static const char __pyx_k_sendmsg[] = "sendmsg";
static const char __pyx_k_servers[] = "servers";
static const char __pyx_k_sockets[] = " sockets=";
static const char __pyx_k_st_mode[] = "st_mode";
static const char __pyx_k_timeout[] = "timeout";
static const char __pyx_k_warning[] = "warning";
static const char __pyx_k_weakref[] = "weakref";
static const char __pyx_k_18_7_7_7[] = "    {: <18} {: >7} | {: >7} | {: >7}";
static const char __pyx_k_AddrInfo[] = "AddrInfo";
static const char __pyx_k_EAI_FAIL[] = "EAI_FAIL";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_LruCache[] = "LruCache";
static const char __pyx_k_Protocol[] = "Protocol";
static const char __pyx_k_SSLError[] = "SSLError";
static const char __pyx_k_S_ISSOCK[] = "S_ISSOCK";
static const char __pyx_k_UVHandle[] = "UVHandle";
static const char __pyx_k_UVStream[] = "UVStream";
static const char __pyx_k_addrinfo[] = "addrinfo";
static const char __pyx_k_ai_local[] = "ai_local";
static const char __pyx_k_asyncgen[] = "asyncgen";
static const char __pyx_k_builtins[] = "__builtins__";
static const char __pyx_k_callback[] = "callback";
static const char __pyx_k_closed_x[] = "<{} closed={} {:#x}>";
static const char __pyx_k_dev_mode[] = "dev_mode";
static const char __pyx_k_exc_info[] = "exc_info";
static const char __pyx_k_executor[] = "executor";
static const char __pyx_k_flowinfo[] = "flowinfo";
static const char __pyx_k_gaierror[] = "gaierror";
static const char __pyx_k_get_loop[] = "get_loop";
static const char __pyx_k_getframe[] = "_getframe";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_is_alive[] = "is_alive";
static const char __pyx_k_isfuture[] = "isfuture";
static const char __pyx_k_lai_iter[] = "lai_iter";
static const char __pyx_k_makefile[] = "makefile";
static const char __pyx_k_new_task[] = "new_task";
static const char __pyx_k_pass_fds[] = "pass_fds";
static const char __pyx_k_peercert[] = "peercert";
static const char __pyx_k_peername[] = "peername";
static const char __pyx_k_platform[] = "platform";
static const char __pyx_k_problems[] = "problems";
static const char __pyx_k_protocol[] = "protocol";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_rai_iter[] = "rai_iter";
static const char __pyx_k_recvfrom[] = "recvfrom";
static const char __pyx_k_scope_id[] = "scope_id";
static const char __pyx_k_sendfile[] = "sendfile";
static const char __pyx_k_server_2[] = "server";
static const char __pyx_k_set_name[] = "set_name";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_shutdown[] = "shutdown";
static const char __pyx_k_sockaddr[] = "sockaddr";
static const char __pyx_k_socket_2[] = "socket";
static const char __pyx_k_sockname[] = "sockname";
static const char __pyx_k_sslproto[] = "sslproto";
static const char __pyx_k_strerror[] = "strerror";
static const char __pyx_k_subtract[] = "subtract";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_wrap_bio[] = "wrap_bio";
static const char __pyx_k_EAI_AGAIN[] = "EAI_AGAIN";
static const char __pyx_k_Loop_stop[] = "Loop.stop";
static const char __pyx_k_Loop_time[] = "Loop time:                 {}";
static const char __pyx_k_MemoryBIO[] = "MemoryBIO";
static const char __pyx_k_Read_EOFs[] = "Read EOFs:                 {}";
static const char __pyx_k_TCPServer[] = "TCPServer";
static const char __pyx_k_Transport[] = "Transport";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_UVFSEvent[] = "UVFSEvent";
static const char __pyx_k_UVProcess[] = "UVProcess";
static const char __pyx_k_UVRequest[] = "UVRequest";
static const char __pyx_k_ai_remote[] = "ai_remote";
static const char __pyx_k_call_soon[] = "call_soon";
static const char __pyx_k_cancelled[] = "cancelled";
static const char __pyx_k_close_fds[] = "close_fds";
static const char __pyx_k_completed[] = "completed";
static const char __pyx_k_exception[] = "exception";
static const char __pyx_k_finalizer[] = "finalizer";
static const char __pyx_k_firstiter[] = "firstiter";
static const char __pyx_k_fs_handle[] = "fs_handle";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_get_debug[] = "get_debug";
static const char __pyx_k_is_closed[] = "is_closed";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_log_lines[] = "log_lines";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_monotonic[] = "monotonic";
static const char __pyx_k_must_be_0[] = ") must be >= 0";
static const char __pyx_k_orig_path[] = "orig_path";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_recv_into[] = "recv_into";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_resume_cb[] = "resume_cb";
static const char __pyx_k_set_debug[] = "set_debug";
static const char __pyx_k_sock_recv[] = "sock_recv";
static const char __pyx_k_sockets_2[] = "sockets";
static const char __pyx_k_start_tls[] = "start_tls";
static const char __pyx_k_stdio_fut[] = "stdio_fut";
static const char __pyx_k_terminate[] = "terminate";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_traceback[] = "__traceback__";
static const char __pyx_k_transport[] = "transport";
static const char __pyx_k_write_eof[] = "write_eof";
static const char __pyx_k_EAI_FAMILY[] = "EAI_FAMILY";
static const char __pyx_k_EAI_MEMORY[] = "EAI_MEMORY";
static const char __pyx_k_EAI_NODATA[] = "EAI_NODATA";
static const char __pyx_k_EAI_NONAME[] = "EAI_NONAME";
static const char __pyx_k_Loop__stop[] = "Loop._stop";
static const char __pyx_k_Loop_close[] = "Loop.close";
static const char __pyx_k_SSLContext[] = "SSLContext";
static const char __pyx_k_SocketKind[] = "SocketKind";
static const char __pyx_k_SystemExit[] = "SystemExit";
static const char __pyx_k_UnixServer[] = "UnixServer";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_add_reader[] = "add_reader";
static const char __pyx_k_add_writer[] = "add_writer";
static const char __pyx_k_c_str_path[] = "c_str_path";
static const char __pyx_k_call_later[] = "call_later";
static const char __pyx_k_concurrent[] = "concurrent";
static const char __pyx_k_conmade_cb[] = "conmade_cb";
static const char __pyx_k_connect_ex[] = "connect_ex";
static const char __pyx_k_coroutines[] = "coroutines";
static const char __pyx_k_created_at[] = "created at {}:{}";
static const char __pyx_k_exceptions[] = "exceptions";
static const char __pyx_k_executable[] = "executable";
static const char __pyx_k_get_buffer[] = "get_buffer";
static const char __pyx_k_getsockopt[] = "getsockopt";
static const char __pyx_k_gettimeout[] = "gettimeout";
static const char __pyx_k_is_closing[] = "is_closing";
static const char __pyx_k_is_reading[] = "is_reading";
static const char __pyx_k_is_running[] = "is_running";
static const char __pyx_k_is_serving[] = "is_serving";
static const char __pyx_k_lai_static[] = "lai_static";
static const char __pyx_k_local_addr[] = "local_addr";
static const char __pyx_k_monitor_fs[] = "_monitor_fs";
static const char __pyx_k_preexec_fn[] = "preexec_fn";
static const char __pyx_k_process_id[] = "process_id";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_rai_static[] = "rai_static";
static const char __pyx_k_reuse_port[] = "reuse_port";
static const char __pyx_k_set_name_2[] = "__set_name__";
static const char __pyx_k_set_result[] = "set_result";
static const char __pyx_k_setsockopt[] = "setsockopt";
static const char __pyx_k_settimeout[] = "settimeout";
static const char __pyx_k_sighandler[] = "__sighandler";
static const char __pyx_k_socketpair[] = "socketpair";
static const char __pyx_k_ssl_object[] = "ssl_object";
static const char __pyx_k_ssl_waiter[] = "ssl_waiter";
static const char __pyx_k_sslcontext[] = "sslcontext";
static const char __pyx_k_stacklevel[] = "stacklevel";
static const char __pyx_k_subprocess[] = "subprocess";
static const char __pyx_k_transports[] = "transports";
static const char __pyx_k_walk_stack[] = "walk_stack";
static const char __pyx_k_writelines[] = "writelines";
static const char __pyx_k_EAI_SERVICE[] = "EAI_SERVICE";
static const char __pyx_k_IPV6_V6ONLY[] = "IPV6_V6ONLY";
static const char __pyx_k_Loop_time_2[] = "Loop.time";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_Read_errors[] = "Read errors:               {}";
static const char __pyx_k_Read_events[] = "Read events:               {}";
static const char __pyx_k_SSLProtocol[] = "SSLProtocol";
static const char __pyx_k_TimerHandle[] = "TimerHandle";
static const char __pyx_k_asyncio_log[] = "asyncio.log";
static const char __pyx_k_base_events[] = "base_events";
static const char __pyx_k_bind_method[] = "bind() method";
static const char __pyx_k_cfunc_to_py[] = "cfunc.to_py";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_compression[] = "compression";
static const char __pyx_k_create_task[] = "create_task";
static const char __pyx_k_debug_flags[] = "debug_flags";
static const char __pyx_k_do_shutdown[] = "_do_shutdown";
static const char __pyx_k_force_close[] = "_force_close";
static const char __pyx_k_format_list[] = "format_list";
static const char __pyx_k_getaddrinfo[] = "getaddrinfo";
static const char __pyx_k_getnameinfo[] = "getnameinfo";
static const char __pyx_k_getpeercert[] = "getpeercert";
static const char __pyx_k_getpeername[] = "getpeername";
static const char __pyx_k_getsockname[] = "getsockname";
static const char __pyx_k_is_closed_2[] = " is closed";
static const char __pyx_k_iscoroutine[] = "iscoroutine";
static const char __pyx_k_isgenerator[] = "isgenerator";
static const char __pyx_k_main_thread[] = "main_thread";
static const char __pyx_k_move_to_end[] = "move_to_end";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_must_be_low[] = ") must be >= low (";
static const char __pyx_k_recv_method[] = "recv() method";
static const char __pyx_k_remote_addr[] = "remote_addr";
static const char __pyx_k_run_forever[] = "run_forever";
static const char __pyx_k_send_method[] = "send() method";
static const char __pyx_k_send_signal[] = "send_signal";
static const char __pyx_k_server_side[] = "server_side";
static const char __pyx_k_setblocking[] = "setblocking";
static const char __pyx_k_sock_accept[] = "sock_accept";
static const char __pyx_k_sock_sendto[] = "sock_sendto";
static const char __pyx_k_startupinfo[] = "startupinfo";
static const char __pyx_k_traceback_2[] = "traceback";
static const char __pyx_k_transport_2[] = "transport ";
static const char __pyx_k_uvloop_loop[] = "uvloop.loop";
static const char __pyx_k_wait_closed[] = "wait_closed";
static const char __pyx_k_wrap_future[] = "wrap_future";
static const char __pyx_k_AI_CANONNAME[] = "AI_CANONNAME";
static const char __pyx_k_BaseProtocol[] = "BaseProtocol";
static const char __pyx_k_EAI_BADFLAGS[] = "EAI_BADFLAGS";
static const char __pyx_k_EAI_BADHINTS[] = "EAI_BADHINTS";
static const char __pyx_k_EAI_CANCELED[] = "EAI_CANCELED";
static const char __pyx_k_EAI_OVERFLOW[] = "EAI_OVERFLOW";
static const char __pyx_k_EAI_PROTOCOL[] = "EAI_PROTOCOL";
static const char __pyx_k_EAI_SOCKTYPE[] = "EAI_SOCKTYPE";
static const char __pyx_k_Loop_call_at[] = "Loop.call_at";
static const char __pyx_k_Process_info[] = "---- Process info: -----";
static const char __pyx_k_PseudoSocket[] = "PseudoSocket";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_SO_BROADCAST[] = "SO_BROADCAST";
static const char __pyx_k_SO_REUSEPORT[] = "SO_REUSEPORT";
static const char __pyx_k_Server_close[] = "Server.close";
static const char __pyx_k_StackSummary[] = "StackSummary";
static const char __pyx_k_TCPTransport[] = "TCPTransport";
static const char __pyx_k_TimeoutError[] = "TimeoutError";
static const char __pyx_k_UDPTransport[] = "UDPTransport";
static const char __pyx_k_Write_errors[] = "Write errors:              {}";
static const char __pyx_k_Write_events[] = "Write events:              {}";
static const char __pyx_k_alive_closed[] = "                        alive  | closed  |";
static const char __pyx_k_app_protocol[] = "app_protocol";
static const char __pyx_k_close_method[] = "close() method";
static const char __pyx_k_disconnected[] = "disconnected";
static const char __pyx_k_do_handshake[] = "do_handshake";
static const char __pyx_k_eof_received[] = "eof_received";
static const char __pyx_k_get_protocol[] = "get_protocol";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_list_of_data[] = "list_of_data";
static const char __pyx_k_lookup_lines[] = "lookup_lines";
static const char __pyx_k_objs_handles[] = "                        objs   | handles |";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_recvmsg_into[] = "recvmsg_into";
static const char __pyx_k_set_protocol[] = "set_protocol";
static const char __pyx_k_siginterrupt[] = "siginterrupt";
static const char __pyx_k_sock_connect[] = "sock_connect";
static const char __pyx_k_sock_sendall[] = "sock_sendall";
static const char __pyx_k_ssl_protocol[] = "ssl_protocol";
static const char __pyx_k_stdio_inited[] = "__stdio_inited";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_version_info[] = "version_info";
static const char __pyx_k_AddressFamily[] = "AddressFamily";
static const char __pyx_k_BaseException[] = "BaseException";
static const char __pyx_k_Errors_logged[] = "Errors logged:             {}";
static const char __pyx_k_Handle_cancel[] = "Handle.cancel";
static const char __pyx_k_Listen_errors[] = "Listen errors:             {}";
static const char __pyx_k_OverflowError[] = "OverflowError";
static const char __pyx_k_RENAME_CHANGE[] = "RENAME_CHANGE";
static const char __pyx_k_SOCK_NONBLOCK[] = "SOCK_NONBLOCK";
static const char __pyx_k_UnixTransport[] = "UnixTransport";
static const char __pyx_k_accept_method[] = "accept() method";
static const char __pyx_k_app_transport[] = "app_transport";
static const char __pyx_k_asyncio_tasks[] = "asyncio.tasks";
static const char __pyx_k_can_write_eof[] = "can_write_eof";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_closing_agens[] = "closing_agens";
static const char __pyx_k_create_future[] = "create_future";
static const char __pyx_k_create_server[] = "create_server";
static const char __pyx_k_creationflags[] = "creationflags";
static const char __pyx_k_data_received[] = "data_received";
static const char __pyx_k_debug_wrapper[] = "debug_wrapper";
static const char __pyx_k_detach_method[] = "detach() method";
static const char __pyx_k_ensure_future[] = "ensure_future";
static const char __pyx_k_getservbyname[] = "getservbyname";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_listen_method[] = "listen() method";
static const char __pyx_k_loop__on_idle[] = "loop._on_idle";
static const char __pyx_k_pause_reading[] = "pause_reading";
static const char __pyx_k_pause_writing[] = "pause_writing";
static const char __pyx_k_proto_factory[] = "proto_factory";
static const char __pyx_k_recvfrom_into[] = "recvfrom_into";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_remove_reader[] = "remove_reader";
static const char __pyx_k_remove_writer[] = "remove_writer";
static const char __pyx_k_reuse_address[] = "reuse_address";
static const char __pyx_k_sendmsg_afalg[] = "sendmsg_afalg";
static const char __pyx_k_sendto_method[] = "sendto() method";
static const char __pyx_k_serve_forever[] = "serve_forever";
static const char __pyx_k_set_exception[] = "set_exception";
static const char __pyx_k_set_wakeup_fd[] = "set_wakeup_fd";
static const char __pyx_k_shutdown_coro[] = "shutdown_coro";
static const char __pyx_k_sock_recvfrom[] = "sock_recvfrom";
static const char __pyx_k_start_serving[] = "start_serving";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_CancelledError[] = "CancelledError";
static const char __pyx_k_EAI_ADDRFAMILY[] = "EAI_ADDRFAMILY";
static const char __pyx_k_Loop_call_soon[] = "Loop.call_soon";
static const char __pyx_k_Loop_get_debug[] = "Loop.get_debug";
static const char __pyx_k_Loop_is_closed[] = "Loop.is_closed";
static const char __pyx_k_Loop_set_debug[] = "Loop.set_debug";
static const char __pyx_k_Loop_sock_recv[] = "Loop.sock_recv";
static const char __pyx_k_Loop_start_tls[] = "Loop.start_tls";
static const char __pyx_k_Process_memory[] = "Process memory:            {}";
static const char __pyx_k_Read_callbacks[] = "Read callbacks:            {}";
static const char __pyx_k_RuntimeWarning[] = "RuntimeWarning";
static const char __pyx_k_Server___aexit[] = "Server.__aexit__";
static const char __pyx_k_SockAddrHolder[] = "SockAddrHolder";
static const char __pyx_k_UDPSendContext[] = "_UDPSendContext";
static const char __pyx_k_UVSocketHandle[] = "UVSocketHandle";
static const char __pyx_k_UVStreamServer[] = "UVStreamServer";
static const char __pyx_k_UVStream_write[] = "UVStream.write";
static const char __pyx_k_asyncio_Handle[] = "asyncio.Handle";
static const char __pyx_k_buffer_updated[] = "buffer_updated";
static const char __pyx_k_check_hostname[] = "check_hostname";
static const char __pyx_k_cleanup_socket[] = "cleanup_socket";
static const char __pyx_k_connect_failed[] = "connect failed";
static const char __pyx_k_connect_method[] = "connect() method";
static const char __pyx_k_error_received[] = "error_received";
static const char __pyx_k_get_backend_id[] = "_get_backend_id";
static const char __pyx_k_get_event_loop[] = "get_event_loop";
static const char __pyx_k_get_extra_info[] = "get_extra_info";
static const char __pyx_k_get_returncode[] = "get_returncode";
static const char __pyx_k_old_agen_hooks[] = "old_agen_hooks";
static const char __pyx_k_process_exited[] = "process_exited";
static const char __pyx_k_r_received_EOF[] = "%r received EOF";
static const char __pyx_k_recvmsg_method[] = "recvmsg() method";
static const char __pyx_k_resume_reading[] = "resume_reading";
static const char __pyx_k_resume_writing[] = "resume_writing";
static const char __pyx_k_sendall_method[] = "sendall() method";
static const char __pyx_k_sendmsg_method[] = "sendmsg() method";
static const char __pyx_k_sock_recv_into[] = "sock_recv_into";
static const char __pyx_k_subprocess_run[] = "__subprocess_run";
static const char __pyx_k_AddrInfoRequest[] = "AddrInfoRequest";
static const char __pyx_k_BlockingIOError[] = "BlockingIOError";
static const char __pyx_k_BrokenPipeError[] = "BrokenPipeError";
static const char __pyx_k_FileExistsError[] = "FileExistsError";
static const char __pyx_k_FileSystemEvent[] = "FileSystemEvent";
static const char __pyx_k_Loop__sock_recv[] = "Loop._sock_recv";
static const char __pyx_k_Loop_add_reader[] = "Loop.add_reader";
static const char __pyx_k_Loop_add_writer[] = "Loop.add_writer";
static const char __pyx_k_Loop_call_later[] = "Loop.call_later";
static const char __pyx_k_Loop_debug_info[] = "--- Loop debug info: ---";
static const char __pyx_k_Loop_is_running[] = "Loop.is_running";
static const char __pyx_k_NameInfoRequest[] = "NameInfoRequest";
static const char __pyx_k_PermissionError[] = "PermissionError";
static const char __pyx_k_ResourceWarning[] = "ResourceWarning";
static const char __pyx_k_SSLSyscallError[] = "SSLSyscallError";
static const char __pyx_k_Server___aenter[] = "Server.__aenter__";
static const char __pyx_k_Server_get_loop[] = "Server.get_loop";
static const char __pyx_k_Shutdown_errors[] = "Shutdown errors            {}";
static const char __pyx_k_SubprocessError[] = "SubprocessError";
static const char __pyx_k_Timer_handles_8[] = "Timer handles:             {: <8} | {}";
static const char __pyx_k_UVBaseTransport[] = "UVBaseTransport";
static const char __pyx_k_allow_broadcast[] = "allow_broadcast";
static const char __pyx_k_asyncio_futures[] = "asyncio.futures";
static const char __pyx_k_collections_abc[] = "collections.abc";
static const char __pyx_k_connection_lost[] = "connection_lost";
static const char __pyx_k_connection_made[] = "connection_made";
static const char __pyx_k_get_inheritable[] = "get_inheritable";
static const char __pyx_k_lai_addr_static[] = "lai_addr_static";
static const char __pyx_k_makefile_method[] = "makefile() method";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_rai_addr_static[] = "rai_addr_static";
static const char __pyx_k_recvfrom_method[] = "recvfrom() method";
static const char __pyx_k_remove_reader_2[] = "__remove_reader";
static const char __pyx_k_remove_writer_2[] = "__remove_writer";
static const char __pyx_k_restore_signals[] = "restore_signals";
static const char __pyx_k_run_in_executor[] = "run_in_executor";
static const char __pyx_k_sendfile_method[] = "sendfile() method";
static const char __pyx_k_server_hostname[] = "server_hostname";
static const char __pyx_k_set_inheritable[] = "set_inheritable";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_shutdown_method[] = "shutdown() method";
static const char __pyx_k_subprocess_exec[] = "subprocess_exec";
static const char __pyx_k_uvloop_loop_pyx[] = "uvloop/loop.pyx";
static const char __pyx_k_uvloop_sslproto[] = "uvloop.sslproto";
static const char __pyx_k_CertificateError[] = "CertificateError";
static const char __pyx_k_FlowControlMixin[] = "_FlowControlMixin";
static const char __pyx_k_Handle_cancelled[] = "Handle.cancelled";
static const char __pyx_k_InterruptedError[] = "InterruptedError";
static const char __pyx_k_Loop__monitor_fs[] = "Loop._monitor_fs";
static const char __pyx_k_Loop__sighandler[] = "_Loop__sighandler";
static const char __pyx_k_Loop_create_task[] = "Loop.create_task";
static const char __pyx_k_Loop_getaddrinfo[] = "Loop.getaddrinfo";
static const char __pyx_k_Loop_getnameinfo[] = "Loop.getnameinfo";
static const char __pyx_k_Loop_run_forever[] = "Loop.run_forever";
static const char __pyx_k_Loop_sock_accept[] = "Loop.sock_accept";
static const char __pyx_k_Loop_sock_sendto[] = "Loop.sock_sendto";
static const char __pyx_k_Polls_debug_info[] = "--- Polls debug info: ---";
static const char __pyx_k_PseudoSocket_dup[] = "PseudoSocket.dup";
static const char __pyx_k_SSLWantReadError[] = "SSLWantReadError";
static const char __pyx_k_TimerHandle_when[] = "TimerHandle.when";
static const char __pyx_k_UVFSEvent_cancel[] = "UVFSEvent.cancel";
static const char __pyx_k_Write_contexts_8[] = "Write contexts:            {: <8} | {}";
static const char __pyx_k_asyncio_sslproto[] = "asyncio.sslproto";
static const char __pyx_k_closed_reading_x[] = "<{} closed={} reading={} {:#x}>";
static const char __pyx_k_decref_socketios[] = "_decref_socketios";
static const char __pyx_k_get_running_loop[] = "_get_running_loop";
static const char __pyx_k_get_task_factory[] = "get_task_factory";
static const char __pyx_k_on_ssl_connected[] = "__on_ssl_connected";
static const char __pyx_k_print_debug_info[] = "print_debug_info";
static const char __pyx_k_protocol_factory[] = "protocol_factory";
static const char __pyx_k_recv_into_method[] = "recv_into() method";
static const char __pyx_k_set_running_loop[] = "_set_running_loop";
static const char __pyx_k_set_task_factory[] = "set_task_factory";
static const char __pyx_k_source_traceback[] = "source_traceback";
static const char __pyx_k_subprocess_shell[] = "subprocess_shell";
static const char __pyx_k_test_coroutine_1[] = "_test_coroutine_1";
static const char __pyx_k_transport_waiter[] = "transport_waiter";
static const char __pyx_k_unix_socket_path[] = "unix socket path ";
static const char __pyx_k_FileNotFoundError[] = "FileNotFoundError";
static const char __pyx_k_InvalidStateError[] = "InvalidStateError";
static const char __pyx_k_IsADirectoryError[] = "IsADirectoryError";
static const char __pyx_k_KeyboardInterrupt[] = "KeyboardInterrupt";
static const char __pyx_k_Loop___sighandler[] = "Loop.__sighandler";
static const char __pyx_k_Loop__do_shutdown[] = "Loop._do_shutdown";
static const char __pyx_k_Loop__sock_accept[] = "Loop._sock_accept";
static const char __pyx_k_Loop_sock_connect[] = "Loop.sock_connect";
static const char __pyx_k_Loop_sock_sendall[] = "Loop.sock_sendall";
static const char __pyx_k_Number_of_signals[] = "Number of signals:         {}";
static const char __pyx_k_PseudoSocket_bind[] = "PseudoSocket.bind";
static const char __pyx_k_PseudoSocket_recv[] = "PseudoSocket.recv";
static const char __pyx_k_PseudoSocket_send[] = "PseudoSocket.send";
static const char __pyx_k_ReadUnixTransport[] = "ReadUnixTransport";
static const char __pyx_k_SSL_ERROR_SYSCALL[] = "SSL_ERROR_SYSCALL";
static const char __pyx_k_Server_is_serving[] = "Server.is_serving";
static const char __pyx_k_Socket_try_writes[] = "Socket try-writes:         {}";
static const char __pyx_k_TCPConnectRequest[] = "_TCPConnectRequest";
static const char __pyx_k_add_done_callback[] = "add_done_callback";
static const char __pyx_k_bufsize_must_be_0[] = "bufsize must be 0";
static const char __pyx_k_connect_ex_method[] = "connect_ex() method";
static const char __pyx_k_connect_read_pipe[] = "connect_read_pipe";
static const char __pyx_k_create_connection[] = "create_connection";
static const char __pyx_k_datagram_received[] = "datagram_received";
static const char __pyx_k_get_app_transport[] = "_get_app_transport";
static const char __pyx_k_libuv_get_version[] = "libuv_get_version";
static const char __pyx_k_return_exceptions[] = "return_exceptions";
static const char __pyx_k_start_new_session[] = "start_new_session";
static const char __pyx_k_unexpected_kwargs[] = "unexpected kwargs: {}";
static const char __pyx_k_uvloop_server_pyx[] = "uvloop/server.pyx";
static const char __pyx_k_Callback_handles_8[] = "Callback handles:          {: <8} | {}";
static const char __pyx_k_DeprecationWarning[] = "DeprecationWarning";
static const char __pyx_k_Loop__sock_connect[] = "Loop._sock_connect";
static const char __pyx_k_Loop__sock_sendall[] = "Loop._sock_sendall";
static const char __pyx_k_Loop_create_future[] = "Loop.create_future";
static const char __pyx_k_Loop_create_server[] = "Loop.create_server";
static const char __pyx_k_Loop_remove_reader[] = "Loop.remove_reader";
static const char __pyx_k_Loop_remove_writer[] = "Loop.remove_writer";
static const char __pyx_k_Loop_sock_recvfrom[] = "Loop.sock_recvfrom";
static const char __pyx_k_PYTHONASYNCIODEBUG[] = "PYTHONASYNCIODEBUG";
static const char __pyx_k_PipeConnectRequest[] = "_PipeConnectRequest";
static const char __pyx_k_ProcessLookupError[] = "ProcessLookupError";
static const char __pyx_k_PseudoSocket_close[] = "PseudoSocket.close";
static const char __pyx_k_PseudoSocket_ioctl[] = "PseudoSocket.ioctl";
static const char __pyx_k_PseudoSocket_share[] = "PseudoSocket.share";
static const char __pyx_k_SSLZeroReturnError[] = "SSLZeroReturnError";
static const char __pyx_k_Server_wait_closed[] = "Server.wait_closed";
static const char __pyx_k_StreamWriteContext[] = "_StreamWriteContext";
static const char __pyx_k_Streams_debug_info[] = "--- Streams debug info: ---";
static const char __pyx_k_ThreadPoolExecutor[] = "ThreadPoolExecutor";
static const char __pyx_k_TimerHandle_cancel[] = "TimerHandle.cancel";
static const char __pyx_k_UVProcessTransport[] = "UVProcessTransport";
static const char __pyx_k_UVStream_write_eof[] = "UVStream.write_eof";
static const char __pyx_k_UnicodeEncodeError[] = "UnicodeEncodeError";
static const char __pyx_k_WriteUnixTransport[] = "WriteUnixTransport";
static const char __pyx_k_Write_without_poll[] = "Write without poll:        {}";
static const char __pyx_k_add_signal_handler[] = "add_signal_handler";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_asyncio_transports[] = "asyncio.transports";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_concurrent_futures[] = "concurrent.futures";
static const char __pyx_k_connect_write_pipe[] = "connect_write_pipe";
static const char __pyx_k_create_unix_server[] = "create_unix_server";
static const char __pyx_k_get_asyncgen_hooks[] = "get_asyncgen_hooks";
static const char __pyx_k_get_pipe_transport[] = "get_pipe_transport";
static const char __pyx_k_ignore_environment[] = "ignore_environment";
static const char __pyx_k_pipe_data_received[] = "pipe_data_received";
static const char __pyx_k_run_until_complete[] = "run_until_complete";
static const char __pyx_k_set_asyncgen_hooks[] = "set_asyncgen_hooks";
static const char __pyx_k_shell_must_be_True[] = "shell must be True";
static const char __pyx_k_shutdown_asyncgens[] = "shutdown_asyncgens";
static const char __pyx_k_sig_out_of_range_1[] = "sig {} out of range(1, {})";
static const char __pyx_k_sock_recvfrom_into[] = "sock_recvfrom_into";
static const char __pyx_k_universal_newlines[] = "universal_newlines";
static const char __pyx_k_unopened_TCPServer[] = "unopened TCPServer";
static const char __pyx_k_Connect_call_failed[] = "Connect call failed ";
static const char __pyx_k_Loop_sock_recv_into[] = "Loop.sock_recv_into";
static const char __pyx_k_Multiple_exceptions[] = "Multiple exceptions: {}";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_PseudoSocket___exit[] = "PseudoSocket.__exit__";
static const char __pyx_k_PseudoSocket_accept[] = "PseudoSocket.accept";
static const char __pyx_k_PseudoSocket_detach[] = "PseudoSocket.detach";
static const char __pyx_k_PseudoSocket_fileno[] = "PseudoSocket.fileno";
static const char __pyx_k_PseudoSocket_listen[] = "PseudoSocket.listen";
static const char __pyx_k_PseudoSocket_sendto[] = "PseudoSocket.sendto";
static const char __pyx_k_SSL_ERROR_WANT_READ[] = "SSL_ERROR_WANT_READ";
static const char __pyx_k_UDPTransport_sendto[] = "UDPTransport.sendto";
static const char __pyx_k_UVFSEvent_cancelled[] = "UVFSEvent.cancelled";
static const char __pyx_k_UVStream_is_reading[] = "UVStream.is_reading";
static const char __pyx_k_UVStream_writelines[] = "UVStream.writelines";
static const char __pyx_k_WeakValueDictionary[] = "WeakValueDictionary";
static const char __pyx_k_asyncio_base_events[] = "asyncio.base_events";
static const char __pyx_k_default_int_handler[] = "default_int_handler";
static const char __pyx_k_iscoroutinefunction[] = "iscoroutinefunction";
static const char __pyx_k_log_destroy_pending[] = "_log_destroy_pending";
static const char __pyx_k_recvmsg_into_method[] = "recvmsg_into() method";
static const char __pyx_k_shell_must_be_False[] = "shell must be False";
static const char __pyx_k_unclosed_resource_r[] = "unclosed resource {!r}; {}";
static const char __pyx_k_uvloop_sslproto_pyx[] = "uvloop/sslproto.pyx";
static const char __pyx_k_warn_on_full_buffer[] = "warn_on_full_buffer";
static const char __pyx_k_AbstractChildWatcher[] = "AbstractChildWatcher";
static const char __pyx_k_ConnectionResetError[] = "ConnectionResetError";
static const char __pyx_k_Event_loop_is_closed[] = "Event loop is closed";
static const char __pyx_k_Loop___reduce_cython[] = "Loop.__reduce_cython__";
static const char __pyx_k_Loop__get_backend_id[] = "Loop._get_backend_id";
static const char __pyx_k_Loop__read_from_self[] = "Loop._read_from_self";
static const char __pyx_k_Loop__sock_recv_into[] = "Loop._sock_recv_into";
static const char __pyx_k_Loop__subprocess_run[] = "_Loop__subprocess_run";
static const char __pyx_k_Loop_run_in_executor[] = "Loop.run_in_executor";
static const char __pyx_k_Loop_subprocess_exec[] = "Loop.subprocess_exec";
static const char __pyx_k_PseudoSocket___enter[] = "PseudoSocket.__enter__";
static const char __pyx_k_PseudoSocket_connect[] = "PseudoSocket.connect";
static const char __pyx_k_PseudoSocket_recvmsg[] = "PseudoSocket.recvmsg";
static const char __pyx_k_PseudoSocket_sendall[] = "PseudoSocket.sendall";
static const char __pyx_k_PseudoSocket_sendmsg[] = "PseudoSocket.sendmsg";
static const char __pyx_k_SSLProtocolTransport[] = "_SSLProtocolTransport";
static const char __pyx_k_SSLProtocol__do_read[] = "SSLProtocol._do_read";
static const char __pyx_k_SSL_ERROR_WANT_WRITE[] = "SSL_ERROR_WANT_WRITE";
static const char __pyx_k_SSL_handshake_failed[] = "SSL handshake failed";
static const char __pyx_k_Server_serve_forever[] = "Server.serve_forever";
static const char __pyx_k_Server_start_serving[] = "Server.start_serving";
static const char __pyx_k_call_connection_made[] = "call_connection_made";
static const char __pyx_k_call_soon_threadsafe[] = "call_soon_threadsafe";
static const char __pyx_k_get_read_buffer_size[] = "get_read_buffer_size";
static const char __pyx_k_libuv_get_loop_t_ptr[] = "libuv_get_loop_t_ptr";
static const char __pyx_k_pipe_connection_lost[] = "pipe_connection_lost";
static const char __pyx_k_protocol_is_required[] = "protocol is required";
static const char __pyx_k_recvfrom_into_method[] = "recvfrom_into() method";
static const char __pyx_k_remove_done_callback[] = "remove_done_callback";
static const char __pyx_k_running_closed_debug[] = "<{}.{} running={} closed={} debug={}>";
static const char __pyx_k_sendmsg_afalg_method[] = "sendmsg_afalg() method";
static const char __pyx_k_set_default_executor[] = "set_default_executor";
static const char __pyx_k_sig_cannot_be_caught[] = "sig {} cannot be caught";
static const char __pyx_k_ssl_shutdown_timeout[] = "ssl_shutdown_timeout";
static const char __pyx_k_uvloop_cbhandles_pyx[] = "uvloop/cbhandles.pyx";
static const char __pyx_k_AddrInfo_data_is_NULL[] = "AddrInfo.data is NULL";
static const char __pyx_k_Exception_in_callback[] = "Exception in callback {}";
static const char __pyx_k_Fatal_error_on_server[] = "Fatal error on server ";
static const char __pyx_k_Invalid_file_object_r[] = "Invalid file object: {!r}";
static const char __pyx_k_Loop___subprocess_run[] = "Loop.__subprocess_run";
static const char __pyx_k_Loop_get_task_factory[] = "Loop.get_task_factory";
static const char __pyx_k_Loop_set_task_factory[] = "Loop.set_task_factory";
static const char __pyx_k_Loop_subprocess_shell[] = "Loop.subprocess_shell";
static const char __pyx_k_PseudoSocket_makefile[] = "PseudoSocket.makefile";
static const char __pyx_k_PseudoSocket_recvfrom[] = "PseudoSocket.recvfrom";
static const char __pyx_k_PseudoSocket_sendfile[] = "PseudoSocket.sendfile";
static const char __pyx_k_PseudoSocket_shutdown[] = "PseudoSocket.shutdown";
static const char __pyx_k_Read_callbacks_failed[] = "Read callbacks failed:     {}";
static const char __pyx_k_Read_failed_callbacks[] = "Read failed callbacks:     {}";
static const char __pyx_k_TimerHandle_cancelled[] = "TimerHandle.cancelled";
static const char __pyx_k_UVBaseTransport_abort[] = "UVBaseTransport.abort";
static const char __pyx_k_UVBaseTransport_close[] = "UVBaseTransport.close";
static const char __pyx_k_get_coroutine_wrapper[] = "get_coroutine_wrapper";
static const char __pyx_k_get_exception_handler[] = "get_exception_handler";
static const char __pyx_k_get_write_buffer_size[] = "get_write_buffer_size";
static const char __pyx_k_getfilesystemencoding[] = "getfilesystemencoding";
static const char __pyx_k_invalid_waiter_object[] = "invalid waiter object ";
static const char __pyx_k_pyx_unpickle_LruCache[] = "__pyx_unpickle_LruCache";
static const char __pyx_k_remove_signal_handler[] = "remove_signal_handler";
static const char __pyx_k_set_coroutine_wrapper[] = "set_coroutine_wrapper";
static const char __pyx_k_set_exception_handler[] = "set_exception_handler";
static const char __pyx_k_ssl_handshake_timeout[] = "ssl_handshake_timeout";
static const char __pyx_k_unclosed_resource_r_2[] = "unclosed resource {!r}";
static const char __pyx_k_uvloop_pseudosock_pyx[] = "uvloop/pseudosock.pyx";
static const char __pyx_k_ConnectionAbortedError[] = "ConnectionAbortedError";
static const char __pyx_k_ConnectionRefusedError[] = "ConnectionRefusedError";
static const char __pyx_k_Handle___reduce_cython[] = "Handle.__reduce_cython__";
static const char __pyx_k_Loop___setstate_cython[] = "Loop.__setstate_cython__";
static const char __pyx_k_Loop_connect_read_pipe[] = "Loop.connect_read_pipe";
static const char __pyx_k_Loop_create_connection[] = "Loop.create_connection";
static const char __pyx_k_PseudoSocket_recv_into[] = "PseudoSocket.recv_into";
static const char __pyx_k_SSLProtocol_get_buffer[] = "SSLProtocol.get_buffer";
static const char __pyx_k_SSL_shutdown_timed_out[] = "SSL shutdown timed out";
static const char __pyx_k_Server___reduce_cython[] = "Server.__reduce_cython__";
static const char __pyx_k_SyncSocketReaderFuture[] = "_SyncSocketReaderFuture";
static const char __pyx_k_SyncSocketWriterFuture[] = "_SyncSocketWriterFuture";
static const char __pyx_k_UVIdle___reduce_cython[] = "UVIdle.__reduce_cython__";
static const char __pyx_k_UVPoll___reduce_cython[] = "UVPoll.__reduce_cython__";
static const char __pyx_k_UVStream_can_write_eof[] = "UVStream.can_write_eof";
static const char __pyx_k_UVStream_pause_reading[] = "UVStream.pause_reading";
static const char __pyx_k_Write_callbacks_failed[] = "Write callbacks failed:    {}";
static const char __pyx_k_Write_failed_callbacks[] = "Write failed callbacks:    {}";
static const char __pyx_k_call_exception_handler[] = "call_exception_handler";
static const char __pyx_k_cancel_request_is_NULL[] = "{}.cancel: .request is NULL";
static const char __pyx_k_check_default_executor[] = "_check_default_executor";
static const char __pyx_k_concurrent_allocations[] = "concurrent allocations";
static const char __pyx_k_create_default_context[] = "create_default_context";
static const char __pyx_k_create_unix_connection[] = "create_unix_connection";
static const char __pyx_k_get_read_buffer_limits[] = "get_read_buffer_limits";
static const char __pyx_k_invalid_Handle_cb_type[] = "invalid Handle.cb_type: {}";
static const char __pyx_k_r_starts_SSL_handshake[] = "%r starts SSL handshake";
static const char __pyx_k_set_read_buffer_limits[] = "set_read_buffer_limits";
static const char __pyx_k_uvloop_handles_tcp_pyx[] = "uvloop/handles/tcp.pyx";
static const char __pyx_k_uvloop_handles_udp_pyx[] = "uvloop/handles/udp.pyx";
static const char __pyx_k_Invalid_file_descriptor[] = "Invalid file descriptor: {}";
static const char __pyx_k_Loop_add_signal_handler[] = "Loop.add_signal_handler";
static const char __pyx_k_Loop_connect_write_pipe[] = "Loop.connect_write_pipe";
static const char __pyx_k_Loop_create_unix_server[] = "Loop.create_unix_server";
static const char __pyx_k_Loop_run_until_complete[] = "Loop.run_until_complete";
static const char __pyx_k_Loop_shutdown_asyncgens[] = "Loop.shutdown_asyncgens";
static const char __pyx_k_Loop_sock_recvfrom_into[] = "Loop.sock_recvfrom_into";
static const char __pyx_k_PseudoSocket___getstate[] = "PseudoSocket.__getstate__";
static const char __pyx_k_PseudoSocket_connect_ex[] = "PseudoSocket.connect_ex";
static const char __pyx_k_PseudoSocket_getsockopt[] = "PseudoSocket.getsockopt";
static const char __pyx_k_PseudoSocket_gettimeout[] = "PseudoSocket.gettimeout";
static const char __pyx_k_PseudoSocket_setsockopt[] = "PseudoSocket.setsockopt";
static const char __pyx_k_PseudoSocket_settimeout[] = "PseudoSocket.settimeout";
static const char __pyx_k_ReadSubprocessPipeProto[] = "ReadSubprocessPipeProto";
static const char __pyx_k_ReadUnixTransport_abort[] = "ReadUnixTransport.abort";
static const char __pyx_k_ReadUnixTransport_write[] = "ReadUnixTransport.write";
static const char __pyx_k_UVAsync___reduce_cython[] = "UVAsync.__reduce_cython__";
static const char __pyx_k_UVCheck___reduce_cython[] = "UVCheck.__reduce_cython__";
static const char __pyx_k_UVProcessTransport_kill[] = "UVProcessTransport.kill";
static const char __pyx_k_UVStream_resume_reading[] = "UVStream.resume_reading";
static const char __pyx_k_UVTimer___reduce_cython[] = "UVTimer.__reduce_cython__";
static const char __pyx_k_asyncgen_finalizer_hook[] = "_asyncgen_finalizer_hook";
static const char __pyx_k_asyncgen_firstiter_hook[] = "_asyncgen_firstiter_hook";
static const char __pyx_k_connect_accepted_socket[] = "connect_accepted_socket";
static const char __pyx_k_expected_asyncio_Future[] = ", expected asyncio.Future";
static const char __pyx_k_get_write_buffer_limits[] = "get_write_buffer_limits";
static const char __pyx_k_service_proto_not_found[] = "service/proto not found";
static const char __pyx_k_set_write_buffer_limits[] = "set_write_buffer_limits";
static const char __pyx_k_uvloop_handles_pipe_pyx[] = "uvloop/handles/pipe.pyx";
static const char __pyx_k_uvloop_sleep_after_fork[] = "__uvloop_sleep_after_fork";
static const char __pyx_k_AF_UNSPEC_in_DNS_results[] = "AF_UNSPEC in DNS results";
static const char __pyx_k_AddrInfo___reduce_cython[] = "AddrInfo.__reduce_cython__";
static const char __pyx_k_Fatal_error_on_transport[] = "Fatal error on transport ";
static const char __pyx_k_Handle___setstate_cython[] = "Handle.__setstate_cython__";
static const char __pyx_k_LruCache___reduce_cython[] = "LruCache.__reduce_cython__";
static const char __pyx_k_PseudoSocket_getpeername[] = "PseudoSocket.getpeername";
static const char __pyx_k_PseudoSocket_getsockname[] = "PseudoSocket.getsockname";
static const char __pyx_k_PseudoSocket_setblocking[] = "PseudoSocket.setblocking";
static const char __pyx_k_SSLProtocol_eof_received[] = "SSLProtocol.eof_received";
static const char __pyx_k_SSL_connection_is_closed[] = "SSL connection is closed";
static const char __pyx_k_Server___setstate_cython[] = "Server.__setstate_cython__";
static const char __pyx_k_UVHandle___reduce_cython[] = "UVHandle.__reduce_cython__";
static const char __pyx_k_UVIdle___setstate_cython[] = "UVIdle.__setstate_cython__";
static const char __pyx_k_UVPoll___setstate_cython[] = "UVPoll.__setstate_cython__";
static const char __pyx_k_UVProcessTransport__wait[] = "UVProcessTransport._wait";
static const char __pyx_k_UVProcessTransport_close[] = "UVProcessTransport.close";
static const char __pyx_k_UVStream___reduce_cython[] = "UVStream.__reduce_cython__";
static const char __pyx_k_WriteSubprocessPipeProto[] = "WriteSubprocessPipeProto";
static const char __pyx_k_context_manager_protocol[] = "context manager protocol";
static const char __pyx_k_create_datagram_endpoint[] = "create_datagram_endpoint";
static const char __pyx_k_exception_in_Task___repr[] = "exception in Task.__repr__";
static const char __pyx_k_loop__exec_queued_writes[] = "loop._exec_queued_writes";
static const char __pyx_k_set_wakeup_fd_1_failed_s[] = "set_wakeup_fd(-1) failed: %s";
static const char __pyx_k_sig_must_be_an_int_not_r[] = "sig must be an int, not {!r}";
static const char __pyx_k_Loop_call_soon_threadsafe[] = "Loop.call_soon_threadsafe";
static const char __pyx_k_Loop_set_default_executor[] = "Loop.set_default_executor";
static const char __pyx_k_PseudoSocket_recvmsg_into[] = "PseudoSocket.recvmsg_into";
static const char __pyx_k_Read_EOF_failed_callbacks[] = "Read EOF failed callbacks: {}";
static const char __pyx_k_SSLProtocol_pause_writing[] = "SSLProtocol.pause_writing";
static const char __pyx_k_TCPServer___reduce_cython[] = "TCPServer.__reduce_cython__";
static const char __pyx_k_UVAsync___setstate_cython[] = "UVAsync.__setstate_cython__";
static const char __pyx_k_UVCheck___setstate_cython[] = "UVCheck.__setstate_cython__";
static const char __pyx_k_UVFSEvent___reduce_cython[] = "UVFSEvent.__reduce_cython__";
static const char __pyx_k_UVProcess___reduce_cython[] = "UVProcess.__reduce_cython__";
static const char __pyx_k_UVRequest___reduce_cython[] = "UVRequest.__reduce_cython__";
static const char __pyx_k_UVTimer___setstate_cython[] = "UVTimer.__setstate_cython__";
static const char __pyx_k_abort_init__closed_is_set[] = "_abort_init: {}._closed is set";
static const char __pyx_k_abort_init__inited_is_set[] = "_abort_init: {}._inited is set";
static const char __pyx_k_default_exception_handler[] = "default_exception_handler";
static const char __pyx_k_pyx_unpickle_PseudoSocket[] = "__pyx_unpickle_PseudoSocket";
static const char __pyx_k_shutdown_default_executor[] = "shutdown_default_executor";
static const char __pyx_k_unexpected_address_family[] = "unexpected address family";
static const char __pyx_k_uvloop_handles_stream_pyx[] = "uvloop/handles/stream.pyx";
static const char __pyx_k_AddrInfo___setstate_cython[] = "AddrInfo.__setstate_cython__";
static const char __pyx_k_Error_calling_eof_received[] = "Error calling eof_received()";
static const char __pyx_k_Fatal_error_on_transport_2[] = "Fatal error on transport";
static const char __pyx_k_Loop___get___locals_lambda[] = "Loop.__get__.<locals>.<lambda>";
static const char __pyx_k_Loop_get_exception_handler[] = "Loop.get_exception_handler";
static const char __pyx_k_Loop_remove_signal_handler[] = "Loop.remove_signal_handler";
static const char __pyx_k_Loop_set_exception_handler[] = "Loop.set_exception_handler";
static const char __pyx_k_LruCache___setstate_cython[] = "LruCache.__setstate_cython__";
static const char __pyx_k_PseudoSocket_recvfrom_into[] = "PseudoSocket.recvfrom_into";
static const char __pyx_k_PseudoSocket_sendmsg_afalg[] = "PseudoSocket.sendmsg_afalg";
static const char __pyx_k_SSLProtocolTransport_abort[] = "_SSLProtocolTransport.abort";
static const char __pyx_k_SSLProtocolTransport_close[] = "_SSLProtocolTransport.close";
static const char __pyx_k_SSLProtocolTransport_write[] = "_SSLProtocolTransport.write";
static const char __pyx_k_SSLProtocol_buffer_updated[] = "SSLProtocol.buffer_updated";
static const char __pyx_k_SSLProtocol_resume_writing[] = "SSLProtocol.resume_writing";
static const char __pyx_k_UVBaseTransport_is_closing[] = "UVBaseTransport.is_closing";
static const char __pyx_k_UVHandle___setstate_cython[] = "UVHandle.__setstate_cython__";
static const char __pyx_k_UVProcessTransport_get_pid[] = "UVProcessTransport.get_pid";
static const char __pyx_k_UVStream___setstate_cython[] = "UVStream.__setstate_cython__";
static const char __pyx_k_UnixServer___reduce_cython[] = "UnixServer.__reduce_cython__";
static const char __pyx_k_close__handle_data_is_NULL[] = "{}._close: _handle.data is NULL";
static const char __pyx_k_is_alive_but__loop_is_None[] = "{} is alive, but _loop is None";
static const char __pyx_k_r_SSL_handshake_took_1f_ms[] = "%r: SSL handshake took %.1f ms";
static const char __pyx_k_test__append_write_backlog[] = "_test__append_write_backlog";
static const char __pyx_k_uvloop_handles_fsevent_pyx[] = "uvloop/handles/fsevent.pyx";
static const char __pyx_k_uvloop_handles_process_pyx[] = "uvloop/handles/process.pyx";
static const char __pyx_k_Address_r_is_already_in_use[] = "Address {!r} is already in use";
static const char __pyx_k_Executing_r_took_3f_seconds[] = "Executing %r took %.3f seconds";
static const char __pyx_k_Executing_s_took_3f_seconds[] = "Executing %s took %.3f seconds";
static const char __pyx_k_Fatal_error_on_SSL_protocol[] = "Fatal error on SSL protocol";
static const char __pyx_k_Loop_call_exception_handler[] = "Loop.call_exception_handler";
static const char __pyx_k_Loop_create_unix_connection[] = "Loop.create_unix_connection";
static const char __pyx_k_ReadUnixTransport_write_eof[] = "ReadUnixTransport.write_eof";
static const char __pyx_k_SSLProtocol___reduce_cython[] = "SSLProtocol.__reduce_cython__";
static const char __pyx_k_SSLProtocol_connection_lost[] = "SSLProtocol.connection_lost";
static const char __pyx_k_SSLProtocol_connection_made[] = "SSLProtocol.connection_made";
static const char __pyx_k_TCPServer___setstate_cython[] = "TCPServer.__setstate_cython__";
static const char __pyx_k_TCPTransport_get_extra_info[] = "TCPTransport.get_extra_info";
static const char __pyx_k_TimerHandle___reduce_cython[] = "TimerHandle.__reduce_cython__";
static const char __pyx_k_UVFSEvent___setstate_cython[] = "UVFSEvent.__setstate_cython__";
static const char __pyx_k_UVProcess___setstate_cython[] = "UVProcess.__setstate_cython__";
static const char __pyx_k_UVRequest___setstate_cython[] = "UVRequest.__setstate_cython__";
static const char __pyx_k_cancel_request_data_is_NULL[] = "{}.cancel: .request.data is NULL";
static const char __pyx_k_cannot_switch_state_from_to[] = "cannot switch state from {} to {}";
static const char __pyx_k_close_uv_is_closing_is_true[] = "{}._close: uv_is_closing() is true";
static const char __pyx_k_host_must_be_a_str_or_bytes[] = "host must be a str or bytes";
static const char __pyx_k_invalid__init_protocol_call[] = "invalid _init_protocol call";
static const char __pyx_k_Loop__check_default_executor[] = "Loop._check_default_executor";
static const char __pyx_k_Loop_connect_accepted_socket[] = "Loop.connect_accepted_socket";
static const char __pyx_k_PseudoSocket___reduce_cython[] = "PseudoSocket.__reduce_cython__";
static const char __pyx_k_PseudoSocket_get_inheritable[] = "PseudoSocket.get_inheritable";
static const char __pyx_k_PseudoSocket_set_inheritable[] = "PseudoSocket.set_inheritable";
static const char __pyx_k_ReadUnixTransport_writelines[] = "ReadUnixTransport.writelines";
static const char __pyx_k_SyncSocketReaderFuture__loop[] = "_SyncSocketReaderFuture__loop";
static const char __pyx_k_SyncSocketReaderFuture__sock[] = "_SyncSocketReaderFuture__sock";
static const char __pyx_k_SyncSocketWriterFuture__loop[] = "_SyncSocketWriterFuture__loop";
static const char __pyx_k_SyncSocketWriterFuture__sock[] = "_SyncSocketWriterFuture__sock";
static const char __pyx_k_TCPTransport___reduce_cython[] = "TCPTransport.__reduce_cython__";
static const char __pyx_k_UDPTransport___reduce_cython[] = "UDPTransport.__reduce_cython__";
static const char __pyx_k_UVBaseTransport__force_close[] = "UVBaseTransport._force_close";
static const char __pyx_k_UVBaseTransport_get_protocol[] = "UVBaseTransport.get_protocol";
static const char __pyx_k_UVBaseTransport_set_protocol[] = "UVBaseTransport.set_protocol";
static const char __pyx_k_UVHandles_python_libuv_total[] = "UVHandles               python | libuv   | total";
static const char __pyx_k_UVProcessTransport_terminate[] = "UVProcessTransport.terminate";
static const char __pyx_k_UnixServer___setstate_cython[] = "UnixServer.__setstate_cython__";
static const char __pyx_k_called_with_handle_data_NULL[] = "{} called with handle.data == NULL";
static const char __pyx_k_could_not_bind_to_local_addr[] = "could not bind to local_addr {}";
static const char __pyx_k_is_alive_but__handle_is_NULL[] = "{} is alive, but _handle is NULL";
static const char __pyx_k_socket_send_raised_exception[] = "socket.send() raised exception.";
static const char __pyx_k_startupinfo_is_not_supported[] = "startupinfo is not supported";
static const char __pyx_k_AF_INET_address_must_be_tuple[] = "AF_INET address must be tuple";
static const char __pyx_k_Handle_loop_is_None_in_Handle[] = "Handle.loop is None in Handle.__dealloc__";
static const char __pyx_k_IPv4_sockaddr_must_be_2_tuple[] = "IPv4 sockaddr must be 2 tuple";
static const char __pyx_k_Loop__asyncgen_finalizer_hook[] = "Loop._asyncgen_finalizer_hook";
static const char __pyx_k_Loop__asyncgen_firstiter_hook[] = "Loop._asyncgen_firstiter_hook";
static const char __pyx_k_Loop_create_datagram_endpoint[] = "Loop.create_datagram_endpoint";
static const char __pyx_k_Pyx_CFunc_6uvloop_4loop_4Loop[] = "__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled.<locals>.wrap";
static const char __pyx_k_SSLProtocol___setstate_cython[] = "SSLProtocol.__setstate_cython__";
static const char __pyx_k_SyncSocketReaderFuture___init[] = "_SyncSocketReaderFuture.__init__";
static const char __pyx_k_SyncSocketReaderFuture_cancel[] = "_SyncSocketReaderFuture.cancel";
static const char __pyx_k_SyncSocketWriterFuture___init[] = "_SyncSocketWriterFuture.__init__";
static const char __pyx_k_SyncSocketWriterFuture_cancel[] = "_SyncSocketWriterFuture.cancel";
static const char __pyx_k_TimerHandle___setstate_cython[] = "TimerHandle.__setstate_cython__";
static const char __pyx_k_UVProcessTransport_is_closing[] = "UVProcessTransport.is_closing";
static const char __pyx_k_UnixTransport___reduce_cython[] = "UnixTransport.__reduce_cython__";
static const char __pyx_k_all_args_must_be_str_or_bytes[] = "all args must be str or bytes";
static const char __pyx_k_could_not_bind_to_remote_addr[] = "could not bind to remote_addr {}";
static const char __pyx_k_dealloc____handle_is_NULL__cl[] = "{}.__dealloc__: _handle is NULL, _closed == 1";
static const char __pyx_k_does_not_have_a_reader_in_the[] = " does not have a reader in the _sock_recv callback";
static const char __pyx_k_does_not_have_a_writer_in_the[] = " does not have a writer in the _sock_sendall callback";
static const char __pyx_k_integer_argument_expected_got[] = "integer argument expected, got {}";
static const char __pyx_k_is_longer_than_107_characters[] = " is longer than 107 characters";
static const char __pyx_k_is_not_supported_by_start_tls[] = " is not supported by start_tls()";
static const char __pyx_k_protocol_pause_writing_failed[] = "protocol.pause_writing() failed";
static const char __pyx_k_r_fileno_returned_non_integer[] = "{!r}.fileno() returned non-integer";
static const char __pyx_k_start_init_can_only_be_called[] = "{}._start_init can only be called once";
static const char __pyx_k_AF_INET6_address_must_be_tuple[] = "AF_INET6 address must be tuple";
static const char __pyx_k_Cannot_serialize_socket_object[] = "Cannot serialize socket object";
static const char __pyx_k_Error_occurred_during_shutdown[] = "Error occurred during shutdown";
static const char __pyx_k_Loop_default_exception_handler[] = "Loop.default_exception_handler";
static const char __pyx_k_Loop_shutdown_default_executor[] = "Loop.shutdown_default_executor";
static const char __pyx_k_PseudoSocket___setstate_cython[] = "PseudoSocket.__setstate_cython__";
static const char __pyx_k_SSLProtocolTransport_can_write[] = "_SSLProtocolTransport.can_write_eof";
static const char __pyx_k_SSLProtocolTransport_get_extra[] = "_SSLProtocolTransport.get_extra_info";
static const char __pyx_k_SSLProtocolTransport_get_write[] = "_SSLProtocolTransport.get_write_buffer_limits";
static const char __pyx_k_SSLProtocolTransport_set_write[] = "_SSLProtocolTransport.set_write_buffer_limits";
static const char __pyx_k_SSLProtocolTransport_write_eof[] = "_SSLProtocolTransport.write_eof";
static const char __pyx_k_SSLProtocol__get_app_transport[] = "SSLProtocol._get_app_transport";
static const char __pyx_k_SockAddrHolder___reduce_cython[] = "SockAddrHolder.__reduce_cython__";
static const char __pyx_k_Sock_ops_successful_on_1st_try[] = "--- Sock ops successful on 1st try: ---";
static const char __pyx_k_SyncSocketReaderFuture__remove[] = "_SyncSocketReaderFuture__remove_reader";
static const char __pyx_k_SyncSocketWriterFuture__remove[] = "_SyncSocketWriterFuture__remove_writer";
static const char __pyx_k_TCPTransport___setstate_cython[] = "TCPTransport.__setstate_cython__";
static const char __pyx_k_UDPSendContext___reduce_cython[] = "_UDPSendContext.__reduce_cython__";
static const char __pyx_k_UDPTransport___setstate_cython[] = "UDPTransport.__setstate_cython__";
static const char __pyx_k_UVBaseTransport_get_extra_info[] = "UVBaseTransport.get_extra_info";
static const char __pyx_k_UVProcessTransport_send_signal[] = "UVProcessTransport.send_signal";
static const char __pyx_k_UVSocketHandle___reduce_cython[] = "UVSocketHandle.__reduce_cython__";
static const char __pyx_k_UVStreamServer___reduce_cython[] = "UVStreamServer.__reduce_cython__";
static const char __pyx_k_cannot_apply_subprocess_STDOUT[] = "cannot apply subprocess.STDOUT";
static const char __pyx_k_close__handle_data_is_not_UVHa[] = "{}._close: _handle.data is not UVHandle/self";
static const char __pyx_k_creationflags_is_not_supported[] = "creationflags is not supported";
static const char __pyx_k_fatal_sent_0_in_advance_uv_buf[] = "fatal: sent < 0 in advance_uv_buf";
static const char __pyx_k_protocol_resume_writing_failed[] = "protocol.resume_writing() failed";
static const char __pyx_k_sock_connect_cb_is_called_on_a[] = "_sock_connect_cb is called on a cancelled Future";
static const char __pyx_k_unclosed_transport_uvloop_loop[] = "unclosed transport <uvloop.loop._SSLProtocolTransport object>";
static const char __pyx_k_uv_close_handle_cb_handle_loop[] = "__uv_close_handle_cb: handle.loop is invalid";
static const char __pyx_k_AF_INET6_must_be_a_tuple_of_2_4[] = "AF_INET6 must be a tuple of 2-4 parameters: (host, port, flowinfo?, scope_id?)";
static const char __pyx_k_A_UDP_Socket_was_expected_got_r[] = "A UDP Socket was expected, got {!r}";
static const char __pyx_k_A_UNIX_Domain_Stream_Socket_was[] = "A UNIX Domain Stream Socket was expected, got {!r}";
static const char __pyx_k_AddrInfoRequest___reduce_cython[] = "AddrInfoRequest.__reduce_cython__";
static const char __pyx_k_AddrInfoRequest_callback_called[] = "AddrInfoRequest callback called with NULL resolver.data";
static const char __pyx_k_Bad_exception_data_from_child_r[] = "Bad exception data from child: {!r}";
static const char __pyx_k_Cannot_run_the_event_loop_while[] = "Cannot run the event loop while another loop is running";
static const char __pyx_k_Invalid_address_must_be_None_or[] = "Invalid address: must be None or ";
static const char __pyx_k_NameInfoRequest___reduce_cython[] = "NameInfoRequest.__reduce_cython__";
static const char __pyx_k_Neither_host_port_nor_sock_were[] = "Neither host/port nor sock were specified";
static const char __pyx_k_Pickling_of_struct_members_such[] = "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)";
static const char __pyx_k_PipeConnectRequest___reduce_cyt[] = "_PipeConnectRequest.__reduce_cython__";
static const char __pyx_k_PipeConnectRequest___setstate_c[] = "_PipeConnectRequest.__setstate_cython__";
static const char __pyx_k_ReadUnixTransport_can_write_eof[] = "ReadUnixTransport.can_write_eof";
static const char __pyx_k_SSLProtocolTransport___reduce_c[] = "_SSLProtocolTransport.__reduce_cython__";
static const char __pyx_k_SSLProtocolTransport___setstate[] = "_SSLProtocolTransport.__setstate_cython__";
static const char __pyx_k_SSLProtocolTransport__force_clo[] = "_SSLProtocolTransport._force_close";
static const char __pyx_k_SSLProtocolTransport__test__app[] = "_SSLProtocolTransport._test__append_write_backlog";
static const char __pyx_k_SSLProtocolTransport_get_protoc[] = "_SSLProtocolTransport.get_protocol";
static const char __pyx_k_SSLProtocolTransport_get_read_b[] = "_SSLProtocolTransport.get_read_buffer_limits";
static const char __pyx_k_SSLProtocolTransport_is_closing[] = "_SSLProtocolTransport.is_closing";
static const char __pyx_k_SSLProtocolTransport_is_reading[] = "_SSLProtocolTransport.is_reading";
static const char __pyx_k_SSLProtocolTransport_pause_read[] = "_SSLProtocolTransport.pause_reading";
static const char __pyx_k_SSLProtocolTransport_resume_rea[] = "_SSLProtocolTransport.resume_reading";
static const char __pyx_k_SSLProtocolTransport_set_protoc[] = "_SSLProtocolTransport.set_protocol";
static const char __pyx_k_SSLProtocolTransport_set_read_b[] = "_SSLProtocolTransport.set_read_buffer_limits";
static const char __pyx_k_SSLProtocolTransport_writelines[] = "_SSLProtocolTransport.writelines";
static const char __pyx_k_StreamWriteContext___reduce_cyt[] = "_StreamWriteContext.__reduce_cython__";
static const char __pyx_k_StreamWriteContext___setstate_c[] = "_StreamWriteContext.__setstate_cython__";
static const char __pyx_k_StreamWriteContext_close_py_buf[] = "_StreamWriteContext.close: py_bufs != NULL and py_bufs_sml_inuse is True";
static const char __pyx_k_StreamWriteContext_close_uv_buf[] = "_StreamWriteContext.close: uv_bufs != NULL and py_bufs_sml_inuse is True";
static const char __pyx_k_SyncSocketReaderFuture___remove[] = "_SyncSocketReaderFuture.__remove_reader";
static const char __pyx_k_SyncSocketWriterFuture___remove[] = "_SyncSocketWriterFuture.__remove_writer";
static const char __pyx_k_TCPConnectRequest___reduce_cyth[] = "_TCPConnectRequest.__reduce_cython__";
static const char __pyx_k_TCPConnectRequest___setstate_cy[] = "_TCPConnectRequest.__setstate_cython__";
static const char __pyx_k_The_reuse_address_parameter_has[] = "The *reuse_address* parameter has been deprecated as of 0.15.";
static const char __pyx_k_UDPSendContext___setstate_cytho[] = "_UDPSendContext.__setstate_cython__";
static const char __pyx_k_UVBaseTransport___reduce_cython[] = "UVBaseTransport.__reduce_cython__";
static const char __pyx_k_UVProcessTransport__stdio_inite[] = "_UVProcessTransport__stdio_inited";
static const char __pyx_k_UVProcessTransport_get_protocol[] = "UVProcessTransport.get_protocol";
static const char __pyx_k_UVProcessTransport_set_protocol[] = "UVProcessTransport.set_protocol";
static const char __pyx_k_UVStreamServer__on_ssl_connecte[] = "_UVStreamServer__on_ssl_connected";
static const char __pyx_k_UVStream__try_write_called_with[] = "UVStream._try_write called with data in uv buffers";
static const char __pyx_k_Unable_to_check_or_remove_stale[] = "Unable to check or remove stale UNIX socket %r: %r";
static const char __pyx_k_UnixTransport___setstate_cython[] = "UnixTransport.__setstate_cython__";
static const char __pyx_k_WriteSubprocessPipeProto___init[] = "WriteSubprocessPipeProto.__init__";
static const char __pyx_k_WriteSubprocessPipeProto___repr[] = "WriteSubprocessPipeProto.__repr__";
static const char __pyx_k_WriteSubprocessPipeProto_resume[] = "WriteSubprocessPipeProto.resume_writing";
static const char __pyx_k_asyncio_is_trying_to_install_it[] = "!!! asyncio is trying to install its ChildWatcher for SIGCHLD signal !!!\n\nThis is probably because a uvloop instance is used with asyncio.set_event_loop(). The correct way to use uvloop is to install its policy: `asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())`\n\n";
static const char __pyx_k_buffer_size_is_not_0_after_a_su[] = "_buffer_size is not 0 after a successful _exec_write";
static const char __pyx_k_called_after_destroying_the_UVH[] = "{} called after destroying the UVHandle";
static const char __pyx_k_cancel_request_data_is_not_UVRe[] = "{}.cancel: .request.data is not UVRequest";
static const char __pyx_k_cannot_add_a_signal_handler_for[] = "cannot add a signal handler for SIGCHLD: it is used by the event loop to track subprocesses";
static const char __pyx_k_cannot_run_TimerHandle_callback[] = "cannot run TimerHandle; callback is not set";
static const char __pyx_k_cannot_spawn_a_process_args_are[] = "cannot spawn a process: args are empty";
static const char __pyx_k_create_datagram_endpoint_locals[] = "create_datagram_endpoint.<locals>.genexpr";
static const char __pyx_k_deallocating_an_open_event_loop[] = "deallocating an open event loop";
static const char __pyx_k_does_not_have_a_reader_in_the_2[] = " does not have a reader in the _sock_recv_into callback";
static const char __pyx_k_does_not_have_a_writer_in_the_2[] = " does not have a writer in the _sock_connect_cb callback";
static const char __pyx_k_expected_AF_INET_AF_INET6_or_AF[] = "expected AF_INET, AF_INET6, or AF_UNIX family, got ";
static const char __pyx_k_getaddrinfo_returned_empty_list[] = "getaddrinfo() returned empty list";
static const char __pyx_k_getsockaddrarg_scope_id_must_be[] = "getsockaddrarg: scope_id must be unsigned 32 bit integer";
static const char __pyx_k_handle_loop_is_NULL_in___ensure[] = "handle.loop is NULL in __ensure_handle_data";
static const char __pyx_k_host_and_port_was_not_specified[] = "host and port was not specified and no sock specified";
static const char __pyx_k_is_alive_but__handle_data_is_no[] = "{} is alive, but _handle.data is not initialized";
static const char __pyx_k_is_alive_but__handle_loop_is_no[] = "{} is alive, but _handle.loop is not initialized";
static const char __pyx_k_is_already_being_awaited_on_ser[] = " is already being awaited on serve_forever()";
static const char __pyx_k_is_not_supposed_to_be_instantia[] = "{} is not supposed to be instantiated from Python";
static const char __pyx_k_is_open_in___dealloc___with_loo[] = "{} is open in __dealloc__ with loop set to NULL";
static const char __pyx_k_no_path_and_sock_were_specified[] = "no path and sock were specified";
static const char __pyx_k_port_must_be_a_str_bytes_or_int[] = "port must be a str, bytes or int";
static const char __pyx_k_protocol_is_not_set_cannot_call[] = "protocol is not set, cannot call connection_made()";
static const char __pyx_k_seconds_aborting_the_connection[] = " seconds: aborting the connection";
static const char __pyx_k_sock_recv_into_is_called_on_a_c[] = "_sock_recv_into is called on a cancelled Future";
static const char __pyx_k_sock_recv_is_called_on_a_cancel[] = "_sock_recv is called on a cancelled Future";
static const char __pyx_k_sock_sendall_is_called_on_a_can[] = "_sock_sendall is called on a cancelled Future";
static const char __pyx_k_sockaddr_must_be_a_tuple_of_2_3[] = "sockaddr must be a tuple of 2, 3 or 4 values";
static const char __pyx_k_socket_family_mismatch_or_a_DNS[] = ": socket family mismatch or a DNS lookup is required";
static const char __pyx_k_ssl_handshake_timeout_should_be[] = "ssl_handshake_timeout should be a positive number, got ";
static const char __pyx_k_sslcontext_is_expected_to_be_an[] = "sslcontext is expected to be an instance of ssl.SSLContext, got ";
static const char __pyx_k_task_factory_must_be_a_callable[] = "task factory must be a callable or None";
static const char __pyx_k_testhelper_unwrap_capsuled_poin[] = "_testhelper_unwrap_capsuled_pointer";
static const char __pyx_k_the_socket_must_be_non_blocking[] = "the socket must be non-blocking";
static const char __pyx_k_try_write_sent_all_data_and_ret[] = "_try_write sent all data and returned non-zero";
static const char __pyx_k_unable_to_start_the_loop_it_was[] = "unable to start the loop; it was closed";
static const char __pyx_k_uv_handle_t_current_freed_total[] = "uv_handle_t (current: {}; freed: {}; total: {})";
static const char __pyx_k_uvloop_PseudoSocket_fd_family_s[] = "<uvloop.PseudoSocket fd={}, family={!s}, type={!s}, proto={}";
static const char __pyx_k_uvloop_handles_streamserver_pyx[] = "uvloop/handles/streamserver.pyx";
static const char __pyx_k_without_no_gc_clear_loop_was_se[] = "{} without @no_gc_clear; loop was set to None by GC";
static const char __pyx_k_AF_INET_address_must_be_tuple_of[] = "AF_INET address must be tuple of (host, port)";
static const char __pyx_k_AF_UNIX_address_must_be_a_str_or[] = "AF_UNIX address must be a str or a bytes object";
static const char __pyx_k_A_Stream_Socket_was_expected_got[] = "A Stream Socket was expected, got {!r}";
static const char __pyx_k_A_callable_object_or_None_is_exp[] = "A callable object or None is expected, got {!r}";
static const char __pyx_k_AddrInfoRequest___setstate_cytho[] = "AddrInfoRequest.__setstate_cython__";
static const char __pyx_k_Cannot_call_write_after_write_eo[] = "Cannot call write() after write_eof()";
static const char __pyx_k_Cannot_call_writelines_after_wri[] = "Cannot call writelines() after write_eof()";
static const char __pyx_k_Cannot_close_a_running_event_loo[] = "Cannot close a running event loop";
static const char __pyx_k_Creating__SSLProtocolTransport_t[] = "Creating _SSLProtocolTransport twice";
static const char __pyx_k_Event_loop_stopped_before_Future[] = "Event loop stopped before Future completed.";
static const char __pyx_k_Exception_in___repr___r_value_ty[] = "Exception in __repr__ {!r}; value type: {!r}";
static const char __pyx_k_Exception_in_default_exception_h[] = "Exception in default exception handler";
static const char __pyx_k_Exception_occurred_in_preexec_fn[] = "Exception occurred in preexec_fn.";
static const char __pyx_k_Executor_shutdown_has_been_calle[] = "Executor shutdown has been called";
static const char __pyx_k_Fatal_write_error_on_datagram_tr[] = "Fatal write error on datagram transport";
static const char __pyx_k_File_descriptor_r_is_used_by_tra[] = "File descriptor {!r} is used by transport {!r}";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))";
static const char __pyx_k_Loop__getaddrinfo_locals_callbac[] = "Loop._getaddrinfo.<locals>.callback";
static const char __pyx_k_Loop__getnameinfo_locals_callbac[] = "Loop._getnameinfo.<locals>.callback";
static const char __pyx_k_NameInfoRequest___setstate_cytho[] = "NameInfoRequest.__setstate_cython__";
static const char __pyx_k_Non_thread_safe_operation_invoke[] = "Non-thread-safe operation invoked on an event loop other than the current one";
static const char __pyx_k_Object_created_at_most_recent_ca[] = "Object created at (most recent call last):\n";
static const char __pyx_k_Passing_reuse_address_True_is_no[] = "Passing `reuse_address=True` is no longer supported, as the usage of SO_REUSEPORT in UDP poses a significant security concern.";
static const char __pyx_k_Racing_with_another_loop_to_spaw[] = "Racing with another loop to spawn a process.";
static const char __pyx_k_ReadSubprocessPipeProto_data_rec[] = "ReadSubprocessPipeProto.data_received";
static const char __pyx_k_ReadUnixTransport___reduce_cytho[] = "ReadUnixTransport.__reduce_cython__";
static const char __pyx_k_ReadUnixTransport___setstate_cyt[] = "ReadUnixTransport.__setstate_cython__";
static const char __pyx_k_ReadUnixTransport_get_write_buff[] = "ReadUnixTransport.get_write_buffer_limits";
static const char __pyx_k_ReadUnixTransport_set_write_buff[] = "ReadUnixTransport.set_write_buffer_limits";
static const char __pyx_k_SSLProtocolTransport_get_write_2[] = "_SSLProtocolTransport.get_write_buffer_size";
static const char __pyx_k_SSLProtocol__start_handshake_loc[] = "SSLProtocol._start_handshake.<locals>.<lambda>";
static const char __pyx_k_SSLProtocol__start_shutdown_loca[] = "SSLProtocol._start_shutdown.<locals>.<lambda>";
static const char __pyx_k_SSL_handshake_failed_on_verifyin[] = "SSL handshake failed on verifying the certificate";
static const char __pyx_k_SSL_handshake_is_taking_longer_t[] = "SSL handshake is taking longer than ";
static const char __pyx_k_Server_side_SSL_needs_a_valid_SS[] = "Server side SSL needs a valid SSLContext";
static const char __pyx_k_SockAddrHolder___setstate_cython[] = "SockAddrHolder.__setstate_cython__";
static const char __pyx_k_The_executor_did_not_finishing_j[] = "The executor did not finishing joining its threads within ";
static const char __pyx_k_UDPTransport_family_is_undefined[] = "UDPTransport.family is undefined; cannot create python socket";
static const char __pyx_k_UVBaseTransport___setstate_cytho[] = "UVBaseTransport.__setstate_cython__";
static const char __pyx_k_UVBaseTransport_get_write_buffer[] = "UVBaseTransport.get_write_buffer_size";
static const char __pyx_k_UVBaseTransport_set_write_buffer[] = "UVBaseTransport.set_write_buffer_limits";
static const char __pyx_k_UVProcessTransport___reduce_cyth[] = "UVProcessTransport.__reduce_cython__";
static const char __pyx_k_UVProcessTransport___setstate_cy[] = "UVProcessTransport.__setstate_cython__";
static const char __pyx_k_UVProcessTransport___stdio_inite[] = "UVProcessTransport.__stdio_inited";
static const char __pyx_k_UVProcessTransport__call_connect[] = "UVProcessTransport._call_connection_made";
static const char __pyx_k_UVProcessTransport_get_extra_inf[] = "UVProcessTransport.get_extra_info";
static const char __pyx_k_UVProcessTransport_get_pipe_tran[] = "UVProcessTransport.get_pipe_transport";
static const char __pyx_k_UVProcessTransport_get_returncod[] = "UVProcessTransport.get_returncode";
static const char __pyx_k_UVProcess__close_after_spawn_cal[] = "UVProcess._close_after_spawn called after uv_spawn";
static const char __pyx_k_UVSocketHandle___setstate_cython[] = "UVSocketHandle.__setstate_cython__";
static const char __pyx_k_UVStreamServer___on_ssl_connecte[] = "UVStreamServer.__on_ssl_connected";
static const char __pyx_k_UVStreamServer___setstate_cython[] = "UVStreamServer.__setstate_cython__";
static const char __pyx_k_UVStream_shutdown_callback_calle[] = "UVStream.shutdown callback called with NULL req.data, status=%r";
static const char __pyx_k_UVStream_write_callback_called_w[] = "UVStream.write callback called with NULL req.data, status=%r";
static const char __pyx_k_UVTransport__call_connection_los[] = "UVTransport._call_connection_lost";
static const char __pyx_k_UVTransport__call_connection_mad[] = "UVTransport._call_connection_made";
static const char __pyx_k_Unable_to_clean_up_listening_UNI[] = "Unable to clean up listening UNIX socket %r: %r";
static const char __pyx_k_Unhandled_error_in_exception_han[] = "Unhandled error in exception handler";
static const char __pyx_k_Unhandled_exception_in_event_loo[] = "Unhandled exception in event loop";
static const char __pyx_k_WriteSubprocessPipeProto_connect[] = "WriteSubprocessPipeProto.connection_made";
static const char __pyx_k_WriteSubprocessPipeProto_pause_w[] = "WriteSubprocessPipeProto.pause_writing";
static const char __pyx_k_WriteUnixTransport___reduce_cyth[] = "WriteUnixTransport.__reduce_cython__";
static const char __pyx_k_WriteUnixTransport___setstate_cy[] = "WriteUnixTransport.__setstate_cython__";
static const char __pyx_k_WriteUnixTransport_pause_reading[] = "WriteUnixTransport.pause_reading";
static const char __pyx_k_WriteUnixTransport_resume_readin[] = "WriteUnixTransport.resume_reading";
static const char __pyx_k_XXX_an_allocated_buffer_in_trans[] = "XXX: an allocated buffer in transport._close()";
static const char __pyx_k_You_must_set_server_hostname_whe[] = "You must set server_hostname when using ssl without a host";
static const char __pyx_k_active_TimerHandle_is_deallacati[] = "active TimerHandle is deallacating";
static const char __pyx_k_add_signal_handler_can_only_be_c[] = "add_signal_handler() can only be called from the main thread";
static const char __pyx_k_all_environment_values_must_be_b[] = "all environment values must be bytes or str";
static const char __pyx_k_all_environment_vars_must_be_byt[] = "all environment vars must be bytes or str";
static const char __pyx_k_an_error_occurred_during_closing[] = "an error occurred during closing of asynchronous generator {!r}";
static const char __pyx_k_asynchronous_generator_r_was_sch[] = "asynchronous generator {!r} was scheduled after loop.shutdown_asyncgens() call";
static const char __pyx_k_cannot_cleanup_signal_handlers_c[] = "cannot cleanup signal handlers: closing the event loop in a non-main OS thread";
static const char __pyx_k_cannot_convert_sockaddr_into_Pyt[] = "cannot convert sockaddr into Python object";
static const char __pyx_k_cannot_open_a_UDP_handle_invalid[] = "cannot open a UDP handle, invalid family {}";
static const char __pyx_k_cannot_pause_signals_handling_no[] = "cannot pause signals handling; no longer running in the main thread";
static const char __pyx_k_cannot_run_Handle_callback_is_no[] = "cannot run Handle; callback is not set";
static const char __pyx_k_cannot_shutdown_signals_handling[] = "cannot shutdown signals handling as it has not been paused";
static const char __pyx_k_closed_Transport_handle_and_unse[] = "closed Transport handle and unset waiter";
static const char __pyx_k_coroutines_cannot_be_used_with_a[] = "coroutines cannot be used with add_signal_handler()";
static const char __pyx_k_coroutines_cannot_be_used_with_r[] = "coroutines cannot be used with run_in_executor()";
static const char __pyx_k_could_not_close_attached_file_ob[] = "could not close attached file object ";
static const char __pyx_k_create_connection_locals_genexpr[] = "create_connection.<locals>.genexpr";
static const char __pyx_k_create_server_failed_to_create_s[] = "create_server() failed to create socket.socket(%r, %r, %r)";
static const char __pyx_k_cwd_must_be_a_str_or_bytes_objec[] = "cwd must be a str or bytes object";
static const char __pyx_k_data_expecting_a_bytes_like_inst[] = "data: expecting a bytes-like instance, got ";
static const char __pyx_k_deallocating_a_running_event_loo[] = "deallocating a running event loop!";
static const char __pyx_k_error_status_in_uv_stream_t_list[] = "error status in uv_stream_t.listen callback";
static const char __pyx_k_error_status_in_uv_stream_t_read[] = "error status in uv_stream_t.read callback";
static const char __pyx_k_error_status_in_uv_stream_t_shut[] = "error status in uv_stream_t.shutdown callback";
static const char __pyx_k_error_status_in_uv_stream_t_writ[] = "error status in uv_stream_t.write callback";
static const char __pyx_k_error_while_attempting_to_bind_o[] = "error while attempting to bind on address ";
static const char __pyx_k_fatal_Could_not_advance__StreamW[] = "fatal: Could not advance _StreamWriteContext";
static const char __pyx_k_get_coroutine_origin_tracking_de[] = "get_coroutine_origin_tracking_depth";
static const char __pyx_k_getnameinfo_argument_1_must_be_a[] = "getnameinfo() argument 1 must be a tuple";
static const char __pyx_k_getnameinfo_flowinfo_must_be_0_1[] = "getnameinfo(): flowinfo must be 0-1048575.";
static const char __pyx_k_handle_data_is_NULL_in___close_a[] = "handle.data is NULL in __close_all_handles_cb";
static const char __pyx_k_handle_loop_data_is_NULL_in___en[] = "handle.loop.data is NULL in __ensure_handle_data";
static const char __pyx_k_has_handle_0_in___uv_close_handl[] = "has_handle=0 in __uv_close_handle_cb";
static const char __pyx_k_host_must_be_a_string_or_bytes_o[] = "host must be a string or bytes object";
static const char __pyx_k_host_port_and_sock_can_not_be_sp[] = "host/port and sock can not be specified at the same time";
static const char __pyx_k_invalid_data_in_writebuf_an_inst[] = "invalid data in writebuf: an instance of bytes, bytearray or memoryview was expected, got {}";
static const char __pyx_k_invalid_socket_family_expected_A[] = "invalid socket family, expected AF_UNIX, AF_INET or AF_INET6";
static const char __pyx_k_local_addr_must_be_a_tuple_of_ho[] = "local_addr must be a tuple of host and port";
static const char __pyx_k_maxsize_is_expected_to_be_greate[] = "maxsize is expected to be greater than 0, got ";
static const char __pyx_k_new_callbacks_were_queued_during[] = "new callbacks were queued during loop closing: ";
static const char __pyx_k_new_poll_handles_were_queued_dur[] = "new poll handles were queued during loop closing: ";
static const char __pyx_k_new_timers_were_queued_during_lo[] = "new timers were queued during loop closing: ";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_no_python_buffer_is_allocated_in[] = "no python buffer is allocated in on_read; nread=";
static const char __pyx_k_object_created_at_most_recent_ca[] = "object created at (most recent call last):\n{}";
static const char __pyx_k_open__StreamWriteContext_is_bein[] = "open _StreamWriteContext is being deallocated";
static const char __pyx_k_open__UDPSendContext_is_being_de[] = "open _UDPSendContext is being deallocated";
static const char __pyx_k_path_and_sock_can_not_be_specifi[] = "path and sock can not be specified at the same time";
static const char __pyx_k_path_was_not_specified_and_no_so[] = "path was not specified, and no sock specified";
static const char __pyx_k_remote_addr_must_be_a_tuple_of_h[] = "remote_addr must be a tuple of (host, port)";
static const char __pyx_k_remove_signal_handler_can_only_b[] = "remove_signal_handler() can only be called from the main thread";
static const char __pyx_k_returning_true_from_eof_received[] = "returning true from eof_received() has no effect when using ssl";
static const char __pyx_k_reuse_port_not_supported_by_sock[] = "reuse_port not supported by socket module";
static const char __pyx_k_run_until_complete_locals_done_c[] = "run_until_complete.<locals>.done_cb";
static const char __pyx_k_self_pipe_exists_before_loop_run[] = "self-pipe exists before loop run";
static const char __pyx_k_self_pipe_was_not_cleaned_up_aft[] = "self-pipe was not cleaned up after loop was run";
static const char __pyx_k_self_py_buf_self_req_cannot_be_c[] = "self.py_buf,self.req cannot be converted to a Python object for pickling";
static const char __pyx_k_self_py_bufs_self_py_bufs_sml_se[] = "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling";
static const char __pyx_k_server_hostname_is_only_meaningf[] = "server_hostname is only meaningful with ssl";
static const char __pyx_k_set_coroutine_origin_tracking_de[] = "set_coroutine_origin_tracking_depth";
static const char __pyx_k_setblocking_transport_sockets_ca[] = "setblocking(): transport sockets cannot be blocking";
static const char __pyx_k_settimeout_only_0_timeout_is_all[] = "settimeout(): only 0 timeout is allowed on transport sockets";
static const char __pyx_k_signals_handling_has_been_alread[] = "signals handling has been already setup";
static const char __pyx_k_signals_handling_has_not_been_se[] = "signals handling has not been setup";
static const char __pyx_k_sockaddr_resolved_to_multiple_ad[] = "sockaddr resolved to multiple addresses";
static const char __pyx_k_socket_modifier_keyword_argument[] = "socket modifier keyword arguments can not be used when sock is specified. ({})";
static const char __pyx_k_ssl_argument_must_be_an_SSLConte[] = "ssl argument must be an SSLContext or None";
static const char __pyx_k_ssl_handshake_timeout_is_only_me[] = "ssl_handshake_timeout is only meaningful with ssl";
static const char __pyx_k_ssl_is_expected_to_be_None_or_an[] = "ssl is expected to be None or an instance of ssl.SSLContext, got {!r}";
static const char __pyx_k_ssl_shutdown_timeout_is_only_mea[] = "ssl_shutdown_timeout is only meaningful with ssl";
static const char __pyx_k_ssl_shutdown_timeout_should_be_a[] = "ssl_shutdown_timeout should be a positive number, got ";
static const char __pyx_k_stream_is_open_after_UVStream__t[] = "stream is open after UVStream._try_write returned None";
static const char __pyx_k_subprocess_STDOUT_is_supported_o[] = "subprocess.STDOUT is supported only by stderr parameter";
static const char __pyx_k_this_event_loop_is_already_runni[] = "this event loop is already running.";
static const char __pyx_k_transport_sockets_do_not_support[] = "transport sockets do not support {}";
static const char __pyx_k_unable_to_listen_no_protocol_fac[] = "unable to listen(); no protocol_factory";
static const char __pyx_k_unable_to_perform_operation_on_r[] = "unable to perform operation on {!r}; the handler is closed";
static const char __pyx_k_unclosed_resource_r_could_not_se[] = "unclosed resource {!r}; could not serialize debug traceback: {}: {}";
static const char __pyx_k_unhandled_error_or_an_empty_buff[] = "unhandled error (or an empty buffer) in get_buffer()";
static const char __pyx_k_universal_newlines_must_be_False[] = "universal_newlines must be False";
static const char __pyx_k_uvloop_handles_basetransport_pyx[] = "uvloop/handles/basetransport.pyx";
static const char __pyx_k_SSLProtocolTransport_get_read_b_2[] = "_SSLProtocolTransport.get_read_buffer_size";
static const char __pyx_k_getaddrinfo_returned_empty_list_2[] = "getaddrinfo() returned empty list for local_addr";
static const char __pyx_k_Exception_in_default_exception_h_2[] = "Exception in default exception handler while handling an unexpected error in custom exception handler";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))";
static const char __pyx_k_ReadUnixTransport_get_write_buff_2[] = "ReadUnixTransport.get_write_buffer_size";
static const char __pyx_k_UDPTransport_family_is_undefined_2[] = "UDPTransport.family is undefined; cannot send";
static const char __pyx_k_UVBaseTransport_get_write_buffer_2[] = "UVBaseTransport.get_write_buffer_limits";
static const char __pyx_k_WriteSubprocessPipeProto_connect_2[] = "WriteSubprocessPipeProto.connection_lost";
static const char __pyx_k_local_addr_must_be_a_tuple_of_ho_2[] = "local_addr must be a tuple of (host, port)";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_wrap(PyObject *__pyx_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, int __pyx_v_enabled); /* proto */
static int __pyx_pf_6uvloop_4loop_4Loop___cinit__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static int __pyx_pf_6uvloop_4loop_4Loop_2__init__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static void __pyx_pf_6uvloop_4loop_4Loop_4__dealloc__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_6__sighandler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_signum, CYTHON_UNUSED PyObject *__pyx_v_frame); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_12_getaddrinfo_callback(PyObject *__pyx_self, PyObject *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_12_getnameinfo_callback(PyObject *__pyx_self, PyObject *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_8_get_backend_id(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_16print_debug_info___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_10__repr__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_12call_soon(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_14call_soon_threadsafe(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_16call_later(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_delay, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_18call_at(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_when, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_20time(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_22stop(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_24run_forever(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_26close(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_28get_debug(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_30set_debug(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_enabled); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_32is_running(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_34is_closed(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_36create_future(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_38create_task(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_coro, PyObject *__pyx_v_name, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_40set_task_factory(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_factory); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_42get_task_factory(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_18run_until_complete_done_cb(PyObject *__pyx_self, PyObject *__pyx_v_fut); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_44run_until_complete(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_future); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_46getaddrinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_49getnameinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sockaddr, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_52start_tls(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_transport, PyObject *__pyx_v_protocol, PyObject *__pyx_v_sslcontext, PyObject *__pyx_v_server_side, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_55create_server(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_flags, PyObject *__pyx_v_sock, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_reuse_address, PyObject *__pyx_v_reuse_port, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout, PyObject *__pyx_v_start_serving); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_17create_connection_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_17create_connection_3genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_58create_connection(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_ssl, PyObject *__pyx_v_family, PyObject *__pyx_v_proto, PyObject *__pyx_v_flags, PyObject *__pyx_v_sock, PyObject *__pyx_v_local_addr, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_61create_unix_server(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_path, PyObject *__pyx_v_backlog, PyObject *__pyx_v_sock, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout, PyObject *__pyx_v_start_serving, PyObject *__pyx_v_cleanup_socket); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_64create_unix_connection(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_path, PyObject *__pyx_v_ssl, PyObject *__pyx_v_sock, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_67default_exception_handler(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_69get_exception_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_71set_exception_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_handler); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_73call_exception_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_75add_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, PyObject *__pyx_v_callback, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_77remove_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_79add_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, PyObject *__pyx_v_callback, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_81remove_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_83sock_recv(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_n); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_86sock_recv_into(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_buf); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_89sock_sendall(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_92sock_accept(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_95sock_connect(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_address); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_98sock_recvfrom(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_sock, CYTHON_UNUSED PyObject *__pyx_v_bufsize); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_101sock_recvfrom_into(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_sock, CYTHON_UNUSED PyObject *__pyx_v_buf, CYTHON_UNUSED PyObject *__pyx_v_nbytes); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_104sock_sendto(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_sock, CYTHON_UNUSED PyObject *__pyx_v_data, CYTHON_UNUSED PyObject *__pyx_v_address); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_107connect_accepted_socket(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_sock, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_110run_in_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_executor, PyObject *__pyx_v_func, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_112set_default_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_executor); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_114__subprocess_run(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_args, PyObject *__pyx_v_stdin, PyObject *__pyx_v_stdout, PyObject *__pyx_v_stderr, PyObject *__pyx_v_universal_newlines, CYTHON_UNUSED PyObject *__pyx_v_shell, PyObject *__pyx_v_bufsize, PyObject *__pyx_v_preexec_fn, CYTHON_UNUSED PyObject *__pyx_v_close_fds, PyObject *__pyx_v_cwd, PyObject *__pyx_v_env, PyObject *__pyx_v_startupinfo, PyObject *__pyx_v_creationflags, PyObject *__pyx_v_restore_signals, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v_executable, PyObject *__pyx_v_pass_fds, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_117subprocess_shell(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_cmd, PyObject *__pyx_v_shell, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_120subprocess_exec(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_program, PyObject *__pyx_v_shell, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_123connect_read_pipe(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_proto_factory, PyObject *__pyx_v_pipe); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_126connect_write_pipe(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_proto_factory, PyObject *__pyx_v_pipe); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_129add_signal_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig, PyObject *__pyx_v_callback, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_131remove_signal_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_24create_datagram_endpoint_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_133create_datagram_endpoint(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_local_addr, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_family, PyObject *__pyx_v_proto, PyObject *__pyx_v_flags, PyObject *__pyx_v_reuse_address, PyObject *__pyx_v_reuse_port, PyObject *__pyx_v_allow_broadcast, PyObject *__pyx_v_sock); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_136_monitor_fs(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_path, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_138_check_default_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_140_asyncgen_finalizer_hook(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_agen); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_142_asyncgen_firstiter_hook(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_agen); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_144shutdown_asyncgens(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_147shutdown_default_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_timeout); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_150_do_shutdown(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_future); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static int __pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration_2__set__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration_4__del__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_7_closed___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_9_debug_cc___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_20_debug_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_21_debug_handles_closed___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_22_debug_handles_current___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_uv_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_uv_handles_freed___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_cb_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_cb_handles_count___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_count___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_35_debug_stream_shutdown_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_33_debug_stream_listen_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_27_debug_stream_read_cb_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_34_debug_stream_read_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_28_debug_stream_read_eof_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_38_debug_stream_read_eof_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_31_debug_stream_read_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_25_debug_stream_write_tries___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_32_debug_stream_write_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_29_debug_stream_write_ctx_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_27_debug_stream_write_ctx_cnt___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_35_debug_stream_write_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_poll_read_events_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_26_poll_read_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_24_poll_write_events_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_27_poll_write_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_21_sock_try_write_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_28_debug_exception_handler_cnt___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_152__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4Loop_154__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_libuv_get_loop_t_ptr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loop); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_2libuv_get_version(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_4_testhelper_unwrap_capsuled_pointer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_loop); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_2__remove_reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_4cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_msg); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_6cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_loop); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_2__remove_writer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_4cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_msg); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_6cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static int __pyx_pf_6uvloop_4loop_6Handle___cinit__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static void __pyx_pf_6uvloop_4loop_6Handle_2__dealloc__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static int __pyx_pf_6uvloop_4loop_6Handle_4__init__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Handle_6__repr__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Handle_8cancel(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Handle_10cancelled(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Handle_17_source_traceback___get__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Handle_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Handle_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_11TimerHandle___cinit__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, uint64_t __pyx_v_delay, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_17_source_traceback___get__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto */
static void __pyx_pf_6uvloop_4loop_11TimerHandle_2__dealloc__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_4__repr__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_6cancelled(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_8cancel(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_10when(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_12PseudoSocket___init__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_fd); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_6family___get__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_4type___get__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_5proto___get__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_2__repr__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_4__getstate__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_6fileno(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_8dup(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_10get_inheritable(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_12set_inheritable(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_14ioctl(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_16getsockopt(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_18setsockopt(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_20getpeername(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_22getsockname(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_24share(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_process_id); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_26accept(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_28connect(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_30connect_ex(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_32bind(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_34listen(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_36makefile(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_38sendfile(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_40close(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_42detach(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_44shutdown(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_46sendmsg_afalg(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_48sendmsg(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_50sendto(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_52send(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_54sendall(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_56recv_into(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_58recvfrom_into(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_60recvmsg_into(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_62recvmsg(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_64recvfrom(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_66recv(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_68settimeout(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_70gettimeout(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_72setblocking(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_flag); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_74__enter__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_76__exit__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_err); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_78__reduce_cython__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_80__setstate_cython__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_8LruCache___init__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_maxsize); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_2__getitem__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_6uvloop_4loop_8LruCache_4__setitem__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_o); /* proto */
static int __pyx_pf_6uvloop_4loop_8LruCache_6__delitem__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_6uvloop_4loop_8LruCache_8__contains__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static Py_ssize_t __pyx_pf_6uvloop_4loop_8LruCache_10__len__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_12__iter__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_14__reduce_cython__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_16__setstate_cython__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_8UVHandle___cinit__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto */
static int __pyx_pf_6uvloop_4loop_8UVHandle_2__init__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto */
static void __pyx_pf_6uvloop_4loop_8UVHandle_4__dealloc__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_6__repr__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_17_source_traceback___get__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_14UVSocketHandle___cinit__(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_14UVSocketHandle_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_14UVSocketHandle_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_7UVAsync___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_7UVAsync_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6UVIdle___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6UVIdle_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_7UVCheck___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_7UVCheck_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_7UVTimer___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_7UVTimer_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6UVPoll___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6UVPoll_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_15UVBaseTransport___cinit__(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_7_paused___get__(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_2get_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_4set_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_6_force_close(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_8abort(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_10close(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_12is_closing(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_14get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_16set_write_buffer_limits(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_high, PyObject *__pyx_v_low); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_18get_write_buffer_limits(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_20get_extra_info(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_8_closing___get__(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_6uvloop_4loop_19_StreamWriteContext___dealloc__(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_19_StreamWriteContext_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_19_StreamWriteContext_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_8UVStream___cinit__(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_2__repr__(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_4write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_buf); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_6writelines(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_bufs); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_8write_eof(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_10can_write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_12is_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_14pause_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_16resume_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_14UVStreamServer___cinit__(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_14UVStreamServer_2__on_ssl_connected(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, PyObject *__pyx_v_transport, PyObject *__pyx_v_fut); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_14UVStreamServer_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_14UVStreamServer_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9TCPServer___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9TCPServer_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12TCPTransport_get_extra_info(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12TCPTransport_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12TCPTransport_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_18_TCPConnectRequest___cinit__(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_loop, PyObject *__pyx_v_transport); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18_TCPConnectRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18_TCPConnectRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_10UnixServer___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_10UnixServer_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_13UnixTransport___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_13UnixTransport_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_get_write_buffer_limits(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_2set_write_buffer_limits(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_high, CYTHON_UNUSED PyObject *__pyx_v_low); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_4get_write_buffer_size(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_6write(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_8writelines(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_list_of_data); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_10write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_12can_write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_14abort(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_pause_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_2resume_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_19_PipeConnectRequest___cinit__(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_loop, PyObject *__pyx_v_transport); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_19_PipeConnectRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_19_PipeConnectRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_9UVProcess___cinit__(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self); /* proto */
static void __pyx_pf_6uvloop_4loop_9UVProcess_2__dealloc__(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVProcess_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVProcess_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_18UVProcessTransport___cinit__(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_2__stdio_inited(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_waiter, PyObject *__pyx_v_stdio_fut); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_4get_protocol(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_6set_protocol(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_8get_pid(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_10get_returncode(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_12get_pipe_transport(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_fd); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_14terminate(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_16kill(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_18send_signal(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_signal); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_20is_closing(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_22close(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_24get_extra_info(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_name, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_26_wait(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_28__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_30__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_proc, PyObject *__pyx_v_fd); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_2connection_made(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_transport); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_6connection_lost(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exc); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_8pause_writing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_10resume_writing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_23ReadSubprocessPipeProto_data_received(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_cancel(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_2cancelled(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_9UVRequest___cinit__(struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9UVRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_14SockAddrHolder___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SockAddrHolder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_14SockAddrHolder_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SockAddrHolder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_8AddrInfo___cinit__(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self); /* proto */
static void __pyx_pf_6uvloop_4loop_8AddrInfo_2__dealloc__(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8AddrInfo_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_8AddrInfo_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_15AddrInfoRequest___cinit__(struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_flags, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15AddrInfoRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15AddrInfoRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_15NameInfoRequest___cinit__(struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15NameInfoRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15NameInfoRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport___cinit__(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_ssl_protocol, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_2get_extra_info(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_4set_protocol(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_protocol); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_6get_protocol(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_8is_closing(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_10close(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static void __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_12__dealloc__(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_14is_reading(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_16pause_reading(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_18resume_reading(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_20set_write_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_high, PyObject *__pyx_v_low); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_22get_write_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_24get_write_buffer_size(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_26set_read_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_high, PyObject *__pyx_v_low); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_28get_read_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_30get_read_buffer_size(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_16_protocol_paused___get__(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_32write(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_34writelines(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_list_of_data); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_36write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_38can_write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_40abort(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_42_force_close(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_exc); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_44_test__append_write_backlog(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_46__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_48__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_11SSLProtocol___cinit__(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static void __pyx_pf_6uvloop_4loop_11SSLProtocol_2__dealloc__(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto */
static int __pyx_pf_6uvloop_4loop_11SSLProtocol_4__init__(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_loop, PyObject *__pyx_v_app_protocol, PyObject *__pyx_v_sslcontext, PyObject *__pyx_v_waiter, PyObject *__pyx_v_server_side, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_call_connection_made, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_6_get_app_transport(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_8connection_made(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_transport); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_10connection_lost(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_exc); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_12get_buffer(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_n); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_14buffer_updated(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_nbytes); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_16eof_received(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda4(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda5(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_18pause_writing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_20resume_writing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_6uvloop_4loop_15_UDPSendContext___dealloc__(struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15_UDPSendContext_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_15_UDPSendContext_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_12UDPTransport___cinit__(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12UDPTransport_2sendto(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_addr); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12UDPTransport_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_12UDPTransport_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6uvloop_4loop_6Server___cinit__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_2__aenter__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_5__aexit__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_8__repr__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_10get_loop(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_12wait_closed(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_15close(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_17is_serving(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_19start_serving(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_22serve_forever(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_7sockets___get__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_25__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6Server_27__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_6_test_coroutine_1(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_9__pyx_unpickle_PseudoSocket(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6uvloop_4loop_11__pyx_unpickle_LruCache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6uvloop_4loop_UVHandle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVSocketHandle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVAsync(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVTimer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVIdle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVBaseTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_Loop(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_Handle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_TimerHandle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVCheck(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVPoll(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVStream(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVStreamServer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_TCPServer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_TCPTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UnixServer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UnixTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_ReadUnixTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_WriteUnixTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVProcess(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVProcessTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVFSEvent(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UVRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop__SSLProtocolTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_SSLProtocol(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_UDPTransport(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_Server(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_PseudoSocket(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_LruCache(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop__StreamWriteContext(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop__TCPConnectRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop__PipeConnectRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_SockAddrHolder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_AddrInfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_AddrInfoRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop_NameInfoRequest(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop__UDPSendContext(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct___getaddrinfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_1__getnameinfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_2___get__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_3_run_until_complete(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_5_getnameinfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_6_start_tls(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_7_create_server(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_8_create_connection(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_9_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_10_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_11_create_unix_server(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_13_sock_recv(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_15_sock_sendall(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_16_sock_accept(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_17_sock_connect(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_20_sock_sendto(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_28_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_31__start_handshake(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_32__start_shutdown(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_33___aenter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_34___aexit__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_35_wait_closed(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_36_start_serving(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_37_serve_forever(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_copy = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_rstrip = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_6uvloop_4loop_UVHandle;
  PyObject *__pyx_type_6uvloop_4loop_UVSocketHandle;
  PyObject *__pyx_type_6uvloop_4loop_UVAsync;
  PyObject *__pyx_type_6uvloop_4loop_UVTimer;
  PyObject *__pyx_type_6uvloop_4loop_UVIdle;
  PyObject *__pyx_type_6uvloop_4loop_UVBaseTransport;
  PyObject *__pyx_type_6uvloop_4loop_Loop;
  PyObject *__pyx_type_6uvloop_4loop_Handle;
  PyObject *__pyx_type_6uvloop_4loop_TimerHandle;
  PyObject *__pyx_type_6uvloop_4loop_UVCheck;
  PyObject *__pyx_type_6uvloop_4loop_UVPoll;
  PyObject *__pyx_type_6uvloop_4loop_UVStream;
  PyObject *__pyx_type_6uvloop_4loop_UVStreamServer;
  PyObject *__pyx_type_6uvloop_4loop_TCPServer;
  PyObject *__pyx_type_6uvloop_4loop_TCPTransport;
  PyObject *__pyx_type_6uvloop_4loop_UnixServer;
  PyObject *__pyx_type_6uvloop_4loop_UnixTransport;
  PyObject *__pyx_type_6uvloop_4loop_ReadUnixTransport;
  PyObject *__pyx_type_6uvloop_4loop_WriteUnixTransport;
  PyObject *__pyx_type_6uvloop_4loop_UVProcess;
  PyObject *__pyx_type_6uvloop_4loop_UVProcessTransport;
  PyObject *__pyx_type_6uvloop_4loop_UVFSEvent;
  PyObject *__pyx_type_6uvloop_4loop_UVRequest;
  PyObject *__pyx_type_6uvloop_4loop__SSLProtocolTransport;
  PyObject *__pyx_type_6uvloop_4loop_SSLProtocol;
  PyObject *__pyx_type_6uvloop_4loop_UDPTransport;
  PyObject *__pyx_type_6uvloop_4loop_Server;
  PyObject *__pyx_type_6uvloop_4loop_PseudoSocket;
  PyObject *__pyx_type_6uvloop_4loop_LruCache;
  PyObject *__pyx_type_6uvloop_4loop__StreamWriteContext;
  PyObject *__pyx_type_6uvloop_4loop__TCPConnectRequest;
  PyObject *__pyx_type_6uvloop_4loop__PipeConnectRequest;
  PyObject *__pyx_type_6uvloop_4loop_SockAddrHolder;
  PyObject *__pyx_type_6uvloop_4loop_AddrInfo;
  PyObject *__pyx_type_6uvloop_4loop_AddrInfoRequest;
  PyObject *__pyx_type_6uvloop_4loop_NameInfoRequest;
  PyObject *__pyx_type_6uvloop_4loop__UDPSendContext;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_2___get__;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter__;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit__;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever;
  PyObject *__pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1;
  PyObject *__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled;
  #endif
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVHandle;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVSocketHandle;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVAsync;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVTimer;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVIdle;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVBaseTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_Loop;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_Handle;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_TimerHandle;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVCheck;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVPoll;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVStream;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVStreamServer;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_TCPServer;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_TCPTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UnixServer;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UnixTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_ReadUnixTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_WriteUnixTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVProcess;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVProcessTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVFSEvent;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UVRequest;
  PyTypeObject *__pyx_ptype_6uvloop_4loop__SSLProtocolTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_SSLProtocol;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_UDPTransport;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_Server;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_PseudoSocket;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_LruCache;
  PyTypeObject *__pyx_ptype_6uvloop_4loop__StreamWriteContext;
  PyTypeObject *__pyx_ptype_6uvloop_4loop__TCPConnectRequest;
  PyTypeObject *__pyx_ptype_6uvloop_4loop__PipeConnectRequest;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_SockAddrHolder;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_AddrInfo;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_AddrInfoRequest;
  PyTypeObject *__pyx_ptype_6uvloop_4loop_NameInfoRequest;
  PyTypeObject *__pyx_ptype_6uvloop_4loop__UDPSendContext;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever;
  PyTypeObject *__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1;
  PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled;
  PyObject *__pyx_kp_b_1;
  PyObject *__pyx_kp_u_18_7_7_7;
  PyObject *__pyx_kp_u_AF_INET6_address_must_be_tuple;
  PyObject *__pyx_kp_u_AF_INET6_must_be_a_tuple_of_2_4;
  PyObject *__pyx_kp_u_AF_INET_address_must_be_tuple;
  PyObject *__pyx_kp_u_AF_INET_address_must_be_tuple_of;
  PyObject *__pyx_kp_u_AF_UNIX_address_must_be_a_str_or;
  PyObject *__pyx_kp_u_AF_UNSPEC_in_DNS_results;
  PyObject *__pyx_n_u_AI_CANONNAME;
  PyObject *__pyx_kp_u_A_Stream_Socket_was_expected_got;
  PyObject *__pyx_kp_u_A_UDP_Socket_was_expected_got_r;
  PyObject *__pyx_kp_u_A_UNIX_Domain_Stream_Socket_was;
  PyObject *__pyx_kp_u_A_callable_object_or_None_is_exp;
  PyObject *__pyx_n_s_AbstractChildWatcher;
  PyObject *__pyx_n_s_AddrInfo;
  PyObject *__pyx_n_s_AddrInfoRequest;
  PyObject *__pyx_n_s_AddrInfoRequest___reduce_cython;
  PyObject *__pyx_n_s_AddrInfoRequest___setstate_cytho;
  PyObject *__pyx_kp_u_AddrInfoRequest_callback_called;
  PyObject *__pyx_n_s_AddrInfo___reduce_cython;
  PyObject *__pyx_n_s_AddrInfo___setstate_cython;
  PyObject *__pyx_kp_u_AddrInfo_data_is_NULL;
  PyObject *__pyx_n_s_AddressFamily;
  PyObject *__pyx_kp_u_Address_r_is_already_in_use;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_AttributeError;
  PyObject *__pyx_kp_u_Bad_exception_data_from_child_r;
  PyObject *__pyx_n_s_BaseException;
  PyObject *__pyx_n_s_BaseProtocol;
  PyObject *__pyx_n_s_BlockingIOError;
  PyObject *__pyx_n_s_BrokenPipeError;
  PyObject *__pyx_n_s_CHANGE;
  PyObject *__pyx_kp_u_Callback_handles_8;
  PyObject *__pyx_n_s_CancelledError;
  PyObject *__pyx_kp_u_Cannot_call_write_after_write_eo;
  PyObject *__pyx_kp_u_Cannot_call_writelines_after_wri;
  PyObject *__pyx_kp_u_Cannot_close_a_running_event_loo;
  PyObject *__pyx_kp_u_Cannot_run_the_event_loop_while;
  PyObject *__pyx_kp_u_Cannot_serialize_socket_object;
  PyObject *__pyx_n_s_CertificateError;
  PyObject *__pyx_kp_u_Connect_call_failed;
  PyObject *__pyx_n_s_ConnectionAbortedError;
  PyObject *__pyx_n_s_ConnectionRefusedError;
  PyObject *__pyx_n_s_ConnectionResetError;
  PyObject *__pyx_n_s_Counter;
  PyObject *__pyx_kp_u_Creating__SSLProtocolTransport_t;
  PyObject *__pyx_n_s_DEVNULL;
  PyObject *__pyx_n_s_DeprecationWarning;
  PyObject *__pyx_n_u_EAI_ADDRFAMILY;
  PyObject *__pyx_n_u_EAI_AGAIN;
  PyObject *__pyx_n_u_EAI_BADFLAGS;
  PyObject *__pyx_n_u_EAI_BADHINTS;
  PyObject *__pyx_n_u_EAI_CANCELED;
  PyObject *__pyx_n_u_EAI_FAIL;
  PyObject *__pyx_n_u_EAI_FAMILY;
  PyObject *__pyx_n_u_EAI_MEMORY;
  PyObject *__pyx_n_u_EAI_NODATA;
  PyObject *__pyx_n_u_EAI_NONAME;
  PyObject *__pyx_n_u_EAI_OVERFLOW;
  PyObject *__pyx_n_u_EAI_PROTOCOL;
  PyObject *__pyx_n_u_EAI_SERVICE;
  PyObject *__pyx_n_u_EAI_SOCKTYPE;
  PyObject *__pyx_kp_u_Error_calling_eof_received;
  PyObject *__pyx_kp_u_Error_occurred_during_shutdown;
  PyObject *__pyx_kp_u_Errors_logged;
  PyObject *__pyx_kp_u_Event_loop_is_closed;
  PyObject *__pyx_kp_u_Event_loop_stopped_before_Future;
  PyObject *__pyx_kp_u_Exception_in___repr___r_value_ty;
  PyObject *__pyx_kp_u_Exception_in_callback;
  PyObject *__pyx_kp_u_Exception_in_default_exception_h;
  PyObject *__pyx_kp_u_Exception_in_default_exception_h_2;
  PyObject *__pyx_kp_u_Exception_occurred_in_preexec_fn;
  PyObject *__pyx_kp_u_Executing_r_took_3f_seconds;
  PyObject *__pyx_kp_u_Executing_s_took_3f_seconds;
  PyObject *__pyx_kp_u_Executor_shutdown_has_been_calle;
  PyObject *__pyx_kp_u_Fatal_error_on_SSL_protocol;
  PyObject *__pyx_kp_u_Fatal_error_on_server;
  PyObject *__pyx_kp_u_Fatal_error_on_transport;
  PyObject *__pyx_kp_u_Fatal_error_on_transport_2;
  PyObject *__pyx_kp_u_Fatal_write_error_on_datagram_tr;
  PyObject *__pyx_n_s_FileExistsError;
  PyObject *__pyx_n_s_FileNotFoundError;
  PyObject *__pyx_n_s_FileSystemEvent;
  PyObject *__pyx_kp_u_File_descriptor_r_is_used_by_tra;
  PyObject *__pyx_n_s_FlowControlMixin;
  PyObject *__pyx_n_s_Future;
  PyObject *__pyx_n_s_Handle;
  PyObject *__pyx_n_s_Handle___reduce_cython;
  PyObject *__pyx_n_s_Handle___setstate_cython;
  PyObject *__pyx_n_s_Handle_cancel;
  PyObject *__pyx_n_s_Handle_cancelled;
  PyObject *__pyx_kp_u_Handle_loop_is_None_in_Handle;
  PyObject *__pyx_n_u_IPV6_V6ONLY;
  PyObject *__pyx_kp_u_IPv4_sockaddr_must_be_2_tuple;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_n_s_IntEnum;
  PyObject *__pyx_n_s_InterruptedError;
  PyObject *__pyx_n_s_InvalidStateError;
  PyObject *__pyx_kp_u_Invalid_address_must_be_None_or;
  PyObject *__pyx_kp_u_Invalid_file_descriptor;
  PyObject *__pyx_kp_u_Invalid_file_object_r;
  PyObject *__pyx_n_s_IsADirectoryError;
  PyObject *__pyx_n_s_Iterable;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_KeyboardInterrupt;
  PyObject *__pyx_kp_u_Listen_errors;
  PyObject *__pyx_n_s_Loop;
  PyObject *__pyx_n_s_Loop___get___locals_lambda;
  PyObject *__pyx_n_s_Loop___reduce_cython;
  PyObject *__pyx_n_s_Loop___setstate_cython;
  PyObject *__pyx_n_s_Loop___sighandler;
  PyObject *__pyx_n_s_Loop___subprocess_run;
  PyObject *__pyx_n_s_Loop__asyncgen_finalizer_hook;
  PyObject *__pyx_n_s_Loop__asyncgen_firstiter_hook;
  PyObject *__pyx_n_s_Loop__check_default_executor;
  PyObject *__pyx_n_s_Loop__do_shutdown;
  PyObject *__pyx_n_s_Loop__get_backend_id;
  PyObject *__pyx_n_s_Loop__getaddrinfo_locals_callbac;
  PyObject *__pyx_n_s_Loop__getnameinfo_locals_callbac;
  PyObject *__pyx_n_s_Loop__monitor_fs;
  PyObject *__pyx_kp_u_Loop__read_from_self;
  PyObject *__pyx_n_s_Loop__sighandler;
  PyObject *__pyx_kp_u_Loop__sock_accept;
  PyObject *__pyx_kp_u_Loop__sock_connect;
  PyObject *__pyx_kp_u_Loop__sock_recv;
  PyObject *__pyx_kp_u_Loop__sock_recv_into;
  PyObject *__pyx_kp_u_Loop__sock_sendall;
  PyObject *__pyx_kp_u_Loop__stop;
  PyObject *__pyx_n_s_Loop__subprocess_run;
  PyObject *__pyx_n_s_Loop_add_reader;
  PyObject *__pyx_n_s_Loop_add_signal_handler;
  PyObject *__pyx_n_s_Loop_add_writer;
  PyObject *__pyx_n_s_Loop_call_at;
  PyObject *__pyx_n_s_Loop_call_exception_handler;
  PyObject *__pyx_n_s_Loop_call_later;
  PyObject *__pyx_n_s_Loop_call_soon;
  PyObject *__pyx_n_s_Loop_call_soon_threadsafe;
  PyObject *__pyx_n_s_Loop_close;
  PyObject *__pyx_n_s_Loop_connect_accepted_socket;
  PyObject *__pyx_n_s_Loop_connect_read_pipe;
  PyObject *__pyx_n_s_Loop_connect_write_pipe;
  PyObject *__pyx_n_s_Loop_create_connection;
  PyObject *__pyx_n_s_Loop_create_datagram_endpoint;
  PyObject *__pyx_n_s_Loop_create_future;
  PyObject *__pyx_n_s_Loop_create_server;
  PyObject *__pyx_n_s_Loop_create_task;
  PyObject *__pyx_n_s_Loop_create_unix_connection;
  PyObject *__pyx_n_s_Loop_create_unix_server;
  PyObject *__pyx_kp_u_Loop_debug_info;
  PyObject *__pyx_n_s_Loop_default_exception_handler;
  PyObject *__pyx_n_s_Loop_get_debug;
  PyObject *__pyx_n_s_Loop_get_exception_handler;
  PyObject *__pyx_n_s_Loop_get_task_factory;
  PyObject *__pyx_n_s_Loop_getaddrinfo;
  PyObject *__pyx_n_s_Loop_getnameinfo;
  PyObject *__pyx_n_s_Loop_is_closed;
  PyObject *__pyx_n_s_Loop_is_running;
  PyObject *__pyx_n_s_Loop_remove_reader;
  PyObject *__pyx_n_s_Loop_remove_signal_handler;
  PyObject *__pyx_n_s_Loop_remove_writer;
  PyObject *__pyx_n_s_Loop_run_forever;
  PyObject *__pyx_n_s_Loop_run_in_executor;
  PyObject *__pyx_n_s_Loop_run_until_complete;
  PyObject *__pyx_n_s_Loop_set_debug;
  PyObject *__pyx_n_s_Loop_set_default_executor;
  PyObject *__pyx_n_s_Loop_set_exception_handler;
  PyObject *__pyx_n_s_Loop_set_task_factory;
  PyObject *__pyx_n_s_Loop_shutdown_asyncgens;
  PyObject *__pyx_n_s_Loop_shutdown_default_executor;
  PyObject *__pyx_n_s_Loop_sock_accept;
  PyObject *__pyx_n_s_Loop_sock_connect;
  PyObject *__pyx_n_s_Loop_sock_recv;
  PyObject *__pyx_n_s_Loop_sock_recv_into;
  PyObject *__pyx_n_s_Loop_sock_recvfrom;
  PyObject *__pyx_n_s_Loop_sock_recvfrom_into;
  PyObject *__pyx_n_s_Loop_sock_sendall;
  PyObject *__pyx_n_s_Loop_sock_sendto;
  PyObject *__pyx_n_s_Loop_start_tls;
  PyObject *__pyx_n_s_Loop_stop;
  PyObject *__pyx_n_s_Loop_subprocess_exec;
  PyObject *__pyx_n_s_Loop_subprocess_shell;
  PyObject *__pyx_kp_u_Loop_time;
  PyObject *__pyx_n_s_Loop_time_2;
  PyObject *__pyx_n_s_LruCache;
  PyObject *__pyx_n_s_LruCache___reduce_cython;
  PyObject *__pyx_n_s_LruCache___setstate_cython;
  PyObject *__pyx_n_s_MemoryBIO;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_kp_u_Multiple_exceptions;
  PyObject *__pyx_n_s_NSIG;
  PyObject *__pyx_n_s_NameInfoRequest;
  PyObject *__pyx_n_s_NameInfoRequest___reduce_cython;
  PyObject *__pyx_n_s_NameInfoRequest___setstate_cytho;
  PyObject *__pyx_kp_u_Neither_host_port_nor_sock_were;
  PyObject *__pyx_kp_u_Non_thread_safe_operation_invoke;
  PyObject *__pyx_n_s_NotImplementedError;
  PyObject *__pyx_kp_u_Number_of_signals;
  PyObject *__pyx_n_s_OSError;
  PyObject *__pyx_n_s_O_RDWR;
  PyObject *__pyx_kp_u_Object_created_at_most_recent_ca;
  PyObject *__pyx_n_s_OrderedDict;
  PyObject *__pyx_n_s_OverflowError;
  PyObject *__pyx_n_s_PIPE;
  PyObject *__pyx_n_u_PYTHONASYNCIODEBUG;
  PyObject *__pyx_kp_u_Passing_reuse_address_True_is_no;
  PyObject *__pyx_n_s_PermissionError;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_kp_s_Pickling_of_struct_members_such;
  PyObject *__pyx_n_s_PipeConnectRequest;
  PyObject *__pyx_n_s_PipeConnectRequest___reduce_cyt;
  PyObject *__pyx_n_s_PipeConnectRequest___setstate_c;
  PyObject *__pyx_kp_u_Polls_debug_info;
  PyObject *__pyx_n_s_ProcessLookupError;
  PyObject *__pyx_kp_u_Process_info;
  PyObject *__pyx_kp_u_Process_memory;
  PyObject *__pyx_n_s_Protocol;
  PyObject *__pyx_n_s_PseudoSocket;
  PyObject *__pyx_n_s_PseudoSocket___enter;
  PyObject *__pyx_n_s_PseudoSocket___exit;
  PyObject *__pyx_n_s_PseudoSocket___getstate;
  PyObject *__pyx_n_s_PseudoSocket___reduce_cython;
  PyObject *__pyx_n_s_PseudoSocket___setstate_cython;
  PyObject *__pyx_n_s_PseudoSocket_accept;
  PyObject *__pyx_n_s_PseudoSocket_bind;
  PyObject *__pyx_n_s_PseudoSocket_close;
  PyObject *__pyx_n_s_PseudoSocket_connect;
  PyObject *__pyx_n_s_PseudoSocket_connect_ex;
  PyObject *__pyx_n_s_PseudoSocket_detach;
  PyObject *__pyx_n_s_PseudoSocket_dup;
  PyObject *__pyx_n_s_PseudoSocket_fileno;
  PyObject *__pyx_n_s_PseudoSocket_get_inheritable;
  PyObject *__pyx_n_s_PseudoSocket_getpeername;
  PyObject *__pyx_n_s_PseudoSocket_getsockname;
  PyObject *__pyx_n_s_PseudoSocket_getsockopt;
  PyObject *__pyx_n_s_PseudoSocket_gettimeout;
  PyObject *__pyx_n_s_PseudoSocket_ioctl;
  PyObject *__pyx_n_s_PseudoSocket_listen;
  PyObject *__pyx_n_s_PseudoSocket_makefile;
  PyObject *__pyx_n_s_PseudoSocket_recv;
  PyObject *__pyx_n_s_PseudoSocket_recv_into;
  PyObject *__pyx_n_s_PseudoSocket_recvfrom;
  PyObject *__pyx_n_s_PseudoSocket_recvfrom_into;
  PyObject *__pyx_n_s_PseudoSocket_recvmsg;
  PyObject *__pyx_n_s_PseudoSocket_recvmsg_into;
  PyObject *__pyx_n_s_PseudoSocket_send;
  PyObject *__pyx_n_s_PseudoSocket_sendall;
  PyObject *__pyx_n_s_PseudoSocket_sendfile;
  PyObject *__pyx_n_s_PseudoSocket_sendmsg;
  PyObject *__pyx_n_s_PseudoSocket_sendmsg_afalg;
  PyObject *__pyx_n_s_PseudoSocket_sendto;
  PyObject *__pyx_n_s_PseudoSocket_set_inheritable;
  PyObject *__pyx_n_s_PseudoSocket_setblocking;
  PyObject *__pyx_n_s_PseudoSocket_setsockopt;
  PyObject *__pyx_n_s_PseudoSocket_settimeout;
  PyObject *__pyx_n_s_PseudoSocket_share;
  PyObject *__pyx_n_s_PseudoSocket_shutdown;
  PyObject *__pyx_n_s_Pyx_CFunc_6uvloop_4loop_4Loop;
  PyObject *__pyx_n_s_RENAME;
  PyObject *__pyx_n_s_RENAME_CHANGE;
  PyObject *__pyx_kp_u_Racing_with_another_loop_to_spaw;
  PyObject *__pyx_n_s_ReadSubprocessPipeProto;
  PyObject *__pyx_n_s_ReadSubprocessPipeProto_data_rec;
  PyObject *__pyx_n_s_ReadUnixTransport;
  PyObject *__pyx_n_s_ReadUnixTransport___reduce_cytho;
  PyObject *__pyx_n_s_ReadUnixTransport___setstate_cyt;
  PyObject *__pyx_n_s_ReadUnixTransport_abort;
  PyObject *__pyx_n_s_ReadUnixTransport_can_write_eof;
  PyObject *__pyx_n_s_ReadUnixTransport_get_write_buff;
  PyObject *__pyx_n_s_ReadUnixTransport_get_write_buff_2;
  PyObject *__pyx_n_s_ReadUnixTransport_set_write_buff;
  PyObject *__pyx_n_s_ReadUnixTransport_write;
  PyObject *__pyx_n_s_ReadUnixTransport_write_eof;
  PyObject *__pyx_n_s_ReadUnixTransport_writelines;
  PyObject *__pyx_kp_u_Read_EOF_failed_callbacks;
  PyObject *__pyx_kp_u_Read_EOFs;
  PyObject *__pyx_kp_u_Read_callbacks;
  PyObject *__pyx_kp_u_Read_callbacks_failed;
  PyObject *__pyx_kp_u_Read_errors;
  PyObject *__pyx_kp_u_Read_events;
  PyObject *__pyx_kp_u_Read_failed_callbacks;
  PyObject *__pyx_n_s_ResourceWarning;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_n_s_RuntimeWarning;
  PyObject *__pyx_n_s_SIG_DFL;
  PyObject *__pyx_n_u_SOCK_NONBLOCK;
  PyObject *__pyx_n_u_SO_BROADCAST;
  PyObject *__pyx_n_u_SO_REUSEPORT;
  PyObject *__pyx_n_s_SSLContext;
  PyObject *__pyx_n_s_SSLError;
  PyObject *__pyx_n_s_SSLProtocol;
  PyObject *__pyx_n_s_SSLProtocolTransport;
  PyObject *__pyx_n_s_SSLProtocolTransport___reduce_c;
  PyObject *__pyx_n_s_SSLProtocolTransport___setstate;
  PyObject *__pyx_n_s_SSLProtocolTransport__force_clo;
  PyObject *__pyx_n_s_SSLProtocolTransport__test__app;
  PyObject *__pyx_n_s_SSLProtocolTransport_abort;
  PyObject *__pyx_n_s_SSLProtocolTransport_can_write;
  PyObject *__pyx_n_s_SSLProtocolTransport_close;
  PyObject *__pyx_n_s_SSLProtocolTransport_get_extra;
  PyObject *__pyx_n_s_SSLProtocolTransport_get_protoc;
  PyObject *__pyx_n_s_SSLProtocolTransport_get_read_b;
  PyObject *__pyx_n_s_SSLProtocolTransport_get_read_b_2;
  PyObject *__pyx_n_s_SSLProtocolTransport_get_write;
  PyObject *__pyx_n_s_SSLProtocolTransport_get_write_2;
  PyObject *__pyx_n_s_SSLProtocolTransport_is_closing;
  PyObject *__pyx_n_s_SSLProtocolTransport_is_reading;
  PyObject *__pyx_n_s_SSLProtocolTransport_pause_read;
  PyObject *__pyx_n_s_SSLProtocolTransport_resume_rea;
  PyObject *__pyx_n_s_SSLProtocolTransport_set_protoc;
  PyObject *__pyx_n_s_SSLProtocolTransport_set_read_b;
  PyObject *__pyx_n_s_SSLProtocolTransport_set_write;
  PyObject *__pyx_n_s_SSLProtocolTransport_write;
  PyObject *__pyx_n_s_SSLProtocolTransport_write_eof;
  PyObject *__pyx_n_s_SSLProtocolTransport_writelines;
  PyObject *__pyx_n_s_SSLProtocol___reduce_cython;
  PyObject *__pyx_n_s_SSLProtocol___setstate_cython;
  PyObject *__pyx_kp_u_SSLProtocol__do_read;
  PyObject *__pyx_n_s_SSLProtocol__get_app_transport;
  PyObject *__pyx_n_s_SSLProtocol__start_handshake_loc;
  PyObject *__pyx_n_s_SSLProtocol__start_shutdown_loca;
  PyObject *__pyx_n_s_SSLProtocol_buffer_updated;
  PyObject *__pyx_n_s_SSLProtocol_connection_lost;
  PyObject *__pyx_n_s_SSLProtocol_connection_made;
  PyObject *__pyx_n_s_SSLProtocol_eof_received;
  PyObject *__pyx_n_s_SSLProtocol_get_buffer;
  PyObject *__pyx_n_s_SSLProtocol_pause_writing;
  PyObject *__pyx_n_s_SSLProtocol_resume_writing;
  PyObject *__pyx_n_s_SSLSyscallError;
  PyObject *__pyx_n_s_SSLWantReadError;
  PyObject *__pyx_n_s_SSLZeroReturnError;
  PyObject *__pyx_n_s_SSL_ERROR_SYSCALL;
  PyObject *__pyx_n_s_SSL_ERROR_WANT_READ;
  PyObject *__pyx_n_s_SSL_ERROR_WANT_WRITE;
  PyObject *__pyx_kp_u_SSL_connection_is_closed;
  PyObject *__pyx_kp_u_SSL_handshake_failed;
  PyObject *__pyx_kp_u_SSL_handshake_failed_on_verifyin;
  PyObject *__pyx_kp_u_SSL_handshake_is_taking_longer_t;
  PyObject *__pyx_kp_u_SSL_shutdown_timed_out;
  PyObject *__pyx_n_s_STDOUT;
  PyObject *__pyx_n_s_S_ISSOCK;
  PyObject *__pyx_n_s_Server;
  PyObject *__pyx_n_s_Server___aenter;
  PyObject *__pyx_n_s_Server___aexit;
  PyObject *__pyx_n_s_Server___reduce_cython;
  PyObject *__pyx_n_s_Server___setstate_cython;
  PyObject *__pyx_n_s_Server_close;
  PyObject *__pyx_n_s_Server_get_loop;
  PyObject *__pyx_n_s_Server_is_serving;
  PyObject *__pyx_n_s_Server_serve_forever;
  PyObject *__pyx_kp_u_Server_side_SSL_needs_a_valid_SS;
  PyObject *__pyx_n_s_Server_start_serving;
  PyObject *__pyx_n_s_Server_wait_closed;
  PyObject *__pyx_kp_u_Shutdown_errors;
  PyObject *__pyx_n_s_SockAddrHolder;
  PyObject *__pyx_n_s_SockAddrHolder___reduce_cython;
  PyObject *__pyx_n_s_SockAddrHolder___setstate_cython;
  PyObject *__pyx_kp_u_Sock_ops_successful_on_1st_try;
  PyObject *__pyx_n_s_SocketKind;
  PyObject *__pyx_kp_u_Socket_try_writes;
  PyObject *__pyx_n_s_StackSummary;
  PyObject *__pyx_n_s_StreamWriteContext;
  PyObject *__pyx_n_s_StreamWriteContext___reduce_cyt;
  PyObject *__pyx_n_s_StreamWriteContext___setstate_c;
  PyObject *__pyx_kp_u_StreamWriteContext_close_py_buf;
  PyObject *__pyx_kp_u_StreamWriteContext_close_uv_buf;
  PyObject *__pyx_kp_u_Streams_debug_info;
  PyObject *__pyx_n_s_SubprocessError;
  PyObject *__pyx_n_s_SyncSocketReaderFuture;
  PyObject *__pyx_n_s_SyncSocketReaderFuture___init;
  PyObject *__pyx_n_s_SyncSocketReaderFuture___remove;
  PyObject *__pyx_n_s_SyncSocketReaderFuture__loop;
  PyObject *__pyx_n_s_SyncSocketReaderFuture__remove;
  PyObject *__pyx_n_s_SyncSocketReaderFuture__sock;
  PyObject *__pyx_n_s_SyncSocketReaderFuture_cancel;
  PyObject *__pyx_n_s_SyncSocketWriterFuture;
  PyObject *__pyx_n_s_SyncSocketWriterFuture___init;
  PyObject *__pyx_n_s_SyncSocketWriterFuture___remove;
  PyObject *__pyx_n_s_SyncSocketWriterFuture__loop;
  PyObject *__pyx_n_s_SyncSocketWriterFuture__remove;
  PyObject *__pyx_n_s_SyncSocketWriterFuture__sock;
  PyObject *__pyx_n_s_SyncSocketWriterFuture_cancel;
  PyObject *__pyx_n_s_SystemExit;
  PyObject *__pyx_n_s_TCPConnectRequest;
  PyObject *__pyx_n_s_TCPConnectRequest___reduce_cyth;
  PyObject *__pyx_n_s_TCPConnectRequest___setstate_cy;
  PyObject *__pyx_n_s_TCPServer;
  PyObject *__pyx_n_s_TCPServer___reduce_cython;
  PyObject *__pyx_n_s_TCPServer___setstate_cython;
  PyObject *__pyx_n_s_TCPTransport;
  PyObject *__pyx_n_s_TCPTransport___reduce_cython;
  PyObject *__pyx_n_s_TCPTransport___setstate_cython;
  PyObject *__pyx_n_s_TCPTransport_get_extra_info;
  PyObject *__pyx_n_s_Task;
  PyObject *__pyx_kp_u_The_executor_did_not_finishing_j;
  PyObject *__pyx_kp_u_The_reuse_address_parameter_has;
  PyObject *__pyx_n_s_Thread;
  PyObject *__pyx_n_s_ThreadPoolExecutor;
  PyObject *__pyx_n_s_TimeoutError;
  PyObject *__pyx_n_s_TimerHandle;
  PyObject *__pyx_n_s_TimerHandle___reduce_cython;
  PyObject *__pyx_n_s_TimerHandle___setstate_cython;
  PyObject *__pyx_n_s_TimerHandle_cancel;
  PyObject *__pyx_n_s_TimerHandle_cancelled;
  PyObject *__pyx_n_s_TimerHandle_when;
  PyObject *__pyx_kp_u_Timer_handles_8;
  PyObject *__pyx_n_s_Transport;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_UDPSendContext;
  PyObject *__pyx_n_s_UDPSendContext___reduce_cython;
  PyObject *__pyx_n_s_UDPSendContext___setstate_cytho;
  PyObject *__pyx_n_s_UDPTransport;
  PyObject *__pyx_n_s_UDPTransport___reduce_cython;
  PyObject *__pyx_n_s_UDPTransport___setstate_cython;
  PyObject *__pyx_kp_u_UDPTransport_family_is_undefined;
  PyObject *__pyx_kp_u_UDPTransport_family_is_undefined_2;
  PyObject *__pyx_n_s_UDPTransport_sendto;
  PyObject *__pyx_n_s_UVAsync;
  PyObject *__pyx_n_s_UVAsync___reduce_cython;
  PyObject *__pyx_n_s_UVAsync___setstate_cython;
  PyObject *__pyx_n_s_UVBaseTransport;
  PyObject *__pyx_n_s_UVBaseTransport___reduce_cython;
  PyObject *__pyx_n_s_UVBaseTransport___setstate_cytho;
  PyObject *__pyx_n_s_UVBaseTransport__force_close;
  PyObject *__pyx_n_s_UVBaseTransport_abort;
  PyObject *__pyx_n_s_UVBaseTransport_close;
  PyObject *__pyx_n_s_UVBaseTransport_get_extra_info;
  PyObject *__pyx_n_s_UVBaseTransport_get_protocol;
  PyObject *__pyx_n_s_UVBaseTransport_get_write_buffer;
  PyObject *__pyx_n_s_UVBaseTransport_get_write_buffer_2;
  PyObject *__pyx_n_s_UVBaseTransport_is_closing;
  PyObject *__pyx_n_s_UVBaseTransport_set_protocol;
  PyObject *__pyx_n_s_UVBaseTransport_set_write_buffer;
  PyObject *__pyx_n_s_UVCheck;
  PyObject *__pyx_n_s_UVCheck___reduce_cython;
  PyObject *__pyx_n_s_UVCheck___setstate_cython;
  PyObject *__pyx_n_s_UVFSEvent;
  PyObject *__pyx_n_s_UVFSEvent___reduce_cython;
  PyObject *__pyx_n_s_UVFSEvent___setstate_cython;
  PyObject *__pyx_n_s_UVFSEvent_cancel;
  PyObject *__pyx_n_s_UVFSEvent_cancelled;
  PyObject *__pyx_n_s_UVHandle;
  PyObject *__pyx_n_s_UVHandle___reduce_cython;
  PyObject *__pyx_n_s_UVHandle___setstate_cython;
  PyObject *__pyx_kp_u_UVHandles_python_libuv_total;
  PyObject *__pyx_n_s_UVIdle;
  PyObject *__pyx_n_s_UVIdle___reduce_cython;
  PyObject *__pyx_n_s_UVIdle___setstate_cython;
  PyObject *__pyx_n_s_UVPoll;
  PyObject *__pyx_n_s_UVPoll___reduce_cython;
  PyObject *__pyx_n_s_UVPoll___setstate_cython;
  PyObject *__pyx_n_s_UVProcess;
  PyObject *__pyx_n_s_UVProcessTransport;
  PyObject *__pyx_n_s_UVProcessTransport___reduce_cyth;
  PyObject *__pyx_n_s_UVProcessTransport___setstate_cy;
  PyObject *__pyx_n_s_UVProcessTransport___stdio_inite;
  PyObject *__pyx_kp_u_UVProcessTransport__call_connect;
  PyObject *__pyx_n_s_UVProcessTransport__stdio_inite;
  PyObject *__pyx_n_s_UVProcessTransport__wait;
  PyObject *__pyx_n_s_UVProcessTransport_close;
  PyObject *__pyx_n_s_UVProcessTransport_get_extra_inf;
  PyObject *__pyx_n_s_UVProcessTransport_get_pid;
  PyObject *__pyx_n_s_UVProcessTransport_get_pipe_tran;
  PyObject *__pyx_n_s_UVProcessTransport_get_protocol;
  PyObject *__pyx_n_s_UVProcessTransport_get_returncod;
  PyObject *__pyx_n_s_UVProcessTransport_is_closing;
  PyObject *__pyx_n_s_UVProcessTransport_kill;
  PyObject *__pyx_n_s_UVProcessTransport_send_signal;
  PyObject *__pyx_n_s_UVProcessTransport_set_protocol;
  PyObject *__pyx_n_s_UVProcessTransport_terminate;
  PyObject *__pyx_n_s_UVProcess___reduce_cython;
  PyObject *__pyx_n_s_UVProcess___setstate_cython;
  PyObject *__pyx_kp_u_UVProcess__close_after_spawn_cal;
  PyObject *__pyx_n_s_UVRequest;
  PyObject *__pyx_n_s_UVRequest___reduce_cython;
  PyObject *__pyx_n_s_UVRequest___setstate_cython;
  PyObject *__pyx_n_s_UVSocketHandle;
  PyObject *__pyx_n_s_UVSocketHandle___reduce_cython;
  PyObject *__pyx_n_s_UVSocketHandle___setstate_cython;
  PyObject *__pyx_n_s_UVStream;
  PyObject *__pyx_n_s_UVStreamServer;
  PyObject *__pyx_n_s_UVStreamServer___on_ssl_connecte;
  PyObject *__pyx_n_s_UVStreamServer___reduce_cython;
  PyObject *__pyx_n_s_UVStreamServer___setstate_cython;
  PyObject *__pyx_n_s_UVStreamServer__on_ssl_connecte;
  PyObject *__pyx_n_s_UVStream___reduce_cython;
  PyObject *__pyx_n_s_UVStream___setstate_cython;
  PyObject *__pyx_kp_u_UVStream__try_write_called_with;
  PyObject *__pyx_n_s_UVStream_can_write_eof;
  PyObject *__pyx_n_s_UVStream_is_reading;
  PyObject *__pyx_n_s_UVStream_pause_reading;
  PyObject *__pyx_n_s_UVStream_resume_reading;
  PyObject *__pyx_kp_u_UVStream_shutdown_callback_calle;
  PyObject *__pyx_n_s_UVStream_write;
  PyObject *__pyx_kp_u_UVStream_write_callback_called_w;
  PyObject *__pyx_n_s_UVStream_write_eof;
  PyObject *__pyx_n_s_UVStream_writelines;
  PyObject *__pyx_n_s_UVTimer;
  PyObject *__pyx_n_s_UVTimer___reduce_cython;
  PyObject *__pyx_n_s_UVTimer___setstate_cython;
  PyObject *__pyx_kp_u_UVTransport__call_connection_los;
  PyObject *__pyx_kp_u_UVTransport__call_connection_mad;
  PyObject *__pyx_kp_u_Unable_to_check_or_remove_stale;
  PyObject *__pyx_kp_u_Unable_to_clean_up_listening_UNI;
  PyObject *__pyx_kp_u_Unhandled_error_in_exception_han;
  PyObject *__pyx_kp_u_Unhandled_exception_in_event_loo;
  PyObject *__pyx_n_s_UnicodeEncodeError;
  PyObject *__pyx_n_s_UnixServer;
  PyObject *__pyx_n_s_UnixServer___reduce_cython;
  PyObject *__pyx_n_s_UnixServer___setstate_cython;
  PyObject *__pyx_n_s_UnixTransport;
  PyObject *__pyx_n_s_UnixTransport___reduce_cython;
  PyObject *__pyx_n_s_UnixTransport___setstate_cython;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s_WeakSet;
  PyObject *__pyx_n_s_WeakValueDictionary;
  PyObject *__pyx_n_s_WriteSubprocessPipeProto;
  PyObject *__pyx_n_s_WriteSubprocessPipeProto___init;
  PyObject *__pyx_n_s_WriteSubprocessPipeProto___repr;
  PyObject *__pyx_n_s_WriteSubprocessPipeProto_connect;
  PyObject *__pyx_n_s_WriteSubprocessPipeProto_connect_2;
  PyObject *__pyx_n_s_WriteSubprocessPipeProto_pause_w;
  PyObject *__pyx_n_s_WriteSubprocessPipeProto_resume;
  PyObject *__pyx_n_s_WriteUnixTransport;
  PyObject *__pyx_n_s_WriteUnixTransport___reduce_cyth;
  PyObject *__pyx_n_s_WriteUnixTransport___setstate_cy;
  PyObject *__pyx_n_s_WriteUnixTransport_pause_reading;
  PyObject *__pyx_n_s_WriteUnixTransport_resume_readin;
  PyObject *__pyx_kp_u_Write_callbacks_failed;
  PyObject *__pyx_kp_u_Write_contexts_8;
  PyObject *__pyx_kp_u_Write_errors;
  PyObject *__pyx_kp_u_Write_events;
  PyObject *__pyx_kp_u_Write_failed_callbacks;
  PyObject *__pyx_kp_u_Write_without_poll;
  PyObject *__pyx_kp_u_XXX_an_allocated_buffer_in_trans;
  PyObject *__pyx_kp_u_You_must_set_server_hostname_whe;
  PyObject *__pyx_n_s__10;
  PyObject *__pyx_kp_b__10;
  PyObject *__pyx_kp_u__10;
  PyObject *__pyx_kp_u__115;
  PyObject *__pyx_kp_u__128;
  PyObject *__pyx_kp_u__129;
  PyObject *__pyx_kp_u__130;
  PyObject *__pyx_kp_u__140;
  PyObject *__pyx_kp_u__141;
  PyObject *__pyx_kp_b__161;
  PyObject *__pyx_kp_b__170;
  PyObject *__pyx_kp_u__196;
  PyObject *__pyx_n_s__198;
  PyObject *__pyx_n_s__263;
  PyObject *__pyx_kp_u__33;
  PyObject *__pyx_n_s__542;
  PyObject *__pyx_kp_u__61;
  PyObject *__pyx_kp_u__69;
  PyObject *__pyx_kp_u__75;
  PyObject *__pyx_kp_u__78;
  PyObject *__pyx_kp_u__79;
  PyObject *__pyx_n_s_abc;
  PyObject *__pyx_n_s_abort;
  PyObject *__pyx_kp_u_abort_init__closed_is_set;
  PyObject *__pyx_kp_u_abort_init__inited_is_set;
  PyObject *__pyx_n_s_accept;
  PyObject *__pyx_kp_u_accept_method;
  PyObject *__pyx_n_s_aclose;
  PyObject *__pyx_kp_u_active_TimerHandle_is_deallacati;
  PyObject *__pyx_n_s_add;
  PyObject *__pyx_n_s_add_done_callback;
  PyObject *__pyx_n_s_add_reader;
  PyObject *__pyx_n_s_add_signal_handler;
  PyObject *__pyx_kp_u_add_signal_handler_can_only_be_c;
  PyObject *__pyx_n_s_add_writer;
  PyObject *__pyx_n_s_addr;
  PyObject *__pyx_n_s_address;
  PyObject *__pyx_n_s_addrinfo;
  PyObject *__pyx_n_s_addrs;
  PyObject *__pyx_n_s_aenter;
  PyObject *__pyx_n_s_aexit;
  PyObject *__pyx_n_s_ag;
  PyObject *__pyx_n_s_agen;
  PyObject *__pyx_n_s_ai;
  PyObject *__pyx_n_s_ai_cnt;
  PyObject *__pyx_n_s_ai_local;
  PyObject *__pyx_n_s_ai_remote;
  PyObject *__pyx_kp_u_alive_closed;
  PyObject *__pyx_kp_u_all_args_must_be_str_or_bytes;
  PyObject *__pyx_kp_u_all_environment_values_must_be_b;
  PyObject *__pyx_kp_u_all_environment_vars_must_be_byt;
  PyObject *__pyx_n_s_allow_broadcast;
  PyObject *__pyx_kp_u_an_error_occurred_during_closing;
  PyObject *__pyx_n_s_app_protocol;
  PyObject *__pyx_n_s_app_transport;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_u_ascii;
  PyObject *__pyx_n_u_asyncgen;
  PyObject *__pyx_n_s_asyncgen_finalizer_hook;
  PyObject *__pyx_n_s_asyncgen_firstiter_hook;
  PyObject *__pyx_kp_u_asynchronous_generator_r_was_sch;
  PyObject *__pyx_n_s_asyncio;
  PyObject *__pyx_kp_s_asyncio_Handle;
  PyObject *__pyx_n_s_asyncio_base_events;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_asyncio_futures;
  PyObject *__pyx_kp_u_asyncio_is_trying_to_install_it;
  PyObject *__pyx_n_s_asyncio_log;
  PyObject *__pyx_n_s_asyncio_sslproto;
  PyObject *__pyx_n_s_asyncio_tasks;
  PyObject *__pyx_n_s_asyncio_transports;
  PyObject *__pyx_n_s_await;
  PyObject *__pyx_n_u_b;
  PyObject *__pyx_n_s_backlog;
  PyObject *__pyx_n_s_base_events;
  PyObject *__pyx_kp_b_bin_sh;
  PyObject *__pyx_n_s_bind;
  PyObject *__pyx_kp_u_bind_method;
  PyObject *__pyx_n_s_buf;
  PyObject *__pyx_kp_u_buffer_size_is_not_0_after_a_su;
  PyObject *__pyx_n_s_buffer_updated;
  PyObject *__pyx_n_s_bufs;
  PyObject *__pyx_n_s_bufsize;
  PyObject *__pyx_kp_u_bufsize_must_be_0;
  PyObject *__pyx_n_s_builtins;
  PyObject *__pyx_kp_b_c;
  PyObject *__pyx_n_s_c_str_path;
  PyObject *__pyx_n_s_call_at;
  PyObject *__pyx_n_s_call_connection_made;
  PyObject *__pyx_n_s_call_exception_handler;
  PyObject *__pyx_n_s_call_later;
  PyObject *__pyx_n_s_call_soon;
  PyObject *__pyx_n_s_call_soon_threadsafe;
  PyObject *__pyx_n_s_callback;
  PyObject *__pyx_kp_u_called_after_destroying_the_UVH;
  PyObject *__pyx_kp_u_called_with_handle_data_NULL;
  PyObject *__pyx_n_s_can_write_eof;
  PyObject *__pyx_n_s_cancel;
  PyObject *__pyx_kp_u_cancel_request_data_is_NULL;
  PyObject *__pyx_kp_u_cancel_request_data_is_not_UVRe;
  PyObject *__pyx_kp_u_cancel_request_is_NULL;
  PyObject *__pyx_n_s_cancelled;
  PyObject *__pyx_n_u_cancelled;
  PyObject *__pyx_kp_u_cannot_add_a_signal_handler_for;
  PyObject *__pyx_kp_u_cannot_apply_subprocess_STDOUT;
  PyObject *__pyx_kp_u_cannot_cleanup_signal_handlers_c;
  PyObject *__pyx_kp_u_cannot_convert_sockaddr_into_Pyt;
  PyObject *__pyx_kp_u_cannot_open_a_UDP_handle_invalid;
  PyObject *__pyx_kp_u_cannot_pause_signals_handling_no;
  PyObject *__pyx_kp_u_cannot_run_Handle_callback_is_no;
  PyObject *__pyx_kp_u_cannot_run_TimerHandle_callback;
  PyObject *__pyx_kp_u_cannot_shutdown_signals_handling;
  PyObject *__pyx_kp_u_cannot_spawn_a_process_args_are;
  PyObject *__pyx_kp_u_cannot_switch_state_from_to;
  PyObject *__pyx_n_s_cast;
  PyObject *__pyx_n_s_cause;
  PyObject *__pyx_n_s_cfunc_to_py;
  PyObject *__pyx_n_s_chain;
  PyObject *__pyx_n_s_check_default_executor;
  PyObject *__pyx_n_s_check_hostname;
  PyObject *__pyx_n_s_cipher;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cleanup_socket;
  PyObject *__pyx_n_s_clear;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_kp_u_close__handle_data_is_NULL;
  PyObject *__pyx_kp_u_close__handle_data_is_not_UVHa;
  PyObject *__pyx_n_s_close_fds;
  PyObject *__pyx_kp_u_close_method;
  PyObject *__pyx_kp_u_close_uv_is_closing_is_true;
  PyObject *__pyx_kp_u_closed_Transport_handle_and_unse;
  PyObject *__pyx_kp_u_closed_reading_x;
  PyObject *__pyx_kp_u_closed_x;
  PyObject *__pyx_n_s_closing_agens;
  PyObject *__pyx_n_s_cmd;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_s_collections_abc;
  PyObject *__pyx_n_s_completed;
  PyObject *__pyx_n_s_compression;
  PyObject *__pyx_n_s_concurrent;
  PyObject *__pyx_kp_u_concurrent_allocations;
  PyObject *__pyx_n_s_concurrent_futures;
  PyObject *__pyx_n_s_conmade_cb;
  PyObject *__pyx_n_s_connect;
  PyObject *__pyx_n_s_connect_accepted_socket;
  PyObject *__pyx_n_s_connect_ex;
  PyObject *__pyx_kp_u_connect_ex_method;
  PyObject *__pyx_kp_u_connect_failed;
  PyObject *__pyx_kp_u_connect_method;
  PyObject *__pyx_n_s_connect_read_pipe;
  PyObject *__pyx_n_s_connect_write_pipe;
  PyObject *__pyx_n_s_connection_lost;
  PyObject *__pyx_n_s_connection_made;
  PyObject *__pyx_n_s_context;
  PyObject *__pyx_n_u_context;
  PyObject *__pyx_kp_u_context_manager_protocol;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_coro;
  PyObject *__pyx_n_s_coroutines;
  PyObject *__pyx_kp_u_coroutines_cannot_be_used_with_a;
  PyObject *__pyx_kp_u_coroutines_cannot_be_used_with_r;
  PyObject *__pyx_kp_u_could_not_bind_to_local_addr;
  PyObject *__pyx_kp_u_could_not_bind_to_remote_addr;
  PyObject *__pyx_kp_u_could_not_close_attached_file_ob;
  PyObject *__pyx_n_s_create_connection;
  PyObject *__pyx_n_s_create_connection_locals_genexpr;
  PyObject *__pyx_n_s_create_datagram_endpoint;
  PyObject *__pyx_n_s_create_datagram_endpoint_locals;
  PyObject *__pyx_n_s_create_default_context;
  PyObject *__pyx_n_s_create_future;
  PyObject *__pyx_n_s_create_server;
  PyObject *__pyx_kp_u_create_server_failed_to_create_s;
  PyObject *__pyx_n_s_create_task;
  PyObject *__pyx_n_s_create_unix_connection;
  PyObject *__pyx_n_s_create_unix_server;
  PyObject *__pyx_kp_u_created_at;
  PyObject *__pyx_n_s_creationflags;
  PyObject *__pyx_kp_u_creationflags_is_not_supported;
  PyObject *__pyx_n_s_ctx;
  PyObject *__pyx_n_s_cwd;
  PyObject *__pyx_kp_u_cwd_must_be_a_str_or_bytes_objec;
  PyObject *__pyx_n_u_cygwin;
  PyObject *__pyx_n_u_darwin;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_kp_u_data_expecting_a_bytes_like_inst;
  PyObject *__pyx_n_s_data_received;
  PyObject *__pyx_n_s_datagram_received;
  PyObject *__pyx_kp_u_dealloc____handle_is_NULL__cl;
  PyObject *__pyx_kp_u_deallocating_a_running_event_loo;
  PyObject *__pyx_kp_u_deallocating_an_open_event_loop;
  PyObject *__pyx_n_s_debug;
  PyObject *__pyx_n_s_debug_flags;
  PyObject *__pyx_n_u_debug_wrapper;
  PyObject *__pyx_n_s_decode;
  PyObject *__pyx_n_s_decref_socketios;
  PyObject *__pyx_n_s_default;
  PyObject *__pyx_n_s_default_exception_handler;
  PyObject *__pyx_n_s_default_int_handler;
  PyObject *__pyx_n_s_delay;
  PyObject *__pyx_n_s_deque;
  PyObject *__pyx_n_s_detach;
  PyObject *__pyx_kp_u_detach_method;
  PyObject *__pyx_n_s_dev_mode;
  PyObject *__pyx_n_s_devnull;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_n_s_disable;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_discard;
  PyObject *__pyx_n_s_disconnected;
  PyObject *__pyx_n_s_do_handshake;
  PyObject *__pyx_n_s_do_shutdown;
  PyObject *__pyx_n_s_doc;
  PyObject *__pyx_kp_u_does_not_have_a_reader_in_the;
  PyObject *__pyx_kp_u_does_not_have_a_reader_in_the_2;
  PyObject *__pyx_kp_u_does_not_have_a_writer_in_the;
  PyObject *__pyx_kp_u_does_not_have_a_writer_in_the_2;
  PyObject *__pyx_n_s_done;
  PyObject *__pyx_n_s_done_cb;
  PyObject *__pyx_n_s_dup;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_enabled;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_ensure_future;
  PyObject *__pyx_n_s_enter;
  PyObject *__pyx_n_s_enum;
  PyObject *__pyx_n_s_env;
  PyObject *__pyx_n_s_environ;
  PyObject *__pyx_n_s_eof_received;
  PyObject *__pyx_n_s_err;
  PyObject *__pyx_n_s_errno;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_n_s_error_received;
  PyObject *__pyx_kp_u_error_status_in_uv_stream_t_list;
  PyObject *__pyx_kp_u_error_status_in_uv_stream_t_read;
  PyObject *__pyx_kp_u_error_status_in_uv_stream_t_shut;
  PyObject *__pyx_kp_u_error_status_in_uv_stream_t_writ;
  PyObject *__pyx_kp_u_error_while_attempting_to_bind_o;
  PyObject *__pyx_n_s_ex;
  PyObject *__pyx_n_s_exc;
  PyObject *__pyx_n_s_exc_info;
  PyObject *__pyx_n_s_exception;
  PyObject *__pyx_n_u_exception;
  PyObject *__pyx_kp_u_exception_in_Task___repr;
  PyObject *__pyx_n_s_exceptions;
  PyObject *__pyx_n_s_excs;
  PyObject *__pyx_n_s_executable;
  PyObject *__pyx_n_s_executor;
  PyObject *__pyx_n_s_exit;
  PyObject *__pyx_kp_u_expected_AF_INET_AF_INET6_or_AF;
  PyObject *__pyx_kp_u_expected_asyncio_Future;
  PyObject *__pyx_n_s_extract;
  PyObject *__pyx_n_s_f1;
  PyObject *__pyx_n_s_f2;
  PyObject *__pyx_n_s_factory;
  PyObject *__pyx_n_s_family;
  PyObject *__pyx_kp_u_fatal_Could_not_advance__StreamW;
  PyObject *__pyx_kp_u_fatal_sent_0_in_advance_uv_buf;
  PyObject *__pyx_n_s_fd;
  PyObject *__pyx_kp_u_fd_2;
  PyObject *__pyx_n_s_fileno;
  PyObject *__pyx_n_s_fileobj;
  PyObject *__pyx_n_s_finalizer;
  PyObject *__pyx_n_s_firstiter;
  PyObject *__pyx_n_s_flag;
  PyObject *__pyx_n_s_flags;
  PyObject *__pyx_n_s_flowinfo;
  PyObject *__pyx_n_s_flush;
  PyObject *__pyx_n_s_force_close;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_format_list;
  PyObject *__pyx_n_s_frame;
  PyObject *__pyx_n_s_fs;
  PyObject *__pyx_n_s_fs_handle;
  PyObject *__pyx_n_s_fspath;
  PyObject *__pyx_n_s_func;
  PyObject *__pyx_n_s_functools;
  PyObject *__pyx_n_s_fut;
  PyObject *__pyx_n_s_future;
  PyObject *__pyx_n_s_futures;
  PyObject *__pyx_n_s_gaierror;
  PyObject *__pyx_n_s_gather;
  PyObject *__pyx_n_s_gc;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_app_transport;
  PyObject *__pyx_n_s_get_asyncgen_hooks;
  PyObject *__pyx_n_s_get_backend_id;
  PyObject *__pyx_n_s_get_buffer;
  PyObject *__pyx_n_u_get_buffer;
  PyObject *__pyx_n_s_get_coroutine_origin_tracking_de;
  PyObject *__pyx_n_u_get_coroutine_wrapper;
  PyObject *__pyx_n_s_get_debug;
  PyObject *__pyx_n_s_get_event_loop;
  PyObject *__pyx_n_s_get_exception_handler;
  PyObject *__pyx_n_s_get_extra_info;
  PyObject *__pyx_n_s_get_inheritable;
  PyObject *__pyx_n_s_get_loop;
  PyObject *__pyx_n_s_get_pid;
  PyObject *__pyx_n_s_get_pipe_transport;
  PyObject *__pyx_n_s_get_protocol;
  PyObject *__pyx_n_s_get_read_buffer_limits;
  PyObject *__pyx_n_s_get_read_buffer_size;
  PyObject *__pyx_n_s_get_returncode;
  PyObject *__pyx_n_u_get_running_loop;
  PyObject *__pyx_n_s_get_task_factory;
  PyObject *__pyx_n_s_get_write_buffer_limits;
  PyObject *__pyx_n_s_get_write_buffer_size;
  PyObject *__pyx_n_s_getaddrinfo;
  PyObject *__pyx_kp_u_getaddrinfo_returned_empty_list;
  PyObject *__pyx_kp_u_getaddrinfo_returned_empty_list_2;
  PyObject *__pyx_n_s_getfilesystemencoding;
  PyObject *__pyx_n_s_getframe;
  PyObject *__pyx_n_s_getnameinfo;
  PyObject *__pyx_kp_u_getnameinfo_argument_1_must_be_a;
  PyObject *__pyx_kp_u_getnameinfo_flowinfo_must_be_0_1;
  PyObject *__pyx_n_s_getpeercert;
  PyObject *__pyx_n_s_getpeername;
  PyObject *__pyx_n_s_getservbyname;
  PyObject *__pyx_kp_u_getsockaddrarg_scope_id_must_be;
  PyObject *__pyx_n_s_getsockname;
  PyObject *__pyx_n_s_getsockopt;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_gettimeout;
  PyObject *__pyx_n_s_h;
  PyObject *__pyx_n_s_handle;
  PyObject *__pyx_n_u_handle;
  PyObject *__pyx_kp_u_handle_data_is_NULL_in___close_a;
  PyObject *__pyx_kp_u_handle_loop_data_is_NULL_in___en;
  PyObject *__pyx_kp_u_handle_loop_is_NULL_in___ensure;
  PyObject *__pyx_n_s_handler;
  PyObject *__pyx_kp_u_has_handle_0_in___uv_close_handl;
  PyObject *__pyx_kp_u_high;
  PyObject *__pyx_n_s_high_2;
  PyObject *__pyx_n_s_host;
  PyObject *__pyx_kp_u_host_and_port_was_not_specified;
  PyObject *__pyx_kp_u_host_must_be_a_str_or_bytes;
  PyObject *__pyx_kp_u_host_must_be_a_string_or_bytes_o;
  PyObject *__pyx_kp_u_host_port_and_sock_can_not_be_sp;
  PyObject *__pyx_n_s_hosts;
  PyObject *__pyx_n_s_id;
  PyObject *__pyx_n_s_ident;
  PyObject *__pyx_n_u_idna;
  PyObject *__pyx_n_s_ignore_environment;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_u_inf;
  PyObject *__pyx_n_s_info;
  PyObject *__pyx_n_s_infos;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_inspect;
  PyObject *__pyx_kp_u_integer_argument_expected_got;
  PyObject *__pyx_kp_u_invalid_Handle_cb_type;
  PyObject *__pyx_kp_u_invalid__init_protocol_call;
  PyObject *__pyx_kp_u_invalid_data_in_writebuf_an_inst;
  PyObject *__pyx_kp_u_invalid_socket_family_expected_A;
  PyObject *__pyx_kp_u_invalid_waiter_object;
  PyObject *__pyx_n_s_io_refs;
  PyObject *__pyx_n_s_ioctl;
  PyObject *__pyx_n_s_is_alive;
  PyObject *__pyx_kp_u_is_alive_but__handle_data_is_no;
  PyObject *__pyx_kp_u_is_alive_but__handle_is_NULL;
  PyObject *__pyx_kp_u_is_alive_but__handle_loop_is_no;
  PyObject *__pyx_kp_u_is_alive_but__loop_is_None;
  PyObject *__pyx_kp_u_is_already_being_awaited_on_ser;
  PyObject *__pyx_n_s_is_closed;
  PyObject *__pyx_kp_u_is_closed_2;
  PyObject *__pyx_n_s_is_closing;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_is_longer_than_107_characters;
  PyObject *__pyx_kp_u_is_not_supported_by_start_tls;
  PyObject *__pyx_kp_u_is_not_supposed_to_be_instantia;
  PyObject *__pyx_kp_u_is_open_in___dealloc___with_loo;
  PyObject *__pyx_n_s_is_reading;
  PyObject *__pyx_n_s_is_running;
  PyObject *__pyx_n_s_is_serving;
  PyObject *__pyx_n_s_iscoroutine;
  PyObject *__pyx_n_s_iscoroutinefunction;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_u_isfuture;
  PyObject *__pyx_n_s_isgenerator;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_itertools;
  PyObject *__pyx_n_s_join;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_kill;
  PyObject *__pyx_n_s_kwargs;
  PyObject *__pyx_kp_u_laddr_s;
  PyObject *__pyx_n_s_lads;
  PyObject *__pyx_n_s_lai;
  PyObject *__pyx_n_s_lai_addr_static;
  PyObject *__pyx_n_s_lai_iter;
  PyObject *__pyx_n_s_lai_static;
  PyObject *__pyx_n_s_last;
  PyObject *__pyx_n_s_libuv_get_loop_t_ptr;
  PyObject *__pyx_n_s_libuv_get_version;
  PyObject *__pyx_n_s_limit;
  PyObject *__pyx_n_s_list_of_data;
  PyObject *__pyx_n_s_listen;
  PyObject *__pyx_kp_u_listen_method;
  PyObject *__pyx_n_s_local_addr;
  PyObject *__pyx_kp_u_local_addr_must_be_a_tuple_of_ho;
  PyObject *__pyx_kp_u_local_addr_must_be_a_tuple_of_ho_2;
  PyObject *__pyx_n_s_log;
  PyObject *__pyx_n_s_log_destroy_pending;
  PyObject *__pyx_n_s_log_lines;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_lookup_lines;
  PyObject *__pyx_n_s_loop;
  PyObject *__pyx_kp_u_loop__exec_queued_writes;
  PyObject *__pyx_kp_u_loop__on_idle;
  PyObject *__pyx_n_s_low;
  PyObject *__pyx_n_s_lower;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_main_thread;
  PyObject *__pyx_n_s_makefile;
  PyObject *__pyx_kp_u_makefile_method;
  PyObject *__pyx_n_s_maxsize;
  PyObject *__pyx_kp_u_maxsize_is_expected_to_be_greate;
  PyObject *__pyx_n_s_message;
  PyObject *__pyx_n_u_message;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_mode;
  PyObject *__pyx_n_s_model;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_monitor_fs;
  PyObject *__pyx_n_s_monotonic;
  PyObject *__pyx_n_s_move_to_end;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_msg;
  PyObject *__pyx_kp_u_must_be_0;
  PyObject *__pyx_kp_u_must_be_low;
  PyObject *__pyx_n_s_n;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_u_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_nbytes;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_kp_u_new_callbacks_were_queued_during;
  PyObject *__pyx_kp_u_new_poll_handles_were_queued_dur;
  PyObject *__pyx_n_s_new_task;
  PyObject *__pyx_kp_u_new_timers_were_queued_during_lo;
  PyObject *__pyx_n_s_nexc;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_kp_u_no_path_and_sock_were_specified;
  PyObject *__pyx_kp_u_no_python_buffer_is_allocated_in;
  PyObject *__pyx_n_s_noop;
  PyObject *__pyx_n_s_noop_2;
  PyObject *__pyx_n_s_obj;
  PyObject *__pyx_n_s_object;
  PyObject *__pyx_kp_u_object_created_at_most_recent_ca;
  PyObject *__pyx_kp_u_objs_handles;
  PyObject *__pyx_n_s_old_agen_hooks;
  PyObject *__pyx_n_s_on_ssl_connected;
  PyObject *__pyx_n_s_open;
  PyObject *__pyx_kp_u_open__StreamWriteContext_is_bein;
  PyObject *__pyx_kp_u_open__UDPSendContext_is_being_de;
  PyObject *__pyx_n_s_opts;
  PyObject *__pyx_n_s_orig_path;
  PyObject *__pyx_n_s_os;
  PyObject *__pyx_n_s_p_bytes;
  PyObject *__pyx_n_s_partial;
  PyObject *__pyx_n_s_pass_fds;
  PyObject *__pyx_n_s_path;
  PyObject *__pyx_kp_u_path_and_sock_can_not_be_specifi;
  PyObject *__pyx_kp_u_path_was_not_specified_and_no_so;
  PyObject *__pyx_n_s_pause_reading;
  PyObject *__pyx_n_s_pause_writing;
  PyObject *__pyx_n_s_peercert;
  PyObject *__pyx_n_u_peername;
  PyObject *__pyx_n_s_pending;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_pipe;
  PyObject *__pyx_n_u_pipe;
  PyObject *__pyx_kp_u_pipe_2;
  PyObject *__pyx_n_s_pipe_connection_lost;
  PyObject *__pyx_n_s_pipe_data_received;
  PyObject *__pyx_n_s_platform;
  PyObject *__pyx_n_s_pop;
  PyObject *__pyx_n_s_popitem;
  PyObject *__pyx_n_s_popleft;
  PyObject *__pyx_n_s_port;
  PyObject *__pyx_kp_u_port_must_be_a_str_bytes_or_int;
  PyObject *__pyx_n_u_posix;
  PyObject *__pyx_n_s_preexec_fn;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_print;
  PyObject *__pyx_n_u_print_debug_info;
  PyObject *__pyx_n_s_problems;
  PyObject *__pyx_n_s_proc;
  PyObject *__pyx_n_s_process_exited;
  PyObject *__pyx_n_s_process_id;
  PyObject *__pyx_n_s_program;
  PyObject *__pyx_n_s_proto;
  PyObject *__pyx_n_s_proto_factory;
  PyObject *__pyx_n_s_protocol;
  PyObject *__pyx_n_u_protocol;
  PyObject *__pyx_n_s_protocol_factory;
  PyObject *__pyx_kp_u_protocol_is_not_set_cannot_call;
  PyObject *__pyx_kp_u_protocol_is_required;
  PyObject *__pyx_kp_u_protocol_pause_writing_failed;
  PyObject *__pyx_kp_u_protocol_resume_writing_failed;
  PyObject *__pyx_n_s_pyaddr;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_LruCache;
  PyObject *__pyx_n_s_pyx_unpickle_PseudoSocket;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_u_qualname;
  PyObject *__pyx_kp_u_r_SSL_handshake_took_1f_ms;
  PyObject *__pyx_kp_u_r_fileno_returned_non_integer;
  PyObject *__pyx_kp_u_r_received_EOF;
  PyObject *__pyx_kp_u_r_s;
  PyObject *__pyx_kp_u_r_starts_SSL_handshake;
  PyObject *__pyx_kp_u_raddr_s;
  PyObject *__pyx_n_s_rads;
  PyObject *__pyx_n_s_rai;
  PyObject *__pyx_n_s_rai_addr_static;
  PyObject *__pyx_n_s_rai_iter;
  PyObject *__pyx_n_s_rai_static;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_read;
  PyObject *__pyx_n_s_recv;
  PyObject *__pyx_n_s_recv_into;
  PyObject *__pyx_kp_u_recv_into_method;
  PyObject *__pyx_kp_u_recv_method;
  PyObject *__pyx_n_s_recvfrom;
  PyObject *__pyx_n_s_recvfrom_into;
  PyObject *__pyx_kp_u_recvfrom_into_method;
  PyObject *__pyx_kp_u_recvfrom_method;
  PyObject *__pyx_n_s_recvmsg;
  PyObject *__pyx_n_s_recvmsg_into;
  PyObject *__pyx_kp_u_recvmsg_into_method;
  PyObject *__pyx_kp_u_recvmsg_method;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_remote_addr;
  PyObject *__pyx_kp_u_remote_addr_must_be_a_tuple_of_h;
  PyObject *__pyx_n_s_remove;
  PyObject *__pyx_n_s_remove_done_callback;
  PyObject *__pyx_n_s_remove_reader;
  PyObject *__pyx_n_s_remove_reader_2;
  PyObject *__pyx_n_s_remove_signal_handler;
  PyObject *__pyx_kp_u_remove_signal_handler_can_only_b;
  PyObject *__pyx_n_s_remove_writer;
  PyObject *__pyx_n_s_remove_writer_2;
  PyObject *__pyx_n_s_repr;
  PyObject *__pyx_n_s_restore_signals;
  PyObject *__pyx_n_s_result;
  PyObject *__pyx_n_s_results;
  PyObject *__pyx_n_s_resume_cb;
  PyObject *__pyx_n_s_resume_reading;
  PyObject *__pyx_n_s_resume_writing;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_return_exceptions;
  PyObject *__pyx_kp_u_returning_true_from_eof_received;
  PyObject *__pyx_n_s_reuse_address;
  PyObject *__pyx_n_s_reuse_port;
  PyObject *__pyx_kp_u_reuse_port_not_supported_by_sock;
  PyObject *__pyx_n_s_reverse;
  PyObject *__pyx_n_s_round;
  PyObject *__pyx_n_s_rstrip;
  PyObject *__pyx_n_s_run;
  PyObject *__pyx_n_s_run_forever;
  PyObject *__pyx_n_s_run_in_executor;
  PyObject *__pyx_n_s_run_until_complete;
  PyObject *__pyx_n_s_run_until_complete_locals_done_c;
  PyObject *__pyx_kp_u_running_closed_debug;
  PyObject *__pyx_n_s_scope_id;
  PyObject *__pyx_kp_u_seconds;
  PyObject *__pyx_kp_u_seconds_aborting_the_connection;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_self_2;
  PyObject *__pyx_n_u_self_2;
  PyObject *__pyx_kp_u_self_pipe_exists_before_loop_run;
  PyObject *__pyx_kp_u_self_pipe_was_not_cleaned_up_aft;
  PyObject *__pyx_kp_s_self_py_buf_self_req_cannot_be_c;
  PyObject *__pyx_kp_s_self_py_bufs_self_py_bufs_sml_se;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_kp_u_send_method;
  PyObject *__pyx_n_s_send_signal;
  PyObject *__pyx_n_s_sendall;
  PyObject *__pyx_kp_u_sendall_method;
  PyObject *__pyx_n_s_sendfile;
  PyObject *__pyx_kp_u_sendfile_method;
  PyObject *__pyx_n_s_sendmsg;
  PyObject *__pyx_n_s_sendmsg_afalg;
  PyObject *__pyx_kp_u_sendmsg_afalg_method;
  PyObject *__pyx_kp_u_sendmsg_method;
  PyObject *__pyx_n_s_sendto;
  PyObject *__pyx_kp_u_sendto_method;
  PyObject *__pyx_n_s_serve_forever;
  PyObject *__pyx_kp_u_server;
  PyObject *__pyx_n_s_server_2;
  PyObject *__pyx_n_s_server_hostname;
  PyObject *__pyx_kp_u_server_hostname_is_only_meaningf;
  PyObject *__pyx_n_s_server_side;
  PyObject *__pyx_n_s_servers;
  PyObject *__pyx_kp_u_service_proto_not_found;
  PyObject *__pyx_n_s_set_asyncgen_hooks;
  PyObject *__pyx_n_s_set_coroutine_origin_tracking_de;
  PyObject *__pyx_n_u_set_coroutine_wrapper;
  PyObject *__pyx_n_s_set_debug;
  PyObject *__pyx_n_s_set_default_executor;
  PyObject *__pyx_n_s_set_exception;
  PyObject *__pyx_n_s_set_exception_handler;
  PyObject *__pyx_n_s_set_inheritable;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_set_name_2;
  PyObject *__pyx_n_s_set_protocol;
  PyObject *__pyx_n_s_set_read_buffer_limits;
  PyObject *__pyx_n_s_set_result;
  PyObject *__pyx_n_u_set_running_loop;
  PyObject *__pyx_n_s_set_task_factory;
  PyObject *__pyx_n_s_set_wakeup_fd;
  PyObject *__pyx_kp_u_set_wakeup_fd_1_failed_s;
  PyObject *__pyx_n_s_set_write_buffer_limits;
  PyObject *__pyx_n_s_setblocking;
  PyObject *__pyx_kp_u_setblocking_transport_sockets_ca;
  PyObject *__pyx_n_s_setsockopt;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_settimeout;
  PyObject *__pyx_kp_u_settimeout_only_0_timeout_is_all;
  PyObject *__pyx_n_s_share;
  PyObject *__pyx_n_s_shell;
  PyObject *__pyx_kp_u_shell_must_be_False;
  PyObject *__pyx_kp_u_shell_must_be_True;
  PyObject *__pyx_n_s_shutdown;
  PyObject *__pyx_n_s_shutdown_asyncgens;
  PyObject *__pyx_n_s_shutdown_coro;
  PyObject *__pyx_n_s_shutdown_default_executor;
  PyObject *__pyx_kp_u_shutdown_method;
  PyObject *__pyx_n_s_sig;
  PyObject *__pyx_kp_u_sig_cannot_be_caught;
  PyObject *__pyx_kp_u_sig_must_be_an_int_not_r;
  PyObject *__pyx_kp_u_sig_out_of_range_1;
  PyObject *__pyx_n_s_sighandler;
  PyObject *__pyx_n_s_siginterrupt;
  PyObject *__pyx_n_s_signal;
  PyObject *__pyx_kp_u_signals_handling_has_been_alread;
  PyObject *__pyx_kp_u_signals_handling_has_not_been_se;
  PyObject *__pyx_n_s_signum;
  PyObject *__pyx_n_s_sin6;
  PyObject *__pyx_n_s_sl;
  PyObject *__pyx_n_s_sleep;
  PyObject *__pyx_n_s_sock;
  PyObject *__pyx_n_s_sock_accept;
  PyObject *__pyx_n_s_sock_connect;
  PyObject *__pyx_kp_u_sock_connect_cb_is_called_on_a;
  PyObject *__pyx_n_s_sock_recv;
  PyObject *__pyx_n_s_sock_recv_into;
  PyObject *__pyx_kp_u_sock_recv_into_is_called_on_a_c;
  PyObject *__pyx_kp_u_sock_recv_is_called_on_a_cancel;
  PyObject *__pyx_n_s_sock_recvfrom;
  PyObject *__pyx_n_s_sock_recvfrom_into;
  PyObject *__pyx_n_s_sock_sendall;
  PyObject *__pyx_kp_u_sock_sendall_is_called_on_a_can;
  PyObject *__pyx_n_s_sock_sendto;
  PyObject *__pyx_n_s_sockaddr;
  PyObject *__pyx_kp_u_sockaddr_must_be_a_tuple_of_2_3;
  PyObject *__pyx_kp_u_sockaddr_resolved_to_multiple_ad;
  PyObject *__pyx_kp_u_socket;
  PyObject *__pyx_n_s_socket_2;
  PyObject *__pyx_n_u_socket_2;
  PyObject *__pyx_kp_u_socket_family_mismatch_or_a_DNS;
  PyObject *__pyx_kp_u_socket_modifier_keyword_argument;
  PyObject *__pyx_kp_u_socket_send_raised_exception;
  PyObject *__pyx_n_s_socketpair;
  PyObject *__pyx_kp_u_sockets;
  PyObject *__pyx_n_s_sockets_2;
  PyObject *__pyx_n_u_sockname;
  PyObject *__pyx_n_s_source;
  PyObject *__pyx_n_u_source_traceback;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_split;
  PyObject *__pyx_n_s_ssl;
  PyObject *__pyx_kp_u_ssl_argument_must_be_an_SSLConte;
  PyObject *__pyx_n_s_ssl_handshake_timeout;
  PyObject *__pyx_kp_u_ssl_handshake_timeout_is_only_me;
  PyObject *__pyx_kp_u_ssl_handshake_timeout_should_be;
  PyObject *__pyx_kp_u_ssl_is_expected_to_be_None_or_an;
  PyObject *__pyx_n_s_ssl_object;
  PyObject *__pyx_n_s_ssl_protocol;
  PyObject *__pyx_n_s_ssl_shutdown_timeout;
  PyObject *__pyx_kp_u_ssl_shutdown_timeout_is_only_mea;
  PyObject *__pyx_kp_u_ssl_shutdown_timeout_should_be_a;
  PyObject *__pyx_n_s_ssl_waiter;
  PyObject *__pyx_n_s_sslcontext;
  PyObject *__pyx_kp_u_sslcontext_is_expected_to_be_an;
  PyObject *__pyx_n_s_sslproto;
  PyObject *__pyx_n_s_st_ino;
  PyObject *__pyx_n_s_st_mode;
  PyObject *__pyx_n_s_stacklevel;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_kp_u_start_init_can_only_be_called;
  PyObject *__pyx_n_s_start_new_session;
  PyObject *__pyx_n_s_start_serving;
  PyObject *__pyx_n_s_start_tls;
  PyObject *__pyx_n_s_startupinfo;
  PyObject *__pyx_kp_u_startupinfo_is_not_supported;
  PyObject *__pyx_n_s_stat;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_stderr;
  PyObject *__pyx_n_s_stdin;
  PyObject *__pyx_n_s_stdio_fut;
  PyObject *__pyx_n_s_stdio_inited;
  PyObject *__pyx_n_s_stdout;
  PyObject *__pyx_n_s_stop;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_kp_u_stream_is_open_after_UVStream__t;
  PyObject *__pyx_n_s_strerror;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_submit;
  PyObject *__pyx_n_s_subprocess;
  PyObject *__pyx_kp_u_subprocess_STDOUT_is_supported_o;
  PyObject *__pyx_n_s_subprocess_exec;
  PyObject *__pyx_n_s_subprocess_run;
  PyObject *__pyx_n_s_subprocess_shell;
  PyObject *__pyx_n_s_subtract;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_sys;
  PyObject *__pyx_n_s_target;
  PyObject *__pyx_n_s_task;
  PyObject *__pyx_n_u_task;
  PyObject *__pyx_kp_u_task_factory_must_be_a_callable;
  PyObject *__pyx_n_s_tb;
  PyObject *__pyx_n_s_tcp;
  PyObject *__pyx_n_u_tcp;
  PyObject *__pyx_n_s_terminate;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_test__append_write_backlog;
  PyObject *__pyx_n_s_test_coroutine_1;
  PyObject *__pyx_n_s_testhelper_unwrap_capsuled_poin;
  PyObject *__pyx_kp_u_the_socket_must_be_non_blocking;
  PyObject *__pyx_kp_u_this_event_loop_is_already_runni;
  PyObject *__pyx_n_s_thread;
  PyObject *__pyx_n_s_threading;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_s_timeout;
  PyObject *__pyx_n_s_tr;
  PyObject *__pyx_n_s_traceback;
  PyObject *__pyx_n_s_traceback_2;
  PyObject *__pyx_n_s_transp;
  PyObject *__pyx_n_s_transport;
  PyObject *__pyx_n_u_transport;
  PyObject *__pyx_kp_u_transport_2;
  PyObject *__pyx_kp_u_transport_sockets_do_not_support;
  PyObject *__pyx_n_s_transport_waiter;
  PyObject *__pyx_n_s_transports;
  PyObject *__pyx_kp_u_try_write_sent_all_data_and_ret;
  PyObject *__pyx_n_s_type;
  PyObject *__pyx_n_s_udp;
  PyObject *__pyx_n_u_udp;
  PyObject *__pyx_kp_u_unable_to_listen_no_protocol_fac;
  PyObject *__pyx_kp_u_unable_to_perform_operation_on_r;
  PyObject *__pyx_kp_u_unable_to_start_the_loop_it_was;
  PyObject *__pyx_kp_u_unclosed_resource_r;
  PyObject *__pyx_kp_u_unclosed_resource_r_2;
  PyObject *__pyx_kp_u_unclosed_resource_r_could_not_se;
  PyObject *__pyx_kp_u_unclosed_transport_uvloop_loop;
  PyObject *__pyx_kp_u_unexpected_address_family;
  PyObject *__pyx_kp_u_unexpected_kwargs;
  PyObject *__pyx_kp_u_unhandled_error_or_an_empty_buff;
  PyObject *__pyx_n_s_universal_newlines;
  PyObject *__pyx_kp_u_universal_newlines_must_be_False;
  PyObject *__pyx_kp_u_unix_socket_path;
  PyObject *__pyx_n_s_unlink;
  PyObject *__pyx_kp_u_unopened_TCPServer;
  PyObject *__pyx_n_s_unset;
  PyObject *__pyx_n_s_unwrap;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_kp_u_uv_close_handle_cb_handle_loop;
  PyObject *__pyx_kp_u_uv_handle_t_current_freed_total;
  PyObject *__pyx_kp_u_uvloop_PseudoSocket_fd_family_s;
  PyObject *__pyx_kp_s_uvloop_cbhandles_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_basetransport_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_fsevent_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_pipe_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_process_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_stream_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_streamserver_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_tcp_pyx;
  PyObject *__pyx_kp_s_uvloop_handles_udp_pyx;
  PyObject *__pyx_n_s_uvloop_loop;
  PyObject *__pyx_kp_s_uvloop_loop_pyx;
  PyObject *__pyx_kp_s_uvloop_pseudosock_pyx;
  PyObject *__pyx_kp_s_uvloop_server_pyx;
  PyObject *__pyx_n_u_uvloop_sleep_after_fork;
  PyObject *__pyx_kp_u_uvloop_sslproto;
  PyObject *__pyx_kp_s_uvloop_sslproto_pyx;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_n_s_version_info;
  PyObject *__pyx_n_s_wait;
  PyObject *__pyx_n_s_wait_2;
  PyObject *__pyx_n_s_wait_closed;
  PyObject *__pyx_n_s_waiter;
  PyObject *__pyx_n_s_walk_stack;
  PyObject *__pyx_n_s_want;
  PyObject *__pyx_n_s_warn;
  PyObject *__pyx_n_s_warn_on_full_buffer;
  PyObject *__pyx_n_s_warning;
  PyObject *__pyx_n_s_warnings;
  PyObject *__pyx_n_u_wb;
  PyObject *__pyx_n_s_weakref;
  PyObject *__pyx_n_s_when;
  PyObject *__pyx_kp_u_without_no_gc_clear_loop_was_se;
  PyObject *__pyx_n_s_wrap;
  PyObject *__pyx_n_s_wrap_bio;
  PyObject *__pyx_n_s_wrap_future;
  PyObject *__pyx_n_s_write;
  PyObject *__pyx_n_s_write_eof;
  PyObject *__pyx_n_s_writelines;
  PyObject *__pyx_n_s_zip;
  PyObject *__pyx_float_0_1;
  PyObject *__pyx_float_1e3;
  PyObject *__pyx_float_1eneg_3;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_3;
  PyObject *__pyx_int_15;
  PyObject *__pyx_int_42;
  PyObject *__pyx_int_100;
  PyObject *__pyx_int_1000;
  PyObject *__pyx_int_50000;
  PyObject *__pyx_int_65536;
  PyObject *__pyx_int_1048575;
  PyObject *__pyx_int_14808274;
  PyObject *__pyx_int_87255255;
  PyObject *__pyx_int_121062814;
  PyObject *__pyx_int_143395587;
  PyObject *__pyx_int_173349396;
  PyObject *__pyx_int_200977156;
  PyObject *__pyx_int_4294967296;
  PyObject *__pyx_int_neg_1;
  int __pyx_k__52;
  int __pyx_k__53;
  PyObject *__pyx_k__71;
  PyObject *__pyx_k__95;
  PyObject *__pyx_k__96;
  PyObject *__pyx_k__97;
  PyObject *__pyx_k__114;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__3;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_tuple__5;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_slice__88;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__18;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__23;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__27;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__29;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__31;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__39;
  PyObject *__pyx_tuple__40;
  PyObject *__pyx_tuple__42;
  PyObject *__pyx_tuple__45;
  PyObject *__pyx_tuple__46;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__48;
  PyObject *__pyx_tuple__49;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__55;
  PyObject *__pyx_tuple__56;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__58;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__60;
  PyObject *__pyx_tuple__62;
  PyObject *__pyx_tuple__64;
  PyObject *__pyx_tuple__65;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__67;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_tuple__70;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__74;
  PyObject *__pyx_tuple__77;
  PyObject *__pyx_tuple__80;
  PyObject *__pyx_tuple__81;
  PyObject *__pyx_tuple__83;
  PyObject *__pyx_tuple__93;
  PyObject *__pyx_tuple__94;
  PyObject *__pyx_tuple__99;
  PyObject *__pyx_codeobj__2;
  PyObject *__pyx_tuple__100;
  PyObject *__pyx_tuple__101;
  PyObject *__pyx_tuple__102;
  PyObject *__pyx_tuple__104;
  PyObject *__pyx_tuple__106;
  PyObject *__pyx_tuple__109;
  PyObject *__pyx_tuple__110;
  PyObject *__pyx_tuple__111;
  PyObject *__pyx_tuple__112;
  PyObject *__pyx_tuple__113;
  PyObject *__pyx_tuple__117;
  PyObject *__pyx_tuple__118;
  PyObject *__pyx_tuple__119;
  PyObject *__pyx_tuple__120;
  PyObject *__pyx_tuple__121;
  PyObject *__pyx_tuple__122;
  PyObject *__pyx_tuple__125;
  PyObject *__pyx_tuple__126;
  PyObject *__pyx_tuple__127;
  PyObject *__pyx_tuple__131;
  PyObject *__pyx_tuple__132;
  PyObject *__pyx_tuple__133;
  PyObject *__pyx_tuple__134;
  PyObject *__pyx_tuple__135;
  PyObject *__pyx_tuple__136;
  PyObject *__pyx_tuple__137;
  PyObject *__pyx_tuple__138;
  PyObject *__pyx_tuple__139;
  PyObject *__pyx_tuple__142;
  PyObject *__pyx_tuple__143;
  PyObject *__pyx_tuple__144;
  PyObject *__pyx_tuple__145;
  PyObject *__pyx_tuple__146;
  PyObject *__pyx_tuple__147;
  PyObject *__pyx_tuple__148;
  PyObject *__pyx_tuple__149;
  PyObject *__pyx_tuple__150;
  PyObject *__pyx_tuple__151;
  PyObject *__pyx_tuple__152;
  PyObject *__pyx_tuple__153;
  PyObject *__pyx_tuple__154;
  PyObject *__pyx_tuple__155;
  PyObject *__pyx_tuple__156;
  PyObject *__pyx_tuple__157;
  PyObject *__pyx_tuple__158;
  PyObject *__pyx_tuple__159;
  PyObject *__pyx_tuple__160;
  PyObject *__pyx_tuple__162;
  PyObject *__pyx_tuple__163;
  PyObject *__pyx_tuple__164;
  PyObject *__pyx_tuple__165;
  PyObject *__pyx_tuple__166;
  PyObject *__pyx_tuple__167;
  PyObject *__pyx_tuple__168;
  PyObject *__pyx_tuple__169;
  PyObject *__pyx_tuple__171;
  PyObject *__pyx_tuple__172;
  PyObject *__pyx_tuple__173;
  PyObject *__pyx_tuple__174;
  PyObject *__pyx_tuple__175;
  PyObject *__pyx_tuple__176;
  PyObject *__pyx_tuple__177;
  PyObject *__pyx_tuple__178;
  PyObject *__pyx_tuple__179;
  PyObject *__pyx_tuple__180;
  PyObject *__pyx_tuple__181;
  PyObject *__pyx_tuple__182;
  PyObject *__pyx_tuple__183;
  PyObject *__pyx_tuple__184;
  PyObject *__pyx_tuple__185;
  PyObject *__pyx_tuple__186;
  PyObject *__pyx_tuple__187;
  PyObject *__pyx_tuple__188;
  PyObject *__pyx_tuple__195;
  PyObject *__pyx_tuple__197;
  PyObject *__pyx_tuple__199;
  PyObject *__pyx_tuple__200;
  PyObject *__pyx_tuple__201;
  PyObject *__pyx_tuple__202;
  PyObject *__pyx_tuple__203;
  PyObject *__pyx_tuple__204;
  PyObject *__pyx_tuple__205;
  PyObject *__pyx_tuple__206;
  PyObject *__pyx_tuple__207;
  PyObject *__pyx_tuple__209;
  PyObject *__pyx_tuple__211;
  PyObject *__pyx_tuple__213;
  PyObject *__pyx_tuple__215;
  PyObject *__pyx_tuple__217;
  PyObject *__pyx_tuple__221;
  PyObject *__pyx_tuple__229;
  PyObject *__pyx_tuple__231;
  PyObject *__pyx_tuple__234;
  PyObject *__pyx_tuple__236;
  PyObject *__pyx_tuple__237;
  PyObject *__pyx_tuple__238;
  PyObject *__pyx_tuple__239;
  PyObject *__pyx_tuple__240;
  PyObject *__pyx_tuple__241;
  PyObject *__pyx_tuple__242;
  PyObject *__pyx_tuple__243;
  PyObject *__pyx_tuple__244;
  PyObject *__pyx_tuple__245;
  PyObject *__pyx_tuple__246;
  PyObject *__pyx_tuple__249;
  PyObject *__pyx_tuple__251;
  PyObject *__pyx_tuple__253;
  PyObject *__pyx_tuple__255;
  PyObject *__pyx_tuple__259;
  PyObject *__pyx_tuple__260;
  PyObject *__pyx_tuple__261;
  PyObject *__pyx_tuple__262;
  PyObject *__pyx_tuple__264;
  PyObject *__pyx_tuple__265;
  PyObject *__pyx_tuple__266;
  PyObject *__pyx_tuple__267;
  PyObject *__pyx_tuple__268;
  PyObject *__pyx_tuple__269;
  PyObject *__pyx_tuple__271;
  PyObject *__pyx_tuple__273;
  PyObject *__pyx_tuple__274;
  PyObject *__pyx_tuple__275;
  PyObject *__pyx_tuple__276;
  PyObject *__pyx_tuple__277;
  PyObject *__pyx_tuple__279;
  PyObject *__pyx_tuple__281;
  PyObject *__pyx_tuple__282;
  PyObject *__pyx_tuple__285;
  PyObject *__pyx_tuple__288;
  PyObject *__pyx_tuple__289;
  PyObject *__pyx_tuple__290;
  PyObject *__pyx_tuple__293;
  PyObject *__pyx_tuple__295;
  PyObject *__pyx_tuple__298;
  PyObject *__pyx_tuple__300;
  PyObject *__pyx_tuple__303;
  PyObject *__pyx_tuple__321;
  PyObject *__pyx_tuple__325;
  PyObject *__pyx_tuple__327;
  PyObject *__pyx_tuple__330;
  PyObject *__pyx_tuple__333;
  PyObject *__pyx_tuple__336;
  PyObject *__pyx_tuple__357;
  PyObject *__pyx_tuple__360;
  PyObject *__pyx_tuple__363;
  PyObject *__pyx_tuple__365;
  PyObject *__pyx_tuple__385;
  PyObject *__pyx_tuple__387;
  PyObject *__pyx_tuple__393;
  PyObject *__pyx_tuple__396;
  PyObject *__pyx_tuple__402;
  PyObject *__pyx_tuple__404;
  PyObject *__pyx_tuple__413;
  PyObject *__pyx_tuple__431;
  PyObject *__pyx_tuple__433;
  PyObject *__pyx_tuple__448;
  PyObject *__pyx_tuple__454;
  PyObject *__pyx_tuple__458;
  PyObject *__pyx_tuple__463;
  PyObject *__pyx_tuple__467;
  PyObject *__pyx_tuple__469;
  PyObject *__pyx_tuple__513;
  PyObject *__pyx_tuple__517;
  PyObject *__pyx_tuple__519;
  PyObject *__pyx_tuple__528;
  PyObject *__pyx_tuple__533;
  PyObject *__pyx_tuple__534;
  PyObject *__pyx_tuple__539;
  PyObject *__pyx_codeobj__20;
  PyObject *__pyx_codeobj__22;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__84;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__87;
  PyObject *__pyx_codeobj__89;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__91;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__98;
  PyObject *__pyx_codeobj__103;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__107;
  PyObject *__pyx_codeobj__108;
  PyObject *__pyx_codeobj__116;
  PyObject *__pyx_codeobj__123;
  PyObject *__pyx_codeobj__124;
  PyObject *__pyx_codeobj__189;
  PyObject *__pyx_codeobj__190;
  PyObject *__pyx_codeobj__191;
  PyObject *__pyx_codeobj__192;
  PyObject *__pyx_codeobj__193;
  PyObject *__pyx_codeobj__194;
  PyObject *__pyx_codeobj__208;
  PyObject *__pyx_codeobj__210;
  PyObject *__pyx_codeobj__212;
  PyObject *__pyx_codeobj__214;
  PyObject *__pyx_codeobj__216;
  PyObject *__pyx_codeobj__218;
  PyObject *__pyx_codeobj__219;
  PyObject *__pyx_codeobj__220;
  PyObject *__pyx_codeobj__222;
  PyObject *__pyx_codeobj__223;
  PyObject *__pyx_codeobj__224;
  PyObject *__pyx_codeobj__225;
  PyObject *__pyx_codeobj__226;
  PyObject *__pyx_codeobj__227;
  PyObject *__pyx_codeobj__228;
  PyObject *__pyx_codeobj__230;
  PyObject *__pyx_codeobj__232;
  PyObject *__pyx_codeobj__233;
  PyObject *__pyx_codeobj__235;
  PyObject *__pyx_codeobj__247;
  PyObject *__pyx_codeobj__248;
  PyObject *__pyx_codeobj__250;
  PyObject *__pyx_codeobj__252;
  PyObject *__pyx_codeobj__254;
  PyObject *__pyx_codeobj__256;
  PyObject *__pyx_codeobj__257;
  PyObject *__pyx_codeobj__258;
  PyObject *__pyx_codeobj__270;
  PyObject *__pyx_codeobj__272;
  PyObject *__pyx_codeobj__278;
  PyObject *__pyx_codeobj__280;
  PyObject *__pyx_codeobj__283;
  PyObject *__pyx_codeobj__284;
  PyObject *__pyx_codeobj__286;
  PyObject *__pyx_codeobj__287;
  PyObject *__pyx_codeobj__291;
  PyObject *__pyx_codeobj__292;
  PyObject *__pyx_codeobj__294;
  PyObject *__pyx_codeobj__296;
  PyObject *__pyx_codeobj__297;
  PyObject *__pyx_codeobj__299;
  PyObject *__pyx_codeobj__301;
  PyObject *__pyx_codeobj__302;
  PyObject *__pyx_codeobj__304;
  PyObject *__pyx_codeobj__305;
  PyObject *__pyx_codeobj__306;
  PyObject *__pyx_codeobj__307;
  PyObject *__pyx_codeobj__308;
  PyObject *__pyx_codeobj__309;
  PyObject *__pyx_codeobj__310;
  PyObject *__pyx_codeobj__311;
  PyObject *__pyx_codeobj__312;
  PyObject *__pyx_codeobj__313;
  PyObject *__pyx_codeobj__314;
  PyObject *__pyx_codeobj__315;
  PyObject *__pyx_codeobj__316;
  PyObject *__pyx_codeobj__317;
  PyObject *__pyx_codeobj__318;
  PyObject *__pyx_codeobj__319;
  PyObject *__pyx_codeobj__320;
  PyObject *__pyx_codeobj__322;
  PyObject *__pyx_codeobj__323;
  PyObject *__pyx_codeobj__324;
  PyObject *__pyx_codeobj__326;
  PyObject *__pyx_codeobj__328;
  PyObject *__pyx_codeobj__329;
  PyObject *__pyx_codeobj__331;
  PyObject *__pyx_codeobj__332;
  PyObject *__pyx_codeobj__334;
  PyObject *__pyx_codeobj__335;
  PyObject *__pyx_codeobj__337;
  PyObject *__pyx_codeobj__338;
  PyObject *__pyx_codeobj__339;
  PyObject *__pyx_codeobj__340;
  PyObject *__pyx_codeobj__341;
  PyObject *__pyx_codeobj__342;
  PyObject *__pyx_codeobj__343;
  PyObject *__pyx_codeobj__344;
  PyObject *__pyx_codeobj__345;
  PyObject *__pyx_codeobj__346;
  PyObject *__pyx_codeobj__347;
  PyObject *__pyx_codeobj__348;
  PyObject *__pyx_codeobj__349;
  PyObject *__pyx_codeobj__350;
  PyObject *__pyx_codeobj__351;
  PyObject *__pyx_codeobj__352;
  PyObject *__pyx_codeobj__353;
  PyObject *__pyx_codeobj__354;
  PyObject *__pyx_codeobj__355;
  PyObject *__pyx_codeobj__356;
  PyObject *__pyx_codeobj__358;
  PyObject *__pyx_codeobj__359;
  PyObject *__pyx_codeobj__361;
  PyObject *__pyx_codeobj__362;
  PyObject *__pyx_codeobj__364;
  PyObject *__pyx_codeobj__366;
  PyObject *__pyx_codeobj__367;
  PyObject *__pyx_codeobj__368;
  PyObject *__pyx_codeobj__369;
  PyObject *__pyx_codeobj__370;
  PyObject *__pyx_codeobj__371;
  PyObject *__pyx_codeobj__372;
  PyObject *__pyx_codeobj__373;
  PyObject *__pyx_codeobj__374;
  PyObject *__pyx_codeobj__375;
  PyObject *__pyx_codeobj__376;
  PyObject *__pyx_codeobj__377;
  PyObject *__pyx_codeobj__378;
  PyObject *__pyx_codeobj__379;
  PyObject *__pyx_codeobj__380;
  PyObject *__pyx_codeobj__381;
  PyObject *__pyx_codeobj__382;
  PyObject *__pyx_codeobj__383;
  PyObject *__pyx_codeobj__384;
  PyObject *__pyx_codeobj__386;
  PyObject *__pyx_codeobj__388;
  PyObject *__pyx_codeobj__389;
  PyObject *__pyx_codeobj__390;
  PyObject *__pyx_codeobj__391;
  PyObject *__pyx_codeobj__392;
  PyObject *__pyx_codeobj__394;
  PyObject *__pyx_codeobj__395;
  PyObject *__pyx_codeobj__397;
  PyObject *__pyx_codeobj__398;
  PyObject *__pyx_codeobj__399;
  PyObject *__pyx_codeobj__400;
  PyObject *__pyx_codeobj__401;
  PyObject *__pyx_codeobj__403;
  PyObject *__pyx_codeobj__405;
  PyObject *__pyx_codeobj__406;
  PyObject *__pyx_codeobj__407;
  PyObject *__pyx_codeobj__408;
  PyObject *__pyx_codeobj__409;
  PyObject *__pyx_codeobj__410;
  PyObject *__pyx_codeobj__411;
  PyObject *__pyx_codeobj__412;
  PyObject *__pyx_codeobj__414;
  PyObject *__pyx_codeobj__415;
  PyObject *__pyx_codeobj__416;
  PyObject *__pyx_codeobj__417;
  PyObject *__pyx_codeobj__418;
  PyObject *__pyx_codeobj__419;
  PyObject *__pyx_codeobj__420;
  PyObject *__pyx_codeobj__421;
  PyObject *__pyx_codeobj__422;
  PyObject *__pyx_codeobj__423;
  PyObject *__pyx_codeobj__424;
  PyObject *__pyx_codeobj__425;
  PyObject *__pyx_codeobj__426;
  PyObject *__pyx_codeobj__427;
  PyObject *__pyx_codeobj__428;
  PyObject *__pyx_codeobj__429;
  PyObject *__pyx_codeobj__430;
  PyObject *__pyx_codeobj__432;
  PyObject *__pyx_codeobj__434;
  PyObject *__pyx_codeobj__435;
  PyObject *__pyx_codeobj__436;
  PyObject *__pyx_codeobj__437;
  PyObject *__pyx_codeobj__438;
  PyObject *__pyx_codeobj__439;
  PyObject *__pyx_codeobj__440;
  PyObject *__pyx_codeobj__441;
  PyObject *__pyx_codeobj__442;
  PyObject *__pyx_codeobj__443;
  PyObject *__pyx_codeobj__444;
  PyObject *__pyx_codeobj__445;
  PyObject *__pyx_codeobj__446;
  PyObject *__pyx_codeobj__447;
  PyObject *__pyx_codeobj__449;
  PyObject *__pyx_codeobj__450;
  PyObject *__pyx_codeobj__451;
  PyObject *__pyx_codeobj__452;
  PyObject *__pyx_codeobj__453;
  PyObject *__pyx_codeobj__455;
  PyObject *__pyx_codeobj__456;
  PyObject *__pyx_codeobj__457;
  PyObject *__pyx_codeobj__459;
  PyObject *__pyx_codeobj__460;
  PyObject *__pyx_codeobj__461;
  PyObject *__pyx_codeobj__462;
  PyObject *__pyx_codeobj__464;
  PyObject *__pyx_codeobj__465;
  PyObject *__pyx_codeobj__466;
  PyObject *__pyx_codeobj__468;
  PyObject *__pyx_codeobj__470;
  PyObject *__pyx_codeobj__471;
  PyObject *__pyx_codeobj__472;
  PyObject *__pyx_codeobj__473;
  PyObject *__pyx_codeobj__474;
  PyObject *__pyx_codeobj__475;
  PyObject *__pyx_codeobj__476;
  PyObject *__pyx_codeobj__477;
  PyObject *__pyx_codeobj__478;
  PyObject *__pyx_codeobj__479;
  PyObject *__pyx_codeobj__480;
  PyObject *__pyx_codeobj__481;
  PyObject *__pyx_codeobj__482;
  PyObject *__pyx_codeobj__483;
  PyObject *__pyx_codeobj__484;
  PyObject *__pyx_codeobj__485;
  PyObject *__pyx_codeobj__486;
  PyObject *__pyx_codeobj__487;
  PyObject *__pyx_codeobj__488;
  PyObject *__pyx_codeobj__489;
  PyObject *__pyx_codeobj__490;
  PyObject *__pyx_codeobj__491;
  PyObject *__pyx_codeobj__492;
  PyObject *__pyx_codeobj__493;
  PyObject *__pyx_codeobj__494;
  PyObject *__pyx_codeobj__495;
  PyObject *__pyx_codeobj__496;
  PyObject *__pyx_codeobj__497;
  PyObject *__pyx_codeobj__498;
  PyObject *__pyx_codeobj__499;
  PyObject *__pyx_codeobj__500;
  PyObject *__pyx_codeobj__501;
  PyObject *__pyx_codeobj__502;
  PyObject *__pyx_codeobj__503;
  PyObject *__pyx_codeobj__504;
  PyObject *__pyx_codeobj__505;
  PyObject *__pyx_codeobj__506;
  PyObject *__pyx_codeobj__507;
  PyObject *__pyx_codeobj__508;
  PyObject *__pyx_codeobj__509;
  PyObject *__pyx_codeobj__510;
  PyObject *__pyx_codeobj__511;
  PyObject *__pyx_codeobj__512;
  PyObject *__pyx_codeobj__514;
  PyObject *__pyx_codeobj__515;
  PyObject *__pyx_codeobj__516;
  PyObject *__pyx_codeobj__518;
  PyObject *__pyx_codeobj__520;
  PyObject *__pyx_codeobj__521;
  PyObject *__pyx_codeobj__522;
  PyObject *__pyx_codeobj__523;
  PyObject *__pyx_codeobj__524;
  PyObject *__pyx_codeobj__525;
  PyObject *__pyx_codeobj__526;
  PyObject *__pyx_codeobj__527;
  PyObject *__pyx_codeobj__529;
  PyObject *__pyx_codeobj__530;
  PyObject *__pyx_codeobj__531;
  PyObject *__pyx_codeobj__532;
  PyObject *__pyx_codeobj__535;
  PyObject *__pyx_codeobj__536;
  PyObject *__pyx_codeobj__537;
  PyObject *__pyx_codeobj__538;
  PyObject *__pyx_codeobj__540;
  PyObject *__pyx_codeobj__541;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVHandle);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVSocketHandle);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVSocketHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVAsync);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVAsync);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVTimer);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVTimer);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVIdle);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVIdle);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVBaseTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVBaseTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_Loop);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_Loop);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_Handle);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_Handle);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_TimerHandle);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_TimerHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVCheck);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVCheck);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVPoll);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVPoll);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVStream);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVStreamServer);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVStreamServer);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_TCPServer);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_TCPServer);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_TCPTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_TCPTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UnixServer);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UnixServer);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UnixTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UnixTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_ReadUnixTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_ReadUnixTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_WriteUnixTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_WriteUnixTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVProcess);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVProcess);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVProcessTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVProcessTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVFSEvent);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVFSEvent);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UVRequest);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UVRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop__SSLProtocolTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_SSLProtocol);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_SSLProtocol);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_UDPTransport);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_UDPTransport);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_Server);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_Server);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_PseudoSocket);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_PseudoSocket);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_LruCache);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_LruCache);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop__StreamWriteContext);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop__StreamWriteContext);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop__TCPConnectRequest);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop__TCPConnectRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop__PipeConnectRequest);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop__PipeConnectRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_SockAddrHolder);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_SockAddrHolder);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_AddrInfo);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_AddrInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_AddrInfoRequest);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_AddrInfoRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop_NameInfoRequest);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop_NameInfoRequest);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop__UDPSendContext);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop__UDPSendContext);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_2___get__);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter__);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit__);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever);
  Py_CLEAR(clear_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1);
  Py_CLEAR(clear_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1);
  Py_CLEAR(clear_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled);
  Py_CLEAR(clear_module_state->__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled);
  Py_CLEAR(clear_module_state->__pyx_kp_b_1);
  Py_CLEAR(clear_module_state->__pyx_kp_u_18_7_7_7);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AF_INET6_address_must_be_tuple);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AF_INET6_must_be_a_tuple_of_2_4);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AF_INET_address_must_be_tuple);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AF_INET_address_must_be_tuple_of);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AF_UNIX_address_must_be_a_str_or);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AF_UNSPEC_in_DNS_results);
  Py_CLEAR(clear_module_state->__pyx_n_u_AI_CANONNAME);
  Py_CLEAR(clear_module_state->__pyx_kp_u_A_Stream_Socket_was_expected_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_A_UDP_Socket_was_expected_got_r);
  Py_CLEAR(clear_module_state->__pyx_kp_u_A_UNIX_Domain_Stream_Socket_was);
  Py_CLEAR(clear_module_state->__pyx_kp_u_A_callable_object_or_None_is_exp);
  Py_CLEAR(clear_module_state->__pyx_n_s_AbstractChildWatcher);
  Py_CLEAR(clear_module_state->__pyx_n_s_AddrInfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_AddrInfoRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_AddrInfoRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_AddrInfoRequest___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AddrInfoRequest_callback_called);
  Py_CLEAR(clear_module_state->__pyx_n_s_AddrInfo___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_AddrInfo___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_AddrInfo_data_is_NULL);
  Py_CLEAR(clear_module_state->__pyx_n_s_AddressFamily);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Address_r_is_already_in_use);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_AttributeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Bad_exception_data_from_child_r);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseException);
  Py_CLEAR(clear_module_state->__pyx_n_s_BaseProtocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_BlockingIOError);
  Py_CLEAR(clear_module_state->__pyx_n_s_BrokenPipeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CHANGE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Callback_handles_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_CancelledError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_call_write_after_write_eo);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_call_writelines_after_wri);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_close_a_running_event_loo);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_run_the_event_loop_while);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Cannot_serialize_socket_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_CertificateError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Connect_call_failed);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConnectionAbortedError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConnectionRefusedError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConnectionResetError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Counter);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Creating__SSLProtocolTransport_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEVNULL);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeprecationWarning);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_ADDRFAMILY);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_AGAIN);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_BADFLAGS);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_BADHINTS);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_CANCELED);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_FAIL);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_FAMILY);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_MEMORY);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_NODATA);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_NONAME);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_OVERFLOW);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_PROTOCOL);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_SERVICE);
  Py_CLEAR(clear_module_state->__pyx_n_u_EAI_SOCKTYPE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_calling_eof_received);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_occurred_during_shutdown);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Errors_logged);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Event_loop_is_closed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Event_loop_stopped_before_Future);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Exception_in___repr___r_value_ty);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Exception_in_callback);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Exception_in_default_exception_h);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Exception_in_default_exception_h_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Exception_occurred_in_preexec_fn);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Executing_r_took_3f_seconds);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Executing_s_took_3f_seconds);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Executor_shutdown_has_been_calle);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Fatal_error_on_SSL_protocol);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Fatal_error_on_server);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Fatal_error_on_transport);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Fatal_error_on_transport_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Fatal_write_error_on_datagram_tr);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileExistsError);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileNotFoundError);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileSystemEvent);
  Py_CLEAR(clear_module_state->__pyx_kp_u_File_descriptor_r_is_used_by_tra);
  Py_CLEAR(clear_module_state->__pyx_n_s_FlowControlMixin);
  Py_CLEAR(clear_module_state->__pyx_n_s_Future);
  Py_CLEAR(clear_module_state->__pyx_n_s_Handle);
  Py_CLEAR(clear_module_state->__pyx_n_s_Handle___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Handle___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Handle_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_Handle_cancelled);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Handle_loop_is_None_in_Handle);
  Py_CLEAR(clear_module_state->__pyx_n_u_IPV6_V6ONLY);
  Py_CLEAR(clear_module_state->__pyx_kp_u_IPv4_sockaddr_must_be_2_tuple);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntEnum);
  Py_CLEAR(clear_module_state->__pyx_n_s_InterruptedError);
  Py_CLEAR(clear_module_state->__pyx_n_s_InvalidStateError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_address_must_be_None_or);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_file_descriptor);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Invalid_file_object_r);
  Py_CLEAR(clear_module_state->__pyx_n_s_IsADirectoryError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Iterable);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyboardInterrupt);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Listen_errors);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop___get___locals_lambda);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop___sighandler);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop___subprocess_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__asyncgen_finalizer_hook);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__asyncgen_firstiter_hook);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__check_default_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__do_shutdown);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__get_backend_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__getaddrinfo_locals_callbac);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__getnameinfo_locals_callbac);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__monitor_fs);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop__read_from_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__sighandler);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop__sock_accept);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop__sock_connect);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop__sock_recv);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop__sock_recv_into);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop__sock_sendall);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop__stop);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop__subprocess_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_add_reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_add_signal_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_add_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_call_at);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_call_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_call_later);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_call_soon);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_call_soon_threadsafe);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_connect_accepted_socket);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_connect_read_pipe);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_connect_write_pipe);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_create_connection);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_create_datagram_endpoint);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_create_future);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_create_server);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_create_task);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_create_unix_connection);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_create_unix_server);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop_debug_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_default_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_get_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_get_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_get_task_factory);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_getaddrinfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_getnameinfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_is_closed);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_is_running);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_remove_reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_remove_signal_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_remove_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_run_forever);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_run_in_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_run_until_complete);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_set_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_set_default_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_set_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_set_task_factory);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_shutdown_asyncgens);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_shutdown_default_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_accept);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_connect);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_recv);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_recv_into);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_recvfrom);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_recvfrom_into);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_sendall);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_sock_sendto);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_start_tls);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_stop);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_subprocess_exec);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_subprocess_shell);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Loop_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_Loop_time_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_LruCache);
  Py_CLEAR(clear_module_state->__pyx_n_s_LruCache___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_LruCache___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryBIO);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Multiple_exceptions);
  Py_CLEAR(clear_module_state->__pyx_n_s_NSIG);
  Py_CLEAR(clear_module_state->__pyx_n_s_NameInfoRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_NameInfoRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_NameInfoRequest___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Neither_host_port_nor_sock_were);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Non_thread_safe_operation_invoke);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplementedError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Number_of_signals);
  Py_CLEAR(clear_module_state->__pyx_n_s_OSError);
  Py_CLEAR(clear_module_state->__pyx_n_s_O_RDWR);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Object_created_at_most_recent_ca);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderedDict);
  Py_CLEAR(clear_module_state->__pyx_n_s_OverflowError);
  Py_CLEAR(clear_module_state->__pyx_n_s_PIPE);
  Py_CLEAR(clear_module_state->__pyx_n_u_PYTHONASYNCIODEBUG);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Passing_reuse_address_True_is_no);
  Py_CLEAR(clear_module_state->__pyx_n_s_PermissionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Pickling_of_struct_members_such);
  Py_CLEAR(clear_module_state->__pyx_n_s_PipeConnectRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_PipeConnectRequest___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_PipeConnectRequest___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Polls_debug_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_ProcessLookupError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Process_info);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Process_memory);
  Py_CLEAR(clear_module_state->__pyx_n_s_Protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket___enter);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket___exit);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket___getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_accept);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_bind);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_connect);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_connect_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_detach);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_dup);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_fileno);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_get_inheritable);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_getpeername);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_getsockname);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_getsockopt);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_gettimeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_ioctl);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_listen);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_makefile);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_recv);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_recv_into);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_recvfrom);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_recvfrom_into);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_recvmsg);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_recvmsg_into);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_sendall);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_sendfile);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_sendmsg);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_sendmsg_afalg);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_sendto);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_set_inheritable);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_setblocking);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_setsockopt);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_settimeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_share);
  Py_CLEAR(clear_module_state->__pyx_n_s_PseudoSocket_shutdown);
  Py_CLEAR(clear_module_state->__pyx_n_s_Pyx_CFunc_6uvloop_4loop_4Loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_RENAME);
  Py_CLEAR(clear_module_state->__pyx_n_s_RENAME_CHANGE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Racing_with_another_loop_to_spaw);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadSubprocessPipeProto);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadSubprocessPipeProto_data_rec);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_abort);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_can_write_eof);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_get_write_buff);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_get_write_buff_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_set_write_buff);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_write_eof);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadUnixTransport_writelines);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Read_EOF_failed_callbacks);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Read_EOFs);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Read_callbacks);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Read_callbacks_failed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Read_errors);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Read_events);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Read_failed_callbacks);
  Py_CLEAR(clear_module_state->__pyx_n_s_ResourceWarning);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeWarning);
  Py_CLEAR(clear_module_state->__pyx_n_s_SIG_DFL);
  Py_CLEAR(clear_module_state->__pyx_n_u_SOCK_NONBLOCK);
  Py_CLEAR(clear_module_state->__pyx_n_u_SO_BROADCAST);
  Py_CLEAR(clear_module_state->__pyx_n_u_SO_REUSEPORT);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLContext);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport___reduce_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport__force_clo);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport__test__app);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_abort);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_can_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_get_extra);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_get_protoc);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_get_read_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_get_read_b_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_get_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_get_write_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_is_closing);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_is_reading);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_pause_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_resume_rea);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_set_protoc);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_set_read_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_set_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_write_eof);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocolTransport_writelines);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SSLProtocol__do_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol__get_app_transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol__start_handshake_loc);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol__start_shutdown_loca);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol_buffer_updated);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol_connection_lost);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol_connection_made);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol_eof_received);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol_get_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol_pause_writing);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLProtocol_resume_writing);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLSyscallError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLWantReadError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSLZeroReturnError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSL_ERROR_SYSCALL);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSL_ERROR_WANT_READ);
  Py_CLEAR(clear_module_state->__pyx_n_s_SSL_ERROR_WANT_WRITE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SSL_connection_is_closed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SSL_handshake_failed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SSL_handshake_failed_on_verifyin);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SSL_handshake_is_taking_longer_t);
  Py_CLEAR(clear_module_state->__pyx_kp_u_SSL_shutdown_timed_out);
  Py_CLEAR(clear_module_state->__pyx_n_s_STDOUT);
  Py_CLEAR(clear_module_state->__pyx_n_s_S_ISSOCK);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server___aenter);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server___aexit);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server_get_loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server_is_serving);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server_serve_forever);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Server_side_SSL_needs_a_valid_SS);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server_start_serving);
  Py_CLEAR(clear_module_state->__pyx_n_s_Server_wait_closed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Shutdown_errors);
  Py_CLEAR(clear_module_state->__pyx_n_s_SockAddrHolder);
  Py_CLEAR(clear_module_state->__pyx_n_s_SockAddrHolder___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SockAddrHolder___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Sock_ops_successful_on_1st_try);
  Py_CLEAR(clear_module_state->__pyx_n_s_SocketKind);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Socket_try_writes);
  Py_CLEAR(clear_module_state->__pyx_n_s_StackSummary);
  Py_CLEAR(clear_module_state->__pyx_n_s_StreamWriteContext);
  Py_CLEAR(clear_module_state->__pyx_n_s_StreamWriteContext___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_StreamWriteContext___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_kp_u_StreamWriteContext_close_py_buf);
  Py_CLEAR(clear_module_state->__pyx_kp_u_StreamWriteContext_close_uv_buf);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Streams_debug_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubprocessError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketReaderFuture);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketReaderFuture___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketReaderFuture___remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketReaderFuture__loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketReaderFuture__remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketReaderFuture__sock);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketReaderFuture_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketWriterFuture);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketWriterFuture___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketWriterFuture___remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketWriterFuture__loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketWriterFuture__remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketWriterFuture__sock);
  Py_CLEAR(clear_module_state->__pyx_n_s_SyncSocketWriterFuture_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_SystemExit);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPConnectRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPConnectRequest___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPConnectRequest___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPServer);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPServer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPServer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPTransport___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPTransport___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TCPTransport_get_extra_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_Task);
  Py_CLEAR(clear_module_state->__pyx_kp_u_The_executor_did_not_finishing_j);
  Py_CLEAR(clear_module_state->__pyx_kp_u_The_reuse_address_parameter_has);
  Py_CLEAR(clear_module_state->__pyx_n_s_Thread);
  Py_CLEAR(clear_module_state->__pyx_n_s_ThreadPoolExecutor);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimeoutError);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimerHandle);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimerHandle___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimerHandle___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimerHandle_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimerHandle_cancelled);
  Py_CLEAR(clear_module_state->__pyx_n_s_TimerHandle_when);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Timer_handles_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_Transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_UDPSendContext);
  Py_CLEAR(clear_module_state->__pyx_n_s_UDPSendContext___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UDPSendContext___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_UDPTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_UDPTransport___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UDPTransport___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UDPTransport_family_is_undefined);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UDPTransport_family_is_undefined_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_UDPTransport_sendto);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVAsync);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVAsync___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVAsync___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport___setstate_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport__force_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_abort);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_get_extra_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_get_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_get_write_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_get_write_buffer_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_is_closing);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_set_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVBaseTransport_set_write_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVCheck);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVCheck___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVCheck___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVFSEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVFSEvent___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVFSEvent___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVFSEvent_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVFSEvent_cancelled);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVHandle);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVHandle___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVHandle___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVHandles_python_libuv_total);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVIdle);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVIdle___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVIdle___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVPoll);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVPoll___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVPoll___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcess);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport___stdio_inite);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVProcessTransport__call_connect);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport__stdio_inite);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport__wait);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_get_extra_inf);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_get_pid);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_get_pipe_tran);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_get_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_get_returncod);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_is_closing);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_kill);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_send_signal);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_set_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcessTransport_terminate);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcess___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVProcess___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVProcess__close_after_spawn_cal);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVRequest);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVRequest___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVRequest___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVSocketHandle);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVSocketHandle___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVSocketHandle___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStreamServer);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStreamServer___on_ssl_connecte);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStreamServer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStreamServer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStreamServer__on_ssl_connecte);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVStream__try_write_called_with);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream_can_write_eof);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream_is_reading);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream_pause_reading);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream_resume_reading);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVStream_shutdown_callback_calle);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream_write);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVStream_write_callback_called_w);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream_write_eof);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVStream_writelines);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVTimer);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVTimer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UVTimer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVTransport__call_connection_los);
  Py_CLEAR(clear_module_state->__pyx_kp_u_UVTransport__call_connection_mad);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unable_to_check_or_remove_stale);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unable_to_clean_up_listening_UNI);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unhandled_error_in_exception_han);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Unhandled_exception_in_event_loo);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnicodeEncodeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnixServer);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnixServer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnixServer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnixTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnixTransport___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnixTransport___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s_WeakSet);
  Py_CLEAR(clear_module_state->__pyx_n_s_WeakValueDictionary);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteSubprocessPipeProto);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteSubprocessPipeProto___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteSubprocessPipeProto___repr);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteSubprocessPipeProto_connect);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteSubprocessPipeProto_connect_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteSubprocessPipeProto_pause_w);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteSubprocessPipeProto_resume);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteUnixTransport);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteUnixTransport___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteUnixTransport___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteUnixTransport_pause_reading);
  Py_CLEAR(clear_module_state->__pyx_n_s_WriteUnixTransport_resume_readin);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Write_callbacks_failed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Write_contexts_8);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Write_errors);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Write_events);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Write_failed_callbacks);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Write_without_poll);
  Py_CLEAR(clear_module_state->__pyx_kp_u_XXX_an_allocated_buffer_in_trans);
  Py_CLEAR(clear_module_state->__pyx_kp_u_You_must_set_server_hostname_whe);
  Py_CLEAR(clear_module_state->__pyx_n_s__10);
  Py_CLEAR(clear_module_state->__pyx_kp_b__10);
  Py_CLEAR(clear_module_state->__pyx_kp_u__10);
  Py_CLEAR(clear_module_state->__pyx_kp_u__115);
  Py_CLEAR(clear_module_state->__pyx_kp_u__128);
  Py_CLEAR(clear_module_state->__pyx_kp_u__129);
  Py_CLEAR(clear_module_state->__pyx_kp_u__130);
  Py_CLEAR(clear_module_state->__pyx_kp_u__140);
  Py_CLEAR(clear_module_state->__pyx_kp_u__141);
  Py_CLEAR(clear_module_state->__pyx_kp_b__161);
  Py_CLEAR(clear_module_state->__pyx_kp_b__170);
  Py_CLEAR(clear_module_state->__pyx_kp_u__196);
  Py_CLEAR(clear_module_state->__pyx_n_s__198);
  Py_CLEAR(clear_module_state->__pyx_n_s__263);
  Py_CLEAR(clear_module_state->__pyx_kp_u__33);
  Py_CLEAR(clear_module_state->__pyx_n_s__542);
  Py_CLEAR(clear_module_state->__pyx_kp_u__61);
  Py_CLEAR(clear_module_state->__pyx_kp_u__69);
  Py_CLEAR(clear_module_state->__pyx_kp_u__75);
  Py_CLEAR(clear_module_state->__pyx_kp_u__78);
  Py_CLEAR(clear_module_state->__pyx_kp_u__79);
  Py_CLEAR(clear_module_state->__pyx_n_s_abc);
  Py_CLEAR(clear_module_state->__pyx_n_s_abort);
  Py_CLEAR(clear_module_state->__pyx_kp_u_abort_init__closed_is_set);
  Py_CLEAR(clear_module_state->__pyx_kp_u_abort_init__inited_is_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_accept);
  Py_CLEAR(clear_module_state->__pyx_kp_u_accept_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_aclose);
  Py_CLEAR(clear_module_state->__pyx_kp_u_active_TimerHandle_is_deallacati);
  Py_CLEAR(clear_module_state->__pyx_n_s_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_done_callback);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_signal_handler);
  Py_CLEAR(clear_module_state->__pyx_kp_u_add_signal_handler_can_only_be_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_addr);
  Py_CLEAR(clear_module_state->__pyx_n_s_address);
  Py_CLEAR(clear_module_state->__pyx_n_s_addrinfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_addrs);
  Py_CLEAR(clear_module_state->__pyx_n_s_aenter);
  Py_CLEAR(clear_module_state->__pyx_n_s_aexit);
  Py_CLEAR(clear_module_state->__pyx_n_s_ag);
  Py_CLEAR(clear_module_state->__pyx_n_s_agen);
  Py_CLEAR(clear_module_state->__pyx_n_s_ai);
  Py_CLEAR(clear_module_state->__pyx_n_s_ai_cnt);
  Py_CLEAR(clear_module_state->__pyx_n_s_ai_local);
  Py_CLEAR(clear_module_state->__pyx_n_s_ai_remote);
  Py_CLEAR(clear_module_state->__pyx_kp_u_alive_closed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_all_args_must_be_str_or_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_all_environment_values_must_be_b);
  Py_CLEAR(clear_module_state->__pyx_kp_u_all_environment_vars_must_be_byt);
  Py_CLEAR(clear_module_state->__pyx_n_s_allow_broadcast);
  Py_CLEAR(clear_module_state->__pyx_kp_u_an_error_occurred_during_closing);
  Py_CLEAR(clear_module_state->__pyx_n_s_app_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_app_transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_u_ascii);
  Py_CLEAR(clear_module_state->__pyx_n_u_asyncgen);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncgen_finalizer_hook);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncgen_firstiter_hook);
  Py_CLEAR(clear_module_state->__pyx_kp_u_asynchronous_generator_r_was_sch);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio);
  Py_CLEAR(clear_module_state->__pyx_kp_s_asyncio_Handle);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_base_events);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_futures);
  Py_CLEAR(clear_module_state->__pyx_kp_u_asyncio_is_trying_to_install_it);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_log);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_sslproto);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_tasks);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_transports);
  Py_CLEAR(clear_module_state->__pyx_n_s_await);
  Py_CLEAR(clear_module_state->__pyx_n_u_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_backlog);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_events);
  Py_CLEAR(clear_module_state->__pyx_kp_b_bin_sh);
  Py_CLEAR(clear_module_state->__pyx_n_s_bind);
  Py_CLEAR(clear_module_state->__pyx_kp_u_bind_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_buf);
  Py_CLEAR(clear_module_state->__pyx_kp_u_buffer_size_is_not_0_after_a_su);
  Py_CLEAR(clear_module_state->__pyx_n_s_buffer_updated);
  Py_CLEAR(clear_module_state->__pyx_n_s_bufs);
  Py_CLEAR(clear_module_state->__pyx_n_s_bufsize);
  Py_CLEAR(clear_module_state->__pyx_kp_u_bufsize_must_be_0);
  Py_CLEAR(clear_module_state->__pyx_n_s_builtins);
  Py_CLEAR(clear_module_state->__pyx_kp_b_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_str_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_at);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_connection_made);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_later);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_soon);
  Py_CLEAR(clear_module_state->__pyx_n_s_call_soon_threadsafe);
  Py_CLEAR(clear_module_state->__pyx_n_s_callback);
  Py_CLEAR(clear_module_state->__pyx_kp_u_called_after_destroying_the_UVH);
  Py_CLEAR(clear_module_state->__pyx_kp_u_called_with_handle_data_NULL);
  Py_CLEAR(clear_module_state->__pyx_n_s_can_write_eof);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancel);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cancel_request_data_is_NULL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cancel_request_data_is_not_UVRe);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cancel_request_is_NULL);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancelled);
  Py_CLEAR(clear_module_state->__pyx_n_u_cancelled);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_add_a_signal_handler_for);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_apply_subprocess_STDOUT);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_cleanup_signal_handlers_c);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_convert_sockaddr_into_Pyt);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_open_a_UDP_handle_invalid);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_pause_signals_handling_no);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_run_Handle_callback_is_no);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_run_TimerHandle_callback);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_shutdown_signals_handling);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_spawn_a_process_args_are);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cannot_switch_state_from_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_cast);
  Py_CLEAR(clear_module_state->__pyx_n_s_cause);
  Py_CLEAR(clear_module_state->__pyx_n_s_cfunc_to_py);
  Py_CLEAR(clear_module_state->__pyx_n_s_chain);
  Py_CLEAR(clear_module_state->__pyx_n_s_check_default_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_check_hostname);
  Py_CLEAR(clear_module_state->__pyx_n_s_cipher);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cleanup_socket);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_kp_u_close__handle_data_is_NULL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_close__handle_data_is_not_UVHa);
  Py_CLEAR(clear_module_state->__pyx_n_s_close_fds);
  Py_CLEAR(clear_module_state->__pyx_kp_u_close_method);
  Py_CLEAR(clear_module_state->__pyx_kp_u_close_uv_is_closing_is_true);
  Py_CLEAR(clear_module_state->__pyx_kp_u_closed_Transport_handle_and_unse);
  Py_CLEAR(clear_module_state->__pyx_kp_u_closed_reading_x);
  Py_CLEAR(clear_module_state->__pyx_kp_u_closed_x);
  Py_CLEAR(clear_module_state->__pyx_n_s_closing_agens);
  Py_CLEAR(clear_module_state->__pyx_n_s_cmd);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections_abc);
  Py_CLEAR(clear_module_state->__pyx_n_s_completed);
  Py_CLEAR(clear_module_state->__pyx_n_s_compression);
  Py_CLEAR(clear_module_state->__pyx_n_s_concurrent);
  Py_CLEAR(clear_module_state->__pyx_kp_u_concurrent_allocations);
  Py_CLEAR(clear_module_state->__pyx_n_s_concurrent_futures);
  Py_CLEAR(clear_module_state->__pyx_n_s_conmade_cb);
  Py_CLEAR(clear_module_state->__pyx_n_s_connect);
  Py_CLEAR(clear_module_state->__pyx_n_s_connect_accepted_socket);
  Py_CLEAR(clear_module_state->__pyx_n_s_connect_ex);
  Py_CLEAR(clear_module_state->__pyx_kp_u_connect_ex_method);
  Py_CLEAR(clear_module_state->__pyx_kp_u_connect_failed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_connect_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_connect_read_pipe);
  Py_CLEAR(clear_module_state->__pyx_n_s_connect_write_pipe);
  Py_CLEAR(clear_module_state->__pyx_n_s_connection_lost);
  Py_CLEAR(clear_module_state->__pyx_n_s_connection_made);
  Py_CLEAR(clear_module_state->__pyx_n_s_context);
  Py_CLEAR(clear_module_state->__pyx_n_u_context);
  Py_CLEAR(clear_module_state->__pyx_kp_u_context_manager_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_coro);
  Py_CLEAR(clear_module_state->__pyx_n_s_coroutines);
  Py_CLEAR(clear_module_state->__pyx_kp_u_coroutines_cannot_be_used_with_a);
  Py_CLEAR(clear_module_state->__pyx_kp_u_coroutines_cannot_be_used_with_r);
  Py_CLEAR(clear_module_state->__pyx_kp_u_could_not_bind_to_local_addr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_could_not_bind_to_remote_addr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_could_not_close_attached_file_ob);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_connection);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_connection_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_datagram_endpoint);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_datagram_endpoint_locals);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_default_context);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_future);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_server);
  Py_CLEAR(clear_module_state->__pyx_kp_u_create_server_failed_to_create_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_task);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_unix_connection);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_unix_server);
  Py_CLEAR(clear_module_state->__pyx_kp_u_created_at);
  Py_CLEAR(clear_module_state->__pyx_n_s_creationflags);
  Py_CLEAR(clear_module_state->__pyx_kp_u_creationflags_is_not_supported);
  Py_CLEAR(clear_module_state->__pyx_n_s_ctx);
  Py_CLEAR(clear_module_state->__pyx_n_s_cwd);
  Py_CLEAR(clear_module_state->__pyx_kp_u_cwd_must_be_a_str_or_bytes_objec);
  Py_CLEAR(clear_module_state->__pyx_n_u_cygwin);
  Py_CLEAR(clear_module_state->__pyx_n_u_darwin);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_kp_u_data_expecting_a_bytes_like_inst);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_received);
  Py_CLEAR(clear_module_state->__pyx_n_s_datagram_received);
  Py_CLEAR(clear_module_state->__pyx_kp_u_dealloc____handle_is_NULL__cl);
  Py_CLEAR(clear_module_state->__pyx_kp_u_deallocating_a_running_event_loo);
  Py_CLEAR(clear_module_state->__pyx_kp_u_deallocating_an_open_event_loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_debug_flags);
  Py_CLEAR(clear_module_state->__pyx_n_u_debug_wrapper);
  Py_CLEAR(clear_module_state->__pyx_n_s_decode);
  Py_CLEAR(clear_module_state->__pyx_n_s_decref_socketios);
  Py_CLEAR(clear_module_state->__pyx_n_s_default);
  Py_CLEAR(clear_module_state->__pyx_n_s_default_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_default_int_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_delay);
  Py_CLEAR(clear_module_state->__pyx_n_s_deque);
  Py_CLEAR(clear_module_state->__pyx_n_s_detach);
  Py_CLEAR(clear_module_state->__pyx_kp_u_detach_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_dev_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_devnull);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_disable);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_discard);
  Py_CLEAR(clear_module_state->__pyx_n_s_disconnected);
  Py_CLEAR(clear_module_state->__pyx_n_s_do_handshake);
  Py_CLEAR(clear_module_state->__pyx_n_s_do_shutdown);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_have_a_reader_in_the);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_have_a_reader_in_the_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_have_a_writer_in_the);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_have_a_writer_in_the_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_done);
  Py_CLEAR(clear_module_state->__pyx_n_s_done_cb);
  Py_CLEAR(clear_module_state->__pyx_n_s_dup);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_enabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_ensure_future);
  Py_CLEAR(clear_module_state->__pyx_n_s_enter);
  Py_CLEAR(clear_module_state->__pyx_n_s_enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_env);
  Py_CLEAR(clear_module_state->__pyx_n_s_environ);
  Py_CLEAR(clear_module_state->__pyx_n_s_eof_received);
  Py_CLEAR(clear_module_state->__pyx_n_s_err);
  Py_CLEAR(clear_module_state->__pyx_n_s_errno);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_error_received);
  Py_CLEAR(clear_module_state->__pyx_kp_u_error_status_in_uv_stream_t_list);
  Py_CLEAR(clear_module_state->__pyx_kp_u_error_status_in_uv_stream_t_read);
  Py_CLEAR(clear_module_state->__pyx_kp_u_error_status_in_uv_stream_t_shut);
  Py_CLEAR(clear_module_state->__pyx_kp_u_error_status_in_uv_stream_t_writ);
  Py_CLEAR(clear_module_state->__pyx_kp_u_error_while_attempting_to_bind_o);
  Py_CLEAR(clear_module_state->__pyx_n_s_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_exception);
  Py_CLEAR(clear_module_state->__pyx_n_u_exception);
  Py_CLEAR(clear_module_state->__pyx_kp_u_exception_in_Task___repr);
  Py_CLEAR(clear_module_state->__pyx_n_s_exceptions);
  Py_CLEAR(clear_module_state->__pyx_n_s_excs);
  Py_CLEAR(clear_module_state->__pyx_n_s_executable);
  Py_CLEAR(clear_module_state->__pyx_n_s_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_exit);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expected_AF_INET_AF_INET6_or_AF);
  Py_CLEAR(clear_module_state->__pyx_kp_u_expected_asyncio_Future);
  Py_CLEAR(clear_module_state->__pyx_n_s_extract);
  Py_CLEAR(clear_module_state->__pyx_n_s_f1);
  Py_CLEAR(clear_module_state->__pyx_n_s_f2);
  Py_CLEAR(clear_module_state->__pyx_n_s_factory);
  Py_CLEAR(clear_module_state->__pyx_n_s_family);
  Py_CLEAR(clear_module_state->__pyx_kp_u_fatal_Could_not_advance__StreamW);
  Py_CLEAR(clear_module_state->__pyx_kp_u_fatal_sent_0_in_advance_uv_buf);
  Py_CLEAR(clear_module_state->__pyx_n_s_fd);
  Py_CLEAR(clear_module_state->__pyx_kp_u_fd_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_fileno);
  Py_CLEAR(clear_module_state->__pyx_n_s_fileobj);
  Py_CLEAR(clear_module_state->__pyx_n_s_finalizer);
  Py_CLEAR(clear_module_state->__pyx_n_s_firstiter);
  Py_CLEAR(clear_module_state->__pyx_n_s_flag);
  Py_CLEAR(clear_module_state->__pyx_n_s_flags);
  Py_CLEAR(clear_module_state->__pyx_n_s_flowinfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_flush);
  Py_CLEAR(clear_module_state->__pyx_n_s_force_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_format_list);
  Py_CLEAR(clear_module_state->__pyx_n_s_frame);
  Py_CLEAR(clear_module_state->__pyx_n_s_fs);
  Py_CLEAR(clear_module_state->__pyx_n_s_fs_handle);
  Py_CLEAR(clear_module_state->__pyx_n_s_fspath);
  Py_CLEAR(clear_module_state->__pyx_n_s_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_functools);
  Py_CLEAR(clear_module_state->__pyx_n_s_fut);
  Py_CLEAR(clear_module_state->__pyx_n_s_future);
  Py_CLEAR(clear_module_state->__pyx_n_s_futures);
  Py_CLEAR(clear_module_state->__pyx_n_s_gaierror);
  Py_CLEAR(clear_module_state->__pyx_n_s_gather);
  Py_CLEAR(clear_module_state->__pyx_n_s_gc);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_app_transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_asyncgen_hooks);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_backend_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_u_get_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_coroutine_origin_tracking_de);
  Py_CLEAR(clear_module_state->__pyx_n_u_get_coroutine_wrapper);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_event_loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_extra_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_inheritable);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_pid);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_pipe_transport);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_read_buffer_limits);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_read_buffer_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_returncode);
  Py_CLEAR(clear_module_state->__pyx_n_u_get_running_loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_task_factory);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_write_buffer_limits);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_write_buffer_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_getaddrinfo);
  Py_CLEAR(clear_module_state->__pyx_kp_u_getaddrinfo_returned_empty_list);
  Py_CLEAR(clear_module_state->__pyx_kp_u_getaddrinfo_returned_empty_list_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_getfilesystemencoding);
  Py_CLEAR(clear_module_state->__pyx_n_s_getframe);
  Py_CLEAR(clear_module_state->__pyx_n_s_getnameinfo);
  Py_CLEAR(clear_module_state->__pyx_kp_u_getnameinfo_argument_1_must_be_a);
  Py_CLEAR(clear_module_state->__pyx_kp_u_getnameinfo_flowinfo_must_be_0_1);
  Py_CLEAR(clear_module_state->__pyx_n_s_getpeercert);
  Py_CLEAR(clear_module_state->__pyx_n_s_getpeername);
  Py_CLEAR(clear_module_state->__pyx_n_s_getservbyname);
  Py_CLEAR(clear_module_state->__pyx_kp_u_getsockaddrarg_scope_id_must_be);
  Py_CLEAR(clear_module_state->__pyx_n_s_getsockname);
  Py_CLEAR(clear_module_state->__pyx_n_s_getsockopt);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_gettimeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_h);
  Py_CLEAR(clear_module_state->__pyx_n_s_handle);
  Py_CLEAR(clear_module_state->__pyx_n_u_handle);
  Py_CLEAR(clear_module_state->__pyx_kp_u_handle_data_is_NULL_in___close_a);
  Py_CLEAR(clear_module_state->__pyx_kp_u_handle_loop_data_is_NULL_in___en);
  Py_CLEAR(clear_module_state->__pyx_kp_u_handle_loop_is_NULL_in___ensure);
  Py_CLEAR(clear_module_state->__pyx_n_s_handler);
  Py_CLEAR(clear_module_state->__pyx_kp_u_has_handle_0_in___uv_close_handl);
  Py_CLEAR(clear_module_state->__pyx_kp_u_high);
  Py_CLEAR(clear_module_state->__pyx_n_s_high_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_host);
  Py_CLEAR(clear_module_state->__pyx_kp_u_host_and_port_was_not_specified);
  Py_CLEAR(clear_module_state->__pyx_kp_u_host_must_be_a_str_or_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_host_must_be_a_string_or_bytes_o);
  Py_CLEAR(clear_module_state->__pyx_kp_u_host_port_and_sock_can_not_be_sp);
  Py_CLEAR(clear_module_state->__pyx_n_s_hosts);
  Py_CLEAR(clear_module_state->__pyx_n_s_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_ident);
  Py_CLEAR(clear_module_state->__pyx_n_u_idna);
  Py_CLEAR(clear_module_state->__pyx_n_s_ignore_environment);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_u_inf);
  Py_CLEAR(clear_module_state->__pyx_n_s_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_infos);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_inspect);
  Py_CLEAR(clear_module_state->__pyx_kp_u_integer_argument_expected_got);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_Handle_cb_type);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid__init_protocol_call);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_data_in_writebuf_an_inst);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_socket_family_expected_A);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_waiter_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_io_refs);
  Py_CLEAR(clear_module_state->__pyx_n_s_ioctl);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_alive);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_alive_but__handle_data_is_no);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_alive_but__handle_is_NULL);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_alive_but__handle_loop_is_no);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_alive_but__loop_is_None);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_already_being_awaited_on_ser);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_closed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_closed_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_closing);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_longer_than_107_characters);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_supported_by_start_tls);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_not_supposed_to_be_instantia);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_open_in___dealloc___with_loo);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_reading);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_running);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_serving);
  Py_CLEAR(clear_module_state->__pyx_n_s_iscoroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_iscoroutinefunction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_u_isfuture);
  Py_CLEAR(clear_module_state->__pyx_n_s_isgenerator);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_itertools);
  Py_CLEAR(clear_module_state->__pyx_n_s_join);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_kill);
  Py_CLEAR(clear_module_state->__pyx_n_s_kwargs);
  Py_CLEAR(clear_module_state->__pyx_kp_u_laddr_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_lads);
  Py_CLEAR(clear_module_state->__pyx_n_s_lai);
  Py_CLEAR(clear_module_state->__pyx_n_s_lai_addr_static);
  Py_CLEAR(clear_module_state->__pyx_n_s_lai_iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_lai_static);
  Py_CLEAR(clear_module_state->__pyx_n_s_last);
  Py_CLEAR(clear_module_state->__pyx_n_s_libuv_get_loop_t_ptr);
  Py_CLEAR(clear_module_state->__pyx_n_s_libuv_get_version);
  Py_CLEAR(clear_module_state->__pyx_n_s_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_list_of_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_listen);
  Py_CLEAR(clear_module_state->__pyx_kp_u_listen_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_local_addr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_local_addr_must_be_a_tuple_of_ho);
  Py_CLEAR(clear_module_state->__pyx_kp_u_local_addr_must_be_a_tuple_of_ho_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_log);
  Py_CLEAR(clear_module_state->__pyx_n_s_log_destroy_pending);
  Py_CLEAR(clear_module_state->__pyx_n_s_log_lines);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_lookup_lines);
  Py_CLEAR(clear_module_state->__pyx_n_s_loop);
  Py_CLEAR(clear_module_state->__pyx_kp_u_loop__exec_queued_writes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_loop__on_idle);
  Py_CLEAR(clear_module_state->__pyx_n_s_low);
  Py_CLEAR(clear_module_state->__pyx_n_s_lower);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_main_thread);
  Py_CLEAR(clear_module_state->__pyx_n_s_makefile);
  Py_CLEAR(clear_module_state->__pyx_kp_u_makefile_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_maxsize);
  Py_CLEAR(clear_module_state->__pyx_kp_u_maxsize_is_expected_to_be_greate);
  Py_CLEAR(clear_module_state->__pyx_n_s_message);
  Py_CLEAR(clear_module_state->__pyx_n_u_message);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_model);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_monitor_fs);
  Py_CLEAR(clear_module_state->__pyx_n_s_monotonic);
  Py_CLEAR(clear_module_state->__pyx_n_s_move_to_end);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_msg);
  Py_CLEAR(clear_module_state->__pyx_kp_u_must_be_0);
  Py_CLEAR(clear_module_state->__pyx_kp_u_must_be_low);
  Py_CLEAR(clear_module_state->__pyx_n_s_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_u_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_nbytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_kp_u_new_callbacks_were_queued_during);
  Py_CLEAR(clear_module_state->__pyx_kp_u_new_poll_handles_were_queued_dur);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_task);
  Py_CLEAR(clear_module_state->__pyx_kp_u_new_timers_were_queued_during_lo);
  Py_CLEAR(clear_module_state->__pyx_n_s_nexc);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_kp_u_no_path_and_sock_were_specified);
  Py_CLEAR(clear_module_state->__pyx_kp_u_no_python_buffer_is_allocated_in);
  Py_CLEAR(clear_module_state->__pyx_n_s_noop);
  Py_CLEAR(clear_module_state->__pyx_n_s_noop_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_obj);
  Py_CLEAR(clear_module_state->__pyx_n_s_object);
  Py_CLEAR(clear_module_state->__pyx_kp_u_object_created_at_most_recent_ca);
  Py_CLEAR(clear_module_state->__pyx_kp_u_objs_handles);
  Py_CLEAR(clear_module_state->__pyx_n_s_old_agen_hooks);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_ssl_connected);
  Py_CLEAR(clear_module_state->__pyx_n_s_open);
  Py_CLEAR(clear_module_state->__pyx_kp_u_open__StreamWriteContext_is_bein);
  Py_CLEAR(clear_module_state->__pyx_kp_u_open__UDPSendContext_is_being_de);
  Py_CLEAR(clear_module_state->__pyx_n_s_opts);
  Py_CLEAR(clear_module_state->__pyx_n_s_orig_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_os);
  Py_CLEAR(clear_module_state->__pyx_n_s_p_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_partial);
  Py_CLEAR(clear_module_state->__pyx_n_s_pass_fds);
  Py_CLEAR(clear_module_state->__pyx_n_s_path);
  Py_CLEAR(clear_module_state->__pyx_kp_u_path_and_sock_can_not_be_specifi);
  Py_CLEAR(clear_module_state->__pyx_kp_u_path_was_not_specified_and_no_so);
  Py_CLEAR(clear_module_state->__pyx_n_s_pause_reading);
  Py_CLEAR(clear_module_state->__pyx_n_s_pause_writing);
  Py_CLEAR(clear_module_state->__pyx_n_s_peercert);
  Py_CLEAR(clear_module_state->__pyx_n_u_peername);
  Py_CLEAR(clear_module_state->__pyx_n_s_pending);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_pipe);
  Py_CLEAR(clear_module_state->__pyx_n_u_pipe);
  Py_CLEAR(clear_module_state->__pyx_kp_u_pipe_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_pipe_connection_lost);
  Py_CLEAR(clear_module_state->__pyx_n_s_pipe_data_received);
  Py_CLEAR(clear_module_state->__pyx_n_s_platform);
  Py_CLEAR(clear_module_state->__pyx_n_s_pop);
  Py_CLEAR(clear_module_state->__pyx_n_s_popitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_popleft);
  Py_CLEAR(clear_module_state->__pyx_n_s_port);
  Py_CLEAR(clear_module_state->__pyx_kp_u_port_must_be_a_str_bytes_or_int);
  Py_CLEAR(clear_module_state->__pyx_n_u_posix);
  Py_CLEAR(clear_module_state->__pyx_n_s_preexec_fn);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_print);
  Py_CLEAR(clear_module_state->__pyx_n_u_print_debug_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_problems);
  Py_CLEAR(clear_module_state->__pyx_n_s_proc);
  Py_CLEAR(clear_module_state->__pyx_n_s_process_exited);
  Py_CLEAR(clear_module_state->__pyx_n_s_process_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_program);
  Py_CLEAR(clear_module_state->__pyx_n_s_proto);
  Py_CLEAR(clear_module_state->__pyx_n_s_proto_factory);
  Py_CLEAR(clear_module_state->__pyx_n_s_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_u_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_protocol_factory);
  Py_CLEAR(clear_module_state->__pyx_kp_u_protocol_is_not_set_cannot_call);
  Py_CLEAR(clear_module_state->__pyx_kp_u_protocol_is_required);
  Py_CLEAR(clear_module_state->__pyx_kp_u_protocol_pause_writing_failed);
  Py_CLEAR(clear_module_state->__pyx_kp_u_protocol_resume_writing_failed);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyaddr);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_LruCache);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PseudoSocket);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_u_qualname);
  Py_CLEAR(clear_module_state->__pyx_kp_u_r_SSL_handshake_took_1f_ms);
  Py_CLEAR(clear_module_state->__pyx_kp_u_r_fileno_returned_non_integer);
  Py_CLEAR(clear_module_state->__pyx_kp_u_r_received_EOF);
  Py_CLEAR(clear_module_state->__pyx_kp_u_r_s);
  Py_CLEAR(clear_module_state->__pyx_kp_u_r_starts_SSL_handshake);
  Py_CLEAR(clear_module_state->__pyx_kp_u_raddr_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_rads);
  Py_CLEAR(clear_module_state->__pyx_n_s_rai);
  Py_CLEAR(clear_module_state->__pyx_n_s_rai_addr_static);
  Py_CLEAR(clear_module_state->__pyx_n_s_rai_iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_rai_static);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_read);
  Py_CLEAR(clear_module_state->__pyx_n_s_recv);
  Py_CLEAR(clear_module_state->__pyx_n_s_recv_into);
  Py_CLEAR(clear_module_state->__pyx_kp_u_recv_into_method);
  Py_CLEAR(clear_module_state->__pyx_kp_u_recv_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_recvfrom);
  Py_CLEAR(clear_module_state->__pyx_n_s_recvfrom_into);
  Py_CLEAR(clear_module_state->__pyx_kp_u_recvfrom_into_method);
  Py_CLEAR(clear_module_state->__pyx_kp_u_recvfrom_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_recvmsg);
  Py_CLEAR(clear_module_state->__pyx_n_s_recvmsg_into);
  Py_CLEAR(clear_module_state->__pyx_kp_u_recvmsg_into_method);
  Py_CLEAR(clear_module_state->__pyx_kp_u_recvmsg_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_remote_addr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_remote_addr_must_be_a_tuple_of_h);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_done_callback);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_reader);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_reader_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_signal_handler);
  Py_CLEAR(clear_module_state->__pyx_kp_u_remove_signal_handler_can_only_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_writer);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove_writer_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_repr);
  Py_CLEAR(clear_module_state->__pyx_n_s_restore_signals);
  Py_CLEAR(clear_module_state->__pyx_n_s_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_results);
  Py_CLEAR(clear_module_state->__pyx_n_s_resume_cb);
  Py_CLEAR(clear_module_state->__pyx_n_s_resume_reading);
  Py_CLEAR(clear_module_state->__pyx_n_s_resume_writing);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_return_exceptions);
  Py_CLEAR(clear_module_state->__pyx_kp_u_returning_true_from_eof_received);
  Py_CLEAR(clear_module_state->__pyx_n_s_reuse_address);
  Py_CLEAR(clear_module_state->__pyx_n_s_reuse_port);
  Py_CLEAR(clear_module_state->__pyx_kp_u_reuse_port_not_supported_by_sock);
  Py_CLEAR(clear_module_state->__pyx_n_s_reverse);
  Py_CLEAR(clear_module_state->__pyx_n_s_round);
  Py_CLEAR(clear_module_state->__pyx_n_s_rstrip);
  Py_CLEAR(clear_module_state->__pyx_n_s_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_run_forever);
  Py_CLEAR(clear_module_state->__pyx_n_s_run_in_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_run_until_complete);
  Py_CLEAR(clear_module_state->__pyx_n_s_run_until_complete_locals_done_c);
  Py_CLEAR(clear_module_state->__pyx_kp_u_running_closed_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_scope_id);
  Py_CLEAR(clear_module_state->__pyx_kp_u_seconds);
  Py_CLEAR(clear_module_state->__pyx_kp_u_seconds_aborting_the_connection);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_self_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_self_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_self_pipe_exists_before_loop_run);
  Py_CLEAR(clear_module_state->__pyx_kp_u_self_pipe_was_not_cleaned_up_aft);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self_py_buf_self_req_cannot_be_c);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self_py_bufs_self_py_bufs_sml_se);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_kp_u_send_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_send_signal);
  Py_CLEAR(clear_module_state->__pyx_n_s_sendall);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sendall_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_sendfile);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sendfile_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_sendmsg);
  Py_CLEAR(clear_module_state->__pyx_n_s_sendmsg_afalg);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sendmsg_afalg_method);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sendmsg_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_sendto);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sendto_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_serve_forever);
  Py_CLEAR(clear_module_state->__pyx_kp_u_server);
  Py_CLEAR(clear_module_state->__pyx_n_s_server_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_server_hostname);
  Py_CLEAR(clear_module_state->__pyx_kp_u_server_hostname_is_only_meaningf);
  Py_CLEAR(clear_module_state->__pyx_n_s_server_side);
  Py_CLEAR(clear_module_state->__pyx_n_s_servers);
  Py_CLEAR(clear_module_state->__pyx_kp_u_service_proto_not_found);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_asyncgen_hooks);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_coroutine_origin_tracking_de);
  Py_CLEAR(clear_module_state->__pyx_n_u_set_coroutine_wrapper);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_debug);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_default_executor);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_exception);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_exception_handler);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_inheritable);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_read_buffer_limits);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_result);
  Py_CLEAR(clear_module_state->__pyx_n_u_set_running_loop);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_task_factory);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_wakeup_fd);
  Py_CLEAR(clear_module_state->__pyx_kp_u_set_wakeup_fd_1_failed_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_write_buffer_limits);
  Py_CLEAR(clear_module_state->__pyx_n_s_setblocking);
  Py_CLEAR(clear_module_state->__pyx_kp_u_setblocking_transport_sockets_ca);
  Py_CLEAR(clear_module_state->__pyx_n_s_setsockopt);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_settimeout);
  Py_CLEAR(clear_module_state->__pyx_kp_u_settimeout_only_0_timeout_is_all);
  Py_CLEAR(clear_module_state->__pyx_n_s_share);
  Py_CLEAR(clear_module_state->__pyx_n_s_shell);
  Py_CLEAR(clear_module_state->__pyx_kp_u_shell_must_be_False);
  Py_CLEAR(clear_module_state->__pyx_kp_u_shell_must_be_True);
  Py_CLEAR(clear_module_state->__pyx_n_s_shutdown);
  Py_CLEAR(clear_module_state->__pyx_n_s_shutdown_asyncgens);
  Py_CLEAR(clear_module_state->__pyx_n_s_shutdown_coro);
  Py_CLEAR(clear_module_state->__pyx_n_s_shutdown_default_executor);
  Py_CLEAR(clear_module_state->__pyx_kp_u_shutdown_method);
  Py_CLEAR(clear_module_state->__pyx_n_s_sig);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sig_cannot_be_caught);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sig_must_be_an_int_not_r);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sig_out_of_range_1);
  Py_CLEAR(clear_module_state->__pyx_n_s_sighandler);
  Py_CLEAR(clear_module_state->__pyx_n_s_siginterrupt);
  Py_CLEAR(clear_module_state->__pyx_n_s_signal);
  Py_CLEAR(clear_module_state->__pyx_kp_u_signals_handling_has_been_alread);
  Py_CLEAR(clear_module_state->__pyx_kp_u_signals_handling_has_not_been_se);
  Py_CLEAR(clear_module_state->__pyx_n_s_signum);
  Py_CLEAR(clear_module_state->__pyx_n_s_sin6);
  Py_CLEAR(clear_module_state->__pyx_n_s_sl);
  Py_CLEAR(clear_module_state->__pyx_n_s_sleep);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_accept);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_connect);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sock_connect_cb_is_called_on_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_recv);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_recv_into);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sock_recv_into_is_called_on_a_c);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sock_recv_is_called_on_a_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_recvfrom);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_recvfrom_into);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_sendall);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sock_sendall_is_called_on_a_can);
  Py_CLEAR(clear_module_state->__pyx_n_s_sock_sendto);
  Py_CLEAR(clear_module_state->__pyx_n_s_sockaddr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sockaddr_must_be_a_tuple_of_2_3);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sockaddr_resolved_to_multiple_ad);
  Py_CLEAR(clear_module_state->__pyx_kp_u_socket);
  Py_CLEAR(clear_module_state->__pyx_n_s_socket_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_socket_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_socket_family_mismatch_or_a_DNS);
  Py_CLEAR(clear_module_state->__pyx_kp_u_socket_modifier_keyword_argument);
  Py_CLEAR(clear_module_state->__pyx_kp_u_socket_send_raised_exception);
  Py_CLEAR(clear_module_state->__pyx_n_s_socketpair);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sockets);
  Py_CLEAR(clear_module_state->__pyx_n_s_sockets_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_sockname);
  Py_CLEAR(clear_module_state->__pyx_n_s_source);
  Py_CLEAR(clear_module_state->__pyx_n_u_source_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_split);
  Py_CLEAR(clear_module_state->__pyx_n_s_ssl);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ssl_argument_must_be_an_SSLConte);
  Py_CLEAR(clear_module_state->__pyx_n_s_ssl_handshake_timeout);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ssl_handshake_timeout_is_only_me);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ssl_handshake_timeout_should_be);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ssl_is_expected_to_be_None_or_an);
  Py_CLEAR(clear_module_state->__pyx_n_s_ssl_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_ssl_protocol);
  Py_CLEAR(clear_module_state->__pyx_n_s_ssl_shutdown_timeout);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ssl_shutdown_timeout_is_only_mea);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ssl_shutdown_timeout_should_be_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_ssl_waiter);
  Py_CLEAR(clear_module_state->__pyx_n_s_sslcontext);
  Py_CLEAR(clear_module_state->__pyx_kp_u_sslcontext_is_expected_to_be_an);
  Py_CLEAR(clear_module_state->__pyx_n_s_sslproto);
  Py_CLEAR(clear_module_state->__pyx_n_s_st_ino);
  Py_CLEAR(clear_module_state->__pyx_n_s_st_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_stacklevel);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_kp_u_start_init_can_only_be_called);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_new_session);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_serving);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_tls);
  Py_CLEAR(clear_module_state->__pyx_n_s_startupinfo);
  Py_CLEAR(clear_module_state->__pyx_kp_u_startupinfo_is_not_supported);
  Py_CLEAR(clear_module_state->__pyx_n_s_stat);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_stderr);
  Py_CLEAR(clear_module_state->__pyx_n_s_stdin);
  Py_CLEAR(clear_module_state->__pyx_n_s_stdio_fut);
  Py_CLEAR(clear_module_state->__pyx_n_s_stdio_inited);
  Py_CLEAR(clear_module_state->__pyx_n_s_stdout);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_kp_u_stream_is_open_after_UVStream__t);
  Py_CLEAR(clear_module_state->__pyx_n_s_strerror);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_submit);
  Py_CLEAR(clear_module_state->__pyx_n_s_subprocess);
  Py_CLEAR(clear_module_state->__pyx_kp_u_subprocess_STDOUT_is_supported_o);
  Py_CLEAR(clear_module_state->__pyx_n_s_subprocess_exec);
  Py_CLEAR(clear_module_state->__pyx_n_s_subprocess_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_subprocess_shell);
  Py_CLEAR(clear_module_state->__pyx_n_s_subtract);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_sys);
  Py_CLEAR(clear_module_state->__pyx_n_s_target);
  Py_CLEAR(clear_module_state->__pyx_n_s_task);
  Py_CLEAR(clear_module_state->__pyx_n_u_task);
  Py_CLEAR(clear_module_state->__pyx_kp_u_task_factory_must_be_a_callable);
  Py_CLEAR(clear_module_state->__pyx_n_s_tb);
  Py_CLEAR(clear_module_state->__pyx_n_s_tcp);
  Py_CLEAR(clear_module_state->__pyx_n_u_tcp);
  Py_CLEAR(clear_module_state->__pyx_n_s_terminate);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_test__append_write_backlog);
  Py_CLEAR(clear_module_state->__pyx_n_s_test_coroutine_1);
  Py_CLEAR(clear_module_state->__pyx_n_s_testhelper_unwrap_capsuled_poin);
  Py_CLEAR(clear_module_state->__pyx_kp_u_the_socket_must_be_non_blocking);
  Py_CLEAR(clear_module_state->__pyx_kp_u_this_event_loop_is_already_runni);
  Py_CLEAR(clear_module_state->__pyx_n_s_thread);
  Py_CLEAR(clear_module_state->__pyx_n_s_threading);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_timeout);
  Py_CLEAR(clear_module_state->__pyx_n_s_tr);
  Py_CLEAR(clear_module_state->__pyx_n_s_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_traceback_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_transp);
  Py_CLEAR(clear_module_state->__pyx_n_s_transport);
  Py_CLEAR(clear_module_state->__pyx_n_u_transport);
  Py_CLEAR(clear_module_state->__pyx_kp_u_transport_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_transport_sockets_do_not_support);
  Py_CLEAR(clear_module_state->__pyx_n_s_transport_waiter);
  Py_CLEAR(clear_module_state->__pyx_n_s_transports);
  Py_CLEAR(clear_module_state->__pyx_kp_u_try_write_sent_all_data_and_ret);
  Py_CLEAR(clear_module_state->__pyx_n_s_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_udp);
  Py_CLEAR(clear_module_state->__pyx_n_u_udp);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unable_to_listen_no_protocol_fac);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unable_to_perform_operation_on_r);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unable_to_start_the_loop_it_was);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unclosed_resource_r);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unclosed_resource_r_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unclosed_resource_r_could_not_se);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unclosed_transport_uvloop_loop);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unexpected_address_family);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unexpected_kwargs);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unhandled_error_or_an_empty_buff);
  Py_CLEAR(clear_module_state->__pyx_n_s_universal_newlines);
  Py_CLEAR(clear_module_state->__pyx_kp_u_universal_newlines_must_be_False);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unix_socket_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_unlink);
  Py_CLEAR(clear_module_state->__pyx_kp_u_unopened_TCPServer);
  Py_CLEAR(clear_module_state->__pyx_n_s_unset);
  Py_CLEAR(clear_module_state->__pyx_n_s_unwrap);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_kp_u_uv_close_handle_cb_handle_loop);
  Py_CLEAR(clear_module_state->__pyx_kp_u_uv_handle_t_current_freed_total);
  Py_CLEAR(clear_module_state->__pyx_kp_u_uvloop_PseudoSocket_fd_family_s);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_cbhandles_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_basetransport_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_fsevent_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_pipe_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_process_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_stream_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_streamserver_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_tcp_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_handles_udp_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_uvloop_loop);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_loop_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_pseudosock_pyx);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_server_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_u_uvloop_sleep_after_fork);
  Py_CLEAR(clear_module_state->__pyx_kp_u_uvloop_sslproto);
  Py_CLEAR(clear_module_state->__pyx_kp_s_uvloop_sslproto_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_version_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_wait);
  Py_CLEAR(clear_module_state->__pyx_n_s_wait_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_wait_closed);
  Py_CLEAR(clear_module_state->__pyx_n_s_waiter);
  Py_CLEAR(clear_module_state->__pyx_n_s_walk_stack);
  Py_CLEAR(clear_module_state->__pyx_n_s_want);
  Py_CLEAR(clear_module_state->__pyx_n_s_warn);
  Py_CLEAR(clear_module_state->__pyx_n_s_warn_on_full_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_warning);
  Py_CLEAR(clear_module_state->__pyx_n_s_warnings);
  Py_CLEAR(clear_module_state->__pyx_n_u_wb);
  Py_CLEAR(clear_module_state->__pyx_n_s_weakref);
  Py_CLEAR(clear_module_state->__pyx_n_s_when);
  Py_CLEAR(clear_module_state->__pyx_kp_u_without_no_gc_clear_loop_was_se);
  Py_CLEAR(clear_module_state->__pyx_n_s_wrap);
  Py_CLEAR(clear_module_state->__pyx_n_s_wrap_bio);
  Py_CLEAR(clear_module_state->__pyx_n_s_wrap_future);
  Py_CLEAR(clear_module_state->__pyx_n_s_write);
  Py_CLEAR(clear_module_state->__pyx_n_s_write_eof);
  Py_CLEAR(clear_module_state->__pyx_n_s_writelines);
  Py_CLEAR(clear_module_state->__pyx_n_s_zip);
  Py_CLEAR(clear_module_state->__pyx_float_0_1);
  Py_CLEAR(clear_module_state->__pyx_float_1e3);
  Py_CLEAR(clear_module_state->__pyx_float_1eneg_3);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_3);
  Py_CLEAR(clear_module_state->__pyx_int_15);
  Py_CLEAR(clear_module_state->__pyx_int_42);
  Py_CLEAR(clear_module_state->__pyx_int_100);
  Py_CLEAR(clear_module_state->__pyx_int_1000);
  Py_CLEAR(clear_module_state->__pyx_int_50000);
  Py_CLEAR(clear_module_state->__pyx_int_65536);
  Py_CLEAR(clear_module_state->__pyx_int_1048575);
  Py_CLEAR(clear_module_state->__pyx_int_14808274);
  Py_CLEAR(clear_module_state->__pyx_int_87255255);
  Py_CLEAR(clear_module_state->__pyx_int_121062814);
  Py_CLEAR(clear_module_state->__pyx_int_143395587);
  Py_CLEAR(clear_module_state->__pyx_int_173349396);
  Py_CLEAR(clear_module_state->__pyx_int_200977156);
  Py_CLEAR(clear_module_state->__pyx_int_4294967296);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_k__71);
  Py_CLEAR(clear_module_state->__pyx_k__95);
  Py_CLEAR(clear_module_state->__pyx_k__96);
  Py_CLEAR(clear_module_state->__pyx_k__97);
  Py_CLEAR(clear_module_state->__pyx_k__114);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__3);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_tuple__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_slice__88);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__18);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__23);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__27);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__29);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__31);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__40);
  Py_CLEAR(clear_module_state->__pyx_tuple__42);
  Py_CLEAR(clear_module_state->__pyx_tuple__45);
  Py_CLEAR(clear_module_state->__pyx_tuple__46);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__48);
  Py_CLEAR(clear_module_state->__pyx_tuple__49);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__55);
  Py_CLEAR(clear_module_state->__pyx_tuple__56);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__58);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__60);
  Py_CLEAR(clear_module_state->__pyx_tuple__62);
  Py_CLEAR(clear_module_state->__pyx_tuple__64);
  Py_CLEAR(clear_module_state->__pyx_tuple__65);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__67);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_tuple__70);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__74);
  Py_CLEAR(clear_module_state->__pyx_tuple__77);
  Py_CLEAR(clear_module_state->__pyx_tuple__80);
  Py_CLEAR(clear_module_state->__pyx_tuple__81);
  Py_CLEAR(clear_module_state->__pyx_tuple__83);
  Py_CLEAR(clear_module_state->__pyx_tuple__93);
  Py_CLEAR(clear_module_state->__pyx_tuple__94);
  Py_CLEAR(clear_module_state->__pyx_tuple__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__100);
  Py_CLEAR(clear_module_state->__pyx_tuple__101);
  Py_CLEAR(clear_module_state->__pyx_tuple__102);
  Py_CLEAR(clear_module_state->__pyx_tuple__104);
  Py_CLEAR(clear_module_state->__pyx_tuple__106);
  Py_CLEAR(clear_module_state->__pyx_tuple__109);
  Py_CLEAR(clear_module_state->__pyx_tuple__110);
  Py_CLEAR(clear_module_state->__pyx_tuple__111);
  Py_CLEAR(clear_module_state->__pyx_tuple__112);
  Py_CLEAR(clear_module_state->__pyx_tuple__113);
  Py_CLEAR(clear_module_state->__pyx_tuple__117);
  Py_CLEAR(clear_module_state->__pyx_tuple__118);
  Py_CLEAR(clear_module_state->__pyx_tuple__119);
  Py_CLEAR(clear_module_state->__pyx_tuple__120);
  Py_CLEAR(clear_module_state->__pyx_tuple__121);
  Py_CLEAR(clear_module_state->__pyx_tuple__122);
  Py_CLEAR(clear_module_state->__pyx_tuple__125);
  Py_CLEAR(clear_module_state->__pyx_tuple__126);
  Py_CLEAR(clear_module_state->__pyx_tuple__127);
  Py_CLEAR(clear_module_state->__pyx_tuple__131);
  Py_CLEAR(clear_module_state->__pyx_tuple__132);
  Py_CLEAR(clear_module_state->__pyx_tuple__133);
  Py_CLEAR(clear_module_state->__pyx_tuple__134);
  Py_CLEAR(clear_module_state->__pyx_tuple__135);
  Py_CLEAR(clear_module_state->__pyx_tuple__136);
  Py_CLEAR(clear_module_state->__pyx_tuple__137);
  Py_CLEAR(clear_module_state->__pyx_tuple__138);
  Py_CLEAR(clear_module_state->__pyx_tuple__139);
  Py_CLEAR(clear_module_state->__pyx_tuple__142);
  Py_CLEAR(clear_module_state->__pyx_tuple__143);
  Py_CLEAR(clear_module_state->__pyx_tuple__144);
  Py_CLEAR(clear_module_state->__pyx_tuple__145);
  Py_CLEAR(clear_module_state->__pyx_tuple__146);
  Py_CLEAR(clear_module_state->__pyx_tuple__147);
  Py_CLEAR(clear_module_state->__pyx_tuple__148);
  Py_CLEAR(clear_module_state->__pyx_tuple__149);
  Py_CLEAR(clear_module_state->__pyx_tuple__150);
  Py_CLEAR(clear_module_state->__pyx_tuple__151);
  Py_CLEAR(clear_module_state->__pyx_tuple__152);
  Py_CLEAR(clear_module_state->__pyx_tuple__153);
  Py_CLEAR(clear_module_state->__pyx_tuple__154);
  Py_CLEAR(clear_module_state->__pyx_tuple__155);
  Py_CLEAR(clear_module_state->__pyx_tuple__156);
  Py_CLEAR(clear_module_state->__pyx_tuple__157);
  Py_CLEAR(clear_module_state->__pyx_tuple__158);
  Py_CLEAR(clear_module_state->__pyx_tuple__159);
  Py_CLEAR(clear_module_state->__pyx_tuple__160);
  Py_CLEAR(clear_module_state->__pyx_tuple__162);
  Py_CLEAR(clear_module_state->__pyx_tuple__163);
  Py_CLEAR(clear_module_state->__pyx_tuple__164);
  Py_CLEAR(clear_module_state->__pyx_tuple__165);
  Py_CLEAR(clear_module_state->__pyx_tuple__166);
  Py_CLEAR(clear_module_state->__pyx_tuple__167);
  Py_CLEAR(clear_module_state->__pyx_tuple__168);
  Py_CLEAR(clear_module_state->__pyx_tuple__169);
  Py_CLEAR(clear_module_state->__pyx_tuple__171);
  Py_CLEAR(clear_module_state->__pyx_tuple__172);
  Py_CLEAR(clear_module_state->__pyx_tuple__173);
  Py_CLEAR(clear_module_state->__pyx_tuple__174);
  Py_CLEAR(clear_module_state->__pyx_tuple__175);
  Py_CLEAR(clear_module_state->__pyx_tuple__176);
  Py_CLEAR(clear_module_state->__pyx_tuple__177);
  Py_CLEAR(clear_module_state->__pyx_tuple__178);
  Py_CLEAR(clear_module_state->__pyx_tuple__179);
  Py_CLEAR(clear_module_state->__pyx_tuple__180);
  Py_CLEAR(clear_module_state->__pyx_tuple__181);
  Py_CLEAR(clear_module_state->__pyx_tuple__182);
  Py_CLEAR(clear_module_state->__pyx_tuple__183);
  Py_CLEAR(clear_module_state->__pyx_tuple__184);
  Py_CLEAR(clear_module_state->__pyx_tuple__185);
  Py_CLEAR(clear_module_state->__pyx_tuple__186);
  Py_CLEAR(clear_module_state->__pyx_tuple__187);
  Py_CLEAR(clear_module_state->__pyx_tuple__188);
  Py_CLEAR(clear_module_state->__pyx_tuple__195);
  Py_CLEAR(clear_module_state->__pyx_tuple__197);
  Py_CLEAR(clear_module_state->__pyx_tuple__199);
  Py_CLEAR(clear_module_state->__pyx_tuple__200);
  Py_CLEAR(clear_module_state->__pyx_tuple__201);
  Py_CLEAR(clear_module_state->__pyx_tuple__202);
  Py_CLEAR(clear_module_state->__pyx_tuple__203);
  Py_CLEAR(clear_module_state->__pyx_tuple__204);
  Py_CLEAR(clear_module_state->__pyx_tuple__205);
  Py_CLEAR(clear_module_state->__pyx_tuple__206);
  Py_CLEAR(clear_module_state->__pyx_tuple__207);
  Py_CLEAR(clear_module_state->__pyx_tuple__209);
  Py_CLEAR(clear_module_state->__pyx_tuple__211);
  Py_CLEAR(clear_module_state->__pyx_tuple__213);
  Py_CLEAR(clear_module_state->__pyx_tuple__215);
  Py_CLEAR(clear_module_state->__pyx_tuple__217);
  Py_CLEAR(clear_module_state->__pyx_tuple__221);
  Py_CLEAR(clear_module_state->__pyx_tuple__229);
  Py_CLEAR(clear_module_state->__pyx_tuple__231);
  Py_CLEAR(clear_module_state->__pyx_tuple__234);
  Py_CLEAR(clear_module_state->__pyx_tuple__236);
  Py_CLEAR(clear_module_state->__pyx_tuple__237);
  Py_CLEAR(clear_module_state->__pyx_tuple__238);
  Py_CLEAR(clear_module_state->__pyx_tuple__239);
  Py_CLEAR(clear_module_state->__pyx_tuple__240);
  Py_CLEAR(clear_module_state->__pyx_tuple__241);
  Py_CLEAR(clear_module_state->__pyx_tuple__242);
  Py_CLEAR(clear_module_state->__pyx_tuple__243);
  Py_CLEAR(clear_module_state->__pyx_tuple__244);
  Py_CLEAR(clear_module_state->__pyx_tuple__245);
  Py_CLEAR(clear_module_state->__pyx_tuple__246);
  Py_CLEAR(clear_module_state->__pyx_tuple__249);
  Py_CLEAR(clear_module_state->__pyx_tuple__251);
  Py_CLEAR(clear_module_state->__pyx_tuple__253);
  Py_CLEAR(clear_module_state->__pyx_tuple__255);
  Py_CLEAR(clear_module_state->__pyx_tuple__259);
  Py_CLEAR(clear_module_state->__pyx_tuple__260);
  Py_CLEAR(clear_module_state->__pyx_tuple__261);
  Py_CLEAR(clear_module_state->__pyx_tuple__262);
  Py_CLEAR(clear_module_state->__pyx_tuple__264);
  Py_CLEAR(clear_module_state->__pyx_tuple__265);
  Py_CLEAR(clear_module_state->__pyx_tuple__266);
  Py_CLEAR(clear_module_state->__pyx_tuple__267);
  Py_CLEAR(clear_module_state->__pyx_tuple__268);
  Py_CLEAR(clear_module_state->__pyx_tuple__269);
  Py_CLEAR(clear_module_state->__pyx_tuple__271);
  Py_CLEAR(clear_module_state->__pyx_tuple__273);
  Py_CLEAR(clear_module_state->__pyx_tuple__274);
  Py_CLEAR(clear_module_state->__pyx_tuple__275);
  Py_CLEAR(clear_module_state->__pyx_tuple__276);
  Py_CLEAR(clear_module_state->__pyx_tuple__277);
  Py_CLEAR(clear_module_state->__pyx_tuple__279);
  Py_CLEAR(clear_module_state->__pyx_tuple__281);
  Py_CLEAR(clear_module_state->__pyx_tuple__282);
  Py_CLEAR(clear_module_state->__pyx_tuple__285);
  Py_CLEAR(clear_module_state->__pyx_tuple__288);
  Py_CLEAR(clear_module_state->__pyx_tuple__289);
  Py_CLEAR(clear_module_state->__pyx_tuple__290);
  Py_CLEAR(clear_module_state->__pyx_tuple__293);
  Py_CLEAR(clear_module_state->__pyx_tuple__295);
  Py_CLEAR(clear_module_state->__pyx_tuple__298);
  Py_CLEAR(clear_module_state->__pyx_tuple__300);
  Py_CLEAR(clear_module_state->__pyx_tuple__303);
  Py_CLEAR(clear_module_state->__pyx_tuple__321);
  Py_CLEAR(clear_module_state->__pyx_tuple__325);
  Py_CLEAR(clear_module_state->__pyx_tuple__327);
  Py_CLEAR(clear_module_state->__pyx_tuple__330);
  Py_CLEAR(clear_module_state->__pyx_tuple__333);
  Py_CLEAR(clear_module_state->__pyx_tuple__336);
  Py_CLEAR(clear_module_state->__pyx_tuple__357);
  Py_CLEAR(clear_module_state->__pyx_tuple__360);
  Py_CLEAR(clear_module_state->__pyx_tuple__363);
  Py_CLEAR(clear_module_state->__pyx_tuple__365);
  Py_CLEAR(clear_module_state->__pyx_tuple__385);
  Py_CLEAR(clear_module_state->__pyx_tuple__387);
  Py_CLEAR(clear_module_state->__pyx_tuple__393);
  Py_CLEAR(clear_module_state->__pyx_tuple__396);
  Py_CLEAR(clear_module_state->__pyx_tuple__402);
  Py_CLEAR(clear_module_state->__pyx_tuple__404);
  Py_CLEAR(clear_module_state->__pyx_tuple__413);
  Py_CLEAR(clear_module_state->__pyx_tuple__431);
  Py_CLEAR(clear_module_state->__pyx_tuple__433);
  Py_CLEAR(clear_module_state->__pyx_tuple__448);
  Py_CLEAR(clear_module_state->__pyx_tuple__454);
  Py_CLEAR(clear_module_state->__pyx_tuple__458);
  Py_CLEAR(clear_module_state->__pyx_tuple__463);
  Py_CLEAR(clear_module_state->__pyx_tuple__467);
  Py_CLEAR(clear_module_state->__pyx_tuple__469);
  Py_CLEAR(clear_module_state->__pyx_tuple__513);
  Py_CLEAR(clear_module_state->__pyx_tuple__517);
  Py_CLEAR(clear_module_state->__pyx_tuple__519);
  Py_CLEAR(clear_module_state->__pyx_tuple__528);
  Py_CLEAR(clear_module_state->__pyx_tuple__533);
  Py_CLEAR(clear_module_state->__pyx_tuple__534);
  Py_CLEAR(clear_module_state->__pyx_tuple__539);
  Py_CLEAR(clear_module_state->__pyx_codeobj__20);
  Py_CLEAR(clear_module_state->__pyx_codeobj__22);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__84);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  Py_CLEAR(clear_module_state->__pyx_codeobj__89);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__91);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__98);
  Py_CLEAR(clear_module_state->__pyx_codeobj__103);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__107);
  Py_CLEAR(clear_module_state->__pyx_codeobj__108);
  Py_CLEAR(clear_module_state->__pyx_codeobj__116);
  Py_CLEAR(clear_module_state->__pyx_codeobj__123);
  Py_CLEAR(clear_module_state->__pyx_codeobj__124);
  Py_CLEAR(clear_module_state->__pyx_codeobj__189);
  Py_CLEAR(clear_module_state->__pyx_codeobj__190);
  Py_CLEAR(clear_module_state->__pyx_codeobj__191);
  Py_CLEAR(clear_module_state->__pyx_codeobj__192);
  Py_CLEAR(clear_module_state->__pyx_codeobj__193);
  Py_CLEAR(clear_module_state->__pyx_codeobj__194);
  Py_CLEAR(clear_module_state->__pyx_codeobj__208);
  Py_CLEAR(clear_module_state->__pyx_codeobj__210);
  Py_CLEAR(clear_module_state->__pyx_codeobj__212);
  Py_CLEAR(clear_module_state->__pyx_codeobj__214);
  Py_CLEAR(clear_module_state->__pyx_codeobj__216);
  Py_CLEAR(clear_module_state->__pyx_codeobj__218);
  Py_CLEAR(clear_module_state->__pyx_codeobj__219);
  Py_CLEAR(clear_module_state->__pyx_codeobj__220);
  Py_CLEAR(clear_module_state->__pyx_codeobj__222);
  Py_CLEAR(clear_module_state->__pyx_codeobj__223);
  Py_CLEAR(clear_module_state->__pyx_codeobj__224);
  Py_CLEAR(clear_module_state->__pyx_codeobj__225);
  Py_CLEAR(clear_module_state->__pyx_codeobj__226);
  Py_CLEAR(clear_module_state->__pyx_codeobj__227);
  Py_CLEAR(clear_module_state->__pyx_codeobj__228);
  Py_CLEAR(clear_module_state->__pyx_codeobj__230);
  Py_CLEAR(clear_module_state->__pyx_codeobj__232);
  Py_CLEAR(clear_module_state->__pyx_codeobj__233);
  Py_CLEAR(clear_module_state->__pyx_codeobj__235);
  Py_CLEAR(clear_module_state->__pyx_codeobj__247);
  Py_CLEAR(clear_module_state->__pyx_codeobj__248);
  Py_CLEAR(clear_module_state->__pyx_codeobj__250);
  Py_CLEAR(clear_module_state->__pyx_codeobj__252);
  Py_CLEAR(clear_module_state->__pyx_codeobj__254);
  Py_CLEAR(clear_module_state->__pyx_codeobj__256);
  Py_CLEAR(clear_module_state->__pyx_codeobj__257);
  Py_CLEAR(clear_module_state->__pyx_codeobj__258);
  Py_CLEAR(clear_module_state->__pyx_codeobj__270);
  Py_CLEAR(clear_module_state->__pyx_codeobj__272);
  Py_CLEAR(clear_module_state->__pyx_codeobj__278);
  Py_CLEAR(clear_module_state->__pyx_codeobj__280);
  Py_CLEAR(clear_module_state->__pyx_codeobj__283);
  Py_CLEAR(clear_module_state->__pyx_codeobj__284);
  Py_CLEAR(clear_module_state->__pyx_codeobj__286);
  Py_CLEAR(clear_module_state->__pyx_codeobj__287);
  Py_CLEAR(clear_module_state->__pyx_codeobj__291);
  Py_CLEAR(clear_module_state->__pyx_codeobj__292);
  Py_CLEAR(clear_module_state->__pyx_codeobj__294);
  Py_CLEAR(clear_module_state->__pyx_codeobj__296);
  Py_CLEAR(clear_module_state->__pyx_codeobj__297);
  Py_CLEAR(clear_module_state->__pyx_codeobj__299);
  Py_CLEAR(clear_module_state->__pyx_codeobj__301);
  Py_CLEAR(clear_module_state->__pyx_codeobj__302);
  Py_CLEAR(clear_module_state->__pyx_codeobj__304);
  Py_CLEAR(clear_module_state->__pyx_codeobj__305);
  Py_CLEAR(clear_module_state->__pyx_codeobj__306);
  Py_CLEAR(clear_module_state->__pyx_codeobj__307);
  Py_CLEAR(clear_module_state->__pyx_codeobj__308);
  Py_CLEAR(clear_module_state->__pyx_codeobj__309);
  Py_CLEAR(clear_module_state->__pyx_codeobj__310);
  Py_CLEAR(clear_module_state->__pyx_codeobj__311);
  Py_CLEAR(clear_module_state->__pyx_codeobj__312);
  Py_CLEAR(clear_module_state->__pyx_codeobj__313);
  Py_CLEAR(clear_module_state->__pyx_codeobj__314);
  Py_CLEAR(clear_module_state->__pyx_codeobj__315);
  Py_CLEAR(clear_module_state->__pyx_codeobj__316);
  Py_CLEAR(clear_module_state->__pyx_codeobj__317);
  Py_CLEAR(clear_module_state->__pyx_codeobj__318);
  Py_CLEAR(clear_module_state->__pyx_codeobj__319);
  Py_CLEAR(clear_module_state->__pyx_codeobj__320);
  Py_CLEAR(clear_module_state->__pyx_codeobj__322);
  Py_CLEAR(clear_module_state->__pyx_codeobj__323);
  Py_CLEAR(clear_module_state->__pyx_codeobj__324);
  Py_CLEAR(clear_module_state->__pyx_codeobj__326);
  Py_CLEAR(clear_module_state->__pyx_codeobj__328);
  Py_CLEAR(clear_module_state->__pyx_codeobj__329);
  Py_CLEAR(clear_module_state->__pyx_codeobj__331);
  Py_CLEAR(clear_module_state->__pyx_codeobj__332);
  Py_CLEAR(clear_module_state->__pyx_codeobj__334);
  Py_CLEAR(clear_module_state->__pyx_codeobj__335);
  Py_CLEAR(clear_module_state->__pyx_codeobj__337);
  Py_CLEAR(clear_module_state->__pyx_codeobj__338);
  Py_CLEAR(clear_module_state->__pyx_codeobj__339);
  Py_CLEAR(clear_module_state->__pyx_codeobj__340);
  Py_CLEAR(clear_module_state->__pyx_codeobj__341);
  Py_CLEAR(clear_module_state->__pyx_codeobj__342);
  Py_CLEAR(clear_module_state->__pyx_codeobj__343);
  Py_CLEAR(clear_module_state->__pyx_codeobj__344);
  Py_CLEAR(clear_module_state->__pyx_codeobj__345);
  Py_CLEAR(clear_module_state->__pyx_codeobj__346);
  Py_CLEAR(clear_module_state->__pyx_codeobj__347);
  Py_CLEAR(clear_module_state->__pyx_codeobj__348);
  Py_CLEAR(clear_module_state->__pyx_codeobj__349);
  Py_CLEAR(clear_module_state->__pyx_codeobj__350);
  Py_CLEAR(clear_module_state->__pyx_codeobj__351);
  Py_CLEAR(clear_module_state->__pyx_codeobj__352);
  Py_CLEAR(clear_module_state->__pyx_codeobj__353);
  Py_CLEAR(clear_module_state->__pyx_codeobj__354);
  Py_CLEAR(clear_module_state->__pyx_codeobj__355);
  Py_CLEAR(clear_module_state->__pyx_codeobj__356);
  Py_CLEAR(clear_module_state->__pyx_codeobj__358);
  Py_CLEAR(clear_module_state->__pyx_codeobj__359);
  Py_CLEAR(clear_module_state->__pyx_codeobj__361);
  Py_CLEAR(clear_module_state->__pyx_codeobj__362);
  Py_CLEAR(clear_module_state->__pyx_codeobj__364);
  Py_CLEAR(clear_module_state->__pyx_codeobj__366);
  Py_CLEAR(clear_module_state->__pyx_codeobj__367);
  Py_CLEAR(clear_module_state->__pyx_codeobj__368);
  Py_CLEAR(clear_module_state->__pyx_codeobj__369);
  Py_CLEAR(clear_module_state->__pyx_codeobj__370);
  Py_CLEAR(clear_module_state->__pyx_codeobj__371);
  Py_CLEAR(clear_module_state->__pyx_codeobj__372);
  Py_CLEAR(clear_module_state->__pyx_codeobj__373);
  Py_CLEAR(clear_module_state->__pyx_codeobj__374);
  Py_CLEAR(clear_module_state->__pyx_codeobj__375);
  Py_CLEAR(clear_module_state->__pyx_codeobj__376);
  Py_CLEAR(clear_module_state->__pyx_codeobj__377);
  Py_CLEAR(clear_module_state->__pyx_codeobj__378);
  Py_CLEAR(clear_module_state->__pyx_codeobj__379);
  Py_CLEAR(clear_module_state->__pyx_codeobj__380);
  Py_CLEAR(clear_module_state->__pyx_codeobj__381);
  Py_CLEAR(clear_module_state->__pyx_codeobj__382);
  Py_CLEAR(clear_module_state->__pyx_codeobj__383);
  Py_CLEAR(clear_module_state->__pyx_codeobj__384);
  Py_CLEAR(clear_module_state->__pyx_codeobj__386);
  Py_CLEAR(clear_module_state->__pyx_codeobj__388);
  Py_CLEAR(clear_module_state->__pyx_codeobj__389);
  Py_CLEAR(clear_module_state->__pyx_codeobj__390);
  Py_CLEAR(clear_module_state->__pyx_codeobj__391);
  Py_CLEAR(clear_module_state->__pyx_codeobj__392);
  Py_CLEAR(clear_module_state->__pyx_codeobj__394);
  Py_CLEAR(clear_module_state->__pyx_codeobj__395);
  Py_CLEAR(clear_module_state->__pyx_codeobj__397);
  Py_CLEAR(clear_module_state->__pyx_codeobj__398);
  Py_CLEAR(clear_module_state->__pyx_codeobj__399);
  Py_CLEAR(clear_module_state->__pyx_codeobj__400);
  Py_CLEAR(clear_module_state->__pyx_codeobj__401);
  Py_CLEAR(clear_module_state->__pyx_codeobj__403);
  Py_CLEAR(clear_module_state->__pyx_codeobj__405);
  Py_CLEAR(clear_module_state->__pyx_codeobj__406);
  Py_CLEAR(clear_module_state->__pyx_codeobj__407);
  Py_CLEAR(clear_module_state->__pyx_codeobj__408);
  Py_CLEAR(clear_module_state->__pyx_codeobj__409);
  Py_CLEAR(clear_module_state->__pyx_codeobj__410);
  Py_CLEAR(clear_module_state->__pyx_codeobj__411);
  Py_CLEAR(clear_module_state->__pyx_codeobj__412);
  Py_CLEAR(clear_module_state->__pyx_codeobj__414);
  Py_CLEAR(clear_module_state->__pyx_codeobj__415);
  Py_CLEAR(clear_module_state->__pyx_codeobj__416);
  Py_CLEAR(clear_module_state->__pyx_codeobj__417);
  Py_CLEAR(clear_module_state->__pyx_codeobj__418);
  Py_CLEAR(clear_module_state->__pyx_codeobj__419);
  Py_CLEAR(clear_module_state->__pyx_codeobj__420);
  Py_CLEAR(clear_module_state->__pyx_codeobj__421);
  Py_CLEAR(clear_module_state->__pyx_codeobj__422);
  Py_CLEAR(clear_module_state->__pyx_codeobj__423);
  Py_CLEAR(clear_module_state->__pyx_codeobj__424);
  Py_CLEAR(clear_module_state->__pyx_codeobj__425);
  Py_CLEAR(clear_module_state->__pyx_codeobj__426);
  Py_CLEAR(clear_module_state->__pyx_codeobj__427);
  Py_CLEAR(clear_module_state->__pyx_codeobj__428);
  Py_CLEAR(clear_module_state->__pyx_codeobj__429);
  Py_CLEAR(clear_module_state->__pyx_codeobj__430);
  Py_CLEAR(clear_module_state->__pyx_codeobj__432);
  Py_CLEAR(clear_module_state->__pyx_codeobj__434);
  Py_CLEAR(clear_module_state->__pyx_codeobj__435);
  Py_CLEAR(clear_module_state->__pyx_codeobj__436);
  Py_CLEAR(clear_module_state->__pyx_codeobj__437);
  Py_CLEAR(clear_module_state->__pyx_codeobj__438);
  Py_CLEAR(clear_module_state->__pyx_codeobj__439);
  Py_CLEAR(clear_module_state->__pyx_codeobj__440);
  Py_CLEAR(clear_module_state->__pyx_codeobj__441);
  Py_CLEAR(clear_module_state->__pyx_codeobj__442);
  Py_CLEAR(clear_module_state->__pyx_codeobj__443);
  Py_CLEAR(clear_module_state->__pyx_codeobj__444);
  Py_CLEAR(clear_module_state->__pyx_codeobj__445);
  Py_CLEAR(clear_module_state->__pyx_codeobj__446);
  Py_CLEAR(clear_module_state->__pyx_codeobj__447);
  Py_CLEAR(clear_module_state->__pyx_codeobj__449);
  Py_CLEAR(clear_module_state->__pyx_codeobj__450);
  Py_CLEAR(clear_module_state->__pyx_codeobj__451);
  Py_CLEAR(clear_module_state->__pyx_codeobj__452);
  Py_CLEAR(clear_module_state->__pyx_codeobj__453);
  Py_CLEAR(clear_module_state->__pyx_codeobj__455);
  Py_CLEAR(clear_module_state->__pyx_codeobj__456);
  Py_CLEAR(clear_module_state->__pyx_codeobj__457);
  Py_CLEAR(clear_module_state->__pyx_codeobj__459);
  Py_CLEAR(clear_module_state->__pyx_codeobj__460);
  Py_CLEAR(clear_module_state->__pyx_codeobj__461);
  Py_CLEAR(clear_module_state->__pyx_codeobj__462);
  Py_CLEAR(clear_module_state->__pyx_codeobj__464);
  Py_CLEAR(clear_module_state->__pyx_codeobj__465);
  Py_CLEAR(clear_module_state->__pyx_codeobj__466);
  Py_CLEAR(clear_module_state->__pyx_codeobj__468);
  Py_CLEAR(clear_module_state->__pyx_codeobj__470);
  Py_CLEAR(clear_module_state->__pyx_codeobj__471);
  Py_CLEAR(clear_module_state->__pyx_codeobj__472);
  Py_CLEAR(clear_module_state->__pyx_codeobj__473);
  Py_CLEAR(clear_module_state->__pyx_codeobj__474);
  Py_CLEAR(clear_module_state->__pyx_codeobj__475);
  Py_CLEAR(clear_module_state->__pyx_codeobj__476);
  Py_CLEAR(clear_module_state->__pyx_codeobj__477);
  Py_CLEAR(clear_module_state->__pyx_codeobj__478);
  Py_CLEAR(clear_module_state->__pyx_codeobj__479);
  Py_CLEAR(clear_module_state->__pyx_codeobj__480);
  Py_CLEAR(clear_module_state->__pyx_codeobj__481);
  Py_CLEAR(clear_module_state->__pyx_codeobj__482);
  Py_CLEAR(clear_module_state->__pyx_codeobj__483);
  Py_CLEAR(clear_module_state->__pyx_codeobj__484);
  Py_CLEAR(clear_module_state->__pyx_codeobj__485);
  Py_CLEAR(clear_module_state->__pyx_codeobj__486);
  Py_CLEAR(clear_module_state->__pyx_codeobj__487);
  Py_CLEAR(clear_module_state->__pyx_codeobj__488);
  Py_CLEAR(clear_module_state->__pyx_codeobj__489);
  Py_CLEAR(clear_module_state->__pyx_codeobj__490);
  Py_CLEAR(clear_module_state->__pyx_codeobj__491);
  Py_CLEAR(clear_module_state->__pyx_codeobj__492);
  Py_CLEAR(clear_module_state->__pyx_codeobj__493);
  Py_CLEAR(clear_module_state->__pyx_codeobj__494);
  Py_CLEAR(clear_module_state->__pyx_codeobj__495);
  Py_CLEAR(clear_module_state->__pyx_codeobj__496);
  Py_CLEAR(clear_module_state->__pyx_codeobj__497);
  Py_CLEAR(clear_module_state->__pyx_codeobj__498);
  Py_CLEAR(clear_module_state->__pyx_codeobj__499);
  Py_CLEAR(clear_module_state->__pyx_codeobj__500);
  Py_CLEAR(clear_module_state->__pyx_codeobj__501);
  Py_CLEAR(clear_module_state->__pyx_codeobj__502);
  Py_CLEAR(clear_module_state->__pyx_codeobj__503);
  Py_CLEAR(clear_module_state->__pyx_codeobj__504);
  Py_CLEAR(clear_module_state->__pyx_codeobj__505);
  Py_CLEAR(clear_module_state->__pyx_codeobj__506);
  Py_CLEAR(clear_module_state->__pyx_codeobj__507);
  Py_CLEAR(clear_module_state->__pyx_codeobj__508);
  Py_CLEAR(clear_module_state->__pyx_codeobj__509);
  Py_CLEAR(clear_module_state->__pyx_codeobj__510);
  Py_CLEAR(clear_module_state->__pyx_codeobj__511);
  Py_CLEAR(clear_module_state->__pyx_codeobj__512);
  Py_CLEAR(clear_module_state->__pyx_codeobj__514);
  Py_CLEAR(clear_module_state->__pyx_codeobj__515);
  Py_CLEAR(clear_module_state->__pyx_codeobj__516);
  Py_CLEAR(clear_module_state->__pyx_codeobj__518);
  Py_CLEAR(clear_module_state->__pyx_codeobj__520);
  Py_CLEAR(clear_module_state->__pyx_codeobj__521);
  Py_CLEAR(clear_module_state->__pyx_codeobj__522);
  Py_CLEAR(clear_module_state->__pyx_codeobj__523);
  Py_CLEAR(clear_module_state->__pyx_codeobj__524);
  Py_CLEAR(clear_module_state->__pyx_codeobj__525);
  Py_CLEAR(clear_module_state->__pyx_codeobj__526);
  Py_CLEAR(clear_module_state->__pyx_codeobj__527);
  Py_CLEAR(clear_module_state->__pyx_codeobj__529);
  Py_CLEAR(clear_module_state->__pyx_codeobj__530);
  Py_CLEAR(clear_module_state->__pyx_codeobj__531);
  Py_CLEAR(clear_module_state->__pyx_codeobj__532);
  Py_CLEAR(clear_module_state->__pyx_codeobj__535);
  Py_CLEAR(clear_module_state->__pyx_codeobj__536);
  Py_CLEAR(clear_module_state->__pyx_codeobj__537);
  Py_CLEAR(clear_module_state->__pyx_codeobj__538);
  Py_CLEAR(clear_module_state->__pyx_codeobj__540);
  Py_CLEAR(clear_module_state->__pyx_codeobj__541);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVHandle);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVSocketHandle);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVSocketHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVAsync);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVAsync);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVTimer);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVTimer);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVIdle);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVIdle);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVBaseTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVBaseTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_Loop);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_Loop);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_Handle);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_Handle);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_TimerHandle);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_TimerHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVCheck);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVCheck);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVPoll);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVPoll);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVStream);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVStreamServer);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVStreamServer);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_TCPServer);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_TCPServer);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_TCPTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_TCPTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UnixServer);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UnixServer);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UnixTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UnixTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_ReadUnixTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_ReadUnixTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_WriteUnixTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_WriteUnixTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVProcess);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVProcess);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVProcessTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVProcessTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVFSEvent);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVFSEvent);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UVRequest);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UVRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop__SSLProtocolTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_SSLProtocol);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_SSLProtocol);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_UDPTransport);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_UDPTransport);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_Server);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_Server);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_PseudoSocket);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_PseudoSocket);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_LruCache);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_LruCache);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop__StreamWriteContext);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop__StreamWriteContext);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop__TCPConnectRequest);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop__TCPConnectRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop__PipeConnectRequest);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop__PipeConnectRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_SockAddrHolder);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_SockAddrHolder);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_AddrInfo);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_AddrInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_AddrInfoRequest);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_AddrInfoRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop_NameInfoRequest);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop_NameInfoRequest);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop__UDPSendContext);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop__UDPSendContext);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_2___get__);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter__);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit__);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever);
  Py_VISIT(traverse_module_state->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1);
  Py_VISIT(traverse_module_state->__pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1);
  Py_VISIT(traverse_module_state->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled);
  Py_VISIT(traverse_module_state->__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled);
  Py_VISIT(traverse_module_state->__pyx_kp_b_1);
  Py_VISIT(traverse_module_state->__pyx_kp_u_18_7_7_7);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AF_INET6_address_must_be_tuple);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AF_INET6_must_be_a_tuple_of_2_4);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AF_INET_address_must_be_tuple);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AF_INET_address_must_be_tuple_of);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AF_UNIX_address_must_be_a_str_or);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AF_UNSPEC_in_DNS_results);
  Py_VISIT(traverse_module_state->__pyx_n_u_AI_CANONNAME);
  Py_VISIT(traverse_module_state->__pyx_kp_u_A_Stream_Socket_was_expected_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_A_UDP_Socket_was_expected_got_r);
  Py_VISIT(traverse_module_state->__pyx_kp_u_A_UNIX_Domain_Stream_Socket_was);
  Py_VISIT(traverse_module_state->__pyx_kp_u_A_callable_object_or_None_is_exp);
  Py_VISIT(traverse_module_state->__pyx_n_s_AbstractChildWatcher);
  Py_VISIT(traverse_module_state->__pyx_n_s_AddrInfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_AddrInfoRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_AddrInfoRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_AddrInfoRequest___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AddrInfoRequest_callback_called);
  Py_VISIT(traverse_module_state->__pyx_n_s_AddrInfo___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_AddrInfo___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_AddrInfo_data_is_NULL);
  Py_VISIT(traverse_module_state->__pyx_n_s_AddressFamily);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Address_r_is_already_in_use);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_AttributeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Bad_exception_data_from_child_r);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseException);
  Py_VISIT(traverse_module_state->__pyx_n_s_BaseProtocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_BlockingIOError);
  Py_VISIT(traverse_module_state->__pyx_n_s_BrokenPipeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CHANGE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Callback_handles_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_CancelledError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_call_write_after_write_eo);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_call_writelines_after_wri);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_close_a_running_event_loo);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_run_the_event_loop_while);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Cannot_serialize_socket_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_CertificateError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Connect_call_failed);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConnectionAbortedError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConnectionRefusedError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConnectionResetError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Counter);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Creating__SSLProtocolTransport_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEVNULL);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeprecationWarning);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_ADDRFAMILY);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_AGAIN);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_BADFLAGS);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_BADHINTS);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_CANCELED);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_FAIL);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_FAMILY);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_MEMORY);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_NODATA);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_NONAME);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_OVERFLOW);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_PROTOCOL);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_SERVICE);
  Py_VISIT(traverse_module_state->__pyx_n_u_EAI_SOCKTYPE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_calling_eof_received);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_occurred_during_shutdown);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Errors_logged);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Event_loop_is_closed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Event_loop_stopped_before_Future);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Exception_in___repr___r_value_ty);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Exception_in_callback);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Exception_in_default_exception_h);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Exception_in_default_exception_h_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Exception_occurred_in_preexec_fn);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Executing_r_took_3f_seconds);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Executing_s_took_3f_seconds);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Executor_shutdown_has_been_calle);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Fatal_error_on_SSL_protocol);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Fatal_error_on_server);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Fatal_error_on_transport);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Fatal_error_on_transport_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Fatal_write_error_on_datagram_tr);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileExistsError);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileNotFoundError);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileSystemEvent);
  Py_VISIT(traverse_module_state->__pyx_kp_u_File_descriptor_r_is_used_by_tra);
  Py_VISIT(traverse_module_state->__pyx_n_s_FlowControlMixin);
  Py_VISIT(traverse_module_state->__pyx_n_s_Future);
  Py_VISIT(traverse_module_state->__pyx_n_s_Handle);
  Py_VISIT(traverse_module_state->__pyx_n_s_Handle___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Handle___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Handle_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_Handle_cancelled);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Handle_loop_is_None_in_Handle);
  Py_VISIT(traverse_module_state->__pyx_n_u_IPV6_V6ONLY);
  Py_VISIT(traverse_module_state->__pyx_kp_u_IPv4_sockaddr_must_be_2_tuple);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntEnum);
  Py_VISIT(traverse_module_state->__pyx_n_s_InterruptedError);
  Py_VISIT(traverse_module_state->__pyx_n_s_InvalidStateError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_address_must_be_None_or);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_file_descriptor);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Invalid_file_object_r);
  Py_VISIT(traverse_module_state->__pyx_n_s_IsADirectoryError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Iterable);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyboardInterrupt);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Listen_errors);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop___get___locals_lambda);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop___sighandler);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop___subprocess_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__asyncgen_finalizer_hook);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__asyncgen_firstiter_hook);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__check_default_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__do_shutdown);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__get_backend_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__getaddrinfo_locals_callbac);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__getnameinfo_locals_callbac);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__monitor_fs);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop__read_from_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__sighandler);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop__sock_accept);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop__sock_connect);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop__sock_recv);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop__sock_recv_into);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop__sock_sendall);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop__stop);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop__subprocess_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_add_reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_add_signal_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_add_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_call_at);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_call_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_call_later);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_call_soon);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_call_soon_threadsafe);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_connect_accepted_socket);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_connect_read_pipe);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_connect_write_pipe);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_create_connection);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_create_datagram_endpoint);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_create_future);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_create_server);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_create_task);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_create_unix_connection);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_create_unix_server);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop_debug_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_default_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_get_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_get_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_get_task_factory);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_getaddrinfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_getnameinfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_is_closed);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_is_running);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_remove_reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_remove_signal_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_remove_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_run_forever);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_run_in_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_run_until_complete);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_set_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_set_default_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_set_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_set_task_factory);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_shutdown_asyncgens);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_shutdown_default_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_accept);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_connect);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_recv);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_recv_into);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_recvfrom);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_recvfrom_into);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_sendall);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_sock_sendto);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_start_tls);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_stop);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_subprocess_exec);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_subprocess_shell);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Loop_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_Loop_time_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_LruCache);
  Py_VISIT(traverse_module_state->__pyx_n_s_LruCache___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_LruCache___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryBIO);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Multiple_exceptions);
  Py_VISIT(traverse_module_state->__pyx_n_s_NSIG);
  Py_VISIT(traverse_module_state->__pyx_n_s_NameInfoRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_NameInfoRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_NameInfoRequest___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Neither_host_port_nor_sock_were);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Non_thread_safe_operation_invoke);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplementedError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Number_of_signals);
  Py_VISIT(traverse_module_state->__pyx_n_s_OSError);
  Py_VISIT(traverse_module_state->__pyx_n_s_O_RDWR);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Object_created_at_most_recent_ca);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderedDict);
  Py_VISIT(traverse_module_state->__pyx_n_s_OverflowError);
  Py_VISIT(traverse_module_state->__pyx_n_s_PIPE);
  Py_VISIT(traverse_module_state->__pyx_n_u_PYTHONASYNCIODEBUG);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Passing_reuse_address_True_is_no);
  Py_VISIT(traverse_module_state->__pyx_n_s_PermissionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Pickling_of_struct_members_such);
  Py_VISIT(traverse_module_state->__pyx_n_s_PipeConnectRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_PipeConnectRequest___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_PipeConnectRequest___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Polls_debug_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_ProcessLookupError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Process_info);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Process_memory);
  Py_VISIT(traverse_module_state->__pyx_n_s_Protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket___enter);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket___exit);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket___getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_accept);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_bind);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_connect);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_connect_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_detach);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_dup);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_fileno);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_get_inheritable);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_getpeername);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_getsockname);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_getsockopt);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_gettimeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_ioctl);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_listen);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_makefile);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_recv);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_recv_into);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_recvfrom);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_recvfrom_into);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_recvmsg);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_recvmsg_into);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_sendall);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_sendfile);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_sendmsg);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_sendmsg_afalg);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_sendto);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_set_inheritable);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_setblocking);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_setsockopt);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_settimeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_share);
  Py_VISIT(traverse_module_state->__pyx_n_s_PseudoSocket_shutdown);
  Py_VISIT(traverse_module_state->__pyx_n_s_Pyx_CFunc_6uvloop_4loop_4Loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_RENAME);
  Py_VISIT(traverse_module_state->__pyx_n_s_RENAME_CHANGE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Racing_with_another_loop_to_spaw);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadSubprocessPipeProto);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadSubprocessPipeProto_data_rec);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_abort);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_can_write_eof);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_get_write_buff);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_get_write_buff_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_set_write_buff);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_write_eof);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadUnixTransport_writelines);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Read_EOF_failed_callbacks);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Read_EOFs);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Read_callbacks);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Read_callbacks_failed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Read_errors);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Read_events);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Read_failed_callbacks);
  Py_VISIT(traverse_module_state->__pyx_n_s_ResourceWarning);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeWarning);
  Py_VISIT(traverse_module_state->__pyx_n_s_SIG_DFL);
  Py_VISIT(traverse_module_state->__pyx_n_u_SOCK_NONBLOCK);
  Py_VISIT(traverse_module_state->__pyx_n_u_SO_BROADCAST);
  Py_VISIT(traverse_module_state->__pyx_n_u_SO_REUSEPORT);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLContext);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport___reduce_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport__force_clo);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport__test__app);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_abort);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_can_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_get_extra);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_get_protoc);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_get_read_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_get_read_b_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_get_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_get_write_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_is_closing);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_is_reading);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_pause_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_resume_rea);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_set_protoc);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_set_read_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_set_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_write_eof);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocolTransport_writelines);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SSLProtocol__do_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol__get_app_transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol__start_handshake_loc);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol__start_shutdown_loca);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol_buffer_updated);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol_connection_lost);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol_connection_made);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol_eof_received);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol_get_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol_pause_writing);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLProtocol_resume_writing);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLSyscallError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLWantReadError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSLZeroReturnError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSL_ERROR_SYSCALL);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSL_ERROR_WANT_READ);
  Py_VISIT(traverse_module_state->__pyx_n_s_SSL_ERROR_WANT_WRITE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SSL_connection_is_closed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SSL_handshake_failed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SSL_handshake_failed_on_verifyin);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SSL_handshake_is_taking_longer_t);
  Py_VISIT(traverse_module_state->__pyx_kp_u_SSL_shutdown_timed_out);
  Py_VISIT(traverse_module_state->__pyx_n_s_STDOUT);
  Py_VISIT(traverse_module_state->__pyx_n_s_S_ISSOCK);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server___aenter);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server___aexit);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server_get_loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server_is_serving);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server_serve_forever);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Server_side_SSL_needs_a_valid_SS);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server_start_serving);
  Py_VISIT(traverse_module_state->__pyx_n_s_Server_wait_closed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Shutdown_errors);
  Py_VISIT(traverse_module_state->__pyx_n_s_SockAddrHolder);
  Py_VISIT(traverse_module_state->__pyx_n_s_SockAddrHolder___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SockAddrHolder___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Sock_ops_successful_on_1st_try);
  Py_VISIT(traverse_module_state->__pyx_n_s_SocketKind);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Socket_try_writes);
  Py_VISIT(traverse_module_state->__pyx_n_s_StackSummary);
  Py_VISIT(traverse_module_state->__pyx_n_s_StreamWriteContext);
  Py_VISIT(traverse_module_state->__pyx_n_s_StreamWriteContext___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_StreamWriteContext___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_kp_u_StreamWriteContext_close_py_buf);
  Py_VISIT(traverse_module_state->__pyx_kp_u_StreamWriteContext_close_uv_buf);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Streams_debug_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubprocessError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketReaderFuture);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketReaderFuture___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketReaderFuture___remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketReaderFuture__loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketReaderFuture__remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketReaderFuture__sock);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketReaderFuture_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketWriterFuture);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketWriterFuture___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketWriterFuture___remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketWriterFuture__loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketWriterFuture__remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketWriterFuture__sock);
  Py_VISIT(traverse_module_state->__pyx_n_s_SyncSocketWriterFuture_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_SystemExit);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPConnectRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPConnectRequest___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPConnectRequest___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPServer);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPServer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPServer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPTransport___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPTransport___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TCPTransport_get_extra_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_Task);
  Py_VISIT(traverse_module_state->__pyx_kp_u_The_executor_did_not_finishing_j);
  Py_VISIT(traverse_module_state->__pyx_kp_u_The_reuse_address_parameter_has);
  Py_VISIT(traverse_module_state->__pyx_n_s_Thread);
  Py_VISIT(traverse_module_state->__pyx_n_s_ThreadPoolExecutor);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimeoutError);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimerHandle);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimerHandle___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimerHandle___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimerHandle_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimerHandle_cancelled);
  Py_VISIT(traverse_module_state->__pyx_n_s_TimerHandle_when);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Timer_handles_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_Transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_UDPSendContext);
  Py_VISIT(traverse_module_state->__pyx_n_s_UDPSendContext___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UDPSendContext___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_UDPTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_UDPTransport___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UDPTransport___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UDPTransport_family_is_undefined);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UDPTransport_family_is_undefined_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_UDPTransport_sendto);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVAsync);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVAsync___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVAsync___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport___setstate_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport__force_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_abort);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_get_extra_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_get_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_get_write_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_get_write_buffer_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_is_closing);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_set_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVBaseTransport_set_write_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVCheck);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVCheck___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVCheck___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVFSEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVFSEvent___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVFSEvent___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVFSEvent_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVFSEvent_cancelled);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVHandle);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVHandle___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVHandle___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVHandles_python_libuv_total);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVIdle);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVIdle___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVIdle___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVPoll);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVPoll___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVPoll___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcess);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport___stdio_inite);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVProcessTransport__call_connect);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport__stdio_inite);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport__wait);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_get_extra_inf);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_get_pid);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_get_pipe_tran);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_get_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_get_returncod);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_is_closing);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_kill);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_send_signal);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_set_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcessTransport_terminate);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcess___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVProcess___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVProcess__close_after_spawn_cal);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVRequest);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVRequest___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVRequest___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVSocketHandle);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVSocketHandle___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVSocketHandle___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStreamServer);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStreamServer___on_ssl_connecte);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStreamServer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStreamServer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStreamServer__on_ssl_connecte);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVStream__try_write_called_with);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream_can_write_eof);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream_is_reading);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream_pause_reading);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream_resume_reading);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVStream_shutdown_callback_calle);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream_write);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVStream_write_callback_called_w);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream_write_eof);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVStream_writelines);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVTimer);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVTimer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UVTimer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVTransport__call_connection_los);
  Py_VISIT(traverse_module_state->__pyx_kp_u_UVTransport__call_connection_mad);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unable_to_check_or_remove_stale);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unable_to_clean_up_listening_UNI);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unhandled_error_in_exception_han);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Unhandled_exception_in_event_loo);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnicodeEncodeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnixServer);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnixServer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnixServer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnixTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnixTransport___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnixTransport___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s_WeakSet);
  Py_VISIT(traverse_module_state->__pyx_n_s_WeakValueDictionary);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteSubprocessPipeProto);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteSubprocessPipeProto___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteSubprocessPipeProto___repr);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteSubprocessPipeProto_connect);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteSubprocessPipeProto_connect_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteSubprocessPipeProto_pause_w);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteSubprocessPipeProto_resume);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteUnixTransport);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteUnixTransport___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteUnixTransport___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteUnixTransport_pause_reading);
  Py_VISIT(traverse_module_state->__pyx_n_s_WriteUnixTransport_resume_readin);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Write_callbacks_failed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Write_contexts_8);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Write_errors);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Write_events);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Write_failed_callbacks);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Write_without_poll);
  Py_VISIT(traverse_module_state->__pyx_kp_u_XXX_an_allocated_buffer_in_trans);
  Py_VISIT(traverse_module_state->__pyx_kp_u_You_must_set_server_hostname_whe);
  Py_VISIT(traverse_module_state->__pyx_n_s__10);
  Py_VISIT(traverse_module_state->__pyx_kp_b__10);
  Py_VISIT(traverse_module_state->__pyx_kp_u__10);
  Py_VISIT(traverse_module_state->__pyx_kp_u__115);
  Py_VISIT(traverse_module_state->__pyx_kp_u__128);
  Py_VISIT(traverse_module_state->__pyx_kp_u__129);
  Py_VISIT(traverse_module_state->__pyx_kp_u__130);
  Py_VISIT(traverse_module_state->__pyx_kp_u__140);
  Py_VISIT(traverse_module_state->__pyx_kp_u__141);
  Py_VISIT(traverse_module_state->__pyx_kp_b__161);
  Py_VISIT(traverse_module_state->__pyx_kp_b__170);
  Py_VISIT(traverse_module_state->__pyx_kp_u__196);
  Py_VISIT(traverse_module_state->__pyx_n_s__198);
  Py_VISIT(traverse_module_state->__pyx_n_s__263);
  Py_VISIT(traverse_module_state->__pyx_kp_u__33);
  Py_VISIT(traverse_module_state->__pyx_n_s__542);
  Py_VISIT(traverse_module_state->__pyx_kp_u__61);
  Py_VISIT(traverse_module_state->__pyx_kp_u__69);
  Py_VISIT(traverse_module_state->__pyx_kp_u__75);
  Py_VISIT(traverse_module_state->__pyx_kp_u__78);
  Py_VISIT(traverse_module_state->__pyx_kp_u__79);
  Py_VISIT(traverse_module_state->__pyx_n_s_abc);
  Py_VISIT(traverse_module_state->__pyx_n_s_abort);
  Py_VISIT(traverse_module_state->__pyx_kp_u_abort_init__closed_is_set);
  Py_VISIT(traverse_module_state->__pyx_kp_u_abort_init__inited_is_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_accept);
  Py_VISIT(traverse_module_state->__pyx_kp_u_accept_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_aclose);
  Py_VISIT(traverse_module_state->__pyx_kp_u_active_TimerHandle_is_deallacati);
  Py_VISIT(traverse_module_state->__pyx_n_s_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_done_callback);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_signal_handler);
  Py_VISIT(traverse_module_state->__pyx_kp_u_add_signal_handler_can_only_be_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_addr);
  Py_VISIT(traverse_module_state->__pyx_n_s_address);
  Py_VISIT(traverse_module_state->__pyx_n_s_addrinfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_addrs);
  Py_VISIT(traverse_module_state->__pyx_n_s_aenter);
  Py_VISIT(traverse_module_state->__pyx_n_s_aexit);
  Py_VISIT(traverse_module_state->__pyx_n_s_ag);
  Py_VISIT(traverse_module_state->__pyx_n_s_agen);
  Py_VISIT(traverse_module_state->__pyx_n_s_ai);
  Py_VISIT(traverse_module_state->__pyx_n_s_ai_cnt);
  Py_VISIT(traverse_module_state->__pyx_n_s_ai_local);
  Py_VISIT(traverse_module_state->__pyx_n_s_ai_remote);
  Py_VISIT(traverse_module_state->__pyx_kp_u_alive_closed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_all_args_must_be_str_or_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_all_environment_values_must_be_b);
  Py_VISIT(traverse_module_state->__pyx_kp_u_all_environment_vars_must_be_byt);
  Py_VISIT(traverse_module_state->__pyx_n_s_allow_broadcast);
  Py_VISIT(traverse_module_state->__pyx_kp_u_an_error_occurred_during_closing);
  Py_VISIT(traverse_module_state->__pyx_n_s_app_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_app_transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_u_ascii);
  Py_VISIT(traverse_module_state->__pyx_n_u_asyncgen);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncgen_finalizer_hook);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncgen_firstiter_hook);
  Py_VISIT(traverse_module_state->__pyx_kp_u_asynchronous_generator_r_was_sch);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio);
  Py_VISIT(traverse_module_state->__pyx_kp_s_asyncio_Handle);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_base_events);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_futures);
  Py_VISIT(traverse_module_state->__pyx_kp_u_asyncio_is_trying_to_install_it);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_log);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_sslproto);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_tasks);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_transports);
  Py_VISIT(traverse_module_state->__pyx_n_s_await);
  Py_VISIT(traverse_module_state->__pyx_n_u_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_backlog);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_events);
  Py_VISIT(traverse_module_state->__pyx_kp_b_bin_sh);
  Py_VISIT(traverse_module_state->__pyx_n_s_bind);
  Py_VISIT(traverse_module_state->__pyx_kp_u_bind_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_buf);
  Py_VISIT(traverse_module_state->__pyx_kp_u_buffer_size_is_not_0_after_a_su);
  Py_VISIT(traverse_module_state->__pyx_n_s_buffer_updated);
  Py_VISIT(traverse_module_state->__pyx_n_s_bufs);
  Py_VISIT(traverse_module_state->__pyx_n_s_bufsize);
  Py_VISIT(traverse_module_state->__pyx_kp_u_bufsize_must_be_0);
  Py_VISIT(traverse_module_state->__pyx_n_s_builtins);
  Py_VISIT(traverse_module_state->__pyx_kp_b_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_str_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_at);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_connection_made);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_later);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_soon);
  Py_VISIT(traverse_module_state->__pyx_n_s_call_soon_threadsafe);
  Py_VISIT(traverse_module_state->__pyx_n_s_callback);
  Py_VISIT(traverse_module_state->__pyx_kp_u_called_after_destroying_the_UVH);
  Py_VISIT(traverse_module_state->__pyx_kp_u_called_with_handle_data_NULL);
  Py_VISIT(traverse_module_state->__pyx_n_s_can_write_eof);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancel);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cancel_request_data_is_NULL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cancel_request_data_is_not_UVRe);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cancel_request_is_NULL);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancelled);
  Py_VISIT(traverse_module_state->__pyx_n_u_cancelled);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_add_a_signal_handler_for);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_apply_subprocess_STDOUT);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_cleanup_signal_handlers_c);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_convert_sockaddr_into_Pyt);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_open_a_UDP_handle_invalid);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_pause_signals_handling_no);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_run_Handle_callback_is_no);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_run_TimerHandle_callback);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_shutdown_signals_handling);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_spawn_a_process_args_are);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cannot_switch_state_from_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_cast);
  Py_VISIT(traverse_module_state->__pyx_n_s_cause);
  Py_VISIT(traverse_module_state->__pyx_n_s_cfunc_to_py);
  Py_VISIT(traverse_module_state->__pyx_n_s_chain);
  Py_VISIT(traverse_module_state->__pyx_n_s_check_default_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_check_hostname);
  Py_VISIT(traverse_module_state->__pyx_n_s_cipher);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cleanup_socket);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_kp_u_close__handle_data_is_NULL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_close__handle_data_is_not_UVHa);
  Py_VISIT(traverse_module_state->__pyx_n_s_close_fds);
  Py_VISIT(traverse_module_state->__pyx_kp_u_close_method);
  Py_VISIT(traverse_module_state->__pyx_kp_u_close_uv_is_closing_is_true);
  Py_VISIT(traverse_module_state->__pyx_kp_u_closed_Transport_handle_and_unse);
  Py_VISIT(traverse_module_state->__pyx_kp_u_closed_reading_x);
  Py_VISIT(traverse_module_state->__pyx_kp_u_closed_x);
  Py_VISIT(traverse_module_state->__pyx_n_s_closing_agens);
  Py_VISIT(traverse_module_state->__pyx_n_s_cmd);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections_abc);
  Py_VISIT(traverse_module_state->__pyx_n_s_completed);
  Py_VISIT(traverse_module_state->__pyx_n_s_compression);
  Py_VISIT(traverse_module_state->__pyx_n_s_concurrent);
  Py_VISIT(traverse_module_state->__pyx_kp_u_concurrent_allocations);
  Py_VISIT(traverse_module_state->__pyx_n_s_concurrent_futures);
  Py_VISIT(traverse_module_state->__pyx_n_s_conmade_cb);
  Py_VISIT(traverse_module_state->__pyx_n_s_connect);
  Py_VISIT(traverse_module_state->__pyx_n_s_connect_accepted_socket);
  Py_VISIT(traverse_module_state->__pyx_n_s_connect_ex);
  Py_VISIT(traverse_module_state->__pyx_kp_u_connect_ex_method);
  Py_VISIT(traverse_module_state->__pyx_kp_u_connect_failed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_connect_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_connect_read_pipe);
  Py_VISIT(traverse_module_state->__pyx_n_s_connect_write_pipe);
  Py_VISIT(traverse_module_state->__pyx_n_s_connection_lost);
  Py_VISIT(traverse_module_state->__pyx_n_s_connection_made);
  Py_VISIT(traverse_module_state->__pyx_n_s_context);
  Py_VISIT(traverse_module_state->__pyx_n_u_context);
  Py_VISIT(traverse_module_state->__pyx_kp_u_context_manager_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_coro);
  Py_VISIT(traverse_module_state->__pyx_n_s_coroutines);
  Py_VISIT(traverse_module_state->__pyx_kp_u_coroutines_cannot_be_used_with_a);
  Py_VISIT(traverse_module_state->__pyx_kp_u_coroutines_cannot_be_used_with_r);
  Py_VISIT(traverse_module_state->__pyx_kp_u_could_not_bind_to_local_addr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_could_not_bind_to_remote_addr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_could_not_close_attached_file_ob);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_connection);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_connection_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_datagram_endpoint);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_datagram_endpoint_locals);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_default_context);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_future);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_server);
  Py_VISIT(traverse_module_state->__pyx_kp_u_create_server_failed_to_create_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_task);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_unix_connection);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_unix_server);
  Py_VISIT(traverse_module_state->__pyx_kp_u_created_at);
  Py_VISIT(traverse_module_state->__pyx_n_s_creationflags);
  Py_VISIT(traverse_module_state->__pyx_kp_u_creationflags_is_not_supported);
  Py_VISIT(traverse_module_state->__pyx_n_s_ctx);
  Py_VISIT(traverse_module_state->__pyx_n_s_cwd);
  Py_VISIT(traverse_module_state->__pyx_kp_u_cwd_must_be_a_str_or_bytes_objec);
  Py_VISIT(traverse_module_state->__pyx_n_u_cygwin);
  Py_VISIT(traverse_module_state->__pyx_n_u_darwin);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_kp_u_data_expecting_a_bytes_like_inst);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_received);
  Py_VISIT(traverse_module_state->__pyx_n_s_datagram_received);
  Py_VISIT(traverse_module_state->__pyx_kp_u_dealloc____handle_is_NULL__cl);
  Py_VISIT(traverse_module_state->__pyx_kp_u_deallocating_a_running_event_loo);
  Py_VISIT(traverse_module_state->__pyx_kp_u_deallocating_an_open_event_loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_debug_flags);
  Py_VISIT(traverse_module_state->__pyx_n_u_debug_wrapper);
  Py_VISIT(traverse_module_state->__pyx_n_s_decode);
  Py_VISIT(traverse_module_state->__pyx_n_s_decref_socketios);
  Py_VISIT(traverse_module_state->__pyx_n_s_default);
  Py_VISIT(traverse_module_state->__pyx_n_s_default_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_default_int_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_delay);
  Py_VISIT(traverse_module_state->__pyx_n_s_deque);
  Py_VISIT(traverse_module_state->__pyx_n_s_detach);
  Py_VISIT(traverse_module_state->__pyx_kp_u_detach_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_dev_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_devnull);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_disable);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_discard);
  Py_VISIT(traverse_module_state->__pyx_n_s_disconnected);
  Py_VISIT(traverse_module_state->__pyx_n_s_do_handshake);
  Py_VISIT(traverse_module_state->__pyx_n_s_do_shutdown);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_have_a_reader_in_the);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_have_a_reader_in_the_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_have_a_writer_in_the);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_have_a_writer_in_the_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_done);
  Py_VISIT(traverse_module_state->__pyx_n_s_done_cb);
  Py_VISIT(traverse_module_state->__pyx_n_s_dup);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_enabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_ensure_future);
  Py_VISIT(traverse_module_state->__pyx_n_s_enter);
  Py_VISIT(traverse_module_state->__pyx_n_s_enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_env);
  Py_VISIT(traverse_module_state->__pyx_n_s_environ);
  Py_VISIT(traverse_module_state->__pyx_n_s_eof_received);
  Py_VISIT(traverse_module_state->__pyx_n_s_err);
  Py_VISIT(traverse_module_state->__pyx_n_s_errno);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_error_received);
  Py_VISIT(traverse_module_state->__pyx_kp_u_error_status_in_uv_stream_t_list);
  Py_VISIT(traverse_module_state->__pyx_kp_u_error_status_in_uv_stream_t_read);
  Py_VISIT(traverse_module_state->__pyx_kp_u_error_status_in_uv_stream_t_shut);
  Py_VISIT(traverse_module_state->__pyx_kp_u_error_status_in_uv_stream_t_writ);
  Py_VISIT(traverse_module_state->__pyx_kp_u_error_while_attempting_to_bind_o);
  Py_VISIT(traverse_module_state->__pyx_n_s_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_exception);
  Py_VISIT(traverse_module_state->__pyx_n_u_exception);
  Py_VISIT(traverse_module_state->__pyx_kp_u_exception_in_Task___repr);
  Py_VISIT(traverse_module_state->__pyx_n_s_exceptions);
  Py_VISIT(traverse_module_state->__pyx_n_s_excs);
  Py_VISIT(traverse_module_state->__pyx_n_s_executable);
  Py_VISIT(traverse_module_state->__pyx_n_s_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_exit);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expected_AF_INET_AF_INET6_or_AF);
  Py_VISIT(traverse_module_state->__pyx_kp_u_expected_asyncio_Future);
  Py_VISIT(traverse_module_state->__pyx_n_s_extract);
  Py_VISIT(traverse_module_state->__pyx_n_s_f1);
  Py_VISIT(traverse_module_state->__pyx_n_s_f2);
  Py_VISIT(traverse_module_state->__pyx_n_s_factory);
  Py_VISIT(traverse_module_state->__pyx_n_s_family);
  Py_VISIT(traverse_module_state->__pyx_kp_u_fatal_Could_not_advance__StreamW);
  Py_VISIT(traverse_module_state->__pyx_kp_u_fatal_sent_0_in_advance_uv_buf);
  Py_VISIT(traverse_module_state->__pyx_n_s_fd);
  Py_VISIT(traverse_module_state->__pyx_kp_u_fd_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_fileno);
  Py_VISIT(traverse_module_state->__pyx_n_s_fileobj);
  Py_VISIT(traverse_module_state->__pyx_n_s_finalizer);
  Py_VISIT(traverse_module_state->__pyx_n_s_firstiter);
  Py_VISIT(traverse_module_state->__pyx_n_s_flag);
  Py_VISIT(traverse_module_state->__pyx_n_s_flags);
  Py_VISIT(traverse_module_state->__pyx_n_s_flowinfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_flush);
  Py_VISIT(traverse_module_state->__pyx_n_s_force_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_format_list);
  Py_VISIT(traverse_module_state->__pyx_n_s_frame);
  Py_VISIT(traverse_module_state->__pyx_n_s_fs);
  Py_VISIT(traverse_module_state->__pyx_n_s_fs_handle);
  Py_VISIT(traverse_module_state->__pyx_n_s_fspath);
  Py_VISIT(traverse_module_state->__pyx_n_s_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_functools);
  Py_VISIT(traverse_module_state->__pyx_n_s_fut);
  Py_VISIT(traverse_module_state->__pyx_n_s_future);
  Py_VISIT(traverse_module_state->__pyx_n_s_futures);
  Py_VISIT(traverse_module_state->__pyx_n_s_gaierror);
  Py_VISIT(traverse_module_state->__pyx_n_s_gather);
  Py_VISIT(traverse_module_state->__pyx_n_s_gc);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_app_transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_asyncgen_hooks);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_backend_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_u_get_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_coroutine_origin_tracking_de);
  Py_VISIT(traverse_module_state->__pyx_n_u_get_coroutine_wrapper);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_event_loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_extra_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_inheritable);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_pid);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_pipe_transport);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_read_buffer_limits);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_read_buffer_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_returncode);
  Py_VISIT(traverse_module_state->__pyx_n_u_get_running_loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_task_factory);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_write_buffer_limits);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_write_buffer_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_getaddrinfo);
  Py_VISIT(traverse_module_state->__pyx_kp_u_getaddrinfo_returned_empty_list);
  Py_VISIT(traverse_module_state->__pyx_kp_u_getaddrinfo_returned_empty_list_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_getfilesystemencoding);
  Py_VISIT(traverse_module_state->__pyx_n_s_getframe);
  Py_VISIT(traverse_module_state->__pyx_n_s_getnameinfo);
  Py_VISIT(traverse_module_state->__pyx_kp_u_getnameinfo_argument_1_must_be_a);
  Py_VISIT(traverse_module_state->__pyx_kp_u_getnameinfo_flowinfo_must_be_0_1);
  Py_VISIT(traverse_module_state->__pyx_n_s_getpeercert);
  Py_VISIT(traverse_module_state->__pyx_n_s_getpeername);
  Py_VISIT(traverse_module_state->__pyx_n_s_getservbyname);
  Py_VISIT(traverse_module_state->__pyx_kp_u_getsockaddrarg_scope_id_must_be);
  Py_VISIT(traverse_module_state->__pyx_n_s_getsockname);
  Py_VISIT(traverse_module_state->__pyx_n_s_getsockopt);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_gettimeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_h);
  Py_VISIT(traverse_module_state->__pyx_n_s_handle);
  Py_VISIT(traverse_module_state->__pyx_n_u_handle);
  Py_VISIT(traverse_module_state->__pyx_kp_u_handle_data_is_NULL_in___close_a);
  Py_VISIT(traverse_module_state->__pyx_kp_u_handle_loop_data_is_NULL_in___en);
  Py_VISIT(traverse_module_state->__pyx_kp_u_handle_loop_is_NULL_in___ensure);
  Py_VISIT(traverse_module_state->__pyx_n_s_handler);
  Py_VISIT(traverse_module_state->__pyx_kp_u_has_handle_0_in___uv_close_handl);
  Py_VISIT(traverse_module_state->__pyx_kp_u_high);
  Py_VISIT(traverse_module_state->__pyx_n_s_high_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_host);
  Py_VISIT(traverse_module_state->__pyx_kp_u_host_and_port_was_not_specified);
  Py_VISIT(traverse_module_state->__pyx_kp_u_host_must_be_a_str_or_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_host_must_be_a_string_or_bytes_o);
  Py_VISIT(traverse_module_state->__pyx_kp_u_host_port_and_sock_can_not_be_sp);
  Py_VISIT(traverse_module_state->__pyx_n_s_hosts);
  Py_VISIT(traverse_module_state->__pyx_n_s_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_ident);
  Py_VISIT(traverse_module_state->__pyx_n_u_idna);
  Py_VISIT(traverse_module_state->__pyx_n_s_ignore_environment);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_u_inf);
  Py_VISIT(traverse_module_state->__pyx_n_s_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_infos);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_inspect);
  Py_VISIT(traverse_module_state->__pyx_kp_u_integer_argument_expected_got);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_Handle_cb_type);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid__init_protocol_call);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_data_in_writebuf_an_inst);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_socket_family_expected_A);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_waiter_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_io_refs);
  Py_VISIT(traverse_module_state->__pyx_n_s_ioctl);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_alive);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_alive_but__handle_data_is_no);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_alive_but__handle_is_NULL);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_alive_but__handle_loop_is_no);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_alive_but__loop_is_None);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_already_being_awaited_on_ser);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_closed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_closed_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_closing);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_longer_than_107_characters);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_supported_by_start_tls);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_not_supposed_to_be_instantia);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_open_in___dealloc___with_loo);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_reading);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_running);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_serving);
  Py_VISIT(traverse_module_state->__pyx_n_s_iscoroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_iscoroutinefunction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_u_isfuture);
  Py_VISIT(traverse_module_state->__pyx_n_s_isgenerator);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_itertools);
  Py_VISIT(traverse_module_state->__pyx_n_s_join);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_kill);
  Py_VISIT(traverse_module_state->__pyx_n_s_kwargs);
  Py_VISIT(traverse_module_state->__pyx_kp_u_laddr_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_lads);
  Py_VISIT(traverse_module_state->__pyx_n_s_lai);
  Py_VISIT(traverse_module_state->__pyx_n_s_lai_addr_static);
  Py_VISIT(traverse_module_state->__pyx_n_s_lai_iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_lai_static);
  Py_VISIT(traverse_module_state->__pyx_n_s_last);
  Py_VISIT(traverse_module_state->__pyx_n_s_libuv_get_loop_t_ptr);
  Py_VISIT(traverse_module_state->__pyx_n_s_libuv_get_version);
  Py_VISIT(traverse_module_state->__pyx_n_s_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_list_of_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_listen);
  Py_VISIT(traverse_module_state->__pyx_kp_u_listen_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_local_addr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_local_addr_must_be_a_tuple_of_ho);
  Py_VISIT(traverse_module_state->__pyx_kp_u_local_addr_must_be_a_tuple_of_ho_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_log);
  Py_VISIT(traverse_module_state->__pyx_n_s_log_destroy_pending);
  Py_VISIT(traverse_module_state->__pyx_n_s_log_lines);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_lookup_lines);
  Py_VISIT(traverse_module_state->__pyx_n_s_loop);
  Py_VISIT(traverse_module_state->__pyx_kp_u_loop__exec_queued_writes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_loop__on_idle);
  Py_VISIT(traverse_module_state->__pyx_n_s_low);
  Py_VISIT(traverse_module_state->__pyx_n_s_lower);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_main_thread);
  Py_VISIT(traverse_module_state->__pyx_n_s_makefile);
  Py_VISIT(traverse_module_state->__pyx_kp_u_makefile_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_maxsize);
  Py_VISIT(traverse_module_state->__pyx_kp_u_maxsize_is_expected_to_be_greate);
  Py_VISIT(traverse_module_state->__pyx_n_s_message);
  Py_VISIT(traverse_module_state->__pyx_n_u_message);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_model);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_monitor_fs);
  Py_VISIT(traverse_module_state->__pyx_n_s_monotonic);
  Py_VISIT(traverse_module_state->__pyx_n_s_move_to_end);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_msg);
  Py_VISIT(traverse_module_state->__pyx_kp_u_must_be_0);
  Py_VISIT(traverse_module_state->__pyx_kp_u_must_be_low);
  Py_VISIT(traverse_module_state->__pyx_n_s_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_u_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_nbytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_kp_u_new_callbacks_were_queued_during);
  Py_VISIT(traverse_module_state->__pyx_kp_u_new_poll_handles_were_queued_dur);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_task);
  Py_VISIT(traverse_module_state->__pyx_kp_u_new_timers_were_queued_during_lo);
  Py_VISIT(traverse_module_state->__pyx_n_s_nexc);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_kp_u_no_path_and_sock_were_specified);
  Py_VISIT(traverse_module_state->__pyx_kp_u_no_python_buffer_is_allocated_in);
  Py_VISIT(traverse_module_state->__pyx_n_s_noop);
  Py_VISIT(traverse_module_state->__pyx_n_s_noop_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_obj);
  Py_VISIT(traverse_module_state->__pyx_n_s_object);
  Py_VISIT(traverse_module_state->__pyx_kp_u_object_created_at_most_recent_ca);
  Py_VISIT(traverse_module_state->__pyx_kp_u_objs_handles);
  Py_VISIT(traverse_module_state->__pyx_n_s_old_agen_hooks);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_ssl_connected);
  Py_VISIT(traverse_module_state->__pyx_n_s_open);
  Py_VISIT(traverse_module_state->__pyx_kp_u_open__StreamWriteContext_is_bein);
  Py_VISIT(traverse_module_state->__pyx_kp_u_open__UDPSendContext_is_being_de);
  Py_VISIT(traverse_module_state->__pyx_n_s_opts);
  Py_VISIT(traverse_module_state->__pyx_n_s_orig_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_os);
  Py_VISIT(traverse_module_state->__pyx_n_s_p_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_partial);
  Py_VISIT(traverse_module_state->__pyx_n_s_pass_fds);
  Py_VISIT(traverse_module_state->__pyx_n_s_path);
  Py_VISIT(traverse_module_state->__pyx_kp_u_path_and_sock_can_not_be_specifi);
  Py_VISIT(traverse_module_state->__pyx_kp_u_path_was_not_specified_and_no_so);
  Py_VISIT(traverse_module_state->__pyx_n_s_pause_reading);
  Py_VISIT(traverse_module_state->__pyx_n_s_pause_writing);
  Py_VISIT(traverse_module_state->__pyx_n_s_peercert);
  Py_VISIT(traverse_module_state->__pyx_n_u_peername);
  Py_VISIT(traverse_module_state->__pyx_n_s_pending);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_pipe);
  Py_VISIT(traverse_module_state->__pyx_n_u_pipe);
  Py_VISIT(traverse_module_state->__pyx_kp_u_pipe_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_pipe_connection_lost);
  Py_VISIT(traverse_module_state->__pyx_n_s_pipe_data_received);
  Py_VISIT(traverse_module_state->__pyx_n_s_platform);
  Py_VISIT(traverse_module_state->__pyx_n_s_pop);
  Py_VISIT(traverse_module_state->__pyx_n_s_popitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_popleft);
  Py_VISIT(traverse_module_state->__pyx_n_s_port);
  Py_VISIT(traverse_module_state->__pyx_kp_u_port_must_be_a_str_bytes_or_int);
  Py_VISIT(traverse_module_state->__pyx_n_u_posix);
  Py_VISIT(traverse_module_state->__pyx_n_s_preexec_fn);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_print);
  Py_VISIT(traverse_module_state->__pyx_n_u_print_debug_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_problems);
  Py_VISIT(traverse_module_state->__pyx_n_s_proc);
  Py_VISIT(traverse_module_state->__pyx_n_s_process_exited);
  Py_VISIT(traverse_module_state->__pyx_n_s_process_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_program);
  Py_VISIT(traverse_module_state->__pyx_n_s_proto);
  Py_VISIT(traverse_module_state->__pyx_n_s_proto_factory);
  Py_VISIT(traverse_module_state->__pyx_n_s_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_u_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_protocol_factory);
  Py_VISIT(traverse_module_state->__pyx_kp_u_protocol_is_not_set_cannot_call);
  Py_VISIT(traverse_module_state->__pyx_kp_u_protocol_is_required);
  Py_VISIT(traverse_module_state->__pyx_kp_u_protocol_pause_writing_failed);
  Py_VISIT(traverse_module_state->__pyx_kp_u_protocol_resume_writing_failed);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyaddr);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_LruCache);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PseudoSocket);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_u_qualname);
  Py_VISIT(traverse_module_state->__pyx_kp_u_r_SSL_handshake_took_1f_ms);
  Py_VISIT(traverse_module_state->__pyx_kp_u_r_fileno_returned_non_integer);
  Py_VISIT(traverse_module_state->__pyx_kp_u_r_received_EOF);
  Py_VISIT(traverse_module_state->__pyx_kp_u_r_s);
  Py_VISIT(traverse_module_state->__pyx_kp_u_r_starts_SSL_handshake);
  Py_VISIT(traverse_module_state->__pyx_kp_u_raddr_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_rads);
  Py_VISIT(traverse_module_state->__pyx_n_s_rai);
  Py_VISIT(traverse_module_state->__pyx_n_s_rai_addr_static);
  Py_VISIT(traverse_module_state->__pyx_n_s_rai_iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_rai_static);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_read);
  Py_VISIT(traverse_module_state->__pyx_n_s_recv);
  Py_VISIT(traverse_module_state->__pyx_n_s_recv_into);
  Py_VISIT(traverse_module_state->__pyx_kp_u_recv_into_method);
  Py_VISIT(traverse_module_state->__pyx_kp_u_recv_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_recvfrom);
  Py_VISIT(traverse_module_state->__pyx_n_s_recvfrom_into);
  Py_VISIT(traverse_module_state->__pyx_kp_u_recvfrom_into_method);
  Py_VISIT(traverse_module_state->__pyx_kp_u_recvfrom_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_recvmsg);
  Py_VISIT(traverse_module_state->__pyx_n_s_recvmsg_into);
  Py_VISIT(traverse_module_state->__pyx_kp_u_recvmsg_into_method);
  Py_VISIT(traverse_module_state->__pyx_kp_u_recvmsg_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_remote_addr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_remote_addr_must_be_a_tuple_of_h);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_done_callback);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_reader);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_reader_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_signal_handler);
  Py_VISIT(traverse_module_state->__pyx_kp_u_remove_signal_handler_can_only_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_writer);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove_writer_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_repr);
  Py_VISIT(traverse_module_state->__pyx_n_s_restore_signals);
  Py_VISIT(traverse_module_state->__pyx_n_s_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_results);
  Py_VISIT(traverse_module_state->__pyx_n_s_resume_cb);
  Py_VISIT(traverse_module_state->__pyx_n_s_resume_reading);
  Py_VISIT(traverse_module_state->__pyx_n_s_resume_writing);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_return_exceptions);
  Py_VISIT(traverse_module_state->__pyx_kp_u_returning_true_from_eof_received);
  Py_VISIT(traverse_module_state->__pyx_n_s_reuse_address);
  Py_VISIT(traverse_module_state->__pyx_n_s_reuse_port);
  Py_VISIT(traverse_module_state->__pyx_kp_u_reuse_port_not_supported_by_sock);
  Py_VISIT(traverse_module_state->__pyx_n_s_reverse);
  Py_VISIT(traverse_module_state->__pyx_n_s_round);
  Py_VISIT(traverse_module_state->__pyx_n_s_rstrip);
  Py_VISIT(traverse_module_state->__pyx_n_s_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_run_forever);
  Py_VISIT(traverse_module_state->__pyx_n_s_run_in_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_run_until_complete);
  Py_VISIT(traverse_module_state->__pyx_n_s_run_until_complete_locals_done_c);
  Py_VISIT(traverse_module_state->__pyx_kp_u_running_closed_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_scope_id);
  Py_VISIT(traverse_module_state->__pyx_kp_u_seconds);
  Py_VISIT(traverse_module_state->__pyx_kp_u_seconds_aborting_the_connection);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_self_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_self_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_self_pipe_exists_before_loop_run);
  Py_VISIT(traverse_module_state->__pyx_kp_u_self_pipe_was_not_cleaned_up_aft);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self_py_buf_self_req_cannot_be_c);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self_py_bufs_self_py_bufs_sml_se);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_kp_u_send_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_send_signal);
  Py_VISIT(traverse_module_state->__pyx_n_s_sendall);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sendall_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_sendfile);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sendfile_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_sendmsg);
  Py_VISIT(traverse_module_state->__pyx_n_s_sendmsg_afalg);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sendmsg_afalg_method);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sendmsg_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_sendto);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sendto_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_serve_forever);
  Py_VISIT(traverse_module_state->__pyx_kp_u_server);
  Py_VISIT(traverse_module_state->__pyx_n_s_server_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_server_hostname);
  Py_VISIT(traverse_module_state->__pyx_kp_u_server_hostname_is_only_meaningf);
  Py_VISIT(traverse_module_state->__pyx_n_s_server_side);
  Py_VISIT(traverse_module_state->__pyx_n_s_servers);
  Py_VISIT(traverse_module_state->__pyx_kp_u_service_proto_not_found);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_asyncgen_hooks);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_coroutine_origin_tracking_de);
  Py_VISIT(traverse_module_state->__pyx_n_u_set_coroutine_wrapper);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_debug);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_default_executor);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_exception);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_exception_handler);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_inheritable);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_read_buffer_limits);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_result);
  Py_VISIT(traverse_module_state->__pyx_n_u_set_running_loop);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_task_factory);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_wakeup_fd);
  Py_VISIT(traverse_module_state->__pyx_kp_u_set_wakeup_fd_1_failed_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_write_buffer_limits);
  Py_VISIT(traverse_module_state->__pyx_n_s_setblocking);
  Py_VISIT(traverse_module_state->__pyx_kp_u_setblocking_transport_sockets_ca);
  Py_VISIT(traverse_module_state->__pyx_n_s_setsockopt);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_settimeout);
  Py_VISIT(traverse_module_state->__pyx_kp_u_settimeout_only_0_timeout_is_all);
  Py_VISIT(traverse_module_state->__pyx_n_s_share);
  Py_VISIT(traverse_module_state->__pyx_n_s_shell);
  Py_VISIT(traverse_module_state->__pyx_kp_u_shell_must_be_False);
  Py_VISIT(traverse_module_state->__pyx_kp_u_shell_must_be_True);
  Py_VISIT(traverse_module_state->__pyx_n_s_shutdown);
  Py_VISIT(traverse_module_state->__pyx_n_s_shutdown_asyncgens);
  Py_VISIT(traverse_module_state->__pyx_n_s_shutdown_coro);
  Py_VISIT(traverse_module_state->__pyx_n_s_shutdown_default_executor);
  Py_VISIT(traverse_module_state->__pyx_kp_u_shutdown_method);
  Py_VISIT(traverse_module_state->__pyx_n_s_sig);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sig_cannot_be_caught);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sig_must_be_an_int_not_r);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sig_out_of_range_1);
  Py_VISIT(traverse_module_state->__pyx_n_s_sighandler);
  Py_VISIT(traverse_module_state->__pyx_n_s_siginterrupt);
  Py_VISIT(traverse_module_state->__pyx_n_s_signal);
  Py_VISIT(traverse_module_state->__pyx_kp_u_signals_handling_has_been_alread);
  Py_VISIT(traverse_module_state->__pyx_kp_u_signals_handling_has_not_been_se);
  Py_VISIT(traverse_module_state->__pyx_n_s_signum);
  Py_VISIT(traverse_module_state->__pyx_n_s_sin6);
  Py_VISIT(traverse_module_state->__pyx_n_s_sl);
  Py_VISIT(traverse_module_state->__pyx_n_s_sleep);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_accept);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_connect);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sock_connect_cb_is_called_on_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_recv);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_recv_into);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sock_recv_into_is_called_on_a_c);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sock_recv_is_called_on_a_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_recvfrom);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_recvfrom_into);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_sendall);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sock_sendall_is_called_on_a_can);
  Py_VISIT(traverse_module_state->__pyx_n_s_sock_sendto);
  Py_VISIT(traverse_module_state->__pyx_n_s_sockaddr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sockaddr_must_be_a_tuple_of_2_3);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sockaddr_resolved_to_multiple_ad);
  Py_VISIT(traverse_module_state->__pyx_kp_u_socket);
  Py_VISIT(traverse_module_state->__pyx_n_s_socket_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_socket_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_socket_family_mismatch_or_a_DNS);
  Py_VISIT(traverse_module_state->__pyx_kp_u_socket_modifier_keyword_argument);
  Py_VISIT(traverse_module_state->__pyx_kp_u_socket_send_raised_exception);
  Py_VISIT(traverse_module_state->__pyx_n_s_socketpair);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sockets);
  Py_VISIT(traverse_module_state->__pyx_n_s_sockets_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_sockname);
  Py_VISIT(traverse_module_state->__pyx_n_s_source);
  Py_VISIT(traverse_module_state->__pyx_n_u_source_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_split);
  Py_VISIT(traverse_module_state->__pyx_n_s_ssl);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ssl_argument_must_be_an_SSLConte);
  Py_VISIT(traverse_module_state->__pyx_n_s_ssl_handshake_timeout);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ssl_handshake_timeout_is_only_me);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ssl_handshake_timeout_should_be);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ssl_is_expected_to_be_None_or_an);
  Py_VISIT(traverse_module_state->__pyx_n_s_ssl_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_ssl_protocol);
  Py_VISIT(traverse_module_state->__pyx_n_s_ssl_shutdown_timeout);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ssl_shutdown_timeout_is_only_mea);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ssl_shutdown_timeout_should_be_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_ssl_waiter);
  Py_VISIT(traverse_module_state->__pyx_n_s_sslcontext);
  Py_VISIT(traverse_module_state->__pyx_kp_u_sslcontext_is_expected_to_be_an);
  Py_VISIT(traverse_module_state->__pyx_n_s_sslproto);
  Py_VISIT(traverse_module_state->__pyx_n_s_st_ino);
  Py_VISIT(traverse_module_state->__pyx_n_s_st_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_stacklevel);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_kp_u_start_init_can_only_be_called);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_new_session);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_serving);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_tls);
  Py_VISIT(traverse_module_state->__pyx_n_s_startupinfo);
  Py_VISIT(traverse_module_state->__pyx_kp_u_startupinfo_is_not_supported);
  Py_VISIT(traverse_module_state->__pyx_n_s_stat);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_stderr);
  Py_VISIT(traverse_module_state->__pyx_n_s_stdin);
  Py_VISIT(traverse_module_state->__pyx_n_s_stdio_fut);
  Py_VISIT(traverse_module_state->__pyx_n_s_stdio_inited);
  Py_VISIT(traverse_module_state->__pyx_n_s_stdout);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_kp_u_stream_is_open_after_UVStream__t);
  Py_VISIT(traverse_module_state->__pyx_n_s_strerror);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_submit);
  Py_VISIT(traverse_module_state->__pyx_n_s_subprocess);
  Py_VISIT(traverse_module_state->__pyx_kp_u_subprocess_STDOUT_is_supported_o);
  Py_VISIT(traverse_module_state->__pyx_n_s_subprocess_exec);
  Py_VISIT(traverse_module_state->__pyx_n_s_subprocess_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_subprocess_shell);
  Py_VISIT(traverse_module_state->__pyx_n_s_subtract);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_sys);
  Py_VISIT(traverse_module_state->__pyx_n_s_target);
  Py_VISIT(traverse_module_state->__pyx_n_s_task);
  Py_VISIT(traverse_module_state->__pyx_n_u_task);
  Py_VISIT(traverse_module_state->__pyx_kp_u_task_factory_must_be_a_callable);
  Py_VISIT(traverse_module_state->__pyx_n_s_tb);
  Py_VISIT(traverse_module_state->__pyx_n_s_tcp);
  Py_VISIT(traverse_module_state->__pyx_n_u_tcp);
  Py_VISIT(traverse_module_state->__pyx_n_s_terminate);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_test__append_write_backlog);
  Py_VISIT(traverse_module_state->__pyx_n_s_test_coroutine_1);
  Py_VISIT(traverse_module_state->__pyx_n_s_testhelper_unwrap_capsuled_poin);
  Py_VISIT(traverse_module_state->__pyx_kp_u_the_socket_must_be_non_blocking);
  Py_VISIT(traverse_module_state->__pyx_kp_u_this_event_loop_is_already_runni);
  Py_VISIT(traverse_module_state->__pyx_n_s_thread);
  Py_VISIT(traverse_module_state->__pyx_n_s_threading);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_timeout);
  Py_VISIT(traverse_module_state->__pyx_n_s_tr);
  Py_VISIT(traverse_module_state->__pyx_n_s_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_traceback_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_transp);
  Py_VISIT(traverse_module_state->__pyx_n_s_transport);
  Py_VISIT(traverse_module_state->__pyx_n_u_transport);
  Py_VISIT(traverse_module_state->__pyx_kp_u_transport_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_transport_sockets_do_not_support);
  Py_VISIT(traverse_module_state->__pyx_n_s_transport_waiter);
  Py_VISIT(traverse_module_state->__pyx_n_s_transports);
  Py_VISIT(traverse_module_state->__pyx_kp_u_try_write_sent_all_data_and_ret);
  Py_VISIT(traverse_module_state->__pyx_n_s_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_udp);
  Py_VISIT(traverse_module_state->__pyx_n_u_udp);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unable_to_listen_no_protocol_fac);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unable_to_perform_operation_on_r);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unable_to_start_the_loop_it_was);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unclosed_resource_r);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unclosed_resource_r_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unclosed_resource_r_could_not_se);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unclosed_transport_uvloop_loop);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unexpected_address_family);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unexpected_kwargs);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unhandled_error_or_an_empty_buff);
  Py_VISIT(traverse_module_state->__pyx_n_s_universal_newlines);
  Py_VISIT(traverse_module_state->__pyx_kp_u_universal_newlines_must_be_False);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unix_socket_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_unlink);
  Py_VISIT(traverse_module_state->__pyx_kp_u_unopened_TCPServer);
  Py_VISIT(traverse_module_state->__pyx_n_s_unset);
  Py_VISIT(traverse_module_state->__pyx_n_s_unwrap);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_kp_u_uv_close_handle_cb_handle_loop);
  Py_VISIT(traverse_module_state->__pyx_kp_u_uv_handle_t_current_freed_total);
  Py_VISIT(traverse_module_state->__pyx_kp_u_uvloop_PseudoSocket_fd_family_s);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_cbhandles_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_basetransport_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_fsevent_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_pipe_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_process_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_stream_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_streamserver_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_tcp_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_handles_udp_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_uvloop_loop);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_loop_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_pseudosock_pyx);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_server_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_u_uvloop_sleep_after_fork);
  Py_VISIT(traverse_module_state->__pyx_kp_u_uvloop_sslproto);
  Py_VISIT(traverse_module_state->__pyx_kp_s_uvloop_sslproto_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_version_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_wait);
  Py_VISIT(traverse_module_state->__pyx_n_s_wait_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_wait_closed);
  Py_VISIT(traverse_module_state->__pyx_n_s_waiter);
  Py_VISIT(traverse_module_state->__pyx_n_s_walk_stack);
  Py_VISIT(traverse_module_state->__pyx_n_s_want);
  Py_VISIT(traverse_module_state->__pyx_n_s_warn);
  Py_VISIT(traverse_module_state->__pyx_n_s_warn_on_full_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_warning);
  Py_VISIT(traverse_module_state->__pyx_n_s_warnings);
  Py_VISIT(traverse_module_state->__pyx_n_u_wb);
  Py_VISIT(traverse_module_state->__pyx_n_s_weakref);
  Py_VISIT(traverse_module_state->__pyx_n_s_when);
  Py_VISIT(traverse_module_state->__pyx_kp_u_without_no_gc_clear_loop_was_se);
  Py_VISIT(traverse_module_state->__pyx_n_s_wrap);
  Py_VISIT(traverse_module_state->__pyx_n_s_wrap_bio);
  Py_VISIT(traverse_module_state->__pyx_n_s_wrap_future);
  Py_VISIT(traverse_module_state->__pyx_n_s_write);
  Py_VISIT(traverse_module_state->__pyx_n_s_write_eof);
  Py_VISIT(traverse_module_state->__pyx_n_s_writelines);
  Py_VISIT(traverse_module_state->__pyx_n_s_zip);
  Py_VISIT(traverse_module_state->__pyx_float_0_1);
  Py_VISIT(traverse_module_state->__pyx_float_1e3);
  Py_VISIT(traverse_module_state->__pyx_float_1eneg_3);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_3);
  Py_VISIT(traverse_module_state->__pyx_int_15);
  Py_VISIT(traverse_module_state->__pyx_int_42);
  Py_VISIT(traverse_module_state->__pyx_int_100);
  Py_VISIT(traverse_module_state->__pyx_int_1000);
  Py_VISIT(traverse_module_state->__pyx_int_50000);
  Py_VISIT(traverse_module_state->__pyx_int_65536);
  Py_VISIT(traverse_module_state->__pyx_int_1048575);
  Py_VISIT(traverse_module_state->__pyx_int_14808274);
  Py_VISIT(traverse_module_state->__pyx_int_87255255);
  Py_VISIT(traverse_module_state->__pyx_int_121062814);
  Py_VISIT(traverse_module_state->__pyx_int_143395587);
  Py_VISIT(traverse_module_state->__pyx_int_173349396);
  Py_VISIT(traverse_module_state->__pyx_int_200977156);
  Py_VISIT(traverse_module_state->__pyx_int_4294967296);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_k__71);
  Py_VISIT(traverse_module_state->__pyx_k__95);
  Py_VISIT(traverse_module_state->__pyx_k__96);
  Py_VISIT(traverse_module_state->__pyx_k__97);
  Py_VISIT(traverse_module_state->__pyx_k__114);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__3);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_tuple__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_slice__88);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__18);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__23);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__27);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__29);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__31);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__40);
  Py_VISIT(traverse_module_state->__pyx_tuple__42);
  Py_VISIT(traverse_module_state->__pyx_tuple__45);
  Py_VISIT(traverse_module_state->__pyx_tuple__46);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__48);
  Py_VISIT(traverse_module_state->__pyx_tuple__49);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__55);
  Py_VISIT(traverse_module_state->__pyx_tuple__56);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__58);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__60);
  Py_VISIT(traverse_module_state->__pyx_tuple__62);
  Py_VISIT(traverse_module_state->__pyx_tuple__64);
  Py_VISIT(traverse_module_state->__pyx_tuple__65);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__67);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_tuple__70);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__74);
  Py_VISIT(traverse_module_state->__pyx_tuple__77);
  Py_VISIT(traverse_module_state->__pyx_tuple__80);
  Py_VISIT(traverse_module_state->__pyx_tuple__81);
  Py_VISIT(traverse_module_state->__pyx_tuple__83);
  Py_VISIT(traverse_module_state->__pyx_tuple__93);
  Py_VISIT(traverse_module_state->__pyx_tuple__94);
  Py_VISIT(traverse_module_state->__pyx_tuple__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__100);
  Py_VISIT(traverse_module_state->__pyx_tuple__101);
  Py_VISIT(traverse_module_state->__pyx_tuple__102);
  Py_VISIT(traverse_module_state->__pyx_tuple__104);
  Py_VISIT(traverse_module_state->__pyx_tuple__106);
  Py_VISIT(traverse_module_state->__pyx_tuple__109);
  Py_VISIT(traverse_module_state->__pyx_tuple__110);
  Py_VISIT(traverse_module_state->__pyx_tuple__111);
  Py_VISIT(traverse_module_state->__pyx_tuple__112);
  Py_VISIT(traverse_module_state->__pyx_tuple__113);
  Py_VISIT(traverse_module_state->__pyx_tuple__117);
  Py_VISIT(traverse_module_state->__pyx_tuple__118);
  Py_VISIT(traverse_module_state->__pyx_tuple__119);
  Py_VISIT(traverse_module_state->__pyx_tuple__120);
  Py_VISIT(traverse_module_state->__pyx_tuple__121);
  Py_VISIT(traverse_module_state->__pyx_tuple__122);
  Py_VISIT(traverse_module_state->__pyx_tuple__125);
  Py_VISIT(traverse_module_state->__pyx_tuple__126);
  Py_VISIT(traverse_module_state->__pyx_tuple__127);
  Py_VISIT(traverse_module_state->__pyx_tuple__131);
  Py_VISIT(traverse_module_state->__pyx_tuple__132);
  Py_VISIT(traverse_module_state->__pyx_tuple__133);
  Py_VISIT(traverse_module_state->__pyx_tuple__134);
  Py_VISIT(traverse_module_state->__pyx_tuple__135);
  Py_VISIT(traverse_module_state->__pyx_tuple__136);
  Py_VISIT(traverse_module_state->__pyx_tuple__137);
  Py_VISIT(traverse_module_state->__pyx_tuple__138);
  Py_VISIT(traverse_module_state->__pyx_tuple__139);
  Py_VISIT(traverse_module_state->__pyx_tuple__142);
  Py_VISIT(traverse_module_state->__pyx_tuple__143);
  Py_VISIT(traverse_module_state->__pyx_tuple__144);
  Py_VISIT(traverse_module_state->__pyx_tuple__145);
  Py_VISIT(traverse_module_state->__pyx_tuple__146);
  Py_VISIT(traverse_module_state->__pyx_tuple__147);
  Py_VISIT(traverse_module_state->__pyx_tuple__148);
  Py_VISIT(traverse_module_state->__pyx_tuple__149);
  Py_VISIT(traverse_module_state->__pyx_tuple__150);
  Py_VISIT(traverse_module_state->__pyx_tuple__151);
  Py_VISIT(traverse_module_state->__pyx_tuple__152);
  Py_VISIT(traverse_module_state->__pyx_tuple__153);
  Py_VISIT(traverse_module_state->__pyx_tuple__154);
  Py_VISIT(traverse_module_state->__pyx_tuple__155);
  Py_VISIT(traverse_module_state->__pyx_tuple__156);
  Py_VISIT(traverse_module_state->__pyx_tuple__157);
  Py_VISIT(traverse_module_state->__pyx_tuple__158);
  Py_VISIT(traverse_module_state->__pyx_tuple__159);
  Py_VISIT(traverse_module_state->__pyx_tuple__160);
  Py_VISIT(traverse_module_state->__pyx_tuple__162);
  Py_VISIT(traverse_module_state->__pyx_tuple__163);
  Py_VISIT(traverse_module_state->__pyx_tuple__164);
  Py_VISIT(traverse_module_state->__pyx_tuple__165);
  Py_VISIT(traverse_module_state->__pyx_tuple__166);
  Py_VISIT(traverse_module_state->__pyx_tuple__167);
  Py_VISIT(traverse_module_state->__pyx_tuple__168);
  Py_VISIT(traverse_module_state->__pyx_tuple__169);
  Py_VISIT(traverse_module_state->__pyx_tuple__171);
  Py_VISIT(traverse_module_state->__pyx_tuple__172);
  Py_VISIT(traverse_module_state->__pyx_tuple__173);
  Py_VISIT(traverse_module_state->__pyx_tuple__174);
  Py_VISIT(traverse_module_state->__pyx_tuple__175);
  Py_VISIT(traverse_module_state->__pyx_tuple__176);
  Py_VISIT(traverse_module_state->__pyx_tuple__177);
  Py_VISIT(traverse_module_state->__pyx_tuple__178);
  Py_VISIT(traverse_module_state->__pyx_tuple__179);
  Py_VISIT(traverse_module_state->__pyx_tuple__180);
  Py_VISIT(traverse_module_state->__pyx_tuple__181);
  Py_VISIT(traverse_module_state->__pyx_tuple__182);
  Py_VISIT(traverse_module_state->__pyx_tuple__183);
  Py_VISIT(traverse_module_state->__pyx_tuple__184);
  Py_VISIT(traverse_module_state->__pyx_tuple__185);
  Py_VISIT(traverse_module_state->__pyx_tuple__186);
  Py_VISIT(traverse_module_state->__pyx_tuple__187);
  Py_VISIT(traverse_module_state->__pyx_tuple__188);
  Py_VISIT(traverse_module_state->__pyx_tuple__195);
  Py_VISIT(traverse_module_state->__pyx_tuple__197);
  Py_VISIT(traverse_module_state->__pyx_tuple__199);
  Py_VISIT(traverse_module_state->__pyx_tuple__200);
  Py_VISIT(traverse_module_state->__pyx_tuple__201);
  Py_VISIT(traverse_module_state->__pyx_tuple__202);
  Py_VISIT(traverse_module_state->__pyx_tuple__203);
  Py_VISIT(traverse_module_state->__pyx_tuple__204);
  Py_VISIT(traverse_module_state->__pyx_tuple__205);
  Py_VISIT(traverse_module_state->__pyx_tuple__206);
  Py_VISIT(traverse_module_state->__pyx_tuple__207);
  Py_VISIT(traverse_module_state->__pyx_tuple__209);
  Py_VISIT(traverse_module_state->__pyx_tuple__211);
  Py_VISIT(traverse_module_state->__pyx_tuple__213);
  Py_VISIT(traverse_module_state->__pyx_tuple__215);
  Py_VISIT(traverse_module_state->__pyx_tuple__217);
  Py_VISIT(traverse_module_state->__pyx_tuple__221);
  Py_VISIT(traverse_module_state->__pyx_tuple__229);
  Py_VISIT(traverse_module_state->__pyx_tuple__231);
  Py_VISIT(traverse_module_state->__pyx_tuple__234);
  Py_VISIT(traverse_module_state->__pyx_tuple__236);
  Py_VISIT(traverse_module_state->__pyx_tuple__237);
  Py_VISIT(traverse_module_state->__pyx_tuple__238);
  Py_VISIT(traverse_module_state->__pyx_tuple__239);
  Py_VISIT(traverse_module_state->__pyx_tuple__240);
  Py_VISIT(traverse_module_state->__pyx_tuple__241);
  Py_VISIT(traverse_module_state->__pyx_tuple__242);
  Py_VISIT(traverse_module_state->__pyx_tuple__243);
  Py_VISIT(traverse_module_state->__pyx_tuple__244);
  Py_VISIT(traverse_module_state->__pyx_tuple__245);
  Py_VISIT(traverse_module_state->__pyx_tuple__246);
  Py_VISIT(traverse_module_state->__pyx_tuple__249);
  Py_VISIT(traverse_module_state->__pyx_tuple__251);
  Py_VISIT(traverse_module_state->__pyx_tuple__253);
  Py_VISIT(traverse_module_state->__pyx_tuple__255);
  Py_VISIT(traverse_module_state->__pyx_tuple__259);
  Py_VISIT(traverse_module_state->__pyx_tuple__260);
  Py_VISIT(traverse_module_state->__pyx_tuple__261);
  Py_VISIT(traverse_module_state->__pyx_tuple__262);
  Py_VISIT(traverse_module_state->__pyx_tuple__264);
  Py_VISIT(traverse_module_state->__pyx_tuple__265);
  Py_VISIT(traverse_module_state->__pyx_tuple__266);
  Py_VISIT(traverse_module_state->__pyx_tuple__267);
  Py_VISIT(traverse_module_state->__pyx_tuple__268);
  Py_VISIT(traverse_module_state->__pyx_tuple__269);
  Py_VISIT(traverse_module_state->__pyx_tuple__271);
  Py_VISIT(traverse_module_state->__pyx_tuple__273);
  Py_VISIT(traverse_module_state->__pyx_tuple__274);
  Py_VISIT(traverse_module_state->__pyx_tuple__275);
  Py_VISIT(traverse_module_state->__pyx_tuple__276);
  Py_VISIT(traverse_module_state->__pyx_tuple__277);
  Py_VISIT(traverse_module_state->__pyx_tuple__279);
  Py_VISIT(traverse_module_state->__pyx_tuple__281);
  Py_VISIT(traverse_module_state->__pyx_tuple__282);
  Py_VISIT(traverse_module_state->__pyx_tuple__285);
  Py_VISIT(traverse_module_state->__pyx_tuple__288);
  Py_VISIT(traverse_module_state->__pyx_tuple__289);
  Py_VISIT(traverse_module_state->__pyx_tuple__290);
  Py_VISIT(traverse_module_state->__pyx_tuple__293);
  Py_VISIT(traverse_module_state->__pyx_tuple__295);
  Py_VISIT(traverse_module_state->__pyx_tuple__298);
  Py_VISIT(traverse_module_state->__pyx_tuple__300);
  Py_VISIT(traverse_module_state->__pyx_tuple__303);
  Py_VISIT(traverse_module_state->__pyx_tuple__321);
  Py_VISIT(traverse_module_state->__pyx_tuple__325);
  Py_VISIT(traverse_module_state->__pyx_tuple__327);
  Py_VISIT(traverse_module_state->__pyx_tuple__330);
  Py_VISIT(traverse_module_state->__pyx_tuple__333);
  Py_VISIT(traverse_module_state->__pyx_tuple__336);
  Py_VISIT(traverse_module_state->__pyx_tuple__357);
  Py_VISIT(traverse_module_state->__pyx_tuple__360);
  Py_VISIT(traverse_module_state->__pyx_tuple__363);
  Py_VISIT(traverse_module_state->__pyx_tuple__365);
  Py_VISIT(traverse_module_state->__pyx_tuple__385);
  Py_VISIT(traverse_module_state->__pyx_tuple__387);
  Py_VISIT(traverse_module_state->__pyx_tuple__393);
  Py_VISIT(traverse_module_state->__pyx_tuple__396);
  Py_VISIT(traverse_module_state->__pyx_tuple__402);
  Py_VISIT(traverse_module_state->__pyx_tuple__404);
  Py_VISIT(traverse_module_state->__pyx_tuple__413);
  Py_VISIT(traverse_module_state->__pyx_tuple__431);
  Py_VISIT(traverse_module_state->__pyx_tuple__433);
  Py_VISIT(traverse_module_state->__pyx_tuple__448);
  Py_VISIT(traverse_module_state->__pyx_tuple__454);
  Py_VISIT(traverse_module_state->__pyx_tuple__458);
  Py_VISIT(traverse_module_state->__pyx_tuple__463);
  Py_VISIT(traverse_module_state->__pyx_tuple__467);
  Py_VISIT(traverse_module_state->__pyx_tuple__469);
  Py_VISIT(traverse_module_state->__pyx_tuple__513);
  Py_VISIT(traverse_module_state->__pyx_tuple__517);
  Py_VISIT(traverse_module_state->__pyx_tuple__519);
  Py_VISIT(traverse_module_state->__pyx_tuple__528);
  Py_VISIT(traverse_module_state->__pyx_tuple__533);
  Py_VISIT(traverse_module_state->__pyx_tuple__534);
  Py_VISIT(traverse_module_state->__pyx_tuple__539);
  Py_VISIT(traverse_module_state->__pyx_codeobj__20);
  Py_VISIT(traverse_module_state->__pyx_codeobj__22);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__84);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  Py_VISIT(traverse_module_state->__pyx_codeobj__89);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__91);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__98);
  Py_VISIT(traverse_module_state->__pyx_codeobj__103);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__107);
  Py_VISIT(traverse_module_state->__pyx_codeobj__108);
  Py_VISIT(traverse_module_state->__pyx_codeobj__116);
  Py_VISIT(traverse_module_state->__pyx_codeobj__123);
  Py_VISIT(traverse_module_state->__pyx_codeobj__124);
  Py_VISIT(traverse_module_state->__pyx_codeobj__189);
  Py_VISIT(traverse_module_state->__pyx_codeobj__190);
  Py_VISIT(traverse_module_state->__pyx_codeobj__191);
  Py_VISIT(traverse_module_state->__pyx_codeobj__192);
  Py_VISIT(traverse_module_state->__pyx_codeobj__193);
  Py_VISIT(traverse_module_state->__pyx_codeobj__194);
  Py_VISIT(traverse_module_state->__pyx_codeobj__208);
  Py_VISIT(traverse_module_state->__pyx_codeobj__210);
  Py_VISIT(traverse_module_state->__pyx_codeobj__212);
  Py_VISIT(traverse_module_state->__pyx_codeobj__214);
  Py_VISIT(traverse_module_state->__pyx_codeobj__216);
  Py_VISIT(traverse_module_state->__pyx_codeobj__218);
  Py_VISIT(traverse_module_state->__pyx_codeobj__219);
  Py_VISIT(traverse_module_state->__pyx_codeobj__220);
  Py_VISIT(traverse_module_state->__pyx_codeobj__222);
  Py_VISIT(traverse_module_state->__pyx_codeobj__223);
  Py_VISIT(traverse_module_state->__pyx_codeobj__224);
  Py_VISIT(traverse_module_state->__pyx_codeobj__225);
  Py_VISIT(traverse_module_state->__pyx_codeobj__226);
  Py_VISIT(traverse_module_state->__pyx_codeobj__227);
  Py_VISIT(traverse_module_state->__pyx_codeobj__228);
  Py_VISIT(traverse_module_state->__pyx_codeobj__230);
  Py_VISIT(traverse_module_state->__pyx_codeobj__232);
  Py_VISIT(traverse_module_state->__pyx_codeobj__233);
  Py_VISIT(traverse_module_state->__pyx_codeobj__235);
  Py_VISIT(traverse_module_state->__pyx_codeobj__247);
  Py_VISIT(traverse_module_state->__pyx_codeobj__248);
  Py_VISIT(traverse_module_state->__pyx_codeobj__250);
  Py_VISIT(traverse_module_state->__pyx_codeobj__252);
  Py_VISIT(traverse_module_state->__pyx_codeobj__254);
  Py_VISIT(traverse_module_state->__pyx_codeobj__256);
  Py_VISIT(traverse_module_state->__pyx_codeobj__257);
  Py_VISIT(traverse_module_state->__pyx_codeobj__258);
  Py_VISIT(traverse_module_state->__pyx_codeobj__270);
  Py_VISIT(traverse_module_state->__pyx_codeobj__272);
  Py_VISIT(traverse_module_state->__pyx_codeobj__278);
  Py_VISIT(traverse_module_state->__pyx_codeobj__280);
  Py_VISIT(traverse_module_state->__pyx_codeobj__283);
  Py_VISIT(traverse_module_state->__pyx_codeobj__284);
  Py_VISIT(traverse_module_state->__pyx_codeobj__286);
  Py_VISIT(traverse_module_state->__pyx_codeobj__287);
  Py_VISIT(traverse_module_state->__pyx_codeobj__291);
  Py_VISIT(traverse_module_state->__pyx_codeobj__292);
  Py_VISIT(traverse_module_state->__pyx_codeobj__294);
  Py_VISIT(traverse_module_state->__pyx_codeobj__296);
  Py_VISIT(traverse_module_state->__pyx_codeobj__297);
  Py_VISIT(traverse_module_state->__pyx_codeobj__299);
  Py_VISIT(traverse_module_state->__pyx_codeobj__301);
  Py_VISIT(traverse_module_state->__pyx_codeobj__302);
  Py_VISIT(traverse_module_state->__pyx_codeobj__304);
  Py_VISIT(traverse_module_state->__pyx_codeobj__305);
  Py_VISIT(traverse_module_state->__pyx_codeobj__306);
  Py_VISIT(traverse_module_state->__pyx_codeobj__307);
  Py_VISIT(traverse_module_state->__pyx_codeobj__308);
  Py_VISIT(traverse_module_state->__pyx_codeobj__309);
  Py_VISIT(traverse_module_state->__pyx_codeobj__310);
  Py_VISIT(traverse_module_state->__pyx_codeobj__311);
  Py_VISIT(traverse_module_state->__pyx_codeobj__312);
  Py_VISIT(traverse_module_state->__pyx_codeobj__313);
  Py_VISIT(traverse_module_state->__pyx_codeobj__314);
  Py_VISIT(traverse_module_state->__pyx_codeobj__315);
  Py_VISIT(traverse_module_state->__pyx_codeobj__316);
  Py_VISIT(traverse_module_state->__pyx_codeobj__317);
  Py_VISIT(traverse_module_state->__pyx_codeobj__318);
  Py_VISIT(traverse_module_state->__pyx_codeobj__319);
  Py_VISIT(traverse_module_state->__pyx_codeobj__320);
  Py_VISIT(traverse_module_state->__pyx_codeobj__322);
  Py_VISIT(traverse_module_state->__pyx_codeobj__323);
  Py_VISIT(traverse_module_state->__pyx_codeobj__324);
  Py_VISIT(traverse_module_state->__pyx_codeobj__326);
  Py_VISIT(traverse_module_state->__pyx_codeobj__328);
  Py_VISIT(traverse_module_state->__pyx_codeobj__329);
  Py_VISIT(traverse_module_state->__pyx_codeobj__331);
  Py_VISIT(traverse_module_state->__pyx_codeobj__332);
  Py_VISIT(traverse_module_state->__pyx_codeobj__334);
  Py_VISIT(traverse_module_state->__pyx_codeobj__335);
  Py_VISIT(traverse_module_state->__pyx_codeobj__337);
  Py_VISIT(traverse_module_state->__pyx_codeobj__338);
  Py_VISIT(traverse_module_state->__pyx_codeobj__339);
  Py_VISIT(traverse_module_state->__pyx_codeobj__340);
  Py_VISIT(traverse_module_state->__pyx_codeobj__341);
  Py_VISIT(traverse_module_state->__pyx_codeobj__342);
  Py_VISIT(traverse_module_state->__pyx_codeobj__343);
  Py_VISIT(traverse_module_state->__pyx_codeobj__344);
  Py_VISIT(traverse_module_state->__pyx_codeobj__345);
  Py_VISIT(traverse_module_state->__pyx_codeobj__346);
  Py_VISIT(traverse_module_state->__pyx_codeobj__347);
  Py_VISIT(traverse_module_state->__pyx_codeobj__348);
  Py_VISIT(traverse_module_state->__pyx_codeobj__349);
  Py_VISIT(traverse_module_state->__pyx_codeobj__350);
  Py_VISIT(traverse_module_state->__pyx_codeobj__351);
  Py_VISIT(traverse_module_state->__pyx_codeobj__352);
  Py_VISIT(traverse_module_state->__pyx_codeobj__353);
  Py_VISIT(traverse_module_state->__pyx_codeobj__354);
  Py_VISIT(traverse_module_state->__pyx_codeobj__355);
  Py_VISIT(traverse_module_state->__pyx_codeobj__356);
  Py_VISIT(traverse_module_state->__pyx_codeobj__358);
  Py_VISIT(traverse_module_state->__pyx_codeobj__359);
  Py_VISIT(traverse_module_state->__pyx_codeobj__361);
  Py_VISIT(traverse_module_state->__pyx_codeobj__362);
  Py_VISIT(traverse_module_state->__pyx_codeobj__364);
  Py_VISIT(traverse_module_state->__pyx_codeobj__366);
  Py_VISIT(traverse_module_state->__pyx_codeobj__367);
  Py_VISIT(traverse_module_state->__pyx_codeobj__368);
  Py_VISIT(traverse_module_state->__pyx_codeobj__369);
  Py_VISIT(traverse_module_state->__pyx_codeobj__370);
  Py_VISIT(traverse_module_state->__pyx_codeobj__371);
  Py_VISIT(traverse_module_state->__pyx_codeobj__372);
  Py_VISIT(traverse_module_state->__pyx_codeobj__373);
  Py_VISIT(traverse_module_state->__pyx_codeobj__374);
  Py_VISIT(traverse_module_state->__pyx_codeobj__375);
  Py_VISIT(traverse_module_state->__pyx_codeobj__376);
  Py_VISIT(traverse_module_state->__pyx_codeobj__377);
  Py_VISIT(traverse_module_state->__pyx_codeobj__378);
  Py_VISIT(traverse_module_state->__pyx_codeobj__379);
  Py_VISIT(traverse_module_state->__pyx_codeobj__380);
  Py_VISIT(traverse_module_state->__pyx_codeobj__381);
  Py_VISIT(traverse_module_state->__pyx_codeobj__382);
  Py_VISIT(traverse_module_state->__pyx_codeobj__383);
  Py_VISIT(traverse_module_state->__pyx_codeobj__384);
  Py_VISIT(traverse_module_state->__pyx_codeobj__386);
  Py_VISIT(traverse_module_state->__pyx_codeobj__388);
  Py_VISIT(traverse_module_state->__pyx_codeobj__389);
  Py_VISIT(traverse_module_state->__pyx_codeobj__390);
  Py_VISIT(traverse_module_state->__pyx_codeobj__391);
  Py_VISIT(traverse_module_state->__pyx_codeobj__392);
  Py_VISIT(traverse_module_state->__pyx_codeobj__394);
  Py_VISIT(traverse_module_state->__pyx_codeobj__395);
  Py_VISIT(traverse_module_state->__pyx_codeobj__397);
  Py_VISIT(traverse_module_state->__pyx_codeobj__398);
  Py_VISIT(traverse_module_state->__pyx_codeobj__399);
  Py_VISIT(traverse_module_state->__pyx_codeobj__400);
  Py_VISIT(traverse_module_state->__pyx_codeobj__401);
  Py_VISIT(traverse_module_state->__pyx_codeobj__403);
  Py_VISIT(traverse_module_state->__pyx_codeobj__405);
  Py_VISIT(traverse_module_state->__pyx_codeobj__406);
  Py_VISIT(traverse_module_state->__pyx_codeobj__407);
  Py_VISIT(traverse_module_state->__pyx_codeobj__408);
  Py_VISIT(traverse_module_state->__pyx_codeobj__409);
  Py_VISIT(traverse_module_state->__pyx_codeobj__410);
  Py_VISIT(traverse_module_state->__pyx_codeobj__411);
  Py_VISIT(traverse_module_state->__pyx_codeobj__412);
  Py_VISIT(traverse_module_state->__pyx_codeobj__414);
  Py_VISIT(traverse_module_state->__pyx_codeobj__415);
  Py_VISIT(traverse_module_state->__pyx_codeobj__416);
  Py_VISIT(traverse_module_state->__pyx_codeobj__417);
  Py_VISIT(traverse_module_state->__pyx_codeobj__418);
  Py_VISIT(traverse_module_state->__pyx_codeobj__419);
  Py_VISIT(traverse_module_state->__pyx_codeobj__420);
  Py_VISIT(traverse_module_state->__pyx_codeobj__421);
  Py_VISIT(traverse_module_state->__pyx_codeobj__422);
  Py_VISIT(traverse_module_state->__pyx_codeobj__423);
  Py_VISIT(traverse_module_state->__pyx_codeobj__424);
  Py_VISIT(traverse_module_state->__pyx_codeobj__425);
  Py_VISIT(traverse_module_state->__pyx_codeobj__426);
  Py_VISIT(traverse_module_state->__pyx_codeobj__427);
  Py_VISIT(traverse_module_state->__pyx_codeobj__428);
  Py_VISIT(traverse_module_state->__pyx_codeobj__429);
  Py_VISIT(traverse_module_state->__pyx_codeobj__430);
  Py_VISIT(traverse_module_state->__pyx_codeobj__432);
  Py_VISIT(traverse_module_state->__pyx_codeobj__434);
  Py_VISIT(traverse_module_state->__pyx_codeobj__435);
  Py_VISIT(traverse_module_state->__pyx_codeobj__436);
  Py_VISIT(traverse_module_state->__pyx_codeobj__437);
  Py_VISIT(traverse_module_state->__pyx_codeobj__438);
  Py_VISIT(traverse_module_state->__pyx_codeobj__439);
  Py_VISIT(traverse_module_state->__pyx_codeobj__440);
  Py_VISIT(traverse_module_state->__pyx_codeobj__441);
  Py_VISIT(traverse_module_state->__pyx_codeobj__442);
  Py_VISIT(traverse_module_state->__pyx_codeobj__443);
  Py_VISIT(traverse_module_state->__pyx_codeobj__444);
  Py_VISIT(traverse_module_state->__pyx_codeobj__445);
  Py_VISIT(traverse_module_state->__pyx_codeobj__446);
  Py_VISIT(traverse_module_state->__pyx_codeobj__447);
  Py_VISIT(traverse_module_state->__pyx_codeobj__449);
  Py_VISIT(traverse_module_state->__pyx_codeobj__450);
  Py_VISIT(traverse_module_state->__pyx_codeobj__451);
  Py_VISIT(traverse_module_state->__pyx_codeobj__452);
  Py_VISIT(traverse_module_state->__pyx_codeobj__453);
  Py_VISIT(traverse_module_state->__pyx_codeobj__455);
  Py_VISIT(traverse_module_state->__pyx_codeobj__456);
  Py_VISIT(traverse_module_state->__pyx_codeobj__457);
  Py_VISIT(traverse_module_state->__pyx_codeobj__459);
  Py_VISIT(traverse_module_state->__pyx_codeobj__460);
  Py_VISIT(traverse_module_state->__pyx_codeobj__461);
  Py_VISIT(traverse_module_state->__pyx_codeobj__462);
  Py_VISIT(traverse_module_state->__pyx_codeobj__464);
  Py_VISIT(traverse_module_state->__pyx_codeobj__465);
  Py_VISIT(traverse_module_state->__pyx_codeobj__466);
  Py_VISIT(traverse_module_state->__pyx_codeobj__468);
  Py_VISIT(traverse_module_state->__pyx_codeobj__470);
  Py_VISIT(traverse_module_state->__pyx_codeobj__471);
  Py_VISIT(traverse_module_state->__pyx_codeobj__472);
  Py_VISIT(traverse_module_state->__pyx_codeobj__473);
  Py_VISIT(traverse_module_state->__pyx_codeobj__474);
  Py_VISIT(traverse_module_state->__pyx_codeobj__475);
  Py_VISIT(traverse_module_state->__pyx_codeobj__476);
  Py_VISIT(traverse_module_state->__pyx_codeobj__477);
  Py_VISIT(traverse_module_state->__pyx_codeobj__478);
  Py_VISIT(traverse_module_state->__pyx_codeobj__479);
  Py_VISIT(traverse_module_state->__pyx_codeobj__480);
  Py_VISIT(traverse_module_state->__pyx_codeobj__481);
  Py_VISIT(traverse_module_state->__pyx_codeobj__482);
  Py_VISIT(traverse_module_state->__pyx_codeobj__483);
  Py_VISIT(traverse_module_state->__pyx_codeobj__484);
  Py_VISIT(traverse_module_state->__pyx_codeobj__485);
  Py_VISIT(traverse_module_state->__pyx_codeobj__486);
  Py_VISIT(traverse_module_state->__pyx_codeobj__487);
  Py_VISIT(traverse_module_state->__pyx_codeobj__488);
  Py_VISIT(traverse_module_state->__pyx_codeobj__489);
  Py_VISIT(traverse_module_state->__pyx_codeobj__490);
  Py_VISIT(traverse_module_state->__pyx_codeobj__491);
  Py_VISIT(traverse_module_state->__pyx_codeobj__492);
  Py_VISIT(traverse_module_state->__pyx_codeobj__493);
  Py_VISIT(traverse_module_state->__pyx_codeobj__494);
  Py_VISIT(traverse_module_state->__pyx_codeobj__495);
  Py_VISIT(traverse_module_state->__pyx_codeobj__496);
  Py_VISIT(traverse_module_state->__pyx_codeobj__497);
  Py_VISIT(traverse_module_state->__pyx_codeobj__498);
  Py_VISIT(traverse_module_state->__pyx_codeobj__499);
  Py_VISIT(traverse_module_state->__pyx_codeobj__500);
  Py_VISIT(traverse_module_state->__pyx_codeobj__501);
  Py_VISIT(traverse_module_state->__pyx_codeobj__502);
  Py_VISIT(traverse_module_state->__pyx_codeobj__503);
  Py_VISIT(traverse_module_state->__pyx_codeobj__504);
  Py_VISIT(traverse_module_state->__pyx_codeobj__505);
  Py_VISIT(traverse_module_state->__pyx_codeobj__506);
  Py_VISIT(traverse_module_state->__pyx_codeobj__507);
  Py_VISIT(traverse_module_state->__pyx_codeobj__508);
  Py_VISIT(traverse_module_state->__pyx_codeobj__509);
  Py_VISIT(traverse_module_state->__pyx_codeobj__510);
  Py_VISIT(traverse_module_state->__pyx_codeobj__511);
  Py_VISIT(traverse_module_state->__pyx_codeobj__512);
  Py_VISIT(traverse_module_state->__pyx_codeobj__514);
  Py_VISIT(traverse_module_state->__pyx_codeobj__515);
  Py_VISIT(traverse_module_state->__pyx_codeobj__516);
  Py_VISIT(traverse_module_state->__pyx_codeobj__518);
  Py_VISIT(traverse_module_state->__pyx_codeobj__520);
  Py_VISIT(traverse_module_state->__pyx_codeobj__521);
  Py_VISIT(traverse_module_state->__pyx_codeobj__522);
  Py_VISIT(traverse_module_state->__pyx_codeobj__523);
  Py_VISIT(traverse_module_state->__pyx_codeobj__524);
  Py_VISIT(traverse_module_state->__pyx_codeobj__525);
  Py_VISIT(traverse_module_state->__pyx_codeobj__526);
  Py_VISIT(traverse_module_state->__pyx_codeobj__527);
  Py_VISIT(traverse_module_state->__pyx_codeobj__529);
  Py_VISIT(traverse_module_state->__pyx_codeobj__530);
  Py_VISIT(traverse_module_state->__pyx_codeobj__531);
  Py_VISIT(traverse_module_state->__pyx_codeobj__532);
  Py_VISIT(traverse_module_state->__pyx_codeobj__535);
  Py_VISIT(traverse_module_state->__pyx_codeobj__536);
  Py_VISIT(traverse_module_state->__pyx_codeobj__537);
  Py_VISIT(traverse_module_state->__pyx_codeobj__538);
  Py_VISIT(traverse_module_state->__pyx_codeobj__540);
  Py_VISIT(traverse_module_state->__pyx_codeobj__541);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_6uvloop_4loop_UVHandle __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVHandle
#define __pyx_type_6uvloop_4loop_UVSocketHandle __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVSocketHandle
#define __pyx_type_6uvloop_4loop_UVAsync __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVAsync
#define __pyx_type_6uvloop_4loop_UVTimer __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVTimer
#define __pyx_type_6uvloop_4loop_UVIdle __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVIdle
#define __pyx_type_6uvloop_4loop_UVBaseTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVBaseTransport
#define __pyx_type_6uvloop_4loop_Loop __pyx_mstate_global->__pyx_type_6uvloop_4loop_Loop
#define __pyx_type_6uvloop_4loop_Handle __pyx_mstate_global->__pyx_type_6uvloop_4loop_Handle
#define __pyx_type_6uvloop_4loop_TimerHandle __pyx_mstate_global->__pyx_type_6uvloop_4loop_TimerHandle
#define __pyx_type_6uvloop_4loop_UVCheck __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVCheck
#define __pyx_type_6uvloop_4loop_UVPoll __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVPoll
#define __pyx_type_6uvloop_4loop_UVStream __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVStream
#define __pyx_type_6uvloop_4loop_UVStreamServer __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVStreamServer
#define __pyx_type_6uvloop_4loop_TCPServer __pyx_mstate_global->__pyx_type_6uvloop_4loop_TCPServer
#define __pyx_type_6uvloop_4loop_TCPTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop_TCPTransport
#define __pyx_type_6uvloop_4loop_UnixServer __pyx_mstate_global->__pyx_type_6uvloop_4loop_UnixServer
#define __pyx_type_6uvloop_4loop_UnixTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop_UnixTransport
#define __pyx_type_6uvloop_4loop_ReadUnixTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop_ReadUnixTransport
#define __pyx_type_6uvloop_4loop_WriteUnixTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop_WriteUnixTransport
#define __pyx_type_6uvloop_4loop_UVProcess __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVProcess
#define __pyx_type_6uvloop_4loop_UVProcessTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVProcessTransport
#define __pyx_type_6uvloop_4loop_UVFSEvent __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVFSEvent
#define __pyx_type_6uvloop_4loop_UVRequest __pyx_mstate_global->__pyx_type_6uvloop_4loop_UVRequest
#define __pyx_type_6uvloop_4loop__SSLProtocolTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop__SSLProtocolTransport
#define __pyx_type_6uvloop_4loop_SSLProtocol __pyx_mstate_global->__pyx_type_6uvloop_4loop_SSLProtocol
#define __pyx_type_6uvloop_4loop_UDPTransport __pyx_mstate_global->__pyx_type_6uvloop_4loop_UDPTransport
#define __pyx_type_6uvloop_4loop_Server __pyx_mstate_global->__pyx_type_6uvloop_4loop_Server
#define __pyx_type_6uvloop_4loop_PseudoSocket __pyx_mstate_global->__pyx_type_6uvloop_4loop_PseudoSocket
#define __pyx_type_6uvloop_4loop_LruCache __pyx_mstate_global->__pyx_type_6uvloop_4loop_LruCache
#define __pyx_type_6uvloop_4loop__StreamWriteContext __pyx_mstate_global->__pyx_type_6uvloop_4loop__StreamWriteContext
#define __pyx_type_6uvloop_4loop__TCPConnectRequest __pyx_mstate_global->__pyx_type_6uvloop_4loop__TCPConnectRequest
#define __pyx_type_6uvloop_4loop__PipeConnectRequest __pyx_mstate_global->__pyx_type_6uvloop_4loop__PipeConnectRequest
#define __pyx_type_6uvloop_4loop_SockAddrHolder __pyx_mstate_global->__pyx_type_6uvloop_4loop_SockAddrHolder
#define __pyx_type_6uvloop_4loop_AddrInfo __pyx_mstate_global->__pyx_type_6uvloop_4loop_AddrInfo
#define __pyx_type_6uvloop_4loop_AddrInfoRequest __pyx_mstate_global->__pyx_type_6uvloop_4loop_AddrInfoRequest
#define __pyx_type_6uvloop_4loop_NameInfoRequest __pyx_mstate_global->__pyx_type_6uvloop_4loop_NameInfoRequest
#define __pyx_type_6uvloop_4loop__UDPSendContext __pyx_mstate_global->__pyx_type_6uvloop_4loop__UDPSendContext
#define __pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_2___get__ __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_2___get__
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter__ __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter__
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit__ __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit__
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever
#define __pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 __pyx_mstate_global->__pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1
#define __pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled __pyx_mstate_global->__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled
#endif
#define __pyx_ptype_6uvloop_4loop_UVHandle __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVHandle
#define __pyx_ptype_6uvloop_4loop_UVSocketHandle __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVSocketHandle
#define __pyx_ptype_6uvloop_4loop_UVAsync __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVAsync
#define __pyx_ptype_6uvloop_4loop_UVTimer __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVTimer
#define __pyx_ptype_6uvloop_4loop_UVIdle __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVIdle
#define __pyx_ptype_6uvloop_4loop_UVBaseTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVBaseTransport
#define __pyx_ptype_6uvloop_4loop_Loop __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_Loop
#define __pyx_ptype_6uvloop_4loop_Handle __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_Handle
#define __pyx_ptype_6uvloop_4loop_TimerHandle __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_TimerHandle
#define __pyx_ptype_6uvloop_4loop_UVCheck __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVCheck
#define __pyx_ptype_6uvloop_4loop_UVPoll __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVPoll
#define __pyx_ptype_6uvloop_4loop_UVStream __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVStream
#define __pyx_ptype_6uvloop_4loop_UVStreamServer __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVStreamServer
#define __pyx_ptype_6uvloop_4loop_TCPServer __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_TCPServer
#define __pyx_ptype_6uvloop_4loop_TCPTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_TCPTransport
#define __pyx_ptype_6uvloop_4loop_UnixServer __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UnixServer
#define __pyx_ptype_6uvloop_4loop_UnixTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UnixTransport
#define __pyx_ptype_6uvloop_4loop_ReadUnixTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_ReadUnixTransport
#define __pyx_ptype_6uvloop_4loop_WriteUnixTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_WriteUnixTransport
#define __pyx_ptype_6uvloop_4loop_UVProcess __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVProcess
#define __pyx_ptype_6uvloop_4loop_UVProcessTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVProcessTransport
#define __pyx_ptype_6uvloop_4loop_UVFSEvent __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVFSEvent
#define __pyx_ptype_6uvloop_4loop_UVRequest __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UVRequest
#define __pyx_ptype_6uvloop_4loop__SSLProtocolTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop__SSLProtocolTransport
#define __pyx_ptype_6uvloop_4loop_SSLProtocol __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_SSLProtocol
#define __pyx_ptype_6uvloop_4loop_UDPTransport __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_UDPTransport
#define __pyx_ptype_6uvloop_4loop_Server __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_Server
#define __pyx_ptype_6uvloop_4loop_PseudoSocket __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_PseudoSocket
#define __pyx_ptype_6uvloop_4loop_LruCache __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_LruCache
#define __pyx_ptype_6uvloop_4loop__StreamWriteContext __pyx_mstate_global->__pyx_ptype_6uvloop_4loop__StreamWriteContext
#define __pyx_ptype_6uvloop_4loop__TCPConnectRequest __pyx_mstate_global->__pyx_ptype_6uvloop_4loop__TCPConnectRequest
#define __pyx_ptype_6uvloop_4loop__PipeConnectRequest __pyx_mstate_global->__pyx_ptype_6uvloop_4loop__PipeConnectRequest
#define __pyx_ptype_6uvloop_4loop_SockAddrHolder __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_SockAddrHolder
#define __pyx_ptype_6uvloop_4loop_AddrInfo __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_AddrInfo
#define __pyx_ptype_6uvloop_4loop_AddrInfoRequest __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_AddrInfoRequest
#define __pyx_ptype_6uvloop_4loop_NameInfoRequest __pyx_mstate_global->__pyx_ptype_6uvloop_4loop_NameInfoRequest
#define __pyx_ptype_6uvloop_4loop__UDPSendContext __pyx_mstate_global->__pyx_ptype_6uvloop_4loop__UDPSendContext
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__ __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__ __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__ __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever
#define __pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 __pyx_mstate_global->__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1
#define __pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled __pyx_mstate_global->__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled
#define __pyx_kp_b_1 __pyx_mstate_global->__pyx_kp_b_1
#define __pyx_kp_u_18_7_7_7 __pyx_mstate_global->__pyx_kp_u_18_7_7_7
#define __pyx_kp_u_AF_INET6_address_must_be_tuple __pyx_mstate_global->__pyx_kp_u_AF_INET6_address_must_be_tuple
#define __pyx_kp_u_AF_INET6_must_be_a_tuple_of_2_4 __pyx_mstate_global->__pyx_kp_u_AF_INET6_must_be_a_tuple_of_2_4
#define __pyx_kp_u_AF_INET_address_must_be_tuple __pyx_mstate_global->__pyx_kp_u_AF_INET_address_must_be_tuple
#define __pyx_kp_u_AF_INET_address_must_be_tuple_of __pyx_mstate_global->__pyx_kp_u_AF_INET_address_must_be_tuple_of
#define __pyx_kp_u_AF_UNIX_address_must_be_a_str_or __pyx_mstate_global->__pyx_kp_u_AF_UNIX_address_must_be_a_str_or
#define __pyx_kp_u_AF_UNSPEC_in_DNS_results __pyx_mstate_global->__pyx_kp_u_AF_UNSPEC_in_DNS_results
#define __pyx_n_u_AI_CANONNAME __pyx_mstate_global->__pyx_n_u_AI_CANONNAME
#define __pyx_kp_u_A_Stream_Socket_was_expected_got __pyx_mstate_global->__pyx_kp_u_A_Stream_Socket_was_expected_got
#define __pyx_kp_u_A_UDP_Socket_was_expected_got_r __pyx_mstate_global->__pyx_kp_u_A_UDP_Socket_was_expected_got_r
#define __pyx_kp_u_A_UNIX_Domain_Stream_Socket_was __pyx_mstate_global->__pyx_kp_u_A_UNIX_Domain_Stream_Socket_was
#define __pyx_kp_u_A_callable_object_or_None_is_exp __pyx_mstate_global->__pyx_kp_u_A_callable_object_or_None_is_exp
#define __pyx_n_s_AbstractChildWatcher __pyx_mstate_global->__pyx_n_s_AbstractChildWatcher
#define __pyx_n_s_AddrInfo __pyx_mstate_global->__pyx_n_s_AddrInfo
#define __pyx_n_s_AddrInfoRequest __pyx_mstate_global->__pyx_n_s_AddrInfoRequest
#define __pyx_n_s_AddrInfoRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_AddrInfoRequest___reduce_cython
#define __pyx_n_s_AddrInfoRequest___setstate_cytho __pyx_mstate_global->__pyx_n_s_AddrInfoRequest___setstate_cytho
#define __pyx_kp_u_AddrInfoRequest_callback_called __pyx_mstate_global->__pyx_kp_u_AddrInfoRequest_callback_called
#define __pyx_n_s_AddrInfo___reduce_cython __pyx_mstate_global->__pyx_n_s_AddrInfo___reduce_cython
#define __pyx_n_s_AddrInfo___setstate_cython __pyx_mstate_global->__pyx_n_s_AddrInfo___setstate_cython
#define __pyx_kp_u_AddrInfo_data_is_NULL __pyx_mstate_global->__pyx_kp_u_AddrInfo_data_is_NULL
#define __pyx_n_s_AddressFamily __pyx_mstate_global->__pyx_n_s_AddressFamily
#define __pyx_kp_u_Address_r_is_already_in_use __pyx_mstate_global->__pyx_kp_u_Address_r_is_already_in_use
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_AttributeError __pyx_mstate_global->__pyx_n_s_AttributeError
#define __pyx_kp_u_Bad_exception_data_from_child_r __pyx_mstate_global->__pyx_kp_u_Bad_exception_data_from_child_r
#define __pyx_n_s_BaseException __pyx_mstate_global->__pyx_n_s_BaseException
#define __pyx_n_s_BaseProtocol __pyx_mstate_global->__pyx_n_s_BaseProtocol
#define __pyx_n_s_BlockingIOError __pyx_mstate_global->__pyx_n_s_BlockingIOError
#define __pyx_n_s_BrokenPipeError __pyx_mstate_global->__pyx_n_s_BrokenPipeError
#define __pyx_n_s_CHANGE __pyx_mstate_global->__pyx_n_s_CHANGE
#define __pyx_kp_u_Callback_handles_8 __pyx_mstate_global->__pyx_kp_u_Callback_handles_8
#define __pyx_n_s_CancelledError __pyx_mstate_global->__pyx_n_s_CancelledError
#define __pyx_kp_u_Cannot_call_write_after_write_eo __pyx_mstate_global->__pyx_kp_u_Cannot_call_write_after_write_eo
#define __pyx_kp_u_Cannot_call_writelines_after_wri __pyx_mstate_global->__pyx_kp_u_Cannot_call_writelines_after_wri
#define __pyx_kp_u_Cannot_close_a_running_event_loo __pyx_mstate_global->__pyx_kp_u_Cannot_close_a_running_event_loo
#define __pyx_kp_u_Cannot_run_the_event_loop_while __pyx_mstate_global->__pyx_kp_u_Cannot_run_the_event_loop_while
#define __pyx_kp_u_Cannot_serialize_socket_object __pyx_mstate_global->__pyx_kp_u_Cannot_serialize_socket_object
#define __pyx_n_s_CertificateError __pyx_mstate_global->__pyx_n_s_CertificateError
#define __pyx_kp_u_Connect_call_failed __pyx_mstate_global->__pyx_kp_u_Connect_call_failed
#define __pyx_n_s_ConnectionAbortedError __pyx_mstate_global->__pyx_n_s_ConnectionAbortedError
#define __pyx_n_s_ConnectionRefusedError __pyx_mstate_global->__pyx_n_s_ConnectionRefusedError
#define __pyx_n_s_ConnectionResetError __pyx_mstate_global->__pyx_n_s_ConnectionResetError
#define __pyx_n_s_Counter __pyx_mstate_global->__pyx_n_s_Counter
#define __pyx_kp_u_Creating__SSLProtocolTransport_t __pyx_mstate_global->__pyx_kp_u_Creating__SSLProtocolTransport_t
#define __pyx_n_s_DEVNULL __pyx_mstate_global->__pyx_n_s_DEVNULL
#define __pyx_n_s_DeprecationWarning __pyx_mstate_global->__pyx_n_s_DeprecationWarning
#define __pyx_n_u_EAI_ADDRFAMILY __pyx_mstate_global->__pyx_n_u_EAI_ADDRFAMILY
#define __pyx_n_u_EAI_AGAIN __pyx_mstate_global->__pyx_n_u_EAI_AGAIN
#define __pyx_n_u_EAI_BADFLAGS __pyx_mstate_global->__pyx_n_u_EAI_BADFLAGS
#define __pyx_n_u_EAI_BADHINTS __pyx_mstate_global->__pyx_n_u_EAI_BADHINTS
#define __pyx_n_u_EAI_CANCELED __pyx_mstate_global->__pyx_n_u_EAI_CANCELED
#define __pyx_n_u_EAI_FAIL __pyx_mstate_global->__pyx_n_u_EAI_FAIL
#define __pyx_n_u_EAI_FAMILY __pyx_mstate_global->__pyx_n_u_EAI_FAMILY
#define __pyx_n_u_EAI_MEMORY __pyx_mstate_global->__pyx_n_u_EAI_MEMORY
#define __pyx_n_u_EAI_NODATA __pyx_mstate_global->__pyx_n_u_EAI_NODATA
#define __pyx_n_u_EAI_NONAME __pyx_mstate_global->__pyx_n_u_EAI_NONAME
#define __pyx_n_u_EAI_OVERFLOW __pyx_mstate_global->__pyx_n_u_EAI_OVERFLOW
#define __pyx_n_u_EAI_PROTOCOL __pyx_mstate_global->__pyx_n_u_EAI_PROTOCOL
#define __pyx_n_u_EAI_SERVICE __pyx_mstate_global->__pyx_n_u_EAI_SERVICE
#define __pyx_n_u_EAI_SOCKTYPE __pyx_mstate_global->__pyx_n_u_EAI_SOCKTYPE
#define __pyx_kp_u_Error_calling_eof_received __pyx_mstate_global->__pyx_kp_u_Error_calling_eof_received
#define __pyx_kp_u_Error_occurred_during_shutdown __pyx_mstate_global->__pyx_kp_u_Error_occurred_during_shutdown
#define __pyx_kp_u_Errors_logged __pyx_mstate_global->__pyx_kp_u_Errors_logged
#define __pyx_kp_u_Event_loop_is_closed __pyx_mstate_global->__pyx_kp_u_Event_loop_is_closed
#define __pyx_kp_u_Event_loop_stopped_before_Future __pyx_mstate_global->__pyx_kp_u_Event_loop_stopped_before_Future
#define __pyx_kp_u_Exception_in___repr___r_value_ty __pyx_mstate_global->__pyx_kp_u_Exception_in___repr___r_value_ty
#define __pyx_kp_u_Exception_in_callback __pyx_mstate_global->__pyx_kp_u_Exception_in_callback
#define __pyx_kp_u_Exception_in_default_exception_h __pyx_mstate_global->__pyx_kp_u_Exception_in_default_exception_h
#define __pyx_kp_u_Exception_in_default_exception_h_2 __pyx_mstate_global->__pyx_kp_u_Exception_in_default_exception_h_2
#define __pyx_kp_u_Exception_occurred_in_preexec_fn __pyx_mstate_global->__pyx_kp_u_Exception_occurred_in_preexec_fn
#define __pyx_kp_u_Executing_r_took_3f_seconds __pyx_mstate_global->__pyx_kp_u_Executing_r_took_3f_seconds
#define __pyx_kp_u_Executing_s_took_3f_seconds __pyx_mstate_global->__pyx_kp_u_Executing_s_took_3f_seconds
#define __pyx_kp_u_Executor_shutdown_has_been_calle __pyx_mstate_global->__pyx_kp_u_Executor_shutdown_has_been_calle
#define __pyx_kp_u_Fatal_error_on_SSL_protocol __pyx_mstate_global->__pyx_kp_u_Fatal_error_on_SSL_protocol
#define __pyx_kp_u_Fatal_error_on_server __pyx_mstate_global->__pyx_kp_u_Fatal_error_on_server
#define __pyx_kp_u_Fatal_error_on_transport __pyx_mstate_global->__pyx_kp_u_Fatal_error_on_transport
#define __pyx_kp_u_Fatal_error_on_transport_2 __pyx_mstate_global->__pyx_kp_u_Fatal_error_on_transport_2
#define __pyx_kp_u_Fatal_write_error_on_datagram_tr __pyx_mstate_global->__pyx_kp_u_Fatal_write_error_on_datagram_tr
#define __pyx_n_s_FileExistsError __pyx_mstate_global->__pyx_n_s_FileExistsError
#define __pyx_n_s_FileNotFoundError __pyx_mstate_global->__pyx_n_s_FileNotFoundError
#define __pyx_n_s_FileSystemEvent __pyx_mstate_global->__pyx_n_s_FileSystemEvent
#define __pyx_kp_u_File_descriptor_r_is_used_by_tra __pyx_mstate_global->__pyx_kp_u_File_descriptor_r_is_used_by_tra
#define __pyx_n_s_FlowControlMixin __pyx_mstate_global->__pyx_n_s_FlowControlMixin
#define __pyx_n_s_Future __pyx_mstate_global->__pyx_n_s_Future
#define __pyx_n_s_Handle __pyx_mstate_global->__pyx_n_s_Handle
#define __pyx_n_s_Handle___reduce_cython __pyx_mstate_global->__pyx_n_s_Handle___reduce_cython
#define __pyx_n_s_Handle___setstate_cython __pyx_mstate_global->__pyx_n_s_Handle___setstate_cython
#define __pyx_n_s_Handle_cancel __pyx_mstate_global->__pyx_n_s_Handle_cancel
#define __pyx_n_s_Handle_cancelled __pyx_mstate_global->__pyx_n_s_Handle_cancelled
#define __pyx_kp_u_Handle_loop_is_None_in_Handle __pyx_mstate_global->__pyx_kp_u_Handle_loop_is_None_in_Handle
#define __pyx_n_u_IPV6_V6ONLY __pyx_mstate_global->__pyx_n_u_IPV6_V6ONLY
#define __pyx_kp_u_IPv4_sockaddr_must_be_2_tuple __pyx_mstate_global->__pyx_kp_u_IPv4_sockaddr_must_be_2_tuple
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_n_s_IntEnum __pyx_mstate_global->__pyx_n_s_IntEnum
#define __pyx_n_s_InterruptedError __pyx_mstate_global->__pyx_n_s_InterruptedError
#define __pyx_n_s_InvalidStateError __pyx_mstate_global->__pyx_n_s_InvalidStateError
#define __pyx_kp_u_Invalid_address_must_be_None_or __pyx_mstate_global->__pyx_kp_u_Invalid_address_must_be_None_or
#define __pyx_kp_u_Invalid_file_descriptor __pyx_mstate_global->__pyx_kp_u_Invalid_file_descriptor
#define __pyx_kp_u_Invalid_file_object_r __pyx_mstate_global->__pyx_kp_u_Invalid_file_object_r
#define __pyx_n_s_IsADirectoryError __pyx_mstate_global->__pyx_n_s_IsADirectoryError
#define __pyx_n_s_Iterable __pyx_mstate_global->__pyx_n_s_Iterable
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_KeyboardInterrupt __pyx_mstate_global->__pyx_n_s_KeyboardInterrupt
#define __pyx_kp_u_Listen_errors __pyx_mstate_global->__pyx_kp_u_Listen_errors
#define __pyx_n_s_Loop __pyx_mstate_global->__pyx_n_s_Loop
#define __pyx_n_s_Loop___get___locals_lambda __pyx_mstate_global->__pyx_n_s_Loop___get___locals_lambda
#define __pyx_n_s_Loop___reduce_cython __pyx_mstate_global->__pyx_n_s_Loop___reduce_cython
#define __pyx_n_s_Loop___setstate_cython __pyx_mstate_global->__pyx_n_s_Loop___setstate_cython
#define __pyx_n_s_Loop___sighandler __pyx_mstate_global->__pyx_n_s_Loop___sighandler
#define __pyx_n_s_Loop___subprocess_run __pyx_mstate_global->__pyx_n_s_Loop___subprocess_run
#define __pyx_n_s_Loop__asyncgen_finalizer_hook __pyx_mstate_global->__pyx_n_s_Loop__asyncgen_finalizer_hook
#define __pyx_n_s_Loop__asyncgen_firstiter_hook __pyx_mstate_global->__pyx_n_s_Loop__asyncgen_firstiter_hook
#define __pyx_n_s_Loop__check_default_executor __pyx_mstate_global->__pyx_n_s_Loop__check_default_executor
#define __pyx_n_s_Loop__do_shutdown __pyx_mstate_global->__pyx_n_s_Loop__do_shutdown
#define __pyx_n_s_Loop__get_backend_id __pyx_mstate_global->__pyx_n_s_Loop__get_backend_id
#define __pyx_n_s_Loop__getaddrinfo_locals_callbac __pyx_mstate_global->__pyx_n_s_Loop__getaddrinfo_locals_callbac
#define __pyx_n_s_Loop__getnameinfo_locals_callbac __pyx_mstate_global->__pyx_n_s_Loop__getnameinfo_locals_callbac
#define __pyx_n_s_Loop__monitor_fs __pyx_mstate_global->__pyx_n_s_Loop__monitor_fs
#define __pyx_kp_u_Loop__read_from_self __pyx_mstate_global->__pyx_kp_u_Loop__read_from_self
#define __pyx_n_s_Loop__sighandler __pyx_mstate_global->__pyx_n_s_Loop__sighandler
#define __pyx_kp_u_Loop__sock_accept __pyx_mstate_global->__pyx_kp_u_Loop__sock_accept
#define __pyx_kp_u_Loop__sock_connect __pyx_mstate_global->__pyx_kp_u_Loop__sock_connect
#define __pyx_kp_u_Loop__sock_recv __pyx_mstate_global->__pyx_kp_u_Loop__sock_recv
#define __pyx_kp_u_Loop__sock_recv_into __pyx_mstate_global->__pyx_kp_u_Loop__sock_recv_into
#define __pyx_kp_u_Loop__sock_sendall __pyx_mstate_global->__pyx_kp_u_Loop__sock_sendall
#define __pyx_kp_u_Loop__stop __pyx_mstate_global->__pyx_kp_u_Loop__stop
#define __pyx_n_s_Loop__subprocess_run __pyx_mstate_global->__pyx_n_s_Loop__subprocess_run
#define __pyx_n_s_Loop_add_reader __pyx_mstate_global->__pyx_n_s_Loop_add_reader
#define __pyx_n_s_Loop_add_signal_handler __pyx_mstate_global->__pyx_n_s_Loop_add_signal_handler
#define __pyx_n_s_Loop_add_writer __pyx_mstate_global->__pyx_n_s_Loop_add_writer
#define __pyx_n_s_Loop_call_at __pyx_mstate_global->__pyx_n_s_Loop_call_at
#define __pyx_n_s_Loop_call_exception_handler __pyx_mstate_global->__pyx_n_s_Loop_call_exception_handler
#define __pyx_n_s_Loop_call_later __pyx_mstate_global->__pyx_n_s_Loop_call_later
#define __pyx_n_s_Loop_call_soon __pyx_mstate_global->__pyx_n_s_Loop_call_soon
#define __pyx_n_s_Loop_call_soon_threadsafe __pyx_mstate_global->__pyx_n_s_Loop_call_soon_threadsafe
#define __pyx_n_s_Loop_close __pyx_mstate_global->__pyx_n_s_Loop_close
#define __pyx_n_s_Loop_connect_accepted_socket __pyx_mstate_global->__pyx_n_s_Loop_connect_accepted_socket
#define __pyx_n_s_Loop_connect_read_pipe __pyx_mstate_global->__pyx_n_s_Loop_connect_read_pipe
#define __pyx_n_s_Loop_connect_write_pipe __pyx_mstate_global->__pyx_n_s_Loop_connect_write_pipe
#define __pyx_n_s_Loop_create_connection __pyx_mstate_global->__pyx_n_s_Loop_create_connection
#define __pyx_n_s_Loop_create_datagram_endpoint __pyx_mstate_global->__pyx_n_s_Loop_create_datagram_endpoint
#define __pyx_n_s_Loop_create_future __pyx_mstate_global->__pyx_n_s_Loop_create_future
#define __pyx_n_s_Loop_create_server __pyx_mstate_global->__pyx_n_s_Loop_create_server
#define __pyx_n_s_Loop_create_task __pyx_mstate_global->__pyx_n_s_Loop_create_task
#define __pyx_n_s_Loop_create_unix_connection __pyx_mstate_global->__pyx_n_s_Loop_create_unix_connection
#define __pyx_n_s_Loop_create_unix_server __pyx_mstate_global->__pyx_n_s_Loop_create_unix_server
#define __pyx_kp_u_Loop_debug_info __pyx_mstate_global->__pyx_kp_u_Loop_debug_info
#define __pyx_n_s_Loop_default_exception_handler __pyx_mstate_global->__pyx_n_s_Loop_default_exception_handler
#define __pyx_n_s_Loop_get_debug __pyx_mstate_global->__pyx_n_s_Loop_get_debug
#define __pyx_n_s_Loop_get_exception_handler __pyx_mstate_global->__pyx_n_s_Loop_get_exception_handler
#define __pyx_n_s_Loop_get_task_factory __pyx_mstate_global->__pyx_n_s_Loop_get_task_factory
#define __pyx_n_s_Loop_getaddrinfo __pyx_mstate_global->__pyx_n_s_Loop_getaddrinfo
#define __pyx_n_s_Loop_getnameinfo __pyx_mstate_global->__pyx_n_s_Loop_getnameinfo
#define __pyx_n_s_Loop_is_closed __pyx_mstate_global->__pyx_n_s_Loop_is_closed
#define __pyx_n_s_Loop_is_running __pyx_mstate_global->__pyx_n_s_Loop_is_running
#define __pyx_n_s_Loop_remove_reader __pyx_mstate_global->__pyx_n_s_Loop_remove_reader
#define __pyx_n_s_Loop_remove_signal_handler __pyx_mstate_global->__pyx_n_s_Loop_remove_signal_handler
#define __pyx_n_s_Loop_remove_writer __pyx_mstate_global->__pyx_n_s_Loop_remove_writer
#define __pyx_n_s_Loop_run_forever __pyx_mstate_global->__pyx_n_s_Loop_run_forever
#define __pyx_n_s_Loop_run_in_executor __pyx_mstate_global->__pyx_n_s_Loop_run_in_executor
#define __pyx_n_s_Loop_run_until_complete __pyx_mstate_global->__pyx_n_s_Loop_run_until_complete
#define __pyx_n_s_Loop_set_debug __pyx_mstate_global->__pyx_n_s_Loop_set_debug
#define __pyx_n_s_Loop_set_default_executor __pyx_mstate_global->__pyx_n_s_Loop_set_default_executor
#define __pyx_n_s_Loop_set_exception_handler __pyx_mstate_global->__pyx_n_s_Loop_set_exception_handler
#define __pyx_n_s_Loop_set_task_factory __pyx_mstate_global->__pyx_n_s_Loop_set_task_factory
#define __pyx_n_s_Loop_shutdown_asyncgens __pyx_mstate_global->__pyx_n_s_Loop_shutdown_asyncgens
#define __pyx_n_s_Loop_shutdown_default_executor __pyx_mstate_global->__pyx_n_s_Loop_shutdown_default_executor
#define __pyx_n_s_Loop_sock_accept __pyx_mstate_global->__pyx_n_s_Loop_sock_accept
#define __pyx_n_s_Loop_sock_connect __pyx_mstate_global->__pyx_n_s_Loop_sock_connect
#define __pyx_n_s_Loop_sock_recv __pyx_mstate_global->__pyx_n_s_Loop_sock_recv
#define __pyx_n_s_Loop_sock_recv_into __pyx_mstate_global->__pyx_n_s_Loop_sock_recv_into
#define __pyx_n_s_Loop_sock_recvfrom __pyx_mstate_global->__pyx_n_s_Loop_sock_recvfrom
#define __pyx_n_s_Loop_sock_recvfrom_into __pyx_mstate_global->__pyx_n_s_Loop_sock_recvfrom_into
#define __pyx_n_s_Loop_sock_sendall __pyx_mstate_global->__pyx_n_s_Loop_sock_sendall
#define __pyx_n_s_Loop_sock_sendto __pyx_mstate_global->__pyx_n_s_Loop_sock_sendto
#define __pyx_n_s_Loop_start_tls __pyx_mstate_global->__pyx_n_s_Loop_start_tls
#define __pyx_n_s_Loop_stop __pyx_mstate_global->__pyx_n_s_Loop_stop
#define __pyx_n_s_Loop_subprocess_exec __pyx_mstate_global->__pyx_n_s_Loop_subprocess_exec
#define __pyx_n_s_Loop_subprocess_shell __pyx_mstate_global->__pyx_n_s_Loop_subprocess_shell
#define __pyx_kp_u_Loop_time __pyx_mstate_global->__pyx_kp_u_Loop_time
#define __pyx_n_s_Loop_time_2 __pyx_mstate_global->__pyx_n_s_Loop_time_2
#define __pyx_n_s_LruCache __pyx_mstate_global->__pyx_n_s_LruCache
#define __pyx_n_s_LruCache___reduce_cython __pyx_mstate_global->__pyx_n_s_LruCache___reduce_cython
#define __pyx_n_s_LruCache___setstate_cython __pyx_mstate_global->__pyx_n_s_LruCache___setstate_cython
#define __pyx_n_s_MemoryBIO __pyx_mstate_global->__pyx_n_s_MemoryBIO
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_kp_u_Multiple_exceptions __pyx_mstate_global->__pyx_kp_u_Multiple_exceptions
#define __pyx_n_s_NSIG __pyx_mstate_global->__pyx_n_s_NSIG
#define __pyx_n_s_NameInfoRequest __pyx_mstate_global->__pyx_n_s_NameInfoRequest
#define __pyx_n_s_NameInfoRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_NameInfoRequest___reduce_cython
#define __pyx_n_s_NameInfoRequest___setstate_cytho __pyx_mstate_global->__pyx_n_s_NameInfoRequest___setstate_cytho
#define __pyx_kp_u_Neither_host_port_nor_sock_were __pyx_mstate_global->__pyx_kp_u_Neither_host_port_nor_sock_were
#define __pyx_kp_u_Non_thread_safe_operation_invoke __pyx_mstate_global->__pyx_kp_u_Non_thread_safe_operation_invoke
#define __pyx_n_s_NotImplementedError __pyx_mstate_global->__pyx_n_s_NotImplementedError
#define __pyx_kp_u_Number_of_signals __pyx_mstate_global->__pyx_kp_u_Number_of_signals
#define __pyx_n_s_OSError __pyx_mstate_global->__pyx_n_s_OSError
#define __pyx_n_s_O_RDWR __pyx_mstate_global->__pyx_n_s_O_RDWR
#define __pyx_kp_u_Object_created_at_most_recent_ca __pyx_mstate_global->__pyx_kp_u_Object_created_at_most_recent_ca
#define __pyx_n_s_OrderedDict __pyx_mstate_global->__pyx_n_s_OrderedDict
#define __pyx_n_s_OverflowError __pyx_mstate_global->__pyx_n_s_OverflowError
#define __pyx_n_s_PIPE __pyx_mstate_global->__pyx_n_s_PIPE
#define __pyx_n_u_PYTHONASYNCIODEBUG __pyx_mstate_global->__pyx_n_u_PYTHONASYNCIODEBUG
#define __pyx_kp_u_Passing_reuse_address_True_is_no __pyx_mstate_global->__pyx_kp_u_Passing_reuse_address_True_is_no
#define __pyx_n_s_PermissionError __pyx_mstate_global->__pyx_n_s_PermissionError
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_kp_s_Pickling_of_struct_members_such __pyx_mstate_global->__pyx_kp_s_Pickling_of_struct_members_such
#define __pyx_n_s_PipeConnectRequest __pyx_mstate_global->__pyx_n_s_PipeConnectRequest
#define __pyx_n_s_PipeConnectRequest___reduce_cyt __pyx_mstate_global->__pyx_n_s_PipeConnectRequest___reduce_cyt
#define __pyx_n_s_PipeConnectRequest___setstate_c __pyx_mstate_global->__pyx_n_s_PipeConnectRequest___setstate_c
#define __pyx_kp_u_Polls_debug_info __pyx_mstate_global->__pyx_kp_u_Polls_debug_info
#define __pyx_n_s_ProcessLookupError __pyx_mstate_global->__pyx_n_s_ProcessLookupError
#define __pyx_kp_u_Process_info __pyx_mstate_global->__pyx_kp_u_Process_info
#define __pyx_kp_u_Process_memory __pyx_mstate_global->__pyx_kp_u_Process_memory
#define __pyx_n_s_Protocol __pyx_mstate_global->__pyx_n_s_Protocol
#define __pyx_n_s_PseudoSocket __pyx_mstate_global->__pyx_n_s_PseudoSocket
#define __pyx_n_s_PseudoSocket___enter __pyx_mstate_global->__pyx_n_s_PseudoSocket___enter
#define __pyx_n_s_PseudoSocket___exit __pyx_mstate_global->__pyx_n_s_PseudoSocket___exit
#define __pyx_n_s_PseudoSocket___getstate __pyx_mstate_global->__pyx_n_s_PseudoSocket___getstate
#define __pyx_n_s_PseudoSocket___reduce_cython __pyx_mstate_global->__pyx_n_s_PseudoSocket___reduce_cython
#define __pyx_n_s_PseudoSocket___setstate_cython __pyx_mstate_global->__pyx_n_s_PseudoSocket___setstate_cython
#define __pyx_n_s_PseudoSocket_accept __pyx_mstate_global->__pyx_n_s_PseudoSocket_accept
#define __pyx_n_s_PseudoSocket_bind __pyx_mstate_global->__pyx_n_s_PseudoSocket_bind
#define __pyx_n_s_PseudoSocket_close __pyx_mstate_global->__pyx_n_s_PseudoSocket_close
#define __pyx_n_s_PseudoSocket_connect __pyx_mstate_global->__pyx_n_s_PseudoSocket_connect
#define __pyx_n_s_PseudoSocket_connect_ex __pyx_mstate_global->__pyx_n_s_PseudoSocket_connect_ex
#define __pyx_n_s_PseudoSocket_detach __pyx_mstate_global->__pyx_n_s_PseudoSocket_detach
#define __pyx_n_s_PseudoSocket_dup __pyx_mstate_global->__pyx_n_s_PseudoSocket_dup
#define __pyx_n_s_PseudoSocket_fileno __pyx_mstate_global->__pyx_n_s_PseudoSocket_fileno
#define __pyx_n_s_PseudoSocket_get_inheritable __pyx_mstate_global->__pyx_n_s_PseudoSocket_get_inheritable
#define __pyx_n_s_PseudoSocket_getpeername __pyx_mstate_global->__pyx_n_s_PseudoSocket_getpeername
#define __pyx_n_s_PseudoSocket_getsockname __pyx_mstate_global->__pyx_n_s_PseudoSocket_getsockname
#define __pyx_n_s_PseudoSocket_getsockopt __pyx_mstate_global->__pyx_n_s_PseudoSocket_getsockopt
#define __pyx_n_s_PseudoSocket_gettimeout __pyx_mstate_global->__pyx_n_s_PseudoSocket_gettimeout
#define __pyx_n_s_PseudoSocket_ioctl __pyx_mstate_global->__pyx_n_s_PseudoSocket_ioctl
#define __pyx_n_s_PseudoSocket_listen __pyx_mstate_global->__pyx_n_s_PseudoSocket_listen
#define __pyx_n_s_PseudoSocket_makefile __pyx_mstate_global->__pyx_n_s_PseudoSocket_makefile
#define __pyx_n_s_PseudoSocket_recv __pyx_mstate_global->__pyx_n_s_PseudoSocket_recv
#define __pyx_n_s_PseudoSocket_recv_into __pyx_mstate_global->__pyx_n_s_PseudoSocket_recv_into
#define __pyx_n_s_PseudoSocket_recvfrom __pyx_mstate_global->__pyx_n_s_PseudoSocket_recvfrom
#define __pyx_n_s_PseudoSocket_recvfrom_into __pyx_mstate_global->__pyx_n_s_PseudoSocket_recvfrom_into
#define __pyx_n_s_PseudoSocket_recvmsg __pyx_mstate_global->__pyx_n_s_PseudoSocket_recvmsg
#define __pyx_n_s_PseudoSocket_recvmsg_into __pyx_mstate_global->__pyx_n_s_PseudoSocket_recvmsg_into
#define __pyx_n_s_PseudoSocket_send __pyx_mstate_global->__pyx_n_s_PseudoSocket_send
#define __pyx_n_s_PseudoSocket_sendall __pyx_mstate_global->__pyx_n_s_PseudoSocket_sendall
#define __pyx_n_s_PseudoSocket_sendfile __pyx_mstate_global->__pyx_n_s_PseudoSocket_sendfile
#define __pyx_n_s_PseudoSocket_sendmsg __pyx_mstate_global->__pyx_n_s_PseudoSocket_sendmsg
#define __pyx_n_s_PseudoSocket_sendmsg_afalg __pyx_mstate_global->__pyx_n_s_PseudoSocket_sendmsg_afalg
#define __pyx_n_s_PseudoSocket_sendto __pyx_mstate_global->__pyx_n_s_PseudoSocket_sendto
#define __pyx_n_s_PseudoSocket_set_inheritable __pyx_mstate_global->__pyx_n_s_PseudoSocket_set_inheritable
#define __pyx_n_s_PseudoSocket_setblocking __pyx_mstate_global->__pyx_n_s_PseudoSocket_setblocking
#define __pyx_n_s_PseudoSocket_setsockopt __pyx_mstate_global->__pyx_n_s_PseudoSocket_setsockopt
#define __pyx_n_s_PseudoSocket_settimeout __pyx_mstate_global->__pyx_n_s_PseudoSocket_settimeout
#define __pyx_n_s_PseudoSocket_share __pyx_mstate_global->__pyx_n_s_PseudoSocket_share
#define __pyx_n_s_PseudoSocket_shutdown __pyx_mstate_global->__pyx_n_s_PseudoSocket_shutdown
#define __pyx_n_s_Pyx_CFunc_6uvloop_4loop_4Loop __pyx_mstate_global->__pyx_n_s_Pyx_CFunc_6uvloop_4loop_4Loop
#define __pyx_n_s_RENAME __pyx_mstate_global->__pyx_n_s_RENAME
#define __pyx_n_s_RENAME_CHANGE __pyx_mstate_global->__pyx_n_s_RENAME_CHANGE
#define __pyx_kp_u_Racing_with_another_loop_to_spaw __pyx_mstate_global->__pyx_kp_u_Racing_with_another_loop_to_spaw
#define __pyx_n_s_ReadSubprocessPipeProto __pyx_mstate_global->__pyx_n_s_ReadSubprocessPipeProto
#define __pyx_n_s_ReadSubprocessPipeProto_data_rec __pyx_mstate_global->__pyx_n_s_ReadSubprocessPipeProto_data_rec
#define __pyx_n_s_ReadUnixTransport __pyx_mstate_global->__pyx_n_s_ReadUnixTransport
#define __pyx_n_s_ReadUnixTransport___reduce_cytho __pyx_mstate_global->__pyx_n_s_ReadUnixTransport___reduce_cytho
#define __pyx_n_s_ReadUnixTransport___setstate_cyt __pyx_mstate_global->__pyx_n_s_ReadUnixTransport___setstate_cyt
#define __pyx_n_s_ReadUnixTransport_abort __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_abort
#define __pyx_n_s_ReadUnixTransport_can_write_eof __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_can_write_eof
#define __pyx_n_s_ReadUnixTransport_get_write_buff __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_get_write_buff
#define __pyx_n_s_ReadUnixTransport_get_write_buff_2 __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_get_write_buff_2
#define __pyx_n_s_ReadUnixTransport_set_write_buff __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_set_write_buff
#define __pyx_n_s_ReadUnixTransport_write __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_write
#define __pyx_n_s_ReadUnixTransport_write_eof __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_write_eof
#define __pyx_n_s_ReadUnixTransport_writelines __pyx_mstate_global->__pyx_n_s_ReadUnixTransport_writelines
#define __pyx_kp_u_Read_EOF_failed_callbacks __pyx_mstate_global->__pyx_kp_u_Read_EOF_failed_callbacks
#define __pyx_kp_u_Read_EOFs __pyx_mstate_global->__pyx_kp_u_Read_EOFs
#define __pyx_kp_u_Read_callbacks __pyx_mstate_global->__pyx_kp_u_Read_callbacks
#define __pyx_kp_u_Read_callbacks_failed __pyx_mstate_global->__pyx_kp_u_Read_callbacks_failed
#define __pyx_kp_u_Read_errors __pyx_mstate_global->__pyx_kp_u_Read_errors
#define __pyx_kp_u_Read_events __pyx_mstate_global->__pyx_kp_u_Read_events
#define __pyx_kp_u_Read_failed_callbacks __pyx_mstate_global->__pyx_kp_u_Read_failed_callbacks
#define __pyx_n_s_ResourceWarning __pyx_mstate_global->__pyx_n_s_ResourceWarning
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_n_s_RuntimeWarning __pyx_mstate_global->__pyx_n_s_RuntimeWarning
#define __pyx_n_s_SIG_DFL __pyx_mstate_global->__pyx_n_s_SIG_DFL
#define __pyx_n_u_SOCK_NONBLOCK __pyx_mstate_global->__pyx_n_u_SOCK_NONBLOCK
#define __pyx_n_u_SO_BROADCAST __pyx_mstate_global->__pyx_n_u_SO_BROADCAST
#define __pyx_n_u_SO_REUSEPORT __pyx_mstate_global->__pyx_n_u_SO_REUSEPORT
#define __pyx_n_s_SSLContext __pyx_mstate_global->__pyx_n_s_SSLContext
#define __pyx_n_s_SSLError __pyx_mstate_global->__pyx_n_s_SSLError
#define __pyx_n_s_SSLProtocol __pyx_mstate_global->__pyx_n_s_SSLProtocol
#define __pyx_n_s_SSLProtocolTransport __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport
#define __pyx_n_s_SSLProtocolTransport___reduce_c __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport___reduce_c
#define __pyx_n_s_SSLProtocolTransport___setstate __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport___setstate
#define __pyx_n_s_SSLProtocolTransport__force_clo __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport__force_clo
#define __pyx_n_s_SSLProtocolTransport__test__app __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport__test__app
#define __pyx_n_s_SSLProtocolTransport_abort __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_abort
#define __pyx_n_s_SSLProtocolTransport_can_write __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_can_write
#define __pyx_n_s_SSLProtocolTransport_close __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_close
#define __pyx_n_s_SSLProtocolTransport_get_extra __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_get_extra
#define __pyx_n_s_SSLProtocolTransport_get_protoc __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_get_protoc
#define __pyx_n_s_SSLProtocolTransport_get_read_b __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_get_read_b
#define __pyx_n_s_SSLProtocolTransport_get_read_b_2 __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_get_read_b_2
#define __pyx_n_s_SSLProtocolTransport_get_write __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_get_write
#define __pyx_n_s_SSLProtocolTransport_get_write_2 __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_get_write_2
#define __pyx_n_s_SSLProtocolTransport_is_closing __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_is_closing
#define __pyx_n_s_SSLProtocolTransport_is_reading __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_is_reading
#define __pyx_n_s_SSLProtocolTransport_pause_read __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_pause_read
#define __pyx_n_s_SSLProtocolTransport_resume_rea __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_resume_rea
#define __pyx_n_s_SSLProtocolTransport_set_protoc __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_set_protoc
#define __pyx_n_s_SSLProtocolTransport_set_read_b __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_set_read_b
#define __pyx_n_s_SSLProtocolTransport_set_write __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_set_write
#define __pyx_n_s_SSLProtocolTransport_write __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_write
#define __pyx_n_s_SSLProtocolTransport_write_eof __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_write_eof
#define __pyx_n_s_SSLProtocolTransport_writelines __pyx_mstate_global->__pyx_n_s_SSLProtocolTransport_writelines
#define __pyx_n_s_SSLProtocol___reduce_cython __pyx_mstate_global->__pyx_n_s_SSLProtocol___reduce_cython
#define __pyx_n_s_SSLProtocol___setstate_cython __pyx_mstate_global->__pyx_n_s_SSLProtocol___setstate_cython
#define __pyx_kp_u_SSLProtocol__do_read __pyx_mstate_global->__pyx_kp_u_SSLProtocol__do_read
#define __pyx_n_s_SSLProtocol__get_app_transport __pyx_mstate_global->__pyx_n_s_SSLProtocol__get_app_transport
#define __pyx_n_s_SSLProtocol__start_handshake_loc __pyx_mstate_global->__pyx_n_s_SSLProtocol__start_handshake_loc
#define __pyx_n_s_SSLProtocol__start_shutdown_loca __pyx_mstate_global->__pyx_n_s_SSLProtocol__start_shutdown_loca
#define __pyx_n_s_SSLProtocol_buffer_updated __pyx_mstate_global->__pyx_n_s_SSLProtocol_buffer_updated
#define __pyx_n_s_SSLProtocol_connection_lost __pyx_mstate_global->__pyx_n_s_SSLProtocol_connection_lost
#define __pyx_n_s_SSLProtocol_connection_made __pyx_mstate_global->__pyx_n_s_SSLProtocol_connection_made
#define __pyx_n_s_SSLProtocol_eof_received __pyx_mstate_global->__pyx_n_s_SSLProtocol_eof_received
#define __pyx_n_s_SSLProtocol_get_buffer __pyx_mstate_global->__pyx_n_s_SSLProtocol_get_buffer
#define __pyx_n_s_SSLProtocol_pause_writing __pyx_mstate_global->__pyx_n_s_SSLProtocol_pause_writing
#define __pyx_n_s_SSLProtocol_resume_writing __pyx_mstate_global->__pyx_n_s_SSLProtocol_resume_writing
#define __pyx_n_s_SSLSyscallError __pyx_mstate_global->__pyx_n_s_SSLSyscallError
#define __pyx_n_s_SSLWantReadError __pyx_mstate_global->__pyx_n_s_SSLWantReadError
#define __pyx_n_s_SSLZeroReturnError __pyx_mstate_global->__pyx_n_s_SSLZeroReturnError
#define __pyx_n_s_SSL_ERROR_SYSCALL __pyx_mstate_global->__pyx_n_s_SSL_ERROR_SYSCALL
#define __pyx_n_s_SSL_ERROR_WANT_READ __pyx_mstate_global->__pyx_n_s_SSL_ERROR_WANT_READ
#define __pyx_n_s_SSL_ERROR_WANT_WRITE __pyx_mstate_global->__pyx_n_s_SSL_ERROR_WANT_WRITE
#define __pyx_kp_u_SSL_connection_is_closed __pyx_mstate_global->__pyx_kp_u_SSL_connection_is_closed
#define __pyx_kp_u_SSL_handshake_failed __pyx_mstate_global->__pyx_kp_u_SSL_handshake_failed
#define __pyx_kp_u_SSL_handshake_failed_on_verifyin __pyx_mstate_global->__pyx_kp_u_SSL_handshake_failed_on_verifyin
#define __pyx_kp_u_SSL_handshake_is_taking_longer_t __pyx_mstate_global->__pyx_kp_u_SSL_handshake_is_taking_longer_t
#define __pyx_kp_u_SSL_shutdown_timed_out __pyx_mstate_global->__pyx_kp_u_SSL_shutdown_timed_out
#define __pyx_n_s_STDOUT __pyx_mstate_global->__pyx_n_s_STDOUT
#define __pyx_n_s_S_ISSOCK __pyx_mstate_global->__pyx_n_s_S_ISSOCK
#define __pyx_n_s_Server __pyx_mstate_global->__pyx_n_s_Server
#define __pyx_n_s_Server___aenter __pyx_mstate_global->__pyx_n_s_Server___aenter
#define __pyx_n_s_Server___aexit __pyx_mstate_global->__pyx_n_s_Server___aexit
#define __pyx_n_s_Server___reduce_cython __pyx_mstate_global->__pyx_n_s_Server___reduce_cython
#define __pyx_n_s_Server___setstate_cython __pyx_mstate_global->__pyx_n_s_Server___setstate_cython
#define __pyx_n_s_Server_close __pyx_mstate_global->__pyx_n_s_Server_close
#define __pyx_n_s_Server_get_loop __pyx_mstate_global->__pyx_n_s_Server_get_loop
#define __pyx_n_s_Server_is_serving __pyx_mstate_global->__pyx_n_s_Server_is_serving
#define __pyx_n_s_Server_serve_forever __pyx_mstate_global->__pyx_n_s_Server_serve_forever
#define __pyx_kp_u_Server_side_SSL_needs_a_valid_SS __pyx_mstate_global->__pyx_kp_u_Server_side_SSL_needs_a_valid_SS
#define __pyx_n_s_Server_start_serving __pyx_mstate_global->__pyx_n_s_Server_start_serving
#define __pyx_n_s_Server_wait_closed __pyx_mstate_global->__pyx_n_s_Server_wait_closed
#define __pyx_kp_u_Shutdown_errors __pyx_mstate_global->__pyx_kp_u_Shutdown_errors
#define __pyx_n_s_SockAddrHolder __pyx_mstate_global->__pyx_n_s_SockAddrHolder
#define __pyx_n_s_SockAddrHolder___reduce_cython __pyx_mstate_global->__pyx_n_s_SockAddrHolder___reduce_cython
#define __pyx_n_s_SockAddrHolder___setstate_cython __pyx_mstate_global->__pyx_n_s_SockAddrHolder___setstate_cython
#define __pyx_kp_u_Sock_ops_successful_on_1st_try __pyx_mstate_global->__pyx_kp_u_Sock_ops_successful_on_1st_try
#define __pyx_n_s_SocketKind __pyx_mstate_global->__pyx_n_s_SocketKind
#define __pyx_kp_u_Socket_try_writes __pyx_mstate_global->__pyx_kp_u_Socket_try_writes
#define __pyx_n_s_StackSummary __pyx_mstate_global->__pyx_n_s_StackSummary
#define __pyx_n_s_StreamWriteContext __pyx_mstate_global->__pyx_n_s_StreamWriteContext
#define __pyx_n_s_StreamWriteContext___reduce_cyt __pyx_mstate_global->__pyx_n_s_StreamWriteContext___reduce_cyt
#define __pyx_n_s_StreamWriteContext___setstate_c __pyx_mstate_global->__pyx_n_s_StreamWriteContext___setstate_c
#define __pyx_kp_u_StreamWriteContext_close_py_buf __pyx_mstate_global->__pyx_kp_u_StreamWriteContext_close_py_buf
#define __pyx_kp_u_StreamWriteContext_close_uv_buf __pyx_mstate_global->__pyx_kp_u_StreamWriteContext_close_uv_buf
#define __pyx_kp_u_Streams_debug_info __pyx_mstate_global->__pyx_kp_u_Streams_debug_info
#define __pyx_n_s_SubprocessError __pyx_mstate_global->__pyx_n_s_SubprocessError
#define __pyx_n_s_SyncSocketReaderFuture __pyx_mstate_global->__pyx_n_s_SyncSocketReaderFuture
#define __pyx_n_s_SyncSocketReaderFuture___init __pyx_mstate_global->__pyx_n_s_SyncSocketReaderFuture___init
#define __pyx_n_s_SyncSocketReaderFuture___remove __pyx_mstate_global->__pyx_n_s_SyncSocketReaderFuture___remove
#define __pyx_n_s_SyncSocketReaderFuture__loop __pyx_mstate_global->__pyx_n_s_SyncSocketReaderFuture__loop
#define __pyx_n_s_SyncSocketReaderFuture__remove __pyx_mstate_global->__pyx_n_s_SyncSocketReaderFuture__remove
#define __pyx_n_s_SyncSocketReaderFuture__sock __pyx_mstate_global->__pyx_n_s_SyncSocketReaderFuture__sock
#define __pyx_n_s_SyncSocketReaderFuture_cancel __pyx_mstate_global->__pyx_n_s_SyncSocketReaderFuture_cancel
#define __pyx_n_s_SyncSocketWriterFuture __pyx_mstate_global->__pyx_n_s_SyncSocketWriterFuture
#define __pyx_n_s_SyncSocketWriterFuture___init __pyx_mstate_global->__pyx_n_s_SyncSocketWriterFuture___init
#define __pyx_n_s_SyncSocketWriterFuture___remove __pyx_mstate_global->__pyx_n_s_SyncSocketWriterFuture___remove
#define __pyx_n_s_SyncSocketWriterFuture__loop __pyx_mstate_global->__pyx_n_s_SyncSocketWriterFuture__loop
#define __pyx_n_s_SyncSocketWriterFuture__remove __pyx_mstate_global->__pyx_n_s_SyncSocketWriterFuture__remove
#define __pyx_n_s_SyncSocketWriterFuture__sock __pyx_mstate_global->__pyx_n_s_SyncSocketWriterFuture__sock
#define __pyx_n_s_SyncSocketWriterFuture_cancel __pyx_mstate_global->__pyx_n_s_SyncSocketWriterFuture_cancel
#define __pyx_n_s_SystemExit __pyx_mstate_global->__pyx_n_s_SystemExit
#define __pyx_n_s_TCPConnectRequest __pyx_mstate_global->__pyx_n_s_TCPConnectRequest
#define __pyx_n_s_TCPConnectRequest___reduce_cyth __pyx_mstate_global->__pyx_n_s_TCPConnectRequest___reduce_cyth
#define __pyx_n_s_TCPConnectRequest___setstate_cy __pyx_mstate_global->__pyx_n_s_TCPConnectRequest___setstate_cy
#define __pyx_n_s_TCPServer __pyx_mstate_global->__pyx_n_s_TCPServer
#define __pyx_n_s_TCPServer___reduce_cython __pyx_mstate_global->__pyx_n_s_TCPServer___reduce_cython
#define __pyx_n_s_TCPServer___setstate_cython __pyx_mstate_global->__pyx_n_s_TCPServer___setstate_cython
#define __pyx_n_s_TCPTransport __pyx_mstate_global->__pyx_n_s_TCPTransport
#define __pyx_n_s_TCPTransport___reduce_cython __pyx_mstate_global->__pyx_n_s_TCPTransport___reduce_cython
#define __pyx_n_s_TCPTransport___setstate_cython __pyx_mstate_global->__pyx_n_s_TCPTransport___setstate_cython
#define __pyx_n_s_TCPTransport_get_extra_info __pyx_mstate_global->__pyx_n_s_TCPTransport_get_extra_info
#define __pyx_n_s_Task __pyx_mstate_global->__pyx_n_s_Task
#define __pyx_kp_u_The_executor_did_not_finishing_j __pyx_mstate_global->__pyx_kp_u_The_executor_did_not_finishing_j
#define __pyx_kp_u_The_reuse_address_parameter_has __pyx_mstate_global->__pyx_kp_u_The_reuse_address_parameter_has
#define __pyx_n_s_Thread __pyx_mstate_global->__pyx_n_s_Thread
#define __pyx_n_s_ThreadPoolExecutor __pyx_mstate_global->__pyx_n_s_ThreadPoolExecutor
#define __pyx_n_s_TimeoutError __pyx_mstate_global->__pyx_n_s_TimeoutError
#define __pyx_n_s_TimerHandle __pyx_mstate_global->__pyx_n_s_TimerHandle
#define __pyx_n_s_TimerHandle___reduce_cython __pyx_mstate_global->__pyx_n_s_TimerHandle___reduce_cython
#define __pyx_n_s_TimerHandle___setstate_cython __pyx_mstate_global->__pyx_n_s_TimerHandle___setstate_cython
#define __pyx_n_s_TimerHandle_cancel __pyx_mstate_global->__pyx_n_s_TimerHandle_cancel
#define __pyx_n_s_TimerHandle_cancelled __pyx_mstate_global->__pyx_n_s_TimerHandle_cancelled
#define __pyx_n_s_TimerHandle_when __pyx_mstate_global->__pyx_n_s_TimerHandle_when
#define __pyx_kp_u_Timer_handles_8 __pyx_mstate_global->__pyx_kp_u_Timer_handles_8
#define __pyx_n_s_Transport __pyx_mstate_global->__pyx_n_s_Transport
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_UDPSendContext __pyx_mstate_global->__pyx_n_s_UDPSendContext
#define __pyx_n_s_UDPSendContext___reduce_cython __pyx_mstate_global->__pyx_n_s_UDPSendContext___reduce_cython
#define __pyx_n_s_UDPSendContext___setstate_cytho __pyx_mstate_global->__pyx_n_s_UDPSendContext___setstate_cytho
#define __pyx_n_s_UDPTransport __pyx_mstate_global->__pyx_n_s_UDPTransport
#define __pyx_n_s_UDPTransport___reduce_cython __pyx_mstate_global->__pyx_n_s_UDPTransport___reduce_cython
#define __pyx_n_s_UDPTransport___setstate_cython __pyx_mstate_global->__pyx_n_s_UDPTransport___setstate_cython
#define __pyx_kp_u_UDPTransport_family_is_undefined __pyx_mstate_global->__pyx_kp_u_UDPTransport_family_is_undefined
#define __pyx_kp_u_UDPTransport_family_is_undefined_2 __pyx_mstate_global->__pyx_kp_u_UDPTransport_family_is_undefined_2
#define __pyx_n_s_UDPTransport_sendto __pyx_mstate_global->__pyx_n_s_UDPTransport_sendto
#define __pyx_n_s_UVAsync __pyx_mstate_global->__pyx_n_s_UVAsync
#define __pyx_n_s_UVAsync___reduce_cython __pyx_mstate_global->__pyx_n_s_UVAsync___reduce_cython
#define __pyx_n_s_UVAsync___setstate_cython __pyx_mstate_global->__pyx_n_s_UVAsync___setstate_cython
#define __pyx_n_s_UVBaseTransport __pyx_mstate_global->__pyx_n_s_UVBaseTransport
#define __pyx_n_s_UVBaseTransport___reduce_cython __pyx_mstate_global->__pyx_n_s_UVBaseTransport___reduce_cython
#define __pyx_n_s_UVBaseTransport___setstate_cytho __pyx_mstate_global->__pyx_n_s_UVBaseTransport___setstate_cytho
#define __pyx_n_s_UVBaseTransport__force_close __pyx_mstate_global->__pyx_n_s_UVBaseTransport__force_close
#define __pyx_n_s_UVBaseTransport_abort __pyx_mstate_global->__pyx_n_s_UVBaseTransport_abort
#define __pyx_n_s_UVBaseTransport_close __pyx_mstate_global->__pyx_n_s_UVBaseTransport_close
#define __pyx_n_s_UVBaseTransport_get_extra_info __pyx_mstate_global->__pyx_n_s_UVBaseTransport_get_extra_info
#define __pyx_n_s_UVBaseTransport_get_protocol __pyx_mstate_global->__pyx_n_s_UVBaseTransport_get_protocol
#define __pyx_n_s_UVBaseTransport_get_write_buffer __pyx_mstate_global->__pyx_n_s_UVBaseTransport_get_write_buffer
#define __pyx_n_s_UVBaseTransport_get_write_buffer_2 __pyx_mstate_global->__pyx_n_s_UVBaseTransport_get_write_buffer_2
#define __pyx_n_s_UVBaseTransport_is_closing __pyx_mstate_global->__pyx_n_s_UVBaseTransport_is_closing
#define __pyx_n_s_UVBaseTransport_set_protocol __pyx_mstate_global->__pyx_n_s_UVBaseTransport_set_protocol
#define __pyx_n_s_UVBaseTransport_set_write_buffer __pyx_mstate_global->__pyx_n_s_UVBaseTransport_set_write_buffer
#define __pyx_n_s_UVCheck __pyx_mstate_global->__pyx_n_s_UVCheck
#define __pyx_n_s_UVCheck___reduce_cython __pyx_mstate_global->__pyx_n_s_UVCheck___reduce_cython
#define __pyx_n_s_UVCheck___setstate_cython __pyx_mstate_global->__pyx_n_s_UVCheck___setstate_cython
#define __pyx_n_s_UVFSEvent __pyx_mstate_global->__pyx_n_s_UVFSEvent
#define __pyx_n_s_UVFSEvent___reduce_cython __pyx_mstate_global->__pyx_n_s_UVFSEvent___reduce_cython
#define __pyx_n_s_UVFSEvent___setstate_cython __pyx_mstate_global->__pyx_n_s_UVFSEvent___setstate_cython
#define __pyx_n_s_UVFSEvent_cancel __pyx_mstate_global->__pyx_n_s_UVFSEvent_cancel
#define __pyx_n_s_UVFSEvent_cancelled __pyx_mstate_global->__pyx_n_s_UVFSEvent_cancelled
#define __pyx_n_s_UVHandle __pyx_mstate_global->__pyx_n_s_UVHandle
#define __pyx_n_s_UVHandle___reduce_cython __pyx_mstate_global->__pyx_n_s_UVHandle___reduce_cython
#define __pyx_n_s_UVHandle___setstate_cython __pyx_mstate_global->__pyx_n_s_UVHandle___setstate_cython
#define __pyx_kp_u_UVHandles_python_libuv_total __pyx_mstate_global->__pyx_kp_u_UVHandles_python_libuv_total
#define __pyx_n_s_UVIdle __pyx_mstate_global->__pyx_n_s_UVIdle
#define __pyx_n_s_UVIdle___reduce_cython __pyx_mstate_global->__pyx_n_s_UVIdle___reduce_cython
#define __pyx_n_s_UVIdle___setstate_cython __pyx_mstate_global->__pyx_n_s_UVIdle___setstate_cython
#define __pyx_n_s_UVPoll __pyx_mstate_global->__pyx_n_s_UVPoll
#define __pyx_n_s_UVPoll___reduce_cython __pyx_mstate_global->__pyx_n_s_UVPoll___reduce_cython
#define __pyx_n_s_UVPoll___setstate_cython __pyx_mstate_global->__pyx_n_s_UVPoll___setstate_cython
#define __pyx_n_s_UVProcess __pyx_mstate_global->__pyx_n_s_UVProcess
#define __pyx_n_s_UVProcessTransport __pyx_mstate_global->__pyx_n_s_UVProcessTransport
#define __pyx_n_s_UVProcessTransport___reduce_cyth __pyx_mstate_global->__pyx_n_s_UVProcessTransport___reduce_cyth
#define __pyx_n_s_UVProcessTransport___setstate_cy __pyx_mstate_global->__pyx_n_s_UVProcessTransport___setstate_cy
#define __pyx_n_s_UVProcessTransport___stdio_inite __pyx_mstate_global->__pyx_n_s_UVProcessTransport___stdio_inite
#define __pyx_kp_u_UVProcessTransport__call_connect __pyx_mstate_global->__pyx_kp_u_UVProcessTransport__call_connect
#define __pyx_n_s_UVProcessTransport__stdio_inite __pyx_mstate_global->__pyx_n_s_UVProcessTransport__stdio_inite
#define __pyx_n_s_UVProcessTransport__wait __pyx_mstate_global->__pyx_n_s_UVProcessTransport__wait
#define __pyx_n_s_UVProcessTransport_close __pyx_mstate_global->__pyx_n_s_UVProcessTransport_close
#define __pyx_n_s_UVProcessTransport_get_extra_inf __pyx_mstate_global->__pyx_n_s_UVProcessTransport_get_extra_inf
#define __pyx_n_s_UVProcessTransport_get_pid __pyx_mstate_global->__pyx_n_s_UVProcessTransport_get_pid
#define __pyx_n_s_UVProcessTransport_get_pipe_tran __pyx_mstate_global->__pyx_n_s_UVProcessTransport_get_pipe_tran
#define __pyx_n_s_UVProcessTransport_get_protocol __pyx_mstate_global->__pyx_n_s_UVProcessTransport_get_protocol
#define __pyx_n_s_UVProcessTransport_get_returncod __pyx_mstate_global->__pyx_n_s_UVProcessTransport_get_returncod
#define __pyx_n_s_UVProcessTransport_is_closing __pyx_mstate_global->__pyx_n_s_UVProcessTransport_is_closing
#define __pyx_n_s_UVProcessTransport_kill __pyx_mstate_global->__pyx_n_s_UVProcessTransport_kill
#define __pyx_n_s_UVProcessTransport_send_signal __pyx_mstate_global->__pyx_n_s_UVProcessTransport_send_signal
#define __pyx_n_s_UVProcessTransport_set_protocol __pyx_mstate_global->__pyx_n_s_UVProcessTransport_set_protocol
#define __pyx_n_s_UVProcessTransport_terminate __pyx_mstate_global->__pyx_n_s_UVProcessTransport_terminate
#define __pyx_n_s_UVProcess___reduce_cython __pyx_mstate_global->__pyx_n_s_UVProcess___reduce_cython
#define __pyx_n_s_UVProcess___setstate_cython __pyx_mstate_global->__pyx_n_s_UVProcess___setstate_cython
#define __pyx_kp_u_UVProcess__close_after_spawn_cal __pyx_mstate_global->__pyx_kp_u_UVProcess__close_after_spawn_cal
#define __pyx_n_s_UVRequest __pyx_mstate_global->__pyx_n_s_UVRequest
#define __pyx_n_s_UVRequest___reduce_cython __pyx_mstate_global->__pyx_n_s_UVRequest___reduce_cython
#define __pyx_n_s_UVRequest___setstate_cython __pyx_mstate_global->__pyx_n_s_UVRequest___setstate_cython
#define __pyx_n_s_UVSocketHandle __pyx_mstate_global->__pyx_n_s_UVSocketHandle
#define __pyx_n_s_UVSocketHandle___reduce_cython __pyx_mstate_global->__pyx_n_s_UVSocketHandle___reduce_cython
#define __pyx_n_s_UVSocketHandle___setstate_cython __pyx_mstate_global->__pyx_n_s_UVSocketHandle___setstate_cython
#define __pyx_n_s_UVStream __pyx_mstate_global->__pyx_n_s_UVStream
#define __pyx_n_s_UVStreamServer __pyx_mstate_global->__pyx_n_s_UVStreamServer
#define __pyx_n_s_UVStreamServer___on_ssl_connecte __pyx_mstate_global->__pyx_n_s_UVStreamServer___on_ssl_connecte
#define __pyx_n_s_UVStreamServer___reduce_cython __pyx_mstate_global->__pyx_n_s_UVStreamServer___reduce_cython
#define __pyx_n_s_UVStreamServer___setstate_cython __pyx_mstate_global->__pyx_n_s_UVStreamServer___setstate_cython
#define __pyx_n_s_UVStreamServer__on_ssl_connecte __pyx_mstate_global->__pyx_n_s_UVStreamServer__on_ssl_connecte
#define __pyx_n_s_UVStream___reduce_cython __pyx_mstate_global->__pyx_n_s_UVStream___reduce_cython
#define __pyx_n_s_UVStream___setstate_cython __pyx_mstate_global->__pyx_n_s_UVStream___setstate_cython
#define __pyx_kp_u_UVStream__try_write_called_with __pyx_mstate_global->__pyx_kp_u_UVStream__try_write_called_with
#define __pyx_n_s_UVStream_can_write_eof __pyx_mstate_global->__pyx_n_s_UVStream_can_write_eof
#define __pyx_n_s_UVStream_is_reading __pyx_mstate_global->__pyx_n_s_UVStream_is_reading
#define __pyx_n_s_UVStream_pause_reading __pyx_mstate_global->__pyx_n_s_UVStream_pause_reading
#define __pyx_n_s_UVStream_resume_reading __pyx_mstate_global->__pyx_n_s_UVStream_resume_reading
#define __pyx_kp_u_UVStream_shutdown_callback_calle __pyx_mstate_global->__pyx_kp_u_UVStream_shutdown_callback_calle
#define __pyx_n_s_UVStream_write __pyx_mstate_global->__pyx_n_s_UVStream_write
#define __pyx_kp_u_UVStream_write_callback_called_w __pyx_mstate_global->__pyx_kp_u_UVStream_write_callback_called_w
#define __pyx_n_s_UVStream_write_eof __pyx_mstate_global->__pyx_n_s_UVStream_write_eof
#define __pyx_n_s_UVStream_writelines __pyx_mstate_global->__pyx_n_s_UVStream_writelines
#define __pyx_n_s_UVTimer __pyx_mstate_global->__pyx_n_s_UVTimer
#define __pyx_n_s_UVTimer___reduce_cython __pyx_mstate_global->__pyx_n_s_UVTimer___reduce_cython
#define __pyx_n_s_UVTimer___setstate_cython __pyx_mstate_global->__pyx_n_s_UVTimer___setstate_cython
#define __pyx_kp_u_UVTransport__call_connection_los __pyx_mstate_global->__pyx_kp_u_UVTransport__call_connection_los
#define __pyx_kp_u_UVTransport__call_connection_mad __pyx_mstate_global->__pyx_kp_u_UVTransport__call_connection_mad
#define __pyx_kp_u_Unable_to_check_or_remove_stale __pyx_mstate_global->__pyx_kp_u_Unable_to_check_or_remove_stale
#define __pyx_kp_u_Unable_to_clean_up_listening_UNI __pyx_mstate_global->__pyx_kp_u_Unable_to_clean_up_listening_UNI
#define __pyx_kp_u_Unhandled_error_in_exception_han __pyx_mstate_global->__pyx_kp_u_Unhandled_error_in_exception_han
#define __pyx_kp_u_Unhandled_exception_in_event_loo __pyx_mstate_global->__pyx_kp_u_Unhandled_exception_in_event_loo
#define __pyx_n_s_UnicodeEncodeError __pyx_mstate_global->__pyx_n_s_UnicodeEncodeError
#define __pyx_n_s_UnixServer __pyx_mstate_global->__pyx_n_s_UnixServer
#define __pyx_n_s_UnixServer___reduce_cython __pyx_mstate_global->__pyx_n_s_UnixServer___reduce_cython
#define __pyx_n_s_UnixServer___setstate_cython __pyx_mstate_global->__pyx_n_s_UnixServer___setstate_cython
#define __pyx_n_s_UnixTransport __pyx_mstate_global->__pyx_n_s_UnixTransport
#define __pyx_n_s_UnixTransport___reduce_cython __pyx_mstate_global->__pyx_n_s_UnixTransport___reduce_cython
#define __pyx_n_s_UnixTransport___setstate_cython __pyx_mstate_global->__pyx_n_s_UnixTransport___setstate_cython
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s_WeakSet __pyx_mstate_global->__pyx_n_s_WeakSet
#define __pyx_n_s_WeakValueDictionary __pyx_mstate_global->__pyx_n_s_WeakValueDictionary
#define __pyx_n_s_WriteSubprocessPipeProto __pyx_mstate_global->__pyx_n_s_WriteSubprocessPipeProto
#define __pyx_n_s_WriteSubprocessPipeProto___init __pyx_mstate_global->__pyx_n_s_WriteSubprocessPipeProto___init
#define __pyx_n_s_WriteSubprocessPipeProto___repr __pyx_mstate_global->__pyx_n_s_WriteSubprocessPipeProto___repr
#define __pyx_n_s_WriteSubprocessPipeProto_connect __pyx_mstate_global->__pyx_n_s_WriteSubprocessPipeProto_connect
#define __pyx_n_s_WriteSubprocessPipeProto_connect_2 __pyx_mstate_global->__pyx_n_s_WriteSubprocessPipeProto_connect_2
#define __pyx_n_s_WriteSubprocessPipeProto_pause_w __pyx_mstate_global->__pyx_n_s_WriteSubprocessPipeProto_pause_w
#define __pyx_n_s_WriteSubprocessPipeProto_resume __pyx_mstate_global->__pyx_n_s_WriteSubprocessPipeProto_resume
#define __pyx_n_s_WriteUnixTransport __pyx_mstate_global->__pyx_n_s_WriteUnixTransport
#define __pyx_n_s_WriteUnixTransport___reduce_cyth __pyx_mstate_global->__pyx_n_s_WriteUnixTransport___reduce_cyth
#define __pyx_n_s_WriteUnixTransport___setstate_cy __pyx_mstate_global->__pyx_n_s_WriteUnixTransport___setstate_cy
#define __pyx_n_s_WriteUnixTransport_pause_reading __pyx_mstate_global->__pyx_n_s_WriteUnixTransport_pause_reading
#define __pyx_n_s_WriteUnixTransport_resume_readin __pyx_mstate_global->__pyx_n_s_WriteUnixTransport_resume_readin
#define __pyx_kp_u_Write_callbacks_failed __pyx_mstate_global->__pyx_kp_u_Write_callbacks_failed
#define __pyx_kp_u_Write_contexts_8 __pyx_mstate_global->__pyx_kp_u_Write_contexts_8
#define __pyx_kp_u_Write_errors __pyx_mstate_global->__pyx_kp_u_Write_errors
#define __pyx_kp_u_Write_events __pyx_mstate_global->__pyx_kp_u_Write_events
#define __pyx_kp_u_Write_failed_callbacks __pyx_mstate_global->__pyx_kp_u_Write_failed_callbacks
#define __pyx_kp_u_Write_without_poll __pyx_mstate_global->__pyx_kp_u_Write_without_poll
#define __pyx_kp_u_XXX_an_allocated_buffer_in_trans __pyx_mstate_global->__pyx_kp_u_XXX_an_allocated_buffer_in_trans
#define __pyx_kp_u_You_must_set_server_hostname_whe __pyx_mstate_global->__pyx_kp_u_You_must_set_server_hostname_whe
#define __pyx_n_s__10 __pyx_mstate_global->__pyx_n_s__10
#define __pyx_kp_b__10 __pyx_mstate_global->__pyx_kp_b__10
#define __pyx_kp_u__10 __pyx_mstate_global->__pyx_kp_u__10
#define __pyx_kp_u__115 __pyx_mstate_global->__pyx_kp_u__115
#define __pyx_kp_u__128 __pyx_mstate_global->__pyx_kp_u__128
#define __pyx_kp_u__129 __pyx_mstate_global->__pyx_kp_u__129
#define __pyx_kp_u__130 __pyx_mstate_global->__pyx_kp_u__130
#define __pyx_kp_u__140 __pyx_mstate_global->__pyx_kp_u__140
#define __pyx_kp_u__141 __pyx_mstate_global->__pyx_kp_u__141
#define __pyx_kp_b__161 __pyx_mstate_global->__pyx_kp_b__161
#define __pyx_kp_b__170 __pyx_mstate_global->__pyx_kp_b__170
#define __pyx_kp_u__196 __pyx_mstate_global->__pyx_kp_u__196
#define __pyx_n_s__198 __pyx_mstate_global->__pyx_n_s__198
#define __pyx_n_s__263 __pyx_mstate_global->__pyx_n_s__263
#define __pyx_kp_u__33 __pyx_mstate_global->__pyx_kp_u__33
#define __pyx_n_s__542 __pyx_mstate_global->__pyx_n_s__542
#define __pyx_kp_u__61 __pyx_mstate_global->__pyx_kp_u__61
#define __pyx_kp_u__69 __pyx_mstate_global->__pyx_kp_u__69
#define __pyx_kp_u__75 __pyx_mstate_global->__pyx_kp_u__75
#define __pyx_kp_u__78 __pyx_mstate_global->__pyx_kp_u__78
#define __pyx_kp_u__79 __pyx_mstate_global->__pyx_kp_u__79
#define __pyx_n_s_abc __pyx_mstate_global->__pyx_n_s_abc
#define __pyx_n_s_abort __pyx_mstate_global->__pyx_n_s_abort
#define __pyx_kp_u_abort_init__closed_is_set __pyx_mstate_global->__pyx_kp_u_abort_init__closed_is_set
#define __pyx_kp_u_abort_init__inited_is_set __pyx_mstate_global->__pyx_kp_u_abort_init__inited_is_set
#define __pyx_n_s_accept __pyx_mstate_global->__pyx_n_s_accept
#define __pyx_kp_u_accept_method __pyx_mstate_global->__pyx_kp_u_accept_method
#define __pyx_n_s_aclose __pyx_mstate_global->__pyx_n_s_aclose
#define __pyx_kp_u_active_TimerHandle_is_deallacati __pyx_mstate_global->__pyx_kp_u_active_TimerHandle_is_deallacati
#define __pyx_n_s_add __pyx_mstate_global->__pyx_n_s_add
#define __pyx_n_s_add_done_callback __pyx_mstate_global->__pyx_n_s_add_done_callback
#define __pyx_n_s_add_reader __pyx_mstate_global->__pyx_n_s_add_reader
#define __pyx_n_s_add_signal_handler __pyx_mstate_global->__pyx_n_s_add_signal_handler
#define __pyx_kp_u_add_signal_handler_can_only_be_c __pyx_mstate_global->__pyx_kp_u_add_signal_handler_can_only_be_c
#define __pyx_n_s_add_writer __pyx_mstate_global->__pyx_n_s_add_writer
#define __pyx_n_s_addr __pyx_mstate_global->__pyx_n_s_addr
#define __pyx_n_s_address __pyx_mstate_global->__pyx_n_s_address
#define __pyx_n_s_addrinfo __pyx_mstate_global->__pyx_n_s_addrinfo
#define __pyx_n_s_addrs __pyx_mstate_global->__pyx_n_s_addrs
#define __pyx_n_s_aenter __pyx_mstate_global->__pyx_n_s_aenter
#define __pyx_n_s_aexit __pyx_mstate_global->__pyx_n_s_aexit
#define __pyx_n_s_ag __pyx_mstate_global->__pyx_n_s_ag
#define __pyx_n_s_agen __pyx_mstate_global->__pyx_n_s_agen
#define __pyx_n_s_ai __pyx_mstate_global->__pyx_n_s_ai
#define __pyx_n_s_ai_cnt __pyx_mstate_global->__pyx_n_s_ai_cnt
#define __pyx_n_s_ai_local __pyx_mstate_global->__pyx_n_s_ai_local
#define __pyx_n_s_ai_remote __pyx_mstate_global->__pyx_n_s_ai_remote
#define __pyx_kp_u_alive_closed __pyx_mstate_global->__pyx_kp_u_alive_closed
#define __pyx_kp_u_all_args_must_be_str_or_bytes __pyx_mstate_global->__pyx_kp_u_all_args_must_be_str_or_bytes
#define __pyx_kp_u_all_environment_values_must_be_b __pyx_mstate_global->__pyx_kp_u_all_environment_values_must_be_b
#define __pyx_kp_u_all_environment_vars_must_be_byt __pyx_mstate_global->__pyx_kp_u_all_environment_vars_must_be_byt
#define __pyx_n_s_allow_broadcast __pyx_mstate_global->__pyx_n_s_allow_broadcast
#define __pyx_kp_u_an_error_occurred_during_closing __pyx_mstate_global->__pyx_kp_u_an_error_occurred_during_closing
#define __pyx_n_s_app_protocol __pyx_mstate_global->__pyx_n_s_app_protocol
#define __pyx_n_s_app_transport __pyx_mstate_global->__pyx_n_s_app_transport
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_u_ascii __pyx_mstate_global->__pyx_n_u_ascii
#define __pyx_n_u_asyncgen __pyx_mstate_global->__pyx_n_u_asyncgen
#define __pyx_n_s_asyncgen_finalizer_hook __pyx_mstate_global->__pyx_n_s_asyncgen_finalizer_hook
#define __pyx_n_s_asyncgen_firstiter_hook __pyx_mstate_global->__pyx_n_s_asyncgen_firstiter_hook
#define __pyx_kp_u_asynchronous_generator_r_was_sch __pyx_mstate_global->__pyx_kp_u_asynchronous_generator_r_was_sch
#define __pyx_n_s_asyncio __pyx_mstate_global->__pyx_n_s_asyncio
#define __pyx_kp_s_asyncio_Handle __pyx_mstate_global->__pyx_kp_s_asyncio_Handle
#define __pyx_n_s_asyncio_base_events __pyx_mstate_global->__pyx_n_s_asyncio_base_events
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_asyncio_futures __pyx_mstate_global->__pyx_n_s_asyncio_futures
#define __pyx_kp_u_asyncio_is_trying_to_install_it __pyx_mstate_global->__pyx_kp_u_asyncio_is_trying_to_install_it
#define __pyx_n_s_asyncio_log __pyx_mstate_global->__pyx_n_s_asyncio_log
#define __pyx_n_s_asyncio_sslproto __pyx_mstate_global->__pyx_n_s_asyncio_sslproto
#define __pyx_n_s_asyncio_tasks __pyx_mstate_global->__pyx_n_s_asyncio_tasks
#define __pyx_n_s_asyncio_transports __pyx_mstate_global->__pyx_n_s_asyncio_transports
#define __pyx_n_s_await __pyx_mstate_global->__pyx_n_s_await
#define __pyx_n_u_b __pyx_mstate_global->__pyx_n_u_b
#define __pyx_n_s_backlog __pyx_mstate_global->__pyx_n_s_backlog
#define __pyx_n_s_base_events __pyx_mstate_global->__pyx_n_s_base_events
#define __pyx_kp_b_bin_sh __pyx_mstate_global->__pyx_kp_b_bin_sh
#define __pyx_n_s_bind __pyx_mstate_global->__pyx_n_s_bind
#define __pyx_kp_u_bind_method __pyx_mstate_global->__pyx_kp_u_bind_method
#define __pyx_n_s_buf __pyx_mstate_global->__pyx_n_s_buf
#define __pyx_kp_u_buffer_size_is_not_0_after_a_su __pyx_mstate_global->__pyx_kp_u_buffer_size_is_not_0_after_a_su
#define __pyx_n_s_buffer_updated __pyx_mstate_global->__pyx_n_s_buffer_updated
#define __pyx_n_s_bufs __pyx_mstate_global->__pyx_n_s_bufs
#define __pyx_n_s_bufsize __pyx_mstate_global->__pyx_n_s_bufsize
#define __pyx_kp_u_bufsize_must_be_0 __pyx_mstate_global->__pyx_kp_u_bufsize_must_be_0
#define __pyx_n_s_builtins __pyx_mstate_global->__pyx_n_s_builtins
#define __pyx_kp_b_c __pyx_mstate_global->__pyx_kp_b_c
#define __pyx_n_s_c_str_path __pyx_mstate_global->__pyx_n_s_c_str_path
#define __pyx_n_s_call_at __pyx_mstate_global->__pyx_n_s_call_at
#define __pyx_n_s_call_connection_made __pyx_mstate_global->__pyx_n_s_call_connection_made
#define __pyx_n_s_call_exception_handler __pyx_mstate_global->__pyx_n_s_call_exception_handler
#define __pyx_n_s_call_later __pyx_mstate_global->__pyx_n_s_call_later
#define __pyx_n_s_call_soon __pyx_mstate_global->__pyx_n_s_call_soon
#define __pyx_n_s_call_soon_threadsafe __pyx_mstate_global->__pyx_n_s_call_soon_threadsafe
#define __pyx_n_s_callback __pyx_mstate_global->__pyx_n_s_callback
#define __pyx_kp_u_called_after_destroying_the_UVH __pyx_mstate_global->__pyx_kp_u_called_after_destroying_the_UVH
#define __pyx_kp_u_called_with_handle_data_NULL __pyx_mstate_global->__pyx_kp_u_called_with_handle_data_NULL
#define __pyx_n_s_can_write_eof __pyx_mstate_global->__pyx_n_s_can_write_eof
#define __pyx_n_s_cancel __pyx_mstate_global->__pyx_n_s_cancel
#define __pyx_kp_u_cancel_request_data_is_NULL __pyx_mstate_global->__pyx_kp_u_cancel_request_data_is_NULL
#define __pyx_kp_u_cancel_request_data_is_not_UVRe __pyx_mstate_global->__pyx_kp_u_cancel_request_data_is_not_UVRe
#define __pyx_kp_u_cancel_request_is_NULL __pyx_mstate_global->__pyx_kp_u_cancel_request_is_NULL
#define __pyx_n_s_cancelled __pyx_mstate_global->__pyx_n_s_cancelled
#define __pyx_n_u_cancelled __pyx_mstate_global->__pyx_n_u_cancelled
#define __pyx_kp_u_cannot_add_a_signal_handler_for __pyx_mstate_global->__pyx_kp_u_cannot_add_a_signal_handler_for
#define __pyx_kp_u_cannot_apply_subprocess_STDOUT __pyx_mstate_global->__pyx_kp_u_cannot_apply_subprocess_STDOUT
#define __pyx_kp_u_cannot_cleanup_signal_handlers_c __pyx_mstate_global->__pyx_kp_u_cannot_cleanup_signal_handlers_c
#define __pyx_kp_u_cannot_convert_sockaddr_into_Pyt __pyx_mstate_global->__pyx_kp_u_cannot_convert_sockaddr_into_Pyt
#define __pyx_kp_u_cannot_open_a_UDP_handle_invalid __pyx_mstate_global->__pyx_kp_u_cannot_open_a_UDP_handle_invalid
#define __pyx_kp_u_cannot_pause_signals_handling_no __pyx_mstate_global->__pyx_kp_u_cannot_pause_signals_handling_no
#define __pyx_kp_u_cannot_run_Handle_callback_is_no __pyx_mstate_global->__pyx_kp_u_cannot_run_Handle_callback_is_no
#define __pyx_kp_u_cannot_run_TimerHandle_callback __pyx_mstate_global->__pyx_kp_u_cannot_run_TimerHandle_callback
#define __pyx_kp_u_cannot_shutdown_signals_handling __pyx_mstate_global->__pyx_kp_u_cannot_shutdown_signals_handling
#define __pyx_kp_u_cannot_spawn_a_process_args_are __pyx_mstate_global->__pyx_kp_u_cannot_spawn_a_process_args_are
#define __pyx_kp_u_cannot_switch_state_from_to __pyx_mstate_global->__pyx_kp_u_cannot_switch_state_from_to
#define __pyx_n_s_cast __pyx_mstate_global->__pyx_n_s_cast
#define __pyx_n_s_cause __pyx_mstate_global->__pyx_n_s_cause
#define __pyx_n_s_cfunc_to_py __pyx_mstate_global->__pyx_n_s_cfunc_to_py
#define __pyx_n_s_chain __pyx_mstate_global->__pyx_n_s_chain
#define __pyx_n_s_check_default_executor __pyx_mstate_global->__pyx_n_s_check_default_executor
#define __pyx_n_s_check_hostname __pyx_mstate_global->__pyx_n_s_check_hostname
#define __pyx_n_s_cipher __pyx_mstate_global->__pyx_n_s_cipher
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cleanup_socket __pyx_mstate_global->__pyx_n_s_cleanup_socket
#define __pyx_n_s_clear __pyx_mstate_global->__pyx_n_s_clear
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_kp_u_close__handle_data_is_NULL __pyx_mstate_global->__pyx_kp_u_close__handle_data_is_NULL
#define __pyx_kp_u_close__handle_data_is_not_UVHa __pyx_mstate_global->__pyx_kp_u_close__handle_data_is_not_UVHa
#define __pyx_n_s_close_fds __pyx_mstate_global->__pyx_n_s_close_fds
#define __pyx_kp_u_close_method __pyx_mstate_global->__pyx_kp_u_close_method
#define __pyx_kp_u_close_uv_is_closing_is_true __pyx_mstate_global->__pyx_kp_u_close_uv_is_closing_is_true
#define __pyx_kp_u_closed_Transport_handle_and_unse __pyx_mstate_global->__pyx_kp_u_closed_Transport_handle_and_unse
#define __pyx_kp_u_closed_reading_x __pyx_mstate_global->__pyx_kp_u_closed_reading_x
#define __pyx_kp_u_closed_x __pyx_mstate_global->__pyx_kp_u_closed_x
#define __pyx_n_s_closing_agens __pyx_mstate_global->__pyx_n_s_closing_agens
#define __pyx_n_s_cmd __pyx_mstate_global->__pyx_n_s_cmd
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_s_collections_abc __pyx_mstate_global->__pyx_n_s_collections_abc
#define __pyx_n_s_completed __pyx_mstate_global->__pyx_n_s_completed
#define __pyx_n_s_compression __pyx_mstate_global->__pyx_n_s_compression
#define __pyx_n_s_concurrent __pyx_mstate_global->__pyx_n_s_concurrent
#define __pyx_kp_u_concurrent_allocations __pyx_mstate_global->__pyx_kp_u_concurrent_allocations
#define __pyx_n_s_concurrent_futures __pyx_mstate_global->__pyx_n_s_concurrent_futures
#define __pyx_n_s_conmade_cb __pyx_mstate_global->__pyx_n_s_conmade_cb
#define __pyx_n_s_connect __pyx_mstate_global->__pyx_n_s_connect
#define __pyx_n_s_connect_accepted_socket __pyx_mstate_global->__pyx_n_s_connect_accepted_socket
#define __pyx_n_s_connect_ex __pyx_mstate_global->__pyx_n_s_connect_ex
#define __pyx_kp_u_connect_ex_method __pyx_mstate_global->__pyx_kp_u_connect_ex_method
#define __pyx_kp_u_connect_failed __pyx_mstate_global->__pyx_kp_u_connect_failed
#define __pyx_kp_u_connect_method __pyx_mstate_global->__pyx_kp_u_connect_method
#define __pyx_n_s_connect_read_pipe __pyx_mstate_global->__pyx_n_s_connect_read_pipe
#define __pyx_n_s_connect_write_pipe __pyx_mstate_global->__pyx_n_s_connect_write_pipe
#define __pyx_n_s_connection_lost __pyx_mstate_global->__pyx_n_s_connection_lost
#define __pyx_n_s_connection_made __pyx_mstate_global->__pyx_n_s_connection_made
#define __pyx_n_s_context __pyx_mstate_global->__pyx_n_s_context
#define __pyx_n_u_context __pyx_mstate_global->__pyx_n_u_context
#define __pyx_kp_u_context_manager_protocol __pyx_mstate_global->__pyx_kp_u_context_manager_protocol
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_coro __pyx_mstate_global->__pyx_n_s_coro
#define __pyx_n_s_coroutines __pyx_mstate_global->__pyx_n_s_coroutines
#define __pyx_kp_u_coroutines_cannot_be_used_with_a __pyx_mstate_global->__pyx_kp_u_coroutines_cannot_be_used_with_a
#define __pyx_kp_u_coroutines_cannot_be_used_with_r __pyx_mstate_global->__pyx_kp_u_coroutines_cannot_be_used_with_r
#define __pyx_kp_u_could_not_bind_to_local_addr __pyx_mstate_global->__pyx_kp_u_could_not_bind_to_local_addr
#define __pyx_kp_u_could_not_bind_to_remote_addr __pyx_mstate_global->__pyx_kp_u_could_not_bind_to_remote_addr
#define __pyx_kp_u_could_not_close_attached_file_ob __pyx_mstate_global->__pyx_kp_u_could_not_close_attached_file_ob
#define __pyx_n_s_create_connection __pyx_mstate_global->__pyx_n_s_create_connection
#define __pyx_n_s_create_connection_locals_genexpr __pyx_mstate_global->__pyx_n_s_create_connection_locals_genexpr
#define __pyx_n_s_create_datagram_endpoint __pyx_mstate_global->__pyx_n_s_create_datagram_endpoint
#define __pyx_n_s_create_datagram_endpoint_locals __pyx_mstate_global->__pyx_n_s_create_datagram_endpoint_locals
#define __pyx_n_s_create_default_context __pyx_mstate_global->__pyx_n_s_create_default_context
#define __pyx_n_s_create_future __pyx_mstate_global->__pyx_n_s_create_future
#define __pyx_n_s_create_server __pyx_mstate_global->__pyx_n_s_create_server
#define __pyx_kp_u_create_server_failed_to_create_s __pyx_mstate_global->__pyx_kp_u_create_server_failed_to_create_s
#define __pyx_n_s_create_task __pyx_mstate_global->__pyx_n_s_create_task
#define __pyx_n_s_create_unix_connection __pyx_mstate_global->__pyx_n_s_create_unix_connection
#define __pyx_n_s_create_unix_server __pyx_mstate_global->__pyx_n_s_create_unix_server
#define __pyx_kp_u_created_at __pyx_mstate_global->__pyx_kp_u_created_at
#define __pyx_n_s_creationflags __pyx_mstate_global->__pyx_n_s_creationflags
#define __pyx_kp_u_creationflags_is_not_supported __pyx_mstate_global->__pyx_kp_u_creationflags_is_not_supported
#define __pyx_n_s_ctx __pyx_mstate_global->__pyx_n_s_ctx
#define __pyx_n_s_cwd __pyx_mstate_global->__pyx_n_s_cwd
#define __pyx_kp_u_cwd_must_be_a_str_or_bytes_objec __pyx_mstate_global->__pyx_kp_u_cwd_must_be_a_str_or_bytes_objec
#define __pyx_n_u_cygwin __pyx_mstate_global->__pyx_n_u_cygwin
#define __pyx_n_u_darwin __pyx_mstate_global->__pyx_n_u_darwin
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_kp_u_data_expecting_a_bytes_like_inst __pyx_mstate_global->__pyx_kp_u_data_expecting_a_bytes_like_inst
#define __pyx_n_s_data_received __pyx_mstate_global->__pyx_n_s_data_received
#define __pyx_n_s_datagram_received __pyx_mstate_global->__pyx_n_s_datagram_received
#define __pyx_kp_u_dealloc____handle_is_NULL__cl __pyx_mstate_global->__pyx_kp_u_dealloc____handle_is_NULL__cl
#define __pyx_kp_u_deallocating_a_running_event_loo __pyx_mstate_global->__pyx_kp_u_deallocating_a_running_event_loo
#define __pyx_kp_u_deallocating_an_open_event_loop __pyx_mstate_global->__pyx_kp_u_deallocating_an_open_event_loop
#define __pyx_n_s_debug __pyx_mstate_global->__pyx_n_s_debug
#define __pyx_n_s_debug_flags __pyx_mstate_global->__pyx_n_s_debug_flags
#define __pyx_n_u_debug_wrapper __pyx_mstate_global->__pyx_n_u_debug_wrapper
#define __pyx_n_s_decode __pyx_mstate_global->__pyx_n_s_decode
#define __pyx_n_s_decref_socketios __pyx_mstate_global->__pyx_n_s_decref_socketios
#define __pyx_n_s_default __pyx_mstate_global->__pyx_n_s_default
#define __pyx_n_s_default_exception_handler __pyx_mstate_global->__pyx_n_s_default_exception_handler
#define __pyx_n_s_default_int_handler __pyx_mstate_global->__pyx_n_s_default_int_handler
#define __pyx_n_s_delay __pyx_mstate_global->__pyx_n_s_delay
#define __pyx_n_s_deque __pyx_mstate_global->__pyx_n_s_deque
#define __pyx_n_s_detach __pyx_mstate_global->__pyx_n_s_detach
#define __pyx_kp_u_detach_method __pyx_mstate_global->__pyx_kp_u_detach_method
#define __pyx_n_s_dev_mode __pyx_mstate_global->__pyx_n_s_dev_mode
#define __pyx_n_s_devnull __pyx_mstate_global->__pyx_n_s_devnull
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_n_s_disable __pyx_mstate_global->__pyx_n_s_disable
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_discard __pyx_mstate_global->__pyx_n_s_discard
#define __pyx_n_s_disconnected __pyx_mstate_global->__pyx_n_s_disconnected
#define __pyx_n_s_do_handshake __pyx_mstate_global->__pyx_n_s_do_handshake
#define __pyx_n_s_do_shutdown __pyx_mstate_global->__pyx_n_s_do_shutdown
#define __pyx_n_s_doc __pyx_mstate_global->__pyx_n_s_doc
#define __pyx_kp_u_does_not_have_a_reader_in_the __pyx_mstate_global->__pyx_kp_u_does_not_have_a_reader_in_the
#define __pyx_kp_u_does_not_have_a_reader_in_the_2 __pyx_mstate_global->__pyx_kp_u_does_not_have_a_reader_in_the_2
#define __pyx_kp_u_does_not_have_a_writer_in_the __pyx_mstate_global->__pyx_kp_u_does_not_have_a_writer_in_the
#define __pyx_kp_u_does_not_have_a_writer_in_the_2 __pyx_mstate_global->__pyx_kp_u_does_not_have_a_writer_in_the_2
#define __pyx_n_s_done __pyx_mstate_global->__pyx_n_s_done
#define __pyx_n_s_done_cb __pyx_mstate_global->__pyx_n_s_done_cb
#define __pyx_n_s_dup __pyx_mstate_global->__pyx_n_s_dup
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_enabled __pyx_mstate_global->__pyx_n_s_enabled
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_ensure_future __pyx_mstate_global->__pyx_n_s_ensure_future
#define __pyx_n_s_enter __pyx_mstate_global->__pyx_n_s_enter
#define __pyx_n_s_enum __pyx_mstate_global->__pyx_n_s_enum
#define __pyx_n_s_env __pyx_mstate_global->__pyx_n_s_env
#define __pyx_n_s_environ __pyx_mstate_global->__pyx_n_s_environ
#define __pyx_n_s_eof_received __pyx_mstate_global->__pyx_n_s_eof_received
#define __pyx_n_s_err __pyx_mstate_global->__pyx_n_s_err
#define __pyx_n_s_errno __pyx_mstate_global->__pyx_n_s_errno
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_n_s_error_received __pyx_mstate_global->__pyx_n_s_error_received
#define __pyx_kp_u_error_status_in_uv_stream_t_list __pyx_mstate_global->__pyx_kp_u_error_status_in_uv_stream_t_list
#define __pyx_kp_u_error_status_in_uv_stream_t_read __pyx_mstate_global->__pyx_kp_u_error_status_in_uv_stream_t_read
#define __pyx_kp_u_error_status_in_uv_stream_t_shut __pyx_mstate_global->__pyx_kp_u_error_status_in_uv_stream_t_shut
#define __pyx_kp_u_error_status_in_uv_stream_t_writ __pyx_mstate_global->__pyx_kp_u_error_status_in_uv_stream_t_writ
#define __pyx_kp_u_error_while_attempting_to_bind_o __pyx_mstate_global->__pyx_kp_u_error_while_attempting_to_bind_o
#define __pyx_n_s_ex __pyx_mstate_global->__pyx_n_s_ex
#define __pyx_n_s_exc __pyx_mstate_global->__pyx_n_s_exc
#define __pyx_n_s_exc_info __pyx_mstate_global->__pyx_n_s_exc_info
#define __pyx_n_s_exception __pyx_mstate_global->__pyx_n_s_exception
#define __pyx_n_u_exception __pyx_mstate_global->__pyx_n_u_exception
#define __pyx_kp_u_exception_in_Task___repr __pyx_mstate_global->__pyx_kp_u_exception_in_Task___repr
#define __pyx_n_s_exceptions __pyx_mstate_global->__pyx_n_s_exceptions
#define __pyx_n_s_excs __pyx_mstate_global->__pyx_n_s_excs
#define __pyx_n_s_executable __pyx_mstate_global->__pyx_n_s_executable
#define __pyx_n_s_executor __pyx_mstate_global->__pyx_n_s_executor
#define __pyx_n_s_exit __pyx_mstate_global->__pyx_n_s_exit
#define __pyx_kp_u_expected_AF_INET_AF_INET6_or_AF __pyx_mstate_global->__pyx_kp_u_expected_AF_INET_AF_INET6_or_AF
#define __pyx_kp_u_expected_asyncio_Future __pyx_mstate_global->__pyx_kp_u_expected_asyncio_Future
#define __pyx_n_s_extract __pyx_mstate_global->__pyx_n_s_extract
#define __pyx_n_s_f1 __pyx_mstate_global->__pyx_n_s_f1
#define __pyx_n_s_f2 __pyx_mstate_global->__pyx_n_s_f2
#define __pyx_n_s_factory __pyx_mstate_global->__pyx_n_s_factory
#define __pyx_n_s_family __pyx_mstate_global->__pyx_n_s_family
#define __pyx_kp_u_fatal_Could_not_advance__StreamW __pyx_mstate_global->__pyx_kp_u_fatal_Could_not_advance__StreamW
#define __pyx_kp_u_fatal_sent_0_in_advance_uv_buf __pyx_mstate_global->__pyx_kp_u_fatal_sent_0_in_advance_uv_buf
#define __pyx_n_s_fd __pyx_mstate_global->__pyx_n_s_fd
#define __pyx_kp_u_fd_2 __pyx_mstate_global->__pyx_kp_u_fd_2
#define __pyx_n_s_fileno __pyx_mstate_global->__pyx_n_s_fileno
#define __pyx_n_s_fileobj __pyx_mstate_global->__pyx_n_s_fileobj
#define __pyx_n_s_finalizer __pyx_mstate_global->__pyx_n_s_finalizer
#define __pyx_n_s_firstiter __pyx_mstate_global->__pyx_n_s_firstiter
#define __pyx_n_s_flag __pyx_mstate_global->__pyx_n_s_flag
#define __pyx_n_s_flags __pyx_mstate_global->__pyx_n_s_flags
#define __pyx_n_s_flowinfo __pyx_mstate_global->__pyx_n_s_flowinfo
#define __pyx_n_s_flush __pyx_mstate_global->__pyx_n_s_flush
#define __pyx_n_s_force_close __pyx_mstate_global->__pyx_n_s_force_close
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_format_list __pyx_mstate_global->__pyx_n_s_format_list
#define __pyx_n_s_frame __pyx_mstate_global->__pyx_n_s_frame
#define __pyx_n_s_fs __pyx_mstate_global->__pyx_n_s_fs
#define __pyx_n_s_fs_handle __pyx_mstate_global->__pyx_n_s_fs_handle
#define __pyx_n_s_fspath __pyx_mstate_global->__pyx_n_s_fspath
#define __pyx_n_s_func __pyx_mstate_global->__pyx_n_s_func
#define __pyx_n_s_functools __pyx_mstate_global->__pyx_n_s_functools
#define __pyx_n_s_fut __pyx_mstate_global->__pyx_n_s_fut
#define __pyx_n_s_future __pyx_mstate_global->__pyx_n_s_future
#define __pyx_n_s_futures __pyx_mstate_global->__pyx_n_s_futures
#define __pyx_n_s_gaierror __pyx_mstate_global->__pyx_n_s_gaierror
#define __pyx_n_s_gather __pyx_mstate_global->__pyx_n_s_gather
#define __pyx_n_s_gc __pyx_mstate_global->__pyx_n_s_gc
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_app_transport __pyx_mstate_global->__pyx_n_s_get_app_transport
#define __pyx_n_s_get_asyncgen_hooks __pyx_mstate_global->__pyx_n_s_get_asyncgen_hooks
#define __pyx_n_s_get_backend_id __pyx_mstate_global->__pyx_n_s_get_backend_id
#define __pyx_n_s_get_buffer __pyx_mstate_global->__pyx_n_s_get_buffer
#define __pyx_n_u_get_buffer __pyx_mstate_global->__pyx_n_u_get_buffer
#define __pyx_n_s_get_coroutine_origin_tracking_de __pyx_mstate_global->__pyx_n_s_get_coroutine_origin_tracking_de
#define __pyx_n_u_get_coroutine_wrapper __pyx_mstate_global->__pyx_n_u_get_coroutine_wrapper
#define __pyx_n_s_get_debug __pyx_mstate_global->__pyx_n_s_get_debug
#define __pyx_n_s_get_event_loop __pyx_mstate_global->__pyx_n_s_get_event_loop
#define __pyx_n_s_get_exception_handler __pyx_mstate_global->__pyx_n_s_get_exception_handler
#define __pyx_n_s_get_extra_info __pyx_mstate_global->__pyx_n_s_get_extra_info
#define __pyx_n_s_get_inheritable __pyx_mstate_global->__pyx_n_s_get_inheritable
#define __pyx_n_s_get_loop __pyx_mstate_global->__pyx_n_s_get_loop
#define __pyx_n_s_get_pid __pyx_mstate_global->__pyx_n_s_get_pid
#define __pyx_n_s_get_pipe_transport __pyx_mstate_global->__pyx_n_s_get_pipe_transport
#define __pyx_n_s_get_protocol __pyx_mstate_global->__pyx_n_s_get_protocol
#define __pyx_n_s_get_read_buffer_limits __pyx_mstate_global->__pyx_n_s_get_read_buffer_limits
#define __pyx_n_s_get_read_buffer_size __pyx_mstate_global->__pyx_n_s_get_read_buffer_size
#define __pyx_n_s_get_returncode __pyx_mstate_global->__pyx_n_s_get_returncode
#define __pyx_n_u_get_running_loop __pyx_mstate_global->__pyx_n_u_get_running_loop
#define __pyx_n_s_get_task_factory __pyx_mstate_global->__pyx_n_s_get_task_factory
#define __pyx_n_s_get_write_buffer_limits __pyx_mstate_global->__pyx_n_s_get_write_buffer_limits
#define __pyx_n_s_get_write_buffer_size __pyx_mstate_global->__pyx_n_s_get_write_buffer_size
#define __pyx_n_s_getaddrinfo __pyx_mstate_global->__pyx_n_s_getaddrinfo
#define __pyx_kp_u_getaddrinfo_returned_empty_list __pyx_mstate_global->__pyx_kp_u_getaddrinfo_returned_empty_list
#define __pyx_kp_u_getaddrinfo_returned_empty_list_2 __pyx_mstate_global->__pyx_kp_u_getaddrinfo_returned_empty_list_2
#define __pyx_n_s_getfilesystemencoding __pyx_mstate_global->__pyx_n_s_getfilesystemencoding
#define __pyx_n_s_getframe __pyx_mstate_global->__pyx_n_s_getframe
#define __pyx_n_s_getnameinfo __pyx_mstate_global->__pyx_n_s_getnameinfo
#define __pyx_kp_u_getnameinfo_argument_1_must_be_a __pyx_mstate_global->__pyx_kp_u_getnameinfo_argument_1_must_be_a
#define __pyx_kp_u_getnameinfo_flowinfo_must_be_0_1 __pyx_mstate_global->__pyx_kp_u_getnameinfo_flowinfo_must_be_0_1
#define __pyx_n_s_getpeercert __pyx_mstate_global->__pyx_n_s_getpeercert
#define __pyx_n_s_getpeername __pyx_mstate_global->__pyx_n_s_getpeername
#define __pyx_n_s_getservbyname __pyx_mstate_global->__pyx_n_s_getservbyname
#define __pyx_kp_u_getsockaddrarg_scope_id_must_be __pyx_mstate_global->__pyx_kp_u_getsockaddrarg_scope_id_must_be
#define __pyx_n_s_getsockname __pyx_mstate_global->__pyx_n_s_getsockname
#define __pyx_n_s_getsockopt __pyx_mstate_global->__pyx_n_s_getsockopt
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_gettimeout __pyx_mstate_global->__pyx_n_s_gettimeout
#define __pyx_n_s_h __pyx_mstate_global->__pyx_n_s_h
#define __pyx_n_s_handle __pyx_mstate_global->__pyx_n_s_handle
#define __pyx_n_u_handle __pyx_mstate_global->__pyx_n_u_handle
#define __pyx_kp_u_handle_data_is_NULL_in___close_a __pyx_mstate_global->__pyx_kp_u_handle_data_is_NULL_in___close_a
#define __pyx_kp_u_handle_loop_data_is_NULL_in___en __pyx_mstate_global->__pyx_kp_u_handle_loop_data_is_NULL_in___en
#define __pyx_kp_u_handle_loop_is_NULL_in___ensure __pyx_mstate_global->__pyx_kp_u_handle_loop_is_NULL_in___ensure
#define __pyx_n_s_handler __pyx_mstate_global->__pyx_n_s_handler
#define __pyx_kp_u_has_handle_0_in___uv_close_handl __pyx_mstate_global->__pyx_kp_u_has_handle_0_in___uv_close_handl
#define __pyx_kp_u_high __pyx_mstate_global->__pyx_kp_u_high
#define __pyx_n_s_high_2 __pyx_mstate_global->__pyx_n_s_high_2
#define __pyx_n_s_host __pyx_mstate_global->__pyx_n_s_host
#define __pyx_kp_u_host_and_port_was_not_specified __pyx_mstate_global->__pyx_kp_u_host_and_port_was_not_specified
#define __pyx_kp_u_host_must_be_a_str_or_bytes __pyx_mstate_global->__pyx_kp_u_host_must_be_a_str_or_bytes
#define __pyx_kp_u_host_must_be_a_string_or_bytes_o __pyx_mstate_global->__pyx_kp_u_host_must_be_a_string_or_bytes_o
#define __pyx_kp_u_host_port_and_sock_can_not_be_sp __pyx_mstate_global->__pyx_kp_u_host_port_and_sock_can_not_be_sp
#define __pyx_n_s_hosts __pyx_mstate_global->__pyx_n_s_hosts
#define __pyx_n_s_id __pyx_mstate_global->__pyx_n_s_id
#define __pyx_n_s_ident __pyx_mstate_global->__pyx_n_s_ident
#define __pyx_n_u_idna __pyx_mstate_global->__pyx_n_u_idna
#define __pyx_n_s_ignore_environment __pyx_mstate_global->__pyx_n_s_ignore_environment
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_u_inf __pyx_mstate_global->__pyx_n_u_inf
#define __pyx_n_s_info __pyx_mstate_global->__pyx_n_s_info
#define __pyx_n_s_infos __pyx_mstate_global->__pyx_n_s_infos
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_inspect __pyx_mstate_global->__pyx_n_s_inspect
#define __pyx_kp_u_integer_argument_expected_got __pyx_mstate_global->__pyx_kp_u_integer_argument_expected_got
#define __pyx_kp_u_invalid_Handle_cb_type __pyx_mstate_global->__pyx_kp_u_invalid_Handle_cb_type
#define __pyx_kp_u_invalid__init_protocol_call __pyx_mstate_global->__pyx_kp_u_invalid__init_protocol_call
#define __pyx_kp_u_invalid_data_in_writebuf_an_inst __pyx_mstate_global->__pyx_kp_u_invalid_data_in_writebuf_an_inst
#define __pyx_kp_u_invalid_socket_family_expected_A __pyx_mstate_global->__pyx_kp_u_invalid_socket_family_expected_A
#define __pyx_kp_u_invalid_waiter_object __pyx_mstate_global->__pyx_kp_u_invalid_waiter_object
#define __pyx_n_s_io_refs __pyx_mstate_global->__pyx_n_s_io_refs
#define __pyx_n_s_ioctl __pyx_mstate_global->__pyx_n_s_ioctl
#define __pyx_n_s_is_alive __pyx_mstate_global->__pyx_n_s_is_alive
#define __pyx_kp_u_is_alive_but__handle_data_is_no __pyx_mstate_global->__pyx_kp_u_is_alive_but__handle_data_is_no
#define __pyx_kp_u_is_alive_but__handle_is_NULL __pyx_mstate_global->__pyx_kp_u_is_alive_but__handle_is_NULL
#define __pyx_kp_u_is_alive_but__handle_loop_is_no __pyx_mstate_global->__pyx_kp_u_is_alive_but__handle_loop_is_no
#define __pyx_kp_u_is_alive_but__loop_is_None __pyx_mstate_global->__pyx_kp_u_is_alive_but__loop_is_None
#define __pyx_kp_u_is_already_being_awaited_on_ser __pyx_mstate_global->__pyx_kp_u_is_already_being_awaited_on_ser
#define __pyx_n_s_is_closed __pyx_mstate_global->__pyx_n_s_is_closed
#define __pyx_kp_u_is_closed_2 __pyx_mstate_global->__pyx_kp_u_is_closed_2
#define __pyx_n_s_is_closing __pyx_mstate_global->__pyx_n_s_is_closing
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_is_longer_than_107_characters __pyx_mstate_global->__pyx_kp_u_is_longer_than_107_characters
#define __pyx_kp_u_is_not_supported_by_start_tls __pyx_mstate_global->__pyx_kp_u_is_not_supported_by_start_tls
#define __pyx_kp_u_is_not_supposed_to_be_instantia __pyx_mstate_global->__pyx_kp_u_is_not_supposed_to_be_instantia
#define __pyx_kp_u_is_open_in___dealloc___with_loo __pyx_mstate_global->__pyx_kp_u_is_open_in___dealloc___with_loo
#define __pyx_n_s_is_reading __pyx_mstate_global->__pyx_n_s_is_reading
#define __pyx_n_s_is_running __pyx_mstate_global->__pyx_n_s_is_running
#define __pyx_n_s_is_serving __pyx_mstate_global->__pyx_n_s_is_serving
#define __pyx_n_s_iscoroutine __pyx_mstate_global->__pyx_n_s_iscoroutine
#define __pyx_n_s_iscoroutinefunction __pyx_mstate_global->__pyx_n_s_iscoroutinefunction
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_u_isfuture __pyx_mstate_global->__pyx_n_u_isfuture
#define __pyx_n_s_isgenerator __pyx_mstate_global->__pyx_n_s_isgenerator
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_itertools __pyx_mstate_global->__pyx_n_s_itertools
#define __pyx_n_s_join __pyx_mstate_global->__pyx_n_s_join
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_kill __pyx_mstate_global->__pyx_n_s_kill
#define __pyx_n_s_kwargs __pyx_mstate_global->__pyx_n_s_kwargs
#define __pyx_kp_u_laddr_s __pyx_mstate_global->__pyx_kp_u_laddr_s
#define __pyx_n_s_lads __pyx_mstate_global->__pyx_n_s_lads
#define __pyx_n_s_lai __pyx_mstate_global->__pyx_n_s_lai
#define __pyx_n_s_lai_addr_static __pyx_mstate_global->__pyx_n_s_lai_addr_static
#define __pyx_n_s_lai_iter __pyx_mstate_global->__pyx_n_s_lai_iter
#define __pyx_n_s_lai_static __pyx_mstate_global->__pyx_n_s_lai_static
#define __pyx_n_s_last __pyx_mstate_global->__pyx_n_s_last
#define __pyx_n_s_libuv_get_loop_t_ptr __pyx_mstate_global->__pyx_n_s_libuv_get_loop_t_ptr
#define __pyx_n_s_libuv_get_version __pyx_mstate_global->__pyx_n_s_libuv_get_version
#define __pyx_n_s_limit __pyx_mstate_global->__pyx_n_s_limit
#define __pyx_n_s_list_of_data __pyx_mstate_global->__pyx_n_s_list_of_data
#define __pyx_n_s_listen __pyx_mstate_global->__pyx_n_s_listen
#define __pyx_kp_u_listen_method __pyx_mstate_global->__pyx_kp_u_listen_method
#define __pyx_n_s_local_addr __pyx_mstate_global->__pyx_n_s_local_addr
#define __pyx_kp_u_local_addr_must_be_a_tuple_of_ho __pyx_mstate_global->__pyx_kp_u_local_addr_must_be_a_tuple_of_ho
#define __pyx_kp_u_local_addr_must_be_a_tuple_of_ho_2 __pyx_mstate_global->__pyx_kp_u_local_addr_must_be_a_tuple_of_ho_2
#define __pyx_n_s_log __pyx_mstate_global->__pyx_n_s_log
#define __pyx_n_s_log_destroy_pending __pyx_mstate_global->__pyx_n_s_log_destroy_pending
#define __pyx_n_s_log_lines __pyx_mstate_global->__pyx_n_s_log_lines
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_lookup_lines __pyx_mstate_global->__pyx_n_s_lookup_lines
#define __pyx_n_s_loop __pyx_mstate_global->__pyx_n_s_loop
#define __pyx_kp_u_loop__exec_queued_writes __pyx_mstate_global->__pyx_kp_u_loop__exec_queued_writes
#define __pyx_kp_u_loop__on_idle __pyx_mstate_global->__pyx_kp_u_loop__on_idle
#define __pyx_n_s_low __pyx_mstate_global->__pyx_n_s_low
#define __pyx_n_s_lower __pyx_mstate_global->__pyx_n_s_lower
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_main_thread __pyx_mstate_global->__pyx_n_s_main_thread
#define __pyx_n_s_makefile __pyx_mstate_global->__pyx_n_s_makefile
#define __pyx_kp_u_makefile_method __pyx_mstate_global->__pyx_kp_u_makefile_method
#define __pyx_n_s_maxsize __pyx_mstate_global->__pyx_n_s_maxsize
#define __pyx_kp_u_maxsize_is_expected_to_be_greate __pyx_mstate_global->__pyx_kp_u_maxsize_is_expected_to_be_greate
#define __pyx_n_s_message __pyx_mstate_global->__pyx_n_s_message
#define __pyx_n_u_message __pyx_mstate_global->__pyx_n_u_message
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_mode __pyx_mstate_global->__pyx_n_s_mode
#define __pyx_n_s_model __pyx_mstate_global->__pyx_n_s_model
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_monitor_fs __pyx_mstate_global->__pyx_n_s_monitor_fs
#define __pyx_n_s_monotonic __pyx_mstate_global->__pyx_n_s_monotonic
#define __pyx_n_s_move_to_end __pyx_mstate_global->__pyx_n_s_move_to_end
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_msg __pyx_mstate_global->__pyx_n_s_msg
#define __pyx_kp_u_must_be_0 __pyx_mstate_global->__pyx_kp_u_must_be_0
#define __pyx_kp_u_must_be_low __pyx_mstate_global->__pyx_kp_u_must_be_low
#define __pyx_n_s_n __pyx_mstate_global->__pyx_n_s_n
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_u_name __pyx_mstate_global->__pyx_n_u_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_nbytes __pyx_mstate_global->__pyx_n_s_nbytes
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_kp_u_new_callbacks_were_queued_during __pyx_mstate_global->__pyx_kp_u_new_callbacks_were_queued_during
#define __pyx_kp_u_new_poll_handles_were_queued_dur __pyx_mstate_global->__pyx_kp_u_new_poll_handles_were_queued_dur
#define __pyx_n_s_new_task __pyx_mstate_global->__pyx_n_s_new_task
#define __pyx_kp_u_new_timers_were_queued_during_lo __pyx_mstate_global->__pyx_kp_u_new_timers_were_queued_during_lo
#define __pyx_n_s_nexc __pyx_mstate_global->__pyx_n_s_nexc
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_kp_u_no_path_and_sock_were_specified __pyx_mstate_global->__pyx_kp_u_no_path_and_sock_were_specified
#define __pyx_kp_u_no_python_buffer_is_allocated_in __pyx_mstate_global->__pyx_kp_u_no_python_buffer_is_allocated_in
#define __pyx_n_s_noop __pyx_mstate_global->__pyx_n_s_noop
#define __pyx_n_s_noop_2 __pyx_mstate_global->__pyx_n_s_noop_2
#define __pyx_n_s_obj __pyx_mstate_global->__pyx_n_s_obj
#define __pyx_n_s_object __pyx_mstate_global->__pyx_n_s_object
#define __pyx_kp_u_object_created_at_most_recent_ca __pyx_mstate_global->__pyx_kp_u_object_created_at_most_recent_ca
#define __pyx_kp_u_objs_handles __pyx_mstate_global->__pyx_kp_u_objs_handles
#define __pyx_n_s_old_agen_hooks __pyx_mstate_global->__pyx_n_s_old_agen_hooks
#define __pyx_n_s_on_ssl_connected __pyx_mstate_global->__pyx_n_s_on_ssl_connected
#define __pyx_n_s_open __pyx_mstate_global->__pyx_n_s_open
#define __pyx_kp_u_open__StreamWriteContext_is_bein __pyx_mstate_global->__pyx_kp_u_open__StreamWriteContext_is_bein
#define __pyx_kp_u_open__UDPSendContext_is_being_de __pyx_mstate_global->__pyx_kp_u_open__UDPSendContext_is_being_de
#define __pyx_n_s_opts __pyx_mstate_global->__pyx_n_s_opts
#define __pyx_n_s_orig_path __pyx_mstate_global->__pyx_n_s_orig_path
#define __pyx_n_s_os __pyx_mstate_global->__pyx_n_s_os
#define __pyx_n_s_p_bytes __pyx_mstate_global->__pyx_n_s_p_bytes
#define __pyx_n_s_partial __pyx_mstate_global->__pyx_n_s_partial
#define __pyx_n_s_pass_fds __pyx_mstate_global->__pyx_n_s_pass_fds
#define __pyx_n_s_path __pyx_mstate_global->__pyx_n_s_path
#define __pyx_kp_u_path_and_sock_can_not_be_specifi __pyx_mstate_global->__pyx_kp_u_path_and_sock_can_not_be_specifi
#define __pyx_kp_u_path_was_not_specified_and_no_so __pyx_mstate_global->__pyx_kp_u_path_was_not_specified_and_no_so
#define __pyx_n_s_pause_reading __pyx_mstate_global->__pyx_n_s_pause_reading
#define __pyx_n_s_pause_writing __pyx_mstate_global->__pyx_n_s_pause_writing
#define __pyx_n_s_peercert __pyx_mstate_global->__pyx_n_s_peercert
#define __pyx_n_u_peername __pyx_mstate_global->__pyx_n_u_peername
#define __pyx_n_s_pending __pyx_mstate_global->__pyx_n_s_pending
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_pipe __pyx_mstate_global->__pyx_n_s_pipe
#define __pyx_n_u_pipe __pyx_mstate_global->__pyx_n_u_pipe
#define __pyx_kp_u_pipe_2 __pyx_mstate_global->__pyx_kp_u_pipe_2
#define __pyx_n_s_pipe_connection_lost __pyx_mstate_global->__pyx_n_s_pipe_connection_lost
#define __pyx_n_s_pipe_data_received __pyx_mstate_global->__pyx_n_s_pipe_data_received
#define __pyx_n_s_platform __pyx_mstate_global->__pyx_n_s_platform
#define __pyx_n_s_pop __pyx_mstate_global->__pyx_n_s_pop
#define __pyx_n_s_popitem __pyx_mstate_global->__pyx_n_s_popitem
#define __pyx_n_s_popleft __pyx_mstate_global->__pyx_n_s_popleft
#define __pyx_n_s_port __pyx_mstate_global->__pyx_n_s_port
#define __pyx_kp_u_port_must_be_a_str_bytes_or_int __pyx_mstate_global->__pyx_kp_u_port_must_be_a_str_bytes_or_int
#define __pyx_n_u_posix __pyx_mstate_global->__pyx_n_u_posix
#define __pyx_n_s_preexec_fn __pyx_mstate_global->__pyx_n_s_preexec_fn
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_print __pyx_mstate_global->__pyx_n_s_print
#define __pyx_n_u_print_debug_info __pyx_mstate_global->__pyx_n_u_print_debug_info
#define __pyx_n_s_problems __pyx_mstate_global->__pyx_n_s_problems
#define __pyx_n_s_proc __pyx_mstate_global->__pyx_n_s_proc
#define __pyx_n_s_process_exited __pyx_mstate_global->__pyx_n_s_process_exited
#define __pyx_n_s_process_id __pyx_mstate_global->__pyx_n_s_process_id
#define __pyx_n_s_program __pyx_mstate_global->__pyx_n_s_program
#define __pyx_n_s_proto __pyx_mstate_global->__pyx_n_s_proto
#define __pyx_n_s_proto_factory __pyx_mstate_global->__pyx_n_s_proto_factory
#define __pyx_n_s_protocol __pyx_mstate_global->__pyx_n_s_protocol
#define __pyx_n_u_protocol __pyx_mstate_global->__pyx_n_u_protocol
#define __pyx_n_s_protocol_factory __pyx_mstate_global->__pyx_n_s_protocol_factory
#define __pyx_kp_u_protocol_is_not_set_cannot_call __pyx_mstate_global->__pyx_kp_u_protocol_is_not_set_cannot_call
#define __pyx_kp_u_protocol_is_required __pyx_mstate_global->__pyx_kp_u_protocol_is_required
#define __pyx_kp_u_protocol_pause_writing_failed __pyx_mstate_global->__pyx_kp_u_protocol_pause_writing_failed
#define __pyx_kp_u_protocol_resume_writing_failed __pyx_mstate_global->__pyx_kp_u_protocol_resume_writing_failed
#define __pyx_n_s_pyaddr __pyx_mstate_global->__pyx_n_s_pyaddr
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_LruCache __pyx_mstate_global->__pyx_n_s_pyx_unpickle_LruCache
#define __pyx_n_s_pyx_unpickle_PseudoSocket __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PseudoSocket
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_u_qualname __pyx_mstate_global->__pyx_n_u_qualname
#define __pyx_kp_u_r_SSL_handshake_took_1f_ms __pyx_mstate_global->__pyx_kp_u_r_SSL_handshake_took_1f_ms
#define __pyx_kp_u_r_fileno_returned_non_integer __pyx_mstate_global->__pyx_kp_u_r_fileno_returned_non_integer
#define __pyx_kp_u_r_received_EOF __pyx_mstate_global->__pyx_kp_u_r_received_EOF
#define __pyx_kp_u_r_s __pyx_mstate_global->__pyx_kp_u_r_s
#define __pyx_kp_u_r_starts_SSL_handshake __pyx_mstate_global->__pyx_kp_u_r_starts_SSL_handshake
#define __pyx_kp_u_raddr_s __pyx_mstate_global->__pyx_kp_u_raddr_s
#define __pyx_n_s_rads __pyx_mstate_global->__pyx_n_s_rads
#define __pyx_n_s_rai __pyx_mstate_global->__pyx_n_s_rai
#define __pyx_n_s_rai_addr_static __pyx_mstate_global->__pyx_n_s_rai_addr_static
#define __pyx_n_s_rai_iter __pyx_mstate_global->__pyx_n_s_rai_iter
#define __pyx_n_s_rai_static __pyx_mstate_global->__pyx_n_s_rai_static
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_read __pyx_mstate_global->__pyx_n_s_read
#define __pyx_n_s_recv __pyx_mstate_global->__pyx_n_s_recv
#define __pyx_n_s_recv_into __pyx_mstate_global->__pyx_n_s_recv_into
#define __pyx_kp_u_recv_into_method __pyx_mstate_global->__pyx_kp_u_recv_into_method
#define __pyx_kp_u_recv_method __pyx_mstate_global->__pyx_kp_u_recv_method
#define __pyx_n_s_recvfrom __pyx_mstate_global->__pyx_n_s_recvfrom
#define __pyx_n_s_recvfrom_into __pyx_mstate_global->__pyx_n_s_recvfrom_into
#define __pyx_kp_u_recvfrom_into_method __pyx_mstate_global->__pyx_kp_u_recvfrom_into_method
#define __pyx_kp_u_recvfrom_method __pyx_mstate_global->__pyx_kp_u_recvfrom_method
#define __pyx_n_s_recvmsg __pyx_mstate_global->__pyx_n_s_recvmsg
#define __pyx_n_s_recvmsg_into __pyx_mstate_global->__pyx_n_s_recvmsg_into
#define __pyx_kp_u_recvmsg_into_method __pyx_mstate_global->__pyx_kp_u_recvmsg_into_method
#define __pyx_kp_u_recvmsg_method __pyx_mstate_global->__pyx_kp_u_recvmsg_method
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_remote_addr __pyx_mstate_global->__pyx_n_s_remote_addr
#define __pyx_kp_u_remote_addr_must_be_a_tuple_of_h __pyx_mstate_global->__pyx_kp_u_remote_addr_must_be_a_tuple_of_h
#define __pyx_n_s_remove __pyx_mstate_global->__pyx_n_s_remove
#define __pyx_n_s_remove_done_callback __pyx_mstate_global->__pyx_n_s_remove_done_callback
#define __pyx_n_s_remove_reader __pyx_mstate_global->__pyx_n_s_remove_reader
#define __pyx_n_s_remove_reader_2 __pyx_mstate_global->__pyx_n_s_remove_reader_2
#define __pyx_n_s_remove_signal_handler __pyx_mstate_global->__pyx_n_s_remove_signal_handler
#define __pyx_kp_u_remove_signal_handler_can_only_b __pyx_mstate_global->__pyx_kp_u_remove_signal_handler_can_only_b
#define __pyx_n_s_remove_writer __pyx_mstate_global->__pyx_n_s_remove_writer
#define __pyx_n_s_remove_writer_2 __pyx_mstate_global->__pyx_n_s_remove_writer_2
#define __pyx_n_s_repr __pyx_mstate_global->__pyx_n_s_repr
#define __pyx_n_s_restore_signals __pyx_mstate_global->__pyx_n_s_restore_signals
#define __pyx_n_s_result __pyx_mstate_global->__pyx_n_s_result
#define __pyx_n_s_results __pyx_mstate_global->__pyx_n_s_results
#define __pyx_n_s_resume_cb __pyx_mstate_global->__pyx_n_s_resume_cb
#define __pyx_n_s_resume_reading __pyx_mstate_global->__pyx_n_s_resume_reading
#define __pyx_n_s_resume_writing __pyx_mstate_global->__pyx_n_s_resume_writing
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_return_exceptions __pyx_mstate_global->__pyx_n_s_return_exceptions
#define __pyx_kp_u_returning_true_from_eof_received __pyx_mstate_global->__pyx_kp_u_returning_true_from_eof_received
#define __pyx_n_s_reuse_address __pyx_mstate_global->__pyx_n_s_reuse_address
#define __pyx_n_s_reuse_port __pyx_mstate_global->__pyx_n_s_reuse_port
#define __pyx_kp_u_reuse_port_not_supported_by_sock __pyx_mstate_global->__pyx_kp_u_reuse_port_not_supported_by_sock
#define __pyx_n_s_reverse __pyx_mstate_global->__pyx_n_s_reverse
#define __pyx_n_s_round __pyx_mstate_global->__pyx_n_s_round
#define __pyx_n_s_rstrip __pyx_mstate_global->__pyx_n_s_rstrip
#define __pyx_n_s_run __pyx_mstate_global->__pyx_n_s_run
#define __pyx_n_s_run_forever __pyx_mstate_global->__pyx_n_s_run_forever
#define __pyx_n_s_run_in_executor __pyx_mstate_global->__pyx_n_s_run_in_executor
#define __pyx_n_s_run_until_complete __pyx_mstate_global->__pyx_n_s_run_until_complete
#define __pyx_n_s_run_until_complete_locals_done_c __pyx_mstate_global->__pyx_n_s_run_until_complete_locals_done_c
#define __pyx_kp_u_running_closed_debug __pyx_mstate_global->__pyx_kp_u_running_closed_debug
#define __pyx_n_s_scope_id __pyx_mstate_global->__pyx_n_s_scope_id
#define __pyx_kp_u_seconds __pyx_mstate_global->__pyx_kp_u_seconds
#define __pyx_kp_u_seconds_aborting_the_connection __pyx_mstate_global->__pyx_kp_u_seconds_aborting_the_connection
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_self_2 __pyx_mstate_global->__pyx_n_s_self_2
#define __pyx_n_u_self_2 __pyx_mstate_global->__pyx_n_u_self_2
#define __pyx_kp_u_self_pipe_exists_before_loop_run __pyx_mstate_global->__pyx_kp_u_self_pipe_exists_before_loop_run
#define __pyx_kp_u_self_pipe_was_not_cleaned_up_aft __pyx_mstate_global->__pyx_kp_u_self_pipe_was_not_cleaned_up_aft
#define __pyx_kp_s_self_py_buf_self_req_cannot_be_c __pyx_mstate_global->__pyx_kp_s_self_py_buf_self_req_cannot_be_c
#define __pyx_kp_s_self_py_bufs_self_py_bufs_sml_se __pyx_mstate_global->__pyx_kp_s_self_py_bufs_self_py_bufs_sml_se
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_kp_u_send_method __pyx_mstate_global->__pyx_kp_u_send_method
#define __pyx_n_s_send_signal __pyx_mstate_global->__pyx_n_s_send_signal
#define __pyx_n_s_sendall __pyx_mstate_global->__pyx_n_s_sendall
#define __pyx_kp_u_sendall_method __pyx_mstate_global->__pyx_kp_u_sendall_method
#define __pyx_n_s_sendfile __pyx_mstate_global->__pyx_n_s_sendfile
#define __pyx_kp_u_sendfile_method __pyx_mstate_global->__pyx_kp_u_sendfile_method
#define __pyx_n_s_sendmsg __pyx_mstate_global->__pyx_n_s_sendmsg
#define __pyx_n_s_sendmsg_afalg __pyx_mstate_global->__pyx_n_s_sendmsg_afalg
#define __pyx_kp_u_sendmsg_afalg_method __pyx_mstate_global->__pyx_kp_u_sendmsg_afalg_method
#define __pyx_kp_u_sendmsg_method __pyx_mstate_global->__pyx_kp_u_sendmsg_method
#define __pyx_n_s_sendto __pyx_mstate_global->__pyx_n_s_sendto
#define __pyx_kp_u_sendto_method __pyx_mstate_global->__pyx_kp_u_sendto_method
#define __pyx_n_s_serve_forever __pyx_mstate_global->__pyx_n_s_serve_forever
#define __pyx_kp_u_server __pyx_mstate_global->__pyx_kp_u_server
#define __pyx_n_s_server_2 __pyx_mstate_global->__pyx_n_s_server_2
#define __pyx_n_s_server_hostname __pyx_mstate_global->__pyx_n_s_server_hostname
#define __pyx_kp_u_server_hostname_is_only_meaningf __pyx_mstate_global->__pyx_kp_u_server_hostname_is_only_meaningf
#define __pyx_n_s_server_side __pyx_mstate_global->__pyx_n_s_server_side
#define __pyx_n_s_servers __pyx_mstate_global->__pyx_n_s_servers
#define __pyx_kp_u_service_proto_not_found __pyx_mstate_global->__pyx_kp_u_service_proto_not_found
#define __pyx_n_s_set_asyncgen_hooks __pyx_mstate_global->__pyx_n_s_set_asyncgen_hooks
#define __pyx_n_s_set_coroutine_origin_tracking_de __pyx_mstate_global->__pyx_n_s_set_coroutine_origin_tracking_de
#define __pyx_n_u_set_coroutine_wrapper __pyx_mstate_global->__pyx_n_u_set_coroutine_wrapper
#define __pyx_n_s_set_debug __pyx_mstate_global->__pyx_n_s_set_debug
#define __pyx_n_s_set_default_executor __pyx_mstate_global->__pyx_n_s_set_default_executor
#define __pyx_n_s_set_exception __pyx_mstate_global->__pyx_n_s_set_exception
#define __pyx_n_s_set_exception_handler __pyx_mstate_global->__pyx_n_s_set_exception_handler
#define __pyx_n_s_set_inheritable __pyx_mstate_global->__pyx_n_s_set_inheritable
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_set_name_2 __pyx_mstate_global->__pyx_n_s_set_name_2
#define __pyx_n_s_set_protocol __pyx_mstate_global->__pyx_n_s_set_protocol
#define __pyx_n_s_set_read_buffer_limits __pyx_mstate_global->__pyx_n_s_set_read_buffer_limits
#define __pyx_n_s_set_result __pyx_mstate_global->__pyx_n_s_set_result
#define __pyx_n_u_set_running_loop __pyx_mstate_global->__pyx_n_u_set_running_loop
#define __pyx_n_s_set_task_factory __pyx_mstate_global->__pyx_n_s_set_task_factory
#define __pyx_n_s_set_wakeup_fd __pyx_mstate_global->__pyx_n_s_set_wakeup_fd
#define __pyx_kp_u_set_wakeup_fd_1_failed_s __pyx_mstate_global->__pyx_kp_u_set_wakeup_fd_1_failed_s
#define __pyx_n_s_set_write_buffer_limits __pyx_mstate_global->__pyx_n_s_set_write_buffer_limits
#define __pyx_n_s_setblocking __pyx_mstate_global->__pyx_n_s_setblocking
#define __pyx_kp_u_setblocking_transport_sockets_ca __pyx_mstate_global->__pyx_kp_u_setblocking_transport_sockets_ca
#define __pyx_n_s_setsockopt __pyx_mstate_global->__pyx_n_s_setsockopt
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_settimeout __pyx_mstate_global->__pyx_n_s_settimeout
#define __pyx_kp_u_settimeout_only_0_timeout_is_all __pyx_mstate_global->__pyx_kp_u_settimeout_only_0_timeout_is_all
#define __pyx_n_s_share __pyx_mstate_global->__pyx_n_s_share
#define __pyx_n_s_shell __pyx_mstate_global->__pyx_n_s_shell
#define __pyx_kp_u_shell_must_be_False __pyx_mstate_global->__pyx_kp_u_shell_must_be_False
#define __pyx_kp_u_shell_must_be_True __pyx_mstate_global->__pyx_kp_u_shell_must_be_True
#define __pyx_n_s_shutdown __pyx_mstate_global->__pyx_n_s_shutdown
#define __pyx_n_s_shutdown_asyncgens __pyx_mstate_global->__pyx_n_s_shutdown_asyncgens
#define __pyx_n_s_shutdown_coro __pyx_mstate_global->__pyx_n_s_shutdown_coro
#define __pyx_n_s_shutdown_default_executor __pyx_mstate_global->__pyx_n_s_shutdown_default_executor
#define __pyx_kp_u_shutdown_method __pyx_mstate_global->__pyx_kp_u_shutdown_method
#define __pyx_n_s_sig __pyx_mstate_global->__pyx_n_s_sig
#define __pyx_kp_u_sig_cannot_be_caught __pyx_mstate_global->__pyx_kp_u_sig_cannot_be_caught
#define __pyx_kp_u_sig_must_be_an_int_not_r __pyx_mstate_global->__pyx_kp_u_sig_must_be_an_int_not_r
#define __pyx_kp_u_sig_out_of_range_1 __pyx_mstate_global->__pyx_kp_u_sig_out_of_range_1
#define __pyx_n_s_sighandler __pyx_mstate_global->__pyx_n_s_sighandler
#define __pyx_n_s_siginterrupt __pyx_mstate_global->__pyx_n_s_siginterrupt
#define __pyx_n_s_signal __pyx_mstate_global->__pyx_n_s_signal
#define __pyx_kp_u_signals_handling_has_been_alread __pyx_mstate_global->__pyx_kp_u_signals_handling_has_been_alread
#define __pyx_kp_u_signals_handling_has_not_been_se __pyx_mstate_global->__pyx_kp_u_signals_handling_has_not_been_se
#define __pyx_n_s_signum __pyx_mstate_global->__pyx_n_s_signum
#define __pyx_n_s_sin6 __pyx_mstate_global->__pyx_n_s_sin6
#define __pyx_n_s_sl __pyx_mstate_global->__pyx_n_s_sl
#define __pyx_n_s_sleep __pyx_mstate_global->__pyx_n_s_sleep
#define __pyx_n_s_sock __pyx_mstate_global->__pyx_n_s_sock
#define __pyx_n_s_sock_accept __pyx_mstate_global->__pyx_n_s_sock_accept
#define __pyx_n_s_sock_connect __pyx_mstate_global->__pyx_n_s_sock_connect
#define __pyx_kp_u_sock_connect_cb_is_called_on_a __pyx_mstate_global->__pyx_kp_u_sock_connect_cb_is_called_on_a
#define __pyx_n_s_sock_recv __pyx_mstate_global->__pyx_n_s_sock_recv
#define __pyx_n_s_sock_recv_into __pyx_mstate_global->__pyx_n_s_sock_recv_into
#define __pyx_kp_u_sock_recv_into_is_called_on_a_c __pyx_mstate_global->__pyx_kp_u_sock_recv_into_is_called_on_a_c
#define __pyx_kp_u_sock_recv_is_called_on_a_cancel __pyx_mstate_global->__pyx_kp_u_sock_recv_is_called_on_a_cancel
#define __pyx_n_s_sock_recvfrom __pyx_mstate_global->__pyx_n_s_sock_recvfrom
#define __pyx_n_s_sock_recvfrom_into __pyx_mstate_global->__pyx_n_s_sock_recvfrom_into
#define __pyx_n_s_sock_sendall __pyx_mstate_global->__pyx_n_s_sock_sendall
#define __pyx_kp_u_sock_sendall_is_called_on_a_can __pyx_mstate_global->__pyx_kp_u_sock_sendall_is_called_on_a_can
#define __pyx_n_s_sock_sendto __pyx_mstate_global->__pyx_n_s_sock_sendto
#define __pyx_n_s_sockaddr __pyx_mstate_global->__pyx_n_s_sockaddr
#define __pyx_kp_u_sockaddr_must_be_a_tuple_of_2_3 __pyx_mstate_global->__pyx_kp_u_sockaddr_must_be_a_tuple_of_2_3
#define __pyx_kp_u_sockaddr_resolved_to_multiple_ad __pyx_mstate_global->__pyx_kp_u_sockaddr_resolved_to_multiple_ad
#define __pyx_kp_u_socket __pyx_mstate_global->__pyx_kp_u_socket
#define __pyx_n_s_socket_2 __pyx_mstate_global->__pyx_n_s_socket_2
#define __pyx_n_u_socket_2 __pyx_mstate_global->__pyx_n_u_socket_2
#define __pyx_kp_u_socket_family_mismatch_or_a_DNS __pyx_mstate_global->__pyx_kp_u_socket_family_mismatch_or_a_DNS
#define __pyx_kp_u_socket_modifier_keyword_argument __pyx_mstate_global->__pyx_kp_u_socket_modifier_keyword_argument
#define __pyx_kp_u_socket_send_raised_exception __pyx_mstate_global->__pyx_kp_u_socket_send_raised_exception
#define __pyx_n_s_socketpair __pyx_mstate_global->__pyx_n_s_socketpair
#define __pyx_kp_u_sockets __pyx_mstate_global->__pyx_kp_u_sockets
#define __pyx_n_s_sockets_2 __pyx_mstate_global->__pyx_n_s_sockets_2
#define __pyx_n_u_sockname __pyx_mstate_global->__pyx_n_u_sockname
#define __pyx_n_s_source __pyx_mstate_global->__pyx_n_s_source
#define __pyx_n_u_source_traceback __pyx_mstate_global->__pyx_n_u_source_traceback
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_split __pyx_mstate_global->__pyx_n_s_split
#define __pyx_n_s_ssl __pyx_mstate_global->__pyx_n_s_ssl
#define __pyx_kp_u_ssl_argument_must_be_an_SSLConte __pyx_mstate_global->__pyx_kp_u_ssl_argument_must_be_an_SSLConte
#define __pyx_n_s_ssl_handshake_timeout __pyx_mstate_global->__pyx_n_s_ssl_handshake_timeout
#define __pyx_kp_u_ssl_handshake_timeout_is_only_me __pyx_mstate_global->__pyx_kp_u_ssl_handshake_timeout_is_only_me
#define __pyx_kp_u_ssl_handshake_timeout_should_be __pyx_mstate_global->__pyx_kp_u_ssl_handshake_timeout_should_be
#define __pyx_kp_u_ssl_is_expected_to_be_None_or_an __pyx_mstate_global->__pyx_kp_u_ssl_is_expected_to_be_None_or_an
#define __pyx_n_s_ssl_object __pyx_mstate_global->__pyx_n_s_ssl_object
#define __pyx_n_s_ssl_protocol __pyx_mstate_global->__pyx_n_s_ssl_protocol
#define __pyx_n_s_ssl_shutdown_timeout __pyx_mstate_global->__pyx_n_s_ssl_shutdown_timeout
#define __pyx_kp_u_ssl_shutdown_timeout_is_only_mea __pyx_mstate_global->__pyx_kp_u_ssl_shutdown_timeout_is_only_mea
#define __pyx_kp_u_ssl_shutdown_timeout_should_be_a __pyx_mstate_global->__pyx_kp_u_ssl_shutdown_timeout_should_be_a
#define __pyx_n_s_ssl_waiter __pyx_mstate_global->__pyx_n_s_ssl_waiter
#define __pyx_n_s_sslcontext __pyx_mstate_global->__pyx_n_s_sslcontext
#define __pyx_kp_u_sslcontext_is_expected_to_be_an __pyx_mstate_global->__pyx_kp_u_sslcontext_is_expected_to_be_an
#define __pyx_n_s_sslproto __pyx_mstate_global->__pyx_n_s_sslproto
#define __pyx_n_s_st_ino __pyx_mstate_global->__pyx_n_s_st_ino
#define __pyx_n_s_st_mode __pyx_mstate_global->__pyx_n_s_st_mode
#define __pyx_n_s_stacklevel __pyx_mstate_global->__pyx_n_s_stacklevel
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_kp_u_start_init_can_only_be_called __pyx_mstate_global->__pyx_kp_u_start_init_can_only_be_called
#define __pyx_n_s_start_new_session __pyx_mstate_global->__pyx_n_s_start_new_session
#define __pyx_n_s_start_serving __pyx_mstate_global->__pyx_n_s_start_serving
#define __pyx_n_s_start_tls __pyx_mstate_global->__pyx_n_s_start_tls
#define __pyx_n_s_startupinfo __pyx_mstate_global->__pyx_n_s_startupinfo
#define __pyx_kp_u_startupinfo_is_not_supported __pyx_mstate_global->__pyx_kp_u_startupinfo_is_not_supported
#define __pyx_n_s_stat __pyx_mstate_global->__pyx_n_s_stat
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_stderr __pyx_mstate_global->__pyx_n_s_stderr
#define __pyx_n_s_stdin __pyx_mstate_global->__pyx_n_s_stdin
#define __pyx_n_s_stdio_fut __pyx_mstate_global->__pyx_n_s_stdio_fut
#define __pyx_n_s_stdio_inited __pyx_mstate_global->__pyx_n_s_stdio_inited
#define __pyx_n_s_stdout __pyx_mstate_global->__pyx_n_s_stdout
#define __pyx_n_s_stop __pyx_mstate_global->__pyx_n_s_stop
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_kp_u_stream_is_open_after_UVStream__t __pyx_mstate_global->__pyx_kp_u_stream_is_open_after_UVStream__t
#define __pyx_n_s_strerror __pyx_mstate_global->__pyx_n_s_strerror
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_submit __pyx_mstate_global->__pyx_n_s_submit
#define __pyx_n_s_subprocess __pyx_mstate_global->__pyx_n_s_subprocess
#define __pyx_kp_u_subprocess_STDOUT_is_supported_o __pyx_mstate_global->__pyx_kp_u_subprocess_STDOUT_is_supported_o
#define __pyx_n_s_subprocess_exec __pyx_mstate_global->__pyx_n_s_subprocess_exec
#define __pyx_n_s_subprocess_run __pyx_mstate_global->__pyx_n_s_subprocess_run
#define __pyx_n_s_subprocess_shell __pyx_mstate_global->__pyx_n_s_subprocess_shell
#define __pyx_n_s_subtract __pyx_mstate_global->__pyx_n_s_subtract
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_sys __pyx_mstate_global->__pyx_n_s_sys
#define __pyx_n_s_target __pyx_mstate_global->__pyx_n_s_target
#define __pyx_n_s_task __pyx_mstate_global->__pyx_n_s_task
#define __pyx_n_u_task __pyx_mstate_global->__pyx_n_u_task
#define __pyx_kp_u_task_factory_must_be_a_callable __pyx_mstate_global->__pyx_kp_u_task_factory_must_be_a_callable
#define __pyx_n_s_tb __pyx_mstate_global->__pyx_n_s_tb
#define __pyx_n_s_tcp __pyx_mstate_global->__pyx_n_s_tcp
#define __pyx_n_u_tcp __pyx_mstate_global->__pyx_n_u_tcp
#define __pyx_n_s_terminate __pyx_mstate_global->__pyx_n_s_terminate
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_test__append_write_backlog __pyx_mstate_global->__pyx_n_s_test__append_write_backlog
#define __pyx_n_s_test_coroutine_1 __pyx_mstate_global->__pyx_n_s_test_coroutine_1
#define __pyx_n_s_testhelper_unwrap_capsuled_poin __pyx_mstate_global->__pyx_n_s_testhelper_unwrap_capsuled_poin
#define __pyx_kp_u_the_socket_must_be_non_blocking __pyx_mstate_global->__pyx_kp_u_the_socket_must_be_non_blocking
#define __pyx_kp_u_this_event_loop_is_already_runni __pyx_mstate_global->__pyx_kp_u_this_event_loop_is_already_runni
#define __pyx_n_s_thread __pyx_mstate_global->__pyx_n_s_thread
#define __pyx_n_s_threading __pyx_mstate_global->__pyx_n_s_threading
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_s_timeout __pyx_mstate_global->__pyx_n_s_timeout
#define __pyx_n_s_tr __pyx_mstate_global->__pyx_n_s_tr
#define __pyx_n_s_traceback __pyx_mstate_global->__pyx_n_s_traceback
#define __pyx_n_s_traceback_2 __pyx_mstate_global->__pyx_n_s_traceback_2
#define __pyx_n_s_transp __pyx_mstate_global->__pyx_n_s_transp
#define __pyx_n_s_transport __pyx_mstate_global->__pyx_n_s_transport
#define __pyx_n_u_transport __pyx_mstate_global->__pyx_n_u_transport
#define __pyx_kp_u_transport_2 __pyx_mstate_global->__pyx_kp_u_transport_2
#define __pyx_kp_u_transport_sockets_do_not_support __pyx_mstate_global->__pyx_kp_u_transport_sockets_do_not_support
#define __pyx_n_s_transport_waiter __pyx_mstate_global->__pyx_n_s_transport_waiter
#define __pyx_n_s_transports __pyx_mstate_global->__pyx_n_s_transports
#define __pyx_kp_u_try_write_sent_all_data_and_ret __pyx_mstate_global->__pyx_kp_u_try_write_sent_all_data_and_ret
#define __pyx_n_s_type __pyx_mstate_global->__pyx_n_s_type
#define __pyx_n_s_udp __pyx_mstate_global->__pyx_n_s_udp
#define __pyx_n_u_udp __pyx_mstate_global->__pyx_n_u_udp
#define __pyx_kp_u_unable_to_listen_no_protocol_fac __pyx_mstate_global->__pyx_kp_u_unable_to_listen_no_protocol_fac
#define __pyx_kp_u_unable_to_perform_operation_on_r __pyx_mstate_global->__pyx_kp_u_unable_to_perform_operation_on_r
#define __pyx_kp_u_unable_to_start_the_loop_it_was __pyx_mstate_global->__pyx_kp_u_unable_to_start_the_loop_it_was
#define __pyx_kp_u_unclosed_resource_r __pyx_mstate_global->__pyx_kp_u_unclosed_resource_r
#define __pyx_kp_u_unclosed_resource_r_2 __pyx_mstate_global->__pyx_kp_u_unclosed_resource_r_2
#define __pyx_kp_u_unclosed_resource_r_could_not_se __pyx_mstate_global->__pyx_kp_u_unclosed_resource_r_could_not_se
#define __pyx_kp_u_unclosed_transport_uvloop_loop __pyx_mstate_global->__pyx_kp_u_unclosed_transport_uvloop_loop
#define __pyx_kp_u_unexpected_address_family __pyx_mstate_global->__pyx_kp_u_unexpected_address_family
#define __pyx_kp_u_unexpected_kwargs __pyx_mstate_global->__pyx_kp_u_unexpected_kwargs
#define __pyx_kp_u_unhandled_error_or_an_empty_buff __pyx_mstate_global->__pyx_kp_u_unhandled_error_or_an_empty_buff
#define __pyx_n_s_universal_newlines __pyx_mstate_global->__pyx_n_s_universal_newlines
#define __pyx_kp_u_universal_newlines_must_be_False __pyx_mstate_global->__pyx_kp_u_universal_newlines_must_be_False
#define __pyx_kp_u_unix_socket_path __pyx_mstate_global->__pyx_kp_u_unix_socket_path
#define __pyx_n_s_unlink __pyx_mstate_global->__pyx_n_s_unlink
#define __pyx_kp_u_unopened_TCPServer __pyx_mstate_global->__pyx_kp_u_unopened_TCPServer
#define __pyx_n_s_unset __pyx_mstate_global->__pyx_n_s_unset
#define __pyx_n_s_unwrap __pyx_mstate_global->__pyx_n_s_unwrap
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_kp_u_uv_close_handle_cb_handle_loop __pyx_mstate_global->__pyx_kp_u_uv_close_handle_cb_handle_loop
#define __pyx_kp_u_uv_handle_t_current_freed_total __pyx_mstate_global->__pyx_kp_u_uv_handle_t_current_freed_total
#define __pyx_kp_u_uvloop_PseudoSocket_fd_family_s __pyx_mstate_global->__pyx_kp_u_uvloop_PseudoSocket_fd_family_s
#define __pyx_kp_s_uvloop_cbhandles_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_cbhandles_pyx
#define __pyx_kp_s_uvloop_handles_basetransport_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_basetransport_pyx
#define __pyx_kp_s_uvloop_handles_fsevent_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_fsevent_pyx
#define __pyx_kp_s_uvloop_handles_pipe_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_pipe_pyx
#define __pyx_kp_s_uvloop_handles_process_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_process_pyx
#define __pyx_kp_s_uvloop_handles_stream_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_stream_pyx
#define __pyx_kp_s_uvloop_handles_streamserver_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_streamserver_pyx
#define __pyx_kp_s_uvloop_handles_tcp_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_tcp_pyx
#define __pyx_kp_s_uvloop_handles_udp_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_handles_udp_pyx
#define __pyx_n_s_uvloop_loop __pyx_mstate_global->__pyx_n_s_uvloop_loop
#define __pyx_kp_s_uvloop_loop_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_loop_pyx
#define __pyx_kp_s_uvloop_pseudosock_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_pseudosock_pyx
#define __pyx_kp_s_uvloop_server_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_server_pyx
#define __pyx_n_u_uvloop_sleep_after_fork __pyx_mstate_global->__pyx_n_u_uvloop_sleep_after_fork
#define __pyx_kp_u_uvloop_sslproto __pyx_mstate_global->__pyx_kp_u_uvloop_sslproto
#define __pyx_kp_s_uvloop_sslproto_pyx __pyx_mstate_global->__pyx_kp_s_uvloop_sslproto_pyx
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_n_s_version_info __pyx_mstate_global->__pyx_n_s_version_info
#define __pyx_n_s_wait __pyx_mstate_global->__pyx_n_s_wait
#define __pyx_n_s_wait_2 __pyx_mstate_global->__pyx_n_s_wait_2
#define __pyx_n_s_wait_closed __pyx_mstate_global->__pyx_n_s_wait_closed
#define __pyx_n_s_waiter __pyx_mstate_global->__pyx_n_s_waiter
#define __pyx_n_s_walk_stack __pyx_mstate_global->__pyx_n_s_walk_stack
#define __pyx_n_s_want __pyx_mstate_global->__pyx_n_s_want
#define __pyx_n_s_warn __pyx_mstate_global->__pyx_n_s_warn
#define __pyx_n_s_warn_on_full_buffer __pyx_mstate_global->__pyx_n_s_warn_on_full_buffer
#define __pyx_n_s_warning __pyx_mstate_global->__pyx_n_s_warning
#define __pyx_n_s_warnings __pyx_mstate_global->__pyx_n_s_warnings
#define __pyx_n_u_wb __pyx_mstate_global->__pyx_n_u_wb
#define __pyx_n_s_weakref __pyx_mstate_global->__pyx_n_s_weakref
#define __pyx_n_s_when __pyx_mstate_global->__pyx_n_s_when
#define __pyx_kp_u_without_no_gc_clear_loop_was_se __pyx_mstate_global->__pyx_kp_u_without_no_gc_clear_loop_was_se
#define __pyx_n_s_wrap __pyx_mstate_global->__pyx_n_s_wrap
#define __pyx_n_s_wrap_bio __pyx_mstate_global->__pyx_n_s_wrap_bio
#define __pyx_n_s_wrap_future __pyx_mstate_global->__pyx_n_s_wrap_future
#define __pyx_n_s_write __pyx_mstate_global->__pyx_n_s_write
#define __pyx_n_s_write_eof __pyx_mstate_global->__pyx_n_s_write_eof
#define __pyx_n_s_writelines __pyx_mstate_global->__pyx_n_s_writelines
#define __pyx_n_s_zip __pyx_mstate_global->__pyx_n_s_zip
#define __pyx_float_0_1 __pyx_mstate_global->__pyx_float_0_1
#define __pyx_float_1e3 __pyx_mstate_global->__pyx_float_1e3
#define __pyx_float_1eneg_3 __pyx_mstate_global->__pyx_float_1eneg_3
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_3 __pyx_mstate_global->__pyx_int_3
#define __pyx_int_15 __pyx_mstate_global->__pyx_int_15
#define __pyx_int_42 __pyx_mstate_global->__pyx_int_42
#define __pyx_int_100 __pyx_mstate_global->__pyx_int_100
#define __pyx_int_1000 __pyx_mstate_global->__pyx_int_1000
#define __pyx_int_50000 __pyx_mstate_global->__pyx_int_50000
#define __pyx_int_65536 __pyx_mstate_global->__pyx_int_65536
#define __pyx_int_1048575 __pyx_mstate_global->__pyx_int_1048575
#define __pyx_int_14808274 __pyx_mstate_global->__pyx_int_14808274
#define __pyx_int_87255255 __pyx_mstate_global->__pyx_int_87255255
#define __pyx_int_121062814 __pyx_mstate_global->__pyx_int_121062814
#define __pyx_int_143395587 __pyx_mstate_global->__pyx_int_143395587
#define __pyx_int_173349396 __pyx_mstate_global->__pyx_int_173349396
#define __pyx_int_200977156 __pyx_mstate_global->__pyx_int_200977156
#define __pyx_int_4294967296 __pyx_mstate_global->__pyx_int_4294967296
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_k__52 __pyx_mstate_global->__pyx_k__52
#define __pyx_k__53 __pyx_mstate_global->__pyx_k__53
#define __pyx_k__71 __pyx_mstate_global->__pyx_k__71
#define __pyx_k__95 __pyx_mstate_global->__pyx_k__95
#define __pyx_k__96 __pyx_mstate_global->__pyx_k__96
#define __pyx_k__97 __pyx_mstate_global->__pyx_k__97
#define __pyx_k__114 __pyx_mstate_global->__pyx_k__114
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__3 __pyx_mstate_global->__pyx_tuple__3
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_tuple__5 __pyx_mstate_global->__pyx_tuple__5
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_slice__88 __pyx_mstate_global->__pyx_slice__88
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__18 __pyx_mstate_global->__pyx_tuple__18
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__23 __pyx_mstate_global->__pyx_tuple__23
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__27 __pyx_mstate_global->__pyx_tuple__27
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__29 __pyx_mstate_global->__pyx_tuple__29
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__31 __pyx_mstate_global->__pyx_tuple__31
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__39 __pyx_mstate_global->__pyx_tuple__39
#define __pyx_tuple__40 __pyx_mstate_global->__pyx_tuple__40
#define __pyx_tuple__42 __pyx_mstate_global->__pyx_tuple__42
#define __pyx_tuple__45 __pyx_mstate_global->__pyx_tuple__45
#define __pyx_tuple__46 __pyx_mstate_global->__pyx_tuple__46
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__48 __pyx_mstate_global->__pyx_tuple__48
#define __pyx_tuple__49 __pyx_mstate_global->__pyx_tuple__49
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__55 __pyx_mstate_global->__pyx_tuple__55
#define __pyx_tuple__56 __pyx_mstate_global->__pyx_tuple__56
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__58 __pyx_mstate_global->__pyx_tuple__58
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__60 __pyx_mstate_global->__pyx_tuple__60
#define __pyx_tuple__62 __pyx_mstate_global->__pyx_tuple__62
#define __pyx_tuple__64 __pyx_mstate_global->__pyx_tuple__64
#define __pyx_tuple__65 __pyx_mstate_global->__pyx_tuple__65
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__67 __pyx_mstate_global->__pyx_tuple__67
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_tuple__70 __pyx_mstate_global->__pyx_tuple__70
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__74 __pyx_mstate_global->__pyx_tuple__74
#define __pyx_tuple__77 __pyx_mstate_global->__pyx_tuple__77
#define __pyx_tuple__80 __pyx_mstate_global->__pyx_tuple__80
#define __pyx_tuple__81 __pyx_mstate_global->__pyx_tuple__81
#define __pyx_tuple__83 __pyx_mstate_global->__pyx_tuple__83
#define __pyx_tuple__93 __pyx_mstate_global->__pyx_tuple__93
#define __pyx_tuple__94 __pyx_mstate_global->__pyx_tuple__94
#define __pyx_tuple__99 __pyx_mstate_global->__pyx_tuple__99
#define __pyx_codeobj__2 __pyx_mstate_global->__pyx_codeobj__2
#define __pyx_tuple__100 __pyx_mstate_global->__pyx_tuple__100
#define __pyx_tuple__101 __pyx_mstate_global->__pyx_tuple__101
#define __pyx_tuple__102 __pyx_mstate_global->__pyx_tuple__102
#define __pyx_tuple__104 __pyx_mstate_global->__pyx_tuple__104
#define __pyx_tuple__106 __pyx_mstate_global->__pyx_tuple__106
#define __pyx_tuple__109 __pyx_mstate_global->__pyx_tuple__109
#define __pyx_tuple__110 __pyx_mstate_global->__pyx_tuple__110
#define __pyx_tuple__111 __pyx_mstate_global->__pyx_tuple__111
#define __pyx_tuple__112 __pyx_mstate_global->__pyx_tuple__112
#define __pyx_tuple__113 __pyx_mstate_global->__pyx_tuple__113
#define __pyx_tuple__117 __pyx_mstate_global->__pyx_tuple__117
#define __pyx_tuple__118 __pyx_mstate_global->__pyx_tuple__118
#define __pyx_tuple__119 __pyx_mstate_global->__pyx_tuple__119
#define __pyx_tuple__120 __pyx_mstate_global->__pyx_tuple__120
#define __pyx_tuple__121 __pyx_mstate_global->__pyx_tuple__121
#define __pyx_tuple__122 __pyx_mstate_global->__pyx_tuple__122
#define __pyx_tuple__125 __pyx_mstate_global->__pyx_tuple__125
#define __pyx_tuple__126 __pyx_mstate_global->__pyx_tuple__126
#define __pyx_tuple__127 __pyx_mstate_global->__pyx_tuple__127
#define __pyx_tuple__131 __pyx_mstate_global->__pyx_tuple__131
#define __pyx_tuple__132 __pyx_mstate_global->__pyx_tuple__132
#define __pyx_tuple__133 __pyx_mstate_global->__pyx_tuple__133
#define __pyx_tuple__134 __pyx_mstate_global->__pyx_tuple__134
#define __pyx_tuple__135 __pyx_mstate_global->__pyx_tuple__135
#define __pyx_tuple__136 __pyx_mstate_global->__pyx_tuple__136
#define __pyx_tuple__137 __pyx_mstate_global->__pyx_tuple__137
#define __pyx_tuple__138 __pyx_mstate_global->__pyx_tuple__138
#define __pyx_tuple__139 __pyx_mstate_global->__pyx_tuple__139
#define __pyx_tuple__142 __pyx_mstate_global->__pyx_tuple__142
#define __pyx_tuple__143 __pyx_mstate_global->__pyx_tuple__143
#define __pyx_tuple__144 __pyx_mstate_global->__pyx_tuple__144
#define __pyx_tuple__145 __pyx_mstate_global->__pyx_tuple__145
#define __pyx_tuple__146 __pyx_mstate_global->__pyx_tuple__146
#define __pyx_tuple__147 __pyx_mstate_global->__pyx_tuple__147
#define __pyx_tuple__148 __pyx_mstate_global->__pyx_tuple__148
#define __pyx_tuple__149 __pyx_mstate_global->__pyx_tuple__149
#define __pyx_tuple__150 __pyx_mstate_global->__pyx_tuple__150
#define __pyx_tuple__151 __pyx_mstate_global->__pyx_tuple__151
#define __pyx_tuple__152 __pyx_mstate_global->__pyx_tuple__152
#define __pyx_tuple__153 __pyx_mstate_global->__pyx_tuple__153
#define __pyx_tuple__154 __pyx_mstate_global->__pyx_tuple__154
#define __pyx_tuple__155 __pyx_mstate_global->__pyx_tuple__155
#define __pyx_tuple__156 __pyx_mstate_global->__pyx_tuple__156
#define __pyx_tuple__157 __pyx_mstate_global->__pyx_tuple__157
#define __pyx_tuple__158 __pyx_mstate_global->__pyx_tuple__158
#define __pyx_tuple__159 __pyx_mstate_global->__pyx_tuple__159
#define __pyx_tuple__160 __pyx_mstate_global->__pyx_tuple__160
#define __pyx_tuple__162 __pyx_mstate_global->__pyx_tuple__162
#define __pyx_tuple__163 __pyx_mstate_global->__pyx_tuple__163
#define __pyx_tuple__164 __pyx_mstate_global->__pyx_tuple__164
#define __pyx_tuple__165 __pyx_mstate_global->__pyx_tuple__165
#define __pyx_tuple__166 __pyx_mstate_global->__pyx_tuple__166
#define __pyx_tuple__167 __pyx_mstate_global->__pyx_tuple__167
#define __pyx_tuple__168 __pyx_mstate_global->__pyx_tuple__168
#define __pyx_tuple__169 __pyx_mstate_global->__pyx_tuple__169
#define __pyx_tuple__171 __pyx_mstate_global->__pyx_tuple__171
#define __pyx_tuple__172 __pyx_mstate_global->__pyx_tuple__172
#define __pyx_tuple__173 __pyx_mstate_global->__pyx_tuple__173
#define __pyx_tuple__174 __pyx_mstate_global->__pyx_tuple__174
#define __pyx_tuple__175 __pyx_mstate_global->__pyx_tuple__175
#define __pyx_tuple__176 __pyx_mstate_global->__pyx_tuple__176
#define __pyx_tuple__177 __pyx_mstate_global->__pyx_tuple__177
#define __pyx_tuple__178 __pyx_mstate_global->__pyx_tuple__178
#define __pyx_tuple__179 __pyx_mstate_global->__pyx_tuple__179
#define __pyx_tuple__180 __pyx_mstate_global->__pyx_tuple__180
#define __pyx_tuple__181 __pyx_mstate_global->__pyx_tuple__181
#define __pyx_tuple__182 __pyx_mstate_global->__pyx_tuple__182
#define __pyx_tuple__183 __pyx_mstate_global->__pyx_tuple__183
#define __pyx_tuple__184 __pyx_mstate_global->__pyx_tuple__184
#define __pyx_tuple__185 __pyx_mstate_global->__pyx_tuple__185
#define __pyx_tuple__186 __pyx_mstate_global->__pyx_tuple__186
#define __pyx_tuple__187 __pyx_mstate_global->__pyx_tuple__187
#define __pyx_tuple__188 __pyx_mstate_global->__pyx_tuple__188
#define __pyx_tuple__195 __pyx_mstate_global->__pyx_tuple__195
#define __pyx_tuple__197 __pyx_mstate_global->__pyx_tuple__197
#define __pyx_tuple__199 __pyx_mstate_global->__pyx_tuple__199
#define __pyx_tuple__200 __pyx_mstate_global->__pyx_tuple__200
#define __pyx_tuple__201 __pyx_mstate_global->__pyx_tuple__201
#define __pyx_tuple__202 __pyx_mstate_global->__pyx_tuple__202
#define __pyx_tuple__203 __pyx_mstate_global->__pyx_tuple__203
#define __pyx_tuple__204 __pyx_mstate_global->__pyx_tuple__204
#define __pyx_tuple__205 __pyx_mstate_global->__pyx_tuple__205
#define __pyx_tuple__206 __pyx_mstate_global->__pyx_tuple__206
#define __pyx_tuple__207 __pyx_mstate_global->__pyx_tuple__207
#define __pyx_tuple__209 __pyx_mstate_global->__pyx_tuple__209
#define __pyx_tuple__211 __pyx_mstate_global->__pyx_tuple__211
#define __pyx_tuple__213 __pyx_mstate_global->__pyx_tuple__213
#define __pyx_tuple__215 __pyx_mstate_global->__pyx_tuple__215
#define __pyx_tuple__217 __pyx_mstate_global->__pyx_tuple__217
#define __pyx_tuple__221 __pyx_mstate_global->__pyx_tuple__221
#define __pyx_tuple__229 __pyx_mstate_global->__pyx_tuple__229
#define __pyx_tuple__231 __pyx_mstate_global->__pyx_tuple__231
#define __pyx_tuple__234 __pyx_mstate_global->__pyx_tuple__234
#define __pyx_tuple__236 __pyx_mstate_global->__pyx_tuple__236
#define __pyx_tuple__237 __pyx_mstate_global->__pyx_tuple__237
#define __pyx_tuple__238 __pyx_mstate_global->__pyx_tuple__238
#define __pyx_tuple__239 __pyx_mstate_global->__pyx_tuple__239
#define __pyx_tuple__240 __pyx_mstate_global->__pyx_tuple__240
#define __pyx_tuple__241 __pyx_mstate_global->__pyx_tuple__241
#define __pyx_tuple__242 __pyx_mstate_global->__pyx_tuple__242
#define __pyx_tuple__243 __pyx_mstate_global->__pyx_tuple__243
#define __pyx_tuple__244 __pyx_mstate_global->__pyx_tuple__244
#define __pyx_tuple__245 __pyx_mstate_global->__pyx_tuple__245
#define __pyx_tuple__246 __pyx_mstate_global->__pyx_tuple__246
#define __pyx_tuple__249 __pyx_mstate_global->__pyx_tuple__249
#define __pyx_tuple__251 __pyx_mstate_global->__pyx_tuple__251
#define __pyx_tuple__253 __pyx_mstate_global->__pyx_tuple__253
#define __pyx_tuple__255 __pyx_mstate_global->__pyx_tuple__255
#define __pyx_tuple__259 __pyx_mstate_global->__pyx_tuple__259
#define __pyx_tuple__260 __pyx_mstate_global->__pyx_tuple__260
#define __pyx_tuple__261 __pyx_mstate_global->__pyx_tuple__261
#define __pyx_tuple__262 __pyx_mstate_global->__pyx_tuple__262
#define __pyx_tuple__264 __pyx_mstate_global->__pyx_tuple__264
#define __pyx_tuple__265 __pyx_mstate_global->__pyx_tuple__265
#define __pyx_tuple__266 __pyx_mstate_global->__pyx_tuple__266
#define __pyx_tuple__267 __pyx_mstate_global->__pyx_tuple__267
#define __pyx_tuple__268 __pyx_mstate_global->__pyx_tuple__268
#define __pyx_tuple__269 __pyx_mstate_global->__pyx_tuple__269
#define __pyx_tuple__271 __pyx_mstate_global->__pyx_tuple__271
#define __pyx_tuple__273 __pyx_mstate_global->__pyx_tuple__273
#define __pyx_tuple__274 __pyx_mstate_global->__pyx_tuple__274
#define __pyx_tuple__275 __pyx_mstate_global->__pyx_tuple__275
#define __pyx_tuple__276 __pyx_mstate_global->__pyx_tuple__276
#define __pyx_tuple__277 __pyx_mstate_global->__pyx_tuple__277
#define __pyx_tuple__279 __pyx_mstate_global->__pyx_tuple__279
#define __pyx_tuple__281 __pyx_mstate_global->__pyx_tuple__281
#define __pyx_tuple__282 __pyx_mstate_global->__pyx_tuple__282
#define __pyx_tuple__285 __pyx_mstate_global->__pyx_tuple__285
#define __pyx_tuple__288 __pyx_mstate_global->__pyx_tuple__288
#define __pyx_tuple__289 __pyx_mstate_global->__pyx_tuple__289
#define __pyx_tuple__290 __pyx_mstate_global->__pyx_tuple__290
#define __pyx_tuple__293 __pyx_mstate_global->__pyx_tuple__293
#define __pyx_tuple__295 __pyx_mstate_global->__pyx_tuple__295
#define __pyx_tuple__298 __pyx_mstate_global->__pyx_tuple__298
#define __pyx_tuple__300 __pyx_mstate_global->__pyx_tuple__300
#define __pyx_tuple__303 __pyx_mstate_global->__pyx_tuple__303
#define __pyx_tuple__321 __pyx_mstate_global->__pyx_tuple__321
#define __pyx_tuple__325 __pyx_mstate_global->__pyx_tuple__325
#define __pyx_tuple__327 __pyx_mstate_global->__pyx_tuple__327
#define __pyx_tuple__330 __pyx_mstate_global->__pyx_tuple__330
#define __pyx_tuple__333 __pyx_mstate_global->__pyx_tuple__333
#define __pyx_tuple__336 __pyx_mstate_global->__pyx_tuple__336
#define __pyx_tuple__357 __pyx_mstate_global->__pyx_tuple__357
#define __pyx_tuple__360 __pyx_mstate_global->__pyx_tuple__360
#define __pyx_tuple__363 __pyx_mstate_global->__pyx_tuple__363
#define __pyx_tuple__365 __pyx_mstate_global->__pyx_tuple__365
#define __pyx_tuple__385 __pyx_mstate_global->__pyx_tuple__385
#define __pyx_tuple__387 __pyx_mstate_global->__pyx_tuple__387
#define __pyx_tuple__393 __pyx_mstate_global->__pyx_tuple__393
#define __pyx_tuple__396 __pyx_mstate_global->__pyx_tuple__396
#define __pyx_tuple__402 __pyx_mstate_global->__pyx_tuple__402
#define __pyx_tuple__404 __pyx_mstate_global->__pyx_tuple__404
#define __pyx_tuple__413 __pyx_mstate_global->__pyx_tuple__413
#define __pyx_tuple__431 __pyx_mstate_global->__pyx_tuple__431
#define __pyx_tuple__433 __pyx_mstate_global->__pyx_tuple__433
#define __pyx_tuple__448 __pyx_mstate_global->__pyx_tuple__448
#define __pyx_tuple__454 __pyx_mstate_global->__pyx_tuple__454
#define __pyx_tuple__458 __pyx_mstate_global->__pyx_tuple__458
#define __pyx_tuple__463 __pyx_mstate_global->__pyx_tuple__463
#define __pyx_tuple__467 __pyx_mstate_global->__pyx_tuple__467
#define __pyx_tuple__469 __pyx_mstate_global->__pyx_tuple__469
#define __pyx_tuple__513 __pyx_mstate_global->__pyx_tuple__513
#define __pyx_tuple__517 __pyx_mstate_global->__pyx_tuple__517
#define __pyx_tuple__519 __pyx_mstate_global->__pyx_tuple__519
#define __pyx_tuple__528 __pyx_mstate_global->__pyx_tuple__528
#define __pyx_tuple__533 __pyx_mstate_global->__pyx_tuple__533
#define __pyx_tuple__534 __pyx_mstate_global->__pyx_tuple__534
#define __pyx_tuple__539 __pyx_mstate_global->__pyx_tuple__539
#define __pyx_codeobj__20 __pyx_mstate_global->__pyx_codeobj__20
#define __pyx_codeobj__22 __pyx_mstate_global->__pyx_codeobj__22
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__84 __pyx_mstate_global->__pyx_codeobj__84
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
#define __pyx_codeobj__89 __pyx_mstate_global->__pyx_codeobj__89
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__91 __pyx_mstate_global->__pyx_codeobj__91
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__98 __pyx_mstate_global->__pyx_codeobj__98
#define __pyx_codeobj__103 __pyx_mstate_global->__pyx_codeobj__103
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__107 __pyx_mstate_global->__pyx_codeobj__107
#define __pyx_codeobj__108 __pyx_mstate_global->__pyx_codeobj__108
#define __pyx_codeobj__116 __pyx_mstate_global->__pyx_codeobj__116
#define __pyx_codeobj__123 __pyx_mstate_global->__pyx_codeobj__123
#define __pyx_codeobj__124 __pyx_mstate_global->__pyx_codeobj__124
#define __pyx_codeobj__189 __pyx_mstate_global->__pyx_codeobj__189
#define __pyx_codeobj__190 __pyx_mstate_global->__pyx_codeobj__190
#define __pyx_codeobj__191 __pyx_mstate_global->__pyx_codeobj__191
#define __pyx_codeobj__192 __pyx_mstate_global->__pyx_codeobj__192
#define __pyx_codeobj__193 __pyx_mstate_global->__pyx_codeobj__193
#define __pyx_codeobj__194 __pyx_mstate_global->__pyx_codeobj__194
#define __pyx_codeobj__208 __pyx_mstate_global->__pyx_codeobj__208
#define __pyx_codeobj__210 __pyx_mstate_global->__pyx_codeobj__210
#define __pyx_codeobj__212 __pyx_mstate_global->__pyx_codeobj__212
#define __pyx_codeobj__214 __pyx_mstate_global->__pyx_codeobj__214
#define __pyx_codeobj__216 __pyx_mstate_global->__pyx_codeobj__216
#define __pyx_codeobj__218 __pyx_mstate_global->__pyx_codeobj__218
#define __pyx_codeobj__219 __pyx_mstate_global->__pyx_codeobj__219
#define __pyx_codeobj__220 __pyx_mstate_global->__pyx_codeobj__220
#define __pyx_codeobj__222 __pyx_mstate_global->__pyx_codeobj__222
#define __pyx_codeobj__223 __pyx_mstate_global->__pyx_codeobj__223
#define __pyx_codeobj__224 __pyx_mstate_global->__pyx_codeobj__224
#define __pyx_codeobj__225 __pyx_mstate_global->__pyx_codeobj__225
#define __pyx_codeobj__226 __pyx_mstate_global->__pyx_codeobj__226
#define __pyx_codeobj__227 __pyx_mstate_global->__pyx_codeobj__227
#define __pyx_codeobj__228 __pyx_mstate_global->__pyx_codeobj__228
#define __pyx_codeobj__230 __pyx_mstate_global->__pyx_codeobj__230
#define __pyx_codeobj__232 __pyx_mstate_global->__pyx_codeobj__232
#define __pyx_codeobj__233 __pyx_mstate_global->__pyx_codeobj__233
#define __pyx_codeobj__235 __pyx_mstate_global->__pyx_codeobj__235
#define __pyx_codeobj__247 __pyx_mstate_global->__pyx_codeobj__247
#define __pyx_codeobj__248 __pyx_mstate_global->__pyx_codeobj__248
#define __pyx_codeobj__250 __pyx_mstate_global->__pyx_codeobj__250
#define __pyx_codeobj__252 __pyx_mstate_global->__pyx_codeobj__252
#define __pyx_codeobj__254 __pyx_mstate_global->__pyx_codeobj__254
#define __pyx_codeobj__256 __pyx_mstate_global->__pyx_codeobj__256
#define __pyx_codeobj__257 __pyx_mstate_global->__pyx_codeobj__257
#define __pyx_codeobj__258 __pyx_mstate_global->__pyx_codeobj__258
#define __pyx_codeobj__270 __pyx_mstate_global->__pyx_codeobj__270
#define __pyx_codeobj__272 __pyx_mstate_global->__pyx_codeobj__272
#define __pyx_codeobj__278 __pyx_mstate_global->__pyx_codeobj__278
#define __pyx_codeobj__280 __pyx_mstate_global->__pyx_codeobj__280
#define __pyx_codeobj__283 __pyx_mstate_global->__pyx_codeobj__283
#define __pyx_codeobj__284 __pyx_mstate_global->__pyx_codeobj__284
#define __pyx_codeobj__286 __pyx_mstate_global->__pyx_codeobj__286
#define __pyx_codeobj__287 __pyx_mstate_global->__pyx_codeobj__287
#define __pyx_codeobj__291 __pyx_mstate_global->__pyx_codeobj__291
#define __pyx_codeobj__292 __pyx_mstate_global->__pyx_codeobj__292
#define __pyx_codeobj__294 __pyx_mstate_global->__pyx_codeobj__294
#define __pyx_codeobj__296 __pyx_mstate_global->__pyx_codeobj__296
#define __pyx_codeobj__297 __pyx_mstate_global->__pyx_codeobj__297
#define __pyx_codeobj__299 __pyx_mstate_global->__pyx_codeobj__299
#define __pyx_codeobj__301 __pyx_mstate_global->__pyx_codeobj__301
#define __pyx_codeobj__302 __pyx_mstate_global->__pyx_codeobj__302
#define __pyx_codeobj__304 __pyx_mstate_global->__pyx_codeobj__304
#define __pyx_codeobj__305 __pyx_mstate_global->__pyx_codeobj__305
#define __pyx_codeobj__306 __pyx_mstate_global->__pyx_codeobj__306
#define __pyx_codeobj__307 __pyx_mstate_global->__pyx_codeobj__307
#define __pyx_codeobj__308 __pyx_mstate_global->__pyx_codeobj__308
#define __pyx_codeobj__309 __pyx_mstate_global->__pyx_codeobj__309
#define __pyx_codeobj__310 __pyx_mstate_global->__pyx_codeobj__310
#define __pyx_codeobj__311 __pyx_mstate_global->__pyx_codeobj__311
#define __pyx_codeobj__312 __pyx_mstate_global->__pyx_codeobj__312
#define __pyx_codeobj__313 __pyx_mstate_global->__pyx_codeobj__313
#define __pyx_codeobj__314 __pyx_mstate_global->__pyx_codeobj__314
#define __pyx_codeobj__315 __pyx_mstate_global->__pyx_codeobj__315
#define __pyx_codeobj__316 __pyx_mstate_global->__pyx_codeobj__316
#define __pyx_codeobj__317 __pyx_mstate_global->__pyx_codeobj__317
#define __pyx_codeobj__318 __pyx_mstate_global->__pyx_codeobj__318
#define __pyx_codeobj__319 __pyx_mstate_global->__pyx_codeobj__319
#define __pyx_codeobj__320 __pyx_mstate_global->__pyx_codeobj__320
#define __pyx_codeobj__322 __pyx_mstate_global->__pyx_codeobj__322
#define __pyx_codeobj__323 __pyx_mstate_global->__pyx_codeobj__323
#define __pyx_codeobj__324 __pyx_mstate_global->__pyx_codeobj__324
#define __pyx_codeobj__326 __pyx_mstate_global->__pyx_codeobj__326
#define __pyx_codeobj__328 __pyx_mstate_global->__pyx_codeobj__328
#define __pyx_codeobj__329 __pyx_mstate_global->__pyx_codeobj__329
#define __pyx_codeobj__331 __pyx_mstate_global->__pyx_codeobj__331
#define __pyx_codeobj__332 __pyx_mstate_global->__pyx_codeobj__332
#define __pyx_codeobj__334 __pyx_mstate_global->__pyx_codeobj__334
#define __pyx_codeobj__335 __pyx_mstate_global->__pyx_codeobj__335
#define __pyx_codeobj__337 __pyx_mstate_global->__pyx_codeobj__337
#define __pyx_codeobj__338 __pyx_mstate_global->__pyx_codeobj__338
#define __pyx_codeobj__339 __pyx_mstate_global->__pyx_codeobj__339
#define __pyx_codeobj__340 __pyx_mstate_global->__pyx_codeobj__340
#define __pyx_codeobj__341 __pyx_mstate_global->__pyx_codeobj__341
#define __pyx_codeobj__342 __pyx_mstate_global->__pyx_codeobj__342
#define __pyx_codeobj__343 __pyx_mstate_global->__pyx_codeobj__343
#define __pyx_codeobj__344 __pyx_mstate_global->__pyx_codeobj__344
#define __pyx_codeobj__345 __pyx_mstate_global->__pyx_codeobj__345
#define __pyx_codeobj__346 __pyx_mstate_global->__pyx_codeobj__346
#define __pyx_codeobj__347 __pyx_mstate_global->__pyx_codeobj__347
#define __pyx_codeobj__348 __pyx_mstate_global->__pyx_codeobj__348
#define __pyx_codeobj__349 __pyx_mstate_global->__pyx_codeobj__349
#define __pyx_codeobj__350 __pyx_mstate_global->__pyx_codeobj__350
#define __pyx_codeobj__351 __pyx_mstate_global->__pyx_codeobj__351
#define __pyx_codeobj__352 __pyx_mstate_global->__pyx_codeobj__352
#define __pyx_codeobj__353 __pyx_mstate_global->__pyx_codeobj__353
#define __pyx_codeobj__354 __pyx_mstate_global->__pyx_codeobj__354
#define __pyx_codeobj__355 __pyx_mstate_global->__pyx_codeobj__355
#define __pyx_codeobj__356 __pyx_mstate_global->__pyx_codeobj__356
#define __pyx_codeobj__358 __pyx_mstate_global->__pyx_codeobj__358
#define __pyx_codeobj__359 __pyx_mstate_global->__pyx_codeobj__359
#define __pyx_codeobj__361 __pyx_mstate_global->__pyx_codeobj__361
#define __pyx_codeobj__362 __pyx_mstate_global->__pyx_codeobj__362
#define __pyx_codeobj__364 __pyx_mstate_global->__pyx_codeobj__364
#define __pyx_codeobj__366 __pyx_mstate_global->__pyx_codeobj__366
#define __pyx_codeobj__367 __pyx_mstate_global->__pyx_codeobj__367
#define __pyx_codeobj__368 __pyx_mstate_global->__pyx_codeobj__368
#define __pyx_codeobj__369 __pyx_mstate_global->__pyx_codeobj__369
#define __pyx_codeobj__370 __pyx_mstate_global->__pyx_codeobj__370
#define __pyx_codeobj__371 __pyx_mstate_global->__pyx_codeobj__371
#define __pyx_codeobj__372 __pyx_mstate_global->__pyx_codeobj__372
#define __pyx_codeobj__373 __pyx_mstate_global->__pyx_codeobj__373
#define __pyx_codeobj__374 __pyx_mstate_global->__pyx_codeobj__374
#define __pyx_codeobj__375 __pyx_mstate_global->__pyx_codeobj__375
#define __pyx_codeobj__376 __pyx_mstate_global->__pyx_codeobj__376
#define __pyx_codeobj__377 __pyx_mstate_global->__pyx_codeobj__377
#define __pyx_codeobj__378 __pyx_mstate_global->__pyx_codeobj__378
#define __pyx_codeobj__379 __pyx_mstate_global->__pyx_codeobj__379
#define __pyx_codeobj__380 __pyx_mstate_global->__pyx_codeobj__380
#define __pyx_codeobj__381 __pyx_mstate_global->__pyx_codeobj__381
#define __pyx_codeobj__382 __pyx_mstate_global->__pyx_codeobj__382
#define __pyx_codeobj__383 __pyx_mstate_global->__pyx_codeobj__383
#define __pyx_codeobj__384 __pyx_mstate_global->__pyx_codeobj__384
#define __pyx_codeobj__386 __pyx_mstate_global->__pyx_codeobj__386
#define __pyx_codeobj__388 __pyx_mstate_global->__pyx_codeobj__388
#define __pyx_codeobj__389 __pyx_mstate_global->__pyx_codeobj__389
#define __pyx_codeobj__390 __pyx_mstate_global->__pyx_codeobj__390
#define __pyx_codeobj__391 __pyx_mstate_global->__pyx_codeobj__391
#define __pyx_codeobj__392 __pyx_mstate_global->__pyx_codeobj__392
#define __pyx_codeobj__394 __pyx_mstate_global->__pyx_codeobj__394
#define __pyx_codeobj__395 __pyx_mstate_global->__pyx_codeobj__395
#define __pyx_codeobj__397 __pyx_mstate_global->__pyx_codeobj__397
#define __pyx_codeobj__398 __pyx_mstate_global->__pyx_codeobj__398
#define __pyx_codeobj__399 __pyx_mstate_global->__pyx_codeobj__399
#define __pyx_codeobj__400 __pyx_mstate_global->__pyx_codeobj__400
#define __pyx_codeobj__401 __pyx_mstate_global->__pyx_codeobj__401
#define __pyx_codeobj__403 __pyx_mstate_global->__pyx_codeobj__403
#define __pyx_codeobj__405 __pyx_mstate_global->__pyx_codeobj__405
#define __pyx_codeobj__406 __pyx_mstate_global->__pyx_codeobj__406
#define __pyx_codeobj__407 __pyx_mstate_global->__pyx_codeobj__407
#define __pyx_codeobj__408 __pyx_mstate_global->__pyx_codeobj__408
#define __pyx_codeobj__409 __pyx_mstate_global->__pyx_codeobj__409
#define __pyx_codeobj__410 __pyx_mstate_global->__pyx_codeobj__410
#define __pyx_codeobj__411 __pyx_mstate_global->__pyx_codeobj__411
#define __pyx_codeobj__412 __pyx_mstate_global->__pyx_codeobj__412
#define __pyx_codeobj__414 __pyx_mstate_global->__pyx_codeobj__414
#define __pyx_codeobj__415 __pyx_mstate_global->__pyx_codeobj__415
#define __pyx_codeobj__416 __pyx_mstate_global->__pyx_codeobj__416
#define __pyx_codeobj__417 __pyx_mstate_global->__pyx_codeobj__417
#define __pyx_codeobj__418 __pyx_mstate_global->__pyx_codeobj__418
#define __pyx_codeobj__419 __pyx_mstate_global->__pyx_codeobj__419
#define __pyx_codeobj__420 __pyx_mstate_global->__pyx_codeobj__420
#define __pyx_codeobj__421 __pyx_mstate_global->__pyx_codeobj__421
#define __pyx_codeobj__422 __pyx_mstate_global->__pyx_codeobj__422
#define __pyx_codeobj__423 __pyx_mstate_global->__pyx_codeobj__423
#define __pyx_codeobj__424 __pyx_mstate_global->__pyx_codeobj__424
#define __pyx_codeobj__425 __pyx_mstate_global->__pyx_codeobj__425
#define __pyx_codeobj__426 __pyx_mstate_global->__pyx_codeobj__426
#define __pyx_codeobj__427 __pyx_mstate_global->__pyx_codeobj__427
#define __pyx_codeobj__428 __pyx_mstate_global->__pyx_codeobj__428
#define __pyx_codeobj__429 __pyx_mstate_global->__pyx_codeobj__429
#define __pyx_codeobj__430 __pyx_mstate_global->__pyx_codeobj__430
#define __pyx_codeobj__432 __pyx_mstate_global->__pyx_codeobj__432
#define __pyx_codeobj__434 __pyx_mstate_global->__pyx_codeobj__434
#define __pyx_codeobj__435 __pyx_mstate_global->__pyx_codeobj__435
#define __pyx_codeobj__436 __pyx_mstate_global->__pyx_codeobj__436
#define __pyx_codeobj__437 __pyx_mstate_global->__pyx_codeobj__437
#define __pyx_codeobj__438 __pyx_mstate_global->__pyx_codeobj__438
#define __pyx_codeobj__439 __pyx_mstate_global->__pyx_codeobj__439
#define __pyx_codeobj__440 __pyx_mstate_global->__pyx_codeobj__440
#define __pyx_codeobj__441 __pyx_mstate_global->__pyx_codeobj__441
#define __pyx_codeobj__442 __pyx_mstate_global->__pyx_codeobj__442
#define __pyx_codeobj__443 __pyx_mstate_global->__pyx_codeobj__443
#define __pyx_codeobj__444 __pyx_mstate_global->__pyx_codeobj__444
#define __pyx_codeobj__445 __pyx_mstate_global->__pyx_codeobj__445
#define __pyx_codeobj__446 __pyx_mstate_global->__pyx_codeobj__446
#define __pyx_codeobj__447 __pyx_mstate_global->__pyx_codeobj__447
#define __pyx_codeobj__449 __pyx_mstate_global->__pyx_codeobj__449
#define __pyx_codeobj__450 __pyx_mstate_global->__pyx_codeobj__450
#define __pyx_codeobj__451 __pyx_mstate_global->__pyx_codeobj__451
#define __pyx_codeobj__452 __pyx_mstate_global->__pyx_codeobj__452
#define __pyx_codeobj__453 __pyx_mstate_global->__pyx_codeobj__453
#define __pyx_codeobj__455 __pyx_mstate_global->__pyx_codeobj__455
#define __pyx_codeobj__456 __pyx_mstate_global->__pyx_codeobj__456
#define __pyx_codeobj__457 __pyx_mstate_global->__pyx_codeobj__457
#define __pyx_codeobj__459 __pyx_mstate_global->__pyx_codeobj__459
#define __pyx_codeobj__460 __pyx_mstate_global->__pyx_codeobj__460
#define __pyx_codeobj__461 __pyx_mstate_global->__pyx_codeobj__461
#define __pyx_codeobj__462 __pyx_mstate_global->__pyx_codeobj__462
#define __pyx_codeobj__464 __pyx_mstate_global->__pyx_codeobj__464
#define __pyx_codeobj__465 __pyx_mstate_global->__pyx_codeobj__465
#define __pyx_codeobj__466 __pyx_mstate_global->__pyx_codeobj__466
#define __pyx_codeobj__468 __pyx_mstate_global->__pyx_codeobj__468
#define __pyx_codeobj__470 __pyx_mstate_global->__pyx_codeobj__470
#define __pyx_codeobj__471 __pyx_mstate_global->__pyx_codeobj__471
#define __pyx_codeobj__472 __pyx_mstate_global->__pyx_codeobj__472
#define __pyx_codeobj__473 __pyx_mstate_global->__pyx_codeobj__473
#define __pyx_codeobj__474 __pyx_mstate_global->__pyx_codeobj__474
#define __pyx_codeobj__475 __pyx_mstate_global->__pyx_codeobj__475
#define __pyx_codeobj__476 __pyx_mstate_global->__pyx_codeobj__476
#define __pyx_codeobj__477 __pyx_mstate_global->__pyx_codeobj__477
#define __pyx_codeobj__478 __pyx_mstate_global->__pyx_codeobj__478
#define __pyx_codeobj__479 __pyx_mstate_global->__pyx_codeobj__479
#define __pyx_codeobj__480 __pyx_mstate_global->__pyx_codeobj__480
#define __pyx_codeobj__481 __pyx_mstate_global->__pyx_codeobj__481
#define __pyx_codeobj__482 __pyx_mstate_global->__pyx_codeobj__482
#define __pyx_codeobj__483 __pyx_mstate_global->__pyx_codeobj__483
#define __pyx_codeobj__484 __pyx_mstate_global->__pyx_codeobj__484
#define __pyx_codeobj__485 __pyx_mstate_global->__pyx_codeobj__485
#define __pyx_codeobj__486 __pyx_mstate_global->__pyx_codeobj__486
#define __pyx_codeobj__487 __pyx_mstate_global->__pyx_codeobj__487
#define __pyx_codeobj__488 __pyx_mstate_global->__pyx_codeobj__488
#define __pyx_codeobj__489 __pyx_mstate_global->__pyx_codeobj__489
#define __pyx_codeobj__490 __pyx_mstate_global->__pyx_codeobj__490
#define __pyx_codeobj__491 __pyx_mstate_global->__pyx_codeobj__491
#define __pyx_codeobj__492 __pyx_mstate_global->__pyx_codeobj__492
#define __pyx_codeobj__493 __pyx_mstate_global->__pyx_codeobj__493
#define __pyx_codeobj__494 __pyx_mstate_global->__pyx_codeobj__494
#define __pyx_codeobj__495 __pyx_mstate_global->__pyx_codeobj__495
#define __pyx_codeobj__496 __pyx_mstate_global->__pyx_codeobj__496
#define __pyx_codeobj__497 __pyx_mstate_global->__pyx_codeobj__497
#define __pyx_codeobj__498 __pyx_mstate_global->__pyx_codeobj__498
#define __pyx_codeobj__499 __pyx_mstate_global->__pyx_codeobj__499
#define __pyx_codeobj__500 __pyx_mstate_global->__pyx_codeobj__500
#define __pyx_codeobj__501 __pyx_mstate_global->__pyx_codeobj__501
#define __pyx_codeobj__502 __pyx_mstate_global->__pyx_codeobj__502
#define __pyx_codeobj__503 __pyx_mstate_global->__pyx_codeobj__503
#define __pyx_codeobj__504 __pyx_mstate_global->__pyx_codeobj__504
#define __pyx_codeobj__505 __pyx_mstate_global->__pyx_codeobj__505
#define __pyx_codeobj__506 __pyx_mstate_global->__pyx_codeobj__506
#define __pyx_codeobj__507 __pyx_mstate_global->__pyx_codeobj__507
#define __pyx_codeobj__508 __pyx_mstate_global->__pyx_codeobj__508
#define __pyx_codeobj__509 __pyx_mstate_global->__pyx_codeobj__509
#define __pyx_codeobj__510 __pyx_mstate_global->__pyx_codeobj__510
#define __pyx_codeobj__511 __pyx_mstate_global->__pyx_codeobj__511
#define __pyx_codeobj__512 __pyx_mstate_global->__pyx_codeobj__512
#define __pyx_codeobj__514 __pyx_mstate_global->__pyx_codeobj__514
#define __pyx_codeobj__515 __pyx_mstate_global->__pyx_codeobj__515
#define __pyx_codeobj__516 __pyx_mstate_global->__pyx_codeobj__516
#define __pyx_codeobj__518 __pyx_mstate_global->__pyx_codeobj__518
#define __pyx_codeobj__520 __pyx_mstate_global->__pyx_codeobj__520
#define __pyx_codeobj__521 __pyx_mstate_global->__pyx_codeobj__521
#define __pyx_codeobj__522 __pyx_mstate_global->__pyx_codeobj__522
#define __pyx_codeobj__523 __pyx_mstate_global->__pyx_codeobj__523
#define __pyx_codeobj__524 __pyx_mstate_global->__pyx_codeobj__524
#define __pyx_codeobj__525 __pyx_mstate_global->__pyx_codeobj__525
#define __pyx_codeobj__526 __pyx_mstate_global->__pyx_codeobj__526
#define __pyx_codeobj__527 __pyx_mstate_global->__pyx_codeobj__527
#define __pyx_codeobj__529 __pyx_mstate_global->__pyx_codeobj__529
#define __pyx_codeobj__530 __pyx_mstate_global->__pyx_codeobj__530
#define __pyx_codeobj__531 __pyx_mstate_global->__pyx_codeobj__531
#define __pyx_codeobj__532 __pyx_mstate_global->__pyx_codeobj__532
#define __pyx_codeobj__535 __pyx_mstate_global->__pyx_codeobj__535
#define __pyx_codeobj__536 __pyx_mstate_global->__pyx_codeobj__536
#define __pyx_codeobj__537 __pyx_mstate_global->__pyx_codeobj__537
#define __pyx_codeobj__538 __pyx_mstate_global->__pyx_codeobj__538
#define __pyx_codeobj__540 __pyx_mstate_global->__pyx_codeobj__540
#define __pyx_codeobj__541 __pyx_mstate_global->__pyx_codeobj__541
/* #### Code section: module_code ### */

/* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled")
 * cdef object __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(object (*f)(Loop, bint) ):
 *     def wrap(Loop self, bint enabled):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Loop', enabled: bool)"""
 *         return f(self, enabled)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_wrap, "wrap(self: 'Loop', enabled: bool)");
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_1wrap = {"wrap", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_1wrap, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_1wrap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self = 0;
  int __pyx_v_enabled;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_enabled,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_enabled)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("wrap", 1, 2, 2, 1); __PYX_ERR(6, 67, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "wrap") < 0)) __PYX_ERR(6, 67, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = ((struct __pyx_obj_6uvloop_4loop_Loop *)values[0]);
    __pyx_v_enabled = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_enabled == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 67, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wrap", 1, 2, 2, __pyx_nargs); __PYX_ERR(6, 67, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6uvloop_4loop_Loop, 1, "self", 0))) __PYX_ERR(6, 67, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_wrap(__pyx_self, __pyx_v_self, __pyx_v_enabled);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_wrap(PyObject *__pyx_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, int __pyx_v_enabled) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap", 1);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":69
 *     def wrap(Loop self, bint enabled):
 *         """wrap(self: 'Loop', enabled: bool)"""
 *         return f(self, enabled)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_f(__pyx_v_self, __pyx_v_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled")
 * cdef object __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(object (*f)(Loop, bint) ):
 *     def wrap(Loop self, bint enabled):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Loop', enabled: bool)"""
 *         return f(self, enabled)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled")
 * cdef object __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(object (*f)(Loop, bint) ):             # <<<<<<<<<<<<<<
 *     def wrap(Loop self, bint enabled):
 *         """wrap(self: 'Loop', enabled: bool)"""
 */

static PyObject *__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(PyObject *(*__pyx_v_f)(struct __pyx_obj_6uvloop_4loop_Loop *, int)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(6, 66, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled")
 * cdef object __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(object (*f)(Loop, bint) ):
 *     def wrap(Loop self, bint enabled):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Loop', enabled: bool)"""
 *         return f(self, enabled)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_11cfunc_dot_to_py_91__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_1wrap, 0, __pyx_n_s_Pyx_CFunc_6uvloop_4loop_4Loop, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":70
 *         """wrap(self: 'Loop', enabled: bool)"""
 *         return f(self, enabled)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":66
 * 
 * @cname("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled")
 * cdef object __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(object (*f)(Loop, bint) ):             # <<<<<<<<<<<<<<
 *     def wrap(Loop self, bint enabled):
 *         """wrap(self: 'Loop', enabled: bool)"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "includes/flowcontrol.pxd":4
 * 
 * 
 * cdef inline add_flowcontrol_defaults(high, low, int kb):             # <<<<<<<<<<<<<<
 *     cdef int h, l
 *     if high is None:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_8includes_11flowcontrol_add_flowcontrol_defaults(PyObject *__pyx_v_high, PyObject *__pyx_v_low, int __pyx_v_kb) {
  int __pyx_v_h;
  int __pyx_v_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_flowcontrol_defaults", 1);

  /* "includes/flowcontrol.pxd":6
 * cdef inline add_flowcontrol_defaults(high, low, int kb):
 *     cdef int h, l
 *     if high is None:             # <<<<<<<<<<<<<<
 *         if low is None:
 *             h = kb * 1024
 */
  __pyx_t_1 = (__pyx_v_high == Py_None);
  if (__pyx_t_1) {

    /* "includes/flowcontrol.pxd":7
 *     cdef int h, l
 *     if high is None:
 *         if low is None:             # <<<<<<<<<<<<<<
 *             h = kb * 1024
 *         else:
 */
    __pyx_t_1 = (__pyx_v_low == Py_None);
    if (__pyx_t_1) {

      /* "includes/flowcontrol.pxd":8
 *     if high is None:
 *         if low is None:
 *             h = kb * 1024             # <<<<<<<<<<<<<<
 *         else:
 *             l = low
 */
      __pyx_v_h = (__pyx_v_kb * 0x400);

      /* "includes/flowcontrol.pxd":7
 *     cdef int h, l
 *     if high is None:
 *         if low is None:             # <<<<<<<<<<<<<<
 *             h = kb * 1024
 *         else:
 */
      goto __pyx_L4;
    }

    /* "includes/flowcontrol.pxd":10
 *             h = kb * 1024
 *         else:
 *             l = low             # <<<<<<<<<<<<<<
 *             h = 4 * l
 *     else:
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_low); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(7, 10, __pyx_L1_error)
      __pyx_v_l = __pyx_t_2;

      /* "includes/flowcontrol.pxd":11
 *         else:
 *             l = low
 *             h = 4 * l             # <<<<<<<<<<<<<<
 *     else:
 *         h = high
 */
      __pyx_v_h = (4 * __pyx_v_l);
    }
    __pyx_L4:;

    /* "includes/flowcontrol.pxd":6
 * cdef inline add_flowcontrol_defaults(high, low, int kb):
 *     cdef int h, l
 *     if high is None:             # <<<<<<<<<<<<<<
 *         if low is None:
 *             h = kb * 1024
 */
    goto __pyx_L3;
  }

  /* "includes/flowcontrol.pxd":13
 *             h = 4 * l
 *     else:
 *         h = high             # <<<<<<<<<<<<<<
 *     if low is None:
 *         l = h // 4
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_high); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(7, 13, __pyx_L1_error)
    __pyx_v_h = __pyx_t_2;
  }
  __pyx_L3:;

  /* "includes/flowcontrol.pxd":14
 *     else:
 *         h = high
 *     if low is None:             # <<<<<<<<<<<<<<
 *         l = h // 4
 *     else:
 */
  __pyx_t_1 = (__pyx_v_low == Py_None);
  if (__pyx_t_1) {

    /* "includes/flowcontrol.pxd":15
 *         h = high
 *     if low is None:
 *         l = h // 4             # <<<<<<<<<<<<<<
 *     else:
 *         l = low
 */
    __pyx_v_l = __Pyx_div_long(__pyx_v_h, 4);

    /* "includes/flowcontrol.pxd":14
 *     else:
 *         h = high
 *     if low is None:             # <<<<<<<<<<<<<<
 *         l = h // 4
 *     else:
 */
    goto __pyx_L5;
  }

  /* "includes/flowcontrol.pxd":17
 *         l = h // 4
 *     else:
 *         l = low             # <<<<<<<<<<<<<<
 * 
 *     if not h >= l >= 0:
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_low); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(7, 17, __pyx_L1_error)
    __pyx_v_l = __pyx_t_2;
  }
  __pyx_L5:;

  /* "includes/flowcontrol.pxd":19
 *         l = low
 * 
 *     if not h >= l >= 0:             # <<<<<<<<<<<<<<
 *         raise ValueError('high (%r) must be >= low (%r) must be >= 0' %
 *                          (h, l))
 */
  __pyx_t_1 = (__pyx_v_h >= __pyx_v_l);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_l >= 0);
  }
  __pyx_t_3 = (!__pyx_t_1);
  if (unlikely(__pyx_t_3)) {

    /* "includes/flowcontrol.pxd":20
 * 
 *     if not h >= l >= 0:
 *         raise ValueError('high (%r) must be >= low (%r) must be >= 0' %             # <<<<<<<<<<<<<<
 *                          (h, l))
 * 
 */
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_high);
    __pyx_t_5 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_high);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_high);

    /* "includes/flowcontrol.pxd":21
 *     if not h >= l >= 0:
 *         raise ValueError('high (%r) must be >= low (%r) must be >= 0' %
 *                          (h, l))             # <<<<<<<<<<<<<<
 * 
 *     return h, l
 */
    __pyx_t_7 = __Pyx_PyUnicode_From_int(__pyx_v_h, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_must_be_low);
    __pyx_t_5 += 18;
    __Pyx_GIVEREF(__pyx_kp_u_must_be_low);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_must_be_low);
    __pyx_t_7 = __Pyx_PyUnicode_From_int(__pyx_v_l, 0, ' ', 'd'); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_must_be_0);
    __pyx_t_5 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_must_be_0);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_must_be_0);

    /* "includes/flowcontrol.pxd":20
 * 
 *     if not h >= l >= 0:
 *         raise ValueError('high (%r) must be >= low (%r) must be >= 0' %             # <<<<<<<<<<<<<<
 *                          (h, l))
 * 
 */
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(7, 20, __pyx_L1_error)

    /* "includes/flowcontrol.pxd":19
 *         l = low
 * 
 *     if not h >= l >= 0:             # <<<<<<<<<<<<<<
 *         raise ValueError('high (%r) must be >= low (%r) must be >= 0' %
 *                          (h, l))
 */
  }

  /* "includes/flowcontrol.pxd":23
 *                          (h, l))
 * 
 *     return h, l             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_h); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_l); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(7, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4)) __PYX_ERR(7, 23, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_7);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7)) __PYX_ERR(7, 23, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "includes/flowcontrol.pxd":4
 * 
 * 
 * cdef inline add_flowcontrol_defaults(high, low, int kb):             # <<<<<<<<<<<<<<
 *     cdef int h, l
 *     if high is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.includes.flowcontrol.add_flowcontrol_defaults", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(8, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(8, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/errors.pyx":1
 * cdef str __strerr(int errno):             # <<<<<<<<<<<<<<
 *     return strerror(errno).decode()
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop___strerr(int __pyx_v_errno) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__strerr", 1);

  /* "uvloop/errors.pyx":2
 * cdef str __strerr(int errno):
 *     return strerror(errno).decode()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = strerror(__pyx_v_errno);
  __pyx_t_2 = __Pyx_ssize_strlen(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_t_3 = __Pyx_decode_c_string(__pyx_t_1, 0, __pyx_t_2, NULL, NULL, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3)) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_INCREF(__pyx_t_3);
  __pyx_r = ((PyObject*)__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uvloop/errors.pyx":1
 * cdef str __strerr(int errno):             # <<<<<<<<<<<<<<
 *     return strerror(errno).decode()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.__strerr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/errors.pyx":5
 * 
 * 
 * cdef __convert_python_error(int uverr):             # <<<<<<<<<<<<<<
 *     # XXX Won't work for Windows:
 *     # From libuv docs:
 */

static PyObject *__pyx_f_6uvloop_4loop___convert_python_error(int __pyx_v_uverr) {
  int __pyx_v_oserr;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__convert_python_error", 1);

  /* "uvloop/errors.pyx":11
 *     #      negated errno (or -errno), while on Windows they
 *     #      are defined by libuv to arbitrary negative numbers.
 *     cdef int oserr = -uverr             # <<<<<<<<<<<<<<
 * 
 *     exc = OSError
 */
  __pyx_v_oserr = (-__pyx_v_uverr);

  /* "uvloop/errors.pyx":13
 *     cdef int oserr = -uverr
 * 
 *     exc = OSError             # <<<<<<<<<<<<<<
 * 
 *     if uverr in (uv.UV_EACCES, uv.UV_EPERM):
 */
  __Pyx_INCREF(__pyx_builtin_OSError);
  __pyx_v_exc = __pyx_builtin_OSError;

  /* "uvloop/errors.pyx":15
 *     exc = OSError
 * 
 *     if uverr in (uv.UV_EACCES, uv.UV_EPERM):             # <<<<<<<<<<<<<<
 *         exc = PermissionError
 * 
 */
  __pyx_t_1 = __pyx_v_uverr;
  __pyx_t_3 = (__pyx_t_1 == UV_EACCES);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 == UV_EPERM);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __pyx_t_3 = __pyx_t_2;
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":16
 * 
 *     if uverr in (uv.UV_EACCES, uv.UV_EPERM):
 *         exc = PermissionError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr in (uv.UV_EAGAIN, uv.UV_EALREADY):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PermissionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 16, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":15
 *     exc = OSError
 * 
 *     if uverr in (uv.UV_EACCES, uv.UV_EPERM):             # <<<<<<<<<<<<<<
 *         exc = PermissionError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":18
 *         exc = PermissionError
 * 
 *     elif uverr in (uv.UV_EAGAIN, uv.UV_EALREADY):             # <<<<<<<<<<<<<<
 *         exc = BlockingIOError
 * 
 */
  __pyx_t_1 = __pyx_v_uverr;
  __pyx_t_2 = (__pyx_t_1 == UV_EAGAIN);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_t_1 == UV_EALREADY);
  __pyx_t_3 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  __pyx_t_2 = __pyx_t_3;
  if (__pyx_t_2) {

    /* "uvloop/errors.pyx":19
 * 
 *     elif uverr in (uv.UV_EAGAIN, uv.UV_EALREADY):
 *         exc = BlockingIOError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr in (uv.UV_EPIPE, uv.UV_ESHUTDOWN):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":18
 *         exc = PermissionError
 * 
 *     elif uverr in (uv.UV_EAGAIN, uv.UV_EALREADY):             # <<<<<<<<<<<<<<
 *         exc = BlockingIOError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":21
 *         exc = BlockingIOError
 * 
 *     elif uverr in (uv.UV_EPIPE, uv.UV_ESHUTDOWN):             # <<<<<<<<<<<<<<
 *         exc = BrokenPipeError
 * 
 */
  __pyx_t_1 = __pyx_v_uverr;
  __pyx_t_3 = (__pyx_t_1 == UV_EPIPE);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 == UV_ESHUTDOWN);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L8_bool_binop_done:;
  __pyx_t_3 = __pyx_t_2;
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":22
 * 
 *     elif uverr in (uv.UV_EPIPE, uv.UV_ESHUTDOWN):
 *         exc = BrokenPipeError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_ECONNABORTED:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_BrokenPipeError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 22, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":21
 *         exc = BlockingIOError
 * 
 *     elif uverr in (uv.UV_EPIPE, uv.UV_ESHUTDOWN):             # <<<<<<<<<<<<<<
 *         exc = BrokenPipeError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":24
 *         exc = BrokenPipeError
 * 
 *     elif uverr == uv.UV_ECONNABORTED:             # <<<<<<<<<<<<<<
 *         exc = ConnectionAbortedError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_ECONNABORTED);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":25
 * 
 *     elif uverr == uv.UV_ECONNABORTED:
 *         exc = ConnectionAbortedError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_ECONNREFUSED:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ConnectionAbortedError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":24
 *         exc = BrokenPipeError
 * 
 *     elif uverr == uv.UV_ECONNABORTED:             # <<<<<<<<<<<<<<
 *         exc = ConnectionAbortedError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":27
 *         exc = ConnectionAbortedError
 * 
 *     elif uverr == uv.UV_ECONNREFUSED:             # <<<<<<<<<<<<<<
 *         exc = ConnectionRefusedError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_ECONNREFUSED);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":28
 * 
 *     elif uverr == uv.UV_ECONNREFUSED:
 *         exc = ConnectionRefusedError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_ECONNRESET:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ConnectionRefusedError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":27
 *         exc = ConnectionAbortedError
 * 
 *     elif uverr == uv.UV_ECONNREFUSED:             # <<<<<<<<<<<<<<
 *         exc = ConnectionRefusedError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":30
 *         exc = ConnectionRefusedError
 * 
 *     elif uverr == uv.UV_ECONNRESET:             # <<<<<<<<<<<<<<
 *         exc = ConnectionResetError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_ECONNRESET);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":31
 * 
 *     elif uverr == uv.UV_ECONNRESET:
 *         exc = ConnectionResetError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EEXIST:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ConnectionResetError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 31, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":30
 *         exc = ConnectionRefusedError
 * 
 *     elif uverr == uv.UV_ECONNRESET:             # <<<<<<<<<<<<<<
 *         exc = ConnectionResetError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":33
 *         exc = ConnectionResetError
 * 
 *     elif uverr == uv.UV_EEXIST:             # <<<<<<<<<<<<<<
 *         exc = FileExistsError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_EEXIST);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":34
 * 
 *     elif uverr == uv.UV_EEXIST:
 *         exc = FileExistsError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_ENOENT:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_FileExistsError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":33
 *         exc = ConnectionResetError
 * 
 *     elif uverr == uv.UV_EEXIST:             # <<<<<<<<<<<<<<
 *         exc = FileExistsError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":36
 *         exc = FileExistsError
 * 
 *     elif uverr == uv.UV_ENOENT:             # <<<<<<<<<<<<<<
 *         exc = FileNotFoundError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_ENOENT);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":37
 * 
 *     elif uverr == uv.UV_ENOENT:
 *         exc = FileNotFoundError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EINTR:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_FileNotFoundError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":36
 *         exc = FileExistsError
 * 
 *     elif uverr == uv.UV_ENOENT:             # <<<<<<<<<<<<<<
 *         exc = FileNotFoundError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":39
 *         exc = FileNotFoundError
 * 
 *     elif uverr == uv.UV_EINTR:             # <<<<<<<<<<<<<<
 *         exc = InterruptedError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_EINTR);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":40
 * 
 *     elif uverr == uv.UV_EINTR:
 *         exc = InterruptedError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EISDIR:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":39
 *         exc = FileNotFoundError
 * 
 *     elif uverr == uv.UV_EINTR:             # <<<<<<<<<<<<<<
 *         exc = InterruptedError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":42
 *         exc = InterruptedError
 * 
 *     elif uverr == uv.UV_EISDIR:             # <<<<<<<<<<<<<<
 *         exc = IsADirectoryError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_EISDIR);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":43
 * 
 *     elif uverr == uv.UV_EISDIR:
 *         exc = IsADirectoryError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_ESRCH:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_IsADirectoryError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":42
 *         exc = InterruptedError
 * 
 *     elif uverr == uv.UV_EISDIR:             # <<<<<<<<<<<<<<
 *         exc = IsADirectoryError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":45
 *         exc = IsADirectoryError
 * 
 *     elif uverr == uv.UV_ESRCH:             # <<<<<<<<<<<<<<
 *         exc = ProcessLookupError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_ESRCH);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":46
 * 
 *     elif uverr == uv.UV_ESRCH:
 *         exc = ProcessLookupError             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_ETIMEDOUT:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ProcessLookupError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":45
 *         exc = IsADirectoryError
 * 
 *     elif uverr == uv.UV_ESRCH:             # <<<<<<<<<<<<<<
 *         exc = ProcessLookupError
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":48
 *         exc = ProcessLookupError
 * 
 *     elif uverr == uv.UV_ETIMEDOUT:             # <<<<<<<<<<<<<<
 *         exc = TimeoutError
 * 
 */
  __pyx_t_3 = (__pyx_v_uverr == UV_ETIMEDOUT);
  if (__pyx_t_3) {

    /* "uvloop/errors.pyx":49
 * 
 *     elif uverr == uv.UV_ETIMEDOUT:
 *         exc = TimeoutError             # <<<<<<<<<<<<<<
 * 
 *     return exc(oserr, __strerr(oserr))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TimeoutError); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/errors.pyx":48
 *         exc = ProcessLookupError
 * 
 *     elif uverr == uv.UV_ETIMEDOUT:             # <<<<<<<<<<<<<<
 *         exc = TimeoutError
 * 
 */
  }
  __pyx_L3:;

  /* "uvloop/errors.pyx":51
 *         exc = TimeoutError
 * 
 *     return exc(oserr, __strerr(oserr))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_oserr); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __pyx_f_6uvloop_4loop___strerr(__pyx_v_oserr); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_exc);
  __pyx_t_7 = __pyx_v_exc; __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_t_5, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "uvloop/errors.pyx":5
 * 
 * 
 * cdef __convert_python_error(int uverr):             # <<<<<<<<<<<<<<
 *     # XXX Won't work for Windows:
 *     # From libuv docs:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.__convert_python_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/errors.pyx":54
 * 
 * 
 * cdef int __convert_socket_error(int uverr):             # <<<<<<<<<<<<<<
 *     cdef int sock_err = 0
 * 
 */

static int __pyx_f_6uvloop_4loop___convert_socket_error(int __pyx_v_uverr) {
  int __pyx_v_sock_err;
  int __pyx_r;
  int __pyx_t_1;

  /* "uvloop/errors.pyx":55
 * 
 * cdef int __convert_socket_error(int uverr):
 *     cdef int sock_err = 0             # <<<<<<<<<<<<<<
 * 
 *     if uverr == uv.UV_EAI_ADDRFAMILY:
 */
  __pyx_v_sock_err = 0;

  /* "uvloop/errors.pyx":57
 *     cdef int sock_err = 0
 * 
 *     if uverr == uv.UV_EAI_ADDRFAMILY:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_ADDRFAMILY
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_ADDRFAMILY);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":58
 * 
 *     if uverr == uv.UV_EAI_ADDRFAMILY:
 *         sock_err = socket_EAI_ADDRFAMILY             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_AGAIN:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_ADDRFAMILY;

    /* "uvloop/errors.pyx":57
 *     cdef int sock_err = 0
 * 
 *     if uverr == uv.UV_EAI_ADDRFAMILY:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_ADDRFAMILY
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":60
 *         sock_err = socket_EAI_ADDRFAMILY
 * 
 *     elif uverr == uv.UV_EAI_AGAIN:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_AGAIN
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_AGAIN);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":61
 * 
 *     elif uverr == uv.UV_EAI_AGAIN:
 *         sock_err = socket_EAI_AGAIN             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_BADFLAGS:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_AGAIN;

    /* "uvloop/errors.pyx":60
 *         sock_err = socket_EAI_ADDRFAMILY
 * 
 *     elif uverr == uv.UV_EAI_AGAIN:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_AGAIN
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":63
 *         sock_err = socket_EAI_AGAIN
 * 
 *     elif uverr == uv.UV_EAI_BADFLAGS:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_BADFLAGS
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_BADFLAGS);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":64
 * 
 *     elif uverr == uv.UV_EAI_BADFLAGS:
 *         sock_err = socket_EAI_BADFLAGS             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_BADHINTS:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_BADFLAGS;

    /* "uvloop/errors.pyx":63
 *         sock_err = socket_EAI_AGAIN
 * 
 *     elif uverr == uv.UV_EAI_BADFLAGS:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_BADFLAGS
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":66
 *         sock_err = socket_EAI_BADFLAGS
 * 
 *     elif uverr == uv.UV_EAI_BADHINTS:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_BADHINTS
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_BADHINTS);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":67
 * 
 *     elif uverr == uv.UV_EAI_BADHINTS:
 *         sock_err = socket_EAI_BADHINTS             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_CANCELED:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_BADHINTS;

    /* "uvloop/errors.pyx":66
 *         sock_err = socket_EAI_BADFLAGS
 * 
 *     elif uverr == uv.UV_EAI_BADHINTS:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_BADHINTS
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":69
 *         sock_err = socket_EAI_BADHINTS
 * 
 *     elif uverr == uv.UV_EAI_CANCELED:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_CANCELED
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_CANCELED);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":70
 * 
 *     elif uverr == uv.UV_EAI_CANCELED:
 *         sock_err = socket_EAI_CANCELED             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_FAIL:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_CANCELED;

    /* "uvloop/errors.pyx":69
 *         sock_err = socket_EAI_BADHINTS
 * 
 *     elif uverr == uv.UV_EAI_CANCELED:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_CANCELED
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":72
 *         sock_err = socket_EAI_CANCELED
 * 
 *     elif uverr == uv.UV_EAI_FAIL:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_FAIL
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_FAIL);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":73
 * 
 *     elif uverr == uv.UV_EAI_FAIL:
 *         sock_err = socket_EAI_FAIL             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_FAMILY:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_FAIL;

    /* "uvloop/errors.pyx":72
 *         sock_err = socket_EAI_CANCELED
 * 
 *     elif uverr == uv.UV_EAI_FAIL:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_FAIL
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":75
 *         sock_err = socket_EAI_FAIL
 * 
 *     elif uverr == uv.UV_EAI_FAMILY:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_FAMILY
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_FAMILY);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":76
 * 
 *     elif uverr == uv.UV_EAI_FAMILY:
 *         sock_err = socket_EAI_FAMILY             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_MEMORY:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_FAMILY;

    /* "uvloop/errors.pyx":75
 *         sock_err = socket_EAI_FAIL
 * 
 *     elif uverr == uv.UV_EAI_FAMILY:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_FAMILY
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":78
 *         sock_err = socket_EAI_FAMILY
 * 
 *     elif uverr == uv.UV_EAI_MEMORY:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_MEMORY
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_MEMORY);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":79
 * 
 *     elif uverr == uv.UV_EAI_MEMORY:
 *         sock_err = socket_EAI_MEMORY             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_NODATA:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_MEMORY;

    /* "uvloop/errors.pyx":78
 *         sock_err = socket_EAI_FAMILY
 * 
 *     elif uverr == uv.UV_EAI_MEMORY:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_MEMORY
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":81
 *         sock_err = socket_EAI_MEMORY
 * 
 *     elif uverr == uv.UV_EAI_NODATA:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_NODATA
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_NODATA);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":82
 * 
 *     elif uverr == uv.UV_EAI_NODATA:
 *         sock_err = socket_EAI_NODATA             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_NONAME:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_NODATA;

    /* "uvloop/errors.pyx":81
 *         sock_err = socket_EAI_MEMORY
 * 
 *     elif uverr == uv.UV_EAI_NODATA:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_NODATA
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":84
 *         sock_err = socket_EAI_NODATA
 * 
 *     elif uverr == uv.UV_EAI_NONAME:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_NONAME
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_NONAME);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":85
 * 
 *     elif uverr == uv.UV_EAI_NONAME:
 *         sock_err = socket_EAI_NONAME             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_OVERFLOW:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_NONAME;

    /* "uvloop/errors.pyx":84
 *         sock_err = socket_EAI_NODATA
 * 
 *     elif uverr == uv.UV_EAI_NONAME:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_NONAME
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":87
 *         sock_err = socket_EAI_NONAME
 * 
 *     elif uverr == uv.UV_EAI_OVERFLOW:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_OVERFLOW
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_OVERFLOW);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":88
 * 
 *     elif uverr == uv.UV_EAI_OVERFLOW:
 *         sock_err = socket_EAI_OVERFLOW             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_PROTOCOL:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_OVERFLOW;

    /* "uvloop/errors.pyx":87
 *         sock_err = socket_EAI_NONAME
 * 
 *     elif uverr == uv.UV_EAI_OVERFLOW:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_OVERFLOW
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":90
 *         sock_err = socket_EAI_OVERFLOW
 * 
 *     elif uverr == uv.UV_EAI_PROTOCOL:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_PROTOCOL
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_PROTOCOL);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":91
 * 
 *     elif uverr == uv.UV_EAI_PROTOCOL:
 *         sock_err = socket_EAI_PROTOCOL             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_SERVICE:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_PROTOCOL;

    /* "uvloop/errors.pyx":90
 *         sock_err = socket_EAI_OVERFLOW
 * 
 *     elif uverr == uv.UV_EAI_PROTOCOL:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_PROTOCOL
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":93
 *         sock_err = socket_EAI_PROTOCOL
 * 
 *     elif uverr == uv.UV_EAI_SERVICE:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_SERVICE
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_SERVICE);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":94
 * 
 *     elif uverr == uv.UV_EAI_SERVICE:
 *         sock_err = socket_EAI_SERVICE             # <<<<<<<<<<<<<<
 * 
 *     elif uverr == uv.UV_EAI_SOCKTYPE:
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_SERVICE;

    /* "uvloop/errors.pyx":93
 *         sock_err = socket_EAI_PROTOCOL
 * 
 *     elif uverr == uv.UV_EAI_SERVICE:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_SERVICE
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/errors.pyx":96
 *         sock_err = socket_EAI_SERVICE
 * 
 *     elif uverr == uv.UV_EAI_SOCKTYPE:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_SOCKTYPE
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_EAI_SOCKTYPE);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":97
 * 
 *     elif uverr == uv.UV_EAI_SOCKTYPE:
 *         sock_err = socket_EAI_SOCKTYPE             # <<<<<<<<<<<<<<
 * 
 *     return sock_err
 */
    __pyx_v_sock_err = __pyx_v_6uvloop_4loop_socket_EAI_SOCKTYPE;

    /* "uvloop/errors.pyx":96
 *         sock_err = socket_EAI_SERVICE
 * 
 *     elif uverr == uv.UV_EAI_SOCKTYPE:             # <<<<<<<<<<<<<<
 *         sock_err = socket_EAI_SOCKTYPE
 * 
 */
  }
  __pyx_L3:;

  /* "uvloop/errors.pyx":99
 *         sock_err = socket_EAI_SOCKTYPE
 * 
 *     return sock_err             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_sock_err;
  goto __pyx_L0;

  /* "uvloop/errors.pyx":54
 * 
 * 
 * cdef int __convert_socket_error(int uverr):             # <<<<<<<<<<<<<<
 *     cdef int sock_err = 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/errors.pyx":102
 * 
 * 
 * cdef convert_error(int uverr):             # <<<<<<<<<<<<<<
 *     cdef int sock_err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_convert_error(int __pyx_v_uverr) {
  int __pyx_v_sock_err;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_error", 1);

  /* "uvloop/errors.pyx":105
 *     cdef int sock_err
 * 
 *     if uverr == uv.UV_ECANCELED:             # <<<<<<<<<<<<<<
 *         return aio_CancelledError()
 * 
 */
  __pyx_t_1 = (__pyx_v_uverr == UV_ECANCELED);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":106
 * 
 *     if uverr == uv.UV_ECANCELED:
 *         return aio_CancelledError()             # <<<<<<<<<<<<<<
 * 
 *     sock_err = __convert_socket_error(uverr)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_CancelledError);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_aio_CancelledError; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/errors.pyx":105
 *     cdef int sock_err
 * 
 *     if uverr == uv.UV_ECANCELED:             # <<<<<<<<<<<<<<
 *         return aio_CancelledError()
 * 
 */
  }

  /* "uvloop/errors.pyx":108
 *         return aio_CancelledError()
 * 
 *     sock_err = __convert_socket_error(uverr)             # <<<<<<<<<<<<<<
 *     if sock_err:
 *         msg = system.gai_strerror(sock_err).decode('utf-8')
 */
  __pyx_t_6 = __pyx_f_6uvloop_4loop___convert_socket_error(__pyx_v_uverr); if (unlikely(__pyx_t_6 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(1, 108, __pyx_L1_error)
  __pyx_v_sock_err = __pyx_t_6;

  /* "uvloop/errors.pyx":109
 * 
 *     sock_err = __convert_socket_error(uverr)
 *     if sock_err:             # <<<<<<<<<<<<<<
 *         msg = system.gai_strerror(sock_err).decode('utf-8')
 *         return socket_gaierror(sock_err, msg)
 */
  __pyx_t_1 = (__pyx_v_sock_err != 0);
  if (__pyx_t_1) {

    /* "uvloop/errors.pyx":110
 *     sock_err = __convert_socket_error(uverr)
 *     if sock_err:
 *         msg = system.gai_strerror(sock_err).decode('utf-8')             # <<<<<<<<<<<<<<
 *         return socket_gaierror(sock_err, msg)
 * 
 */
    __pyx_t_7 = gai_strerror(__pyx_v_sock_err);
    __pyx_t_8 = __Pyx_ssize_strlen(__pyx_t_7); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(1, 110, __pyx_L1_error)
    __pyx_t_2 = __Pyx_decode_c_string(__pyx_t_7, 0, __pyx_t_8, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_msg = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/errors.pyx":111
 *     if sock_err:
 *         msg = system.gai_strerror(sock_err).decode('utf-8')
 *         return socket_gaierror(sock_err, msg)             # <<<<<<<<<<<<<<
 * 
 *     return __convert_python_error(uverr)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_sock_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_gaierror);
    __pyx_t_4 = __pyx_v_6uvloop_4loop_socket_gaierror; __pyx_t_9 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_t_3, __pyx_v_msg};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/errors.pyx":109
 * 
 *     sock_err = __convert_socket_error(uverr)
 *     if sock_err:             # <<<<<<<<<<<<<<
 *         msg = system.gai_strerror(sock_err).decode('utf-8')
 *         return socket_gaierror(sock_err, msg)
 */
  }

  /* "uvloop/errors.pyx":113
 *         return socket_gaierror(sock_err, msg)
 * 
 *     return __convert_python_error(uverr)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_f_6uvloop_4loop___convert_python_error(__pyx_v_uverr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/errors.pyx":102
 * 
 * 
 * cdef convert_error(int uverr):             # <<<<<<<<<<<<<<
 *     cdef int sock_err
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.convert_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":57
 * 
 * 
 * cdef _is_sock_stream(sock_type):             # <<<<<<<<<<<<<<
 *     if SOCK_NONBLOCK == -1:
 *         return sock_type == uv.SOCK_STREAM
 */

static PyObject *__pyx_f_6uvloop_4loop__is_sock_stream(PyObject *__pyx_v_sock_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_sock_stream", 1);

  /* "uvloop/loop.pyx":58
 * 
 * cdef _is_sock_stream(sock_type):
 *     if SOCK_NONBLOCK == -1:             # <<<<<<<<<<<<<<
 *         return sock_type == uv.SOCK_STREAM
 *     else:
 */
  __pyx_t_1 = (__pyx_v_6uvloop_4loop_SOCK_NONBLOCK == -1L);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":59
 * cdef _is_sock_stream(sock_type):
 *     if SOCK_NONBLOCK == -1:
 *         return sock_type == uv.SOCK_STREAM             # <<<<<<<<<<<<<<
 *     else:
 *         # Linux's socket.type is a bitmask that can include extra info
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(SOCK_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_sock_type, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 59, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":58
 * 
 * cdef _is_sock_stream(sock_type):
 *     if SOCK_NONBLOCK == -1:             # <<<<<<<<<<<<<<
 *         return sock_type == uv.SOCK_STREAM
 *     else:
 */
  }

  /* "uvloop/loop.pyx":66
 *         # https://github.com/torvalds/linux/blob/v4.13/include/linux/net.h#L77
 *         # for more details.
 *         return (sock_type & 0xF) == uv.SOCK_STREAM             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_v_sock_type, __pyx_int_15, 0xF, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_From_int(SOCK_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "uvloop/loop.pyx":57
 * 
 * 
 * cdef _is_sock_stream(sock_type):             # <<<<<<<<<<<<<<
 *     if SOCK_NONBLOCK == -1:
 *         return sock_type == uv.SOCK_STREAM
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop._is_sock_stream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":69
 * 
 * 
 * cdef _is_sock_dgram(sock_type):             # <<<<<<<<<<<<<<
 *     if SOCK_NONBLOCK == -1:
 *         return sock_type == uv.SOCK_DGRAM
 */

static PyObject *__pyx_f_6uvloop_4loop__is_sock_dgram(PyObject *__pyx_v_sock_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_sock_dgram", 1);

  /* "uvloop/loop.pyx":70
 * 
 * cdef _is_sock_dgram(sock_type):
 *     if SOCK_NONBLOCK == -1:             # <<<<<<<<<<<<<<
 *         return sock_type == uv.SOCK_DGRAM
 *     else:
 */
  __pyx_t_1 = (__pyx_v_6uvloop_4loop_SOCK_NONBLOCK == -1L);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":71
 * cdef _is_sock_dgram(sock_type):
 *     if SOCK_NONBLOCK == -1:
 *         return sock_type == uv.SOCK_DGRAM             # <<<<<<<<<<<<<<
 *     else:
 *         # Read the comment in `_is_sock_stream`.
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(SOCK_DGRAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_sock_type, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":70
 * 
 * cdef _is_sock_dgram(sock_type):
 *     if SOCK_NONBLOCK == -1:             # <<<<<<<<<<<<<<
 *         return sock_type == uv.SOCK_DGRAM
 *     else:
 */
  }

  /* "uvloop/loop.pyx":74
 *     else:
 *         # Read the comment in `_is_sock_stream`.
 *         return (sock_type & 0xF) == uv.SOCK_DGRAM             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_v_sock_type, __pyx_int_15, 0xF, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_From_int(SOCK_DGRAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "uvloop/loop.pyx":69
 * 
 * 
 * cdef _is_sock_dgram(sock_type):             # <<<<<<<<<<<<<<
 *     if SOCK_NONBLOCK == -1:
 *         return sock_type == uv.SOCK_DGRAM
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop._is_sock_dgram", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":77
 * 
 * 
 * cdef isfuture(obj):             # <<<<<<<<<<<<<<
 *     if aio_isfuture is None:
 *         return isinstance(obj, aio_Future)
 */

static PyObject *__pyx_f_6uvloop_4loop_isfuture(PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("isfuture", 1);

  /* "uvloop/loop.pyx":78
 * 
 * cdef isfuture(obj):
 *     if aio_isfuture is None:             # <<<<<<<<<<<<<<
 *         return isinstance(obj, aio_Future)
 *     else:
 */
  __pyx_t_1 = (__pyx_v_6uvloop_4loop_aio_isfuture == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":79
 * cdef isfuture(obj):
 *     if aio_isfuture is None:
 *         return isinstance(obj, aio_Future)             # <<<<<<<<<<<<<<
 *     else:
 *         return aio_isfuture(obj)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_6uvloop_4loop_aio_Future;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = PyObject_IsInstance(__pyx_v_obj, __pyx_t_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":78
 * 
 * cdef isfuture(obj):
 *     if aio_isfuture is None:             # <<<<<<<<<<<<<<
 *         return isinstance(obj, aio_Future)
 *     else:
 */
  }

  /* "uvloop/loop.pyx":81
 *         return isinstance(obj, aio_Future)
 *     else:
 *         return aio_isfuture(obj)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_isfuture);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_aio_isfuture; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "uvloop/loop.pyx":77
 * 
 * 
 * cdef isfuture(obj):             # <<<<<<<<<<<<<<
 *     if aio_isfuture is None:
 *         return isinstance(obj, aio_Future)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.isfuture", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":84
 * 
 * 
 * cdef inline socket_inc_io_ref(sock):             # <<<<<<<<<<<<<<
 *     if isinstance(sock, socket_socket):
 *         sock._io_refs += 1
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_socket_inc_io_ref(PyObject *__pyx_v_sock) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("socket_inc_io_ref", 1);

  /* "uvloop/loop.pyx":85
 * 
 * cdef inline socket_inc_io_ref(sock):
 *     if isinstance(sock, socket_socket):             # <<<<<<<<<<<<<<
 *         sock._io_refs += 1
 * 
 */
  __pyx_t_1 = __pyx_v_6uvloop_4loop_socket_socket;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_sock, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":86
 * cdef inline socket_inc_io_ref(sock):
 *     if isinstance(sock, socket_socket):
 *         sock._io_refs += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_io_refs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_sock, __pyx_n_s_io_refs, __pyx_t_3) < 0) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":85
 * 
 * cdef inline socket_inc_io_ref(sock):
 *     if isinstance(sock, socket_socket):             # <<<<<<<<<<<<<<
 *         sock._io_refs += 1
 * 
 */
  }

  /* "uvloop/loop.pyx":84
 * 
 * 
 * cdef inline socket_inc_io_ref(sock):             # <<<<<<<<<<<<<<
 *     if isinstance(sock, socket_socket):
 *         sock._io_refs += 1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.socket_inc_io_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":89
 * 
 * 
 * cdef inline socket_dec_io_ref(sock):             # <<<<<<<<<<<<<<
 *     if isinstance(sock, socket_socket):
 *         sock._decref_socketios()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_socket_dec_io_ref(PyObject *__pyx_v_sock) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("socket_dec_io_ref", 1);

  /* "uvloop/loop.pyx":90
 * 
 * cdef inline socket_dec_io_ref(sock):
 *     if isinstance(sock, socket_socket):             # <<<<<<<<<<<<<<
 *         sock._decref_socketios()
 * 
 */
  __pyx_t_1 = __pyx_v_6uvloop_4loop_socket_socket;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_sock, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":91
 * cdef inline socket_dec_io_ref(sock):
 *     if isinstance(sock, socket_socket):
 *         sock._decref_socketios()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_decref_socketios); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":90
 * 
 * cdef inline socket_dec_io_ref(sock):
 *     if isinstance(sock, socket_socket):             # <<<<<<<<<<<<<<
 *         sock._decref_socketios()
 * 
 */
  }

  /* "uvloop/loop.pyx":89
 * 
 * 
 * cdef inline socket_dec_io_ref(sock):             # <<<<<<<<<<<<<<
 *     if isinstance(sock, socket_socket):
 *         sock._decref_socketios()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.socket_dec_io_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":94
 * 
 * 
 * cdef inline run_in_context(context, method):             # <<<<<<<<<<<<<<
 *     # This method is internally used to workaround a reference issue that in
 *     # certain circumstances, inlined context.run() will not hold a reference to
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_run_in_context(PyObject *__pyx_v_context, PyObject *__pyx_v_method) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_in_context", 1);

  /* "uvloop/loop.pyx":99
 *     # the given method instance, which - if deallocated - will cause segfault.
 *     # See also: edgedb/edgedb#2222
 *     Py_INCREF(method)             # <<<<<<<<<<<<<<
 *     try:
 *         return context.run(method)
 */
  Py_INCREF(__pyx_v_method);

  /* "uvloop/loop.pyx":100
 *     # See also: edgedb/edgedb#2222
 *     Py_INCREF(method)
 *     try:             # <<<<<<<<<<<<<<
 *         return context.run(method)
 *     finally:
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":101
 *     Py_INCREF(method)
 *     try:
 *         return context.run(method)             # <<<<<<<<<<<<<<
 *     finally:
 *         Py_DECREF(method)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_context, __pyx_n_s_run); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_method};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "uvloop/loop.pyx":103
 *         return context.run(method)
 *     finally:
 *         Py_DECREF(method)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        Py_DECREF(__pyx_v_method);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_13 = __pyx_r;
      __pyx_r = 0;
      Py_DECREF(__pyx_v_method);
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/loop.pyx":94
 * 
 * 
 * cdef inline run_in_context(context, method):             # <<<<<<<<<<<<<<
 *     # This method is internally used to workaround a reference issue that in
 *     # certain circumstances, inlined context.run() will not hold a reference to
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.run_in_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":106
 * 
 * 
 * cdef inline run_in_context1(context, method, arg):             # <<<<<<<<<<<<<<
 *     Py_INCREF(method)
 *     try:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_run_in_context1(PyObject *__pyx_v_context, PyObject *__pyx_v_method, PyObject *__pyx_v_arg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_in_context1", 1);

  /* "uvloop/loop.pyx":107
 * 
 * cdef inline run_in_context1(context, method, arg):
 *     Py_INCREF(method)             # <<<<<<<<<<<<<<
 *     try:
 *         return context.run(method, arg)
 */
  Py_INCREF(__pyx_v_method);

  /* "uvloop/loop.pyx":108
 * cdef inline run_in_context1(context, method, arg):
 *     Py_INCREF(method)
 *     try:             # <<<<<<<<<<<<<<
 *         return context.run(method, arg)
 *     finally:
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":109
 *     Py_INCREF(method)
 *     try:
 *         return context.run(method, arg)             # <<<<<<<<<<<<<<
 *     finally:
 *         Py_DECREF(method)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_context, __pyx_n_s_run); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 109, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_method, __pyx_v_arg};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "uvloop/loop.pyx":111
 *         return context.run(method, arg)
 *     finally:
 *         Py_DECREF(method)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        Py_DECREF(__pyx_v_method);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_13 = __pyx_r;
      __pyx_r = 0;
      Py_DECREF(__pyx_v_method);
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/loop.pyx":106
 * 
 * 
 * cdef inline run_in_context1(context, method, arg):             # <<<<<<<<<<<<<<
 *     Py_INCREF(method)
 *     try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.run_in_context1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":114
 * 
 * 
 * cdef inline run_in_context2(context, method, arg1, arg2):             # <<<<<<<<<<<<<<
 *     Py_INCREF(method)
 *     try:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_run_in_context2(PyObject *__pyx_v_context, PyObject *__pyx_v_method, PyObject *__pyx_v_arg1, PyObject *__pyx_v_arg2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_in_context2", 1);

  /* "uvloop/loop.pyx":115
 * 
 * cdef inline run_in_context2(context, method, arg1, arg2):
 *     Py_INCREF(method)             # <<<<<<<<<<<<<<
 *     try:
 *         return context.run(method, arg1, arg2)
 */
  Py_INCREF(__pyx_v_method);

  /* "uvloop/loop.pyx":116
 * cdef inline run_in_context2(context, method, arg1, arg2):
 *     Py_INCREF(method)
 *     try:             # <<<<<<<<<<<<<<
 *         return context.run(method, arg1, arg2)
 *     finally:
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":117
 *     Py_INCREF(method)
 *     try:
 *         return context.run(method, arg1, arg2)             # <<<<<<<<<<<<<<
 *     finally:
 *         Py_DECREF(method)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_context, __pyx_n_s_run); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_v_method, __pyx_v_arg1, __pyx_v_arg2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "uvloop/loop.pyx":119
 *         return context.run(method, arg1, arg2)
 *     finally:
 *         Py_DECREF(method)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        Py_DECREF(__pyx_v_method);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_13 = __pyx_r;
      __pyx_r = 0;
      Py_DECREF(__pyx_v_method);
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/loop.pyx":114
 * 
 * 
 * cdef inline run_in_context2(context, method, arg1, arg2):             # <<<<<<<<<<<<<<
 *     Py_INCREF(method)
 *     try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.run_in_context2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":129
 * @cython.no_gc_clear
 * cdef class Loop:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_4Loop_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_4Loop_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop___cinit__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_4Loop___cinit__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  int __pyx_v_err;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/loop.pyx":133
 * 
 *         # Install PyMem* memory allocators if they aren't installed yet.
 *         __install_pymem()             # <<<<<<<<<<<<<<
 * 
 *         # Install pthread_atfork handlers
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___install_pymem(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":136
 * 
 *         # Install pthread_atfork handlers
 *         __install_atfork()             # <<<<<<<<<<<<<<
 * 
 *         self.uvloop = <uv.uv_loop_t*>PyMem_RawMalloc(sizeof(uv.uv_loop_t))
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___install_atfork(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":138
 *         __install_atfork()
 * 
 *         self.uvloop = <uv.uv_loop_t*>PyMem_RawMalloc(sizeof(uv.uv_loop_t))             # <<<<<<<<<<<<<<
 *         if self.uvloop is NULL:
 *             raise MemoryError()
 */
  __pyx_v_self->uvloop = ((uv_loop_t *)PyMem_RawMalloc((sizeof(uv_loop_t))));

  /* "uvloop/loop.pyx":139
 * 
 *         self.uvloop = <uv.uv_loop_t*>PyMem_RawMalloc(sizeof(uv.uv_loop_t))
 *         if self.uvloop is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (__pyx_v_self->uvloop == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/loop.pyx":140
 *         self.uvloop = <uv.uv_loop_t*>PyMem_RawMalloc(sizeof(uv.uv_loop_t))
 *         if self.uvloop is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self.slow_callback_duration = 0.1
 */
    PyErr_NoMemory(); __PYX_ERR(0, 140, __pyx_L1_error)

    /* "uvloop/loop.pyx":139
 * 
 *         self.uvloop = <uv.uv_loop_t*>PyMem_RawMalloc(sizeof(uv.uv_loop_t))
 *         if self.uvloop is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uvloop/loop.pyx":142
 *             raise MemoryError()
 * 
 *         self.slow_callback_duration = 0.1             # <<<<<<<<<<<<<<
 * 
 *         self._closed = 0
 */
  __Pyx_INCREF(__pyx_float_0_1);
  __Pyx_GIVEREF(__pyx_float_0_1);
  __Pyx_GOTREF(__pyx_v_self->slow_callback_duration);
  __Pyx_DECREF(__pyx_v_self->slow_callback_duration);
  __pyx_v_self->slow_callback_duration = __pyx_float_0_1;

  /* "uvloop/loop.pyx":144
 *         self.slow_callback_duration = 0.1
 * 
 *         self._closed = 0             # <<<<<<<<<<<<<<
 *         self._debug = 0
 *         self._thread_id = 0
 */
  __pyx_v_self->_closed = 0;

  /* "uvloop/loop.pyx":145
 * 
 *         self._closed = 0
 *         self._debug = 0             # <<<<<<<<<<<<<<
 *         self._thread_id = 0
 *         self._running = 0
 */
  __pyx_v_self->_debug = 0;

  /* "uvloop/loop.pyx":146
 *         self._closed = 0
 *         self._debug = 0
 *         self._thread_id = 0             # <<<<<<<<<<<<<<
 *         self._running = 0
 *         self._stopping = 0
 */
  __pyx_v_self->_thread_id = 0;

  /* "uvloop/loop.pyx":147
 *         self._debug = 0
 *         self._thread_id = 0
 *         self._running = 0             # <<<<<<<<<<<<<<
 *         self._stopping = 0
 * 
 */
  __pyx_v_self->_running = 0;

  /* "uvloop/loop.pyx":148
 *         self._thread_id = 0
 *         self._running = 0
 *         self._stopping = 0             # <<<<<<<<<<<<<<
 * 
 *         self._transports = weakref_WeakValueDictionary()
 */
  __pyx_v_self->_stopping = 0;

  /* "uvloop/loop.pyx":150
 *         self._stopping = 0
 * 
 *         self._transports = weakref_WeakValueDictionary()             # <<<<<<<<<<<<<<
 *         self._processes = set()
 * 
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_weakref_WeakValueDictionary);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_weakref_WeakValueDictionary; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_transports);
  __Pyx_DECREF(__pyx_v_self->_transports);
  __pyx_v_self->_transports = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":151
 * 
 *         self._transports = weakref_WeakValueDictionary()
 *         self._processes = set()             # <<<<<<<<<<<<<<
 * 
 *         # Used to keep a reference (and hence keep the fileobj alive)
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_processes);
  __Pyx_DECREF(__pyx_v_self->_processes);
  __pyx_v_self->_processes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":156
 *         # for as long as its registered by add_reader or add_writer.
 *         # This is how the selector module and hence asyncio behaves.
 *         self._fd_to_reader_fileobj = {}             # <<<<<<<<<<<<<<
 *         self._fd_to_writer_fileobj = {}
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_fd_to_reader_fileobj);
  __Pyx_DECREF(__pyx_v_self->_fd_to_reader_fileobj);
  __pyx_v_self->_fd_to_reader_fileobj = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":157
 *         # This is how the selector module and hence asyncio behaves.
 *         self._fd_to_reader_fileobj = {}
 *         self._fd_to_writer_fileobj = {}             # <<<<<<<<<<<<<<
 * 
 *         self._unix_server_sockets = {}
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_fd_to_writer_fileobj);
  __Pyx_DECREF(__pyx_v_self->_fd_to_writer_fileobj);
  __pyx_v_self->_fd_to_writer_fileobj = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":159
 *         self._fd_to_writer_fileobj = {}
 * 
 *         self._unix_server_sockets = {}             # <<<<<<<<<<<<<<
 * 
 *         self._timers = set()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_unix_server_sockets);
  __Pyx_DECREF(__pyx_v_self->_unix_server_sockets);
  __pyx_v_self->_unix_server_sockets = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":161
 *         self._unix_server_sockets = {}
 * 
 *         self._timers = set()             # <<<<<<<<<<<<<<
 *         self._polls = {}
 * 
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_timers);
  __Pyx_DECREF(__pyx_v_self->_timers);
  __pyx_v_self->_timers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":162
 * 
 *         self._timers = set()
 *         self._polls = {}             # <<<<<<<<<<<<<<
 * 
 *         self._recv_buffer_in_use = 0
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_polls);
  __Pyx_DECREF(__pyx_v_self->_polls);
  __pyx_v_self->_polls = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":164
 *         self._polls = {}
 * 
 *         self._recv_buffer_in_use = 0             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_loop_init(self.uvloop)
 */
  __pyx_v_self->_recv_buffer_in_use = 0;

  /* "uvloop/loop.pyx":166
 *         self._recv_buffer_in_use = 0
 * 
 *         err = uv.uv_loop_init(self.uvloop)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
  __pyx_v_err = uv_loop_init(__pyx_v_self->uvloop);

  /* "uvloop/loop.pyx":167
 * 
 *         err = uv.uv_loop_init(self.uvloop)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 *         self.uvloop.data = <void*> self
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/loop.pyx":168
 *         err = uv.uv_loop_init(self.uvloop)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 *         self.uvloop.data = <void*> self
 * 
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 168, __pyx_L1_error)

    /* "uvloop/loop.pyx":167
 * 
 *         err = uv.uv_loop_init(self.uvloop)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 *         self.uvloop.data = <void*> self
 */
  }

  /* "uvloop/loop.pyx":169
 *         if err < 0:
 *             raise convert_error(err)
 *         self.uvloop.data = <void*> self             # <<<<<<<<<<<<<<
 * 
 *         self._init_debug_fields()
 */
  __pyx_v_self->uvloop->data = ((void *)__pyx_v_self);

  /* "uvloop/loop.pyx":171
 *         self.uvloop.data = <void*> self
 * 
 *         self._init_debug_fields()             # <<<<<<<<<<<<<<
 * 
 *         self.active_process_handler = None
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_init_debug_fields(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":173
 *         self._init_debug_fields()
 * 
 *         self.active_process_handler = None             # <<<<<<<<<<<<<<
 * 
 *         self._last_error = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->active_process_handler);
  __Pyx_DECREF((PyObject *)__pyx_v_self->active_process_handler);
  __pyx_v_self->active_process_handler = ((struct __pyx_obj_6uvloop_4loop_UVProcess *)Py_None);

  /* "uvloop/loop.pyx":175
 *         self.active_process_handler = None
 * 
 *         self._last_error = None             # <<<<<<<<<<<<<<
 * 
 *         self._task_factory = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_error);
  __Pyx_DECREF(__pyx_v_self->_last_error);
  __pyx_v_self->_last_error = Py_None;

  /* "uvloop/loop.pyx":177
 *         self._last_error = None
 * 
 *         self._task_factory = None             # <<<<<<<<<<<<<<
 *         self._exception_handler = None
 *         self._default_executor = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_task_factory);
  __Pyx_DECREF(__pyx_v_self->_task_factory);
  __pyx_v_self->_task_factory = Py_None;

  /* "uvloop/loop.pyx":178
 * 
 *         self._task_factory = None
 *         self._exception_handler = None             # <<<<<<<<<<<<<<
 *         self._default_executor = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_exception_handler);
  __Pyx_DECREF(__pyx_v_self->_exception_handler);
  __pyx_v_self->_exception_handler = Py_None;

  /* "uvloop/loop.pyx":179
 *         self._task_factory = None
 *         self._exception_handler = None
 *         self._default_executor = None             # <<<<<<<<<<<<<<
 * 
 *         self._queued_streams = set()
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_default_executor);
  __Pyx_DECREF(__pyx_v_self->_default_executor);
  __pyx_v_self->_default_executor = Py_None;

  /* "uvloop/loop.pyx":181
 *         self._default_executor = None
 * 
 *         self._queued_streams = set()             # <<<<<<<<<<<<<<
 *         self._executing_streams = set()
 *         self._ready = col_deque()
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_queued_streams);
  __Pyx_DECREF(__pyx_v_self->_queued_streams);
  __pyx_v_self->_queued_streams = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":182
 * 
 *         self._queued_streams = set()
 *         self._executing_streams = set()             # <<<<<<<<<<<<<<
 *         self._ready = col_deque()
 *         self._ready_len = 0
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_executing_streams);
  __Pyx_DECREF(__pyx_v_self->_executing_streams);
  __pyx_v_self->_executing_streams = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":183
 *         self._queued_streams = set()
 *         self._executing_streams = set()
 *         self._ready = col_deque()             # <<<<<<<<<<<<<<
 *         self._ready_len = 0
 * 
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_col_deque);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_col_deque; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_ready);
  __Pyx_DECREF(__pyx_v_self->_ready);
  __pyx_v_self->_ready = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":184
 *         self._executing_streams = set()
 *         self._ready = col_deque()
 *         self._ready_len = 0             # <<<<<<<<<<<<<<
 * 
 *         self.handler_async = UVAsync.new(
 */
  __pyx_v_self->_ready_len = 0;

  /* "uvloop/loop.pyx":186
 *         self._ready_len = 0
 * 
 *         self.handler_async = UVAsync.new(             # <<<<<<<<<<<<<<
 *             self, <method_t>self._on_wake, self)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_7UVAsync_new(__pyx_v_self, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_on_wake), ((PyObject *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->handler_async);
  __Pyx_DECREF((PyObject *)__pyx_v_self->handler_async);
  __pyx_v_self->handler_async = ((struct __pyx_obj_6uvloop_4loop_UVAsync *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":191
 *         self.handler_idle = UVIdle.new(
 *             self,
 *             new_MethodHandle(             # <<<<<<<<<<<<<<
 *                 self, "loop._on_idle", <method_t>self._on_idle, None, self))
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_new_MethodHandle(__pyx_v_self, __pyx_kp_u_loop__on_idle, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_on_idle), Py_None, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 191, __pyx_L1_error)

  /* "uvloop/loop.pyx":189
 *             self, <method_t>self._on_wake, self)
 * 
 *         self.handler_idle = UVIdle.new(             # <<<<<<<<<<<<<<
 *             self,
 *             new_MethodHandle(
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_6uvloop_4loop_6UVIdle_new(__pyx_v_self, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->handler_idle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->handler_idle);
  __pyx_v_self->handler_idle = ((struct __pyx_obj_6uvloop_4loop_UVIdle *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":198
 *         self.handler_check__exec_writes = UVCheck.new(
 *             self,
 *             new_MethodHandle(             # <<<<<<<<<<<<<<
 *                 self, "loop._exec_queued_writes",
 *                 <method_t>self._exec_queued_writes, None, self))
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_new_MethodHandle(__pyx_v_self, __pyx_kp_u_loop__exec_queued_writes, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_exec_queued_writes), Py_None, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 198, __pyx_L1_error)

  /* "uvloop/loop.pyx":196
 *         # Needed to call `UVStream._exec_write` for writes scheduled
 *         # during `Protocol.data_received`.
 *         self.handler_check__exec_writes = UVCheck.new(             # <<<<<<<<<<<<<<
 *             self,
 *             new_MethodHandle(
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_7UVCheck_new(__pyx_v_self, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->handler_check__exec_writes);
  __Pyx_DECREF((PyObject *)__pyx_v_self->handler_check__exec_writes);
  __pyx_v_self->handler_check__exec_writes = ((struct __pyx_obj_6uvloop_4loop_UVCheck *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":202
 *                 <method_t>self._exec_queued_writes, None, self))
 * 
 *         self._signals = set()             # <<<<<<<<<<<<<<
 *         self._ssock = self._csock = None
 *         self._signal_handlers = {}
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_signals);
  __Pyx_DECREF(__pyx_v_self->_signals);
  __pyx_v_self->_signals = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":203
 * 
 *         self._signals = set()
 *         self._ssock = self._csock = None             # <<<<<<<<<<<<<<
 *         self._signal_handlers = {}
 *         self._listening_signals = False
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_ssock);
  __Pyx_DECREF(__pyx_v_self->_ssock);
  __pyx_v_self->_ssock = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_csock);
  __Pyx_DECREF(__pyx_v_self->_csock);
  __pyx_v_self->_csock = Py_None;

  /* "uvloop/loop.pyx":204
 *         self._signals = set()
 *         self._ssock = self._csock = None
 *         self._signal_handlers = {}             # <<<<<<<<<<<<<<
 *         self._listening_signals = False
 *         self._old_signal_wakeup_id = -1
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_signal_handlers);
  __Pyx_DECREF(__pyx_v_self->_signal_handlers);
  __pyx_v_self->_signal_handlers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":205
 *         self._ssock = self._csock = None
 *         self._signal_handlers = {}
 *         self._listening_signals = False             # <<<<<<<<<<<<<<
 *         self._old_signal_wakeup_id = -1
 * 
 */
  __pyx_v_self->_listening_signals = 0;

  /* "uvloop/loop.pyx":206
 *         self._signal_handlers = {}
 *         self._listening_signals = False
 *         self._old_signal_wakeup_id = -1             # <<<<<<<<<<<<<<
 * 
 *         self._coroutine_debug_set = False
 */
  __pyx_v_self->_old_signal_wakeup_id = -1;

  /* "uvloop/loop.pyx":208
 *         self._old_signal_wakeup_id = -1
 * 
 *         self._coroutine_debug_set = False             # <<<<<<<<<<<<<<
 * 
 *         # A weak set of all asynchronous generators that are
 */
  __pyx_v_self->_coroutine_debug_set = 0;

  /* "uvloop/loop.pyx":212
 *         # A weak set of all asynchronous generators that are
 *         # being iterated by the loop.
 *         self._asyncgens = weakref_WeakSet()             # <<<<<<<<<<<<<<
 * 
 *         # Set to True when `loop.shutdown_asyncgens` is called.
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_weakref_WeakSet);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_weakref_WeakSet; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_asyncgens);
  __Pyx_DECREF(__pyx_v_self->_asyncgens);
  __pyx_v_self->_asyncgens = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":215
 * 
 *         # Set to True when `loop.shutdown_asyncgens` is called.
 *         self._asyncgens_shutdown_called = False             # <<<<<<<<<<<<<<
 *         # Set to True when `loop.shutdown_default_executor` is called.
 *         self._executor_shutdown_called = False
 */
  __pyx_v_self->_asyncgens_shutdown_called = 0;

  /* "uvloop/loop.pyx":217
 *         self._asyncgens_shutdown_called = False
 *         # Set to True when `loop.shutdown_default_executor` is called.
 *         self._executor_shutdown_called = False             # <<<<<<<<<<<<<<
 * 
 *         self._servers = set()
 */
  __pyx_v_self->_executor_shutdown_called = 0;

  /* "uvloop/loop.pyx":219
 *         self._executor_shutdown_called = False
 * 
 *         self._servers = set()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _is_main_thread(self):
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_servers);
  __Pyx_DECREF(__pyx_v_self->_servers);
  __pyx_v_self->_servers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":129
 * @cython.no_gc_clear
 * cdef class Loop:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":221
 *         self._servers = set()
 * 
 *     cdef inline _is_main_thread(self):             # <<<<<<<<<<<<<<
 *         cdef uint64_t main_thread_id = system.MAIN_THREAD_ID
 *         if system.MAIN_THREAD_ID_SET == 0:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__is_main_thread(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  uint64_t __pyx_v_main_thread_id;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  uint64_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_main_thread", 1);

  /* "uvloop/loop.pyx":222
 * 
 *     cdef inline _is_main_thread(self):
 *         cdef uint64_t main_thread_id = system.MAIN_THREAD_ID             # <<<<<<<<<<<<<<
 *         if system.MAIN_THREAD_ID_SET == 0:
 *             main_thread_id = <uint64_t>threading_main_thread().ident
 */
  __pyx_v_main_thread_id = MAIN_THREAD_ID;

  /* "uvloop/loop.pyx":223
 *     cdef inline _is_main_thread(self):
 *         cdef uint64_t main_thread_id = system.MAIN_THREAD_ID
 *         if system.MAIN_THREAD_ID_SET == 0:             # <<<<<<<<<<<<<<
 *             main_thread_id = <uint64_t>threading_main_thread().ident
 *             system.setMainThreadID(main_thread_id)
 */
  __pyx_t_1 = (MAIN_THREAD_ID_SET == 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":224
 *         cdef uint64_t main_thread_id = system.MAIN_THREAD_ID
 *         if system.MAIN_THREAD_ID_SET == 0:
 *             main_thread_id = <uint64_t>threading_main_thread().ident             # <<<<<<<<<<<<<<
 *             system.setMainThreadID(main_thread_id)
 *         return main_thread_id == PyThread_get_thread_ident()
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_threading_main_thread);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_threading_main_thread; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ident); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_uint64_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 224, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_main_thread_id = ((uint64_t)__pyx_t_6);

    /* "uvloop/loop.pyx":225
 *         if system.MAIN_THREAD_ID_SET == 0:
 *             main_thread_id = <uint64_t>threading_main_thread().ident
 *             system.setMainThreadID(main_thread_id)             # <<<<<<<<<<<<<<
 *         return main_thread_id == PyThread_get_thread_ident()
 * 
 */
    setMainThreadID(__pyx_v_main_thread_id);

    /* "uvloop/loop.pyx":223
 *     cdef inline _is_main_thread(self):
 *         cdef uint64_t main_thread_id = system.MAIN_THREAD_ID
 *         if system.MAIN_THREAD_ID_SET == 0:             # <<<<<<<<<<<<<<
 *             main_thread_id = <uint64_t>threading_main_thread().ident
 *             system.setMainThreadID(main_thread_id)
 */
  }

  /* "uvloop/loop.pyx":226
 *             main_thread_id = <uint64_t>threading_main_thread().ident
 *             system.setMainThreadID(main_thread_id)
 *         return main_thread_id == PyThread_get_thread_ident()             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_main_thread_id == PyThread_get_thread_ident())); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":221
 *         self._servers = set()
 * 
 *     cdef inline _is_main_thread(self):             # <<<<<<<<<<<<<<
 *         cdef uint64_t main_thread_id = system.MAIN_THREAD_ID
 *         if system.MAIN_THREAD_ID_SET == 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop._is_main_thread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":228
 *         return main_thread_id == PyThread_get_thread_ident()
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.set_debug(
 *             sys_dev_mode or (not sys_ignore_environment
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_4Loop_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_4Loop_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_2__init__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_4Loop_2__init__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/loop.pyx":229
 * 
 *     def __init__(self):
 *         self.set_debug(             # <<<<<<<<<<<<<<
 *             sys_dev_mode or (not sys_ignore_environment
 *                              and bool(os_environ.get('PYTHONASYNCIODEBUG'))))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/loop.pyx":230
 *     def __init__(self):
 *         self.set_debug(
 *             sys_dev_mode or (not sys_ignore_environment             # <<<<<<<<<<<<<<
 *                              and bool(os_environ.get('PYTHONASYNCIODEBUG'))))
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_6uvloop_4loop_sys_dev_mode); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 230, __pyx_L1_error)
  if (!__pyx_t_4) {
  } else {
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_sys_dev_mode);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_sys_dev_mode;
    goto __pyx_L3_bool_binop_done;
  }

  /* "uvloop/loop.pyx":231
 *         self.set_debug(
 *             sys_dev_mode or (not sys_ignore_environment
 *                              and bool(os_environ.get('PYTHONASYNCIODEBUG'))))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_6uvloop_4loop_sys_ignore_environment); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 230, __pyx_L1_error)

  /* "uvloop/loop.pyx":230
 *     def __init__(self):
 *         self.set_debug(
 *             sys_dev_mode or (not sys_ignore_environment             # <<<<<<<<<<<<<<
 *                              and bool(os_environ.get('PYTHONASYNCIODEBUG'))))
 * 
 */
  __pyx_t_5 = (!__pyx_t_4);
  if (__pyx_t_5) {
  } else {
    __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "uvloop/loop.pyx":231
 *         self.set_debug(
 *             sys_dev_mode or (not sys_ignore_environment
 *                              and bool(os_environ.get('PYTHONASYNCIODEBUG'))))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_os_environ, __pyx_n_s_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_n_u_PYTHONASYNCIODEBUG};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyBool_FromLong((!(!__pyx_t_5))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_3 = __pyx_t_6;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_t_6 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":228
 *         return main_thread_id == PyThread_get_thread_ident()
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.set_debug(
 *             sys_dev_mode or (not sys_ignore_environment
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Loop.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":233
 *                              and bool(os_environ.get('PYTHONASYNCIODEBUG'))))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._running == 1:
 *             raise RuntimeError('deallocating a running event loop!')
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_4Loop_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_4Loop_5__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_4Loop_4__dealloc__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_4Loop_4__dealloc__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/loop.pyx":234
 * 
 *     def __dealloc__(self):
 *         if self._running == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('deallocating a running event loop!')
 *         if self._closed == 0:
 */
  __pyx_t_1 = (__pyx_v_self->_running == 1);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":235
 *     def __dealloc__(self):
 *         if self._running == 1:
 *             raise RuntimeError('deallocating a running event loop!')             # <<<<<<<<<<<<<<
 *         if self._closed == 0:
 *             aio_logger.error("deallocating an open event loop")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 235, __pyx_L1_error)

    /* "uvloop/loop.pyx":234
 * 
 *     def __dealloc__(self):
 *         if self._running == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('deallocating a running event loop!')
 *         if self._closed == 0:
 */
  }

  /* "uvloop/loop.pyx":236
 *         if self._running == 1:
 *             raise RuntimeError('deallocating a running event loop!')
 *         if self._closed == 0:             # <<<<<<<<<<<<<<
 *             aio_logger.error("deallocating an open event loop")
 *             return
 */
  __pyx_t_1 = (__pyx_v_self->_closed == 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":237
 *             raise RuntimeError('deallocating a running event loop!')
 *         if self._closed == 0:
 *             aio_logger.error("deallocating an open event loop")             # <<<<<<<<<<<<<<
 *             return
 *         PyMem_RawFree(self.uvloop)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_deallocating_an_open_event_loop};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":238
 *         if self._closed == 0:
 *             aio_logger.error("deallocating an open event loop")
 *             return             # <<<<<<<<<<<<<<
 *         PyMem_RawFree(self.uvloop)
 *         self.uvloop = NULL
 */
    goto __pyx_L0;

    /* "uvloop/loop.pyx":236
 *         if self._running == 1:
 *             raise RuntimeError('deallocating a running event loop!')
 *         if self._closed == 0:             # <<<<<<<<<<<<<<
 *             aio_logger.error("deallocating an open event loop")
 *             return
 */
  }

  /* "uvloop/loop.pyx":239
 *             aio_logger.error("deallocating an open event loop")
 *             return
 *         PyMem_RawFree(self.uvloop)             # <<<<<<<<<<<<<<
 *         self.uvloop = NULL
 * 
 */
  PyMem_RawFree(__pyx_v_self->uvloop);

  /* "uvloop/loop.pyx":240
 *             return
 *         PyMem_RawFree(self.uvloop)
 *         self.uvloop = NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_debug_fields(self):
 */
  __pyx_v_self->uvloop = NULL;

  /* "uvloop/loop.pyx":233
 *                              and bool(os_environ.get('PYTHONASYNCIODEBUG'))))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._running == 1:
 *             raise RuntimeError('deallocating a running event loop!')
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uvloop.loop.Loop.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/loop.pyx":242
 *         self.uvloop = NULL
 * 
 *     cdef _init_debug_fields(self):             # <<<<<<<<<<<<<<
 *         self._debug_cc = bool(UVLOOP_DEBUG)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__init_debug_fields(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init_debug_fields", 1);

  /* "uvloop/loop.pyx":243
 * 
 *     cdef _init_debug_fields(self):
 *         self._debug_cc = bool(UVLOOP_DEBUG)             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(UVLOOP_DEBUG); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->_debug_cc = (!(!__pyx_t_2));

  /* "uvloop/loop.pyx":245
 *         self._debug_cc = bool(UVLOOP_DEBUG)
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self._debug_handles_current = col_Counter()
 *             self._debug_handles_closed = col_Counter()
 */
  __pyx_t_2 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":246
 * 
 *         if UVLOOP_DEBUG:
 *             self._debug_handles_current = col_Counter()             # <<<<<<<<<<<<<<
 *             self._debug_handles_closed = col_Counter()
 *             self._debug_handles_total = col_Counter()
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_col_Counter);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_col_Counter; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_debug_handles_current);
    __Pyx_DECREF(__pyx_v_self->_debug_handles_current);
    __pyx_v_self->_debug_handles_current = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":247
 *         if UVLOOP_DEBUG:
 *             self._debug_handles_current = col_Counter()
 *             self._debug_handles_closed = col_Counter()             # <<<<<<<<<<<<<<
 *             self._debug_handles_total = col_Counter()
 *         else:
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_col_Counter);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_col_Counter; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_debug_handles_closed);
    __Pyx_DECREF(__pyx_v_self->_debug_handles_closed);
    __pyx_v_self->_debug_handles_closed = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":248
 *             self._debug_handles_current = col_Counter()
 *             self._debug_handles_closed = col_Counter()
 *             self._debug_handles_total = col_Counter()             # <<<<<<<<<<<<<<
 *         else:
 *             self._debug_handles_current = None
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_col_Counter);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_col_Counter; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_debug_handles_total);
    __Pyx_DECREF(__pyx_v_self->_debug_handles_total);
    __pyx_v_self->_debug_handles_total = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":245
 *         self._debug_cc = bool(UVLOOP_DEBUG)
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self._debug_handles_current = col_Counter()
 *             self._debug_handles_closed = col_Counter()
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":250
 *             self._debug_handles_total = col_Counter()
 *         else:
 *             self._debug_handles_current = None             # <<<<<<<<<<<<<<
 *             self._debug_handles_closed = None
 *             self._debug_handles_total = None
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_debug_handles_current);
    __Pyx_DECREF(__pyx_v_self->_debug_handles_current);
    __pyx_v_self->_debug_handles_current = Py_None;

    /* "uvloop/loop.pyx":251
 *         else:
 *             self._debug_handles_current = None
 *             self._debug_handles_closed = None             # <<<<<<<<<<<<<<
 *             self._debug_handles_total = None
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_debug_handles_closed);
    __Pyx_DECREF(__pyx_v_self->_debug_handles_closed);
    __pyx_v_self->_debug_handles_closed = Py_None;

    /* "uvloop/loop.pyx":252
 *             self._debug_handles_current = None
 *             self._debug_handles_closed = None
 *             self._debug_handles_total = None             # <<<<<<<<<<<<<<
 * 
 *         self._debug_uv_handles_total = 0
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_debug_handles_total);
    __Pyx_DECREF(__pyx_v_self->_debug_handles_total);
    __pyx_v_self->_debug_handles_total = Py_None;
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":254
 *             self._debug_handles_total = None
 * 
 *         self._debug_uv_handles_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_uv_handles_freed = 0
 * 
 */
  __pyx_v_self->_debug_uv_handles_total = 0;

  /* "uvloop/loop.pyx":255
 * 
 *         self._debug_uv_handles_total = 0
 *         self._debug_uv_handles_freed = 0             # <<<<<<<<<<<<<<
 * 
 *         self._debug_stream_read_cb_total = 0
 */
  __pyx_v_self->_debug_uv_handles_freed = 0;

  /* "uvloop/loop.pyx":257
 *         self._debug_uv_handles_freed = 0
 * 
 *         self._debug_stream_read_cb_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_read_eof_total = 0
 *         self._debug_stream_read_errors_total = 0
 */
  __pyx_v_self->_debug_stream_read_cb_total = 0;

  /* "uvloop/loop.pyx":258
 * 
 *         self._debug_stream_read_cb_total = 0
 *         self._debug_stream_read_eof_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_read_errors_total = 0
 *         self._debug_stream_read_cb_errors_total = 0
 */
  __pyx_v_self->_debug_stream_read_eof_total = 0;

  /* "uvloop/loop.pyx":259
 *         self._debug_stream_read_cb_total = 0
 *         self._debug_stream_read_eof_total = 0
 *         self._debug_stream_read_errors_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_read_cb_errors_total = 0
 *         self._debug_stream_read_eof_cb_errors_total = 0
 */
  __pyx_v_self->_debug_stream_read_errors_total = 0;

  /* "uvloop/loop.pyx":260
 *         self._debug_stream_read_eof_total = 0
 *         self._debug_stream_read_errors_total = 0
 *         self._debug_stream_read_cb_errors_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_read_eof_cb_errors_total = 0
 * 
 */
  __pyx_v_self->_debug_stream_read_cb_errors_total = 0;

  /* "uvloop/loop.pyx":261
 *         self._debug_stream_read_errors_total = 0
 *         self._debug_stream_read_cb_errors_total = 0
 *         self._debug_stream_read_eof_cb_errors_total = 0             # <<<<<<<<<<<<<<
 * 
 *         self._debug_stream_shutdown_errors_total = 0
 */
  __pyx_v_self->_debug_stream_read_eof_cb_errors_total = 0;

  /* "uvloop/loop.pyx":263
 *         self._debug_stream_read_eof_cb_errors_total = 0
 * 
 *         self._debug_stream_shutdown_errors_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_listen_errors_total = 0
 * 
 */
  __pyx_v_self->_debug_stream_shutdown_errors_total = 0;

  /* "uvloop/loop.pyx":264
 * 
 *         self._debug_stream_shutdown_errors_total = 0
 *         self._debug_stream_listen_errors_total = 0             # <<<<<<<<<<<<<<
 * 
 *         self._debug_stream_write_tries = 0
 */
  __pyx_v_self->_debug_stream_listen_errors_total = 0;

  /* "uvloop/loop.pyx":266
 *         self._debug_stream_listen_errors_total = 0
 * 
 *         self._debug_stream_write_tries = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_write_errors_total = 0
 *         self._debug_stream_write_ctx_total = 0
 */
  __pyx_v_self->_debug_stream_write_tries = 0;

  /* "uvloop/loop.pyx":267
 * 
 *         self._debug_stream_write_tries = 0
 *         self._debug_stream_write_errors_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_write_ctx_total = 0
 *         self._debug_stream_write_ctx_cnt = 0
 */
  __pyx_v_self->_debug_stream_write_errors_total = 0;

  /* "uvloop/loop.pyx":268
 *         self._debug_stream_write_tries = 0
 *         self._debug_stream_write_errors_total = 0
 *         self._debug_stream_write_ctx_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_write_ctx_cnt = 0
 *         self._debug_stream_write_cb_errors_total = 0
 */
  __pyx_v_self->_debug_stream_write_ctx_total = 0;

  /* "uvloop/loop.pyx":269
 *         self._debug_stream_write_errors_total = 0
 *         self._debug_stream_write_ctx_total = 0
 *         self._debug_stream_write_ctx_cnt = 0             # <<<<<<<<<<<<<<
 *         self._debug_stream_write_cb_errors_total = 0
 * 
 */
  __pyx_v_self->_debug_stream_write_ctx_cnt = 0;

  /* "uvloop/loop.pyx":270
 *         self._debug_stream_write_ctx_total = 0
 *         self._debug_stream_write_ctx_cnt = 0
 *         self._debug_stream_write_cb_errors_total = 0             # <<<<<<<<<<<<<<
 * 
 *         self._debug_cb_handles_total = 0
 */
  __pyx_v_self->_debug_stream_write_cb_errors_total = 0;

  /* "uvloop/loop.pyx":272
 *         self._debug_stream_write_cb_errors_total = 0
 * 
 *         self._debug_cb_handles_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_cb_handles_count = 0
 * 
 */
  __pyx_v_self->_debug_cb_handles_total = 0;

  /* "uvloop/loop.pyx":273
 * 
 *         self._debug_cb_handles_total = 0
 *         self._debug_cb_handles_count = 0             # <<<<<<<<<<<<<<
 * 
 *         self._debug_cb_timer_handles_total = 0
 */
  __pyx_v_self->_debug_cb_handles_count = 0;

  /* "uvloop/loop.pyx":275
 *         self._debug_cb_handles_count = 0
 * 
 *         self._debug_cb_timer_handles_total = 0             # <<<<<<<<<<<<<<
 *         self._debug_cb_timer_handles_count = 0
 * 
 */
  __pyx_v_self->_debug_cb_timer_handles_total = 0;

  /* "uvloop/loop.pyx":276
 * 
 *         self._debug_cb_timer_handles_total = 0
 *         self._debug_cb_timer_handles_count = 0             # <<<<<<<<<<<<<<
 * 
 *         self._poll_read_events_total = 0
 */
  __pyx_v_self->_debug_cb_timer_handles_count = 0;

  /* "uvloop/loop.pyx":278
 *         self._debug_cb_timer_handles_count = 0
 * 
 *         self._poll_read_events_total = 0             # <<<<<<<<<<<<<<
 *         self._poll_read_cb_errors_total = 0
 *         self._poll_write_events_total = 0
 */
  __pyx_v_self->_poll_read_events_total = 0;

  /* "uvloop/loop.pyx":279
 * 
 *         self._poll_read_events_total = 0
 *         self._poll_read_cb_errors_total = 0             # <<<<<<<<<<<<<<
 *         self._poll_write_events_total = 0
 *         self._poll_write_cb_errors_total = 0
 */
  __pyx_v_self->_poll_read_cb_errors_total = 0;

  /* "uvloop/loop.pyx":280
 *         self._poll_read_events_total = 0
 *         self._poll_read_cb_errors_total = 0
 *         self._poll_write_events_total = 0             # <<<<<<<<<<<<<<
 *         self._poll_write_cb_errors_total = 0
 * 
 */
  __pyx_v_self->_poll_write_events_total = 0;

  /* "uvloop/loop.pyx":281
 *         self._poll_read_cb_errors_total = 0
 *         self._poll_write_events_total = 0
 *         self._poll_write_cb_errors_total = 0             # <<<<<<<<<<<<<<
 * 
 *         self._sock_try_write_total = 0
 */
  __pyx_v_self->_poll_write_cb_errors_total = 0;

  /* "uvloop/loop.pyx":283
 *         self._poll_write_cb_errors_total = 0
 * 
 *         self._sock_try_write_total = 0             # <<<<<<<<<<<<<<
 * 
 *         self._debug_exception_handler_cnt = 0
 */
  __pyx_v_self->_sock_try_write_total = 0;

  /* "uvloop/loop.pyx":285
 *         self._sock_try_write_total = 0
 * 
 *         self._debug_exception_handler_cnt = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef _setup_or_resume_signals(self):
 */
  __pyx_v_self->_debug_exception_handler_cnt = 0;

  /* "uvloop/loop.pyx":242
 *         self.uvloop = NULL
 * 
 *     cdef _init_debug_fields(self):             # <<<<<<<<<<<<<<
 *         self._debug_cc = bool(UVLOOP_DEBUG)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop._init_debug_fields", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":287
 *         self._debug_exception_handler_cnt = 0
 * 
 *     cdef _setup_or_resume_signals(self):             # <<<<<<<<<<<<<<
 *         if not self._is_main_thread():
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__setup_or_resume_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_v_fileno = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_setup_or_resume_signals", 1);

  /* "uvloop/loop.pyx":288
 * 
 *     cdef _setup_or_resume_signals(self):
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__is_main_thread(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":289
 *     cdef _setup_or_resume_signals(self):
 *         if not self._is_main_thread():
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self._listening_signals:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":288
 * 
 *     cdef _setup_or_resume_signals(self):
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/loop.pyx":291
 *             return
 * 
 *         if self._listening_signals:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('signals handling has been already setup')
 * 
 */
  if (unlikely(__pyx_v_self->_listening_signals)) {

    /* "uvloop/loop.pyx":292
 * 
 *         if self._listening_signals:
 *             raise RuntimeError('signals handling has been already setup')             # <<<<<<<<<<<<<<
 * 
 *         if self._ssock is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 292, __pyx_L1_error)

    /* "uvloop/loop.pyx":291
 *             return
 * 
 *         if self._listening_signals:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('signals handling has been already setup')
 * 
 */
  }

  /* "uvloop/loop.pyx":294
 *             raise RuntimeError('signals handling has been already setup')
 * 
 *         if self._ssock is not None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('self-pipe exists before loop run')
 * 
 */
  __pyx_t_3 = (__pyx_v_self->_ssock != Py_None);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":295
 * 
 *         if self._ssock is not None:
 *             raise RuntimeError('self-pipe exists before loop run')             # <<<<<<<<<<<<<<
 * 
 *         # Create a self-pipe and call set_signal_wakeup_fd() with one
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 295, __pyx_L1_error)

    /* "uvloop/loop.pyx":294
 *             raise RuntimeError('signals handling has been already setup')
 * 
 *         if self._ssock is not None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('self-pipe exists before loop run')
 * 
 */
  }

  /* "uvloop/loop.pyx":302
 *         # standard Python's one or or user set their own.)
 * 
 *         self._ssock, self._csock = socket_socketpair()             # <<<<<<<<<<<<<<
 *         try:
 *             self._ssock.setblocking(False)
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_socketpair);
  __pyx_t_4 = __pyx_v_6uvloop_4loop_socket_socketpair; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 302, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
    index = 0; __pyx_t_4 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 302, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 302, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_ssock);
  __Pyx_DECREF(__pyx_v_self->_ssock);
  __pyx_v_self->_ssock = __pyx_t_4;
  __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->_csock);
  __Pyx_DECREF(__pyx_v_self->_csock);
  __pyx_v_self->_csock = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":303
 * 
 *         self._ssock, self._csock = socket_socketpair()
 *         try:             # <<<<<<<<<<<<<<
 *             self._ssock.setblocking(False)
 *             self._csock.setblocking(False)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "uvloop/loop.pyx":304
 *         self._ssock, self._csock = socket_socketpair()
 *         try:
 *             self._ssock.setblocking(False)             # <<<<<<<<<<<<<<
 *             self._csock.setblocking(False)
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_ssock, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 304, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, Py_False};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":305
 *         try:
 *             self._ssock.setblocking(False)
 *             self._csock.setblocking(False)             # <<<<<<<<<<<<<<
 * 
 *             fileno = self._csock.fileno()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_csock, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 305, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, Py_False};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":307
 *             self._csock.setblocking(False)
 * 
 *             fileno = self._csock.fileno()             # <<<<<<<<<<<<<<
 * 
 *             self._old_signal_wakeup_id = _set_signal_wakeup_fd(fileno)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_csock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 307, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_v_fileno = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":309
 *             fileno = self._csock.fileno()
 * 
 *             self._old_signal_wakeup_id = _set_signal_wakeup_fd(fileno)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             # Out of all statements in the try block, only the
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop__set_signal_wakeup_fd(__pyx_v_fileno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 309, __pyx_L8_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_self->_old_signal_wakeup_id = __pyx_t_12;

      /* "uvloop/loop.pyx":303
 * 
 *         self._ssock, self._csock = socket_socketpair()
 *         try:             # <<<<<<<<<<<<<<
 *             self._ssock.setblocking(False)
 *             self._csock.setblocking(False)
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L13_try_end;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":310
 * 
 *             self._old_signal_wakeup_id = _set_signal_wakeup_fd(fileno)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             # Out of all statements in the try block, only the
 *             # "_set_signal_wakeup_fd()" call can fail, but it shouldn't,
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_12) {
      __Pyx_AddTraceback("uvloop.loop.Loop._setup_or_resume_signals", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 310, __pyx_L10_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "uvloop/loop.pyx":316
 *             # Still, if something goes horribly wrong we want to clean up
 *             # the socket pair.
 *             self._ssock.close()             # <<<<<<<<<<<<<<
 *             self._csock.close()
 *             self._ssock = None
 */
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_ssock, __pyx_n_s_close); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 316, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 316, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":317
 *             # the socket pair.
 *             self._ssock.close()
 *             self._csock.close()             # <<<<<<<<<<<<<<
 *             self._ssock = None
 *             self._csock = None
 */
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_csock, __pyx_n_s_close); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 317, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 317, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":318
 *             self._ssock.close()
 *             self._csock.close()
 *             self._ssock = None             # <<<<<<<<<<<<<<
 *             self._csock = None
 *             raise
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->_ssock);
      __Pyx_DECREF(__pyx_v_self->_ssock);
      __pyx_v_self->_ssock = Py_None;

      /* "uvloop/loop.pyx":319
 *             self._csock.close()
 *             self._ssock = None
 *             self._csock = None             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->_csock);
      __Pyx_DECREF(__pyx_v_self->_csock);
      __pyx_v_self->_csock = Py_None;

      /* "uvloop/loop.pyx":320
 *             self._ssock = None
 *             self._csock = None
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         self._add_reader(
 */
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_5, __pyx_t_4);
      __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
      __PYX_ERR(0, 320, __pyx_L10_except_error)
    }
    goto __pyx_L10_except_error;

    /* "uvloop/loop.pyx":303
 * 
 *         self._ssock, self._csock = socket_socketpair()
 *         try:             # <<<<<<<<<<<<<<
 *             self._ssock.setblocking(False)
 *             self._csock.setblocking(False)
 */
    __pyx_L10_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L13_try_end:;
  }

  /* "uvloop/loop.pyx":323
 * 
 *         self._add_reader(
 *             self._ssock,             # <<<<<<<<<<<<<<
 *             new_MethodHandle(
 *                 self,
 */
  __pyx_t_4 = __pyx_v_self->_ssock;
  __Pyx_INCREF(__pyx_t_4);

  /* "uvloop/loop.pyx":324
 *         self._add_reader(
 *             self._ssock,
 *             new_MethodHandle(             # <<<<<<<<<<<<<<
 *                 self,
 *                 "Loop._read_from_self",
 */
  __pyx_t_5 = __pyx_f_6uvloop_4loop_new_MethodHandle(__pyx_v_self, __pyx_kp_u_Loop__read_from_self, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_read_from_self), Py_None, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "uvloop/loop.pyx":322
 *             raise
 * 
 *         self._add_reader(             # <<<<<<<<<<<<<<
 *             self._ssock,
 *             new_MethodHandle(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_add_reader(__pyx_v_self, __pyx_t_4, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":331
 *                 self))
 * 
 *         self._listening_signals = True             # <<<<<<<<<<<<<<
 * 
 *     cdef _pause_signals(self):
 */
  __pyx_v_self->_listening_signals = 1;

  /* "uvloop/loop.pyx":287
 *         self._debug_exception_handler_cnt = 0
 * 
 *     cdef _setup_or_resume_signals(self):             # <<<<<<<<<<<<<<
 *         if not self._is_main_thread():
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("uvloop.loop.Loop._setup_or_resume_signals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fileno);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":333
 *         self._listening_signals = True
 * 
 *     cdef _pause_signals(self):             # <<<<<<<<<<<<<<
 *         if not self._is_main_thread():
 *             if self._listening_signals:
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__pause_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pause_signals", 1);

  /* "uvloop/loop.pyx":334
 * 
 *     cdef _pause_signals(self):
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             if self._listening_signals:
 *                 raise RuntimeError(
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__is_main_thread(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":335
 *     cdef _pause_signals(self):
 *         if not self._is_main_thread():
 *             if self._listening_signals:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     'cannot pause signals handling; no longer running in '
 */
    if (unlikely(__pyx_v_self->_listening_signals)) {

      /* "uvloop/loop.pyx":336
 *         if not self._is_main_thread():
 *             if self._listening_signals:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     'cannot pause signals handling; no longer running in '
 *                     'the main thread')
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 336, __pyx_L1_error)

      /* "uvloop/loop.pyx":335
 *     cdef _pause_signals(self):
 *         if not self._is_main_thread():
 *             if self._listening_signals:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     'cannot pause signals handling; no longer running in '
 */
    }

    /* "uvloop/loop.pyx":340
 *                     'the main thread')
 *             else:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         if not self._listening_signals:
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;
    }

    /* "uvloop/loop.pyx":334
 * 
 *     cdef _pause_signals(self):
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             if self._listening_signals:
 *                 raise RuntimeError(
 */
  }

  /* "uvloop/loop.pyx":342
 *                 return
 * 
 *         if not self._listening_signals:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('signals handling has not been setup')
 * 
 */
  __pyx_t_3 = (!__pyx_v_self->_listening_signals);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":343
 * 
 *         if not self._listening_signals:
 *             raise RuntimeError('signals handling has not been setup')             # <<<<<<<<<<<<<<
 * 
 *         self._listening_signals = False
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 343, __pyx_L1_error)

    /* "uvloop/loop.pyx":342
 *                 return
 * 
 *         if not self._listening_signals:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('signals handling has not been setup')
 * 
 */
  }

  /* "uvloop/loop.pyx":345
 *             raise RuntimeError('signals handling has not been setup')
 * 
 *         self._listening_signals = False             # <<<<<<<<<<<<<<
 * 
 *         _set_signal_wakeup_fd(self._old_signal_wakeup_id)
 */
  __pyx_v_self->_listening_signals = 0;

  /* "uvloop/loop.pyx":347
 *         self._listening_signals = False
 * 
 *         _set_signal_wakeup_fd(self._old_signal_wakeup_id)             # <<<<<<<<<<<<<<
 * 
 *         self._remove_reader(self._ssock)
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_old_signal_wakeup_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_f_6uvloop_4loop__set_signal_wakeup_fd(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":349
 *         _set_signal_wakeup_fd(self._old_signal_wakeup_id)
 * 
 *         self._remove_reader(self._ssock)             # <<<<<<<<<<<<<<
 *         self._ssock.close()
 *         self._csock.close()
 */
  __pyx_t_4 = __pyx_v_self->_ssock;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":350
 * 
 *         self._remove_reader(self._ssock)
 *         self._ssock.close()             # <<<<<<<<<<<<<<
 *         self._csock.close()
 *         self._ssock = None
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_ssock, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":351
 *         self._remove_reader(self._ssock)
 *         self._ssock.close()
 *         self._csock.close()             # <<<<<<<<<<<<<<
 *         self._ssock = None
 *         self._csock = None
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_csock, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":352
 *         self._ssock.close()
 *         self._csock.close()
 *         self._ssock = None             # <<<<<<<<<<<<<<
 *         self._csock = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_ssock);
  __Pyx_DECREF(__pyx_v_self->_ssock);
  __pyx_v_self->_ssock = Py_None;

  /* "uvloop/loop.pyx":353
 *         self._csock.close()
 *         self._ssock = None
 *         self._csock = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _shutdown_signals(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_csock);
  __Pyx_DECREF(__pyx_v_self->_csock);
  __pyx_v_self->_csock = Py_None;

  /* "uvloop/loop.pyx":333
 *         self._listening_signals = True
 * 
 *     cdef _pause_signals(self):             # <<<<<<<<<<<<<<
 *         if not self._is_main_thread():
 *             if self._listening_signals:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.Loop._pause_signals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":355
 *         self._csock = None
 * 
 *     cdef _shutdown_signals(self):             # <<<<<<<<<<<<<<
 *         if not self._is_main_thread():
 *             if self._signal_handlers:
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__shutdown_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_v_sig = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_shutdown_signals", 1);

  /* "uvloop/loop.pyx":356
 * 
 *     cdef _shutdown_signals(self):
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             if self._signal_handlers:
 *                 aio_logger.warning(
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__is_main_thread(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":357
 *     cdef _shutdown_signals(self):
 *         if not self._is_main_thread():
 *             if self._signal_handlers:             # <<<<<<<<<<<<<<
 *                 aio_logger.warning(
 *                     'cannot cleanup signal handlers: closing the event loop '
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_self->_signal_handlers); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 357, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "uvloop/loop.pyx":358
 *         if not self._is_main_thread():
 *             if self._signal_handlers:
 *                 aio_logger.warning(             # <<<<<<<<<<<<<<
 *                     'cannot cleanup signal handlers: closing the event loop '
 *                     'in a non-main OS thread')
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_kp_u_cannot_cleanup_signal_handlers_c};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":357
 *     cdef _shutdown_signals(self):
 *         if not self._is_main_thread():
 *             if self._signal_handlers:             # <<<<<<<<<<<<<<
 *                 aio_logger.warning(
 *                     'cannot cleanup signal handlers: closing the event loop '
 */
    }

    /* "uvloop/loop.pyx":361
 *                     'cannot cleanup signal handlers: closing the event loop '
 *                     'in a non-main OS thread')
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self._listening_signals:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":356
 * 
 *     cdef _shutdown_signals(self):
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             if self._signal_handlers:
 *                 aio_logger.warning(
 */
  }

  /* "uvloop/loop.pyx":363
 *             return
 * 
 *         if self._listening_signals:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'cannot shutdown signals handling as it has not been paused')
 */
  if (unlikely(__pyx_v_self->_listening_signals)) {

    /* "uvloop/loop.pyx":364
 * 
 *         if self._listening_signals:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'cannot shutdown signals handling as it has not been paused')
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 364, __pyx_L1_error)

    /* "uvloop/loop.pyx":363
 *             return
 * 
 *         if self._listening_signals:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'cannot shutdown signals handling as it has not been paused')
 */
  }

  /* "uvloop/loop.pyx":367
 *                 'cannot shutdown signals handling as it has not been paused')
 * 
 *         if self._ssock:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'self-pipe was not cleaned up after loop was run')
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_self->_ssock); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 367, __pyx_L1_error)
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":368
 * 
 *         if self._ssock:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'self-pipe was not cleaned up after loop was run')
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 368, __pyx_L1_error)

    /* "uvloop/loop.pyx":367
 *                 'cannot shutdown signals handling as it has not been paused')
 * 
 *         if self._ssock:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'self-pipe was not cleaned up after loop was run')
 */
  }

  /* "uvloop/loop.pyx":371
 *                 'self-pipe was not cleaned up after loop was run')
 * 
 *         for sig in list(self._signal_handlers):             # <<<<<<<<<<<<<<
 *             self.remove_signal_handler(sig)
 * 
 */
  __pyx_t_1 = PySequence_List(__pyx_v_self->_signal_handlers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 371, __pyx_L1_error)
      #endif
      if (__pyx_t_7 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 371, __pyx_L1_error)
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":372
 * 
 *         for sig in list(self._signal_handlers):
 *             self.remove_signal_handler(sig)             # <<<<<<<<<<<<<<
 * 
 *     def __sighandler(self, signum, frame):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove_signal_handler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_sig};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":371
 *                 'self-pipe was not cleaned up after loop was run')
 * 
 *         for sig in list(self._signal_handlers):             # <<<<<<<<<<<<<<
 *             self.remove_signal_handler(sig)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":355
 *         self._csock = None
 * 
 *     cdef _shutdown_signals(self):             # <<<<<<<<<<<<<<
 *         if not self._is_main_thread():
 *             if self._signal_handlers:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Loop._shutdown_signals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":374
 *             self.remove_signal_handler(sig)
 * 
 *     def __sighandler(self, signum, frame):             # <<<<<<<<<<<<<<
 *         self._signals.add(signum)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_7_Loop__sighandler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_6__sighandler, "Loop.__sighandler(self, signum, frame)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_7_Loop__sighandler = {"_Loop__sighandler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_7_Loop__sighandler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_6__sighandler};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_7_Loop__sighandler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_signum = 0;
  CYTHON_UNUSED PyObject *__pyx_v_frame = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Loop__sighandler (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signum,&__pyx_n_s_frame,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_signum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_frame)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_Loop__sighandler", 1, 2, 2, 1); __PYX_ERR(0, 374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_Loop__sighandler") < 0)) __PYX_ERR(0, 374, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_signum = values[0];
    __pyx_v_frame = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_Loop__sighandler", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 374, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop._Loop__sighandler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_6__sighandler(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_signum, __pyx_v_frame);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_6__sighandler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_signum, CYTHON_UNUSED PyObject *__pyx_v_frame) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Loop__sighandler", 1);

  /* "uvloop/loop.pyx":375
 * 
 *     def __sighandler(self, signum, frame):
 *         self._signals.add(signum)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _ceval_process_signals(self):
 */
  if (unlikely(__pyx_v_self->_signals == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(0, 375, __pyx_L1_error)
  }
  __pyx_t_1 = PySet_Add(__pyx_v_self->_signals, __pyx_v_signum); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 375, __pyx_L1_error)

  /* "uvloop/loop.pyx":374
 *             self.remove_signal_handler(sig)
 * 
 *     def __sighandler(self, signum, frame):             # <<<<<<<<<<<<<<
 *         self._signals.add(signum)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop._Loop__sighandler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":377
 *         self._signals.add(signum)
 * 
 *     cdef inline _ceval_process_signals(self):             # <<<<<<<<<<<<<<
 *         # Invoke CPython eval loop to let process signals.
 *         PyErr_CheckSignals()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__ceval_process_signals(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ceval_process_signals", 1);

  /* "uvloop/loop.pyx":379
 *     cdef inline _ceval_process_signals(self):
 *         # Invoke CPython eval loop to let process signals.
 *         PyErr_CheckSignals()             # <<<<<<<<<<<<<<
 *         # Calling a pure-Python function will invoke
 *         # _PyEval_EvalFrameDefault which will process
 */
  __pyx_t_1 = PyErr_CheckSignals(); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 379, __pyx_L1_error)

  /* "uvloop/loop.pyx":383
 *         # _PyEval_EvalFrameDefault which will process
 *         # pending signal callbacks.
 *         _noop.noop()  # Might raise ^C             # <<<<<<<<<<<<<<
 * 
 *     cdef _read_from_self(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_noop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_noop_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":377
 *         self._signals.add(signum)
 * 
 *     cdef inline _ceval_process_signals(self):             # <<<<<<<<<<<<<<
 *         # Invoke CPython eval loop to let process signals.
 *         PyErr_CheckSignals()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop._ceval_process_signals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":385
 *         _noop.noop()  # Might raise ^C
 * 
 *     cdef _read_from_self(self):             # <<<<<<<<<<<<<<
 *         cdef bytes sigdata
 *         sigdata = b''
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__read_from_self(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_v_sigdata = 0;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_from_self", 1);

  /* "uvloop/loop.pyx":387
 *     cdef _read_from_self(self):
 *         cdef bytes sigdata
 *         sigdata = b''             # <<<<<<<<<<<<<<
 *         while True:
 *             try:
 */
  __Pyx_INCREF(__pyx_kp_b__10);
  __pyx_v_sigdata = __pyx_kp_b__10;

  /* "uvloop/loop.pyx":388
 *         cdef bytes sigdata
 *         sigdata = b''
 *         while True:             # <<<<<<<<<<<<<<
 *             try:
 *                 data = self._ssock.recv(65536)
 */
  while (1) {

    /* "uvloop/loop.pyx":389
 *         sigdata = b''
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 data = self._ssock.recv(65536)
 *                 if not data:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      /*try:*/ {

        /* "uvloop/loop.pyx":390
 *         while True:
 *             try:
 *                 data = self._ssock.recv(65536)             # <<<<<<<<<<<<<<
 *                 if not data:
 *                     break
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_ssock, __pyx_n_s_recv); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 390, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_int_65536};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":391
 *             try:
 *                 data = self._ssock.recv(65536)
 *                 if not data:             # <<<<<<<<<<<<<<
 *                     break
 *                 sigdata += data
 */
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_data); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 391, __pyx_L5_error)
        __pyx_t_9 = (!__pyx_t_8);
        if (__pyx_t_9) {

          /* "uvloop/loop.pyx":392
 *                 data = self._ssock.recv(65536)
 *                 if not data:
 *                     break             # <<<<<<<<<<<<<<
 *                 sigdata += data
 *             except InterruptedError:
 */
          goto __pyx_L10_try_break;

          /* "uvloop/loop.pyx":391
 *             try:
 *                 data = self._ssock.recv(65536)
 *                 if not data:             # <<<<<<<<<<<<<<
 *                     break
 *                 sigdata += data
 */
        }

        /* "uvloop/loop.pyx":393
 *                 if not data:
 *                     break
 *                 sigdata += data             # <<<<<<<<<<<<<<
 *             except InterruptedError:
 *                 continue
 */
        __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_sigdata, __pyx_v_data); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 393, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 393, __pyx_L5_error)
        __Pyx_DECREF_SET(__pyx_v_sigdata, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":389
 *         sigdata = b''
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 data = self._ssock.recv(65536)
 *                 if not data:
 */
      }
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":394
 *                     break
 *                 sigdata += data
 *             except InterruptedError:             # <<<<<<<<<<<<<<
 *                 continue
 *             except BlockingIOError:
 */
      __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 394, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.Loop._read_from_self", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 394, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);

        /* "uvloop/loop.pyx":395
 *                 sigdata += data
 *             except InterruptedError:
 *                 continue             # <<<<<<<<<<<<<<
 *             except BlockingIOError:
 *                 break
 */
        goto __pyx_L14_except_continue;
        __pyx_L14_except_continue:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L11_try_continue;
      }

      /* "uvloop/loop.pyx":396
 *             except InterruptedError:
 *                 continue
 *             except BlockingIOError:             # <<<<<<<<<<<<<<
 *                 break
 *         if sigdata:
 */
      __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 396, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.Loop._read_from_self", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 396, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);

        /* "uvloop/loop.pyx":397
 *                 continue
 *             except BlockingIOError:
 *                 break             # <<<<<<<<<<<<<<
 *         if sigdata:
 *             self._invoke_signals(sigdata)
 */
        goto __pyx_L17_except_break;
        __pyx_L17_except_break:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L10_try_break;
      }
      goto __pyx_L7_except_error;

      /* "uvloop/loop.pyx":389
 *         sigdata = b''
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 data = self._ssock.recv(65536)
 *                 if not data:
 */
      __pyx_L7_except_error:;
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
      goto __pyx_L1_error;
      __pyx_L10_try_break:;
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
      goto __pyx_L4_break;
      __pyx_L11_try_continue:;
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
      goto __pyx_L3_continue;
      __pyx_L12_try_end:;
    }
    __pyx_L3_continue:;
  }
  __pyx_L4_break:;

  /* "uvloop/loop.pyx":398
 *             except BlockingIOError:
 *                 break
 *         if sigdata:             # <<<<<<<<<<<<<<
 *             self._invoke_signals(sigdata)
 * 
 */
  __pyx_t_9 = (__pyx_v_sigdata != Py_None)&&(PyBytes_GET_SIZE(__pyx_v_sigdata) != 0);
  if (__pyx_t_9) {

    /* "uvloop/loop.pyx":399
 *                 break
 *         if sigdata:
 *             self._invoke_signals(sigdata)             # <<<<<<<<<<<<<<
 * 
 *     cdef _invoke_signals(self, bytes data):
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_invoke_signals(__pyx_v_self, __pyx_v_sigdata); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":398
 *             except BlockingIOError:
 *                 break
 *         if sigdata:             # <<<<<<<<<<<<<<
 *             self._invoke_signals(sigdata)
 * 
 */
  }

  /* "uvloop/loop.pyx":385
 *         _noop.noop()  # Might raise ^C
 * 
 *     cdef _read_from_self(self):             # <<<<<<<<<<<<<<
 *         cdef bytes sigdata
 *         sigdata = b''
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uvloop.loop.Loop._read_from_self", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sigdata);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":401
 *             self._invoke_signals(sigdata)
 * 
 *     cdef _invoke_signals(self, bytes data):             # <<<<<<<<<<<<<<
 *         cdef set sigs
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__invoke_signals(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_v_sigs = 0;
  PyObject *__pyx_v_signum = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_invoke_signals", 1);

  /* "uvloop/loop.pyx":404
 *         cdef set sigs
 * 
 *         self._ceval_process_signals()             # <<<<<<<<<<<<<<
 * 
 *         sigs = self._signals.copy()
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__ceval_process_signals(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":406
 *         self._ceval_process_signals()
 * 
 *         sigs = self._signals.copy()             # <<<<<<<<<<<<<<
 *         self._signals.clear()
 *         for signum in data:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_signals, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PySet_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_t_1))) __PYX_ERR(0, 406, __pyx_L1_error)
  __pyx_v_sigs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":407
 * 
 *         sigs = self._signals.copy()
 *         self._signals.clear()             # <<<<<<<<<<<<<<
 *         for signum in data:
 *             if not signum:
 */
  if (unlikely(__pyx_v_self->_signals == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
    __PYX_ERR(0, 407, __pyx_L1_error)
  }
  __pyx_t_5 = PySet_Clear(__pyx_v_self->_signals); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 407, __pyx_L1_error)

  /* "uvloop/loop.pyx":408
 *         sigs = self._signals.copy()
 *         self._signals.clear()
 *         for signum in data:             # <<<<<<<<<<<<<<
 *             if not signum:
 *                 # ignore null bytes written by set_wakeup_fd()
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 408, __pyx_L1_error)
  for (;;) {
    {
      __pyx_t_2 = __pyx_t_6(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 408, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_signum, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":409
 *         self._signals.clear()
 *         for signum in data:
 *             if not signum:             # <<<<<<<<<<<<<<
 *                 # ignore null bytes written by set_wakeup_fd()
 *                 continue
 */
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_signum); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 409, __pyx_L1_error)
    __pyx_t_8 = (!__pyx_t_7);
    if (__pyx_t_8) {

      /* "uvloop/loop.pyx":411
 *             if not signum:
 *                 # ignore null bytes written by set_wakeup_fd()
 *                 continue             # <<<<<<<<<<<<<<
 *             sigs.discard(signum)
 *             self._handle_signal(signum)
 */
      goto __pyx_L3_continue;

      /* "uvloop/loop.pyx":409
 *         self._signals.clear()
 *         for signum in data:
 *             if not signum:             # <<<<<<<<<<<<<<
 *                 # ignore null bytes written by set_wakeup_fd()
 *                 continue
 */
    }

    /* "uvloop/loop.pyx":412
 *                 # ignore null bytes written by set_wakeup_fd()
 *                 continue
 *             sigs.discard(signum)             # <<<<<<<<<<<<<<
 *             self._handle_signal(signum)
 * 
 */
    if (unlikely(__pyx_v_sigs == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "discard");
      __PYX_ERR(0, 412, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PySet_Discard(__pyx_v_sigs, __pyx_v_signum); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 412, __pyx_L1_error)

    /* "uvloop/loop.pyx":413
 *                 continue
 *             sigs.discard(signum)
 *             self._handle_signal(signum)             # <<<<<<<<<<<<<<
 * 
 *         for signum in sigs:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_handle_signal(__pyx_v_self, __pyx_v_signum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":408
 *         sigs = self._signals.copy()
 *         self._signals.clear()
 *         for signum in data:             # <<<<<<<<<<<<<<
 *             if not signum:
 *                 # ignore null bytes written by set_wakeup_fd()
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":415
 *             self._handle_signal(signum)
 * 
 *         for signum in sigs:             # <<<<<<<<<<<<<<
 *             # Since not all signals are registered by add_signal_handler()
 *             # (for instance, we use the default SIGINT handler) not all
 */
  __pyx_t_9 = 0;
  __pyx_t_2 = __Pyx_set_iterator(__pyx_v_sigs, 1, (&__pyx_t_10), (&__pyx_t_11)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_2;
  __pyx_t_2 = 0;
  while (1) {
    __pyx_t_12 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_10, &__pyx_t_9, &__pyx_t_2, __pyx_t_11);
    if (unlikely(__pyx_t_12 == 0)) break;
    if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_signum, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":422
 *             # from __sighandler; this ensures that signals shouldn't be
 *             # lost even if set_wakeup_fd() couldn't write to the self-pipe.
 *             self._handle_signal(signum)             # <<<<<<<<<<<<<<
 * 
 *     cdef _handle_signal(self, sig):
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_handle_signal(__pyx_v_self, __pyx_v_signum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":401
 *             self._invoke_signals(sigdata)
 * 
 *     cdef _invoke_signals(self, bytes data):             # <<<<<<<<<<<<<<
 *         cdef set sigs
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Loop._invoke_signals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sigs);
  __Pyx_XDECREF(__pyx_v_signum);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":424
 *             self._handle_signal(signum)
 * 
 *     cdef _handle_signal(self, sig):             # <<<<<<<<<<<<<<
 *         cdef Handle handle
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__handle_signal(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_handle_signal", 1);

  /* "uvloop/loop.pyx":427
 *         cdef Handle handle
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             handle = <Handle>(self._signal_handlers[sig])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/loop.pyx":428
 * 
 *         try:
 *             handle = <Handle>(self._signal_handlers[sig])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             handle = None
 */
      if (unlikely(__pyx_v_self->_signal_handlers == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 428, __pyx_L3_error)
      }
      __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_self->_signal_handlers, __pyx_v_sig); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 428, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __pyx_t_4;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":427
 *         cdef Handle handle
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             handle = <Handle>(self._signal_handlers[sig])
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":429
 *         try:
 *             handle = <Handle>(self._signal_handlers[sig])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             handle = None
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("uvloop.loop.Loop._handle_signal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 429, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "uvloop/loop.pyx":430
 *             handle = <Handle>(self._signal_handlers[sig])
 *         except KeyError:
 *             handle = None             # <<<<<<<<<<<<<<
 * 
 *         if handle is None:
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_handle, ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":427
 *         cdef Handle handle
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             handle = <Handle>(self._signal_handlers[sig])
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":432
 *             handle = None
 * 
 *         if handle is None:             # <<<<<<<<<<<<<<
 *             self._ceval_process_signals()
 *             return
 */
  __pyx_t_8 = (((PyObject *)__pyx_v_handle) == Py_None);
  if (__pyx_t_8) {

    /* "uvloop/loop.pyx":433
 * 
 *         if handle is None:
 *             self._ceval_process_signals()             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_7 = __pyx_f_6uvloop_4loop_4Loop__ceval_process_signals(__pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":434
 *         if handle is None:
 *             self._ceval_process_signals()
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if handle._cancelled:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":432
 *             handle = None
 * 
 *         if handle is None:             # <<<<<<<<<<<<<<
 *             self._ceval_process_signals()
 *             return
 */
  }

  /* "uvloop/loop.pyx":436
 *             return
 * 
 *         if handle._cancelled:             # <<<<<<<<<<<<<<
 *             self.remove_signal_handler(sig)  # Remove it properly.
 *         else:
 */
  if (__pyx_v_handle->_cancelled) {

    /* "uvloop/loop.pyx":437
 * 
 *         if handle._cancelled:
 *             self.remove_signal_handler(sig)  # Remove it properly.             # <<<<<<<<<<<<<<
 *         else:
 *             self._append_ready_handle(handle)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove_signal_handler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_sig};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":436
 *             return
 * 
 *         if handle._cancelled:             # <<<<<<<<<<<<<<
 *             self.remove_signal_handler(sig)  # Remove it properly.
 *         else:
 */
    goto __pyx_L12;
  }

  /* "uvloop/loop.pyx":439
 *             self.remove_signal_handler(sig)  # Remove it properly.
 *         else:
 *             self._append_ready_handle(handle)             # <<<<<<<<<<<<<<
 *             self.handler_async.send()
 * 
 */
  /*else*/ {
    __pyx_t_7 = __pyx_f_6uvloop_4loop_4Loop__append_ready_handle(__pyx_v_self, __pyx_v_handle); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":440
 *         else:
 *             self._append_ready_handle(handle)
 *             self.handler_async.send()             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_wake(self):
 */
    __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->handler_async->__pyx_base.__pyx_vtab)->send(__pyx_v_self->handler_async); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L12:;

  /* "uvloop/loop.pyx":424
 *             self._handle_signal(signum)
 * 
 *     cdef _handle_signal(self, sig):             # <<<<<<<<<<<<<<
 *         cdef Handle handle
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.Loop._handle_signal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":442
 *             self.handler_async.send()
 * 
 *     cdef _on_wake(self):             # <<<<<<<<<<<<<<
 *         if ((self._ready_len > 0 or self._stopping) and
 *                 not self.handler_idle.running):
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__on_wake(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_wake", 1);

  /* "uvloop/loop.pyx":443
 * 
 *     cdef _on_wake(self):
 *         if ((self._ready_len > 0 or self._stopping) and             # <<<<<<<<<<<<<<
 *                 not self.handler_idle.running):
 *             self.handler_idle.start()
 */
  __pyx_t_2 = (__pyx_v_self->_ready_len > 0);
  if (!__pyx_t_2) {
  } else {
    goto __pyx_L5_next_and;
  }
  if (__pyx_v_self->_stopping) {
  } else {
    __pyx_t_1 = __pyx_v_self->_stopping;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_L5_next_and:;

  /* "uvloop/loop.pyx":444
 *     cdef _on_wake(self):
 *         if ((self._ready_len > 0 or self._stopping) and
 *                 not self.handler_idle.running):             # <<<<<<<<<<<<<<
 *             self.handler_idle.start()
 * 
 */
  __pyx_t_2 = (!__pyx_v_self->handler_idle->running);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "uvloop/loop.pyx":443
 * 
 *     cdef _on_wake(self):
 *         if ((self._ready_len > 0 or self._stopping) and             # <<<<<<<<<<<<<<
 *                 not self.handler_idle.running):
 *             self.handler_idle.start()
 */
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":445
 *         if ((self._ready_len > 0 or self._stopping) and
 *                 not self.handler_idle.running):
 *             self.handler_idle.start()             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_idle(self):
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_6UVIdle_start(__pyx_v_self->handler_idle); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":443
 * 
 *     cdef _on_wake(self):
 *         if ((self._ready_len > 0 or self._stopping) and             # <<<<<<<<<<<<<<
 *                 not self.handler_idle.running):
 *             self.handler_idle.start()
 */
  }

  /* "uvloop/loop.pyx":442
 *             self.handler_async.send()
 * 
 *     cdef _on_wake(self):             # <<<<<<<<<<<<<<
 *         if ((self._ready_len > 0 or self._stopping) and
 *                 not self.handler_idle.running):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Loop._on_wake", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":447
 *             self.handler_idle.start()
 * 
 *     cdef _on_idle(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, ntodo
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__on_idle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  CYTHON_UNUSED int __pyx_v_i;
  int __pyx_v_ntodo;
  PyObject *__pyx_v_popleft = 0;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handler = 0;
  PyObject *__pyx_v_started = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_delta = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_idle", 1);

  /* "uvloop/loop.pyx":450
 *         cdef:
 *             int i, ntodo
 *             object popleft = self._ready.popleft             # <<<<<<<<<<<<<<
 *             Handle handler
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_ready, __pyx_n_s_popleft); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_popleft = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":453
 *             Handle handler
 * 
 *         ntodo = len(self._ready)             # <<<<<<<<<<<<<<
 *         if self._debug:
 *             for i from 0 <= i < ntodo:
 */
  __pyx_t_1 = __pyx_v_self->_ready;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ntodo = __pyx_t_2;

  /* "uvloop/loop.pyx":454
 * 
 *         ntodo = len(self._ready)
 *         if self._debug:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()
 */
  if (__pyx_v_self->_debug) {

    /* "uvloop/loop.pyx":455
 *         ntodo = len(self._ready)
 *         if self._debug:
 *             for i from 0 <= i < ntodo:             # <<<<<<<<<<<<<<
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 */
    __pyx_t_3 = __pyx_v_ntodo;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

      /* "uvloop/loop.pyx":456
 *         if self._debug:
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()             # <<<<<<<<<<<<<<
 *                 if handler._cancelled == 0:
 *                     try:
 */
      __Pyx_INCREF(__pyx_v_popleft);
      __pyx_t_4 = __pyx_v_popleft; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_t_4 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_handler, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":457
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:             # <<<<<<<<<<<<<<
 *                     try:
 *                         started = time_monotonic()
 */
      __pyx_t_7 = (__pyx_v_handler->_cancelled == 0);
      if (__pyx_t_7) {

        /* "uvloop/loop.pyx":458
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 *                     try:             # <<<<<<<<<<<<<<
 *                         started = time_monotonic()
 *                         handler._run()
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          /*try:*/ {

            /* "uvloop/loop.pyx":459
 *                 if handler._cancelled == 0:
 *                     try:
 *                         started = time_monotonic()             # <<<<<<<<<<<<<<
 *                         handler._run()
 *                     except BaseException as ex:
 */
            __Pyx_INCREF(__pyx_v_6uvloop_4loop_time_monotonic);
            __pyx_t_1 = __pyx_v_6uvloop_4loop_time_monotonic; __pyx_t_5 = NULL;
            __pyx_t_6 = 0;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
                __pyx_t_6 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
              __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 459, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
            __Pyx_XDECREF_SET(__pyx_v_started, __pyx_t_4);
            __pyx_t_4 = 0;

            /* "uvloop/loop.pyx":460
 *                     try:
 *                         started = time_monotonic()
 *                         handler._run()             # <<<<<<<<<<<<<<
 *                     except BaseException as ex:
 *                         self._stop(ex)
 */
            __pyx_t_4 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_handler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 460, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "uvloop/loop.pyx":458
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 *                     try:             # <<<<<<<<<<<<<<
 *                         started = time_monotonic()
 *                         handler._run()
 */
          }

          /* "uvloop/loop.pyx":465
 *                         return
 *                     else:
 *                         delta = time_monotonic() - started             # <<<<<<<<<<<<<<
 *                         if delta > self.slow_callback_duration:
 *                             aio_logger.warning(
 */
          /*else:*/ {
            __Pyx_INCREF(__pyx_v_6uvloop_4loop_time_monotonic);
            __pyx_t_1 = __pyx_v_6uvloop_4loop_time_monotonic; __pyx_t_5 = NULL;
            __pyx_t_6 = 0;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
                __pyx_t_6 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
              __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L9_except_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
            __pyx_t_1 = PyNumber_Subtract(__pyx_t_4, __pyx_v_started); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L9_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF_SET(__pyx_v_delta, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "uvloop/loop.pyx":466
 *                     else:
 *                         delta = time_monotonic() - started
 *                         if delta > self.slow_callback_duration:             # <<<<<<<<<<<<<<
 *                             aio_logger.warning(
 *                                 'Executing %s took %.3f seconds',
 */
            __pyx_t_1 = PyObject_RichCompare(__pyx_v_delta, __pyx_v_self->slow_callback_duration, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L9_except_error)
            __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 466, __pyx_L9_except_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_7) {

              /* "uvloop/loop.pyx":467
 *                         delta = time_monotonic() - started
 *                         if delta > self.slow_callback_duration:
 *                             aio_logger.warning(             # <<<<<<<<<<<<<<
 *                                 'Executing %s took %.3f seconds',
 *                                 handler._format_handle(), delta)
 */
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 467, __pyx_L9_except_error)
              __Pyx_GOTREF(__pyx_t_4);

              /* "uvloop/loop.pyx":469
 *                             aio_logger.warning(
 *                                 'Executing %s took %.3f seconds',
 *                                 handler._format_handle(), delta)             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
              __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_handler->__pyx_vtab)->_format_handle(__pyx_v_handler); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 469, __pyx_L9_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_11 = NULL;
              __pyx_t_6 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_11)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_11);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                  __pyx_t_6 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[4] = {__pyx_t_11, __pyx_kp_u_Executing_s_took_3f_seconds, __pyx_t_5, __pyx_v_delta};
                __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L9_except_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "uvloop/loop.pyx":466
 *                     else:
 *                         delta = time_monotonic() - started
 *                         if delta > self.slow_callback_duration:             # <<<<<<<<<<<<<<
 *                             aio_logger.warning(
 *                                 'Executing %s took %.3f seconds',
 */
            }
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L14_try_end;
          __pyx_L7_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/loop.pyx":461
 *                         started = time_monotonic()
 *                         handler._run()
 *                     except BaseException as ex:             # <<<<<<<<<<<<<<
 *                         self._stop(ex)
 *                         return
 */
          __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
          if (__pyx_t_12) {
            __Pyx_AddTraceback("uvloop.loop.Loop._on_idle", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 461, __pyx_L9_except_error)
            __Pyx_XGOTREF(__pyx_t_1);
            __Pyx_XGOTREF(__pyx_t_4);
            __Pyx_XGOTREF(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __pyx_v_ex = __pyx_t_4;
            /*try:*/ {

              /* "uvloop/loop.pyx":462
 *                         handler._run()
 *                     except BaseException as ex:
 *                         self._stop(ex)             # <<<<<<<<<<<<<<
 *                         return
 *                     else:
 */
              __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_stop(__pyx_v_self, __pyx_v_ex); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 462, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "uvloop/loop.pyx":463
 *                     except BaseException as ex:
 *                         self._stop(ex)
 *                         return             # <<<<<<<<<<<<<<
 *                     else:
 *                         delta = time_monotonic() - started
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_r = Py_None; __Pyx_INCREF(Py_None);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              goto __pyx_L20_return;
            }

            /* "uvloop/loop.pyx":461
 *                         started = time_monotonic()
 *                         handler._run()
 *                     except BaseException as ex:             # <<<<<<<<<<<<<<
 *                         self._stop(ex)
 *                         return
 */
            /*finally:*/ {
              __pyx_L21_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
                if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_15);
                __Pyx_XGOTREF(__pyx_t_16);
                __Pyx_XGOTREF(__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                __pyx_t_12 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
                {
                  __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                }
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_18);
                  __Pyx_XGIVEREF(__pyx_t_19);
                  __Pyx_XGIVEREF(__pyx_t_20);
                  __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
                }
                __Pyx_XGIVEREF(__pyx_t_15);
                __Pyx_XGIVEREF(__pyx_t_16);
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
                __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
                __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
                goto __pyx_L9_except_error;
              }
              __pyx_L20_return: {
                __pyx_t_20 = __pyx_r;
                __pyx_r = 0;
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                __pyx_r = __pyx_t_20;
                __pyx_t_20 = 0;
                goto __pyx_L10_except_return;
              }
            }
          }
          goto __pyx_L9_except_error;

          /* "uvloop/loop.pyx":458
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 *                     try:             # <<<<<<<<<<<<<<
 *                         started = time_monotonic()
 *                         handler._run()
 */
          __pyx_L9_except_error:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L1_error;
          __pyx_L10_except_return:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L0;
          __pyx_L14_try_end:;
        }

        /* "uvloop/loop.pyx":457
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:             # <<<<<<<<<<<<<<
 *                     try:
 *                         started = time_monotonic()
 */
      }
    }

    /* "uvloop/loop.pyx":454
 * 
 *         ntodo = len(self._ready)
 *         if self._debug:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":472
 * 
 *         else:
 *             for i from 0 <= i < ntodo:             # <<<<<<<<<<<<<<
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 */
  /*else*/ {
    __pyx_t_3 = __pyx_v_ntodo;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

      /* "uvloop/loop.pyx":473
 *         else:
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()             # <<<<<<<<<<<<<<
 *                 if handler._cancelled == 0:
 *                     try:
 */
      __Pyx_INCREF(__pyx_v_popleft);
      __pyx_t_4 = __pyx_v_popleft; __pyx_t_1 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_t_4 = __pyx_t_5;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_handler, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":474
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:             # <<<<<<<<<<<<<<
 *                     try:
 *                         handler._run()
 */
      __pyx_t_7 = (__pyx_v_handler->_cancelled == 0);
      if (__pyx_t_7) {

        /* "uvloop/loop.pyx":475
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 *                     try:             # <<<<<<<<<<<<<<
 *                         handler._run()
 *                     except BaseException as ex:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_8);
          /*try:*/ {

            /* "uvloop/loop.pyx":476
 *                 if handler._cancelled == 0:
 *                     try:
 *                         handler._run()             # <<<<<<<<<<<<<<
 *                     except BaseException as ex:
 *                         self._stop(ex)
 */
            __pyx_t_4 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_handler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 476, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "uvloop/loop.pyx":475
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 *                     try:             # <<<<<<<<<<<<<<
 *                         handler._run()
 *                     except BaseException as ex:
 */
          }
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L37_try_end;
          __pyx_L30_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/loop.pyx":477
 *                     try:
 *                         handler._run()
 *                     except BaseException as ex:             # <<<<<<<<<<<<<<
 *                         self._stop(ex)
 *                         return
 */
          __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
          if (__pyx_t_13) {
            __Pyx_AddTraceback("uvloop.loop.Loop._on_idle", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(0, 477, __pyx_L32_except_error)
            __Pyx_XGOTREF(__pyx_t_4);
            __Pyx_XGOTREF(__pyx_t_5);
            __Pyx_XGOTREF(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_5);
            __pyx_v_ex = __pyx_t_5;
            /*try:*/ {

              /* "uvloop/loop.pyx":478
 *                         handler._run()
 *                     except BaseException as ex:
 *                         self._stop(ex)             # <<<<<<<<<<<<<<
 *                         return
 * 
 */
              __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_stop(__pyx_v_self, __pyx_v_ex); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 478, __pyx_L43_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "uvloop/loop.pyx":479
 *                     except BaseException as ex:
 *                         self._stop(ex)
 *                         return             # <<<<<<<<<<<<<<
 * 
 *         if len(self._queued_streams):
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_r = Py_None; __Pyx_INCREF(Py_None);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              goto __pyx_L42_return;
            }

            /* "uvloop/loop.pyx":477
 *                     try:
 *                         handler._run()
 *                     except BaseException as ex:             # <<<<<<<<<<<<<<
 *                         self._stop(ex)
 *                         return
 */
            /*finally:*/ {
              __pyx_L43_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
                if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_16);
                __Pyx_XGOTREF(__pyx_t_15);
                __pyx_t_13 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
                {
                  __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                }
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_17);
                  __Pyx_XGIVEREF(__pyx_t_16);
                  __Pyx_XGIVEREF(__pyx_t_15);
                  __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
                }
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_ErrRestore(__pyx_t_20, __pyx_t_19, __pyx_t_18);
                __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
                __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_21;
                goto __pyx_L32_except_error;
              }
              __pyx_L42_return: {
                __pyx_t_15 = __pyx_r;
                __pyx_r = 0;
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                __pyx_r = __pyx_t_15;
                __pyx_t_15 = 0;
                goto __pyx_L33_except_return;
              }
            }
          }
          goto __pyx_L32_except_error;

          /* "uvloop/loop.pyx":475
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:
 *                     try:             # <<<<<<<<<<<<<<
 *                         handler._run()
 *                     except BaseException as ex:
 */
          __pyx_L32_except_error:;
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
          goto __pyx_L1_error;
          __pyx_L33_except_return:;
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
          goto __pyx_L0;
          __pyx_L37_try_end:;
        }

        /* "uvloop/loop.pyx":474
 *             for i from 0 <= i < ntodo:
 *                 handler = <Handle> popleft()
 *                 if handler._cancelled == 0:             # <<<<<<<<<<<<<<
 *                     try:
 *                         handler._run()
 */
      }
    }
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":481
 *                         return
 * 
 *         if len(self._queued_streams):             # <<<<<<<<<<<<<<
 *             self._exec_queued_writes()
 * 
 */
  __pyx_t_1 = __pyx_v_self->_queued_streams;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 481, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PySet_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__pyx_t_2 != 0);
  if (__pyx_t_7) {

    /* "uvloop/loop.pyx":482
 * 
 *         if len(self._queued_streams):
 *             self._exec_queued_writes()             # <<<<<<<<<<<<<<
 * 
 *         self._ready_len = len(self._ready)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_exec_queued_writes(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":481
 *                         return
 * 
 *         if len(self._queued_streams):             # <<<<<<<<<<<<<<
 *             self._exec_queued_writes()
 * 
 */
  }

  /* "uvloop/loop.pyx":484
 *             self._exec_queued_writes()
 * 
 *         self._ready_len = len(self._ready)             # <<<<<<<<<<<<<<
 *         if self._ready_len == 0 and self.handler_idle.running:
 *             self.handler_idle.stop()
 */
  __pyx_t_1 = __pyx_v_self->_ready;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->_ready_len = __pyx_t_2;

  /* "uvloop/loop.pyx":485
 * 
 *         self._ready_len = len(self._ready)
 *         if self._ready_len == 0 and self.handler_idle.running:             # <<<<<<<<<<<<<<
 *             self.handler_idle.stop()
 * 
 */
  __pyx_t_22 = (__pyx_v_self->_ready_len == 0);
  if (__pyx_t_22) {
  } else {
    __pyx_t_7 = __pyx_t_22;
    goto __pyx_L51_bool_binop_done;
  }
  __pyx_t_7 = __pyx_v_self->handler_idle->running;
  __pyx_L51_bool_binop_done:;
  if (__pyx_t_7) {

    /* "uvloop/loop.pyx":486
 *         self._ready_len = len(self._ready)
 *         if self._ready_len == 0 and self.handler_idle.running:
 *             self.handler_idle.stop()             # <<<<<<<<<<<<<<
 * 
 *         if self._stopping:
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_6UVIdle_stop(__pyx_v_self->handler_idle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":485
 * 
 *         self._ready_len = len(self._ready)
 *         if self._ready_len == 0 and self.handler_idle.running:             # <<<<<<<<<<<<<<
 *             self.handler_idle.stop()
 * 
 */
  }

  /* "uvloop/loop.pyx":488
 *             self.handler_idle.stop()
 * 
 *         if self._stopping:             # <<<<<<<<<<<<<<
 *             uv.uv_stop(self.uvloop)  # void
 * 
 */
  if (__pyx_v_self->_stopping) {

    /* "uvloop/loop.pyx":489
 * 
 *         if self._stopping:
 *             uv.uv_stop(self.uvloop)  # void             # <<<<<<<<<<<<<<
 * 
 *     cdef _stop(self, exc):
 */
    uv_stop(__pyx_v_self->uvloop);

    /* "uvloop/loop.pyx":488
 *             self.handler_idle.stop()
 * 
 *         if self._stopping:             # <<<<<<<<<<<<<<
 *             uv.uv_stop(self.uvloop)  # void
 * 
 */
  }

  /* "uvloop/loop.pyx":447
 *             self.handler_idle.start()
 * 
 *     cdef _on_idle(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i, ntodo
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("uvloop.loop.Loop._on_idle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_popleft);
  __Pyx_XDECREF((PyObject *)__pyx_v_handler);
  __Pyx_XDECREF(__pyx_v_started);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_delta);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":491
 *             uv.uv_stop(self.uvloop)  # void
 * 
 *     cdef _stop(self, exc):             # <<<<<<<<<<<<<<
 *         if exc is not None:
 *             self._last_error = exc
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__stop(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_stop", 1);

  /* "uvloop/loop.pyx":492
 * 
 *     cdef _stop(self, exc):
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             self._last_error = exc
 *         if self._stopping == 1:
 */
  __pyx_t_1 = (__pyx_v_exc != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":493
 *     cdef _stop(self, exc):
 *         if exc is not None:
 *             self._last_error = exc             # <<<<<<<<<<<<<<
 *         if self._stopping == 1:
 *             return
 */
    __Pyx_INCREF(__pyx_v_exc);
    __Pyx_GIVEREF(__pyx_v_exc);
    __Pyx_GOTREF(__pyx_v_self->_last_error);
    __Pyx_DECREF(__pyx_v_self->_last_error);
    __pyx_v_self->_last_error = __pyx_v_exc;

    /* "uvloop/loop.pyx":492
 * 
 *     cdef _stop(self, exc):
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             self._last_error = exc
 *         if self._stopping == 1:
 */
  }

  /* "uvloop/loop.pyx":494
 *         if exc is not None:
 *             self._last_error = exc
 *         if self._stopping == 1:             # <<<<<<<<<<<<<<
 *             return
 *         self._stopping = 1
 */
  __pyx_t_1 = (__pyx_v_self->_stopping == 1);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":495
 *             self._last_error = exc
 *         if self._stopping == 1:
 *             return             # <<<<<<<<<<<<<<
 *         self._stopping = 1
 *         if not self.handler_idle.running:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":494
 *         if exc is not None:
 *             self._last_error = exc
 *         if self._stopping == 1:             # <<<<<<<<<<<<<<
 *             return
 *         self._stopping = 1
 */
  }

  /* "uvloop/loop.pyx":496
 *         if self._stopping == 1:
 *             return
 *         self._stopping = 1             # <<<<<<<<<<<<<<
 *         if not self.handler_idle.running:
 *             self.handler_idle.start()
 */
  __pyx_v_self->_stopping = 1;

  /* "uvloop/loop.pyx":497
 *             return
 *         self._stopping = 1
 *         if not self.handler_idle.running:             # <<<<<<<<<<<<<<
 *             self.handler_idle.start()
 * 
 */
  __pyx_t_1 = (!__pyx_v_self->handler_idle->running);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":498
 *         self._stopping = 1
 *         if not self.handler_idle.running:
 *             self.handler_idle.start()             # <<<<<<<<<<<<<<
 * 
 *     cdef __run(self, uv.uv_run_mode mode):
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVIdle_start(__pyx_v_self->handler_idle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":497
 *             return
 *         self._stopping = 1
 *         if not self.handler_idle.running:             # <<<<<<<<<<<<<<
 *             self.handler_idle.start()
 * 
 */
  }

  /* "uvloop/loop.pyx":491
 *             uv.uv_stop(self.uvloop)  # void
 * 
 *     cdef _stop(self, exc):             # <<<<<<<<<<<<<<
 *         if exc is not None:
 *             self._last_error = exc
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":500
 *             self.handler_idle.start()
 * 
 *     cdef __run(self, uv.uv_run_mode mode):             # <<<<<<<<<<<<<<
 *         # Although every UVHandle holds a reference to the loop,
 *         # we want to do everything to ensure that the loop will
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__Loop__run(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, uv_run_mode __pyx_v_mode) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Loop__run", 1);

  /* "uvloop/loop.pyx":505
 *         # never deallocate during the run -- so we do some
 *         # manual refs management.
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         with nogil:
 *             err = uv.uv_run(self.uvloop, mode)
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "uvloop/loop.pyx":506
 *         # manual refs management.
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             err = uv.uv_run(self.uvloop, mode)
 *         Py_DECREF(self)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      _save = NULL;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "uvloop/loop.pyx":507
 *         Py_INCREF(self)
 *         with nogil:
 *             err = uv.uv_run(self.uvloop, mode)             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
 */
        __pyx_v_err = uv_run(__pyx_v_self->uvloop, __pyx_v_mode);
      }

      /* "uvloop/loop.pyx":506
 *         # manual refs management.
 *         Py_INCREF(self)
 *         with nogil:             # <<<<<<<<<<<<<<
 *             err = uv.uv_run(self.uvloop, mode)
 *         Py_DECREF(self)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "uvloop/loop.pyx":508
 *         with nogil:
 *             err = uv.uv_run(self.uvloop, mode)
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *         if err < 0:
 */
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "uvloop/loop.pyx":510
 *         Py_DECREF(self)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":511
 * 
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *     cdef _run(self, uv.uv_run_mode mode):
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 511, __pyx_L1_error)

    /* "uvloop/loop.pyx":510
 *         Py_DECREF(self)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  }

  /* "uvloop/loop.pyx":500
 *             self.handler_idle.start()
 * 
 *     cdef __run(self, uv.uv_run_mode mode):             # <<<<<<<<<<<<<<
 *         # Although every UVHandle holds a reference to the loop,
 *         # we want to do everything to ensure that the loop will
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._Loop__run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":513
 *             raise convert_error(err)
 * 
 *     cdef _run(self, uv.uv_run_mode mode):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__run(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, uv_run_mode __pyx_v_mode) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  char const *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_run", 1);

  /* "uvloop/loop.pyx":516
 *         cdef int err
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('unable to start the loop; it was closed')
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_closed == 1);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":517
 * 
 *         if self._closed == 1:
 *             raise RuntimeError('unable to start the loop; it was closed')             # <<<<<<<<<<<<<<
 * 
 *         if self._running == 1:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 517, __pyx_L1_error)

    /* "uvloop/loop.pyx":516
 *         cdef int err
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('unable to start the loop; it was closed')
 * 
 */
  }

  /* "uvloop/loop.pyx":519
 *             raise RuntimeError('unable to start the loop; it was closed')
 * 
 *         if self._running == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('this event loop is already running.')
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_running == 1);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":520
 * 
 *         if self._running == 1:
 *             raise RuntimeError('this event loop is already running.')             # <<<<<<<<<<<<<<
 * 
 *         if (aio_get_running_loop is not None and
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 520, __pyx_L1_error)

    /* "uvloop/loop.pyx":519
 *             raise RuntimeError('unable to start the loop; it was closed')
 * 
 *         if self._running == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('this event loop is already running.')
 * 
 */
  }

  /* "uvloop/loop.pyx":522
 *             raise RuntimeError('this event loop is already running.')
 * 
 *         if (aio_get_running_loop is not None and             # <<<<<<<<<<<<<<
 *                 aio_get_running_loop() is not None):
 *             raise RuntimeError(
 */
  __pyx_t_3 = (__pyx_v_6uvloop_4loop_aio_get_running_loop != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }

  /* "uvloop/loop.pyx":523
 * 
 *         if (aio_get_running_loop is not None and
 *                 aio_get_running_loop() is not None):             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'Cannot run the event loop while another loop is running')
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_get_running_loop);
  __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_get_running_loop; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_3 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;

  /* "uvloop/loop.pyx":522
 *             raise RuntimeError('this event loop is already running.')
 * 
 *         if (aio_get_running_loop is not None and             # <<<<<<<<<<<<<<
 *                 aio_get_running_loop() is not None):
 *             raise RuntimeError(
 */
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":524
 *         if (aio_get_running_loop is not None and
 *                 aio_get_running_loop() is not None):
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'Cannot run the event loop while another loop is running')
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 524, __pyx_L1_error)

    /* "uvloop/loop.pyx":522
 *             raise RuntimeError('this event loop is already running.')
 * 
 *         if (aio_get_running_loop is not None and             # <<<<<<<<<<<<<<
 *                 aio_get_running_loop() is not None):
 *             raise RuntimeError(
 */
  }

  /* "uvloop/loop.pyx":528
 * 
 *         # reset _last_error
 *         self._last_error = None             # <<<<<<<<<<<<<<
 * 
 *         self._thread_id = PyThread_get_thread_ident()
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_error);
  __Pyx_DECREF(__pyx_v_self->_last_error);
  __pyx_v_self->_last_error = Py_None;

  /* "uvloop/loop.pyx":530
 *         self._last_error = None
 * 
 *         self._thread_id = PyThread_get_thread_ident()             # <<<<<<<<<<<<<<
 *         self._running = 1
 * 
 */
  __pyx_v_self->_thread_id = PyThread_get_thread_ident();

  /* "uvloop/loop.pyx":531
 * 
 *         self._thread_id = PyThread_get_thread_ident()
 *         self._running = 1             # <<<<<<<<<<<<<<
 * 
 *         self.handler_check__exec_writes.start()
 */
  __pyx_v_self->_running = 1;

  /* "uvloop/loop.pyx":533
 *         self._running = 1
 * 
 *         self.handler_check__exec_writes.start()             # <<<<<<<<<<<<<<
 *         self.handler_idle.start()
 * 
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_7UVCheck_start(__pyx_v_self->handler_check__exec_writes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":534
 * 
 *         self.handler_check__exec_writes.start()
 *         self.handler_idle.start()             # <<<<<<<<<<<<<<
 * 
 *         self._setup_or_resume_signals()
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVIdle_start(__pyx_v_self->handler_idle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":536
 *         self.handler_idle.start()
 * 
 *         self._setup_or_resume_signals()             # <<<<<<<<<<<<<<
 * 
 *         if aio_set_running_loop is not None:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_setup_or_resume_signals(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":538
 *         self._setup_or_resume_signals()
 * 
 *         if aio_set_running_loop is not None:             # <<<<<<<<<<<<<<
 *             aio_set_running_loop(self)
 *         try:
 */
  __pyx_t_1 = (__pyx_v_6uvloop_4loop_aio_set_running_loop != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":539
 * 
 *         if aio_set_running_loop is not None:
 *             aio_set_running_loop(self)             # <<<<<<<<<<<<<<
 *         try:
 *             self.__run(mode)
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_set_running_loop);
    __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_set_running_loop; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":538
 *         self._setup_or_resume_signals()
 * 
 *         if aio_set_running_loop is not None:             # <<<<<<<<<<<<<<
 *             aio_set_running_loop(self)
 *         try:
 */
  }

  /* "uvloop/loop.pyx":540
 *         if aio_set_running_loop is not None:
 *             aio_set_running_loop(self)
 *         try:             # <<<<<<<<<<<<<<
 *             self.__run(mode)
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":541
 *             aio_set_running_loop(self)
 *         try:
 *             self.__run(mode)             # <<<<<<<<<<<<<<
 *         finally:
 *             if aio_set_running_loop is not None:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_Loop__run(__pyx_v_self, __pyx_v_mode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L10_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/loop.pyx":543
 *             self.__run(mode)
 *         finally:
 *             if aio_set_running_loop is not None:             # <<<<<<<<<<<<<<
 *                 aio_set_running_loop(None)
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = (__pyx_v_6uvloop_4loop_aio_set_running_loop != Py_None);
      if (__pyx_t_1) {

        /* "uvloop/loop.pyx":544
 *         finally:
 *             if aio_set_running_loop is not None:
 *                 aio_set_running_loop(None)             # <<<<<<<<<<<<<<
 * 
 *             self.handler_check__exec_writes.stop()
 */
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_set_running_loop);
        __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_set_running_loop; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_None};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":543
 *             self.__run(mode)
 *         finally:
 *             if aio_set_running_loop is not None:             # <<<<<<<<<<<<<<
 *                 aio_set_running_loop(None)
 * 
 */
      }

      /* "uvloop/loop.pyx":546
 *                 aio_set_running_loop(None)
 * 
 *             self.handler_check__exec_writes.stop()             # <<<<<<<<<<<<<<
 *             self.handler_idle.stop()
 * 
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop_7UVCheck_stop(__pyx_v_self->handler_check__exec_writes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 546, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":547
 * 
 *             self.handler_check__exec_writes.stop()
 *             self.handler_idle.stop()             # <<<<<<<<<<<<<<
 * 
 *             self._pause_signals()
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVIdle_stop(__pyx_v_self->handler_idle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 547, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":549
 *             self.handler_idle.stop()
 * 
 *             self._pause_signals()             # <<<<<<<<<<<<<<
 * 
 *             self._thread_id = 0
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_pause_signals(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":551
 *             self._pause_signals()
 * 
 *             self._thread_id = 0             # <<<<<<<<<<<<<<
 *             self._running = 0
 *             self._stopping = 0
 */
      __pyx_v_self->_thread_id = 0;

      /* "uvloop/loop.pyx":552
 * 
 *             self._thread_id = 0
 *             self._running = 0             # <<<<<<<<<<<<<<
 *             self._stopping = 0
 * 
 */
      __pyx_v_self->_running = 0;

      /* "uvloop/loop.pyx":553
 *             self._thread_id = 0
 *             self._running = 0
 *             self._stopping = 0             # <<<<<<<<<<<<<<
 * 
 *         if self._last_error is not None:
 */
      __pyx_v_self->_stopping = 0;
      goto __pyx_L11;
    }
    __pyx_L10_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_7 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_9 = __pyx_filename;
      {

        /* "uvloop/loop.pyx":543
 *             self.__run(mode)
 *         finally:
 *             if aio_set_running_loop is not None:             # <<<<<<<<<<<<<<
 *                 aio_set_running_loop(None)
 * 
 */
        __pyx_t_1 = (__pyx_v_6uvloop_4loop_aio_set_running_loop != Py_None);
        if (__pyx_t_1) {

          /* "uvloop/loop.pyx":544
 *         finally:
 *             if aio_set_running_loop is not None:
 *                 aio_set_running_loop(None)             # <<<<<<<<<<<<<<
 * 
 *             self.handler_check__exec_writes.stop()
 */
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_set_running_loop);
          __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_set_running_loop; __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_None};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 544, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "uvloop/loop.pyx":543
 *             self.__run(mode)
 *         finally:
 *             if aio_set_running_loop is not None:             # <<<<<<<<<<<<<<
 *                 aio_set_running_loop(None)
 * 
 */
        }

        /* "uvloop/loop.pyx":546
 *                 aio_set_running_loop(None)
 * 
 *             self.handler_check__exec_writes.stop()             # <<<<<<<<<<<<<<
 *             self.handler_idle.stop()
 * 
 */
        __pyx_t_2 = __pyx_f_6uvloop_4loop_7UVCheck_stop(__pyx_v_self->handler_check__exec_writes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 546, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":547
 * 
 *             self.handler_check__exec_writes.stop()
 *             self.handler_idle.stop()             # <<<<<<<<<<<<<<
 * 
 *             self._pause_signals()
 */
        __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVIdle_stop(__pyx_v_self->handler_idle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 547, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":549
 *             self.handler_idle.stop()
 * 
 *             self._pause_signals()             # <<<<<<<<<<<<<<
 * 
 *             self._thread_id = 0
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_pause_signals(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":551
 *             self._pause_signals()
 * 
 *             self._thread_id = 0             # <<<<<<<<<<<<<<
 *             self._running = 0
 *             self._stopping = 0
 */
        __pyx_v_self->_thread_id = 0;

        /* "uvloop/loop.pyx":552
 * 
 *             self._thread_id = 0
 *             self._running = 0             # <<<<<<<<<<<<<<
 *             self._stopping = 0
 * 
 */
        __pyx_v_self->_running = 0;

        /* "uvloop/loop.pyx":553
 *             self._thread_id = 0
 *             self._running = 0
 *             self._stopping = 0             # <<<<<<<<<<<<<<
 * 
 *         if self._last_error is not None:
 */
        __pyx_v_self->_stopping = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_9;
      goto __pyx_L1_error;
      __pyx_L14_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L11:;
  }

  /* "uvloop/loop.pyx":555
 *             self._stopping = 0
 * 
 *         if self._last_error is not None:             # <<<<<<<<<<<<<<
 *             # The loop was stopped with an error with 'loop._stop(error)' call
 *             raise self._last_error
 */
  __pyx_t_1 = (__pyx_v_self->_last_error != Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":557
 *         if self._last_error is not None:
 *             # The loop was stopped with an error with 'loop._stop(error)' call
 *             raise self._last_error             # <<<<<<<<<<<<<<
 * 
 *     cdef _close(self):
 */
    __Pyx_Raise(__pyx_v_self->_last_error, 0, 0, 0);
    __PYX_ERR(0, 557, __pyx_L1_error)

    /* "uvloop/loop.pyx":555
 *             self._stopping = 0
 * 
 *         if self._last_error is not None:             # <<<<<<<<<<<<<<
 *             # The loop was stopped with an error with 'loop._stop(error)' call
 *             raise self._last_error
 */
  }

  /* "uvloop/loop.pyx":513
 *             raise convert_error(err)
 * 
 *     cdef _run(self, uv.uv_run_mode mode):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.Loop._run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":559
 *             raise self._last_error
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__close(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_cb_handle = NULL;
  PyObject *__pyx_v_poll_handle = NULL;
  PyObject *__pyx_v_timer_cbhandle = NULL;
  PyObject *__pyx_v_fileobj = NULL;
  PyObject *__pyx_v_executor = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/loop.pyx":562
 *         cdef int err
 * 
 *         if self._running == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Cannot close a running event loop")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_running == 1);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":563
 * 
 *         if self._running == 1:
 *             raise RuntimeError("Cannot close a running event loop")             # <<<<<<<<<<<<<<
 * 
 *         if self._closed == 1:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 563, __pyx_L1_error)

    /* "uvloop/loop.pyx":562
 *         cdef int err
 * 
 *         if self._running == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Cannot close a running event loop")
 * 
 */
  }

  /* "uvloop/loop.pyx":565
 *             raise RuntimeError("Cannot close a running event loop")
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_closed == 1);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":566
 * 
 *         if self._closed == 1:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._closed = 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":565
 *             raise RuntimeError("Cannot close a running event loop")
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/loop.pyx":568
 *             return
 * 
 *         self._closed = 1             # <<<<<<<<<<<<<<
 * 
 *         for cb_handle in self._ready:
 */
  __pyx_v_self->_closed = 1;

  /* "uvloop/loop.pyx":570
 *         self._closed = 1
 * 
 *         for cb_handle in self._ready:             # <<<<<<<<<<<<<<
 *             cb_handle.cancel()
 *         self._ready.clear()
 */
  if (likely(PyList_CheckExact(__pyx_v_self->_ready)) || PyTuple_CheckExact(__pyx_v_self->_ready)) {
    __pyx_t_2 = __pyx_v_self->_ready; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->_ready); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 570, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 570, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 570, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 570, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 570, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 570, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_cb_handle, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":571
 * 
 *         for cb_handle in self._ready:
 *             cb_handle.cancel()             # <<<<<<<<<<<<<<
 *         self._ready.clear()
 *         self._ready_len = 0
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_cb_handle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":570
 *         self._closed = 1
 * 
 *         for cb_handle in self._ready:             # <<<<<<<<<<<<<<
 *             cb_handle.cancel()
 *         self._ready.clear()
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":572
 *         for cb_handle in self._ready:
 *             cb_handle.cancel()
 *         self._ready.clear()             # <<<<<<<<<<<<<<
 *         self._ready_len = 0
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_ready, __pyx_n_s_clear); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":573
 *             cb_handle.cancel()
 *         self._ready.clear()
 *         self._ready_len = 0             # <<<<<<<<<<<<<<
 * 
 *         if self._polls:
 */
  __pyx_v_self->_ready_len = 0;

  /* "uvloop/loop.pyx":575
 *         self._ready_len = 0
 * 
 *         if self._polls:             # <<<<<<<<<<<<<<
 *             for poll_handle in self._polls.values():
 *                 (<UVHandle>poll_handle)._close()
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_polls); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 575, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":576
 * 
 *         if self._polls:
 *             for poll_handle in self._polls.values():             # <<<<<<<<<<<<<<
 *                 (<UVHandle>poll_handle)._close()
 * 
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_self->_polls == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
      __PYX_ERR(0, 576, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_self->_polls, 1, __pyx_n_s_values, (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;
    while (1) {
      __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_9, &__pyx_t_3, NULL, &__pyx_t_5, NULL, __pyx_t_10);
      if (unlikely(__pyx_t_11 == 0)) break;
      if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_poll_handle, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":577
 *         if self._polls:
 *             for poll_handle in self._polls.values():
 *                 (<UVHandle>poll_handle)._close()             # <<<<<<<<<<<<<<
 * 
 *             self._polls.clear()
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_poll_handle)->__pyx_vtab)->_close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_poll_handle)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":579
 *                 (<UVHandle>poll_handle)._close()
 * 
 *             self._polls.clear()             # <<<<<<<<<<<<<<
 * 
 *         if self._timers:
 */
    if (unlikely(__pyx_v_self->_polls == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
      __PYX_ERR(0, 579, __pyx_L1_error)
    }
    __pyx_t_12 = __Pyx_PyDict_Clear(__pyx_v_self->_polls); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 579, __pyx_L1_error)

    /* "uvloop/loop.pyx":575
 *         self._ready_len = 0
 * 
 *         if self._polls:             # <<<<<<<<<<<<<<
 *             for poll_handle in self._polls.values():
 *                 (<UVHandle>poll_handle)._close()
 */
  }

  /* "uvloop/loop.pyx":581
 *             self._polls.clear()
 * 
 *         if self._timers:             # <<<<<<<<<<<<<<
 *             for timer_cbhandle in tuple(self._timers):
 *                 timer_cbhandle.cancel()
 */
  __pyx_t_1 = (__pyx_v_self->_timers != Py_None)&&(PySet_GET_SIZE(__pyx_v_self->_timers) != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":582
 * 
 *         if self._timers:
 *             for timer_cbhandle in tuple(self._timers):             # <<<<<<<<<<<<<<
 *                 timer_cbhandle.cancel()
 * 
 */
    __pyx_t_2 = PySequence_Tuple(__pyx_v_self->_timers); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 582, __pyx_L1_error)
        #endif
        if (__pyx_t_9 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 582, __pyx_L1_error)
      #else
      __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_timer_cbhandle, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":583
 *         if self._timers:
 *             for timer_cbhandle in tuple(self._timers):
 *                 timer_cbhandle.cancel()             # <<<<<<<<<<<<<<
 * 
 *         # Close all remaining handles
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_timer_cbhandle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":582
 * 
 *         if self._timers:
 *             for timer_cbhandle in tuple(self._timers):             # <<<<<<<<<<<<<<
 *                 timer_cbhandle.cancel()
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":581
 *             self._polls.clear()
 * 
 *         if self._timers:             # <<<<<<<<<<<<<<
 *             for timer_cbhandle in tuple(self._timers):
 *                 timer_cbhandle.cancel()
 */
  }

  /* "uvloop/loop.pyx":586
 * 
 *         # Close all remaining handles
 *         self.handler_async._close()             # <<<<<<<<<<<<<<
 *         self.handler_idle._close()
 *         self.handler_check__exec_writes._close()
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->handler_async->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self->handler_async)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":587
 *         # Close all remaining handles
 *         self.handler_async._close()
 *         self.handler_idle._close()             # <<<<<<<<<<<<<<
 *         self.handler_check__exec_writes._close()
 *         __close_all_handles(self)
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->handler_idle->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self->handler_idle)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":588
 *         self.handler_async._close()
 *         self.handler_idle._close()
 *         self.handler_check__exec_writes._close()             # <<<<<<<<<<<<<<
 *         __close_all_handles(self)
 *         self._shutdown_signals()
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->handler_check__exec_writes->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self->handler_check__exec_writes)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":589
 *         self.handler_idle._close()
 *         self.handler_check__exec_writes._close()
 *         __close_all_handles(self)             # <<<<<<<<<<<<<<
 *         self._shutdown_signals()
 *         # During this run there should be no open handles,
 */
  __pyx_f_6uvloop_4loop___close_all_handles(__pyx_v_self);

  /* "uvloop/loop.pyx":590
 *         self.handler_check__exec_writes._close()
 *         __close_all_handles(self)
 *         self._shutdown_signals()             # <<<<<<<<<<<<<<
 *         # During this run there should be no open handles,
 *         # so it should finish right away
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_shutdown_signals(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":593
 *         # During this run there should be no open handles,
 *         # so it should finish right away
 *         self.__run(uv.UV_RUN_DEFAULT)             # <<<<<<<<<<<<<<
 * 
 *         if self._fd_to_writer_fileobj:
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_Loop__run(__pyx_v_self, UV_RUN_DEFAULT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":595
 *         self.__run(uv.UV_RUN_DEFAULT)
 * 
 *         if self._fd_to_writer_fileobj:             # <<<<<<<<<<<<<<
 *             for fileobj in self._fd_to_writer_fileobj.values():
 *                 socket_dec_io_ref(fileobj)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_fd_to_writer_fileobj); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 595, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":596
 * 
 *         if self._fd_to_writer_fileobj:
 *             for fileobj in self._fd_to_writer_fileobj.values():             # <<<<<<<<<<<<<<
 *                 socket_dec_io_ref(fileobj)
 *             self._fd_to_writer_fileobj.clear()
 */
    __pyx_t_9 = 0;
    if (unlikely(__pyx_v_self->_fd_to_writer_fileobj == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
      __PYX_ERR(0, 596, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_dict_iterator(__pyx_v_self->_fd_to_writer_fileobj, 1, __pyx_n_s_values, (&__pyx_t_3), (&__pyx_t_10)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_2;
    __pyx_t_2 = 0;
    while (1) {
      __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_5, __pyx_t_3, &__pyx_t_9, NULL, &__pyx_t_2, NULL, __pyx_t_10);
      if (unlikely(__pyx_t_11 == 0)) break;
      if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 596, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_fileobj, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":597
 *         if self._fd_to_writer_fileobj:
 *             for fileobj in self._fd_to_writer_fileobj.values():
 *                 socket_dec_io_ref(fileobj)             # <<<<<<<<<<<<<<
 *             self._fd_to_writer_fileobj.clear()
 * 
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_v_fileobj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":598
 *             for fileobj in self._fd_to_writer_fileobj.values():
 *                 socket_dec_io_ref(fileobj)
 *             self._fd_to_writer_fileobj.clear()             # <<<<<<<<<<<<<<
 * 
 *         if self._fd_to_reader_fileobj:
 */
    if (unlikely(__pyx_v_self->_fd_to_writer_fileobj == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
      __PYX_ERR(0, 598, __pyx_L1_error)
    }
    __pyx_t_12 = __Pyx_PyDict_Clear(__pyx_v_self->_fd_to_writer_fileobj); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 598, __pyx_L1_error)

    /* "uvloop/loop.pyx":595
 *         self.__run(uv.UV_RUN_DEFAULT)
 * 
 *         if self._fd_to_writer_fileobj:             # <<<<<<<<<<<<<<
 *             for fileobj in self._fd_to_writer_fileobj.values():
 *                 socket_dec_io_ref(fileobj)
 */
  }

  /* "uvloop/loop.pyx":600
 *             self._fd_to_writer_fileobj.clear()
 * 
 *         if self._fd_to_reader_fileobj:             # <<<<<<<<<<<<<<
 *             for fileobj in self._fd_to_reader_fileobj.values():
 *                 socket_dec_io_ref(fileobj)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_fd_to_reader_fileobj); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 600, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":601
 * 
 *         if self._fd_to_reader_fileobj:
 *             for fileobj in self._fd_to_reader_fileobj.values():             # <<<<<<<<<<<<<<
 *                 socket_dec_io_ref(fileobj)
 *             self._fd_to_reader_fileobj.clear()
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_self->_fd_to_reader_fileobj == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
      __PYX_ERR(0, 601, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_dict_iterator(__pyx_v_self->_fd_to_reader_fileobj, 1, __pyx_n_s_values, (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_2;
    __pyx_t_2 = 0;
    while (1) {
      __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_5, __pyx_t_9, &__pyx_t_3, NULL, &__pyx_t_2, NULL, __pyx_t_10);
      if (unlikely(__pyx_t_11 == 0)) break;
      if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_fileobj, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":602
 *         if self._fd_to_reader_fileobj:
 *             for fileobj in self._fd_to_reader_fileobj.values():
 *                 socket_dec_io_ref(fileobj)             # <<<<<<<<<<<<<<
 *             self._fd_to_reader_fileobj.clear()
 * 
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_v_fileobj); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":603
 *             for fileobj in self._fd_to_reader_fileobj.values():
 *                 socket_dec_io_ref(fileobj)
 *             self._fd_to_reader_fileobj.clear()             # <<<<<<<<<<<<<<
 * 
 *         if self._timers:
 */
    if (unlikely(__pyx_v_self->_fd_to_reader_fileobj == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
      __PYX_ERR(0, 603, __pyx_L1_error)
    }
    __pyx_t_12 = __Pyx_PyDict_Clear(__pyx_v_self->_fd_to_reader_fileobj); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 603, __pyx_L1_error)

    /* "uvloop/loop.pyx":600
 *             self._fd_to_writer_fileobj.clear()
 * 
 *         if self._fd_to_reader_fileobj:             # <<<<<<<<<<<<<<
 *             for fileobj in self._fd_to_reader_fileobj.values():
 *                 socket_dec_io_ref(fileobj)
 */
  }

  /* "uvloop/loop.pyx":605
 *             self._fd_to_reader_fileobj.clear()
 * 
 *         if self._timers:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f"new timers were queued during loop closing: {self._timers}")
 */
  __pyx_t_1 = (__pyx_v_self->_timers != Py_None)&&(PySet_GET_SIZE(__pyx_v_self->_timers) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":607
 *         if self._timers:
 *             raise RuntimeError(
 *                 f"new timers were queued during loop closing: {self._timers}")             # <<<<<<<<<<<<<<
 * 
 *         if self._polls:
 */
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_self->_timers, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_new_timers_were_queued_during_lo, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":606
 * 
 *         if self._timers:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 f"new timers were queued during loop closing: {self._timers}")
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 606, __pyx_L1_error)

    /* "uvloop/loop.pyx":605
 *             self._fd_to_reader_fileobj.clear()
 * 
 *         if self._timers:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f"new timers were queued during loop closing: {self._timers}")
 */
  }

  /* "uvloop/loop.pyx":609
 *                 f"new timers were queued during loop closing: {self._timers}")
 * 
 *         if self._polls:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f"new poll handles were queued during loop closing: "
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_polls); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 609, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":612
 *             raise RuntimeError(
 *                 f"new poll handles were queued during loop closing: "
 *                 f"{self._polls}")             # <<<<<<<<<<<<<<
 * 
 *         if self._ready:
 */
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_self->_polls, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "uvloop/loop.pyx":611
 *         if self._polls:
 *             raise RuntimeError(
 *                 f"new poll handles were queued during loop closing: "             # <<<<<<<<<<<<<<
 *                 f"{self._polls}")
 * 
 */
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_new_poll_handles_were_queued_dur, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":610
 * 
 *         if self._polls:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 f"new poll handles were queued during loop closing: "
 *                 f"{self._polls}")
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 610, __pyx_L1_error)

    /* "uvloop/loop.pyx":609
 *                 f"new timers were queued during loop closing: {self._timers}")
 * 
 *         if self._polls:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f"new poll handles were queued during loop closing: "
 */
  }

  /* "uvloop/loop.pyx":614
 *                 f"{self._polls}")
 * 
 *         if self._ready:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f"new callbacks were queued during loop closing: "
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_ready); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 614, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":617
 *             raise RuntimeError(
 *                 f"new callbacks were queued during loop closing: "
 *                 f"{self._ready}")             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_loop_close(self.uvloop)
 */
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_self->_ready, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "uvloop/loop.pyx":616
 *         if self._ready:
 *             raise RuntimeError(
 *                 f"new callbacks were queued during loop closing: "             # <<<<<<<<<<<<<<
 *                 f"{self._ready}")
 * 
 */
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_new_callbacks_were_queued_during, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":615
 * 
 *         if self._ready:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 f"new callbacks were queued during loop closing: "
 *                 f"{self._ready}")
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 615, __pyx_L1_error)

    /* "uvloop/loop.pyx":614
 *                 f"{self._polls}")
 * 
 *         if self._ready:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f"new callbacks were queued during loop closing: "
 */
  }

  /* "uvloop/loop.pyx":619
 *                 f"{self._ready}")
 * 
 *         err = uv.uv_loop_close(self.uvloop)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
  __pyx_v_err = uv_loop_close(__pyx_v_self->uvloop);

  /* "uvloop/loop.pyx":620
 * 
 *         err = uv.uv_loop_close(self.uvloop)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":621
 *         err = uv.uv_loop_close(self.uvloop)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         self.handler_async = None
 */
    __pyx_t_5 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 621, __pyx_L1_error)

    /* "uvloop/loop.pyx":620
 * 
 *         err = uv.uv_loop_close(self.uvloop)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  }

  /* "uvloop/loop.pyx":623
 *             raise convert_error(err)
 * 
 *         self.handler_async = None             # <<<<<<<<<<<<<<
 *         self.handler_idle = None
 *         self.handler_check__exec_writes = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->handler_async);
  __Pyx_DECREF((PyObject *)__pyx_v_self->handler_async);
  __pyx_v_self->handler_async = ((struct __pyx_obj_6uvloop_4loop_UVAsync *)Py_None);

  /* "uvloop/loop.pyx":624
 * 
 *         self.handler_async = None
 *         self.handler_idle = None             # <<<<<<<<<<<<<<
 *         self.handler_check__exec_writes = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->handler_idle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->handler_idle);
  __pyx_v_self->handler_idle = ((struct __pyx_obj_6uvloop_4loop_UVIdle *)Py_None);

  /* "uvloop/loop.pyx":625
 *         self.handler_async = None
 *         self.handler_idle = None
 *         self.handler_check__exec_writes = None             # <<<<<<<<<<<<<<
 * 
 *         self._executor_shutdown_called = True
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->handler_check__exec_writes);
  __Pyx_DECREF((PyObject *)__pyx_v_self->handler_check__exec_writes);
  __pyx_v_self->handler_check__exec_writes = ((struct __pyx_obj_6uvloop_4loop_UVCheck *)Py_None);

  /* "uvloop/loop.pyx":627
 *         self.handler_check__exec_writes = None
 * 
 *         self._executor_shutdown_called = True             # <<<<<<<<<<<<<<
 *         executor = self._default_executor
 *         if executor is not None:
 */
  __pyx_v_self->_executor_shutdown_called = 1;

  /* "uvloop/loop.pyx":628
 * 
 *         self._executor_shutdown_called = True
 *         executor = self._default_executor             # <<<<<<<<<<<<<<
 *         if executor is not None:
 *             self._default_executor = None
 */
  __pyx_t_5 = __pyx_v_self->_default_executor;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_executor = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":629
 *         self._executor_shutdown_called = True
 *         executor = self._default_executor
 *         if executor is not None:             # <<<<<<<<<<<<<<
 *             self._default_executor = None
 *             executor.shutdown(wait=False)
 */
  __pyx_t_1 = (__pyx_v_executor != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":630
 *         executor = self._default_executor
 *         if executor is not None:
 *             self._default_executor = None             # <<<<<<<<<<<<<<
 *             executor.shutdown(wait=False)
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_default_executor);
    __Pyx_DECREF(__pyx_v_self->_default_executor);
    __pyx_v_self->_default_executor = Py_None;

    /* "uvloop/loop.pyx":631
 *         if executor is not None:
 *             self._default_executor = None
 *             executor.shutdown(wait=False)             # <<<<<<<<<<<<<<
 * 
 *     cdef uint64_t _time(self):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_executor, __pyx_n_s_shutdown); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_wait, Py_False) < 0) __PYX_ERR(0, 631, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":629
 *         self._executor_shutdown_called = True
 *         executor = self._default_executor
 *         if executor is not None:             # <<<<<<<<<<<<<<
 *             self._default_executor = None
 *             executor.shutdown(wait=False)
 */
  }

  /* "uvloop/loop.pyx":559
 *             raise self._last_error
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.Loop._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cb_handle);
  __Pyx_XDECREF(__pyx_v_poll_handle);
  __Pyx_XDECREF(__pyx_v_timer_cbhandle);
  __Pyx_XDECREF(__pyx_v_fileobj);
  __Pyx_XDECREF(__pyx_v_executor);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":633
 *             executor.shutdown(wait=False)
 * 
 *     cdef uint64_t _time(self):             # <<<<<<<<<<<<<<
 *         # asyncio doesn't have a time cache, neither should uvloop.
 *         uv.uv_update_time(self.uvloop)  # void
 */

static uint64_t __pyx_f_6uvloop_4loop_4Loop__time(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  uint64_t __pyx_r;

  /* "uvloop/loop.pyx":635
 *     cdef uint64_t _time(self):
 *         # asyncio doesn't have a time cache, neither should uvloop.
 *         uv.uv_update_time(self.uvloop)  # void             # <<<<<<<<<<<<<<
 *         return uv.uv_now(self.uvloop)
 * 
 */
  uv_update_time(__pyx_v_self->uvloop);

  /* "uvloop/loop.pyx":636
 *         # asyncio doesn't have a time cache, neither should uvloop.
 *         uv.uv_update_time(self.uvloop)  # void
 *         return uv.uv_now(self.uvloop)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _queue_write(self, UVStream stream):
 */
  __pyx_r = uv_now(__pyx_v_self->uvloop);
  goto __pyx_L0;

  /* "uvloop/loop.pyx":633
 *             executor.shutdown(wait=False)
 * 
 *     cdef uint64_t _time(self):             # <<<<<<<<<<<<<<
 *         # asyncio doesn't have a time cache, neither should uvloop.
 *         uv.uv_update_time(self.uvloop)  # void
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/loop.pyx":638
 *         return uv.uv_now(self.uvloop)
 * 
 *     cdef inline _queue_write(self, UVStream stream):             # <<<<<<<<<<<<<<
 *         self._queued_streams.add(stream)
 *         if not self.handler_check__exec_writes.running:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__queue_write(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_queue_write", 1);

  /* "uvloop/loop.pyx":639
 * 
 *     cdef inline _queue_write(self, UVStream stream):
 *         self._queued_streams.add(stream)             # <<<<<<<<<<<<<<
 *         if not self.handler_check__exec_writes.running:
 *             self.handler_check__exec_writes.start()
 */
  if (unlikely(__pyx_v_self->_queued_streams == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(0, 639, __pyx_L1_error)
  }
  __pyx_t_1 = PySet_Add(__pyx_v_self->_queued_streams, ((PyObject *)__pyx_v_stream)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 639, __pyx_L1_error)

  /* "uvloop/loop.pyx":640
 *     cdef inline _queue_write(self, UVStream stream):
 *         self._queued_streams.add(stream)
 *         if not self.handler_check__exec_writes.running:             # <<<<<<<<<<<<<<
 *             self.handler_check__exec_writes.start()
 * 
 */
  __pyx_t_2 = (!__pyx_v_self->handler_check__exec_writes->running);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":641
 *         self._queued_streams.add(stream)
 *         if not self.handler_check__exec_writes.running:
 *             self.handler_check__exec_writes.start()             # <<<<<<<<<<<<<<
 * 
 *     cdef _exec_queued_writes(self):
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_7UVCheck_start(__pyx_v_self->handler_check__exec_writes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":640
 *     cdef inline _queue_write(self, UVStream stream):
 *         self._queued_streams.add(stream)
 *         if not self.handler_check__exec_writes.running:             # <<<<<<<<<<<<<<
 *             self.handler_check__exec_writes.start()
 * 
 */
  }

  /* "uvloop/loop.pyx":638
 *         return uv.uv_now(self.uvloop)
 * 
 *     cdef inline _queue_write(self, UVStream stream):             # <<<<<<<<<<<<<<
 *         self._queued_streams.add(stream)
 *         if not self.handler_check__exec_writes.running:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Loop._queue_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":643
 *             self.handler_check__exec_writes.start()
 * 
 *     cdef _exec_queued_writes(self):             # <<<<<<<<<<<<<<
 *         if len(self._queued_streams) == 0:
 *             if self.handler_check__exec_writes.running:
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__exec_queued_writes(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_stream = 0;
  PyObject *__pyx_v_streams = NULL;
  PyObject *__pyx_v_pystream = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  char const *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_exec_queued_writes", 1);

  /* "uvloop/loop.pyx":644
 * 
 *     cdef _exec_queued_writes(self):
 *         if len(self._queued_streams) == 0:             # <<<<<<<<<<<<<<
 *             if self.handler_check__exec_writes.running:
 *                 self.handler_check__exec_writes.stop()
 */
  __pyx_t_1 = __pyx_v_self->_queued_streams;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 644, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PySet_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 == 0);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":645
 *     cdef _exec_queued_writes(self):
 *         if len(self._queued_streams) == 0:
 *             if self.handler_check__exec_writes.running:             # <<<<<<<<<<<<<<
 *                 self.handler_check__exec_writes.stop()
 *             return
 */
    if (__pyx_v_self->handler_check__exec_writes->running) {

      /* "uvloop/loop.pyx":646
 *         if len(self._queued_streams) == 0:
 *             if self.handler_check__exec_writes.running:
 *                 self.handler_check__exec_writes.stop()             # <<<<<<<<<<<<<<
 *             return
 * 
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_7UVCheck_stop(__pyx_v_self->handler_check__exec_writes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":645
 *     cdef _exec_queued_writes(self):
 *         if len(self._queued_streams) == 0:
 *             if self.handler_check__exec_writes.running:             # <<<<<<<<<<<<<<
 *                 self.handler_check__exec_writes.stop()
 *             return
 */
    }

    /* "uvloop/loop.pyx":647
 *             if self.handler_check__exec_writes.running:
 *                 self.handler_check__exec_writes.stop()
 *             return             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":644
 * 
 *     cdef _exec_queued_writes(self):
 *         if len(self._queued_streams) == 0:             # <<<<<<<<<<<<<<
 *             if self.handler_check__exec_writes.running:
 *                 self.handler_check__exec_writes.stop()
 */
  }

  /* "uvloop/loop.pyx":652
 *             UVStream stream
 * 
 *         streams = self._queued_streams             # <<<<<<<<<<<<<<
 *         self._queued_streams = self._executing_streams
 *         self._executing_streams = streams
 */
  __pyx_t_1 = __pyx_v_self->_queued_streams;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_streams = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":653
 * 
 *         streams = self._queued_streams
 *         self._queued_streams = self._executing_streams             # <<<<<<<<<<<<<<
 *         self._executing_streams = streams
 *         try:
 */
  __pyx_t_1 = __pyx_v_self->_executing_streams;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_queued_streams);
  __Pyx_DECREF(__pyx_v_self->_queued_streams);
  __pyx_v_self->_queued_streams = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":654
 *         streams = self._queued_streams
 *         self._queued_streams = self._executing_streams
 *         self._executing_streams = streams             # <<<<<<<<<<<<<<
 *         try:
 *             for pystream in streams:
 */
  __Pyx_INCREF(__pyx_v_streams);
  __Pyx_GIVEREF(__pyx_v_streams);
  __Pyx_GOTREF(__pyx_v_self->_executing_streams);
  __Pyx_DECREF(__pyx_v_self->_executing_streams);
  __pyx_v_self->_executing_streams = __pyx_v_streams;

  /* "uvloop/loop.pyx":655
 *         self._queued_streams = self._executing_streams
 *         self._executing_streams = streams
 *         try:             # <<<<<<<<<<<<<<
 *             for pystream in streams:
 *                 stream = <UVStream>pystream
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":656
 *         self._executing_streams = streams
 *         try:
 *             for pystream in streams:             # <<<<<<<<<<<<<<
 *                 stream = <UVStream>pystream
 *                 stream._exec_write()
 */
    __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_set_iterator(__pyx_v_streams, 1, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 656, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_7 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_4, &__pyx_t_2, &__pyx_t_6, __pyx_t_5);
      if (unlikely(__pyx_t_7 == 0)) break;
      if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 656, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_pystream, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":657
 *         try:
 *             for pystream in streams:
 *                 stream = <UVStream>pystream             # <<<<<<<<<<<<<<
 *                 stream._exec_write()
 *         finally:
 */
      __pyx_t_6 = __pyx_v_pystream;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_stream, ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":658
 *             for pystream in streams:
 *                 stream = <UVStream>pystream
 *                 stream._exec_write()             # <<<<<<<<<<<<<<
 *         finally:
 *             streams.clear()
 */
      __pyx_t_6 = __pyx_f_6uvloop_4loop_8UVStream__exec_write(__pyx_v_stream); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 658, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "uvloop/loop.pyx":660
 *                 stream._exec_write()
 *         finally:
 *             streams.clear()             # <<<<<<<<<<<<<<
 * 
 *         if self.handler_check__exec_writes.running:
 */
  /*finally:*/ {
    /*normal exit:*/{
      if (unlikely(__pyx_v_streams == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
        __PYX_ERR(0, 660, __pyx_L1_error)
      }
      __pyx_t_8 = PySet_Clear(__pyx_v_streams); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 660, __pyx_L1_error)
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_5 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_9 = __pyx_filename;
      {
        if (unlikely(__pyx_v_streams == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
          __PYX_ERR(0, 660, __pyx_L11_error)
        }
        __pyx_t_8 = PySet_Clear(__pyx_v_streams); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 660, __pyx_L11_error)
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_9;
      goto __pyx_L1_error;
      __pyx_L11_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "uvloop/loop.pyx":662
 *             streams.clear()
 * 
 *         if self.handler_check__exec_writes.running:             # <<<<<<<<<<<<<<
 *             if len(self._queued_streams) == 0:
 *                 self.handler_check__exec_writes.stop()
 */
  if (__pyx_v_self->handler_check__exec_writes->running) {

    /* "uvloop/loop.pyx":663
 * 
 *         if self.handler_check__exec_writes.running:
 *             if len(self._queued_streams) == 0:             # <<<<<<<<<<<<<<
 *                 self.handler_check__exec_writes.stop()
 * 
 */
    __pyx_t_1 = __pyx_v_self->_queued_streams;
    __Pyx_INCREF(__pyx_t_1);
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 663, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PySet_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 663, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = (__pyx_t_4 == 0);
    if (__pyx_t_3) {

      /* "uvloop/loop.pyx":664
 *         if self.handler_check__exec_writes.running:
 *             if len(self._queued_streams) == 0:
 *                 self.handler_check__exec_writes.stop()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _call_soon(self, object callback, object args, object context):
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_7UVCheck_stop(__pyx_v_self->handler_check__exec_writes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":663
 * 
 *         if self.handler_check__exec_writes.running:
 *             if len(self._queued_streams) == 0:             # <<<<<<<<<<<<<<
 *                 self.handler_check__exec_writes.stop()
 * 
 */
    }

    /* "uvloop/loop.pyx":662
 *             streams.clear()
 * 
 *         if self.handler_check__exec_writes.running:             # <<<<<<<<<<<<<<
 *             if len(self._queued_streams) == 0:
 *                 self.handler_check__exec_writes.stop()
 */
  }

  /* "uvloop/loop.pyx":643
 *             self.handler_check__exec_writes.start()
 * 
 *     cdef _exec_queued_writes(self):             # <<<<<<<<<<<<<<
 *         if len(self._queued_streams) == 0:
 *             if self.handler_check__exec_writes.running:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.Loop._exec_queued_writes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XDECREF(__pyx_v_streams);
  __Pyx_XDECREF(__pyx_v_pystream);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":666
 *                 self.handler_check__exec_writes.stop()
 * 
 *     cdef inline _call_soon(self, object callback, object args, object context):             # <<<<<<<<<<<<<<
 *         cdef Handle handle
 *         handle = new_Handle(self, callback, args, context)
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__call_soon(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_soon", 1);

  /* "uvloop/loop.pyx":668
 *     cdef inline _call_soon(self, object callback, object args, object context):
 *         cdef Handle handle
 *         handle = new_Handle(self, callback, args, context)             # <<<<<<<<<<<<<<
 *         self._call_soon_handle(handle)
 *         return handle
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_new_Handle(__pyx_v_self, __pyx_v_callback, __pyx_v_args, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 668, __pyx_L1_error)
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":669
 *         cdef Handle handle
 *         handle = new_Handle(self, callback, args, context)
 *         self._call_soon_handle(handle)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self, __pyx_v_handle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":670
 *         handle = new_Handle(self, callback, args, context)
 *         self._call_soon_handle(handle)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _append_ready_handle(self, Handle handle):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = ((PyObject *)__pyx_v_handle);
  goto __pyx_L0;

  /* "uvloop/loop.pyx":666
 *                 self.handler_check__exec_writes.stop()
 * 
 *     cdef inline _call_soon(self, object callback, object args, object context):             # <<<<<<<<<<<<<<
 *         cdef Handle handle
 *         handle = new_Handle(self, callback, args, context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._call_soon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":672
 *         return handle
 * 
 *     cdef inline _append_ready_handle(self, Handle handle):             # <<<<<<<<<<<<<<
 *         self._check_closed()
 *         self._ready.append(handle)
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__append_ready_handle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_append_ready_handle", 1);

  /* "uvloop/loop.pyx":673
 * 
 *     cdef inline _append_ready_handle(self, Handle handle):
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         self._ready.append(handle)
 *         self._ready_len += 1
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":674
 *     cdef inline _append_ready_handle(self, Handle handle):
 *         self._check_closed()
 *         self._ready.append(handle)             # <<<<<<<<<<<<<<
 *         self._ready_len += 1
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Append(__pyx_v_self->_ready, ((PyObject *)__pyx_v_handle)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 674, __pyx_L1_error)

  /* "uvloop/loop.pyx":675
 *         self._check_closed()
 *         self._ready.append(handle)
 *         self._ready_len += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _call_soon_handle(self, Handle handle):
 */
  __pyx_v_self->_ready_len = (__pyx_v_self->_ready_len + 1);

  /* "uvloop/loop.pyx":672
 *         return handle
 * 
 *     cdef inline _append_ready_handle(self, Handle handle):             # <<<<<<<<<<<<<<
 *         self._check_closed()
 *         self._ready.append(handle)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._append_ready_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":677
 *         self._ready_len += 1
 * 
 *     cdef inline _call_soon_handle(self, Handle handle):             # <<<<<<<<<<<<<<
 *         self._append_ready_handle(handle)
 *         if not self.handler_idle.running:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__call_soon_handle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_soon_handle", 1);

  /* "uvloop/loop.pyx":678
 * 
 *     cdef inline _call_soon_handle(self, Handle handle):
 *         self._append_ready_handle(handle)             # <<<<<<<<<<<<<<
 *         if not self.handler_idle.running:
 *             self.handler_idle.start()
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__append_ready_handle(__pyx_v_self, __pyx_v_handle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":679
 *     cdef inline _call_soon_handle(self, Handle handle):
 *         self._append_ready_handle(handle)
 *         if not self.handler_idle.running:             # <<<<<<<<<<<<<<
 *             self.handler_idle.start()
 * 
 */
  __pyx_t_2 = (!__pyx_v_self->handler_idle->running);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":680
 *         self._append_ready_handle(handle)
 *         if not self.handler_idle.running:
 *             self.handler_idle.start()             # <<<<<<<<<<<<<<
 * 
 *     cdef _call_later(self, uint64_t delay, object callback, object args,
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_6UVIdle_start(__pyx_v_self->handler_idle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":679
 *     cdef inline _call_soon_handle(self, Handle handle):
 *         self._append_ready_handle(handle)
 *         if not self.handler_idle.running:             # <<<<<<<<<<<<<<
 *             self.handler_idle.start()
 * 
 */
  }

  /* "uvloop/loop.pyx":677
 *         self._ready_len += 1
 * 
 *     cdef inline _call_soon_handle(self, Handle handle):             # <<<<<<<<<<<<<<
 *         self._append_ready_handle(handle)
 *         if not self.handler_idle.running:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._call_soon_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":682
 *             self.handler_idle.start()
 * 
 *     cdef _call_later(self, uint64_t delay, object callback, object args,             # <<<<<<<<<<<<<<
 *                      object context):
 *         return TimerHandle(self, callback, args, delay, context)
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__call_later(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, uint64_t __pyx_v_delay, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_later", 1);

  /* "uvloop/loop.pyx":684
 *     cdef _call_later(self, uint64_t delay, object callback, object args,
 *                      object context):
 *         return TimerHandle(self, callback, args, delay, context)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _handle_exception(self, object ex):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_delay); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 684, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_callback)) __PYX_ERR(0, 684, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_args)) __PYX_ERR(0, 684, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_context);
  __Pyx_GIVEREF(__pyx_v_context);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_v_context)) __PYX_ERR(0, 684, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_TimerHandle), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":682
 *             self.handler_idle.start()
 * 
 *     cdef _call_later(self, uint64_t delay, object callback, object args,             # <<<<<<<<<<<<<<
 *                      object context):
 *         return TimerHandle(self, callback, args, delay, context)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._call_later", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":686
 *         return TimerHandle(self, callback, args, delay, context)
 * 
 *     cdef void _handle_exception(self, object ex):             # <<<<<<<<<<<<<<
 *         if isinstance(ex, Exception):
 *             self.call_exception_handler({'exception': ex})
 */

static void __pyx_f_6uvloop_4loop_4Loop__handle_exception(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_ex) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_handle_exception", 1);

  /* "uvloop/loop.pyx":687
 * 
 *     cdef void _handle_exception(self, object ex):
 *         if isinstance(ex, Exception):             # <<<<<<<<<<<<<<
 *             self.call_exception_handler({'exception': ex})
 *         else:
 */
  __pyx_t_1 = __Pyx_PyException_Check(__pyx_v_ex); 
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":688
 *     cdef void _handle_exception(self, object ex):
 *         if isinstance(ex, Exception):
 *             self.call_exception_handler({'exception': ex})             # <<<<<<<<<<<<<<
 *         else:
 *             # BaseException
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_exception, __pyx_v_ex) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":687
 * 
 *     cdef void _handle_exception(self, object ex):
 *         if isinstance(ex, Exception):             # <<<<<<<<<<<<<<
 *             self.call_exception_handler({'exception': ex})
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":691
 *         else:
 *             # BaseException
 *             self._last_error = ex             # <<<<<<<<<<<<<<
 *             # Exit ASAP
 *             self._stop(None)
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_ex);
    __Pyx_GIVEREF(__pyx_v_ex);
    __Pyx_GOTREF(__pyx_v_self->_last_error);
    __Pyx_DECREF(__pyx_v_self->_last_error);
    __pyx_v_self->_last_error = __pyx_v_ex;

    /* "uvloop/loop.pyx":693
 *             self._last_error = ex
 *             # Exit ASAP
 *             self._stop(None)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _check_signal(self, sig):
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_stop(__pyx_v_self, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":686
 *         return TimerHandle(self, callback, args, delay, context)
 * 
 *     cdef void _handle_exception(self, object ex):             # <<<<<<<<<<<<<<
 *         if isinstance(ex, Exception):
 *             self.call_exception_handler({'exception': ex})
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.Loop._handle_exception", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/loop.pyx":695
 *             self._stop(None)
 * 
 *     cdef inline _check_signal(self, sig):             # <<<<<<<<<<<<<<
 *         if not isinstance(sig, int):
 *             raise TypeError('sig must be an int, not {!r}'.format(sig))
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_signal(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_signal", 1);

  /* "uvloop/loop.pyx":696
 * 
 *     cdef inline _check_signal(self, sig):
 *         if not isinstance(sig, int):             # <<<<<<<<<<<<<<
 *             raise TypeError('sig must be an int, not {!r}'.format(sig))
 * 
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_sig); 
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/loop.pyx":697
 *     cdef inline _check_signal(self, sig):
 *         if not isinstance(sig, int):
 *             raise TypeError('sig must be an int, not {!r}'.format(sig))             # <<<<<<<<<<<<<<
 * 
 *         if not (1 <= sig < signal_NSIG):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_sig_must_be_an_int_not_r, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_sig};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 697, __pyx_L1_error)

    /* "uvloop/loop.pyx":696
 * 
 *     cdef inline _check_signal(self, sig):
 *         if not isinstance(sig, int):             # <<<<<<<<<<<<<<
 *             raise TypeError('sig must be an int, not {!r}'.format(sig))
 * 
 */
  }

  /* "uvloop/loop.pyx":699
 *             raise TypeError('sig must be an int, not {!r}'.format(sig))
 * 
 *         if not (1 <= sig < signal_NSIG):             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'sig {} out of range(1, {})'.format(sig, signal_NSIG))
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_int_1, __pyx_v_sig, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 699, __pyx_L1_error)
  if (__Pyx_PyObject_IsTrue(__pyx_t_4)) {
    __Pyx_DECREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_signal_NSIG); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_sig, __pyx_t_3, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (!__pyx_t_2);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":701
 *         if not (1 <= sig < signal_NSIG):
 *             raise ValueError(
 *                 'sig {} out of range(1, {})'.format(sig, signal_NSIG))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _check_closed(self):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_sig_out_of_range_1, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_signal_NSIG); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_v_sig, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 701, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "uvloop/loop.pyx":700
 * 
 *         if not (1 <= sig < signal_NSIG):
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'sig {} out of range(1, {})'.format(sig, signal_NSIG))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 700, __pyx_L1_error)

    /* "uvloop/loop.pyx":699
 *             raise TypeError('sig must be an int, not {!r}'.format(sig))
 * 
 *         if not (1 <= sig < signal_NSIG):             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'sig {} out of range(1, {})'.format(sig, signal_NSIG))
 */
  }

  /* "uvloop/loop.pyx":695
 *             self._stop(None)
 * 
 *     cdef inline _check_signal(self, sig):             # <<<<<<<<<<<<<<
 *         if not isinstance(sig, int):
 *             raise TypeError('sig must be an int, not {!r}'.format(sig))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.Loop._check_signal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":703
 *                 'sig {} out of range(1, {})'.format(sig, signal_NSIG))
 * 
 *     cdef inline _check_closed(self):             # <<<<<<<<<<<<<<
 *         if self._closed == 1:
 *             raise RuntimeError('Event loop is closed')
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_closed(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_closed", 1);

  /* "uvloop/loop.pyx":704
 * 
 *     cdef inline _check_closed(self):
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Event loop is closed')
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_closed == 1);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":705
 *     cdef inline _check_closed(self):
 *         if self._closed == 1:
 *             raise RuntimeError('Event loop is closed')             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _check_thread(self):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 705, __pyx_L1_error)

    /* "uvloop/loop.pyx":704
 * 
 *     cdef inline _check_closed(self):
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Event loop is closed')
 * 
 */
  }

  /* "uvloop/loop.pyx":703
 *                 'sig {} out of range(1, {})'.format(sig, signal_NSIG))
 * 
 *     cdef inline _check_closed(self):             # <<<<<<<<<<<<<<
 *         if self._closed == 1:
 *             raise RuntimeError('Event loop is closed')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._check_closed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":707
 *             raise RuntimeError('Event loop is closed')
 * 
 *     cdef inline _check_thread(self):             # <<<<<<<<<<<<<<
 *         if self._thread_id == 0:
 *             return
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__check_thread(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  uint64_t __pyx_v_thread_id;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_thread", 1);

  /* "uvloop/loop.pyx":708
 * 
 *     cdef inline _check_thread(self):
 *         if self._thread_id == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_thread_id == 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":709
 *     cdef inline _check_thread(self):
 *         if self._thread_id == 0:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         cdef uint64_t thread_id
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":708
 * 
 *     cdef inline _check_thread(self):
 *         if self._thread_id == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/loop.pyx":712
 * 
 *         cdef uint64_t thread_id
 *         thread_id = <uint64_t>PyThread_get_thread_ident()             # <<<<<<<<<<<<<<
 * 
 *         if thread_id != self._thread_id:
 */
  __pyx_v_thread_id = ((uint64_t)PyThread_get_thread_ident());

  /* "uvloop/loop.pyx":714
 *         thread_id = <uint64_t>PyThread_get_thread_ident()
 * 
 *         if thread_id != self._thread_id:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 "Non-thread-safe operation invoked on an event loop other "
 */
  __pyx_t_1 = (__pyx_v_thread_id != __pyx_v_self->_thread_id);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":715
 * 
 *         if thread_id != self._thread_id:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 "Non-thread-safe operation invoked on an event loop other "
 *                 "than the current one")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 715, __pyx_L1_error)

    /* "uvloop/loop.pyx":714
 *         thread_id = <uint64_t>PyThread_get_thread_ident()
 * 
 *         if thread_id != self._thread_id:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 "Non-thread-safe operation invoked on an event loop other "
 */
  }

  /* "uvloop/loop.pyx":707
 *             raise RuntimeError('Event loop is closed')
 * 
 *     cdef inline _check_thread(self):             # <<<<<<<<<<<<<<
 *         if self._thread_id == 0:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._check_thread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":719
 *                 "than the current one")
 * 
 *     cdef inline _new_future(self):             # <<<<<<<<<<<<<<
 *         return aio_Future(loop=self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_4Loop__new_future(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_future", 1);

  /* "uvloop/loop.pyx":720
 * 
 *     cdef inline _new_future(self):
 *         return aio_Future(loop=self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _track_transport(self, UVBaseTransport transport):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_loop, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 720, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_Future, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":719
 *                 "than the current one")
 * 
 *     cdef inline _new_future(self):             # <<<<<<<<<<<<<<
 *         return aio_Future(loop=self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._new_future", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":722
 *         return aio_Future(loop=self)
 * 
 *     cdef _track_transport(self, UVBaseTransport transport):             # <<<<<<<<<<<<<<
 *         self._transports[transport._fileno()] = transport
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__track_transport(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_transport) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_track_transport", 1);

  /* "uvloop/loop.pyx":723
 * 
 *     cdef _track_transport(self, UVBaseTransport transport):
 *         self._transports[transport._fileno()] = transport             # <<<<<<<<<<<<<<
 * 
 *     cdef _track_process(self, UVProcess proc):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_transport->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._fileno(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_transport)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely((PyObject_SetItem(__pyx_v_self->_transports, __pyx_t_1, ((PyObject *)__pyx_v_transport)) < 0))) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":722
 *         return aio_Future(loop=self)
 * 
 *     cdef _track_transport(self, UVBaseTransport transport):             # <<<<<<<<<<<<<<
 *         self._transports[transport._fileno()] = transport
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._track_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":725
 *         self._transports[transport._fileno()] = transport
 * 
 *     cdef _track_process(self, UVProcess proc):             # <<<<<<<<<<<<<<
 *         self._processes.add(proc)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__track_process(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_proc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_track_process", 1);

  /* "uvloop/loop.pyx":726
 * 
 *     cdef _track_process(self, UVProcess proc):
 *         self._processes.add(proc)             # <<<<<<<<<<<<<<
 * 
 *     cdef _untrack_process(self, UVProcess proc):
 */
  if (unlikely(__pyx_v_self->_processes == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(0, 726, __pyx_L1_error)
  }
  __pyx_t_1 = PySet_Add(__pyx_v_self->_processes, ((PyObject *)__pyx_v_proc)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 726, __pyx_L1_error)

  /* "uvloop/loop.pyx":725
 *         self._transports[transport._fileno()] = transport
 * 
 *     cdef _track_process(self, UVProcess proc):             # <<<<<<<<<<<<<<
 *         self._processes.add(proc)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop._track_process", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":728
 *         self._processes.add(proc)
 * 
 *     cdef _untrack_process(self, UVProcess proc):             # <<<<<<<<<<<<<<
 *         self._processes.discard(proc)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__untrack_process(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_proc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_untrack_process", 1);

  /* "uvloop/loop.pyx":729
 * 
 *     cdef _untrack_process(self, UVProcess proc):
 *         self._processes.discard(proc)             # <<<<<<<<<<<<<<
 * 
 *     cdef _fileobj_to_fd(self, fileobj):
 */
  if (unlikely(__pyx_v_self->_processes == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "discard");
    __PYX_ERR(0, 729, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PySet_Discard(__pyx_v_self->_processes, ((PyObject *)__pyx_v_proc)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 729, __pyx_L1_error)

  /* "uvloop/loop.pyx":728
 *         self._processes.add(proc)
 * 
 *     cdef _untrack_process(self, UVProcess proc):             # <<<<<<<<<<<<<<
 *         self._processes.discard(proc)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop._untrack_process", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":731
 *         self._processes.discard(proc)
 * 
 *     cdef _fileobj_to_fd(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Return a file descriptor from a file object.
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__fileobj_to_fd(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj) {
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fileobj_to_fd", 1);

  /* "uvloop/loop.pyx":744
 *         """
 *         # Copy of the `selectors._fileobj_to_fd()` function.
 *         if isinstance(fileobj, int):             # <<<<<<<<<<<<<<
 *             fd = fileobj
 *         else:
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_fileobj); 
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":745
 *         # Copy of the `selectors._fileobj_to_fd()` function.
 *         if isinstance(fileobj, int):
 *             fd = fileobj             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
    __Pyx_INCREF(__pyx_v_fileobj);
    __pyx_v_fd = __pyx_v_fileobj;

    /* "uvloop/loop.pyx":744
 *         """
 *         # Copy of the `selectors._fileobj_to_fd()` function.
 *         if isinstance(fileobj, int):             # <<<<<<<<<<<<<<
 *             fd = fileobj
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":747
 *             fd = fileobj
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 fd = int(fileobj.fileno())
 *             except (AttributeError, TypeError, ValueError):
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/loop.pyx":748
 *         else:
 *             try:
 *                 fd = int(fileobj.fileno())             # <<<<<<<<<<<<<<
 *             except (AttributeError, TypeError, ValueError):
 *                 raise ValueError("Invalid file object: "
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_fileobj, __pyx_n_s_fileno); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 748, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 748, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 748, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_fd = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "uvloop/loop.pyx":747
 *             fd = fileobj
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 fd = int(fileobj.fileno())
 *             except (AttributeError, TypeError, ValueError):
 */
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":749
 *             try:
 *                 fd = int(fileobj.fileno())
 *             except (AttributeError, TypeError, ValueError):             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid file object: "
 *                                  "{!r}".format(fileobj)) from None
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop._fileobj_to_fd", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_7) < 0) __PYX_ERR(0, 749, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "uvloop/loop.pyx":751
 *             except (AttributeError, TypeError, ValueError):
 *                 raise ValueError("Invalid file object: "
 *                                  "{!r}".format(fileobj)) from None             # <<<<<<<<<<<<<<
 *         if fd < 0:
 *             raise ValueError("Invalid file descriptor: {}".format(fd))
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Invalid_file_object_r, __pyx_n_s_format); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 751, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_fileobj};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 751, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }

        /* "uvloop/loop.pyx":750
 *                 fd = int(fileobj.fileno())
 *             except (AttributeError, TypeError, ValueError):
 *                 raise ValueError("Invalid file object: "             # <<<<<<<<<<<<<<
 *                                  "{!r}".format(fileobj)) from None
 *         if fd < 0:
 */
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 750, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/loop.pyx":751
 *             except (AttributeError, TypeError, ValueError):
 *                 raise ValueError("Invalid file object: "
 *                                  "{!r}".format(fileobj)) from None             # <<<<<<<<<<<<<<
 *         if fd < 0:
 *             raise ValueError("Invalid file descriptor: {}".format(fd))
 */
        __Pyx_Raise(__pyx_t_11, 0, 0, Py_None);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 750, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;

      /* "uvloop/loop.pyx":747
 *             fd = fileobj
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 fd = int(fileobj.fileno())
 *             except (AttributeError, TypeError, ValueError):
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":752
 *                 raise ValueError("Invalid file object: "
 *                                  "{!r}".format(fileobj)) from None
 *         if fd < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Invalid file descriptor: {}".format(fd))
 *         return fd
 */
  __pyx_t_7 = PyObject_RichCompare(__pyx_v_fd, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 752, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":753
 *                                  "{!r}".format(fileobj)) from None
 *         if fd < 0:
 *             raise ValueError("Invalid file descriptor: {}".format(fd))             # <<<<<<<<<<<<<<
 *         return fd
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Invalid_file_descriptor, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_fd};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 753, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 753, __pyx_L1_error)

    /* "uvloop/loop.pyx":752
 *                 raise ValueError("Invalid file object: "
 *                                  "{!r}".format(fileobj)) from None
 *         if fd < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Invalid file descriptor: {}".format(fd))
 *         return fd
 */
  }

  /* "uvloop/loop.pyx":754
 *         if fd < 0:
 *             raise ValueError("Invalid file descriptor: {}".format(fd))
 *         return fd             # <<<<<<<<<<<<<<
 * 
 *     cdef _ensure_fd_no_transport(self, fd):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fd);
  __pyx_r = __pyx_v_fd;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":731
 *         self._processes.discard(proc)
 * 
 *     cdef _fileobj_to_fd(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Return a file descriptor from a file object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uvloop.loop.Loop._fileobj_to_fd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":756
 *         return fd
 * 
 *     cdef _ensure_fd_no_transport(self, fd):             # <<<<<<<<<<<<<<
 *         cdef UVBaseTransport tr
 *         try:
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__ensure_fd_no_transport(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fd) {
  struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_tr = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ensure_fd_no_transport", 1);

  /* "uvloop/loop.pyx":758
 *     cdef _ensure_fd_no_transport(self, fd):
 *         cdef UVBaseTransport tr
 *         try:             # <<<<<<<<<<<<<<
 *             tr = <UVBaseTransport>(self._transports[fd])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/loop.pyx":759
 *         cdef UVBaseTransport tr
 *         try:
 *             tr = <UVBaseTransport>(self._transports[fd])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             pass
 */
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_self->_transports, __pyx_v_fd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 759, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __pyx_t_4;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_tr = ((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":758
 *     cdef _ensure_fd_no_transport(self, fd):
 *         cdef UVBaseTransport tr
 *         try:             # <<<<<<<<<<<<<<
 *             tr = <UVBaseTransport>(self._transports[fd])
 *         except KeyError:
 */
    }

    /* "uvloop/loop.pyx":763
 *             pass
 *         else:
 *             if tr._is_alive():             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     'File descriptor {!r} is used by transport {!r}'.format(
 */
    /*else:*/ {
      __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_tr->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_tr)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 763, __pyx_L5_except_error)
      if (unlikely(__pyx_t_6)) {

        /* "uvloop/loop.pyx":765
 *             if tr._is_alive():
 *                 raise RuntimeError(
 *                     'File descriptor {!r} is used by transport {!r}'.format(             # <<<<<<<<<<<<<<
 *                         fd, tr))
 * 
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_File_descriptor_r_is_used_by_tra, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 765, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "uvloop/loop.pyx":766
 *                 raise RuntimeError(
 *                     'File descriptor {!r} is used by transport {!r}'.format(
 *                         fd, tr))             # <<<<<<<<<<<<<<
 * 
 *     cdef _add_reader(self, fileobj, Handle handle):
 */
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_v_fd, ((PyObject *)__pyx_v_tr)};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 765, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "uvloop/loop.pyx":764
 *         else:
 *             if tr._is_alive():
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     'File descriptor {!r} is used by transport {!r}'.format(
 *                         fd, tr))
 */
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 764, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 764, __pyx_L5_except_error)

        /* "uvloop/loop.pyx":763
 *             pass
 *         else:
 *             if tr._is_alive():             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     'File descriptor {!r} is used by transport {!r}'.format(
 */
      }
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":760
 *         try:
 *             tr = <UVBaseTransport>(self._transports[fd])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_9) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":758
 *     cdef _ensure_fd_no_transport(self, fd):
 *         cdef UVBaseTransport tr
 *         try:             # <<<<<<<<<<<<<<
 *             tr = <UVBaseTransport>(self._transports[fd])
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":756
 *         return fd
 * 
 *     cdef _ensure_fd_no_transport(self, fd):             # <<<<<<<<<<<<<<
 *         cdef UVBaseTransport tr
 *         try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.Loop._ensure_fd_no_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":768
 *                         fd, tr))
 * 
 *     cdef _add_reader(self, fileobj, Handle handle):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__add_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_poll = 0;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_v_old_fileobj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_add_reader", 1);

  /* "uvloop/loop.pyx":772
 *             UVPoll poll
 * 
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         fd = self._fileobj_to_fd(fileobj)
 *         self._ensure_fd_no_transport(fd)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":773
 * 
 *         self._check_closed()
 *         fd = self._fileobj_to_fd(fileobj)             # <<<<<<<<<<<<<<
 *         self._ensure_fd_no_transport(fd)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_fileobj_to_fd(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":774
 *         self._check_closed()
 *         fd = self._fileobj_to_fd(fileobj)
 *         self._ensure_fd_no_transport(fd)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_ensure_fd_no_transport(__pyx_v_self, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":776
 *         self._ensure_fd_no_transport(fd)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/loop.pyx":777
 * 
 *         try:
 *             poll = <UVPoll>(self._polls[fd])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             poll = UVPoll.new(self, fd)
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 777, __pyx_L3_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_polls, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_poll = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":776
 *         self._ensure_fd_no_transport(fd)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":778
 *         try:
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             poll = UVPoll.new(self, fd)
 *             self._polls[fd] = poll
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("uvloop.loop.Loop._add_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 778, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "uvloop/loop.pyx":779
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 *             poll = UVPoll.new(self, fd)             # <<<<<<<<<<<<<<
 *             self._polls[fd] = poll
 * 
 */
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_fd); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 779, __pyx_L5_except_error)
      __pyx_t_8 = ((PyObject *)__pyx_f_6uvloop_4loop_6UVPoll_new(__pyx_v_self, __pyx_t_6)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 779, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_poll, ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "uvloop/loop.pyx":780
 *         except KeyError:
 *             poll = UVPoll.new(self, fd)
 *             self._polls[fd] = poll             # <<<<<<<<<<<<<<
 * 
 *         poll.start_reading(handle)
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 780, __pyx_L5_except_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_self->_polls, __pyx_v_fd, ((PyObject *)__pyx_v_poll)) < 0))) __PYX_ERR(0, 780, __pyx_L5_except_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":776
 *         self._ensure_fd_no_transport(fd)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":782
 *             self._polls[fd] = poll
 * 
 *         poll.start_reading(handle)             # <<<<<<<<<<<<<<
 * 
 *         old_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->start_reading(__pyx_v_poll, __pyx_v_handle); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":784
 *         poll.start_reading(handle)
 * 
 *         old_fileobj = self._fd_to_reader_fileobj.pop(fd, None)             # <<<<<<<<<<<<<<
 *         if old_fileobj is not None:
 *             socket_dec_io_ref(old_fileobj)
 */
  if (unlikely(__pyx_v_self->_fd_to_reader_fileobj == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
    __PYX_ERR(0, 784, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyDict_Pop(__pyx_v_self->_fd_to_reader_fileobj, __pyx_v_fd, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_old_fileobj = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":785
 * 
 *         old_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 *         if old_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(old_fileobj)
 * 
 */
  __pyx_t_9 = (__pyx_v_old_fileobj != Py_None);
  if (__pyx_t_9) {

    /* "uvloop/loop.pyx":786
 *         old_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 *         if old_fileobj is not None:
 *             socket_dec_io_ref(old_fileobj)             # <<<<<<<<<<<<<<
 * 
 *         self._fd_to_reader_fileobj[fd] = fileobj
 */
    __pyx_t_7 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_v_old_fileobj); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":785
 * 
 *         old_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 *         if old_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(old_fileobj)
 * 
 */
  }

  /* "uvloop/loop.pyx":788
 *             socket_dec_io_ref(old_fileobj)
 * 
 *         self._fd_to_reader_fileobj[fd] = fileobj             # <<<<<<<<<<<<<<
 *         socket_inc_io_ref(fileobj)
 * 
 */
  if (unlikely(__pyx_v_self->_fd_to_reader_fileobj == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 788, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->_fd_to_reader_fileobj, __pyx_v_fd, __pyx_v_fileobj) < 0))) __PYX_ERR(0, 788, __pyx_L1_error)

  /* "uvloop/loop.pyx":789
 * 
 *         self._fd_to_reader_fileobj[fd] = fileobj
 *         socket_inc_io_ref(fileobj)             # <<<<<<<<<<<<<<
 * 
 *     cdef _remove_reader(self, fileobj):
 */
  __pyx_t_7 = __pyx_f_6uvloop_4loop_socket_inc_io_ref(__pyx_v_fileobj); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":768
 *                         fd, tr))
 * 
 *     cdef _add_reader(self, fileobj, Handle handle):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Loop._add_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_poll);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XDECREF(__pyx_v_old_fileobj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":791
 *         socket_inc_io_ref(fileobj)
 * 
 *     cdef _remove_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__remove_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_poll = 0;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_v_mapped_fileobj = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_remove_reader", 1);

  /* "uvloop/loop.pyx":795
 *             UVPoll poll
 * 
 *         fd = self._fileobj_to_fd(fileobj)             # <<<<<<<<<<<<<<
 *         self._ensure_fd_no_transport(fd)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_fileobj_to_fd(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":796
 * 
 *         fd = self._fileobj_to_fd(fileobj)
 *         self._ensure_fd_no_transport(fd)             # <<<<<<<<<<<<<<
 * 
 *         mapped_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_ensure_fd_no_transport(__pyx_v_self, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":798
 *         self._ensure_fd_no_transport(fd)
 * 
 *         mapped_fileobj = self._fd_to_reader_fileobj.pop(fd, None)             # <<<<<<<<<<<<<<
 *         if mapped_fileobj is not None:
 *             socket_dec_io_ref(mapped_fileobj)
 */
  if (unlikely(__pyx_v_self->_fd_to_reader_fileobj == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
    __PYX_ERR(0, 798, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_self->_fd_to_reader_fileobj, __pyx_v_fd, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_mapped_fileobj = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":799
 * 
 *         mapped_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 *         if mapped_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 */
  __pyx_t_2 = (__pyx_v_mapped_fileobj != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":800
 *         mapped_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 *         if mapped_fileobj is not None:
 *             socket_dec_io_ref(mapped_fileobj)             # <<<<<<<<<<<<<<
 * 
 *         if self._closed == 1:
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_v_mapped_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":799
 * 
 *         mapped_fileobj = self._fd_to_reader_fileobj.pop(fd, None)
 *         if mapped_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 */
  }

  /* "uvloop/loop.pyx":802
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_closed == 1);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":803
 * 
 *         if self._closed == 1:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":802
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "uvloop/loop.pyx":805
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "uvloop/loop.pyx":806
 * 
 *         try:
 *             poll = <UVPoll>(self._polls[fd])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             return False
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 806, __pyx_L5_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_polls, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_poll = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":805
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":807
 *         try:
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("uvloop.loop.Loop._remove_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_1, &__pyx_t_8) < 0) __PYX_ERR(0, 807, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "uvloop/loop.pyx":808
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         result = poll.stop_reading()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L8_except_return;
    }
    goto __pyx_L7_except_error;

    /* "uvloop/loop.pyx":805
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L8_except_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L0;
    __pyx_L10_try_end:;
  }

  /* "uvloop/loop.pyx":810
 *             return False
 * 
 *         result = poll.stop_reading()             # <<<<<<<<<<<<<<
 *         if not poll.is_active():
 *             del self._polls[fd]
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->stop_reading(__pyx_v_poll); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_result = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "uvloop/loop.pyx":811
 * 
 *         result = poll.stop_reading()
 *         if not poll.is_active():             # <<<<<<<<<<<<<<
 *             del self._polls[fd]
 *             poll._close()
 */
  __pyx_t_2 = (!(((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->is_active(__pyx_v_poll) != 0));
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":812
 *         result = poll.stop_reading()
 *         if not poll.is_active():
 *             del self._polls[fd]             # <<<<<<<<<<<<<<
 *             poll._close()
 * 
 */
    if (unlikely(__pyx_v_self->_polls == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 812, __pyx_L1_error)
    }
    if (unlikely((PyDict_DelItem(__pyx_v_self->_polls, __pyx_v_fd) < 0))) __PYX_ERR(0, 812, __pyx_L1_error)

    /* "uvloop/loop.pyx":813
 *         if not poll.is_active():
 *             del self._polls[fd]
 *             poll._close()             # <<<<<<<<<<<<<<
 * 
 *         return result
 */
    __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_poll)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/loop.pyx":811
 * 
 *         result = poll.stop_reading()
 *         if not poll.is_active():             # <<<<<<<<<<<<<<
 *             del self._polls[fd]
 *             poll._close()
 */
  }

  /* "uvloop/loop.pyx":815
 *             poll._close()
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef _has_reader(self, fileobj):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":791
 *         socket_inc_io_ref(fileobj)
 * 
 *     cdef _remove_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Loop._remove_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_poll);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XDECREF(__pyx_v_mapped_fileobj);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":817
 *         return result
 * 
 *     cdef _has_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__has_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_poll = 0;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_has_reader", 1);

  /* "uvloop/loop.pyx":821
 *             UVPoll poll
 * 
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":822
 * 
 *         self._check_closed()
 *         fd = self._fileobj_to_fd(fileobj)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_fileobj_to_fd(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":824
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/loop.pyx":825
 * 
 *         try:
 *             poll = <UVPoll>(self._polls[fd])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             return False
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 825, __pyx_L3_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_polls, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_poll = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":824
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":826
 *         try:
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("uvloop.loop.Loop._has_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 826, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "uvloop/loop.pyx":827
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         return poll.is_reading()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":824
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":829
 *             return False
 * 
 *         return poll.is_reading()             # <<<<<<<<<<<<<<
 * 
 *     cdef _add_writer(self, fileobj, Handle handle):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->is_reading(__pyx_v_poll); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":817
 *         return result
 * 
 *     cdef _has_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.Loop._has_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_poll);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":831
 *         return poll.is_reading()
 * 
 *     cdef _add_writer(self, fileobj, Handle handle):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__add_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_poll = 0;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_v_old_fileobj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_add_writer", 1);

  /* "uvloop/loop.pyx":835
 *             UVPoll poll
 * 
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         fd = self._fileobj_to_fd(fileobj)
 *         self._ensure_fd_no_transport(fd)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":836
 * 
 *         self._check_closed()
 *         fd = self._fileobj_to_fd(fileobj)             # <<<<<<<<<<<<<<
 *         self._ensure_fd_no_transport(fd)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_fileobj_to_fd(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":837
 *         self._check_closed()
 *         fd = self._fileobj_to_fd(fileobj)
 *         self._ensure_fd_no_transport(fd)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_ensure_fd_no_transport(__pyx_v_self, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":839
 *         self._ensure_fd_no_transport(fd)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/loop.pyx":840
 * 
 *         try:
 *             poll = <UVPoll>(self._polls[fd])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             poll = UVPoll.new(self, fd)
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 840, __pyx_L3_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_polls, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_poll = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":839
 *         self._ensure_fd_no_transport(fd)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":841
 *         try:
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             poll = UVPoll.new(self, fd)
 *             self._polls[fd] = poll
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("uvloop.loop.Loop._add_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 841, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "uvloop/loop.pyx":842
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 *             poll = UVPoll.new(self, fd)             # <<<<<<<<<<<<<<
 *             self._polls[fd] = poll
 * 
 */
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_fd); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 842, __pyx_L5_except_error)
      __pyx_t_8 = ((PyObject *)__pyx_f_6uvloop_4loop_6UVPoll_new(__pyx_v_self, __pyx_t_6)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 842, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_poll, ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "uvloop/loop.pyx":843
 *         except KeyError:
 *             poll = UVPoll.new(self, fd)
 *             self._polls[fd] = poll             # <<<<<<<<<<<<<<
 * 
 *         poll.start_writing(handle)
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 843, __pyx_L5_except_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_self->_polls, __pyx_v_fd, ((PyObject *)__pyx_v_poll)) < 0))) __PYX_ERR(0, 843, __pyx_L5_except_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":839
 *         self._ensure_fd_no_transport(fd)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":845
 *             self._polls[fd] = poll
 * 
 *         poll.start_writing(handle)             # <<<<<<<<<<<<<<
 * 
 *         old_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->start_writing(__pyx_v_poll, __pyx_v_handle); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":847
 *         poll.start_writing(handle)
 * 
 *         old_fileobj = self._fd_to_writer_fileobj.pop(fd, None)             # <<<<<<<<<<<<<<
 *         if old_fileobj is not None:
 *             socket_dec_io_ref(old_fileobj)
 */
  if (unlikely(__pyx_v_self->_fd_to_writer_fileobj == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
    __PYX_ERR(0, 847, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyDict_Pop(__pyx_v_self->_fd_to_writer_fileobj, __pyx_v_fd, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_old_fileobj = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":848
 * 
 *         old_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 *         if old_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(old_fileobj)
 * 
 */
  __pyx_t_9 = (__pyx_v_old_fileobj != Py_None);
  if (__pyx_t_9) {

    /* "uvloop/loop.pyx":849
 *         old_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 *         if old_fileobj is not None:
 *             socket_dec_io_ref(old_fileobj)             # <<<<<<<<<<<<<<
 * 
 *         self._fd_to_writer_fileobj[fd] = fileobj
 */
    __pyx_t_7 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_v_old_fileobj); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":848
 * 
 *         old_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 *         if old_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(old_fileobj)
 * 
 */
  }

  /* "uvloop/loop.pyx":851
 *             socket_dec_io_ref(old_fileobj)
 * 
 *         self._fd_to_writer_fileobj[fd] = fileobj             # <<<<<<<<<<<<<<
 *         socket_inc_io_ref(fileobj)
 * 
 */
  if (unlikely(__pyx_v_self->_fd_to_writer_fileobj == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 851, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->_fd_to_writer_fileobj, __pyx_v_fd, __pyx_v_fileobj) < 0))) __PYX_ERR(0, 851, __pyx_L1_error)

  /* "uvloop/loop.pyx":852
 * 
 *         self._fd_to_writer_fileobj[fd] = fileobj
 *         socket_inc_io_ref(fileobj)             # <<<<<<<<<<<<<<
 * 
 *     cdef _remove_writer(self, fileobj):
 */
  __pyx_t_7 = __pyx_f_6uvloop_4loop_socket_inc_io_ref(__pyx_v_fileobj); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":831
 *         return poll.is_reading()
 * 
 *     cdef _add_writer(self, fileobj, Handle handle):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Loop._add_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_poll);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XDECREF(__pyx_v_old_fileobj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":854
 *         socket_inc_io_ref(fileobj)
 * 
 *     cdef _remove_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__remove_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_poll = 0;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_v_mapped_fileobj = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_remove_writer", 1);

  /* "uvloop/loop.pyx":858
 *             UVPoll poll
 * 
 *         fd = self._fileobj_to_fd(fileobj)             # <<<<<<<<<<<<<<
 *         self._ensure_fd_no_transport(fd)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_fileobj_to_fd(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":859
 * 
 *         fd = self._fileobj_to_fd(fileobj)
 *         self._ensure_fd_no_transport(fd)             # <<<<<<<<<<<<<<
 * 
 *         mapped_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_ensure_fd_no_transport(__pyx_v_self, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":861
 *         self._ensure_fd_no_transport(fd)
 * 
 *         mapped_fileobj = self._fd_to_writer_fileobj.pop(fd, None)             # <<<<<<<<<<<<<<
 *         if mapped_fileobj is not None:
 *             socket_dec_io_ref(mapped_fileobj)
 */
  if (unlikely(__pyx_v_self->_fd_to_writer_fileobj == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
    __PYX_ERR(0, 861, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_self->_fd_to_writer_fileobj, __pyx_v_fd, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_mapped_fileobj = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":862
 * 
 *         mapped_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 *         if mapped_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 */
  __pyx_t_2 = (__pyx_v_mapped_fileobj != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":863
 *         mapped_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 *         if mapped_fileobj is not None:
 *             socket_dec_io_ref(mapped_fileobj)             # <<<<<<<<<<<<<<
 * 
 *         if self._closed == 1:
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_v_mapped_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":862
 * 
 *         mapped_fileobj = self._fd_to_writer_fileobj.pop(fd, None)
 *         if mapped_fileobj is not None:             # <<<<<<<<<<<<<<
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 */
  }

  /* "uvloop/loop.pyx":865
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_closed == 1);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":866
 * 
 *         if self._closed == 1:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":865
 *             socket_dec_io_ref(mapped_fileobj)
 * 
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "uvloop/loop.pyx":868
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "uvloop/loop.pyx":869
 * 
 *         try:
 *             poll = <UVPoll>(self._polls[fd])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             return False
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 869, __pyx_L5_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_polls, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 869, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_poll = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":868
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":870
 *         try:
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("uvloop.loop.Loop._remove_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_1, &__pyx_t_8) < 0) __PYX_ERR(0, 870, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "uvloop/loop.pyx":871
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         result = poll.stop_writing()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L8_except_return;
    }
    goto __pyx_L7_except_error;

    /* "uvloop/loop.pyx":868
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L8_except_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L0;
    __pyx_L10_try_end:;
  }

  /* "uvloop/loop.pyx":873
 *             return False
 * 
 *         result = poll.stop_writing()             # <<<<<<<<<<<<<<
 *         if not poll.is_active():
 *             del self._polls[fd]
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->stop_writing(__pyx_v_poll); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_result = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "uvloop/loop.pyx":874
 * 
 *         result = poll.stop_writing()
 *         if not poll.is_active():             # <<<<<<<<<<<<<<
 *             del self._polls[fd]
 *             poll._close()
 */
  __pyx_t_2 = (!(((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->is_active(__pyx_v_poll) != 0));
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":875
 *         result = poll.stop_writing()
 *         if not poll.is_active():
 *             del self._polls[fd]             # <<<<<<<<<<<<<<
 *             poll._close()
 * 
 */
    if (unlikely(__pyx_v_self->_polls == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 875, __pyx_L1_error)
    }
    if (unlikely((PyDict_DelItem(__pyx_v_self->_polls, __pyx_v_fd) < 0))) __PYX_ERR(0, 875, __pyx_L1_error)

    /* "uvloop/loop.pyx":876
 *         if not poll.is_active():
 *             del self._polls[fd]
 *             poll._close()             # <<<<<<<<<<<<<<
 * 
 *         return result
 */
    __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_poll)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/loop.pyx":874
 * 
 *         result = poll.stop_writing()
 *         if not poll.is_active():             # <<<<<<<<<<<<<<
 *             del self._polls[fd]
 *             poll._close()
 */
  }

  /* "uvloop/loop.pyx":878
 *             poll._close()
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef _has_writer(self, fileobj):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":854
 *         socket_inc_io_ref(fileobj)
 * 
 *     cdef _remove_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Loop._remove_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_poll);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XDECREF(__pyx_v_mapped_fileobj);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":880
 *         return result
 * 
 *     cdef _has_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__has_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_poll = 0;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_has_writer", 1);

  /* "uvloop/loop.pyx":884
 *             UVPoll poll
 * 
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":885
 * 
 *         self._check_closed()
 *         fd = self._fileobj_to_fd(fileobj)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_fileobj_to_fd(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":887
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/loop.pyx":888
 * 
 *         try:
 *             poll = <UVPoll>(self._polls[fd])             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             return False
 */
      if (unlikely(__pyx_v_self->_polls == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 888, __pyx_L3_error)
      }
      __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_polls, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_poll = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":887
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":889
 *         try:
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("uvloop.loop.Loop._has_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 889, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "uvloop/loop.pyx":890
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         return poll.is_writing()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":887
 *         fd = self._fileobj_to_fd(fileobj)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             poll = <UVPoll>(self._polls[fd])
 *         except KeyError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":892
 *             return False
 * 
 *         return poll.is_writing()             # <<<<<<<<<<<<<<
 * 
 *     cdef _getaddrinfo(self, object host, object port,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->is_writing(__pyx_v_poll); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":880
 *         return result
 * 
 *     cdef _has_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVPoll poll
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.Loop._has_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_poll);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":914
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if AddrInfo.isinstance(result):
 *                 try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_12_getaddrinfo_1callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_12_getaddrinfo_1callback = {"callback", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_12_getaddrinfo_1callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_12_getaddrinfo_1callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_result = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_result)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 914, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "callback") < 0)) __PYX_ERR(0, 914, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_result = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("callback", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 914, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop._getaddrinfo.callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_12_getaddrinfo_callback(__pyx_self, __pyx_v_result);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_12_getaddrinfo_callback(PyObject *__pyx_self, PyObject *__pyx_v_result) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *__pyx_cur_scope;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *__pyx_outer_scope;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("callback", 1);
  __pyx_outer_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "uvloop/loop.pyx":915
 * 
 *         def callback(result):
 *             if AddrInfo.isinstance(result):             # <<<<<<<<<<<<<<
 *                 try:
 *                     if unpack == 0:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_8AddrInfo_isinstance(__pyx_v_result); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 915, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":916
 *         def callback(result):
 *             if AddrInfo.isinstance(result):
 *                 try:             # <<<<<<<<<<<<<<
 *                     if unpack == 0:
 *                         data = result
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "uvloop/loop.pyx":917
 *             if AddrInfo.isinstance(result):
 *                 try:
 *                     if unpack == 0:             # <<<<<<<<<<<<<<
 *                         data = result
 *                     else:
 */
        __pyx_t_2 = (__pyx_cur_scope->__pyx_v_unpack == 0);
        if (__pyx_t_2) {

          /* "uvloop/loop.pyx":918
 *                 try:
 *                     if unpack == 0:
 *                         data = result             # <<<<<<<<<<<<<<
 *                     else:
 *                         data = (<AddrInfo>result).unpack()
 */
          __Pyx_INCREF(__pyx_v_result);
          __pyx_v_data = __pyx_v_result;

          /* "uvloop/loop.pyx":917
 *             if AddrInfo.isinstance(result):
 *                 try:
 *                     if unpack == 0:             # <<<<<<<<<<<<<<
 *                         data = result
 *                     else:
 */
          goto __pyx_L10;
        }

        /* "uvloop/loop.pyx":920
 *                         data = result
 *                     else:
 *                         data = (<AddrInfo>result).unpack()             # <<<<<<<<<<<<<<
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 */
        /*else*/ {
          __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_AddrInfo *)((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_v_result)->__pyx_vtab)->unpack(((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_v_result)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 920, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_data = __pyx_t_6;
          __pyx_t_6 = 0;
        }
        __pyx_L10:;

        /* "uvloop/loop.pyx":916
 *         def callback(result):
 *             if AddrInfo.isinstance(result):
 *                 try:             # <<<<<<<<<<<<<<
 *                     if unpack == 0:
 *                         data = result
 */
      }

      /* "uvloop/loop.pyx":927
 *                         fut.set_exception(ex)
 *                 else:
 *                     if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                         fut.set_result(data)
 *             else:
 */
      /*else:*/ {
        if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 927, __pyx_L6_except_error) }
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 927, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 927, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 927, __pyx_L6_except_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_10 = (!__pyx_t_2);
        if (__pyx_t_10) {

          /* "uvloop/loop.pyx":928
 *                 else:
 *                     if not fut.cancelled():
 *                         fut.set_result(data)             # <<<<<<<<<<<<<<
 *             else:
 *                 if not fut.cancelled():
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 928, __pyx_L6_except_error) }
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 928, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_8 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_data};
            __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 928, __pyx_L6_except_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "uvloop/loop.pyx":927
 *                         fut.set_exception(ex)
 *                 else:
 *                     if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                         fut.set_result(data)
 *             else:
 */
        }
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":921
 *                     else:
 *                         data = (<AddrInfo>result).unpack()
 *                 except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                     raise
 *                 except BaseException as ex:
 */
      __pyx_t_1 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_1) {
        __Pyx_AddTraceback("uvloop.loop.Loop._getaddrinfo.callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 921, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);

        /* "uvloop/loop.pyx":922
 *                         data = (<AddrInfo>result).unpack()
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise             # <<<<<<<<<<<<<<
 *                 except BaseException as ex:
 *                     if not fut.cancelled():
 */
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; 
        __PYX_ERR(0, 922, __pyx_L6_except_error)
      }

      /* "uvloop/loop.pyx":923
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     if not fut.cancelled():
 *                         fut.set_exception(ex)
 */
      __pyx_t_1 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_1) {
        __Pyx_AddTraceback("uvloop.loop.Loop._getaddrinfo.callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 923, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_ex = __pyx_t_7;
        /*try:*/ {

          /* "uvloop/loop.pyx":924
 *                     raise
 *                 except BaseException as ex:
 *                     if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                         fut.set_exception(ex)
 *                 else:
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 924, __pyx_L19_error) }
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 924, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 924, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 924, __pyx_L19_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_2 = (!__pyx_t_10);
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":925
 *                 except BaseException as ex:
 *                     if not fut.cancelled():
 *                         fut.set_exception(ex)             # <<<<<<<<<<<<<<
 *                 else:
 *                     if not fut.cancelled():
 */
            if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 925, __pyx_L19_error) }
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 925, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = NULL;
            __pyx_t_9 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
                __pyx_t_9 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_v_ex};
              __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 925, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

            /* "uvloop/loop.pyx":924
 *                     raise
 *                 except BaseException as ex:
 *                     if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                         fut.set_exception(ex)
 *                 else:
 */
          }
        }

        /* "uvloop/loop.pyx":923
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     if not fut.cancelled():
 *                         fut.set_exception(ex)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L20;
          }
          __pyx_L19_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_1 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L6_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "uvloop/loop.pyx":916
 *         def callback(result):
 *             if AddrInfo.isinstance(result):
 *                 try:             # <<<<<<<<<<<<<<
 *                     if unpack == 0:
 *                         data = result
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      __pyx_L9_try_end:;
    }

    /* "uvloop/loop.pyx":915
 * 
 *         def callback(result):
 *             if AddrInfo.isinstance(result):             # <<<<<<<<<<<<<<
 *                 try:
 *                     if unpack == 0:
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":930
 *                         fut.set_result(data)
 *             else:
 *                 if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                     fut.set_exception(result)
 * 
 */
  /*else*/ {
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 930, __pyx_L1_error) }
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = (!__pyx_t_2);
    if (__pyx_t_10) {

      /* "uvloop/loop.pyx":931
 *             else:
 *                 if not fut.cancelled():
 *                     fut.set_exception(result)             # <<<<<<<<<<<<<<
 * 
 *         AddrInfoRequest(self, host, port, family, type, proto, flags, callback)
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 931, __pyx_L1_error) }
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 931, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_result};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 931, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":930
 *                         fut.set_result(data)
 *             else:
 *                 if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                     fut.set_exception(result)
 * 
 */
    }
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":914
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if AddrInfo.isinstance(result):
 *                 try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("uvloop.loop.Loop._getaddrinfo.callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":894
 *         return poll.is_writing()
 * 
 *     cdef _getaddrinfo(self, object host, object port,             # <<<<<<<<<<<<<<
 *                       int family, int type,
 *                       int proto, int flags,
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__getaddrinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_flags, int __pyx_v_unpack) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *__pyx_cur_scope;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_getaddrinfo", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct___getaddrinfo(__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 894, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_unpack = __pyx_v_unpack;
  __Pyx_INCREF(__pyx_v_host);
  __Pyx_INCREF(__pyx_v_port);

  /* "uvloop/loop.pyx":899
 *                       int unpack):
 * 
 *         if isinstance(port, str):             # <<<<<<<<<<<<<<
 *             port = port.encode()
 *         elif isinstance(port, int):
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_port); 
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":900
 * 
 *         if isinstance(port, str):
 *             port = port.encode()             # <<<<<<<<<<<<<<
 *         elif isinstance(port, int):
 *             port = str(port).encode()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_port, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 900, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_port, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":899
 *                       int unpack):
 * 
 *         if isinstance(port, str):             # <<<<<<<<<<<<<<
 *             port = port.encode()
 *         elif isinstance(port, int):
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":901
 *         if isinstance(port, str):
 *             port = port.encode()
 *         elif isinstance(port, int):             # <<<<<<<<<<<<<<
 *             port = str(port).encode()
 *         if port is not None and not isinstance(port, bytes):
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_port); 
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":902
 *             port = port.encode()
 *         elif isinstance(port, int):
 *             port = str(port).encode()             # <<<<<<<<<<<<<<
 *         if port is not None and not isinstance(port, bytes):
 *             raise TypeError('port must be a str, bytes or int')
 */
    __pyx_t_2 = __Pyx_PyObject_Unicode(__pyx_v_port); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_AsEncodedString(((PyObject*)__pyx_t_2), NULL, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_port, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":901
 *         if isinstance(port, str):
 *             port = port.encode()
 *         elif isinstance(port, int):             # <<<<<<<<<<<<<<
 *             port = str(port).encode()
 *         if port is not None and not isinstance(port, bytes):
 */
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":903
 *         elif isinstance(port, int):
 *             port = str(port).encode()
 *         if port is not None and not isinstance(port, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError('port must be a str, bytes or int')
 * 
 */
  __pyx_t_6 = (__pyx_v_port != Py_None);
  if (__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_6 = PyBytes_Check(__pyx_v_port); 
  __pyx_t_7 = (!__pyx_t_6);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":904
 *             port = str(port).encode()
 *         if port is not None and not isinstance(port, bytes):
 *             raise TypeError('port must be a str, bytes or int')             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(host, str):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 904, __pyx_L1_error)

    /* "uvloop/loop.pyx":903
 *         elif isinstance(port, int):
 *             port = str(port).encode()
 *         if port is not None and not isinstance(port, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError('port must be a str, bytes or int')
 * 
 */
  }

  /* "uvloop/loop.pyx":906
 *             raise TypeError('port must be a str, bytes or int')
 * 
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             host = host.encode('idna')
 *         if host is not None:
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_host); 
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":907
 * 
 *         if isinstance(host, str):
 *             host = host.encode('idna')             # <<<<<<<<<<<<<<
 *         if host is not None:
 *             if not isinstance(host, bytes):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_host, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_idna};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_host, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":906
 *             raise TypeError('port must be a str, bytes or int')
 * 
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             host = host.encode('idna')
 *         if host is not None:
 */
  }

  /* "uvloop/loop.pyx":908
 *         if isinstance(host, str):
 *             host = host.encode('idna')
 *         if host is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(host, bytes):
 *                 raise TypeError('host must be a str or bytes')
 */
  __pyx_t_1 = (__pyx_v_host != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":909
 *             host = host.encode('idna')
 *         if host is not None:
 *             if not isinstance(host, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError('host must be a str or bytes')
 * 
 */
    __pyx_t_1 = PyBytes_Check(__pyx_v_host); 
    __pyx_t_7 = (!__pyx_t_1);
    if (unlikely(__pyx_t_7)) {

      /* "uvloop/loop.pyx":910
 *         if host is not None:
 *             if not isinstance(host, bytes):
 *                 raise TypeError('host must be a str or bytes')             # <<<<<<<<<<<<<<
 * 
 *         fut = self._new_future()
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 910, __pyx_L1_error)

      /* "uvloop/loop.pyx":909
 *             host = host.encode('idna')
 *         if host is not None:
 *             if not isinstance(host, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError('host must be a str or bytes')
 * 
 */
    }

    /* "uvloop/loop.pyx":908
 *         if isinstance(host, str):
 *             host = host.encode('idna')
 *         if host is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(host, bytes):
 *                 raise TypeError('host must be a str or bytes')
 */
  }

  /* "uvloop/loop.pyx":912
 *                 raise TypeError('host must be a str or bytes')
 * 
 *         fut = self._new_future()             # <<<<<<<<<<<<<<
 * 
 *         def callback(result):
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_fut = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":914
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if AddrInfo.isinstance(result):
 *                 try:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_12_getaddrinfo_1callback, 0, __pyx_n_s_Loop__getaddrinfo_locals_callbac, ((PyObject*)__pyx_cur_scope), __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_callback = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":933
 *                     fut.set_exception(result)
 * 
 *         AddrInfoRequest(self, host, port, family, type, proto, flags, callback)             # <<<<<<<<<<<<<<
 *         return fut
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_family); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_proto); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyTuple_New(8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 933, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_host);
  __Pyx_GIVEREF(__pyx_v_host);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_host)) __PYX_ERR(0, 933, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_port);
  __Pyx_GIVEREF(__pyx_v_port);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_v_port)) __PYX_ERR(0, 933, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_4)) __PYX_ERR(0, 933, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 6, __pyx_t_8)) __PYX_ERR(0, 933, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 7, __pyx_v_callback)) __PYX_ERR(0, 933, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_AddrInfoRequest), __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/loop.pyx":934
 * 
 *         AddrInfoRequest(self, host, port, family, type, proto, flags, callback)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     cdef _getnameinfo(self, system.sockaddr *addr, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_fut);
  __pyx_r = __pyx_cur_scope->__pyx_v_fut;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":894
 *         return poll.is_writing()
 * 
 *     cdef _getaddrinfo(self, object host, object port,             # <<<<<<<<<<<<<<
 *                       int family, int type,
 *                       int proto, int flags,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.Loop._getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_XDECREF(__pyx_v_host);
  __Pyx_XDECREF(__pyx_v_port);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":940
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if isinstance(result, tuple):
 *                 fut.set_result(result)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_12_getnameinfo_1callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_12_getnameinfo_1callback = {"callback", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_12_getnameinfo_1callback, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_12_getnameinfo_1callback(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_result = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("callback (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_result)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 940, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "callback") < 0)) __PYX_ERR(0, 940, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_result = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("callback", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 940, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop._getnameinfo.callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_12_getnameinfo_callback(__pyx_self, __pyx_v_result);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_12_getnameinfo_callback(PyObject *__pyx_self, PyObject *__pyx_v_result) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *__pyx_cur_scope;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("callback", 1);
  __pyx_outer_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "uvloop/loop.pyx":941
 * 
 *         def callback(result):
 *             if isinstance(result, tuple):             # <<<<<<<<<<<<<<
 *                 fut.set_result(result)
 *             else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_result); 
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":942
 *         def callback(result):
 *             if isinstance(result, tuple):
 *                 fut.set_result(result)             # <<<<<<<<<<<<<<
 *             else:
 *                 fut.set_exception(result)
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 942, __pyx_L1_error) }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_result};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 942, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":941
 * 
 *         def callback(result):
 *             if isinstance(result, tuple):             # <<<<<<<<<<<<<<
 *                 fut.set_result(result)
 *             else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":944
 *                 fut.set_result(result)
 *             else:
 *                 fut.set_exception(result)             # <<<<<<<<<<<<<<
 * 
 *         nr = NameInfoRequest(self, callback)
 */
  /*else*/ {
    if (unlikely(!__pyx_cur_scope->__pyx_v_fut)) { __Pyx_RaiseClosureNameError("fut"); __PYX_ERR(0, 944, __pyx_L1_error) }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 944, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_result};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 944, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":940
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if isinstance(result, tuple):
 *                 fut.set_result(result)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop._getnameinfo.callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":936
 *         return fut
 * 
 *     cdef _getnameinfo(self, system.sockaddr *addr, int flags):             # <<<<<<<<<<<<<<
 *         cdef NameInfoRequest nr
 *         fut = self._new_future()
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__getnameinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, struct sockaddr *__pyx_v_addr, int __pyx_v_flags) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *__pyx_cur_scope;
  struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_nr = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_getnameinfo", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_1__getnameinfo(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 936, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }

  /* "uvloop/loop.pyx":938
 *     cdef _getnameinfo(self, system.sockaddr *addr, int flags):
 *         cdef NameInfoRequest nr
 *         fut = self._new_future()             # <<<<<<<<<<<<<<
 * 
 *         def callback(result):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":940
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if isinstance(result, tuple):
 *                 fut.set_result(result)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_12_getnameinfo_1callback, 0, __pyx_n_s_Loop__getnameinfo_locals_callbac, ((PyObject*)__pyx_cur_scope), __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_callback = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":946
 *                 fut.set_exception(result)
 * 
 *         nr = NameInfoRequest(self, callback)             # <<<<<<<<<<<<<<
 *         nr.query(addr, flags)
 *         return fut
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 946, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_callback)) __PYX_ERR(0, 946, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_NameInfoRequest), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nr = ((struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":947
 * 
 *         nr = NameInfoRequest(self, callback)
 *         nr.query(addr, flags)             # <<<<<<<<<<<<<<
 *         return fut
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_NameInfoRequest *)__pyx_v_nr->__pyx_base.__pyx_vtab)->query(__pyx_v_nr, __pyx_v_addr, __pyx_v_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":948
 *         nr = NameInfoRequest(self, callback)
 *         nr.query(addr, flags)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     cdef _sock_recv(self, fut, sock, n):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_fut);
  __pyx_r = __pyx_cur_scope->__pyx_v_fut;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":936
 *         return fut
 * 
 *     cdef _getnameinfo(self, system.sockaddr *addr, int flags):             # <<<<<<<<<<<<<<
 *         cdef NameInfoRequest nr
 *         fut = self._new_future()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._getnameinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_nr);
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":950
 *         return fut
 * 
 *     cdef _sock_recv(self, fut, sock, n):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_recv(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_n) {
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sock_recv", 1);

  /* "uvloop/loop.pyx":951
 * 
 *     cdef _sock_recv(self, fut, sock, n):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":952
 *     cdef _sock_recv(self, fut, sock, n):
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 952, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 952, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 952, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/loop.pyx":954
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_recv is called on a cancelled Future')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 954, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 954, __pyx_L1_error)

      /* "uvloop/loop.pyx":952
 *     cdef _sock_recv(self, fut, sock, n):
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(
 */
    }

    /* "uvloop/loop.pyx":957
 *                     f'_sock_recv is called on a cancelled Future')
 * 
 *             if not self._has_reader(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a reader '
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_has_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 957, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_1);
    if (unlikely(__pyx_t_6)) {

      /* "uvloop/loop.pyx":959
 *             if not self._has_reader(sock):
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a reader '             # <<<<<<<<<<<<<<
 *                     f'in the _sock_recv callback')
 * 
 */
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 959, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = 0;
      __pyx_t_8 = 127;
      __Pyx_INCREF(__pyx_kp_u_socket);
      __pyx_t_7 += 7;
      __Pyx_GIVEREF(__pyx_kp_u_socket);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_socket);
      __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_sock), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 959, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_does_not_have_a_reader_in_the);
      __pyx_t_7 += 50;
      __Pyx_GIVEREF(__pyx_kp_u_does_not_have_a_reader_in_the);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_does_not_have_a_reader_in_the);
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 959, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":958
 * 
 *             if not self._has_reader(sock):
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'socket {sock!r} does not have a reader '
 *                     f'in the _sock_recv callback')
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 958, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 958, __pyx_L1_error)

      /* "uvloop/loop.pyx":957
 *                     f'_sock_recv is called on a cancelled Future')
 * 
 *             if not self._has_reader(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a reader '
 */
    }

    /* "uvloop/loop.pyx":951
 * 
 *     cdef _sock_recv(self, fut, sock, n):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 */
  }

  /* "uvloop/loop.pyx":962
 *                     f'in the _sock_recv callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             data = sock.recv(n)
 *         except (BlockingIOError, InterruptedError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "uvloop/loop.pyx":963
 * 
 *         try:
 *             data = sock.recv(n)             # <<<<<<<<<<<<<<
 *         except (BlockingIOError, InterruptedError):
 *             # No need to re-add the reader, let's just wait until
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_recv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 963, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_n};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 963, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_v_data = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":962
 *                     f'in the _sock_recv callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             data = sock.recv(n)
 *         except (BlockingIOError, InterruptedError):
 */
    }

    /* "uvloop/loop.pyx":974
 *             self._remove_reader(sock)
 *         else:
 *             fut.set_result(data)             # <<<<<<<<<<<<<<
 *             self._remove_reader(sock)
 * 
 */
    /*else:*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 974, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_data};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 974, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":975
 *         else:
 *             fut.set_result(data)
 *             self._remove_reader(sock)             # <<<<<<<<<<<<<<
 * 
 *     cdef _sock_recv_into(self, fut, sock, buf):
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":964
 *         try:
 *             data = sock.recv(n)
 *         except (BlockingIOError, InterruptedError):             # <<<<<<<<<<<<<<
 *             # No need to re-add the reader, let's just wait until
 *             # the poll handler calls this callback again.
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 964, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 964, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_2, __pyx_t_12, __pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0;
    if (__pyx_t_14) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L7_exception_handled;
    }

    /* "uvloop/loop.pyx":968
 *             # the poll handler calls this callback again.
 *             pass
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as exc:
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 968, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);

      /* "uvloop/loop.pyx":969
 *             pass
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 */
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
      __PYX_ERR(0, 969, __pyx_L8_except_error)
    }

    /* "uvloop/loop.pyx":970
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 970, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/loop.pyx":971
 *             raise
 *         except BaseException as exc:
 *             fut.set_exception(exc)             # <<<<<<<<<<<<<<
 *             self._remove_reader(sock)
 *         else:
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 971, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_15 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_v_exc};
          __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 971, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "uvloop/loop.pyx":972
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)             # <<<<<<<<<<<<<<
 *         else:
 *             fut.set_result(data)
 */
        __pyx_t_13 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 972, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }

      /* "uvloop/loop.pyx":970
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L20;
        }
        __pyx_L19_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __pyx_t_14 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
          }
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
          goto __pyx_L8_except_error;
        }
        __pyx_L20:;
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L7_exception_handled;
    }
    goto __pyx_L8_except_error;

    /* "uvloop/loop.pyx":962
 *                     f'in the _sock_recv callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             data = sock.recv(n)
 *         except (BlockingIOError, InterruptedError):
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L7_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    __pyx_L11_try_end:;
  }

  /* "uvloop/loop.pyx":950
 *         return fut
 * 
 *     cdef _sock_recv(self, fut, sock, n):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":977
 *             self._remove_reader(sock)
 * 
 *     cdef _sock_recv_into(self, fut, sock, buf):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_recv_into(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_buf) {
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sock_recv_into", 1);

  /* "uvloop/loop.pyx":978
 * 
 *     cdef _sock_recv_into(self, fut, sock, buf):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":979
 *     cdef _sock_recv_into(self, fut, sock, buf):
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 979, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 979, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 979, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/loop.pyx":981
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_recv_into is called on a cancelled Future')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 981, __pyx_L1_error)

      /* "uvloop/loop.pyx":979
 *     cdef _sock_recv_into(self, fut, sock, buf):
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(
 */
    }

    /* "uvloop/loop.pyx":984
 *                     f'_sock_recv_into is called on a cancelled Future')
 * 
 *             if not self._has_reader(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a reader '
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_has_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_1);
    if (unlikely(__pyx_t_6)) {

      /* "uvloop/loop.pyx":986
 *             if not self._has_reader(sock):
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a reader '             # <<<<<<<<<<<<<<
 *                     f'in the _sock_recv_into callback')
 * 
 */
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = 0;
      __pyx_t_8 = 127;
      __Pyx_INCREF(__pyx_kp_u_socket);
      __pyx_t_7 += 7;
      __Pyx_GIVEREF(__pyx_kp_u_socket);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_socket);
      __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_sock), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_does_not_have_a_reader_in_the_2);
      __pyx_t_7 += 55;
      __Pyx_GIVEREF(__pyx_kp_u_does_not_have_a_reader_in_the_2);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_does_not_have_a_reader_in_the_2);
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":985
 * 
 *             if not self._has_reader(sock):
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'socket {sock!r} does not have a reader '
 *                     f'in the _sock_recv_into callback')
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 985, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 985, __pyx_L1_error)

      /* "uvloop/loop.pyx":984
 *                     f'_sock_recv_into is called on a cancelled Future')
 * 
 *             if not self._has_reader(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a reader '
 */
    }

    /* "uvloop/loop.pyx":978
 * 
 *     cdef _sock_recv_into(self, fut, sock, buf):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 */
  }

  /* "uvloop/loop.pyx":989
 *                     f'in the _sock_recv_into callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             data = sock.recv_into(buf)
 *         except (BlockingIOError, InterruptedError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "uvloop/loop.pyx":990
 * 
 *         try:
 *             data = sock.recv_into(buf)             # <<<<<<<<<<<<<<
 *         except (BlockingIOError, InterruptedError):
 *             # No need to re-add the reader, let's just wait until
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_recv_into); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_buf};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_v_data = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":989
 *                     f'in the _sock_recv_into callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             data = sock.recv_into(buf)
 *         except (BlockingIOError, InterruptedError):
 */
    }

    /* "uvloop/loop.pyx":1001
 *             self._remove_reader(sock)
 *         else:
 *             fut.set_result(data)             # <<<<<<<<<<<<<<
 *             self._remove_reader(sock)
 * 
 */
    /*else:*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1001, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_data};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1001, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":1002
 *         else:
 *             fut.set_result(data)
 *             self._remove_reader(sock)             # <<<<<<<<<<<<<<
 * 
 *     cdef _sock_sendall(self, fut, sock, data):
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1002, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":991
 *         try:
 *             data = sock.recv_into(buf)
 *         except (BlockingIOError, InterruptedError):             # <<<<<<<<<<<<<<
 *             # No need to re-add the reader, let's just wait until
 *             # the poll handler calls this callback again.
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 991, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 991, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_2, __pyx_t_12, __pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0;
    if (__pyx_t_14) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L7_exception_handled;
    }

    /* "uvloop/loop.pyx":995
 *             # the poll handler calls this callback again.
 *             pass
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as exc:
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_recv_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 995, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);

      /* "uvloop/loop.pyx":996
 *             pass
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 */
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
      __PYX_ERR(0, 996, __pyx_L8_except_error)
    }

    /* "uvloop/loop.pyx":997
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_recv_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 997, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/loop.pyx":998
 *             raise
 *         except BaseException as exc:
 *             fut.set_exception(exc)             # <<<<<<<<<<<<<<
 *             self._remove_reader(sock)
 *         else:
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 998, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_15 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_v_exc};
          __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 998, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "uvloop/loop.pyx":999
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)             # <<<<<<<<<<<<<<
 *         else:
 *             fut.set_result(data)
 */
        __pyx_t_13 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 999, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }

      /* "uvloop/loop.pyx":997
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L20;
        }
        __pyx_L19_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __pyx_t_14 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
          }
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
          goto __pyx_L8_except_error;
        }
        __pyx_L20:;
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L7_exception_handled;
    }
    goto __pyx_L8_except_error;

    /* "uvloop/loop.pyx":989
 *                     f'in the _sock_recv_into callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             data = sock.recv_into(buf)
 *         except (BlockingIOError, InterruptedError):
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L7_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    __pyx_L11_try_end:;
  }

  /* "uvloop/loop.pyx":977
 *             self._remove_reader(sock)
 * 
 *     cdef _sock_recv_into(self, fut, sock, buf):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_recv_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1004
 *             self._remove_reader(sock)
 * 
 *     cdef _sock_sendall(self, fut, sock, data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Handle handle
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_sendall(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_data) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  int __pyx_v_n;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sock_sendall", 0);
  __Pyx_INCREF(__pyx_v_data);

  /* "uvloop/loop.pyx":1009
 *             int n
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1010
 * 
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/loop.pyx":1012
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_sendall is called on a cancelled Future')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1012, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1012, __pyx_L1_error)

      /* "uvloop/loop.pyx":1010
 * 
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(
 */
    }

    /* "uvloop/loop.pyx":1015
 *                     f'_sock_sendall is called on a cancelled Future')
 * 
 *             if not self._has_writer(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a writer '
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_has_writer(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_1);
    if (unlikely(__pyx_t_6)) {

      /* "uvloop/loop.pyx":1017
 *             if not self._has_writer(sock):
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a writer '             # <<<<<<<<<<<<<<
 *                     f'in the _sock_sendall callback')
 * 
 */
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = 0;
      __pyx_t_8 = 127;
      __Pyx_INCREF(__pyx_kp_u_socket);
      __pyx_t_7 += 7;
      __Pyx_GIVEREF(__pyx_kp_u_socket);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_socket);
      __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_sock), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_does_not_have_a_writer_in_the);
      __pyx_t_7 += 53;
      __Pyx_GIVEREF(__pyx_kp_u_does_not_have_a_writer_in_the);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_does_not_have_a_writer_in_the);
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":1016
 * 
 *             if not self._has_writer(sock):
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'socket {sock!r} does not have a writer '
 *                     f'in the _sock_sendall callback')
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1016, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1016, __pyx_L1_error)

      /* "uvloop/loop.pyx":1015
 *                     f'_sock_sendall is called on a cancelled Future')
 * 
 *             if not self._has_writer(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a writer '
 */
    }

    /* "uvloop/loop.pyx":1009
 *             int n
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 */
  }

  /* "uvloop/loop.pyx":1020
 *                     f'in the _sock_sendall callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             n = sock.send(data)
 *         except (BlockingIOError, InterruptedError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "uvloop/loop.pyx":1021
 * 
 *         try:
 *             n = sock.send(data)             # <<<<<<<<<<<<<<
 *         except (BlockingIOError, InterruptedError):
 *             # Try next time.
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_send); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1021, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_data};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1021, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_n = __pyx_t_12;

      /* "uvloop/loop.pyx":1020
 *                     f'in the _sock_sendall callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             n = sock.send(data)
 *         except (BlockingIOError, InterruptedError):
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1022
 *         try:
 *             n = sock.send(data)
 *         except (BlockingIOError, InterruptedError):             # <<<<<<<<<<<<<<
 *             # Try next time.
 *             return
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1022, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1022, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_12 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_2, __pyx_t_13, __pyx_t_14);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0;
    if (__pyx_t_12) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 1022, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);

      /* "uvloop/loop.pyx":1024
 *         except (BlockingIOError, InterruptedError):
 *             # Try next time.
 *             return             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_except_return;
    }

    /* "uvloop/loop.pyx":1025
 *             # Try next time.
 *             return
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as exc:
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 1025, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "uvloop/loop.pyx":1026
 *             return
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 */
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0; 
      __PYX_ERR(0, 1026, __pyx_L8_except_error)
    }

    /* "uvloop/loop.pyx":1027
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_writer(sock)
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 1027, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/loop.pyx":1028
 *             raise
 *         except BaseException as exc:
 *             fut.set_exception(exc)             # <<<<<<<<<<<<<<
 *             self._remove_writer(sock)
 *             return
 */
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1028, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_15 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_v_exc};
          __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1028, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

        /* "uvloop/loop.pyx":1029
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 *             self._remove_writer(sock)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
        __pyx_t_14 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_writer(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1029, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

        /* "uvloop/loop.pyx":1030
 *             fut.set_exception(exc)
 *             self._remove_writer(sock)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._remove_writer(sock)
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L20_return;
      }

      /* "uvloop/loop.pyx":1027
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_writer(sock)
 */
      /*finally:*/ {
        __pyx_L21_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __pyx_t_12 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
          }
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
          goto __pyx_L8_except_error;
        }
        __pyx_L20_return: {
          __pyx_t_23 = __pyx_r;
          __pyx_r = 0;
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          __pyx_r = __pyx_t_23;
          __pyx_t_23 = 0;
          goto __pyx_L9_except_return;
        }
      }
    }
    goto __pyx_L8_except_error;

    /* "uvloop/loop.pyx":1020
 *                     f'in the _sock_sendall callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             n = sock.send(data)
 *         except (BlockingIOError, InterruptedError):
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L9_except_return:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L0;
    __pyx_L11_try_end:;
  }

  /* "uvloop/loop.pyx":1032
 *             return
 * 
 *         self._remove_writer(sock)             # <<<<<<<<<<<<<<
 * 
 *         if n == len(data):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_writer(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1034
 *         self._remove_writer(sock)
 * 
 *         if n == len(data):             # <<<<<<<<<<<<<<
 *             fut.set_result(None)
 *         else:
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1034, __pyx_L1_error)
  __pyx_t_6 = (__pyx_v_n == __pyx_t_7);
  if (__pyx_t_6) {

    /* "uvloop/loop.pyx":1035
 * 
 *         if n == len(data):
 *             fut.set_result(None)             # <<<<<<<<<<<<<<
 *         else:
 *             if n:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, Py_None};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1035, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":1034
 *         self._remove_writer(sock)
 * 
 *         if n == len(data):             # <<<<<<<<<<<<<<
 *             fut.set_result(None)
 *         else:
 */
    goto __pyx_L27;
  }

  /* "uvloop/loop.pyx":1037
 *             fut.set_result(None)
 *         else:
 *             if n:             # <<<<<<<<<<<<<<
 *                 if not isinstance(data, memoryview):
 *                     data = memoryview(data)
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_n != 0);
    if (__pyx_t_6) {

      /* "uvloop/loop.pyx":1038
 *         else:
 *             if n:
 *                 if not isinstance(data, memoryview):             # <<<<<<<<<<<<<<
 *                     data = memoryview(data)
 *                 data = data[n:]
 */
      __pyx_t_6 = PyMemoryView_Check(__pyx_v_data); 
      __pyx_t_1 = (!__pyx_t_6);
      if (__pyx_t_1) {

        /* "uvloop/loop.pyx":1039
 *             if n:
 *                 if not isinstance(data, memoryview):
 *                     data = memoryview(data)             # <<<<<<<<<<<<<<
 *                 data = data[n:]
 * 
 */
        __pyx_t_2 = PyMemoryView_FromObject(__pyx_v_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":1038
 *         else:
 *             if n:
 *                 if not isinstance(data, memoryview):             # <<<<<<<<<<<<<<
 *                     data = memoryview(data)
 *                 data = data[n:]
 */
      }

      /* "uvloop/loop.pyx":1040
 *                 if not isinstance(data, memoryview):
 *                     data = memoryview(data)
 *                 data = data[n:]             # <<<<<<<<<<<<<<
 * 
 *             handle = new_MethodHandle3(
 */
      __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_data, __pyx_v_n, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":1037
 *             fut.set_result(None)
 *         else:
 *             if n:             # <<<<<<<<<<<<<<
 *                 if not isinstance(data, memoryview):
 *                     data = memoryview(data)
 */
    }

    /* "uvloop/loop.pyx":1042
 *                 data = data[n:]
 * 
 *             handle = new_MethodHandle3(             # <<<<<<<<<<<<<<
 *                 self,
 *                 "Loop._sock_sendall",
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_new_MethodHandle3(__pyx_v_self, __pyx_kp_u_Loop__sock_sendall, ((__pyx_t_6uvloop_4loop_method3_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_sock_sendall), Py_None, ((PyObject *)__pyx_v_self), __pyx_v_fut, __pyx_v_sock, __pyx_v_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 1042, __pyx_L1_error)
    __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":1050
 *                 fut, sock, data)
 * 
 *             self._add_writer(sock, handle)             # <<<<<<<<<<<<<<
 * 
 *     cdef _sock_accept(self, fut, sock):
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_add_writer(__pyx_v_self, __pyx_v_sock, __pyx_v_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1050, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L27:;

  /* "uvloop/loop.pyx":1004
 *             self._remove_reader(sock)
 * 
 *     cdef _sock_sendall(self, fut, sock, data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Handle handle
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1052
 *             self._add_writer(sock, handle)
 * 
 *     cdef _sock_accept(self, fut, sock):             # <<<<<<<<<<<<<<
 *         try:
 *             conn, address = sock.accept()
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_accept(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock) {
  PyObject *__pyx_v_conn = NULL;
  PyObject *__pyx_v_address = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sock_accept", 1);

  /* "uvloop/loop.pyx":1053
 * 
 *     cdef _sock_accept(self, fut, sock):
 *         try:             # <<<<<<<<<<<<<<
 *             conn, address = sock.accept()
 *             conn.setblocking(False)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/loop.pyx":1054
 *     cdef _sock_accept(self, fut, sock):
 *         try:
 *             conn, address = sock.accept()             # <<<<<<<<<<<<<<
 *             conn.setblocking(False)
 *         except (BlockingIOError, InterruptedError):
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_accept); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1054, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1054, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1054, __pyx_L3_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1054, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1054, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1054, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
        index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 1054, __pyx_L3_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1054, __pyx_L3_error)
        __pyx_L10_unpacking_done:;
      }
      __pyx_v_conn = __pyx_t_5;
      __pyx_t_5 = 0;
      __pyx_v_address = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":1055
 *         try:
 *             conn, address = sock.accept()
 *             conn.setblocking(False)             # <<<<<<<<<<<<<<
 *         except (BlockingIOError, InterruptedError):
 *             # There is an active reader for _sock_accept, so
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_conn, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1055, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_False};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1055, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1053
 * 
 *     cdef _sock_accept(self, fut, sock):
 *         try:             # <<<<<<<<<<<<<<
 *             conn, address = sock.accept()
 *             conn.setblocking(False)
 */
    }

    /* "uvloop/loop.pyx":1066
 *             self._remove_reader(sock)
 *         else:
 *             fut.set_result((conn, address))             # <<<<<<<<<<<<<<
 *             self._remove_reader(sock)
 * 
 */
    /*else:*/ {
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1066, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1066, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_conn);
      __Pyx_GIVEREF(__pyx_v_conn);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_conn)) __PYX_ERR(0, 1066, __pyx_L5_except_error);
      __Pyx_INCREF(__pyx_v_address);
      __Pyx_GIVEREF(__pyx_v_address);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_address)) __PYX_ERR(0, 1066, __pyx_L5_except_error);
      __pyx_t_8 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_5};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1066, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1067
 *         else:
 *             fut.set_result((conn, address))
 *             self._remove_reader(sock)             # <<<<<<<<<<<<<<
 * 
 *     cdef _sock_connect(self, sock, address):
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1067, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/loop.pyx":1056
 *             conn, address = sock.accept()
 *             conn.setblocking(False)
 *         except (BlockingIOError, InterruptedError):             # <<<<<<<<<<<<<<
 *             # There is an active reader for _sock_accept, so
 *             # do nothing, it will be called again.
 */
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1056, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1056, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_4, __pyx_t_8, __pyx_t_10);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_6, __pyx_t_5);
    __pyx_t_4 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0;
    if (__pyx_t_11) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }

    /* "uvloop/loop.pyx":1060
 *             # do nothing, it will be called again.
 *             pass
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as exc:
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 1060, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "uvloop/loop.pyx":1061
 *             pass
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 */
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_6, __pyx_t_4);
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_4 = 0; 
      __PYX_ERR(0, 1061, __pyx_L5_except_error)
    }

    /* "uvloop/loop.pyx":1062
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 1062, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_exc = __pyx_t_6;
      /*try:*/ {

        /* "uvloop/loop.pyx":1063
 *             raise
 *         except BaseException as exc:
 *             fut.set_exception(exc)             # <<<<<<<<<<<<<<
 *             self._remove_reader(sock)
 *         else:
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1063, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_12 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_exc};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1063, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/loop.pyx":1064
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)             # <<<<<<<<<<<<<<
 *         else:
 *             fut.set_result((conn, address))
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1064, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }

      /* "uvloop/loop.pyx":1062
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_reader(sock)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L19;
        }
        __pyx_L18_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_11 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L5_except_error;
        }
        __pyx_L19:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":1053
 * 
 *     cdef _sock_accept(self, fut, sock):
 *         try:             # <<<<<<<<<<<<<<
 *             conn, address = sock.accept()
 *             conn.setblocking(False)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":1052
 *             self._add_writer(sock, handle)
 * 
 *     cdef _sock_accept(self, fut, sock):             # <<<<<<<<<<<<<<
 *         try:
 *             conn, address = sock.accept()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_conn);
  __Pyx_XDECREF(__pyx_v_address);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1069
 *             self._remove_reader(sock)
 * 
 *     cdef _sock_connect(self, sock, address):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Handle handle
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_connect(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_address) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_v_fut = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sock_connect", 1);

  /* "uvloop/loop.pyx":1073
 *             Handle handle
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             sock.connect(address)
 *         except (BlockingIOError, InterruptedError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/loop.pyx":1074
 * 
 *         try:
 *             sock.connect(address)             # <<<<<<<<<<<<<<
 *         except (BlockingIOError, InterruptedError):
 *             pass
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_connect); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1074, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_address};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1074, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1073
 *             Handle handle
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             sock.connect(address)
 *         except (BlockingIOError, InterruptedError):
 */
    }

    /* "uvloop/loop.pyx":1078
 *             pass
 *         else:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         fut = _SyncSocketWriterFuture(sock, self)
 */
    /*else:*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":1075
 *         try:
 *             sock.connect(address)
 *         except (BlockingIOError, InterruptedError):             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1075, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1075, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_4, __pyx_t_8, __pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_ErrRestore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
    if (__pyx_t_10) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":1073
 *             Handle handle
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             sock.connect(address)
 *         except (BlockingIOError, InterruptedError):
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  }

  /* "uvloop/loop.pyx":1080
 *             return
 * 
 *         fut = _SyncSocketWriterFuture(sock, self)             # <<<<<<<<<<<<<<
 *         handle = new_MethodHandle3(
 *             self,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_SyncSocketWriterFuture); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_sock, ((PyObject *)__pyx_v_self)};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1080, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_v_fut = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1081
 * 
 *         fut = _SyncSocketWriterFuture(sock, self)
 *         handle = new_MethodHandle3(             # <<<<<<<<<<<<<<
 *             self,
 *             "Loop._sock_connect",
 */
  __pyx_t_6 = __pyx_f_6uvloop_4loop_new_MethodHandle3(__pyx_v_self, __pyx_kp_u_Loop__sock_connect, ((__pyx_t_6uvloop_4loop_method3_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_sock_connect_cb), Py_None, ((PyObject *)__pyx_v_self), __pyx_v_fut, __pyx_v_sock, __pyx_v_address); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 1081, __pyx_L1_error)
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1089
 *             fut, sock, address)
 * 
 *         self._add_writer(sock, handle)             # <<<<<<<<<<<<<<
 *         return fut
 * 
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_add_writer(__pyx_v_self, __pyx_v_sock, __pyx_v_handle); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1090
 * 
 *         self._add_writer(sock, handle)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 *     cdef _sock_connect_cb(self, fut, sock, address):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1069
 *             self._remove_reader(sock)
 * 
 *     cdef _sock_connect(self, sock, address):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Handle handle
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1092
 *         return fut
 * 
 *     cdef _sock_connect_cb(self, fut, sock, address):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_connect_cb(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fut, PyObject *__pyx_v_sock, PyObject *__pyx_v_address) {
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sock_connect_cb", 1);

  /* "uvloop/loop.pyx":1093
 * 
 *     cdef _sock_connect_cb(self, fut, sock, address):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1094
 *     cdef _sock_connect_cb(self, fut, sock, address):
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1094, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1094, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/loop.pyx":1096
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_connect_cb is called on a cancelled Future')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1096, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1096, __pyx_L1_error)

      /* "uvloop/loop.pyx":1094
 *     cdef _sock_connect_cb(self, fut, sock, address):
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():             # <<<<<<<<<<<<<<
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(
 */
    }

    /* "uvloop/loop.pyx":1099
 *                     f'_sock_connect_cb is called on a cancelled Future')
 * 
 *             if not self._has_writer(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a writer '
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_has_writer(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_1);
    if (unlikely(__pyx_t_6)) {

      /* "uvloop/loop.pyx":1101
 *             if not self._has_writer(sock):
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a writer '             # <<<<<<<<<<<<<<
 *                     f'in the _sock_connect_cb callback')
 * 
 */
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = 0;
      __pyx_t_8 = 127;
      __Pyx_INCREF(__pyx_kp_u_socket);
      __pyx_t_7 += 7;
      __Pyx_GIVEREF(__pyx_kp_u_socket);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_socket);
      __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_sock), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_does_not_have_a_writer_in_the_2);
      __pyx_t_7 += 56;
      __Pyx_GIVEREF(__pyx_kp_u_does_not_have_a_writer_in_the_2);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_does_not_have_a_writer_in_the_2);
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":1100
 * 
 *             if not self._has_writer(sock):
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'socket {sock!r} does not have a writer '
 *                     f'in the _sock_connect_cb callback')
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1100, __pyx_L1_error)

      /* "uvloop/loop.pyx":1099
 *                     f'_sock_connect_cb is called on a cancelled Future')
 * 
 *             if not self._has_writer(sock):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f'socket {sock!r} does not have a writer '
 */
    }

    /* "uvloop/loop.pyx":1093
 * 
 *     cdef _sock_connect_cb(self, fut, sock, address):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 */
  }

  /* "uvloop/loop.pyx":1104
 *                     f'in the _sock_connect_cb callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             err = sock.getsockopt(uv.SOL_SOCKET, uv.SO_ERROR)
 *             if err != 0:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "uvloop/loop.pyx":1105
 * 
 *         try:
 *             err = sock.getsockopt(uv.SOL_SOCKET, uv.SO_ERROR)             # <<<<<<<<<<<<<<
 *             if err != 0:
 *                 # Jump to any except clause below.
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_getsockopt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1105, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_int(SOL_SOCKET); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1105, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = __Pyx_PyInt_From_int(SO_ERROR); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1105, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_t_4, __pyx_t_12};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1105, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_v_err = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":1106
 *         try:
 *             err = sock.getsockopt(uv.SOL_SOCKET, uv.SO_ERROR)
 *             if err != 0:             # <<<<<<<<<<<<<<
 *                 # Jump to any except clause below.
 *                 raise OSError(err, 'Connect call failed %s' % (address,))
 */
      __pyx_t_6 = (__Pyx_PyInt_BoolNeObjC(__pyx_v_err, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 1106, __pyx_L6_error)
      if (unlikely(__pyx_t_6)) {

        /* "uvloop/loop.pyx":1108
 *             if err != 0:
 *                 # Jump to any except clause below.
 *                 raise OSError(err, 'Connect call failed %s' % (address,))             # <<<<<<<<<<<<<<
 *         except (BlockingIOError, InterruptedError):
 *             # socket is still registered, the callback will be retried later
 */
        __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_address), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Connect_call_failed, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_err);
        __Pyx_GIVEREF(__pyx_v_err);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_err)) __PYX_ERR(0, 1108, __pyx_L6_error);
        __Pyx_GIVEREF(__pyx_t_3);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L6_error);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 1108, __pyx_L6_error)

        /* "uvloop/loop.pyx":1106
 *         try:
 *             err = sock.getsockopt(uv.SOL_SOCKET, uv.SO_ERROR)
 *             if err != 0:             # <<<<<<<<<<<<<<
 *                 # Jump to any except clause below.
 *                 raise OSError(err, 'Connect call failed %s' % (address,))
 */
      }

      /* "uvloop/loop.pyx":1104
 *                     f'in the _sock_connect_cb callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             err = sock.getsockopt(uv.SOL_SOCKET, uv.SO_ERROR)
 *             if err != 0:
 */
    }

    /* "uvloop/loop.pyx":1118
 *             self._remove_writer(sock)
 *         else:
 *             fut.set_result(None)             # <<<<<<<<<<<<<<
 *             self._remove_writer(sock)
 * 
 */
    /*else:*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1118, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, Py_None};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1118, __pyx_L8_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":1119
 *         else:
 *             fut.set_result(None)
 *             self._remove_writer(sock)             # <<<<<<<<<<<<<<
 * 
 *     cdef _sock_set_reuseport(self, int fd):
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_writer(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1119, __pyx_L8_except_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1109
 *                 # Jump to any except clause below.
 *                 raise OSError(err, 'Connect call failed %s' % (address,))
 *         except (BlockingIOError, InterruptedError):             # <<<<<<<<<<<<<<
 *             # socket is still registered, the callback will be retried later
 *             pass
 */
    __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_2, &__pyx_t_12);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1109, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1109, __pyx_L8_except_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_3, __pyx_t_4, __pyx_t_13);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_ErrRestore(__pyx_t_3, __pyx_t_2, __pyx_t_12);
    __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_12 = 0;
    if (__pyx_t_14) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L7_exception_handled;
    }

    /* "uvloop/loop.pyx":1112
 *             # socket is still registered, the callback will be retried later
 *             pass
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as exc:
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_connect_cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 1112, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":1113
 *             pass
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 */
      __Pyx_GIVEREF(__pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestoreWithState(__pyx_t_12, __pyx_t_2, __pyx_t_3);
      __pyx_t_12 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0; 
      __PYX_ERR(0, 1113, __pyx_L8_except_error)
    }

    /* "uvloop/loop.pyx":1114
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_writer(sock)
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.Loop._sock_connect_cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_12) < 0) __PYX_ERR(0, 1114, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_exc = __pyx_t_2;
      /*try:*/ {

        /* "uvloop/loop.pyx":1115
 *             raise
 *         except BaseException as exc:
 *             fut.set_exception(exc)             # <<<<<<<<<<<<<<
 *             self._remove_writer(sock)
 *         else:
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1115, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_15 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_v_exc};
          __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1115, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "uvloop/loop.pyx":1116
 *         except BaseException as exc:
 *             fut.set_exception(exc)
 *             self._remove_writer(sock)             # <<<<<<<<<<<<<<
 *         else:
 *             fut.set_result(None)
 */
        __pyx_t_13 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_writer(__pyx_v_self, __pyx_v_sock); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1116, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }

      /* "uvloop/loop.pyx":1114
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             fut.set_exception(exc)
 *             self._remove_writer(sock)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L21;
        }
        __pyx_L20_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __pyx_t_14 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
          }
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
          __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
          goto __pyx_L8_except_error;
        }
        __pyx_L21:;
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L7_exception_handled;
    }
    goto __pyx_L8_except_error;

    /* "uvloop/loop.pyx":1104
 *                     f'in the _sock_connect_cb callback')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             err = sock.getsockopt(uv.SOL_SOCKET, uv.SO_ERROR)
 *             if err != 0:
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L7_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    __pyx_L11_try_end:;
  }

  /* "uvloop/loop.pyx":1092
 *         return fut
 * 
 *     cdef _sock_connect_cb(self, fut, sock, address):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if fut.cancelled():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_connect_cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1121
 *             self._remove_writer(sock)
 * 
 *     cdef _sock_set_reuseport(self, int fd):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err = 0
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__sock_set_reuseport(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, int __pyx_v_fd) {
  int __pyx_v_err;
  int __pyx_v_reuseport_flag;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sock_set_reuseport", 1);

  /* "uvloop/loop.pyx":1123
 *     cdef _sock_set_reuseport(self, int fd):
 *         cdef:
 *             int err = 0             # <<<<<<<<<<<<<<
 *             int reuseport_flag = 1
 * 
 */
  __pyx_v_err = 0;

  /* "uvloop/loop.pyx":1124
 *         cdef:
 *             int err = 0
 *             int reuseport_flag = 1             # <<<<<<<<<<<<<<
 * 
 *         err = system.setsockopt(
 */
  __pyx_v_reuseport_flag = 1;

  /* "uvloop/loop.pyx":1126
 *             int reuseport_flag = 1
 * 
 *         err = system.setsockopt(             # <<<<<<<<<<<<<<
 *             fd,
 *             uv.SOL_SOCKET,
 */
  __pyx_v_err = setsockopt(__pyx_v_fd, SOL_SOCKET, __pyx_v_6uvloop_4loop_SO_REUSEPORT, ((char *)(&__pyx_v_reuseport_flag)), (sizeof(__pyx_v_reuseport_flag)));

  /* "uvloop/loop.pyx":1133
 *             sizeof(reuseport_flag))
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(-errno.errno)
 * 
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":1134
 * 
 *         if err < 0:
 *             raise convert_error(-errno.errno)             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_coroutine_debug(self, bint enabled):
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error((-errno)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1134, __pyx_L1_error)

    /* "uvloop/loop.pyx":1133
 *             sizeof(reuseport_flag))
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(-errno.errno)
 * 
 */
  }

  /* "uvloop/loop.pyx":1121
 *             self._remove_writer(sock)
 * 
 *     cdef _sock_set_reuseport(self, int fd):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err = 0
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_set_reuseport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1136
 *             raise convert_error(-errno.errno)
 * 
 *     cdef _set_coroutine_debug(self, bint enabled):             # <<<<<<<<<<<<<<
 *         enabled = bool(enabled)
 *         if self._coroutine_debug_set == enabled:
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__set_coroutine_debug(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, int __pyx_v_enabled) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_coroutine_debug", 1);

  /* "uvloop/loop.pyx":1137
 * 
 *     cdef _set_coroutine_debug(self, bint enabled):
 *         enabled = bool(enabled)             # <<<<<<<<<<<<<<
 *         if self._coroutine_debug_set == enabled:
 *             return
 */
  __pyx_t_1 = __pyx_v_enabled;
  __pyx_v_enabled = (!(!__pyx_t_1));

  /* "uvloop/loop.pyx":1138
 *     cdef _set_coroutine_debug(self, bint enabled):
 *         enabled = bool(enabled)
 *         if self._coroutine_debug_set == enabled:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_coroutine_debug_set == __pyx_v_enabled);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1139
 *         enabled = bool(enabled)
 *         if self._coroutine_debug_set == enabled:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if enabled:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":1138
 *     cdef _set_coroutine_debug(self, bint enabled):
 *         enabled = bool(enabled)
 *         if self._coroutine_debug_set == enabled:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/loop.pyx":1141
 *             return
 * 
 *         if enabled:             # <<<<<<<<<<<<<<
 *             self._coroutine_origin_tracking_saved_depth = (
 *                 sys.get_coroutine_origin_tracking_depth())
 */
  if (__pyx_v_enabled) {

    /* "uvloop/loop.pyx":1143
 *         if enabled:
 *             self._coroutine_origin_tracking_saved_depth = (
 *                 sys.get_coroutine_origin_tracking_depth())             # <<<<<<<<<<<<<<
 *             sys.set_coroutine_origin_tracking_depth(
 *                 DEBUG_STACK_DEPTH)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_coroutine_origin_tracking_de); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1143, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":1142
 * 
 *         if enabled:
 *             self._coroutine_origin_tracking_saved_depth = (             # <<<<<<<<<<<<<<
 *                 sys.get_coroutine_origin_tracking_depth())
 *             sys.set_coroutine_origin_tracking_depth(
 */
    __pyx_v_self->_coroutine_origin_tracking_saved_depth = __pyx_t_6;

    /* "uvloop/loop.pyx":1144
 *             self._coroutine_origin_tracking_saved_depth = (
 *                 sys.get_coroutine_origin_tracking_depth())
 *             sys.set_coroutine_origin_tracking_depth(             # <<<<<<<<<<<<<<
 *                 DEBUG_STACK_DEPTH)
 *         else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_set_coroutine_origin_tracking_de); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1145
 *                 sys.get_coroutine_origin_tracking_depth())
 *             sys.set_coroutine_origin_tracking_depth(
 *                 DEBUG_STACK_DEPTH)             # <<<<<<<<<<<<<<
 *         else:
 *             sys.set_coroutine_origin_tracking_depth(
 */
    __pyx_t_4 = __Pyx_PyInt_From___pyx_anon_enum(__pyx_e_6uvloop_4loop_DEBUG_STACK_DEPTH); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":1141
 *             return
 * 
 *         if enabled:             # <<<<<<<<<<<<<<
 *             self._coroutine_origin_tracking_saved_depth = (
 *                 sys.get_coroutine_origin_tracking_depth())
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":1147
 *                 DEBUG_STACK_DEPTH)
 *         else:
 *             sys.set_coroutine_origin_tracking_depth(             # <<<<<<<<<<<<<<
 *                 self._coroutine_origin_tracking_saved_depth)
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_set_coroutine_origin_tracking_de); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1148
 *         else:
 *             sys.set_coroutine_origin_tracking_depth(
 *                 self._coroutine_origin_tracking_saved_depth)             # <<<<<<<<<<<<<<
 * 
 *         self._coroutine_debug_set = enabled
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_coroutine_origin_tracking_saved_depth); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":1150
 *                 self._coroutine_origin_tracking_saved_depth)
 * 
 *         self._coroutine_debug_set = enabled             # <<<<<<<<<<<<<<
 * 
 *     def _get_backend_id(self):
 */
  __pyx_v_self->_coroutine_debug_set = __pyx_v_enabled;

  /* "uvloop/loop.pyx":1136
 *             raise convert_error(-errno.errno)
 * 
 *     cdef _set_coroutine_debug(self, bint enabled):             # <<<<<<<<<<<<<<
 *         enabled = bool(enabled)
 *         if self._coroutine_debug_set == enabled:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.Loop._set_coroutine_debug", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1152
 *         self._coroutine_debug_set = enabled
 * 
 *     def _get_backend_id(self):             # <<<<<<<<<<<<<<
 *         """This method is used by uvloop tests and is not part of the API."""
 *         return uv.uv_backend_fd(self.uvloop)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_9_get_backend_id(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_8_get_backend_id, "Loop._get_backend_id(self)\nThis method is used by uvloop tests and is not part of the API.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_9_get_backend_id = {"_get_backend_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_9_get_backend_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_8_get_backend_id};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_9_get_backend_id(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_backend_id (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_get_backend_id", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_get_backend_id", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_8_get_backend_id(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_8_get_backend_id(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_backend_id", 1);

  /* "uvloop/loop.pyx":1154
 *     def _get_backend_id(self):
 *         """This method is used by uvloop tests and is not part of the API."""
 *         return uv.uv_backend_fd(self.uvloop)             # <<<<<<<<<<<<<<
 * 
 *     cdef _print_debug_info(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uv_os_fd_t(uv_backend_fd(__pyx_v_self->uvloop)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1152
 *         self._coroutine_debug_set = enabled
 * 
 *     def _get_backend_id(self):             # <<<<<<<<<<<<<<
 *         """This method is used by uvloop tests and is not part of the API."""
 *         return uv.uv_backend_fd(self.uvloop)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._get_backend_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1156
 *         return uv.uv_backend_fd(self.uvloop)
 * 
 *     cdef _print_debug_info(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err
 */

static PyObject *__pyx_f_6uvloop_4loop_4Loop__print_debug_info(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  int __pyx_v_err;
  uv_rusage_t __pyx_v_rusage;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_print_debug_info", 1);

  /* "uvloop/loop.pyx":1161
 *             uv.uv_rusage_t rusage
 * 
 *         err = uv.uv_getrusage(&rusage)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
  __pyx_v_err = uv_getrusage((&__pyx_v_rusage));

  /* "uvloop/loop.pyx":1162
 * 
 *         err = uv.uv_getrusage(&rusage)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":1163
 *         err = uv.uv_getrusage(&rusage)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         # OS
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1163, __pyx_L1_error)

    /* "uvloop/loop.pyx":1162
 * 
 *         err = uv.uv_getrusage(&rusage)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  }

  /* "uvloop/loop.pyx":1167
 *         # OS
 * 
 *         print('---- Process info: -----')             # <<<<<<<<<<<<<<
 *         print('Process memory:            {}'.format(rusage.ru_maxrss))
 *         print('Number of signals:         {}'.format(rusage.ru_nsignals))
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1168
 * 
 *         print('---- Process info: -----')
 *         print('Process memory:            {}'.format(rusage.ru_maxrss))             # <<<<<<<<<<<<<<
 *         print('Number of signals:         {}'.format(rusage.ru_nsignals))
 *         print('')
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Process_memory, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_rusage.ru_maxrss); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1169
 *         print('---- Process info: -----')
 *         print('Process memory:            {}'.format(rusage.ru_maxrss))
 *         print('Number of signals:         {}'.format(rusage.ru_nsignals))             # <<<<<<<<<<<<<<
 *         print('')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Number_of_signals, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_rusage.ru_nsignals); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1170
 *         print('Process memory:            {}'.format(rusage.ru_maxrss))
 *         print('Number of signals:         {}'.format(rusage.ru_nsignals))
 *         print('')             # <<<<<<<<<<<<<<
 * 
 *         # Loop
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1174
 *         # Loop
 * 
 *         print('--- Loop debug info: ---')             # <<<<<<<<<<<<<<
 *         print('Loop time:                 {}'.format(self.time()))
 *         print('Errors logged:             {}'.format(
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1175
 * 
 *         print('--- Loop debug info: ---')
 *         print('Loop time:                 {}'.format(self.time()))             # <<<<<<<<<<<<<<
 *         print('Errors logged:             {}'.format(
 *             self._debug_exception_handler_cnt))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Loop_time, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1176
 *         print('--- Loop debug info: ---')
 *         print('Loop time:                 {}'.format(self.time()))
 *         print('Errors logged:             {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_exception_handler_cnt))
 *         print()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Errors_logged, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/loop.pyx":1177
 *         print('Loop time:                 {}'.format(self.time()))
 *         print('Errors logged:             {}'.format(
 *             self._debug_exception_handler_cnt))             # <<<<<<<<<<<<<<
 *         print()
 *         print('Callback handles:          {: <8} | {}'.format(
 */
  __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_exception_handler_cnt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/loop.pyx":1176
 *         print('--- Loop debug info: ---')
 *         print('Loop time:                 {}'.format(self.time()))
 *         print('Errors logged:             {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_exception_handler_cnt))
 *         print()
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1178
 *         print('Errors logged:             {}'.format(
 *             self._debug_exception_handler_cnt))
 *         print()             # <<<<<<<<<<<<<<
 *         print('Callback handles:          {: <8} | {}'.format(
 *             self._debug_cb_handles_count,
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1179
 *             self._debug_exception_handler_cnt))
 *         print()
 *         print('Callback handles:          {: <8} | {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_cb_handles_count,
 *             self._debug_cb_handles_total))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Callback_handles_8, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1180
 *         print()
 *         print('Callback handles:          {: <8} | {}'.format(
 *             self._debug_cb_handles_count,             # <<<<<<<<<<<<<<
 *             self._debug_cb_handles_total))
 *         print('Timer handles:             {: <8} | {}'.format(
 */
  __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_handles_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1181
 *         print('Callback handles:          {: <8} | {}'.format(
 *             self._debug_cb_handles_count,
 *             self._debug_cb_handles_total))             # <<<<<<<<<<<<<<
 *         print('Timer handles:             {: <8} | {}'.format(
 *             self._debug_cb_timer_handles_count,
 */
  __pyx_t_5 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_handles_total); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_4, __pyx_t_5};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1179
 *             self._debug_exception_handler_cnt))
 *         print()
 *         print('Callback handles:          {: <8} | {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_cb_handles_count,
 *             self._debug_cb_handles_total))
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1182
 *             self._debug_cb_handles_count,
 *             self._debug_cb_handles_total))
 *         print('Timer handles:             {: <8} | {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_cb_timer_handles_count,
 *             self._debug_cb_timer_handles_total))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Timer_handles_8, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/loop.pyx":1183
 *             self._debug_cb_handles_total))
 *         print('Timer handles:             {: <8} | {}'.format(
 *             self._debug_cb_timer_handles_count,             # <<<<<<<<<<<<<<
 *             self._debug_cb_timer_handles_total))
 *         print()
 */
  __pyx_t_5 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_timer_handles_count); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "uvloop/loop.pyx":1184
 *         print('Timer handles:             {: <8} | {}'.format(
 *             self._debug_cb_timer_handles_count,
 *             self._debug_cb_timer_handles_total))             # <<<<<<<<<<<<<<
 *         print()
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_timer_handles_total); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_5, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/loop.pyx":1182
 *             self._debug_cb_handles_count,
 *             self._debug_cb_handles_total))
 *         print('Timer handles:             {: <8} | {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_cb_timer_handles_count,
 *             self._debug_cb_timer_handles_total))
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1185
 *             self._debug_cb_timer_handles_count,
 *             self._debug_cb_timer_handles_total))
 *         print()             # <<<<<<<<<<<<<<
 * 
 *         print('                        alive  | closed  |')
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1187
 *         print()
 * 
 *         print('                        alive  | closed  |')             # <<<<<<<<<<<<<<
 *         print('UVHandles               python | libuv   | total')
 *         print('                        objs   | handles |')
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1188
 * 
 *         print('                        alive  | closed  |')
 *         print('UVHandles               python | libuv   | total')             # <<<<<<<<<<<<<<
 *         print('                        objs   | handles |')
 *         print('-------------------------------+---------+---------')
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1189
 *         print('                        alive  | closed  |')
 *         print('UVHandles               python | libuv   | total')
 *         print('                        objs   | handles |')             # <<<<<<<<<<<<<<
 *         print('-------------------------------+---------+---------')
 *         for name in sorted(self._debug_handles_total):
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1190
 *         print('UVHandles               python | libuv   | total')
 *         print('                        objs   | handles |')
 *         print('-------------------------------+---------+---------')             # <<<<<<<<<<<<<<
 *         for name in sorted(self._debug_handles_total):
 *             print('    {: <18} {: >7} | {: >7} | {: >7}'.format(
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1191
 *         print('                        objs   | handles |')
 *         print('-------------------------------+---------+---------')
 *         for name in sorted(self._debug_handles_total):             # <<<<<<<<<<<<<<
 *             print('    {: <18} {: >7} | {: >7} | {: >7}'.format(
 *                 name,
 */
  __pyx_t_3 = __pyx_v_self->_debug_handles_total;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_8 = PyList_Sort(__pyx_t_2); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1191, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1191, __pyx_L1_error)
      #endif
      if (__pyx_t_9 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_2); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 1191, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":1192
 *         print('-------------------------------+---------+---------')
 *         for name in sorted(self._debug_handles_total):
 *             print('    {: <18} {: >7} | {: >7} | {: >7}'.format(             # <<<<<<<<<<<<<<
 *                 name,
 *                 self._debug_handles_current[name],
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_18_7_7_7, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/loop.pyx":1194
 *             print('    {: <18} {: >7} | {: >7} | {: >7}'.format(
 *                 name,
 *                 self._debug_handles_current[name],             # <<<<<<<<<<<<<<
 *                 self._debug_handles_closed[name],
 *                 self._debug_handles_total[name]))
 */
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_self->_debug_handles_current, __pyx_v_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "uvloop/loop.pyx":1195
 *                 name,
 *                 self._debug_handles_current[name],
 *                 self._debug_handles_closed[name],             # <<<<<<<<<<<<<<
 *                 self._debug_handles_total[name]))
 *         print()
 */
    __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_self->_debug_handles_closed, __pyx_v_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "uvloop/loop.pyx":1196
 *                 self._debug_handles_current[name],
 *                 self._debug_handles_closed[name],
 *                 self._debug_handles_total[name]))             # <<<<<<<<<<<<<<
 *         print()
 * 
 */
    __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_self->_debug_handles_total, __pyx_v_name); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_11, __pyx_v_name, __pyx_t_5, __pyx_t_7, __pyx_t_10};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 4+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "uvloop/loop.pyx":1192
 *         print('-------------------------------+---------+---------')
 *         for name in sorted(self._debug_handles_total):
 *             print('    {: <18} {: >7} | {: >7} | {: >7}'.format(             # <<<<<<<<<<<<<<
 *                 name,
 *                 self._debug_handles_current[name],
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1191
 *         print('                        objs   | handles |')
 *         print('-------------------------------+---------+---------')
 *         for name in sorted(self._debug_handles_total):             # <<<<<<<<<<<<<<
 *             print('    {: <18} {: >7} | {: >7} | {: >7}'.format(
 *                 name,
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1197
 *                 self._debug_handles_closed[name],
 *                 self._debug_handles_total[name]))
 *         print()             # <<<<<<<<<<<<<<
 * 
 *         print('uv_handle_t (current: {}; freed: {}; total: {})'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1199
 *         print()
 * 
 *         print('uv_handle_t (current: {}; freed: {}; total: {})'.format(             # <<<<<<<<<<<<<<
 *             self._debug_uv_handles_total - self._debug_uv_handles_freed,
 *             self._debug_uv_handles_freed,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_uv_handle_t_current_freed_total, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1200
 * 
 *         print('uv_handle_t (current: {}; freed: {}; total: {})'.format(
 *             self._debug_uv_handles_total - self._debug_uv_handles_freed,             # <<<<<<<<<<<<<<
 *             self._debug_uv_handles_freed,
 *             self._debug_uv_handles_total))
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t((__pyx_v_self->_debug_uv_handles_total - __pyx_v_self->_debug_uv_handles_freed)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/loop.pyx":1201
 *         print('uv_handle_t (current: {}; freed: {}; total: {})'.format(
 *             self._debug_uv_handles_total - self._debug_uv_handles_freed,
 *             self._debug_uv_handles_freed,             # <<<<<<<<<<<<<<
 *             self._debug_uv_handles_total))
 *         print()
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_uv_handles_freed); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "uvloop/loop.pyx":1202
 *             self._debug_uv_handles_total - self._debug_uv_handles_freed,
 *             self._debug_uv_handles_freed,
 *             self._debug_uv_handles_total))             # <<<<<<<<<<<<<<
 *         print()
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_uv_handles_total); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_2, __pyx_t_10, __pyx_t_7};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1199
 *         print()
 * 
 *         print('uv_handle_t (current: {}; freed: {}; total: {})'.format(             # <<<<<<<<<<<<<<
 *             self._debug_uv_handles_total - self._debug_uv_handles_freed,
 *             self._debug_uv_handles_freed,
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1203
 *             self._debug_uv_handles_freed,
 *             self._debug_uv_handles_total))
 *         print()             # <<<<<<<<<<<<<<
 * 
 *         print('--- Streams debug info: ---')
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1205
 *         print()
 * 
 *         print('--- Streams debug info: ---')             # <<<<<<<<<<<<<<
 *         print('Write errors:              {}'.format(
 *             self._debug_stream_write_errors_total))
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1206
 * 
 *         print('--- Streams debug info: ---')
 *         print('Write errors:              {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_errors_total))
 *         print('Write without poll:        {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Write_errors, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1207
 *         print('--- Streams debug info: ---')
 *         print('Write errors:              {}'.format(
 *             self._debug_stream_write_errors_total))             # <<<<<<<<<<<<<<
 *         print('Write without poll:        {}'.format(
 *             self._debug_stream_write_tries))
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_errors_total); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_7};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1206
 * 
 *         print('--- Streams debug info: ---')
 *         print('Write errors:              {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_errors_total))
 *         print('Write without poll:        {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1208
 *         print('Write errors:              {}'.format(
 *             self._debug_stream_write_errors_total))
 *         print('Write without poll:        {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_tries))
 *         print('Write contexts:            {: <8} | {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Write_without_poll, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1209
 *             self._debug_stream_write_errors_total))
 *         print('Write without poll:        {}'.format(
 *             self._debug_stream_write_tries))             # <<<<<<<<<<<<<<
 *         print('Write contexts:            {: <8} | {}'.format(
 *             self._debug_stream_write_ctx_cnt,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_tries); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_10 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_7};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1208
 *         print('Write errors:              {}'.format(
 *             self._debug_stream_write_errors_total))
 *         print('Write without poll:        {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_tries))
 *         print('Write contexts:            {: <8} | {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1210
 *         print('Write without poll:        {}'.format(
 *             self._debug_stream_write_tries))
 *         print('Write contexts:            {: <8} | {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_ctx_cnt,
 *             self._debug_stream_write_ctx_total))
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Write_contexts_8, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1211
 *             self._debug_stream_write_tries))
 *         print('Write contexts:            {: <8} | {}'.format(
 *             self._debug_stream_write_ctx_cnt,             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_ctx_total))
 *         print('Write failed callbacks:    {}'.format(
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_ctx_cnt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "uvloop/loop.pyx":1212
 *         print('Write contexts:            {: <8} | {}'.format(
 *             self._debug_stream_write_ctx_cnt,
 *             self._debug_stream_write_ctx_total))             # <<<<<<<<<<<<<<
 *         print('Write failed callbacks:    {}'.format(
 *             self._debug_stream_write_cb_errors_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_ctx_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_7, __pyx_t_10};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1210
 *         print('Write without poll:        {}'.format(
 *             self._debug_stream_write_tries))
 *         print('Write contexts:            {: <8} | {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_ctx_cnt,
 *             self._debug_stream_write_ctx_total))
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1213
 *             self._debug_stream_write_ctx_cnt,
 *             self._debug_stream_write_ctx_total))
 *         print('Write failed callbacks:    {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_cb_errors_total))
 *         print()
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Write_failed_callbacks, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1214
 *             self._debug_stream_write_ctx_total))
 *         print('Write failed callbacks:    {}'.format(
 *             self._debug_stream_write_cb_errors_total))             # <<<<<<<<<<<<<<
 *         print()
 *         print('Read errors:               {}'.format(
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_cb_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1213
 *             self._debug_stream_write_ctx_cnt,
 *             self._debug_stream_write_ctx_total))
 *         print('Write failed callbacks:    {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_write_cb_errors_total))
 *         print()
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1215
 *         print('Write failed callbacks:    {}'.format(
 *             self._debug_stream_write_cb_errors_total))
 *         print()             # <<<<<<<<<<<<<<
 *         print('Read errors:               {}'.format(
 *             self._debug_stream_read_errors_total))
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1216
 *             self._debug_stream_write_cb_errors_total))
 *         print()
 *         print('Read errors:               {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_errors_total))
 *         print('Read callbacks:            {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Read_errors, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1217
 *         print()
 *         print('Read errors:               {}'.format(
 *             self._debug_stream_read_errors_total))             # <<<<<<<<<<<<<<
 *         print('Read callbacks:            {}'.format(
 *             self._debug_stream_read_cb_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1216
 *             self._debug_stream_write_cb_errors_total))
 *         print()
 *         print('Read errors:               {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_errors_total))
 *         print('Read callbacks:            {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1218
 *         print('Read errors:               {}'.format(
 *             self._debug_stream_read_errors_total))
 *         print('Read callbacks:            {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_cb_total))
 *         print('Read failed callbacks:     {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Read_callbacks, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1219
 *             self._debug_stream_read_errors_total))
 *         print('Read callbacks:            {}'.format(
 *             self._debug_stream_read_cb_total))             # <<<<<<<<<<<<<<
 *         print('Read failed callbacks:     {}'.format(
 *             self._debug_stream_read_cb_errors_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_cb_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1218
 *         print('Read errors:               {}'.format(
 *             self._debug_stream_read_errors_total))
 *         print('Read callbacks:            {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_cb_total))
 *         print('Read failed callbacks:     {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1220
 *         print('Read callbacks:            {}'.format(
 *             self._debug_stream_read_cb_total))
 *         print('Read failed callbacks:     {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_cb_errors_total))
 *         print('Read EOFs:                 {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Read_failed_callbacks, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1221
 *             self._debug_stream_read_cb_total))
 *         print('Read failed callbacks:     {}'.format(
 *             self._debug_stream_read_cb_errors_total))             # <<<<<<<<<<<<<<
 *         print('Read EOFs:                 {}'.format(
 *             self._debug_stream_read_eof_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_cb_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1220
 *         print('Read callbacks:            {}'.format(
 *             self._debug_stream_read_cb_total))
 *         print('Read failed callbacks:     {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_cb_errors_total))
 *         print('Read EOFs:                 {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1222
 *         print('Read failed callbacks:     {}'.format(
 *             self._debug_stream_read_cb_errors_total))
 *         print('Read EOFs:                 {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_eof_total))
 *         print('Read EOF failed callbacks: {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Read_EOFs, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1223
 *             self._debug_stream_read_cb_errors_total))
 *         print('Read EOFs:                 {}'.format(
 *             self._debug_stream_read_eof_total))             # <<<<<<<<<<<<<<
 *         print('Read EOF failed callbacks: {}'.format(
 *             self._debug_stream_read_eof_cb_errors_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_eof_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1222
 *         print('Read failed callbacks:     {}'.format(
 *             self._debug_stream_read_cb_errors_total))
 *         print('Read EOFs:                 {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_eof_total))
 *         print('Read EOF failed callbacks: {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1224
 *         print('Read EOFs:                 {}'.format(
 *             self._debug_stream_read_eof_total))
 *         print('Read EOF failed callbacks: {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_eof_cb_errors_total))
 *         print()
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Read_EOF_failed_callbacks, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1225
 *             self._debug_stream_read_eof_total))
 *         print('Read EOF failed callbacks: {}'.format(
 *             self._debug_stream_read_eof_cb_errors_total))             # <<<<<<<<<<<<<<
 *         print()
 *         print('Listen errors:             {}'.format(
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_eof_cb_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1224
 *         print('Read EOFs:                 {}'.format(
 *             self._debug_stream_read_eof_total))
 *         print('Read EOF failed callbacks: {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_read_eof_cb_errors_total))
 *         print()
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1226
 *         print('Read EOF failed callbacks: {}'.format(
 *             self._debug_stream_read_eof_cb_errors_total))
 *         print()             # <<<<<<<<<<<<<<
 *         print('Listen errors:             {}'.format(
 *             self._debug_stream_listen_errors_total))
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1227
 *             self._debug_stream_read_eof_cb_errors_total))
 *         print()
 *         print('Listen errors:             {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_listen_errors_total))
 *         print('Shutdown errors            {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Listen_errors, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1228
 *         print()
 *         print('Listen errors:             {}'.format(
 *             self._debug_stream_listen_errors_total))             # <<<<<<<<<<<<<<
 *         print('Shutdown errors            {}'.format(
 *             self._debug_stream_shutdown_errors_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_listen_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1227
 *             self._debug_stream_read_eof_cb_errors_total))
 *         print()
 *         print('Listen errors:             {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_listen_errors_total))
 *         print('Shutdown errors            {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1229
 *         print('Listen errors:             {}'.format(
 *             self._debug_stream_listen_errors_total))
 *         print('Shutdown errors            {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_shutdown_errors_total))
 *         print()
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Shutdown_errors, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1230
 *             self._debug_stream_listen_errors_total))
 *         print('Shutdown errors            {}'.format(
 *             self._debug_stream_shutdown_errors_total))             # <<<<<<<<<<<<<<
 *         print()
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_shutdown_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1229
 *         print('Listen errors:             {}'.format(
 *             self._debug_stream_listen_errors_total))
 *         print('Shutdown errors            {}'.format(             # <<<<<<<<<<<<<<
 *             self._debug_stream_shutdown_errors_total))
 *         print()
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1231
 *         print('Shutdown errors            {}'.format(
 *             self._debug_stream_shutdown_errors_total))
 *         print()             # <<<<<<<<<<<<<<
 * 
 *         print('--- Polls debug info: ---')
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1233
 *         print()
 * 
 *         print('--- Polls debug info: ---')             # <<<<<<<<<<<<<<
 *         print('Read events:               {}'.format(
 *             self._poll_read_events_total))
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1234
 * 
 *         print('--- Polls debug info: ---')
 *         print('Read events:               {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_read_events_total))
 *         print('Read callbacks failed:     {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Read_events, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1235
 *         print('--- Polls debug info: ---')
 *         print('Read events:               {}'.format(
 *             self._poll_read_events_total))             # <<<<<<<<<<<<<<
 *         print('Read callbacks failed:     {}'.format(
 *             self._poll_read_cb_errors_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_read_events_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1234
 * 
 *         print('--- Polls debug info: ---')
 *         print('Read events:               {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_read_events_total))
 *         print('Read callbacks failed:     {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1236
 *         print('Read events:               {}'.format(
 *             self._poll_read_events_total))
 *         print('Read callbacks failed:     {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_read_cb_errors_total))
 *         print('Write events:              {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Read_callbacks_failed, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1237
 *             self._poll_read_events_total))
 *         print('Read callbacks failed:     {}'.format(
 *             self._poll_read_cb_errors_total))             # <<<<<<<<<<<<<<
 *         print('Write events:              {}'.format(
 *             self._poll_write_events_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_read_cb_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1236
 *         print('Read events:               {}'.format(
 *             self._poll_read_events_total))
 *         print('Read callbacks failed:     {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_read_cb_errors_total))
 *         print('Write events:              {}'.format(
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1238
 *         print('Read callbacks failed:     {}'.format(
 *             self._poll_read_cb_errors_total))
 *         print('Write events:              {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_write_events_total))
 *         print('Write callbacks failed:    {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Write_events, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1239
 *             self._poll_read_cb_errors_total))
 *         print('Write events:              {}'.format(
 *             self._poll_write_events_total))             # <<<<<<<<<<<<<<
 *         print('Write callbacks failed:    {}'.format(
 *             self._poll_write_cb_errors_total))
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_write_events_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1238
 *         print('Read callbacks failed:     {}'.format(
 *             self._poll_read_cb_errors_total))
 *         print('Write events:              {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_write_events_total))
 *         print('Write callbacks failed:    {}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1240
 *         print('Write events:              {}'.format(
 *             self._poll_write_events_total))
 *         print('Write callbacks failed:    {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_write_cb_errors_total))
 *         print()
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Write_callbacks_failed, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/loop.pyx":1241
 *             self._poll_write_events_total))
 *         print('Write callbacks failed:    {}'.format(
 *             self._poll_write_cb_errors_total))             # <<<<<<<<<<<<<<
 *         print()
 * 
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_write_cb_errors_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uvloop/loop.pyx":1240
 *         print('Write events:              {}'.format(
 *             self._poll_write_events_total))
 *         print('Write callbacks failed:    {}'.format(             # <<<<<<<<<<<<<<
 *             self._poll_write_cb_errors_total))
 *         print()
 */
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1242
 *         print('Write callbacks failed:    {}'.format(
 *             self._poll_write_cb_errors_total))
 *         print()             # <<<<<<<<<<<<<<
 * 
 *         print('--- Sock ops successful on 1st try: ---')
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_builtin_print); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1244
 *         print()
 * 
 *         print('--- Sock ops successful on 1st try: ---')             # <<<<<<<<<<<<<<
 *         print('Socket try-writes:         {}'.format(
 *             self._sock_try_write_total))
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1245
 * 
 *         print('--- Sock ops successful on 1st try: ---')
 *         print('Socket try-writes:         {}'.format(             # <<<<<<<<<<<<<<
 *             self._sock_try_write_total))
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Socket_try_writes, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1246
 *         print('--- Sock ops successful on 1st try: ---')
 *         print('Socket try-writes:         {}'.format(
 *             self._sock_try_write_total))             # <<<<<<<<<<<<<<
 * 
 *         print(flush=True)
 */
  __pyx_t_10 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_sock_try_write_total); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_10};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "uvloop/loop.pyx":1245
 * 
 *         print('--- Sock ops successful on 1st try: ---')
 *         print('Socket try-writes:         {}'.format(             # <<<<<<<<<<<<<<
 *             self._sock_try_write_total))
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1248
 *             self._sock_try_write_total))
 * 
 *         print(flush=True)             # <<<<<<<<<<<<<<
 * 
 *     property print_debug_info:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_flush, Py_True) < 0) __PYX_ERR(0, 1248, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1156
 *         return uv.uv_backend_fd(self.uvloop)
 * 
 *     cdef _print_debug_info(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("uvloop.loop.Loop._print_debug_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1251
 * 
 *     property print_debug_info:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 return lambda: self._print_debug_info()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_16print_debug_info_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_16print_debug_info_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_16print_debug_info___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1253
 *         def __get__(self):
 *             if UVLOOP_DEBUG:
 *                 return lambda: self._print_debug_info()             # <<<<<<<<<<<<<<
 *             else:
 *                 raise AttributeError('print_debug_info')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_16print_debug_info_7__get___lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_16print_debug_info_7__get___lambda = {"lambda", (PyCFunction)__pyx_pw_6uvloop_4loop_4Loop_16print_debug_info_7__get___lambda, METH_NOARGS, 0};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_16print_debug_info_7__get___lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *__pyx_cur_scope;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 1);
  __pyx_outer_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 1253, __pyx_L1_error) }
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_print_debug_info(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop.print_debug_info.__get__.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1251
 * 
 *     property print_debug_info:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 return lambda: self._print_debug_info()
 */

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_16print_debug_info___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_2___get__(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1251, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "uvloop/loop.pyx":1252
 *     property print_debug_info:
 *         def __get__(self):
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 return lambda: self._print_debug_info()
 *             else:
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (likely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":1253
 *         def __get__(self):
 *             if UVLOOP_DEBUG:
 *                 return lambda: self._print_debug_info()             # <<<<<<<<<<<<<<
 *             else:
 *                 raise AttributeError('print_debug_info')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_16print_debug_info_7__get___lambda, 0, __pyx_n_s_Loop___get___locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_uvloop_loop, __pyx_d, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":1252
 *     property print_debug_info:
 *         def __get__(self):
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 return lambda: self._print_debug_info()
 *             else:
 */
  }

  /* "uvloop/loop.pyx":1255
 *                 return lambda: self._print_debug_info()
 *             else:
 *                 raise AttributeError('print_debug_info')             # <<<<<<<<<<<<<<
 * 
 *     # Public API
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_AttributeError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1255, __pyx_L1_error)
  }

  /* "uvloop/loop.pyx":1251
 * 
 *     property print_debug_info:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 return lambda: self._print_debug_info()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop.print_debug_info.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1259
 *     # Public API
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{}.{} running={} closed={} debug={}>'.format(
 *             self.__class__.__module__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_11__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_11__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_10__repr__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_10__repr__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/loop.pyx":1260
 * 
 *     def __repr__(self):
 *         return '<{}.{} running={} closed={} debug={}>'.format(             # <<<<<<<<<<<<<<
 *             self.__class__.__module__,
 *             self.__class__.__name__,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_running_closed_debug, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/loop.pyx":1261
 *     def __repr__(self):
 *         return '<{}.{} running={} closed={} debug={}>'.format(
 *             self.__class__.__module__,             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.is_running(),
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_module); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1262
 *         return '<{}.{} running={} closed={} debug={}>'.format(
 *             self.__class__.__module__,
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.is_running(),
 *             self.is_closed(),
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1263
 *             self.__class__.__module__,
 *             self.__class__.__name__,
 *             self.is_running(),             # <<<<<<<<<<<<<<
 *             self.is_closed(),
 *             self.get_debug()
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_running); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "uvloop/loop.pyx":1264
 *             self.__class__.__name__,
 *             self.is_running(),
 *             self.is_closed(),             # <<<<<<<<<<<<<<
 *             self.get_debug()
 *         )
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_closed); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }

  /* "uvloop/loop.pyx":1265
 *             self.is_running(),
 *             self.is_closed(),
 *             self.get_debug()             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_debug); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __pyx_t_9 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[6] = {__pyx_t_9, __pyx_t_4, __pyx_t_5, __pyx_t_3, __pyx_t_6, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 5+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1259
 *     # Public API
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{}.{} running={} closed={} debug={}>'.format(
 *             self.__class__.__module__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uvloop.loop.Loop.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1268
 *         )
 * 
 *     def call_soon(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called as soon as possible.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_13call_soon(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_12call_soon, "Loop.call_soon(self, callback, *args, context=None)\nArrange for a callback to be called as soon as possible.\n\n        This operates as a FIFO queue: callbacks are called in the\n        order in which they are registered.  Each callback will be\n        called exactly once.\n\n        Any positional arguments after the callback will be passed to\n        the callback when it is called.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_13call_soon = {"call_soon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_13call_soon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_12call_soon};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_13call_soon(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_context = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_soon (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_callback,&__pyx_n_s_context,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1268, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 1;
        PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
        if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1268, __pyx_L3_error)
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "call_soon") < 0)) __PYX_ERR(0, 1268, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_callback = values[0];
    __pyx_v_context = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_soon", 0, 1, 1, __pyx_nargs); __PYX_ERR(0, 1268, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.call_soon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_12call_soon(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_callback, __pyx_v_context, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_12call_soon(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_soon", 1);

  /* "uvloop/loop.pyx":1278
 *         the callback when it is called.
 *         """
 *         if self._debug == 1:             # <<<<<<<<<<<<<<
 *             self._check_thread()
 *         if args:
 */
  __pyx_t_1 = (__pyx_v_self->_debug == 1);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1279
 *         """
 *         if self._debug == 1:
 *             self._check_thread()             # <<<<<<<<<<<<<<
 *         if args:
 *             return self._call_soon(callback, args, context)
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__check_thread(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":1278
 *         the callback when it is called.
 *         """
 *         if self._debug == 1:             # <<<<<<<<<<<<<<
 *             self._check_thread()
 *         if args:
 */
  }

  /* "uvloop/loop.pyx":1280
 *         if self._debug == 1:
 *             self._check_thread()
 *         if args:             # <<<<<<<<<<<<<<
 *             return self._call_soon(callback, args, context)
 *         else:
 */
  __pyx_t_1 = (PyTuple_GET_SIZE(__pyx_v_args) != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1281
 *             self._check_thread()
 *         if args:
 *             return self._call_soon(callback, args, context)             # <<<<<<<<<<<<<<
 *         else:
 *             return self._call_soon(callback, None, context)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__call_soon(__pyx_v_self, __pyx_v_callback, __pyx_v_args, __pyx_v_context); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":1280
 *         if self._debug == 1:
 *             self._check_thread()
 *         if args:             # <<<<<<<<<<<<<<
 *             return self._call_soon(callback, args, context)
 *         else:
 */
  }

  /* "uvloop/loop.pyx":1283
 *             return self._call_soon(callback, args, context)
 *         else:
 *             return self._call_soon(callback, None, context)             # <<<<<<<<<<<<<<
 * 
 *     def call_soon_threadsafe(self, callback, *args, context=None):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__call_soon(__pyx_v_self, __pyx_v_callback, Py_None, __pyx_v_context); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "uvloop/loop.pyx":1268
 *         )
 * 
 *     def call_soon(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called as soon as possible.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop.call_soon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1285
 *             return self._call_soon(callback, None, context)
 * 
 *     def call_soon_threadsafe(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_soon(), but thread-safe."""
 *         if not args:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_15call_soon_threadsafe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_14call_soon_threadsafe, "Loop.call_soon_threadsafe(self, callback, *args, context=None)\nLike call_soon(), but thread-safe.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_15call_soon_threadsafe = {"call_soon_threadsafe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_15call_soon_threadsafe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_14call_soon_threadsafe};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_15call_soon_threadsafe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_context = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_soon_threadsafe (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_callback,&__pyx_n_s_context,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1285, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 1;
        PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
        if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1285, __pyx_L3_error)
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "call_soon_threadsafe") < 0)) __PYX_ERR(0, 1285, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_callback = values[0];
    __pyx_v_context = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_soon_threadsafe", 0, 1, 1, __pyx_nargs); __PYX_ERR(0, 1285, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.call_soon_threadsafe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_14call_soon_threadsafe(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_callback, __pyx_v_context, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_14call_soon_threadsafe(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_soon_threadsafe", 0);
  __Pyx_INCREF(__pyx_v_args);

  /* "uvloop/loop.pyx":1287
 *     def call_soon_threadsafe(self, callback, *args, context=None):
 *         """Like call_soon(), but thread-safe."""
 *         if not args:             # <<<<<<<<<<<<<<
 *             args = None
 *         cdef Handle handle = new_Handle(self, callback, args, context)
 */
  __pyx_t_1 = (PyTuple_GET_SIZE(__pyx_v_args) != 0);
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1288
 *         """Like call_soon(), but thread-safe."""
 *         if not args:
 *             args = None             # <<<<<<<<<<<<<<
 *         cdef Handle handle = new_Handle(self, callback, args, context)
 *         self._append_ready_handle(handle)  # deque append is atomic
 */
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_args, ((PyObject*)Py_None));

    /* "uvloop/loop.pyx":1287
 *     def call_soon_threadsafe(self, callback, *args, context=None):
 *         """Like call_soon(), but thread-safe."""
 *         if not args:             # <<<<<<<<<<<<<<
 *             args = None
 *         cdef Handle handle = new_Handle(self, callback, args, context)
 */
  }

  /* "uvloop/loop.pyx":1289
 *         if not args:
 *             args = None
 *         cdef Handle handle = new_Handle(self, callback, args, context)             # <<<<<<<<<<<<<<
 *         self._append_ready_handle(handle)  # deque append is atomic
 *         # libuv async handler is thread-safe while the idle handler is not -
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_new_Handle(__pyx_v_self, __pyx_v_callback, __pyx_v_args, __pyx_v_context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1290
 *             args = None
 *         cdef Handle handle = new_Handle(self, callback, args, context)
 *         self._append_ready_handle(handle)  # deque append is atomic             # <<<<<<<<<<<<<<
 *         # libuv async handler is thread-safe while the idle handler is not -
 *         # we only set the async handler here, which will start the idle handler
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_4Loop__append_ready_handle(__pyx_v_self, __pyx_v_handle); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1294
 *         # we only set the async handler here, which will start the idle handler
 *         # in _on_wake() from the loop and eventually call the callback.
 *         self.handler_async.send()             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->handler_async->__pyx_base.__pyx_vtab)->send(__pyx_v_self->handler_async); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1295
 *         # in _on_wake() from the loop and eventually call the callback.
 *         self.handler_async.send()
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     def call_later(self, delay, callback, *args, context=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = ((PyObject *)__pyx_v_handle);
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1285
 *             return self._call_soon(callback, None, context)
 * 
 *     def call_soon_threadsafe(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_soon(), but thread-safe."""
 *         if not args:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Loop.call_soon_threadsafe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1297
 *         return handle
 * 
 *     def call_later(self, delay, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called at a given time.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_17call_later(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_16call_later, "Loop.call_later(self, delay, callback, *args, context=None)\nArrange for a callback to be called at a given time.\n\n        Return a Handle: an opaque object with a cancel() method that\n        can be used to cancel the call.\n\n        The delay can be an int or float, expressed in seconds.  It is\n        always relative to the current time.\n\n        Each callback will be called exactly once.  If two callbacks\n        are scheduled for exactly the same time, it undefined which\n        will be called first.\n\n        Any positional arguments after the callback will be passed to\n        the callback when it is called.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_17call_later = {"call_later", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_17call_later, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_16call_later};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_17call_later(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_delay = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_context = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_later (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 2, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_delay,&__pyx_n_s_callback,&__pyx_n_s_context,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_delay)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("call_later", 0, 2, 2, 1); __PYX_ERR(0, 1297, __pyx_L3_error)
        }
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 2;
        PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
        if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 2) ? kwd_pos_args : 2;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "call_later") < 0)) __PYX_ERR(0, 1297, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_delay = values[0];
    __pyx_v_callback = values[1];
    __pyx_v_context = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_later", 0, 2, 2, __pyx_nargs); __PYX_ERR(0, 1297, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.call_later", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_16call_later(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_delay, __pyx_v_callback, __pyx_v_context, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_16call_later(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_delay, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args) {
  uint64_t __pyx_v_when;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  uint64_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_later", 0);
  __Pyx_INCREF(__pyx_v_delay);
  __Pyx_INCREF(__pyx_v_args);

  /* "uvloop/loop.pyx":1315
 *         cdef uint64_t when
 * 
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         if self._debug == 1:
 *             self._check_thread()
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1316
 * 
 *         self._check_closed()
 *         if self._debug == 1:             # <<<<<<<<<<<<<<
 *             self._check_thread()
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_debug == 1);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1317
 *         self._check_closed()
 *         if self._debug == 1:
 *             self._check_thread()             # <<<<<<<<<<<<<<
 * 
 *         if delay < 0:
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_thread(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":1316
 * 
 *         self._check_closed()
 *         if self._debug == 1:             # <<<<<<<<<<<<<<
 *             self._check_thread()
 * 
 */
  }

  /* "uvloop/loop.pyx":1319
 *             self._check_thread()
 * 
 *         if delay < 0:             # <<<<<<<<<<<<<<
 *             delay = 0
 *         elif delay == py_inf or delay > MAX_SLEEP:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_delay, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1319, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1320
 * 
 *         if delay < 0:
 *             delay = 0             # <<<<<<<<<<<<<<
 *         elif delay == py_inf or delay > MAX_SLEEP:
 *             # ~100 years sounds like a good approximation of
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_DECREF_SET(__pyx_v_delay, __pyx_int_0);

    /* "uvloop/loop.pyx":1319
 *             self._check_thread()
 * 
 *         if delay < 0:             # <<<<<<<<<<<<<<
 *             delay = 0
 *         elif delay == py_inf or delay > MAX_SLEEP:
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":1321
 *         if delay < 0:
 *             delay = 0
 *         elif delay == py_inf or delay > MAX_SLEEP:             # <<<<<<<<<<<<<<
 *             # ~100 years sounds like a good approximation of
 *             # infinity for a Python application.
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_delay, __pyx_v_6uvloop_4loop_py_inf, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_6uvloop_4loop_MAX_SLEEP); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_delay, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1324
 *             # ~100 years sounds like a good approximation of
 *             # infinity for a Python application.
 *             delay = MAX_SLEEP             # <<<<<<<<<<<<<<
 * 
 *         when = <uint64_t>round(delay * 1000)
 */
    __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_6uvloop_4loop_MAX_SLEEP); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_delay, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1321
 *         if delay < 0:
 *             delay = 0
 *         elif delay == py_inf or delay > MAX_SLEEP:             # <<<<<<<<<<<<<<
 *             # ~100 years sounds like a good approximation of
 *             # infinity for a Python application.
 */
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":1326
 *             delay = MAX_SLEEP
 * 
 *         when = <uint64_t>round(delay * 1000)             # <<<<<<<<<<<<<<
 *         if not args:
 *             args = None
 */
  __pyx_t_4 = __Pyx_PyInt_MultiplyObjC(__pyx_v_delay, __pyx_int_1000, 0x3E8, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_round, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_when = ((uint64_t)__pyx_t_5);

  /* "uvloop/loop.pyx":1327
 * 
 *         when = <uint64_t>round(delay * 1000)
 *         if not args:             # <<<<<<<<<<<<<<
 *             args = None
 *         if when == 0:
 */
  __pyx_t_2 = (PyTuple_GET_SIZE(__pyx_v_args) != 0);
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":1328
 *         when = <uint64_t>round(delay * 1000)
 *         if not args:
 *             args = None             # <<<<<<<<<<<<<<
 *         if when == 0:
 *             return self._call_soon(callback, args, context)
 */
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_args, ((PyObject*)Py_None));

    /* "uvloop/loop.pyx":1327
 * 
 *         when = <uint64_t>round(delay * 1000)
 *         if not args:             # <<<<<<<<<<<<<<
 *             args = None
 *         if when == 0:
 */
  }

  /* "uvloop/loop.pyx":1329
 *         if not args:
 *             args = None
 *         if when == 0:             # <<<<<<<<<<<<<<
 *             return self._call_soon(callback, args, context)
 *         else:
 */
  __pyx_t_3 = (__pyx_v_when == 0);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":1330
 *             args = None
 *         if when == 0:
 *             return self._call_soon(callback, args, context)             # <<<<<<<<<<<<<<
 *         else:
 *             return self._call_later(when, callback, args, context)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__call_soon(__pyx_v_self, __pyx_v_callback, __pyx_v_args, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":1329
 *         if not args:
 *             args = None
 *         if when == 0:             # <<<<<<<<<<<<<<
 *             return self._call_soon(callback, args, context)
 *         else:
 */
  }

  /* "uvloop/loop.pyx":1332
 *             return self._call_soon(callback, args, context)
 *         else:
 *             return self._call_later(when, callback, args, context)             # <<<<<<<<<<<<<<
 * 
 *     def call_at(self, when, callback, *args, context=None):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_call_later(__pyx_v_self, __pyx_v_when, __pyx_v_callback, __pyx_v_args, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "uvloop/loop.pyx":1297
 *         return handle
 * 
 *     def call_later(self, delay, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called at a given time.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.call_later", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_delay);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1334
 *             return self._call_later(when, callback, args, context)
 * 
 *     def call_at(self, when, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_later(), but uses an absolute time.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_19call_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_18call_at, "Loop.call_at(self, when, callback, *args, context=None)\nLike call_later(), but uses an absolute time.\n\n        Absolute time corresponds to the event loop's time() method.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_19call_at = {"call_at", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_19call_at, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_18call_at};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_19call_at(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_when = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_context = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_at (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 2, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_when,&__pyx_n_s_callback,&__pyx_n_s_context,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_when)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("call_at", 0, 2, 2, 1); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 2;
        PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
        if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 2) ? kwd_pos_args : 2;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "call_at") < 0)) __PYX_ERR(0, 1334, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_when = values[0];
    __pyx_v_callback = values[1];
    __pyx_v_context = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_at", 0, 2, 2, __pyx_nargs); __PYX_ERR(0, 1334, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.call_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_18call_at(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_when, __pyx_v_callback, __pyx_v_context, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_18call_at(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_when, PyObject *__pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_at", 1);

  /* "uvloop/loop.pyx":1339
 *         Absolute time corresponds to the event loop's time() method.
 *         """
 *         return self.call_later(             # <<<<<<<<<<<<<<
 *             when - self.time(), callback, *args, context=context)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call_later); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "uvloop/loop.pyx":1340
 *         """
 *         return self.call_later(
 *             when - self.time(), callback, *args, context=context)             # <<<<<<<<<<<<<<
 * 
 *     def time(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = PyNumber_Subtract(__pyx_v_when, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1339
 *         Absolute time corresponds to the event loop's time() method.
 *         """
 *         return self.call_later(             # <<<<<<<<<<<<<<
 *             when - self.time(), callback, *args, context=context)
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(0, 1339, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_callback)) __PYX_ERR(0, 1339, __pyx_L1_error);
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1340
 *         """
 *         return self.call_later(
 *             when - self.time(), callback, *args, context=context)             # <<<<<<<<<<<<<<
 * 
 *     def time(self):
 */
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_context, __pyx_v_context) < 0) __PYX_ERR(0, 1340, __pyx_L1_error)

  /* "uvloop/loop.pyx":1339
 *         Absolute time corresponds to the event loop's time() method.
 *         """
 *         return self.call_later(             # <<<<<<<<<<<<<<
 *             when - self.time(), callback, *args, context=context)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1334
 *             return self._call_later(when, callback, args, context)
 * 
 *     def call_at(self, when, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_later(), but uses an absolute time.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.call_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1342
 *             when - self.time(), callback, *args, context=context)
 * 
 *     def time(self):             # <<<<<<<<<<<<<<
 *         """Return the time according to the event loop's clock.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_21time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_20time, "Loop.time(self)\nReturn the time according to the event loop's clock.\n\n        This is a float expressed in seconds since an epoch, but the\n        epoch, precision, accuracy and drift are unspecified and may\n        differ per event loop.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_21time = {"time", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_21time, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_20time};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_21time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("time (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("time", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "time", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_20time(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_20time(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("time", 1);

  /* "uvloop/loop.pyx":1349
 *         differ per event loop.
 *         """
 *         return self._time() / 1000             # <<<<<<<<<<<<<<
 * 
 *     def stop(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_time(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble((((double)__pyx_t_1) / 1000.0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1342
 *             when - self.time(), callback, *args, context=context)
 * 
 *     def time(self):             # <<<<<<<<<<<<<<
 *         """Return the time according to the event loop's clock.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop.time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1351
 *         return self._time() / 1000
 * 
 *     def stop(self):             # <<<<<<<<<<<<<<
 *         """Stop running the event loop.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23stop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_22stop, "Loop.stop(self)\nStop running the event loop.\n\n        Every callback already scheduled will still run.  This simply informs\n        run_forever to stop looping after a complete iteration.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_23stop = {"stop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_23stop, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_22stop};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23stop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stop (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("stop", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "stop", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_22stop(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_22stop(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop", 1);

  /* "uvloop/loop.pyx":1358
 *         """
 *         self._call_soon_handle(
 *             new_MethodHandle1(             # <<<<<<<<<<<<<<
 *                 self,
 *                 "Loop._stop",
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_new_MethodHandle1(__pyx_v_self, __pyx_kp_u_Loop__stop, ((__pyx_t_6uvloop_4loop_method1_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_stop), Py_None, ((PyObject *)__pyx_v_self), Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 1358, __pyx_L1_error)

  /* "uvloop/loop.pyx":1357
 *         run_forever to stop looping after a complete iteration.
 *         """
 *         self._call_soon_handle(             # <<<<<<<<<<<<<<
 *             new_MethodHandle1(
 *                 self,
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1351
 *         return self._time() / 1000
 * 
 *     def stop(self):             # <<<<<<<<<<<<<<
 *         """Stop running the event loop.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1366
 *                 None))
 * 
 *     def run_forever(self):             # <<<<<<<<<<<<<<
 *         """Run the event loop until stop() is called."""
 *         self._check_closed()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_25run_forever(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_24run_forever, "Loop.run_forever(self)\nRun the event loop until stop() is called.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_25run_forever = {"run_forever", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_25run_forever, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_24run_forever};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_25run_forever(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_forever (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("run_forever", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "run_forever", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_24run_forever(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_24run_forever(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  uv_run_mode __pyx_v_mode;
  PyObject *__pyx_v_old_agen_hooks = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_forever", 1);

  /* "uvloop/loop.pyx":1368
 *     def run_forever(self):
 *         """Run the event loop until stop() is called."""
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         mode = uv.UV_RUN_DEFAULT
 *         if self._stopping:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1369
 *         """Run the event loop until stop() is called."""
 *         self._check_closed()
 *         mode = uv.UV_RUN_DEFAULT             # <<<<<<<<<<<<<<
 *         if self._stopping:
 *             # loop.stop() was called right before loop.run_forever().
 */
  __pyx_v_mode = UV_RUN_DEFAULT;

  /* "uvloop/loop.pyx":1370
 *         self._check_closed()
 *         mode = uv.UV_RUN_DEFAULT
 *         if self._stopping:             # <<<<<<<<<<<<<<
 *             # loop.stop() was called right before loop.run_forever().
 *             # This is how asyncio loop behaves.
 */
  if (__pyx_v_self->_stopping) {

    /* "uvloop/loop.pyx":1373
 *             # loop.stop() was called right before loop.run_forever().
 *             # This is how asyncio loop behaves.
 *             mode = uv.UV_RUN_NOWAIT             # <<<<<<<<<<<<<<
 *         self._set_coroutine_debug(self._debug)
 *         old_agen_hooks = sys.get_asyncgen_hooks()
 */
    __pyx_v_mode = UV_RUN_NOWAIT;

    /* "uvloop/loop.pyx":1370
 *         self._check_closed()
 *         mode = uv.UV_RUN_DEFAULT
 *         if self._stopping:             # <<<<<<<<<<<<<<
 *             # loop.stop() was called right before loop.run_forever().
 *             # This is how asyncio loop behaves.
 */
  }

  /* "uvloop/loop.pyx":1374
 *             # This is how asyncio loop behaves.
 *             mode = uv.UV_RUN_NOWAIT
 *         self._set_coroutine_debug(self._debug)             # <<<<<<<<<<<<<<
 *         old_agen_hooks = sys.get_asyncgen_hooks()
 *         sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_set_coroutine_debug(__pyx_v_self, __pyx_v_self->_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1375
 *             mode = uv.UV_RUN_NOWAIT
 *         self._set_coroutine_debug(self._debug)
 *         old_agen_hooks = sys.get_asyncgen_hooks()             # <<<<<<<<<<<<<<
 *         sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,
 *                                finalizer=self._asyncgen_finalizer_hook)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get_asyncgen_hooks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_old_agen_hooks = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1376
 *         self._set_coroutine_debug(self._debug)
 *         old_agen_hooks = sys.get_asyncgen_hooks()
 *         sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,             # <<<<<<<<<<<<<<
 *                                finalizer=self._asyncgen_finalizer_hook)
 *         try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_set_asyncgen_hooks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asyncgen_firstiter_hook); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_firstiter, __pyx_t_2) < 0) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1377
 *         old_agen_hooks = sys.get_asyncgen_hooks()
 *         sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,
 *                                finalizer=self._asyncgen_finalizer_hook)             # <<<<<<<<<<<<<<
 *         try:
 *             self._run(mode)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asyncgen_finalizer_hook); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_finalizer, __pyx_t_2) < 0) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1376
 *         self._set_coroutine_debug(self._debug)
 *         old_agen_hooks = sys.get_asyncgen_hooks()
 *         sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,             # <<<<<<<<<<<<<<
 *                                finalizer=self._asyncgen_finalizer_hook)
 *         try:
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":1378
 *         sys.set_asyncgen_hooks(firstiter=self._asyncgen_firstiter_hook,
 *                                finalizer=self._asyncgen_finalizer_hook)
 *         try:             # <<<<<<<<<<<<<<
 *             self._run(mode)
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":1379
 *                                finalizer=self._asyncgen_finalizer_hook)
 *         try:
 *             self._run(mode)             # <<<<<<<<<<<<<<
 *         finally:
 *             self._set_coroutine_debug(False)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_run(__pyx_v_self, __pyx_v_mode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1379, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/loop.pyx":1381
 *             self._run(mode)
 *         finally:
 *             self._set_coroutine_debug(False)             # <<<<<<<<<<<<<<
 *             sys.set_asyncgen_hooks(*old_agen_hooks)
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_set_coroutine_debug(__pyx_v_self, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1381, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":1382
 *         finally:
 *             self._set_coroutine_debug(False)
 *             sys.set_asyncgen_hooks(*old_agen_hooks)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_set_asyncgen_hooks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_v_old_agen_hooks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {

        /* "uvloop/loop.pyx":1381
 *             self._run(mode)
 *         finally:
 *             self._set_coroutine_debug(False)             # <<<<<<<<<<<<<<
 *             sys.set_asyncgen_hooks(*old_agen_hooks)
 * 
 */
        __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_set_coroutine_debug(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1381, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "uvloop/loop.pyx":1382
 *         finally:
 *             self._set_coroutine_debug(False)
 *             sys.set_asyncgen_hooks(*old_agen_hooks)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1382, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_set_asyncgen_hooks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1382, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_old_agen_hooks); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1382, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1382, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
      __pyx_L8_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/loop.pyx":1366
 *                 None))
 * 
 *     def run_forever(self):             # <<<<<<<<<<<<<<
 *         """Run the event loop until stop() is called."""
 *         self._check_closed()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Loop.run_forever", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_old_agen_hooks);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1384
 *             sys.set_asyncgen_hooks(*old_agen_hooks)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the event loop.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_26close, "Loop.close(self)\nClose the event loop.\n\n        The event loop must not be running.\n\n        This is idempotent and irreversible.\n\n        No other methods should be called after this one.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_27close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_27close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_26close};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_26close(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_26close(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/loop.pyx":1393
 *         No other methods should be called after this one.
 *         """
 *         self._close()             # <<<<<<<<<<<<<<
 * 
 *     def get_debug(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_close(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1384
 *             sys.set_asyncgen_hooks(*old_agen_hooks)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the event loop.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1395
 *         self._close()
 * 
 *     def get_debug(self):             # <<<<<<<<<<<<<<
 *         return bool(self._debug)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29get_debug(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_28get_debug, "Loop.get_debug(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_29get_debug = {"get_debug", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_29get_debug, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_28get_debug};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29get_debug(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_debug (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_debug", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_debug", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_28get_debug(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_28get_debug(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_debug", 1);

  /* "uvloop/loop.pyx":1396
 * 
 *     def get_debug(self):
 *         return bool(self._debug)             # <<<<<<<<<<<<<<
 * 
 *     def set_debug(self, enabled):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->_debug;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1395
 *         self._close()
 * 
 *     def get_debug(self):             # <<<<<<<<<<<<<<
 *         return bool(self._debug)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop.get_debug", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1398
 *         return bool(self._debug)
 * 
 *     def set_debug(self, enabled):             # <<<<<<<<<<<<<<
 *         self._debug = bool(enabled)
 *         if self.is_running():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_31set_debug(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_30set_debug, "Loop.set_debug(self, enabled)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_31set_debug = {"set_debug", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_31set_debug, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_30set_debug};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_31set_debug(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_enabled = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_debug (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_enabled,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_enabled)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_debug") < 0)) __PYX_ERR(0, 1398, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_enabled = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_debug", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1398, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.set_debug", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_30set_debug(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_enabled);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_30set_debug(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_enabled) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_debug", 1);

  /* "uvloop/loop.pyx":1399
 * 
 *     def set_debug(self, enabled):
 *         self._debug = bool(enabled)             # <<<<<<<<<<<<<<
 *         if self.is_running():
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_enabled); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1399, __pyx_L1_error)
  __pyx_v_self->_debug = (!(!__pyx_t_1));

  /* "uvloop/loop.pyx":1400
 *     def set_debug(self, enabled):
 *         self._debug = bool(enabled)
 *         if self.is_running():             # <<<<<<<<<<<<<<
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_running); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1401
 *         self._debug = bool(enabled)
 *         if self.is_running():
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)             # <<<<<<<<<<<<<<
 * 
 *     def is_running(self):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call_soon_threadsafe); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_t_6 = __pyx_v_self;
    __pyx_t_4 = __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_t_6->__pyx_vtab)->_set_coroutine_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = ((PyObject *)__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyMethod_New2Arg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF((PyObject *)__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_8, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1401, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":1400
 *     def set_debug(self, enabled):
 *         self._debug = bool(enabled)
 *         if self.is_running():             # <<<<<<<<<<<<<<
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)
 * 
 */
  }

  /* "uvloop/loop.pyx":1398
 *         return bool(self._debug)
 * 
 *     def set_debug(self, enabled):             # <<<<<<<<<<<<<<
 *         self._debug = bool(enabled)
 *         if self.is_running():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF((PyObject *)__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Loop.set_debug", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1403
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)
 * 
 *     def is_running(self):             # <<<<<<<<<<<<<<
 *         """Return whether the event loop is currently running."""
 *         return bool(self._running)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_33is_running(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_32is_running, "Loop.is_running(self)\nReturn whether the event loop is currently running.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_33is_running = {"is_running", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_33is_running, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_32is_running};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_33is_running(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_running (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_running", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_running", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_32is_running(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_32is_running(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_running", 1);

  /* "uvloop/loop.pyx":1405
 *     def is_running(self):
 *         """Return whether the event loop is currently running."""
 *         return bool(self._running)             # <<<<<<<<<<<<<<
 * 
 *     def is_closed(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->_running;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1403
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)
 * 
 *     def is_running(self):             # <<<<<<<<<<<<<<
 *         """Return whether the event loop is currently running."""
 *         return bool(self._running)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop.is_running", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1407
 *         return bool(self._running)
 * 
 *     def is_closed(self):             # <<<<<<<<<<<<<<
 *         """Returns True if the event loop was closed."""
 *         return bool(self._closed)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_35is_closed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_34is_closed, "Loop.is_closed(self)\nReturns True if the event loop was closed.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_35is_closed = {"is_closed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_35is_closed, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_34is_closed};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_35is_closed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_closed (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_closed", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_closed", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_34is_closed(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_34is_closed(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_closed", 1);

  /* "uvloop/loop.pyx":1409
 *     def is_closed(self):
 *         """Returns True if the event loop was closed."""
 *         return bool(self._closed)             # <<<<<<<<<<<<<<
 * 
 *     def create_future(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->_closed;
  __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1407
 *         return bool(self._running)
 * 
 *     def is_closed(self):             # <<<<<<<<<<<<<<
 *         """Returns True if the event loop was closed."""
 *         return bool(self._closed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Loop.is_closed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1411
 *         return bool(self._closed)
 * 
 *     def create_future(self):             # <<<<<<<<<<<<<<
 *         """Create a Future object attached to the loop."""
 *         return self._new_future()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_37create_future(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_36create_future, "Loop.create_future(self)\nCreate a Future object attached to the loop.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_37create_future = {"create_future", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_37create_future, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_36create_future};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_37create_future(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_future (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("create_future", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "create_future", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_36create_future(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_36create_future(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_future", 1);

  /* "uvloop/loop.pyx":1413
 *     def create_future(self):
 *         """Create a Future object attached to the loop."""
 *         return self._new_future()             # <<<<<<<<<<<<<<
 * 
 *     def create_task(self, coro, *, name=None, context=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1411
 *         return bool(self._closed)
 * 
 *     def create_future(self):             # <<<<<<<<<<<<<<
 *         """Create a Future object attached to the loop."""
 *         return self._new_future()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop.create_future", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1415
 *         return self._new_future()
 * 
 *     def create_task(self, coro, *, name=None, context=None):             # <<<<<<<<<<<<<<
 *         """Schedule a coroutine object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_39create_task(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_38create_task, "Loop.create_task(self, coro, *, name=None, context=None)\nSchedule a coroutine object.\n\n        Return a task object.\n\n        If name is not None, task.set_name(name) will be called if the task\n        object has the set_name attribute, true for default Task in CPython.\n\n        An optional keyword-only context argument allows specifying a custom\n        contextvars.Context for the coro to run in. The current context copy is\n        created when no context is provided.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_39create_task = {"create_task", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_39create_task, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_38create_task};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_39create_task(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_coro = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_context = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_task (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_coro,&__pyx_n_s_name_2,&__pyx_n_s_context,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_coro)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1415, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (kw_args > 0 && likely(kw_args <= 2)) {
        Py_ssize_t index;
        for (index = 1; index < 3 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1415, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_task") < 0)) __PYX_ERR(0, 1415, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_coro = values[0];
    __pyx_v_name = values[1];
    __pyx_v_context = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_task", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1415, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.create_task", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_38create_task(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_coro, __pyx_v_name, __pyx_v_context);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_38create_task(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_coro, PyObject *__pyx_v_name, PyObject *__pyx_v_context) {
  PyObject *__pyx_v_task = NULL;
  PyObject *__pyx_v_set_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_task", 1);

  /* "uvloop/loop.pyx":1427
 *         created when no context is provided.
 *         """
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         if PY311:
 *             if self._task_factory is None:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1428
 *         """
 *         self._check_closed()
 *         if PY311:             # <<<<<<<<<<<<<<
 *             if self._task_factory is None:
 *                 task = aio_Task(coro, loop=self, context=context)
 */
  __pyx_t_2 = (__pyx_v_6uvloop_4loop_PY311 != 0);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1429
 *         self._check_closed()
 *         if PY311:
 *             if self._task_factory is None:             # <<<<<<<<<<<<<<
 *                 task = aio_Task(coro, loop=self, context=context)
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->_task_factory == Py_None);
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1430
 *         if PY311:
 *             if self._task_factory is None:
 *                 task = aio_Task(coro, loop=self, context=context)             # <<<<<<<<<<<<<<
 *             else:
 *                 task = self._task_factory(self, coro, context=context)
 */
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_coro);
      __Pyx_GIVEREF(__pyx_v_coro);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_coro)) __PYX_ERR(0, 1430, __pyx_L1_error);
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_loop, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 1430, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_context, __pyx_v_context) < 0) __PYX_ERR(0, 1430, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_Task, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1430, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_task = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1429
 *         self._check_closed()
 *         if PY311:
 *             if self._task_factory is None:             # <<<<<<<<<<<<<<
 *                 task = aio_Task(coro, loop=self, context=context)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uvloop/loop.pyx":1432
 *                 task = aio_Task(coro, loop=self, context=context)
 *             else:
 *                 task = self._task_factory(self, coro, context=context)             # <<<<<<<<<<<<<<
 *         else:
 *             if context is None:
 */
    /*else*/ {
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF((PyObject *)__pyx_v_self);
      __Pyx_GIVEREF((PyObject *)__pyx_v_self);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 1432, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_coro);
      __Pyx_GIVEREF(__pyx_v_coro);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_coro)) __PYX_ERR(0, 1432, __pyx_L1_error);
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_context, __pyx_v_context) < 0) __PYX_ERR(0, 1432, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_self->_task_factory, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_task = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_L4:;

    /* "uvloop/loop.pyx":1428
 *         """
 *         self._check_closed()
 *         if PY311:             # <<<<<<<<<<<<<<
 *             if self._task_factory is None:
 *                 task = aio_Task(coro, loop=self, context=context)
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":1434
 *                 task = self._task_factory(self, coro, context=context)
 *         else:
 *             if context is None:             # <<<<<<<<<<<<<<
 *                 if self._task_factory is None:
 *                     task = aio_Task(coro, loop=self)
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_context == Py_None);
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1435
 *         else:
 *             if context is None:
 *                 if self._task_factory is None:             # <<<<<<<<<<<<<<
 *                     task = aio_Task(coro, loop=self)
 *                 else:
 */
      __pyx_t_2 = (__pyx_v_self->_task_factory == Py_None);
      if (__pyx_t_2) {

        /* "uvloop/loop.pyx":1436
 *             if context is None:
 *                 if self._task_factory is None:
 *                     task = aio_Task(coro, loop=self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     task = self._task_factory(self, coro)
 */
        __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_coro);
        __Pyx_GIVEREF(__pyx_v_coro);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_coro)) __PYX_ERR(0, 1436, __pyx_L1_error);
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_loop, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 1436, __pyx_L1_error)
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_Task, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1436, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_task = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":1435
 *         else:
 *             if context is None:
 *                 if self._task_factory is None:             # <<<<<<<<<<<<<<
 *                     task = aio_Task(coro, loop=self)
 *                 else:
 */
        goto __pyx_L6;
      }

      /* "uvloop/loop.pyx":1438
 *                     task = aio_Task(coro, loop=self)
 *                 else:
 *                     task = self._task_factory(self, coro)             # <<<<<<<<<<<<<<
 *             else:
 *                 if self._task_factory is None:
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_v_self->_task_factory);
        __pyx_t_3 = __pyx_v_self->_task_factory; __pyx_t_1 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_1, ((PyObject *)__pyx_v_self), __pyx_v_coro};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1438, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_v_task = __pyx_t_4;
        __pyx_t_4 = 0;
      }
      __pyx_L6:;

      /* "uvloop/loop.pyx":1434
 *                 task = self._task_factory(self, coro, context=context)
 *         else:
 *             if context is None:             # <<<<<<<<<<<<<<
 *                 if self._task_factory is None:
 *                     task = aio_Task(coro, loop=self)
 */
      goto __pyx_L5;
    }

    /* "uvloop/loop.pyx":1440
 *                     task = self._task_factory(self, coro)
 *             else:
 *                 if self._task_factory is None:             # <<<<<<<<<<<<<<
 *                     task = context.run(aio_Task, coro, self)
 *                 else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_self->_task_factory == Py_None);
      if (__pyx_t_2) {

        /* "uvloop/loop.pyx":1441
 *             else:
 *                 if self._task_factory is None:
 *                     task = context.run(aio_Task, coro, self)             # <<<<<<<<<<<<<<
 *                 else:
 *                     task = context.run(self._task_factory, self, coro)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_context, __pyx_n_s_run); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1441, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_v_6uvloop_4loop_aio_Task, __pyx_v_coro, ((PyObject *)__pyx_v_self)};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1441, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_v_task = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":1440
 *                     task = self._task_factory(self, coro)
 *             else:
 *                 if self._task_factory is None:             # <<<<<<<<<<<<<<
 *                     task = context.run(aio_Task, coro, self)
 *                 else:
 */
        goto __pyx_L7;
      }

      /* "uvloop/loop.pyx":1443
 *                     task = context.run(aio_Task, coro, self)
 *                 else:
 *                     task = context.run(self._task_factory, self, coro)             # <<<<<<<<<<<<<<
 * 
 *         # copied from asyncio.tasks._set_task_name (bpo-34270)
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_context, __pyx_n_s_run); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1443, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_v_self->_task_factory, ((PyObject *)__pyx_v_self), __pyx_v_coro};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1443, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_v_task = __pyx_t_4;
        __pyx_t_4 = 0;
      }
      __pyx_L7:;
    }
    __pyx_L5:;
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":1446
 * 
 *         # copied from asyncio.tasks._set_task_name (bpo-34270)
 *         if name is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 set_name = task.set_name
 */
  __pyx_t_2 = (__pyx_v_name != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1447
 *         # copied from asyncio.tasks._set_task_name (bpo-34270)
 *         if name is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 set_name = task.set_name
 *             except AttributeError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "uvloop/loop.pyx":1448
 *         if name is not None:
 *             try:
 *                 set_name = task.set_name             # <<<<<<<<<<<<<<
 *             except AttributeError:
 *                 pass
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_set_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1448, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_v_set_name = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":1447
 *         # copied from asyncio.tasks._set_task_name (bpo-34270)
 *         if name is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 set_name = task.set_name
 *             except AttributeError:
 */
      }

      /* "uvloop/loop.pyx":1452
 *                 pass
 *             else:
 *                 set_name(name)             # <<<<<<<<<<<<<<
 * 
 *         return task
 */
      /*else:*/ {
        __Pyx_INCREF(__pyx_v_set_name);
        __pyx_t_3 = __pyx_v_set_name; __pyx_t_1 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_name};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1452, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L14_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1449
 *             try:
 *                 set_name = task.set_name
 *             except AttributeError:             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
      if (__pyx_t_9) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L10_exception_handled;
      }
      goto __pyx_L11_except_error;

      /* "uvloop/loop.pyx":1447
 *         # copied from asyncio.tasks._set_task_name (bpo-34270)
 *         if name is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 set_name = task.set_name
 *             except AttributeError:
 */
      __pyx_L11_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L10_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L14_try_end:;
    }

    /* "uvloop/loop.pyx":1446
 * 
 *         # copied from asyncio.tasks._set_task_name (bpo-34270)
 *         if name is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 set_name = task.set_name
 */
  }

  /* "uvloop/loop.pyx":1454
 *                 set_name(name)
 * 
 *         return task             # <<<<<<<<<<<<<<
 * 
 *     def set_task_factory(self, factory):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_task);
  __pyx_r = __pyx_v_task;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1415
 *         return self._new_future()
 * 
 *     def create_task(self, coro, *, name=None, context=None):             # <<<<<<<<<<<<<<
 *         """Schedule a coroutine object.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.create_task", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_task);
  __Pyx_XDECREF(__pyx_v_set_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1456
 *         return task
 * 
 *     def set_task_factory(self, factory):             # <<<<<<<<<<<<<<
 *         """Set a task factory that will be used by loop.create_task().
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_41set_task_factory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_40set_task_factory, "Loop.set_task_factory(self, factory)\nSet a task factory that will be used by loop.create_task().\n\n        If factory is None the default task factory will be set.\n\n        If factory is a callable, it should have a signature matching\n        '(loop, coro)', where 'loop' will be a reference to the active\n        event loop, 'coro' will be a coroutine object.  The callable\n        must return a Future.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_41set_task_factory = {"set_task_factory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_41set_task_factory, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_40set_task_factory};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_41set_task_factory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_factory = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_task_factory (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_factory,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1456, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_task_factory") < 0)) __PYX_ERR(0, 1456, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_factory = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_task_factory", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1456, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.set_task_factory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_40set_task_factory(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_factory);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_40set_task_factory(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_factory) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_task_factory", 1);

  /* "uvloop/loop.pyx":1466
 *         must return a Future.
 *         """
 *         if factory is not None and not callable(factory):             # <<<<<<<<<<<<<<
 *             raise TypeError('task factory must be a callable or None')
 *         self._task_factory = factory
 */
  __pyx_t_2 = (__pyx_v_factory != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_v_factory); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1466, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":1467
 *         """
 *         if factory is not None and not callable(factory):
 *             raise TypeError('task factory must be a callable or None')             # <<<<<<<<<<<<<<
 *         self._task_factory = factory
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1467, __pyx_L1_error)

    /* "uvloop/loop.pyx":1466
 *         must return a Future.
 *         """
 *         if factory is not None and not callable(factory):             # <<<<<<<<<<<<<<
 *             raise TypeError('task factory must be a callable or None')
 *         self._task_factory = factory
 */
  }

  /* "uvloop/loop.pyx":1468
 *         if factory is not None and not callable(factory):
 *             raise TypeError('task factory must be a callable or None')
 *         self._task_factory = factory             # <<<<<<<<<<<<<<
 * 
 *     def get_task_factory(self):
 */
  __Pyx_INCREF(__pyx_v_factory);
  __Pyx_GIVEREF(__pyx_v_factory);
  __Pyx_GOTREF(__pyx_v_self->_task_factory);
  __Pyx_DECREF(__pyx_v_self->_task_factory);
  __pyx_v_self->_task_factory = __pyx_v_factory;

  /* "uvloop/loop.pyx":1456
 *         return task
 * 
 *     def set_task_factory(self, factory):             # <<<<<<<<<<<<<<
 *         """Set a task factory that will be used by loop.create_task().
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.set_task_factory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1470
 *         self._task_factory = factory
 * 
 *     def get_task_factory(self):             # <<<<<<<<<<<<<<
 *         """Return a task factory, or None if the default one is in use."""
 *         return self._task_factory
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_43get_task_factory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_42get_task_factory, "Loop.get_task_factory(self)\nReturn a task factory, or None if the default one is in use.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_43get_task_factory = {"get_task_factory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_43get_task_factory, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_42get_task_factory};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_43get_task_factory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_task_factory (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_task_factory", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_task_factory", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_42get_task_factory(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_42get_task_factory(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_task_factory", 1);

  /* "uvloop/loop.pyx":1472
 *     def get_task_factory(self):
 *         """Return a task factory, or None if the default one is in use."""
 *         return self._task_factory             # <<<<<<<<<<<<<<
 * 
 *     def run_until_complete(self, future):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_task_factory);
  __pyx_r = __pyx_v_self->_task_factory;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1470
 *         self._task_factory = factory
 * 
 *     def get_task_factory(self):             # <<<<<<<<<<<<<<
 *         """Return a task factory, or None if the default one is in use."""
 *         return self._task_factory
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1474
 *         return self._task_factory
 * 
 *     def run_until_complete(self, future):             # <<<<<<<<<<<<<<
 *         """Run until the Future is done.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_45run_until_complete(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_44run_until_complete, "Loop.run_until_complete(self, future)\nRun until the Future is done.\n\n        If the argument is a coroutine, it is wrapped in a Task.\n\n        WARNING: It would be disastrous to call run_until_complete()\n        with the same coroutine twice -- it would wrap it in two\n        different Tasks and that can't be good.\n\n        Return the Future's result, or raise its exception.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_45run_until_complete = {"run_until_complete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_45run_until_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_44run_until_complete};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_45run_until_complete(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_future = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_until_complete (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_future,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_future)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1474, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "run_until_complete") < 0)) __PYX_ERR(0, 1474, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_future = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run_until_complete", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1474, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.run_until_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_44run_until_complete(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_future);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1494
 *             future._log_destroy_pending = False
 * 
 *         def done_cb(fut):             # <<<<<<<<<<<<<<
 *             if not fut.cancelled():
 *                 exc = fut.exception()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_18run_until_complete_1done_cb(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_18run_until_complete_1done_cb = {"done_cb", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_18run_until_complete_1done_cb, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_18run_until_complete_1done_cb(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fut = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("done_cb (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fut,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fut)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1494, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "done_cb") < 0)) __PYX_ERR(0, 1494, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fut = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("done_cb", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1494, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.run_until_complete.done_cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_18run_until_complete_done_cb(__pyx_self, __pyx_v_fut);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_18run_until_complete_done_cb(PyObject *__pyx_self, PyObject *__pyx_v_fut) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *__pyx_cur_scope;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *__pyx_outer_scope;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("done_cb", 1);
  __pyx_outer_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "uvloop/loop.pyx":1495
 * 
 *         def done_cb(fut):
 *             if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                 exc = fut.exception()
 *                 if isinstance(exc, (SystemExit, KeyboardInterrupt)):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "uvloop/loop.pyx":1496
 *         def done_cb(fut):
 *             if not fut.cancelled():
 *                 exc = fut.exception()             # <<<<<<<<<<<<<<
 *                 if isinstance(exc, (SystemExit, KeyboardInterrupt)):
 *                     # Issue #336: run_forever() already finished,
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":1497
 *             if not fut.cancelled():
 *                 exc = fut.exception()
 *                 if isinstance(exc, (SystemExit, KeyboardInterrupt)):             # <<<<<<<<<<<<<<
 *                     # Issue #336: run_forever() already finished,
 *                     # no need to stop it.
 */
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_exc, __pyx_builtin_SystemExit); 
    if (!__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_5 = PyObject_IsInstance(__pyx_v_exc, __pyx_builtin_KeyboardInterrupt); 
    __pyx_t_6 = __pyx_t_5;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_6) {

      /* "uvloop/loop.pyx":1500
 *                     # Issue #336: run_forever() already finished,
 *                     # no need to stop it.
 *                     return             # <<<<<<<<<<<<<<
 *             self.stop()
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/loop.pyx":1497
 *             if not fut.cancelled():
 *                 exc = fut.exception()
 *                 if isinstance(exc, (SystemExit, KeyboardInterrupt)):             # <<<<<<<<<<<<<<
 *                     # Issue #336: run_forever() already finished,
 *                     # no need to stop it.
 */
    }

    /* "uvloop/loop.pyx":1495
 * 
 *         def done_cb(fut):
 *             if not fut.cancelled():             # <<<<<<<<<<<<<<
 *                 exc = fut.exception()
 *                 if isinstance(exc, (SystemExit, KeyboardInterrupt)):
 */
  }

  /* "uvloop/loop.pyx":1501
 *                     # no need to stop it.
 *                     return
 *             self.stop()             # <<<<<<<<<<<<<<
 * 
 *         future.add_done_callback(done_cb)
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 1501, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_stop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1494
 *             future._log_destroy_pending = False
 * 
 *         def done_cb(fut):             # <<<<<<<<<<<<<<
 *             if not fut.cancelled():
 *                 exc = fut.exception()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Loop.run_until_complete.done_cb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1474
 *         return self._task_factory
 * 
 *     def run_until_complete(self, future):             # <<<<<<<<<<<<<<
 *         """Run until the Future is done.
 * 
 */

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_44run_until_complete(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_future) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *__pyx_cur_scope;
  int __pyx_v_new_task;
  PyObject *__pyx_v_done_cb = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_until_complete", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_3_run_until_complete(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1474, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_INCREF(__pyx_v_future);

  /* "uvloop/loop.pyx":1485
 *         Return the Future's result, or raise its exception.
 *         """
 *         self._check_closed()             # <<<<<<<<<<<<<<
 * 
 *         new_task = not isfuture(future)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1487
 *         self._check_closed()
 * 
 *         new_task = not isfuture(future)             # <<<<<<<<<<<<<<
 *         future = aio_ensure_future(future, loop=self)
 *         if new_task:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_isfuture(__pyx_v_future); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_new_task = (!__pyx_t_2);

  /* "uvloop/loop.pyx":1488
 * 
 *         new_task = not isfuture(future)
 *         future = aio_ensure_future(future, loop=self)             # <<<<<<<<<<<<<<
 *         if new_task:
 *             # An exception is raised if the future didn't complete, so there
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_future);
  __Pyx_GIVEREF(__pyx_v_future);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_future)) __PYX_ERR(0, 1488, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_loop, ((PyObject *)__pyx_cur_scope->__pyx_v_self)) < 0) __PYX_ERR(0, 1488, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_ensure_future, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_future, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1489
 *         new_task = not isfuture(future)
 *         future = aio_ensure_future(future, loop=self)
 *         if new_task:             # <<<<<<<<<<<<<<
 *             # An exception is raised if the future didn't complete, so there
 *             # is no need to log the "destroy pending task" message
 */
  if (__pyx_v_new_task) {

    /* "uvloop/loop.pyx":1492
 *             # An exception is raised if the future didn't complete, so there
 *             # is no need to log the "destroy pending task" message
 *             future._log_destroy_pending = False             # <<<<<<<<<<<<<<
 * 
 *         def done_cb(fut):
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_future, __pyx_n_s_log_destroy_pending, Py_False) < 0) __PYX_ERR(0, 1492, __pyx_L1_error)

    /* "uvloop/loop.pyx":1489
 *         new_task = not isfuture(future)
 *         future = aio_ensure_future(future, loop=self)
 *         if new_task:             # <<<<<<<<<<<<<<
 *             # An exception is raised if the future didn't complete, so there
 *             # is no need to log the "destroy pending task" message
 */
  }

  /* "uvloop/loop.pyx":1494
 *             future._log_destroy_pending = False
 * 
 *         def done_cb(fut):             # <<<<<<<<<<<<<<
 *             if not fut.cancelled():
 *                 exc = fut.exception()
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_18run_until_complete_1done_cb, 0, __pyx_n_s_run_until_complete_locals_done_c, ((PyObject*)__pyx_cur_scope), __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_done_cb = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1503
 *             self.stop()
 * 
 *         future.add_done_callback(done_cb)             # <<<<<<<<<<<<<<
 *         try:
 *             self.run_forever()
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_add_done_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_done_cb};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1504
 * 
 *         future.add_done_callback(done_cb)
 *         try:             # <<<<<<<<<<<<<<
 *             self.run_forever()
 *         except BaseException:
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "uvloop/loop.pyx":1505
 *         future.add_done_callback(done_cb)
 *         try:
 *             self.run_forever()             # <<<<<<<<<<<<<<
 *         except BaseException:
 *             if new_task and future.done() and not future.cancelled():
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_run_forever); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1505, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1505, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":1504
 * 
 *         future.add_done_callback(done_cb)
 *         try:             # <<<<<<<<<<<<<<
 *             self.run_forever()
 *         except BaseException:
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1506
 *         try:
 *             self.run_forever()
 *         except BaseException:             # <<<<<<<<<<<<<<
 *             if new_task and future.done() and not future.cancelled():
 *                 # The coroutine raised a BaseException. Consume the exception
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.run_until_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 1506, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_1);

        /* "uvloop/loop.pyx":1507
 *             self.run_forever()
 *         except BaseException:
 *             if new_task and future.done() and not future.cancelled():             # <<<<<<<<<<<<<<
 *                 # The coroutine raised a BaseException. Consume the exception
 *                 # to not log a warning, the caller doesn't have access to the
 */
        if (__pyx_v_new_task) {
        } else {
          __pyx_t_2 = __pyx_v_new_task;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_done); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1507, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1507, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1507, __pyx_L9_except_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_13) {
        } else {
          __pyx_t_2 = __pyx_t_13;
          goto __pyx_L16_bool_binop_done;
        }
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1507, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1507, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1507, __pyx_L9_except_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_14 = (!__pyx_t_13);
        __pyx_t_2 = __pyx_t_14;
        __pyx_L16_bool_binop_done:;
        if (__pyx_t_2) {

          /* "uvloop/loop.pyx":1511
 *                 # to not log a warning, the caller doesn't have access to the
 *                 # local task.
 *                 future.exception()             # <<<<<<<<<<<<<<
 *             raise
 *         finally:
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_exception); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1511, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          __pyx_t_5 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_5 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
            __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1511, __pyx_L9_except_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "uvloop/loop.pyx":1507
 *             self.run_forever()
 *         except BaseException:
 *             if new_task and future.done() and not future.cancelled():             # <<<<<<<<<<<<<<
 *                 # The coroutine raised a BaseException. Consume the exception
 *                 # to not log a warning, the caller doesn't have access to the
 */
        }

        /* "uvloop/loop.pyx":1512
 *                 # local task.
 *                 future.exception()
 *             raise             # <<<<<<<<<<<<<<
 *         finally:
 *             future.remove_done_callback(done_cb)
 */
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_1);
        __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_3, __pyx_t_1);
        __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_1 = 0; 
        __PYX_ERR(0, 1512, __pyx_L9_except_error)
      }
      goto __pyx_L9_except_error;

      /* "uvloop/loop.pyx":1504
 * 
 *         future.add_done_callback(done_cb)
 *         try:             # <<<<<<<<<<<<<<
 *             self.run_forever()
 *         except BaseException:
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L5_error;
      __pyx_L12_try_end:;
    }
  }

  /* "uvloop/loop.pyx":1514
 *             raise
 *         finally:
 *             future.remove_done_callback(done_cb)             # <<<<<<<<<<<<<<
 *         if not future.done():
 *             raise RuntimeError('Event loop stopped before Future completed.')
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_remove_done_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_done_cb};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1514, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_9 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_remove_done_callback); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1514, __pyx_L20_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_done_cb};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1514, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
      goto __pyx_L1_error;
      __pyx_L20_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/loop.pyx":1515
 *         finally:
 *             future.remove_done_callback(done_cb)
 *         if not future.done():             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Event loop stopped before Future completed.')
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_done); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_14 = (!__pyx_t_2);
  if (unlikely(__pyx_t_14)) {

    /* "uvloop/loop.pyx":1516
 *             future.remove_done_callback(done_cb)
 *         if not future.done():
 *             raise RuntimeError('Event loop stopped before Future completed.')             # <<<<<<<<<<<<<<
 * 
 *         return future.result()
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1516, __pyx_L1_error)

    /* "uvloop/loop.pyx":1515
 *         finally:
 *             future.remove_done_callback(done_cb)
 *         if not future.done():             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Event loop stopped before Future completed.')
 * 
 */
  }

  /* "uvloop/loop.pyx":1518
 *             raise RuntimeError('Event loop stopped before Future completed.')
 * 
 *         return future.result()             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":1474
 *         return self._task_factory
 * 
 *     def run_until_complete(self, future):             # <<<<<<<<<<<<<<
 *         """Run until the Future is done.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uvloop.loop.Loop.run_until_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_done_cb);
  __Pyx_XDECREF(__pyx_v_future);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_48generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":1520
 *         return future.result()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getaddrinfo(self, object host, object port, *,
 *                           int family=0, int type=0, int proto=0, int flags=0):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_47getaddrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_46getaddrinfo, "Loop.getaddrinfo(self, host, port, *, int family=0, int type=0, int proto=0, int flags=0)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_47getaddrinfo = {"getaddrinfo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_47getaddrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_46getaddrinfo};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_47getaddrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_host = 0;
  PyObject *__pyx_v_port = 0;
  int __pyx_v_family;
  int __pyx_v_type;
  int __pyx_v_proto;
  int __pyx_v_flags;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getaddrinfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_host,&__pyx_n_s_port,&__pyx_n_s_family,&__pyx_n_s_type,&__pyx_n_s_proto,&__pyx_n_s_flags,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_host)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1520, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_port)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1520, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("getaddrinfo", 1, 2, 2, 1); __PYX_ERR(0, 1520, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 4)) {
        Py_ssize_t index;
        for (index = 2; index < 6 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1520, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "getaddrinfo") < 0)) __PYX_ERR(0, 1520, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_host = values[0];
    __pyx_v_port = values[1];
    if (values[2]) {
      __pyx_v_family = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_family == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1522, __pyx_L3_error)
    } else {
      __pyx_v_family = ((int)0);
    }
    if (values[3]) {
      __pyx_v_type = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1522, __pyx_L3_error)
    } else {
      __pyx_v_type = ((int)0);
    }
    if (values[4]) {
      __pyx_v_proto = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_proto == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1522, __pyx_L3_error)
    } else {
      __pyx_v_proto = ((int)0);
    }
    if (values[5]) {
      __pyx_v_flags = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1522, __pyx_L3_error)
    } else {
      __pyx_v_flags = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getaddrinfo", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1520, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_46getaddrinfo(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_host, __pyx_v_port, __pyx_v_family, __pyx_v_type, __pyx_v_proto, __pyx_v_flags);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_46getaddrinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_flags) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getaddrinfo", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1520, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_host = __pyx_v_host;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_host);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_host);
  __pyx_cur_scope->__pyx_v_port = __pyx_v_port;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_port);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_port);
  __pyx_cur_scope->__pyx_v_family = __pyx_v_family;
  __pyx_cur_scope->__pyx_v_type = __pyx_v_type;
  __pyx_cur_scope->__pyx_v_proto = __pyx_v_proto;
  __pyx_cur_scope->__pyx_v_flags = __pyx_v_flags;
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_48generator, __pyx_codeobj__43, (PyObject *) __pyx_cur_scope, __pyx_n_s_getaddrinfo, __pyx_n_s_Loop_getaddrinfo, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 1520, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_48generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getaddrinfo", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1520, __pyx_L1_error)

  /* "uvloop/loop.pyx":1524
 *                           int family=0, int type=0, int proto=0, int flags=0):
 * 
 *         addr = __static_getaddrinfo_pyaddr(host, port, family,             # <<<<<<<<<<<<<<
 *                                            type, proto, flags)
 *         if addr is not None:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___static_getaddrinfo_pyaddr(__pyx_cur_scope->__pyx_v_host, __pyx_cur_scope->__pyx_v_port, __pyx_cur_scope->__pyx_v_family, __pyx_cur_scope->__pyx_v_type, __pyx_cur_scope->__pyx_v_proto, __pyx_cur_scope->__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_addr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1526
 *         addr = __static_getaddrinfo_pyaddr(host, port, family,
 *                                            type, proto, flags)
 *         if addr is not None:             # <<<<<<<<<<<<<<
 *             return [addr]
 * 
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_addr != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1527
 *                                            type, proto, flags)
 *         if addr is not None:
 *             return [addr]             # <<<<<<<<<<<<<<
 * 
 *         return await self._getaddrinfo(
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_addr);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_addr);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_v_addr)) __PYX_ERR(0, 1527, __pyx_L1_error);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":1526
 *         addr = __static_getaddrinfo_pyaddr(host, port, family,
 *                                            type, proto, flags)
 *         if addr is not None:             # <<<<<<<<<<<<<<
 *             return [addr]
 * 
 */
  }

  /* "uvloop/loop.pyx":1529
 *             return [addr]
 * 
 *         return await self._getaddrinfo(             # <<<<<<<<<<<<<<
 *             host, port, family, type, proto, flags, 1)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "uvloop/loop.pyx":1530
 * 
 *         return await self._getaddrinfo(
 *             host, port, family, type, proto, flags, 1)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getaddrinfo(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_host, __pyx_cur_scope->__pyx_v_port, __pyx_cur_scope->__pyx_v_family, __pyx_cur_scope->__pyx_v_type, __pyx_cur_scope->__pyx_v_proto, __pyx_cur_scope->__pyx_v_flags, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L5_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1529, __pyx_L1_error)
    __pyx_t_1 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_1);
  } else {
    __pyx_t_1 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_1) < 0) __PYX_ERR(0, 1529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":1520
 *         return future.result()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getaddrinfo(self, object host, object port, *,
 *                           int family=0, int type=0, int proto=0, int flags=0):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_51generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":1532
 *             host, port, family, type, proto, flags, 1)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getnameinfo(self, sockaddr, int flags=0):
 *         cdef:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_50getnameinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_49getnameinfo, "Loop.getnameinfo(self, sockaddr, int flags=0)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_50getnameinfo = {"getnameinfo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_50getnameinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_49getnameinfo};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_50getnameinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sockaddr = 0;
  int __pyx_v_flags;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getnameinfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sockaddr,&__pyx_n_s_flags,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sockaddr)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1532, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_flags);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1532, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "getnameinfo") < 0)) __PYX_ERR(0, 1532, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_sockaddr = values[0];
    if (values[1]) {
      __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1533, __pyx_L3_error)
    } else {
      __pyx_v_flags = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("getnameinfo", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1532, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.getnameinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_49getnameinfo(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sockaddr, __pyx_v_flags);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_49getnameinfo(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sockaddr, int __pyx_v_flags) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getnameinfo", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_5_getnameinfo(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1532, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sockaddr = __pyx_v_sockaddr;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sockaddr);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sockaddr);
  __pyx_cur_scope->__pyx_v_flags = __pyx_v_flags;
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_51generator1, __pyx_codeobj__44, (PyObject *) __pyx_cur_scope, __pyx_n_s_getnameinfo, __pyx_n_s_Loop_getnameinfo, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 1532, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.getnameinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_51generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  struct addrinfo *__pyx_t_7;
  int __pyx_t_8;
  unsigned long __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getnameinfo", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L16_resume_from_await;
    case 2: goto __pyx_L20_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1532, __pyx_L1_error)

  /* "uvloop/loop.pyx":1539
 *             system.sockaddr_in6 *sin6
 * 
 *         if not isinstance(sockaddr, tuple):             # <<<<<<<<<<<<<<
 *             raise TypeError('getnameinfo() argument 1 must be a tuple')
 * 
 */
  __pyx_t_1 = PyTuple_Check(__pyx_cur_scope->__pyx_v_sockaddr); 
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/loop.pyx":1540
 * 
 *         if not isinstance(sockaddr, tuple):
 *             raise TypeError('getnameinfo() argument 1 must be a tuple')             # <<<<<<<<<<<<<<
 * 
 *         sl = len(sockaddr)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1540, __pyx_L1_error)

    /* "uvloop/loop.pyx":1539
 *             system.sockaddr_in6 *sin6
 * 
 *         if not isinstance(sockaddr, tuple):             # <<<<<<<<<<<<<<
 *             raise TypeError('getnameinfo() argument 1 must be a tuple')
 * 
 */
  }

  /* "uvloop/loop.pyx":1542
 *             raise TypeError('getnameinfo() argument 1 must be a tuple')
 * 
 *         sl = len(sockaddr)             # <<<<<<<<<<<<<<
 * 
 *         if sl < 2 or sl > 4:
 */
  __pyx_t_4 = PyObject_Length(__pyx_cur_scope->__pyx_v_sockaddr); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1542, __pyx_L1_error)
  __pyx_cur_scope->__pyx_v_sl = __pyx_t_4;

  /* "uvloop/loop.pyx":1544
 *         sl = len(sockaddr)
 * 
 *         if sl < 2 or sl > 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('sockaddr must be a tuple of 2, 3 or 4 values')
 * 
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_sl < 2);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_sl > 4);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/loop.pyx":1545
 * 
 *         if sl < 2 or sl > 4:
 *             raise ValueError('sockaddr must be a tuple of 2, 3 or 4 values')             # <<<<<<<<<<<<<<
 * 
 *         if sl > 2:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1545, __pyx_L1_error)

    /* "uvloop/loop.pyx":1544
 *         sl = len(sockaddr)
 * 
 *         if sl < 2 or sl > 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('sockaddr must be a tuple of 2, 3 or 4 values')
 * 
 */
  }

  /* "uvloop/loop.pyx":1547
 *             raise ValueError('sockaddr must be a tuple of 2, 3 or 4 values')
 * 
 *         if sl > 2:             # <<<<<<<<<<<<<<
 *             flowinfo = sockaddr[2]
 *             if flowinfo < 0 or flowinfo > 0xfffff:
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sl > 2);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1548
 * 
 *         if sl > 2:
 *             flowinfo = sockaddr[2]             # <<<<<<<<<<<<<<
 *             if flowinfo < 0 or flowinfo > 0xfffff:
 *                 raise OverflowError(
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_sockaddr, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_flowinfo = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1549
 *         if sl > 2:
 *             flowinfo = sockaddr[2]
 *             if flowinfo < 0 or flowinfo > 0xfffff:             # <<<<<<<<<<<<<<
 *                 raise OverflowError(
 *                     'getnameinfo(): flowinfo must be 0-1048575.')
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_flowinfo, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1549, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1549, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_flowinfo, __pyx_int_1048575, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1549, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1549, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_1;
    __pyx_L10_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1550
 *             flowinfo = sockaddr[2]
 *             if flowinfo < 0 or flowinfo > 0xfffff:
 *                 raise OverflowError(             # <<<<<<<<<<<<<<
 *                     'getnameinfo(): flowinfo must be 0-1048575.')
 *         else:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1550, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1550, __pyx_L1_error)

      /* "uvloop/loop.pyx":1549
 *         if sl > 2:
 *             flowinfo = sockaddr[2]
 *             if flowinfo < 0 or flowinfo > 0xfffff:             # <<<<<<<<<<<<<<
 *                 raise OverflowError(
 *                     'getnameinfo(): flowinfo must be 0-1048575.')
 */
    }

    /* "uvloop/loop.pyx":1547
 *             raise ValueError('sockaddr must be a tuple of 2, 3 or 4 values')
 * 
 *         if sl > 2:             # <<<<<<<<<<<<<<
 *             flowinfo = sockaddr[2]
 *             if flowinfo < 0 or flowinfo > 0xfffff:
 */
    goto __pyx_L8;
  }

  /* "uvloop/loop.pyx":1553
 *                     'getnameinfo(): flowinfo must be 0-1048575.')
 *         else:
 *             flowinfo = 0             # <<<<<<<<<<<<<<
 * 
 *         if sl > 3:
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    __pyx_cur_scope->__pyx_v_flowinfo = __pyx_int_0;
  }
  __pyx_L8:;

  /* "uvloop/loop.pyx":1555
 *             flowinfo = 0
 * 
 *         if sl > 3:             # <<<<<<<<<<<<<<
 *             scope_id = sockaddr[3]
 *             if scope_id < 0 or scope_id > 2 ** 32:
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sl > 3);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1556
 * 
 *         if sl > 3:
 *             scope_id = sockaddr[3]             # <<<<<<<<<<<<<<
 *             if scope_id < 0 or scope_id > 2 ** 32:
 *                 raise OverflowError(
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_sockaddr, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_scope_id = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1557
 *         if sl > 3:
 *             scope_id = sockaddr[3]
 *             if scope_id < 0 or scope_id > 2 ** 32:             # <<<<<<<<<<<<<<
 *                 raise OverflowError(
 *                     'getsockaddrarg: scope_id must be unsigned 32 bit integer')
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_scope_id, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1557, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_scope_id, __pyx_int_4294967296, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1557, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1557, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_1;
    __pyx_L14_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1558
 *             scope_id = sockaddr[3]
 *             if scope_id < 0 or scope_id > 2 ** 32:
 *                 raise OverflowError(             # <<<<<<<<<<<<<<
 *                     'getsockaddrarg: scope_id must be unsigned 32 bit integer')
 *         else:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1558, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1558, __pyx_L1_error)

      /* "uvloop/loop.pyx":1557
 *         if sl > 3:
 *             scope_id = sockaddr[3]
 *             if scope_id < 0 or scope_id > 2 ** 32:             # <<<<<<<<<<<<<<
 *                 raise OverflowError(
 *                     'getsockaddrarg: scope_id must be unsigned 32 bit integer')
 */
    }

    /* "uvloop/loop.pyx":1555
 *             flowinfo = 0
 * 
 *         if sl > 3:             # <<<<<<<<<<<<<<
 *             scope_id = sockaddr[3]
 *             if scope_id < 0 or scope_id > 2 ** 32:
 */
    goto __pyx_L12;
  }

  /* "uvloop/loop.pyx":1561
 *                     'getsockaddrarg: scope_id must be unsigned 32 bit integer')
 *         else:
 *             scope_id = 0             # <<<<<<<<<<<<<<
 * 
 *         ai_cnt = await self._getaddrinfo(
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    __pyx_cur_scope->__pyx_v_scope_id = __pyx_int_0;
  }
  __pyx_L12:;

  /* "uvloop/loop.pyx":1564
 * 
 *         ai_cnt = await self._getaddrinfo(
 *             sockaddr[0], sockaddr[1],             # <<<<<<<<<<<<<<
 *             uv.AF_UNSPEC,         # family
 *             uv.SOCK_DGRAM,        # type
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_sockaddr, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_sockaddr, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "uvloop/loop.pyx":1563
 *             scope_id = 0
 * 
 *         ai_cnt = await self._getaddrinfo(             # <<<<<<<<<<<<<<
 *             sockaddr[0], sockaddr[1],
 *             uv.AF_UNSPEC,         # family
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getaddrinfo(__pyx_cur_scope->__pyx_v_self, __pyx_t_3, __pyx_t_5, AF_UNSPEC, SOCK_DGRAM, 0, AI_NUMERICHOST, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L16_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1563, __pyx_L1_error)
    __pyx_t_6 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_6);
  } else {
    __pyx_t_6 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_6) < 0) __PYX_ERR(0, 1563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_6uvloop_4loop_AddrInfo))))) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_cur_scope->__pyx_v_ai_cnt = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1571
 *             0)                    # unpack
 * 
 *         ai = ai_cnt.data             # <<<<<<<<<<<<<<
 * 
 *         if ai.ai_next:
 */
  __pyx_t_7 = __pyx_cur_scope->__pyx_v_ai_cnt->data;
  __pyx_cur_scope->__pyx_v_ai = __pyx_t_7;

  /* "uvloop/loop.pyx":1573
 *         ai = ai_cnt.data
 * 
 *         if ai.ai_next:             # <<<<<<<<<<<<<<
 *             raise OSError("sockaddr resolved to multiple addresses")
 * 
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ai->ai_next != 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/loop.pyx":1574
 * 
 *         if ai.ai_next:
 *             raise OSError("sockaddr resolved to multiple addresses")             # <<<<<<<<<<<<<<
 * 
 *         if ai.ai_family == uv.AF_INET:
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__49, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1574, __pyx_L1_error)

    /* "uvloop/loop.pyx":1573
 *         ai = ai_cnt.data
 * 
 *         if ai.ai_next:             # <<<<<<<<<<<<<<
 *             raise OSError("sockaddr resolved to multiple addresses")
 * 
 */
  }

  /* "uvloop/loop.pyx":1576
 *             raise OSError("sockaddr resolved to multiple addresses")
 * 
 *         if ai.ai_family == uv.AF_INET:             # <<<<<<<<<<<<<<
 *             if sl > 2:
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ai->ai_family == AF_INET);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1577
 * 
 *         if ai.ai_family == uv.AF_INET:
 *             if sl > 2:             # <<<<<<<<<<<<<<
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")
 *         elif ai.ai_family == uv.AF_INET6:
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sl > 2);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1578
 *         if ai.ai_family == uv.AF_INET:
 *             if sl > 2:
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")             # <<<<<<<<<<<<<<
 *         elif ai.ai_family == uv.AF_INET6:
 *             # Modify some fields in `ai`
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1578, __pyx_L1_error)

      /* "uvloop/loop.pyx":1577
 * 
 *         if ai.ai_family == uv.AF_INET:
 *             if sl > 2:             # <<<<<<<<<<<<<<
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")
 *         elif ai.ai_family == uv.AF_INET6:
 */
    }

    /* "uvloop/loop.pyx":1576
 *             raise OSError("sockaddr resolved to multiple addresses")
 * 
 *         if ai.ai_family == uv.AF_INET:             # <<<<<<<<<<<<<<
 *             if sl > 2:
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")
 */
    goto __pyx_L18;
  }

  /* "uvloop/loop.pyx":1579
 *             if sl > 2:
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")
 *         elif ai.ai_family == uv.AF_INET6:             # <<<<<<<<<<<<<<
 *             # Modify some fields in `ai`
 *             sin6 = <system.sockaddr_in6*> ai.ai_addr
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ai->ai_family == AF_INET6);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1581
 *         elif ai.ai_family == uv.AF_INET6:
 *             # Modify some fields in `ai`
 *             sin6 = <system.sockaddr_in6*> ai.ai_addr             # <<<<<<<<<<<<<<
 *             sin6.sin6_flowinfo = system.htonl(flowinfo)
 *             sin6.sin6_scope_id = scope_id
 */
    __pyx_cur_scope->__pyx_v_sin6 = ((struct sockaddr_in6 *)__pyx_cur_scope->__pyx_v_ai->ai_addr);

    /* "uvloop/loop.pyx":1582
 *             # Modify some fields in `ai`
 *             sin6 = <system.sockaddr_in6*> ai.ai_addr
 *             sin6.sin6_flowinfo = system.htonl(flowinfo)             # <<<<<<<<<<<<<<
 *             sin6.sin6_scope_id = scope_id
 * 
 */
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_flowinfo); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1582, __pyx_L1_error)
    __pyx_cur_scope->__pyx_v_sin6->sin6_flowinfo = htonl(__pyx_t_8);

    /* "uvloop/loop.pyx":1583
 *             sin6 = <system.sockaddr_in6*> ai.ai_addr
 *             sin6.sin6_flowinfo = system.htonl(flowinfo)
 *             sin6.sin6_scope_id = scope_id             # <<<<<<<<<<<<<<
 * 
 *         return await self._getnameinfo(ai.ai_addr, flags)
 */
    __pyx_t_9 = __Pyx_PyInt_As_unsigned_long(__pyx_cur_scope->__pyx_v_scope_id); if (unlikely((__pyx_t_9 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1583, __pyx_L1_error)
    __pyx_cur_scope->__pyx_v_sin6->sin6_scope_id = __pyx_t_9;

    /* "uvloop/loop.pyx":1579
 *             if sl > 2:
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")
 *         elif ai.ai_family == uv.AF_INET6:             # <<<<<<<<<<<<<<
 *             # Modify some fields in `ai`
 *             sin6 = <system.sockaddr_in6*> ai.ai_addr
 */
  }
  __pyx_L18:;

  /* "uvloop/loop.pyx":1585
 *             sin6.sin6_scope_id = scope_id
 * 
 *         return await self._getnameinfo(ai.ai_addr, flags)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getnameinfo(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_ai->ai_addr, __pyx_cur_scope->__pyx_v_flags); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 2;
    return __pyx_r;
    __pyx_L20_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1585, __pyx_L1_error)
    __pyx_t_6 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_6);
  } else {
    __pyx_t_6 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_6) < 0) __PYX_ERR(0, 1585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":1532
 *             host, port, family, type, proto, flags, 1)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getnameinfo(self, sockaddr, int flags=0):
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("getnameinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_54generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":1587
 *         return await self._getnameinfo(ai.ai_addr, flags)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_53start_tls(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_52start_tls, "Loop.start_tls(self, transport, protocol, sslcontext, *, server_side=False, server_hostname=None, ssl_handshake_timeout=None, ssl_shutdown_timeout=None)\nUpgrade transport to TLS.\n\n        Return a new transport that *protocol* should start using\n        immediately.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_53start_tls = {"start_tls", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_53start_tls, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_52start_tls};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_53start_tls(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_transport = 0;
  PyObject *__pyx_v_protocol = 0;
  PyObject *__pyx_v_sslcontext = 0;
  PyObject *__pyx_v_server_side = 0;
  PyObject *__pyx_v_server_hostname = 0;
  PyObject *__pyx_v_ssl_handshake_timeout = 0;
  PyObject *__pyx_v_ssl_shutdown_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_tls (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_transport,&__pyx_n_s_protocol,&__pyx_n_s_sslcontext,&__pyx_n_s_server_side,&__pyx_n_s_server_hostname,&__pyx_n_s_ssl_handshake_timeout,&__pyx_n_s_ssl_shutdown_timeout,0};

    /* "uvloop/loop.pyx":1589
 *     @cython.iterable_coroutine
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,             # <<<<<<<<<<<<<<
 *                         server_hostname=None,
 *                         ssl_handshake_timeout=None,
 */
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));

    /* "uvloop/loop.pyx":1590
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 *                         server_hostname=None,             # <<<<<<<<<<<<<<
 *                         ssl_handshake_timeout=None,
 *                         ssl_shutdown_timeout=None):
 */
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1591
 *                         server_side=False,
 *                         server_hostname=None,
 *                         ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                         ssl_shutdown_timeout=None):
 *         """Upgrade transport to TLS.
 */
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1592
 *                         server_hostname=None,
 *                         ssl_handshake_timeout=None,
 *                         ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 *         """Upgrade transport to TLS.
 * 
 */
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_transport)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("start_tls", 1, 3, 3, 1); __PYX_ERR(0, 1587, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sslcontext)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("start_tls", 1, 3, 3, 2); __PYX_ERR(0, 1587, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 4)) {
        Py_ssize_t index;
        for (index = 3; index < 7 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "start_tls") < 0)) __PYX_ERR(0, 1587, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_transport = values[0];
    __pyx_v_protocol = values[1];
    __pyx_v_sslcontext = values[2];
    __pyx_v_server_side = values[3];
    __pyx_v_server_hostname = values[4];
    __pyx_v_ssl_handshake_timeout = values[5];
    __pyx_v_ssl_shutdown_timeout = values[6];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start_tls", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1587, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.start_tls", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_52start_tls(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_transport, __pyx_v_protocol, __pyx_v_sslcontext, __pyx_v_server_side, __pyx_v_server_hostname, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout);

  /* "uvloop/loop.pyx":1587
 *         return await self._getnameinfo(ai.ai_addr, flags)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_52start_tls(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_transport, PyObject *__pyx_v_protocol, PyObject *__pyx_v_sslcontext, PyObject *__pyx_v_server_side, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start_tls", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_6_start_tls(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1587, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_transport = __pyx_v_transport;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_transport);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_transport);
  __pyx_cur_scope->__pyx_v_protocol = __pyx_v_protocol;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
  __pyx_cur_scope->__pyx_v_sslcontext = __pyx_v_sslcontext;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sslcontext);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sslcontext);
  __pyx_cur_scope->__pyx_v_server_side = __pyx_v_server_side;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_side);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_side);
  __pyx_cur_scope->__pyx_v_server_hostname = __pyx_v_server_hostname;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_hostname);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_hostname);
  __pyx_cur_scope->__pyx_v_ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_54generator2, __pyx_codeobj__51, (PyObject *) __pyx_cur_scope, __pyx_n_s_start_tls, __pyx_n_s_Loop_start_tls, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.start_tls", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_54generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_tls", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L14_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1587, __pyx_L1_error)

  /* "uvloop/loop.pyx":1598
 *         immediately.
 *         """
 *         if not isinstance(sslcontext, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 f'sslcontext is expected to be an instance of ssl.SSLContext, '
 */
  __pyx_t_1 = __pyx_v_6uvloop_4loop_ssl_SSLContext;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_sslcontext, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":1601
 *             raise TypeError(
 *                 f'sslcontext is expected to be an instance of ssl.SSLContext, '
 *                 f'got {sslcontext!r}')             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(transport, (TCPTransport, UnixTransport)):
 */
    __pyx_t_1 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_cur_scope->__pyx_v_sslcontext), __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "uvloop/loop.pyx":1600
 *         if not isinstance(sslcontext, ssl_SSLContext):
 *             raise TypeError(
 *                 f'sslcontext is expected to be an instance of ssl.SSLContext, '             # <<<<<<<<<<<<<<
 *                 f'got {sslcontext!r}')
 * 
 */
    __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_sslcontext_is_expected_to_be_an, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":1599
 *         """
 *         if not isinstance(sslcontext, ssl_SSLContext):
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 f'sslcontext is expected to be an instance of ssl.SSLContext, '
 *                 f'got {sslcontext!r}')
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1599, __pyx_L1_error)

    /* "uvloop/loop.pyx":1598
 *         immediately.
 *         """
 *         if not isinstance(sslcontext, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 f'sslcontext is expected to be an instance of ssl.SSLContext, '
 */
  }

  /* "uvloop/loop.pyx":1603
 *                 f'got {sslcontext!r}')
 * 
 *         if isinstance(transport, (TCPTransport, UnixTransport)):             # <<<<<<<<<<<<<<
 *             context = (<UVStream>transport).context
 *         elif isinstance(transport, _SSLProtocolTransport):
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_transport, __pyx_ptype_6uvloop_4loop_TCPTransport); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_transport, __pyx_ptype_6uvloop_4loop_UnixTransport); 
  __pyx_t_3 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":1604
 * 
 *         if isinstance(transport, (TCPTransport, UnixTransport)):
 *             context = (<UVStream>transport).context             # <<<<<<<<<<<<<<
 *         elif isinstance(transport, _SSLProtocolTransport):
 *             context = (<_SSLProtocolTransport>transport).context
 */
    __pyx_t_1 = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_cur_scope->__pyx_v_transport)->__pyx_base.__pyx_base.__pyx_base.context;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_context = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":1603
 *                 f'got {sslcontext!r}')
 * 
 *         if isinstance(transport, (TCPTransport, UnixTransport)):             # <<<<<<<<<<<<<<
 *             context = (<UVStream>transport).context
 *         elif isinstance(transport, _SSLProtocolTransport):
 */
    goto __pyx_L5;
  }

  /* "uvloop/loop.pyx":1605
 *         if isinstance(transport, (TCPTransport, UnixTransport)):
 *             context = (<UVStream>transport).context
 *         elif isinstance(transport, _SSLProtocolTransport):             # <<<<<<<<<<<<<<
 *             context = (<_SSLProtocolTransport>transport).context
 *         else:
 */
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_cur_scope->__pyx_v_transport, __pyx_ptype_6uvloop_4loop__SSLProtocolTransport); 
  if (likely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":1606
 *             context = (<UVStream>transport).context
 *         elif isinstance(transport, _SSLProtocolTransport):
 *             context = (<_SSLProtocolTransport>transport).context             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError(
 */
    __pyx_t_1 = ((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_cur_scope->__pyx_v_transport)->context;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_context = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":1605
 *         if isinstance(transport, (TCPTransport, UnixTransport)):
 *             context = (<UVStream>transport).context
 *         elif isinstance(transport, _SSLProtocolTransport):             # <<<<<<<<<<<<<<
 *             context = (<_SSLProtocolTransport>transport).context
 *         else:
 */
    goto __pyx_L5;
  }

  /* "uvloop/loop.pyx":1608
 *             context = (<_SSLProtocolTransport>transport).context
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 f'transport {transport!r} is not supported by start_tls()')
 * 
 */
  /*else*/ {

    /* "uvloop/loop.pyx":1609
 *         else:
 *             raise TypeError(
 *                 f'transport {transport!r} is not supported by start_tls()')             # <<<<<<<<<<<<<<
 * 
 *         waiter = self._new_future()
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_transport_2);
    __pyx_t_5 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_transport_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_transport_2);
    __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_cur_scope->__pyx_v_transport), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_not_supported_by_start_tls);
    __pyx_t_5 += 32;
    __Pyx_GIVEREF(__pyx_kp_u_is_not_supported_by_start_tls);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_is_not_supported_by_start_tls);
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":1608
 *             context = (<_SSLProtocolTransport>transport).context
 *         else:
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 f'transport {transport!r} is not supported by start_tls()')
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1608, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "uvloop/loop.pyx":1611
 *                 f'transport {transport!r} is not supported by start_tls()')
 * 
 *         waiter = self._new_future()             # <<<<<<<<<<<<<<
 *         ssl_protocol = SSLProtocol(
 *             self, protocol, sslcontext, waiter,
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":1612
 * 
 *         waiter = self._new_future()
 *         ssl_protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *             self, protocol, sslcontext, waiter,
 *             server_side, server_hostname,
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(0, 1612, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_protocol)) __PYX_ERR(0, 1612, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sslcontext);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sslcontext);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_cur_scope->__pyx_v_sslcontext)) __PYX_ERR(0, 1612, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_waiter);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_waiter);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_cur_scope->__pyx_v_waiter)) __PYX_ERR(0, 1612, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_side);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_side);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_cur_scope->__pyx_v_server_side)) __PYX_ERR(0, 1612, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_hostname);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_hostname);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_cur_scope->__pyx_v_server_hostname)) __PYX_ERR(0, 1612, __pyx_L1_error);

  /* "uvloop/loop.pyx":1615
 *             self, protocol, sslcontext, waiter,
 *             server_side, server_hostname,
 *             ssl_handshake_timeout=ssl_handshake_timeout,             # <<<<<<<<<<<<<<
 *             ssl_shutdown_timeout=ssl_shutdown_timeout,
 *             call_connection_made=False)
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_ssl_handshake_timeout, __pyx_cur_scope->__pyx_v_ssl_handshake_timeout) < 0) __PYX_ERR(0, 1615, __pyx_L1_error)

  /* "uvloop/loop.pyx":1616
 *             server_side, server_hostname,
 *             ssl_handshake_timeout=ssl_handshake_timeout,
 *             ssl_shutdown_timeout=ssl_shutdown_timeout,             # <<<<<<<<<<<<<<
 *             call_connection_made=False)
 * 
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_ssl_shutdown_timeout, __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout) < 0) __PYX_ERR(0, 1615, __pyx_L1_error)

  /* "uvloop/loop.pyx":1617
 *             ssl_handshake_timeout=ssl_handshake_timeout,
 *             ssl_shutdown_timeout=ssl_shutdown_timeout,
 *             call_connection_made=False)             # <<<<<<<<<<<<<<
 * 
 *         # Pause early so that "ssl_protocol.data_received()" doesn't
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_call_connection_made, Py_False) < 0) __PYX_ERR(0, 1615, __pyx_L1_error)

  /* "uvloop/loop.pyx":1612
 * 
 *         waiter = self._new_future()
 *         ssl_protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *             self, protocol, sslcontext, waiter,
 *             server_side, server_hostname,
 */
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol), __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_cur_scope->__pyx_v_ssl_protocol = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":1621
 *         # Pause early so that "ssl_protocol.data_received()" doesn't
 *         # have a chance to get called before "ssl_protocol.connection_made()".
 *         transport.pause_reading()             # <<<<<<<<<<<<<<
 * 
 *         transport.set_protocol(ssl_protocol)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_transport, __pyx_n_s_pause_reading); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":1623
 *         transport.pause_reading()
 * 
 *         transport.set_protocol(ssl_protocol)             # <<<<<<<<<<<<<<
 *         conmade_cb = self.call_soon(ssl_protocol.connection_made, transport,
 *                                     context=context)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_transport, __pyx_n_s_set_protocol); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, ((PyObject *)__pyx_cur_scope->__pyx_v_ssl_protocol)};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uvloop/loop.pyx":1624
 * 
 *         transport.set_protocol(ssl_protocol)
 *         conmade_cb = self.call_soon(ssl_protocol.connection_made, transport,             # <<<<<<<<<<<<<<
 *                                     context=context)
 *         # transport.resume_reading() will use the right context
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_ssl_protocol), __pyx_n_s_connection_made); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(0, 1624, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_transport);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_transport);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_transport)) __PYX_ERR(0, 1624, __pyx_L1_error);
  __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":1625
 *         transport.set_protocol(ssl_protocol)
 *         conmade_cb = self.call_soon(ssl_protocol.connection_made, transport,
 *                                     context=context)             # <<<<<<<<<<<<<<
 *         # transport.resume_reading() will use the right context
 *         # (transport.context) to call e.g. data_received()
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_context, __pyx_cur_scope->__pyx_v_context) < 0) __PYX_ERR(0, 1625, __pyx_L1_error)

  /* "uvloop/loop.pyx":1624
 * 
 *         transport.set_protocol(ssl_protocol)
 *         conmade_cb = self.call_soon(ssl_protocol.connection_made, transport,             # <<<<<<<<<<<<<<
 *                                     context=context)
 *         # transport.resume_reading() will use the right context
 */
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_conmade_cb = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "uvloop/loop.pyx":1628
 *         # transport.resume_reading() will use the right context
 *         # (transport.context) to call e.g. data_received()
 *         resume_cb = self.call_soon(transport.resume_reading)             # <<<<<<<<<<<<<<
 *         app_transport = ssl_protocol._get_app_transport(context)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_transport, __pyx_n_s_resume_reading); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_1};
    __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_resume_cb = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "uvloop/loop.pyx":1629
 *         # (transport.context) to call e.g. data_received()
 *         resume_cb = self.call_soon(transport.resume_reading)
 *         app_transport = ssl_protocol._get_app_transport(context)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_ssl_protocol), __pyx_n_s_get_app_transport); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_cur_scope->__pyx_v_context};
    __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_app_transport = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "uvloop/loop.pyx":1631
 *         app_transport = ssl_protocol._get_app_transport(context)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_12);
    /*try:*/ {

      /* "uvloop/loop.pyx":1632
 * 
 *         try:
 *             await waiter             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_10;
        __Pyx_XGIVEREF(__pyx_t_11);
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_11;
        __Pyx_XGIVEREF(__pyx_t_12);
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_12;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L14_resume_from_await:;
        __pyx_t_10 = __pyx_cur_scope->__pyx_t_0;
        __pyx_cur_scope->__pyx_t_0 = 0;
        __Pyx_XGOTREF(__pyx_t_10);
        __pyx_t_11 = __pyx_cur_scope->__pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = 0;
        __Pyx_XGOTREF(__pyx_t_11);
        __pyx_t_12 = __pyx_cur_scope->__pyx_t_2;
        __pyx_cur_scope->__pyx_t_2 = 0;
        __Pyx_XGOTREF(__pyx_t_12);
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1632, __pyx_L8_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 1632, __pyx_L8_error)
        }
      }

      /* "uvloop/loop.pyx":1631
 *         app_transport = ssl_protocol._get_app_transport(context)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
    }
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    goto __pyx_L13_try_end;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "uvloop/loop.pyx":1633
 *         try:
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException:
 */
    __pyx_t_13 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_13) {
      __Pyx_AddTraceback("uvloop.loop.Loop.start_tls", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_4, &__pyx_t_1) < 0) __PYX_ERR(0, 1633, __pyx_L10_except_error)
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_1);

      /* "uvloop/loop.pyx":1634
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException:
 *             app_transport.close()
 */
      __Pyx_GIVEREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ErrRestoreWithState(__pyx_t_9, __pyx_t_4, __pyx_t_1);
      __pyx_t_9 = 0; __pyx_t_4 = 0; __pyx_t_1 = 0; 
      __PYX_ERR(0, 1634, __pyx_L10_except_error)
    }

    /* "uvloop/loop.pyx":1635
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException:             # <<<<<<<<<<<<<<
 *             app_transport.close()
 *             conmade_cb.cancel()
 */
    __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_13) {
      __Pyx_AddTraceback("uvloop.loop.Loop.start_tls", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(0, 1635, __pyx_L10_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_9);

      /* "uvloop/loop.pyx":1636
 *             raise
 *         except BaseException:
 *             app_transport.close()             # <<<<<<<<<<<<<<
 *             conmade_cb.cancel()
 *             resume_cb.cancel()
 */
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_app_transport, __pyx_n_s_close); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1636, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1636, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":1637
 *         except BaseException:
 *             app_transport.close()
 *             conmade_cb.cancel()             # <<<<<<<<<<<<<<
 *             resume_cb.cancel()
 *             raise
 */
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_conmade_cb, __pyx_n_s_cancel); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1637, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1637, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":1638
 *             app_transport.close()
 *             conmade_cb.cancel()
 *             resume_cb.cancel()             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_resume_cb, __pyx_n_s_cancel); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1638, __pyx_L10_except_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1638, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":1639
 *             conmade_cb.cancel()
 *             resume_cb.cancel()
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         return app_transport
 */
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_4, __pyx_t_9);
      __pyx_t_1 = 0; __pyx_t_4 = 0; __pyx_t_9 = 0; 
      __PYX_ERR(0, 1639, __pyx_L10_except_error)
    }
    goto __pyx_L10_except_error;

    /* "uvloop/loop.pyx":1631
 *         app_transport = ssl_protocol._get_app_transport(context)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
    __pyx_L10_except_error:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    goto __pyx_L1_error;
    __pyx_L13_try_end:;
  }

  /* "uvloop/loop.pyx":1641
 *             raise
 * 
 *         return app_transport             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_cur_scope->__pyx_v_app_transport);
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":1587
 *         return await self._getnameinfo(ai.ai_addr, flags)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("start_tls", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_57generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":1643
 *         return app_transport
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_56create_server(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_55create_server, "Loop.create_server(self, protocol_factory, host=None, port=None, *, int family=uv.AF_UNSPEC, int flags=uv.AI_PASSIVE, sock=None, backlog=100, ssl=None, reuse_address=None, reuse_port=None, ssl_handshake_timeout=None, ssl_shutdown_timeout=None, start_serving=True)\nA coroutine which creates a TCP server bound to host and port.\n\n        The return value is a Server object which can be used to stop\n        the service.\n\n        If host is an empty string or None all interfaces are assumed\n        and a list of multiple sockets will be returned (most likely\n        one for IPv4 and another one for IPv6). The host parameter can also be\n        a sequence (e.g. list) of hosts to bind to.\n\n        family can be set to either AF_INET or AF_INET6 to force the\n        socket to use IPv4 or IPv6. If not set it will be determined\n        from host (defaults to AF_UNSPEC).\n\n        flags is a bitmask for getaddrinfo().\n\n        sock can optionally be specified in order to use a preexisting\n        socket object.\n\n        backlog is the maximum number of queued connections passed to\n        listen() (defaults to 100).\n\n        ssl can be set to an SSLContext to enable SSL over the\n        accepted connections.\n\n        reuse_address tells the kernel to reuse a local socket in\n        TIME_WAIT state, without waiting for its natural timeout to\n        expire. If not specified will automatically be set to True on\n        UNIX.\n\n        reuse_port tells the kernel to allow this endpoint to be bound to\n        the same port as other existing endpoints are bound to, so long as\n        they all set this flag when being created. This option is not\n        supported on Windows.\n\n        ssl_handshake_timeout is the time in seconds that an SSL server\n        will wait for completion of the SSL handshake before aborting the\n        connection. Default is 60s.\n\n        ssl_shutdown_timeout is the time in seconds that an SSL server\n        will wait fo""r completion of the SSL shutdown before aborting the\n        connection. Default is 30s.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_56create_server = {"create_server", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_56create_server, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_55create_server};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_56create_server(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_host = 0;
  PyObject *__pyx_v_port = 0;
  int __pyx_v_family;
  int __pyx_v_flags;
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_backlog = 0;
  PyObject *__pyx_v_ssl = 0;
  PyObject *__pyx_v_reuse_address = 0;
  PyObject *__pyx_v_reuse_port = 0;
  PyObject *__pyx_v_ssl_handshake_timeout = 0;
  PyObject *__pyx_v_ssl_shutdown_timeout = 0;
  PyObject *__pyx_v_start_serving = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_server (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_host,&__pyx_n_s_port,&__pyx_n_s_family,&__pyx_n_s_flags,&__pyx_n_s_sock,&__pyx_n_s_backlog,&__pyx_n_s_ssl,&__pyx_n_s_reuse_address,&__pyx_n_s_reuse_port,&__pyx_n_s_ssl_handshake_timeout,&__pyx_n_s_ssl_shutdown_timeout,&__pyx_n_s_start_serving,0};

    /* "uvloop/loop.pyx":1644
 * 
 *     @cython.iterable_coroutine
 *     async def create_server(self, protocol_factory, host=None, port=None,             # <<<<<<<<<<<<<<
 *                             *,
 *                             int family=uv.AF_UNSPEC,
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1648
 *                             int family=uv.AF_UNSPEC,
 *                             int flags=uv.AI_PASSIVE,
 *                             sock=None,             # <<<<<<<<<<<<<<
 *                             backlog=100,
 *                             ssl=None,
 */
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_100));

    /* "uvloop/loop.pyx":1650
 *                             sock=None,
 *                             backlog=100,
 *                             ssl=None,             # <<<<<<<<<<<<<<
 *                             reuse_address=None,
 *                             reuse_port=None,
 */
    values[7] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1651
 *                             backlog=100,
 *                             ssl=None,
 *                             reuse_address=None,             # <<<<<<<<<<<<<<
 *                             reuse_port=None,
 *                             ssl_handshake_timeout=None,
 */
    values[8] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1652
 *                             ssl=None,
 *                             reuse_address=None,
 *                             reuse_port=None,             # <<<<<<<<<<<<<<
 *                             ssl_handshake_timeout=None,
 *                             ssl_shutdown_timeout=None,
 */
    values[9] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1653
 *                             reuse_address=None,
 *                             reuse_port=None,
 *                             ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                             ssl_shutdown_timeout=None,
 *                             start_serving=True):
 */
    values[10] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1654
 *                             reuse_port=None,
 *                             ssl_handshake_timeout=None,
 *                             ssl_shutdown_timeout=None,             # <<<<<<<<<<<<<<
 *                             start_serving=True):
 *         """A coroutine which creates a TCP server bound to host and port.
 */
    values[11] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1655
 *                             ssl_handshake_timeout=None,
 *                             ssl_shutdown_timeout=None,
 *                             start_serving=True):             # <<<<<<<<<<<<<<
 *         """A coroutine which creates a TCP server bound to host and port.
 * 
 */
    values[12] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_host);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_port);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 10)) {
        Py_ssize_t index;
        for (index = 3; index < 13 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_server") < 0)) __PYX_ERR(0, 1643, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_host = values[1];
    __pyx_v_port = values[2];
    if (values[3]) {
      __pyx_v_family = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_family == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1646, __pyx_L3_error)
    } else {
      __pyx_v_family = __pyx_k__52;
    }
    if (values[4]) {
      __pyx_v_flags = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1647, __pyx_L3_error)
    } else {
      __pyx_v_flags = __pyx_k__53;
    }
    __pyx_v_sock = values[5];
    __pyx_v_backlog = values[6];
    __pyx_v_ssl = values[7];
    __pyx_v_reuse_address = values[8];
    __pyx_v_reuse_port = values[9];
    __pyx_v_ssl_handshake_timeout = values[10];
    __pyx_v_ssl_shutdown_timeout = values[11];
    __pyx_v_start_serving = values[12];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_server", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 1643, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_55create_server(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_host, __pyx_v_port, __pyx_v_family, __pyx_v_flags, __pyx_v_sock, __pyx_v_backlog, __pyx_v_ssl, __pyx_v_reuse_address, __pyx_v_reuse_port, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout, __pyx_v_start_serving);

  /* "uvloop/loop.pyx":1643
 *         return app_transport
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_55create_server(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_flags, PyObject *__pyx_v_sock, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_reuse_address, PyObject *__pyx_v_reuse_port, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout, PyObject *__pyx_v_start_serving) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_server", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_7_create_server(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1643, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_host = __pyx_v_host;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_host);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_host);
  __pyx_cur_scope->__pyx_v_port = __pyx_v_port;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_port);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_port);
  __pyx_cur_scope->__pyx_v_family = __pyx_v_family;
  __pyx_cur_scope->__pyx_v_flags = __pyx_v_flags;
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_backlog = __pyx_v_backlog;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_backlog);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_backlog);
  __pyx_cur_scope->__pyx_v_ssl = __pyx_v_ssl;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl);
  __pyx_cur_scope->__pyx_v_reuse_address = __pyx_v_reuse_address;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_reuse_address);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_reuse_address);
  __pyx_cur_scope->__pyx_v_reuse_port = __pyx_v_reuse_port;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_reuse_port);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_reuse_port);
  __pyx_cur_scope->__pyx_v_ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __pyx_cur_scope->__pyx_v_start_serving = __pyx_v_start_serving;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_start_serving);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_start_serving);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_57generator3, __pyx_codeobj__54, (PyObject *) __pyx_cur_scope, __pyx_n_s_create_server, __pyx_n_s_Loop_create_server, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 1643, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_57generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  struct addrinfo *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  unsigned int __pyx_t_17;
  int __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  Py_ssize_t __pyx_t_21;
  Py_UCS4 __pyx_t_22;
  int __pyx_t_23;
  char const *__pyx_t_24;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  char const *__pyx_t_31;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_server", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L10_resume_from_await;
    case 2: goto __pyx_L31_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1704
 *             Server server
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *             if host is not None or port is not None:
 *                 raise ValueError(
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1704, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1704, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1705
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:
 *             if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_host != Py_None);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_port != Py_None);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/loop.pyx":1706
 *         if sock is not None and sock.family == uv.AF_UNIX:
 *             if host is not None or port is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_server(
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1706, __pyx_L1_error)

      /* "uvloop/loop.pyx":1705
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:
 *             if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    }

    /* "uvloop/loop.pyx":1708
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_server(             # <<<<<<<<<<<<<<
 *                 protocol_factory, sock=sock, backlog=backlog, ssl=ssl,
 *                 start_serving=start_serving,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_create_unix_server); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "uvloop/loop.pyx":1709
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_server(
 *                 protocol_factory, sock=sock, backlog=backlog, ssl=ssl,             # <<<<<<<<<<<<<<
 *                 start_serving=start_serving,
 *                 # asyncio won't clean up socket file using create_server() API
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_cur_scope->__pyx_v_protocol_factory)) __PYX_ERR(0, 1708, __pyx_L1_error);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sock, __pyx_cur_scope->__pyx_v_sock) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_backlog, __pyx_cur_scope->__pyx_v_backlog) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl, __pyx_cur_scope->__pyx_v_ssl) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)

    /* "uvloop/loop.pyx":1710
 *             return await self.create_unix_server(
 *                 protocol_factory, sock=sock, backlog=backlog, ssl=ssl,
 *                 start_serving=start_serving,             # <<<<<<<<<<<<<<
 *                 # asyncio won't clean up socket file using create_server() API
 *                 cleanup_socket=False,
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_start_serving, __pyx_cur_scope->__pyx_v_start_serving) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)

    /* "uvloop/loop.pyx":1712
 *                 start_serving=start_serving,
 *                 # asyncio won't clean up socket file using create_server() API
 *                 cleanup_socket=False,             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_cleanup_socket, Py_False) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)

    /* "uvloop/loop.pyx":1708
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_server(             # <<<<<<<<<<<<<<
 *                 protocol_factory, sock=sock, backlog=backlog, ssl=ssl,
 *                 start_serving=start_serving,
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L10_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1708, __pyx_L1_error)
      __pyx_t_6 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_6);
    } else {
      __pyx_t_6 = NULL;
      if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_6) < 0) __PYX_ERR(0, 1708, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":1704
 *             Server server
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *             if host is not None or port is not None:
 *                 raise ValueError(
 */
  }

  /* "uvloop/loop.pyx":1715
 *             )
 * 
 *         server = Server(self)             # <<<<<<<<<<<<<<
 * 
 *         if ssl is not None:
 */
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6uvloop_4loop_Server), ((PyObject *)__pyx_cur_scope->__pyx_v_self)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_cur_scope->__pyx_v_server = ((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1717
 *         server = Server(self)
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_ssl != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1718
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 */
    __pyx_t_6 = __pyx_v_6uvloop_4loop_ssl_SSLContext;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_1 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_ssl, __pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1718, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_2 = (!__pyx_t_1);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1719
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')             # <<<<<<<<<<<<<<
 *         else:
 *             if ssl_handshake_timeout is not None:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1719, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1719, __pyx_L1_error)

      /* "uvloop/loop.pyx":1718
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 */
    }

    /* "uvloop/loop.pyx":1717
 *         server = Server(self)
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 */
    goto __pyx_L11;
  }

  /* "uvloop/loop.pyx":1721
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ssl_handshake_timeout != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1722
 *         else:
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1722, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1722, __pyx_L1_error)

      /* "uvloop/loop.pyx":1721
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    }

    /* "uvloop/loop.pyx":1724
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1725
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1725, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1725, __pyx_L1_error)

      /* "uvloop/loop.pyx":1724
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    }
  }
  __pyx_L11:;

  /* "uvloop/loop.pyx":1728
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_host != Py_None);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_port != Py_None);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L16_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1729
 * 
 *         if host is not None or port is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1730
 *         if host is not None or port is not None:
 *             if sock is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'host/port and sock can not be specified at the same time')
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1730, __pyx_L1_error)

      /* "uvloop/loop.pyx":1729
 * 
 *         if host is not None or port is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    }

    /* "uvloop/loop.pyx":1733
 *                     'host/port and sock can not be specified at the same time')
 * 
 *             if reuse_address is None:             # <<<<<<<<<<<<<<
 *                 reuse_address = os_name == 'posix' and sys_platform != 'cygwin'
 *             reuse_port = bool(reuse_port)
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_reuse_address == Py_None);
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1734
 * 
 *             if reuse_address is None:
 *                 reuse_address = os_name == 'posix' and sys_platform != 'cygwin'             # <<<<<<<<<<<<<<
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:
 */
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_6uvloop_4loop_os_name, __pyx_n_u_posix, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1734, __pyx_L1_error)
      if (__pyx_t_2) {
      } else {
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1734, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = __pyx_t_3;
        __pyx_t_3 = 0;
        goto __pyx_L20_bool_binop_done;
      }
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_6uvloop_4loop_sys_platform, __pyx_n_u_cygwin, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1734, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1734, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __pyx_t_3;
      __pyx_t_3 = 0;
      __pyx_L20_bool_binop_done:;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_reuse_address);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_reuse_address, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":1733
 *                     'host/port and sock can not be specified at the same time')
 * 
 *             if reuse_address is None:             # <<<<<<<<<<<<<<
 *                 reuse_address = os_name == 'posix' and sys_platform != 'cygwin'
 *             reuse_port = bool(reuse_port)
 */
    }

    /* "uvloop/loop.pyx":1735
 *             if reuse_address is None:
 *                 reuse_address = os_name == 'posix' and sys_platform != 'cygwin'
 *             reuse_port = bool(reuse_port)             # <<<<<<<<<<<<<<
 *             if reuse_port and not has_SO_REUSEPORT:
 *                 raise ValueError(
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1735, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_reuse_port);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_reuse_port, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":1736
 *                 reuse_address = os_name == 'posix' and sys_platform != 'cygwin'
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'reuse_port not supported by socket module')
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1736, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L23_bool_binop_done;
    }
    __pyx_t_1 = (!(__pyx_v_6uvloop_4loop_has_SO_REUSEPORT != 0));
    __pyx_t_2 = __pyx_t_1;
    __pyx_L23_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1737
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'reuse_port not supported by socket module')
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1737, __pyx_L1_error)

      /* "uvloop/loop.pyx":1736
 *                 reuse_address = os_name == 'posix' and sys_platform != 'cygwin'
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'reuse_port not supported by socket module')
 */
    }

    /* "uvloop/loop.pyx":1740
 *                     'reuse_port not supported by socket module')
 * 
 *             if host == '':             # <<<<<<<<<<<<<<
 *                 hosts = [None]
 *             elif (isinstance(host, str) or not isinstance(host, col_Iterable)):
 */
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_cur_scope->__pyx_v_host, __pyx_kp_u__10, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1740, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1741
 * 
 *             if host == '':
 *                 hosts = [None]             # <<<<<<<<<<<<<<
 *             elif (isinstance(host, str) or not isinstance(host, col_Iterable)):
 *                 hosts = [host]
 */
      __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1741, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, Py_None)) __PYX_ERR(0, 1741, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_cur_scope->__pyx_v_hosts = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":1740
 *                     'reuse_port not supported by socket module')
 * 
 *             if host == '':             # <<<<<<<<<<<<<<
 *                 hosts = [None]
 *             elif (isinstance(host, str) or not isinstance(host, col_Iterable)):
 */
      goto __pyx_L25;
    }

    /* "uvloop/loop.pyx":1742
 *             if host == '':
 *                 hosts = [None]
 *             elif (isinstance(host, str) or not isinstance(host, col_Iterable)):             # <<<<<<<<<<<<<<
 *                 hosts = [host]
 *             else:
 */
    __pyx_t_1 = PyUnicode_Check(__pyx_cur_scope->__pyx_v_host); 
    if (!__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_6 = __pyx_v_6uvloop_4loop_col_Iterable;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_1 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_host, __pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1742, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = (!__pyx_t_1);
    __pyx_t_2 = __pyx_t_7;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1743
 *                 hosts = [None]
 *             elif (isinstance(host, str) or not isinstance(host, col_Iterable)):
 *                 hosts = [host]             # <<<<<<<<<<<<<<
 *             else:
 *                 hosts = host
 */
      __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1743, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_host);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_host);
      if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_cur_scope->__pyx_v_host)) __PYX_ERR(0, 1743, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_cur_scope->__pyx_v_hosts = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":1742
 *             if host == '':
 *                 hosts = [None]
 *             elif (isinstance(host, str) or not isinstance(host, col_Iterable)):             # <<<<<<<<<<<<<<
 *                 hosts = [host]
 *             else:
 */
      goto __pyx_L25;
    }

    /* "uvloop/loop.pyx":1745
 *                 hosts = [host]
 *             else:
 *                 hosts = host             # <<<<<<<<<<<<<<
 * 
 *             fs = [self._getaddrinfo(host, port, family,
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_host);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_host);
      __pyx_cur_scope->__pyx_v_hosts = __pyx_cur_scope->__pyx_v_host;
    }
    __pyx_L25:;

    /* "uvloop/loop.pyx":1747
 *                 hosts = host
 * 
 *             fs = [self._getaddrinfo(host, port, family,             # <<<<<<<<<<<<<<
 *                                     uv.SOCK_STREAM, 0, flags,
 *                                     0) for host in hosts]
 */
    { /* enter inner scope */
      __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1747, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "uvloop/loop.pyx":1749
 *             fs = [self._getaddrinfo(host, port, family,
 *                                     uv.SOCK_STREAM, 0, flags,
 *                                     0) for host in hosts]             # <<<<<<<<<<<<<<
 * 
 *             infos = await aio_gather(*fs)
 */
      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_hosts)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_hosts)) {
        __pyx_t_3 = __pyx_cur_scope->__pyx_v_hosts; __Pyx_INCREF(__pyx_t_3);
        __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_hosts); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1749, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1749, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1749, __pyx_L1_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1749, __pyx_L1_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1749, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1749, __pyx_L1_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1749, __pyx_L1_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1749, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_9(__pyx_t_3);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1749, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_7genexpr__pyx_v_host);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_7genexpr__pyx_v_host, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":1747
 *                 hosts = host
 * 
 *             fs = [self._getaddrinfo(host, port, family,             # <<<<<<<<<<<<<<
 *                                     uv.SOCK_STREAM, 0, flags,
 *                                     0) for host in hosts]
 */
        __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getaddrinfo(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_7genexpr__pyx_v_host, __pyx_cur_scope->__pyx_v_port, __pyx_cur_scope->__pyx_v_family, SOCK_STREAM, 0, __pyx_cur_scope->__pyx_v_flags, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1747, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1747, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":1749
 *             fs = [self._getaddrinfo(host, port, family,
 *                                     uv.SOCK_STREAM, 0, flags,
 *                                     0) for host in hosts]             # <<<<<<<<<<<<<<
 * 
 *             infos = await aio_gather(*fs)
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } /* exit inner scope */
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_cur_scope->__pyx_v_fs = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":1751
 *                                     0) for host in hosts]
 * 
 *             infos = await aio_gather(*fs)             # <<<<<<<<<<<<<<
 * 
 *             completed = False
 */
    __pyx_t_6 = PySequence_Tuple(__pyx_cur_scope->__pyx_v_fs); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_gather, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 2;
      return __pyx_r;
      __pyx_L31_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1751, __pyx_L1_error)
      __pyx_t_3 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_3);
    } else {
      __pyx_t_3 = NULL;
      if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_3) < 0) __PYX_ERR(0, 1751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_infos = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1753
 *             infos = await aio_gather(*fs)
 * 
 *             completed = False             # <<<<<<<<<<<<<<
 *             sock = None
 *             try:
 */
    __pyx_cur_scope->__pyx_v_completed = 0;

    /* "uvloop/loop.pyx":1754
 * 
 *             completed = False
 *             sock = None             # <<<<<<<<<<<<<<
 *             try:
 *                 for info in infos:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_sock);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_sock, Py_None);
    __Pyx_GIVEREF(Py_None);

    /* "uvloop/loop.pyx":1755
 *             completed = False
 *             sock = None
 *             try:             # <<<<<<<<<<<<<<
 *                 for info in infos:
 *                     addrinfo = (<AddrInfo>info).data
 */
    /*try:*/ {

      /* "uvloop/loop.pyx":1756
 *             sock = None
 *             try:
 *                 for info in infos:             # <<<<<<<<<<<<<<
 *                     addrinfo = (<AddrInfo>info).data
 *                     while addrinfo != NULL:
 */
      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_infos)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_infos)) {
        __pyx_t_3 = __pyx_cur_scope->__pyx_v_infos; __Pyx_INCREF(__pyx_t_3);
        __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_infos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1756, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1756, __pyx_L33_error)
      }
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1756, __pyx_L33_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1756, __pyx_L33_error)
            #else
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1756, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1756, __pyx_L33_error)
              #endif
              if (__pyx_t_8 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1756, __pyx_L33_error)
            #else
            __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1756, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_9(__pyx_t_3);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1756, __pyx_L33_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_info);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_info, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        __pyx_t_6 = 0;

        /* "uvloop/loop.pyx":1757
 *             try:
 *                 for info in infos:
 *                     addrinfo = (<AddrInfo>info).data             # <<<<<<<<<<<<<<
 *                     while addrinfo != NULL:
 *                         if addrinfo.ai_family == uv.AF_UNSPEC:
 */
        __pyx_t_10 = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_cur_scope->__pyx_v_info)->data;
        __pyx_cur_scope->__pyx_v_addrinfo = __pyx_t_10;

        /* "uvloop/loop.pyx":1758
 *                 for info in infos:
 *                     addrinfo = (<AddrInfo>info).data
 *                     while addrinfo != NULL:             # <<<<<<<<<<<<<<
 *                         if addrinfo.ai_family == uv.AF_UNSPEC:
 *                             raise RuntimeError('AF_UNSPEC in DNS results')
 */
        while (1) {
          __pyx_t_2 = (__pyx_cur_scope->__pyx_v_addrinfo != NULL);
          if (!__pyx_t_2) break;

          /* "uvloop/loop.pyx":1759
 *                     addrinfo = (<AddrInfo>info).data
 *                     while addrinfo != NULL:
 *                         if addrinfo.ai_family == uv.AF_UNSPEC:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError('AF_UNSPEC in DNS results')
 * 
 */
          __pyx_t_2 = (__pyx_cur_scope->__pyx_v_addrinfo->ai_family == AF_UNSPEC);
          if (unlikely(__pyx_t_2)) {

            /* "uvloop/loop.pyx":1760
 *                     while addrinfo != NULL:
 *                         if addrinfo.ai_family == uv.AF_UNSPEC:
 *                             raise RuntimeError('AF_UNSPEC in DNS results')             # <<<<<<<<<<<<<<
 * 
 *                         try:
 */
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1760, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_Raise(__pyx_t_6, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __PYX_ERR(0, 1760, __pyx_L33_error)

            /* "uvloop/loop.pyx":1759
 *                     addrinfo = (<AddrInfo>info).data
 *                     while addrinfo != NULL:
 *                         if addrinfo.ai_family == uv.AF_UNSPEC:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError('AF_UNSPEC in DNS results')
 * 
 */
          }

          /* "uvloop/loop.pyx":1762
 *                             raise RuntimeError('AF_UNSPEC in DNS results')
 * 
 *                         try:             # <<<<<<<<<<<<<<
 *                             sock = socket_socket(addrinfo.ai_family,
 *                                                  addrinfo.ai_socktype,
 */
          {
            __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            /*try:*/ {

              /* "uvloop/loop.pyx":1763
 * 
 *                         try:
 *                             sock = socket_socket(addrinfo.ai_family,             # <<<<<<<<<<<<<<
 *                                                  addrinfo.ai_socktype,
 *                                                  addrinfo.ai_protocol)
 */
              __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_addrinfo->ai_family); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1763, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_4);

              /* "uvloop/loop.pyx":1764
 *                         try:
 *                             sock = socket_socket(addrinfo.ai_family,
 *                                                  addrinfo.ai_socktype,             # <<<<<<<<<<<<<<
 *                                                  addrinfo.ai_protocol)
 *                         except socket_error:
 */
              __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_addrinfo->ai_socktype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1764, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_5);

              /* "uvloop/loop.pyx":1765
 *                             sock = socket_socket(addrinfo.ai_family,
 *                                                  addrinfo.ai_socktype,
 *                                                  addrinfo.ai_protocol)             # <<<<<<<<<<<<<<
 *                         except socket_error:
 *                             # Assume it's a bad family/type/protocol
 */
              __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_addrinfo->ai_protocol); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1765, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_socket);
              __pyx_t_15 = __pyx_v_6uvloop_4loop_socket_socket; __pyx_t_16 = NULL;
              __pyx_t_17 = 0;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_15))) {
                __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
                if (likely(__pyx_t_16)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                  __Pyx_INCREF(__pyx_t_16);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_15, function);
                  __pyx_t_17 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[4] = {__pyx_t_16, __pyx_t_4, __pyx_t_5, __pyx_t_14};
                __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_17, 3+__pyx_t_17);
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1763, __pyx_L40_error)
                __Pyx_GOTREF(__pyx_t_6);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              }
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_sock);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_sock, __pyx_t_6);
              __Pyx_GIVEREF(__pyx_t_6);
              __pyx_t_6 = 0;

              /* "uvloop/loop.pyx":1762
 *                             raise RuntimeError('AF_UNSPEC in DNS results')
 * 
 *                         try:             # <<<<<<<<<<<<<<
 *                             sock = socket_socket(addrinfo.ai_family,
 *                                                  addrinfo.ai_socktype,
 */
            }
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            goto __pyx_L47_try_end;
            __pyx_L40_error:;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "uvloop/loop.pyx":1766
 *                                                  addrinfo.ai_socktype,
 *                                                  addrinfo.ai_protocol)
 *                         except socket_error:             # <<<<<<<<<<<<<<
 *                             # Assume it's a bad family/type/protocol
 *                             # combination.
 */
            __pyx_t_18 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_socket_error);
            if (__pyx_t_18) {
              __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_15, &__pyx_t_14) < 0) __PYX_ERR(0, 1766, __pyx_L42_except_error)
              __Pyx_XGOTREF(__pyx_t_6);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_14);

              /* "uvloop/loop.pyx":1769
 *                             # Assume it's a bad family/type/protocol
 *                             # combination.
 *                             if self._debug:             # <<<<<<<<<<<<<<
 *                                 aio_logger.warning(
 *                                     'create_server() failed to create '
 */
              if (__pyx_cur_scope->__pyx_v_self->_debug) {

                /* "uvloop/loop.pyx":1770
 *                             # combination.
 *                             if self._debug:
 *                                 aio_logger.warning(             # <<<<<<<<<<<<<<
 *                                     'create_server() failed to create '
 *                                     'socket.socket(%r, %r, %r)',
 */
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1770, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_5);

                /* "uvloop/loop.pyx":1773
 *                                     'create_server() failed to create '
 *                                     'socket.socket(%r, %r, %r)',
 *                                     addrinfo.ai_family,             # <<<<<<<<<<<<<<
 *                                     addrinfo.ai_socktype,
 *                                     addrinfo.ai_protocol, exc_info=True)
 */
                __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_addrinfo->ai_family); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1773, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_4);

                /* "uvloop/loop.pyx":1774
 *                                     'socket.socket(%r, %r, %r)',
 *                                     addrinfo.ai_family,
 *                                     addrinfo.ai_socktype,             # <<<<<<<<<<<<<<
 *                                     addrinfo.ai_protocol, exc_info=True)
 *                             addrinfo = addrinfo.ai_next
 */
                __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_addrinfo->ai_socktype); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1774, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_16);

                /* "uvloop/loop.pyx":1775
 *                                     addrinfo.ai_family,
 *                                     addrinfo.ai_socktype,
 *                                     addrinfo.ai_protocol, exc_info=True)             # <<<<<<<<<<<<<<
 *                             addrinfo = addrinfo.ai_next
 *                             continue
 */
                __pyx_t_19 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_addrinfo->ai_protocol); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1775, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_19);

                /* "uvloop/loop.pyx":1770
 *                             # combination.
 *                             if self._debug:
 *                                 aio_logger.warning(             # <<<<<<<<<<<<<<
 *                                     'create_server() failed to create '
 *                                     'socket.socket(%r, %r, %r)',
 */
                __pyx_t_20 = PyTuple_New(4); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1770, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_20);
                __Pyx_INCREF(__pyx_kp_u_create_server_failed_to_create_s);
                __Pyx_GIVEREF(__pyx_kp_u_create_server_failed_to_create_s);
                if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_kp_u_create_server_failed_to_create_s)) __PYX_ERR(0, 1770, __pyx_L42_except_error);
                __Pyx_GIVEREF(__pyx_t_4);
                if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_4)) __PYX_ERR(0, 1770, __pyx_L42_except_error);
                __Pyx_GIVEREF(__pyx_t_16);
                if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 2, __pyx_t_16)) __PYX_ERR(0, 1770, __pyx_L42_except_error);
                __Pyx_GIVEREF(__pyx_t_19);
                if (__Pyx_PyTuple_SET_ITEM(__pyx_t_20, 3, __pyx_t_19)) __PYX_ERR(0, 1770, __pyx_L42_except_error);
                __pyx_t_4 = 0;
                __pyx_t_16 = 0;
                __pyx_t_19 = 0;

                /* "uvloop/loop.pyx":1775
 *                                     addrinfo.ai_family,
 *                                     addrinfo.ai_socktype,
 *                                     addrinfo.ai_protocol, exc_info=True)             # <<<<<<<<<<<<<<
 *                             addrinfo = addrinfo.ai_next
 *                             continue
 */
                __pyx_t_19 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1775, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_19);
                if (PyDict_SetItem(__pyx_t_19, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(0, 1775, __pyx_L42_except_error)

                /* "uvloop/loop.pyx":1770
 *                             # combination.
 *                             if self._debug:
 *                                 aio_logger.warning(             # <<<<<<<<<<<<<<
 *                                     'create_server() failed to create '
 *                                     'socket.socket(%r, %r, %r)',
 */
                __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_20, __pyx_t_19); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1770, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_16);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

                /* "uvloop/loop.pyx":1769
 *                             # Assume it's a bad family/type/protocol
 *                             # combination.
 *                             if self._debug:             # <<<<<<<<<<<<<<
 *                                 aio_logger.warning(
 *                                     'create_server() failed to create '
 */
              }

              /* "uvloop/loop.pyx":1776
 *                                     addrinfo.ai_socktype,
 *                                     addrinfo.ai_protocol, exc_info=True)
 *                             addrinfo = addrinfo.ai_next             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
              __pyx_t_10 = __pyx_cur_scope->__pyx_v_addrinfo->ai_next;
              __pyx_cur_scope->__pyx_v_addrinfo = __pyx_t_10;

              /* "uvloop/loop.pyx":1777
 *                                     addrinfo.ai_protocol, exc_info=True)
 *                             addrinfo = addrinfo.ai_next
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                         if reuse_address:
 */
              goto __pyx_L48_except_continue;
              __pyx_L48_except_continue:;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              goto __pyx_L46_try_continue;
            }
            goto __pyx_L42_except_error;

            /* "uvloop/loop.pyx":1762
 *                             raise RuntimeError('AF_UNSPEC in DNS results')
 * 
 *                         try:             # <<<<<<<<<<<<<<
 *                             sock = socket_socket(addrinfo.ai_family,
 *                                                  addrinfo.ai_socktype,
 */
            __pyx_L42_except_error:;
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
            goto __pyx_L33_error;
            __pyx_L46_try_continue:;
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
            goto __pyx_L37_continue;
            __pyx_L47_try_end:;
          }

          /* "uvloop/loop.pyx":1779
 *                             continue
 * 
 *                         if reuse_address:             # <<<<<<<<<<<<<<
 *                             sock.setsockopt(uv.SOL_SOCKET, uv.SO_REUSEADDR, 1)
 *                         if reuse_port:
 */
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_address); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1779, __pyx_L33_error)
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":1780
 * 
 *                         if reuse_address:
 *                             sock.setsockopt(uv.SOL_SOCKET, uv.SO_REUSEADDR, 1)             # <<<<<<<<<<<<<<
 *                         if reuse_port:
 *                             sock.setsockopt(uv.SOL_SOCKET, SO_REUSEPORT, 1)
 */
            __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setsockopt); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1780, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_6 = __Pyx_PyInt_From_int(SOL_SOCKET); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1780, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_16 = __Pyx_PyInt_From_int(SO_REUSEADDR); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1780, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_19 = NULL;
            __pyx_t_17 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_15))) {
              __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_15);
              if (likely(__pyx_t_19)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                __Pyx_INCREF(__pyx_t_19);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_15, function);
                __pyx_t_17 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[4] = {__pyx_t_19, __pyx_t_6, __pyx_t_16, __pyx_int_1};
              __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_17, 3+__pyx_t_17);
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1780, __pyx_L33_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

            /* "uvloop/loop.pyx":1779
 *                             continue
 * 
 *                         if reuse_address:             # <<<<<<<<<<<<<<
 *                             sock.setsockopt(uv.SOL_SOCKET, uv.SO_REUSEADDR, 1)
 *                         if reuse_port:
 */
          }

          /* "uvloop/loop.pyx":1781
 *                         if reuse_address:
 *                             sock.setsockopt(uv.SOL_SOCKET, uv.SO_REUSEADDR, 1)
 *                         if reuse_port:             # <<<<<<<<<<<<<<
 *                             sock.setsockopt(uv.SOL_SOCKET, SO_REUSEPORT, 1)
 *                         # Disable IPv4/IPv6 dual stack support (enabled by
 */
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1781, __pyx_L33_error)
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":1782
 *                             sock.setsockopt(uv.SOL_SOCKET, uv.SO_REUSEADDR, 1)
 *                         if reuse_port:
 *                             sock.setsockopt(uv.SOL_SOCKET, SO_REUSEPORT, 1)             # <<<<<<<<<<<<<<
 *                         # Disable IPv4/IPv6 dual stack support (enabled by
 *                         # default on Linux) which makes a single socket
 */
            __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setsockopt); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1782, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_16 = __Pyx_PyInt_From_int(SOL_SOCKET); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1782, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_SO_REUSEPORT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1782, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_19 = NULL;
            __pyx_t_17 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_15))) {
              __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_15);
              if (likely(__pyx_t_19)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                __Pyx_INCREF(__pyx_t_19);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_15, function);
                __pyx_t_17 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[4] = {__pyx_t_19, __pyx_t_16, __pyx_t_6, __pyx_int_1};
              __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_17, 3+__pyx_t_17);
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1782, __pyx_L33_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

            /* "uvloop/loop.pyx":1781
 *                         if reuse_address:
 *                             sock.setsockopt(uv.SOL_SOCKET, uv.SO_REUSEADDR, 1)
 *                         if reuse_port:             # <<<<<<<<<<<<<<
 *                             sock.setsockopt(uv.SOL_SOCKET, SO_REUSEPORT, 1)
 *                         # Disable IPv4/IPv6 dual stack support (enabled by
 */
          }

          /* "uvloop/loop.pyx":1786
 *                         # default on Linux) which makes a single socket
 *                         # listen on both address families.
 *                         if (addrinfo.ai_family == uv.AF_INET6 and             # <<<<<<<<<<<<<<
 *                                 has_IPV6_V6ONLY):
 *                             sock.setsockopt(uv.IPPROTO_IPV6, IPV6_V6ONLY, 1)
 */
          __pyx_t_7 = (__pyx_cur_scope->__pyx_v_addrinfo->ai_family == AF_INET6);
          if (__pyx_t_7) {
          } else {
            __pyx_t_2 = __pyx_t_7;
            goto __pyx_L54_bool_binop_done;
          }

          /* "uvloop/loop.pyx":1787
 *                         # listen on both address families.
 *                         if (addrinfo.ai_family == uv.AF_INET6 and
 *                                 has_IPV6_V6ONLY):             # <<<<<<<<<<<<<<
 *                             sock.setsockopt(uv.IPPROTO_IPV6, IPV6_V6ONLY, 1)
 * 
 */
          __pyx_t_7 = (__pyx_v_6uvloop_4loop_has_IPV6_V6ONLY != 0);
          __pyx_t_2 = __pyx_t_7;
          __pyx_L54_bool_binop_done:;

          /* "uvloop/loop.pyx":1786
 *                         # default on Linux) which makes a single socket
 *                         # listen on both address families.
 *                         if (addrinfo.ai_family == uv.AF_INET6 and             # <<<<<<<<<<<<<<
 *                                 has_IPV6_V6ONLY):
 *                             sock.setsockopt(uv.IPPROTO_IPV6, IPV6_V6ONLY, 1)
 */
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":1788
 *                         if (addrinfo.ai_family == uv.AF_INET6 and
 *                                 has_IPV6_V6ONLY):
 *                             sock.setsockopt(uv.IPPROTO_IPV6, IPV6_V6ONLY, 1)             # <<<<<<<<<<<<<<
 * 
 *                         pyaddr = __convert_sockaddr_to_pyaddr(addrinfo.ai_addr)
 */
            __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setsockopt); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1788, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_6 = __Pyx_PyInt_From_int(IPPROTO_IPV6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1788, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_IPV6_V6ONLY); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1788, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_19 = NULL;
            __pyx_t_17 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_15))) {
              __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_15);
              if (likely(__pyx_t_19)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                __Pyx_INCREF(__pyx_t_19);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_15, function);
                __pyx_t_17 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[4] = {__pyx_t_19, __pyx_t_6, __pyx_t_16, __pyx_int_1};
              __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_17, 3+__pyx_t_17);
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1788, __pyx_L33_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

            /* "uvloop/loop.pyx":1786
 *                         # default on Linux) which makes a single socket
 *                         # listen on both address families.
 *                         if (addrinfo.ai_family == uv.AF_INET6 and             # <<<<<<<<<<<<<<
 *                                 has_IPV6_V6ONLY):
 *                             sock.setsockopt(uv.IPPROTO_IPV6, IPV6_V6ONLY, 1)
 */
          }

          /* "uvloop/loop.pyx":1790
 *                             sock.setsockopt(uv.IPPROTO_IPV6, IPV6_V6ONLY, 1)
 * 
 *                         pyaddr = __convert_sockaddr_to_pyaddr(addrinfo.ai_addr)             # <<<<<<<<<<<<<<
 *                         try:
 *                             sock.bind(pyaddr)
 */
          __pyx_t_14 = __pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(__pyx_cur_scope->__pyx_v_addrinfo->ai_addr); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1790, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pyaddr);
          __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pyaddr, __pyx_t_14);
          __Pyx_GIVEREF(__pyx_t_14);
          __pyx_t_14 = 0;

          /* "uvloop/loop.pyx":1791
 * 
 *                         pyaddr = __convert_sockaddr_to_pyaddr(addrinfo.ai_addr)
 *                         try:             # <<<<<<<<<<<<<<
 *                             sock.bind(pyaddr)
 *                         except OSError as err:
 */
          {
            __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_11);
            /*try:*/ {

              /* "uvloop/loop.pyx":1792
 *                         pyaddr = __convert_sockaddr_to_pyaddr(addrinfo.ai_addr)
 *                         try:
 *                             sock.bind(pyaddr)             # <<<<<<<<<<<<<<
 *                         except OSError as err:
 *                             raise OSError(
 */
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_bind); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1792, __pyx_L56_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_16 = NULL;
              __pyx_t_17 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_15))) {
                __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
                if (likely(__pyx_t_16)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                  __Pyx_INCREF(__pyx_t_16);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_15, function);
                  __pyx_t_17 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_cur_scope->__pyx_v_pyaddr};
                __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1792, __pyx_L56_error)
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              }
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "uvloop/loop.pyx":1791
 * 
 *                         pyaddr = __convert_sockaddr_to_pyaddr(addrinfo.ai_addr)
 *                         try:             # <<<<<<<<<<<<<<
 *                             sock.bind(pyaddr)
 *                         except OSError as err:
 */
            }
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            goto __pyx_L63_try_end;
            __pyx_L56_error:;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "uvloop/loop.pyx":1793
 *                         try:
 *                             sock.bind(pyaddr)
 *                         except OSError as err:             # <<<<<<<<<<<<<<
 *                             raise OSError(
 *                                 err.errno, 'error while attempting '
 */
            __pyx_t_18 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
            if (__pyx_t_18) {
              __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0) __PYX_ERR(0, 1793, __pyx_L58_except_error)
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_GIVEREF(__pyx_t_15);
              __pyx_cur_scope->__pyx_v_err = __pyx_t_15;
              /*try:*/ {

                /* "uvloop/loop.pyx":1795
 *                         except OSError as err:
 *                             raise OSError(
 *                                 err.errno, 'error while attempting '             # <<<<<<<<<<<<<<
 *                                 'to bind on address %r: %s'
 *                                 % (pyaddr, err.strerror.lower())) from None
 */
                __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_err, __pyx_n_s_errno); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1795, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_19 = PyTuple_New(4); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1795, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_19);
                __pyx_t_21 = 0;
                __pyx_t_22 = 127;
                __Pyx_INCREF(__pyx_kp_u_error_while_attempting_to_bind_o);
                __pyx_t_21 += 42;
                __Pyx_GIVEREF(__pyx_kp_u_error_while_attempting_to_bind_o);
                PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_kp_u_error_while_attempting_to_bind_o);

                /* "uvloop/loop.pyx":1797
 *                                 err.errno, 'error while attempting '
 *                                 'to bind on address %r: %s'
 *                                 % (pyaddr, err.strerror.lower())) from None             # <<<<<<<<<<<<<<
 * 
 *                         tcp = TCPServer.new(self, protocol_factory, server,
 */
                __pyx_t_20 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_cur_scope->__pyx_v_pyaddr), __pyx_empty_unicode); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1797, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_20);
                __pyx_t_22 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_20) > __pyx_t_22) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_20) : __pyx_t_22;
                __pyx_t_21 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_20);
                __Pyx_GIVEREF(__pyx_t_20);
                PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_20);
                __pyx_t_20 = 0;
                __Pyx_INCREF(__pyx_kp_u__61);
                __pyx_t_21 += 2;
                __Pyx_GIVEREF(__pyx_kp_u__61);
                PyTuple_SET_ITEM(__pyx_t_19, 2, __pyx_kp_u__61);
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_err, __pyx_n_s_strerror); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1797, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_lower); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1797, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                __pyx_t_5 = NULL;
                __pyx_t_17 = 0;
                #if CYTHON_UNPACK_METHODS
                if (likely(PyMethod_Check(__pyx_t_4))) {
                  __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
                  if (likely(__pyx_t_5)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                    __Pyx_INCREF(__pyx_t_5);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_4, function);
                    __pyx_t_17 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
                  __pyx_t_20 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1797, __pyx_L69_error)
                  __Pyx_GOTREF(__pyx_t_20);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                }
                __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_20), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1797, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
                __pyx_t_22 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_22) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_22;
                __pyx_t_21 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_4);
                PyTuple_SET_ITEM(__pyx_t_19, 3, __pyx_t_4);
                __pyx_t_4 = 0;

                /* "uvloop/loop.pyx":1795
 *                         except OSError as err:
 *                             raise OSError(
 *                                 err.errno, 'error while attempting '             # <<<<<<<<<<<<<<
 *                                 'to bind on address %r: %s'
 *                                 % (pyaddr, err.strerror.lower())) from None
 */
                __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_19, 4, __pyx_t_21, __pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1795, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                /* "uvloop/loop.pyx":1794
 *                             sock.bind(pyaddr)
 *                         except OSError as err:
 *                             raise OSError(             # <<<<<<<<<<<<<<
 *                                 err.errno, 'error while attempting '
 *                                 'to bind on address %r: %s'
 */
                __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1794, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_19);
                __Pyx_GIVEREF(__pyx_t_6);
                if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_6)) __PYX_ERR(0, 1794, __pyx_L69_error);
                __Pyx_GIVEREF(__pyx_t_4);
                if (__Pyx_PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_4)) __PYX_ERR(0, 1794, __pyx_L69_error);
                __pyx_t_6 = 0;
                __pyx_t_4 = 0;
                __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_t_19, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1794, __pyx_L69_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

                /* "uvloop/loop.pyx":1797
 *                                 err.errno, 'error while attempting '
 *                                 'to bind on address %r: %s'
 *                                 % (pyaddr, err.strerror.lower())) from None             # <<<<<<<<<<<<<<
 * 
 *                         tcp = TCPServer.new(self, protocol_factory, server,
 */
                __Pyx_Raise(__pyx_t_4, 0, 0, Py_None);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                __PYX_ERR(0, 1794, __pyx_L69_error)
              }

              /* "uvloop/loop.pyx":1793
 *                         try:
 *                             sock.bind(pyaddr)
 *                         except OSError as err:             # <<<<<<<<<<<<<<
 *                             raise OSError(
 *                                 err.errno, 'error while attempting '
 */
              /*finally:*/ {
                __pyx_L69_error:;
                /*exception exit:*/{
                  __Pyx_PyThreadState_assign
                  __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0; __pyx_t_28 = 0; __pyx_t_29 = 0; __pyx_t_30 = 0;
                  __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                  __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                  if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_28, &__pyx_t_29, &__pyx_t_30);
                  if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_25, &__pyx_t_26, &__pyx_t_27) < 0)) __Pyx_ErrFetch(&__pyx_t_25, &__pyx_t_26, &__pyx_t_27);
                  __Pyx_XGOTREF(__pyx_t_25);
                  __Pyx_XGOTREF(__pyx_t_26);
                  __Pyx_XGOTREF(__pyx_t_27);
                  __Pyx_XGOTREF(__pyx_t_28);
                  __Pyx_XGOTREF(__pyx_t_29);
                  __Pyx_XGOTREF(__pyx_t_30);
                  __pyx_t_18 = __pyx_lineno; __pyx_t_23 = __pyx_clineno; __pyx_t_24 = __pyx_filename;
                  {
                    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_err);
                    __Pyx_DECREF(__pyx_cur_scope->__pyx_v_err); __pyx_cur_scope->__pyx_v_err = 0;
                  }
                  if (PY_MAJOR_VERSION >= 3) {
                    __Pyx_XGIVEREF(__pyx_t_28);
                    __Pyx_XGIVEREF(__pyx_t_29);
                    __Pyx_XGIVEREF(__pyx_t_30);
                    __Pyx_ExceptionReset(__pyx_t_28, __pyx_t_29, __pyx_t_30);
                  }
                  __Pyx_XGIVEREF(__pyx_t_25);
                  __Pyx_XGIVEREF(__pyx_t_26);
                  __Pyx_XGIVEREF(__pyx_t_27);
                  __Pyx_ErrRestore(__pyx_t_25, __pyx_t_26, __pyx_t_27);
                  __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0; __pyx_t_28 = 0; __pyx_t_29 = 0; __pyx_t_30 = 0;
                  __pyx_lineno = __pyx_t_18; __pyx_clineno = __pyx_t_23; __pyx_filename = __pyx_t_24;
                  goto __pyx_L58_except_error;
                }
              }
            }
            goto __pyx_L58_except_error;

            /* "uvloop/loop.pyx":1791
 * 
 *                         pyaddr = __convert_sockaddr_to_pyaddr(addrinfo.ai_addr)
 *                         try:             # <<<<<<<<<<<<<<
 *                             sock.bind(pyaddr)
 *                         except OSError as err:
 */
            __pyx_L58_except_error:;
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
            goto __pyx_L33_error;
            __pyx_L63_try_end:;
          }

          /* "uvloop/loop.pyx":1799
 *                                 % (pyaddr, err.strerror.lower())) from None
 * 
 *                         tcp = TCPServer.new(self, protocol_factory, server,             # <<<<<<<<<<<<<<
 *                                             uv.AF_UNSPEC, backlog,
 *                                             ssl, ssl_handshake_timeout,
 */
          __pyx_t_16 = ((PyObject *)__pyx_f_6uvloop_4loop_9TCPServer_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol_factory, __pyx_cur_scope->__pyx_v_server, AF_UNSPEC, __pyx_cur_scope->__pyx_v_backlog, __pyx_cur_scope->__pyx_v_ssl, __pyx_cur_scope->__pyx_v_ssl_handshake_timeout, __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1799, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_tcp);
          __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_tcp, ((struct __pyx_obj_6uvloop_4loop_TCPServer *)__pyx_t_16));
          __Pyx_GIVEREF(__pyx_t_16);
          __pyx_t_16 = 0;

          /* "uvloop/loop.pyx":1804
 *                                             ssl_shutdown_timeout)
 * 
 *                         try:             # <<<<<<<<<<<<<<
 *                             tcp._open(sock.fileno())
 *                         except (KeyboardInterrupt, SystemExit):
 */
          {
            __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            /*try:*/ {

              /* "uvloop/loop.pyx":1805
 * 
 *                         try:
 *                             tcp._open(sock.fileno())             # <<<<<<<<<<<<<<
 *                         except (KeyboardInterrupt, SystemExit):
 *                             raise
 */
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1805, __pyx_L75_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_14 = NULL;
              __pyx_t_17 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_15))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_15);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_15, function);
                  __pyx_t_17 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
                __pyx_t_16 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1805, __pyx_L75_error)
                __Pyx_GOTREF(__pyx_t_16);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              }
              __pyx_t_23 = __Pyx_PyInt_As_int(__pyx_t_16); if (unlikely((__pyx_t_23 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1805, __pyx_L75_error)
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_16 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_cur_scope->__pyx_v_tcp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_tcp), __pyx_t_23); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1805, __pyx_L75_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

              /* "uvloop/loop.pyx":1804
 *                                             ssl_shutdown_timeout)
 * 
 *                         try:             # <<<<<<<<<<<<<<
 *                             tcp._open(sock.fileno())
 *                         except (KeyboardInterrupt, SystemExit):
 */
            }
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            goto __pyx_L82_try_end;
            __pyx_L75_error:;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "uvloop/loop.pyx":1806
 *                         try:
 *                             tcp._open(sock.fileno())
 *                         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                             raise
 *                         except BaseException:
 */
            __pyx_t_23 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
            if (__pyx_t_23) {
              __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14) < 0) __PYX_ERR(0, 1806, __pyx_L77_except_error)
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_14);

              /* "uvloop/loop.pyx":1807
 *                             tcp._open(sock.fileno())
 *                         except (KeyboardInterrupt, SystemExit):
 *                             raise             # <<<<<<<<<<<<<<
 *                         except BaseException:
 *                             tcp._close()
 */
              __Pyx_GIVEREF(__pyx_t_16);
              __Pyx_GIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_ErrRestoreWithState(__pyx_t_16, __pyx_t_15, __pyx_t_14);
              __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; 
              __PYX_ERR(0, 1807, __pyx_L77_except_error)
            }

            /* "uvloop/loop.pyx":1808
 *                         except (KeyboardInterrupt, SystemExit):
 *                             raise
 *                         except BaseException:             # <<<<<<<<<<<<<<
 *                             tcp._close()
 *                             raise
 */
            __pyx_t_23 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
            if (__pyx_t_23) {
              __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0) __PYX_ERR(0, 1808, __pyx_L77_except_error)
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);

              /* "uvloop/loop.pyx":1809
 *                             raise
 *                         except BaseException:
 *                             tcp._close()             # <<<<<<<<<<<<<<
 *                             raise
 * 
 */
              __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_cur_scope->__pyx_v_tcp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tcp)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1809, __pyx_L77_except_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

              /* "uvloop/loop.pyx":1810
 *                         except BaseException:
 *                             tcp._close()
 *                             raise             # <<<<<<<<<<<<<<
 * 
 *                         server._add_server(tcp)
 */
              __Pyx_GIVEREF(__pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_ErrRestoreWithState(__pyx_t_14, __pyx_t_15, __pyx_t_16);
              __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; 
              __PYX_ERR(0, 1810, __pyx_L77_except_error)
            }
            goto __pyx_L77_except_error;

            /* "uvloop/loop.pyx":1804
 *                                             ssl_shutdown_timeout)
 * 
 *                         try:             # <<<<<<<<<<<<<<
 *                             tcp._open(sock.fileno())
 *                         except (KeyboardInterrupt, SystemExit):
 */
            __pyx_L77_except_error:;
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
            goto __pyx_L33_error;
            __pyx_L82_try_end:;
          }

          /* "uvloop/loop.pyx":1812
 *                             raise
 * 
 *                         server._add_server(tcp)             # <<<<<<<<<<<<<<
 *                         sock.detach()
 *                         sock = None
 */
          __pyx_t_16 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_server->__pyx_vtab)->_add_server(__pyx_cur_scope->__pyx_v_server, ((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_cur_scope->__pyx_v_tcp)); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1812, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "uvloop/loop.pyx":1813
 * 
 *                         server._add_server(tcp)
 *                         sock.detach()             # <<<<<<<<<<<<<<
 *                         sock = None
 * 
 */
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1813, __pyx_L33_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_14 = NULL;
          __pyx_t_17 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_15))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_15);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_15, function);
              __pyx_t_17 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
            __pyx_t_16 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1813, __pyx_L33_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "uvloop/loop.pyx":1814
 *                         server._add_server(tcp)
 *                         sock.detach()
 *                         sock = None             # <<<<<<<<<<<<<<
 * 
 *                         addrinfo = addrinfo.ai_next
 */
          __Pyx_INCREF(Py_None);
          __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_sock);
          __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_sock, Py_None);
          __Pyx_GIVEREF(Py_None);

          /* "uvloop/loop.pyx":1816
 *                         sock = None
 * 
 *                         addrinfo = addrinfo.ai_next             # <<<<<<<<<<<<<<
 * 
 *                 completed = True
 */
          __pyx_t_10 = __pyx_cur_scope->__pyx_v_addrinfo->ai_next;
          __pyx_cur_scope->__pyx_v_addrinfo = __pyx_t_10;
          __pyx_L37_continue:;
        }

        /* "uvloop/loop.pyx":1756
 *             sock = None
 *             try:
 *                 for info in infos:             # <<<<<<<<<<<<<<
 *                     addrinfo = (<AddrInfo>info).data
 *                     while addrinfo != NULL:
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":1818
 *                         addrinfo = addrinfo.ai_next
 * 
 *                 completed = True             # <<<<<<<<<<<<<<
 *             finally:
 *                 if not completed:
 */
      __pyx_cur_scope->__pyx_v_completed = 1;
    }

    /* "uvloop/loop.pyx":1820
 *                 completed = True
 *             finally:
 *                 if not completed:             # <<<<<<<<<<<<<<
 *                     if sock is not None:
 *                         sock.close()
 */
    /*finally:*/ {
      /*normal exit:*/{
        __pyx_t_2 = (!__pyx_cur_scope->__pyx_v_completed);
        if (__pyx_t_2) {

          /* "uvloop/loop.pyx":1821
 *             finally:
 *                 if not completed:
 *                     if sock is not None:             # <<<<<<<<<<<<<<
 *                         sock.close()
 *                     server.close()
 */
          __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":1822
 *                 if not completed:
 *                     if sock is not None:
 *                         sock.close()             # <<<<<<<<<<<<<<
 *                     server.close()
 *         else:
 */
            __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_close); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1822, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_15 = NULL;
            __pyx_t_17 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_16))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
              if (likely(__pyx_t_15)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
                __Pyx_INCREF(__pyx_t_15);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_16, function);
                __pyx_t_17 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
              __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1822, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "uvloop/loop.pyx":1821
 *             finally:
 *                 if not completed:
 *                     if sock is not None:             # <<<<<<<<<<<<<<
 *                         sock.close()
 *                     server.close()
 */
          }

          /* "uvloop/loop.pyx":1823
 *                     if sock is not None:
 *                         sock.close()
 *                     server.close()             # <<<<<<<<<<<<<<
 *         else:
 *             if sock is None:
 */
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_server), __pyx_n_s_close); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1823, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_15 = NULL;
          __pyx_t_17 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
              __pyx_t_17 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1823, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "uvloop/loop.pyx":1820
 *                 completed = True
 *             finally:
 *                 if not completed:             # <<<<<<<<<<<<<<
 *                     if sock is not None:
 *                         sock.close()
 */
        }
        goto __pyx_L34;
      }
      __pyx_L33_error:;
      /*exception exit:*/{
        __Pyx_PyThreadState_assign
        __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_30 = 0; __pyx_t_29 = 0; __pyx_t_28 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_30, &__pyx_t_29, &__pyx_t_28);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_30);
        __Pyx_XGOTREF(__pyx_t_29);
        __Pyx_XGOTREF(__pyx_t_28);
        __pyx_t_23 = __pyx_lineno; __pyx_t_18 = __pyx_clineno; __pyx_t_31 = __pyx_filename;
        {
          __pyx_t_2 = (!__pyx_cur_scope->__pyx_v_completed);
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":1821
 *             finally:
 *                 if not completed:
 *                     if sock is not None:             # <<<<<<<<<<<<<<
 *                         sock.close()
 *                     server.close()
 */
            __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
            if (__pyx_t_2) {

              /* "uvloop/loop.pyx":1822
 *                 if not completed:
 *                     if sock is not None:
 *                         sock.close()             # <<<<<<<<<<<<<<
 *                     server.close()
 *         else:
 */
              __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_close); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1822, __pyx_L91_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_15 = NULL;
              __pyx_t_17 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_16))) {
                __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
                if (likely(__pyx_t_15)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
                  __Pyx_INCREF(__pyx_t_15);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_16, function);
                  __pyx_t_17 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
                __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
                __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1822, __pyx_L91_error)
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

              /* "uvloop/loop.pyx":1821
 *             finally:
 *                 if not completed:
 *                     if sock is not None:             # <<<<<<<<<<<<<<
 *                         sock.close()
 *                     server.close()
 */
            }

            /* "uvloop/loop.pyx":1823
 *                     if sock is not None:
 *                         sock.close()
 *                     server.close()             # <<<<<<<<<<<<<<
 *         else:
 *             if sock is None:
 */
            __pyx_t_16 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_server), __pyx_n_s_close); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1823, __pyx_L91_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_15 = NULL;
            __pyx_t_17 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_16))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
              if (likely(__pyx_t_15)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
                __Pyx_INCREF(__pyx_t_15);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_16, function);
                __pyx_t_17 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
              __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1823, __pyx_L91_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "uvloop/loop.pyx":1820
 *                 completed = True
 *             finally:
 *                 if not completed:             # <<<<<<<<<<<<<<
 *                     if sock is not None:
 *                         sock.close()
 */
          }
        }
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_30);
          __Pyx_XGIVEREF(__pyx_t_29);
          __Pyx_XGIVEREF(__pyx_t_28);
          __Pyx_ExceptionReset(__pyx_t_30, __pyx_t_29, __pyx_t_28);
        }
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ErrRestore(__pyx_t_13, __pyx_t_12, __pyx_t_11);
        __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_30 = 0; __pyx_t_29 = 0; __pyx_t_28 = 0;
        __pyx_lineno = __pyx_t_23; __pyx_clineno = __pyx_t_18; __pyx_filename = __pyx_t_31;
        goto __pyx_L1_error;
        __pyx_L91_error:;
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_30);
          __Pyx_XGIVEREF(__pyx_t_29);
          __Pyx_XGIVEREF(__pyx_t_28);
          __Pyx_ExceptionReset(__pyx_t_30, __pyx_t_29, __pyx_t_28);
        }
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_30 = 0; __pyx_t_29 = 0; __pyx_t_28 = 0;
        goto __pyx_L1_error;
      }
      __pyx_L34:;
    }

    /* "uvloop/loop.pyx":1728
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
    goto __pyx_L15;
  }

  /* "uvloop/loop.pyx":1825
 *                     server.close()
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('Neither host/port nor sock were specified')
 *             if not _is_sock_stream(sock.type):
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock == Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1826
 *         else:
 *             if sock is None:
 *                 raise ValueError('Neither host/port nor sock were specified')             # <<<<<<<<<<<<<<
 *             if not _is_sock_stream(sock.type):
 *                 raise ValueError(
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1826, __pyx_L1_error)

      /* "uvloop/loop.pyx":1825
 *                     server.close()
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('Neither host/port nor sock were specified')
 *             if not _is_sock_stream(sock.type):
 */
    }

    /* "uvloop/loop.pyx":1827
 *             if sock is None:
 *                 raise ValueError('Neither host/port nor sock were specified')
 *             if not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A Stream Socket was expected, got {!r}'.format(sock))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_16 = __pyx_f_6uvloop_4loop__is_sock_stream(__pyx_t_3); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_16); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_7 = (!__pyx_t_2);
    if (unlikely(__pyx_t_7)) {

      /* "uvloop/loop.pyx":1829
 *             if not _is_sock_stream(sock.type):
 *                 raise ValueError(
 *                     'A Stream Socket was expected, got {!r}'.format(sock))             # <<<<<<<<<<<<<<
 * 
 *             # libuv will set the socket to non-blocking mode, but
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_Stream_Socket_was_expected_got, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1829, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_15 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_cur_scope->__pyx_v_sock};
        __pyx_t_16 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1829, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/loop.pyx":1828
 *                 raise ValueError('Neither host/port nor sock were specified')
 *             if not _is_sock_stream(sock.type):
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'A Stream Socket was expected, got {!r}'.format(sock))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1828, __pyx_L1_error)

      /* "uvloop/loop.pyx":1827
 *             if sock is None:
 *                 raise ValueError('Neither host/port nor sock were specified')
 *             if not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A Stream Socket was expected, got {!r}'.format(sock))
 */
    }

    /* "uvloop/loop.pyx":1833
 *             # libuv will set the socket to non-blocking mode, but
 *             # we want Python socket object to notice that.
 *             sock.setblocking(False)             # <<<<<<<<<<<<<<
 * 
 *             tcp = TCPServer.new(self, protocol_factory, server,
 */
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_15 = NULL;
    __pyx_t_17 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_16))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_16, function);
        __pyx_t_17 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_15, Py_False};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1835
 *             sock.setblocking(False)
 * 
 *             tcp = TCPServer.new(self, protocol_factory, server,             # <<<<<<<<<<<<<<
 *                                 uv.AF_UNSPEC, backlog,
 *                                 ssl, ssl_handshake_timeout,
 */
    __pyx_t_3 = ((PyObject *)__pyx_f_6uvloop_4loop_9TCPServer_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol_factory, __pyx_cur_scope->__pyx_v_server, AF_UNSPEC, __pyx_cur_scope->__pyx_v_backlog, __pyx_cur_scope->__pyx_v_ssl, __pyx_cur_scope->__pyx_v_ssl_handshake_timeout, __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1835, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_tcp = ((struct __pyx_obj_6uvloop_4loop_TCPServer *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1840
 *                                 ssl_shutdown_timeout)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 tcp._open(sock.fileno())
 *             except (KeyboardInterrupt, SystemExit):
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_28, &__pyx_t_29, &__pyx_t_30);
      __Pyx_XGOTREF(__pyx_t_28);
      __Pyx_XGOTREF(__pyx_t_29);
      __Pyx_XGOTREF(__pyx_t_30);
      /*try:*/ {

        /* "uvloop/loop.pyx":1841
 * 
 *             try:
 *                 tcp._open(sock.fileno())             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1841, __pyx_L96_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_15 = NULL;
        __pyx_t_17 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_16))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_16);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_16, function);
            __pyx_t_17 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_17, 0+__pyx_t_17);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1841, __pyx_L96_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __pyx_t_18 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_18 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1841, __pyx_L96_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_cur_scope->__pyx_v_tcp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_tcp), __pyx_t_18); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1841, __pyx_L96_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "uvloop/loop.pyx":1840
 *                                 ssl_shutdown_timeout)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 tcp._open(sock.fileno())
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
      __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
      __Pyx_XDECREF(__pyx_t_30); __pyx_t_30 = 0;
      goto __pyx_L101_try_end;
      __pyx_L96_error:;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":1842
 *             try:
 *                 tcp._open(sock.fileno())
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_18 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_18) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_16, &__pyx_t_15) < 0) __PYX_ERR(0, 1842, __pyx_L98_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_15);

        /* "uvloop/loop.pyx":1843
 *                 tcp._open(sock.fileno())
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 tcp._close()
 */
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_16, __pyx_t_15);
        __pyx_t_3 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; 
        __PYX_ERR(0, 1843, __pyx_L98_except_error)
      }

      /* "uvloop/loop.pyx":1844
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 tcp._close()
 *                 raise
 */
      __pyx_t_18 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_18) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_3) < 0) __PYX_ERR(0, 1844, __pyx_L98_except_error)
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "uvloop/loop.pyx":1845
 *                 raise
 *             except BaseException:
 *                 tcp._close()             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
        __pyx_t_14 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_cur_scope->__pyx_v_tcp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tcp)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1845, __pyx_L98_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

        /* "uvloop/loop.pyx":1846
 *             except BaseException:
 *                 tcp._close()
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *             tcp._attach_fileobj(sock)
 */
        __Pyx_GIVEREF(__pyx_t_15);
        __Pyx_GIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ErrRestoreWithState(__pyx_t_15, __pyx_t_16, __pyx_t_3);
        __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_3 = 0; 
        __PYX_ERR(0, 1846, __pyx_L98_except_error)
      }
      goto __pyx_L98_except_error;

      /* "uvloop/loop.pyx":1840
 *                                 ssl_shutdown_timeout)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 tcp._open(sock.fileno())
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L98_except_error:;
      __Pyx_XGIVEREF(__pyx_t_28);
      __Pyx_XGIVEREF(__pyx_t_29);
      __Pyx_XGIVEREF(__pyx_t_30);
      __Pyx_ExceptionReset(__pyx_t_28, __pyx_t_29, __pyx_t_30);
      goto __pyx_L1_error;
      __pyx_L101_try_end:;
    }

    /* "uvloop/loop.pyx":1848
 *                 raise
 * 
 *             tcp._attach_fileobj(sock)             # <<<<<<<<<<<<<<
 *             server._add_server(tcp)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_cur_scope->__pyx_v_tcp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_tcp), __pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1848, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1849
 * 
 *             tcp._attach_fileobj(sock)
 *             server._add_server(tcp)             # <<<<<<<<<<<<<<
 * 
 *         if start_serving:
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_server->__pyx_vtab)->_add_server(__pyx_cur_scope->__pyx_v_server, ((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_cur_scope->__pyx_v_tcp)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L15:;

  /* "uvloop/loop.pyx":1851
 *             server._add_server(tcp)
 * 
 *         if start_serving:             # <<<<<<<<<<<<<<
 *             server._start_serving()
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_start_serving); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1851, __pyx_L1_error)
  if (__pyx_t_7) {

    /* "uvloop/loop.pyx":1852
 * 
 *         if start_serving:
 *             server._start_serving()             # <<<<<<<<<<<<<<
 * 
 *         server._ref()
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_server->__pyx_vtab)->_start_serving(__pyx_cur_scope->__pyx_v_server); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":1851
 *             server._add_server(tcp)
 * 
 *         if start_serving:             # <<<<<<<<<<<<<<
 *             server._start_serving()
 * 
 */
  }

  /* "uvloop/loop.pyx":1854
 *             server._start_serving()
 * 
 *         server._ref()             # <<<<<<<<<<<<<<
 *         return server
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_server->__pyx_vtab)->_ref(__pyx_cur_scope->__pyx_v_server); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1855
 * 
 *         server._ref()
 *         return server             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(((PyObject *)__pyx_cur_scope->__pyx_v_server));
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":1643
 *         return app_transport
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("create_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_60generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_59create_connection(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_58create_connection, "Loop.create_connection(self, protocol_factory, host=None, port=None, *, ssl=None, family=0, proto=0, flags=0, sock=None, local_addr=None, server_hostname=None, ssl_handshake_timeout=None, ssl_shutdown_timeout=None)\nConnect to a TCP server.\n\n        Create a streaming transport connection to a given Internet host and\n        port: socket family AF_INET or socket.AF_INET6 depending on host (or\n        family if specified), socket type SOCK_STREAM. protocol_factory must be\n        a callable returning a protocol instance.\n\n        This method is a coroutine which will try to establish the connection\n        in the background.  When successful, the coroutine returns a\n        (transport, protocol) pair.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_59create_connection = {"create_connection", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_59create_connection, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_58create_connection};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_59create_connection(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_host = 0;
  PyObject *__pyx_v_port = 0;
  PyObject *__pyx_v_ssl = 0;
  PyObject *__pyx_v_family = 0;
  PyObject *__pyx_v_proto = 0;
  PyObject *__pyx_v_flags = 0;
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_local_addr = 0;
  PyObject *__pyx_v_server_hostname = 0;
  PyObject *__pyx_v_ssl_handshake_timeout = 0;
  PyObject *__pyx_v_ssl_shutdown_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_connection (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_host,&__pyx_n_s_port,&__pyx_n_s_ssl,&__pyx_n_s_family,&__pyx_n_s_proto,&__pyx_n_s_flags,&__pyx_n_s_sock,&__pyx_n_s_local_addr,&__pyx_n_s_server_hostname,&__pyx_n_s_ssl_handshake_timeout,&__pyx_n_s_ssl_shutdown_timeout,0};

    /* "uvloop/loop.pyx":1858
 * 
 *     @cython.iterable_coroutine
 *     async def create_connection(self, protocol_factory, host=None, port=None,             # <<<<<<<<<<<<<<
 *                                 *,
 *                                 ssl=None,
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1860
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 *                                 ssl=None,             # <<<<<<<<<<<<<<
 *                                 family=0, proto=0, flags=0, sock=None,
 *                                 local_addr=None, server_hostname=None,
 */
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));

    /* "uvloop/loop.pyx":1861
 *                                 *,
 *                                 ssl=None,
 *                                 family=0, proto=0, flags=0, sock=None,             # <<<<<<<<<<<<<<
 *                                 local_addr=None, server_hostname=None,
 *                                 ssl_handshake_timeout=None,
 */
    values[7] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1862
 *                                 ssl=None,
 *                                 family=0, proto=0, flags=0, sock=None,
 *                                 local_addr=None, server_hostname=None,             # <<<<<<<<<<<<<<
 *                                 ssl_handshake_timeout=None,
 *                                 ssl_shutdown_timeout=None):
 */
    values[8] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[9] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1863
 *                                 family=0, proto=0, flags=0, sock=None,
 *                                 local_addr=None, server_hostname=None,
 *                                 ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                 ssl_shutdown_timeout=None):
 *         """Connect to a TCP server.
 */
    values[10] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":1864
 *                                 local_addr=None, server_hostname=None,
 *                                 ssl_handshake_timeout=None,
 *                                 ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 *         """Connect to a TCP server.
 * 
 */
    values[11] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1857, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_host);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1857, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_port);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1857, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 9)) {
        Py_ssize_t index;
        for (index = 3; index < 12 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1857, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_connection") < 0)) __PYX_ERR(0, 1857, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_host = values[1];
    __pyx_v_port = values[2];
    __pyx_v_ssl = values[3];
    __pyx_v_family = values[4];
    __pyx_v_proto = values[5];
    __pyx_v_flags = values[6];
    __pyx_v_sock = values[7];
    __pyx_v_local_addr = values[8];
    __pyx_v_server_hostname = values[9];
    __pyx_v_ssl_handshake_timeout = values[10];
    __pyx_v_ssl_shutdown_timeout = values[11];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_connection", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 1857, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_58create_connection(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_host, __pyx_v_port, __pyx_v_ssl, __pyx_v_family, __pyx_v_proto, __pyx_v_flags, __pyx_v_sock, __pyx_v_local_addr, __pyx_v_server_hostname, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout);

  /* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_17create_connection_2generator30(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2042
 *                 # If they all have the same str(), raise one.
 *                 model = str(exceptions[0])
 *                 if all(str(exc) == model for exc in exceptions):             # <<<<<<<<<<<<<<
 *                     raise exceptions[0]
 *                 # Raise a combined exception so the user can see all
 */

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_17create_connection_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_9_genexpr(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2042, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_17create_connection_2generator30, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_create_connection_locals_genexpr, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2042, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_connection.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_17create_connection_2generator30(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2042, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 2042, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2042, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2042, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_exc);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_exc, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_cur_scope->__pyx_v_exc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_model)) { __Pyx_RaiseClosureNameError("model"); __PYX_ERR(0, 2042, __pyx_L1_error) }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_model, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 2042, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = (!__pyx_t_4);
    if (__pyx_t_5) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_17create_connection_5generator31(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2047
 *                 # the various error messages.
 *                 raise OSError('Multiple exceptions: {}'.format(
 *                     ', '.join(str(exc) for exc in exceptions)))             # <<<<<<<<<<<<<<
 *         else:
 *             if sock is None:
 */

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_17create_connection_3genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_10_genexpr(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2047, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_17create_connection_5generator31, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_create_connection_locals_genexpr, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2047, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_connection.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_17create_connection_5generator31(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2047, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 2047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 2047, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2047, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2047, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_exc);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_exc, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_cur_scope->__pyx_v_exc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 2047, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_58create_connection(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_host, PyObject *__pyx_v_port, PyObject *__pyx_v_ssl, PyObject *__pyx_v_family, PyObject *__pyx_v_proto, PyObject *__pyx_v_flags, PyObject *__pyx_v_sock, PyObject *__pyx_v_local_addr, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *__pyx_cur_scope;
  PyObject *__pyx_gb_6uvloop_4loop_4Loop_17create_connection_2generator30 = 0;
  PyObject *__pyx_gb_6uvloop_4loop_4Loop_17create_connection_5generator31 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_connection", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_8_create_connection(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1857, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_host = __pyx_v_host;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_host);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_host);
  __pyx_cur_scope->__pyx_v_port = __pyx_v_port;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_port);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_port);
  __pyx_cur_scope->__pyx_v_ssl = __pyx_v_ssl;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl);
  __pyx_cur_scope->__pyx_v_family = __pyx_v_family;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_family);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_family);
  __pyx_cur_scope->__pyx_v_proto = __pyx_v_proto;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_proto);
  __pyx_cur_scope->__pyx_v_flags = __pyx_v_flags;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_flags);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_flags);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_local_addr = __pyx_v_local_addr;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_local_addr);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_local_addr);
  __pyx_cur_scope->__pyx_v_server_hostname = __pyx_v_server_hostname;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_hostname);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_hostname);
  __pyx_cur_scope->__pyx_v_ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_60generator4, __pyx_codeobj__63, (PyObject *) __pyx_cur_scope, __pyx_n_s_create_connection, __pyx_n_s_Loop_create_connection, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 1857, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XDECREF(__pyx_gb_6uvloop_4loop_4Loop_17create_connection_2generator30);
  __Pyx_XDECREF(__pyx_gb_6uvloop_4loop_4Loop_17create_connection_5generator31);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_60generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  struct addrinfo *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  char const *__pyx_t_28;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_connection", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L10_resume_from_await;
    case 2: goto __pyx_L30_resume_from_await;
    case 3: goto __pyx_L69_resume_from_await;
    case 4: goto __pyx_L96_resume_from_await;
    case 5: goto __pyx_L108_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1857, __pyx_L1_error)

  /* "uvloop/loop.pyx":1877
 *         """
 *         cdef:
 *             AddrInfo ai_local = None             # <<<<<<<<<<<<<<
 *             AddrInfo ai_remote
 *             TCPTransport tr
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_ai_local = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)Py_None);

  /* "uvloop/loop.pyx":1881
 *             TCPTransport tr
 * 
 *             system.addrinfo *rai = NULL             # <<<<<<<<<<<<<<
 *             system.addrinfo *lai = NULL
 * 
 */
  __pyx_cur_scope->__pyx_v_rai = NULL;

  /* "uvloop/loop.pyx":1882
 * 
 *             system.addrinfo *rai = NULL
 *             system.addrinfo *lai = NULL             # <<<<<<<<<<<<<<
 * 
 *             system.addrinfo *rai_iter = NULL
 */
  __pyx_cur_scope->__pyx_v_lai = NULL;

  /* "uvloop/loop.pyx":1884
 *             system.addrinfo *lai = NULL
 * 
 *             system.addrinfo *rai_iter = NULL             # <<<<<<<<<<<<<<
 *             system.addrinfo *lai_iter = NULL
 * 
 */
  __pyx_cur_scope->__pyx_v_rai_iter = NULL;

  /* "uvloop/loop.pyx":1885
 * 
 *             system.addrinfo *rai_iter = NULL
 *             system.addrinfo *lai_iter = NULL             # <<<<<<<<<<<<<<
 * 
 *             system.addrinfo rai_static
 */
  __pyx_cur_scope->__pyx_v_lai_iter = NULL;

  /* "uvloop/loop.pyx":1897
 *             object ssl_waiter
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *             if host is not None or port is not None:
 *                 raise ValueError(
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1897, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1898
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:
 *             if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_host != Py_None);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_port != Py_None);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/loop.pyx":1899
 *         if sock is not None and sock.family == uv.AF_UNIX:
 *             if host is not None or port is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_connection(
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1899, __pyx_L1_error)

      /* "uvloop/loop.pyx":1898
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:
 *             if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    }

    /* "uvloop/loop.pyx":1901
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_connection(             # <<<<<<<<<<<<<<
 *                 protocol_factory, None,
 *                 sock=sock, ssl=ssl, server_hostname=server_hostname)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_create_unix_connection); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "uvloop/loop.pyx":1902
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_connection(
 *                 protocol_factory, None,             # <<<<<<<<<<<<<<
 *                 sock=sock, ssl=ssl, server_hostname=server_hostname)
 * 
 */
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_cur_scope->__pyx_v_protocol_factory)) __PYX_ERR(0, 1901, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, Py_None)) __PYX_ERR(0, 1901, __pyx_L1_error);

    /* "uvloop/loop.pyx":1903
 *             return await self.create_unix_connection(
 *                 protocol_factory, None,
 *                 sock=sock, ssl=ssl, server_hostname=server_hostname)             # <<<<<<<<<<<<<<
 * 
 *         app_protocol = protocol = protocol_factory()
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sock, __pyx_cur_scope->__pyx_v_sock) < 0) __PYX_ERR(0, 1903, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl, __pyx_cur_scope->__pyx_v_ssl) < 0) __PYX_ERR(0, 1903, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_server_hostname, __pyx_cur_scope->__pyx_v_server_hostname) < 0) __PYX_ERR(0, 1903, __pyx_L1_error)

    /* "uvloop/loop.pyx":1901
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_connection(             # <<<<<<<<<<<<<<
 *                 protocol_factory, None,
 *                 sock=sock, ssl=ssl, server_hostname=server_hostname)
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L10_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1901, __pyx_L1_error)
      __pyx_t_6 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_6);
    } else {
      __pyx_t_6 = NULL;
      if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_6) < 0) __PYX_ERR(0, 1901, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":1897
 *             object ssl_waiter
 * 
 *         if sock is not None and sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *             if host is not None or port is not None:
 *                 raise ValueError(
 */
  }

  /* "uvloop/loop.pyx":1905
 *                 sock=sock, ssl=ssl, server_hostname=server_hostname)
 * 
 *         app_protocol = protocol = protocol_factory()             # <<<<<<<<<<<<<<
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_protocol_factory; __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1905, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_INCREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_cur_scope->__pyx_v_app_protocol = __pyx_t_6;
  __Pyx_INCREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_cur_scope->__pyx_v_protocol = __pyx_t_6;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1906
 * 
 *         app_protocol = protocol = protocol_factory()
 *         ssl_waiter = None             # <<<<<<<<<<<<<<
 *         context = Context_CopyCurrent()
 *         if ssl:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_ssl_waiter = Py_None;

  /* "uvloop/loop.pyx":1907
 *         app_protocol = protocol = protocol_factory()
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 *         if ssl:
 *             if server_hostname is None:
 */
  __pyx_t_6 = Context_CopyCurrent(); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_cur_scope->__pyx_v_context = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1908
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()
 *         if ssl:             # <<<<<<<<<<<<<<
 *             if server_hostname is None:
 *                 if not host:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_ssl); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1908, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":1909
 *         context = Context_CopyCurrent()
 *         if ssl:
 *             if server_hostname is None:             # <<<<<<<<<<<<<<
 *                 if not host:
 *                     raise ValueError('You must set server_hostname '
 */
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_server_hostname == Py_None);
    if (__pyx_t_1) {

      /* "uvloop/loop.pyx":1910
 *         if ssl:
 *             if server_hostname is None:
 *                 if not host:             # <<<<<<<<<<<<<<
 *                     raise ValueError('You must set server_hostname '
 *                                      'when using ssl without a host')
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_host); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1910, __pyx_L1_error)
      __pyx_t_2 = (!__pyx_t_1);
      if (unlikely(__pyx_t_2)) {

        /* "uvloop/loop.pyx":1911
 *             if server_hostname is None:
 *                 if not host:
 *                     raise ValueError('You must set server_hostname '             # <<<<<<<<<<<<<<
 *                                      'when using ssl without a host')
 *                 server_hostname = host
 */
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(0, 1911, __pyx_L1_error)

        /* "uvloop/loop.pyx":1910
 *         if ssl:
 *             if server_hostname is None:
 *                 if not host:             # <<<<<<<<<<<<<<
 *                     raise ValueError('You must set server_hostname '
 *                                      'when using ssl without a host')
 */
      }

      /* "uvloop/loop.pyx":1913
 *                     raise ValueError('You must set server_hostname '
 *                                      'when using ssl without a host')
 *                 server_hostname = host             # <<<<<<<<<<<<<<
 * 
 *             ssl_waiter = self._new_future()
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_host);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_server_hostname);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_server_hostname, __pyx_cur_scope->__pyx_v_host);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_host);

      /* "uvloop/loop.pyx":1909
 *         context = Context_CopyCurrent()
 *         if ssl:
 *             if server_hostname is None:             # <<<<<<<<<<<<<<
 *                 if not host:
 *                     raise ValueError('You must set server_hostname '
 */
    }

    /* "uvloop/loop.pyx":1915
 *                 server_hostname = host
 * 
 *             ssl_waiter = self._new_future()             # <<<<<<<<<<<<<<
 *             sslcontext = None if isinstance(ssl, bool) else ssl
 *             protocol = SSLProtocol(
 */
    __pyx_t_6 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_ssl_waiter);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_ssl_waiter, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":1916
 * 
 *             ssl_waiter = self._new_future()
 *             sslcontext = None if isinstance(ssl, bool) else ssl             # <<<<<<<<<<<<<<
 *             protocol = SSLProtocol(
 *                 self, app_protocol, sslcontext, ssl_waiter,
 */
    __pyx_t_3 = ((PyObject*)&PyBool_Type);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_ssl, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1916, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {
      __Pyx_INCREF(Py_None);
      __pyx_t_6 = Py_None;
    } else {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
      __pyx_t_6 = __pyx_cur_scope->__pyx_v_ssl;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_cur_scope->__pyx_v_sslcontext = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "uvloop/loop.pyx":1917
 *             ssl_waiter = self._new_future()
 *             sslcontext = None if isinstance(ssl, bool) else ssl
 *             protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self, app_protocol, sslcontext, ssl_waiter,
 *                 False, server_hostname,
 */
    __pyx_t_6 = PyTuple_New(6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(0, 1917, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_cur_scope->__pyx_v_app_protocol)) __PYX_ERR(0, 1917, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sslcontext);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sslcontext);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_cur_scope->__pyx_v_sslcontext)) __PYX_ERR(0, 1917, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_waiter);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_waiter);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_cur_scope->__pyx_v_ssl_waiter)) __PYX_ERR(0, 1917, __pyx_L1_error);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 4, Py_False)) __PYX_ERR(0, 1917, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_hostname);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_hostname);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_cur_scope->__pyx_v_server_hostname)) __PYX_ERR(0, 1917, __pyx_L1_error);

    /* "uvloop/loop.pyx":1920
 *                 self, app_protocol, sslcontext, ssl_waiter,
 *                 False, server_hostname,
 *                 ssl_handshake_timeout=ssl_handshake_timeout,             # <<<<<<<<<<<<<<
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *         else:
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_handshake_timeout, __pyx_cur_scope->__pyx_v_ssl_handshake_timeout) < 0) __PYX_ERR(0, 1920, __pyx_L1_error)

    /* "uvloop/loop.pyx":1921
 *                 False, server_hostname,
 *                 ssl_handshake_timeout=ssl_handshake_timeout,
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)             # <<<<<<<<<<<<<<
 *         else:
 *             if server_hostname is not None:
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_shutdown_timeout, __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout) < 0) __PYX_ERR(0, 1920, __pyx_L1_error)

    /* "uvloop/loop.pyx":1917
 *             ssl_waiter = self._new_future()
 *             sslcontext = None if isinstance(ssl, bool) else ssl
 *             protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self, app_protocol, sslcontext, ssl_waiter,
 *                 False, server_hostname,
 */
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol), __pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_protocol);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_protocol, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1908
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()
 *         if ssl:             # <<<<<<<<<<<<<<
 *             if server_hostname is None:
 *                 if not host:
 */
    goto __pyx_L11;
  }

  /* "uvloop/loop.pyx":1923
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *         else:
 *             if server_hostname is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_server_hostname != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1924
 *         else:
 *             if server_hostname is not None:
 *                 raise ValueError('server_hostname is only meaningful with ssl')             # <<<<<<<<<<<<<<
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1924, __pyx_L1_error)

      /* "uvloop/loop.pyx":1923
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *         else:
 *             if server_hostname is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:
 */
    }

    /* "uvloop/loop.pyx":1925
 *             if server_hostname is not None:
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ssl_handshake_timeout != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1926
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1926, __pyx_L1_error)

      /* "uvloop/loop.pyx":1925
 *             if server_hostname is not None:
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    }

    /* "uvloop/loop.pyx":1928
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1929
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1929, __pyx_L1_error)

      /* "uvloop/loop.pyx":1928
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    }
  }
  __pyx_L11:;

  /* "uvloop/loop.pyx":1932
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_host != Py_None);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L18_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_port != Py_None);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L18_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":1933
 * 
 *         if host is not None or port is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":1934
 *         if host is not None or port is not None:
 *             if sock is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'host/port and sock can not be specified at the same time')
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1934, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1934, __pyx_L1_error)

      /* "uvloop/loop.pyx":1933
 * 
 *         if host is not None or port is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host/port and sock can not be specified at the same time')
 */
    }

    /* "uvloop/loop.pyx":1937
 *                     'host/port and sock can not be specified at the same time')
 * 
 *             fs = []             # <<<<<<<<<<<<<<
 *             f1 = f2 = None
 * 
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_fs = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1938
 * 
 *             fs = []
 *             f1 = f2 = None             # <<<<<<<<<<<<<<
 * 
 *             addr = __static_getaddrinfo(
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_f1 = Py_None;
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_f2 = Py_None;

    /* "uvloop/loop.pyx":1941
 * 
 *             addr = __static_getaddrinfo(
 *                 host, port, family, uv.SOCK_STREAM,             # <<<<<<<<<<<<<<
 *                 proto, <system.sockaddr*>&rai_addr_static)
 * 
 */
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1941, __pyx_L1_error)

    /* "uvloop/loop.pyx":1942
 *             addr = __static_getaddrinfo(
 *                 host, port, family, uv.SOCK_STREAM,
 *                 proto, <system.sockaddr*>&rai_addr_static)             # <<<<<<<<<<<<<<
 * 
 *             if addr is None:
 */
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_proto); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1942, __pyx_L1_error)

    /* "uvloop/loop.pyx":1940
 *             f1 = f2 = None
 * 
 *             addr = __static_getaddrinfo(             # <<<<<<<<<<<<<<
 *                 host, port, family, uv.SOCK_STREAM,
 *                 proto, <system.sockaddr*>&rai_addr_static)
 */
    __pyx_t_4 = __pyx_f_6uvloop_4loop___static_getaddrinfo(__pyx_cur_scope->__pyx_v_host, __pyx_cur_scope->__pyx_v_port, __pyx_t_8, SOCK_STREAM, __pyx_t_9, ((struct sockaddr *)(&__pyx_cur_scope->__pyx_v_rai_addr_static))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1940, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_addr = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1944
 *                 proto, <system.sockaddr*>&rai_addr_static)
 * 
 *             if addr is None:             # <<<<<<<<<<<<<<
 *                 f1 = self._getaddrinfo(
 *                     host, port, family,
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_addr == Py_None);
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1946
 *             if addr is None:
 *                 f1 = self._getaddrinfo(
 *                     host, port, family,             # <<<<<<<<<<<<<<
 *                     uv.SOCK_STREAM, proto, flags,
 *                     0)  # 0 == don't unpack
 */
      __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1946, __pyx_L1_error)

      /* "uvloop/loop.pyx":1947
 *                 f1 = self._getaddrinfo(
 *                     host, port, family,
 *                     uv.SOCK_STREAM, proto, flags,             # <<<<<<<<<<<<<<
 *                     0)  # 0 == don't unpack
 * 
 */
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_proto); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1947, __pyx_L1_error)
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_flags); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1947, __pyx_L1_error)

      /* "uvloop/loop.pyx":1945
 * 
 *             if addr is None:
 *                 f1 = self._getaddrinfo(             # <<<<<<<<<<<<<<
 *                     host, port, family,
 *                     uv.SOCK_STREAM, proto, flags,
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getaddrinfo(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_host, __pyx_cur_scope->__pyx_v_port, __pyx_t_9, SOCK_STREAM, __pyx_t_8, __pyx_t_10, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1945, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_f1);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_f1, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1950
 *                     0)  # 0 == don't unpack
 * 
 *                 fs.append(f1)             # <<<<<<<<<<<<<<
 *             else:
 *                 rai_static.ai_addr = <system.sockaddr*>&rai_addr_static
 */
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_fs, __pyx_cur_scope->__pyx_v_f1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1950, __pyx_L1_error)

      /* "uvloop/loop.pyx":1944
 *                 proto, <system.sockaddr*>&rai_addr_static)
 * 
 *             if addr is None:             # <<<<<<<<<<<<<<
 *                 f1 = self._getaddrinfo(
 *                     host, port, family,
 */
      goto __pyx_L21;
    }

    /* "uvloop/loop.pyx":1952
 *                 fs.append(f1)
 *             else:
 *                 rai_static.ai_addr = <system.sockaddr*>&rai_addr_static             # <<<<<<<<<<<<<<
 *                 rai_static.ai_next = NULL
 *                 rai = &rai_static
 */
    /*else*/ {
      __pyx_cur_scope->__pyx_v_rai_static.ai_addr = ((struct sockaddr *)(&__pyx_cur_scope->__pyx_v_rai_addr_static));

      /* "uvloop/loop.pyx":1953
 *             else:
 *                 rai_static.ai_addr = <system.sockaddr*>&rai_addr_static
 *                 rai_static.ai_next = NULL             # <<<<<<<<<<<<<<
 *                 rai = &rai_static
 * 
 */
      __pyx_cur_scope->__pyx_v_rai_static.ai_next = NULL;

      /* "uvloop/loop.pyx":1954
 *                 rai_static.ai_addr = <system.sockaddr*>&rai_addr_static
 *                 rai_static.ai_next = NULL
 *                 rai = &rai_static             # <<<<<<<<<<<<<<
 * 
 *             if local_addr is not None:
 */
      __pyx_cur_scope->__pyx_v_rai = (&__pyx_cur_scope->__pyx_v_rai_static);
    }
    __pyx_L21:;

    /* "uvloop/loop.pyx":1956
 *                 rai = &rai_static
 * 
 *             if local_addr is not None:             # <<<<<<<<<<<<<<
 *                 if not isinstance(local_addr, (tuple, list)) or \
 *                         len(local_addr) != 2:
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_local_addr != Py_None);
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1957
 * 
 *             if local_addr is not None:
 *                 if not isinstance(local_addr, (tuple, list)) or \             # <<<<<<<<<<<<<<
 *                         len(local_addr) != 2:
 *                     raise ValueError(
 */
      __pyx_t_12 = PyTuple_Check(__pyx_cur_scope->__pyx_v_local_addr); 
      if (!__pyx_t_12) {
      } else {
        __pyx_t_1 = __pyx_t_12;
        goto __pyx_L26_bool_binop_done;
      }
      __pyx_t_12 = PyList_Check(__pyx_cur_scope->__pyx_v_local_addr); 
      __pyx_t_1 = __pyx_t_12;
      __pyx_L26_bool_binop_done:;
      __pyx_t_12 = (!__pyx_t_1);
      if (!__pyx_t_12) {
      } else {
        __pyx_t_2 = __pyx_t_12;
        goto __pyx_L24_bool_binop_done;
      }

      /* "uvloop/loop.pyx":1958
 *             if local_addr is not None:
 *                 if not isinstance(local_addr, (tuple, list)) or \
 *                         len(local_addr) != 2:             # <<<<<<<<<<<<<<
 *                     raise ValueError(
 *                         'local_addr must be a tuple of host and port')
 */
      __pyx_t_13 = PyObject_Length(__pyx_cur_scope->__pyx_v_local_addr); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1958, __pyx_L1_error)
      __pyx_t_12 = (__pyx_t_13 != 2);
      __pyx_t_2 = __pyx_t_12;
      __pyx_L24_bool_binop_done:;

      /* "uvloop/loop.pyx":1957
 * 
 *             if local_addr is not None:
 *                 if not isinstance(local_addr, (tuple, list)) or \             # <<<<<<<<<<<<<<
 *                         len(local_addr) != 2:
 *                     raise ValueError(
 */
      if (unlikely(__pyx_t_2)) {

        /* "uvloop/loop.pyx":1959
 *                 if not isinstance(local_addr, (tuple, list)) or \
 *                         len(local_addr) != 2:
 *                     raise ValueError(             # <<<<<<<<<<<<<<
 *                         'local_addr must be a tuple of host and port')
 * 
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__66, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1959, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1959, __pyx_L1_error)

        /* "uvloop/loop.pyx":1957
 * 
 *             if local_addr is not None:
 *                 if not isinstance(local_addr, (tuple, list)) or \             # <<<<<<<<<<<<<<
 *                         len(local_addr) != 2:
 *                     raise ValueError(
 */
      }

      /* "uvloop/loop.pyx":1963
 * 
 *                 addr = __static_getaddrinfo(
 *                     local_addr[0], local_addr[1],             # <<<<<<<<<<<<<<
 *                     family, uv.SOCK_STREAM,
 *                     proto, <system.sockaddr*>&lai_addr_static)
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_local_addr, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_local_addr, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1963, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":1964
 *                 addr = __static_getaddrinfo(
 *                     local_addr[0], local_addr[1],
 *                     family, uv.SOCK_STREAM,             # <<<<<<<<<<<<<<
 *                     proto, <system.sockaddr*>&lai_addr_static)
 *                 if addr is None:
 */
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1964, __pyx_L1_error)

      /* "uvloop/loop.pyx":1965
 *                     local_addr[0], local_addr[1],
 *                     family, uv.SOCK_STREAM,
 *                     proto, <system.sockaddr*>&lai_addr_static)             # <<<<<<<<<<<<<<
 *                 if addr is None:
 *                     f2 = self._getaddrinfo(
 */
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_proto); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1965, __pyx_L1_error)

      /* "uvloop/loop.pyx":1962
 *                         'local_addr must be a tuple of host and port')
 * 
 *                 addr = __static_getaddrinfo(             # <<<<<<<<<<<<<<
 *                     local_addr[0], local_addr[1],
 *                     family, uv.SOCK_STREAM,
 */
      __pyx_t_6 = __pyx_f_6uvloop_4loop___static_getaddrinfo(__pyx_t_4, __pyx_t_3, __pyx_t_10, SOCK_STREAM, __pyx_t_8, ((struct sockaddr *)(&__pyx_cur_scope->__pyx_v_lai_addr_static))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1962, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_addr);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_addr, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":1966
 *                     family, uv.SOCK_STREAM,
 *                     proto, <system.sockaddr*>&lai_addr_static)
 *                 if addr is None:             # <<<<<<<<<<<<<<
 *                     f2 = self._getaddrinfo(
 *                         local_addr[0], local_addr[1], family,
 */
      __pyx_t_2 = (__pyx_cur_scope->__pyx_v_addr == Py_None);
      if (__pyx_t_2) {

        /* "uvloop/loop.pyx":1968
 *                 if addr is None:
 *                     f2 = self._getaddrinfo(
 *                         local_addr[0], local_addr[1], family,             # <<<<<<<<<<<<<<
 *                         uv.SOCK_STREAM, proto, flags,
 *                         0)  # 0 == don't unpack
 */
        __pyx_t_6 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_local_addr, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1968, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_local_addr, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1968, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1968, __pyx_L1_error)

        /* "uvloop/loop.pyx":1969
 *                     f2 = self._getaddrinfo(
 *                         local_addr[0], local_addr[1], family,
 *                         uv.SOCK_STREAM, proto, flags,             # <<<<<<<<<<<<<<
 *                         0)  # 0 == don't unpack
 * 
 */
        __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_proto); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1969, __pyx_L1_error)
        __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_flags); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1969, __pyx_L1_error)

        /* "uvloop/loop.pyx":1967
 *                     proto, <system.sockaddr*>&lai_addr_static)
 *                 if addr is None:
 *                     f2 = self._getaddrinfo(             # <<<<<<<<<<<<<<
 *                         local_addr[0], local_addr[1], family,
 *                         uv.SOCK_STREAM, proto, flags,
 */
        __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getaddrinfo(__pyx_cur_scope->__pyx_v_self, __pyx_t_6, __pyx_t_3, __pyx_t_8, SOCK_STREAM, __pyx_t_10, __pyx_t_9, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1967, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_f2);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_f2, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;

        /* "uvloop/loop.pyx":1972
 *                         0)  # 0 == don't unpack
 * 
 *                     fs.append(f2)             # <<<<<<<<<<<<<<
 *                 else:
 *                     lai_static.ai_addr = <system.sockaddr*>&lai_addr_static
 */
        __pyx_t_11 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_fs, __pyx_cur_scope->__pyx_v_f2); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1972, __pyx_L1_error)

        /* "uvloop/loop.pyx":1966
 *                     family, uv.SOCK_STREAM,
 *                     proto, <system.sockaddr*>&lai_addr_static)
 *                 if addr is None:             # <<<<<<<<<<<<<<
 *                     f2 = self._getaddrinfo(
 *                         local_addr[0], local_addr[1], family,
 */
        goto __pyx_L28;
      }

      /* "uvloop/loop.pyx":1974
 *                     fs.append(f2)
 *                 else:
 *                     lai_static.ai_addr = <system.sockaddr*>&lai_addr_static             # <<<<<<<<<<<<<<
 *                     lai_static.ai_next = NULL
 *                     lai = &lai_static
 */
      /*else*/ {
        __pyx_cur_scope->__pyx_v_lai_static.ai_addr = ((struct sockaddr *)(&__pyx_cur_scope->__pyx_v_lai_addr_static));

        /* "uvloop/loop.pyx":1975
 *                 else:
 *                     lai_static.ai_addr = <system.sockaddr*>&lai_addr_static
 *                     lai_static.ai_next = NULL             # <<<<<<<<<<<<<<
 *                     lai = &lai_static
 * 
 */
        __pyx_cur_scope->__pyx_v_lai_static.ai_next = NULL;

        /* "uvloop/loop.pyx":1976
 *                     lai_static.ai_addr = <system.sockaddr*>&lai_addr_static
 *                     lai_static.ai_next = NULL
 *                     lai = &lai_static             # <<<<<<<<<<<<<<
 * 
 *             if len(fs):
 */
        __pyx_cur_scope->__pyx_v_lai = (&__pyx_cur_scope->__pyx_v_lai_static);
      }
      __pyx_L28:;

      /* "uvloop/loop.pyx":1956
 *                 rai = &rai_static
 * 
 *             if local_addr is not None:             # <<<<<<<<<<<<<<
 *                 if not isinstance(local_addr, (tuple, list)) or \
 *                         len(local_addr) != 2:
 */
    }

    /* "uvloop/loop.pyx":1978
 *                     lai = &lai_static
 * 
 *             if len(fs):             # <<<<<<<<<<<<<<
 *                 await aio_wait(fs)
 * 
 */
    __pyx_t_13 = __Pyx_PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_fs); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1978, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_13 != 0);
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1979
 * 
 *             if len(fs):
 *                 await aio_wait(fs)             # <<<<<<<<<<<<<<
 * 
 *             if rai is NULL:
 */
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_wait);
      __pyx_t_3 = __pyx_v_6uvloop_4loop_aio_wait; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_fs};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1979, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 2;
        return __pyx_r;
        __pyx_L30_resume_from_await:;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1979, __pyx_L1_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 1979, __pyx_L1_error)
        }
      }

      /* "uvloop/loop.pyx":1978
 *                     lai = &lai_static
 * 
 *             if len(fs):             # <<<<<<<<<<<<<<
 *                 await aio_wait(fs)
 * 
 */
    }

    /* "uvloop/loop.pyx":1981
 *                 await aio_wait(fs)
 * 
 *             if rai is NULL:             # <<<<<<<<<<<<<<
 *                 ai_remote = f1.result()
 *                 if ai_remote.data is NULL:
 */
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_rai == NULL);
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1982
 * 
 *             if rai is NULL:
 *                 ai_remote = f1.result()             # <<<<<<<<<<<<<<
 *                 if ai_remote.data is NULL:
 *                     raise OSError('getaddrinfo() returned empty list')
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f1, __pyx_n_s_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1982, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1982, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_6uvloop_4loop_AddrInfo))))) __PYX_ERR(0, 1982, __pyx_L1_error)
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_cur_scope->__pyx_v_ai_remote = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1983
 *             if rai is NULL:
 *                 ai_remote = f1.result()
 *                 if ai_remote.data is NULL:             # <<<<<<<<<<<<<<
 *                     raise OSError('getaddrinfo() returned empty list')
 *                 rai = ai_remote.data
 */
      __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ai_remote->data == NULL);
      if (unlikely(__pyx_t_2)) {

        /* "uvloop/loop.pyx":1984
 *                 ai_remote = f1.result()
 *                 if ai_remote.data is NULL:
 *                     raise OSError('getaddrinfo() returned empty list')             # <<<<<<<<<<<<<<
 *                 rai = ai_remote.data
 * 
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1984, __pyx_L1_error)

        /* "uvloop/loop.pyx":1983
 *             if rai is NULL:
 *                 ai_remote = f1.result()
 *                 if ai_remote.data is NULL:             # <<<<<<<<<<<<<<
 *                     raise OSError('getaddrinfo() returned empty list')
 *                 rai = ai_remote.data
 */
      }

      /* "uvloop/loop.pyx":1985
 *                 if ai_remote.data is NULL:
 *                     raise OSError('getaddrinfo() returned empty list')
 *                 rai = ai_remote.data             # <<<<<<<<<<<<<<
 * 
 *             if lai is NULL and f2 is not None:
 */
      __pyx_t_14 = __pyx_cur_scope->__pyx_v_ai_remote->data;
      __pyx_cur_scope->__pyx_v_rai = __pyx_t_14;

      /* "uvloop/loop.pyx":1981
 *                 await aio_wait(fs)
 * 
 *             if rai is NULL:             # <<<<<<<<<<<<<<
 *                 ai_remote = f1.result()
 *                 if ai_remote.data is NULL:
 */
    }

    /* "uvloop/loop.pyx":1987
 *                 rai = ai_remote.data
 * 
 *             if lai is NULL and f2 is not None:             # <<<<<<<<<<<<<<
 *                 ai_local = f2.result()
 *                 if ai_local.data is NULL:
 */
    __pyx_t_12 = (__pyx_cur_scope->__pyx_v_lai == NULL);
    if (__pyx_t_12) {
    } else {
      __pyx_t_2 = __pyx_t_12;
      goto __pyx_L34_bool_binop_done;
    }
    __pyx_t_12 = (__pyx_cur_scope->__pyx_v_f2 != Py_None);
    __pyx_t_2 = __pyx_t_12;
    __pyx_L34_bool_binop_done:;
    if (__pyx_t_2) {

      /* "uvloop/loop.pyx":1988
 * 
 *             if lai is NULL and f2 is not None:
 *                 ai_local = f2.result()             # <<<<<<<<<<<<<<
 *                 if ai_local.data is NULL:
 *                     raise OSError(
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f2, __pyx_n_s_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1988, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1988, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_6uvloop_4loop_AddrInfo))))) __PYX_ERR(0, 1988, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_ai_local);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_ai_local, ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_t_4));
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":1989
 *             if lai is NULL and f2 is not None:
 *                 ai_local = f2.result()
 *                 if ai_local.data is NULL:             # <<<<<<<<<<<<<<
 *                     raise OSError(
 *                         'getaddrinfo() returned empty list for local_addr')
 */
      __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ai_local->data == NULL);
      if (unlikely(__pyx_t_2)) {

        /* "uvloop/loop.pyx":1990
 *                 ai_local = f2.result()
 *                 if ai_local.data is NULL:
 *                     raise OSError(             # <<<<<<<<<<<<<<
 *                         'getaddrinfo() returned empty list for local_addr')
 *                 lai = ai_local.data
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__68, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1990, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1990, __pyx_L1_error)

        /* "uvloop/loop.pyx":1989
 *             if lai is NULL and f2 is not None:
 *                 ai_local = f2.result()
 *                 if ai_local.data is NULL:             # <<<<<<<<<<<<<<
 *                     raise OSError(
 *                         'getaddrinfo() returned empty list for local_addr')
 */
      }

      /* "uvloop/loop.pyx":1992
 *                     raise OSError(
 *                         'getaddrinfo() returned empty list for local_addr')
 *                 lai = ai_local.data             # <<<<<<<<<<<<<<
 * 
 *             exceptions = []
 */
      __pyx_t_14 = __pyx_cur_scope->__pyx_v_ai_local->data;
      __pyx_cur_scope->__pyx_v_lai = __pyx_t_14;

      /* "uvloop/loop.pyx":1987
 *                 rai = ai_remote.data
 * 
 *             if lai is NULL and f2 is not None:             # <<<<<<<<<<<<<<
 *                 ai_local = f2.result()
 *                 if ai_local.data is NULL:
 */
    }

    /* "uvloop/loop.pyx":1994
 *                 lai = ai_local.data
 * 
 *             exceptions = []             # <<<<<<<<<<<<<<
 *             rai_iter = rai
 *             while rai_iter is not NULL:
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_exceptions = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":1995
 * 
 *             exceptions = []
 *             rai_iter = rai             # <<<<<<<<<<<<<<
 *             while rai_iter is not NULL:
 *                 tr = None
 */
    __pyx_cur_scope->__pyx_v_rai_iter = __pyx_cur_scope->__pyx_v_rai;

    /* "uvloop/loop.pyx":1996
 *             exceptions = []
 *             rai_iter = rai
 *             while rai_iter is not NULL:             # <<<<<<<<<<<<<<
 *                 tr = None
 *                 try:
 */
    while (1) {
      __pyx_t_2 = (__pyx_cur_scope->__pyx_v_rai_iter != NULL);
      if (!__pyx_t_2) break;

      /* "uvloop/loop.pyx":1997
 *             rai_iter = rai
 *             while rai_iter is not NULL:
 *                 tr = None             # <<<<<<<<<<<<<<
 *                 try:
 *                     waiter = self._new_future()
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_tr, ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)Py_None));
      __Pyx_GIVEREF(Py_None);

      /* "uvloop/loop.pyx":1998
 *             while rai_iter is not NULL:
 *                 tr = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     waiter = self._new_future()
 *                     tr = TCPTransport.new(self, protocol, None, waiter,
 */
      {
        __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_17);
        /*try:*/ {

          /* "uvloop/loop.pyx":1999
 *                 tr = None
 *                 try:
 *                     waiter = self._new_future()             # <<<<<<<<<<<<<<
 *                     tr = TCPTransport.new(self, protocol, None, waiter,
 *                                           context)
 */
          __pyx_t_4 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1999, __pyx_L39_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_waiter);
          __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_waiter, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_4);
          __pyx_t_4 = 0;

          /* "uvloop/loop.pyx":2000
 *                 try:
 *                     waiter = self._new_future()
 *                     tr = TCPTransport.new(self, protocol, None, waiter,             # <<<<<<<<<<<<<<
 *                                           context)
 * 
 */
          __pyx_t_4 = ((PyObject *)__pyx_f_6uvloop_4loop_12TCPTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_waiter, __pyx_cur_scope->__pyx_v_context)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2000, __pyx_L39_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
          __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_tr, ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_t_4));
          __Pyx_GIVEREF(__pyx_t_4);
          __pyx_t_4 = 0;

          /* "uvloop/loop.pyx":2003
 *                                           context)
 * 
 *                     if lai is not NULL:             # <<<<<<<<<<<<<<
 *                         lai_iter = lai
 *                         while lai_iter is not NULL:
 */
          __pyx_t_2 = (__pyx_cur_scope->__pyx_v_lai != NULL);
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":2004
 * 
 *                     if lai is not NULL:
 *                         lai_iter = lai             # <<<<<<<<<<<<<<
 *                         while lai_iter is not NULL:
 *                             try:
 */
            __pyx_cur_scope->__pyx_v_lai_iter = __pyx_cur_scope->__pyx_v_lai;

            /* "uvloop/loop.pyx":2005
 *                     if lai is not NULL:
 *                         lai_iter = lai
 *                         while lai_iter is not NULL:             # <<<<<<<<<<<<<<
 *                             try:
 *                                 tr.bind(lai_iter.ai_addr)
 */
            while (1) {
              __pyx_t_2 = (__pyx_cur_scope->__pyx_v_lai_iter != NULL);
              if (!__pyx_t_2) break;

              /* "uvloop/loop.pyx":2006
 *                         lai_iter = lai
 *                         while lai_iter is not NULL:
 *                             try:             # <<<<<<<<<<<<<<
 *                                 tr.bind(lai_iter.ai_addr)
 *                                 break
 */
              {
                __Pyx_ExceptionSave(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                /*try:*/ {

                  /* "uvloop/loop.pyx":2007
 *                         while lai_iter is not NULL:
 *                             try:
 *                                 tr.bind(lai_iter.ai_addr)             # <<<<<<<<<<<<<<
 *                                 break
 *                             except OSError as exc:
 */
                  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->bind(__pyx_cur_scope->__pyx_v_tr, __pyx_cur_scope->__pyx_v_lai_iter->ai_addr, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2007, __pyx_L50_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

                  /* "uvloop/loop.pyx":2008
 *                             try:
 *                                 tr.bind(lai_iter.ai_addr)
 *                                 break             # <<<<<<<<<<<<<<
 *                             except OSError as exc:
 *                                 exceptions.append(exc)
 */
                  goto __pyx_L55_try_break;

                  /* "uvloop/loop.pyx":2006
 *                         lai_iter = lai
 *                         while lai_iter is not NULL:
 *                             try:             # <<<<<<<<<<<<<<
 *                                 tr.bind(lai_iter.ai_addr)
 *                                 break
 */
                }
                __pyx_L50_error:;
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "uvloop/loop.pyx":2009
 *                                 tr.bind(lai_iter.ai_addr)
 *                                 break
 *                             except OSError as exc:             # <<<<<<<<<<<<<<
 *                                 exceptions.append(exc)
 *                             lai_iter = lai_iter.ai_next
 */
                __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
                if (__pyx_t_9) {
                  __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
                  if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_6) < 0) __PYX_ERR(0, 2009, __pyx_L52_except_error)
                  __Pyx_XGOTREF(__pyx_t_4);
                  __Pyx_XGOTREF(__pyx_t_3);
                  __Pyx_XGOTREF(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_3);
                  __pyx_cur_scope->__pyx_v_exc = __pyx_t_3;
                  /*try:*/ {

                    /* "uvloop/loop.pyx":2010
 *                                 break
 *                             except OSError as exc:
 *                                 exceptions.append(exc)             # <<<<<<<<<<<<<<
 *                             lai_iter = lai_iter.ai_next
 *                         else:
 */
                    __pyx_t_11 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_exceptions, __pyx_cur_scope->__pyx_v_exc); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 2010, __pyx_L63_error)
                  }

                  /* "uvloop/loop.pyx":2009
 *                                 tr.bind(lai_iter.ai_addr)
 *                                 break
 *                             except OSError as exc:             # <<<<<<<<<<<<<<
 *                                 exceptions.append(exc)
 *                             lai_iter = lai_iter.ai_next
 */
                  /*finally:*/ {
                    /*normal exit:*/{
                      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exc);
                      __Pyx_DECREF(__pyx_cur_scope->__pyx_v_exc); __pyx_cur_scope->__pyx_v_exc = 0;
                      goto __pyx_L64;
                    }
                    __pyx_L63_error:;
                    /*exception exit:*/{
                      __Pyx_PyThreadState_assign
                      __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
                      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_25, &__pyx_t_26, &__pyx_t_27);
                      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24) < 0)) __Pyx_ErrFetch(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
                      __Pyx_XGOTREF(__pyx_t_22);
                      __Pyx_XGOTREF(__pyx_t_23);
                      __Pyx_XGOTREF(__pyx_t_24);
                      __Pyx_XGOTREF(__pyx_t_25);
                      __Pyx_XGOTREF(__pyx_t_26);
                      __Pyx_XGOTREF(__pyx_t_27);
                      __pyx_t_9 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
                      {
                        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exc);
                        __Pyx_DECREF(__pyx_cur_scope->__pyx_v_exc); __pyx_cur_scope->__pyx_v_exc = 0;
                      }
                      if (PY_MAJOR_VERSION >= 3) {
                        __Pyx_XGIVEREF(__pyx_t_25);
                        __Pyx_XGIVEREF(__pyx_t_26);
                        __Pyx_XGIVEREF(__pyx_t_27);
                        __Pyx_ExceptionReset(__pyx_t_25, __pyx_t_26, __pyx_t_27);
                      }
                      __Pyx_XGIVEREF(__pyx_t_22);
                      __Pyx_XGIVEREF(__pyx_t_23);
                      __Pyx_XGIVEREF(__pyx_t_24);
                      __Pyx_ErrRestore(__pyx_t_22, __pyx_t_23, __pyx_t_24);
                      __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
                      __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_21;
                      goto __pyx_L52_except_error;
                    }
                    __pyx_L64:;
                  }
                  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                  goto __pyx_L51_exception_handled;
                }
                goto __pyx_L52_except_error;

                /* "uvloop/loop.pyx":2006
 *                         lai_iter = lai
 *                         while lai_iter is not NULL:
 *                             try:             # <<<<<<<<<<<<<<
 *                                 tr.bind(lai_iter.ai_addr)
 *                                 break
 */
                __pyx_L52_except_error:;
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
                goto __pyx_L39_error;
                __pyx_L55_try_break:;
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
                goto __pyx_L49_break;
                __pyx_L51_exception_handled:;
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
              }

              /* "uvloop/loop.pyx":2011
 *                             except OSError as exc:
 *                                 exceptions.append(exc)
 *                             lai_iter = lai_iter.ai_next             # <<<<<<<<<<<<<<
 *                         else:
 *                             tr._close()
 */
              __pyx_t_14 = __pyx_cur_scope->__pyx_v_lai_iter->ai_next;
              __pyx_cur_scope->__pyx_v_lai_iter = __pyx_t_14;
            }

            /* "uvloop/loop.pyx":2013
 *                             lai_iter = lai_iter.ai_next
 *                         else:
 *                             tr._close()             # <<<<<<<<<<<<<<
 *                             tr = None
 * 
 */
            /*else*/ {
              __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2013, __pyx_L39_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

              /* "uvloop/loop.pyx":2014
 *                         else:
 *                             tr._close()
 *                             tr = None             # <<<<<<<<<<<<<<
 * 
 *                             rai_iter = rai_iter.ai_next
 */
              __Pyx_INCREF(Py_None);
              __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_tr, ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)Py_None));
              __Pyx_GIVEREF(Py_None);

              /* "uvloop/loop.pyx":2016
 *                             tr = None
 * 
 *                             rai_iter = rai_iter.ai_next             # <<<<<<<<<<<<<<
 *                             continue
 * 
 */
              __pyx_t_14 = __pyx_cur_scope->__pyx_v_rai_iter->ai_next;
              __pyx_cur_scope->__pyx_v_rai_iter = __pyx_t_14;

              /* "uvloop/loop.pyx":2017
 * 
 *                             rai_iter = rai_iter.ai_next
 *                             continue             # <<<<<<<<<<<<<<
 * 
 *                     tr.connect(rai_iter.ai_addr)
 */
              goto __pyx_L45_try_continue;
            }
            __pyx_L49_break:;

            /* "uvloop/loop.pyx":2003
 *                                           context)
 * 
 *                     if lai is not NULL:             # <<<<<<<<<<<<<<
 *                         lai_iter = lai
 *                         while lai_iter is not NULL:
 */
          }

          /* "uvloop/loop.pyx":2019
 *                             continue
 * 
 *                     tr.connect(rai_iter.ai_addr)             # <<<<<<<<<<<<<<
 *                     await waiter
 * 
 */
          __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->connect(__pyx_cur_scope->__pyx_v_tr, __pyx_cur_scope->__pyx_v_rai_iter->ai_addr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2019, __pyx_L39_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "uvloop/loop.pyx":2020
 * 
 *                     tr.connect(rai_iter.ai_addr)
 *                     await waiter             # <<<<<<<<<<<<<<
 * 
 *                 except OSError as exc:
 */
          __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
          __Pyx_XGOTREF(__pyx_r);
          if (likely(__pyx_r)) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __pyx_cur_scope->__pyx_t_0 = __pyx_t_15;
            __Pyx_XGIVEREF(__pyx_t_16);
            __pyx_cur_scope->__pyx_t_1 = __pyx_t_16;
            __Pyx_XGIVEREF(__pyx_t_17);
            __pyx_cur_scope->__pyx_t_2 = __pyx_t_17;
            __Pyx_XGIVEREF(__pyx_r);
            __Pyx_RefNannyFinishContext();
            __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
            /* return from generator, awaiting value */
            __pyx_generator->resume_label = 3;
            return __pyx_r;
            __pyx_L69_resume_from_await:;
            __pyx_t_15 = __pyx_cur_scope->__pyx_t_0;
            __pyx_cur_scope->__pyx_t_0 = 0;
            __Pyx_XGOTREF(__pyx_t_15);
            __pyx_t_16 = __pyx_cur_scope->__pyx_t_1;
            __pyx_cur_scope->__pyx_t_1 = 0;
            __Pyx_XGOTREF(__pyx_t_16);
            __pyx_t_17 = __pyx_cur_scope->__pyx_t_2;
            __pyx_cur_scope->__pyx_t_2 = 0;
            __Pyx_XGOTREF(__pyx_t_17);
            if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2020, __pyx_L39_error)
          } else {
            PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
              else __PYX_ERR(0, 2020, __pyx_L39_error)
            }
          }

          /* "uvloop/loop.pyx":1998
 *             while rai_iter is not NULL:
 *                 tr = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     waiter = self._new_future()
 *                     tr = TCPTransport.new(self, protocol, None, waiter,
 */
        }

        /* "uvloop/loop.pyx":2035
 *                     raise
 *                 else:
 *                     break             # <<<<<<<<<<<<<<
 * 
 *                 rai_iter = rai_iter.ai_next
 */
        /*else:*/ {
          goto __pyx_L44_try_break;
        }
        __pyx_L39_error:;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "uvloop/loop.pyx":2022
 *                     await waiter
 * 
 *                 except OSError as exc:             # <<<<<<<<<<<<<<
 *                     if tr is not None:
 *                         tr._close()
 */
        __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
        if (__pyx_t_10) {
          __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 2022, __pyx_L41_except_error)
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_exc);
          __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_exc, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_3);
          /*try:*/ {

            /* "uvloop/loop.pyx":2023
 * 
 *                 except OSError as exc:
 *                     if tr is not None:             # <<<<<<<<<<<<<<
 *                         tr._close()
 *                         tr = None
 */
            __pyx_t_2 = (((PyObject *)__pyx_cur_scope->__pyx_v_tr) != Py_None);
            if (__pyx_t_2) {

              /* "uvloop/loop.pyx":2024
 *                 except OSError as exc:
 *                     if tr is not None:
 *                         tr._close()             # <<<<<<<<<<<<<<
 *                         tr = None
 *                     exceptions.append(exc)
 */
              __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2024, __pyx_L75_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "uvloop/loop.pyx":2025
 *                     if tr is not None:
 *                         tr._close()
 *                         tr = None             # <<<<<<<<<<<<<<
 *                     exceptions.append(exc)
 *                 except (KeyboardInterrupt, SystemExit):
 */
              __Pyx_INCREF(Py_None);
              __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
              __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_tr, ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)Py_None));
              __Pyx_GIVEREF(Py_None);

              /* "uvloop/loop.pyx":2023
 * 
 *                 except OSError as exc:
 *                     if tr is not None:             # <<<<<<<<<<<<<<
 *                         tr._close()
 *                         tr = None
 */
            }

            /* "uvloop/loop.pyx":2026
 *                         tr._close()
 *                         tr = None
 *                     exceptions.append(exc)             # <<<<<<<<<<<<<<
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 */
            __pyx_t_11 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_exceptions, __pyx_cur_scope->__pyx_v_exc); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 2026, __pyx_L75_error)
          }

          /* "uvloop/loop.pyx":2022
 *                     await waiter
 * 
 *                 except OSError as exc:             # <<<<<<<<<<<<<<
 *                     if tr is not None:
 *                         tr._close()
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exc);
              __Pyx_DECREF(__pyx_cur_scope->__pyx_v_exc); __pyx_cur_scope->__pyx_v_exc = 0;
              goto __pyx_L76;
            }
            __pyx_L75_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_assign
              __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_27 = 0; __pyx_t_26 = 0; __pyx_t_25 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_27, &__pyx_t_26, &__pyx_t_25);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_27);
              __Pyx_XGOTREF(__pyx_t_26);
              __Pyx_XGOTREF(__pyx_t_25);
              __pyx_t_10 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_28 = __pyx_filename;
              {
                __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exc);
                __Pyx_DECREF(__pyx_cur_scope->__pyx_v_exc); __pyx_cur_scope->__pyx_v_exc = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_27);
                __Pyx_XGIVEREF(__pyx_t_26);
                __Pyx_XGIVEREF(__pyx_t_25);
                __Pyx_ExceptionReset(__pyx_t_27, __pyx_t_26, __pyx_t_25);
              }
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ErrRestore(__pyx_t_20, __pyx_t_19, __pyx_t_18);
              __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_27 = 0; __pyx_t_26 = 0; __pyx_t_25 = 0;
              __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_28;
              goto __pyx_L41_except_error;
            }
            __pyx_L76:;
          }
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L40_exception_handled;
        }

        /* "uvloop/loop.pyx":2027
 *                         tr = None
 *                     exceptions.append(exc)
 *                 except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                     raise
 *                 except BaseException:
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_6) < 0) __PYX_ERR(0, 2027, __pyx_L41_except_error)
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_6);

          /* "uvloop/loop.pyx":2028
 *                     exceptions.append(exc)
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise             # <<<<<<<<<<<<<<
 *                 except BaseException:
 *                     if tr is not None:
 */
          __Pyx_GIVEREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_3);
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_3, __pyx_t_6);
          __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_6 = 0; 
          __PYX_ERR(0, 2028, __pyx_L41_except_error)
        }

        /* "uvloop/loop.pyx":2029
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 *                 except BaseException:             # <<<<<<<<<<<<<<
 *                     if tr is not None:
 *                         tr._close()
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 2029, __pyx_L41_except_error)
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_4);

          /* "uvloop/loop.pyx":2030
 *                     raise
 *                 except BaseException:
 *                     if tr is not None:             # <<<<<<<<<<<<<<
 *                         tr._close()
 *                         tr = None
 */
          __pyx_t_2 = (((PyObject *)__pyx_cur_scope->__pyx_v_tr) != Py_None);
          if (__pyx_t_2) {

            /* "uvloop/loop.pyx":2031
 *                 except BaseException:
 *                     if tr is not None:
 *                         tr._close()             # <<<<<<<<<<<<<<
 *                         tr = None
 *                     raise
 */
            __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2031, __pyx_L41_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

            /* "uvloop/loop.pyx":2032
 *                     if tr is not None:
 *                         tr._close()
 *                         tr = None             # <<<<<<<<<<<<<<
 *                     raise
 *                 else:
 */
            __Pyx_INCREF(Py_None);
            __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
            __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_tr, ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)Py_None));
            __Pyx_GIVEREF(Py_None);

            /* "uvloop/loop.pyx":2030
 *                     raise
 *                 except BaseException:
 *                     if tr is not None:             # <<<<<<<<<<<<<<
 *                         tr._close()
 *                         tr = None
 */
          }

          /* "uvloop/loop.pyx":2033
 *                         tr._close()
 *                         tr = None
 *                     raise             # <<<<<<<<<<<<<<
 *                 else:
 *                     break
 */
          __Pyx_GIVEREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_3);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_3, __pyx_t_4);
          __pyx_t_6 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0; 
          __PYX_ERR(0, 2033, __pyx_L41_except_error)
        }
        goto __pyx_L41_except_error;

        /* "uvloop/loop.pyx":1998
 *             while rai_iter is not NULL:
 *                 tr = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     waiter = self._new_future()
 *                     tr = TCPTransport.new(self, protocol, None, waiter,
 */
        __pyx_L41_except_error:;
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
        goto __pyx_L1_error;
        __pyx_L44_try_break:;
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
        goto __pyx_L38_break;
        __pyx_L45_try_continue:;
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
        goto __pyx_L37_continue;
        __pyx_L40_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      }

      /* "uvloop/loop.pyx":2037
 *                     break
 * 
 *                 rai_iter = rai_iter.ai_next             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __pyx_t_14 = __pyx_cur_scope->__pyx_v_rai_iter->ai_next;
      __pyx_cur_scope->__pyx_v_rai_iter = __pyx_t_14;
      __pyx_L37_continue:;
    }

    /* "uvloop/loop.pyx":2041
 *             else:
 *                 # If they all have the same str(), raise one.
 *                 model = str(exceptions[0])             # <<<<<<<<<<<<<<
 *                 if all(str(exc) == model for exc in exceptions):
 *                     raise exceptions[0]
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_exceptions, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_v_model = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":2042
 *                 # If they all have the same str(), raise one.
 *                 model = str(exceptions[0])
 *                 if all(str(exc) == model for exc in exceptions):             # <<<<<<<<<<<<<<
 *                     raise exceptions[0]
 *                 # Raise a combined exception so the user can see all
 */
      __pyx_t_3 = __pyx_pf_6uvloop_4loop_4Loop_17create_connection_genexpr(((PyObject*)__pyx_cur_scope), __pyx_cur_scope->__pyx_v_exceptions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2042, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2042, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2042, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(__pyx_t_2)) {

        /* "uvloop/loop.pyx":2043
 *                 model = str(exceptions[0])
 *                 if all(str(exc) == model for exc in exceptions):
 *                     raise exceptions[0]             # <<<<<<<<<<<<<<
 *                 # Raise a combined exception so the user can see all
 *                 # the various error messages.
 */
        __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_exceptions, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 2043, __pyx_L1_error)

        /* "uvloop/loop.pyx":2042
 *                 # If they all have the same str(), raise one.
 *                 model = str(exceptions[0])
 *                 if all(str(exc) == model for exc in exceptions):             # <<<<<<<<<<<<<<
 *                     raise exceptions[0]
 *                 # Raise a combined exception so the user can see all
 */
      }

      /* "uvloop/loop.pyx":2046
 *                 # Raise a combined exception so the user can see all
 *                 # the various error messages.
 *                 raise OSError('Multiple exceptions: {}'.format(             # <<<<<<<<<<<<<<
 *                     ', '.join(str(exc) for exc in exceptions)))
 *         else:
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Multiple_exceptions, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":2047
 *                 # the various error messages.
 *                 raise OSError('Multiple exceptions: {}'.format(
 *                     ', '.join(str(exc) for exc in exceptions)))             # <<<<<<<<<<<<<<
 *         else:
 *             if sock is None:
 */
      __pyx_t_6 = __pyx_pf_6uvloop_4loop_4Loop_17create_connection_3genexpr(NULL, __pyx_cur_scope->__pyx_v_exceptions); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2047, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_Generator_Next(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2047, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyUnicode_Join(__pyx_kp_u__69, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2047, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2046, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/loop.pyx":2046
 *                 # Raise a combined exception so the user can see all
 *                 # the various error messages.
 *                 raise OSError('Multiple exceptions: {}'.format(             # <<<<<<<<<<<<<<
 *                     ', '.join(str(exc) for exc in exceptions)))
 *         else:
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_OSError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2046, __pyx_L1_error)
    }
    __pyx_L38_break:;

    /* "uvloop/loop.pyx":1932
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if host is not None or port is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
    goto __pyx_L17;
  }

  /* "uvloop/loop.pyx":2049
 *                     ', '.join(str(exc) for exc in exceptions)))
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host and port was not specified and no sock specified')
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_cur_scope->__pyx_v_sock == Py_None);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/loop.pyx":2050
 *         else:
 *             if sock is None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'host and port was not specified and no sock specified')
 *             if not _is_sock_stream(sock.type):
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__70, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2050, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2050, __pyx_L1_error)

      /* "uvloop/loop.pyx":2049
 *                     ', '.join(str(exc) for exc in exceptions)))
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'host and port was not specified and no sock specified')
 */
    }

    /* "uvloop/loop.pyx":2052
 *                 raise ValueError(
 *                     'host and port was not specified and no sock specified')
 *             if not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A Stream Socket was expected, got {!r}'.format(sock))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_f_6uvloop_4loop__is_sock_stream(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2052, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_12 = (!__pyx_t_2);
    if (unlikely(__pyx_t_12)) {

      /* "uvloop/loop.pyx":2054
 *             if not _is_sock_stream(sock.type):
 *                 raise ValueError(
 *                     'A Stream Socket was expected, got {!r}'.format(sock))             # <<<<<<<<<<<<<<
 * 
 *             # libuv will set the socket to non-blocking mode, but
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_Stream_Socket_was_expected_got, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_sock};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/loop.pyx":2053
 *                     'host and port was not specified and no sock specified')
 *             if not _is_sock_stream(sock.type):
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'A Stream Socket was expected, got {!r}'.format(sock))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2053, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2053, __pyx_L1_error)

      /* "uvloop/loop.pyx":2052
 *                 raise ValueError(
 *                     'host and port was not specified and no sock specified')
 *             if not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A Stream Socket was expected, got {!r}'.format(sock))
 */
    }

    /* "uvloop/loop.pyx":2058
 *             # libuv will set the socket to non-blocking mode, but
 *             # we want Python socket object to notice that.
 *             sock.setblocking(False)             # <<<<<<<<<<<<<<
 * 
 *             waiter = self._new_future()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2058, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, Py_False};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2058, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2060
 *             sock.setblocking(False)
 * 
 *             waiter = self._new_future()             # <<<<<<<<<<<<<<
 *             tr = TCPTransport.new(self, protocol, None, waiter, context)
 *             try:
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_waiter = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2061
 * 
 *             waiter = self._new_future()
 *             tr = TCPTransport.new(self, protocol, None, waiter, context)             # <<<<<<<<<<<<<<
 *             try:
 *                 # libuv will make socket non-blocking
 */
    __pyx_t_3 = ((PyObject *)__pyx_f_6uvloop_4loop_12TCPTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_waiter, __pyx_cur_scope->__pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_tr = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2062
 *             waiter = self._new_future()
 *             tr = TCPTransport.new(self, protocol, None, waiter, context)
 *             try:             # <<<<<<<<<<<<<<
 *                 # libuv will make socket non-blocking
 *                 tr._open(sock.fileno())
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_15);
      /*try:*/ {

        /* "uvloop/loop.pyx":2064
 *             try:
 *                 # libuv will make socket non-blocking
 *                 tr._open(sock.fileno())             # <<<<<<<<<<<<<<
 *                 tr._init_protocol()
 *                 await waiter
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2064, __pyx_L90_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2064, __pyx_L90_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2064, __pyx_L90_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_tr), __pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2064, __pyx_L90_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "uvloop/loop.pyx":2065
 *                 # libuv will make socket non-blocking
 *                 tr._open(sock.fileno())
 *                 tr._init_protocol()             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
        __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2065, __pyx_L90_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "uvloop/loop.pyx":2066
 *                 tr._open(sock.fileno())
 *                 tr._init_protocol()
 *                 await waiter             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_15);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_15;
          __Pyx_XGIVEREF(__pyx_t_16);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_16;
          __Pyx_XGIVEREF(__pyx_t_17);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_17;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 4;
          return __pyx_r;
          __pyx_L96_resume_from_await:;
          __pyx_t_15 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_15);
          __pyx_t_16 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_16);
          __pyx_t_17 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_17);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2066, __pyx_L90_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(0, 2066, __pyx_L90_error)
          }
        }

        /* "uvloop/loop.pyx":2062
 *             waiter = self._new_future()
 *             tr = TCPTransport.new(self, protocol, None, waiter, context)
 *             try:             # <<<<<<<<<<<<<<
 *                 # libuv will make socket non-blocking
 *                 tr._open(sock.fileno())
 */
      }
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      goto __pyx_L95_try_end;
      __pyx_L90_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":2067
 *                 tr._init_protocol()
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_6) < 0) __PYX_ERR(0, 2067, __pyx_L92_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "uvloop/loop.pyx":2068
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 # It's OK to call `_close()` here, as opposed to
 */
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_4, __pyx_t_6);
        __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_6 = 0; 
        __PYX_ERR(0, 2068, __pyx_L92_except_error)
      }

      /* "uvloop/loop.pyx":2069
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 # It's OK to call `_close()` here, as opposed to
 *                 # `_force_close()` or `close()` as we want to terminate the
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 2069, __pyx_L92_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "uvloop/loop.pyx":2075
 *                 # up in `Transport._call_connection_made()`, and calling
 *                 # `_close()` before it is fine.
 *                 tr._close()             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2075, __pyx_L92_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/loop.pyx":2076
 *                 # `_close()` before it is fine.
 *                 tr._close()
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *             tr._attach_fileobj(sock)
 */
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_4, __pyx_t_3);
        __pyx_t_6 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; 
        __PYX_ERR(0, 2076, __pyx_L92_except_error)
      }
      goto __pyx_L92_except_error;

      /* "uvloop/loop.pyx":2062
 *             waiter = self._new_future()
 *             tr = TCPTransport.new(self, protocol, None, waiter, context)
 *             try:             # <<<<<<<<<<<<<<
 *                 # libuv will make socket non-blocking
 *                 tr._open(sock.fileno())
 */
      __pyx_L92_except_error:;
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
      goto __pyx_L1_error;
      __pyx_L95_try_end:;
    }

    /* "uvloop/loop.pyx":2078
 *                 raise
 * 
 *             tr._attach_fileobj(sock)             # <<<<<<<<<<<<<<
 * 
 *         if ssl:
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_tr), __pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L17:;

  /* "uvloop/loop.pyx":2080
 *             tr._attach_fileobj(sock)
 * 
 *         if ssl:             # <<<<<<<<<<<<<<
 *             app_transport = protocol._get_app_transport(context)
 *             try:
 */
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_ssl); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 2080, __pyx_L1_error)
  if (__pyx_t_12) {

    /* "uvloop/loop.pyx":2081
 * 
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)             # <<<<<<<<<<<<<<
 *             try:
 *                 await ssl_waiter
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_protocol, __pyx_n_s_get_app_transport); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_context};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2081, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_app_transport = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2082
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      /*try:*/ {

        /* "uvloop/loop.pyx":2083
 *             app_transport = protocol._get_app_transport(context)
 *             try:
 *                 await ssl_waiter             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_ssl_waiter);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_15);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_15;
          __Pyx_XGIVEREF(__pyx_t_16);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_16;
          __Pyx_XGIVEREF(__pyx_t_17);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_17;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 5;
          return __pyx_r;
          __pyx_L108_resume_from_await:;
          __pyx_t_15 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_15);
          __pyx_t_16 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_16);
          __pyx_t_17 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_17);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2083, __pyx_L102_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(0, 2083, __pyx_L102_error)
          }
        }

        /* "uvloop/loop.pyx":2082
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      goto __pyx_L107_try_end;
      __pyx_L102_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":2084
 *             try:
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_6) < 0) __PYX_ERR(0, 2084, __pyx_L104_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "uvloop/loop.pyx":2085
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 app_transport.close()
 */
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_4, __pyx_t_6);
        __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_6 = 0; 
        __PYX_ERR(0, 2085, __pyx_L104_except_error)
      }

      /* "uvloop/loop.pyx":2086
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 app_transport.close()
 *                 raise
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 2086, __pyx_L104_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "uvloop/loop.pyx":2087
 *                 raise
 *             except BaseException:
 *                 app_transport.close()             # <<<<<<<<<<<<<<
 *                 raise
 *             return app_transport, app_protocol
 */
        __pyx_t_29 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_app_transport, __pyx_n_s_close); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 2087, __pyx_L104_except_error)
        __Pyx_GOTREF(__pyx_t_29);
        __pyx_t_30 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_29))) {
          __pyx_t_30 = PyMethod_GET_SELF(__pyx_t_29);
          if (likely(__pyx_t_30)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_29);
            __Pyx_INCREF(__pyx_t_30);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_29, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_30, NULL};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_29, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_30); __pyx_t_30 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2087, __pyx_L104_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/loop.pyx":2088
 *             except BaseException:
 *                 app_transport.close()
 *                 raise             # <<<<<<<<<<<<<<
 *             return app_transport, app_protocol
 *         else:
 */
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_4, __pyx_t_3);
        __pyx_t_6 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; 
        __PYX_ERR(0, 2088, __pyx_L104_except_error)
      }
      goto __pyx_L104_except_error;

      /* "uvloop/loop.pyx":2082
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L104_except_error:;
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      goto __pyx_L1_error;
      __pyx_L107_try_end:;
    }

    /* "uvloop/loop.pyx":2089
 *                 app_transport.close()
 *                 raise
 *             return app_transport, app_protocol             # <<<<<<<<<<<<<<
 *         else:
 *             return tr, protocol
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2089, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_transport);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_transport);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_app_transport)) __PYX_ERR(0, 2089, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_app_protocol)) __PYX_ERR(0, 2089, __pyx_L1_error);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":2080
 *             tr._attach_fileobj(sock)
 * 
 *         if ssl:             # <<<<<<<<<<<<<<
 *             app_transport = protocol._get_app_transport(context)
 *             try:
 */
  }

  /* "uvloop/loop.pyx":2091
 *             return app_transport, app_protocol
 *         else:
 *             return tr, protocol             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2091, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
    __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_tr))) __PYX_ERR(0, 2091, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_protocol)) __PYX_ERR(0, 2091, __pyx_L1_error);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_AddTraceback("create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_63generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2093
 *             return tr, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_62create_unix_server(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_61create_unix_server, "Loop.create_unix_server(self, protocol_factory, path=None, *, backlog=100, sock=None, ssl=None, ssl_handshake_timeout=None, ssl_shutdown_timeout=None, start_serving=True, cleanup_socket=PY313)\nA coroutine which creates a UNIX Domain Socket server.\n\n        The return value is a Server object, which can be used to stop\n        the service.\n\n        path is a str, representing a file systsem path to bind the\n        server socket to.\n\n        sock can optionally be specified in order to use a preexisting\n        socket object.\n\n        backlog is the maximum number of queued connections passed to\n        listen() (defaults to 100).\n\n        ssl can be set to an SSLContext to enable SSL over the\n        accepted connections.\n\n        ssl_handshake_timeout is the time in seconds that an SSL server\n        will wait for completion of the SSL handshake before aborting the\n        connection. Default is 60s.\n\n        ssl_shutdown_timeout is the time in seconds that an SSL server\n        will wait for completion of the SSL shutdown before aborting the\n        connection. Default is 30s.\n\n        If *cleanup_socket* is true then the Unix socket will automatically\n        be removed from the filesystem when the server is closed, unless the\n        socket has been replaced after the server has been created.\n        This defaults to True on Python 3.13 and above, or False otherwise.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_62create_unix_server = {"create_unix_server", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_62create_unix_server, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_61create_unix_server};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_62create_unix_server(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_backlog = 0;
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_ssl = 0;
  PyObject *__pyx_v_ssl_handshake_timeout = 0;
  PyObject *__pyx_v_ssl_shutdown_timeout = 0;
  PyObject *__pyx_v_start_serving = 0;
  PyObject *__pyx_v_cleanup_socket = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_unix_server (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_path,&__pyx_n_s_backlog,&__pyx_n_s_sock,&__pyx_n_s_ssl,&__pyx_n_s_ssl_handshake_timeout,&__pyx_n_s_ssl_shutdown_timeout,&__pyx_n_s_start_serving,&__pyx_n_s_cleanup_socket,0};

    /* "uvloop/loop.pyx":2094
 * 
 *     @cython.iterable_coroutine
 *     async def create_unix_server(self, protocol_factory, path=None,             # <<<<<<<<<<<<<<
 *                                  *, backlog=100, sock=None, ssl=None,
 *                                  ssl_handshake_timeout=None,
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_100));

    /* "uvloop/loop.pyx":2095
 *     @cython.iterable_coroutine
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,             # <<<<<<<<<<<<<<
 *                                  ssl_handshake_timeout=None,
 *                                  ssl_shutdown_timeout=None,
 */
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2096
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 *                                  ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                  ssl_shutdown_timeout=None,
 *                                  start_serving=True, cleanup_socket=PY313):
 */
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2097
 *                                  *, backlog=100, sock=None, ssl=None,
 *                                  ssl_handshake_timeout=None,
 *                                  ssl_shutdown_timeout=None,             # <<<<<<<<<<<<<<
 *                                  start_serving=True, cleanup_socket=PY313):
 *         """A coroutine which creates a UNIX Domain Socket server.
 */
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2098
 *                                  ssl_handshake_timeout=None,
 *                                  ssl_shutdown_timeout=None,
 *                                  start_serving=True, cleanup_socket=PY313):             # <<<<<<<<<<<<<<
 *         """A coroutine which creates a UNIX Domain Socket server.
 * 
 */
    values[7] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    values[8] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__71);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2093, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_path);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2093, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 7)) {
        Py_ssize_t index;
        for (index = 2; index < 9 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2093, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_unix_server") < 0)) __PYX_ERR(0, 2093, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_path = values[1];
    __pyx_v_backlog = values[2];
    __pyx_v_sock = values[3];
    __pyx_v_ssl = values[4];
    __pyx_v_ssl_handshake_timeout = values[5];
    __pyx_v_ssl_shutdown_timeout = values[6];
    __pyx_v_start_serving = values[7];
    __pyx_v_cleanup_socket = values[8];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_unix_server", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 2093, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_61create_unix_server(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_path, __pyx_v_backlog, __pyx_v_sock, __pyx_v_ssl, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout, __pyx_v_start_serving, __pyx_v_cleanup_socket);

  /* "uvloop/loop.pyx":2093
 *             return tr, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_61create_unix_server(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_path, PyObject *__pyx_v_backlog, PyObject *__pyx_v_sock, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout, PyObject *__pyx_v_start_serving, PyObject *__pyx_v_cleanup_socket) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_unix_server", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_11_create_unix_server(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2093, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_path = __pyx_v_path;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_path);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_path);
  __pyx_cur_scope->__pyx_v_backlog = __pyx_v_backlog;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_backlog);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_backlog);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_ssl = __pyx_v_ssl;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl);
  __pyx_cur_scope->__pyx_v_ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __pyx_cur_scope->__pyx_v_start_serving = __pyx_v_start_serving;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_start_serving);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_start_serving);
  __pyx_cur_scope->__pyx_v_cleanup_socket = __pyx_v_cleanup_socket;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_cleanup_socket);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_cleanup_socket);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_63generator5, __pyx_codeobj__72, (PyObject *) __pyx_cur_scope, __pyx_n_s_create_unix_server, __pyx_n_s_Loop_create_unix_server, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2093, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_63generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  int __pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_unix_server", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2093, __pyx_L1_error)

  /* "uvloop/loop.pyx":2131
 *         cdef:
 *             UnixServer pipe
 *             Server server = Server(self)             # <<<<<<<<<<<<<<
 * 
 *         if ssl is not None:
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6uvloop_4loop_Server), ((PyObject *)__pyx_cur_scope->__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_server = ((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2133
 *             Server server = Server(self)
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_ssl != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":2134
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 */
    __pyx_t_1 = __pyx_v_6uvloop_4loop_ssl_SSLContext;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_ssl, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = (!__pyx_t_2);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2135
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')             # <<<<<<<<<<<<<<
 *         else:
 *             if ssl_handshake_timeout is not None:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 2135, __pyx_L1_error)

      /* "uvloop/loop.pyx":2134
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 */
    }

    /* "uvloop/loop.pyx":2133
 *             Server server = Server(self)
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":2137
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_ssl_handshake_timeout != Py_None);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2138
 *         else:
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 2138, __pyx_L1_error)

      /* "uvloop/loop.pyx":2137
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    }

    /* "uvloop/loop.pyx":2140
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout != Py_None);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2141
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 2141, __pyx_L1_error)

      /* "uvloop/loop.pyx":2140
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    }
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":2144
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if path is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
  __pyx_t_3 = (__pyx_cur_scope->__pyx_v_path != Py_None);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":2145
 * 
 *         if path is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'path and sock can not be specified at the same time')
 */
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2146
 *         if path is not None:
 *             if sock is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'path and sock can not be specified at the same time')
 *             orig_path = path
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 2146, __pyx_L1_error)

      /* "uvloop/loop.pyx":2145
 * 
 *         if path is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'path and sock can not be specified at the same time')
 */
    }

    /* "uvloop/loop.pyx":2148
 *                 raise ValueError(
 *                     'path and sock can not be specified at the same time')
 *             orig_path = path             # <<<<<<<<<<<<<<
 * 
 *             path = os_fspath(path)
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_path);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_path);
    __pyx_cur_scope->__pyx_v_orig_path = __pyx_cur_scope->__pyx_v_path;

    /* "uvloop/loop.pyx":2150
 *             orig_path = path
 * 
 *             path = os_fspath(path)             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(path, str):
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_fspath);
    __pyx_t_4 = __pyx_v_6uvloop_4loop_os_fspath; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_path};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_path);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_path, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2152
 *             path = os_fspath(path)
 * 
 *             if isinstance(path, str):             # <<<<<<<<<<<<<<
 *                 path = PyUnicode_EncodeFSDefault(path)
 * 
 */
    __pyx_t_3 = PyUnicode_Check(__pyx_cur_scope->__pyx_v_path); 
    if (__pyx_t_3) {

      /* "uvloop/loop.pyx":2153
 * 
 *             if isinstance(path, str):
 *                 path = PyUnicode_EncodeFSDefault(path)             # <<<<<<<<<<<<<<
 * 
 *             # Check for abstract socket.
 */
      __pyx_t_1 = PyUnicode_EncodeFSDefault(__pyx_cur_scope->__pyx_v_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_path);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_path, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":2152
 *             path = os_fspath(path)
 * 
 *             if isinstance(path, str):             # <<<<<<<<<<<<<<
 *                 path = PyUnicode_EncodeFSDefault(path)
 * 
 */
    }

    /* "uvloop/loop.pyx":2156
 * 
 *             # Check for abstract socket.
 *             if path[0] != 0:             # <<<<<<<<<<<<<<
 *                 try:
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_path, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_1, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "uvloop/loop.pyx":2157
 *             # Check for abstract socket.
 *             if path[0] != 0:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):
 *                         os_remove(path)
 */
      {
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "uvloop/loop.pyx":2158
 *             if path[0] != 0:
 *                 try:
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):             # <<<<<<<<<<<<<<
 *                         os_remove(path)
 *                 except FileNotFoundError:
 */
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_stat);
          __pyx_t_5 = __pyx_v_6uvloop_4loop_os_stat; __pyx_t_10 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_cur_scope->__pyx_v_path};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2158, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2158, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_stat_S_ISSOCK);
          __pyx_t_4 = __pyx_v_6uvloop_4loop_stat_S_ISSOCK; __pyx_t_10 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2158, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2158, __pyx_L12_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_3) {

            /* "uvloop/loop.pyx":2159
 *                 try:
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):
 *                         os_remove(path)             # <<<<<<<<<<<<<<
 *                 except FileNotFoundError:
 *                     pass
 */
            __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_remove);
            __pyx_t_4 = __pyx_v_6uvloop_4loop_os_remove; __pyx_t_5 = NULL;
            __pyx_t_6 = 0;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
                __pyx_t_6 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_path};
              __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2159, __pyx_L12_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "uvloop/loop.pyx":2158
 *             if path[0] != 0:
 *                 try:
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):             # <<<<<<<<<<<<<<
 *                         os_remove(path)
 *                 except FileNotFoundError:
 */
          }

          /* "uvloop/loop.pyx":2157
 *             # Check for abstract socket.
 *             if path[0] != 0:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):
 *                         os_remove(path)
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L17_try_end;
        __pyx_L12_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/loop.pyx":2160
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):
 *                         os_remove(path)
 *                 except FileNotFoundError:             # <<<<<<<<<<<<<<
 *                     pass
 *                 except OSError as err:
 */
        __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_4, &__pyx_t_5);
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_FileNotFoundError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2160, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_ErrRestore(__pyx_t_1, __pyx_t_4, __pyx_t_5);
        __pyx_t_1 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0;
        if (__pyx_t_11) {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L13_exception_handled;
        }

        /* "uvloop/loop.pyx":2162
 *                 except FileNotFoundError:
 *                     pass
 *                 except OSError as err:             # <<<<<<<<<<<<<<
 *                     # Directory may have permissions only to create socket.
 *                     aio_logger.error(
 */
        __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
        if (__pyx_t_11) {
          __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_1) < 0) __PYX_ERR(0, 2162, __pyx_L14_except_error)
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_4);
          __pyx_cur_scope->__pyx_v_err = __pyx_t_4;
          /*try:*/ {

            /* "uvloop/loop.pyx":2164
 *                 except OSError as err:
 *                     # Directory may have permissions only to create socket.
 *                     aio_logger.error(             # <<<<<<<<<<<<<<
 *                         'Unable to check or remove stale UNIX socket %r: %r',
 *                         orig_path, err)
 */
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2164, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "uvloop/loop.pyx":2166
 *                     aio_logger.error(
 *                         'Unable to check or remove stale UNIX socket %r: %r',
 *                         orig_path, err)             # <<<<<<<<<<<<<<
 * 
 *             # We use Python sockets to create a UNIX server socket because
 */
            __pyx_t_13 = NULL;
            __pyx_t_6 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
                __pyx_t_6 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[4] = {__pyx_t_13, __pyx_kp_u_Unable_to_check_or_remove_stale, __pyx_cur_scope->__pyx_v_orig_path, __pyx_cur_scope->__pyx_v_err};
              __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2164, __pyx_L24_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }

          /* "uvloop/loop.pyx":2162
 *                 except FileNotFoundError:
 *                     pass
 *                 except OSError as err:             # <<<<<<<<<<<<<<
 *                     # Directory may have permissions only to create socket.
 *                     aio_logger.error(
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_err);
              __Pyx_DECREF(__pyx_cur_scope->__pyx_v_err); __pyx_cur_scope->__pyx_v_err = 0;
              goto __pyx_L25;
            }
            __pyx_L24_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_assign
              __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __pyx_t_11 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
              {
                __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_err);
                __Pyx_DECREF(__pyx_cur_scope->__pyx_v_err); __pyx_cur_scope->__pyx_v_err = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
              }
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
              __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
              __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
              goto __pyx_L14_except_error;
            }
            __pyx_L25:;
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L13_exception_handled;
        }
        goto __pyx_L14_except_error;

        /* "uvloop/loop.pyx":2157
 *             # Check for abstract socket.
 *             if path[0] != 0:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):
 *                         os_remove(path)
 */
        __pyx_L14_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L13_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L17_try_end:;
      }

      /* "uvloop/loop.pyx":2156
 * 
 *             # Check for abstract socket.
 *             if path[0] != 0:             # <<<<<<<<<<<<<<
 *                 try:
 *                     if stat_S_ISSOCK(os_stat(path).st_mode):
 */
    }

    /* "uvloop/loop.pyx":2172
 *             # they were bound to.  This is different from asyncio, which
 *             # doesn't cleanup the socket path.
 *             sock = socket_socket(uv.AF_UNIX)             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_socket);
    __pyx_t_5 = __pyx_v_6uvloop_4loop_socket_socket; __pyx_t_10 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_sock);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_sock, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2174
 *             sock = socket_socket(uv.AF_UNIX)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 sock.bind(path)
 *             except OSError as exc:
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "uvloop/loop.pyx":2175
 * 
 *             try:
 *                 sock.bind(path)             # <<<<<<<<<<<<<<
 *             except OSError as exc:
 *                 sock.close()
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_bind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2175, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_path};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2175, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/loop.pyx":2174
 *             sock = socket_socket(uv.AF_UNIX)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 sock.bind(path)
 *             except OSError as exc:
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L35_try_end;
      __pyx_L30_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":2176
 *             try:
 *                 sock.bind(path)
 *             except OSError as exc:             # <<<<<<<<<<<<<<
 *                 sock.close()
 *                 if exc.errno == errno.EADDRINUSE:
 */
      __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
      if (__pyx_t_14) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 2176, __pyx_L32_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_5);
        __pyx_cur_scope->__pyx_v_exc = __pyx_t_5;
        /*try:*/ {

          /* "uvloop/loop.pyx":2177
 *                 sock.bind(path)
 *             except OSError as exc:
 *                 sock.close()             # <<<<<<<<<<<<<<
 *                 if exc.errno == errno.EADDRINUSE:
 *                     # Let's improve the error message by adding
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_close); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2177, __pyx_L41_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
            __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2177, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "uvloop/loop.pyx":2178
 *             except OSError as exc:
 *                 sock.close()
 *                 if exc.errno == errno.EADDRINUSE:             # <<<<<<<<<<<<<<
 *                     # Let's improve the error message by adding
 *                     # with what exact address it occurs.
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_exc, __pyx_n_s_errno); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2178, __pyx_L41_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyInt_From___pyx_anon_enum(EADDRINUSE); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2178, __pyx_L41_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = PyObject_RichCompare(__pyx_t_10, __pyx_t_12, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2178, __pyx_L41_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2178, __pyx_L41_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(__pyx_t_3)) {

            /* "uvloop/loop.pyx":2181
 *                     # Let's improve the error message by adding
 *                     # with what exact address it occurs.
 *                     msg = 'Address {!r} is already in use'.format(orig_path)             # <<<<<<<<<<<<<<
 *                     raise OSError(errno.EADDRINUSE, msg) from None
 *                 else:
 */
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Address_r_is_already_in_use, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2181, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_10 = NULL;
            __pyx_t_6 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
                __pyx_t_6 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_cur_scope->__pyx_v_orig_path};
              __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2181, __pyx_L41_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __Pyx_GIVEREF(__pyx_t_13);
            __pyx_cur_scope->__pyx_v_msg = __pyx_t_13;
            __pyx_t_13 = 0;

            /* "uvloop/loop.pyx":2182
 *                     # with what exact address it occurs.
 *                     msg = 'Address {!r} is already in use'.format(orig_path)
 *                     raise OSError(errno.EADDRINUSE, msg) from None             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise
 */
            __pyx_t_13 = __Pyx_PyInt_From___pyx_anon_enum(EADDRINUSE); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2182, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2182, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_13)) __PYX_ERR(0, 2182, __pyx_L41_error);
            __Pyx_INCREF(__pyx_cur_scope->__pyx_v_msg);
            __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_msg);
            if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_cur_scope->__pyx_v_msg)) __PYX_ERR(0, 2182, __pyx_L41_error);
            __pyx_t_13 = 0;
            __pyx_t_13 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_t_12, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2182, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_Raise(__pyx_t_13, 0, 0, Py_None);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __PYX_ERR(0, 2182, __pyx_L41_error)

            /* "uvloop/loop.pyx":2178
 *             except OSError as exc:
 *                 sock.close()
 *                 if exc.errno == errno.EADDRINUSE:             # <<<<<<<<<<<<<<
 *                     # Let's improve the error message by adding
 *                     # with what exact address it occurs.
 */
          }

          /* "uvloop/loop.pyx":2184
 *                     raise OSError(errno.EADDRINUSE, msg) from None
 *                 else:
 *                     raise             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
          /*else*/ {
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_XGIVEREF(__pyx_t_4);
            __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_5, __pyx_t_4);
            __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
            __PYX_ERR(0, 2184, __pyx_L41_error)
          }
        }

        /* "uvloop/loop.pyx":2176
 *             try:
 *                 sock.bind(path)
 *             except OSError as exc:             # <<<<<<<<<<<<<<
 *                 sock.close()
 *                 if exc.errno == errno.EADDRINUSE:
 */
        /*finally:*/ {
          __pyx_L41_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_assign
            __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_16);
            __pyx_t_14 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
            {
              __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_exc);
              __Pyx_DECREF(__pyx_cur_scope->__pyx_v_exc); __pyx_cur_scope->__pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_17, __pyx_t_16);
            }
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ErrRestore(__pyx_t_21, __pyx_t_20, __pyx_t_19);
            __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
            __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_22;
            goto __pyx_L32_except_error;
          }
        }
      }

      /* "uvloop/loop.pyx":2185
 *                 else:
 *                     raise
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(0, 2185, __pyx_L32_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_1);

        /* "uvloop/loop.pyx":2186
 *                     raise
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 sock.close()
 */
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_1);
        __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_5, __pyx_t_1);
        __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_1 = 0; 
        __PYX_ERR(0, 2186, __pyx_L32_except_error)
      }

      /* "uvloop/loop.pyx":2187
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 sock.close()
 *                 raise
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 2187, __pyx_L32_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);

        /* "uvloop/loop.pyx":2188
 *                 raise
 *             except BaseException:
 *                 sock.close()             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_close); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2188, __pyx_L32_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_10 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
          __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2188, __pyx_L32_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "uvloop/loop.pyx":2189
 *             except BaseException:
 *                 sock.close()
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_5, __pyx_t_4);
        __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
        __PYX_ERR(0, 2189, __pyx_L32_except_error)
      }
      goto __pyx_L32_except_error;

      /* "uvloop/loop.pyx":2174
 *             sock = socket_socket(uv.AF_UNIX)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 sock.bind(path)
 *             except OSError as exc:
 */
      __pyx_L32_except_error:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L35_try_end:;
    }

    /* "uvloop/loop.pyx":2144
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if path is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
    goto __pyx_L8;
  }

  /* "uvloop/loop.pyx":2192
 * 
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'path was not specified, and no sock specified')
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_sock == Py_None);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2193
 *         else:
 *             if sock is None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'path was not specified, and no sock specified')
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 2193, __pyx_L1_error)

      /* "uvloop/loop.pyx":2192
 * 
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'path was not specified, and no sock specified')
 */
    }

    /* "uvloop/loop.pyx":2196
 *                     'path was not specified, and no sock specified')
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L54_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_f_6uvloop_4loop__is_sock_stream(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2196, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_23 = (!__pyx_t_2);
    __pyx_t_3 = __pyx_t_23;
    __pyx_L54_bool_binop_done:;
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2199
 *                 raise ValueError(
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 *                     .format(sock))             # <<<<<<<<<<<<<<
 * 
 *             # libuv will set the socket to non-blocking mode, but
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_UNIX_Domain_Stream_Socket_was, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_sock};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "uvloop/loop.pyx":2197
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 *                     .format(sock))
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 2197, __pyx_L1_error)

      /* "uvloop/loop.pyx":2196
 *                     'path was not specified, and no sock specified')
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 */
    }

    /* "uvloop/loop.pyx":2203
 *             # libuv will set the socket to non-blocking mode, but
 *             # we want Python socket object to notice that.
 *             sock.setblocking(False)             # <<<<<<<<<<<<<<
 * 
 *         if cleanup_socket:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, Py_False};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L8:;

  /* "uvloop/loop.pyx":2205
 *             sock.setblocking(False)
 * 
 *         if cleanup_socket:             # <<<<<<<<<<<<<<
 *             path = sock.getsockname()
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_cleanup_socket); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2205, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":2206
 * 
 *         if cleanup_socket:
 *             path = sock.getsockname()             # <<<<<<<<<<<<<<
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 *             if path[0] not in (0, '\x00'):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_getsockname); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_path);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_path, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2208
 *             path = sock.getsockname()
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 *             if path[0] not in (0, '\x00'):             # <<<<<<<<<<<<<<
 *                 try:
 *                     self._unix_server_sockets[sock] = os_stat(path).st_ino
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_path, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_23 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_1, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_23 < 0))) __PYX_ERR(0, 2208, __pyx_L1_error)
    if (__pyx_t_23) {
    } else {
      __pyx_t_3 = __pyx_t_23;
      goto __pyx_L58_bool_binop_done;
    }
    __pyx_t_23 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u__75, Py_NE)); if (unlikely((__pyx_t_23 < 0))) __PYX_ERR(0, 2208, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_23;
    __pyx_L58_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_23 = __pyx_t_3;
    if (__pyx_t_23) {

      /* "uvloop/loop.pyx":2209
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 *             if path[0] not in (0, '\x00'):
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._unix_server_sockets[sock] = os_stat(path).st_ino
 *                 except FileNotFoundError:
 */
      {
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {

          /* "uvloop/loop.pyx":2210
 *             if path[0] not in (0, '\x00'):
 *                 try:
 *                     self._unix_server_sockets[sock] = os_stat(path).st_ino             # <<<<<<<<<<<<<<
 *                 except FileNotFoundError:
 *                     pass
 */
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_stat);
          __pyx_t_5 = __pyx_v_6uvloop_4loop_os_stat; __pyx_t_4 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_path};
            __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2210, __pyx_L60_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_st_ino); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2210, __pyx_L60_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(__pyx_cur_scope->__pyx_v_self->_unix_server_sockets == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 2210, __pyx_L60_error)
          }
          if (unlikely((PyDict_SetItem(__pyx_cur_scope->__pyx_v_self->_unix_server_sockets, __pyx_cur_scope->__pyx_v_sock, __pyx_t_5) < 0))) __PYX_ERR(0, 2210, __pyx_L60_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/loop.pyx":2209
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 *             if path[0] not in (0, '\x00'):
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._unix_server_sockets[sock] = os_stat(path).st_ino
 *                 except FileNotFoundError:
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L65_try_end;
        __pyx_L60_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/loop.pyx":2211
 *                 try:
 *                     self._unix_server_sockets[sock] = os_stat(path).st_ino
 *                 except FileNotFoundError:             # <<<<<<<<<<<<<<
 *                     pass
 * 
 */
        __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_1, &__pyx_t_4);
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_FileNotFoundError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2211, __pyx_L62_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_5, __pyx_t_13);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_ErrRestore(__pyx_t_5, __pyx_t_1, __pyx_t_4);
        __pyx_t_5 = 0; __pyx_t_1 = 0; __pyx_t_4 = 0;
        if (__pyx_t_11) {
          __Pyx_ErrRestore(0,0,0);
          goto __pyx_L61_exception_handled;
        }
        goto __pyx_L62_except_error;

        /* "uvloop/loop.pyx":2209
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 *             if path[0] not in (0, '\x00'):
 *                 try:             # <<<<<<<<<<<<<<
 *                     self._unix_server_sockets[sock] = os_stat(path).st_ino
 *                 except FileNotFoundError:
 */
        __pyx_L62_except_error:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L61_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L65_try_end:;
      }

      /* "uvloop/loop.pyx":2208
 *             path = sock.getsockname()
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 *             if path[0] not in (0, '\x00'):             # <<<<<<<<<<<<<<
 *                 try:
 *                     self._unix_server_sockets[sock] = os_stat(path).st_ino
 */
    }

    /* "uvloop/loop.pyx":2205
 *             sock.setblocking(False)
 * 
 *         if cleanup_socket:             # <<<<<<<<<<<<<<
 *             path = sock.getsockname()
 *             # Check for abstract socket. `str` and `bytes` paths are supported.
 */
  }

  /* "uvloop/loop.pyx":2214
 *                     pass
 * 
 *         pipe = UnixServer.new(             # <<<<<<<<<<<<<<
 *             self, protocol_factory, server, backlog,
 *             ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 */
  __pyx_t_4 = ((PyObject *)__pyx_f_6uvloop_4loop_10UnixServer_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol_factory, __pyx_cur_scope->__pyx_v_server, __pyx_cur_scope->__pyx_v_backlog, __pyx_cur_scope->__pyx_v_ssl, __pyx_cur_scope->__pyx_v_ssl_handshake_timeout, __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_pipe = ((struct __pyx_obj_6uvloop_4loop_UnixServer *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":2218
 *             ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             pipe._open(sock.fileno())
 *         except (KeyboardInterrupt, SystemExit):
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "uvloop/loop.pyx":2219
 * 
 *         try:
 *             pipe._open(sock.fileno())             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2219, __pyx_L66_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2219, __pyx_L66_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2219, __pyx_L66_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_cur_scope->__pyx_v_pipe->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_pipe), __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2219, __pyx_L66_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":2218
 *             ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             pipe._open(sock.fileno())
 *         except (KeyboardInterrupt, SystemExit):
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L71_try_end;
    __pyx_L66_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":2220
 *         try:
 *             pipe._open(sock.fileno())
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException:
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_5) < 0) __PYX_ERR(0, 2220, __pyx_L68_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "uvloop/loop.pyx":2221
 *             pipe._open(sock.fileno())
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException:
 *             pipe._close()
 */
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_1, __pyx_t_5);
      __pyx_t_4 = 0; __pyx_t_1 = 0; __pyx_t_5 = 0; 
      __PYX_ERR(0, 2221, __pyx_L68_except_error)
    }

    /* "uvloop/loop.pyx":2222
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException:             # <<<<<<<<<<<<<<
 *             pipe._close()
 *             sock.close()
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_4) < 0) __PYX_ERR(0, 2222, __pyx_L68_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "uvloop/loop.pyx":2223
 *             raise
 *         except BaseException:
 *             pipe._close()             # <<<<<<<<<<<<<<
 *             sock.close()
 *             raise
 */
      __pyx_t_13 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_cur_scope->__pyx_v_pipe->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_pipe)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2223, __pyx_L68_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "uvloop/loop.pyx":2224
 *         except BaseException:
 *             pipe._close()
 *             sock.close()             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_close); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2224, __pyx_L68_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_10 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
        __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2224, __pyx_L68_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "uvloop/loop.pyx":2225
 *             pipe._close()
 *             sock.close()
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         pipe._attach_fileobj(sock)
 */
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_1, __pyx_t_4);
      __pyx_t_5 = 0; __pyx_t_1 = 0; __pyx_t_4 = 0; 
      __PYX_ERR(0, 2225, __pyx_L68_except_error)
    }
    goto __pyx_L68_except_error;

    /* "uvloop/loop.pyx":2218
 *             ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             pipe._open(sock.fileno())
 *         except (KeyboardInterrupt, SystemExit):
 */
    __pyx_L68_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L71_try_end:;
  }

  /* "uvloop/loop.pyx":2227
 *             raise
 * 
 *         pipe._attach_fileobj(sock)             # <<<<<<<<<<<<<<
 *         server._add_server(pipe)
 * 
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_cur_scope->__pyx_v_pipe->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_pipe), __pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":2228
 * 
 *         pipe._attach_fileobj(sock)
 *         server._add_server(pipe)             # <<<<<<<<<<<<<<
 * 
 *         if start_serving:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_server->__pyx_vtab)->_add_server(__pyx_cur_scope->__pyx_v_server, ((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_cur_scope->__pyx_v_pipe)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":2230
 *         server._add_server(pipe)
 * 
 *         if start_serving:             # <<<<<<<<<<<<<<
 *             server._start_serving()
 * 
 */
  __pyx_t_23 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_start_serving); if (unlikely((__pyx_t_23 < 0))) __PYX_ERR(0, 2230, __pyx_L1_error)
  if (__pyx_t_23) {

    /* "uvloop/loop.pyx":2231
 * 
 *         if start_serving:
 *             server._start_serving()             # <<<<<<<<<<<<<<
 * 
 *         return server
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_server->__pyx_vtab)->_start_serving(__pyx_cur_scope->__pyx_v_server); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":2230
 *         server._add_server(pipe)
 * 
 *         if start_serving:             # <<<<<<<<<<<<<<
 *             server._start_serving()
 * 
 */
  }

  /* "uvloop/loop.pyx":2233
 *             server._start_serving()
 * 
 *         return server             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(((PyObject *)__pyx_cur_scope->__pyx_v_server));
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2093
 *             return tr, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("create_unix_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_66generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2235
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_65create_unix_connection(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_64create_unix_connection, "Loop.create_unix_connection(self, protocol_factory, path=None, *, ssl=None, sock=None, server_hostname=None, ssl_handshake_timeout=None, ssl_shutdown_timeout=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_65create_unix_connection = {"create_unix_connection", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_65create_unix_connection, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_64create_unix_connection};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_65create_unix_connection(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_ssl = 0;
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_server_hostname = 0;
  PyObject *__pyx_v_ssl_handshake_timeout = 0;
  PyObject *__pyx_v_ssl_shutdown_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_unix_connection (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_path,&__pyx_n_s_ssl,&__pyx_n_s_sock,&__pyx_n_s_server_hostname,&__pyx_n_s_ssl_handshake_timeout,&__pyx_n_s_ssl_shutdown_timeout,0};

    /* "uvloop/loop.pyx":2236
 * 
 *     @cython.iterable_coroutine
 *     async def create_unix_connection(self, protocol_factory, path=None, *,             # <<<<<<<<<<<<<<
 *                                      ssl=None, sock=None,
 *                                      server_hostname=None,
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2237
 *     @cython.iterable_coroutine
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,             # <<<<<<<<<<<<<<
 *                                      server_hostname=None,
 *                                      ssl_handshake_timeout=None,
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2238
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 *                                      server_hostname=None,             # <<<<<<<<<<<<<<
 *                                      ssl_handshake_timeout=None,
 *                                      ssl_shutdown_timeout=None):
 */
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2239
 *                                      ssl=None, sock=None,
 *                                      server_hostname=None,
 *                                      ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                      ssl_shutdown_timeout=None):
 * 
 */
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2240
 *                                      server_hostname=None,
 *                                      ssl_handshake_timeout=None,
 *                                      ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2235, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_path);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2235, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 5)) {
        Py_ssize_t index;
        for (index = 2; index < 7 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2235, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_unix_connection") < 0)) __PYX_ERR(0, 2235, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_path = values[1];
    __pyx_v_ssl = values[2];
    __pyx_v_sock = values[3];
    __pyx_v_server_hostname = values[4];
    __pyx_v_ssl_handshake_timeout = values[5];
    __pyx_v_ssl_shutdown_timeout = values[6];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_unix_connection", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 2235, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_64create_unix_connection(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_path, __pyx_v_ssl, __pyx_v_sock, __pyx_v_server_hostname, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout);

  /* "uvloop/loop.pyx":2235
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_64create_unix_connection(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_path, PyObject *__pyx_v_ssl, PyObject *__pyx_v_sock, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_unix_connection", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2235, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_path = __pyx_v_path;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_path);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_path);
  __pyx_cur_scope->__pyx_v_ssl = __pyx_v_ssl;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_server_hostname = __pyx_v_server_hostname;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_hostname);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_hostname);
  __pyx_cur_scope->__pyx_v_ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_66generator6, __pyx_codeobj__76, (PyObject *) __pyx_cur_scope, __pyx_n_s_create_unix_connection, __pyx_n_s_Loop_create_unix_connection, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2235, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_66generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  char *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_unix_connection", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L18_resume_from_await;
    case 2: goto __pyx_L33_resume_from_await;
    case 3: goto __pyx_L45_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2235, __pyx_L1_error)

  /* "uvloop/loop.pyx":2249
 *             object ssl_waiter
 * 
 *         app_protocol = protocol = protocol_factory()             # <<<<<<<<<<<<<<
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_protocol_factory; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_app_protocol = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_protocol = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2250
 * 
 *         app_protocol = protocol = protocol_factory()
 *         ssl_waiter = None             # <<<<<<<<<<<<<<
 *         context = Context_CopyCurrent()
 *         if ssl:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_ssl_waiter = Py_None;

  /* "uvloop/loop.pyx":2251
 *         app_protocol = protocol = protocol_factory()
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 *         if ssl:
 *             if server_hostname is None:
 */
  __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_context = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2252
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()
 *         if ssl:             # <<<<<<<<<<<<<<
 *             if server_hostname is None:
 *                 raise ValueError('You must set server_hostname '
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_ssl); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 2252, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "uvloop/loop.pyx":2253
 *         context = Context_CopyCurrent()
 *         if ssl:
 *             if server_hostname is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('You must set server_hostname '
 *                                  'when using ssl without a host')
 */
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_server_hostname == Py_None);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/loop.pyx":2254
 *         if ssl:
 *             if server_hostname is None:
 *                 raise ValueError('You must set server_hostname '             # <<<<<<<<<<<<<<
 *                                  'when using ssl without a host')
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 2254, __pyx_L1_error)

      /* "uvloop/loop.pyx":2253
 *         context = Context_CopyCurrent()
 *         if ssl:
 *             if server_hostname is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('You must set server_hostname '
 *                                  'when using ssl without a host')
 */
    }

    /* "uvloop/loop.pyx":2257
 *                                  'when using ssl without a host')
 * 
 *             ssl_waiter = self._new_future()             # <<<<<<<<<<<<<<
 *             sslcontext = None if isinstance(ssl, bool) else ssl
 *             protocol = SSLProtocol(
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_ssl_waiter);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_ssl_waiter, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2258
 * 
 *             ssl_waiter = self._new_future()
 *             sslcontext = None if isinstance(ssl, bool) else ssl             # <<<<<<<<<<<<<<
 *             protocol = SSLProtocol(
 *                 self, app_protocol, sslcontext, ssl_waiter,
 */
    __pyx_t_2 = ((PyObject*)&PyBool_Type);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_ssl, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 2258, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(Py_None);
      __pyx_t_1 = Py_None;
    } else {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
      __pyx_t_1 = __pyx_cur_scope->__pyx_v_ssl;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_sslcontext = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2259
 *             ssl_waiter = self._new_future()
 *             sslcontext = None if isinstance(ssl, bool) else ssl
 *             protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self, app_protocol, sslcontext, ssl_waiter,
 *                 False, server_hostname,
 */
    __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(0, 2259, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_app_protocol)) __PYX_ERR(0, 2259, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sslcontext);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sslcontext);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_cur_scope->__pyx_v_sslcontext)) __PYX_ERR(0, 2259, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_waiter);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_waiter);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_cur_scope->__pyx_v_ssl_waiter)) __PYX_ERR(0, 2259, __pyx_L1_error);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, Py_False)) __PYX_ERR(0, 2259, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_server_hostname);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_server_hostname);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_cur_scope->__pyx_v_server_hostname)) __PYX_ERR(0, 2259, __pyx_L1_error);

    /* "uvloop/loop.pyx":2262
 *                 self, app_protocol, sslcontext, ssl_waiter,
 *                 False, server_hostname,
 *                 ssl_handshake_timeout=ssl_handshake_timeout,             # <<<<<<<<<<<<<<
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *         else:
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ssl_handshake_timeout, __pyx_cur_scope->__pyx_v_ssl_handshake_timeout) < 0) __PYX_ERR(0, 2262, __pyx_L1_error)

    /* "uvloop/loop.pyx":2263
 *                 False, server_hostname,
 *                 ssl_handshake_timeout=ssl_handshake_timeout,
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)             # <<<<<<<<<<<<<<
 *         else:
 *             if server_hostname is not None:
 */
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ssl_shutdown_timeout, __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout) < 0) __PYX_ERR(0, 2262, __pyx_L1_error)

    /* "uvloop/loop.pyx":2259
 *             ssl_waiter = self._new_future()
 *             sslcontext = None if isinstance(ssl, bool) else ssl
 *             protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self, app_protocol, sslcontext, ssl_waiter,
 *                 False, server_hostname,
 */
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_protocol);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_protocol, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2252
 *         ssl_waiter = None
 *         context = Context_CopyCurrent()
 *         if ssl:             # <<<<<<<<<<<<<<
 *             if server_hostname is None:
 *                 raise ValueError('You must set server_hostname '
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":2265
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *         else:
 *             if server_hostname is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_server_hostname != Py_None);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/loop.pyx":2266
 *         else:
 *             if server_hostname is not None:
 *                 raise ValueError('server_hostname is only meaningful with ssl')             # <<<<<<<<<<<<<<
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2266, __pyx_L1_error)

      /* "uvloop/loop.pyx":2265
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *         else:
 *             if server_hostname is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:
 */
    }

    /* "uvloop/loop.pyx":2267
 *             if server_hostname is not None:
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_ssl_handshake_timeout != Py_None);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/loop.pyx":2268
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2268, __pyx_L1_error)

      /* "uvloop/loop.pyx":2267
 *             if server_hostname is not None:
 *                 raise ValueError('server_hostname is only meaningful with ssl')
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    }

    /* "uvloop/loop.pyx":2270
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout != Py_None);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/loop.pyx":2271
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2271, __pyx_L1_error)

      /* "uvloop/loop.pyx":2270
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    }
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":2274
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if path is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
  __pyx_t_5 = (__pyx_cur_scope->__pyx_v_path != Py_None);
  if (__pyx_t_5) {

    /* "uvloop/loop.pyx":2275
 * 
 *         if path is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'path and sock can not be specified at the same time')
 */
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/loop.pyx":2276
 *         if path is not None:
 *             if sock is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'path and sock can not be specified at the same time')
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2276, __pyx_L1_error)

      /* "uvloop/loop.pyx":2275
 * 
 *         if path is not None:
 *             if sock is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'path and sock can not be specified at the same time')
 */
    }

    /* "uvloop/loop.pyx":2279
 *                     'path and sock can not be specified at the same time')
 * 
 *             path = os_fspath(path)             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(path, str):
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_fspath);
    __pyx_t_2 = __pyx_v_6uvloop_4loop_os_fspath; __pyx_t_1 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_cur_scope->__pyx_v_path};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_path);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_path, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2281
 *             path = os_fspath(path)
 * 
 *             if isinstance(path, str):             # <<<<<<<<<<<<<<
 *                 path = PyUnicode_EncodeFSDefault(path)
 * 
 */
    __pyx_t_5 = PyUnicode_Check(__pyx_cur_scope->__pyx_v_path); 
    if (__pyx_t_5) {

      /* "uvloop/loop.pyx":2282
 * 
 *             if isinstance(path, str):
 *                 path = PyUnicode_EncodeFSDefault(path)             # <<<<<<<<<<<<<<
 * 
 *             waiter = self._new_future()
 */
      __pyx_t_3 = PyUnicode_EncodeFSDefault(__pyx_cur_scope->__pyx_v_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_path);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_path, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":2281
 *             path = os_fspath(path)
 * 
 *             if isinstance(path, str):             # <<<<<<<<<<<<<<
 *                 path = PyUnicode_EncodeFSDefault(path)
 * 
 */
    }

    /* "uvloop/loop.pyx":2284
 *                 path = PyUnicode_EncodeFSDefault(path)
 * 
 *             waiter = self._new_future()             # <<<<<<<<<<<<<<
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             tr.connect(path)
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_waiter = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2285
 * 
 *             waiter = self._new_future()
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)             # <<<<<<<<<<<<<<
 *             tr.connect(path)
 *             try:
 */
    __pyx_t_3 = ((PyObject *)__pyx_f_6uvloop_4loop_13UnixTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_waiter, __pyx_cur_scope->__pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_v_tr = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2286
 *             waiter = self._new_future()
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             tr.connect(path)             # <<<<<<<<<<<<<<
 *             try:
 *                 await waiter
 */
    __pyx_t_6 = __Pyx_PyObject_AsWritableString(__pyx_cur_scope->__pyx_v_path); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 2286, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->connect(__pyx_cur_scope->__pyx_v_tr, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2287
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             tr.connect(path)
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "uvloop/loop.pyx":2288
 *             tr.connect(path)
 *             try:
 *                 await waiter             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_7);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_7;
          __Pyx_XGIVEREF(__pyx_t_8);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_8;
          __Pyx_XGIVEREF(__pyx_t_9);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_9;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 1;
          return __pyx_r;
          __pyx_L18_resume_from_await:;
          __pyx_t_7 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_7);
          __pyx_t_8 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_8);
          __pyx_t_9 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_9);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2288, __pyx_L12_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(0, 2288, __pyx_L12_error)
          }
        }

        /* "uvloop/loop.pyx":2287
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             tr.connect(path)
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L17_try_end;
      __pyx_L12_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":2289
 *             try:
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 2289, __pyx_L14_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_1);

        /* "uvloop/loop.pyx":2290
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 tr._close()
 */
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_1);
        __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_2, __pyx_t_1);
        __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
        __PYX_ERR(0, 2290, __pyx_L14_except_error)
      }

      /* "uvloop/loop.pyx":2291
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 tr._close()
 *                 raise
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 2291, __pyx_L14_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "uvloop/loop.pyx":2292
 *                 raise
 *             except BaseException:
 *                 tr._close()             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
        __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2292, __pyx_L14_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "uvloop/loop.pyx":2293
 *             except BaseException:
 *                 tr._close()
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_2, __pyx_t_3);
        __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0; 
        __PYX_ERR(0, 2293, __pyx_L14_except_error)
      }
      goto __pyx_L14_except_error;

      /* "uvloop/loop.pyx":2287
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             tr.connect(path)
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L14_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L17_try_end:;
    }

    /* "uvloop/loop.pyx":2274
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if path is not None:             # <<<<<<<<<<<<<<
 *             if sock is not None:
 *                 raise ValueError(
 */
    goto __pyx_L9;
  }

  /* "uvloop/loop.pyx":2296
 * 
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('no path and sock were specified')
 * 
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_sock == Py_None);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/loop.pyx":2297
 *         else:
 *             if sock is None:
 *                 raise ValueError('no path and sock were specified')             # <<<<<<<<<<<<<<
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__77, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2297, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 2297, __pyx_L1_error)

      /* "uvloop/loop.pyx":2296
 * 
 *         else:
 *             if sock is None:             # <<<<<<<<<<<<<<
 *                 raise ValueError('no path and sock were specified')
 * 
 */
    }

    /* "uvloop/loop.pyx":2299
 *                 raise ValueError('no path and sock were specified')
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_12) {
    } else {
      __pyx_t_5 = __pyx_t_12;
      goto __pyx_L25_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_f_6uvloop_4loop__is_sock_stream(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 2299, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = (!__pyx_t_12);
    __pyx_t_5 = __pyx_t_13;
    __pyx_L25_bool_binop_done:;
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/loop.pyx":2302
 *                 raise ValueError(
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 *                     .format(sock))             # <<<<<<<<<<<<<<
 * 
 *             # libuv will set the socket to non-blocking mode, but
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_UNIX_Domain_Stream_Socket_was, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_cur_scope->__pyx_v_sock};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "uvloop/loop.pyx":2300
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 *                     .format(sock))
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 2300, __pyx_L1_error)

      /* "uvloop/loop.pyx":2299
 *                 raise ValueError('no path and sock were specified')
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A UNIX Domain Stream Socket was expected, got {!r}'
 */
    }

    /* "uvloop/loop.pyx":2306
 *             # libuv will set the socket to non-blocking mode, but
 *             # we want Python socket object to notice that.
 *             sock.setblocking(False)             # <<<<<<<<<<<<<<
 * 
 *             waiter = self._new_future()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, Py_False};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2308
 *             sock.setblocking(False)
 * 
 *             waiter = self._new_future()             # <<<<<<<<<<<<<<
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             try:
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_waiter = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2309
 * 
 *             waiter = self._new_future()
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)             # <<<<<<<<<<<<<<
 *             try:
 *                 tr._open(sock.fileno())
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_13UnixTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_waiter, __pyx_cur_scope->__pyx_v_context)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_tr = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2310
 *             waiter = self._new_future()
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             try:             # <<<<<<<<<<<<<<
 *                 tr._open(sock.fileno())
 *                 tr._init_protocol()
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "uvloop/loop.pyx":2311
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             try:
 *                 tr._open(sock.fileno())             # <<<<<<<<<<<<<<
 *                 tr._init_protocol()
 *                 await waiter
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2311, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2311, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2311, __pyx_L27_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_tr), __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2311, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/loop.pyx":2312
 *             try:
 *                 tr._open(sock.fileno())
 *                 tr._init_protocol()             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2312, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/loop.pyx":2313
 *                 tr._open(sock.fileno())
 *                 tr._init_protocol()
 *                 await waiter             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_7);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_7;
          __Pyx_XGIVEREF(__pyx_t_8);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_8;
          __Pyx_XGIVEREF(__pyx_t_9);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_9;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 2;
          return __pyx_r;
          __pyx_L33_resume_from_await:;
          __pyx_t_7 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_7);
          __pyx_t_8 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_8);
          __pyx_t_9 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_9);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2313, __pyx_L27_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(0, 2313, __pyx_L27_error)
          }
        }

        /* "uvloop/loop.pyx":2310
 *             waiter = self._new_future()
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             try:             # <<<<<<<<<<<<<<
 *                 tr._open(sock.fileno())
 *                 tr._init_protocol()
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L32_try_end;
      __pyx_L27_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":2314
 *                 tr._init_protocol()
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 2314, __pyx_L29_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "uvloop/loop.pyx":2315
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 tr._close()
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_2, __pyx_t_3);
        __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0; 
        __PYX_ERR(0, 2315, __pyx_L29_except_error)
      }

      /* "uvloop/loop.pyx":2316
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 tr._close()
 *                 raise
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 2316, __pyx_L29_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_1);

        /* "uvloop/loop.pyx":2317
 *                 raise
 *             except BaseException:
 *                 tr._close()             # <<<<<<<<<<<<<<
 *                 raise
 * 
 */
        __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_tr)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2317, __pyx_L29_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "uvloop/loop.pyx":2318
 *             except BaseException:
 *                 tr._close()
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *             tr._attach_fileobj(sock)
 */
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_1);
        __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_2, __pyx_t_1);
        __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
        __PYX_ERR(0, 2318, __pyx_L29_except_error)
      }
      goto __pyx_L29_except_error;

      /* "uvloop/loop.pyx":2310
 *             waiter = self._new_future()
 *             tr = UnixTransport.new(self, protocol, None, waiter, context)
 *             try:             # <<<<<<<<<<<<<<
 *                 tr._open(sock.fileno())
 *                 tr._init_protocol()
 */
      __pyx_L29_except_error:;
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L32_try_end:;
    }

    /* "uvloop/loop.pyx":2320
 *                 raise
 * 
 *             tr._attach_fileobj(sock)             # <<<<<<<<<<<<<<
 * 
 *         if ssl:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_cur_scope->__pyx_v_tr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_tr), __pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L9:;

  /* "uvloop/loop.pyx":2322
 *             tr._attach_fileobj(sock)
 * 
 *         if ssl:             # <<<<<<<<<<<<<<
 *             app_transport = protocol._get_app_transport(Context_CopyCurrent())
 *             try:
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_ssl); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 2322, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "uvloop/loop.pyx":2323
 * 
 *         if ssl:
 *             app_transport = protocol._get_app_transport(Context_CopyCurrent())             # <<<<<<<<<<<<<<
 *             try:
 *                 await ssl_waiter
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_protocol, __pyx_n_s_get_app_transport); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = Context_CopyCurrent(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_v_app_transport = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2324
 *         if ssl:
 *             app_transport = protocol._get_app_transport(Context_CopyCurrent())
 *             try:             # <<<<<<<<<<<<<<
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "uvloop/loop.pyx":2325
 *             app_transport = protocol._get_app_transport(Context_CopyCurrent())
 *             try:
 *                 await ssl_waiter             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_ssl_waiter);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_7);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_7;
          __Pyx_XGIVEREF(__pyx_t_8);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_8;
          __Pyx_XGIVEREF(__pyx_t_9);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_9;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 3;
          return __pyx_r;
          __pyx_L45_resume_from_await:;
          __pyx_t_7 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_7);
          __pyx_t_8 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_8);
          __pyx_t_9 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_9);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2325, __pyx_L39_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(0, 2325, __pyx_L39_error)
          }
        }

        /* "uvloop/loop.pyx":2324
 *         if ssl:
 *             app_transport = protocol._get_app_transport(Context_CopyCurrent())
 *             try:             # <<<<<<<<<<<<<<
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L44_try_end;
      __pyx_L39_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":2326
 *             try:
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 2326, __pyx_L41_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "uvloop/loop.pyx":2327
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 app_transport.close()
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_2, __pyx_t_3);
        __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0; 
        __PYX_ERR(0, 2327, __pyx_L41_except_error)
      }

      /* "uvloop/loop.pyx":2328
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 app_transport.close()
 *                 raise
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.Loop.create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 2328, __pyx_L41_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_1);

        /* "uvloop/loop.pyx":2329
 *                 raise
 *             except BaseException:
 *                 app_transport.close()             # <<<<<<<<<<<<<<
 *                 raise
 *             return app_transport, app_protocol
 */
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_app_transport, __pyx_n_s_close); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2329, __pyx_L41_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_15 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_14, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_15, NULL};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2329, __pyx_L41_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "uvloop/loop.pyx":2330
 *             except BaseException:
 *                 app_transport.close()
 *                 raise             # <<<<<<<<<<<<<<
 *             return app_transport, app_protocol
 *         else:
 */
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_1);
        __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_2, __pyx_t_1);
        __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
        __PYX_ERR(0, 2330, __pyx_L41_except_error)
      }
      goto __pyx_L41_except_error;

      /* "uvloop/loop.pyx":2324
 *         if ssl:
 *             app_transport = protocol._get_app_transport(Context_CopyCurrent())
 *             try:             # <<<<<<<<<<<<<<
 *                 await ssl_waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L41_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L44_try_end:;
    }

    /* "uvloop/loop.pyx":2331
 *                 app_transport.close()
 *                 raise
 *             return app_transport, app_protocol             # <<<<<<<<<<<<<<
 *         else:
 *             return tr, protocol
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_transport);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_transport);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_v_app_transport)) __PYX_ERR(0, 2331, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_app_protocol)) __PYX_ERR(0, 2331, __pyx_L1_error);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":2322
 *             tr._attach_fileobj(sock)
 * 
 *         if ssl:             # <<<<<<<<<<<<<<
 *             app_transport = protocol._get_app_transport(Context_CopyCurrent())
 *             try:
 */
  }

  /* "uvloop/loop.pyx":2333
 *             return app_transport, app_protocol
 *         else:
 *             return tr, protocol             # <<<<<<<<<<<<<<
 * 
 *     def default_exception_handler(self, context):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
    __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_tr);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_tr))) __PYX_ERR(0, 2333, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_protocol)) __PYX_ERR(0, 2333, __pyx_L1_error);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2235
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("create_unix_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2335
 *             return tr, protocol
 * 
 *     def default_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Default exception handler.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_68default_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_67default_exception_handler, "Loop.default_exception_handler(self, context)\nDefault exception handler.\n\n        This is called when an exception occurs and no exception\n        handler is set, and can be called by a custom exception\n        handler that wants to defer to the default behavior.\n\n        The context parameter has the same meaning as in\n        `call_exception_handler()`.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_68default_exception_handler = {"default_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_68default_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_67default_exception_handler};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_68default_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_context = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("default_exception_handler (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2335, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "default_exception_handler") < 0)) __PYX_ERR(0, 2335, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_context = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("default_exception_handler", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2335, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.default_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_67default_exception_handler(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_context);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_67default_exception_handler(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_v_message = NULL;
  PyObject *__pyx_v_exception = NULL;
  PyObject *__pyx_v_exc_info = NULL;
  PyObject *__pyx_v_log_lines = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_tb = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("default_exception_handler", 1);

  /* "uvloop/loop.pyx":2345
 *         `call_exception_handler()`.
 *         """
 *         message = context.get('message')             # <<<<<<<<<<<<<<
 *         if not message:
 *             message = 'Unhandled exception in event loop'
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_context, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_message};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_message = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2346
 *         """
 *         message = context.get('message')
 *         if not message:             # <<<<<<<<<<<<<<
 *             message = 'Unhandled exception in event loop'
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_message); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 2346, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "uvloop/loop.pyx":2347
 *         message = context.get('message')
 *         if not message:
 *             message = 'Unhandled exception in event loop'             # <<<<<<<<<<<<<<
 * 
 *         exception = context.get('exception')
 */
    __Pyx_INCREF(__pyx_kp_u_Unhandled_exception_in_event_loo);
    __Pyx_DECREF_SET(__pyx_v_message, __pyx_kp_u_Unhandled_exception_in_event_loo);

    /* "uvloop/loop.pyx":2346
 *         """
 *         message = context.get('message')
 *         if not message:             # <<<<<<<<<<<<<<
 *             message = 'Unhandled exception in event loop'
 * 
 */
  }

  /* "uvloop/loop.pyx":2349
 *             message = 'Unhandled exception in event loop'
 * 
 *         exception = context.get('exception')             # <<<<<<<<<<<<<<
 *         if exception is not None:
 *             exc_info = (type(exception), exception, exception.__traceback__)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_context, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_exception};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_exception = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2350
 * 
 *         exception = context.get('exception')
 *         if exception is not None:             # <<<<<<<<<<<<<<
 *             exc_info = (type(exception), exception, exception.__traceback__)
 *         else:
 */
  __pyx_t_6 = (__pyx_v_exception != Py_None);
  if (__pyx_t_6) {

    /* "uvloop/loop.pyx":2351
 *         exception = context.get('exception')
 *         if exception is not None:
 *             exc_info = (type(exception), exception, exception.__traceback__)             # <<<<<<<<<<<<<<
 *         else:
 *             exc_info = False
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exception, __pyx_n_s_traceback); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_exception)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_exception)));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_exception)))) __PYX_ERR(0, 2351, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_exception);
    __Pyx_GIVEREF(__pyx_v_exception);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_exception)) __PYX_ERR(0, 2351, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1)) __PYX_ERR(0, 2351, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_v_exc_info = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2350
 * 
 *         exception = context.get('exception')
 *         if exception is not None:             # <<<<<<<<<<<<<<
 *             exc_info = (type(exception), exception, exception.__traceback__)
 *         else:
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":2353
 *             exc_info = (type(exception), exception, exception.__traceback__)
 *         else:
 *             exc_info = False             # <<<<<<<<<<<<<<
 * 
 *         log_lines = [message]
 */
  /*else*/ {
    __Pyx_INCREF(Py_False);
    __pyx_v_exc_info = Py_False;
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":2355
 *             exc_info = False
 * 
 *         log_lines = [message]             # <<<<<<<<<<<<<<
 *         for key in sorted(context):
 *             if key in {'message', 'exception'}:
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_message);
  __Pyx_GIVEREF(__pyx_v_message);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_message)) __PYX_ERR(0, 2355, __pyx_L1_error);
  __pyx_v_log_lines = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2356
 * 
 *         log_lines = [message]
 *         for key in sorted(context):             # <<<<<<<<<<<<<<
 *             if key in {'message', 'exception'}:
 *                 continue
 */
  __pyx_t_1 = PySequence_List(__pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_7 = PyList_Sort(__pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 2356, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2356, __pyx_L1_error)
      #endif
      if (__pyx_t_8 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_2); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 2356, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2357
 *         log_lines = [message]
 *         for key in sorted(context):
 *             if key in {'message', 'exception'}:             # <<<<<<<<<<<<<<
 *                 continue
 *             value = context[key]
 */
    __Pyx_INCREF(__pyx_v_key);
    __pyx_t_2 = __pyx_v_key;
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_message, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 2357, __pyx_L1_error)
    if (!__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_2, __pyx_n_u_exception, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 2357, __pyx_L1_error)
    __pyx_t_6 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __pyx_t_6;
    if (__pyx_t_5) {

      /* "uvloop/loop.pyx":2358
 *         for key in sorted(context):
 *             if key in {'message', 'exception'}:
 *                 continue             # <<<<<<<<<<<<<<
 *             value = context[key]
 *             if key == 'source_traceback':
 */
      goto __pyx_L5_continue;

      /* "uvloop/loop.pyx":2357
 *         log_lines = [message]
 *         for key in sorted(context):
 *             if key in {'message', 'exception'}:             # <<<<<<<<<<<<<<
 *                 continue
 *             value = context[key]
 */
    }

    /* "uvloop/loop.pyx":2359
 *             if key in {'message', 'exception'}:
 *                 continue
 *             value = context[key]             # <<<<<<<<<<<<<<
 *             if key == 'source_traceback':
 *                 tb = ''.join(tb_format_list(value))
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_context, __pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2360
 *                 continue
 *             value = context[key]
 *             if key == 'source_traceback':             # <<<<<<<<<<<<<<
 *                 tb = ''.join(tb_format_list(value))
 *                 value = 'Object created at (most recent call last):\n'
 */
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_key, __pyx_n_u_source_traceback, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 2360, __pyx_L1_error)
    if (__pyx_t_5) {

      /* "uvloop/loop.pyx":2361
 *             value = context[key]
 *             if key == 'source_traceback':
 *                 tb = ''.join(tb_format_list(value))             # <<<<<<<<<<<<<<
 *                 value = 'Object created at (most recent call last):\n'
 *                 value += tb.rstrip()
 */
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_tb_format_list);
      __pyx_t_3 = __pyx_v_6uvloop_4loop_tb_format_list; __pyx_t_9 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_value};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_3 = PyUnicode_Join(__pyx_kp_u__10, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2361, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF_SET(__pyx_v_tb, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":2362
 *             if key == 'source_traceback':
 *                 tb = ''.join(tb_format_list(value))
 *                 value = 'Object created at (most recent call last):\n'             # <<<<<<<<<<<<<<
 *                 value += tb.rstrip()
 *             else:
 */
      __Pyx_INCREF(__pyx_kp_u_Object_created_at_most_recent_ca);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_kp_u_Object_created_at_most_recent_ca);

      /* "uvloop/loop.pyx":2363
 *                 tb = ''.join(tb_format_list(value))
 *                 value = 'Object created at (most recent call last):\n'
 *                 value += tb.rstrip()             # <<<<<<<<<<<<<<
 *             else:
 *                 try:
 */
      __pyx_t_3 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_rstrip, __pyx_v_tb); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_value, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":2360
 *                 continue
 *             value = context[key]
 *             if key == 'source_traceback':             # <<<<<<<<<<<<<<
 *                 tb = ''.join(tb_format_list(value))
 *                 value = 'Object created at (most recent call last):\n'
 */
      goto __pyx_L10;
    }

    /* "uvloop/loop.pyx":2365
 *                 value += tb.rstrip()
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = repr(value)
 *                 except (KeyboardInterrupt, SystemExit):
 */
    /*else*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {

          /* "uvloop/loop.pyx":2366
 *             else:
 *                 try:
 *                     value = repr(value)             # <<<<<<<<<<<<<<
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 */
          __pyx_t_2 = PyObject_Repr(__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2366, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "uvloop/loop.pyx":2365
 *                 value += tb.rstrip()
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = repr(value)
 *                 except (KeyboardInterrupt, SystemExit):
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L18_try_end;
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "uvloop/loop.pyx":2367
 *                 try:
 *                     value = repr(value)
 *                 except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                     raise
 *                 except BaseException as ex:
 */
        __pyx_t_13 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
        if (__pyx_t_13) {
          __Pyx_AddTraceback("uvloop.loop.Loop.default_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_9) < 0) __PYX_ERR(0, 2367, __pyx_L13_except_error)
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_9);

          /* "uvloop/loop.pyx":2368
 *                     value = repr(value)
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise             # <<<<<<<<<<<<<<
 *                 except BaseException as ex:
 *                     value = ('Exception in __repr__ {!r}; '
 */
          __Pyx_GIVEREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_3);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_3, __pyx_t_9);
          __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_9 = 0; 
          __PYX_ERR(0, 2368, __pyx_L13_except_error)
        }

        /* "uvloop/loop.pyx":2369
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     value = ('Exception in __repr__ {!r}; '
 *                              'value type: {!r}'.format(ex, type(value)))
 */
        __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
        if (__pyx_t_13) {
          __Pyx_AddTraceback("uvloop.loop.Loop.default_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 2369, __pyx_L13_except_error)
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __pyx_v_ex = __pyx_t_3;
          /*try:*/ {

            /* "uvloop/loop.pyx":2371
 *                 except BaseException as ex:
 *                     value = ('Exception in __repr__ {!r}; '
 *                              'value type: {!r}'.format(ex, type(value)))             # <<<<<<<<<<<<<<
 *             log_lines.append('{}: {}'.format(key, value))
 * 
 */
            __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Exception_in___repr___r_value_ty, __pyx_n_s_format); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2371, __pyx_L26_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_16 = NULL;
            __pyx_t_4 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_15))) {
              __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
              if (likely(__pyx_t_16)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                __Pyx_INCREF(__pyx_t_16);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_15, function);
                __pyx_t_4 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_16, __pyx_v_ex, ((PyObject *)Py_TYPE(__pyx_v_value))};
              __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 2371, __pyx_L26_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_14);
            __pyx_t_14 = 0;
          }

          /* "uvloop/loop.pyx":2369
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     value = ('Exception in __repr__ {!r}; '
 *                              'value type: {!r}'.format(ex, type(value)))
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              goto __pyx_L27;
            }
            __pyx_L26_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_23);
              __Pyx_XGOTREF(__pyx_t_24);
              __pyx_t_13 = __pyx_lineno; __pyx_t_17 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_XGIVEREF(__pyx_t_23);
                __Pyx_XGIVEREF(__pyx_t_24);
                __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
              }
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ErrRestore(__pyx_t_19, __pyx_t_20, __pyx_t_21);
              __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
              __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_17; __pyx_filename = __pyx_t_18;
              goto __pyx_L13_except_error;
            }
            __pyx_L27:;
          }
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L12_exception_handled;
        }
        goto __pyx_L13_except_error;

        /* "uvloop/loop.pyx":2365
 *                 value += tb.rstrip()
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     value = repr(value)
 *                 except (KeyboardInterrupt, SystemExit):
 */
        __pyx_L13_except_error:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        goto __pyx_L1_error;
        __pyx_L12_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        __pyx_L18_try_end:;
      }
    }
    __pyx_L10:;

    /* "uvloop/loop.pyx":2372
 *                     value = ('Exception in __repr__ {!r}; '
 *                              'value type: {!r}'.format(ex, type(value)))
 *             log_lines.append('{}: {}'.format(key, value))             # <<<<<<<<<<<<<<
 * 
 *         aio_logger.error('\n'.join(log_lines), exc_info=exc_info)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__78, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_v_key, __pyx_v_value};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_log_lines, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 2372, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2356
 * 
 *         log_lines = [message]
 *         for key in sorted(context):             # <<<<<<<<<<<<<<
 *             if key in {'message', 'exception'}:
 *                 continue
 */
    __pyx_L5_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2374
 *             log_lines.append('{}: {}'.format(key, value))
 * 
 *         aio_logger.error('\n'.join(log_lines), exc_info=exc_info)             # <<<<<<<<<<<<<<
 * 
 *     def get_exception_handler(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__79, __pyx_v_log_lines); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 2374, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_exc_info, __pyx_v_exc_info) < 0) __PYX_ERR(0, 2374, __pyx_L1_error)
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "uvloop/loop.pyx":2335
 *             return tr, protocol
 * 
 *     def default_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Default exception handler.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("uvloop.loop.Loop.default_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_message);
  __Pyx_XDECREF(__pyx_v_exception);
  __Pyx_XDECREF(__pyx_v_exc_info);
  __Pyx_XDECREF(__pyx_v_log_lines);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_tb);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2376
 *         aio_logger.error('\n'.join(log_lines), exc_info=exc_info)
 * 
 *     def get_exception_handler(self):             # <<<<<<<<<<<<<<
 *         """Return an exception handler, or None if the default one is in use.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_70get_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_69get_exception_handler, "Loop.get_exception_handler(self)\nReturn an exception handler, or None if the default one is in use.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_70get_exception_handler = {"get_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_70get_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_69get_exception_handler};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_70get_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_exception_handler (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_exception_handler", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_exception_handler", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_69get_exception_handler(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_69get_exception_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_exception_handler", 1);

  /* "uvloop/loop.pyx":2379
 *         """Return an exception handler, or None if the default one is in use.
 *         """
 *         return self._exception_handler             # <<<<<<<<<<<<<<
 * 
 *     def set_exception_handler(self, handler):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_exception_handler);
  __pyx_r = __pyx_v_self->_exception_handler;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":2376
 *         aio_logger.error('\n'.join(log_lines), exc_info=exc_info)
 * 
 *     def get_exception_handler(self):             # <<<<<<<<<<<<<<
 *         """Return an exception handler, or None if the default one is in use.
 *         """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2381
 *         return self._exception_handler
 * 
 *     def set_exception_handler(self, handler):             # <<<<<<<<<<<<<<
 *         """Set handler as the new event loop exception handler.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_72set_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_71set_exception_handler, "Loop.set_exception_handler(self, handler)\nSet handler as the new event loop exception handler.\n\n        If handler is None, the default exception handler will\n        be set.\n\n        If handler is a callable object, it should have a\n        signature matching '(loop, context)', where 'loop'\n        will be a reference to the active event loop, 'context'\n        will be a dict object (see `call_exception_handler()`\n        documentation for details about context).\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_72set_exception_handler = {"set_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_72set_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_71set_exception_handler};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_72set_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_handler = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_exception_handler (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handler,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_handler)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2381, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_exception_handler") < 0)) __PYX_ERR(0, 2381, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_handler = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_exception_handler", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2381, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.set_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_71set_exception_handler(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_handler);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_71set_exception_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_handler) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_exception_handler", 1);

  /* "uvloop/loop.pyx":2393
 *         documentation for details about context).
 *         """
 *         if handler is not None and not callable(handler):             # <<<<<<<<<<<<<<
 *             raise TypeError('A callable object or None is expected, '
 *                             'got {!r}'.format(handler))
 */
  __pyx_t_2 = (__pyx_v_handler != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyCallable_Check(__pyx_v_handler); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2393, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2395
 *         if handler is not None and not callable(handler):
 *             raise TypeError('A callable object or None is expected, '
 *                             'got {!r}'.format(handler))             # <<<<<<<<<<<<<<
 *         self._exception_handler = handler
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_callable_object_or_None_is_exp, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_handler};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2395, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }

    /* "uvloop/loop.pyx":2394
 *         """
 *         if handler is not None and not callable(handler):
 *             raise TypeError('A callable object or None is expected, '             # <<<<<<<<<<<<<<
 *                             'got {!r}'.format(handler))
 *         self._exception_handler = handler
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 2394, __pyx_L1_error)

    /* "uvloop/loop.pyx":2393
 *         documentation for details about context).
 *         """
 *         if handler is not None and not callable(handler):             # <<<<<<<<<<<<<<
 *             raise TypeError('A callable object or None is expected, '
 *                             'got {!r}'.format(handler))
 */
  }

  /* "uvloop/loop.pyx":2396
 *             raise TypeError('A callable object or None is expected, '
 *                             'got {!r}'.format(handler))
 *         self._exception_handler = handler             # <<<<<<<<<<<<<<
 * 
 *     def call_exception_handler(self, context):
 */
  __Pyx_INCREF(__pyx_v_handler);
  __Pyx_GIVEREF(__pyx_v_handler);
  __Pyx_GOTREF(__pyx_v_self->_exception_handler);
  __Pyx_DECREF(__pyx_v_self->_exception_handler);
  __pyx_v_self->_exception_handler = __pyx_v_handler;

  /* "uvloop/loop.pyx":2381
 *         return self._exception_handler
 * 
 *     def set_exception_handler(self, handler):             # <<<<<<<<<<<<<<
 *         """Set handler as the new event loop exception handler.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.Loop.set_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2398
 *         self._exception_handler = handler
 * 
 *     def call_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Call the current event loop's exception handler.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_74call_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_73call_exception_handler, "Loop.call_exception_handler(self, context)\nCall the current event loop's exception handler.\n\n        The context argument is a dict containing the following keys:\n\n        - 'message': Error message;\n        - 'exception' (optional): Exception object;\n        - 'future' (optional): Future instance;\n        - 'handle' (optional): Handle instance;\n        - 'protocol' (optional): Protocol instance;\n        - 'transport' (optional): Transport instance;\n        - 'socket' (optional): Socket instance.\n\n        New keys maybe introduced in the future.\n\n        Note: do not overload this method in an event loop subclass.\n        For custom exception handling, use the\n        `set_exception_handler()` method.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_74call_exception_handler = {"call_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_74call_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_73call_exception_handler};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_74call_exception_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_context = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("call_exception_handler (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2398, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "call_exception_handler") < 0)) __PYX_ERR(0, 2398, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_context = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("call_exception_handler", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2398, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_73call_exception_handler(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_context);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_73call_exception_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  char const *__pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("call_exception_handler", 1);

  /* "uvloop/loop.pyx":2417
 *         `set_exception_handler()` method.
 *         """
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self._debug_exception_handler_cnt += 1
 * 
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2418
 *         """
 *         if UVLOOP_DEBUG:
 *             self._debug_exception_handler_cnt += 1             # <<<<<<<<<<<<<<
 * 
 *         if self._exception_handler is None:
 */
    __pyx_v_self->_debug_exception_handler_cnt = (__pyx_v_self->_debug_exception_handler_cnt + 1);

    /* "uvloop/loop.pyx":2417
 *         `set_exception_handler()` method.
 *         """
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self._debug_exception_handler_cnt += 1
 * 
 */
  }

  /* "uvloop/loop.pyx":2420
 *             self._debug_exception_handler_cnt += 1
 * 
 *         if self._exception_handler is None:             # <<<<<<<<<<<<<<
 *             try:
 *                 self.default_exception_handler(context)
 */
  __pyx_t_1 = (__pyx_v_self->_exception_handler == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2421
 * 
 *         if self._exception_handler is None:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.default_exception_handler(context)
 *             except (KeyboardInterrupt, SystemExit):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/loop.pyx":2422
 *         if self._exception_handler is None:
 *             try:
 *                 self.default_exception_handler(context)             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_default_exception_handler); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2422, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_context};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2422, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/loop.pyx":2421
 * 
 *         if self._exception_handler is None:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.default_exception_handler(context)
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L10_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":2423
 *             try:
 *                 self.default_exception_handler(context)
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 2423, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "uvloop/loop.pyx":2424
 *                 self.default_exception_handler(context)
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 # Second protection layer for unexpected errors
 */
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; 
        __PYX_ERR(0, 2424, __pyx_L7_except_error)
      }

      /* "uvloop/loop.pyx":2425
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 # Second protection layer for unexpected errors
 *                 # in the default implementation, as well as for subclassed
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 2425, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);

        /* "uvloop/loop.pyx":2429
 *                 # in the default implementation, as well as for subclassed
 *                 # event loops with overloaded "default_exception_handler".
 *                 aio_logger.error('Exception in default exception handler',             # <<<<<<<<<<<<<<
 *                                  exc_info=True)
 *         else:
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2429, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "uvloop/loop.pyx":2430
 *                 # event loops with overloaded "default_exception_handler".
 *                 aio_logger.error('Exception in default exception handler',
 *                                  exc_info=True)             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
        __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2430, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(0, 2430, __pyx_L7_except_error)

        /* "uvloop/loop.pyx":2429
 *                 # in the default implementation, as well as for subclassed
 *                 # event loops with overloaded "default_exception_handler".
 *                 aio_logger.error('Exception in default exception handler',             # <<<<<<<<<<<<<<
 *                                  exc_info=True)
 *         else:
 */
        __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__80, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2429, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;

      /* "uvloop/loop.pyx":2421
 * 
 *         if self._exception_handler is None:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.default_exception_handler(context)
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L7_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L10_try_end:;
    }

    /* "uvloop/loop.pyx":2420
 *             self._debug_exception_handler_cnt += 1
 * 
 *         if self._exception_handler is None:             # <<<<<<<<<<<<<<
 *             try:
 *                 self.default_exception_handler(context)
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":2432
 *                                  exc_info=True)
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 self._exception_handler(self, context)
 *             except (KeyboardInterrupt, SystemExit):
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      /*try:*/ {

        /* "uvloop/loop.pyx":2433
 *         else:
 *             try:
 *                 self._exception_handler(self, context)             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __Pyx_INCREF(__pyx_v_self->_exception_handler);
        __pyx_t_6 = __pyx_v_self->_exception_handler; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_7, ((PyObject *)__pyx_v_self), __pyx_v_context};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2433, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/loop.pyx":2432
 *                                  exc_info=True)
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 self._exception_handler(self, context)
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L20_try_end;
      __pyx_L15_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":2434
 *             try:
 *                 self._exception_handler(self, context)
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException as exc:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 2434, __pyx_L17_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "uvloop/loop.pyx":2435
 *                 self._exception_handler(self, context)
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException as exc:
 *                 # Exception in the user set custom exception handler.
 */
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; 
        __PYX_ERR(0, 2435, __pyx_L17_except_error)
      }

      /* "uvloop/loop.pyx":2436
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 # Exception in the user set custom exception handler.
 *                 try:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 2436, __pyx_L17_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_v_exc = __pyx_t_6;
        /*try:*/ {

          /* "uvloop/loop.pyx":2438
 *             except BaseException as exc:
 *                 # Exception in the user set custom exception handler.
 *                 try:             # <<<<<<<<<<<<<<
 *                     # Let's try default handler.
 *                     self.default_exception_handler({
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            /*try:*/ {

              /* "uvloop/loop.pyx":2440
 *                 try:
 *                     # Let's try default handler.
 *                     self.default_exception_handler({             # <<<<<<<<<<<<<<
 *                         'message': 'Unhandled error in exception handler',
 *                         'exception': exc,
 */
              __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_default_exception_handler); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2440, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_11);

              /* "uvloop/loop.pyx":2441
 *                     # Let's try default handler.
 *                     self.default_exception_handler({
 *                         'message': 'Unhandled error in exception handler',             # <<<<<<<<<<<<<<
 *                         'exception': exc,
 *                         'context': context,
 */
              __pyx_t_10 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2441, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_10);
              if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_message, __pyx_kp_u_Unhandled_error_in_exception_han) < 0) __PYX_ERR(0, 2441, __pyx_L30_error)

              /* "uvloop/loop.pyx":2442
 *                     self.default_exception_handler({
 *                         'message': 'Unhandled error in exception handler',
 *                         'exception': exc,             # <<<<<<<<<<<<<<
 *                         'context': context,
 *                     })
 */
              if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(0, 2441, __pyx_L30_error)

              /* "uvloop/loop.pyx":2443
 *                         'message': 'Unhandled error in exception handler',
 *                         'exception': exc,
 *                         'context': context,             # <<<<<<<<<<<<<<
 *                     })
 *                 except (KeyboardInterrupt, SystemExit):
 */
              if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_context, __pyx_v_context) < 0) __PYX_ERR(0, 2441, __pyx_L30_error)
              __pyx_t_16 = NULL;
              __pyx_t_8 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_11))) {
                __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_11);
                if (likely(__pyx_t_16)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                  __Pyx_INCREF(__pyx_t_16);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_11, function);
                  __pyx_t_8 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_t_10};
                __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2440, __pyx_L30_error)
                __Pyx_GOTREF(__pyx_t_12);
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              }
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

              /* "uvloop/loop.pyx":2438
 *             except BaseException as exc:
 *                 # Exception in the user set custom exception handler.
 *                 try:             # <<<<<<<<<<<<<<
 *                     # Let's try default handler.
 *                     self.default_exception_handler({
 */
            }
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            goto __pyx_L37_try_end;
            __pyx_L30_error:;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "uvloop/loop.pyx":2445
 *                         'context': context,
 *                     })
 *                 except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                     raise
 *                 except BaseException:
 */
            __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
            if (__pyx_t_9) {
              __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10) < 0) __PYX_ERR(0, 2445, __pyx_L32_except_error)
              __Pyx_XGOTREF(__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_11);
              __Pyx_XGOTREF(__pyx_t_10);

              /* "uvloop/loop.pyx":2446
 *                     })
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise             # <<<<<<<<<<<<<<
 *                 except BaseException:
 *                     # Guard 'default_exception_handler' in case it is
 */
              __Pyx_GIVEREF(__pyx_t_12);
              __Pyx_GIVEREF(__pyx_t_11);
              __Pyx_XGIVEREF(__pyx_t_10);
              __Pyx_ErrRestoreWithState(__pyx_t_12, __pyx_t_11, __pyx_t_10);
              __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; 
              __PYX_ERR(0, 2446, __pyx_L32_except_error)
            }

            /* "uvloop/loop.pyx":2447
 *                 except (KeyboardInterrupt, SystemExit):
 *                     raise
 *                 except BaseException:             # <<<<<<<<<<<<<<
 *                     # Guard 'default_exception_handler' in case it is
 *                     # overloaded.
 */
            __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
            if (__pyx_t_9) {
              __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0) __PYX_ERR(0, 2447, __pyx_L32_except_error)
              __Pyx_XGOTREF(__pyx_t_10);
              __Pyx_XGOTREF(__pyx_t_11);
              __Pyx_XGOTREF(__pyx_t_12);

              /* "uvloop/loop.pyx":2450
 *                     # Guard 'default_exception_handler' in case it is
 *                     # overloaded.
 *                     aio_logger.error('Exception in default exception handler '             # <<<<<<<<<<<<<<
 *                                      'while handling an unexpected error '
 *                                      'in custom exception handler',
 */
              __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 2450, __pyx_L32_except_error)
              __Pyx_GOTREF(__pyx_t_16);

              /* "uvloop/loop.pyx":2453
 *                                      'while handling an unexpected error '
 *                                      'in custom exception handler',
 *                                      exc_info=True)             # <<<<<<<<<<<<<<
 * 
 *     def add_reader(self, fileobj, callback, *args):
 */
              __pyx_t_17 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 2453, __pyx_L32_except_error)
              __Pyx_GOTREF(__pyx_t_17);
              if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(0, 2453, __pyx_L32_except_error)

              /* "uvloop/loop.pyx":2450
 *                     # Guard 'default_exception_handler' in case it is
 *                     # overloaded.
 *                     aio_logger.error('Exception in default exception handler '             # <<<<<<<<<<<<<<
 *                                      'while handling an unexpected error '
 *                                      'in custom exception handler',
 */
              __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_tuple__81, __pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 2450, __pyx_L32_except_error)
              __Pyx_GOTREF(__pyx_t_18);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              goto __pyx_L31_exception_handled;
            }
            goto __pyx_L32_except_error;

            /* "uvloop/loop.pyx":2438
 *             except BaseException as exc:
 *                 # Exception in the user set custom exception handler.
 *                 try:             # <<<<<<<<<<<<<<
 *                     # Let's try default handler.
 *                     self.default_exception_handler({
 */
            __pyx_L32_except_error:;
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            goto __pyx_L28_error;
            __pyx_L31_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            __pyx_L37_try_end:;
          }
        }

        /* "uvloop/loop.pyx":2436
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 # Exception in the user set custom exception handler.
 *                 try:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L29;
          }
          __pyx_L28_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_23);
            __pyx_t_9 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_14, __pyx_t_13);
            __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
            __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_20;
            goto __pyx_L17_except_error;
          }
          __pyx_L29:;
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L16_exception_handled;
      }
      goto __pyx_L17_except_error;

      /* "uvloop/loop.pyx":2432
 *                                  exc_info=True)
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 self._exception_handler(self, context)
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L17_except_error:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      goto __pyx_L1_error;
      __pyx_L16_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      __pyx_L20_try_end:;
    }
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":2398
 *         self._exception_handler = handler
 * 
 *     def call_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Call the current event loop's exception handler.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("uvloop.loop.Loop.call_exception_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2455
 *                                      exc_info=True)
 * 
 *     def add_reader(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a reader callback."""
 *         if len(args) == 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_76add_reader(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_75add_reader, "Loop.add_reader(self, fileobj, callback, *args)\nAdd a reader callback.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_76add_reader = {"add_reader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_76add_reader, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_75add_reader};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_76add_reader(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fileobj = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_reader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 2, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fileobj,&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fileobj)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2455, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2455, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("add_reader", 0, 2, 2, 1); __PYX_ERR(0, 2455, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 2) ? kwd_pos_args : 2;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "add_reader") < 0)) __PYX_ERR(0, 2455, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_fileobj = values[0];
    __pyx_v_callback = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_reader", 0, 2, 2, __pyx_nargs); __PYX_ERR(0, 2455, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.add_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_75add_reader(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_fileobj, __pyx_v_callback, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_75add_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, PyObject *__pyx_v_callback, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_reader", 0);
  __Pyx_INCREF(__pyx_v_args);

  /* "uvloop/loop.pyx":2457
 *     def add_reader(self, fileobj, callback, *args):
 *         """Add a reader callback."""
 *         if len(args) == 0:             # <<<<<<<<<<<<<<
 *             args = None
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))
 */
  __pyx_t_1 = __Pyx_PyTuple_GET_SIZE(__pyx_v_args); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2457, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 0);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":2458
 *         """Add a reader callback."""
 *         if len(args) == 0:
 *             args = None             # <<<<<<<<<<<<<<
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_args, ((PyObject*)Py_None));

    /* "uvloop/loop.pyx":2457
 *     def add_reader(self, fileobj, callback, *args):
 *         """Add a reader callback."""
 *         if len(args) == 0:             # <<<<<<<<<<<<<<
 *             args = None
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))
 */
  }

  /* "uvloop/loop.pyx":2459
 *         if len(args) == 0:
 *             args = None
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))             # <<<<<<<<<<<<<<
 * 
 *     def remove_reader(self, fileobj):
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_new_Handle(__pyx_v_self, __pyx_v_callback, __pyx_v_args, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 2459, __pyx_L1_error)
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_add_reader(__pyx_v_self, __pyx_v_fileobj, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":2455
 *                                      exc_info=True)
 * 
 *     def add_reader(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a reader callback."""
 *         if len(args) == 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.add_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2461
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a reader callback."""
 *         self._remove_reader(fileobj)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_78remove_reader(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_77remove_reader, "Loop.remove_reader(self, fileobj)\nRemove a reader callback.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_78remove_reader = {"remove_reader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_78remove_reader, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_77remove_reader};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_78remove_reader(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fileobj = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_reader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fileobj,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fileobj)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2461, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_reader") < 0)) __PYX_ERR(0, 2461, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fileobj = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_reader", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2461, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.remove_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_77remove_reader(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_fileobj);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_77remove_reader(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_reader", 1);

  /* "uvloop/loop.pyx":2463
 *     def remove_reader(self, fileobj):
 *         """Remove a reader callback."""
 *         self._remove_reader(fileobj)             # <<<<<<<<<<<<<<
 * 
 *     def add_writer(self, fileobj, callback, *args):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_reader(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2461
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a reader callback."""
 *         self._remove_reader(fileobj)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop.remove_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2465
 *         self._remove_reader(fileobj)
 * 
 *     def add_writer(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a writer callback.."""
 *         if len(args) == 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_80add_writer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_79add_writer, "Loop.add_writer(self, fileobj, callback, *args)\nAdd a writer callback..");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_80add_writer = {"add_writer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_80add_writer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_79add_writer};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_80add_writer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fileobj = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_writer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 2, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fileobj,&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fileobj)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2465, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2465, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("add_writer", 0, 2, 2, 1); __PYX_ERR(0, 2465, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 2) ? kwd_pos_args : 2;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "add_writer") < 0)) __PYX_ERR(0, 2465, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_fileobj = values[0];
    __pyx_v_callback = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_writer", 0, 2, 2, __pyx_nargs); __PYX_ERR(0, 2465, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.add_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_79add_writer(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_fileobj, __pyx_v_callback, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_79add_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj, PyObject *__pyx_v_callback, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_writer", 0);
  __Pyx_INCREF(__pyx_v_args);

  /* "uvloop/loop.pyx":2467
 *     def add_writer(self, fileobj, callback, *args):
 *         """Add a writer callback.."""
 *         if len(args) == 0:             # <<<<<<<<<<<<<<
 *             args = None
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))
 */
  __pyx_t_1 = __Pyx_PyTuple_GET_SIZE(__pyx_v_args); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2467, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 0);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":2468
 *         """Add a writer callback.."""
 *         if len(args) == 0:
 *             args = None             # <<<<<<<<<<<<<<
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_args, ((PyObject*)Py_None));

    /* "uvloop/loop.pyx":2467
 *     def add_writer(self, fileobj, callback, *args):
 *         """Add a writer callback.."""
 *         if len(args) == 0:             # <<<<<<<<<<<<<<
 *             args = None
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))
 */
  }

  /* "uvloop/loop.pyx":2469
 *         if len(args) == 0:
 *             args = None
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))             # <<<<<<<<<<<<<<
 * 
 *     def remove_writer(self, fileobj):
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_new_Handle(__pyx_v_self, __pyx_v_callback, __pyx_v_args, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 2469, __pyx_L1_error)
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_add_writer(__pyx_v_self, __pyx_v_fileobj, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":2465
 *         self._remove_reader(fileobj)
 * 
 *     def add_writer(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a writer callback.."""
 *         if len(args) == 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.add_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2471
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a writer callback."""
 *         self._remove_writer(fileobj)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_82remove_writer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_81remove_writer, "Loop.remove_writer(self, fileobj)\nRemove a writer callback.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_82remove_writer = {"remove_writer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_82remove_writer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_81remove_writer};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_82remove_writer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fileobj = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_writer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fileobj,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fileobj)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2471, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_writer") < 0)) __PYX_ERR(0, 2471, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fileobj = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_writer", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2471, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.remove_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_81remove_writer(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_fileobj);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_81remove_writer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_fileobj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_writer", 1);

  /* "uvloop/loop.pyx":2473
 *     def remove_writer(self, fileobj):
 *         """Remove a writer callback."""
 *         self._remove_writer(fileobj)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_vtab)->_remove_writer(__pyx_v_self, __pyx_v_fileobj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2471
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a writer callback."""
 *         self._remove_writer(fileobj)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop.remove_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_85generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2475
 *         self._remove_writer(fileobj)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv(self, sock, n):
 *         """Receive data from the socket.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_84sock_recv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_83sock_recv, "Loop.sock_recv(self, sock, n)\nReceive data from the socket.\n\n        The return value is a bytes object representing the data received.\n        The maximum amount of data to be received at once is specified by\n        nbytes.\n\n        This method is a coroutine.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_84sock_recv = {"sock_recv", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_84sock_recv, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_83sock_recv};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_84sock_recv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_n = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recv (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,&__pyx_n_s_n,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2475, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_n)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2475, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_recv", 1, 2, 2, 1); __PYX_ERR(0, 2475, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_recv") < 0)) __PYX_ERR(0, 2475, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_sock = values[0];
    __pyx_v_n = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_recv", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2475, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_83sock_recv(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock, __pyx_v_n);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_83sock_recv(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_n) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_recv", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_13_sock_recv(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2475, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_n = __pyx_v_n;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_n);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_n);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_85generator7, __pyx_codeobj__82, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_recv, __pyx_n_s_Loop_sock_recv, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2475, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_85generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recv", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2475, __pyx_L1_error)

  /* "uvloop/loop.pyx":2488
 *             Handle handle
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  if (__pyx_cur_scope->__pyx_v_self->_debug) {
  } else {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_debug;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_gettimeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2489
 * 
 *         if self._debug and sock.gettimeout() != 0:
 *             raise ValueError("the socket must be non-blocking")             # <<<<<<<<<<<<<<
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2489, __pyx_L1_error)

    /* "uvloop/loop.pyx":2488
 *             Handle handle
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  }

  /* "uvloop/loop.pyx":2491
 *             raise ValueError("the socket must be non-blocking")
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)             # <<<<<<<<<<<<<<
 *         handle = new_MethodHandle3(
 *             self,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SyncSocketReaderFuture); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_sock, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_fut = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2492
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)
 *         handle = new_MethodHandle3(             # <<<<<<<<<<<<<<
 *             self,
 *             "Loop._sock_recv",
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_new_MethodHandle3(__pyx_cur_scope->__pyx_v_self, __pyx_kp_u_Loop__sock_recv, ((__pyx_t_6uvloop_4loop_method3_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_recv), Py_None, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_fut, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 2492, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2500
 *             fut, sock, n)
 * 
 *         self._add_reader(sock, handle)             # <<<<<<<<<<<<<<
 *         return await fut
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_add_reader(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2501
 * 
 *         self._add_reader(sock, handle)
 *         return await fut             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_fut);
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2501, __pyx_L1_error)
    __pyx_t_2 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_2);
  } else {
    __pyx_t_2 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_2) < 0) __PYX_ERR(0, 2501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2475
 *         self._remove_writer(fileobj)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv(self, sock, n):
 *         """Receive data from the socket.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("sock_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_88generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2503
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv_into(self, sock, buf):
 *         """Receive data from the socket.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_87sock_recv_into(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_86sock_recv_into, "Loop.sock_recv_into(self, sock, buf)\nReceive data from the socket.\n\n        The received data is written into *buf* (a writable buffer).\n        The return value is the number of bytes written.\n\n        This method is a coroutine.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_87sock_recv_into = {"sock_recv_into", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_87sock_recv_into, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_86sock_recv_into};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_87sock_recv_into(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_buf = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recv_into (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,&__pyx_n_s_buf,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2503, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_buf)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2503, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_recv_into", 1, 2, 2, 1); __PYX_ERR(0, 2503, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_recv_into") < 0)) __PYX_ERR(0, 2503, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_sock = values[0];
    __pyx_v_buf = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_recv_into", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2503, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recv_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_86sock_recv_into(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock, __pyx_v_buf);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_86sock_recv_into(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_buf) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_recv_into", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2503, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_buf = __pyx_v_buf;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_buf);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_88generator8, __pyx_codeobj__84, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_recv_into, __pyx_n_s_Loop_sock_recv_into, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2503, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recv_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_88generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recv_into", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2503, __pyx_L1_error)

  /* "uvloop/loop.pyx":2515
 *             Handle handle
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  if (__pyx_cur_scope->__pyx_v_self->_debug) {
  } else {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_debug;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_gettimeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2516
 * 
 *         if self._debug and sock.gettimeout() != 0:
 *             raise ValueError("the socket must be non-blocking")             # <<<<<<<<<<<<<<
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2516, __pyx_L1_error)

    /* "uvloop/loop.pyx":2515
 *             Handle handle
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  }

  /* "uvloop/loop.pyx":2518
 *             raise ValueError("the socket must be non-blocking")
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)             # <<<<<<<<<<<<<<
 *         handle = new_MethodHandle3(
 *             self,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SyncSocketReaderFuture); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_sock, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_fut = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2519
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)
 *         handle = new_MethodHandle3(             # <<<<<<<<<<<<<<
 *             self,
 *             "Loop._sock_recv_into",
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_new_MethodHandle3(__pyx_cur_scope->__pyx_v_self, __pyx_kp_u_Loop__sock_recv_into, ((__pyx_t_6uvloop_4loop_method3_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_recv_into), Py_None, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_fut, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_buf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 2519, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2527
 *             fut, sock, buf)
 * 
 *         self._add_reader(sock, handle)             # <<<<<<<<<<<<<<
 *         return await fut
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_add_reader(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2528
 * 
 *         self._add_reader(sock, handle)
 *         return await fut             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_fut);
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2528, __pyx_L1_error)
    __pyx_t_2 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_2);
  } else {
    __pyx_t_2 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_2) < 0) __PYX_ERR(0, 2528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2503
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv_into(self, sock, buf):
 *         """Receive data from the socket.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("sock_recv_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_91generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2530
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendall(self, sock, data):
 *         """Send data to the socket.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_90sock_sendall(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_89sock_sendall, "Loop.sock_sendall(self, sock, data)\nSend data to the socket.\n\n        The socket must be connected to a remote socket. This method continues\n        to send data from data until either all data has been sent or an\n        error occurs. None is returned on success. On error, an exception is\n        raised, and there is no way to determine how much data, if any, was\n        successfully processed by the receiving end of the connection.\n\n        This method is a coroutine.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_90sock_sendall = {"sock_sendall", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_90sock_sendall, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_89sock_sendall};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_90sock_sendall(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_sendall (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2530, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2530, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_sendall", 1, 2, 2, 1); __PYX_ERR(0, 2530, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_sendall") < 0)) __PYX_ERR(0, 2530, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_sock = values[0];
    __pyx_v_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_sendall", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2530, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_89sock_sendall(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock, __pyx_v_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_89sock_sendall(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_data) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_sendall", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_15_sock_sendall(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2530, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_data = __pyx_v_data;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_data);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_data);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_91generator9, __pyx_codeobj__85, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_sendall, __pyx_n_s_Loop_sock_sendall, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2530, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_91generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_sendall", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L20_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2530, __pyx_L1_error)

  /* "uvloop/loop.pyx":2546
 *             ssize_t n
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  if (__pyx_cur_scope->__pyx_v_self->_debug) {
  } else {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_debug;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_gettimeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2547
 * 
 *         if self._debug and sock.gettimeout() != 0:
 *             raise ValueError("the socket must be non-blocking")             # <<<<<<<<<<<<<<
 * 
 *         if not data:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2547, __pyx_L1_error)

    /* "uvloop/loop.pyx":2546
 *             ssize_t n
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  }

  /* "uvloop/loop.pyx":2549
 *             raise ValueError("the socket must be non-blocking")
 * 
 *         if not data:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_data); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2549, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_1);
  if (__pyx_t_6) {

    /* "uvloop/loop.pyx":2550
 * 
 *         if not data:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         socket_inc_io_ref(sock)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":2549
 *             raise ValueError("the socket must be non-blocking")
 * 
 *         if not data:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/loop.pyx":2552
 *             return
 * 
 *         socket_inc_io_ref(sock)             # <<<<<<<<<<<<<<
 *         try:
 *             try:
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_socket_inc_io_ref(__pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2553
 * 
 *         socket_inc_io_ref(sock)
 *         try:             # <<<<<<<<<<<<<<
 *             try:
 *                 n = sock.send(data)
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":2554
 *         socket_inc_io_ref(sock)
 *         try:
 *             try:             # <<<<<<<<<<<<<<
 *                 n = sock.send(data)
 *             except (BlockingIOError, InterruptedError):
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "uvloop/loop.pyx":2555
 *         try:
 *             try:
 *                 n = sock.send(data)             # <<<<<<<<<<<<<<
 *             except (BlockingIOError, InterruptedError):
 *                 pass
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_send); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2555, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_data};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2555, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_10 = PyInt_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2555, __pyx_L11_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_cur_scope->__pyx_v_n = __pyx_t_10;

        /* "uvloop/loop.pyx":2554
 *         socket_inc_io_ref(sock)
 *         try:
 *             try:             # <<<<<<<<<<<<<<
 *                 n = sock.send(data)
 *             except (BlockingIOError, InterruptedError):
 */
      }

      /* "uvloop/loop.pyx":2559
 *                 pass
 *             else:
 *                 if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                     # This can be a partial success, i.e. only part
 *                     # of the data was sent
 */
      /*else:*/ {
        __pyx_t_6 = (UVLOOP_DEBUG != 0);
        if (__pyx_t_6) {

          /* "uvloop/loop.pyx":2562
 *                     # This can be a partial success, i.e. only part
 *                     # of the data was sent
 *                     self._sock_try_write_total += 1             # <<<<<<<<<<<<<<
 * 
 *                 if n == len(data):
 */
          __pyx_cur_scope->__pyx_v_self->_sock_try_write_total = (__pyx_cur_scope->__pyx_v_self->_sock_try_write_total + 1);

          /* "uvloop/loop.pyx":2559
 *                 pass
 *             else:
 *                 if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                     # This can be a partial success, i.e. only part
 *                     # of the data was sent
 */
        }

        /* "uvloop/loop.pyx":2564
 *                     self._sock_try_write_total += 1
 * 
 *                 if n == len(data):             # <<<<<<<<<<<<<<
 *                     return
 *                 if not isinstance(data, memoryview):
 */
        __pyx_t_11 = PyObject_Length(__pyx_cur_scope->__pyx_v_data); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2564, __pyx_L13_except_error)
        __pyx_t_6 = (__pyx_cur_scope->__pyx_v_n == __pyx_t_11);
        if (__pyx_t_6) {

          /* "uvloop/loop.pyx":2565
 * 
 *                 if n == len(data):
 *                     return             # <<<<<<<<<<<<<<
 *                 if not isinstance(data, memoryview):
 *                     data = memoryview(data)
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_r = NULL;
          goto __pyx_L14_except_return;

          /* "uvloop/loop.pyx":2564
 *                     self._sock_try_write_total += 1
 * 
 *                 if n == len(data):             # <<<<<<<<<<<<<<
 *                     return
 *                 if not isinstance(data, memoryview):
 */
        }

        /* "uvloop/loop.pyx":2566
 *                 if n == len(data):
 *                     return
 *                 if not isinstance(data, memoryview):             # <<<<<<<<<<<<<<
 *                     data = memoryview(data)
 *                 data = data[n:]
 */
        __pyx_t_6 = PyMemoryView_Check(__pyx_cur_scope->__pyx_v_data); 
        __pyx_t_1 = (!__pyx_t_6);
        if (__pyx_t_1) {

          /* "uvloop/loop.pyx":2567
 *                     return
 *                 if not isinstance(data, memoryview):
 *                     data = memoryview(data)             # <<<<<<<<<<<<<<
 *                 data = data[n:]
 * 
 */
          __pyx_t_2 = PyMemoryView_FromObject(__pyx_cur_scope->__pyx_v_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2567, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_data);
          __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_data, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_2);
          __pyx_t_2 = 0;

          /* "uvloop/loop.pyx":2566
 *                 if n == len(data):
 *                     return
 *                 if not isinstance(data, memoryview):             # <<<<<<<<<<<<<<
 *                     data = memoryview(data)
 *                 data = data[n:]
 */
        }

        /* "uvloop/loop.pyx":2568
 *                 if not isinstance(data, memoryview):
 *                     data = memoryview(data)
 *                 data = data[n:]             # <<<<<<<<<<<<<<
 * 
 *             fut = _SyncSocketWriterFuture(sock, self)
 */
        __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_data, __pyx_cur_scope->__pyx_v_n, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2568, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_data);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_data, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __pyx_t_2 = 0;
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L16_try_end;
      __pyx_L11_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":2556
 *             try:
 *                 n = sock.send(data)
 *             except (BlockingIOError, InterruptedError):             # <<<<<<<<<<<<<<
 *                 pass
 *             else:
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_BlockingIOError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2556, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_InterruptedError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 2556, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = __Pyx_PyErr_GivenExceptionMatches2(__pyx_t_2, __pyx_t_12, __pyx_t_13);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0;
      if (__pyx_t_14) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L12_exception_handled;
      }
      goto __pyx_L13_except_error;

      /* "uvloop/loop.pyx":2554
 *         socket_inc_io_ref(sock)
 *         try:
 *             try:             # <<<<<<<<<<<<<<
 *                 n = sock.send(data)
 *             except (BlockingIOError, InterruptedError):
 */
      __pyx_L13_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L9_error;
      __pyx_L14_except_return:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L8_return;
      __pyx_L12_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_L16_try_end:;
    }

    /* "uvloop/loop.pyx":2570
 *                 data = data[n:]
 * 
 *             fut = _SyncSocketWriterFuture(sock, self)             # <<<<<<<<<<<<<<
 *             handle = new_MethodHandle3(
 *                 self,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SyncSocketWriterFuture); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2570, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_cur_scope->__pyx_v_sock, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2570, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_fut = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":2571
 * 
 *             fut = _SyncSocketWriterFuture(sock, self)
 *             handle = new_MethodHandle3(             # <<<<<<<<<<<<<<
 *                 self,
 *                 "Loop._sock_sendall",
 */
    __pyx_t_4 = __pyx_f_6uvloop_4loop_new_MethodHandle3(__pyx_cur_scope->__pyx_v_self, __pyx_kp_u_Loop__sock_sendall, ((__pyx_t_6uvloop_4loop_method3_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_sendall), Py_None, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_fut, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_data); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2571, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 2571, __pyx_L9_error)
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_cur_scope->__pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":2579
 *                 fut, sock, data)
 * 
 *             self._add_writer(sock, handle)             # <<<<<<<<<<<<<<
 *             return await fut
 *         finally:
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_add_writer(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_handle); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2579, __pyx_L9_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":2580
 * 
 *             self._add_writer(sock, handle)
 *             return await fut             # <<<<<<<<<<<<<<
 *         finally:
 *             socket_dec_io_ref(sock)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_fut);
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L20_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2580, __pyx_L9_error)
      __pyx_t_4 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_4);
    } else {
      __pyx_t_4 = NULL;
      if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_4) < 0) __PYX_ERR(0, 2580, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_return;
  }

  /* "uvloop/loop.pyx":2582
 *             return await fut
 *         finally:
 *             socket_dec_io_ref(sock)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  /*finally:*/ {
    __pyx_L9_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
      {
        __pyx_t_4 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2582, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
      goto __pyx_L1_error;
      __pyx_L22_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L8_return: {
      __Pyx_PyThreadState_assign
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __pyx_t_20 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_4 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_20;
      __pyx_t_20 = 0;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __Pyx_XGIVEREF(__pyx_t_19);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ErrRestore(__pyx_t_19, __pyx_t_18, __pyx_t_17);
      __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0;
      goto __pyx_L0;
    }
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2530
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendall(self, sock, data):
 *         """Send data to the socket.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("sock_sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_94generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2584
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_accept(self, sock):
 *         """Accept a connection.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_93sock_accept(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_92sock_accept, "Loop.sock_accept(self, sock)\nAccept a connection.\n\n        The socket must be bound to an address and listening for connections.\n        The return value is a pair (conn, address) where conn is a new socket\n        object usable to send and receive data on the connection, and address\n        is the address bound to the socket on the other end of the connection.\n\n        This method is a coroutine.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_93sock_accept = {"sock_accept", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_93sock_accept, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_92sock_accept};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_93sock_accept(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sock = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_accept (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2584, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_accept") < 0)) __PYX_ERR(0, 2584, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_sock = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_accept", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2584, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_92sock_accept(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_92sock_accept(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_accept", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_16_sock_accept(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2584, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_94generator10, __pyx_codeobj__86, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_accept, __pyx_n_s_Loop_sock_accept, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2584, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_94generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_accept", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2584, __pyx_L1_error)

  /* "uvloop/loop.pyx":2598
 *             Handle handle
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  if (__pyx_cur_scope->__pyx_v_self->_debug) {
  } else {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_debug;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_gettimeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2599
 * 
 *         if self._debug and sock.gettimeout() != 0:
 *             raise ValueError("the socket must be non-blocking")             # <<<<<<<<<<<<<<
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2599, __pyx_L1_error)

    /* "uvloop/loop.pyx":2598
 *             Handle handle
 * 
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  }

  /* "uvloop/loop.pyx":2601
 *             raise ValueError("the socket must be non-blocking")
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)             # <<<<<<<<<<<<<<
 *         handle = new_MethodHandle2(
 *             self,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SyncSocketReaderFuture); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_sock, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_fut = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2602
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)
 *         handle = new_MethodHandle2(             # <<<<<<<<<<<<<<
 *             self,
 *             "Loop._sock_accept",
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_new_MethodHandle2(__pyx_cur_scope->__pyx_v_self, __pyx_kp_u_Loop__sock_accept, ((__pyx_t_6uvloop_4loop_method2_t)((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_accept), Py_None, ((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_fut, __pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 2602, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2610
 *             fut, sock)
 * 
 *         self._add_reader(sock, handle)             # <<<<<<<<<<<<<<
 *         return await fut
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_add_reader(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2611
 * 
 *         self._add_reader(sock, handle)
 *         return await fut             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_fut);
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2611, __pyx_L1_error)
    __pyx_t_2 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_2);
  } else {
    __pyx_t_2 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_2) < 0) __PYX_ERR(0, 2611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2584
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_accept(self, sock):
 *         """Accept a connection.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("sock_accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_97generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2613
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_connect(self, sock, address):
 *         """Connect to a remote socket at address.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_96sock_connect(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_95sock_connect, "Loop.sock_connect(self, sock, address)\nConnect to a remote socket at address.\n\n        This method is a coroutine.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_96sock_connect = {"sock_connect", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_96sock_connect, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_95sock_connect};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_96sock_connect(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_address = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_connect (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,&__pyx_n_s_address,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2613, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_address)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2613, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_connect", 1, 2, 2, 1); __PYX_ERR(0, 2613, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_connect") < 0)) __PYX_ERR(0, 2613, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_sock = values[0];
    __pyx_v_address = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_connect", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2613, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_95sock_connect(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock, __pyx_v_address);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_95sock_connect(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_address) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_connect", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_17_sock_connect(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2613, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_address = __pyx_v_address;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_address);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_address);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_97generator11, __pyx_codeobj__87, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_connect, __pyx_n_s_Loop_sock_connect, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_97generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_connect", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L11_resume_from_await;
    case 2: goto __pyx_L15_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2613, __pyx_L1_error)

  /* "uvloop/loop.pyx":2619
 *         This method is a coroutine.
 *         """
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  if (__pyx_cur_scope->__pyx_v_self->_debug) {
  } else {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_debug;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_gettimeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2620
 *         """
 *         if self._debug and sock.gettimeout() != 0:
 *             raise ValueError("the socket must be non-blocking")             # <<<<<<<<<<<<<<
 * 
 *         socket_inc_io_ref(sock)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2620, __pyx_L1_error)

    /* "uvloop/loop.pyx":2619
 *         This method is a coroutine.
 *         """
 *         if self._debug and sock.gettimeout() != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("the socket must be non-blocking")
 * 
 */
  }

  /* "uvloop/loop.pyx":2622
 *             raise ValueError("the socket must be non-blocking")
 * 
 *         socket_inc_io_ref(sock)             # <<<<<<<<<<<<<<
 *         try:
 *             if sock.family == uv.AF_UNIX:
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_socket_inc_io_ref(__pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2623
 * 
 *         socket_inc_io_ref(sock)
 *         try:             # <<<<<<<<<<<<<<
 *             if sock.family == uv.AF_UNIX:
 *                 fut = self._sock_connect(sock, address)
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":2624
 *         socket_inc_io_ref(sock)
 *         try:
 *             if sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *                 fut = self._sock_connect(sock, address)
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2624, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2624, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2624, __pyx_L8_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2624, __pyx_L8_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {

      /* "uvloop/loop.pyx":2625
 *         try:
 *             if sock.family == uv.AF_UNIX:
 *                 fut = self._sock_connect(sock, address)             # <<<<<<<<<<<<<<
 *             else:
 *                 addrs = await self.getaddrinfo(
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_connect(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_address); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2625, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_cur_scope->__pyx_v_fut = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":2624
 *         socket_inc_io_ref(sock)
 *         try:
 *             if sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *                 fut = self._sock_connect(sock, address)
 *             else:
 */
      goto __pyx_L10;
    }

    /* "uvloop/loop.pyx":2627
 *                 fut = self._sock_connect(sock, address)
 *             else:
 *                 addrs = await self.getaddrinfo(             # <<<<<<<<<<<<<<
 *                     *address[:2], family=sock.family)
 * 
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_getaddrinfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2627, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "uvloop/loop.pyx":2628
 *             else:
 *                 addrs = await self.getaddrinfo(
 *                     *address[:2], family=sock.family)             # <<<<<<<<<<<<<<
 * 
 *                 _, _, _, _, address = addrs[0]
 */
      __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_address, 0, 2, NULL, NULL, &__pyx_slice__88, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2628, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":2627
 *                 fut = self._sock_connect(sock, address)
 *             else:
 *                 addrs = await self.getaddrinfo(             # <<<<<<<<<<<<<<
 *                     *address[:2], family=sock.family)
 * 
 */
      __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2627, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/loop.pyx":2628
 *             else:
 *                 addrs = await self.getaddrinfo(
 *                     *address[:2], family=sock.family)             # <<<<<<<<<<<<<<
 * 
 *                 _, _, _, _, address = addrs[0]
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2628, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2628, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_family, __pyx_t_7) < 0) __PYX_ERR(0, 2628, __pyx_L8_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":2627
 *                 fut = self._sock_connect(sock, address)
 *             else:
 *                 addrs = await self.getaddrinfo(             # <<<<<<<<<<<<<<
 *                     *address[:2], family=sock.family)
 * 
 */
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2627, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L11_resume_from_await:;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2627, __pyx_L8_error)
        __pyx_t_7 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_7);
      } else {
        __pyx_t_7 = NULL;
        if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_7) < 0) __PYX_ERR(0, 2627, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_cur_scope->__pyx_v_addrs = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "uvloop/loop.pyx":2630
 *                     *address[:2], family=sock.family)
 * 
 *                 _, _, _, _, address = addrs[0]             # <<<<<<<<<<<<<<
 *                 fut = self._sock_connect(sock, address)
 *             if fut is not None:
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_addrs, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2630, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_7);
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 5)) {
          if (size > 5) __Pyx_RaiseTooManyValuesError(5);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 2630, __pyx_L8_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 3); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 4); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 3); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 4); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_8,&__pyx_t_9};
          for (i=0; i < 5; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 2630, __pyx_L8_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[5] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_8,&__pyx_t_9};
        __pyx_t_10 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2630, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
        for (index=0; index < 5; index++) {
          PyObject* item = __pyx_t_11(__pyx_t_10); if (unlikely(!item)) goto __pyx_L12_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 5) < 0) __PYX_ERR(0, 2630, __pyx_L8_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L13_unpacking_done;
        __pyx_L12_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 2630, __pyx_L8_error)
        __pyx_L13_unpacking_done:;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_v__ = __pyx_t_3;
      __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v__);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v__, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v__);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v__, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v__);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v__, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_address);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_address, __pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_9);
      __pyx_t_9 = 0;

      /* "uvloop/loop.pyx":2631
 * 
 *                 _, _, _, _, address = addrs[0]
 *                 fut = self._sock_connect(sock, address)             # <<<<<<<<<<<<<<
 *             if fut is not None:
 *                 await fut
 */
      __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_connect(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_sock, __pyx_cur_scope->__pyx_v_address); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2631, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_cur_scope->__pyx_v_fut = __pyx_t_7;
      __pyx_t_7 = 0;
    }
    __pyx_L10:;

    /* "uvloop/loop.pyx":2632
 *                 _, _, _, _, address = addrs[0]
 *                 fut = self._sock_connect(sock, address)
 *             if fut is not None:             # <<<<<<<<<<<<<<
 *                 await fut
 *         finally:
 */
    __pyx_t_1 = (__pyx_cur_scope->__pyx_v_fut != Py_None);
    if (__pyx_t_1) {

      /* "uvloop/loop.pyx":2633
 *                 fut = self._sock_connect(sock, address)
 *             if fut is not None:
 *                 await fut             # <<<<<<<<<<<<<<
 *         finally:
 *             socket_dec_io_ref(sock)
 */
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_fut);
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 2;
        return __pyx_r;
        __pyx_L15_resume_from_await:;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2633, __pyx_L8_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 2633, __pyx_L8_error)
        }
      }

      /* "uvloop/loop.pyx":2632
 *                 _, _, _, _, address = addrs[0]
 *                 fut = self._sock_connect(sock, address)
 *             if fut is not None:             # <<<<<<<<<<<<<<
 *                 await fut
 *         finally:
 */
    }
  }

  /* "uvloop/loop.pyx":2635
 *                 await fut
 *         finally:
 *             socket_dec_io_ref(sock)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_7 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2635, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L9;
    }
    __pyx_L8_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_assign
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __pyx_t_12 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
      {
        __pyx_t_7 = __pyx_f_6uvloop_4loop_socket_dec_io_ref(__pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2635, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
      goto __pyx_L1_error;
      __pyx_L17_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L9:;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2613
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_connect(self, sock, address):
 *         """Connect to a remote socket at address.
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("sock_connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_100generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2637
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom(self, sock, bufsize):
 *         raise NotImplementedError
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_99sock_recvfrom(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_98sock_recvfrom, "Loop.sock_recvfrom(self, sock, bufsize)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_99sock_recvfrom = {"sock_recvfrom", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_99sock_recvfrom, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_98sock_recvfrom};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_99sock_recvfrom(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_sock = 0;
  CYTHON_UNUSED PyObject *__pyx_v_bufsize = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recvfrom (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,&__pyx_n_s_bufsize,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2637, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bufsize)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2637, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_recvfrom", 1, 2, 2, 1); __PYX_ERR(0, 2637, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_recvfrom") < 0)) __PYX_ERR(0, 2637, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_sock = values[0];
    __pyx_v_bufsize = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_recvfrom", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2637, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recvfrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_98sock_recvfrom(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock, __pyx_v_bufsize);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_98sock_recvfrom(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_sock, CYTHON_UNUSED PyObject *__pyx_v_bufsize) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_recvfrom", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2637, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_bufsize = __pyx_v_bufsize;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_bufsize);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_bufsize);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_100generator12, __pyx_codeobj__89, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_recvfrom, __pyx_n_s_Loop_sock_recvfrom, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2637, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recvfrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_100generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recvfrom", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2637, __pyx_L1_error)

  /* "uvloop/loop.pyx":2639
 *     @cython.iterable_coroutine
 *     async def sock_recvfrom(self, sock, bufsize):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 2639, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2637
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom(self, sock, bufsize):
 *         raise NotImplementedError
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_AddTraceback("sock_recvfrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_103generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2641
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom_into(self, sock, buf, nbytes=0):
 *         raise NotImplementedError
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_102sock_recvfrom_into(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_101sock_recvfrom_into, "Loop.sock_recvfrom_into(self, sock, buf, nbytes=0)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_102sock_recvfrom_into = {"sock_recvfrom_into", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_102sock_recvfrom_into, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_101sock_recvfrom_into};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_102sock_recvfrom_into(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_sock = 0;
  CYTHON_UNUSED PyObject *__pyx_v_buf = 0;
  CYTHON_UNUSED PyObject *__pyx_v_nbytes = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recvfrom_into (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,&__pyx_n_s_buf,&__pyx_n_s_nbytes,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2641, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_buf)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2641, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_recvfrom_into", 0, 2, 3, 1); __PYX_ERR(0, 2641, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_nbytes);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2641, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_recvfrom_into") < 0)) __PYX_ERR(0, 2641, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_sock = values[0];
    __pyx_v_buf = values[1];
    __pyx_v_nbytes = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_recvfrom_into", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 2641, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recvfrom_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_101sock_recvfrom_into(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock, __pyx_v_buf, __pyx_v_nbytes);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_101sock_recvfrom_into(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_sock, CYTHON_UNUSED PyObject *__pyx_v_buf, CYTHON_UNUSED PyObject *__pyx_v_nbytes) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_recvfrom_into", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2641, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_buf = __pyx_v_buf;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_buf);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_buf);
  __pyx_cur_scope->__pyx_v_nbytes = __pyx_v_nbytes;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_nbytes);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_nbytes);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_103generator13, __pyx_codeobj__90, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_recvfrom_into, __pyx_n_s_Loop_sock_recvfrom_into, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2641, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_recvfrom_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_103generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_recvfrom_into", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2641, __pyx_L1_error)

  /* "uvloop/loop.pyx":2643
 *     @cython.iterable_coroutine
 *     async def sock_recvfrom_into(self, sock, buf, nbytes=0):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 2643, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2641
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom_into(self, sock, buf, nbytes=0):
 *         raise NotImplementedError
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_AddTraceback("sock_recvfrom_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_106generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2645
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendto(self, sock, data, address):
 *         raise NotImplementedError
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_105sock_sendto(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_104sock_sendto, "Loop.sock_sendto(self, sock, data, address)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_105sock_sendto = {"sock_sendto", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_105sock_sendto, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_104sock_sendto};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_105sock_sendto(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_sock = 0;
  CYTHON_UNUSED PyObject *__pyx_v_data = 0;
  CYTHON_UNUSED PyObject *__pyx_v_address = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_sendto (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sock,&__pyx_n_s_data,&__pyx_n_s_address,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2645, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2645, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_sendto", 1, 3, 3, 1); __PYX_ERR(0, 2645, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_address)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2645, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sock_sendto", 1, 3, 3, 2); __PYX_ERR(0, 2645, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sock_sendto") < 0)) __PYX_ERR(0, 2645, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_sock = values[0];
    __pyx_v_data = values[1];
    __pyx_v_address = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sock_sendto", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 2645, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_sendto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_104sock_sendto(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sock, __pyx_v_data, __pyx_v_address);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_104sock_sendto(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_sock, CYTHON_UNUSED PyObject *__pyx_v_data, CYTHON_UNUSED PyObject *__pyx_v_address) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sock_sendto", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_20_sock_sendto(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2645, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_data = __pyx_v_data;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_data);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_data);
  __pyx_cur_scope->__pyx_v_address = __pyx_v_address;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_address);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_address);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_106generator14, __pyx_codeobj__91, (PyObject *) __pyx_cur_scope, __pyx_n_s_sock_sendto, __pyx_n_s_Loop_sock_sendto, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2645, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.sock_sendto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_106generator14(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sock_sendto", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2645, __pyx_L1_error)

  /* "uvloop/loop.pyx":2647
 *     @cython.iterable_coroutine
 *     async def sock_sendto(self, sock, data, address):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 2647, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2645
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendto(self, sock, data, address):
 *         raise NotImplementedError
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_AddTraceback("sock_sendto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_109generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2649
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_108connect_accepted_socket(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_107connect_accepted_socket, "Loop.connect_accepted_socket(self, protocol_factory, sock, *, ssl=None, ssl_handshake_timeout=None, ssl_shutdown_timeout=None)\nHandle an accepted connection.\n\n        This is used by servers that accept connections outside of\n        asyncio but that use asyncio to handle connections.\n\n        This method is a coroutine.  When completed, the coroutine\n        returns a (transport, protocol) pair.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_108connect_accepted_socket = {"connect_accepted_socket", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_108connect_accepted_socket, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_107connect_accepted_socket};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_108connect_accepted_socket(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_ssl = 0;
  PyObject *__pyx_v_ssl_handshake_timeout = 0;
  PyObject *__pyx_v_ssl_shutdown_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_accepted_socket (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_sock,&__pyx_n_s_ssl,&__pyx_n_s_ssl_handshake_timeout,&__pyx_n_s_ssl_shutdown_timeout,0};

    /* "uvloop/loop.pyx":2651
 *     @cython.iterable_coroutine
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,             # <<<<<<<<<<<<<<
 *                                       ssl_handshake_timeout=None,
 *                                       ssl_shutdown_timeout=None):
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2652
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 *                                       ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                       ssl_shutdown_timeout=None):
 *         """Handle an accepted connection.
 */
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2653
 *                                       ssl=None,
 *                                       ssl_handshake_timeout=None,
 *                                       ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 *         """Handle an accepted connection.
 * 
 */
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2649, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2649, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("connect_accepted_socket", 1, 2, 2, 1); __PYX_ERR(0, 2649, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 3)) {
        Py_ssize_t index;
        for (index = 2; index < 5 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2649, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connect_accepted_socket") < 0)) __PYX_ERR(0, 2649, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_sock = values[1];
    __pyx_v_ssl = values[2];
    __pyx_v_ssl_handshake_timeout = values[3];
    __pyx_v_ssl_shutdown_timeout = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connect_accepted_socket", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2649, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.connect_accepted_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_107connect_accepted_socket(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_sock, __pyx_v_ssl, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout);

  /* "uvloop/loop.pyx":2649
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_107connect_accepted_socket(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_sock, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect_accepted_socket", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2649, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  __pyx_cur_scope->__pyx_v_ssl = __pyx_v_ssl;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl);
  __pyx_cur_scope->__pyx_v_ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_handshake_timeout);
  __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_109generator15, __pyx_codeobj__92, (PyObject *) __pyx_cur_scope, __pyx_n_s_connect_accepted_socket, __pyx_n_s_Loop_connect_accepted_socket, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2649, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.connect_accepted_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_109generator15(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_accepted_socket", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L21_resume_from_await;
    case 2: goto __pyx_L32_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2649, __pyx_L1_error)

  /* "uvloop/loop.pyx":2664
 * 
 *         cdef:
 *             UVStream transport = None             # <<<<<<<<<<<<<<
 * 
 *         if ssl is not None:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_transport = ((struct __pyx_obj_6uvloop_4loop_UVStream *)Py_None);

  /* "uvloop/loop.pyx":2666
 *             UVStream transport = None
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_ssl != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2667
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 */
    __pyx_t_2 = __pyx_v_6uvloop_4loop_ssl_SSLContext;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_1 = PyObject_IsInstance(__pyx_cur_scope->__pyx_v_ssl, __pyx_t_2); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 2667, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (!__pyx_t_1);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2668
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')             # <<<<<<<<<<<<<<
 *         else:
 *             if ssl_handshake_timeout is not None:
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2668, __pyx_L1_error)

      /* "uvloop/loop.pyx":2667
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 */
    }

    /* "uvloop/loop.pyx":2666
 *             UVStream transport = None
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":2670
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_ssl_handshake_timeout != Py_None);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2671
 *         else:
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2671, __pyx_L1_error)

      /* "uvloop/loop.pyx":2670
 *                 raise TypeError('ssl argument must be an SSLContext or None')
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    }

    /* "uvloop/loop.pyx":2673
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    __pyx_t_3 = (__pyx_cur_scope->__pyx_v_ssl_shutdown_timeout != Py_None);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/loop.pyx":2674
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 2674, __pyx_L1_error)

      /* "uvloop/loop.pyx":2673
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    }
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":2677
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'A Stream Socket was expected, got {!r}'.format(sock))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_f_6uvloop_4loop__is_sock_stream(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2677, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (!__pyx_t_3);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2679
 *         if not _is_sock_stream(sock.type):
 *             raise ValueError(
 *                 'A Stream Socket was expected, got {!r}'.format(sock))             # <<<<<<<<<<<<<<
 * 
 *         app_protocol = protocol_factory()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_Stream_Socket_was_expected_got, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_sock};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "uvloop/loop.pyx":2678
 * 
 *         if not _is_sock_stream(sock.type):
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'A Stream Socket was expected, got {!r}'.format(sock))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2678, __pyx_L1_error)

    /* "uvloop/loop.pyx":2677
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         if not _is_sock_stream(sock.type):             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'A Stream Socket was expected, got {!r}'.format(sock))
 */
  }

  /* "uvloop/loop.pyx":2681
 *                 'A Stream Socket was expected, got {!r}'.format(sock))
 * 
 *         app_protocol = protocol_factory()             # <<<<<<<<<<<<<<
 *         waiter = self._new_future()
 *         transport_waiter = None
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_protocol_factory; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_app_protocol = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2682
 * 
 *         app_protocol = protocol_factory()
 *         waiter = self._new_future()             # <<<<<<<<<<<<<<
 *         transport_waiter = None
 *         context = Context_CopyCurrent()
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_waiter = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2683
 *         app_protocol = protocol_factory()
 *         waiter = self._new_future()
 *         transport_waiter = None             # <<<<<<<<<<<<<<
 *         context = Context_CopyCurrent()
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_transport_waiter = Py_None;

  /* "uvloop/loop.pyx":2684
 *         waiter = self._new_future()
 *         transport_waiter = None
 *         context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 * 
 *         if ssl is None:
 */
  __pyx_t_2 = Context_CopyCurrent(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_context = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2686
 *         context = Context_CopyCurrent()
 * 
 *         if ssl is None:             # <<<<<<<<<<<<<<
 *             protocol = app_protocol
 *             transport_waiter = waiter
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_ssl == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2687
 * 
 *         if ssl is None:
 *             protocol = app_protocol             # <<<<<<<<<<<<<<
 *             transport_waiter = waiter
 *         else:
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_protocol);
    __pyx_cur_scope->__pyx_v_protocol = __pyx_cur_scope->__pyx_v_app_protocol;

    /* "uvloop/loop.pyx":2688
 *         if ssl is None:
 *             protocol = app_protocol
 *             transport_waiter = waiter             # <<<<<<<<<<<<<<
 *         else:
 *             protocol = SSLProtocol(
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_waiter);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_transport_waiter);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_transport_waiter, __pyx_cur_scope->__pyx_v_waiter);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_waiter);

    /* "uvloop/loop.pyx":2686
 *         context = Context_CopyCurrent()
 * 
 *         if ssl is None:             # <<<<<<<<<<<<<<
 *             protocol = app_protocol
 *             transport_waiter = waiter
 */
    goto __pyx_L9;
  }

  /* "uvloop/loop.pyx":2690
 *             transport_waiter = waiter
 *         else:
 *             protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self, app_protocol, ssl, waiter,
 *                 server_side=True,
 */
  /*else*/ {

    /* "uvloop/loop.pyx":2691
 *         else:
 *             protocol = SSLProtocol(
 *                 self, app_protocol, ssl, waiter,             # <<<<<<<<<<<<<<
 *                 server_side=True,
 *                 server_hostname=None,
 */
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_self))) __PYX_ERR(0, 2690, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_app_protocol)) __PYX_ERR(0, 2690, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ssl);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ssl);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_cur_scope->__pyx_v_ssl)) __PYX_ERR(0, 2690, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_waiter);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_waiter);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_cur_scope->__pyx_v_waiter)) __PYX_ERR(0, 2690, __pyx_L1_error);

    /* "uvloop/loop.pyx":2692
 *             protocol = SSLProtocol(
 *                 self, app_protocol, ssl, waiter,
 *                 server_side=True,             # <<<<<<<<<<<<<<
 *                 server_hostname=None,
 *                 ssl_handshake_timeout=ssl_handshake_timeout,
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_server_side, Py_True) < 0) __PYX_ERR(0, 2692, __pyx_L1_error)

    /* "uvloop/loop.pyx":2693
 *                 self, app_protocol, ssl, waiter,
 *                 server_side=True,
 *                 server_hostname=None,             # <<<<<<<<<<<<<<
 *                 ssl_handshake_timeout=ssl_handshake_timeout,
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_server_hostname, Py_None) < 0) __PYX_ERR(0, 2692, __pyx_L1_error)

    /* "uvloop/loop.pyx":2694
 *                 server_side=True,
 *                 server_hostname=None,
 *                 ssl_handshake_timeout=ssl_handshake_timeout,             # <<<<<<<<<<<<<<
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *             transport_waiter = None
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_ssl_handshake_timeout, __pyx_cur_scope->__pyx_v_ssl_handshake_timeout) < 0) __PYX_ERR(0, 2692, __pyx_L1_error)

    /* "uvloop/loop.pyx":2695
 *                 server_hostname=None,
 *                 ssl_handshake_timeout=ssl_handshake_timeout,
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)             # <<<<<<<<<<<<<<
 *             transport_waiter = None
 * 
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_ssl_shutdown_timeout, __pyx_cur_scope->__pyx_v_ssl_shutdown_timeout) < 0) __PYX_ERR(0, 2692, __pyx_L1_error)

    /* "uvloop/loop.pyx":2690
 *             transport_waiter = waiter
 *         else:
 *             protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self, app_protocol, ssl, waiter,
 *                 server_side=True,
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol), __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_cur_scope->__pyx_v_protocol = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":2696
 *                 ssl_handshake_timeout=ssl_handshake_timeout,
 *                 ssl_shutdown_timeout=ssl_shutdown_timeout)
 *             transport_waiter = None             # <<<<<<<<<<<<<<
 * 
 *         if sock.family == uv.AF_UNIX:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_transport_waiter);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_transport_waiter, Py_None);
    __Pyx_GIVEREF(Py_None);
  }
  __pyx_L9:;

  /* "uvloop/loop.pyx":2698
 *             transport_waiter = None
 * 
 *         if sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *             transport = <UVStream>UnixTransport.new(
 *                 self, protocol, None, transport_waiter, context)
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_5, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2698, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2698, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2699
 * 
 *         if sock.family == uv.AF_UNIX:
 *             transport = <UVStream>UnixTransport.new(             # <<<<<<<<<<<<<<
 *                 self, protocol, None, transport_waiter, context)
 *         elif sock.family in (uv.AF_INET, uv.AF_INET6):
 */
    __pyx_t_2 = ((PyObject *)__pyx_f_6uvloop_4loop_13UnixTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_transport_waiter, __pyx_cur_scope->__pyx_v_context)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_transport);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_transport, ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_4));
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":2698
 *             transport_waiter = None
 * 
 *         if sock.family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *             transport = <UVStream>UnixTransport.new(
 *                 self, protocol, None, transport_waiter, context)
 */
    goto __pyx_L10;
  }

  /* "uvloop/loop.pyx":2701
 *             transport = <UVStream>UnixTransport.new(
 *                 self, protocol, None, transport_waiter, context)
 *         elif sock.family in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *             transport = <UVStream>TCPTransport.new(
 *                 self, protocol, None, transport_waiter, context)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyInt_From_int(AF_INET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyInt_From_int(AF_INET6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L11_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __pyx_t_1;
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":2702
 *                 self, protocol, None, transport_waiter, context)
 *         elif sock.family in (uv.AF_INET, uv.AF_INET6):
 *             transport = <UVStream>TCPTransport.new(             # <<<<<<<<<<<<<<
 *                 self, protocol, None, transport_waiter, context)
 * 
 */
    __pyx_t_4 = ((PyObject *)__pyx_f_6uvloop_4loop_12TCPTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_transport_waiter, __pyx_cur_scope->__pyx_v_context)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __pyx_t_4;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_transport);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_transport, ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_2));
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2701
 *             transport = <UVStream>UnixTransport.new(
 *                 self, protocol, None, transport_waiter, context)
 *         elif sock.family in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *             transport = <UVStream>TCPTransport.new(
 *                 self, protocol, None, transport_waiter, context)
 */
  }
  __pyx_L10:;

  /* "uvloop/loop.pyx":2705
 *                 self, protocol, None, transport_waiter, context)
 * 
 *         if transport is None:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'invalid socket family, expected AF_UNIX, AF_INET or AF_INET6')
 */
  __pyx_t_3 = (((PyObject *)__pyx_cur_scope->__pyx_v_transport) == Py_None);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":2706
 * 
 *         if transport is None:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'invalid socket family, expected AF_UNIX, AF_INET or AF_INET6')
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__93, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2706, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2706, __pyx_L1_error)

    /* "uvloop/loop.pyx":2705
 *                 self, protocol, None, transport_waiter, context)
 * 
 *         if transport is None:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'invalid socket family, expected AF_UNIX, AF_INET or AF_INET6')
 */
  }

  /* "uvloop/loop.pyx":2709
 *                 'invalid socket family, expected AF_UNIX, AF_INET or AF_INET6')
 * 
 *         transport._open(sock.fileno())             # <<<<<<<<<<<<<<
 *         transport._init_protocol()
 *         transport._attach_fileobj(sock)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_cur_scope->__pyx_v_transport->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_transport), __pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2710
 * 
 *         transport._open(sock.fileno())
 *         transport._init_protocol()             # <<<<<<<<<<<<<<
 *         transport._attach_fileobj(sock)
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_cur_scope->__pyx_v_transport->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_transport)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2711
 *         transport._open(sock.fileno())
 *         transport._init_protocol()
 *         transport._attach_fileobj(sock)             # <<<<<<<<<<<<<<
 * 
 *         if ssl:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_cur_scope->__pyx_v_transport->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_transport), __pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2713
 *         transport._attach_fileobj(sock)
 * 
 *         if ssl:             # <<<<<<<<<<<<<<
 *             app_transport = protocol._get_app_transport(context)
 *             try:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_ssl); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2713, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":2714
 * 
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)             # <<<<<<<<<<<<<<
 *             try:
 *                 await waiter
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_protocol, __pyx_n_s_get_app_transport); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_context};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_v_app_transport = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2715
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
    {
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "uvloop/loop.pyx":2716
 *             app_transport = protocol._get_app_transport(context)
 *             try:
 *                 await waiter             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_8);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_8;
          __Pyx_XGIVEREF(__pyx_t_9);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_9;
          __Pyx_XGIVEREF(__pyx_t_10);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_10;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 1;
          return __pyx_r;
          __pyx_L21_resume_from_await:;
          __pyx_t_8 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_8);
          __pyx_t_9 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_9);
          __pyx_t_10 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_10);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2716, __pyx_L15_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(0, 2716, __pyx_L15_error)
          }
        }

        /* "uvloop/loop.pyx":2715
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L20_try_end;
      __pyx_L15_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":2717
 *             try:
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("uvloop.loop.Loop.connect_accepted_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 2717, __pyx_L17_except_error)
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);

        /* "uvloop/loop.pyx":2718
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 app_transport.close()
 */
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_4, __pyx_t_5);
        __pyx_t_2 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0; 
        __PYX_ERR(0, 2718, __pyx_L17_except_error)
      }

      /* "uvloop/loop.pyx":2719
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 app_transport.close()
 *                 raise
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("uvloop.loop.Loop.connect_accepted_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(0, 2719, __pyx_L17_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);

        /* "uvloop/loop.pyx":2720
 *                 raise
 *             except BaseException:
 *                 app_transport.close()             # <<<<<<<<<<<<<<
 *                 raise
 *             return app_transport, protocol
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_app_transport, __pyx_n_s_close); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2720, __pyx_L17_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2720, __pyx_L17_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "uvloop/loop.pyx":2721
 *             except BaseException:
 *                 app_transport.close()
 *                 raise             # <<<<<<<<<<<<<<
 *             return app_transport, protocol
 *         else:
 */
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_4, __pyx_t_2);
        __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_2 = 0; 
        __PYX_ERR(0, 2721, __pyx_L17_except_error)
      }
      goto __pyx_L17_except_error;

      /* "uvloop/loop.pyx":2715
 *         if ssl:
 *             app_transport = protocol._get_app_transport(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L17_except_error:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L1_error;
      __pyx_L20_try_end:;
    }

    /* "uvloop/loop.pyx":2722
 *                 app_transport.close()
 *                 raise
 *             return app_transport, protocol             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_app_transport);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_app_transport);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_cur_scope->__pyx_v_app_transport)) __PYX_ERR(0, 2722, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_protocol)) __PYX_ERR(0, 2722, __pyx_L1_error);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":2713
 *         transport._attach_fileobj(sock)
 * 
 *         if ssl:             # <<<<<<<<<<<<<<
 *             app_transport = protocol._get_app_transport(context)
 *             try:
 */
  }

  /* "uvloop/loop.pyx":2724
 *             return app_transport, protocol
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
  /*else*/ {
    {
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "uvloop/loop.pyx":2725
 *         else:
 *             try:
 *                 await waiter             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_8);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_8;
          __Pyx_XGIVEREF(__pyx_t_9);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_9;
          __Pyx_XGIVEREF(__pyx_t_10);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_10;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 2;
          return __pyx_r;
          __pyx_L32_resume_from_await:;
          __pyx_t_8 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_8);
          __pyx_t_9 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_9);
          __pyx_t_10 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_10);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2725, __pyx_L26_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(0, 2725, __pyx_L26_error)
          }
        }

        /* "uvloop/loop.pyx":2724
 *             return app_transport, protocol
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L31_try_end;
      __pyx_L26_error:;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":2726
 *             try:
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException:
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("uvloop.loop.Loop.connect_accepted_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(0, 2726, __pyx_L28_except_error)
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);

        /* "uvloop/loop.pyx":2727
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException:
 *                 transport._close()
 */
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_4, __pyx_t_5);
        __pyx_t_2 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0; 
        __PYX_ERR(0, 2727, __pyx_L28_except_error)
      }

      /* "uvloop/loop.pyx":2728
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException:             # <<<<<<<<<<<<<<
 *                 transport._close()
 *                 raise
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("uvloop.loop.Loop.connect_accepted_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(0, 2728, __pyx_L28_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);

        /* "uvloop/loop.pyx":2729
 *                 raise
 *             except BaseException:
 *                 transport._close()             # <<<<<<<<<<<<<<
 *                 raise
 *             return transport, protocol
 */
        __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_cur_scope->__pyx_v_transport->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_transport)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2729, __pyx_L28_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

        /* "uvloop/loop.pyx":2730
 *             except BaseException:
 *                 transport._close()
 *                 raise             # <<<<<<<<<<<<<<
 *             return transport, protocol
 * 
 */
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_4, __pyx_t_2);
        __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_2 = 0; 
        __PYX_ERR(0, 2730, __pyx_L28_except_error)
      }
      goto __pyx_L28_except_error;

      /* "uvloop/loop.pyx":2724
 *             return app_transport, protocol
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 await waiter
 *             except (KeyboardInterrupt, SystemExit):
 */
      __pyx_L28_except_error:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L31_try_end:;
    }

    /* "uvloop/loop.pyx":2731
 *                 transport._close()
 *                 raise
 *             return transport, protocol             # <<<<<<<<<<<<<<
 * 
 *     def run_in_executor(self, executor, func, *args):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2731, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_transport);
    __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_transport);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_transport))) __PYX_ERR(0, 2731, __pyx_L1_error);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_protocol)) __PYX_ERR(0, 2731, __pyx_L1_error);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2649
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("connect_accepted_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2733
 *             return transport, protocol
 * 
 *     def run_in_executor(self, executor, func, *args):             # <<<<<<<<<<<<<<
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):
 *             raise TypeError("coroutines cannot be used with run_in_executor()")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_111run_in_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_110run_in_executor, "Loop.run_in_executor(self, executor, func, *args)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_111run_in_executor = {"run_in_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_111run_in_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_110run_in_executor};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_111run_in_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_executor = 0;
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("run_in_executor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 2, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_executor,&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_executor)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2733, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2733, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("run_in_executor", 0, 2, 2, 1); __PYX_ERR(0, 2733, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 2) ? kwd_pos_args : 2;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "run_in_executor") < 0)) __PYX_ERR(0, 2733, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_executor = values[0];
    __pyx_v_func = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("run_in_executor", 0, 2, 2, __pyx_nargs); __PYX_ERR(0, 2733, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.run_in_executor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_110run_in_executor(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_executor, __pyx_v_func, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_110run_in_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_executor, PyObject *__pyx_v_func, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_in_executor", 0);
  __Pyx_INCREF(__pyx_v_executor);

  /* "uvloop/loop.pyx":2734
 * 
 *     def run_in_executor(self, executor, func, *args):
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):             # <<<<<<<<<<<<<<
 *             raise TypeError("coroutines cannot be used with run_in_executor()")
 * 
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_iscoroutine);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_aio_iscoroutine; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_func};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2734, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_iscoroutinefunction);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_aio_iscoroutinefunction; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_func};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2734, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 2734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2735
 *     def run_in_executor(self, executor, func, *args):
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):
 *             raise TypeError("coroutines cannot be used with run_in_executor()")             # <<<<<<<<<<<<<<
 * 
 *         self._check_closed()
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2735, __pyx_L1_error)

    /* "uvloop/loop.pyx":2734
 * 
 *     def run_in_executor(self, executor, func, *args):
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):             # <<<<<<<<<<<<<<
 *             raise TypeError("coroutines cannot be used with run_in_executor()")
 * 
 */
  }

  /* "uvloop/loop.pyx":2737
 *             raise TypeError("coroutines cannot be used with run_in_executor()")
 * 
 *         self._check_closed()             # <<<<<<<<<<<<<<
 * 
 *         if executor is None:
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2739
 *         self._check_closed()
 * 
 *         if executor is None:             # <<<<<<<<<<<<<<
 *             executor = self._default_executor
 *             # Only check when the default executor is being used
 */
  __pyx_t_1 = (__pyx_v_executor == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2740
 * 
 *         if executor is None:
 *             executor = self._default_executor             # <<<<<<<<<<<<<<
 *             # Only check when the default executor is being used
 *             self._check_default_executor()
 */
    __pyx_t_2 = __pyx_v_self->_default_executor;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_executor, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2742
 *             executor = self._default_executor
 *             # Only check when the default executor is being used
 *             self._check_default_executor()             # <<<<<<<<<<<<<<
 *             if executor is None:
 *                 executor = cc_ThreadPoolExecutor()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_check_default_executor); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2742, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":2743
 *             # Only check when the default executor is being used
 *             self._check_default_executor()
 *             if executor is None:             # <<<<<<<<<<<<<<
 *                 executor = cc_ThreadPoolExecutor()
 *                 self._default_executor = executor
 */
    __pyx_t_1 = (__pyx_v_executor == Py_None);
    if (__pyx_t_1) {

      /* "uvloop/loop.pyx":2744
 *             self._check_default_executor()
 *             if executor is None:
 *                 executor = cc_ThreadPoolExecutor()             # <<<<<<<<<<<<<<
 *                 self._default_executor = executor
 * 
 */
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_cc_ThreadPoolExecutor);
      __pyx_t_3 = __pyx_v_6uvloop_4loop_cc_ThreadPoolExecutor; __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2744, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_executor, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":2745
 *             if executor is None:
 *                 executor = cc_ThreadPoolExecutor()
 *                 self._default_executor = executor             # <<<<<<<<<<<<<<
 * 
 *         return aio_wrap_future(executor.submit(func, *args), loop=self)
 */
      __Pyx_INCREF(__pyx_v_executor);
      __Pyx_GIVEREF(__pyx_v_executor);
      __Pyx_GOTREF(__pyx_v_self->_default_executor);
      __Pyx_DECREF(__pyx_v_self->_default_executor);
      __pyx_v_self->_default_executor = __pyx_v_executor;

      /* "uvloop/loop.pyx":2743
 *             # Only check when the default executor is being used
 *             self._check_default_executor()
 *             if executor is None:             # <<<<<<<<<<<<<<
 *                 executor = cc_ThreadPoolExecutor()
 *                 self._default_executor = executor
 */
    }

    /* "uvloop/loop.pyx":2739
 *         self._check_closed()
 * 
 *         if executor is None:             # <<<<<<<<<<<<<<
 *             executor = self._default_executor
 *             # Only check when the default executor is being used
 */
  }

  /* "uvloop/loop.pyx":2747
 *                 self._default_executor = executor
 * 
 *         return aio_wrap_future(executor.submit(func, *args), loop=self)             # <<<<<<<<<<<<<<
 * 
 *     def set_default_executor(self, executor):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_executor, __pyx_n_s_submit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_func)) __PYX_ERR(0, 2747, __pyx_L1_error);
  __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_v_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(0, 2747, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_loop, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 2747, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_wrap_future, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":2733
 *             return transport, protocol
 * 
 *     def run_in_executor(self, executor, func, *args):             # <<<<<<<<<<<<<<
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):
 *             raise TypeError("coroutines cannot be used with run_in_executor()")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Loop.run_in_executor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_executor);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2749
 *         return aio_wrap_future(executor.submit(func, *args), loop=self)
 * 
 *     def set_default_executor(self, executor):             # <<<<<<<<<<<<<<
 *         self._default_executor = executor
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_113set_default_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_112set_default_executor, "Loop.set_default_executor(self, executor)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_113set_default_executor = {"set_default_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_113set_default_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_112set_default_executor};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_113set_default_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_executor = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_default_executor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_executor,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_executor)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2749, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_default_executor") < 0)) __PYX_ERR(0, 2749, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_executor = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_default_executor", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2749, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.set_default_executor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_112set_default_executor(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_executor);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_112set_default_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_executor) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_default_executor", 1);

  /* "uvloop/loop.pyx":2750
 * 
 *     def set_default_executor(self, executor):
 *         self._default_executor = executor             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_INCREF(__pyx_v_executor);
  __Pyx_GIVEREF(__pyx_v_executor);
  __Pyx_GOTREF(__pyx_v_self->_default_executor);
  __Pyx_DECREF(__pyx_v_self->_default_executor);
  __pyx_v_self->_default_executor = __pyx_v_executor;

  /* "uvloop/loop.pyx":2749
 *         return aio_wrap_future(executor.submit(func, *args), loop=self)
 * 
 *     def set_default_executor(self, executor):             # <<<<<<<<<<<<<<
 *         self._default_executor = executor
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_116generator16(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2752
 *         self._default_executor = executor
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_115_Loop__subprocess_run(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_114__subprocess_run, "Loop.__subprocess_run(self, protocol_factory, args, stdin=subprocess_PIPE, stdout=subprocess_PIPE, stderr=subprocess_PIPE, universal_newlines=False, shell=True, bufsize=0, preexec_fn=None, close_fds=None, cwd=None, env=None, startupinfo=None, creationflags=0, restore_signals=True, start_new_session=False, executable=None, pass_fds=(), **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_115_Loop__subprocess_run = {"_Loop__subprocess_run", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_115_Loop__subprocess_run, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_114__subprocess_run};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_115_Loop__subprocess_run(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_stdin = 0;
  PyObject *__pyx_v_stdout = 0;
  PyObject *__pyx_v_stderr = 0;
  PyObject *__pyx_v_universal_newlines = 0;
  CYTHON_UNUSED PyObject *__pyx_v_shell = 0;
  PyObject *__pyx_v_bufsize = 0;
  PyObject *__pyx_v_preexec_fn = 0;
  CYTHON_UNUSED PyObject *__pyx_v_close_fds = 0;
  PyObject *__pyx_v_cwd = 0;
  PyObject *__pyx_v_env = 0;
  PyObject *__pyx_v_startupinfo = 0;
  PyObject *__pyx_v_creationflags = 0;
  PyObject *__pyx_v_restore_signals = 0;
  PyObject *__pyx_v_start_new_session = 0;
  PyObject *__pyx_v_executable = 0;
  PyObject *__pyx_v_pass_fds = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[18] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Loop__subprocess_run (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_args,&__pyx_n_s_stdin,&__pyx_n_s_stdout,&__pyx_n_s_stderr,&__pyx_n_s_universal_newlines,&__pyx_n_s_shell,&__pyx_n_s_bufsize,&__pyx_n_s_preexec_fn,&__pyx_n_s_close_fds,&__pyx_n_s_cwd,&__pyx_n_s_env,&__pyx_n_s_startupinfo,&__pyx_n_s_creationflags,&__pyx_n_s_restore_signals,&__pyx_n_s_start_new_session,&__pyx_n_s_executable,&__pyx_n_s_pass_fds,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__95);
    values[3] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__96);
    values[4] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__97);

    /* "uvloop/loop.pyx":2757
 *                                stdout=subprocess_PIPE,
 *                                stderr=subprocess_PIPE,
 *                                universal_newlines=False,             # <<<<<<<<<<<<<<
 *                                shell=True,
 *                                bufsize=0,
 */
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));

    /* "uvloop/loop.pyx":2758
 *                                stderr=subprocess_PIPE,
 *                                universal_newlines=False,
 *                                shell=True,             # <<<<<<<<<<<<<<
 *                                bufsize=0,
 *                                preexec_fn=None,
 */
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    values[7] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));

    /* "uvloop/loop.pyx":2760
 *                                shell=True,
 *                                bufsize=0,
 *                                preexec_fn=None,             # <<<<<<<<<<<<<<
 *                                close_fds=None,
 *                                cwd=None,
 */
    values[8] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2761
 *                                bufsize=0,
 *                                preexec_fn=None,
 *                                close_fds=None,             # <<<<<<<<<<<<<<
 *                                cwd=None,
 *                                env=None,
 */
    values[9] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2762
 *                                preexec_fn=None,
 *                                close_fds=None,
 *                                cwd=None,             # <<<<<<<<<<<<<<
 *                                env=None,
 *                                startupinfo=None,
 */
    values[10] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2763
 *                                close_fds=None,
 *                                cwd=None,
 *                                env=None,             # <<<<<<<<<<<<<<
 *                                startupinfo=None,
 *                                creationflags=0,
 */
    values[11] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2764
 *                                cwd=None,
 *                                env=None,
 *                                startupinfo=None,             # <<<<<<<<<<<<<<
 *                                creationflags=0,
 *                                restore_signals=True,
 */
    values[12] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[13] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));

    /* "uvloop/loop.pyx":2766
 *                                startupinfo=None,
 *                                creationflags=0,
 *                                restore_signals=True,             # <<<<<<<<<<<<<<
 *                                start_new_session=False,
 *                                executable=None,
 */
    values[14] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));

    /* "uvloop/loop.pyx":2767
 *                                creationflags=0,
 *                                restore_signals=True,
 *                                start_new_session=False,             # <<<<<<<<<<<<<<
 *                                executable=None,
 *                                pass_fds=(),
 */
    values[15] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));

    /* "uvloop/loop.pyx":2768
 *                                restore_signals=True,
 *                                start_new_session=False,
 *                                executable=None,             # <<<<<<<<<<<<<<
 *                                pass_fds=(),
 *                                **kwargs):
 */
    values[16] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":2769
 *                                start_new_session=False,
 *                                executable=None,
 *                                pass_fds=(),             # <<<<<<<<<<<<<<
 *                                **kwargs):
 * 
 */
    values[17] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_empty_tuple));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case 18: values[17] = __Pyx_Arg_FASTCALL(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = __Pyx_Arg_FASTCALL(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = __Pyx_Arg_FASTCALL(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = __Pyx_Arg_FASTCALL(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = __Pyx_Arg_FASTCALL(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = __Pyx_Arg_FASTCALL(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = __Pyx_Arg_FASTCALL(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = __Pyx_Arg_FASTCALL(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = __Pyx_Arg_FASTCALL(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = __Pyx_Arg_FASTCALL(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_FASTCALL(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_args)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_Loop__subprocess_run", 0, 2, 18, 1); __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_stdin);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_stdout);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_stderr);
          if (value) { values[4] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_universal_newlines);
          if (value) { values[5] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_shell);
          if (value) { values[6] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bufsize);
          if (value) { values[7] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_preexec_fn);
          if (value) { values[8] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_close_fds);
          if (value) { values[9] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cwd);
          if (value) { values[10] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_env);
          if (value) { values[11] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_startupinfo);
          if (value) { values[12] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_creationflags);
          if (value) { values[13] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_restore_signals);
          if (value) { values[14] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_start_new_session);
          if (value) { values[15] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_executable);
          if (value) { values[16] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pass_fds);
          if (value) { values[17] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "_Loop__subprocess_run") < 0)) __PYX_ERR(0, 2752, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case 18: values[17] = __Pyx_Arg_FASTCALL(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = __Pyx_Arg_FASTCALL(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = __Pyx_Arg_FASTCALL(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = __Pyx_Arg_FASTCALL(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = __Pyx_Arg_FASTCALL(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = __Pyx_Arg_FASTCALL(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = __Pyx_Arg_FASTCALL(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = __Pyx_Arg_FASTCALL(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = __Pyx_Arg_FASTCALL(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = __Pyx_Arg_FASTCALL(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_FASTCALL(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_args = values[1];
    __pyx_v_stdin = values[2];
    __pyx_v_stdout = values[3];
    __pyx_v_stderr = values[4];
    __pyx_v_universal_newlines = values[5];
    __pyx_v_shell = values[6];
    __pyx_v_bufsize = values[7];
    __pyx_v_preexec_fn = values[8];
    __pyx_v_close_fds = values[9];
    __pyx_v_cwd = values[10];
    __pyx_v_env = values[11];
    __pyx_v_startupinfo = values[12];
    __pyx_v_creationflags = values[13];
    __pyx_v_restore_signals = values[14];
    __pyx_v_start_new_session = values[15];
    __pyx_v_executable = values[16];
    __pyx_v_pass_fds = values[17];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_Loop__subprocess_run", 0, 2, 18, __pyx_nargs); __PYX_ERR(0, 2752, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_CLEAR(__pyx_v_kwargs);
  __Pyx_AddTraceback("uvloop.loop.Loop._Loop__subprocess_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_114__subprocess_run(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_args, __pyx_v_stdin, __pyx_v_stdout, __pyx_v_stderr, __pyx_v_universal_newlines, __pyx_v_shell, __pyx_v_bufsize, __pyx_v_preexec_fn, __pyx_v_close_fds, __pyx_v_cwd, __pyx_v_env, __pyx_v_startupinfo, __pyx_v_creationflags, __pyx_v_restore_signals, __pyx_v_start_new_session, __pyx_v_executable, __pyx_v_pass_fds, __pyx_v_kwargs);

  /* "uvloop/loop.pyx":2752
 *         self._default_executor = executor
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 */

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_114__subprocess_run(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_args, PyObject *__pyx_v_stdin, PyObject *__pyx_v_stdout, PyObject *__pyx_v_stderr, PyObject *__pyx_v_universal_newlines, CYTHON_UNUSED PyObject *__pyx_v_shell, PyObject *__pyx_v_bufsize, PyObject *__pyx_v_preexec_fn, CYTHON_UNUSED PyObject *__pyx_v_close_fds, PyObject *__pyx_v_cwd, PyObject *__pyx_v_env, PyObject *__pyx_v_startupinfo, PyObject *__pyx_v_creationflags, PyObject *__pyx_v_restore_signals, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v_executable, PyObject *__pyx_v_pass_fds, PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Loop__subprocess_run", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2752, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_args = __pyx_v_args;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_args);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_args);
  __pyx_cur_scope->__pyx_v_stdin = __pyx_v_stdin;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_stdin);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_stdin);
  __pyx_cur_scope->__pyx_v_stdout = __pyx_v_stdout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_stdout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_stdout);
  __pyx_cur_scope->__pyx_v_stderr = __pyx_v_stderr;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_stderr);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_stderr);
  __pyx_cur_scope->__pyx_v_universal_newlines = __pyx_v_universal_newlines;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_universal_newlines);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_universal_newlines);
  __pyx_cur_scope->__pyx_v_shell = __pyx_v_shell;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_shell);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_shell);
  __pyx_cur_scope->__pyx_v_bufsize = __pyx_v_bufsize;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_bufsize);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_bufsize);
  __pyx_cur_scope->__pyx_v_preexec_fn = __pyx_v_preexec_fn;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_preexec_fn);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_preexec_fn);
  __pyx_cur_scope->__pyx_v_close_fds = __pyx_v_close_fds;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_close_fds);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_close_fds);
  __pyx_cur_scope->__pyx_v_cwd = __pyx_v_cwd;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_cwd);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_cwd);
  __pyx_cur_scope->__pyx_v_env = __pyx_v_env;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_env);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_env);
  __pyx_cur_scope->__pyx_v_startupinfo = __pyx_v_startupinfo;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_startupinfo);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_startupinfo);
  __pyx_cur_scope->__pyx_v_creationflags = __pyx_v_creationflags;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_creationflags);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_creationflags);
  __pyx_cur_scope->__pyx_v_restore_signals = __pyx_v_restore_signals;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_restore_signals);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_restore_signals);
  __pyx_cur_scope->__pyx_v_start_new_session = __pyx_v_start_new_session;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_start_new_session);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_start_new_session);
  __pyx_cur_scope->__pyx_v_executable = __pyx_v_executable;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_executable);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_executable);
  __pyx_cur_scope->__pyx_v_pass_fds = __pyx_v_pass_fds;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_pass_fds);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_pass_fds);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_116generator16, __pyx_codeobj__98, (PyObject *) __pyx_cur_scope, __pyx_n_s_subprocess_run, __pyx_n_s_Loop___subprocess_run, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2752, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop._Loop__subprocess_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_116generator16(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__subprocess_run", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L17_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2752, __pyx_L1_error)

  /* "uvloop/loop.pyx":2776
 * 
 *         cdef:
 *             int debug_flags = 0             # <<<<<<<<<<<<<<
 * 
 *         if universal_newlines:
 */
  __pyx_cur_scope->__pyx_v_debug_flags = 0;

  /* "uvloop/loop.pyx":2778
 *             int debug_flags = 0
 * 
 *         if universal_newlines:             # <<<<<<<<<<<<<<
 *             raise ValueError("universal_newlines must be False")
 *         if bufsize != 0:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_universal_newlines); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2778, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2779
 * 
 *         if universal_newlines:
 *             raise ValueError("universal_newlines must be False")             # <<<<<<<<<<<<<<
 *         if bufsize != 0:
 *             raise ValueError("bufsize must be 0")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__99, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2779, __pyx_L1_error)

    /* "uvloop/loop.pyx":2778
 *             int debug_flags = 0
 * 
 *         if universal_newlines:             # <<<<<<<<<<<<<<
 *             raise ValueError("universal_newlines must be False")
 *         if bufsize != 0:
 */
  }

  /* "uvloop/loop.pyx":2780
 *         if universal_newlines:
 *             raise ValueError("universal_newlines must be False")
 *         if bufsize != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("bufsize must be 0")
 *         if startupinfo is not None:
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolNeObjC(__pyx_cur_scope->__pyx_v_bufsize, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2780, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2781
 *             raise ValueError("universal_newlines must be False")
 *         if bufsize != 0:
 *             raise ValueError("bufsize must be 0")             # <<<<<<<<<<<<<<
 *         if startupinfo is not None:
 *             raise ValueError('startupinfo is not supported')
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__100, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2781, __pyx_L1_error)

    /* "uvloop/loop.pyx":2780
 *         if universal_newlines:
 *             raise ValueError("universal_newlines must be False")
 *         if bufsize != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("bufsize must be 0")
 *         if startupinfo is not None:
 */
  }

  /* "uvloop/loop.pyx":2782
 *         if bufsize != 0:
 *             raise ValueError("bufsize must be 0")
 *         if startupinfo is not None:             # <<<<<<<<<<<<<<
 *             raise ValueError('startupinfo is not supported')
 *         if creationflags != 0:
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_startupinfo != Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2783
 *             raise ValueError("bufsize must be 0")
 *         if startupinfo is not None:
 *             raise ValueError('startupinfo is not supported')             # <<<<<<<<<<<<<<
 *         if creationflags != 0:
 *             raise ValueError('creationflags is not supported')
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__101, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2783, __pyx_L1_error)

    /* "uvloop/loop.pyx":2782
 *         if bufsize != 0:
 *             raise ValueError("bufsize must be 0")
 *         if startupinfo is not None:             # <<<<<<<<<<<<<<
 *             raise ValueError('startupinfo is not supported')
 *         if creationflags != 0:
 */
  }

  /* "uvloop/loop.pyx":2784
 *         if startupinfo is not None:
 *             raise ValueError('startupinfo is not supported')
 *         if creationflags != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('creationflags is not supported')
 * 
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolNeObjC(__pyx_cur_scope->__pyx_v_creationflags, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2784, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2785
 *             raise ValueError('startupinfo is not supported')
 *         if creationflags != 0:
 *             raise ValueError('creationflags is not supported')             # <<<<<<<<<<<<<<
 * 
 *         if executable is not None:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__102, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2785, __pyx_L1_error)

    /* "uvloop/loop.pyx":2784
 *         if startupinfo is not None:
 *             raise ValueError('startupinfo is not supported')
 *         if creationflags != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('creationflags is not supported')
 * 
 */
  }

  /* "uvloop/loop.pyx":2787
 *             raise ValueError('creationflags is not supported')
 * 
 *         if executable is not None:             # <<<<<<<<<<<<<<
 *             args[0] = executable
 * 
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_executable != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2788
 * 
 *         if executable is not None:
 *             args[0] = executable             # <<<<<<<<<<<<<<
 * 
 *         # For tests only! Do not use in your code. Ever.
 */
    if (unlikely((__Pyx_SetItemInt(__pyx_cur_scope->__pyx_v_args, 0, __pyx_cur_scope->__pyx_v_executable, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0))) __PYX_ERR(0, 2788, __pyx_L1_error)

    /* "uvloop/loop.pyx":2787
 *             raise ValueError('creationflags is not supported')
 * 
 *         if executable is not None:             # <<<<<<<<<<<<<<
 *             args[0] = executable
 * 
 */
  }

  /* "uvloop/loop.pyx":2791
 * 
 *         # For tests only! Do not use in your code. Ever.
 *         if kwargs.pop("__uvloop_sleep_after_fork", False):             # <<<<<<<<<<<<<<
 *             debug_flags |= __PROCESS_DEBUG_SLEEP_AFTER_FORK
 *         if kwargs:
 */
  __pyx_t_2 = __Pyx_PyDict_Pop(__pyx_cur_scope->__pyx_v_kwargs, __pyx_n_u_uvloop_sleep_after_fork, Py_False); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2791, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":2792
 *         # For tests only! Do not use in your code. Ever.
 *         if kwargs.pop("__uvloop_sleep_after_fork", False):
 *             debug_flags |= __PROCESS_DEBUG_SLEEP_AFTER_FORK             # <<<<<<<<<<<<<<
 *         if kwargs:
 *             raise ValueError(
 */
    __pyx_cur_scope->__pyx_v_debug_flags = (__pyx_cur_scope->__pyx_v_debug_flags | __pyx_e_6uvloop_4loop___PROCESS_DEBUG_SLEEP_AFTER_FORK);

    /* "uvloop/loop.pyx":2791
 * 
 *         # For tests only! Do not use in your code. Ever.
 *         if kwargs.pop("__uvloop_sleep_after_fork", False):             # <<<<<<<<<<<<<<
 *             debug_flags |= __PROCESS_DEBUG_SLEEP_AFTER_FORK
 *         if kwargs:
 */
  }

  /* "uvloop/loop.pyx":2793
 *         if kwargs.pop("__uvloop_sleep_after_fork", False):
 *             debug_flags |= __PROCESS_DEBUG_SLEEP_AFTER_FORK
 *         if kwargs:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'unexpected kwargs: {}'.format(', '.join(kwargs.keys())))
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_kwargs); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2793, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2795
 *         if kwargs:
 *             raise ValueError(
 *                 'unexpected kwargs: {}'.format(', '.join(kwargs.keys())))             # <<<<<<<<<<<<<<
 * 
 *         waiter = self._new_future()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unexpected_kwargs, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyDict_Keys(__pyx_cur_scope->__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__69, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "uvloop/loop.pyx":2794
 *             debug_flags |= __PROCESS_DEBUG_SLEEP_AFTER_FORK
 *         if kwargs:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'unexpected kwargs: {}'.format(', '.join(kwargs.keys())))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 2794, __pyx_L1_error)

    /* "uvloop/loop.pyx":2793
 *         if kwargs.pop("__uvloop_sleep_after_fork", False):
 *             debug_flags |= __PROCESS_DEBUG_SLEEP_AFTER_FORK
 *         if kwargs:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'unexpected kwargs: {}'.format(', '.join(kwargs.keys())))
 */
  }

  /* "uvloop/loop.pyx":2797
 *                 'unexpected kwargs: {}'.format(', '.join(kwargs.keys())))
 * 
 *         waiter = self._new_future()             # <<<<<<<<<<<<<<
 *         protocol = protocol_factory()
 *         proc = UVProcessTransport.new(self, protocol,
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_waiter = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":2798
 * 
 *         waiter = self._new_future()
 *         protocol = protocol_factory()             # <<<<<<<<<<<<<<
 *         proc = UVProcessTransport.new(self, protocol,
 *                                       args, env, cwd, start_new_session,
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_protocol_factory; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_protocol = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":2803
 *                                       stdin, stdout, stderr, pass_fds,
 *                                       waiter,
 *                                       debug_flags,             # <<<<<<<<<<<<<<
 *                                       preexec_fn,
 *                                       restore_signals)
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_cur_scope->__pyx_v_debug_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":2799
 *         waiter = self._new_future()
 *         protocol = protocol_factory()
 *         proc = UVProcessTransport.new(self, protocol,             # <<<<<<<<<<<<<<
 *                                       args, env, cwd, start_new_session,
 *                                       stdin, stdout, stderr, pass_fds,
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_6uvloop_4loop_18UVProcessTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_protocol, __pyx_cur_scope->__pyx_v_args, __pyx_cur_scope->__pyx_v_env, __pyx_cur_scope->__pyx_v_cwd, __pyx_cur_scope->__pyx_v_start_new_session, __pyx_cur_scope->__pyx_v_stdin, __pyx_cur_scope->__pyx_v_stdout, __pyx_cur_scope->__pyx_v_stderr, __pyx_cur_scope->__pyx_v_pass_fds, __pyx_cur_scope->__pyx_v_waiter, __pyx_t_3, __pyx_cur_scope->__pyx_v_preexec_fn, __pyx_cur_scope->__pyx_v_restore_signals)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_proc = ((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2807
 *                                       restore_signals)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "uvloop/loop.pyx":2808
 * 
 *         try:
 *             await waiter             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_t_7);
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_7;
        __Pyx_XGIVEREF(__pyx_t_8);
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_8;
        __Pyx_XGIVEREF(__pyx_t_9);
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_9;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L17_resume_from_await:;
        __pyx_t_7 = __pyx_cur_scope->__pyx_t_0;
        __pyx_cur_scope->__pyx_t_0 = 0;
        __Pyx_XGOTREF(__pyx_t_7);
        __pyx_t_8 = __pyx_cur_scope->__pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = 0;
        __Pyx_XGOTREF(__pyx_t_8);
        __pyx_t_9 = __pyx_cur_scope->__pyx_t_2;
        __pyx_cur_scope->__pyx_t_2 = 0;
        __Pyx_XGOTREF(__pyx_t_9);
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2808, __pyx_L11_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 2808, __pyx_L11_error)
        }
      }

      /* "uvloop/loop.pyx":2807
 *                                       restore_signals)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L16_try_end;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":2809
 *         try:
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException:
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("uvloop.loop.Loop._Loop__subprocess_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_5) < 0) __PYX_ERR(0, 2809, __pyx_L13_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "uvloop/loop.pyx":2810
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException:
 *             proc.close()
 */
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_3, __pyx_t_5);
      __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_5 = 0; 
      __PYX_ERR(0, 2810, __pyx_L13_except_error)
    }

    /* "uvloop/loop.pyx":2811
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException:             # <<<<<<<<<<<<<<
 *             proc.close()
 *             raise
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("uvloop.loop.Loop._Loop__subprocess_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 2811, __pyx_L13_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);

      /* "uvloop/loop.pyx":2812
 *             raise
 *         except BaseException:
 *             proc.close()             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_proc), __pyx_n_s_close); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2812, __pyx_L13_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2812, __pyx_L13_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":2813
 *         except BaseException:
 *             proc.close()
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         return proc, protocol
 */
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_3, __pyx_t_2);
      __pyx_t_5 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
      __PYX_ERR(0, 2813, __pyx_L13_except_error)
    }
    goto __pyx_L13_except_error;

    /* "uvloop/loop.pyx":2807
 *                                       restore_signals)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
    __pyx_L13_except_error:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L16_try_end:;
  }

  /* "uvloop/loop.pyx":2815
 *             raise
 * 
 *         return proc, protocol             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_proc);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_proc);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_proc))) __PYX_ERR(0, 2815, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_protocol)) __PYX_ERR(0, 2815, __pyx_L1_error);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2752
 *         self._default_executor = executor
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("__subprocess_run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_119generator17(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2817
 *         return proc, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_118subprocess_shell(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_117subprocess_shell, "Loop.subprocess_shell(self, protocol_factory, cmd, *, shell=True, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_118subprocess_shell = {"subprocess_shell", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_118subprocess_shell, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_117subprocess_shell};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_118subprocess_shell(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_cmd = 0;
  PyObject *__pyx_v_shell = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subprocess_shell (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_cmd,&__pyx_n_s_shell,0};

    /* "uvloop/loop.pyx":2819
 *     @cython.iterable_coroutine
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,             # <<<<<<<<<<<<<<
 *                                **kwargs):
 * 
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_True));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2817, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_cmd)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2817, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("subprocess_shell", 1, 2, 2, 1); __PYX_ERR(0, 2817, __pyx_L3_error)
        }
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 2;
        PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
        if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2817, __pyx_L3_error)
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "subprocess_shell") < 0)) __PYX_ERR(0, 2817, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_cmd = values[1];
    __pyx_v_shell = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subprocess_shell", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2817, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_CLEAR(__pyx_v_kwargs);
  __Pyx_AddTraceback("uvloop.loop.Loop.subprocess_shell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_117subprocess_shell(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_cmd, __pyx_v_shell, __pyx_v_kwargs);

  /* "uvloop/loop.pyx":2817
 *         return proc, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,
 */

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_117subprocess_shell(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_cmd, PyObject *__pyx_v_shell, PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subprocess_shell", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2817, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_cmd = __pyx_v_cmd;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_cmd);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_cmd);
  __pyx_cur_scope->__pyx_v_shell = __pyx_v_shell;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_shell);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_shell);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_119generator17, __pyx_codeobj__103, (PyObject *) __pyx_cur_scope, __pyx_n_s_subprocess_shell, __pyx_n_s_Loop_subprocess_shell, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2817, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.subprocess_shell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_119generator17(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subprocess_shell", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2817, __pyx_L1_error)

  /* "uvloop/loop.pyx":2822
 *                                **kwargs):
 * 
 *         if not shell:             # <<<<<<<<<<<<<<
 *             raise ValueError("shell must be True")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_shell); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2822, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/loop.pyx":2823
 * 
 *         if not shell:
 *             raise ValueError("shell must be True")             # <<<<<<<<<<<<<<
 * 
 *         args = [cmd]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__104, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 2823, __pyx_L1_error)

    /* "uvloop/loop.pyx":2822
 *                                **kwargs):
 * 
 *         if not shell:             # <<<<<<<<<<<<<<
 *             raise ValueError("shell must be True")
 * 
 */
  }

  /* "uvloop/loop.pyx":2825
 *             raise ValueError("shell must be True")
 * 
 *         args = [cmd]             # <<<<<<<<<<<<<<
 *         if shell:
 *             args = [b'/bin/sh', b'-c'] + args
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_cmd);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_cmd);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_cmd)) __PYX_ERR(0, 2825, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_args = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":2826
 * 
 *         args = [cmd]
 *         if shell:             # <<<<<<<<<<<<<<
 *             args = [b'/bin/sh', b'-c'] + args
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_shell); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2826, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":2827
 *         args = [cmd]
 *         if shell:
 *             args = [b'/bin/sh', b'-c'] + args             # <<<<<<<<<<<<<<
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=True,
 */
    __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_kp_b_bin_sh);
    __Pyx_GIVEREF(__pyx_kp_b_bin_sh);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_kp_b_bin_sh)) __PYX_ERR(0, 2827, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_b_c);
    __Pyx_GIVEREF(__pyx_kp_b_c);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_kp_b_c)) __PYX_ERR(0, 2827, __pyx_L1_error);
    __pyx_t_4 = PyNumber_Add(__pyx_t_3, __pyx_cur_scope->__pyx_v_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_args);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_args, ((PyObject*)__pyx_t_4));
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":2826
 * 
 *         args = [cmd]
 *         if shell:             # <<<<<<<<<<<<<<
 *             args = [b'/bin/sh', b'-c'] + args
 * 
 */
  }

  /* "uvloop/loop.pyx":2829
 *             args = [b'/bin/sh', b'-c'] + args
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=True,             # <<<<<<<<<<<<<<
 *                                            **kwargs)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_Loop__subprocess_run); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_protocol_factory)) __PYX_ERR(0, 2829, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_args);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_args)) __PYX_ERR(0, 2829, __pyx_L1_error);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_shell, Py_True) < 0) __PYX_ERR(0, 2829, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":2830
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=True,
 *                                            **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  if (__Pyx_MergeKeywords(__pyx_t_5, __pyx_cur_scope->__pyx_v_kwargs) < 0) __PYX_ERR(0, 2830, __pyx_L1_error)

  /* "uvloop/loop.pyx":2829
 *             args = [b'/bin/sh', b'-c'] + args
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=True,             # <<<<<<<<<<<<<<
 *                                            **kwargs)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2829, __pyx_L1_error)
    __pyx_t_6 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_6);
  } else {
    __pyx_t_6 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_6) < 0) __PYX_ERR(0, 2829, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2817
 *         return proc, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("subprocess_shell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_122generator18(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2832
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_121subprocess_exec(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_120subprocess_exec, "Loop.subprocess_exec(self, protocol_factory, program, *args, shell=False, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_121subprocess_exec = {"subprocess_exec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_121subprocess_exec, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_120subprocess_exec};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_121subprocess_exec(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_program = 0;
  PyObject *__pyx_v_shell = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subprocess_exec (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 2, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_CLEAR(__pyx_v_kwargs);
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_program,&__pyx_n_s_shell,0};

    /* "uvloop/loop.pyx":2834
 *     @cython.iterable_coroutine
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):             # <<<<<<<<<<<<<<
 * 
 *         if shell:
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2832, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_program)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2832, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("subprocess_exec", 0, 2, 2, 1); __PYX_ERR(0, 2832, __pyx_L3_error)
        }
      }
      if (kw_args == 1) {
        const Py_ssize_t index = 2;
        PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
        if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2832, __pyx_L3_error)
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 2) ? kwd_pos_args : 2;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, used_pos_args, "subprocess_exec") < 0)) __PYX_ERR(0, 2832, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_program = values[1];
    __pyx_v_shell = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subprocess_exec", 0, 2, 2, __pyx_nargs); __PYX_ERR(0, 2832, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_CLEAR(__pyx_v_args);
  __Pyx_CLEAR(__pyx_v_kwargs);
  __Pyx_AddTraceback("uvloop.loop.Loop.subprocess_exec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_120subprocess_exec(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_program, __pyx_v_shell, __pyx_v_args, __pyx_v_kwargs);

  /* "uvloop/loop.pyx":2832
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):
 */

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_120subprocess_exec(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_program, PyObject *__pyx_v_shell, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subprocess_exec", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2832, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_program = __pyx_v_program;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_program);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_program);
  __pyx_cur_scope->__pyx_v_shell = __pyx_v_shell;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_shell);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_shell);
  __pyx_cur_scope->__pyx_v_args = __pyx_v_args;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_args);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_args);
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_122generator18, __pyx_codeobj__105, (PyObject *) __pyx_cur_scope, __pyx_n_s_subprocess_exec, __pyx_n_s_Loop_subprocess_exec, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2832, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.subprocess_exec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_122generator18(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subprocess_exec", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L5_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2832, __pyx_L1_error)

  /* "uvloop/loop.pyx":2836
 *                               shell=False, **kwargs):
 * 
 *         if shell:             # <<<<<<<<<<<<<<
 *             raise ValueError("shell must be False")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_shell); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2836, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":2837
 * 
 *         if shell:
 *             raise ValueError("shell must be False")             # <<<<<<<<<<<<<<
 * 
 *         args = list((program,) + args)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__106, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2837, __pyx_L1_error)

    /* "uvloop/loop.pyx":2836
 *                               shell=False, **kwargs):
 * 
 *         if shell:             # <<<<<<<<<<<<<<
 *             raise ValueError("shell must be False")
 * 
 */
  }

  /* "uvloop/loop.pyx":2839
 *             raise ValueError("shell must be False")
 * 
 *         args = list((program,) + args)             # <<<<<<<<<<<<<<
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=False,
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_program);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_program);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_cur_scope->__pyx_v_program)) __PYX_ERR(0, 2839, __pyx_L1_error);
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_cur_scope->__pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_args);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_args, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":2841
 *         args = list((program,) + args)
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=False,             # <<<<<<<<<<<<<<
 *                                            **kwargs)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_Loop__subprocess_run); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_protocol_factory)) __PYX_ERR(0, 2841, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_args);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_args);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_cur_scope->__pyx_v_args)) __PYX_ERR(0, 2841, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_shell, Py_False) < 0) __PYX_ERR(0, 2841, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":2842
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=False,
 *                                            **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  if (__Pyx_MergeKeywords(__pyx_t_4, __pyx_cur_scope->__pyx_v_kwargs) < 0) __PYX_ERR(0, 2842, __pyx_L1_error)

  /* "uvloop/loop.pyx":2841
 *         args = list((program,) + args)
 * 
 *         return await self.__subprocess_run(protocol_factory, args, shell=False,             # <<<<<<<<<<<<<<
 *                                            **kwargs)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L5_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2841, __pyx_L1_error)
    __pyx_t_5 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_5);
  } else {
    __pyx_t_5 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_5) < 0) __PYX_ERR(0, 2841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2832
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("subprocess_exec", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_125generator19(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2844
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_read_pipe(self, proto_factory, pipe):
 *         """Register read pipe in event loop. Set the pipe to non-blocking mode.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_124connect_read_pipe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_123connect_read_pipe, "Loop.connect_read_pipe(self, proto_factory, pipe)\nRegister read pipe in event loop. Set the pipe to non-blocking mode.\n\n        protocol_factory should instantiate object with Protocol interface.\n        pipe is a file-like object.\n        Return pair (transport, protocol), where transport supports the\n        ReadTransport interface.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_124connect_read_pipe = {"connect_read_pipe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_124connect_read_pipe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_123connect_read_pipe};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_124connect_read_pipe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proto_factory = 0;
  PyObject *__pyx_v_pipe = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_read_pipe (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proto_factory,&__pyx_n_s_pipe,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proto_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2844, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pipe)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2844, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("connect_read_pipe", 1, 2, 2, 1); __PYX_ERR(0, 2844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connect_read_pipe") < 0)) __PYX_ERR(0, 2844, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_proto_factory = values[0];
    __pyx_v_pipe = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connect_read_pipe", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2844, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.connect_read_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_123connect_read_pipe(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_proto_factory, __pyx_v_pipe);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_123connect_read_pipe(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_proto_factory, PyObject *__pyx_v_pipe) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect_read_pipe", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2844, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_proto_factory = __pyx_v_proto_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_proto_factory);
  __pyx_cur_scope->__pyx_v_pipe = __pyx_v_pipe;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_pipe);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_pipe);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_125generator19, __pyx_codeobj__107, (PyObject *) __pyx_cur_scope, __pyx_n_s_connect_read_pipe, __pyx_n_s_Loop_connect_read_pipe, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2844, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.connect_read_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_125generator19(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_read_pipe", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L10_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2844, __pyx_L1_error)

  /* "uvloop/loop.pyx":2855
 *             ReadUnixTransport transp
 * 
 *         waiter = self._new_future()             # <<<<<<<<<<<<<<
 *         proto = proto_factory()
 *         transp = ReadUnixTransport.new(self, proto, None, waiter)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2856
 * 
 *         waiter = self._new_future()
 *         proto = proto_factory()             # <<<<<<<<<<<<<<
 *         transp = ReadUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto_factory);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_proto_factory; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2856, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_proto = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2857
 *         waiter = self._new_future()
 *         proto = proto_factory()
 *         transp = ReadUnixTransport.new(self, proto, None, waiter)             # <<<<<<<<<<<<<<
 *         transp._add_extra_info('pipe', pipe)
 *         try:
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_17ReadUnixTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_proto, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_waiter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_transp = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2858
 *         proto = proto_factory()
 *         transp = ReadUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)             # <<<<<<<<<<<<<<
 *         try:
 *             transp._open(pipe.fileno())
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._add_extra_info(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_transp), __pyx_n_u_pipe, __pyx_cur_scope->__pyx_v_pipe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2859
 *         transp = ReadUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 *         try:             # <<<<<<<<<<<<<<
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "uvloop/loop.pyx":2860
 *         transp._add_extra_info('pipe', pipe)
 *         try:
 *             transp._open(pipe.fileno())             # <<<<<<<<<<<<<<
 *             transp._init_protocol()
 *             await waiter
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_pipe, __pyx_n_s_fileno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2860, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2860, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2860, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_transp), __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2860, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":2861
 *         try:
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_transp)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2861, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":2862
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 *             await waiter             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_t_5);
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_5;
        __Pyx_XGIVEREF(__pyx_t_6);
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_6;
        __Pyx_XGIVEREF(__pyx_t_7);
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_7;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L10_resume_from_await:;
        __pyx_t_5 = __pyx_cur_scope->__pyx_t_0;
        __pyx_cur_scope->__pyx_t_0 = 0;
        __Pyx_XGOTREF(__pyx_t_5);
        __pyx_t_6 = __pyx_cur_scope->__pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = 0;
        __Pyx_XGOTREF(__pyx_t_6);
        __pyx_t_7 = __pyx_cur_scope->__pyx_t_2;
        __pyx_cur_scope->__pyx_t_2 = 0;
        __Pyx_XGOTREF(__pyx_t_7);
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2862, __pyx_L4_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 2862, __pyx_L4_error)
        }
      }

      /* "uvloop/loop.pyx":2859
 *         transp = ReadUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 *         try:             # <<<<<<<<<<<<<<
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2863
 *             transp._init_protocol()
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.Loop.connect_read_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 2863, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":2864
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException:
 *             transp._close()
 */
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_2, __pyx_t_3);
      __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0; 
      __PYX_ERR(0, 2864, __pyx_L6_except_error)
    }

    /* "uvloop/loop.pyx":2865
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException:             # <<<<<<<<<<<<<<
 *             transp._close()
 *             raise
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.Loop.connect_read_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 2865, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);

      /* "uvloop/loop.pyx":2866
 *             raise
 *         except BaseException:
 *             transp._close()             # <<<<<<<<<<<<<<
 *             raise
 *         transp._attach_fileobj(pipe)
 */
      __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_transp)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2866, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/loop.pyx":2867
 *         except BaseException:
 *             transp._close()
 *             raise             # <<<<<<<<<<<<<<
 *         transp._attach_fileobj(pipe)
 *         return transp, proto
 */
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
      __PYX_ERR(0, 2867, __pyx_L6_except_error)
    }
    goto __pyx_L6_except_error;

    /* "uvloop/loop.pyx":2859
 *         transp = ReadUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 *         try:             # <<<<<<<<<<<<<<
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "uvloop/loop.pyx":2868
 *             transp._close()
 *             raise
 *         transp._attach_fileobj(pipe)             # <<<<<<<<<<<<<<
 *         return transp, proto
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_transp), __pyx_cur_scope->__pyx_v_pipe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2869
 *             raise
 *         transp._attach_fileobj(pipe)
 *         return transp, proto             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_transp);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_transp);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_transp))) __PYX_ERR(0, 2869, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_proto);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_proto)) __PYX_ERR(0, 2869, __pyx_L1_error);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2844
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_read_pipe(self, proto_factory, pipe):
 *         """Register read pipe in event loop. Set the pipe to non-blocking mode.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("connect_read_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_128generator20(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":2871
 *         return transp, proto
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_write_pipe(self, proto_factory, pipe):
 *         """Register write pipe in event loop.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_127connect_write_pipe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_126connect_write_pipe, "Loop.connect_write_pipe(self, proto_factory, pipe)\nRegister write pipe in event loop.\n\n        protocol_factory should instantiate object with BaseProtocol interface.\n        Pipe is file-like object already switched to nonblocking.\n        Return pair (transport, protocol), where transport support\n        WriteTransport interface.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_127connect_write_pipe = {"connect_write_pipe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_127connect_write_pipe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_126connect_write_pipe};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_127connect_write_pipe(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proto_factory = 0;
  PyObject *__pyx_v_pipe = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_write_pipe (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proto_factory,&__pyx_n_s_pipe,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proto_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2871, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pipe)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2871, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("connect_write_pipe", 1, 2, 2, 1); __PYX_ERR(0, 2871, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connect_write_pipe") < 0)) __PYX_ERR(0, 2871, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_proto_factory = values[0];
    __pyx_v_pipe = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connect_write_pipe", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2871, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.connect_write_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_126connect_write_pipe(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_proto_factory, __pyx_v_pipe);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_126connect_write_pipe(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_proto_factory, PyObject *__pyx_v_pipe) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect_write_pipe", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2871, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_proto_factory = __pyx_v_proto_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_proto_factory);
  __pyx_cur_scope->__pyx_v_pipe = __pyx_v_pipe;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_pipe);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_pipe);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_128generator20, __pyx_codeobj__108, (PyObject *) __pyx_cur_scope, __pyx_n_s_connect_write_pipe, __pyx_n_s_Loop_connect_write_pipe, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 2871, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.connect_write_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_128generator20(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_write_pipe", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L10_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2871, __pyx_L1_error)

  /* "uvloop/loop.pyx":2882
 *             WriteUnixTransport transp
 * 
 *         waiter = self._new_future()             # <<<<<<<<<<<<<<
 *         proto = proto_factory()
 *         transp = WriteUnixTransport.new(self, proto, None, waiter)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_waiter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2883
 * 
 *         waiter = self._new_future()
 *         proto = proto_factory()             # <<<<<<<<<<<<<<
 *         transp = WriteUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto_factory);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_proto_factory; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_proto = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2884
 *         waiter = self._new_future()
 *         proto = proto_factory()
 *         transp = WriteUnixTransport.new(self, proto, None, waiter)             # <<<<<<<<<<<<<<
 *         transp._add_extra_info('pipe', pipe)
 *         try:
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_18WriteUnixTransport_new(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_proto, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_cur_scope->__pyx_v_waiter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_transp = ((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2885
 *         proto = proto_factory()
 *         transp = WriteUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)             # <<<<<<<<<<<<<<
 *         try:
 *             transp._open(pipe.fileno())
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._add_extra_info(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_transp), __pyx_n_u_pipe, __pyx_cur_scope->__pyx_v_pipe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2886
 *         transp = WriteUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 *         try:             # <<<<<<<<<<<<<<
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "uvloop/loop.pyx":2887
 *         transp._add_extra_info('pipe', pipe)
 *         try:
 *             transp._open(pipe.fileno())             # <<<<<<<<<<<<<<
 *             transp._init_protocol()
 *             await waiter
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_pipe, __pyx_n_s_fileno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2887, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2887, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2887, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_transp), __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2887, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":2888
 *         try:
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()             # <<<<<<<<<<<<<<
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_transp)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2888, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":2889
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 *             await waiter             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_t_5);
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_5;
        __Pyx_XGIVEREF(__pyx_t_6);
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_6;
        __Pyx_XGIVEREF(__pyx_t_7);
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_7;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L10_resume_from_await:;
        __pyx_t_5 = __pyx_cur_scope->__pyx_t_0;
        __pyx_cur_scope->__pyx_t_0 = 0;
        __Pyx_XGOTREF(__pyx_t_5);
        __pyx_t_6 = __pyx_cur_scope->__pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = 0;
        __Pyx_XGOTREF(__pyx_t_6);
        __pyx_t_7 = __pyx_cur_scope->__pyx_t_2;
        __pyx_cur_scope->__pyx_t_2 = 0;
        __Pyx_XGOTREF(__pyx_t_7);
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2889, __pyx_L4_error)
      } else {
        PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
          else __PYX_ERR(0, 2889, __pyx_L4_error)
        }
      }

      /* "uvloop/loop.pyx":2886
 *         transp = WriteUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 *         try:             # <<<<<<<<<<<<<<
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":2890
 *             transp._init_protocol()
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.Loop.connect_write_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 2890, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":2891
 *             await waiter
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException:
 *             transp._close()
 */
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_2, __pyx_t_3);
      __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0; 
      __PYX_ERR(0, 2891, __pyx_L6_except_error)
    }

    /* "uvloop/loop.pyx":2892
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException:             # <<<<<<<<<<<<<<
 *             transp._close()
 *             raise
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.Loop.connect_write_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0) __PYX_ERR(0, 2892, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);

      /* "uvloop/loop.pyx":2893
 *             raise
 *         except BaseException:
 *             transp._close()             # <<<<<<<<<<<<<<
 *             raise
 *         transp._attach_fileobj(pipe)
 */
      __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_cur_scope->__pyx_v_transp)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2893, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/loop.pyx":2894
 *         except BaseException:
 *             transp._close()
 *             raise             # <<<<<<<<<<<<<<
 *         transp._attach_fileobj(pipe)
 *         return transp, proto
 */
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; 
      __PYX_ERR(0, 2894, __pyx_L6_except_error)
    }
    goto __pyx_L6_except_error;

    /* "uvloop/loop.pyx":2886
 *         transp = WriteUnixTransport.new(self, proto, None, waiter)
 *         transp._add_extra_info('pipe', pipe)
 *         try:             # <<<<<<<<<<<<<<
 *             transp._open(pipe.fileno())
 *             transp._init_protocol()
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "uvloop/loop.pyx":2895
 *             transp._close()
 *             raise
 *         transp._attach_fileobj(pipe)             # <<<<<<<<<<<<<<
 *         return transp, proto
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_cur_scope->__pyx_v_transp->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_transp), __pyx_cur_scope->__pyx_v_pipe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2896
 *             raise
 *         transp._attach_fileobj(pipe)
 *         return transp, proto             # <<<<<<<<<<<<<<
 * 
 *     def add_signal_handler(self, sig, callback, *args):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_transp);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_transp);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_transp))) __PYX_ERR(0, 2896, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_proto);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_proto)) __PYX_ERR(0, 2896, __pyx_L1_error);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":2871
 *         return transp, proto
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_write_pipe(self, proto_factory, pipe):
 *         """Register write pipe in event loop.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("connect_write_pipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2898
 *         return transp, proto
 * 
 *     def add_signal_handler(self, sig, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a handler for a signal.  UNIX only.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_130add_signal_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_129add_signal_handler, "Loop.add_signal_handler(self, sig, callback, *args)\nAdd a handler for a signal.  UNIX only.\n\n        Raise ValueError if the signal number is invalid or uncatchable.\n        Raise RuntimeError if there is a problem setting up the handler.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_130add_signal_handler = {"add_signal_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_130add_signal_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_129add_signal_handler};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_130add_signal_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sig = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_args = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_signal_handler (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_args = __Pyx_ArgsSlice_FASTCALL(__pyx_args, 2, __pyx_nargs);
  if (unlikely(!__pyx_v_args)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_v_args);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sig,&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sig)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2898, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2898, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("add_signal_handler", 0, 2, 2, 1); __PYX_ERR(0, 2898, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 2) ? kwd_pos_args : 2;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "add_signal_handler") < 0)) __PYX_ERR(0, 2898, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_sig = values[0];
    __pyx_v_callback = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_signal_handler", 0, 2, 2, __pyx_nargs); __PYX_ERR(0, 2898, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_AddTraceback("uvloop.loop.Loop.add_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_129add_signal_handler(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sig, __pyx_v_callback, __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_129add_signal_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig, PyObject *__pyx_v_callback, PyObject *__pyx_v_args) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h = 0;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_nexc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_t_20;
  char const *__pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  char const *__pyx_t_28;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_signal_handler", 1);

  /* "uvloop/loop.pyx":2907
 *             Handle h
 * 
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'add_signal_handler() can only be called from '
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__is_main_thread(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":2908
 * 
 *         if not self._is_main_thread():
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'add_signal_handler() can only be called from '
 *                 'the main thread')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__109, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2908, __pyx_L1_error)

    /* "uvloop/loop.pyx":2907
 *             Handle h
 * 
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'add_signal_handler() can only be called from '
 */
  }

  /* "uvloop/loop.pyx":2912
 *                 'the main thread')
 * 
 *         if (aio_iscoroutine(callback)             # <<<<<<<<<<<<<<
 *                 or aio_iscoroutinefunction(callback)):
 *             raise TypeError(
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_iscoroutine);
  __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_iscoroutine; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_callback};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2912, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }

  /* "uvloop/loop.pyx":2913
 * 
 *         if (aio_iscoroutine(callback)
 *                 or aio_iscoroutinefunction(callback)):             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 "coroutines cannot be used with add_signal_handler()")
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_iscoroutinefunction);
  __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_iscoroutinefunction; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_callback};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2913, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2913, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;

  /* "uvloop/loop.pyx":2912
 *                 'the main thread')
 * 
 *         if (aio_iscoroutine(callback)             # <<<<<<<<<<<<<<
 *                 or aio_iscoroutinefunction(callback)):
 *             raise TypeError(
 */
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":2914
 *         if (aio_iscoroutine(callback)
 *                 or aio_iscoroutinefunction(callback)):
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 "coroutines cannot be used with add_signal_handler()")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__110, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2914, __pyx_L1_error)

    /* "uvloop/loop.pyx":2912
 *                 'the main thread')
 * 
 *         if (aio_iscoroutine(callback)             # <<<<<<<<<<<<<<
 *                 or aio_iscoroutinefunction(callback)):
 *             raise TypeError(
 */
  }

  /* "uvloop/loop.pyx":2917
 *                 "coroutines cannot be used with add_signal_handler()")
 * 
 *         if sig == uv.SIGCHLD:             # <<<<<<<<<<<<<<
 *             if (hasattr(callback, '__self__') and
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(SIGCHLD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_sig, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":2918
 * 
 *         if sig == uv.SIGCHLD:
 *             if (hasattr(callback, '__self__') and             # <<<<<<<<<<<<<<
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):
 * 
 */
    __pyx_t_2 = __Pyx_HasAttr(__pyx_v_callback, __pyx_n_u_self_2); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2918, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L9_bool_binop_done;
    }

    /* "uvloop/loop.pyx":2919
 *         if sig == uv.SIGCHLD:
 *             if (hasattr(callback, '__self__') and
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):             # <<<<<<<<<<<<<<
 * 
 *                 warnings_warn(
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_callback, __pyx_n_s_self_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __pyx_v_6uvloop_4loop_aio_AbstractChildWatcher;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_IsInstance(__pyx_t_4, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 2919, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_L9_bool_binop_done:;

    /* "uvloop/loop.pyx":2918
 * 
 *         if sig == uv.SIGCHLD:
 *             if (hasattr(callback, '__self__') and             # <<<<<<<<<<<<<<
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):
 * 
 */
    if (__pyx_t_3) {

      /* "uvloop/loop.pyx":2927
 *                     "The correct way to use uvloop is to install its policy: "
 *                     "`asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())`"
 *                     "\n\n", RuntimeWarning, source=self)             # <<<<<<<<<<<<<<
 * 
 *                 # TODO: ideally we should always raise an error here,
 */
      __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2927, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_source, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 2927, __pyx_L1_error)

      /* "uvloop/loop.pyx":2921
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):
 * 
 *                 warnings_warn(             # <<<<<<<<<<<<<<
 *                     "!!! asyncio is trying to install its ChildWatcher for "
 *                     "SIGCHLD signal !!!\n\nThis is probably because a uvloop "
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_warnings_warn, __pyx_tuple__111, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2921, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/loop.pyx":2934
 *                 # in our README.  Need to start a deprecation period
 *                 # at some point to turn this warning into an error.
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             raise RuntimeError(
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/loop.pyx":2918
 * 
 *         if sig == uv.SIGCHLD:
 *             if (hasattr(callback, '__self__') and             # <<<<<<<<<<<<<<
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):
 * 
 */
    }

    /* "uvloop/loop.pyx":2936
 *                 return
 * 
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'cannot add a signal handler for SIGCHLD: it is used '
 *                 'by the event loop to track subprocesses')
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__112, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2936, __pyx_L1_error)

    /* "uvloop/loop.pyx":2917
 *                 "coroutines cannot be used with add_signal_handler()")
 * 
 *         if sig == uv.SIGCHLD:             # <<<<<<<<<<<<<<
 *             if (hasattr(callback, '__self__') and
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):
 */
  }

  /* "uvloop/loop.pyx":2940
 *                 'by the event loop to track subprocesses')
 * 
 *         self._check_signal(sig)             # <<<<<<<<<<<<<<
 *         self._check_closed()
 * 
 */
  __pyx_t_4 = __pyx_f_6uvloop_4loop_4Loop__check_signal(__pyx_v_self, __pyx_v_sig); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":2941
 * 
 *         self._check_signal(sig)
 *         self._check_closed()             # <<<<<<<<<<<<<<
 * 
 *         h = new_Handle(self, callback, args or None, None)
 */
  __pyx_t_4 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":2943
 *         self._check_closed()
 * 
 *         h = new_Handle(self, callback, args or None, None)             # <<<<<<<<<<<<<<
 *         self._signal_handlers[sig] = h
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_args); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2943, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __Pyx_INCREF(__pyx_v_args);
    __pyx_t_4 = __pyx_v_args;
    goto __pyx_L11_bool_binop_done;
  }
  __Pyx_INCREF(Py_None);
  __pyx_t_4 = Py_None;
  __pyx_L11_bool_binop_done:;
  __pyx_t_1 = __pyx_f_6uvloop_4loop_new_Handle(__pyx_v_self, __pyx_v_callback, __pyx_t_4, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(0, 2943, __pyx_L1_error)
  __pyx_v_h = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2944
 * 
 *         h = new_Handle(self, callback, args or None, None)
 *         self._signal_handlers[sig] = h             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  if (unlikely(__pyx_v_self->_signal_handlers == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 2944, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->_signal_handlers, __pyx_v_sig, ((PyObject *)__pyx_v_h)) < 0))) __PYX_ERR(0, 2944, __pyx_L1_error)

  /* "uvloop/loop.pyx":2946
 *         self._signal_handlers[sig] = h
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             # Register a dummy signal handler to ask Python to write the signal
 *             # number in the wakeup file descriptor.
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "uvloop/loop.pyx":2949
 *             # Register a dummy signal handler to ask Python to write the signal
 *             # number in the wakeup file descriptor.
 *             signal_signal(sig, self.__sighandler)             # <<<<<<<<<<<<<<
 * 
 *             # Set SA_RESTART to limit EINTR occurrences.
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Loop__sighandler); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2949, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_signal_signal);
      __pyx_t_5 = __pyx_v_6uvloop_4loop_signal_signal; __pyx_t_10 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_10, __pyx_v_sig, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2949, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":2952
 * 
 *             # Set SA_RESTART to limit EINTR occurrences.
 *             signal_siginterrupt(sig, False)             # <<<<<<<<<<<<<<
 *         except OSError as exc:
 *             del self._signal_handlers[sig]
 */
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_signal_siginterrupt);
      __pyx_t_5 = __pyx_v_6uvloop_4loop_signal_siginterrupt; __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_sig, Py_False};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2952, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":2946
 *         self._signal_handlers[sig] = h
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             # Register a dummy signal handler to ask Python to write the signal
 *             # number in the wakeup file descriptor.
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L18_try_end;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/loop.pyx":2953
 *             # Set SA_RESTART to limit EINTR occurrences.
 *             signal_siginterrupt(sig, False)
 *         except OSError as exc:             # <<<<<<<<<<<<<<
 *             del self._signal_handlers[sig]
 *             if not self._signal_handlers:
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.Loop.add_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 2953, __pyx_L15_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_exc = __pyx_t_5;
      /*try:*/ {

        /* "uvloop/loop.pyx":2954
 *             signal_siginterrupt(sig, False)
 *         except OSError as exc:
 *             del self._signal_handlers[sig]             # <<<<<<<<<<<<<<
 *             if not self._signal_handlers:
 *                 try:
 */
        if (unlikely(__pyx_v_self->_signal_handlers == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 2954, __pyx_L24_error)
        }
        if (unlikely((PyDict_DelItem(__pyx_v_self->_signal_handlers, __pyx_v_sig) < 0))) __PYX_ERR(0, 2954, __pyx_L24_error)

        /* "uvloop/loop.pyx":2955
 *         except OSError as exc:
 *             del self._signal_handlers[sig]
 *             if not self._signal_handlers:             # <<<<<<<<<<<<<<
 *                 try:
 *                     signal_set_wakeup_fd(-1)
 */
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_self->_signal_handlers); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2955, __pyx_L24_error)
        __pyx_t_2 = (!__pyx_t_3);
        if (__pyx_t_2) {

          /* "uvloop/loop.pyx":2956
 *             del self._signal_handlers[sig]
 *             if not self._signal_handlers:
 *                 try:             # <<<<<<<<<<<<<<
 *                     signal_set_wakeup_fd(-1)
 *                 except (ValueError, OSError) as nexc:
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            /*try:*/ {

              /* "uvloop/loop.pyx":2957
 *             if not self._signal_handlers:
 *                 try:
 *                     signal_set_wakeup_fd(-1)             # <<<<<<<<<<<<<<
 *                 except (ValueError, OSError) as nexc:
 *                     aio_logger.info('set_wakeup_fd(-1) failed: %s', nexc)
 */
              __Pyx_INCREF(__pyx_v_6uvloop_4loop_signal_set_wakeup_fd);
              __pyx_t_15 = __pyx_v_6uvloop_4loop_signal_set_wakeup_fd; __pyx_t_16 = NULL;
              __pyx_t_6 = 0;
              #if CYTHON_UNPACK_METHODS
              if (unlikely(PyMethod_Check(__pyx_t_15))) {
                __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
                if (likely(__pyx_t_16)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                  __Pyx_INCREF(__pyx_t_16);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_15, function);
                  __pyx_t_6 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_int_neg_1};
                __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2957, __pyx_L27_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

              /* "uvloop/loop.pyx":2956
 *             del self._signal_handlers[sig]
 *             if not self._signal_handlers:
 *                 try:             # <<<<<<<<<<<<<<
 *                     signal_set_wakeup_fd(-1)
 *                 except (ValueError, OSError) as nexc:
 */
            }
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            goto __pyx_L34_try_end;
            __pyx_L27_error:;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;

            /* "uvloop/loop.pyx":2958
 *                 try:
 *                     signal_set_wakeup_fd(-1)
 *                 except (ValueError, OSError) as nexc:             # <<<<<<<<<<<<<<
 *                     aio_logger.info('set_wakeup_fd(-1) failed: %s', nexc)
 * 
 */
            __pyx_t_11 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_ValueError, __pyx_builtin_OSError);
            if (__pyx_t_11) {
              __Pyx_AddTraceback("uvloop.loop.Loop.add_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_15, &__pyx_t_16) < 0) __PYX_ERR(0, 2958, __pyx_L29_except_error)
              __Pyx_XGOTREF(__pyx_t_10);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_15);
              __pyx_v_nexc = __pyx_t_15;
              /*try:*/ {

                /* "uvloop/loop.pyx":2959
 *                     signal_set_wakeup_fd(-1)
 *                 except (ValueError, OSError) as nexc:
 *                     aio_logger.info('set_wakeup_fd(-1) failed: %s', nexc)             # <<<<<<<<<<<<<<
 * 
 *             if exc.errno == errno_EINVAL:
 */
                __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_info); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 2959, __pyx_L40_error)
                __Pyx_GOTREF(__pyx_t_18);
                __pyx_t_19 = NULL;
                __pyx_t_6 = 0;
                #if CYTHON_UNPACK_METHODS
                if (likely(PyMethod_Check(__pyx_t_18))) {
                  __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_18);
                  if (likely(__pyx_t_19)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                    __Pyx_INCREF(__pyx_t_19);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_18, function);
                    __pyx_t_6 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[3] = {__pyx_t_19, __pyx_kp_u_set_wakeup_fd_1_failed_s, __pyx_v_nexc};
                  __pyx_t_17 = __Pyx_PyObject_FastCall(__pyx_t_18, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
                  __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                  if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 2959, __pyx_L40_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                }
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              }

              /* "uvloop/loop.pyx":2958
 *                 try:
 *                     signal_set_wakeup_fd(-1)
 *                 except (ValueError, OSError) as nexc:             # <<<<<<<<<<<<<<
 *                     aio_logger.info('set_wakeup_fd(-1) failed: %s', nexc)
 * 
 */
              /*finally:*/ {
                /*normal exit:*/{
                  __Pyx_DECREF(__pyx_v_nexc); __pyx_v_nexc = 0;
                  goto __pyx_L41;
                }
                __pyx_L40_error:;
                /*exception exit:*/{
                  __Pyx_PyThreadState_declare
                  __Pyx_PyThreadState_assign
                  __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
                  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                  __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                  __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                  if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_25, &__pyx_t_26, &__pyx_t_27);
                  if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24) < 0)) __Pyx_ErrFetch(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
                  __Pyx_XGOTREF(__pyx_t_22);
                  __Pyx_XGOTREF(__pyx_t_23);
                  __Pyx_XGOTREF(__pyx_t_24);
                  __Pyx_XGOTREF(__pyx_t_25);
                  __Pyx_XGOTREF(__pyx_t_26);
                  __Pyx_XGOTREF(__pyx_t_27);
                  __pyx_t_11 = __pyx_lineno; __pyx_t_20 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
                  {
                    __Pyx_DECREF(__pyx_v_nexc); __pyx_v_nexc = 0;
                  }
                  if (PY_MAJOR_VERSION >= 3) {
                    __Pyx_XGIVEREF(__pyx_t_25);
                    __Pyx_XGIVEREF(__pyx_t_26);
                    __Pyx_XGIVEREF(__pyx_t_27);
                    __Pyx_ExceptionReset(__pyx_t_25, __pyx_t_26, __pyx_t_27);
                  }
                  __Pyx_XGIVEREF(__pyx_t_22);
                  __Pyx_XGIVEREF(__pyx_t_23);
                  __Pyx_XGIVEREF(__pyx_t_24);
                  __Pyx_ErrRestore(__pyx_t_22, __pyx_t_23, __pyx_t_24);
                  __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0;
                  __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_20; __pyx_filename = __pyx_t_21;
                  goto __pyx_L29_except_error;
                }
                __pyx_L41:;
              }
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              goto __pyx_L28_exception_handled;
            }
            goto __pyx_L29_except_error;

            /* "uvloop/loop.pyx":2956
 *             del self._signal_handlers[sig]
 *             if not self._signal_handlers:
 *                 try:             # <<<<<<<<<<<<<<
 *                     signal_set_wakeup_fd(-1)
 *                 except (ValueError, OSError) as nexc:
 */
            __pyx_L29_except_error:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            goto __pyx_L24_error;
            __pyx_L28_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            __pyx_L34_try_end:;
          }

          /* "uvloop/loop.pyx":2955
 *         except OSError as exc:
 *             del self._signal_handlers[sig]
 *             if not self._signal_handlers:             # <<<<<<<<<<<<<<
 *                 try:
 *                     signal_set_wakeup_fd(-1)
 */
        }

        /* "uvloop/loop.pyx":2961
 *                     aio_logger.info('set_wakeup_fd(-1) failed: %s', nexc)
 * 
 *             if exc.errno == errno_EINVAL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 *             else:
 */
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_errno); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 2961, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_15 = PyObject_RichCompare(__pyx_t_16, __pyx_v_6uvloop_4loop_errno_EINVAL, Py_EQ); __Pyx_XGOTREF(__pyx_t_15); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2961, __pyx_L24_error)
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_15); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2961, __pyx_L24_error)
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(__pyx_t_2)) {

          /* "uvloop/loop.pyx":2962
 * 
 *             if exc.errno == errno_EINVAL:
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))             # <<<<<<<<<<<<<<
 *             else:
 *                 raise
 */
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_sig_cannot_be_caught, __pyx_n_s_format); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 2962, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_10 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_sig};
            __pyx_t_15 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 2962, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __pyx_t_16 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 2962, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_Raise(__pyx_t_16, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __PYX_ERR(0, 2962, __pyx_L24_error)

          /* "uvloop/loop.pyx":2961
 *                     aio_logger.info('set_wakeup_fd(-1) failed: %s', nexc)
 * 
 *             if exc.errno == errno_EINVAL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 *             else:
 */
        }

        /* "uvloop/loop.pyx":2964
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 *             else:
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *     def remove_signal_handler(self, sig):
 */
        /*else*/ {
          __Pyx_GIVEREF(__pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_5, __pyx_t_4);
          __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
          __PYX_ERR(0, 2964, __pyx_L24_error)
        }
      }

      /* "uvloop/loop.pyx":2953
 *             # Set SA_RESTART to limit EINTR occurrences.
 *             signal_siginterrupt(sig, False)
 *         except OSError as exc:             # <<<<<<<<<<<<<<
 *             del self._signal_handlers[sig]
 *             if not self._signal_handlers:
 */
      /*finally:*/ {
        __pyx_L24_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_27 = 0; __pyx_t_26 = 0; __pyx_t_25 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_27, &__pyx_t_26, &__pyx_t_25);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_27);
          __Pyx_XGOTREF(__pyx_t_26);
          __Pyx_XGOTREF(__pyx_t_25);
          __pyx_t_20 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_28 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_27);
            __Pyx_XGIVEREF(__pyx_t_26);
            __Pyx_XGIVEREF(__pyx_t_25);
            __Pyx_ExceptionReset(__pyx_t_27, __pyx_t_26, __pyx_t_25);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_13, __pyx_t_12);
          __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_27 = 0; __pyx_t_26 = 0; __pyx_t_25 = 0;
          __pyx_lineno = __pyx_t_20; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_28;
          goto __pyx_L15_except_error;
        }
      }
    }
    goto __pyx_L15_except_error;

    /* "uvloop/loop.pyx":2946
 *         self._signal_handlers[sig] = h
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             # Register a dummy signal handler to ask Python to write the signal
 *             # number in the wakeup file descriptor.
 */
    __pyx_L15_except_error:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L18_try_end:;
  }

  /* "uvloop/loop.pyx":2898
 *         return transp, proto
 * 
 *     def add_signal_handler(self, sig, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a handler for a signal.  UNIX only.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("uvloop.loop.Loop.add_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_h);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_nexc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":2966
 *                 raise
 * 
 *     def remove_signal_handler(self, sig):             # <<<<<<<<<<<<<<
 *         """Remove a handler for a signal.  UNIX only.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_132remove_signal_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_131remove_signal_handler, "Loop.remove_signal_handler(self, sig)\nRemove a handler for a signal.  UNIX only.\n\n        Return True if a signal handler was removed, False if not.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_132remove_signal_handler = {"remove_signal_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_132remove_signal_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_131remove_signal_handler};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_132remove_signal_handler(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_sig = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_signal_handler (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sig,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sig)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2966, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove_signal_handler") < 0)) __PYX_ERR(0, 2966, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_sig = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove_signal_handler", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2966, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.remove_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_131remove_signal_handler(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_sig);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_131remove_signal_handler(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_sig) {
  PyObject *__pyx_v_handler = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove_signal_handler", 1);

  /* "uvloop/loop.pyx":2972
 *         """
 * 
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'remove_signal_handler() can only be called from '
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__is_main_thread(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!__pyx_t_2);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/loop.pyx":2973
 * 
 *         if not self._is_main_thread():
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'remove_signal_handler() can only be called from '
 *                 'the main thread')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__113, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 2973, __pyx_L1_error)

    /* "uvloop/loop.pyx":2972
 *         """
 * 
 *         if not self._is_main_thread():             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'remove_signal_handler() can only be called from '
 */
  }

  /* "uvloop/loop.pyx":2977
 *                 'the main thread')
 * 
 *         self._check_signal(sig)             # <<<<<<<<<<<<<<
 * 
 *         if not self._listening_signals:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_signal(__pyx_v_self, __pyx_v_sig); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":2979
 *         self._check_signal(sig)
 * 
 *         if not self._listening_signals:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_3 = (!__pyx_v_self->_listening_signals);
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":2980
 * 
 *         if not self._listening_signals:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":2979
 *         self._check_signal(sig)
 * 
 *         if not self._listening_signals:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "uvloop/loop.pyx":2982
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             del self._signal_handlers[sig]
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "uvloop/loop.pyx":2983
 * 
 *         try:
 *             del self._signal_handlers[sig]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             return False
 */
      if (unlikely(__pyx_v_self->_signal_handlers == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 2983, __pyx_L5_error)
      }
      if (unlikely((PyDict_DelItem(__pyx_v_self->_signal_handlers, __pyx_v_sig) < 0))) __PYX_ERR(0, 2983, __pyx_L5_error)

      /* "uvloop/loop.pyx":2982
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             del self._signal_handlers[sig]
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":2984
 *         try:
 *             del self._signal_handlers[sig]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("uvloop.loop.Loop.remove_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(0, 2984, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);

      /* "uvloop/loop.pyx":2985
 *             del self._signal_handlers[sig]
 *         except KeyError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         if sig == uv.SIGINT:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L8_except_return;
    }
    goto __pyx_L7_except_error;

    /* "uvloop/loop.pyx":2982
 *             return False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             del self._signal_handlers[sig]
 *         except KeyError:
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L8_except_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
    __pyx_L10_try_end:;
  }

  /* "uvloop/loop.pyx":2987
 *             return False
 * 
 *         if sig == uv.SIGINT:             # <<<<<<<<<<<<<<
 *             handler = signal_default_int_handler
 *         else:
 */
  __pyx_t_9 = __Pyx_PyInt_From_int(SIGINT); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = PyObject_RichCompare(__pyx_v_sig, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2987, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2987, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":2988
 * 
 *         if sig == uv.SIGINT:
 *             handler = signal_default_int_handler             # <<<<<<<<<<<<<<
 *         else:
 *             handler = signal_SIG_DFL
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_signal_default_int_handler);
    __pyx_v_handler = __pyx_v_6uvloop_4loop_signal_default_int_handler;

    /* "uvloop/loop.pyx":2987
 *             return False
 * 
 *         if sig == uv.SIGINT:             # <<<<<<<<<<<<<<
 *             handler = signal_default_int_handler
 *         else:
 */
    goto __pyx_L13;
  }

  /* "uvloop/loop.pyx":2990
 *             handler = signal_default_int_handler
 *         else:
 *             handler = signal_SIG_DFL             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_signal_SIG_DFL);
    __pyx_v_handler = __pyx_v_6uvloop_4loop_signal_SIG_DFL;
  }
  __pyx_L13:;

  /* "uvloop/loop.pyx":2992
 *             handler = signal_SIG_DFL
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             signal_signal(sig, handler)
 *         except OSError as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/loop.pyx":2993
 * 
 *         try:
 *             signal_signal(sig, handler)             # <<<<<<<<<<<<<<
 *         except OSError as exc:
 *             if exc.errno == errno_EINVAL:
 */
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_signal_signal);
      __pyx_t_9 = __pyx_v_6uvloop_4loop_signal_signal; __pyx_t_1 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_sig, __pyx_v_handler};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_10, 2+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2993, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/loop.pyx":2992
 *             handler = signal_SIG_DFL
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             signal_signal(sig, handler)
 *         except OSError as exc:
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L19_try_end;
    __pyx_L14_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "uvloop/loop.pyx":2994
 *         try:
 *             signal_signal(sig, handler)
 *         except OSError as exc:             # <<<<<<<<<<<<<<
 *             if exc.errno == errno_EINVAL:
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("uvloop.loop.Loop.remove_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_1) < 0) __PYX_ERR(0, 2994, __pyx_L16_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_9);
      __pyx_v_exc = __pyx_t_9;
      /*try:*/ {

        /* "uvloop/loop.pyx":2995
 *             signal_signal(sig, handler)
 *         except OSError as exc:
 *             if exc.errno == errno_EINVAL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 *             else:
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc, __pyx_n_s_errno); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2995, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = PyObject_RichCompare(__pyx_t_11, __pyx_v_6uvloop_4loop_errno_EINVAL, Py_EQ); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2995, __pyx_L25_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2995, __pyx_L25_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(__pyx_t_3)) {

          /* "uvloop/loop.pyx":2996
 *         except OSError as exc:
 *             if exc.errno == errno_EINVAL:
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))             # <<<<<<<<<<<<<<
 *             else:
 *                 raise
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_sig_cannot_be_caught, __pyx_n_s_format); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2996, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_13 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_v_sig};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 2996, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2996, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_Raise(__pyx_t_11, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __PYX_ERR(0, 2996, __pyx_L25_error)

          /* "uvloop/loop.pyx":2995
 *             signal_signal(sig, handler)
 *         except OSError as exc:
 *             if exc.errno == errno_EINVAL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 *             else:
 */
        }

        /* "uvloop/loop.pyx":2998
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 *             else:
 *                 raise             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
        /*else*/ {
          __Pyx_GIVEREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_1);
          __Pyx_ErrRestoreWithState(__pyx_t_8, __pyx_t_9, __pyx_t_1);
          __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_1 = 0; 
          __PYX_ERR(0, 2998, __pyx_L25_error)
        }
      }

      /* "uvloop/loop.pyx":2994
 *         try:
 *             signal_signal(sig, handler)
 *         except OSError as exc:             # <<<<<<<<<<<<<<
 *             if exc.errno == errno_EINVAL:
 *                 raise RuntimeError('sig {} cannot be caught'.format(sig))
 */
      /*finally:*/ {
        __pyx_L25_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __pyx_t_7 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
          }
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
          goto __pyx_L16_except_error;
        }
      }
    }
    goto __pyx_L16_except_error;

    /* "uvloop/loop.pyx":2992
 *             handler = signal_SIG_DFL
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             signal_signal(sig, handler)
 *         except OSError as exc:
 */
    __pyx_L16_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L19_try_end:;
  }

  /* "uvloop/loop.pyx":3000
 *                 raise
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":2966
 *                 raise
 * 
 *     def remove_signal_handler(self, sig):             # <<<<<<<<<<<<<<
 *         """Remove a handler for a signal.  UNIX only.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("uvloop.loop.Loop.remove_signal_handler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_handler);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_135generator21(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_134create_datagram_endpoint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_133create_datagram_endpoint, "Loop.create_datagram_endpoint(self, protocol_factory, local_addr=None, remote_addr=None, *, family=0, proto=0, flags=0, reuse_address=_unset, reuse_port=None, allow_broadcast=None, sock=None)\nA coroutine which creates a datagram endpoint.\n\n        This method will try to establish the endpoint in the background.\n        When successful, the coroutine returns a (transport, protocol) pair.\n\n        protocol_factory must be a callable returning a protocol instance.\n\n        socket family AF_INET or socket.AF_INET6 depending on host (or\n        family if specified), socket type SOCK_DGRAM.\n\n        reuse_port tells the kernel to allow this endpoint to be bound to\n        the same port as other existing endpoints are bound to, so long as\n        they all set this flag when being created. This option is not\n        supported on Windows and some UNIX's. If the\n        :py:data:`~socket.SO_REUSEPORT` constant is not defined then this\n        capability is unsupported.\n\n        allow_broadcast tells the kernel to allow this endpoint to send\n        messages to the broadcast address.\n\n        sock can optionally be specified in order to use a preexisting\n        socket object.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_134create_datagram_endpoint = {"create_datagram_endpoint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_134create_datagram_endpoint, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_133create_datagram_endpoint};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_134create_datagram_endpoint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol_factory = 0;
  PyObject *__pyx_v_local_addr = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_family = 0;
  PyObject *__pyx_v_proto = 0;
  PyObject *__pyx_v_flags = 0;
  PyObject *__pyx_v_reuse_address = 0;
  PyObject *__pyx_v_reuse_port = 0;
  PyObject *__pyx_v_allow_broadcast = 0;
  PyObject *__pyx_v_sock = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_datagram_endpoint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol_factory,&__pyx_n_s_local_addr,&__pyx_n_s_remote_addr,&__pyx_n_s_family,&__pyx_n_s_proto,&__pyx_n_s_flags,&__pyx_n_s_reuse_address,&__pyx_n_s_reuse_port,&__pyx_n_s_allow_broadcast,&__pyx_n_s_sock,0};

    /* "uvloop/loop.pyx":3004
 *     @cython.iterable_coroutine
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,             # <<<<<<<<<<<<<<
 *                                        family=0, proto=0, flags=0,
 *                                        reuse_address=_unset, reuse_port=None,
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_0));
    values[6] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__114);

    /* "uvloop/loop.pyx":3006
 *                                        local_addr=None, remote_addr=None, *,
 *                                        family=0, proto=0, flags=0,
 *                                        reuse_address=_unset, reuse_port=None,             # <<<<<<<<<<<<<<
 *                                        allow_broadcast=None, sock=None):
 *         """A coroutine which creates a datagram endpoint.
 */
    values[7] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));

    /* "uvloop/loop.pyx":3007
 *                                        family=0, proto=0, flags=0,
 *                                        reuse_address=_unset, reuse_port=None,
 *                                        allow_broadcast=None, sock=None):             # <<<<<<<<<<<<<<
 *         """A coroutine which creates a datagram endpoint.
 * 
 */
    values[8] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[9] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol_factory)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3002, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_local_addr);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3002, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_remote_addr);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3002, __pyx_L3_error)
        }
      }
      if (kw_args > 0 && likely(kw_args <= 7)) {
        Py_ssize_t index;
        for (index = 3; index < 10 && kw_args > 0; index++) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, *__pyx_pyargnames[index]);
          if (value) { values[index] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3002, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_datagram_endpoint") < 0)) __PYX_ERR(0, 3002, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_protocol_factory = values[0];
    __pyx_v_local_addr = values[1];
    __pyx_v_remote_addr = values[2];
    __pyx_v_family = values[3];
    __pyx_v_proto = values[4];
    __pyx_v_flags = values[5];
    __pyx_v_reuse_address = values[6];
    __pyx_v_reuse_port = values[7];
    __pyx_v_allow_broadcast = values[8];
    __pyx_v_sock = values[9];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_datagram_endpoint", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 3002, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.create_datagram_endpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_133create_datagram_endpoint(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_protocol_factory, __pyx_v_local_addr, __pyx_v_remote_addr, __pyx_v_family, __pyx_v_proto, __pyx_v_flags, __pyx_v_reuse_address, __pyx_v_reuse_port, __pyx_v_allow_broadcast, __pyx_v_sock);

  /* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_24create_datagram_endpoint_2generator32(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":3049
 *                             allow_broadcast=allow_broadcast)
 *                 problems = ', '.join(
 *                     '{}={}'.format(k, v) for k, v in opts.items() if v)             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'socket modifier keyword arguments can not be used '
 */

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_24create_datagram_endpoint_genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_28_genexpr(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3049, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_24create_datagram_endpoint_2generator32, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_create_datagram_endpoint_locals, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 3049, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_datagram_endpoint.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_24create_datagram_endpoint_2generator32(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3049, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 3049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 3049, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_genexpr_arg_0, 1, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 3049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_v); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 3049, __pyx_L1_error)
    if (__pyx_t_8) {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__115, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_cur_scope->__pyx_v_k, __pyx_cur_scope->__pyx_v_v};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_10, 2+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3049, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 3049, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_133create_datagram_endpoint(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_protocol_factory, PyObject *__pyx_v_local_addr, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_family, PyObject *__pyx_v_proto, PyObject *__pyx_v_flags, PyObject *__pyx_v_reuse_address, PyObject *__pyx_v_reuse_port, PyObject *__pyx_v_allow_broadcast, PyObject *__pyx_v_sock) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *__pyx_cur_scope;
  PyObject *__pyx_gb_6uvloop_4loop_4Loop_24create_datagram_endpoint_2generator32 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_datagram_endpoint", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3002, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_protocol_factory = __pyx_v_protocol_factory;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_cur_scope->__pyx_v_local_addr = __pyx_v_local_addr;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_local_addr);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_local_addr);
  __pyx_cur_scope->__pyx_v_remote_addr = __pyx_v_remote_addr;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_remote_addr);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_remote_addr);
  __pyx_cur_scope->__pyx_v_family = __pyx_v_family;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_family);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_family);
  __pyx_cur_scope->__pyx_v_proto = __pyx_v_proto;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_proto);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_proto);
  __pyx_cur_scope->__pyx_v_flags = __pyx_v_flags;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_flags);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_flags);
  __pyx_cur_scope->__pyx_v_reuse_address = __pyx_v_reuse_address;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_reuse_address);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_reuse_address);
  __pyx_cur_scope->__pyx_v_reuse_port = __pyx_v_reuse_port;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_reuse_port);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_reuse_port);
  __pyx_cur_scope->__pyx_v_allow_broadcast = __pyx_v_allow_broadcast;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_allow_broadcast);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_allow_broadcast);
  __pyx_cur_scope->__pyx_v_sock = __pyx_v_sock;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sock);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sock);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_135generator21, __pyx_codeobj__116, (PyObject *) __pyx_cur_scope, __pyx_n_s_create_datagram_endpoint, __pyx_n_s_Loop_create_datagram_endpoint, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 3002, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.create_datagram_endpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XDECREF(__pyx_gb_6uvloop_4loop_4Loop_24create_datagram_endpoint_2generator32);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_135generator21(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  struct addrinfo *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_datagram_endpoint", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L25_resume_from_await;
    case 2: goto __pyx_L32_resume_from_await;
    case 3: goto __pyx_L70_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3002, __pyx_L1_error)

  /* "uvloop/loop.pyx":3032
 *         """
 *         cdef:
 *             UDPTransport udp = None             # <<<<<<<<<<<<<<
 *             system.addrinfo * lai
 *             system.addrinfo * rai
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_udp = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)Py_None);

  /* "uvloop/loop.pyx":3036
 *             system.addrinfo * rai
 * 
 *         if sock is not None:             # <<<<<<<<<<<<<<
 *             if not _is_sock_dgram(sock.type):
 *                 raise ValueError(
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_sock != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3037
 * 
 *         if sock is not None:
 *             if not _is_sock_dgram(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A UDP Socket was expected, got {!r}'.format(sock))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_f_6uvloop_4loop__is_sock_dgram(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3037, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (!__pyx_t_1);
    if (unlikely(__pyx_t_4)) {

      /* "uvloop/loop.pyx":3039
 *             if not _is_sock_dgram(sock.type):
 *                 raise ValueError(
 *                     'A UDP Socket was expected, got {!r}'.format(sock))             # <<<<<<<<<<<<<<
 *             if (local_addr or remote_addr or
 *                     family or proto or flags or
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_UDP_Socket_was_expected_got_r, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3039, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_sock};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3039, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }

      /* "uvloop/loop.pyx":3038
 *         if sock is not None:
 *             if not _is_sock_dgram(sock.type):
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'A UDP Socket was expected, got {!r}'.format(sock))
 *             if (local_addr or remote_addr or
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3038, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 3038, __pyx_L1_error)

      /* "uvloop/loop.pyx":3037
 * 
 *         if sock is not None:
 *             if not _is_sock_dgram(sock.type):             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'A UDP Socket was expected, got {!r}'.format(sock))
 */
    }

    /* "uvloop/loop.pyx":3040
 *                 raise ValueError(
 *                     'A UDP Socket was expected, got {!r}'.format(sock))
 *             if (local_addr or remote_addr or             # <<<<<<<<<<<<<<
 *                     family or proto or flags or
 *                     reuse_port or allow_broadcast):
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_local_addr); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3040, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_remote_addr); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3040, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L7_bool_binop_done;
    }

    /* "uvloop/loop.pyx":3041
 *                     'A UDP Socket was expected, got {!r}'.format(sock))
 *             if (local_addr or remote_addr or
 *                     family or proto or flags or             # <<<<<<<<<<<<<<
 *                     reuse_port or allow_broadcast):
 *                 # show the problematic kwargs in exception msg
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3041, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_proto); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3041, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_flags); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3041, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L7_bool_binop_done;
    }

    /* "uvloop/loop.pyx":3042
 *             if (local_addr or remote_addr or
 *                     family or proto or flags or
 *                     reuse_port or allow_broadcast):             # <<<<<<<<<<<<<<
 *                 # show the problematic kwargs in exception msg
 *                 opts = dict(local_addr=local_addr, remote_addr=remote_addr,
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3042, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_allow_broadcast); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3042, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_1;
    __pyx_L7_bool_binop_done:;

    /* "uvloop/loop.pyx":3040
 *                 raise ValueError(
 *                     'A UDP Socket was expected, got {!r}'.format(sock))
 *             if (local_addr or remote_addr or             # <<<<<<<<<<<<<<
 *                     family or proto or flags or
 *                     reuse_port or allow_broadcast):
 */
    if (unlikely(__pyx_t_4)) {

      /* "uvloop/loop.pyx":3044
 *                     reuse_port or allow_broadcast):
 *                 # show the problematic kwargs in exception msg
 *                 opts = dict(local_addr=local_addr, remote_addr=remote_addr,             # <<<<<<<<<<<<<<
 *                             family=family, proto=proto, flags=flags,
 *                             reuse_address=reuse_address, reuse_port=reuse_port,
 */
      __pyx_t_2 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_local_addr, __pyx_cur_scope->__pyx_v_local_addr) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_remote_addr, __pyx_cur_scope->__pyx_v_remote_addr) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)

      /* "uvloop/loop.pyx":3045
 *                 # show the problematic kwargs in exception msg
 *                 opts = dict(local_addr=local_addr, remote_addr=remote_addr,
 *                             family=family, proto=proto, flags=flags,             # <<<<<<<<<<<<<<
 *                             reuse_address=reuse_address, reuse_port=reuse_port,
 *                             allow_broadcast=allow_broadcast)
 */
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_family, __pyx_cur_scope->__pyx_v_family) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_proto, __pyx_cur_scope->__pyx_v_proto) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_flags, __pyx_cur_scope->__pyx_v_flags) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)

      /* "uvloop/loop.pyx":3046
 *                 opts = dict(local_addr=local_addr, remote_addr=remote_addr,
 *                             family=family, proto=proto, flags=flags,
 *                             reuse_address=reuse_address, reuse_port=reuse_port,             # <<<<<<<<<<<<<<
 *                             allow_broadcast=allow_broadcast)
 *                 problems = ', '.join(
 */
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_reuse_address, __pyx_cur_scope->__pyx_v_reuse_address) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_reuse_port, __pyx_cur_scope->__pyx_v_reuse_port) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)

      /* "uvloop/loop.pyx":3047
 *                             family=family, proto=proto, flags=flags,
 *                             reuse_address=reuse_address, reuse_port=reuse_port,
 *                             allow_broadcast=allow_broadcast)             # <<<<<<<<<<<<<<
 *                 problems = ', '.join(
 *                     '{}={}'.format(k, v) for k, v in opts.items() if v)
 */
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_allow_broadcast, __pyx_cur_scope->__pyx_v_allow_broadcast) < 0) __PYX_ERR(0, 3044, __pyx_L1_error)
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_v_opts = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":3049
 *                             allow_broadcast=allow_broadcast)
 *                 problems = ', '.join(
 *                     '{}={}'.format(k, v) for k, v in opts.items() if v)             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'socket modifier keyword arguments can not be used '
 */
      __pyx_t_2 = __pyx_pf_6uvloop_4loop_4Loop_24create_datagram_endpoint_genexpr(NULL, __pyx_cur_scope->__pyx_v_opts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "uvloop/loop.pyx":3048
 *                             reuse_address=reuse_address, reuse_port=reuse_port,
 *                             allow_broadcast=allow_broadcast)
 *                 problems = ', '.join(             # <<<<<<<<<<<<<<
 *                     '{}={}'.format(k, v) for k, v in opts.items() if v)
 *                 raise ValueError(
 */
      __pyx_t_3 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__69, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_cur_scope->__pyx_v_problems = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":3052
 *                 raise ValueError(
 *                     'socket modifier keyword arguments can not be used '
 *                     'when sock is specified. ({})'.format(problems))             # <<<<<<<<<<<<<<
 *             sock.setblocking(False)
 *             udp = UDPTransport.__new__(UDPTransport)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_socket_modifier_keyword_argument, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_problems};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3052, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/loop.pyx":3050
 *                 problems = ', '.join(
 *                     '{}={}'.format(k, v) for k, v in opts.items() if v)
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'socket modifier keyword arguments can not be used '
 *                     'when sock is specified. ({})'.format(problems))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3050, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 3050, __pyx_L1_error)

      /* "uvloop/loop.pyx":3040
 *                 raise ValueError(
 *                     'A UDP Socket was expected, got {!r}'.format(sock))
 *             if (local_addr or remote_addr or             # <<<<<<<<<<<<<<
 *                     family or proto or flags or
 *                     reuse_port or allow_broadcast):
 */
    }

    /* "uvloop/loop.pyx":3053
 *                     'socket modifier keyword arguments can not be used '
 *                     'when sock is specified. ({})'.format(problems))
 *             sock.setblocking(False)             # <<<<<<<<<<<<<<
 *             udp = UDPTransport.__new__(UDPTransport)
 *             udp._init(self, uv.AF_UNSPEC)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_setblocking); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_False};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3053, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":3054
 *                     'when sock is specified. ({})'.format(problems))
 *             sock.setblocking(False)
 *             udp = UDPTransport.__new__(UDPTransport)             # <<<<<<<<<<<<<<
 *             udp._init(self, uv.AF_UNSPEC)
 *             udp.open(sock.family, sock.fileno())
 */
    __pyx_t_3 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UDPTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UDPTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3054, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_udp);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_udp, ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_t_3));
    __Pyx_GIVEREF((PyObject *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":3055
 *             sock.setblocking(False)
 *             udp = UDPTransport.__new__(UDPTransport)
 *             udp._init(self, uv.AF_UNSPEC)             # <<<<<<<<<<<<<<
 *             udp.open(sock.family, sock.fileno())
 *             udp._attach_fileobj(sock)
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_init(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_self, AF_UNSPEC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":3056
 *             udp = UDPTransport.__new__(UDPTransport)
 *             udp._init(self, uv.AF_UNSPEC)
 *             udp.open(sock.family, sock.fileno())             # <<<<<<<<<<<<<<
 *             udp._attach_fileobj(sock)
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_family); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3056, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_sock, __pyx_n_s_fileno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3056, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->open(__pyx_cur_scope->__pyx_v_udp, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":3057
 *             udp._init(self, uv.AF_UNSPEC)
 *             udp.open(sock.family, sock.fileno())
 *             udp._attach_fileobj(sock)             # <<<<<<<<<<<<<<
 *         else:
 *             if reuse_address is not _unset:
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._attach_fileobj(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_udp), __pyx_cur_scope->__pyx_v_sock); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":3036
 *             system.addrinfo * rai
 * 
 *         if sock is not None:             # <<<<<<<<<<<<<<
 *             if not _is_sock_dgram(sock.type):
 *                 raise ValueError(
 */
    goto __pyx_L4;
  }

  /* "uvloop/loop.pyx":3059
 *             udp._attach_fileobj(sock)
 *         else:
 *             if reuse_address is not _unset:             # <<<<<<<<<<<<<<
 *                 if reuse_address:
 *                     raise ValueError("Passing `reuse_address=True` is no "
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_unset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = (__pyx_cur_scope->__pyx_v_reuse_address != __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {

      /* "uvloop/loop.pyx":3060
 *         else:
 *             if reuse_address is not _unset:
 *                 if reuse_address:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Passing `reuse_address=True` is no "
 *                                      "longer supported, as the usage of "
 */
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_address); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 3060, __pyx_L1_error)
      if (unlikely(__pyx_t_4)) {

        /* "uvloop/loop.pyx":3061
 *             if reuse_address is not _unset:
 *                 if reuse_address:
 *                     raise ValueError("Passing `reuse_address=True` is no "             # <<<<<<<<<<<<<<
 *                                      "longer supported, as the usage of "
 *                                      "SO_REUSEPORT in UDP poses a significant "
 */
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__117, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 3061, __pyx_L1_error)

        /* "uvloop/loop.pyx":3060
 *         else:
 *             if reuse_address is not _unset:
 *                 if reuse_address:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Passing `reuse_address=True` is no "
 *                                      "longer supported, as the usage of "
 */
      }

      /* "uvloop/loop.pyx":3066
 *                                      "security concern.")
 *                 else:
 *                     warnings_warn("The *reuse_address* parameter has been "             # <<<<<<<<<<<<<<
 *                                   "deprecated as of 0.15.", DeprecationWarning,
 *                                   stacklevel=2)
 */
      /*else*/ {

        /* "uvloop/loop.pyx":3068
 *                     warnings_warn("The *reuse_address* parameter has been "
 *                                   "deprecated as of 0.15.", DeprecationWarning,
 *                                   stacklevel=2)             # <<<<<<<<<<<<<<
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:
 */
        __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3068, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3068, __pyx_L1_error)

        /* "uvloop/loop.pyx":3066
 *                                      "security concern.")
 *                 else:
 *                     warnings_warn("The *reuse_address* parameter has been "             # <<<<<<<<<<<<<<
 *                                   "deprecated as of 0.15.", DeprecationWarning,
 *                                   stacklevel=2)
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_warnings_warn, __pyx_tuple__118, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3066, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }

      /* "uvloop/loop.pyx":3059
 *             udp._attach_fileobj(sock)
 *         else:
 *             if reuse_address is not _unset:             # <<<<<<<<<<<<<<
 *                 if reuse_address:
 *                     raise ValueError("Passing `reuse_address=True` is no "
 */
    }

    /* "uvloop/loop.pyx":3069
 *                                   "deprecated as of 0.15.", DeprecationWarning,
 *                                   stacklevel=2)
 *             reuse_port = bool(reuse_port)             # <<<<<<<<<<<<<<
 *             if reuse_port and not has_SO_REUSEPORT:
 *                 raise ValueError(
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 3069, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyBool_FromLong((!(!__pyx_t_4))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_reuse_port);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_reuse_port, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3070
 *                                   stacklevel=2)
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'reuse_port not supported by socket module')
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3070, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_4 = __pyx_t_1;
      goto __pyx_L17_bool_binop_done;
    }
    __pyx_t_1 = (!(__pyx_v_6uvloop_4loop_has_SO_REUSEPORT != 0));
    __pyx_t_4 = __pyx_t_1;
    __pyx_L17_bool_binop_done:;
    if (unlikely(__pyx_t_4)) {

      /* "uvloop/loop.pyx":3071
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'reuse_port not supported by socket module')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3071, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 3071, __pyx_L1_error)

      /* "uvloop/loop.pyx":3070
 *                                   stacklevel=2)
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'reuse_port not supported by socket module')
 */
    }

    /* "uvloop/loop.pyx":3074
 *                     'reuse_port not supported by socket module')
 * 
 *             lads = None             # <<<<<<<<<<<<<<
 *             if local_addr is not None:
 *                 if (not isinstance(local_addr, (tuple, list)) or
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_lads = Py_None;

    /* "uvloop/loop.pyx":3075
 * 
 *             lads = None
 *             if local_addr is not None:             # <<<<<<<<<<<<<<
 *                 if (not isinstance(local_addr, (tuple, list)) or
 *                         len(local_addr) != 2):
 */
    __pyx_t_4 = (__pyx_cur_scope->__pyx_v_local_addr != Py_None);
    if (__pyx_t_4) {

      /* "uvloop/loop.pyx":3076
 *             lads = None
 *             if local_addr is not None:
 *                 if (not isinstance(local_addr, (tuple, list)) or             # <<<<<<<<<<<<<<
 *                         len(local_addr) != 2):
 *                     raise TypeError(
 */
      __pyx_t_9 = PyTuple_Check(__pyx_cur_scope->__pyx_v_local_addr); 
      if (!__pyx_t_9) {
      } else {
        __pyx_t_1 = __pyx_t_9;
        goto __pyx_L23_bool_binop_done;
      }
      __pyx_t_9 = PyList_Check(__pyx_cur_scope->__pyx_v_local_addr); 
      __pyx_t_1 = __pyx_t_9;
      __pyx_L23_bool_binop_done:;
      __pyx_t_9 = (!__pyx_t_1);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_4 = __pyx_t_9;
        goto __pyx_L21_bool_binop_done;
      }

      /* "uvloop/loop.pyx":3077
 *             if local_addr is not None:
 *                 if (not isinstance(local_addr, (tuple, list)) or
 *                         len(local_addr) != 2):             # <<<<<<<<<<<<<<
 *                     raise TypeError(
 *                         'local_addr must be a tuple of (host, port)')
 */
      __pyx_t_10 = PyObject_Length(__pyx_cur_scope->__pyx_v_local_addr); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3077, __pyx_L1_error)
      __pyx_t_9 = (__pyx_t_10 != 2);
      __pyx_t_4 = __pyx_t_9;
      __pyx_L21_bool_binop_done:;

      /* "uvloop/loop.pyx":3076
 *             lads = None
 *             if local_addr is not None:
 *                 if (not isinstance(local_addr, (tuple, list)) or             # <<<<<<<<<<<<<<
 *                         len(local_addr) != 2):
 *                     raise TypeError(
 */
      if (unlikely(__pyx_t_4)) {

        /* "uvloop/loop.pyx":3078
 *                 if (not isinstance(local_addr, (tuple, list)) or
 *                         len(local_addr) != 2):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'local_addr must be a tuple of (host, port)')
 *                 lads = await self._getaddrinfo(
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__119, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3078, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 3078, __pyx_L1_error)

        /* "uvloop/loop.pyx":3076
 *             lads = None
 *             if local_addr is not None:
 *                 if (not isinstance(local_addr, (tuple, list)) or             # <<<<<<<<<<<<<<
 *                         len(local_addr) != 2):
 *                     raise TypeError(
 */
      }

      /* "uvloop/loop.pyx":3081
 *                         'local_addr must be a tuple of (host, port)')
 *                 lads = await self._getaddrinfo(
 *                     local_addr[0], local_addr[1],             # <<<<<<<<<<<<<<
 *                     family, uv.SOCK_DGRAM, proto, flags,
 *                     0)
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_local_addr, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3081, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_local_addr, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3081, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":3082
 *                 lads = await self._getaddrinfo(
 *                     local_addr[0], local_addr[1],
 *                     family, uv.SOCK_DGRAM, proto, flags,             # <<<<<<<<<<<<<<
 *                     0)
 * 
 */
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3082, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_proto); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3082, __pyx_L1_error)
      __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_flags); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3082, __pyx_L1_error)

      /* "uvloop/loop.pyx":3080
 *                     raise TypeError(
 *                         'local_addr must be a tuple of (host, port)')
 *                 lads = await self._getaddrinfo(             # <<<<<<<<<<<<<<
 *                     local_addr[0], local_addr[1],
 *                     family, uv.SOCK_DGRAM, proto, flags,
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getaddrinfo(__pyx_cur_scope->__pyx_v_self, __pyx_t_2, __pyx_t_3, __pyx_t_8, SOCK_DGRAM, __pyx_t_7, __pyx_t_11, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3080, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L25_resume_from_await:;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3080, __pyx_L1_error)
        __pyx_t_5 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_5);
      } else {
        __pyx_t_5 = NULL;
        if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_5) < 0) __PYX_ERR(0, 3080, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_lads);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_lads, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/loop.pyx":3075
 * 
 *             lads = None
 *             if local_addr is not None:             # <<<<<<<<<<<<<<
 *                 if (not isinstance(local_addr, (tuple, list)) or
 *                         len(local_addr) != 2):
 */
    }

    /* "uvloop/loop.pyx":3085
 *                     0)
 * 
 *             rads = None             # <<<<<<<<<<<<<<
 *             if remote_addr is not None:
 *                 if (not isinstance(remote_addr, (tuple, list)) or
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __pyx_cur_scope->__pyx_v_rads = Py_None;

    /* "uvloop/loop.pyx":3086
 * 
 *             rads = None
 *             if remote_addr is not None:             # <<<<<<<<<<<<<<
 *                 if (not isinstance(remote_addr, (tuple, list)) or
 *                         len(remote_addr) != 2):
 */
    __pyx_t_4 = (__pyx_cur_scope->__pyx_v_remote_addr != Py_None);
    if (__pyx_t_4) {

      /* "uvloop/loop.pyx":3087
 *             rads = None
 *             if remote_addr is not None:
 *                 if (not isinstance(remote_addr, (tuple, list)) or             # <<<<<<<<<<<<<<
 *                         len(remote_addr) != 2):
 *                     raise TypeError(
 */
      __pyx_t_1 = PyTuple_Check(__pyx_cur_scope->__pyx_v_remote_addr); 
      if (!__pyx_t_1) {
      } else {
        __pyx_t_9 = __pyx_t_1;
        goto __pyx_L30_bool_binop_done;
      }
      __pyx_t_1 = PyList_Check(__pyx_cur_scope->__pyx_v_remote_addr); 
      __pyx_t_9 = __pyx_t_1;
      __pyx_L30_bool_binop_done:;
      __pyx_t_1 = (!__pyx_t_9);
      if (!__pyx_t_1) {
      } else {
        __pyx_t_4 = __pyx_t_1;
        goto __pyx_L28_bool_binop_done;
      }

      /* "uvloop/loop.pyx":3088
 *             if remote_addr is not None:
 *                 if (not isinstance(remote_addr, (tuple, list)) or
 *                         len(remote_addr) != 2):             # <<<<<<<<<<<<<<
 *                     raise TypeError(
 *                         'remote_addr must be a tuple of (host, port)')
 */
      __pyx_t_10 = PyObject_Length(__pyx_cur_scope->__pyx_v_remote_addr); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3088, __pyx_L1_error)
      __pyx_t_1 = (__pyx_t_10 != 2);
      __pyx_t_4 = __pyx_t_1;
      __pyx_L28_bool_binop_done:;

      /* "uvloop/loop.pyx":3087
 *             rads = None
 *             if remote_addr is not None:
 *                 if (not isinstance(remote_addr, (tuple, list)) or             # <<<<<<<<<<<<<<
 *                         len(remote_addr) != 2):
 *                     raise TypeError(
 */
      if (unlikely(__pyx_t_4)) {

        /* "uvloop/loop.pyx":3089
 *                 if (not isinstance(remote_addr, (tuple, list)) or
 *                         len(remote_addr) != 2):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'remote_addr must be a tuple of (host, port)')
 *                 rads = await self._getaddrinfo(
 */
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__120, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3089, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __PYX_ERR(0, 3089, __pyx_L1_error)

        /* "uvloop/loop.pyx":3087
 *             rads = None
 *             if remote_addr is not None:
 *                 if (not isinstance(remote_addr, (tuple, list)) or             # <<<<<<<<<<<<<<
 *                         len(remote_addr) != 2):
 *                     raise TypeError(
 */
      }

      /* "uvloop/loop.pyx":3092
 *                         'remote_addr must be a tuple of (host, port)')
 *                 rads = await self._getaddrinfo(
 *                     remote_addr[0], remote_addr[1],             # <<<<<<<<<<<<<<
 *                     family, uv.SOCK_DGRAM, proto, flags,
 *                     0)
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_remote_addr, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3092, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_remote_addr, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3092, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/loop.pyx":3093
 *                 rads = await self._getaddrinfo(
 *                     remote_addr[0], remote_addr[1],
 *                     family, uv.SOCK_DGRAM, proto, flags,             # <<<<<<<<<<<<<<
 *                     0)
 * 
 */
      __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3093, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_proto); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3093, __pyx_L1_error)
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_flags); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3093, __pyx_L1_error)

      /* "uvloop/loop.pyx":3091
 *                     raise TypeError(
 *                         'remote_addr must be a tuple of (host, port)')
 *                 rads = await self._getaddrinfo(             # <<<<<<<<<<<<<<
 *                     remote_addr[0], remote_addr[1],
 *                     family, uv.SOCK_DGRAM, proto, flags,
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_getaddrinfo(__pyx_cur_scope->__pyx_v_self, __pyx_t_5, __pyx_t_3, __pyx_t_11, SOCK_DGRAM, __pyx_t_7, __pyx_t_8, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3091, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XGOTREF(__pyx_r);
      if (likely(__pyx_r)) {
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, awaiting value */
        __pyx_generator->resume_label = 2;
        return __pyx_r;
        __pyx_L32_resume_from_await:;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3091, __pyx_L1_error)
        __pyx_t_2 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_2);
      } else {
        __pyx_t_2 = NULL;
        if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_2) < 0) __PYX_ERR(0, 3091, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_rads);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_rads, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/loop.pyx":3086
 * 
 *             rads = None
 *             if remote_addr is not None:             # <<<<<<<<<<<<<<
 *                 if (not isinstance(remote_addr, (tuple, list)) or
 *                         len(remote_addr) != 2):
 */
    }

    /* "uvloop/loop.pyx":3096
 *                     0)
 * 
 *             excs = []             # <<<<<<<<<<<<<<
 *             if lads is None:
 *                 if rads is not None:
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_v_excs = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3097
 * 
 *             excs = []
 *             if lads is None:             # <<<<<<<<<<<<<<
 *                 if rads is not None:
 *                     udp = UDPTransport.__new__(UDPTransport)
 */
    __pyx_t_4 = (__pyx_cur_scope->__pyx_v_lads == Py_None);
    if (__pyx_t_4) {

      /* "uvloop/loop.pyx":3098
 *             excs = []
 *             if lads is None:
 *                 if rads is not None:             # <<<<<<<<<<<<<<
 *                     udp = UDPTransport.__new__(UDPTransport)
 *                     rai = (<AddrInfo>rads).data
 */
      __pyx_t_4 = (__pyx_cur_scope->__pyx_v_rads != Py_None);
      if (__pyx_t_4) {

        /* "uvloop/loop.pyx":3099
 *             if lads is None:
 *                 if rads is not None:
 *                     udp = UDPTransport.__new__(UDPTransport)             # <<<<<<<<<<<<<<
 *                     rai = (<AddrInfo>rads).data
 *                     udp._init(self, rai.ai_family)
 */
        __pyx_t_2 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UDPTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UDPTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3099, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_2);
        __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_udp);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_udp, ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_t_2));
        __Pyx_GIVEREF((PyObject *)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":3100
 *                 if rads is not None:
 *                     udp = UDPTransport.__new__(UDPTransport)
 *                     rai = (<AddrInfo>rads).data             # <<<<<<<<<<<<<<
 *                     udp._init(self, rai.ai_family)
 *                     udp._connect(rai.ai_addr, rai.ai_addrlen)
 */
        __pyx_t_12 = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_cur_scope->__pyx_v_rads)->data;
        __pyx_cur_scope->__pyx_v_rai = __pyx_t_12;

        /* "uvloop/loop.pyx":3101
 *                     udp = UDPTransport.__new__(UDPTransport)
 *                     rai = (<AddrInfo>rads).data
 *                     udp._init(self, rai.ai_family)             # <<<<<<<<<<<<<<
 *                     udp._connect(rai.ai_addr, rai.ai_addrlen)
 *                     udp._set_address(rai)
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_init(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_rai->ai_family); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3101, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":3102
 *                     rai = (<AddrInfo>rads).data
 *                     udp._init(self, rai.ai_family)
 *                     udp._connect(rai.ai_addr, rai.ai_addrlen)             # <<<<<<<<<<<<<<
 *                     udp._set_address(rai)
 *                 else:
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_connect(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_rai->ai_addr, __pyx_cur_scope->__pyx_v_rai->ai_addrlen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3102, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":3103
 *                     udp._init(self, rai.ai_family)
 *                     udp._connect(rai.ai_addr, rai.ai_addrlen)
 *                     udp._set_address(rai)             # <<<<<<<<<<<<<<
 *                 else:
 *                     if family not in (uv.AF_INET, uv.AF_INET6):
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_set_address(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_rai); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3103, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":3098
 *             excs = []
 *             if lads is None:
 *                 if rads is not None:             # <<<<<<<<<<<<<<
 *                     udp = UDPTransport.__new__(UDPTransport)
 *                     rai = (<AddrInfo>rads).data
 */
        goto __pyx_L34;
      }

      /* "uvloop/loop.pyx":3105
 *                     udp._set_address(rai)
 *                 else:
 *                     if family not in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *                         raise ValueError('unexpected address family')
 *                     udp = UDPTransport.__new__(UDPTransport)
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_family);
        __pyx_t_2 = __pyx_cur_scope->__pyx_v_family;
        __pyx_t_3 = __Pyx_PyInt_From_int(AF_INET); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3105, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3105, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_1) {
        } else {
          __pyx_t_4 = __pyx_t_1;
          goto __pyx_L36_bool_binop_done;
        }
        __pyx_t_5 = __Pyx_PyInt_From_int(AF_INET6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3105, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3105, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_4 = __pyx_t_1;
        __pyx_L36_bool_binop_done:;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_1 = __pyx_t_4;
        if (unlikely(__pyx_t_1)) {

          /* "uvloop/loop.pyx":3106
 *                 else:
 *                     if family not in (uv.AF_INET, uv.AF_INET6):
 *                         raise ValueError('unexpected address family')             # <<<<<<<<<<<<<<
 *                     udp = UDPTransport.__new__(UDPTransport)
 *                     udp._init(self, family)
 */
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__121, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3106, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 3106, __pyx_L1_error)

          /* "uvloop/loop.pyx":3105
 *                     udp._set_address(rai)
 *                 else:
 *                     if family not in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *                         raise ValueError('unexpected address family')
 *                     udp = UDPTransport.__new__(UDPTransport)
 */
        }

        /* "uvloop/loop.pyx":3107
 *                     if family not in (uv.AF_INET, uv.AF_INET6):
 *                         raise ValueError('unexpected address family')
 *                     udp = UDPTransport.__new__(UDPTransport)             # <<<<<<<<<<<<<<
 *                     udp._init(self, family)
 * 
 */
        __pyx_t_2 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UDPTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UDPTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3107, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_2);
        __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_udp);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_udp, ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_t_2));
        __Pyx_GIVEREF((PyObject *)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":3108
 *                         raise ValueError('unexpected address family')
 *                     udp = UDPTransport.__new__(UDPTransport)
 *                     udp._init(self, family)             # <<<<<<<<<<<<<<
 * 
 *                 if reuse_port:
 */
        __pyx_t_6 = __Pyx_PyInt_As_unsigned_int(__pyx_cur_scope->__pyx_v_family); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3108, __pyx_L1_error)
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_init(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_self, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_L34:;

      /* "uvloop/loop.pyx":3110
 *                     udp._init(self, family)
 * 
 *                 if reuse_port:             # <<<<<<<<<<<<<<
 *                     self._sock_set_reuseport(udp._fileno())
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3110, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "uvloop/loop.pyx":3111
 * 
 *                 if reuse_port:
 *                     self._sock_set_reuseport(udp._fileno())             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._fileno(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_udp)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3111, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_set_reuseport(__pyx_cur_scope->__pyx_v_self, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":3110
 *                     udp._init(self, family)
 * 
 *                 if reuse_port:             # <<<<<<<<<<<<<<
 *                     self._sock_set_reuseport(udp._fileno())
 * 
 */
      }

      /* "uvloop/loop.pyx":3097
 * 
 *             excs = []
 *             if lads is None:             # <<<<<<<<<<<<<<
 *                 if rads is not None:
 *                     udp = UDPTransport.__new__(UDPTransport)
 */
      goto __pyx_L33;
    }

    /* "uvloop/loop.pyx":3114
 * 
 *             else:
 *                 lai = (<AddrInfo>lads).data             # <<<<<<<<<<<<<<
 *                 while lai is not NULL:
 *                     try:
 */
    /*else*/ {
      __pyx_t_12 = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_cur_scope->__pyx_v_lads)->data;
      __pyx_cur_scope->__pyx_v_lai = __pyx_t_12;

      /* "uvloop/loop.pyx":3115
 *             else:
 *                 lai = (<AddrInfo>lads).data
 *                 while lai is not NULL:             # <<<<<<<<<<<<<<
 *                     try:
 *                         udp = UDPTransport.__new__(UDPTransport)
 */
      while (1) {
        __pyx_t_1 = (__pyx_cur_scope->__pyx_v_lai != NULL);
        if (!__pyx_t_1) break;

        /* "uvloop/loop.pyx":3116
 *                 lai = (<AddrInfo>lads).data
 *                 while lai is not NULL:
 *                     try:             # <<<<<<<<<<<<<<
 *                         udp = UDPTransport.__new__(UDPTransport)
 *                         udp._init(self, lai.ai_family)
 */
        {
          __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          /*try:*/ {

            /* "uvloop/loop.pyx":3117
 *                 while lai is not NULL:
 *                     try:
 *                         udp = UDPTransport.__new__(UDPTransport)             # <<<<<<<<<<<<<<
 *                         udp._init(self, lai.ai_family)
 *                         if reuse_port:
 */
            __pyx_t_2 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UDPTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UDPTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3117, __pyx_L41_error)
            __Pyx_GOTREF((PyObject *)__pyx_t_2);
            __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_udp);
            __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_udp, ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_t_2));
            __Pyx_GIVEREF((PyObject *)__pyx_t_2);
            __pyx_t_2 = 0;

            /* "uvloop/loop.pyx":3118
 *                     try:
 *                         udp = UDPTransport.__new__(UDPTransport)
 *                         udp._init(self, lai.ai_family)             # <<<<<<<<<<<<<<
 *                         if reuse_port:
 *                             self._sock_set_reuseport(udp._fileno())
 */
            __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_init(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_lai->ai_family); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3118, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "uvloop/loop.pyx":3119
 *                         udp = UDPTransport.__new__(UDPTransport)
 *                         udp._init(self, lai.ai_family)
 *                         if reuse_port:             # <<<<<<<<<<<<<<
 *                             self._sock_set_reuseport(udp._fileno())
 *                         udp._bind(lai.ai_addr)
 */
            __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_reuse_port); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3119, __pyx_L41_error)
            if (__pyx_t_1) {

              /* "uvloop/loop.pyx":3120
 *                         udp._init(self, lai.ai_family)
 *                         if reuse_port:
 *                             self._sock_set_reuseport(udp._fileno())             # <<<<<<<<<<<<<<
 *                         udp._bind(lai.ai_addr)
 *                     except (KeyboardInterrupt, SystemExit):
 */
              __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._fileno(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_cur_scope->__pyx_v_udp)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3120, __pyx_L41_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 3120, __pyx_L41_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_sock_set_reuseport(__pyx_cur_scope->__pyx_v_self, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3120, __pyx_L41_error)
              __Pyx_GOTREF(__pyx_t_2);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

              /* "uvloop/loop.pyx":3119
 *                         udp = UDPTransport.__new__(UDPTransport)
 *                         udp._init(self, lai.ai_family)
 *                         if reuse_port:             # <<<<<<<<<<<<<<
 *                             self._sock_set_reuseport(udp._fileno())
 *                         udp._bind(lai.ai_addr)
 */
            }

            /* "uvloop/loop.pyx":3121
 *                         if reuse_port:
 *                             self._sock_set_reuseport(udp._fileno())
 *                         udp._bind(lai.ai_addr)             # <<<<<<<<<<<<<<
 *                     except (KeyboardInterrupt, SystemExit):
 *                         raise
 */
            __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_bind(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_lai->ai_addr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3121, __pyx_L41_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "uvloop/loop.pyx":3116
 *                 lai = (<AddrInfo>lads).data
 *                 while lai is not NULL:
 *                     try:             # <<<<<<<<<<<<<<
 *                         udp = UDPTransport.__new__(UDPTransport)
 *                         udp._init(self, lai.ai_family)
 */
          }

          /* "uvloop/loop.pyx":3129
 *                         continue
 *                     else:
 *                         break             # <<<<<<<<<<<<<<
 *                 else:
 *                     ctx = None
 */
          /*else:*/ {
            goto __pyx_L46_try_break;
          }
          __pyx_L41_error:;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/loop.pyx":3122
 *                             self._sock_set_reuseport(udp._fileno())
 *                         udp._bind(lai.ai_addr)
 *                     except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                         raise
 *                     except BaseException as ex:
 */
          __pyx_t_8 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
          if (__pyx_t_8) {
            __Pyx_AddTraceback("uvloop.loop.Loop.create_datagram_endpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_5) < 0) __PYX_ERR(0, 3122, __pyx_L43_except_error)
            __Pyx_XGOTREF(__pyx_t_2);
            __Pyx_XGOTREF(__pyx_t_3);
            __Pyx_XGOTREF(__pyx_t_5);

            /* "uvloop/loop.pyx":3123
 *                         udp._bind(lai.ai_addr)
 *                     except (KeyboardInterrupt, SystemExit):
 *                         raise             # <<<<<<<<<<<<<<
 *                     except BaseException as ex:
 *                         lai = lai.ai_next
 */
            __Pyx_GIVEREF(__pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_3);
            __Pyx_XGIVEREF(__pyx_t_5);
            __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_3, __pyx_t_5);
            __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_5 = 0; 
            __PYX_ERR(0, 3123, __pyx_L43_except_error)
          }

          /* "uvloop/loop.pyx":3124
 *                     except (KeyboardInterrupt, SystemExit):
 *                         raise
 *                     except BaseException as ex:             # <<<<<<<<<<<<<<
 *                         lai = lai.ai_next
 *                         excs.append(ex)
 */
          __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
          if (__pyx_t_8) {
            __Pyx_AddTraceback("uvloop.loop.Loop.create_datagram_endpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 3124, __pyx_L43_except_error)
            __Pyx_XGOTREF(__pyx_t_5);
            __Pyx_XGOTREF(__pyx_t_3);
            __Pyx_XGOTREF(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_3);
            __pyx_cur_scope->__pyx_v_ex = __pyx_t_3;
            /*try:*/ {

              /* "uvloop/loop.pyx":3125
 *                         raise
 *                     except BaseException as ex:
 *                         lai = lai.ai_next             # <<<<<<<<<<<<<<
 *                         excs.append(ex)
 *                         continue
 */
              __pyx_t_12 = __pyx_cur_scope->__pyx_v_lai->ai_next;
              __pyx_cur_scope->__pyx_v_lai = __pyx_t_12;

              /* "uvloop/loop.pyx":3126
 *                     except BaseException as ex:
 *                         lai = lai.ai_next
 *                         excs.append(ex)             # <<<<<<<<<<<<<<
 *                         continue
 *                     else:
 */
              __pyx_t_16 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_excs, __pyx_cur_scope->__pyx_v_ex); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 3126, __pyx_L57_error)

              /* "uvloop/loop.pyx":3127
 *                         lai = lai.ai_next
 *                         excs.append(ex)
 *                         continue             # <<<<<<<<<<<<<<
 *                     else:
 *                         break
 */
              goto __pyx_L54_continue;
            }

            /* "uvloop/loop.pyx":3124
 *                     except (KeyboardInterrupt, SystemExit):
 *                         raise
 *                     except BaseException as ex:             # <<<<<<<<<<<<<<
 *                         lai = lai.ai_next
 *                         excs.append(ex)
 */
            /*finally:*/ {
              __pyx_L57_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_assign
                __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
                if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
                if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_21);
                __Pyx_XGOTREF(__pyx_t_22);
                __Pyx_XGOTREF(__pyx_t_23);
                __pyx_t_8 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
                {
                  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_ex);
                  __Pyx_DECREF(__pyx_cur_scope->__pyx_v_ex); __pyx_cur_scope->__pyx_v_ex = 0;
                }
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_21);
                  __Pyx_XGIVEREF(__pyx_t_22);
                  __Pyx_XGIVEREF(__pyx_t_23);
                  __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
                }
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
                __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
                __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_17;
                goto __pyx_L43_except_error;
              }
              __pyx_L54_continue: {
                __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_ex);
                __Pyx_DECREF(__pyx_cur_scope->__pyx_v_ex); __pyx_cur_scope->__pyx_v_ex = 0;
                goto __pyx_L52_except_continue;
              }
            }
            __pyx_L52_except_continue:;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            goto __pyx_L47_try_continue;
          }
          goto __pyx_L43_except_error;

          /* "uvloop/loop.pyx":3116
 *                 lai = (<AddrInfo>lads).data
 *                 while lai is not NULL:
 *                     try:             # <<<<<<<<<<<<<<
 *                         udp = UDPTransport.__new__(UDPTransport)
 *                         udp._init(self, lai.ai_family)
 */
          __pyx_L43_except_error:;
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          goto __pyx_L1_error;
          __pyx_L46_try_break:;
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          goto __pyx_L40_break;
          __pyx_L47_try_continue:;
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          goto __pyx_L39_continue;
        }
        __pyx_L39_continue:;
      }

      /* "uvloop/loop.pyx":3131
 *                         break
 *                 else:
 *                     ctx = None             # <<<<<<<<<<<<<<
 *                     if len(excs):
 *                         ctx = excs[0]
 */
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __pyx_cur_scope->__pyx_v_ctx = Py_None;

        /* "uvloop/loop.pyx":3132
 *                 else:
 *                     ctx = None
 *                     if len(excs):             # <<<<<<<<<<<<<<
 *                         ctx = excs[0]
 *                     raise OSError('could not bind to local_addr {}'.format(
 */
        __pyx_t_10 = __Pyx_PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_excs); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3132, __pyx_L1_error)
        __pyx_t_1 = (__pyx_t_10 != 0);
        if (__pyx_t_1) {

          /* "uvloop/loop.pyx":3133
 *                     ctx = None
 *                     if len(excs):
 *                         ctx = excs[0]             # <<<<<<<<<<<<<<
 *                     raise OSError('could not bind to local_addr {}'.format(
 *                         local_addr)) from ctx
 */
          __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_excs, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3133, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_ctx);
          __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_ctx, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_2);
          __pyx_t_2 = 0;

          /* "uvloop/loop.pyx":3132
 *                 else:
 *                     ctx = None
 *                     if len(excs):             # <<<<<<<<<<<<<<
 *                         ctx = excs[0]
 *                     raise OSError('could not bind to local_addr {}'.format(
 */
        }

        /* "uvloop/loop.pyx":3134
 *                     if len(excs):
 *                         ctx = excs[0]
 *                     raise OSError('could not bind to local_addr {}'.format(             # <<<<<<<<<<<<<<
 *                         local_addr)) from ctx
 * 
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_could_not_bind_to_local_addr, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);

        /* "uvloop/loop.pyx":3135
 *                         ctx = excs[0]
 *                     raise OSError('could not bind to local_addr {}'.format(
 *                         local_addr)) from ctx             # <<<<<<<<<<<<<<
 * 
 *                 if rads is not None:
 */
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_local_addr};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3134, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "uvloop/loop.pyx":3134
 *                     if len(excs):
 *                         ctx = excs[0]
 *                     raise OSError('could not bind to local_addr {}'.format(             # <<<<<<<<<<<<<<
 *                         local_addr)) from ctx
 * 
 */
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_OSError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/loop.pyx":3135
 *                         ctx = excs[0]
 *                     raise OSError('could not bind to local_addr {}'.format(
 *                         local_addr)) from ctx             # <<<<<<<<<<<<<<
 * 
 *                 if rads is not None:
 */
        __Pyx_Raise(__pyx_t_3, 0, 0, __pyx_cur_scope->__pyx_v_ctx);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 3134, __pyx_L1_error)
      }
      __pyx_L40_break:;

      /* "uvloop/loop.pyx":3137
 *                         local_addr)) from ctx
 * 
 *                 if rads is not None:             # <<<<<<<<<<<<<<
 *                     rai = (<AddrInfo>rads).data
 *                     while rai is not NULL:
 */
      __pyx_t_1 = (__pyx_cur_scope->__pyx_v_rads != Py_None);
      if (__pyx_t_1) {

        /* "uvloop/loop.pyx":3138
 * 
 *                 if rads is not None:
 *                     rai = (<AddrInfo>rads).data             # <<<<<<<<<<<<<<
 *                     while rai is not NULL:
 *                         if rai.ai_family != lai.ai_family:
 */
        __pyx_t_12 = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_cur_scope->__pyx_v_rads)->data;
        __pyx_cur_scope->__pyx_v_rai = __pyx_t_12;

        /* "uvloop/loop.pyx":3139
 *                 if rads is not None:
 *                     rai = (<AddrInfo>rads).data
 *                     while rai is not NULL:             # <<<<<<<<<<<<<<
 *                         if rai.ai_family != lai.ai_family:
 *                             rai = rai.ai_next
 */
        while (1) {
          __pyx_t_1 = (__pyx_cur_scope->__pyx_v_rai != NULL);
          if (!__pyx_t_1) break;

          /* "uvloop/loop.pyx":3140
 *                     rai = (<AddrInfo>rads).data
 *                     while rai is not NULL:
 *                         if rai.ai_family != lai.ai_family:             # <<<<<<<<<<<<<<
 *                             rai = rai.ai_next
 *                             continue
 */
          __pyx_t_1 = (__pyx_cur_scope->__pyx_v_rai->ai_family != __pyx_cur_scope->__pyx_v_lai->ai_family);
          if (__pyx_t_1) {

            /* "uvloop/loop.pyx":3141
 *                     while rai is not NULL:
 *                         if rai.ai_family != lai.ai_family:
 *                             rai = rai.ai_next             # <<<<<<<<<<<<<<
 *                             continue
 *                         if rai.ai_protocol != lai.ai_protocol:
 */
            __pyx_t_12 = __pyx_cur_scope->__pyx_v_rai->ai_next;
            __pyx_cur_scope->__pyx_v_rai = __pyx_t_12;

            /* "uvloop/loop.pyx":3142
 *                         if rai.ai_family != lai.ai_family:
 *                             rai = rai.ai_next
 *                             continue             # <<<<<<<<<<<<<<
 *                         if rai.ai_protocol != lai.ai_protocol:
 *                             rai = rai.ai_next
 */
            goto __pyx_L65_continue;

            /* "uvloop/loop.pyx":3140
 *                     rai = (<AddrInfo>rads).data
 *                     while rai is not NULL:
 *                         if rai.ai_family != lai.ai_family:             # <<<<<<<<<<<<<<
 *                             rai = rai.ai_next
 *                             continue
 */
          }

          /* "uvloop/loop.pyx":3143
 *                             rai = rai.ai_next
 *                             continue
 *                         if rai.ai_protocol != lai.ai_protocol:             # <<<<<<<<<<<<<<
 *                             rai = rai.ai_next
 *                             continue
 */
          __pyx_t_1 = (__pyx_cur_scope->__pyx_v_rai->ai_protocol != __pyx_cur_scope->__pyx_v_lai->ai_protocol);
          if (__pyx_t_1) {

            /* "uvloop/loop.pyx":3144
 *                             continue
 *                         if rai.ai_protocol != lai.ai_protocol:
 *                             rai = rai.ai_next             # <<<<<<<<<<<<<<
 *                             continue
 *                         udp._connect(rai.ai_addr, rai.ai_addrlen)
 */
            __pyx_t_12 = __pyx_cur_scope->__pyx_v_rai->ai_next;
            __pyx_cur_scope->__pyx_v_rai = __pyx_t_12;

            /* "uvloop/loop.pyx":3145
 *                         if rai.ai_protocol != lai.ai_protocol:
 *                             rai = rai.ai_next
 *                             continue             # <<<<<<<<<<<<<<
 *                         udp._connect(rai.ai_addr, rai.ai_addrlen)
 *                         udp._set_address(rai)
 */
            goto __pyx_L65_continue;

            /* "uvloop/loop.pyx":3143
 *                             rai = rai.ai_next
 *                             continue
 *                         if rai.ai_protocol != lai.ai_protocol:             # <<<<<<<<<<<<<<
 *                             rai = rai.ai_next
 *                             continue
 */
          }

          /* "uvloop/loop.pyx":3146
 *                             rai = rai.ai_next
 *                             continue
 *                         udp._connect(rai.ai_addr, rai.ai_addrlen)             # <<<<<<<<<<<<<<
 *                         udp._set_address(rai)
 *                         break
 */
          __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_connect(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_rai->ai_addr, __pyx_cur_scope->__pyx_v_rai->ai_addrlen); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3146, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "uvloop/loop.pyx":3147
 *                             continue
 *                         udp._connect(rai.ai_addr, rai.ai_addrlen)
 *                         udp._set_address(rai)             # <<<<<<<<<<<<<<
 *                         break
 *                     else:
 */
          __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_set_address(__pyx_cur_scope->__pyx_v_udp, __pyx_cur_scope->__pyx_v_rai); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3147, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "uvloop/loop.pyx":3148
 *                         udp._connect(rai.ai_addr, rai.ai_addrlen)
 *                         udp._set_address(rai)
 *                         break             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise OSError(
 */
          goto __pyx_L66_break;
          __pyx_L65_continue:;
        }

        /* "uvloop/loop.pyx":3150
 *                         break
 *                     else:
 *                         raise OSError(             # <<<<<<<<<<<<<<
 *                             'could not bind to remote_addr {}'.format(
 *                                 remote_addr))
 */
        /*else*/ {

          /* "uvloop/loop.pyx":3151
 *                     else:
 *                         raise OSError(
 *                             'could not bind to remote_addr {}'.format(             # <<<<<<<<<<<<<<
 *                                 remote_addr))
 * 
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_could_not_bind_to_remote_addr, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3151, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);

          /* "uvloop/loop.pyx":3152
 *                         raise OSError(
 *                             'could not bind to remote_addr {}'.format(
 *                                 remote_addr))             # <<<<<<<<<<<<<<
 * 
 *         if allow_broadcast:
 */
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_remote_addr};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3151, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }

          /* "uvloop/loop.pyx":3150
 *                         break
 *                     else:
 *                         raise OSError(             # <<<<<<<<<<<<<<
 *                             'could not bind to remote_addr {}'.format(
 *                                 remote_addr))
 */
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_OSError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3150, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 3150, __pyx_L1_error)
        }
        __pyx_L66_break:;

        /* "uvloop/loop.pyx":3137
 *                         local_addr)) from ctx
 * 
 *                 if rads is not None:             # <<<<<<<<<<<<<<
 *                     rai = (<AddrInfo>rads).data
 *                     while rai is not NULL:
 */
      }
    }
    __pyx_L33:;
  }
  __pyx_L4:;

  /* "uvloop/loop.pyx":3154
 *                                 remote_addr))
 * 
 *         if allow_broadcast:             # <<<<<<<<<<<<<<
 *             udp._set_broadcast(1)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_allow_broadcast); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3154, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3155
 * 
 *         if allow_broadcast:
 *             udp._set_broadcast(1)             # <<<<<<<<<<<<<<
 * 
 *         protocol = protocol_factory()
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_set_broadcast(__pyx_cur_scope->__pyx_v_udp, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3154
 *                                 remote_addr))
 * 
 *         if allow_broadcast:             # <<<<<<<<<<<<<<
 *             udp._set_broadcast(1)
 * 
 */
  }

  /* "uvloop/loop.pyx":3157
 *             udp._set_broadcast(1)
 * 
 *         protocol = protocol_factory()             # <<<<<<<<<<<<<<
 *         waiter = self._new_future()
 *         assert udp is not None
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol_factory);
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_protocol_factory; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_protocol = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3158
 * 
 *         protocol = protocol_factory()
 *         waiter = self._new_future()             # <<<<<<<<<<<<<<
 *         assert udp is not None
 *         udp._set_protocol(protocol)
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_waiter = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3159
 *         protocol = protocol_factory()
 *         waiter = self._new_future()
 *         assert udp is not None             # <<<<<<<<<<<<<<
 *         udp._set_protocol(protocol)
 *         udp._set_waiter(waiter)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (((PyObject *)__pyx_cur_scope->__pyx_v_udp) != Py_None);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 3159, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 3159, __pyx_L1_error)
  #endif

  /* "uvloop/loop.pyx":3160
 *         waiter = self._new_future()
 *         assert udp is not None
 *         udp._set_protocol(protocol)             # <<<<<<<<<<<<<<
 *         udp._set_waiter(waiter)
 *         udp._init_protocol()
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._set_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_udp), __pyx_cur_scope->__pyx_v_protocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3161
 *         assert udp is not None
 *         udp._set_protocol(protocol)
 *         udp._set_waiter(waiter)             # <<<<<<<<<<<<<<
 *         udp._init_protocol()
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._set_waiter(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_udp), __pyx_cur_scope->__pyx_v_waiter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3162
 *         udp._set_protocol(protocol)
 *         udp._set_waiter(waiter)
 *         udp._init_protocol()             # <<<<<<<<<<<<<<
 * 
 *         await waiter
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_cur_scope->__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_cur_scope->__pyx_v_udp)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3164
 *         udp._init_protocol()
 * 
 *         await waiter             # <<<<<<<<<<<<<<
 *         return udp, protocol
 * 
 */
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 3;
    return __pyx_r;
    __pyx_L70_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3164, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(0, 3164, __pyx_L1_error)
    }
  }

  /* "uvloop/loop.pyx":3165
 * 
 *         await waiter
 *         return udp, protocol             # <<<<<<<<<<<<<<
 * 
 *     def _monitor_fs(self, path: str, callback) -> asyncio.Handle:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_udp);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_udp);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_udp))) __PYX_ERR(0, 3165, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_protocol);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_protocol);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_protocol)) __PYX_ERR(0, 3165, __pyx_L1_error);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("create_datagram_endpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3167
 *         return udp, protocol
 * 
 *     def _monitor_fs(self, path: str, callback) -> asyncio.Handle:             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVFSEvent fs_handle
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_137_monitor_fs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_136_monitor_fs, "Loop._monitor_fs(self, unicode path: str, callback) -> asyncio.Handle");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_137_monitor_fs = {"_monitor_fs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_137_monitor_fs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_136_monitor_fs};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_137_monitor_fs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_callback = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_monitor_fs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_path)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3167, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3167, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_monitor_fs", 1, 2, 2, 1); __PYX_ERR(0, 3167, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_monitor_fs") < 0)) __PYX_ERR(0, 3167, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_path = ((PyObject*)values[0]);
    __pyx_v_callback = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_monitor_fs", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 3167, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop._monitor_fs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_path), (&PyUnicode_Type), 0, "path", 1))) __PYX_ERR(0, 3167, __pyx_L1_error)
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_136_monitor_fs(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_path, __pyx_v_callback);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_136_monitor_fs(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_path, PyObject *__pyx_v_callback) {
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_fs_handle = 0;
  char *__pyx_v_c_str_path;
  PyObject *__pyx_v_p_bytes = NULL;
  long __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_monitor_fs", 1);

  /* "uvloop/loop.pyx":3172
 *             char* c_str_path
 * 
 *         self._check_closed()             # <<<<<<<<<<<<<<
 *         fs_handle = UVFSEvent.new(self, callback, None)
 *         p_bytes = path.encode('UTF-8')
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__check_closed(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3173
 * 
 *         self._check_closed()
 *         fs_handle = UVFSEvent.new(self, callback, None)             # <<<<<<<<<<<<<<
 *         p_bytes = path.encode('UTF-8')
 *         c_str_path = p_bytes
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_9UVFSEvent_new(__pyx_v_self, __pyx_v_callback, Py_None)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fs_handle = ((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3174
 *         self._check_closed()
 *         fs_handle = UVFSEvent.new(self, callback, None)
 *         p_bytes = path.encode('UTF-8')             # <<<<<<<<<<<<<<
 *         c_str_path = p_bytes
 *         flags = 0
 */
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_p_bytes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3175
 *         fs_handle = UVFSEvent.new(self, callback, None)
 *         p_bytes = path.encode('UTF-8')
 *         c_str_path = p_bytes             # <<<<<<<<<<<<<<
 *         flags = 0
 *         fs_handle.start(c_str_path, flags)
 */
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_v_p_bytes); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 3175, __pyx_L1_error)
  __pyx_v_c_str_path = __pyx_t_2;

  /* "uvloop/loop.pyx":3176
 *         p_bytes = path.encode('UTF-8')
 *         c_str_path = p_bytes
 *         flags = 0             # <<<<<<<<<<<<<<
 *         fs_handle.start(c_str_path, flags)
 *         return fs_handle
 */
  __pyx_v_flags = 0;

  /* "uvloop/loop.pyx":3177
 *         c_str_path = p_bytes
 *         flags = 0
 *         fs_handle.start(c_str_path, flags)             # <<<<<<<<<<<<<<
 *         return fs_handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_fs_handle->__pyx_base.__pyx_vtab)->start(__pyx_v_fs_handle, __pyx_v_c_str_path, __pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3178
 *         flags = 0
 *         fs_handle.start(c_str_path, flags)
 *         return fs_handle             # <<<<<<<<<<<<<<
 * 
 *     def _check_default_executor(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_fs_handle);
  __pyx_r = ((PyObject *)__pyx_v_fs_handle);
  goto __pyx_L0;

  /* "uvloop/loop.pyx":3167
 *         return udp, protocol
 * 
 *     def _monitor_fs(self, path: str, callback) -> asyncio.Handle:             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVFSEvent fs_handle
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._monitor_fs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fs_handle);
  __Pyx_XDECREF(__pyx_v_p_bytes);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3180
 *         return fs_handle
 * 
 *     def _check_default_executor(self):             # <<<<<<<<<<<<<<
 *         if self._executor_shutdown_called:
 *             raise RuntimeError('Executor shutdown has been called')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_139_check_default_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_138_check_default_executor, "Loop._check_default_executor(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_139_check_default_executor = {"_check_default_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_139_check_default_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_138_check_default_executor};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_139_check_default_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_check_default_executor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_check_default_executor", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_check_default_executor", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_138_check_default_executor(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_138_check_default_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_default_executor", 1);

  /* "uvloop/loop.pyx":3181
 * 
 *     def _check_default_executor(self):
 *         if self._executor_shutdown_called:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Executor shutdown has been called')
 * 
 */
  if (unlikely(__pyx_v_self->_executor_shutdown_called)) {

    /* "uvloop/loop.pyx":3182
 *     def _check_default_executor(self):
 *         if self._executor_shutdown_called:
 *             raise RuntimeError('Executor shutdown has been called')             # <<<<<<<<<<<<<<
 * 
 *     def _asyncgen_finalizer_hook(self, agen):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__122, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 3182, __pyx_L1_error)

    /* "uvloop/loop.pyx":3181
 * 
 *     def _check_default_executor(self):
 *         if self._executor_shutdown_called:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Executor shutdown has been called')
 * 
 */
  }

  /* "uvloop/loop.pyx":3180
 *         return fs_handle
 * 
 *     def _check_default_executor(self):             # <<<<<<<<<<<<<<
 *         if self._executor_shutdown_called:
 *             raise RuntimeError('Executor shutdown has been called')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._check_default_executor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3184
 *             raise RuntimeError('Executor shutdown has been called')
 * 
 *     def _asyncgen_finalizer_hook(self, agen):             # <<<<<<<<<<<<<<
 *         self._asyncgens.discard(agen)
 *         if not self.is_closed():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_141_asyncgen_finalizer_hook(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_140_asyncgen_finalizer_hook, "Loop._asyncgen_finalizer_hook(self, agen)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_141_asyncgen_finalizer_hook = {"_asyncgen_finalizer_hook", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_141_asyncgen_finalizer_hook, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_140_asyncgen_finalizer_hook};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_141_asyncgen_finalizer_hook(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_agen = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_asyncgen_finalizer_hook (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_agen,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_agen)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3184, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_asyncgen_finalizer_hook") < 0)) __PYX_ERR(0, 3184, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_agen = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_asyncgen_finalizer_hook", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3184, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop._asyncgen_finalizer_hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_140_asyncgen_finalizer_hook(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_agen);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_140_asyncgen_finalizer_hook(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_agen) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_asyncgen_finalizer_hook", 1);

  /* "uvloop/loop.pyx":3185
 * 
 *     def _asyncgen_finalizer_hook(self, agen):
 *         self._asyncgens.discard(agen)             # <<<<<<<<<<<<<<
 *         if not self.is_closed():
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_asyncgens, __pyx_n_s_discard); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_agen};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3186
 *     def _asyncgen_finalizer_hook(self, agen):
 *         self._asyncgens.discard(agen)
 *         if not self.is_closed():             # <<<<<<<<<<<<<<
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_closed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 3186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "uvloop/loop.pyx":3187
 *         self._asyncgens.discard(agen)
 *         if not self.is_closed():
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())             # <<<<<<<<<<<<<<
 * 
 *     def _asyncgen_firstiter_hook(self, agen):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call_soon_threadsafe); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_create_task); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_agen, __pyx_n_s_aclose); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __pyx_t_8 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_t_3, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":3186
 *     def _asyncgen_finalizer_hook(self, agen):
 *         self._asyncgens.discard(agen)
 *         if not self.is_closed():             # <<<<<<<<<<<<<<
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())
 * 
 */
  }

  /* "uvloop/loop.pyx":3184
 *             raise RuntimeError('Executor shutdown has been called')
 * 
 *     def _asyncgen_finalizer_hook(self, agen):             # <<<<<<<<<<<<<<
 *         self._asyncgens.discard(agen)
 *         if not self.is_closed():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.Loop._asyncgen_finalizer_hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3189
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())
 * 
 *     def _asyncgen_firstiter_hook(self, agen):             # <<<<<<<<<<<<<<
 *         if self._asyncgens_shutdown_called:
 *             warnings_warn(
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_143_asyncgen_firstiter_hook(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_142_asyncgen_firstiter_hook, "Loop._asyncgen_firstiter_hook(self, agen)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_143_asyncgen_firstiter_hook = {"_asyncgen_firstiter_hook", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_143_asyncgen_firstiter_hook, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_142_asyncgen_firstiter_hook};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_143_asyncgen_firstiter_hook(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_agen = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_asyncgen_firstiter_hook (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_agen,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_agen)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3189, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_asyncgen_firstiter_hook") < 0)) __PYX_ERR(0, 3189, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_agen = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_asyncgen_firstiter_hook", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3189, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop._asyncgen_firstiter_hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_142_asyncgen_firstiter_hook(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_agen);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_142_asyncgen_firstiter_hook(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_agen) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_asyncgen_firstiter_hook", 1);

  /* "uvloop/loop.pyx":3190
 * 
 *     def _asyncgen_firstiter_hook(self, agen):
 *         if self._asyncgens_shutdown_called:             # <<<<<<<<<<<<<<
 *             warnings_warn(
 *                 "asynchronous generator {!r} was scheduled after "
 */
  if (__pyx_v_self->_asyncgens_shutdown_called) {

    /* "uvloop/loop.pyx":3193
 *             warnings_warn(
 *                 "asynchronous generator {!r} was scheduled after "
 *                 "loop.shutdown_asyncgens() call".format(agen),             # <<<<<<<<<<<<<<
 *                 ResourceWarning, source=self)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_asynchronous_generator_r_was_sch, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_agen};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "uvloop/loop.pyx":3194
 *                 "asynchronous generator {!r} was scheduled after "
 *                 "loop.shutdown_asyncgens() call".format(agen),
 *                 ResourceWarning, source=self)             # <<<<<<<<<<<<<<
 * 
 *         self._asyncgens.add(agen)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ResourceWarning); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "uvloop/loop.pyx":3191
 *     def _asyncgen_firstiter_hook(self, agen):
 *         if self._asyncgens_shutdown_called:
 *             warnings_warn(             # <<<<<<<<<<<<<<
 *                 "asynchronous generator {!r} was scheduled after "
 *                 "loop.shutdown_asyncgens() call".format(agen),
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 3191, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 3191, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3194
 *                 "asynchronous generator {!r} was scheduled after "
 *                 "loop.shutdown_asyncgens() call".format(agen),
 *                 ResourceWarning, source=self)             # <<<<<<<<<<<<<<
 * 
 *         self._asyncgens.add(agen)
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_source, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 3194, __pyx_L1_error)

    /* "uvloop/loop.pyx":3191
 *     def _asyncgen_firstiter_hook(self, agen):
 *         if self._asyncgens_shutdown_called:
 *             warnings_warn(             # <<<<<<<<<<<<<<
 *                 "asynchronous generator {!r} was scheduled after "
 *                 "loop.shutdown_asyncgens() call".format(agen),
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_warnings_warn, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":3190
 * 
 *     def _asyncgen_firstiter_hook(self, agen):
 *         if self._asyncgens_shutdown_called:             # <<<<<<<<<<<<<<
 *             warnings_warn(
 *                 "asynchronous generator {!r} was scheduled after "
 */
  }

  /* "uvloop/loop.pyx":3196
 *                 ResourceWarning, source=self)
 * 
 *         self._asyncgens.add(agen)             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_asyncgens, __pyx_n_s_add); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_agen};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3189
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())
 * 
 *     def _asyncgen_firstiter_hook(self, agen):             # <<<<<<<<<<<<<<
 *         if self._asyncgens_shutdown_called:
 *             warnings_warn(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Loop._asyncgen_firstiter_hook", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_146generator22(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":3198
 *         self._asyncgens.add(agen)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_asyncgens(self):
 *         """Shutdown all active asynchronous generators."""
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_145shutdown_asyncgens(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_144shutdown_asyncgens, "Loop.shutdown_asyncgens(self)\nShutdown all active asynchronous generators.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_145shutdown_asyncgens = {"shutdown_asyncgens", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_145shutdown_asyncgens, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_144shutdown_asyncgens};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_145shutdown_asyncgens(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown_asyncgens (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("shutdown_asyncgens", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "shutdown_asyncgens", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_144shutdown_asyncgens(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_144shutdown_asyncgens(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shutdown_asyncgens", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3198, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_146generator22, __pyx_codeobj__123, (PyObject *) __pyx_cur_scope, __pyx_n_s_shutdown_asyncgens, __pyx_n_s_Loop_shutdown_asyncgens, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 3198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.shutdown_asyncgens", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_146generator22(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown_asyncgens", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3198, __pyx_L1_error)

  /* "uvloop/loop.pyx":3201
 *     async def shutdown_asyncgens(self):
 *         """Shutdown all active asynchronous generators."""
 *         self._asyncgens_shutdown_called = True             # <<<<<<<<<<<<<<
 * 
 *         if not len(self._asyncgens):
 */
  __pyx_cur_scope->__pyx_v_self->_asyncgens_shutdown_called = 1;

  /* "uvloop/loop.pyx":3203
 *         self._asyncgens_shutdown_called = True
 * 
 *         if not len(self._asyncgens):             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->_asyncgens;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 3203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (!(__pyx_t_2 != 0));
  if (__pyx_t_3) {

    /* "uvloop/loop.pyx":3204
 * 
 *         if not len(self._asyncgens):
 *             return             # <<<<<<<<<<<<<<
 * 
 *         closing_agens = list(self._asyncgens)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":3203
 *         self._asyncgens_shutdown_called = True
 * 
 *         if not len(self._asyncgens):             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/loop.pyx":3206
 *             return
 * 
 *         closing_agens = list(self._asyncgens)             # <<<<<<<<<<<<<<
 *         self._asyncgens.clear()
 * 
 */
  __pyx_t_1 = PySequence_List(__pyx_cur_scope->__pyx_v_self->_asyncgens); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_closing_agens = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3207
 * 
 *         closing_agens = list(self._asyncgens)
 *         self._asyncgens.clear()             # <<<<<<<<<<<<<<
 * 
 *         shutdown_coro = aio_gather(
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self->_asyncgens, __pyx_n_s_clear); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3209
 *         self._asyncgens.clear()
 * 
 *         shutdown_coro = aio_gather(             # <<<<<<<<<<<<<<
 *             *[ag.aclose() for ag in closing_agens],
 *             return_exceptions=True)
 */
  { /* enter inner scope */

    /* "uvloop/loop.pyx":3210
 * 
 *         shutdown_coro = aio_gather(
 *             *[ag.aclose() for ag in closing_agens],             # <<<<<<<<<<<<<<
 *             return_exceptions=True)
 * 
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_closing_agens; __Pyx_INCREF(__pyx_t_4);
    __pyx_t_2 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 3210, __pyx_L1_error)
        #endif
        if (__pyx_t_2 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 3210, __pyx_L1_error)
      #else
      __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_8genexpr4__pyx_v_ag);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_8genexpr4__pyx_v_ag, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_8genexpr4__pyx_v_ag, __pyx_n_s_aclose); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 3210, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } /* exit inner scope */

  /* "uvloop/loop.pyx":3209
 *         self._asyncgens.clear()
 * 
 *         shutdown_coro = aio_gather(             # <<<<<<<<<<<<<<
 *             *[ag.aclose() for ag in closing_agens],
 *             return_exceptions=True)
 */
  __pyx_t_4 = PySequence_Tuple(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3211
 *         shutdown_coro = aio_gather(
 *             *[ag.aclose() for ag in closing_agens],
 *             return_exceptions=True)             # <<<<<<<<<<<<<<
 * 
 *         results = await shutdown_coro
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_return_exceptions, Py_True) < 0) __PYX_ERR(0, 3211, __pyx_L1_error)

  /* "uvloop/loop.pyx":3209
 *         self._asyncgens.clear()
 * 
 *         shutdown_coro = aio_gather(             # <<<<<<<<<<<<<<
 *             *[ag.aclose() for ag in closing_agens],
 *             return_exceptions=True)
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_gather, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_shutdown_coro = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":3213
 *             return_exceptions=True)
 * 
 *         results = await shutdown_coro             # <<<<<<<<<<<<<<
 *         for result, agen in zip(results, closing_agens):
 *             if isinstance(result, Exception):
 */
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_shutdown_coro);
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L8_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3213, __pyx_L1_error)
    __pyx_t_5 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_5);
  } else {
    __pyx_t_5 = NULL;
    if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_5) < 0) __PYX_ERR(0, 3213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_results = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":3214
 * 
 *         results = await shutdown_coro
 *         for result, agen in zip(results, closing_agens):             # <<<<<<<<<<<<<<
 *             if isinstance(result, Exception):
 *                 self.call_exception_handler({
 */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_results);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_results);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_cur_scope->__pyx_v_results)) __PYX_ERR(0, 3214, __pyx_L1_error);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_closing_agens);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_closing_agens);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_cur_scope->__pyx_v_closing_agens)) __PYX_ERR(0, 3214, __pyx_L1_error);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_5 = __pyx_t_1; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_2 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 3214, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 3214, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 3214, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 3214, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 3214, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3214, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_5);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 3214, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 3214, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
      index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_8); if (unlikely(!__pyx_t_4)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_7 = __pyx_t_10(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L11_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_8), 2) < 0) __PYX_ERR(0, 3214, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L12_unpacking_done;
      __pyx_L11_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 3214, __pyx_L1_error)
      __pyx_L12_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_result);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_result, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_agen);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_agen, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":3215
 *         results = await shutdown_coro
 *         for result, agen in zip(results, closing_agens):
 *             if isinstance(result, Exception):             # <<<<<<<<<<<<<<
 *                 self.call_exception_handler({
 *                     'message': 'an error occurred during closing of '
 */
    __pyx_t_3 = __Pyx_PyException_Check(__pyx_cur_scope->__pyx_v_result); 
    if (__pyx_t_3) {

      /* "uvloop/loop.pyx":3216
 *         for result, agen in zip(results, closing_agens):
 *             if isinstance(result, Exception):
 *                 self.call_exception_handler({             # <<<<<<<<<<<<<<
 *                     'message': 'an error occurred during closing of '
 *                                'asynchronous generator {!r}'.format(agen),
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3216, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "uvloop/loop.pyx":3217
 *             if isinstance(result, Exception):
 *                 self.call_exception_handler({
 *                     'message': 'an error occurred during closing of '             # <<<<<<<<<<<<<<
 *                                'asynchronous generator {!r}'.format(agen),
 *                     'exception': result,
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "uvloop/loop.pyx":3218
 *                 self.call_exception_handler({
 *                     'message': 'an error occurred during closing of '
 *                                'asynchronous generator {!r}'.format(agen),             # <<<<<<<<<<<<<<
 *                     'exception': result,
 *                     'asyncgen': agen
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_an_error_occurred_during_closing, __pyx_n_s_format); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_cur_scope->__pyx_v_agen};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3218, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_message, __pyx_t_8) < 0) __PYX_ERR(0, 3217, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/loop.pyx":3219
 *                     'message': 'an error occurred during closing of '
 *                                'asynchronous generator {!r}'.format(agen),
 *                     'exception': result,             # <<<<<<<<<<<<<<
 *                     'asyncgen': agen
 *                 })
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_exception, __pyx_cur_scope->__pyx_v_result) < 0) __PYX_ERR(0, 3217, __pyx_L1_error)

      /* "uvloop/loop.pyx":3220
 *                                'asynchronous generator {!r}'.format(agen),
 *                     'exception': result,
 *                     'asyncgen': agen             # <<<<<<<<<<<<<<
 *                 })
 * 
 */
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_asyncgen, __pyx_cur_scope->__pyx_v_agen) < 0) __PYX_ERR(0, 3217, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3216, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/loop.pyx":3215
 *         results = await shutdown_coro
 *         for result, agen in zip(results, closing_agens):
 *             if isinstance(result, Exception):             # <<<<<<<<<<<<<<
 *                 self.call_exception_handler({
 *                     'message': 'an error occurred during closing of '
 */
    }

    /* "uvloop/loop.pyx":3214
 * 
 *         results = await shutdown_coro
 *         for result, agen in zip(results, closing_agens):             # <<<<<<<<<<<<<<
 *             if isinstance(result, Exception):
 *                 self.call_exception_handler({
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":3198
 *         self._asyncgens.add(agen)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_asyncgens(self):
 *         """Shutdown all active asynchronous generators."""
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("shutdown_asyncgens", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_4Loop_149generator23(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":3223
 *                 })
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_default_executor(self, timeout=None):
 *         """Schedule the shutdown of the default executor.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_148shutdown_default_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_147shutdown_default_executor, "Loop.shutdown_default_executor(self, timeout=None)\nSchedule the shutdown of the default executor.\n\n        The timeout parameter specifies the amount of time the executor will\n        be given to finish joining. The default value is None, which means\n        that the executor will be given an unlimited amount of time.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_148shutdown_default_executor = {"shutdown_default_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_148shutdown_default_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_147shutdown_default_executor};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_148shutdown_default_executor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timeout = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown_default_executor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_timeout,0};

    /* "uvloop/loop.pyx":3224
 * 
 *     @cython.iterable_coroutine
 *     async def shutdown_default_executor(self, timeout=None):             # <<<<<<<<<<<<<<
 *         """Schedule the shutdown of the default executor.
 * 
 */
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3223, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "shutdown_default_executor") < 0)) __PYX_ERR(0, 3223, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_timeout = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("shutdown_default_executor", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 3223, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.shutdown_default_executor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_147shutdown_default_executor(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_timeout);

  /* "uvloop/loop.pyx":3223
 *                 })
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_default_executor(self, timeout=None):
 *         """Schedule the shutdown of the default executor.
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_147shutdown_default_executor(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_timeout) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shutdown_default_executor", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3223, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_timeout = __pyx_v_timeout;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timeout);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timeout);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_4Loop_149generator23, __pyx_codeobj__124, (PyObject *) __pyx_cur_scope, __pyx_n_s_shutdown_default_executor, __pyx_n_s_Loop_shutdown_default_executor, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 3223, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.shutdown_default_executor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_4Loop_149generator23(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_UCS4 __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown_default_executor", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3223, __pyx_L1_error)

  /* "uvloop/loop.pyx":3231
 *         that the executor will be given an unlimited amount of time.
 *         """
 *         self._executor_shutdown_called = True             # <<<<<<<<<<<<<<
 *         if self._default_executor is None:
 *             return
 */
  __pyx_cur_scope->__pyx_v_self->_executor_shutdown_called = 1;

  /* "uvloop/loop.pyx":3232
 *         """
 *         self._executor_shutdown_called = True
 *         if self._default_executor is None:             # <<<<<<<<<<<<<<
 *             return
 *         future = self.create_future()
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_self->_default_executor == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3233
 *         self._executor_shutdown_called = True
 *         if self._default_executor is None:
 *             return             # <<<<<<<<<<<<<<
 *         future = self.create_future()
 *         thread = threading_Thread(target=self._do_shutdown, args=(future,))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":3232
 *         """
 *         self._executor_shutdown_called = True
 *         if self._default_executor is None:             # <<<<<<<<<<<<<<
 *             return
 *         future = self.create_future()
 */
  }

  /* "uvloop/loop.pyx":3234
 *         if self._default_executor is None:
 *             return
 *         future = self.create_future()             # <<<<<<<<<<<<<<
 *         thread = threading_Thread(target=self._do_shutdown, args=(future,))
 *         thread.start()
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_create_future); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_future = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3235
 *             return
 *         future = self.create_future()
 *         thread = threading_Thread(target=self._do_shutdown, args=(future,))             # <<<<<<<<<<<<<<
 *         thread.start()
 *         try:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_do_shutdown); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_target, __pyx_t_3) < 0) __PYX_ERR(0, 3235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_future);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_future);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_cur_scope->__pyx_v_future)) __PYX_ERR(0, 3235, __pyx_L1_error);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_args, __pyx_t_3) < 0) __PYX_ERR(0, 3235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_threading_Thread, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_thread = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":3236
 *         future = self.create_future()
 *         thread = threading_Thread(target=self._do_shutdown, args=(future,))
 *         thread.start()             # <<<<<<<<<<<<<<
 *         try:
 *             await future
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_thread, __pyx_n_s_start); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":3237
 *         thread = threading_Thread(target=self._do_shutdown, args=(future,))
 *         thread.start()
 *         try:             # <<<<<<<<<<<<<<
 *             await future
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/loop.pyx":3238
 *         thread.start()
 *         try:
 *             await future             # <<<<<<<<<<<<<<
 *         finally:
 *             thread.join(timeout)
 */
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_future);
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L8_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3238, __pyx_L6_error)
    } else {
      PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
      if (exc_type) {
        if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
        else __PYX_ERR(0, 3238, __pyx_L6_error)
      }
    }
  }

  /* "uvloop/loop.pyx":3240
 *             await future
 *         finally:
 *             thread.join(timeout)             # <<<<<<<<<<<<<<
 * 
 *         if thread.is_alive():
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_thread, __pyx_n_s_join); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_timeout};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_thread, __pyx_n_s_join); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3240, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_timeout};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3240, __pyx_L10_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
      __pyx_L10_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "uvloop/loop.pyx":3242
 *             thread.join(timeout)
 * 
 *         if thread.is_alive():             # <<<<<<<<<<<<<<
 *             warnings_warn(
 *                 "The executor did not finishing joining "
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_thread, __pyx_n_s_is_alive); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 3242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3244
 *         if thread.is_alive():
 *             warnings_warn(
 *                 "The executor did not finishing joining "             # <<<<<<<<<<<<<<
 *                 f"its threads within {timeout} seconds.",
 *                 RuntimeWarning,
 */
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_15 = 0;
    __pyx_t_16 = 127;
    __Pyx_INCREF(__pyx_kp_u_The_executor_did_not_finishing_j);
    __pyx_t_15 += 58;
    __Pyx_GIVEREF(__pyx_kp_u_The_executor_did_not_finishing_j);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_The_executor_did_not_finishing_j);

    /* "uvloop/loop.pyx":3245
 *             warnings_warn(
 *                 "The executor did not finishing joining "
 *                 f"its threads within {timeout} seconds.",             # <<<<<<<<<<<<<<
 *                 RuntimeWarning,
 *                 stacklevel=2
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_cur_scope->__pyx_v_timeout, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_16;
    __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_seconds);
    __pyx_t_15 += 9;
    __Pyx_GIVEREF(__pyx_kp_u_seconds);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_seconds);

    /* "uvloop/loop.pyx":3244
 *         if thread.is_alive():
 *             warnings_warn(
 *                 "The executor did not finishing joining "             # <<<<<<<<<<<<<<
 *                 f"its threads within {timeout} seconds.",
 *                 RuntimeWarning,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":3243
 * 
 *         if thread.is_alive():
 *             warnings_warn(             # <<<<<<<<<<<<<<
 *                 "The executor did not finishing joining "
 *                 f"its threads within {timeout} seconds.",
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 3243, __pyx_L1_error);
    __Pyx_INCREF(__pyx_builtin_RuntimeWarning);
    __Pyx_GIVEREF(__pyx_builtin_RuntimeWarning);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_builtin_RuntimeWarning)) __PYX_ERR(0, 3243, __pyx_L1_error);
    __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3247
 *                 f"its threads within {timeout} seconds.",
 *                 RuntimeWarning,
 *                 stacklevel=2             # <<<<<<<<<<<<<<
 *             )
 *             self._default_executor.shutdown(wait=False)
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_stacklevel, __pyx_int_2) < 0) __PYX_ERR(0, 3247, __pyx_L1_error)

    /* "uvloop/loop.pyx":3243
 * 
 *         if thread.is_alive():
 *             warnings_warn(             # <<<<<<<<<<<<<<
 *                 "The executor did not finishing joining "
 *                 f"its threads within {timeout} seconds.",
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_warnings_warn, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/loop.pyx":3249
 *                 stacklevel=2
 *             )
 *             self._default_executor.shutdown(wait=False)             # <<<<<<<<<<<<<<
 * 
 *     def _do_shutdown(self, future):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self->_default_executor, __pyx_n_s_shutdown); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_wait, Py_False) < 0) __PYX_ERR(0, 3249, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/loop.pyx":3242
 *             thread.join(timeout)
 * 
 *         if thread.is_alive():             # <<<<<<<<<<<<<<
 *             warnings_warn(
 *                 "The executor did not finishing joining "
 */
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/loop.pyx":3223
 *                 })
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_default_executor(self, timeout=None):
 *         """Schedule the shutdown of the default executor.
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("shutdown_default_executor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3251
 *             self._default_executor.shutdown(wait=False)
 * 
 *     def _do_shutdown(self, future):             # <<<<<<<<<<<<<<
 *         try:
 *             self._default_executor.shutdown(wait=True)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_151_do_shutdown(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_150_do_shutdown, "Loop._do_shutdown(self, future)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_151_do_shutdown = {"_do_shutdown", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_151_do_shutdown, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_150_do_shutdown};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_151_do_shutdown(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_future = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_do_shutdown (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_future,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_future)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3251, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_do_shutdown") < 0)) __PYX_ERR(0, 3251, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_future = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_do_shutdown", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3251, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop._do_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_150_do_shutdown(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v_future);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_150_do_shutdown(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_future) {
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_shutdown", 1);

  /* "uvloop/loop.pyx":3252
 * 
 *     def _do_shutdown(self, future):
 *         try:             # <<<<<<<<<<<<<<
 *             self._default_executor.shutdown(wait=True)
 *             self.call_soon_threadsafe(future.set_result, None)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/loop.pyx":3253
 *     def _do_shutdown(self, future):
 *         try:
 *             self._default_executor.shutdown(wait=True)             # <<<<<<<<<<<<<<
 *             self.call_soon_threadsafe(future.set_result, None)
 *         except Exception as ex:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_default_executor, __pyx_n_s_shutdown); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3253, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3253, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_wait, Py_True) < 0) __PYX_ERR(0, 3253, __pyx_L3_error)
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3253, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":3254
 *         try:
 *             self._default_executor.shutdown(wait=True)
 *             self.call_soon_threadsafe(future.set_result, None)             # <<<<<<<<<<<<<<
 *         except Exception as ex:
 *             self.call_soon_threadsafe(future.set_exception, ex)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call_soon_threadsafe); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3254, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_set_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3254, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_4, Py_None};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3254, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/loop.pyx":3252
 * 
 *     def _do_shutdown(self, future):
 *         try:             # <<<<<<<<<<<<<<
 *             self._default_executor.shutdown(wait=True)
 *             self.call_soon_threadsafe(future.set_result, None)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/loop.pyx":3255
 *             self._default_executor.shutdown(wait=True)
 *             self.call_soon_threadsafe(future.set_result, None)
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self.call_soon_threadsafe(future.set_exception, ex)
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.Loop._do_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(0, 3255, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_ex = __pyx_t_5;
      /*try:*/ {

        /* "uvloop/loop.pyx":3256
 *             self.call_soon_threadsafe(future.set_result, None)
 *         except Exception as ex:
 *             self.call_soon_threadsafe(future.set_exception, ex)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_call_soon_threadsafe); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 3256, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_future, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 3256, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_t_11, __pyx_v_ex};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 3256, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }

      /* "uvloop/loop.pyx":3255
 *             self._default_executor.shutdown(wait=True)
 *             self.call_soon_threadsafe(future.set_result, None)
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self.call_soon_threadsafe(future.set_exception, ex)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L15;
        }
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_9 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L5_except_error;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/loop.pyx":3252
 * 
 *     def _do_shutdown(self, future):
 *         try:             # <<<<<<<<<<<<<<
 *             self._default_executor.shutdown(wait=True)
 *             self.call_soon_threadsafe(future.set_result, None)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/loop.pyx":3251
 *             self._default_executor.shutdown(wait=False)
 * 
 *     def _do_shutdown(self, future):             # <<<<<<<<<<<<<<
 *         try:
 *             self._default_executor.shutdown(wait=True)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uvloop.loop.Loop._do_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":39
 *         int _coroutine_origin_tracking_saved_depth
 * 
 *         public slow_callback_duration             # <<<<<<<<<<<<<<
 * 
 *         readonly bint _closed
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->slow_callback_duration);
  __pyx_r = __pyx_v_self->slow_callback_duration;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration_2__set__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration_2__set__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->slow_callback_duration);
  __Pyx_DECREF(__pyx_v_self->slow_callback_duration);
  __pyx_v_self->slow_callback_duration = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_5__del__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration_4__del__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_4Loop_22slow_callback_duration_4__del__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->slow_callback_duration);
  __Pyx_DECREF(__pyx_v_self->slow_callback_duration);
  __pyx_v_self->slow_callback_duration = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":41
 *         public slow_callback_duration
 * 
 *         readonly bint _closed             # <<<<<<<<<<<<<<
 *         bint _debug
 *         bint _running
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_7_closed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_7_closed_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_7_closed___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_7_closed___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._closed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":94
 *         # True when compiled with DEBUG.
 *         # Used only in unittests.
 *         readonly bint _debug_cc             # <<<<<<<<<<<<<<
 * 
 *         readonly object _debug_handles_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_9_debug_cc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_9_debug_cc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_9_debug_cc___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_9_debug_cc___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_debug_cc); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_cc.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":96
 *         readonly bint _debug_cc
 * 
 *         readonly object _debug_handles_total             # <<<<<<<<<<<<<<
 *         readonly object _debug_handles_closed
 *         readonly object _debug_handles_current
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_20_debug_handles_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_20_debug_handles_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_20_debug_handles_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_20_debug_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_debug_handles_total);
  __pyx_r = __pyx_v_self->_debug_handles_total;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":97
 * 
 *         readonly object _debug_handles_total
 *         readonly object _debug_handles_closed             # <<<<<<<<<<<<<<
 *         readonly object _debug_handles_current
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_21_debug_handles_closed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_21_debug_handles_closed_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_21_debug_handles_closed___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_21_debug_handles_closed___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_debug_handles_closed);
  __pyx_r = __pyx_v_self->_debug_handles_closed;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":98
 *         readonly object _debug_handles_total
 *         readonly object _debug_handles_closed
 *         readonly object _debug_handles_current             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _debug_uv_handles_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_22_debug_handles_current_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_22_debug_handles_current_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_22_debug_handles_current___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_22_debug_handles_current___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_debug_handles_current);
  __pyx_r = __pyx_v_self->_debug_handles_current;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":100
 *         readonly object _debug_handles_current
 * 
 *         readonly uint64_t _debug_uv_handles_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_uv_handles_freed
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_uv_handles_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_uv_handles_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_23_debug_uv_handles_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_uv_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_uv_handles_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_uv_handles_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":101
 * 
 *         readonly uint64_t _debug_uv_handles_total
 *         readonly uint64_t _debug_uv_handles_freed             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _debug_cb_handles_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_uv_handles_freed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_uv_handles_freed_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_23_debug_uv_handles_freed___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_uv_handles_freed___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_uv_handles_freed); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_uv_handles_freed.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":103
 *         readonly uint64_t _debug_uv_handles_freed
 * 
 *         readonly uint64_t _debug_cb_handles_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_cb_handles_count
 *         readonly uint64_t _debug_cb_timer_handles_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_cb_handles_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_cb_handles_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_23_debug_cb_handles_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_cb_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_handles_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_cb_handles_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":104
 * 
 *         readonly uint64_t _debug_cb_handles_total
 *         readonly uint64_t _debug_cb_handles_count             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_cb_timer_handles_total
 *         readonly uint64_t _debug_cb_timer_handles_count
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_cb_handles_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_debug_cb_handles_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_23_debug_cb_handles_count___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_debug_cb_handles_count___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_handles_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_cb_handles_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":105
 *         readonly uint64_t _debug_cb_handles_total
 *         readonly uint64_t _debug_cb_handles_count
 *         readonly uint64_t _debug_cb_timer_handles_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_cb_timer_handles_count
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_timer_handles_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_cb_timer_handles_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":106
 *         readonly uint64_t _debug_cb_handles_count
 *         readonly uint64_t _debug_cb_timer_handles_total
 *         readonly uint64_t _debug_cb_timer_handles_count             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _debug_stream_shutdown_errors_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_count___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_count___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_cb_timer_handles_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_cb_timer_handles_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":108
 *         readonly uint64_t _debug_cb_timer_handles_count
 * 
 *         readonly uint64_t _debug_stream_shutdown_errors_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_listen_errors_total
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_35_debug_stream_shutdown_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_35_debug_stream_shutdown_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_35_debug_stream_shutdown_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_35_debug_stream_shutdown_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_shutdown_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_shutdown_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":109
 * 
 *         readonly uint64_t _debug_stream_shutdown_errors_total
 *         readonly uint64_t _debug_stream_listen_errors_total             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _debug_stream_read_cb_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_33_debug_stream_listen_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_33_debug_stream_listen_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_33_debug_stream_listen_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_33_debug_stream_listen_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_listen_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_listen_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":111
 *         readonly uint64_t _debug_stream_listen_errors_total
 * 
 *         readonly uint64_t _debug_stream_read_cb_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_read_cb_errors_total
 *         readonly uint64_t _debug_stream_read_eof_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27_debug_stream_read_cb_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27_debug_stream_read_cb_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_27_debug_stream_read_cb_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_27_debug_stream_read_cb_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_cb_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_read_cb_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":112
 * 
 *         readonly uint64_t _debug_stream_read_cb_total
 *         readonly uint64_t _debug_stream_read_cb_errors_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_read_eof_total
 *         readonly uint64_t _debug_stream_read_eof_cb_errors_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_34_debug_stream_read_cb_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_34_debug_stream_read_cb_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_34_debug_stream_read_cb_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_34_debug_stream_read_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_cb_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_read_cb_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":113
 *         readonly uint64_t _debug_stream_read_cb_total
 *         readonly uint64_t _debug_stream_read_cb_errors_total
 *         readonly uint64_t _debug_stream_read_eof_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_read_eof_cb_errors_total
 *         readonly uint64_t _debug_stream_read_errors_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_28_debug_stream_read_eof_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_28_debug_stream_read_eof_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_28_debug_stream_read_eof_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_28_debug_stream_read_eof_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_eof_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_read_eof_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":114
 *         readonly uint64_t _debug_stream_read_cb_errors_total
 *         readonly uint64_t _debug_stream_read_eof_total
 *         readonly uint64_t _debug_stream_read_eof_cb_errors_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_read_errors_total
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_38_debug_stream_read_eof_cb_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_38_debug_stream_read_eof_cb_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_38_debug_stream_read_eof_cb_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_38_debug_stream_read_eof_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_eof_cb_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_read_eof_cb_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":115
 *         readonly uint64_t _debug_stream_read_eof_total
 *         readonly uint64_t _debug_stream_read_eof_cb_errors_total
 *         readonly uint64_t _debug_stream_read_errors_total             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _debug_stream_write_tries
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_31_debug_stream_read_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_31_debug_stream_read_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_31_debug_stream_read_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_31_debug_stream_read_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_read_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_read_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":117
 *         readonly uint64_t _debug_stream_read_errors_total
 * 
 *         readonly uint64_t _debug_stream_write_tries             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_write_errors_total
 *         readonly uint64_t _debug_stream_write_ctx_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_25_debug_stream_write_tries_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_25_debug_stream_write_tries_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_25_debug_stream_write_tries___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_25_debug_stream_write_tries___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_tries); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_write_tries.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":118
 * 
 *         readonly uint64_t _debug_stream_write_tries
 *         readonly uint64_t _debug_stream_write_errors_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_write_ctx_total
 *         readonly uint64_t _debug_stream_write_ctx_cnt
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_32_debug_stream_write_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_32_debug_stream_write_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_32_debug_stream_write_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_32_debug_stream_write_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_write_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":119
 *         readonly uint64_t _debug_stream_write_tries
 *         readonly uint64_t _debug_stream_write_errors_total
 *         readonly uint64_t _debug_stream_write_ctx_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_write_ctx_cnt
 *         readonly uint64_t _debug_stream_write_cb_errors_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29_debug_stream_write_ctx_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_29_debug_stream_write_ctx_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_29_debug_stream_write_ctx_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_29_debug_stream_write_ctx_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_ctx_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_write_ctx_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":120
 *         readonly uint64_t _debug_stream_write_errors_total
 *         readonly uint64_t _debug_stream_write_ctx_total
 *         readonly uint64_t _debug_stream_write_ctx_cnt             # <<<<<<<<<<<<<<
 *         readonly uint64_t _debug_stream_write_cb_errors_total
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27_debug_stream_write_ctx_cnt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27_debug_stream_write_ctx_cnt_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_27_debug_stream_write_ctx_cnt___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_27_debug_stream_write_ctx_cnt___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_ctx_cnt); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_write_ctx_cnt.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":121
 *         readonly uint64_t _debug_stream_write_ctx_total
 *         readonly uint64_t _debug_stream_write_ctx_cnt
 *         readonly uint64_t _debug_stream_write_cb_errors_total             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _poll_read_events_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_35_debug_stream_write_cb_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_35_debug_stream_write_cb_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_35_debug_stream_write_cb_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_35_debug_stream_write_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_stream_write_cb_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_stream_write_cb_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":123
 *         readonly uint64_t _debug_stream_write_cb_errors_total
 * 
 *         readonly uint64_t _poll_read_events_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _poll_read_cb_errors_total
 *         readonly uint64_t _poll_write_events_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_poll_read_events_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_23_poll_read_events_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_23_poll_read_events_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_23_poll_read_events_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_read_events_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._poll_read_events_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":124
 * 
 *         readonly uint64_t _poll_read_events_total
 *         readonly uint64_t _poll_read_cb_errors_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _poll_write_events_total
 *         readonly uint64_t _poll_write_cb_errors_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_26_poll_read_cb_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_26_poll_read_cb_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_26_poll_read_cb_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_26_poll_read_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_read_cb_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._poll_read_cb_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":125
 *         readonly uint64_t _poll_read_events_total
 *         readonly uint64_t _poll_read_cb_errors_total
 *         readonly uint64_t _poll_write_events_total             # <<<<<<<<<<<<<<
 *         readonly uint64_t _poll_write_cb_errors_total
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_24_poll_write_events_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_24_poll_write_events_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_24_poll_write_events_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_24_poll_write_events_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_write_events_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._poll_write_events_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":126
 *         readonly uint64_t _poll_read_cb_errors_total
 *         readonly uint64_t _poll_write_events_total
 *         readonly uint64_t _poll_write_cb_errors_total             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _sock_try_write_total
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27_poll_write_cb_errors_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_27_poll_write_cb_errors_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_27_poll_write_cb_errors_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_27_poll_write_cb_errors_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_poll_write_cb_errors_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._poll_write_cb_errors_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":128
 *         readonly uint64_t _poll_write_cb_errors_total
 * 
 *         readonly uint64_t _sock_try_write_total             # <<<<<<<<<<<<<<
 * 
 *         readonly uint64_t _debug_exception_handler_cnt
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_21_sock_try_write_total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_21_sock_try_write_total_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_21_sock_try_write_total___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_21_sock_try_write_total___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_sock_try_write_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._sock_try_write_total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pxd":130
 *         readonly uint64_t _sock_try_write_total
 * 
 *         readonly uint64_t _debug_exception_handler_cnt             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_debug_fields(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_28_debug_exception_handler_cnt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_28_debug_exception_handler_cnt_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_28_debug_exception_handler_cnt___get__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_28_debug_exception_handler_cnt___get__(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t(__pyx_v_self->_debug_exception_handler_cnt); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Loop._debug_exception_handler_cnt.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_153__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_152__reduce_cython__, "Loop.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_153__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_153__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_152__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_153__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_152__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_152__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_155__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4Loop_154__setstate_cython__, "Loop.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_4Loop_155__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_155__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_154__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_4Loop_155__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Loop.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4Loop_154__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4Loop_154__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Loop.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3260
 * 
 * # Expose pointer for integration with other C-extensions
 * def libuv_get_loop_t_ptr(loop):             # <<<<<<<<<<<<<<
 *     return PyCapsule_New(<void *>(<Loop>loop).uvloop, NULL, NULL)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_1libuv_get_loop_t_ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_libuv_get_loop_t_ptr, "libuv_get_loop_t_ptr(loop)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_1libuv_get_loop_t_ptr = {"libuv_get_loop_t_ptr", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_1libuv_get_loop_t_ptr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_libuv_get_loop_t_ptr};
static PyObject *__pyx_pw_6uvloop_4loop_1libuv_get_loop_t_ptr(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_loop = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("libuv_get_loop_t_ptr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3260, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "libuv_get_loop_t_ptr") < 0)) __PYX_ERR(0, 3260, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_loop = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("libuv_get_loop_t_ptr", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3260, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.libuv_get_loop_t_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_libuv_get_loop_t_ptr(__pyx_self, __pyx_v_loop);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_libuv_get_loop_t_ptr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_loop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("libuv_get_loop_t_ptr", 1);

  /* "uvloop/loop.pyx":3261
 * # Expose pointer for integration with other C-extensions
 * def libuv_get_loop_t_ptr(loop):
 *     return PyCapsule_New(<void *>(<Loop>loop).uvloop, NULL, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyCapsule_New(((void *)((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_loop)->uvloop), NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":3260
 * 
 * # Expose pointer for integration with other C-extensions
 * def libuv_get_loop_t_ptr(loop):             # <<<<<<<<<<<<<<
 *     return PyCapsule_New(<void *>(<Loop>loop).uvloop, NULL, NULL)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.libuv_get_loop_t_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3264
 * 
 * 
 * def libuv_get_version():             # <<<<<<<<<<<<<<
 *     return uv.uv_version()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_3libuv_get_version(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_2libuv_get_version, "libuv_get_version()");
static PyMethodDef __pyx_mdef_6uvloop_4loop_3libuv_get_version = {"libuv_get_version", (PyCFunction)__pyx_pw_6uvloop_4loop_3libuv_get_version, METH_NOARGS, __pyx_doc_6uvloop_4loop_2libuv_get_version};
static PyObject *__pyx_pw_6uvloop_4loop_3libuv_get_version(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("libuv_get_version (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_2libuv_get_version(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_2libuv_get_version(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("libuv_get_version", 1);

  /* "uvloop/loop.pyx":3265
 * 
 * def libuv_get_version():
 *     return uv.uv_version()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(uv_version()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":3264
 * 
 * 
 * def libuv_get_version():             # <<<<<<<<<<<<<<
 *     return uv.uv_version()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.libuv_get_version", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3268
 * 
 * 
 * def _testhelper_unwrap_capsuled_pointer(obj):             # <<<<<<<<<<<<<<
 *     return <uint64_t>PyCapsule_GetPointer(obj, NULL)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_5_testhelper_unwrap_capsuled_pointer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_4_testhelper_unwrap_capsuled_pointer, "_testhelper_unwrap_capsuled_pointer(obj)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_5_testhelper_unwrap_capsuled_pointer = {"_testhelper_unwrap_capsuled_pointer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_5_testhelper_unwrap_capsuled_pointer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4_testhelper_unwrap_capsuled_pointer};
static PyObject *__pyx_pw_6uvloop_4loop_5_testhelper_unwrap_capsuled_pointer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_obj = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_testhelper_unwrap_capsuled_pointer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_obj)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3268, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_testhelper_unwrap_capsuled_pointer") < 0)) __PYX_ERR(0, 3268, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_obj = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_testhelper_unwrap_capsuled_pointer", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3268, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._testhelper_unwrap_capsuled_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_4_testhelper_unwrap_capsuled_pointer(__pyx_self, __pyx_v_obj);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_4_testhelper_unwrap_capsuled_pointer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  void *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_testhelper_unwrap_capsuled_pointer", 1);

  /* "uvloop/loop.pyx":3269
 * 
 * def _testhelper_unwrap_capsuled_pointer(obj):
 *     return <uint64_t>PyCapsule_GetPointer(obj, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyCapsule_GetPointer(__pyx_v_obj, NULL); if (unlikely(__pyx_t_1 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 3269, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_uint64_t(((uint64_t)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/loop.pyx":3268
 * 
 * 
 * def _testhelper_unwrap_capsuled_pointer(obj):             # <<<<<<<<<<<<<<
 *     return <uint64_t>PyCapsule_GetPointer(obj, NULL)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop._testhelper_unwrap_capsuled_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3272
 * 
 * 
 * cdef void __loop_alloc_buffer(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* uvhandle,
 *     size_t suggested_size,
 */

static void __pyx_f_6uvloop_4loop___loop_alloc_buffer(uv_handle_t *__pyx_v_uvhandle, CYTHON_UNUSED size_t __pyx_v_suggested_size, uv_buf_t *__pyx_v_buf) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__loop_alloc_buffer", 0);

  /* "uvloop/loop.pyx":3278
 * ) noexcept with gil:
 *     cdef:
 *         Loop loop = (<UVHandle>uvhandle.data)._loop             # <<<<<<<<<<<<<<
 * 
 *     if loop._recv_buffer_in_use == 1:
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_uvhandle->data)->_loop);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3280
 *         Loop loop = (<UVHandle>uvhandle.data)._loop
 * 
 *     if loop._recv_buffer_in_use == 1:             # <<<<<<<<<<<<<<
 *         buf.len = 0
 *         exc = RuntimeError('concurrent allocations')
 */
  __pyx_t_2 = (__pyx_v_loop->_recv_buffer_in_use == 1);
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":3281
 * 
 *     if loop._recv_buffer_in_use == 1:
 *         buf.len = 0             # <<<<<<<<<<<<<<
 *         exc = RuntimeError('concurrent allocations')
 *         loop._handle_exception(exc)
 */
    __pyx_v_buf->len = 0;

    /* "uvloop/loop.pyx":3282
 *     if loop._recv_buffer_in_use == 1:
 *         buf.len = 0
 *         exc = RuntimeError('concurrent allocations')             # <<<<<<<<<<<<<<
 *         loop._handle_exception(exc)
 *         return
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__125, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/loop.pyx":3283
 *         buf.len = 0
 *         exc = RuntimeError('concurrent allocations')
 *         loop._handle_exception(exc)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_loop->__pyx_vtab)->_handle_exception(__pyx_v_loop, __pyx_v_exc); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3283, __pyx_L1_error)

    /* "uvloop/loop.pyx":3284
 *         exc = RuntimeError('concurrent allocations')
 *         loop._handle_exception(exc)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     loop._recv_buffer_in_use = 1
 */
    goto __pyx_L0;

    /* "uvloop/loop.pyx":3280
 *         Loop loop = (<UVHandle>uvhandle.data)._loop
 * 
 *     if loop._recv_buffer_in_use == 1:             # <<<<<<<<<<<<<<
 *         buf.len = 0
 *         exc = RuntimeError('concurrent allocations')
 */
  }

  /* "uvloop/loop.pyx":3286
 *         return
 * 
 *     loop._recv_buffer_in_use = 1             # <<<<<<<<<<<<<<
 *     buf.base = loop._recv_buffer
 *     buf.len = sizeof(loop._recv_buffer)
 */
  __pyx_v_loop->_recv_buffer_in_use = 1;

  /* "uvloop/loop.pyx":3287
 * 
 *     loop._recv_buffer_in_use = 1
 *     buf.base = loop._recv_buffer             # <<<<<<<<<<<<<<
 *     buf.len = sizeof(loop._recv_buffer)
 * 
 */
  __pyx_t_3 = __pyx_v_loop->_recv_buffer;
  __pyx_v_buf->base = __pyx_t_3;

  /* "uvloop/loop.pyx":3288
 *     loop._recv_buffer_in_use = 1
 *     buf.base = loop._recv_buffer
 *     buf.len = sizeof(loop._recv_buffer)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(!__pyx_v_loop)) { __Pyx_RaiseUnboundLocalError("loop"); __PYX_ERR(0, 3288, __pyx_L1_error) }
  __pyx_v_buf->len = (sizeof(__pyx_v_loop->_recv_buffer));

  /* "uvloop/loop.pyx":3272
 * 
 * 
 * cdef void __loop_alloc_buffer(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* uvhandle,
 *     size_t suggested_size,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("uvloop.loop.__loop_alloc_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/loop.pyx":3291
 * 
 * 
 * cdef inline void __loop_free_buffer(Loop loop):             # <<<<<<<<<<<<<<
 *     loop._recv_buffer_in_use = 0
 * 
 */

static CYTHON_INLINE void __pyx_f_6uvloop_4loop___loop_free_buffer(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop) {

  /* "uvloop/loop.pyx":3292
 * 
 * cdef inline void __loop_free_buffer(Loop loop):
 *     loop._recv_buffer_in_use = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_loop->_recv_buffer_in_use = 0;

  /* "uvloop/loop.pyx":3291
 * 
 * 
 * cdef inline void __loop_free_buffer(Loop loop):             # <<<<<<<<<<<<<<
 *     loop._recv_buffer_in_use = 0
 * 
 */

  /* function exit code */
}

/* "uvloop/loop.pyx":3297
 * class _SyncSocketReaderFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture___init__, "_SyncSocketReaderFuture.__init__(self, sock, loop)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture___init__};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_loop = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_sock,&__pyx_n_s_loop,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3297, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3297, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 3297, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3297, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 3297, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 3297, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_sock = values[1];
    __pyx_v_loop = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 3297, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture___init__(__pyx_self, __pyx_v_self, __pyx_v_sock, __pyx_v_loop);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_loop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/loop.pyx":3298
 * 
 *     def __init__(self, sock, loop):
 *         aio_Future.__init__(self, loop=loop)             # <<<<<<<<<<<<<<
 *         self.__sock = sock
 *         self.__loop = loop
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_Future, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self)) __PYX_ERR(0, 3298, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_loop, __pyx_v_loop) < 0) __PYX_ERR(0, 3298, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":3299
 *     def __init__(self, sock, loop):
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock             # <<<<<<<<<<<<<<
 *         self.__loop = loop
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__sock, __pyx_v_sock) < 0) __PYX_ERR(0, 3299, __pyx_L1_error)

  /* "uvloop/loop.pyx":3300
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 *         self.__loop = loop             # <<<<<<<<<<<<<<
 * 
 *     def __remove_reader(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__loop, __pyx_v_loop) < 0) __PYX_ERR(0, 3300, __pyx_L1_error)

  /* "uvloop/loop.pyx":3297
 * class _SyncSocketReaderFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3302
 *         self.__loop = loop
 * 
 *     def __remove_reader(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_reader(self.__sock)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_3__remove_reader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture_2__remove_reader, "_SyncSocketReaderFuture.__remove_reader(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_3__remove_reader = {"__remove_reader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_3__remove_reader, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture_2__remove_reader};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_3__remove_reader(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__remove_reader (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3302, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__remove_reader") < 0)) __PYX_ERR(0, 3302, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__remove_reader", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3302, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.__remove_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_2__remove_reader(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_2__remove_reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__remove_reader", 1);

  /* "uvloop/loop.pyx":3303
 * 
 *     def __remove_reader(self):
 *         if self.__sock is not None and self.__sock.fileno() != -1:             # <<<<<<<<<<<<<<
 *             self.__loop.remove_reader(self.__sock)
 *             self.__sock = None
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__sock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_fileno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_3 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_neg_1, -1L, 0)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 3303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3304
 *     def __remove_reader(self):
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_reader(self.__sock)             # <<<<<<<<<<<<<<
 *             self.__sock = None
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__loop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_remove_reader); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__sock); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3305
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_reader(self.__sock)
 *             self.__sock = None             # <<<<<<<<<<<<<<
 * 
 *     if PY39:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__sock, Py_None) < 0) __PYX_ERR(0, 3305, __pyx_L1_error)

    /* "uvloop/loop.pyx":3303
 * 
 *     def __remove_reader(self):
 *         if self.__sock is not None and self.__sock.fileno() != -1:             # <<<<<<<<<<<<<<
 *             self.__loop.remove_reader(self.__sock)
 *             self.__sock = None
 */
  }

  /* "uvloop/loop.pyx":3302
 *         self.__loop = loop
 * 
 *     def __remove_reader(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_reader(self.__sock)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.__remove_reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3308
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self, msg=msg)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_5cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture_4cancel, "_SyncSocketReaderFuture.cancel(self, msg=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_5cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_5cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture_4cancel};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_5cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_msg = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_msg,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3308, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_msg);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3308, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel") < 0)) __PYX_ERR(0, 3308, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_msg = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 3308, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_4cancel(__pyx_self, __pyx_v_self, __pyx_v_msg);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_4cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/loop.pyx":3309
 *     if PY39:
 *         def cancel(self, msg=None):
 *             self.__remove_reader()             # <<<<<<<<<<<<<<
 *             aio_Future.cancel(self, msg=msg)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3310
 *         def cancel(self, msg=None):
 *             self.__remove_reader()
 *             aio_Future.cancel(self, msg=msg)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_Future, __pyx_n_s_cancel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self)) __PYX_ERR(0, 3310, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_msg, __pyx_v_msg) < 0) __PYX_ERR(0, 3310, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":3308
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self, msg=msg)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3313
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_7cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture_6cancel, "_SyncSocketReaderFuture.cancel(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_7cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_7cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketReaderFuture_6cancel};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketReaderFuture_7cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3313, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel") < 0)) __PYX_ERR(0, 3313, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3313, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_6cancel(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketReaderFuture_6cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/loop.pyx":3314
 *     else:
 *         def cancel(self):
 *             self.__remove_reader()             # <<<<<<<<<<<<<<
 *             aio_Future.cancel(self)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketReaderFuture__remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3315
 *         def cancel(self):
 *             self.__remove_reader()
 *             aio_Future.cancel(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_Future, __pyx_n_s_cancel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3313
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketReaderFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3320
 * class _SyncSocketWriterFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture___init__, "_SyncSocketWriterFuture.__init__(self, sock, loop)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture___init__};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_sock = 0;
  PyObject *__pyx_v_loop = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_sock,&__pyx_n_s_loop,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3320, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3320, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 3320, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3320, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 3320, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 3320, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_sock = values[1];
    __pyx_v_loop = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 3320, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture___init__(__pyx_self, __pyx_v_self, __pyx_v_sock, __pyx_v_loop);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_sock, PyObject *__pyx_v_loop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/loop.pyx":3321
 * 
 *     def __init__(self, sock, loop):
 *         aio_Future.__init__(self, loop=loop)             # <<<<<<<<<<<<<<
 *         self.__sock = sock
 *         self.__loop = loop
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_Future, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self)) __PYX_ERR(0, 3321, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_loop, __pyx_v_loop) < 0) __PYX_ERR(0, 3321, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/loop.pyx":3322
 *     def __init__(self, sock, loop):
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock             # <<<<<<<<<<<<<<
 *         self.__loop = loop
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__sock, __pyx_v_sock) < 0) __PYX_ERR(0, 3322, __pyx_L1_error)

  /* "uvloop/loop.pyx":3323
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 *         self.__loop = loop             # <<<<<<<<<<<<<<
 * 
 *     def __remove_writer(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__loop, __pyx_v_loop) < 0) __PYX_ERR(0, 3323, __pyx_L1_error)

  /* "uvloop/loop.pyx":3320
 * class _SyncSocketWriterFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3325
 *         self.__loop = loop
 * 
 *     def __remove_writer(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_writer(self.__sock)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_3__remove_writer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture_2__remove_writer, "_SyncSocketWriterFuture.__remove_writer(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_3__remove_writer = {"__remove_writer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_3__remove_writer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture_2__remove_writer};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_3__remove_writer(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__remove_writer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3325, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__remove_writer") < 0)) __PYX_ERR(0, 3325, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__remove_writer", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3325, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.__remove_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_2__remove_writer(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_2__remove_writer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__remove_writer", 1);

  /* "uvloop/loop.pyx":3326
 * 
 *     def __remove_writer(self):
 *         if self.__sock is not None and self.__sock.fileno() != -1:             # <<<<<<<<<<<<<<
 *             self.__loop.remove_writer(self.__sock)
 *             self.__sock = None
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__sock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__sock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_fileno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_3 = (__Pyx_PyInt_BoolNeObjC(__pyx_t_2, __pyx_int_neg_1, -1L, 0)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 3326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3327
 *     def __remove_writer(self):
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_writer(self.__sock)             # <<<<<<<<<<<<<<
 *             self.__sock = None
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__loop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_remove_writer); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__sock); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3327, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3328
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_writer(self.__sock)
 *             self.__sock = None             # <<<<<<<<<<<<<<
 * 
 *     if PY39:
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__sock, Py_None) < 0) __PYX_ERR(0, 3328, __pyx_L1_error)

    /* "uvloop/loop.pyx":3326
 * 
 *     def __remove_writer(self):
 *         if self.__sock is not None and self.__sock.fileno() != -1:             # <<<<<<<<<<<<<<
 *             self.__loop.remove_writer(self.__sock)
 *             self.__sock = None
 */
  }

  /* "uvloop/loop.pyx":3325
 *         self.__loop = loop
 * 
 *     def __remove_writer(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_writer(self.__sock)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.__remove_writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3331
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self, msg=msg)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_5cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture_4cancel, "_SyncSocketWriterFuture.cancel(self, msg=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_5cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_5cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture_4cancel};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_5cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_msg = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_msg,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3331, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_msg);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3331, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel") < 0)) __PYX_ERR(0, 3331, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_msg = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 3331, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_4cancel(__pyx_self, __pyx_v_self, __pyx_v_msg);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_4cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/loop.pyx":3332
 *     if PY39:
 *         def cancel(self, msg=None):
 *             self.__remove_writer()             # <<<<<<<<<<<<<<
 *             aio_Future.cancel(self, msg=msg)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3333
 *         def cancel(self, msg=None):
 *             self.__remove_writer()
 *             aio_Future.cancel(self, msg=msg)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_Future, __pyx_n_s_cancel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self)) __PYX_ERR(0, 3333, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_msg, __pyx_v_msg) < 0) __PYX_ERR(0, 3333, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 3333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/loop.pyx":3331
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self, msg=msg)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3336
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_7cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture_6cancel, "_SyncSocketWriterFuture.cancel(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_7cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_7cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23_SyncSocketWriterFuture_6cancel};
static PyObject *__pyx_pw_6uvloop_4loop_23_SyncSocketWriterFuture_7cancel(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3336, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel") < 0)) __PYX_ERR(0, 3336, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3336, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_6cancel(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23_SyncSocketWriterFuture_6cancel(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/loop.pyx":3337
 *     else:
 *         def cancel(self):
 *             self.__remove_writer()             # <<<<<<<<<<<<<<
 *             aio_Future.cancel(self)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_SyncSocketWriterFuture__remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3338
 *         def cancel(self):
 *             self.__remove_writer()
 *             aio_Future.cancel(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_Future, __pyx_n_s_cancel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_self};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/loop.pyx":3336
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SyncSocketWriterFuture.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":4
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class Handle:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._cancelled = 0
 *         self.cb_type = 0
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_6Handle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_6Handle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle___cinit__(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_6Handle___cinit__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/cbhandles.pyx":5
 * cdef class Handle:
 *     def __cinit__(self):
 *         self._cancelled = 0             # <<<<<<<<<<<<<<
 *         self.cb_type = 0
 *         self._source_traceback = None
 */
  __pyx_v_self->_cancelled = 0;

  /* "uvloop/cbhandles.pyx":6
 *     def __cinit__(self):
 *         self._cancelled = 0
 *         self.cb_type = 0             # <<<<<<<<<<<<<<
 *         self._source_traceback = None
 * 
 */
  __pyx_v_self->cb_type = 0;

  /* "uvloop/cbhandles.pyx":7
 *         self._cancelled = 0
 *         self.cb_type = 0
 *         self._source_traceback = None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _set_loop(self, Loop loop):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_source_traceback);
  __Pyx_DECREF(__pyx_v_self->_source_traceback);
  __pyx_v_self->_source_traceback = Py_None;

  /* "uvloop/cbhandles.pyx":4
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class Handle:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._cancelled = 0
 *         self.cb_type = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":9
 *         self._source_traceback = None
 * 
 *     cdef inline _set_loop(self, Loop loop):             # <<<<<<<<<<<<<<
 *         self.loop = loop
 *         if UVLOOP_DEBUG:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__set_loop(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_loop", 1);

  /* "uvloop/cbhandles.pyx":10
 * 
 *     cdef inline _set_loop(self, Loop loop):
 *         self.loop = loop             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_cb_handles_total += 1
 */
  __Pyx_INCREF((PyObject *)__pyx_v_loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_loop);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->loop);
  __pyx_v_self->loop = __pyx_v_loop;

  /* "uvloop/cbhandles.pyx":11
 *     cdef inline _set_loop(self, Loop loop):
 *         self.loop = loop
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_cb_handles_total += 1
 *             loop._debug_cb_handles_count += 1
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":12
 *         self.loop = loop
 *         if UVLOOP_DEBUG:
 *             loop._debug_cb_handles_total += 1             # <<<<<<<<<<<<<<
 *             loop._debug_cb_handles_count += 1
 *         if loop._debug:
 */
    __pyx_v_loop->_debug_cb_handles_total = (__pyx_v_loop->_debug_cb_handles_total + 1);

    /* "uvloop/cbhandles.pyx":13
 *         if UVLOOP_DEBUG:
 *             loop._debug_cb_handles_total += 1
 *             loop._debug_cb_handles_count += 1             # <<<<<<<<<<<<<<
 *         if loop._debug:
 *             self._source_traceback = extract_stack()
 */
    __pyx_v_loop->_debug_cb_handles_count = (__pyx_v_loop->_debug_cb_handles_count + 1);

    /* "uvloop/cbhandles.pyx":11
 *     cdef inline _set_loop(self, Loop loop):
 *         self.loop = loop
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_cb_handles_total += 1
 *             loop._debug_cb_handles_count += 1
 */
  }

  /* "uvloop/cbhandles.pyx":14
 *             loop._debug_cb_handles_total += 1
 *             loop._debug_cb_handles_count += 1
 *         if loop._debug:             # <<<<<<<<<<<<<<
 *             self._source_traceback = extract_stack()
 * 
 */
  if (__pyx_v_loop->_debug) {

    /* "uvloop/cbhandles.pyx":15
 *             loop._debug_cb_handles_count += 1
 *         if loop._debug:
 *             self._source_traceback = extract_stack()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _set_context(self, object context):
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_extract_stack(); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_source_traceback);
    __Pyx_DECREF(__pyx_v_self->_source_traceback);
    __pyx_v_self->_source_traceback = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":14
 *             loop._debug_cb_handles_total += 1
 *             loop._debug_cb_handles_count += 1
 *         if loop._debug:             # <<<<<<<<<<<<<<
 *             self._source_traceback = extract_stack()
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":9
 *         self._source_traceback = None
 * 
 *     cdef inline _set_loop(self, Loop loop):             # <<<<<<<<<<<<<<
 *         self.loop = loop
 *         if UVLOOP_DEBUG:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Handle._set_loop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":17
 *             self._source_traceback = extract_stack()
 * 
 *     cdef inline _set_context(self, object context):             # <<<<<<<<<<<<<<
 *         if context is None:
 *             context = Context_CopyCurrent()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__set_context(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_context", 0);
  __Pyx_INCREF(__pyx_v_context);

  /* "uvloop/cbhandles.pyx":18
 * 
 *     cdef inline _set_context(self, object context):
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  __pyx_t_1 = (__pyx_v_context == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":19
 *     cdef inline _set_context(self, object context):
 *         if context is None:
 *             context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 *         self.context = context
 * 
 */
    __pyx_t_2 = Context_CopyCurrent(); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_context, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":18
 * 
 *     cdef inline _set_context(self, object context):
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  }

  /* "uvloop/cbhandles.pyx":20
 *         if context is None:
 *             context = Context_CopyCurrent()
 *         self.context = context             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(__pyx_v_context);
  __Pyx_GIVEREF(__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->context);
  __Pyx_DECREF(__pyx_v_self->context);
  __pyx_v_self->context = __pyx_v_context;

  /* "uvloop/cbhandles.pyx":17
 *             self._source_traceback = extract_stack()
 * 
 *     cdef inline _set_context(self, object context):             # <<<<<<<<<<<<<<
 *         if context is None:
 *             context = Context_CopyCurrent()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.Handle._set_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_context);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":22
 *         self.context = context
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG and self.loop is not None:
 *             self.loop._debug_cb_handles_count -= 1
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_6Handle_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_6Handle_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_6Handle_2__dealloc__(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_6Handle_2__dealloc__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/cbhandles.pyx":23
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG and self.loop is not None:             # <<<<<<<<<<<<<<
 *             self.loop._debug_cb_handles_count -= 1
 *         if self.loop is None:
 */
  __pyx_t_2 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (((PyObject *)__pyx_v_self->loop) != Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":24
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG and self.loop is not None:
 *             self.loop._debug_cb_handles_count -= 1             # <<<<<<<<<<<<<<
 *         if self.loop is None:
 *             raise RuntimeError('Handle.loop is None in Handle.__dealloc__')
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self->loop);
    __pyx_t_3 = __pyx_v_self->loop;
    __pyx_t_3->_debug_cb_handles_count = (__pyx_t_3->_debug_cb_handles_count - 1);
    __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/cbhandles.pyx":23
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG and self.loop is not None:             # <<<<<<<<<<<<<<
 *             self.loop._debug_cb_handles_count -= 1
 *         if self.loop is None:
 */
  }

  /* "uvloop/cbhandles.pyx":25
 *         if UVLOOP_DEBUG and self.loop is not None:
 *             self.loop._debug_cb_handles_count -= 1
 *         if self.loop is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Handle.loop is None in Handle.__dealloc__')
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->loop) == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/cbhandles.pyx":26
 *             self.loop._debug_cb_handles_count -= 1
 *         if self.loop is None:
 *             raise RuntimeError('Handle.loop is None in Handle.__dealloc__')             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__126, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(10, 26, __pyx_L1_error)

    /* "uvloop/cbhandles.pyx":25
 *         if UVLOOP_DEBUG and self.loop is not None:
 *             self.loop._debug_cb_handles_count -= 1
 *         if self.loop is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Handle.loop is None in Handle.__dealloc__')
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":22
 *         self.context = context
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG and self.loop is not None:
 *             self.loop._debug_cb_handles_count -= 1
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF((PyObject *)__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uvloop.loop.Handle.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/cbhandles.pyx":28
 *             raise RuntimeError('Handle.loop is None in Handle.__dealloc__')
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_6Handle_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_6Handle_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle_4__init__(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_6Handle_4__init__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/cbhandles.pyx":30
 *     def __init__(self):
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(             # <<<<<<<<<<<<<<
 *                 self.__class__.__name__))
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_is_not_supposed_to_be_instantia, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/cbhandles.pyx":31
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(
 *                 self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _run(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/cbhandles.pyx":29
 * 
 *     def __init__(self):
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             '{} is not supposed to be instantiated from Python'.format(
 *                 self.__class__.__name__))
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(10, 29, __pyx_L1_error)

  /* "uvloop/cbhandles.pyx":28
 *             raise RuntimeError('Handle.loop is None in Handle.__dealloc__')
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Handle.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":33
 *                 self.__class__.__name__))
 * 
 *     cdef inline _run(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int cb_type
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6Handle__run(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  int __pyx_v_cb_type;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_v_context = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  int __pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_run", 1);

  /* "uvloop/cbhandles.pyx":38
 *             object callback
 * 
 *         if self._cancelled:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_self->_cancelled) {

    /* "uvloop/cbhandles.pyx":39
 * 
 *         if self._cancelled:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         cb_type = self.cb_type
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/cbhandles.pyx":38
 *             object callback
 * 
 *         if self._cancelled:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":41
 *             return
 * 
 *         cb_type = self.cb_type             # <<<<<<<<<<<<<<
 * 
 *         # Since _run is a cdef and there's no BoundMethod,
 */
  __pyx_t_1 = __pyx_v_self->cb_type;
  __pyx_v_cb_type = __pyx_t_1;

  /* "uvloop/cbhandles.pyx":46
 *         # we guard 'self' manually (since the callback
 *         # might cause GC of the handle.)
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "uvloop/cbhandles.pyx":48
 *         Py_INCREF(self)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.context is not None
 *             Context_Enter(self.context)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/cbhandles.pyx":49
 * 
 *         try:
 *             assert self.context is not None             # <<<<<<<<<<<<<<
 *             Context_Enter(self.context)
 * 
 */
        #ifndef CYTHON_WITHOUT_ASSERTIONS
        if (unlikely(__pyx_assertions_enabled())) {
          __pyx_t_5 = (__pyx_v_self->context != Py_None);
          if (unlikely(!__pyx_t_5)) {
            __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
            __PYX_ERR(10, 49, __pyx_L7_error)
          }
        }
        #else
        if ((1)); else __PYX_ERR(10, 49, __pyx_L7_error)
        #endif

        /* "uvloop/cbhandles.pyx":50
 *         try:
 *             assert self.context is not None
 *             Context_Enter(self.context)             # <<<<<<<<<<<<<<
 * 
 *             if cb_type == 1:
 */
        __pyx_t_6 = __pyx_v_self->context;
        __Pyx_INCREF(__pyx_t_6);
        __pyx_t_1 = Context_Enter(__pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(10, 50, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "uvloop/cbhandles.pyx":52
 *             Context_Enter(self.context)
 * 
 *             if cb_type == 1:             # <<<<<<<<<<<<<<
 *                 callback = self.arg1
 *                 if callback is None:
 */
        switch (__pyx_v_cb_type) {
          case 1:

          /* "uvloop/cbhandles.pyx":53
 * 
 *             if cb_type == 1:
 *                 callback = self.arg1             # <<<<<<<<<<<<<<
 *                 if callback is None:
 *                     raise RuntimeError(
 */
          __pyx_t_6 = __pyx_v_self->arg1;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_v_callback = __pyx_t_6;
          __pyx_t_6 = 0;

          /* "uvloop/cbhandles.pyx":54
 *             if cb_type == 1:
 *                 callback = self.arg1
 *                 if callback is None:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         'cannot run Handle; callback is not set')
 */
          __pyx_t_5 = (__pyx_v_callback == Py_None);
          if (unlikely(__pyx_t_5)) {

            /* "uvloop/cbhandles.pyx":55
 *                 callback = self.arg1
 *                 if callback is None:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         'cannot run Handle; callback is not set')
 * 
 */
            __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__127, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 55, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_Raise(__pyx_t_6, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __PYX_ERR(10, 55, __pyx_L7_error)

            /* "uvloop/cbhandles.pyx":54
 *             if cb_type == 1:
 *                 callback = self.arg1
 *                 if callback is None:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         'cannot run Handle; callback is not set')
 */
          }

          /* "uvloop/cbhandles.pyx":58
 *                         'cannot run Handle; callback is not set')
 * 
 *                 args = self.arg2             # <<<<<<<<<<<<<<
 * 
 *                 if args is None:
 */
          __pyx_t_6 = __pyx_v_self->arg2;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_v_args = __pyx_t_6;
          __pyx_t_6 = 0;

          /* "uvloop/cbhandles.pyx":60
 *                 args = self.arg2
 * 
 *                 if args is None:             # <<<<<<<<<<<<<<
 *                     callback()
 *                 else:
 */
          __pyx_t_5 = (__pyx_v_args == Py_None);
          if (__pyx_t_5) {

            /* "uvloop/cbhandles.pyx":61
 * 
 *                 if args is None:
 *                     callback()             # <<<<<<<<<<<<<<
 *                 else:
 *                     callback(*args)
 */
            __Pyx_INCREF(__pyx_v_callback);
            __pyx_t_7 = __pyx_v_callback; __pyx_t_8 = NULL;
            __pyx_t_9 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_7))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_7, function);
                __pyx_t_9 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
              __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 61, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "uvloop/cbhandles.pyx":60
 *                 args = self.arg2
 * 
 *                 if args is None:             # <<<<<<<<<<<<<<
 *                     callback()
 *                 else:
 */
            goto __pyx_L14;
          }

          /* "uvloop/cbhandles.pyx":63
 *                     callback()
 *                 else:
 *                     callback(*args)             # <<<<<<<<<<<<<<
 * 
 *             elif cb_type == 2:
 */
          /*else*/ {
            __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 63, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_v_callback, __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 63, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __pyx_L14:;

          /* "uvloop/cbhandles.pyx":52
 *             Context_Enter(self.context)
 * 
 *             if cb_type == 1:             # <<<<<<<<<<<<<<
 *                 callback = self.arg1
 *                 if callback is None:
 */
          break;
          case 2:

          /* "uvloop/cbhandles.pyx":66
 * 
 *             elif cb_type == 2:
 *                 (<method_t>self.callback)(self.arg1)             # <<<<<<<<<<<<<<
 * 
 *             elif cb_type == 3:
 */
          __pyx_t_7 = __pyx_v_self->arg1;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_6 = ((__pyx_t_6uvloop_4loop_method_t)__pyx_v_self->callback)(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 66, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "uvloop/cbhandles.pyx":65
 *                     callback(*args)
 * 
 *             elif cb_type == 2:             # <<<<<<<<<<<<<<
 *                 (<method_t>self.callback)(self.arg1)
 * 
 */
          break;
          case 3:

          /* "uvloop/cbhandles.pyx":69
 * 
 *             elif cb_type == 3:
 *                 (<method1_t>self.callback)(self.arg1, self.arg2)             # <<<<<<<<<<<<<<
 * 
 *             elif cb_type == 4:
 */
          __pyx_t_6 = __pyx_v_self->arg1;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_7 = __pyx_v_self->arg2;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_8 = ((__pyx_t_6uvloop_4loop_method1_t)__pyx_v_self->callback)(__pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 69, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "uvloop/cbhandles.pyx":68
 *                 (<method_t>self.callback)(self.arg1)
 * 
 *             elif cb_type == 3:             # <<<<<<<<<<<<<<
 *                 (<method1_t>self.callback)(self.arg1, self.arg2)
 * 
 */
          break;
          case 4:

          /* "uvloop/cbhandles.pyx":72
 * 
 *             elif cb_type == 4:
 *                 (<method2_t>self.callback)(self.arg1, self.arg2, self.arg3)             # <<<<<<<<<<<<<<
 * 
 *             elif cb_type == 5:
 */
          __pyx_t_8 = __pyx_v_self->arg1;
          __Pyx_INCREF(__pyx_t_8);
          __pyx_t_7 = __pyx_v_self->arg2;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_6 = __pyx_v_self->arg3;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_10 = ((__pyx_t_6uvloop_4loop_method2_t)__pyx_v_self->callback)(__pyx_t_8, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 72, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "uvloop/cbhandles.pyx":71
 *                 (<method1_t>self.callback)(self.arg1, self.arg2)
 * 
 *             elif cb_type == 4:             # <<<<<<<<<<<<<<
 *                 (<method2_t>self.callback)(self.arg1, self.arg2, self.arg3)
 * 
 */
          break;
          case 5:

          /* "uvloop/cbhandles.pyx":76
 *             elif cb_type == 5:
 *                 (<method3_t>self.callback)(
 *                     self.arg1, self.arg2, self.arg3, self.arg4)             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
          __pyx_t_10 = __pyx_v_self->arg1;
          __Pyx_INCREF(__pyx_t_10);
          __pyx_t_6 = __pyx_v_self->arg2;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_7 = __pyx_v_self->arg3;
          __Pyx_INCREF(__pyx_t_7);
          __pyx_t_8 = __pyx_v_self->arg4;
          __Pyx_INCREF(__pyx_t_8);

          /* "uvloop/cbhandles.pyx":75
 * 
 *             elif cb_type == 5:
 *                 (<method3_t>self.callback)(             # <<<<<<<<<<<<<<
 *                     self.arg1, self.arg2, self.arg3, self.arg4)
 * 
 */
          __pyx_t_11 = ((__pyx_t_6uvloop_4loop_method3_t)__pyx_v_self->callback)(__pyx_t_10, __pyx_t_6, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(10, 75, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "uvloop/cbhandles.pyx":74
 *                 (<method2_t>self.callback)(self.arg1, self.arg2, self.arg3)
 * 
 *             elif cb_type == 5:             # <<<<<<<<<<<<<<
 *                 (<method3_t>self.callback)(
 *                     self.arg1, self.arg2, self.arg3, self.arg4)
 */
          break;
          default:

          /* "uvloop/cbhandles.pyx":79
 * 
 *             else:
 *                 raise RuntimeError('invalid Handle.cb_type: {}'.format(             # <<<<<<<<<<<<<<
 *                     cb_type))
 * 
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_Handle_cb_type, __pyx_n_s_format); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 79, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);

          /* "uvloop/cbhandles.pyx":80
 *             else:
 *                 raise RuntimeError('invalid Handle.cb_type: {}'.format(
 *                     cb_type))             # <<<<<<<<<<<<<<
 * 
 *         except (KeyboardInterrupt, SystemExit):
 */
          __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_cb_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 80, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_6 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(10, 79, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }

          /* "uvloop/cbhandles.pyx":79
 * 
 *             else:
 *                 raise RuntimeError('invalid Handle.cb_type: {}'.format(             # <<<<<<<<<<<<<<
 *                     cb_type))
 * 
 */
          __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 79, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_Raise(__pyx_t_8, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __PYX_ERR(10, 79, __pyx_L7_error)
          break;
        }

        /* "uvloop/cbhandles.pyx":48
 *         Py_INCREF(self)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.context is not None
 *             Context_Enter(self.context)
 */
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/cbhandles.pyx":82
 *                     cb_type))
 * 
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as ex:
 */
      __pyx_t_1 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_1) {
        __Pyx_AddTraceback("uvloop.loop.Handle._run", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_11, &__pyx_t_7) < 0) __PYX_ERR(10, 82, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "uvloop/cbhandles.pyx":83
 * 
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as ex:
 *             if cb_type == 1:
 */
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestoreWithState(__pyx_t_8, __pyx_t_11, __pyx_t_7);
        __pyx_t_8 = 0; __pyx_t_11 = 0; __pyx_t_7 = 0; 
        __PYX_ERR(10, 83, __pyx_L9_except_error)
      }

      /* "uvloop/cbhandles.pyx":84
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if cb_type == 1:
 *                 msg = 'Exception in callback {}'.format(callback)
 */
      __pyx_t_1 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_1) {
        __Pyx_AddTraceback("uvloop.loop.Handle._run", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_11, &__pyx_t_8) < 0) __PYX_ERR(10, 84, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_11);
        __pyx_v_ex = __pyx_t_11;
        /*try:*/ {

          /* "uvloop/cbhandles.pyx":85
 *             raise
 *         except BaseException as ex:
 *             if cb_type == 1:             # <<<<<<<<<<<<<<
 *                 msg = 'Exception in callback {}'.format(callback)
 *             else:
 */
          __pyx_t_5 = (__pyx_v_cb_type == 1);
          if (__pyx_t_5) {

            /* "uvloop/cbhandles.pyx":86
 *         except BaseException as ex:
 *             if cb_type == 1:
 *                 msg = 'Exception in callback {}'.format(callback)             # <<<<<<<<<<<<<<
 *             else:
 *                 msg = 'Exception in callback {}'.format(self.meth_name)
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Exception_in_callback, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 86, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (unlikely(!__pyx_v_callback)) { __Pyx_RaiseUnboundLocalError("callback"); __PYX_ERR(10, 86, __pyx_L22_error) }
            __pyx_t_12 = NULL;
            __pyx_t_9 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
                __pyx_t_9 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_callback};
              __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 86, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __pyx_v_msg = __pyx_t_6;
            __pyx_t_6 = 0;

            /* "uvloop/cbhandles.pyx":85
 *             raise
 *         except BaseException as ex:
 *             if cb_type == 1:             # <<<<<<<<<<<<<<
 *                 msg = 'Exception in callback {}'.format(callback)
 *             else:
 */
            goto __pyx_L24;
          }

          /* "uvloop/cbhandles.pyx":88
 *                 msg = 'Exception in callback {}'.format(callback)
 *             else:
 *                 msg = 'Exception in callback {}'.format(self.meth_name)             # <<<<<<<<<<<<<<
 * 
 *             context = {
 */
          /*else*/ {
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Exception_in_callback, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 88, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_12 = NULL;
            __pyx_t_9 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
                __pyx_t_9 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_self->meth_name};
              __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 88, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __pyx_v_msg = __pyx_t_6;
            __pyx_t_6 = 0;
          }
          __pyx_L24:;

          /* "uvloop/cbhandles.pyx":91
 * 
 *             context = {
 *                 'message': msg,             # <<<<<<<<<<<<<<
 *                 'exception': ex,
 *                 'handle': self,
 */
          __pyx_t_6 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 91, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_message, __pyx_v_msg) < 0) __PYX_ERR(10, 91, __pyx_L22_error)

          /* "uvloop/cbhandles.pyx":92
 *             context = {
 *                 'message': msg,
 *                 'exception': ex,             # <<<<<<<<<<<<<<
 *                 'handle': self,
 *             }
 */
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_exception, __pyx_v_ex) < 0) __PYX_ERR(10, 91, __pyx_L22_error)

          /* "uvloop/cbhandles.pyx":93
 *                 'message': msg,
 *                 'exception': ex,
 *                 'handle': self,             # <<<<<<<<<<<<<<
 *             }
 * 
 */
          if (PyDict_SetItem(__pyx_t_6, __pyx_n_u_handle, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(10, 91, __pyx_L22_error)
          __pyx_v_context = __pyx_t_6;
          __pyx_t_6 = 0;

          /* "uvloop/cbhandles.pyx":96
 *             }
 * 
 *             if self._source_traceback is not None:             # <<<<<<<<<<<<<<
 *                 context['source_traceback'] = self._source_traceback
 * 
 */
          __pyx_t_5 = (__pyx_v_self->_source_traceback != Py_None);
          if (__pyx_t_5) {

            /* "uvloop/cbhandles.pyx":97
 * 
 *             if self._source_traceback is not None:
 *                 context['source_traceback'] = self._source_traceback             # <<<<<<<<<<<<<<
 * 
 *             self.loop.call_exception_handler(context)
 */
            __pyx_t_6 = __pyx_v_self->_source_traceback;
            __Pyx_INCREF(__pyx_t_6);
            if (unlikely((PyObject_SetItem(__pyx_v_context, __pyx_n_u_source_traceback, __pyx_t_6) < 0))) __PYX_ERR(10, 97, __pyx_L22_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "uvloop/cbhandles.pyx":96
 *             }
 * 
 *             if self._source_traceback is not None:             # <<<<<<<<<<<<<<
 *                 context['source_traceback'] = self._source_traceback
 * 
 */
          }

          /* "uvloop/cbhandles.pyx":99
 *                 context['source_traceback'] = self._source_traceback
 * 
 *             self.loop.call_exception_handler(context)             # <<<<<<<<<<<<<<
 * 
 *         finally:
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 99, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_context};
            __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 99, __pyx_L22_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }

        /* "uvloop/cbhandles.pyx":84
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if cb_type == 1:
 *                 msg = 'Exception in callback {}'.format(callback)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L23;
          }
          __pyx_L22_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_1 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L9_except_error;
          }
          __pyx_L23:;
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "uvloop/cbhandles.pyx":48
 *         Py_INCREF(self)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.context is not None
 *             Context_Enter(self.context)
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L5_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L12_try_end:;
    }
  }

  /* "uvloop/cbhandles.pyx":102
 * 
 *         finally:
 *             context = self.context             # <<<<<<<<<<<<<<
 *             Py_DECREF(self)
 *             Context_Exit(context)
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_8 = __pyx_v_self->context;
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_context, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "uvloop/cbhandles.pyx":103
 *         finally:
 *             context = self.context
 *             Py_DECREF(self)             # <<<<<<<<<<<<<<
 *             Context_Exit(context)
 * 
 */
      Py_DECREF(((PyObject *)__pyx_v_self));

      /* "uvloop/cbhandles.pyx":104
 *             context = self.context
 *             Py_DECREF(self)
 *             Context_Exit(context)             # <<<<<<<<<<<<<<
 * 
 *     cdef _cancel(self):
 */
      __pyx_t_13 = Context_Exit(__pyx_v_context); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(10, 104, __pyx_L1_error)
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0)) __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_1 = __pyx_clineno; __pyx_t_21 = __pyx_filename;
      {

        /* "uvloop/cbhandles.pyx":102
 * 
 *         finally:
 *             context = self.context             # <<<<<<<<<<<<<<
 *             Py_DECREF(self)
 *             Context_Exit(context)
 */
        __pyx_t_8 = __pyx_v_self->context;
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_context, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "uvloop/cbhandles.pyx":103
 *         finally:
 *             context = self.context
 *             Py_DECREF(self)             # <<<<<<<<<<<<<<
 *             Context_Exit(context)
 * 
 */
        Py_DECREF(((PyObject *)__pyx_v_self));

        /* "uvloop/cbhandles.pyx":104
 *             context = self.context
 *             Py_DECREF(self)
 *             Context_Exit(context)             # <<<<<<<<<<<<<<
 * 
 *     cdef _cancel(self):
 */
        __pyx_t_22 = Context_Exit(__pyx_v_context); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(10, 104, __pyx_L31_error)
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestore(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_1; __pyx_filename = __pyx_t_21;
      goto __pyx_L1_error;
      __pyx_L31_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/cbhandles.pyx":33
 *                 self.__class__.__name__))
 * 
 *     cdef inline _run(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int cb_type
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uvloop.loop.Handle._run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_context);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":106
 *             Context_Exit(context)
 * 
 *     cdef _cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancelled = 1
 *         self.callback = NULL
 */

static PyObject *__pyx_f_6uvloop_4loop_6Handle__cancel(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_cancel", 1);

  /* "uvloop/cbhandles.pyx":107
 * 
 *     cdef _cancel(self):
 *         self._cancelled = 1             # <<<<<<<<<<<<<<
 *         self.callback = NULL
 *         self.arg1 = self.arg2 = self.arg3 = self.arg4 = None
 */
  __pyx_v_self->_cancelled = 1;

  /* "uvloop/cbhandles.pyx":108
 *     cdef _cancel(self):
 *         self._cancelled = 1
 *         self.callback = NULL             # <<<<<<<<<<<<<<
 *         self.arg1 = self.arg2 = self.arg3 = self.arg4 = None
 * 
 */
  __pyx_v_self->callback = NULL;

  /* "uvloop/cbhandles.pyx":109
 *         self._cancelled = 1
 *         self.callback = NULL
 *         self.arg1 = self.arg2 = self.arg3 = self.arg4 = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _format_handle(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->arg1);
  __Pyx_DECREF(__pyx_v_self->arg1);
  __pyx_v_self->arg1 = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->arg2);
  __Pyx_DECREF(__pyx_v_self->arg2);
  __pyx_v_self->arg2 = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->arg3);
  __Pyx_DECREF(__pyx_v_self->arg3);
  __pyx_v_self->arg3 = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->arg4);
  __Pyx_DECREF(__pyx_v_self->arg4);
  __pyx_v_self->arg4 = Py_None;

  /* "uvloop/cbhandles.pyx":106
 *             Context_Exit(context)
 * 
 *     cdef _cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancelled = 1
 *         self.callback = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":111
 *         self.arg1 = self.arg2 = self.arg3 = self.arg4 = None
 * 
 *     cdef _format_handle(self):             # <<<<<<<<<<<<<<
 *         # Mirrors `asyncio.base_events._format_handle`.
 *         if self.cb_type == 1 and self.arg1 is not None:
 */

static PyObject *__pyx_f_6uvloop_4loop_6Handle__format_handle(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  PyObject *__pyx_v_cb = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  unsigned int __pyx_t_14;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_format_handle", 1);

  /* "uvloop/cbhandles.pyx":113
 *     cdef _format_handle(self):
 *         # Mirrors `asyncio.base_events._format_handle`.
 *         if self.cb_type == 1 and self.arg1 is not None:             # <<<<<<<<<<<<<<
 *             cb = self.arg1
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):
 */
  __pyx_t_2 = (__pyx_v_self->cb_type == 1);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->arg1 != Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":114
 *         # Mirrors `asyncio.base_events._format_handle`.
 *         if self.cb_type == 1 and self.arg1 is not None:
 *             cb = self.arg1             # <<<<<<<<<<<<<<
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):
 *                 try:
 */
    __pyx_t_3 = __pyx_v_self->arg1;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_cb = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/cbhandles.pyx":115
 *         if self.cb_type == 1 and self.arg1 is not None:
 *             cb = self.arg1
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):             # <<<<<<<<<<<<<<
 *                 try:
 *                     return repr(cb.__self__)
 */
    __pyx_t_3 = __Pyx_GetAttr3(__pyx_v_cb, __pyx_n_u_self_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_Task;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_1 = PyObject_IsInstance(__pyx_t_3, __pyx_t_4); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(10, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {

      /* "uvloop/cbhandles.pyx":116
 *             cb = self.arg1
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):
 *                 try:             # <<<<<<<<<<<<<<
 *                     return repr(cb.__self__)
 *                 except (AttributeError, TypeError, ValueError) as ex:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        /*try:*/ {

          /* "uvloop/cbhandles.pyx":117
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):
 *                 try:
 *                     return repr(cb.__self__)             # <<<<<<<<<<<<<<
 *                 except (AttributeError, TypeError, ValueError) as ex:
 *                     # Cython generates empty __code__ objects for coroutines
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cb, __pyx_n_s_self_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 117, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 117, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_r = __pyx_t_3;
          __pyx_t_3 = 0;
          goto __pyx_L11_try_return;

          /* "uvloop/cbhandles.pyx":116
 *             cb = self.arg1
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):
 *                 try:             # <<<<<<<<<<<<<<
 *                     return repr(cb.__self__)
 *                 except (AttributeError, TypeError, ValueError) as ex:
 */
        }
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/cbhandles.pyx":118
 *                 try:
 *                     return repr(cb.__self__)
 *                 except (AttributeError, TypeError, ValueError) as ex:             # <<<<<<<<<<<<<<
 *                     # Cython generates empty __code__ objects for coroutines
 *                     # that can crash asyncio.Task.__repr__ with an
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
        if (__pyx_t_8) {
          __Pyx_AddTraceback("uvloop.loop.Handle._format_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(10, 118, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_3);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_4);
          __pyx_v_ex = __pyx_t_4;
          /*try:*/ {

            /* "uvloop/cbhandles.pyx":122
 *                     # that can crash asyncio.Task.__repr__ with an
 *                     # AttributeError etc.  Guard against that.
 *                     self.loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                         'message': 'exception in Task.__repr__',
 *                         'task': cb.__self__,
 */
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_11)) __PYX_ERR(10, 122, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "uvloop/cbhandles.pyx":123
 *                     # AttributeError etc.  Guard against that.
 *                     self.loop.call_exception_handler({
 *                         'message': 'exception in Task.__repr__',             # <<<<<<<<<<<<<<
 *                         'task': cb.__self__,
 *                         'exception': ex,
 */
            __pyx_t_12 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_12)) __PYX_ERR(10, 123, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_12);
            if (PyDict_SetItem(__pyx_t_12, __pyx_n_u_message, __pyx_kp_u_exception_in_Task___repr) < 0) __PYX_ERR(10, 123, __pyx_L18_error)

            /* "uvloop/cbhandles.pyx":124
 *                     self.loop.call_exception_handler({
 *                         'message': 'exception in Task.__repr__',
 *                         'task': cb.__self__,             # <<<<<<<<<<<<<<
 *                         'exception': ex,
 *                         'handle': self,
 */
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_cb, __pyx_n_s_self_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(10, 124, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (PyDict_SetItem(__pyx_t_12, __pyx_n_u_task, __pyx_t_13) < 0) __PYX_ERR(10, 123, __pyx_L18_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

            /* "uvloop/cbhandles.pyx":125
 *                         'message': 'exception in Task.__repr__',
 *                         'task': cb.__self__,
 *                         'exception': ex,             # <<<<<<<<<<<<<<
 *                         'handle': self,
 *                     })
 */
            if (PyDict_SetItem(__pyx_t_12, __pyx_n_u_exception, __pyx_v_ex) < 0) __PYX_ERR(10, 123, __pyx_L18_error)

            /* "uvloop/cbhandles.pyx":126
 *                         'task': cb.__self__,
 *                         'exception': ex,
 *                         'handle': self,             # <<<<<<<<<<<<<<
 *                     })
 *         return repr(self)
 */
            if (PyDict_SetItem(__pyx_t_12, __pyx_n_u_handle, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(10, 123, __pyx_L18_error)
            __pyx_t_13 = NULL;
            __pyx_t_14 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_11))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_11, function);
                __pyx_t_14 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_12};
              __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 122, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }

          /* "uvloop/cbhandles.pyx":118
 *                 try:
 *                     return repr(cb.__self__)
 *                 except (AttributeError, TypeError, ValueError) as ex:             # <<<<<<<<<<<<<<
 *                     # Cython generates empty __code__ objects for coroutines
 *                     # that can crash asyncio.Task.__repr__ with an
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              goto __pyx_L19;
            }
            __pyx_L18_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __pyx_t_8 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
              }
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
              goto __pyx_L9_except_error;
            }
            __pyx_L19:;
          }
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L8_exception_handled;
        }
        goto __pyx_L9_except_error;

        /* "uvloop/cbhandles.pyx":116
 *             cb = self.arg1
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):
 *                 try:             # <<<<<<<<<<<<<<
 *                     return repr(cb.__self__)
 *                 except (AttributeError, TypeError, ValueError) as ex:
 */
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        goto __pyx_L1_error;
        __pyx_L11_try_return:;
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        goto __pyx_L0;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      }

      /* "uvloop/cbhandles.pyx":115
 *         if self.cb_type == 1 and self.arg1 is not None:
 *             cb = self.arg1
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):             # <<<<<<<<<<<<<<
 *                 try:
 *                     return repr(cb.__self__)
 */
    }

    /* "uvloop/cbhandles.pyx":113
 *     cdef _format_handle(self):
 *         # Mirrors `asyncio.base_events._format_handle`.
 *         if self.cb_type == 1 and self.arg1 is not None:             # <<<<<<<<<<<<<<
 *             cb = self.arg1
 *             if isinstance(getattr(cb, '__self__', None), aio_Task):
 */
  }

  /* "uvloop/cbhandles.pyx":128
 *                         'handle': self,
 *                     })
 *         return repr(self)             # <<<<<<<<<<<<<<
 * 
 *     # Public API
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_9 = PyObject_Repr(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_9)) __PYX_ERR(10, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_r = __pyx_t_9;
  __pyx_t_9 = 0;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":111
 *         self.arg1 = self.arg2 = self.arg3 = self.arg4 = None
 * 
 *     cdef _format_handle(self):             # <<<<<<<<<<<<<<
 *         # Mirrors `asyncio.base_events._format_handle`.
 *         if self.cb_type == 1 and self.arg1 is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("uvloop.loop.Handle._format_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cb);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":132
 *     # Public API
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         info = [self.__class__.__name__]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_7__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_7__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle_6__repr__(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Handle_6__repr__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_func = NULL;
  PyObject *__pyx_v_cb_name = NULL;
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/cbhandles.pyx":133
 * 
 *     def __repr__(self):
 *         info = [self.__class__.__name__]             # <<<<<<<<<<<<<<
 * 
 *         if self._cancelled:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(10, 133, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_v_info = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":135
 *         info = [self.__class__.__name__]
 * 
 *         if self._cancelled:             # <<<<<<<<<<<<<<
 *             info.append('cancelled')
 * 
 */
  if (__pyx_v_self->_cancelled) {

    /* "uvloop/cbhandles.pyx":136
 * 
 *         if self._cancelled:
 *             info.append('cancelled')             # <<<<<<<<<<<<<<
 * 
 *         if self.cb_type == 1 and self.arg1 is not None:
 */
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_n_u_cancelled); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 136, __pyx_L1_error)

    /* "uvloop/cbhandles.pyx":135
 *         info = [self.__class__.__name__]
 * 
 *         if self._cancelled:             # <<<<<<<<<<<<<<
 *             info.append('cancelled')
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":138
 *             info.append('cancelled')
 * 
 *         if self.cb_type == 1 and self.arg1 is not None:             # <<<<<<<<<<<<<<
 *             func = self.arg1
 *             # Cython can unset func.__qualname__/__name__, hence the checks.
 */
  __pyx_t_5 = (__pyx_v_self->cb_type == 1);
  if (__pyx_t_5) {
  } else {
    __pyx_t_4 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__pyx_v_self->arg1 != Py_None);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_4) {

    /* "uvloop/cbhandles.pyx":139
 * 
 *         if self.cb_type == 1 and self.arg1 is not None:
 *             func = self.arg1             # <<<<<<<<<<<<<<
 *             # Cython can unset func.__qualname__/__name__, hence the checks.
 *             if hasattr(func, '__qualname__') and func.__qualname__:
 */
    __pyx_t_1 = __pyx_v_self->arg1;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_func = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/cbhandles.pyx":141
 *             func = self.arg1
 *             # Cython can unset func.__qualname__/__name__, hence the checks.
 *             if hasattr(func, '__qualname__') and func.__qualname__:             # <<<<<<<<<<<<<<
 *                 cb_name = func.__qualname__
 *             elif hasattr(func, '__name__') and func.__name__:
 */
    __pyx_t_5 = __Pyx_HasAttr(__pyx_v_func, __pyx_n_u_qualname); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(10, 141, __pyx_L1_error)
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_qualname); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(10, 141, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_4) {

      /* "uvloop/cbhandles.pyx":142
 *             # Cython can unset func.__qualname__/__name__, hence the checks.
 *             if hasattr(func, '__qualname__') and func.__qualname__:
 *                 cb_name = func.__qualname__             # <<<<<<<<<<<<<<
 *             elif hasattr(func, '__name__') and func.__name__:
 *                 cb_name = func.__name__
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_qualname); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_cb_name = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/cbhandles.pyx":141
 *             func = self.arg1
 *             # Cython can unset func.__qualname__/__name__, hence the checks.
 *             if hasattr(func, '__qualname__') and func.__qualname__:             # <<<<<<<<<<<<<<
 *                 cb_name = func.__qualname__
 *             elif hasattr(func, '__name__') and func.__name__:
 */
      goto __pyx_L7;
    }

    /* "uvloop/cbhandles.pyx":143
 *             if hasattr(func, '__qualname__') and func.__qualname__:
 *                 cb_name = func.__qualname__
 *             elif hasattr(func, '__name__') and func.__name__:             # <<<<<<<<<<<<<<
 *                 cb_name = func.__name__
 *             else:
 */
    __pyx_t_5 = __Pyx_HasAttr(__pyx_v_func, __pyx_n_u_name); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(10, 143, __pyx_L1_error)
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(10, 143, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __pyx_t_5;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_4) {

      /* "uvloop/cbhandles.pyx":144
 *                 cb_name = func.__qualname__
 *             elif hasattr(func, '__name__') and func.__name__:
 *                 cb_name = func.__name__             # <<<<<<<<<<<<<<
 *             else:
 *                 cb_name = repr(func)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_func, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_cb_name = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/cbhandles.pyx":143
 *             if hasattr(func, '__qualname__') and func.__qualname__:
 *                 cb_name = func.__qualname__
 *             elif hasattr(func, '__name__') and func.__name__:             # <<<<<<<<<<<<<<
 *                 cb_name = func.__name__
 *             else:
 */
      goto __pyx_L7;
    }

    /* "uvloop/cbhandles.pyx":146
 *                 cb_name = func.__name__
 *             else:
 *                 cb_name = repr(func)             # <<<<<<<<<<<<<<
 * 
 *             info.append(cb_name)
 */
    /*else*/ {
      __pyx_t_1 = PyObject_Repr(__pyx_v_func); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_cb_name = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_L7:;

    /* "uvloop/cbhandles.pyx":148
 *                 cb_name = repr(func)
 * 
 *             info.append(cb_name)             # <<<<<<<<<<<<<<
 *         elif self.meth_name is not None:
 *             info.append(self.meth_name)
 */
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_v_cb_name); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 148, __pyx_L1_error)

    /* "uvloop/cbhandles.pyx":138
 *             info.append('cancelled')
 * 
 *         if self.cb_type == 1 and self.arg1 is not None:             # <<<<<<<<<<<<<<
 *             func = self.arg1
 *             # Cython can unset func.__qualname__/__name__, hence the checks.
 */
    goto __pyx_L4;
  }

  /* "uvloop/cbhandles.pyx":149
 * 
 *             info.append(cb_name)
 *         elif self.meth_name is not None:             # <<<<<<<<<<<<<<
 *             info.append(self.meth_name)
 * 
 */
  __pyx_t_4 = (__pyx_v_self->meth_name != ((PyObject*)Py_None));
  if (__pyx_t_4) {

    /* "uvloop/cbhandles.pyx":150
 *             info.append(cb_name)
 *         elif self.meth_name is not None:
 *             info.append(self.meth_name)             # <<<<<<<<<<<<<<
 * 
 *         if self._source_traceback is not None:
 */
    __pyx_t_1 = __pyx_v_self->meth_name;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 150, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/cbhandles.pyx":149
 * 
 *             info.append(cb_name)
 *         elif self.meth_name is not None:             # <<<<<<<<<<<<<<
 *             info.append(self.meth_name)
 * 
 */
  }
  __pyx_L4:;

  /* "uvloop/cbhandles.pyx":152
 *             info.append(self.meth_name)
 * 
 *         if self._source_traceback is not None:             # <<<<<<<<<<<<<<
 *             frame = self._source_traceback[-1]
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 */
  __pyx_t_4 = (__pyx_v_self->_source_traceback != Py_None);
  if (__pyx_t_4) {

    /* "uvloop/cbhandles.pyx":153
 * 
 *         if self._source_traceback is not None:
 *             frame = self._source_traceback[-1]             # <<<<<<<<<<<<<<
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 * 
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->_source_traceback, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_frame = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/cbhandles.pyx":154
 *         if self._source_traceback is not None:
 *             frame = self._source_traceback[-1]
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))             # <<<<<<<<<<<<<<
 * 
 *         return '<' + ' '.join(info) + '>'
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_created_at, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_frame, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_frame, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(10, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 154, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/cbhandles.pyx":152
 *             info.append(self.meth_name)
 * 
 *         if self._source_traceback is not None:             # <<<<<<<<<<<<<<
 *             frame = self._source_traceback[-1]
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 */
  }

  /* "uvloop/cbhandles.pyx":156
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 * 
 *         return '<' + ' '.join(info) + '>'             # <<<<<<<<<<<<<<
 * 
 *     def cancel(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__129, __pyx_v_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u__128, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_2, __pyx_kp_u__130); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":132
 *     # Public API
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         info = [self.__class__.__name__]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.Handle.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_func);
  __Pyx_XDECREF(__pyx_v_cb_name);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":158
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_9cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Handle_8cancel, "Handle.cancel(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Handle_9cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_9cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_8cancel};
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_9cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("cancel", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "cancel", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle_8cancel(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Handle_8cancel(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/cbhandles.pyx":159
 * 
 *     def cancel(self):
 *         self._cancel()             # <<<<<<<<<<<<<<
 * 
 *     def cancelled(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_self->__pyx_vtab)->_cancel(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":158
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Handle.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":161
 *         self._cancel()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_11cancelled(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Handle_10cancelled, "Handle.cancelled(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Handle_11cancelled = {"cancelled", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_11cancelled, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_10cancelled};
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_11cancelled(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancelled (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("cancelled", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "cancelled", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle_10cancelled(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Handle_10cancelled(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancelled", 1);

  /* "uvloop/cbhandles.pyx":162
 * 
 *     def cancelled(self):
 *         return self._cancelled             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_cancelled); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":161
 *         self._cancel()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Handle.cancelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pxd":14
 *         object __weakref__
 * 
 *         readonly _source_traceback             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _set_loop(self, Loop loop)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_17_source_traceback_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_17_source_traceback_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle_17_source_traceback___get__(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Handle_17_source_traceback___get__(struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_source_traceback);
  __pyx_r = __pyx_v_self->_source_traceback;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Handle_12__reduce_cython__, "Handle.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Handle_13__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_12__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle_12__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Handle_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Handle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Handle_14__setstate_cython__, "Handle.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Handle_15__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_14__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6Handle_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Handle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Handle_14__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Handle_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Handle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":168
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class TimerHandle:
 *     def __cinit__(self, Loop loop, object callback, object args,             # <<<<<<<<<<<<<<
 *                   uint64_t delay, object context):
 * 
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_11TimerHandle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_11TimerHandle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_args = 0;
  uint64_t __pyx_v_delay;
  PyObject *__pyx_v_context = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,&__pyx_n_s_callback,&__pyx_n_s_args,&__pyx_n_s_delay,&__pyx_n_s_context,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 168, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 168, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 5, 5, 1); __PYX_ERR(10, 168, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_args)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 168, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 5, 5, 2); __PYX_ERR(10, 168, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_delay)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 168, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 5, 5, 3); __PYX_ERR(10, 168, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(10, 168, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 5, 5, 4); __PYX_ERR(10, 168, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(10, 168, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
    }
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)values[0]);
    __pyx_v_callback = values[1];
    __pyx_v_args = values[2];
    __pyx_v_delay = __Pyx_PyInt_As_uint64_t(values[3]); if (unlikely((__pyx_v_delay == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(10, 169, __pyx_L3_error)
    __pyx_v_context = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 5, 5, __pyx_nargs); __PYX_ERR(10, 168, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_loop), __pyx_ptype_6uvloop_4loop_Loop, 1, "loop", 0))) __PYX_ERR(10, 168, __pyx_L1_error)
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle___cinit__(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self), __pyx_v_loop, __pyx_v_callback, __pyx_v_args, __pyx_v_delay, __pyx_v_context);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_11TimerHandle___cinit__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, uint64_t __pyx_v_delay, PyObject *__pyx_v_context) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_context);

  /* "uvloop/cbhandles.pyx":171
 *                   uint64_t delay, object context):
 * 
 *         self.loop = loop             # <<<<<<<<<<<<<<
 *         self.callback = callback
 *         self.args = args
 */
  __Pyx_INCREF((PyObject *)__pyx_v_loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_loop);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->loop);
  __pyx_v_self->loop = __pyx_v_loop;

  /* "uvloop/cbhandles.pyx":172
 * 
 *         self.loop = loop
 *         self.callback = callback             # <<<<<<<<<<<<<<
 *         self.args = args
 *         self._cancelled = 0
 */
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  __Pyx_GOTREF(__pyx_v_self->callback);
  __Pyx_DECREF(__pyx_v_self->callback);
  __pyx_v_self->callback = __pyx_v_callback;

  /* "uvloop/cbhandles.pyx":173
 *         self.loop = loop
 *         self.callback = callback
 *         self.args = args             # <<<<<<<<<<<<<<
 *         self._cancelled = 0
 * 
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v_args))||((__pyx_v_args) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v_args))) __PYX_ERR(10, 173, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_args;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":174
 *         self.callback = callback
 *         self.args = args
 *         self._cancelled = 0             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG:
 */
  __pyx_v_self->_cancelled = 0;

  /* "uvloop/cbhandles.pyx":176
 *         self._cancelled = 0
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self.loop._debug_cb_timer_handles_total += 1
 *             self.loop._debug_cb_timer_handles_count += 1
 */
  __pyx_t_2 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_2) {

    /* "uvloop/cbhandles.pyx":177
 * 
 *         if UVLOOP_DEBUG:
 *             self.loop._debug_cb_timer_handles_total += 1             # <<<<<<<<<<<<<<
 *             self.loop._debug_cb_timer_handles_count += 1
 * 
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self->loop);
    __pyx_t_3 = __pyx_v_self->loop;
    __pyx_t_3->_debug_cb_timer_handles_total = (__pyx_t_3->_debug_cb_timer_handles_total + 1);
    __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/cbhandles.pyx":178
 *         if UVLOOP_DEBUG:
 *             self.loop._debug_cb_timer_handles_total += 1
 *             self.loop._debug_cb_timer_handles_count += 1             # <<<<<<<<<<<<<<
 * 
 *         if context is None:
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self->loop);
    __pyx_t_3 = __pyx_v_self->loop;
    __pyx_t_3->_debug_cb_timer_handles_count = (__pyx_t_3->_debug_cb_timer_handles_count + 1);
    __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/cbhandles.pyx":176
 *         self._cancelled = 0
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self.loop._debug_cb_timer_handles_total += 1
 *             self.loop._debug_cb_timer_handles_count += 1
 */
  }

  /* "uvloop/cbhandles.pyx":180
 *             self.loop._debug_cb_timer_handles_count += 1
 * 
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  __pyx_t_2 = (__pyx_v_context == Py_None);
  if (__pyx_t_2) {

    /* "uvloop/cbhandles.pyx":181
 * 
 *         if context is None:
 *             context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 *         self.context = context
 * 
 */
    __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_context, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/cbhandles.pyx":180
 *             self.loop._debug_cb_timer_handles_count += 1
 * 
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  }

  /* "uvloop/cbhandles.pyx":182
 *         if context is None:
 *             context = Context_CopyCurrent()
 *         self.context = context             # <<<<<<<<<<<<<<
 * 
 *         if loop._debug:
 */
  __Pyx_INCREF(__pyx_v_context);
  __Pyx_GIVEREF(__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->context);
  __Pyx_DECREF(__pyx_v_self->context);
  __pyx_v_self->context = __pyx_v_context;

  /* "uvloop/cbhandles.pyx":184
 *         self.context = context
 * 
 *         if loop._debug:             # <<<<<<<<<<<<<<
 *             self._debug_info = (
 *                 format_callback_name(callback),
 */
  if (__pyx_v_loop->_debug) {

    /* "uvloop/cbhandles.pyx":186
 *         if loop._debug:
 *             self._debug_info = (
 *                 format_callback_name(callback),             # <<<<<<<<<<<<<<
 *                 extract_stack()
 *             )
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_format_callback_name(__pyx_v_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "uvloop/cbhandles.pyx":187
 *             self._debug_info = (
 *                 format_callback_name(callback),
 *                 extract_stack()             # <<<<<<<<<<<<<<
 *             )
 *         else:
 */
    __pyx_t_4 = __pyx_f_6uvloop_4loop_extract_stack(); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uvloop/cbhandles.pyx":186
 *         if loop._debug:
 *             self._debug_info = (
 *                 format_callback_name(callback),             # <<<<<<<<<<<<<<
 *                 extract_stack()
 *             )
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(10, 186, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4)) __PYX_ERR(10, 186, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;

    /* "uvloop/cbhandles.pyx":185
 * 
 *         if loop._debug:
 *             self._debug_info = (             # <<<<<<<<<<<<<<
 *                 format_callback_name(callback),
 *                 extract_stack()
 */
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_self->_debug_info);
    __Pyx_DECREF(__pyx_v_self->_debug_info);
    __pyx_v_self->_debug_info = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "uvloop/cbhandles.pyx":184
 *         self.context = context
 * 
 *         if loop._debug:             # <<<<<<<<<<<<<<
 *             self._debug_info = (
 *                 format_callback_name(callback),
 */
    goto __pyx_L5;
  }

  /* "uvloop/cbhandles.pyx":190
 *             )
 *         else:
 *             self._debug_info = None             # <<<<<<<<<<<<<<
 * 
 *         self.timer = UVTimer.new(
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_debug_info);
    __Pyx_DECREF(__pyx_v_self->_debug_info);
    __pyx_v_self->_debug_info = ((PyObject*)Py_None);
  }
  __pyx_L5:;

  /* "uvloop/cbhandles.pyx":192
 *             self._debug_info = None
 * 
 *         self.timer = UVTimer.new(             # <<<<<<<<<<<<<<
 *             loop, <method_t>self._run, self, delay)
 * 
 */
  __pyx_t_5 = ((PyObject *)__pyx_f_6uvloop_4loop_7UVTimer_new(__pyx_v_loop, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_TimerHandle *)__pyx_v_self->__pyx_vtab)->_run), ((PyObject *)__pyx_v_self), __pyx_v_delay)); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->timer);
  __Pyx_DECREF((PyObject *)__pyx_v_self->timer);
  __pyx_v_self->timer = ((struct __pyx_obj_6uvloop_4loop_UVTimer *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "uvloop/cbhandles.pyx":195
 *             loop, <method_t>self._run, self, delay)
 * 
 *         self.timer.start()             # <<<<<<<<<<<<<<
 *         self._when = self.timer.get_when() * 1e-3
 * 
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->timer->__pyx_base.__pyx_vtab)->start(__pyx_v_self->timer); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/cbhandles.pyx":196
 * 
 *         self.timer.start()
 *         self._when = self.timer.get_when() * 1e-3             # <<<<<<<<<<<<<<
 * 
 *         # Only add to loop._timers when `self.timer` is successfully created
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->timer->__pyx_base.__pyx_vtab)->get_when(__pyx_v_self->timer); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_5, __pyx_float_1eneg_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_when);
  __Pyx_DECREF(__pyx_v_self->_when);
  __pyx_v_self->_when = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "uvloop/cbhandles.pyx":199
 * 
 *         # Only add to loop._timers when `self.timer` is successfully created
 *         loop._timers.add(self)             # <<<<<<<<<<<<<<
 * 
 *     property _source_traceback:
 */
  if (unlikely(__pyx_v_loop->_timers == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(10, 199, __pyx_L1_error)
  }
  __pyx_t_6 = PySet_Add(__pyx_v_loop->_timers, ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(10, 199, __pyx_L1_error)

  /* "uvloop/cbhandles.pyx":168
 * @cython.freelist(DEFAULT_FREELIST_SIZE)
 * cdef class TimerHandle:
 *     def __cinit__(self, Loop loop, object callback, object args,             # <<<<<<<<<<<<<<
 *                   uint64_t delay, object context):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF((PyObject *)__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_context);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":202
 * 
 *     property _source_traceback:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self._debug_info is not None:
 *                 return self._debug_info[1]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_17_source_traceback_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_17_source_traceback_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle_17_source_traceback___get__(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_17_source_traceback___get__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uvloop/cbhandles.pyx":203
 *     property _source_traceback:
 *         def __get__(self):
 *             if self._debug_info is not None:             # <<<<<<<<<<<<<<
 *                 return self._debug_info[1]
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_debug_info != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":204
 *         def __get__(self):
 *             if self._debug_info is not None:
 *                 return self._debug_info[1]             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_self->_debug_info == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(10, 204, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_GetItemInt_Tuple(__pyx_v_self->_debug_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/cbhandles.pyx":203
 *     property _source_traceback:
 *         def __get__(self):
 *             if self._debug_info is not None:             # <<<<<<<<<<<<<<
 *                 return self._debug_info[1]
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":202
 * 
 *     property _source_traceback:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self._debug_info is not None:
 *                 return self._debug_info[1]
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle._source_traceback.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":206
 *                 return self._debug_info[1]
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             self.loop._debug_cb_timer_handles_count -= 1
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_11TimerHandle_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_11TimerHandle_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_11TimerHandle_2__dealloc__(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_11TimerHandle_2__dealloc__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/cbhandles.pyx":207
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self.loop._debug_cb_timer_handles_count -= 1
 *         if self.timer is not None:
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":208
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:
 *             self.loop._debug_cb_timer_handles_count -= 1             # <<<<<<<<<<<<<<
 *         if self.timer is not None:
 *             raise RuntimeError('active TimerHandle is deallacating')
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self->loop);
    __pyx_t_2 = __pyx_v_self->loop;
    __pyx_t_2->_debug_cb_timer_handles_count = (__pyx_t_2->_debug_cb_timer_handles_count - 1);
    __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":207
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self.loop._debug_cb_timer_handles_count -= 1
 *         if self.timer is not None:
 */
  }

  /* "uvloop/cbhandles.pyx":209
 *         if UVLOOP_DEBUG:
 *             self.loop._debug_cb_timer_handles_count -= 1
 *         if self.timer is not None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('active TimerHandle is deallacating')
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->timer) != Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/cbhandles.pyx":210
 *             self.loop._debug_cb_timer_handles_count -= 1
 *         if self.timer is not None:
 *             raise RuntimeError('active TimerHandle is deallacating')             # <<<<<<<<<<<<<<
 * 
 *     cdef _cancel(self):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__131, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(10, 210, __pyx_L1_error)

    /* "uvloop/cbhandles.pyx":209
 *         if UVLOOP_DEBUG:
 *             self.loop._debug_cb_timer_handles_count -= 1
 *         if self.timer is not None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('active TimerHandle is deallacating')
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":206
 *                 return self._debug_info[1]
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             self.loop._debug_cb_timer_handles_count -= 1
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF((PyObject *)__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uvloop.loop.TimerHandle.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/cbhandles.pyx":212
 *             raise RuntimeError('active TimerHandle is deallacating')
 * 
 *     cdef _cancel(self):             # <<<<<<<<<<<<<<
 *         if self._cancelled == 1:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_11TimerHandle__cancel(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_cancel", 1);

  /* "uvloop/cbhandles.pyx":213
 * 
 *     cdef _cancel(self):
 *         if self._cancelled == 1:             # <<<<<<<<<<<<<<
 *             return
 *         self._cancelled = 1
 */
  __pyx_t_1 = (__pyx_v_self->_cancelled == 1);
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":214
 *     cdef _cancel(self):
 *         if self._cancelled == 1:
 *             return             # <<<<<<<<<<<<<<
 *         self._cancelled = 1
 *         self._clear()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/cbhandles.pyx":213
 * 
 *     cdef _cancel(self):
 *         if self._cancelled == 1:             # <<<<<<<<<<<<<<
 *             return
 *         self._cancelled = 1
 */
  }

  /* "uvloop/cbhandles.pyx":215
 *         if self._cancelled == 1:
 *             return
 *         self._cancelled = 1             # <<<<<<<<<<<<<<
 *         self._clear()
 * 
 */
  __pyx_v_self->_cancelled = 1;

  /* "uvloop/cbhandles.pyx":216
 *             return
 *         self._cancelled = 1
 *         self._clear()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _clear(self):
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_11TimerHandle__clear(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/cbhandles.pyx":212
 *             raise RuntimeError('active TimerHandle is deallacating')
 * 
 *     cdef _cancel(self):             # <<<<<<<<<<<<<<
 *         if self._cancelled == 1:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle._cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":218
 *         self._clear()
 * 
 *     cdef inline _clear(self):             # <<<<<<<<<<<<<<
 *         if self.timer is None:
 *             return
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_11TimerHandle__clear(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_clear", 1);

  /* "uvloop/cbhandles.pyx":219
 * 
 *     cdef inline _clear(self):
 *         if self.timer is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->timer) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":220
 *     cdef inline _clear(self):
 *         if self.timer is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self.callback = None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/cbhandles.pyx":219
 * 
 *     cdef inline _clear(self):
 *         if self.timer is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":222
 *             return
 * 
 *         self.callback = None             # <<<<<<<<<<<<<<
 *         self.args = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->callback);
  __Pyx_DECREF(__pyx_v_self->callback);
  __pyx_v_self->callback = Py_None;

  /* "uvloop/cbhandles.pyx":223
 * 
 *         self.callback = None
 *         self.args = None             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = ((PyObject*)Py_None);

  /* "uvloop/cbhandles.pyx":225
 *         self.args = None
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self.loop._timers.remove(self)
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/cbhandles.pyx":226
 * 
 *         try:
 *             self.loop._timers.remove(self)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.timer._close()
 */
    if (unlikely(__pyx_v_self->loop->_timers == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "remove");
      __PYX_ERR(10, 226, __pyx_L5_error)
    }
    __pyx_t_2 = __Pyx_PySet_Remove(__pyx_v_self->loop->_timers, ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(10, 226, __pyx_L5_error)
  }

  /* "uvloop/cbhandles.pyx":228
 *             self.loop._timers.remove(self)
 *         finally:
 *             self.timer._close()             # <<<<<<<<<<<<<<
 *             self.timer = None  # let the UVTimer handle GC
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->timer->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self->timer)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/cbhandles.pyx":229
 *         finally:
 *             self.timer._close()
 *             self.timer = None  # let the UVTimer handle GC             # <<<<<<<<<<<<<<
 * 
 *     cdef _run(self):
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->timer);
      __Pyx_DECREF((PyObject *)__pyx_v_self->timer);
      __pyx_v_self->timer = ((struct __pyx_obj_6uvloop_4loop_UVTimer *)Py_None);
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {

        /* "uvloop/cbhandles.pyx":228
 *             self.loop._timers.remove(self)
 *         finally:
 *             self.timer._close()             # <<<<<<<<<<<<<<
 *             self.timer = None  # let the UVTimer handle GC
 * 
 */
        __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->timer->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self->timer)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 228, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "uvloop/cbhandles.pyx":229
 *         finally:
 *             self.timer._close()
 *             self.timer = None  # let the UVTimer handle GC             # <<<<<<<<<<<<<<
 * 
 *     cdef _run(self):
 */
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF((PyObject *)__pyx_v_self->timer);
        __Pyx_DECREF((PyObject *)__pyx_v_self->timer);
        __pyx_v_self->timer = ((struct __pyx_obj_6uvloop_4loop_UVTimer *)Py_None);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
      __pyx_L8_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/cbhandles.pyx":218
 *         self._clear()
 * 
 *     cdef inline _clear(self):             # <<<<<<<<<<<<<<
 *         if self.timer is None:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle._clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":231
 *             self.timer = None  # let the UVTimer handle GC
 * 
 *     cdef _run(self):             # <<<<<<<<<<<<<<
 *         if self._cancelled == 1:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_11TimerHandle__run(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_v_callback = NULL;
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_v_started = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_context = NULL;
  PyObject *__pyx_v_delta = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  char const *__pyx_t_22;
  int __pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_run", 1);

  /* "uvloop/cbhandles.pyx":232
 * 
 *     cdef _run(self):
 *         if self._cancelled == 1:             # <<<<<<<<<<<<<<
 *             return
 *         if self.callback is None:
 */
  __pyx_t_1 = (__pyx_v_self->_cancelled == 1);
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":233
 *     cdef _run(self):
 *         if self._cancelled == 1:
 *             return             # <<<<<<<<<<<<<<
 *         if self.callback is None:
 *             raise RuntimeError('cannot run TimerHandle; callback is not set')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/cbhandles.pyx":232
 * 
 *     cdef _run(self):
 *         if self._cancelled == 1:             # <<<<<<<<<<<<<<
 *             return
 *         if self.callback is None:
 */
  }

  /* "uvloop/cbhandles.pyx":234
 *         if self._cancelled == 1:
 *             return
 *         if self.callback is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('cannot run TimerHandle; callback is not set')
 * 
 */
  __pyx_t_1 = (__pyx_v_self->callback == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/cbhandles.pyx":235
 *             return
 *         if self.callback is None:
 *             raise RuntimeError('cannot run TimerHandle; callback is not set')             # <<<<<<<<<<<<<<
 * 
 *         callback = self.callback
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__132, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(10, 235, __pyx_L1_error)

    /* "uvloop/cbhandles.pyx":234
 *         if self._cancelled == 1:
 *             return
 *         if self.callback is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('cannot run TimerHandle; callback is not set')
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":237
 *             raise RuntimeError('cannot run TimerHandle; callback is not set')
 * 
 *         callback = self.callback             # <<<<<<<<<<<<<<
 *         args = self.args
 * 
 */
  __pyx_t_2 = __pyx_v_self->callback;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_callback = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/cbhandles.pyx":238
 * 
 *         callback = self.callback
 *         args = self.args             # <<<<<<<<<<<<<<
 * 
 *         # Since _run is a cdef and there's no BoundMethod,
 */
  __pyx_t_2 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_args = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/cbhandles.pyx":242
 *         # Since _run is a cdef and there's no BoundMethod,
 *         # we guard 'self' manually.
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 * 
 *         if self.loop._debug:
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "uvloop/cbhandles.pyx":244
 *         Py_INCREF(self)
 * 
 *         if self.loop._debug:             # <<<<<<<<<<<<<<
 *             started = time_monotonic()
 *         try:
 */
  if (__pyx_v_self->loop->_debug) {

    /* "uvloop/cbhandles.pyx":245
 * 
 *         if self.loop._debug:
 *             started = time_monotonic()             # <<<<<<<<<<<<<<
 *         try:
 *             assert self.context is not None
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_time_monotonic);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_time_monotonic; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_started = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":244
 *         Py_INCREF(self)
 * 
 *         if self.loop._debug:             # <<<<<<<<<<<<<<
 *             started = time_monotonic()
 *         try:
 */
  }

  /* "uvloop/cbhandles.pyx":246
 *         if self.loop._debug:
 *             started = time_monotonic()
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.context is not None
 *             Context_Enter(self.context)
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "uvloop/cbhandles.pyx":247
 *             started = time_monotonic()
 *         try:
 *             assert self.context is not None             # <<<<<<<<<<<<<<
 *             Context_Enter(self.context)
 * 
 */
        #ifndef CYTHON_WITHOUT_ASSERTIONS
        if (unlikely(__pyx_assertions_enabled())) {
          __pyx_t_1 = (__pyx_v_self->context != Py_None);
          if (unlikely(!__pyx_t_1)) {
            __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
            __PYX_ERR(10, 247, __pyx_L9_error)
          }
        }
        #else
        if ((1)); else __PYX_ERR(10, 247, __pyx_L9_error)
        #endif

        /* "uvloop/cbhandles.pyx":248
 *         try:
 *             assert self.context is not None
 *             Context_Enter(self.context)             # <<<<<<<<<<<<<<
 * 
 *             if args is not None:
 */
        __pyx_t_2 = __pyx_v_self->context;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_9 = Context_Enter(__pyx_t_2); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(10, 248, __pyx_L9_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/cbhandles.pyx":250
 *             Context_Enter(self.context)
 * 
 *             if args is not None:             # <<<<<<<<<<<<<<
 *                 callback(*args)
 *             else:
 */
        __pyx_t_1 = (__pyx_v_args != ((PyObject*)Py_None));
        if (__pyx_t_1) {

          /* "uvloop/cbhandles.pyx":251
 * 
 *             if args is not None:
 *                 callback(*args)             # <<<<<<<<<<<<<<
 *             else:
 *                 callback()
 */
          if (unlikely(__pyx_v_args == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
            __PYX_ERR(10, 251, __pyx_L9_error)
          }
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_v_callback, __pyx_v_args, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 251, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "uvloop/cbhandles.pyx":250
 *             Context_Enter(self.context)
 * 
 *             if args is not None:             # <<<<<<<<<<<<<<
 *                 callback(*args)
 *             else:
 */
          goto __pyx_L15;
        }

        /* "uvloop/cbhandles.pyx":253
 *                 callback(*args)
 *             else:
 *                 callback()             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_v_callback);
          __pyx_t_3 = __pyx_v_callback; __pyx_t_4 = NULL;
          __pyx_t_5 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_5 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 253, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_L15:;

        /* "uvloop/cbhandles.pyx":246
 *         if self.loop._debug:
 *             started = time_monotonic()
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.context is not None
 *             Context_Enter(self.context)
 */
      }

      /* "uvloop/cbhandles.pyx":268
 *             self.loop.call_exception_handler(context)
 *         else:
 *             if self.loop._debug:             # <<<<<<<<<<<<<<
 *                 delta = time_monotonic() - started
 *                 if delta > self.loop.slow_callback_duration:
 */
      /*else:*/ {
        if (__pyx_v_self->loop->_debug) {

          /* "uvloop/cbhandles.pyx":269
 *         else:
 *             if self.loop._debug:
 *                 delta = time_monotonic() - started             # <<<<<<<<<<<<<<
 *                 if delta > self.loop.slow_callback_duration:
 *                     aio_logger.warning(
 */
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_time_monotonic);
          __pyx_t_3 = __pyx_v_6uvloop_4loop_time_monotonic; __pyx_t_4 = NULL;
          __pyx_t_5 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_5 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 269, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          if (unlikely(!__pyx_v_started)) { __Pyx_RaiseUnboundLocalError("started"); __PYX_ERR(10, 269, __pyx_L11_except_error) }
          __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_started); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 269, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v_delta = __pyx_t_3;
          __pyx_t_3 = 0;

          /* "uvloop/cbhandles.pyx":270
 *             if self.loop._debug:
 *                 delta = time_monotonic() - started
 *                 if delta > self.loop.slow_callback_duration:             # <<<<<<<<<<<<<<
 *                     aio_logger.warning(
 *                         'Executing %r took %.3f seconds',
 */
          __pyx_t_3 = PyObject_RichCompare(__pyx_v_delta, __pyx_v_self->loop->slow_callback_duration, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 270, __pyx_L11_except_error)
          __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(10, 270, __pyx_L11_except_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (__pyx_t_1) {

            /* "uvloop/cbhandles.pyx":271
 *                 delta = time_monotonic() - started
 *                 if delta > self.loop.slow_callback_duration:
 *                     aio_logger.warning(             # <<<<<<<<<<<<<<
 *                         'Executing %r took %.3f seconds',
 *                         self, delta)
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_warning); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 271, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_2);

            /* "uvloop/cbhandles.pyx":273
 *                     aio_logger.warning(
 *                         'Executing %r took %.3f seconds',
 *                         self, delta)             # <<<<<<<<<<<<<<
 *         finally:
 *             context = self.context
 */
            __pyx_t_4 = NULL;
            __pyx_t_5 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
                __pyx_t_5 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_kp_u_Executing_r_took_3f_seconds, ((PyObject *)__pyx_v_self), __pyx_v_delta};
              __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 271, __pyx_L11_except_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "uvloop/cbhandles.pyx":270
 *             if self.loop._debug:
 *                 delta = time_monotonic() - started
 *                 if delta > self.loop.slow_callback_duration:             # <<<<<<<<<<<<<<
 *                     aio_logger.warning(
 *                         'Executing %r took %.3f seconds',
 */
          }

          /* "uvloop/cbhandles.pyx":268
 *             self.loop.call_exception_handler(context)
 *         else:
 *             if self.loop._debug:             # <<<<<<<<<<<<<<
 *                 delta = time_monotonic() - started
 *                 if delta > self.loop.slow_callback_duration:
 */
        }
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L14_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/cbhandles.pyx":254
 *             else:
 *                 callback()
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as ex:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.TimerHandle._run", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_4) < 0) __PYX_ERR(10, 254, __pyx_L11_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_4);

        /* "uvloop/cbhandles.pyx":255
 *                 callback()
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as ex:
 *             context = {
 */
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_2, __pyx_t_4);
        __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_4 = 0; 
        __PYX_ERR(10, 255, __pyx_L11_except_error)
      }

      /* "uvloop/cbhandles.pyx":256
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             context = {
 *                 'message': 'Exception in callback {}'.format(callback),
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.TimerHandle._run", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(10, 256, __pyx_L11_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_ex = __pyx_t_2;
        /*try:*/ {

          /* "uvloop/cbhandles.pyx":258
 *         except BaseException as ex:
 *             context = {
 *                 'message': 'Exception in callback {}'.format(callback),             # <<<<<<<<<<<<<<
 *                 'exception': ex,
 *                 'handle': self,
 */
          __pyx_t_10 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 258, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Exception_in_callback, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(10, 258, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = NULL;
          __pyx_t_5 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_5 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_v_callback};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(10, 258, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_message, __pyx_t_11) < 0) __PYX_ERR(10, 258, __pyx_L25_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "uvloop/cbhandles.pyx":259
 *             context = {
 *                 'message': 'Exception in callback {}'.format(callback),
 *                 'exception': ex,             # <<<<<<<<<<<<<<
 *                 'handle': self,
 *             }
 */
          if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_exception, __pyx_v_ex) < 0) __PYX_ERR(10, 258, __pyx_L25_error)

          /* "uvloop/cbhandles.pyx":260
 *                 'message': 'Exception in callback {}'.format(callback),
 *                 'exception': ex,
 *                 'handle': self,             # <<<<<<<<<<<<<<
 *             }
 * 
 */
          if (PyDict_SetItem(__pyx_t_10, __pyx_n_u_handle, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(10, 258, __pyx_L25_error)
          __pyx_v_context = __pyx_t_10;
          __pyx_t_10 = 0;

          /* "uvloop/cbhandles.pyx":263
 *             }
 * 
 *             if self._debug_info is not None:             # <<<<<<<<<<<<<<
 *                 context['source_traceback'] = self._debug_info[1]
 * 
 */
          __pyx_t_1 = (__pyx_v_self->_debug_info != ((PyObject*)Py_None));
          if (__pyx_t_1) {

            /* "uvloop/cbhandles.pyx":264
 * 
 *             if self._debug_info is not None:
 *                 context['source_traceback'] = self._debug_info[1]             # <<<<<<<<<<<<<<
 * 
 *             self.loop.call_exception_handler(context)
 */
            if (unlikely(__pyx_v_self->_debug_info == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
              __PYX_ERR(10, 264, __pyx_L25_error)
            }
            __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v_self->_debug_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 264, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_10);
            if (unlikely((PyObject_SetItem(__pyx_v_context, __pyx_n_u_source_traceback, __pyx_t_10) < 0))) __PYX_ERR(10, 264, __pyx_L25_error)
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "uvloop/cbhandles.pyx":263
 *             }
 * 
 *             if self._debug_info is not None:             # <<<<<<<<<<<<<<
 *                 context['source_traceback'] = self._debug_info[1]
 * 
 */
          }

          /* "uvloop/cbhandles.pyx":266
 *                 context['source_traceback'] = self._debug_info[1]
 * 
 *             self.loop.call_exception_handler(context)             # <<<<<<<<<<<<<<
 *         else:
 *             if self.loop._debug:
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_11)) __PYX_ERR(10, 266, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          __pyx_t_5 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_5 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_v_context};
            __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 266, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }

        /* "uvloop/cbhandles.pyx":256
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             context = {
 *                 'message': 'Exception in callback {}'.format(callback),
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L26;
          }
          __pyx_L25_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_9 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L11_except_error;
          }
          __pyx_L26:;
        }
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L10_exception_handled;
      }
      goto __pyx_L11_except_error;

      /* "uvloop/cbhandles.pyx":246
 *         if self.loop._debug:
 *             started = time_monotonic()
 *         try:             # <<<<<<<<<<<<<<
 *             assert self.context is not None
 *             Context_Enter(self.context)
 */
      __pyx_L11_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L7_error;
      __pyx_L10_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L14_try_end:;
    }
  }

  /* "uvloop/cbhandles.pyx":275
 *                         self, delta)
 *         finally:
 *             context = self.context             # <<<<<<<<<<<<<<
 *             Py_DECREF(self)
 *             Context_Exit(context)
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_3 = __pyx_v_self->context;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_context, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "uvloop/cbhandles.pyx":276
 *         finally:
 *             context = self.context
 *             Py_DECREF(self)             # <<<<<<<<<<<<<<
 *             Context_Exit(context)
 *             self._clear()
 */
      Py_DECREF(((PyObject *)__pyx_v_self));

      /* "uvloop/cbhandles.pyx":277
 *             context = self.context
 *             Py_DECREF(self)
 *             Context_Exit(context)             # <<<<<<<<<<<<<<
 *             self._clear()
 * 
 */
      __pyx_t_14 = Context_Exit(__pyx_v_context); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(10, 277, __pyx_L1_error)

      /* "uvloop/cbhandles.pyx":278
 *             Py_DECREF(self)
 *             Context_Exit(context)
 *             self._clear()             # <<<<<<<<<<<<<<
 * 
 *     # Public API
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_11TimerHandle__clear(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L8;
    }
    __pyx_L7_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_14 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {

        /* "uvloop/cbhandles.pyx":275
 *                         self, delta)
 *         finally:
 *             context = self.context             # <<<<<<<<<<<<<<
 *             Py_DECREF(self)
 *             Context_Exit(context)
 */
        __pyx_t_3 = __pyx_v_self->context;
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_XDECREF_SET(__pyx_v_context, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "uvloop/cbhandles.pyx":276
 *         finally:
 *             context = self.context
 *             Py_DECREF(self)             # <<<<<<<<<<<<<<
 *             Context_Exit(context)
 *             self._clear()
 */
        Py_DECREF(((PyObject *)__pyx_v_self));

        /* "uvloop/cbhandles.pyx":277
 *             context = self.context
 *             Py_DECREF(self)
 *             Context_Exit(context)             # <<<<<<<<<<<<<<
 *             self._clear()
 * 
 */
        __pyx_t_23 = Context_Exit(__pyx_v_context); if (unlikely(__pyx_t_23 == ((int)-1))) __PYX_ERR(10, 277, __pyx_L33_error)

        /* "uvloop/cbhandles.pyx":278
 *             Py_DECREF(self)
 *             Context_Exit(context)
 *             self._clear()             # <<<<<<<<<<<<<<
 * 
 *     # Public API
 */
        __pyx_t_3 = __pyx_f_6uvloop_4loop_11TimerHandle__clear(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 278, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
      __pyx_L33_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L8:;
  }

  /* "uvloop/cbhandles.pyx":231
 *             self.timer = None  # let the UVTimer handle GC
 * 
 *     cdef _run(self):             # <<<<<<<<<<<<<<
 *         if self._cancelled == 1:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle._run", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_started);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_context);
  __Pyx_XDECREF(__pyx_v_delta);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":282
 *     # Public API
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         info = [self.__class__.__name__]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_5__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle_4__repr__(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_4__repr__(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_v_info = NULL;
  PyObject *__pyx_v_callback_name = NULL;
  PyObject *__pyx_v_source_traceback = NULL;
  PyObject *__pyx_v_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/cbhandles.pyx":283
 * 
 *     def __repr__(self):
 *         info = [self.__class__.__name__]             # <<<<<<<<<<<<<<
 * 
 *         if self._cancelled:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(10, 283, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_v_info = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":285
 *         info = [self.__class__.__name__]
 * 
 *         if self._cancelled:             # <<<<<<<<<<<<<<
 *             info.append('cancelled')
 * 
 */
  if (__pyx_v_self->_cancelled) {

    /* "uvloop/cbhandles.pyx":286
 * 
 *         if self._cancelled:
 *             info.append('cancelled')             # <<<<<<<<<<<<<<
 * 
 *         if self._debug_info is not None:
 */
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_n_u_cancelled); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 286, __pyx_L1_error)

    /* "uvloop/cbhandles.pyx":285
 *         info = [self.__class__.__name__]
 * 
 *         if self._cancelled:             # <<<<<<<<<<<<<<
 *             info.append('cancelled')
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":288
 *             info.append('cancelled')
 * 
 *         if self._debug_info is not None:             # <<<<<<<<<<<<<<
 *             callback_name = self._debug_info[0]
 *             source_traceback = self._debug_info[1]
 */
  __pyx_t_4 = (__pyx_v_self->_debug_info != ((PyObject*)Py_None));
  if (__pyx_t_4) {

    /* "uvloop/cbhandles.pyx":289
 * 
 *         if self._debug_info is not None:
 *             callback_name = self._debug_info[0]             # <<<<<<<<<<<<<<
 *             source_traceback = self._debug_info[1]
 *         else:
 */
    if (unlikely(__pyx_v_self->_debug_info == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(10, 289, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_self->_debug_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_callback_name = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/cbhandles.pyx":290
 *         if self._debug_info is not None:
 *             callback_name = self._debug_info[0]
 *             source_traceback = self._debug_info[1]             # <<<<<<<<<<<<<<
 *         else:
 *             callback_name = None
 */
    if (unlikely(__pyx_v_self->_debug_info == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(10, 290, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_self->_debug_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_source_traceback = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/cbhandles.pyx":288
 *             info.append('cancelled')
 * 
 *         if self._debug_info is not None:             # <<<<<<<<<<<<<<
 *             callback_name = self._debug_info[0]
 *             source_traceback = self._debug_info[1]
 */
    goto __pyx_L4;
  }

  /* "uvloop/cbhandles.pyx":292
 *             source_traceback = self._debug_info[1]
 *         else:
 *             callback_name = None             # <<<<<<<<<<<<<<
 *             source_traceback = None
 * 
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_callback_name = Py_None;

    /* "uvloop/cbhandles.pyx":293
 *         else:
 *             callback_name = None
 *             source_traceback = None             # <<<<<<<<<<<<<<
 * 
 *         if callback_name is not None:
 */
    __Pyx_INCREF(Py_None);
    __pyx_v_source_traceback = Py_None;
  }
  __pyx_L4:;

  /* "uvloop/cbhandles.pyx":295
 *             source_traceback = None
 * 
 *         if callback_name is not None:             # <<<<<<<<<<<<<<
 *             info.append(callback_name)
 *         elif self.callback is not None:
 */
  __pyx_t_4 = (__pyx_v_callback_name != Py_None);
  if (__pyx_t_4) {

    /* "uvloop/cbhandles.pyx":296
 * 
 *         if callback_name is not None:
 *             info.append(callback_name)             # <<<<<<<<<<<<<<
 *         elif self.callback is not None:
 *             info.append(format_callback_name(self.callback))
 */
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_v_callback_name); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 296, __pyx_L1_error)

    /* "uvloop/cbhandles.pyx":295
 *             source_traceback = None
 * 
 *         if callback_name is not None:             # <<<<<<<<<<<<<<
 *             info.append(callback_name)
 *         elif self.callback is not None:
 */
    goto __pyx_L5;
  }

  /* "uvloop/cbhandles.pyx":297
 *         if callback_name is not None:
 *             info.append(callback_name)
 *         elif self.callback is not None:             # <<<<<<<<<<<<<<
 *             info.append(format_callback_name(self.callback))
 * 
 */
  __pyx_t_4 = (__pyx_v_self->callback != Py_None);
  if (__pyx_t_4) {

    /* "uvloop/cbhandles.pyx":298
 *             info.append(callback_name)
 *         elif self.callback is not None:
 *             info.append(format_callback_name(self.callback))             # <<<<<<<<<<<<<<
 * 
 *         if source_traceback is not None:
 */
    __pyx_t_1 = __pyx_v_self->callback;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_f_6uvloop_4loop_format_callback_name(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":297
 *         if callback_name is not None:
 *             info.append(callback_name)
 *         elif self.callback is not None:             # <<<<<<<<<<<<<<
 *             info.append(format_callback_name(self.callback))
 * 
 */
  }
  __pyx_L5:;

  /* "uvloop/cbhandles.pyx":300
 *             info.append(format_callback_name(self.callback))
 * 
 *         if source_traceback is not None:             # <<<<<<<<<<<<<<
 *             frame = source_traceback[-1]
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 */
  __pyx_t_4 = (__pyx_v_source_traceback != Py_None);
  if (__pyx_t_4) {

    /* "uvloop/cbhandles.pyx":301
 * 
 *         if source_traceback is not None:
 *             frame = source_traceback[-1]             # <<<<<<<<<<<<<<
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 * 
 */
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_source_traceback, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_frame = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":302
 *         if source_traceback is not None:
 *             frame = source_traceback[-1]
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))             # <<<<<<<<<<<<<<
 * 
 *         return '<' + ' '.join(info) + '>'
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_created_at, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_frame, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_frame, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_5, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 302, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_info, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(10, 302, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":300
 *             info.append(format_callback_name(self.callback))
 * 
 *         if source_traceback is not None:             # <<<<<<<<<<<<<<
 *             frame = source_traceback[-1]
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 */
  }

  /* "uvloop/cbhandles.pyx":304
 *             info.append('created at {}:{}'.format(frame[0], frame[1]))
 * 
 *         return '<' + ' '.join(info) + '>'             # <<<<<<<<<<<<<<
 * 
 *     def cancelled(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__129, __pyx_v_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_kp_u__128, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_1, __pyx_kp_u__130); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":282
 *     # Public API
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         info = [self.__class__.__name__]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_callback_name);
  __Pyx_XDECREF(__pyx_v_source_traceback);
  __Pyx_XDECREF(__pyx_v_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":306
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_7cancelled(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11TimerHandle_6cancelled, "TimerHandle.cancelled(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11TimerHandle_7cancelled = {"cancelled", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_7cancelled, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_6cancelled};
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_7cancelled(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancelled (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("cancelled", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "cancelled", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle_6cancelled(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_6cancelled(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancelled", 1);

  /* "uvloop/cbhandles.pyx":307
 * 
 *     def cancelled(self):
 *         return self._cancelled             # <<<<<<<<<<<<<<
 * 
 *     def cancel(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_cancelled); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":306
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.cancelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":309
 *         return self._cancelled
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_9cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11TimerHandle_8cancel, "TimerHandle.cancel(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11TimerHandle_9cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_9cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_8cancel};
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_9cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("cancel", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "cancel", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle_8cancel(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_8cancel(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/cbhandles.pyx":310
 * 
 *     def cancel(self):
 *         self._cancel()             # <<<<<<<<<<<<<<
 * 
 *     def when(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TimerHandle *)__pyx_v_self->__pyx_vtab)->_cancel(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":309
 *         return self._cancelled
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":312
 *         self._cancel()
 * 
 *     def when(self):             # <<<<<<<<<<<<<<
 *         return self._when
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_11when(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11TimerHandle_10when, "TimerHandle.when(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11TimerHandle_11when = {"when", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_11when, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_10when};
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_11when(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("when (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("when", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "when", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle_10when(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_10when(struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("when", 1);

  /* "uvloop/cbhandles.pyx":313
 * 
 *     def when(self):
 *         return self._when             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_when);
  __pyx_r = __pyx_v_self->_when;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":312
 *         self._cancel()
 * 
 *     def when(self):             # <<<<<<<<<<<<<<
 *         return self._when
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11TimerHandle_12__reduce_cython__, "TimerHandle.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11TimerHandle_13__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_12__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_13__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle_12__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11TimerHandle_14__setstate_cython__, "TimerHandle.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11TimerHandle_15__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_14__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_11TimerHandle_15__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11TimerHandle_14__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_TimerHandle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11TimerHandle_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.TimerHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":316
 * 
 * 
 * cdef format_callback_name(func):             # <<<<<<<<<<<<<<
 *     if hasattr(func, '__qualname__'):
 *         cb_name = getattr(func, '__qualname__')
 */

static PyObject *__pyx_f_6uvloop_4loop_format_callback_name(PyObject *__pyx_v_func) {
  PyObject *__pyx_v_cb_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("format_callback_name", 1);

  /* "uvloop/cbhandles.pyx":317
 * 
 * cdef format_callback_name(func):
 *     if hasattr(func, '__qualname__'):             # <<<<<<<<<<<<<<
 *         cb_name = getattr(func, '__qualname__')
 *     elif hasattr(func, '__name__'):
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_func, __pyx_n_u_qualname); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(10, 317, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":318
 * cdef format_callback_name(func):
 *     if hasattr(func, '__qualname__'):
 *         cb_name = getattr(func, '__qualname__')             # <<<<<<<<<<<<<<
 *     elif hasattr(func, '__name__'):
 *         cb_name = getattr(func, '__name__')
 */
    __pyx_t_2 = __Pyx_GetAttr(__pyx_v_func, __pyx_n_u_qualname); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_cb_name = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":317
 * 
 * cdef format_callback_name(func):
 *     if hasattr(func, '__qualname__'):             # <<<<<<<<<<<<<<
 *         cb_name = getattr(func, '__qualname__')
 *     elif hasattr(func, '__name__'):
 */
    goto __pyx_L3;
  }

  /* "uvloop/cbhandles.pyx":319
 *     if hasattr(func, '__qualname__'):
 *         cb_name = getattr(func, '__qualname__')
 *     elif hasattr(func, '__name__'):             # <<<<<<<<<<<<<<
 *         cb_name = getattr(func, '__name__')
 *     else:
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_func, __pyx_n_u_name); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(10, 319, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/cbhandles.pyx":320
 *         cb_name = getattr(func, '__qualname__')
 *     elif hasattr(func, '__name__'):
 *         cb_name = getattr(func, '__name__')             # <<<<<<<<<<<<<<
 *     else:
 *         cb_name = repr(func)
 */
    __pyx_t_2 = __Pyx_GetAttr(__pyx_v_func, __pyx_n_u_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_cb_name = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/cbhandles.pyx":319
 *     if hasattr(func, '__qualname__'):
 *         cb_name = getattr(func, '__qualname__')
 *     elif hasattr(func, '__name__'):             # <<<<<<<<<<<<<<
 *         cb_name = getattr(func, '__name__')
 *     else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/cbhandles.pyx":322
 *         cb_name = getattr(func, '__name__')
 *     else:
 *         cb_name = repr(func)             # <<<<<<<<<<<<<<
 *     return cb_name
 * 
 */
  /*else*/ {
    __pyx_t_2 = PyObject_Repr(__pyx_v_func); if (unlikely(!__pyx_t_2)) __PYX_ERR(10, 322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_cb_name = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uvloop/cbhandles.pyx":323
 *     else:
 *         cb_name = repr(func)
 *     return cb_name             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cb_name);
  __pyx_r = __pyx_v_cb_name;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":316
 * 
 * 
 * cdef format_callback_name(func):             # <<<<<<<<<<<<<<
 *     if hasattr(func, '__qualname__'):
 *         cb_name = getattr(func, '__qualname__')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.format_callback_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cb_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":326
 * 
 * 
 * cdef new_Handle(Loop loop, object callback, object args, object context):             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)
 */

static PyObject *__pyx_f_6uvloop_4loop_new_Handle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback, PyObject *__pyx_v_args, PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_Handle", 1);

  /* "uvloop/cbhandles.pyx":328
 * cdef new_Handle(Loop loop, object callback, object args, object context):
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)             # <<<<<<<<<<<<<<
 *     handle._set_loop(loop)
 *     handle._set_context(context)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_Handle(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_Handle), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 328, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":329
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)             # <<<<<<<<<<<<<<
 *     handle._set_context(context)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_loop(__pyx_v_handle, __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":330
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)
 *     handle._set_context(context)             # <<<<<<<<<<<<<<
 * 
 *     handle.cb_type = 1
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_context(__pyx_v_handle, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":332
 *     handle._set_context(context)
 * 
 *     handle.cb_type = 1             # <<<<<<<<<<<<<<
 * 
 *     handle.arg1 = callback
 */
  __pyx_v_handle->cb_type = 1;

  /* "uvloop/cbhandles.pyx":334
 *     handle.cb_type = 1
 * 
 *     handle.arg1 = callback             # <<<<<<<<<<<<<<
 *     handle.arg2 = args
 * 
 */
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  __Pyx_GOTREF(__pyx_v_handle->arg1);
  __Pyx_DECREF(__pyx_v_handle->arg1);
  __pyx_v_handle->arg1 = __pyx_v_callback;

  /* "uvloop/cbhandles.pyx":335
 * 
 *     handle.arg1 = callback
 *     handle.arg2 = args             # <<<<<<<<<<<<<<
 * 
 *     return handle
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_handle->arg2);
  __Pyx_DECREF(__pyx_v_handle->arg2);
  __pyx_v_handle->arg2 = __pyx_v_args;

  /* "uvloop/cbhandles.pyx":337
 *     handle.arg2 = args
 * 
 *     return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = ((PyObject *)__pyx_v_handle);
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":326
 * 
 * 
 * cdef new_Handle(Loop loop, object callback, object args, object context):             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.new_Handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":340
 * 
 * 
 * cdef new_MethodHandle(Loop loop, str name, method_t callback, object context,             # <<<<<<<<<<<<<<
 *                       object bound_to):
 *     cdef Handle handle
 */

static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_name, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_bound_to) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_MethodHandle", 1);

  /* "uvloop/cbhandles.pyx":343
 *                       object bound_to):
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)             # <<<<<<<<<<<<<<
 *     handle._set_loop(loop)
 *     handle._set_context(context)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_Handle(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_Handle), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 343, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":344
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)             # <<<<<<<<<<<<<<
 *     handle._set_context(context)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_loop(__pyx_v_handle, __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":345
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)
 *     handle._set_context(context)             # <<<<<<<<<<<<<<
 * 
 *     handle.cb_type = 2
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_context(__pyx_v_handle, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":347
 *     handle._set_context(context)
 * 
 *     handle.cb_type = 2             # <<<<<<<<<<<<<<
 *     handle.meth_name = name
 * 
 */
  __pyx_v_handle->cb_type = 2;

  /* "uvloop/cbhandles.pyx":348
 * 
 *     handle.cb_type = 2
 *     handle.meth_name = name             # <<<<<<<<<<<<<<
 * 
 *     handle.callback = <void*> callback
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_handle->meth_name);
  __Pyx_DECREF(__pyx_v_handle->meth_name);
  __pyx_v_handle->meth_name = __pyx_v_name;

  /* "uvloop/cbhandles.pyx":350
 *     handle.meth_name = name
 * 
 *     handle.callback = <void*> callback             # <<<<<<<<<<<<<<
 *     handle.arg1 = bound_to
 * 
 */
  __pyx_v_handle->callback = ((void *)__pyx_v_callback);

  /* "uvloop/cbhandles.pyx":351
 * 
 *     handle.callback = <void*> callback
 *     handle.arg1 = bound_to             # <<<<<<<<<<<<<<
 * 
 *     return handle
 */
  __Pyx_INCREF(__pyx_v_bound_to);
  __Pyx_GIVEREF(__pyx_v_bound_to);
  __Pyx_GOTREF(__pyx_v_handle->arg1);
  __Pyx_DECREF(__pyx_v_handle->arg1);
  __pyx_v_handle->arg1 = __pyx_v_bound_to;

  /* "uvloop/cbhandles.pyx":353
 *     handle.arg1 = bound_to
 * 
 *     return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = ((PyObject *)__pyx_v_handle);
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":340
 * 
 * 
 * cdef new_MethodHandle(Loop loop, str name, method_t callback, object context,             # <<<<<<<<<<<<<<
 *                       object bound_to):
 *     cdef Handle handle
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.new_MethodHandle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":356
 * 
 * 
 * cdef new_MethodHandle1(Loop loop, str name, method1_t callback, object context,             # <<<<<<<<<<<<<<
 *                        object bound_to, object arg):
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle1(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_name, __pyx_t_6uvloop_4loop_method1_t __pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_bound_to, PyObject *__pyx_v_arg) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_MethodHandle1", 1);

  /* "uvloop/cbhandles.pyx":360
 * 
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)             # <<<<<<<<<<<<<<
 *     handle._set_loop(loop)
 *     handle._set_context(context)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_Handle(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_Handle), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 360, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":361
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)             # <<<<<<<<<<<<<<
 *     handle._set_context(context)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_loop(__pyx_v_handle, __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":362
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)
 *     handle._set_context(context)             # <<<<<<<<<<<<<<
 * 
 *     handle.cb_type = 3
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_context(__pyx_v_handle, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":364
 *     handle._set_context(context)
 * 
 *     handle.cb_type = 3             # <<<<<<<<<<<<<<
 *     handle.meth_name = name
 * 
 */
  __pyx_v_handle->cb_type = 3;

  /* "uvloop/cbhandles.pyx":365
 * 
 *     handle.cb_type = 3
 *     handle.meth_name = name             # <<<<<<<<<<<<<<
 * 
 *     handle.callback = <void*> callback
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_handle->meth_name);
  __Pyx_DECREF(__pyx_v_handle->meth_name);
  __pyx_v_handle->meth_name = __pyx_v_name;

  /* "uvloop/cbhandles.pyx":367
 *     handle.meth_name = name
 * 
 *     handle.callback = <void*> callback             # <<<<<<<<<<<<<<
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg
 */
  __pyx_v_handle->callback = ((void *)__pyx_v_callback);

  /* "uvloop/cbhandles.pyx":368
 * 
 *     handle.callback = <void*> callback
 *     handle.arg1 = bound_to             # <<<<<<<<<<<<<<
 *     handle.arg2 = arg
 * 
 */
  __Pyx_INCREF(__pyx_v_bound_to);
  __Pyx_GIVEREF(__pyx_v_bound_to);
  __Pyx_GOTREF(__pyx_v_handle->arg1);
  __Pyx_DECREF(__pyx_v_handle->arg1);
  __pyx_v_handle->arg1 = __pyx_v_bound_to;

  /* "uvloop/cbhandles.pyx":369
 *     handle.callback = <void*> callback
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg             # <<<<<<<<<<<<<<
 * 
 *     return handle
 */
  __Pyx_INCREF(__pyx_v_arg);
  __Pyx_GIVEREF(__pyx_v_arg);
  __Pyx_GOTREF(__pyx_v_handle->arg2);
  __Pyx_DECREF(__pyx_v_handle->arg2);
  __pyx_v_handle->arg2 = __pyx_v_arg;

  /* "uvloop/cbhandles.pyx":371
 *     handle.arg2 = arg
 * 
 *     return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = ((PyObject *)__pyx_v_handle);
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":356
 * 
 * 
 * cdef new_MethodHandle1(Loop loop, str name, method1_t callback, object context,             # <<<<<<<<<<<<<<
 *                        object bound_to, object arg):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.new_MethodHandle1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":374
 * 
 * 
 * cdef new_MethodHandle2(Loop loop, str name, method2_t callback, object context,             # <<<<<<<<<<<<<<
 *                        object bound_to, object arg1, object arg2):
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle2(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_name, __pyx_t_6uvloop_4loop_method2_t __pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_bound_to, PyObject *__pyx_v_arg1, PyObject *__pyx_v_arg2) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_MethodHandle2", 1);

  /* "uvloop/cbhandles.pyx":378
 * 
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)             # <<<<<<<<<<<<<<
 *     handle._set_loop(loop)
 *     handle._set_context(context)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_Handle(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_Handle), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 378, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":379
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)             # <<<<<<<<<<<<<<
 *     handle._set_context(context)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_loop(__pyx_v_handle, __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":380
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)
 *     handle._set_context(context)             # <<<<<<<<<<<<<<
 * 
 *     handle.cb_type = 4
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_context(__pyx_v_handle, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":382
 *     handle._set_context(context)
 * 
 *     handle.cb_type = 4             # <<<<<<<<<<<<<<
 *     handle.meth_name = name
 * 
 */
  __pyx_v_handle->cb_type = 4;

  /* "uvloop/cbhandles.pyx":383
 * 
 *     handle.cb_type = 4
 *     handle.meth_name = name             # <<<<<<<<<<<<<<
 * 
 *     handle.callback = <void*> callback
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_handle->meth_name);
  __Pyx_DECREF(__pyx_v_handle->meth_name);
  __pyx_v_handle->meth_name = __pyx_v_name;

  /* "uvloop/cbhandles.pyx":385
 *     handle.meth_name = name
 * 
 *     handle.callback = <void*> callback             # <<<<<<<<<<<<<<
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg1
 */
  __pyx_v_handle->callback = ((void *)__pyx_v_callback);

  /* "uvloop/cbhandles.pyx":386
 * 
 *     handle.callback = <void*> callback
 *     handle.arg1 = bound_to             # <<<<<<<<<<<<<<
 *     handle.arg2 = arg1
 *     handle.arg3 = arg2
 */
  __Pyx_INCREF(__pyx_v_bound_to);
  __Pyx_GIVEREF(__pyx_v_bound_to);
  __Pyx_GOTREF(__pyx_v_handle->arg1);
  __Pyx_DECREF(__pyx_v_handle->arg1);
  __pyx_v_handle->arg1 = __pyx_v_bound_to;

  /* "uvloop/cbhandles.pyx":387
 *     handle.callback = <void*> callback
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg1             # <<<<<<<<<<<<<<
 *     handle.arg3 = arg2
 * 
 */
  __Pyx_INCREF(__pyx_v_arg1);
  __Pyx_GIVEREF(__pyx_v_arg1);
  __Pyx_GOTREF(__pyx_v_handle->arg2);
  __Pyx_DECREF(__pyx_v_handle->arg2);
  __pyx_v_handle->arg2 = __pyx_v_arg1;

  /* "uvloop/cbhandles.pyx":388
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg1
 *     handle.arg3 = arg2             # <<<<<<<<<<<<<<
 * 
 *     return handle
 */
  __Pyx_INCREF(__pyx_v_arg2);
  __Pyx_GIVEREF(__pyx_v_arg2);
  __Pyx_GOTREF(__pyx_v_handle->arg3);
  __Pyx_DECREF(__pyx_v_handle->arg3);
  __pyx_v_handle->arg3 = __pyx_v_arg2;

  /* "uvloop/cbhandles.pyx":390
 *     handle.arg3 = arg2
 * 
 *     return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = ((PyObject *)__pyx_v_handle);
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":374
 * 
 * 
 * cdef new_MethodHandle2(Loop loop, str name, method2_t callback, object context,             # <<<<<<<<<<<<<<
 *                        object bound_to, object arg1, object arg2):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.new_MethodHandle2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":393
 * 
 * 
 * cdef new_MethodHandle3(Loop loop, str name, method3_t callback, object context,             # <<<<<<<<<<<<<<
 *                        object bound_to, object arg1, object arg2, object arg3):
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_new_MethodHandle3(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_name, __pyx_t_6uvloop_4loop_method3_t __pyx_v_callback, PyObject *__pyx_v_context, PyObject *__pyx_v_bound_to, PyObject *__pyx_v_arg1, PyObject *__pyx_v_arg2, PyObject *__pyx_v_arg3) {
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_handle = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new_MethodHandle3", 1);

  /* "uvloop/cbhandles.pyx":397
 * 
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)             # <<<<<<<<<<<<<<
 *     handle._set_loop(loop)
 *     handle._set_context(context)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_Handle(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_Handle), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 397, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":398
 *     cdef Handle handle
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)             # <<<<<<<<<<<<<<
 *     handle._set_context(context)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_loop(__pyx_v_handle, __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":399
 *     handle = Handle.__new__(Handle)
 *     handle._set_loop(loop)
 *     handle._set_context(context)             # <<<<<<<<<<<<<<
 * 
 *     handle.cb_type = 5
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_6Handle__set_context(__pyx_v_handle, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/cbhandles.pyx":401
 *     handle._set_context(context)
 * 
 *     handle.cb_type = 5             # <<<<<<<<<<<<<<
 *     handle.meth_name = name
 * 
 */
  __pyx_v_handle->cb_type = 5;

  /* "uvloop/cbhandles.pyx":402
 * 
 *     handle.cb_type = 5
 *     handle.meth_name = name             # <<<<<<<<<<<<<<
 * 
 *     handle.callback = <void*> callback
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_handle->meth_name);
  __Pyx_DECREF(__pyx_v_handle->meth_name);
  __pyx_v_handle->meth_name = __pyx_v_name;

  /* "uvloop/cbhandles.pyx":404
 *     handle.meth_name = name
 * 
 *     handle.callback = <void*> callback             # <<<<<<<<<<<<<<
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg1
 */
  __pyx_v_handle->callback = ((void *)__pyx_v_callback);

  /* "uvloop/cbhandles.pyx":405
 * 
 *     handle.callback = <void*> callback
 *     handle.arg1 = bound_to             # <<<<<<<<<<<<<<
 *     handle.arg2 = arg1
 *     handle.arg3 = arg2
 */
  __Pyx_INCREF(__pyx_v_bound_to);
  __Pyx_GIVEREF(__pyx_v_bound_to);
  __Pyx_GOTREF(__pyx_v_handle->arg1);
  __Pyx_DECREF(__pyx_v_handle->arg1);
  __pyx_v_handle->arg1 = __pyx_v_bound_to;

  /* "uvloop/cbhandles.pyx":406
 *     handle.callback = <void*> callback
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg1             # <<<<<<<<<<<<<<
 *     handle.arg3 = arg2
 *     handle.arg4 = arg3
 */
  __Pyx_INCREF(__pyx_v_arg1);
  __Pyx_GIVEREF(__pyx_v_arg1);
  __Pyx_GOTREF(__pyx_v_handle->arg2);
  __Pyx_DECREF(__pyx_v_handle->arg2);
  __pyx_v_handle->arg2 = __pyx_v_arg1;

  /* "uvloop/cbhandles.pyx":407
 *     handle.arg1 = bound_to
 *     handle.arg2 = arg1
 *     handle.arg3 = arg2             # <<<<<<<<<<<<<<
 *     handle.arg4 = arg3
 * 
 */
  __Pyx_INCREF(__pyx_v_arg2);
  __Pyx_GIVEREF(__pyx_v_arg2);
  __Pyx_GOTREF(__pyx_v_handle->arg3);
  __Pyx_DECREF(__pyx_v_handle->arg3);
  __pyx_v_handle->arg3 = __pyx_v_arg2;

  /* "uvloop/cbhandles.pyx":408
 *     handle.arg2 = arg1
 *     handle.arg3 = arg2
 *     handle.arg4 = arg3             # <<<<<<<<<<<<<<
 * 
 *     return handle
 */
  __Pyx_INCREF(__pyx_v_arg3);
  __Pyx_GIVEREF(__pyx_v_arg3);
  __Pyx_GOTREF(__pyx_v_handle->arg4);
  __Pyx_DECREF(__pyx_v_handle->arg4);
  __pyx_v_handle->arg4 = __pyx_v_arg3;

  /* "uvloop/cbhandles.pyx":410
 *     handle.arg4 = arg3
 * 
 *     return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = ((PyObject *)__pyx_v_handle);
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":393
 * 
 * 
 * cdef new_MethodHandle3(Loop loop, str name, method3_t callback, object context,             # <<<<<<<<<<<<<<
 *                        object bound_to, object arg1, object arg2, object arg3):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.new_MethodHandle3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/cbhandles.pyx":413
 * 
 * 
 * cdef extract_stack():             # <<<<<<<<<<<<<<
 *     """Replacement for traceback.extract_stack() that only does the
 *     necessary work for asyncio debug mode.
 */

static PyObject *__pyx_f_6uvloop_4loop_extract_stack(void) {
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_stack = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extract_stack", 1);

  /* "uvloop/cbhandles.pyx":417
 *     necessary work for asyncio debug mode.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         f = sys_getframe()
 *     # sys._getframe() might raise ValueError if being called without a frame, e.g.
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/cbhandles.pyx":418
 *     """
 *     try:
 *         f = sys_getframe()             # <<<<<<<<<<<<<<
 *     # sys._getframe() might raise ValueError if being called without a frame, e.g.
 *     # from Cython or similar C extensions.
 */
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_sys_getframe);
      __pyx_t_5 = __pyx_v_6uvloop_4loop_sys_getframe; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 418, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_v_f = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/cbhandles.pyx":417
 *     necessary work for asyncio debug mode.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         f = sys_getframe()
 *     # sys._getframe() might raise ValueError if being called without a frame, e.g.
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/cbhandles.pyx":421
 *     # sys._getframe() might raise ValueError if being called without a frame, e.g.
 *     # from Cython or similar C extensions.
 *     except ValueError:             # <<<<<<<<<<<<<<
 *         return None
 *     if f is None:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.extract_stack", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(10, 421, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "uvloop/cbhandles.pyx":422
 *     # from Cython or similar C extensions.
 *     except ValueError:
 *         return None             # <<<<<<<<<<<<<<
 *     if f is None:
 *         return
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/cbhandles.pyx":417
 *     necessary work for asyncio debug mode.
 *     """
 *     try:             # <<<<<<<<<<<<<<
 *         f = sys_getframe()
 *     # sys._getframe() might raise ValueError if being called without a frame, e.g.
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "uvloop/cbhandles.pyx":423
 *     except ValueError:
 *         return None
 *     if f is None:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_9 = (__pyx_v_f == Py_None);
  if (__pyx_t_9) {

    /* "uvloop/cbhandles.pyx":424
 *         return None
 *     if f is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/cbhandles.pyx":423
 *     except ValueError:
 *         return None
 *     if f is None:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/cbhandles.pyx":426
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         stack = tb_StackSummary.extract(tb_walk_stack(f),
 *                                         limit=DEBUG_STACK_DEPTH,
 */
  /*try:*/ {

    /* "uvloop/cbhandles.pyx":427
 * 
 *     try:
 *         stack = tb_StackSummary.extract(tb_walk_stack(f),             # <<<<<<<<<<<<<<
 *                                         limit=DEBUG_STACK_DEPTH,
 *                                         lookup_lines=False)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_tb_StackSummary, __pyx_n_s_extract); if (unlikely(!__pyx_t_6)) __PYX_ERR(10, 427, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_tb_walk_stack);
    __pyx_t_4 = __pyx_v_6uvloop_4loop_tb_walk_stack; __pyx_t_10 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_v_f};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 427, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 427, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5)) __PYX_ERR(10, 427, __pyx_L13_error);
    __pyx_t_5 = 0;

    /* "uvloop/cbhandles.pyx":428
 *     try:
 *         stack = tb_StackSummary.extract(tb_walk_stack(f),
 *                                         limit=DEBUG_STACK_DEPTH,             # <<<<<<<<<<<<<<
 *                                         lookup_lines=False)
 *     finally:
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 428, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From___pyx_anon_enum(__pyx_e_6uvloop_4loop_DEBUG_STACK_DEPTH); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 428, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_limit, __pyx_t_10) < 0) __PYX_ERR(10, 428, __pyx_L13_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "uvloop/cbhandles.pyx":429
 *         stack = tb_StackSummary.extract(tb_walk_stack(f),
 *                                         limit=DEBUG_STACK_DEPTH,
 *                                         lookup_lines=False)             # <<<<<<<<<<<<<<
 *     finally:
 *         f = None
 */
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_lookup_lines, Py_False) < 0) __PYX_ERR(10, 428, __pyx_L13_error)

    /* "uvloop/cbhandles.pyx":427
 * 
 *     try:
 *         stack = tb_StackSummary.extract(tb_walk_stack(f),             # <<<<<<<<<<<<<<
 *                                         limit=DEBUG_STACK_DEPTH,
 *                                         lookup_lines=False)
 */
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 427, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_stack = __pyx_t_10;
    __pyx_t_10 = 0;
  }

  /* "uvloop/cbhandles.pyx":431
 *                                         lookup_lines=False)
 *     finally:
 *         f = None             # <<<<<<<<<<<<<<
 * 
 *     stack.reverse()
 */
  /*finally:*/ {
    /*normal exit:*/{
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_f, Py_None);
      goto __pyx_L14;
    }
    __pyx_L13_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1) < 0)) __Pyx_ErrFetch(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_8 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
      {
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_f, Py_None);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_1);
      __Pyx_ErrRestore(__pyx_t_3, __pyx_t_2, __pyx_t_1);
      __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_1 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
      goto __pyx_L1_error;
    }
    __pyx_L14:;
  }

  /* "uvloop/cbhandles.pyx":433
 *         f = None
 * 
 *     stack.reverse()             # <<<<<<<<<<<<<<
 *     return stack
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_stack, __pyx_n_s_reverse); if (unlikely(!__pyx_t_5)) __PYX_ERR(10, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(10, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "uvloop/cbhandles.pyx":434
 * 
 *     stack.reverse()
 *     return stack             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_stack);
  __pyx_r = __pyx_v_stack;
  goto __pyx_L0;

  /* "uvloop/cbhandles.pyx":413
 * 
 * 
 * cdef extract_stack():             # <<<<<<<<<<<<<<
 *     """Replacement for traceback.extract_stack() that only does the
 *     necessary work for asyncio debug mode.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uvloop.loop.extract_stack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_stack);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":10
 *         object _sockname
 * 
 *     def __init__(self, int family, int type, int proto, int fd):             # <<<<<<<<<<<<<<
 *         self._family = family
 *         self._type = type
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_12PseudoSocket_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_12PseudoSocket_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_family;
  int __pyx_v_type;
  int __pyx_v_proto;
  int __pyx_v_fd;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_family,&__pyx_n_s_type,&__pyx_n_s_proto,&__pyx_n_s_fd,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_family)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(11, 10, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proto)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(11, 10, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fd)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(11, 10, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(11, 10, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
    }
    __pyx_v_family = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_family == (int)-1) && PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
    __pyx_v_type = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
    __pyx_v_proto = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_proto == (int)-1) && PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
    __pyx_v_fd = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_fd == (int)-1) && PyErr_Occurred())) __PYX_ERR(11, 10, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, __pyx_nargs); __PYX_ERR(11, 10, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket___init__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_family, __pyx_v_type, __pyx_v_proto, __pyx_v_fd);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_12PseudoSocket___init__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_fd) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/pseudosock.pyx":11
 * 
 *     def __init__(self, int family, int type, int proto, int fd):
 *         self._family = family             # <<<<<<<<<<<<<<
 *         self._type = type
 *         self._proto = proto
 */
  __pyx_v_self->_family = __pyx_v_family;

  /* "uvloop/pseudosock.pyx":12
 *     def __init__(self, int family, int type, int proto, int fd):
 *         self._family = family
 *         self._type = type             # <<<<<<<<<<<<<<
 *         self._proto = proto
 *         self._fd = fd
 */
  __pyx_v_self->_type = __pyx_v_type;

  /* "uvloop/pseudosock.pyx":13
 *         self._family = family
 *         self._type = type
 *         self._proto = proto             # <<<<<<<<<<<<<<
 *         self._fd = fd
 *         self._peername = None
 */
  __pyx_v_self->_proto = __pyx_v_proto;

  /* "uvloop/pseudosock.pyx":14
 *         self._type = type
 *         self._proto = proto
 *         self._fd = fd             # <<<<<<<<<<<<<<
 *         self._peername = None
 *         self._sockname = None
 */
  __pyx_v_self->_fd = __pyx_v_fd;

  /* "uvloop/pseudosock.pyx":15
 *         self._proto = proto
 *         self._fd = fd
 *         self._peername = None             # <<<<<<<<<<<<<<
 *         self._sockname = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_peername);
  __Pyx_DECREF(__pyx_v_self->_peername);
  __pyx_v_self->_peername = Py_None;

  /* "uvloop/pseudosock.pyx":16
 *         self._fd = fd
 *         self._peername = None
 *         self._sockname = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _na(self, what):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_sockname);
  __Pyx_DECREF(__pyx_v_self->_sockname);
  __pyx_v_self->_sockname = Py_None;

  /* "uvloop/pseudosock.pyx":10
 *         object _sockname
 * 
 *     def __init__(self, int family, int type, int proto, int fd):             # <<<<<<<<<<<<<<
 *         self._family = family
 *         self._type = type
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":18
 *         self._sockname = None
 * 
 *     cdef _na(self, what):             # <<<<<<<<<<<<<<
 *         raise TypeError('transport sockets do not support {}'.format(what))
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12PseudoSocket__na(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_what) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_na", 1);

  /* "uvloop/pseudosock.pyx":19
 * 
 *     cdef _na(self, what):
 *         raise TypeError('transport sockets do not support {}'.format(what))             # <<<<<<<<<<<<<<
 * 
 *     cdef _make_sock(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_transport_sockets_do_not_support, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_what};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(11, 19, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":18
 *         self._sockname = None
 * 
 *     cdef _na(self, what):             # <<<<<<<<<<<<<<
 *         raise TypeError('transport sockets do not support {}'.format(what))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket._na", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":21
 *         raise TypeError('transport sockets do not support {}'.format(what))
 * 
 *     cdef _make_sock(self):             # <<<<<<<<<<<<<<
 *         return socket_socket(self._family, self._type, self._proto, self._fd)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12PseudoSocket__make_sock(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_sock", 1);

  /* "uvloop/pseudosock.pyx":22
 * 
 *     cdef _make_sock(self):
 *         return socket_socket(self._family, self._type, self._proto, self._fd)             # <<<<<<<<<<<<<<
 * 
 *     property family:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_family); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->_proto); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->_fd); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_socket);
  __pyx_t_6 = __pyx_v_6uvloop_4loop_socket_socket; __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 4+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 22, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/pseudosock.pyx":21
 *         raise TypeError('transport sockets do not support {}'.format(what))
 * 
 *     cdef _make_sock(self):             # <<<<<<<<<<<<<<
 *         return socket_socket(self._family, self._type, self._proto, self._fd)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket._make_sock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":25
 * 
 *     property family:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             try:
 *                 return socket_AddressFamily(self._family)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_6family_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_6family_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_6family___get__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_6family___get__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uvloop/pseudosock.pyx":26
 *     property family:
 *         def __get__(self):
 *             try:             # <<<<<<<<<<<<<<
 *                 return socket_AddressFamily(self._family)
 *             except ValueError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/pseudosock.pyx":27
 *         def __get__(self):
 *             try:
 *                 return socket_AddressFamily(self._family)             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 return self._family
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->_family); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 27, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_AddressFamily);
      __pyx_t_6 = __pyx_v_6uvloop_4loop_socket_AddressFamily; __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 27, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "uvloop/pseudosock.pyx":26
 *     property family:
 *         def __get__(self):
 *             try:             # <<<<<<<<<<<<<<
 *                 return socket_AddressFamily(self._family)
 *             except ValueError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/pseudosock.pyx":28
 *             try:
 *                 return socket_AddressFamily(self._family)
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 return self._family
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.PseudoSocket.family.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(11, 28, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "uvloop/pseudosock.pyx":29
 *                 return socket_AddressFamily(self._family)
 *             except ValueError:
 *                 return self._family             # <<<<<<<<<<<<<<
 * 
 *     property type:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->_family); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 29, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/pseudosock.pyx":26
 *     property family:
 *         def __get__(self):
 *             try:             # <<<<<<<<<<<<<<
 *                 return socket_AddressFamily(self._family)
 *             except ValueError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "uvloop/pseudosock.pyx":25
 * 
 *     property family:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             try:
 *                 return socket_AddressFamily(self._family)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.family.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":32
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             try:
 *                 return socket_SocketKind(self._type)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_4type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_4type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_4type___get__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_4type___get__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uvloop/pseudosock.pyx":33
 *     property type:
 *         def __get__(self):
 *             try:             # <<<<<<<<<<<<<<
 *                 return socket_SocketKind(self._type)
 *             except ValueError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/pseudosock.pyx":34
 *         def __get__(self):
 *             try:
 *                 return socket_SocketKind(self._type)             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 return self._type
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 34, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_SocketKind);
      __pyx_t_6 = __pyx_v_6uvloop_4loop_socket_SocketKind; __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 34, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "uvloop/pseudosock.pyx":33
 *     property type:
 *         def __get__(self):
 *             try:             # <<<<<<<<<<<<<<
 *                 return socket_SocketKind(self._type)
 *             except ValueError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/pseudosock.pyx":35
 *             try:
 *                 return socket_SocketKind(self._type)
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 return self._type
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.PseudoSocket.type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(11, 35, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);

      /* "uvloop/pseudosock.pyx":36
 *                 return socket_SocketKind(self._type)
 *             except ValueError:
 *                 return self._type             # <<<<<<<<<<<<<<
 * 
 *     property proto:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 36, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/pseudosock.pyx":33
 *     property type:
 *         def __get__(self):
 *             try:             # <<<<<<<<<<<<<<
 *                 return socket_SocketKind(self._type)
 *             except ValueError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "uvloop/pseudosock.pyx":32
 * 
 *     property type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             try:
 *                 return socket_SocketKind(self._type)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":39
 * 
 *     property proto:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._proto
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_5proto_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_5proto_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_5proto___get__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_5proto___get__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uvloop/pseudosock.pyx":40
 *     property proto:
 *         def __get__(self):
 *             return self._proto             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_proto); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/pseudosock.pyx":39
 * 
 *     property proto:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._proto
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.proto.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":42
 *             return self._proto
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         s = ("<uvloop.PseudoSocket fd={}, family={!s}, "
 *              "type={!s}, proto={}").format(self.fileno(), self.family.name,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_2__repr__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_2__repr__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_laddr = NULL;
  PyObject *__pyx_v_raddr = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/pseudosock.pyx":44
 *     def __repr__(self):
 *         s = ("<uvloop.PseudoSocket fd={}, family={!s}, "
 *              "type={!s}, proto={}").format(self.fileno(), self.family.name,             # <<<<<<<<<<<<<<
 *                                            self.type.name, self.proto)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_uvloop_PseudoSocket_fd_family_s, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_fileno); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_family); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(11, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/pseudosock.pyx":45
 *         s = ("<uvloop.PseudoSocket fd={}, family={!s}, "
 *              "type={!s}, proto={}").format(self.fileno(), self.family.name,
 *                                            self.type.name, self.proto)             # <<<<<<<<<<<<<<
 * 
 *         if self._fd != -1:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(11, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_proto); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_8, __pyx_t_3, __pyx_t_5, __pyx_t_7, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 4+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_s = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":47
 *                                            self.type.name, self.proto)
 * 
 *         if self._fd != -1:             # <<<<<<<<<<<<<<
 *             try:
 *                 laddr = self.getsockname()
 */
  __pyx_t_9 = (__pyx_v_self->_fd != -1L);
  if (__pyx_t_9) {

    /* "uvloop/pseudosock.pyx":48
 * 
 *         if self._fd != -1:
 *             try:             # <<<<<<<<<<<<<<
 *                 laddr = self.getsockname()
 *                 if laddr:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      /*try:*/ {

        /* "uvloop/pseudosock.pyx":49
 *         if self._fd != -1:
 *             try:
 *                 laddr = self.getsockname()             # <<<<<<<<<<<<<<
 *                 if laddr:
 *                     s += ", laddr=%s" % str(laddr)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getsockname); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 49, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 49, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_v_laddr = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "uvloop/pseudosock.pyx":50
 *             try:
 *                 laddr = self.getsockname()
 *                 if laddr:             # <<<<<<<<<<<<<<
 *                     s += ", laddr=%s" % str(laddr)
 *             except socket_error:
 */
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_laddr); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(11, 50, __pyx_L4_error)
        if (__pyx_t_9) {

          /* "uvloop/pseudosock.pyx":51
 *                 laddr = self.getsockname()
 *                 if laddr:
 *                     s += ", laddr=%s" % str(laddr)             # <<<<<<<<<<<<<<
 *             except socket_error:
 *                 pass
 */
          __pyx_t_1 = __Pyx_PyObject_Unicode(__pyx_v_laddr); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 51, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_laddr_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 51, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 51, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_s, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "uvloop/pseudosock.pyx":50
 *             try:
 *                 laddr = self.getsockname()
 *                 if laddr:             # <<<<<<<<<<<<<<
 *                     s += ", laddr=%s" % str(laddr)
 *             except socket_error:
 */
        }

        /* "uvloop/pseudosock.pyx":48
 * 
 *         if self._fd != -1:
 *             try:             # <<<<<<<<<<<<<<
 *                 laddr = self.getsockname()
 *                 if laddr:
 */
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/pseudosock.pyx":52
 *                 if laddr:
 *                     s += ", laddr=%s" % str(laddr)
 *             except socket_error:             # <<<<<<<<<<<<<<
 *                 pass
 *             try:
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_socket_error);
      if (__pyx_t_13) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "uvloop/pseudosock.pyx":48
 * 
 *         if self._fd != -1:
 *             try:             # <<<<<<<<<<<<<<
 *                 laddr = self.getsockname()
 *                 if laddr:
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_L9_try_end:;
    }

    /* "uvloop/pseudosock.pyx":54
 *             except socket_error:
 *                 pass
 *             try:             # <<<<<<<<<<<<<<
 *                 raddr = self.getpeername()
 *                 if raddr:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "uvloop/pseudosock.pyx":55
 *                 pass
 *             try:
 *                 raddr = self.getpeername()             # <<<<<<<<<<<<<<
 *                 if raddr:
 *                     s += ", raddr=%s" % str(raddr)
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getpeername); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 55, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 55, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_v_raddr = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "uvloop/pseudosock.pyx":56
 *             try:
 *                 raddr = self.getpeername()
 *                 if raddr:             # <<<<<<<<<<<<<<
 *                     s += ", raddr=%s" % str(raddr)
 *             except socket_error:
 */
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_v_raddr); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(11, 56, __pyx_L11_error)
        if (__pyx_t_9) {

          /* "uvloop/pseudosock.pyx":57
 *                 raddr = self.getpeername()
 *                 if raddr:
 *                     s += ", raddr=%s" % str(raddr)             # <<<<<<<<<<<<<<
 *             except socket_error:
 *                 pass
 */
          __pyx_t_1 = __Pyx_PyObject_Unicode(__pyx_v_raddr); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 57, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_raddr_s, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 57, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_s, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 57, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_s, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "uvloop/pseudosock.pyx":56
 *             try:
 *                 raddr = self.getpeername()
 *                 if raddr:             # <<<<<<<<<<<<<<
 *                     s += ", raddr=%s" % str(raddr)
 *             except socket_error:
 */
        }

        /* "uvloop/pseudosock.pyx":54
 *             except socket_error:
 *                 pass
 *             try:             # <<<<<<<<<<<<<<
 *                 raddr = self.getpeername()
 *                 if raddr:
 */
      }
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L16_try_end;
      __pyx_L11_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/pseudosock.pyx":58
 *                 if raddr:
 *                     s += ", raddr=%s" % str(raddr)
 *             except socket_error:             # <<<<<<<<<<<<<<
 *                 pass
 *         s += '>'
 */
      __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_socket_error);
      if (__pyx_t_13) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L12_exception_handled;
      }
      goto __pyx_L13_except_error;

      /* "uvloop/pseudosock.pyx":54
 *             except socket_error:
 *                 pass
 *             try:             # <<<<<<<<<<<<<<
 *                 raddr = self.getpeername()
 *                 if raddr:
 */
      __pyx_L13_except_error:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      goto __pyx_L1_error;
      __pyx_L12_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      __pyx_L16_try_end:;
    }

    /* "uvloop/pseudosock.pyx":47
 *                                            self.type.name, self.proto)
 * 
 *         if self._fd != -1:             # <<<<<<<<<<<<<<
 *             try:
 *                 laddr = self.getsockname()
 */
  }

  /* "uvloop/pseudosock.pyx":60
 *             except socket_error:
 *                 pass
 *         s += '>'             # <<<<<<<<<<<<<<
 *         return s
 * 
 */
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_s, __pyx_kp_u__130); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_s, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":61
 *                 pass
 *         s += '>'
 *         return s             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_s);
  __pyx_r = __pyx_v_s;
  goto __pyx_L0;

  /* "uvloop/pseudosock.pyx":42
 *             return self._proto
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         s = ("<uvloop.PseudoSocket fd={}, family={!s}, "
 *              "type={!s}, proto={}").format(self.fileno(), self.family.name,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_laddr);
  __Pyx_XDECREF(__pyx_v_raddr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":63
 *         return s
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError("Cannot serialize socket object")
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_5__getstate__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_4__getstate__, "PseudoSocket.__getstate__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_5__getstate__ = {"__getstate__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_5__getstate__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_4__getstate__};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_5__getstate__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__getstate__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__getstate__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_4__getstate__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_4__getstate__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 1);

  /* "uvloop/pseudosock.pyx":64
 * 
 *     def __getstate__(self):
 *         raise TypeError("Cannot serialize socket object")             # <<<<<<<<<<<<<<
 * 
 *     def fileno(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__133, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(11, 64, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":63
 *         return s
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError("Cannot serialize socket object")
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":66
 *         raise TypeError("Cannot serialize socket object")
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         return self._fd
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_7fileno(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_6fileno, "PseudoSocket.fileno(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_7fileno = {"fileno", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_7fileno, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_6fileno};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_7fileno(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fileno (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("fileno", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "fileno", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_6fileno(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_6fileno(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fileno", 1);

  /* "uvloop/pseudosock.pyx":67
 * 
 *     def fileno(self):
 *         return self._fd             # <<<<<<<<<<<<<<
 * 
 *     def dup(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/pseudosock.pyx":66
 *         raise TypeError("Cannot serialize socket object")
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         return self._fd
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.fileno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":69
 *         return self._fd
 * 
 *     def dup(self):             # <<<<<<<<<<<<<<
 *         fd = os_dup(self._fd)
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_9dup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_8dup, "PseudoSocket.dup(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_9dup = {"dup", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_9dup, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_8dup};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_9dup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dup (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("dup", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "dup", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_8dup(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_8dup(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dup", 1);

  /* "uvloop/pseudosock.pyx":70
 * 
 *     def dup(self):
 *         fd = os_dup(self._fd)             # <<<<<<<<<<<<<<
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)
 *         sock.settimeout(0)
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_dup);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_os_dup; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_fd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":71
 *     def dup(self):
 *         fd = os_dup(self._fd)
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)             # <<<<<<<<<<<<<<
 *         sock.settimeout(0)
 *         return sock
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_family); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_proto); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(11, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(11, 71, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(11, 71, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(11, 71, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fileno, __pyx_v_fd) < 0) __PYX_ERR(11, 71, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_socket_socket, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sock = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/pseudosock.pyx":72
 *         fd = os_dup(self._fd)
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)
 *         sock.settimeout(0)             # <<<<<<<<<<<<<<
 *         return sock
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_settimeout); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_0};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/pseudosock.pyx":73
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)
 *         sock.settimeout(0)
 *         return sock             # <<<<<<<<<<<<<<
 * 
 *     def get_inheritable(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_sock);
  __pyx_r = __pyx_v_sock;
  goto __pyx_L0;

  /* "uvloop/pseudosock.pyx":69
 *         return self._fd
 * 
 *     def dup(self):             # <<<<<<<<<<<<<<
 *         fd = os_dup(self._fd)
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.dup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":75
 *         return sock
 * 
 *     def get_inheritable(self):             # <<<<<<<<<<<<<<
 *         return os_get_inheritable(self._fd)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_11get_inheritable(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_10get_inheritable, "PseudoSocket.get_inheritable(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_11get_inheritable = {"get_inheritable", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_11get_inheritable, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_10get_inheritable};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_11get_inheritable(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_inheritable (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_inheritable", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_inheritable", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_10get_inheritable(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_10get_inheritable(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_inheritable", 1);

  /* "uvloop/pseudosock.pyx":76
 * 
 *     def get_inheritable(self):
 *         return os_get_inheritable(self._fd)             # <<<<<<<<<<<<<<
 * 
 *     def set_inheritable(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_get_inheritable);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_os_get_inheritable; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/pseudosock.pyx":75
 *         return sock
 * 
 *     def get_inheritable(self):             # <<<<<<<<<<<<<<
 *         return os_get_inheritable(self._fd)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.get_inheritable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":78
 *         return os_get_inheritable(self._fd)
 * 
 *     def set_inheritable(self):             # <<<<<<<<<<<<<<
 *         os_set_inheritable(self._fd)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_13set_inheritable(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_12set_inheritable, "PseudoSocket.set_inheritable(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_13set_inheritable = {"set_inheritable", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_13set_inheritable, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_12set_inheritable};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_13set_inheritable(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_inheritable (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("set_inheritable", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "set_inheritable", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_12set_inheritable(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_12set_inheritable(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_inheritable", 1);

  /* "uvloop/pseudosock.pyx":79
 * 
 *     def set_inheritable(self):
 *         os_set_inheritable(self._fd)             # <<<<<<<<<<<<<<
 * 
 *     def ioctl(self, *args, **kwargs):
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":78
 *         return os_get_inheritable(self._fd)
 * 
 *     def set_inheritable(self):             # <<<<<<<<<<<<<<
 *         os_set_inheritable(self._fd)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.set_inheritable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":81
 *         os_set_inheritable(self._fd)
 * 
 *     def ioctl(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_15ioctl(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_14ioctl, "PseudoSocket.ioctl(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_15ioctl = {"ioctl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_15ioctl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_14ioctl};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_15ioctl(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ioctl (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "ioctl", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_14ioctl(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_14ioctl(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ioctl", 1);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":84
 *         pass
 * 
 *     def getsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_17getsockopt(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_16getsockopt, "PseudoSocket.getsockopt(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_17getsockopt = {"getsockopt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_17getsockopt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_16getsockopt};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_17getsockopt(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getsockopt (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "getsockopt", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_16getsockopt(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_16getsockopt(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  unsigned int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getsockopt", 1);

  /* "uvloop/pseudosock.pyx":85
 * 
 *     def getsockopt(self, *args, **kwargs):
 *         sock = self._make_sock()             # <<<<<<<<<<<<<<
 *         try:
 *             return sock.getsockopt(*args, **kwargs)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_make_sock(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sock = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":86
 *     def getsockopt(self, *args, **kwargs):
 *         sock = self._make_sock()
 *         try:             # <<<<<<<<<<<<<<
 *             return sock.getsockopt(*args, **kwargs)
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/pseudosock.pyx":87
 *         sock = self._make_sock()
 *         try:
 *             return sock.getsockopt(*args, **kwargs)             # <<<<<<<<<<<<<<
 *         finally:
 *             sock.detach()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_getsockopt); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 87, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 87, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 87, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_return;
  }

  /* "uvloop/pseudosock.pyx":89
 *             return sock.getsockopt(*args, **kwargs)
 *         finally:
 *             sock.detach()             # <<<<<<<<<<<<<<
 * 
 *     def setsockopt(self, *args, **kwargs):
 */
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 89, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        __pyx_t_13 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_13 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_13, 0+__pyx_t_13);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 89, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
      __pyx_L7_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_12 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 89, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_13 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_13 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_13, 0+__pyx_t_13);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 89, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_12;
      __pyx_t_12 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/pseudosock.pyx":84
 *         pass
 * 
 *     def getsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.getsockopt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":91
 *             sock.detach()
 * 
 *     def setsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_19setsockopt(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_18setsockopt, "PseudoSocket.setsockopt(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_19setsockopt = {"setsockopt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_19setsockopt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_18setsockopt};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_19setsockopt(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setsockopt (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "setsockopt", 1))) return NULL;
  if (unlikely(__pyx_kwds)) {
    __pyx_v_kwargs = __Pyx_KwargsAsDict_VARARGS(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_18setsockopt(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_18setsockopt(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  unsigned int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setsockopt", 1);

  /* "uvloop/pseudosock.pyx":92
 * 
 *     def setsockopt(self, *args, **kwargs):
 *         sock = self._make_sock()             # <<<<<<<<<<<<<<
 *         try:
 *             return sock.setsockopt(*args, **kwargs)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_make_sock(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sock = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":93
 *     def setsockopt(self, *args, **kwargs):
 *         sock = self._make_sock()
 *         try:             # <<<<<<<<<<<<<<
 *             return sock.setsockopt(*args, **kwargs)
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/pseudosock.pyx":94
 *         sock = self._make_sock()
 *         try:
 *             return sock.setsockopt(*args, **kwargs)             # <<<<<<<<<<<<<<
 *         finally:
 *             sock.detach()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_setsockopt); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 94, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 94, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 94, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_return;
  }

  /* "uvloop/pseudosock.pyx":96
 *             return sock.setsockopt(*args, **kwargs)
 *         finally:
 *             sock.detach()             # <<<<<<<<<<<<<<
 * 
 *     def getpeername(self):
 */
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 96, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = NULL;
        __pyx_t_13 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_13 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_13, 0+__pyx_t_13);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 96, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
      __pyx_L7_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_12 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_13 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_13 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_13, 0+__pyx_t_13);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 96, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_12;
      __pyx_t_12 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/pseudosock.pyx":91
 *             sock.detach()
 * 
 *     def setsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.setsockopt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":98
 *             sock.detach()
 * 
 *     def getpeername(self):             # <<<<<<<<<<<<<<
 *         if self._peername is not None:
 *             return self._peername
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_21getpeername(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_20getpeername, "PseudoSocket.getpeername(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_21getpeername = {"getpeername", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_21getpeername, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_20getpeername};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_21getpeername(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getpeername (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("getpeername", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "getpeername", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_20getpeername(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_20getpeername(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getpeername", 1);

  /* "uvloop/pseudosock.pyx":99
 * 
 *     def getpeername(self):
 *         if self._peername is not None:             # <<<<<<<<<<<<<<
 *             return self._peername
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_peername != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/pseudosock.pyx":100
 *     def getpeername(self):
 *         if self._peername is not None:
 *             return self._peername             # <<<<<<<<<<<<<<
 * 
 *         sock = self._make_sock()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_peername);
    __pyx_r = __pyx_v_self->_peername;
    goto __pyx_L0;

    /* "uvloop/pseudosock.pyx":99
 * 
 *     def getpeername(self):
 *         if self._peername is not None:             # <<<<<<<<<<<<<<
 *             return self._peername
 * 
 */
  }

  /* "uvloop/pseudosock.pyx":102
 *             return self._peername
 * 
 *         sock = self._make_sock()             # <<<<<<<<<<<<<<
 *         try:
 *             self._peername = sock.getpeername()
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_make_sock(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sock = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/pseudosock.pyx":103
 * 
 *         sock = self._make_sock()
 *         try:             # <<<<<<<<<<<<<<
 *             self._peername = sock.getpeername()
 *             return self._peername
 */
  /*try:*/ {

    /* "uvloop/pseudosock.pyx":104
 *         sock = self._make_sock()
 *         try:
 *             self._peername = sock.getpeername()             # <<<<<<<<<<<<<<
 *             return self._peername
 *         finally:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_getpeername); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 104, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 104, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_peername);
    __Pyx_DECREF(__pyx_v_self->_peername);
    __pyx_v_self->_peername = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/pseudosock.pyx":105
 *         try:
 *             self._peername = sock.getpeername()
 *             return self._peername             # <<<<<<<<<<<<<<
 *         finally:
 *             sock.detach()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_peername);
    __pyx_r = __pyx_v_self->_peername;
    goto __pyx_L4_return;
  }

  /* "uvloop/pseudosock.pyx":107
 *             return self._peername
 *         finally:
 *             sock.detach()             # <<<<<<<<<<<<<<
 * 
 *     def getsockname(self):
 */
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 107, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 107, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
      __pyx_L8_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_14 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 107, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_14;
      __pyx_t_14 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/pseudosock.pyx":98
 *             sock.detach()
 * 
 *     def getpeername(self):             # <<<<<<<<<<<<<<
 *         if self._peername is not None:
 *             return self._peername
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.getpeername", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":109
 *             sock.detach()
 * 
 *     def getsockname(self):             # <<<<<<<<<<<<<<
 *         if self._sockname is not None:
 *             return self._sockname
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_23getsockname(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_22getsockname, "PseudoSocket.getsockname(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_23getsockname = {"getsockname", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_23getsockname, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_22getsockname};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_23getsockname(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getsockname (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("getsockname", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "getsockname", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_22getsockname(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_22getsockname(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("getsockname", 1);

  /* "uvloop/pseudosock.pyx":110
 * 
 *     def getsockname(self):
 *         if self._sockname is not None:             # <<<<<<<<<<<<<<
 *             return self._sockname
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_sockname != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/pseudosock.pyx":111
 *     def getsockname(self):
 *         if self._sockname is not None:
 *             return self._sockname             # <<<<<<<<<<<<<<
 * 
 *         sock = self._make_sock()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_sockname);
    __pyx_r = __pyx_v_self->_sockname;
    goto __pyx_L0;

    /* "uvloop/pseudosock.pyx":110
 * 
 *     def getsockname(self):
 *         if self._sockname is not None:             # <<<<<<<<<<<<<<
 *             return self._sockname
 * 
 */
  }

  /* "uvloop/pseudosock.pyx":113
 *             return self._sockname
 * 
 *         sock = self._make_sock()             # <<<<<<<<<<<<<<
 *         try:
 *             self._sockname = sock.getsockname()
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_make_sock(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sock = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/pseudosock.pyx":114
 * 
 *         sock = self._make_sock()
 *         try:             # <<<<<<<<<<<<<<
 *             self._sockname = sock.getsockname()
 *             return self._sockname
 */
  /*try:*/ {

    /* "uvloop/pseudosock.pyx":115
 *         sock = self._make_sock()
 *         try:
 *             self._sockname = sock.getsockname()             # <<<<<<<<<<<<<<
 *             return self._sockname
 *         finally:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_getsockname); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 115, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 115, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_sockname);
    __Pyx_DECREF(__pyx_v_self->_sockname);
    __pyx_v_self->_sockname = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/pseudosock.pyx":116
 *         try:
 *             self._sockname = sock.getsockname()
 *             return self._sockname             # <<<<<<<<<<<<<<
 *         finally:
 *             sock.detach()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_sockname);
    __pyx_r = __pyx_v_self->_sockname;
    goto __pyx_L4_return;
  }

  /* "uvloop/pseudosock.pyx":118
 *             return self._sockname
 *         finally:
 *             sock.detach()             # <<<<<<<<<<<<<<
 * 
 *     def share(self, process_id):
 */
  /*finally:*/ {
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 118, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 118, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
      __pyx_L8_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_14 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 118, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_14;
      __pyx_t_14 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/pseudosock.pyx":109
 *             sock.detach()
 * 
 *     def getsockname(self):             # <<<<<<<<<<<<<<
 *         if self._sockname is not None:
 *             return self._sockname
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.getsockname", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":120
 *             sock.detach()
 * 
 *     def share(self, process_id):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_25share(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_24share, "PseudoSocket.share(self, process_id)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_25share = {"share", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_25share, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_24share};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_25share(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_process_id = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("share (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_process_id,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_process_id)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 120, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "share") < 0)) __PYX_ERR(11, 120, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_process_id = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("share", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 120, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.share", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_24share(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_process_id);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_24share(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_process_id) {
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("share", 1);

  /* "uvloop/pseudosock.pyx":121
 * 
 *     def share(self, process_id):
 *         sock = self._make_sock()             # <<<<<<<<<<<<<<
 *         try:
 *             return sock.share(process_id)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_make_sock(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sock = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":122
 *     def share(self, process_id):
 *         sock = self._make_sock()
 *         try:             # <<<<<<<<<<<<<<
 *             return sock.share(process_id)
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/pseudosock.pyx":123
 *         sock = self._make_sock()
 *         try:
 *             return sock.share(process_id)             # <<<<<<<<<<<<<<
 *         finally:
 *             sock.detach()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_share); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 123, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_process_id};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 123, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L3_return;
  }

  /* "uvloop/pseudosock.pyx":125
 *             return sock.share(process_id)
 *         finally:
 *             sock.detach()             # <<<<<<<<<<<<<<
 * 
 *     def accept(self):
 */
  /*finally:*/ {
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 125, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_3 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 125, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
      __pyx_L7_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_13 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_detach); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;
    }
  }

  /* "uvloop/pseudosock.pyx":120
 *             sock.detach()
 * 
 *     def share(self, process_id):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.share", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":127
 *             sock.detach()
 * 
 *     def accept(self):             # <<<<<<<<<<<<<<
 *         self._na('accept() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_27accept(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_26accept, "PseudoSocket.accept(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_27accept = {"accept", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_27accept, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_26accept};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_27accept(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("accept", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "accept", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_26accept(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_26accept(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("accept", 1);

  /* "uvloop/pseudosock.pyx":128
 * 
 *     def accept(self):
 *         self._na('accept() method')             # <<<<<<<<<<<<<<
 * 
 *     def connect(self, *args):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_accept_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":127
 *             sock.detach()
 * 
 *     def accept(self):             # <<<<<<<<<<<<<<
 *         self._na('accept() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":130
 *         self._na('accept() method')
 * 
 *     def connect(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_29connect(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_28connect, "PseudoSocket.connect(self, *args)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_29connect = {"connect", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_29connect, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_28connect};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_29connect(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "connect", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_28connect(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_28connect(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect", 1);

  /* "uvloop/pseudosock.pyx":131
 * 
 *     def connect(self, *args):
 *         self._na('connect() method')             # <<<<<<<<<<<<<<
 * 
 *     def connect_ex(self, *args):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_connect_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":130
 *         self._na('accept() method')
 * 
 *     def connect(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":133
 *         self._na('connect() method')
 * 
 *     def connect_ex(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect_ex() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_31connect_ex(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_30connect_ex, "PseudoSocket.connect_ex(self, *args)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_31connect_ex = {"connect_ex", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_31connect_ex, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_30connect_ex};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_31connect_ex(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect_ex (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "connect_ex", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_30connect_ex(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_30connect_ex(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect_ex", 1);

  /* "uvloop/pseudosock.pyx":134
 * 
 *     def connect_ex(self, *args):
 *         self._na('connect_ex() method')             # <<<<<<<<<<<<<<
 * 
 *     def bind(self, *args):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_connect_ex_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":133
 *         self._na('connect() method')
 * 
 *     def connect_ex(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect_ex() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.connect_ex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":136
 *         self._na('connect_ex() method')
 * 
 *     def bind(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('bind() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_33bind(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_32bind, "PseudoSocket.bind(self, *args)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_33bind = {"bind", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_33bind, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_32bind};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_33bind(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bind (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "bind", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_32bind(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_32bind(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind", 1);

  /* "uvloop/pseudosock.pyx":137
 * 
 *     def bind(self, *args):
 *         self._na('bind() method')             # <<<<<<<<<<<<<<
 * 
 *     def listen(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_bind_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":136
 *         self._na('connect_ex() method')
 * 
 *     def bind(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('bind() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":139
 *         self._na('bind() method')
 * 
 *     def listen(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('listen() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_35listen(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_34listen, "PseudoSocket.listen(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_35listen = {"listen", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_35listen, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_34listen};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_35listen(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("listen (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "listen", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_34listen(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_34listen(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("listen", 1);

  /* "uvloop/pseudosock.pyx":140
 * 
 *     def listen(self, *args, **kwargs):
 *         self._na('listen() method')             # <<<<<<<<<<<<<<
 * 
 *     def makefile(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_listen_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":139
 *         self._na('bind() method')
 * 
 *     def listen(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('listen() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.listen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":142
 *         self._na('listen() method')
 * 
 *     def makefile(self):             # <<<<<<<<<<<<<<
 *         self._na('makefile() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_37makefile(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_36makefile, "PseudoSocket.makefile(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_37makefile = {"makefile", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_37makefile, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_36makefile};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_37makefile(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("makefile (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("makefile", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "makefile", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_36makefile(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_36makefile(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("makefile", 1);

  /* "uvloop/pseudosock.pyx":143
 * 
 *     def makefile(self):
 *         self._na('makefile() method')             # <<<<<<<<<<<<<<
 * 
 *     def sendfile(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_makefile_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":142
 *         self._na('listen() method')
 * 
 *     def makefile(self):             # <<<<<<<<<<<<<<
 *         self._na('makefile() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.makefile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":145
 *         self._na('makefile() method')
 * 
 *     def sendfile(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendfile() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_39sendfile(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_38sendfile, "PseudoSocket.sendfile(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_39sendfile = {"sendfile", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_39sendfile, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_38sendfile};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_39sendfile(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sendfile (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sendfile", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_38sendfile(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_38sendfile(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sendfile", 1);

  /* "uvloop/pseudosock.pyx":146
 * 
 *     def sendfile(self, *args, **kwargs):
 *         self._na('sendfile() method')             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_sendfile_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":145
 *         self._na('makefile() method')
 * 
 *     def sendfile(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendfile() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.sendfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":148
 *         self._na('sendfile() method')
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._na('close() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_41close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_40close, "PseudoSocket.close(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_41close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_41close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_40close};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_41close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_40close(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_40close(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/pseudosock.pyx":149
 * 
 *     def close(self):
 *         self._na('close() method')             # <<<<<<<<<<<<<<
 * 
 *     def detach(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_close_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":148
 *         self._na('sendfile() method')
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._na('close() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":151
 *         self._na('close() method')
 * 
 *     def detach(self):             # <<<<<<<<<<<<<<
 *         self._na('detach() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_43detach(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_42detach, "PseudoSocket.detach(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_43detach = {"detach", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_43detach, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_42detach};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_43detach(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("detach (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("detach", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "detach", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_42detach(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_42detach(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("detach", 1);

  /* "uvloop/pseudosock.pyx":152
 * 
 *     def detach(self):
 *         self._na('detach() method')             # <<<<<<<<<<<<<<
 * 
 *     def shutdown(self, *args):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_detach_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":151
 *         self._na('close() method')
 * 
 *     def detach(self):             # <<<<<<<<<<<<<<
 *         self._na('detach() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.detach", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":154
 *         self._na('detach() method')
 * 
 *     def shutdown(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('shutdown() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_45shutdown(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_44shutdown, "PseudoSocket.shutdown(self, *args)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_45shutdown = {"shutdown", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_45shutdown, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_44shutdown};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_45shutdown(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shutdown (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "shutdown", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_44shutdown(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_44shutdown(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shutdown", 1);

  /* "uvloop/pseudosock.pyx":155
 * 
 *     def shutdown(self, *args):
 *         self._na('shutdown() method')             # <<<<<<<<<<<<<<
 * 
 *     def sendmsg_afalg(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_shutdown_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":154
 *         self._na('detach() method')
 * 
 *     def shutdown(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('shutdown() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":157
 *         self._na('shutdown() method')
 * 
 *     def sendmsg_afalg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg_afalg() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_47sendmsg_afalg(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_46sendmsg_afalg, "PseudoSocket.sendmsg_afalg(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_47sendmsg_afalg = {"sendmsg_afalg", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_47sendmsg_afalg, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_46sendmsg_afalg};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_47sendmsg_afalg(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sendmsg_afalg (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sendmsg_afalg", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_46sendmsg_afalg(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_46sendmsg_afalg(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sendmsg_afalg", 1);

  /* "uvloop/pseudosock.pyx":158
 * 
 *     def sendmsg_afalg(self, *args, **kwargs):
 *         self._na('sendmsg_afalg() method')             # <<<<<<<<<<<<<<
 * 
 *     def sendmsg(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_sendmsg_afalg_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":157
 *         self._na('shutdown() method')
 * 
 *     def sendmsg_afalg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg_afalg() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.sendmsg_afalg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":160
 *         self._na('sendmsg_afalg() method')
 * 
 *     def sendmsg(self):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_49sendmsg(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_48sendmsg, "PseudoSocket.sendmsg(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_49sendmsg = {"sendmsg", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_49sendmsg, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_48sendmsg};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_49sendmsg(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sendmsg (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("sendmsg", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sendmsg", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_48sendmsg(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_48sendmsg(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sendmsg", 1);

  /* "uvloop/pseudosock.pyx":161
 * 
 *     def sendmsg(self):
 *         self._na('sendmsg() method')             # <<<<<<<<<<<<<<
 * 
 *     def sendto(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_sendmsg_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":160
 *         self._na('sendmsg_afalg() method')
 * 
 *     def sendmsg(self):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.sendmsg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":163
 *         self._na('sendmsg() method')
 * 
 *     def sendto(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendto() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_51sendto(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_50sendto, "PseudoSocket.sendto(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_51sendto = {"sendto", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_51sendto, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_50sendto};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_51sendto(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sendto (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sendto", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_50sendto(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_50sendto(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sendto", 1);

  /* "uvloop/pseudosock.pyx":164
 * 
 *     def sendto(self, *args, **kwargs):
 *         self._na('sendto() method')             # <<<<<<<<<<<<<<
 * 
 *     def send(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_sendto_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":163
 *         self._na('sendmsg() method')
 * 
 *     def sendto(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendto() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.sendto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":166
 *         self._na('sendto() method')
 * 
 *     def send(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('send() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_53send(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_52send, "PseudoSocket.send(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_53send = {"send", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_53send, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_52send};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_53send(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("send (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "send", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_52send(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_52send(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("send", 1);

  /* "uvloop/pseudosock.pyx":167
 * 
 *     def send(self, *args, **kwargs):
 *         self._na('send() method')             # <<<<<<<<<<<<<<
 * 
 *     def sendall(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_send_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":166
 *         self._na('sendto() method')
 * 
 *     def send(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('send() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":169
 *         self._na('send() method')
 * 
 *     def sendall(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendall() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_55sendall(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_54sendall, "PseudoSocket.sendall(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_55sendall = {"sendall", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_55sendall, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_54sendall};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_55sendall(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sendall (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "sendall", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_54sendall(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_54sendall(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sendall", 1);

  /* "uvloop/pseudosock.pyx":170
 * 
 *     def sendall(self, *args, **kwargs):
 *         self._na('sendall() method')             # <<<<<<<<<<<<<<
 * 
 *     def recv_into(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_sendall_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":169
 *         self._na('send() method')
 * 
 *     def sendall(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendall() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.sendall", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":172
 *         self._na('sendall() method')
 * 
 *     def recv_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv_into() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_57recv_into(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_56recv_into, "PseudoSocket.recv_into(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_57recv_into = {"recv_into", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_57recv_into, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_56recv_into};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_57recv_into(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recv_into (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "recv_into", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_56recv_into(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_56recv_into(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recv_into", 1);

  /* "uvloop/pseudosock.pyx":173
 * 
 *     def recv_into(self, *args, **kwargs):
 *         self._na('recv_into() method')             # <<<<<<<<<<<<<<
 * 
 *     def recvfrom_into(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_recv_into_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":172
 *         self._na('sendall() method')
 * 
 *     def recv_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv_into() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.recv_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":175
 *         self._na('recv_into() method')
 * 
 *     def recvfrom_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom_into() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_59recvfrom_into(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_58recvfrom_into, "PseudoSocket.recvfrom_into(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_59recvfrom_into = {"recvfrom_into", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_59recvfrom_into, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_58recvfrom_into};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_59recvfrom_into(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recvfrom_into (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "recvfrom_into", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_58recvfrom_into(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_58recvfrom_into(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recvfrom_into", 1);

  /* "uvloop/pseudosock.pyx":176
 * 
 *     def recvfrom_into(self, *args, **kwargs):
 *         self._na('recvfrom_into() method')             # <<<<<<<<<<<<<<
 * 
 *     def recvmsg_into(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_recvfrom_into_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":175
 *         self._na('recv_into() method')
 * 
 *     def recvfrom_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom_into() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.recvfrom_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":178
 *         self._na('recvfrom_into() method')
 * 
 *     def recvmsg_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg_into() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_61recvmsg_into(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_60recvmsg_into, "PseudoSocket.recvmsg_into(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_61recvmsg_into = {"recvmsg_into", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_61recvmsg_into, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_60recvmsg_into};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_61recvmsg_into(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recvmsg_into (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "recvmsg_into", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_60recvmsg_into(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_60recvmsg_into(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recvmsg_into", 1);

  /* "uvloop/pseudosock.pyx":179
 * 
 *     def recvmsg_into(self, *args, **kwargs):
 *         self._na('recvmsg_into() method')             # <<<<<<<<<<<<<<
 * 
 *     def recvmsg(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_recvmsg_into_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":178
 *         self._na('recvfrom_into() method')
 * 
 *     def recvmsg_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg_into() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.recvmsg_into", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":181
 *         self._na('recvmsg_into() method')
 * 
 *     def recvmsg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_63recvmsg(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_62recvmsg, "PseudoSocket.recvmsg(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_63recvmsg = {"recvmsg", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_63recvmsg, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_62recvmsg};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_63recvmsg(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recvmsg (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "recvmsg", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_62recvmsg(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_62recvmsg(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recvmsg", 1);

  /* "uvloop/pseudosock.pyx":182
 * 
 *     def recvmsg(self, *args, **kwargs):
 *         self._na('recvmsg() method')             # <<<<<<<<<<<<<<
 * 
 *     def recvfrom(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_recvmsg_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":181
 *         self._na('recvmsg_into() method')
 * 
 *     def recvmsg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.recvmsg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":184
 *         self._na('recvmsg() method')
 * 
 *     def recvfrom(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_65recvfrom(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_64recvfrom, "PseudoSocket.recvfrom(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_65recvfrom = {"recvfrom", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_65recvfrom, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_64recvfrom};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_65recvfrom(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recvfrom (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "recvfrom", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_64recvfrom(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_64recvfrom(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recvfrom", 1);

  /* "uvloop/pseudosock.pyx":185
 * 
 *     def recvfrom(self, *args, **kwargs):
 *         self._na('recvfrom() method')             # <<<<<<<<<<<<<<
 * 
 *     def recv(self, *args, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_recvfrom_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":184
 *         self._na('recvmsg() method')
 * 
 *     def recvfrom(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.recvfrom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":187
 *         self._na('recvfrom() method')
 * 
 *     def recv(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv() method')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_67recv(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_66recv, "PseudoSocket.recv(self, *args, **kwargs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_67recv = {"recv", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_67recv, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_66recv};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_67recv(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recv (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "recv", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_66recv(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_66recv(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recv", 1);

  /* "uvloop/pseudosock.pyx":188
 * 
 *     def recv(self, *args, **kwargs):
 *         self._na('recv() method')             # <<<<<<<<<<<<<<
 * 
 *     def settimeout(self, value):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_recv_method); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":187
 *         self._na('recvfrom() method')
 * 
 *     def recv(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv() method')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":190
 *         self._na('recv() method')
 * 
 *     def settimeout(self, value):             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_69settimeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_68settimeout, "PseudoSocket.settimeout(self, value)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_69settimeout = {"settimeout", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_69settimeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_68settimeout};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_69settimeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("settimeout (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 190, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "settimeout") < 0)) __PYX_ERR(11, 190, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("settimeout", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 190, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.settimeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_68settimeout(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_68settimeout(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("settimeout", 1);

  /* "uvloop/pseudosock.pyx":191
 * 
 *     def settimeout(self, value):
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return
 *         raise ValueError(
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_value, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 191, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/pseudosock.pyx":192
 *     def settimeout(self, value):
 *         if value == 0:
 *             return             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             'settimeout(): only 0 timeout is allowed on transport sockets')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/pseudosock.pyx":191
 * 
 *     def settimeout(self, value):
 *         if value == 0:             # <<<<<<<<<<<<<<
 *             return
 *         raise ValueError(
 */
  }

  /* "uvloop/pseudosock.pyx":193
 *         if value == 0:
 *             return
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             'settimeout(): only 0 timeout is allowed on transport sockets')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__134, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(11, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(11, 193, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":190
 *         self._na('recv() method')
 * 
 *     def settimeout(self, value):             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             return
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.settimeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":196
 *             'settimeout(): only 0 timeout is allowed on transport sockets')
 * 
 *     def gettimeout(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_71gettimeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_70gettimeout, "PseudoSocket.gettimeout(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_71gettimeout = {"gettimeout", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_71gettimeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_70gettimeout};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_71gettimeout(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gettimeout (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("gettimeout", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "gettimeout", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_70gettimeout(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_70gettimeout(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gettimeout", 1);

  /* "uvloop/pseudosock.pyx":197
 * 
 *     def gettimeout(self):
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def setblocking(self, flag):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "uvloop/pseudosock.pyx":196
 *             'settimeout(): only 0 timeout is allowed on transport sockets')
 * 
 *     def gettimeout(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":199
 *         return 0
 * 
 *     def setblocking(self, flag):             # <<<<<<<<<<<<<<
 *         if not flag:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_73setblocking(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_72setblocking, "PseudoSocket.setblocking(self, flag)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_73setblocking = {"setblocking", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_73setblocking, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_72setblocking};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_73setblocking(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_flag = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setblocking (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_flag,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_flag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(11, 199, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "setblocking") < 0)) __PYX_ERR(11, 199, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_flag = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setblocking", 1, 1, 1, __pyx_nargs); __PYX_ERR(11, 199, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.setblocking", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_72setblocking(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_flag);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_72setblocking(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v_flag) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setblocking", 1);

  /* "uvloop/pseudosock.pyx":200
 * 
 *     def setblocking(self, flag):
 *         if not flag:             # <<<<<<<<<<<<<<
 *             return
 *         raise ValueError(
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_flag); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(11, 200, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/pseudosock.pyx":201
 *     def setblocking(self, flag):
 *         if not flag:
 *             return             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             'setblocking(): transport sockets cannot be blocking')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/pseudosock.pyx":200
 * 
 *     def setblocking(self, flag):
 *         if not flag:             # <<<<<<<<<<<<<<
 *             return
 *         raise ValueError(
 */
  }

  /* "uvloop/pseudosock.pyx":202
 *         if not flag:
 *             return
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             'setblocking(): transport sockets cannot be blocking')
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__135, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(11, 202, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":199
 *         return 0
 * 
 *     def setblocking(self, flag):             # <<<<<<<<<<<<<<
 *         if not flag:
 *             return
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.setblocking", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":205
 *             'setblocking(): transport sockets cannot be blocking')
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_75__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_74__enter__, "PseudoSocket.__enter__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_75__enter__ = {"__enter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_75__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_74__enter__};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_75__enter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__enter__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__enter__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_74__enter__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_74__enter__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__enter__", 1);

  /* "uvloop/pseudosock.pyx":206
 * 
 *     def __enter__(self):
 *         self._na('context manager protocol')             # <<<<<<<<<<<<<<
 * 
 *     def __exit__(self, *err):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_context_manager_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":205
 *             'setblocking(): transport sockets cannot be blocking')
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__enter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/pseudosock.pyx":208
 *         self._na('context manager protocol')
 * 
 *     def __exit__(self, *err):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_77__exit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_76__exit__, "PseudoSocket.__exit__(self, *err)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_77__exit__ = {"__exit__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_77__exit__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_76__exit__};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_77__exit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_err = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__exit__", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_err = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_76__exit__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v_err);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_err);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_76__exit__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_err) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__exit__", 1);

  /* "uvloop/pseudosock.pyx":209
 * 
 *     def __exit__(self, *err):
 *         self._na('context manager protocol')             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket *)__pyx_v_self->__pyx_vtab)->_na(__pyx_v_self, __pyx_kp_u_context_manager_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/pseudosock.pyx":208
 *         self._na('context manager protocol')
 * 
 *     def __exit__(self, *err):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_79__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_78__reduce_cython__, "PseudoSocket.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_79__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_79__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_78__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_79__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_78__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_78__reduce_cython__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._family, self._fd, self._peername, self._proto, self._sockname, self._type)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_family); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_proto); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_peername);
  __Pyx_GIVEREF(__pyx_v_self->_peername);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_self->_peername)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_sockname);
  __Pyx_GIVEREF(__pyx_v_self->_sockname);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->_sockname)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_4)) __PYX_ERR(6, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._family, self._fd, self._peername, self._proto, self._sockname, self._type)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self._family, self._fd, self._peername, self._proto, self._sockname, self._type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(6, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._peername is not None or self._sockname is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._family, self._fd, self._peername, self._proto, self._sockname, self._type)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._peername is not None or self._sockname is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, None), state
 */
  /*else*/ {
    __pyx_t_7 = (__pyx_v_self->_peername != Py_None);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (__pyx_v_self->_sockname != Py_None);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._peername is not None or self._sockname is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._peername is not None or self._sockname is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_PseudoSocket); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_200977156);
    __Pyx_GIVEREF(__pyx_int_200977156);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_200977156)) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(6, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(6, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._peername is not None or self._sockname is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, None), state
 *     else:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PseudoSocket__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PseudoSocket); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(6, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_200977156);
    __Pyx_GIVEREF(__pyx_int_200977156);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_200977156)) __PYX_ERR(6, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(6, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(6, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(6, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PseudoSocket__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_81__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12PseudoSocket_80__setstate_cython__, "PseudoSocket.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12PseudoSocket_81__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_81__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_80__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_12PseudoSocket_81__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12PseudoSocket_80__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12PseudoSocket_80__setstate_cython__(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PseudoSocket__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pyx_unpickle_PseudoSocket__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PseudoSocket__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.PseudoSocket.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":31
 *     # beginning of it.
 * 
 *     def __init__(self, *, maxsize):             # <<<<<<<<<<<<<<
 *         if maxsize <= 0:
 *             raise ValueError(
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8LruCache_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_8LruCache_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_maxsize = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_maxsize,0};
    if (likely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_maxsize)) != 0)) {
        (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
        kw_args--;
      }
      else if (unlikely(PyErr_Occurred())) __PYX_ERR(12, 31, __pyx_L3_error)
      else {
        __Pyx_RaiseKeywordRequired("__init__", __pyx_n_s_maxsize); __PYX_ERR(12, 31, __pyx_L3_error)
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, 0, "__init__") < 0)) __PYX_ERR(12, 31, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 0)) {
      goto __pyx_L5_argtuple_error;
    } else {
      __Pyx_RaiseKeywordRequired("__init__", __pyx_n_s_maxsize); __PYX_ERR(12, 31, __pyx_L3_error)
    }
    __pyx_v_maxsize = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); __PYX_ERR(12, 31, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.LruCache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache___init__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self), __pyx_v_maxsize);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8LruCache___init__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_maxsize) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/lru.pyx":32
 * 
 *     def __init__(self, *, maxsize):
 *         if maxsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f'maxsize is expected to be greater than 0, got {maxsize}')
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_maxsize, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 32, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(12, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/lru.pyx":34
 *         if maxsize <= 0:
 *             raise ValueError(
 *                 f'maxsize is expected to be greater than 0, got {maxsize}')             # <<<<<<<<<<<<<<
 * 
 *         self._dict = col_OrderedDict()
 */
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_maxsize, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_maxsize_is_expected_to_be_greate, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/lru.pyx":33
 *     def __init__(self, *, maxsize):
 *         if maxsize <= 0:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 f'maxsize is expected to be greater than 0, got {maxsize}')
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(12, 33, __pyx_L1_error)

    /* "uvloop/lru.pyx":32
 * 
 *     def __init__(self, *, maxsize):
 *         if maxsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f'maxsize is expected to be greater than 0, got {maxsize}')
 */
  }

  /* "uvloop/lru.pyx":36
 *                 f'maxsize is expected to be greater than 0, got {maxsize}')
 * 
 *         self._dict = col_OrderedDict()             # <<<<<<<<<<<<<<
 *         self._dict_move_to_end = self._dict.move_to_end
 *         self._dict_get = self._dict.get
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_col_OrderedDict);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_col_OrderedDict; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_dict);
  __Pyx_DECREF(__pyx_v_self->_dict);
  __pyx_v_self->_dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":37
 * 
 *         self._dict = col_OrderedDict()
 *         self._dict_move_to_end = self._dict.move_to_end             # <<<<<<<<<<<<<<
 *         self._dict_get = self._dict.get
 *         self._maxsize = maxsize
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_dict, __pyx_n_s_move_to_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_dict_move_to_end);
  __Pyx_DECREF(__pyx_v_self->_dict_move_to_end);
  __pyx_v_self->_dict_move_to_end = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":38
 *         self._dict = col_OrderedDict()
 *         self._dict_move_to_end = self._dict.move_to_end
 *         self._dict_get = self._dict.get             # <<<<<<<<<<<<<<
 *         self._maxsize = maxsize
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_dict, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_dict_get);
  __Pyx_DECREF(__pyx_v_self->_dict_get);
  __pyx_v_self->_dict_get = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":39
 *         self._dict_move_to_end = self._dict.move_to_end
 *         self._dict_get = self._dict.get
 *         self._maxsize = maxsize             # <<<<<<<<<<<<<<
 * 
 *     cdef get(self, key, default):
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_maxsize); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(12, 39, __pyx_L1_error)
  __pyx_v_self->_maxsize = __pyx_t_6;

  /* "uvloop/lru.pyx":31
 *     # beginning of it.
 * 
 *     def __init__(self, *, maxsize):             # <<<<<<<<<<<<<<
 *         if maxsize <= 0:
 *             raise ValueError(
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.LruCache.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":41
 *         self._maxsize = maxsize
 * 
 *     cdef get(self, key, default):             # <<<<<<<<<<<<<<
 *         o = self._dict_get(key, _LRU_MARKER)
 *         if o is _LRU_MARKER:
 */

static PyObject *__pyx_f_6uvloop_4loop_8LruCache_get(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default) {
  PyObject *__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 1);

  /* "uvloop/lru.pyx":42
 * 
 *     cdef get(self, key, default):
 *         o = self._dict_get(key, _LRU_MARKER)             # <<<<<<<<<<<<<<
 *         if o is _LRU_MARKER:
 *             return default
 */
  __Pyx_INCREF(__pyx_v_self->_dict_get);
  __pyx_t_2 = __pyx_v_self->_dict_get; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_key, __pyx_v_6uvloop_4loop__LRU_MARKER};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":43
 *     cdef get(self, key, default):
 *         o = self._dict_get(key, _LRU_MARKER)
 *         if o is _LRU_MARKER:             # <<<<<<<<<<<<<<
 *             return default
 *         self._dict_move_to_end(key)  # last=True
 */
  __pyx_t_5 = (__pyx_v_o == __pyx_v_6uvloop_4loop__LRU_MARKER);
  if (__pyx_t_5) {

    /* "uvloop/lru.pyx":44
 *         o = self._dict_get(key, _LRU_MARKER)
 *         if o is _LRU_MARKER:
 *             return default             # <<<<<<<<<<<<<<
 *         self._dict_move_to_end(key)  # last=True
 *         return o
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_default);
    __pyx_r = __pyx_v_default;
    goto __pyx_L0;

    /* "uvloop/lru.pyx":43
 *     cdef get(self, key, default):
 *         o = self._dict_get(key, _LRU_MARKER)
 *         if o is _LRU_MARKER:             # <<<<<<<<<<<<<<
 *             return default
 *         self._dict_move_to_end(key)  # last=True
 */
  }

  /* "uvloop/lru.pyx":45
 *         if o is _LRU_MARKER:
 *             return default
 *         self._dict_move_to_end(key)  # last=True             # <<<<<<<<<<<<<<
 *         return o
 * 
 */
  __Pyx_INCREF(__pyx_v_self->_dict_move_to_end);
  __pyx_t_2 = __pyx_v_self->_dict_move_to_end; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":46
 *             return default
 *         self._dict_move_to_end(key)  # last=True
 *         return o             # <<<<<<<<<<<<<<
 * 
 *     cdef inline needs_cleanup(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_o);
  __pyx_r = __pyx_v_o;
  goto __pyx_L0;

  /* "uvloop/lru.pyx":41
 *         self._maxsize = maxsize
 * 
 *     cdef get(self, key, default):             # <<<<<<<<<<<<<<
 *         o = self._dict_get(key, _LRU_MARKER)
 *         if o is _LRU_MARKER:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.LruCache.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":48
 *         return o
 * 
 *     cdef inline needs_cleanup(self):             # <<<<<<<<<<<<<<
 *         return len(self._dict) > self._maxsize
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8LruCache_needs_cleanup(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("needs_cleanup", 1);

  /* "uvloop/lru.pyx":49
 * 
 *     cdef inline needs_cleanup(self):
 *         return len(self._dict) > self._maxsize             # <<<<<<<<<<<<<<
 * 
 *     cdef inline cleanup_one(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->_dict;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(12, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_t_2 > __pyx_v_self->_maxsize)); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/lru.pyx":48
 *         return o
 * 
 *     cdef inline needs_cleanup(self):             # <<<<<<<<<<<<<<
 *         return len(self._dict) > self._maxsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.LruCache.needs_cleanup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":51
 *         return len(self._dict) > self._maxsize
 * 
 *     cdef inline cleanup_one(self):             # <<<<<<<<<<<<<<
 *         k, _ = self._dict.popitem(last=False)
 *         return k
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8LruCache_cleanup_one(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self) {
  PyObject *__pyx_v_k = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cleanup_one", 1);

  /* "uvloop/lru.pyx":52
 * 
 *     cdef inline cleanup_one(self):
 *         k, _ = self._dict.popitem(last=False)             # <<<<<<<<<<<<<<
 *         return k
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_dict, __pyx_n_s_popitem); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_last, Py_False) < 0) __PYX_ERR(12, 52, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(12, 52, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(12, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(12, 52, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(12, 52, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_k = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v__ = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":53
 *     cdef inline cleanup_one(self):
 *         k, _ = self._dict.popitem(last=False)
 *         return k             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_k);
  __pyx_r = __pyx_v_k;
  goto __pyx_L0;

  /* "uvloop/lru.pyx":51
 *         return len(self._dict) > self._maxsize
 * 
 *     cdef inline cleanup_one(self):             # <<<<<<<<<<<<<<
 *         k, _ = self._dict.popitem(last=False)
 *         return k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.LruCache.cleanup_one", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":55
 *         return k
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         o = self._dict[key]
 *         self._dict_move_to_end(key)  # last=True
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_2__getitem__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_2__getitem__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "uvloop/lru.pyx":56
 * 
 *     def __getitem__(self, key):
 *         o = self._dict[key]             # <<<<<<<<<<<<<<
 *         self._dict_move_to_end(key)  # last=True
 *         return o
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->_dict, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_o = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":57
 *     def __getitem__(self, key):
 *         o = self._dict[key]
 *         self._dict_move_to_end(key)  # last=True             # <<<<<<<<<<<<<<
 *         return o
 * 
 */
  __Pyx_INCREF(__pyx_v_self->_dict_move_to_end);
  __pyx_t_2 = __pyx_v_self->_dict_move_to_end; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/lru.pyx":58
 *         o = self._dict[key]
 *         self._dict_move_to_end(key)  # last=True
 *         return o             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, key, o):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_o);
  __pyx_r = __pyx_v_o;
  goto __pyx_L0;

  /* "uvloop/lru.pyx":55
 *         return k
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         o = self._dict[key]
 *         self._dict_move_to_end(key)  # last=True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.LruCache.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":60
 *         return o
 * 
 *     def __setitem__(self, key, o):             # <<<<<<<<<<<<<<
 *         if key in self._dict:
 *             self._dict[key] = o
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8LruCache_5__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_o); /*proto*/
static int __pyx_pw_6uvloop_4loop_8LruCache_5__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_o) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_4__setitem__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self), ((PyObject *)__pyx_v_key), ((PyObject *)__pyx_v_o));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8LruCache_4__setitem__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 1);

  /* "uvloop/lru.pyx":61
 * 
 *     def __setitem__(self, key, o):
 *         if key in self._dict:             # <<<<<<<<<<<<<<
 *             self._dict[key] = o
 *             self._dict_move_to_end(key)  # last=True
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_self->_dict, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(12, 61, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/lru.pyx":62
 *     def __setitem__(self, key, o):
 *         if key in self._dict:
 *             self._dict[key] = o             # <<<<<<<<<<<<<<
 *             self._dict_move_to_end(key)  # last=True
 *         else:
 */
    if (unlikely((PyObject_SetItem(__pyx_v_self->_dict, __pyx_v_key, __pyx_v_o) < 0))) __PYX_ERR(12, 62, __pyx_L1_error)

    /* "uvloop/lru.pyx":63
 *         if key in self._dict:
 *             self._dict[key] = o
 *             self._dict_move_to_end(key)  # last=True             # <<<<<<<<<<<<<<
 *         else:
 *             self._dict[key] = o
 */
    __Pyx_INCREF(__pyx_v_self->_dict_move_to_end);
    __pyx_t_3 = __pyx_v_self->_dict_move_to_end; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_key};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 63, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/lru.pyx":61
 * 
 *     def __setitem__(self, key, o):
 *         if key in self._dict:             # <<<<<<<<<<<<<<
 *             self._dict[key] = o
 *             self._dict_move_to_end(key)  # last=True
 */
    goto __pyx_L3;
  }

  /* "uvloop/lru.pyx":65
 *             self._dict_move_to_end(key)  # last=True
 *         else:
 *             self._dict[key] = o             # <<<<<<<<<<<<<<
 *         while self.needs_cleanup():
 *             self.cleanup_one()
 */
  /*else*/ {
    if (unlikely((PyObject_SetItem(__pyx_v_self->_dict, __pyx_v_key, __pyx_v_o) < 0))) __PYX_ERR(12, 65, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "uvloop/lru.pyx":66
 *         else:
 *             self._dict[key] = o
 *         while self.needs_cleanup():             # <<<<<<<<<<<<<<
 *             self.cleanup_one()
 * 
 */
  while (1) {
    __pyx_t_2 = __pyx_f_6uvloop_4loop_8LruCache_needs_cleanup(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(12, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!__pyx_t_1) break;

    /* "uvloop/lru.pyx":67
 *             self._dict[key] = o
 *         while self.needs_cleanup():
 *             self.cleanup_one()             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, key):
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_8LruCache_cleanup_one(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/lru.pyx":60
 *         return o
 * 
 *     def __setitem__(self, key, o):             # <<<<<<<<<<<<<<
 *         if key in self._dict:
 *             self._dict[key] = o
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.LruCache.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":69
 *             self.cleanup_one()
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         del self._dict[key]
 * 
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8LruCache_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_6uvloop_4loop_8LruCache_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_6__delitem__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8LruCache_6__delitem__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uvloop/lru.pyx":70
 * 
 *     def __delitem__(self, key):
 *         del self._dict[key]             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  if (unlikely((PyObject_DelItem(__pyx_v_self->_dict, __pyx_v_key) < 0))) __PYX_ERR(12, 70, __pyx_L1_error)

  /* "uvloop/lru.pyx":69
 *             self.cleanup_one()
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         del self._dict[key]
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.LruCache.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/lru.pyx":72
 *         del self._dict[key]
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return key in self._dict
 * 
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8LruCache_9__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_6uvloop_4loop_8LruCache_9__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_8__contains__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8LruCache_8__contains__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uvloop/lru.pyx":73
 * 
 *     def __contains__(self, key):
 *         return key in self._dict             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_self->_dict, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(12, 73, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "uvloop/lru.pyx":72
 *         del self._dict[key]
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return key in self._dict
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.LruCache.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/lru.pyx":75
 *         return key in self._dict
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self._dict)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6uvloop_4loop_8LruCache_11__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6uvloop_4loop_8LruCache_11__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_10__len__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6uvloop_4loop_8LruCache_10__len__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 1);

  /* "uvloop/lru.pyx":76
 * 
 *     def __len__(self):
 *         return len(self._dict)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_1 = __pyx_v_self->_dict;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(12, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "uvloop/lru.pyx":75
 *         return key in self._dict
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self._dict)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.LruCache.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/lru.pyx":78
 *         return len(self._dict)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self._dict)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_13__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_13__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_12__iter__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_12__iter__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "uvloop/lru.pyx":79
 * 
 *     def __iter__(self):
 *         return iter(self._dict)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->_dict;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(12, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/lru.pyx":78
 *         return len(self._dict)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self._dict)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.LruCache.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8LruCache_14__reduce_cython__, "LruCache.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8LruCache_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8LruCache_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8LruCache_14__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_14__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_14__reduce_cython__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._dict, self._dict_get, self._dict_move_to_end, self._maxsize)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_maxsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_dict);
  __Pyx_GIVEREF(__pyx_v_self->_dict);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_dict)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_dict_get);
  __Pyx_GIVEREF(__pyx_v_self->_dict_get);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->_dict_get)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_dict_move_to_end);
  __Pyx_GIVEREF(__pyx_v_self->_dict_move_to_end);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->_dict_move_to_end)) __PYX_ERR(6, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._dict, self._dict_get, self._dict_move_to_end, self._maxsize)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._dict, self._dict_get, self._dict_move_to_end, self._maxsize)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(6, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._dict is not None or self._dict_get is not None or self._dict_move_to_end is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._dict, self._dict_get, self._dict_move_to_end, self._maxsize)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._dict is not None or self._dict_get is not None or self._dict_move_to_end is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, None), state
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_self->_dict != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->_dict_get != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->_dict_move_to_end != Py_None);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._dict is not None or self._dict_get is not None or self._dict_move_to_end is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._dict is not None or self._dict_get is not None or self._dict_move_to_end is not None
 *     if use_setstate:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_LruCache); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_14808274);
    __Pyx_GIVEREF(__pyx_int_14808274);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_14808274)) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(6, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(6, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(6, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._dict is not None or self._dict_get is not None or self._dict_move_to_end is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, None), state
 *     else:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LruCache__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_LruCache); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(6, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_14808274);
    __Pyx_GIVEREF(__pyx_int_14808274);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_14808274)) __PYX_ERR(6, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(6, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5)) __PYX_ERR(6, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(6, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.LruCache.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LruCache__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8LruCache_16__setstate_cython__, "LruCache.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8LruCache_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8LruCache_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8LruCache_16__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8LruCache_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.LruCache.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_8LruCache_16__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8LruCache_16__setstate_cython__(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_LruCache__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pyx_unpickle_LruCache__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LruCache__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.LruCache.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":14
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._closed = 0
 *         self._inited = 0
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8UVHandle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_8UVHandle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVHandle___cinit__(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8UVHandle___cinit__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/handle.pyx":15
 * 
 *     def __cinit__(self):
 *         self._closed = 0             # <<<<<<<<<<<<<<
 *         self._inited = 0
 *         self._has_handle = 1
 */
  __pyx_v_self->_closed = 0;

  /* "uvloop/handles/handle.pyx":16
 *     def __cinit__(self):
 *         self._closed = 0
 *         self._inited = 0             # <<<<<<<<<<<<<<
 *         self._has_handle = 1
 *         self._handle = NULL
 */
  __pyx_v_self->_inited = 0;

  /* "uvloop/handles/handle.pyx":17
 *         self._closed = 0
 *         self._inited = 0
 *         self._has_handle = 1             # <<<<<<<<<<<<<<
 *         self._handle = NULL
 *         self._loop = None
 */
  __pyx_v_self->_has_handle = 1;

  /* "uvloop/handles/handle.pyx":18
 *         self._inited = 0
 *         self._has_handle = 1
 *         self._handle = NULL             # <<<<<<<<<<<<<<
 *         self._loop = None
 *         self._source_traceback = None
 */
  __pyx_v_self->_handle = NULL;

  /* "uvloop/handles/handle.pyx":19
 *         self._has_handle = 1
 *         self._handle = NULL
 *         self._loop = None             # <<<<<<<<<<<<<<
 *         self._source_traceback = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_loop);
  __pyx_v_self->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None);

  /* "uvloop/handles/handle.pyx":20
 *         self._handle = NULL
 *         self._loop = None
 *         self._source_traceback = None             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_source_traceback);
  __Pyx_DECREF(__pyx_v_self->_source_traceback);
  __pyx_v_self->_source_traceback = Py_None;

  /* "uvloop/handles/handle.pyx":14
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._closed = 0
 *         self._inited = 0
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":22
 *         self._source_traceback = None
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8UVHandle_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_8UVHandle_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVHandle_2__init__(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8UVHandle_2__init__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/handles/handle.pyx":24
 *     def __init__(self):
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(             # <<<<<<<<<<<<<<
 *                 self.__class__.__name__))
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_is_not_supposed_to_be_instantia, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/handles/handle.pyx":25
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(
 *                 self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/handles/handle.pyx":23
 * 
 *     def __init__(self):
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             '{} is not supposed to be instantiated from Python'.format(
 *                 self.__class__.__name__))
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(2, 23, __pyx_L1_error)

  /* "uvloop/handles/handle.pyx":22
 *         self._source_traceback = None
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             '{} is not supposed to be instantiated from Python'.format(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVHandle.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":27
 *                 self.__class__.__name__))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_8UVHandle_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_8UVHandle_5__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_8UVHandle_4__dealloc__(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_8UVHandle_4__dealloc__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/handles/handle.pyx":28
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self._loop is not None:
 *                 if self._inited:
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":29
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:             # <<<<<<<<<<<<<<
 *                 if self._inited:
 *                     self._loop._debug_handles_current.subtract([
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->_loop) != Py_None);
    if (likely(__pyx_t_1)) {

      /* "uvloop/handles/handle.pyx":30
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:
 *                 if self._inited:             # <<<<<<<<<<<<<<
 *                     self._loop._debug_handles_current.subtract([
 *                         self.__class__.__name__])
 */
      if (__pyx_v_self->_inited) {

        /* "uvloop/handles/handle.pyx":31
 *             if self._loop is not None:
 *                 if self._inited:
 *                     self._loop._debug_handles_current.subtract([             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__])
 *             else:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_loop->_debug_handles_current, __pyx_n_s_subtract); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 31, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);

        /* "uvloop/handles/handle.pyx":32
 *                 if self._inited:
 *                     self._loop._debug_handles_current.subtract([
 *                         self.__class__.__name__])             # <<<<<<<<<<<<<<
 *             else:
 *                 # No "@cython.no_gc_clear" decorator on this UVHandle
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 32, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 32, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/handles/handle.pyx":31
 *             if self._loop is not None:
 *                 if self._inited:
 *                     self._loop._debug_handles_current.subtract([             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__])
 *             else:
 */
        __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 31, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_5);
        if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_5)) __PYX_ERR(2, 31, __pyx_L1_error);
        __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 31, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/handles/handle.pyx":30
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:
 *                 if self._inited:             # <<<<<<<<<<<<<<
 *                     self._loop._debug_handles_current.subtract([
 *                         self.__class__.__name__])
 */
      }

      /* "uvloop/handles/handle.pyx":29
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:             # <<<<<<<<<<<<<<
 *                 if self._inited:
 *                     self._loop._debug_handles_current.subtract([
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/handle.pyx":35
 *             else:
 *                 # No "@cython.no_gc_clear" decorator on this UVHandle
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{} without @no_gc_clear; loop was set to None by GC'
 *                     .format(self.__class__.__name__))
 */
    /*else*/ {

      /* "uvloop/handles/handle.pyx":37
 *                 raise RuntimeError(
 *                     '{} without @no_gc_clear; loop was set to None by GC'
 *                     .format(self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *         if self._handle is NULL:
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_without_no_gc_clear_loop_was_se, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 37, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 37, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 37, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 37, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/handles/handle.pyx":35
 *             else:
 *                 # No "@cython.no_gc_clear" decorator on this UVHandle
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{} without @no_gc_clear; loop was set to None by GC'
 *                     .format(self.__class__.__name__))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 35, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "uvloop/handles/handle.pyx":28
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self._loop is not None:
 *                 if self._inited:
 */
  }

  /* "uvloop/handles/handle.pyx":39
 *                     .format(self.__class__.__name__))
 * 
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_handle == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":40
 * 
 *         if self._handle is NULL:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # -> When we're at this point, something is wrong <-
 */
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":39
 *                     .format(self.__class__.__name__))
 * 
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":44
 *         # -> When we're at this point, something is wrong <-
 * 
 *         if self._handle.loop is NULL:             # <<<<<<<<<<<<<<
 *             # The handle wasn't initialized with "uv_{handle}_init"
 *             self._closed = 1
 */
  __pyx_t_1 = (__pyx_v_self->_handle->loop == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/handle.pyx":46
 *         if self._handle.loop is NULL:
 *             # The handle wasn't initialized with "uv_{handle}_init"
 *             self._closed = 1             # <<<<<<<<<<<<<<
 *             self._free()
 *             raise RuntimeError(
 */
    __pyx_v_self->_closed = 1;

    /* "uvloop/handles/handle.pyx":47
 *             # The handle wasn't initialized with "uv_{handle}_init"
 *             self._closed = 1
 *             self._free()             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 '{} is open in __dealloc__ with loop set to NULL'
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_self->__pyx_vtab)->_free(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/handle.pyx":50
 *             raise RuntimeError(
 *                 '{} is open in __dealloc__ with loop set to NULL'
 *                 .format(self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *         if self._closed:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_is_open_in___dealloc___with_loo, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 50, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "uvloop/handles/handle.pyx":48
 *             self._closed = 1
 *             self._free()
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 '{} is open in __dealloc__ with loop set to NULL'
 *                 .format(self.__class__.__name__))
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 48, __pyx_L1_error)

    /* "uvloop/handles/handle.pyx":44
 *         # -> When we're at this point, something is wrong <-
 * 
 *         if self._handle.loop is NULL:             # <<<<<<<<<<<<<<
 *             # The handle wasn't initialized with "uv_{handle}_init"
 *             self._closed = 1
 */
  }

  /* "uvloop/handles/handle.pyx":52
 *                 .format(self.__class__.__name__))
 * 
 *         if self._closed:             # <<<<<<<<<<<<<<
 *             # So _handle is not NULL and self._closed == 1?
 *             raise RuntimeError(
 */
  if (unlikely(__pyx_v_self->_closed)) {

    /* "uvloop/handles/handle.pyx":55
 *             # So _handle is not NULL and self._closed == 1?
 *             raise RuntimeError(
 *                 '{}.__dealloc__: _handle is NULL, _closed == 1'.format(             # <<<<<<<<<<<<<<
 *                     self.__class__.__name__))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_dealloc____handle_is_NULL__cl, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/handle.pyx":56
 *             raise RuntimeError(
 *                 '{}.__dealloc__: _handle is NULL, _closed == 1'.format(
 *                     self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *         # The handle is dealloced while open.  Let's try to close it.
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "uvloop/handles/handle.pyx":54
 *         if self._closed:
 *             # So _handle is not NULL and self._closed == 1?
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 '{}.__dealloc__: _handle is NULL, _closed == 1'.format(
 *                     self.__class__.__name__))
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 54, __pyx_L1_error)

    /* "uvloop/handles/handle.pyx":52
 *                 .format(self.__class__.__name__))
 * 
 *         if self._closed:             # <<<<<<<<<<<<<<
 *             # So _handle is not NULL and self._closed == 1?
 *             raise RuntimeError(
 */
  }

  /* "uvloop/handles/handle.pyx":61
 *         # Situations when this is possible include unhandled exceptions,
 *         # errors during Handle.__cinit__/__init__ etc.
 *         if self._inited:             # <<<<<<<<<<<<<<
 *             self._handle.data = NULL
 *             uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 */
  if (__pyx_v_self->_inited) {

    /* "uvloop/handles/handle.pyx":62
 *         # errors during Handle.__cinit__/__init__ etc.
 *         if self._inited:
 *             self._handle.data = NULL             # <<<<<<<<<<<<<<
 *             uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 *             self._handle = NULL
 */
    __pyx_v_self->_handle->data = NULL;

    /* "uvloop/handles/handle.pyx":63
 *         if self._inited:
 *             self._handle.data = NULL
 *             uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors             # <<<<<<<<<<<<<<
 *             self._handle = NULL
 *             self._warn_unclosed()
 */
    uv_close(__pyx_v_self->_handle, __pyx_f_6uvloop_4loop___uv_close_handle_cb);

    /* "uvloop/handles/handle.pyx":64
 *             self._handle.data = NULL
 *             uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 *             self._handle = NULL             # <<<<<<<<<<<<<<
 *             self._warn_unclosed()
 *         else:
 */
    __pyx_v_self->_handle = NULL;

    /* "uvloop/handles/handle.pyx":65
 *             uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 *             self._handle = NULL
 *             self._warn_unclosed()             # <<<<<<<<<<<<<<
 *         else:
 *             # The handle was allocated, but not initialized
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_self->__pyx_vtab)->_warn_unclosed(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/handle.pyx":61
 *         # Situations when this is possible include unhandled exceptions,
 *         # errors during Handle.__cinit__/__init__ etc.
 *         if self._inited:             # <<<<<<<<<<<<<<
 *             self._handle.data = NULL
 *             uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 */
    goto __pyx_L9;
  }

  /* "uvloop/handles/handle.pyx":68
 *         else:
 *             # The handle was allocated, but not initialized
 *             self._closed = 1             # <<<<<<<<<<<<<<
 *             self._free()
 * 
 */
  /*else*/ {
    __pyx_v_self->_closed = 1;

    /* "uvloop/handles/handle.pyx":69
 *             # The handle was allocated, but not initialized
 *             self._closed = 1
 *             self._free()             # <<<<<<<<<<<<<<
 * 
 *     cdef _free(self):
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_self->__pyx_vtab)->_free(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L9:;

  /* "uvloop/handles/handle.pyx":27
 *                 self.__class__.__name__))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uvloop.loop.UVHandle.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/handles/handle.pyx":71
 *             self._free()
 * 
 *     cdef _free(self):             # <<<<<<<<<<<<<<
 *         if self._handle == NULL:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__free(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_free", 1);

  /* "uvloop/handles/handle.pyx":72
 * 
 *     cdef _free(self):
 *         if self._handle == NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_handle == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":73
 *     cdef _free(self):
 *         if self._handle == NULL:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG and self._inited:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":72
 * 
 *     cdef _free(self):
 *         if self._handle == NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":75
 *             return
 * 
 *         if UVLOOP_DEBUG and self._inited:             # <<<<<<<<<<<<<<
 *             self._loop._debug_uv_handles_freed += 1
 * 
 */
  __pyx_t_2 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->_inited;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":76
 * 
 *         if UVLOOP_DEBUG and self._inited:
 *             self._loop._debug_uv_handles_freed += 1             # <<<<<<<<<<<<<<
 * 
 *         PyMem_RawFree(self._handle)
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self->_loop);
    __pyx_t_3 = __pyx_v_self->_loop;
    __pyx_t_3->_debug_uv_handles_freed = (__pyx_t_3->_debug_uv_handles_freed + 1);
    __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/handle.pyx":75
 *             return
 * 
 *         if UVLOOP_DEBUG and self._inited:             # <<<<<<<<<<<<<<
 *             self._loop._debug_uv_handles_freed += 1
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":78
 *             self._loop._debug_uv_handles_freed += 1
 * 
 *         PyMem_RawFree(self._handle)             # <<<<<<<<<<<<<<
 *         self._handle = NULL
 * 
 */
  PyMem_RawFree(__pyx_v_self->_handle);

  /* "uvloop/handles/handle.pyx":79
 * 
 *         PyMem_RawFree(self._handle)
 *         self._handle = NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef _warn_unclosed(self):
 */
  __pyx_v_self->_handle = NULL;

  /* "uvloop/handles/handle.pyx":71
 *             self._free()
 * 
 *     cdef _free(self):             # <<<<<<<<<<<<<<
 *         if self._handle == NULL:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":81
 *         self._handle = NULL
 * 
 *     cdef _warn_unclosed(self):             # <<<<<<<<<<<<<<
 *         if self._source_traceback is not None:
 *             try:
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__warn_unclosed(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_v_tb = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_warn_unclosed", 1);

  /* "uvloop/handles/handle.pyx":82
 * 
 *     cdef _warn_unclosed(self):
 *         if self._source_traceback is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 tb = ''.join(tb_format_list(self._source_traceback))
 */
  __pyx_t_1 = (__pyx_v_self->_source_traceback != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":83
 *     cdef _warn_unclosed(self):
 *         if self._source_traceback is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 tb = ''.join(tb_format_list(self._source_traceback))
 *                 tb = 'object created at (most recent call last):\n{}'.format(
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/handles/handle.pyx":84
 *         if self._source_traceback is not None:
 *             try:
 *                 tb = ''.join(tb_format_list(self._source_traceback))             # <<<<<<<<<<<<<<
 *                 tb = 'object created at (most recent call last):\n{}'.format(
 *                     tb.rstrip())
 */
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_tb_format_list);
        __pyx_t_6 = __pyx_v_6uvloop_4loop_tb_format_list; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_self->_source_traceback};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 84, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __pyx_t_6 = PyUnicode_Join(__pyx_kp_u__10, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 84, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_tb = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "uvloop/handles/handle.pyx":85
 *             try:
 *                 tb = ''.join(tb_format_list(self._source_traceback))
 *                 tb = 'object created at (most recent call last):\n{}'.format(             # <<<<<<<<<<<<<<
 *                     tb.rstrip())
 *             except Exception as ex:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_object_created_at_most_recent_ca, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 85, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "uvloop/handles/handle.pyx":86
 *                 tb = ''.join(tb_format_list(self._source_traceback))
 *                 tb = 'object created at (most recent call last):\n{}'.format(
 *                     tb.rstrip())             # <<<<<<<<<<<<<<
 *             except Exception as ex:
 *                 msg = (
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_rstrip); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 86, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 86, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __pyx_t_9 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 85, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_tb, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "uvloop/handles/handle.pyx":83
 *     cdef _warn_unclosed(self):
 *         if self._source_traceback is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 tb = ''.join(tb_format_list(self._source_traceback))
 *                 tb = 'object created at (most recent call last):\n{}'.format(
 */
      }

      /* "uvloop/handles/handle.pyx":93
 *                 ).format(self, type(ex).__name__, ex)
 *             else:
 *                 msg = 'unclosed resource {!r}; {}'.format(self, tb)             # <<<<<<<<<<<<<<
 *         else:
 *             msg = 'unclosed resource {!r}'.format(self)
 */
      /*else:*/ {
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unclosed_resource_r, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 93, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_7, ((PyObject *)__pyx_v_self), __pyx_v_tb};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 93, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __pyx_v_msg = __pyx_t_6;
        __pyx_t_6 = 0;
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/handles/handle.pyx":87
 *                 tb = 'object created at (most recent call last):\n{}'.format(
 *                     tb.rstrip())
 *             except Exception as ex:             # <<<<<<<<<<<<<<
 *                 msg = (
 *                     'unclosed resource {!r}; could not serialize '
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.UVHandle._warn_unclosed", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_7) < 0) __PYX_ERR(2, 87, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __pyx_v_ex = __pyx_t_5;
        /*try:*/ {

          /* "uvloop/handles/handle.pyx":91
 *                     'unclosed resource {!r}; could not serialize '
 *                     'debug traceback: {}: {}'
 *                 ).format(self, type(ex).__name__, ex)             # <<<<<<<<<<<<<<
 *             else:
 *                 msg = 'unclosed resource {!r}; {}'.format(self, tb)
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unclosed_resource_r_could_not_se, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 91, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_ex)), __pyx_n_s_name); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 91, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[4] = {__pyx_t_13, ((PyObject *)__pyx_v_self), __pyx_t_12, __pyx_v_ex};
            __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 91, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __pyx_v_msg = __pyx_t_9;
          __pyx_t_9 = 0;
        }

        /* "uvloop/handles/handle.pyx":87
 *                 tb = 'object created at (most recent call last):\n{}'.format(
 *                     tb.rstrip())
 *             except Exception as ex:             # <<<<<<<<<<<<<<
 *                 msg = (
 *                     'unclosed resource {!r}; could not serialize '
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L16;
          }
          __pyx_L15_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_11 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L6_except_error;
          }
          __pyx_L16:;
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "uvloop/handles/handle.pyx":83
 *     cdef _warn_unclosed(self):
 *         if self._source_traceback is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 tb = ''.join(tb_format_list(self._source_traceback))
 *                 tb = 'object created at (most recent call last):\n{}'.format(
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "uvloop/handles/handle.pyx":82
 * 
 *     cdef _warn_unclosed(self):
 *         if self._source_traceback is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 tb = ''.join(tb_format_list(self._source_traceback))
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/handle.pyx":95
 *                 msg = 'unclosed resource {!r}; {}'.format(self, tb)
 *         else:
 *             msg = 'unclosed resource {!r}'.format(self)             # <<<<<<<<<<<<<<
 *         warnings_warn(msg, ResourceWarning)
 * 
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unclosed_resource_r_2, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, ((PyObject *)__pyx_v_self)};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_v_msg = __pyx_t_7;
    __pyx_t_7 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/handle.pyx":96
 *         else:
 *             msg = 'unclosed resource {!r}'.format(self)
 *         warnings_warn(msg, ResourceWarning)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _abort_init(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ResourceWarning); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_warnings_warn);
  __pyx_t_6 = __pyx_v_6uvloop_4loop_warnings_warn; __pyx_t_9 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_v_msg, __pyx_t_5};
    __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uvloop/handles/handle.pyx":81
 *         self._handle = NULL
 * 
 *     cdef _warn_unclosed(self):             # <<<<<<<<<<<<<<
 *         if self._source_traceback is not None:
 *             try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._warn_unclosed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tb);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":98
 *         warnings_warn(msg, ResourceWarning)
 * 
 *     cdef inline _abort_init(self):             # <<<<<<<<<<<<<<
 *         if self._handle is not NULL:
 *             self._free()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__abort_init(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_abort_init", 1);

  /* "uvloop/handles/handle.pyx":99
 * 
 *     cdef inline _abort_init(self):
 *         if self._handle is not NULL:             # <<<<<<<<<<<<<<
 *             self._free()
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_handle != NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":100
 *     cdef inline _abort_init(self):
 *         if self._handle is not NULL:
 *             self._free()             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_self->__pyx_vtab)->_free(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/handle.pyx":99
 * 
 *     cdef inline _abort_init(self):
 *         if self._handle is not NULL:             # <<<<<<<<<<<<<<
 *             self._free()
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":102
 *             self._free()
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 name = self.__class__.__name__
 */
  /*try:*/ {

    /* "uvloop/handles/handle.pyx":103
 * 
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 name = self.__class__.__name__
 *                 if self._inited:
 */
    __pyx_t_1 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_1) {

      /* "uvloop/handles/handle.pyx":104
 *         try:
 *             if UVLOOP_DEBUG:
 *                 name = self.__class__.__name__             # <<<<<<<<<<<<<<
 *                 if self._inited:
 *                     raise RuntimeError(
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 104, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 104, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_name = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "uvloop/handles/handle.pyx":105
 *             if UVLOOP_DEBUG:
 *                 name = self.__class__.__name__
 *                 if self._inited:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_abort_init: {}._inited is set'.format(name))
 */
      if (unlikely(__pyx_v_self->_inited)) {

        /* "uvloop/handles/handle.pyx":107
 *                 if self._inited:
 *                     raise RuntimeError(
 *                         '_abort_init: {}._inited is set'.format(name))             # <<<<<<<<<<<<<<
 *                 if self._closed:
 *                     raise RuntimeError(
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_abort_init__inited_is_set, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 107, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_name};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 107, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "uvloop/handles/handle.pyx":106
 *                 name = self.__class__.__name__
 *                 if self._inited:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_abort_init: {}._inited is set'.format(name))
 *                 if self._closed:
 */
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 106, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(2, 106, __pyx_L5_error)

        /* "uvloop/handles/handle.pyx":105
 *             if UVLOOP_DEBUG:
 *                 name = self.__class__.__name__
 *                 if self._inited:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_abort_init: {}._inited is set'.format(name))
 */
      }

      /* "uvloop/handles/handle.pyx":108
 *                     raise RuntimeError(
 *                         '_abort_init: {}._inited is set'.format(name))
 *                 if self._closed:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_abort_init: {}._closed is set'.format(name))
 */
      if (unlikely(__pyx_v_self->_closed)) {

        /* "uvloop/handles/handle.pyx":110
 *                 if self._closed:
 *                     raise RuntimeError(
 *                         '_abort_init: {}._closed is set'.format(name))             # <<<<<<<<<<<<<<
 *         finally:
 *             self._closed = 1
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_abort_init__closed_is_set, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 110, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_name};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 110, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "uvloop/handles/handle.pyx":109
 *                         '_abort_init: {}._inited is set'.format(name))
 *                 if self._closed:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_abort_init: {}._closed is set'.format(name))
 *         finally:
 */
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 109, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(2, 109, __pyx_L5_error)

        /* "uvloop/handles/handle.pyx":108
 *                     raise RuntimeError(
 *                         '_abort_init: {}._inited is set'.format(name))
 *                 if self._closed:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_abort_init: {}._closed is set'.format(name))
 */
      }

      /* "uvloop/handles/handle.pyx":103
 * 
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 name = self.__class__.__name__
 *                 if self._inited:
 */
    }
  }

  /* "uvloop/handles/handle.pyx":112
 *                         '_abort_init: {}._closed is set'.format(name))
 *         finally:
 *             self._closed = 1             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _finish_init(self):
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_v_self->_closed = 1;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        __pyx_v_self->_closed = 1;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/handles/handle.pyx":98
 *         warnings_warn(msg, ResourceWarning)
 * 
 *     cdef inline _abort_init(self):             # <<<<<<<<<<<<<<
 *         if self._handle is not NULL:
 *             self._free()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._abort_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":114
 *             self._closed = 1
 * 
 *     cdef inline _finish_init(self):             # <<<<<<<<<<<<<<
 *         self._inited = 1
 *         if self._has_handle == 1:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__finish_init(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_v_cls_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_finish_init", 1);

  /* "uvloop/handles/handle.pyx":115
 * 
 *     cdef inline _finish_init(self):
 *         self._inited = 1             # <<<<<<<<<<<<<<
 *         if self._has_handle == 1:
 *             self._handle.data = <void*>self
 */
  __pyx_v_self->_inited = 1;

  /* "uvloop/handles/handle.pyx":116
 *     cdef inline _finish_init(self):
 *         self._inited = 1
 *         if self._has_handle == 1:             # <<<<<<<<<<<<<<
 *             self._handle.data = <void*>self
 *         if self._loop._debug:
 */
  __pyx_t_1 = (__pyx_v_self->_has_handle == 1);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":117
 *         self._inited = 1
 *         if self._has_handle == 1:
 *             self._handle.data = <void*>self             # <<<<<<<<<<<<<<
 *         if self._loop._debug:
 *             self._source_traceback = extract_stack()
 */
    __pyx_v_self->_handle->data = ((void *)__pyx_v_self);

    /* "uvloop/handles/handle.pyx":116
 *     cdef inline _finish_init(self):
 *         self._inited = 1
 *         if self._has_handle == 1:             # <<<<<<<<<<<<<<
 *             self._handle.data = <void*>self
 *         if self._loop._debug:
 */
  }

  /* "uvloop/handles/handle.pyx":118
 *         if self._has_handle == 1:
 *             self._handle.data = <void*>self
 *         if self._loop._debug:             # <<<<<<<<<<<<<<
 *             self._source_traceback = extract_stack()
 *         if UVLOOP_DEBUG:
 */
  if (__pyx_v_self->_loop->_debug) {

    /* "uvloop/handles/handle.pyx":119
 *             self._handle.data = <void*>self
 *         if self._loop._debug:
 *             self._source_traceback = extract_stack()             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             cls_name = self.__class__.__name__
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_extract_stack(); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_source_traceback);
    __Pyx_DECREF(__pyx_v_self->_source_traceback);
    __pyx_v_self->_source_traceback = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/handle.pyx":118
 *         if self._has_handle == 1:
 *             self._handle.data = <void*>self
 *         if self._loop._debug:             # <<<<<<<<<<<<<<
 *             self._source_traceback = extract_stack()
 *         if UVLOOP_DEBUG:
 */
  }

  /* "uvloop/handles/handle.pyx":120
 *         if self._loop._debug:
 *             self._source_traceback = extract_stack()
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             cls_name = self.__class__.__name__
 *             self._loop._debug_uv_handles_total += 1
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":121
 *             self._source_traceback = extract_stack()
 *         if UVLOOP_DEBUG:
 *             cls_name = self.__class__.__name__             # <<<<<<<<<<<<<<
 *             self._loop._debug_uv_handles_total += 1
 *             self._loop._debug_handles_total.update([cls_name])
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_cls_name = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/handles/handle.pyx":122
 *         if UVLOOP_DEBUG:
 *             cls_name = self.__class__.__name__
 *             self._loop._debug_uv_handles_total += 1             # <<<<<<<<<<<<<<
 *             self._loop._debug_handles_total.update([cls_name])
 *             self._loop._debug_handles_current.update([cls_name])
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self->_loop);
    __pyx_t_4 = __pyx_v_self->_loop;
    __pyx_t_4->_debug_uv_handles_total = (__pyx_t_4->_debug_uv_handles_total + 1);
    __Pyx_DECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/handle.pyx":123
 *             cls_name = self.__class__.__name__
 *             self._loop._debug_uv_handles_total += 1
 *             self._loop._debug_handles_total.update([cls_name])             # <<<<<<<<<<<<<<
 *             self._loop._debug_handles_current.update([cls_name])
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_loop->_debug_handles_total, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_cls_name);
    __Pyx_GIVEREF(__pyx_v_cls_name);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_cls_name)) __PYX_ERR(2, 123, __pyx_L1_error);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/handle.pyx":124
 *             self._loop._debug_uv_handles_total += 1
 *             self._loop._debug_handles_total.update([cls_name])
 *             self._loop._debug_handles_current.update([cls_name])             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _start_init(self, Loop loop):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_loop->_debug_handles_current, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_cls_name);
    __Pyx_GIVEREF(__pyx_v_cls_name);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_cls_name)) __PYX_ERR(2, 124, __pyx_L1_error);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/handle.pyx":120
 *         if self._loop._debug:
 *             self._source_traceback = extract_stack()
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             cls_name = self.__class__.__name__
 *             self._loop._debug_uv_handles_total += 1
 */
  }

  /* "uvloop/handles/handle.pyx":114
 *             self._closed = 1
 * 
 *     cdef inline _finish_init(self):             # <<<<<<<<<<<<<<
 *         self._inited = 1
 *         if self._has_handle == 1:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF((PyObject *)__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._finish_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cls_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":126
 *             self._loop._debug_handles_current.update([cls_name])
 * 
 *     cdef inline _start_init(self, Loop loop):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__start_init(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_start_init", 1);

  /* "uvloop/handles/handle.pyx":127
 * 
 *     cdef inline _start_init(self, Loop loop):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self._loop is not None:
 *                 raise RuntimeError(
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":128
 *     cdef inline _start_init(self, Loop loop):
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._start_init can only be called once'.format(
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->_loop) != Py_None);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/handle.pyx":130
 *             if self._loop is not None:
 *                 raise RuntimeError(
 *                     '{}._start_init can only be called once'.format(             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_start_init_can_only_be_called, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 130, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/handles/handle.pyx":131
 *                 raise RuntimeError(
 *                     '{}._start_init can only be called once'.format(
 *                         self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *         self._loop = loop
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/handles/handle.pyx":129
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{}._start_init can only be called once'.format(
 *                         self.__class__.__name__))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 129, __pyx_L1_error)

      /* "uvloop/handles/handle.pyx":128
 *     cdef inline _start_init(self, Loop loop):
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._start_init can only be called once'.format(
 */
    }

    /* "uvloop/handles/handle.pyx":127
 * 
 *     cdef inline _start_init(self, Loop loop):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self._loop is not None:
 *                 raise RuntimeError(
 */
  }

  /* "uvloop/handles/handle.pyx":133
 *                         self.__class__.__name__))
 * 
 *         self._loop = loop             # <<<<<<<<<<<<<<
 * 
 *     cdef inline bint _is_alive(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_loop);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_loop);
  __pyx_v_self->_loop = __pyx_v_loop;

  /* "uvloop/handles/handle.pyx":126
 *             self._loop._debug_handles_current.update([cls_name])
 * 
 *     cdef inline _start_init(self, Loop loop):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if self._loop is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._start_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":135
 *         self._loop = loop
 * 
 *     cdef inline bint _is_alive(self):             # <<<<<<<<<<<<<<
 *         cdef bint res
 *         res = self._closed != 1 and self._inited == 1
 */

static CYTHON_INLINE int __pyx_f_6uvloop_4loop_8UVHandle__is_alive(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  int __pyx_v_res;
  PyObject *__pyx_v_name = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_alive", 1);

  /* "uvloop/handles/handle.pyx":137
 *     cdef inline bint _is_alive(self):
 *         cdef bint res
 *         res = self._closed != 1 and self._inited == 1             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if res and self._has_handle == 1:
 */
  __pyx_t_2 = (__pyx_v_self->_closed != 1);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->_inited == 1);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_v_res = __pyx_t_1;

  /* "uvloop/handles/handle.pyx":138
 *         cdef bint res
 *         res = self._closed != 1 and self._inited == 1
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if res and self._has_handle == 1:
 *                 name = self.__class__.__name__
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":139
 *         res = self._closed != 1 and self._inited == 1
 *         if UVLOOP_DEBUG:
 *             if res and self._has_handle == 1:             # <<<<<<<<<<<<<<
 *                 name = self.__class__.__name__
 *                 if self._handle is NULL:
 */
    if (__pyx_v_res) {
    } else {
      __pyx_t_1 = __pyx_v_res;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->_has_handle == 1);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "uvloop/handles/handle.pyx":140
 *         if UVLOOP_DEBUG:
 *             if res and self._has_handle == 1:
 *                 name = self.__class__.__name__             # <<<<<<<<<<<<<<
 *                 if self._handle is NULL:
 *                     raise RuntimeError(
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_name = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/handles/handle.pyx":141
 *             if res and self._has_handle == 1:
 *                 name = self.__class__.__name__
 *                 if self._handle is NULL:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle is NULL'.format(name))
 */
      __pyx_t_1 = (__pyx_v_self->_handle == NULL);
      if (unlikely(__pyx_t_1)) {

        /* "uvloop/handles/handle.pyx":143
 *                 if self._handle is NULL:
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle is NULL'.format(name))             # <<<<<<<<<<<<<<
 *                 if self._loop is None:
 *                     raise RuntimeError(
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_is_alive_but__handle_is_NULL, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_name};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 143, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "uvloop/handles/handle.pyx":142
 *                 name = self.__class__.__name__
 *                 if self._handle is NULL:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '{} is alive, but _handle is NULL'.format(name))
 *                 if self._loop is None:
 */
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 142, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(2, 142, __pyx_L1_error)

        /* "uvloop/handles/handle.pyx":141
 *             if res and self._has_handle == 1:
 *                 name = self.__class__.__name__
 *                 if self._handle is NULL:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle is NULL'.format(name))
 */
      }

      /* "uvloop/handles/handle.pyx":144
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle is NULL'.format(name))
 *                 if self._loop is None:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _loop is None'.format(name))
 */
      __pyx_t_1 = (((PyObject *)__pyx_v_self->_loop) == Py_None);
      if (unlikely(__pyx_t_1)) {

        /* "uvloop/handles/handle.pyx":146
 *                 if self._loop is None:
 *                     raise RuntimeError(
 *                         '{} is alive, but _loop is None'.format(name))             # <<<<<<<<<<<<<<
 *                 if self._handle.loop is not self._loop.uvloop:
 *                     raise RuntimeError(
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_is_alive_but__loop_is_None, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 146, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_name};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 146, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "uvloop/handles/handle.pyx":145
 *                         '{} is alive, but _handle is NULL'.format(name))
 *                 if self._loop is None:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '{} is alive, but _loop is None'.format(name))
 *                 if self._handle.loop is not self._loop.uvloop:
 */
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 145, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 145, __pyx_L1_error)

        /* "uvloop/handles/handle.pyx":144
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle is NULL'.format(name))
 *                 if self._loop is None:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _loop is None'.format(name))
 */
      }

      /* "uvloop/handles/handle.pyx":147
 *                     raise RuntimeError(
 *                         '{} is alive, but _loop is None'.format(name))
 *                 if self._handle.loop is not self._loop.uvloop:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle.loop is not '
 */
      __pyx_t_1 = (__pyx_v_self->_handle->loop != __pyx_v_self->_loop->uvloop);
      if (unlikely(__pyx_t_1)) {

        /* "uvloop/handles/handle.pyx":150
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle.loop is not '
 *                         'initialized'.format(name))             # <<<<<<<<<<<<<<
 *                 if self._handle.data is not <void*>self:
 *                     raise RuntimeError(
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_is_alive_but__handle_loop_is_no, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 150, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_name};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 150, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }

        /* "uvloop/handles/handle.pyx":148
 *                         '{} is alive, but _loop is None'.format(name))
 *                 if self._handle.loop is not self._loop.uvloop:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '{} is alive, but _handle.loop is not '
 *                         'initialized'.format(name))
 */
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(2, 148, __pyx_L1_error)

        /* "uvloop/handles/handle.pyx":147
 *                     raise RuntimeError(
 *                         '{} is alive, but _loop is None'.format(name))
 *                 if self._handle.loop is not self._loop.uvloop:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle.loop is not '
 */
      }

      /* "uvloop/handles/handle.pyx":151
 *                         '{} is alive, but _handle.loop is not '
 *                         'initialized'.format(name))
 *                 if self._handle.data is not <void*>self:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle.data is not '
 */
      __pyx_t_1 = (__pyx_v_self->_handle->data != ((void *)__pyx_v_self));
      if (unlikely(__pyx_t_1)) {

        /* "uvloop/handles/handle.pyx":154
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle.data is not '
 *                         'initialized'.format(name))             # <<<<<<<<<<<<<<
 *         return res
 * 
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_is_alive_but__handle_data_is_no, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_name};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 154, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "uvloop/handles/handle.pyx":152
 *                         'initialized'.format(name))
 *                 if self._handle.data is not <void*>self:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '{} is alive, but _handle.data is not '
 *                         'initialized'.format(name))
 */
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 152, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 152, __pyx_L1_error)

        /* "uvloop/handles/handle.pyx":151
 *                         '{} is alive, but _handle.loop is not '
 *                         'initialized'.format(name))
 *                 if self._handle.data is not <void*>self:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '{} is alive, but _handle.data is not '
 */
      }

      /* "uvloop/handles/handle.pyx":139
 *         res = self._closed != 1 and self._inited == 1
 *         if UVLOOP_DEBUG:
 *             if res and self._has_handle == 1:             # <<<<<<<<<<<<<<
 *                 name = self.__class__.__name__
 *                 if self._handle is NULL:
 */
    }

    /* "uvloop/handles/handle.pyx":138
 *         cdef bint res
 *         res = self._closed != 1 and self._inited == 1
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if res and self._has_handle == 1:
 *                 name = self.__class__.__name__
 */
  }

  /* "uvloop/handles/handle.pyx":155
 *                         '{} is alive, but _handle.data is not '
 *                         'initialized'.format(name))
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _ensure_alive(self):
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "uvloop/handles/handle.pyx":135
 *         self._loop = loop
 * 
 *     cdef inline bint _is_alive(self):             # <<<<<<<<<<<<<<
 *         cdef bint res
 *         res = self._closed != 1 and self._inited == 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._is_alive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":157
 *         return res
 * 
 *     cdef inline _ensure_alive(self):             # <<<<<<<<<<<<<<
 *         if not self._is_alive():
 *             raise RuntimeError(
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVHandle__ensure_alive(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ensure_alive", 1);

  /* "uvloop/handles/handle.pyx":158
 * 
 *     cdef inline _ensure_alive(self):
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'unable to perform operation on {!r}; '
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVHandle__is_alive(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 158, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/handle.pyx":161
 *             raise RuntimeError(
 *                 'unable to perform operation on {!r}; '
 *                 'the handler is closed'.format(self))             # <<<<<<<<<<<<<<
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unable_to_perform_operation_on_r, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, ((PyObject *)__pyx_v_self)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "uvloop/handles/handle.pyx":159
 *     cdef inline _ensure_alive(self):
 *         if not self._is_alive():
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'unable to perform operation on {!r}; '
 *                 'the handler is closed'.format(self))
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 159, __pyx_L1_error)

    /* "uvloop/handles/handle.pyx":158
 * 
 *     cdef inline _ensure_alive(self):
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'unable to perform operation on {!r}; '
 */
  }

  /* "uvloop/handles/handle.pyx":157
 *         return res
 * 
 *     cdef inline _ensure_alive(self):             # <<<<<<<<<<<<<<
 *         if not self._is_alive():
 *             raise RuntimeError(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._ensure_alive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":163
 *                 'the handler is closed'.format(self))
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         # Fatal error means an error that was returned by the
 *         # underlying libuv handle function.  We usually can't
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__fatal_error(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error *__pyx_optional_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fatal_error", 1);
  if (__pyx_optional_args) {
  }

  /* "uvloop/handles/handle.pyx":167
 *         # underlying libuv handle function.  We usually can't
 *         # recover from that, hence we just close the handle.
 *         self._close()             # <<<<<<<<<<<<<<
 * 
 *         if throw or self._loop is None:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_self->__pyx_vtab)->_close(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/handle.pyx":169
 *         self._close()
 * 
 *         if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *             raise exc
 *         else:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_throw); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(2, 169, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((PyObject *)__pyx_v_self->_loop) == Py_None);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/handle.pyx":170
 * 
 *         if throw or self._loop is None:
 *             raise exc             # <<<<<<<<<<<<<<
 *         else:
 *             self._loop._handle_exception(exc)
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(2, 170, __pyx_L1_error)

    /* "uvloop/handles/handle.pyx":169
 *         self._close()
 * 
 *         if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *             raise exc
 *         else:
 */
  }

  /* "uvloop/handles/handle.pyx":172
 *             raise exc
 *         else:
 *             self._loop._handle_exception(exc)             # <<<<<<<<<<<<<<
 * 
 *     cdef _error(self, exc, throw):
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->_loop->__pyx_vtab)->_handle_exception(__pyx_v_self->_loop, __pyx_v_exc); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 172, __pyx_L1_error)
  }

  /* "uvloop/handles/handle.pyx":163
 *                 'the handler is closed'.format(self))
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         # Fatal error means an error that was returned by the
 *         # underlying libuv handle function.  We usually can't
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._fatal_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":174
 *             self._loop._handle_exception(exc)
 * 
 *     cdef _error(self, exc, throw):             # <<<<<<<<<<<<<<
 *         # A non-fatal error is usually an error that was caught
 *         # by the handler, but was originated in the client code
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__error(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_error", 1);

  /* "uvloop/handles/handle.pyx":179
 *         # (not in libuv).  In this case we either want to simply
 *         # raise or log it.
 *         if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *             raise exc
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_throw); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 179, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (((PyObject *)__pyx_v_self->_loop) == Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/handle.pyx":180
 *         # raise or log it.
 *         if throw or self._loop is None:
 *             raise exc             # <<<<<<<<<<<<<<
 *         else:
 *             self._loop._handle_exception(exc)
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(2, 180, __pyx_L1_error)

    /* "uvloop/handles/handle.pyx":179
 *         # (not in libuv).  In this case we either want to simply
 *         # raise or log it.
 *         if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *             raise exc
 *         else:
 */
  }

  /* "uvloop/handles/handle.pyx":182
 *             raise exc
 *         else:
 *             self._loop._handle_exception(exc)             # <<<<<<<<<<<<<<
 * 
 *     cdef _close(self):
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->_loop->__pyx_vtab)->_handle_exception(__pyx_v_self->_loop, __pyx_v_exc); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 182, __pyx_L1_error)
  }

  /* "uvloop/handles/handle.pyx":174
 *             self._loop._handle_exception(exc)
 * 
 *     cdef _error(self, exc, throw):             # <<<<<<<<<<<<<<
 *         # A non-fatal error is usually an error that was caught
 *         # by the handler, but was originated in the client code
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVHandle._error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":184
 *             self._loop._handle_exception(exc)
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         if self._closed == 1:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVHandle__close(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/handles/handle.pyx":185
 * 
 *     cdef _close(self):
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_closed == 1);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":186
 *     cdef _close(self):
 *         if self._closed == 1:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._closed = 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":185
 * 
 *     cdef _close(self):
 *         if self._closed == 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":188
 *             return
 * 
 *         self._closed = 1             # <<<<<<<<<<<<<<
 * 
 *         if self._handle is NULL:
 */
  __pyx_v_self->_closed = 1;

  /* "uvloop/handles/handle.pyx":190
 *         self._closed = 1
 * 
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_handle == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":191
 * 
 *         if self._handle is NULL:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":190
 *         self._closed = 1
 * 
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":193
 *             return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self._handle.data is NULL:
 *                 raise RuntimeError(
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":194
 * 
 *         if UVLOOP_DEBUG:
 *             if self._handle.data is NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is NULL'.format(
 */
    __pyx_t_1 = (__pyx_v_self->_handle->data == NULL);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/handle.pyx":196
 *             if self._handle.data is NULL:
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is NULL'.format(             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_close__handle_data_is_NULL, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/handles/handle.pyx":197
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is NULL'.format(
 *                         self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *             if <object>self._handle.data is not self:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 196, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/handles/handle.pyx":195
 *         if UVLOOP_DEBUG:
 *             if self._handle.data is NULL:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{}._close: _handle.data is NULL'.format(
 *                         self.__class__.__name__))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 195, __pyx_L1_error)

      /* "uvloop/handles/handle.pyx":194
 * 
 *         if UVLOOP_DEBUG:
 *             if self._handle.data is NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is NULL'.format(
 */
    }

    /* "uvloop/handles/handle.pyx":199
 *                         self.__class__.__name__))
 * 
 *             if <object>self._handle.data is not self:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is not UVHandle/self'.format(
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->_handle->data) != ((PyObject *)__pyx_v_self));
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/handle.pyx":201
 *             if <object>self._handle.data is not self:
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is not UVHandle/self'.format(             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__))
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_close__handle_data_is_not_UVHa, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "uvloop/handles/handle.pyx":202
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is not UVHandle/self'.format(
 *                         self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *             if uv.uv_is_closing(self._handle):
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 201, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }

      /* "uvloop/handles/handle.pyx":200
 * 
 *             if <object>self._handle.data is not self:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{}._close: _handle.data is not UVHandle/self'.format(
 *                         self.__class__.__name__))
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(2, 200, __pyx_L1_error)

      /* "uvloop/handles/handle.pyx":199
 *                         self.__class__.__name__))
 * 
 *             if <object>self._handle.data is not self:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._close: _handle.data is not UVHandle/self'.format(
 */
    }

    /* "uvloop/handles/handle.pyx":204
 *                         self.__class__.__name__))
 * 
 *             if uv.uv_is_closing(self._handle):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._close: uv_is_closing() is true'.format(
 */
    __pyx_t_1 = (uv_is_closing(__pyx_v_self->_handle) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/handle.pyx":206
 *             if uv.uv_is_closing(self._handle):
 *                 raise RuntimeError(
 *                     '{}._close: uv_is_closing() is true'.format(             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_close_uv_is_closing_is_true, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/handles/handle.pyx":207
 *                 raise RuntimeError(
 *                     '{}._close: uv_is_closing() is true'.format(
 *                         self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *         # We want the handle wrapper (UVHandle) to stay alive until
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/handles/handle.pyx":205
 * 
 *             if uv.uv_is_closing(self._handle):
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{}._close: uv_is_closing() is true'.format(
 *                         self.__class__.__name__))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 205, __pyx_L1_error)

      /* "uvloop/handles/handle.pyx":204
 *                         self.__class__.__name__))
 * 
 *             if uv.uv_is_closing(self._handle):             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}._close: uv_is_closing() is true'.format(
 */
    }

    /* "uvloop/handles/handle.pyx":193
 *             return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self._handle.data is NULL:
 *                 raise RuntimeError(
 */
  }

  /* "uvloop/handles/handle.pyx":211
 *         # We want the handle wrapper (UVHandle) to stay alive until
 *         # the closing callback fires.
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 *         uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 * 
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "uvloop/handles/handle.pyx":212
 *         # the closing callback fires.
 *         Py_INCREF(self)
 *         uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  uv_close(__pyx_v_self->_handle, __pyx_f_6uvloop_4loop___uv_close_handle_cb);

  /* "uvloop/handles/handle.pyx":184
 *             self._loop._handle_exception(exc)
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         if self._closed == 1:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVHandle._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":214
 *         uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} closed={} {:#x}>'.format(
 *             self.__class__.__name__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_7__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_7__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVHandle_6__repr__(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_6__repr__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/handles/handle.pyx":215
 * 
 *     def __repr__(self):
 *         return '<{} closed={} {:#x}>'.format(             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self._closed,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_closed_x, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/handles/handle.pyx":216
 *     def __repr__(self):
 *         return '<{} closed={} {:#x}>'.format(
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self._closed,
 *             id(self))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/handle.pyx":217
 *         return '<{} closed={} {:#x}>'.format(
 *             self.__class__.__name__,
 *             self._closed,             # <<<<<<<<<<<<<<
 *             id(self))
 * 
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->_closed); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/handles/handle.pyx":218
 *             self.__class__.__name__,
 *             self._closed,
 *             id(self))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_6, __pyx_t_4, __pyx_t_3, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 3+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/handle.pyx":214
 *         uv.uv_close(self._handle, __uv_close_handle_cb)  # void; no errors
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} closed={} {:#x}>'.format(
 *             self.__class__.__name__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UVHandle.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pxd":5
 *         uv.uv_handle_t *_handle
 *         Loop _loop
 *         readonly _source_traceback             # <<<<<<<<<<<<<<
 *         bint _closed
 *         bint _inited
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_17_source_traceback_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_17_source_traceback_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVHandle_17_source_traceback___get__(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_17_source_traceback___get__(struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_source_traceback);
  __pyx_r = __pyx_v_self->_source_traceback;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVHandle_8__reduce_cython__, "UVHandle.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVHandle_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVHandle_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVHandle_8__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVHandle_8__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVHandle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVHandle_10__setstate_cython__, "UVHandle.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVHandle_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVHandle_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVHandle_10__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8UVHandle_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVHandle_10__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVHandle_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":223
 * cdef class UVSocketHandle(UVHandle):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._fileobj = None
 *         self.__cached_socket = None
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_14UVSocketHandle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_14UVSocketHandle_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_14UVSocketHandle___cinit__(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_14UVSocketHandle___cinit__(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/handle.pyx":224
 * 
 *     def __cinit__(self):
 *         self._fileobj = None             # <<<<<<<<<<<<<<
 *         self.__cached_socket = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_fileobj);
  __Pyx_DECREF(__pyx_v_self->_fileobj);
  __pyx_v_self->_fileobj = Py_None;

  /* "uvloop/handles/handle.pyx":225
 *     def __cinit__(self):
 *         self._fileobj = None
 *         self.__cached_socket = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _fileno(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_UVSocketHandle__cached_socket);
  __Pyx_DECREF(__pyx_v_self->_UVSocketHandle__cached_socket);
  __pyx_v_self->_UVSocketHandle__cached_socket = Py_None;

  /* "uvloop/handles/handle.pyx":223
 * cdef class UVSocketHandle(UVHandle):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._fileobj = None
 *         self.__cached_socket = None
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":227
 *         self.__cached_socket = None
 * 
 *     cdef _fileno(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int fd
 */

static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__fileno(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self) {
  int __pyx_v_fd;
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fileno", 1);

  /* "uvloop/handles/handle.pyx":232
 *             int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         err = uv.uv_fileno(self._handle, <uv.uv_os_fd_t*>&fd)
 *         if err < 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/handle.pyx":233
 * 
 *         self._ensure_alive()
 *         err = uv.uv_fileno(self._handle, <uv.uv_os_fd_t*>&fd)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
  __pyx_v_err = uv_fileno(__pyx_v_self->__pyx_base._handle, ((uv_os_fd_t *)(&__pyx_v_fd)));

  /* "uvloop/handles/handle.pyx":234
 *         self._ensure_alive()
 *         err = uv.uv_fileno(self._handle, <uv.uv_os_fd_t*>&fd)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/handle.pyx":235
 *         err = uv.uv_fileno(self._handle, <uv.uv_os_fd_t*>&fd)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         return fd
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(2, 235, __pyx_L1_error)

    /* "uvloop/handles/handle.pyx":234
 *         self._ensure_alive()
 *         err = uv.uv_fileno(self._handle, <uv.uv_os_fd_t*>&fd)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":237
 *             raise convert_error(err)
 * 
 *         return fd             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/handle.pyx":227
 *         self.__cached_socket = None
 * 
 *     cdef _fileno(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int fd
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._fileno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":239
 *         return fd
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__new_socket(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/handle.pyx":240
 * 
 *     cdef _new_socket(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _get_socket(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(2, 240, __pyx_L1_error)

  /* "uvloop/handles/handle.pyx":239
 *         return fd
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":242
 *         raise NotImplementedError
 * 
 *     cdef inline _get_socket(self):             # <<<<<<<<<<<<<<
 *         if self.__cached_socket is not None:
 *             return self.__cached_socket
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__get_socket(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_socket", 1);

  /* "uvloop/handles/handle.pyx":243
 * 
 *     cdef inline _get_socket(self):
 *         if self.__cached_socket is not None:             # <<<<<<<<<<<<<<
 *             return self.__cached_socket
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_UVSocketHandle__cached_socket != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":244
 *     cdef inline _get_socket(self):
 *         if self.__cached_socket is not None:
 *             return self.__cached_socket             # <<<<<<<<<<<<<<
 * 
 *         if not self._is_alive():
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_UVSocketHandle__cached_socket);
    __pyx_r = __pyx_v_self->_UVSocketHandle__cached_socket;
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":243
 * 
 *     cdef inline _get_socket(self):
 *         if self.__cached_socket is not None:             # <<<<<<<<<<<<<<
 *             return self.__cached_socket
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":246
 *             return self.__cached_socket
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 246, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/handles/handle.pyx":247
 * 
 *         if not self._is_alive():
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         self.__cached_socket = self._new_socket()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":246
 *             return self.__cached_socket
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":249
 *             return None
 * 
 *         self.__cached_socket = self._new_socket()             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             # We don't "dup" for the "__cached_socket".
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle *)__pyx_v_self->__pyx_base.__pyx_vtab)->_new_socket(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_UVSocketHandle__cached_socket);
  __Pyx_DECREF(__pyx_v_self->_UVSocketHandle__cached_socket);
  __pyx_v_self->_UVSocketHandle__cached_socket = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/handles/handle.pyx":250
 * 
 *         self.__cached_socket = self._new_socket()
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             # We don't "dup" for the "__cached_socket".
 *             assert self.__cached_socket.fileno() == self._fileno()
 */
  __pyx_t_2 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/handle.pyx":252
 *         if UVLOOP_DEBUG:
 *             # We don't "dup" for the "__cached_socket".
 *             assert self.__cached_socket.fileno() == self._fileno()             # <<<<<<<<<<<<<<
 *         return self.__cached_socket
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_UVSocketHandle__cached_socket, __pyx_n_s_fileno); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 252, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle *)__pyx_v_self->__pyx_base.__pyx_vtab)->_fileno(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 252, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(2, 252, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(2, 252, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(2, 252, __pyx_L1_error)
    #endif

    /* "uvloop/handles/handle.pyx":250
 * 
 *         self.__cached_socket = self._new_socket()
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             # We don't "dup" for the "__cached_socket".
 *             assert self.__cached_socket.fileno() == self._fileno()
 */
  }

  /* "uvloop/handles/handle.pyx":253
 *             # We don't "dup" for the "__cached_socket".
 *             assert self.__cached_socket.fileno() == self._fileno()
 *         return self.__cached_socket             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _attach_fileobj(self, object file):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_UVSocketHandle__cached_socket);
  __pyx_r = __pyx_v_self->_UVSocketHandle__cached_socket;
  goto __pyx_L0;

  /* "uvloop/handles/handle.pyx":242
 *         raise NotImplementedError
 * 
 *     cdef inline _get_socket(self):             # <<<<<<<<<<<<<<
 *         if self.__cached_socket is not None:
 *             return self.__cached_socket
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._get_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":255
 *         return self.__cached_socket
 * 
 *     cdef inline _attach_fileobj(self, object file):             # <<<<<<<<<<<<<<
 *         # When we create a TCP/PIPE/etc connection/server based on
 *         # a Python file object, we need to close the file object when
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__attach_fileobj(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self, PyObject *__pyx_v_file) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_attach_fileobj", 1);

  /* "uvloop/handles/handle.pyx":259
 *         # a Python file object, we need to close the file object when
 *         # the uv handle is closed.
 *         socket_inc_io_ref(file)             # <<<<<<<<<<<<<<
 *         self._fileobj = file
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_socket_inc_io_ref(__pyx_v_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/handle.pyx":260
 *         # the uv handle is closed.
 *         socket_inc_io_ref(file)
 *         self._fileobj = file             # <<<<<<<<<<<<<<
 * 
 *     cdef _close(self):
 */
  __Pyx_INCREF(__pyx_v_file);
  __Pyx_GIVEREF(__pyx_v_file);
  __Pyx_GOTREF(__pyx_v_self->_fileobj);
  __Pyx_DECREF(__pyx_v_self->_fileobj);
  __pyx_v_self->_fileobj = __pyx_v_file;

  /* "uvloop/handles/handle.pyx":255
 *         return self.__cached_socket
 * 
 *     cdef inline _attach_fileobj(self, object file):             # <<<<<<<<<<<<<<
 *         # When we create a TCP/PIPE/etc connection/server based on
 *         # a Python file object, we need to close the file object when
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._attach_fileobj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":262
 *         self._fileobj = file
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         if self.__cached_socket is not None:
 *             (<PseudoSocket>self.__cached_socket)._fd = -1
 */

static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__close(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self) {
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  char const *__pyx_t_26;
  char const *__pyx_t_27;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/handles/handle.pyx":263
 * 
 *     cdef _close(self):
 *         if self.__cached_socket is not None:             # <<<<<<<<<<<<<<
 *             (<PseudoSocket>self.__cached_socket)._fd = -1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_UVSocketHandle__cached_socket != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":264
 *     cdef _close(self):
 *         if self.__cached_socket is not None:
 *             (<PseudoSocket>self.__cached_socket)._fd = -1             # <<<<<<<<<<<<<<
 * 
 *         UVHandle._close(self)
 */
    ((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v_self->_UVSocketHandle__cached_socket)->_fd = -1;

    /* "uvloop/handles/handle.pyx":263
 * 
 *     cdef _close(self):
 *         if self.__cached_socket is not None:             # <<<<<<<<<<<<<<
 *             (<PseudoSocket>self.__cached_socket)._fd = -1
 * 
 */
  }

  /* "uvloop/handles/handle.pyx":266
 *             (<PseudoSocket>self.__cached_socket)._fd = -1
 * 
 *         UVHandle._close(self)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_8UVHandle__close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/handle.pyx":268
 *         UVHandle._close(self)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             # This code will only run for transports created from
 *             # Python sockets, i.e. with `loop.create_server(sock=sock)` etc.
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "uvloop/handles/handle.pyx":271
 *             # This code will only run for transports created from
 *             # Python sockets, i.e. with `loop.create_server(sock=sock)` etc.
 *             if self._fileobj is not None:             # <<<<<<<<<<<<<<
 *                 if isinstance(self._fileobj, socket_socket):
 *                     # Detaching the socket object is the ideal solution:
 */
        __pyx_t_1 = (__pyx_v_self->_fileobj != Py_None);
        if (__pyx_t_1) {

          /* "uvloop/handles/handle.pyx":272
 *             # Python sockets, i.e. with `loop.create_server(sock=sock)` etc.
 *             if self._fileobj is not None:
 *                 if isinstance(self._fileobj, socket_socket):             # <<<<<<<<<<<<<<
 *                     # Detaching the socket object is the ideal solution:
 *                     # * libuv will actually close the FD;
 */
          __pyx_t_2 = __pyx_v_self->_fileobj;
          __Pyx_INCREF(__pyx_t_2);
          __pyx_t_6 = __pyx_v_6uvloop_4loop_socket_socket;
          __Pyx_INCREF(__pyx_t_6);
          __pyx_t_1 = PyObject_IsInstance(__pyx_t_2, __pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 272, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_1) {

            /* "uvloop/handles/handle.pyx":281
 *                     # No need to call `socket_dec_io_ref()`, as
 *                     # `socket.detach()` ignores `socket._io_refs`.
 *                     self._fileobj.detach()             # <<<<<<<<<<<<<<
 *                 else:
 *                     try:
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fileobj, __pyx_n_s_detach); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 281, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = NULL;
            __pyx_t_8 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_2))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_2, function);
                __pyx_t_8 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
              __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 281, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

            /* "uvloop/handles/handle.pyx":272
 *             # Python sockets, i.e. with `loop.create_server(sock=sock)` etc.
 *             if self._fileobj is not None:
 *                 if isinstance(self._fileobj, socket_socket):             # <<<<<<<<<<<<<<
 *                     # Detaching the socket object is the ideal solution:
 *                     # * libuv will actually close the FD;
 */
            goto __pyx_L14;
          }

          /* "uvloop/handles/handle.pyx":283
 *                     self._fileobj.detach()
 *                 else:
 *                     try:             # <<<<<<<<<<<<<<
 *                         # `socket.close()` will raise an EBADF because libuv
 *                         # has already closed the underlying FD.
 */
          /*else*/ {
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
              __Pyx_XGOTREF(__pyx_t_9);
              __Pyx_XGOTREF(__pyx_t_10);
              __Pyx_XGOTREF(__pyx_t_11);
              /*try:*/ {

                /* "uvloop/handles/handle.pyx":286
 *                         # `socket.close()` will raise an EBADF because libuv
 *                         # has already closed the underlying FD.
 *                         self._fileobj.close()             # <<<<<<<<<<<<<<
 *                     except OSError as ex:
 *                         if ex.errno != errno_EBADF:
 */
                __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fileobj, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 286, __pyx_L15_error)
                __Pyx_GOTREF(__pyx_t_2);
                __pyx_t_7 = NULL;
                __pyx_t_8 = 0;
                #if CYTHON_UNPACK_METHODS
                if (likely(PyMethod_Check(__pyx_t_2))) {
                  __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
                  if (likely(__pyx_t_7)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
                    __Pyx_INCREF(__pyx_t_7);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_2, function);
                    __pyx_t_8 = 1;
                  }
                }
                #endif
                {
                  PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
                  __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
                  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                  if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 286, __pyx_L15_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                }
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

                /* "uvloop/handles/handle.pyx":283
 *                     self._fileobj.detach()
 *                 else:
 *                     try:             # <<<<<<<<<<<<<<
 *                         # `socket.close()` will raise an EBADF because libuv
 *                         # has already closed the underlying FD.
 */
              }
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              goto __pyx_L20_try_end;
              __pyx_L15_error:;
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "uvloop/handles/handle.pyx":287
 *                         # has already closed the underlying FD.
 *                         self._fileobj.close()
 *                     except OSError as ex:             # <<<<<<<<<<<<<<
 *                         if ex.errno != errno_EBADF:
 *                             raise
 */
              __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
              if (__pyx_t_12) {
                __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_2, &__pyx_t_7) < 0) __PYX_ERR(2, 287, __pyx_L17_except_error)
                __Pyx_XGOTREF(__pyx_t_6);
                __Pyx_XGOTREF(__pyx_t_2);
                __Pyx_XGOTREF(__pyx_t_7);
                __Pyx_INCREF(__pyx_t_2);
                __pyx_v_ex = __pyx_t_2;
                /*try:*/ {

                  /* "uvloop/handles/handle.pyx":288
 *                         self._fileobj.close()
 *                     except OSError as ex:
 *                         if ex.errno != errno_EBADF:             # <<<<<<<<<<<<<<
 *                             raise
 *         except Exception as ex:
 */
                  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_ex, __pyx_n_s_errno); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 288, __pyx_L26_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __pyx_t_14 = PyObject_RichCompare(__pyx_t_13, __pyx_v_6uvloop_4loop_errno_EBADF, Py_NE); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 288, __pyx_L26_error)
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(2, 288, __pyx_L26_error)
                  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                  if (unlikely(__pyx_t_1)) {

                    /* "uvloop/handles/handle.pyx":289
 *                     except OSError as ex:
 *                         if ex.errno != errno_EBADF:
 *                             raise             # <<<<<<<<<<<<<<
 *         except Exception as ex:
 *             self._loop.call_exception_handler({
 */
                    __Pyx_GIVEREF(__pyx_t_6);
                    __Pyx_GIVEREF(__pyx_t_2);
                    __Pyx_XGIVEREF(__pyx_t_7);
                    __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_2, __pyx_t_7);
                    __pyx_t_6 = 0; __pyx_t_2 = 0; __pyx_t_7 = 0; 
                    __PYX_ERR(2, 289, __pyx_L26_error)

                    /* "uvloop/handles/handle.pyx":288
 *                         self._fileobj.close()
 *                     except OSError as ex:
 *                         if ex.errno != errno_EBADF:             # <<<<<<<<<<<<<<
 *                             raise
 *         except Exception as ex:
 */
                  }
                }

                /* "uvloop/handles/handle.pyx":287
 *                         # has already closed the underlying FD.
 *                         self._fileobj.close()
 *                     except OSError as ex:             # <<<<<<<<<<<<<<
 *                         if ex.errno != errno_EBADF:
 *                             raise
 */
                /*finally:*/ {
                  /*normal exit:*/{
                    __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                    goto __pyx_L27;
                  }
                  __pyx_L26_error:;
                  /*exception exit:*/{
                    __Pyx_PyThreadState_declare
                    __Pyx_PyThreadState_assign
                    __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
                    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                    if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
                    if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
                    __Pyx_XGOTREF(__pyx_t_17);
                    __Pyx_XGOTREF(__pyx_t_18);
                    __Pyx_XGOTREF(__pyx_t_19);
                    __Pyx_XGOTREF(__pyx_t_20);
                    __Pyx_XGOTREF(__pyx_t_21);
                    __Pyx_XGOTREF(__pyx_t_22);
                    __pyx_t_12 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
                    {
                      __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                    }
                    if (PY_MAJOR_VERSION >= 3) {
                      __Pyx_XGIVEREF(__pyx_t_20);
                      __Pyx_XGIVEREF(__pyx_t_21);
                      __Pyx_XGIVEREF(__pyx_t_22);
                      __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
                    }
                    __Pyx_XGIVEREF(__pyx_t_17);
                    __Pyx_XGIVEREF(__pyx_t_18);
                    __Pyx_XGIVEREF(__pyx_t_19);
                    __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
                    __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
                    __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
                    goto __pyx_L17_except_error;
                  }
                  __pyx_L27:;
                }
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
                goto __pyx_L16_exception_handled;
              }
              goto __pyx_L17_except_error;

              /* "uvloop/handles/handle.pyx":283
 *                     self._fileobj.detach()
 *                 else:
 *                     try:             # <<<<<<<<<<<<<<
 *                         # `socket.close()` will raise an EBADF because libuv
 *                         # has already closed the underlying FD.
 */
              __pyx_L17_except_error:;
              __Pyx_XGIVEREF(__pyx_t_9);
              __Pyx_XGIVEREF(__pyx_t_10);
              __Pyx_XGIVEREF(__pyx_t_11);
              __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
              goto __pyx_L7_error;
              __pyx_L16_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_9);
              __Pyx_XGIVEREF(__pyx_t_10);
              __Pyx_XGIVEREF(__pyx_t_11);
              __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
              __pyx_L20_try_end:;
            }
          }
          __pyx_L14:;

          /* "uvloop/handles/handle.pyx":271
 *             # This code will only run for transports created from
 *             # Python sockets, i.e. with `loop.create_server(sock=sock)` etc.
 *             if self._fileobj is not None:             # <<<<<<<<<<<<<<
 *                 if isinstance(self._fileobj, socket_socket):
 *                     # Detaching the socket object is the ideal solution:
 */
        }

        /* "uvloop/handles/handle.pyx":268
 *         UVHandle._close(self)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             # This code will only run for transports created from
 *             # Python sockets, i.e. with `loop.create_server(sock=sock)` etc.
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/handles/handle.pyx":290
 *                         if ex.errno != errno_EBADF:
 *                             raise
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._loop.call_exception_handler({
 *                 'exception': ex,
 */
      __pyx_t_15 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_15) {
        __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_2, &__pyx_t_6) < 0) __PYX_ERR(2, 290, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_ex, __pyx_t_2);
        /*try:*/ {

          /* "uvloop/handles/handle.pyx":291
 *                             raise
 *         except Exception as ex:
 *             self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                 'exception': ex,
 *                 'transport': self,
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base._loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 291, __pyx_L38_error)
          __Pyx_GOTREF(__pyx_t_13);

          /* "uvloop/handles/handle.pyx":292
 *         except Exception as ex:
 *             self._loop.call_exception_handler({
 *                 'exception': ex,             # <<<<<<<<<<<<<<
 *                 'transport': self,
 *                 'message': f'could not close attached file object '
 */
          __pyx_t_23 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_23)) __PYX_ERR(2, 292, __pyx_L38_error)
          __Pyx_GOTREF(__pyx_t_23);
          if (PyDict_SetItem(__pyx_t_23, __pyx_n_u_exception, __pyx_v_ex) < 0) __PYX_ERR(2, 292, __pyx_L38_error)

          /* "uvloop/handles/handle.pyx":293
 *             self._loop.call_exception_handler({
 *                 'exception': ex,
 *                 'transport': self,             # <<<<<<<<<<<<<<
 *                 'message': f'could not close attached file object '
 *                            f'{self._fileobj!r}',
 */
          if (PyDict_SetItem(__pyx_t_23, __pyx_n_u_transport, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(2, 292, __pyx_L38_error)

          /* "uvloop/handles/handle.pyx":295
 *                 'transport': self,
 *                 'message': f'could not close attached file object '
 *                            f'{self._fileobj!r}',             # <<<<<<<<<<<<<<
 *             })
 *         finally:
 */
          __pyx_t_24 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_self->_fileobj), __pyx_empty_unicode); if (unlikely(!__pyx_t_24)) __PYX_ERR(2, 295, __pyx_L38_error)
          __Pyx_GOTREF(__pyx_t_24);

          /* "uvloop/handles/handle.pyx":294
 *                 'exception': ex,
 *                 'transport': self,
 *                 'message': f'could not close attached file object '             # <<<<<<<<<<<<<<
 *                            f'{self._fileobj!r}',
 *             })
 */
          __pyx_t_25 = __Pyx_PyUnicode_Concat(__pyx_kp_u_could_not_close_attached_file_ob, __pyx_t_24); if (unlikely(!__pyx_t_25)) __PYX_ERR(2, 294, __pyx_L38_error)
          __Pyx_GOTREF(__pyx_t_25);
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          if (PyDict_SetItem(__pyx_t_23, __pyx_n_u_message, __pyx_t_25) < 0) __PYX_ERR(2, 292, __pyx_L38_error)
          __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
          __pyx_t_25 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_25)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_25);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_25, __pyx_t_23};
            __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
            __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 291, __pyx_L38_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        }

        /* "uvloop/handles/handle.pyx":290
 *                         if ex.errno != errno_EBADF:
 *                             raise
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._loop.call_exception_handler({
 *                 'exception': ex,
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L39;
          }
          __pyx_L38_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
            __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
            __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_21, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_10);
            __Pyx_XGOTREF(__pyx_t_9);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_15 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_26 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_21, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_10);
            __Pyx_XGIVEREF(__pyx_t_9);
            __Pyx_ErrRestore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
            __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_15; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_26;
            goto __pyx_L9_except_error;
          }
          __pyx_L39:;
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "uvloop/handles/handle.pyx":268
 *         UVHandle._close(self)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             # This code will only run for transports created from
 *             # Python sockets, i.e. with `loop.create_server(sock=sock)` etc.
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L5_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      __pyx_L12_try_end:;
    }
  }

  /* "uvloop/handles/handle.pyx":298
 *             })
 *         finally:
 *             self._fileobj = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _open(self, int sockfd):
 */
  /*finally:*/ {
    /*normal exit:*/{
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->_fileobj);
      __Pyx_DECREF(__pyx_v_self->_fileobj);
      __pyx_v_self->_fileobj = Py_None;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
      __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
      __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_22);
      __pyx_t_12 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_27 = __pyx_filename;
      {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF(__pyx_v_self->_fileobj);
        __Pyx_DECREF(__pyx_v_self->_fileobj);
        __pyx_v_self->_fileobj = Py_None;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
      }
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
      __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_27;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/handles/handle.pyx":262
 *         self._fileobj = file
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         if self.__cached_socket is not None:
 *             (<PseudoSocket>self.__cached_socket)._fd = -1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":300
 *             self._fileobj = None
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_14UVSocketHandle__open(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self, CYTHON_UNUSED int __pyx_v_sockfd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open", 1);

  /* "uvloop/handles/handle.pyx":301
 * 
 *     cdef _open(self, int sockfd):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(2, 301, __pyx_L1_error)

  /* "uvloop/handles/handle.pyx":300
 *             self._fileobj = None
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_14UVSocketHandle_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_14UVSocketHandle_2__reduce_cython__, "UVSocketHandle.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_14UVSocketHandle_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVSocketHandle_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVSocketHandle_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_14UVSocketHandle_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_14UVSocketHandle_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_14UVSocketHandle_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_14UVSocketHandle_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_14UVSocketHandle_4__setstate_cython__, "UVSocketHandle.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_14UVSocketHandle_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVSocketHandle_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVSocketHandle_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_14UVSocketHandle_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_14UVSocketHandle_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_14UVSocketHandle_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVSocketHandle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":304
 * 
 * 
 * cdef inline bint __ensure_handle_data(uv.uv_handle_t* handle,             # <<<<<<<<<<<<<<
 *                                       const char* handle_ctx):
 * 
 */

static CYTHON_INLINE int __pyx_f_6uvloop_4loop___ensure_handle_data(uv_handle_t *__pyx_v_handle, char const *__pyx_v_handle_ctx) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ensure_handle_data", 1);

  /* "uvloop/handles/handle.pyx":309
 *     cdef Loop loop
 * 
 *     if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *         if handle.loop is NULL:
 *             raise RuntimeError(
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":310
 * 
 *     if UVLOOP_DEBUG:
 *         if handle.loop is NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'handle.loop is NULL in __ensure_handle_data')
 */
    __pyx_t_1 = (__pyx_v_handle->loop == NULL);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/handle.pyx":311
 *     if UVLOOP_DEBUG:
 *         if handle.loop is NULL:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'handle.loop is NULL in __ensure_handle_data')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__136, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(2, 311, __pyx_L1_error)

      /* "uvloop/handles/handle.pyx":310
 * 
 *     if UVLOOP_DEBUG:
 *         if handle.loop is NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'handle.loop is NULL in __ensure_handle_data')
 */
    }

    /* "uvloop/handles/handle.pyx":314
 *                 'handle.loop is NULL in __ensure_handle_data')
 * 
 *         if handle.loop.data is NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'handle.loop.data is NULL in __ensure_handle_data')
 */
    __pyx_t_1 = (__pyx_v_handle->loop->data == NULL);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/handle.pyx":315
 * 
 *         if handle.loop.data is NULL:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'handle.loop.data is NULL in __ensure_handle_data')
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__137, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 315, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(2, 315, __pyx_L1_error)

      /* "uvloop/handles/handle.pyx":314
 *                 'handle.loop is NULL in __ensure_handle_data')
 * 
 *         if handle.loop.data is NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'handle.loop.data is NULL in __ensure_handle_data')
 */
    }

    /* "uvloop/handles/handle.pyx":309
 *     cdef Loop loop
 * 
 *     if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *         if handle.loop is NULL:
 *             raise RuntimeError(
 */
  }

  /* "uvloop/handles/handle.pyx":318
 *                 'handle.loop.data is NULL in __ensure_handle_data')
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         loop = <Loop>handle.loop.data
 *         loop.call_exception_handler({
 */
  __pyx_t_1 = (__pyx_v_handle->data == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":319
 * 
 *     if handle.data is NULL:
 *         loop = <Loop>handle.loop.data             # <<<<<<<<<<<<<<
 *         loop.call_exception_handler({
 *             'message': '{} called with handle.data == NULL'.format(
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_handle->loop->data);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/handle.pyx":320
 *     if handle.data is NULL:
 *         loop = <Loop>handle.loop.data
 *         loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *             'message': '{} called with handle.data == NULL'.format(
 *                 handle_ctx.decode('latin-1'))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/handle.pyx":321
 *         loop = <Loop>handle.loop.data
 *         loop.call_exception_handler({
 *             'message': '{} called with handle.data == NULL'.format(             # <<<<<<<<<<<<<<
 *                 handle_ctx.decode('latin-1'))
 *         })
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_called_with_handle_data_NULL, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "uvloop/handles/handle.pyx":322
 *         loop.call_exception_handler({
 *             'message': '{} called with handle.data == NULL'.format(
 *                 handle_ctx.decode('latin-1'))             # <<<<<<<<<<<<<<
 *         })
 *         return 0
 */
    __pyx_t_7 = __Pyx_ssize_strlen(__pyx_v_handle_ctx); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 322, __pyx_L1_error)
    __pyx_t_8 = __Pyx_decode_c_string(__pyx_v_handle_ctx, 0, __pyx_t_7, NULL, NULL, PyUnicode_DecodeLatin1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 321, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_message, __pyx_t_5) < 0) __PYX_ERR(2, 321, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/handle.pyx":324
 *                 handle_ctx.decode('latin-1'))
 *         })
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     if handle.data is NULL:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":318
 *                 'handle.loop.data is NULL in __ensure_handle_data')
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         loop = <Loop>handle.loop.data
 *         loop.call_exception_handler({
 */
  }

  /* "uvloop/handles/handle.pyx":326
 *         return 0
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         # The underlying UVHandle object was GCed with an open uv_handle_t.
 *         loop = <Loop>handle.loop.data
 */
  __pyx_t_1 = (__pyx_v_handle->data == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":328
 *     if handle.data is NULL:
 *         # The underlying UVHandle object was GCed with an open uv_handle_t.
 *         loop = <Loop>handle.loop.data             # <<<<<<<<<<<<<<
 *         loop.call_exception_handler({
 *             'message': '{} called after destroying the UVHandle'.format(
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_handle->loop->data);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/handle.pyx":329
 *         # The underlying UVHandle object was GCed with an open uv_handle_t.
 *         loop = <Loop>handle.loop.data
 *         loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *             'message': '{} called after destroying the UVHandle'.format(
 *                 handle_ctx.decode('latin-1'))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/handle.pyx":330
 *         loop = <Loop>handle.loop.data
 *         loop.call_exception_handler({
 *             'message': '{} called after destroying the UVHandle'.format(             # <<<<<<<<<<<<<<
 *                 handle_ctx.decode('latin-1'))
 *         })
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_called_after_destroying_the_UVH, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "uvloop/handles/handle.pyx":331
 *         loop.call_exception_handler({
 *             'message': '{} called after destroying the UVHandle'.format(
 *                 handle_ctx.decode('latin-1'))             # <<<<<<<<<<<<<<
 *         })
 *         return 0
 */
    __pyx_t_7 = __Pyx_ssize_strlen(__pyx_v_handle_ctx); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 331, __pyx_L1_error)
    __pyx_t_8 = __Pyx_decode_c_string(__pyx_v_handle_ctx, 0, __pyx_t_7, NULL, NULL, PyUnicode_DecodeLatin1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_message, __pyx_t_5) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 329, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/handle.pyx":333
 *                 handle_ctx.decode('latin-1'))
 *         })
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":326
 *         return 0
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         # The underlying UVHandle object was GCed with an open uv_handle_t.
 *         loop = <Loop>handle.loop.data
 */
  }

  /* "uvloop/handles/handle.pyx":335
 *         return 0
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "uvloop/handles/handle.pyx":304
 * 
 * 
 * cdef inline bint __ensure_handle_data(uv.uv_handle_t* handle,             # <<<<<<<<<<<<<<
 *                                       const char* handle_ctx):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.__ensure_handle_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/handle.pyx":338
 * 
 * 
 * cdef void __uv_close_handle_cb(uv.uv_handle_t* handle) noexcept with gil:             # <<<<<<<<<<<<<<
 *     cdef UVHandle h
 * 
 */

static void __pyx_f_6uvloop_4loop___uv_close_handle_cb(uv_handle_t *__pyx_v_handle) {
  struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_h = 0;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_close_handle_cb", 0);

  /* "uvloop/handles/handle.pyx":341
 *     cdef UVHandle h
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         # The original UVHandle is long dead. Just free the mem of
 *         # the uv_handle_t* handler.
 */
  __pyx_t_1 = (__pyx_v_handle->data == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/handle.pyx":345
 *         # the uv_handle_t* handler.
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if handle.loop == NULL or handle.loop.data == NULL:
 *                 raise RuntimeError(
 */
    __pyx_t_1 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_1) {

      /* "uvloop/handles/handle.pyx":346
 * 
 *         if UVLOOP_DEBUG:
 *             if handle.loop == NULL or handle.loop.data == NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '__uv_close_handle_cb: handle.loop is invalid')
 */
      __pyx_t_2 = (__pyx_v_handle->loop == NULL);
      if (!__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L6_bool_binop_done;
      }
      __pyx_t_2 = (__pyx_v_handle->loop->data == NULL);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L6_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "uvloop/handles/handle.pyx":347
 *         if UVLOOP_DEBUG:
 *             if handle.loop == NULL or handle.loop.data == NULL:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '__uv_close_handle_cb: handle.loop is invalid')
 *             (<Loop>handle.loop.data)._debug_uv_handles_freed += 1
 */
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__138, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(2, 347, __pyx_L1_error)

        /* "uvloop/handles/handle.pyx":346
 * 
 *         if UVLOOP_DEBUG:
 *             if handle.loop == NULL or handle.loop.data == NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '__uv_close_handle_cb: handle.loop is invalid')
 */
      }

      /* "uvloop/handles/handle.pyx":349
 *                 raise RuntimeError(
 *                     '__uv_close_handle_cb: handle.loop is invalid')
 *             (<Loop>handle.loop.data)._debug_uv_handles_freed += 1             # <<<<<<<<<<<<<<
 * 
 *         PyMem_RawFree(handle)
 */
      __Pyx_INCREF((PyObject *)((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_handle->loop->data));
      __pyx_t_4 = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_v_handle->loop->data);
      __pyx_t_4->_debug_uv_handles_freed = (__pyx_t_4->_debug_uv_handles_freed + 1);
      __Pyx_DECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/handles/handle.pyx":345
 *         # the uv_handle_t* handler.
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if handle.loop == NULL or handle.loop.data == NULL:
 *                 raise RuntimeError(
 */
    }

    /* "uvloop/handles/handle.pyx":351
 *             (<Loop>handle.loop.data)._debug_uv_handles_freed += 1
 * 
 *         PyMem_RawFree(handle)             # <<<<<<<<<<<<<<
 *     else:
 *         h = <UVHandle>handle.data
 */
    PyMem_RawFree(__pyx_v_handle);

    /* "uvloop/handles/handle.pyx":341
 *     cdef UVHandle h
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         # The original UVHandle is long dead. Just free the mem of
 *         # the uv_handle_t* handler.
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/handle.pyx":353
 *         PyMem_RawFree(handle)
 *     else:
 *         h = <UVHandle>handle.data             # <<<<<<<<<<<<<<
 *         try:
 *             if UVLOOP_DEBUG:
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_h = ((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/handles/handle.pyx":354
 *     else:
 *         h = <UVHandle>handle.data
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 if not h._has_handle:
 */
    /*try:*/ {

      /* "uvloop/handles/handle.pyx":355
 *         h = <UVHandle>handle.data
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if not h._has_handle:
 *                     raise RuntimeError(
 */
      __pyx_t_1 = (UVLOOP_DEBUG != 0);
      if (__pyx_t_1) {

        /* "uvloop/handles/handle.pyx":356
 *         try:
 *             if UVLOOP_DEBUG:
 *                 if not h._has_handle:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         'has_handle=0 in __uv_close_handle_cb')
 */
        __pyx_t_1 = (!__pyx_v_h->_has_handle);
        if (unlikely(__pyx_t_1)) {

          /* "uvloop/handles/handle.pyx":357
 *             if UVLOOP_DEBUG:
 *                 if not h._has_handle:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         'has_handle=0 in __uv_close_handle_cb')
 *                 h._loop._debug_handles_closed.update([
 */
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__139, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 357, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(2, 357, __pyx_L9_error)

          /* "uvloop/handles/handle.pyx":356
 *         try:
 *             if UVLOOP_DEBUG:
 *                 if not h._has_handle:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         'has_handle=0 in __uv_close_handle_cb')
 */
        }

        /* "uvloop/handles/handle.pyx":359
 *                     raise RuntimeError(
 *                         'has_handle=0 in __uv_close_handle_cb')
 *                 h._loop._debug_handles_closed.update([             # <<<<<<<<<<<<<<
 *                     h.__class__.__name__])
 *             h._free()
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_h->_loop->_debug_handles_closed, __pyx_n_s_update); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 359, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_5);

        /* "uvloop/handles/handle.pyx":360
 *                         'has_handle=0 in __uv_close_handle_cb')
 *                 h._loop._debug_handles_closed.update([
 *                     h.__class__.__name__])             # <<<<<<<<<<<<<<
 *             h._free()
 *         finally:
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_h), __pyx_n_s_class); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 360, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_name); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 360, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "uvloop/handles/handle.pyx":359
 *                     raise RuntimeError(
 *                         'has_handle=0 in __uv_close_handle_cb')
 *                 h._loop._debug_handles_closed.update([             # <<<<<<<<<<<<<<
 *                     h.__class__.__name__])
 *             h._free()
 */
        __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 359, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_7)) __PYX_ERR(2, 359, __pyx_L9_error);
        __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 359, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "uvloop/handles/handle.pyx":355
 *         h = <UVHandle>handle.data
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if not h._has_handle:
 *                     raise RuntimeError(
 */
      }

      /* "uvloop/handles/handle.pyx":361
 *                 h._loop._debug_handles_closed.update([
 *                     h.__class__.__name__])
 *             h._free()             # <<<<<<<<<<<<<<
 *         finally:
 *             Py_DECREF(h)  # Was INCREFed in UVHandle._close
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_h->__pyx_vtab)->_free(__pyx_v_h); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 361, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "uvloop/handles/handle.pyx":363
 *             h._free()
 *         finally:
 *             Py_DECREF(h)  # Was INCREFed in UVHandle._close             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*finally:*/ {
      /*normal exit:*/{
        Py_DECREF(((PyObject *)__pyx_v_h));
        goto __pyx_L10;
      }
      __pyx_L9_error:;
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_17);
        __pyx_t_9 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
        {
          Py_DECREF(((PyObject *)__pyx_v_h));
        }
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
        }
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
        __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
        goto __pyx_L1_error;
      }
      __pyx_L10:;
    }
  }
  __pyx_L3:;

  /* "uvloop/handles/handle.pyx":338
 * 
 * 
 * cdef void __uv_close_handle_cb(uv.uv_handle_t* handle) noexcept with gil:             # <<<<<<<<<<<<<<
 *     cdef UVHandle h
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF((PyObject *)__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_close_handle_cb", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_h);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/handle.pyx":366
 * 
 * 
 * cdef void __close_all_handles(Loop loop) noexcept:             # <<<<<<<<<<<<<<
 *     uv.uv_walk(loop.uvloop,
 *                __uv_walk_close_all_handles_cb,
 */

static void __pyx_f_6uvloop_4loop___close_all_handles(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop) {

  /* "uvloop/handles/handle.pyx":367
 * 
 * cdef void __close_all_handles(Loop loop) noexcept:
 *     uv.uv_walk(loop.uvloop,             # <<<<<<<<<<<<<<
 *                __uv_walk_close_all_handles_cb,
 *                <void*>loop)  # void
 */
  uv_walk(__pyx_v_loop->uvloop, __pyx_f_6uvloop_4loop___uv_walk_close_all_handles_cb, ((void *)__pyx_v_loop));

  /* "uvloop/handles/handle.pyx":366
 * 
 * 
 * cdef void __close_all_handles(Loop loop) noexcept:             # <<<<<<<<<<<<<<
 *     uv.uv_walk(loop.uvloop,
 *                __uv_walk_close_all_handles_cb,
 */

  /* function exit code */
}

/* "uvloop/handles/handle.pyx":372
 * 
 * 
 * cdef void __uv_walk_close_all_handles_cb(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* handle,
 *     void* arg,
 */

static void __pyx_f_6uvloop_4loop___uv_walk_close_all_handles_cb(uv_handle_t *__pyx_v_handle, void *__pyx_v_arg) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  struct __pyx_obj_6uvloop_4loop_UVHandle *__pyx_v_h = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_walk_close_all_handles_cb", 0);

  /* "uvloop/handles/handle.pyx":378
 * 
 *     cdef:
 *         Loop loop = <Loop>arg             # <<<<<<<<<<<<<<
 *         UVHandle h
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_arg);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/handle.pyx":381
 *         UVHandle h
 * 
 *     if uv.uv_is_closing(handle):             # <<<<<<<<<<<<<<
 *         # The handle is closed or is closing.
 *         return
 */
  __pyx_t_2 = (uv_is_closing(__pyx_v_handle) != 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/handle.pyx":383
 *     if uv.uv_is_closing(handle):
 *         # The handle is closed or is closing.
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if handle.data is NULL:
 */
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":381
 *         UVHandle h
 * 
 *     if uv.uv_is_closing(handle):             # <<<<<<<<<<<<<<
 *         # The handle is closed or is closing.
 *         return
 */
  }

  /* "uvloop/handles/handle.pyx":385
 *         return
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         # This shouldn't happen. Ever.
 *         loop.call_exception_handler({
 */
  __pyx_t_2 = (__pyx_v_handle->data == NULL);
  if (__pyx_t_2) {

    /* "uvloop/handles/handle.pyx":387
 *     if handle.data is NULL:
 *         # This shouldn't happen. Ever.
 *         loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *             'message': 'handle.data is NULL in __close_all_handles_cb'
 *         })
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/handle.pyx":388
 *         # This shouldn't happen. Ever.
 *         loop.call_exception_handler({
 *             'message': 'handle.data is NULL in __close_all_handles_cb'             # <<<<<<<<<<<<<<
 *         })
 *         return
 */
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_u_message, __pyx_kp_u_handle_data_is_NULL_in___close_a) < 0) __PYX_ERR(2, 388, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/handle.pyx":390
 *             'message': 'handle.data is NULL in __close_all_handles_cb'
 *         })
 *         return             # <<<<<<<<<<<<<<
 * 
 *     h = <UVHandle>handle.data
 */
    goto __pyx_L0;

    /* "uvloop/handles/handle.pyx":385
 *         return
 * 
 *     if handle.data is NULL:             # <<<<<<<<<<<<<<
 *         # This shouldn't happen. Ever.
 *         loop.call_exception_handler({
 */
  }

  /* "uvloop/handles/handle.pyx":392
 *         return
 * 
 *     h = <UVHandle>handle.data             # <<<<<<<<<<<<<<
 *     if not h._closed:
 *         h._warn_unclosed()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_h = ((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/handle.pyx":393
 * 
 *     h = <UVHandle>handle.data
 *     if not h._closed:             # <<<<<<<<<<<<<<
 *         h._warn_unclosed()
 *         h._close()
 */
  __pyx_t_2 = (!__pyx_v_h->_closed);
  if (__pyx_t_2) {

    /* "uvloop/handles/handle.pyx":394
 *     h = <UVHandle>handle.data
 *     if not h._closed:
 *         h._warn_unclosed()             # <<<<<<<<<<<<<<
 *         h._close()
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_h->__pyx_vtab)->_warn_unclosed(__pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/handle.pyx":395
 *     if not h._closed:
 *         h._warn_unclosed()
 *         h._close()             # <<<<<<<<<<<<<<
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVHandle *)__pyx_v_h->__pyx_vtab)->_close(__pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/handle.pyx":393
 * 
 *     h = <UVHandle>handle.data
 *     if not h._closed:             # <<<<<<<<<<<<<<
 *         h._warn_unclosed()
 *         h._close()
 */
  }

  /* "uvloop/handles/handle.pyx":372
 * 
 * 
 * cdef void __uv_walk_close_all_handles_cb(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* handle,
 *     void* arg,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_walk_close_all_handles_cb", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF((PyObject *)__pyx_v_h);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/async_.pyx":3
 * @cython.no_gc_clear
 * cdef class UVAsync(UVHandle):
 *     cdef _init(self, Loop loop, method_t callback, object ctx):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_7UVAsync__init(struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/async_.pyx":6
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_async_t))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/async_.pyx":8
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_async_t))             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             self._abort_init()
 */
  __pyx_v_self->__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_async_t))));

  /* "uvloop/handles/async_.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_async_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/async_.pyx":10
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_async_t))
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/async_.pyx":11
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_async_init(self._loop.uvloop,
 */
    PyErr_NoMemory(); __PYX_ERR(13, 11, __pyx_L1_error)

    /* "uvloop/handles/async_.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_async_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/async_.pyx":13
 *             raise MemoryError()
 * 
 *         err = uv.uv_async_init(self._loop.uvloop,             # <<<<<<<<<<<<<<
 *                                <uv.uv_async_t*>self._handle,
 *                                __uvasync_callback)
 */
  __pyx_v_err = uv_async_init(__pyx_v_self->__pyx_base._loop->uvloop, ((uv_async_t *)__pyx_v_self->__pyx_base._handle), __pyx_f_6uvloop_4loop___uvasync_callback);

  /* "uvloop/handles/async_.pyx":16
 *                                <uv.uv_async_t*>self._handle,
 *                                __uvasync_callback)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/async_.pyx":17
 *                                __uvasync_callback)
 *         if err < 0:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/async_.pyx":18
 *         if err < 0:
 *             self._abort_init()
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         self._finish_init()
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 18, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(13, 18, __pyx_L1_error)

    /* "uvloop/handles/async_.pyx":16
 *                                <uv.uv_async_t*>self._handle,
 *                                __uvasync_callback)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  }

  /* "uvloop/handles/async_.pyx":20
 *             raise convert_error(err)
 * 
 *         self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *         self.callback = callback
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/async_.pyx":22
 *         self._finish_init()
 * 
 *         self.callback = callback             # <<<<<<<<<<<<<<
 *         self.ctx = ctx
 * 
 */
  __pyx_v_self->callback = __pyx_v_callback;

  /* "uvloop/handles/async_.pyx":23
 * 
 *         self.callback = callback
 *         self.ctx = ctx             # <<<<<<<<<<<<<<
 * 
 *     cdef send(self):
 */
  __Pyx_INCREF(__pyx_v_ctx);
  __Pyx_GIVEREF(__pyx_v_ctx);
  __Pyx_GOTREF(__pyx_v_self->ctx);
  __Pyx_DECREF(__pyx_v_self->ctx);
  __pyx_v_self->ctx = __pyx_v_ctx;

  /* "uvloop/handles/async_.pyx":3
 * @cython.no_gc_clear
 * cdef class UVAsync(UVHandle):
 *     cdef _init(self, Loop loop, method_t callback, object ctx):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVAsync._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/async_.pyx":25
 *         self.ctx = ctx
 * 
 *     cdef send(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_7UVAsync_send(struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("send", 1);

  /* "uvloop/handles/async_.pyx":28
 *         cdef int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_async_send(<uv.uv_async_t*>self._handle)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/async_.pyx":30
 *         self._ensure_alive()
 * 
 *         err = uv.uv_async_send(<uv.uv_async_t*>self._handle)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  __pyx_v_err = uv_async_send(((uv_async_t *)__pyx_v_self->__pyx_base._handle));

  /* "uvloop/handles/async_.pyx":31
 * 
 *         err = uv.uv_async_send(<uv.uv_async_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/async_.pyx":32
 *         err = uv.uv_async_send(<uv.uv_async_t*>self._handle)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/async_.pyx":33
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/async_.pyx":34
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/async_.pyx":31
 * 
 *         err = uv.uv_async_send(<uv.uv_async_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/async_.pyx":25
 *         self.ctx = ctx
 * 
 *     cdef send(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVAsync.send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/async_.pyx":37
 * 
 *     @staticmethod
 *     cdef UVAsync new(Loop loop, method_t callback, object ctx):             # <<<<<<<<<<<<<<
 *         cdef UVAsync handle
 *         handle = UVAsync.__new__(UVAsync)
 */

static struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_f_6uvloop_4loop_7UVAsync_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx) {
  struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/async_.pyx":39
 *     cdef UVAsync new(Loop loop, method_t callback, object ctx):
 *         cdef UVAsync handle
 *         handle = UVAsync.__new__(UVAsync)             # <<<<<<<<<<<<<<
 *         handle._init(loop, callback, ctx)
 *         return handle
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UVAsync(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UVAsync), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 39, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UVAsync *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/async_.pyx":40
 *         cdef UVAsync handle
 *         handle = UVAsync.__new__(UVAsync)
 *         handle._init(loop, callback, ctx)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_init(__pyx_v_handle, __pyx_v_loop, __pyx_v_callback, __pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/async_.pyx":41
 *         handle = UVAsync.__new__(UVAsync)
 *         handle._init(loop, callback, ctx)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/async_.pyx":37
 * 
 *     @staticmethod
 *     cdef UVAsync new(Loop loop, method_t callback, object ctx):             # <<<<<<<<<<<<<<
 *         cdef UVAsync handle
 *         handle = UVAsync.__new__(UVAsync)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVAsync.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_7UVAsync_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_7UVAsync___reduce_cython__, "UVAsync.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_7UVAsync_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVAsync_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVAsync___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_7UVAsync_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_7UVAsync___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVAsync *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_7UVAsync___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVAsync.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_7UVAsync_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_7UVAsync_2__setstate_cython__, "UVAsync.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_7UVAsync_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVAsync_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVAsync_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_7UVAsync_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVAsync.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_7UVAsync_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVAsync *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_7UVAsync_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVAsync.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/async_.pyx":44
 * 
 * 
 * cdef void __uvasync_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_async_t* handle,
 * ) noexcept with gil:
 */

static void __pyx_f_6uvloop_4loop___uvasync_callback(uv_async_t *__pyx_v_handle) {
  struct __pyx_obj_6uvloop_4loop_UVAsync *__pyx_v_async_ = 0;
  __pyx_t_6uvloop_4loop_method_t __pyx_v_cb;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_6uvloop_4loop_method_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uvasync_callback", 0);

  /* "uvloop/handles/async_.pyx":47
 *     uv.uv_async_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVAsync callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVAsync callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(13, 47, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/async_.pyx":48
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVAsync callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/async_.pyx":47
 *     uv.uv_async_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVAsync callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/handles/async_.pyx":51
 * 
 *     cdef:
 *         UVAsync async_ = <UVAsync> handle.data             # <<<<<<<<<<<<<<
 *         method_t cb = async_.callback
 *     try:
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_async_ = ((struct __pyx_obj_6uvloop_4loop_UVAsync *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/async_.pyx":52
 *     cdef:
 *         UVAsync async_ = <UVAsync> handle.data
 *         method_t cb = async_.callback             # <<<<<<<<<<<<<<
 *     try:
 *         cb(async_.ctx)
 */
  __pyx_t_4 = __pyx_v_async_->callback;
  __pyx_v_cb = __pyx_t_4;

  /* "uvloop/handles/async_.pyx":53
 *         UVAsync async_ = <UVAsync> handle.data
 *         method_t cb = async_.callback
 *     try:             # <<<<<<<<<<<<<<
 *         cb(async_.ctx)
 *     except BaseException as ex:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "uvloop/handles/async_.pyx":54
 *         method_t cb = async_.callback
 *     try:
 *         cb(async_.ctx)             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         async_._error(ex, False)
 */
      __pyx_t_3 = __pyx_v_async_->ctx;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_8 = __pyx_v_cb(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(13, 54, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/handles/async_.pyx":53
 *         UVAsync async_ = <UVAsync> handle.data
 *         method_t cb = async_.callback
 *     try:             # <<<<<<<<<<<<<<
 *         cb(async_.ctx)
 *     except BaseException as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/handles/async_.pyx":55
 *     try:
 *         cb(async_.ctx)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         async_._error(ex, False)
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.__uvasync_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_3, &__pyx_t_10) < 0) __PYX_ERR(13, 55, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_ex = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/handles/async_.pyx":56
 *         cb(async_.ctx)
 *     except BaseException as ex:
 *         async_._error(ex, False)             # <<<<<<<<<<<<<<
 */
        __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVAsync *)__pyx_v_async_->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_async_), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_11)) __PYX_ERR(13, 56, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }

      /* "uvloop/handles/async_.pyx":55
 *     try:
 *         cb(async_.ctx)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         async_._error(ex, False)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_9 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L6_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/handles/async_.pyx":53
 *         UVAsync async_ = <UVAsync> handle.data
 *         method_t cb = async_.callback
 *     try:             # <<<<<<<<<<<<<<
 *         cb(async_.ctx)
 *     except BaseException as ex:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    __pyx_L9_try_end:;
  }

  /* "uvloop/handles/async_.pyx":44
 * 
 * 
 * cdef void __uvasync_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_async_t* handle,
 * ) noexcept with gil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_WriteUnraisable("uvloop.loop.__uvasync_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_async_);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/idle.pyx":3
 * @cython.no_gc_clear
 * cdef class UVIdle(UVHandle):
 *     cdef _init(self, Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVIdle__init(struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/idle.pyx":6
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_idle_t))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/idle.pyx":8
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_idle_t))             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             self._abort_init()
 */
  __pyx_v_self->__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_idle_t))));

  /* "uvloop/handles/idle.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_idle_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/idle.pyx":10
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_idle_t))
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/idle.pyx":11
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_idle_init(self._loop.uvloop, <uv.uv_idle_t*>self._handle)
 */
    PyErr_NoMemory(); __PYX_ERR(14, 11, __pyx_L1_error)

    /* "uvloop/handles/idle.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_idle_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/idle.pyx":13
 *             raise MemoryError()
 * 
 *         err = uv.uv_idle_init(self._loop.uvloop, <uv.uv_idle_t*>self._handle)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             self._abort_init()
 */
  __pyx_v_err = uv_idle_init(__pyx_v_self->__pyx_base._loop->uvloop, ((uv_idle_t *)__pyx_v_self->__pyx_base._handle));

  /* "uvloop/handles/idle.pyx":14
 * 
 *         err = uv.uv_idle_init(self._loop.uvloop, <uv.uv_idle_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/idle.pyx":15
 *         err = uv.uv_idle_init(self._loop.uvloop, <uv.uv_idle_t*>self._handle)
 *         if err < 0:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/idle.pyx":16
 *         if err < 0:
 *             self._abort_init()
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         self._finish_init()
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 16, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(14, 16, __pyx_L1_error)

    /* "uvloop/handles/idle.pyx":14
 * 
 *         err = uv.uv_idle_init(self._loop.uvloop, <uv.uv_idle_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  }

  /* "uvloop/handles/idle.pyx":18
 *             raise convert_error(err)
 * 
 *         self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *         self.h = h
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/idle.pyx":20
 *         self._finish_init()
 * 
 *         self.h = h             # <<<<<<<<<<<<<<
 *         self.running = 0
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_h);
  __Pyx_GIVEREF((PyObject *)__pyx_v_h);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->h);
  __Pyx_DECREF((PyObject *)__pyx_v_self->h);
  __pyx_v_self->h = __pyx_v_h;

  /* "uvloop/handles/idle.pyx":21
 * 
 *         self.h = h
 *         self.running = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline stop(self):
 */
  __pyx_v_self->running = 0;

  /* "uvloop/handles/idle.pyx":3
 * @cython.no_gc_clear
 * cdef class UVIdle(UVHandle):
 *     cdef _init(self, Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVIdle._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/idle.pyx":23
 *         self.running = 0
 * 
 *     cdef inline stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVIdle_stop(struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop", 1);

  /* "uvloop/handles/idle.pyx":26
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(14, 26, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/handles/idle.pyx":27
 * 
 *         if not self._is_alive():
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/idle.pyx":28
 *         if not self._is_alive():
 *             self.running = 0
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 1:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/idle.pyx":26
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  }

  /* "uvloop/handles/idle.pyx":30
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_idle_stop(<uv.uv_idle_t*>self._handle)
 *             self.running = 0
 */
  __pyx_t_2 = (__pyx_v_self->running == 1);
  if (__pyx_t_2) {

    /* "uvloop/handles/idle.pyx":31
 * 
 *         if self.running == 1:
 *             err = uv.uv_idle_stop(<uv.uv_idle_t*>self._handle)             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             if err < 0:
 */
    __pyx_v_err = uv_idle_stop(((uv_idle_t *)__pyx_v_self->__pyx_base._handle));

    /* "uvloop/handles/idle.pyx":32
 *         if self.running == 1:
 *             err = uv.uv_idle_stop(<uv.uv_idle_t*>self._handle)
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             if err < 0:
 *                 exc = convert_error(err)
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/idle.pyx":33
 *             err = uv.uv_idle_stop(<uv.uv_idle_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/idle.pyx":34
 *             self.running = 0
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(14, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "uvloop/handles/idle.pyx":35
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(14, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/idle.pyx":36
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 * 
 *     cdef inline start(self):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/idle.pyx":33
 *             err = uv.uv_idle_stop(<uv.uv_idle_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/idle.pyx":30
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_idle_stop(<uv.uv_idle_t*>self._handle)
 *             self.running = 0
 */
  }

  /* "uvloop/handles/idle.pyx":23
 *         self.running = 0
 * 
 *     cdef inline stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVIdle.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/idle.pyx":38
 *                 return
 * 
 *     cdef inline start(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVIdle_start(struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start", 1);

  /* "uvloop/handles/idle.pyx":41
 *         cdef int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/idle.pyx":43
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             err = uv.uv_idle_start(<uv.uv_idle_t*>self._handle,
 *                                    cb_idle_callback)
 */
  __pyx_t_2 = (__pyx_v_self->running == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/idle.pyx":44
 * 
 *         if self.running == 0:
 *             err = uv.uv_idle_start(<uv.uv_idle_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                    cb_idle_callback)
 *             if err < 0:
 */
    __pyx_v_err = uv_idle_start(((uv_idle_t *)__pyx_v_self->__pyx_base._handle), __pyx_f_6uvloop_4loop_cb_idle_callback);

    /* "uvloop/handles/idle.pyx":46
 *             err = uv.uv_idle_start(<uv.uv_idle_t*>self._handle,
 *                                    cb_idle_callback)
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/idle.pyx":47
 *                                    cb_idle_callback)
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_exc = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/handles/idle.pyx":48
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 *             self.running = 1
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 48, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/idle.pyx":49
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 *             self.running = 1
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/idle.pyx":46
 *             err = uv.uv_idle_start(<uv.uv_idle_t*>self._handle,
 *                                    cb_idle_callback)
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/idle.pyx":50
 *                 self._fatal_error(exc, True)
 *                 return
 *             self.running = 1             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_v_self->running = 1;

    /* "uvloop/handles/idle.pyx":43
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             err = uv.uv_idle_start(<uv.uv_idle_t*>self._handle,
 *                                    cb_idle_callback)
 */
  }

  /* "uvloop/handles/idle.pyx":38
 *                 return
 * 
 *     cdef inline start(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVIdle.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/idle.pyx":53
 * 
 *     @staticmethod
 *     cdef UVIdle new(Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef UVIdle handle
 *         handle = UVIdle.__new__(UVIdle)
 */

static struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_f_6uvloop_4loop_6UVIdle_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h) {
  struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/idle.pyx":55
 *     cdef UVIdle new(Loop loop, Handle h):
 *         cdef UVIdle handle
 *         handle = UVIdle.__new__(UVIdle)             # <<<<<<<<<<<<<<
 *         handle._init(loop, h)
 *         return handle
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UVIdle(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UVIdle), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 55, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UVIdle *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/idle.pyx":56
 *         cdef UVIdle handle
 *         handle = UVIdle.__new__(UVIdle)
 *         handle._init(loop, h)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_init(__pyx_v_handle, __pyx_v_loop, __pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/idle.pyx":57
 *         handle = UVIdle.__new__(UVIdle)
 *         handle._init(loop, h)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/idle.pyx":53
 * 
 *     @staticmethod
 *     cdef UVIdle new(Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef UVIdle handle
 *         handle = UVIdle.__new__(UVIdle)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVIdle.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6UVIdle_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6UVIdle___reduce_cython__, "UVIdle.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6UVIdle_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVIdle_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVIdle___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6UVIdle_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6UVIdle___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVIdle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6UVIdle___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVIdle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6UVIdle_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6UVIdle_2__setstate_cython__, "UVIdle.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6UVIdle_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVIdle_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVIdle_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6UVIdle_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVIdle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_6UVIdle_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVIdle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6UVIdle_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVIdle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/idle.pyx":60
 * 
 * 
 * cdef void cb_idle_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_idle_t* handle,
 * ) noexcept with gil:
 */

static void __pyx_f_6uvloop_4loop_cb_idle_callback(uv_idle_t *__pyx_v_handle) {
  struct __pyx_obj_6uvloop_4loop_UVIdle *__pyx_v_idle = 0;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h = 0;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("cb_idle_callback", 0);

  /* "uvloop/handles/idle.pyx":63
 *     uv.uv_idle_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVIdle callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVIdle callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(14, 63, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/idle.pyx":64
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVIdle callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/idle.pyx":63
 *     uv.uv_idle_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVIdle callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/handles/idle.pyx":67
 * 
 *     cdef:
 *         UVIdle idle = <UVIdle> handle.data             # <<<<<<<<<<<<<<
 *         Handle h = idle.h
 *     try:
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_idle = ((struct __pyx_obj_6uvloop_4loop_UVIdle *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/idle.pyx":68
 *     cdef:
 *         UVIdle idle = <UVIdle> handle.data
 *         Handle h = idle.h             # <<<<<<<<<<<<<<
 *     try:
 *         h._run()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_idle->h);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_h = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/idle.pyx":69
 *         UVIdle idle = <UVIdle> handle.data
 *         Handle h = idle.h
 *     try:             # <<<<<<<<<<<<<<
 *         h._run()
 *     except BaseException as ex:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "uvloop/handles/idle.pyx":70
 *         Handle h = idle.h
 *     try:
 *         h._run()             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         idle._error(ex, False)
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_h); if (unlikely(!__pyx_t_3)) __PYX_ERR(14, 70, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/idle.pyx":69
 *         UVIdle idle = <UVIdle> handle.data
 *         Handle h = idle.h
 *     try:             # <<<<<<<<<<<<<<
 *         h._run()
 *     except BaseException as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/idle.pyx":71
 *     try:
 *         h._run()
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         idle._error(ex, False)
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("uvloop.loop.cb_idle_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(14, 71, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_ex = __pyx_t_8;
      /*try:*/ {

        /* "uvloop/handles/idle.pyx":72
 *         h._run()
 *     except BaseException as ex:
 *         idle._error(ex, False)             # <<<<<<<<<<<<<<
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVIdle *)__pyx_v_idle->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_idle), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_10)) __PYX_ERR(14, 72, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }

      /* "uvloop/handles/idle.pyx":71
 *     try:
 *         h._run()
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         idle._error(ex, False)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_7 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L6_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/handles/idle.pyx":69
 *         UVIdle idle = <UVIdle> handle.data
 *         Handle h = idle.h
 *     try:             # <<<<<<<<<<<<<<
 *         h._run()
 *     except BaseException as ex:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_L9_try_end:;
  }

  /* "uvloop/handles/idle.pyx":60
 * 
 * 
 * cdef void cb_idle_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_idle_t* handle,
 * ) noexcept with gil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("uvloop.loop.cb_idle_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_idle);
  __Pyx_XDECREF((PyObject *)__pyx_v_h);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/check.pyx":3
 * @cython.no_gc_clear
 * cdef class UVCheck(UVHandle):
 *     cdef _init(self, Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_7UVCheck__init(struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/check.pyx":6
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_check_t))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/check.pyx":8
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_check_t))             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             self._abort_init()
 */
  __pyx_v_self->__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_check_t))));

  /* "uvloop/handles/check.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_check_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/check.pyx":10
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_check_t))
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/check.pyx":11
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_check_init(self._loop.uvloop, <uv.uv_check_t*>self._handle)
 */
    PyErr_NoMemory(); __PYX_ERR(15, 11, __pyx_L1_error)

    /* "uvloop/handles/check.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_check_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/check.pyx":13
 *             raise MemoryError()
 * 
 *         err = uv.uv_check_init(self._loop.uvloop, <uv.uv_check_t*>self._handle)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             self._abort_init()
 */
  __pyx_v_err = uv_check_init(__pyx_v_self->__pyx_base._loop->uvloop, ((uv_check_t *)__pyx_v_self->__pyx_base._handle));

  /* "uvloop/handles/check.pyx":14
 * 
 *         err = uv.uv_check_init(self._loop.uvloop, <uv.uv_check_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/check.pyx":15
 *         err = uv.uv_check_init(self._loop.uvloop, <uv.uv_check_t*>self._handle)
 *         if err < 0:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/check.pyx":16
 *         if err < 0:
 *             self._abort_init()
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         self._finish_init()
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 16, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(15, 16, __pyx_L1_error)

    /* "uvloop/handles/check.pyx":14
 * 
 *         err = uv.uv_check_init(self._loop.uvloop, <uv.uv_check_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  }

  /* "uvloop/handles/check.pyx":18
 *             raise convert_error(err)
 * 
 *         self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *         self.h = h
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/check.pyx":20
 *         self._finish_init()
 * 
 *         self.h = h             # <<<<<<<<<<<<<<
 *         self.running = 0
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_h);
  __Pyx_GIVEREF((PyObject *)__pyx_v_h);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->h);
  __Pyx_DECREF((PyObject *)__pyx_v_self->h);
  __pyx_v_self->h = __pyx_v_h;

  /* "uvloop/handles/check.pyx":21
 * 
 *         self.h = h
 *         self.running = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline stop(self):
 */
  __pyx_v_self->running = 0;

  /* "uvloop/handles/check.pyx":3
 * @cython.no_gc_clear
 * cdef class UVCheck(UVHandle):
 *     cdef _init(self, Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVCheck._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/check.pyx":23
 *         self.running = 0
 * 
 *     cdef inline stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_7UVCheck_stop(struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop", 1);

  /* "uvloop/handles/check.pyx":26
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(15, 26, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/handles/check.pyx":27
 * 
 *         if not self._is_alive():
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/check.pyx":28
 *         if not self._is_alive():
 *             self.running = 0
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 1:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/check.pyx":26
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  }

  /* "uvloop/handles/check.pyx":30
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_check_stop(<uv.uv_check_t*>self._handle)
 *             self.running = 0
 */
  __pyx_t_2 = (__pyx_v_self->running == 1);
  if (__pyx_t_2) {

    /* "uvloop/handles/check.pyx":31
 * 
 *         if self.running == 1:
 *             err = uv.uv_check_stop(<uv.uv_check_t*>self._handle)             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             if err < 0:
 */
    __pyx_v_err = uv_check_stop(((uv_check_t *)__pyx_v_self->__pyx_base._handle));

    /* "uvloop/handles/check.pyx":32
 *         if self.running == 1:
 *             err = uv.uv_check_stop(<uv.uv_check_t*>self._handle)
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             if err < 0:
 *                 exc = convert_error(err)
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/check.pyx":33
 *             err = uv.uv_check_stop(<uv.uv_check_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/check.pyx":34
 *             self.running = 0
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(15, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "uvloop/handles/check.pyx":35
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(15, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/check.pyx":36
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 * 
 *     cdef inline start(self):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/check.pyx":33
 *             err = uv.uv_check_stop(<uv.uv_check_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/check.pyx":30
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_check_stop(<uv.uv_check_t*>self._handle)
 *             self.running = 0
 */
  }

  /* "uvloop/handles/check.pyx":23
 *         self.running = 0
 * 
 *     cdef inline stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVCheck.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/check.pyx":38
 *                 return
 * 
 *     cdef inline start(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_7UVCheck_start(struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start", 1);

  /* "uvloop/handles/check.pyx":41
 *         cdef int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/check.pyx":43
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             err = uv.uv_check_start(<uv.uv_check_t*>self._handle,
 *                                     cb_check_callback)
 */
  __pyx_t_2 = (__pyx_v_self->running == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/check.pyx":44
 * 
 *         if self.running == 0:
 *             err = uv.uv_check_start(<uv.uv_check_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                     cb_check_callback)
 *             if err < 0:
 */
    __pyx_v_err = uv_check_start(((uv_check_t *)__pyx_v_self->__pyx_base._handle), __pyx_f_6uvloop_4loop_cb_check_callback);

    /* "uvloop/handles/check.pyx":46
 *             err = uv.uv_check_start(<uv.uv_check_t*>self._handle,
 *                                     cb_check_callback)
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/check.pyx":47
 *                                     cb_check_callback)
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_exc = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/handles/check.pyx":48
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 *             self.running = 1
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 48, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/check.pyx":49
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 *             self.running = 1
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/check.pyx":46
 *             err = uv.uv_check_start(<uv.uv_check_t*>self._handle,
 *                                     cb_check_callback)
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/check.pyx":50
 *                 self._fatal_error(exc, True)
 *                 return
 *             self.running = 1             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_v_self->running = 1;

    /* "uvloop/handles/check.pyx":43
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             err = uv.uv_check_start(<uv.uv_check_t*>self._handle,
 *                                     cb_check_callback)
 */
  }

  /* "uvloop/handles/check.pyx":38
 *                 return
 * 
 *     cdef inline start(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVCheck.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/check.pyx":53
 * 
 *     @staticmethod
 *     cdef UVCheck new(Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef UVCheck handle
 *         handle = UVCheck.__new__(UVCheck)
 */

static struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_f_6uvloop_4loop_7UVCheck_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h) {
  struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/check.pyx":55
 *     cdef UVCheck new(Loop loop, Handle h):
 *         cdef UVCheck handle
 *         handle = UVCheck.__new__(UVCheck)             # <<<<<<<<<<<<<<
 *         handle._init(loop, h)
 *         return handle
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UVCheck(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UVCheck), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 55, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UVCheck *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/check.pyx":56
 *         cdef UVCheck handle
 *         handle = UVCheck.__new__(UVCheck)
 *         handle._init(loop, h)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_init(__pyx_v_handle, __pyx_v_loop, __pyx_v_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/check.pyx":57
 *         handle = UVCheck.__new__(UVCheck)
 *         handle._init(loop, h)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/check.pyx":53
 * 
 *     @staticmethod
 *     cdef UVCheck new(Loop loop, Handle h):             # <<<<<<<<<<<<<<
 *         cdef UVCheck handle
 *         handle = UVCheck.__new__(UVCheck)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVCheck.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_7UVCheck_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_7UVCheck___reduce_cython__, "UVCheck.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_7UVCheck_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVCheck_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVCheck___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_7UVCheck_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_7UVCheck___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVCheck *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_7UVCheck___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVCheck.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_7UVCheck_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_7UVCheck_2__setstate_cython__, "UVCheck.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_7UVCheck_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVCheck_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVCheck_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_7UVCheck_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVCheck.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_7UVCheck_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVCheck *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_7UVCheck_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVCheck.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/check.pyx":60
 * 
 * 
 * cdef void cb_check_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_check_t* handle,
 * ) noexcept with gil:
 */

static void __pyx_f_6uvloop_4loop_cb_check_callback(uv_check_t *__pyx_v_handle) {
  struct __pyx_obj_6uvloop_4loop_UVCheck *__pyx_v_check = 0;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h = 0;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("cb_check_callback", 0);

  /* "uvloop/handles/check.pyx":63
 *     uv.uv_check_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVCheck callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVCheck callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(15, 63, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/check.pyx":64
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVCheck callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/check.pyx":63
 *     uv.uv_check_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVCheck callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/handles/check.pyx":67
 * 
 *     cdef:
 *         UVCheck check = <UVCheck> handle.data             # <<<<<<<<<<<<<<
 *         Handle h = check.h
 *     try:
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_check = ((struct __pyx_obj_6uvloop_4loop_UVCheck *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/check.pyx":68
 *     cdef:
 *         UVCheck check = <UVCheck> handle.data
 *         Handle h = check.h             # <<<<<<<<<<<<<<
 *     try:
 *         h._run()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_check->h);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_h = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/check.pyx":69
 *         UVCheck check = <UVCheck> handle.data
 *         Handle h = check.h
 *     try:             # <<<<<<<<<<<<<<
 *         h._run()
 *     except BaseException as ex:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "uvloop/handles/check.pyx":70
 *         Handle h = check.h
 *     try:
 *         h._run()             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         check._error(ex, False)
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_h); if (unlikely(!__pyx_t_3)) __PYX_ERR(15, 70, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/check.pyx":69
 *         UVCheck check = <UVCheck> handle.data
 *         Handle h = check.h
 *     try:             # <<<<<<<<<<<<<<
 *         h._run()
 *     except BaseException as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/check.pyx":71
 *     try:
 *         h._run()
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         check._error(ex, False)
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("uvloop.loop.cb_check_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(15, 71, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_ex = __pyx_t_8;
      /*try:*/ {

        /* "uvloop/handles/check.pyx":72
 *         h._run()
 *     except BaseException as ex:
 *         check._error(ex, False)             # <<<<<<<<<<<<<<
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVCheck *)__pyx_v_check->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_check), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_10)) __PYX_ERR(15, 72, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }

      /* "uvloop/handles/check.pyx":71
 *     try:
 *         h._run()
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         check._error(ex, False)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_7 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L6_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/handles/check.pyx":69
 *         UVCheck check = <UVCheck> handle.data
 *         Handle h = check.h
 *     try:             # <<<<<<<<<<<<<<
 *         h._run()
 *     except BaseException as ex:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_L9_try_end:;
  }

  /* "uvloop/handles/check.pyx":60
 * 
 * 
 * cdef void cb_check_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_check_t* handle,
 * ) noexcept with gil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("uvloop.loop.cb_check_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_check);
  __Pyx_XDECREF((PyObject *)__pyx_v_h);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/timer.pyx":3
 * @cython.no_gc_clear
 * cdef class UVTimer(UVHandle):
 *     cdef _init(self, Loop loop, method_t callback, object ctx,             # <<<<<<<<<<<<<<
 *                uint64_t timeout):
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_7UVTimer__init(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx, uint64_t __pyx_v_timeout) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/timer.pyx":8
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*> PyMem_RawMalloc(sizeof(uv.uv_timer_t))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/timer.pyx":10
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*> PyMem_RawMalloc(sizeof(uv.uv_timer_t))             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             self._abort_init()
 */
  __pyx_v_self->__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_timer_t))));

  /* "uvloop/handles/timer.pyx":11
 * 
 *         self._handle = <uv.uv_handle_t*> PyMem_RawMalloc(sizeof(uv.uv_timer_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/timer.pyx":12
 *         self._handle = <uv.uv_handle_t*> PyMem_RawMalloc(sizeof(uv.uv_timer_t))
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/timer.pyx":13
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_timer_init(self._loop.uvloop, <uv.uv_timer_t*>self._handle)
 */
    PyErr_NoMemory(); __PYX_ERR(16, 13, __pyx_L1_error)

    /* "uvloop/handles/timer.pyx":11
 * 
 *         self._handle = <uv.uv_handle_t*> PyMem_RawMalloc(sizeof(uv.uv_timer_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/timer.pyx":15
 *             raise MemoryError()
 * 
 *         err = uv.uv_timer_init(self._loop.uvloop, <uv.uv_timer_t*>self._handle)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             self._abort_init()
 */
  __pyx_v_err = uv_timer_init(__pyx_v_self->__pyx_base._loop->uvloop, ((uv_timer_t *)__pyx_v_self->__pyx_base._handle));

  /* "uvloop/handles/timer.pyx":16
 * 
 *         err = uv.uv_timer_init(self._loop.uvloop, <uv.uv_timer_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/timer.pyx":17
 *         err = uv.uv_timer_init(self._loop.uvloop, <uv.uv_timer_t*>self._handle)
 *         if err < 0:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/timer.pyx":18
 *         if err < 0:
 *             self._abort_init()
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         self._finish_init()
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 18, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(16, 18, __pyx_L1_error)

    /* "uvloop/handles/timer.pyx":16
 * 
 *         err = uv.uv_timer_init(self._loop.uvloop, <uv.uv_timer_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  }

  /* "uvloop/handles/timer.pyx":20
 *             raise convert_error(err)
 * 
 *         self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *         self.callback = callback
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/timer.pyx":22
 *         self._finish_init()
 * 
 *         self.callback = callback             # <<<<<<<<<<<<<<
 *         self.ctx = ctx
 *         self.running = 0
 */
  __pyx_v_self->callback = __pyx_v_callback;

  /* "uvloop/handles/timer.pyx":23
 * 
 *         self.callback = callback
 *         self.ctx = ctx             # <<<<<<<<<<<<<<
 *         self.running = 0
 *         self.timeout = timeout
 */
  __Pyx_INCREF(__pyx_v_ctx);
  __Pyx_GIVEREF(__pyx_v_ctx);
  __Pyx_GOTREF(__pyx_v_self->ctx);
  __Pyx_DECREF(__pyx_v_self->ctx);
  __pyx_v_self->ctx = __pyx_v_ctx;

  /* "uvloop/handles/timer.pyx":24
 *         self.callback = callback
 *         self.ctx = ctx
 *         self.running = 0             # <<<<<<<<<<<<<<
 *         self.timeout = timeout
 *         self.start_t = 0
 */
  __pyx_v_self->running = 0;

  /* "uvloop/handles/timer.pyx":25
 *         self.ctx = ctx
 *         self.running = 0
 *         self.timeout = timeout             # <<<<<<<<<<<<<<
 *         self.start_t = 0
 * 
 */
  __pyx_v_self->timeout = __pyx_v_timeout;

  /* "uvloop/handles/timer.pyx":26
 *         self.running = 0
 *         self.timeout = timeout
 *         self.start_t = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef stop(self):
 */
  __pyx_v_self->start_t = 0;

  /* "uvloop/handles/timer.pyx":3
 * @cython.no_gc_clear
 * cdef class UVTimer(UVHandle):
 *     cdef _init(self, Loop loop, method_t callback, object ctx,             # <<<<<<<<<<<<<<
 *                uint64_t timeout):
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVTimer._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/timer.pyx":28
 *         self.start_t = 0
 * 
 *     cdef stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_7UVTimer_stop(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop", 1);

  /* "uvloop/handles/timer.pyx":31
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(16, 31, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/handles/timer.pyx":32
 * 
 *         if not self._is_alive():
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/timer.pyx":33
 *         if not self._is_alive():
 *             self.running = 0
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 1:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/timer.pyx":31
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  }

  /* "uvloop/handles/timer.pyx":35
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_timer_stop(<uv.uv_timer_t*>self._handle)
 *             self.running = 0
 */
  __pyx_t_2 = (__pyx_v_self->running == 1);
  if (__pyx_t_2) {

    /* "uvloop/handles/timer.pyx":36
 * 
 *         if self.running == 1:
 *             err = uv.uv_timer_stop(<uv.uv_timer_t*>self._handle)             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             if err < 0:
 */
    __pyx_v_err = uv_timer_stop(((uv_timer_t *)__pyx_v_self->__pyx_base._handle));

    /* "uvloop/handles/timer.pyx":37
 *         if self.running == 1:
 *             err = uv.uv_timer_stop(<uv.uv_timer_t*>self._handle)
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             if err < 0:
 *                 exc = convert_error(err)
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/timer.pyx":38
 *             err = uv.uv_timer_stop(<uv.uv_timer_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/timer.pyx":39
 *             self.running = 0
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "uvloop/handles/timer.pyx":40
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(16, 40, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/timer.pyx":41
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 * 
 *     cdef start(self):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/timer.pyx":38
 *             err = uv.uv_timer_stop(<uv.uv_timer_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/timer.pyx":35
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_timer_stop(<uv.uv_timer_t*>self._handle)
 *             self.running = 0
 */
  }

  /* "uvloop/handles/timer.pyx":28
 *         self.start_t = 0
 * 
 *     cdef stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVTimer.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/timer.pyx":43
 *                 return
 * 
 *     cdef start(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_7UVTimer_start(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start", 1);

  /* "uvloop/handles/timer.pyx":46
 *         cdef int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/timer.pyx":48
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             # Update libuv internal time.
 *             uv.uv_update_time(self._loop.uvloop)  # void
 */
  __pyx_t_2 = (__pyx_v_self->running == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/timer.pyx":50
 *         if self.running == 0:
 *             # Update libuv internal time.
 *             uv.uv_update_time(self._loop.uvloop)  # void             # <<<<<<<<<<<<<<
 *             self.start_t = uv.uv_now(self._loop.uvloop)
 * 
 */
    uv_update_time(__pyx_v_self->__pyx_base._loop->uvloop);

    /* "uvloop/handles/timer.pyx":51
 *             # Update libuv internal time.
 *             uv.uv_update_time(self._loop.uvloop)  # void
 *             self.start_t = uv.uv_now(self._loop.uvloop)             # <<<<<<<<<<<<<<
 * 
 *             err = uv.uv_timer_start(<uv.uv_timer_t*>self._handle,
 */
    __pyx_v_self->start_t = uv_now(__pyx_v_self->__pyx_base._loop->uvloop);

    /* "uvloop/handles/timer.pyx":53
 *             self.start_t = uv.uv_now(self._loop.uvloop)
 * 
 *             err = uv.uv_timer_start(<uv.uv_timer_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                     __uvtimer_callback,
 *                                     self.timeout, 0)
 */
    __pyx_v_err = uv_timer_start(((uv_timer_t *)__pyx_v_self->__pyx_base._handle), __pyx_f_6uvloop_4loop___uvtimer_callback, __pyx_v_self->timeout, 0);

    /* "uvloop/handles/timer.pyx":56
 *                                     __uvtimer_callback,
 *                                     self.timeout, 0)
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/timer.pyx":57
 *                                     self.timeout, 0)
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 57, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_exc = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/handles/timer.pyx":58
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 *             self.running = 1
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 58, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/timer.pyx":59
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 *             self.running = 1
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/timer.pyx":56
 *                                     __uvtimer_callback,
 *                                     self.timeout, 0)
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/timer.pyx":60
 *                 self._fatal_error(exc, True)
 *                 return
 *             self.running = 1             # <<<<<<<<<<<<<<
 * 
 *     cdef get_when(self):
 */
    __pyx_v_self->running = 1;

    /* "uvloop/handles/timer.pyx":48
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             # Update libuv internal time.
 *             uv.uv_update_time(self._loop.uvloop)  # void
 */
  }

  /* "uvloop/handles/timer.pyx":43
 *                 return
 * 
 *     cdef start(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVTimer.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/timer.pyx":62
 *             self.running = 1
 * 
 *     cdef get_when(self):             # <<<<<<<<<<<<<<
 *         return self.start_t + self.timeout
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_7UVTimer_get_when(struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_when", 1);

  /* "uvloop/handles/timer.pyx":63
 * 
 *     cdef get_when(self):
 *         return self.start_t + self.timeout             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t((__pyx_v_self->start_t + __pyx_v_self->timeout)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/timer.pyx":62
 *             self.running = 1
 * 
 *     cdef get_when(self):             # <<<<<<<<<<<<<<
 *         return self.start_t + self.timeout
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVTimer.get_when", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/timer.pyx":66
 * 
 *     @staticmethod
 *     cdef UVTimer new(Loop loop, method_t callback, object ctx,             # <<<<<<<<<<<<<<
 *                      uint64_t timeout):
 * 
 */

static struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_f_6uvloop_4loop_7UVTimer_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, __pyx_t_6uvloop_4loop_method_t __pyx_v_callback, PyObject *__pyx_v_ctx, uint64_t __pyx_v_timeout) {
  struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/timer.pyx":70
 * 
 *         cdef UVTimer handle
 *         handle = UVTimer.__new__(UVTimer)             # <<<<<<<<<<<<<<
 *         handle._init(loop, callback, ctx, timeout)
 *         return handle
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UVTimer(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UVTimer), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 70, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UVTimer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/timer.pyx":71
 *         cdef UVTimer handle
 *         handle = UVTimer.__new__(UVTimer)
 *         handle._init(loop, callback, ctx, timeout)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_init(__pyx_v_handle, __pyx_v_loop, __pyx_v_callback, __pyx_v_ctx, __pyx_v_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/timer.pyx":72
 *         handle = UVTimer.__new__(UVTimer)
 *         handle._init(loop, callback, ctx, timeout)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/timer.pyx":66
 * 
 *     @staticmethod
 *     cdef UVTimer new(Loop loop, method_t callback, object ctx,             # <<<<<<<<<<<<<<
 *                      uint64_t timeout):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVTimer.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_7UVTimer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_7UVTimer___reduce_cython__, "UVTimer.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_7UVTimer_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVTimer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVTimer___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_7UVTimer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_7UVTimer___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVTimer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_7UVTimer___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVTimer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_7UVTimer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_7UVTimer_2__setstate_cython__, "UVTimer.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_7UVTimer_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVTimer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVTimer_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_7UVTimer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVTimer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_7UVTimer_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVTimer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_7UVTimer_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVTimer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/timer.pyx":75
 * 
 * 
 * cdef void __uvtimer_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_timer_t* handle,
 * ) noexcept with gil:
 */

static void __pyx_f_6uvloop_4loop___uvtimer_callback(uv_timer_t *__pyx_v_handle) {
  struct __pyx_obj_6uvloop_4loop_UVTimer *__pyx_v_timer = 0;
  __pyx_t_6uvloop_4loop_method_t __pyx_v_cb;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_6uvloop_4loop_method_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uvtimer_callback", 0);

  /* "uvloop/handles/timer.pyx":78
 *     uv.uv_timer_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVTimer callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVTimer callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(16, 78, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/timer.pyx":79
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVTimer callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/timer.pyx":78
 *     uv.uv_timer_t* handle,
 * ) noexcept with gil:
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVTimer callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/handles/timer.pyx":82
 * 
 *     cdef:
 *         UVTimer timer = <UVTimer> handle.data             # <<<<<<<<<<<<<<
 *         method_t cb = timer.callback
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_timer = ((struct __pyx_obj_6uvloop_4loop_UVTimer *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/timer.pyx":83
 *     cdef:
 *         UVTimer timer = <UVTimer> handle.data
 *         method_t cb = timer.callback             # <<<<<<<<<<<<<<
 * 
 *     timer.running = 0
 */
  __pyx_t_4 = __pyx_v_timer->callback;
  __pyx_v_cb = __pyx_t_4;

  /* "uvloop/handles/timer.pyx":85
 *         method_t cb = timer.callback
 * 
 *     timer.running = 0             # <<<<<<<<<<<<<<
 *     try:
 *         cb(timer.ctx)
 */
  __pyx_v_timer->running = 0;

  /* "uvloop/handles/timer.pyx":86
 * 
 *     timer.running = 0
 *     try:             # <<<<<<<<<<<<<<
 *         cb(timer.ctx)
 *     except BaseException as ex:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "uvloop/handles/timer.pyx":87
 *     timer.running = 0
 *     try:
 *         cb(timer.ctx)             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         timer._error(ex, False)
 */
      __pyx_t_3 = __pyx_v_timer->ctx;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_8 = __pyx_v_cb(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(16, 87, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/handles/timer.pyx":86
 * 
 *     timer.running = 0
 *     try:             # <<<<<<<<<<<<<<
 *         cb(timer.ctx)
 *     except BaseException as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/handles/timer.pyx":88
 *     try:
 *         cb(timer.ctx)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         timer._error(ex, False)
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.__uvtimer_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_3, &__pyx_t_10) < 0) __PYX_ERR(16, 88, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_ex = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/handles/timer.pyx":89
 *         cb(timer.ctx)
 *     except BaseException as ex:
 *         timer._error(ex, False)             # <<<<<<<<<<<<<<
 */
        __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVTimer *)__pyx_v_timer->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_timer), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_11)) __PYX_ERR(16, 89, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }

      /* "uvloop/handles/timer.pyx":88
 *     try:
 *         cb(timer.ctx)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         timer._error(ex, False)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_9 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L6_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/handles/timer.pyx":86
 * 
 *     timer.running = 0
 *     try:             # <<<<<<<<<<<<<<
 *         cb(timer.ctx)
 *     except BaseException as ex:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    __pyx_L9_try_end:;
  }

  /* "uvloop/handles/timer.pyx":75
 * 
 * 
 * cdef void __uvtimer_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_timer_t* handle,
 * ) noexcept with gil:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_WriteUnraisable("uvloop.loop.__uvtimer_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_timer);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/poll.pyx":3
 * @cython.no_gc_clear
 * cdef class UVPoll(UVHandle):
 *     cdef _init(self, Loop loop, int fd):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll__init(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, int __pyx_v_fd) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/poll.pyx":6
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_poll_t))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/poll.pyx":8
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_poll_t))             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             self._abort_init()
 */
  __pyx_v_self->__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_poll_t))));

  /* "uvloop/handles/poll.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_poll_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/poll.pyx":10
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_poll_t))
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/poll.pyx":11
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_poll_init(self._loop.uvloop,
 */
    PyErr_NoMemory(); __PYX_ERR(17, 11, __pyx_L1_error)

    /* "uvloop/handles/poll.pyx":9
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_poll_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/poll.pyx":13
 *             raise MemoryError()
 * 
 *         err = uv.uv_poll_init(self._loop.uvloop,             # <<<<<<<<<<<<<<
 *                               <uv.uv_poll_t *>self._handle, fd)
 *         if err < 0:
 */
  __pyx_v_err = uv_poll_init(__pyx_v_self->__pyx_base._loop->uvloop, ((uv_poll_t *)__pyx_v_self->__pyx_base._handle), __pyx_v_fd);

  /* "uvloop/handles/poll.pyx":15
 *         err = uv.uv_poll_init(self._loop.uvloop,
 *                               <uv.uv_poll_t *>self._handle, fd)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/poll.pyx":16
 *                               <uv.uv_poll_t *>self._handle, fd)
 *         if err < 0:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 16, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/poll.pyx":17
 *         if err < 0:
 *             self._abort_init()
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         self._finish_init()
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(17, 17, __pyx_L1_error)

    /* "uvloop/handles/poll.pyx":15
 *         err = uv.uv_poll_init(self._loop.uvloop,
 *                               <uv.uv_poll_t *>self._handle, fd)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  }

  /* "uvloop/handles/poll.pyx":19
 *             raise convert_error(err)
 * 
 *         self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *         self.fd = fd
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/poll.pyx":21
 *         self._finish_init()
 * 
 *         self.fd = fd             # <<<<<<<<<<<<<<
 *         self.reading_handle = None
 *         self.writing_handle = None
 */
  __pyx_v_self->fd = __pyx_v_fd;

  /* "uvloop/handles/poll.pyx":22
 * 
 *         self.fd = fd
 *         self.reading_handle = None             # <<<<<<<<<<<<<<
 *         self.writing_handle = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->reading_handle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->reading_handle);
  __pyx_v_self->reading_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

  /* "uvloop/handles/poll.pyx":23
 *         self.fd = fd
 *         self.reading_handle = None
 *         self.writing_handle = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->writing_handle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->writing_handle);
  __pyx_v_self->writing_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

  /* "uvloop/handles/poll.pyx":3
 * @cython.no_gc_clear
 * cdef class UVPoll(UVHandle):
 *     cdef _init(self, Loop loop, int fd):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVPoll._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":26
 * 
 *     @staticmethod
 *     cdef UVPoll new(Loop loop, int fd):             # <<<<<<<<<<<<<<
 *         cdef UVPoll handle
 *         handle = UVPoll.__new__(UVPoll)
 */

static struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_f_6uvloop_4loop_6UVPoll_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, int __pyx_v_fd) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/poll.pyx":28
 *     cdef UVPoll new(Loop loop, int fd):
 *         cdef UVPoll handle
 *         handle = UVPoll.__new__(UVPoll)             # <<<<<<<<<<<<<<
 *         handle._init(loop, fd)
 *         return handle
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UVPoll(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UVPoll), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 28, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/poll.pyx":29
 *         cdef UVPoll handle
 *         handle = UVPoll.__new__(UVPoll)
 *         handle._init(loop, fd)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_init(__pyx_v_handle, __pyx_v_loop, __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/poll.pyx":30
 *         handle = UVPoll.__new__(UVPoll)
 *         handle._init(loop, fd)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef int is_active(self) noexcept:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/poll.pyx":26
 * 
 *     @staticmethod
 *     cdef UVPoll new(Loop loop, int fd):             # <<<<<<<<<<<<<<
 *         cdef UVPoll handle
 *         handle = UVPoll.__new__(UVPoll)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":32
 *         return handle
 * 
 *     cdef int is_active(self) noexcept:             # <<<<<<<<<<<<<<
 *         return (self.reading_handle is not None or
 *                 self.writing_handle is not None)
 */

static int __pyx_f_6uvloop_4loop_6UVPoll_is_active(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "uvloop/handles/poll.pyx":33
 * 
 *     cdef int is_active(self) noexcept:
 *         return (self.reading_handle is not None or             # <<<<<<<<<<<<<<
 *                 self.writing_handle is not None)
 * 
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_self->reading_handle) != Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "uvloop/handles/poll.pyx":34
 *     cdef int is_active(self) noexcept:
 *         return (self.reading_handle is not None or
 *                 self.writing_handle is not None)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _poll_start(self, int flags):
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_self->writing_handle) != Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "uvloop/handles/poll.pyx":32
 *         return handle
 * 
 *     cdef int is_active(self) noexcept:             # <<<<<<<<<<<<<<
 *         return (self.reading_handle is not None or
 *                 self.writing_handle is not None)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":36
 *                 self.writing_handle is not None)
 * 
 *     cdef inline _poll_start(self, int flags):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVPoll__poll_start(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, int __pyx_v_flags) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_poll_start", 1);

  /* "uvloop/handles/poll.pyx":39
 *         cdef int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_poll_start(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/poll.pyx":41
 *         self._ensure_alive()
 * 
 *         err = uv.uv_poll_start(             # <<<<<<<<<<<<<<
 *             <uv.uv_poll_t*>self._handle,
 *             flags,
 */
  __pyx_v_err = uv_poll_start(((uv_poll_t *)__pyx_v_self->__pyx_base._handle), __pyx_v_flags, __pyx_f_6uvloop_4loop___on_uvpoll_event);

  /* "uvloop/handles/poll.pyx":46
 *             __on_uvpoll_event)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":47
 * 
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/poll.pyx":48
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/poll.pyx":49
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _poll_stop(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/poll.pyx":46
 *             __on_uvpoll_event)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/poll.pyx":36
 *                 self.writing_handle is not None)
 * 
 *     cdef inline _poll_start(self, int flags):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVPoll._poll_start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":51
 *             return
 * 
 *     cdef inline _poll_stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_6UVPoll__poll_stop(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  int __pyx_v_backend_id;
  struct epoll_event __pyx_v_dummy_event;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_poll_stop", 1);

  /* "uvloop/handles/poll.pyx":54
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(17, 54, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":55
 * 
 *         if not self._is_alive():
 *             return             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_poll_stop(<uv.uv_poll_t*>self._handle)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/poll.pyx":54
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/poll.pyx":57
 *             return
 * 
 *         err = uv.uv_poll_stop(<uv.uv_poll_t*>self._handle)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  __pyx_v_err = uv_poll_stop(((uv_poll_t *)__pyx_v_self->__pyx_base._handle));

  /* "uvloop/handles/poll.pyx":58
 * 
 *         err = uv.uv_poll_stop(<uv.uv_poll_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":59
 *         err = uv.uv_poll_stop(<uv.uv_poll_t*>self._handle)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_exc = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/handles/poll.pyx":60
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/poll.pyx":61
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/poll.pyx":58
 * 
 *         err = uv.uv_poll_stop(<uv.uv_poll_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/poll.pyx":67
 *             system.epoll_event dummy_event
 * 
 *         if system.PLATFORM_IS_LINUX:             # <<<<<<<<<<<<<<
 *             # libuv doesn't remove the FD from epoll immediately
 *             # after uv_poll_stop or uv_poll_close, causing hard
 */
  __pyx_t_2 = (PLATFORM_IS_LINUX != 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":77
 *             # after calling uv_poll_stop.
 * 
 *             backend_id = uv.uv_backend_fd(self._loop.uvloop)             # <<<<<<<<<<<<<<
 *             if backend_id != -1:
 *                 memset(&dummy_event, 0, sizeof(dummy_event))
 */
    __pyx_v_backend_id = uv_backend_fd(__pyx_v_self->__pyx_base._loop->uvloop);

    /* "uvloop/handles/poll.pyx":78
 * 
 *             backend_id = uv.uv_backend_fd(self._loop.uvloop)
 *             if backend_id != -1:             # <<<<<<<<<<<<<<
 *                 memset(&dummy_event, 0, sizeof(dummy_event))
 *                 system.epoll_ctl(
 */
    __pyx_t_2 = (__pyx_v_backend_id != -1L);
    if (__pyx_t_2) {

      /* "uvloop/handles/poll.pyx":79
 *             backend_id = uv.uv_backend_fd(self._loop.uvloop)
 *             if backend_id != -1:
 *                 memset(&dummy_event, 0, sizeof(dummy_event))             # <<<<<<<<<<<<<<
 *                 system.epoll_ctl(
 *                     backend_id,
 */
      (void)(memset((&__pyx_v_dummy_event), 0, (sizeof(__pyx_v_dummy_event))));

      /* "uvloop/handles/poll.pyx":80
 *             if backend_id != -1:
 *                 memset(&dummy_event, 0, sizeof(dummy_event))
 *                 system.epoll_ctl(             # <<<<<<<<<<<<<<
 *                     backend_id,
 *                     system.EPOLL_CTL_DEL,
 */
      (void)(epoll_ctl(__pyx_v_backend_id, EPOLL_CTL_DEL, __pyx_v_self->fd, (&__pyx_v_dummy_event)));

      /* "uvloop/handles/poll.pyx":78
 * 
 *             backend_id = uv.uv_backend_fd(self._loop.uvloop)
 *             if backend_id != -1:             # <<<<<<<<<<<<<<
 *                 memset(&dummy_event, 0, sizeof(dummy_event))
 *                 system.epoll_ctl(
 */
    }

    /* "uvloop/handles/poll.pyx":67
 *             system.epoll_event dummy_event
 * 
 *         if system.PLATFORM_IS_LINUX:             # <<<<<<<<<<<<<<
 *             # libuv doesn't remove the FD from epoll immediately
 *             # after uv_poll_stop or uv_poll_close, causing hard
 */
  }

  /* "uvloop/handles/poll.pyx":51
 *             return
 * 
 *     cdef inline _poll_stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVPoll._poll_stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":86
 *                     &dummy_event)  # ignore errors
 * 
 *     cdef is_reading(self):             # <<<<<<<<<<<<<<
 *         return self._is_alive() and self.reading_handle is not None
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_is_reading(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_reading", 1);

  /* "uvloop/handles/poll.pyx":87
 * 
 *     cdef is_reading(self):
 *         return self._is_alive() and self.reading_handle is not None             # <<<<<<<<<<<<<<
 * 
 *     cdef is_writing(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(17, 87, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = (((PyObject *)__pyx_v_self->reading_handle) != Py_None);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/poll.pyx":86
 *                     &dummy_event)  # ignore errors
 * 
 *     cdef is_reading(self):             # <<<<<<<<<<<<<<
 *         return self._is_alive() and self.reading_handle is not None
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.is_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":89
 *         return self._is_alive() and self.reading_handle is not None
 * 
 *     cdef is_writing(self):             # <<<<<<<<<<<<<<
 *         return self._is_alive() and self.writing_handle is not None
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_is_writing(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_writing", 1);

  /* "uvloop/handles/poll.pyx":90
 * 
 *     cdef is_writing(self):
 *         return self._is_alive() and self.writing_handle is not None             # <<<<<<<<<<<<<<
 * 
 *     cdef start_reading(self, Handle callback):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(17, 90, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = (((PyObject *)__pyx_v_self->writing_handle) != Py_None);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/poll.pyx":89
 *         return self._is_alive() and self.reading_handle is not None
 * 
 *     cdef is_writing(self):             # <<<<<<<<<<<<<<
 *         return self._is_alive() and self.writing_handle is not None
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.is_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":92
 *         return self._is_alive() and self.writing_handle is not None
 * 
 *     cdef start_reading(self, Handle callback):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int mask = 0
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_start_reading(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_callback) {
  int __pyx_v_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start_reading", 1);

  /* "uvloop/handles/poll.pyx":94
 *     cdef start_reading(self, Handle callback):
 *         cdef:
 *             int mask = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.reading_handle is None:
 */
  __pyx_v_mask = 0;

  /* "uvloop/handles/poll.pyx":96
 *             int mask = 0
 * 
 *         if self.reading_handle is None:             # <<<<<<<<<<<<<<
 *             # not reading right now, setup the handle
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->reading_handle) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":99
 *             # not reading right now, setup the handle
 * 
 *             mask = uv.UV_READABLE             # <<<<<<<<<<<<<<
 *             if self.writing_handle is not None:
 *                 # are we writing right now?
 */
    __pyx_v_mask = UV_READABLE;

    /* "uvloop/handles/poll.pyx":100
 * 
 *             mask = uv.UV_READABLE
 *             if self.writing_handle is not None:             # <<<<<<<<<<<<<<
 *                 # are we writing right now?
 *                 mask |= uv.UV_WRITABLE
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->writing_handle) != Py_None);
    if (__pyx_t_1) {

      /* "uvloop/handles/poll.pyx":102
 *             if self.writing_handle is not None:
 *                 # are we writing right now?
 *                 mask |= uv.UV_WRITABLE             # <<<<<<<<<<<<<<
 * 
 *             self._poll_start(mask)
 */
      __pyx_v_mask = (__pyx_v_mask | UV_WRITABLE);

      /* "uvloop/handles/poll.pyx":100
 * 
 *             mask = uv.UV_READABLE
 *             if self.writing_handle is not None:             # <<<<<<<<<<<<<<
 *                 # are we writing right now?
 *                 mask |= uv.UV_WRITABLE
 */
    }

    /* "uvloop/handles/poll.pyx":104
 *                 mask |= uv.UV_WRITABLE
 * 
 *             self._poll_start(mask)             # <<<<<<<<<<<<<<
 *         else:
 *             self.reading_handle._cancel()
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVPoll__poll_start(__pyx_v_self, __pyx_v_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/poll.pyx":96
 *             int mask = 0
 * 
 *         if self.reading_handle is None:             # <<<<<<<<<<<<<<
 *             # not reading right now, setup the handle
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/poll.pyx":106
 *             self._poll_start(mask)
 *         else:
 *             self.reading_handle._cancel()             # <<<<<<<<<<<<<<
 * 
 *         self.reading_handle = callback
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_self->reading_handle->__pyx_vtab)->_cancel(__pyx_v_self->reading_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/poll.pyx":108
 *             self.reading_handle._cancel()
 * 
 *         self.reading_handle = callback             # <<<<<<<<<<<<<<
 * 
 *     cdef start_writing(self, Handle callback):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_callback);
  __Pyx_GIVEREF((PyObject *)__pyx_v_callback);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->reading_handle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->reading_handle);
  __pyx_v_self->reading_handle = __pyx_v_callback;

  /* "uvloop/handles/poll.pyx":92
 *         return self._is_alive() and self.writing_handle is not None
 * 
 *     cdef start_reading(self, Handle callback):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int mask = 0
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.start_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":110
 *         self.reading_handle = callback
 * 
 *     cdef start_writing(self, Handle callback):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int mask = 0
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_start_writing(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_callback) {
  int __pyx_v_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start_writing", 1);

  /* "uvloop/handles/poll.pyx":112
 *     cdef start_writing(self, Handle callback):
 *         cdef:
 *             int mask = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.writing_handle is None:
 */
  __pyx_v_mask = 0;

  /* "uvloop/handles/poll.pyx":114
 *             int mask = 0
 * 
 *         if self.writing_handle is None:             # <<<<<<<<<<<<<<
 *             # not writing right now, setup the handle
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writing_handle) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":117
 *             # not writing right now, setup the handle
 * 
 *             mask = uv.UV_WRITABLE             # <<<<<<<<<<<<<<
 *             if self.reading_handle is not None:
 *                 # are we reading right now?
 */
    __pyx_v_mask = UV_WRITABLE;

    /* "uvloop/handles/poll.pyx":118
 * 
 *             mask = uv.UV_WRITABLE
 *             if self.reading_handle is not None:             # <<<<<<<<<<<<<<
 *                 # are we reading right now?
 *                 mask |= uv.UV_READABLE
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->reading_handle) != Py_None);
    if (__pyx_t_1) {

      /* "uvloop/handles/poll.pyx":120
 *             if self.reading_handle is not None:
 *                 # are we reading right now?
 *                 mask |= uv.UV_READABLE             # <<<<<<<<<<<<<<
 * 
 *             self._poll_start(mask)
 */
      __pyx_v_mask = (__pyx_v_mask | UV_READABLE);

      /* "uvloop/handles/poll.pyx":118
 * 
 *             mask = uv.UV_WRITABLE
 *             if self.reading_handle is not None:             # <<<<<<<<<<<<<<
 *                 # are we reading right now?
 *                 mask |= uv.UV_READABLE
 */
    }

    /* "uvloop/handles/poll.pyx":122
 *                 mask |= uv.UV_READABLE
 * 
 *             self._poll_start(mask)             # <<<<<<<<<<<<<<
 *         else:
 *             self.writing_handle._cancel()
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVPoll__poll_start(__pyx_v_self, __pyx_v_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/poll.pyx":114
 *             int mask = 0
 * 
 *         if self.writing_handle is None:             # <<<<<<<<<<<<<<
 *             # not writing right now, setup the handle
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/poll.pyx":124
 *             self._poll_start(mask)
 *         else:
 *             self.writing_handle._cancel()             # <<<<<<<<<<<<<<
 * 
 *         self.writing_handle = callback
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_self->writing_handle->__pyx_vtab)->_cancel(__pyx_v_self->writing_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/poll.pyx":126
 *             self.writing_handle._cancel()
 * 
 *         self.writing_handle = callback             # <<<<<<<<<<<<<<
 * 
 *     cdef stop_reading(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_callback);
  __Pyx_GIVEREF((PyObject *)__pyx_v_callback);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->writing_handle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->writing_handle);
  __pyx_v_self->writing_handle = __pyx_v_callback;

  /* "uvloop/handles/poll.pyx":110
 *         self.reading_handle = callback
 * 
 *     cdef start_writing(self, Handle callback):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int mask = 0
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.start_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":128
 *         self.writing_handle = callback
 * 
 *     cdef stop_reading(self):             # <<<<<<<<<<<<<<
 *         if self.reading_handle is None:
 *             return False
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_stop_reading(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop_reading", 1);

  /* "uvloop/handles/poll.pyx":129
 * 
 *     cdef stop_reading(self):
 *         if self.reading_handle is None:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->reading_handle) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":130
 *     cdef stop_reading(self):
 *         if self.reading_handle is None:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         self.reading_handle._cancel()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "uvloop/handles/poll.pyx":129
 * 
 *     cdef stop_reading(self):
 *         if self.reading_handle is None:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "uvloop/handles/poll.pyx":132
 *             return False
 * 
 *         self.reading_handle._cancel()             # <<<<<<<<<<<<<<
 *         self.reading_handle = None
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_self->reading_handle->__pyx_vtab)->_cancel(__pyx_v_self->reading_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/poll.pyx":133
 * 
 *         self.reading_handle._cancel()
 *         self.reading_handle = None             # <<<<<<<<<<<<<<
 * 
 *         if self.writing_handle is None:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->reading_handle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->reading_handle);
  __pyx_v_self->reading_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

  /* "uvloop/handles/poll.pyx":135
 *         self.reading_handle = None
 * 
 *         if self.writing_handle is None:             # <<<<<<<<<<<<<<
 *             self.stop()
 *         else:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writing_handle) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":136
 * 
 *         if self.writing_handle is None:
 *             self.stop()             # <<<<<<<<<<<<<<
 *         else:
 *             self._poll_start(uv.UV_WRITABLE)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->stop(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/poll.pyx":135
 *         self.reading_handle = None
 * 
 *         if self.writing_handle is None:             # <<<<<<<<<<<<<<
 *             self.stop()
 *         else:
 */
    goto __pyx_L4;
  }

  /* "uvloop/handles/poll.pyx":138
 *             self.stop()
 *         else:
 *             self._poll_start(uv.UV_WRITABLE)             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
  /*else*/ {
    __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVPoll__poll_start(__pyx_v_self, UV_WRITABLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "uvloop/handles/poll.pyx":140
 *             self._poll_start(uv.UV_WRITABLE)
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef stop_writing(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "uvloop/handles/poll.pyx":128
 *         self.writing_handle = callback
 * 
 *     cdef stop_reading(self):             # <<<<<<<<<<<<<<
 *         if self.reading_handle is None:
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.stop_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":142
 *         return True
 * 
 *     cdef stop_writing(self):             # <<<<<<<<<<<<<<
 *         if self.writing_handle is None:
 *             return False
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_stop_writing(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop_writing", 1);

  /* "uvloop/handles/poll.pyx":143
 * 
 *     cdef stop_writing(self):
 *         if self.writing_handle is None:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writing_handle) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":144
 *     cdef stop_writing(self):
 *         if self.writing_handle is None:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         self.writing_handle._cancel()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "uvloop/handles/poll.pyx":143
 * 
 *     cdef stop_writing(self):
 *         if self.writing_handle is None:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
  }

  /* "uvloop/handles/poll.pyx":146
 *             return False
 * 
 *         self.writing_handle._cancel()             # <<<<<<<<<<<<<<
 *         self.writing_handle = None
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_self->writing_handle->__pyx_vtab)->_cancel(__pyx_v_self->writing_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/poll.pyx":147
 * 
 *         self.writing_handle._cancel()
 *         self.writing_handle = None             # <<<<<<<<<<<<<<
 * 
 *         if self.reading_handle is None:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->writing_handle);
  __Pyx_DECREF((PyObject *)__pyx_v_self->writing_handle);
  __pyx_v_self->writing_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

  /* "uvloop/handles/poll.pyx":149
 *         self.writing_handle = None
 * 
 *         if self.reading_handle is None:             # <<<<<<<<<<<<<<
 *             self.stop()
 *         else:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->reading_handle) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":150
 * 
 *         if self.reading_handle is None:
 *             self.stop()             # <<<<<<<<<<<<<<
 *         else:
 *             self._poll_start(uv.UV_READABLE)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->stop(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/poll.pyx":149
 *         self.writing_handle = None
 * 
 *         if self.reading_handle is None:             # <<<<<<<<<<<<<<
 *             self.stop()
 *         else:
 */
    goto __pyx_L4;
  }

  /* "uvloop/handles/poll.pyx":152
 *             self.stop()
 *         else:
 *             self._poll_start(uv.UV_READABLE)             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
  /*else*/ {
    __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVPoll__poll_start(__pyx_v_self, UV_READABLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "uvloop/handles/poll.pyx":154
 *             self._poll_start(uv.UV_READABLE)
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef stop(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "uvloop/handles/poll.pyx":142
 *         return True
 * 
 *     cdef stop_writing(self):             # <<<<<<<<<<<<<<
 *         if self.writing_handle is None:
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.stop_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":156
 *         return True
 * 
 *     cdef stop(self):             # <<<<<<<<<<<<<<
 *         if self.reading_handle is not None:
 *             self.reading_handle._cancel()
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll_stop(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop", 1);

  /* "uvloop/handles/poll.pyx":157
 * 
 *     cdef stop(self):
 *         if self.reading_handle is not None:             # <<<<<<<<<<<<<<
 *             self.reading_handle._cancel()
 *             self.reading_handle = None
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->reading_handle) != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":158
 *     cdef stop(self):
 *         if self.reading_handle is not None:
 *             self.reading_handle._cancel()             # <<<<<<<<<<<<<<
 *             self.reading_handle = None
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_self->reading_handle->__pyx_vtab)->_cancel(__pyx_v_self->reading_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/poll.pyx":159
 *         if self.reading_handle is not None:
 *             self.reading_handle._cancel()
 *             self.reading_handle = None             # <<<<<<<<<<<<<<
 * 
 *         if self.writing_handle is not None:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->reading_handle);
    __Pyx_DECREF((PyObject *)__pyx_v_self->reading_handle);
    __pyx_v_self->reading_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

    /* "uvloop/handles/poll.pyx":157
 * 
 *     cdef stop(self):
 *         if self.reading_handle is not None:             # <<<<<<<<<<<<<<
 *             self.reading_handle._cancel()
 *             self.reading_handle = None
 */
  }

  /* "uvloop/handles/poll.pyx":161
 *             self.reading_handle = None
 * 
 *         if self.writing_handle is not None:             # <<<<<<<<<<<<<<
 *             self.writing_handle._cancel()
 *             self.writing_handle = None
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->writing_handle) != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":162
 * 
 *         if self.writing_handle is not None:
 *             self.writing_handle._cancel()             # <<<<<<<<<<<<<<
 *             self.writing_handle = None
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Handle *)__pyx_v_self->writing_handle->__pyx_vtab)->_cancel(__pyx_v_self->writing_handle); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/poll.pyx":163
 *         if self.writing_handle is not None:
 *             self.writing_handle._cancel()
 *             self.writing_handle = None             # <<<<<<<<<<<<<<
 * 
 *         self._poll_stop()
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->writing_handle);
    __Pyx_DECREF((PyObject *)__pyx_v_self->writing_handle);
    __pyx_v_self->writing_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

    /* "uvloop/handles/poll.pyx":161
 *             self.reading_handle = None
 * 
 *         if self.writing_handle is not None:             # <<<<<<<<<<<<<<
 *             self.writing_handle._cancel()
 *             self.writing_handle = None
 */
  }

  /* "uvloop/handles/poll.pyx":165
 *             self.writing_handle = None
 * 
 *         self._poll_stop()             # <<<<<<<<<<<<<<
 * 
 *     cdef _close(self):
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_6UVPoll__poll_stop(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/poll.pyx":156
 *         return True
 * 
 *     cdef stop(self):             # <<<<<<<<<<<<<<
 *         if self.reading_handle is not None:
 *             self.reading_handle._cancel()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVPoll.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":167
 *         self._poll_stop()
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         if self.is_active():
 *             self.stop()
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll__close(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/handles/poll.pyx":168
 * 
 *     cdef _close(self):
 *         if self.is_active():             # <<<<<<<<<<<<<<
 *             self.stop()
 * 
 */
  __pyx_t_1 = (((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->is_active(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/poll.pyx":169
 *     cdef _close(self):
 *         if self.is_active():
 *             self.stop()             # <<<<<<<<<<<<<<
 * 
 *         UVHandle._close(<UVHandle>self)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->stop(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/poll.pyx":168
 * 
 *     cdef _close(self):
 *         if self.is_active():             # <<<<<<<<<<<<<<
 *             self.stop()
 * 
 */
  }

  /* "uvloop/handles/poll.pyx":171
 *             self.stop()
 * 
 *         UVHandle._close(<UVHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_8UVHandle__close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(17, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/poll.pyx":167
 *         self._poll_stop()
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         if self.is_active():
 *             self.stop()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVPoll._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":173
 *         UVHandle._close(<UVHandle>self)
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         try:
 *             if self.reading_handle is not None:
 */

static PyObject *__pyx_f_6uvloop_4loop_6UVPoll__fatal_error(struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc, CYTHON_UNUSED PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_6UVPoll__fatal_error *__pyx_optional_args) {
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  char const *__pyx_t_17;
  char const *__pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fatal_error", 1);
  if (__pyx_optional_args) {
  }

  /* "uvloop/handles/poll.pyx":174
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):
 *         try:             # <<<<<<<<<<<<<<
 *             if self.reading_handle is not None:
 *                 try:
 */
  /*try:*/ {

    /* "uvloop/handles/poll.pyx":175
 *     cdef _fatal_error(self, exc, throw, reason=None):
 *         try:
 *             if self.reading_handle is not None:             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.reading_handle._run()
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->reading_handle) != Py_None);
    if (__pyx_t_1) {

      /* "uvloop/handles/poll.pyx":176
 *         try:
 *             if self.reading_handle is not None:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.reading_handle._run()
 *                 except BaseException as ex:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        /*try:*/ {

          /* "uvloop/handles/poll.pyx":177
 *             if self.reading_handle is not None:
 *                 try:
 *                     self.reading_handle._run()             # <<<<<<<<<<<<<<
 *                 except BaseException as ex:
 *                     self._loop._handle_exception(ex)
 */
          __pyx_t_5 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_self->reading_handle); if (unlikely(!__pyx_t_5)) __PYX_ERR(17, 177, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/handles/poll.pyx":176
 *         try:
 *             if self.reading_handle is not None:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.reading_handle._run()
 *                 except BaseException as ex:
 */
        }
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/poll.pyx":178
 *                 try:
 *                     self.reading_handle._run()
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     self._loop._handle_exception(ex)
 *                 self.reading_handle = None
 */
        __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
        if (__pyx_t_6) {
          __Pyx_AddTraceback("uvloop.loop.UVPoll._fatal_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(17, 178, __pyx_L9_except_error)
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __pyx_v_ex = __pyx_t_7;
          /*try:*/ {

            /* "uvloop/handles/poll.pyx":179
 *                     self.reading_handle._run()
 *                 except BaseException as ex:
 *                     self._loop._handle_exception(ex)             # <<<<<<<<<<<<<<
 *                 self.reading_handle = None
 * 
 */
            ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_base._loop->__pyx_vtab)->_handle_exception(__pyx_v_self->__pyx_base._loop, __pyx_v_ex); if (unlikely(PyErr_Occurred())) __PYX_ERR(17, 179, __pyx_L18_error)
          }

          /* "uvloop/handles/poll.pyx":178
 *                 try:
 *                     self.reading_handle._run()
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     self._loop._handle_exception(ex)
 *                 self.reading_handle = None
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              goto __pyx_L19;
            }
            __pyx_L18_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
              __Pyx_XGOTREF(__pyx_t_11);
              __Pyx_XGOTREF(__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_13);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);
              __pyx_t_6 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_14);
                __Pyx_XGIVEREF(__pyx_t_15);
                __Pyx_XGIVEREF(__pyx_t_16);
                __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
              }
              __Pyx_XGIVEREF(__pyx_t_11);
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
              __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
              __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
              goto __pyx_L9_except_error;
            }
            __pyx_L19:;
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L8_exception_handled;
        }
        goto __pyx_L9_except_error;

        /* "uvloop/handles/poll.pyx":176
 *         try:
 *             if self.reading_handle is not None:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.reading_handle._run()
 *                 except BaseException as ex:
 */
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
        goto __pyx_L4_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
        __pyx_L12_try_end:;
      }

      /* "uvloop/handles/poll.pyx":180
 *                 except BaseException as ex:
 *                     self._loop._handle_exception(ex)
 *                 self.reading_handle = None             # <<<<<<<<<<<<<<
 * 
 *             if self.writing_handle is not None:
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->reading_handle);
      __Pyx_DECREF((PyObject *)__pyx_v_self->reading_handle);
      __pyx_v_self->reading_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

      /* "uvloop/handles/poll.pyx":175
 *     cdef _fatal_error(self, exc, throw, reason=None):
 *         try:
 *             if self.reading_handle is not None:             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.reading_handle._run()
 */
    }

    /* "uvloop/handles/poll.pyx":182
 *                 self.reading_handle = None
 * 
 *             if self.writing_handle is not None:             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.writing_handle._run()
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->writing_handle) != Py_None);
    if (__pyx_t_1) {

      /* "uvloop/handles/poll.pyx":183
 * 
 *             if self.writing_handle is not None:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.writing_handle._run()
 *                 except BaseException as ex:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_2);
        /*try:*/ {

          /* "uvloop/handles/poll.pyx":184
 *             if self.writing_handle is not None:
 *                 try:
 *                     self.writing_handle._run()             # <<<<<<<<<<<<<<
 *                 except BaseException as ex:
 *                     self._loop._handle_exception(ex)
 */
          __pyx_t_8 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_self->writing_handle); if (unlikely(!__pyx_t_8)) __PYX_ERR(17, 184, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "uvloop/handles/poll.pyx":183
 * 
 *             if self.writing_handle is not None:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.writing_handle._run()
 *                 except BaseException as ex:
 */
        }
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L30_try_end;
        __pyx_L25_error:;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "uvloop/handles/poll.pyx":185
 *                 try:
 *                     self.writing_handle._run()
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     self._loop._handle_exception(ex)
 *                 self.writing_handle = None
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("uvloop.loop.UVPoll._fatal_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_5) < 0) __PYX_ERR(17, 185, __pyx_L27_except_error)
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __pyx_v_ex = __pyx_t_7;
          /*try:*/ {

            /* "uvloop/handles/poll.pyx":186
 *                     self.writing_handle._run()
 *                 except BaseException as ex:
 *                     self._loop._handle_exception(ex)             # <<<<<<<<<<<<<<
 *                 self.writing_handle = None
 * 
 */
            ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_base._loop->__pyx_vtab)->_handle_exception(__pyx_v_self->__pyx_base._loop, __pyx_v_ex); if (unlikely(PyErr_Occurred())) __PYX_ERR(17, 186, __pyx_L36_error)
          }

          /* "uvloop/handles/poll.pyx":185
 *                 try:
 *                     self.writing_handle._run()
 *                 except BaseException as ex:             # <<<<<<<<<<<<<<
 *                     self._loop._handle_exception(ex)
 *                 self.writing_handle = None
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              goto __pyx_L37;
            }
            __pyx_L36_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_13);
              __Pyx_XGOTREF(__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_11);
              __pyx_t_9 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_13);
                __Pyx_XGIVEREF(__pyx_t_12);
                __Pyx_XGIVEREF(__pyx_t_11);
                __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
              }
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_ErrRestore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
              __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0;
              __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_17;
              goto __pyx_L27_except_error;
            }
            __pyx_L37:;
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L26_exception_handled;
        }
        goto __pyx_L27_except_error;

        /* "uvloop/handles/poll.pyx":183
 * 
 *             if self.writing_handle is not None:
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.writing_handle._run()
 *                 except BaseException as ex:
 */
        __pyx_L27_except_error:;
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_3, __pyx_t_2);
        goto __pyx_L4_error;
        __pyx_L26_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_3, __pyx_t_2);
        __pyx_L30_try_end:;
      }

      /* "uvloop/handles/poll.pyx":187
 *                 except BaseException as ex:
 *                     self._loop._handle_exception(ex)
 *                 self.writing_handle = None             # <<<<<<<<<<<<<<
 * 
 *         finally:
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->writing_handle);
      __Pyx_DECREF((PyObject *)__pyx_v_self->writing_handle);
      __pyx_v_self->writing_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None);

      /* "uvloop/handles/poll.pyx":182
 *                 self.reading_handle = None
 * 
 *             if self.writing_handle is not None:             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.writing_handle._run()
 */
    }
  }

  /* "uvloop/handles/poll.pyx":190
 * 
 *         finally:
 *             self._close()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(17, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0)) __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_6 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
      {
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(17, 190, __pyx_L43_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_t_2 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_18;
      goto __pyx_L1_error;
      __pyx_L43_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "uvloop/handles/poll.pyx":173
 *         UVHandle._close(<UVHandle>self)
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         try:
 *             if self.reading_handle is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.UVPoll._fatal_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6UVPoll_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6UVPoll___reduce_cython__, "UVPoll.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6UVPoll_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVPoll_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVPoll___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6UVPoll_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6UVPoll___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6UVPoll___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVPoll.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6UVPoll_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6UVPoll_2__setstate_cython__, "UVPoll.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6UVPoll_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVPoll_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVPoll_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6UVPoll_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVPoll.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_6UVPoll_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6UVPoll_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVPoll.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/poll.pyx":193
 * 
 * 
 * cdef void __on_uvpoll_event(             # <<<<<<<<<<<<<<
 *     uv.uv_poll_t* handle,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___on_uvpoll_event(uv_poll_t *__pyx_v_handle, int __pyx_v_status, int __pyx_v_events) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *__pyx_v_poll = 0;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  char const *__pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__on_uvpoll_event", 0);

  /* "uvloop/handles/poll.pyx":199
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVPoll callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVPoll callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(17, 199, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":200
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVPoll callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/poll.pyx":199
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle, "UVPoll callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/handles/poll.pyx":203
 * 
 *     cdef:
 *         UVPoll poll = <UVPoll> handle.data             # <<<<<<<<<<<<<<
 * 
 *     if status < 0:
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_poll = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/poll.pyx":205
 *         UVPoll poll = <UVPoll> handle.data
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *         poll._fatal_error(exc, False)
 */
  __pyx_t_2 = (__pyx_v_status < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":206
 * 
 *     if status < 0:
 *         exc = convert_error(status)             # <<<<<<<<<<<<<<
 *         poll._fatal_error(exc, False)
 *         return
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_exc = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/handles/poll.pyx":207
 *     if status < 0:
 *         exc = convert_error(status)
 *         poll._fatal_error(exc, False)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_poll), __pyx_v_exc, Py_False, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/poll.pyx":208
 *         exc = convert_error(status)
 *         poll._fatal_error(exc, False)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if ((events & (uv.UV_READABLE | uv.UV_DISCONNECT)) and
 */
    goto __pyx_L0;

    /* "uvloop/handles/poll.pyx":205
 *         UVPoll poll = <UVPoll> handle.data
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *         poll._fatal_error(exc, False)
 */
  }

  /* "uvloop/handles/poll.pyx":210
 *         return
 * 
 *     if ((events & (uv.UV_READABLE | uv.UV_DISCONNECT)) and             # <<<<<<<<<<<<<<
 *             poll.reading_handle is not None):
 * 
 */
  __pyx_t_1 = ((__pyx_v_events & (UV_READABLE | UV_DISCONNECT)) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }

  /* "uvloop/handles/poll.pyx":211
 * 
 *     if ((events & (uv.UV_READABLE | uv.UV_DISCONNECT)) and
 *             poll.reading_handle is not None):             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_poll->reading_handle) != Py_None);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;

  /* "uvloop/handles/poll.pyx":210
 *         return
 * 
 *     if ((events & (uv.UV_READABLE | uv.UV_DISCONNECT)) and             # <<<<<<<<<<<<<<
 *             poll.reading_handle is not None):
 * 
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":213
 *             poll.reading_handle is not None):
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_events_total += 1
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "uvloop/handles/poll.pyx":214
 * 
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_read_events_total += 1
 *             poll.reading_handle._run()
 */
        __pyx_t_2 = (UVLOOP_DEBUG != 0);
        if (__pyx_t_2) {

          /* "uvloop/handles/poll.pyx":215
 *         try:
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_events_total += 1             # <<<<<<<<<<<<<<
 *             poll.reading_handle._run()
 *         except BaseException as ex:
 */
          __Pyx_INCREF((PyObject *)__pyx_v_poll->__pyx_base._loop);
          __pyx_t_7 = __pyx_v_poll->__pyx_base._loop;
          __pyx_t_7->_poll_read_events_total = (__pyx_t_7->_poll_read_events_total + 1);
          __Pyx_DECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;

          /* "uvloop/handles/poll.pyx":214
 * 
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_read_events_total += 1
 *             poll.reading_handle._run()
 */
        }

        /* "uvloop/handles/poll.pyx":216
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_events_total += 1
 *             poll.reading_handle._run()             # <<<<<<<<<<<<<<
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:
 */
        __pyx_t_3 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_poll->reading_handle); if (unlikely(!__pyx_t_3)) __PYX_ERR(17, 216, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "uvloop/handles/poll.pyx":213
 *             poll.reading_handle is not None):
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_events_total += 1
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L13_try_end;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/handles/poll.pyx":217
 *                 poll._loop._poll_read_events_total += 1
 *             poll.reading_handle._run()
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_cb_errors_total += 1
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("uvloop.loop.__on_uvpoll_event", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_9, &__pyx_t_10) < 0) __PYX_ERR(17, 217, __pyx_L10_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_v_ex = __pyx_t_9;
        /*try:*/ {

          /* "uvloop/handles/poll.pyx":218
 *             poll.reading_handle._run()
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_read_cb_errors_total += 1
 *             poll._error(ex, False)
 */
          __pyx_t_2 = (UVLOOP_DEBUG != 0);
          if (__pyx_t_2) {

            /* "uvloop/handles/poll.pyx":219
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_cb_errors_total += 1             # <<<<<<<<<<<<<<
 *             poll._error(ex, False)
 *             # continue code execution
 */
            __Pyx_INCREF((PyObject *)__pyx_v_poll->__pyx_base._loop);
            __pyx_t_7 = __pyx_v_poll->__pyx_base._loop;
            __pyx_t_7->_poll_read_cb_errors_total = (__pyx_t_7->_poll_read_cb_errors_total + 1);
            __Pyx_DECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;

            /* "uvloop/handles/poll.pyx":218
 *             poll.reading_handle._run()
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_read_cb_errors_total += 1
 *             poll._error(ex, False)
 */
          }

          /* "uvloop/handles/poll.pyx":220
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_cb_errors_total += 1
 *             poll._error(ex, False)             # <<<<<<<<<<<<<<
 *             # continue code execution
 * 
 */
          __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_poll), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_11)) __PYX_ERR(17, 220, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }

        /* "uvloop/handles/poll.pyx":217
 *                 poll._loop._poll_read_events_total += 1
 *             poll.reading_handle._run()
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_cb_errors_total += 1
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L21;
          }
          __pyx_L20_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L10_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L9_exception_handled;
      }
      goto __pyx_L10_except_error;

      /* "uvloop/handles/poll.pyx":213
 *             poll.reading_handle is not None):
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_read_events_total += 1
 */
      __pyx_L10_except_error:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L9_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_L13_try_end:;
    }

    /* "uvloop/handles/poll.pyx":210
 *         return
 * 
 *     if ((events & (uv.UV_READABLE | uv.UV_DISCONNECT)) and             # <<<<<<<<<<<<<<
 *             poll.reading_handle is not None):
 * 
 */
  }

  /* "uvloop/handles/poll.pyx":223
 *             # continue code execution
 * 
 *     if ((events & (uv.UV_WRITABLE | uv.UV_DISCONNECT)) and             # <<<<<<<<<<<<<<
 *             poll.writing_handle is not None):
 * 
 */
  __pyx_t_1 = ((__pyx_v_events & (UV_WRITABLE | UV_DISCONNECT)) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L28_bool_binop_done;
  }

  /* "uvloop/handles/poll.pyx":224
 * 
 *     if ((events & (uv.UV_WRITABLE | uv.UV_DISCONNECT)) and
 *             poll.writing_handle is not None):             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_poll->writing_handle) != Py_None);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L28_bool_binop_done:;

  /* "uvloop/handles/poll.pyx":223
 *             # continue code execution
 * 
 *     if ((events & (uv.UV_WRITABLE | uv.UV_DISCONNECT)) and             # <<<<<<<<<<<<<<
 *             poll.writing_handle is not None):
 * 
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/poll.pyx":226
 *             poll.writing_handle is not None):
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_events_total += 1
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/handles/poll.pyx":227
 * 
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_write_events_total += 1
 *             poll.writing_handle._run()
 */
        __pyx_t_2 = (UVLOOP_DEBUG != 0);
        if (__pyx_t_2) {

          /* "uvloop/handles/poll.pyx":228
 *         try:
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_events_total += 1             # <<<<<<<<<<<<<<
 *             poll.writing_handle._run()
 *         except BaseException as ex:
 */
          __Pyx_INCREF((PyObject *)__pyx_v_poll->__pyx_base._loop);
          __pyx_t_7 = __pyx_v_poll->__pyx_base._loop;
          __pyx_t_7->_poll_write_events_total = (__pyx_t_7->_poll_write_events_total + 1);
          __Pyx_DECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;

          /* "uvloop/handles/poll.pyx":227
 * 
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_write_events_total += 1
 *             poll.writing_handle._run()
 */
        }

        /* "uvloop/handles/poll.pyx":229
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_events_total += 1
 *             poll.writing_handle._run()             # <<<<<<<<<<<<<<
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:
 */
        __pyx_t_10 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_poll->writing_handle); if (unlikely(!__pyx_t_10)) __PYX_ERR(17, 229, __pyx_L30_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/handles/poll.pyx":226
 *             poll.writing_handle is not None):
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_events_total += 1
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L35_try_end;
      __pyx_L30_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/handles/poll.pyx":230
 *                 poll._loop._poll_write_events_total += 1
 *             poll.writing_handle._run()
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_cb_errors_total += 1
 */
      __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_12) {
        __Pyx_AddTraceback("uvloop.loop.__on_uvpoll_event", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_3) < 0) __PYX_ERR(17, 230, __pyx_L32_except_error)
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_v_ex = __pyx_t_9;
        /*try:*/ {

          /* "uvloop/handles/poll.pyx":231
 *             poll.writing_handle._run()
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_write_cb_errors_total += 1
 *             poll._error(ex, False)
 */
          __pyx_t_2 = (UVLOOP_DEBUG != 0);
          if (__pyx_t_2) {

            /* "uvloop/handles/poll.pyx":232
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_cb_errors_total += 1             # <<<<<<<<<<<<<<
 *             poll._error(ex, False)
 */
            __Pyx_INCREF((PyObject *)__pyx_v_poll->__pyx_base._loop);
            __pyx_t_7 = __pyx_v_poll->__pyx_base._loop;
            __pyx_t_7->_poll_write_cb_errors_total = (__pyx_t_7->_poll_write_cb_errors_total + 1);
            __Pyx_DECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;

            /* "uvloop/handles/poll.pyx":231
 *             poll.writing_handle._run()
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 poll._loop._poll_write_cb_errors_total += 1
 *             poll._error(ex, False)
 */
          }

          /* "uvloop/handles/poll.pyx":233
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_cb_errors_total += 1
 *             poll._error(ex, False)             # <<<<<<<<<<<<<<
 */
          __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVPoll *)__pyx_v_poll->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_poll), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_11)) __PYX_ERR(17, 233, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }

        /* "uvloop/handles/poll.pyx":230
 *                 poll._loop._poll_write_events_total += 1
 *             poll.writing_handle._run()
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_cb_errors_total += 1
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L43;
          }
          __pyx_L42_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_18, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_14);
            __pyx_t_12 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_15, __pyx_t_14);
            }
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_19, __pyx_t_18, __pyx_t_17);
            __pyx_t_19 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0;
            __pyx_lineno = __pyx_t_12; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_20;
            goto __pyx_L32_except_error;
          }
          __pyx_L43:;
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L31_exception_handled;
      }
      goto __pyx_L32_except_error;

      /* "uvloop/handles/poll.pyx":226
 *             poll.writing_handle is not None):
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 poll._loop._poll_write_events_total += 1
 */
      __pyx_L32_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L31_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      __pyx_L35_try_end:;
    }

    /* "uvloop/handles/poll.pyx":223
 *             # continue code execution
 * 
 *     if ((events & (uv.UV_WRITABLE | uv.UV_DISCONNECT)) and             # <<<<<<<<<<<<<<
 *             poll.writing_handle is not None):
 * 
 */
  }

  /* "uvloop/handles/poll.pyx":193
 * 
 * 
 * cdef void __on_uvpoll_event(             # <<<<<<<<<<<<<<
 *     uv.uv_poll_t* handle,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF((PyObject *)__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_WriteUnraisable("uvloop.loop.__on_uvpoll_event", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_poll);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/basetransport.pyx":3
 * cdef class UVBaseTransport(UVSocketHandle):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         # Flow control
 *         self._high_water = FLOW_CONTROL_HIGH_WATER * 1024
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_15UVBaseTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_15UVBaseTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport___cinit__(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_15UVBaseTransport___cinit__(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/basetransport.pyx":5
 *     def __cinit__(self):
 *         # Flow control
 *         self._high_water = FLOW_CONTROL_HIGH_WATER * 1024             # <<<<<<<<<<<<<<
 *         self._low_water = FLOW_CONTROL_HIGH_WATER // 4
 * 
 */
  __pyx_v_self->_high_water = (__pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER * 0x400);

  /* "uvloop/handles/basetransport.pyx":6
 *         # Flow control
 *         self._high_water = FLOW_CONTROL_HIGH_WATER * 1024
 *         self._low_water = FLOW_CONTROL_HIGH_WATER // 4             # <<<<<<<<<<<<<<
 * 
 *         self._protocol = None
 */
  __pyx_v_self->_low_water = __Pyx_div_long(__pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER, 4);

  /* "uvloop/handles/basetransport.pyx":8
 *         self._low_water = FLOW_CONTROL_HIGH_WATER // 4
 * 
 *         self._protocol = None             # <<<<<<<<<<<<<<
 *         self._protocol_connected = 0
 *         self._protocol_paused = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol);
  __Pyx_DECREF(__pyx_v_self->_protocol);
  __pyx_v_self->_protocol = Py_None;

  /* "uvloop/handles/basetransport.pyx":9
 * 
 *         self._protocol = None
 *         self._protocol_connected = 0             # <<<<<<<<<<<<<<
 *         self._protocol_paused = 0
 *         self._protocol_data_received = None
 */
  __pyx_v_self->_protocol_connected = 0;

  /* "uvloop/handles/basetransport.pyx":10
 *         self._protocol = None
 *         self._protocol_connected = 0
 *         self._protocol_paused = 0             # <<<<<<<<<<<<<<
 *         self._protocol_data_received = None
 * 
 */
  __pyx_v_self->_protocol_paused = 0;

  /* "uvloop/handles/basetransport.pyx":11
 *         self._protocol_connected = 0
 *         self._protocol_paused = 0
 *         self._protocol_data_received = None             # <<<<<<<<<<<<<<
 * 
 *         self._server = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol_data_received);
  __Pyx_DECREF(__pyx_v_self->_protocol_data_received);
  __pyx_v_self->_protocol_data_received = Py_None;

  /* "uvloop/handles/basetransport.pyx":13
 *         self._protocol_data_received = None
 * 
 *         self._server = None             # <<<<<<<<<<<<<<
 *         self._waiter = None
 *         self._extra_info = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_server);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_server);
  __pyx_v_self->_server = ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None);

  /* "uvloop/handles/basetransport.pyx":14
 * 
 *         self._server = None
 *         self._waiter = None             # <<<<<<<<<<<<<<
 *         self._extra_info = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_waiter);
  __Pyx_DECREF(__pyx_v_self->_waiter);
  __pyx_v_self->_waiter = Py_None;

  /* "uvloop/handles/basetransport.pyx":15
 *         self._server = None
 *         self._waiter = None
 *         self._extra_info = None             # <<<<<<<<<<<<<<
 * 
 *         self._conn_lost = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_extra_info);
  __Pyx_DECREF(__pyx_v_self->_extra_info);
  __pyx_v_self->_extra_info = ((PyObject*)Py_None);

  /* "uvloop/handles/basetransport.pyx":17
 *         self._extra_info = None
 * 
 *         self._conn_lost = 0             # <<<<<<<<<<<<<<
 * 
 *         self._closing = 0
 */
  __pyx_v_self->_conn_lost = 0;

  /* "uvloop/handles/basetransport.pyx":19
 *         self._conn_lost = 0
 * 
 *         self._closing = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t _get_write_buffer_size(self):
 */
  __pyx_v_self->_closing = 0;

  /* "uvloop/handles/basetransport.pyx":3
 * cdef class UVBaseTransport(UVSocketHandle):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         # Flow control
 *         self._high_water = FLOW_CONTROL_HIGH_WATER * 1024
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":21
 *         self._closing = 0
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

static size_t __pyx_f_6uvloop_4loop_15UVBaseTransport__get_write_buffer_size(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  size_t __pyx_r;

  /* "uvloop/handles/basetransport.pyx":22
 * 
 *     cdef size_t _get_write_buffer_size(self):
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _schedule_call_connection_made(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "uvloop/handles/basetransport.pyx":21
 *         self._closing = 0
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":24
 *         return 0
 * 
 *     cdef inline _schedule_call_connection_made(self):             # <<<<<<<<<<<<<<
 *         self._loop._call_soon_handle(
 *             new_MethodHandle(self._loop,
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_made(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_schedule_call_connection_made", 1);

  /* "uvloop/handles/basetransport.pyx":26
 *     cdef inline _schedule_call_connection_made(self):
 *         self._loop._call_soon_handle(
 *             new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                              "UVTransport._call_connection_made",
 *                              <method_t>self._call_connection_made,
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_1);

  /* "uvloop/handles/basetransport.pyx":29
 *                              "UVTransport._call_connection_made",
 *                              <method_t>self._call_connection_made,
 *                              self.context,             # <<<<<<<<<<<<<<
 *                              self))
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.context;
  __Pyx_INCREF(__pyx_t_2);

  /* "uvloop/handles/basetransport.pyx":26
 *     cdef inline _schedule_call_connection_made(self):
 *         self._loop._call_soon_handle(
 *             new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                              "UVTransport._call_connection_made",
 *                              <method_t>self._call_connection_made,
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_new_MethodHandle(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1), __pyx_kp_u_UVTransport__call_connection_mad, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_call_connection_made), __pyx_t_2, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(18, 26, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":25
 * 
 *     cdef inline _schedule_call_connection_made(self):
 *         self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *             new_MethodHandle(self._loop,
 *                              "UVTransport._call_connection_made",
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self->__pyx_base.__pyx_base._loop, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/basetransport.pyx":24
 *         return 0
 * 
 *     cdef inline _schedule_call_connection_made(self):             # <<<<<<<<<<<<<<
 *         self._loop._call_soon_handle(
 *             new_MethodHandle(self._loop,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._schedule_call_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":32
 *                              self))
 * 
 *     cdef inline _schedule_call_connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self._loop._call_soon_handle(
 *             new_MethodHandle1(self._loop,
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_lost(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_schedule_call_connection_lost", 1);

  /* "uvloop/handles/basetransport.pyx":34
 *     cdef inline _schedule_call_connection_lost(self, exc):
 *         self._loop._call_soon_handle(
 *             new_MethodHandle1(self._loop,             # <<<<<<<<<<<<<<
 *                               "UVTransport._call_connection_lost",
 *                               <method1_t>self._call_connection_lost,
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_1);

  /* "uvloop/handles/basetransport.pyx":37
 *                               "UVTransport._call_connection_lost",
 *                               <method1_t>self._call_connection_lost,
 *                               self.context,             # <<<<<<<<<<<<<<
 *                               self, exc))
 * 
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.context;
  __Pyx_INCREF(__pyx_t_2);

  /* "uvloop/handles/basetransport.pyx":34
 *     cdef inline _schedule_call_connection_lost(self, exc):
 *         self._loop._call_soon_handle(
 *             new_MethodHandle1(self._loop,             # <<<<<<<<<<<<<<
 *                               "UVTransport._call_connection_lost",
 *                               <method1_t>self._call_connection_lost,
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_new_MethodHandle1(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1), __pyx_kp_u_UVTransport__call_connection_los, ((__pyx_t_6uvloop_4loop_method1_t)((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_call_connection_lost), __pyx_t_2, ((PyObject *)__pyx_v_self), __pyx_v_exc); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(18, 34, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":33
 * 
 *     cdef inline _schedule_call_connection_lost(self, exc):
 *         self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *             new_MethodHandle1(self._loop,
 *                               "UVTransport._call_connection_lost",
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self->__pyx_base.__pyx_base._loop, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/basetransport.pyx":32
 *                              self))
 * 
 *     cdef inline _schedule_call_connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self._loop._call_soon_handle(
 *             new_MethodHandle1(self._loop,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._schedule_call_connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":40
 *                               self, exc))
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         # Overload UVHandle._fatal_error
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__fatal_error(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_15UVBaseTransport__fatal_error *__pyx_optional_args) {
  PyObject *__pyx_v_reason = ((PyObject *)Py_None);
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fatal_error", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_reason = __pyx_optional_args->reason;
    }
  }

  /* "uvloop/handles/basetransport.pyx":43
 *         # Overload UVHandle._fatal_error
 * 
 *         self._force_close(exc)             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(exc, OSError):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_force_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_exc};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":45
 *         self._force_close(exc)
 * 
 *         if not isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 * 
 *             if throw or self._loop is None:
 */
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_exc, __pyx_builtin_OSError); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(18, 45, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "uvloop/handles/basetransport.pyx":47
 *         if not isinstance(exc, OSError):
 * 
 *             if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_throw); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(18, 47, __pyx_L1_error)
    if (!__pyx_t_5) {
    } else {
      __pyx_t_6 = __pyx_t_5;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop) == Py_None);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L5_bool_binop_done:;
    if (unlikely(__pyx_t_6)) {

      /* "uvloop/handles/basetransport.pyx":48
 * 
 *             if throw or self._loop is None:
 *                 raise exc             # <<<<<<<<<<<<<<
 * 
 *             msg = f'Fatal error on transport {self.__class__.__name__}'
 */
      __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
      __PYX_ERR(18, 48, __pyx_L1_error)

      /* "uvloop/handles/basetransport.pyx":47
 *         if not isinstance(exc, OSError):
 * 
 *             if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
    }

    /* "uvloop/handles/basetransport.pyx":50
 *                 raise exc
 * 
 *             msg = f'Fatal error on transport {self.__class__.__name__}'             # <<<<<<<<<<<<<<
 *             if reason is not None:
 *                 msg = f'{msg} ({reason})'
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Fatal_error_on_transport, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_msg = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/basetransport.pyx":51
 * 
 *             msg = f'Fatal error on transport {self.__class__.__name__}'
 *             if reason is not None:             # <<<<<<<<<<<<<<
 *                 msg = f'{msg} ({reason})'
 * 
 */
    __pyx_t_6 = (__pyx_v_reason != Py_None);
    if (__pyx_t_6) {

      /* "uvloop/handles/basetransport.pyx":52
 *             msg = f'Fatal error on transport {self.__class__.__name__}'
 *             if reason is not None:
 *                 msg = f'{msg} ({reason})'             # <<<<<<<<<<<<<<
 * 
 *             self._loop.call_exception_handler({
 */
      __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = 0;
      __pyx_t_8 = 127;
      __Pyx_INCREF(__pyx_v_msg);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_msg) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_msg) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_v_msg);
      __Pyx_GIVEREF(__pyx_v_msg);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_msg);
      __Pyx_INCREF(__pyx_kp_u__140);
      __pyx_t_7 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__140);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u__140);
      __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_reason, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_8;
      __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u__141);
      __pyx_t_7 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__141);
      PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_kp_u__141);
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_2, 4, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_msg, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "uvloop/handles/basetransport.pyx":51
 * 
 *             msg = f'Fatal error on transport {self.__class__.__name__}'
 *             if reason is not None:             # <<<<<<<<<<<<<<
 *                 msg = f'{msg} ({reason})'
 * 
 */
    }

    /* "uvloop/handles/basetransport.pyx":54
 *                 msg = f'{msg} ({reason})'
 * 
 *             self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                 'message': msg,
 *                 'exception': exc,
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "uvloop/handles/basetransport.pyx":55
 * 
 *             self._loop.call_exception_handler({
 *                 'message': msg,             # <<<<<<<<<<<<<<
 *                 'exception': exc,
 *                 'transport': self,
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_message, __pyx_v_msg) < 0) __PYX_ERR(18, 55, __pyx_L1_error)

    /* "uvloop/handles/basetransport.pyx":56
 *             self._loop.call_exception_handler({
 *                 'message': msg,
 *                 'exception': exc,             # <<<<<<<<<<<<<<
 *                 'transport': self,
 *                 'protocol': self._protocol,
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(18, 55, __pyx_L1_error)

    /* "uvloop/handles/basetransport.pyx":57
 *                 'message': msg,
 *                 'exception': exc,
 *                 'transport': self,             # <<<<<<<<<<<<<<
 *                 'protocol': self._protocol,
 *             })
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_transport, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(18, 55, __pyx_L1_error)

    /* "uvloop/handles/basetransport.pyx":58
 *                 'exception': exc,
 *                 'transport': self,
 *                 'protocol': self._protocol,             # <<<<<<<<<<<<<<
 *             })
 * 
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_protocol, __pyx_v_self->_protocol) < 0) __PYX_ERR(18, 55, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 54, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/basetransport.pyx":45
 *         self._force_close(exc)
 * 
 *         if not isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 * 
 *             if throw or self._loop is None:
 */
  }

  /* "uvloop/handles/basetransport.pyx":40
 *                               self, exc))
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         # Overload UVHandle._fatal_error
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._fatal_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":61
 *             })
 * 
 *     cdef inline _maybe_pause_protocol(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t size = self._get_write_buffer_size()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_pause_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_maybe_pause_protocol", 1);

  /* "uvloop/handles/basetransport.pyx":63
 *     cdef inline _maybe_pause_protocol(self):
 *         cdef:
 *             size_t size = self._get_write_buffer_size()             # <<<<<<<<<<<<<<
 * 
 *         if size <= self._high_water:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_get_write_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 63, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "uvloop/handles/basetransport.pyx":65
 *             size_t size = self._get_write_buffer_size()
 * 
 *         if size <= self._high_water:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_2 = (__pyx_v_size <= __pyx_v_self->_high_water);
  if (__pyx_t_2) {

    /* "uvloop/handles/basetransport.pyx":66
 * 
 *         if size <= self._high_water:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if not self._protocol_paused:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":65
 *             size_t size = self._get_write_buffer_size()
 * 
 *         if size <= self._high_water:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/basetransport.pyx":68
 *             return
 * 
 *         if not self._protocol_paused:             # <<<<<<<<<<<<<<
 *             self._protocol_paused = 1
 *             try:
 */
  __pyx_t_2 = (!__pyx_v_self->_protocol_paused);
  if (__pyx_t_2) {

    /* "uvloop/handles/basetransport.pyx":69
 * 
 *         if not self._protocol_paused:
 *             self._protocol_paused = 1             # <<<<<<<<<<<<<<
 *             try:
 *                 # _maybe_pause_protocol() is always triggered from user-calls,
 */
    __pyx_v_self->_protocol_paused = 1;

    /* "uvloop/handles/basetransport.pyx":70
 *         if not self._protocol_paused:
 *             self._protocol_paused = 1
 *             try:             # <<<<<<<<<<<<<<
 *                 # _maybe_pause_protocol() is always triggered from user-calls,
 *                 # so we must copy the context to avoid entering context twice
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "uvloop/handles/basetransport.pyx":74
 *                 # so we must copy the context to avoid entering context twice
 *                 run_in_context(
 *                     self.context.copy(), self._protocol.pause_writing,             # <<<<<<<<<<<<<<
 *                 )
 *             except (KeyboardInterrupt, SystemExit):
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.context, __pyx_n_s_copy); if (unlikely(!__pyx_t_7)) __PYX_ERR(18, 74, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_9 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(18, 74, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_pause_writing); if (unlikely(!__pyx_t_7)) __PYX_ERR(18, 74, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "uvloop/handles/basetransport.pyx":73
 *                 # _maybe_pause_protocol() is always triggered from user-calls,
 *                 # so we must copy the context to avoid entering context twice
 *                 run_in_context(             # <<<<<<<<<<<<<<
 *                     self.context.copy(), self._protocol.pause_writing,
 *                 )
 */
        __pyx_t_8 = __pyx_f_6uvloop_4loop_run_in_context(__pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(18, 73, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "uvloop/handles/basetransport.pyx":70
 *         if not self._protocol_paused:
 *             self._protocol_paused = 1
 *             try:             # <<<<<<<<<<<<<<
 *                 # _maybe_pause_protocol() is always triggered from user-calls,
 *                 # so we must copy the context to avoid entering context twice
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/handles/basetransport.pyx":76
 *                     self.context.copy(), self._protocol.pause_writing,
 *                 )
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException as exc:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._maybe_pause_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(18, 76, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "uvloop/handles/basetransport.pyx":77
 *                 )
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 */
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ErrRestoreWithState(__pyx_t_8, __pyx_t_7, __pyx_t_6);
        __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; 
        __PYX_ERR(18, 77, __pyx_L7_except_error)
      }

      /* "uvloop/handles/basetransport.pyx":78
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._maybe_pause_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(18, 78, __pyx_L7_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_exc = __pyx_t_7;
        /*try:*/ {

          /* "uvloop/handles/basetransport.pyx":79
 *                 raise
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                     'message': 'protocol.pause_writing() failed',
 *                     'exception': exc,
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_12)) __PYX_ERR(18, 79, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_12);

          /* "uvloop/handles/basetransport.pyx":80
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',             # <<<<<<<<<<<<<<
 *                     'exception': exc,
 *                     'transport': self,
 */
          __pyx_t_13 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_13)) __PYX_ERR(18, 80, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (PyDict_SetItem(__pyx_t_13, __pyx_n_u_message, __pyx_kp_u_protocol_pause_writing_failed) < 0) __PYX_ERR(18, 80, __pyx_L18_error)

          /* "uvloop/handles/basetransport.pyx":81
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',
 *                     'exception': exc,             # <<<<<<<<<<<<<<
 *                     'transport': self,
 *                     'protocol': self._protocol,
 */
          if (PyDict_SetItem(__pyx_t_13, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(18, 80, __pyx_L18_error)

          /* "uvloop/handles/basetransport.pyx":82
 *                     'message': 'protocol.pause_writing() failed',
 *                     'exception': exc,
 *                     'transport': self,             # <<<<<<<<<<<<<<
 *                     'protocol': self._protocol,
 *                 })
 */
          if (PyDict_SetItem(__pyx_t_13, __pyx_n_u_transport, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(18, 80, __pyx_L18_error)

          /* "uvloop/handles/basetransport.pyx":83
 *                     'exception': exc,
 *                     'transport': self,
 *                     'protocol': self._protocol,             # <<<<<<<<<<<<<<
 *                 })
 * 
 */
          if (PyDict_SetItem(__pyx_t_13, __pyx_n_u_protocol, __pyx_v_self->_protocol) < 0) __PYX_ERR(18, 80, __pyx_L18_error)
          __pyx_t_14 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_t_13};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(18, 79, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }

        /* "uvloop/handles/basetransport.pyx":78
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L19;
          }
          __pyx_L18_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __pyx_t_10 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
            }
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
            __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
            goto __pyx_L7_except_error;
          }
          __pyx_L19:;
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;

      /* "uvloop/handles/basetransport.pyx":70
 *         if not self._protocol_paused:
 *             self._protocol_paused = 1
 *             try:             # <<<<<<<<<<<<<<
 *                 # _maybe_pause_protocol() is always triggered from user-calls,
 *                 # so we must copy the context to avoid entering context twice
 */
      __pyx_L7_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      __pyx_L10_try_end:;
    }

    /* "uvloop/handles/basetransport.pyx":68
 *             return
 * 
 *         if not self._protocol_paused:             # <<<<<<<<<<<<<<
 *             self._protocol_paused = 1
 *             try:
 */
  }

  /* "uvloop/handles/basetransport.pyx":61
 *             })
 * 
 *     cdef inline _maybe_pause_protocol(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t size = self._get_write_buffer_size()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._maybe_pause_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":86
 *                 })
 * 
 *     cdef inline _maybe_resume_protocol(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t size = self._get_write_buffer_size()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_resume_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_maybe_resume_protocol", 1);

  /* "uvloop/handles/basetransport.pyx":88
 *     cdef inline _maybe_resume_protocol(self):
 *         cdef:
 *             size_t size = self._get_write_buffer_size()             # <<<<<<<<<<<<<<
 * 
 *         if self._protocol_paused and size <= self._low_water:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_get_write_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 88, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "uvloop/handles/basetransport.pyx":90
 *             size_t size = self._get_write_buffer_size()
 * 
 *         if self._protocol_paused and size <= self._low_water:             # <<<<<<<<<<<<<<
 *             self._protocol_paused = 0
 *             try:
 */
  if (__pyx_v_self->_protocol_paused) {
  } else {
    __pyx_t_2 = __pyx_v_self->_protocol_paused;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_size <= __pyx_v_self->_low_water);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/handles/basetransport.pyx":91
 * 
 *         if self._protocol_paused and size <= self._low_water:
 *             self._protocol_paused = 0             # <<<<<<<<<<<<<<
 *             try:
 *                 # We're copying the context to avoid entering context twice,
 */
    __pyx_v_self->_protocol_paused = 0;

    /* "uvloop/handles/basetransport.pyx":92
 *         if self._protocol_paused and size <= self._low_water:
 *             self._protocol_paused = 0
 *             try:             # <<<<<<<<<<<<<<
 *                 # We're copying the context to avoid entering context twice,
 *                 # even though it's not always necessary to copy - it's easier
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "uvloop/handles/basetransport.pyx":97
 *                 # to copy here than passing down a copied context.
 *                 run_in_context(
 *                     self.context.copy(), self._protocol.resume_writing,             # <<<<<<<<<<<<<<
 *                 )
 *             except (KeyboardInterrupt, SystemExit):
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.context, __pyx_n_s_copy); if (unlikely(!__pyx_t_8)) __PYX_ERR(18, 97, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
          __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 0+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(18, 97, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_resume_writing); if (unlikely(!__pyx_t_8)) __PYX_ERR(18, 97, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_8);

        /* "uvloop/handles/basetransport.pyx":96
 *                 # even though it's not always necessary to copy - it's easier
 *                 # to copy here than passing down a copied context.
 *                 run_in_context(             # <<<<<<<<<<<<<<
 *                     self.context.copy(), self._protocol.resume_writing,
 *                 )
 */
        __pyx_t_9 = __pyx_f_6uvloop_4loop_run_in_context(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(18, 96, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "uvloop/handles/basetransport.pyx":92
 *         if self._protocol_paused and size <= self._low_water:
 *             self._protocol_paused = 0
 *             try:             # <<<<<<<<<<<<<<
 *                 # We're copying the context to avoid entering context twice,
 *                 # even though it's not always necessary to copy - it's easier
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/handles/basetransport.pyx":99
 *                     self.context.copy(), self._protocol.resume_writing,
 *                 )
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException as exc:
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._maybe_resume_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(18, 99, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "uvloop/handles/basetransport.pyx":100
 *                 )
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 */
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestoreWithState(__pyx_t_9, __pyx_t_8, __pyx_t_7);
        __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; 
        __PYX_ERR(18, 100, __pyx_L8_except_error)
      }

      /* "uvloop/handles/basetransport.pyx":101
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._maybe_resume_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(18, 101, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_v_exc = __pyx_t_8;
        /*try:*/ {

          /* "uvloop/handles/basetransport.pyx":102
 *                 raise
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                     'message': 'protocol.resume_writing() failed',
 *                     'exception': exc,
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_13)) __PYX_ERR(18, 102, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_13);

          /* "uvloop/handles/basetransport.pyx":103
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',             # <<<<<<<<<<<<<<
 *                     'exception': exc,
 *                     'transport': self,
 */
          __pyx_t_14 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_14)) __PYX_ERR(18, 103, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_message, __pyx_kp_u_protocol_resume_writing_failed) < 0) __PYX_ERR(18, 103, __pyx_L19_error)

          /* "uvloop/handles/basetransport.pyx":104
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',
 *                     'exception': exc,             # <<<<<<<<<<<<<<
 *                     'transport': self,
 *                     'protocol': self._protocol,
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(18, 103, __pyx_L19_error)

          /* "uvloop/handles/basetransport.pyx":105
 *                     'message': 'protocol.resume_writing() failed',
 *                     'exception': exc,
 *                     'transport': self,             # <<<<<<<<<<<<<<
 *                     'protocol': self._protocol,
 *                 })
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_transport, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(18, 103, __pyx_L19_error)

          /* "uvloop/handles/basetransport.pyx":106
 *                     'exception': exc,
 *                     'transport': self,
 *                     'protocol': self._protocol,             # <<<<<<<<<<<<<<
 *                 })
 * 
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_protocol, __pyx_v_self->_protocol) < 0) __PYX_ERR(18, 103, __pyx_L19_error)
          __pyx_t_15 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_14};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(18, 102, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }

        /* "uvloop/handles/basetransport.pyx":101
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L20;
          }
          __pyx_L19_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_23);
            __pyx_t_11 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
            }
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
            __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
            goto __pyx_L8_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;

      /* "uvloop/handles/basetransport.pyx":92
 *         if self._protocol_paused and size <= self._low_water:
 *             self._protocol_paused = 0
 *             try:             # <<<<<<<<<<<<<<
 *                 # We're copying the context to avoid entering context twice,
 *                 # even though it's not always necessary to copy - it's easier
 */
      __pyx_L8_except_error:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L7_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_L11_try_end:;
    }

    /* "uvloop/handles/basetransport.pyx":90
 *             size_t size = self._get_write_buffer_size()
 * 
 *         if self._protocol_paused and size <= self._low_water:             # <<<<<<<<<<<<<<
 *             self._protocol_paused = 0
 *             try:
 */
  }

  /* "uvloop/handles/basetransport.pyx":86
 *                 })
 * 
 *     cdef inline _maybe_resume_protocol(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             size_t size = self._get_write_buffer_size()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._maybe_resume_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":109
 *                 })
 * 
 *     cdef _wakeup_waiter(self):             # <<<<<<<<<<<<<<
 *         if self._waiter is not None:
 *             if not self._waiter.cancelled():
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__wakeup_waiter(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_wakeup_waiter", 1);

  /* "uvloop/handles/basetransport.pyx":110
 * 
 *     cdef _wakeup_waiter(self):
 *         if self._waiter is not None:             # <<<<<<<<<<<<<<
 *             if not self._waiter.cancelled():
 *                 if not self._is_alive():
 */
  __pyx_t_1 = (__pyx_v_self->_waiter != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":111
 *     cdef _wakeup_waiter(self):
 *         if self._waiter is not None:
 *             if not self._waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 if not self._is_alive():
 *                     self._waiter.set_exception(
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(18, 111, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_1);
    if (__pyx_t_6) {

      /* "uvloop/handles/basetransport.pyx":112
 *         if self._waiter is not None:
 *             if not self._waiter.cancelled():
 *                 if not self._is_alive():             # <<<<<<<<<<<<<<
 *                     self._waiter.set_exception(
 *                         RuntimeError(
 */
      __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 112, __pyx_L1_error)
      __pyx_t_1 = (!__pyx_t_6);
      if (__pyx_t_1) {

        /* "uvloop/handles/basetransport.pyx":113
 *             if not self._waiter.cancelled():
 *                 if not self._is_alive():
 *                     self._waiter.set_exception(             # <<<<<<<<<<<<<<
 *                         RuntimeError(
 *                             'closed Transport handle and unset waiter'))
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 113, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);

        /* "uvloop/handles/basetransport.pyx":114
 *                 if not self._is_alive():
 *                     self._waiter.set_exception(
 *                         RuntimeError(             # <<<<<<<<<<<<<<
 *                             'closed Transport handle and unset waiter'))
 *                 else:
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__142, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(18, 114, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 113, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/handles/basetransport.pyx":112
 *         if self._waiter is not None:
 *             if not self._waiter.cancelled():
 *                 if not self._is_alive():             # <<<<<<<<<<<<<<
 *                     self._waiter.set_exception(
 *                         RuntimeError(
 */
        goto __pyx_L5;
      }

      /* "uvloop/handles/basetransport.pyx":117
 *                             'closed Transport handle and unset waiter'))
 *                 else:
 *                     self._waiter.set_result(True)             # <<<<<<<<<<<<<<
 *             self._waiter = None
 * 
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, Py_True};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 117, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_L5:;

      /* "uvloop/handles/basetransport.pyx":111
 *     cdef _wakeup_waiter(self):
 *         if self._waiter is not None:
 *             if not self._waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 if not self._is_alive():
 *                     self._waiter.set_exception(
 */
    }

    /* "uvloop/handles/basetransport.pyx":118
 *                 else:
 *                     self._waiter.set_result(True)
 *             self._waiter = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _call_connection_made(self):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_waiter);
    __Pyx_DECREF(__pyx_v_self->_waiter);
    __pyx_v_self->_waiter = Py_None;

    /* "uvloop/handles/basetransport.pyx":110
 * 
 *     cdef _wakeup_waiter(self):
 *         if self._waiter is not None:             # <<<<<<<<<<<<<<
 *             if not self._waiter.cancelled():
 *                 if not self._is_alive():
 */
  }

  /* "uvloop/handles/basetransport.pyx":109
 *                 })
 * 
 *     cdef _wakeup_waiter(self):             # <<<<<<<<<<<<<<
 *         if self._waiter is not None:
 *             if not self._waiter.cancelled():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._wakeup_waiter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":120
 *             self._waiter = None
 * 
 *     cdef _call_connection_made(self):             # <<<<<<<<<<<<<<
 *         if self._protocol is None:
 *             raise RuntimeError(
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__call_connection_made(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_connection_made", 1);

  /* "uvloop/handles/basetransport.pyx":121
 * 
 *     cdef _call_connection_made(self):
 *         if self._protocol is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'protocol is not set, cannot call connection_made()')
 */
  __pyx_t_1 = (__pyx_v_self->_protocol == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/basetransport.pyx":122
 *     cdef _call_connection_made(self):
 *         if self._protocol is None:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'protocol is not set, cannot call connection_made()')
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__143, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(18, 122, __pyx_L1_error)

    /* "uvloop/handles/basetransport.pyx":121
 * 
 *     cdef _call_connection_made(self):
 *         if self._protocol is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'protocol is not set, cannot call connection_made()')
 */
  }

  /* "uvloop/handles/basetransport.pyx":128
 *         # `transport._close()` in `loop.create_connection()` if an
 *         # exception happens during `await waiter`.
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             # A connection waiter can be cancelled between
 *             # 'await loop.create_connection()' and
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 128, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_1);
  if (__pyx_t_3) {

    /* "uvloop/handles/basetransport.pyx":133
 *             # `_schedule_call_connection_made` and
 *             # the actual `_call_connection_made`.
 *             self._wakeup_waiter()             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_wakeup_waiter(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/basetransport.pyx":134
 *             # the actual `_call_connection_made`.
 *             self._wakeup_waiter()
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # Set _protocol_connected to 1 before calling "connection_made":
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":128
 *         # `transport._close()` in `loop.create_connection()` if an
 *         # exception happens during `await waiter`.
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             # A connection waiter can be cancelled between
 *             # 'await loop.create_connection()' and
 */
  }

  /* "uvloop/handles/basetransport.pyx":139
 *         # if transport is aborted or closed, "connection_lost" will
 *         # still be scheduled.
 *         self._protocol_connected = 1             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_v_self->_protocol_connected = 1;

  /* "uvloop/handles/basetransport.pyx":141
 *         self._protocol_connected = 1
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._protocol.connection_made(self)
 *         except BaseException:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "uvloop/handles/basetransport.pyx":142
 * 
 *         try:
 *             self._protocol.connection_made(self)             # <<<<<<<<<<<<<<
 *         except BaseException:
 *             self._wakeup_waiter()
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_connection_made); if (unlikely(!__pyx_t_7)) __PYX_ERR(18, 142, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, ((PyObject *)__pyx_v_self)};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 142, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/basetransport.pyx":141
 *         self._protocol_connected = 1
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._protocol.connection_made(self)
 *         except BaseException:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/handles/basetransport.pyx":143
 *         try:
 *             self._protocol.connection_made(self)
 *         except BaseException:             # <<<<<<<<<<<<<<
 *             self._wakeup_waiter()
 *             raise
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._call_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(18, 143, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);

      /* "uvloop/handles/basetransport.pyx":144
 *             self._protocol.connection_made(self)
 *         except BaseException:
 *             self._wakeup_waiter()             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_wakeup_waiter(__pyx_v_self); if (unlikely(!__pyx_t_11)) __PYX_ERR(18, 144, __pyx_L7_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "uvloop/handles/basetransport.pyx":145
 *         except BaseException:
 *             self._wakeup_waiter()
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if not self._is_alive():
 */
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_7, __pyx_t_8);
      __pyx_t_2 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; 
      __PYX_ERR(18, 145, __pyx_L7_except_error)
    }
    goto __pyx_L7_except_error;

    /* "uvloop/handles/basetransport.pyx":141
 *         self._protocol_connected = 1
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._protocol.connection_made(self)
 *         except BaseException:
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "uvloop/handles/basetransport.pyx":147
 *             raise
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             # This might happen when "transport.abort()" is called
 *             # from "Protocol.connection_made".
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 147, __pyx_L1_error)
  __pyx_t_1 = (!__pyx_t_3);
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":150
 *             # This might happen when "transport.abort()" is called
 *             # from "Protocol.connection_made".
 *             self._wakeup_waiter()             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_wakeup_waiter(__pyx_v_self); if (unlikely(!__pyx_t_8)) __PYX_ERR(18, 150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/handles/basetransport.pyx":151
 *             # from "Protocol.connection_made".
 *             self._wakeup_waiter()
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._start_reading()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":147
 *             raise
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             # This might happen when "transport.abort()" is called
 *             # from "Protocol.connection_made".
 */
  }

  /* "uvloop/handles/basetransport.pyx":153
 *             return
 * 
 *         self._start_reading()             # <<<<<<<<<<<<<<
 *         self._wakeup_waiter()
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_start_reading(__pyx_v_self); if (unlikely(!__pyx_t_8)) __PYX_ERR(18, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/handles/basetransport.pyx":154
 * 
 *         self._start_reading()
 *         self._wakeup_waiter()             # <<<<<<<<<<<<<<
 * 
 *     cdef _call_connection_lost(self, exc):
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_wakeup_waiter(__pyx_v_self); if (unlikely(!__pyx_t_8)) __PYX_ERR(18, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/handles/basetransport.pyx":120
 *             self._waiter = None
 * 
 *     cdef _call_connection_made(self):             # <<<<<<<<<<<<<<
 *         if self._protocol is None:
 *             raise RuntimeError(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._call_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":156
 *         self._wakeup_waiter()
 * 
 *     cdef _call_connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         if self._waiter is not None:
 *             if not self._waiter.done():
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__call_connection_lost(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc) {
  struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  char const *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_connection_lost", 1);

  /* "uvloop/handles/basetransport.pyx":157
 * 
 *     cdef _call_connection_lost(self, exc):
 *         if self._waiter is not None:             # <<<<<<<<<<<<<<
 *             if not self._waiter.done():
 *                 self._waiter.set_exception(exc)
 */
  __pyx_t_1 = (__pyx_v_self->_waiter != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":158
 *     cdef _call_connection_lost(self, exc):
 *         if self._waiter is not None:
 *             if not self._waiter.done():             # <<<<<<<<<<<<<<
 *                 self._waiter.set_exception(exc)
 *             self._waiter = None
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_done); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(18, 158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = (!__pyx_t_1);
    if (__pyx_t_6) {

      /* "uvloop/handles/basetransport.pyx":159
 *         if self._waiter is not None:
 *             if not self._waiter.done():
 *                 self._waiter.set_exception(exc)             # <<<<<<<<<<<<<<
 *             self._waiter = None
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_exc};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/basetransport.pyx":158
 *     cdef _call_connection_lost(self, exc):
 *         if self._waiter is not None:
 *             if not self._waiter.done():             # <<<<<<<<<<<<<<
 *                 self._waiter.set_exception(exc)
 *             self._waiter = None
 */
    }

    /* "uvloop/handles/basetransport.pyx":160
 *             if not self._waiter.done():
 *                 self._waiter.set_exception(exc)
 *             self._waiter = None             # <<<<<<<<<<<<<<
 * 
 *         if self._closed:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_waiter);
    __Pyx_DECREF(__pyx_v_self->_waiter);
    __pyx_v_self->_waiter = Py_None;

    /* "uvloop/handles/basetransport.pyx":157
 * 
 *     cdef _call_connection_lost(self, exc):
 *         if self._waiter is not None:             # <<<<<<<<<<<<<<
 *             if not self._waiter.done():
 *                 self._waiter.set_exception(exc)
 */
  }

  /* "uvloop/handles/basetransport.pyx":162
 *             self._waiter = None
 * 
 *         if self._closed:             # <<<<<<<<<<<<<<
 *             # The handle is closed -- likely, _call_connection_lost
 *             # was already called before.
 */
  if (__pyx_v_self->__pyx_base.__pyx_base._closed) {

    /* "uvloop/handles/basetransport.pyx":165
 *             # The handle is closed -- likely, _call_connection_lost
 *             # was already called before.
 *             return             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":162
 *             self._waiter = None
 * 
 *         if self._closed:             # <<<<<<<<<<<<<<
 *             # The handle is closed -- likely, _call_connection_lost
 *             # was already called before.
 */
  }

  /* "uvloop/handles/basetransport.pyx":167
 *             return
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self._protocol_connected:
 *                 self._protocol.connection_lost(exc)
 */
  /*try:*/ {

    /* "uvloop/handles/basetransport.pyx":168
 * 
 *         try:
 *             if self._protocol_connected:             # <<<<<<<<<<<<<<
 *                 self._protocol.connection_lost(exc)
 *         finally:
 */
    if (__pyx_v_self->_protocol_connected) {

      /* "uvloop/handles/basetransport.pyx":169
 *         try:
 *             if self._protocol_connected:
 *                 self._protocol.connection_lost(exc)             # <<<<<<<<<<<<<<
 *         finally:
 *             self._clear_protocol()
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_connection_lost); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 169, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_exc};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 169, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/basetransport.pyx":168
 * 
 *         try:
 *             if self._protocol_connected:             # <<<<<<<<<<<<<<
 *                 self._protocol.connection_lost(exc)
 *         finally:
 */
    }
  }

  /* "uvloop/handles/basetransport.pyx":171
 *                 self._protocol.connection_lost(exc)
 *         finally:
 *             self._clear_protocol()             # <<<<<<<<<<<<<<
 * 
 *             self._close()
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_clear_protocol(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/basetransport.pyx":173
 *             self._clear_protocol()
 * 
 *             self._close()             # <<<<<<<<<<<<<<
 * 
 *             server = self._server
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/basetransport.pyx":175
 *             self._close()
 * 
 *             server = self._server             # <<<<<<<<<<<<<<
 *             if server is not None:
 *                 (<Server>server)._detach()
 */
      __pyx_t_2 = ((PyObject *)__pyx_v_self->_server);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_v_server = ((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "uvloop/handles/basetransport.pyx":176
 * 
 *             server = self._server
 *             if server is not None:             # <<<<<<<<<<<<<<
 *                 (<Server>server)._detach()
 *                 self._server = None
 */
      __pyx_t_6 = (((PyObject *)__pyx_v_server) != Py_None);
      if (__pyx_t_6) {

        /* "uvloop/handles/basetransport.pyx":177
 *             server = self._server
 *             if server is not None:
 *                 (<Server>server)._detach()             # <<<<<<<<<<<<<<
 *                 self._server = None
 * 
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_server)->__pyx_vtab)->_detach(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_server)); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/handles/basetransport.pyx":178
 *             if server is not None:
 *                 (<Server>server)._detach()
 *                 self._server = None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _set_server(self, Server server):
 */
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF((PyObject *)__pyx_v_self->_server);
        __Pyx_DECREF((PyObject *)__pyx_v_self->_server);
        __pyx_v_self->_server = ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None);

        /* "uvloop/handles/basetransport.pyx":176
 * 
 *             server = self._server
 *             if server is not None:             # <<<<<<<<<<<<<<
 *                 (<Server>server)._detach()
 *                 self._server = None
 */
      }
      goto __pyx_L8;
    }
    __pyx_L7_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_7 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_9 = __pyx_filename;
      {

        /* "uvloop/handles/basetransport.pyx":171
 *                 self._protocol.connection_lost(exc)
 *         finally:
 *             self._clear_protocol()             # <<<<<<<<<<<<<<
 * 
 *             self._close()
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_clear_protocol(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 171, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/handles/basetransport.pyx":173
 *             self._clear_protocol()
 * 
 *             self._close()             # <<<<<<<<<<<<<<
 * 
 *             server = self._server
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 173, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/handles/basetransport.pyx":175
 *             self._close()
 * 
 *             server = self._server             # <<<<<<<<<<<<<<
 *             if server is not None:
 *                 (<Server>server)._detach()
 */
        __pyx_t_2 = ((PyObject *)__pyx_v_self->_server);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_server = ((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "uvloop/handles/basetransport.pyx":176
 * 
 *             server = self._server
 *             if server is not None:             # <<<<<<<<<<<<<<
 *                 (<Server>server)._detach()
 *                 self._server = None
 */
        __pyx_t_6 = (((PyObject *)__pyx_v_server) != Py_None);
        if (__pyx_t_6) {

          /* "uvloop/handles/basetransport.pyx":177
 *             server = self._server
 *             if server is not None:
 *                 (<Server>server)._detach()             # <<<<<<<<<<<<<<
 *                 self._server = None
 * 
 */
          __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_server)->__pyx_vtab)->_detach(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_server)); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 177, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "uvloop/handles/basetransport.pyx":178
 *             if server is not None:
 *                 (<Server>server)._detach()
 *                 self._server = None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _set_server(self, Server server):
 */
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          __Pyx_GOTREF((PyObject *)__pyx_v_self->_server);
          __Pyx_DECREF((PyObject *)__pyx_v_self->_server);
          __pyx_v_self->_server = ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None);

          /* "uvloop/handles/basetransport.pyx":176
 * 
 *             server = self._server
 *             if server is not None:             # <<<<<<<<<<<<<<
 *                 (<Server>server)._detach()
 *                 self._server = None
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_9;
      goto __pyx_L1_error;
      __pyx_L12_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L8:;
  }

  /* "uvloop/handles/basetransport.pyx":156
 *         self._wakeup_waiter()
 * 
 *     cdef _call_connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         if self._waiter is not None:
 *             if not self._waiter.done():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._call_connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_server);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":180
 *                 self._server = None
 * 
 *     cdef inline _set_server(self, Server server):             # <<<<<<<<<<<<<<
 *         self._server = server
 *         (<Server>server)._attach()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_server(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_server", 1);

  /* "uvloop/handles/basetransport.pyx":181
 * 
 *     cdef inline _set_server(self, Server server):
 *         self._server = server             # <<<<<<<<<<<<<<
 *         (<Server>server)._attach()
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_server);
  __Pyx_GIVEREF((PyObject *)__pyx_v_server);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_server);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_server);
  __pyx_v_self->_server = __pyx_v_server;

  /* "uvloop/handles/basetransport.pyx":182
 *     cdef inline _set_server(self, Server server):
 *         self._server = server
 *         (<Server>server)._attach()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _set_waiter(self, object waiter):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_server)->__pyx_vtab)->_attach(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_server)); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":180
 *                 self._server = None
 * 
 *     cdef inline _set_server(self, Server server):             # <<<<<<<<<<<<<<
 *         self._server = server
 *         (<Server>server)._attach()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._set_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":184
 *         (<Server>server)._attach()
 * 
 *     cdef inline _set_waiter(self, object waiter):             # <<<<<<<<<<<<<<
 *         if waiter is not None and not isfuture(waiter):
 *             raise TypeError(
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_waiter(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_waiter) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_waiter", 1);

  /* "uvloop/handles/basetransport.pyx":185
 * 
 *     cdef inline _set_waiter(self, object waiter):
 *         if waiter is not None and not isfuture(waiter):             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 f'invalid waiter object {waiter!r}, expected asyncio.Future')
 */
  __pyx_t_2 = (__pyx_v_waiter != Py_None);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __pyx_f_6uvloop_4loop_isfuture(__pyx_v_waiter); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(18, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/basetransport.pyx":187
 *         if waiter is not None and not isfuture(waiter):
 *             raise TypeError(
 *                 f'invalid waiter object {waiter!r}, expected asyncio.Future')             # <<<<<<<<<<<<<<
 * 
 *         self._waiter = waiter
 */
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = 0;
    __pyx_t_6 = 127;
    __Pyx_INCREF(__pyx_kp_u_invalid_waiter_object);
    __pyx_t_5 += 22;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_waiter_object);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_invalid_waiter_object);
    __pyx_t_7 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_waiter), __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(18, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_6;
    __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_kp_u_expected_asyncio_Future);
    __pyx_t_5 += 25;
    __Pyx_GIVEREF(__pyx_kp_u_expected_asyncio_Future);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_expected_asyncio_Future);
    __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(18, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/basetransport.pyx":186
 *     cdef inline _set_waiter(self, object waiter):
 *         if waiter is not None and not isfuture(waiter):
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 f'invalid waiter object {waiter!r}, expected asyncio.Future')
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(18, 186, __pyx_L1_error)

    /* "uvloop/handles/basetransport.pyx":185
 * 
 *     cdef inline _set_waiter(self, object waiter):
 *         if waiter is not None and not isfuture(waiter):             # <<<<<<<<<<<<<<
 *             raise TypeError(
 *                 f'invalid waiter object {waiter!r}, expected asyncio.Future')
 */
  }

  /* "uvloop/handles/basetransport.pyx":189
 *                 f'invalid waiter object {waiter!r}, expected asyncio.Future')
 * 
 *         self._waiter = waiter             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_protocol(self, object protocol):
 */
  __Pyx_INCREF(__pyx_v_waiter);
  __Pyx_GIVEREF(__pyx_v_waiter);
  __Pyx_GOTREF(__pyx_v_self->_waiter);
  __Pyx_DECREF(__pyx_v_self->_waiter);
  __pyx_v_self->_waiter = __pyx_v_waiter;

  /* "uvloop/handles/basetransport.pyx":184
 *         (<Server>server)._attach()
 * 
 *     cdef inline _set_waiter(self, object waiter):             # <<<<<<<<<<<<<<
 *         if waiter is not None and not isfuture(waiter):
 *             raise TypeError(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._set_waiter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":191
 *         self._waiter = waiter
 * 
 *     cdef _set_protocol(self, object protocol):             # <<<<<<<<<<<<<<
 *         self._protocol = protocol
 *         # Store a reference to the bound method directly
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__set_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_protocol", 1);

  /* "uvloop/handles/basetransport.pyx":192
 * 
 *     cdef _set_protocol(self, object protocol):
 *         self._protocol = protocol             # <<<<<<<<<<<<<<
 *         # Store a reference to the bound method directly
 *         try:
 */
  __Pyx_INCREF(__pyx_v_protocol);
  __Pyx_GIVEREF(__pyx_v_protocol);
  __Pyx_GOTREF(__pyx_v_self->_protocol);
  __Pyx_DECREF(__pyx_v_self->_protocol);
  __pyx_v_self->_protocol = __pyx_v_protocol;

  /* "uvloop/handles/basetransport.pyx":194
 *         self._protocol = protocol
 *         # Store a reference to the bound method directly
 *         try:             # <<<<<<<<<<<<<<
 *             self._protocol_data_received = protocol.data_received
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/handles/basetransport.pyx":195
 *         # Store a reference to the bound method directly
 *         try:
 *             self._protocol_data_received = protocol.data_received             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             pass
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_protocol, __pyx_n_s_data_received); if (unlikely(!__pyx_t_4)) __PYX_ERR(18, 195, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_v_self->_protocol_data_received);
      __Pyx_DECREF(__pyx_v_self->_protocol_data_received);
      __pyx_v_self->_protocol_data_received = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/handles/basetransport.pyx":194
 *         self._protocol = protocol
 *         # Store a reference to the bound method directly
 *         try:             # <<<<<<<<<<<<<<
 *             self._protocol_data_received = protocol.data_received
 *         except AttributeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/basetransport.pyx":196
 *         try:
 *             self._protocol_data_received = protocol.data_received
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_5) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/handles/basetransport.pyx":194
 *         self._protocol = protocol
 *         # Store a reference to the bound method directly
 *         try:             # <<<<<<<<<<<<<<
 *             self._protocol_data_received = protocol.data_received
 *         except AttributeError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/handles/basetransport.pyx":191
 *         self._waiter = waiter
 * 
 *     cdef _set_protocol(self, object protocol):             # <<<<<<<<<<<<<<
 *         self._protocol = protocol
 *         # Store a reference to the bound method directly
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._set_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":199
 *             pass
 * 
 *     cdef _clear_protocol(self):             # <<<<<<<<<<<<<<
 *         self._protocol = None
 *         self._protocol_data_received = None
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__clear_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_clear_protocol", 1);

  /* "uvloop/handles/basetransport.pyx":200
 * 
 *     cdef _clear_protocol(self):
 *         self._protocol = None             # <<<<<<<<<<<<<<
 *         self._protocol_data_received = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol);
  __Pyx_DECREF(__pyx_v_self->_protocol);
  __pyx_v_self->_protocol = Py_None;

  /* "uvloop/handles/basetransport.pyx":201
 *     cdef _clear_protocol(self):
 *         self._protocol = None
 *         self._protocol_data_received = None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _init_protocol(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol_data_received);
  __Pyx_DECREF(__pyx_v_self->_protocol_data_received);
  __pyx_v_self->_protocol_data_received = Py_None;

  /* "uvloop/handles/basetransport.pyx":199
 *             pass
 * 
 *     cdef _clear_protocol(self):             # <<<<<<<<<<<<<<
 *         self._protocol = None
 *         self._protocol_data_received = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":203
 *         self._protocol_data_received = None
 * 
 *     cdef inline _init_protocol(self):             # <<<<<<<<<<<<<<
 *         self._loop._track_transport(self)
 *         if self._protocol is None:
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__init_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init_protocol", 1);

  /* "uvloop/handles/basetransport.pyx":204
 * 
 *     cdef inline _init_protocol(self):
 *         self._loop._track_transport(self)             # <<<<<<<<<<<<<<
 *         if self._protocol is None:
 *             raise RuntimeError('invalid _init_protocol call')
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_base.__pyx_base._loop->__pyx_vtab)->_track_transport(__pyx_v_self->__pyx_base.__pyx_base._loop, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":205
 *     cdef inline _init_protocol(self):
 *         self._loop._track_transport(self)
 *         if self._protocol is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('invalid _init_protocol call')
 *         self._schedule_call_connection_made()
 */
  __pyx_t_2 = (__pyx_v_self->_protocol == Py_None);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/basetransport.pyx":206
 *         self._loop._track_transport(self)
 *         if self._protocol is None:
 *             raise RuntimeError('invalid _init_protocol call')             # <<<<<<<<<<<<<<
 *         self._schedule_call_connection_made()
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__144, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(18, 206, __pyx_L1_error)

    /* "uvloop/handles/basetransport.pyx":205
 *     cdef inline _init_protocol(self):
 *         self._loop._track_transport(self)
 *         if self._protocol is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('invalid _init_protocol call')
 *         self._schedule_call_connection_made()
 */
  }

  /* "uvloop/handles/basetransport.pyx":207
 *         if self._protocol is None:
 *             raise RuntimeError('invalid _init_protocol call')
 *         self._schedule_call_connection_made()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _add_extra_info(self, str name, object obj):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_made(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":203
 *         self._protocol_data_received = None
 * 
 *     cdef inline _init_protocol(self):             # <<<<<<<<<<<<<<
 *         self._loop._track_transport(self)
 *         if self._protocol is None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._init_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":209
 *         self._schedule_call_connection_made()
 * 
 *     cdef inline _add_extra_info(self, str name, object obj):             # <<<<<<<<<<<<<<
 *         if self._extra_info is None:
 *             self._extra_info = {}
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__add_extra_info(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_add_extra_info", 1);

  /* "uvloop/handles/basetransport.pyx":210
 * 
 *     cdef inline _add_extra_info(self, str name, object obj):
 *         if self._extra_info is None:             # <<<<<<<<<<<<<<
 *             self._extra_info = {}
 *         self._extra_info[name] = obj
 */
  __pyx_t_1 = (__pyx_v_self->_extra_info == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":211
 *     cdef inline _add_extra_info(self, str name, object obj):
 *         if self._extra_info is None:
 *             self._extra_info = {}             # <<<<<<<<<<<<<<
 *         self._extra_info[name] = obj
 * 
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_extra_info);
    __Pyx_DECREF(__pyx_v_self->_extra_info);
    __pyx_v_self->_extra_info = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/basetransport.pyx":210
 * 
 *     cdef inline _add_extra_info(self, str name, object obj):
 *         if self._extra_info is None:             # <<<<<<<<<<<<<<
 *             self._extra_info = {}
 *         self._extra_info[name] = obj
 */
  }

  /* "uvloop/handles/basetransport.pyx":212
 *         if self._extra_info is None:
 *             self._extra_info = {}
 *         self._extra_info[name] = obj             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_reading(self):
 */
  if (unlikely(__pyx_v_self->_extra_info == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(18, 212, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_self->_extra_info, __pyx_v_name, __pyx_v_obj) < 0))) __PYX_ERR(18, 212, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":209
 *         self._schedule_call_connection_made()
 * 
 *     cdef inline _add_extra_info(self, str name, object obj):             # <<<<<<<<<<<<<<
 *         if self._extra_info is None:
 *             self._extra_info = {}
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._add_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":214
 *         self._extra_info[name] = obj
 * 
 *     cdef bint _is_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static int __pyx_f_6uvloop_4loop_15UVBaseTransport__is_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uvloop/handles/basetransport.pyx":215
 * 
 *     cdef bint _is_reading(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     cdef _start_reading(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(18, 215, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":214
 *         self._extra_info[name] = obj
 * 
 *     cdef bint _is_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._is_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":217
 *         raise NotImplementedError
 * 
 *     cdef _start_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__start_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_start_reading", 1);

  /* "uvloop/handles/basetransport.pyx":218
 * 
 *     cdef _start_reading(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     cdef _stop_reading(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(18, 218, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":217
 *         raise NotImplementedError
 * 
 *     cdef _start_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._start_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":220
 *         raise NotImplementedError
 * 
 *     cdef _stop_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_15UVBaseTransport__stop_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_stop_reading", 1);

  /* "uvloop/handles/basetransport.pyx":221
 * 
 *     cdef _stop_reading(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     # === Public API ===
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(18, 221, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":220
 *         raise NotImplementedError
 * 
 *     cdef _stop_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._stop_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":227
 *     property _paused:
 *         # Used by SSLProto.  Might be removed in the future.
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return bool(not self._is_reading())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_7_paused_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_7_paused_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_7_paused___get__(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_7_paused___get__(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uvloop/handles/basetransport.pyx":228
 *         # Used by SSLProto.  Might be removed in the future.
 *         def __get__(self):
 *             return bool(not self._is_reading())             # <<<<<<<<<<<<<<
 * 
 *     def get_protocol(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_is_reading(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 228, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  __pyx_t_3 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/basetransport.pyx":227
 *     property _paused:
 *         # Used by SSLProto.  Might be removed in the future.
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return bool(not self._is_reading())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._paused.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":230
 *             return bool(not self._is_reading())
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_3get_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_2get_protocol, "UVBaseTransport.get_protocol(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_3get_protocol = {"get_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_3get_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_2get_protocol};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_3get_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_protocol (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_protocol", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_protocol", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_2get_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_2get_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_protocol", 1);

  /* "uvloop/handles/basetransport.pyx":231
 * 
 *     def get_protocol(self):
 *         return self._protocol             # <<<<<<<<<<<<<<
 * 
 *     def set_protocol(self, protocol):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_protocol);
  __pyx_r = __pyx_v_self->_protocol;
  goto __pyx_L0;

  /* "uvloop/handles/basetransport.pyx":230
 *             return bool(not self._is_reading())
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":233
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._set_protocol(protocol)
 *         if self._is_reading():
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_5set_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_4set_protocol, "UVBaseTransport.set_protocol(self, protocol)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_5set_protocol = {"set_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_5set_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_4set_protocol};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_5set_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_protocol (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 233, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_protocol") < 0)) __PYX_ERR(18, 233, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_protocol = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_protocol", 1, 1, 1, __pyx_nargs); __PYX_ERR(18, 233, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.set_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_4set_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_protocol);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_4set_protocol(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_protocol", 1);

  /* "uvloop/handles/basetransport.pyx":234
 * 
 *     def set_protocol(self, protocol):
 *         self._set_protocol(protocol)             # <<<<<<<<<<<<<<
 *         if self._is_reading():
 *             self._stop_reading()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_set_protocol(__pyx_v_self, __pyx_v_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":235
 *     def set_protocol(self, protocol):
 *         self._set_protocol(protocol)
 *         if self._is_reading():             # <<<<<<<<<<<<<<
 *             self._stop_reading()
 *             self._start_reading()
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_is_reading(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 235, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "uvloop/handles/basetransport.pyx":236
 *         self._set_protocol(protocol)
 *         if self._is_reading():
 *             self._stop_reading()             # <<<<<<<<<<<<<<
 *             self._start_reading()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_stop_reading(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 236, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/basetransport.pyx":237
 *         if self._is_reading():
 *             self._stop_reading()
 *             self._start_reading()             # <<<<<<<<<<<<<<
 * 
 *     def _force_close(self, exc):
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_start_reading(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/basetransport.pyx":235
 *     def set_protocol(self, protocol):
 *         self._set_protocol(protocol)
 *         if self._is_reading():             # <<<<<<<<<<<<<<
 *             self._stop_reading()
 *             self._start_reading()
 */
  }

  /* "uvloop/handles/basetransport.pyx":233
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._set_protocol(protocol)
 *         if self._is_reading():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.set_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":239
 *             self._start_reading()
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         # Used by SSLProto.  Might be removed in the future.
 *         if self._conn_lost or self._closed:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_7_force_close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_6_force_close, "UVBaseTransport._force_close(self, exc)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_7_force_close = {"_force_close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_7_force_close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_6_force_close};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_7_force_close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_exc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_force_close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 239, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_force_close") < 0)) __PYX_ERR(18, 239, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_exc = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_force_close", 1, 1, 1, __pyx_nargs); __PYX_ERR(18, 239, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._force_close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_6_force_close(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_exc);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_6_force_close(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_force_close", 1);

  /* "uvloop/handles/basetransport.pyx":241
 *     def _force_close(self, exc):
 *         # Used by SSLProto.  Might be removed in the future.
 *         if self._conn_lost or self._closed:             # <<<<<<<<<<<<<<
 *             return
 *         if not self._closing:
 */
  __pyx_t_2 = (__pyx_v_self->_conn_lost != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base._closed;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":242
 *         # Used by SSLProto.  Might be removed in the future.
 *         if self._conn_lost or self._closed:
 *             return             # <<<<<<<<<<<<<<
 *         if not self._closing:
 *             self._closing = 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":241
 *     def _force_close(self, exc):
 *         # Used by SSLProto.  Might be removed in the future.
 *         if self._conn_lost or self._closed:             # <<<<<<<<<<<<<<
 *             return
 *         if not self._closing:
 */
  }

  /* "uvloop/handles/basetransport.pyx":243
 *         if self._conn_lost or self._closed:
 *             return
 *         if not self._closing:             # <<<<<<<<<<<<<<
 *             self._closing = 1
 *             self._stop_reading()
 */
  __pyx_t_1 = (!__pyx_v_self->_closing);
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":244
 *             return
 *         if not self._closing:
 *             self._closing = 1             # <<<<<<<<<<<<<<
 *             self._stop_reading()
 *         self._conn_lost += 1
 */
    __pyx_v_self->_closing = 1;

    /* "uvloop/handles/basetransport.pyx":245
 *         if not self._closing:
 *             self._closing = 1
 *             self._stop_reading()             # <<<<<<<<<<<<<<
 *         self._conn_lost += 1
 *         self._schedule_call_connection_lost(exc)
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_stop_reading(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/basetransport.pyx":243
 *         if self._conn_lost or self._closed:
 *             return
 *         if not self._closing:             # <<<<<<<<<<<<<<
 *             self._closing = 1
 *             self._stop_reading()
 */
  }

  /* "uvloop/handles/basetransport.pyx":246
 *             self._closing = 1
 *             self._stop_reading()
 *         self._conn_lost += 1             # <<<<<<<<<<<<<<
 *         self._schedule_call_connection_lost(exc)
 * 
 */
  __pyx_v_self->_conn_lost = (__pyx_v_self->_conn_lost + 1);

  /* "uvloop/handles/basetransport.pyx":247
 *             self._stop_reading()
 *         self._conn_lost += 1
 *         self._schedule_call_connection_lost(exc)             # <<<<<<<<<<<<<<
 * 
 *     def abort(self):
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_lost(__pyx_v_self, __pyx_v_exc); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/basetransport.pyx":239
 *             self._start_reading()
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         # Used by SSLProto.  Might be removed in the future.
 *         if self._conn_lost or self._closed:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._force_close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":249
 *         self._schedule_call_connection_lost(exc)
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         self._force_close(None)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_9abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_8abort, "UVBaseTransport.abort(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_9abort = {"abort", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_9abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_8abort};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_9abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("abort (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("abort", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "abort", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_8abort(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_8abort(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("abort", 1);

  /* "uvloop/handles/basetransport.pyx":250
 * 
 *     def abort(self):
 *         self._force_close(None)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_force_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, Py_None};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":249
 *         self._schedule_call_connection_lost(exc)
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         self._force_close(None)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.abort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":252
 *         self._force_close(None)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._closing or self._closed:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_11close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_10close, "UVBaseTransport.close(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_11close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_11close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_10close};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_11close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_10close(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_10close(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/handles/basetransport.pyx":253
 * 
 *     def close(self):
 *         if self._closing or self._closed:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (!__pyx_v_self->_closing) {
  } else {
    __pyx_t_1 = __pyx_v_self->_closing;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base._closed;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":254
 *     def close(self):
 *         if self._closing or self._closed:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._closing = 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":253
 * 
 *     def close(self):
 *         if self._closing or self._closed:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/basetransport.pyx":256
 *             return
 * 
 *         self._closing = 1             # <<<<<<<<<<<<<<
 *         self._stop_reading()
 * 
 */
  __pyx_v_self->_closing = 1;

  /* "uvloop/handles/basetransport.pyx":257
 * 
 *         self._closing = 1
 *         self._stop_reading()             # <<<<<<<<<<<<<<
 * 
 *         if not self._get_write_buffer_size():
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_stop_reading(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/basetransport.pyx":259
 *         self._stop_reading()
 * 
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             # The write buffer is empty
 *             self._conn_lost += 1
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_get_write_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 259, __pyx_L1_error)
  __pyx_t_1 = (!(__pyx_t_3 != 0));
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":261
 *         if not self._get_write_buffer_size():
 *             # The write buffer is empty
 *             self._conn_lost += 1             # <<<<<<<<<<<<<<
 *             self._schedule_call_connection_lost(None)
 * 
 */
    __pyx_v_self->_conn_lost = (__pyx_v_self->_conn_lost + 1);

    /* "uvloop/handles/basetransport.pyx":262
 *             # The write buffer is empty
 *             self._conn_lost += 1
 *             self._schedule_call_connection_lost(None)             # <<<<<<<<<<<<<<
 * 
 *     def is_closing(self):
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_lost(__pyx_v_self, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/basetransport.pyx":259
 *         self._stop_reading()
 * 
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             # The write buffer is empty
 *             self._conn_lost += 1
 */
  }

  /* "uvloop/handles/basetransport.pyx":252
 *         self._force_close(None)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._closing or self._closed:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":264
 *             self._schedule_call_connection_lost(None)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closing
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_13is_closing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_12is_closing, "UVBaseTransport.is_closing(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_13is_closing = {"is_closing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_13is_closing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_12is_closing};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_13is_closing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_closing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_closing", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_closing", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_12is_closing(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_12is_closing(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_closing", 1);

  /* "uvloop/handles/basetransport.pyx":265
 * 
 *     def is_closing(self):
 *         return self._closing             # <<<<<<<<<<<<<<
 * 
 *     def get_write_buffer_size(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_closing); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/basetransport.pyx":264
 *             self._schedule_call_connection_lost(None)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closing
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.is_closing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":267
 *         return self._closing
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._get_write_buffer_size()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_15get_write_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_14get_write_buffer_size, "UVBaseTransport.get_write_buffer_size(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_15get_write_buffer_size = {"get_write_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_15get_write_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_14get_write_buffer_size};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_15get_write_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_write_buffer_size (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_write_buffer_size", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_write_buffer_size", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_14get_write_buffer_size(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_14get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_write_buffer_size", 1);

  /* "uvloop/handles/basetransport.pyx":268
 * 
 *     def get_write_buffer_size(self):
 *         return self._get_write_buffer_size()             # <<<<<<<<<<<<<<
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_get_write_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 268, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/basetransport.pyx":267
 *         return self._closing
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._get_write_buffer_size()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.get_write_buffer_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":270
 *         return self._get_write_buffer_size()
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_17set_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_16set_write_buffer_limits, "UVBaseTransport.set_write_buffer_limits(self, high=None, low=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_17set_write_buffer_limits = {"set_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_17set_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_16set_write_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_17set_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_high = 0;
  PyObject *__pyx_v_low = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_write_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_high_2,&__pyx_n_s_low,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_high_2);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 270, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_low);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 270, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_write_buffer_limits") < 0)) __PYX_ERR(18, 270, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_high = values[0];
    __pyx_v_low = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_write_buffer_limits", 0, 0, 2, __pyx_nargs); __PYX_ERR(18, 270, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.set_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_16set_write_buffer_limits(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_high, __pyx_v_low);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_16set_write_buffer_limits(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_high, PyObject *__pyx_v_low) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_write_buffer_limits", 1);

  /* "uvloop/handles/basetransport.pyx":271
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         self._high_water, self._low_water = add_flowcontrol_defaults(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":273
 *         self._ensure_alive()
 * 
 *         self._high_water, self._low_water = add_flowcontrol_defaults(             # <<<<<<<<<<<<<<
 *             high, low, FLOW_CONTROL_HIGH_WATER)
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_8includes_11flowcontrol_add_flowcontrol_defaults(__pyx_v_high, __pyx_v_low, __pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(18, 273, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(18, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(18, 273, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(18, 273, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(18, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_t_3); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(18, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->_high_water = __pyx_t_6;
  __pyx_v_self->_low_water = __pyx_t_7;

  /* "uvloop/handles/basetransport.pyx":276
 *             high, low, FLOW_CONTROL_HIGH_WATER)
 * 
 *         self._maybe_pause_protocol()             # <<<<<<<<<<<<<<
 * 
 *     def get_write_buffer_limits(self):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_pause_protocol(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/basetransport.pyx":270
 *         return self._get_write_buffer_size()
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.set_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":278
 *         self._maybe_pause_protocol()
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._low_water, self._high_water)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_19get_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_18get_write_buffer_limits, "UVBaseTransport.get_write_buffer_limits(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_19get_write_buffer_limits = {"get_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_19get_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_18get_write_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_19get_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_write_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_write_buffer_limits", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_write_buffer_limits", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_18get_write_buffer_limits(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_18get_write_buffer_limits(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_write_buffer_limits", 1);

  /* "uvloop/handles/basetransport.pyx":279
 * 
 *     def get_write_buffer_limits(self):
 *         return (self._low_water, self._high_water)             # <<<<<<<<<<<<<<
 * 
 *     def get_extra_info(self, name, default=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_low_water); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_high_water); if (unlikely(!__pyx_t_2)) __PYX_ERR(18, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(18, 279, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(18, 279, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/basetransport.pyx":278
 *         self._maybe_pause_protocol()
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._low_water, self._high_water)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.get_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pyx":281
 *         return (self._low_water, self._high_water)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if self._extra_info is not None and name in self._extra_info:
 *             return self._extra_info[name]
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_21get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_20get_extra_info, "UVBaseTransport.get_extra_info(self, name, default=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_21get_extra_info = {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_21get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_20get_extra_info};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_21get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_default = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_extra_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_default,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name_2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 281, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_default);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(18, 281, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_extra_info") < 0)) __PYX_ERR(18, 281, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_extra_info", 0, 1, 2, __pyx_nargs); __PYX_ERR(18, 281, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_20get_extra_info(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_name, __pyx_v_default);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_20get_extra_info(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_extra_info", 1);

  /* "uvloop/handles/basetransport.pyx":282
 * 
 *     def get_extra_info(self, name, default=None):
 *         if self._extra_info is not None and name in self._extra_info:             # <<<<<<<<<<<<<<
 *             return self._extra_info[name]
 *         if name == 'socket':
 */
  __pyx_t_2 = (__pyx_v_self->_extra_info != ((PyObject*)Py_None));
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  if (unlikely(__pyx_v_self->_extra_info == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(18, 282, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_v_name, __pyx_v_self->_extra_info, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(18, 282, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":283
 *     def get_extra_info(self, name, default=None):
 *         if self._extra_info is not None and name in self._extra_info:
 *             return self._extra_info[name]             # <<<<<<<<<<<<<<
 *         if name == 'socket':
 *             return self._get_socket()
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_self->_extra_info == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(18, 283, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_self->_extra_info, __pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":282
 * 
 *     def get_extra_info(self, name, default=None):
 *         if self._extra_info is not None and name in self._extra_info:             # <<<<<<<<<<<<<<
 *             return self._extra_info[name]
 *         if name == 'socket':
 */
  }

  /* "uvloop/handles/basetransport.pyx":284
 *         if self._extra_info is not None and name in self._extra_info:
 *             return self._extra_info[name]
 *         if name == 'socket':             # <<<<<<<<<<<<<<
 *             return self._get_socket()
 *         if name == 'sockname':
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_n_u_socket_2, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(18, 284, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":285
 *             return self._extra_info[name]
 *         if name == 'socket':
 *             return self._get_socket()             # <<<<<<<<<<<<<<
 *         if name == 'sockname':
 *             return self._get_socket().getsockname()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":284
 *         if self._extra_info is not None and name in self._extra_info:
 *             return self._extra_info[name]
 *         if name == 'socket':             # <<<<<<<<<<<<<<
 *             return self._get_socket()
 *         if name == 'sockname':
 */
  }

  /* "uvloop/handles/basetransport.pyx":286
 *         if name == 'socket':
 *             return self._get_socket()
 *         if name == 'sockname':             # <<<<<<<<<<<<<<
 *             return self._get_socket().getsockname()
 *         if name == 'peername':
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_n_u_sockname, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(18, 286, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":287
 *             return self._get_socket()
 *         if name == 'sockname':
 *             return self._get_socket().getsockname()             # <<<<<<<<<<<<<<
 *         if name == 'peername':
 *             try:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(18, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_getsockname); if (unlikely(!__pyx_t_5)) __PYX_ERR(18, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uvloop/handles/basetransport.pyx":286
 *         if name == 'socket':
 *             return self._get_socket()
 *         if name == 'sockname':             # <<<<<<<<<<<<<<
 *             return self._get_socket().getsockname()
 *         if name == 'peername':
 */
  }

  /* "uvloop/handles/basetransport.pyx":288
 *         if name == 'sockname':
 *             return self._get_socket().getsockname()
 *         if name == 'peername':             # <<<<<<<<<<<<<<
 *             try:
 *                 return self._get_socket().getpeername()
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_n_u_peername, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(18, 288, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/basetransport.pyx":289
 *             return self._get_socket().getsockname()
 *         if name == 'peername':
 *             try:             # <<<<<<<<<<<<<<
 *                 return self._get_socket().getpeername()
 *             except socket_error:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "uvloop/handles/basetransport.pyx":290
 *         if name == 'peername':
 *             try:
 *                 return self._get_socket().getpeername()             # <<<<<<<<<<<<<<
 *             except socket_error:
 *                 return default
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_5)) __PYX_ERR(18, 290, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_getpeername); if (unlikely(!__pyx_t_4)) __PYX_ERR(18, 290, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 290, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __pyx_r = __pyx_t_3;
        __pyx_t_3 = 0;
        goto __pyx_L13_try_return;

        /* "uvloop/handles/basetransport.pyx":289
 *             return self._get_socket().getsockname()
 *         if name == 'peername':
 *             try:             # <<<<<<<<<<<<<<
 *                 return self._get_socket().getpeername()
 *             except socket_error:
 */
      }
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/handles/basetransport.pyx":291
 *             try:
 *                 return self._get_socket().getpeername()
 *             except socket_error:             # <<<<<<<<<<<<<<
 *                 return default
 *         return default
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_socket_error);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5) < 0) __PYX_ERR(18, 291, __pyx_L11_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);

        /* "uvloop/handles/basetransport.pyx":292
 *                 return self._get_socket().getpeername()
 *             except socket_error:
 *                 return default             # <<<<<<<<<<<<<<
 *         return default
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_default);
        __pyx_r = __pyx_v_default;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L12_except_return;
      }
      goto __pyx_L11_except_error;

      /* "uvloop/handles/basetransport.pyx":289
 *             return self._get_socket().getsockname()
 *         if name == 'peername':
 *             try:             # <<<<<<<<<<<<<<
 *                 return self._get_socket().getpeername()
 *             except socket_error:
 */
      __pyx_L11_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L13_try_return:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L0;
      __pyx_L12_except_return:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L0;
    }

    /* "uvloop/handles/basetransport.pyx":288
 *         if name == 'sockname':
 *             return self._get_socket().getsockname()
 *         if name == 'peername':             # <<<<<<<<<<<<<<
 *             try:
 *                 return self._get_socket().getpeername()
 */
  }

  /* "uvloop/handles/basetransport.pyx":293
 *             except socket_error:
 *                 return default
 *         return default             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_default);
  __pyx_r = __pyx_v_default;
  goto __pyx_L0;

  /* "uvloop/handles/basetransport.pyx":281
 *         return (self._low_water, self._high_water)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if self._extra_info is not None and name in self._extra_info:
 *             return self._extra_info[name]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/basetransport.pxd":4
 * 
 *     cdef:
 *         readonly bint _closing             # <<<<<<<<<<<<<<
 * 
 *         bint _protocol_connected
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_8_closing_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_8_closing_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_8_closing___get__(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_8_closing___get__(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_closing); if (unlikely(!__pyx_t_1)) __PYX_ERR(19, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport._closing.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_22__reduce_cython__, "UVBaseTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_23__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_22__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_22__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15UVBaseTransport_24__setstate_cython__, "UVBaseTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15UVBaseTransport_25__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_24__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15UVBaseTransport_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15UVBaseTransport_24__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15UVBaseTransport_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVBaseTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVBaseTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":33
 *         bint            closed
 * 
 *     cdef free_bufs(self):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_19_StreamWriteContext_free_bufs(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("free_bufs", 1);

  /* "uvloop/handles/stream.pyx":36
 *         cdef size_t i
 * 
 *         if self.uv_bufs is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_RawFree(self.uv_bufs)
 *             self.uv_bufs = NULL
 */
  __pyx_t_1 = (__pyx_v_self->uv_bufs != NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":37
 * 
 *         if self.uv_bufs is not NULL:
 *             PyMem_RawFree(self.uv_bufs)             # <<<<<<<<<<<<<<
 *             self.uv_bufs = NULL
 *             if UVLOOP_DEBUG:
 */
    PyMem_RawFree(__pyx_v_self->uv_bufs);

    /* "uvloop/handles/stream.pyx":38
 *         if self.uv_bufs is not NULL:
 *             PyMem_RawFree(self.uv_bufs)
 *             self.uv_bufs = NULL             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:
 */
    __pyx_v_self->uv_bufs = NULL;

    /* "uvloop/handles/stream.pyx":39
 *             PyMem_RawFree(self.uv_bufs)
 *             self.uv_bufs = NULL
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(
 */
    __pyx_t_1 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":40
 *             self.uv_bufs = NULL
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_StreamWriteContext.close: uv_bufs != NULL and '
 */
      if (unlikely(__pyx_v_self->py_bufs_sml_inuse)) {

        /* "uvloop/handles/stream.pyx":41
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_StreamWriteContext.close: uv_bufs != NULL and '
 *                         'py_bufs_sml_inuse is True')
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__145, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 41, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(20, 41, __pyx_L1_error)

        /* "uvloop/handles/stream.pyx":40
 *             self.uv_bufs = NULL
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_StreamWriteContext.close: uv_bufs != NULL and '
 */
      }

      /* "uvloop/handles/stream.pyx":39
 *             PyMem_RawFree(self.uv_bufs)
 *             self.uv_bufs = NULL
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(
 */
    }

    /* "uvloop/handles/stream.pyx":36
 *         cdef size_t i
 * 
 *         if self.uv_bufs is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_RawFree(self.uv_bufs)
 *             self.uv_bufs = NULL
 */
  }

  /* "uvloop/handles/stream.pyx":45
 *                         'py_bufs_sml_inuse is True')
 * 
 *         if self.py_bufs is not NULL:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs[i])
 */
  __pyx_t_1 = (__pyx_v_self->py_bufs != NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":46
 * 
 *         if self.py_bufs is not NULL:
 *             for i from 0 <= i < self.py_bufs_len:             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&self.py_bufs[i])
 *             PyMem_RawFree(self.py_bufs)
 */
    __pyx_t_3 = __pyx_v_self->py_bufs_len;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

      /* "uvloop/handles/stream.pyx":47
 *         if self.py_bufs is not NULL:
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs[i])             # <<<<<<<<<<<<<<
 *             PyMem_RawFree(self.py_bufs)
 *             self.py_bufs = NULL
 */
      PyBuffer_Release((&(__pyx_v_self->py_bufs[__pyx_v_i])));
    }

    /* "uvloop/handles/stream.pyx":48
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs[i])
 *             PyMem_RawFree(self.py_bufs)             # <<<<<<<<<<<<<<
 *             self.py_bufs = NULL
 *             if UVLOOP_DEBUG:
 */
    PyMem_RawFree(__pyx_v_self->py_bufs);

    /* "uvloop/handles/stream.pyx":49
 *                 PyBuffer_Release(&self.py_bufs[i])
 *             PyMem_RawFree(self.py_bufs)
 *             self.py_bufs = NULL             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:
 */
    __pyx_v_self->py_bufs = NULL;

    /* "uvloop/handles/stream.pyx":50
 *             PyMem_RawFree(self.py_bufs)
 *             self.py_bufs = NULL
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(
 */
    __pyx_t_1 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":51
 *             self.py_bufs = NULL
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_StreamWriteContext.close: py_bufs != NULL and '
 */
      if (unlikely(__pyx_v_self->py_bufs_sml_inuse)) {

        /* "uvloop/handles/stream.pyx":52
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_StreamWriteContext.close: py_bufs != NULL and '
 *                         'py_bufs_sml_inuse is True')
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__146, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 52, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(20, 52, __pyx_L1_error)

        /* "uvloop/handles/stream.pyx":51
 *             self.py_bufs = NULL
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_StreamWriteContext.close: py_bufs != NULL and '
 */
      }

      /* "uvloop/handles/stream.pyx":50
 *             PyMem_RawFree(self.py_bufs)
 *             self.py_bufs = NULL
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(
 */
    }

    /* "uvloop/handles/stream.pyx":45
 *                         'py_bufs_sml_inuse is True')
 * 
 *         if self.py_bufs is not NULL:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs[i])
 */
  }

  /* "uvloop/handles/stream.pyx":56
 *                         'py_bufs_sml_inuse is True')
 * 
 *         if self.py_bufs_sml_inuse:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs_sml[i])
 */
  if (__pyx_v_self->py_bufs_sml_inuse) {

    /* "uvloop/handles/stream.pyx":57
 * 
 *         if self.py_bufs_sml_inuse:
 *             for i from 0 <= i < self.py_bufs_len:             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&self.py_bufs_sml[i])
 *             self.py_bufs_sml_inuse = 0
 */
    __pyx_t_3 = __pyx_v_self->py_bufs_len;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

      /* "uvloop/handles/stream.pyx":58
 *         if self.py_bufs_sml_inuse:
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs_sml[i])             # <<<<<<<<<<<<<<
 *             self.py_bufs_sml_inuse = 0
 * 
 */
      PyBuffer_Release((&(__pyx_v_self->py_bufs_sml[__pyx_v_i])));
    }

    /* "uvloop/handles/stream.pyx":59
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs_sml[i])
 *             self.py_bufs_sml_inuse = 0             # <<<<<<<<<<<<<<
 * 
 *         self.py_bufs_len = 0
 */
    __pyx_v_self->py_bufs_sml_inuse = 0;

    /* "uvloop/handles/stream.pyx":56
 *                         'py_bufs_sml_inuse is True')
 * 
 *         if self.py_bufs_sml_inuse:             # <<<<<<<<<<<<<<
 *             for i from 0 <= i < self.py_bufs_len:
 *                 PyBuffer_Release(&self.py_bufs_sml[i])
 */
  }

  /* "uvloop/handles/stream.pyx":61
 *             self.py_bufs_sml_inuse = 0
 * 
 *         self.py_bufs_len = 0             # <<<<<<<<<<<<<<
 *         self.buffers = None
 * 
 */
  __pyx_v_self->py_bufs_len = 0;

  /* "uvloop/handles/stream.pyx":62
 * 
 *         self.py_bufs_len = 0
 *         self.buffers = None             # <<<<<<<<<<<<<<
 * 
 *     cdef close(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->buffers);
  __Pyx_DECREF(__pyx_v_self->buffers);
  __pyx_v_self->buffers = ((PyObject*)Py_None);

  /* "uvloop/handles/stream.pyx":33
 *         bint            closed
 * 
 *     cdef free_bufs(self):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.free_bufs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":64
 *         self.buffers = None
 * 
 *     cdef close(self):             # <<<<<<<<<<<<<<
 *         if self.closed:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_19_StreamWriteContext_close(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/handles/stream.pyx":65
 * 
 *     cdef close(self):
 *         if self.closed:             # <<<<<<<<<<<<<<
 *             return
 *         self.closed = 1
 */
  if (__pyx_v_self->closed) {

    /* "uvloop/handles/stream.pyx":66
 *     cdef close(self):
 *         if self.closed:
 *             return             # <<<<<<<<<<<<<<
 *         self.closed = 1
 *         self.free_bufs()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":65
 * 
 *     cdef close(self):
 *         if self.closed:             # <<<<<<<<<<<<<<
 *             return
 *         self.closed = 1
 */
  }

  /* "uvloop/handles/stream.pyx":67
 *         if self.closed:
 *             return
 *         self.closed = 1             # <<<<<<<<<<<<<<
 *         self.free_bufs()
 *         Py_DECREF(self)
 */
  __pyx_v_self->closed = 1;

  /* "uvloop/handles/stream.pyx":68
 *             return
 *         self.closed = 1
 *         self.free_bufs()             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_self->__pyx_vtab)->free_bufs(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":69
 *         self.closed = 1
 *         self.free_bufs()
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef advance_uv_buf(self, size_t sent):
 */
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "uvloop/handles/stream.pyx":64
 *         self.buffers = None
 * 
 *     cdef close(self):             # <<<<<<<<<<<<<<
 *         if self.closed:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":71
 *         Py_DECREF(self)
 * 
 *     cdef advance_uv_buf(self, size_t sent):             # <<<<<<<<<<<<<<
 *         # Advance the pointer to first uv_buf and the
 *         # pointer to first byte in that buffer.
 */

static PyObject *__pyx_f_6uvloop_4loop_19_StreamWriteContext_advance_uv_buf(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self, size_t __pyx_v_sent) {
  uv_buf_t *__pyx_v_buf;
  size_t __pyx_v_idx;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("advance_uv_buf", 1);

  /* "uvloop/handles/stream.pyx":84
 *             size_t idx
 * 
 *         for idx from 0 <= idx < self.uv_bufs_len:             # <<<<<<<<<<<<<<
 *             buf = &self.uv_bufs_start[idx]
 *             if buf.len > sent:
 */
  __pyx_t_1 = __pyx_v_self->uv_bufs_len;
  for (__pyx_v_idx = 0; __pyx_v_idx < __pyx_t_1; __pyx_v_idx++) {

    /* "uvloop/handles/stream.pyx":85
 * 
 *         for idx from 0 <= idx < self.uv_bufs_len:
 *             buf = &self.uv_bufs_start[idx]             # <<<<<<<<<<<<<<
 *             if buf.len > sent:
 *                 buf.len -= sent
 */
    __pyx_v_buf = (&(__pyx_v_self->uv_bufs_start[__pyx_v_idx]));

    /* "uvloop/handles/stream.pyx":86
 *         for idx from 0 <= idx < self.uv_bufs_len:
 *             buf = &self.uv_bufs_start[idx]
 *             if buf.len > sent:             # <<<<<<<<<<<<<<
 *                 buf.len -= sent
 *                 buf.base = buf.base + sent
 */
    __pyx_t_2 = (__pyx_v_buf->len > __pyx_v_sent);
    if (__pyx_t_2) {

      /* "uvloop/handles/stream.pyx":87
 *             buf = &self.uv_bufs_start[idx]
 *             if buf.len > sent:
 *                 buf.len -= sent             # <<<<<<<<<<<<<<
 *                 buf.base = buf.base + sent
 *                 self.uv_bufs_start = buf
 */
      __pyx_v_buf->len = (__pyx_v_buf->len - __pyx_v_sent);

      /* "uvloop/handles/stream.pyx":88
 *             if buf.len > sent:
 *                 buf.len -= sent
 *                 buf.base = buf.base + sent             # <<<<<<<<<<<<<<
 *                 self.uv_bufs_start = buf
 *                 self.uv_bufs_len -= idx
 */
      __pyx_v_buf->base = (__pyx_v_buf->base + __pyx_v_sent);

      /* "uvloop/handles/stream.pyx":89
 *                 buf.len -= sent
 *                 buf.base = buf.base + sent
 *                 self.uv_bufs_start = buf             # <<<<<<<<<<<<<<
 *                 self.uv_bufs_len -= idx
 *                 return
 */
      __pyx_v_self->uv_bufs_start = __pyx_v_buf;

      /* "uvloop/handles/stream.pyx":90
 *                 buf.base = buf.base + sent
 *                 self.uv_bufs_start = buf
 *                 self.uv_bufs_len -= idx             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
      __pyx_v_self->uv_bufs_len = (__pyx_v_self->uv_bufs_len - __pyx_v_idx);

      /* "uvloop/handles/stream.pyx":91
 *                 self.uv_bufs_start = buf
 *                 self.uv_bufs_len -= idx
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 sent -= self.uv_bufs_start[idx].len
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/stream.pyx":86
 *         for idx from 0 <= idx < self.uv_bufs_len:
 *             buf = &self.uv_bufs_start[idx]
 *             if buf.len > sent:             # <<<<<<<<<<<<<<
 *                 buf.len -= sent
 *                 buf.base = buf.base + sent
 */
    }

    /* "uvloop/handles/stream.pyx":93
 *                 return
 *             else:
 *                 sent -= self.uv_bufs_start[idx].len             # <<<<<<<<<<<<<<
 * 
 *             if UVLOOP_DEBUG:
 */
    /*else*/ {
      __pyx_v_sent = (__pyx_v_sent - (__pyx_v_self->uv_bufs_start[__pyx_v_idx]).len);
    }

    /* "uvloop/handles/stream.pyx":95
 *                 sent -= self.uv_bufs_start[idx].len
 * 
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if sent < 0:
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')
 */
    __pyx_t_2 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/stream.pyx":96
 * 
 *             if UVLOOP_DEBUG:
 *                 if sent < 0:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')
 * 
 */
      __pyx_t_2 = (__pyx_v_sent < 0);
      if (unlikely(__pyx_t_2)) {

        /* "uvloop/handles/stream.pyx":97
 *             if UVLOOP_DEBUG:
 *                 if sent < 0:
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')             # <<<<<<<<<<<<<<
 * 
 *         raise RuntimeError('fatal: Could not advance _StreamWriteContext')
 */
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__147, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 97, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(20, 97, __pyx_L1_error)

        /* "uvloop/handles/stream.pyx":96
 * 
 *             if UVLOOP_DEBUG:
 *                 if sent < 0:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')
 * 
 */
      }

      /* "uvloop/handles/stream.pyx":95
 *                 sent -= self.uv_bufs_start[idx].len
 * 
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if sent < 0:
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')
 */
    }
  }

  /* "uvloop/handles/stream.pyx":99
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')
 * 
 *         raise RuntimeError('fatal: Could not advance _StreamWriteContext')             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__148, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __PYX_ERR(20, 99, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":71
 *         Py_DECREF(self)
 * 
 *     cdef advance_uv_buf(self, size_t sent):             # <<<<<<<<<<<<<<
 *         # Advance the pointer to first uv_buf and the
 *         # pointer to first byte in that buffer.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.advance_uv_buf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":102
 * 
 *     @staticmethod
 *     cdef _StreamWriteContext new(UVStream stream, list buffers):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _StreamWriteContext ctx
 */

static struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_f_6uvloop_4loop_19_StreamWriteContext_new(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_stream, PyObject *__pyx_v_buffers) {
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_ctx = 0;
  int __pyx_v_uv_bufs_idx;
  size_t __pyx_v_py_bufs_len;
  Py_buffer *__pyx_v_p_pybufs;
  uv_buf_t *__pyx_v_p_uvbufs;
  PyObject *__pyx_v_buf = NULL;
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  Py_buffer *__pyx_t_9;
  uv_buf_t *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/stream.pyx":105
 *         cdef:
 *             _StreamWriteContext ctx
 *             int uv_bufs_idx = 0             # <<<<<<<<<<<<<<
 *             size_t py_bufs_len = 0
 *             int i
 */
  __pyx_v_uv_bufs_idx = 0;

  /* "uvloop/handles/stream.pyx":106
 *             _StreamWriteContext ctx
 *             int uv_bufs_idx = 0
 *             size_t py_bufs_len = 0             # <<<<<<<<<<<<<<
 *             int i
 * 
 */
  __pyx_v_py_bufs_len = 0;

  /* "uvloop/handles/stream.pyx":112
 *             uv.uv_buf_t* p_uvbufs
 * 
 *         ctx = _StreamWriteContext.__new__(_StreamWriteContext)             # <<<<<<<<<<<<<<
 *         ctx.stream = None
 *         ctx.closed = 1
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop__StreamWriteContext(((PyTypeObject *)__pyx_ptype_6uvloop_4loop__StreamWriteContext), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 112, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_ctx = ((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":113
 * 
 *         ctx = _StreamWriteContext.__new__(_StreamWriteContext)
 *         ctx.stream = None             # <<<<<<<<<<<<<<
 *         ctx.closed = 1
 *         ctx.py_bufs_len = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_ctx->stream);
  __Pyx_DECREF((PyObject *)__pyx_v_ctx->stream);
  __pyx_v_ctx->stream = ((struct __pyx_obj_6uvloop_4loop_UVStream *)Py_None);

  /* "uvloop/handles/stream.pyx":114
 *         ctx = _StreamWriteContext.__new__(_StreamWriteContext)
 *         ctx.stream = None
 *         ctx.closed = 1             # <<<<<<<<<<<<<<
 *         ctx.py_bufs_len = 0
 *         ctx.py_bufs_sml_inuse = 0
 */
  __pyx_v_ctx->closed = 1;

  /* "uvloop/handles/stream.pyx":115
 *         ctx.stream = None
 *         ctx.closed = 1
 *         ctx.py_bufs_len = 0             # <<<<<<<<<<<<<<
 *         ctx.py_bufs_sml_inuse = 0
 *         ctx.uv_bufs = NULL
 */
  __pyx_v_ctx->py_bufs_len = 0;

  /* "uvloop/handles/stream.pyx":116
 *         ctx.closed = 1
 *         ctx.py_bufs_len = 0
 *         ctx.py_bufs_sml_inuse = 0             # <<<<<<<<<<<<<<
 *         ctx.uv_bufs = NULL
 *         ctx.py_bufs = NULL
 */
  __pyx_v_ctx->py_bufs_sml_inuse = 0;

  /* "uvloop/handles/stream.pyx":117
 *         ctx.py_bufs_len = 0
 *         ctx.py_bufs_sml_inuse = 0
 *         ctx.uv_bufs = NULL             # <<<<<<<<<<<<<<
 *         ctx.py_bufs = NULL
 *         ctx.buffers = buffers
 */
  __pyx_v_ctx->uv_bufs = NULL;

  /* "uvloop/handles/stream.pyx":118
 *         ctx.py_bufs_sml_inuse = 0
 *         ctx.uv_bufs = NULL
 *         ctx.py_bufs = NULL             # <<<<<<<<<<<<<<
 *         ctx.buffers = buffers
 *         ctx.stream = stream
 */
  __pyx_v_ctx->py_bufs = NULL;

  /* "uvloop/handles/stream.pyx":119
 *         ctx.uv_bufs = NULL
 *         ctx.py_bufs = NULL
 *         ctx.buffers = buffers             # <<<<<<<<<<<<<<
 *         ctx.stream = stream
 * 
 */
  __Pyx_INCREF(__pyx_v_buffers);
  __Pyx_GIVEREF(__pyx_v_buffers);
  __Pyx_GOTREF(__pyx_v_ctx->buffers);
  __Pyx_DECREF(__pyx_v_ctx->buffers);
  __pyx_v_ctx->buffers = __pyx_v_buffers;

  /* "uvloop/handles/stream.pyx":120
 *         ctx.py_bufs = NULL
 *         ctx.buffers = buffers
 *         ctx.stream = stream             # <<<<<<<<<<<<<<
 * 
 *         if len(buffers) <= __PREALLOCED_BUFS:
 */
  __Pyx_INCREF((PyObject *)__pyx_v_stream);
  __Pyx_GIVEREF((PyObject *)__pyx_v_stream);
  __Pyx_GOTREF((PyObject *)__pyx_v_ctx->stream);
  __Pyx_DECREF((PyObject *)__pyx_v_ctx->stream);
  __pyx_v_ctx->stream = __pyx_v_stream;

  /* "uvloop/handles/stream.pyx":122
 *         ctx.stream = stream
 * 
 *         if len(buffers) <= __PREALLOCED_BUFS:             # <<<<<<<<<<<<<<
 *             # We've got a small number of buffers to write, don't
 *             # need to use malloc.
 */
  if (unlikely(__pyx_v_buffers == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(20, 122, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_buffers); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(20, 122, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 <= __PREALLOCED_BUFS);
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":125
 *             # We've got a small number of buffers to write, don't
 *             # need to use malloc.
 *             ctx.py_bufs_sml_inuse = 1             # <<<<<<<<<<<<<<
 *             p_pybufs = <Py_buffer*>&ctx.py_bufs_sml
 *             p_uvbufs = <uv.uv_buf_t*>&ctx.uv_bufs_sml
 */
    __pyx_v_ctx->py_bufs_sml_inuse = 1;

    /* "uvloop/handles/stream.pyx":126
 *             # need to use malloc.
 *             ctx.py_bufs_sml_inuse = 1
 *             p_pybufs = <Py_buffer*>&ctx.py_bufs_sml             # <<<<<<<<<<<<<<
 *             p_uvbufs = <uv.uv_buf_t*>&ctx.uv_bufs_sml
 * 
 */
    __pyx_v_p_pybufs = ((Py_buffer *)(&__pyx_v_ctx->py_bufs_sml));

    /* "uvloop/handles/stream.pyx":127
 *             ctx.py_bufs_sml_inuse = 1
 *             p_pybufs = <Py_buffer*>&ctx.py_bufs_sml
 *             p_uvbufs = <uv.uv_buf_t*>&ctx.uv_bufs_sml             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    __pyx_v_p_uvbufs = ((uv_buf_t *)(&__pyx_v_ctx->uv_bufs_sml));

    /* "uvloop/handles/stream.pyx":122
 *         ctx.stream = stream
 * 
 *         if len(buffers) <= __PREALLOCED_BUFS:             # <<<<<<<<<<<<<<
 *             # We've got a small number of buffers to write, don't
 *             # need to use malloc.
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/stream.pyx":130
 * 
 *         else:
 *             for buf in buffers:             # <<<<<<<<<<<<<<
 *                 if UVLOOP_DEBUG:
 *                     if not isinstance(buf, (bytes, bytearray, memoryview)):
 */
  /*else*/ {
    if (unlikely(__pyx_v_buffers == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(20, 130, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_buffers; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 130, __pyx_L1_error)
        #endif
        if (__pyx_t_2 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(20, 130, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 130, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_buf, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/handles/stream.pyx":131
 *         else:
 *             for buf in buffers:
 *                 if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                     if not isinstance(buf, (bytes, bytearray, memoryview)):
 *                         raise RuntimeError(
 */
      __pyx_t_3 = (UVLOOP_DEBUG != 0);
      if (__pyx_t_3) {

        /* "uvloop/handles/stream.pyx":132
 *             for buf in buffers:
 *                 if UVLOOP_DEBUG:
 *                     if not isinstance(buf, (bytes, bytearray, memoryview)):             # <<<<<<<<<<<<<<
 *                         raise RuntimeError(
 *                             'invalid data in writebuf: an instance of '
 */
        __pyx_t_5 = PyBytes_Check(__pyx_v_buf); 
        if (!__pyx_t_5) {
        } else {
          __pyx_t_3 = __pyx_t_5;
          goto __pyx_L8_bool_binop_done;
        }
        __pyx_t_5 = PyByteArray_Check(__pyx_v_buf); 
        if (!__pyx_t_5) {
        } else {
          __pyx_t_3 = __pyx_t_5;
          goto __pyx_L8_bool_binop_done;
        }
        __pyx_t_5 = PyMemoryView_Check(__pyx_v_buf); 
        __pyx_t_3 = __pyx_t_5;
        __pyx_L8_bool_binop_done:;
        __pyx_t_5 = (!__pyx_t_3);
        if (unlikely(__pyx_t_5)) {

          /* "uvloop/handles/stream.pyx":136
 *                             'invalid data in writebuf: an instance of '
 *                             'bytes, bytearray or memoryview was expected, '
 *                             'got {}'.format(type(buf)))             # <<<<<<<<<<<<<<
 * 
 *                 if not PyBytes_CheckExact(buf):
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_data_in_writebuf_an_inst, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 136, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_7, ((PyObject *)Py_TYPE(__pyx_v_buf))};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 136, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }

          /* "uvloop/handles/stream.pyx":133
 *                 if UVLOOP_DEBUG:
 *                     if not isinstance(buf, (bytes, bytearray, memoryview)):
 *                         raise RuntimeError(             # <<<<<<<<<<<<<<
 *                             'invalid data in writebuf: an instance of '
 *                             'bytes, bytearray or memoryview was expected, '
 */
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 133, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_Raise(__pyx_t_6, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __PYX_ERR(20, 133, __pyx_L1_error)

          /* "uvloop/handles/stream.pyx":132
 *             for buf in buffers:
 *                 if UVLOOP_DEBUG:
 *                     if not isinstance(buf, (bytes, bytearray, memoryview)):             # <<<<<<<<<<<<<<
 *                         raise RuntimeError(
 *                             'invalid data in writebuf: an instance of '
 */
        }

        /* "uvloop/handles/stream.pyx":131
 *         else:
 *             for buf in buffers:
 *                 if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                     if not isinstance(buf, (bytes, bytearray, memoryview)):
 *                         raise RuntimeError(
 */
      }

      /* "uvloop/handles/stream.pyx":138
 *                             'got {}'.format(type(buf)))
 * 
 *                 if not PyBytes_CheckExact(buf):             # <<<<<<<<<<<<<<
 *                     py_bufs_len += 1
 * 
 */
      __pyx_t_5 = (!PyBytes_CheckExact(__pyx_v_buf));
      if (__pyx_t_5) {

        /* "uvloop/handles/stream.pyx":139
 * 
 *                 if not PyBytes_CheckExact(buf):
 *                     py_bufs_len += 1             # <<<<<<<<<<<<<<
 * 
 *             if py_bufs_len > 0:
 */
        __pyx_v_py_bufs_len = (__pyx_v_py_bufs_len + 1);

        /* "uvloop/handles/stream.pyx":138
 *                             'got {}'.format(type(buf)))
 * 
 *                 if not PyBytes_CheckExact(buf):             # <<<<<<<<<<<<<<
 *                     py_bufs_len += 1
 * 
 */
      }

      /* "uvloop/handles/stream.pyx":130
 * 
 *         else:
 *             for buf in buffers:             # <<<<<<<<<<<<<<
 *                 if UVLOOP_DEBUG:
 *                     if not isinstance(buf, (bytes, bytearray, memoryview)):
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":141
 *                     py_bufs_len += 1
 * 
 *             if py_bufs_len > 0:             # <<<<<<<<<<<<<<
 *                 ctx.py_bufs = <Py_buffer*>PyMem_RawMalloc(
 *                     py_bufs_len * sizeof(Py_buffer))
 */
    __pyx_t_5 = (__pyx_v_py_bufs_len > 0);
    if (__pyx_t_5) {

      /* "uvloop/handles/stream.pyx":142
 * 
 *             if py_bufs_len > 0:
 *                 ctx.py_bufs = <Py_buffer*>PyMem_RawMalloc(             # <<<<<<<<<<<<<<
 *                     py_bufs_len * sizeof(Py_buffer))
 *                 if ctx.py_bufs is NULL:
 */
      __pyx_v_ctx->py_bufs = ((Py_buffer *)PyMem_RawMalloc((__pyx_v_py_bufs_len * (sizeof(Py_buffer)))));

      /* "uvloop/handles/stream.pyx":144
 *                 ctx.py_bufs = <Py_buffer*>PyMem_RawMalloc(
 *                     py_bufs_len * sizeof(Py_buffer))
 *                 if ctx.py_bufs is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 * 
 */
      __pyx_t_5 = (__pyx_v_ctx->py_bufs == NULL);
      if (unlikely(__pyx_t_5)) {

        /* "uvloop/handles/stream.pyx":145
 *                     py_bufs_len * sizeof(Py_buffer))
 *                 if ctx.py_bufs is NULL:
 *                     raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             ctx.uv_bufs = <uv.uv_buf_t*>PyMem_RawMalloc(
 */
        PyErr_NoMemory(); __PYX_ERR(20, 145, __pyx_L1_error)

        /* "uvloop/handles/stream.pyx":144
 *                 ctx.py_bufs = <Py_buffer*>PyMem_RawMalloc(
 *                     py_bufs_len * sizeof(Py_buffer))
 *                 if ctx.py_bufs is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 * 
 */
      }

      /* "uvloop/handles/stream.pyx":141
 *                     py_bufs_len += 1
 * 
 *             if py_bufs_len > 0:             # <<<<<<<<<<<<<<
 *                 ctx.py_bufs = <Py_buffer*>PyMem_RawMalloc(
 *                     py_bufs_len * sizeof(Py_buffer))
 */
    }

    /* "uvloop/handles/stream.pyx":148
 * 
 *             ctx.uv_bufs = <uv.uv_buf_t*>PyMem_RawMalloc(
 *                 len(buffers) * sizeof(uv.uv_buf_t))             # <<<<<<<<<<<<<<
 *             if ctx.uv_bufs is NULL:
 *                 raise MemoryError()
 */
    if (unlikely(__pyx_v_buffers == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(20, 148, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_buffers); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(20, 148, __pyx_L1_error)

    /* "uvloop/handles/stream.pyx":147
 *                     raise MemoryError()
 * 
 *             ctx.uv_bufs = <uv.uv_buf_t*>PyMem_RawMalloc(             # <<<<<<<<<<<<<<
 *                 len(buffers) * sizeof(uv.uv_buf_t))
 *             if ctx.uv_bufs is NULL:
 */
    __pyx_v_ctx->uv_bufs = ((uv_buf_t *)PyMem_RawMalloc((__pyx_t_2 * (sizeof(uv_buf_t)))));

    /* "uvloop/handles/stream.pyx":149
 *             ctx.uv_bufs = <uv.uv_buf_t*>PyMem_RawMalloc(
 *                 len(buffers) * sizeof(uv.uv_buf_t))
 *             if ctx.uv_bufs is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_5 = (__pyx_v_ctx->uv_bufs == NULL);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/handles/stream.pyx":150
 *                 len(buffers) * sizeof(uv.uv_buf_t))
 *             if ctx.uv_bufs is NULL:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             p_pybufs = ctx.py_bufs
 */
      PyErr_NoMemory(); __PYX_ERR(20, 150, __pyx_L1_error)

      /* "uvloop/handles/stream.pyx":149
 *             ctx.uv_bufs = <uv.uv_buf_t*>PyMem_RawMalloc(
 *                 len(buffers) * sizeof(uv.uv_buf_t))
 *             if ctx.uv_bufs is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "uvloop/handles/stream.pyx":152
 *                 raise MemoryError()
 * 
 *             p_pybufs = ctx.py_bufs             # <<<<<<<<<<<<<<
 *             p_uvbufs = ctx.uv_bufs
 * 
 */
    __pyx_t_9 = __pyx_v_ctx->py_bufs;
    __pyx_v_p_pybufs = __pyx_t_9;

    /* "uvloop/handles/stream.pyx":153
 * 
 *             p_pybufs = ctx.py_bufs
 *             p_uvbufs = ctx.uv_bufs             # <<<<<<<<<<<<<<
 * 
 *         py_bufs_len = 0
 */
    __pyx_t_10 = __pyx_v_ctx->uv_bufs;
    __pyx_v_p_uvbufs = __pyx_t_10;
  }
  __pyx_L3:;

  /* "uvloop/handles/stream.pyx":155
 *             p_uvbufs = ctx.uv_bufs
 * 
 *         py_bufs_len = 0             # <<<<<<<<<<<<<<
 *         for buf in buffers:
 *             if PyBytes_CheckExact(buf):
 */
  __pyx_v_py_bufs_len = 0;

  /* "uvloop/handles/stream.pyx":156
 * 
 *         py_bufs_len = 0
 *         for buf in buffers:             # <<<<<<<<<<<<<<
 *             if PyBytes_CheckExact(buf):
 *                 # We can only use this hack for bytes since it's
 */
  if (unlikely(__pyx_v_buffers == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(20, 156, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_buffers; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 156, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_6); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(20, 156, __pyx_L1_error)
    #else
    __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_buf, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "uvloop/handles/stream.pyx":157
 *         py_bufs_len = 0
 *         for buf in buffers:
 *             if PyBytes_CheckExact(buf):             # <<<<<<<<<<<<<<
 *                 # We can only use this hack for bytes since it's
 *                 # immutable.  For everything else it is only safe to
 */
    __pyx_t_5 = PyBytes_CheckExact(__pyx_v_buf);
    if (__pyx_t_5) {

      /* "uvloop/handles/stream.pyx":161
 *                 # immutable.  For everything else it is only safe to
 *                 # use buffer protocol.
 *                 p_uvbufs[uv_bufs_idx].base = PyBytes_AS_STRING(buf)             # <<<<<<<<<<<<<<
 *                 p_uvbufs[uv_bufs_idx].len = Py_SIZE(buf)
 * 
 */
      (__pyx_v_p_uvbufs[__pyx_v_uv_bufs_idx]).base = PyBytes_AS_STRING(__pyx_v_buf);

      /* "uvloop/handles/stream.pyx":162
 *                 # use buffer protocol.
 *                 p_uvbufs[uv_bufs_idx].base = PyBytes_AS_STRING(buf)
 *                 p_uvbufs[uv_bufs_idx].len = Py_SIZE(buf)             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      (__pyx_v_p_uvbufs[__pyx_v_uv_bufs_idx]).len = Py_SIZE(__pyx_v_buf);

      /* "uvloop/handles/stream.pyx":157
 *         py_bufs_len = 0
 *         for buf in buffers:
 *             if PyBytes_CheckExact(buf):             # <<<<<<<<<<<<<<
 *                 # We can only use this hack for bytes since it's
 *                 # immutable.  For everything else it is only safe to
 */
      goto __pyx_L18;
    }

    /* "uvloop/handles/stream.pyx":165
 * 
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     PyObject_GetBuffer(
 *                         buf, &p_pybufs[py_bufs_len], PyBUF_SIMPLE)
 */
    /*else*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        /*try:*/ {

          /* "uvloop/handles/stream.pyx":166
 *             else:
 *                 try:
 *                     PyObject_GetBuffer(             # <<<<<<<<<<<<<<
 *                         buf, &p_pybufs[py_bufs_len], PyBUF_SIMPLE)
 *                 except Exception:
 */
          __pyx_t_14 = PyObject_GetBuffer(__pyx_v_buf, (&(__pyx_v_p_pybufs[__pyx_v_py_bufs_len])), PyBUF_SIMPLE); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(20, 166, __pyx_L19_error)

          /* "uvloop/handles/stream.pyx":165
 * 
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     PyObject_GetBuffer(
 *                         buf, &p_pybufs[py_bufs_len], PyBUF_SIMPLE)
 */
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L26_try_end;
        __pyx_L19_error:;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "uvloop/handles/stream.pyx":168
 *                     PyObject_GetBuffer(
 *                         buf, &p_pybufs[py_bufs_len], PyBUF_SIMPLE)
 *                 except Exception:             # <<<<<<<<<<<<<<
 *                     # This shouldn't ever happen, as `UVStream._buffer_write`
 *                     # casts non-bytes objects to `memoryviews`.
 */
        __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_14) {
          __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(20, 168, __pyx_L21_except_error)
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_7);

          /* "uvloop/handles/stream.pyx":171
 *                     # This shouldn't ever happen, as `UVStream._buffer_write`
 *                     # casts non-bytes objects to `memoryviews`.
 *                     ctx.py_bufs_len = py_bufs_len             # <<<<<<<<<<<<<<
 *                     ctx.free_bufs()
 *                     raise
 */
          __pyx_v_ctx->py_bufs_len = __pyx_v_py_bufs_len;

          /* "uvloop/handles/stream.pyx":172
 *                     # casts non-bytes objects to `memoryviews`.
 *                     ctx.py_bufs_len = py_bufs_len
 *                     ctx.free_bufs()             # <<<<<<<<<<<<<<
 *                     raise
 * 
 */
          __pyx_t_15 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_ctx->__pyx_vtab)->free_bufs(__pyx_v_ctx); if (unlikely(!__pyx_t_15)) __PYX_ERR(20, 172, __pyx_L21_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "uvloop/handles/stream.pyx":173
 *                     ctx.py_bufs_len = py_bufs_len
 *                     ctx.free_bufs()
 *                     raise             # <<<<<<<<<<<<<<
 * 
 *                 p_uvbufs[uv_bufs_idx].base = <char*>p_pybufs[py_bufs_len].buf
 */
          __Pyx_GIVEREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_4);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_4, __pyx_t_7);
          __pyx_t_6 = 0; __pyx_t_4 = 0; __pyx_t_7 = 0; 
          __PYX_ERR(20, 173, __pyx_L21_except_error)
        }
        goto __pyx_L21_except_error;

        /* "uvloop/handles/stream.pyx":165
 * 
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     PyObject_GetBuffer(
 *                         buf, &p_pybufs[py_bufs_len], PyBUF_SIMPLE)
 */
        __pyx_L21_except_error:;
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        goto __pyx_L1_error;
        __pyx_L26_try_end:;
      }

      /* "uvloop/handles/stream.pyx":175
 *                     raise
 * 
 *                 p_uvbufs[uv_bufs_idx].base = <char*>p_pybufs[py_bufs_len].buf             # <<<<<<<<<<<<<<
 *                 p_uvbufs[uv_bufs_idx].len = p_pybufs[py_bufs_len].len
 * 
 */
      (__pyx_v_p_uvbufs[__pyx_v_uv_bufs_idx]).base = ((char *)(__pyx_v_p_pybufs[__pyx_v_py_bufs_len]).buf);

      /* "uvloop/handles/stream.pyx":176
 * 
 *                 p_uvbufs[uv_bufs_idx].base = <char*>p_pybufs[py_bufs_len].buf
 *                 p_uvbufs[uv_bufs_idx].len = p_pybufs[py_bufs_len].len             # <<<<<<<<<<<<<<
 * 
 *                 py_bufs_len += 1
 */
      __pyx_t_16 = (__pyx_v_p_pybufs[__pyx_v_py_bufs_len]).len;
      (__pyx_v_p_uvbufs[__pyx_v_uv_bufs_idx]).len = __pyx_t_16;

      /* "uvloop/handles/stream.pyx":178
 *                 p_uvbufs[uv_bufs_idx].len = p_pybufs[py_bufs_len].len
 * 
 *                 py_bufs_len += 1             # <<<<<<<<<<<<<<
 * 
 *             uv_bufs_idx += 1
 */
      __pyx_v_py_bufs_len = (__pyx_v_py_bufs_len + 1);
    }
    __pyx_L18:;

    /* "uvloop/handles/stream.pyx":180
 *                 py_bufs_len += 1
 * 
 *             uv_bufs_idx += 1             # <<<<<<<<<<<<<<
 * 
 *         ctx.uv_bufs_start = p_uvbufs
 */
    __pyx_v_uv_bufs_idx = (__pyx_v_uv_bufs_idx + 1);

    /* "uvloop/handles/stream.pyx":156
 * 
 *         py_bufs_len = 0
 *         for buf in buffers:             # <<<<<<<<<<<<<<
 *             if PyBytes_CheckExact(buf):
 *                 # We can only use this hack for bytes since it's
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":182
 *             uv_bufs_idx += 1
 * 
 *         ctx.uv_bufs_start = p_uvbufs             # <<<<<<<<<<<<<<
 *         ctx.uv_bufs_len = uv_bufs_idx
 * 
 */
  __pyx_v_ctx->uv_bufs_start = __pyx_v_p_uvbufs;

  /* "uvloop/handles/stream.pyx":183
 * 
 *         ctx.uv_bufs_start = p_uvbufs
 *         ctx.uv_bufs_len = uv_bufs_idx             # <<<<<<<<<<<<<<
 * 
 *         ctx.py_bufs_len = py_bufs_len
 */
  __pyx_v_ctx->uv_bufs_len = __pyx_v_uv_bufs_idx;

  /* "uvloop/handles/stream.pyx":185
 *         ctx.uv_bufs_len = uv_bufs_idx
 * 
 *         ctx.py_bufs_len = py_bufs_len             # <<<<<<<<<<<<<<
 *         ctx.req.data = <void*> ctx
 * 
 */
  __pyx_v_ctx->py_bufs_len = __pyx_v_py_bufs_len;

  /* "uvloop/handles/stream.pyx":186
 * 
 *         ctx.py_bufs_len = py_bufs_len
 *         ctx.req.data = <void*> ctx             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG:
 */
  __pyx_v_ctx->req.data = ((void *)__pyx_v_ctx);

  /* "uvloop/handles/stream.pyx":188
 *         ctx.req.data = <void*> ctx
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_write_ctx_total += 1
 *             stream._loop._debug_stream_write_ctx_cnt += 1
 */
  __pyx_t_5 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_5) {

    /* "uvloop/handles/stream.pyx":189
 * 
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_ctx_total += 1             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_write_ctx_cnt += 1
 * 
 */
    __Pyx_INCREF((PyObject *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop);
    __pyx_t_17 = __pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop;
    __pyx_t_17->_debug_stream_write_ctx_total = (__pyx_t_17->_debug_stream_write_ctx_total + 1);
    __Pyx_DECREF((PyObject *)__pyx_t_17); __pyx_t_17 = 0;

    /* "uvloop/handles/stream.pyx":190
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_ctx_total += 1
 *             stream._loop._debug_stream_write_ctx_cnt += 1             # <<<<<<<<<<<<<<
 * 
 *         # Do incref after everything else is done.
 */
    __Pyx_INCREF((PyObject *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop);
    __pyx_t_17 = __pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop;
    __pyx_t_17->_debug_stream_write_ctx_cnt = (__pyx_t_17->_debug_stream_write_ctx_cnt + 1);
    __Pyx_DECREF((PyObject *)__pyx_t_17); __pyx_t_17 = 0;

    /* "uvloop/handles/stream.pyx":188
 *         ctx.req.data = <void*> ctx
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_write_ctx_total += 1
 *             stream._loop._debug_stream_write_ctx_cnt += 1
 */
  }

  /* "uvloop/handles/stream.pyx":195
 *         # Under no circumstances we want `ctx` to be GCed while
 *         # libuv is still working with `ctx.uv_bufs`.
 *         Py_INCREF(ctx)             # <<<<<<<<<<<<<<
 *         ctx.closed = 0
 *         return ctx
 */
  Py_INCREF(((PyObject *)__pyx_v_ctx));

  /* "uvloop/handles/stream.pyx":196
 *         # libuv is still working with `ctx.uv_bufs`.
 *         Py_INCREF(ctx)
 *         ctx.closed = 0             # <<<<<<<<<<<<<<
 *         return ctx
 * 
 */
  __pyx_v_ctx->closed = 0;

  /* "uvloop/handles/stream.pyx":197
 *         Py_INCREF(ctx)
 *         ctx.closed = 0
 *         return ctx             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ctx);
  __pyx_r = __pyx_v_ctx;
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":102
 * 
 *     @staticmethod
 *     cdef _StreamWriteContext new(UVStream stream, list buffers):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _StreamWriteContext ctx
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF((PyObject *)__pyx_t_17);
  __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":199
 *         return ctx
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if not self.closed:
 *             # Because we do an INCREF in _StreamWriteContext.new,
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_19_StreamWriteContext_1__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_19_StreamWriteContext_1__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_19_StreamWriteContext___dealloc__(((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_19_StreamWriteContext___dealloc__(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/handles/stream.pyx":200
 * 
 *     def __dealloc__(self):
 *         if not self.closed:             # <<<<<<<<<<<<<<
 *             # Because we do an INCREF in _StreamWriteContext.new,
 *             # __dealloc__ shouldn't ever happen with `self.closed == 1`
 */
  __pyx_t_1 = (!__pyx_v_self->closed);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/stream.pyx":203
 *             # Because we do an INCREF in _StreamWriteContext.new,
 *             # __dealloc__ shouldn't ever happen with `self.closed == 1`
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'open _StreamWriteContext is being deallocated')
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__149, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(20, 203, __pyx_L1_error)

    /* "uvloop/handles/stream.pyx":200
 * 
 *     def __dealloc__(self):
 *         if not self.closed:             # <<<<<<<<<<<<<<
 *             # Because we do an INCREF in _StreamWriteContext.new,
 *             # __dealloc__ shouldn't ever happen with `self.closed == 1`
 */
  }

  /* "uvloop/handles/stream.pyx":206
 *                 'open _StreamWriteContext is being deallocated')
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self.stream is not None:
 *                 self.stream._loop._debug_stream_write_ctx_cnt -= 1
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":207
 * 
 *         if UVLOOP_DEBUG:
 *             if self.stream is not None:             # <<<<<<<<<<<<<<
 *                 self.stream._loop._debug_stream_write_ctx_cnt -= 1
 *                 self.stream = None
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->stream) != Py_None);
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":208
 *         if UVLOOP_DEBUG:
 *             if self.stream is not None:
 *                 self.stream._loop._debug_stream_write_ctx_cnt -= 1             # <<<<<<<<<<<<<<
 *                 self.stream = None
 * 
 */
      __Pyx_INCREF((PyObject *)__pyx_v_self->stream->__pyx_base.__pyx_base.__pyx_base._loop);
      __pyx_t_3 = __pyx_v_self->stream->__pyx_base.__pyx_base.__pyx_base._loop;
      __pyx_t_3->_debug_stream_write_ctx_cnt = (__pyx_t_3->_debug_stream_write_ctx_cnt - 1);
      __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/stream.pyx":209
 *             if self.stream is not None:
 *                 self.stream._loop._debug_stream_write_ctx_cnt -= 1
 *                 self.stream = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->stream);
      __Pyx_DECREF((PyObject *)__pyx_v_self->stream);
      __pyx_v_self->stream = ((struct __pyx_obj_6uvloop_4loop_UVStream *)Py_None);

      /* "uvloop/handles/stream.pyx":207
 * 
 *         if UVLOOP_DEBUG:
 *             if self.stream is not None:             # <<<<<<<<<<<<<<
 *                 self.stream._loop._debug_stream_write_ctx_cnt -= 1
 *                 self.stream = None
 */
    }

    /* "uvloop/handles/stream.pyx":206
 *                 'open _StreamWriteContext is being deallocated')
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self.stream is not None:
 *                 self.stream._loop._debug_stream_write_ctx_cnt -= 1
 */
  }

  /* "uvloop/handles/stream.pyx":199
 *         return ctx
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if not self.closed:
 *             # Because we do an INCREF in _StreamWriteContext.new,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF((PyObject *)__pyx_t_3);
  __Pyx_WriteUnraisable("uvloop.loop._StreamWriteContext.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_19_StreamWriteContext_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_19_StreamWriteContext_2__reduce_cython__, "_StreamWriteContext.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_19_StreamWriteContext_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_StreamWriteContext_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_StreamWriteContext_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_19_StreamWriteContext_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_19_StreamWriteContext_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_19_StreamWriteContext_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_py_bufs_self_py_bufs_sml_se, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_19_StreamWriteContext_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_19_StreamWriteContext_4__setstate_cython__, "_StreamWriteContext.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_19_StreamWriteContext_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_StreamWriteContext_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_StreamWriteContext_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_19_StreamWriteContext_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_19_StreamWriteContext_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_19_StreamWriteContext_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_py_bufs_self_py_bufs_sml_se, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._StreamWriteContext.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":215
 * cdef class UVStream(UVBaseTransport):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.__shutting_down = 0
 *         self.__reading = 0
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8UVStream_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_8UVStream_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream___cinit__(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8UVStream___cinit__(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/stream.pyx":216
 * 
 *     def __cinit__(self):
 *         self.__shutting_down = 0             # <<<<<<<<<<<<<<
 *         self.__reading = 0
 *         self.__read_error_close = 0
 */
  __pyx_v_self->_UVStream__shutting_down = 0;

  /* "uvloop/handles/stream.pyx":217
 *     def __cinit__(self):
 *         self.__shutting_down = 0
 *         self.__reading = 0             # <<<<<<<<<<<<<<
 *         self.__read_error_close = 0
 *         self.__buffered = 0
 */
  __pyx_v_self->_UVStream__reading = 0;

  /* "uvloop/handles/stream.pyx":218
 *         self.__shutting_down = 0
 *         self.__reading = 0
 *         self.__read_error_close = 0             # <<<<<<<<<<<<<<
 *         self.__buffered = 0
 *         self._eof = 0
 */
  __pyx_v_self->_UVStream__read_error_close = 0;

  /* "uvloop/handles/stream.pyx":219
 *         self.__reading = 0
 *         self.__read_error_close = 0
 *         self.__buffered = 0             # <<<<<<<<<<<<<<
 *         self._eof = 0
 *         self._buffer = []
 */
  __pyx_v_self->_UVStream__buffered = 0;

  /* "uvloop/handles/stream.pyx":220
 *         self.__read_error_close = 0
 *         self.__buffered = 0
 *         self._eof = 0             # <<<<<<<<<<<<<<
 *         self._buffer = []
 *         self._buffer_size = 0
 */
  __pyx_v_self->_eof = 0;

  /* "uvloop/handles/stream.pyx":221
 *         self.__buffered = 0
 *         self._eof = 0
 *         self._buffer = []             # <<<<<<<<<<<<<<
 *         self._buffer_size = 0
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_buffer);
  __Pyx_DECREF(__pyx_v_self->_buffer);
  __pyx_v_self->_buffer = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":222
 *         self._eof = 0
 *         self._buffer = []
 *         self._buffer_size = 0             # <<<<<<<<<<<<<<
 * 
 *         self._protocol_get_buffer = None
 */
  __pyx_v_self->_buffer_size = 0;

  /* "uvloop/handles/stream.pyx":224
 *         self._buffer_size = 0
 * 
 *         self._protocol_get_buffer = None             # <<<<<<<<<<<<<<
 *         self._protocol_buffer_updated = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol_get_buffer);
  __Pyx_DECREF(__pyx_v_self->_protocol_get_buffer);
  __pyx_v_self->_protocol_get_buffer = Py_None;

  /* "uvloop/handles/stream.pyx":225
 * 
 *         self._protocol_get_buffer = None
 *         self._protocol_buffer_updated = None             # <<<<<<<<<<<<<<
 * 
 *         self._read_pybuf_acquired = False
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol_buffer_updated);
  __Pyx_DECREF(__pyx_v_self->_protocol_buffer_updated);
  __pyx_v_self->_protocol_buffer_updated = Py_None;

  /* "uvloop/handles/stream.pyx":227
 *         self._protocol_buffer_updated = None
 * 
 *         self._read_pybuf_acquired = False             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_protocol(self, object protocol):
 */
  __pyx_v_self->_read_pybuf_acquired = 0;

  /* "uvloop/handles/stream.pyx":215
 * cdef class UVStream(UVBaseTransport):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.__shutting_down = 0
 *         self.__reading = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":229
 *         self._read_pybuf_acquired = False
 * 
 *     cdef _set_protocol(self, object protocol):             # <<<<<<<<<<<<<<
 *         if protocol is None:
 *             raise TypeError('protocol is required')
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVStream__set_protocol(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_protocol", 1);

  /* "uvloop/handles/stream.pyx":230
 * 
 *     cdef _set_protocol(self, object protocol):
 *         if protocol is None:             # <<<<<<<<<<<<<<
 *             raise TypeError('protocol is required')
 * 
 */
  __pyx_t_1 = (__pyx_v_protocol == Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/stream.pyx":231
 *     cdef _set_protocol(self, object protocol):
 *         if protocol is None:
 *             raise TypeError('protocol is required')             # <<<<<<<<<<<<<<
 * 
 *         UVBaseTransport._set_protocol(self, protocol)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__150, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(20, 231, __pyx_L1_error)

    /* "uvloop/handles/stream.pyx":230
 * 
 *     cdef _set_protocol(self, object protocol):
 *         if protocol is None:             # <<<<<<<<<<<<<<
 *             raise TypeError('protocol is required')
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":233
 *             raise TypeError('protocol is required')
 * 
 *         UVBaseTransport._set_protocol(self, protocol)             # <<<<<<<<<<<<<<
 * 
 *         if (hasattr(protocol, 'get_buffer') and
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_15UVBaseTransport__set_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_protocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/stream.pyx":235
 *         UVBaseTransport._set_protocol(self, protocol)
 * 
 *         if (hasattr(protocol, 'get_buffer') and             # <<<<<<<<<<<<<<
 *                 not isinstance(protocol, aio_Protocol)):
 *             try:
 */
  __pyx_t_3 = __Pyx_HasAttr(__pyx_v_protocol, __pyx_n_u_get_buffer); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(20, 235, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }

  /* "uvloop/handles/stream.pyx":236
 * 
 *         if (hasattr(protocol, 'get_buffer') and
 *                 not isinstance(protocol, aio_Protocol)):             # <<<<<<<<<<<<<<
 *             try:
 *                 self._protocol_get_buffer = protocol.get_buffer
 */
  __pyx_t_2 = __pyx_v_6uvloop_4loop_aio_Protocol;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_protocol, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(20, 236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (!__pyx_t_3);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;

  /* "uvloop/handles/stream.pyx":235
 *         UVBaseTransport._set_protocol(self, protocol)
 * 
 *         if (hasattr(protocol, 'get_buffer') and             # <<<<<<<<<<<<<<
 *                 not isinstance(protocol, aio_Protocol)):
 *             try:
 */
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":237
 *         if (hasattr(protocol, 'get_buffer') and
 *                 not isinstance(protocol, aio_Protocol)):
 *             try:             # <<<<<<<<<<<<<<
 *                 self._protocol_get_buffer = protocol.get_buffer
 *                 self._protocol_buffer_updated = protocol.buffer_updated
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "uvloop/handles/stream.pyx":238
 *                 not isinstance(protocol, aio_Protocol)):
 *             try:
 *                 self._protocol_get_buffer = protocol.get_buffer             # <<<<<<<<<<<<<<
 *                 self._protocol_buffer_updated = protocol.buffer_updated
 *                 self.__buffered = 1
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_protocol, __pyx_n_s_get_buffer); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 238, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_self->_protocol_get_buffer);
        __Pyx_DECREF(__pyx_v_self->_protocol_get_buffer);
        __pyx_v_self->_protocol_get_buffer = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "uvloop/handles/stream.pyx":239
 *             try:
 *                 self._protocol_get_buffer = protocol.get_buffer
 *                 self._protocol_buffer_updated = protocol.buffer_updated             # <<<<<<<<<<<<<<
 *                 self.__buffered = 1
 *             except AttributeError:
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_protocol, __pyx_n_s_buffer_updated); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 239, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_v_self->_protocol_buffer_updated);
        __Pyx_DECREF(__pyx_v_self->_protocol_buffer_updated);
        __pyx_v_self->_protocol_buffer_updated = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "uvloop/handles/stream.pyx":240
 *                 self._protocol_get_buffer = protocol.get_buffer
 *                 self._protocol_buffer_updated = protocol.buffer_updated
 *                 self.__buffered = 1             # <<<<<<<<<<<<<<
 *             except AttributeError:
 *                 pass
 */
        __pyx_v_self->_UVStream__buffered = 1;

        /* "uvloop/handles/stream.pyx":237
 *         if (hasattr(protocol, 'get_buffer') and
 *                 not isinstance(protocol, aio_Protocol)):
 *             try:             # <<<<<<<<<<<<<<
 *                 self._protocol_get_buffer = protocol.get_buffer
 *                 self._protocol_buffer_updated = protocol.buffer_updated
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":241
 *                 self._protocol_buffer_updated = protocol.buffer_updated
 *                 self.__buffered = 1
 *             except AttributeError:             # <<<<<<<<<<<<<<
 *                 pass
 *         else:
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
      if (__pyx_t_8) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "uvloop/handles/stream.pyx":237
 *         if (hasattr(protocol, 'get_buffer') and
 *                 not isinstance(protocol, aio_Protocol)):
 *             try:             # <<<<<<<<<<<<<<
 *                 self._protocol_get_buffer = protocol.get_buffer
 *                 self._protocol_buffer_updated = protocol.buffer_updated
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_L12_try_end:;
    }

    /* "uvloop/handles/stream.pyx":235
 *         UVBaseTransport._set_protocol(self, protocol)
 * 
 *         if (hasattr(protocol, 'get_buffer') and             # <<<<<<<<<<<<<<
 *                 not isinstance(protocol, aio_Protocol)):
 *             try:
 */
    goto __pyx_L4;
  }

  /* "uvloop/handles/stream.pyx":244
 *                 pass
 *         else:
 *             self.__buffered = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef _clear_protocol(self):
 */
  /*else*/ {
    __pyx_v_self->_UVStream__buffered = 0;
  }
  __pyx_L4:;

  /* "uvloop/handles/stream.pyx":229
 *         self._read_pybuf_acquired = False
 * 
 *     cdef _set_protocol(self, object protocol):             # <<<<<<<<<<<<<<
 *         if protocol is None:
 *             raise TypeError('protocol is required')
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVStream._set_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":246
 *             self.__buffered = 0
 * 
 *     cdef _clear_protocol(self):             # <<<<<<<<<<<<<<
 *         UVBaseTransport._clear_protocol(self)
 *         self._protocol_get_buffer = None
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVStream__clear_protocol(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_clear_protocol", 1);

  /* "uvloop/handles/stream.pyx":247
 * 
 *     cdef _clear_protocol(self):
 *         UVBaseTransport._clear_protocol(self)             # <<<<<<<<<<<<<<
 *         self._protocol_get_buffer = None
 *         self._protocol_buffer_updated = None
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_15UVBaseTransport__clear_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":248
 *     cdef _clear_protocol(self):
 *         UVBaseTransport._clear_protocol(self)
 *         self._protocol_get_buffer = None             # <<<<<<<<<<<<<<
 *         self._protocol_buffer_updated = None
 *         self.__buffered = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol_get_buffer);
  __Pyx_DECREF(__pyx_v_self->_protocol_get_buffer);
  __pyx_v_self->_protocol_get_buffer = Py_None;

  /* "uvloop/handles/stream.pyx":249
 *         UVBaseTransport._clear_protocol(self)
 *         self._protocol_get_buffer = None
 *         self._protocol_buffer_updated = None             # <<<<<<<<<<<<<<
 *         self.__buffered = 0
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol_buffer_updated);
  __Pyx_DECREF(__pyx_v_self->_protocol_buffer_updated);
  __pyx_v_self->_protocol_buffer_updated = Py_None;

  /* "uvloop/handles/stream.pyx":250
 *         self._protocol_get_buffer = None
 *         self._protocol_buffer_updated = None
 *         self.__buffered = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _shutdown(self):
 */
  __pyx_v_self->_UVStream__buffered = 0;

  /* "uvloop/handles/stream.pyx":246
 *             self.__buffered = 0
 * 
 *     cdef _clear_protocol(self):             # <<<<<<<<<<<<<<
 *         UVBaseTransport._clear_protocol(self)
 *         self._protocol_get_buffer = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream._clear_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":252
 *         self.__buffered = 0
 * 
 *     cdef inline _shutdown(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__shutdown(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_shutdown", 1);

  /* "uvloop/handles/stream.pyx":255
 *         cdef int err
 * 
 *         if self.__shutting_down:             # <<<<<<<<<<<<<<
 *             return
 *         self.__shutting_down = 1
 */
  if (__pyx_v_self->_UVStream__shutting_down) {

    /* "uvloop/handles/stream.pyx":256
 * 
 *         if self.__shutting_down:
 *             return             # <<<<<<<<<<<<<<
 *         self.__shutting_down = 1
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":255
 *         cdef int err
 * 
 *         if self.__shutting_down:             # <<<<<<<<<<<<<<
 *             return
 *         self.__shutting_down = 1
 */
  }

  /* "uvloop/handles/stream.pyx":257
 *         if self.__shutting_down:
 *             return
 *         self.__shutting_down = 1             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_alive()
 */
  __pyx_v_self->_UVStream__shutting_down = 1;

  /* "uvloop/handles/stream.pyx":259
 *         self.__shutting_down = 1
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         self._shutdown_req.data = <void*> self
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":261
 *         self._ensure_alive()
 * 
 *         self._shutdown_req.data = <void*> self             # <<<<<<<<<<<<<<
 *         err = uv.uv_shutdown(&self._shutdown_req,
 *                              <uv.uv_stream_t*> self._handle,
 */
  __pyx_v_self->_shutdown_req.data = ((void *)__pyx_v_self);

  /* "uvloop/handles/stream.pyx":262
 * 
 *         self._shutdown_req.data = <void*> self
 *         err = uv.uv_shutdown(&self._shutdown_req,             # <<<<<<<<<<<<<<
 *                              <uv.uv_stream_t*> self._handle,
 *                              __uv_stream_on_shutdown)
 */
  __pyx_v_err = uv_shutdown((&__pyx_v_self->_shutdown_req), ((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_f_6uvloop_4loop___uv_stream_on_shutdown);

  /* "uvloop/handles/stream.pyx":265
 *                              <uv.uv_stream_t*> self._handle,
 *                              __uv_stream_on_shutdown)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":266
 *                              __uv_stream_on_shutdown)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":267
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":268
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _accept(self, UVStream server):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":265
 *                              <uv.uv_stream_t*> self._handle,
 *                              __uv_stream_on_shutdown)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/stream.pyx":252
 *         self.__buffered = 0
 * 
 *     cdef inline _shutdown(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream._shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":270
 *             return
 * 
 *     cdef inline _accept(self, UVStream server):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__accept(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_server) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_accept", 1);

  /* "uvloop/handles/stream.pyx":272
 *     cdef inline _accept(self, UVStream server):
 *         cdef int err
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_accept(<uv.uv_stream_t*>server._handle,
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":274
 *         self._ensure_alive()
 * 
 *         err = uv.uv_accept(<uv.uv_stream_t*>server._handle,             # <<<<<<<<<<<<<<
 *                            <uv.uv_stream_t*>self._handle)
 *         if err < 0:
 */
  __pyx_v_err = uv_accept(((uv_stream_t *)__pyx_v_server->__pyx_base.__pyx_base.__pyx_base._handle), ((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle));

  /* "uvloop/handles/stream.pyx":276
 *         err = uv.uv_accept(<uv.uv_stream_t*>server._handle,
 *                            <uv.uv_stream_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":277
 *                            <uv.uv_stream_t*>self._handle)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":278
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":279
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._on_accept()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":276
 *         err = uv.uv_accept(<uv.uv_stream_t*>server._handle,
 *                            <uv.uv_stream_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/stream.pyx":281
 *             return
 * 
 *         self._on_accept()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _close_on_read_error(self):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__on_accept(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":270
 *             return
 * 
 *     cdef inline _accept(self, UVStream server):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream._accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":283
 *         self._on_accept()
 * 
 *     cdef inline _close_on_read_error(self):             # <<<<<<<<<<<<<<
 *         self.__read_error_close = 1
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__close_on_read_error(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_close_on_read_error", 1);

  /* "uvloop/handles/stream.pyx":284
 * 
 *     cdef inline _close_on_read_error(self):
 *         self.__read_error_close = 1             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_reading(self):
 */
  __pyx_v_self->_UVStream__read_error_close = 1;

  /* "uvloop/handles/stream.pyx":283
 *         self._on_accept()
 * 
 *     cdef inline _close_on_read_error(self):             # <<<<<<<<<<<<<<
 *         self.__read_error_close = 1
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":286
 *         self.__read_error_close = 1
 * 
 *     cdef bint _is_reading(self):             # <<<<<<<<<<<<<<
 *         return self.__reading
 * 
 */

static int __pyx_f_6uvloop_4loop_8UVStream__is_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  int __pyx_r;

  /* "uvloop/handles/stream.pyx":287
 * 
 *     cdef bint _is_reading(self):
 *         return self.__reading             # <<<<<<<<<<<<<<
 * 
 *     cdef _start_reading(self):
 */
  __pyx_r = __pyx_v_self->_UVStream__reading;
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":286
 *         self.__read_error_close = 1
 * 
 *     cdef bint _is_reading(self):             # <<<<<<<<<<<<<<
 *         return self.__reading
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":289
 *         return self.__reading
 * 
 *     cdef _start_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVStream__start_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_start_reading", 1);

  /* "uvloop/handles/stream.pyx":292
 *         cdef int err
 * 
 *         if self._closing:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_self->__pyx_base._closing) {

    /* "uvloop/handles/stream.pyx":293
 * 
 *         if self._closing:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_alive()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":292
 *         cdef int err
 * 
 *         if self._closing:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":295
 *             return
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self.__reading:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":297
 *         self._ensure_alive()
 * 
 *         if self.__reading:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_self->_UVStream__reading) {

    /* "uvloop/handles/stream.pyx":298
 * 
 *         if self.__reading:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self.__buffered:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":297
 *         self._ensure_alive()
 * 
 *         if self.__reading:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":300
 *             return
 * 
 *         if self.__buffered:             # <<<<<<<<<<<<<<
 *             err = uv.uv_read_start(<uv.uv_stream_t*>self._handle,
 *                                    __uv_stream_buffered_alloc,
 */
  if (__pyx_v_self->_UVStream__buffered) {

    /* "uvloop/handles/stream.pyx":301
 * 
 *         if self.__buffered:
 *             err = uv.uv_read_start(<uv.uv_stream_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                    __uv_stream_buffered_alloc,
 *                                    __uv_stream_buffered_on_read)
 */
    __pyx_v_err = uv_read_start(((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_f_6uvloop_4loop___uv_stream_buffered_alloc, __pyx_f_6uvloop_4loop___uv_stream_buffered_on_read);

    /* "uvloop/handles/stream.pyx":300
 *             return
 * 
 *         if self.__buffered:             # <<<<<<<<<<<<<<
 *             err = uv.uv_read_start(<uv.uv_stream_t*>self._handle,
 *                                    __uv_stream_buffered_alloc,
 */
    goto __pyx_L5;
  }

  /* "uvloop/handles/stream.pyx":305
 *                                    __uv_stream_buffered_on_read)
 *         else:
 *             err = uv.uv_read_start(<uv.uv_stream_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                    __loop_alloc_buffer,
 *                                    __uv_stream_on_read)
 */
  /*else*/ {

    /* "uvloop/handles/stream.pyx":307
 *             err = uv.uv_read_start(<uv.uv_stream_t*>self._handle,
 *                                    __loop_alloc_buffer,
 *                                    __uv_stream_on_read)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
    __pyx_v_err = uv_read_start(((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_f_6uvloop_4loop___loop_alloc_buffer, __pyx_f_6uvloop_4loop___uv_stream_on_read);
  }
  __pyx_L5:;

  /* "uvloop/handles/stream.pyx":308
 *                                    __loop_alloc_buffer,
 *                                    __uv_stream_on_read)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":309
 *                                    __uv_stream_on_read)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":310
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 *         else:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":311
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 *         else:
 *             # UVStream must live until the read callback is called
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":308
 *                                    __loop_alloc_buffer,
 *                                    __uv_stream_on_read)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/stream.pyx":314
 *         else:
 *             # UVStream must live until the read callback is called
 *             self.__reading_started()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline __reading_started(self):
 */
  /*else*/ {
    __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_started(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "uvloop/handles/stream.pyx":289
 *         return self.__reading
 * 
 *     cdef _start_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream._start_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":316
 *             self.__reading_started()
 * 
 *     cdef inline __reading_started(self):             # <<<<<<<<<<<<<<
 *         if self.__reading:
 *             return
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_started(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_UVStream__reading_started", 1);

  /* "uvloop/handles/stream.pyx":317
 * 
 *     cdef inline __reading_started(self):
 *         if self.__reading:             # <<<<<<<<<<<<<<
 *             return
 *         self.__reading = 1
 */
  if (__pyx_v_self->_UVStream__reading) {

    /* "uvloop/handles/stream.pyx":318
 *     cdef inline __reading_started(self):
 *         if self.__reading:
 *             return             # <<<<<<<<<<<<<<
 *         self.__reading = 1
 *         Py_INCREF(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":317
 * 
 *     cdef inline __reading_started(self):
 *         if self.__reading:             # <<<<<<<<<<<<<<
 *             return
 *         self.__reading = 1
 */
  }

  /* "uvloop/handles/stream.pyx":319
 *         if self.__reading:
 *             return
 *         self.__reading = 1             # <<<<<<<<<<<<<<
 *         Py_INCREF(self)
 * 
 */
  __pyx_v_self->_UVStream__reading = 1;

  /* "uvloop/handles/stream.pyx":320
 *             return
 *         self.__reading = 1
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline __reading_stopped(self):
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "uvloop/handles/stream.pyx":316
 *             self.__reading_started()
 * 
 *     cdef inline __reading_started(self):             # <<<<<<<<<<<<<<
 *         if self.__reading:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":322
 *         Py_INCREF(self)
 * 
 *     cdef inline __reading_stopped(self):             # <<<<<<<<<<<<<<
 *         if not self.__reading:
 *             return
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_stopped(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_UVStream__reading_stopped", 1);

  /* "uvloop/handles/stream.pyx":323
 * 
 *     cdef inline __reading_stopped(self):
 *         if not self.__reading:             # <<<<<<<<<<<<<<
 *             return
 *         self.__reading = 0
 */
  __pyx_t_1 = (!__pyx_v_self->_UVStream__reading);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":324
 *     cdef inline __reading_stopped(self):
 *         if not self.__reading:
 *             return             # <<<<<<<<<<<<<<
 *         self.__reading = 0
 *         Py_DECREF(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":323
 * 
 *     cdef inline __reading_stopped(self):
 *         if not self.__reading:             # <<<<<<<<<<<<<<
 *             return
 *         self.__reading = 0
 */
  }

  /* "uvloop/handles/stream.pyx":325
 *         if not self.__reading:
 *             return
 *         self.__reading = 0             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
 */
  __pyx_v_self->_UVStream__reading = 0;

  /* "uvloop/handles/stream.pyx":326
 *             return
 *         self.__reading = 0
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _stop_reading(self):
 */
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "uvloop/handles/stream.pyx":322
 *         Py_INCREF(self)
 * 
 *     cdef inline __reading_stopped(self):             # <<<<<<<<<<<<<<
 *         if not self.__reading:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":328
 *         Py_DECREF(self)
 * 
 *     cdef _stop_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVStream__stop_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_stop_reading", 1);

  /* "uvloop/handles/stream.pyx":331
 *         cdef int err
 * 
 *         if not self.__reading:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (!__pyx_v_self->_UVStream__reading);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":332
 * 
 *         if not self.__reading:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_alive()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":331
 *         cdef int err
 * 
 *         if not self.__reading:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":334
 *             return
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         # From libuv docs:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/stream.pyx":339
 *         #    This function is idempotent and may be safely
 *         #    called on a stopped stream.
 *         err = uv.uv_read_stop(<uv.uv_stream_t*>self._handle)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  __pyx_v_err = uv_read_stop(((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle));

  /* "uvloop/handles/stream.pyx":340
 *         #    called on a stopped stream.
 *         err = uv.uv_read_stop(<uv.uv_stream_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":341
 *         err = uv.uv_read_stop(<uv.uv_stream_t*>self._handle)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/stream.pyx":342
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 *         else:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/stream.pyx":343
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 *         else:
 *             self.__reading_stopped()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":340
 *         #    called on a stopped stream.
 *         err = uv.uv_read_stop(<uv.uv_stream_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/stream.pyx":345
 *             return
 *         else:
 *             self.__reading_stopped()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _try_write(self, object data):
 */
  /*else*/ {
    __pyx_t_2 = __pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_stopped(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/handles/stream.pyx":328
 *         Py_DECREF(self)
 * 
 *     cdef _stop_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVStream._stop_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":347
 *             self.__reading_stopped()
 * 
 *     cdef inline _try_write(self, object data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t written
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__try_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_data) {
  Py_ssize_t __pyx_v_written;
  int __pyx_v_used_buf;
  Py_buffer __pyx_v_py_buf;
  void *__pyx_v_buf;
  size_t __pyx_v_blen;
  int __pyx_v_saved_errno;
  int __pyx_v_fd;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  void *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_try_write", 1);

  /* "uvloop/handles/stream.pyx":350
 *         cdef:
 *             ssize_t written
 *             bint used_buf = 0             # <<<<<<<<<<<<<<
 *             Py_buffer py_buf
 *             void* buf
 */
  __pyx_v_used_buf = 0;

  /* "uvloop/handles/stream.pyx":357
 *             int fd
 * 
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size != 0:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'UVStream._try_write called with data in uv buffers')
 */
  __pyx_t_1 = (((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle)->write_queue_size != 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/stream.pyx":358
 * 
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size != 0:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'UVStream._try_write called with data in uv buffers')
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__151, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(20, 358, __pyx_L1_error)

    /* "uvloop/handles/stream.pyx":357
 *             int fd
 * 
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size != 0:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'UVStream._try_write called with data in uv buffers')
 */
  }

  /* "uvloop/handles/stream.pyx":361
 *                 'UVStream._try_write called with data in uv buffers')
 * 
 *         if PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             # We can only use this hack for bytes since it's
 *             # immutable.  For everything else it is only safe to
 */
  __pyx_t_1 = PyBytes_CheckExact(__pyx_v_data);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":365
 *             # immutable.  For everything else it is only safe to
 *             # use buffer protocol.
 *             buf = <void*>PyBytes_AS_STRING(data)             # <<<<<<<<<<<<<<
 *             blen = Py_SIZE(data)
 *         else:
 */
    __pyx_v_buf = ((void *)PyBytes_AS_STRING(__pyx_v_data));

    /* "uvloop/handles/stream.pyx":366
 *             # use buffer protocol.
 *             buf = <void*>PyBytes_AS_STRING(data)
 *             blen = Py_SIZE(data)             # <<<<<<<<<<<<<<
 *         else:
 *             PyObject_GetBuffer(data, &py_buf, PyBUF_SIMPLE)
 */
    __pyx_v_blen = Py_SIZE(__pyx_v_data);

    /* "uvloop/handles/stream.pyx":361
 *                 'UVStream._try_write called with data in uv buffers')
 * 
 *         if PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             # We can only use this hack for bytes since it's
 *             # immutable.  For everything else it is only safe to
 */
    goto __pyx_L4;
  }

  /* "uvloop/handles/stream.pyx":368
 *             blen = Py_SIZE(data)
 *         else:
 *             PyObject_GetBuffer(data, &py_buf, PyBUF_SIMPLE)             # <<<<<<<<<<<<<<
 *             used_buf = 1
 *             buf = py_buf.buf
 */
  /*else*/ {
    __pyx_t_3 = PyObject_GetBuffer(__pyx_v_data, (&__pyx_v_py_buf), PyBUF_SIMPLE); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(20, 368, __pyx_L1_error)

    /* "uvloop/handles/stream.pyx":369
 *         else:
 *             PyObject_GetBuffer(data, &py_buf, PyBUF_SIMPLE)
 *             used_buf = 1             # <<<<<<<<<<<<<<
 *             buf = py_buf.buf
 *             blen = py_buf.len
 */
    __pyx_v_used_buf = 1;

    /* "uvloop/handles/stream.pyx":370
 *             PyObject_GetBuffer(data, &py_buf, PyBUF_SIMPLE)
 *             used_buf = 1
 *             buf = py_buf.buf             # <<<<<<<<<<<<<<
 *             blen = py_buf.len
 * 
 */
    __pyx_t_4 = __pyx_v_py_buf.buf;
    __pyx_v_buf = __pyx_t_4;

    /* "uvloop/handles/stream.pyx":371
 *             used_buf = 1
 *             buf = py_buf.buf
 *             blen = py_buf.len             # <<<<<<<<<<<<<<
 * 
 *         if blen == 0:
 */
    __pyx_t_5 = __pyx_v_py_buf.len;
    __pyx_v_blen = __pyx_t_5;
  }
  __pyx_L4:;

  /* "uvloop/handles/stream.pyx":373
 *             blen = py_buf.len
 * 
 *         if blen == 0:             # <<<<<<<<<<<<<<
 *             # Empty data, do nothing.
 *             return 0
 */
  __pyx_t_1 = (__pyx_v_blen == 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":375
 *         if blen == 0:
 *             # Empty data, do nothing.
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         fd = self._fileno()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":373
 *             blen = py_buf.len
 * 
 *         if blen == 0:             # <<<<<<<<<<<<<<
 *             # Empty data, do nothing.
 *             return 0
 */
  }

  /* "uvloop/handles/stream.pyx":377
 *             return 0
 * 
 *         fd = self._fileno()             # <<<<<<<<<<<<<<
 *         # Use `unistd.h/write` directly, it's faster than
 *         # uv_try_write -- less layers of code.  The error
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._fileno(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(20, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_fd = __pyx_t_3;

  /* "uvloop/handles/stream.pyx":381
 *         # uv_try_write -- less layers of code.  The error
 *         # checking logic is copied from libuv.
 *         written = system.write(fd, buf, blen)             # <<<<<<<<<<<<<<
 *         while written == -1 and (
 *                 errno.errno == errno.EINTR or
 */
  __pyx_v_written = write(__pyx_v_fd, __pyx_v_buf, __pyx_v_blen);

  /* "uvloop/handles/stream.pyx":382
 *         # checking logic is copied from libuv.
 *         written = system.write(fd, buf, blen)
 *         while written == -1 and (             # <<<<<<<<<<<<<<
 *                 errno.errno == errno.EINTR or
 *                 (system.PLATFORM_IS_APPLE and
 */
  while (1) {
    __pyx_t_6 = (__pyx_v_written == -1L);
    if (__pyx_t_6) {
    } else {
      __pyx_t_1 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }

    /* "uvloop/handles/stream.pyx":383
 *         written = system.write(fd, buf, blen)
 *         while written == -1 and (
 *                 errno.errno == errno.EINTR or             # <<<<<<<<<<<<<<
 *                 (system.PLATFORM_IS_APPLE and
 *                     errno.errno == errno.EPROTOTYPE)):
 */
    __pyx_t_6 = (errno == EINTR);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_1 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }

    /* "uvloop/handles/stream.pyx":384
 *         while written == -1 and (
 *                 errno.errno == errno.EINTR or
 *                 (system.PLATFORM_IS_APPLE and             # <<<<<<<<<<<<<<
 *                     errno.errno == errno.EPROTOTYPE)):
 *             # From libuv code (unix/stream.c):
 */
    __pyx_t_6 = (PLATFORM_IS_APPLE != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_1 = __pyx_t_6;
      goto __pyx_L8_bool_binop_done;
    }

    /* "uvloop/handles/stream.pyx":385
 *                 errno.errno == errno.EINTR or
 *                 (system.PLATFORM_IS_APPLE and
 *                     errno.errno == errno.EPROTOTYPE)):             # <<<<<<<<<<<<<<
 *             # From libuv code (unix/stream.c):
 *             #   Due to a possible kernel bug at least in OS X 10.10 "Yosemite",
 */
    __pyx_t_6 = (errno == EPROTOTYPE);
    __pyx_t_1 = __pyx_t_6;
    __pyx_L8_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "uvloop/handles/stream.pyx":391
 *             #   that is shutting down. If we retry the write, we should get
 *             #   the expected EPIPE instead.
 *             written = system.write(fd, buf, blen)             # <<<<<<<<<<<<<<
 *         saved_errno = errno.errno
 * 
 */
    __pyx_v_written = write(__pyx_v_fd, __pyx_v_buf, __pyx_v_blen);
  }

  /* "uvloop/handles/stream.pyx":392
 *             #   the expected EPIPE instead.
 *             written = system.write(fd, buf, blen)
 *         saved_errno = errno.errno             # <<<<<<<<<<<<<<
 * 
 *         if used_buf:
 */
  __pyx_v_saved_errno = errno;

  /* "uvloop/handles/stream.pyx":394
 *         saved_errno = errno.errno
 * 
 *         if used_buf:             # <<<<<<<<<<<<<<
 *             PyBuffer_Release(&py_buf)
 * 
 */
  if (__pyx_v_used_buf) {

    /* "uvloop/handles/stream.pyx":395
 * 
 *         if used_buf:
 *             PyBuffer_Release(&py_buf)             # <<<<<<<<<<<<<<
 * 
 *         if written < 0:
 */
    PyBuffer_Release((&__pyx_v_py_buf));

    /* "uvloop/handles/stream.pyx":394
 *         saved_errno = errno.errno
 * 
 *         if used_buf:             # <<<<<<<<<<<<<<
 *             PyBuffer_Release(&py_buf)
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":397
 *             PyBuffer_Release(&py_buf)
 * 
 *         if written < 0:             # <<<<<<<<<<<<<<
 *             if saved_errno == errno.EAGAIN or \
 *                     saved_errno == system.EWOULDBLOCK:
 */
  __pyx_t_1 = (__pyx_v_written < 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":398
 * 
 *         if written < 0:
 *             if saved_errno == errno.EAGAIN or \             # <<<<<<<<<<<<<<
 *                     saved_errno == system.EWOULDBLOCK:
 *                 return -1
 */
    __pyx_t_6 = (__pyx_v_saved_errno == EAGAIN);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_1 = __pyx_t_6;
      goto __pyx_L15_bool_binop_done;
    }

    /* "uvloop/handles/stream.pyx":399
 *         if written < 0:
 *             if saved_errno == errno.EAGAIN or \
 *                     saved_errno == system.EWOULDBLOCK:             # <<<<<<<<<<<<<<
 *                 return -1
 *             else:
 */
    __pyx_t_6 = (__pyx_v_saved_errno == EWOULDBLOCK);
    __pyx_t_1 = __pyx_t_6;
    __pyx_L15_bool_binop_done:;

    /* "uvloop/handles/stream.pyx":398
 * 
 *         if written < 0:
 *             if saved_errno == errno.EAGAIN or \             # <<<<<<<<<<<<<<
 *                     saved_errno == system.EWOULDBLOCK:
 *                 return -1
 */
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":400
 *             if saved_errno == errno.EAGAIN or \
 *                     saved_errno == system.EWOULDBLOCK:
 *                 return -1             # <<<<<<<<<<<<<<
 *             else:
 *                 exc = convert_error(-saved_errno)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_int_neg_1);
      __pyx_r = __pyx_int_neg_1;
      goto __pyx_L0;

      /* "uvloop/handles/stream.pyx":398
 * 
 *         if written < 0:
 *             if saved_errno == errno.EAGAIN or \             # <<<<<<<<<<<<<<
 *                     saved_errno == system.EWOULDBLOCK:
 *                 return -1
 */
    }

    /* "uvloop/handles/stream.pyx":402
 *                 return -1
 *             else:
 *                 exc = convert_error(-saved_errno)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
    /*else*/ {
      __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error((-__pyx_v_saved_errno)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_exc = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":403
 *             else:
 *                 exc = convert_error(-saved_errno)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":404
 *                 exc = convert_error(-saved_errno)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG:
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;
    }

    /* "uvloop/handles/stream.pyx":397
 *             PyBuffer_Release(&py_buf)
 * 
 *         if written < 0:             # <<<<<<<<<<<<<<
 *             if saved_errno == errno.EAGAIN or \
 *                     saved_errno == system.EWOULDBLOCK:
 */
  }

  /* "uvloop/handles/stream.pyx":406
 *                 return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self._loop._debug_stream_write_tries += 1
 * 
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":407
 * 
 *         if UVLOOP_DEBUG:
 *             self._loop._debug_stream_write_tries += 1             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>written == blen:
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop);
    __pyx_t_7 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop;
    __pyx_t_7->_debug_stream_write_tries = (__pyx_t_7->_debug_stream_write_tries + 1);
    __Pyx_DECREF((PyObject *)__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/handles/stream.pyx":406
 *                 return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             self._loop._debug_stream_write_tries += 1
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":409
 *             self._loop._debug_stream_write_tries += 1
 * 
 *         if <size_t>written == blen:             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
  __pyx_t_1 = (((size_t)__pyx_v_written) == __pyx_v_blen);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":410
 * 
 *         if <size_t>written == blen:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         return written
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":409
 *             self._loop._debug_stream_write_tries += 1
 * 
 *         if <size_t>written == blen:             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":412
 *             return 0
 * 
 *         return written             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _buffer_write(self, object data):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_written); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":347
 *             self.__reading_stopped()
 * 
 *     cdef inline _try_write(self, object data):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ssize_t written
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF((PyObject *)__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UVStream._try_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":414
 *         return written
 * 
 *     cdef inline _buffer_write(self, object data):             # <<<<<<<<<<<<<<
 *         cdef int dlen
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__buffer_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_data) {
  int __pyx_v_dlen;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_buffer_write", 0);
  __Pyx_INCREF(__pyx_v_data);

  /* "uvloop/handles/stream.pyx":417
 *         cdef int dlen
 * 
 *         if not PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             data = memoryview(data).cast('b')
 * 
 */
  __pyx_t_1 = (!PyBytes_CheckExact(__pyx_v_data));
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":418
 * 
 *         if not PyBytes_CheckExact(data):
 *             data = memoryview(data).cast('b')             # <<<<<<<<<<<<<<
 * 
 *         dlen = len(data)
 */
    __pyx_t_3 = PyMemoryView_FromObject(__pyx_v_data); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_cast); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_n_u_b};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 418, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/stream.pyx":417
 *         cdef int dlen
 * 
 *         if not PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             data = memoryview(data).cast('b')
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":420
 *             data = memoryview(data).cast('b')
 * 
 *         dlen = len(data)             # <<<<<<<<<<<<<<
 *         if not dlen:
 *             return
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(20, 420, __pyx_L1_error)
  __pyx_v_dlen = __pyx_t_6;

  /* "uvloop/handles/stream.pyx":421
 * 
 *         dlen = len(data)
 *         if not dlen:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (!(__pyx_v_dlen != 0));
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":422
 *         dlen = len(data)
 *         if not dlen:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._buffer_size += dlen
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":421
 * 
 *         dlen = len(data)
 *         if not dlen:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":424
 *             return
 * 
 *         self._buffer_size += dlen             # <<<<<<<<<<<<<<
 *         self._buffer.append(data)
 * 
 */
  __pyx_v_self->_buffer_size = (__pyx_v_self->_buffer_size + __pyx_v_dlen);

  /* "uvloop/handles/stream.pyx":425
 * 
 *         self._buffer_size += dlen
 *         self._buffer.append(data)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _initiate_write(self):
 */
  if (unlikely(__pyx_v_self->_buffer == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(20, 425, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->_buffer, __pyx_v_data); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(20, 425, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":414
 *         return written
 * 
 *     cdef inline _buffer_write(self, object data):             # <<<<<<<<<<<<<<
 *         cdef int dlen
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVStream._buffer_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":427
 *         self._buffer.append(data)
 * 
 *     cdef inline _initiate_write(self):             # <<<<<<<<<<<<<<
 *         if (not self._protocol_paused and
 *                 (<uv.uv_stream_t*>self._handle).write_queue_size == 0 and
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__initiate_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_v_all_sent = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_initiate_write", 1);

  /* "uvloop/handles/stream.pyx":428
 * 
 *     cdef inline _initiate_write(self):
 *         if (not self._protocol_paused and             # <<<<<<<<<<<<<<
 *                 (<uv.uv_stream_t*>self._handle).write_queue_size == 0 and
 *                 self._buffer_size > self._high_water):
 */
  __pyx_t_2 = (!__pyx_v_self->__pyx_base._protocol_paused);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "uvloop/handles/stream.pyx":429
 *     cdef inline _initiate_write(self):
 *         if (not self._protocol_paused and
 *                 (<uv.uv_stream_t*>self._handle).write_queue_size == 0 and             # <<<<<<<<<<<<<<
 *                 self._buffer_size > self._high_water):
 *             # Fast-path.  If:
 */
  __pyx_t_2 = (((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle)->write_queue_size == 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "uvloop/handles/stream.pyx":430
 *         if (not self._protocol_paused and
 *                 (<uv.uv_stream_t*>self._handle).write_queue_size == 0 and
 *                 self._buffer_size > self._high_water):             # <<<<<<<<<<<<<<
 *             # Fast-path.  If:
 *             #   - the protocol isn't yet paused,
 */
  __pyx_t_2 = (__pyx_v_self->_buffer_size > __pyx_v_self->__pyx_base._high_water);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "uvloop/handles/stream.pyx":428
 * 
 *     cdef inline _initiate_write(self):
 *         if (not self._protocol_paused and             # <<<<<<<<<<<<<<
 *                 (<uv.uv_stream_t*>self._handle).write_queue_size == 0 and
 *                 self._buffer_size > self._high_water):
 */
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":438
 *             # Then:
 *             #   - Try to write all buffered data right now.
 *             all_sent = self._exec_write()             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 if self._buffer_size != 0 or self._buffer != []:
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_8UVStream__exec_write(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_all_sent = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/handles/stream.pyx":439
 *             #   - Try to write all buffered data right now.
 *             all_sent = self._exec_write()
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if self._buffer_size != 0 or self._buffer != []:
 *                     raise RuntimeError(
 */
    __pyx_t_1 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":440
 *             all_sent = self._exec_write()
 *             if UVLOOP_DEBUG:
 *                 if self._buffer_size != 0 or self._buffer != []:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_buffer_size is not 0 after a successful _exec_write')
 */
      __pyx_t_2 = (__pyx_v_self->_buffer_size != 0);
      if (!__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 440, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->_buffer, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 440, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(20, 440, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_1 = __pyx_t_2;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "uvloop/handles/stream.pyx":441
 *             if UVLOOP_DEBUG:
 *                 if self._buffer_size != 0 or self._buffer != []:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_buffer_size is not 0 after a successful _exec_write')
 * 
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__152, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 441, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(20, 441, __pyx_L1_error)

        /* "uvloop/handles/stream.pyx":440
 *             all_sent = self._exec_write()
 *             if UVLOOP_DEBUG:
 *                 if self._buffer_size != 0 or self._buffer != []:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError(
 *                         '_buffer_size is not 0 after a successful _exec_write')
 */
      }

      /* "uvloop/handles/stream.pyx":439
 *             #   - Try to write all buffered data right now.
 *             all_sent = self._exec_write()
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 if self._buffer_size != 0 or self._buffer != []:
 *                     raise RuntimeError(
 */
    }

    /* "uvloop/handles/stream.pyx":447
 *             # as `uv_write` should be called already.
 * 
 *             if not all_sent:             # <<<<<<<<<<<<<<
 *                 # If not all of the data was sent successfully,
 *                 # we might need to pause the protocol.
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_all_sent); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(20, 447, __pyx_L1_error)
    __pyx_t_2 = (!__pyx_t_1);
    if (__pyx_t_2) {

      /* "uvloop/handles/stream.pyx":450
 *                 # If not all of the data was sent successfully,
 *                 # we might need to pause the protocol.
 *                 self._maybe_pause_protocol()             # <<<<<<<<<<<<<<
 * 
 *         elif self._buffer_size > 0:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._maybe_pause_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 450, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/handles/stream.pyx":447
 *             # as `uv_write` should be called already.
 * 
 *             if not all_sent:             # <<<<<<<<<<<<<<
 *                 # If not all of the data was sent successfully,
 *                 # we might need to pause the protocol.
 */
    }

    /* "uvloop/handles/stream.pyx":428
 * 
 *     cdef inline _initiate_write(self):
 *         if (not self._protocol_paused and             # <<<<<<<<<<<<<<
 *                 (<uv.uv_stream_t*>self._handle).write_queue_size == 0 and
 *                 self._buffer_size > self._high_water):
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/stream.pyx":452
 *                 self._maybe_pause_protocol()
 * 
 *         elif self._buffer_size > 0:             # <<<<<<<<<<<<<<
 *             self._maybe_pause_protocol()
 *             self._loop._queue_write(self)
 */
  __pyx_t_2 = (__pyx_v_self->_buffer_size > 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":453
 * 
 *         elif self._buffer_size > 0:
 *             self._maybe_pause_protocol()             # <<<<<<<<<<<<<<
 *             self._loop._queue_write(self)
 * 
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._maybe_pause_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/stream.pyx":454
 *         elif self._buffer_size > 0:
 *             self._maybe_pause_protocol()
 *             self._loop._queue_write(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _exec_write(self):
 */
    __pyx_t_4 = __pyx_f_6uvloop_4loop_4Loop__queue_write(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop, __pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/stream.pyx":452
 *                 self._maybe_pause_protocol()
 * 
 *         elif self._buffer_size > 0:             # <<<<<<<<<<<<<<
 *             self._maybe_pause_protocol()
 *             self._loop._queue_write(self)
 */
  }
  __pyx_L3:;

  /* "uvloop/handles/stream.pyx":427
 *         self._buffer.append(data)
 * 
 *     cdef inline _initiate_write(self):             # <<<<<<<<<<<<<<
 *         if (not self._protocol_paused and
 *                 (<uv.uv_stream_t*>self._handle).write_queue_size == 0 and
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVStream._initiate_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_all_sent);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":456
 *             self._loop._queue_write(self)
 * 
 *     cdef inline _exec_write(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__exec_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  int __pyx_v_err;
  int __pyx_v_buf_len;
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_ctx = 0;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_sent = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_exec_write", 1);

  /* "uvloop/handles/stream.pyx":460
 *             int err
 *             int buf_len
 *             _StreamWriteContext ctx = None             # <<<<<<<<<<<<<<
 * 
 *         if self._closed:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_ctx = ((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)Py_None);

  /* "uvloop/handles/stream.pyx":462
 *             _StreamWriteContext ctx = None
 * 
 *         if self._closed:             # <<<<<<<<<<<<<<
 *             # If the handle is closed, just return, it's too
 *             # late to do anything.
 */
  if (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._closed) {

    /* "uvloop/handles/stream.pyx":465
 *             # If the handle is closed, just return, it's too
 *             # late to do anything.
 *             return             # <<<<<<<<<<<<<<
 * 
 *         buf_len = len(self._buffer)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":462
 *             _StreamWriteContext ctx = None
 * 
 *         if self._closed:             # <<<<<<<<<<<<<<
 *             # If the handle is closed, just return, it's too
 *             # late to do anything.
 */
  }

  /* "uvloop/handles/stream.pyx":467
 *             return
 * 
 *         buf_len = len(self._buffer)             # <<<<<<<<<<<<<<
 *         if not buf_len:
 *             return
 */
  __pyx_t_1 = __pyx_v_self->_buffer;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(20, 467, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(20, 467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_buf_len = __pyx_t_2;

  /* "uvloop/handles/stream.pyx":468
 * 
 *         buf_len = len(self._buffer)
 *         if not buf_len:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_3 = (!(__pyx_v_buf_len != 0));
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":469
 *         buf_len = len(self._buffer)
 *         if not buf_len:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size == 0:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":468
 * 
 *         buf_len = len(self._buffer)
 *         if not buf_len:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":471
 *             return
 * 
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size == 0:             # <<<<<<<<<<<<<<
 *             # libuv internal write buffers for this stream are empty.
 *             if buf_len == 1:
 */
  __pyx_t_3 = (((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle)->write_queue_size == 0);
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":473
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size == 0:
 *             # libuv internal write buffers for this stream are empty.
 *             if buf_len == 1:             # <<<<<<<<<<<<<<
 *                 # If we only have one piece of data to send, let's
 *                 # use our fast implementation of try_write.
 */
    __pyx_t_3 = (__pyx_v_buf_len == 1);
    if (__pyx_t_3) {

      /* "uvloop/handles/stream.pyx":476
 *                 # If we only have one piece of data to send, let's
 *                 # use our fast implementation of try_write.
 *                 data = self._buffer[0]             # <<<<<<<<<<<<<<
 *                 sent = self._try_write(data)
 * 
 */
      if (unlikely(__pyx_v_self->_buffer == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(20, 476, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_self->_buffer, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_data = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/handles/stream.pyx":477
 *                 # use our fast implementation of try_write.
 *                 data = self._buffer[0]
 *                 sent = self._try_write(data)             # <<<<<<<<<<<<<<
 * 
 *                 if sent is None:
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__try_write(__pyx_v_self, __pyx_v_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 477, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_sent = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/handles/stream.pyx":479
 *                 sent = self._try_write(data)
 * 
 *                 if sent is None:             # <<<<<<<<<<<<<<
 *                     # A `self._fatal_error` was called.
 *                     # It might not raise an exception under some
 */
      __pyx_t_3 = (__pyx_v_sent == Py_None);
      if (__pyx_t_3) {

        /* "uvloop/handles/stream.pyx":483
 *                     # It might not raise an exception under some
 *                     # conditions.
 *                     self._buffer_size = 0             # <<<<<<<<<<<<<<
 *                     self._buffer.clear()
 *                     if not self._closing:
 */
        __pyx_v_self->_buffer_size = 0;

        /* "uvloop/handles/stream.pyx":484
 *                     # conditions.
 *                     self._buffer_size = 0
 *                     self._buffer.clear()             # <<<<<<<<<<<<<<
 *                     if not self._closing:
 *                         # This should never happen.
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_buffer, __pyx_n_s_clear); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 484, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/stream.pyx":485
 *                     self._buffer_size = 0
 *                     self._buffer.clear()
 *                     if not self._closing:             # <<<<<<<<<<<<<<
 *                         # This should never happen.
 *                         raise RuntimeError(
 */
        __pyx_t_3 = (!__pyx_v_self->__pyx_base._closing);
        if (unlikely(__pyx_t_3)) {

          /* "uvloop/handles/stream.pyx":487
 *                     if not self._closing:
 *                         # This should never happen.
 *                         raise RuntimeError(             # <<<<<<<<<<<<<<
 *                             'stream is open after UVStream._try_write '
 *                             'returned None')
 */
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__153, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 487, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_Raise(__pyx_t_1, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(20, 487, __pyx_L1_error)

          /* "uvloop/handles/stream.pyx":485
 *                     self._buffer_size = 0
 *                     self._buffer.clear()
 *                     if not self._closing:             # <<<<<<<<<<<<<<
 *                         # This should never happen.
 *                         raise RuntimeError(
 */
        }

        /* "uvloop/handles/stream.pyx":490
 *                             'stream is open after UVStream._try_write '
 *                             'returned None')
 *                     return             # <<<<<<<<<<<<<<
 * 
 *                 if sent == 0:
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L0;

        /* "uvloop/handles/stream.pyx":479
 *                 sent = self._try_write(data)
 * 
 *                 if sent is None:             # <<<<<<<<<<<<<<
 *                     # A `self._fatal_error` was called.
 *                     # It might not raise an exception under some
 */
      }

      /* "uvloop/handles/stream.pyx":492
 *                     return
 * 
 *                 if sent == 0:             # <<<<<<<<<<<<<<
 *                     # All data was successfully written.
 *                     self._buffer_size = 0
 */
      __pyx_t_3 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_sent, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(20, 492, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "uvloop/handles/stream.pyx":494
 *                 if sent == 0:
 *                     # All data was successfully written.
 *                     self._buffer_size = 0             # <<<<<<<<<<<<<<
 *                     self._buffer.clear()
 *                     # on_write will call "maybe_resume_protocol".
 */
        __pyx_v_self->_buffer_size = 0;

        /* "uvloop/handles/stream.pyx":495
 *                     # All data was successfully written.
 *                     self._buffer_size = 0
 *                     self._buffer.clear()             # <<<<<<<<<<<<<<
 *                     # on_write will call "maybe_resume_protocol".
 *                     self._on_write()
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_buffer, __pyx_n_s_clear); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 495, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 495, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/stream.pyx":497
 *                     self._buffer.clear()
 *                     # on_write will call "maybe_resume_protocol".
 *                     self._on_write()             # <<<<<<<<<<<<<<
 *                     return True
 * 
 */
        __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__on_write(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 497, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/stream.pyx":498
 *                     # on_write will call "maybe_resume_protocol".
 *                     self._on_write()
 *                     return True             # <<<<<<<<<<<<<<
 * 
 *                 if sent > 0:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_True);
        __pyx_r = Py_True;
        goto __pyx_L0;

        /* "uvloop/handles/stream.pyx":492
 *                     return
 * 
 *                 if sent == 0:             # <<<<<<<<<<<<<<
 *                     # All data was successfully written.
 *                     self._buffer_size = 0
 */
      }

      /* "uvloop/handles/stream.pyx":500
 *                     return True
 * 
 *                 if sent > 0:             # <<<<<<<<<<<<<<
 *                     if UVLOOP_DEBUG:
 *                         if sent == len(data):
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_sent, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 500, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(20, 500, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_3) {

        /* "uvloop/handles/stream.pyx":501
 * 
 *                 if sent > 0:
 *                     if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                         if sent == len(data):
 *                             raise RuntimeError(
 */
        __pyx_t_3 = (UVLOOP_DEBUG != 0);
        if (__pyx_t_3) {

          /* "uvloop/handles/stream.pyx":502
 *                 if sent > 0:
 *                     if UVLOOP_DEBUG:
 *                         if sent == len(data):             # <<<<<<<<<<<<<<
 *                             raise RuntimeError(
 *                                 '_try_write sent all data and returned '
 */
          __pyx_t_2 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(20, 502, __pyx_L1_error)
          __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 502, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = PyObject_RichCompare(__pyx_v_sent, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 502, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(20, 502, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(__pyx_t_3)) {

            /* "uvloop/handles/stream.pyx":503
 *                     if UVLOOP_DEBUG:
 *                         if sent == len(data):
 *                             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                                 '_try_write sent all data and returned '
 *                                 'non-zero')
 */
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__154, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 503, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_Raise(__pyx_t_4, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_ERR(20, 503, __pyx_L1_error)

            /* "uvloop/handles/stream.pyx":502
 *                 if sent > 0:
 *                     if UVLOOP_DEBUG:
 *                         if sent == len(data):             # <<<<<<<<<<<<<<
 *                             raise RuntimeError(
 *                                 '_try_write sent all data and returned '
 */
          }

          /* "uvloop/handles/stream.pyx":501
 * 
 *                 if sent > 0:
 *                     if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                         if sent == len(data):
 *                             raise RuntimeError(
 */
        }

        /* "uvloop/handles/stream.pyx":507
 *                                 'non-zero')
 * 
 *                     if PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *                         # Cast bytes to memoryview to avoid copying
 *                         # data that wasn't sent.
 */
        __pyx_t_3 = PyBytes_CheckExact(__pyx_v_data);
        if (__pyx_t_3) {

          /* "uvloop/handles/stream.pyx":510
 *                         # Cast bytes to memoryview to avoid copying
 *                         # data that wasn't sent.
 *                         data = memoryview(data)             # <<<<<<<<<<<<<<
 *                     data = data[sent:]
 * 
 */
          __pyx_t_4 = PyMemoryView_FromObject(__pyx_v_data); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 510, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_4);
          __pyx_t_4 = 0;

          /* "uvloop/handles/stream.pyx":507
 *                                 'non-zero')
 * 
 *                     if PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *                         # Cast bytes to memoryview to avoid copying
 *                         # data that wasn't sent.
 */
        }

        /* "uvloop/handles/stream.pyx":511
 *                         # data that wasn't sent.
 *                         data = memoryview(data)
 *                     data = data[sent:]             # <<<<<<<<<<<<<<
 * 
 *                     self._buffer_size -= sent
 */
        __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_data, 0, 0, &__pyx_v_sent, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 511, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "uvloop/handles/stream.pyx":513
 *                     data = data[sent:]
 * 
 *                     self._buffer_size -= sent             # <<<<<<<<<<<<<<
 *                     self._buffer[0] = data
 * 
 */
        __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_buffer_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = PyNumber_InPlaceSubtract(__pyx_t_4, __pyx_v_sent); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(20, 513, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_self->_buffer_size = __pyx_t_7;

        /* "uvloop/handles/stream.pyx":514
 * 
 *                     self._buffer_size -= sent
 *                     self._buffer[0] = data             # <<<<<<<<<<<<<<
 * 
 *                 # At this point it's either data was sent partially,
 */
        if (unlikely(__pyx_v_self->_buffer == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(20, 514, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_self->_buffer, 0, __pyx_v_data, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(20, 514, __pyx_L1_error)

        /* "uvloop/handles/stream.pyx":500
 *                     return True
 * 
 *                 if sent > 0:             # <<<<<<<<<<<<<<
 *                     if UVLOOP_DEBUG:
 *                         if sent == len(data):
 */
      }

      /* "uvloop/handles/stream.pyx":473
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size == 0:
 *             # libuv internal write buffers for this stream are empty.
 *             if buf_len == 1:             # <<<<<<<<<<<<<<
 *                 # If we only have one piece of data to send, let's
 *                 # use our fast implementation of try_write.
 */
      goto __pyx_L6;
    }

    /* "uvloop/handles/stream.pyx":520
 * 
 *             else:
 *                 ctx = _StreamWriteContext.new(self, self._buffer)             # <<<<<<<<<<<<<<
 * 
 *                 err = uv.uv_try_write(<uv.uv_stream_t*>self._handle,
 */
    /*else*/ {
      __pyx_t_1 = __pyx_v_self->_buffer;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = ((PyObject *)__pyx_f_6uvloop_4loop_19_StreamWriteContext_new(__pyx_v_self, ((PyObject*)__pyx_t_1))); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 520, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_ctx, ((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "uvloop/handles/stream.pyx":522
 *                 ctx = _StreamWriteContext.new(self, self._buffer)
 * 
 *                 err = uv.uv_try_write(<uv.uv_stream_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                       ctx.uv_bufs_start,
 *                                       ctx.uv_bufs_len)
 */
      __pyx_v_err = uv_try_write(((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_ctx->uv_bufs_start, __pyx_v_ctx->uv_bufs_len);

      /* "uvloop/handles/stream.pyx":526
 *                                       ctx.uv_bufs_len)
 * 
 *                 if err > 0:             # <<<<<<<<<<<<<<
 *                     # Some data was successfully sent.
 * 
 */
      __pyx_t_3 = (__pyx_v_err > 0);
      if (__pyx_t_3) {

        /* "uvloop/handles/stream.pyx":529
 *                     # Some data was successfully sent.
 * 
 *                     if <size_t>err == self._buffer_size:             # <<<<<<<<<<<<<<
 *                         # Everything was sent.
 *                         ctx.close()
 */
        __pyx_t_3 = (((size_t)__pyx_v_err) == __pyx_v_self->_buffer_size);
        if (__pyx_t_3) {

          /* "uvloop/handles/stream.pyx":531
 *                     if <size_t>err == self._buffer_size:
 *                         # Everything was sent.
 *                         ctx.close()             # <<<<<<<<<<<<<<
 *                         self._buffer.clear()
 *                         self._buffer_size = 0
 */
          __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_ctx->__pyx_vtab)->close(__pyx_v_ctx); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 531, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "uvloop/handles/stream.pyx":532
 *                         # Everything was sent.
 *                         ctx.close()
 *                         self._buffer.clear()             # <<<<<<<<<<<<<<
 *                         self._buffer_size = 0
 *                         # on_write will call "maybe_resume_protocol".
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_buffer, __pyx_n_s_clear); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 532, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 532, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "uvloop/handles/stream.pyx":533
 *                         ctx.close()
 *                         self._buffer.clear()
 *                         self._buffer_size = 0             # <<<<<<<<<<<<<<
 *                         # on_write will call "maybe_resume_protocol".
 *                         self._on_write()
 */
          __pyx_v_self->_buffer_size = 0;

          /* "uvloop/handles/stream.pyx":535
 *                         self._buffer_size = 0
 *                         # on_write will call "maybe_resume_protocol".
 *                         self._on_write()             # <<<<<<<<<<<<<<
 *                         return True
 * 
 */
          __pyx_t_4 = __pyx_f_6uvloop_4loop_8UVStream__on_write(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 535, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "uvloop/handles/stream.pyx":536
 *                         # on_write will call "maybe_resume_protocol".
 *                         self._on_write()
 *                         return True             # <<<<<<<<<<<<<<
 * 
 *                     try:
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(Py_True);
          __pyx_r = Py_True;
          goto __pyx_L0;

          /* "uvloop/handles/stream.pyx":529
 *                     # Some data was successfully sent.
 * 
 *                     if <size_t>err == self._buffer_size:             # <<<<<<<<<<<<<<
 *                         # Everything was sent.
 *                         ctx.close()
 */
        }

        /* "uvloop/handles/stream.pyx":538
 *                         return True
 * 
 *                     try:             # <<<<<<<<<<<<<<
 *                         # Advance pointers to uv_bufs in `ctx`,
 *                         # we will reuse it soon for a uv_write
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_10);
          /*try:*/ {

            /* "uvloop/handles/stream.pyx":542
 *                         # we will reuse it soon for a uv_write
 *                         # call.
 *                         ctx.advance_uv_buf(<ssize_t>err)             # <<<<<<<<<<<<<<
 *                     except Exception as ex:  # This should never happen.
 *                         # Let's try to close the `ctx` anyways.
 */
            __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_ctx->__pyx_vtab)->advance_uv_buf(__pyx_v_ctx, ((Py_ssize_t)__pyx_v_err)); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 542, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "uvloop/handles/stream.pyx":538
 *                         return True
 * 
 *                     try:             # <<<<<<<<<<<<<<
 *                         # Advance pointers to uv_bufs in `ctx`,
 *                         # we will reuse it soon for a uv_write
 */
          }
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          goto __pyx_L21_try_end;
          __pyx_L16_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/handles/stream.pyx":543
 *                         # call.
 *                         ctx.advance_uv_buf(<ssize_t>err)
 *                     except Exception as ex:  # This should never happen.             # <<<<<<<<<<<<<<
 *                         # Let's try to close the `ctx` anyways.
 *                         ctx.close()
 */
          __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
          if (__pyx_t_11) {
            __Pyx_AddTraceback("uvloop.loop.UVStream._exec_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_5) < 0) __PYX_ERR(20, 543, __pyx_L18_except_error)
            __Pyx_XGOTREF(__pyx_t_4);
            __Pyx_XGOTREF(__pyx_t_1);
            __Pyx_XGOTREF(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_1);
            __pyx_v_ex = __pyx_t_1;
            /*try:*/ {

              /* "uvloop/handles/stream.pyx":545
 *                     except Exception as ex:  # This should never happen.
 *                         # Let's try to close the `ctx` anyways.
 *                         ctx.close()             # <<<<<<<<<<<<<<
 *                         self._fatal_error(ex, True)
 *                         self._buffer.clear()
 */
              __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_ctx->__pyx_vtab)->close(__pyx_v_ctx); if (unlikely(!__pyx_t_12)) __PYX_ERR(20, 545, __pyx_L27_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

              /* "uvloop/handles/stream.pyx":546
 *                         # Let's try to close the `ctx` anyways.
 *                         ctx.close()
 *                         self._fatal_error(ex, True)             # <<<<<<<<<<<<<<
 *                         self._buffer.clear()
 *                         self._buffer_size = 0
 */
              __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_ex, Py_True, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(20, 546, __pyx_L27_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

              /* "uvloop/handles/stream.pyx":547
 *                         ctx.close()
 *                         self._fatal_error(ex, True)
 *                         self._buffer.clear()             # <<<<<<<<<<<<<<
 *                         self._buffer_size = 0
 *                         return
 */
              __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_buffer, __pyx_n_s_clear); if (unlikely(!__pyx_t_13)) __PYX_ERR(20, 547, __pyx_L27_error)
              __Pyx_GOTREF(__pyx_t_13);
              __pyx_t_14 = NULL;
              __pyx_t_6 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_13))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_13, function);
                  __pyx_t_6 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
                __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                if (unlikely(!__pyx_t_12)) __PYX_ERR(20, 547, __pyx_L27_error)
                __Pyx_GOTREF(__pyx_t_12);
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              }
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

              /* "uvloop/handles/stream.pyx":548
 *                         self._fatal_error(ex, True)
 *                         self._buffer.clear()
 *                         self._buffer_size = 0             # <<<<<<<<<<<<<<
 *                         return
 * 
 */
              __pyx_v_self->_buffer_size = 0;

              /* "uvloop/handles/stream.pyx":549
 *                         self._buffer.clear()
 *                         self._buffer_size = 0
 *                         return             # <<<<<<<<<<<<<<
 * 
 *                 elif err != uv.UV_EAGAIN:
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_r = Py_None; __Pyx_INCREF(Py_None);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              goto __pyx_L26_return;
            }

            /* "uvloop/handles/stream.pyx":543
 *                         # call.
 *                         ctx.advance_uv_buf(<ssize_t>err)
 *                     except Exception as ex:  # This should never happen.             # <<<<<<<<<<<<<<
 *                         # Let's try to close the `ctx` anyways.
 *                         ctx.close()
 */
            /*finally:*/ {
              __pyx_L27_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
                __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
                __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
                if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_17);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_21);
                __Pyx_XGOTREF(__pyx_t_22);
                __pyx_t_11 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
                {
                  __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                }
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_20);
                  __Pyx_XGIVEREF(__pyx_t_21);
                  __Pyx_XGIVEREF(__pyx_t_22);
                  __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
                }
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
                __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
                __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
                goto __pyx_L18_except_error;
              }
              __pyx_L26_return: {
                __pyx_t_22 = __pyx_r;
                __pyx_r = 0;
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
                __pyx_r = __pyx_t_22;
                __pyx_t_22 = 0;
                goto __pyx_L19_except_return;
              }
            }
          }
          goto __pyx_L18_except_error;

          /* "uvloop/handles/stream.pyx":538
 *                         return True
 * 
 *                     try:             # <<<<<<<<<<<<<<
 *                         # Advance pointers to uv_bufs in `ctx`,
 *                         # we will reuse it soon for a uv_write
 */
          __pyx_L18_except_error:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L1_error;
          __pyx_L19_except_return:;
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
          goto __pyx_L0;
          __pyx_L21_try_end:;
        }

        /* "uvloop/handles/stream.pyx":526
 *                                       ctx.uv_bufs_len)
 * 
 *                 if err > 0:             # <<<<<<<<<<<<<<
 *                     # Some data was successfully sent.
 * 
 */
        goto __pyx_L14;
      }

      /* "uvloop/handles/stream.pyx":551
 *                         return
 * 
 *                 elif err != uv.UV_EAGAIN:             # <<<<<<<<<<<<<<
 *                     ctx.close()
 *                     exc = convert_error(err)
 */
      __pyx_t_3 = (__pyx_v_err != UV_EAGAIN);
      if (__pyx_t_3) {

        /* "uvloop/handles/stream.pyx":552
 * 
 *                 elif err != uv.UV_EAGAIN:
 *                     ctx.close()             # <<<<<<<<<<<<<<
 *                     exc = convert_error(err)
 *                     self._fatal_error(exc, True)
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_ctx->__pyx_vtab)->close(__pyx_v_ctx); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 552, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/stream.pyx":553
 *                 elif err != uv.UV_EAGAIN:
 *                     ctx.close()
 *                     exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                     self._fatal_error(exc, True)
 *                     self._buffer.clear()
 */
        __pyx_t_5 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_v_exc = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "uvloop/handles/stream.pyx":554
 *                     ctx.close()
 *                     exc = convert_error(err)
 *                     self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                     self._buffer.clear()
 *                     self._buffer_size = 0
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 554, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/stream.pyx":555
 *                     exc = convert_error(err)
 *                     self._fatal_error(exc, True)
 *                     self._buffer.clear()             # <<<<<<<<<<<<<<
 *                     self._buffer_size = 0
 *                     return
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_buffer, __pyx_n_s_clear); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 555, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/stream.pyx":556
 *                     self._fatal_error(exc, True)
 *                     self._buffer.clear()
 *                     self._buffer_size = 0             # <<<<<<<<<<<<<<
 *                     return
 * 
 */
        __pyx_v_self->_buffer_size = 0;

        /* "uvloop/handles/stream.pyx":557
 *                     self._buffer.clear()
 *                     self._buffer_size = 0
 *                     return             # <<<<<<<<<<<<<<
 * 
 *                 # fall through
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L0;

        /* "uvloop/handles/stream.pyx":551
 *                         return
 * 
 *                 elif err != uv.UV_EAGAIN:             # <<<<<<<<<<<<<<
 *                     ctx.close()
 *                     exc = convert_error(err)
 */
      }
      __pyx_L14:;
    }
    __pyx_L6:;

    /* "uvloop/handles/stream.pyx":471
 *             return
 * 
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size == 0:             # <<<<<<<<<<<<<<
 *             # libuv internal write buffers for this stream are empty.
 *             if buf_len == 1:
 */
  }

  /* "uvloop/handles/stream.pyx":561
 *                 # fall through
 * 
 *         if ctx is None:             # <<<<<<<<<<<<<<
 *             ctx = _StreamWriteContext.new(self, self._buffer)
 * 
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_ctx) == Py_None);
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":562
 * 
 *         if ctx is None:
 *             ctx = _StreamWriteContext.new(self, self._buffer)             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_write(&ctx.req,
 */
    __pyx_t_5 = __pyx_v_self->_buffer;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_19_StreamWriteContext_new(__pyx_v_self, ((PyObject*)__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_ctx, ((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":561
 *                 # fall through
 * 
 *         if ctx is None:             # <<<<<<<<<<<<<<
 *             ctx = _StreamWriteContext.new(self, self._buffer)
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":564
 *             ctx = _StreamWriteContext.new(self, self._buffer)
 * 
 *         err = uv.uv_write(&ctx.req,             # <<<<<<<<<<<<<<
 *                           <uv.uv_stream_t*>self._handle,
 *                           ctx.uv_bufs_start,
 */
  __pyx_v_err = uv_write((&__pyx_v_ctx->req), ((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_ctx->uv_bufs_start, __pyx_v_ctx->uv_bufs_len, __pyx_f_6uvloop_4loop___uv_stream_on_write);

  /* "uvloop/handles/stream.pyx":570
 *                           __uv_stream_on_write)
 * 
 *         self._buffer_size = 0             # <<<<<<<<<<<<<<
 *         # Can't use `_buffer.clear()` here: `ctx` holds a reference to
 *         # the `_buffer`.
 */
  __pyx_v_self->_buffer_size = 0;

  /* "uvloop/handles/stream.pyx":573
 *         # Can't use `_buffer.clear()` here: `ctx` holds a reference to
 *         # the `_buffer`.
 *         self._buffer = []             # <<<<<<<<<<<<<<
 * 
 *         if err < 0:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_buffer);
  __Pyx_DECREF(__pyx_v_self->_buffer);
  __pyx_v_self->_buffer = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":575
 *         self._buffer = []
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             # close write context
 *             ctx.close()
 */
  __pyx_t_3 = (__pyx_v_err < 0);
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":577
 *         if err < 0:
 *             # close write context
 *             ctx.close()             # <<<<<<<<<<<<<<
 * 
 *             exc = convert_error(err)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_ctx->__pyx_vtab)->close(__pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":579
 *             ctx.close()
 * 
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":580
 * 
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":581
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._maybe_resume_protocol()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":575
 *         self._buffer = []
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             # close write context
 *             ctx.close()
 */
  }

  /* "uvloop/handles/stream.pyx":583
 *             return
 * 
 *         self._maybe_resume_protocol()             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t _get_write_buffer_size(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._maybe_resume_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":456
 *             self._loop._queue_write(self)
 * 
 *     cdef inline _exec_write(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("uvloop.loop.UVStream._exec_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_sent);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":585
 *         self._maybe_resume_protocol()
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             return 0
 */

static size_t __pyx_f_6uvloop_4loop_8UVStream__get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  size_t __pyx_r;
  int __pyx_t_1;

  /* "uvloop/handles/stream.pyx":586
 * 
 *     cdef size_t _get_write_buffer_size(self):
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return ((<uv.uv_stream_t*>self._handle).write_queue_size +
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":587
 *     cdef size_t _get_write_buffer_size(self):
 *         if self._handle is NULL:
 *             return 0             # <<<<<<<<<<<<<<
 *         return ((<uv.uv_stream_t*>self._handle).write_queue_size +
 *                 self._buffer_size)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":586
 * 
 *     cdef size_t _get_write_buffer_size(self):
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return ((<uv.uv_stream_t*>self._handle).write_queue_size +
 */
  }

  /* "uvloop/handles/stream.pyx":588
 *         if self._handle is NULL:
 *             return 0
 *         return ((<uv.uv_stream_t*>self._handle).write_queue_size +             # <<<<<<<<<<<<<<
 *                 self._buffer_size)
 * 
 */
  __pyx_r = (((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle)->write_queue_size + __pyx_v_self->_buffer_size);
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":585
 *         self._maybe_resume_protocol()
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":591
 *                 self._buffer_size)
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         try:
 *             if self._read_pybuf_acquired:
 */

static PyObject *__pyx_f_6uvloop_4loop_8UVStream__close(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/handles/stream.pyx":592
 * 
 *     cdef _close(self):
 *         try:             # <<<<<<<<<<<<<<
 *             if self._read_pybuf_acquired:
 *                 # Should never happen. libuv always calls uv_alloc/uv_read
 */
  /*try:*/ {

    /* "uvloop/handles/stream.pyx":593
 *     cdef _close(self):
 *         try:
 *             if self._read_pybuf_acquired:             # <<<<<<<<<<<<<<
 *                 # Should never happen. libuv always calls uv_alloc/uv_read
 *                 # in pairs.
 */
    if (__pyx_v_self->_read_pybuf_acquired) {

      /* "uvloop/handles/stream.pyx":596
 *                 # Should never happen. libuv always calls uv_alloc/uv_read
 *                 # in pairs.
 *                 self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                     'transport': self,
 *                     'message': 'XXX: an allocated buffer in transport._close()'
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 596, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "uvloop/handles/stream.pyx":597
 *                 # in pairs.
 *                 self._loop.call_exception_handler({
 *                     'transport': self,             # <<<<<<<<<<<<<<
 *                     'message': 'XXX: an allocated buffer in transport._close()'
 *                 })
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 597, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_transport, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(20, 597, __pyx_L4_error)
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_message, __pyx_kp_u_XXX_an_allocated_buffer_in_trans) < 0) __PYX_ERR(20, 597, __pyx_L4_error)
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 596, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/stream.pyx":600
 *                     'message': 'XXX: an allocated buffer in transport._close()'
 *                 })
 *                 self._read_pybuf_acquired = 0             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&self._read_pybuf)
 * 
 */
      __pyx_v_self->_read_pybuf_acquired = 0;

      /* "uvloop/handles/stream.pyx":601
 *                 })
 *                 self._read_pybuf_acquired = 0
 *                 PyBuffer_Release(&self._read_pybuf)             # <<<<<<<<<<<<<<
 * 
 *             self._stop_reading()
 */
      PyBuffer_Release((&__pyx_v_self->_read_pybuf));

      /* "uvloop/handles/stream.pyx":593
 *     cdef _close(self):
 *         try:
 *             if self._read_pybuf_acquired:             # <<<<<<<<<<<<<<
 *                 # Should never happen. libuv always calls uv_alloc/uv_read
 *                 # in pairs.
 */
    }

    /* "uvloop/handles/stream.pyx":603
 *                 PyBuffer_Release(&self._read_pybuf)
 * 
 *             self._stop_reading()             # <<<<<<<<<<<<<<
 *         finally:
 *             UVSocketHandle._close(<UVHandle>self)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._stop_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 603, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "uvloop/handles/stream.pyx":605
 *             self._stop_reading()
 *         finally:
 *             UVSocketHandle._close(<UVHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _on_accept(self):
 */
  /*finally:*/ {
    /*normal exit:*/{
      if (!(likely(((((PyObject *)__pyx_v_self)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6uvloop_4loop_UVSocketHandle))))) __PYX_ERR(20, 605, __pyx_L1_error)
      __pyx_t_1 = __pyx_f_6uvloop_4loop_14UVSocketHandle__close(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_8 = __pyx_filename;
      {
        if (!(likely(((((PyObject *)__pyx_v_self)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6uvloop_4loop_UVSocketHandle))))) __PYX_ERR(20, 605, __pyx_L8_error)
        __pyx_t_1 = __pyx_f_6uvloop_4loop_14UVSocketHandle__close(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 605, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_8;
      goto __pyx_L1_error;
      __pyx_L8_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "uvloop/handles/stream.pyx":591
 *                 self._buffer_size)
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         try:
 *             if self._read_pybuf_acquired:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVStream._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":607
 *             UVSocketHandle._close(<UVHandle>self)
 * 
 *     cdef inline _on_accept(self):             # <<<<<<<<<<<<<<
 *         # Ultimately called by __uv_stream_on_listen.
 *         self._init_protocol()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_accept(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_accept", 1);

  /* "uvloop/handles/stream.pyx":609
 *     cdef inline _on_accept(self):
 *         # Ultimately called by __uv_stream_on_listen.
 *         self._init_protocol()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _on_eof(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":607
 *             UVSocketHandle._close(<UVHandle>self)
 * 
 *     cdef inline _on_accept(self):             # <<<<<<<<<<<<<<
 *         # Ultimately called by __uv_stream_on_listen.
 *         self._init_protocol()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream._on_accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":611
 *         self._init_protocol()
 * 
 *     cdef inline _on_eof(self):             # <<<<<<<<<<<<<<
 *         # Any exception raised here will be caught in
 *         # __uv_stream_on_read.
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_eof(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_v_meth = NULL;
  PyObject *__pyx_v_keep_open = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_eof", 1);

  /* "uvloop/handles/stream.pyx":615
 *         # __uv_stream_on_read.
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             meth = self._protocol.eof_received
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/handles/stream.pyx":616
 * 
 *         try:
 *             meth = self._protocol.eof_received             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             keep_open = False
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._protocol, __pyx_n_s_eof_received); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 616, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_meth = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/handles/stream.pyx":615
 *         # __uv_stream_on_read.
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             meth = self._protocol.eof_received
 *         except AttributeError:
 */
    }

    /* "uvloop/handles/stream.pyx":620
 *             keep_open = False
 *         else:
 *             keep_open = run_in_context(self.context, meth)             # <<<<<<<<<<<<<<
 * 
 *         if keep_open:
 */
    /*else:*/ {
      __pyx_t_4 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = __pyx_f_6uvloop_4loop_run_in_context(__pyx_t_4, __pyx_v_meth); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 620, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_keep_open = __pyx_t_5;
      __pyx_t_5 = 0;
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/stream.pyx":617
 *         try:
 *             meth = self._protocol.eof_received
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             keep_open = False
 *         else:
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("uvloop.loop.UVStream._on_eof", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(20, 617, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "uvloop/handles/stream.pyx":618
 *             meth = self._protocol.eof_received
 *         except AttributeError:
 *             keep_open = False             # <<<<<<<<<<<<<<
 *         else:
 *             keep_open = run_in_context(self.context, meth)
 */
      __Pyx_INCREF(Py_False);
      __pyx_v_keep_open = Py_False;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/handles/stream.pyx":615
 *         # __uv_stream_on_read.
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             meth = self._protocol.eof_received
 *         except AttributeError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/handles/stream.pyx":622
 *             keep_open = run_in_context(self.context, meth)
 * 
 *         if keep_open:             # <<<<<<<<<<<<<<
 *             # We're keeping the connection open so the
 *             # protocol can write more, but we still can't
 */
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_keep_open); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(20, 622, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "uvloop/handles/stream.pyx":626
 *             # protocol can write more, but we still can't
 *             # receive more, so remove the reader callback.
 *             self._stop_reading()             # <<<<<<<<<<<<<<
 *         else:
 *             self.close()
 */
    __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._stop_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/handles/stream.pyx":622
 *             keep_open = run_in_context(self.context, meth)
 * 
 *         if keep_open:             # <<<<<<<<<<<<<<
 *             # We're keeping the connection open so the
 *             # protocol can write more, but we still can't
 */
    goto __pyx_L11;
  }

  /* "uvloop/handles/stream.pyx":628
 *             self._stop_reading()
 *         else:
 *             self.close()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _on_write(self):
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 628, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L11:;

  /* "uvloop/handles/stream.pyx":611
 *         self._init_protocol()
 * 
 *     cdef inline _on_eof(self):             # <<<<<<<<<<<<<<
 *         # Any exception raised here will be caught in
 *         # __uv_stream_on_read.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UVStream._on_eof", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_meth);
  __Pyx_XDECREF(__pyx_v_keep_open);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":630
 *             self.close()
 * 
 *     cdef inline _on_write(self):             # <<<<<<<<<<<<<<
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_write", 1);

  /* "uvloop/handles/stream.pyx":631
 * 
 *     cdef inline _on_write(self):
 *         self._maybe_resume_protocol()             # <<<<<<<<<<<<<<
 *         if not self._get_write_buffer_size():
 *             if self._closing:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._maybe_resume_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":632
 *     cdef inline _on_write(self):
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_write_buffer_size(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 632, __pyx_L1_error)
  __pyx_t_3 = (!(__pyx_t_2 != 0));
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":633
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():
 *             if self._closing:             # <<<<<<<<<<<<<<
 *                 self._schedule_call_connection_lost(None)
 *             elif self._eof:
 */
    if (__pyx_v_self->__pyx_base._closing) {

      /* "uvloop/handles/stream.pyx":634
 *         if not self._get_write_buffer_size():
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)             # <<<<<<<<<<<<<<
 *             elif self._eof:
 *                 self._shutdown()
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._schedule_call_connection_lost(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 634, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/stream.pyx":633
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():
 *             if self._closing:             # <<<<<<<<<<<<<<
 *                 self._schedule_call_connection_lost(None)
 *             elif self._eof:
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/stream.pyx":635
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)
 *             elif self._eof:             # <<<<<<<<<<<<<<
 *                 self._shutdown()
 * 
 */
    if (__pyx_v_self->_eof) {

      /* "uvloop/handles/stream.pyx":636
 *                 self._schedule_call_connection_lost(None)
 *             elif self._eof:
 *                 self._shutdown()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _init(self, Loop loop, object protocol, Server server,
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__shutdown(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/stream.pyx":635
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)
 *             elif self._eof:             # <<<<<<<<<<<<<<
 *                 self._shutdown()
 * 
 */
    }
    __pyx_L4:;

    /* "uvloop/handles/stream.pyx":632
 *     cdef inline _on_write(self):
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)
 */
  }

  /* "uvloop/handles/stream.pyx":630
 *             self.close()
 * 
 *     cdef inline _on_write(self):             # <<<<<<<<<<<<<<
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream._on_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":638
 *                 self._shutdown()
 * 
 *     cdef inline _init(self, Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                       object waiter, object context):
 *         self.context = context
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__init(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/stream.pyx":640
 *     cdef inline _init(self, Loop loop, object protocol, Server server,
 *                       object waiter, object context):
 *         self.context = context             # <<<<<<<<<<<<<<
 *         self._set_protocol(protocol)
 *         self._start_init(loop)
 */
  __Pyx_INCREF(__pyx_v_context);
  __Pyx_GIVEREF(__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context);
  __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context = __pyx_v_context;

  /* "uvloop/handles/stream.pyx":641
 *                       object waiter, object context):
 *         self.context = context
 *         self._set_protocol(protocol)             # <<<<<<<<<<<<<<
 *         self._start_init(loop)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._set_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":642
 *         self.context = context
 *         self._set_protocol(protocol)
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         if server is not None:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":644
 *         self._start_init(loop)
 * 
 *         if server is not None:             # <<<<<<<<<<<<<<
 *             self._set_server(server)
 * 
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_server) != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":645
 * 
 *         if server is not None:
 *             self._set_server(server)             # <<<<<<<<<<<<<<
 * 
 *         if waiter is not None:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._set_server(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_server); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":644
 *         self._start_init(loop)
 * 
 *         if server is not None:             # <<<<<<<<<<<<<<
 *             self._set_server(server)
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":647
 *             self._set_server(server)
 * 
 *         if waiter is not None:             # <<<<<<<<<<<<<<
 *             self._set_waiter(waiter)
 * 
 */
  __pyx_t_2 = (__pyx_v_waiter != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":648
 * 
 *         if waiter is not None:
 *             self._set_waiter(waiter)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _on_connect(self, object exc):
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._set_waiter(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), __pyx_v_waiter); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":647
 *             self._set_server(server)
 * 
 *         if waiter is not None:             # <<<<<<<<<<<<<<
 *             self._set_waiter(waiter)
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":638
 *                 self._shutdown()
 * 
 *     cdef inline _init(self, Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                       object waiter, object context):
 *         self.context = context
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":650
 *             self._set_waiter(waiter)
 * 
 *     cdef inline _on_connect(self, object exc):             # <<<<<<<<<<<<<<
 *         # Called from __tcp_connect_callback (tcp.pyx) and
 *         # __pipe_connect_callback (pipe.pyx).
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_8UVStream__on_connect(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_connect", 1);

  /* "uvloop/handles/stream.pyx":653
 *         # Called from __tcp_connect_callback (tcp.pyx) and
 *         # __pipe_connect_callback (pipe.pyx).
 *         if exc is None:             # <<<<<<<<<<<<<<
 *             self._init_protocol()
 *         else:
 */
  __pyx_t_1 = (__pyx_v_exc == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":654
 *         # __pipe_connect_callback (pipe.pyx).
 *         if exc is None:
 *             self._init_protocol()             # <<<<<<<<<<<<<<
 *         else:
 *             if self._waiter is None:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/stream.pyx":653
 *         # Called from __tcp_connect_callback (tcp.pyx) and
 *         # __pipe_connect_callback (pipe.pyx).
 *         if exc is None:             # <<<<<<<<<<<<<<
 *             self._init_protocol()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/stream.pyx":656
 *             self._init_protocol()
 *         else:
 *             if self._waiter is None:             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, False, "connect failed")
 *             elif self._waiter.cancelled():
 */
  /*else*/ {
    __pyx_t_1 = (__pyx_v_self->__pyx_base._waiter == Py_None);
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":657
 *         else:
 *             if self._waiter is None:
 *                 self._fatal_error(exc, False, "connect failed")             # <<<<<<<<<<<<<<
 *             elif self._waiter.cancelled():
 *                 # Connect call was cancelled; just close the transport
 */
      __pyx_t_3.__pyx_n = 1;
      __pyx_t_3.reason = __pyx_kp_u_connect_failed;
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_False, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 657, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":656
 *             self._init_protocol()
 *         else:
 *             if self._waiter is None:             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, False, "connect failed")
 *             elif self._waiter.cancelled():
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/stream.pyx":658
 *             if self._waiter is None:
 *                 self._fatal_error(exc, False, "connect failed")
 *             elif self._waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 # Connect call was cancelled; just close the transport
 *                 # silently.
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._waiter, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 658, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(20, 658, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":661
 *                 # Connect call was cancelled; just close the transport
 *                 # silently.
 *                 self._close()             # <<<<<<<<<<<<<<
 *             elif self._waiter.done():
 *                 self._fatal_error(exc, False, "connect failed")
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 661, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":658
 *             if self._waiter is None:
 *                 self._fatal_error(exc, False, "connect failed")
 *             elif self._waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 # Connect call was cancelled; just close the transport
 *                 # silently.
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/stream.pyx":662
 *                 # silently.
 *                 self._close()
 *             elif self._waiter.done():             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, False, "connect failed")
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._waiter, __pyx_n_s_done); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(20, 662, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":663
 *                 self._close()
 *             elif self._waiter.done():
 *                 self._fatal_error(exc, False, "connect failed")             # <<<<<<<<<<<<<<
 *             else:
 *                 self._waiter.set_exception(exc)
 */
      __pyx_t_3.__pyx_n = 1;
      __pyx_t_3.reason = __pyx_kp_u_connect_failed;
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_False, &__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 663, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":662
 *                 # silently.
 *                 self._close()
 *             elif self._waiter.done():             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, False, "connect failed")
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/stream.pyx":665
 *                 self._fatal_error(exc, False, "connect failed")
 *             else:
 *                 self._waiter.set_exception(exc)             # <<<<<<<<<<<<<<
 *                 self._close()
 * 
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 665, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_exc};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 665, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":666
 *             else:
 *                 self._waiter.set_exception(exc)
 *                 self._close()             # <<<<<<<<<<<<<<
 * 
 *     # === Public API ===
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 666, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L4:;
  }
  __pyx_L3:;

  /* "uvloop/handles/stream.pyx":650
 *             self._set_waiter(waiter)
 * 
 *     cdef inline _on_connect(self, object exc):             # <<<<<<<<<<<<<<
 *         # Called from __tcp_connect_callback (tcp.pyx) and
 *         # __pipe_connect_callback (pipe.pyx).
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVStream._on_connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":670
 *     # === Public API ===
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} closed={} reading={} {:#x}>'.format(
 *             self.__class__.__name__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_2__repr__(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_2__repr__(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/handles/stream.pyx":671
 * 
 *     def __repr__(self):
 *         return '<{} closed={} reading={} {:#x}>'.format(             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self._closed,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_closed_reading_x, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/handles/stream.pyx":672
 *     def __repr__(self):
 *         return '<{} closed={} reading={} {:#x}>'.format(
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self._closed,
 *             self.__reading,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/stream.pyx":673
 *         return '<{} closed={} reading={} {:#x}>'.format(
 *             self.__class__.__name__,
 *             self._closed,             # <<<<<<<<<<<<<<
 *             self.__reading,
 *             id(self))
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._closed); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/handles/stream.pyx":674
 *             self.__class__.__name__,
 *             self._closed,
 *             self.__reading,             # <<<<<<<<<<<<<<
 *             id(self))
 * 
 */
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->_UVStream__reading); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "uvloop/handles/stream.pyx":675
 *             self._closed,
 *             self.__reading,
 *             id(self))             # <<<<<<<<<<<<<<
 * 
 *     def write(self, object buf):
 */
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_6)) __PYX_ERR(20, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_t_4, __pyx_t_3, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 4+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":670
 *     # === Public API ===
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<{} closed={} reading={} {:#x}>'.format(
 *             self.__class__.__name__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UVStream.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":677
 *             id(self))
 * 
 *     def write(self, object buf):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_5write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_4write, "UVStream.write(self, buf)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_5write = {"write", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_5write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_4write};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_5write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_buf = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_buf,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_buf)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 677, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "write") < 0)) __PYX_ERR(20, 677, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_buf = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write", 1, 1, 1, __pyx_nargs); __PYX_ERR(20, 677, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVStream.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_4write(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_buf);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_4write(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_buf) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 1);

  /* "uvloop/handles/stream.pyx":678
 * 
 *     def write(self, object buf):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self._eof:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":680
 *         self._ensure_alive()
 * 
 *         if self._eof:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot call write() after write_eof()')
 *         if not buf:
 */
  if (unlikely(__pyx_v_self->_eof)) {

    /* "uvloop/handles/stream.pyx":681
 * 
 *         if self._eof:
 *             raise RuntimeError('Cannot call write() after write_eof()')             # <<<<<<<<<<<<<<
 *         if not buf:
 *             return
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__155, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(20, 681, __pyx_L1_error)

    /* "uvloop/handles/stream.pyx":680
 *         self._ensure_alive()
 * 
 *         if self._eof:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot call write() after write_eof()')
 *         if not buf:
 */
  }

  /* "uvloop/handles/stream.pyx":682
 *         if self._eof:
 *             raise RuntimeError('Cannot call write() after write_eof()')
 *         if not buf:             # <<<<<<<<<<<<<<
 *             return
 *         if self._conn_lost:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_buf); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(20, 682, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":683
 *             raise RuntimeError('Cannot call write() after write_eof()')
 *         if not buf:
 *             return             # <<<<<<<<<<<<<<
 *         if self._conn_lost:
 *             self._conn_lost += 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":682
 *         if self._eof:
 *             raise RuntimeError('Cannot call write() after write_eof()')
 *         if not buf:             # <<<<<<<<<<<<<<
 *             return
 *         if self._conn_lost:
 */
  }

  /* "uvloop/handles/stream.pyx":684
 *         if not buf:
 *             return
 *         if self._conn_lost:             # <<<<<<<<<<<<<<
 *             self._conn_lost += 1
 *             return
 */
  __pyx_t_3 = (__pyx_v_self->__pyx_base._conn_lost != 0);
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":685
 *             return
 *         if self._conn_lost:
 *             self._conn_lost += 1             # <<<<<<<<<<<<<<
 *             return
 *         self._buffer_write(buf)
 */
    __pyx_v_self->__pyx_base._conn_lost = (__pyx_v_self->__pyx_base._conn_lost + 1);

    /* "uvloop/handles/stream.pyx":686
 *         if self._conn_lost:
 *             self._conn_lost += 1
 *             return             # <<<<<<<<<<<<<<
 *         self._buffer_write(buf)
 *         self._initiate_write()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":684
 *         if not buf:
 *             return
 *         if self._conn_lost:             # <<<<<<<<<<<<<<
 *             self._conn_lost += 1
 *             return
 */
  }

  /* "uvloop/handles/stream.pyx":687
 *             self._conn_lost += 1
 *             return
 *         self._buffer_write(buf)             # <<<<<<<<<<<<<<
 *         self._initiate_write()
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__buffer_write(__pyx_v_self, __pyx_v_buf); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":688
 *             return
 *         self._buffer_write(buf)
 *         self._initiate_write()             # <<<<<<<<<<<<<<
 * 
 *     def writelines(self, bufs):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__initiate_write(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":677
 *             id(self))
 * 
 *     def write(self, object buf):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":690
 *         self._initiate_write()
 * 
 *     def writelines(self, bufs):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_7writelines(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_6writelines, "UVStream.writelines(self, bufs)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_7writelines = {"writelines", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_7writelines, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_6writelines};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_7writelines(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_bufs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writelines (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bufs,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_bufs)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 690, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "writelines") < 0)) __PYX_ERR(20, 690, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_bufs = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writelines", 1, 1, 1, __pyx_nargs); __PYX_ERR(20, 690, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVStream.writelines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_6writelines(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_bufs);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_6writelines(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, PyObject *__pyx_v_bufs) {
  PyObject *__pyx_v_buf = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writelines", 1);

  /* "uvloop/handles/stream.pyx":691
 * 
 *     def writelines(self, bufs):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self._eof:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":693
 *         self._ensure_alive()
 * 
 *         if self._eof:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot call writelines() after write_eof()')
 *         if self._conn_lost:
 */
  if (unlikely(__pyx_v_self->_eof)) {

    /* "uvloop/handles/stream.pyx":694
 * 
 *         if self._eof:
 *             raise RuntimeError('Cannot call writelines() after write_eof()')             # <<<<<<<<<<<<<<
 *         if self._conn_lost:
 *             self._conn_lost += 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__156, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(20, 694, __pyx_L1_error)

    /* "uvloop/handles/stream.pyx":693
 *         self._ensure_alive()
 * 
 *         if self._eof:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('Cannot call writelines() after write_eof()')
 *         if self._conn_lost:
 */
  }

  /* "uvloop/handles/stream.pyx":695
 *         if self._eof:
 *             raise RuntimeError('Cannot call writelines() after write_eof()')
 *         if self._conn_lost:             # <<<<<<<<<<<<<<
 *             self._conn_lost += 1
 *             return
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._conn_lost != 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":696
 *             raise RuntimeError('Cannot call writelines() after write_eof()')
 *         if self._conn_lost:
 *             self._conn_lost += 1             # <<<<<<<<<<<<<<
 *             return
 *         for buf in bufs:
 */
    __pyx_v_self->__pyx_base._conn_lost = (__pyx_v_self->__pyx_base._conn_lost + 1);

    /* "uvloop/handles/stream.pyx":697
 *         if self._conn_lost:
 *             self._conn_lost += 1
 *             return             # <<<<<<<<<<<<<<
 *         for buf in bufs:
 *             self._buffer_write(buf)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":695
 *         if self._eof:
 *             raise RuntimeError('Cannot call writelines() after write_eof()')
 *         if self._conn_lost:             # <<<<<<<<<<<<<<
 *             self._conn_lost += 1
 *             return
 */
  }

  /* "uvloop/handles/stream.pyx":698
 *             self._conn_lost += 1
 *             return
 *         for buf in bufs:             # <<<<<<<<<<<<<<
 *             self._buffer_write(buf)
 *         self._initiate_write()
 */
  if (likely(PyList_CheckExact(__pyx_v_bufs)) || PyTuple_CheckExact(__pyx_v_bufs)) {
    __pyx_t_1 = __pyx_v_bufs; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_bufs); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 698, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 698, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(20, 698, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(20, 698, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(20, 698, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 698, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(20, 698, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_buf, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "uvloop/handles/stream.pyx":699
 *             return
 *         for buf in bufs:
 *             self._buffer_write(buf)             # <<<<<<<<<<<<<<
 *         self._initiate_write()
 * 
 */
    __pyx_t_5 = __pyx_f_6uvloop_4loop_8UVStream__buffer_write(__pyx_v_self, __pyx_v_buf); if (unlikely(!__pyx_t_5)) __PYX_ERR(20, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/handles/stream.pyx":698
 *             self._conn_lost += 1
 *             return
 *         for buf in bufs:             # <<<<<<<<<<<<<<
 *             self._buffer_write(buf)
 *         self._initiate_write()
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":700
 *         for buf in bufs:
 *             self._buffer_write(buf)
 *         self._initiate_write()             # <<<<<<<<<<<<<<
 * 
 *     def write_eof(self):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__initiate_write(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":690
 *         self._initiate_write()
 * 
 *     def writelines(self, bufs):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVStream.writelines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":702
 *         self._initiate_write()
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_9write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_8write_eof, "UVStream.write_eof(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_9write_eof = {"write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_9write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_8write_eof};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_9write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_eof (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("write_eof", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "write_eof", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_8write_eof(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_8write_eof(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_eof", 1);

  /* "uvloop/handles/stream.pyx":703
 * 
 *     def write_eof(self):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self._eof:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":705
 *         self._ensure_alive()
 * 
 *         if self._eof:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_self->_eof) {

    /* "uvloop/handles/stream.pyx":706
 * 
 *         if self._eof:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._eof = 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":705
 *         self._ensure_alive()
 * 
 *         if self._eof:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":708
 *             return
 * 
 *         self._eof = 1             # <<<<<<<<<<<<<<
 *         if not self._get_write_buffer_size():
 *             self._shutdown()
 */
  __pyx_v_self->_eof = 1;

  /* "uvloop/handles/stream.pyx":709
 * 
 *         self._eof = 1
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             self._shutdown()
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_write_buffer_size(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 709, __pyx_L1_error)
  __pyx_t_3 = (!(__pyx_t_2 != 0));
  if (__pyx_t_3) {

    /* "uvloop/handles/stream.pyx":710
 *         self._eof = 1
 *         if not self._get_write_buffer_size():
 *             self._shutdown()             # <<<<<<<<<<<<<<
 * 
 *     def can_write_eof(self):
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__shutdown(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":709
 * 
 *         self._eof = 1
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             self._shutdown()
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":702
 *         self._initiate_write()
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStream.write_eof", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":712
 *             self._shutdown()
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_11can_write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_10can_write_eof, "UVStream.can_write_eof(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_11can_write_eof = {"can_write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_11can_write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_10can_write_eof};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_11can_write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_write_eof (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("can_write_eof", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "can_write_eof", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_10can_write_eof(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_10can_write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_write_eof", 1);

  /* "uvloop/handles/stream.pyx":713
 * 
 *     def can_write_eof(self):
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def is_reading(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":712
 *             self._shutdown()
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":715
 *         return True
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return self._is_reading()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_13is_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_12is_reading, "UVStream.is_reading(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_13is_reading = {"is_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_13is_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_12is_reading};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_13is_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_12is_reading(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_12is_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_reading", 1);

  /* "uvloop/handles/stream.pyx":716
 * 
 *     def is_reading(self):
 *         return self._is_reading()             # <<<<<<<<<<<<<<
 * 
 *     def pause_reading(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._is_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 716, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":715
 *         return True
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return self._is_reading()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVStream.is_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":718
 *         return self._is_reading()
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         if self._closing or not self._is_reading():
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_15pause_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_14pause_reading, "UVStream.pause_reading(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_15pause_reading = {"pause_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_15pause_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_14pause_reading};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_15pause_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pause_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("pause_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pause_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_14pause_reading(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_14pause_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pause_reading", 1);

  /* "uvloop/handles/stream.pyx":719
 * 
 *     def pause_reading(self):
 *         if self._closing or not self._is_reading():             # <<<<<<<<<<<<<<
 *             return
 *         self._stop_reading()
 */
  if (!__pyx_v_self->__pyx_base._closing) {
  } else {
    __pyx_t_1 = __pyx_v_self->__pyx_base._closing;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._is_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 719, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":720
 *     def pause_reading(self):
 *         if self._closing or not self._is_reading():
 *             return             # <<<<<<<<<<<<<<
 *         self._stop_reading()
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":719
 * 
 *     def pause_reading(self):
 *         if self._closing or not self._is_reading():             # <<<<<<<<<<<<<<
 *             return
 *         self._stop_reading()
 */
  }

  /* "uvloop/handles/stream.pyx":721
 *         if self._closing or not self._is_reading():
 *             return
 *         self._stop_reading()             # <<<<<<<<<<<<<<
 * 
 *     def resume_reading(self):
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._stop_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/handles/stream.pyx":718
 *         return self._is_reading()
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         if self._closing or not self._is_reading():
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVStream.pause_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":723
 *         self._stop_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         if self._is_reading() or self._closing:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_17resume_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_16resume_reading, "UVStream.resume_reading(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_17resume_reading = {"resume_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_17resume_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_16resume_reading};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_17resume_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resume_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("resume_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "resume_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_16resume_reading(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_16resume_reading(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resume_reading", 1);

  /* "uvloop/handles/stream.pyx":724
 * 
 *     def resume_reading(self):
 *         if self._is_reading() or self._closing:             # <<<<<<<<<<<<<<
 *             return
 *         self._start_reading()
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._is_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 724, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->__pyx_base._closing;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":725
 *     def resume_reading(self):
 *         if self._is_reading() or self._closing:
 *             return             # <<<<<<<<<<<<<<
 *         self._start_reading()
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":724
 * 
 *     def resume_reading(self):
 *         if self._is_reading() or self._closing:             # <<<<<<<<<<<<<<
 *             return
 *         self._start_reading()
 */
  }

  /* "uvloop/handles/stream.pyx":726
 *         if self._is_reading() or self._closing:
 *             return
 *         self._start_reading()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._start_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/stream.pyx":723
 *         self._stop_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         if self._is_reading() or self._closing:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVStream.resume_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_18__reduce_cython__, "UVStream.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_19__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_18__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_18__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVStream.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8UVStream_20__setstate_cython__, "UVStream.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8UVStream_21__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_20__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8UVStream_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVStream.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_8UVStream_20__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8UVStream_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVStream.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":729
 * 
 * 
 * cdef void __uv_stream_on_shutdown(uv.uv_shutdown_t* req,             # <<<<<<<<<<<<<<
 *                                   int status) noexcept with gil:
 * 
 */

static void __pyx_f_6uvloop_4loop___uv_stream_on_shutdown(uv_shutdown_t *__pyx_v_req, int __pyx_v_status) {
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_stream = 0;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_8 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_stream_on_shutdown", 0);

  /* "uvloop/handles/stream.pyx":734
 *     # callback for uv_shutdown
 * 
 *     if req.data is NULL:             # <<<<<<<<<<<<<<
 *         aio_logger.error(
 *             'UVStream.shutdown callback called with NULL req.data, status=%r',
 */
  __pyx_t_1 = (__pyx_v_req->data == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":735
 * 
 *     if req.data is NULL:
 *         aio_logger.error(             # <<<<<<<<<<<<<<
 *             'UVStream.shutdown callback called with NULL req.data, status=%r',
 *             status)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/stream.pyx":737
 *         aio_logger.error(
 *             'UVStream.shutdown callback called with NULL req.data, status=%r',
 *             status)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_UVStream_shutdown_callback_calle, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 735, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/stream.pyx":738
 *             'UVStream.shutdown callback called with NULL req.data, status=%r',
 *             status)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef UVStream stream = <UVStream> req.data
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":734
 *     # callback for uv_shutdown
 * 
 *     if req.data is NULL:             # <<<<<<<<<<<<<<
 *         aio_logger.error(
 *             'UVStream.shutdown callback called with NULL req.data, status=%r',
 */
  }

  /* "uvloop/handles/stream.pyx":740
 *         return
 * 
 *     cdef UVStream stream = <UVStream> req.data             # <<<<<<<<<<<<<<
 * 
 *     if status < 0 and status != uv.UV_ECANCELED:
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_req->data);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_stream = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/handles/stream.pyx":742
 *     cdef UVStream stream = <UVStream> req.data
 * 
 *     if status < 0 and status != uv.UV_ECANCELED:             # <<<<<<<<<<<<<<
 *         # From libuv source code:
 *         #     The ECANCELED error code is a lie, the shutdown(2) syscall is a
 */
  __pyx_t_7 = (__pyx_v_status < 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_1 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_status != UV_ECANCELED);
  __pyx_t_1 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":749
 *         #     informs the callee that the handle has been destroyed.
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_shutdown_errors_total += 1
 * 
 */
    __pyx_t_1 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":750
 * 
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_shutdown_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *         exc = convert_error(status)
 */
      __Pyx_INCREF((PyObject *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop);
      __pyx_t_8 = __pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop;
      __pyx_t_8->_debug_stream_shutdown_errors_total = (__pyx_t_8->_debug_stream_shutdown_errors_total + 1);
      __Pyx_DECREF((PyObject *)__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/handles/stream.pyx":749
 *         #     informs the callee that the handle has been destroyed.
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_shutdown_errors_total += 1
 * 
 */
    }

    /* "uvloop/handles/stream.pyx":752
 *             stream._loop._debug_stream_shutdown_errors_total += 1
 * 
 *         exc = convert_error(status)             # <<<<<<<<<<<<<<
 *         stream._fatal_error(
 *             exc, False, "error status in uv_stream_t.shutdown callback")
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/stream.pyx":753
 * 
 *         exc = convert_error(status)
 *         stream._fatal_error(             # <<<<<<<<<<<<<<
 *             exc, False, "error status in uv_stream_t.shutdown callback")
 *         return
 */
    __pyx_t_9.__pyx_n = 1;
    __pyx_t_9.reason = __pyx_kp_u_error_status_in_uv_stream_t_shut;
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_stream), __pyx_v_exc, Py_False, &__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/stream.pyx":755
 *         stream._fatal_error(
 *             exc, False, "error status in uv_stream_t.shutdown callback")
 *         return             # <<<<<<<<<<<<<<
 * 
 * 
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":742
 *     cdef UVStream stream = <UVStream> req.data
 * 
 *     if status < 0 and status != uv.UV_ECANCELED:             # <<<<<<<<<<<<<<
 *         # From libuv source code:
 *         #     The ECANCELED error code is a lie, the shutdown(2) syscall is a
 */
  }

  /* "uvloop/handles/stream.pyx":729
 * 
 * 
 * cdef void __uv_stream_on_shutdown(uv.uv_shutdown_t* req,             # <<<<<<<<<<<<<<
 *                                   int status) noexcept with gil:
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF((PyObject *)__pyx_t_8);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_stream_on_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/stream.pyx":758
 * 
 * 
 * cdef inline bint __uv_stream_on_read_common(             # <<<<<<<<<<<<<<
 *     UVStream sc,
 *     Loop loop,
 */

static CYTHON_INLINE int __pyx_f_6uvloop_4loop___uv_stream_on_read_common(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_sc, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, Py_ssize_t __pyx_v_nread) {
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_exc = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__uv_stream_on_read_common", 1);

  /* "uvloop/handles/stream.pyx":763
 *     ssize_t nread,
 * ):
 *     if sc._closed:             # <<<<<<<<<<<<<<
 *         # The stream was closed, there is no reason to
 *         # do any work now.
 */
  if (__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base._closed) {

    /* "uvloop/handles/stream.pyx":766
 *         # The stream was closed, there is no reason to
 *         # do any work now.
 *         sc.__reading_stopped()  # Just in case.             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_stopped(__pyx_v_sc); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":767
 *         # do any work now.
 *         sc.__reading_stopped()  # Just in case.
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     if nread == uv.UV_EOF:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":763
 *     ssize_t nread,
 * ):
 *     if sc._closed:             # <<<<<<<<<<<<<<
 *         # The stream was closed, there is no reason to
 *         # do any work now.
 */
  }

  /* "uvloop/handles/stream.pyx":769
 *         return True
 * 
 *     if nread == uv.UV_EOF:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #     The callee is responsible for stopping closing the stream
 */
  __pyx_t_2 = (__pyx_v_nread == UV_EOF);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":774
 *         #     when an error happens by calling uv_read_stop() or uv_close().
 *         #     Trying to read from the stream again is undefined.
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 loop._debug_stream_read_eof_total += 1
 */
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_5);
        /*try:*/ {

          /* "uvloop/handles/stream.pyx":775
 *         #     Trying to read from the stream again is undefined.
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 loop._debug_stream_read_eof_total += 1
 * 
 */
          __pyx_t_2 = (UVLOOP_DEBUG != 0);
          if (__pyx_t_2) {

            /* "uvloop/handles/stream.pyx":776
 *         try:
 *             if UVLOOP_DEBUG:
 *                 loop._debug_stream_read_eof_total += 1             # <<<<<<<<<<<<<<
 * 
 *             sc._stop_reading()
 */
            __pyx_v_loop->_debug_stream_read_eof_total = (__pyx_v_loop->_debug_stream_read_eof_total + 1);

            /* "uvloop/handles/stream.pyx":775
 *         #     Trying to read from the stream again is undefined.
 *         try:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 loop._debug_stream_read_eof_total += 1
 * 
 */
          }

          /* "uvloop/handles/stream.pyx":778
 *                 loop._debug_stream_read_eof_total += 1
 * 
 *             sc._stop_reading()             # <<<<<<<<<<<<<<
 *             sc._on_eof()
 *         except BaseException as ex:
 */
          __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._stop_reading(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_sc)); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 778, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "uvloop/handles/stream.pyx":779
 * 
 *             sc._stop_reading()
 *             sc._on_eof()             # <<<<<<<<<<<<<<
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:
 */
          __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__on_eof(__pyx_v_sc); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 779, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "uvloop/handles/stream.pyx":774
 *         #     when an error happens by calling uv_read_stop() or uv_close().
 *         #     Trying to read from the stream again is undefined.
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 loop._debug_stream_read_eof_total += 1
 */
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L13_try_end;
        __pyx_L8_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/stream.pyx":780
 *             sc._stop_reading()
 *             sc._on_eof()
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 loop._debug_stream_read_eof_cb_errors_total += 1
 */
        __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
        if (__pyx_t_6) {
          __Pyx_AddTraceback("uvloop.loop.__uv_stream_on_read_common", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(20, 780, __pyx_L10_except_error)
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __pyx_v_ex = __pyx_t_7;
          /*try:*/ {

            /* "uvloop/handles/stream.pyx":781
 *             sc._on_eof()
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 loop._debug_stream_read_eof_cb_errors_total += 1
 * 
 */
            __pyx_t_2 = (UVLOOP_DEBUG != 0);
            if (__pyx_t_2) {

              /* "uvloop/handles/stream.pyx":782
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:
 *                 loop._debug_stream_read_eof_cb_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *             sc._fatal_error(ex, False)
 */
              __pyx_v_loop->_debug_stream_read_eof_cb_errors_total = (__pyx_v_loop->_debug_stream_read_eof_cb_errors_total + 1);

              /* "uvloop/handles/stream.pyx":781
 *             sc._on_eof()
 *         except BaseException as ex:
 *             if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *                 loop._debug_stream_read_eof_cb_errors_total += 1
 * 
 */
            }

            /* "uvloop/handles/stream.pyx":784
 *                 loop._debug_stream_read_eof_cb_errors_total += 1
 * 
 *             sc._fatal_error(ex, False)             # <<<<<<<<<<<<<<
 *         finally:
 *             return True
 */
            __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_sc), __pyx_v_ex, Py_False, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(20, 784, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          }

          /* "uvloop/handles/stream.pyx":780
 *             sc._stop_reading()
 *             sc._on_eof()
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 loop._debug_stream_read_eof_cb_errors_total += 1
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              goto __pyx_L21;
            }
            __pyx_L20_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_13);
              __Pyx_XGOTREF(__pyx_t_14);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_17);
              __pyx_t_6 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_15);
                __Pyx_XGIVEREF(__pyx_t_16);
                __Pyx_XGIVEREF(__pyx_t_17);
                __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
              }
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_XGIVEREF(__pyx_t_14);
              __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
              __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
              __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
              goto __pyx_L10_except_error;
            }
            __pyx_L21:;
          }
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L9_exception_handled;
        }
        goto __pyx_L10_except_error;

        /* "uvloop/handles/stream.pyx":774
 *         #     when an error happens by calling uv_read_stop() or uv_close().
 *         #     Trying to read from the stream again is undefined.
 *         try:             # <<<<<<<<<<<<<<
 *             if UVLOOP_DEBUG:
 *                 loop._debug_stream_read_eof_total += 1
 */
        __pyx_L10_except_error:;
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
        goto __pyx_L6_error;
        __pyx_L9_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
        __pyx_L13_try_end:;
      }
    }

    /* "uvloop/handles/stream.pyx":786
 *             sc._fatal_error(ex, False)
 *         finally:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     if nread == 0:
 */
    /*finally:*/ {
      /*normal exit:*/{
        __pyx_r = 1;
        goto __pyx_L0;
      }
      __pyx_L6_error:;
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_17);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_15);
        {
          __pyx_r = 1;
          goto __pyx_L27_return;
        }
        __pyx_L27_return:;
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
        goto __pyx_L0;
      }
    }

    /* "uvloop/handles/stream.pyx":769
 *         return True
 * 
 *     if nread == uv.UV_EOF:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #     The callee is responsible for stopping closing the stream
 */
  }

  /* "uvloop/handles/stream.pyx":788
 *             return True
 * 
 *     if nread == 0:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #     nread might be 0, which does not indicate an error or EOF.
 */
  __pyx_t_2 = (__pyx_v_nread == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":792
 *         #     nread might be 0, which does not indicate an error or EOF.
 *         #     This is equivalent to EAGAIN or EWOULDBLOCK under read(2).
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     if nread < 0:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":788
 *             return True
 * 
 *     if nread == 0:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #     nread might be 0, which does not indicate an error or EOF.
 */
  }

  /* "uvloop/handles/stream.pyx":794
 *         return True
 * 
 *     if nread < 0:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #     The callee is responsible for stopping closing the stream
 */
  __pyx_t_2 = (__pyx_v_nread < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":804
 *         # we don't need try...finally here.
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_errors_total += 1
 * 
 */
    __pyx_t_2 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/stream.pyx":805
 * 
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *         if sc.__read_error_close:
 */
      __pyx_v_loop->_debug_stream_read_errors_total = (__pyx_v_loop->_debug_stream_read_errors_total + 1);

      /* "uvloop/handles/stream.pyx":804
 *         # we don't need try...finally here.
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_errors_total += 1
 * 
 */
    }

    /* "uvloop/handles/stream.pyx":807
 *             loop._debug_stream_read_errors_total += 1
 * 
 *         if sc.__read_error_close:             # <<<<<<<<<<<<<<
 *             # Used for getting notified when a pipe is closed.
 *             # See WriteUnixTransport for the explanation.
 */
    if (__pyx_v_sc->_UVStream__read_error_close) {

      /* "uvloop/handles/stream.pyx":810
 *             # Used for getting notified when a pipe is closed.
 *             # See WriteUnixTransport for the explanation.
 *             sc._on_eof()             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
      __pyx_t_8 = __pyx_f_6uvloop_4loop_8UVStream__on_eof(__pyx_v_sc); if (unlikely(!__pyx_t_8)) __PYX_ERR(20, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/handles/stream.pyx":811
 *             # See WriteUnixTransport for the explanation.
 *             sc._on_eof()
 *             return True             # <<<<<<<<<<<<<<
 * 
 *         exc = convert_error(nread)
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "uvloop/handles/stream.pyx":807
 *             loop._debug_stream_read_errors_total += 1
 * 
 *         if sc.__read_error_close:             # <<<<<<<<<<<<<<
 *             # Used for getting notified when a pipe is closed.
 *             # See WriteUnixTransport for the explanation.
 */
    }

    /* "uvloop/handles/stream.pyx":813
 *             return True
 * 
 *         exc = convert_error(nread)             # <<<<<<<<<<<<<<
 *         sc._fatal_error(
 *             exc, False, "error status in uv_stream_t.read callback")
 */
    __pyx_t_8 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_nread); if (unlikely(!__pyx_t_8)) __PYX_ERR(20, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_v_exc = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "uvloop/handles/stream.pyx":814
 * 
 *         exc = convert_error(nread)
 *         sc._fatal_error(             # <<<<<<<<<<<<<<
 *             exc, False, "error status in uv_stream_t.read callback")
 *         return True
 */
    __pyx_t_18.__pyx_n = 1;
    __pyx_t_18.reason = __pyx_kp_u_error_status_in_uv_stream_t_read;
    __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_sc), __pyx_v_exc, Py_False, &__pyx_t_18); if (unlikely(!__pyx_t_8)) __PYX_ERR(20, 814, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/handles/stream.pyx":816
 *         sc._fatal_error(
 *             exc, False, "error status in uv_stream_t.read callback")
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     return False
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":794
 *         return True
 * 
 *     if nread < 0:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #     The callee is responsible for stopping closing the stream
 */
  }

  /* "uvloop/handles/stream.pyx":818
 *         return True
 * 
 *     return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "uvloop/handles/stream.pyx":758
 * 
 * 
 * cdef inline bint __uv_stream_on_read_common(             # <<<<<<<<<<<<<<
 *     UVStream sc,
 *     Loop loop,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.__uv_stream_on_read_common", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/stream.pyx":821
 * 
 * 
 * cdef inline void __uv_stream_on_read_impl(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* stream,
 *     ssize_t nread,
 */

static CYTHON_INLINE void __pyx_f_6uvloop_4loop___uv_stream_on_read_impl(uv_stream_t *__pyx_v_stream, Py_ssize_t __pyx_v_nread, CYTHON_UNUSED uv_buf_t const *__pyx_v_buf) {
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_sc = 0;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__uv_stream_on_read_impl", 1);

  /* "uvloop/handles/stream.pyx":827
 * ):
 *     cdef:
 *         UVStream sc = <UVStream>stream.data             # <<<<<<<<<<<<<<
 *         Loop loop = sc._loop
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_stream->data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_sc = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":828
 *     cdef:
 *         UVStream sc = <UVStream>stream.data
 *         Loop loop = sc._loop             # <<<<<<<<<<<<<<
 * 
 *     # It's OK to free the buffer early, since nothing will
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":832
 *     # It's OK to free the buffer early, since nothing will
 *     # be able to touch it until this method is done.
 *     __loop_free_buffer(loop)             # <<<<<<<<<<<<<<
 * 
 *     if __uv_stream_on_read_common(sc, loop, nread):
 */
  __pyx_f_6uvloop_4loop___loop_free_buffer(__pyx_v_loop); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 832, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":834
 *     __loop_free_buffer(loop)
 * 
 *     if __uv_stream_on_read_common(sc, loop, nread):             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop___uv_stream_on_read_common(__pyx_v_sc, __pyx_v_loop, __pyx_v_nread); if (unlikely(__pyx_t_2 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(20, 834, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":835
 * 
 *     if __uv_stream_on_read_common(sc, loop, nread):
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":834
 *     __loop_free_buffer(loop)
 * 
 *     if __uv_stream_on_read_common(sc, loop, nread):             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/handles/stream.pyx":837
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_total += 1
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "uvloop/handles/stream.pyx":838
 * 
 *     try:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_total += 1
 * 
 */
      __pyx_t_2 = (UVLOOP_DEBUG != 0);
      if (__pyx_t_2) {

        /* "uvloop/handles/stream.pyx":839
 *     try:
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_total += 1             # <<<<<<<<<<<<<<
 * 
 *         run_in_context1(
 */
        __pyx_v_loop->_debug_stream_read_cb_total = (__pyx_v_loop->_debug_stream_read_cb_total + 1);

        /* "uvloop/handles/stream.pyx":838
 * 
 *     try:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_total += 1
 * 
 */
      }

      /* "uvloop/handles/stream.pyx":842
 * 
 *         run_in_context1(
 *             sc.context,             # <<<<<<<<<<<<<<
 *             sc._protocol_data_received,
 *             loop._recv_buffer[:nread],
 */
      __pyx_t_1 = __pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.context;
      __Pyx_INCREF(__pyx_t_1);

      /* "uvloop/handles/stream.pyx":843
 *         run_in_context1(
 *             sc.context,
 *             sc._protocol_data_received,             # <<<<<<<<<<<<<<
 *             loop._recv_buffer[:nread],
 *         )
 */
      __pyx_t_6 = __pyx_v_sc->__pyx_base._protocol_data_received;
      __Pyx_INCREF(__pyx_t_6);

      /* "uvloop/handles/stream.pyx":844
 *             sc.context,
 *             sc._protocol_data_received,
 *             loop._recv_buffer[:nread],             # <<<<<<<<<<<<<<
 *         )
 *     except BaseException as exc:
 */
      __pyx_t_7 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_loop->_recv_buffer) + 0, __pyx_v_nread - 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(20, 844, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "uvloop/handles/stream.pyx":841
 *             loop._debug_stream_read_cb_total += 1
 * 
 *         run_in_context1(             # <<<<<<<<<<<<<<
 *             sc.context,
 *             sc._protocol_data_received,
 */
      __pyx_t_8 = __pyx_f_6uvloop_4loop_run_in_context1(__pyx_t_1, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(20, 841, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/handles/stream.pyx":837
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_total += 1
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/handles/stream.pyx":846
 *             loop._recv_buffer[:nread],
 *         )
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_errors_total += 1
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.__uv_stream_on_read_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(20, 846, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_exc = __pyx_t_7;
      /*try:*/ {

        /* "uvloop/handles/stream.pyx":847
 *         )
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_errors_total += 1
 * 
 */
        __pyx_t_2 = (UVLOOP_DEBUG != 0);
        if (__pyx_t_2) {

          /* "uvloop/handles/stream.pyx":848
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *         sc._fatal_error(exc, False)
 */
          __pyx_v_loop->_debug_stream_read_cb_errors_total = (__pyx_v_loop->_debug_stream_read_cb_errors_total + 1);

          /* "uvloop/handles/stream.pyx":847
 *         )
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_errors_total += 1
 * 
 */
        }

        /* "uvloop/handles/stream.pyx":850
 *             loop._debug_stream_read_cb_errors_total += 1
 * 
 *         sc._fatal_error(exc, False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_sc), __pyx_v_exc, Py_False, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 850, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "uvloop/handles/stream.pyx":846
 *             loop._recv_buffer[:nread],
 *         )
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_errors_total += 1
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L17;
        }
        __pyx_L16_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __pyx_t_9 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          }
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
          goto __pyx_L6_except_error;
        }
        __pyx_L17:;
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/handles/stream.pyx":837
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_total += 1
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L9_try_end:;
  }

  /* "uvloop/handles/stream.pyx":821
 * 
 * 
 * cdef inline void __uv_stream_on_read_impl(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* stream,
 *     ssize_t nread,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.__uv_stream_on_read_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sc);
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/handles/stream.pyx":853
 * 
 * 
 * cdef inline void __uv_stream_on_write_impl(             # <<<<<<<<<<<<<<
 *     uv.uv_write_t* req,
 *     int status,
 */

static CYTHON_INLINE void __pyx_f_6uvloop_4loop___uv_stream_on_write_impl(uv_write_t *__pyx_v_req, int __pyx_v_status) {
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_v_ctx = 0;
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_stream = 0;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_3 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__uv_stream_on_write_impl", 1);

  /* "uvloop/handles/stream.pyx":858
 * ):
 *     cdef:
 *         _StreamWriteContext ctx = <_StreamWriteContext> req.data             # <<<<<<<<<<<<<<
 *         UVStream stream = <UVStream>ctx.stream
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_req->data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_ctx = ((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":859
 *     cdef:
 *         _StreamWriteContext ctx = <_StreamWriteContext> req.data
 *         UVStream stream = <UVStream>ctx.stream             # <<<<<<<<<<<<<<
 * 
 *     ctx.close()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_ctx->stream);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_stream = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":861
 *         UVStream stream = <UVStream>ctx.stream
 * 
 *     ctx.close()             # <<<<<<<<<<<<<<
 * 
 *     if stream._closed:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext *)__pyx_v_ctx->__pyx_vtab)->close(__pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/stream.pyx":863
 *     ctx.close()
 * 
 *     if stream._closed:             # <<<<<<<<<<<<<<
 *         # The stream was closed, there is nothing to do.
 *         # Even if there is an error, like EPIPE, there
 */
  if (__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._closed) {

    /* "uvloop/handles/stream.pyx":867
 *         # Even if there is an error, like EPIPE, there
 *         # is no reason to report it.
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if status < 0:
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":863
 *     ctx.close()
 * 
 *     if stream._closed:             # <<<<<<<<<<<<<<
 *         # The stream was closed, there is nothing to do.
 *         # Even if there is an error, like EPIPE, there
 */
  }

  /* "uvloop/handles/stream.pyx":869
 *         return
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_errors_total += 1
 */
  __pyx_t_2 = (__pyx_v_status < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":870
 * 
 *     if status < 0:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_write_errors_total += 1
 * 
 */
    __pyx_t_2 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/stream.pyx":871
 *     if status < 0:
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *         exc = convert_error(status)
 */
      __Pyx_INCREF((PyObject *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop);
      __pyx_t_3 = __pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop;
      __pyx_t_3->_debug_stream_write_errors_total = (__pyx_t_3->_debug_stream_write_errors_total + 1);
      __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/stream.pyx":870
 * 
 *     if status < 0:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_write_errors_total += 1
 * 
 */
    }

    /* "uvloop/handles/stream.pyx":873
 *             stream._loop._debug_stream_write_errors_total += 1
 * 
 *         exc = convert_error(status)             # <<<<<<<<<<<<<<
 *         stream._fatal_error(
 *             exc, False, "error status in uv_stream_t.write callback")
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":874
 * 
 *         exc = convert_error(status)
 *         stream._fatal_error(             # <<<<<<<<<<<<<<
 *             exc, False, "error status in uv_stream_t.write callback")
 *         return
 */
    __pyx_t_4.__pyx_n = 1;
    __pyx_t_4.reason = __pyx_kp_u_error_status_in_uv_stream_t_writ;
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_stream), __pyx_v_exc, Py_False, &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/stream.pyx":876
 *         stream._fatal_error(
 *             exc, False, "error status in uv_stream_t.write callback")
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":869
 *         return
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_errors_total += 1
 */
  }

  /* "uvloop/handles/stream.pyx":878
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         stream._on_write()
 *     except BaseException as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "uvloop/handles/stream.pyx":879
 * 
 *     try:
 *         stream._on_write()             # <<<<<<<<<<<<<<
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVStream__on_write(__pyx_v_stream); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 879, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/stream.pyx":878
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         stream._on_write()
 *     except BaseException as exc:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/stream.pyx":880
 *     try:
 *         stream._on_write()
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_cb_errors_total += 1
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.__uv_stream_on_write_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_9, &__pyx_t_10) < 0) __PYX_ERR(20, 880, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      __pyx_v_exc = __pyx_t_9;
      /*try:*/ {

        /* "uvloop/handles/stream.pyx":881
 *         stream._on_write()
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_write_cb_errors_total += 1
 * 
 */
        __pyx_t_2 = (UVLOOP_DEBUG != 0);
        if (__pyx_t_2) {

          /* "uvloop/handles/stream.pyx":882
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_cb_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *         stream._fatal_error(exc, False)
 */
          __Pyx_INCREF((PyObject *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop);
          __pyx_t_3 = __pyx_v_stream->__pyx_base.__pyx_base.__pyx_base._loop;
          __pyx_t_3->_debug_stream_write_cb_errors_total = (__pyx_t_3->_debug_stream_write_cb_errors_total + 1);
          __Pyx_DECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;

          /* "uvloop/handles/stream.pyx":881
 *         stream._on_write()
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_write_cb_errors_total += 1
 * 
 */
        }

        /* "uvloop/handles/stream.pyx":884
 *             stream._loop._debug_stream_write_cb_errors_total += 1
 * 
 *         stream._fatal_error(exc, False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_stream), __pyx_v_exc, Py_False, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(20, 884, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }

      /* "uvloop/handles/stream.pyx":880
 *     try:
 *         stream._on_write()
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_write_cb_errors_total += 1
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L18;
        }
        __pyx_L17_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF((PyObject *)__pyx_t_3); __pyx_t_3 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L8_except_error;
        }
        __pyx_L18:;
      }
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L7_exception_handled;
    }
    goto __pyx_L8_except_error;

    /* "uvloop/handles/stream.pyx":878
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         stream._on_write()
 *     except BaseException as exc:
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L7_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    __pyx_L11_try_end:;
  }

  /* "uvloop/handles/stream.pyx":853
 * 
 * 
 * cdef inline void __uv_stream_on_write_impl(             # <<<<<<<<<<<<<<
 *     uv.uv_write_t* req,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF((PyObject *)__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("uvloop.loop.__uv_stream_on_write_impl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/handles/stream.pyx":887
 * 
 * 
 * cdef void __uv_stream_on_read(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* stream,
 *     ssize_t nread,
 */

static void __pyx_f_6uvloop_4loop___uv_stream_on_read(uv_stream_t *__pyx_v_stream, Py_ssize_t __pyx_v_nread, uv_buf_t const *__pyx_v_buf) {
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif

  /* "uvloop/handles/stream.pyx":893
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream read callback") == 0:
 *         return
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_stream), ((char const *)"UVStream read callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(20, 893, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":894
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,
 *                             "UVStream read callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_t_1 == 0);

  /* "uvloop/handles/stream.pyx":893
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream read callback") == 0:
 *         return
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":895
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,
 *                             "UVStream read callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     # Don't need try-finally, __uv_stream_on_read_impl is void
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":893
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream read callback") == 0:
 *         return
 */
  }

  /* "uvloop/handles/stream.pyx":898
 * 
 *     # Don't need try-finally, __uv_stream_on_read_impl is void
 *     __uv_stream_on_read_impl(stream, nread, buf)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_6uvloop_4loop___uv_stream_on_read_impl(__pyx_v_stream, __pyx_v_nread, __pyx_v_buf); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 898, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":887
 * 
 * 
 * cdef void __uv_stream_on_read(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* stream,
 *     ssize_t nread,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("uvloop.loop.__uv_stream_on_read", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/stream.pyx":901
 * 
 * 
 * cdef void __uv_stream_on_write(             # <<<<<<<<<<<<<<
 *     uv.uv_write_t* req,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___uv_stream_on_write(uv_write_t *__pyx_v_req, int __pyx_v_status) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_stream_on_write", 0);

  /* "uvloop/handles/stream.pyx":906
 * ) noexcept with gil:
 * 
 *     if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *         if req.data is NULL:
 *             aio_logger.error(
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/stream.pyx":907
 * 
 *     if UVLOOP_DEBUG:
 *         if req.data is NULL:             # <<<<<<<<<<<<<<
 *             aio_logger.error(
 *                 'UVStream.write callback called with NULL req.data, status=%r',
 */
    __pyx_t_1 = (__pyx_v_req->data == NULL);
    if (__pyx_t_1) {

      /* "uvloop/handles/stream.pyx":908
 *     if UVLOOP_DEBUG:
 *         if req.data is NULL:
 *             aio_logger.error(             # <<<<<<<<<<<<<<
 *                 'UVStream.write callback called with NULL req.data, status=%r',
 *                 status)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/handles/stream.pyx":910
 *             aio_logger.error(
 *                 'UVStream.write callback called with NULL req.data, status=%r',
 *                 status)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_UVStream_write_callback_called_w, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(20, 908, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/stream.pyx":911
 *                 'UVStream.write callback called with NULL req.data, status=%r',
 *                 status)
 *             return             # <<<<<<<<<<<<<<
 * 
 *     # Don't need try-finally, __uv_stream_on_write_impl is void
 */
      goto __pyx_L0;

      /* "uvloop/handles/stream.pyx":907
 * 
 *     if UVLOOP_DEBUG:
 *         if req.data is NULL:             # <<<<<<<<<<<<<<
 *             aio_logger.error(
 *                 'UVStream.write callback called with NULL req.data, status=%r',
 */
    }

    /* "uvloop/handles/stream.pyx":906
 * ) noexcept with gil:
 * 
 *     if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *         if req.data is NULL:
 *             aio_logger.error(
 */
  }

  /* "uvloop/handles/stream.pyx":914
 * 
 *     # Don't need try-finally, __uv_stream_on_write_impl is void
 *     __uv_stream_on_write_impl(req, status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_6uvloop_4loop___uv_stream_on_write_impl(__pyx_v_req, __pyx_v_status); if (unlikely(PyErr_Occurred())) __PYX_ERR(20, 914, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":901
 * 
 * 
 * cdef void __uv_stream_on_write(             # <<<<<<<<<<<<<<
 *     uv.uv_write_t* req,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_stream_on_write", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/stream.pyx":917
 * 
 * 
 * cdef void __uv_stream_buffered_alloc(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* stream,
 *     size_t suggested_size,
 */

static void __pyx_f_6uvloop_4loop___uv_stream_buffered_alloc(uv_handle_t *__pyx_v_stream, size_t __pyx_v_suggested_size, uv_buf_t *__pyx_v_uvbuf) {
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_sc = 0;
  CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  Py_buffer *__pyx_v_pybuf;
  int __pyx_v_got_buf;
  PyObject *__pyx_v_buf = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_stream_buffered_alloc", 0);

  /* "uvloop/handles/stream.pyx":923
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream alloc buffer callback") == 0:
 *         return
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_stream), ((char const *)"UVStream alloc buffer callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(20, 923, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":924
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,
 *                             "UVStream alloc buffer callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_t_1 == 0);

  /* "uvloop/handles/stream.pyx":923
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream alloc buffer callback") == 0:
 *         return
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":925
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,
 *                             "UVStream alloc buffer callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":923
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream alloc buffer callback") == 0:
 *         return
 */
  }

  /* "uvloop/handles/stream.pyx":928
 * 
 *     cdef:
 *         UVStream sc = <UVStream>stream.data             # <<<<<<<<<<<<<<
 *         Loop loop = sc._loop
 *         Py_buffer* pybuf = &sc._read_pybuf
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_stream->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_sc = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/stream.pyx":929
 *     cdef:
 *         UVStream sc = <UVStream>stream.data
 *         Loop loop = sc._loop             # <<<<<<<<<<<<<<
 *         Py_buffer* pybuf = &sc._read_pybuf
 *         int got_buf = 0
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/stream.pyx":930
 *         UVStream sc = <UVStream>stream.data
 *         Loop loop = sc._loop
 *         Py_buffer* pybuf = &sc._read_pybuf             # <<<<<<<<<<<<<<
 *         int got_buf = 0
 * 
 */
  __pyx_v_pybuf = (&__pyx_v_sc->_read_pybuf);

  /* "uvloop/handles/stream.pyx":931
 *         Loop loop = sc._loop
 *         Py_buffer* pybuf = &sc._read_pybuf
 *         int got_buf = 0             # <<<<<<<<<<<<<<
 * 
 *     if sc._read_pybuf_acquired:
 */
  __pyx_v_got_buf = 0;

  /* "uvloop/handles/stream.pyx":933
 *         int got_buf = 0
 * 
 *     if sc._read_pybuf_acquired:             # <<<<<<<<<<<<<<
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 */
  if (__pyx_v_sc->_read_pybuf_acquired) {

    /* "uvloop/handles/stream.pyx":934
 * 
 *     if sc._read_pybuf_acquired:
 *         uvbuf.len = 0             # <<<<<<<<<<<<<<
 *         uvbuf.base = NULL
 *         return
 */
    __pyx_v_uvbuf->len = 0;

    /* "uvloop/handles/stream.pyx":935
 *     if sc._read_pybuf_acquired:
 *         uvbuf.len = 0
 *         uvbuf.base = NULL             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_v_uvbuf->base = NULL;

    /* "uvloop/handles/stream.pyx":936
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 *         return             # <<<<<<<<<<<<<<
 * 
 *     sc._read_pybuf_acquired = 0
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":933
 *         int got_buf = 0
 * 
 *     if sc._read_pybuf_acquired:             # <<<<<<<<<<<<<<
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 */
  }

  /* "uvloop/handles/stream.pyx":938
 *         return
 * 
 *     sc._read_pybuf_acquired = 0             # <<<<<<<<<<<<<<
 *     try:
 *         buf = run_in_context1(
 */
  __pyx_v_sc->_read_pybuf_acquired = 0;

  /* "uvloop/handles/stream.pyx":939
 * 
 *     sc._read_pybuf_acquired = 0
 *     try:             # <<<<<<<<<<<<<<
 *         buf = run_in_context1(
 *             sc.context,
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "uvloop/handles/stream.pyx":941
 *     try:
 *         buf = run_in_context1(
 *             sc.context,             # <<<<<<<<<<<<<<
 *             sc._protocol_get_buffer,
 *             suggested_size,
 */
      __pyx_t_3 = __pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.context;
      __Pyx_INCREF(__pyx_t_3);

      /* "uvloop/handles/stream.pyx":942
 *         buf = run_in_context1(
 *             sc.context,
 *             sc._protocol_get_buffer,             # <<<<<<<<<<<<<<
 *             suggested_size,
 *         )
 */
      __pyx_t_7 = __pyx_v_sc->_protocol_get_buffer;
      __Pyx_INCREF(__pyx_t_7);

      /* "uvloop/handles/stream.pyx":943
 *             sc.context,
 *             sc._protocol_get_buffer,
 *             suggested_size,             # <<<<<<<<<<<<<<
 *         )
 *         PyObject_GetBuffer(buf, pybuf, PyBUF_WRITABLE)
 */
      __pyx_t_8 = __Pyx_PyInt_FromSize_t(__pyx_v_suggested_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(20, 943, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);

      /* "uvloop/handles/stream.pyx":940
 *     sc._read_pybuf_acquired = 0
 *     try:
 *         buf = run_in_context1(             # <<<<<<<<<<<<<<
 *             sc.context,
 *             sc._protocol_get_buffer,
 */
      __pyx_t_9 = __pyx_f_6uvloop_4loop_run_in_context1(__pyx_t_3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(20, 940, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_buf = __pyx_t_9;
      __pyx_t_9 = 0;

      /* "uvloop/handles/stream.pyx":945
 *             suggested_size,
 *         )
 *         PyObject_GetBuffer(buf, pybuf, PyBUF_WRITABLE)             # <<<<<<<<<<<<<<
 *         got_buf = 1
 *     except BaseException as exc:
 */
      __pyx_t_10 = PyObject_GetBuffer(__pyx_v_buf, __pyx_v_pybuf, PyBUF_WRITABLE); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(20, 945, __pyx_L5_error)

      /* "uvloop/handles/stream.pyx":946
 *         )
 *         PyObject_GetBuffer(buf, pybuf, PyBUF_WRITABLE)
 *         got_buf = 1             # <<<<<<<<<<<<<<
 *     except BaseException as exc:
 *         # Can't call 'sc._fatal_error' or 'sc._close', libuv will SF.
 */
      __pyx_v_got_buf = 1;

      /* "uvloop/handles/stream.pyx":939
 * 
 *     sc._read_pybuf_acquired = 0
 *     try:             # <<<<<<<<<<<<<<
 *         buf = run_in_context1(
 *             sc.context,
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "uvloop/handles/stream.pyx":947
 *         PyObject_GetBuffer(buf, pybuf, PyBUF_WRITABLE)
 *         got_buf = 1
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         # Can't call 'sc._fatal_error' or 'sc._close', libuv will SF.
 *         # We'll do it later in __uv_stream_buffered_on_read when we
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("uvloop.loop.__uv_stream_buffered_alloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(20, 947, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_exc = __pyx_t_8;
      /*try:*/ {

        /* "uvloop/handles/stream.pyx":951
 *         # We'll do it later in __uv_stream_buffered_on_read when we
 *         # receive UV_ENOBUFS.
 *         uvbuf.len = 0             # <<<<<<<<<<<<<<
 *         uvbuf.base = NULL
 *         return
 */
        __pyx_v_uvbuf->len = 0;

        /* "uvloop/handles/stream.pyx":952
 *         # receive UV_ENOBUFS.
 *         uvbuf.len = 0
 *         uvbuf.base = NULL             # <<<<<<<<<<<<<<
 *         return
 * 
 */
        __pyx_v_uvbuf->base = NULL;

        /* "uvloop/handles/stream.pyx":953
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if not pybuf.len:
 */
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L15_return;
      }

      /* "uvloop/handles/stream.pyx":947
 *         PyObject_GetBuffer(buf, pybuf, PyBUF_WRITABLE)
 *         got_buf = 1
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         # Can't call 'sc._fatal_error' or 'sc._close', libuv will SF.
 *         # We'll do it later in __uv_stream_buffered_on_read when we
 */
      /*finally:*/ {
        __pyx_L15_return: {
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L8_except_return;
        }
      }
    }
    goto __pyx_L7_except_error;

    /* "uvloop/handles/stream.pyx":939
 * 
 *     sc._read_pybuf_acquired = 0
 *     try:             # <<<<<<<<<<<<<<
 *         buf = run_in_context1(
 *             sc.context,
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L8_except_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
    __pyx_L10_try_end:;
  }

  /* "uvloop/handles/stream.pyx":955
 *         return
 * 
 *     if not pybuf.len:             # <<<<<<<<<<<<<<
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 */
  __pyx_t_2 = (!(__pyx_v_pybuf->len != 0));
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":956
 * 
 *     if not pybuf.len:
 *         uvbuf.len = 0             # <<<<<<<<<<<<<<
 *         uvbuf.base = NULL
 *         if got_buf:
 */
    __pyx_v_uvbuf->len = 0;

    /* "uvloop/handles/stream.pyx":957
 *     if not pybuf.len:
 *         uvbuf.len = 0
 *         uvbuf.base = NULL             # <<<<<<<<<<<<<<
 *         if got_buf:
 *             PyBuffer_Release(pybuf)
 */
    __pyx_v_uvbuf->base = NULL;

    /* "uvloop/handles/stream.pyx":958
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 *         if got_buf:             # <<<<<<<<<<<<<<
 *             PyBuffer_Release(pybuf)
 *         return
 */
    __pyx_t_2 = (__pyx_v_got_buf != 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/stream.pyx":959
 *         uvbuf.base = NULL
 *         if got_buf:
 *             PyBuffer_Release(pybuf)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
      PyBuffer_Release(__pyx_v_pybuf);

      /* "uvloop/handles/stream.pyx":958
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 *         if got_buf:             # <<<<<<<<<<<<<<
 *             PyBuffer_Release(pybuf)
 *         return
 */
    }

    /* "uvloop/handles/stream.pyx":960
 *         if got_buf:
 *             PyBuffer_Release(pybuf)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     sc._read_pybuf_acquired = 1
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":955
 *         return
 * 
 *     if not pybuf.len:             # <<<<<<<<<<<<<<
 *         uvbuf.len = 0
 *         uvbuf.base = NULL
 */
  }

  /* "uvloop/handles/stream.pyx":962
 *         return
 * 
 *     sc._read_pybuf_acquired = 1             # <<<<<<<<<<<<<<
 *     uvbuf.base = <char*>pybuf.buf
 *     uvbuf.len = pybuf.len
 */
  __pyx_v_sc->_read_pybuf_acquired = 1;

  /* "uvloop/handles/stream.pyx":963
 * 
 *     sc._read_pybuf_acquired = 1
 *     uvbuf.base = <char*>pybuf.buf             # <<<<<<<<<<<<<<
 *     uvbuf.len = pybuf.len
 * 
 */
  __pyx_v_uvbuf->base = ((char *)__pyx_v_pybuf->buf);

  /* "uvloop/handles/stream.pyx":964
 *     sc._read_pybuf_acquired = 1
 *     uvbuf.base = <char*>pybuf.buf
 *     uvbuf.len = pybuf.len             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_11 = __pyx_v_pybuf->len;
  __pyx_v_uvbuf->len = __pyx_t_11;

  /* "uvloop/handles/stream.pyx":917
 * 
 * 
 * cdef void __uv_stream_buffered_alloc(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* stream,
 *     size_t suggested_size,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_stream_buffered_alloc", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sc);
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/stream.pyx":967
 * 
 * 
 * cdef void __uv_stream_buffered_on_read(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* stream,
 *     ssize_t nread,
 */

static void __pyx_f_6uvloop_4loop___uv_stream_buffered_on_read(uv_stream_t *__pyx_v_stream, Py_ssize_t __pyx_v_nread, CYTHON_UNUSED uv_buf_t const *__pyx_v_buf) {
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_sc = 0;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  Py_buffer *__pyx_v_pybuf;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  char const *__pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_stream_buffered_on_read", 0);

  /* "uvloop/handles/stream.pyx":973
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream buffered read callback") == 0:
 *         return
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_stream), ((char const *)"UVStream buffered read callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(20, 973, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":974
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,
 *                             "UVStream buffered read callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_t_1 == 0);

  /* "uvloop/handles/stream.pyx":973
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream buffered read callback") == 0:
 *         return
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":975
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,
 *                             "UVStream buffered read callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":973
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>stream,             # <<<<<<<<<<<<<<
 *                             "UVStream buffered read callback") == 0:
 *         return
 */
  }

  /* "uvloop/handles/stream.pyx":978
 * 
 *     cdef:
 *         UVStream sc = <UVStream>stream.data             # <<<<<<<<<<<<<<
 *         Loop loop = sc._loop
 *         Py_buffer* pybuf = &sc._read_pybuf
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_stream->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_sc = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/stream.pyx":979
 *     cdef:
 *         UVStream sc = <UVStream>stream.data
 *         Loop loop = sc._loop             # <<<<<<<<<<<<<<
 *         Py_buffer* pybuf = &sc._read_pybuf
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/stream.pyx":980
 *         UVStream sc = <UVStream>stream.data
 *         Loop loop = sc._loop
 *         Py_buffer* pybuf = &sc._read_pybuf             # <<<<<<<<<<<<<<
 * 
 *     if nread == uv.UV_ENOBUFS:
 */
  __pyx_v_pybuf = (&__pyx_v_sc->_read_pybuf);

  /* "uvloop/handles/stream.pyx":982
 *         Py_buffer* pybuf = &sc._read_pybuf
 * 
 *     if nread == uv.UV_ENOBUFS:             # <<<<<<<<<<<<<<
 *         sc._fatal_error(
 *             RuntimeError(
 */
  __pyx_t_2 = (__pyx_v_nread == UV_ENOBUFS);
  if (__pyx_t_2) {

    /* "uvloop/handles/stream.pyx":984
 *     if nread == uv.UV_ENOBUFS:
 *         sc._fatal_error(
 *             RuntimeError(             # <<<<<<<<<<<<<<
 *                 'unhandled error (or an empty buffer) in get_buffer()'),
 *             False)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__157, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/stream.pyx":983
 * 
 *     if nread == uv.UV_ENOBUFS:
 *         sc._fatal_error(             # <<<<<<<<<<<<<<
 *             RuntimeError(
 *                 'unhandled error (or an empty buffer) in get_buffer()'),
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_sc), __pyx_t_3, Py_False, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/stream.pyx":987
 *                 'unhandled error (or an empty buffer) in get_buffer()'),
 *             False)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    goto __pyx_L0;

    /* "uvloop/handles/stream.pyx":982
 *         Py_buffer* pybuf = &sc._read_pybuf
 * 
 *     if nread == uv.UV_ENOBUFS:             # <<<<<<<<<<<<<<
 *         sc._fatal_error(
 *             RuntimeError(
 */
  }

  /* "uvloop/handles/stream.pyx":989
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if nread > 0 and not sc._read_pybuf_acquired:
 *             # From libuv docs:
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "uvloop/handles/stream.pyx":990
 * 
 *     try:
 *         if nread > 0 and not sc._read_pybuf_acquired:             # <<<<<<<<<<<<<<
 *             # From libuv docs:
 *             #     nread is > 0 if there is data available or < 0 on error. When
 */
        __pyx_t_1 = (__pyx_v_nread > 0);
        if (__pyx_t_1) {
        } else {
          __pyx_t_2 = __pyx_t_1;
          goto __pyx_L15_bool_binop_done;
        }
        __pyx_t_1 = (!__pyx_v_sc->_read_pybuf_acquired);
        __pyx_t_2 = __pyx_t_1;
        __pyx_L15_bool_binop_done:;
        if (unlikely(__pyx_t_2)) {

          /* "uvloop/handles/stream.pyx":997
 *             #     buffer; in that case buf.len and buf.base are both set to 0.
 *             raise RuntimeError(
 *                 f'no python buffer is allocated in on_read; nread={nread}')             # <<<<<<<<<<<<<<
 * 
 *         if nread == 0:
 */
          __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_v_nread, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 997, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_no_python_buffer_is_allocated_in, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 997, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "uvloop/handles/stream.pyx":996
 *             #     nread < 0, the buf parameter might not point to a valid
 *             #     buffer; in that case buf.len and buf.base are both set to 0.
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 f'no python buffer is allocated in on_read; nread={nread}')
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 996, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(20, 996, __pyx_L8_error)

          /* "uvloop/handles/stream.pyx":990
 * 
 *     try:
 *         if nread > 0 and not sc._read_pybuf_acquired:             # <<<<<<<<<<<<<<
 *             # From libuv docs:
 *             #     nread is > 0 if there is data available or < 0 on error. When
 */
        }

        /* "uvloop/handles/stream.pyx":999
 *                 f'no python buffer is allocated in on_read; nread={nread}')
 * 
 *         if nread == 0:             # <<<<<<<<<<<<<<
 *             # From libuv docs:
 *             #     nread might be 0, which does not indicate an error or EOF.
 */
        __pyx_t_2 = (__pyx_v_nread == 0);
        if (__pyx_t_2) {

          /* "uvloop/handles/stream.pyx":1003
 *             #     nread might be 0, which does not indicate an error or EOF.
 *             #     This is equivalent to EAGAIN or EWOULDBLOCK under read(2).
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if __uv_stream_on_read_common(sc, loop, nread):
 */
          goto __pyx_L12_try_return;

          /* "uvloop/handles/stream.pyx":999
 *                 f'no python buffer is allocated in on_read; nread={nread}')
 * 
 *         if nread == 0:             # <<<<<<<<<<<<<<
 *             # From libuv docs:
 *             #     nread might be 0, which does not indicate an error or EOF.
 */
        }

        /* "uvloop/handles/stream.pyx":1005
 *             return
 * 
 *         if __uv_stream_on_read_common(sc, loop, nread):             # <<<<<<<<<<<<<<
 *             return
 * 
 */
        __pyx_t_2 = __pyx_f_6uvloop_4loop___uv_stream_on_read_common(__pyx_v_sc, __pyx_v_loop, __pyx_v_nread); if (unlikely(__pyx_t_2 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(20, 1005, __pyx_L8_error)
        if (__pyx_t_2) {

          /* "uvloop/handles/stream.pyx":1006
 * 
 *         if __uv_stream_on_read_common(sc, loop, nread):
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG:
 */
          goto __pyx_L12_try_return;

          /* "uvloop/handles/stream.pyx":1005
 *             return
 * 
 *         if __uv_stream_on_read_common(sc, loop, nread):             # <<<<<<<<<<<<<<
 *             return
 * 
 */
        }

        /* "uvloop/handles/stream.pyx":1008
 *             return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_total += 1
 * 
 */
        __pyx_t_2 = (UVLOOP_DEBUG != 0);
        if (__pyx_t_2) {

          /* "uvloop/handles/stream.pyx":1009
 * 
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_total += 1             # <<<<<<<<<<<<<<
 * 
 *         run_in_context1(sc.context, sc._protocol_buffer_updated, nread)
 */
          __pyx_v_loop->_debug_stream_read_cb_total = (__pyx_v_loop->_debug_stream_read_cb_total + 1);

          /* "uvloop/handles/stream.pyx":1008
 *             return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_total += 1
 * 
 */
        }

        /* "uvloop/handles/stream.pyx":1011
 *             loop._debug_stream_read_cb_total += 1
 * 
 *         run_in_context1(sc.context, sc._protocol_buffer_updated, nread)             # <<<<<<<<<<<<<<
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:
 */
        __pyx_t_4 = __pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.context;
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_3 = __pyx_v_sc->_protocol_buffer_updated;
        __Pyx_INCREF(__pyx_t_3);
        __pyx_t_8 = PyInt_FromSsize_t(__pyx_v_nread); if (unlikely(!__pyx_t_8)) __PYX_ERR(20, 1011, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __pyx_f_6uvloop_4loop_run_in_context1(__pyx_t_4, __pyx_t_3, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(20, 1011, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "uvloop/handles/stream.pyx":989
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if nread > 0 and not sc._read_pybuf_acquired:
 *             # From libuv docs:
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L13_try_end;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/handles/stream.pyx":1012
 * 
 *         run_in_context1(sc.context, sc._protocol_buffer_updated, nread)
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_errors_total += 1
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.__uv_stream_buffered_on_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_3) < 0) __PYX_ERR(20, 1012, __pyx_L10_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_v_exc = __pyx_t_8;
        /*try:*/ {

          /* "uvloop/handles/stream.pyx":1013
 *         run_in_context1(sc.context, sc._protocol_buffer_updated, nread)
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_errors_total += 1
 * 
 */
          __pyx_t_2 = (UVLOOP_DEBUG != 0);
          if (__pyx_t_2) {

            /* "uvloop/handles/stream.pyx":1014
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *         sc._fatal_error(exc, False)
 */
            __pyx_v_loop->_debug_stream_read_cb_errors_total = (__pyx_v_loop->_debug_stream_read_cb_errors_total + 1);

            /* "uvloop/handles/stream.pyx":1013
 *         run_in_context1(sc.context, sc._protocol_buffer_updated, nread)
 *     except BaseException as exc:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             loop._debug_stream_read_cb_errors_total += 1
 * 
 */
          }

          /* "uvloop/handles/stream.pyx":1016
 *             loop._debug_stream_read_cb_errors_total += 1
 * 
 *         sc._fatal_error(exc, False)             # <<<<<<<<<<<<<<
 *     finally:
 *         sc._read_pybuf_acquired = 0
 */
          __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_sc->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_sc), __pyx_v_exc, Py_False, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(20, 1016, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "uvloop/handles/stream.pyx":1012
 * 
 *         run_in_context1(sc.context, sc._protocol_buffer_updated, nread)
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             loop._debug_stream_read_cb_errors_total += 1
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L26;
          }
          __pyx_L25_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __pyx_t_10 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            }
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
            goto __pyx_L10_except_error;
          }
          __pyx_L26:;
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L9_exception_handled;
      }
      goto __pyx_L10_except_error;

      /* "uvloop/handles/stream.pyx":989
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if nread > 0 and not sc._read_pybuf_acquired:
 *             # From libuv docs:
 */
      __pyx_L10_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L6_error;
      __pyx_L12_try_return:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L5_return;
      __pyx_L9_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_L13_try_end:;
    }
  }

  /* "uvloop/handles/stream.pyx":1018
 *         sc._fatal_error(exc, False)
 *     finally:
 *         sc._read_pybuf_acquired = 0             # <<<<<<<<<<<<<<
 *         PyBuffer_Release(pybuf)
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_v_sc->_read_pybuf_acquired = 0;

      /* "uvloop/handles/stream.pyx":1019
 *     finally:
 *         sc._read_pybuf_acquired = 0
 *         PyBuffer_Release(pybuf)             # <<<<<<<<<<<<<<
 */
      PyBuffer_Release(__pyx_v_pybuf);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_16);
      __pyx_t_11 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
      {

        /* "uvloop/handles/stream.pyx":1018
 *         sc._fatal_error(exc, False)
 *     finally:
 *         sc._read_pybuf_acquired = 0             # <<<<<<<<<<<<<<
 *         PyBuffer_Release(pybuf)
 */
        __pyx_v_sc->_read_pybuf_acquired = 0;

        /* "uvloop/handles/stream.pyx":1019
 *     finally:
 *         sc._read_pybuf_acquired = 0
 *         PyBuffer_Release(pybuf)             # <<<<<<<<<<<<<<
 */
        PyBuffer_Release(__pyx_v_pybuf);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_17, __pyx_t_16);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_6, __pyx_t_5);
      __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_19;
      goto __pyx_L1_error;
    }
    __pyx_L5_return: {

      /* "uvloop/handles/stream.pyx":1018
 *         sc._fatal_error(exc, False)
 *     finally:
 *         sc._read_pybuf_acquired = 0             # <<<<<<<<<<<<<<
 *         PyBuffer_Release(pybuf)
 */
      __pyx_v_sc->_read_pybuf_acquired = 0;

      /* "uvloop/handles/stream.pyx":1019
 *     finally:
 *         sc._read_pybuf_acquired = 0
 *         PyBuffer_Release(pybuf)             # <<<<<<<<<<<<<<
 */
      PyBuffer_Release(__pyx_v_pybuf);
      goto __pyx_L0;
    }
    __pyx_L7:;
  }

  /* "uvloop/handles/stream.pyx":967
 * 
 * 
 * cdef void __uv_stream_buffered_on_read(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* stream,
 *     ssize_t nread,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_stream_buffered_on_read", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sc);
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/streamserver.pyx":4
 * cdef class UVStreamServer(UVSocketHandle):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.opened = 0
 *         self._server = None
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_14UVStreamServer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_14UVStreamServer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_14UVStreamServer___cinit__(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_14UVStreamServer___cinit__(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/streamserver.pyx":5
 * 
 *     def __cinit__(self):
 *         self.opened = 0             # <<<<<<<<<<<<<<
 *         self._server = None
 *         self.ssl = None
 */
  __pyx_v_self->opened = 0;

  /* "uvloop/handles/streamserver.pyx":6
 *     def __cinit__(self):
 *         self.opened = 0
 *         self._server = None             # <<<<<<<<<<<<<<
 *         self.ssl = None
 *         self.ssl_handshake_timeout = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_server);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_server);
  __pyx_v_self->_server = ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None);

  /* "uvloop/handles/streamserver.pyx":7
 *         self.opened = 0
 *         self._server = None
 *         self.ssl = None             # <<<<<<<<<<<<<<
 *         self.ssl_handshake_timeout = None
 *         self.ssl_shutdown_timeout = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ssl);
  __Pyx_DECREF(__pyx_v_self->ssl);
  __pyx_v_self->ssl = Py_None;

  /* "uvloop/handles/streamserver.pyx":8
 *         self._server = None
 *         self.ssl = None
 *         self.ssl_handshake_timeout = None             # <<<<<<<<<<<<<<
 *         self.ssl_shutdown_timeout = None
 *         self.protocol_factory = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ssl_handshake_timeout);
  __Pyx_DECREF(__pyx_v_self->ssl_handshake_timeout);
  __pyx_v_self->ssl_handshake_timeout = Py_None;

  /* "uvloop/handles/streamserver.pyx":9
 *         self.ssl = None
 *         self.ssl_handshake_timeout = None
 *         self.ssl_shutdown_timeout = None             # <<<<<<<<<<<<<<
 *         self.protocol_factory = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ssl_shutdown_timeout);
  __Pyx_DECREF(__pyx_v_self->ssl_shutdown_timeout);
  __pyx_v_self->ssl_shutdown_timeout = Py_None;

  /* "uvloop/handles/streamserver.pyx":10
 *         self.ssl_handshake_timeout = None
 *         self.ssl_shutdown_timeout = None
 *         self.protocol_factory = None             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _init(self, Loop loop, object protocol_factory,
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->protocol_factory);
  __Pyx_DECREF(__pyx_v_self->protocol_factory);
  __pyx_v_self->protocol_factory = Py_None;

  /* "uvloop/handles/streamserver.pyx":4
 * cdef class UVStreamServer(UVSocketHandle):
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.opened = 0
 *         self._server = None
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":12
 *         self.protocol_factory = None
 * 
 *     cdef inline _init(self, Loop loop, object protocol_factory,             # <<<<<<<<<<<<<<
 *                       Server server,
 *                       object backlog,
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__init(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol_factory, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/streamserver.pyx":19
 *                       object ssl_shutdown_timeout):
 * 
 *         if not isinstance(backlog, int):             # <<<<<<<<<<<<<<
 *             # Don't allow floats
 *             raise TypeError('integer argument expected, got {}'.format(
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_backlog); 
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/streamserver.pyx":21
 *         if not isinstance(backlog, int):
 *             # Don't allow floats
 *             raise TypeError('integer argument expected, got {}'.format(             # <<<<<<<<<<<<<<
 *                 type(backlog).__name__))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_integer_argument_expected_got, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uvloop/handles/streamserver.pyx":22
 *             # Don't allow floats
 *             raise TypeError('integer argument expected, got {}'.format(
 *                 type(backlog).__name__))             # <<<<<<<<<<<<<<
 * 
 *         if ssl is not None:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_backlog)), __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(21, 22, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 21, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "uvloop/handles/streamserver.pyx":21
 *         if not isinstance(backlog, int):
 *             # Don't allow floats
 *             raise TypeError('integer argument expected, got {}'.format(             # <<<<<<<<<<<<<<
 *                 type(backlog).__name__))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(21, 21, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":19
 *                       object ssl_shutdown_timeout):
 * 
 *         if not isinstance(backlog, int):             # <<<<<<<<<<<<<<
 *             # Don't allow floats
 *             raise TypeError('integer argument expected, got {}'.format(
 */
  }

  /* "uvloop/handles/streamserver.pyx":24
 *                 type(backlog).__name__))
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError(
 */
  __pyx_t_2 = (__pyx_v_ssl != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/handles/streamserver.pyx":25
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     'ssl is expected to be None or an instance of '
 */
    __pyx_t_4 = __pyx_v_6uvloop_4loop_ssl_SSLContext;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_2 = PyObject_IsInstance(__pyx_v_ssl, __pyx_t_4); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(21, 25, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = (!__pyx_t_2);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/streamserver.pyx":28
 *                 raise TypeError(
 *                     'ssl is expected to be None or an instance of '
 *                     'ssl.SSLContext, got {!r}'.format(ssl))             # <<<<<<<<<<<<<<
 *         else:
 *             if ssl_handshake_timeout is not None:
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_ssl_is_expected_to_be_None_or_an, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 28, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_ssl};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 28, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/handles/streamserver.pyx":26
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     'ssl is expected to be None or an instance of '
 *                     'ssl.SSLContext, got {!r}'.format(ssl))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 26, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(21, 26, __pyx_L1_error)

      /* "uvloop/handles/streamserver.pyx":25
 * 
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     'ssl is expected to be None or an instance of '
 */
    }

    /* "uvloop/handles/streamserver.pyx":24
 *                 type(backlog).__name__))
 * 
 *         if ssl is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError(
 */
    goto __pyx_L4;
  }

  /* "uvloop/handles/streamserver.pyx":30
 *                     'ssl.SSLContext, got {!r}'.format(ssl))
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
  /*else*/ {
    __pyx_t_1 = (__pyx_v_ssl_handshake_timeout != Py_None);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/streamserver.pyx":31
 *         else:
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 31, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(21, 31, __pyx_L1_error)

      /* "uvloop/handles/streamserver.pyx":30
 *                     'ssl.SSLContext, got {!r}'.format(ssl))
 *         else:
 *             if ssl_handshake_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 */
    }

    /* "uvloop/handles/streamserver.pyx":33
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    __pyx_t_1 = (__pyx_v_ssl_shutdown_timeout != Py_None);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/streamserver.pyx":34
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(21, 34, __pyx_L1_error)

      /* "uvloop/handles/streamserver.pyx":33
 *                 raise ValueError(
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 */
    }
  }
  __pyx_L4:;

  /* "uvloop/handles/streamserver.pyx":37
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 *         self.backlog = backlog             # <<<<<<<<<<<<<<
 *         self.ssl = ssl
 *         self.ssl_handshake_timeout = ssl_handshake_timeout
 */
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_backlog); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(21, 37, __pyx_L1_error)
  __pyx_v_self->backlog = __pyx_t_8;

  /* "uvloop/handles/streamserver.pyx":38
 * 
 *         self.backlog = backlog
 *         self.ssl = ssl             # <<<<<<<<<<<<<<
 *         self.ssl_handshake_timeout = ssl_handshake_timeout
 *         self.ssl_shutdown_timeout = ssl_shutdown_timeout
 */
  __Pyx_INCREF(__pyx_v_ssl);
  __Pyx_GIVEREF(__pyx_v_ssl);
  __Pyx_GOTREF(__pyx_v_self->ssl);
  __Pyx_DECREF(__pyx_v_self->ssl);
  __pyx_v_self->ssl = __pyx_v_ssl;

  /* "uvloop/handles/streamserver.pyx":39
 *         self.backlog = backlog
 *         self.ssl = ssl
 *         self.ssl_handshake_timeout = ssl_handshake_timeout             # <<<<<<<<<<<<<<
 *         self.ssl_shutdown_timeout = ssl_shutdown_timeout
 * 
 */
  __Pyx_INCREF(__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_v_ssl_handshake_timeout);
  __Pyx_GOTREF(__pyx_v_self->ssl_handshake_timeout);
  __Pyx_DECREF(__pyx_v_self->ssl_handshake_timeout);
  __pyx_v_self->ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;

  /* "uvloop/handles/streamserver.pyx":40
 *         self.ssl = ssl
 *         self.ssl_handshake_timeout = ssl_handshake_timeout
 *         self.ssl_shutdown_timeout = ssl_shutdown_timeout             # <<<<<<<<<<<<<<
 * 
 *         self._start_init(loop)
 */
  __Pyx_INCREF(__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_v_ssl_shutdown_timeout);
  __Pyx_GOTREF(__pyx_v_self->ssl_shutdown_timeout);
  __Pyx_DECREF(__pyx_v_self->ssl_shutdown_timeout);
  __pyx_v_self->ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;

  /* "uvloop/handles/streamserver.pyx":42
 *         self.ssl_shutdown_timeout = ssl_shutdown_timeout
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 *         self.protocol_factory = protocol_factory
 *         self._server = server
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/streamserver.pyx":43
 * 
 *         self._start_init(loop)
 *         self.protocol_factory = protocol_factory             # <<<<<<<<<<<<<<
 *         self._server = server
 * 
 */
  __Pyx_INCREF(__pyx_v_protocol_factory);
  __Pyx_GIVEREF(__pyx_v_protocol_factory);
  __Pyx_GOTREF(__pyx_v_self->protocol_factory);
  __Pyx_DECREF(__pyx_v_self->protocol_factory);
  __pyx_v_self->protocol_factory = __pyx_v_protocol_factory;

  /* "uvloop/handles/streamserver.pyx":44
 *         self._start_init(loop)
 *         self.protocol_factory = protocol_factory
 *         self._server = server             # <<<<<<<<<<<<<<
 * 
 *     cdef inline listen(self):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_server);
  __Pyx_GIVEREF((PyObject *)__pyx_v_server);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_server);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_server);
  __pyx_v_self->_server = __pyx_v_server;

  /* "uvloop/handles/streamserver.pyx":12
 *         self.protocol_factory = None
 * 
 *     cdef inline _init(self, Loop loop, object protocol_factory,             # <<<<<<<<<<<<<<
 *                       Server server,
 *                       object backlog,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":46
 *         self._server = server
 * 
 *     cdef inline listen(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer_listen(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("listen", 1);

  /* "uvloop/handles/streamserver.pyx":48
 *     cdef inline listen(self):
 *         cdef int err
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self.protocol_factory is None:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/streamserver.pyx":50
 *         self._ensure_alive()
 * 
 *         if self.protocol_factory is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('unable to listen(); no protocol_factory')
 * 
 */
  __pyx_t_2 = (__pyx_v_self->protocol_factory == Py_None);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/streamserver.pyx":51
 * 
 *         if self.protocol_factory is None:
 *             raise RuntimeError('unable to listen(); no protocol_factory')             # <<<<<<<<<<<<<<
 * 
 *         if self.opened != 1:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__158, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(21, 51, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":50
 *         self._ensure_alive()
 * 
 *         if self.protocol_factory is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('unable to listen(); no protocol_factory')
 * 
 */
  }

  /* "uvloop/handles/streamserver.pyx":53
 *             raise RuntimeError('unable to listen(); no protocol_factory')
 * 
 *         if self.opened != 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('unopened TCPServer')
 * 
 */
  __pyx_t_2 = (__pyx_v_self->opened != 1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/streamserver.pyx":54
 * 
 *         if self.opened != 1:
 *             raise RuntimeError('unopened TCPServer')             # <<<<<<<<<<<<<<
 * 
 *         self.context = Context_CopyCurrent()
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__159, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(21, 54, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":53
 *             raise RuntimeError('unable to listen(); no protocol_factory')
 * 
 *         if self.opened != 1:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('unopened TCPServer')
 * 
 */
  }

  /* "uvloop/handles/streamserver.pyx":56
 *             raise RuntimeError('unopened TCPServer')
 * 
 *         self.context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_listen(<uv.uv_stream_t*> self._handle,
 */
  __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.context);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.context);
  __pyx_v_self->__pyx_base.__pyx_base.context = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/streamserver.pyx":58
 *         self.context = Context_CopyCurrent()
 * 
 *         err = uv.uv_listen(<uv.uv_stream_t*> self._handle,             # <<<<<<<<<<<<<<
 *                            self.backlog,
 *                            __uv_streamserver_on_listen)
 */
  __pyx_v_err = uv_listen(((uv_stream_t *)__pyx_v_self->__pyx_base.__pyx_base._handle), __pyx_v_self->backlog, __pyx_f_6uvloop_4loop___uv_streamserver_on_listen);

  /* "uvloop/handles/streamserver.pyx":61
 *                            self.backlog,
 *                            __uv_streamserver_on_listen)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/streamserver.pyx":62
 *                            __uv_streamserver_on_listen)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/streamserver.pyx":63
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/streamserver.pyx":64
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _on_listen(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/streamserver.pyx":61
 *                            self.backlog,
 *                            __uv_streamserver_on_listen)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/streamserver.pyx":46
 *         self._server = server
 * 
 *     cdef inline listen(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer.listen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":66
 *             return
 * 
 *     cdef inline _on_listen(self):             # <<<<<<<<<<<<<<
 *         cdef UVStream client
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__on_listen(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_client = 0;
  PyObject *__pyx_v_protocol = NULL;
  PyObject *__pyx_v_waiter = NULL;
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_ssl_protocol = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_listen", 1);

  /* "uvloop/handles/streamserver.pyx":69
 *         cdef UVStream client
 * 
 *         protocol = run_in_context(self.context, self.protocol_factory)             # <<<<<<<<<<<<<<
 * 
 *         if self.ssl is None:
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.context;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->protocol_factory;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_6uvloop_4loop_run_in_context(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_protocol = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/handles/streamserver.pyx":71
 *         protocol = run_in_context(self.context, self.protocol_factory)
 * 
 *         if self.ssl is None:             # <<<<<<<<<<<<<<
 *             client = self._make_new_transport(protocol, None, self.context)
 * 
 */
  __pyx_t_4 = (__pyx_v_self->ssl == Py_None);
  if (__pyx_t_4) {

    /* "uvloop/handles/streamserver.pyx":72
 * 
 *         if self.ssl is None:
 *             client = self._make_new_transport(protocol, None, self.context)             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    __pyx_t_3 = __pyx_v_self->__pyx_base.__pyx_base.context;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_make_new_transport(__pyx_v_self, __pyx_v_protocol, Py_None, __pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_client = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/streamserver.pyx":71
 *         protocol = run_in_context(self.context, self.protocol_factory)
 * 
 *         if self.ssl is None:             # <<<<<<<<<<<<<<
 *             client = self._make_new_transport(protocol, None, self.context)
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/streamserver.pyx":75
 * 
 *         else:
 *             waiter = self._loop._new_future()             # <<<<<<<<<<<<<<
 * 
 *             ssl_protocol = SSLProtocol(
 */
  /*else*/ {
    __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self->__pyx_base.__pyx_base._loop); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_waiter = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/streamserver.pyx":77
 *             waiter = self._loop._new_future()
 * 
 *             ssl_protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self._loop, protocol, self.ssl,
 *                 waiter,
 */
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop))) __PYX_ERR(21, 77, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_protocol);
    __Pyx_GIVEREF(__pyx_v_protocol);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_protocol)) __PYX_ERR(21, 77, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_self->ssl);
    __Pyx_GIVEREF(__pyx_v_self->ssl);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->ssl)) __PYX_ERR(21, 77, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_waiter);
    __Pyx_GIVEREF(__pyx_v_waiter);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_waiter)) __PYX_ERR(21, 77, __pyx_L1_error);

    /* "uvloop/handles/streamserver.pyx":80
 *                 self._loop, protocol, self.ssl,
 *                 waiter,
 *                 server_side=True,             # <<<<<<<<<<<<<<
 *                 server_hostname=None,
 *                 ssl_handshake_timeout=self.ssl_handshake_timeout,
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_server_side, Py_True) < 0) __PYX_ERR(21, 80, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":81
 *                 waiter,
 *                 server_side=True,
 *                 server_hostname=None,             # <<<<<<<<<<<<<<
 *                 ssl_handshake_timeout=self.ssl_handshake_timeout,
 *                 ssl_shutdown_timeout=self.ssl_shutdown_timeout)
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_server_hostname, Py_None) < 0) __PYX_ERR(21, 80, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":82
 *                 server_side=True,
 *                 server_hostname=None,
 *                 ssl_handshake_timeout=self.ssl_handshake_timeout,             # <<<<<<<<<<<<<<
 *                 ssl_shutdown_timeout=self.ssl_shutdown_timeout)
 * 
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_handshake_timeout, __pyx_v_self->ssl_handshake_timeout) < 0) __PYX_ERR(21, 80, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":83
 *                 server_hostname=None,
 *                 ssl_handshake_timeout=self.ssl_handshake_timeout,
 *                 ssl_shutdown_timeout=self.ssl_shutdown_timeout)             # <<<<<<<<<<<<<<
 * 
 *             client = self._make_new_transport(ssl_protocol, None, self.context)
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_shutdown_timeout, __pyx_v_self->ssl_shutdown_timeout) < 0) __PYX_ERR(21, 80, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":77
 *             waiter = self._loop._new_future()
 * 
 *             ssl_protocol = SSLProtocol(             # <<<<<<<<<<<<<<
 *                 self._loop, protocol, self.ssl,
 *                 waiter,
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_ssl_protocol = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/handles/streamserver.pyx":85
 *                 ssl_shutdown_timeout=self.ssl_shutdown_timeout)
 * 
 *             client = self._make_new_transport(ssl_protocol, None, self.context)             # <<<<<<<<<<<<<<
 * 
 *             waiter.add_done_callback(
 */
    __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base.context;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_make_new_transport(__pyx_v_self, ((PyObject *)__pyx_v_ssl_protocol), Py_None, __pyx_t_1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_client = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/handles/streamserver.pyx":87
 *             client = self._make_new_transport(ssl_protocol, None, self.context)
 * 
 *             waiter.add_done_callback(             # <<<<<<<<<<<<<<
 *                 ft_partial(self.__on_ssl_connected, client))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_add_done_callback); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "uvloop/handles/streamserver.pyx":88
 * 
 *             waiter.add_done_callback(
 *                 ft_partial(self.__on_ssl_connected, client))             # <<<<<<<<<<<<<<
 * 
 *         client._accept(<UVStream>self)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_UVStreamServer__on_ssl_connecte); if (unlikely(!__pyx_t_5)) __PYX_ERR(21, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_ft_partial);
    __pyx_t_6 = __pyx_v_6uvloop_4loop_ft_partial; __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_5, ((PyObject *)__pyx_v_client)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 88, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_2};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/streamserver.pyx":90
 *                 ft_partial(self.__on_ssl_connected, client))
 * 
 *         client._accept(<UVStream>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_8UVStream__accept(__pyx_v_client, ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/streamserver.pyx":66
 *             return
 * 
 *     cdef inline _on_listen(self):             # <<<<<<<<<<<<<<
 *         cdef UVStream client
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer._on_listen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_client);
  __Pyx_XDECREF(__pyx_v_protocol);
  __Pyx_XDECREF(__pyx_v_waiter);
  __Pyx_XDECREF((PyObject *)__pyx_v_ssl_protocol);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":92
 *         client._accept(<UVStream>self)
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         # Overload UVHandle._fatal_error
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__fatal_error(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, PyObject *__pyx_v_exc, PyObject *__pyx_v_throw, struct __pyx_opt_args_6uvloop_4loop_14UVStreamServer__fatal_error *__pyx_optional_args) {
  PyObject *__pyx_v_reason = ((PyObject *)Py_None);
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  Py_UCS4 __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fatal_error", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_reason = __pyx_optional_args->reason;
    }
  }

  /* "uvloop/handles/streamserver.pyx":95
 *         # Overload UVHandle._fatal_error
 * 
 *         self._close()             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(exc, OSError):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/streamserver.pyx":97
 *         self._close()
 * 
 *         if not isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 * 
 *             if throw or self._loop is None:
 */
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_exc, __pyx_builtin_OSError); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(21, 97, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {

    /* "uvloop/handles/streamserver.pyx":99
 *         if not isinstance(exc, OSError):
 * 
 *             if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_throw); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(21, 99, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop) == Py_None);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/handles/streamserver.pyx":100
 * 
 *             if throw or self._loop is None:
 *                 raise exc             # <<<<<<<<<<<<<<
 * 
 *             msg = f'Fatal error on server {self.__class__.__name__}'
 */
      __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
      __PYX_ERR(21, 100, __pyx_L1_error)

      /* "uvloop/handles/streamserver.pyx":99
 *         if not isinstance(exc, OSError):
 * 
 *             if throw or self._loop is None:             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
    }

    /* "uvloop/handles/streamserver.pyx":102
 *                 raise exc
 * 
 *             msg = f'Fatal error on server {self.__class__.__name__}'             # <<<<<<<<<<<<<<
 *             if reason is not None:
 *                 msg = f'{msg} ({reason})'
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Fatal_error_on_server, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_msg = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "uvloop/handles/streamserver.pyx":103
 * 
 *             msg = f'Fatal error on server {self.__class__.__name__}'
 *             if reason is not None:             # <<<<<<<<<<<<<<
 *                 msg = f'{msg} ({reason})'
 * 
 */
    __pyx_t_3 = (__pyx_v_reason != Py_None);
    if (__pyx_t_3) {

      /* "uvloop/handles/streamserver.pyx":104
 *             msg = f'Fatal error on server {self.__class__.__name__}'
 *             if reason is not None:
 *                 msg = f'{msg} ({reason})'             # <<<<<<<<<<<<<<
 * 
 *             self._loop.call_exception_handler({
 */
      __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 0;
      __pyx_t_6 = 127;
      __Pyx_INCREF(__pyx_v_msg);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_msg) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_msg) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_v_msg);
      __Pyx_GIVEREF(__pyx_v_msg);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_msg);
      __Pyx_INCREF(__pyx_kp_u__140);
      __pyx_t_5 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__140);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_kp_u__140);
      __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_reason, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_6) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_6;
      __pyx_t_5 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u__141);
      __pyx_t_5 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__141);
      PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_kp_u__141);
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_4, 4, __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_msg, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "uvloop/handles/streamserver.pyx":103
 * 
 *             msg = f'Fatal error on server {self.__class__.__name__}'
 *             if reason is not None:             # <<<<<<<<<<<<<<
 *                 msg = f'{msg} ({reason})'
 * 
 */
    }

    /* "uvloop/handles/streamserver.pyx":106
 *                 msg = f'{msg} ({reason})'
 * 
 *             self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                 'message': msg,
 *                 'exception': exc,
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_4)) __PYX_ERR(21, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uvloop/handles/streamserver.pyx":107
 * 
 *             self._loop.call_exception_handler({
 *                 'message': msg,             # <<<<<<<<<<<<<<
 *                 'exception': exc,
 *             })
 */
    __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(21, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_message, __pyx_v_msg) < 0) __PYX_ERR(21, 107, __pyx_L1_error)

    /* "uvloop/handles/streamserver.pyx":108
 *             self._loop.call_exception_handler({
 *                 'message': msg,
 *                 'exception': exc,             # <<<<<<<<<<<<<<
 *             })
 * 
 */
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(21, 107, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/streamserver.pyx":97
 *         self._close()
 * 
 *         if not isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 * 
 *             if throw or self._loop is None:
 */
  }

  /* "uvloop/handles/streamserver.pyx":92
 *         client._accept(<UVStream>self)
 * 
 *     cdef _fatal_error(self, exc, throw, reason=None):             # <<<<<<<<<<<<<<
 *         # Overload UVHandle._fatal_error
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer._fatal_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":111
 *             })
 * 
 *     cdef inline _mark_as_open(self):             # <<<<<<<<<<<<<<
 *         self.opened = 1
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_14UVStreamServer__mark_as_open(struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_mark_as_open", 1);

  /* "uvloop/handles/streamserver.pyx":112
 * 
 *     cdef inline _mark_as_open(self):
 *         self.opened = 1             # <<<<<<<<<<<<<<
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,
 */
  __pyx_v_self->opened = 1;

  /* "uvloop/handles/streamserver.pyx":111
 *             })
 * 
 *     cdef inline _mark_as_open(self):             # <<<<<<<<<<<<<<
 *         self.opened = 1
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":114
 *         self.opened = 1
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,             # <<<<<<<<<<<<<<
 *                                       object context):
 *         raise NotImplementedError
 */

static struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_f_6uvloop_4loop_14UVStreamServer__make_new_transport(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_protocol, CYTHON_UNUSED PyObject *__pyx_v_waiter, CYTHON_UNUSED PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_new_transport", 1);

  /* "uvloop/handles/streamserver.pyx":116
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,
 *                                       object context):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def __on_ssl_connected(self, transport, fut):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(21, 116, __pyx_L1_error)

  /* "uvloop/handles/streamserver.pyx":114
 *         self.opened = 1
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,             # <<<<<<<<<<<<<<
 *                                       object context):
 *         raise NotImplementedError
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer._make_new_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":118
 *         raise NotImplementedError
 * 
 *     def __on_ssl_connected(self, transport, fut):             # <<<<<<<<<<<<<<
 *         exc = fut.exception()
 *         if exc is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_14UVStreamServer_3_UVStreamServer__on_ssl_connected(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_14UVStreamServer_2__on_ssl_connected, "UVStreamServer.__on_ssl_connected(self, transport, fut)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_14UVStreamServer_3_UVStreamServer__on_ssl_connected = {"_UVStreamServer__on_ssl_connected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVStreamServer_3_UVStreamServer__on_ssl_connected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVStreamServer_2__on_ssl_connected};
static PyObject *__pyx_pw_6uvloop_4loop_14UVStreamServer_3_UVStreamServer__on_ssl_connected(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_transport = 0;
  PyObject *__pyx_v_fut = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_UVStreamServer__on_ssl_connected (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_transport,&__pyx_n_s_fut,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_transport)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(21, 118, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fut)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(21, 118, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_UVStreamServer__on_ssl_connected", 1, 2, 2, 1); __PYX_ERR(21, 118, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_UVStreamServer__on_ssl_connected") < 0)) __PYX_ERR(21, 118, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_transport = values[0];
    __pyx_v_fut = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_UVStreamServer__on_ssl_connected", 1, 2, 2, __pyx_nargs); __PYX_ERR(21, 118, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer._UVStreamServer__on_ssl_connected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_14UVStreamServer_2__on_ssl_connected(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self), __pyx_v_transport, __pyx_v_fut);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_14UVStreamServer_2__on_ssl_connected(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, PyObject *__pyx_v_transport, PyObject *__pyx_v_fut) {
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_UVStreamServer__on_ssl_connected", 1);

  /* "uvloop/handles/streamserver.pyx":119
 * 
 *     def __on_ssl_connected(self, transport, fut):
 *         exc = fut.exception()             # <<<<<<<<<<<<<<
 *         if exc is not None:
 *             transport._force_close(exc)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_exc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/streamserver.pyx":120
 *     def __on_ssl_connected(self, transport, fut):
 *         exc = fut.exception()
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             transport._force_close(exc)
 * 
 */
  __pyx_t_5 = (__pyx_v_exc != Py_None);
  if (__pyx_t_5) {

    /* "uvloop/handles/streamserver.pyx":121
 *         exc = fut.exception()
 *         if exc is not None:
 *             transport._force_close(exc)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_transport, __pyx_n_s_force_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(21, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_exc};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/streamserver.pyx":120
 *     def __on_ssl_connected(self, transport, fut):
 *         exc = fut.exception()
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             transport._force_close(exc)
 * 
 */
  }

  /* "uvloop/handles/streamserver.pyx":118
 *         raise NotImplementedError
 * 
 *     def __on_ssl_connected(self, transport, fut):             # <<<<<<<<<<<<<<
 *         exc = fut.exception()
 *         if exc is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer._UVStreamServer__on_ssl_connected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_14UVStreamServer_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_14UVStreamServer_4__reduce_cython__, "UVStreamServer.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_14UVStreamServer_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVStreamServer_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVStreamServer_4__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_14UVStreamServer_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_14UVStreamServer_4__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_14UVStreamServer_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_14UVStreamServer_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_14UVStreamServer_6__setstate_cython__, "UVStreamServer.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_14UVStreamServer_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVStreamServer_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVStreamServer_6__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_14UVStreamServer_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_14UVStreamServer_6__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_14UVStreamServer_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVStreamServer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/streamserver.pyx":124
 * 
 * 
 * cdef void __uv_streamserver_on_listen(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* handle,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___uv_streamserver_on_listen(uv_stream_t *__pyx_v_handle, int __pyx_v_status) {
  struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_stream = 0;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_t_4 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_streamserver_on_listen", 0);

  /* "uvloop/handles/streamserver.pyx":131
 *     # callback for uv_listen
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UVStream listen callback") == 0:
 *         return
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVStream listen callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(21, 131, __pyx_L1_error)

  /* "uvloop/handles/streamserver.pyx":132
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,
 *                             "UVStream listen callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_t_1 == 0);

  /* "uvloop/handles/streamserver.pyx":131
 *     # callback for uv_listen
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UVStream listen callback") == 0:
 *         return
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/streamserver.pyx":133
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,
 *                             "UVStream listen callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/streamserver.pyx":131
 *     # callback for uv_listen
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UVStream listen callback") == 0:
 *         return
 */
  }

  /* "uvloop/handles/streamserver.pyx":136
 * 
 *     cdef:
 *         UVStreamServer stream = <UVStreamServer> handle.data             # <<<<<<<<<<<<<<
 * 
 *     if status < 0:
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_stream = ((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/streamserver.pyx":138
 *         UVStreamServer stream = <UVStreamServer> handle.data
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_listen_errors_total += 1
 */
  __pyx_t_2 = (__pyx_v_status < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/streamserver.pyx":139
 * 
 *     if status < 0:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_listen_errors_total += 1
 * 
 */
    __pyx_t_2 = (UVLOOP_DEBUG != 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/streamserver.pyx":140
 *     if status < 0:
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_listen_errors_total += 1             # <<<<<<<<<<<<<<
 * 
 *         exc = convert_error(status)
 */
      __Pyx_INCREF((PyObject *)__pyx_v_stream->__pyx_base.__pyx_base._loop);
      __pyx_t_4 = __pyx_v_stream->__pyx_base.__pyx_base._loop;
      __pyx_t_4->_debug_stream_listen_errors_total = (__pyx_t_4->_debug_stream_listen_errors_total + 1);
      __Pyx_DECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/handles/streamserver.pyx":139
 * 
 *     if status < 0:
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             stream._loop._debug_stream_listen_errors_total += 1
 * 
 */
    }

    /* "uvloop/handles/streamserver.pyx":142
 *             stream._loop._debug_stream_listen_errors_total += 1
 * 
 *         exc = convert_error(status)             # <<<<<<<<<<<<<<
 *         stream._fatal_error(
 *             exc, False, "error status in uv_stream_t.listen callback")
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_exc = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/handles/streamserver.pyx":143
 * 
 *         exc = convert_error(status)
 *         stream._fatal_error(             # <<<<<<<<<<<<<<
 *             exc, False, "error status in uv_stream_t.listen callback")
 *         return
 */
    __pyx_t_5.__pyx_n = 1;
    __pyx_t_5.reason = __pyx_kp_u_error_status_in_uv_stream_t_list;
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_stream), __pyx_v_exc, Py_False, &__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/streamserver.pyx":145
 *         stream._fatal_error(
 *             exc, False, "error status in uv_stream_t.listen callback")
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    goto __pyx_L0;

    /* "uvloop/handles/streamserver.pyx":138
 *         UVStreamServer stream = <UVStreamServer> handle.data
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             stream._loop._debug_stream_listen_errors_total += 1
 */
  }

  /* "uvloop/handles/streamserver.pyx":147
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         stream._on_listen()
 *     except BaseException as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "uvloop/handles/streamserver.pyx":148
 * 
 *     try:
 *         stream._on_listen()             # <<<<<<<<<<<<<<
 *     except BaseException as exc:
 *         stream._error(exc, False)
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_14UVStreamServer__on_listen(__pyx_v_stream); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 148, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/streamserver.pyx":147
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         stream._on_listen()
 *     except BaseException as exc:
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/streamserver.pyx":149
 *     try:
 *         stream._on_listen()
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         stream._error(exc, False)
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.__uv_streamserver_on_listen", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_10, &__pyx_t_11) < 0) __PYX_ERR(21, 149, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_10);
      __pyx_v_exc = __pyx_t_10;
      /*try:*/ {

        /* "uvloop/handles/streamserver.pyx":150
 *         stream._on_listen()
 *     except BaseException as exc:
 *         stream._error(exc, False)             # <<<<<<<<<<<<<<
 */
        __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)__pyx_v_stream->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_stream), __pyx_v_exc, Py_False); if (unlikely(!__pyx_t_12)) __PYX_ERR(21, 150, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }

      /* "uvloop/handles/streamserver.pyx":149
 *     try:
 *         stream._on_listen()
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         stream._error(exc, False)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L18;
        }
        __pyx_L17_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF((PyObject *)__pyx_t_4); __pyx_t_4 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_9 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L8_except_error;
        }
        __pyx_L18:;
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L7_exception_handled;
    }
    goto __pyx_L8_except_error;

    /* "uvloop/handles/streamserver.pyx":147
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         stream._on_listen()
 *     except BaseException as exc:
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    __pyx_L11_try_end:;
  }

  /* "uvloop/handles/streamserver.pyx":124
 * 
 * 
 * cdef void __uv_streamserver_on_listen(             # <<<<<<<<<<<<<<
 *     uv.uv_stream_t* handle,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF((PyObject *)__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_streamserver_on_listen", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stream);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/tcp.pyx":1
 * cdef __tcp_init_uv_handle(UVStream handle, Loop loop, unsigned int flags):             # <<<<<<<<<<<<<<
 *     cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop___tcp_init_uv_handle(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_handle, CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, unsigned int __pyx_v_flags) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__tcp_init_uv_handle", 1);

  /* "uvloop/handles/tcp.pyx":4
 *     cdef int err
 * 
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_tcp_t))             # <<<<<<<<<<<<<<
 *     if handle._handle is NULL:
 *         handle._abort_init()
 */
  __pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_tcp_t))));

  /* "uvloop/handles/tcp.pyx":5
 * 
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_tcp_t))
 *     if handle._handle is NULL:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise MemoryError()
 */
  __pyx_t_1 = (__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/tcp.pyx":6
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_tcp_t))
 *     if handle._handle is NULL:
 *         handle._abort_init()             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_handle)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/tcp.pyx":7
 *     if handle._handle is NULL:
 *         handle._abort_init()
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     err = uv.uv_tcp_init_ex(handle._loop.uvloop,
 */
    PyErr_NoMemory(); __PYX_ERR(3, 7, __pyx_L1_error)

    /* "uvloop/handles/tcp.pyx":5
 * 
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_tcp_t))
 *     if handle._handle is NULL:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise MemoryError()
 */
  }

  /* "uvloop/handles/tcp.pyx":9
 *         raise MemoryError()
 * 
 *     err = uv.uv_tcp_init_ex(handle._loop.uvloop,             # <<<<<<<<<<<<<<
 *                             <uv.uv_tcp_t*>handle._handle,
 *                             flags)
 */
  __pyx_v_err = uv_tcp_init_ex(__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._loop->uvloop, ((uv_tcp_t *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_flags);

  /* "uvloop/handles/tcp.pyx":12
 *                             <uv.uv_tcp_t*>handle._handle,
 *                             flags)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise convert_error(err)
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/tcp.pyx":13
 *                             flags)
 *     if err < 0:
 *         handle._abort_init()             # <<<<<<<<<<<<<<
 *         raise convert_error(err)
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_handle)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/tcp.pyx":14
 *     if err < 0:
 *         handle._abort_init()
 *         raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *     handle._finish_init()
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(3, 14, __pyx_L1_error)

    /* "uvloop/handles/tcp.pyx":12
 *                             <uv.uv_tcp_t*>handle._handle,
 *                             flags)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise convert_error(err)
 */
  }

  /* "uvloop/handles/tcp.pyx":16
 *         raise convert_error(err)
 * 
 *     handle._finish_init()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_handle)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/tcp.pyx":1
 * cdef __tcp_init_uv_handle(UVStream handle, Loop loop, unsigned int flags):             # <<<<<<<<<<<<<<
 *     cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.__tcp_init_uv_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":19
 * 
 * 
 * cdef __tcp_bind(UVStream handle, system.sockaddr* addr, unsigned int flags):             # <<<<<<<<<<<<<<
 *     cdef int err
 *     err = uv.uv_tcp_bind(<uv.uv_tcp_t *>handle._handle,
 */

static PyObject *__pyx_f_6uvloop_4loop___tcp_bind(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_handle, struct sockaddr *__pyx_v_addr, unsigned int __pyx_v_flags) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__tcp_bind", 1);

  /* "uvloop/handles/tcp.pyx":21
 * cdef __tcp_bind(UVStream handle, system.sockaddr* addr, unsigned int flags):
 *     cdef int err
 *     err = uv.uv_tcp_bind(<uv.uv_tcp_t *>handle._handle,             # <<<<<<<<<<<<<<
 *                          addr, flags)
 *     if err < 0:
 */
  __pyx_v_err = uv_tcp_bind(((uv_tcp_t *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_addr, __pyx_v_flags);

  /* "uvloop/handles/tcp.pyx":23
 *     err = uv.uv_tcp_bind(<uv.uv_tcp_t *>handle._handle,
 *                          addr, flags)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(err)
 *         raise exc
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/tcp.pyx":24
 *                          addr, flags)
 *     if err < 0:
 *         exc = convert_error(err)             # <<<<<<<<<<<<<<
 *         raise exc
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/tcp.pyx":25
 *     if err < 0:
 *         exc = convert_error(err)
 *         raise exc             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(3, 25, __pyx_L1_error)

    /* "uvloop/handles/tcp.pyx":23
 *     err = uv.uv_tcp_bind(<uv.uv_tcp_t *>handle._handle,
 *                          addr, flags)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(err)
 *         raise exc
 */
  }

  /* "uvloop/handles/tcp.pyx":19
 * 
 * 
 * cdef __tcp_bind(UVStream handle, system.sockaddr* addr, unsigned int flags):             # <<<<<<<<<<<<<<
 *     cdef int err
 *     err = uv.uv_tcp_bind(<uv.uv_tcp_t *>handle._handle,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.__tcp_bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":28
 * 
 * 
 * cdef __tcp_open(UVStream handle, int sockfd):             # <<<<<<<<<<<<<<
 *     cdef int err
 *     err = uv.uv_tcp_open(<uv.uv_tcp_t *>handle._handle,
 */

static PyObject *__pyx_f_6uvloop_4loop___tcp_open(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_handle, int __pyx_v_sockfd) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__tcp_open", 1);

  /* "uvloop/handles/tcp.pyx":30
 * cdef __tcp_open(UVStream handle, int sockfd):
 *     cdef int err
 *     err = uv.uv_tcp_open(<uv.uv_tcp_t *>handle._handle,             # <<<<<<<<<<<<<<
 *                          <uv.uv_os_sock_t>sockfd)
 *     if err < 0:
 */
  __pyx_v_err = uv_tcp_open(((uv_tcp_t *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle), ((uv_os_sock_t)__pyx_v_sockfd));

  /* "uvloop/handles/tcp.pyx":32
 *     err = uv.uv_tcp_open(<uv.uv_tcp_t *>handle._handle,
 *                          <uv.uv_os_sock_t>sockfd)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(err)
 *         raise exc
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/tcp.pyx":33
 *                          <uv.uv_os_sock_t>sockfd)
 *     if err < 0:
 *         exc = convert_error(err)             # <<<<<<<<<<<<<<
 *         raise exc
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/tcp.pyx":34
 *     if err < 0:
 *         exc = convert_error(err)
 *         raise exc             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(3, 34, __pyx_L1_error)

    /* "uvloop/handles/tcp.pyx":32
 *     err = uv.uv_tcp_open(<uv.uv_tcp_t *>handle._handle,
 *                          <uv.uv_os_sock_t>sockfd)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(err)
 *         raise exc
 */
  }

  /* "uvloop/handles/tcp.pyx":28
 * 
 * 
 * cdef __tcp_open(UVStream handle, int sockfd):             # <<<<<<<<<<<<<<
 *     cdef int err
 *     err = uv.uv_tcp_open(<uv.uv_tcp_t *>handle._handle,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.__tcp_open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":37
 * 
 * 
 * cdef __tcp_get_socket(UVSocketHandle handle):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int buf_len = sizeof(system.sockaddr_storage)
 */

static PyObject *__pyx_f_6uvloop_4loop___tcp_get_socket(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_handle) {
  int __pyx_v_buf_len;
  int __pyx_v_fileno;
  int __pyx_v_err;
  struct sockaddr_storage __pyx_v_buf;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__tcp_get_socket", 1);

  /* "uvloop/handles/tcp.pyx":39
 * cdef __tcp_get_socket(UVSocketHandle handle):
 *     cdef:
 *         int buf_len = sizeof(system.sockaddr_storage)             # <<<<<<<<<<<<<<
 *         int fileno
 *         int err
 */
  __pyx_v_buf_len = (sizeof(struct sockaddr_storage));

  /* "uvloop/handles/tcp.pyx":44
 *         system.sockaddr_storage buf
 * 
 *     fileno = handle._fileno()             # <<<<<<<<<<<<<<
 * 
 *     err = uv.uv_tcp_getsockname(<uv.uv_tcp_t*>handle._handle,
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_fileno(__pyx_v_handle); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fileno = __pyx_t_2;

  /* "uvloop/handles/tcp.pyx":46
 *     fileno = handle._fileno()
 * 
 *     err = uv.uv_tcp_getsockname(<uv.uv_tcp_t*>handle._handle,             # <<<<<<<<<<<<<<
 *                                 <system.sockaddr*>&buf,
 *                                 &buf_len)
 */
  __pyx_v_err = uv_tcp_getsockname(((uv_tcp_t *)__pyx_v_handle->__pyx_base._handle), ((struct sockaddr *)(&__pyx_v_buf)), (&__pyx_v_buf_len));

  /* "uvloop/handles/tcp.pyx":49
 *                                 <system.sockaddr*>&buf,
 *                                 &buf_len)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         raise convert_error(err)
 * 
 */
  __pyx_t_3 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/handles/tcp.pyx":50
 *                                 &buf_len)
 *     if err < 0:
 *         raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *     return PseudoSocket(buf.ss_family, uv.SOCK_STREAM, 0, fileno)
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 50, __pyx_L1_error)

    /* "uvloop/handles/tcp.pyx":49
 *                                 <system.sockaddr*>&buf,
 *                                 &buf_len)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         raise convert_error(err)
 * 
 */
  }

  /* "uvloop/handles/tcp.pyx":52
 *         raise convert_error(err)
 * 
 *     return PseudoSocket(buf.ss_family, uv.SOCK_STREAM, 0, fileno)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_short(__pyx_v_buf.ss_family); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_From_int(SOCK_STREAM); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_fileno); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(3, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(3, 52, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4)) __PYX_ERR(3, 52, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_int_0)) __PYX_ERR(3, 52, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_5)) __PYX_ERR(3, 52, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/tcp.pyx":37
 * 
 * 
 * cdef __tcp_get_socket(UVSocketHandle handle):             # <<<<<<<<<<<<<<
 *     cdef:
 *         int buf_len = sizeof(system.sockaddr_storage)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.__tcp_get_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":59
 * 
 *     @staticmethod
 *     cdef TCPServer new(Loop loop, object protocol_factory, Server server,             # <<<<<<<<<<<<<<
 *                        unsigned int flags,
 *                        object backlog,
 */

static struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_f_6uvloop_4loop_9TCPServer_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol_factory, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, unsigned int __pyx_v_flags, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/tcp.pyx":67
 * 
 *         cdef TCPServer handle
 *         handle = TCPServer.__new__(TCPServer)             # <<<<<<<<<<<<<<
 *         handle._init(loop, protocol_factory, server, backlog,
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_TCPServer(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_TCPServer), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 67, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_TCPServer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":68
 *         cdef TCPServer handle
 *         handle = TCPServer.__new__(TCPServer)
 *         handle._init(loop, protocol_factory, server, backlog,             # <<<<<<<<<<<<<<
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 *         __tcp_init_uv_handle(<UVStream>handle, loop, flags)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_handle), __pyx_v_loop, __pyx_v_protocol_factory, __pyx_v_server, __pyx_v_backlog, __pyx_v_ssl, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":70
 *         handle._init(loop, protocol_factory, server, backlog,
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 *         __tcp_init_uv_handle(<UVStream>handle, loop, flags)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_init_uv_handle(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop, __pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":71
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 *         __tcp_init_uv_handle(<UVStream>handle, loop, flags)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/tcp.pyx":59
 * 
 *     @staticmethod
 *     cdef TCPServer new(Loop loop, object protocol_factory, Server server,             # <<<<<<<<<<<<<<
 *                        unsigned int flags,
 *                        object backlog,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TCPServer.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":73
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_9TCPServer__new_socket(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/tcp.pyx":74
 * 
 *     cdef _new_socket(self):
 *         return __tcp_get_socket(<UVSocketHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _open(self, int sockfd):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/tcp.pyx":73
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TCPServer._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":76
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         try:
 */

static PyObject *__pyx_f_6uvloop_4loop_9TCPServer__open(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, int __pyx_v_sockfd) {
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open", 1);

  /* "uvloop/handles/tcp.pyx":77
 * 
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         try:
 *             __tcp_open(<UVStream>self, sockfd)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":78
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()
 *         try:             # <<<<<<<<<<<<<<
 *             __tcp_open(<UVStream>self, sockfd)
 *         except Exception as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/handles/tcp.pyx":79
 *         self._ensure_alive()
 *         try:
 *             __tcp_open(<UVStream>self, sockfd)             # <<<<<<<<<<<<<<
 *         except Exception as exc:
 *             self._fatal_error(exc, True)
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_open(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_sockfd); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 79, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/tcp.pyx":78
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()
 *         try:             # <<<<<<<<<<<<<<
 *             __tcp_open(<UVStream>self, sockfd)
 *         except Exception as exc:
 */
    }

    /* "uvloop/handles/tcp.pyx":83
 *             self._fatal_error(exc, True)
 *         else:
 *             self._mark_as_open()             # <<<<<<<<<<<<<<
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 */
    /*else:*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._mark_as_open(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 83, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/tcp.pyx":80
 *         try:
 *             __tcp_open(<UVStream>self, sockfd)
 *         except Exception as exc:             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *         else:
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("uvloop.loop.TCPServer._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 80, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_exc = __pyx_t_6;
      /*try:*/ {

        /* "uvloop/handles/tcp.pyx":81
 *             __tcp_open(<UVStream>self, sockfd)
 *         except Exception as exc:
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *         else:
 *             self._mark_as_open()
 */
        __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 81, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }

      /* "uvloop/handles/tcp.pyx":80
 *         try:
 *             __tcp_open(<UVStream>self, sockfd)
 *         except Exception as exc:             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *         else:
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L15;
        }
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __pyx_t_5 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          }
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
          goto __pyx_L5_except_error;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/handles/tcp.pyx":78
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()
 *         try:             # <<<<<<<<<<<<<<
 *             __tcp_open(<UVStream>self, sockfd)
 *         except Exception as exc:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "uvloop/handles/tcp.pyx":76
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.TCPServer._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":85
 *             self._mark_as_open()
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         try:
 */

static PyObject *__pyx_f_6uvloop_4loop_9TCPServer_bind(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, struct sockaddr *__pyx_v_addr, struct __pyx_opt_args_6uvloop_4loop_9TCPServer_bind *__pyx_optional_args) {
  unsigned int __pyx_v_flags = ((unsigned int)0);
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_flags = __pyx_optional_args->flags;
    }
  }

  /* "uvloop/handles/tcp.pyx":86
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         try:
 *             __tcp_bind(<UVStream>self, addr, flags)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":87
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 *         self._ensure_alive()
 *         try:             # <<<<<<<<<<<<<<
 *             __tcp_bind(<UVStream>self, addr, flags)
 *         except Exception as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/handles/tcp.pyx":88
 *         self._ensure_alive()
 *         try:
 *             __tcp_bind(<UVStream>self, addr, flags)             # <<<<<<<<<<<<<<
 *         except Exception as exc:
 *             self._fatal_error(exc, True)
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_bind(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_addr, __pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 88, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/tcp.pyx":87
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 *         self._ensure_alive()
 *         try:             # <<<<<<<<<<<<<<
 *             __tcp_bind(<UVStream>self, addr, flags)
 *         except Exception as exc:
 */
    }

    /* "uvloop/handles/tcp.pyx":92
 *             self._fatal_error(exc, True)
 *         else:
 *             self._mark_as_open()             # <<<<<<<<<<<<<<
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,
 */
    /*else:*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._mark_as_open(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 92, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/tcp.pyx":89
 *         try:
 *             __tcp_bind(<UVStream>self, addr, flags)
 *         except Exception as exc:             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *         else:
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("uvloop.loop.TCPServer.bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 89, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_exc = __pyx_t_6;
      /*try:*/ {

        /* "uvloop/handles/tcp.pyx":90
 *             __tcp_bind(<UVStream>self, addr, flags)
 *         except Exception as exc:
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *         else:
 *             self._mark_as_open()
 */
        __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 90, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }

      /* "uvloop/handles/tcp.pyx":89
 *         try:
 *             __tcp_bind(<UVStream>self, addr, flags)
 *         except Exception as exc:             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *         else:
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L15;
        }
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __pyx_t_5 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          }
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
          __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
          goto __pyx_L5_except_error;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/handles/tcp.pyx":87
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 *         self._ensure_alive()
 *         try:             # <<<<<<<<<<<<<<
 *             __tcp_bind(<UVStream>self, addr, flags)
 *         except Exception as exc:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L8_try_end:;
  }

  /* "uvloop/handles/tcp.pyx":85
 *             self._mark_as_open()
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.TCPServer.bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":94
 *             self._mark_as_open()
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,             # <<<<<<<<<<<<<<
 *                                       object context):
 *         cdef TCPTransport tr
 */

static struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_f_6uvloop_4loop_9TCPServer__make_new_transport(struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, PyObject *__pyx_v_protocol, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_tr = 0;
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_new_transport", 1);

  /* "uvloop/handles/tcp.pyx":97
 *                                       object context):
 *         cdef TCPTransport tr
 *         tr = TCPTransport.new(self._loop, protocol, self._server, waiter,             # <<<<<<<<<<<<<<
 *                               context)
 *         return <UVStream>tr
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_v_self->__pyx_base._server);
  __Pyx_INCREF(__pyx_t_2);

  /* "uvloop/handles/tcp.pyx":98
 *         cdef TCPTransport tr
 *         tr = TCPTransport.new(self._loop, protocol, self._server, waiter,
 *                               context)             # <<<<<<<<<<<<<<
 *         return <UVStream>tr
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_6uvloop_4loop_12TCPTransport_new(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1), __pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_t_2), __pyx_v_waiter, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tr = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/tcp.pyx":99
 *         tr = TCPTransport.new(self._loop, protocol, self._server, waiter,
 *                               context)
 *         return <UVStream>tr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_tr));
  __pyx_r = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_tr);
  goto __pyx_L0;

  /* "uvloop/handles/tcp.pyx":94
 *             self._mark_as_open()
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,             # <<<<<<<<<<<<<<
 *                                       object context):
 *         cdef TCPTransport tr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.TCPServer._make_new_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9TCPServer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9TCPServer___reduce_cython__, "TCPServer.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9TCPServer_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9TCPServer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9TCPServer___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9TCPServer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_9TCPServer___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_TCPServer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9TCPServer___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.TCPServer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9TCPServer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9TCPServer_2__setstate_cython__, "TCPServer.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9TCPServer_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9TCPServer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9TCPServer_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9TCPServer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.TCPServer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_9TCPServer_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_TCPServer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9TCPServer_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.TCPServer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":106
 * 
 *     @staticmethod
 *     cdef TCPTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                           object waiter, object context):
 * 
 */

static struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_f_6uvloop_4loop_12TCPTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/tcp.pyx":110
 * 
 *         cdef TCPTransport handle
 *         handle = TCPTransport.__new__(TCPTransport)             # <<<<<<<<<<<<<<
 *         handle._init(loop, protocol, server, waiter, context)
 *         __tcp_init_uv_handle(<UVStream>handle, loop, uv.AF_UNSPEC)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_TCPTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_TCPTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 110, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":111
 *         cdef TCPTransport handle
 *         handle = TCPTransport.__new__(TCPTransport)
 *         handle._init(loop, protocol, server, waiter, context)             # <<<<<<<<<<<<<<
 *         __tcp_init_uv_handle(<UVStream>handle, loop, uv.AF_UNSPEC)
 *         handle.__peername_set = 0
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop, __pyx_v_protocol, __pyx_v_server, __pyx_v_waiter, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":112
 *         handle = TCPTransport.__new__(TCPTransport)
 *         handle._init(loop, protocol, server, waiter, context)
 *         __tcp_init_uv_handle(<UVStream>handle, loop, uv.AF_UNSPEC)             # <<<<<<<<<<<<<<
 *         handle.__peername_set = 0
 *         handle.__sockname_set = 0
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_init_uv_handle(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop, AF_UNSPEC); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":113
 *         handle._init(loop, protocol, server, waiter, context)
 *         __tcp_init_uv_handle(<UVStream>handle, loop, uv.AF_UNSPEC)
 *         handle.__peername_set = 0             # <<<<<<<<<<<<<<
 *         handle.__sockname_set = 0
 *         handle._set_nodelay()
 */
  __pyx_v_handle->_TCPTransport__peername_set = 0;

  /* "uvloop/handles/tcp.pyx":114
 *         __tcp_init_uv_handle(<UVStream>handle, loop, uv.AF_UNSPEC)
 *         handle.__peername_set = 0
 *         handle.__sockname_set = 0             # <<<<<<<<<<<<<<
 *         handle._set_nodelay()
 *         return handle
 */
  __pyx_v_handle->_TCPTransport__sockname_set = 0;

  /* "uvloop/handles/tcp.pyx":115
 *         handle.__peername_set = 0
 *         handle.__sockname_set = 0
 *         handle._set_nodelay()             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_set_nodelay(__pyx_v_handle); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":116
 *         handle.__sockname_set = 0
 *         handle._set_nodelay()
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_nodelay(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/tcp.pyx":106
 * 
 *     @staticmethod
 *     cdef TCPTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                           object waiter, object context):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":118
 *         return handle
 * 
 *     cdef _set_nodelay(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__set_nodelay(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_nodelay", 1);

  /* "uvloop/handles/tcp.pyx":120
 *     cdef _set_nodelay(self):
 *         cdef int err
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         err = uv.uv_tcp_nodelay(<uv.uv_tcp_t*>self._handle, 1)
 *         if err < 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":121
 *         cdef int err
 *         self._ensure_alive()
 *         err = uv.uv_tcp_nodelay(<uv.uv_tcp_t*>self._handle, 1)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
  __pyx_v_err = uv_tcp_nodelay(((uv_tcp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._handle), 1);

  /* "uvloop/handles/tcp.pyx":122
 *         self._ensure_alive()
 *         err = uv.uv_tcp_nodelay(<uv.uv_tcp_t*>self._handle, 1)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/tcp.pyx":123
 *         err = uv.uv_tcp_nodelay(<uv.uv_tcp_t*>self._handle, 1)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *     cdef _call_connection_made(self):
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(3, 123, __pyx_L1_error)

    /* "uvloop/handles/tcp.pyx":122
 *         self._ensure_alive()
 *         err = uv.uv_tcp_nodelay(<uv.uv_tcp_t*>self._handle, 1)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  }

  /* "uvloop/handles/tcp.pyx":118
 *         return handle
 * 
 *     cdef _set_nodelay(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport._set_nodelay", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":125
 *             raise convert_error(err)
 * 
 *     cdef _call_connection_made(self):             # <<<<<<<<<<<<<<
 *         # asyncio saves peername & sockname when transports are instantiated,
 *         # so that they're accessible even after the transport is closed.
 */

static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__call_connection_made(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self) {
  int __pyx_v_err;
  int __pyx_v_buf_len;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_connection_made", 1);

  /* "uvloop/handles/tcp.pyx":135
 *             int buf_len
 * 
 *         buf_len = sizeof(system.sockaddr_storage)             # <<<<<<<<<<<<<<
 *         err = uv.uv_tcp_getsockname(<uv.uv_tcp_t*>self._handle,
 *                                     <system.sockaddr*>&self.__sockname,
 */
  __pyx_v_buf_len = (sizeof(struct sockaddr_storage));

  /* "uvloop/handles/tcp.pyx":136
 * 
 *         buf_len = sizeof(system.sockaddr_storage)
 *         err = uv.uv_tcp_getsockname(<uv.uv_tcp_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                     <system.sockaddr*>&self.__sockname,
 *                                     &buf_len)
 */
  __pyx_v_err = uv_tcp_getsockname(((uv_tcp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._handle), ((struct sockaddr *)(&__pyx_v_self->_TCPTransport__sockname)), (&__pyx_v_buf_len));

  /* "uvloop/handles/tcp.pyx":139
 *                                     <system.sockaddr*>&self.__sockname,
 *                                     &buf_len)
 *         if err >= 0:             # <<<<<<<<<<<<<<
 *             # Ignore errors, this is an optional thing.
 *             # If something serious is going on, the transport
 */
  __pyx_t_1 = (__pyx_v_err >= 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/tcp.pyx":144
 *             # will crash later (in roughly the same way how
 *             # an asyncio transport would.)
 *             self.__sockname_set = 1             # <<<<<<<<<<<<<<
 * 
 *         buf_len = sizeof(system.sockaddr_storage)
 */
    __pyx_v_self->_TCPTransport__sockname_set = 1;

    /* "uvloop/handles/tcp.pyx":139
 *                                     <system.sockaddr*>&self.__sockname,
 *                                     &buf_len)
 *         if err >= 0:             # <<<<<<<<<<<<<<
 *             # Ignore errors, this is an optional thing.
 *             # If something serious is going on, the transport
 */
  }

  /* "uvloop/handles/tcp.pyx":146
 *             self.__sockname_set = 1
 * 
 *         buf_len = sizeof(system.sockaddr_storage)             # <<<<<<<<<<<<<<
 *         err = uv.uv_tcp_getpeername(<uv.uv_tcp_t*>self._handle,
 *                                     <system.sockaddr*>&self.__peername,
 */
  __pyx_v_buf_len = (sizeof(struct sockaddr_storage));

  /* "uvloop/handles/tcp.pyx":147
 * 
 *         buf_len = sizeof(system.sockaddr_storage)
 *         err = uv.uv_tcp_getpeername(<uv.uv_tcp_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                     <system.sockaddr*>&self.__peername,
 *                                     &buf_len)
 */
  __pyx_v_err = uv_tcp_getpeername(((uv_tcp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._handle), ((struct sockaddr *)(&__pyx_v_self->_TCPTransport__peername)), (&__pyx_v_buf_len));

  /* "uvloop/handles/tcp.pyx":150
 *                                     <system.sockaddr*>&self.__peername,
 *                                     &buf_len)
 *         if err >= 0:             # <<<<<<<<<<<<<<
 *             # Same as few lines above -- we don't really care
 *             # about error case here.
 */
  __pyx_t_1 = (__pyx_v_err >= 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/tcp.pyx":153
 *             # Same as few lines above -- we don't really care
 *             # about error case here.
 *             self.__peername_set = 1             # <<<<<<<<<<<<<<
 * 
 *         UVBaseTransport._call_connection_made(self)
 */
    __pyx_v_self->_TCPTransport__peername_set = 1;

    /* "uvloop/handles/tcp.pyx":150
 *                                     <system.sockaddr*>&self.__peername,
 *                                     &buf_len)
 *         if err >= 0:             # <<<<<<<<<<<<<<
 *             # Same as few lines above -- we don't really care
 *             # about error case here.
 */
  }

  /* "uvloop/handles/tcp.pyx":155
 *             self.__peername_set = 1
 * 
 *         UVBaseTransport._call_connection_made(self)             # <<<<<<<<<<<<<<
 * 
 *     def get_extra_info(self, name, default=None):
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_15UVBaseTransport__call_connection_made(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/tcp.pyx":125
 *             raise convert_error(err)
 * 
 *     cdef _call_connection_made(self):             # <<<<<<<<<<<<<<
 *         # asyncio saves peername & sockname when transports are instantiated,
 *         # so that they're accessible even after the transport is closed.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport._call_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":157
 *         UVBaseTransport._call_connection_made(self)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if name == 'sockname':
 *             if self.__sockname_set:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12TCPTransport_1get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12TCPTransport_get_extra_info, "TCPTransport.get_extra_info(self, name, default=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12TCPTransport_1get_extra_info = {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12TCPTransport_1get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12TCPTransport_get_extra_info};
static PyObject *__pyx_pw_6uvloop_4loop_12TCPTransport_1get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_default = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_extra_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_default,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name_2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 157, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_default);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 157, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_extra_info") < 0)) __PYX_ERR(3, 157, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_extra_info", 0, 1, 2, __pyx_nargs); __PYX_ERR(3, 157, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12TCPTransport_get_extra_info(((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_v_self), __pyx_v_name, __pyx_v_default);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12TCPTransport_get_extra_info(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_extra_info", 1);

  /* "uvloop/handles/tcp.pyx":158
 * 
 *     def get_extra_info(self, name, default=None):
 *         if name == 'sockname':             # <<<<<<<<<<<<<<
 *             if self.__sockname_set:
 *                 return __convert_sockaddr_to_pyaddr(
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_n_u_sockname, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 158, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/tcp.pyx":159
 *     def get_extra_info(self, name, default=None):
 *         if name == 'sockname':
 *             if self.__sockname_set:             # <<<<<<<<<<<<<<
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__sockname)
 */
    if (__pyx_v_self->_TCPTransport__sockname_set) {

      /* "uvloop/handles/tcp.pyx":160
 *         if name == 'sockname':
 *             if self.__sockname_set:
 *                 return __convert_sockaddr_to_pyaddr(             # <<<<<<<<<<<<<<
 *                     <system.sockaddr*>&self.__sockname)
 *         elif name == 'peername':
 */
      __Pyx_XDECREF(__pyx_r);

      /* "uvloop/handles/tcp.pyx":161
 *             if self.__sockname_set:
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__sockname)             # <<<<<<<<<<<<<<
 *         elif name == 'peername':
 *             if self.__peername_set:
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(((struct sockaddr *)(&__pyx_v_self->_TCPTransport__sockname))); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "uvloop/handles/tcp.pyx":159
 *     def get_extra_info(self, name, default=None):
 *         if name == 'sockname':
 *             if self.__sockname_set:             # <<<<<<<<<<<<<<
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__sockname)
 */
    }

    /* "uvloop/handles/tcp.pyx":158
 * 
 *     def get_extra_info(self, name, default=None):
 *         if name == 'sockname':             # <<<<<<<<<<<<<<
 *             if self.__sockname_set:
 *                 return __convert_sockaddr_to_pyaddr(
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/tcp.pyx":162
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__sockname)
 *         elif name == 'peername':             # <<<<<<<<<<<<<<
 *             if self.__peername_set:
 *                 return __convert_sockaddr_to_pyaddr(
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_n_u_peername, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(3, 162, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/tcp.pyx":163
 *                     <system.sockaddr*>&self.__sockname)
 *         elif name == 'peername':
 *             if self.__peername_set:             # <<<<<<<<<<<<<<
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__peername)
 */
    if (__pyx_v_self->_TCPTransport__peername_set) {

      /* "uvloop/handles/tcp.pyx":164
 *         elif name == 'peername':
 *             if self.__peername_set:
 *                 return __convert_sockaddr_to_pyaddr(             # <<<<<<<<<<<<<<
 *                     <system.sockaddr*>&self.__peername)
 *         return super().get_extra_info(name, default)
 */
      __Pyx_XDECREF(__pyx_r);

      /* "uvloop/handles/tcp.pyx":165
 *             if self.__peername_set:
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__peername)             # <<<<<<<<<<<<<<
 *         return super().get_extra_info(name, default)
 * 
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(((struct sockaddr *)(&__pyx_v_self->_TCPTransport__peername))); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "uvloop/handles/tcp.pyx":163
 *                     <system.sockaddr*>&self.__sockname)
 *         elif name == 'peername':
 *             if self.__peername_set:             # <<<<<<<<<<<<<<
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__peername)
 */
    }

    /* "uvloop/handles/tcp.pyx":162
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__sockname)
 *         elif name == 'peername':             # <<<<<<<<<<<<<<
 *             if self.__peername_set:
 *                 return __convert_sockaddr_to_pyaddr(
 */
  }
  __pyx_L3:;

  /* "uvloop/handles/tcp.pyx":166
 *                 return __convert_sockaddr_to_pyaddr(
 *                     <system.sockaddr*>&self.__peername)
 *         return super().get_extra_info(name, default)             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)__pyx_ptype_6uvloop_4loop_TCPTransport);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_6uvloop_4loop_TCPTransport);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_6uvloop_4loop_TCPTransport))) __PYX_ERR(3, 166, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(3, 166, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get_extra_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_name, __pyx_v_default};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/tcp.pyx":157
 *         UVBaseTransport._call_connection_made(self)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if name == 'sockname':
 *             if self.__sockname_set:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":168
 *         return super().get_extra_info(name, default)
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__new_socket(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/tcp.pyx":169
 * 
 *     cdef _new_socket(self):
 *         return __tcp_get_socket(<UVSocketHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/tcp.pyx":168
 *         return super().get_extra_info(name, default)
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":171
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         __tcp_bind(<UVStream>self, addr, flags)
 */

static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport_bind(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr, struct __pyx_opt_args_6uvloop_4loop_12TCPTransport_bind *__pyx_optional_args) {
  unsigned int __pyx_v_flags = ((unsigned int)0);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_flags = __pyx_optional_args->flags;
    }
  }

  /* "uvloop/handles/tcp.pyx":172
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         __tcp_bind(<UVStream>self, addr, flags)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":173
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):
 *         self._ensure_alive()
 *         __tcp_bind(<UVStream>self, addr, flags)             # <<<<<<<<<<<<<<
 * 
 *     cdef _open(self, int sockfd):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_bind(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_addr, __pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":171
 *         return __tcp_get_socket(<UVSocketHandle>self)
 * 
 *     cdef bind(self, system.sockaddr* addr, unsigned int flags=0):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         __tcp_bind(<UVStream>self, addr, flags)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":175
 *         __tcp_bind(<UVStream>self, addr, flags)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         __tcp_open(<UVStream>self, sockfd)
 */

static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport__open(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, int __pyx_v_sockfd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open", 1);

  /* "uvloop/handles/tcp.pyx":176
 * 
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         __tcp_open(<UVStream>self, sockfd)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":177
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()
 *         __tcp_open(<UVStream>self, sockfd)             # <<<<<<<<<<<<<<
 * 
 *     cdef connect(self, system.sockaddr* addr):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___tcp_open(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_sockfd); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":175
 *         __tcp_bind(<UVStream>self, addr, flags)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         __tcp_open(<UVStream>self, sockfd)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":179
 *         __tcp_open(<UVStream>self, sockfd)
 * 
 *     cdef connect(self, system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *         cdef _TCPConnectRequest req
 *         req = _TCPConnectRequest(self._loop, self)
 */

static PyObject *__pyx_f_6uvloop_4loop_12TCPTransport_connect(struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr) {
  struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect", 1);

  /* "uvloop/handles/tcp.pyx":181
 *     cdef connect(self, system.sockaddr* addr):
 *         cdef _TCPConnectRequest req
 *         req = _TCPConnectRequest(self._loop, self)             # <<<<<<<<<<<<<<
 *         req.connect(addr)
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._loop);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._loop))) __PYX_ERR(3, 181, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(3, 181, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop__TCPConnectRequest), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_req = ((struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/handles/tcp.pyx":182
 *         cdef _TCPConnectRequest req
 *         req = _TCPConnectRequest(self._loop, self)
 *         req.connect(addr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop__TCPConnectRequest *)__pyx_v_req->__pyx_base.__pyx_vtab)->connect(__pyx_v_req, __pyx_v_addr); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/tcp.pyx":179
 *         __tcp_open(<UVStream>self, sockfd)
 * 
 *     cdef connect(self, system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *         cdef _TCPConnectRequest req
 *         req = _TCPConnectRequest(self._loop, self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12TCPTransport_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12TCPTransport_2__reduce_cython__, "TCPTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12TCPTransport_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12TCPTransport_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12TCPTransport_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_12TCPTransport_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12TCPTransport_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12TCPTransport_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12TCPTransport_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12TCPTransport_4__setstate_cython__, "TCPTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12TCPTransport_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12TCPTransport_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12TCPTransport_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_12TCPTransport_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12TCPTransport_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12TCPTransport_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.TCPTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":190
 *         uv.uv_connect_t _req_data
 * 
 *     def __cinit__(self, loop, transport):             # <<<<<<<<<<<<<<
 *         self.request = <uv.uv_req_t*>&self._req_data
 *         self.request.data = <void*>self
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_18_TCPConnectRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_18_TCPConnectRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_loop = 0;
  PyObject *__pyx_v_transport = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,&__pyx_n_s_transport,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 190, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_transport)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 190, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(3, 190, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(3, 190, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_loop = values[0];
    __pyx_v_transport = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, __pyx_nargs); __PYX_ERR(3, 190, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._TCPConnectRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18_TCPConnectRequest___cinit__(((struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)__pyx_v_self), __pyx_v_loop, __pyx_v_transport);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_18_TCPConnectRequest___cinit__(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_loop, PyObject *__pyx_v_transport) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/tcp.pyx":191
 * 
 *     def __cinit__(self, loop, transport):
 *         self.request = <uv.uv_req_t*>&self._req_data             # <<<<<<<<<<<<<<
 *         self.request.data = <void*>self
 *         self.transport = transport
 */
  __pyx_v_self->__pyx_base.request = ((uv_req_t *)(&__pyx_v_self->_req_data));

  /* "uvloop/handles/tcp.pyx":192
 *     def __cinit__(self, loop, transport):
 *         self.request = <uv.uv_req_t*>&self._req_data
 *         self.request.data = <void*>self             # <<<<<<<<<<<<<<
 *         self.transport = transport
 * 
 */
  __pyx_v_self->__pyx_base.request->data = ((void *)__pyx_v_self);

  /* "uvloop/handles/tcp.pyx":193
 *         self.request = <uv.uv_req_t*>&self._req_data
 *         self.request.data = <void*>self
 *         self.transport = transport             # <<<<<<<<<<<<<<
 * 
 *     cdef connect(self, system.sockaddr* addr):
 */
  if (!(likely(((__pyx_v_transport) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_transport, __pyx_ptype_6uvloop_4loop_TCPTransport))))) __PYX_ERR(3, 193, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_transport;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->transport);
  __Pyx_DECREF((PyObject *)__pyx_v_self->transport);
  __pyx_v_self->transport = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":190
 *         uv.uv_connect_t _req_data
 * 
 *     def __cinit__(self, loop, transport):             # <<<<<<<<<<<<<<
 *         self.request = <uv.uv_req_t*>&self._req_data
 *         self.request.data = <void*>self
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._TCPConnectRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":195
 *         self.transport = transport
 * 
 *     cdef connect(self, system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         err = uv.uv_tcp_connect(<uv.uv_connect_t*>self.request,
 */

static PyObject *__pyx_f_6uvloop_4loop_18_TCPConnectRequest_connect(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self, struct sockaddr *__pyx_v_addr) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect", 1);

  /* "uvloop/handles/tcp.pyx":197
 *     cdef connect(self, system.sockaddr* addr):
 *         cdef int err
 *         err = uv.uv_tcp_connect(<uv.uv_connect_t*>self.request,             # <<<<<<<<<<<<<<
 *                                 <uv.uv_tcp_t*>self.transport._handle,
 *                                 addr,
 */
  __pyx_v_err = uv_tcp_connect(((uv_connect_t *)__pyx_v_self->__pyx_base.request), ((uv_tcp_t *)__pyx_v_self->transport->__pyx_base.__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_addr, __pyx_f_6uvloop_4loop___tcp_connect_callback);

  /* "uvloop/handles/tcp.pyx":201
 *                                 addr,
 *                                 __tcp_connect_callback)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self.on_done()
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/tcp.pyx":202
 *                                 __tcp_connect_callback)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self.on_done()
 *             raise exc
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/tcp.pyx":203
 *         if err < 0:
 *             exc = convert_error(err)
 *             self.on_done()             # <<<<<<<<<<<<<<
 *             raise exc
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop__TCPConnectRequest *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/tcp.pyx":204
 *             exc = convert_error(err)
 *             self.on_done()
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(3, 204, __pyx_L1_error)

    /* "uvloop/handles/tcp.pyx":201
 *                                 addr,
 *                                 __tcp_connect_callback)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self.on_done()
 */
  }

  /* "uvloop/handles/tcp.pyx":195
 *         self.transport = transport
 * 
 *     cdef connect(self, system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         err = uv.uv_tcp_connect(<uv.uv_connect_t*>self.request,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop._TCPConnectRequest.connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18_TCPConnectRequest_2__reduce_cython__, "_TCPConnectRequest.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18_TCPConnectRequest_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18_TCPConnectRequest_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18_TCPConnectRequest_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18_TCPConnectRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._TCPConnectRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18_TCPConnectRequest_4__setstate_cython__, "_TCPConnectRequest.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18_TCPConnectRequest_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18_TCPConnectRequest_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._TCPConnectRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18_TCPConnectRequest_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18_TCPConnectRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._TCPConnectRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/tcp.pyx":207
 * 
 * 
 * cdef void __tcp_connect_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_connect_t* req,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___tcp_connect_callback(uv_connect_t *__pyx_v_req, int __pyx_v_status) {
  struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *__pyx_v_wrapper = 0;
  struct __pyx_obj_6uvloop_4loop_TCPTransport *__pyx_v_transport = 0;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  char const *__pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__tcp_connect_callback", 0);

  /* "uvloop/handles/tcp.pyx":215
 *         TCPTransport transport
 * 
 *     wrapper = <_TCPConnectRequest> req.data             # <<<<<<<<<<<<<<
 *     transport = wrapper.transport
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_req->data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":216
 * 
 *     wrapper = <_TCPConnectRequest> req.data
 *     transport = wrapper.transport             # <<<<<<<<<<<<<<
 * 
 *     if status < 0:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_wrapper->transport);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_transport = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/tcp.pyx":218
 *     transport = wrapper.transport
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *     else:
 */
  __pyx_t_2 = (__pyx_v_status < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/tcp.pyx":219
 * 
 *     if status < 0:
 *         exc = convert_error(status)             # <<<<<<<<<<<<<<
 *     else:
 *         exc = None
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/tcp.pyx":218
 *     transport = wrapper.transport
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/tcp.pyx":221
 *         exc = convert_error(status)
 *     else:
 *         exc = None             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_exc = Py_None;
  }
  __pyx_L3:;

  /* "uvloop/handles/tcp.pyx":223
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "uvloop/handles/tcp.pyx":224
 * 
 *     try:
 *         transport._on_connect(exc)             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         wrapper.transport._fatal_error(ex, False)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_v_transport->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._on_connect(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_transport), __pyx_v_exc); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 224, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/tcp.pyx":223
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/tcp.pyx":225
 *     try:
 *         transport._on_connect(exc)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         wrapper.transport._fatal_error(ex, False)
 *     finally:
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("uvloop.loop.__tcp_connect_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(3, 225, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_ex = __pyx_t_7;
        /*try:*/ {

          /* "uvloop/handles/tcp.pyx":226
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 *         wrapper.transport._fatal_error(ex, False)             # <<<<<<<<<<<<<<
 *     finally:
 *         wrapper.on_done()
 */
          __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport *)__pyx_v_wrapper->transport->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_wrapper->transport), __pyx_v_ex, Py_False, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(3, 226, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }

        /* "uvloop/handles/tcp.pyx":225
 *     try:
 *         transport._on_connect(exc)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         wrapper.transport._fatal_error(ex, False)
 *     finally:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L19;
          }
          __pyx_L18_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __pyx_t_6 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            }
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
            goto __pyx_L9_except_error;
          }
          __pyx_L19:;
        }
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "uvloop/handles/tcp.pyx":223
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L5_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      __pyx_L12_try_end:;
    }
  }

  /* "uvloop/handles/tcp.pyx":228
 *         wrapper.transport._fatal_error(ex, False)
 *     finally:
 *         wrapper.on_done()             # <<<<<<<<<<<<<<
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop__TCPConnectRequest *)__pyx_v_wrapper->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_wrapper)); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_10 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
      {
        __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop__TCPConnectRequest *)__pyx_v_wrapper->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_wrapper)); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 228, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_18;
      goto __pyx_L1_error;
      __pyx_L25_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/handles/tcp.pyx":207
 * 
 * 
 * cdef void __tcp_connect_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_connect_t* req,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uvloop.loop.__tcp_connect_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XDECREF((PyObject *)__pyx_v_transport);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/pipe.pyx":1
 * cdef __pipe_init_uv_handle(UVStream handle, Loop loop):             # <<<<<<<<<<<<<<
 *     cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop___pipe_init_uv_handle(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_handle, CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pipe_init_uv_handle", 1);

  /* "uvloop/handles/pipe.pyx":4
 *     cdef int err
 * 
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_pipe_t))             # <<<<<<<<<<<<<<
 *     if handle._handle is NULL:
 *         handle._abort_init()
 */
  __pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_pipe_t))));

  /* "uvloop/handles/pipe.pyx":5
 * 
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_pipe_t))
 *     if handle._handle is NULL:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise MemoryError()
 */
  __pyx_t_1 = (__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/pipe.pyx":6
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_pipe_t))
 *     if handle._handle is NULL:
 *         handle._abort_init()             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_handle)); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/pipe.pyx":7
 *     if handle._handle is NULL:
 *         handle._abort_init()
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     # Initialize pipe handle with ipc=0.
 */
    PyErr_NoMemory(); __PYX_ERR(22, 7, __pyx_L1_error)

    /* "uvloop/handles/pipe.pyx":5
 * 
 *     handle._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_pipe_t))
 *     if handle._handle is NULL:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise MemoryError()
 */
  }

  /* "uvloop/handles/pipe.pyx":12
 *     # ipc=1 means that libuv will use recvmsg/sendmsg
 *     # instead of recv/send.
 *     err = uv.uv_pipe_init(handle._loop.uvloop,             # <<<<<<<<<<<<<<
 *                           <uv.uv_pipe_t*>handle._handle,
 *                           0)
 */
  __pyx_v_err = uv_pipe_init(__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._loop->uvloop, ((uv_pipe_t *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle), 0);

  /* "uvloop/handles/pipe.pyx":17
 *     # UV_HANDLE_READABLE allows calling uv_read_start() on this pipe
 *     # even if it is O_WRONLY, see also #317, libuv/libuv#2058
 *     handle._handle.flags |= uv.UV_INTERNAL_HANDLE_READABLE             # <<<<<<<<<<<<<<
 *     if err < 0:
 *         handle._abort_init()
 */
  __pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle->flags = (__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle->flags | __pyx_e_6uvloop_8includes_2uv_UV_INTERNAL_HANDLE_READABLE);

  /* "uvloop/handles/pipe.pyx":18
 *     # even if it is O_WRONLY, see also #317, libuv/libuv#2058
 *     handle._handle.flags |= uv.UV_INTERNAL_HANDLE_READABLE
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise convert_error(err)
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/pipe.pyx":19
 *     handle._handle.flags |= uv.UV_INTERNAL_HANDLE_READABLE
 *     if err < 0:
 *         handle._abort_init()             # <<<<<<<<<<<<<<
 *         raise convert_error(err)
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_handle)); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/pipe.pyx":20
 *     if err < 0:
 *         handle._abort_init()
 *         raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *     handle._finish_init()
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(22, 20, __pyx_L1_error)

    /* "uvloop/handles/pipe.pyx":18
 *     # even if it is O_WRONLY, see also #317, libuv/libuv#2058
 *     handle._handle.flags |= uv.UV_INTERNAL_HANDLE_READABLE
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         handle._abort_init()
 *         raise convert_error(err)
 */
  }

  /* "uvloop/handles/pipe.pyx":22
 *         raise convert_error(err)
 * 
 *     handle._finish_init()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStream *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_handle)); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/pipe.pyx":1
 * cdef __pipe_init_uv_handle(UVStream handle, Loop loop):             # <<<<<<<<<<<<<<
 *     cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.__pipe_init_uv_handle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":25
 * 
 * 
 * cdef __pipe_open(UVStream handle, int fd):             # <<<<<<<<<<<<<<
 *     cdef int err
 *     err = uv.uv_pipe_open(<uv.uv_pipe_t *>handle._handle,
 */

static PyObject *__pyx_f_6uvloop_4loop___pipe_open(struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_v_handle, int __pyx_v_fd) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pipe_open", 1);

  /* "uvloop/handles/pipe.pyx":27
 * cdef __pipe_open(UVStream handle, int fd):
 *     cdef int err
 *     err = uv.uv_pipe_open(<uv.uv_pipe_t *>handle._handle,             # <<<<<<<<<<<<<<
 *                           <uv.uv_os_fd_t>fd)
 *     if err < 0:
 */
  __pyx_v_err = uv_pipe_open(((uv_pipe_t *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base._handle), ((uv_os_fd_t)__pyx_v_fd));

  /* "uvloop/handles/pipe.pyx":29
 *     err = uv.uv_pipe_open(<uv.uv_pipe_t *>handle._handle,
 *                           <uv.uv_os_fd_t>fd)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(err)
 *         raise exc
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/pipe.pyx":30
 *                           <uv.uv_os_fd_t>fd)
 *     if err < 0:
 *         exc = convert_error(err)             # <<<<<<<<<<<<<<
 *         raise exc
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/pipe.pyx":31
 *     if err < 0:
 *         exc = convert_error(err)
 *         raise exc             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(22, 31, __pyx_L1_error)

    /* "uvloop/handles/pipe.pyx":29
 *     err = uv.uv_pipe_open(<uv.uv_pipe_t *>handle._handle,
 *                           <uv.uv_os_fd_t>fd)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(err)
 *         raise exc
 */
  }

  /* "uvloop/handles/pipe.pyx":25
 * 
 * 
 * cdef __pipe_open(UVStream handle, int fd):             # <<<<<<<<<<<<<<
 *     cdef int err
 *     err = uv.uv_pipe_open(<uv.uv_pipe_t *>handle._handle,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.__pipe_open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":34
 * 
 * 
 * cdef __pipe_get_socket(UVSocketHandle handle):             # <<<<<<<<<<<<<<
 *     fileno = handle._fileno()
 *     return PseudoSocket(uv.AF_UNIX, uv.SOCK_STREAM, 0, fileno)
 */

static PyObject *__pyx_f_6uvloop_4loop___pipe_get_socket(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *__pyx_v_handle) {
  PyObject *__pyx_v_fileno = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pipe_get_socket", 1);

  /* "uvloop/handles/pipe.pyx":35
 * 
 * cdef __pipe_get_socket(UVSocketHandle handle):
 *     fileno = handle._fileno()             # <<<<<<<<<<<<<<
 *     return PseudoSocket(uv.AF_UNIX, uv.SOCK_STREAM, 0, fileno)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_fileno(__pyx_v_handle); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fileno = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":36
 * cdef __pipe_get_socket(UVSocketHandle handle):
 *     fileno = handle._fileno()
 *     return PseudoSocket(uv.AF_UNIX, uv.SOCK_STREAM, 0, fileno)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(AF_UNIX); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(SOCK_STREAM); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(22, 36, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(22, 36, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_int_0)) __PYX_ERR(22, 36, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_fileno);
  __Pyx_GIVEREF(__pyx_v_fileno);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_fileno)) __PYX_ERR(22, 36, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":34
 * 
 * 
 * cdef __pipe_get_socket(UVSocketHandle handle):             # <<<<<<<<<<<<<<
 *     fileno = handle._fileno()
 *     return PseudoSocket(uv.AF_UNIX, uv.SOCK_STREAM, 0, fileno)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.__pipe_get_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fileno);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":43
 * 
 *     @staticmethod
 *     cdef UnixServer new(Loop loop, object protocol_factory, Server server,             # <<<<<<<<<<<<<<
 *                         object backlog,
 *                         object ssl,
 */

static struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_f_6uvloop_4loop_10UnixServer_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol_factory, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_backlog, PyObject *__pyx_v_ssl, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/pipe.pyx":50
 * 
 *         cdef UnixServer handle
 *         handle = UnixServer.__new__(UnixServer)             # <<<<<<<<<<<<<<
 *         handle._init(loop, protocol_factory, server, backlog,
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UnixServer(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UnixServer), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 50, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UnixServer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":51
 *         cdef UnixServer handle
 *         handle = UnixServer.__new__(UnixServer)
 *         handle._init(loop, protocol_factory, server, backlog,             # <<<<<<<<<<<<<<
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_handle), __pyx_v_loop, __pyx_v_protocol_factory, __pyx_v_server, __pyx_v_backlog, __pyx_v_ssl, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":53
 *         handle._init(loop, protocol_factory, server, backlog,
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 *         __pipe_init_uv_handle(<UVStream>handle, loop)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_init_uv_handle(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":54
 *                      ssl, ssl_handshake_timeout, ssl_shutdown_timeout)
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":43
 * 
 *     @staticmethod
 *     cdef UnixServer new(Loop loop, object protocol_factory, Server server,             # <<<<<<<<<<<<<<
 *                         object backlog,
 *                         object ssl,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UnixServer.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":56
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_10UnixServer__new_socket(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/pipe.pyx":57
 * 
 *     cdef _new_socket(self):
 *         return __pipe_get_socket(<UVSocketHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _open(self, int sockfd):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":56
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UnixServer._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":59
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         __pipe_open(<UVStream>self, sockfd)
 */

static PyObject *__pyx_f_6uvloop_4loop_10UnixServer__open(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, int __pyx_v_sockfd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open", 1);

  /* "uvloop/handles/pipe.pyx":60
 * 
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         __pipe_open(<UVStream>self, sockfd)
 *         self._mark_as_open()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":61
 *     cdef _open(self, int sockfd):
 *         self._ensure_alive()
 *         __pipe_open(<UVStream>self, sockfd)             # <<<<<<<<<<<<<<
 *         self._mark_as_open()
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_open(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_sockfd); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":62
 *         self._ensure_alive()
 *         __pipe_open(<UVStream>self, sockfd)
 *         self._mark_as_open()             # <<<<<<<<<<<<<<
 * 
 *     cdef bind(self, str path):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._mark_as_open(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":59
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 *         __pipe_open(<UVStream>self, sockfd)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UnixServer._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":64
 *         self._mark_as_open()
 * 
 *     cdef bind(self, str path):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

static PyObject *__pyx_f_6uvloop_4loop_10UnixServer_bind(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, PyObject *__pyx_v_path) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind", 1);

  /* "uvloop/handles/pipe.pyx":66
 *     cdef bind(self, str path):
 *         cdef int err
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         err = uv.uv_pipe_bind(<uv.uv_pipe_t *>self._handle,
 *                               path.encode())
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":68
 *         self._ensure_alive()
 *         err = uv.uv_pipe_bind(<uv.uv_pipe_t *>self._handle,
 *                               path.encode())             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  if (unlikely(__pyx_v_path == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(22, 68, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_path, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_AsString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(22, 68, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":67
 *         cdef int err
 *         self._ensure_alive()
 *         err = uv.uv_pipe_bind(<uv.uv_pipe_t *>self._handle,             # <<<<<<<<<<<<<<
 *                               path.encode())
 *         if err < 0:
 */
  __pyx_v_err = uv_pipe_bind(((uv_pipe_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":69
 *         err = uv.uv_pipe_bind(<uv.uv_pipe_t *>self._handle,
 *                               path.encode())
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_3 = (__pyx_v_err < 0);
  if (__pyx_t_3) {

    /* "uvloop/handles/pipe.pyx":70
 *                               path.encode())
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/pipe.pyx":71
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/pipe.pyx":72
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._mark_as_open()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/pipe.pyx":69
 *         err = uv.uv_pipe_bind(<uv.uv_pipe_t *>self._handle,
 *                               path.encode())
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/pipe.pyx":74
 *             return
 * 
 *         self._mark_as_open()             # <<<<<<<<<<<<<<
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixServer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._mark_as_open(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":64
 *         self._mark_as_open()
 * 
 *     cdef bind(self, str path):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UnixServer.bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":76
 *         self._mark_as_open()
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,             # <<<<<<<<<<<<<<
 *                                       object context):
 *         cdef UnixTransport tr
 */

static struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_f_6uvloop_4loop_10UnixServer__make_new_transport(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, PyObject *__pyx_v_protocol, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_tr = 0;
  struct __pyx_obj_6uvloop_4loop_UVStream *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_new_transport", 1);

  /* "uvloop/handles/pipe.pyx":79
 *                                       object context):
 *         cdef UnixTransport tr
 *         tr = UnixTransport.new(self._loop, protocol, self._server, waiter,             # <<<<<<<<<<<<<<
 *                                context)
 *         return <UVStream>tr
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)__pyx_v_self->__pyx_base._server);
  __Pyx_INCREF(__pyx_t_2);

  /* "uvloop/handles/pipe.pyx":80
 *         cdef UnixTransport tr
 *         tr = UnixTransport.new(self._loop, protocol, self._server, waiter,
 *                                context)             # <<<<<<<<<<<<<<
 *         return <UVStream>tr
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_6uvloop_4loop_13UnixTransport_new(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1), __pyx_v_protocol, ((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_t_2), __pyx_v_waiter, __pyx_v_context)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tr = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/pipe.pyx":81
 *         tr = UnixTransport.new(self._loop, protocol, self._server, waiter,
 *                                context)
 *         return <UVStream>tr             # <<<<<<<<<<<<<<
 * 
 *     cdef _close(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_tr));
  __pyx_r = ((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_tr);
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":76
 *         self._mark_as_open()
 * 
 *     cdef UVStream _make_new_transport(self, object protocol, object waiter,             # <<<<<<<<<<<<<<
 *                                       object context):
 *         cdef UnixTransport tr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UnixServer._make_new_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tr);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":83
 *         return <UVStream>tr
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         sock = self._fileobj
 *         if sock is not None and sock in self._loop._unix_server_sockets:
 */

static PyObject *__pyx_f_6uvloop_4loop_10UnixServer__close(struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self) {
  PyObject *__pyx_v_sock = NULL;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_prev_ino = NULL;
  PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/handles/pipe.pyx":84
 * 
 *     cdef _close(self):
 *         sock = self._fileobj             # <<<<<<<<<<<<<<
 *         if sock is not None and sock in self._loop._unix_server_sockets:
 *             path = sock.getsockname()
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.__pyx_base._fileobj;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_sock = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":85
 *     cdef _close(self):
 *         sock = self._fileobj
 *         if sock is not None and sock in self._loop._unix_server_sockets:             # <<<<<<<<<<<<<<
 *             path = sock.getsockname()
 *         else:
 */
  __pyx_t_3 = (__pyx_v_sock != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop->_unix_server_sockets == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(22, 85, __pyx_L1_error)
  }
  __pyx_t_3 = (__Pyx_PyDict_ContainsTF(__pyx_v_sock, __pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop->_unix_server_sockets, Py_EQ)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(22, 85, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/handles/pipe.pyx":86
 *         sock = self._fileobj
 *         if sock is not None and sock in self._loop._unix_server_sockets:
 *             path = sock.getsockname()             # <<<<<<<<<<<<<<
 *         else:
 *             path = None
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_sock, __pyx_n_s_getsockname); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_v_path = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/pipe.pyx":85
 *     cdef _close(self):
 *         sock = self._fileobj
 *         if sock is not None and sock in self._loop._unix_server_sockets:             # <<<<<<<<<<<<<<
 *             path = sock.getsockname()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/pipe.pyx":88
 *             path = sock.getsockname()
 *         else:
 *             path = None             # <<<<<<<<<<<<<<
 * 
 *         UVStreamServer._close(self)
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_path = Py_None;
  }
  __pyx_L3:;

  /* "uvloop/handles/pipe.pyx":90
 *             path = None
 * 
 *         UVStreamServer._close(self)             # <<<<<<<<<<<<<<
 * 
 *         if path is not None:
 */
  __pyx_t_1 = __pyx_vtabptr_6uvloop_4loop_UVStreamServer->__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":92
 *         UVStreamServer._close(self)
 * 
 *         if path is not None:             # <<<<<<<<<<<<<<
 *             prev_ino = self._loop._unix_server_sockets[sock]
 *             del self._loop._unix_server_sockets[sock]
 */
  __pyx_t_2 = (__pyx_v_path != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/handles/pipe.pyx":93
 * 
 *         if path is not None:
 *             prev_ino = self._loop._unix_server_sockets[sock]             # <<<<<<<<<<<<<<
 *             del self._loop._unix_server_sockets[sock]
 *             try:
 */
    if (unlikely(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop->_unix_server_sockets == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(22, 93, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop->_unix_server_sockets, __pyx_v_sock); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_prev_ino = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/pipe.pyx":94
 *         if path is not None:
 *             prev_ino = self._loop._unix_server_sockets[sock]
 *             del self._loop._unix_server_sockets[sock]             # <<<<<<<<<<<<<<
 *             try:
 *                 if os_stat(path).st_ino == prev_ino:
 */
    if (unlikely(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop->_unix_server_sockets == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(22, 94, __pyx_L1_error)
    }
    if (unlikely((PyDict_DelItem(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._loop->_unix_server_sockets, __pyx_v_sock) < 0))) __PYX_ERR(22, 94, __pyx_L1_error)

    /* "uvloop/handles/pipe.pyx":95
 *             prev_ino = self._loop._unix_server_sockets[sock]
 *             del self._loop._unix_server_sockets[sock]
 *             try:             # <<<<<<<<<<<<<<
 *                 if os_stat(path).st_ino == prev_ino:
 *                     os_unlink(path)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "uvloop/handles/pipe.pyx":96
 *             del self._loop._unix_server_sockets[sock]
 *             try:
 *                 if os_stat(path).st_ino == prev_ino:             # <<<<<<<<<<<<<<
 *                     os_unlink(path)
 *             except FileNotFoundError:
 */
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_stat);
        __pyx_t_4 = __pyx_v_6uvloop_4loop_os_stat; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_path};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 96, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_st_ino); if (unlikely(!__pyx_t_4)) __PYX_ERR(22, 96, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_v_prev_ino, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 96, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(22, 96, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_2) {

          /* "uvloop/handles/pipe.pyx":97
 *             try:
 *                 if os_stat(path).st_ino == prev_ino:
 *                     os_unlink(path)             # <<<<<<<<<<<<<<
 *             except FileNotFoundError:
 *                 pass
 */
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_unlink);
          __pyx_t_4 = __pyx_v_6uvloop_4loop_os_unlink; __pyx_t_5 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_path};
            __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 97, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "uvloop/handles/pipe.pyx":96
 *             del self._loop._unix_server_sockets[sock]
 *             try:
 *                 if os_stat(path).st_ino == prev_ino:             # <<<<<<<<<<<<<<
 *                     os_unlink(path)
 *             except FileNotFoundError:
 */
        }

        /* "uvloop/handles/pipe.pyx":95
 *             prev_ino = self._loop._unix_server_sockets[sock]
 *             del self._loop._unix_server_sockets[sock]
 *             try:             # <<<<<<<<<<<<<<
 *                 if os_stat(path).st_ino == prev_ino:
 *                     os_unlink(path)
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/handles/pipe.pyx":98
 *                 if os_stat(path).st_ino == prev_ino:
 *                     os_unlink(path)
 *             except FileNotFoundError:             # <<<<<<<<<<<<<<
 *                 pass
 *             except OSError as err:
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_4, &__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_FileNotFoundError); if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 98, __pyx_L9_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_10);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_4, __pyx_t_5);
      __pyx_t_1 = 0; __pyx_t_4 = 0; __pyx_t_5 = 0;
      if (__pyx_t_11) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L8_exception_handled;
      }

      /* "uvloop/handles/pipe.pyx":100
 *             except FileNotFoundError:
 *                 pass
 *             except OSError as err:             # <<<<<<<<<<<<<<
 *                 aio_logger.error('Unable to clean up listening UNIX socket '
 *                                  '%r: %r', path, err)
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.UnixServer._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_1) < 0) __PYX_ERR(22, 100, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_err = __pyx_t_4;
        /*try:*/ {

          /* "uvloop/handles/pipe.pyx":101
 *                 pass
 *             except OSError as err:
 *                 aio_logger.error('Unable to clean up listening UNIX socket '             # <<<<<<<<<<<<<<
 *                                  '%r: %r', path, err)
 * 
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_12)) __PYX_ERR(22, 101, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_12);

          /* "uvloop/handles/pipe.pyx":102
 *             except OSError as err:
 *                 aio_logger.error('Unable to clean up listening UNIX socket '
 *                                  '%r: %r', path, err)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_13 = NULL;
          __pyx_t_6 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_6 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[4] = {__pyx_t_13, __pyx_kp_u_Unable_to_clean_up_listening_UNI, __pyx_v_path, __pyx_v_err};
            __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_10)) __PYX_ERR(22, 101, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }

        /* "uvloop/handles/pipe.pyx":100
 *             except FileNotFoundError:
 *                 pass
 *             except OSError as err:             # <<<<<<<<<<<<<<
 *                 aio_logger.error('Unable to clean up listening UNIX socket '
 *                                  '%r: %r', path, err)
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            goto __pyx_L20;
          }
          __pyx_L19_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __pyx_t_11 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            }
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
            __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
            goto __pyx_L9_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "uvloop/handles/pipe.pyx":95
 *             prev_ino = self._loop._unix_server_sockets[sock]
 *             del self._loop._unix_server_sockets[sock]
 *             try:             # <<<<<<<<<<<<<<
 *                 if os_stat(path).st_ino == prev_ino:
 *                     os_unlink(path)
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_L12_try_end:;
    }

    /* "uvloop/handles/pipe.pyx":92
 *         UVStreamServer._close(self)
 * 
 *         if path is not None:             # <<<<<<<<<<<<<<
 *             prev_ino = self._loop._unix_server_sockets[sock]
 *             del self._loop._unix_server_sockets[sock]
 */
  }

  /* "uvloop/handles/pipe.pyx":83
 *         return <UVStream>tr
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         sock = self._fileobj
 *         if sock is not None and sock in self._loop._unix_server_sockets:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("uvloop.loop.UnixServer._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sock);
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_prev_ino);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_10UnixServer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_10UnixServer___reduce_cython__, "UnixServer.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_10UnixServer_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_10UnixServer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_10UnixServer___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_10UnixServer_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_10UnixServer___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UnixServer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_10UnixServer___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UnixServer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_10UnixServer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_10UnixServer_2__setstate_cython__, "UnixServer.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_10UnixServer_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_10UnixServer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_10UnixServer_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_10UnixServer_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UnixServer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_10UnixServer_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UnixServer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_10UnixServer_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixServer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UnixServer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":109
 * 
 *     @staticmethod
 *     cdef UnixTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                            object waiter, object context):
 * 
 */

static struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_f_6uvloop_4loop_13UnixTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter, PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/pipe.pyx":113
 * 
 *         cdef UnixTransport handle
 *         handle = UnixTransport.__new__(UnixTransport)             # <<<<<<<<<<<<<<
 *         handle._init(loop, protocol, server, waiter, context)
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UnixTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UnixTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 113, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":114
 *         cdef UnixTransport handle
 *         handle = UnixTransport.__new__(UnixTransport)
 *         handle._init(loop, protocol, server, waiter, context)             # <<<<<<<<<<<<<<
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 *         return handle
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop, __pyx_v_protocol, __pyx_v_server, __pyx_v_waiter, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":115
 *         handle = UnixTransport.__new__(UnixTransport)
 *         handle._init(loop, protocol, server, waiter, context)
 *         __pipe_init_uv_handle(<UVStream>handle, loop)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_init_uv_handle(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":116
 *         handle._init(loop, protocol, server, waiter, context)
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":109
 * 
 *     @staticmethod
 *     cdef UnixTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                            object waiter, object context):
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UnixTransport.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":118
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_13UnixTransport__new_socket(struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/pipe.pyx":119
 * 
 *     cdef _new_socket(self):
 *         return __pipe_get_socket(<UVSocketHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _open(self, int sockfd):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":118
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UnixTransport._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":121
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_13UnixTransport__open(struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self, int __pyx_v_sockfd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open", 1);

  /* "uvloop/handles/pipe.pyx":122
 * 
 *     cdef _open(self, int sockfd):
 *         __pipe_open(<UVStream>self, sockfd)             # <<<<<<<<<<<<<<
 * 
 *     cdef connect(self, char* addr):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_open(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_sockfd); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":121
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UnixTransport._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":124
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     cdef connect(self, char* addr):             # <<<<<<<<<<<<<<
 *         cdef _PipeConnectRequest req
 *         req = _PipeConnectRequest(self._loop, self)
 */

static PyObject *__pyx_f_6uvloop_4loop_13UnixTransport_connect(struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self, char *__pyx_v_addr) {
  struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_req = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect", 1);

  /* "uvloop/handles/pipe.pyx":126
 *     cdef connect(self, char* addr):
 *         cdef _PipeConnectRequest req
 *         req = _PipeConnectRequest(self._loop, self)             # <<<<<<<<<<<<<<
 *         req.connect(addr)
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._loop);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._loop))) __PYX_ERR(22, 126, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(22, 126, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop__PipeConnectRequest), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_req = ((struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/handles/pipe.pyx":127
 *         cdef _PipeConnectRequest req
 *         req = _PipeConnectRequest(self._loop, self)
 *         req.connect(addr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop__PipeConnectRequest *)__pyx_v_req->__pyx_base.__pyx_vtab)->connect(__pyx_v_req, __pyx_v_addr); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/pipe.pyx":124
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     cdef connect(self, char* addr):             # <<<<<<<<<<<<<<
 *         cdef _PipeConnectRequest req
 *         req = _PipeConnectRequest(self._loop, self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UnixTransport.connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_req);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_13UnixTransport_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_13UnixTransport___reduce_cython__, "UnixTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_13UnixTransport_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_13UnixTransport_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_13UnixTransport___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_13UnixTransport_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_13UnixTransport___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_13UnixTransport___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UnixTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_13UnixTransport_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_13UnixTransport_2__setstate_cython__, "UnixTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_13UnixTransport_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_13UnixTransport_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_13UnixTransport_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_13UnixTransport_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UnixTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_13UnixTransport_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_13UnixTransport_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UnixTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":134
 * 
 *     @staticmethod
 *     cdef ReadUnixTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                                object waiter):
 *         cdef ReadUnixTransport handle
 */

static struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_f_6uvloop_4loop_17ReadUnixTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter) {
  struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/pipe.pyx":137
 *                                object waiter):
 *         cdef ReadUnixTransport handle
 *         handle = ReadUnixTransport.__new__(ReadUnixTransport)             # <<<<<<<<<<<<<<
 *         # This is only used in connect_read_pipe() and subprocess_shell/exec()
 *         # directly, we could simply copy the current context.
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_ReadUnixTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 137, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":140
 *         # This is only used in connect_read_pipe() and subprocess_shell/exec()
 *         # directly, we could simply copy the current context.
 *         handle._init(loop, protocol, server, waiter, Context_CopyCurrent())             # <<<<<<<<<<<<<<
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 *         return handle
 */
  __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop, __pyx_v_protocol, __pyx_v_server, __pyx_v_waiter, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/pipe.pyx":141
 *         # directly, we could simply copy the current context.
 *         handle._init(loop, protocol, server, waiter, Context_CopyCurrent())
 *         __pipe_init_uv_handle(<UVStream>handle, loop)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop___pipe_init_uv_handle(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/pipe.pyx":142
 *         handle._init(loop, protocol, server, waiter, Context_CopyCurrent())
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":134
 * 
 *     @staticmethod
 *     cdef ReadUnixTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                                object waiter):
 *         cdef ReadUnixTransport handle
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":144
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_17ReadUnixTransport__new_socket(struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/pipe.pyx":145
 * 
 *     cdef _new_socket(self):
 *         return __pipe_get_socket(<UVSocketHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _open(self, int sockfd):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":144
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":147
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_17ReadUnixTransport__open(struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, int __pyx_v_sockfd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open", 1);

  /* "uvloop/handles/pipe.pyx":148
 * 
 *     cdef _open(self, int sockfd):
 *         __pipe_open(<UVStream>self, sockfd)             # <<<<<<<<<<<<<<
 * 
 *     def get_write_buffer_limits(self):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_open(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_sockfd); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":147
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":150
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_1get_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_get_write_buffer_limits, "ReadUnixTransport.get_write_buffer_limits(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_1get_write_buffer_limits = {"get_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_1get_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_get_write_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_1get_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_write_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_write_buffer_limits", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_write_buffer_limits", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_get_write_buffer_limits(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_get_write_buffer_limits(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_write_buffer_limits", 1);

  /* "uvloop/handles/pipe.pyx":151
 * 
 *     def get_write_buffer_limits(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 151, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":150
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.get_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":153
 *         raise NotImplementedError
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_3set_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_2set_write_buffer_limits, "ReadUnixTransport.set_write_buffer_limits(self, high=None, low=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_3set_write_buffer_limits = {"set_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_3set_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_2set_write_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_3set_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_high = 0;
  CYTHON_UNUSED PyObject *__pyx_v_low = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_write_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_high_2,&__pyx_n_s_low,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_high_2);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 153, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_low);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 153, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_write_buffer_limits") < 0)) __PYX_ERR(22, 153, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_high = values[0];
    __pyx_v_low = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_write_buffer_limits", 0, 0, 2, __pyx_nargs); __PYX_ERR(22, 153, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.set_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_2set_write_buffer_limits(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self), __pyx_v_high, __pyx_v_low);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_2set_write_buffer_limits(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_high, CYTHON_UNUSED PyObject *__pyx_v_low) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_write_buffer_limits", 1);

  /* "uvloop/handles/pipe.pyx":154
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def get_write_buffer_size(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 154, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":153
 *         raise NotImplementedError
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.set_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":156
 *         raise NotImplementedError
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_5get_write_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_4get_write_buffer_size, "ReadUnixTransport.get_write_buffer_size(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_5get_write_buffer_size = {"get_write_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_5get_write_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_4get_write_buffer_size};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_5get_write_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_write_buffer_size (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_write_buffer_size", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_write_buffer_size", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_4get_write_buffer_size(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_4get_write_buffer_size(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_write_buffer_size", 1);

  /* "uvloop/handles/pipe.pyx":157
 * 
 *     def get_write_buffer_size(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def write(self, data):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 157, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":156
 *         raise NotImplementedError
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.get_write_buffer_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":159
 *         raise NotImplementedError
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_7write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_6write, "ReadUnixTransport.write(self, data)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_7write = {"write", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_7write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_6write};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_7write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 159, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "write") < 0)) __PYX_ERR(22, 159, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write", 1, 1, 1, __pyx_nargs); __PYX_ERR(22, 159, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_6write(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_6write(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 1);

  /* "uvloop/handles/pipe.pyx":160
 * 
 *     def write(self, data):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def writelines(self, list_of_data):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 160, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":159
 *         raise NotImplementedError
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":162
 *         raise NotImplementedError
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_9writelines(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_8writelines, "ReadUnixTransport.writelines(self, list_of_data)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_9writelines = {"writelines", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_9writelines, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_8writelines};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_9writelines(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_list_of_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writelines (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_list_of_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_list_of_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 162, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "writelines") < 0)) __PYX_ERR(22, 162, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_list_of_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writelines", 1, 1, 1, __pyx_nargs); __PYX_ERR(22, 162, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.writelines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_8writelines(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self), __pyx_v_list_of_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_8writelines(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_list_of_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writelines", 1);

  /* "uvloop/handles/pipe.pyx":163
 * 
 *     def writelines(self, list_of_data):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def write_eof(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 163, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":162
 *         raise NotImplementedError
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.writelines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":165
 *         raise NotImplementedError
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_11write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_10write_eof, "ReadUnixTransport.write_eof(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_11write_eof = {"write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_11write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_10write_eof};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_11write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_eof (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("write_eof", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "write_eof", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_10write_eof(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_10write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_eof", 1);

  /* "uvloop/handles/pipe.pyx":166
 * 
 *     def write_eof(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def can_write_eof(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 166, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":165
 *         raise NotImplementedError
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.write_eof", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":168
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_13can_write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_12can_write_eof, "ReadUnixTransport.can_write_eof(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_13can_write_eof = {"can_write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_13can_write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_12can_write_eof};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_13can_write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_write_eof (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("can_write_eof", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "can_write_eof", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_12can_write_eof(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_12can_write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("can_write_eof", 1);

  /* "uvloop/handles/pipe.pyx":169
 * 
 *     def can_write_eof(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def abort(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 169, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":168
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.can_write_eof", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":171
 *         raise NotImplementedError
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_15abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_14abort, "ReadUnixTransport.abort(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_15abort = {"abort", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_15abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_14abort};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_15abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("abort (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("abort", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "abort", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_14abort(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_14abort(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("abort", 1);

  /* "uvloop/handles/pipe.pyx":172
 * 
 *     def abort(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 172, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":171
 *         raise NotImplementedError
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.abort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_16__reduce_cython__, "ReadUnixTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_17__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_16__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_16__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_17ReadUnixTransport_18__setstate_cython__, "ReadUnixTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_17ReadUnixTransport_19__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_18__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_17ReadUnixTransport_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_17ReadUnixTransport_18__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_17ReadUnixTransport_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.ReadUnixTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":179
 * 
 *     @staticmethod
 *     cdef WriteUnixTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                                 object waiter):
 *         cdef WriteUnixTransport handle
 */

static struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_f_6uvloop_4loop_18WriteUnixTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_server, PyObject *__pyx_v_waiter) {
  struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/pipe.pyx":182
 *                                 object waiter):
 *         cdef WriteUnixTransport handle
 *         handle = WriteUnixTransport.__new__(WriteUnixTransport)             # <<<<<<<<<<<<<<
 * 
 *         # We listen for read events on write-end of the pipe. When
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_WriteUnixTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_WriteUnixTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 182, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":188
 *         # receive an error -- we want to silence that error, and just
 *         # close the transport.
 *         handle._close_on_read_error()             # <<<<<<<<<<<<<<
 * 
 *         # This is only used in connect_write_pipe() and subprocess_shell/exec()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._close_on_read_error(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":192
 *         # This is only used in connect_write_pipe() and subprocess_shell/exec()
 *         # directly, we could simply copy the current context.
 *         handle._init(loop, protocol, server, waiter, Context_CopyCurrent())             # <<<<<<<<<<<<<<
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 *         return handle
 */
  __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop, __pyx_v_protocol, __pyx_v_server, __pyx_v_waiter, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/pipe.pyx":193
 *         # directly, we could simply copy the current context.
 *         handle._init(loop, protocol, server, waiter, Context_CopyCurrent())
 *         __pipe_init_uv_handle(<UVStream>handle, loop)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop___pipe_init_uv_handle(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_handle), __pyx_v_loop); if (unlikely(!__pyx_t_2)) __PYX_ERR(22, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/pipe.pyx":194
 *         handle._init(loop, protocol, server, waiter, Context_CopyCurrent())
 *         __pipe_init_uv_handle(<UVStream>handle, loop)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":179
 * 
 *     @staticmethod
 *     cdef WriteUnixTransport new(Loop loop, object protocol, Server server,             # <<<<<<<<<<<<<<
 *                                 object waiter):
 *         cdef WriteUnixTransport handle
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":196
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_18WriteUnixTransport__new_socket(struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/pipe.pyx":197
 * 
 *     cdef _new_socket(self):
 *         return __pipe_get_socket(<UVSocketHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _open(self, int sockfd):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/pipe.pyx":196
 *         return handle
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":199
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_18WriteUnixTransport__open(struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self, int __pyx_v_sockfd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_open", 1);

  /* "uvloop/handles/pipe.pyx":200
 * 
 *     cdef _open(self, int sockfd):
 *         __pipe_open(<UVStream>self, sockfd)             # <<<<<<<<<<<<<<
 * 
 *     def pause_reading(self):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___pipe_open(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_self), __pyx_v_sockfd); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":199
 *         return __pipe_get_socket(<UVSocketHandle>self)
 * 
 *     cdef _open(self, int sockfd):             # <<<<<<<<<<<<<<
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport._open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":202
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_1pause_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18WriteUnixTransport_pause_reading, "WriteUnixTransport.pause_reading(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18WriteUnixTransport_1pause_reading = {"pause_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_1pause_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_pause_reading};
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_1pause_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pause_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("pause_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pause_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18WriteUnixTransport_pause_reading(((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_pause_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pause_reading", 1);

  /* "uvloop/handles/pipe.pyx":203
 * 
 *     def pause_reading(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def resume_reading(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 203, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":202
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport.pause_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":205
 *         raise NotImplementedError
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_3resume_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18WriteUnixTransport_2resume_reading, "WriteUnixTransport.resume_reading(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18WriteUnixTransport_3resume_reading = {"resume_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_3resume_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_2resume_reading};
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_3resume_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resume_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("resume_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "resume_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18WriteUnixTransport_2resume_reading(((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_2resume_reading(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resume_reading", 1);

  /* "uvloop/handles/pipe.pyx":206
 * 
 *     def resume_reading(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(22, 206, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":205
 *         raise NotImplementedError
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport.resume_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18WriteUnixTransport_4__reduce_cython__, "WriteUnixTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18WriteUnixTransport_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_4__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18WriteUnixTransport_4__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18WriteUnixTransport_6__setstate_cython__, "WriteUnixTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18WriteUnixTransport_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_6__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_18WriteUnixTransport_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18WriteUnixTransport_6__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18WriteUnixTransport_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.WriteUnixTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":214
 *         uv.uv_connect_t _req_data
 * 
 *     def __cinit__(self, loop, transport):             # <<<<<<<<<<<<<<
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.request.data = <void*>self
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_19_PipeConnectRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_19_PipeConnectRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_loop = 0;
  PyObject *__pyx_v_transport = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,&__pyx_n_s_transport,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 214, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_transport)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(22, 214, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(22, 214, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(22, 214, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_loop = values[0];
    __pyx_v_transport = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, __pyx_nargs); __PYX_ERR(22, 214, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._PipeConnectRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_19_PipeConnectRequest___cinit__(((struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)__pyx_v_self), __pyx_v_loop, __pyx_v_transport);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_19_PipeConnectRequest___cinit__(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_loop, PyObject *__pyx_v_transport) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/pipe.pyx":215
 * 
 *     def __cinit__(self, loop, transport):
 *         self.request = <uv.uv_req_t*> &self._req_data             # <<<<<<<<<<<<<<
 *         self.request.data = <void*>self
 *         self.transport = transport
 */
  __pyx_v_self->__pyx_base.request = ((uv_req_t *)(&__pyx_v_self->_req_data));

  /* "uvloop/handles/pipe.pyx":216
 *     def __cinit__(self, loop, transport):
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.request.data = <void*>self             # <<<<<<<<<<<<<<
 *         self.transport = transport
 * 
 */
  __pyx_v_self->__pyx_base.request->data = ((void *)__pyx_v_self);

  /* "uvloop/handles/pipe.pyx":217
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.request.data = <void*>self
 *         self.transport = transport             # <<<<<<<<<<<<<<
 * 
 *     cdef connect(self, char* addr):
 */
  if (!(likely(((__pyx_v_transport) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_transport, __pyx_ptype_6uvloop_4loop_UnixTransport))))) __PYX_ERR(22, 217, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_transport;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->transport);
  __Pyx_DECREF((PyObject *)__pyx_v_self->transport);
  __pyx_v_self->transport = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":214
 *         uv.uv_connect_t _req_data
 * 
 *     def __cinit__(self, loop, transport):             # <<<<<<<<<<<<<<
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.request.data = <void*>self
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._PipeConnectRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":219
 *         self.transport = transport
 * 
 *     cdef connect(self, char* addr):             # <<<<<<<<<<<<<<
 *         # uv_pipe_connect returns void
 *         uv.uv_pipe_connect(<uv.uv_connect_t*>self.request,
 */

static PyObject *__pyx_f_6uvloop_4loop_19_PipeConnectRequest_connect(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self, char *__pyx_v_addr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect", 1);

  /* "uvloop/handles/pipe.pyx":221
 *     cdef connect(self, char* addr):
 *         # uv_pipe_connect returns void
 *         uv.uv_pipe_connect(<uv.uv_connect_t*>self.request,             # <<<<<<<<<<<<<<
 *                            <uv.uv_pipe_t*>self.transport._handle,
 *                            addr,
 */
  uv_pipe_connect(((uv_connect_t *)__pyx_v_self->__pyx_base.request), ((uv_pipe_t *)__pyx_v_self->transport->__pyx_base.__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_addr, __pyx_f_6uvloop_4loop___pipe_connect_callback);

  /* "uvloop/handles/pipe.pyx":219
 *         self.transport = transport
 * 
 *     cdef connect(self, char* addr):             # <<<<<<<<<<<<<<
 *         # uv_pipe_connect returns void
 *         uv.uv_pipe_connect(<uv.uv_connect_t*>self.request,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_19_PipeConnectRequest_2__reduce_cython__, "_PipeConnectRequest.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_19_PipeConnectRequest_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_PipeConnectRequest_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_19_PipeConnectRequest_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_19_PipeConnectRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._PipeConnectRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_19_PipeConnectRequest_4__setstate_cython__, "_PipeConnectRequest.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_19_PipeConnectRequest_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_PipeConnectRequest_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._PipeConnectRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_19_PipeConnectRequest_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_19_PipeConnectRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._PipeConnectRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/pipe.pyx":226
 *                            __pipe_connect_callback)
 * 
 * cdef void __pipe_connect_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_connect_t* req,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___pipe_connect_callback(uv_connect_t *__pyx_v_req, int __pyx_v_status) {
  struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *__pyx_v_wrapper = 0;
  struct __pyx_obj_6uvloop_4loop_UnixTransport *__pyx_v_transport = 0;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  char const *__pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__pipe_connect_callback", 0);

  /* "uvloop/handles/pipe.pyx":234
 *         UnixTransport transport
 * 
 *     wrapper = <_PipeConnectRequest> req.data             # <<<<<<<<<<<<<<
 *     transport = wrapper.transport
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_req->data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":235
 * 
 *     wrapper = <_PipeConnectRequest> req.data
 *     transport = wrapper.transport             # <<<<<<<<<<<<<<
 * 
 *     if status < 0:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_wrapper->transport);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_transport = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/pipe.pyx":237
 *     transport = wrapper.transport
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *     else:
 */
  __pyx_t_2 = (__pyx_v_status < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/pipe.pyx":238
 * 
 *     if status < 0:
 *         exc = convert_error(status)             # <<<<<<<<<<<<<<
 *     else:
 *         exc = None
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/pipe.pyx":237
 *     transport = wrapper.transport
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/pipe.pyx":240
 *         exc = convert_error(status)
 *     else:
 *         exc = None             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_exc = Py_None;
  }
  __pyx_L3:;

  /* "uvloop/handles/pipe.pyx":242
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "uvloop/handles/pipe.pyx":243
 * 
 *     try:
 *         transport._on_connect(exc)             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         wrapper.transport._fatal_error(ex, False)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_v_transport->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._on_connect(((struct __pyx_obj_6uvloop_4loop_UVStream *)__pyx_v_transport), __pyx_v_exc); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 243, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/pipe.pyx":242
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/pipe.pyx":244
 *     try:
 *         transport._on_connect(exc)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         wrapper.transport._fatal_error(ex, False)
 *     finally:
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("uvloop.loop.__pipe_connect_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(22, 244, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_ex = __pyx_t_7;
        /*try:*/ {

          /* "uvloop/handles/pipe.pyx":245
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 *         wrapper.transport._fatal_error(ex, False)             # <<<<<<<<<<<<<<
 *     finally:
 *         wrapper.on_done()
 */
          __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport *)__pyx_v_wrapper->transport->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_wrapper->transport), __pyx_v_ex, Py_False, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(22, 245, __pyx_L18_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }

        /* "uvloop/handles/pipe.pyx":244
 *     try:
 *         transport._on_connect(exc)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         wrapper.transport._fatal_error(ex, False)
 *     finally:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L19;
          }
          __pyx_L18_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __pyx_t_6 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_15);
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            }
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
            goto __pyx_L9_except_error;
          }
          __pyx_L19:;
        }
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "uvloop/handles/pipe.pyx":242
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         transport._on_connect(exc)
 *     except BaseException as ex:
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L5_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      __pyx_L12_try_end:;
    }
  }

  /* "uvloop/handles/pipe.pyx":247
 *         wrapper.transport._fatal_error(ex, False)
 *     finally:
 *         wrapper.on_done()             # <<<<<<<<<<<<<<
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop__PipeConnectRequest *)__pyx_v_wrapper->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_wrapper)); if (unlikely(!__pyx_t_8)) __PYX_ERR(22, 247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_15);
      __pyx_t_10 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
      {
        __pyx_t_8 = ((struct __pyx_vtabstruct_6uvloop_4loop__PipeConnectRequest *)__pyx_v_wrapper->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_wrapper)); if (unlikely(!__pyx_t_8)) __PYX_ERR(22, 247, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
      }
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_4, __pyx_t_3);
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
      __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_18;
      goto __pyx_L1_error;
      __pyx_L25_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_16, __pyx_t_15);
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/handles/pipe.pyx":226
 *                            __pipe_connect_callback)
 * 
 * cdef void __pipe_connect_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_connect_t* req,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uvloop.loop.__pipe_connect_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XDECREF((PyObject *)__pyx_v_transport);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/process.pyx":5
 *     """Abstract class; wrapper over uv_process_t handle."""
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.uv_opt_env = NULL
 *         self.uv_opt_args = NULL
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_9UVProcess_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_9UVProcess_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVProcess___cinit__(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_9UVProcess___cinit__(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/process.pyx":6
 * 
 *     def __cinit__(self):
 *         self.uv_opt_env = NULL             # <<<<<<<<<<<<<<
 *         self.uv_opt_args = NULL
 *         self._returncode = None
 */
  __pyx_v_self->uv_opt_env = NULL;

  /* "uvloop/handles/process.pyx":7
 *     def __cinit__(self):
 *         self.uv_opt_env = NULL
 *         self.uv_opt_args = NULL             # <<<<<<<<<<<<<<
 *         self._returncode = None
 *         self._pid = None
 */
  __pyx_v_self->uv_opt_args = NULL;

  /* "uvloop/handles/process.pyx":8
 *         self.uv_opt_env = NULL
 *         self.uv_opt_args = NULL
 *         self._returncode = None             # <<<<<<<<<<<<<<
 *         self._pid = None
 *         self._fds_to_close = list()
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_returncode);
  __Pyx_DECREF(__pyx_v_self->_returncode);
  __pyx_v_self->_returncode = Py_None;

  /* "uvloop/handles/process.pyx":9
 *         self.uv_opt_args = NULL
 *         self._returncode = None
 *         self._pid = None             # <<<<<<<<<<<<<<
 *         self._fds_to_close = list()
 *         self._preexec_fn = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_pid);
  __Pyx_DECREF(__pyx_v_self->_pid);
  __pyx_v_self->_pid = Py_None;

  /* "uvloop/handles/process.pyx":10
 *         self._returncode = None
 *         self._pid = None
 *         self._fds_to_close = list()             # <<<<<<<<<<<<<<
 *         self._preexec_fn = None
 *         self._restore_signals = True
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_fds_to_close);
  __Pyx_DECREF(__pyx_v_self->_fds_to_close);
  __pyx_v_self->_fds_to_close = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":11
 *         self._pid = None
 *         self._fds_to_close = list()
 *         self._preexec_fn = None             # <<<<<<<<<<<<<<
 *         self._restore_signals = True
 *         self.context = Context_CopyCurrent()
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_preexec_fn);
  __Pyx_DECREF(__pyx_v_self->_preexec_fn);
  __pyx_v_self->_preexec_fn = Py_None;

  /* "uvloop/handles/process.pyx":12
 *         self._fds_to_close = list()
 *         self._preexec_fn = None
 *         self._restore_signals = True             # <<<<<<<<<<<<<<
 *         self.context = Context_CopyCurrent()
 * 
 */
  __pyx_v_self->_restore_signals = 1;

  /* "uvloop/handles/process.pyx":13
 *         self._preexec_fn = None
 *         self._restore_signals = True
 *         self.context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 * 
 *     cdef _close_process_handle(self):
 */
  __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.context);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.context);
  __pyx_v_self->__pyx_base.context = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":5
 *     """Abstract class; wrapper over uv_process_t handle."""
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.uv_opt_env = NULL
 *         self.uv_opt_args = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVProcess.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":15
 *         self.context = Context_CopyCurrent()
 * 
 *     cdef _close_process_handle(self):             # <<<<<<<<<<<<<<
 *         # XXX: This is a workaround for a libuv bug:
 *         # - https://github.com/libuv/libuv/issues/1933
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__close_process_handle(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_close_process_handle", 1);

  /* "uvloop/handles/process.pyx":19
 *         # - https://github.com/libuv/libuv/issues/1933
 *         # - https://github.com/libuv/libuv/pull/551
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return
 *         self._handle.data = NULL
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":20
 *         # - https://github.com/libuv/libuv/pull/551
 *         if self._handle is NULL:
 *             return             # <<<<<<<<<<<<<<
 *         self._handle.data = NULL
 *         uv.uv_close(self._handle, __uv_close_process_handle_cb)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":19
 *         # - https://github.com/libuv/libuv/issues/1933
 *         # - https://github.com/libuv/libuv/pull/551
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return
 *         self._handle.data = NULL
 */
  }

  /* "uvloop/handles/process.pyx":21
 *         if self._handle is NULL:
 *             return
 *         self._handle.data = NULL             # <<<<<<<<<<<<<<
 *         uv.uv_close(self._handle, __uv_close_process_handle_cb)
 *         self._handle = NULL  # close callback will free() the memory
 */
  __pyx_v_self->__pyx_base._handle->data = NULL;

  /* "uvloop/handles/process.pyx":22
 *             return
 *         self._handle.data = NULL
 *         uv.uv_close(self._handle, __uv_close_process_handle_cb)             # <<<<<<<<<<<<<<
 *         self._handle = NULL  # close callback will free() the memory
 * 
 */
  uv_close(__pyx_v_self->__pyx_base._handle, __pyx_f_6uvloop_4loop___uv_close_process_handle_cb);

  /* "uvloop/handles/process.pyx":23
 *         self._handle.data = NULL
 *         uv.uv_close(self._handle, __uv_close_process_handle_cb)
 *         self._handle = NULL  # close callback will free() the memory             # <<<<<<<<<<<<<<
 * 
 *     cdef _init(self, Loop loop, list args, dict env,
 */
  __pyx_v_self->__pyx_base._handle = NULL;

  /* "uvloop/handles/process.pyx":15
 *         self.context = Context_CopyCurrent()
 * 
 *     cdef _close_process_handle(self):             # <<<<<<<<<<<<<<
 *         # XXX: This is a workaround for a libuv bug:
 *         # - https://github.com/libuv/libuv/issues/1933
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":25
 *         self._handle = NULL  # close callback will free() the memory
 * 
 *     cdef _init(self, Loop loop, list args, dict env,             # <<<<<<<<<<<<<<
 *                cwd, start_new_session,
 *                _stdin, _stdout, _stderr,  # std* can be defined as macros in C
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_args, PyObject *__pyx_v_env, PyObject *__pyx_v_cwd, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr, PyObject *__pyx_v_pass_fds, PyObject *__pyx_v_debug_flags, PyObject *__pyx_v_preexec_fn, PyObject *__pyx_v_restore_signals) {
  int __pyx_v_err;
  int __pyx_v_force_fork;
  PyObject *__pyx_v_restore_inheritable = NULL;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_v_fds_to_close = NULL;
  PyObject *__pyx_v_errpipe_data = NULL;
  PyObject *__pyx_v_part = NULL;
  PyObject *__pyx_v_exc_name = NULL;
  PyObject *__pyx_v_exc_msg = NULL;
  PyObject *__pyx_v_exc_cls = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  unsigned int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  Py_ssize_t __pyx_t_18;
  int __pyx_t_19;
  char const *__pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_t_24;
  int __pyx_t_25;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/process.pyx":36
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":38
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(             # <<<<<<<<<<<<<<
 *             sizeof(uv.uv_process_t))
 *         if self._handle is NULL:
 */
  __pyx_v_self->__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_process_t))));

  /* "uvloop/handles/process.pyx":40
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(
 *             sizeof(uv.uv_process_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/process.pyx":41
 *             sizeof(uv.uv_process_t))
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/process.pyx":42
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         # Too early to call _finish_init, but still a lot of work to do.
 */
    PyErr_NoMemory(); __PYX_ERR(4, 42, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":40
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(
 *             sizeof(uv.uv_process_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/process.pyx":47
 *         # Let's set handle.data to NULL, so in case something goes wrong,
 *         # callbacks have a chance to avoid casting *something* into UVHandle.
 *         self._handle.data = NULL             # <<<<<<<<<<<<<<
 * 
 *         force_fork = False
 */
  __pyx_v_self->__pyx_base._handle->data = NULL;

  /* "uvloop/handles/process.pyx":49
 *         self._handle.data = NULL
 * 
 *         force_fork = False             # <<<<<<<<<<<<<<
 *         if system.PLATFORM_IS_APPLE and not (
 *             preexec_fn is None
 */
  __pyx_v_force_fork = 0;

  /* "uvloop/handles/process.pyx":50
 * 
 *         force_fork = False
 *         if system.PLATFORM_IS_APPLE and not (             # <<<<<<<<<<<<<<
 *             preexec_fn is None
 *             and not pass_fds
 */
  __pyx_t_3 = (PLATFORM_IS_APPLE != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }

  /* "uvloop/handles/process.pyx":51
 *         force_fork = False
 *         if system.PLATFORM_IS_APPLE and not (
 *             preexec_fn is None             # <<<<<<<<<<<<<<
 *             and not pass_fds
 *         ):
 */
  __pyx_t_4 = (__pyx_v_preexec_fn == Py_None);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }

  /* "uvloop/handles/process.pyx":52
 *         if system.PLATFORM_IS_APPLE and not (
 *             preexec_fn is None
 *             and not pass_fds             # <<<<<<<<<<<<<<
 *         ):
 *             # see _execute_child() in CPython/subprocess.py
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_pass_fds); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(4, 52, __pyx_L1_error)
  __pyx_t_5 = (!__pyx_t_4);
  __pyx_t_3 = __pyx_t_5;
  __pyx_L7_bool_binop_done:;

  /* "uvloop/handles/process.pyx":50
 * 
 *         force_fork = False
 *         if system.PLATFORM_IS_APPLE and not (             # <<<<<<<<<<<<<<
 *             preexec_fn is None
 *             and not pass_fds
 */
  __pyx_t_5 = (!__pyx_t_3);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/handles/process.pyx":55
 *         ):
 *             # see _execute_child() in CPython/subprocess.py
 *             force_fork = True             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __pyx_v_force_fork = 1;

    /* "uvloop/handles/process.pyx":50
 * 
 *         force_fork = False
 *         if system.PLATFORM_IS_APPLE and not (             # <<<<<<<<<<<<<<
 *             preexec_fn is None
 *             and not pass_fds
 */
  }

  /* "uvloop/handles/process.pyx":57
 *             force_fork = True
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._init_options(args, env, cwd, start_new_session,
 *                                _stdin, _stdout, _stderr, force_fork)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "uvloop/handles/process.pyx":58
 * 
 *         try:
 *             self._init_options(args, env, cwd, start_new_session,             # <<<<<<<<<<<<<<
 *                                _stdin, _stdout, _stderr, force_fork)
 * 
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->_init_options(__pyx_v_self, __pyx_v_args, __pyx_v_env, __pyx_v_cwd, __pyx_v_start_new_session, __pyx_v__stdin, __pyx_v__stdout, __pyx_v__stderr, __pyx_v_force_fork); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 58, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":61
 *                                _stdin, _stdout, _stderr, force_fork)
 * 
 *             restore_inheritable = set()             # <<<<<<<<<<<<<<
 *             if pass_fds:
 *                 for fd in pass_fds:
 */
      __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 61, __pyx_L9_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_restore_inheritable = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":62
 * 
 *             restore_inheritable = set()
 *             if pass_fds:             # <<<<<<<<<<<<<<
 *                 for fd in pass_fds:
 *                     if not os_get_inheritable(fd):
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_pass_fds); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 62, __pyx_L9_error)
      if (__pyx_t_2) {

        /* "uvloop/handles/process.pyx":63
 *             restore_inheritable = set()
 *             if pass_fds:
 *                 for fd in pass_fds:             # <<<<<<<<<<<<<<
 *                     if not os_get_inheritable(fd):
 *                         restore_inheritable.add(fd)
 */
        if (likely(PyList_CheckExact(__pyx_v_pass_fds)) || PyTuple_CheckExact(__pyx_v_pass_fds)) {
          __pyx_t_1 = __pyx_v_pass_fds; __Pyx_INCREF(__pyx_t_1);
          __pyx_t_9 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_pass_fds); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 63, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 63, __pyx_L9_error)
        }
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_1))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 63, __pyx_L9_error)
                #endif
                if (__pyx_t_9 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_11 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_11); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(4, 63, __pyx_L9_error)
              #else
              __pyx_t_11 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 63, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_11);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 63, __pyx_L9_error)
                #endif
                if (__pyx_t_9 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_11); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(4, 63, __pyx_L9_error)
              #else
              __pyx_t_11 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 63, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_11);
              #endif
            }
          } else {
            __pyx_t_11 = __pyx_t_10(__pyx_t_1);
            if (unlikely(!__pyx_t_11)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(4, 63, __pyx_L9_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_11);
          }
          __Pyx_XDECREF_SET(__pyx_v_fd, __pyx_t_11);
          __pyx_t_11 = 0;

          /* "uvloop/handles/process.pyx":64
 *             if pass_fds:
 *                 for fd in pass_fds:
 *                     if not os_get_inheritable(fd):             # <<<<<<<<<<<<<<
 *                         restore_inheritable.add(fd)
 *                         os_set_inheritable(fd, True)
 */
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_get_inheritable);
          __pyx_t_12 = __pyx_v_6uvloop_4loop_os_get_inheritable; __pyx_t_13 = NULL;
          __pyx_t_14 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_14 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_v_fd};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 64, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 64, __pyx_L9_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_5 = (!__pyx_t_2);
          if (__pyx_t_5) {

            /* "uvloop/handles/process.pyx":65
 *                 for fd in pass_fds:
 *                     if not os_get_inheritable(fd):
 *                         restore_inheritable.add(fd)             # <<<<<<<<<<<<<<
 *                         os_set_inheritable(fd, True)
 *         except Exception:
 */
            __pyx_t_15 = PySet_Add(__pyx_v_restore_inheritable, __pyx_v_fd); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(4, 65, __pyx_L9_error)

            /* "uvloop/handles/process.pyx":66
 *                     if not os_get_inheritable(fd):
 *                         restore_inheritable.add(fd)
 *                         os_set_inheritable(fd, True)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             self._abort_init()
 */
            __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
            __pyx_t_12 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_13 = NULL;
            __pyx_t_14 = 0;
            #if CYTHON_UNPACK_METHODS
            if (unlikely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
                __pyx_t_14 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_v_fd, Py_True};
              __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_14, 2+__pyx_t_14);
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 66, __pyx_L9_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

            /* "uvloop/handles/process.pyx":64
 *             if pass_fds:
 *                 for fd in pass_fds:
 *                     if not os_get_inheritable(fd):             # <<<<<<<<<<<<<<
 *                         restore_inheritable.add(fd)
 *                         os_set_inheritable(fd, True)
 */
          }

          /* "uvloop/handles/process.pyx":63
 *             restore_inheritable = set()
 *             if pass_fds:
 *                 for fd in pass_fds:             # <<<<<<<<<<<<<<
 *                     if not os_get_inheritable(fd):
 *                         restore_inheritable.add(fd)
 */
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/process.pyx":62
 * 
 *             restore_inheritable = set()
 *             if pass_fds:             # <<<<<<<<<<<<<<
 *                 for fd in pass_fds:
 *                     if not os_get_inheritable(fd):
 */
      }

      /* "uvloop/handles/process.pyx":57
 *             force_fork = True
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._init_options(args, env, cwd, start_new_session,
 *                                _stdin, _stdout, _stderr, force_fork)
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L14_try_end;
    __pyx_L9_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "uvloop/handles/process.pyx":67
 *                         restore_inheritable.add(fd)
 *                         os_set_inheritable(fd, True)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise
 */
    __pyx_t_16 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_16) {
      __Pyx_AddTraceback("uvloop.loop.UVProcess._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_11, &__pyx_t_12) < 0) __PYX_ERR(4, 67, __pyx_L11_except_error)
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);

      /* "uvloop/handles/process.pyx":68
 *                         os_set_inheritable(fd, True)
 *         except Exception:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_13 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 68, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "uvloop/handles/process.pyx":69
 *         except Exception:
 *             self._abort_init()
 *             raise             # <<<<<<<<<<<<<<
 * 
 *         if __forking or loop.active_process_handler is not None:
 */
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_11, __pyx_t_12);
      __pyx_t_1 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; 
      __PYX_ERR(4, 69, __pyx_L11_except_error)
    }
    goto __pyx_L11_except_error;

    /* "uvloop/handles/process.pyx":57
 *             force_fork = True
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._init_options(args, env, cwd, start_new_session,
 *                                _stdin, _stdout, _stderr, force_fork)
 */
    __pyx_L11_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L14_try_end:;
  }

  /* "uvloop/handles/process.pyx":71
 *             raise
 * 
 *         if __forking or loop.active_process_handler is not None:             # <<<<<<<<<<<<<<
 *             # Our pthread_atfork handlers won't work correctly when
 *             # another loop is forking in another thread (even though
 */
  __pyx_t_2 = (__pyx_v_6uvloop_4loop___forking != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_5 = __pyx_t_2;
    goto __pyx_L23_bool_binop_done;
  }
  __pyx_t_2 = (((PyObject *)__pyx_v_loop->active_process_handler) != Py_None);
  __pyx_t_5 = __pyx_t_2;
  __pyx_L23_bool_binop_done:;
  if (unlikely(__pyx_t_5)) {

    /* "uvloop/handles/process.pyx":75
 *             # another loop is forking in another thread (even though
 *             # GIL should help us to avoid that.)
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'Racing with another loop to spawn a process.')
 */
    __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":76
 *             # GIL should help us to avoid that.)
 *             self._abort_init()
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'Racing with another loop to spawn a process.')
 * 
 */
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__160, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_Raise(__pyx_t_12, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __PYX_ERR(4, 76, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":71
 *             raise
 * 
 *         if __forking or loop.active_process_handler is not None:             # <<<<<<<<<<<<<<
 *             # Our pthread_atfork handlers won't work correctly when
 *             # another loop is forking in another thread (even though
 */
  }

  /* "uvloop/handles/process.pyx":79
 *                 'Racing with another loop to spawn a process.')
 * 
 *         self._errpipe_read, self._errpipe_write = os_pipe()             # <<<<<<<<<<<<<<
 *         fds_to_close = self._fds_to_close
 *         self._fds_to_close = None
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_pipe);
  __pyx_t_11 = __pyx_v_6uvloop_4loop_os_pipe; __pyx_t_1 = NULL;
  __pyx_t_14 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_11))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_11, function);
      __pyx_t_14 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_14, 0+__pyx_t_14);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_12))) || (PyList_CheckExact(__pyx_t_12))) {
    PyObject* sequence = __pyx_t_12;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(4, 79, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_11 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_11 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_13 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_17 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_13);
    index = 0; __pyx_t_11 = __pyx_t_17(__pyx_t_13); if (unlikely(!__pyx_t_11)) goto __pyx_L25_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_11);
    index = 1; __pyx_t_1 = __pyx_t_17(__pyx_t_13); if (unlikely(!__pyx_t_1)) goto __pyx_L25_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_13), 2) < 0) __PYX_ERR(4, 79, __pyx_L1_error)
    __pyx_t_17 = NULL;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    goto __pyx_L26_unpacking_done;
    __pyx_L25_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_17 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(4, 79, __pyx_L1_error)
    __pyx_L26_unpacking_done:;
  }
  __Pyx_GIVEREF(__pyx_t_11);
  __Pyx_GOTREF(__pyx_v_self->_errpipe_read);
  __Pyx_DECREF(__pyx_v_self->_errpipe_read);
  __pyx_v_self->_errpipe_read = __pyx_t_11;
  __pyx_t_11 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_errpipe_write);
  __Pyx_DECREF(__pyx_v_self->_errpipe_write);
  __pyx_v_self->_errpipe_write = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":80
 * 
 *         self._errpipe_read, self._errpipe_write = os_pipe()
 *         fds_to_close = self._fds_to_close             # <<<<<<<<<<<<<<
 *         self._fds_to_close = None
 *         fds_to_close.append(self._errpipe_read)
 */
  __pyx_t_12 = __pyx_v_self->_fds_to_close;
  __Pyx_INCREF(__pyx_t_12);
  __pyx_v_fds_to_close = ((PyObject*)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "uvloop/handles/process.pyx":81
 *         self._errpipe_read, self._errpipe_write = os_pipe()
 *         fds_to_close = self._fds_to_close
 *         self._fds_to_close = None             # <<<<<<<<<<<<<<
 *         fds_to_close.append(self._errpipe_read)
 *         # add the write pipe last so we can close it early
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_fds_to_close);
  __Pyx_DECREF(__pyx_v_self->_fds_to_close);
  __pyx_v_self->_fds_to_close = ((PyObject*)Py_None);

  /* "uvloop/handles/process.pyx":82
 *         fds_to_close = self._fds_to_close
 *         self._fds_to_close = None
 *         fds_to_close.append(self._errpipe_read)             # <<<<<<<<<<<<<<
 *         # add the write pipe last so we can close it early
 *         fds_to_close.append(self._errpipe_write)
 */
  if (unlikely(__pyx_v_fds_to_close == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(4, 82, __pyx_L1_error)
  }
  __pyx_t_12 = __pyx_v_self->_errpipe_read;
  __Pyx_INCREF(__pyx_t_12);
  __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_fds_to_close, __pyx_t_12); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(4, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "uvloop/handles/process.pyx":84
 *         fds_to_close.append(self._errpipe_read)
 *         # add the write pipe last so we can close it early
 *         fds_to_close.append(self._errpipe_write)             # <<<<<<<<<<<<<<
 *         try:
 *             os_set_inheritable(self._errpipe_write, True)
 */
  if (unlikely(__pyx_v_fds_to_close == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(4, 84, __pyx_L1_error)
  }
  __pyx_t_12 = __pyx_v_self->_errpipe_write;
  __Pyx_INCREF(__pyx_t_12);
  __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_fds_to_close, __pyx_t_12); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(4, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "uvloop/handles/process.pyx":85
 *         # add the write pipe last so we can close it early
 *         fds_to_close.append(self._errpipe_write)
 *         try:             # <<<<<<<<<<<<<<
 *             os_set_inheritable(self._errpipe_write, True)
 * 
 */
  /*try:*/ {

    /* "uvloop/handles/process.pyx":86
 *         fds_to_close.append(self._errpipe_write)
 *         try:
 *             os_set_inheritable(self._errpipe_write, True)             # <<<<<<<<<<<<<<
 * 
 *             self._preexec_fn = preexec_fn
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
    __pyx_t_1 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_11 = NULL;
    __pyx_t_14 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_14 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_v_self->_errpipe_write, Py_True};
      __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 2+__pyx_t_14);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 86, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":88
 *             os_set_inheritable(self._errpipe_write, True)
 * 
 *             self._preexec_fn = preexec_fn             # <<<<<<<<<<<<<<
 *             self._restore_signals = restore_signals
 * 
 */
    __Pyx_INCREF(__pyx_v_preexec_fn);
    __Pyx_GIVEREF(__pyx_v_preexec_fn);
    __Pyx_GOTREF(__pyx_v_self->_preexec_fn);
    __Pyx_DECREF(__pyx_v_self->_preexec_fn);
    __pyx_v_self->_preexec_fn = __pyx_v_preexec_fn;

    /* "uvloop/handles/process.pyx":89
 * 
 *             self._preexec_fn = preexec_fn
 *             self._restore_signals = restore_signals             # <<<<<<<<<<<<<<
 * 
 *             loop.active_process_handler = self
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_restore_signals); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 89, __pyx_L28_error)
    __pyx_v_self->_restore_signals = __pyx_t_5;

    /* "uvloop/handles/process.pyx":91
 *             self._restore_signals = restore_signals
 * 
 *             loop.active_process_handler = self             # <<<<<<<<<<<<<<
 *             __forking = 1
 *             __forking_loop = loop
 */
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    __Pyx_GOTREF((PyObject *)__pyx_v_loop->active_process_handler);
    __Pyx_DECREF((PyObject *)__pyx_v_loop->active_process_handler);
    __pyx_v_loop->active_process_handler = __pyx_v_self;

    /* "uvloop/handles/process.pyx":92
 * 
 *             loop.active_process_handler = self
 *             __forking = 1             # <<<<<<<<<<<<<<
 *             __forking_loop = loop
 *             system.setForkHandler(<system.OnForkHandler>&__get_fork_handler)
 */
    __pyx_v_6uvloop_4loop___forking = 1;

    /* "uvloop/handles/process.pyx":93
 *             loop.active_process_handler = self
 *             __forking = 1
 *             __forking_loop = loop             # <<<<<<<<<<<<<<
 *             system.setForkHandler(<system.OnForkHandler>&__get_fork_handler)
 * 
 */
    __Pyx_INCREF((PyObject *)__pyx_v_loop);
    __Pyx_XGOTREF((PyObject *)__pyx_v_6uvloop_4loop___forking_loop);
    __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop___forking_loop, __pyx_v_loop);
    __Pyx_GIVEREF((PyObject *)__pyx_v_loop);

    /* "uvloop/handles/process.pyx":94
 *             __forking = 1
 *             __forking_loop = loop
 *             system.setForkHandler(<system.OnForkHandler>&__get_fork_handler)             # <<<<<<<<<<<<<<
 * 
 *             PyOS_BeforeFork()
 */
    setForkHandler(((OnForkHandler)(&__pyx_f_6uvloop_4loop___get_fork_handler)));

    /* "uvloop/handles/process.pyx":96
 *             system.setForkHandler(<system.OnForkHandler>&__get_fork_handler)
 * 
 *             PyOS_BeforeFork()             # <<<<<<<<<<<<<<
 * 
 *             err = uv.uv_spawn(loop.uvloop,
 */
    PyOS_BeforeFork();

    /* "uvloop/handles/process.pyx":98
 *             PyOS_BeforeFork()
 * 
 *             err = uv.uv_spawn(loop.uvloop,             # <<<<<<<<<<<<<<
 *                               <uv.uv_process_t*>self._handle,
 *                               &self.options)
 */
    __pyx_v_err = uv_spawn(__pyx_v_loop->uvloop, ((uv_process_t *)__pyx_v_self->__pyx_base._handle), (&__pyx_v_self->options));

    /* "uvloop/handles/process.pyx":102
 *                               &self.options)
 * 
 *             __forking = 0             # <<<<<<<<<<<<<<
 *             __forking_loop = None
 *             system.resetForkHandler()
 */
    __pyx_v_6uvloop_4loop___forking = 0;

    /* "uvloop/handles/process.pyx":103
 * 
 *             __forking = 0
 *             __forking_loop = None             # <<<<<<<<<<<<<<
 *             system.resetForkHandler()
 *             loop.active_process_handler = None
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XGOTREF((PyObject *)__pyx_v_6uvloop_4loop___forking_loop);
    __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop___forking_loop, ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None));
    __Pyx_GIVEREF(Py_None);

    /* "uvloop/handles/process.pyx":104
 *             __forking = 0
 *             __forking_loop = None
 *             system.resetForkHandler()             # <<<<<<<<<<<<<<
 *             loop.active_process_handler = None
 * 
 */
    resetForkHandler();

    /* "uvloop/handles/process.pyx":105
 *             __forking_loop = None
 *             system.resetForkHandler()
 *             loop.active_process_handler = None             # <<<<<<<<<<<<<<
 * 
 *             PyOS_AfterFork_Parent()
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_loop->active_process_handler);
    __Pyx_DECREF((PyObject *)__pyx_v_loop->active_process_handler);
    __pyx_v_loop->active_process_handler = ((struct __pyx_obj_6uvloop_4loop_UVProcess *)Py_None);

    /* "uvloop/handles/process.pyx":107
 *             loop.active_process_handler = None
 * 
 *             PyOS_AfterFork_Parent()             # <<<<<<<<<<<<<<
 * 
 *             if err < 0:
 */
    PyOS_AfterFork_Parent();

    /* "uvloop/handles/process.pyx":109
 *             PyOS_AfterFork_Parent()
 * 
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 self._close_process_handle()
 *                 self._abort_init()
 */
    __pyx_t_5 = (__pyx_v_err < 0);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/handles/process.pyx":110
 * 
 *             if err < 0:
 *                 self._close_process_handle()             # <<<<<<<<<<<<<<
 *                 self._abort_init()
 *                 raise convert_error(err)
 */
      __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->_close_process_handle(__pyx_v_self); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 110, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

      /* "uvloop/handles/process.pyx":111
 *             if err < 0:
 *                 self._close_process_handle()
 *                 self._abort_init()             # <<<<<<<<<<<<<<
 *                 raise convert_error(err)
 * 
 */
      __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 111, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

      /* "uvloop/handles/process.pyx":112
 *                 self._close_process_handle()
 *                 self._abort_init()
 *                 raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *             self._finish_init()
 */
      __pyx_t_12 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 112, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_Raise(__pyx_t_12, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __PYX_ERR(4, 112, __pyx_L28_error)

      /* "uvloop/handles/process.pyx":109
 *             PyOS_AfterFork_Parent()
 * 
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 self._close_process_handle()
 *                 self._abort_init()
 */
    }

    /* "uvloop/handles/process.pyx":114
 *                 raise convert_error(err)
 * 
 *             self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *             # close the write pipe early
 */
    __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 114, __pyx_L28_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":117
 * 
 *             # close the write pipe early
 *             os_close(fds_to_close.pop())             # <<<<<<<<<<<<<<
 * 
 *             if preexec_fn is not None:
 */
    if (unlikely(__pyx_v_fds_to_close == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
      __PYX_ERR(4, 117, __pyx_L28_error)
    }
    __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_fds_to_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 117, __pyx_L28_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_close);
    __pyx_t_11 = __pyx_v_6uvloop_4loop_os_close; __pyx_t_13 = NULL;
    __pyx_t_14 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_14 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_1};
      __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 117, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":119
 *             os_close(fds_to_close.pop())
 * 
 *             if preexec_fn is not None:             # <<<<<<<<<<<<<<
 *                 errpipe_data = bytearray()
 *                 while True:
 */
    __pyx_t_5 = (__pyx_v_preexec_fn != Py_None);
    if (__pyx_t_5) {

      /* "uvloop/handles/process.pyx":120
 * 
 *             if preexec_fn is not None:
 *                 errpipe_data = bytearray()             # <<<<<<<<<<<<<<
 *                 while True:
 *                     # XXX: This is a blocking code that has to be
 */
      __pyx_t_12 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyByteArray_Type))); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 120, __pyx_L28_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_v_errpipe_data = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "uvloop/handles/process.pyx":121
 *             if preexec_fn is not None:
 *                 errpipe_data = bytearray()
 *                 while True:             # <<<<<<<<<<<<<<
 *                     # XXX: This is a blocking code that has to be
 *                     # rewritten (using loop.connect_read_pipe() or
 */
      while (1) {

        /* "uvloop/handles/process.pyx":125
 *                     # rewritten (using loop.connect_read_pipe() or
 *                     # otherwise.)
 *                     part = os_read(self._errpipe_read, 50000)             # <<<<<<<<<<<<<<
 *                     errpipe_data += part
 *                     if not part or len(errpipe_data) > 50000:
 */
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_read);
        __pyx_t_11 = __pyx_v_6uvloop_4loop_os_read; __pyx_t_1 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_self->_errpipe_read, __pyx_int_50000};
          __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_14, 2+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 125, __pyx_L28_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_XDECREF_SET(__pyx_v_part, __pyx_t_12);
        __pyx_t_12 = 0;

        /* "uvloop/handles/process.pyx":126
 *                     # otherwise.)
 *                     part = os_read(self._errpipe_read, 50000)
 *                     errpipe_data += part             # <<<<<<<<<<<<<<
 *                     if not part or len(errpipe_data) > 50000:
 *                         break
 */
        __pyx_t_12 = PyNumber_InPlaceAdd(__pyx_v_errpipe_data, __pyx_v_part); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 126, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF_SET(__pyx_v_errpipe_data, __pyx_t_12);
        __pyx_t_12 = 0;

        /* "uvloop/handles/process.pyx":127
 *                     part = os_read(self._errpipe_read, 50000)
 *                     errpipe_data += part
 *                     if not part or len(errpipe_data) > 50000:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_part); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 127, __pyx_L28_error)
        __pyx_t_3 = (!__pyx_t_2);
        if (!__pyx_t_3) {
        } else {
          __pyx_t_5 = __pyx_t_3;
          goto __pyx_L35_bool_binop_done;
        }
        __pyx_t_9 = PyObject_Length(__pyx_v_errpipe_data); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(4, 127, __pyx_L28_error)
        __pyx_t_3 = (__pyx_t_9 > 0xC350);
        __pyx_t_5 = __pyx_t_3;
        __pyx_L35_bool_binop_done:;
        if (__pyx_t_5) {

          /* "uvloop/handles/process.pyx":128
 *                     errpipe_data += part
 *                     if not part or len(errpipe_data) > 50000:
 *                         break             # <<<<<<<<<<<<<<
 * 
 *         finally:
 */
          goto __pyx_L33_break;

          /* "uvloop/handles/process.pyx":127
 *                     part = os_read(self._errpipe_read, 50000)
 *                     errpipe_data += part
 *                     if not part or len(errpipe_data) > 50000:             # <<<<<<<<<<<<<<
 *                         break
 * 
 */
        }
      }
      __pyx_L33_break:;

      /* "uvloop/handles/process.pyx":119
 *             os_close(fds_to_close.pop())
 * 
 *             if preexec_fn is not None:             # <<<<<<<<<<<<<<
 *                 errpipe_data = bytearray()
 *                 while True:
 */
    }
  }

  /* "uvloop/handles/process.pyx":131
 * 
 *         finally:
 *             while fds_to_close:             # <<<<<<<<<<<<<<
 *                 os_close(fds_to_close.pop())
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      while (1) {
        __pyx_t_5 = (__pyx_v_fds_to_close != Py_None)&&(PyList_GET_SIZE(__pyx_v_fds_to_close) != 0);
        if (!__pyx_t_5) break;

        /* "uvloop/handles/process.pyx":132
 *         finally:
 *             while fds_to_close:
 *                 os_close(fds_to_close.pop())             # <<<<<<<<<<<<<<
 * 
 *             for fd in restore_inheritable:
 */
        if (unlikely(__pyx_v_fds_to_close == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
          __PYX_ERR(4, 132, __pyx_L1_error)
        }
        __pyx_t_11 = __Pyx_PyList_Pop(__pyx_v_fds_to_close); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_close);
        __pyx_t_1 = __pyx_v_6uvloop_4loop_os_close; __pyx_t_13 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_11};
          __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 132, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }

      /* "uvloop/handles/process.pyx":134
 *                 os_close(fds_to_close.pop())
 * 
 *             for fd in restore_inheritable:             # <<<<<<<<<<<<<<
 *                 os_set_inheritable(fd, False)
 * 
 */
      __pyx_t_9 = 0;
      __pyx_t_1 = __Pyx_set_iterator(__pyx_v_restore_inheritable, 1, (&__pyx_t_18), (&__pyx_t_16)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF(__pyx_t_12);
      __pyx_t_12 = __pyx_t_1;
      __pyx_t_1 = 0;
      while (1) {
        __pyx_t_19 = __Pyx_set_iter_next(__pyx_t_12, __pyx_t_18, &__pyx_t_9, &__pyx_t_1, __pyx_t_16);
        if (unlikely(__pyx_t_19 == 0)) break;
        if (unlikely(__pyx_t_19 == -1)) __PYX_ERR(4, 134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_fd, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "uvloop/handles/process.pyx":135
 * 
 *             for fd in restore_inheritable:
 *                 os_set_inheritable(fd, False)             # <<<<<<<<<<<<<<
 * 
 *         # asyncio caches the PID in BaseSubprocessTransport,
 */
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
        __pyx_t_11 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_13 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_v_fd, Py_False};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_14, 2+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 135, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L29;
    }
    __pyx_L28_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_22);
      __Pyx_XGOTREF(__pyx_t_23);
      __pyx_t_16 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {

        /* "uvloop/handles/process.pyx":131
 * 
 *         finally:
 *             while fds_to_close:             # <<<<<<<<<<<<<<
 *                 os_close(fds_to_close.pop())
 * 
 */
        while (1) {
          __pyx_t_5 = (__pyx_v_fds_to_close != Py_None)&&(PyList_GET_SIZE(__pyx_v_fds_to_close) != 0);
          if (!__pyx_t_5) break;

          /* "uvloop/handles/process.pyx":132
 *         finally:
 *             while fds_to_close:
 *                 os_close(fds_to_close.pop())             # <<<<<<<<<<<<<<
 * 
 *             for fd in restore_inheritable:
 */
          if (unlikely(__pyx_v_fds_to_close == Py_None)) {
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
            __PYX_ERR(4, 132, __pyx_L42_error)
          }
          __pyx_t_1 = __Pyx_PyList_Pop(__pyx_v_fds_to_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 132, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_close);
          __pyx_t_11 = __pyx_v_6uvloop_4loop_os_close; __pyx_t_13 = NULL;
          __pyx_t_14 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_14 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_1};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 132, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }

        /* "uvloop/handles/process.pyx":134
 *                 os_close(fds_to_close.pop())
 * 
 *             for fd in restore_inheritable:             # <<<<<<<<<<<<<<
 *                 os_set_inheritable(fd, False)
 * 
 */
        __pyx_t_18 = 0;
        __pyx_t_11 = __Pyx_set_iterator(__pyx_v_restore_inheritable, 1, (&__pyx_t_9), (&__pyx_t_24)); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 134, __pyx_L42_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_12);
        __pyx_t_12 = __pyx_t_11;
        __pyx_t_11 = 0;
        while (1) {
          __pyx_t_25 = __Pyx_set_iter_next(__pyx_t_12, __pyx_t_9, &__pyx_t_18, &__pyx_t_11, __pyx_t_24);
          if (unlikely(__pyx_t_25 == 0)) break;
          if (unlikely(__pyx_t_25 == -1)) __PYX_ERR(4, 134, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_XDECREF_SET(__pyx_v_fd, __pyx_t_11);
          __pyx_t_11 = 0;

          /* "uvloop/handles/process.pyx":135
 * 
 *             for fd in restore_inheritable:
 *                 os_set_inheritable(fd, False)             # <<<<<<<<<<<<<<
 * 
 *         # asyncio caches the PID in BaseSubprocessTransport,
 */
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
          __pyx_t_1 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_13 = NULL;
          __pyx_t_14 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_14 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_13, __pyx_v_fd, Py_False};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 2+__pyx_t_14);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 135, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_XGIVEREF(__pyx_t_23);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
      __pyx_lineno = __pyx_t_16; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
      __pyx_L42_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_XGIVEREF(__pyx_t_23);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L29:;
  }

  /* "uvloop/handles/process.pyx":140
 *         # so that the transport knows what the PID was even
 *         # after the process is finished.
 *         self._pid = (<uv.uv_process_t*>self._handle).pid             # <<<<<<<<<<<<<<
 * 
 *         # Track the process handle (create a strong ref to it)
 */
  __pyx_t_12 = __Pyx_PyInt_From_int(((uv_process_t *)__pyx_v_self->__pyx_base._handle)->pid); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_12);
  __Pyx_GOTREF(__pyx_v_self->_pid);
  __Pyx_DECREF(__pyx_v_self->_pid);
  __pyx_v_self->_pid = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "uvloop/handles/process.pyx":148
 *         # which will call `UVProcess._close()`, which will, in turn,
 *         # untrack this handle.
 *         self._loop._track_process(self)             # <<<<<<<<<<<<<<
 * 
 *         if debug_flags & __PROCESS_DEBUG_SLEEP_AFTER_FORK:
 */
  __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_base._loop->__pyx_vtab)->_track_process(__pyx_v_self->__pyx_base._loop, __pyx_v_self); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "uvloop/handles/process.pyx":150
 *         self._loop._track_process(self)
 * 
 *         if debug_flags & __PROCESS_DEBUG_SLEEP_AFTER_FORK:             # <<<<<<<<<<<<<<
 *             time_sleep(1)
 * 
 */
  __pyx_t_12 = __Pyx_PyInt_From___pyx_anon_enum(__pyx_e_6uvloop_4loop___PROCESS_DEBUG_SLEEP_AFTER_FORK); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_11 = PyNumber_And(__pyx_v_debug_flags, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(4, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (__pyx_t_5) {

    /* "uvloop/handles/process.pyx":151
 * 
 *         if debug_flags & __PROCESS_DEBUG_SLEEP_AFTER_FORK:
 *             time_sleep(1)             # <<<<<<<<<<<<<<
 * 
 *         if preexec_fn is not None and errpipe_data:
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_time_sleep);
    __pyx_t_12 = __pyx_v_6uvloop_4loop_time_sleep; __pyx_t_1 = NULL;
    __pyx_t_14 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
        __pyx_t_14 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_int_1};
      __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "uvloop/handles/process.pyx":150
 *         self._loop._track_process(self)
 * 
 *         if debug_flags & __PROCESS_DEBUG_SLEEP_AFTER_FORK:             # <<<<<<<<<<<<<<
 *             time_sleep(1)
 * 
 */
  }

  /* "uvloop/handles/process.pyx":153
 *             time_sleep(1)
 * 
 *         if preexec_fn is not None and errpipe_data:             # <<<<<<<<<<<<<<
 *             # preexec_fn has raised an exception.  The child
 *             # process must be dead now.
 */
  __pyx_t_3 = (__pyx_v_preexec_fn != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_5 = __pyx_t_3;
    goto __pyx_L49_bool_binop_done;
  }
  if (unlikely(!__pyx_v_errpipe_data)) { __Pyx_RaiseUnboundLocalError("errpipe_data"); __PYX_ERR(4, 153, __pyx_L1_error) }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_errpipe_data); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 153, __pyx_L1_error)
  __pyx_t_5 = __pyx_t_3;
  __pyx_L49_bool_binop_done:;
  if (__pyx_t_5) {

    /* "uvloop/handles/process.pyx":156
 *             # preexec_fn has raised an exception.  The child
 *             # process must be dead now.
 *             try:             # <<<<<<<<<<<<<<
 *                 exc_name, exc_msg = errpipe_data.split(b':', 1)
 *                 exc_name = exc_name.decode()
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_23, &__pyx_t_22, &__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_23);
      __Pyx_XGOTREF(__pyx_t_22);
      __Pyx_XGOTREF(__pyx_t_21);
      /*try:*/ {

        /* "uvloop/handles/process.pyx":157
 *             # process must be dead now.
 *             try:
 *                 exc_name, exc_msg = errpipe_data.split(b':', 1)             # <<<<<<<<<<<<<<
 *                 exc_name = exc_name.decode()
 *                 exc_msg = exc_msg.decode()
 */
        if (unlikely(!__pyx_v_errpipe_data)) { __Pyx_RaiseUnboundLocalError("errpipe_data"); __PYX_ERR(4, 157, __pyx_L51_error) }
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_errpipe_data, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 157, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__162, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 157, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_12))) || (PyList_CheckExact(__pyx_t_12))) {
          PyObject* sequence = __pyx_t_12;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(4, 157, __pyx_L51_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_11 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_11 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_1);
          #else
          __pyx_t_11 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 157, __pyx_L51_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 157, __pyx_L51_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_13 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 157, __pyx_L51_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_17 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_13);
          index = 0; __pyx_t_11 = __pyx_t_17(__pyx_t_13); if (unlikely(!__pyx_t_11)) goto __pyx_L57_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_11);
          index = 1; __pyx_t_1 = __pyx_t_17(__pyx_t_13); if (unlikely(!__pyx_t_1)) goto __pyx_L57_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_13), 2) < 0) __PYX_ERR(4, 157, __pyx_L51_error)
          __pyx_t_17 = NULL;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L58_unpacking_done;
          __pyx_L57_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_17 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(4, 157, __pyx_L51_error)
          __pyx_L58_unpacking_done:;
        }
        __pyx_v_exc_name = __pyx_t_11;
        __pyx_t_11 = 0;
        __pyx_v_exc_msg = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "uvloop/handles/process.pyx":158
 *             try:
 *                 exc_name, exc_msg = errpipe_data.split(b':', 1)
 *                 exc_name = exc_name.decode()             # <<<<<<<<<<<<<<
 *                 exc_msg = exc_msg.decode()
 *             except Exception:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc_name, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 158, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, NULL};
          __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 0+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 158, __pyx_L51_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_exc_name, __pyx_t_12);
        __pyx_t_12 = 0;

        /* "uvloop/handles/process.pyx":159
 *                 exc_name, exc_msg = errpipe_data.split(b':', 1)
 *                 exc_name = exc_name.decode()
 *                 exc_msg = exc_msg.decode()             # <<<<<<<<<<<<<<
 *             except Exception:
 *                 self._close()
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_exc_msg, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 159, __pyx_L51_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, NULL};
          __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 0+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 159, __pyx_L51_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF_SET(__pyx_v_exc_msg, __pyx_t_12);
        __pyx_t_12 = 0;

        /* "uvloop/handles/process.pyx":156
 *             # preexec_fn has raised an exception.  The child
 *             # process must be dead now.
 *             try:             # <<<<<<<<<<<<<<
 *                 exc_name, exc_msg = errpipe_data.split(b':', 1)
 *                 exc_name = exc_name.decode()
 */
      }
      __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
      goto __pyx_L56_try_end;
      __pyx_L51_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;

      /* "uvloop/handles/process.pyx":160
 *                 exc_name = exc_name.decode()
 *                 exc_msg = exc_msg.decode()
 *             except Exception:             # <<<<<<<<<<<<<<
 *                 self._close()
 *                 raise subprocess_SubprocessError(
 */
      __pyx_t_19 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_19) {
        __Pyx_AddTraceback("uvloop.loop.UVProcess._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_1, &__pyx_t_11) < 0) __PYX_ERR(4, 160, __pyx_L53_except_error)
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_11);

        /* "uvloop/handles/process.pyx":161
 *                 exc_msg = exc_msg.decode()
 *             except Exception:
 *                 self._close()             # <<<<<<<<<<<<<<
 *                 raise subprocess_SubprocessError(
 *                     'Bad exception data from child: {!r}'.format(
 */
        __pyx_t_13 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 161, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

        /* "uvloop/handles/process.pyx":163
 *                 self._close()
 *                 raise subprocess_SubprocessError(
 *                     'Bad exception data from child: {!r}'.format(             # <<<<<<<<<<<<<<
 *                         errpipe_data))
 *             exc_cls = getattr(__builtins__, exc_name,
 */
        __pyx_t_27 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Bad_exception_data_from_child_r, __pyx_n_s_format); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 163, __pyx_L53_except_error)
        __Pyx_GOTREF(__pyx_t_27);

        /* "uvloop/handles/process.pyx":164
 *                 raise subprocess_SubprocessError(
 *                     'Bad exception data from child: {!r}'.format(
 *                         errpipe_data))             # <<<<<<<<<<<<<<
 *             exc_cls = getattr(__builtins__, exc_name,
 *                               subprocess_SubprocessError)
 */
        if (unlikely(!__pyx_v_errpipe_data)) { __Pyx_RaiseUnboundLocalError("errpipe_data"); __PYX_ERR(4, 164, __pyx_L53_except_error) }
        __pyx_t_28 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_27))) {
          __pyx_t_28 = PyMethod_GET_SELF(__pyx_t_27);
          if (likely(__pyx_t_28)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_27);
            __Pyx_INCREF(__pyx_t_28);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_27, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_28, __pyx_v_errpipe_data};
          __pyx_t_26 = __Pyx_PyObject_FastCall(__pyx_t_27, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
          if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 163, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_26);
          __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
        }
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_subprocess_SubprocessError);
        __pyx_t_27 = __pyx_v_6uvloop_4loop_subprocess_SubprocessError; __pyx_t_28 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_27))) {
          __pyx_t_28 = PyMethod_GET_SELF(__pyx_t_27);
          if (likely(__pyx_t_28)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_27);
            __Pyx_INCREF(__pyx_t_28);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_27, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_28, __pyx_t_26};
          __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_27, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
          __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 162, __pyx_L53_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
        }
        __Pyx_Raise(__pyx_t_13, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __PYX_ERR(4, 162, __pyx_L53_except_error)
      }
      goto __pyx_L53_except_error;

      /* "uvloop/handles/process.pyx":156
 *             # preexec_fn has raised an exception.  The child
 *             # process must be dead now.
 *             try:             # <<<<<<<<<<<<<<
 *                 exc_name, exc_msg = errpipe_data.split(b':', 1)
 *                 exc_name = exc_name.decode()
 */
      __pyx_L53_except_error:;
      __Pyx_XGIVEREF(__pyx_t_23);
      __Pyx_XGIVEREF(__pyx_t_22);
      __Pyx_XGIVEREF(__pyx_t_21);
      __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_22, __pyx_t_21);
      goto __pyx_L1_error;
      __pyx_L56_try_end:;
    }

    /* "uvloop/handles/process.pyx":165
 *                     'Bad exception data from child: {!r}'.format(
 *                         errpipe_data))
 *             exc_cls = getattr(__builtins__, exc_name,             # <<<<<<<<<<<<<<
 *                               subprocess_SubprocessError)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_builtins); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "uvloop/handles/process.pyx":166
 *                         errpipe_data))
 *             exc_cls = getattr(__builtins__, exc_name,
 *                               subprocess_SubprocessError)             # <<<<<<<<<<<<<<
 * 
 *             exc = subprocess_SubprocessError(
 */
    __pyx_t_1 = __pyx_v_6uvloop_4loop_subprocess_SubprocessError;
    __Pyx_INCREF(__pyx_t_1);

    /* "uvloop/handles/process.pyx":165
 *                     'Bad exception data from child: {!r}'.format(
 *                         errpipe_data))
 *             exc_cls = getattr(__builtins__, exc_name,             # <<<<<<<<<<<<<<
 *                               subprocess_SubprocessError)
 * 
 */
    __pyx_t_12 = __Pyx_GetAttr3(__pyx_t_11, __pyx_v_exc_name, __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_exc_cls = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":168
 *                               subprocess_SubprocessError)
 * 
 *             exc = subprocess_SubprocessError(             # <<<<<<<<<<<<<<
 *                 'Exception occurred in preexec_fn.')
 *             exc.__cause__ = exc_cls(exc_msg)
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_subprocess_SubprocessError);
    __pyx_t_1 = __pyx_v_6uvloop_4loop_subprocess_SubprocessError; __pyx_t_11 = NULL;
    __pyx_t_14 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_14 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_Exception_occurred_in_preexec_fn};
      __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_v_exc = __pyx_t_12;
    __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":170
 *             exc = subprocess_SubprocessError(
 *                 'Exception occurred in preexec_fn.')
 *             exc.__cause__ = exc_cls(exc_msg)             # <<<<<<<<<<<<<<
 *             self._close()
 *             raise exc
 */
    __Pyx_INCREF(__pyx_v_exc_cls);
    __pyx_t_1 = __pyx_v_exc_cls; __pyx_t_11 = NULL;
    __pyx_t_14 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_14 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_exc_msg};
      __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 1+__pyx_t_14);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_exc, __pyx_n_s_cause, __pyx_t_12) < 0) __PYX_ERR(4, 170, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":171
 *                 'Exception occurred in preexec_fn.')
 *             exc.__cause__ = exc_cls(exc_msg)
 *             self._close()             # <<<<<<<<<<<<<<
 *             raise exc
 * 
 */
    __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "uvloop/handles/process.pyx":172
 *             exc.__cause__ = exc_cls(exc_msg)
 *             self._close()
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 *     cdef _after_fork(self):
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(4, 172, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":153
 *             time_sleep(1)
 * 
 *         if preexec_fn is not None and errpipe_data:             # <<<<<<<<<<<<<<
 *             # preexec_fn has raised an exception.  The child
 *             # process must be dead now.
 */
  }

  /* "uvloop/handles/process.pyx":25
 *         self._handle = NULL  # close callback will free() the memory
 * 
 *     cdef _init(self, Loop loop, list args, dict env,             # <<<<<<<<<<<<<<
 *                cwd, start_new_session,
 *                _stdin, _stdout, _stderr,  # std* can be defined as macros in C
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_restore_inheritable);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XDECREF(__pyx_v_fds_to_close);
  __Pyx_XDECREF(__pyx_v_errpipe_data);
  __Pyx_XDECREF(__pyx_v_part);
  __Pyx_XDECREF(__pyx_v_exc_name);
  __Pyx_XDECREF(__pyx_v_exc_msg);
  __Pyx_XDECREF(__pyx_v_exc_cls);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":174
 *             raise exc
 * 
 *     cdef _after_fork(self):             # <<<<<<<<<<<<<<
 *         # See CPython/_posixsubprocess.c for details
 *         cdef int err
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__after_fork(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_after_fork", 1);

  /* "uvloop/handles/process.pyx":178
 *         cdef int err
 * 
 *         if self._restore_signals:             # <<<<<<<<<<<<<<
 *             _Py_RestoreSignals()
 * 
 */
  if (__pyx_v_self->_restore_signals) {

    /* "uvloop/handles/process.pyx":179
 * 
 *         if self._restore_signals:
 *             _Py_RestoreSignals()             # <<<<<<<<<<<<<<
 * 
 *         PyOS_AfterFork_Child()
 */
    _Py_RestoreSignals();

    /* "uvloop/handles/process.pyx":178
 *         cdef int err
 * 
 *         if self._restore_signals:             # <<<<<<<<<<<<<<
 *             _Py_RestoreSignals()
 * 
 */
  }

  /* "uvloop/handles/process.pyx":181
 *             _Py_RestoreSignals()
 * 
 *         PyOS_AfterFork_Child()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_loop_fork(self._loop.uvloop)
 */
  PyOS_AfterFork_Child();

  /* "uvloop/handles/process.pyx":183
 *         PyOS_AfterFork_Child()
 * 
 *         err = uv.uv_loop_fork(self._loop.uvloop)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
  __pyx_v_err = uv_loop_fork(__pyx_v_self->__pyx_base._loop->uvloop);

  /* "uvloop/handles/process.pyx":184
 * 
 *         err = uv.uv_loop_fork(self._loop.uvloop)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/process.pyx":185
 *         err = uv.uv_loop_fork(self._loop.uvloop)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         if self._preexec_fn is not None:
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 185, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":184
 * 
 *         err = uv.uv_loop_fork(self._loop.uvloop)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  }

  /* "uvloop/handles/process.pyx":187
 *             raise convert_error(err)
 * 
 *         if self._preexec_fn is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 gc_disable()
 */
  __pyx_t_1 = (__pyx_v_self->_preexec_fn != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":188
 * 
 *         if self._preexec_fn is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 gc_disable()
 *                 self._preexec_fn()
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "uvloop/handles/process.pyx":189
 *         if self._preexec_fn is not None:
 *             try:
 *                 gc_disable()             # <<<<<<<<<<<<<<
 *                 self._preexec_fn()
 *             except BaseException as ex:
 */
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_gc_disable);
        __pyx_t_6 = __pyx_v_6uvloop_4loop_gc_disable; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 189, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/handles/process.pyx":190
 *             try:
 *                 gc_disable()
 *                 self._preexec_fn()             # <<<<<<<<<<<<<<
 *             except BaseException as ex:
 *                 try:
 */
        __Pyx_INCREF(__pyx_v_self->_preexec_fn);
        __pyx_t_6 = __pyx_v_self->_preexec_fn; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 190, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/handles/process.pyx":188
 * 
 *         if self._preexec_fn is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 gc_disable()
 *                 self._preexec_fn()
 */
      }

      /* "uvloop/handles/process.pyx":201
 *                     return
 *             else:
 *                 os_close(self._errpipe_write)             # <<<<<<<<<<<<<<
 *         else:
 *             os_close(self._errpipe_write)
 */
      /*else:*/ {
        __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_close);
        __pyx_t_6 = __pyx_v_6uvloop_4loop_os_close; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_self->_errpipe_write};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 201, __pyx_L8_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/handles/process.pyx":191
 *                 gc_disable()
 *                 self._preexec_fn()
 *             except BaseException as ex:             # <<<<<<<<<<<<<<
 *                 try:
 *                     with open(self._errpipe_write, 'wb') as f:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.UVProcess._after_fork", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(4, 191, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __pyx_v_ex = __pyx_t_6;
        /*try:*/ {

          /* "uvloop/handles/process.pyx":192
 *                 self._preexec_fn()
 *             except BaseException as ex:
 *                 try:             # <<<<<<<<<<<<<<
 *                     with open(self._errpipe_write, 'wb') as f:
 *                         f.write(str(ex.__class__.__name__).encode())
 */
          /*try:*/ {

            /* "uvloop/handles/process.pyx":193
 *             except BaseException as ex:
 *                 try:
 *                     with open(self._errpipe_write, 'wb') as f:             # <<<<<<<<<<<<<<
 *                         f.write(str(ex.__class__.__name__).encode())
 *                         f.write(b':')
 */
            /*with:*/ {
              __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 193, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_INCREF(__pyx_v_self->_errpipe_write);
              __Pyx_GIVEREF(__pyx_v_self->_errpipe_write);
              if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_self->_errpipe_write)) __PYX_ERR(4, 193, __pyx_L22_error);
              __Pyx_INCREF(__pyx_n_u_wb);
              __Pyx_GIVEREF(__pyx_n_u_wb);
              if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_n_u_wb)) __PYX_ERR(4, 193, __pyx_L22_error);
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_10, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 193, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __pyx_t_12 = __Pyx_PyObject_LookupSpecial(__pyx_t_11, __pyx_n_s_exit); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 193, __pyx_L22_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_13 = __Pyx_PyObject_LookupSpecial(__pyx_t_11, __pyx_n_s_enter); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 193, __pyx_L24_error)
              __Pyx_GOTREF(__pyx_t_13);
              __pyx_t_14 = NULL;
              __pyx_t_8 = 0;
              #if CYTHON_UNPACK_METHODS
              if (likely(PyMethod_Check(__pyx_t_13))) {
                __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
                if (likely(__pyx_t_14)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                  __Pyx_INCREF(__pyx_t_14);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_13, function);
                  __pyx_t_8 = 1;
                }
              }
              #endif
              {
                PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
                __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 193, __pyx_L24_error)
                __Pyx_GOTREF(__pyx_t_10);
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              }
              __pyx_t_13 = __pyx_t_10;
              __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              /*try:*/ {
                {
                  __Pyx_PyThreadState_declare
                  __Pyx_PyThreadState_assign
                  __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
                  __Pyx_XGOTREF(__pyx_t_15);
                  __Pyx_XGOTREF(__pyx_t_16);
                  __Pyx_XGOTREF(__pyx_t_17);
                  /*try:*/ {
                    __pyx_v_f = __pyx_t_13;
                    __pyx_t_13 = 0;

                    /* "uvloop/handles/process.pyx":194
 *                 try:
 *                     with open(self._errpipe_write, 'wb') as f:
 *                         f.write(str(ex.__class__.__name__).encode())             # <<<<<<<<<<<<<<
 *                         f.write(b':')
 *                         f.write(str(ex.args[0]).encode())
 */
                    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 194, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_11);
                    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_ex, __pyx_n_s_class); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 194, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_name); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 194, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                    __pyx_t_10 = __Pyx_PyObject_Unicode(__pyx_t_14); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 194, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                    __pyx_t_14 = PyUnicode_AsEncodedString(((PyObject*)__pyx_t_10), NULL, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 194, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                    __pyx_t_10 = NULL;
                    __pyx_t_8 = 0;
                    #if CYTHON_UNPACK_METHODS
                    if (likely(PyMethod_Check(__pyx_t_11))) {
                      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
                      if (likely(__pyx_t_10)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                        __Pyx_INCREF(__pyx_t_10);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_11, function);
                        __pyx_t_8 = 1;
                      }
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_14};
                      __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                      if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 194, __pyx_L30_error)
                      __Pyx_GOTREF(__pyx_t_13);
                      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                    /* "uvloop/handles/process.pyx":195
 *                     with open(self._errpipe_write, 'wb') as f:
 *                         f.write(str(ex.__class__.__name__).encode())
 *                         f.write(b':')             # <<<<<<<<<<<<<<
 *                         f.write(str(ex.args[0]).encode())
 *                 finally:
 */
                    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 195, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_11);
                    __pyx_t_14 = NULL;
                    __pyx_t_8 = 0;
                    #if CYTHON_UNPACK_METHODS
                    if (likely(PyMethod_Check(__pyx_t_11))) {
                      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
                      if (likely(__pyx_t_14)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                        __Pyx_INCREF(__pyx_t_14);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_11, function);
                        __pyx_t_8 = 1;
                      }
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_kp_b__161};
                      __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                      if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 195, __pyx_L30_error)
                      __Pyx_GOTREF(__pyx_t_13);
                      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                    /* "uvloop/handles/process.pyx":196
 *                         f.write(str(ex.__class__.__name__).encode())
 *                         f.write(b':')
 *                         f.write(str(ex.args[0]).encode())             # <<<<<<<<<<<<<<
 *                 finally:
 *                     system._exit(255)
 */
                    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 196, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_11);
                    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_ex, __pyx_n_s_args); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 196, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_14, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 196, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                    __pyx_t_14 = __Pyx_PyObject_Unicode(__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 196, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                    __pyx_t_10 = PyUnicode_AsEncodedString(((PyObject*)__pyx_t_14), NULL, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 196, __pyx_L30_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                    __pyx_t_14 = NULL;
                    __pyx_t_8 = 0;
                    #if CYTHON_UNPACK_METHODS
                    if (likely(PyMethod_Check(__pyx_t_11))) {
                      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
                      if (likely(__pyx_t_14)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                        __Pyx_INCREF(__pyx_t_14);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_11, function);
                        __pyx_t_8 = 1;
                      }
                    }
                    #endif
                    {
                      PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_t_10};
                      __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
                      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                      if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 196, __pyx_L30_error)
                      __Pyx_GOTREF(__pyx_t_13);
                      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

                    /* "uvloop/handles/process.pyx":193
 *             except BaseException as ex:
 *                 try:
 *                     with open(self._errpipe_write, 'wb') as f:             # <<<<<<<<<<<<<<
 *                         f.write(str(ex.__class__.__name__).encode())
 *                         f.write(b':')
 */
                  }
                  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                  __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                  goto __pyx_L37_try_end;
                  __pyx_L30_error:;
                  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                  __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                  __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                  /*except:*/ {
                    __Pyx_AddTraceback("uvloop.loop.UVProcess._after_fork", __pyx_clineno, __pyx_lineno, __pyx_filename);
                    if (__Pyx_GetException(&__pyx_t_13, &__pyx_t_11, &__pyx_t_10) < 0) __PYX_ERR(4, 193, __pyx_L32_except_error)
                    __Pyx_XGOTREF(__pyx_t_13);
                    __Pyx_XGOTREF(__pyx_t_11);
                    __Pyx_XGOTREF(__pyx_t_10);
                    __pyx_t_14 = PyTuple_Pack(3, __pyx_t_13, __pyx_t_11, __pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 193, __pyx_L32_except_error)
                    __Pyx_GOTREF(__pyx_t_14);
                    __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_14, NULL);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                    if (unlikely(!__pyx_t_18)) __PYX_ERR(4, 193, __pyx_L32_except_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_18);
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    if (__pyx_t_1 < 0) __PYX_ERR(4, 193, __pyx_L32_except_error)
                    __pyx_t_19 = (!__pyx_t_1);
                    if (unlikely(__pyx_t_19)) {
                      __Pyx_GIVEREF(__pyx_t_13);
                      __Pyx_GIVEREF(__pyx_t_11);
                      __Pyx_XGIVEREF(__pyx_t_10);
                      __Pyx_ErrRestoreWithState(__pyx_t_13, __pyx_t_11, __pyx_t_10);
                      __pyx_t_13 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; 
                      __PYX_ERR(4, 193, __pyx_L32_except_error)
                    }
                    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                    goto __pyx_L31_exception_handled;
                  }
                  __pyx_L32_except_error:;
                  __Pyx_XGIVEREF(__pyx_t_15);
                  __Pyx_XGIVEREF(__pyx_t_16);
                  __Pyx_XGIVEREF(__pyx_t_17);
                  __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
                  goto __pyx_L22_error;
                  __pyx_L31_exception_handled:;
                  __Pyx_XGIVEREF(__pyx_t_15);
                  __Pyx_XGIVEREF(__pyx_t_16);
                  __Pyx_XGIVEREF(__pyx_t_17);
                  __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
                  __pyx_L37_try_end:;
                }
              }
              /*finally:*/ {
                /*normal exit:*/{
                  if (__pyx_t_12) {
                    __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_tuple__163, NULL);
                    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
                    if (unlikely(!__pyx_t_17)) __PYX_ERR(4, 193, __pyx_L22_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  }
                  goto __pyx_L29;
                }
                __pyx_L29:;
              }
              goto __pyx_L41;
              __pyx_L24_error:;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              goto __pyx_L22_error;
              __pyx_L41:;
            }
          }

          /* "uvloop/handles/process.pyx":198
 *                         f.write(str(ex.args[0]).encode())
 *                 finally:
 *                     system._exit(255)             # <<<<<<<<<<<<<<
 *                     return
 *             else:
 */
          /*finally:*/ {
            /*normal exit:*/{
              _exit(0xFF);

              /* "uvloop/handles/process.pyx":199
 *                 finally:
 *                     system._exit(255)
 *                     return             # <<<<<<<<<<<<<<
 *             else:
 *                 os_close(self._errpipe_write)
 */
              __Pyx_XDECREF(__pyx_r);
              __pyx_r = Py_None; __Pyx_INCREF(Py_None);
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              goto __pyx_L16_return;
            }
            __pyx_L22_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_12 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_18 = 0; __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_18, &__pyx_t_20);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_17, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_17, &__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_12);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_15);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_20);
              {

                /* "uvloop/handles/process.pyx":198
 *                         f.write(str(ex.args[0]).encode())
 *                 finally:
 *                     system._exit(255)             # <<<<<<<<<<<<<<
 *                     return
 *             else:
 */
                _exit(0xFF);

                /* "uvloop/handles/process.pyx":199
 *                 finally:
 *                     system._exit(255)
 *                     return             # <<<<<<<<<<<<<<
 *             else:
 *                 os_close(self._errpipe_write)
 */
                __Pyx_XDECREF(__pyx_r);
                __pyx_r = Py_None; __Pyx_INCREF(Py_None);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
                goto __pyx_L44_return;
              }
              __pyx_L44_return:;
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_15);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_18, __pyx_t_20);
              }
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_15 = 0; __pyx_t_18 = 0; __pyx_t_20 = 0;
              goto __pyx_L16_return;
            }
          }
        }

        /* "uvloop/handles/process.pyx":191
 *                 gc_disable()
 *                 self._preexec_fn()
 *             except BaseException as ex:             # <<<<<<<<<<<<<<
 *                 try:
 *                     with open(self._errpipe_write, 'wb') as f:
 */
        /*finally:*/ {
          __pyx_L16_return: {
            __pyx_t_20 = __pyx_r;
            __pyx_r = 0;
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            __pyx_r = __pyx_t_20;
            __pyx_t_20 = 0;
            goto __pyx_L9_except_return;
          }
        }
      }
      goto __pyx_L8_except_error;

      /* "uvloop/handles/process.pyx":188
 * 
 *         if self._preexec_fn is not None:
 *             try:             # <<<<<<<<<<<<<<
 *                 gc_disable()
 *                 self._preexec_fn()
 */
      __pyx_L8_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L9_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L11_try_end:;
    }

    /* "uvloop/handles/process.pyx":187
 *             raise convert_error(err)
 * 
 *         if self._preexec_fn is not None:             # <<<<<<<<<<<<<<
 *             try:
 *                 gc_disable()
 */
    goto __pyx_L5;
  }

  /* "uvloop/handles/process.pyx":203
 *                 os_close(self._errpipe_write)
 *         else:
 *             os_close(self._errpipe_write)             # <<<<<<<<<<<<<<
 * 
 *     cdef _close_after_spawn(self, int fd):
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_close);
    __pyx_t_6 = __pyx_v_6uvloop_4loop_os_close; __pyx_t_2 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->_errpipe_write};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L5:;

  /* "uvloop/handles/process.pyx":174
 *             raise exc
 * 
 *     cdef _after_fork(self):             # <<<<<<<<<<<<<<
 *         # See CPython/_posixsubprocess.c for details
 *         cdef int err
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._after_fork", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":205
 *             os_close(self._errpipe_write)
 * 
 *     cdef _close_after_spawn(self, int fd):             # <<<<<<<<<<<<<<
 *         if self._fds_to_close is None:
 *             raise RuntimeError(
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__close_after_spawn(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, int __pyx_v_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close_after_spawn", 1);

  /* "uvloop/handles/process.pyx":206
 * 
 *     cdef _close_after_spawn(self, int fd):
 *         if self._fds_to_close is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'UVProcess._close_after_spawn called after uv_spawn')
 */
  __pyx_t_1 = (__pyx_v_self->_fds_to_close == ((PyObject*)Py_None));
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/process.pyx":207
 *     cdef _close_after_spawn(self, int fd):
 *         if self._fds_to_close is None:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'UVProcess._close_after_spawn called after uv_spawn')
 *         self._fds_to_close.append(fd)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__164, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 207, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":206
 * 
 *     cdef _close_after_spawn(self, int fd):
 *         if self._fds_to_close is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'UVProcess._close_after_spawn called after uv_spawn')
 */
  }

  /* "uvloop/handles/process.pyx":209
 *             raise RuntimeError(
 *                 'UVProcess._close_after_spawn called after uv_spawn')
 *         self._fds_to_close.append(fd)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  if (unlikely(__pyx_v_self->_fds_to_close == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(4, 209, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_self->_fds_to_close, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(4, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":205
 *             os_close(self._errpipe_write)
 * 
 *     cdef _close_after_spawn(self, int fd):             # <<<<<<<<<<<<<<
 *         if self._fds_to_close is None:
 *             raise RuntimeError(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._close_after_spawn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":211
 *         self._fds_to_close.append(fd)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.uv_opt_env is not NULL:
 *             PyMem_RawFree(self.uv_opt_env)
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_9UVProcess_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_9UVProcess_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_9UVProcess_2__dealloc__(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_9UVProcess_2__dealloc__(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self) {
  int __pyx_t_1;

  /* "uvloop/handles/process.pyx":212
 * 
 *     def __dealloc__(self):
 *         if self.uv_opt_env is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_RawFree(self.uv_opt_env)
 *             self.uv_opt_env = NULL
 */
  __pyx_t_1 = (__pyx_v_self->uv_opt_env != NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":213
 *     def __dealloc__(self):
 *         if self.uv_opt_env is not NULL:
 *             PyMem_RawFree(self.uv_opt_env)             # <<<<<<<<<<<<<<
 *             self.uv_opt_env = NULL
 * 
 */
    PyMem_RawFree(__pyx_v_self->uv_opt_env);

    /* "uvloop/handles/process.pyx":214
 *         if self.uv_opt_env is not NULL:
 *             PyMem_RawFree(self.uv_opt_env)
 *             self.uv_opt_env = NULL             # <<<<<<<<<<<<<<
 * 
 *         if self.uv_opt_args is not NULL:
 */
    __pyx_v_self->uv_opt_env = NULL;

    /* "uvloop/handles/process.pyx":212
 * 
 *     def __dealloc__(self):
 *         if self.uv_opt_env is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_RawFree(self.uv_opt_env)
 *             self.uv_opt_env = NULL
 */
  }

  /* "uvloop/handles/process.pyx":216
 *             self.uv_opt_env = NULL
 * 
 *         if self.uv_opt_args is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_RawFree(self.uv_opt_args)
 *             self.uv_opt_args = NULL
 */
  __pyx_t_1 = (__pyx_v_self->uv_opt_args != NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":217
 * 
 *         if self.uv_opt_args is not NULL:
 *             PyMem_RawFree(self.uv_opt_args)             # <<<<<<<<<<<<<<
 *             self.uv_opt_args = NULL
 * 
 */
    PyMem_RawFree(__pyx_v_self->uv_opt_args);

    /* "uvloop/handles/process.pyx":218
 *         if self.uv_opt_args is not NULL:
 *             PyMem_RawFree(self.uv_opt_args)
 *             self.uv_opt_args = NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef char** __to_cstring_array(self, list arr):
 */
    __pyx_v_self->uv_opt_args = NULL;

    /* "uvloop/handles/process.pyx":216
 *             self.uv_opt_env = NULL
 * 
 *         if self.uv_opt_args is not NULL:             # <<<<<<<<<<<<<<
 *             PyMem_RawFree(self.uv_opt_args)
 *             self.uv_opt_args = NULL
 */
  }

  /* "uvloop/handles/process.pyx":211
 *         self._fds_to_close.append(fd)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.uv_opt_env is not NULL:
 *             PyMem_RawFree(self.uv_opt_env)
 */

  /* function exit code */
}

/* "uvloop/handles/process.pyx":220
 *             self.uv_opt_args = NULL
 * 
 *     cdef char** __to_cstring_array(self, list arr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i
 */

static char **__pyx_f_6uvloop_4loop_9UVProcess__UVProcess__to_cstring_array(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_arr) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_arr_len;
  PyObject *__pyx_v_el = 0;
  char **__pyx_v_ret;
  char **__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_UVProcess__to_cstring_array", 1);

  /* "uvloop/handles/process.pyx":223
 *         cdef:
 *             int i
 *             ssize_t arr_len = len(arr)             # <<<<<<<<<<<<<<
 *             bytes el
 * 
 */
  if (unlikely(__pyx_v_arr == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(4, 223, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_arr); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(4, 223, __pyx_L1_error)
  __pyx_v_arr_len = __pyx_t_1;

  /* "uvloop/handles/process.pyx":228
 *             char **ret
 * 
 *         ret = <char **>PyMem_RawMalloc((arr_len + 1) * sizeof(char *))             # <<<<<<<<<<<<<<
 *         if ret is NULL:
 *             raise MemoryError()
 */
  __pyx_v_ret = ((char **)PyMem_RawMalloc(((__pyx_v_arr_len + 1) * (sizeof(char *)))));

  /* "uvloop/handles/process.pyx":229
 * 
 *         ret = <char **>PyMem_RawMalloc((arr_len + 1) * sizeof(char *))
 *         if ret is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (__pyx_v_ret == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/process.pyx":230
 *         ret = <char **>PyMem_RawMalloc((arr_len + 1) * sizeof(char *))
 *         if ret is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(arr_len):
 */
    PyErr_NoMemory(); __PYX_ERR(4, 230, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":229
 * 
 *         ret = <char **>PyMem_RawMalloc((arr_len + 1) * sizeof(char *))
 *         if ret is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uvloop/handles/process.pyx":232
 *             raise MemoryError()
 * 
 *         for i in range(arr_len):             # <<<<<<<<<<<<<<
 *             el = arr[i]
 *             # NB: PyBytes_AsString doesn't copy the data;
 */
  __pyx_t_3 = __pyx_v_arr_len;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "uvloop/handles/process.pyx":233
 * 
 *         for i in range(arr_len):
 *             el = arr[i]             # <<<<<<<<<<<<<<
 *             # NB: PyBytes_AsString doesn't copy the data;
 *             # we have to be careful when the "arr" is GCed,
 */
    if (unlikely(__pyx_v_arr == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 233, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_arr, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_6))) __PYX_ERR(4, 233, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_el, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "uvloop/handles/process.pyx":237
 *             # we have to be careful when the "arr" is GCed,
 *             # and it shouldn't be ever mutated.
 *             ret[i] = PyBytes_AsString(el)             # <<<<<<<<<<<<<<
 * 
 *         ret[arr_len] = NULL
 */
    __pyx_t_7 = PyBytes_AsString(__pyx_v_el); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(4, 237, __pyx_L1_error)
    (__pyx_v_ret[__pyx_v_i]) = __pyx_t_7;
  }

  /* "uvloop/handles/process.pyx":239
 *             ret[i] = PyBytes_AsString(el)
 * 
 *         ret[arr_len] = NULL             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
  (__pyx_v_ret[__pyx_v_arr_len]) = NULL;

  /* "uvloop/handles/process.pyx":240
 * 
 *         ret[arr_len] = NULL
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_options(self, list args, dict env, cwd, start_new_session,
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":220
 *             self.uv_opt_args = NULL
 * 
 *     cdef char** __to_cstring_array(self, list arr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._UVProcess__to_cstring_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_el);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":242
 *         return ret
 * 
 *     cdef _init_options(self, list args, dict env, cwd, start_new_session,             # <<<<<<<<<<<<<<
 *                        _stdin, _stdout, _stderr, bint force_fork):
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_options(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_env, PyObject *__pyx_v_cwd, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr, int __pyx_v_force_fork) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char **__pyx_t_2;
  char *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init_options", 0);
  __Pyx_INCREF(__pyx_v_cwd);

  /* "uvloop/handles/process.pyx":245
 *                        _stdin, _stdout, _stderr, bint force_fork):
 * 
 *         memset(&self.options, 0, sizeof(uv.uv_process_options_t))             # <<<<<<<<<<<<<<
 * 
 *         self._init_env(env)
 */
  (void)(memset((&__pyx_v_self->options), 0, (sizeof(uv_process_options_t))));

  /* "uvloop/handles/process.pyx":247
 *         memset(&self.options, 0, sizeof(uv.uv_process_options_t))
 * 
 *         self._init_env(env)             # <<<<<<<<<<<<<<
 *         self.options.env = self.uv_opt_env
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->_init_env(__pyx_v_self, __pyx_v_env); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":248
 * 
 *         self._init_env(env)
 *         self.options.env = self.uv_opt_env             # <<<<<<<<<<<<<<
 * 
 *         self._init_args(args)
 */
  __pyx_t_2 = __pyx_v_self->uv_opt_env;
  __pyx_v_self->options.env = __pyx_t_2;

  /* "uvloop/handles/process.pyx":250
 *         self.options.env = self.uv_opt_env
 * 
 *         self._init_args(args)             # <<<<<<<<<<<<<<
 *         self.options.file = self.uv_opt_file
 *         self.options.args = self.uv_opt_args
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->_init_args(__pyx_v_self, __pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":251
 * 
 *         self._init_args(args)
 *         self.options.file = self.uv_opt_file             # <<<<<<<<<<<<<<
 *         self.options.args = self.uv_opt_args
 * 
 */
  __pyx_t_3 = __pyx_v_self->uv_opt_file;
  __pyx_v_self->options.file = __pyx_t_3;

  /* "uvloop/handles/process.pyx":252
 *         self._init_args(args)
 *         self.options.file = self.uv_opt_file
 *         self.options.args = self.uv_opt_args             # <<<<<<<<<<<<<<
 * 
 *         if start_new_session:
 */
  __pyx_t_2 = __pyx_v_self->uv_opt_args;
  __pyx_v_self->options.args = __pyx_t_2;

  /* "uvloop/handles/process.pyx":254
 *         self.options.args = self.uv_opt_args
 * 
 *         if start_new_session:             # <<<<<<<<<<<<<<
 *             self.options.flags |= uv.UV_PROCESS_DETACHED
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_start_new_session); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(4, 254, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "uvloop/handles/process.pyx":255
 * 
 *         if start_new_session:
 *             self.options.flags |= uv.UV_PROCESS_DETACHED             # <<<<<<<<<<<<<<
 * 
 *         if force_fork:
 */
    __pyx_v_self->options.flags = (__pyx_v_self->options.flags | UV_PROCESS_DETACHED);

    /* "uvloop/handles/process.pyx":254
 *         self.options.args = self.uv_opt_args
 * 
 *         if start_new_session:             # <<<<<<<<<<<<<<
 *             self.options.flags |= uv.UV_PROCESS_DETACHED
 * 
 */
  }

  /* "uvloop/handles/process.pyx":257
 *             self.options.flags |= uv.UV_PROCESS_DETACHED
 * 
 *         if force_fork:             # <<<<<<<<<<<<<<
 *             # This is a hack to work around the change in libuv 1.44:
 *             #    > macos: use posix_spawn instead of fork
 */
  if (__pyx_v_force_fork) {

    /* "uvloop/handles/process.pyx":269
 *             # UV_PROCESS_SETUID or UV_PROCESS_SETGID would reliably make
 *             # libuv fallback to use fork, so let's just use it for now.
 *             self.options.flags |= uv.UV_PROCESS_SETUID             # <<<<<<<<<<<<<<
 *             self.options.uid = uv.getuid()
 * 
 */
    __pyx_v_self->options.flags = (__pyx_v_self->options.flags | UV_PROCESS_SETUID);

    /* "uvloop/handles/process.pyx":270
 *             # libuv fallback to use fork, so let's just use it for now.
 *             self.options.flags |= uv.UV_PROCESS_SETUID
 *             self.options.uid = uv.getuid()             # <<<<<<<<<<<<<<
 * 
 *         if cwd is not None:
 */
    __pyx_v_self->options.uid = getuid();

    /* "uvloop/handles/process.pyx":257
 *             self.options.flags |= uv.UV_PROCESS_DETACHED
 * 
 *         if force_fork:             # <<<<<<<<<<<<<<
 *             # This is a hack to work around the change in libuv 1.44:
 *             #    > macos: use posix_spawn instead of fork
 */
  }

  /* "uvloop/handles/process.pyx":272
 *             self.options.uid = uv.getuid()
 * 
 *         if cwd is not None:             # <<<<<<<<<<<<<<
 *             cwd = os_fspath(cwd)
 * 
 */
  __pyx_t_4 = (__pyx_v_cwd != Py_None);
  if (__pyx_t_4) {

    /* "uvloop/handles/process.pyx":273
 * 
 *         if cwd is not None:
 *             cwd = os_fspath(cwd)             # <<<<<<<<<<<<<<
 * 
 *             if isinstance(cwd, str):
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_fspath);
    __pyx_t_5 = __pyx_v_6uvloop_4loop_os_fspath; __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_cwd};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_cwd, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/handles/process.pyx":275
 *             cwd = os_fspath(cwd)
 * 
 *             if isinstance(cwd, str):             # <<<<<<<<<<<<<<
 *                 cwd = PyUnicode_EncodeFSDefault(cwd)
 *             if not isinstance(cwd, bytes):
 */
    __pyx_t_4 = PyUnicode_Check(__pyx_v_cwd); 
    if (__pyx_t_4) {

      /* "uvloop/handles/process.pyx":276
 * 
 *             if isinstance(cwd, str):
 *                 cwd = PyUnicode_EncodeFSDefault(cwd)             # <<<<<<<<<<<<<<
 *             if not isinstance(cwd, bytes):
 *                 raise ValueError('cwd must be a str or bytes object')
 */
      __pyx_t_1 = PyUnicode_EncodeFSDefault(__pyx_v_cwd); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 276, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_cwd, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":275
 *             cwd = os_fspath(cwd)
 * 
 *             if isinstance(cwd, str):             # <<<<<<<<<<<<<<
 *                 cwd = PyUnicode_EncodeFSDefault(cwd)
 *             if not isinstance(cwd, bytes):
 */
    }

    /* "uvloop/handles/process.pyx":277
 *             if isinstance(cwd, str):
 *                 cwd = PyUnicode_EncodeFSDefault(cwd)
 *             if not isinstance(cwd, bytes):             # <<<<<<<<<<<<<<
 *                 raise ValueError('cwd must be a str or bytes object')
 * 
 */
    __pyx_t_4 = PyBytes_Check(__pyx_v_cwd); 
    __pyx_t_8 = (!__pyx_t_4);
    if (unlikely(__pyx_t_8)) {

      /* "uvloop/handles/process.pyx":278
 *                 cwd = PyUnicode_EncodeFSDefault(cwd)
 *             if not isinstance(cwd, bytes):
 *                 raise ValueError('cwd must be a str or bytes object')             # <<<<<<<<<<<<<<
 * 
 *             self.__cwd = cwd
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__165, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(4, 278, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":277
 *             if isinstance(cwd, str):
 *                 cwd = PyUnicode_EncodeFSDefault(cwd)
 *             if not isinstance(cwd, bytes):             # <<<<<<<<<<<<<<
 *                 raise ValueError('cwd must be a str or bytes object')
 * 
 */
    }

    /* "uvloop/handles/process.pyx":280
 *                 raise ValueError('cwd must be a str or bytes object')
 * 
 *             self.__cwd = cwd             # <<<<<<<<<<<<<<
 *             self.options.cwd = PyBytes_AsString(self.__cwd)
 * 
 */
    if (!(likely(PyBytes_CheckExact(__pyx_v_cwd))||((__pyx_v_cwd) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_v_cwd))) __PYX_ERR(4, 280, __pyx_L1_error)
    __pyx_t_1 = __pyx_v_cwd;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_UVProcess__cwd);
    __Pyx_DECREF(__pyx_v_self->_UVProcess__cwd);
    __pyx_v_self->_UVProcess__cwd = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/handles/process.pyx":281
 * 
 *             self.__cwd = cwd
 *             self.options.cwd = PyBytes_AsString(self.__cwd)             # <<<<<<<<<<<<<<
 * 
 *         self.options.exit_cb = &__uvprocess_on_exit_callback
 */
    __pyx_t_1 = __pyx_v_self->_UVProcess__cwd;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = PyBytes_AsString(__pyx_t_1); if (unlikely(__pyx_t_3 == ((char *)NULL))) __PYX_ERR(4, 281, __pyx_L1_error)
    __pyx_v_self->options.cwd = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/process.pyx":272
 *             self.options.uid = uv.getuid()
 * 
 *         if cwd is not None:             # <<<<<<<<<<<<<<
 *             cwd = os_fspath(cwd)
 * 
 */
  }

  /* "uvloop/handles/process.pyx":283
 *             self.options.cwd = PyBytes_AsString(self.__cwd)
 * 
 *         self.options.exit_cb = &__uvprocess_on_exit_callback             # <<<<<<<<<<<<<<
 * 
 *         self._init_files(_stdin, _stdout, _stderr)
 */
  __pyx_v_self->options.exit_cb = (&__pyx_f_6uvloop_4loop___uvprocess_on_exit_callback);

  /* "uvloop/handles/process.pyx":285
 *         self.options.exit_cb = &__uvprocess_on_exit_callback
 * 
 *         self._init_files(_stdin, _stdout, _stderr)             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_args(self, list args):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->_init_files(__pyx_v_self, __pyx_v__stdin, __pyx_v__stdout, __pyx_v__stderr); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":242
 *         return ret
 * 
 *     cdef _init_options(self, list args, dict env, cwd, start_new_session,             # <<<<<<<<<<<<<<
 *                        _stdin, _stdout, _stderr, bint force_fork):
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._init_options", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cwd);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":287
 *         self._init_files(_stdin, _stdout, _stderr)
 * 
 *     cdef _init_args(self, list args):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes path
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_args(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_v_path = 0;
  int __pyx_v_an;
  int __pyx_v_i;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_t_11;
  char *__pyx_t_12;
  char **__pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init_args", 1);

  /* "uvloop/handles/process.pyx":290
 *         cdef:
 *             bytes path
 *             int an = len(args)             # <<<<<<<<<<<<<<
 * 
 *         if an < 1:
 */
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(4, 290, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_args); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(4, 290, __pyx_L1_error)
  __pyx_v_an = __pyx_t_1;

  /* "uvloop/handles/process.pyx":292
 *             int an = len(args)
 * 
 *         if an < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError('cannot spawn a process: args are empty')
 * 
 */
  __pyx_t_2 = (__pyx_v_an < 1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/process.pyx":293
 * 
 *         if an < 1:
 *             raise ValueError('cannot spawn a process: args are empty')             # <<<<<<<<<<<<<<
 * 
 *         self.__args = args.copy()
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__166, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 293, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":292
 *             int an = len(args)
 * 
 *         if an < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError('cannot spawn a process: args are empty')
 * 
 */
  }

  /* "uvloop/handles/process.pyx":295
 *             raise ValueError('cannot spawn a process: args are empty')
 * 
 *         self.__args = args.copy()             # <<<<<<<<<<<<<<
 *         for i in range(an):
 *             arg = os_fspath(args[i])
 */
  __pyx_t_3 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_copy, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_3))) __PYX_ERR(4, 295, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_UVProcess__args);
  __Pyx_DECREF(__pyx_v_self->_UVProcess__args);
  __pyx_v_self->_UVProcess__args = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":296
 * 
 *         self.__args = args.copy()
 *         for i in range(an):             # <<<<<<<<<<<<<<
 *             arg = os_fspath(args[i])
 *             if isinstance(arg, str):
 */
  __pyx_t_4 = __pyx_v_an;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "uvloop/handles/process.pyx":297
 *         self.__args = args.copy()
 *         for i in range(an):
 *             arg = os_fspath(args[i])             # <<<<<<<<<<<<<<
 *             if isinstance(arg, str):
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)
 */
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 297, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_args, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_fspath);
    __pyx_t_8 = __pyx_v_6uvloop_4loop_os_fspath; __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_7};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 297, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/handles/process.pyx":298
 *         for i in range(an):
 *             arg = os_fspath(args[i])
 *             if isinstance(arg, str):             # <<<<<<<<<<<<<<
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)
 *             elif not isinstance(arg, bytes):
 */
    __pyx_t_2 = PyUnicode_Check(__pyx_v_arg); 
    if (__pyx_t_2) {

      /* "uvloop/handles/process.pyx":299
 *             arg = os_fspath(args[i])
 *             if isinstance(arg, str):
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)             # <<<<<<<<<<<<<<
 *             elif not isinstance(arg, bytes):
 *                 raise TypeError('all args must be str or bytes')
 */
      __pyx_t_3 = PyUnicode_EncodeFSDefault(__pyx_v_arg); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__pyx_v_self->_UVProcess__args == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(4, 299, __pyx_L1_error)
      }
      if (unlikely((__Pyx_SetItemInt(__pyx_v_self->_UVProcess__args, __pyx_v_i, __pyx_t_3, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(4, 299, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/process.pyx":298
 *         for i in range(an):
 *             arg = os_fspath(args[i])
 *             if isinstance(arg, str):             # <<<<<<<<<<<<<<
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)
 *             elif not isinstance(arg, bytes):
 */
      goto __pyx_L6;
    }

    /* "uvloop/handles/process.pyx":300
 *             if isinstance(arg, str):
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)
 *             elif not isinstance(arg, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError('all args must be str or bytes')
 * 
 */
    __pyx_t_2 = PyBytes_Check(__pyx_v_arg); 
    __pyx_t_11 = (!__pyx_t_2);
    if (unlikely(__pyx_t_11)) {

      /* "uvloop/handles/process.pyx":301
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)
 *             elif not isinstance(arg, bytes):
 *                 raise TypeError('all args must be str or bytes')             # <<<<<<<<<<<<<<
 * 
 *         path = self.__args[0]
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__167, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(4, 301, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":300
 *             if isinstance(arg, str):
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)
 *             elif not isinstance(arg, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError('all args must be str or bytes')
 * 
 */
    }
    __pyx_L6:;
  }

  /* "uvloop/handles/process.pyx":303
 *                 raise TypeError('all args must be str or bytes')
 * 
 *         path = self.__args[0]             # <<<<<<<<<<<<<<
 *         self.uv_opt_file = PyBytes_AsString(path)
 *         self.uv_opt_args = self.__to_cstring_array(self.__args)
 */
  if (unlikely(__pyx_v_self->_UVProcess__args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(4, 303, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_v_self->_UVProcess__args, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_3))) __PYX_ERR(4, 303, __pyx_L1_error)
  __pyx_v_path = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":304
 * 
 *         path = self.__args[0]
 *         self.uv_opt_file = PyBytes_AsString(path)             # <<<<<<<<<<<<<<
 *         self.uv_opt_args = self.__to_cstring_array(self.__args)
 * 
 */
  __pyx_t_12 = PyBytes_AsString(__pyx_v_path); if (unlikely(__pyx_t_12 == ((char *)NULL))) __PYX_ERR(4, 304, __pyx_L1_error)
  __pyx_v_self->uv_opt_file = __pyx_t_12;

  /* "uvloop/handles/process.pyx":305
 *         path = self.__args[0]
 *         self.uv_opt_file = PyBytes_AsString(path)
 *         self.uv_opt_args = self.__to_cstring_array(self.__args)             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_env(self, dict env):
 */
  __pyx_t_3 = __pyx_v_self->_UVProcess__args;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_13 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->_UVProcess__to_cstring_array(__pyx_v_self, ((PyObject*)__pyx_t_3)); if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->uv_opt_args = __pyx_t_13;

  /* "uvloop/handles/process.pyx":287
 *         self._init_files(_stdin, _stdout, _stderr)
 * 
 *     cdef _init_args(self, list args):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes path
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._init_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":307
 *         self.uv_opt_args = self.__to_cstring_array(self.__args)
 * 
 *     cdef _init_env(self, dict env):             # <<<<<<<<<<<<<<
 *         if env is not None:
 *             self.__env = list()
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_env(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, PyObject *__pyx_v_env) {
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char **__pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init_env", 1);

  /* "uvloop/handles/process.pyx":308
 * 
 *     cdef _init_env(self, dict env):
 *         if env is not None:             # <<<<<<<<<<<<<<
 *             self.__env = list()
 *             for key in env:
 */
  __pyx_t_1 = (__pyx_v_env != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":309
 *     cdef _init_env(self, dict env):
 *         if env is not None:
 *             self.__env = list()             # <<<<<<<<<<<<<<
 *             for key in env:
 *                 val = env[key]
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_UVProcess__env);
    __Pyx_DECREF(__pyx_v_self->_UVProcess__env);
    __pyx_v_self->_UVProcess__env = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":310
 *         if env is not None:
 *             self.__env = list()
 *             for key in env:             # <<<<<<<<<<<<<<
 *                 val = env[key]
 * 
 */
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_env == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(4, 310, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_env, 1, ((PyObject *)NULL), (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, NULL, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_7 == 0)) break;
      if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(4, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "uvloop/handles/process.pyx":311
 *             self.__env = list()
 *             for key in env:
 *                 val = env[key]             # <<<<<<<<<<<<<<
 * 
 *                 if isinstance(key, str):
 */
      if (unlikely(__pyx_v_env == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(4, 311, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_env, __pyx_v_key); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "uvloop/handles/process.pyx":313
 *                 val = env[key]
 * 
 *                 if isinstance(key, str):             # <<<<<<<<<<<<<<
 *                     key = PyUnicode_EncodeFSDefault(key)
 *                 elif not isinstance(key, bytes):
 */
      __pyx_t_1 = PyUnicode_Check(__pyx_v_key); 
      if (__pyx_t_1) {

        /* "uvloop/handles/process.pyx":314
 * 
 *                 if isinstance(key, str):
 *                     key = PyUnicode_EncodeFSDefault(key)             # <<<<<<<<<<<<<<
 *                 elif not isinstance(key, bytes):
 *                     raise TypeError(
 */
        __pyx_t_6 = PyUnicode_EncodeFSDefault(__pyx_v_key); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 314, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_key, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "uvloop/handles/process.pyx":313
 *                 val = env[key]
 * 
 *                 if isinstance(key, str):             # <<<<<<<<<<<<<<
 *                     key = PyUnicode_EncodeFSDefault(key)
 *                 elif not isinstance(key, bytes):
 */
        goto __pyx_L6;
      }

      /* "uvloop/handles/process.pyx":315
 *                 if isinstance(key, str):
 *                     key = PyUnicode_EncodeFSDefault(key)
 *                 elif not isinstance(key, bytes):             # <<<<<<<<<<<<<<
 *                     raise TypeError(
 *                         'all environment vars must be bytes or str')
 */
      __pyx_t_1 = PyBytes_Check(__pyx_v_key); 
      __pyx_t_8 = (!__pyx_t_1);
      if (unlikely(__pyx_t_8)) {

        /* "uvloop/handles/process.pyx":316
 *                     key = PyUnicode_EncodeFSDefault(key)
 *                 elif not isinstance(key, bytes):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'all environment vars must be bytes or str')
 * 
 */
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__168, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 316, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(4, 316, __pyx_L1_error)

        /* "uvloop/handles/process.pyx":315
 *                 if isinstance(key, str):
 *                     key = PyUnicode_EncodeFSDefault(key)
 *                 elif not isinstance(key, bytes):             # <<<<<<<<<<<<<<
 *                     raise TypeError(
 *                         'all environment vars must be bytes or str')
 */
      }
      __pyx_L6:;

      /* "uvloop/handles/process.pyx":319
 *                         'all environment vars must be bytes or str')
 * 
 *                 if isinstance(val, str):             # <<<<<<<<<<<<<<
 *                     val = PyUnicode_EncodeFSDefault(val)
 *                 elif not isinstance(val, bytes):
 */
      __pyx_t_8 = PyUnicode_Check(__pyx_v_val); 
      if (__pyx_t_8) {

        /* "uvloop/handles/process.pyx":320
 * 
 *                 if isinstance(val, str):
 *                     val = PyUnicode_EncodeFSDefault(val)             # <<<<<<<<<<<<<<
 *                 elif not isinstance(val, bytes):
 *                     raise TypeError(
 */
        __pyx_t_6 = PyUnicode_EncodeFSDefault(__pyx_v_val); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "uvloop/handles/process.pyx":319
 *                         'all environment vars must be bytes or str')
 * 
 *                 if isinstance(val, str):             # <<<<<<<<<<<<<<
 *                     val = PyUnicode_EncodeFSDefault(val)
 *                 elif not isinstance(val, bytes):
 */
        goto __pyx_L7;
      }

      /* "uvloop/handles/process.pyx":321
 *                 if isinstance(val, str):
 *                     val = PyUnicode_EncodeFSDefault(val)
 *                 elif not isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                     raise TypeError(
 *                         'all environment values must be bytes or str')
 */
      __pyx_t_8 = PyBytes_Check(__pyx_v_val); 
      __pyx_t_1 = (!__pyx_t_8);
      if (unlikely(__pyx_t_1)) {

        /* "uvloop/handles/process.pyx":322
 *                     val = PyUnicode_EncodeFSDefault(val)
 *                 elif not isinstance(val, bytes):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'all environment values must be bytes or str')
 * 
 */
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__169, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 322, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_Raise(__pyx_t_6, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __PYX_ERR(4, 322, __pyx_L1_error)

        /* "uvloop/handles/process.pyx":321
 *                 if isinstance(val, str):
 *                     val = PyUnicode_EncodeFSDefault(val)
 *                 elif not isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                     raise TypeError(
 *                         'all environment values must be bytes or str')
 */
      }
      __pyx_L7:;

      /* "uvloop/handles/process.pyx":325
 *                         'all environment values must be bytes or str')
 * 
 *                 self.__env.append(key + b'=' + val)             # <<<<<<<<<<<<<<
 * 
 *             self.uv_opt_env = self.__to_cstring_array(self.__env)
 */
      if (unlikely(__pyx_v_self->_UVProcess__env == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(4, 325, __pyx_L1_error)
      }
      __pyx_t_6 = PyNumber_Add(__pyx_v_key, __pyx_kp_b__170); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = PyNumber_Add(__pyx_t_6, __pyx_v_val); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_self->_UVProcess__env, __pyx_t_9); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(4, 325, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":327
 *                 self.__env.append(key + b'=' + val)
 * 
 *             self.uv_opt_env = self.__to_cstring_array(self.__env)             # <<<<<<<<<<<<<<
 *         else:
 *             self.__env = None
 */
    __pyx_t_2 = __pyx_v_self->_UVProcess__env;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->_UVProcess__to_cstring_array(__pyx_v_self, ((PyObject*)__pyx_t_2)); if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 327, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_self->uv_opt_env = __pyx_t_11;

    /* "uvloop/handles/process.pyx":308
 * 
 *     cdef _init_env(self, dict env):
 *         if env is not None:             # <<<<<<<<<<<<<<
 *             self.__env = list()
 *             for key in env:
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":329
 *             self.uv_opt_env = self.__to_cstring_array(self.__env)
 *         else:
 *             self.__env = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_files(self, _stdin, _stdout, _stderr):
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_UVProcess__env);
    __Pyx_DECREF(__pyx_v_self->_UVProcess__env);
    __pyx_v_self->_UVProcess__env = ((PyObject*)Py_None);
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":307
 *         self.uv_opt_args = self.__to_cstring_array(self.__args)
 * 
 *     cdef _init_env(self, dict env):             # <<<<<<<<<<<<<<
 *         if env is not None:
 *             self.__env = list()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._init_env", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":331
 *             self.__env = None
 * 
 *     cdef _init_files(self, _stdin, _stdout, _stderr):             # <<<<<<<<<<<<<<
 *         self.options.stdio_count = 0
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__init_files(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__stdin, CYTHON_UNUSED PyObject *__pyx_v__stdout, CYTHON_UNUSED PyObject *__pyx_v__stderr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_init_files", 1);

  /* "uvloop/handles/process.pyx":332
 * 
 *     cdef _init_files(self, _stdin, _stdout, _stderr):
 *         self.options.stdio_count = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef _kill(self, int signum):
 */
  __pyx_v_self->options.stdio_count = 0;

  /* "uvloop/handles/process.pyx":331
 *             self.__env = None
 * 
 *     cdef _init_files(self, _stdin, _stdout, _stderr):             # <<<<<<<<<<<<<<
 *         self.options.stdio_count = 0
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":334
 *         self.options.stdio_count = 0
 * 
 *     cdef _kill(self, int signum):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__kill(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, int __pyx_v_signum) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_kill", 1);

  /* "uvloop/handles/process.pyx":336
 *     cdef _kill(self, int signum):
 *         cdef int err
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 *         err = uv.uv_process_kill(<uv.uv_process_t*>self._handle, signum)
 *         if err < 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":337
 *         cdef int err
 *         self._ensure_alive()
 *         err = uv.uv_process_kill(<uv.uv_process_t*>self._handle, signum)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
  __pyx_v_err = uv_process_kill(((uv_process_t *)__pyx_v_self->__pyx_base._handle), __pyx_v_signum);

  /* "uvloop/handles/process.pyx":338
 *         self._ensure_alive()
 *         err = uv.uv_process_kill(<uv.uv_process_t*>self._handle, signum)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/process.pyx":339
 *         err = uv.uv_process_kill(<uv.uv_process_t*>self._handle, signum)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(4, 339, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":338
 *         self._ensure_alive()
 *         err = uv.uv_process_kill(<uv.uv_process_t*>self._handle, signum)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
  }

  /* "uvloop/handles/process.pyx":334
 *         self.options.stdio_count = 0
 * 
 *     cdef _kill(self, int signum):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         self._ensure_alive()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._kill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":341
 *             raise convert_error(err)
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):             # <<<<<<<<<<<<<<
 *         if term_signal:
 *             # From Python docs:
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__on_exit(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, int64_t __pyx_v_exit_status, int __pyx_v_term_signal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_exit", 1);

  /* "uvloop/handles/process.pyx":342
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):
 *         if term_signal:             # <<<<<<<<<<<<<<
 *             # From Python docs:
 *             #    A negative value -N indicates that the child was
 */
  __pyx_t_1 = (__pyx_v_term_signal != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":346
 *             #    A negative value -N indicates that the child was
 *             #    terminated by signal N (POSIX only).
 *             self._returncode = -term_signal             # <<<<<<<<<<<<<<
 *         else:
 *             self._returncode = exit_status
 */
    __pyx_t_2 = __Pyx_PyInt_From_int((-__pyx_v_term_signal)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_returncode);
    __Pyx_DECREF(__pyx_v_self->_returncode);
    __pyx_v_self->_returncode = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":342
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):
 *         if term_signal:             # <<<<<<<<<<<<<<
 *             # From Python docs:
 *             #    A negative value -N indicates that the child was
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":348
 *             self._returncode = -term_signal
 *         else:
 *             self._returncode = exit_status             # <<<<<<<<<<<<<<
 * 
 *         self._close()
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_exit_status); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_returncode);
    __Pyx_DECREF(__pyx_v_self->_returncode);
    __pyx_v_self->_returncode = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":350
 *             self._returncode = exit_status
 * 
 *         self._close()             # <<<<<<<<<<<<<<
 * 
 *     cdef _close(self):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":341
 *             raise convert_error(err)
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):             # <<<<<<<<<<<<<<
 *         if term_signal:
 *             # From Python docs:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._on_exit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":352
 *         self._close()
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         try:
 *             if self._loop is not None:
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVProcess__close(struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  char const *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/handles/process.pyx":353
 * 
 *     cdef _close(self):
 *         try:             # <<<<<<<<<<<<<<
 *             if self._loop is not None:
 *                 self._loop._untrack_process(self)
 */
  /*try:*/ {

    /* "uvloop/handles/process.pyx":354
 *     cdef _close(self):
 *         try:
 *             if self._loop is not None:             # <<<<<<<<<<<<<<
 *                 self._loop._untrack_process(self)
 *         finally:
 */
    __pyx_t_1 = (((PyObject *)__pyx_v_self->__pyx_base._loop) != Py_None);
    if (__pyx_t_1) {

      /* "uvloop/handles/process.pyx":355
 *         try:
 *             if self._loop is not None:
 *                 self._loop._untrack_process(self)             # <<<<<<<<<<<<<<
 *         finally:
 *             UVHandle._close(self)
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->__pyx_base._loop->__pyx_vtab)->_untrack_process(__pyx_v_self->__pyx_base._loop, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 355, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/process.pyx":354
 *     cdef _close(self):
 *         try:
 *             if self._loop is not None:             # <<<<<<<<<<<<<<
 *                 self._loop._untrack_process(self)
 *         finally:
 */
    }
  }

  /* "uvloop/handles/process.pyx":357
 *                 self._loop._untrack_process(self)
 *         finally:
 *             UVHandle._close(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = __pyx_f_6uvloop_4loop_8UVHandle__close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __pyx_t_3 = __pyx_lineno; __pyx_t_4 = __pyx_clineno; __pyx_t_5 = __pyx_filename;
      {
        __pyx_t_2 = __pyx_f_6uvloop_4loop_8UVHandle__close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 357, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __pyx_lineno = __pyx_t_3; __pyx_clineno = __pyx_t_4; __pyx_filename = __pyx_t_5;
      goto __pyx_L1_error;
      __pyx_L8_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "uvloop/handles/process.pyx":352
 *         self._close()
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         try:
 *             if self._loop is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UVProcess._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVProcess_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVProcess_4__reduce_cython__, "UVProcess.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVProcess_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVProcess_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVProcess_4__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9UVProcess_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVProcess_4__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVProcess_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVProcess.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVProcess_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVProcess_6__setstate_cython__, "UVProcess.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVProcess_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVProcess_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVProcess_6__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9UVProcess_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVProcess.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVProcess_6__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVProcess_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVProcess.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":368
 * @cython.no_gc_clear
 * cdef class UVProcessTransport(UVProcess):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._exit_waiters = []
 *         self._protocol = None
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_18UVProcessTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_18UVProcessTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport___cinit__(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_18UVProcessTransport___cinit__(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/process.pyx":369
 * cdef class UVProcessTransport(UVProcess):
 *     def __cinit__(self):
 *         self._exit_waiters = []             # <<<<<<<<<<<<<<
 *         self._protocol = None
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_exit_waiters);
  __Pyx_DECREF(__pyx_v_self->_exit_waiters);
  __pyx_v_self->_exit_waiters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":370
 *     def __cinit__(self):
 *         self._exit_waiters = []
 *         self._protocol = None             # <<<<<<<<<<<<<<
 * 
 *         self._init_futs = []
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_protocol);
  __Pyx_DECREF(__pyx_v_self->_protocol);
  __pyx_v_self->_protocol = Py_None;

  /* "uvloop/handles/process.pyx":372
 *         self._protocol = None
 * 
 *         self._init_futs = []             # <<<<<<<<<<<<<<
 *         self._pending_calls = []
 *         self._stdio_ready = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_init_futs);
  __Pyx_DECREF(__pyx_v_self->_init_futs);
  __pyx_v_self->_init_futs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":373
 * 
 *         self._init_futs = []
 *         self._pending_calls = []             # <<<<<<<<<<<<<<
 *         self._stdio_ready = 0
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_pending_calls);
  __Pyx_DECREF(__pyx_v_self->_pending_calls);
  __pyx_v_self->_pending_calls = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":374
 *         self._init_futs = []
 *         self._pending_calls = []
 *         self._stdio_ready = 0             # <<<<<<<<<<<<<<
 * 
 *         self._stdin = self._stdout = self._stderr = None
 */
  __pyx_v_self->_stdio_ready = 0;

  /* "uvloop/handles/process.pyx":376
 *         self._stdio_ready = 0
 * 
 *         self._stdin = self._stdout = self._stderr = None             # <<<<<<<<<<<<<<
 *         self.stdin_proto = self.stdout_proto = self.stderr_proto = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_stdin);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_stdin);
  __pyx_v_self->_stdin = ((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)Py_None);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_stdout);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_stdout);
  __pyx_v_self->_stdout = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)Py_None);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_stderr);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_stderr);
  __pyx_v_self->_stderr = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)Py_None);

  /* "uvloop/handles/process.pyx":377
 * 
 *         self._stdin = self._stdout = self._stderr = None
 *         self.stdin_proto = self.stdout_proto = self.stderr_proto = None             # <<<<<<<<<<<<<<
 * 
 *         self._finished = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->stdin_proto);
  __Pyx_DECREF(__pyx_v_self->stdin_proto);
  __pyx_v_self->stdin_proto = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->stdout_proto);
  __Pyx_DECREF(__pyx_v_self->stdout_proto);
  __pyx_v_self->stdout_proto = Py_None;
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->stderr_proto);
  __Pyx_DECREF(__pyx_v_self->stderr_proto);
  __pyx_v_self->stderr_proto = Py_None;

  /* "uvloop/handles/process.pyx":379
 *         self.stdin_proto = self.stdout_proto = self.stderr_proto = None
 * 
 *         self._finished = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):
 */
  __pyx_v_self->_finished = 0;

  /* "uvloop/handles/process.pyx":368
 * @cython.no_gc_clear
 * cdef class UVProcessTransport(UVProcess):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._exit_waiters = []
 *         self._protocol = None
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":381
 *         self._finished = 0
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):             # <<<<<<<<<<<<<<
 *         UVProcess._on_exit(self, exit_status, term_signal)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__on_exit(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int64_t __pyx_v_exit_status, int __pyx_v_term_signal) {
  PyObject *__pyx_v_waiter = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_exit", 1);

  /* "uvloop/handles/process.pyx":382
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):
 *         UVProcess._on_exit(self, exit_status, term_signal)             # <<<<<<<<<<<<<<
 * 
 *         if self._stdio_ready:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_9UVProcess__on_exit(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_v_exit_status, __pyx_v_term_signal); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":384
 *         UVProcess._on_exit(self, exit_status, term_signal)
 * 
 *         if self._stdio_ready:             # <<<<<<<<<<<<<<
 *             self._loop.call_soon(self._protocol.process_exited,
 *                                  context=self.context)
 */
  if (__pyx_v_self->_stdio_ready) {

    /* "uvloop/handles/process.pyx":385
 * 
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.process_exited,             # <<<<<<<<<<<<<<
 *                                  context=self.context)
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_process_exited); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(4, 385, __pyx_L1_error);
    __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":386
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.process_exited,
 *                                  context=self.context)             # <<<<<<<<<<<<<<
 *         else:
 *             self._pending_calls.append((_CALL_PROCESS_EXITED, None, None))
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_context, __pyx_v_self->__pyx_base.__pyx_base.context) < 0) __PYX_ERR(4, 386, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":385
 * 
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.process_exited,             # <<<<<<<<<<<<<<
 *                                  context=self.context)
 *         else:
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/process.pyx":384
 *         UVProcess._on_exit(self, exit_status, term_signal)
 * 
 *         if self._stdio_ready:             # <<<<<<<<<<<<<<
 *             self._loop.call_soon(self._protocol.process_exited,
 *                                  context=self.context)
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":388
 *                                  context=self.context)
 *         else:
 *             self._pending_calls.append((_CALL_PROCESS_EXITED, None, None))             # <<<<<<<<<<<<<<
 * 
 *         self._try_finish()
 */
  /*else*/ {
    if (unlikely(__pyx_v_self->_pending_calls == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(4, 388, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->_pending_calls, __pyx_tuple__171); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(4, 388, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":390
 *             self._pending_calls.append((_CALL_PROCESS_EXITED, None, None))
 * 
 *         self._try_finish()             # <<<<<<<<<<<<<<
 * 
 *         for waiter in self._exit_waiters:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_try_finish(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/handles/process.pyx":392
 *         self._try_finish()
 * 
 *         for waiter in self._exit_waiters:             # <<<<<<<<<<<<<<
 *             if not waiter.cancelled():
 *                 waiter.set_result(self._returncode)
 */
  if (unlikely(__pyx_v_self->_exit_waiters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(4, 392, __pyx_L1_error)
  }
  __pyx_t_4 = __pyx_v_self->_exit_waiters; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_6 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 392, __pyx_L1_error)
      #endif
      if (__pyx_t_6 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(4, 392, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_waiter, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":393
 * 
 *         for waiter in self._exit_waiters:
 *             if not waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 waiter.set_result(self._returncode)
 *         self._exit_waiters.clear()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(4, 393, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = (!__pyx_t_8);
    if (__pyx_t_9) {

      /* "uvloop/handles/process.pyx":394
 *         for waiter in self._exit_waiters:
 *             if not waiter.cancelled():
 *                 waiter.set_result(self._returncode)             # <<<<<<<<<<<<<<
 *         self._exit_waiters.clear()
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_self->__pyx_base._returncode};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/process.pyx":393
 * 
 *         for waiter in self._exit_waiters:
 *             if not waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 waiter.set_result(self._returncode)
 *         self._exit_waiters.clear()
 */
    }

    /* "uvloop/handles/process.pyx":392
 *         self._try_finish()
 * 
 *         for waiter in self._exit_waiters:             # <<<<<<<<<<<<<<
 *             if not waiter.cancelled():
 *                 waiter.set_result(self._returncode)
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/handles/process.pyx":395
 *             if not waiter.cancelled():
 *                 waiter.set_result(self._returncode)
 *         self._exit_waiters.clear()             # <<<<<<<<<<<<<<
 * 
 *         self._close()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_exit_waiters, __pyx_n_s_clear); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/handles/process.pyx":397
 *         self._exit_waiters.clear()
 * 
 *         self._close()             # <<<<<<<<<<<<<<
 * 
 *     cdef _check_proc(self):
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/handles/process.pyx":381
 *         self._finished = 0
 * 
 *     cdef _on_exit(self, int64_t exit_status, int term_signal):             # <<<<<<<<<<<<<<
 *         UVProcess._on_exit(self, exit_status, term_signal)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._on_exit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_waiter);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":399
 *         self._close()
 * 
 *     cdef _check_proc(self):             # <<<<<<<<<<<<<<
 *         if not self._is_alive() or self._returncode is not None:
 *             raise ProcessLookupError()
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__check_proc(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_proc", 1);

  /* "uvloop/handles/process.pyx":400
 * 
 *     cdef _check_proc(self):
 *         if not self._is_alive() or self._returncode is not None:             # <<<<<<<<<<<<<<
 *             raise ProcessLookupError()
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 400, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_self->__pyx_base._returncode != Py_None);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/process.pyx":401
 *     cdef _check_proc(self):
 *         if not self._is_alive() or self._returncode is not None:
 *             raise ProcessLookupError()             # <<<<<<<<<<<<<<
 * 
 *     cdef _pipe_connection_lost(self, int fd, exc):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ProcessLookupError); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(4, 401, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":400
 * 
 *     cdef _check_proc(self):
 *         if not self._is_alive() or self._returncode is not None:             # <<<<<<<<<<<<<<
 *             raise ProcessLookupError()
 * 
 */
  }

  /* "uvloop/handles/process.pyx":399
 *         self._close()
 * 
 *     cdef _check_proc(self):             # <<<<<<<<<<<<<<
 *         if not self._is_alive() or self._returncode is not None:
 *             raise ProcessLookupError()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._check_proc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":403
 *             raise ProcessLookupError()
 * 
 *     cdef _pipe_connection_lost(self, int fd, exc):             # <<<<<<<<<<<<<<
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__pipe_connection_lost(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_fd, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pipe_connection_lost", 1);

  /* "uvloop/handles/process.pyx":404
 * 
 *     cdef _pipe_connection_lost(self, int fd, exc):
 *         if self._stdio_ready:             # <<<<<<<<<<<<<<
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,
 *                                  context=self.context)
 */
  if (__pyx_v_self->_stdio_ready) {

    /* "uvloop/handles/process.pyx":405
 *     cdef _pipe_connection_lost(self, int fd, exc):
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,             # <<<<<<<<<<<<<<
 *                                  context=self.context)
 *             self._try_finish()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_pipe_connection_lost); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(4, 405, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(4, 405, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_exc);
    __Pyx_GIVEREF(__pyx_v_exc);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_exc)) __PYX_ERR(4, 405, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;

    /* "uvloop/handles/process.pyx":406
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,
 *                                  context=self.context)             # <<<<<<<<<<<<<<
 *             self._try_finish()
 *         else:
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_context, __pyx_v_self->__pyx_base.__pyx_base.context) < 0) __PYX_ERR(4, 406, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":405
 *     cdef _pipe_connection_lost(self, int fd, exc):
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,             # <<<<<<<<<<<<<<
 *                                  context=self.context)
 *             self._try_finish()
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":407
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,
 *                                  context=self.context)
 *             self._try_finish()             # <<<<<<<<<<<<<<
 *         else:
 *             self._pending_calls.append((_CALL_PIPE_CONNECTION_LOST, fd, exc))
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_try_finish(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":404
 * 
 *     cdef _pipe_connection_lost(self, int fd, exc):
 *         if self._stdio_ready:             # <<<<<<<<<<<<<<
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,
 *                                  context=self.context)
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":409
 *             self._try_finish()
 *         else:
 *             self._pending_calls.append((_CALL_PIPE_CONNECTION_LOST, fd, exc))             # <<<<<<<<<<<<<<
 * 
 *     cdef _pipe_data_received(self, int fd, data):
 */
  /*else*/ {
    if (unlikely(__pyx_v_self->_pending_calls == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(4, 409, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_1)) __PYX_ERR(4, 409, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(4, 409, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_exc);
    __Pyx_GIVEREF(__pyx_v_exc);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_exc)) __PYX_ERR(4, 409, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->_pending_calls, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(4, 409, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":403
 *             raise ProcessLookupError()
 * 
 *     cdef _pipe_connection_lost(self, int fd, exc):             # <<<<<<<<<<<<<<
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_connection_lost, fd, exc,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._pipe_connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":411
 *             self._pending_calls.append((_CALL_PIPE_CONNECTION_LOST, fd, exc))
 * 
 *     cdef _pipe_data_received(self, int fd, data):             # <<<<<<<<<<<<<<
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_data_received, fd, data,
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__pipe_data_received(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_fd, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pipe_data_received", 1);

  /* "uvloop/handles/process.pyx":412
 * 
 *     cdef _pipe_data_received(self, int fd, data):
 *         if self._stdio_ready:             # <<<<<<<<<<<<<<
 *             self._loop.call_soon(self._protocol.pipe_data_received, fd, data,
 *                                  context=self.context)
 */
  if (__pyx_v_self->_stdio_ready) {

    /* "uvloop/handles/process.pyx":413
 *     cdef _pipe_data_received(self, int fd, data):
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_data_received, fd, data,             # <<<<<<<<<<<<<<
 *                                  context=self.context)
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_pipe_data_received); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(4, 413, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(4, 413, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_data)) __PYX_ERR(4, 413, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;

    /* "uvloop/handles/process.pyx":414
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_data_received, fd, data,
 *                                  context=self.context)             # <<<<<<<<<<<<<<
 *         else:
 *             self._pending_calls.append((_CALL_PIPE_DATA_RECEIVED, fd, data))
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_context, __pyx_v_self->__pyx_base.__pyx_base.context) < 0) __PYX_ERR(4, 414, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":413
 *     cdef _pipe_data_received(self, int fd, data):
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_data_received, fd, data,             # <<<<<<<<<<<<<<
 *                                  context=self.context)
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":412
 * 
 *     cdef _pipe_data_received(self, int fd, data):
 *         if self._stdio_ready:             # <<<<<<<<<<<<<<
 *             self._loop.call_soon(self._protocol.pipe_data_received, fd, data,
 *                                  context=self.context)
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":416
 *                                  context=self.context)
 *         else:
 *             self._pending_calls.append((_CALL_PIPE_DATA_RECEIVED, fd, data))             # <<<<<<<<<<<<<<
 * 
 *     cdef _file_redirect_stdio(self, int fd):
 */
  /*else*/ {
    if (unlikely(__pyx_v_self->_pending_calls == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(4, 416, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_0)) __PYX_ERR(4, 416, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(4, 416, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_data)) __PYX_ERR(4, 416, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->_pending_calls, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(4, 416, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":411
 *             self._pending_calls.append((_CALL_PIPE_CONNECTION_LOST, fd, exc))
 * 
 *     cdef _pipe_data_received(self, int fd, data):             # <<<<<<<<<<<<<<
 *         if self._stdio_ready:
 *             self._loop.call_soon(self._protocol.pipe_data_received, fd, data,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._pipe_data_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":418
 *             self._pending_calls.append((_CALL_PIPE_DATA_RECEIVED, fd, data))
 * 
 *     cdef _file_redirect_stdio(self, int fd):             # <<<<<<<<<<<<<<
 *         fd = os_dup(fd)
 *         os_set_inheritable(fd, True)
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_redirect_stdio(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_file_redirect_stdio", 1);

  /* "uvloop/handles/process.pyx":419
 * 
 *     cdef _file_redirect_stdio(self, int fd):
 *         fd = os_dup(fd)             # <<<<<<<<<<<<<<
 *         os_set_inheritable(fd, True)
 *         self._close_after_spawn(fd)
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_dup);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_os_dup; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 419, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fd = __pyx_t_6;

  /* "uvloop/handles/process.pyx":420
 *     cdef _file_redirect_stdio(self, int fd):
 *         fd = os_dup(fd)
 *         os_set_inheritable(fd, True)             # <<<<<<<<<<<<<<
 *         self._close_after_spawn(fd)
 *         return fd
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __pyx_t_2 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_3, Py_True};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":421
 *         fd = os_dup(fd)
 *         os_set_inheritable(fd, True)
 *         self._close_after_spawn(fd)             # <<<<<<<<<<<<<<
 *         return fd
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._close_after_spawn(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":422
 *         os_set_inheritable(fd, True)
 *         self._close_after_spawn(fd)
 *         return fd             # <<<<<<<<<<<<<<
 * 
 *     cdef _file_devnull(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":418
 *             self._pending_calls.append((_CALL_PIPE_DATA_RECEIVED, fd, data))
 * 
 *     cdef _file_redirect_stdio(self, int fd):             # <<<<<<<<<<<<<<
 *         fd = os_dup(fd)
 *         os_set_inheritable(fd, True)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._file_redirect_stdio", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":424
 *         return fd
 * 
 *     cdef _file_devnull(self):             # <<<<<<<<<<<<<<
 *         dn = os_open(os_devnull, os_O_RDWR)
 *         os_set_inheritable(dn, True)
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_devnull(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_v_dn = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_file_devnull", 1);

  /* "uvloop/handles/process.pyx":425
 * 
 *     cdef _file_devnull(self):
 *         dn = os_open(os_devnull, os_O_RDWR)             # <<<<<<<<<<<<<<
 *         os_set_inheritable(dn, True)
 *         self._close_after_spawn(dn)
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_open);
  __pyx_t_2 = __pyx_v_6uvloop_4loop_os_open; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_6uvloop_4loop_os_devnull, __pyx_v_6uvloop_4loop_os_O_RDWR};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_dn = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":426
 *     cdef _file_devnull(self):
 *         dn = os_open(os_devnull, os_O_RDWR)
 *         os_set_inheritable(dn, True)             # <<<<<<<<<<<<<<
 *         self._close_after_spawn(dn)
 *         return dn
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __pyx_t_2 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_dn, Py_True};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":427
 *         dn = os_open(os_devnull, os_O_RDWR)
 *         os_set_inheritable(dn, True)
 *         self._close_after_spawn(dn)             # <<<<<<<<<<<<<<
 *         return dn
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_dn); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 427, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._close_after_spawn(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":428
 *         os_set_inheritable(dn, True)
 *         self._close_after_spawn(dn)
 *         return dn             # <<<<<<<<<<<<<<
 * 
 *     cdef _file_outpipe(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dn);
  __pyx_r = __pyx_v_dn;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":424
 *         return fd
 * 
 *     cdef _file_devnull(self):             # <<<<<<<<<<<<<<
 *         dn = os_open(os_devnull, os_O_RDWR)
 *         os_set_inheritable(dn, True)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._file_devnull", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dn);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":430
 *         return dn
 * 
 *     cdef _file_outpipe(self):             # <<<<<<<<<<<<<<
 *         r, w = __socketpair()
 *         os_set_inheritable(w, True)
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_outpipe(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_w = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_file_outpipe", 1);

  /* "uvloop/handles/process.pyx":431
 * 
 *     cdef _file_outpipe(self):
 *         r, w = __socketpair()             # <<<<<<<<<<<<<<
 *         os_set_inheritable(w, True)
 *         self._close_after_spawn(w)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___socketpair(); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(4, 431, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(4, 431, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(4, 431, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_r = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_w = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":432
 *     cdef _file_outpipe(self):
 *         r, w = __socketpair()
 *         os_set_inheritable(w, True)             # <<<<<<<<<<<<<<
 *         self._close_after_spawn(w)
 *         return r, w
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_w, Py_True};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":433
 *         r, w = __socketpair()
 *         os_set_inheritable(w, True)
 *         self._close_after_spawn(w)             # <<<<<<<<<<<<<<
 *         return r, w
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_w); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 433, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._close_after_spawn(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":434
 *         os_set_inheritable(w, True)
 *         self._close_after_spawn(w)
 *         return r, w             # <<<<<<<<<<<<<<
 * 
 *     cdef _file_inpipe(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_r);
  __Pyx_GIVEREF(__pyx_v_r);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_r)) __PYX_ERR(4, 434, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_w);
  __Pyx_GIVEREF(__pyx_v_w);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w)) __PYX_ERR(4, 434, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":430
 *         return dn
 * 
 *     cdef _file_outpipe(self):             # <<<<<<<<<<<<<<
 *         r, w = __socketpair()
 *         os_set_inheritable(w, True)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._file_outpipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_w);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":436
 *         return r, w
 * 
 *     cdef _file_inpipe(self):             # <<<<<<<<<<<<<<
 *         r, w = __socketpair()
 *         os_set_inheritable(r, True)
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__file_inpipe(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_w = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_file_inpipe", 1);

  /* "uvloop/handles/process.pyx":437
 * 
 *     cdef _file_inpipe(self):
 *         r, w = __socketpair()             # <<<<<<<<<<<<<<
 *         os_set_inheritable(r, True)
 *         self._close_after_spawn(r)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___socketpair(); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(4, 437, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(4, 437, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(4, 437, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_r = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_w = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":438
 *     cdef _file_inpipe(self):
 *         r, w = __socketpair()
 *         os_set_inheritable(r, True)             # <<<<<<<<<<<<<<
 *         self._close_after_spawn(r)
 *         return r, w
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_r, Py_True};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":439
 *         r, w = __socketpair()
 *         os_set_inheritable(r, True)
 *         self._close_after_spawn(r)             # <<<<<<<<<<<<<<
 *         return r, w
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_r); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 439, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._close_after_spawn(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":440
 *         os_set_inheritable(r, True)
 *         self._close_after_spawn(r)
 *         return r, w             # <<<<<<<<<<<<<<
 * 
 *     cdef _init_files(self, _stdin, _stdout, _stderr):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_r);
  __Pyx_GIVEREF(__pyx_v_r);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_r)) __PYX_ERR(4, 440, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_w);
  __Pyx_GIVEREF(__pyx_v_w);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w)) __PYX_ERR(4, 440, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":436
 *         return r, w
 * 
 *     cdef _file_inpipe(self):             # <<<<<<<<<<<<<<
 *         r, w = __socketpair()
 *         os_set_inheritable(r, True)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._file_inpipe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_w);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":442
 *         return r, w
 * 
 *     cdef _init_files(self, _stdin, _stdout, _stderr):             # <<<<<<<<<<<<<<
 *         cdef uv.uv_stdio_container_t *iocnt
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__init_files(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr) {
  uv_stdio_container_t *__pyx_v_iocnt;
  PyObject *__pyx_v_io = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_v_w = NULL;
  PyObject *__pyx_v_waiter = NULL;
  long __pyx_v_idx;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uv_stdio_container_t *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  long __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init_files", 1);

  /* "uvloop/handles/process.pyx":445
 *         cdef uv.uv_stdio_container_t *iocnt
 * 
 *         UVProcess._init_files(self, _stdin, _stdout, _stderr)             # <<<<<<<<<<<<<<
 * 
 *         io = [None, None, None]
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_9UVProcess__init_files(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_v__stdin, __pyx_v__stdout, __pyx_v__stderr); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":447
 *         UVProcess._init_files(self, _stdin, _stdout, _stderr)
 * 
 *         io = [None, None, None]             # <<<<<<<<<<<<<<
 * 
 *         self.options.stdio_count = 3
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, Py_None)) __PYX_ERR(4, 447, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, Py_None)) __PYX_ERR(4, 447, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, Py_None)) __PYX_ERR(4, 447, __pyx_L1_error);
  __pyx_v_io = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":449
 *         io = [None, None, None]
 * 
 *         self.options.stdio_count = 3             # <<<<<<<<<<<<<<
 *         self.options.stdio = self.iocnt
 * 
 */
  __pyx_v_self->__pyx_base.options.stdio_count = 3;

  /* "uvloop/handles/process.pyx":450
 * 
 *         self.options.stdio_count = 3
 *         self.options.stdio = self.iocnt             # <<<<<<<<<<<<<<
 * 
 *         if _stdin is not None:
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.iocnt;
  __pyx_v_self->__pyx_base.options.stdio = __pyx_t_2;

  /* "uvloop/handles/process.pyx":452
 *         self.options.stdio = self.iocnt
 * 
 *         if _stdin is not None:             # <<<<<<<<<<<<<<
 *             if _stdin == subprocess_PIPE:
 *                 r, w = self._file_inpipe()
 */
  __pyx_t_3 = (__pyx_v__stdin != Py_None);
  if (__pyx_t_3) {

    /* "uvloop/handles/process.pyx":453
 * 
 *         if _stdin is not None:
 *             if _stdin == subprocess_PIPE:             # <<<<<<<<<<<<<<
 *                 r, w = self._file_inpipe()
 *                 io[0] = r
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v__stdin, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 453, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 453, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":454
 *         if _stdin is not None:
 *             if _stdin == subprocess_PIPE:
 *                 r, w = self._file_inpipe()             # <<<<<<<<<<<<<<
 *                 io[0] = r
 * 
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_inpipe(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 454, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(4, 454, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
        index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(4, 454, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L6_unpacking_done;
        __pyx_L5_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(4, 454, __pyx_L1_error)
        __pyx_L6_unpacking_done:;
      }
      __pyx_v_r = __pyx_t_1;
      __pyx_t_1 = 0;
      __pyx_v_w = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "uvloop/handles/process.pyx":455
 *             if _stdin == subprocess_PIPE:
 *                 r, w = self._file_inpipe()
 *                 io[0] = r             # <<<<<<<<<<<<<<
 * 
 *                 self.stdin_proto = WriteSubprocessPipeProto(self, 0)
 */
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 0, __pyx_v_r, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 455, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":457
 *                 io[0] = r
 * 
 *                 self.stdin_proto = WriteSubprocessPipeProto(self, 0)             # <<<<<<<<<<<<<<
 *                 waiter = self._loop._new_future()
 *                 self._stdin = WriteUnixTransport.new(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_WriteSubprocessPipeProto); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, ((PyObject *)__pyx_v_self), __pyx_int_0};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 457, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_v_self->stdin_proto);
      __Pyx_DECREF(__pyx_v_self->stdin_proto);
      __pyx_v_self->stdin_proto = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":458
 * 
 *                 self.stdin_proto = WriteSubprocessPipeProto(self, 0)
 *                 waiter = self._loop._new_future()             # <<<<<<<<<<<<<<
 *                 self._stdin = WriteUnixTransport.new(
 *                     self._loop, self.stdin_proto, None, waiter)
 */
      __pyx_t_4 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self->__pyx_base.__pyx_base._loop); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_waiter = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":460
 *                 waiter = self._loop._new_future()
 *                 self._stdin = WriteUnixTransport.new(
 *                     self._loop, self.stdin_proto, None, waiter)             # <<<<<<<<<<<<<<
 *                 self._init_futs.append(waiter)
 *                 self._stdin._open(w)
 */
      __pyx_t_4 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = __pyx_v_self->stdin_proto;
      __Pyx_INCREF(__pyx_t_5);

      /* "uvloop/handles/process.pyx":459
 *                 self.stdin_proto = WriteSubprocessPipeProto(self, 0)
 *                 waiter = self._loop._new_future()
 *                 self._stdin = WriteUnixTransport.new(             # <<<<<<<<<<<<<<
 *                     self._loop, self.stdin_proto, None, waiter)
 *                 self._init_futs.append(waiter)
 */
      __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_18WriteUnixTransport_new(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_4), __pyx_t_5, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_v_waiter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->_stdin);
      __Pyx_DECREF((PyObject *)__pyx_v_self->_stdin);
      __pyx_v_self->_stdin = ((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":461
 *                 self._stdin = WriteUnixTransport.new(
 *                     self._loop, self.stdin_proto, None, waiter)
 *                 self._init_futs.append(waiter)             # <<<<<<<<<<<<<<
 *                 self._stdin._open(w)
 *                 self._stdin._init_protocol()
 */
      if (unlikely(__pyx_v_self->_init_futs == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(4, 461, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_self->_init_futs, __pyx_v_waiter); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 461, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":462
 *                     self._loop, self.stdin_proto, None, waiter)
 *                 self._init_futs.append(waiter)
 *                 self._stdin._open(w)             # <<<<<<<<<<<<<<
 *                 self._stdin._init_protocol()
 *             elif _stdin == subprocess_DEVNULL:
 */
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_w); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 462, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_v_self->_stdin->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self->_stdin), __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 462, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":463
 *                 self._init_futs.append(waiter)
 *                 self._stdin._open(w)
 *                 self._stdin._init_protocol()             # <<<<<<<<<<<<<<
 *             elif _stdin == subprocess_DEVNULL:
 *                 io[0] = self._file_devnull()
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport *)__pyx_v_self->_stdin->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->_stdin)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":453
 * 
 *         if _stdin is not None:
 *             if _stdin == subprocess_PIPE:             # <<<<<<<<<<<<<<
 *                 r, w = self._file_inpipe()
 *                 io[0] = r
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/process.pyx":464
 *                 self._stdin._open(w)
 *                 self._stdin._init_protocol()
 *             elif _stdin == subprocess_DEVNULL:             # <<<<<<<<<<<<<<
 *                 io[0] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_DEVNULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v__stdin, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 464, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 464, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":465
 *                 self._stdin._init_protocol()
 *             elif _stdin == subprocess_DEVNULL:
 *                 io[0] = self._file_devnull()             # <<<<<<<<<<<<<<
 *             elif _stdout == subprocess_STDOUT:
 *                 raise ValueError(
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_devnull(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 0, __pyx_t_5, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 465, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/handles/process.pyx":464
 *                 self._stdin._open(w)
 *                 self._stdin._init_protocol()
 *             elif _stdin == subprocess_DEVNULL:             # <<<<<<<<<<<<<<
 *                 io[0] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/process.pyx":466
 *             elif _stdin == subprocess_DEVNULL:
 *                 io[0] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_STDOUT); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v__stdout, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 466, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 466, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/handles/process.pyx":467
 *                 io[0] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 *             else:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__172, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 467, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(4, 467, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":466
 *             elif _stdin == subprocess_DEVNULL:
 *                 io[0] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 */
    }

    /* "uvloop/handles/process.pyx":470
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 *             else:
 *                 io[0] = self._file_redirect_stdio(_stdin)             # <<<<<<<<<<<<<<
 *         else:
 *             io[0] = self._file_redirect_stdio(0)
 */
    /*else*/ {
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v__stdin); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 470, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_redirect_stdio(__pyx_v_self, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 470, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 0, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 470, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L4:;

    /* "uvloop/handles/process.pyx":452
 *         self.options.stdio = self.iocnt
 * 
 *         if _stdin is not None:             # <<<<<<<<<<<<<<
 *             if _stdin == subprocess_PIPE:
 *                 r, w = self._file_inpipe()
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":472
 *                 io[0] = self._file_redirect_stdio(_stdin)
 *         else:
 *             io[0] = self._file_redirect_stdio(0)             # <<<<<<<<<<<<<<
 * 
 *         if _stdout is not None:
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_redirect_stdio(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 0, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 472, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":474
 *             io[0] = self._file_redirect_stdio(0)
 * 
 *         if _stdout is not None:             # <<<<<<<<<<<<<<
 *             if _stdout == subprocess_PIPE:
 *                 # We can't use UV_CREATE_PIPE here, since 'stderr' might be
 */
  __pyx_t_3 = (__pyx_v__stdout != Py_None);
  if (__pyx_t_3) {

    /* "uvloop/handles/process.pyx":475
 * 
 *         if _stdout is not None:
 *             if _stdout == subprocess_PIPE:             # <<<<<<<<<<<<<<
 *                 # We can't use UV_CREATE_PIPE here, since 'stderr' might be
 *                 # set to 'subprocess.STDOUT', and there is no way to
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v__stdout, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 475, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 475, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":482
 *                 # stderr manually.
 * 
 *                 r, w = self._file_outpipe()             # <<<<<<<<<<<<<<
 *                 io[1] = w
 * 
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_outpipe(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 482, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(4, 482, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 482, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 482, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 482, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
        index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(4, 482, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(4, 482, __pyx_L1_error)
        __pyx_L10_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_w, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":483
 * 
 *                 r, w = self._file_outpipe()
 *                 io[1] = w             # <<<<<<<<<<<<<<
 * 
 *                 self.stdout_proto = ReadSubprocessPipeProto(self, 1)
 */
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 1, __pyx_v_w, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 483, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":485
 *                 io[1] = w
 * 
 *                 self.stdout_proto = ReadSubprocessPipeProto(self, 1)             # <<<<<<<<<<<<<<
 *                 waiter = self._loop._new_future()
 *                 self._stdout = ReadUnixTransport.new(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ReadSubprocessPipeProto); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, ((PyObject *)__pyx_v_self), __pyx_int_1};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 485, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_v_self->stdout_proto);
      __Pyx_DECREF(__pyx_v_self->stdout_proto);
      __pyx_v_self->stdout_proto = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "uvloop/handles/process.pyx":486
 * 
 *                 self.stdout_proto = ReadSubprocessPipeProto(self, 1)
 *                 waiter = self._loop._new_future()             # <<<<<<<<<<<<<<
 *                 self._stdout = ReadUnixTransport.new(
 *                     self._loop, self.stdout_proto, None, waiter)
 */
      __pyx_t_5 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self->__pyx_base.__pyx_base._loop); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_waiter, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/handles/process.pyx":488
 *                 waiter = self._loop._new_future()
 *                 self._stdout = ReadUnixTransport.new(
 *                     self._loop, self.stdout_proto, None, waiter)             # <<<<<<<<<<<<<<
 *                 self._init_futs.append(waiter)
 *                 self._stdout._open(r)
 */
      __pyx_t_5 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_4 = __pyx_v_self->stdout_proto;
      __Pyx_INCREF(__pyx_t_4);

      /* "uvloop/handles/process.pyx":487
 *                 self.stdout_proto = ReadSubprocessPipeProto(self, 1)
 *                 waiter = self._loop._new_future()
 *                 self._stdout = ReadUnixTransport.new(             # <<<<<<<<<<<<<<
 *                     self._loop, self.stdout_proto, None, waiter)
 *                 self._init_futs.append(waiter)
 */
      __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_17ReadUnixTransport_new(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_5), __pyx_t_4, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_v_waiter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->_stdout);
      __Pyx_DECREF((PyObject *)__pyx_v_self->_stdout);
      __pyx_v_self->_stdout = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":489
 *                 self._stdout = ReadUnixTransport.new(
 *                     self._loop, self.stdout_proto, None, waiter)
 *                 self._init_futs.append(waiter)             # <<<<<<<<<<<<<<
 *                 self._stdout._open(r)
 *                 self._stdout._init_protocol()
 */
      if (unlikely(__pyx_v_self->_init_futs == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(4, 489, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_self->_init_futs, __pyx_v_waiter); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 489, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":490
 *                     self._loop, self.stdout_proto, None, waiter)
 *                 self._init_futs.append(waiter)
 *                 self._stdout._open(r)             # <<<<<<<<<<<<<<
 *                 self._stdout._init_protocol()
 *             elif _stdout == subprocess_DEVNULL:
 */
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_r); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 490, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self->_stdout->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self->_stdout), __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":491
 *                 self._init_futs.append(waiter)
 *                 self._stdout._open(r)
 *                 self._stdout._init_protocol()             # <<<<<<<<<<<<<<
 *             elif _stdout == subprocess_DEVNULL:
 *                 io[1] = self._file_devnull()
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self->_stdout->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->_stdout)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":475
 * 
 *         if _stdout is not None:
 *             if _stdout == subprocess_PIPE:             # <<<<<<<<<<<<<<
 *                 # We can't use UV_CREATE_PIPE here, since 'stderr' might be
 *                 # set to 'subprocess.STDOUT', and there is no way to
 */
      goto __pyx_L8;
    }

    /* "uvloop/handles/process.pyx":492
 *                 self._stdout._open(r)
 *                 self._stdout._init_protocol()
 *             elif _stdout == subprocess_DEVNULL:             # <<<<<<<<<<<<<<
 *                 io[1] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_DEVNULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v__stdout, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 492, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 492, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":493
 *                 self._stdout._init_protocol()
 *             elif _stdout == subprocess_DEVNULL:
 *                 io[1] = self._file_devnull()             # <<<<<<<<<<<<<<
 *             elif _stdout == subprocess_STDOUT:
 *                 raise ValueError(
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_devnull(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 1, __pyx_t_4, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 493, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":492
 *                 self._stdout._open(r)
 *                 self._stdout._init_protocol()
 *             elif _stdout == subprocess_DEVNULL:             # <<<<<<<<<<<<<<
 *                 io[1] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:
 */
      goto __pyx_L8;
    }

    /* "uvloop/handles/process.pyx":494
 *             elif _stdout == subprocess_DEVNULL:
 *                 io[1] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_STDOUT); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v__stdout, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 494, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 494, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/handles/process.pyx":495
 *                 io[1] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 *             else:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__172, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 495, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(4, 495, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":494
 *             elif _stdout == subprocess_DEVNULL:
 *                 io[1] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 */
    }

    /* "uvloop/handles/process.pyx":498
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 *             else:
 *                 io[1] = self._file_redirect_stdio(_stdout)             # <<<<<<<<<<<<<<
 *         else:
 *             io[1] = self._file_redirect_stdio(1)
 */
    /*else*/ {
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v__stdout); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 498, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_redirect_stdio(__pyx_v_self, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 1, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 498, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L8:;

    /* "uvloop/handles/process.pyx":474
 *             io[0] = self._file_redirect_stdio(0)
 * 
 *         if _stdout is not None:             # <<<<<<<<<<<<<<
 *             if _stdout == subprocess_PIPE:
 *                 # We can't use UV_CREATE_PIPE here, since 'stderr' might be
 */
    goto __pyx_L7;
  }

  /* "uvloop/handles/process.pyx":500
 *                 io[1] = self._file_redirect_stdio(_stdout)
 *         else:
 *             io[1] = self._file_redirect_stdio(1)             # <<<<<<<<<<<<<<
 * 
 *         if _stderr is not None:
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_redirect_stdio(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 1, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 500, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L7:;

  /* "uvloop/handles/process.pyx":502
 *             io[1] = self._file_redirect_stdio(1)
 * 
 *         if _stderr is not None:             # <<<<<<<<<<<<<<
 *             if _stderr == subprocess_PIPE:
 *                 r, w = self._file_outpipe()
 */
  __pyx_t_3 = (__pyx_v__stderr != Py_None);
  if (__pyx_t_3) {

    /* "uvloop/handles/process.pyx":503
 * 
 *         if _stderr is not None:
 *             if _stderr == subprocess_PIPE:             # <<<<<<<<<<<<<<
 *                 r, w = self._file_outpipe()
 *                 io[2] = w
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v__stderr, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 503, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 503, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":504
 *         if _stderr is not None:
 *             if _stderr == subprocess_PIPE:
 *                 r, w = self._file_outpipe()             # <<<<<<<<<<<<<<
 *                 io[2] = w
 * 
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_outpipe(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(4, 504, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
        index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(4, 504, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L14_unpacking_done;
        __pyx_L13_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(4, 504, __pyx_L1_error)
        __pyx_L14_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_w, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/handles/process.pyx":505
 *             if _stderr == subprocess_PIPE:
 *                 r, w = self._file_outpipe()
 *                 io[2] = w             # <<<<<<<<<<<<<<
 * 
 *                 self.stderr_proto = ReadSubprocessPipeProto(self, 2)
 */
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 2, __pyx_v_w, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 505, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":507
 *                 io[2] = w
 * 
 *                 self.stderr_proto = ReadSubprocessPipeProto(self, 2)             # <<<<<<<<<<<<<<
 *                 waiter = self._loop._new_future()
 *                 self._stderr = ReadUnixTransport.new(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ReadSubprocessPipeProto); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 507, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, ((PyObject *)__pyx_v_self), __pyx_int_2};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 507, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_v_self->stderr_proto);
      __Pyx_DECREF(__pyx_v_self->stderr_proto);
      __pyx_v_self->stderr_proto = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":508
 * 
 *                 self.stderr_proto = ReadSubprocessPipeProto(self, 2)
 *                 waiter = self._loop._new_future()             # <<<<<<<<<<<<<<
 *                 self._stderr = ReadUnixTransport.new(
 *                     self._loop, self.stderr_proto, None, waiter)
 */
      __pyx_t_4 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self->__pyx_base.__pyx_base._loop); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_waiter, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":510
 *                 waiter = self._loop._new_future()
 *                 self._stderr = ReadUnixTransport.new(
 *                     self._loop, self.stderr_proto, None, waiter)             # <<<<<<<<<<<<<<
 *                 self._init_futs.append(waiter)
 *                 self._stderr._open(r)
 */
      __pyx_t_4 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_5 = __pyx_v_self->stderr_proto;
      __Pyx_INCREF(__pyx_t_5);

      /* "uvloop/handles/process.pyx":509
 *                 self.stderr_proto = ReadSubprocessPipeProto(self, 2)
 *                 waiter = self._loop._new_future()
 *                 self._stderr = ReadUnixTransport.new(             # <<<<<<<<<<<<<<
 *                     self._loop, self.stderr_proto, None, waiter)
 *                 self._init_futs.append(waiter)
 */
      __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_17ReadUnixTransport_new(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_4), __pyx_t_5, ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None), __pyx_v_waiter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->_stderr);
      __Pyx_DECREF((PyObject *)__pyx_v_self->_stderr);
      __pyx_v_self->_stderr = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":511
 *                 self._stderr = ReadUnixTransport.new(
 *                     self._loop, self.stderr_proto, None, waiter)
 *                 self._init_futs.append(waiter)             # <<<<<<<<<<<<<<
 *                 self._stderr._open(r)
 *                 self._stderr._init_protocol()
 */
      if (unlikely(__pyx_v_self->_init_futs == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(4, 511, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_self->_init_futs, __pyx_v_waiter); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 511, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":512
 *                     self._loop, self.stderr_proto, None, waiter)
 *                 self._init_futs.append(waiter)
 *                 self._stderr._open(r)             # <<<<<<<<<<<<<<
 *                 self._stderr._init_protocol()
 *             elif _stderr == subprocess_STDOUT:
 */
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_r); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 512, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self->_stderr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._open(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self->_stderr), __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":513
 *                 self._init_futs.append(waiter)
 *                 self._stderr._open(r)
 *                 self._stderr._init_protocol()             # <<<<<<<<<<<<<<
 *             elif _stderr == subprocess_STDOUT:
 *                 if io[1] is None:
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport *)__pyx_v_self->_stderr->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._init_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self->_stderr)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 513, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":503
 * 
 *         if _stderr is not None:
 *             if _stderr == subprocess_PIPE:             # <<<<<<<<<<<<<<
 *                 r, w = self._file_outpipe()
 *                 io[2] = w
 */
      goto __pyx_L12;
    }

    /* "uvloop/handles/process.pyx":514
 *                 self._stderr._open(r)
 *                 self._stderr._init_protocol()
 *             elif _stderr == subprocess_STDOUT:             # <<<<<<<<<<<<<<
 *                 if io[1] is None:
 *                     # shouldn't ever happen
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_STDOUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v__stderr, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 514, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 514, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":515
 *                 self._stderr._init_protocol()
 *             elif _stderr == subprocess_STDOUT:
 *                 if io[1] is None:             # <<<<<<<<<<<<<<
 *                     # shouldn't ever happen
 *                     raise RuntimeError('cannot apply subprocess.STDOUT')
 */
      __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_io, 1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 515, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = (__pyx_t_5 == Py_None);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(__pyx_t_3)) {

        /* "uvloop/handles/process.pyx":517
 *                 if io[1] is None:
 *                     # shouldn't ever happen
 *                     raise RuntimeError('cannot apply subprocess.STDOUT')             # <<<<<<<<<<<<<<
 * 
 *                 io[2] = self._file_redirect_stdio(io[1])
 */
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__173, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 517, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __PYX_ERR(4, 517, __pyx_L1_error)

        /* "uvloop/handles/process.pyx":515
 *                 self._stderr._init_protocol()
 *             elif _stderr == subprocess_STDOUT:
 *                 if io[1] is None:             # <<<<<<<<<<<<<<
 *                     # shouldn't ever happen
 *                     raise RuntimeError('cannot apply subprocess.STDOUT')
 */
      }

      /* "uvloop/handles/process.pyx":519
 *                     raise RuntimeError('cannot apply subprocess.STDOUT')
 * 
 *                 io[2] = self._file_redirect_stdio(io[1])             # <<<<<<<<<<<<<<
 *             elif _stderr == subprocess_DEVNULL:
 *                 io[2] = self._file_devnull()
 */
      __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_io, 1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 519, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 519, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_redirect_stdio(__pyx_v_self, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 519, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 2, __pyx_t_5, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 519, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/handles/process.pyx":514
 *                 self._stderr._open(r)
 *                 self._stderr._init_protocol()
 *             elif _stderr == subprocess_STDOUT:             # <<<<<<<<<<<<<<
 *                 if io[1] is None:
 *                     # shouldn't ever happen
 */
      goto __pyx_L12;
    }

    /* "uvloop/handles/process.pyx":520
 * 
 *                 io[2] = self._file_redirect_stdio(io[1])
 *             elif _stderr == subprocess_DEVNULL:             # <<<<<<<<<<<<<<
 *                 io[2] = self._file_devnull()
 *             else:
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_DEVNULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v__stderr, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 520, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(4, 520, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":521
 *                 io[2] = self._file_redirect_stdio(io[1])
 *             elif _stderr == subprocess_DEVNULL:
 *                 io[2] = self._file_devnull()             # <<<<<<<<<<<<<<
 *             else:
 *                 io[2] = self._file_redirect_stdio(_stderr)
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_devnull(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 521, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 2, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 521, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/process.pyx":520
 * 
 *                 io[2] = self._file_redirect_stdio(io[1])
 *             elif _stderr == subprocess_DEVNULL:             # <<<<<<<<<<<<<<
 *                 io[2] = self._file_devnull()
 *             else:
 */
      goto __pyx_L12;
    }

    /* "uvloop/handles/process.pyx":523
 *                 io[2] = self._file_devnull()
 *             else:
 *                 io[2] = self._file_redirect_stdio(_stderr)             # <<<<<<<<<<<<<<
 *         else:
 *             io[2] = self._file_redirect_stdio(2)
 */
    /*else*/ {
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v__stderr); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 523, __pyx_L1_error)
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_redirect_stdio(__pyx_v_self, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 523, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 2, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 523, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L12:;

    /* "uvloop/handles/process.pyx":502
 *             io[1] = self._file_redirect_stdio(1)
 * 
 *         if _stderr is not None:             # <<<<<<<<<<<<<<
 *             if _stderr == subprocess_PIPE:
 *                 r, w = self._file_outpipe()
 */
    goto __pyx_L11;
  }

  /* "uvloop/handles/process.pyx":525
 *                 io[2] = self._file_redirect_stdio(_stderr)
 *         else:
 *             io[2] = self._file_redirect_stdio(2)             # <<<<<<<<<<<<<<
 * 
 *         assert len(io) == 3
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_file_redirect_stdio(__pyx_v_self, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((__Pyx_SetItemInt(__pyx_v_io, 2, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0))) __PYX_ERR(4, 525, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L11:;

  /* "uvloop/handles/process.pyx":527
 *             io[2] = self._file_redirect_stdio(2)
 * 
 *         assert len(io) == 3             # <<<<<<<<<<<<<<
 *         for idx in range(3):
 *             iocnt = &self.iocnt[idx]
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_11 = __Pyx_PyList_GET_SIZE(__pyx_v_io); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(4, 527, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_11 == 3);
    if (unlikely(!__pyx_t_3)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(4, 527, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(4, 527, __pyx_L1_error)
  #endif

  /* "uvloop/handles/process.pyx":528
 * 
 *         assert len(io) == 3
 *         for idx in range(3):             # <<<<<<<<<<<<<<
 *             iocnt = &self.iocnt[idx]
 *             if io[idx] is not None:
 */
  for (__pyx_t_12 = 0; __pyx_t_12 < 3; __pyx_t_12+=1) {
    __pyx_v_idx = __pyx_t_12;

    /* "uvloop/handles/process.pyx":529
 *         assert len(io) == 3
 *         for idx in range(3):
 *             iocnt = &self.iocnt[idx]             # <<<<<<<<<<<<<<
 *             if io[idx] is not None:
 *                 iocnt.flags = uv.UV_INHERIT_FD
 */
    __pyx_v_iocnt = (&(__pyx_v_self->__pyx_base.iocnt[__pyx_v_idx]));

    /* "uvloop/handles/process.pyx":530
 *         for idx in range(3):
 *             iocnt = &self.iocnt[idx]
 *             if io[idx] is not None:             # <<<<<<<<<<<<<<
 *                 iocnt.flags = uv.UV_INHERIT_FD
 *                 iocnt.data.fd = io[idx]
 */
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_io, __pyx_v_idx, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "uvloop/handles/process.pyx":531
 *             iocnt = &self.iocnt[idx]
 *             if io[idx] is not None:
 *                 iocnt.flags = uv.UV_INHERIT_FD             # <<<<<<<<<<<<<<
 *                 iocnt.data.fd = io[idx]
 *             else:
 */
      __pyx_v_iocnt->flags = UV_INHERIT_FD;

      /* "uvloop/handles/process.pyx":532
 *             if io[idx] is not None:
 *                 iocnt.flags = uv.UV_INHERIT_FD
 *                 iocnt.data.fd = io[idx]             # <<<<<<<<<<<<<<
 *             else:
 *                 iocnt.flags = uv.UV_IGNORE
 */
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_io, __pyx_v_idx, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 532, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 532, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_iocnt->data.fd = __pyx_t_10;

      /* "uvloop/handles/process.pyx":530
 *         for idx in range(3):
 *             iocnt = &self.iocnt[idx]
 *             if io[idx] is not None:             # <<<<<<<<<<<<<<
 *                 iocnt.flags = uv.UV_INHERIT_FD
 *                 iocnt.data.fd = io[idx]
 */
      goto __pyx_L18;
    }

    /* "uvloop/handles/process.pyx":534
 *                 iocnt.data.fd = io[idx]
 *             else:
 *                 iocnt.flags = uv.UV_IGNORE             # <<<<<<<<<<<<<<
 * 
 *     cdef _call_connection_made(self, waiter):
 */
    /*else*/ {
      __pyx_v_iocnt->flags = UV_IGNORE;
    }
    __pyx_L18:;
  }

  /* "uvloop/handles/process.pyx":442
 *         return r, w
 * 
 *     cdef _init_files(self, _stdin, _stdout, _stderr):             # <<<<<<<<<<<<<<
 *         cdef uv.uv_stdio_container_t *iocnt
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._init_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_io);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XDECREF(__pyx_v_w);
  __Pyx_XDECREF(__pyx_v_waiter);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":536
 *                 iocnt.flags = uv.UV_IGNORE
 * 
 *     cdef _call_connection_made(self, waiter):             # <<<<<<<<<<<<<<
 *         try:
 *             # we're always called in the right context, so just call the user's
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__call_connection_made(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_waiter) {
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_pending_calls = NULL;
  PyObject *__pyx_v_type = NULL;
  PyObject *__pyx_v_fd = NULL;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  Py_ssize_t __pyx_t_23;
  PyObject *(*__pyx_t_24)(PyObject *);
  PyObject *(*__pyx_t_25)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_connection_made", 1);

  /* "uvloop/handles/process.pyx":537
 * 
 *     cdef _call_connection_made(self, waiter):
 *         try:             # <<<<<<<<<<<<<<
 *             # we're always called in the right context, so just call the user's
 *             self._protocol.connection_made(self)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/handles/process.pyx":539
 *         try:
 *             # we're always called in the right context, so just call the user's
 *             self._protocol.connection_made(self)             # <<<<<<<<<<<<<<
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_connection_made); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 539, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, ((PyObject *)__pyx_v_self)};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 539, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":537
 * 
 *     cdef _call_connection_made(self, waiter):
 *         try:             # <<<<<<<<<<<<<<
 *             # we're always called in the right context, so just call the user's
 *             self._protocol.connection_made(self)
 */
    }

    /* "uvloop/handles/process.pyx":548
 *                 raise
 *         else:
 *             if waiter is not None and not waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 waiter.set_result(True)
 * 
 */
    /*else:*/ {
      __pyx_t_9 = (__pyx_v_waiter != Py_None);
      if (__pyx_t_9) {
      } else {
        __pyx_t_8 = __pyx_t_9;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 548, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 548, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(4, 548, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_10 = (!__pyx_t_9);
      __pyx_t_8 = __pyx_t_10;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_8) {

        /* "uvloop/handles/process.pyx":549
 *         else:
 *             if waiter is not None and not waiter.cancelled():
 *                 waiter.set_result(True)             # <<<<<<<<<<<<<<
 * 
 *         self._stdio_ready = 1
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_result); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 549, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, Py_True};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 549, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/handles/process.pyx":548
 *                 raise
 *         else:
 *             if waiter is not None and not waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 waiter.set_result(True)
 * 
 */
      }
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/handles/process.pyx":540
 *             # we're always called in the right context, so just call the user's
 *             self._protocol.connection_made(self)
 *         except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *             raise
 *         except BaseException as ex:
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._call_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(4, 540, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "uvloop/handles/process.pyx":541
 *             self._protocol.connection_made(self)
 *         except (KeyboardInterrupt, SystemExit):
 *             raise             # <<<<<<<<<<<<<<
 *         except BaseException as ex:
 *             if waiter is not None and not waiter.cancelled():
 */
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_t_4 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0; 
      __PYX_ERR(4, 541, __pyx_L5_except_error)
    }

    /* "uvloop/handles/process.pyx":542
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if waiter is not None and not waiter.cancelled():
 *                 waiter.set_exception(ex)
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._call_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(4, 542, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_ex = __pyx_t_5;
      /*try:*/ {

        /* "uvloop/handles/process.pyx":543
 *             raise
 *         except BaseException as ex:
 *             if waiter is not None and not waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 waiter.set_exception(ex)
 *             else:
 */
        __pyx_t_10 = (__pyx_v_waiter != Py_None);
        if (__pyx_t_10) {
        } else {
          __pyx_t_8 = __pyx_t_10;
          goto __pyx_L22_bool_binop_done;
        }
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 543, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
          __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 543, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(4, 543, __pyx_L19_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_9 = (!__pyx_t_10);
        __pyx_t_8 = __pyx_t_9;
        __pyx_L22_bool_binop_done:;
        if (likely(__pyx_t_8)) {

          /* "uvloop/handles/process.pyx":544
 *         except BaseException as ex:
 *             if waiter is not None and not waiter.cancelled():
 *                 waiter.set_exception(ex)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 544, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = NULL;
          __pyx_t_7 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_7 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_v_ex};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 544, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "uvloop/handles/process.pyx":543
 *             raise
 *         except BaseException as ex:
 *             if waiter is not None and not waiter.cancelled():             # <<<<<<<<<<<<<<
 *                 waiter.set_exception(ex)
 *             else:
 */
          goto __pyx_L21;
        }

        /* "uvloop/handles/process.pyx":546
 *                 waiter.set_exception(ex)
 *             else:
 *                 raise             # <<<<<<<<<<<<<<
 *         else:
 *             if waiter is not None and not waiter.cancelled():
 */
        /*else*/ {
          __Pyx_GIVEREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5);
          __Pyx_XGIVEREF(__pyx_t_4);
          __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_4);
          __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_4 = 0; 
          __PYX_ERR(4, 546, __pyx_L19_error)
        }
        __pyx_L21:;
      }

      /* "uvloop/handles/process.pyx":542
 *         except (KeyboardInterrupt, SystemExit):
 *             raise
 *         except BaseException as ex:             # <<<<<<<<<<<<<<
 *             if waiter is not None and not waiter.cancelled():
 *                 waiter.set_exception(ex)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L20;
        }
        __pyx_L19_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __pyx_t_11 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
          }
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
          goto __pyx_L5_except_error;
        }
        __pyx_L20:;
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/handles/process.pyx":537
 * 
 *     cdef _call_connection_made(self, waiter):
 *         try:             # <<<<<<<<<<<<<<
 *             # we're always called in the right context, so just call the user's
 *             self._protocol.connection_made(self)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/handles/process.pyx":551
 *                 waiter.set_result(True)
 * 
 *         self._stdio_ready = 1             # <<<<<<<<<<<<<<
 *         if self._pending_calls:
 *             pending_calls = self._pending_calls.copy()
 */
  __pyx_v_self->_stdio_ready = 1;

  /* "uvloop/handles/process.pyx":552
 * 
 *         self._stdio_ready = 1
 *         if self._pending_calls:             # <<<<<<<<<<<<<<
 *             pending_calls = self._pending_calls.copy()
 *             self._pending_calls.clear()
 */
  __pyx_t_8 = (__pyx_v_self->_pending_calls != Py_None)&&(PyList_GET_SIZE(__pyx_v_self->_pending_calls) != 0);
  if (__pyx_t_8) {

    /* "uvloop/handles/process.pyx":553
 *         self._stdio_ready = 1
 *         if self._pending_calls:
 *             pending_calls = self._pending_calls.copy()             # <<<<<<<<<<<<<<
 *             self._pending_calls.clear()
 *             for (type, fd, arg) in pending_calls:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_pending_calls, __pyx_n_s_copy); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 553, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_v_pending_calls = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "uvloop/handles/process.pyx":554
 *         if self._pending_calls:
 *             pending_calls = self._pending_calls.copy()
 *             self._pending_calls.clear()             # <<<<<<<<<<<<<<
 *             for (type, fd, arg) in pending_calls:
 *                 if type == _CALL_PIPE_CONNECTION_LOST:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_pending_calls, __pyx_n_s_clear); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 554, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/process.pyx":555
 *             pending_calls = self._pending_calls.copy()
 *             self._pending_calls.clear()
 *             for (type, fd, arg) in pending_calls:             # <<<<<<<<<<<<<<
 *                 if type == _CALL_PIPE_CONNECTION_LOST:
 *                     self._pipe_connection_lost(fd, arg)
 */
    if (likely(PyList_CheckExact(__pyx_v_pending_calls)) || PyTuple_CheckExact(__pyx_v_pending_calls)) {
      __pyx_t_4 = __pyx_v_pending_calls; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_23 = 0;
      __pyx_t_24 = NULL;
    } else {
      __pyx_t_23 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_pending_calls); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 555, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_24 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 555, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_24)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 555, __pyx_L1_error)
            #endif
            if (__pyx_t_23 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_23); __Pyx_INCREF(__pyx_t_5); __pyx_t_23++; if (unlikely((0 < 0))) __PYX_ERR(4, 555, __pyx_L1_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_23); __pyx_t_23++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 555, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(4, 555, __pyx_L1_error)
            #endif
            if (__pyx_t_23 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_23); __Pyx_INCREF(__pyx_t_5); __pyx_t_23++; if (unlikely((0 < 0))) __PYX_ERR(4, 555, __pyx_L1_error)
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_23); __pyx_t_23++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 555, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_24(__pyx_t_4);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(4, 555, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 3)) {
          if (size > 3) __Pyx_RaiseTooManyValuesError(3);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(4, 555, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_12 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_13 = PyTuple_GET_ITEM(sequence, 2); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_12 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_13 = PyList_GET_ITEM(sequence, 2); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_13);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_12 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_14 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_25 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_14);
        index = 0; __pyx_t_6 = __pyx_t_25(__pyx_t_14); if (unlikely(!__pyx_t_6)) goto __pyx_L31_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_12 = __pyx_t_25(__pyx_t_14); if (unlikely(!__pyx_t_12)) goto __pyx_L31_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_12);
        index = 2; __pyx_t_13 = __pyx_t_25(__pyx_t_14); if (unlikely(!__pyx_t_13)) goto __pyx_L31_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_13);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_25(__pyx_t_14), 3) < 0) __PYX_ERR(4, 555, __pyx_L1_error)
        __pyx_t_25 = NULL;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L32_unpacking_done;
        __pyx_L31_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_25 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(4, 555, __pyx_L1_error)
        __pyx_L32_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_type, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_fd, __pyx_t_12);
      __pyx_t_12 = 0;
      __Pyx_XDECREF_SET(__pyx_v_arg, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "uvloop/handles/process.pyx":556
 *             self._pending_calls.clear()
 *             for (type, fd, arg) in pending_calls:
 *                 if type == _CALL_PIPE_CONNECTION_LOST:             # <<<<<<<<<<<<<<
 *                     self._pipe_connection_lost(fd, arg)
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:
 */
      __pyx_t_8 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_type, __pyx_int_1, 1, 0)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(4, 556, __pyx_L1_error)
      if (__pyx_t_8) {

        /* "uvloop/handles/process.pyx":557
 *             for (type, fd, arg) in pending_calls:
 *                 if type == _CALL_PIPE_CONNECTION_LOST:
 *                     self._pipe_connection_lost(fd, arg)             # <<<<<<<<<<<<<<
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:
 *                     self._pipe_data_received(fd, arg)
 */
        __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_v_fd); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 557, __pyx_L1_error)
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_pipe_connection_lost(__pyx_v_self, __pyx_t_15, __pyx_v_arg); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 557, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/process.pyx":556
 *             self._pending_calls.clear()
 *             for (type, fd, arg) in pending_calls:
 *                 if type == _CALL_PIPE_CONNECTION_LOST:             # <<<<<<<<<<<<<<
 *                     self._pipe_connection_lost(fd, arg)
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:
 */
        goto __pyx_L33;
      }

      /* "uvloop/handles/process.pyx":558
 *                 if type == _CALL_PIPE_CONNECTION_LOST:
 *                     self._pipe_connection_lost(fd, arg)
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:             # <<<<<<<<<<<<<<
 *                     self._pipe_data_received(fd, arg)
 *                 elif type == _CALL_PROCESS_EXITED:
 */
      __pyx_t_8 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_type, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(4, 558, __pyx_L1_error)
      if (__pyx_t_8) {

        /* "uvloop/handles/process.pyx":559
 *                     self._pipe_connection_lost(fd, arg)
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:
 *                     self._pipe_data_received(fd, arg)             # <<<<<<<<<<<<<<
 *                 elif type == _CALL_PROCESS_EXITED:
 *                     self._loop.call_soon(self._protocol.process_exited)
 */
        __pyx_t_15 = __Pyx_PyInt_As_int(__pyx_v_fd); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 559, __pyx_L1_error)
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_pipe_data_received(__pyx_v_self, __pyx_t_15, __pyx_v_arg); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/process.pyx":558
 *                 if type == _CALL_PIPE_CONNECTION_LOST:
 *                     self._pipe_connection_lost(fd, arg)
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:             # <<<<<<<<<<<<<<
 *                     self._pipe_data_received(fd, arg)
 *                 elif type == _CALL_PROCESS_EXITED:
 */
        goto __pyx_L33;
      }

      /* "uvloop/handles/process.pyx":560
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:
 *                     self._pipe_data_received(fd, arg)
 *                 elif type == _CALL_PROCESS_EXITED:             # <<<<<<<<<<<<<<
 *                     self._loop.call_soon(self._protocol.process_exited)
 *                 elif type == _CALL_CONNECTION_LOST:
 */
      __pyx_t_8 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_type, __pyx_int_2, 2, 0)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(4, 560, __pyx_L1_error)
      if (__pyx_t_8) {

        /* "uvloop/handles/process.pyx":561
 *                     self._pipe_data_received(fd, arg)
 *                 elif type == _CALL_PROCESS_EXITED:
 *                     self._loop.call_soon(self._protocol.process_exited)             # <<<<<<<<<<<<<<
 *                 elif type == _CALL_CONNECTION_LOST:
 *                     self._loop.call_soon(self._protocol.connection_lost, None)
 */
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 561, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_process_exited); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 561, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_12};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 561, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/process.pyx":560
 *                 elif type == _CALL_PIPE_DATA_RECEIVED:
 *                     self._pipe_data_received(fd, arg)
 *                 elif type == _CALL_PROCESS_EXITED:             # <<<<<<<<<<<<<<
 *                     self._loop.call_soon(self._protocol.process_exited)
 *                 elif type == _CALL_CONNECTION_LOST:
 */
        goto __pyx_L33;
      }

      /* "uvloop/handles/process.pyx":562
 *                 elif type == _CALL_PROCESS_EXITED:
 *                     self._loop.call_soon(self._protocol.process_exited)
 *                 elif type == _CALL_CONNECTION_LOST:             # <<<<<<<<<<<<<<
 *                     self._loop.call_soon(self._protocol.connection_lost, None)
 * 
 */
      __pyx_t_8 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_type, __pyx_int_3, 3, 0)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(4, 562, __pyx_L1_error)
      if (__pyx_t_8) {

        /* "uvloop/handles/process.pyx":563
 *                     self._loop.call_soon(self._protocol.process_exited)
 *                 elif type == _CALL_CONNECTION_LOST:
 *                     self._loop.call_soon(self._protocol.connection_lost, None)             # <<<<<<<<<<<<<<
 * 
 *     cdef _try_finish(self):
 */
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 563, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_connection_lost); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 563, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_12, Py_None};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 563, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/handles/process.pyx":562
 *                 elif type == _CALL_PROCESS_EXITED:
 *                     self._loop.call_soon(self._protocol.process_exited)
 *                 elif type == _CALL_CONNECTION_LOST:             # <<<<<<<<<<<<<<
 *                     self._loop.call_soon(self._protocol.connection_lost, None)
 * 
 */
      }
      __pyx_L33:;

      /* "uvloop/handles/process.pyx":555
 *             pending_calls = self._pending_calls.copy()
 *             self._pending_calls.clear()
 *             for (type, fd, arg) in pending_calls:             # <<<<<<<<<<<<<<
 *                 if type == _CALL_PIPE_CONNECTION_LOST:
 *                     self._pipe_connection_lost(fd, arg)
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/process.pyx":552
 * 
 *         self._stdio_ready = 1
 *         if self._pending_calls:             # <<<<<<<<<<<<<<
 *             pending_calls = self._pending_calls.copy()
 *             self._pending_calls.clear()
 */
  }

  /* "uvloop/handles/process.pyx":536
 *                 iocnt.flags = uv.UV_IGNORE
 * 
 *     cdef _call_connection_made(self, waiter):             # <<<<<<<<<<<<<<
 *         try:
 *             # we're always called in the right context, so just call the user's
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._call_connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_pending_calls);
  __Pyx_XDECREF(__pyx_v_type);
  __Pyx_XDECREF(__pyx_v_fd);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":565
 *                     self._loop.call_soon(self._protocol.connection_lost, None)
 * 
 *     cdef _try_finish(self):             # <<<<<<<<<<<<<<
 *         if self._returncode is None or self._finished:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_18UVProcessTransport__try_finish(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_try_finish", 1);

  /* "uvloop/handles/process.pyx":566
 * 
 *     cdef _try_finish(self):
 *         if self._returncode is None or self._finished:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._returncode == Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __pyx_v_self->_finished;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":567
 *     cdef _try_finish(self):
 *         if self._returncode is None or self._finished:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if ((self.stdin_proto is None or self.stdin_proto.disconnected) and
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":566
 * 
 *     cdef _try_finish(self):
 *         if self._returncode is None or self._finished:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/process.pyx":569
 *             return
 * 
 *         if ((self.stdin_proto is None or self.stdin_proto.disconnected) and             # <<<<<<<<<<<<<<
 *                 (self.stdout_proto is None or
 *                     self.stdout_proto.disconnected) and
 */
  __pyx_t_2 = (__pyx_v_self->stdin_proto == Py_None);
  if (!__pyx_t_2) {
  } else {
    goto __pyx_L8_next_and;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->stdin_proto, __pyx_n_s_disconnected); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 569, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_L8_next_and:;

  /* "uvloop/handles/process.pyx":570
 * 
 *         if ((self.stdin_proto is None or self.stdin_proto.disconnected) and
 *                 (self.stdout_proto is None or             # <<<<<<<<<<<<<<
 *                     self.stdout_proto.disconnected) and
 *                 (self.stderr_proto is None or
 */
  __pyx_t_2 = (__pyx_v_self->stdout_proto == Py_None);
  if (!__pyx_t_2) {
  } else {
    goto __pyx_L10_next_and;
  }

  /* "uvloop/handles/process.pyx":571
 *         if ((self.stdin_proto is None or self.stdin_proto.disconnected) and
 *                 (self.stdout_proto is None or
 *                     self.stdout_proto.disconnected) and             # <<<<<<<<<<<<<<
 *                 (self.stderr_proto is None or
 *                     self.stderr_proto.disconnected)):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->stdout_proto, __pyx_n_s_disconnected); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 571, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_L10_next_and:;

  /* "uvloop/handles/process.pyx":572
 *                 (self.stdout_proto is None or
 *                     self.stdout_proto.disconnected) and
 *                 (self.stderr_proto is None or             # <<<<<<<<<<<<<<
 *                     self.stderr_proto.disconnected)):
 * 
 */
  __pyx_t_2 = (__pyx_v_self->stderr_proto == Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "uvloop/handles/process.pyx":573
 *                     self.stdout_proto.disconnected) and
 *                 (self.stderr_proto is None or
 *                     self.stderr_proto.disconnected)):             # <<<<<<<<<<<<<<
 * 
 *             self._finished = 1
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->stderr_proto, __pyx_n_s_disconnected); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(4, 573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "uvloop/handles/process.pyx":569
 *             return
 * 
 *         if ((self.stdin_proto is None or self.stdin_proto.disconnected) and             # <<<<<<<<<<<<<<
 *                 (self.stdout_proto is None or
 *                     self.stdout_proto.disconnected) and
 */
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":575
 *                     self.stderr_proto.disconnected)):
 * 
 *             self._finished = 1             # <<<<<<<<<<<<<<
 * 
 *             if self._stdio_ready:
 */
    __pyx_v_self->_finished = 1;

    /* "uvloop/handles/process.pyx":577
 *             self._finished = 1
 * 
 *             if self._stdio_ready:             # <<<<<<<<<<<<<<
 *                 # copy self.context for simplicity
 *                 self._loop.call_soon(self._protocol.connection_lost, None,
 */
    if (__pyx_v_self->_stdio_ready) {

      /* "uvloop/handles/process.pyx":579
 *             if self._stdio_ready:
 *                 # copy self.context for simplicity
 *                 self._loop.call_soon(self._protocol.connection_lost, None,             # <<<<<<<<<<<<<<
 *                                      context=self.context)
 *             else:
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_protocol, __pyx_n_s_connection_lost); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(4, 579, __pyx_L1_error);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, Py_None)) __PYX_ERR(4, 579, __pyx_L1_error);
      __pyx_t_4 = 0;

      /* "uvloop/handles/process.pyx":580
 *                 # copy self.context for simplicity
 *                 self._loop.call_soon(self._protocol.connection_lost, None,
 *                                      context=self.context)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._pending_calls.append((_CALL_CONNECTION_LOST, None, None))
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_context, __pyx_v_self->__pyx_base.__pyx_base.context) < 0) __PYX_ERR(4, 580, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":579
 *             if self._stdio_ready:
 *                 # copy self.context for simplicity
 *                 self._loop.call_soon(self._protocol.connection_lost, None,             # <<<<<<<<<<<<<<
 *                                      context=self.context)
 *             else:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/handles/process.pyx":577
 *             self._finished = 1
 * 
 *             if self._stdio_ready:             # <<<<<<<<<<<<<<
 *                 # copy self.context for simplicity
 *                 self._loop.call_soon(self._protocol.connection_lost, None,
 */
      goto __pyx_L13;
    }

    /* "uvloop/handles/process.pyx":582
 *                                      context=self.context)
 *             else:
 *                 self._pending_calls.append((_CALL_CONNECTION_LOST, None, None))             # <<<<<<<<<<<<<<
 * 
 *     def __stdio_inited(self, waiter, stdio_fut):
 */
    /*else*/ {
      if (unlikely(__pyx_v_self->_pending_calls == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(4, 582, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_self->_pending_calls, __pyx_tuple__174); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(4, 582, __pyx_L1_error)
    }
    __pyx_L13:;

    /* "uvloop/handles/process.pyx":569
 *             return
 * 
 *         if ((self.stdin_proto is None or self.stdin_proto.disconnected) and             # <<<<<<<<<<<<<<
 *                 (self.stdout_proto is None or
 *                     self.stdout_proto.disconnected) and
 */
  }

  /* "uvloop/handles/process.pyx":565
 *                     self._loop.call_soon(self._protocol.connection_lost, None)
 * 
 *     cdef _try_finish(self):             # <<<<<<<<<<<<<<
 *         if self._returncode is None or self._finished:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._try_finish", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":584
 *                 self._pending_calls.append((_CALL_CONNECTION_LOST, None, None))
 * 
 *     def __stdio_inited(self, waiter, stdio_fut):             # <<<<<<<<<<<<<<
 *         exc = stdio_fut.exception()
 *         if exc is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_3_UVProcessTransport__stdio_inited(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_2__stdio_inited, "UVProcessTransport.__stdio_inited(self, waiter, stdio_fut)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_3_UVProcessTransport__stdio_inited = {"_UVProcessTransport__stdio_inited", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_3_UVProcessTransport__stdio_inited, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_2__stdio_inited};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_3_UVProcessTransport__stdio_inited(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_waiter = 0;
  PyObject *__pyx_v_stdio_fut = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_UVProcessTransport__stdio_inited (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_waiter,&__pyx_n_s_stdio_fut,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_waiter)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 584, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_stdio_fut)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 584, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("_UVProcessTransport__stdio_inited", 1, 2, 2, 1); __PYX_ERR(4, 584, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_UVProcessTransport__stdio_inited") < 0)) __PYX_ERR(4, 584, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_waiter = values[0];
    __pyx_v_stdio_fut = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_UVProcessTransport__stdio_inited", 1, 2, 2, __pyx_nargs); __PYX_ERR(4, 584, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._UVProcessTransport__stdio_inited", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_2__stdio_inited(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self), __pyx_v_waiter, __pyx_v_stdio_fut);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_2__stdio_inited(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_waiter, PyObject *__pyx_v_stdio_fut) {
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_UVProcessTransport__stdio_inited", 1);

  /* "uvloop/handles/process.pyx":585
 * 
 *     def __stdio_inited(self, waiter, stdio_fut):
 *         exc = stdio_fut.exception()             # <<<<<<<<<<<<<<
 *         if exc is not None:
 *             if waiter is None:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_stdio_fut, __pyx_n_s_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_exc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":586
 *     def __stdio_inited(self, waiter, stdio_fut):
 *         exc = stdio_fut.exception()
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             if waiter is None:
 *                 raise exc
 */
  __pyx_t_5 = (__pyx_v_exc != Py_None);
  if (__pyx_t_5) {

    /* "uvloop/handles/process.pyx":587
 *         exc = stdio_fut.exception()
 *         if exc is not None:
 *             if waiter is None:             # <<<<<<<<<<<<<<
 *                 raise exc
 *             else:
 */
    __pyx_t_5 = (__pyx_v_waiter == Py_None);
    if (unlikely(__pyx_t_5)) {

      /* "uvloop/handles/process.pyx":588
 *         if exc is not None:
 *             if waiter is None:
 *                 raise exc             # <<<<<<<<<<<<<<
 *             else:
 *                 waiter.set_exception(exc)
 */
      __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
      __PYX_ERR(4, 588, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":587
 *         exc = stdio_fut.exception()
 *         if exc is not None:
 *             if waiter is None:             # <<<<<<<<<<<<<<
 *                 raise exc
 *             else:
 */
    }

    /* "uvloop/handles/process.pyx":590
 *                 raise exc
 *             else:
 *                 waiter.set_exception(exc)             # <<<<<<<<<<<<<<
 *         else:
 *             self._loop._call_soon_handle(
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 590, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_exc};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 590, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uvloop/handles/process.pyx":586
 *     def __stdio_inited(self, waiter, stdio_fut):
 *         exc = stdio_fut.exception()
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             if waiter is None:
 *                 raise exc
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":592
 *                 waiter.set_exception(exc)
 *         else:
 *             self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *                 new_MethodHandle1(self._loop,
 *                                   "UVProcessTransport._call_connection_made",
 */
  /*else*/ {

    /* "uvloop/handles/process.pyx":593
 *         else:
 *             self._loop._call_soon_handle(
 *                 new_MethodHandle1(self._loop,             # <<<<<<<<<<<<<<
 *                                   "UVProcessTransport._call_connection_made",
 *                                   <method1_t>self._call_connection_made,
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_self->__pyx_base.__pyx_base._loop);
    __Pyx_INCREF(__pyx_t_1);

    /* "uvloop/handles/process.pyx":597
 *                                   <method1_t>self._call_connection_made,
 *                                   None,  # means to copy the current context
 *                                   self, waiter))             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_new_MethodHandle1(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1), __pyx_kp_u_UVProcessTransport__call_connect, ((__pyx_t_6uvloop_4loop_method1_t)((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_call_connection_made), Py_None, ((PyObject *)__pyx_v_self), __pyx_v_waiter); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/process.pyx":593
 *         else:
 *             self._loop._call_soon_handle(
 *                 new_MethodHandle1(self._loop,             # <<<<<<<<<<<<<<
 *                                   "UVProcessTransport._call_connection_made",
 *                                   <method1_t>self._call_connection_made,
 */
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(4, 593, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":592
 *                 waiter.set_exception(exc)
 *         else:
 *             self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *                 new_MethodHandle1(self._loop,
 *                                   "UVProcessTransport._call_connection_made",
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self->__pyx_base.__pyx_base._loop, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":584
 *                 self._pending_calls.append((_CALL_CONNECTION_LOST, None, None))
 * 
 *     def __stdio_inited(self, waiter, stdio_fut):             # <<<<<<<<<<<<<<
 *         exc = stdio_fut.exception()
 *         if exc is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._UVProcessTransport__stdio_inited", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":600
 * 
 *     @staticmethod
 *     cdef UVProcessTransport new(Loop loop, protocol, args, env,             # <<<<<<<<<<<<<<
 *                                 cwd, start_new_session,
 *                                 _stdin, _stdout, _stderr, pass_fds,
 */

static struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_f_6uvloop_4loop_18UVProcessTransport_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_protocol, PyObject *__pyx_v_args, PyObject *__pyx_v_env, PyObject *__pyx_v_cwd, PyObject *__pyx_v_start_new_session, PyObject *__pyx_v__stdin, PyObject *__pyx_v__stdout, PyObject *__pyx_v__stderr, PyObject *__pyx_v_pass_fds, PyObject *__pyx_v_waiter, PyObject *__pyx_v_debug_flags, PyObject *__pyx_v_preexec_fn, PyObject *__pyx_v_restore_signals) {
  struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_handle = 0;
  PyObject *__pyx_v_init_fut = NULL;
  struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/process.pyx":609
 * 
 *         cdef UVProcessTransport handle
 *         handle = UVProcessTransport.__new__(UVProcessTransport)             # <<<<<<<<<<<<<<
 *         handle._protocol = protocol
 *         handle._init(loop, args, env, cwd, start_new_session,
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UVProcessTransport(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 609, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":610
 *         cdef UVProcessTransport handle
 *         handle = UVProcessTransport.__new__(UVProcessTransport)
 *         handle._protocol = protocol             # <<<<<<<<<<<<<<
 *         handle._init(loop, args, env, cwd, start_new_session,
 *                      __process_convert_fileno(_stdin),
 */
  __Pyx_INCREF(__pyx_v_protocol);
  __Pyx_GIVEREF(__pyx_v_protocol);
  __Pyx_GOTREF(__pyx_v_handle->_protocol);
  __Pyx_DECREF(__pyx_v_handle->_protocol);
  __pyx_v_handle->_protocol = __pyx_v_protocol;

  /* "uvloop/handles/process.pyx":611
 *         handle = UVProcessTransport.__new__(UVProcessTransport)
 *         handle._protocol = protocol
 *         handle._init(loop, args, env, cwd, start_new_session,             # <<<<<<<<<<<<<<
 *                      __process_convert_fileno(_stdin),
 *                      __process_convert_fileno(_stdout),
 */
  if (!(likely(PyList_CheckExact(__pyx_v_args))||((__pyx_v_args) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_args))) __PYX_ERR(4, 611, __pyx_L1_error)
  if (!(likely(PyDict_CheckExact(__pyx_v_env))||((__pyx_v_env) == Py_None) || __Pyx_RaiseUnexpectedTypeError("dict", __pyx_v_env))) __PYX_ERR(4, 611, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":612
 *         handle._protocol = protocol
 *         handle._init(loop, args, env, cwd, start_new_session,
 *                      __process_convert_fileno(_stdin),             # <<<<<<<<<<<<<<
 *                      __process_convert_fileno(_stdout),
 *                      __process_convert_fileno(_stderr),
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___process_convert_fileno(__pyx_v__stdin); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "uvloop/handles/process.pyx":613
 *         handle._init(loop, args, env, cwd, start_new_session,
 *                      __process_convert_fileno(_stdin),
 *                      __process_convert_fileno(_stdout),             # <<<<<<<<<<<<<<
 *                      __process_convert_fileno(_stderr),
 *                      pass_fds,
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop___process_convert_fileno(__pyx_v__stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/handles/process.pyx":614
 *                      __process_convert_fileno(_stdin),
 *                      __process_convert_fileno(_stdout),
 *                      __process_convert_fileno(_stderr),             # <<<<<<<<<<<<<<
 *                      pass_fds,
 *                      debug_flags,
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop___process_convert_fileno(__pyx_v__stderr); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/handles/process.pyx":611
 *         handle = UVProcessTransport.__new__(UVProcessTransport)
 *         handle._protocol = protocol
 *         handle._init(loop, args, env, cwd, start_new_session,             # <<<<<<<<<<<<<<
 *                      __process_convert_fileno(_stdin),
 *                      __process_convert_fileno(_stdout),
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._init(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_handle), __pyx_v_loop, ((PyObject*)__pyx_v_args), ((PyObject*)__pyx_v_env), __pyx_v_cwd, __pyx_v_start_new_session, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_v_pass_fds, __pyx_v_debug_flags, __pyx_v_preexec_fn, __pyx_v_restore_signals); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/handles/process.pyx":620
 *                      restore_signals)
 * 
 *         if handle._init_futs:             # <<<<<<<<<<<<<<
 *             handle._stdio_ready = 0
 *             init_fut = aio_gather(*handle._init_futs)
 */
  __pyx_t_5 = (__pyx_v_handle->_init_futs != Py_None)&&(PyList_GET_SIZE(__pyx_v_handle->_init_futs) != 0);
  if (__pyx_t_5) {

    /* "uvloop/handles/process.pyx":621
 * 
 *         if handle._init_futs:
 *             handle._stdio_ready = 0             # <<<<<<<<<<<<<<
 *             init_fut = aio_gather(*handle._init_futs)
 *             # add_done_callback will copy the current context and run the
 */
    __pyx_v_handle->_stdio_ready = 0;

    /* "uvloop/handles/process.pyx":622
 *         if handle._init_futs:
 *             handle._stdio_ready = 0
 *             init_fut = aio_gather(*handle._init_futs)             # <<<<<<<<<<<<<<
 *             # add_done_callback will copy the current context and run the
 *             # callback within the context
 */
    __pyx_t_4 = PySequence_Tuple(__pyx_v_handle->_init_futs); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_aio_gather, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_init_fut = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/handles/process.pyx":625
 *             # add_done_callback will copy the current context and run the
 *             # callback within the context
 *             init_fut.add_done_callback(             # <<<<<<<<<<<<<<
 *                 ft_partial(handle.__stdio_inited, waiter))
 *         else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_fut, __pyx_n_s_add_done_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uvloop/handles/process.pyx":626
 *             # callback within the context
 *             init_fut.add_done_callback(
 *                 ft_partial(handle.__stdio_inited, waiter))             # <<<<<<<<<<<<<<
 *         else:
 *             handle._stdio_ready = 1
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_handle), __pyx_n_s_UVProcessTransport__stdio_inite); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_ft_partial);
    __pyx_t_6 = __pyx_v_6uvloop_4loop_ft_partial; __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_1, __pyx_v_waiter};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 626, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_2};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/process.pyx":620
 *                      restore_signals)
 * 
 *         if handle._init_futs:             # <<<<<<<<<<<<<<
 *             handle._stdio_ready = 0
 *             init_fut = aio_gather(*handle._init_futs)
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/process.pyx":628
 *                 ft_partial(handle.__stdio_inited, waiter))
 *         else:
 *             handle._stdio_ready = 1             # <<<<<<<<<<<<<<
 *             loop._call_soon_handle(
 *                 new_MethodHandle1(loop,
 */
  /*else*/ {
    __pyx_v_handle->_stdio_ready = 1;

    /* "uvloop/handles/process.pyx":630
 *             handle._stdio_ready = 1
 *             loop._call_soon_handle(
 *                 new_MethodHandle1(loop,             # <<<<<<<<<<<<<<
 *                                   "UVProcessTransport._call_connection_made",
 *                                   <method1_t>handle._call_connection_made,
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_new_MethodHandle1(__pyx_v_loop, __pyx_kp_u_UVProcessTransport__call_connect, ((__pyx_t_6uvloop_4loop_method1_t)((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_handle->__pyx_base.__pyx_base.__pyx_vtab)->_call_connection_made), Py_None, ((PyObject *)__pyx_v_handle), __pyx_v_waiter); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(4, 630, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":629
 *         else:
 *             handle._stdio_ready = 1
 *             loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *                 new_MethodHandle1(loop,
 *                                   "UVProcessTransport._call_connection_made",
 */
    __pyx_t_4 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_loop, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/process.pyx":636
 *                                   handle, waiter))
 * 
 *         return handle             # <<<<<<<<<<<<<<
 * 
 *     def get_protocol(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":600
 * 
 *     @staticmethod
 *     cdef UVProcessTransport new(Loop loop, protocol, args, env,             # <<<<<<<<<<<<<<
 *                                 cwd, start_new_session,
 *                                 _stdin, _stdout, _stderr, pass_fds,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XDECREF(__pyx_v_init_fut);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":638
 *         return handle
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_5get_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_4get_protocol, "UVProcessTransport.get_protocol(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_5get_protocol = {"get_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_5get_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_4get_protocol};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_5get_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_protocol (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_protocol", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_protocol", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_4get_protocol(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_4get_protocol(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_protocol", 1);

  /* "uvloop/handles/process.pyx":639
 * 
 *     def get_protocol(self):
 *         return self._protocol             # <<<<<<<<<<<<<<
 * 
 *     def set_protocol(self, protocol):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_protocol);
  __pyx_r = __pyx_v_self->_protocol;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":638
 *         return handle
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":641
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._protocol = protocol
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_7set_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_6set_protocol, "UVProcessTransport.set_protocol(self, protocol)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_7set_protocol = {"set_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_7set_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_6set_protocol};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_7set_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_protocol (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 641, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_protocol") < 0)) __PYX_ERR(4, 641, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_protocol = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_protocol", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 641, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.set_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_6set_protocol(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self), __pyx_v_protocol);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_6set_protocol(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_protocol", 1);

  /* "uvloop/handles/process.pyx":642
 * 
 *     def set_protocol(self, protocol):
 *         self._protocol = protocol             # <<<<<<<<<<<<<<
 * 
 *     def get_pid(self):
 */
  __Pyx_INCREF(__pyx_v_protocol);
  __Pyx_GIVEREF(__pyx_v_protocol);
  __Pyx_GOTREF(__pyx_v_self->_protocol);
  __Pyx_DECREF(__pyx_v_self->_protocol);
  __pyx_v_self->_protocol = __pyx_v_protocol;

  /* "uvloop/handles/process.pyx":641
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._protocol = protocol
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":644
 *         self._protocol = protocol
 * 
 *     def get_pid(self):             # <<<<<<<<<<<<<<
 *         return self._pid
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_9get_pid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_8get_pid, "UVProcessTransport.get_pid(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_9get_pid = {"get_pid", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_9get_pid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_8get_pid};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_9get_pid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pid (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_pid", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_pid", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_8get_pid(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_8get_pid(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pid", 1);

  /* "uvloop/handles/process.pyx":645
 * 
 *     def get_pid(self):
 *         return self._pid             # <<<<<<<<<<<<<<
 * 
 *     def get_returncode(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._pid);
  __pyx_r = __pyx_v_self->__pyx_base._pid;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":644
 *         self._protocol = protocol
 * 
 *     def get_pid(self):             # <<<<<<<<<<<<<<
 *         return self._pid
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":647
 *         return self._pid
 * 
 *     def get_returncode(self):             # <<<<<<<<<<<<<<
 *         return self._returncode
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_11get_returncode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_10get_returncode, "UVProcessTransport.get_returncode(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_11get_returncode = {"get_returncode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_11get_returncode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_10get_returncode};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_11get_returncode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_returncode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_returncode", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_returncode", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_10get_returncode(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_10get_returncode(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_returncode", 1);

  /* "uvloop/handles/process.pyx":648
 * 
 *     def get_returncode(self):
 *         return self._returncode             # <<<<<<<<<<<<<<
 * 
 *     def get_pipe_transport(self, fd):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._returncode);
  __pyx_r = __pyx_v_self->__pyx_base._returncode;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":647
 *         return self._pid
 * 
 *     def get_returncode(self):             # <<<<<<<<<<<<<<
 *         return self._returncode
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":650
 *         return self._returncode
 * 
 *     def get_pipe_transport(self, fd):             # <<<<<<<<<<<<<<
 *         if fd == 0:
 *             return self._stdin
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_13get_pipe_transport(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_12get_pipe_transport, "UVProcessTransport.get_pipe_transport(self, fd)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_13get_pipe_transport = {"get_pipe_transport", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_13get_pipe_transport, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_12get_pipe_transport};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_13get_pipe_transport(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_fd = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pipe_transport (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fd,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fd)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 650, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_pipe_transport") < 0)) __PYX_ERR(4, 650, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_fd = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_pipe_transport", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 650, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.get_pipe_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_12get_pipe_transport(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self), __pyx_v_fd);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_12get_pipe_transport(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, PyObject *__pyx_v_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pipe_transport", 1);

  /* "uvloop/handles/process.pyx":651
 * 
 *     def get_pipe_transport(self, fd):
 *         if fd == 0:             # <<<<<<<<<<<<<<
 *             return self._stdin
 *         elif fd == 1:
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_fd, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(4, 651, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":652
 *     def get_pipe_transport(self, fd):
 *         if fd == 0:
 *             return self._stdin             # <<<<<<<<<<<<<<
 *         elif fd == 1:
 *             return self._stdout
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self->_stdin);
    __pyx_r = ((PyObject *)__pyx_v_self->_stdin);
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":651
 * 
 *     def get_pipe_transport(self, fd):
 *         if fd == 0:             # <<<<<<<<<<<<<<
 *             return self._stdin
 *         elif fd == 1:
 */
  }

  /* "uvloop/handles/process.pyx":653
 *         if fd == 0:
 *             return self._stdin
 *         elif fd == 1:             # <<<<<<<<<<<<<<
 *             return self._stdout
 *         elif fd == 2:
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_fd, __pyx_int_1, 1, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(4, 653, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":654
 *             return self._stdin
 *         elif fd == 1:
 *             return self._stdout             # <<<<<<<<<<<<<<
 *         elif fd == 2:
 *             return self._stderr
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self->_stdout);
    __pyx_r = ((PyObject *)__pyx_v_self->_stdout);
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":653
 *         if fd == 0:
 *             return self._stdin
 *         elif fd == 1:             # <<<<<<<<<<<<<<
 *             return self._stdout
 *         elif fd == 2:
 */
  }

  /* "uvloop/handles/process.pyx":655
 *         elif fd == 1:
 *             return self._stdout
 *         elif fd == 2:             # <<<<<<<<<<<<<<
 *             return self._stderr
 * 
 */
  __pyx_t_1 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_fd, __pyx_int_2, 2, 0)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(4, 655, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":656
 *             return self._stdout
 *         elif fd == 2:
 *             return self._stderr             # <<<<<<<<<<<<<<
 * 
 *     def terminate(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self->_stderr);
    __pyx_r = ((PyObject *)__pyx_v_self->_stderr);
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":655
 *         elif fd == 1:
 *             return self._stdout
 *         elif fd == 2:             # <<<<<<<<<<<<<<
 *             return self._stderr
 * 
 */
  }

  /* "uvloop/handles/process.pyx":650
 *         return self._returncode
 * 
 *     def get_pipe_transport(self, fd):             # <<<<<<<<<<<<<<
 *         if fd == 0:
 *             return self._stdin
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.get_pipe_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":658
 *             return self._stderr
 * 
 *     def terminate(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGTERM)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_15terminate(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_14terminate, "UVProcessTransport.terminate(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_15terminate = {"terminate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_15terminate, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_14terminate};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_15terminate(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("terminate (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("terminate", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "terminate", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_14terminate(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_14terminate(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("terminate", 1);

  /* "uvloop/handles/process.pyx":659
 * 
 *     def terminate(self):
 *         self._check_proc()             # <<<<<<<<<<<<<<
 *         self._kill(uv.SIGTERM)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_check_proc(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":660
 *     def terminate(self):
 *         self._check_proc()
 *         self._kill(uv.SIGTERM)             # <<<<<<<<<<<<<<
 * 
 *     def kill(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._kill(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), SIGTERM); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":658
 *             return self._stderr
 * 
 *     def terminate(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGTERM)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.terminate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":662
 *         self._kill(uv.SIGTERM)
 * 
 *     def kill(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGKILL)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_17kill(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_16kill, "UVProcessTransport.kill(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_17kill = {"kill", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_17kill, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_16kill};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_17kill(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("kill (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("kill", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "kill", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_16kill(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_16kill(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("kill", 1);

  /* "uvloop/handles/process.pyx":663
 * 
 *     def kill(self):
 *         self._check_proc()             # <<<<<<<<<<<<<<
 *         self._kill(uv.SIGKILL)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_check_proc(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":664
 *     def kill(self):
 *         self._check_proc()
 *         self._kill(uv.SIGKILL)             # <<<<<<<<<<<<<<
 * 
 *     def send_signal(self, int signal):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._kill(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), SIGKILL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":662
 *         self._kill(uv.SIGTERM)
 * 
 *     def kill(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGKILL)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.kill", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":666
 *         self._kill(uv.SIGKILL)
 * 
 *     def send_signal(self, int signal):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(signal)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_19send_signal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_18send_signal, "UVProcessTransport.send_signal(self, int signal)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_19send_signal = {"send_signal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_19send_signal, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_18send_signal};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_19send_signal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_signal;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("send_signal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_signal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_signal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 666, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "send_signal") < 0)) __PYX_ERR(4, 666, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_signal = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_signal == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 666, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("send_signal", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 666, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.send_signal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_18send_signal(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self), __pyx_v_signal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_18send_signal(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, int __pyx_v_signal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("send_signal", 1);

  /* "uvloop/handles/process.pyx":667
 * 
 *     def send_signal(self, int signal):
 *         self._check_proc()             # <<<<<<<<<<<<<<
 *         self._kill(signal)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_check_proc(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":668
 *     def send_signal(self, int signal):
 *         self._check_proc()
 *         self._kill(signal)             # <<<<<<<<<<<<<<
 * 
 *     def is_closing(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._kill(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), __pyx_v_signal); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":666
 *         self._kill(uv.SIGKILL)
 * 
 *     def send_signal(self, int signal):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(signal)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.send_signal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":670
 *         self._kill(signal)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_21is_closing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_20is_closing, "UVProcessTransport.is_closing(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_21is_closing = {"is_closing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_21is_closing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_20is_closing};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_21is_closing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_closing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_closing", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_closing", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_20is_closing(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_20is_closing(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_closing", 1);

  /* "uvloop/handles/process.pyx":671
 * 
 *     def is_closing(self):
 *         return self._closed             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->__pyx_base.__pyx_base._closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":670
 *         self._kill(signal)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.is_closing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":673
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._returncode is None:
 *             self._kill(uv.SIGKILL)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_23close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_22close, "UVProcessTransport.close(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_23close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_23close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_22close};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_23close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_22close(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_22close(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/handles/process.pyx":674
 * 
 *     def close(self):
 *         if self._returncode is None:             # <<<<<<<<<<<<<<
 *             self._kill(uv.SIGKILL)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base._returncode == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":675
 *     def close(self):
 *         if self._returncode is None:
 *             self._kill(uv.SIGKILL)             # <<<<<<<<<<<<<<
 * 
 *         if self._stdin is not None:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._kill(((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_v_self), SIGKILL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":674
 * 
 *     def close(self):
 *         if self._returncode is None:             # <<<<<<<<<<<<<<
 *             self._kill(uv.SIGKILL)
 * 
 */
  }

  /* "uvloop/handles/process.pyx":677
 *             self._kill(uv.SIGKILL)
 * 
 *         if self._stdin is not None:             # <<<<<<<<<<<<<<
 *             self._stdin.close()
 *         if self._stdout is not None:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->_stdin) != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":678
 * 
 *         if self._stdin is not None:
 *             self._stdin.close()             # <<<<<<<<<<<<<<
 *         if self._stdout is not None:
 *             self._stdout.close()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_stdin), __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":677
 *             self._kill(uv.SIGKILL)
 * 
 *         if self._stdin is not None:             # <<<<<<<<<<<<<<
 *             self._stdin.close()
 *         if self._stdout is not None:
 */
  }

  /* "uvloop/handles/process.pyx":679
 *         if self._stdin is not None:
 *             self._stdin.close()
 *         if self._stdout is not None:             # <<<<<<<<<<<<<<
 *             self._stdout.close()
 *         if self._stderr is not None:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->_stdout) != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":680
 *             self._stdin.close()
 *         if self._stdout is not None:
 *             self._stdout.close()             # <<<<<<<<<<<<<<
 *         if self._stderr is not None:
 *             self._stderr.close()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_stdout), __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":679
 *         if self._stdin is not None:
 *             self._stdin.close()
 *         if self._stdout is not None:             # <<<<<<<<<<<<<<
 *             self._stdout.close()
 *         if self._stderr is not None:
 */
  }

  /* "uvloop/handles/process.pyx":681
 *         if self._stdout is not None:
 *             self._stdout.close()
 *         if self._stderr is not None:             # <<<<<<<<<<<<<<
 *             self._stderr.close()
 * 
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->_stderr) != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":682
 *             self._stdout.close()
 *         if self._stderr is not None:
 *             self._stderr.close()             # <<<<<<<<<<<<<<
 * 
 *         if self._returncode is not None:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_stderr), __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 682, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 682, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":681
 *         if self._stdout is not None:
 *             self._stdout.close()
 *         if self._stderr is not None:             # <<<<<<<<<<<<<<
 *             self._stderr.close()
 * 
 */
  }

  /* "uvloop/handles/process.pyx":684
 *             self._stderr.close()
 * 
 *         if self._returncode is not None:             # <<<<<<<<<<<<<<
 *             # The process is dead, just close the UV handle.
 *             #
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base._returncode != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":691
 *             # which the transport will be GC'ed and the uvhandle will be
 *             # closed in UVHandle.__dealloc__.)
 *             self._close()             # <<<<<<<<<<<<<<
 * 
 *     def get_extra_info(self, name, default=None):
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 691, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":684
 *             self._stderr.close()
 * 
 *         if self._returncode is not None:             # <<<<<<<<<<<<<<
 *             # The process is dead, just close the UV handle.
 *             #
 */
  }

  /* "uvloop/handles/process.pyx":673
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._returncode is None:
 *             self._kill(uv.SIGKILL)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":693
 *             self._close()
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         return default
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_25get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_24get_extra_info, "UVProcessTransport.get_extra_info(self, name, default=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_25get_extra_info = {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_25get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_24get_extra_info};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_25get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_default = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_extra_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_default,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name_2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 693, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_default);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 693, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_extra_info") < 0)) __PYX_ERR(4, 693, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_extra_info", 0, 1, 2, __pyx_nargs); __PYX_ERR(4, 693, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_24get_extra_info(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self), __pyx_v_name, __pyx_v_default);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_24get_extra_info(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_name, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_extra_info", 1);

  /* "uvloop/handles/process.pyx":694
 * 
 *     def get_extra_info(self, name, default=None):
 *         return default             # <<<<<<<<<<<<<<
 * 
 *     def _wait(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_default);
  __pyx_r = __pyx_v_default;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":693
 *             self._close()
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         return default
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":696
 *         return default
 * 
 *     def _wait(self):             # <<<<<<<<<<<<<<
 *         fut = self._loop._new_future()
 *         if self._returncode is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_27_wait(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_26_wait, "UVProcessTransport._wait(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_27_wait = {"_wait", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_27_wait, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_26_wait};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_27_wait(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_wait (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_wait", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_wait", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_26_wait(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_26_wait(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_v_fut = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_wait", 1);

  /* "uvloop/handles/process.pyx":697
 * 
 *     def _wait(self):
 *         fut = self._loop._new_future()             # <<<<<<<<<<<<<<
 *         if self._returncode is not None:
 *             fut.set_result(self._returncode)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_v_self->__pyx_base.__pyx_base._loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fut = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":698
 *     def _wait(self):
 *         fut = self._loop._new_future()
 *         if self._returncode is not None:             # <<<<<<<<<<<<<<
 *             fut.set_result(self._returncode)
 *             return fut
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._returncode != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/handles/process.pyx":699
 *         fut = self._loop._new_future()
 *         if self._returncode is not None:
 *             fut.set_result(self._returncode)             # <<<<<<<<<<<<<<
 *             return fut
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_fut, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_self->__pyx_base._returncode};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/process.pyx":700
 *         if self._returncode is not None:
 *             fut.set_result(self._returncode)
 *             return fut             # <<<<<<<<<<<<<<
 * 
 *         self._exit_waiters.append(fut)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_fut);
    __pyx_r = __pyx_v_fut;
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":698
 *     def _wait(self):
 *         fut = self._loop._new_future()
 *         if self._returncode is not None:             # <<<<<<<<<<<<<<
 *             fut.set_result(self._returncode)
 *             return fut
 */
  }

  /* "uvloop/handles/process.pyx":702
 *             return fut
 * 
 *         self._exit_waiters.append(fut)             # <<<<<<<<<<<<<<
 *         return fut
 * 
 */
  if (unlikely(__pyx_v_self->_exit_waiters == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(4, 702, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_self->_exit_waiters, __pyx_v_fut); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(4, 702, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":703
 * 
 *         self._exit_waiters.append(fut)
 *         return fut             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fut);
  __pyx_r = __pyx_v_fut;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":696
 *         return default
 * 
 *     def _wait(self):             # <<<<<<<<<<<<<<
 *         fut = self._loop._new_future()
 *         if self._returncode is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport._wait", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fut);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_29__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_28__reduce_cython__, "UVProcessTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_29__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_29__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_28__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_29__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_28__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_28__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_31__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_18UVProcessTransport_30__setstate_cython__, "UVProcessTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_18UVProcessTransport_31__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_31__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_30__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_18UVProcessTransport_31__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_18UVProcessTransport_30__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_18UVProcessTransport_30__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVProcessTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVProcessTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":708
 * class WriteSubprocessPipeProto(aio_BaseProtocol):
 * 
 *     def __init__(self, proc, fd):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if type(proc) is not UVProcessTransport:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto___init__, "WriteSubprocessPipeProto.__init__(self, proc, fd)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto___init__};
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_proc = 0;
  PyObject *__pyx_v_fd = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_proc,&__pyx_n_s_fd,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 708, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proc)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 708, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(4, 708, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_fd)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 708, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(4, 708, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(4, 708, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_self = values[0];
    __pyx_v_proc = values[1];
    __pyx_v_fd = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, __pyx_nargs); __PYX_ERR(4, 708, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto___init__(__pyx_self, __pyx_v_self, __pyx_v_proc, __pyx_v_fd);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_proc, PyObject *__pyx_v_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "uvloop/handles/process.pyx":709
 * 
 *     def __init__(self, proc, fd):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if type(proc) is not UVProcessTransport:
 *                 raise TypeError
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":710
 *     def __init__(self, proc, fd):
 *         if UVLOOP_DEBUG:
 *             if type(proc) is not UVProcessTransport:             # <<<<<<<<<<<<<<
 *                 raise TypeError
 *             if not isinstance(fd, int):
 */
    __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_proc)) != ((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport));
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/process.pyx":711
 *         if UVLOOP_DEBUG:
 *             if type(proc) is not UVProcessTransport:
 *                 raise TypeError             # <<<<<<<<<<<<<<
 *             if not isinstance(fd, int):
 *                 raise TypeError
 */
      __Pyx_Raise(__pyx_builtin_TypeError, 0, 0, 0);
      __PYX_ERR(4, 711, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":710
 *     def __init__(self, proc, fd):
 *         if UVLOOP_DEBUG:
 *             if type(proc) is not UVProcessTransport:             # <<<<<<<<<<<<<<
 *                 raise TypeError
 *             if not isinstance(fd, int):
 */
    }

    /* "uvloop/handles/process.pyx":712
 *             if type(proc) is not UVProcessTransport:
 *                 raise TypeError
 *             if not isinstance(fd, int):             # <<<<<<<<<<<<<<
 *                 raise TypeError
 *         self.proc = proc
 */
    __pyx_t_1 = PyInt_Check(__pyx_v_fd); 
    __pyx_t_2 = (!__pyx_t_1);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/handles/process.pyx":713
 *                 raise TypeError
 *             if not isinstance(fd, int):
 *                 raise TypeError             # <<<<<<<<<<<<<<
 *         self.proc = proc
 *         self.fd = fd
 */
      __Pyx_Raise(__pyx_builtin_TypeError, 0, 0, 0);
      __PYX_ERR(4, 713, __pyx_L1_error)

      /* "uvloop/handles/process.pyx":712
 *             if type(proc) is not UVProcessTransport:
 *                 raise TypeError
 *             if not isinstance(fd, int):             # <<<<<<<<<<<<<<
 *                 raise TypeError
 *         self.proc = proc
 */
    }

    /* "uvloop/handles/process.pyx":709
 * 
 *     def __init__(self, proc, fd):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if type(proc) is not UVProcessTransport:
 *                 raise TypeError
 */
  }

  /* "uvloop/handles/process.pyx":714
 *             if not isinstance(fd, int):
 *                 raise TypeError
 *         self.proc = proc             # <<<<<<<<<<<<<<
 *         self.fd = fd
 *         self.pipe = None
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_proc, __pyx_v_proc) < 0) __PYX_ERR(4, 714, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":715
 *                 raise TypeError
 *         self.proc = proc
 *         self.fd = fd             # <<<<<<<<<<<<<<
 *         self.pipe = None
 *         self.disconnected = False
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_fd, __pyx_v_fd) < 0) __PYX_ERR(4, 715, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":716
 *         self.proc = proc
 *         self.fd = fd
 *         self.pipe = None             # <<<<<<<<<<<<<<
 *         self.disconnected = False
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pipe, Py_None) < 0) __PYX_ERR(4, 716, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":717
 *         self.fd = fd
 *         self.pipe = None
 *         self.disconnected = False             # <<<<<<<<<<<<<<
 * 
 *     def connection_made(self, transport):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_disconnected, Py_False) < 0) __PYX_ERR(4, 717, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":708
 * class WriteSubprocessPipeProto(aio_BaseProtocol):
 * 
 *     def __init__(self, proc, fd):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if type(proc) is not UVProcessTransport:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":719
 *         self.disconnected = False
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.pipe = transport
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_3connection_made(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_2connection_made, "WriteSubprocessPipeProto.connection_made(self, transport)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_3connection_made = {"connection_made", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_3connection_made, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_2connection_made};
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_3connection_made(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_transport = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connection_made (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_transport,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 719, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_transport)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 719, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("connection_made", 1, 2, 2, 1); __PYX_ERR(4, 719, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connection_made") < 0)) __PYX_ERR(4, 719, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_transport = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connection_made", 1, 2, 2, __pyx_nargs); __PYX_ERR(4, 719, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_2connection_made(__pyx_self, __pyx_v_self, __pyx_v_transport);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_2connection_made(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_transport) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connection_made", 1);

  /* "uvloop/handles/process.pyx":720
 * 
 *     def connection_made(self, transport):
 *         self.pipe = transport             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_pipe, __pyx_v_transport) < 0) __PYX_ERR(4, 720, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":719
 *         self.disconnected = False
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.pipe = transport
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":722
 *         self.pipe = transport
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ('<%s fd=%s pipe=%r>'
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_5__repr__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_4__repr__, "WriteSubprocessPipeProto.__repr__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_5__repr__ = {"__repr__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_5__repr__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_4__repr__};
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_5__repr__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 722, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__repr__") < 0)) __PYX_ERR(4, 722, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__repr__", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 722, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_4__repr__(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_4__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/handles/process.pyx":723
 * 
 *     def __repr__(self):
 *         return ('<%s fd=%s pipe=%r>'             # <<<<<<<<<<<<<<
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u__128);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__128);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__128);

  /* "uvloop/handles/process.pyx":724
 *     def __repr__(self):
 *         return ('<%s fd=%s pipe=%r>'
 *                 % (self.__class__.__name__, self.fd, self.pipe))             # <<<<<<<<<<<<<<
 * 
 *     def connection_lost(self, exc):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_5), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_fd_2);
  __pyx_t_2 += 4;
  __Pyx_GIVEREF(__pyx_kp_u_fd_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_fd_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_fd); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_pipe_2);
  __pyx_t_2 += 6;
  __Pyx_GIVEREF(__pyx_kp_u_pipe_2);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_pipe_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_pipe); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_5), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 724, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__130);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__130);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__130);

  /* "uvloop/handles/process.pyx":723
 * 
 *     def __repr__(self):
 *         return ('<%s fd=%s pipe=%r>'             # <<<<<<<<<<<<<<
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 * 
 */
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 7, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":722
 *         self.pipe = transport
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ('<%s fd=%s pipe=%r>'
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":726
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.disconnected = True
 *         (<UVProcessTransport>self.proc)._pipe_connection_lost(self.fd, exc)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_7connection_lost(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_6connection_lost, "WriteSubprocessPipeProto.connection_lost(self, exc)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_7connection_lost = {"connection_lost", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_7connection_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_6connection_lost};
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_7connection_lost(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_exc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connection_lost (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_exc,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 726, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 726, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("connection_lost", 1, 2, 2, 1); __PYX_ERR(4, 726, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connection_lost") < 0)) __PYX_ERR(4, 726, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_exc = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connection_lost", 1, 2, 2, __pyx_nargs); __PYX_ERR(4, 726, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_6connection_lost(__pyx_self, __pyx_v_self, __pyx_v_exc);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_6connection_lost(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connection_lost", 1);

  /* "uvloop/handles/process.pyx":727
 * 
 *     def connection_lost(self, exc):
 *         self.disconnected = True             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._pipe_connection_lost(self.fd, exc)
 *         self.proc = None
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_disconnected, Py_True) < 0) __PYX_ERR(4, 727, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":728
 *     def connection_lost(self, exc):
 *         self.disconnected = True
 *         (<UVProcessTransport>self.proc)._pipe_connection_lost(self.fd, exc)             # <<<<<<<<<<<<<<
 *         self.proc = None
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_proc); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 728, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_1)->__pyx_base.__pyx_base.__pyx_vtab)->_pipe_connection_lost(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_1), __pyx_t_3, __pyx_v_exc); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":729
 *         self.disconnected = True
 *         (<UVProcessTransport>self.proc)._pipe_connection_lost(self.fd, exc)
 *         self.proc = None             # <<<<<<<<<<<<<<
 * 
 *     def pause_writing(self):
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_proc, Py_None) < 0) __PYX_ERR(4, 729, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":726
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.disconnected = True
 *         (<UVProcessTransport>self.proc)._pipe_connection_lost(self.fd, exc)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":731
 *         self.proc = None
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_9pause_writing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_8pause_writing, "WriteSubprocessPipeProto.pause_writing(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_9pause_writing = {"pause_writing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_9pause_writing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_8pause_writing};
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_9pause_writing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pause_writing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 731, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pause_writing") < 0)) __PYX_ERR(4, 731, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pause_writing", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 731, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.pause_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_8pause_writing(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_8pause_writing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pause_writing", 1);

  /* "uvloop/handles/process.pyx":732
 * 
 *     def pause_writing(self):
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()             # <<<<<<<<<<<<<<
 * 
 *     def resume_writing(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_proc); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_2)->_protocol, __pyx_n_s_pause_writing); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 732, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":731
 *         self.proc = None
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.pause_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":734
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.resume_writing()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_11resume_writing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_10resume_writing, "WriteSubprocessPipeProto.resume_writing(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_11resume_writing = {"resume_writing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_11resume_writing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_24WriteSubprocessPipeProto_10resume_writing};
static PyObject *__pyx_pw_6uvloop_4loop_24WriteSubprocessPipeProto_11resume_writing(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resume_writing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 734, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "resume_writing") < 0)) __PYX_ERR(4, 734, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resume_writing", 1, 1, 1, __pyx_nargs); __PYX_ERR(4, 734, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.resume_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_10resume_writing(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_24WriteSubprocessPipeProto_10resume_writing(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resume_writing", 1);

  /* "uvloop/handles/process.pyx":735
 * 
 *     def resume_writing(self):
 *         (<UVProcessTransport>self.proc)._protocol.resume_writing()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_proc); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_2)->_protocol, __pyx_n_s_resume_writing); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/process.pyx":734
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.resume_writing()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.WriteSubprocessPipeProto.resume_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":741
 *                               aio_Protocol):
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._pipe_data_received(self.fd, data)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_23ReadSubprocessPipeProto_1data_received(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_23ReadSubprocessPipeProto_data_received, "ReadSubprocessPipeProto.data_received(self, data)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_23ReadSubprocessPipeProto_1data_received = {"data_received", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_23ReadSubprocessPipeProto_1data_received, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_23ReadSubprocessPipeProto_data_received};
static PyObject *__pyx_pw_6uvloop_4loop_23ReadSubprocessPipeProto_1data_received(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("data_received (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 741, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 741, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("data_received", 1, 2, 2, 1); __PYX_ERR(4, 741, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "data_received") < 0)) __PYX_ERR(4, 741, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("data_received", 1, 2, 2, __pyx_nargs); __PYX_ERR(4, 741, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.ReadSubprocessPipeProto.data_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_23ReadSubprocessPipeProto_data_received(__pyx_self, __pyx_v_self, __pyx_v_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_23ReadSubprocessPipeProto_data_received(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("data_received", 1);

  /* "uvloop/handles/process.pyx":742
 * 
 *     def data_received(self, data):
 *         (<UVProcessTransport>self.proc)._pipe_data_received(self.fd, data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_proc); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_fd); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport *)((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_1)->__pyx_base.__pyx_base.__pyx_vtab)->_pipe_data_received(((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)__pyx_t_1), __pyx_t_3, __pyx_v_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":741
 *                               aio_Protocol):
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._pipe_data_received(self.fd, data)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.ReadSubprocessPipeProto.data_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":745
 * 
 * 
 * cdef __process_convert_fileno(object obj):             # <<<<<<<<<<<<<<
 *     if obj is None or isinstance(obj, int):
 *         return obj
 */

static PyObject *__pyx_f_6uvloop_4loop___process_convert_fileno(PyObject *__pyx_v_obj) {
  PyObject *__pyx_v_fileno = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__process_convert_fileno", 1);

  /* "uvloop/handles/process.pyx":746
 * 
 * cdef __process_convert_fileno(object obj):
 *     if obj is None or isinstance(obj, int):             # <<<<<<<<<<<<<<
 *         return obj
 * 
 */
  __pyx_t_2 = (__pyx_v_obj == Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyInt_Check(__pyx_v_obj); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/handles/process.pyx":747
 * cdef __process_convert_fileno(object obj):
 *     if obj is None or isinstance(obj, int):
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     fileno = obj.fileno()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_obj);
    __pyx_r = __pyx_v_obj;
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":746
 * 
 * cdef __process_convert_fileno(object obj):
 *     if obj is None or isinstance(obj, int):             # <<<<<<<<<<<<<<
 *         return obj
 * 
 */
  }

  /* "uvloop/handles/process.pyx":749
 *         return obj
 * 
 *     fileno = obj.fileno()             # <<<<<<<<<<<<<<
 *     if not isinstance(fileno, int):
 *         raise TypeError(
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_obj, __pyx_n_s_fileno); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_v_fileno = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":750
 * 
 *     fileno = obj.fileno()
 *     if not isinstance(fileno, int):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             '{!r}.fileno() returned non-integer'.format(obj))
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_fileno); 
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/process.pyx":752
 *     if not isinstance(fileno, int):
 *         raise TypeError(
 *             '{!r}.fileno() returned non-integer'.format(obj))             # <<<<<<<<<<<<<<
 *     return fileno
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_r_fileno_returned_non_integer, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 752, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_obj};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "uvloop/handles/process.pyx":751
 *     fileno = obj.fileno()
 *     if not isinstance(fileno, int):
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             '{!r}.fileno() returned non-integer'.format(obj))
 *     return fileno
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(4, 751, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":750
 * 
 *     fileno = obj.fileno()
 *     if not isinstance(fileno, int):             # <<<<<<<<<<<<<<
 *         raise TypeError(
 *             '{!r}.fileno() returned non-integer'.format(obj))
 */
  }

  /* "uvloop/handles/process.pyx":753
 *         raise TypeError(
 *             '{!r}.fileno() returned non-integer'.format(obj))
 *     return fileno             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fileno);
  __pyx_r = __pyx_v_fileno;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":745
 * 
 * 
 * cdef __process_convert_fileno(object obj):             # <<<<<<<<<<<<<<
 *     if obj is None or isinstance(obj, int):
 *         return obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.__process_convert_fileno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fileno);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":756
 * 
 * 
 * cdef void __uvprocess_on_exit_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_process_t *handle,
 *     int64_t exit_status,
 */

static void __pyx_f_6uvloop_4loop___uvprocess_on_exit_callback(uv_process_t *__pyx_v_handle, int64_t __pyx_v_exit_status, int __pyx_v_term_signal) {
  struct __pyx_obj_6uvloop_4loop_UVProcess *__pyx_v_proc = 0;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uvprocess_on_exit_callback", 0);

  /* "uvloop/handles/process.pyx":762
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UVProcess exit callback") == 0:
 *         return
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVProcess exit callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(4, 762, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":763
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,
 *                             "UVProcess exit callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_t_1 == 0);

  /* "uvloop/handles/process.pyx":762
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UVProcess exit callback") == 0:
 *         return
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/process.pyx":764
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,
 *                             "UVProcess exit callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef UVProcess proc = <UVProcess> handle.data
 */
    goto __pyx_L0;

    /* "uvloop/handles/process.pyx":762
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UVProcess exit callback") == 0:
 *         return
 */
  }

  /* "uvloop/handles/process.pyx":766
 *         return
 * 
 *     cdef UVProcess proc = <UVProcess> handle.data             # <<<<<<<<<<<<<<
 *     try:
 *         proc._on_exit(exit_status, term_signal)
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_proc = ((struct __pyx_obj_6uvloop_4loop_UVProcess *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":767
 * 
 *     cdef UVProcess proc = <UVProcess> handle.data
 *     try:             # <<<<<<<<<<<<<<
 *         proc._on_exit(exit_status, term_signal)
 *     except BaseException as ex:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "uvloop/handles/process.pyx":768
 *     cdef UVProcess proc = <UVProcess> handle.data
 *     try:
 *         proc._on_exit(exit_status, term_signal)             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         proc._error(ex, False)
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_proc->__pyx_base.__pyx_vtab)->_on_exit(__pyx_v_proc, __pyx_v_exit_status, __pyx_v_term_signal); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 768, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/process.pyx":767
 * 
 *     cdef UVProcess proc = <UVProcess> handle.data
 *     try:             # <<<<<<<<<<<<<<
 *         proc._on_exit(exit_status, term_signal)
 *     except BaseException as ex:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/process.pyx":769
 *     try:
 *         proc._on_exit(exit_status, term_signal)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         proc._error(ex, False)
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("uvloop.loop.__uvprocess_on_exit_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(4, 769, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_ex = __pyx_t_8;
      /*try:*/ {

        /* "uvloop/handles/process.pyx":770
 *         proc._on_exit(exit_status, term_signal)
 *     except BaseException as ex:
 *         proc._error(ex, False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_proc->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_proc), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 770, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }

      /* "uvloop/handles/process.pyx":769
 *     try:
 *         proc._on_exit(exit_status, term_signal)
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         proc._error(ex, False)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_7 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L6_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/handles/process.pyx":767
 * 
 *     cdef UVProcess proc = <UVProcess> handle.data
 *     try:             # <<<<<<<<<<<<<<
 *         proc._on_exit(exit_status, term_signal)
 *     except BaseException as ex:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_L9_try_end:;
  }

  /* "uvloop/handles/process.pyx":756
 * 
 * 
 * cdef void __uvprocess_on_exit_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_process_t *handle,
 *     int64_t exit_status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("uvloop.loop.__uvprocess_on_exit_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_proc);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/process.pyx":773
 * 
 * 
 * cdef __socketpair():             # <<<<<<<<<<<<<<
 *     cdef:
 *         int fds[2]
 */

static PyObject *__pyx_f_6uvloop_4loop___socketpair(void) {
  int __pyx_v_fds[2];
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__socketpair", 1);

  /* "uvloop/handles/process.pyx":778
 *         int err
 * 
 *     err = system.socketpair(uv.AF_UNIX, uv.SOCK_STREAM, 0, fds)             # <<<<<<<<<<<<<<
 *     if err:
 *         exc = convert_error(-err)
 */
  __pyx_v_err = socketpair(AF_UNIX, SOCK_STREAM, 0, __pyx_v_fds);

  /* "uvloop/handles/process.pyx":779
 * 
 *     err = system.socketpair(uv.AF_UNIX, uv.SOCK_STREAM, 0, fds)
 *     if err:             # <<<<<<<<<<<<<<
 *         exc = convert_error(-err)
 *         raise exc
 */
  __pyx_t_1 = (__pyx_v_err != 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/process.pyx":780
 *     err = system.socketpair(uv.AF_UNIX, uv.SOCK_STREAM, 0, fds)
 *     if err:
 *         exc = convert_error(-err)             # <<<<<<<<<<<<<<
 *         raise exc
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error((-__pyx_v_err)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/process.pyx":781
 *     if err:
 *         exc = convert_error(-err)
 *         raise exc             # <<<<<<<<<<<<<<
 * 
 *     os_set_inheritable(fds[0], False)
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(4, 781, __pyx_L1_error)

    /* "uvloop/handles/process.pyx":779
 * 
 *     err = system.socketpair(uv.AF_UNIX, uv.SOCK_STREAM, 0, fds)
 *     if err:             # <<<<<<<<<<<<<<
 *         exc = convert_error(-err)
 *         raise exc
 */
  }

  /* "uvloop/handles/process.pyx":783
 *         raise exc
 * 
 *     os_set_inheritable(fds[0], False)             # <<<<<<<<<<<<<<
 *     os_set_inheritable(fds[1], False)
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_fds[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __pyx_t_4 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, Py_False};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":784
 * 
 *     os_set_inheritable(fds[0], False)
 *     os_set_inheritable(fds[1], False)             # <<<<<<<<<<<<<<
 * 
 *     return fds[0], fds[1]
 */
  __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_fds[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_os_set_inheritable; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_4, Py_False};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 784, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":786
 *     os_set_inheritable(fds[1], False)
 * 
 *     return fds[0], fds[1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int((__pyx_v_fds[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_fds[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(4, 786, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(4, 786, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/process.pyx":773
 * 
 * 
 * cdef __socketpair():             # <<<<<<<<<<<<<<
 *     cdef:
 *         int fds[2]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.__socketpair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/process.pyx":789
 * 
 * 
 * cdef void __uv_close_process_handle_cb(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* handle
 * ) noexcept with gil:
 */

static void __pyx_f_6uvloop_4loop___uv_close_process_handle_cb(uv_handle_t *__pyx_v_handle) {
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif

  /* "uvloop/handles/process.pyx":792
 *     uv.uv_handle_t* handle
 * ) noexcept with gil:
 *     PyMem_RawFree(handle)             # <<<<<<<<<<<<<<
 */
  PyMem_RawFree(__pyx_v_handle);

  /* "uvloop/handles/process.pyx":789
 * 
 * 
 * cdef void __uv_close_process_handle_cb(             # <<<<<<<<<<<<<<
 *     uv.uv_handle_t* handle
 * ) noexcept with gil:
 */

  /* function exit code */
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/fsevent.pyx":12
 * @cython.no_gc_clear
 * cdef class UVFSEvent(UVHandle):
 *     cdef _init(self, Loop loop, object callback, object context):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent__init(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback, PyObject *__pyx_v_context) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 0);
  __Pyx_INCREF(__pyx_v_context);

  /* "uvloop/handles/fsevent.pyx":15
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/fsevent.pyx":17
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(             # <<<<<<<<<<<<<<
 *             sizeof(uv.uv_fs_event_t)
 *         )
 */
  __pyx_v_self->__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_fs_event_t))));

  /* "uvloop/handles/fsevent.pyx":20
 *             sizeof(uv.uv_fs_event_t)
 *         )
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/fsevent.pyx":21
 *         )
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/fsevent.pyx":22
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_fs_event_init(
 */
    PyErr_NoMemory(); __PYX_ERR(23, 22, __pyx_L1_error)

    /* "uvloop/handles/fsevent.pyx":20
 *             sizeof(uv.uv_fs_event_t)
 *         )
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/fsevent.pyx":24
 *             raise MemoryError()
 * 
 *         err = uv.uv_fs_event_init(             # <<<<<<<<<<<<<<
 *             self._loop.uvloop, <uv.uv_fs_event_t*>self._handle
 *         )
 */
  __pyx_v_err = uv_fs_event_init(__pyx_v_self->__pyx_base._loop->uvloop, ((uv_fs_event_t *)__pyx_v_self->__pyx_base._handle));

  /* "uvloop/handles/fsevent.pyx":27
 *             self._loop.uvloop, <uv.uv_fs_event_t*>self._handle
 *         )
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/fsevent.pyx":28
 *         )
 *         if err < 0:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/fsevent.pyx":29
 *         if err < 0:
 *             self._abort_init()
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         self._finish_init()
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(23, 29, __pyx_L1_error)

    /* "uvloop/handles/fsevent.pyx":27
 *             self._loop.uvloop, <uv.uv_fs_event_t*>self._handle
 *         )
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  }

  /* "uvloop/handles/fsevent.pyx":31
 *             raise convert_error(err)
 * 
 *         self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *         self.running = 0
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/fsevent.pyx":33
 *         self._finish_init()
 * 
 *         self.running = 0             # <<<<<<<<<<<<<<
 *         self.callback = callback
 *         if context is None:
 */
  __pyx_v_self->running = 0;

  /* "uvloop/handles/fsevent.pyx":34
 * 
 *         self.running = 0
 *         self.callback = callback             # <<<<<<<<<<<<<<
 *         if context is None:
 *             context = Context_CopyCurrent()
 */
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  __Pyx_GOTREF(__pyx_v_self->callback);
  __Pyx_DECREF(__pyx_v_self->callback);
  __pyx_v_self->callback = __pyx_v_callback;

  /* "uvloop/handles/fsevent.pyx":35
 *         self.running = 0
 *         self.callback = callback
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  __pyx_t_2 = (__pyx_v_context == Py_None);
  if (__pyx_t_2) {

    /* "uvloop/handles/fsevent.pyx":36
 *         self.callback = callback
 *         if context is None:
 *             context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 *         self.context = context
 * 
 */
    __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_context, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/handles/fsevent.pyx":35
 *         self.running = 0
 *         self.callback = callback
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  }

  /* "uvloop/handles/fsevent.pyx":37
 *         if context is None:
 *             context = Context_CopyCurrent()
 *         self.context = context             # <<<<<<<<<<<<<<
 * 
 *     cdef start(self, char* path, int flags):
 */
  __Pyx_INCREF(__pyx_v_context);
  __Pyx_GIVEREF(__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.context);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.context);
  __pyx_v_self->__pyx_base.context = __pyx_v_context;

  /* "uvloop/handles/fsevent.pyx":12
 * @cython.no_gc_clear
 * cdef class UVFSEvent(UVHandle):
 *     cdef _init(self, Loop loop, object callback, object context):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_context);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/fsevent.pyx":39
 *         self.context = context
 * 
 *     cdef start(self, char* path, int flags):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent_start(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self, char *__pyx_v_path, int __pyx_v_flags) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start", 1);

  /* "uvloop/handles/fsevent.pyx":42
 *         cdef int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/fsevent.pyx":44
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             err = uv.uv_fs_event_start(
 *                 <uv.uv_fs_event_t*>self._handle,
 */
  __pyx_t_2 = (__pyx_v_self->running == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/fsevent.pyx":45
 * 
 *         if self.running == 0:
 *             err = uv.uv_fs_event_start(             # <<<<<<<<<<<<<<
 *                 <uv.uv_fs_event_t*>self._handle,
 *                 __uvfsevent_callback,
 */
    __pyx_v_err = uv_fs_event_start(((uv_fs_event_t *)__pyx_v_self->__pyx_base._handle), __pyx_f_6uvloop_4loop___uvfsevent_callback, __pyx_v_path, __pyx_v_flags);

    /* "uvloop/handles/fsevent.pyx":51
 *                 flags,
 *             )
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/fsevent.pyx":52
 *             )
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 52, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_exc = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/handles/fsevent.pyx":53
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 *             self.running = 1
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/fsevent.pyx":54
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 *             self.running = 1
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/fsevent.pyx":51
 *                 flags,
 *             )
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/fsevent.pyx":55
 *                 self._fatal_error(exc, True)
 *                 return
 *             self.running = 1             # <<<<<<<<<<<<<<
 * 
 *     cdef stop(self):
 */
    __pyx_v_self->running = 1;

    /* "uvloop/handles/fsevent.pyx":44
 *         self._ensure_alive()
 * 
 *         if self.running == 0:             # <<<<<<<<<<<<<<
 *             err = uv.uv_fs_event_start(
 *                 <uv.uv_fs_event_t*>self._handle,
 */
  }

  /* "uvloop/handles/fsevent.pyx":39
 *         self.context = context
 * 
 *     cdef start(self, char* path, int flags):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/fsevent.pyx":57
 *             self.running = 1
 * 
 *     cdef stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent_stop(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop", 1);

  /* "uvloop/handles/fsevent.pyx":60
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._is_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(23, 60, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/handles/fsevent.pyx":61
 * 
 *         if not self._is_alive():
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/fsevent.pyx":62
 *         if not self._is_alive():
 *             self.running = 0
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self.running == 1:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/fsevent.pyx":60
 *         cdef int err
 * 
 *         if not self._is_alive():             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             return
 */
  }

  /* "uvloop/handles/fsevent.pyx":64
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_fs_event_stop(<uv.uv_fs_event_t*>self._handle)
 *             self.running = 0
 */
  __pyx_t_2 = (__pyx_v_self->running == 1);
  if (__pyx_t_2) {

    /* "uvloop/handles/fsevent.pyx":65
 * 
 *         if self.running == 1:
 *             err = uv.uv_fs_event_stop(<uv.uv_fs_event_t*>self._handle)             # <<<<<<<<<<<<<<
 *             self.running = 0
 *             if err < 0:
 */
    __pyx_v_err = uv_fs_event_stop(((uv_fs_event_t *)__pyx_v_self->__pyx_base._handle));

    /* "uvloop/handles/fsevent.pyx":66
 *         if self.running == 1:
 *             err = uv.uv_fs_event_stop(<uv.uv_fs_event_t*>self._handle)
 *             self.running = 0             # <<<<<<<<<<<<<<
 *             if err < 0:
 *                 exc = convert_error(err)
 */
    __pyx_v_self->running = 0;

    /* "uvloop/handles/fsevent.pyx":67
 *             err = uv.uv_fs_event_stop(<uv.uv_fs_event_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/fsevent.pyx":68
 *             self.running = 0
 *             if err < 0:
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self._fatal_error(exc, True)
 *                 return
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 68, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "uvloop/handles/fsevent.pyx":69
 *             if err < 0:
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 69, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/fsevent.pyx":70
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 *                 return             # <<<<<<<<<<<<<<
 * 
 *     cdef _close(self):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/handles/fsevent.pyx":67
 *             err = uv.uv_fs_event_stop(<uv.uv_fs_event_t*>self._handle)
 *             self.running = 0
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 exc = convert_error(err)
 *                 self._fatal_error(exc, True)
 */
    }

    /* "uvloop/handles/fsevent.pyx":64
 *             return
 * 
 *         if self.running == 1:             # <<<<<<<<<<<<<<
 *             err = uv.uv_fs_event_stop(<uv.uv_fs_event_t*>self._handle)
 *             self.running = 0
 */
  }

  /* "uvloop/handles/fsevent.pyx":57
 *             self.running = 1
 * 
 *     cdef stop(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/fsevent.pyx":72
 *                 return
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.stop()
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVFSEvent__close(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  char const *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close", 1);

  /* "uvloop/handles/fsevent.pyx":73
 * 
 *     cdef _close(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self.stop()
 *         finally:
 */
  /*try:*/ {

    /* "uvloop/handles/fsevent.pyx":74
 *     cdef _close(self):
 *         try:
 *             self.stop()             # <<<<<<<<<<<<<<
 *         finally:
 *             UVHandle._close(<UVHandle>self)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->stop(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 74, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "uvloop/handles/fsevent.pyx":76
 *             self.stop()
 *         finally:
 *             UVHandle._close(<UVHandle>self)             # <<<<<<<<<<<<<<
 * 
 *     def cancel(self):
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVHandle__close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __pyx_t_2 = __pyx_lineno; __pyx_t_3 = __pyx_clineno; __pyx_t_4 = __pyx_filename;
      {
        __pyx_t_1 = __pyx_f_6uvloop_4loop_8UVHandle__close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 76, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __pyx_lineno = __pyx_t_2; __pyx_clineno = __pyx_t_3; __pyx_filename = __pyx_t_4;
      goto __pyx_L1_error;
      __pyx_L7_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "uvloop/handles/fsevent.pyx":72
 *                 return
 * 
 *     cdef _close(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self.stop()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent._close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/fsevent.pyx":78
 *             UVHandle._close(<UVHandle>self)
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._close()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_1cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVFSEvent_cancel, "UVFSEvent.cancel(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVFSEvent_1cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_1cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_cancel};
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_1cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("cancel", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "cancel", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVFSEvent_cancel(((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_cancel(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/handles/fsevent.pyx":79
 * 
 *     def cancel(self):
 *         self._close()             # <<<<<<<<<<<<<<
 * 
 *     def cancelled(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/fsevent.pyx":78
 *             UVHandle._close(<UVHandle>self)
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._close()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/fsevent.pyx":81
 *         self._close()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self.running == 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_3cancelled(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVFSEvent_2cancelled, "UVFSEvent.cancelled(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVFSEvent_3cancelled = {"cancelled", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_3cancelled, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_2cancelled};
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_3cancelled(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancelled (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("cancelled", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "cancelled", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVFSEvent_2cancelled(((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_2cancelled(struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancelled", 1);

  /* "uvloop/handles/fsevent.pyx":82
 * 
 *     def cancelled(self):
 *         return self.running == 0             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->running == 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/fsevent.pyx":81
 *         self._close()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self.running == 0
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.cancelled", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/fsevent.pyx":85
 * 
 *     @staticmethod
 *     cdef UVFSEvent new(Loop loop, object callback, object context):             # <<<<<<<<<<<<<<
 *         cdef UVFSEvent handle
 *         handle = UVFSEvent.__new__(UVFSEvent)
 */

static struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_f_6uvloop_4loop_9UVFSEvent_new(struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback, PyObject *__pyx_v_context) {
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_handle = 0;
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/fsevent.pyx":87
 *     cdef UVFSEvent new(Loop loop, object callback, object context):
 *         cdef UVFSEvent handle
 *         handle = UVFSEvent.__new__(UVFSEvent)             # <<<<<<<<<<<<<<
 *         handle._init(loop, callback, context)
 *         return handle
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_UVFSEvent(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_UVFSEvent), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 87, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_handle = ((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/fsevent.pyx":88
 *         cdef UVFSEvent handle
 *         handle = UVFSEvent.__new__(UVFSEvent)
 *         handle._init(loop, callback, context)             # <<<<<<<<<<<<<<
 *         return handle
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_handle->__pyx_base.__pyx_vtab)->_init(__pyx_v_handle, __pyx_v_loop, __pyx_v_callback, __pyx_v_context); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/fsevent.pyx":89
 *         handle = UVFSEvent.__new__(UVFSEvent)
 *         handle._init(loop, callback, context)
 *         return handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_handle);
  __pyx_r = __pyx_v_handle;
  goto __pyx_L0;

  /* "uvloop/handles/fsevent.pyx":85
 * 
 *     @staticmethod
 *     cdef UVFSEvent new(Loop loop, object callback, object context):             # <<<<<<<<<<<<<<
 *         cdef UVFSEvent handle
 *         handle = UVFSEvent.__new__(UVFSEvent)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_handle);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVFSEvent_4__reduce_cython__, "UVFSEvent.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVFSEvent_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_4__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVFSEvent_4__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVFSEvent_6__setstate_cython__, "UVFSEvent.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVFSEvent_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_6__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9UVFSEvent_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVFSEvent_6__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVFSEvent_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVFSEvent.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/fsevent.pyx":92
 * 
 * 
 * cdef void __uvfsevent_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_fs_event_t* handle,
 *     const char *filename,
 */

static void __pyx_f_6uvloop_4loop___uvfsevent_callback(uv_fs_event_t *__pyx_v_handle, char const *__pyx_v_filename, int __pyx_v_events, CYTHON_UNUSED int __pyx_v_status) {
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *__pyx_v_fs_event = 0;
  struct __pyx_obj_6uvloop_4loop_Handle *__pyx_v_h = 0;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  unsigned int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uvfsevent_callback", 0);

  /* "uvloop/handles/fsevent.pyx":98
 *     int status,
 * ) noexcept with gil:
 *     if __ensure_handle_data(             # <<<<<<<<<<<<<<
 *         <uv.uv_handle_t*>handle, "UVFSEvent callback"
 *     ) == 0:
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UVFSEvent callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(23, 98, __pyx_L1_error)

  /* "uvloop/handles/fsevent.pyx":100
 *     if __ensure_handle_data(
 *         <uv.uv_handle_t*>handle, "UVFSEvent callback"
 *     ) == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_t_1 == 0);

  /* "uvloop/handles/fsevent.pyx":98
 *     int status,
 * ) noexcept with gil:
 *     if __ensure_handle_data(             # <<<<<<<<<<<<<<
 *         <uv.uv_handle_t*>handle, "UVFSEvent callback"
 *     ) == 0:
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/fsevent.pyx":101
 *         <uv.uv_handle_t*>handle, "UVFSEvent callback"
 *     ) == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/fsevent.pyx":98
 *     int status,
 * ) noexcept with gil:
 *     if __ensure_handle_data(             # <<<<<<<<<<<<<<
 *         <uv.uv_handle_t*>handle, "UVFSEvent callback"
 *     ) == 0:
 */
  }

  /* "uvloop/handles/fsevent.pyx":104
 * 
 *     cdef:
 *         UVFSEvent fs_event = <UVFSEvent> handle.data             # <<<<<<<<<<<<<<
 *         Handle h
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_fs_event = ((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/fsevent.pyx":107
 *         Handle h
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         h = new_Handle(
 *             fs_event._loop,
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "uvloop/handles/fsevent.pyx":109
 *     try:
 *         h = new_Handle(
 *             fs_event._loop,             # <<<<<<<<<<<<<<
 *             fs_event.callback,
 *             (filename, FileSystemEvent(events)),
 */
      __pyx_t_3 = ((PyObject *)__pyx_v_fs_event->__pyx_base._loop);
      __Pyx_INCREF(__pyx_t_3);

      /* "uvloop/handles/fsevent.pyx":110
 *         h = new_Handle(
 *             fs_event._loop,
 *             fs_event.callback,             # <<<<<<<<<<<<<<
 *             (filename, FileSystemEvent(events)),
 *             fs_event.context,
 */
      __pyx_t_7 = __pyx_v_fs_event->callback;
      __Pyx_INCREF(__pyx_t_7);

      /* "uvloop/handles/fsevent.pyx":111
 *             fs_event._loop,
 *             fs_event.callback,
 *             (filename, FileSystemEvent(events)),             # <<<<<<<<<<<<<<
 *             fs_event.context,
 *         )
 */
      __pyx_t_8 = __Pyx_PyBytes_FromString(__pyx_v_filename); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 111, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_FileSystemEvent); if (unlikely(!__pyx_t_10)) __PYX_ERR(23, 111, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_events); if (unlikely(!__pyx_t_11)) __PYX_ERR(23, 111, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = NULL;
      __pyx_t_13 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_13 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_11};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_13, 1+__pyx_t_13);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(23, 111, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(23, 111, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8)) __PYX_ERR(23, 111, __pyx_L4_error);
      __Pyx_GIVEREF(__pyx_t_9);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9)) __PYX_ERR(23, 111, __pyx_L4_error);
      __pyx_t_8 = 0;
      __pyx_t_9 = 0;

      /* "uvloop/handles/fsevent.pyx":112
 *             fs_event.callback,
 *             (filename, FileSystemEvent(events)),
 *             fs_event.context,             # <<<<<<<<<<<<<<
 *         )
 *         h._run()
 */
      __pyx_t_9 = __pyx_v_fs_event->__pyx_base.context;
      __Pyx_INCREF(__pyx_t_9);

      /* "uvloop/handles/fsevent.pyx":108
 * 
 *     try:
 *         h = new_Handle(             # <<<<<<<<<<<<<<
 *             fs_event._loop,
 *             fs_event.callback,
 */
      __pyx_t_8 = __pyx_f_6uvloop_4loop_new_Handle(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_3), __pyx_t_7, __pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 108, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(23, 108, __pyx_L4_error)
      __pyx_v_h = ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_8);
      __pyx_t_8 = 0;

      /* "uvloop/handles/fsevent.pyx":114
 *             fs_event.context,
 *         )
 *         h._run()             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         fs_event._error(ex, False)
 */
      __pyx_t_8 = __pyx_f_6uvloop_4loop_6Handle__run(__pyx_v_h); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 114, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/handles/fsevent.pyx":107
 *         Handle h
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         h = new_Handle(
 *             fs_event._loop,
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "uvloop/handles/fsevent.pyx":115
 *         )
 *         h._run()
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         fs_event._error(ex, False)
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.__uvfsevent_callback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0) __PYX_ERR(23, 115, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      __pyx_v_ex = __pyx_t_9;
      /*try:*/ {

        /* "uvloop/handles/fsevent.pyx":116
 *         h._run()
 *     except BaseException as ex:
 *         fs_event._error(ex, False)             # <<<<<<<<<<<<<<
 */
        __pyx_t_7 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent *)__pyx_v_fs_event->__pyx_base.__pyx_vtab)->__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_fs_event), __pyx_v_ex, Py_False); if (unlikely(!__pyx_t_7)) __PYX_ERR(23, 116, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }

      /* "uvloop/handles/fsevent.pyx":115
 *         )
 *         h._run()
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         fs_event._error(ex, False)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __pyx_t_14 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
          }
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
          __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
          goto __pyx_L6_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/handles/fsevent.pyx":107
 *         Handle h
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         h = new_Handle(
 *             fs_event._loop,
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    __pyx_L9_try_end:;
  }

  /* "uvloop/handles/fsevent.pyx":92
 * 
 * 
 * cdef void __uvfsevent_callback(             # <<<<<<<<<<<<<<
 *     uv.uv_fs_event_t* handle,
 *     const char *filename,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_WriteUnraisable("uvloop.loop.__uvfsevent_callback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fs_event);
  __Pyx_XDECREF((PyObject *)__pyx_v_h);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/request.pyx":10
 *     """
 * 
 *     def __cinit__(self, Loop loop, *_):             # <<<<<<<<<<<<<<
 *         self.request = NULL
 *         self.loop = loop
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_9UVRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_9UVRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v__ = __Pyx_ArgsSlice_VARARGS(__pyx_args, 1, __pyx_nargs);
  if (unlikely(!__pyx_v__)) {
    __Pyx_RefNannyFinishContext();
    return -1;
  }
  __Pyx_GOTREF(__pyx_v__);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        default:
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 10, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        const Py_ssize_t used_pos_args = (kwd_pos_args < 1) ? kwd_pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, used_pos_args, "__cinit__") < 0)) __PYX_ERR(24, 10, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs < 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 1, __pyx_nargs); __PYX_ERR(24, 10, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v__); __pyx_v__ = 0;
  __Pyx_AddTraceback("uvloop.loop.UVRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_loop), __pyx_ptype_6uvloop_4loop_Loop, 1, "loop", 0))) __PYX_ERR(24, 10, __pyx_L1_error)
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVRequest___cinit__(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_self), __pyx_v_loop, __pyx_v__);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_DECREF(__pyx_v__);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_9UVRequest___cinit__(struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, CYTHON_UNUSED PyObject *__pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/request.pyx":11
 * 
 *     def __cinit__(self, Loop loop, *_):
 *         self.request = NULL             # <<<<<<<<<<<<<<
 *         self.loop = loop
 *         self.done = 0
 */
  __pyx_v_self->request = NULL;

  /* "uvloop/request.pyx":12
 *     def __cinit__(self, Loop loop, *_):
 *         self.request = NULL
 *         self.loop = loop             # <<<<<<<<<<<<<<
 *         self.done = 0
 *         Py_INCREF(self)
 */
  __Pyx_INCREF((PyObject *)__pyx_v_loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_loop);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->loop);
  __pyx_v_self->loop = __pyx_v_loop;

  /* "uvloop/request.pyx":13
 *         self.request = NULL
 *         self.loop = loop
 *         self.done = 0             # <<<<<<<<<<<<<<
 *         Py_INCREF(self)
 * 
 */
  __pyx_v_self->done = 0;

  /* "uvloop/request.pyx":14
 *         self.loop = loop
 *         self.done = 0
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef on_done(self):
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "uvloop/request.pyx":10
 *     """
 * 
 *     def __cinit__(self, Loop loop, *_):             # <<<<<<<<<<<<<<
 *         self.request = NULL
 *         self.loop = loop
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/request.pyx":16
 *         Py_INCREF(self)
 * 
 *     cdef on_done(self):             # <<<<<<<<<<<<<<
 *         self.done = 1
 *         Py_DECREF(self)
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVRequest_on_done(struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_done", 1);

  /* "uvloop/request.pyx":17
 * 
 *     cdef on_done(self):
 *         self.done = 1             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
 */
  __pyx_v_self->done = 1;

  /* "uvloop/request.pyx":18
 *     cdef on_done(self):
 *         self.done = 1
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef cancel(self):
 */
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "uvloop/request.pyx":16
 *         Py_INCREF(self)
 * 
 *     cdef on_done(self):             # <<<<<<<<<<<<<<
 *         self.done = 1
 *         Py_DECREF(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/request.pyx":20
 *         Py_DECREF(self)
 * 
 *     cdef cancel(self):             # <<<<<<<<<<<<<<
 *         # Most requests are implemented using a threadpool.  It's only
 *         # possible to cancel a request when it's still in a threadpool's
 */

static PyObject *__pyx_f_6uvloop_4loop_9UVRequest_cancel(struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "uvloop/request.pyx":29
 *         cdef int err
 * 
 *         if self.done == 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->done == 1);
  if (__pyx_t_1) {

    /* "uvloop/request.pyx":30
 * 
 *         if self.done == 1:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if UVLOOP_DEBUG:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/request.pyx":29
 *         cdef int err
 * 
 *         if self.done == 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/request.pyx":32
 *             return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self.request is NULL:
 *                 raise RuntimeError(
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/request.pyx":33
 * 
 *         if UVLOOP_DEBUG:
 *             if self.request is NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}.cancel: .request is NULL'.format(
 */
    __pyx_t_1 = (__pyx_v_self->request == NULL);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/request.pyx":35
 *             if self.request is NULL:
 *                 raise RuntimeError(
 *                     '{}.cancel: .request is NULL'.format(             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_cancel_request_is_NULL, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 35, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/request.pyx":36
 *                 raise RuntimeError(
 *                     '{}.cancel: .request is NULL'.format(
 *                         self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *             if self.request.data is NULL:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 36, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 36, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 35, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/request.pyx":34
 *         if UVLOOP_DEBUG:
 *             if self.request is NULL:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{}.cancel: .request is NULL'.format(
 *                         self.__class__.__name__))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 34, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(24, 34, __pyx_L1_error)

      /* "uvloop/request.pyx":33
 * 
 *         if UVLOOP_DEBUG:
 *             if self.request is NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}.cancel: .request is NULL'.format(
 */
    }

    /* "uvloop/request.pyx":38
 *                         self.__class__.__name__))
 * 
 *             if self.request.data is NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is NULL'.format(
 */
    __pyx_t_1 = (__pyx_v_self->request->data == NULL);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/request.pyx":40
 *             if self.request.data is NULL:
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is NULL'.format(             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__))
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_cancel_request_data_is_NULL, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 40, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "uvloop/request.pyx":41
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is NULL'.format(
 *                         self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *             if <UVRequest>self.request.data is not self:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 41, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 41, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 40, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }

      /* "uvloop/request.pyx":39
 * 
 *             if self.request.data is NULL:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{}.cancel: .request.data is NULL'.format(
 *                         self.__class__.__name__))
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(24, 39, __pyx_L1_error)

      /* "uvloop/request.pyx":38
 *                         self.__class__.__name__))
 * 
 *             if self.request.data is NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is NULL'.format(
 */
    }

    /* "uvloop/request.pyx":43
 *                         self.__class__.__name__))
 * 
 *             if <UVRequest>self.request.data is not self:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is not UVRequest'.format(
 */
    __pyx_t_1 = (((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_self->request->data) != __pyx_v_self);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/request.pyx":45
 *             if <UVRequest>self.request.data is not self:
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is not UVRequest'.format(             # <<<<<<<<<<<<<<
 *                         self.__class__.__name__))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_cancel_request_data_is_not_UVRe, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 45, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "uvloop/request.pyx":46
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is not UVRequest'.format(
 *                         self.__class__.__name__))             # <<<<<<<<<<<<<<
 * 
 *         # We only can cancel pending requests.  Let's try.
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(24, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(24, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(24, 45, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "uvloop/request.pyx":44
 * 
 *             if <UVRequest>self.request.data is not self:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '{}.cancel: .request.data is not UVRequest'.format(
 *                         self.__class__.__name__))
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(24, 44, __pyx_L1_error)

      /* "uvloop/request.pyx":43
 *                         self.__class__.__name__))
 * 
 *             if <UVRequest>self.request.data is not self:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     '{}.cancel: .request.data is not UVRequest'.format(
 */
    }

    /* "uvloop/request.pyx":32
 *             return
 * 
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if self.request is NULL:
 *                 raise RuntimeError(
 */
  }

  /* "uvloop/request.pyx":49
 * 
 *         # We only can cancel pending requests.  Let's try.
 *         err = uv.uv_cancel(self.request)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             if err == uv.UV_EBUSY:
 */
  __pyx_v_err = uv_cancel(__pyx_v_self->request);

  /* "uvloop/request.pyx":50
 *         # We only can cancel pending requests.  Let's try.
 *         err = uv.uv_cancel(self.request)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             if err == uv.UV_EBUSY:
 *                 # Can't close the request -- it's executing (see the first
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (__pyx_t_1) {

    /* "uvloop/request.pyx":51
 *         err = uv.uv_cancel(self.request)
 *         if err < 0:
 *             if err == uv.UV_EBUSY:             # <<<<<<<<<<<<<<
 *                 # Can't close the request -- it's executing (see the first
 *                 # comment).  Loop will have to wait until the callback
 */
    __pyx_t_1 = (__pyx_v_err == UV_EBUSY);
    if (__pyx_t_1) {
      goto __pyx_L9;
    }

    /* "uvloop/request.pyx":56
 *                 # fires.
 *                 pass
 *             elif err == uv.UV_EINVAL:             # <<<<<<<<<<<<<<
 *                 # From libuv docs:
 *                 #
 */
    __pyx_t_1 = (__pyx_v_err == UV_EINVAL);
    if (__pyx_t_1) {

      /* "uvloop/request.pyx":62
 *                 #     uv_getnameinfo_t and uv_work_t requests is currently
 *                 #     supported.
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 ex = convert_error(err)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "uvloop/request.pyx":56
 *                 # fires.
 *                 pass
 *             elif err == uv.UV_EINVAL:             # <<<<<<<<<<<<<<
 *                 # From libuv docs:
 *                 #
 */
    }

    /* "uvloop/request.pyx":64
 *                 return
 *             else:
 *                 ex = convert_error(err)             # <<<<<<<<<<<<<<
 *                 self.loop._handle_exception(ex)
 */
    /*else*/ {
      __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(24, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_ex = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "uvloop/request.pyx":65
 *             else:
 *                 ex = convert_error(err)
 *                 self.loop._handle_exception(ex)             # <<<<<<<<<<<<<<
 */
      ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_self->loop->__pyx_vtab)->_handle_exception(__pyx_v_self->loop, __pyx_v_ex); if (unlikely(PyErr_Occurred())) __PYX_ERR(24, 65, __pyx_L1_error)
    }
    __pyx_L9:;

    /* "uvloop/request.pyx":50
 *         # We only can cancel pending requests.  Let's try.
 *         err = uv.uv_cancel(self.request)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             if err == uv.UV_EBUSY:
 *                 # Can't close the request -- it's executing (see the first
 */
  }

  /* "uvloop/request.pyx":20
 *         Py_DECREF(self)
 * 
 *     cdef cancel(self):             # <<<<<<<<<<<<<<
 *         # Most requests are implemented using a threadpool.  It's only
 *         # possible to cancel a request when it's still in a threadpool's
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UVRequest.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVRequest_2__reduce_cython__, "UVRequest.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVRequest_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVRequest_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9UVRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVRequest_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_9UVRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9UVRequest_4__setstate_cython__, "UVRequest.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_9UVRequest_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVRequest_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_9UVRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UVRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_9UVRequest_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9UVRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UVRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UVRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":1
 * cdef __port_to_int(port, proto):             # <<<<<<<<<<<<<<
 *     if type(port) is int:
 *         return port
 */

static PyObject *__pyx_f_6uvloop_4loop___port_to_int(PyObject *__pyx_v_port, PyObject *__pyx_v_proto) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__port_to_int", 0);
  __Pyx_INCREF(__pyx_v_port);

  /* "uvloop/dns.pyx":2
 * cdef __port_to_int(port, proto):
 *     if type(port) is int:             # <<<<<<<<<<<<<<
 *         return port
 * 
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_port)) == ((PyObject *)(&PyInt_Type)));
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":3
 * cdef __port_to_int(port, proto):
 *     if type(port) is int:
 *         return port             # <<<<<<<<<<<<<<
 * 
 *     if port is None or port == '' or port == b'':
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_port);
    __pyx_r = __pyx_v_port;
    goto __pyx_L0;

    /* "uvloop/dns.pyx":2
 * cdef __port_to_int(port, proto):
 *     if type(port) is int:             # <<<<<<<<<<<<<<
 *         return port
 * 
 */
  }

  /* "uvloop/dns.pyx":5
 *         return port
 * 
 *     if port is None or port == '' or port == b'':             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = (__pyx_v_port == Py_None);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_port, __pyx_kp_u__10, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 5, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_v_port, __pyx_kp_b__10, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 5, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":6
 * 
 *     if port is None or port == '' or port == b'':
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;

    /* "uvloop/dns.pyx":5
 *         return port
 * 
 *     if port is None or port == '' or port == b'':             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "uvloop/dns.pyx":8
 *         return 0
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return int(port)
 *     except (ValueError, TypeError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "uvloop/dns.pyx":9
 * 
 *     try:
 *         return int(port)             # <<<<<<<<<<<<<<
 *     except (ValueError, TypeError):
 *         pass
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = __Pyx_PyNumber_Int(__pyx_v_port); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 9, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L12_try_return;

      /* "uvloop/dns.pyx":8
 *         return 0
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return int(port)
 *     except (ValueError, TypeError):
 */
    }
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/dns.pyx":10
 *     try:
 *         return int(port)
 *     except (ValueError, TypeError):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_ValueError, __pyx_builtin_TypeError);
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L9_exception_handled;
    }
    goto __pyx_L10_except_error;

    /* "uvloop/dns.pyx":8
 *         return 0
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         return int(port)
 *     except (ValueError, TypeError):
 */
    __pyx_L10_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L12_try_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L0;
    __pyx_L9_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
  }

  /* "uvloop/dns.pyx":13
 *         pass
 * 
 *     if isinstance(port, bytes):             # <<<<<<<<<<<<<<
 *         port = port.decode()
 * 
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_port); 
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":14
 * 
 *     if isinstance(port, bytes):
 *         port = port.decode()             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(port, str) and proto is not None:
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_port, __pyx_n_s_decode); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 0+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_port, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "uvloop/dns.pyx":13
 *         pass
 * 
 *     if isinstance(port, bytes):             # <<<<<<<<<<<<<<
 *         port = port.decode()
 * 
 */
  }

  /* "uvloop/dns.pyx":16
 *         port = port.decode()
 * 
 *     if isinstance(port, str) and proto is not None:             # <<<<<<<<<<<<<<
 *         if proto == uv.IPPROTO_TCP:
 *             return socket_getservbyname(port, 'tcp')
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_port); 
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_proto != Py_None);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L16_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":17
 * 
 *     if isinstance(port, str) and proto is not None:
 *         if proto == uv.IPPROTO_TCP:             # <<<<<<<<<<<<<<
 *             return socket_getservbyname(port, 'tcp')
 *         elif proto == uv.IPPROTO_UDP:
 */
    __pyx_t_6 = __Pyx_PyInt_From_int(IPPROTO_TCP); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = PyObject_RichCompare(__pyx_v_proto, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 17, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(5, 17, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_1) {

      /* "uvloop/dns.pyx":18
 *     if isinstance(port, str) and proto is not None:
 *         if proto == uv.IPPROTO_TCP:
 *             return socket_getservbyname(port, 'tcp')             # <<<<<<<<<<<<<<
 *         elif proto == uv.IPPROTO_UDP:
 *             return socket_getservbyname(port, 'udp')
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_getservbyname);
      __pyx_t_6 = __pyx_v_6uvloop_4loop_socket_getservbyname; __pyx_t_9 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_v_port, __pyx_n_u_tcp};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_10, 2+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 18, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_r = __pyx_t_8;
      __pyx_t_8 = 0;
      goto __pyx_L0;

      /* "uvloop/dns.pyx":17
 * 
 *     if isinstance(port, str) and proto is not None:
 *         if proto == uv.IPPROTO_TCP:             # <<<<<<<<<<<<<<
 *             return socket_getservbyname(port, 'tcp')
 *         elif proto == uv.IPPROTO_UDP:
 */
    }

    /* "uvloop/dns.pyx":19
 *         if proto == uv.IPPROTO_TCP:
 *             return socket_getservbyname(port, 'tcp')
 *         elif proto == uv.IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *             return socket_getservbyname(port, 'udp')
 * 
 */
    __pyx_t_8 = __Pyx_PyInt_From_int(IPPROTO_UDP); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 19, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_6 = PyObject_RichCompare(__pyx_v_proto, __pyx_t_8, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 19, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(5, 19, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_1) {

      /* "uvloop/dns.pyx":20
 *             return socket_getservbyname(port, 'tcp')
 *         elif proto == uv.IPPROTO_UDP:
 *             return socket_getservbyname(port, 'udp')             # <<<<<<<<<<<<<<
 * 
 *     raise OSError('service/proto not found')
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_getservbyname);
      __pyx_t_8 = __pyx_v_6uvloop_4loop_socket_getservbyname; __pyx_t_9 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_9, __pyx_v_port, __pyx_n_u_udp};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 2+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 20, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L0;

      /* "uvloop/dns.pyx":19
 *         if proto == uv.IPPROTO_TCP:
 *             return socket_getservbyname(port, 'tcp')
 *         elif proto == uv.IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *             return socket_getservbyname(port, 'udp')
 * 
 */
    }

    /* "uvloop/dns.pyx":16
 *         port = port.decode()
 * 
 *     if isinstance(port, str) and proto is not None:             # <<<<<<<<<<<<<<
 *         if proto == uv.IPPROTO_TCP:
 *             return socket_getservbyname(port, 'tcp')
 */
  }

  /* "uvloop/dns.pyx":22
 *             return socket_getservbyname(port, 'udp')
 * 
 *     raise OSError('service/proto not found')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__175, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_Raise(__pyx_t_6, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __PYX_ERR(5, 22, __pyx_L1_error)

  /* "uvloop/dns.pyx":1
 * cdef __port_to_int(port, proto):             # <<<<<<<<<<<<<<
 *     if type(port) is int:
 *         return port
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.__port_to_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_port);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":25
 * 
 * 
 * cdef __convert_sockaddr_to_pyaddr(const system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *     # Converts sockaddr structs into what Python socket
 *     # module can understand:
 */

static PyObject *__pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(struct sockaddr const *__pyx_v_addr) {
  char __pyx_v_buf[0x80];
  int __pyx_v_err;
  struct sockaddr_in *__pyx_v_addr4;
  struct sockaddr_in6 *__pyx_v_addr6;
  struct sockaddr_un *__pyx_v_addr_un;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__convert_sockaddr_to_pyaddr", 1);

  /* "uvloop/dns.pyx":38
 *         system.sockaddr_un *addr_un
 * 
 *     if addr.sa_family == uv.AF_INET:             # <<<<<<<<<<<<<<
 *         addr4 = <system.sockaddr_in*>addr
 * 
 */
  __pyx_t_1 = (__pyx_v_addr->sa_family == AF_INET);
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":39
 * 
 *     if addr.sa_family == uv.AF_INET:
 *         addr4 = <system.sockaddr_in*>addr             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_ip4_name(addr4, buf, sizeof(buf))
 */
    __pyx_v_addr4 = ((struct sockaddr_in *)__pyx_v_addr);

    /* "uvloop/dns.pyx":41
 *         addr4 = <system.sockaddr_in*>addr
 * 
 *         err = uv.uv_ip4_name(addr4, buf, sizeof(buf))             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
    __pyx_v_err = uv_ip4_name(__pyx_v_addr4, __pyx_v_buf, (sizeof(__pyx_v_buf)));

    /* "uvloop/dns.pyx":42
 * 
 *         err = uv.uv_ip4_name(addr4, buf, sizeof(buf))
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = (__pyx_v_err < 0);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/dns.pyx":43
 *         err = uv.uv_ip4_name(addr4, buf, sizeof(buf))
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         return (
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 43, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(5, 43, __pyx_L1_error)

      /* "uvloop/dns.pyx":42
 * 
 *         err = uv.uv_ip4_name(addr4, buf, sizeof(buf))
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    }

    /* "uvloop/dns.pyx":45
 *             raise convert_error(err)
 * 
 *         return (             # <<<<<<<<<<<<<<
 *             PyUnicode_FromString(buf),
 *             system.ntohs(addr4.sin_port)
 */
    __Pyx_XDECREF(__pyx_r);

    /* "uvloop/dns.pyx":46
 * 
 *         return (
 *             PyUnicode_FromString(buf),             # <<<<<<<<<<<<<<
 *             system.ntohs(addr4.sin_port)
 *         )
 */
    __pyx_t_2 = PyUnicode_FromString(__pyx_v_buf); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "uvloop/dns.pyx":47
 *         return (
 *             PyUnicode_FromString(buf),
 *             system.ntohs(addr4.sin_port)             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(ntohs(__pyx_v_addr4->sin_port)); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/dns.pyx":46
 * 
 *         return (
 *             PyUnicode_FromString(buf),             # <<<<<<<<<<<<<<
 *             system.ntohs(addr4.sin_port)
 *         )
 */
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(5, 46, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(5, 46, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "uvloop/dns.pyx":38
 *         system.sockaddr_un *addr_un
 * 
 *     if addr.sa_family == uv.AF_INET:             # <<<<<<<<<<<<<<
 *         addr4 = <system.sockaddr_in*>addr
 * 
 */
  }

  /* "uvloop/dns.pyx":50
 *         )
 * 
 *     elif addr.sa_family == uv.AF_INET6:             # <<<<<<<<<<<<<<
 *         addr6 = <system.sockaddr_in6*>addr
 * 
 */
  __pyx_t_1 = (__pyx_v_addr->sa_family == AF_INET6);
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":51
 * 
 *     elif addr.sa_family == uv.AF_INET6:
 *         addr6 = <system.sockaddr_in6*>addr             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_ip6_name(addr6, buf, sizeof(buf))
 */
    __pyx_v_addr6 = ((struct sockaddr_in6 *)__pyx_v_addr);

    /* "uvloop/dns.pyx":53
 *         addr6 = <system.sockaddr_in6*>addr
 * 
 *         err = uv.uv_ip6_name(addr6, buf, sizeof(buf))             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
    __pyx_v_err = uv_ip6_name(__pyx_v_addr6, __pyx_v_buf, (sizeof(__pyx_v_buf)));

    /* "uvloop/dns.pyx":54
 * 
 *         err = uv.uv_ip6_name(addr6, buf, sizeof(buf))
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = (__pyx_v_err < 0);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/dns.pyx":55
 *         err = uv.uv_ip6_name(addr6, buf, sizeof(buf))
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         return (
 */
      __pyx_t_4 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(5, 55, __pyx_L1_error)

      /* "uvloop/dns.pyx":54
 * 
 *         err = uv.uv_ip6_name(addr6, buf, sizeof(buf))
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    }

    /* "uvloop/dns.pyx":57
 *             raise convert_error(err)
 * 
 *         return (             # <<<<<<<<<<<<<<
 *             PyUnicode_FromString(buf),
 *             system.ntohs(addr6.sin6_port),
 */
    __Pyx_XDECREF(__pyx_r);

    /* "uvloop/dns.pyx":58
 * 
 *         return (
 *             PyUnicode_FromString(buf),             # <<<<<<<<<<<<<<
 *             system.ntohs(addr6.sin6_port),
 *             system.ntohl(addr6.sin6_flowinfo),
 */
    __pyx_t_4 = PyUnicode_FromString(__pyx_v_buf); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uvloop/dns.pyx":59
 *         return (
 *             PyUnicode_FromString(buf),
 *             system.ntohs(addr6.sin6_port),             # <<<<<<<<<<<<<<
 *             system.ntohl(addr6.sin6_flowinfo),
 *             addr6.sin6_scope_id
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(ntohs(__pyx_v_addr6->sin6_port)); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/dns.pyx":60
 *             PyUnicode_FromString(buf),
 *             system.ntohs(addr6.sin6_port),
 *             system.ntohl(addr6.sin6_flowinfo),             # <<<<<<<<<<<<<<
 *             addr6.sin6_scope_id
 *         )
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(ntohl(__pyx_v_addr6->sin6_flowinfo)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "uvloop/dns.pyx":61
 *             system.ntohs(addr6.sin6_port),
 *             system.ntohl(addr6.sin6_flowinfo),
 *             addr6.sin6_scope_id             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_From_unsigned_long(__pyx_v_addr6->sin6_scope_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "uvloop/dns.pyx":58
 * 
 *         return (
 *             PyUnicode_FromString(buf),             # <<<<<<<<<<<<<<
 *             system.ntohs(addr6.sin6_port),
 *             system.ntohl(addr6.sin6_flowinfo),
 */
    __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4)) __PYX_ERR(5, 58, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3)) __PYX_ERR(5, 58, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2)) __PYX_ERR(5, 58, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_5)) __PYX_ERR(5, 58, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_3 = 0;
    __pyx_t_2 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "uvloop/dns.pyx":50
 *         )
 * 
 *     elif addr.sa_family == uv.AF_INET6:             # <<<<<<<<<<<<<<
 *         addr6 = <system.sockaddr_in6*>addr
 * 
 */
  }

  /* "uvloop/dns.pyx":64
 *         )
 * 
 *     elif addr.sa_family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *         addr_un = <system.sockaddr_un*>addr
 *         return system.MakeUnixSockPyAddr(addr_un)
 */
  __pyx_t_1 = (__pyx_v_addr->sa_family == AF_UNIX);
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":65
 * 
 *     elif addr.sa_family == uv.AF_UNIX:
 *         addr_un = <system.sockaddr_un*>addr             # <<<<<<<<<<<<<<
 *         return system.MakeUnixSockPyAddr(addr_un)
 * 
 */
    __pyx_v_addr_un = ((struct sockaddr_un *)__pyx_v_addr);

    /* "uvloop/dns.pyx":66
 *     elif addr.sa_family == uv.AF_UNIX:
 *         addr_un = <system.sockaddr_un*>addr
 *         return system.MakeUnixSockPyAddr(addr_un)             # <<<<<<<<<<<<<<
 * 
 *     raise RuntimeError("cannot convert sockaddr into Python object")
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = MakeUnixSockPyAddr(__pyx_v_addr_un); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "uvloop/dns.pyx":64
 *         )
 * 
 *     elif addr.sa_family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *         addr_un = <system.sockaddr_un*>addr
 *         return system.MakeUnixSockPyAddr(addr_un)
 */
  }

  /* "uvloop/dns.pyx":68
 *         return system.MakeUnixSockPyAddr(addr_un)
 * 
 *     raise RuntimeError("cannot convert sockaddr into Python object")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__176, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_Raise(__pyx_t_6, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __PYX_ERR(5, 68, __pyx_L1_error)

  /* "uvloop/dns.pyx":25
 * 
 * 
 * cdef __convert_sockaddr_to_pyaddr(const system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *     # Converts sockaddr structs into what Python socket
 *     # module can understand:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.__convert_sockaddr_to_pyaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_14SockAddrHolder_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_14SockAddrHolder___reduce_cython__, "SockAddrHolder.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_14SockAddrHolder_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14SockAddrHolder_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14SockAddrHolder___reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_14SockAddrHolder_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_14SockAddrHolder___reduce_cython__(((struct __pyx_obj_6uvloop_4loop_SockAddrHolder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_14SockAddrHolder___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SockAddrHolder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.SockAddrHolder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_14SockAddrHolder_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_14SockAddrHolder_2__setstate_cython__, "SockAddrHolder.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_14SockAddrHolder_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14SockAddrHolder_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14SockAddrHolder_2__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_14SockAddrHolder_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SockAddrHolder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_14SockAddrHolder_2__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_SockAddrHolder *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_14SockAddrHolder_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SockAddrHolder *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.SockAddrHolder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":82
 * 
 * 
 * cdef __convert_pyaddr_to_sockaddr(int family, object addr,             # <<<<<<<<<<<<<<
 *                                   system.sockaddr* res):
 *     cdef:
 */

static PyObject *__pyx_f_6uvloop_4loop___convert_pyaddr_to_sockaddr(int __pyx_v_family, PyObject *__pyx_v_addr, struct sockaddr *__pyx_v_res) {
  int __pyx_v_err;
  int __pyx_v_addr_len;
  int __pyx_v_scope_id;
  int __pyx_v_flowinfo;
  char *__pyx_v_buf;
  Py_ssize_t __pyx_v_buflen;
  struct __pyx_obj_6uvloop_4loop_SockAddrHolder *__pyx_v_ret = 0;
  PyObject *__pyx_v_host = NULL;
  PyObject *__pyx_v_port = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  char const *__pyx_t_16;
  char const *__pyx_t_17;
  Py_UCS4 __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__convert_pyaddr_to_sockaddr", 0);
  __Pyx_INCREF(__pyx_v_addr);

  /* "uvloop/dns.pyx":87
 *         int err
 *         int addr_len
 *         int scope_id = 0             # <<<<<<<<<<<<<<
 *         int flowinfo = 0
 *         char *buf
 */
  __pyx_v_scope_id = 0;

  /* "uvloop/dns.pyx":88
 *         int addr_len
 *         int scope_id = 0
 *         int flowinfo = 0             # <<<<<<<<<<<<<<
 *         char *buf
 *         Py_ssize_t buflen
 */
  __pyx_v_flowinfo = 0;

  /* "uvloop/dns.pyx":93
 *         SockAddrHolder ret
 * 
 *     ret = sockaddrs.get(addr, None)             # <<<<<<<<<<<<<<
 *     if ret is not None and ret.family == family:
 *         memcpy(res, &ret.addr, ret.addr_size)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop_8LruCache_get(__pyx_v_6uvloop_4loop_sockaddrs, __pyx_v_addr, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6uvloop_4loop_SockAddrHolder))))) __PYX_ERR(5, 93, __pyx_L1_error)
  __pyx_v_ret = ((struct __pyx_obj_6uvloop_4loop_SockAddrHolder *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":94
 * 
 *     ret = sockaddrs.get(addr, None)
 *     if ret is not None and ret.family == family:             # <<<<<<<<<<<<<<
 *         memcpy(res, &ret.addr, ret.addr_size)
 *         return
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_ret) != Py_None);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_ret->family == __pyx_v_family);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/dns.pyx":95
 *     ret = sockaddrs.get(addr, None)
 *     if ret is not None and ret.family == family:
 *         memcpy(res, &ret.addr, ret.addr_size)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    (void)(memcpy(__pyx_v_res, (&__pyx_v_ret->addr), __pyx_v_ret->addr_size));

    /* "uvloop/dns.pyx":96
 *     if ret is not None and ret.family == family:
 *         memcpy(res, &ret.addr, ret.addr_size)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     ret = SockAddrHolder.__new__(SockAddrHolder)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/dns.pyx":94
 * 
 *     ret = sockaddrs.get(addr, None)
 *     if ret is not None and ret.family == family:             # <<<<<<<<<<<<<<
 *         memcpy(res, &ret.addr, ret.addr_size)
 *         return
 */
  }

  /* "uvloop/dns.pyx":98
 *         return
 * 
 *     ret = SockAddrHolder.__new__(SockAddrHolder)             # <<<<<<<<<<<<<<
 *     if family == uv.AF_INET:
 *         if not isinstance(addr, tuple):
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop_SockAddrHolder(((PyTypeObject *)__pyx_ptype_6uvloop_4loop_SockAddrHolder), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 98, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_ret, ((struct __pyx_obj_6uvloop_4loop_SockAddrHolder *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":99
 * 
 *     ret = SockAddrHolder.__new__(SockAddrHolder)
 *     if family == uv.AF_INET:             # <<<<<<<<<<<<<<
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET address must be tuple')
 */
  __pyx_t_2 = (__pyx_v_family == AF_INET);
  if (__pyx_t_2) {

    /* "uvloop/dns.pyx":100
 *     ret = SockAddrHolder.__new__(SockAddrHolder)
 *     if family == uv.AF_INET:
 *         if not isinstance(addr, tuple):             # <<<<<<<<<<<<<<
 *             raise TypeError('AF_INET address must be tuple')
 *         if len(addr) != 2:
 */
    __pyx_t_2 = PyTuple_Check(__pyx_v_addr); 
    __pyx_t_3 = (!__pyx_t_2);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/dns.pyx":101
 *     if family == uv.AF_INET:
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET address must be tuple')             # <<<<<<<<<<<<<<
 *         if len(addr) != 2:
 *             raise ValueError('AF_INET address must be tuple of (host, port)')
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__177, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(5, 101, __pyx_L1_error)

      /* "uvloop/dns.pyx":100
 *     ret = SockAddrHolder.__new__(SockAddrHolder)
 *     if family == uv.AF_INET:
 *         if not isinstance(addr, tuple):             # <<<<<<<<<<<<<<
 *             raise TypeError('AF_INET address must be tuple')
 *         if len(addr) != 2:
 */
    }

    /* "uvloop/dns.pyx":102
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET address must be tuple')
 *         if len(addr) != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('AF_INET address must be tuple of (host, port)')
 *         host, port = addr
 */
    __pyx_t_4 = PyObject_Length(__pyx_v_addr); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(5, 102, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_4 != 2);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/dns.pyx":103
 *             raise TypeError('AF_INET address must be tuple')
 *         if len(addr) != 2:
 *             raise ValueError('AF_INET address must be tuple of (host, port)')             # <<<<<<<<<<<<<<
 *         host, port = addr
 *         if isinstance(host, str):
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__178, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 103, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(5, 103, __pyx_L1_error)

      /* "uvloop/dns.pyx":102
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET address must be tuple')
 *         if len(addr) != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError('AF_INET address must be tuple of (host, port)')
 *         host, port = addr
 */
    }

    /* "uvloop/dns.pyx":104
 *         if len(addr) != 2:
 *             raise ValueError('AF_INET address must be tuple of (host, port)')
 *         host, port = addr             # <<<<<<<<<<<<<<
 *         if isinstance(host, str):
 *             try:
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_addr))) || (PyList_CheckExact(__pyx_v_addr))) {
      PyObject* sequence = __pyx_v_addr;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(5, 104, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_v_addr); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
      index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(5, 104, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(5, 104, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }
    __pyx_v_host = __pyx_t_1;
    __pyx_t_1 = 0;
    __pyx_v_port = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "uvloop/dns.pyx":105
 *             raise ValueError('AF_INET address must be tuple of (host, port)')
 *         host, port = addr
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 # idna codec is rather slow, so we try ascii first.
 */
    __pyx_t_3 = PyUnicode_Check(__pyx_v_host); 
    if (__pyx_t_3) {

      /* "uvloop/dns.pyx":106
 *         host, port = addr
 *         if isinstance(host, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        /*try:*/ {

          /* "uvloop/dns.pyx":108
 *             try:
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')             # <<<<<<<<<<<<<<
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_host, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 108, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_n_u_ascii};
            __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 108, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF_SET(__pyx_v_host, __pyx_t_5);
          __pyx_t_5 = 0;

          /* "uvloop/dns.pyx":106
 *         host, port = addr
 *         if isinstance(host, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 */
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L17_try_end;
        __pyx_L12_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "uvloop/dns.pyx":109
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 *             except UnicodeEncodeError:             # <<<<<<<<<<<<<<
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):
 */
        __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeEncodeError);
        if (__pyx_t_12) {
          __Pyx_AddTraceback("uvloop.loop.__convert_pyaddr_to_sockaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(5, 109, __pyx_L14_except_error)
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_6);

          /* "uvloop/dns.pyx":110
 *                 host = host.encode('ascii')
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')             # <<<<<<<<<<<<<<
 *         if not isinstance(host, (bytes, bytearray)):
 *             raise TypeError('host must be a string or bytes object')
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_host, __pyx_n_s_encode); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 110, __pyx_L14_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_15 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_n_u_idna};
            __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(5, 110, __pyx_L14_except_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF_SET(__pyx_v_host, __pyx_t_13);
          __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L13_exception_handled;
        }
        goto __pyx_L14_except_error;

        /* "uvloop/dns.pyx":106
 *         host, port = addr
 *         if isinstance(host, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 */
        __pyx_L14_except_error:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        goto __pyx_L1_error;
        __pyx_L13_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        __pyx_L17_try_end:;
      }

      /* "uvloop/dns.pyx":105
 *             raise ValueError('AF_INET address must be tuple of (host, port)')
 *         host, port = addr
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 # idna codec is rather slow, so we try ascii first.
 */
    }

    /* "uvloop/dns.pyx":111
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             raise TypeError('host must be a string or bytes object')
 * 
 */
    __pyx_t_2 = PyBytes_Check(__pyx_v_host); 
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L21_bool_binop_done;
    }
    __pyx_t_2 = PyByteArray_Check(__pyx_v_host); 
    __pyx_t_3 = __pyx_t_2;
    __pyx_L21_bool_binop_done:;
    __pyx_t_2 = (!__pyx_t_3);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/dns.pyx":112
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):
 *             raise TypeError('host must be a string or bytes object')             # <<<<<<<<<<<<<<
 * 
 *         port = __port_to_int(port, None)
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__179, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(5, 112, __pyx_L1_error)

      /* "uvloop/dns.pyx":111
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             raise TypeError('host must be a string or bytes object')
 * 
 */
    }

    /* "uvloop/dns.pyx":114
 *             raise TypeError('host must be a string or bytes object')
 * 
 *         port = __port_to_int(port, None)             # <<<<<<<<<<<<<<
 * 
 *         ret.addr_size = sizeof(system.sockaddr_in)
 */
    __pyx_t_6 = __pyx_f_6uvloop_4loop___port_to_int(__pyx_v_port, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_port, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "uvloop/dns.pyx":116
 *         port = __port_to_int(port, None)
 * 
 *         ret.addr_size = sizeof(system.sockaddr_in)             # <<<<<<<<<<<<<<
 *         err = uv.uv_ip4_addr(host, <int>port, <system.sockaddr_in*>&ret.addr)
 *         if err < 0:
 */
    __pyx_v_ret->addr_size = (sizeof(struct sockaddr_in));

    /* "uvloop/dns.pyx":117
 * 
 *         ret.addr_size = sizeof(system.sockaddr_in)
 *         err = uv.uv_ip4_addr(host, <int>port, <system.sockaddr_in*>&ret.addr)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
    __pyx_t_16 = __Pyx_PyObject_AsString(__pyx_v_host); if (unlikely((!__pyx_t_16) && PyErr_Occurred())) __PYX_ERR(5, 117, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_port); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 117, __pyx_L1_error)
    __pyx_v_err = uv_ip4_addr(__pyx_t_16, ((int)__pyx_t_12), ((struct sockaddr_in *)(&__pyx_v_ret->addr)));

    /* "uvloop/dns.pyx":118
 *         ret.addr_size = sizeof(system.sockaddr_in)
 *         err = uv.uv_ip4_addr(host, <int>port, <system.sockaddr_in*>&ret.addr)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/dns.pyx":119
 *         err = uv.uv_ip4_addr(host, <int>port, <system.sockaddr_in*>&ret.addr)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *     elif family == uv.AF_INET6:
 */
      __pyx_t_6 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(5, 119, __pyx_L1_error)

      /* "uvloop/dns.pyx":118
 *         ret.addr_size = sizeof(system.sockaddr_in)
 *         err = uv.uv_ip4_addr(host, <int>port, <system.sockaddr_in*>&ret.addr)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    }

    /* "uvloop/dns.pyx":99
 * 
 *     ret = SockAddrHolder.__new__(SockAddrHolder)
 *     if family == uv.AF_INET:             # <<<<<<<<<<<<<<
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET address must be tuple')
 */
    goto __pyx_L6;
  }

  /* "uvloop/dns.pyx":121
 *             raise convert_error(err)
 * 
 *     elif family == uv.AF_INET6:             # <<<<<<<<<<<<<<
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET6 address must be tuple')
 */
  __pyx_t_2 = (__pyx_v_family == AF_INET6);
  if (__pyx_t_2) {

    /* "uvloop/dns.pyx":122
 * 
 *     elif family == uv.AF_INET6:
 *         if not isinstance(addr, tuple):             # <<<<<<<<<<<<<<
 *             raise TypeError('AF_INET6 address must be tuple')
 * 
 */
    __pyx_t_2 = PyTuple_Check(__pyx_v_addr); 
    __pyx_t_3 = (!__pyx_t_2);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/dns.pyx":123
 *     elif family == uv.AF_INET6:
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET6 address must be tuple')             # <<<<<<<<<<<<<<
 * 
 *         addr_len = len(addr)
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__180, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(5, 123, __pyx_L1_error)

      /* "uvloop/dns.pyx":122
 * 
 *     elif family == uv.AF_INET6:
 *         if not isinstance(addr, tuple):             # <<<<<<<<<<<<<<
 *             raise TypeError('AF_INET6 address must be tuple')
 * 
 */
    }

    /* "uvloop/dns.pyx":125
 *             raise TypeError('AF_INET6 address must be tuple')
 * 
 *         addr_len = len(addr)             # <<<<<<<<<<<<<<
 *         if addr_len < 2 or addr_len > 4:
 *             raise ValueError(
 */
    __pyx_t_4 = PyObject_Length(__pyx_v_addr); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(5, 125, __pyx_L1_error)
    __pyx_v_addr_len = __pyx_t_4;

    /* "uvloop/dns.pyx":126
 * 
 *         addr_len = len(addr)
 *         if addr_len < 2 or addr_len > 4:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'AF_INET6 must be a tuple of 2-4 parameters: '
 */
    __pyx_t_2 = (__pyx_v_addr_len < 2);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_addr_len > 4);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L26_bool_binop_done:;
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/dns.pyx":127
 *         addr_len = len(addr)
 *         if addr_len < 2 or addr_len > 4:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'AF_INET6 must be a tuple of 2-4 parameters: '
 *                 '(host, port, flowinfo?, scope_id?)')
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__181, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(5, 127, __pyx_L1_error)

      /* "uvloop/dns.pyx":126
 * 
 *         addr_len = len(addr)
 *         if addr_len < 2 or addr_len > 4:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 'AF_INET6 must be a tuple of 2-4 parameters: '
 */
    }

    /* "uvloop/dns.pyx":131
 *                 '(host, port, flowinfo?, scope_id?)')
 * 
 *         host = addr[0]             # <<<<<<<<<<<<<<
 *         if isinstance(host, str):
 *             try:
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_addr, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_host = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "uvloop/dns.pyx":132
 * 
 *         host = addr[0]
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 # idna codec is rather slow, so we try ascii first.
 */
    __pyx_t_3 = PyUnicode_Check(__pyx_v_host); 
    if (__pyx_t_3) {

      /* "uvloop/dns.pyx":133
 *         host = addr[0]
 *         if isinstance(host, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {

          /* "uvloop/dns.pyx":135
 *             try:
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')             # <<<<<<<<<<<<<<
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_host, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 135, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_5 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_n_u_ascii};
            __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 135, __pyx_L29_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF_SET(__pyx_v_host, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "uvloop/dns.pyx":133
 *         host = addr[0]
 *         if isinstance(host, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L34_try_end;
        __pyx_L29_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "uvloop/dns.pyx":136
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 *             except UnicodeEncodeError:             # <<<<<<<<<<<<<<
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):
 */
        __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeEncodeError);
        if (__pyx_t_12) {
          __Pyx_AddTraceback("uvloop.loop.__convert_pyaddr_to_sockaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_1, &__pyx_t_5) < 0) __PYX_ERR(5, 136, __pyx_L31_except_error)
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_XGOTREF(__pyx_t_5);

          /* "uvloop/dns.pyx":137
 *                 host = host.encode('ascii')
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')             # <<<<<<<<<<<<<<
 *         if not isinstance(host, (bytes, bytearray)):
 *             raise TypeError('host must be a string or bytes object')
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_host, __pyx_n_s_encode); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 137, __pyx_L31_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_15 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_n_u_idna};
            __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(5, 137, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF_SET(__pyx_v_host, __pyx_t_13);
          __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L30_exception_handled;
        }
        goto __pyx_L31_except_error;

        /* "uvloop/dns.pyx":133
 *         host = addr[0]
 *         if isinstance(host, str):
 *             try:             # <<<<<<<<<<<<<<
 *                 # idna codec is rather slow, so we try ascii first.
 *                 host = host.encode('ascii')
 */
        __pyx_L31_except_error:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L30_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
        __pyx_L34_try_end:;
      }

      /* "uvloop/dns.pyx":132
 * 
 *         host = addr[0]
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             try:
 *                 # idna codec is rather slow, so we try ascii first.
 */
    }

    /* "uvloop/dns.pyx":138
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             raise TypeError('host must be a string or bytes object')
 * 
 */
    __pyx_t_2 = PyBytes_Check(__pyx_v_host); 
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L38_bool_binop_done;
    }
    __pyx_t_2 = PyByteArray_Check(__pyx_v_host); 
    __pyx_t_3 = __pyx_t_2;
    __pyx_L38_bool_binop_done:;
    __pyx_t_2 = (!__pyx_t_3);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/dns.pyx":139
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):
 *             raise TypeError('host must be a string or bytes object')             # <<<<<<<<<<<<<<
 * 
 *         port = __port_to_int(addr[1], None)
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__179, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(5, 139, __pyx_L1_error)

      /* "uvloop/dns.pyx":138
 *             except UnicodeEncodeError:
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):             # <<<<<<<<<<<<<<
 *             raise TypeError('host must be a string or bytes object')
 * 
 */
    }

    /* "uvloop/dns.pyx":141
 *             raise TypeError('host must be a string or bytes object')
 * 
 *         port = __port_to_int(addr[1], None)             # <<<<<<<<<<<<<<
 * 
 *         if addr_len > 2:
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_addr, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __pyx_f_6uvloop_4loop___port_to_int(__pyx_t_5, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_port = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/dns.pyx":143
 *         port = __port_to_int(addr[1], None)
 * 
 *         if addr_len > 2:             # <<<<<<<<<<<<<<
 *             flowinfo = addr[2]
 *         if addr_len > 3:
 */
    __pyx_t_2 = (__pyx_v_addr_len > 2);
    if (__pyx_t_2) {

      /* "uvloop/dns.pyx":144
 * 
 *         if addr_len > 2:
 *             flowinfo = addr[2]             # <<<<<<<<<<<<<<
 *         if addr_len > 3:
 *             scope_id = addr[3]
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_addr, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 144, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_flowinfo = __pyx_t_12;

      /* "uvloop/dns.pyx":143
 *         port = __port_to_int(addr[1], None)
 * 
 *         if addr_len > 2:             # <<<<<<<<<<<<<<
 *             flowinfo = addr[2]
 *         if addr_len > 3:
 */
    }

    /* "uvloop/dns.pyx":145
 *         if addr_len > 2:
 *             flowinfo = addr[2]
 *         if addr_len > 3:             # <<<<<<<<<<<<<<
 *             scope_id = addr[3]
 * 
 */
    __pyx_t_2 = (__pyx_v_addr_len > 3);
    if (__pyx_t_2) {

      /* "uvloop/dns.pyx":146
 *             flowinfo = addr[2]
 *         if addr_len > 3:
 *             scope_id = addr[3]             # <<<<<<<<<<<<<<
 * 
 *         ret.addr_size = sizeof(system.sockaddr_in6)
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_addr, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 146, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_scope_id = __pyx_t_12;

      /* "uvloop/dns.pyx":145
 *         if addr_len > 2:
 *             flowinfo = addr[2]
 *         if addr_len > 3:             # <<<<<<<<<<<<<<
 *             scope_id = addr[3]
 * 
 */
    }

    /* "uvloop/dns.pyx":148
 *             scope_id = addr[3]
 * 
 *         ret.addr_size = sizeof(system.sockaddr_in6)             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_ip6_addr(host, port, <system.sockaddr_in6*>&ret.addr)
 */
    __pyx_v_ret->addr_size = (sizeof(struct sockaddr_in6));

    /* "uvloop/dns.pyx":150
 *         ret.addr_size = sizeof(system.sockaddr_in6)
 * 
 *         err = uv.uv_ip6_addr(host, port, <system.sockaddr_in6*>&ret.addr)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             raise convert_error(err)
 */
    __pyx_t_17 = __Pyx_PyObject_AsString(__pyx_v_host); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(5, 150, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_port); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 150, __pyx_L1_error)
    __pyx_v_err = uv_ip6_addr(__pyx_t_17, __pyx_t_12, ((struct sockaddr_in6 *)(&__pyx_v_ret->addr)));

    /* "uvloop/dns.pyx":151
 * 
 *         err = uv.uv_ip6_addr(host, port, <system.sockaddr_in6*>&ret.addr)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_2 = (__pyx_v_err < 0);
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/dns.pyx":152
 *         err = uv.uv_ip6_addr(host, port, <system.sockaddr_in6*>&ret.addr)
 *         if err < 0:
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         (<system.sockaddr_in6*>&ret.addr).sin6_flowinfo = flowinfo
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(5, 152, __pyx_L1_error)

      /* "uvloop/dns.pyx":151
 * 
 *         err = uv.uv_ip6_addr(host, port, <system.sockaddr_in6*>&ret.addr)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    }

    /* "uvloop/dns.pyx":154
 *             raise convert_error(err)
 * 
 *         (<system.sockaddr_in6*>&ret.addr).sin6_flowinfo = flowinfo             # <<<<<<<<<<<<<<
 *         (<system.sockaddr_in6*>&ret.addr).sin6_scope_id = scope_id
 * 
 */
    ((struct sockaddr_in6 *)(&__pyx_v_ret->addr))->sin6_flowinfo = __pyx_v_flowinfo;

    /* "uvloop/dns.pyx":155
 * 
 *         (<system.sockaddr_in6*>&ret.addr).sin6_flowinfo = flowinfo
 *         (<system.sockaddr_in6*>&ret.addr).sin6_scope_id = scope_id             # <<<<<<<<<<<<<<
 * 
 *     elif family == uv.AF_UNIX:
 */
    ((struct sockaddr_in6 *)(&__pyx_v_ret->addr))->sin6_scope_id = __pyx_v_scope_id;

    /* "uvloop/dns.pyx":121
 *             raise convert_error(err)
 * 
 *     elif family == uv.AF_INET6:             # <<<<<<<<<<<<<<
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET6 address must be tuple')
 */
    goto __pyx_L6;
  }

  /* "uvloop/dns.pyx":157
 *         (<system.sockaddr_in6*>&ret.addr).sin6_scope_id = scope_id
 * 
 *     elif family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *         if isinstance(addr, str):
 *             addr = addr.encode(sys_getfilesystemencoding())
 */
  __pyx_t_2 = (__pyx_v_family == AF_UNIX);
  if (likely(__pyx_t_2)) {

    /* "uvloop/dns.pyx":158
 * 
 *     elif family == uv.AF_UNIX:
 *         if isinstance(addr, str):             # <<<<<<<<<<<<<<
 *             addr = addr.encode(sys_getfilesystemencoding())
 *         elif not isinstance(addr, bytes):
 */
    __pyx_t_2 = PyUnicode_Check(__pyx_v_addr); 
    if (__pyx_t_2) {

      /* "uvloop/dns.pyx":159
 *     elif family == uv.AF_UNIX:
 *         if isinstance(addr, str):
 *             addr = addr.encode(sys_getfilesystemencoding())             # <<<<<<<<<<<<<<
 *         elif not isinstance(addr, bytes):
 *             raise TypeError('AF_UNIX address must be a str or a bytes object')
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_addr, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_6uvloop_4loop_sys_getfilesystemencoding);
      __pyx_t_13 = __pyx_v_6uvloop_4loop_sys_getfilesystemencoding; __pyx_t_14 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __pyx_t_13 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 159, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_addr, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "uvloop/dns.pyx":158
 * 
 *     elif family == uv.AF_UNIX:
 *         if isinstance(addr, str):             # <<<<<<<<<<<<<<
 *             addr = addr.encode(sys_getfilesystemencoding())
 *         elif not isinstance(addr, bytes):
 */
      goto __pyx_L43;
    }

    /* "uvloop/dns.pyx":160
 *         if isinstance(addr, str):
 *             addr = addr.encode(sys_getfilesystemencoding())
 *         elif not isinstance(addr, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError('AF_UNIX address must be a str or a bytes object')
 * 
 */
    __pyx_t_2 = PyBytes_Check(__pyx_v_addr); 
    __pyx_t_3 = (!__pyx_t_2);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/dns.pyx":161
 *             addr = addr.encode(sys_getfilesystemencoding())
 *         elif not isinstance(addr, bytes):
 *             raise TypeError('AF_UNIX address must be a str or a bytes object')             # <<<<<<<<<<<<<<
 * 
 *         PyBytes_AsStringAndSize(addr, &buf, &buflen)
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__182, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(5, 161, __pyx_L1_error)

      /* "uvloop/dns.pyx":160
 *         if isinstance(addr, str):
 *             addr = addr.encode(sys_getfilesystemencoding())
 *         elif not isinstance(addr, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError('AF_UNIX address must be a str or a bytes object')
 * 
 */
    }
    __pyx_L43:;

    /* "uvloop/dns.pyx":163
 *             raise TypeError('AF_UNIX address must be a str or a bytes object')
 * 
 *         PyBytes_AsStringAndSize(addr, &buf, &buflen)             # <<<<<<<<<<<<<<
 *         if buflen > 107:
 *             raise ValueError(
 */
    __pyx_t_12 = PyBytes_AsStringAndSize(__pyx_v_addr, (&__pyx_v_buf), (&__pyx_v_buflen)); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(5, 163, __pyx_L1_error)

    /* "uvloop/dns.pyx":164
 * 
 *         PyBytes_AsStringAndSize(addr, &buf, &buflen)
 *         if buflen > 107:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f'unix socket path {addr!r} is longer than 107 characters')
 */
    __pyx_t_3 = (__pyx_v_buflen > 0x6B);
    if (unlikely(__pyx_t_3)) {

      /* "uvloop/dns.pyx":166
 *         if buflen > 107:
 *             raise ValueError(
 *                 f'unix socket path {addr!r} is longer than 107 characters')             # <<<<<<<<<<<<<<
 * 
 *         ret.addr_size = sizeof(system.sockaddr_un)
 */
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_18 = 127;
      __Pyx_INCREF(__pyx_kp_u_unix_socket_path);
      __pyx_t_4 += 17;
      __Pyx_GIVEREF(__pyx_kp_u_unix_socket_path);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_unix_socket_path);
      __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_addr), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_18 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_18) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_18;
      __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_longer_than_107_characters);
      __pyx_t_4 += 30;
      __Pyx_GIVEREF(__pyx_kp_u_is_longer_than_107_characters);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_is_longer_than_107_characters);
      __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_4, __pyx_t_18); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/dns.pyx":165
 *         PyBytes_AsStringAndSize(addr, &buf, &buflen)
 *         if buflen > 107:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 f'unix socket path {addr!r} is longer than 107 characters')
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(5, 165, __pyx_L1_error)

      /* "uvloop/dns.pyx":164
 * 
 *         PyBytes_AsStringAndSize(addr, &buf, &buflen)
 *         if buflen > 107:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f'unix socket path {addr!r} is longer than 107 characters')
 */
    }

    /* "uvloop/dns.pyx":168
 *                 f'unix socket path {addr!r} is longer than 107 characters')
 * 
 *         ret.addr_size = sizeof(system.sockaddr_un)             # <<<<<<<<<<<<<<
 *         memset(&ret.addr, 0, sizeof(system.sockaddr_un))
 *         (<system.sockaddr_un*>&ret.addr).sun_family = uv.AF_UNIX
 */
    __pyx_v_ret->addr_size = (sizeof(struct sockaddr_un));

    /* "uvloop/dns.pyx":169
 * 
 *         ret.addr_size = sizeof(system.sockaddr_un)
 *         memset(&ret.addr, 0, sizeof(system.sockaddr_un))             # <<<<<<<<<<<<<<
 *         (<system.sockaddr_un*>&ret.addr).sun_family = uv.AF_UNIX
 *         memcpy((<system.sockaddr_un*>&ret.addr).sun_path, buf, buflen)
 */
    (void)(memset((&__pyx_v_ret->addr), 0, (sizeof(struct sockaddr_un))));

    /* "uvloop/dns.pyx":170
 *         ret.addr_size = sizeof(system.sockaddr_un)
 *         memset(&ret.addr, 0, sizeof(system.sockaddr_un))
 *         (<system.sockaddr_un*>&ret.addr).sun_family = uv.AF_UNIX             # <<<<<<<<<<<<<<
 *         memcpy((<system.sockaddr_un*>&ret.addr).sun_path, buf, buflen)
 * 
 */
    ((struct sockaddr_un *)(&__pyx_v_ret->addr))->sun_family = AF_UNIX;

    /* "uvloop/dns.pyx":171
 *         memset(&ret.addr, 0, sizeof(system.sockaddr_un))
 *         (<system.sockaddr_un*>&ret.addr).sun_family = uv.AF_UNIX
 *         memcpy((<system.sockaddr_un*>&ret.addr).sun_path, buf, buflen)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
    (void)(memcpy(((struct sockaddr_un *)(&__pyx_v_ret->addr))->sun_path, __pyx_v_buf, __pyx_v_buflen));

    /* "uvloop/dns.pyx":157
 *         (<system.sockaddr_in6*>&ret.addr).sin6_scope_id = scope_id
 * 
 *     elif family == uv.AF_UNIX:             # <<<<<<<<<<<<<<
 *         if isinstance(addr, str):
 *             addr = addr.encode(sys_getfilesystemencoding())
 */
    goto __pyx_L6;
  }

  /* "uvloop/dns.pyx":174
 * 
 *     else:
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             f'expected AF_INET, AF_INET6, or AF_UNIX family, got {family}')
 * 
 */
  /*else*/ {

    /* "uvloop/dns.pyx":175
 *     else:
 *         raise ValueError(
 *             f'expected AF_INET, AF_INET6, or AF_UNIX family, got {family}')             # <<<<<<<<<<<<<<
 * 
 *     ret.family = family
 */
    __pyx_t_1 = __Pyx_PyUnicode_From_int(__pyx_v_family, 0, ' ', 'd'); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_kp_u_expected_AF_INET_AF_INET6_or_AF, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/dns.pyx":174
 * 
 *     else:
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             f'expected AF_INET, AF_INET6, or AF_UNIX family, got {family}')
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(5, 174, __pyx_L1_error)
  }
  __pyx_L6:;

  /* "uvloop/dns.pyx":177
 *             f'expected AF_INET, AF_INET6, or AF_UNIX family, got {family}')
 * 
 *     ret.family = family             # <<<<<<<<<<<<<<
 *     sockaddrs[addr] = ret
 *     memcpy(res, &ret.addr, ret.addr_size)
 */
  __pyx_v_ret->family = __pyx_v_family;

  /* "uvloop/dns.pyx":178
 * 
 *     ret.family = family
 *     sockaddrs[addr] = ret             # <<<<<<<<<<<<<<
 *     memcpy(res, &ret.addr, ret.addr_size)
 * 
 */
  if (unlikely((PyObject_SetItem(((PyObject *)__pyx_v_6uvloop_4loop_sockaddrs), __pyx_v_addr, ((PyObject *)__pyx_v_ret)) < 0))) __PYX_ERR(5, 178, __pyx_L1_error)

  /* "uvloop/dns.pyx":179
 *     ret.family = family
 *     sockaddrs[addr] = ret
 *     memcpy(res, &ret.addr, ret.addr_size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (void)(memcpy(__pyx_v_res, (&__pyx_v_ret->addr), __pyx_v_ret->addr_size));

  /* "uvloop/dns.pyx":82
 * 
 * 
 * cdef __convert_pyaddr_to_sockaddr(int family, object addr,             # <<<<<<<<<<<<<<
 *                                   system.sockaddr* res):
 *     cdef:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.__convert_pyaddr_to_sockaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_host);
  __Pyx_XDECREF(__pyx_v_port);
  __Pyx_XDECREF(__pyx_v_addr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":182
 * 
 * 
 * cdef __static_getaddrinfo(object host, object port,             # <<<<<<<<<<<<<<
 *                           int family, int type,
 *                           int proto,
 */

static PyObject *__pyx_f_6uvloop_4loop___static_getaddrinfo(PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, struct sockaddr *__pyx_v_addr) {
  PyObject *__pyx_v_hp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__static_getaddrinfo", 0);
  __Pyx_INCREF(__pyx_v_port);

  /* "uvloop/dns.pyx":187
 *                           system.sockaddr *addr):
 * 
 *     if proto not in {0, uv.IPPROTO_TCP, uv.IPPROTO_UDP}:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = __pyx_v_proto;
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 != IPPROTO_TCP);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 != IPPROTO_UDP);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __pyx_t_3 = __pyx_t_2;
  if (__pyx_t_3) {

    /* "uvloop/dns.pyx":188
 * 
 *     if proto not in {0, uv.IPPROTO_TCP, uv.IPPROTO_UDP}:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if _is_sock_stream(type):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/dns.pyx":187
 *                           system.sockaddr *addr):
 * 
 *     if proto not in {0, uv.IPPROTO_TCP, uv.IPPROTO_UDP}:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/dns.pyx":190
 *         return
 * 
 *     if _is_sock_stream(type):             # <<<<<<<<<<<<<<
 *         proto = uv.IPPROTO_TCP
 *     elif _is_sock_dgram(type):
 */
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_f_6uvloop_4loop__is_sock_stream(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_3) {

    /* "uvloop/dns.pyx":191
 * 
 *     if _is_sock_stream(type):
 *         proto = uv.IPPROTO_TCP             # <<<<<<<<<<<<<<
 *     elif _is_sock_dgram(type):
 *         proto = uv.IPPROTO_UDP
 */
    __pyx_v_proto = IPPROTO_TCP;

    /* "uvloop/dns.pyx":190
 *         return
 * 
 *     if _is_sock_stream(type):             # <<<<<<<<<<<<<<
 *         proto = uv.IPPROTO_TCP
 *     elif _is_sock_dgram(type):
 */
    goto __pyx_L7;
  }

  /* "uvloop/dns.pyx":192
 *     if _is_sock_stream(type):
 *         proto = uv.IPPROTO_TCP
 *     elif _is_sock_dgram(type):             # <<<<<<<<<<<<<<
 *         proto = uv.IPPROTO_UDP
 *     else:
 */
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __pyx_f_6uvloop_4loop__is_sock_dgram(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(5, 192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_3) {

    /* "uvloop/dns.pyx":193
 *         proto = uv.IPPROTO_TCP
 *     elif _is_sock_dgram(type):
 *         proto = uv.IPPROTO_UDP             # <<<<<<<<<<<<<<
 *     else:
 *         return
 */
    __pyx_v_proto = IPPROTO_UDP;

    /* "uvloop/dns.pyx":192
 *     if _is_sock_stream(type):
 *         proto = uv.IPPROTO_TCP
 *     elif _is_sock_dgram(type):             # <<<<<<<<<<<<<<
 *         proto = uv.IPPROTO_UDP
 *     else:
 */
    goto __pyx_L7;
  }

  /* "uvloop/dns.pyx":195
 *         proto = uv.IPPROTO_UDP
 *     else:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }
  __pyx_L7:;

  /* "uvloop/dns.pyx":197
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         port = __port_to_int(port, proto)
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "uvloop/dns.pyx":198
 * 
 *     try:
 *         port = __port_to_int(port, proto)             # <<<<<<<<<<<<<<
 *     except Exception:
 *         return
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_proto); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 198, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __pyx_f_6uvloop_4loop___port_to_int(__pyx_v_port, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 198, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_port, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "uvloop/dns.pyx":197
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         port = __port_to_int(port, proto)
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L13_try_end;
    __pyx_L8_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/dns.pyx":199
 *     try:
 *         port = __port_to_int(port, proto)
 *     except Exception:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_1 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_1) {
      __Pyx_AddTraceback("uvloop.loop.__static_getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_9) < 0) __PYX_ERR(5, 199, __pyx_L10_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_9);

      /* "uvloop/dns.pyx":200
 *         port = __port_to_int(port, proto)
 *     except Exception:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     hp = (host, port)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L11_except_return;
    }
    goto __pyx_L10_except_error;

    /* "uvloop/dns.pyx":197
 *         return
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         port = __port_to_int(port, proto)
 *     except Exception:
 */
    __pyx_L10_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L11_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L13_try_end:;
  }

  /* "uvloop/dns.pyx":202
 *         return
 * 
 *     hp = (host, port)             # <<<<<<<<<<<<<<
 *     if family == uv.AF_UNSPEC:
 *         try:
 */
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_v_host);
  __Pyx_GIVEREF(__pyx_v_host);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_v_host)) __PYX_ERR(5, 202, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_port);
  __Pyx_GIVEREF(__pyx_v_port);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_v_port)) __PYX_ERR(5, 202, __pyx_L1_error);
  __pyx_v_hp = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "uvloop/dns.pyx":203
 * 
 *     hp = (host, port)
 *     if family == uv.AF_UNSPEC:             # <<<<<<<<<<<<<<
 *         try:
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET, hp, addr)
 */
  __pyx_t_3 = (__pyx_v_family == AF_UNSPEC);
  if (__pyx_t_3) {

    /* "uvloop/dns.pyx":204
 *     hp = (host, port)
 *     if family == uv.AF_UNSPEC:
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET, hp, addr)
 *         except Exception:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "uvloop/dns.pyx":205
 *     if family == uv.AF_UNSPEC:
 *         try:
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET, hp, addr)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
        __pyx_t_9 = __pyx_f_6uvloop_4loop___convert_pyaddr_to_sockaddr(AF_INET, __pyx_v_hp, __pyx_v_addr); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 205, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "uvloop/dns.pyx":204
 *     hp = (host, port)
 *     if family == uv.AF_UNSPEC:
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET, hp, addr)
 *         except Exception:
 */
      }

      /* "uvloop/dns.pyx":209
 *             pass
 *         else:
 *             return (uv.AF_INET, type, proto)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
      /*else:*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_9 = __Pyx_PyInt_From_int(AF_INET); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 209, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 209, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_proto); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 209, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 209, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_9);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9)) __PYX_ERR(5, 209, __pyx_L19_except_error);
        __Pyx_GIVEREF(__pyx_t_4);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_4)) __PYX_ERR(5, 209, __pyx_L19_except_error);
        __Pyx_GIVEREF(__pyx_t_5);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_5)) __PYX_ERR(5, 209, __pyx_L19_except_error);
        __pyx_t_9 = 0;
        __pyx_t_4 = 0;
        __pyx_t_5 = 0;
        __pyx_r = __pyx_t_10;
        __pyx_t_10 = 0;
        goto __pyx_L20_except_return;
      }
      __pyx_L17_error:;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/dns.pyx":206
 *         try:
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET, hp, addr)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
      __pyx_t_1 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_1) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L18_exception_handled;
      }
      goto __pyx_L19_except_error;

      /* "uvloop/dns.pyx":204
 *     hp = (host, port)
 *     if family == uv.AF_UNSPEC:
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET, hp, addr)
 *         except Exception:
 */
      __pyx_L19_except_error:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L20_except_return:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L18_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
    }

    /* "uvloop/dns.pyx":211
 *             return (uv.AF_INET, type, proto)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET6, hp, addr)
 *         except Exception:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "uvloop/dns.pyx":212
 * 
 *         try:
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET6, hp, addr)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
        __pyx_t_10 = __pyx_f_6uvloop_4loop___convert_pyaddr_to_sockaddr(AF_INET6, __pyx_v_hp, __pyx_v_addr); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 212, __pyx_L23_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/dns.pyx":211
 *             return (uv.AF_INET, type, proto)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET6, hp, addr)
 *         except Exception:
 */
      }

      /* "uvloop/dns.pyx":216
 *             pass
 *         else:
 *             return (uv.AF_INET6, type, proto)             # <<<<<<<<<<<<<<
 * 
 *     else:
 */
      /*else:*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_10 = __Pyx_PyInt_From_int(AF_INET6); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 216, __pyx_L25_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 216, __pyx_L25_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_proto); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 216, __pyx_L25_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 216, __pyx_L25_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_10);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10)) __PYX_ERR(5, 216, __pyx_L25_except_error);
        __Pyx_GIVEREF(__pyx_t_5);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_5)) __PYX_ERR(5, 216, __pyx_L25_except_error);
        __Pyx_GIVEREF(__pyx_t_4);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_4)) __PYX_ERR(5, 216, __pyx_L25_except_error);
        __pyx_t_10 = 0;
        __pyx_t_5 = 0;
        __pyx_t_4 = 0;
        __pyx_r = __pyx_t_9;
        __pyx_t_9 = 0;
        goto __pyx_L26_except_return;
      }
      __pyx_L23_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/dns.pyx":213
 *         try:
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET6, hp, addr)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
      __pyx_t_1 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_1) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L24_exception_handled;
      }
      goto __pyx_L25_except_error;

      /* "uvloop/dns.pyx":211
 *             return (uv.AF_INET, type, proto)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET6, hp, addr)
 *         except Exception:
 */
      __pyx_L25_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L26_except_return:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L0;
      __pyx_L24_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    }

    /* "uvloop/dns.pyx":203
 * 
 *     hp = (host, port)
 *     if family == uv.AF_UNSPEC:             # <<<<<<<<<<<<<<
 *         try:
 *             __convert_pyaddr_to_sockaddr(uv.AF_INET, hp, addr)
 */
    goto __pyx_L16;
  }

  /* "uvloop/dns.pyx":219
 * 
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(family, hp, addr)
 *         except Exception:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "uvloop/dns.pyx":220
 *     else:
 *         try:
 *             __convert_pyaddr_to_sockaddr(family, hp, addr)             # <<<<<<<<<<<<<<
 *         except Exception:
 *             pass
 */
        __pyx_t_9 = __pyx_f_6uvloop_4loop___convert_pyaddr_to_sockaddr(__pyx_v_family, __pyx_v_hp, __pyx_v_addr); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 220, __pyx_L29_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "uvloop/dns.pyx":219
 * 
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(family, hp, addr)
 *         except Exception:
 */
      }

      /* "uvloop/dns.pyx":224
 *             pass
 *         else:
 *             return (family, type, proto)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      /*else:*/ {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_family); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 224, __pyx_L31_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 224, __pyx_L31_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_proto); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 224, __pyx_L31_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 224, __pyx_L31_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_9);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9)) __PYX_ERR(5, 224, __pyx_L31_except_error);
        __Pyx_GIVEREF(__pyx_t_4);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_4)) __PYX_ERR(5, 224, __pyx_L31_except_error);
        __Pyx_GIVEREF(__pyx_t_5);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_5)) __PYX_ERR(5, 224, __pyx_L31_except_error);
        __pyx_t_9 = 0;
        __pyx_t_4 = 0;
        __pyx_t_5 = 0;
        __pyx_r = __pyx_t_10;
        __pyx_t_10 = 0;
        goto __pyx_L32_except_return;
      }
      __pyx_L29_error:;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/dns.pyx":221
 *         try:
 *             __convert_pyaddr_to_sockaddr(family, hp, addr)
 *         except Exception:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
      __pyx_t_1 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_1) {
        __Pyx_ErrRestore(0,0,0);
        goto __pyx_L30_exception_handled;
      }
      goto __pyx_L31_except_error;

      /* "uvloop/dns.pyx":219
 * 
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             __convert_pyaddr_to_sockaddr(family, hp, addr)
 *         except Exception:
 */
      __pyx_L31_except_error:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L32_except_return:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L30_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
    }
  }
  __pyx_L16:;

  /* "uvloop/dns.pyx":182
 * 
 * 
 * cdef __static_getaddrinfo(object host, object port,             # <<<<<<<<<<<<<<
 *                           int family, int type,
 *                           int proto,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uvloop.loop.__static_getaddrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hp);
  __Pyx_XDECREF(__pyx_v_port);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":227
 * 
 * 
 * cdef __static_getaddrinfo_pyaddr(object host, object port,             # <<<<<<<<<<<<<<
 *                                  int family, int type,
 *                                  int proto, int flags):
 */

static PyObject *__pyx_f_6uvloop_4loop___static_getaddrinfo_pyaddr(PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_flags) {
  struct sockaddr_storage __pyx_v_addr;
  PyObject *__pyx_v_triplet = 0;
  PyObject *__pyx_v_af = NULL;
  PyObject *__pyx_v_pyaddr = NULL;
  PyObject *__pyx_v_canon_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  unsigned int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__static_getaddrinfo_pyaddr", 1);

  /* "uvloop/dns.pyx":235
 *         object triplet
 * 
 *     triplet = __static_getaddrinfo(             # <<<<<<<<<<<<<<
 *         host, port, family, type,
 *         proto, <system.sockaddr*>&addr)
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___static_getaddrinfo(__pyx_v_host, __pyx_v_port, __pyx_v_family, __pyx_v_type, __pyx_v_proto, ((struct sockaddr *)(&__pyx_v_addr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_triplet = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":238
 *         host, port, family, type,
 *         proto, <system.sockaddr*>&addr)
 *     if triplet is None:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_v_triplet == Py_None);
  if (__pyx_t_2) {

    /* "uvloop/dns.pyx":239
 *         proto, <system.sockaddr*>&addr)
 *     if triplet is None:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     af, type, proto = triplet
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/dns.pyx":238
 *         host, port, family, type,
 *         proto, <system.sockaddr*>&addr)
 *     if triplet is None:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  }

  /* "uvloop/dns.pyx":241
 *         return
 * 
 *     af, type, proto = triplet             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_triplet))) || (PyList_CheckExact(__pyx_v_triplet))) {
    PyObject* sequence = __pyx_v_triplet;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(5, 241, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_v_triplet); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
    index = 0; __pyx_t_1 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(5, 241, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(5, 241, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_af = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_type = __pyx_t_7;
  __pyx_v_proto = __pyx_t_8;

  /* "uvloop/dns.pyx":243
 *     af, type, proto = triplet
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         pyaddr = __convert_sockaddr_to_pyaddr(<system.sockaddr*>&addr)
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    /*try:*/ {

      /* "uvloop/dns.pyx":244
 * 
 *     try:
 *         pyaddr = __convert_sockaddr_to_pyaddr(<system.sockaddr*>&addr)             # <<<<<<<<<<<<<<
 *     except Exception:
 *         return
 */
      __pyx_t_4 = __pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(((struct sockaddr *)(&__pyx_v_addr))); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 244, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_pyaddr = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "uvloop/dns.pyx":243
 *     af, type, proto = triplet
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         pyaddr = __convert_sockaddr_to_pyaddr(<system.sockaddr*>&addr)
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    goto __pyx_L11_try_end;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/dns.pyx":245
 *     try:
 *         pyaddr = __convert_sockaddr_to_pyaddr(<system.sockaddr*>&addr)
 *     except Exception:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.__static_getaddrinfo_pyaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(5, 245, __pyx_L8_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_1);

      /* "uvloop/dns.pyx":246
 *         pyaddr = __convert_sockaddr_to_pyaddr(<system.sockaddr*>&addr)
 *     except Exception:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     # When the host is an IP while type is one of TCP or UDP, different libc
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_except_return;
    }
    goto __pyx_L8_except_error;

    /* "uvloop/dns.pyx":243
 *     af, type, proto = triplet
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         pyaddr = __convert_sockaddr_to_pyaddr(<system.sockaddr*>&addr)
 *     except Exception:
 */
    __pyx_L8_except_error:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L1_error;
    __pyx_L9_except_return:;
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
    goto __pyx_L0;
    __pyx_L11_try_end:;
  }

  /* "uvloop/dns.pyx":271
 *     # is an IP and type is one of TCP or UDP.
 *     # All other cases are still asyncio-compatible.
 *     if flags & socket_AI_CANONNAME:             # <<<<<<<<<<<<<<
 *         if isinstance(host, str):
 *             canon_name = host
 */
  __pyx_t_2 = ((__pyx_v_flags & __pyx_v_6uvloop_4loop_socket_AI_CANONNAME) != 0);
  if (__pyx_t_2) {

    /* "uvloop/dns.pyx":272
 *     # All other cases are still asyncio-compatible.
 *     if flags & socket_AI_CANONNAME:
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             canon_name = host
 *         else:
 */
    __pyx_t_2 = PyUnicode_Check(__pyx_v_host); 
    if (__pyx_t_2) {

      /* "uvloop/dns.pyx":273
 *     if flags & socket_AI_CANONNAME:
 *         if isinstance(host, str):
 *             canon_name = host             # <<<<<<<<<<<<<<
 *         else:
 *             canon_name = host.decode('ascii')
 */
      __Pyx_INCREF(__pyx_v_host);
      __pyx_v_canon_name = __pyx_v_host;

      /* "uvloop/dns.pyx":272
 *     # All other cases are still asyncio-compatible.
 *     if flags & socket_AI_CANONNAME:
 *         if isinstance(host, str):             # <<<<<<<<<<<<<<
 *             canon_name = host
 *         else:
 */
      goto __pyx_L15;
    }

    /* "uvloop/dns.pyx":275
 *             canon_name = host
 *         else:
 *             canon_name = host.decode('ascii')             # <<<<<<<<<<<<<<
 *     else:
 *         canon_name = ''
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_host, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_12 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_12 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_n_u_ascii};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 275, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_v_canon_name = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __pyx_L15:;

    /* "uvloop/dns.pyx":271
 *     # is an IP and type is one of TCP or UDP.
 *     # All other cases are still asyncio-compatible.
 *     if flags & socket_AI_CANONNAME:             # <<<<<<<<<<<<<<
 *         if isinstance(host, str):
 *             canon_name = host
 */
    goto __pyx_L14;
  }

  /* "uvloop/dns.pyx":277
 *             canon_name = host.decode('ascii')
 *     else:
 *         canon_name = ''             # <<<<<<<<<<<<<<
 * 
 *     return (
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_u__10);
    __pyx_v_canon_name = __pyx_kp_u__10;
  }
  __pyx_L14:;

  /* "uvloop/dns.pyx":279
 *         canon_name = ''
 * 
 *     return (             # <<<<<<<<<<<<<<
 *         _intenum_converter(af, socket_AddressFamily),
 *         _intenum_converter(type, socket_SocketKind),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "uvloop/dns.pyx":280
 * 
 *     return (
 *         _intenum_converter(af, socket_AddressFamily),             # <<<<<<<<<<<<<<
 *         _intenum_converter(type, socket_SocketKind),
 *         proto,
 */
  __pyx_t_1 = __pyx_v_6uvloop_4loop_socket_AddressFamily;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = __pyx_f_6uvloop_4loop__intenum_converter(__pyx_v_af, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":281
 *     return (
 *         _intenum_converter(af, socket_AddressFamily),
 *         _intenum_converter(type, socket_SocketKind),             # <<<<<<<<<<<<<<
 *         proto,
 *         canon_name,
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_v_6uvloop_4loop_socket_SocketKind;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = __pyx_f_6uvloop_4loop__intenum_converter(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uvloop/dns.pyx":282
 *         _intenum_converter(af, socket_AddressFamily),
 *         _intenum_converter(type, socket_SocketKind),
 *         proto,             # <<<<<<<<<<<<<<
 *         canon_name,
 *         pyaddr,
 */
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_proto); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uvloop/dns.pyx":280
 * 
 *     return (
 *         _intenum_converter(af, socket_AddressFamily),             # <<<<<<<<<<<<<<
 *         _intenum_converter(type, socket_SocketKind),
 *         proto,
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3)) __PYX_ERR(5, 280, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5)) __PYX_ERR(5, 280, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4)) __PYX_ERR(5, 280, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_canon_name);
  __Pyx_GIVEREF(__pyx_v_canon_name);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_canon_name)) __PYX_ERR(5, 280, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_pyaddr);
  __Pyx_GIVEREF(__pyx_v_pyaddr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_pyaddr)) __PYX_ERR(5, 280, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/dns.pyx":227
 * 
 * 
 * cdef __static_getaddrinfo_pyaddr(object host, object port,             # <<<<<<<<<<<<<<
 *                                  int family, int type,
 *                                  int proto, int flags):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.__static_getaddrinfo_pyaddr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_triplet);
  __Pyx_XDECREF(__pyx_v_af);
  __Pyx_XDECREF(__pyx_v_pyaddr);
  __Pyx_XDECREF(__pyx_v_canon_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":293
 *         system.addrinfo *data
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.data = NULL
 * 
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_8AddrInfo_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_8AddrInfo_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_8AddrInfo___cinit__(((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_8AddrInfo___cinit__(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self) {
  int __pyx_r;

  /* "uvloop/dns.pyx":294
 * 
 *     def __cinit__(self):
 *         self.data = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->data = NULL;

  /* "uvloop/dns.pyx":293
 *         system.addrinfo *data
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.data = NULL
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "uvloop/dns.pyx":296
 *         self.data = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.data is not NULL:
 *             uv.uv_freeaddrinfo(self.data)  # returns void
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_8AddrInfo_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_8AddrInfo_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_8AddrInfo_2__dealloc__(((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_8AddrInfo_2__dealloc__(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self) {
  int __pyx_t_1;

  /* "uvloop/dns.pyx":297
 * 
 *     def __dealloc__(self):
 *         if self.data is not NULL:             # <<<<<<<<<<<<<<
 *             uv.uv_freeaddrinfo(self.data)  # returns void
 *             self.data = NULL
 */
  __pyx_t_1 = (__pyx_v_self->data != NULL);
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":298
 *     def __dealloc__(self):
 *         if self.data is not NULL:
 *             uv.uv_freeaddrinfo(self.data)  # returns void             # <<<<<<<<<<<<<<
 *             self.data = NULL
 * 
 */
    uv_freeaddrinfo(__pyx_v_self->data);

    /* "uvloop/dns.pyx":299
 *         if self.data is not NULL:
 *             uv.uv_freeaddrinfo(self.data)  # returns void
 *             self.data = NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_data(self, system.addrinfo *data) noexcept:
 */
    __pyx_v_self->data = NULL;

    /* "uvloop/dns.pyx":297
 * 
 *     def __dealloc__(self):
 *         if self.data is not NULL:             # <<<<<<<<<<<<<<
 *             uv.uv_freeaddrinfo(self.data)  # returns void
 *             self.data = NULL
 */
  }

  /* "uvloop/dns.pyx":296
 *         self.data = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.data is not NULL:
 *             uv.uv_freeaddrinfo(self.data)  # returns void
 */

  /* function exit code */
}

/* "uvloop/dns.pyx":301
 *             self.data = NULL
 * 
 *     cdef void set_data(self, system.addrinfo *data) noexcept:             # <<<<<<<<<<<<<<
 *         self.data = data
 * 
 */

static void __pyx_f_6uvloop_4loop_8AddrInfo_set_data(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self, struct addrinfo *__pyx_v_data) {

  /* "uvloop/dns.pyx":302
 * 
 *     cdef void set_data(self, system.addrinfo *data) noexcept:
 *         self.data = data             # <<<<<<<<<<<<<<
 * 
 *     cdef unpack(self):
 */
  __pyx_v_self->data = __pyx_v_data;

  /* "uvloop/dns.pyx":301
 *             self.data = NULL
 * 
 *     cdef void set_data(self, system.addrinfo *data) noexcept:             # <<<<<<<<<<<<<<
 *         self.data = data
 * 
 */

  /* function exit code */
}

/* "uvloop/dns.pyx":304
 *         self.data = data
 * 
 *     cdef unpack(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list result = []
 */

static PyObject *__pyx_f_6uvloop_4loop_8AddrInfo_unpack(struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self) {
  PyObject *__pyx_v_result = 0;
  struct addrinfo *__pyx_v_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  struct addrinfo *__pyx_t_3;
  unsigned short __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("unpack", 1);

  /* "uvloop/dns.pyx":306
 *     cdef unpack(self):
 *         cdef:
 *             list result = []             # <<<<<<<<<<<<<<
 *             system.addrinfo *ptr
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":309
 *             system.addrinfo *ptr
 * 
 *         if self.data is NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('AddrInfo.data is NULL')
 * 
 */
  __pyx_t_2 = (__pyx_v_self->data == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/dns.pyx":310
 * 
 *         if self.data is NULL:
 *             raise RuntimeError('AddrInfo.data is NULL')             # <<<<<<<<<<<<<<
 * 
 *         ptr = self.data
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__183, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(5, 310, __pyx_L1_error)

    /* "uvloop/dns.pyx":309
 *             system.addrinfo *ptr
 * 
 *         if self.data is NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError('AddrInfo.data is NULL')
 * 
 */
  }

  /* "uvloop/dns.pyx":312
 *             raise RuntimeError('AddrInfo.data is NULL')
 * 
 *         ptr = self.data             # <<<<<<<<<<<<<<
 *         while ptr != NULL:
 *             if ptr.ai_addr.sa_family in (uv.AF_INET, uv.AF_INET6):
 */
  __pyx_t_3 = __pyx_v_self->data;
  __pyx_v_ptr = __pyx_t_3;

  /* "uvloop/dns.pyx":313
 * 
 *         ptr = self.data
 *         while ptr != NULL:             # <<<<<<<<<<<<<<
 *             if ptr.ai_addr.sa_family in (uv.AF_INET, uv.AF_INET6):
 *                 result.append((
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_ptr != NULL);
    if (!__pyx_t_2) break;

    /* "uvloop/dns.pyx":314
 *         ptr = self.data
 *         while ptr != NULL:
 *             if ptr.ai_addr.sa_family in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *                 result.append((
 *                     _intenum_converter(ptr.ai_family, socket_AddressFamily),
 */
    __pyx_t_4 = __pyx_v_ptr->ai_addr->sa_family;
    __pyx_t_5 = (__pyx_t_4 == AF_INET);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_t_4 == AF_INET6);
    __pyx_t_2 = __pyx_t_5;
    __pyx_L7_bool_binop_done:;
    __pyx_t_5 = __pyx_t_2;
    if (__pyx_t_5) {

      /* "uvloop/dns.pyx":316
 *             if ptr.ai_addr.sa_family in (uv.AF_INET, uv.AF_INET6):
 *                 result.append((
 *                     _intenum_converter(ptr.ai_family, socket_AddressFamily),             # <<<<<<<<<<<<<<
 *                     _intenum_converter(ptr.ai_socktype, socket_SocketKind),
 *                     ptr.ai_protocol,
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ptr->ai_family); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __pyx_v_6uvloop_4loop_socket_AddressFamily;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = __pyx_f_6uvloop_4loop__intenum_converter(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/dns.pyx":317
 *                 result.append((
 *                     _intenum_converter(ptr.ai_family, socket_AddressFamily),
 *                     _intenum_converter(ptr.ai_socktype, socket_SocketKind),             # <<<<<<<<<<<<<<
 *                     ptr.ai_protocol,
 *                     ('' if ptr.ai_canonname is NULL else
 */
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_ptr->ai_socktype); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __pyx_v_6uvloop_4loop_socket_SocketKind;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = __pyx_f_6uvloop_4loop__intenum_converter(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 317, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/dns.pyx":318
 *                     _intenum_converter(ptr.ai_family, socket_AddressFamily),
 *                     _intenum_converter(ptr.ai_socktype, socket_SocketKind),
 *                     ptr.ai_protocol,             # <<<<<<<<<<<<<<
 *                     ('' if ptr.ai_canonname is NULL else
 *                         (<bytes>ptr.ai_canonname).decode()),
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ptr->ai_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "uvloop/dns.pyx":319
 *                     _intenum_converter(ptr.ai_socktype, socket_SocketKind),
 *                     ptr.ai_protocol,
 *                     ('' if ptr.ai_canonname is NULL else             # <<<<<<<<<<<<<<
 *                         (<bytes>ptr.ai_canonname).decode()),
 *                     __convert_sockaddr_to_pyaddr(ptr.ai_addr)
 */
      __pyx_t_5 = (__pyx_v_ptr->ai_canonname == NULL);
      if (__pyx_t_5) {
        __Pyx_INCREF(__pyx_kp_u__10);
        __pyx_t_6 = __pyx_kp_u__10;
      } else {

        /* "uvloop/dns.pyx":320
 *                     ptr.ai_protocol,
 *                     ('' if ptr.ai_canonname is NULL else
 *                         (<bytes>ptr.ai_canonname).decode()),             # <<<<<<<<<<<<<<
 *                     __convert_sockaddr_to_pyaddr(ptr.ai_addr)
 *                 ))
 */
        __pyx_t_9 = __Pyx_PyBytes_FromString(__pyx_v_ptr->ai_canonname); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (unlikely(__pyx_t_9 == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
          __PYX_ERR(5, 320, __pyx_L1_error)
        }
        __pyx_t_10 = __Pyx_decode_bytes(((PyObject*)__pyx_t_9), 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 320, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_6 = __pyx_t_10;
        __pyx_t_10 = 0;
      }

      /* "uvloop/dns.pyx":321
 *                     ('' if ptr.ai_canonname is NULL else
 *                         (<bytes>ptr.ai_canonname).decode()),
 *                     __convert_sockaddr_to_pyaddr(ptr.ai_addr)             # <<<<<<<<<<<<<<
 *                 ))
 * 
 */
      __pyx_t_10 = __pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(__pyx_v_ptr->ai_addr); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 321, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "uvloop/dns.pyx":316
 *             if ptr.ai_addr.sa_family in (uv.AF_INET, uv.AF_INET6):
 *                 result.append((
 *                     _intenum_converter(ptr.ai_family, socket_AddressFamily),             # <<<<<<<<<<<<<<
 *                     _intenum_converter(ptr.ai_socktype, socket_SocketKind),
 *                     ptr.ai_protocol,
 */
      __pyx_t_9 = PyTuple_New(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_7);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7)) __PYX_ERR(5, 316, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_8);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_8)) __PYX_ERR(5, 316, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_1);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_1)) __PYX_ERR(5, 316, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_6);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_6)) __PYX_ERR(5, 316, __pyx_L1_error);
      __Pyx_GIVEREF(__pyx_t_10);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_t_10)) __PYX_ERR(5, 316, __pyx_L1_error);
      __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      __pyx_t_1 = 0;
      __pyx_t_6 = 0;
      __pyx_t_10 = 0;

      /* "uvloop/dns.pyx":315
 *         while ptr != NULL:
 *             if ptr.ai_addr.sa_family in (uv.AF_INET, uv.AF_INET6):
 *                 result.append((             # <<<<<<<<<<<<<<
 *                     _intenum_converter(ptr.ai_family, socket_AddressFamily),
 *                     _intenum_converter(ptr.ai_socktype, socket_SocketKind),
 */
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_9); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(5, 315, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/dns.pyx":314
 *         ptr = self.data
 *         while ptr != NULL:
 *             if ptr.ai_addr.sa_family in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *                 result.append((
 *                     _intenum_converter(ptr.ai_family, socket_AddressFamily),
 */
    }

    /* "uvloop/dns.pyx":324
 *                 ))
 * 
 *             ptr = ptr.ai_next             # <<<<<<<<<<<<<<
 * 
 *         return result
 */
    __pyx_t_3 = __pyx_v_ptr->ai_next;
    __pyx_v_ptr = __pyx_t_3;
  }

  /* "uvloop/dns.pyx":326
 *             ptr = ptr.ai_next
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "uvloop/dns.pyx":304
 *         self.data = data
 * 
 *     cdef unpack(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list result = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uvloop.loop.AddrInfo.unpack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":329
 * 
 *     @staticmethod
 *     cdef int isinstance(object other):             # <<<<<<<<<<<<<<
 *         return type(other) is AddrInfo
 * 
 */

static int __pyx_f_6uvloop_4loop_8AddrInfo_isinstance(PyObject *__pyx_v_other) {
  int __pyx_r;
  int __pyx_t_1;

  /* "uvloop/dns.pyx":330
 *     @staticmethod
 *     cdef int isinstance(object other):
 *         return type(other) is AddrInfo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_other)) == ((PyObject *)__pyx_ptype_6uvloop_4loop_AddrInfo));
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "uvloop/dns.pyx":329
 * 
 *     @staticmethod
 *     cdef int isinstance(object other):             # <<<<<<<<<<<<<<
 *         return type(other) is AddrInfo
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8AddrInfo_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8AddrInfo_4__reduce_cython__, "AddrInfo.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8AddrInfo_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8AddrInfo_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8AddrInfo_4__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8AddrInfo_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_8AddrInfo_4__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8AddrInfo_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.AddrInfo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_8AddrInfo_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_8AddrInfo_6__setstate_cython__, "AddrInfo.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_8AddrInfo_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8AddrInfo_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8AddrInfo_6__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_8AddrInfo_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.AddrInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_8AddrInfo_6__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_8AddrInfo_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.AddrInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":339
 *         uv.uv_getaddrinfo_t _req_data
 * 
 *     def __cinit__(self, Loop loop,             # <<<<<<<<<<<<<<
 *                   bytes host, bytes port,
 *                   int family, int type, int proto, int flags,
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_15AddrInfoRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_15AddrInfoRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_host = 0;
  PyObject *__pyx_v_port = 0;
  int __pyx_v_family;
  int __pyx_v_type;
  int __pyx_v_proto;
  int __pyx_v_flags;
  PyObject *__pyx_v_callback = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,&__pyx_n_s_host,&__pyx_n_s_port,&__pyx_n_s_family,&__pyx_n_s_type,&__pyx_n_s_proto,&__pyx_n_s_flags,&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  8: values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_host)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 1); __PYX_ERR(5, 339, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_port)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 2); __PYX_ERR(5, 339, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_family)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 3); __PYX_ERR(5, 339, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 4); __PYX_ERR(5, 339, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proto)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[5]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 5); __PYX_ERR(5, 339, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_flags)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[6]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 6); __PYX_ERR(5, 339, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[7]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 339, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 7); __PYX_ERR(5, 339, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(5, 339, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 8)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
      values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
      values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
      values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
    }
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)values[0]);
    __pyx_v_host = ((PyObject*)values[1]);
    __pyx_v_port = ((PyObject*)values[2]);
    __pyx_v_family = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_family == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 341, __pyx_L3_error)
    __pyx_v_type = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 341, __pyx_L3_error)
    __pyx_v_proto = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_proto == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 341, __pyx_L3_error)
    __pyx_v_flags = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 341, __pyx_L3_error)
    __pyx_v_callback = values[7];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, __pyx_nargs); __PYX_ERR(5, 339, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.AddrInfoRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_loop), __pyx_ptype_6uvloop_4loop_Loop, 1, "loop", 0))) __PYX_ERR(5, 339, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_host), (&PyBytes_Type), 1, "host", 1))) __PYX_ERR(5, 340, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_port), (&PyBytes_Type), 1, "port", 1))) __PYX_ERR(5, 340, __pyx_L1_error)
  __pyx_r = __pyx_pf_6uvloop_4loop_15AddrInfoRequest___cinit__(((struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)__pyx_v_self), __pyx_v_loop, __pyx_v_host, __pyx_v_port, __pyx_v_family, __pyx_v_type, __pyx_v_proto, __pyx_v_flags, __pyx_v_callback);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_15AddrInfoRequest___cinit__(struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_host, PyObject *__pyx_v_port, int __pyx_v_family, int __pyx_v_type, int __pyx_v_proto, int __pyx_v_flags, PyObject *__pyx_v_callback) {
  int __pyx_v_err;
  char *__pyx_v_chost;
  char *__pyx_v_cport;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_v_ex = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  char const *__pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  unsigned int __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/dns.pyx":349
 *             char *cport
 * 
 *         if host is None:             # <<<<<<<<<<<<<<
 *             chost = NULL
 *         elif host == b'' and sys.platform == 'darwin':
 */
  __pyx_t_1 = (__pyx_v_host == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":350
 * 
 *         if host is None:
 *             chost = NULL             # <<<<<<<<<<<<<<
 *         elif host == b'' and sys.platform == 'darwin':
 *             # It seems `getaddrinfo("", ...)` on macOS is equivalent to
 */
    __pyx_v_chost = NULL;

    /* "uvloop/dns.pyx":349
 *             char *cport
 * 
 *         if host is None:             # <<<<<<<<<<<<<<
 *             chost = NULL
 *         elif host == b'' and sys.platform == 'darwin':
 */
    goto __pyx_L3;
  }

  /* "uvloop/dns.pyx":351
 *         if host is None:
 *             chost = NULL
 *         elif host == b'' and sys.platform == 'darwin':             # <<<<<<<<<<<<<<
 *             # It seems `getaddrinfo("", ...)` on macOS is equivalent to
 *             # `getaddrinfo("localhost", ...)`. This is inconsistent with
 */
  __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_v_host, __pyx_kp_b__10, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 351, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_platform); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_n_u_darwin, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(5, 351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":355
 *             # `getaddrinfo("localhost", ...)`. This is inconsistent with
 *             # libuv 1.48 which treats empty nodename as EINVAL.
 *             chost = <char*>'localhost'             # <<<<<<<<<<<<<<
 *         else:
 *             chost = <char*>host
 */
    __pyx_v_chost = ((char *)((char *)"localhost"));

    /* "uvloop/dns.pyx":351
 *         if host is None:
 *             chost = NULL
 *         elif host == b'' and sys.platform == 'darwin':             # <<<<<<<<<<<<<<
 *             # It seems `getaddrinfo("", ...)` on macOS is equivalent to
 *             # `getaddrinfo("localhost", ...)`. This is inconsistent with
 */
    goto __pyx_L3;
  }

  /* "uvloop/dns.pyx":357
 *             chost = <char*>'localhost'
 *         else:
 *             chost = <char*>host             # <<<<<<<<<<<<<<
 * 
 *         if port is None:
 */
  /*else*/ {
    if (unlikely(__pyx_v_host == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(5, 357, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_host); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(5, 357, __pyx_L1_error)
    __pyx_v_chost = ((char *)__pyx_t_5);
  }
  __pyx_L3:;

  /* "uvloop/dns.pyx":359
 *             chost = <char*>host
 * 
 *         if port is None:             # <<<<<<<<<<<<<<
 *             cport = NULL
 *         else:
 */
  __pyx_t_1 = (__pyx_v_port == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":360
 * 
 *         if port is None:
 *             cport = NULL             # <<<<<<<<<<<<<<
 *         else:
 *             cport = <char*>port
 */
    __pyx_v_cport = NULL;

    /* "uvloop/dns.pyx":359
 *             chost = <char*>host
 * 
 *         if port is None:             # <<<<<<<<<<<<<<
 *             cport = NULL
 *         else:
 */
    goto __pyx_L6;
  }

  /* "uvloop/dns.pyx":362
 *             cport = NULL
 *         else:
 *             cport = <char*>port             # <<<<<<<<<<<<<<
 * 
 *         memset(&self.hints, 0, sizeof(system.addrinfo))
 */
  /*else*/ {
    if (unlikely(__pyx_v_port == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(5, 362, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_port); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(5, 362, __pyx_L1_error)
    __pyx_v_cport = ((char *)__pyx_t_5);
  }
  __pyx_L6:;

  /* "uvloop/dns.pyx":364
 *             cport = <char*>port
 * 
 *         memset(&self.hints, 0, sizeof(system.addrinfo))             # <<<<<<<<<<<<<<
 *         self.hints.ai_flags = flags
 *         self.hints.ai_family = family
 */
  (void)(memset((&__pyx_v_self->hints), 0, (sizeof(struct addrinfo))));

  /* "uvloop/dns.pyx":365
 * 
 *         memset(&self.hints, 0, sizeof(system.addrinfo))
 *         self.hints.ai_flags = flags             # <<<<<<<<<<<<<<
 *         self.hints.ai_family = family
 *         self.hints.ai_socktype = type
 */
  __pyx_v_self->hints.ai_flags = __pyx_v_flags;

  /* "uvloop/dns.pyx":366
 *         memset(&self.hints, 0, sizeof(system.addrinfo))
 *         self.hints.ai_flags = flags
 *         self.hints.ai_family = family             # <<<<<<<<<<<<<<
 *         self.hints.ai_socktype = type
 *         self.hints.ai_protocol = proto
 */
  __pyx_v_self->hints.ai_family = __pyx_v_family;

  /* "uvloop/dns.pyx":367
 *         self.hints.ai_flags = flags
 *         self.hints.ai_family = family
 *         self.hints.ai_socktype = type             # <<<<<<<<<<<<<<
 *         self.hints.ai_protocol = proto
 * 
 */
  __pyx_v_self->hints.ai_socktype = __pyx_v_type;

  /* "uvloop/dns.pyx":368
 *         self.hints.ai_family = family
 *         self.hints.ai_socktype = type
 *         self.hints.ai_protocol = proto             # <<<<<<<<<<<<<<
 * 
 *         self.request = <uv.uv_req_t*> &self._req_data
 */
  __pyx_v_self->hints.ai_protocol = __pyx_v_proto;

  /* "uvloop/dns.pyx":370
 *         self.hints.ai_protocol = proto
 * 
 *         self.request = <uv.uv_req_t*> &self._req_data             # <<<<<<<<<<<<<<
 *         self.callback = callback
 *         self.request.data = <void*>self
 */
  __pyx_v_self->__pyx_base.request = ((uv_req_t *)(&__pyx_v_self->_req_data));

  /* "uvloop/dns.pyx":371
 * 
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.callback = callback             # <<<<<<<<<<<<<<
 *         self.request.data = <void*>self
 * 
 */
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  __Pyx_GOTREF(__pyx_v_self->callback);
  __Pyx_DECREF(__pyx_v_self->callback);
  __pyx_v_self->callback = __pyx_v_callback;

  /* "uvloop/dns.pyx":372
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.callback = callback
 *         self.request.data = <void*>self             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_getaddrinfo(loop.uvloop,
 */
  __pyx_v_self->__pyx_base.request->data = ((void *)__pyx_v_self);

  /* "uvloop/dns.pyx":374
 *         self.request.data = <void*>self
 * 
 *         err = uv.uv_getaddrinfo(loop.uvloop,             # <<<<<<<<<<<<<<
 *                                 <uv.uv_getaddrinfo_t*>self.request,
 *                                 __on_addrinfo_resolved,
 */
  __pyx_v_err = uv_getaddrinfo(__pyx_v_loop->uvloop, ((uv_getaddrinfo_t *)__pyx_v_self->__pyx_base.request), __pyx_f_6uvloop_4loop___on_addrinfo_resolved, __pyx_v_chost, __pyx_v_cport, (&__pyx_v_self->hints));

  /* "uvloop/dns.pyx":381
 *                                 &self.hints)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self.on_done()
 *             try:
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":382
 * 
 *         if err < 0:
 *             self.on_done()             # <<<<<<<<<<<<<<
 *             try:
 *                 if err == uv.UV_EINVAL:
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_AddrInfoRequest *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/dns.pyx":383
 *         if err < 0:
 *             self.on_done()
 *             try:             # <<<<<<<<<<<<<<
 *                 if err == uv.UV_EINVAL:
 *                     # Convert UV_EINVAL to EAI_NONAME to match libc behavior
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "uvloop/dns.pyx":384
 *             self.on_done()
 *             try:
 *                 if err == uv.UV_EINVAL:             # <<<<<<<<<<<<<<
 *                     # Convert UV_EINVAL to EAI_NONAME to match libc behavior
 *                     msg = system.gai_strerror(socket_EAI_NONAME).decode('utf-8')
 */
        __pyx_t_1 = (__pyx_v_err == UV_EINVAL);
        if (__pyx_t_1) {

          /* "uvloop/dns.pyx":386
 *                 if err == uv.UV_EINVAL:
 *                     # Convert UV_EINVAL to EAI_NONAME to match libc behavior
 *                     msg = system.gai_strerror(socket_EAI_NONAME).decode('utf-8')             # <<<<<<<<<<<<<<
 *                     ex = socket_gaierror(socket_EAI_NONAME, msg)
 *                 else:
 */
          __pyx_t_9 = gai_strerror(__pyx_v_6uvloop_4loop_socket_EAI_NONAME);
          __pyx_t_10 = __Pyx_ssize_strlen(__pyx_t_9); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(5, 386, __pyx_L8_error)
          __pyx_t_4 = __Pyx_decode_c_string(__pyx_t_9, 0, __pyx_t_10, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 386, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_4);
          __pyx_v_msg = __pyx_t_4;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "uvloop/dns.pyx":387
 *                     # Convert UV_EINVAL to EAI_NONAME to match libc behavior
 *                     msg = system.gai_strerror(socket_EAI_NONAME).decode('utf-8')
 *                     ex = socket_gaierror(socket_EAI_NONAME, msg)             # <<<<<<<<<<<<<<
 *                 else:
 *                     ex = convert_error(err)
 */
          __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_socket_EAI_NONAME); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 387, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_v_6uvloop_4loop_socket_gaierror);
          __pyx_t_11 = __pyx_v_6uvloop_4loop_socket_gaierror; __pyx_t_12 = NULL;
          __pyx_t_13 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_13 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_t_3, __pyx_v_msg};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_13, 2+__pyx_t_13);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 387, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __pyx_v_ex = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "uvloop/dns.pyx":384
 *             self.on_done()
 *             try:
 *                 if err == uv.UV_EINVAL:             # <<<<<<<<<<<<<<
 *                     # Convert UV_EINVAL to EAI_NONAME to match libc behavior
 *                     msg = system.gai_strerror(socket_EAI_NONAME).decode('utf-8')
 */
          goto __pyx_L14;
        }

        /* "uvloop/dns.pyx":389
 *                     ex = socket_gaierror(socket_EAI_NONAME, msg)
 *                 else:
 *                     ex = convert_error(err)             # <<<<<<<<<<<<<<
 *             except Exception as ex:
 *                 callback(ex)
 */
        /*else*/ {
          __pyx_t_4 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 389, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_v_ex = __pyx_t_4;
          __pyx_t_4 = 0;
        }
        __pyx_L14:;

        /* "uvloop/dns.pyx":383
 *         if err < 0:
 *             self.on_done()
 *             try:             # <<<<<<<<<<<<<<
 *                 if err == uv.UV_EINVAL:
 *                     # Convert UV_EINVAL to EAI_NONAME to match libc behavior
 */
      }

      /* "uvloop/dns.pyx":393
 *                 callback(ex)
 *             else:
 *                 callback(ex)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      /*else:*/ {
        __Pyx_INCREF(__pyx_v_callback);
        __pyx_t_11 = __pyx_v_callback; __pyx_t_3 = NULL;
        __pyx_t_13 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_13 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_ex};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_13, 1+__pyx_t_13);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 393, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L13_try_end;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/dns.pyx":390
 *                 else:
 *                     ex = convert_error(err)
 *             except Exception as ex:             # <<<<<<<<<<<<<<
 *                 callback(ex)
 *             else:
 */
      __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_14) {
        __Pyx_AddTraceback("uvloop.loop.AddrInfoRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_11, &__pyx_t_3) < 0) __PYX_ERR(5, 390, __pyx_L10_except_error)
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_XDECREF_SET(__pyx_v_ex, __pyx_t_11);
        /*try:*/ {

          /* "uvloop/dns.pyx":391
 *                     ex = convert_error(err)
 *             except Exception as ex:
 *                 callback(ex)             # <<<<<<<<<<<<<<
 *             else:
 *                 callback(ex)
 */
          __Pyx_INCREF(__pyx_v_callback);
          __pyx_t_15 = __pyx_v_callback; __pyx_t_16 = NULL;
          __pyx_t_13 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_15))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_15, function);
              __pyx_t_13 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_v_ex};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_15, __pyx_callargs+1-__pyx_t_13, 1+__pyx_t_13);
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(5, 391, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }

        /* "uvloop/dns.pyx":390
 *                 else:
 *                     ex = convert_error(err)
 *             except Exception as ex:             # <<<<<<<<<<<<<<
 *                 callback(ex)
 *             else:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L21;
          }
          __pyx_L20_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_24);
            __pyx_t_14 = __pyx_lineno; __pyx_t_17 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_XGIVEREF(__pyx_t_24);
              __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
            }
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ErrRestore(__pyx_t_19, __pyx_t_20, __pyx_t_21);
            __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
            __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_17; __pyx_filename = __pyx_t_18;
            goto __pyx_L10_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L9_exception_handled;
      }
      goto __pyx_L10_except_error;

      /* "uvloop/dns.pyx":383
 *         if err < 0:
 *             self.on_done()
 *             try:             # <<<<<<<<<<<<<<
 *                 if err == uv.UV_EINVAL:
 *                     # Convert UV_EINVAL to EAI_NONAME to match libc behavior
 */
      __pyx_L10_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L9_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L13_try_end:;
    }

    /* "uvloop/dns.pyx":381
 *                                 &self.hints)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self.on_done()
 *             try:
 */
  }

  /* "uvloop/dns.pyx":339
 *         uv.uv_getaddrinfo_t _req_data
 * 
 *     def __cinit__(self, Loop loop,             # <<<<<<<<<<<<<<
 *                   bytes host, bytes port,
 *                   int family, int type, int proto, int flags,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("uvloop.loop.AddrInfoRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15AddrInfoRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15AddrInfoRequest_2__reduce_cython__, "AddrInfoRequest.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15AddrInfoRequest_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15AddrInfoRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15AddrInfoRequest_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15AddrInfoRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15AddrInfoRequest_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15AddrInfoRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.AddrInfoRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15AddrInfoRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15AddrInfoRequest_4__setstate_cython__, "AddrInfoRequest.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15AddrInfoRequest_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15AddrInfoRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15AddrInfoRequest_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15AddrInfoRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.AddrInfoRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15AddrInfoRequest_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15AddrInfoRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.AddrInfoRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":401
 *         uv.uv_getnameinfo_t _req_data
 * 
 *     def __cinit__(self, Loop loop, callback):             # <<<<<<<<<<<<<<
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.callback = callback
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_15NameInfoRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_15NameInfoRequest_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_callback = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 401, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 401, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(5, 401, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(5, 401, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)values[0]);
    __pyx_v_callback = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, __pyx_nargs); __PYX_ERR(5, 401, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.NameInfoRequest.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_loop), __pyx_ptype_6uvloop_4loop_Loop, 1, "loop", 0))) __PYX_ERR(5, 401, __pyx_L1_error)
  __pyx_r = __pyx_pf_6uvloop_4loop_15NameInfoRequest___cinit__(((struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)__pyx_v_self), __pyx_v_loop, __pyx_v_callback);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_15NameInfoRequest___cinit__(struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_callback) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/dns.pyx":402
 * 
 *     def __cinit__(self, Loop loop, callback):
 *         self.request = <uv.uv_req_t*> &self._req_data             # <<<<<<<<<<<<<<
 *         self.callback = callback
 *         self.request.data = <void*>self
 */
  __pyx_v_self->__pyx_base.request = ((uv_req_t *)(&__pyx_v_self->_req_data));

  /* "uvloop/dns.pyx":403
 *     def __cinit__(self, Loop loop, callback):
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.callback = callback             # <<<<<<<<<<<<<<
 *         self.request.data = <void*>self
 * 
 */
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  __Pyx_GOTREF(__pyx_v_self->callback);
  __Pyx_DECREF(__pyx_v_self->callback);
  __pyx_v_self->callback = __pyx_v_callback;

  /* "uvloop/dns.pyx":404
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.callback = callback
 *         self.request.data = <void*>self             # <<<<<<<<<<<<<<
 * 
 *     cdef query(self, system.sockaddr *addr, int flags):
 */
  __pyx_v_self->__pyx_base.request->data = ((void *)__pyx_v_self);

  /* "uvloop/dns.pyx":401
 *         uv.uv_getnameinfo_t _req_data
 * 
 *     def __cinit__(self, Loop loop, callback):             # <<<<<<<<<<<<<<
 *         self.request = <uv.uv_req_t*> &self._req_data
 *         self.callback = callback
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":406
 *         self.request.data = <void*>self
 * 
 *     cdef query(self, system.sockaddr *addr, int flags):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         err = uv.uv_getnameinfo(self.loop.uvloop,
 */

static PyObject *__pyx_f_6uvloop_4loop_15NameInfoRequest_query(struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self, struct sockaddr *__pyx_v_addr, int __pyx_v_flags) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query", 1);

  /* "uvloop/dns.pyx":408
 *     cdef query(self, system.sockaddr *addr, int flags):
 *         cdef int err
 *         err = uv.uv_getnameinfo(self.loop.uvloop,             # <<<<<<<<<<<<<<
 *                                 <uv.uv_getnameinfo_t*>self.request,
 *                                 __on_nameinfo_resolved,
 */
  __pyx_v_err = uv_getnameinfo(__pyx_v_self->__pyx_base.loop->uvloop, ((uv_getnameinfo_t *)__pyx_v_self->__pyx_base.request), __pyx_f_6uvloop_4loop___on_nameinfo_resolved, __pyx_v_addr, __pyx_v_flags);

  /* "uvloop/dns.pyx":413
 *                                 addr,
 *                                 flags)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self.on_done()
 *             self.callback(convert_error(err))
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":414
 *                                 flags)
 *         if err < 0:
 *             self.on_done()             # <<<<<<<<<<<<<<
 *             self.callback(convert_error(err))
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_NameInfoRequest *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/dns.pyx":415
 *         if err < 0:
 *             self.on_done()
 *             self.callback(convert_error(err))             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->callback);
    __pyx_t_4 = __pyx_v_self->callback; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/dns.pyx":413
 *                                 addr,
 *                                 flags)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self.on_done()
 *             self.callback(convert_error(err))
 */
  }

  /* "uvloop/dns.pyx":406
 *         self.request.data = <void*>self
 * 
 *     cdef query(self, system.sockaddr *addr, int flags):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         err = uv.uv_getnameinfo(self.loop.uvloop,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.NameInfoRequest.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15NameInfoRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15NameInfoRequest_2__reduce_cython__, "NameInfoRequest.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15NameInfoRequest_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15NameInfoRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15NameInfoRequest_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15NameInfoRequest_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15NameInfoRequest_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15NameInfoRequest_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.NameInfoRequest.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15NameInfoRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15NameInfoRequest_4__setstate_cython__, "NameInfoRequest.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15NameInfoRequest_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15NameInfoRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15NameInfoRequest_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15NameInfoRequest_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.NameInfoRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15NameInfoRequest_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15NameInfoRequest_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.NameInfoRequest.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":418
 * 
 * 
 * cdef _intenum_converter(value, enum_klass):             # <<<<<<<<<<<<<<
 *     try:
 *         return enum_klass(value)
 */

static PyObject *__pyx_f_6uvloop_4loop__intenum_converter(PyObject *__pyx_v_value, PyObject *__pyx_v_enum_klass) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_intenum_converter", 1);

  /* "uvloop/dns.pyx":419
 * 
 * cdef _intenum_converter(value, enum_klass):
 *     try:             # <<<<<<<<<<<<<<
 *         return enum_klass(value)
 *     except ValueError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/dns.pyx":420
 * cdef _intenum_converter(value, enum_klass):
 *     try:
 *         return enum_klass(value)             # <<<<<<<<<<<<<<
 *     except ValueError:
 *         return value
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_enum_klass);
      __pyx_t_5 = __pyx_v_enum_klass; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_value};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 420, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "uvloop/dns.pyx":419
 * 
 * cdef _intenum_converter(value, enum_klass):
 *     try:             # <<<<<<<<<<<<<<
 *         return enum_klass(value)
 *     except ValueError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/dns.pyx":421
 *     try:
 *         return enum_klass(value)
 *     except ValueError:             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop._intenum_converter", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(5, 421, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "uvloop/dns.pyx":422
 *         return enum_klass(value)
 *     except ValueError:
 *         return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_value);
      __pyx_r = __pyx_v_value;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/dns.pyx":419
 * 
 * cdef _intenum_converter(value, enum_klass):
 *     try:             # <<<<<<<<<<<<<<
 *         return enum_klass(value)
 *     except ValueError:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "uvloop/dns.pyx":418
 * 
 * 
 * cdef _intenum_converter(value, enum_klass):             # <<<<<<<<<<<<<<
 *     try:
 *         return enum_klass(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop._intenum_converter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/dns.pyx":425
 * 
 * 
 * cdef void __on_addrinfo_resolved(             # <<<<<<<<<<<<<<
 *     uv.uv_getaddrinfo_t *resolver,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___on_addrinfo_resolved(uv_getaddrinfo_t *__pyx_v_resolver, int __pyx_v_status, struct addrinfo *__pyx_v_res) {
  struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *__pyx_v_request = 0;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_callback = 0;
  struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_v_ai = 0;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  char const *__pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__on_addrinfo_resolved", 0);

  /* "uvloop/dns.pyx":431
 * ) noexcept with gil:
 * 
 *     if resolver.data is NULL:             # <<<<<<<<<<<<<<
 *         aio_logger.error(
 *             'AddrInfoRequest callback called with NULL resolver.data')
 */
  __pyx_t_1 = (__pyx_v_resolver->data == NULL);
  if (__pyx_t_1) {

    /* "uvloop/dns.pyx":432
 * 
 *     if resolver.data is NULL:
 *         aio_logger.error(             # <<<<<<<<<<<<<<
 *             'AddrInfoRequest callback called with NULL resolver.data')
 *         return
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_AddrInfoRequest_callback_called};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/dns.pyx":434
 *         aio_logger.error(
 *             'AddrInfoRequest callback called with NULL resolver.data')
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/dns.pyx":431
 * ) noexcept with gil:
 * 
 *     if resolver.data is NULL:             # <<<<<<<<<<<<<<
 *         aio_logger.error(
 *             'AddrInfoRequest callback called with NULL resolver.data')
 */
  }

  /* "uvloop/dns.pyx":437
 * 
 *     cdef:
 *         AddrInfoRequest request = <AddrInfoRequest> resolver.data             # <<<<<<<<<<<<<<
 *         Loop loop = request.loop
 *         object callback = request.callback
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_resolver->data);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_request = ((struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/dns.pyx":438
 *     cdef:
 *         AddrInfoRequest request = <AddrInfoRequest> resolver.data
 *         Loop loop = request.loop             # <<<<<<<<<<<<<<
 *         object callback = request.callback
 *         AddrInfo ai
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_request->__pyx_base.loop);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/dns.pyx":439
 *         AddrInfoRequest request = <AddrInfoRequest> resolver.data
 *         Loop loop = request.loop
 *         object callback = request.callback             # <<<<<<<<<<<<<<
 *         AddrInfo ai
 * 
 */
  __pyx_t_2 = __pyx_v_request->callback;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_callback = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/dns.pyx":442
 *         AddrInfo ai
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if status < 0:
 *             callback(convert_error(status))
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "uvloop/dns.pyx":443
 * 
 *     try:
 *         if status < 0:             # <<<<<<<<<<<<<<
 *             callback(convert_error(status))
 *         else:
 */
        __pyx_t_1 = (__pyx_v_status < 0);
        if (__pyx_t_1) {

          /* "uvloop/dns.pyx":444
 *     try:
 *         if status < 0:
 *             callback(convert_error(status))             # <<<<<<<<<<<<<<
 *         else:
 *             ai = AddrInfo()
 */
          __pyx_t_3 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 444, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_INCREF(__pyx_v_callback);
          __pyx_t_4 = __pyx_v_callback; __pyx_t_9 = NULL;
          __pyx_t_5 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_5 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_3};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 444, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "uvloop/dns.pyx":443
 * 
 *     try:
 *         if status < 0:             # <<<<<<<<<<<<<<
 *             callback(convert_error(status))
 *         else:
 */
          goto __pyx_L13;
        }

        /* "uvloop/dns.pyx":446
 *             callback(convert_error(status))
 *         else:
 *             ai = AddrInfo()             # <<<<<<<<<<<<<<
 *             ai.set_data(res)
 *             callback(ai)
 */
        /*else*/ {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6uvloop_4loop_AddrInfo)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 446, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_v_ai = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)__pyx_t_2);
          __pyx_t_2 = 0;

          /* "uvloop/dns.pyx":447
 *         else:
 *             ai = AddrInfo()
 *             ai.set_data(res)             # <<<<<<<<<<<<<<
 *             callback(ai)
 *     except (KeyboardInterrupt, SystemExit):
 */
          ((struct __pyx_vtabstruct_6uvloop_4loop_AddrInfo *)__pyx_v_ai->__pyx_vtab)->set_data(__pyx_v_ai, __pyx_v_res);

          /* "uvloop/dns.pyx":448
 *             ai = AddrInfo()
 *             ai.set_data(res)
 *             callback(ai)             # <<<<<<<<<<<<<<
 *     except (KeyboardInterrupt, SystemExit):
 *         raise
 */
          __Pyx_INCREF(__pyx_v_callback);
          __pyx_t_4 = __pyx_v_callback; __pyx_t_3 = NULL;
          __pyx_t_5 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_5 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_ai)};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 448, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __pyx_L13:;

        /* "uvloop/dns.pyx":442
 *         AddrInfo ai
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if status < 0:
 *             callback(convert_error(status))
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L12_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/dns.pyx":449
 *             ai.set_data(res)
 *             callback(ai)
 *     except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *         raise
 *     except BaseException as ex:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.__on_addrinfo_resolved", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(5, 449, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);

        /* "uvloop/dns.pyx":450
 *             callback(ai)
 *     except (KeyboardInterrupt, SystemExit):
 *         raise             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         loop._handle_exception(ex)
 */
        __Pyx_GIVEREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ErrRestoreWithState(__pyx_t_2, __pyx_t_4, __pyx_t_3);
        __pyx_t_2 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; 
        __PYX_ERR(5, 450, __pyx_L9_except_error)
      }

      /* "uvloop/dns.pyx":451
 *     except (KeyboardInterrupt, SystemExit):
 *         raise
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         loop._handle_exception(ex)
 *     finally:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.__on_addrinfo_resolved", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_2) < 0) __PYX_ERR(5, 451, __pyx_L9_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_ex = __pyx_t_4;
        /*try:*/ {

          /* "uvloop/dns.pyx":452
 *         raise
 *     except BaseException as ex:
 *         loop._handle_exception(ex)             # <<<<<<<<<<<<<<
 *     finally:
 *         request.on_done()
 */
          ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_loop->__pyx_vtab)->_handle_exception(__pyx_v_loop, __pyx_v_ex); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 452, __pyx_L21_error)
        }

        /* "uvloop/dns.pyx":451
 *     except (KeyboardInterrupt, SystemExit):
 *         raise
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         loop._handle_exception(ex)
 *     finally:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L22;
          }
          __pyx_L21_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __pyx_t_10 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            }
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
            goto __pyx_L9_except_error;
          }
          __pyx_L22:;
        }
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;

      /* "uvloop/dns.pyx":442
 *         AddrInfo ai
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if status < 0:
 *             callback(convert_error(status))
 */
      __pyx_L9_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L5_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L12_try_end:;
    }
  }

  /* "uvloop/dns.pyx":454
 *         loop._handle_exception(ex)
 *     finally:
 *         request.on_done()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_AddrInfoRequest *)__pyx_v_request->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_request)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 454, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_16);
      __pyx_t_11 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
      {
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_AddrInfoRequest *)__pyx_v_request->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_request)); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 454, __pyx_L28_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_17, __pyx_t_16);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_19;
      goto __pyx_L1_error;
      __pyx_L28_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_17, __pyx_t_16);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/dns.pyx":425
 * 
 * 
 * cdef void __on_addrinfo_resolved(             # <<<<<<<<<<<<<<
 *     uv.uv_getaddrinfo_t *resolver,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uvloop.loop.__on_addrinfo_resolved", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_request);
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_XDECREF((PyObject *)__pyx_v_ai);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/dns.pyx":457
 * 
 * 
 * cdef void __on_nameinfo_resolved(             # <<<<<<<<<<<<<<
 *     uv.uv_getnameinfo_t* req,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___on_nameinfo_resolved(uv_getnameinfo_t *__pyx_v_req, int __pyx_v_status, char const *__pyx_v_hostname, char const *__pyx_v_service) {
  struct __pyx_obj_6uvloop_4loop_NameInfoRequest *__pyx_v_request = 0;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_callback = 0;
  PyObject *__pyx_v_ex = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  char const *__pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__on_nameinfo_resolved", 0);

  /* "uvloop/dns.pyx":464
 * ) noexcept with gil:
 *     cdef:
 *         NameInfoRequest request = <NameInfoRequest> req.data             # <<<<<<<<<<<<<<
 *         Loop loop = request.loop
 *         object callback = request.callback
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_req->data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_request = ((struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":465
 *     cdef:
 *         NameInfoRequest request = <NameInfoRequest> req.data
 *         Loop loop = request.loop             # <<<<<<<<<<<<<<
 *         object callback = request.callback
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_request->__pyx_base.loop);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":466
 *         NameInfoRequest request = <NameInfoRequest> req.data
 *         Loop loop = request.loop
 *         object callback = request.callback             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __pyx_t_1 = __pyx_v_request->callback;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_callback = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/dns.pyx":468
 *         object callback = request.callback
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if status < 0:
 *             callback(convert_error(status))
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/dns.pyx":469
 * 
 *     try:
 *         if status < 0:             # <<<<<<<<<<<<<<
 *             callback(convert_error(status))
 *         else:
 */
        __pyx_t_5 = (__pyx_v_status < 0);
        if (__pyx_t_5) {

          /* "uvloop/dns.pyx":470
 *     try:
 *         if status < 0:
 *             callback(convert_error(status))             # <<<<<<<<<<<<<<
 *         else:
 *             callback(((<bytes>hostname).decode(),
 */
          __pyx_t_6 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 470, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_v_callback);
          __pyx_t_7 = __pyx_v_callback; __pyx_t_8 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
            __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 470, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "uvloop/dns.pyx":469
 * 
 *     try:
 *         if status < 0:             # <<<<<<<<<<<<<<
 *             callback(convert_error(status))
 *         else:
 */
          goto __pyx_L12;
        }

        /* "uvloop/dns.pyx":472
 *             callback(convert_error(status))
 *         else:
 *             callback(((<bytes>hostname).decode(),             # <<<<<<<<<<<<<<
 *                       (<bytes>service).decode()))
 *     except (KeyboardInterrupt, SystemExit):
 */
        /*else*/ {
          __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_hostname); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 472, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (unlikely(__pyx_t_7 == Py_None)) {
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
            __PYX_ERR(5, 472, __pyx_L6_error)
          }
          __pyx_t_6 = __Pyx_decode_bytes(((PyObject*)__pyx_t_7), 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 472, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "uvloop/dns.pyx":473
 *         else:
 *             callback(((<bytes>hostname).decode(),
 *                       (<bytes>service).decode()))             # <<<<<<<<<<<<<<
 *     except (KeyboardInterrupt, SystemExit):
 *         raise
 */
          __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_service); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 473, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (unlikely(__pyx_t_7 == Py_None)) {
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
            __PYX_ERR(5, 473, __pyx_L6_error)
          }
          __pyx_t_8 = __Pyx_decode_bytes(((PyObject*)__pyx_t_7), 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 473, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "uvloop/dns.pyx":472
 *             callback(convert_error(status))
 *         else:
 *             callback(((<bytes>hostname).decode(),             # <<<<<<<<<<<<<<
 *                       (<bytes>service).decode()))
 *     except (KeyboardInterrupt, SystemExit):
 */
          __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 472, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_6);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6)) __PYX_ERR(5, 472, __pyx_L6_error);
          __Pyx_GIVEREF(__pyx_t_8);
          if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8)) __PYX_ERR(5, 472, __pyx_L6_error);
          __pyx_t_6 = 0;
          __pyx_t_8 = 0;
          __Pyx_INCREF(__pyx_v_callback);
          __pyx_t_8 = __pyx_v_callback; __pyx_t_6 = NULL;
          __pyx_t_9 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_9 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
            __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 472, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __pyx_L12:;

        /* "uvloop/dns.pyx":468
 *         object callback = request.callback
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if status < 0:
 *             callback(convert_error(status))
 */
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "uvloop/dns.pyx":474
 *             callback(((<bytes>hostname).decode(),
 *                       (<bytes>service).decode()))
 *     except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *         raise
 *     except BaseException as ex:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.__on_nameinfo_resolved", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(5, 474, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "uvloop/dns.pyx":475
 *                       (<bytes>service).decode()))
 *     except (KeyboardInterrupt, SystemExit):
 *         raise             # <<<<<<<<<<<<<<
 *     except BaseException as ex:
 *         loop._handle_exception(ex)
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_8, __pyx_t_7);
        __pyx_t_1 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; 
        __PYX_ERR(5, 475, __pyx_L8_except_error)
      }

      /* "uvloop/dns.pyx":476
 *     except (KeyboardInterrupt, SystemExit):
 *         raise
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         loop._handle_exception(ex)
 *     finally:
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("uvloop.loop.__on_nameinfo_resolved", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_1) < 0) __PYX_ERR(5, 476, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_v_ex = __pyx_t_8;
        /*try:*/ {

          /* "uvloop/dns.pyx":477
 *         raise
 *     except BaseException as ex:
 *         loop._handle_exception(ex)             # <<<<<<<<<<<<<<
 *     finally:
 *         request.on_done()
 */
          ((struct __pyx_vtabstruct_6uvloop_4loop_Loop *)__pyx_v_loop->__pyx_vtab)->_handle_exception(__pyx_v_loop, __pyx_v_ex); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 477, __pyx_L20_error)
        }

        /* "uvloop/dns.pyx":476
 *     except (KeyboardInterrupt, SystemExit):
 *         raise
 *     except BaseException as ex:             # <<<<<<<<<<<<<<
 *         loop._handle_exception(ex)
 *     finally:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L21;
          }
          __pyx_L20_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __pyx_t_10 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            }
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;

      /* "uvloop/dns.pyx":468
 *         object callback = request.callback
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if status < 0:
 *             callback(convert_error(status))
 */
      __pyx_L8_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L11_try_end:;
    }
  }

  /* "uvloop/dns.pyx":479
 *         loop._handle_exception(ex)
 *     finally:
 *         request.on_done()             # <<<<<<<<<<<<<<
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_NameInfoRequest *)__pyx_v_request->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_request)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 479, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2) < 0)) __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_3, &__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_16);
      __pyx_t_11 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
      {
        __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_NameInfoRequest *)__pyx_v_request->__pyx_base.__pyx_vtab)->__pyx_base.on_done(((struct __pyx_obj_6uvloop_4loop_UVRequest *)__pyx_v_request)); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 479, __pyx_L27_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_17, __pyx_t_16);
      }
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_ErrRestore(__pyx_t_4, __pyx_t_3, __pyx_t_2);
      __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_19;
      goto __pyx_L1_error;
      __pyx_L27_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_17, __pyx_t_16);
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "uvloop/dns.pyx":457
 * 
 * 
 * cdef void __on_nameinfo_resolved(             # <<<<<<<<<<<<<<
 *     uv.uv_getnameinfo_t* req,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("uvloop.loop.__on_nameinfo_resolved", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_request);
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF(__pyx_v_callback);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/sslproto.pyx":1
 * cdef _create_transport_context(server_side, server_hostname):             # <<<<<<<<<<<<<<
 *     if server_side:
 *         raise ValueError('Server side SSL needs a valid SSLContext')
 */

static PyObject *__pyx_f_6uvloop_4loop__create_transport_context(PyObject *__pyx_v_server_side, PyObject *__pyx_v_server_hostname) {
  PyObject *__pyx_v_sslcontext = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_transport_context", 1);

  /* "uvloop/sslproto.pyx":2
 * cdef _create_transport_context(server_side, server_hostname):
 *     if server_side:             # <<<<<<<<<<<<<<
 *         raise ValueError('Server side SSL needs a valid SSLContext')
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_server_side); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 2, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/sslproto.pyx":3
 * cdef _create_transport_context(server_side, server_hostname):
 *     if server_side:
 *         raise ValueError('Server side SSL needs a valid SSLContext')             # <<<<<<<<<<<<<<
 * 
 *     # Client side may pass ssl=True to use a default
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__184, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(25, 3, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":2
 * cdef _create_transport_context(server_side, server_hostname):
 *     if server_side:             # <<<<<<<<<<<<<<
 *         raise ValueError('Server side SSL needs a valid SSLContext')
 * 
 */
  }

  /* "uvloop/sslproto.pyx":9
 *     # The default is secure for client connections.
 *     # Python 3.4+: use up-to-date strong settings.
 *     sslcontext = ssl_create_default_context()             # <<<<<<<<<<<<<<
 *     if not server_hostname:
 *         sslcontext.check_hostname = False
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_ssl_create_default_context);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_ssl_create_default_context; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_sslcontext = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":10
 *     # Python 3.4+: use up-to-date strong settings.
 *     sslcontext = ssl_create_default_context()
 *     if not server_hostname:             # <<<<<<<<<<<<<<
 *         sslcontext.check_hostname = False
 *     return sslcontext
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_server_hostname); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 10, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_1);
  if (__pyx_t_6) {

    /* "uvloop/sslproto.pyx":11
 *     sslcontext = ssl_create_default_context()
 *     if not server_hostname:
 *         sslcontext.check_hostname = False             # <<<<<<<<<<<<<<
 *     return sslcontext
 * 
 */
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_sslcontext, __pyx_n_s_check_hostname, Py_False) < 0) __PYX_ERR(25, 11, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":10
 *     # Python 3.4+: use up-to-date strong settings.
 *     sslcontext = ssl_create_default_context()
 *     if not server_hostname:             # <<<<<<<<<<<<<<
 *         sslcontext.check_hostname = False
 *     return sslcontext
 */
  }

  /* "uvloop/sslproto.pyx":12
 *     if not server_hostname:
 *         sslcontext.check_hostname = False
 *     return sslcontext             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_sslcontext);
  __pyx_r = __pyx_v_sslcontext;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":1
 * cdef _create_transport_context(server_side, server_hostname):             # <<<<<<<<<<<<<<
 *     if server_side:
 *         raise ValueError('Server side SSL needs a valid SSLContext')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop._create_transport_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sslcontext);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":20
 *     # _sendfile_compatible = constants._SendfileMode.FALLBACK
 * 
 *     def __cinit__(self, Loop loop, ssl_protocol, context):             # <<<<<<<<<<<<<<
 *         self._loop = loop
 *         # SSLProtocol instance
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_ssl_protocol = 0;
  PyObject *__pyx_v_context = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,&__pyx_n_s_ssl_protocol,&__pyx_n_s_context,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 20, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ssl_protocol)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 20, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(25, 20, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 20, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(25, 20, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(25, 20, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)values[0]);
    __pyx_v_ssl_protocol = values[1];
    __pyx_v_context = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(25, 20, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_loop), __pyx_ptype_6uvloop_4loop_Loop, 1, "loop", 0))) __PYX_ERR(25, 20, __pyx_L1_error)
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport___cinit__(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_loop, __pyx_v_ssl_protocol, __pyx_v_context);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport___cinit__(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, PyObject *__pyx_v_ssl_protocol, PyObject *__pyx_v_context) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_context);

  /* "uvloop/sslproto.pyx":21
 * 
 *     def __cinit__(self, Loop loop, ssl_protocol, context):
 *         self._loop = loop             # <<<<<<<<<<<<<<
 *         # SSLProtocol instance
 *         self._ssl_protocol = ssl_protocol
 */
  __Pyx_INCREF((PyObject *)__pyx_v_loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_loop);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_loop);
  __pyx_v_self->_loop = __pyx_v_loop;

  /* "uvloop/sslproto.pyx":23
 *         self._loop = loop
 *         # SSLProtocol instance
 *         self._ssl_protocol = ssl_protocol             # <<<<<<<<<<<<<<
 *         self._closed = False
 *         if context is None:
 */
  if (!(likely(((__pyx_v_ssl_protocol) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_ssl_protocol, __pyx_ptype_6uvloop_4loop_SSLProtocol))))) __PYX_ERR(25, 23, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_ssl_protocol;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_ssl_protocol);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_ssl_protocol);
  __pyx_v_self->_ssl_protocol = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":24
 *         # SSLProtocol instance
 *         self._ssl_protocol = ssl_protocol
 *         self._closed = False             # <<<<<<<<<<<<<<
 *         if context is None:
 *             context = Context_CopyCurrent()
 */
  __pyx_v_self->_closed = 0;

  /* "uvloop/sslproto.pyx":25
 *         self._ssl_protocol = ssl_protocol
 *         self._closed = False
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  __pyx_t_2 = (__pyx_v_context == Py_None);
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":26
 *         self._closed = False
 *         if context is None:
 *             context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 *         self.context = context
 * 
 */
    __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 26, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_context, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":25
 *         self._ssl_protocol = ssl_protocol
 *         self._closed = False
 *         if context is None:             # <<<<<<<<<<<<<<
 *             context = Context_CopyCurrent()
 *         self.context = context
 */
  }

  /* "uvloop/sslproto.pyx":27
 *         if context is None:
 *             context = Context_CopyCurrent()
 *         self.context = context             # <<<<<<<<<<<<<<
 * 
 *     def get_extra_info(self, name, default=None):
 */
  __Pyx_INCREF(__pyx_v_context);
  __Pyx_GIVEREF(__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->context);
  __Pyx_DECREF(__pyx_v_self->context);
  __pyx_v_self->context = __pyx_v_context;

  /* "uvloop/sslproto.pyx":20
 *     # _sendfile_compatible = constants._SendfileMode.FALLBACK
 * 
 *     def __cinit__(self, Loop loop, ssl_protocol, context):             # <<<<<<<<<<<<<<
 *         self._loop = loop
 *         # SSLProtocol instance
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_context);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":29
 *         self.context = context
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         """Get optional transport information."""
 *         return self._ssl_protocol._get_extra_info(name, default)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_3get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_2get_extra_info, "_SSLProtocolTransport.get_extra_info(self, name, default=None)\nGet optional transport information.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_3get_extra_info = {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_3get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_2get_extra_info};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_3get_extra_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_default = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_extra_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_default,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name_2)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 29, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_default);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 29, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_extra_info") < 0)) __PYX_ERR(25, 29, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_extra_info", 0, 1, 2, __pyx_nargs); __PYX_ERR(25, 29, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_2get_extra_info(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_name, __pyx_v_default);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_2get_extra_info(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__get_extra_info __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_extra_info", 1);

  /* "uvloop/sslproto.pyx":31
 *     def get_extra_info(self, name, default=None):
 *         """Get optional transport information."""
 *         return self._ssl_protocol._get_extra_info(name, default)             # <<<<<<<<<<<<<<
 * 
 *     def set_protocol(self, protocol):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.__pyx_default = __pyx_v_default;
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_get_extra_info(__pyx_v_self->_ssl_protocol, __pyx_v_name, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":29
 *         self.context = context
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         """Get optional transport information."""
 *         return self._ssl_protocol._get_extra_info(name, default)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":33
 *         return self._ssl_protocol._get_extra_info(name, default)
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_5set_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_4set_protocol, "_SSLProtocolTransport.set_protocol(self, protocol)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_5set_protocol = {"set_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_5set_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_4set_protocol};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_5set_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_protocol = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_protocol (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_protocol,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protocol)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 33, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_protocol") < 0)) __PYX_ERR(25, 33, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_protocol = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_protocol", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 33, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.set_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_4set_protocol(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_protocol);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_4set_protocol(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_protocol", 1);

  /* "uvloop/sslproto.pyx":34
 * 
 *     def set_protocol(self, protocol):
 *         self._ssl_protocol._set_app_protocol(protocol)             # <<<<<<<<<<<<<<
 * 
 *     def get_protocol(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_set_app_protocol(__pyx_v_self->_ssl_protocol, __pyx_v_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":33
 *         return self._ssl_protocol._get_extra_info(name, default)
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.set_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":36
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._ssl_protocol._app_protocol
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_7get_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_6get_protocol, "_SSLProtocolTransport.get_protocol(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_7get_protocol = {"get_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_7get_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_6get_protocol};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_7get_protocol(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_protocol (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_protocol", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_protocol", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_6get_protocol(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_6get_protocol(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_protocol", 1);

  /* "uvloop/sslproto.pyx":37
 * 
 *     def get_protocol(self):
 *         return self._ssl_protocol._app_protocol             # <<<<<<<<<<<<<<
 * 
 *     def is_closing(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_ssl_protocol->_app_protocol);
  __pyx_r = __pyx_v_self->_ssl_protocol->_app_protocol;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":36
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._ssl_protocol._app_protocol
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":39
 *         return self._ssl_protocol._app_protocol
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_9is_closing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_8is_closing, "_SSLProtocolTransport.is_closing(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_9is_closing = {"is_closing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_9is_closing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_8is_closing};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_9is_closing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_closing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_closing", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_closing", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_8is_closing(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_8is_closing(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_closing", 1);

  /* "uvloop/sslproto.pyx":40
 * 
 *     def is_closing(self):
 *         return self._closed             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":39
 *         return self._ssl_protocol._app_protocol
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.is_closing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":42
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the transport.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_11close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_10close, "_SSLProtocolTransport.close(self)\nClose the transport.\n\n        Buffered data will be flushed asynchronously.  No more data\n        will be received.  After all buffered data is flushed, the\n        protocol's connection_lost() method will (eventually) called\n        with None as its argument.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_11close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_11close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_10close};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_11close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_10close(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_10close(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__start_shutdown __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/sslproto.pyx":50
 *         with None as its argument.
 *         """
 *         self._closed = True             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._start_shutdown(self.context.copy())
 * 
 */
  __pyx_v_self->_closed = 1;

  /* "uvloop/sslproto.pyx":51
 *         """
 *         self._closed = True
 *         self._ssl_protocol._start_shutdown(self.context.copy())             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->context, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5.__pyx_n = 1;
  __pyx_t_5.context = __pyx_t_1;
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_start_shutdown(__pyx_v_self->_ssl_protocol, &__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":42
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the transport.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":53
 *         self._ssl_protocol._start_shutdown(self.context.copy())
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if not self._closed:
 *             self._closed = True
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_13__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_13__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_12__dealloc__(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_12__dealloc__(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/sslproto.pyx":54
 * 
 *     def __dealloc__(self):
 *         if not self._closed:             # <<<<<<<<<<<<<<
 *             self._closed = True
 *             warnings_warn(
 */
  __pyx_t_1 = (!__pyx_v_self->_closed);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":55
 *     def __dealloc__(self):
 *         if not self._closed:
 *             self._closed = True             # <<<<<<<<<<<<<<
 *             warnings_warn(
 *                 "unclosed transport <uvloop.loop._SSLProtocolTransport "
 */
    __pyx_v_self->_closed = 1;

    /* "uvloop/sslproto.pyx":58
 *             warnings_warn(
 *                 "unclosed transport <uvloop.loop._SSLProtocolTransport "
 *                 "object>", ResourceWarning)             # <<<<<<<<<<<<<<
 * 
 *     def is_reading(self):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ResourceWarning); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_warnings_warn);
    __pyx_t_4 = __pyx_v_6uvloop_4loop_warnings_warn; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_unclosed_transport_uvloop_loop, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 56, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":54
 * 
 *     def __dealloc__(self):
 *         if not self._closed:             # <<<<<<<<<<<<<<
 *             self._closed = True
 *             warnings_warn(
 */
  }

  /* "uvloop/sslproto.pyx":53
 *         self._ssl_protocol._start_shutdown(self.context.copy())
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if not self._closed:
 *             self._closed = True
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uvloop.loop._SSLProtocolTransport.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/sslproto.pyx":60
 *                 "object>", ResourceWarning)
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return not self._ssl_protocol._app_reading_paused
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_15is_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_14is_reading, "_SSLProtocolTransport.is_reading(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_15is_reading = {"is_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_15is_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_14is_reading};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_15is_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_14is_reading(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_14is_reading(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_reading", 1);

  /* "uvloop/sslproto.pyx":61
 * 
 *     def is_reading(self):
 *         return not self._ssl_protocol._app_reading_paused             # <<<<<<<<<<<<<<
 * 
 *     def pause_reading(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_v_self->_ssl_protocol->_app_reading_paused)); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":60
 *                 "object>", ResourceWarning)
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return not self._ssl_protocol._app_reading_paused
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.is_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":63
 *         return not self._ssl_protocol._app_reading_paused
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         """Pause the receiving end.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_17pause_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_16pause_reading, "_SSLProtocolTransport.pause_reading(self)\nPause the receiving end.\n\n        No data will be passed to the protocol's data_received()\n        method until resume_reading() is called.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_17pause_reading = {"pause_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_17pause_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_16pause_reading};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_17pause_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pause_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("pause_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pause_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_16pause_reading(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_16pause_reading(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pause_reading", 1);

  /* "uvloop/sslproto.pyx":69
 *         method until resume_reading() is called.
 *         """
 *         self._ssl_protocol._pause_reading()             # <<<<<<<<<<<<<<
 * 
 *     def resume_reading(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_pause_reading(__pyx_v_self->_ssl_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":63
 *         return not self._ssl_protocol._app_reading_paused
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         """Pause the receiving end.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.pause_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":71
 *         self._ssl_protocol._pause_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         """Resume the receiving end.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_19resume_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_18resume_reading, "_SSLProtocolTransport.resume_reading(self)\nResume the receiving end.\n\n        Data received will once again be passed to the protocol's\n        data_received() method.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_19resume_reading = {"resume_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_19resume_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_18resume_reading};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_19resume_reading(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resume_reading (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("resume_reading", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "resume_reading", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_18resume_reading(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_18resume_reading(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resume_reading", 1);

  /* "uvloop/sslproto.pyx":77
 *         data_received() method.
 *         """
 *         self._ssl_protocol._resume_reading(self.context.copy())             # <<<<<<<<<<<<<<
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->context, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_resume_reading(__pyx_v_self->_ssl_protocol, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":71
 *         self._ssl_protocol._pause_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         """Resume the receiving end.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.resume_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":79
 *         self._ssl_protocol._resume_reading(self.context.copy())
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for write flow control.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_21set_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_20set_write_buffer_limits, "_SSLProtocolTransport.set_write_buffer_limits(self, high=None, low=None)\nSet the high- and low-water limits for write flow control.\n\n        These two values control when to call the protocol's\n        pause_writing() and resume_writing() methods.  If specified,\n        the low-water limit must be less than or equal to the\n        high-water limit.  Neither value can be negative.\n\n        The defaults are implementation-specific.  If only the\n        high-water limit is given, the low-water limit defaults to an\n        implementation-specific value less than or equal to the\n        high-water limit.  Setting high to zero forces low to zero as\n        well, and causes pause_writing() to be called whenever the\n        buffer becomes non-empty.  Setting low to zero causes\n        resume_writing() to be called only once the buffer is empty.\n        Use of zero for either limit is generally sub-optimal as it\n        reduces opportunities for doing I/O and computation\n        concurrently.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_21set_write_buffer_limits = {"set_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_21set_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_20set_write_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_21set_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_high = 0;
  PyObject *__pyx_v_low = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_write_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_high_2,&__pyx_n_s_low,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_high_2);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 79, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_low);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 79, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_write_buffer_limits") < 0)) __PYX_ERR(25, 79, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_high = values[0];
    __pyx_v_low = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_write_buffer_limits", 0, 0, 2, __pyx_nargs); __PYX_ERR(25, 79, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.set_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_20set_write_buffer_limits(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_high, __pyx_v_low);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_20set_write_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_high, PyObject *__pyx_v_low) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_write_buffer_limits", 1);

  /* "uvloop/sslproto.pyx":98
 *         concurrently.
 *         """
 *         self._ssl_protocol._set_write_buffer_limits(high, low)             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._control_app_writing(self.context.copy())
 * 
 */
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.high = __pyx_v_high;
  __pyx_t_2.low = __pyx_v_low;
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_set_write_buffer_limits(__pyx_v_self->_ssl_protocol, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":99
 *         """
 *         self._ssl_protocol._set_write_buffer_limits(high, low)
 *         self._ssl_protocol._control_app_writing(self.context.copy())             # <<<<<<<<<<<<<<
 * 
 *     def get_write_buffer_limits(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->context, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6.__pyx_n = 1;
  __pyx_t_6.context = __pyx_t_1;
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_control_app_writing(__pyx_v_self->_ssl_protocol, &__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/sslproto.pyx":79
 *         self._ssl_protocol._resume_reading(self.context.copy())
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for write flow control.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.set_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":101
 *         self._ssl_protocol._control_app_writing(self.context.copy())
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._outgoing_low_water,
 *                 self._ssl_protocol._outgoing_high_water)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_23get_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_22get_write_buffer_limits, "_SSLProtocolTransport.get_write_buffer_limits(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_23get_write_buffer_limits = {"get_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_23get_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_22get_write_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_23get_write_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_write_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_write_buffer_limits", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_write_buffer_limits", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_22get_write_buffer_limits(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_22get_write_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_write_buffer_limits", 1);

  /* "uvloop/sslproto.pyx":102
 * 
 *     def get_write_buffer_limits(self):
 *         return (self._ssl_protocol._outgoing_low_water,             # <<<<<<<<<<<<<<
 *                 self._ssl_protocol._outgoing_high_water)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_ssl_protocol->_outgoing_low_water); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "uvloop/sslproto.pyx":103
 *     def get_write_buffer_limits(self):
 *         return (self._ssl_protocol._outgoing_low_water,
 *                 self._ssl_protocol._outgoing_high_water)             # <<<<<<<<<<<<<<
 * 
 *     def get_write_buffer_size(self):
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_ssl_protocol->_outgoing_high_water); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/sslproto.pyx":102
 * 
 *     def get_write_buffer_limits(self):
 *         return (self._ssl_protocol._outgoing_low_water,             # <<<<<<<<<<<<<<
 *                 self._ssl_protocol._outgoing_high_water)
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(25, 102, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(25, 102, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":101
 *         self._ssl_protocol._control_app_writing(self.context.copy())
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._outgoing_low_water,
 *                 self._ssl_protocol._outgoing_high_water)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.get_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":105
 *                 self._ssl_protocol._outgoing_high_water)
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the write buffers."""
 *         return self._ssl_protocol._get_write_buffer_size()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_25get_write_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_24get_write_buffer_size, "_SSLProtocolTransport.get_write_buffer_size(self)\nReturn the current size of the write buffers.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_25get_write_buffer_size = {"get_write_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_25get_write_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_24get_write_buffer_size};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_25get_write_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_write_buffer_size (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_write_buffer_size", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_write_buffer_size", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_24get_write_buffer_size(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_24get_write_buffer_size(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_write_buffer_size", 1);

  /* "uvloop/sslproto.pyx":107
 *     def get_write_buffer_size(self):
 *         """Return the current size of the write buffers."""
 *         return self._ssl_protocol._get_write_buffer_size()             # <<<<<<<<<<<<<<
 * 
 *     def set_read_buffer_limits(self, high=None, low=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_get_write_buffer_size(__pyx_v_self->_ssl_protocol); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 107, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":105
 *                 self._ssl_protocol._outgoing_high_water)
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the write buffers."""
 *         return self._ssl_protocol._get_write_buffer_size()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.get_write_buffer_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":109
 *         return self._ssl_protocol._get_write_buffer_size()
 * 
 *     def set_read_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for read flow control.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_27set_read_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_26set_read_buffer_limits, "_SSLProtocolTransport.set_read_buffer_limits(self, high=None, low=None)\nSet the high- and low-water limits for read flow control.\n\n        These two values control when to call the upstream transport's\n        pause_reading() and resume_reading() methods.  If specified,\n        the low-water limit must be less than or equal to the\n        high-water limit.  Neither value can be negative.\n\n        The defaults are implementation-specific.  If only the\n        high-water limit is given, the low-water limit defaults to an\n        implementation-specific value less than or equal to the\n        high-water limit.  Setting high to zero forces low to zero as\n        well, and causes pause_reading() to be called whenever the\n        buffer becomes non-empty.  Setting low to zero causes\n        resume_reading() to be called only once the buffer is empty.\n        Use of zero for either limit is generally sub-optimal as it\n        reduces opportunities for doing I/O and computation\n        concurrently.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_27set_read_buffer_limits = {"set_read_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_27set_read_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_26set_read_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_27set_read_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_high = 0;
  PyObject *__pyx_v_low = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_read_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_high_2,&__pyx_n_s_low,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_high_2);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 109, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_low);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 109, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_read_buffer_limits") < 0)) __PYX_ERR(25, 109, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_high = values[0];
    __pyx_v_low = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_read_buffer_limits", 0, 0, 2, __pyx_nargs); __PYX_ERR(25, 109, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.set_read_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_26set_read_buffer_limits(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_high, __pyx_v_low);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_26set_read_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_high, PyObject *__pyx_v_low) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_read_buffer_limits", 1);

  /* "uvloop/sslproto.pyx":128
 *         concurrently.
 *         """
 *         self._ssl_protocol._set_read_buffer_limits(high, low)             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._control_ssl_reading()
 * 
 */
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.high = __pyx_v_high;
  __pyx_t_2.low = __pyx_v_low;
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_set_read_buffer_limits(__pyx_v_self->_ssl_protocol, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":129
 *         """
 *         self._ssl_protocol._set_read_buffer_limits(high, low)
 *         self._ssl_protocol._control_ssl_reading()             # <<<<<<<<<<<<<<
 * 
 *     def get_read_buffer_limits(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_control_ssl_reading(__pyx_v_self->_ssl_protocol); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":109
 *         return self._ssl_protocol._get_write_buffer_size()
 * 
 *     def set_read_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for read flow control.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.set_read_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":131
 *         self._ssl_protocol._control_ssl_reading()
 * 
 *     def get_read_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._incoming_low_water,
 *                 self._ssl_protocol._incoming_high_water)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_29get_read_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_28get_read_buffer_limits, "_SSLProtocolTransport.get_read_buffer_limits(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_29get_read_buffer_limits = {"get_read_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_29get_read_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_28get_read_buffer_limits};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_29get_read_buffer_limits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_read_buffer_limits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_read_buffer_limits", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_read_buffer_limits", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_28get_read_buffer_limits(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_28get_read_buffer_limits(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_read_buffer_limits", 1);

  /* "uvloop/sslproto.pyx":132
 * 
 *     def get_read_buffer_limits(self):
 *         return (self._ssl_protocol._incoming_low_water,             # <<<<<<<<<<<<<<
 *                 self._ssl_protocol._incoming_high_water)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_ssl_protocol->_incoming_low_water); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "uvloop/sslproto.pyx":133
 *     def get_read_buffer_limits(self):
 *         return (self._ssl_protocol._incoming_low_water,
 *                 self._ssl_protocol._incoming_high_water)             # <<<<<<<<<<<<<<
 * 
 *     def get_read_buffer_size(self):
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_ssl_protocol->_incoming_high_water); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/sslproto.pyx":132
 * 
 *     def get_read_buffer_limits(self):
 *         return (self._ssl_protocol._incoming_low_water,             # <<<<<<<<<<<<<<
 *                 self._ssl_protocol._incoming_high_water)
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(25, 132, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(25, 132, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":131
 *         self._ssl_protocol._control_ssl_reading()
 * 
 *     def get_read_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._incoming_low_water,
 *                 self._ssl_protocol._incoming_high_water)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.get_read_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":135
 *                 self._ssl_protocol._incoming_high_water)
 * 
 *     def get_read_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the read buffer."""
 *         return self._ssl_protocol._get_read_buffer_size()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_31get_read_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_30get_read_buffer_size, "_SSLProtocolTransport.get_read_buffer_size(self)\nReturn the current size of the read buffer.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_31get_read_buffer_size = {"get_read_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_31get_read_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_30get_read_buffer_size};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_31get_read_buffer_size(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_read_buffer_size (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_read_buffer_size", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_read_buffer_size", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_30get_read_buffer_size(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_30get_read_buffer_size(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_read_buffer_size", 1);

  /* "uvloop/sslproto.pyx":137
 *     def get_read_buffer_size(self):
 *         """Return the current size of the read buffer."""
 *         return self._ssl_protocol._get_read_buffer_size()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_get_read_buffer_size(__pyx_v_self->_ssl_protocol); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 137, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":135
 *                 self._ssl_protocol._incoming_high_water)
 * 
 *     def get_read_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the read buffer."""
 *         return self._ssl_protocol._get_read_buffer_size()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.get_read_buffer_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":139
 *         return self._ssl_protocol._get_read_buffer_size()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def _protocol_paused(self):
 *         # Required for sendfile fallback pause_writing/resume_writing logic
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_16_protocol_paused_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_16_protocol_paused_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_16_protocol_paused___get__(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_16_protocol_paused___get__(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uvloop/sslproto.pyx":142
 *     def _protocol_paused(self):
 *         # Required for sendfile fallback pause_writing/resume_writing logic
 *         return self._ssl_protocol._app_writing_paused             # <<<<<<<<<<<<<<
 * 
 *     def write(self, data):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_ssl_protocol->_app_writing_paused); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":139
 *         return self._ssl_protocol._get_read_buffer_size()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def _protocol_paused(self):
 *         # Required for sendfile fallback pause_writing/resume_writing logic
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport._protocol_paused.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":144
 *         return self._ssl_protocol._app_writing_paused
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         """Write some data bytes to the transport.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_33write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_32write, "_SSLProtocolTransport.write(self, data)\nWrite some data bytes to the transport.\n\n        This does not block; it buffers the data and arranges for it\n        to be sent out asynchronously.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_33write = {"write", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_33write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_32write};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_33write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 144, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "write") < 0)) __PYX_ERR(25, 144, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("write", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 144, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_32write(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_32write(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write", 1);

  /* "uvloop/sslproto.pyx":150
 *         to be sent out asynchronously.
 *         """
 *         if not isinstance(data, (bytes, bytearray, memoryview)):             # <<<<<<<<<<<<<<
 *             raise TypeError(f"data: expecting a bytes-like instance, "
 *                             f"got {type(data).__name__}")
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_data); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyByteArray_Check(__pyx_v_data); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = PyMemoryView_Check(__pyx_v_data); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = (!__pyx_t_1);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/sslproto.pyx":152
 *         if not isinstance(data, (bytes, bytearray, memoryview)):
 *             raise TypeError(f"data: expecting a bytes-like instance, "
 *                             f"got {type(data).__name__}")             # <<<<<<<<<<<<<<
 *         if not data:
 *             return
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_data)), __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_3, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":151
 *         """
 *         if not isinstance(data, (bytes, bytearray, memoryview)):
 *             raise TypeError(f"data: expecting a bytes-like instance, "             # <<<<<<<<<<<<<<
 *                             f"got {type(data).__name__}")
 *         if not data:
 */
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_data_expecting_a_bytes_like_inst, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(25, 151, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":150
 *         to be sent out asynchronously.
 *         """
 *         if not isinstance(data, (bytes, bytearray, memoryview)):             # <<<<<<<<<<<<<<
 *             raise TypeError(f"data: expecting a bytes-like instance, "
 *                             f"got {type(data).__name__}")
 */
  }

  /* "uvloop/sslproto.pyx":153
 *             raise TypeError(f"data: expecting a bytes-like instance, "
 *                             f"got {type(data).__name__}")
 *         if not data:             # <<<<<<<<<<<<<<
 *             return
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_data); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(25, 153, __pyx_L1_error)
  __pyx_t_1 = (!__pyx_t_2);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":154
 *                             f"got {type(data).__name__}")
 *         if not data:
 *             return             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":153
 *             raise TypeError(f"data: expecting a bytes-like instance, "
 *                             f"got {type(data).__name__}")
 *         if not data:             # <<<<<<<<<<<<<<
 *             return
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())
 */
  }

  /* "uvloop/sslproto.pyx":155
 *         if not data:
 *             return
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())             # <<<<<<<<<<<<<<
 * 
 *     def writelines(self, list_of_data):
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_data)) __PYX_ERR(25, 155, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->context, __pyx_n_s_copy); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_write_appdata(__pyx_v_self->_ssl_protocol, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uvloop/sslproto.pyx":144
 *         return self._ssl_protocol._app_writing_paused
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         """Write some data bytes to the transport.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":157
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         """Write a list (or any iterable) of data bytes to the transport.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_35writelines(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_34writelines, "_SSLProtocolTransport.writelines(self, list_of_data)\nWrite a list (or any iterable) of data bytes to the transport.\n\n        The default implementation concatenates the arguments and\n        calls write() on the result.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_35writelines = {"writelines", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_35writelines, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_34writelines};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_35writelines(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_list_of_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writelines (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_list_of_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_list_of_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 157, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "writelines") < 0)) __PYX_ERR(25, 157, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_list_of_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writelines", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 157, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.writelines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_34writelines(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_list_of_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_34writelines(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_list_of_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("writelines", 1);

  /* "uvloop/sslproto.pyx":163
 *         calls write() on the result.
 *         """
 *         self._ssl_protocol._write_appdata(list_of_data, self.context.copy())             # <<<<<<<<<<<<<<
 * 
 *     def write_eof(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->context, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_write_appdata(__pyx_v_self->_ssl_protocol, __pyx_v_list_of_data, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":157
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         """Write a list (or any iterable) of data bytes to the transport.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.writelines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":165
 *         self._ssl_protocol._write_appdata(list_of_data, self.context.copy())
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         """Close the write end after flushing buffered data.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_37write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_36write_eof, "_SSLProtocolTransport.write_eof(self)\nClose the write end after flushing buffered data.\n\n        This raises :exc:`NotImplementedError` right now.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_37write_eof = {"write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_37write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_36write_eof};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_37write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_eof (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("write_eof", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "write_eof", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_36write_eof(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_36write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_eof", 1);

  /* "uvloop/sslproto.pyx":170
 *         This raises :exc:`NotImplementedError` right now.
 *         """
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def can_write_eof(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(25, 170, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":165
 *         self._ssl_protocol._write_appdata(list_of_data, self.context.copy())
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         """Close the write end after flushing buffered data.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.write_eof", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":172
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         """Return True if this transport supports write_eof(), False if not."""
 *         return False
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_39can_write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_38can_write_eof, "_SSLProtocolTransport.can_write_eof(self)\nReturn True if this transport supports write_eof(), False if not.");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_39can_write_eof = {"can_write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_39can_write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_38can_write_eof};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_39can_write_eof(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_write_eof (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("can_write_eof", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "can_write_eof", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_38can_write_eof(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_38can_write_eof(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_write_eof", 1);

  /* "uvloop/sslproto.pyx":174
 *     def can_write_eof(self):
 *         """Return True if this transport supports write_eof(), False if not."""
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def abort(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":172
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         """Return True if this transport supports write_eof(), False if not."""
 *         return False
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":176
 *         return False
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         """Close the transport immediately.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_41abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_40abort, "_SSLProtocolTransport.abort(self)\nClose the transport immediately.\n\n        Buffered data will be lost.  No more data will be received.\n        The protocol's connection_lost() method will (eventually) be\n        called with None as its argument.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_41abort = {"abort", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_41abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_40abort};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_41abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("abort (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("abort", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "abort", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_40abort(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_40abort(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("abort", 1);

  /* "uvloop/sslproto.pyx":183
 *         called with None as its argument.
 *         """
 *         self._force_close(None)             # <<<<<<<<<<<<<<
 * 
 *     def _force_close(self, exc):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_force_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, Py_None};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":176
 *         return False
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         """Close the transport immediately.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.abort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":185
 *         self._force_close(None)
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         self._closed = True
 *         self._ssl_protocol._abort(exc)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_43_force_close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_42_force_close, "_SSLProtocolTransport._force_close(self, exc)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_43_force_close = {"_force_close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_43_force_close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_42_force_close};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_43_force_close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_exc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_force_close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 185, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_force_close") < 0)) __PYX_ERR(25, 185, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_exc = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_force_close", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 185, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport._force_close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_42_force_close(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_exc);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_42_force_close(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_force_close", 1);

  /* "uvloop/sslproto.pyx":186
 * 
 *     def _force_close(self, exc):
 *         self._closed = True             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._abort(exc)
 * 
 */
  __pyx_v_self->_closed = 1;

  /* "uvloop/sslproto.pyx":187
 *     def _force_close(self, exc):
 *         self._closed = True
 *         self._ssl_protocol._abort(exc)             # <<<<<<<<<<<<<<
 * 
 *     def _test__append_write_backlog(self, data):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->_ssl_protocol->__pyx_vtab)->_abort(__pyx_v_self->_ssl_protocol, __pyx_v_exc); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":185
 *         self._force_close(None)
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         self._closed = True
 *         self._ssl_protocol._abort(exc)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport._force_close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":189
 *         self._ssl_protocol._abort(exc)
 * 
 *     def _test__append_write_backlog(self, data):             # <<<<<<<<<<<<<<
 *         # for test only
 *         self._ssl_protocol._write_backlog.append(data)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_45_test__append_write_backlog(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_44_test__append_write_backlog, "_SSLProtocolTransport._test__append_write_backlog(self, data)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_45_test__append_write_backlog = {"_test__append_write_backlog", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_45_test__append_write_backlog, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_44_test__append_write_backlog};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_45_test__append_write_backlog(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test__append_write_backlog (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 189, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_test__append_write_backlog") < 0)) __PYX_ERR(25, 189, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_test__append_write_backlog", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 189, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport._test__append_write_backlog", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_44_test__append_write_backlog(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_44_test__append_write_backlog(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test__append_write_backlog", 1);

  /* "uvloop/sslproto.pyx":191
 *     def _test__append_write_backlog(self, data):
 *         # for test only
 *         self._ssl_protocol._write_backlog.append(data)             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._write_buffer_size += len(data)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Append(__pyx_v_self->_ssl_protocol->_write_backlog, __pyx_v_data); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 191, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":192
 *         # for test only
 *         self._ssl_protocol._write_backlog.append(data)
 *         self._ssl_protocol._write_buffer_size += len(data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF((PyObject *)__pyx_v_self->_ssl_protocol);
  __pyx_t_2 = __pyx_v_self->_ssl_protocol;
  __pyx_t_3 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(25, 192, __pyx_L1_error)
  __pyx_t_2->_write_buffer_size = (__pyx_t_2->_write_buffer_size + __pyx_t_3);
  __Pyx_DECREF((PyObject *)__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":189
 *         self._ssl_protocol._abort(exc)
 * 
 *     def _test__append_write_backlog(self, data):             # <<<<<<<<<<<<<<
 *         # for test only
 *         self._ssl_protocol._write_backlog.append(data)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF((PyObject *)__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport._test__append_write_backlog", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_47__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_46__reduce_cython__, "_SSLProtocolTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_47__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_47__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_46__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_47__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_46__reduce_cython__(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_46__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_49__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_48__setstate_cython__, "_SSLProtocolTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_49__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_49__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_48__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_49__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_48__setstate_cython__(((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_21_SSLProtocolTransport_48__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._SSLProtocolTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":202
 *     """
 * 
 *     def __cinit__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._ssl_buffer_len = SSL_READ_MAX_SIZE
 *         self._ssl_buffer = <char*>PyMem_RawMalloc(self._ssl_buffer_len)
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_11SSLProtocol_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_11SSLProtocol_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 1))) return -1;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol___cinit__(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_11SSLProtocol___cinit__(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/sslproto.pyx":203
 * 
 *     def __cinit__(self, *args, **kwargs):
 *         self._ssl_buffer_len = SSL_READ_MAX_SIZE             # <<<<<<<<<<<<<<
 *         self._ssl_buffer = <char*>PyMem_RawMalloc(self._ssl_buffer_len)
 *         if not self._ssl_buffer:
 */
  __pyx_v_self->_ssl_buffer_len = __pyx_e_6uvloop_4loop_SSL_READ_MAX_SIZE;

  /* "uvloop/sslproto.pyx":204
 *     def __cinit__(self, *args, **kwargs):
 *         self._ssl_buffer_len = SSL_READ_MAX_SIZE
 *         self._ssl_buffer = <char*>PyMem_RawMalloc(self._ssl_buffer_len)             # <<<<<<<<<<<<<<
 *         if not self._ssl_buffer:
 *             raise MemoryError()
 */
  __pyx_v_self->_ssl_buffer = ((char *)PyMem_RawMalloc(__pyx_v_self->_ssl_buffer_len));

  /* "uvloop/sslproto.pyx":205
 *         self._ssl_buffer_len = SSL_READ_MAX_SIZE
 *         self._ssl_buffer = <char*>PyMem_RawMalloc(self._ssl_buffer_len)
 *         if not self._ssl_buffer:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self._ssl_buffer_view = PyMemoryView_FromMemory(
 */
  __pyx_t_1 = (!(__pyx_v_self->_ssl_buffer != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/sslproto.pyx":206
 *         self._ssl_buffer = <char*>PyMem_RawMalloc(self._ssl_buffer_len)
 *         if not self._ssl_buffer:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         self._ssl_buffer_view = PyMemoryView_FromMemory(
 *             self._ssl_buffer, self._ssl_buffer_len, PyBUF_WRITE)
 */
    PyErr_NoMemory(); __PYX_ERR(25, 206, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":205
 *         self._ssl_buffer_len = SSL_READ_MAX_SIZE
 *         self._ssl_buffer = <char*>PyMem_RawMalloc(self._ssl_buffer_len)
 *         if not self._ssl_buffer:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self._ssl_buffer_view = PyMemoryView_FromMemory(
 */
  }

  /* "uvloop/sslproto.pyx":207
 *         if not self._ssl_buffer:
 *             raise MemoryError()
 *         self._ssl_buffer_view = PyMemoryView_FromMemory(             # <<<<<<<<<<<<<<
 *             self._ssl_buffer, self._ssl_buffer_len, PyBUF_WRITE)
 * 
 */
  __pyx_t_2 = PyMemoryView_FromMemory(__pyx_v_self->_ssl_buffer, __pyx_v_self->_ssl_buffer_len, PyBUF_WRITE); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_ssl_buffer_view);
  __Pyx_DECREF(__pyx_v_self->_ssl_buffer_view);
  __pyx_v_self->_ssl_buffer_view = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":202
 *     """
 * 
 *     def __cinit__(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._ssl_buffer_len = SSL_READ_MAX_SIZE
 *         self._ssl_buffer = <char*>PyMem_RawMalloc(self._ssl_buffer_len)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":210
 *             self._ssl_buffer, self._ssl_buffer_len, PyBUF_WRITE)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self._ssl_buffer_view = None
 *         PyMem_RawFree(self._ssl_buffer)
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_11SSLProtocol_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_11SSLProtocol_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_11SSLProtocol_2__dealloc__(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_11SSLProtocol_2__dealloc__(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/sslproto.pyx":211
 * 
 *     def __dealloc__(self):
 *         self._ssl_buffer_view = None             # <<<<<<<<<<<<<<
 *         PyMem_RawFree(self._ssl_buffer)
 *         self._ssl_buffer = NULL
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_ssl_buffer_view);
  __Pyx_DECREF(__pyx_v_self->_ssl_buffer_view);
  __pyx_v_self->_ssl_buffer_view = Py_None;

  /* "uvloop/sslproto.pyx":212
 *     def __dealloc__(self):
 *         self._ssl_buffer_view = None
 *         PyMem_RawFree(self._ssl_buffer)             # <<<<<<<<<<<<<<
 *         self._ssl_buffer = NULL
 *         self._ssl_buffer_len = 0
 */
  PyMem_RawFree(__pyx_v_self->_ssl_buffer);

  /* "uvloop/sslproto.pyx":213
 *         self._ssl_buffer_view = None
 *         PyMem_RawFree(self._ssl_buffer)
 *         self._ssl_buffer = NULL             # <<<<<<<<<<<<<<
 *         self._ssl_buffer_len = 0
 * 
 */
  __pyx_v_self->_ssl_buffer = NULL;

  /* "uvloop/sslproto.pyx":214
 *         PyMem_RawFree(self._ssl_buffer)
 *         self._ssl_buffer = NULL
 *         self._ssl_buffer_len = 0             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, loop, app_protocol, sslcontext, waiter,
 */
  __pyx_v_self->_ssl_buffer_len = 0;

  /* "uvloop/sslproto.pyx":210
 *             self._ssl_buffer, self._ssl_buffer_len, PyBUF_WRITE)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self._ssl_buffer_view = None
 *         PyMem_RawFree(self._ssl_buffer)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "uvloop/sslproto.pyx":216
 *         self._ssl_buffer_len = 0
 * 
 *     def __init__(self, loop, app_protocol, sslcontext, waiter,             # <<<<<<<<<<<<<<
 *                  server_side=False, server_hostname=None,
 *                  call_connection_made=True,
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_11SSLProtocol_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_11SSLProtocol_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_loop = 0;
  PyObject *__pyx_v_app_protocol = 0;
  PyObject *__pyx_v_sslcontext = 0;
  PyObject *__pyx_v_waiter = 0;
  PyObject *__pyx_v_server_side = 0;
  PyObject *__pyx_v_server_hostname = 0;
  PyObject *__pyx_v_call_connection_made = 0;
  PyObject *__pyx_v_ssl_handshake_timeout = 0;
  PyObject *__pyx_v_ssl_shutdown_timeout = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,&__pyx_n_s_app_protocol,&__pyx_n_s_sslcontext,&__pyx_n_s_waiter,&__pyx_n_s_server_side,&__pyx_n_s_server_hostname,&__pyx_n_s_call_connection_made,&__pyx_n_s_ssl_handshake_timeout,&__pyx_n_s_ssl_shutdown_timeout,0};

    /* "uvloop/sslproto.pyx":217
 * 
 *     def __init__(self, loop, app_protocol, sslcontext, waiter,
 *                  server_side=False, server_hostname=None,             # <<<<<<<<<<<<<<
 *                  call_connection_made=True,
 *                  ssl_handshake_timeout=None,
 */
    values[4] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_False));
    values[5] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));

    /* "uvloop/sslproto.pyx":218
 *     def __init__(self, loop, app_protocol, sslcontext, waiter,
 *                  server_side=False, server_hostname=None,
 *                  call_connection_made=True,             # <<<<<<<<<<<<<<
 *                  ssl_handshake_timeout=None,
 *                  ssl_shutdown_timeout=None):
 */
    values[6] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_True));

    /* "uvloop/sslproto.pyx":219
 *                  server_side=False, server_hostname=None,
 *                  call_connection_made=True,
 *                  ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                  ssl_shutdown_timeout=None):
 *         if ssl_handshake_timeout is None:
 */
    values[7] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));

    /* "uvloop/sslproto.pyx":220
 *                  call_connection_made=True,
 *                  ssl_handshake_timeout=None,
 *                  ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 *         if ssl_handshake_timeout is None:
 *             ssl_handshake_timeout = SSL_HANDSHAKE_TIMEOUT
 */
    values[8] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  9: values[8] = __Pyx_Arg_VARARGS(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_app_protocol)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 9, 1); __PYX_ERR(25, 216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_sslcontext)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 9, 2); __PYX_ERR(25, 216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_waiter)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 9, 3); __PYX_ERR(25, 216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_server_side);
          if (value) { values[4] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_server_hostname);
          if (value) { values[5] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_call_connection_made);
          if (value) { values[6] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ssl_handshake_timeout);
          if (value) { values[7] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ssl_shutdown_timeout);
          if (value) { values[8] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 216, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(25, 216, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  9: values[8] = __Pyx_Arg_VARARGS(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = __Pyx_Arg_VARARGS(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = __Pyx_Arg_VARARGS(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_VARARGS(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_loop = values[0];
    __pyx_v_app_protocol = values[1];
    __pyx_v_sslcontext = values[2];
    __pyx_v_waiter = values[3];
    __pyx_v_server_side = values[4];
    __pyx_v_server_hostname = values[5];
    __pyx_v_call_connection_made = values[6];
    __pyx_v_ssl_handshake_timeout = values[7];
    __pyx_v_ssl_shutdown_timeout = values[8];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 9, __pyx_nargs); __PYX_ERR(25, 216, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_4__init__(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v_loop, __pyx_v_app_protocol, __pyx_v_sslcontext, __pyx_v_waiter, __pyx_v_server_side, __pyx_v_server_hostname, __pyx_v_call_connection_made, __pyx_v_ssl_handshake_timeout, __pyx_v_ssl_shutdown_timeout);

  /* "uvloop/sslproto.pyx":216
 *         self._ssl_buffer_len = 0
 * 
 *     def __init__(self, loop, app_protocol, sslcontext, waiter,             # <<<<<<<<<<<<<<
 *                  server_side=False, server_hostname=None,
 *                  call_connection_made=True,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_11SSLProtocol_4__init__(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_loop, PyObject *__pyx_v_app_protocol, PyObject *__pyx_v_sslcontext, PyObject *__pyx_v_waiter, PyObject *__pyx_v_server_side, PyObject *__pyx_v_server_hostname, PyObject *__pyx_v_call_connection_made, PyObject *__pyx_v_ssl_handshake_timeout, PyObject *__pyx_v_ssl_shutdown_timeout) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_sslcontext);
  __Pyx_INCREF(__pyx_v_ssl_handshake_timeout);
  __Pyx_INCREF(__pyx_v_ssl_shutdown_timeout);

  /* "uvloop/sslproto.pyx":221
 *                  ssl_handshake_timeout=None,
 *                  ssl_shutdown_timeout=None):
 *         if ssl_handshake_timeout is None:             # <<<<<<<<<<<<<<
 *             ssl_handshake_timeout = SSL_HANDSHAKE_TIMEOUT
 *         elif ssl_handshake_timeout <= 0:
 */
  __pyx_t_1 = (__pyx_v_ssl_handshake_timeout == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":222
 *                  ssl_shutdown_timeout=None):
 *         if ssl_handshake_timeout is None:
 *             ssl_handshake_timeout = SSL_HANDSHAKE_TIMEOUT             # <<<<<<<<<<<<<<
 *         elif ssl_handshake_timeout <= 0:
 *             raise ValueError(
 */
    __pyx_t_2 = PyFloat_FromDouble(SSL_HANDSHAKE_TIMEOUT); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_ssl_handshake_timeout, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":221
 *                  ssl_handshake_timeout=None,
 *                  ssl_shutdown_timeout=None):
 *         if ssl_handshake_timeout is None:             # <<<<<<<<<<<<<<
 *             ssl_handshake_timeout = SSL_HANDSHAKE_TIMEOUT
 *         elif ssl_handshake_timeout <= 0:
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":223
 *         if ssl_handshake_timeout is None:
 *             ssl_handshake_timeout = SSL_HANDSHAKE_TIMEOUT
 *         elif ssl_handshake_timeout <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f"ssl_handshake_timeout should be a positive number, "
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_ssl_handshake_timeout, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 223, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/sslproto.pyx":226
 *             raise ValueError(
 *                 f"ssl_handshake_timeout should be a positive number, "
 *                 f"got {ssl_handshake_timeout}")             # <<<<<<<<<<<<<<
 *         if ssl_shutdown_timeout is None:
 *             ssl_shutdown_timeout = SSL_SHUTDOWN_TIMEOUT
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_ssl_handshake_timeout, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "uvloop/sslproto.pyx":225
 *         elif ssl_handshake_timeout <= 0:
 *             raise ValueError(
 *                 f"ssl_handshake_timeout should be a positive number, "             # <<<<<<<<<<<<<<
 *                 f"got {ssl_handshake_timeout}")
 *         if ssl_shutdown_timeout is None:
 */
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_ssl_handshake_timeout_should_be, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":224
 *             ssl_handshake_timeout = SSL_HANDSHAKE_TIMEOUT
 *         elif ssl_handshake_timeout <= 0:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 f"ssl_handshake_timeout should be a positive number, "
 *                 f"got {ssl_handshake_timeout}")
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(25, 224, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":223
 *         if ssl_handshake_timeout is None:
 *             ssl_handshake_timeout = SSL_HANDSHAKE_TIMEOUT
 *         elif ssl_handshake_timeout <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f"ssl_handshake_timeout should be a positive number, "
 */
  }
  __pyx_L3:;

  /* "uvloop/sslproto.pyx":227
 *                 f"ssl_handshake_timeout should be a positive number, "
 *                 f"got {ssl_handshake_timeout}")
 *         if ssl_shutdown_timeout is None:             # <<<<<<<<<<<<<<
 *             ssl_shutdown_timeout = SSL_SHUTDOWN_TIMEOUT
 *         elif ssl_shutdown_timeout <= 0:
 */
  __pyx_t_1 = (__pyx_v_ssl_shutdown_timeout == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":228
 *                 f"got {ssl_handshake_timeout}")
 *         if ssl_shutdown_timeout is None:
 *             ssl_shutdown_timeout = SSL_SHUTDOWN_TIMEOUT             # <<<<<<<<<<<<<<
 *         elif ssl_shutdown_timeout <= 0:
 *             raise ValueError(
 */
    __pyx_t_2 = PyFloat_FromDouble(SSL_SHUTDOWN_TIMEOUT); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 228, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_ssl_shutdown_timeout, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":227
 *                 f"ssl_handshake_timeout should be a positive number, "
 *                 f"got {ssl_handshake_timeout}")
 *         if ssl_shutdown_timeout is None:             # <<<<<<<<<<<<<<
 *             ssl_shutdown_timeout = SSL_SHUTDOWN_TIMEOUT
 *         elif ssl_shutdown_timeout <= 0:
 */
    goto __pyx_L4;
  }

  /* "uvloop/sslproto.pyx":229
 *         if ssl_shutdown_timeout is None:
 *             ssl_shutdown_timeout = SSL_SHUTDOWN_TIMEOUT
 *         elif ssl_shutdown_timeout <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f"ssl_shutdown_timeout should be a positive number, "
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_ssl_shutdown_timeout, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 229, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 229, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/sslproto.pyx":232
 *             raise ValueError(
 *                 f"ssl_shutdown_timeout should be a positive number, "
 *                 f"got {ssl_shutdown_timeout}")             # <<<<<<<<<<<<<<
 * 
 *         if not sslcontext:
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_ssl_shutdown_timeout, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "uvloop/sslproto.pyx":231
 *         elif ssl_shutdown_timeout <= 0:
 *             raise ValueError(
 *                 f"ssl_shutdown_timeout should be a positive number, "             # <<<<<<<<<<<<<<
 *                 f"got {ssl_shutdown_timeout}")
 * 
 */
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_ssl_shutdown_timeout_should_be_a, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":230
 *             ssl_shutdown_timeout = SSL_SHUTDOWN_TIMEOUT
 *         elif ssl_shutdown_timeout <= 0:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 f"ssl_shutdown_timeout should be a positive number, "
 *                 f"got {ssl_shutdown_timeout}")
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(25, 230, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":229
 *         if ssl_shutdown_timeout is None:
 *             ssl_shutdown_timeout = SSL_SHUTDOWN_TIMEOUT
 *         elif ssl_shutdown_timeout <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError(
 *                 f"ssl_shutdown_timeout should be a positive number, "
 */
  }
  __pyx_L4:;

  /* "uvloop/sslproto.pyx":234
 *                 f"got {ssl_shutdown_timeout}")
 * 
 *         if not sslcontext:             # <<<<<<<<<<<<<<
 *             sslcontext = _create_transport_context(
 *                 server_side, server_hostname)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_sslcontext); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 234, __pyx_L1_error)
  __pyx_t_4 = (!__pyx_t_1);
  if (__pyx_t_4) {

    /* "uvloop/sslproto.pyx":235
 * 
 *         if not sslcontext:
 *             sslcontext = _create_transport_context(             # <<<<<<<<<<<<<<
 *                 server_side, server_hostname)
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop__create_transport_context(__pyx_v_server_side, __pyx_v_server_hostname); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_sslcontext, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":234
 *                 f"got {ssl_shutdown_timeout}")
 * 
 *         if not sslcontext:             # <<<<<<<<<<<<<<
 *             sslcontext = _create_transport_context(
 *                 server_side, server_hostname)
 */
  }

  /* "uvloop/sslproto.pyx":238
 *                 server_side, server_hostname)
 * 
 *         self._server_side = server_side             # <<<<<<<<<<<<<<
 *         if server_hostname and not server_side:
 *             self._server_hostname = server_hostname
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_server_side); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(25, 238, __pyx_L1_error)
  __pyx_v_self->_server_side = __pyx_t_4;

  /* "uvloop/sslproto.pyx":239
 * 
 *         self._server_side = server_side
 *         if server_hostname and not server_side:             # <<<<<<<<<<<<<<
 *             self._server_hostname = server_hostname
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_server_hostname); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 239, __pyx_L1_error)
  if (__pyx_t_1) {
  } else {
    __pyx_t_4 = __pyx_t_1;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_server_side); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 239, __pyx_L1_error)
  __pyx_t_5 = (!__pyx_t_1);
  __pyx_t_4 = __pyx_t_5;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_4) {

    /* "uvloop/sslproto.pyx":240
 *         self._server_side = server_side
 *         if server_hostname and not server_side:
 *             self._server_hostname = server_hostname             # <<<<<<<<<<<<<<
 *         else:
 *             self._server_hostname = None
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_v_server_hostname))||((__pyx_v_server_hostname) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_server_hostname))) __PYX_ERR(25, 240, __pyx_L1_error)
    __pyx_t_2 = __pyx_v_server_hostname;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_server_hostname);
    __Pyx_DECREF(__pyx_v_self->_server_hostname);
    __pyx_v_self->_server_hostname = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":239
 * 
 *         self._server_side = server_side
 *         if server_hostname and not server_side:             # <<<<<<<<<<<<<<
 *             self._server_hostname = server_hostname
 *         else:
 */
    goto __pyx_L6;
  }

  /* "uvloop/sslproto.pyx":242
 *             self._server_hostname = server_hostname
 *         else:
 *             self._server_hostname = None             # <<<<<<<<<<<<<<
 *         self._sslcontext = sslcontext
 *         # SSL-specific extra info. More info are set when the handshake
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_server_hostname);
    __Pyx_DECREF(__pyx_v_self->_server_hostname);
    __pyx_v_self->_server_hostname = ((PyObject*)Py_None);
  }
  __pyx_L6:;

  /* "uvloop/sslproto.pyx":243
 *         else:
 *             self._server_hostname = None
 *         self._sslcontext = sslcontext             # <<<<<<<<<<<<<<
 *         # SSL-specific extra info. More info are set when the handshake
 *         # completes.
 */
  __Pyx_INCREF(__pyx_v_sslcontext);
  __Pyx_GIVEREF(__pyx_v_sslcontext);
  __Pyx_GOTREF(__pyx_v_self->_sslcontext);
  __Pyx_DECREF(__pyx_v_self->_sslcontext);
  __pyx_v_self->_sslcontext = __pyx_v_sslcontext;

  /* "uvloop/sslproto.pyx":246
 *         # SSL-specific extra info. More info are set when the handshake
 *         # completes.
 *         self._extra = dict(sslcontext=sslcontext)             # <<<<<<<<<<<<<<
 * 
 *         # App data write buffering
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_sslcontext, __pyx_v_sslcontext) < 0) __PYX_ERR(25, 246, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_extra);
  __Pyx_DECREF(__pyx_v_self->_extra);
  __pyx_v_self->_extra = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":249
 * 
 *         # App data write buffering
 *         self._write_backlog = col_deque()             # <<<<<<<<<<<<<<
 *         self._write_buffer_size = 0
 * 
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_col_deque);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_col_deque; __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_write_backlog);
  __Pyx_DECREF(__pyx_v_self->_write_backlog);
  __pyx_v_self->_write_backlog = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":250
 *         # App data write buffering
 *         self._write_backlog = col_deque()
 *         self._write_buffer_size = 0             # <<<<<<<<<<<<<<
 * 
 *         self._waiter = waiter
 */
  __pyx_v_self->_write_buffer_size = 0;

  /* "uvloop/sslproto.pyx":252
 *         self._write_buffer_size = 0
 * 
 *         self._waiter = waiter             # <<<<<<<<<<<<<<
 *         self._loop = loop
 *         self._set_app_protocol(app_protocol)
 */
  __Pyx_INCREF(__pyx_v_waiter);
  __Pyx_GIVEREF(__pyx_v_waiter);
  __Pyx_GOTREF(__pyx_v_self->_waiter);
  __Pyx_DECREF(__pyx_v_self->_waiter);
  __pyx_v_self->_waiter = __pyx_v_waiter;

  /* "uvloop/sslproto.pyx":253
 * 
 *         self._waiter = waiter
 *         self._loop = loop             # <<<<<<<<<<<<<<
 *         self._set_app_protocol(app_protocol)
 *         self._app_transport = None
 */
  if (!(likely(((__pyx_v_loop) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_loop, __pyx_ptype_6uvloop_4loop_Loop))))) __PYX_ERR(25, 253, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_loop;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_loop);
  __pyx_v_self->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":254
 *         self._waiter = waiter
 *         self._loop = loop
 *         self._set_app_protocol(app_protocol)             # <<<<<<<<<<<<<<
 *         self._app_transport = None
 *         self._app_transport_created = False
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_app_protocol(__pyx_v_self, __pyx_v_app_protocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":255
 *         self._loop = loop
 *         self._set_app_protocol(app_protocol)
 *         self._app_transport = None             # <<<<<<<<<<<<<<
 *         self._app_transport_created = False
 *         # transport, ex: SelectorSocketTransport
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_app_transport);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_app_transport);
  __pyx_v_self->_app_transport = ((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)Py_None);

  /* "uvloop/sslproto.pyx":256
 *         self._set_app_protocol(app_protocol)
 *         self._app_transport = None
 *         self._app_transport_created = False             # <<<<<<<<<<<<<<
 *         # transport, ex: SelectorSocketTransport
 *         self._transport = None
 */
  __pyx_v_self->_app_transport_created = 0;

  /* "uvloop/sslproto.pyx":258
 *         self._app_transport_created = False
 *         # transport, ex: SelectorSocketTransport
 *         self._transport = None             # <<<<<<<<<<<<<<
 *         self._ssl_handshake_timeout = ssl_handshake_timeout
 *         self._ssl_shutdown_timeout = ssl_shutdown_timeout
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_transport);
  __Pyx_DECREF(__pyx_v_self->_transport);
  __pyx_v_self->_transport = Py_None;

  /* "uvloop/sslproto.pyx":259
 *         # transport, ex: SelectorSocketTransport
 *         self._transport = None
 *         self._ssl_handshake_timeout = ssl_handshake_timeout             # <<<<<<<<<<<<<<
 *         self._ssl_shutdown_timeout = ssl_shutdown_timeout
 *         # SSL and state machine
 */
  __Pyx_INCREF(__pyx_v_ssl_handshake_timeout);
  __Pyx_GIVEREF(__pyx_v_ssl_handshake_timeout);
  __Pyx_GOTREF(__pyx_v_self->_ssl_handshake_timeout);
  __Pyx_DECREF(__pyx_v_self->_ssl_handshake_timeout);
  __pyx_v_self->_ssl_handshake_timeout = __pyx_v_ssl_handshake_timeout;

  /* "uvloop/sslproto.pyx":260
 *         self._transport = None
 *         self._ssl_handshake_timeout = ssl_handshake_timeout
 *         self._ssl_shutdown_timeout = ssl_shutdown_timeout             # <<<<<<<<<<<<<<
 *         # SSL and state machine
 *         self._sslobj = None
 */
  __Pyx_INCREF(__pyx_v_ssl_shutdown_timeout);
  __Pyx_GIVEREF(__pyx_v_ssl_shutdown_timeout);
  __Pyx_GOTREF(__pyx_v_self->_ssl_shutdown_timeout);
  __Pyx_DECREF(__pyx_v_self->_ssl_shutdown_timeout);
  __pyx_v_self->_ssl_shutdown_timeout = __pyx_v_ssl_shutdown_timeout;

  /* "uvloop/sslproto.pyx":262
 *         self._ssl_shutdown_timeout = ssl_shutdown_timeout
 *         # SSL and state machine
 *         self._sslobj = None             # <<<<<<<<<<<<<<
 *         self._incoming = ssl_MemoryBIO()
 *         self._incoming_write = self._incoming.write
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_sslobj);
  __Pyx_DECREF(__pyx_v_self->_sslobj);
  __pyx_v_self->_sslobj = Py_None;

  /* "uvloop/sslproto.pyx":263
 *         # SSL and state machine
 *         self._sslobj = None
 *         self._incoming = ssl_MemoryBIO()             # <<<<<<<<<<<<<<
 *         self._incoming_write = self._incoming.write
 *         self._outgoing = ssl_MemoryBIO()
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_ssl_MemoryBIO);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_ssl_MemoryBIO; __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_incoming);
  __Pyx_DECREF(__pyx_v_self->_incoming);
  __pyx_v_self->_incoming = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":264
 *         self._sslobj = None
 *         self._incoming = ssl_MemoryBIO()
 *         self._incoming_write = self._incoming.write             # <<<<<<<<<<<<<<
 *         self._outgoing = ssl_MemoryBIO()
 *         self._outgoing_read = self._outgoing.read
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_incoming, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_incoming_write);
  __Pyx_DECREF(__pyx_v_self->_incoming_write);
  __pyx_v_self->_incoming_write = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":265
 *         self._incoming = ssl_MemoryBIO()
 *         self._incoming_write = self._incoming.write
 *         self._outgoing = ssl_MemoryBIO()             # <<<<<<<<<<<<<<
 *         self._outgoing_read = self._outgoing.read
 *         self._state = UNWRAPPED
 */
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_ssl_MemoryBIO);
  __pyx_t_3 = __pyx_v_6uvloop_4loop_ssl_MemoryBIO; __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_outgoing);
  __Pyx_DECREF(__pyx_v_self->_outgoing);
  __pyx_v_self->_outgoing = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":266
 *         self._incoming_write = self._incoming.write
 *         self._outgoing = ssl_MemoryBIO()
 *         self._outgoing_read = self._outgoing.read             # <<<<<<<<<<<<<<
 *         self._state = UNWRAPPED
 *         self._conn_lost = 0  # Set when connection_lost called
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_outgoing, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_outgoing_read);
  __Pyx_DECREF(__pyx_v_self->_outgoing_read);
  __pyx_v_self->_outgoing_read = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":267
 *         self._outgoing = ssl_MemoryBIO()
 *         self._outgoing_read = self._outgoing.read
 *         self._state = UNWRAPPED             # <<<<<<<<<<<<<<
 *         self._conn_lost = 0  # Set when connection_lost called
 *         if call_connection_made:
 */
  __pyx_v_self->_state = __pyx_e_6uvloop_4loop_UNWRAPPED;

  /* "uvloop/sslproto.pyx":268
 *         self._outgoing_read = self._outgoing.read
 *         self._state = UNWRAPPED
 *         self._conn_lost = 0  # Set when connection_lost called             # <<<<<<<<<<<<<<
 *         if call_connection_made:
 *             self._app_state = STATE_INIT
 */
  __pyx_v_self->_conn_lost = 0;

  /* "uvloop/sslproto.pyx":269
 *         self._state = UNWRAPPED
 *         self._conn_lost = 0  # Set when connection_lost called
 *         if call_connection_made:             # <<<<<<<<<<<<<<
 *             self._app_state = STATE_INIT
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_call_connection_made); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(25, 269, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "uvloop/sslproto.pyx":270
 *         self._conn_lost = 0  # Set when connection_lost called
 *         if call_connection_made:
 *             self._app_state = STATE_INIT             # <<<<<<<<<<<<<<
 *         else:
 *             self._app_state = STATE_CON_MADE
 */
    __pyx_v_self->_app_state = __pyx_e_6uvloop_4loop_STATE_INIT;

    /* "uvloop/sslproto.pyx":269
 *         self._state = UNWRAPPED
 *         self._conn_lost = 0  # Set when connection_lost called
 *         if call_connection_made:             # <<<<<<<<<<<<<<
 *             self._app_state = STATE_INIT
 *         else:
 */
    goto __pyx_L9;
  }

  /* "uvloop/sslproto.pyx":272
 *             self._app_state = STATE_INIT
 *         else:
 *             self._app_state = STATE_CON_MADE             # <<<<<<<<<<<<<<
 * 
 *         # Flow Control
 */
  /*else*/ {
    __pyx_v_self->_app_state = __pyx_e_6uvloop_4loop_STATE_CON_MADE;
  }
  __pyx_L9:;

  /* "uvloop/sslproto.pyx":276
 *         # Flow Control
 * 
 *         self._ssl_writing_paused = False             # <<<<<<<<<<<<<<
 * 
 *         self._app_reading_paused = False
 */
  __pyx_v_self->_ssl_writing_paused = 0;

  /* "uvloop/sslproto.pyx":278
 *         self._ssl_writing_paused = False
 * 
 *         self._app_reading_paused = False             # <<<<<<<<<<<<<<
 * 
 *         self._ssl_reading_paused = False
 */
  __pyx_v_self->_app_reading_paused = 0;

  /* "uvloop/sslproto.pyx":280
 *         self._app_reading_paused = False
 * 
 *         self._ssl_reading_paused = False             # <<<<<<<<<<<<<<
 *         self._incoming_high_water = 0
 *         self._incoming_low_water = 0
 */
  __pyx_v_self->_ssl_reading_paused = 0;

  /* "uvloop/sslproto.pyx":281
 * 
 *         self._ssl_reading_paused = False
 *         self._incoming_high_water = 0             # <<<<<<<<<<<<<<
 *         self._incoming_low_water = 0
 *         self._set_read_buffer_limits()
 */
  __pyx_v_self->_incoming_high_water = 0;

  /* "uvloop/sslproto.pyx":282
 *         self._ssl_reading_paused = False
 *         self._incoming_high_water = 0
 *         self._incoming_low_water = 0             # <<<<<<<<<<<<<<
 *         self._set_read_buffer_limits()
 * 
 */
  __pyx_v_self->_incoming_low_water = 0;

  /* "uvloop/sslproto.pyx":283
 *         self._incoming_high_water = 0
 *         self._incoming_low_water = 0
 *         self._set_read_buffer_limits()             # <<<<<<<<<<<<<<
 * 
 *         self._app_writing_paused = False
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_read_buffer_limits(__pyx_v_self, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":285
 *         self._set_read_buffer_limits()
 * 
 *         self._app_writing_paused = False             # <<<<<<<<<<<<<<
 *         self._outgoing_high_water = 0
 *         self._outgoing_low_water = 0
 */
  __pyx_v_self->_app_writing_paused = 0;

  /* "uvloop/sslproto.pyx":286
 * 
 *         self._app_writing_paused = False
 *         self._outgoing_high_water = 0             # <<<<<<<<<<<<<<
 *         self._outgoing_low_water = 0
 *         self._set_write_buffer_limits()
 */
  __pyx_v_self->_outgoing_high_water = 0;

  /* "uvloop/sslproto.pyx":287
 *         self._app_writing_paused = False
 *         self._outgoing_high_water = 0
 *         self._outgoing_low_water = 0             # <<<<<<<<<<<<<<
 *         self._set_write_buffer_limits()
 * 
 */
  __pyx_v_self->_outgoing_low_water = 0;

  /* "uvloop/sslproto.pyx":288
 *         self._outgoing_high_water = 0
 *         self._outgoing_low_water = 0
 *         self._set_write_buffer_limits()             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_app_protocol(self, app_protocol):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_write_buffer_limits(__pyx_v_self, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":216
 *         self._ssl_buffer_len = 0
 * 
 *     def __init__(self, loop, app_protocol, sslcontext, waiter,             # <<<<<<<<<<<<<<
 *                  server_side=False, server_hostname=None,
 *                  call_connection_made=True,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sslcontext);
  __Pyx_XDECREF(__pyx_v_ssl_handshake_timeout);
  __Pyx_XDECREF(__pyx_v_ssl_shutdown_timeout);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":290
 *         self._set_write_buffer_limits()
 * 
 *     cdef _set_app_protocol(self, app_protocol):             # <<<<<<<<<<<<<<
 *         self._app_protocol = app_protocol
 *         if (hasattr(app_protocol, 'get_buffer') and
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_app_protocol(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_app_protocol) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_app_protocol", 1);

  /* "uvloop/sslproto.pyx":291
 * 
 *     cdef _set_app_protocol(self, app_protocol):
 *         self._app_protocol = app_protocol             # <<<<<<<<<<<<<<
 *         if (hasattr(app_protocol, 'get_buffer') and
 *                 not isinstance(app_protocol, aio_Protocol)):
 */
  __Pyx_INCREF(__pyx_v_app_protocol);
  __Pyx_GIVEREF(__pyx_v_app_protocol);
  __Pyx_GOTREF(__pyx_v_self->_app_protocol);
  __Pyx_DECREF(__pyx_v_self->_app_protocol);
  __pyx_v_self->_app_protocol = __pyx_v_app_protocol;

  /* "uvloop/sslproto.pyx":292
 *     cdef _set_app_protocol(self, app_protocol):
 *         self._app_protocol = app_protocol
 *         if (hasattr(app_protocol, 'get_buffer') and             # <<<<<<<<<<<<<<
 *                 not isinstance(app_protocol, aio_Protocol)):
 *             self._app_protocol_get_buffer = app_protocol.get_buffer
 */
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_app_protocol, __pyx_n_u_get_buffer); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(25, 292, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "uvloop/sslproto.pyx":293
 *         self._app_protocol = app_protocol
 *         if (hasattr(app_protocol, 'get_buffer') and
 *                 not isinstance(app_protocol, aio_Protocol)):             # <<<<<<<<<<<<<<
 *             self._app_protocol_get_buffer = app_protocol.get_buffer
 *             self._app_protocol_buffer_updated = app_protocol.buffer_updated
 */
  __pyx_t_3 = __pyx_v_6uvloop_4loop_aio_Protocol;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_app_protocol, __pyx_t_3); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(25, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (!__pyx_t_2);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "uvloop/sslproto.pyx":292
 *     cdef _set_app_protocol(self, app_protocol):
 *         self._app_protocol = app_protocol
 *         if (hasattr(app_protocol, 'get_buffer') and             # <<<<<<<<<<<<<<
 *                 not isinstance(app_protocol, aio_Protocol)):
 *             self._app_protocol_get_buffer = app_protocol.get_buffer
 */
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":294
 *         if (hasattr(app_protocol, 'get_buffer') and
 *                 not isinstance(app_protocol, aio_Protocol)):
 *             self._app_protocol_get_buffer = app_protocol.get_buffer             # <<<<<<<<<<<<<<
 *             self._app_protocol_buffer_updated = app_protocol.buffer_updated
 *             self._app_protocol_is_buffer = True
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_app_protocol, __pyx_n_s_get_buffer); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_app_protocol_get_buffer);
    __Pyx_DECREF(__pyx_v_self->_app_protocol_get_buffer);
    __pyx_v_self->_app_protocol_get_buffer = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":295
 *                 not isinstance(app_protocol, aio_Protocol)):
 *             self._app_protocol_get_buffer = app_protocol.get_buffer
 *             self._app_protocol_buffer_updated = app_protocol.buffer_updated             # <<<<<<<<<<<<<<
 *             self._app_protocol_is_buffer = True
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_app_protocol, __pyx_n_s_buffer_updated); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_app_protocol_buffer_updated);
    __Pyx_DECREF(__pyx_v_self->_app_protocol_buffer_updated);
    __pyx_v_self->_app_protocol_buffer_updated = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":296
 *             self._app_protocol_get_buffer = app_protocol.get_buffer
 *             self._app_protocol_buffer_updated = app_protocol.buffer_updated
 *             self._app_protocol_is_buffer = True             # <<<<<<<<<<<<<<
 *         else:
 *             self._app_protocol_is_buffer = False
 */
    __pyx_v_self->_app_protocol_is_buffer = 1;

    /* "uvloop/sslproto.pyx":292
 *     cdef _set_app_protocol(self, app_protocol):
 *         self._app_protocol = app_protocol
 *         if (hasattr(app_protocol, 'get_buffer') and             # <<<<<<<<<<<<<<
 *                 not isinstance(app_protocol, aio_Protocol)):
 *             self._app_protocol_get_buffer = app_protocol.get_buffer
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":298
 *             self._app_protocol_is_buffer = True
 *         else:
 *             self._app_protocol_is_buffer = False             # <<<<<<<<<<<<<<
 * 
 *     cdef _wakeup_waiter(self, exc=None):
 */
  /*else*/ {
    __pyx_v_self->_app_protocol_is_buffer = 0;
  }
  __pyx_L3:;

  /* "uvloop/sslproto.pyx":290
 *         self._set_write_buffer_limits()
 * 
 *     cdef _set_app_protocol(self, app_protocol):             # <<<<<<<<<<<<<<
 *         self._app_protocol = app_protocol
 *         if (hasattr(app_protocol, 'get_buffer') and
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._set_app_protocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":300
 *             self._app_protocol_is_buffer = False
 * 
 *     cdef _wakeup_waiter(self, exc=None):             # <<<<<<<<<<<<<<
 *         if self._waiter is None:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__wakeup_waiter(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter *__pyx_optional_args) {
  PyObject *__pyx_v_exc = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_wakeup_waiter", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_exc = __pyx_optional_args->exc;
    }
  }

  /* "uvloop/sslproto.pyx":301
 * 
 *     cdef _wakeup_waiter(self, exc=None):
 *         if self._waiter is None:             # <<<<<<<<<<<<<<
 *             return
 *         if not self._waiter.cancelled():
 */
  __pyx_t_1 = (__pyx_v_self->_waiter == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":302
 *     cdef _wakeup_waiter(self, exc=None):
 *         if self._waiter is None:
 *             return             # <<<<<<<<<<<<<<
 *         if not self._waiter.cancelled():
 *             if exc is not None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":301
 * 
 *     cdef _wakeup_waiter(self, exc=None):
 *         if self._waiter is None:             # <<<<<<<<<<<<<<
 *             return
 *         if not self._waiter.cancelled():
 */
  }

  /* "uvloop/sslproto.pyx":303
 *         if self._waiter is None:
 *             return
 *         if not self._waiter.cancelled():             # <<<<<<<<<<<<<<
 *             if exc is not None:
 *                 self._waiter.set_exception(exc)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_cancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (!__pyx_t_1);
  if (__pyx_t_6) {

    /* "uvloop/sslproto.pyx":304
 *             return
 *         if not self._waiter.cancelled():
 *             if exc is not None:             # <<<<<<<<<<<<<<
 *                 self._waiter.set_exception(exc)
 *             else:
 */
    __pyx_t_6 = (__pyx_v_exc != Py_None);
    if (__pyx_t_6) {

      /* "uvloop/sslproto.pyx":305
 *         if not self._waiter.cancelled():
 *             if exc is not None:
 *                 self._waiter.set_exception(exc)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._waiter.set_result(None)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_set_exception); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_exc};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/sslproto.pyx":304
 *             return
 *         if not self._waiter.cancelled():
 *             if exc is not None:             # <<<<<<<<<<<<<<
 *                 self._waiter.set_exception(exc)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "uvloop/sslproto.pyx":307
 *                 self._waiter.set_exception(exc)
 *             else:
 *                 self._waiter.set_result(None)             # <<<<<<<<<<<<<<
 *         self._waiter = None
 * 
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_waiter, __pyx_n_s_set_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, Py_None};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 307, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "uvloop/sslproto.pyx":303
 *         if self._waiter is None:
 *             return
 *         if not self._waiter.cancelled():             # <<<<<<<<<<<<<<
 *             if exc is not None:
 *                 self._waiter.set_exception(exc)
 */
  }

  /* "uvloop/sslproto.pyx":308
 *             else:
 *                 self._waiter.set_result(None)
 *         self._waiter = None             # <<<<<<<<<<<<<<
 * 
 *     def _get_app_transport(self, context=None):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_waiter);
  __Pyx_DECREF(__pyx_v_self->_waiter);
  __pyx_v_self->_waiter = Py_None;

  /* "uvloop/sslproto.pyx":300
 *             self._app_protocol_is_buffer = False
 * 
 *     cdef _wakeup_waiter(self, exc=None):             # <<<<<<<<<<<<<<
 *         if self._waiter is None:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._wakeup_waiter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":310
 *         self._waiter = None
 * 
 *     def _get_app_transport(self, context=None):             # <<<<<<<<<<<<<<
 *         if self._app_transport is None:
 *             if self._app_transport_created:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_7_get_app_transport(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_6_get_app_transport, "SSLProtocol._get_app_transport(self, context=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_7_get_app_transport = {"_get_app_transport", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_7_get_app_transport, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_6_get_app_transport};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_7_get_app_transport(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_context = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_app_transport (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_context,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_context);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 310, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_get_app_transport") < 0)) __PYX_ERR(25, 310, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_context = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_app_transport", 0, 0, 1, __pyx_nargs); __PYX_ERR(25, 310, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._get_app_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_6_get_app_transport(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v_context);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_6_get_app_transport(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_app_transport", 1);

  /* "uvloop/sslproto.pyx":311
 * 
 *     def _get_app_transport(self, context=None):
 *         if self._app_transport is None:             # <<<<<<<<<<<<<<
 *             if self._app_transport_created:
 *                 raise RuntimeError('Creating _SSLProtocolTransport twice')
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->_app_transport) == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":312
 *     def _get_app_transport(self, context=None):
 *         if self._app_transport is None:
 *             if self._app_transport_created:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('Creating _SSLProtocolTransport twice')
 *             self._app_transport = _SSLProtocolTransport(self._loop, self,
 */
    if (unlikely(__pyx_v_self->_app_transport_created)) {

      /* "uvloop/sslproto.pyx":313
 *         if self._app_transport is None:
 *             if self._app_transport_created:
 *                 raise RuntimeError('Creating _SSLProtocolTransport twice')             # <<<<<<<<<<<<<<
 *             self._app_transport = _SSLProtocolTransport(self._loop, self,
 *                                                         context)
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__185, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(25, 313, __pyx_L1_error)

      /* "uvloop/sslproto.pyx":312
 *     def _get_app_transport(self, context=None):
 *         if self._app_transport is None:
 *             if self._app_transport_created:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError('Creating _SSLProtocolTransport twice')
 *             self._app_transport = _SSLProtocolTransport(self._loop, self,
 */
    }

    /* "uvloop/sslproto.pyx":314
 *             if self._app_transport_created:
 *                 raise RuntimeError('Creating _SSLProtocolTransport twice')
 *             self._app_transport = _SSLProtocolTransport(self._loop, self,             # <<<<<<<<<<<<<<
 *                                                         context)
 *             self._app_transport_created = True
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_v_self->_loop);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self->_loop);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->_loop))) __PYX_ERR(25, 314, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(25, 314, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_context);
    __Pyx_GIVEREF(__pyx_v_context);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_context)) __PYX_ERR(25, 314, __pyx_L1_error);
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->_app_transport);
    __Pyx_DECREF((PyObject *)__pyx_v_self->_app_transport);
    __pyx_v_self->_app_transport = ((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":316
 *             self._app_transport = _SSLProtocolTransport(self._loop, self,
 *                                                         context)
 *             self._app_transport_created = True             # <<<<<<<<<<<<<<
 *         return self._app_transport
 * 
 */
    __pyx_v_self->_app_transport_created = 1;

    /* "uvloop/sslproto.pyx":311
 * 
 *     def _get_app_transport(self, context=None):
 *         if self._app_transport is None:             # <<<<<<<<<<<<<<
 *             if self._app_transport_created:
 *                 raise RuntimeError('Creating _SSLProtocolTransport twice')
 */
  }

  /* "uvloop/sslproto.pyx":317
 *                                                         context)
 *             self._app_transport_created = True
 *         return self._app_transport             # <<<<<<<<<<<<<<
 * 
 *     def connection_made(self, transport):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_app_transport);
  __pyx_r = ((PyObject *)__pyx_v_self->_app_transport);
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":310
 *         self._waiter = None
 * 
 *     def _get_app_transport(self, context=None):             # <<<<<<<<<<<<<<
 *         if self._app_transport is None:
 *             if self._app_transport_created:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._get_app_transport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":319
 *         return self._app_transport
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is made.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_9connection_made(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_8connection_made, "SSLProtocol.connection_made(self, transport)\nCalled when the low-level connection is made.\n\n        Start the SSL handshake.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_9connection_made = {"connection_made", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_9connection_made, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_8connection_made};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_9connection_made(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_transport = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connection_made (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_transport,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_transport)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 319, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connection_made") < 0)) __PYX_ERR(25, 319, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_transport = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connection_made", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 319, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_8connection_made(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v_transport);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_8connection_made(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_transport) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connection_made", 1);

  /* "uvloop/sslproto.pyx":324
 *         Start the SSL handshake.
 *         """
 *         self._transport = transport             # <<<<<<<<<<<<<<
 *         self._start_handshake()
 * 
 */
  __Pyx_INCREF(__pyx_v_transport);
  __Pyx_GIVEREF(__pyx_v_transport);
  __Pyx_GOTREF(__pyx_v_self->_transport);
  __Pyx_DECREF(__pyx_v_self->_transport);
  __pyx_v_self->_transport = __pyx_v_transport;

  /* "uvloop/sslproto.pyx":325
 *         """
 *         self._transport = transport
 *         self._start_handshake()             # <<<<<<<<<<<<<<
 * 
 *     def connection_lost(self, exc):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_start_handshake(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":319
 *         return self._app_transport
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is made.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.connection_made", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":327
 *         self._start_handshake()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is lost or closed.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_11connection_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_10connection_lost, "SSLProtocol.connection_lost(self, exc)\nCalled when the low-level connection is lost or closed.\n\n        The argument is an exception object or None (the latter\n        meaning a regular EOF is received or the connection was\n        aborted or closed).\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_11connection_lost = {"connection_lost", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_11connection_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_10connection_lost};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_11connection_lost(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_exc = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connection_lost (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_exc,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exc)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 327, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "connection_lost") < 0)) __PYX_ERR(25, 327, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_exc = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connection_lost", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 327, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_10connection_lost(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v_exc);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_10connection_lost(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connection_lost", 1);

  /* "uvloop/sslproto.pyx":334
 *         aborted or closed).
 *         """
 *         self._write_backlog.clear()             # <<<<<<<<<<<<<<
 *         self._outgoing_read()
 *         self._conn_lost += 1
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_write_backlog, __pyx_n_s_clear); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":335
 *         """
 *         self._write_backlog.clear()
 *         self._outgoing_read()             # <<<<<<<<<<<<<<
 *         self._conn_lost += 1
 * 
 */
  __Pyx_INCREF(__pyx_v_self->_outgoing_read);
  __pyx_t_2 = __pyx_v_self->_outgoing_read; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":336
 *         self._write_backlog.clear()
 *         self._outgoing_read()
 *         self._conn_lost += 1             # <<<<<<<<<<<<<<
 * 
 *         # Just mark the app transport as closed so that its __dealloc__
 */
  __pyx_v_self->_conn_lost = (__pyx_v_self->_conn_lost + 1);

  /* "uvloop/sslproto.pyx":340
 *         # Just mark the app transport as closed so that its __dealloc__
 *         # doesn't complain.
 *         if self._app_transport is not None:             # <<<<<<<<<<<<<<
 *             self._app_transport._closed = True
 * 
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->_app_transport) != Py_None);
  if (__pyx_t_5) {

    /* "uvloop/sslproto.pyx":341
 *         # doesn't complain.
 *         if self._app_transport is not None:
 *             self._app_transport._closed = True             # <<<<<<<<<<<<<<
 * 
 *         if self._state != DO_HANDSHAKE:
 */
    __pyx_v_self->_app_transport->_closed = 1;

    /* "uvloop/sslproto.pyx":340
 *         # Just mark the app transport as closed so that its __dealloc__
 *         # doesn't complain.
 *         if self._app_transport is not None:             # <<<<<<<<<<<<<<
 *             self._app_transport._closed = True
 * 
 */
  }

  /* "uvloop/sslproto.pyx":343
 *             self._app_transport._closed = True
 * 
 *         if self._state != DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             if self._app_state == STATE_CON_MADE or \
 *                     self._app_state == STATE_EOF:
 */
  __pyx_t_5 = (__pyx_v_self->_state != __pyx_e_6uvloop_4loop_DO_HANDSHAKE);
  if (__pyx_t_5) {

    /* "uvloop/sslproto.pyx":344
 * 
 *         if self._state != DO_HANDSHAKE:
 *             if self._app_state == STATE_CON_MADE or \             # <<<<<<<<<<<<<<
 *                     self._app_state == STATE_EOF:
 *                 self._app_state = STATE_CON_LOST
 */
    switch (__pyx_v_self->_app_state) {
      case __pyx_e_6uvloop_4loop_STATE_CON_MADE:
      case __pyx_e_6uvloop_4loop_STATE_EOF:

      /* "uvloop/sslproto.pyx":346
 *             if self._app_state == STATE_CON_MADE or \
 *                     self._app_state == STATE_EOF:
 *                 self._app_state = STATE_CON_LOST             # <<<<<<<<<<<<<<
 *                 self._loop.call_soon(self._app_protocol.connection_lost, exc)
 *         self._set_state(UNWRAPPED)
 */
      __pyx_v_self->_app_state = __pyx_e_6uvloop_4loop_STATE_CON_LOST;

      /* "uvloop/sslproto.pyx":347
 *                     self._app_state == STATE_EOF:
 *                 self._app_state = STATE_CON_LOST
 *                 self._loop.call_soon(self._app_protocol.connection_lost, exc)             # <<<<<<<<<<<<<<
 *         self._set_state(UNWRAPPED)
 *         self._transport = None
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_call_soon); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_connection_lost); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_3, __pyx_v_exc};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 347, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/sslproto.pyx":344
 * 
 *         if self._state != DO_HANDSHAKE:
 *             if self._app_state == STATE_CON_MADE or \             # <<<<<<<<<<<<<<
 *                     self._app_state == STATE_EOF:
 *                 self._app_state = STATE_CON_LOST
 */
      break;
      default: break;
    }

    /* "uvloop/sslproto.pyx":343
 *             self._app_transport._closed = True
 * 
 *         if self._state != DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             if self._app_state == STATE_CON_MADE or \
 *                     self._app_state == STATE_EOF:
 */
  }

  /* "uvloop/sslproto.pyx":348
 *                 self._app_state = STATE_CON_LOST
 *                 self._loop.call_soon(self._app_protocol.connection_lost, exc)
 *         self._set_state(UNWRAPPED)             # <<<<<<<<<<<<<<
 *         self._transport = None
 *         self._app_transport = None
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_state(__pyx_v_self, __pyx_e_6uvloop_4loop_UNWRAPPED); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":349
 *                 self._loop.call_soon(self._app_protocol.connection_lost, exc)
 *         self._set_state(UNWRAPPED)
 *         self._transport = None             # <<<<<<<<<<<<<<
 *         self._app_transport = None
 *         self._app_protocol = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_transport);
  __Pyx_DECREF(__pyx_v_self->_transport);
  __pyx_v_self->_transport = Py_None;

  /* "uvloop/sslproto.pyx":350
 *         self._set_state(UNWRAPPED)
 *         self._transport = None
 *         self._app_transport = None             # <<<<<<<<<<<<<<
 *         self._app_protocol = None
 *         self._wakeup_waiter(exc)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_app_transport);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_app_transport);
  __pyx_v_self->_app_transport = ((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)Py_None);

  /* "uvloop/sslproto.pyx":351
 *         self._transport = None
 *         self._app_transport = None
 *         self._app_protocol = None             # <<<<<<<<<<<<<<
 *         self._wakeup_waiter(exc)
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_app_protocol);
  __Pyx_DECREF(__pyx_v_self->_app_protocol);
  __pyx_v_self->_app_protocol = Py_None;

  /* "uvloop/sslproto.pyx":352
 *         self._app_transport = None
 *         self._app_protocol = None
 *         self._wakeup_waiter(exc)             # <<<<<<<<<<<<<<
 * 
 *         if self._shutdown_timeout_handle:
 */
  __pyx_t_7.__pyx_n = 1;
  __pyx_t_7.exc = __pyx_v_exc;
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_wakeup_waiter(__pyx_v_self, &__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":354
 *         self._wakeup_waiter(exc)
 * 
 *         if self._shutdown_timeout_handle:             # <<<<<<<<<<<<<<
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_self->_shutdown_timeout_handle); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(25, 354, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "uvloop/sslproto.pyx":355
 * 
 *         if self._shutdown_timeout_handle:
 *             self._shutdown_timeout_handle.cancel()             # <<<<<<<<<<<<<<
 *             self._shutdown_timeout_handle = None
 *         if self._handshake_timeout_handle:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_shutdown_timeout_handle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":356
 *         if self._shutdown_timeout_handle:
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None             # <<<<<<<<<<<<<<
 *         if self._handshake_timeout_handle:
 *             self._handshake_timeout_handle.cancel()
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_shutdown_timeout_handle);
    __Pyx_DECREF(__pyx_v_self->_shutdown_timeout_handle);
    __pyx_v_self->_shutdown_timeout_handle = Py_None;

    /* "uvloop/sslproto.pyx":354
 *         self._wakeup_waiter(exc)
 * 
 *         if self._shutdown_timeout_handle:             # <<<<<<<<<<<<<<
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None
 */
  }

  /* "uvloop/sslproto.pyx":357
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None
 *         if self._handshake_timeout_handle:             # <<<<<<<<<<<<<<
 *             self._handshake_timeout_handle.cancel()
 *             self._handshake_timeout_handle = None
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_self->_handshake_timeout_handle); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(25, 357, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "uvloop/sslproto.pyx":358
 *             self._shutdown_timeout_handle = None
 *         if self._handshake_timeout_handle:
 *             self._handshake_timeout_handle.cancel()             # <<<<<<<<<<<<<<
 *             self._handshake_timeout_handle = None
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_handshake_timeout_handle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":359
 *         if self._handshake_timeout_handle:
 *             self._handshake_timeout_handle.cancel()
 *             self._handshake_timeout_handle = None             # <<<<<<<<<<<<<<
 * 
 *     def get_buffer(self, n):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_handshake_timeout_handle);
    __Pyx_DECREF(__pyx_v_self->_handshake_timeout_handle);
    __pyx_v_self->_handshake_timeout_handle = Py_None;

    /* "uvloop/sslproto.pyx":357
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None
 *         if self._handshake_timeout_handle:             # <<<<<<<<<<<<<<
 *             self._handshake_timeout_handle.cancel()
 *             self._handshake_timeout_handle = None
 */
  }

  /* "uvloop/sslproto.pyx":327
 *         self._start_handshake()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is lost or closed.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.connection_lost", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":361
 *             self._handshake_timeout_handle = None
 * 
 *     def get_buffer(self, n):             # <<<<<<<<<<<<<<
 *         cdef size_t want = n
 *         if want > SSL_READ_MAX_SIZE:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_13get_buffer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_12get_buffer, "SSLProtocol.get_buffer(self, n)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_13get_buffer = {"get_buffer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_13get_buffer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_12get_buffer};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_13get_buffer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_n = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_buffer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_n)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 361, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_buffer") < 0)) __PYX_ERR(25, 361, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_n = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_buffer", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 361, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.get_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_12get_buffer(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v_n);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_12get_buffer(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_n) {
  size_t __pyx_v_want;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_buffer", 1);

  /* "uvloop/sslproto.pyx":362
 * 
 *     def get_buffer(self, n):
 *         cdef size_t want = n             # <<<<<<<<<<<<<<
 *         if want > SSL_READ_MAX_SIZE:
 *             want = SSL_READ_MAX_SIZE
 */
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_n); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 362, __pyx_L1_error)
  __pyx_v_want = __pyx_t_1;

  /* "uvloop/sslproto.pyx":363
 *     def get_buffer(self, n):
 *         cdef size_t want = n
 *         if want > SSL_READ_MAX_SIZE:             # <<<<<<<<<<<<<<
 *             want = SSL_READ_MAX_SIZE
 *         if self._ssl_buffer_len < want:
 */
  __pyx_t_2 = (__pyx_v_want > __pyx_e_6uvloop_4loop_SSL_READ_MAX_SIZE);
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":364
 *         cdef size_t want = n
 *         if want > SSL_READ_MAX_SIZE:
 *             want = SSL_READ_MAX_SIZE             # <<<<<<<<<<<<<<
 *         if self._ssl_buffer_len < want:
 *             self._ssl_buffer = <char*>PyMem_RawRealloc(self._ssl_buffer, want)
 */
    __pyx_v_want = __pyx_e_6uvloop_4loop_SSL_READ_MAX_SIZE;

    /* "uvloop/sslproto.pyx":363
 *     def get_buffer(self, n):
 *         cdef size_t want = n
 *         if want > SSL_READ_MAX_SIZE:             # <<<<<<<<<<<<<<
 *             want = SSL_READ_MAX_SIZE
 *         if self._ssl_buffer_len < want:
 */
  }

  /* "uvloop/sslproto.pyx":365
 *         if want > SSL_READ_MAX_SIZE:
 *             want = SSL_READ_MAX_SIZE
 *         if self._ssl_buffer_len < want:             # <<<<<<<<<<<<<<
 *             self._ssl_buffer = <char*>PyMem_RawRealloc(self._ssl_buffer, want)
 *             if not self._ssl_buffer:
 */
  __pyx_t_2 = (__pyx_v_self->_ssl_buffer_len < __pyx_v_want);
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":366
 *             want = SSL_READ_MAX_SIZE
 *         if self._ssl_buffer_len < want:
 *             self._ssl_buffer = <char*>PyMem_RawRealloc(self._ssl_buffer, want)             # <<<<<<<<<<<<<<
 *             if not self._ssl_buffer:
 *                 raise MemoryError()
 */
    __pyx_v_self->_ssl_buffer = ((char *)PyMem_RawRealloc(__pyx_v_self->_ssl_buffer, __pyx_v_want));

    /* "uvloop/sslproto.pyx":367
 *         if self._ssl_buffer_len < want:
 *             self._ssl_buffer = <char*>PyMem_RawRealloc(self._ssl_buffer, want)
 *             if not self._ssl_buffer:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             self._ssl_buffer_len = want
 */
    __pyx_t_2 = (!(__pyx_v_self->_ssl_buffer != 0));
    if (unlikely(__pyx_t_2)) {

      /* "uvloop/sslproto.pyx":368
 *             self._ssl_buffer = <char*>PyMem_RawRealloc(self._ssl_buffer, want)
 *             if not self._ssl_buffer:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 *             self._ssl_buffer_len = want
 *             self._ssl_buffer_view = PyMemoryView_FromMemory(
 */
      PyErr_NoMemory(); __PYX_ERR(25, 368, __pyx_L1_error)

      /* "uvloop/sslproto.pyx":367
 *         if self._ssl_buffer_len < want:
 *             self._ssl_buffer = <char*>PyMem_RawRealloc(self._ssl_buffer, want)
 *             if not self._ssl_buffer:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 *             self._ssl_buffer_len = want
 */
    }

    /* "uvloop/sslproto.pyx":369
 *             if not self._ssl_buffer:
 *                 raise MemoryError()
 *             self._ssl_buffer_len = want             # <<<<<<<<<<<<<<
 *             self._ssl_buffer_view = PyMemoryView_FromMemory(
 *                 self._ssl_buffer, want, PyBUF_WRITE)
 */
    __pyx_v_self->_ssl_buffer_len = __pyx_v_want;

    /* "uvloop/sslproto.pyx":370
 *                 raise MemoryError()
 *             self._ssl_buffer_len = want
 *             self._ssl_buffer_view = PyMemoryView_FromMemory(             # <<<<<<<<<<<<<<
 *                 self._ssl_buffer, want, PyBUF_WRITE)
 *         return self._ssl_buffer_view
 */
    __pyx_t_3 = PyMemoryView_FromMemory(__pyx_v_self->_ssl_buffer, __pyx_v_want, PyBUF_WRITE); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_ssl_buffer_view);
    __Pyx_DECREF(__pyx_v_self->_ssl_buffer_view);
    __pyx_v_self->_ssl_buffer_view = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":365
 *         if want > SSL_READ_MAX_SIZE:
 *             want = SSL_READ_MAX_SIZE
 *         if self._ssl_buffer_len < want:             # <<<<<<<<<<<<<<
 *             self._ssl_buffer = <char*>PyMem_RawRealloc(self._ssl_buffer, want)
 *             if not self._ssl_buffer:
 */
  }

  /* "uvloop/sslproto.pyx":372
 *             self._ssl_buffer_view = PyMemoryView_FromMemory(
 *                 self._ssl_buffer, want, PyBUF_WRITE)
 *         return self._ssl_buffer_view             # <<<<<<<<<<<<<<
 * 
 *     def buffer_updated(self, nbytes):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_ssl_buffer_view);
  __pyx_r = __pyx_v_self->_ssl_buffer_view;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":361
 *             self._handshake_timeout_handle = None
 * 
 *     def get_buffer(self, n):             # <<<<<<<<<<<<<<
 *         cdef size_t want = n
 *         if want > SSL_READ_MAX_SIZE:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.get_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":374
 *         return self._ssl_buffer_view
 * 
 *     def buffer_updated(self, nbytes):             # <<<<<<<<<<<<<<
 *         self._incoming_write(PyMemoryView_FromMemory(
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_15buffer_updated(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_14buffer_updated, "SSLProtocol.buffer_updated(self, nbytes)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_15buffer_updated = {"buffer_updated", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_15buffer_updated, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_14buffer_updated};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_15buffer_updated(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_nbytes = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("buffer_updated (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nbytes,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_nbytes)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 374, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "buffer_updated") < 0)) __PYX_ERR(25, 374, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_nbytes = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("buffer_updated", 1, 1, 1, __pyx_nargs); __PYX_ERR(25, 374, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.buffer_updated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_14buffer_updated(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v_nbytes);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_14buffer_updated(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_nbytes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("buffer_updated", 1);

  /* "uvloop/sslproto.pyx":376
 *     def buffer_updated(self, nbytes):
 *         self._incoming_write(PyMemoryView_FromMemory(
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))             # <<<<<<<<<<<<<<
 * 
 *         if self._state == DO_HANDSHAKE:
 */
  __pyx_t_2 = PyInt_AsSsize_t(__pyx_v_nbytes); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 376, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":375
 * 
 *     def buffer_updated(self, nbytes):
 *         self._incoming_write(PyMemoryView_FromMemory(             # <<<<<<<<<<<<<<
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))
 * 
 */
  __pyx_t_3 = PyMemoryView_FromMemory(__pyx_v_self->_ssl_buffer, __pyx_t_2, PyBUF_WRITE); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->_incoming_write);
  __pyx_t_4 = __pyx_v_self->_incoming_write; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":378
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))
 * 
 *         if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             self._do_handshake()
 * 
 */
  switch (__pyx_v_self->_state) {
    case __pyx_e_6uvloop_4loop_DO_HANDSHAKE:

    /* "uvloop/sslproto.pyx":379
 * 
 *         if self._state == DO_HANDSHAKE:
 *             self._do_handshake()             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == WRAPPED:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_handshake(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":378
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))
 * 
 *         if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             self._do_handshake()
 * 
 */
    break;
    case __pyx_e_6uvloop_4loop_WRAPPED:

    /* "uvloop/sslproto.pyx":382
 * 
 *         elif self._state == WRAPPED:
 *             self._do_read()             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == FLUSHING:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":381
 *             self._do_handshake()
 * 
 *         elif self._state == WRAPPED:             # <<<<<<<<<<<<<<
 *             self._do_read()
 * 
 */
    break;
    case __pyx_e_6uvloop_4loop_FLUSHING:

    /* "uvloop/sslproto.pyx":385
 * 
 *         elif self._state == FLUSHING:
 *             self._do_flush()             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == SHUTDOWN:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_flush(__pyx_v_self, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":384
 *             self._do_read()
 * 
 *         elif self._state == FLUSHING:             # <<<<<<<<<<<<<<
 *             self._do_flush()
 * 
 */
    break;
    case __pyx_e_6uvloop_4loop_SHUTDOWN:

    /* "uvloop/sslproto.pyx":388
 * 
 *         elif self._state == SHUTDOWN:
 *             self._do_shutdown()             # <<<<<<<<<<<<<<
 * 
 *     def eof_received(self):
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_shutdown(__pyx_v_self, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":387
 *             self._do_flush()
 * 
 *         elif self._state == SHUTDOWN:             # <<<<<<<<<<<<<<
 *             self._do_shutdown()
 * 
 */
    break;
    default: break;
  }

  /* "uvloop/sslproto.pyx":374
 *         return self._ssl_buffer_view
 * 
 *     def buffer_updated(self, nbytes):             # <<<<<<<<<<<<<<
 *         self._incoming_write(PyMemoryView_FromMemory(
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.buffer_updated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":390
 *             self._do_shutdown()
 * 
 *     def eof_received(self):             # <<<<<<<<<<<<<<
 *         """Called when the other end of the low-level stream
 *         is half-closed.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_17eof_received(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_16eof_received, "SSLProtocol.eof_received(self)\nCalled when the other end of the low-level stream\n        is half-closed.\n\n        If this returns a false value (including None), the transport\n        will close itself.  If it returns a true value, closing the\n        transport is up to the protocol.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_17eof_received = {"eof_received", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_17eof_received, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_16eof_received};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_17eof_received(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("eof_received (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("eof_received", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "eof_received", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_16eof_received(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_16eof_received(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("eof_received", 1);

  /* "uvloop/sslproto.pyx":398
 *         transport is up to the protocol.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r received EOF", self)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":399
 *         """
 *         try:
 *             if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *                 aio_logger.debug("%r received EOF", self)
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_get_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 399, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 399, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(25, 399, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_8) {

        /* "uvloop/sslproto.pyx":400
 *         try:
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r received EOF", self)             # <<<<<<<<<<<<<<
 * 
 *             if self._state == DO_HANDSHAKE:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 400, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_kp_u_r_received_EOF, ((PyObject *)__pyx_v_self)};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 400, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/sslproto.pyx":399
 *         """
 *         try:
 *             if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *                 aio_logger.debug("%r received EOF", self)
 * 
 */
      }

      /* "uvloop/sslproto.pyx":402
 *                 aio_logger.debug("%r received EOF", self)
 * 
 *             if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *                 self._on_handshake_complete(ConnectionResetError)
 * 
 */
      switch (__pyx_v_self->_state) {
        case __pyx_e_6uvloop_4loop_DO_HANDSHAKE:

        /* "uvloop/sslproto.pyx":403
 * 
 *             if self._state == DO_HANDSHAKE:
 *                 self._on_handshake_complete(ConnectionResetError)             # <<<<<<<<<<<<<<
 * 
 *             elif self._state == WRAPPED or self._state == FLUSHING:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ConnectionResetError); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 403, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_handshake_complete(__pyx_v_self, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 403, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/sslproto.pyx":402
 *                 aio_logger.debug("%r received EOF", self)
 * 
 *             if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *                 self._on_handshake_complete(ConnectionResetError)
 * 
 */
        break;
        case __pyx_e_6uvloop_4loop_WRAPPED:

        /* "uvloop/sslproto.pyx":405
 *                 self._on_handshake_complete(ConnectionResetError)
 * 
 *             elif self._state == WRAPPED or self._state == FLUSHING:             # <<<<<<<<<<<<<<
 *                 # We treat a low-level EOF as a critical situation similar to a
 *                 # broken connection - just send whatever is in the buffer and
 */
        case __pyx_e_6uvloop_4loop_FLUSHING:

        /* "uvloop/sslproto.pyx":411
 *                 # because we don't want the user to think that this is a
 *                 # graceful shutdown triggered by SSL "close_notify".
 *                 self._set_state(SHUTDOWN)             # <<<<<<<<<<<<<<
 *                 self._on_shutdown_complete(None)
 * 
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_state(__pyx_v_self, __pyx_e_6uvloop_4loop_SHUTDOWN); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 411, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/sslproto.pyx":412
 *                 # graceful shutdown triggered by SSL "close_notify".
 *                 self._set_state(SHUTDOWN)
 *                 self._on_shutdown_complete(None)             # <<<<<<<<<<<<<<
 * 
 *             elif self._state == SHUTDOWN:
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_shutdown_complete(__pyx_v_self, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 412, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/sslproto.pyx":405
 *                 self._on_handshake_complete(ConnectionResetError)
 * 
 *             elif self._state == WRAPPED or self._state == FLUSHING:             # <<<<<<<<<<<<<<
 *                 # We treat a low-level EOF as a critical situation similar to a
 *                 # broken connection - just send whatever is in the buffer and
 */
        break;
        case __pyx_e_6uvloop_4loop_SHUTDOWN:

        /* "uvloop/sslproto.pyx":415
 * 
 *             elif self._state == SHUTDOWN:
 *                 self._on_shutdown_complete(None)             # <<<<<<<<<<<<<<
 * 
 *         except Exception:
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_shutdown_complete(__pyx_v_self, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 415, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/sslproto.pyx":414
 *                 self._on_shutdown_complete(None)
 * 
 *             elif self._state == SHUTDOWN:             # <<<<<<<<<<<<<<
 *                 self._on_shutdown_complete(None)
 * 
 */
        break;
        default: break;
      }

      /* "uvloop/sslproto.pyx":398
 *         transport is up to the protocol.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r received EOF", self)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/sslproto.pyx":417
 *                 self._on_shutdown_complete(None)
 * 
 *         except Exception:             # <<<<<<<<<<<<<<
 *             self._transport.close()
 *             raise
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol.eof_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_6) < 0) __PYX_ERR(25, 417, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);

      /* "uvloop/sslproto.pyx":418
 * 
 *         except Exception:
 *             self._transport.close()             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_close); if (unlikely(!__pyx_t_11)) __PYX_ERR(25, 418, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
        __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 418, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "uvloop/sslproto.pyx":419
 *         except Exception:
 *             self._transport.close()
 *             raise             # <<<<<<<<<<<<<<
 * 
 *     cdef _get_extra_info(self, name, default=None):
 */
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_4, __pyx_t_6);
      __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_6 = 0; 
      __PYX_ERR(25, 419, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "uvloop/sslproto.pyx":398
 *         transport is up to the protocol.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r received EOF", self)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "uvloop/sslproto.pyx":390
 *             self._do_shutdown()
 * 
 *     def eof_received(self):             # <<<<<<<<<<<<<<
 *         """Called when the other end of the low-level stream
 *         is half-closed.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.eof_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":421
 *             raise
 * 
 *     cdef _get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if name == 'uvloop.sslproto':
 *             return self
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__get_extra_info(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_name, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__get_extra_info *__pyx_optional_args) {
  PyObject *__pyx_v_default = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_extra_info", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default = __pyx_optional_args->__pyx_default;
    }
  }

  /* "uvloop/sslproto.pyx":422
 * 
 *     cdef _get_extra_info(self, name, default=None):
 *         if name == 'uvloop.sslproto':             # <<<<<<<<<<<<<<
 *             return self
 *         elif name in self._extra:
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_name, __pyx_kp_u_uvloop_sslproto, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":423
 *     cdef _get_extra_info(self, name, default=None):
 *         if name == 'uvloop.sslproto':
 *             return self             # <<<<<<<<<<<<<<
 *         elif name in self._extra:
 *             return self._extra[name]
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":422
 * 
 *     cdef _get_extra_info(self, name, default=None):
 *         if name == 'uvloop.sslproto':             # <<<<<<<<<<<<<<
 *             return self
 *         elif name in self._extra:
 */
  }

  /* "uvloop/sslproto.pyx":424
 *         if name == 'uvloop.sslproto':
 *             return self
 *         elif name in self._extra:             # <<<<<<<<<<<<<<
 *             return self._extra[name]
 *         elif self._transport is not None:
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_v_self->_extra, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 424, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":425
 *             return self
 *         elif name in self._extra:
 *             return self._extra[name]             # <<<<<<<<<<<<<<
 *         elif self._transport is not None:
 *             return self._transport.get_extra_info(name, default)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_self->_extra, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":424
 *         if name == 'uvloop.sslproto':
 *             return self
 *         elif name in self._extra:             # <<<<<<<<<<<<<<
 *             return self._extra[name]
 *         elif self._transport is not None:
 */
  }

  /* "uvloop/sslproto.pyx":426
 *         elif name in self._extra:
 *             return self._extra[name]
 *         elif self._transport is not None:             # <<<<<<<<<<<<<<
 *             return self._transport.get_extra_info(name, default)
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->_transport != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":427
 *             return self._extra[name]
 *         elif self._transport is not None:
 *             return self._transport.get_extra_info(name, default)             # <<<<<<<<<<<<<<
 *         else:
 *             return default
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_get_extra_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_name, __pyx_v_default};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":426
 *         elif name in self._extra:
 *             return self._extra[name]
 *         elif self._transport is not None:             # <<<<<<<<<<<<<<
 *             return self._transport.get_extra_info(name, default)
 *         else:
 */
  }

  /* "uvloop/sslproto.pyx":429
 *             return self._transport.get_extra_info(name, default)
 *         else:
 *             return default             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_state(self, SSLProtocolState new_state):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_default);
    __pyx_r = __pyx_v_default;
    goto __pyx_L0;
  }

  /* "uvloop/sslproto.pyx":421
 *             raise
 * 
 *     cdef _get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if name == 'uvloop.sslproto':
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._get_extra_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":431
 *             return default
 * 
 *     cdef _set_state(self, SSLProtocolState new_state):             # <<<<<<<<<<<<<<
 *         cdef bint allowed = False
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_state(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, enum __pyx_t_6uvloop_4loop_SSLProtocolState __pyx_v_new_state) {
  int __pyx_v_allowed;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_state", 1);

  /* "uvloop/sslproto.pyx":432
 * 
 *     cdef _set_state(self, SSLProtocolState new_state):
 *         cdef bint allowed = False             # <<<<<<<<<<<<<<
 * 
 *         if new_state == UNWRAPPED:
 */
  __pyx_v_allowed = 0;

  /* "uvloop/sslproto.pyx":434
 *         cdef bint allowed = False
 * 
 *         if new_state == UNWRAPPED:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
  __pyx_t_1 = (__pyx_v_new_state == __pyx_e_6uvloop_4loop_UNWRAPPED);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":435
 * 
 *         if new_state == UNWRAPPED:
 *             allowed = True             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == UNWRAPPED and new_state == DO_HANDSHAKE:
 */
    __pyx_v_allowed = 1;

    /* "uvloop/sslproto.pyx":434
 *         cdef bint allowed = False
 * 
 *         if new_state == UNWRAPPED:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":437
 *             allowed = True
 * 
 *         elif self._state == UNWRAPPED and new_state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_UNWRAPPED);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_new_state == __pyx_e_6uvloop_4loop_DO_HANDSHAKE);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":438
 * 
 *         elif self._state == UNWRAPPED and new_state == DO_HANDSHAKE:
 *             allowed = True             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == DO_HANDSHAKE and new_state == WRAPPED:
 */
    __pyx_v_allowed = 1;

    /* "uvloop/sslproto.pyx":437
 *             allowed = True
 * 
 *         elif self._state == UNWRAPPED and new_state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":440
 *             allowed = True
 * 
 *         elif self._state == DO_HANDSHAKE and new_state == WRAPPED:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_DO_HANDSHAKE);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_new_state == __pyx_e_6uvloop_4loop_WRAPPED);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":441
 * 
 *         elif self._state == DO_HANDSHAKE and new_state == WRAPPED:
 *             allowed = True             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == WRAPPED and new_state == FLUSHING:
 */
    __pyx_v_allowed = 1;

    /* "uvloop/sslproto.pyx":440
 *             allowed = True
 * 
 *         elif self._state == DO_HANDSHAKE and new_state == WRAPPED:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":443
 *             allowed = True
 * 
 *         elif self._state == WRAPPED and new_state == FLUSHING:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_WRAPPED);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_new_state == __pyx_e_6uvloop_4loop_FLUSHING);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":444
 * 
 *         elif self._state == WRAPPED and new_state == FLUSHING:
 *             allowed = True             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == WRAPPED and new_state == SHUTDOWN:
 */
    __pyx_v_allowed = 1;

    /* "uvloop/sslproto.pyx":443
 *             allowed = True
 * 
 *         elif self._state == WRAPPED and new_state == FLUSHING:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":446
 *             allowed = True
 * 
 *         elif self._state == WRAPPED and new_state == SHUTDOWN:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_WRAPPED);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_new_state == __pyx_e_6uvloop_4loop_SHUTDOWN);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":447
 * 
 *         elif self._state == WRAPPED and new_state == SHUTDOWN:
 *             allowed = True             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == FLUSHING and new_state == SHUTDOWN:
 */
    __pyx_v_allowed = 1;

    /* "uvloop/sslproto.pyx":446
 *             allowed = True
 * 
 *         elif self._state == WRAPPED and new_state == SHUTDOWN:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":449
 *             allowed = True
 * 
 *         elif self._state == FLUSHING and new_state == SHUTDOWN:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_FLUSHING);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_new_state == __pyx_e_6uvloop_4loop_SHUTDOWN);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":450
 * 
 *         elif self._state == FLUSHING and new_state == SHUTDOWN:
 *             allowed = True             # <<<<<<<<<<<<<<
 * 
 *         if allowed:
 */
    __pyx_v_allowed = 1;

    /* "uvloop/sslproto.pyx":449
 *             allowed = True
 * 
 *         elif self._state == FLUSHING and new_state == SHUTDOWN:             # <<<<<<<<<<<<<<
 *             allowed = True
 * 
 */
  }
  __pyx_L3:;

  /* "uvloop/sslproto.pyx":452
 *             allowed = True
 * 
 *         if allowed:             # <<<<<<<<<<<<<<
 *             self._state = new_state
 * 
 */
  if (likely(__pyx_v_allowed)) {

    /* "uvloop/sslproto.pyx":453
 * 
 *         if allowed:
 *             self._state = new_state             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    __pyx_v_self->_state = __pyx_v_new_state;

    /* "uvloop/sslproto.pyx":452
 *             allowed = True
 * 
 *         if allowed:             # <<<<<<<<<<<<<<
 *             self._state = new_state
 * 
 */
    goto __pyx_L14;
  }

  /* "uvloop/sslproto.pyx":456
 * 
 *         else:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'cannot switch state from {} to {}'.format(
 *                     self._state, new_state))
 */
  /*else*/ {

    /* "uvloop/sslproto.pyx":457
 *         else:
 *             raise RuntimeError(
 *                 'cannot switch state from {} to {}'.format(             # <<<<<<<<<<<<<<
 *                     self._state, new_state))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_cannot_switch_state_from_to, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uvloop/sslproto.pyx":458
 *             raise RuntimeError(
 *                 'cannot switch state from {} to {}'.format(
 *                     self._state, new_state))             # <<<<<<<<<<<<<<
 * 
 *     # Handshake flow
 */
    __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_6uvloop_4loop_SSLProtocolState(__pyx_v_self->_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_enum____pyx_t_6uvloop_4loop_SSLProtocolState(__pyx_v_new_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_7, __pyx_t_5, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 457, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "uvloop/sslproto.pyx":456
 * 
 *         else:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'cannot switch state from {} to {}'.format(
 *                     self._state, new_state))
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(25, 456, __pyx_L1_error)
  }
  __pyx_L14:;

  /* "uvloop/sslproto.pyx":431
 *             return default
 * 
 *     cdef _set_state(self, SSLProtocolState new_state):             # <<<<<<<<<<<<<<
 *         cdef bint allowed = False
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":474
 *         self._handshake_timeout_handle = \
 *             self._loop.call_later(self._ssl_handshake_timeout,
 *                                   lambda: self._check_handshake_timeout())             # <<<<<<<<<<<<<<
 * 
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_16_start_handshake_lambda4(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_16_start_handshake_lambda4 = {"lambda4", (PyCFunction)__pyx_pw_6uvloop_4loop_11SSLProtocol_16_start_handshake_lambda4, METH_NOARGS, 0};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_16_start_handshake_lambda4(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda4 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_lambda_funcdef_lambda4(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda4(PyObject *__pyx_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *__pyx_cur_scope;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda4", 1);
  __pyx_outer_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(25, 474, __pyx_L1_error) }
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_check_handshake_timeout(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._start_handshake.lambda4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":462
 *     # Handshake flow
 * 
 *     cdef _start_handshake(self):             # <<<<<<<<<<<<<<
 *         if self._loop.get_debug():
 *             aio_logger.debug("%r starts SSL handshake", self)
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__start_handshake(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *__pyx_cur_scope;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_start_handshake", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_31__start_handshake(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(25, 462, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "uvloop/sslproto.pyx":463
 * 
 *     cdef _start_handshake(self):
 *         if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *             aio_logger.debug("%r starts SSL handshake", self)
 *             self._handshake_start_time = self._loop.time()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->_loop), __pyx_n_s_get_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(25, 463, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "uvloop/sslproto.pyx":464
 *     cdef _start_handshake(self):
 *         if self._loop.get_debug():
 *             aio_logger.debug("%r starts SSL handshake", self)             # <<<<<<<<<<<<<<
 *             self._handshake_start_time = self._loop.time()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_kp_u_r_starts_SSL_handshake, ((PyObject *)__pyx_cur_scope->__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":465
 *         if self._loop.get_debug():
 *             aio_logger.debug("%r starts SSL handshake", self)
 *             self._handshake_start_time = self._loop.time()             # <<<<<<<<<<<<<<
 *         else:
 *             self._handshake_start_time = None
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->_loop), __pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 465, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_handshake_start_time);
    __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_handshake_start_time);
    __pyx_cur_scope->__pyx_v_self->_handshake_start_time = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":463
 * 
 *     cdef _start_handshake(self):
 *         if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *             aio_logger.debug("%r starts SSL handshake", self)
 *             self._handshake_start_time = self._loop.time()
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":467
 *             self._handshake_start_time = self._loop.time()
 *         else:
 *             self._handshake_start_time = None             # <<<<<<<<<<<<<<
 * 
 *         self._set_state(DO_HANDSHAKE)
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_handshake_start_time);
    __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_handshake_start_time);
    __pyx_cur_scope->__pyx_v_self->_handshake_start_time = Py_None;
  }
  __pyx_L3:;

  /* "uvloop/sslproto.pyx":469
 *             self._handshake_start_time = None
 * 
 *         self._set_state(DO_HANDSHAKE)             # <<<<<<<<<<<<<<
 * 
 *         # start handshake timeout count down
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_set_state(__pyx_cur_scope->__pyx_v_self, __pyx_e_6uvloop_4loop_DO_HANDSHAKE); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":473
 *         # start handshake timeout count down
 *         self._handshake_timeout_handle = \
 *             self._loop.call_later(self._ssl_handshake_timeout,             # <<<<<<<<<<<<<<
 *                                   lambda: self._check_handshake_timeout())
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->_loop), __pyx_n_s_call_later); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/sslproto.pyx":474
 *         self._handshake_timeout_handle = \
 *             self._loop.call_later(self._ssl_handshake_timeout,
 *                                   lambda: self._check_handshake_timeout())             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_16_start_handshake_lambda4, 0, __pyx_n_s_SSLProtocol__start_handshake_loc, ((PyObject*)__pyx_cur_scope), __pyx_n_s_uvloop_loop, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_self->_ssl_handshake_timeout, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/sslproto.pyx":472
 * 
 *         # start handshake timeout count down
 *         self._handshake_timeout_handle = \             # <<<<<<<<<<<<<<
 *             self._loop.call_later(self._ssl_handshake_timeout,
 *                                   lambda: self._check_handshake_timeout())
 */
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_handshake_timeout_handle);
  __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_handshake_timeout_handle);
  __pyx_cur_scope->__pyx_v_self->_handshake_timeout_handle = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":476
 *                                   lambda: self._check_handshake_timeout())
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._sslobj = self._sslcontext.wrap_bio(
 *                 self._incoming, self._outgoing,
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":477
 * 
 *         try:
 *             self._sslobj = self._sslcontext.wrap_bio(             # <<<<<<<<<<<<<<
 *                 self._incoming, self._outgoing,
 *                 server_side=self._server_side,
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self->_sslcontext, __pyx_n_s_wrap_bio); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 477, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "uvloop/sslproto.pyx":478
 *         try:
 *             self._sslobj = self._sslcontext.wrap_bio(
 *                 self._incoming, self._outgoing,             # <<<<<<<<<<<<<<
 *                 server_side=self._server_side,
 *                 server_hostname=self._server_hostname)
 */
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 477, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self->_incoming);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self->_incoming);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_cur_scope->__pyx_v_self->_incoming)) __PYX_ERR(25, 477, __pyx_L4_error);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self->_outgoing);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self->_outgoing);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_self->_outgoing)) __PYX_ERR(25, 477, __pyx_L4_error);

      /* "uvloop/sslproto.pyx":479
 *             self._sslobj = self._sslcontext.wrap_bio(
 *                 self._incoming, self._outgoing,
 *                 server_side=self._server_side,             # <<<<<<<<<<<<<<
 *                 server_hostname=self._server_hostname)
 *             self._sslobj_read = self._sslobj.read
 */
      __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 479, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_v_self->_server_side); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 479, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_server_side, __pyx_t_6) < 0) __PYX_ERR(25, 479, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/sslproto.pyx":480
 *                 self._incoming, self._outgoing,
 *                 server_side=self._server_side,
 *                 server_hostname=self._server_hostname)             # <<<<<<<<<<<<<<
 *             self._sslobj_read = self._sslobj.read
 *             self._sslobj_write = self._sslobj.write
 */
      if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_server_hostname, __pyx_cur_scope->__pyx_v_self->_server_hostname) < 0) __PYX_ERR(25, 479, __pyx_L4_error)

      /* "uvloop/sslproto.pyx":477
 * 
 *         try:
 *             self._sslobj = self._sslcontext.wrap_bio(             # <<<<<<<<<<<<<<
 *                 self._incoming, self._outgoing,
 *                 server_side=self._server_side,
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 477, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_sslobj);
      __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_sslobj);
      __pyx_cur_scope->__pyx_v_self->_sslobj = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "uvloop/sslproto.pyx":481
 *                 server_side=self._server_side,
 *                 server_hostname=self._server_hostname)
 *             self._sslobj_read = self._sslobj.read             # <<<<<<<<<<<<<<
 *             self._sslobj_write = self._sslobj.write
 *         except Exception as ex:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self->_sslobj, __pyx_n_s_read); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 481, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_sslobj_read);
      __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_sslobj_read);
      __pyx_cur_scope->__pyx_v_self->_sslobj_read = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "uvloop/sslproto.pyx":482
 *                 server_hostname=self._server_hostname)
 *             self._sslobj_read = self._sslobj.read
 *             self._sslobj_write = self._sslobj.write             # <<<<<<<<<<<<<<
 *         except Exception as ex:
 *             self._on_handshake_complete(ex)
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self->_sslobj, __pyx_n_s_write); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 482, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_sslobj_write);
      __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_sslobj_write);
      __pyx_cur_scope->__pyx_v_self->_sslobj_write = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "uvloop/sslproto.pyx":476
 *                                   lambda: self._check_handshake_timeout())
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._sslobj = self._sslcontext.wrap_bio(
 *                 self._incoming, self._outgoing,
 */
    }

    /* "uvloop/sslproto.pyx":486
 *             self._on_handshake_complete(ex)
 *         else:
 *             self._do_handshake()             # <<<<<<<<<<<<<<
 * 
 *     cdef _check_handshake_timeout(self):
 */
    /*else:*/ {
      __pyx_t_6 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_do_handshake(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 486, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/sslproto.pyx":483
 *             self._sslobj_read = self._sslobj.read
 *             self._sslobj_write = self._sslobj.write
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._on_handshake_complete(ex)
 *         else:
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_10) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._start_handshake", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(25, 483, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_ex = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":484
 *             self._sslobj_write = self._sslobj.write
 *         except Exception as ex:
 *             self._on_handshake_complete(ex)             # <<<<<<<<<<<<<<
 *         else:
 *             self._do_handshake()
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_on_handshake_complete(__pyx_cur_scope->__pyx_v_self, __pyx_v_ex); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 484, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }

      /* "uvloop/sslproto.pyx":483
 *             self._sslobj_read = self._sslobj.read
 *             self._sslobj_write = self._sslobj.write
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._on_handshake_complete(ex)
 *         else:
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_10 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L6_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/sslproto.pyx":476
 *                                   lambda: self._check_handshake_timeout())
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             self._sslobj = self._sslcontext.wrap_bio(
 *                 self._incoming, self._outgoing,
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    __pyx_L9_try_end:;
  }

  /* "uvloop/sslproto.pyx":462
 *     # Handshake flow
 * 
 *     cdef _start_handshake(self):             # <<<<<<<<<<<<<<
 *         if self._loop.get_debug():
 *             aio_logger.debug("%r starts SSL handshake", self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._start_handshake", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":488
 *             self._do_handshake()
 * 
 *     cdef _check_handshake_timeout(self):             # <<<<<<<<<<<<<<
 *         if self._state == DO_HANDSHAKE:
 *             msg = (
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__check_handshake_timeout(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_handshake_timeout", 1);

  /* "uvloop/sslproto.pyx":489
 * 
 *     cdef _check_handshake_timeout(self):
 *         if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             msg = (
 *                 f"SSL handshake is taking longer than "
 */
  __pyx_t_1 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_DO_HANDSHAKE);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":491
 *         if self._state == DO_HANDSHAKE:
 *             msg = (
 *                 f"SSL handshake is taking longer than "             # <<<<<<<<<<<<<<
 *                 f"{self._ssl_handshake_timeout} seconds: "
 *                 f"aborting the connection"
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_SSL_handshake_is_taking_longer_t);
    __pyx_t_3 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_SSL_handshake_is_taking_longer_t);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_SSL_handshake_is_taking_longer_t);

    /* "uvloop/sslproto.pyx":492
 *             msg = (
 *                 f"SSL handshake is taking longer than "
 *                 f"{self._ssl_handshake_timeout} seconds: "             # <<<<<<<<<<<<<<
 *                 f"aborting the connection"
 *             )
 */
    __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_v_self->_ssl_handshake_timeout, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_seconds_aborting_the_connection);
    __pyx_t_3 += 33;
    __Pyx_GIVEREF(__pyx_kp_u_seconds_aborting_the_connection);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_seconds_aborting_the_connection);

    /* "uvloop/sslproto.pyx":491
 *         if self._state == DO_HANDSHAKE:
 *             msg = (
 *                 f"SSL handshake is taking longer than "             # <<<<<<<<<<<<<<
 *                 f"{self._ssl_handshake_timeout} seconds: "
 *                 f"aborting the connection"
 */
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_msg = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "uvloop/sslproto.pyx":495
 *                 f"aborting the connection"
 *             )
 *             self._fatal_error(ConnectionAbortedError(msg))             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_handshake(self):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_ConnectionAbortedError); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_msg); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_fatal_error(__pyx_v_self, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/sslproto.pyx":489
 * 
 *     cdef _check_handshake_timeout(self):
 *         if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             msg = (
 *                 f"SSL handshake is taking longer than "
 */
  }

  /* "uvloop/sslproto.pyx":488
 *             self._do_handshake()
 * 
 *     cdef _check_handshake_timeout(self):             # <<<<<<<<<<<<<<
 *         if self._state == DO_HANDSHAKE:
 *             msg = (
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._check_handshake_timeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":497
 *             self._fatal_error(ConnectionAbortedError(msg))
 * 
 *     cdef _do_handshake(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self._sslobj.do_handshake()
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_handshake(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  char const *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  char const *__pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_handshake", 1);

  /* "uvloop/sslproto.pyx":498
 * 
 *     cdef _do_handshake(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self._sslobj.do_handshake()
 *         except ssl_SSLAgainErrors as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":499
 *     cdef _do_handshake(self):
 *         try:
 *             self._sslobj.do_handshake()             # <<<<<<<<<<<<<<
 *         except ssl_SSLAgainErrors as exc:
 *             self._process_outgoing()
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_sslobj, __pyx_n_s_do_handshake); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 499, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 499, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/sslproto.pyx":498
 * 
 *     cdef _do_handshake(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self._sslobj.do_handshake()
 *         except ssl_SSLAgainErrors as exc:
 */
    }

    /* "uvloop/sslproto.pyx":505
 *             self._on_handshake_complete(exc)
 *         else:
 *             self._on_handshake_complete(None)             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_handshake_complete(self, handshake_exc):
 */
    /*else:*/ {
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_handshake_complete(__pyx_v_self, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 505, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/sslproto.pyx":500
 *         try:
 *             self._sslobj.do_handshake()
 *         except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *             self._process_outgoing()
 *         except ssl_SSLError as exc:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_handshake", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(25, 500, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_exc = __pyx_t_5;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":501
 *             self._sslobj.do_handshake()
 *         except ssl_SSLAgainErrors as exc:
 *             self._process_outgoing()             # <<<<<<<<<<<<<<
 *         except ssl_SSLError as exc:
 *             self._on_handshake_complete(exc)
 */
        __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_process_outgoing(__pyx_v_self); if (unlikely(!__pyx_t_9)) __PYX_ERR(25, 501, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }

      /* "uvloop/sslproto.pyx":500
 *         try:
 *             self._sslobj.do_handshake()
 *         except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *             self._process_outgoing()
 *         except ssl_SSLError as exc:
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L15;
        }
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __pyx_t_8 = __pyx_lineno; __pyx_t_10 = __pyx_clineno; __pyx_t_11 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          }
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ErrRestore(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_10; __pyx_filename = __pyx_t_11;
          goto __pyx_L5_except_error;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L4_exception_handled;
    }

    /* "uvloop/sslproto.pyx":502
 *         except ssl_SSLAgainErrors as exc:
 *             self._process_outgoing()
 *         except ssl_SSLError as exc:             # <<<<<<<<<<<<<<
 *             self._on_handshake_complete(exc)
 *         else:
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLError);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_handshake", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4) < 0) __PYX_ERR(25, 502, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_exc = __pyx_t_5;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":503
 *             self._process_outgoing()
 *         except ssl_SSLError as exc:
 *             self._on_handshake_complete(exc)             # <<<<<<<<<<<<<<
 *         else:
 *             self._on_handshake_complete(None)
 */
        __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_handshake_complete(__pyx_v_self, __pyx_v_exc); if (unlikely(!__pyx_t_9)) __PYX_ERR(25, 503, __pyx_L25_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }

      /* "uvloop/sslproto.pyx":502
 *         except ssl_SSLAgainErrors as exc:
 *             self._process_outgoing()
 *         except ssl_SSLError as exc:             # <<<<<<<<<<<<<<
 *             self._on_handshake_complete(exc)
 *         else:
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L26;
        }
        __pyx_L25_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_16, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_12);
          __pyx_t_10 = __pyx_lineno; __pyx_t_8 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
          }
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_17, __pyx_t_16, __pyx_t_15);
          __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0;
          __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_8; __pyx_filename = __pyx_t_18;
          goto __pyx_L5_except_error;
        }
        __pyx_L26:;
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/sslproto.pyx":498
 * 
 *     cdef _do_handshake(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self._sslobj.do_handshake()
 *         except ssl_SSLAgainErrors as exc:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/sslproto.pyx":497
 *             self._fatal_error(ConnectionAbortedError(msg))
 * 
 *     cdef _do_handshake(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self._sslobj.do_handshake()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_handshake", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":507
 *             self._on_handshake_complete(None)
 * 
 *     cdef _on_handshake_complete(self, handshake_exc):             # <<<<<<<<<<<<<<
 *         if self._handshake_timeout_handle is not None:
 *             self._handshake_timeout_handle.cancel()
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__on_handshake_complete(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_handshake_exc) {
  PyObject *__pyx_v_sslobj = NULL;
  PyObject *__pyx_v_peercert = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_v_dt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error __pyx_t_11;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter __pyx_t_12;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_handshake_complete", 1);

  /* "uvloop/sslproto.pyx":508
 * 
 *     cdef _on_handshake_complete(self, handshake_exc):
 *         if self._handshake_timeout_handle is not None:             # <<<<<<<<<<<<<<
 *             self._handshake_timeout_handle.cancel()
 *             self._handshake_timeout_handle = None
 */
  __pyx_t_1 = (__pyx_v_self->_handshake_timeout_handle != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":509
 *     cdef _on_handshake_complete(self, handshake_exc):
 *         if self._handshake_timeout_handle is not None:
 *             self._handshake_timeout_handle.cancel()             # <<<<<<<<<<<<<<
 *             self._handshake_timeout_handle = None
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_handshake_timeout_handle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 509, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":510
 *         if self._handshake_timeout_handle is not None:
 *             self._handshake_timeout_handle.cancel()
 *             self._handshake_timeout_handle = None             # <<<<<<<<<<<<<<
 * 
 *         sslobj = self._sslobj
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_handshake_timeout_handle);
    __Pyx_DECREF(__pyx_v_self->_handshake_timeout_handle);
    __pyx_v_self->_handshake_timeout_handle = Py_None;

    /* "uvloop/sslproto.pyx":508
 * 
 *     cdef _on_handshake_complete(self, handshake_exc):
 *         if self._handshake_timeout_handle is not None:             # <<<<<<<<<<<<<<
 *             self._handshake_timeout_handle.cancel()
 *             self._handshake_timeout_handle = None
 */
  }

  /* "uvloop/sslproto.pyx":512
 *             self._handshake_timeout_handle = None
 * 
 *         sslobj = self._sslobj             # <<<<<<<<<<<<<<
 *         try:
 *             if handshake_exc is None:
 */
  __pyx_t_2 = __pyx_v_self->_sslobj;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_sslobj = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":513
 * 
 *         sslobj = self._sslobj
 *         try:             # <<<<<<<<<<<<<<
 *             if handshake_exc is None:
 *                 self._set_state(WRAPPED)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":514
 *         sslobj = self._sslobj
 *         try:
 *             if handshake_exc is None:             # <<<<<<<<<<<<<<
 *                 self._set_state(WRAPPED)
 *             else:
 */
      __pyx_t_1 = (__pyx_v_handshake_exc == Py_None);
      if (likely(__pyx_t_1)) {

        /* "uvloop/sslproto.pyx":515
 *         try:
 *             if handshake_exc is None:
 *                 self._set_state(WRAPPED)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise handshake_exc
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_state(__pyx_v_self, __pyx_e_6uvloop_4loop_WRAPPED); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 515, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/sslproto.pyx":514
 *         sslobj = self._sslobj
 *         try:
 *             if handshake_exc is None:             # <<<<<<<<<<<<<<
 *                 self._set_state(WRAPPED)
 *             else:
 */
        goto __pyx_L10;
      }

      /* "uvloop/sslproto.pyx":517
 *                 self._set_state(WRAPPED)
 *             else:
 *                 raise handshake_exc             # <<<<<<<<<<<<<<
 * 
 *             peercert = sslobj.getpeercert()
 */
      /*else*/ {
        __Pyx_Raise(__pyx_v_handshake_exc, 0, 0, 0);
        __PYX_ERR(25, 517, __pyx_L4_error)
      }
      __pyx_L10:;

      /* "uvloop/sslproto.pyx":519
 *                 raise handshake_exc
 * 
 *             peercert = sslobj.getpeercert()             # <<<<<<<<<<<<<<
 *         except Exception as exc:
 *             self._set_state(UNWRAPPED)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sslobj, __pyx_n_s_getpeercert); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 519, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 519, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_v_peercert = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "uvloop/sslproto.pyx":513
 * 
 *         sslobj = self._sslobj
 *         try:             # <<<<<<<<<<<<<<
 *             if handshake_exc is None:
 *                 self._set_state(WRAPPED)
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/sslproto.pyx":520
 * 
 *             peercert = sslobj.getpeercert()
 *         except Exception as exc:             # <<<<<<<<<<<<<<
 *             self._set_state(UNWRAPPED)
 *             if isinstance(exc, ssl_CertificateError):
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_9) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._on_handshake_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(25, 520, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_exc = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":521
 *             peercert = sslobj.getpeercert()
 *         except Exception as exc:
 *             self._set_state(UNWRAPPED)             # <<<<<<<<<<<<<<
 *             if isinstance(exc, ssl_CertificateError):
 *                 msg = 'SSL handshake failed on verifying the certificate'
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_state(__pyx_v_self, __pyx_e_6uvloop_4loop_UNWRAPPED); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 521, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/sslproto.pyx":522
 *         except Exception as exc:
 *             self._set_state(UNWRAPPED)
 *             if isinstance(exc, ssl_CertificateError):             # <<<<<<<<<<<<<<
 *                 msg = 'SSL handshake failed on verifying the certificate'
 *             else:
 */
        __pyx_t_10 = __pyx_v_6uvloop_4loop_ssl_CertificateError;
        __Pyx_INCREF(__pyx_t_10);
        __pyx_t_1 = PyObject_IsInstance(__pyx_v_exc, __pyx_t_10); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 522, __pyx_L16_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (__pyx_t_1) {

          /* "uvloop/sslproto.pyx":523
 *             self._set_state(UNWRAPPED)
 *             if isinstance(exc, ssl_CertificateError):
 *                 msg = 'SSL handshake failed on verifying the certificate'             # <<<<<<<<<<<<<<
 *             else:
 *                 msg = 'SSL handshake failed'
 */
          __Pyx_INCREF(__pyx_kp_u_SSL_handshake_failed_on_verifyin);
          __pyx_v_msg = __pyx_kp_u_SSL_handshake_failed_on_verifyin;

          /* "uvloop/sslproto.pyx":522
 *         except Exception as exc:
 *             self._set_state(UNWRAPPED)
 *             if isinstance(exc, ssl_CertificateError):             # <<<<<<<<<<<<<<
 *                 msg = 'SSL handshake failed on verifying the certificate'
 *             else:
 */
          goto __pyx_L18;
        }

        /* "uvloop/sslproto.pyx":525
 *                 msg = 'SSL handshake failed on verifying the certificate'
 *             else:
 *                 msg = 'SSL handshake failed'             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, msg)
 *             self._wakeup_waiter(exc)
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_kp_u_SSL_handshake_failed);
          __pyx_v_msg = __pyx_kp_u_SSL_handshake_failed;
        }
        __pyx_L18:;

        /* "uvloop/sslproto.pyx":526
 *             else:
 *                 msg = 'SSL handshake failed'
 *             self._fatal_error(exc, msg)             # <<<<<<<<<<<<<<
 *             self._wakeup_waiter(exc)
 *             return
 */
        __pyx_t_11.__pyx_n = 1;
        __pyx_t_11.message = __pyx_v_msg;
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_fatal_error(__pyx_v_self, __pyx_v_exc, &__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 526, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/sslproto.pyx":527
 *                 msg = 'SSL handshake failed'
 *             self._fatal_error(exc, msg)
 *             self._wakeup_waiter(exc)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
        __pyx_t_12.__pyx_n = 1;
        __pyx_t_12.exc = __pyx_v_exc;
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_wakeup_waiter(__pyx_v_self, &__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 527, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/sslproto.pyx":528
 *             self._fatal_error(exc, msg)
 *             self._wakeup_waiter(exc)
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self._loop.get_debug():
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L15_return;
      }

      /* "uvloop/sslproto.pyx":520
 * 
 *             peercert = sslobj.getpeercert()
 *         except Exception as exc:             # <<<<<<<<<<<<<<
 *             self._set_state(UNWRAPPED)
 *             if isinstance(exc, ssl_CertificateError):
 */
      /*finally:*/ {
        __pyx_L16_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_9 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L6_except_error;
        }
        __pyx_L15_return: {
          __pyx_t_20 = __pyx_r;
          __pyx_r = 0;
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          __pyx_r = __pyx_t_20;
          __pyx_t_20 = 0;
          goto __pyx_L7_except_return;
        }
      }
    }
    goto __pyx_L6_except_error;

    /* "uvloop/sslproto.pyx":513
 * 
 *         sslobj = self._sslobj
 *         try:             # <<<<<<<<<<<<<<
 *             if handshake_exc is None:
 *                 self._set_state(WRAPPED)
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L9_try_end:;
  }

  /* "uvloop/sslproto.pyx":530
 *             return
 * 
 *         if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *             dt = self._loop.time() - self._handshake_start_time
 *             aio_logger.debug("%r: SSL handshake took %.1f ms", self, dt * 1e3)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_get_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 530, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":531
 * 
 *         if self._loop.get_debug():
 *             dt = self._loop.time() - self._handshake_start_time             # <<<<<<<<<<<<<<
 *             aio_logger.debug("%r: SSL handshake took %.1f ms", self, dt * 1e3)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 531, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_4, __pyx_v_self->_handshake_start_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_dt = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":532
 *         if self._loop.get_debug():
 *             dt = self._loop.time() - self._handshake_start_time
 *             aio_logger.debug("%r: SSL handshake took %.1f ms", self, dt * 1e3)             # <<<<<<<<<<<<<<
 * 
 *         # Add extra info that becomes available after handshake.
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_dt, __pyx_float_1e3); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_10, __pyx_kp_u_r_SSL_handshake_took_1f_ms, ((PyObject *)__pyx_v_self), __pyx_t_2};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 532, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":530
 *             return
 * 
 *         if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *             dt = self._loop.time() - self._handshake_start_time
 *             aio_logger.debug("%r: SSL handshake took %.1f ms", self, dt * 1e3)
 */
  }

  /* "uvloop/sslproto.pyx":535
 * 
 *         # Add extra info that becomes available after handshake.
 *         self._extra.update(peercert=peercert,             # <<<<<<<<<<<<<<
 *                            cipher=sslobj.cipher(),
 *                            compression=sslobj.compression(),
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_extra, __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_peercert, __pyx_v_peercert) < 0) __PYX_ERR(25, 535, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":536
 *         # Add extra info that becomes available after handshake.
 *         self._extra.update(peercert=peercert,
 *                            cipher=sslobj.cipher(),             # <<<<<<<<<<<<<<
 *                            compression=sslobj.compression(),
 *                            ssl_object=sslobj)
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_sslobj, __pyx_n_s_cipher); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_21 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_21)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_21);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_21, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_cipher, __pyx_t_2) < 0) __PYX_ERR(25, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":537
 *         self._extra.update(peercert=peercert,
 *                            cipher=sslobj.cipher(),
 *                            compression=sslobj.compression(),             # <<<<<<<<<<<<<<
 *                            ssl_object=sslobj)
 *         if self._app_state == STATE_INIT:
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_sslobj, __pyx_n_s_compression); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_21 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_21)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_21);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_21, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_compression, __pyx_t_2) < 0) __PYX_ERR(25, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":538
 *                            cipher=sslobj.cipher(),
 *                            compression=sslobj.compression(),
 *                            ssl_object=sslobj)             # <<<<<<<<<<<<<<
 *         if self._app_state == STATE_INIT:
 *             self._app_state = STATE_CON_MADE
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_ssl_object, __pyx_v_sslobj) < 0) __PYX_ERR(25, 535, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":535
 * 
 *         # Add extra info that becomes available after handshake.
 *         self._extra.update(peercert=peercert,             # <<<<<<<<<<<<<<
 *                            cipher=sslobj.cipher(),
 *                            compression=sslobj.compression(),
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":539
 *                            compression=sslobj.compression(),
 *                            ssl_object=sslobj)
 *         if self._app_state == STATE_INIT:             # <<<<<<<<<<<<<<
 *             self._app_state = STATE_CON_MADE
 *             self._app_protocol.connection_made(self._get_app_transport())
 */
  __pyx_t_1 = (__pyx_v_self->_app_state == __pyx_e_6uvloop_4loop_STATE_INIT);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":540
 *                            ssl_object=sslobj)
 *         if self._app_state == STATE_INIT:
 *             self._app_state = STATE_CON_MADE             # <<<<<<<<<<<<<<
 *             self._app_protocol.connection_made(self._get_app_transport())
 *         self._wakeup_waiter()
 */
    __pyx_v_self->_app_state = __pyx_e_6uvloop_4loop_STATE_CON_MADE;

    /* "uvloop/sslproto.pyx":541
 *         if self._app_state == STATE_INIT:
 *             self._app_state = STATE_CON_MADE
 *             self._app_protocol.connection_made(self._get_app_transport())             # <<<<<<<<<<<<<<
 *         self._wakeup_waiter()
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_connection_made); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_app_transport); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_21 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_21)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_21);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_21, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __pyx_t_10 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 541, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":539
 *                            compression=sslobj.compression(),
 *                            ssl_object=sslobj)
 *         if self._app_state == STATE_INIT:             # <<<<<<<<<<<<<<
 *             self._app_state = STATE_CON_MADE
 *             self._app_protocol.connection_made(self._get_app_transport())
 */
  }

  /* "uvloop/sslproto.pyx":542
 *             self._app_state = STATE_CON_MADE
 *             self._app_protocol.connection_made(self._get_app_transport())
 *         self._wakeup_waiter()             # <<<<<<<<<<<<<<
 * 
 *         # We should wakeup user code before sending the first data below. In
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_wakeup_waiter(__pyx_v_self, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":551
 *         # before having to handle the first data.
 *         self._loop._call_soon_handle(
 *             new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                              "SSLProtocol._do_read",
 *                              <method_t> self._do_read,
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_loop);
  __Pyx_INCREF(__pyx_t_2);

  /* "uvloop/sslproto.pyx":555
 *                              <method_t> self._do_read,
 *                              None,  # current context is good
 *                              self))             # <<<<<<<<<<<<<<
 * 
 *     # Shutdown flow
 */
  __pyx_t_4 = __pyx_f_6uvloop_4loop_new_MethodHandle(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_2), __pyx_kp_u_SSLProtocol__do_read, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read), Py_None, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":551
 *         # before having to handle the first data.
 *         self._loop._call_soon_handle(
 *             new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                              "SSLProtocol._do_read",
 *                              <method_t> self._do_read,
 */
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(25, 551, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":550
 *         # that the user get a chance to e.g. check ALPN with the transport
 *         # before having to handle the first data.
 *         self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *             new_MethodHandle(self._loop,
 *                              "SSLProtocol._do_read",
 */
  __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self->_loop, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":507
 *             self._on_handshake_complete(None)
 * 
 *     cdef _on_handshake_complete(self, handshake_exc):             # <<<<<<<<<<<<<<
 *         if self._handshake_timeout_handle is not None:
 *             self._handshake_timeout_handle.cancel()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._on_handshake_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sslobj);
  __Pyx_XDECREF(__pyx_v_peercert);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_dt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":573
 *             self._shutdown_timeout_handle = \
 *                 self._loop.call_later(self._ssl_shutdown_timeout,
 *                                       lambda: self._check_shutdown_timeout())             # <<<<<<<<<<<<<<
 *             self._do_flush(context)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_15_start_shutdown_lambda5(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_15_start_shutdown_lambda5 = {"lambda5", (PyCFunction)__pyx_pw_6uvloop_4loop_11SSLProtocol_15_start_shutdown_lambda5, METH_NOARGS, 0};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_15_start_shutdown_lambda5(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda5 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_lambda_funcdef_lambda5(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda5(PyObject *__pyx_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *__pyx_cur_scope;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda5", 1);
  __pyx_outer_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(25, 573, __pyx_L1_error) }
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_check_shutdown_timeout(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._start_shutdown.lambda5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":559
 *     # Shutdown flow
 * 
 *     cdef _start_shutdown(self, object context=None):             # <<<<<<<<<<<<<<
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__start_shutdown(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__start_shutdown *__pyx_optional_args) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *__pyx_cur_scope;
  PyObject *__pyx_v_context = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_flush __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_start_shutdown", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_32__start_shutdown(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(25, 559, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_context = __pyx_optional_args->context;
    }
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "uvloop/sslproto.pyx":560
 * 
 *     cdef _start_shutdown(self, object context=None):
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):             # <<<<<<<<<<<<<<
 *             return
 *         # we don't need the context for _abort or the timeout, because
 */
  switch (__pyx_cur_scope->__pyx_v_self->_state) {
    case __pyx_e_6uvloop_4loop_FLUSHING:
    case __pyx_e_6uvloop_4loop_SHUTDOWN:
    case __pyx_e_6uvloop_4loop_UNWRAPPED:

    /* "uvloop/sslproto.pyx":561
 *     cdef _start_shutdown(self, object context=None):
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             return             # <<<<<<<<<<<<<<
 *         # we don't need the context for _abort or the timeout, because
 *         # TCP transport._force_close() should be able to call
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":560
 * 
 *     cdef _start_shutdown(self, object context=None):
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):             # <<<<<<<<<<<<<<
 *             return
 *         # we don't need the context for _abort or the timeout, because
 */
    break;
    default: break;
  }

  /* "uvloop/sslproto.pyx":565
 *         # TCP transport._force_close() should be able to call
 *         # connection_lost() in the right context
 *         if self._app_transport is not None:             # <<<<<<<<<<<<<<
 *             self._app_transport._closed = True
 *         if self._state == DO_HANDSHAKE:
 */
  __pyx_t_1 = (((PyObject *)__pyx_cur_scope->__pyx_v_self->_app_transport) != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":566
 *         # connection_lost() in the right context
 *         if self._app_transport is not None:
 *             self._app_transport._closed = True             # <<<<<<<<<<<<<<
 *         if self._state == DO_HANDSHAKE:
 *             self._abort(None)
 */
    __pyx_cur_scope->__pyx_v_self->_app_transport->_closed = 1;

    /* "uvloop/sslproto.pyx":565
 *         # TCP transport._force_close() should be able to call
 *         # connection_lost() in the right context
 *         if self._app_transport is not None:             # <<<<<<<<<<<<<<
 *             self._app_transport._closed = True
 *         if self._state == DO_HANDSHAKE:
 */
  }

  /* "uvloop/sslproto.pyx":567
 *         if self._app_transport is not None:
 *             self._app_transport._closed = True
 *         if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             self._abort(None)
 *         else:
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_self->_state == __pyx_e_6uvloop_4loop_DO_HANDSHAKE);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":568
 *             self._app_transport._closed = True
 *         if self._state == DO_HANDSHAKE:
 *             self._abort(None)             # <<<<<<<<<<<<<<
 *         else:
 *             self._set_state(FLUSHING)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_abort(__pyx_cur_scope->__pyx_v_self, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":567
 *         if self._app_transport is not None:
 *             self._app_transport._closed = True
 *         if self._state == DO_HANDSHAKE:             # <<<<<<<<<<<<<<
 *             self._abort(None)
 *         else:
 */
    goto __pyx_L4;
  }

  /* "uvloop/sslproto.pyx":570
 *             self._abort(None)
 *         else:
 *             self._set_state(FLUSHING)             # <<<<<<<<<<<<<<
 *             self._shutdown_timeout_handle = \
 *                 self._loop.call_later(self._ssl_shutdown_timeout,
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_set_state(__pyx_cur_scope->__pyx_v_self, __pyx_e_6uvloop_4loop_FLUSHING); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":572
 *             self._set_state(FLUSHING)
 *             self._shutdown_timeout_handle = \
 *                 self._loop.call_later(self._ssl_shutdown_timeout,             # <<<<<<<<<<<<<<
 *                                       lambda: self._check_shutdown_timeout())
 *             self._do_flush(context)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->_loop), __pyx_n_s_call_later); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/sslproto.pyx":573
 *             self._shutdown_timeout_handle = \
 *                 self._loop.call_later(self._ssl_shutdown_timeout,
 *                                       lambda: self._check_shutdown_timeout())             # <<<<<<<<<<<<<<
 *             self._do_flush(context)
 * 
 */
    __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_15_start_shutdown_lambda5, 0, __pyx_n_s_SSLProtocol__start_shutdown_loca, ((PyObject*)__pyx_cur_scope), __pyx_n_s_uvloop_loop, __pyx_d, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_self->_ssl_shutdown_timeout, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "uvloop/sslproto.pyx":571
 *         else:
 *             self._set_state(FLUSHING)
 *             self._shutdown_timeout_handle = \             # <<<<<<<<<<<<<<
 *                 self._loop.call_later(self._ssl_shutdown_timeout,
 *                                       lambda: self._check_shutdown_timeout())
 */
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_shutdown_timeout_handle);
    __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_shutdown_timeout_handle);
    __pyx_cur_scope->__pyx_v_self->_shutdown_timeout_handle = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":574
 *                 self._loop.call_later(self._ssl_shutdown_timeout,
 *                                       lambda: self._check_shutdown_timeout())
 *             self._do_flush(context)             # <<<<<<<<<<<<<<
 * 
 *     cdef _check_shutdown_timeout(self):
 */
    __pyx_t_7.__pyx_n = 1;
    __pyx_t_7.context = __pyx_v_context;
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_do_flush(__pyx_cur_scope->__pyx_v_self, &__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "uvloop/sslproto.pyx":559
 *     # Shutdown flow
 * 
 *     cdef _start_shutdown(self, object context=None):             # <<<<<<<<<<<<<<
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._start_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":576
 *             self._do_flush(context)
 * 
 *     cdef _check_shutdown_timeout(self):             # <<<<<<<<<<<<<<
 *         if self._state in (FLUSHING, SHUTDOWN):
 *             self._transport._force_close(
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__check_shutdown_timeout(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_shutdown_timeout", 1);

  /* "uvloop/sslproto.pyx":577
 * 
 *     cdef _check_shutdown_timeout(self):
 *         if self._state in (FLUSHING, SHUTDOWN):             # <<<<<<<<<<<<<<
 *             self._transport._force_close(
 *                 aio_TimeoutError('SSL shutdown timed out'))
 */
  switch (__pyx_v_self->_state) {
    case __pyx_e_6uvloop_4loop_FLUSHING:
    case __pyx_e_6uvloop_4loop_SHUTDOWN:

    /* "uvloop/sslproto.pyx":578
 *     cdef _check_shutdown_timeout(self):
 *         if self._state in (FLUSHING, SHUTDOWN):
 *             self._transport._force_close(             # <<<<<<<<<<<<<<
 *                 aio_TimeoutError('SSL shutdown timed out'))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_force_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "uvloop/sslproto.pyx":579
 *         if self._state in (FLUSHING, SHUTDOWN):
 *             self._transport._force_close(
 *                 aio_TimeoutError('SSL shutdown timed out'))             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_read_into_void(self, object context):
 */
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_TimeoutError);
    __pyx_t_4 = __pyx_v_6uvloop_4loop_aio_TimeoutError; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_kp_u_SSL_shutdown_timed_out};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":577
 * 
 *     cdef _check_shutdown_timeout(self):
 *         if self._state in (FLUSHING, SHUTDOWN):             # <<<<<<<<<<<<<<
 *             self._transport._force_close(
 *                 aio_TimeoutError('SSL shutdown timed out'))
 */
    break;
    default: break;
  }

  /* "uvloop/sslproto.pyx":576
 *             self._do_flush(context)
 * 
 *     cdef _check_shutdown_timeout(self):             # <<<<<<<<<<<<<<
 *         if self._state in (FLUSHING, SHUTDOWN):
 *             self._transport._force_close(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._check_shutdown_timeout", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":581
 *                 aio_TimeoutError('SSL shutdown timed out'))
 * 
 *     cdef _do_read_into_void(self, object context):             # <<<<<<<<<<<<<<
 *         """Consume and discard incoming application data.
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read_into_void(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_context) {
  int __pyx_v_close_notify;
  CYTHON_UNUSED PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__call_eof_received __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_read_into_void", 1);

  /* "uvloop/sslproto.pyx":587
 *         """
 *         cdef:
 *             bint close_notify = False             # <<<<<<<<<<<<<<
 *         try:
 *             while True:
 */
  __pyx_v_close_notify = 0;

  /* "uvloop/sslproto.pyx":588
 *         cdef:
 *             bint close_notify = False
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":589
 *             bint close_notify = False
 *         try:
 *             while True:             # <<<<<<<<<<<<<<
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):
 *                     close_notify = True
 */
      while (1) {

        /* "uvloop/sslproto.pyx":590
 *         try:
 *             while True:
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):             # <<<<<<<<<<<<<<
 *                     close_notify = True
 *                     break
 */
        __pyx_t_5 = __Pyx_PyInt_From___pyx_anon_enum(__pyx_e_6uvloop_4loop_SSL_READ_MAX_SIZE); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 590, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_self->_sslobj_read);
        __pyx_t_6 = __pyx_v_self->_sslobj_read; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 590, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(25, 590, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_10 = (!__pyx_t_9);
        if (__pyx_t_10) {

          /* "uvloop/sslproto.pyx":591
 *             while True:
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):
 *                     close_notify = True             # <<<<<<<<<<<<<<
 *                     break
 *         except ssl_SSLAgainErrors as exc:
 */
          __pyx_v_close_notify = 1;

          /* "uvloop/sslproto.pyx":592
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):
 *                     close_notify = True
 *                     break             # <<<<<<<<<<<<<<
 *         except ssl_SSLAgainErrors as exc:
 *             pass
 */
          goto __pyx_L10_break;

          /* "uvloop/sslproto.pyx":590
 *         try:
 *             while True:
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):             # <<<<<<<<<<<<<<
 *                     close_notify = True
 *                     break
 */
        }
      }
      __pyx_L10_break:;

      /* "uvloop/sslproto.pyx":588
 *         cdef:
 *             bint close_notify = False
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/sslproto.pyx":593
 *                     close_notify = True
 *                     break
 *         except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *             pass
 *         except ssl_SSLZeroReturnError:
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read_into_void", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(25, 593, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_exc = __pyx_t_6;
      /*try:*/ {
      }
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L18;
        }
        __pyx_L18:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L4_exception_handled;
    }

    /* "uvloop/sslproto.pyx":595
 *         except ssl_SSLAgainErrors as exc:
 *             pass
 *         except ssl_SSLZeroReturnError:             # <<<<<<<<<<<<<<
 *             close_notify = True
 *         if close_notify:
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLZeroReturnError);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read_into_void", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(25, 595, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);

      /* "uvloop/sslproto.pyx":596
 *             pass
 *         except ssl_SSLZeroReturnError:
 *             close_notify = True             # <<<<<<<<<<<<<<
 *         if close_notify:
 *             self._call_eof_received(context)
 */
      __pyx_v_close_notify = 1;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/sslproto.pyx":588
 *         cdef:
 *             bint close_notify = False
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 if not self._sslobj_read(SSL_READ_MAX_SIZE):
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/sslproto.pyx":597
 *         except ssl_SSLZeroReturnError:
 *             close_notify = True
 *         if close_notify:             # <<<<<<<<<<<<<<
 *             self._call_eof_received(context)
 * 
 */
  if (__pyx_v_close_notify) {

    /* "uvloop/sslproto.pyx":598
 *             close_notify = True
 *         if close_notify:
 *             self._call_eof_received(context)             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_flush(self, object context=None):
 */
    __pyx_t_12.__pyx_n = 1;
    __pyx_t_12.context = __pyx_v_context;
    __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_call_eof_received(__pyx_v_self, &__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/sslproto.pyx":597
 *         except ssl_SSLZeroReturnError:
 *             close_notify = True
 *         if close_notify:             # <<<<<<<<<<<<<<
 *             self._call_eof_received(context)
 * 
 */
  }

  /* "uvloop/sslproto.pyx":581
 *                 aio_TimeoutError('SSL shutdown timed out'))
 * 
 *     cdef _do_read_into_void(self, object context):             # <<<<<<<<<<<<<<
 *         """Consume and discard incoming application data.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read_into_void", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":600
 *             self._call_eof_received(context)
 * 
 *     cdef _do_flush(self, object context=None):             # <<<<<<<<<<<<<<
 *         """Flush the write backlog, discarding new data received.
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_flush(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_flush *__pyx_optional_args) {
  PyObject *__pyx_v_context = ((PyObject *)Py_None);
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_shutdown __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_flush", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_context = __pyx_optional_args->context;
    }
  }

  /* "uvloop/sslproto.pyx":608
 *         in FLUSHING, as we could fully manage the flow control internally.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self._do_read_into_void(context)
 *             self._do_write()
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":609
 *         """
 *         try:
 *             self._do_read_into_void(context)             # <<<<<<<<<<<<<<
 *             self._do_write()
 *             self._process_outgoing()
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read_into_void(__pyx_v_self, __pyx_v_context); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 609, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/sslproto.pyx":610
 *         try:
 *             self._do_read_into_void(context)
 *             self._do_write()             # <<<<<<<<<<<<<<
 *             self._process_outgoing()
 *             self._control_ssl_reading()
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_write(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 610, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/sslproto.pyx":611
 *             self._do_read_into_void(context)
 *             self._do_write()
 *             self._process_outgoing()             # <<<<<<<<<<<<<<
 *             self._control_ssl_reading()
 *         except Exception as ex:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_process_outgoing(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 611, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/sslproto.pyx":612
 *             self._do_write()
 *             self._process_outgoing()
 *             self._control_ssl_reading()             # <<<<<<<<<<<<<<
 *         except Exception as ex:
 *             self._on_shutdown_complete(ex)
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_control_ssl_reading(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 612, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/sslproto.pyx":608
 *         in FLUSHING, as we could fully manage the flow control internally.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self._do_read_into_void(context)
 *             self._do_write()
 */
    }

    /* "uvloop/sslproto.pyx":616
 *             self._on_shutdown_complete(ex)
 *         else:
 *             if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *                 self._set_state(SHUTDOWN)
 *                 self._do_shutdown(context)
 */
    /*else:*/ {
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_get_write_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 616, __pyx_L5_except_error)
      __pyx_t_6 = (!(__pyx_t_5 != 0));
      if (__pyx_t_6) {

        /* "uvloop/sslproto.pyx":617
 *         else:
 *             if not self._get_write_buffer_size():
 *                 self._set_state(SHUTDOWN)             # <<<<<<<<<<<<<<
 *                 self._do_shutdown(context)
 * 
 */
        __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_state(__pyx_v_self, __pyx_e_6uvloop_4loop_SHUTDOWN); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 617, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/sslproto.pyx":618
 *             if not self._get_write_buffer_size():
 *                 self._set_state(SHUTDOWN)
 *                 self._do_shutdown(context)             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_shutdown(self, object context=None):
 */
        __pyx_t_7.__pyx_n = 1;
        __pyx_t_7.context = __pyx_v_context;
        __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_shutdown(__pyx_v_self, &__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 618, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "uvloop/sslproto.pyx":616
 *             self._on_shutdown_complete(ex)
 *         else:
 *             if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *                 self._set_state(SHUTDOWN)
 *                 self._do_shutdown(context)
 */
      }
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/sslproto.pyx":613
 *             self._process_outgoing()
 *             self._control_ssl_reading()
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._on_shutdown_complete(ex)
 *         else:
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_9, &__pyx_t_10) < 0) __PYX_ERR(25, 613, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      __pyx_v_ex = __pyx_t_9;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":614
 *             self._control_ssl_reading()
 *         except Exception as ex:
 *             self._on_shutdown_complete(ex)             # <<<<<<<<<<<<<<
 *         else:
 *             if not self._get_write_buffer_size():
 */
        __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_shutdown_complete(__pyx_v_self, __pyx_v_ex); if (unlikely(!__pyx_t_11)) __PYX_ERR(25, 614, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }

      /* "uvloop/sslproto.pyx":613
 *             self._process_outgoing()
 *             self._control_ssl_reading()
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._on_shutdown_complete(ex)
 *         else:
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/sslproto.pyx":608
 *         in FLUSHING, as we could fully manage the flow control internally.
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             self._do_read_into_void(context)
 *             self._do_write()
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/sslproto.pyx":600
 *             self._call_eof_received(context)
 * 
 *     cdef _do_flush(self, object context=None):             # <<<<<<<<<<<<<<
 *         """Flush the write backlog, discarding new data received.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":620
 *                 self._do_shutdown(context)
 * 
 *     cdef _do_shutdown(self, object context=None):             # <<<<<<<<<<<<<<
 *         """Send close_notify and wait for the same from the peer."""
 *         try:
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_shutdown(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_shutdown *__pyx_optional_args) {
  PyObject *__pyx_v_context = ((PyObject *)Py_None);
  CYTHON_UNUSED PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  size_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  char const *__pyx_t_23;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_shutdown", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_context = __pyx_optional_args->context;
    }
  }

  /* "uvloop/sslproto.pyx":622
 *     cdef _do_shutdown(self, object context=None):
 *         """Send close_notify and wait for the same from the peer."""
 *         try:             # <<<<<<<<<<<<<<
 *             # we must skip all application data (if any) before unwrap
 *             self._do_read_into_void(context)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":624
 *         try:
 *             # we must skip all application data (if any) before unwrap
 *             self._do_read_into_void(context)             # <<<<<<<<<<<<<<
 *             try:
 *                 self._sslobj.unwrap()
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read_into_void(__pyx_v_self, __pyx_v_context); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 624, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/sslproto.pyx":625
 *             # we must skip all application data (if any) before unwrap
 *             self._do_read_into_void(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 self._sslobj.unwrap()
 *             except ssl_SSLAgainErrors as exc:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        /*try:*/ {

          /* "uvloop/sslproto.pyx":626
 *             self._do_read_into_void(context)
 *             try:
 *                 self._sslobj.unwrap()             # <<<<<<<<<<<<<<
 *             except ssl_SSLAgainErrors as exc:
 *                 self._process_outgoing()
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_sslobj, __pyx_n_s_unwrap); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 626, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 0+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 626, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "uvloop/sslproto.pyx":625
 *             # we must skip all application data (if any) before unwrap
 *             self._do_read_into_void(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 self._sslobj.unwrap()
 *             except ssl_SSLAgainErrors as exc:
 */
        }

        /* "uvloop/sslproto.pyx":630
 *                 self._process_outgoing()
 *             else:
 *                 self._process_outgoing()             # <<<<<<<<<<<<<<
 *                 if not self._get_write_buffer_size():
 *                     self._on_shutdown_complete(None)
 */
        /*else:*/ {
          __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_process_outgoing(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 630, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "uvloop/sslproto.pyx":631
 *             else:
 *                 self._process_outgoing()
 *                 if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *                     self._on_shutdown_complete(None)
 *         except Exception as ex:
 */
          __pyx_t_11 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_get_write_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 631, __pyx_L11_except_error)
          __pyx_t_12 = (!(__pyx_t_11 != 0));
          if (__pyx_t_12) {

            /* "uvloop/sslproto.pyx":632
 *                 self._process_outgoing()
 *                 if not self._get_write_buffer_size():
 *                     self._on_shutdown_complete(None)             # <<<<<<<<<<<<<<
 *         except Exception as ex:
 *             self._on_shutdown_complete(ex)
 */
            __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_shutdown_complete(__pyx_v_self, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 632, __pyx_L11_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "uvloop/sslproto.pyx":631
 *             else:
 *                 self._process_outgoing()
 *                 if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *                     self._on_shutdown_complete(None)
 *         except Exception as ex:
 */
          }
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L14_try_end;
        __pyx_L9_error:;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "uvloop/sslproto.pyx":627
 *             try:
 *                 self._sslobj.unwrap()
 *             except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *                 self._process_outgoing()
 *             else:
 */
        __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors);
        if (__pyx_t_13) {
          __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(25, 627, __pyx_L11_except_error)
          __Pyx_XGOTREF(__pyx_t_4);
          __Pyx_XGOTREF(__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_8);
          __pyx_v_exc = __pyx_t_8;
          /*try:*/ {

            /* "uvloop/sslproto.pyx":628
 *                 self._sslobj.unwrap()
 *             except ssl_SSLAgainErrors as exc:
 *                 self._process_outgoing()             # <<<<<<<<<<<<<<
 *             else:
 *                 self._process_outgoing()
 */
            __pyx_t_14 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_process_outgoing(__pyx_v_self); if (unlikely(!__pyx_t_14)) __PYX_ERR(25, 628, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }

          /* "uvloop/sslproto.pyx":627
 *             try:
 *                 self._sslobj.unwrap()
 *             except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *                 self._process_outgoing()
 *             else:
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
              goto __pyx_L22;
            }
            __pyx_L21_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __pyx_t_13 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
              }
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
              goto __pyx_L11_except_error;
            }
            __pyx_L22:;
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L10_exception_handled;
        }
        goto __pyx_L11_except_error;

        /* "uvloop/sslproto.pyx":625
 *             # we must skip all application data (if any) before unwrap
 *             self._do_read_into_void(context)
 *             try:             # <<<<<<<<<<<<<<
 *                 self._sslobj.unwrap()
 *             except ssl_SSLAgainErrors as exc:
 */
        __pyx_L11_except_error:;
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        goto __pyx_L3_error;
        __pyx_L10_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        __pyx_L14_try_end:;
      }

      /* "uvloop/sslproto.pyx":622
 *     cdef _do_shutdown(self, object context=None):
 *         """Send close_notify and wait for the same from the peer."""
 *         try:             # <<<<<<<<<<<<<<
 *             # we must skip all application data (if any) before unwrap
 *             self._do_read_into_void(context)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "uvloop/sslproto.pyx":633
 *                 if not self._get_write_buffer_size():
 *                     self._on_shutdown_complete(None)
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._on_shutdown_complete(ex)
 * 
 */
    __pyx_t_15 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_15) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_4) < 0) __PYX_ERR(25, 633, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_ex = __pyx_t_8;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":634
 *                     self._on_shutdown_complete(None)
 *         except Exception as ex:
 *             self._on_shutdown_complete(ex)             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_shutdown_complete(self, shutdown_exc):
 */
        __pyx_t_14 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_on_shutdown_complete(__pyx_v_self, __pyx_v_ex); if (unlikely(!__pyx_t_14)) __PYX_ERR(25, 634, __pyx_L32_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }

      /* "uvloop/sslproto.pyx":633
 *                 if not self._get_write_buffer_size():
 *                     self._on_shutdown_complete(None)
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._on_shutdown_complete(ex)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L33;
        }
        __pyx_L32_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_21, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_15 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_23 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_21, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_XGIVEREF(__pyx_t_5);
          __Pyx_ErrRestore(__pyx_t_7, __pyx_t_6, __pyx_t_5);
          __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_15; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_23;
          goto __pyx_L5_except_error;
        }
        __pyx_L33:;
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/sslproto.pyx":622
 *     cdef _do_shutdown(self, object context=None):
 *         """Send close_notify and wait for the same from the peer."""
 *         try:             # <<<<<<<<<<<<<<
 *             # we must skip all application data (if any) before unwrap
 *             self._do_read_into_void(context)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/sslproto.pyx":620
 *                 self._do_shutdown(context)
 * 
 *     cdef _do_shutdown(self, object context=None):             # <<<<<<<<<<<<<<
 *         """Send close_notify and wait for the same from the peer."""
 *         try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_shutdown", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":636
 *             self._on_shutdown_complete(ex)
 * 
 *     cdef _on_shutdown_complete(self, shutdown_exc):             # <<<<<<<<<<<<<<
 *         if self._shutdown_timeout_handle is not None:
 *             self._shutdown_timeout_handle.cancel()
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__on_shutdown_complete(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_shutdown_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_shutdown_complete", 1);

  /* "uvloop/sslproto.pyx":637
 * 
 *     cdef _on_shutdown_complete(self, shutdown_exc):
 *         if self._shutdown_timeout_handle is not None:             # <<<<<<<<<<<<<<
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None
 */
  __pyx_t_1 = (__pyx_v_self->_shutdown_timeout_handle != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":638
 *     cdef _on_shutdown_complete(self, shutdown_exc):
 *         if self._shutdown_timeout_handle is not None:
 *             self._shutdown_timeout_handle.cancel()             # <<<<<<<<<<<<<<
 *             self._shutdown_timeout_handle = None
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_shutdown_timeout_handle, __pyx_n_s_cancel); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":639
 *         if self._shutdown_timeout_handle is not None:
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None             # <<<<<<<<<<<<<<
 * 
 *         # we don't need the context here because TCP transport.close() should
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_shutdown_timeout_handle);
    __Pyx_DECREF(__pyx_v_self->_shutdown_timeout_handle);
    __pyx_v_self->_shutdown_timeout_handle = Py_None;

    /* "uvloop/sslproto.pyx":637
 * 
 *     cdef _on_shutdown_complete(self, shutdown_exc):
 *         if self._shutdown_timeout_handle is not None:             # <<<<<<<<<<<<<<
 *             self._shutdown_timeout_handle.cancel()
 *             self._shutdown_timeout_handle = None
 */
  }

  /* "uvloop/sslproto.pyx":643
 *         # we don't need the context here because TCP transport.close() should
 *         # be able to call connection_made() in the right context
 *         if shutdown_exc:             # <<<<<<<<<<<<<<
 *             self._fatal_error(shutdown_exc, 'Error occurred during shutdown')
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_shutdown_exc); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 643, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":644
 *         # be able to call connection_made() in the right context
 *         if shutdown_exc:
 *             self._fatal_error(shutdown_exc, 'Error occurred during shutdown')             # <<<<<<<<<<<<<<
 *         else:
 *             self._transport.close()
 */
    __pyx_t_6.__pyx_n = 1;
    __pyx_t_6.message = __pyx_kp_u_Error_occurred_during_shutdown;
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_fatal_error(__pyx_v_self, __pyx_v_shutdown_exc, &__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":643
 *         # we don't need the context here because TCP transport.close() should
 *         # be able to call connection_made() in the right context
 *         if shutdown_exc:             # <<<<<<<<<<<<<<
 *             self._fatal_error(shutdown_exc, 'Error occurred during shutdown')
 *         else:
 */
    goto __pyx_L4;
  }

  /* "uvloop/sslproto.pyx":646
 *             self._fatal_error(shutdown_exc, 'Error occurred during shutdown')
 *         else:
 *             self._transport.close()             # <<<<<<<<<<<<<<
 * 
 *     cdef _abort(self, exc):
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "uvloop/sslproto.pyx":636
 *             self._on_shutdown_complete(ex)
 * 
 *     cdef _on_shutdown_complete(self, shutdown_exc):             # <<<<<<<<<<<<<<
 *         if self._shutdown_timeout_handle is not None:
 *             self._shutdown_timeout_handle.cancel()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._on_shutdown_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":648
 *             self._transport.close()
 * 
 *     cdef _abort(self, exc):             # <<<<<<<<<<<<<<
 *         self._set_state(UNWRAPPED)
 *         if self._transport is not None:
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__abort(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_exc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_abort", 1);

  /* "uvloop/sslproto.pyx":649
 * 
 *     cdef _abort(self, exc):
 *         self._set_state(UNWRAPPED)             # <<<<<<<<<<<<<<
 *         if self._transport is not None:
 *             self._transport._force_close(exc)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_set_state(__pyx_v_self, __pyx_e_6uvloop_4loop_UNWRAPPED); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":650
 *     cdef _abort(self, exc):
 *         self._set_state(UNWRAPPED)
 *         if self._transport is not None:             # <<<<<<<<<<<<<<
 *             self._transport._force_close(exc)
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_transport != Py_None);
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":651
 *         self._set_state(UNWRAPPED)
 *         if self._transport is not None:
 *             self._transport._force_close(exc)             # <<<<<<<<<<<<<<
 * 
 *     # Outgoing flow
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_force_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_exc};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":650
 *     cdef _abort(self, exc):
 *         self._set_state(UNWRAPPED)
 *         if self._transport is not None:             # <<<<<<<<<<<<<<
 *             self._transport._force_close(exc)
 * 
 */
  }

  /* "uvloop/sslproto.pyx":648
 *             self._transport.close()
 * 
 *     cdef _abort(self, exc):             # <<<<<<<<<<<<<<
 *         self._set_state(UNWRAPPED)
 *         if self._transport is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._abort", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":655
 *     # Outgoing flow
 * 
 *     cdef _write_appdata(self, list_of_data, object context):             # <<<<<<<<<<<<<<
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__write_appdata(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_list_of_data, PyObject *__pyx_v_context) {
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error __pyx_t_16;
  int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_write_appdata", 1);

  /* "uvloop/sslproto.pyx":656
 * 
 *     cdef _write_appdata(self, list_of_data, object context):
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):             # <<<<<<<<<<<<<<
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 *                 aio_logger.warning('SSL connection is closed')
 */
  switch (__pyx_v_self->_state) {
    case __pyx_e_6uvloop_4loop_FLUSHING:
    case __pyx_e_6uvloop_4loop_SHUTDOWN:
    case __pyx_e_6uvloop_4loop_UNWRAPPED:

    /* "uvloop/sslproto.pyx":657
 *     cdef _write_appdata(self, list_of_data, object context):
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:             # <<<<<<<<<<<<<<
 *                 aio_logger.warning('SSL connection is closed')
 *             self._conn_lost += 1
 */
    __pyx_t_1 = (__pyx_v_self->_conn_lost >= __pyx_e_6uvloop_4loop_LOG_THRESHOLD_FOR_CONNLOST_WRITES);
    if (__pyx_t_1) {

      /* "uvloop/sslproto.pyx":658
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 *                 aio_logger.warning('SSL connection is closed')             # <<<<<<<<<<<<<<
 *             self._conn_lost += 1
 *             return
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_warning); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_SSL_connection_is_closed};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 658, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/sslproto.pyx":657
 *     cdef _write_appdata(self, list_of_data, object context):
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:             # <<<<<<<<<<<<<<
 *                 aio_logger.warning('SSL connection is closed')
 *             self._conn_lost += 1
 */
    }

    /* "uvloop/sslproto.pyx":659
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 *                 aio_logger.warning('SSL connection is closed')
 *             self._conn_lost += 1             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_v_self->_conn_lost = (__pyx_v_self->_conn_lost + 1);

    /* "uvloop/sslproto.pyx":660
 *                 aio_logger.warning('SSL connection is closed')
 *             self._conn_lost += 1
 *             return             # <<<<<<<<<<<<<<
 * 
 *         for data in list_of_data:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":656
 * 
 *     cdef _write_appdata(self, list_of_data, object context):
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):             # <<<<<<<<<<<<<<
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 *                 aio_logger.warning('SSL connection is closed')
 */
    break;
    default: break;
  }

  /* "uvloop/sslproto.pyx":662
 *             return
 * 
 *         for data in list_of_data:             # <<<<<<<<<<<<<<
 *             self._write_backlog.append(data)
 *             self._write_buffer_size += len(data)
 */
  if (likely(PyList_CheckExact(__pyx_v_list_of_data)) || PyTuple_CheckExact(__pyx_v_list_of_data)) {
    __pyx_t_2 = __pyx_v_list_of_data; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_list_of_data); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 662, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(25, 662, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(25, 662, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 662, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(25, 662, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(25, 662, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 662, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(25, 662, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":663
 * 
 *         for data in list_of_data:
 *             self._write_backlog.append(data)             # <<<<<<<<<<<<<<
 *             self._write_buffer_size += len(data)
 * 
 */
    __pyx_t_8 = __Pyx_PyObject_Append(__pyx_v_self->_write_backlog, __pyx_v_data); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(25, 663, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":664
 *         for data in list_of_data:
 *             self._write_backlog.append(data)
 *             self._write_buffer_size += len(data)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __pyx_t_9 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(25, 664, __pyx_L1_error)
    __pyx_v_self->_write_buffer_size = (__pyx_v_self->_write_buffer_size + __pyx_t_9);

    /* "uvloop/sslproto.pyx":662
 *             return
 * 
 *         for data in list_of_data:             # <<<<<<<<<<<<<<
 *             self._write_backlog.append(data)
 *             self._write_buffer_size += len(data)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/sslproto.pyx":666
 *             self._write_buffer_size += len(data)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self._state == WRAPPED:
 *                 self._do_write()
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_12);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":667
 * 
 *         try:
 *             if self._state == WRAPPED:             # <<<<<<<<<<<<<<
 *                 self._do_write()
 *                 self._process_outgoing()
 */
      __pyx_t_1 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_WRAPPED);
      if (__pyx_t_1) {

        /* "uvloop/sslproto.pyx":668
 *         try:
 *             if self._state == WRAPPED:
 *                 self._do_write()             # <<<<<<<<<<<<<<
 *                 self._process_outgoing()
 *                 self._control_app_writing(context)
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_write(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 668, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/sslproto.pyx":669
 *             if self._state == WRAPPED:
 *                 self._do_write()
 *                 self._process_outgoing()             # <<<<<<<<<<<<<<
 *                 self._control_app_writing(context)
 * 
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_process_outgoing(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 669, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/sslproto.pyx":670
 *                 self._do_write()
 *                 self._process_outgoing()
 *                 self._control_app_writing(context)             # <<<<<<<<<<<<<<
 * 
 *         except Exception as ex:
 */
        __pyx_t_13.__pyx_n = 1;
        __pyx_t_13.context = __pyx_v_context;
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_control_app_writing(__pyx_v_self, &__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 670, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "uvloop/sslproto.pyx":667
 * 
 *         try:
 *             if self._state == WRAPPED:             # <<<<<<<<<<<<<<
 *                 self._do_write()
 *                 self._process_outgoing()
 */
      }

      /* "uvloop/sslproto.pyx":666
 *             self._write_buffer_size += len(data)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self._state == WRAPPED:
 *                 self._do_write()
 */
    }
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    goto __pyx_L12_try_end;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/sslproto.pyx":672
 *                 self._control_app_writing(context)
 * 
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 */
    __pyx_t_14 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_14) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._write_appdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(25, 672, __pyx_L9_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_ex = __pyx_t_3;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":673
 * 
 *         except Exception as ex:
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_write(self):
 */
        __pyx_t_16.__pyx_n = 1;
        __pyx_t_16.message = __pyx_kp_u_Fatal_error_on_SSL_protocol;
        __pyx_t_15 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_fatal_error(__pyx_v_self, __pyx_v_ex, &__pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(25, 673, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      }

      /* "uvloop/sslproto.pyx":672
 *                 self._control_app_writing(context)
 * 
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L20;
        }
        __pyx_L19_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __Pyx_XGOTREF(__pyx_t_24);
          __pyx_t_14 = __pyx_lineno; __pyx_t_17 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
          }
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_XGIVEREF(__pyx_t_21);
          __Pyx_ErrRestore(__pyx_t_19, __pyx_t_20, __pyx_t_21);
          __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
          __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_17; __pyx_filename = __pyx_t_18;
          goto __pyx_L9_except_error;
        }
        __pyx_L20:;
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L8_exception_handled;
    }
    goto __pyx_L9_except_error;

    /* "uvloop/sslproto.pyx":666
 *             self._write_buffer_size += len(data)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self._state == WRAPPED:
 *                 self._do_write()
 */
    __pyx_L9_except_error:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    goto __pyx_L1_error;
    __pyx_L8_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    __pyx_L12_try_end:;
  }

  /* "uvloop/sslproto.pyx":655
 *     # Outgoing flow
 * 
 *     cdef _write_appdata(self, list_of_data, object context):             # <<<<<<<<<<<<<<
 *         if self._state in (FLUSHING, SHUTDOWN, UNWRAPPED):
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._write_appdata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":675
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 *     cdef _do_write(self):             # <<<<<<<<<<<<<<
 *         """Do SSL write, consumes write backlog and fills outgoing BIO."""
 *         cdef size_t data_len, count
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_write(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  size_t __pyx_v_data_len;
  size_t __pyx_v_count;
  PyObject *__pyx_v_data = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  size_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_write", 1);

  /* "uvloop/sslproto.pyx":678
 *         """Do SSL write, consumes write backlog and fills outgoing BIO."""
 *         cdef size_t data_len, count
 *         try:             # <<<<<<<<<<<<<<
 *             while self._write_backlog:
 *                 data = self._write_backlog[0]
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":679
 *         cdef size_t data_len, count
 *         try:
 *             while self._write_backlog:             # <<<<<<<<<<<<<<
 *                 data = self._write_backlog[0]
 *                 count = self._sslobj_write(data)
 */
      while (1) {
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_self->_write_backlog); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(25, 679, __pyx_L3_error)
        if (!__pyx_t_4) break;

        /* "uvloop/sslproto.pyx":680
 *         try:
 *             while self._write_backlog:
 *                 data = self._write_backlog[0]             # <<<<<<<<<<<<<<
 *                 count = self._sslobj_write(data)
 *                 data_len = len(data)
 */
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->_write_backlog, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 680, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_v_data, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "uvloop/sslproto.pyx":681
 *             while self._write_backlog:
 *                 data = self._write_backlog[0]
 *                 count = self._sslobj_write(data)             # <<<<<<<<<<<<<<
 *                 data_len = len(data)
 *                 if count < data_len:
 */
        __Pyx_INCREF(__pyx_v_self->_sslobj_write);
        __pyx_t_6 = __pyx_v_self->_sslobj_write; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_data};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 681, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __pyx_t_9 = __Pyx_PyInt_As_size_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 681, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_count = __pyx_t_9;

        /* "uvloop/sslproto.pyx":682
 *                 data = self._write_backlog[0]
 *                 count = self._sslobj_write(data)
 *                 data_len = len(data)             # <<<<<<<<<<<<<<
 *                 if count < data_len:
 *                     if not PyMemoryView_Check(data):
 */
        __pyx_t_10 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(25, 682, __pyx_L3_error)
        __pyx_v_data_len = __pyx_t_10;

        /* "uvloop/sslproto.pyx":683
 *                 count = self._sslobj_write(data)
 *                 data_len = len(data)
 *                 if count < data_len:             # <<<<<<<<<<<<<<
 *                     if not PyMemoryView_Check(data):
 *                         data = PyMemoryView_FromObject(data)
 */
        __pyx_t_4 = (__pyx_v_count < __pyx_v_data_len);
        if (__pyx_t_4) {

          /* "uvloop/sslproto.pyx":684
 *                 data_len = len(data)
 *                 if count < data_len:
 *                     if not PyMemoryView_Check(data):             # <<<<<<<<<<<<<<
 *                         data = PyMemoryView_FromObject(data)
 *                     self._write_backlog[0] = data[count:]
 */
          __pyx_t_4 = (!(PyMemoryView_Check(__pyx_v_data) != 0));
          if (__pyx_t_4) {

            /* "uvloop/sslproto.pyx":685
 *                 if count < data_len:
 *                     if not PyMemoryView_Check(data):
 *                         data = PyMemoryView_FromObject(data)             # <<<<<<<<<<<<<<
 *                     self._write_backlog[0] = data[count:]
 *                     self._write_buffer_size -= count
 */
            __pyx_t_5 = PyMemoryView_FromObject(__pyx_v_data); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 685, __pyx_L3_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF_SET(__pyx_v_data, __pyx_t_5);
            __pyx_t_5 = 0;

            /* "uvloop/sslproto.pyx":684
 *                 data_len = len(data)
 *                 if count < data_len:
 *                     if not PyMemoryView_Check(data):             # <<<<<<<<<<<<<<
 *                         data = PyMemoryView_FromObject(data)
 *                     self._write_backlog[0] = data[count:]
 */
          }

          /* "uvloop/sslproto.pyx":686
 *                     if not PyMemoryView_Check(data):
 *                         data = PyMemoryView_FromObject(data)
 *                     self._write_backlog[0] = data[count:]             # <<<<<<<<<<<<<<
 *                     self._write_buffer_size -= count
 *                 else:
 */
          __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_data, __pyx_v_count, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 686, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (unlikely((__Pyx_SetItemInt(__pyx_v_self->_write_backlog, 0, __pyx_t_5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0))) __PYX_ERR(25, 686, __pyx_L3_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/sslproto.pyx":687
 *                         data = PyMemoryView_FromObject(data)
 *                     self._write_backlog[0] = data[count:]
 *                     self._write_buffer_size -= count             # <<<<<<<<<<<<<<
 *                 else:
 *                     del self._write_backlog[0]
 */
          __pyx_v_self->_write_buffer_size = (__pyx_v_self->_write_buffer_size - __pyx_v_count);

          /* "uvloop/sslproto.pyx":683
 *                 count = self._sslobj_write(data)
 *                 data_len = len(data)
 *                 if count < data_len:             # <<<<<<<<<<<<<<
 *                     if not PyMemoryView_Check(data):
 *                         data = PyMemoryView_FromObject(data)
 */
          goto __pyx_L11;
        }

        /* "uvloop/sslproto.pyx":689
 *                     self._write_buffer_size -= count
 *                 else:
 *                     del self._write_backlog[0]             # <<<<<<<<<<<<<<
 *                     self._write_buffer_size -= data_len
 *         except ssl_SSLAgainErrors as exc:
 */
        /*else*/ {
          if (unlikely((__Pyx_DelItemInt(__pyx_v_self->_write_backlog, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0))) __PYX_ERR(25, 689, __pyx_L3_error)

          /* "uvloop/sslproto.pyx":690
 *                 else:
 *                     del self._write_backlog[0]
 *                     self._write_buffer_size -= data_len             # <<<<<<<<<<<<<<
 *         except ssl_SSLAgainErrors as exc:
 *             pass
 */
          __pyx_v_self->_write_buffer_size = (__pyx_v_self->_write_buffer_size - __pyx_v_data_len);
        }
        __pyx_L11:;
      }

      /* "uvloop/sslproto.pyx":678
 *         """Do SSL write, consumes write backlog and fills outgoing BIO."""
 *         cdef size_t data_len, count
 *         try:             # <<<<<<<<<<<<<<
 *             while self._write_backlog:
 *                 data = self._write_backlog[0]
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/sslproto.pyx":691
 *                     del self._write_backlog[0]
 *                     self._write_buffer_size -= data_len
 *         except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(25, 691, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_exc = __pyx_t_6;
      /*try:*/ {
      }
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L19;
        }
        __pyx_L19:;
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/sslproto.pyx":678
 *         """Do SSL write, consumes write backlog and fills outgoing BIO."""
 *         cdef size_t data_len, count
 *         try:             # <<<<<<<<<<<<<<
 *             while self._write_backlog:
 *                 data = self._write_backlog[0]
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/sslproto.pyx":675
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 *     cdef _do_write(self):             # <<<<<<<<<<<<<<
 *         """Do SSL write, consumes write backlog and fills outgoing BIO."""
 *         cdef size_t data_len, count
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":694
 *             pass
 * 
 *     cdef _process_outgoing(self):             # <<<<<<<<<<<<<<
 *         """Send bytes from the outgoing BIO."""
 *         if not self._ssl_writing_paused:
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__process_outgoing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_process_outgoing", 1);

  /* "uvloop/sslproto.pyx":696
 *     cdef _process_outgoing(self):
 *         """Send bytes from the outgoing BIO."""
 *         if not self._ssl_writing_paused:             # <<<<<<<<<<<<<<
 *             data = self._outgoing_read()
 *             if len(data):
 */
  __pyx_t_1 = (!__pyx_v_self->_ssl_writing_paused);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":697
 *         """Send bytes from the outgoing BIO."""
 *         if not self._ssl_writing_paused:
 *             data = self._outgoing_read()             # <<<<<<<<<<<<<<
 *             if len(data):
 *                 self._transport.write(data)
 */
    __Pyx_INCREF(__pyx_v_self->_outgoing_read);
    __pyx_t_3 = __pyx_v_self->_outgoing_read; __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 697, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_data = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":698
 *         if not self._ssl_writing_paused:
 *             data = self._outgoing_read()
 *             if len(data):             # <<<<<<<<<<<<<<
 *                 self._transport.write(data)
 * 
 */
    __pyx_t_6 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(25, 698, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_6 != 0);
    if (__pyx_t_1) {

      /* "uvloop/sslproto.pyx":699
 *             data = self._outgoing_read()
 *             if len(data):
 *                 self._transport.write(data)             # <<<<<<<<<<<<<<
 * 
 *     # Incoming flow
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 699, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_data};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 699, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/sslproto.pyx":698
 *         if not self._ssl_writing_paused:
 *             data = self._outgoing_read()
 *             if len(data):             # <<<<<<<<<<<<<<
 *                 self._transport.write(data)
 * 
 */
    }

    /* "uvloop/sslproto.pyx":696
 *     cdef _process_outgoing(self):
 *         """Send bytes from the outgoing BIO."""
 *         if not self._ssl_writing_paused:             # <<<<<<<<<<<<<<
 *             data = self._outgoing_read()
 *             if len(data):
 */
  }

  /* "uvloop/sslproto.pyx":694
 *             pass
 * 
 *     cdef _process_outgoing(self):             # <<<<<<<<<<<<<<
 *         """Send bytes from the outgoing BIO."""
 *         if not self._ssl_writing_paused:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._process_outgoing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":703
 *     # Incoming flow
 * 
 *     cdef _do_read(self):             # <<<<<<<<<<<<<<
 *         if self._state != WRAPPED:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_read", 1);

  /* "uvloop/sslproto.pyx":704
 * 
 *     cdef _do_read(self):
 *         if self._state != WRAPPED:             # <<<<<<<<<<<<<<
 *             return
 *         try:
 */
  __pyx_t_1 = (__pyx_v_self->_state != __pyx_e_6uvloop_4loop_WRAPPED);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":705
 *     cdef _do_read(self):
 *         if self._state != WRAPPED:
 *             return             # <<<<<<<<<<<<<<
 *         try:
 *             if not self._app_reading_paused:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/sslproto.pyx":704
 * 
 *     cdef _do_read(self):
 *         if self._state != WRAPPED:             # <<<<<<<<<<<<<<
 *             return
 *         try:
 */
  }

  /* "uvloop/sslproto.pyx":706
 *         if self._state != WRAPPED:
 *             return
 *         try:             # <<<<<<<<<<<<<<
 *             if not self._app_reading_paused:
 *                 if self._app_protocol_is_buffer:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":707
 *             return
 *         try:
 *             if not self._app_reading_paused:             # <<<<<<<<<<<<<<
 *                 if self._app_protocol_is_buffer:
 *                     self._do_read__buffered()
 */
      __pyx_t_1 = (!__pyx_v_self->_app_reading_paused);
      if (__pyx_t_1) {

        /* "uvloop/sslproto.pyx":708
 *         try:
 *             if not self._app_reading_paused:
 *                 if self._app_protocol_is_buffer:             # <<<<<<<<<<<<<<
 *                     self._do_read__buffered()
 *                 else:
 */
        if (__pyx_v_self->_app_protocol_is_buffer) {

          /* "uvloop/sslproto.pyx":709
 *             if not self._app_reading_paused:
 *                 if self._app_protocol_is_buffer:
 *                     self._do_read__buffered()             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._do_read__copied()
 */
          __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read__buffered(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 709, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/sslproto.pyx":708
 *         try:
 *             if not self._app_reading_paused:
 *                 if self._app_protocol_is_buffer:             # <<<<<<<<<<<<<<
 *                     self._do_read__buffered()
 *                 else:
 */
          goto __pyx_L11;
        }

        /* "uvloop/sslproto.pyx":711
 *                     self._do_read__buffered()
 *                 else:
 *                     self._do_read__copied()             # <<<<<<<<<<<<<<
 *                 if self._write_backlog:
 *                     self._do_write()
 */
        /*else*/ {
          __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read__copied(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 711, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __pyx_L11:;

        /* "uvloop/sslproto.pyx":712
 *                 else:
 *                     self._do_read__copied()
 *                 if self._write_backlog:             # <<<<<<<<<<<<<<
 *                     self._do_write()
 *                 self._process_outgoing()
 */
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_write_backlog); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 712, __pyx_L4_error)
        if (__pyx_t_1) {

          /* "uvloop/sslproto.pyx":713
 *                     self._do_read__copied()
 *                 if self._write_backlog:
 *                     self._do_write()             # <<<<<<<<<<<<<<
 *                 self._process_outgoing()
 *                 self._control_app_writing()
 */
          __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_write(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 713, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "uvloop/sslproto.pyx":712
 *                 else:
 *                     self._do_read__copied()
 *                 if self._write_backlog:             # <<<<<<<<<<<<<<
 *                     self._do_write()
 *                 self._process_outgoing()
 */
        }

        /* "uvloop/sslproto.pyx":714
 *                 if self._write_backlog:
 *                     self._do_write()
 *                 self._process_outgoing()             # <<<<<<<<<<<<<<
 *                 self._control_app_writing()
 *             self._control_ssl_reading()
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_process_outgoing(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 714, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/sslproto.pyx":715
 *                     self._do_write()
 *                 self._process_outgoing()
 *                 self._control_app_writing()             # <<<<<<<<<<<<<<
 *             self._control_ssl_reading()
 *         except Exception as ex:
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_control_app_writing(__pyx_v_self, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 715, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "uvloop/sslproto.pyx":707
 *             return
 *         try:
 *             if not self._app_reading_paused:             # <<<<<<<<<<<<<<
 *                 if self._app_protocol_is_buffer:
 *                     self._do_read__buffered()
 */
      }

      /* "uvloop/sslproto.pyx":716
 *                 self._process_outgoing()
 *                 self._control_app_writing()
 *             self._control_ssl_reading()             # <<<<<<<<<<<<<<
 *         except Exception as ex:
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_control_ssl_reading(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 716, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/sslproto.pyx":706
 *         if self._state != WRAPPED:
 *             return
 *         try:             # <<<<<<<<<<<<<<
 *             if not self._app_reading_paused:
 *                 if self._app_protocol_is_buffer:
 */
    }
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/sslproto.pyx":717
 *                 self._control_app_writing()
 *             self._control_ssl_reading()
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_6) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(25, 717, __pyx_L6_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_ex = __pyx_t_7;
      /*try:*/ {

        /* "uvloop/sslproto.pyx":718
 *             self._control_ssl_reading()
 *         except Exception as ex:
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_read__buffered(self):
 */
        __pyx_t_10.__pyx_n = 1;
        __pyx_t_10.message = __pyx_kp_u_Fatal_error_on_SSL_protocol;
        __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_fatal_error(__pyx_v_self, __pyx_v_ex, &__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(25, 718, __pyx_L18_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }

      /* "uvloop/sslproto.pyx":717
 *                 self._control_app_writing()
 *             self._control_ssl_reading()
 *         except Exception as ex:             # <<<<<<<<<<<<<<
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          goto __pyx_L19;
        }
        __pyx_L18_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_6 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L6_except_error;
        }
        __pyx_L19:;
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;

    /* "uvloop/sslproto.pyx":706
 *         if self._state != WRAPPED:
 *             return
 *         try:             # <<<<<<<<<<<<<<
 *             if not self._app_reading_paused:
 *                 if self._app_protocol_is_buffer:
 */
    __pyx_L6_except_error:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    __pyx_L9_try_end:;
  }

  /* "uvloop/sslproto.pyx":703
 *     # Incoming flow
 * 
 *     cdef _do_read(self):             # <<<<<<<<<<<<<<
 *         if self._state != WRAPPED:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":720
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 *     cdef _do_read__buffered(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_buffer pybuf
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read__buffered(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  Py_buffer __pyx_v_pybuf;
  int __pyx_v_pybuf_inited;
  size_t __pyx_v_wants;
  size_t __pyx_v_offset;
  int __pyx_v_count;
  PyObject *__pyx_v_buf = 0;
  CYTHON_UNUSED PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_read__buffered", 1);

  /* "uvloop/sslproto.pyx":723
 *         cdef:
 *             Py_buffer pybuf
 *             bint pybuf_inited = False             # <<<<<<<<<<<<<<
 *             size_t wants, offset = 0
 *             int count = 1
 */
  __pyx_v_pybuf_inited = 0;

  /* "uvloop/sslproto.pyx":724
 *             Py_buffer pybuf
 *             bint pybuf_inited = False
 *             size_t wants, offset = 0             # <<<<<<<<<<<<<<
 *             int count = 1
 *             object buf
 */
  __pyx_v_offset = 0;

  /* "uvloop/sslproto.pyx":725
 *             bint pybuf_inited = False
 *             size_t wants, offset = 0
 *             int count = 1             # <<<<<<<<<<<<<<
 *             object buf
 * 
 */
  __pyx_v_count = 1;

  /* "uvloop/sslproto.pyx":728
 *             object buf
 * 
 *         buf = self._app_protocol_get_buffer(self._get_read_buffer_size())             # <<<<<<<<<<<<<<
 *         wants = len(buf)
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_get_read_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 728, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->_app_protocol_get_buffer);
  __pyx_t_4 = __pyx_v_self->_app_protocol_get_buffer; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_v_buf = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/sslproto.pyx":729
 * 
 *         buf = self._app_protocol_get_buffer(self._get_read_buffer_size())
 *         wants = len(buf)             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_7 = PyObject_Length(__pyx_v_buf); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(25, 729, __pyx_L1_error)
  __pyx_v_wants = __pyx_t_7;

  /* "uvloop/sslproto.pyx":731
 *         wants = len(buf)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             count = self._sslobj_read(wants, buf)
 * 
 */
  /*try:*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "uvloop/sslproto.pyx":732
 * 
 *         try:
 *             count = self._sslobj_read(wants, buf)             # <<<<<<<<<<<<<<
 * 
 *             if count > 0:
 */
        __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_wants); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 732, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_v_self->_sslobj_read);
        __pyx_t_3 = __pyx_v_self->_sslobj_read; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_4, __pyx_v_buf};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 732, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(25, 732, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_count = __pyx_t_11;

        /* "uvloop/sslproto.pyx":734
 *             count = self._sslobj_read(wants, buf)
 * 
 *             if count > 0:             # <<<<<<<<<<<<<<
 *                 offset = count
 *                 if offset < wants:
 */
        __pyx_t_12 = (__pyx_v_count > 0);
        if (__pyx_t_12) {

          /* "uvloop/sslproto.pyx":735
 * 
 *             if count > 0:
 *                 offset = count             # <<<<<<<<<<<<<<
 *                 if offset < wants:
 *                     PyObject_GetBuffer(buf, &pybuf, PyBUF_WRITABLE)
 */
          __pyx_v_offset = __pyx_v_count;

          /* "uvloop/sslproto.pyx":736
 *             if count > 0:
 *                 offset = count
 *                 if offset < wants:             # <<<<<<<<<<<<<<
 *                     PyObject_GetBuffer(buf, &pybuf, PyBUF_WRITABLE)
 *                     pybuf_inited = True
 */
          __pyx_t_12 = (__pyx_v_offset < __pyx_v_wants);
          if (__pyx_t_12) {

            /* "uvloop/sslproto.pyx":737
 *                 offset = count
 *                 if offset < wants:
 *                     PyObject_GetBuffer(buf, &pybuf, PyBUF_WRITABLE)             # <<<<<<<<<<<<<<
 *                     pybuf_inited = True
 *                 while offset < wants:
 */
            __pyx_t_11 = PyObject_GetBuffer(__pyx_v_buf, (&__pyx_v_pybuf), PyBUF_WRITABLE); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(25, 737, __pyx_L6_error)

            /* "uvloop/sslproto.pyx":738
 *                 if offset < wants:
 *                     PyObject_GetBuffer(buf, &pybuf, PyBUF_WRITABLE)
 *                     pybuf_inited = True             # <<<<<<<<<<<<<<
 *                 while offset < wants:
 *                     buf = PyMemoryView_FromMemory(
 */
            __pyx_v_pybuf_inited = 1;

            /* "uvloop/sslproto.pyx":736
 *             if count > 0:
 *                 offset = count
 *                 if offset < wants:             # <<<<<<<<<<<<<<
 *                     PyObject_GetBuffer(buf, &pybuf, PyBUF_WRITABLE)
 *                     pybuf_inited = True
 */
          }

          /* "uvloop/sslproto.pyx":739
 *                     PyObject_GetBuffer(buf, &pybuf, PyBUF_WRITABLE)
 *                     pybuf_inited = True
 *                 while offset < wants:             # <<<<<<<<<<<<<<
 *                     buf = PyMemoryView_FromMemory(
 *                         (<char*>pybuf.buf) + offset,
 */
          while (1) {
            __pyx_t_12 = (__pyx_v_offset < __pyx_v_wants);
            if (!__pyx_t_12) break;

            /* "uvloop/sslproto.pyx":740
 *                     pybuf_inited = True
 *                 while offset < wants:
 *                     buf = PyMemoryView_FromMemory(             # <<<<<<<<<<<<<<
 *                         (<char*>pybuf.buf) + offset,
 *                         wants - offset,
 */
            __pyx_t_1 = PyMemoryView_FromMemory((((char *)__pyx_v_pybuf.buf) + __pyx_v_offset), (__pyx_v_wants - __pyx_v_offset), PyBUF_WRITE); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 740, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF_SET(__pyx_v_buf, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "uvloop/sslproto.pyx":744
 *                         wants - offset,
 *                         PyBUF_WRITE)
 *                     count = self._sslobj_read(wants - offset, buf)             # <<<<<<<<<<<<<<
 *                     if count > 0:
 *                         offset += count
 */
            __pyx_t_3 = __Pyx_PyInt_FromSize_t((__pyx_v_wants - __pyx_v_offset)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 744, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_INCREF(__pyx_v_self->_sslobj_read);
            __pyx_t_4 = __pyx_v_self->_sslobj_read; __pyx_t_5 = NULL;
            __pyx_t_6 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
                __pyx_t_6 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_buf};
              __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 744, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
            __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(25, 744, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_v_count = __pyx_t_11;

            /* "uvloop/sslproto.pyx":745
 *                         PyBUF_WRITE)
 *                     count = self._sslobj_read(wants - offset, buf)
 *                     if count > 0:             # <<<<<<<<<<<<<<
 *                         offset += count
 *                     else:
 */
            __pyx_t_12 = (__pyx_v_count > 0);
            if (__pyx_t_12) {

              /* "uvloop/sslproto.pyx":746
 *                     count = self._sslobj_read(wants - offset, buf)
 *                     if count > 0:
 *                         offset += count             # <<<<<<<<<<<<<<
 *                     else:
 *                         break
 */
              __pyx_v_offset = (__pyx_v_offset + __pyx_v_count);

              /* "uvloop/sslproto.pyx":745
 *                         PyBUF_WRITE)
 *                     count = self._sslobj_read(wants - offset, buf)
 *                     if count > 0:             # <<<<<<<<<<<<<<
 *                         offset += count
 *                     else:
 */
              goto __pyx_L16;
            }

            /* "uvloop/sslproto.pyx":748
 *                         offset += count
 *                     else:
 *                         break             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._loop._call_soon_handle(
 */
            /*else*/ {
              goto __pyx_L15_break;
            }
            __pyx_L16:;
          }

          /* "uvloop/sslproto.pyx":750
 *                         break
 *                 else:
 *                     self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *                         new_MethodHandle(self._loop,
 *                                          "SSLProtocol._do_read",
 */
          /*else*/ {

            /* "uvloop/sslproto.pyx":751
 *                 else:
 *                     self._loop._call_soon_handle(
 *                         new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                                          "SSLProtocol._do_read",
 *                                          <method_t>self._do_read,
 */
            __pyx_t_1 = ((PyObject *)__pyx_v_self->_loop);
            __Pyx_INCREF(__pyx_t_1);

            /* "uvloop/sslproto.pyx":755
 *                                          <method_t>self._do_read,
 *                                          None,  # current context is good
 *                                          self))             # <<<<<<<<<<<<<<
 *         except ssl_SSLAgainErrors as exc:
 *             pass
 */
            __pyx_t_4 = __pyx_f_6uvloop_4loop_new_MethodHandle(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_1), __pyx_kp_u_SSLProtocol__do_read, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read), Py_None, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 751, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "uvloop/sslproto.pyx":751
 *                 else:
 *                     self._loop._call_soon_handle(
 *                         new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                                          "SSLProtocol._do_read",
 *                                          <method_t>self._do_read,
 */
            if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(25, 751, __pyx_L6_error)

            /* "uvloop/sslproto.pyx":750
 *                         break
 *                 else:
 *                     self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *                         new_MethodHandle(self._loop,
 *                                          "SSLProtocol._do_read",
 */
            __pyx_t_1 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self->_loop, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 750, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_L15_break:;

          /* "uvloop/sslproto.pyx":734
 *             count = self._sslobj_read(wants, buf)
 * 
 *             if count > 0:             # <<<<<<<<<<<<<<
 *                 offset = count
 *                 if offset < wants:
 */
        }

        /* "uvloop/sslproto.pyx":731
 *         wants = len(buf)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             count = self._sslobj_read(wants, buf)
 * 
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "uvloop/sslproto.pyx":756
 *                                          None,  # current context is good
 *                                          self))
 *         except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *             pass
 *         finally:
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read__buffered", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(25, 756, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_exc = __pyx_t_4;
        /*try:*/ {
        }
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L23;
          }
          __pyx_L23:;
        }
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;

      /* "uvloop/sslproto.pyx":731
 *         wants = len(buf)
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             count = self._sslobj_read(wants, buf)
 * 
 */
      __pyx_L8_except_error:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L4_error;
      __pyx_L7_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_L11_try_end:;
    }
  }

  /* "uvloop/sslproto.pyx":759
 *             pass
 *         finally:
 *             if pybuf_inited:             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&pybuf)
 *         if offset > 0:
 */
  /*finally:*/ {
    /*normal exit:*/{
      if (__pyx_v_pybuf_inited) {

        /* "uvloop/sslproto.pyx":760
 *         finally:
 *             if pybuf_inited:
 *                 PyBuffer_Release(&pybuf)             # <<<<<<<<<<<<<<
 *         if offset > 0:
 *             self._app_protocol_buffer_updated(offset)
 */
        PyBuffer_Release((&__pyx_v_pybuf));

        /* "uvloop/sslproto.pyx":759
 *             pass
 *         finally:
 *             if pybuf_inited:             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&pybuf)
 *         if offset > 0:
 */
      }
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_11 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
      {
        if (__pyx_v_pybuf_inited) {

          /* "uvloop/sslproto.pyx":760
 *         finally:
 *             if pybuf_inited:
 *                 PyBuffer_Release(&pybuf)             # <<<<<<<<<<<<<<
 *         if offset > 0:
 *             self._app_protocol_buffer_updated(offset)
 */
          PyBuffer_Release((&__pyx_v_pybuf));

          /* "uvloop/sslproto.pyx":759
 *             pass
 *         finally:
 *             if pybuf_inited:             # <<<<<<<<<<<<<<
 *                 PyBuffer_Release(&pybuf)
 *         if offset > 0:
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "uvloop/sslproto.pyx":761
 *             if pybuf_inited:
 *                 PyBuffer_Release(&pybuf)
 *         if offset > 0:             # <<<<<<<<<<<<<<
 *             self._app_protocol_buffer_updated(offset)
 *         if not count:
 */
  __pyx_t_12 = (__pyx_v_offset > 0);
  if (__pyx_t_12) {

    /* "uvloop/sslproto.pyx":762
 *                 PyBuffer_Release(&pybuf)
 *         if offset > 0:
 *             self._app_protocol_buffer_updated(offset)             # <<<<<<<<<<<<<<
 *         if not count:
 *             # close_notify
 */
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_offset); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_self->_app_protocol_buffer_updated);
    __pyx_t_1 = __pyx_v_self->_app_protocol_buffer_updated; __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":761
 *             if pybuf_inited:
 *                 PyBuffer_Release(&pybuf)
 *         if offset > 0:             # <<<<<<<<<<<<<<
 *             self._app_protocol_buffer_updated(offset)
 *         if not count:
 */
  }

  /* "uvloop/sslproto.pyx":763
 *         if offset > 0:
 *             self._app_protocol_buffer_updated(offset)
 *         if not count:             # <<<<<<<<<<<<<<
 *             # close_notify
 *             self._call_eof_received()
 */
  __pyx_t_12 = (!(__pyx_v_count != 0));
  if (__pyx_t_12) {

    /* "uvloop/sslproto.pyx":765
 *         if not count:
 *             # close_notify
 *             self._call_eof_received()             # <<<<<<<<<<<<<<
 *             self._start_shutdown()
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_call_eof_received(__pyx_v_self, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":766
 *             # close_notify
 *             self._call_eof_received()
 *             self._start_shutdown()             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_read__copied(self):
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_start_shutdown(__pyx_v_self, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/sslproto.pyx":763
 *         if offset > 0:
 *             self._app_protocol_buffer_updated(offset)
 *         if not count:             # <<<<<<<<<<<<<<
 *             # close_notify
 *             self._call_eof_received()
 */
  }

  /* "uvloop/sslproto.pyx":720
 *             self._fatal_error(ex, 'Fatal error on SSL protocol')
 * 
 *     cdef _do_read__buffered(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Py_buffer pybuf
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read__buffered", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buf);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":768
 *             self._start_shutdown()
 * 
 *     cdef _do_read__copied(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list data
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__do_read__copied(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_first = 0;
  PyObject *__pyx_v_chunk = 0;
  int __pyx_v_zero;
  int __pyx_v_one;
  CYTHON_UNUSED PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_read__copied", 1);

  /* "uvloop/sslproto.pyx":771
 *         cdef:
 *             list data
 *             bytes first, chunk = b'1'             # <<<<<<<<<<<<<<
 *             bint zero = True, one = False
 * 
 */
  __Pyx_INCREF(__pyx_kp_b_1);
  __pyx_v_chunk = __pyx_kp_b_1;

  /* "uvloop/sslproto.pyx":772
 *             list data
 *             bytes first, chunk = b'1'
 *             bint zero = True, one = False             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_v_zero = 1;
  __pyx_v_one = 0;

  /* "uvloop/sslproto.pyx":774
 *             bint zero = True, one = False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "uvloop/sslproto.pyx":775
 * 
 *         try:
 *             while True:             # <<<<<<<<<<<<<<
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)
 *                 if not chunk:
 */
      while (1) {

        /* "uvloop/sslproto.pyx":776
 *         try:
 *             while True:
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)             # <<<<<<<<<<<<<<
 *                 if not chunk:
 *                     break
 */
        __pyx_t_5 = __Pyx_PyInt_From___pyx_anon_enum(__pyx_e_6uvloop_4loop_SSL_READ_MAX_SIZE); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 776, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_v_self->_sslobj_read);
        __pyx_t_6 = __pyx_v_self->_sslobj_read; __pyx_t_7 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_5};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 776, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(25, 776, __pyx_L3_error)
        __Pyx_DECREF_SET(__pyx_v_chunk, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "uvloop/sslproto.pyx":777
 *             while True:
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)
 *                 if not chunk:             # <<<<<<<<<<<<<<
 *                     break
 *                 if zero:
 */
        __pyx_t_9 = (__pyx_v_chunk != Py_None)&&(PyBytes_GET_SIZE(__pyx_v_chunk) != 0);
        __pyx_t_10 = (!__pyx_t_9);
        if (__pyx_t_10) {

          /* "uvloop/sslproto.pyx":778
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)
 *                 if not chunk:
 *                     break             # <<<<<<<<<<<<<<
 *                 if zero:
 *                     zero = False
 */
          goto __pyx_L10_break;

          /* "uvloop/sslproto.pyx":777
 *             while True:
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)
 *                 if not chunk:             # <<<<<<<<<<<<<<
 *                     break
 *                 if zero:
 */
        }

        /* "uvloop/sslproto.pyx":779
 *                 if not chunk:
 *                     break
 *                 if zero:             # <<<<<<<<<<<<<<
 *                     zero = False
 *                     one = True
 */
        if (__pyx_v_zero) {

          /* "uvloop/sslproto.pyx":780
 *                     break
 *                 if zero:
 *                     zero = False             # <<<<<<<<<<<<<<
 *                     one = True
 *                     first = chunk
 */
          __pyx_v_zero = 0;

          /* "uvloop/sslproto.pyx":781
 *                 if zero:
 *                     zero = False
 *                     one = True             # <<<<<<<<<<<<<<
 *                     first = chunk
 *                 elif one:
 */
          __pyx_v_one = 1;

          /* "uvloop/sslproto.pyx":782
 *                     zero = False
 *                     one = True
 *                     first = chunk             # <<<<<<<<<<<<<<
 *                 elif one:
 *                     one = False
 */
          __Pyx_INCREF(__pyx_v_chunk);
          __Pyx_XDECREF_SET(__pyx_v_first, __pyx_v_chunk);

          /* "uvloop/sslproto.pyx":779
 *                 if not chunk:
 *                     break
 *                 if zero:             # <<<<<<<<<<<<<<
 *                     zero = False
 *                     one = True
 */
          goto __pyx_L12;
        }

        /* "uvloop/sslproto.pyx":783
 *                     one = True
 *                     first = chunk
 *                 elif one:             # <<<<<<<<<<<<<<
 *                     one = False
 *                     data = [first, chunk]
 */
        if (__pyx_v_one) {

          /* "uvloop/sslproto.pyx":784
 *                     first = chunk
 *                 elif one:
 *                     one = False             # <<<<<<<<<<<<<<
 *                     data = [first, chunk]
 *                 else:
 */
          __pyx_v_one = 0;

          /* "uvloop/sslproto.pyx":785
 *                 elif one:
 *                     one = False
 *                     data = [first, chunk]             # <<<<<<<<<<<<<<
 *                 else:
 *                     data.append(chunk)
 */
          if (unlikely(!__pyx_v_first)) { __Pyx_RaiseUnboundLocalError("first"); __PYX_ERR(25, 785, __pyx_L3_error) }
          __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 785, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_v_first);
          __Pyx_GIVEREF(__pyx_v_first);
          if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_first)) __PYX_ERR(25, 785, __pyx_L3_error);
          __Pyx_INCREF(__pyx_v_chunk);
          __Pyx_GIVEREF(__pyx_v_chunk);
          if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_v_chunk)) __PYX_ERR(25, 785, __pyx_L3_error);
          __Pyx_XDECREF_SET(__pyx_v_data, ((PyObject*)__pyx_t_4));
          __pyx_t_4 = 0;

          /* "uvloop/sslproto.pyx":783
 *                     one = True
 *                     first = chunk
 *                 elif one:             # <<<<<<<<<<<<<<
 *                     one = False
 *                     data = [first, chunk]
 */
          goto __pyx_L12;
        }

        /* "uvloop/sslproto.pyx":787
 *                     data = [first, chunk]
 *                 else:
 *                     data.append(chunk)             # <<<<<<<<<<<<<<
 *         except ssl_SSLAgainErrors as exc:
 *             pass
 */
        /*else*/ {
          if (unlikely(!__pyx_v_data)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(25, 787, __pyx_L3_error) }
          __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_data, __pyx_v_chunk); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(25, 787, __pyx_L3_error)
        }
        __pyx_L12:;
      }
      __pyx_L10_break:;

      /* "uvloop/sslproto.pyx":774
 *             bint zero = True, one = False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "uvloop/sslproto.pyx":788
 *                 else:
 *                     data.append(chunk)
 *         except ssl_SSLAgainErrors as exc:             # <<<<<<<<<<<<<<
 *             pass
 *         if one:
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read__copied", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(25, 788, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_exc = __pyx_t_6;
      /*try:*/ {
      }
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L19;
        }
        __pyx_L19:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "uvloop/sslproto.pyx":774
 *             bint zero = True, one = False
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 chunk = self._sslobj_read(SSL_READ_MAX_SIZE)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "uvloop/sslproto.pyx":790
 *         except ssl_SSLAgainErrors as exc:
 *             pass
 *         if one:             # <<<<<<<<<<<<<<
 *             self._app_protocol.data_received(first)
 *         elif not zero:
 */
  if (__pyx_v_one) {

    /* "uvloop/sslproto.pyx":791
 *             pass
 *         if one:
 *             self._app_protocol.data_received(first)             # <<<<<<<<<<<<<<
 *         elif not zero:
 *             self._app_protocol.data_received(b''.join(data))
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_data_received); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(!__pyx_v_first)) { __Pyx_RaiseUnboundLocalError("first"); __PYX_ERR(25, 791, __pyx_L1_error) }
    __pyx_t_4 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_first};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 791, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/sslproto.pyx":790
 *         except ssl_SSLAgainErrors as exc:
 *             pass
 *         if one:             # <<<<<<<<<<<<<<
 *             self._app_protocol.data_received(first)
 *         elif not zero:
 */
    goto __pyx_L20;
  }

  /* "uvloop/sslproto.pyx":792
 *         if one:
 *             self._app_protocol.data_received(first)
 *         elif not zero:             # <<<<<<<<<<<<<<
 *             self._app_protocol.data_received(b''.join(data))
 *         if not chunk:
 */
  __pyx_t_10 = (!__pyx_v_zero);
  if (__pyx_t_10) {

    /* "uvloop/sslproto.pyx":793
 *             self._app_protocol.data_received(first)
 *         elif not zero:
 *             self._app_protocol.data_received(b''.join(data))             # <<<<<<<<<<<<<<
 *         if not chunk:
 *             # close_notify
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_data_received); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(!__pyx_v_data)) { __Pyx_RaiseUnboundLocalError("data"); __PYX_ERR(25, 793, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__10, __pyx_v_data); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_4};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/sslproto.pyx":792
 *         if one:
 *             self._app_protocol.data_received(first)
 *         elif not zero:             # <<<<<<<<<<<<<<
 *             self._app_protocol.data_received(b''.join(data))
 *         if not chunk:
 */
  }
  __pyx_L20:;

  /* "uvloop/sslproto.pyx":794
 *         elif not zero:
 *             self._app_protocol.data_received(b''.join(data))
 *         if not chunk:             # <<<<<<<<<<<<<<
 *             # close_notify
 *             self._call_eof_received()
 */
  __pyx_t_10 = (__pyx_v_chunk != Py_None)&&(PyBytes_GET_SIZE(__pyx_v_chunk) != 0);
  __pyx_t_9 = (!__pyx_t_10);
  if (__pyx_t_9) {

    /* "uvloop/sslproto.pyx":796
 *         if not chunk:
 *             # close_notify
 *             self._call_eof_received()             # <<<<<<<<<<<<<<
 *             self._start_shutdown()
 * 
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_call_eof_received(__pyx_v_self, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/sslproto.pyx":797
 *             # close_notify
 *             self._call_eof_received()
 *             self._start_shutdown()             # <<<<<<<<<<<<<<
 * 
 *     cdef _call_eof_received(self, object context=None):
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_start_shutdown(__pyx_v_self, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/sslproto.pyx":794
 *         elif not zero:
 *             self._app_protocol.data_received(b''.join(data))
 *         if not chunk:             # <<<<<<<<<<<<<<
 *             # close_notify
 *             self._call_eof_received()
 */
  }

  /* "uvloop/sslproto.pyx":768
 *             self._start_shutdown()
 * 
 *     cdef _do_read__copied(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             list data
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._do_read__copied", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_first);
  __Pyx_XDECREF(__pyx_v_chunk);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":799
 *             self._start_shutdown()
 * 
 *     cdef _call_eof_received(self, object context=None):             # <<<<<<<<<<<<<<
 *         if self._app_state == STATE_CON_MADE:
 *             self._app_state = STATE_EOF
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__call_eof_received(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__call_eof_received *__pyx_optional_args) {
  PyObject *__pyx_v_context = ((PyObject *)Py_None);
  PyObject *__pyx_v_keep_open = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_call_eof_received", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_context = __pyx_optional_args->context;
    }
  }

  /* "uvloop/sslproto.pyx":800
 * 
 *     cdef _call_eof_received(self, object context=None):
 *         if self._app_state == STATE_CON_MADE:             # <<<<<<<<<<<<<<
 *             self._app_state = STATE_EOF
 *             try:
 */
  __pyx_t_1 = (__pyx_v_self->_app_state == __pyx_e_6uvloop_4loop_STATE_CON_MADE);
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":801
 *     cdef _call_eof_received(self, object context=None):
 *         if self._app_state == STATE_CON_MADE:
 *             self._app_state = STATE_EOF             # <<<<<<<<<<<<<<
 *             try:
 *                 if context is None:
 */
    __pyx_v_self->_app_state = __pyx_e_6uvloop_4loop_STATE_EOF;

    /* "uvloop/sslproto.pyx":802
 *         if self._app_state == STATE_CON_MADE:
 *             self._app_state = STATE_EOF
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/sslproto.pyx":803
 *             self._app_state = STATE_EOF
 *             try:
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     # If the caller didn't provide a context, we assume the
 *                     # caller is already in the right context, which is usually
 */
        __pyx_t_1 = (__pyx_v_context == Py_None);
        if (__pyx_t_1) {

          /* "uvloop/sslproto.pyx":807
 *                     # caller is already in the right context, which is usually
 *                     # inside the upstream callbacks like buffer_updated()
 *                     keep_open = self._app_protocol.eof_received()             # <<<<<<<<<<<<<<
 *                 else:
 *                     keep_open = run_in_context(
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_eof_received); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 807, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
            __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 807, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __pyx_v_keep_open = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "uvloop/sslproto.pyx":803
 *             self._app_state = STATE_EOF
 *             try:
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     # If the caller didn't provide a context, we assume the
 *                     # caller is already in the right context, which is usually
 */
          goto __pyx_L10;
        }

        /* "uvloop/sslproto.pyx":809
 *                     keep_open = self._app_protocol.eof_received()
 *                 else:
 *                     keep_open = run_in_context(             # <<<<<<<<<<<<<<
 *                         context, self._app_protocol.eof_received,
 *                     )
 */
        /*else*/ {

          /* "uvloop/sslproto.pyx":810
 *                 else:
 *                     keep_open = run_in_context(
 *                         context, self._app_protocol.eof_received,             # <<<<<<<<<<<<<<
 *                     )
 *             except (KeyboardInterrupt, SystemExit):
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_eof_received); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 810, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "uvloop/sslproto.pyx":809
 *                     keep_open = self._app_protocol.eof_received()
 *                 else:
 *                     keep_open = run_in_context(             # <<<<<<<<<<<<<<
 *                         context, self._app_protocol.eof_received,
 *                     )
 */
          __pyx_t_6 = __pyx_f_6uvloop_4loop_run_in_context(__pyx_v_context, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 809, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_keep_open = __pyx_t_6;
          __pyx_t_6 = 0;
        }
        __pyx_L10:;

        /* "uvloop/sslproto.pyx":802
 *         if self._app_state == STATE_CON_MADE:
 *             self._app_state = STATE_EOF
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
      }

      /* "uvloop/sslproto.pyx":817
 *                 self._fatal_error(ex, 'Error calling eof_received()')
 *             else:
 *                 if keep_open:             # <<<<<<<<<<<<<<
 *                     aio_logger.warning('returning true from eof_received() '
 *                                        'has no effect when using ssl')
 */
      /*else:*/ {
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_keep_open); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 817, __pyx_L6_except_error)
        if (__pyx_t_1) {

          /* "uvloop/sslproto.pyx":818
 *             else:
 *                 if keep_open:
 *                     aio_logger.warning('returning true from eof_received() '             # <<<<<<<<<<<<<<
 *                                        'has no effect when using ssl')
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 818, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_returning_true_from_eof_received};
            __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 818, __pyx_L6_except_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

          /* "uvloop/sslproto.pyx":817
 *                 self._fatal_error(ex, 'Error calling eof_received()')
 *             else:
 *                 if keep_open:             # <<<<<<<<<<<<<<
 *                     aio_logger.warning('returning true from eof_received() '
 *                                        'has no effect when using ssl')
 */
        }
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "uvloop/sslproto.pyx":812
 *                         context, self._app_protocol.eof_received,
 *                     )
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException as ex:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.SSLProtocol._call_eof_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_7) < 0) __PYX_ERR(25, 812, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_7);

        /* "uvloop/sslproto.pyx":813
 *                     )
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException as ex:
 *                 self._fatal_error(ex, 'Error calling eof_received()')
 */
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_7);
        __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_7 = 0; 
        __PYX_ERR(25, 813, __pyx_L6_except_error)
      }

      /* "uvloop/sslproto.pyx":814
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as ex:             # <<<<<<<<<<<<<<
 *                 self._fatal_error(ex, 'Error calling eof_received()')
 *             else:
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("uvloop.loop.SSLProtocol._call_eof_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(25, 814, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __pyx_v_ex = __pyx_t_5;
        /*try:*/ {

          /* "uvloop/sslproto.pyx":815
 *                 raise
 *             except BaseException as ex:
 *                 self._fatal_error(ex, 'Error calling eof_received()')             # <<<<<<<<<<<<<<
 *             else:
 *                 if keep_open:
 */
          __pyx_t_11.__pyx_n = 1;
          __pyx_t_11.message = __pyx_kp_u_Error_calling_eof_received;
          __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_fatal_error(__pyx_v_self, __pyx_v_ex, &__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(25, 815, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }

        /* "uvloop/sslproto.pyx":814
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as ex:             # <<<<<<<<<<<<<<
 *                 self._fatal_error(ex, 'Error calling eof_received()')
 *             else:
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            goto __pyx_L20;
          }
          __pyx_L19_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __pyx_t_9 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_ex); __pyx_v_ex = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
            }
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
            __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
            __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
            goto __pyx_L6_except_error;
          }
          __pyx_L20:;
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "uvloop/sslproto.pyx":802
 *         if self._app_state == STATE_CON_MADE:
 *             self._app_state = STATE_EOF
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "uvloop/sslproto.pyx":800
 * 
 *     cdef _call_eof_received(self, object context=None):
 *         if self._app_state == STATE_CON_MADE:             # <<<<<<<<<<<<<<
 *             self._app_state = STATE_EOF
 *             try:
 */
  }

  /* "uvloop/sslproto.pyx":799
 *             self._start_shutdown()
 * 
 *     cdef _call_eof_received(self, object context=None):             # <<<<<<<<<<<<<<
 *         if self._app_state == STATE_CON_MADE:
 *             self._app_state = STATE_EOF
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._call_eof_received", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_keep_open);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":823
 *     # Flow control for writes from APP socket
 * 
 *     cdef _control_app_writing(self, object context=None):             # <<<<<<<<<<<<<<
 *         cdef size_t size = self._get_write_buffer_size()
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__control_app_writing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing *__pyx_optional_args) {
  PyObject *__pyx_v_context = ((PyObject *)Py_None);
  size_t __pyx_v_size;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  char const *__pyx_t_24;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_control_app_writing", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_context = __pyx_optional_args->context;
    }
  }

  /* "uvloop/sslproto.pyx":824
 * 
 *     cdef _control_app_writing(self, object context=None):
 *         cdef size_t size = self._get_write_buffer_size()             # <<<<<<<<<<<<<<
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:
 *             self._app_writing_paused = True
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_get_write_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 824, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "uvloop/sslproto.pyx":825
 *     cdef _control_app_writing(self, object context=None):
 *         cdef size_t size = self._get_write_buffer_size()
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:             # <<<<<<<<<<<<<<
 *             self._app_writing_paused = True
 *             try:
 */
  __pyx_t_3 = (__pyx_v_size >= __pyx_v_self->_outgoing_high_water);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (!__pyx_v_self->_app_writing_paused);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":826
 *         cdef size_t size = self._get_write_buffer_size()
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:
 *             self._app_writing_paused = True             # <<<<<<<<<<<<<<
 *             try:
 *                 if context is None:
 */
    __pyx_v_self->_app_writing_paused = 1;

    /* "uvloop/sslproto.pyx":827
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:
 *             self._app_writing_paused = True
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "uvloop/sslproto.pyx":828
 *             self._app_writing_paused = True
 *             try:
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     # If the caller didn't provide a context, we assume the
 *                     # caller is already in the right context, which is usually
 */
        __pyx_t_2 = (__pyx_v_context == Py_None);
        if (__pyx_t_2) {

          /* "uvloop/sslproto.pyx":832
 *                     # caller is already in the right context, which is usually
 *                     # inside the upstream callbacks like buffer_updated()
 *                     self._app_protocol.pause_writing()             # <<<<<<<<<<<<<<
 *                 else:
 *                     run_in_context(context, self._app_protocol.pause_writing)
 */
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_pause_writing); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 832, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 0+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 832, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "uvloop/sslproto.pyx":828
 *             self._app_writing_paused = True
 *             try:
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     # If the caller didn't provide a context, we assume the
 *                     # caller is already in the right context, which is usually
 */
          goto __pyx_L12;
        }

        /* "uvloop/sslproto.pyx":834
 *                     self._app_protocol.pause_writing()
 *                 else:
 *                     run_in_context(context, self._app_protocol.pause_writing)             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        /*else*/ {
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_pause_writing); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 834, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_8 = __pyx_f_6uvloop_4loop_run_in_context(__pyx_v_context, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 834, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_L12:;

        /* "uvloop/sslproto.pyx":827
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:
 *             self._app_writing_paused = True
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L11_try_end;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/sslproto.pyx":835
 *                 else:
 *                     run_in_context(context, self._app_protocol.pause_writing)
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException as exc:
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.SSLProtocol._control_app_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_9) < 0) __PYX_ERR(25, 835, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "uvloop/sslproto.pyx":836
 *                     run_in_context(context, self._app_protocol.pause_writing)
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 */
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ErrRestoreWithState(__pyx_t_8, __pyx_t_7, __pyx_t_9);
        __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_9 = 0; 
        __PYX_ERR(25, 836, __pyx_L8_except_error)
      }

      /* "uvloop/sslproto.pyx":837
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("uvloop.loop.SSLProtocol._control_app_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(25, 837, __pyx_L8_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_exc = __pyx_t_7;
        /*try:*/ {

          /* "uvloop/sslproto.pyx":838
 *                 raise
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                     'message': 'protocol.pause_writing() failed',
 *                     'exception': exc,
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_13)) __PYX_ERR(25, 838, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_13);

          /* "uvloop/sslproto.pyx":839
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',             # <<<<<<<<<<<<<<
 *                     'exception': exc,
 *                     'transport': self._app_transport,
 */
          __pyx_t_14 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_14)) __PYX_ERR(25, 839, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_message, __pyx_kp_u_protocol_pause_writing_failed) < 0) __PYX_ERR(25, 839, __pyx_L20_error)

          /* "uvloop/sslproto.pyx":840
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',
 *                     'exception': exc,             # <<<<<<<<<<<<<<
 *                     'transport': self._app_transport,
 *                     'protocol': self,
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(25, 839, __pyx_L20_error)

          /* "uvloop/sslproto.pyx":841
 *                     'message': 'protocol.pause_writing() failed',
 *                     'exception': exc,
 *                     'transport': self._app_transport,             # <<<<<<<<<<<<<<
 *                     'protocol': self,
 *                 })
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_transport, ((PyObject *)__pyx_v_self->_app_transport)) < 0) __PYX_ERR(25, 839, __pyx_L20_error)

          /* "uvloop/sslproto.pyx":842
 *                     'exception': exc,
 *                     'transport': self._app_transport,
 *                     'protocol': self,             # <<<<<<<<<<<<<<
 *                 })
 *         elif size <= self._outgoing_low_water and self._app_writing_paused:
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_protocol, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(25, 839, __pyx_L20_error)
          __pyx_t_15 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_14};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(25, 838, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }

        /* "uvloop/sslproto.pyx":837
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.pause_writing() failed',
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L21;
          }
          __pyx_L20_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_23);
            __pyx_t_11 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
            }
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
            __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
            goto __pyx_L8_except_error;
          }
          __pyx_L21:;
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_exception_handled;
      }
      goto __pyx_L8_except_error;

      /* "uvloop/sslproto.pyx":827
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:
 *             self._app_writing_paused = True
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
      __pyx_L8_except_error:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L7_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_L11_try_end:;
    }

    /* "uvloop/sslproto.pyx":825
 *     cdef _control_app_writing(self, object context=None):
 *         cdef size_t size = self._get_write_buffer_size()
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:             # <<<<<<<<<<<<<<
 *             self._app_writing_paused = True
 *             try:
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":844
 *                     'protocol': self,
 *                 })
 *         elif size <= self._outgoing_low_water and self._app_writing_paused:             # <<<<<<<<<<<<<<
 *             self._app_writing_paused = False
 *             try:
 */
  __pyx_t_3 = (__pyx_v_size <= __pyx_v_self->_outgoing_low_water);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L26_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_self->_app_writing_paused;
  __pyx_L26_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":845
 *                 })
 *         elif size <= self._outgoing_low_water and self._app_writing_paused:
 *             self._app_writing_paused = False             # <<<<<<<<<<<<<<
 *             try:
 *                 if context is None:
 */
    __pyx_v_self->_app_writing_paused = 0;

    /* "uvloop/sslproto.pyx":846
 *         elif size <= self._outgoing_low_water and self._app_writing_paused:
 *             self._app_writing_paused = False
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "uvloop/sslproto.pyx":847
 *             self._app_writing_paused = False
 *             try:
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     # If the caller didn't provide a context, we assume the
 *                     # caller is already in the right context, which is usually
 */
        __pyx_t_2 = (__pyx_v_context == Py_None);
        if (__pyx_t_2) {

          /* "uvloop/sslproto.pyx":851
 *                     # caller is already in the right context, which is usually
 *                     # inside the upstream callbacks like resume_writing()
 *                     self._app_protocol.resume_writing()             # <<<<<<<<<<<<<<
 *                 else:
 *                     run_in_context(context, self._app_protocol.resume_writing)
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_resume_writing); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 851, __pyx_L28_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
            __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_10, 0+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 851, __pyx_L28_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "uvloop/sslproto.pyx":847
 *             self._app_writing_paused = False
 *             try:
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     # If the caller didn't provide a context, we assume the
 *                     # caller is already in the right context, which is usually
 */
          goto __pyx_L34;
        }

        /* "uvloop/sslproto.pyx":853
 *                     self._app_protocol.resume_writing()
 *                 else:
 *                     run_in_context(context, self._app_protocol.resume_writing)             # <<<<<<<<<<<<<<
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 */
        /*else*/ {
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_app_protocol, __pyx_n_s_resume_writing); if (unlikely(!__pyx_t_8)) __PYX_ERR(25, 853, __pyx_L28_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_7 = __pyx_f_6uvloop_4loop_run_in_context(__pyx_v_context, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(25, 853, __pyx_L28_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __pyx_L34:;

        /* "uvloop/sslproto.pyx":846
 *         elif size <= self._outgoing_low_water and self._app_writing_paused:
 *             self._app_writing_paused = False
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L33_try_end;
      __pyx_L28_error:;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/sslproto.pyx":854
 *                 else:
 *                     run_in_context(context, self._app_protocol.resume_writing)
 *             except (KeyboardInterrupt, SystemExit):             # <<<<<<<<<<<<<<
 *                 raise
 *             except BaseException as exc:
 */
      __pyx_t_16 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_KeyboardInterrupt, __pyx_builtin_SystemExit);
      if (__pyx_t_16) {
        __Pyx_AddTraceback("uvloop.loop.SSLProtocol._control_app_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(25, 854, __pyx_L30_except_error)
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "uvloop/sslproto.pyx":855
 *                     run_in_context(context, self._app_protocol.resume_writing)
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise             # <<<<<<<<<<<<<<
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 */
        __Pyx_GIVEREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; 
        __PYX_ERR(25, 855, __pyx_L30_except_error)
      }

      /* "uvloop/sslproto.pyx":856
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',
 */
      __pyx_t_16 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_16) {
        __Pyx_AddTraceback("uvloop.loop.SSLProtocol._control_app_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(25, 856, __pyx_L30_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_v_exc = __pyx_t_8;
        /*try:*/ {

          /* "uvloop/sslproto.pyx":857
 *                 raise
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                     'message': 'protocol.resume_writing() failed',
 *                     'exception': exc,
 */
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_13)) __PYX_ERR(25, 857, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_13);

          /* "uvloop/sslproto.pyx":858
 *             except BaseException as exc:
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',             # <<<<<<<<<<<<<<
 *                     'exception': exc,
 *                     'transport': self._app_transport,
 */
          __pyx_t_14 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_14)) __PYX_ERR(25, 858, __pyx_L42_error)
          __Pyx_GOTREF(__pyx_t_14);
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_message, __pyx_kp_u_protocol_resume_writing_failed) < 0) __PYX_ERR(25, 858, __pyx_L42_error)

          /* "uvloop/sslproto.pyx":859
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',
 *                     'exception': exc,             # <<<<<<<<<<<<<<
 *                     'transport': self._app_transport,
 *                     'protocol': self,
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(25, 858, __pyx_L42_error)

          /* "uvloop/sslproto.pyx":860
 *                     'message': 'protocol.resume_writing() failed',
 *                     'exception': exc,
 *                     'transport': self._app_transport,             # <<<<<<<<<<<<<<
 *                     'protocol': self,
 *                 })
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_transport, ((PyObject *)__pyx_v_self->_app_transport)) < 0) __PYX_ERR(25, 858, __pyx_L42_error)

          /* "uvloop/sslproto.pyx":861
 *                     'exception': exc,
 *                     'transport': self._app_transport,
 *                     'protocol': self,             # <<<<<<<<<<<<<<
 *                 })
 * 
 */
          if (PyDict_SetItem(__pyx_t_14, __pyx_n_u_protocol, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(25, 858, __pyx_L42_error)
          __pyx_t_15 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_14};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(25, 857, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }

        /* "uvloop/sslproto.pyx":856
 *             except (KeyboardInterrupt, SystemExit):
 *                 raise
 *             except BaseException as exc:             # <<<<<<<<<<<<<<
 *                 self._loop.call_exception_handler({
 *                     'message': 'protocol.resume_writing() failed',
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L43;
          }
          __pyx_L42_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_23, &__pyx_t_22, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_23, &__pyx_t_22, &__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_18);
            __pyx_t_16 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_24 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
            }
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_ErrRestore(__pyx_t_23, __pyx_t_22, __pyx_t_21);
            __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
            __pyx_lineno = __pyx_t_16; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_24;
            goto __pyx_L30_except_error;
          }
          __pyx_L43:;
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L29_exception_handled;
      }
      goto __pyx_L30_except_error;

      /* "uvloop/sslproto.pyx":846
 *         elif size <= self._outgoing_low_water and self._app_writing_paused:
 *             self._app_writing_paused = False
 *             try:             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     # If the caller didn't provide a context, we assume the
 */
      __pyx_L30_except_error:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L29_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      __pyx_L33_try_end:;
    }

    /* "uvloop/sslproto.pyx":844
 *                     'protocol': self,
 *                 })
 *         elif size <= self._outgoing_low_water and self._app_writing_paused:             # <<<<<<<<<<<<<<
 *             self._app_writing_paused = False
 *             try:
 */
  }
  __pyx_L3:;

  /* "uvloop/sslproto.pyx":823
 *     # Flow control for writes from APP socket
 * 
 *     cdef _control_app_writing(self, object context=None):             # <<<<<<<<<<<<<<
 *         cdef size_t size = self._get_write_buffer_size()
 *         if size >= self._outgoing_high_water and not self._app_writing_paused:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._control_app_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":864
 *                 })
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._outgoing.pending + self._write_buffer_size
 * 
 */

static size_t __pyx_f_6uvloop_4loop_11SSLProtocol__get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_write_buffer_size", 1);

  /* "uvloop/sslproto.pyx":865
 * 
 *     cdef size_t _get_write_buffer_size(self):
 *         return self._outgoing.pending + self._write_buffer_size             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_write_buffer_limits(self, high=None, low=None):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_outgoing, __pyx_n_s_pending); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_write_buffer_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_size_t(__pyx_t_3); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 865, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":864
 *                 })
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._outgoing.pending + self._write_buffer_size
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._get_write_buffer_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":867
 *         return self._outgoing.pending + self._write_buffer_size
 * 
 *     cdef _set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         high, low = add_flowcontrol_defaults(
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_WRITE)
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits *__pyx_optional_args) {
  PyObject *__pyx_v_high = ((PyObject *)Py_None);
  PyObject *__pyx_v_low = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_write_buffer_limits", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_high = __pyx_optional_args->high;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_low = __pyx_optional_args->low;
      }
    }
  }
  __Pyx_INCREF(__pyx_v_high);
  __Pyx_INCREF(__pyx_v_low);

  /* "uvloop/sslproto.pyx":868
 * 
 *     cdef _set_write_buffer_limits(self, high=None, low=None):
 *         high, low = add_flowcontrol_defaults(             # <<<<<<<<<<<<<<
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_WRITE)
 *         self._outgoing_high_water = high
 */
  __pyx_t_1 = __pyx_f_6uvloop_8includes_11flowcontrol_add_flowcontrol_defaults(__pyx_v_high, __pyx_v_low, __pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER_SSL_WRITE); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(25, 868, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 868, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(25, 868, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(25, 868, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __Pyx_DECREF_SET(__pyx_v_high, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_low, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/sslproto.pyx":870
 *         high, low = add_flowcontrol_defaults(
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_WRITE)
 *         self._outgoing_high_water = high             # <<<<<<<<<<<<<<
 *         self._outgoing_low_water = low
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_v_high); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 870, __pyx_L1_error)
  __pyx_v_self->_outgoing_high_water = __pyx_t_6;

  /* "uvloop/sslproto.pyx":871
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_WRITE)
 *         self._outgoing_high_water = high
 *         self._outgoing_low_water = low             # <<<<<<<<<<<<<<
 * 
 *     # Flow control for reads to APP socket
 */
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_v_low); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 871, __pyx_L1_error)
  __pyx_v_self->_outgoing_low_water = __pyx_t_6;

  /* "uvloop/sslproto.pyx":867
 *         return self._outgoing.pending + self._write_buffer_size
 * 
 *     cdef _set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         high, low = add_flowcontrol_defaults(
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_WRITE)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._set_write_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_high);
  __Pyx_XDECREF(__pyx_v_low);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":875
 *     # Flow control for reads to APP socket
 * 
 *     cdef _pause_reading(self):             # <<<<<<<<<<<<<<
 *         self._app_reading_paused = True
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__pause_reading(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pause_reading", 1);

  /* "uvloop/sslproto.pyx":876
 * 
 *     cdef _pause_reading(self):
 *         self._app_reading_paused = True             # <<<<<<<<<<<<<<
 * 
 *     cdef _resume_reading(self, object context):
 */
  __pyx_v_self->_app_reading_paused = 1;

  /* "uvloop/sslproto.pyx":875
 *     # Flow control for reads to APP socket
 * 
 *     cdef _pause_reading(self):             # <<<<<<<<<<<<<<
 *         self._app_reading_paused = True
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":878
 *         self._app_reading_paused = True
 * 
 *     cdef _resume_reading(self, object context):             # <<<<<<<<<<<<<<
 *         if self._app_reading_paused:
 *             self._app_reading_paused = False
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__resume_reading(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_resume_reading", 1);

  /* "uvloop/sslproto.pyx":879
 * 
 *     cdef _resume_reading(self, object context):
 *         if self._app_reading_paused:             # <<<<<<<<<<<<<<
 *             self._app_reading_paused = False
 *             if self._state == WRAPPED:
 */
  if (__pyx_v_self->_app_reading_paused) {

    /* "uvloop/sslproto.pyx":880
 *     cdef _resume_reading(self, object context):
 *         if self._app_reading_paused:
 *             self._app_reading_paused = False             # <<<<<<<<<<<<<<
 *             if self._state == WRAPPED:
 *                 self._loop._call_soon_handle(
 */
    __pyx_v_self->_app_reading_paused = 0;

    /* "uvloop/sslproto.pyx":881
 *         if self._app_reading_paused:
 *             self._app_reading_paused = False
 *             if self._state == WRAPPED:             # <<<<<<<<<<<<<<
 *                 self._loop._call_soon_handle(
 *                     new_MethodHandle(self._loop,
 */
    __pyx_t_1 = (__pyx_v_self->_state == __pyx_e_6uvloop_4loop_WRAPPED);
    if (__pyx_t_1) {

      /* "uvloop/sslproto.pyx":883
 *             if self._state == WRAPPED:
 *                 self._loop._call_soon_handle(
 *                     new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                                      "SSLProtocol._do_read",
 *                                      <method_t>self._do_read,
 */
      __pyx_t_2 = ((PyObject *)__pyx_v_self->_loop);
      __Pyx_INCREF(__pyx_t_2);

      /* "uvloop/sslproto.pyx":887
 *                                      <method_t>self._do_read,
 *                                      context,
 *                                      self))             # <<<<<<<<<<<<<<
 * 
 *     # Flow control for reads from SSL socket
 */
      __pyx_t_3 = __pyx_f_6uvloop_4loop_new_MethodHandle(((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_2), __pyx_kp_u_SSLProtocol__do_read, ((__pyx_t_6uvloop_4loop_method_t)((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_read), __pyx_v_context, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/sslproto.pyx":883
 *             if self._state == WRAPPED:
 *                 self._loop._call_soon_handle(
 *                     new_MethodHandle(self._loop,             # <<<<<<<<<<<<<<
 *                                      "SSLProtocol._do_read",
 *                                      <method_t>self._do_read,
 */
      if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6uvloop_4loop_Handle))))) __PYX_ERR(25, 883, __pyx_L1_error)

      /* "uvloop/sslproto.pyx":882
 *             self._app_reading_paused = False
 *             if self._state == WRAPPED:
 *                 self._loop._call_soon_handle(             # <<<<<<<<<<<<<<
 *                     new_MethodHandle(self._loop,
 *                                      "SSLProtocol._do_read",
 */
      __pyx_t_2 = __pyx_f_6uvloop_4loop_4Loop__call_soon_handle(__pyx_v_self->_loop, ((struct __pyx_obj_6uvloop_4loop_Handle *)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/sslproto.pyx":881
 *         if self._app_reading_paused:
 *             self._app_reading_paused = False
 *             if self._state == WRAPPED:             # <<<<<<<<<<<<<<
 *                 self._loop._call_soon_handle(
 *                     new_MethodHandle(self._loop,
 */
    }

    /* "uvloop/sslproto.pyx":879
 * 
 *     cdef _resume_reading(self, object context):
 *         if self._app_reading_paused:             # <<<<<<<<<<<<<<
 *             self._app_reading_paused = False
 *             if self._state == WRAPPED:
 */
  }

  /* "uvloop/sslproto.pyx":878
 *         self._app_reading_paused = True
 * 
 *     cdef _resume_reading(self, object context):             # <<<<<<<<<<<<<<
 *         if self._app_reading_paused:
 *             self._app_reading_paused = False
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._resume_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":891
 *     # Flow control for reads from SSL socket
 * 
 *     cdef _control_ssl_reading(self):             # <<<<<<<<<<<<<<
 *         cdef size_t size = self._get_read_buffer_size()
 *         if size >= self._incoming_high_water and not self._ssl_reading_paused:
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__control_ssl_reading(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  size_t __pyx_v_size;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_control_ssl_reading", 1);

  /* "uvloop/sslproto.pyx":892
 * 
 *     cdef _control_ssl_reading(self):
 *         cdef size_t size = self._get_read_buffer_size()             # <<<<<<<<<<<<<<
 *         if size >= self._incoming_high_water and not self._ssl_reading_paused:
 *             self._ssl_reading_paused = True
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_get_read_buffer_size(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(25, 892, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "uvloop/sslproto.pyx":893
 *     cdef _control_ssl_reading(self):
 *         cdef size_t size = self._get_read_buffer_size()
 *         if size >= self._incoming_high_water and not self._ssl_reading_paused:             # <<<<<<<<<<<<<<
 *             self._ssl_reading_paused = True
 *             self._transport.pause_reading()
 */
  __pyx_t_3 = (__pyx_v_size >= __pyx_v_self->_incoming_high_water);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (!__pyx_v_self->_ssl_reading_paused);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":894
 *         cdef size_t size = self._get_read_buffer_size()
 *         if size >= self._incoming_high_water and not self._ssl_reading_paused:
 *             self._ssl_reading_paused = True             # <<<<<<<<<<<<<<
 *             self._transport.pause_reading()
 *         elif size <= self._incoming_low_water and self._ssl_reading_paused:
 */
    __pyx_v_self->_ssl_reading_paused = 1;

    /* "uvloop/sslproto.pyx":895
 *         if size >= self._incoming_high_water and not self._ssl_reading_paused:
 *             self._ssl_reading_paused = True
 *             self._transport.pause_reading()             # <<<<<<<<<<<<<<
 *         elif size <= self._incoming_low_water and self._ssl_reading_paused:
 *             self._ssl_reading_paused = False
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_pause_reading); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/sslproto.pyx":893
 *     cdef _control_ssl_reading(self):
 *         cdef size_t size = self._get_read_buffer_size()
 *         if size >= self._incoming_high_water and not self._ssl_reading_paused:             # <<<<<<<<<<<<<<
 *             self._ssl_reading_paused = True
 *             self._transport.pause_reading()
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":896
 *             self._ssl_reading_paused = True
 *             self._transport.pause_reading()
 *         elif size <= self._incoming_low_water and self._ssl_reading_paused:             # <<<<<<<<<<<<<<
 *             self._ssl_reading_paused = False
 *             self._transport.resume_reading()
 */
  __pyx_t_3 = (__pyx_v_size <= __pyx_v_self->_incoming_low_water);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = __pyx_v_self->_ssl_reading_paused;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/sslproto.pyx":897
 *             self._transport.pause_reading()
 *         elif size <= self._incoming_low_water and self._ssl_reading_paused:
 *             self._ssl_reading_paused = False             # <<<<<<<<<<<<<<
 *             self._transport.resume_reading()
 * 
 */
    __pyx_v_self->_ssl_reading_paused = 0;

    /* "uvloop/sslproto.pyx":898
 *         elif size <= self._incoming_low_water and self._ssl_reading_paused:
 *             self._ssl_reading_paused = False
 *             self._transport.resume_reading()             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_read_buffer_limits(self, high=None, low=None):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_resume_reading); if (unlikely(!__pyx_t_5)) __PYX_ERR(25, 898, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/sslproto.pyx":896
 *             self._ssl_reading_paused = True
 *             self._transport.pause_reading()
 *         elif size <= self._incoming_low_water and self._ssl_reading_paused:             # <<<<<<<<<<<<<<
 *             self._ssl_reading_paused = False
 *             self._transport.resume_reading()
 */
  }
  __pyx_L3:;

  /* "uvloop/sslproto.pyx":891
 *     # Flow control for reads from SSL socket
 * 
 *     cdef _control_ssl_reading(self):             # <<<<<<<<<<<<<<
 *         cdef size_t size = self._get_read_buffer_size()
 *         if size >= self._incoming_high_water and not self._ssl_reading_paused:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._control_ssl_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":900
 *             self._transport.resume_reading()
 * 
 *     cdef _set_read_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         high, low = add_flowcontrol_defaults(
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_READ)
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits *__pyx_optional_args) {
  PyObject *__pyx_v_high = ((PyObject *)Py_None);
  PyObject *__pyx_v_low = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_read_buffer_limits", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_high = __pyx_optional_args->high;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_low = __pyx_optional_args->low;
      }
    }
  }
  __Pyx_INCREF(__pyx_v_high);
  __Pyx_INCREF(__pyx_v_low);

  /* "uvloop/sslproto.pyx":901
 * 
 *     cdef _set_read_buffer_limits(self, high=None, low=None):
 *         high, low = add_flowcontrol_defaults(             # <<<<<<<<<<<<<<
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_READ)
 *         self._incoming_high_water = high
 */
  __pyx_t_1 = __pyx_f_6uvloop_8includes_11flowcontrol_add_flowcontrol_defaults(__pyx_v_high, __pyx_v_low, __pyx_e_6uvloop_4loop_FLOW_CONTROL_HIGH_WATER_SSL_READ); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(25, 901, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(25, 901, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(25, 901, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __Pyx_DECREF_SET(__pyx_v_high, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_low, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/sslproto.pyx":903
 *         high, low = add_flowcontrol_defaults(
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_READ)
 *         self._incoming_high_water = high             # <<<<<<<<<<<<<<
 *         self._incoming_low_water = low
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_v_high); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 903, __pyx_L1_error)
  __pyx_v_self->_incoming_high_water = __pyx_t_6;

  /* "uvloop/sslproto.pyx":904
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_READ)
 *         self._incoming_high_water = high
 *         self._incoming_low_water = low             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t _get_read_buffer_size(self):
 */
  __pyx_t_6 = __Pyx_PyInt_As_size_t(__pyx_v_low); if (unlikely((__pyx_t_6 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 904, __pyx_L1_error)
  __pyx_v_self->_incoming_low_water = __pyx_t_6;

  /* "uvloop/sslproto.pyx":900
 *             self._transport.resume_reading()
 * 
 *     cdef _set_read_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         high, low = add_flowcontrol_defaults(
 *             high, low, FLOW_CONTROL_HIGH_WATER_SSL_READ)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._set_read_buffer_limits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_high);
  __Pyx_XDECREF(__pyx_v_low);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":906
 *         self._incoming_low_water = low
 * 
 *     cdef size_t _get_read_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._incoming.pending
 * 
 */

static size_t __pyx_f_6uvloop_4loop_11SSLProtocol__get_read_buffer_size(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_read_buffer_size", 1);

  /* "uvloop/sslproto.pyx":907
 * 
 *     cdef size_t _get_read_buffer_size(self):
 *         return self._incoming.pending             # <<<<<<<<<<<<<<
 * 
 *     # Flow control for writes to SSL socket
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_incoming, __pyx_n_s_pending); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(25, 907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "uvloop/sslproto.pyx":906
 *         self._incoming_low_water = low
 * 
 *     cdef size_t _get_read_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._incoming.pending
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._get_read_buffer_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":911
 *     # Flow control for writes to SSL socket
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer goes over
 *         the high-water mark.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_19pause_writing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_18pause_writing, "SSLProtocol.pause_writing(self)\nCalled when the low-level transport's buffer goes over\n        the high-water mark.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_19pause_writing = {"pause_writing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_19pause_writing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_18pause_writing};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_19pause_writing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pause_writing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("pause_writing", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "pause_writing", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_18pause_writing(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_18pause_writing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pause_writing", 1);

  /* "uvloop/sslproto.pyx":915
 *         the high-water mark.
 *         """
 *         assert not self._ssl_writing_paused             # <<<<<<<<<<<<<<
 *         self._ssl_writing_paused = True
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (!__pyx_v_self->_ssl_writing_paused);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(25, 915, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(25, 915, __pyx_L1_error)
  #endif

  /* "uvloop/sslproto.pyx":916
 *         """
 *         assert not self._ssl_writing_paused
 *         self._ssl_writing_paused = True             # <<<<<<<<<<<<<<
 * 
 *     def resume_writing(self):
 */
  __pyx_v_self->_ssl_writing_paused = 1;

  /* "uvloop/sslproto.pyx":911
 *     # Flow control for writes to SSL socket
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer goes over
 *         the high-water mark.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.pause_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":918
 *         self._ssl_writing_paused = True
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer drains below
 *         the low-water mark.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_21resume_writing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_20resume_writing, "SSLProtocol.resume_writing(self)\nCalled when the low-level transport's buffer drains below\n        the low-water mark.\n        ");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_21resume_writing = {"resume_writing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_21resume_writing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_20resume_writing};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_21resume_writing(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resume_writing (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("resume_writing", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "resume_writing", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_20resume_writing(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_20resume_writing(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resume_writing", 1);

  /* "uvloop/sslproto.pyx":922
 *         the low-water mark.
 *         """
 *         assert self._ssl_writing_paused             # <<<<<<<<<<<<<<
 *         self._ssl_writing_paused = False
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    if (unlikely(!__pyx_v_self->_ssl_writing_paused)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(25, 922, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(25, 922, __pyx_L1_error)
  #endif

  /* "uvloop/sslproto.pyx":923
 *         """
 *         assert self._ssl_writing_paused
 *         self._ssl_writing_paused = False             # <<<<<<<<<<<<<<
 * 
 *         if self._state == WRAPPED:
 */
  __pyx_v_self->_ssl_writing_paused = 0;

  /* "uvloop/sslproto.pyx":925
 *         self._ssl_writing_paused = False
 * 
 *         if self._state == WRAPPED:             # <<<<<<<<<<<<<<
 *             self._process_outgoing()
 *             self._control_app_writing()
 */
  switch (__pyx_v_self->_state) {
    case __pyx_e_6uvloop_4loop_WRAPPED:

    /* "uvloop/sslproto.pyx":926
 * 
 *         if self._state == WRAPPED:
 *             self._process_outgoing()             # <<<<<<<<<<<<<<
 *             self._control_app_writing()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_process_outgoing(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":927
 *         if self._state == WRAPPED:
 *             self._process_outgoing()
 *             self._control_app_writing()             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == FLUSHING:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_control_app_writing(__pyx_v_self, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":925
 *         self._ssl_writing_paused = False
 * 
 *         if self._state == WRAPPED:             # <<<<<<<<<<<<<<
 *             self._process_outgoing()
 *             self._control_app_writing()
 */
    break;
    case __pyx_e_6uvloop_4loop_FLUSHING:

    /* "uvloop/sslproto.pyx":930
 * 
 *         elif self._state == FLUSHING:
 *             self._do_flush()             # <<<<<<<<<<<<<<
 * 
 *         elif self._state == SHUTDOWN:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_flush(__pyx_v_self, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":929
 *             self._control_app_writing()
 * 
 *         elif self._state == FLUSHING:             # <<<<<<<<<<<<<<
 *             self._do_flush()
 * 
 */
    break;
    case __pyx_e_6uvloop_4loop_SHUTDOWN:

    /* "uvloop/sslproto.pyx":933
 * 
 *         elif self._state == SHUTDOWN:
 *             self._do_shutdown()             # <<<<<<<<<<<<<<
 * 
 *     cdef _fatal_error(self, exc, message='Fatal error on transport'):
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol *)__pyx_v_self->__pyx_vtab)->_do_shutdown(__pyx_v_self, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(25, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/sslproto.pyx":932
 *             self._do_flush()
 * 
 *         elif self._state == SHUTDOWN:             # <<<<<<<<<<<<<<
 *             self._do_shutdown()
 * 
 */
    break;
    default: break;
  }

  /* "uvloop/sslproto.pyx":918
 *         self._ssl_writing_paused = True
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer drains below
 *         the low-water mark.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.resume_writing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/sslproto.pyx":935
 *             self._do_shutdown()
 * 
 *     cdef _fatal_error(self, exc, message='Fatal error on transport'):             # <<<<<<<<<<<<<<
 *         if self._app_transport:
 *             self._app_transport._force_close(exc)
 */

static PyObject *__pyx_f_6uvloop_4loop_11SSLProtocol__fatal_error(struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, PyObject *__pyx_v_exc, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error *__pyx_optional_args) {
  PyObject *__pyx_v_message = ((PyObject *)__pyx_kp_u_Fatal_error_on_transport_2);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_fatal_error", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_message = __pyx_optional_args->message;
    }
  }

  /* "uvloop/sslproto.pyx":936
 * 
 *     cdef _fatal_error(self, exc, message='Fatal error on transport'):
 *         if self._app_transport:             # <<<<<<<<<<<<<<
 *             self._app_transport._force_close(exc)
 *         elif self._transport:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_app_transport)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 936, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":937
 *     cdef _fatal_error(self, exc, message='Fatal error on transport'):
 *         if self._app_transport:
 *             self._app_transport._force_close(exc)             # <<<<<<<<<<<<<<
 *         elif self._transport:
 *             self._transport._force_close(exc)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_app_transport), __pyx_n_s_force_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 937, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_exc};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 937, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":936
 * 
 *     cdef _fatal_error(self, exc, message='Fatal error on transport'):
 *         if self._app_transport:             # <<<<<<<<<<<<<<
 *             self._app_transport._force_close(exc)
 *         elif self._transport:
 */
    goto __pyx_L3;
  }

  /* "uvloop/sslproto.pyx":938
 *         if self._app_transport:
 *             self._app_transport._force_close(exc)
 *         elif self._transport:             # <<<<<<<<<<<<<<
 *             self._transport._force_close(exc)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_transport); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 938, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":939
 *             self._app_transport._force_close(exc)
 *         elif self._transport:
 *             self._transport._force_close(exc)             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(exc, OSError):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_transport, __pyx_n_s_force_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_exc};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/sslproto.pyx":938
 *         if self._app_transport:
 *             self._app_transport._force_close(exc)
 *         elif self._transport:             # <<<<<<<<<<<<<<
 *             self._transport._force_close(exc)
 * 
 */
  }
  __pyx_L3:;

  /* "uvloop/sslproto.pyx":941
 *             self._transport._force_close(exc)
 * 
 *         if isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)
 */
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_exc, __pyx_builtin_OSError); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 941, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uvloop/sslproto.pyx":942
 * 
 *         if isinstance(exc, OSError):
 *             if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)
 *         elif not isinstance(exc, aio_CancelledError):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_get_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 942, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(25, 942, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_1) {

      /* "uvloop/sslproto.pyx":943
 *         if isinstance(exc, OSError):
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)             # <<<<<<<<<<<<<<
 *         elif not isinstance(exc, aio_CancelledError):
 *             self._loop.call_exception_handler({
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(25, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_kp_u_r_s);
      __Pyx_GIVEREF(__pyx_kp_u_r_s);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_r_s)) __PYX_ERR(25, 943, __pyx_L1_error);
      __Pyx_INCREF((PyObject *)__pyx_v_self);
      __Pyx_GIVEREF((PyObject *)__pyx_v_self);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(25, 943, __pyx_L1_error);
      __Pyx_INCREF(__pyx_v_message);
      __Pyx_GIVEREF(__pyx_v_message);
      if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_message)) __PYX_ERR(25, 943, __pyx_L1_error);
      __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(25, 943, __pyx_L1_error)
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/sslproto.pyx":942
 * 
 *         if isinstance(exc, OSError):
 *             if self._loop.get_debug():             # <<<<<<<<<<<<<<
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)
 *         elif not isinstance(exc, aio_CancelledError):
 */
    }

    /* "uvloop/sslproto.pyx":941
 *             self._transport._force_close(exc)
 * 
 *         if isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)
 */
    goto __pyx_L4;
  }

  /* "uvloop/sslproto.pyx":944
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)
 *         elif not isinstance(exc, aio_CancelledError):             # <<<<<<<<<<<<<<
 *             self._loop.call_exception_handler({
 *                 'message': message,
 */
  __pyx_t_6 = __pyx_v_6uvloop_4loop_aio_CancelledError;
  __Pyx_INCREF(__pyx_t_6);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_exc, __pyx_t_6); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(25, 944, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = (!__pyx_t_1);
  if (__pyx_t_7) {

    /* "uvloop/sslproto.pyx":945
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)
 *         elif not isinstance(exc, aio_CancelledError):
 *             self._loop.call_exception_handler({             # <<<<<<<<<<<<<<
 *                 'message': message,
 *                 'exception': exc,
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_loop), __pyx_n_s_call_exception_handler); if (unlikely(!__pyx_t_4)) __PYX_ERR(25, 945, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uvloop/sslproto.pyx":946
 *         elif not isinstance(exc, aio_CancelledError):
 *             self._loop.call_exception_handler({
 *                 'message': message,             # <<<<<<<<<<<<<<
 *                 'exception': exc,
 *                 'transport': self._transport,
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_message, __pyx_v_message) < 0) __PYX_ERR(25, 946, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":947
 *             self._loop.call_exception_handler({
 *                 'message': message,
 *                 'exception': exc,             # <<<<<<<<<<<<<<
 *                 'transport': self._transport,
 *                 'protocol': self,
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_exception, __pyx_v_exc) < 0) __PYX_ERR(25, 946, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":948
 *                 'message': message,
 *                 'exception': exc,
 *                 'transport': self._transport,             # <<<<<<<<<<<<<<
 *                 'protocol': self,
 *             })
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_transport, __pyx_v_self->_transport) < 0) __PYX_ERR(25, 946, __pyx_L1_error)

    /* "uvloop/sslproto.pyx":949
 *                 'exception': exc,
 *                 'transport': self._transport,
 *                 'protocol': self,             # <<<<<<<<<<<<<<
 *             })
 */
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_u_protocol, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(25, 946, __pyx_L1_error)
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(25, 945, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uvloop/sslproto.pyx":944
 *             if self._loop.get_debug():
 *                 aio_logger.debug("%r: %s", self, message, exc_info=True)
 *         elif not isinstance(exc, aio_CancelledError):             # <<<<<<<<<<<<<<
 *             self._loop.call_exception_handler({
 *                 'message': message,
 */
  }
  __pyx_L4:;

  /* "uvloop/sslproto.pyx":935
 *             self._do_shutdown()
 * 
 *     cdef _fatal_error(self, exc, message='Fatal error on transport'):             # <<<<<<<<<<<<<<
 *         if self._app_transport:
 *             self._app_transport._force_close(exc)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol._fatal_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_22__reduce_cython__, "SSLProtocol.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_23__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_22__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_22__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11SSLProtocol_24__setstate_cython__, "SSLProtocol.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_11SSLProtocol_25__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_24__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_11SSLProtocol_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11SSLProtocol_24__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11SSLProtocol_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_SSLProtocol *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.SSLProtocol.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":16
 *         bint            closed
 * 
 *     cdef close(self):             # <<<<<<<<<<<<<<
 *         if self.closed:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_15_UDPSendContext_close(struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/handles/udp.pyx":17
 * 
 *     cdef close(self):
 *         if self.closed:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_self->closed) {

    /* "uvloop/handles/udp.pyx":18
 *     cdef close(self):
 *         if self.closed:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self.closed = 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":17
 * 
 *     cdef close(self):
 *         if self.closed:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/udp.pyx":20
 *             return
 * 
 *         self.closed = 1             # <<<<<<<<<<<<<<
 *         PyBuffer_Release(&self.py_buf)  # void
 *         self.req.data = NULL
 */
  __pyx_v_self->closed = 1;

  /* "uvloop/handles/udp.pyx":21
 * 
 *         self.closed = 1
 *         PyBuffer_Release(&self.py_buf)  # void             # <<<<<<<<<<<<<<
 *         self.req.data = NULL
 *         self.uv_buf.base = NULL
 */
  PyBuffer_Release((&__pyx_v_self->py_buf));

  /* "uvloop/handles/udp.pyx":22
 *         self.closed = 1
 *         PyBuffer_Release(&self.py_buf)  # void
 *         self.req.data = NULL             # <<<<<<<<<<<<<<
 *         self.uv_buf.base = NULL
 *         Py_DECREF(self)
 */
  __pyx_v_self->req.data = NULL;

  /* "uvloop/handles/udp.pyx":23
 *         PyBuffer_Release(&self.py_buf)  # void
 *         self.req.data = NULL
 *         self.uv_buf.base = NULL             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 *         self.udp = None
 */
  __pyx_v_self->uv_buf.base = NULL;

  /* "uvloop/handles/udp.pyx":24
 *         self.req.data = NULL
 *         self.uv_buf.base = NULL
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 *         self.udp = None
 * 
 */
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "uvloop/handles/udp.pyx":25
 *         self.uv_buf.base = NULL
 *         Py_DECREF(self)
 *         self.udp = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->udp);
  __Pyx_DECREF((PyObject *)__pyx_v_self->udp);
  __pyx_v_self->udp = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)Py_None);

  /* "uvloop/handles/udp.pyx":16
 *         bint            closed
 * 
 *     cdef close(self):             # <<<<<<<<<<<<<<
 *         if self.closed:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":28
 * 
 *     @staticmethod
 *     cdef _UDPSendContext new(UDPTransport udp, object data):             # <<<<<<<<<<<<<<
 *         cdef _UDPSendContext ctx
 *         ctx = _UDPSendContext.__new__(_UDPSendContext)
 */

static struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_f_6uvloop_4loop_15_UDPSendContext_new(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_udp, PyObject *__pyx_v_data) {
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_ctx = 0;
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("new", 1);

  /* "uvloop/handles/udp.pyx":30
 *     cdef _UDPSendContext new(UDPTransport udp, object data):
 *         cdef _UDPSendContext ctx
 *         ctx = _UDPSendContext.__new__(_UDPSendContext)             # <<<<<<<<<<<<<<
 *         ctx.udp = None
 *         ctx.closed = 1
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_6uvloop_4loop__UDPSendContext(((PyTypeObject *)__pyx_ptype_6uvloop_4loop__UDPSendContext), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 30, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_ctx = ((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":31
 *         cdef _UDPSendContext ctx
 *         ctx = _UDPSendContext.__new__(_UDPSendContext)
 *         ctx.udp = None             # <<<<<<<<<<<<<<
 *         ctx.closed = 1
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_ctx->udp);
  __Pyx_DECREF((PyObject *)__pyx_v_ctx->udp);
  __pyx_v_ctx->udp = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)Py_None);

  /* "uvloop/handles/udp.pyx":32
 *         ctx = _UDPSendContext.__new__(_UDPSendContext)
 *         ctx.udp = None
 *         ctx.closed = 1             # <<<<<<<<<<<<<<
 * 
 *         ctx.req.data = <void*> ctx
 */
  __pyx_v_ctx->closed = 1;

  /* "uvloop/handles/udp.pyx":34
 *         ctx.closed = 1
 * 
 *         ctx.req.data = <void*> ctx             # <<<<<<<<<<<<<<
 *         Py_INCREF(ctx)
 * 
 */
  __pyx_v_ctx->req.data = ((void *)__pyx_v_ctx);

  /* "uvloop/handles/udp.pyx":35
 * 
 *         ctx.req.data = <void*> ctx
 *         Py_INCREF(ctx)             # <<<<<<<<<<<<<<
 * 
 *         PyObject_GetBuffer(data, &ctx.py_buf, PyBUF_SIMPLE)
 */
  Py_INCREF(((PyObject *)__pyx_v_ctx));

  /* "uvloop/handles/udp.pyx":37
 *         Py_INCREF(ctx)
 * 
 *         PyObject_GetBuffer(data, &ctx.py_buf, PyBUF_SIMPLE)             # <<<<<<<<<<<<<<
 *         ctx.uv_buf.base = <char*>ctx.py_buf.buf
 *         ctx.uv_buf.len = ctx.py_buf.len
 */
  __pyx_t_2 = PyObject_GetBuffer(__pyx_v_data, (&__pyx_v_ctx->py_buf), PyBUF_SIMPLE); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(26, 37, __pyx_L1_error)

  /* "uvloop/handles/udp.pyx":38
 * 
 *         PyObject_GetBuffer(data, &ctx.py_buf, PyBUF_SIMPLE)
 *         ctx.uv_buf.base = <char*>ctx.py_buf.buf             # <<<<<<<<<<<<<<
 *         ctx.uv_buf.len = ctx.py_buf.len
 *         ctx.udp = udp
 */
  __pyx_v_ctx->uv_buf.base = ((char *)__pyx_v_ctx->py_buf.buf);

  /* "uvloop/handles/udp.pyx":39
 *         PyObject_GetBuffer(data, &ctx.py_buf, PyBUF_SIMPLE)
 *         ctx.uv_buf.base = <char*>ctx.py_buf.buf
 *         ctx.uv_buf.len = ctx.py_buf.len             # <<<<<<<<<<<<<<
 *         ctx.udp = udp
 * 
 */
  __pyx_t_3 = __pyx_v_ctx->py_buf.len;
  __pyx_v_ctx->uv_buf.len = __pyx_t_3;

  /* "uvloop/handles/udp.pyx":40
 *         ctx.uv_buf.base = <char*>ctx.py_buf.buf
 *         ctx.uv_buf.len = ctx.py_buf.len
 *         ctx.udp = udp             # <<<<<<<<<<<<<<
 * 
 *         ctx.closed = 0
 */
  __Pyx_INCREF((PyObject *)__pyx_v_udp);
  __Pyx_GIVEREF((PyObject *)__pyx_v_udp);
  __Pyx_GOTREF((PyObject *)__pyx_v_ctx->udp);
  __Pyx_DECREF((PyObject *)__pyx_v_ctx->udp);
  __pyx_v_ctx->udp = __pyx_v_udp;

  /* "uvloop/handles/udp.pyx":42
 *         ctx.udp = udp
 * 
 *         ctx.closed = 0             # <<<<<<<<<<<<<<
 *         return ctx
 * 
 */
  __pyx_v_ctx->closed = 0;

  /* "uvloop/handles/udp.pyx":43
 * 
 *         ctx.closed = 0
 *         return ctx             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ctx);
  __pyx_r = __pyx_v_ctx;
  goto __pyx_L0;

  /* "uvloop/handles/udp.pyx":28
 * 
 *     @staticmethod
 *     cdef _UDPSendContext new(UDPTransport udp, object data):             # <<<<<<<<<<<<<<
 *         cdef _UDPSendContext ctx
 *         ctx = _UDPSendContext.__new__(_UDPSendContext)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop._UDPSendContext.new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":45
 *         return ctx
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if not self.closed:
 */

/* Python wrapper */
static void __pyx_pw_6uvloop_4loop_15_UDPSendContext_1__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6uvloop_4loop_15_UDPSendContext_1__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_6uvloop_4loop_15_UDPSendContext___dealloc__(((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6uvloop_4loop_15_UDPSendContext___dealloc__(struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uvloop/handles/udp.pyx":46
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if not self.closed:
 *                 raise RuntimeError(
 */
  __pyx_t_1 = (UVLOOP_DEBUG != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":47
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:
 *             if not self.closed:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     'open _UDPSendContext is being deallocated')
 */
    __pyx_t_1 = (!__pyx_v_self->closed);
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/udp.pyx":48
 *         if UVLOOP_DEBUG:
 *             if not self.closed:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     'open _UDPSendContext is being deallocated')
 *         self.udp = None
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__186, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 48, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(26, 48, __pyx_L1_error)

      /* "uvloop/handles/udp.pyx":47
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:
 *             if not self.closed:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     'open _UDPSendContext is being deallocated')
 */
    }

    /* "uvloop/handles/udp.pyx":46
 * 
 *     def __dealloc__(self):
 *         if UVLOOP_DEBUG:             # <<<<<<<<<<<<<<
 *             if not self.closed:
 *                 raise RuntimeError(
 */
  }

  /* "uvloop/handles/udp.pyx":50
 *                 raise RuntimeError(
 *                     'open _UDPSendContext is being deallocated')
 *         self.udp = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->udp);
  __Pyx_DECREF((PyObject *)__pyx_v_self->udp);
  __pyx_v_self->udp = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)Py_None);

  /* "uvloop/handles/udp.pyx":45
 *         return ctx
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if not self.closed:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("uvloop.loop._UDPSendContext.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15_UDPSendContext_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15_UDPSendContext_2__reduce_cython__, "_UDPSendContext.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15_UDPSendContext_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15_UDPSendContext_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15_UDPSendContext_2__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15_UDPSendContext_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_15_UDPSendContext_2__reduce_cython__(((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15_UDPSendContext_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_py_buf_self_req_cannot_be_c, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._UDPSendContext.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_15_UDPSendContext_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_15_UDPSendContext_4__setstate_cython__, "_UDPSendContext.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_15_UDPSendContext_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15_UDPSendContext_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15_UDPSendContext_4__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_15_UDPSendContext_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop._UDPSendContext.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_15_UDPSendContext_4__setstate_cython__(((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_15_UDPSendContext_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self_py_buf_self_req_cannot_be_c, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._UDPSendContext.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":55
 * @cython.no_gc_clear
 * cdef class UDPTransport(UVBaseTransport):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._family = uv.AF_UNSPEC
 *         self.__receiving = 0
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_12UDPTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_12UDPTransport_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6uvloop_4loop_12UDPTransport___cinit__(((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_12UDPTransport___cinit__(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/handles/udp.pyx":56
 * cdef class UDPTransport(UVBaseTransport):
 *     def __cinit__(self):
 *         self._family = uv.AF_UNSPEC             # <<<<<<<<<<<<<<
 *         self.__receiving = 0
 *         self._address = None
 */
  __pyx_v_self->_family = AF_UNSPEC;

  /* "uvloop/handles/udp.pyx":57
 *     def __cinit__(self):
 *         self._family = uv.AF_UNSPEC
 *         self.__receiving = 0             # <<<<<<<<<<<<<<
 *         self._address = None
 *         self.context = Context_CopyCurrent()
 */
  __pyx_v_self->_UDPTransport__receiving = 0;

  /* "uvloop/handles/udp.pyx":58
 *         self._family = uv.AF_UNSPEC
 *         self.__receiving = 0
 *         self._address = None             # <<<<<<<<<<<<<<
 *         self.context = Context_CopyCurrent()
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_address);
  __Pyx_DECREF(__pyx_v_self->_address);
  __pyx_v_self->_address = Py_None;

  /* "uvloop/handles/udp.pyx":59
 *         self.__receiving = 0
 *         self._address = None
 *         self.context = Context_CopyCurrent()             # <<<<<<<<<<<<<<
 * 
 *     cdef _init(self, Loop loop, unsigned int family):
 */
  __pyx_t_1 = Context_CopyCurrent(); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context);
  __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":55
 * @cython.no_gc_clear
 * cdef class UDPTransport(UVBaseTransport):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._family = uv.AF_UNSPEC
 *         self.__receiving = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":61
 *         self.context = Context_CopyCurrent()
 * 
 *     cdef _init(self, Loop loop, unsigned int family):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__init(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop, unsigned int __pyx_v_family) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 1);

  /* "uvloop/handles/udp.pyx":64
 *         cdef int err
 * 
 *         self._start_init(loop)             # <<<<<<<<<<<<<<
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_udp_t))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._start_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_loop); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":66
 *         self._start_init(loop)
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_udp_t))             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             self._abort_init()
 */
  __pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle = ((uv_handle_t *)PyMem_RawMalloc((sizeof(uv_udp_t))));

  /* "uvloop/handles/udp.pyx":67
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_udp_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/udp.pyx":68
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_udp_t))
 *         if self._handle is NULL:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/udp.pyx":69
 *         if self._handle is NULL:
 *             self._abort_init()
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_udp_init_ex(loop.uvloop,
 */
    PyErr_NoMemory(); __PYX_ERR(26, 69, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":67
 * 
 *         self._handle = <uv.uv_handle_t*>PyMem_RawMalloc(sizeof(uv.uv_udp_t))
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise MemoryError()
 */
  }

  /* "uvloop/handles/udp.pyx":71
 *             raise MemoryError()
 * 
 *         err = uv.uv_udp_init_ex(loop.uvloop,             # <<<<<<<<<<<<<<
 *                                 <uv.uv_udp_t*>self._handle,
 *                                 family)
 */
  __pyx_v_err = uv_udp_init_ex(__pyx_v_loop->uvloop, ((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_family);

  /* "uvloop/handles/udp.pyx":74
 *                                 <uv.uv_udp_t*>self._handle,
 *                                 family)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/udp.pyx":75
 *                                 family)
 *         if err < 0:
 *             self._abort_init()             # <<<<<<<<<<<<<<
 *             raise convert_error(err)
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._abort_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/udp.pyx":76
 *         if err < 0:
 *             self._abort_init()
 *             raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 *         if family in (uv.AF_INET, uv.AF_INET6):
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(26, 76, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":74
 *                                 <uv.uv_udp_t*>self._handle,
 *                                 family)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             self._abort_init()
 *             raise convert_error(err)
 */
  }

  /* "uvloop/handles/udp.pyx":78
 *             raise convert_error(err)
 * 
 *         if family in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *             self._family = family
 * 
 */
  __pyx_t_3 = __pyx_v_family;
  __pyx_t_4 = (__pyx_t_3 == AF_INET);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_t_3 == AF_INET6);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  __pyx_t_4 = __pyx_t_2;
  if (__pyx_t_4) {

    /* "uvloop/handles/udp.pyx":79
 * 
 *         if family in (uv.AF_INET, uv.AF_INET6):
 *             self._family = family             # <<<<<<<<<<<<<<
 * 
 *         self._finish_init()
 */
    __pyx_v_self->_family = __pyx_v_family;

    /* "uvloop/handles/udp.pyx":78
 *             raise convert_error(err)
 * 
 *         if family in (uv.AF_INET, uv.AF_INET6):             # <<<<<<<<<<<<<<
 *             self._family = family
 * 
 */
  }

  /* "uvloop/handles/udp.pyx":81
 *             self._family = family
 * 
 *         self._finish_init()             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_address(self, system.addrinfo *addr):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._finish_init(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":61
 *         self.context = Context_CopyCurrent()
 * 
 *     cdef _init(self, Loop loop, unsigned int family):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":83
 *         self._finish_init()
 * 
 *     cdef _set_address(self, system.addrinfo *addr):             # <<<<<<<<<<<<<<
 *         self._address = __convert_sockaddr_to_pyaddr(addr.ai_addr)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__set_address(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct addrinfo *__pyx_v_addr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_address", 1);

  /* "uvloop/handles/udp.pyx":84
 * 
 *     cdef _set_address(self, system.addrinfo *addr):
 *         self._address = __convert_sockaddr_to_pyaddr(addr.ai_addr)             # <<<<<<<<<<<<<<
 * 
 *     cdef _connect(self, system.sockaddr* addr, size_t addr_len):
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(__pyx_v_addr->ai_addr); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_address);
  __Pyx_DECREF(__pyx_v_self->_address);
  __pyx_v_self->_address = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":83
 *         self._finish_init()
 * 
 *     cdef _set_address(self, system.addrinfo *addr):             # <<<<<<<<<<<<<<
 *         self._address = __convert_sockaddr_to_pyaddr(addr.ai_addr)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._set_address", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":86
 *         self._address = __convert_sockaddr_to_pyaddr(addr.ai_addr)
 * 
 *     cdef _connect(self, system.sockaddr* addr, size_t addr_len):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         err = uv.uv_udp_connect(<uv.uv_udp_t*>self._handle, addr)
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__connect(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr, CYTHON_UNUSED size_t __pyx_v_addr_len) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_connect", 1);

  /* "uvloop/handles/udp.pyx":88
 *     cdef _connect(self, system.sockaddr* addr, size_t addr_len):
 *         cdef int err
 *         err = uv.uv_udp_connect(<uv.uv_udp_t*>self._handle, addr)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  __pyx_v_err = uv_udp_connect(((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_addr);

  /* "uvloop/handles/udp.pyx":89
 *         cdef int err
 *         err = uv.uv_udp_connect(<uv.uv_udp_t*>self._handle, addr)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/handles/udp.pyx":90
 *         err = uv.uv_udp_connect(<uv.uv_udp_t*>self._handle, addr)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             raise exc
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/udp.pyx":91
 *         if err < 0:
 *             exc = convert_error(err)
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 *     cdef open(self, int family, int sockfd):
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(26, 91, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":89
 *         cdef int err
 *         err = uv.uv_udp_connect(<uv.uv_udp_t*>self._handle, addr)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  }

  /* "uvloop/handles/udp.pyx":86
 *         self._address = __convert_sockaddr_to_pyaddr(addr.ai_addr)
 * 
 *     cdef _connect(self, system.sockaddr* addr, size_t addr_len):             # <<<<<<<<<<<<<<
 *         cdef int err
 *         err = uv.uv_udp_connect(<uv.uv_udp_t*>self._handle, addr)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":93
 *             raise exc
 * 
 *     cdef open(self, int family, int sockfd):             # <<<<<<<<<<<<<<
 *         if family in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             self._family = family
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport_open(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, int __pyx_v_family, int __pyx_v_sockfd) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("open", 1);

  /* "uvloop/handles/udp.pyx":94
 * 
 *     cdef open(self, int family, int sockfd):
 *         if family in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):             # <<<<<<<<<<<<<<
 *             self._family = family
 *         else:
 */
  __pyx_t_1 = __pyx_v_family;
  __pyx_t_3 = (__pyx_t_1 == AF_INET);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 == AF_INET6);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 == AF_UNIX);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __pyx_t_3 = __pyx_t_2;
  if (likely(__pyx_t_3)) {

    /* "uvloop/handles/udp.pyx":95
 *     cdef open(self, int family, int sockfd):
 *         if family in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             self._family = family             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError(
 */
    __pyx_v_self->_family = __pyx_v_family;

    /* "uvloop/handles/udp.pyx":94
 * 
 *     cdef open(self, int family, int sockfd):
 *         if family in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):             # <<<<<<<<<<<<<<
 *             self._family = family
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/udp.pyx":97
 *             self._family = family
 *         else:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'cannot open a UDP handle, invalid family {}'.format(family))
 * 
 */
  /*else*/ {

    /* "uvloop/handles/udp.pyx":98
 *         else:
 *             raise ValueError(
 *                 'cannot open a UDP handle, invalid family {}'.format(family))             # <<<<<<<<<<<<<<
 * 
 *         cdef int err
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_cannot_open_a_UDP_handle_invalid, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(26, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_family); if (unlikely(!__pyx_t_6)) __PYX_ERR(26, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }

    /* "uvloop/handles/udp.pyx":97
 *             self._family = family
 *         else:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'cannot open a UDP handle, invalid family {}'.format(family))
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(26, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(26, 97, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "uvloop/handles/udp.pyx":101
 * 
 *         cdef int err
 *         err = uv.uv_udp_open(<uv.uv_udp_t*>self._handle,             # <<<<<<<<<<<<<<
 *                              <uv.uv_os_sock_t>sockfd)
 * 
 */
  __pyx_v_err = uv_udp_open(((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), ((uv_os_sock_t)__pyx_v_sockfd));

  /* "uvloop/handles/udp.pyx":104
 *                              <uv.uv_os_sock_t>sockfd)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  __pyx_t_3 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/handles/udp.pyx":105
 * 
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             raise exc
 * 
 */
    __pyx_t_5 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_5)) __PYX_ERR(26, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_exc = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "uvloop/handles/udp.pyx":106
 *         if err < 0:
 *             exc = convert_error(err)
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 *     cdef _bind(self, system.sockaddr* addr):
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(26, 106, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":104
 *                              <uv.uv_os_sock_t>sockfd)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  }

  /* "uvloop/handles/udp.pyx":93
 *             raise exc
 * 
 *     cdef open(self, int family, int sockfd):             # <<<<<<<<<<<<<<
 *         if family in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             self._family = family
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":108
 *             raise exc
 * 
 *     cdef _bind(self, system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__bind(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, struct sockaddr *__pyx_v_addr) {
  int __pyx_v_err;
  int __pyx_v_flags;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_bind", 1);

  /* "uvloop/handles/udp.pyx":111
 *         cdef:
 *             int err
 *             int flags = 0             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_alive()
 */
  __pyx_v_flags = 0;

  /* "uvloop/handles/udp.pyx":113
 *             int flags = 0
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_udp_bind(<uv.uv_udp_t*>self._handle, addr, flags)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":115
 *         self._ensure_alive()
 * 
 *         err = uv.uv_udp_bind(<uv.uv_udp_t*>self._handle, addr, flags)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  __pyx_v_err = uv_udp_bind(((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_addr, __pyx_v_flags);

  /* "uvloop/handles/udp.pyx":116
 * 
 *         err = uv.uv_udp_bind(<uv.uv_udp_t*>self._handle, addr, flags)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/udp.pyx":117
 *         err = uv.uv_udp_bind(<uv.uv_udp_t*>self._handle, addr, flags)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             raise exc
 * 
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/udp.pyx":118
 *         if err < 0:
 *             exc = convert_error(err)
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 *     cdef _set_broadcast(self, bint on):
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(26, 118, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":116
 * 
 *         err = uv.uv_udp_bind(<uv.uv_udp_t*>self._handle, addr, flags)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  }

  /* "uvloop/handles/udp.pyx":108
 *             raise exc
 * 
 *     cdef _bind(self, system.sockaddr* addr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int err
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._bind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":120
 *             raise exc
 * 
 *     cdef _set_broadcast(self, bint on):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__set_broadcast(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, int __pyx_v_on) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_broadcast", 1);

  /* "uvloop/handles/udp.pyx":123
 *         cdef int err
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_udp_set_broadcast(<uv.uv_udp_t*>self._handle, on)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":125
 *         self._ensure_alive()
 * 
 *         err = uv.uv_udp_set_broadcast(<uv.uv_udp_t*>self._handle, on)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  __pyx_v_err = uv_udp_set_broadcast(((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_v_on);

  /* "uvloop/handles/udp.pyx":126
 * 
 *         err = uv.uv_udp_set_broadcast(<uv.uv_udp_t*>self._handle, on)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_2)) {

    /* "uvloop/handles/udp.pyx":127
 *         err = uv.uv_udp_set_broadcast(<uv.uv_udp_t*>self._handle, on)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             raise exc
 * 
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/udp.pyx":128
 *         if err < 0:
 *             exc = convert_error(err)
 *             raise exc             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t _get_write_buffer_size(self):
 */
    __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
    __PYX_ERR(26, 128, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":126
 * 
 *         err = uv.uv_udp_set_broadcast(<uv.uv_udp_t*>self._handle, on)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             raise exc
 */
  }

  /* "uvloop/handles/udp.pyx":120
 *             raise exc
 * 
 *     cdef _set_broadcast(self, bint on):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._set_broadcast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":130
 *             raise exc
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             return 0
 */

static size_t __pyx_f_6uvloop_4loop_12UDPTransport__get_write_buffer_size(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  size_t __pyx_r;
  int __pyx_t_1;

  /* "uvloop/handles/udp.pyx":131
 * 
 *     cdef size_t _get_write_buffer_size(self):
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return (<uv.uv_udp_t*>self._handle).send_queue_size
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":132
 *     cdef size_t _get_write_buffer_size(self):
 *         if self._handle is NULL:
 *             return 0             # <<<<<<<<<<<<<<
 *         return (<uv.uv_udp_t*>self._handle).send_queue_size
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":131
 * 
 *     cdef size_t _get_write_buffer_size(self):
 *         if self._handle is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return (<uv.uv_udp_t*>self._handle).send_queue_size
 */
  }

  /* "uvloop/handles/udp.pyx":133
 *         if self._handle is NULL:
 *             return 0
 *         return (<uv.uv_udp_t*>self._handle).send_queue_size             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_reading(self):
 */
  __pyx_r = ((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle)->send_queue_size;
  goto __pyx_L0;

  /* "uvloop/handles/udp.pyx":130
 *             raise exc
 * 
 *     cdef size_t _get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         if self._handle is NULL:
 *             return 0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":135
 *         return (<uv.uv_udp_t*>self._handle).send_queue_size
 * 
 *     cdef bint _is_reading(self):             # <<<<<<<<<<<<<<
 *         return self.__receiving
 * 
 */

static int __pyx_f_6uvloop_4loop_12UDPTransport__is_reading(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  int __pyx_r;

  /* "uvloop/handles/udp.pyx":136
 * 
 *     cdef bint _is_reading(self):
 *         return self.__receiving             # <<<<<<<<<<<<<<
 * 
 *     cdef _start_reading(self):
 */
  __pyx_r = __pyx_v_self->_UDPTransport__receiving;
  goto __pyx_L0;

  /* "uvloop/handles/udp.pyx":135
 *         return (<uv.uv_udp_t*>self._handle).send_queue_size
 * 
 *     cdef bint _is_reading(self):             # <<<<<<<<<<<<<<
 *         return self.__receiving
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":138
 *         return self.__receiving
 * 
 *     cdef _start_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__start_reading(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_start_reading", 1);

  /* "uvloop/handles/udp.pyx":141
 *         cdef int err
 * 
 *         if self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_self->_UDPTransport__receiving) {

    /* "uvloop/handles/udp.pyx":142
 * 
 *         if self.__receiving:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_alive()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":141
 *         cdef int err
 * 
 *         if self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/udp.pyx":144
 *             return
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_udp_recv_start(<uv.uv_udp_t*>self._handle,
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":146
 *         self._ensure_alive()
 * 
 *         err = uv.uv_udp_recv_start(<uv.uv_udp_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                    __loop_alloc_buffer,
 *                                    __uv_udp_on_receive)
 */
  __pyx_v_err = uv_udp_recv_start(((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), __pyx_f_6uvloop_4loop___loop_alloc_buffer, __pyx_f_6uvloop_4loop___uv_udp_on_receive);

  /* "uvloop/handles/udp.pyx":150
 *                                    __uv_udp_on_receive)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_2 = (__pyx_v_err < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":151
 * 
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_exc = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "uvloop/handles/udp.pyx":152
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 *         else:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/handles/udp.pyx":153
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 *         else:
 *             # UDPTransport must live until the read callback is called
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":150
 *                                    __uv_udp_on_receive)
 * 
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/udp.pyx":156
 *         else:
 *             # UDPTransport must live until the read callback is called
 *             self.__receiving_started()             # <<<<<<<<<<<<<<
 * 
 *     cdef _stop_reading(self):
 */
  /*else*/ {
    __pyx_t_1 = __pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_started(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "uvloop/handles/udp.pyx":138
 *         return self.__receiving
 * 
 *     cdef _start_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._start_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":158
 *             self.__receiving_started()
 * 
 *     cdef _stop_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__stop_reading(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_stop_reading", 1);

  /* "uvloop/handles/udp.pyx":161
 *         cdef int err
 * 
 *         if not self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (!__pyx_v_self->_UDPTransport__receiving);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":162
 * 
 *         if not self.__receiving:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_alive()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":161
 *         cdef int err
 * 
 *         if not self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/handles/udp.pyx":164
 *             return
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         err = uv.uv_udp_recv_stop(<uv.uv_udp_t*>self._handle)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/udp.pyx":166
 *         self._ensure_alive()
 * 
 *         err = uv.uv_udp_recv_stop(<uv.uv_udp_t*>self._handle)             # <<<<<<<<<<<<<<
 *         if err < 0:
 *             exc = convert_error(err)
 */
  __pyx_v_err = uv_udp_recv_stop(((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle));

  /* "uvloop/handles/udp.pyx":167
 * 
 *         err = uv.uv_udp_recv_stop(<uv.uv_udp_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":168
 *         err = uv.uv_udp_recv_stop(<uv.uv_udp_t*>self._handle)
 *         if err < 0:
 *             exc = convert_error(err)             # <<<<<<<<<<<<<<
 *             self._fatal_error(exc, True)
 *             return
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/udp.pyx":169
 *         if err < 0:
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             return
 *         else:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/udp.pyx":170
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 *             return             # <<<<<<<<<<<<<<
 *         else:
 *             self.__receiving_stopped()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":167
 * 
 *         err = uv.uv_udp_recv_stop(<uv.uv_udp_t*>self._handle)
 *         if err < 0:             # <<<<<<<<<<<<<<
 *             exc = convert_error(err)
 *             self._fatal_error(exc, True)
 */
  }

  /* "uvloop/handles/udp.pyx":172
 *             return
 *         else:
 *             self.__receiving_stopped()             # <<<<<<<<<<<<<<
 * 
 *     cdef inline __receiving_started(self):
 */
  /*else*/ {
    __pyx_t_2 = __pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_stopped(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "uvloop/handles/udp.pyx":158
 *             self.__receiving_started()
 * 
 *     cdef _stop_reading(self):             # <<<<<<<<<<<<<<
 *         cdef int err
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._stop_reading", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":174
 *             self.__receiving_stopped()
 * 
 *     cdef inline __receiving_started(self):             # <<<<<<<<<<<<<<
 *         if self.__receiving:
 *             return
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_started(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_UDPTransport__receiving_started", 1);

  /* "uvloop/handles/udp.pyx":175
 * 
 *     cdef inline __receiving_started(self):
 *         if self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 *         self.__receiving = 1
 */
  if (__pyx_v_self->_UDPTransport__receiving) {

    /* "uvloop/handles/udp.pyx":176
 *     cdef inline __receiving_started(self):
 *         if self.__receiving:
 *             return             # <<<<<<<<<<<<<<
 *         self.__receiving = 1
 *         Py_INCREF(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":175
 * 
 *     cdef inline __receiving_started(self):
 *         if self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 *         self.__receiving = 1
 */
  }

  /* "uvloop/handles/udp.pyx":177
 *         if self.__receiving:
 *             return
 *         self.__receiving = 1             # <<<<<<<<<<<<<<
 *         Py_INCREF(self)
 * 
 */
  __pyx_v_self->_UDPTransport__receiving = 1;

  /* "uvloop/handles/udp.pyx":178
 *             return
 *         self.__receiving = 1
 *         Py_INCREF(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef inline __receiving_stopped(self):
 */
  Py_INCREF(((PyObject *)__pyx_v_self));

  /* "uvloop/handles/udp.pyx":174
 *             self.__receiving_stopped()
 * 
 *     cdef inline __receiving_started(self):             # <<<<<<<<<<<<<<
 *         if self.__receiving:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":180
 *         Py_INCREF(self)
 * 
 *     cdef inline __receiving_stopped(self):             # <<<<<<<<<<<<<<
 *         if not self.__receiving:
 *             return
 */

static CYTHON_INLINE PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_stopped(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_UDPTransport__receiving_stopped", 1);

  /* "uvloop/handles/udp.pyx":181
 * 
 *     cdef inline __receiving_stopped(self):
 *         if not self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 *         self.__receiving = 0
 */
  __pyx_t_1 = (!__pyx_v_self->_UDPTransport__receiving);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":182
 *     cdef inline __receiving_stopped(self):
 *         if not self.__receiving:
 *             return             # <<<<<<<<<<<<<<
 *         self.__receiving = 0
 *         Py_DECREF(self)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":181
 * 
 *     cdef inline __receiving_stopped(self):
 *         if not self.__receiving:             # <<<<<<<<<<<<<<
 *             return
 *         self.__receiving = 0
 */
  }

  /* "uvloop/handles/udp.pyx":183
 *         if not self.__receiving:
 *             return
 *         self.__receiving = 0             # <<<<<<<<<<<<<<
 *         Py_DECREF(self)
 * 
 */
  __pyx_v_self->_UDPTransport__receiving = 0;

  /* "uvloop/handles/udp.pyx":184
 *             return
 *         self.__receiving = 0
 *         Py_DECREF(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _new_socket(self):
 */
  Py_DECREF(((PyObject *)__pyx_v_self));

  /* "uvloop/handles/udp.pyx":180
 *         Py_INCREF(self)
 * 
 *     cdef inline __receiving_stopped(self):             # <<<<<<<<<<<<<<
 *         if not self.__receiving:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":186
 *         Py_DECREF(self)
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             raise RuntimeError(
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__new_socket(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  PyObject *__pyx_v_fileno = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_new_socket", 1);

  /* "uvloop/handles/udp.pyx":187
 * 
 *     cdef _new_socket(self):
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'UDPTransport.family is undefined; '
 */
  __pyx_t_1 = __pyx_v_self->_family;
  __pyx_t_3 = (__pyx_t_1 != AF_INET);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 != AF_INET6);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_t_1 != AF_UNIX);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __pyx_t_3 = __pyx_t_2;
  if (unlikely(__pyx_t_3)) {

    /* "uvloop/handles/udp.pyx":188
 *     cdef _new_socket(self):
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'UDPTransport.family is undefined; '
 *                 'cannot create python socket')
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__187, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(26, 188, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":187
 * 
 *     cdef _new_socket(self):
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 'UDPTransport.family is undefined; '
 */
  }

  /* "uvloop/handles/udp.pyx":192
 *                 'cannot create python socket')
 * 
 *         fileno = self._fileno()             # <<<<<<<<<<<<<<
 *         return PseudoSocket(self._family, uv.SOCK_DGRAM, 0, fileno)
 * 
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._fileno(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_fileno = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "uvloop/handles/udp.pyx":193
 * 
 *         fileno = self._fileno()
 *         return PseudoSocket(self._family, uv.SOCK_DGRAM, 0, fileno)             # <<<<<<<<<<<<<<
 * 
 *     cdef _send(self, object data, object addr):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->_family); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(SOCK_DGRAM); if (unlikely(!__pyx_t_5)) __PYX_ERR(26, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(26, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4)) __PYX_ERR(26, 193, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5)) __PYX_ERR(26, 193, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_int_0)) __PYX_ERR(26, 193, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_fileno);
  __Pyx_GIVEREF(__pyx_v_fileno);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_fileno)) __PYX_ERR(26, 193, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(26, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "uvloop/handles/udp.pyx":186
 *         Py_DECREF(self)
 * 
 *     cdef _new_socket(self):             # <<<<<<<<<<<<<<
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             raise RuntimeError(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._new_socket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fileno);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":195
 *         return PseudoSocket(self._family, uv.SOCK_DGRAM, 0, fileno)
 * 
 *     cdef _send(self, object data, object addr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _UDPSendContext ctx
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__send(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_addr) {
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_ctx = 0;
  struct sockaddr_storage __pyx_v_saddr_st;
  struct sockaddr *__pyx_v_saddr;
  Py_buffer __pyx_v_try_pybuf;
  uv_buf_t __pyx_v_try_uvbuf;
  int __pyx_v_err;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  size_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  unsigned int __pyx_t_14;
  struct __pyx_opt_args_6uvloop_4loop_12UDPTransport__on_sent __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_send", 1);

  /* "uvloop/handles/udp.pyx":203
 *             uv.uv_buf_t     try_uvbuf
 * 
 *         self._ensure_alive()             # <<<<<<<<<<<<<<
 * 
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._ensure_alive(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/handles/udp.pyx":205
 *         self._ensure_alive()
 * 
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):             # <<<<<<<<<<<<<<
 *             raise RuntimeError('UDPTransport.family is undefined; cannot send')
 * 
 */
  __pyx_t_2 = __pyx_v_self->_family;
  __pyx_t_4 = (__pyx_t_2 != AF_INET);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_t_2 != AF_INET6);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_t_2 != AF_UNIX);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  __pyx_t_4 = __pyx_t_3;
  if (unlikely(__pyx_t_4)) {

    /* "uvloop/handles/udp.pyx":206
 * 
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             raise RuntimeError('UDPTransport.family is undefined; cannot send')             # <<<<<<<<<<<<<<
 * 
 *         if addr is None:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__188, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(26, 206, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":205
 *         self._ensure_alive()
 * 
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):             # <<<<<<<<<<<<<<
 *             raise RuntimeError('UDPTransport.family is undefined; cannot send')
 * 
 */
  }

  /* "uvloop/handles/udp.pyx":208
 *             raise RuntimeError('UDPTransport.family is undefined; cannot send')
 * 
 *         if addr is None:             # <<<<<<<<<<<<<<
 *             saddr = NULL
 *         else:
 */
  __pyx_t_4 = (__pyx_v_addr == Py_None);
  if (__pyx_t_4) {

    /* "uvloop/handles/udp.pyx":209
 * 
 *         if addr is None:
 *             saddr = NULL             # <<<<<<<<<<<<<<
 *         else:
 *             try:
 */
    __pyx_v_saddr = NULL;

    /* "uvloop/handles/udp.pyx":208
 *             raise RuntimeError('UDPTransport.family is undefined; cannot send')
 * 
 *         if addr is None:             # <<<<<<<<<<<<<<
 *             saddr = NULL
 *         else:
 */
    goto __pyx_L7;
  }

  /* "uvloop/handles/udp.pyx":211
 *             saddr = NULL
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 __convert_pyaddr_to_sockaddr(self._family, addr,
 *                                              <system.sockaddr*>&saddr_st)
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "uvloop/handles/udp.pyx":212
 *         else:
 *             try:
 *                 __convert_pyaddr_to_sockaddr(self._family, addr,             # <<<<<<<<<<<<<<
 *                                              <system.sockaddr*>&saddr_st)
 *             except (ValueError, TypeError):
 */
        __pyx_t_1 = __pyx_f_6uvloop_4loop___convert_pyaddr_to_sockaddr(__pyx_v_self->_family, __pyx_v_addr, ((struct sockaddr *)(&__pyx_v_saddr_st))); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 212, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "uvloop/handles/udp.pyx":211
 *             saddr = NULL
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 __convert_pyaddr_to_sockaddr(self._family, addr,
 *                                              <system.sockaddr*>&saddr_st)
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L13_try_end;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/udp.pyx":214
 *                 __convert_pyaddr_to_sockaddr(self._family, addr,
 *                                              <system.sockaddr*>&saddr_st)
 *             except (ValueError, TypeError):             # <<<<<<<<<<<<<<
 *                 raise
 *             except Exception:
 */
      __pyx_t_2 = __Pyx_PyErr_ExceptionMatches2(__pyx_builtin_ValueError, __pyx_builtin_TypeError);
      if (__pyx_t_2) {
        __Pyx_AddTraceback("uvloop.loop.UDPTransport._send", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(26, 214, __pyx_L10_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);

        /* "uvloop/handles/udp.pyx":215
 *                                              <system.sockaddr*>&saddr_st)
 *             except (ValueError, TypeError):
 *                 raise             # <<<<<<<<<<<<<<
 *             except Exception:
 *                 raise ValueError(
 */
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_8, __pyx_t_9);
        __pyx_t_1 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; 
        __PYX_ERR(26, 215, __pyx_L10_except_error)
      }

      /* "uvloop/handles/udp.pyx":216
 *             except (ValueError, TypeError):
 *                 raise
 *             except Exception:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     f'{addr!r}: socket family mismatch or '
 */
      __pyx_t_2 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
      if (__pyx_t_2) {
        __Pyx_AddTraceback("uvloop.loop.UDPTransport._send", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_1) < 0) __PYX_ERR(26, 216, __pyx_L10_except_error)
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_1);

        /* "uvloop/handles/udp.pyx":218
 *             except Exception:
 *                 raise ValueError(
 *                     f'{addr!r}: socket family mismatch or '             # <<<<<<<<<<<<<<
 *                     f'a DNS lookup is required')
 *             saddr = <system.sockaddr*>(&saddr_st)
 */
        __pyx_t_10 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_addr), __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(26, 218, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_10, __pyx_kp_u_socket_family_mismatch_or_a_DNS); if (unlikely(!__pyx_t_11)) __PYX_ERR(26, 218, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "uvloop/handles/udp.pyx":217
 *                 raise
 *             except Exception:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     f'{addr!r}: socket family mismatch or '
 *                     f'a DNS lookup is required')
 */
        __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(26, 217, __pyx_L10_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_Raise(__pyx_t_10, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __PYX_ERR(26, 217, __pyx_L10_except_error)
      }
      goto __pyx_L10_except_error;

      /* "uvloop/handles/udp.pyx":211
 *             saddr = NULL
 *         else:
 *             try:             # <<<<<<<<<<<<<<
 *                 __convert_pyaddr_to_sockaddr(self._family, addr,
 *                                              <system.sockaddr*>&saddr_st)
 */
      __pyx_L10_except_error:;
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L13_try_end:;
    }

    /* "uvloop/handles/udp.pyx":220
 *                     f'{addr!r}: socket family mismatch or '
 *                     f'a DNS lookup is required')
 *             saddr = <system.sockaddr*>(&saddr_st)             # <<<<<<<<<<<<<<
 * 
 *         if self._get_write_buffer_size() == 0:
 */
    __pyx_v_saddr = ((struct sockaddr *)(&__pyx_v_saddr_st));
  }
  __pyx_L7:;

  /* "uvloop/handles/udp.pyx":222
 *             saddr = <system.sockaddr*>(&saddr_st)
 * 
 *         if self._get_write_buffer_size() == 0:             # <<<<<<<<<<<<<<
 *             PyObject_GetBuffer(data, &try_pybuf, PyBUF_SIMPLE)
 *             try_uvbuf.base = <char*>try_pybuf.buf
 */
  __pyx_t_12 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_write_buffer_size(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 222, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_12 == 0);
  if (__pyx_t_4) {

    /* "uvloop/handles/udp.pyx":223
 * 
 *         if self._get_write_buffer_size() == 0:
 *             PyObject_GetBuffer(data, &try_pybuf, PyBUF_SIMPLE)             # <<<<<<<<<<<<<<
 *             try_uvbuf.base = <char*>try_pybuf.buf
 *             try_uvbuf.len = try_pybuf.len
 */
    __pyx_t_2 = PyObject_GetBuffer(__pyx_v_data, (&__pyx_v_try_pybuf), PyBUF_SIMPLE); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(26, 223, __pyx_L1_error)

    /* "uvloop/handles/udp.pyx":224
 *         if self._get_write_buffer_size() == 0:
 *             PyObject_GetBuffer(data, &try_pybuf, PyBUF_SIMPLE)
 *             try_uvbuf.base = <char*>try_pybuf.buf             # <<<<<<<<<<<<<<
 *             try_uvbuf.len = try_pybuf.len
 *             err = uv.uv_udp_try_send(<uv.uv_udp_t*>self._handle,
 */
    __pyx_v_try_uvbuf.base = ((char *)__pyx_v_try_pybuf.buf);

    /* "uvloop/handles/udp.pyx":225
 *             PyObject_GetBuffer(data, &try_pybuf, PyBUF_SIMPLE)
 *             try_uvbuf.base = <char*>try_pybuf.buf
 *             try_uvbuf.len = try_pybuf.len             # <<<<<<<<<<<<<<
 *             err = uv.uv_udp_try_send(<uv.uv_udp_t*>self._handle,
 *                                      &try_uvbuf,
 */
    __pyx_t_13 = __pyx_v_try_pybuf.len;
    __pyx_v_try_uvbuf.len = __pyx_t_13;

    /* "uvloop/handles/udp.pyx":226
 *             try_uvbuf.base = <char*>try_pybuf.buf
 *             try_uvbuf.len = try_pybuf.len
 *             err = uv.uv_udp_try_send(<uv.uv_udp_t*>self._handle,             # <<<<<<<<<<<<<<
 *                                      &try_uvbuf,
 *                                      1,
 */
    __pyx_v_err = uv_udp_try_send(((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), (&__pyx_v_try_uvbuf), 1, __pyx_v_saddr);

    /* "uvloop/handles/udp.pyx":230
 *                                      1,
 *                                      saddr)
 *             PyBuffer_Release(&try_pybuf)             # <<<<<<<<<<<<<<
 *         else:
 *             err = uv.UV_EAGAIN
 */
    PyBuffer_Release((&__pyx_v_try_pybuf));

    /* "uvloop/handles/udp.pyx":222
 *             saddr = <system.sockaddr*>(&saddr_st)
 * 
 *         if self._get_write_buffer_size() == 0:             # <<<<<<<<<<<<<<
 *             PyObject_GetBuffer(data, &try_pybuf, PyBUF_SIMPLE)
 *             try_uvbuf.base = <char*>try_pybuf.buf
 */
    goto __pyx_L18;
  }

  /* "uvloop/handles/udp.pyx":232
 *             PyBuffer_Release(&try_pybuf)
 *         else:
 *             err = uv.UV_EAGAIN             # <<<<<<<<<<<<<<
 * 
 *         if err == uv.UV_EAGAIN:
 */
  /*else*/ {
    __pyx_v_err = UV_EAGAIN;
  }
  __pyx_L18:;

  /* "uvloop/handles/udp.pyx":234
 *             err = uv.UV_EAGAIN
 * 
 *         if err == uv.UV_EAGAIN:             # <<<<<<<<<<<<<<
 *             ctx = _UDPSendContext.new(self, data)
 *             err = uv.uv_udp_send(&ctx.req,
 */
  __pyx_t_4 = (__pyx_v_err == UV_EAGAIN);
  if (__pyx_t_4) {

    /* "uvloop/handles/udp.pyx":235
 * 
 *         if err == uv.UV_EAGAIN:
 *             ctx = _UDPSendContext.new(self, data)             # <<<<<<<<<<<<<<
 *             err = uv.uv_udp_send(&ctx.req,
 *                                  <uv.uv_udp_t*>self._handle,
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_6uvloop_4loop_15_UDPSendContext_new(__pyx_v_self, __pyx_v_data)); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_ctx = ((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "uvloop/handles/udp.pyx":236
 *         if err == uv.UV_EAGAIN:
 *             ctx = _UDPSendContext.new(self, data)
 *             err = uv.uv_udp_send(&ctx.req,             # <<<<<<<<<<<<<<
 *                                  <uv.uv_udp_t*>self._handle,
 *                                  &ctx.uv_buf,
 */
    __pyx_v_err = uv_udp_send((&__pyx_v_ctx->req), ((uv_udp_t *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base._handle), (&__pyx_v_ctx->uv_buf), 1, __pyx_v_saddr, __pyx_f_6uvloop_4loop___uv_udp_on_send);

    /* "uvloop/handles/udp.pyx":243
 *                                  __uv_udp_on_send)
 * 
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 ctx.close()
 * 
 */
    __pyx_t_4 = (__pyx_v_err < 0);
    if (__pyx_t_4) {

      /* "uvloop/handles/udp.pyx":244
 * 
 *             if err < 0:
 *                 ctx.close()             # <<<<<<<<<<<<<<
 * 
 *                 exc = convert_error(err)
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop__UDPSendContext *)__pyx_v_ctx->__pyx_vtab)->close(__pyx_v_ctx); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 244, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "uvloop/handles/udp.pyx":246
 *                 ctx.close()
 * 
 *                 exc = convert_error(err)             # <<<<<<<<<<<<<<
 *                 if isinstance(exc, OSError):
 *                     run_in_context1(self.context.copy(), self._protocol.error_received, exc)
 */
      __pyx_t_1 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_exc = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "uvloop/handles/udp.pyx":247
 * 
 *                 exc = convert_error(err)
 *                 if isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 *                     run_in_context1(self.context.copy(), self._protocol.error_received, exc)
 *                 else:
 */
      __pyx_t_4 = PyObject_IsInstance(__pyx_v_exc, __pyx_builtin_OSError); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(26, 247, __pyx_L1_error)
      if (__pyx_t_4) {

        /* "uvloop/handles/udp.pyx":248
 *                 exc = convert_error(err)
 *                 if isinstance(exc, OSError):
 *                     run_in_context1(self.context.copy(), self._protocol.error_received, exc)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._fatal_error(exc, True)
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context, __pyx_n_s_copy); if (unlikely(!__pyx_t_8)) __PYX_ERR(26, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = NULL;
        __pyx_t_14 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_14 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_9, NULL};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_14, 0+__pyx_t_14);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 248, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._protocol, __pyx_n_s_error_received); if (unlikely(!__pyx_t_8)) __PYX_ERR(26, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __pyx_f_6uvloop_4loop_run_in_context1(__pyx_t_1, __pyx_t_8, __pyx_v_exc); if (unlikely(!__pyx_t_9)) __PYX_ERR(26, 248, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "uvloop/handles/udp.pyx":247
 * 
 *                 exc = convert_error(err)
 *                 if isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 *                     run_in_context1(self.context.copy(), self._protocol.error_received, exc)
 *                 else:
 */
        goto __pyx_L21;
      }

      /* "uvloop/handles/udp.pyx":250
 *                     run_in_context1(self.context.copy(), self._protocol.error_received, exc)
 *                 else:
 *                     self._fatal_error(exc, True)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._maybe_pause_protocol()
 */
      /*else*/ {
        __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_True, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(26, 250, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __pyx_L21:;

      /* "uvloop/handles/udp.pyx":243
 *                                  __uv_udp_on_send)
 * 
 *             if err < 0:             # <<<<<<<<<<<<<<
 *                 ctx.close()
 * 
 */
      goto __pyx_L20;
    }

    /* "uvloop/handles/udp.pyx":252
 *                     self._fatal_error(exc, True)
 *             else:
 *                 self._maybe_pause_protocol()             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    /*else*/ {
      __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._maybe_pause_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_9)) __PYX_ERR(26, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __pyx_L20:;

    /* "uvloop/handles/udp.pyx":234
 *             err = uv.UV_EAGAIN
 * 
 *         if err == uv.UV_EAGAIN:             # <<<<<<<<<<<<<<
 *             ctx = _UDPSendContext.new(self, data)
 *             err = uv.uv_udp_send(&ctx.req,
 */
    goto __pyx_L19;
  }

  /* "uvloop/handles/udp.pyx":255
 * 
 *         else:
 *             self._on_sent(convert_error(err) if err < 0 else None, self.context.copy())             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_receive(self, bytes data, object exc, object addr):
 */
  /*else*/ {
    __pyx_t_4 = (__pyx_v_err < 0);
    if (__pyx_t_4) {
      __pyx_t_8 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_8)) __PYX_ERR(26, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __pyx_t_8;
      __pyx_t_8 = 0;
    } else {
      __Pyx_INCREF(Py_None);
      __pyx_t_9 = Py_None;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context, __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    __pyx_t_14 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_14 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
      __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_14, 0+__pyx_t_14);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(26, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_15.__pyx_n = 1;
    __pyx_t_15.context = __pyx_t_8;
    __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_on_sent(__pyx_v_self, __pyx_t_9, &__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L19:;

  /* "uvloop/handles/udp.pyx":195
 *         return PseudoSocket(self._family, uv.SOCK_DGRAM, 0, fileno)
 * 
 *     cdef _send(self, object data, object addr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _UDPSendContext ctx
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":257
 *             self._on_sent(convert_error(err) if err < 0 else None, self.context.copy())
 * 
 *     cdef _on_receive(self, bytes data, object exc, object addr):             # <<<<<<<<<<<<<<
 *         if exc is None:
 *             run_in_context2(
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__on_receive(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_exc, PyObject *__pyx_v_addr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_receive", 1);

  /* "uvloop/handles/udp.pyx":258
 * 
 *     cdef _on_receive(self, bytes data, object exc, object addr):
 *         if exc is None:             # <<<<<<<<<<<<<<
 *             run_in_context2(
 *                 self.context, self._protocol.datagram_received, data, addr,
 */
  __pyx_t_1 = (__pyx_v_exc == Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":260
 *         if exc is None:
 *             run_in_context2(
 *                 self.context, self._protocol.datagram_received, data, addr,             # <<<<<<<<<<<<<<
 *             )
 *         else:
 */
    __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._protocol, __pyx_n_s_datagram_received); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/udp.pyx":259
 *     cdef _on_receive(self, bytes data, object exc, object addr):
 *         if exc is None:
 *             run_in_context2(             # <<<<<<<<<<<<<<
 *                 self.context, self._protocol.datagram_received, data, addr,
 *             )
 */
    __pyx_t_4 = __pyx_f_6uvloop_4loop_run_in_context2(__pyx_t_2, __pyx_t_3, __pyx_v_data, __pyx_v_addr); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uvloop/handles/udp.pyx":258
 * 
 *     cdef _on_receive(self, bytes data, object exc, object addr):
 *         if exc is None:             # <<<<<<<<<<<<<<
 *             run_in_context2(
 *                 self.context, self._protocol.datagram_received, data, addr,
 */
    goto __pyx_L3;
  }

  /* "uvloop/handles/udp.pyx":263
 *             )
 *         else:
 *             run_in_context1(self.context, self._protocol.error_received, exc)             # <<<<<<<<<<<<<<
 * 
 *     cdef _on_sent(self, object exc, object context=None):
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._protocol, __pyx_n_s_error_received); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_f_6uvloop_4loop_run_in_context1(__pyx_t_4, __pyx_t_3, __pyx_v_exc); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uvloop/handles/udp.pyx":257
 *             self._on_sent(convert_error(err) if err < 0 else None, self.context.copy())
 * 
 *     cdef _on_receive(self, bytes data, object exc, object addr):             # <<<<<<<<<<<<<<
 *         if exc is None:
 *             run_in_context2(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._on_receive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":265
 *             run_in_context1(self.context, self._protocol.error_received, exc)
 * 
 *     cdef _on_sent(self, object exc, object context=None):             # <<<<<<<<<<<<<<
 *         if exc is not None:
 *             if isinstance(exc, OSError):
 */

static PyObject *__pyx_f_6uvloop_4loop_12UDPTransport__on_sent(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_exc, struct __pyx_opt_args_6uvloop_4loop_12UDPTransport__on_sent *__pyx_optional_args) {
  PyObject *__pyx_v_context = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_on_sent", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_context = __pyx_optional_args->context;
    }
  }
  __Pyx_INCREF(__pyx_v_context);

  /* "uvloop/handles/udp.pyx":266
 * 
 *     cdef _on_sent(self, object exc, object context=None):
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(exc, OSError):
 *                 if context is None:
 */
  __pyx_t_1 = (__pyx_v_exc != Py_None);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":267
 *     cdef _on_sent(self, object exc, object context=None):
 *         if exc is not None:
 *             if isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     context = self.context
 */
    __pyx_t_1 = PyObject_IsInstance(__pyx_v_exc, __pyx_builtin_OSError); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(26, 267, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "uvloop/handles/udp.pyx":268
 *         if exc is not None:
 *             if isinstance(exc, OSError):
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     context = self.context
 *                 run_in_context1(context, self._protocol.error_received, exc)
 */
      __pyx_t_1 = (__pyx_v_context == Py_None);
      if (__pyx_t_1) {

        /* "uvloop/handles/udp.pyx":269
 *             if isinstance(exc, OSError):
 *                 if context is None:
 *                     context = self.context             # <<<<<<<<<<<<<<
 *                 run_in_context1(context, self._protocol.error_received, exc)
 *             else:
 */
        __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.context;
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_context, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "uvloop/handles/udp.pyx":268
 *         if exc is not None:
 *             if isinstance(exc, OSError):
 *                 if context is None:             # <<<<<<<<<<<<<<
 *                     context = self.context
 *                 run_in_context1(context, self._protocol.error_received, exc)
 */
      }

      /* "uvloop/handles/udp.pyx":270
 *                 if context is None:
 *                     context = self.context
 *                 run_in_context1(context, self._protocol.error_received, exc)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._fatal_error(
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._protocol, __pyx_n_s_error_received); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __pyx_f_6uvloop_4loop_run_in_context1(__pyx_v_context, __pyx_t_2, __pyx_v_exc); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/udp.pyx":267
 *     cdef _on_sent(self, object exc, object context=None):
 *         if exc is not None:
 *             if isinstance(exc, OSError):             # <<<<<<<<<<<<<<
 *                 if context is None:
 *                     context = self.context
 */
      goto __pyx_L4;
    }

    /* "uvloop/handles/udp.pyx":272
 *                 run_in_context1(context, self._protocol.error_received, exc)
 *             else:
 *                 self._fatal_error(             # <<<<<<<<<<<<<<
 *                     exc, False, 'Fatal write error on datagram transport')
 * 
 */
    /*else*/ {

      /* "uvloop/handles/udp.pyx":273
 *             else:
 *                 self._fatal_error(
 *                     exc, False, 'Fatal write error on datagram transport')             # <<<<<<<<<<<<<<
 * 
 *         self._maybe_resume_protocol()
 */
      __pyx_t_4.__pyx_n = 1;
      __pyx_t_4.reason = __pyx_kp_u_Fatal_write_error_on_datagram_tr;
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._fatal_error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_self), __pyx_v_exc, Py_False, &__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L4:;

    /* "uvloop/handles/udp.pyx":266
 * 
 *     cdef _on_sent(self, object exc, object context=None):
 *         if exc is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(exc, OSError):
 *                 if context is None:
 */
  }

  /* "uvloop/handles/udp.pyx":275
 *                     exc, False, 'Fatal write error on datagram transport')
 * 
 *         self._maybe_resume_protocol()             # <<<<<<<<<<<<<<
 *         if not self._get_write_buffer_size():
 *             if self._closing:
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._maybe_resume_protocol(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/udp.pyx":276
 * 
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_write_buffer_size(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 276, __pyx_L1_error)
  __pyx_t_1 = (!(__pyx_t_5 != 0));
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":277
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():
 *             if self._closing:             # <<<<<<<<<<<<<<
 *                 self._schedule_call_connection_lost(None)
 * 
 */
    if (__pyx_v_self->__pyx_base._closing) {

      /* "uvloop/handles/udp.pyx":278
 *         if not self._get_write_buffer_size():
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)             # <<<<<<<<<<<<<<
 * 
 *     # === Public API ===
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._schedule_call_connection_lost(((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)__pyx_v_self), Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/udp.pyx":277
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():
 *             if self._closing:             # <<<<<<<<<<<<<<
 *                 self._schedule_call_connection_lost(None)
 * 
 */
    }

    /* "uvloop/handles/udp.pyx":276
 * 
 *         self._maybe_resume_protocol()
 *         if not self._get_write_buffer_size():             # <<<<<<<<<<<<<<
 *             if self._closing:
 *                 self._schedule_call_connection_lost(None)
 */
  }

  /* "uvloop/handles/udp.pyx":265
 *             run_in_context1(self.context, self._protocol.error_received, exc)
 * 
 *     cdef _on_sent(self, object exc, object context=None):             # <<<<<<<<<<<<<<
 *         if exc is not None:
 *             if isinstance(exc, OSError):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport._on_sent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_context);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":282
 *     # === Public API ===
 * 
 *     def sendto(self, data, addr=None):             # <<<<<<<<<<<<<<
 *         if not data:
 *             # Replicating asyncio logic here.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12UDPTransport_3sendto(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12UDPTransport_2sendto, "UDPTransport.sendto(self, data, addr=None)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12UDPTransport_3sendto = {"sendto", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12UDPTransport_3sendto, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12UDPTransport_2sendto};
static PyObject *__pyx_pw_6uvloop_4loop_12UDPTransport_3sendto(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_addr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sendto (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,&__pyx_n_s_addr,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 282, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_addr);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 282, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sendto") < 0)) __PYX_ERR(26, 282, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_data = values[0];
    __pyx_v_addr = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sendto", 0, 1, 2, __pyx_nargs); __PYX_ERR(26, 282, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UDPTransport.sendto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12UDPTransport_2sendto(((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_v_self), __pyx_v_data, __pyx_v_addr);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12UDPTransport_2sendto(struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_addr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sendto", 0);
  __Pyx_INCREF(__pyx_v_addr);

  /* "uvloop/handles/udp.pyx":283
 * 
 *     def sendto(self, data, addr=None):
 *         if not data:             # <<<<<<<<<<<<<<
 *             # Replicating asyncio logic here.
 *             return
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_data); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(26, 283, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":285
 *         if not data:
 *             # Replicating asyncio logic here.
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self._address:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":283
 * 
 *     def sendto(self, data, addr=None):
 *         if not data:             # <<<<<<<<<<<<<<
 *             # Replicating asyncio logic here.
 *             return
 */
  }

  /* "uvloop/handles/udp.pyx":287
 *             return
 * 
 *         if self._address:             # <<<<<<<<<<<<<<
 *             if addr not in (None, self._address):
 *                 # Replicating asyncio logic here.
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->_address); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(26, 287, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":288
 * 
 *         if self._address:
 *             if addr not in (None, self._address):             # <<<<<<<<<<<<<<
 *                 # Replicating asyncio logic here.
 *                 raise ValueError(
 */
    __Pyx_INCREF(__pyx_v_addr);
    __pyx_t_3 = __pyx_v_addr;
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, Py_None, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 288, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(26, 288, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_self->_address, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 288, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(26, 288, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_2;
    if (unlikely(__pyx_t_1)) {

      /* "uvloop/handles/udp.pyx":291
 *                 # Replicating asyncio logic here.
 *                 raise ValueError(
 *                     'Invalid address: must be None or %s' % (self._address,))             # <<<<<<<<<<<<<<
 * 
 *             # Instead of setting addr to self._address below like what asyncio
 */
      __pyx_t_3 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_self->_address), __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Invalid_address_must_be_None_or, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/udp.pyx":290
 *             if addr not in (None, self._address):
 *                 # Replicating asyncio logic here.
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'Invalid address: must be None or %s' % (self._address,))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(26, 290, __pyx_L1_error)

      /* "uvloop/handles/udp.pyx":288
 * 
 *         if self._address:
 *             if addr not in (None, self._address):             # <<<<<<<<<<<<<<
 *                 # Replicating asyncio logic here.
 *                 raise ValueError(
 */
    }

    /* "uvloop/handles/udp.pyx":295
 *             # Instead of setting addr to self._address below like what asyncio
 *             # does, we depend on previous uv_udp_connect() to set the address
 *             addr = None             # <<<<<<<<<<<<<<
 * 
 *         if self._conn_lost:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_addr, Py_None);

    /* "uvloop/handles/udp.pyx":287
 *             return
 * 
 *         if self._address:             # <<<<<<<<<<<<<<
 *             if addr not in (None, self._address):
 *                 # Replicating asyncio logic here.
 */
  }

  /* "uvloop/handles/udp.pyx":297
 *             addr = None
 * 
 *         if self._conn_lost:             # <<<<<<<<<<<<<<
 *             # Replicating asyncio logic here.
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base._conn_lost != 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":299
 *         if self._conn_lost:
 *             # Replicating asyncio logic here.
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:             # <<<<<<<<<<<<<<
 *                 aio_logger.warning('socket.send() raised exception.')
 *             self._conn_lost += 1
 */
    __pyx_t_1 = (__pyx_v_self->__pyx_base._conn_lost >= __pyx_e_6uvloop_4loop_LOG_THRESHOLD_FOR_CONNLOST_WRITES);
    if (__pyx_t_1) {

      /* "uvloop/handles/udp.pyx":300
 *             # Replicating asyncio logic here.
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 *                 aio_logger.warning('socket.send() raised exception.')             # <<<<<<<<<<<<<<
 *             self._conn_lost += 1
 *             return
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_kp_u_socket_send_raised_exception};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 300, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/udp.pyx":299
 *         if self._conn_lost:
 *             # Replicating asyncio logic here.
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:             # <<<<<<<<<<<<<<
 *                 aio_logger.warning('socket.send() raised exception.')
 *             self._conn_lost += 1
 */
    }

    /* "uvloop/handles/udp.pyx":301
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 *                 aio_logger.warning('socket.send() raised exception.')
 *             self._conn_lost += 1             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_v_self->__pyx_base._conn_lost = (__pyx_v_self->__pyx_base._conn_lost + 1);

    /* "uvloop/handles/udp.pyx":302
 *                 aio_logger.warning('socket.send() raised exception.')
 *             self._conn_lost += 1
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._send(data, addr)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":297
 *             addr = None
 * 
 *         if self._conn_lost:             # <<<<<<<<<<<<<<
 *             # Replicating asyncio logic here.
 *             if self._conn_lost >= LOG_THRESHOLD_FOR_CONNLOST_WRITES:
 */
  }

  /* "uvloop/handles/udp.pyx":304
 *             return
 * 
 *         self._send(data, addr)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_send(__pyx_v_self, __pyx_v_data, __pyx_v_addr); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/udp.pyx":282
 *     # === Public API ===
 * 
 *     def sendto(self, data, addr=None):             # <<<<<<<<<<<<<<
 *         if not data:
 *             # Replicating asyncio logic here.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.UDPTransport.sendto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_addr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12UDPTransport_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12UDPTransport_4__reduce_cython__, "UDPTransport.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12UDPTransport_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12UDPTransport_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12UDPTransport_4__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_12UDPTransport_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_12UDPTransport_4__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12UDPTransport_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UDPTransport.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12UDPTransport_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_12UDPTransport_6__setstate_cython__, "UDPTransport.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12UDPTransport_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12UDPTransport_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12UDPTransport_6__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_12UDPTransport_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.UDPTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_12UDPTransport_6__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_12UDPTransport_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.UDPTransport.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/handles/udp.pyx":307
 * 
 * 
 * cdef void __uv_udp_on_receive(             # <<<<<<<<<<<<<<
 *     uv.uv_udp_t* handle,
 *     ssize_t nread,
 */

static void __pyx_f_6uvloop_4loop___uv_udp_on_receive(uv_udp_t *__pyx_v_handle, Py_ssize_t __pyx_v_nread, CYTHON_UNUSED uv_buf_t const *__pyx_v_buf, struct sockaddr const *__pyx_v_addr, CYTHON_UNUSED unsigned int __pyx_v_flags) {
  struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_udp = 0;
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_pyaddr = 0;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  char const *__pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_udp_on_receive", 0);

  /* "uvloop/handles/udp.pyx":315
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UDPTransport receive callback") == 0:
 *         return
 */
  __pyx_t_1 = __pyx_f_6uvloop_4loop___ensure_handle_data(((uv_handle_t *)__pyx_v_handle), ((char const *)"UDPTransport receive callback")); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(26, 315, __pyx_L1_error)

  /* "uvloop/handles/udp.pyx":316
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,
 *                             "UDPTransport receive callback") == 0:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_2 = (__pyx_t_1 == 0);

  /* "uvloop/handles/udp.pyx":315
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UDPTransport receive callback") == 0:
 *         return
 */
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":317
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,
 *                             "UDPTransport receive callback") == 0:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":315
 * ) noexcept with gil:
 * 
 *     if __ensure_handle_data(<uv.uv_handle_t*>handle,             # <<<<<<<<<<<<<<
 *                             "UDPTransport receive callback") == 0:
 *         return
 */
  }

  /* "uvloop/handles/udp.pyx":320
 * 
 *     cdef:
 *         UDPTransport udp = <UDPTransport>handle.data             # <<<<<<<<<<<<<<
 *         Loop loop = udp._loop
 *         bytes data
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_handle->data);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_udp = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/udp.pyx":321
 *     cdef:
 *         UDPTransport udp = <UDPTransport>handle.data
 *         Loop loop = udp._loop             # <<<<<<<<<<<<<<
 *         bytes data
 *         object pyaddr
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base._loop);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/handles/udp.pyx":327
 *     # It's OK to free the buffer early, since nothing will
 *     # be able to touch it until this method is done.
 *     __loop_free_buffer(loop)             # <<<<<<<<<<<<<<
 * 
 *     if udp._closed:
 */
  __pyx_f_6uvloop_4loop___loop_free_buffer(__pyx_v_loop); if (unlikely(PyErr_Occurred())) __PYX_ERR(26, 327, __pyx_L1_error)

  /* "uvloop/handles/udp.pyx":329
 *     __loop_free_buffer(loop)
 * 
 *     if udp._closed:             # <<<<<<<<<<<<<<
 *         # The handle was closed, there is no reason to
 *         # do any work now.
 */
  if (__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base._closed) {

    /* "uvloop/handles/udp.pyx":332
 *         # The handle was closed, there is no reason to
 *         # do any work now.
 *         udp.__receiving_stopped()  # Just in case.             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_stopped(__pyx_v_udp); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/handles/udp.pyx":333
 *         # do any work now.
 *         udp.__receiving_stopped()  # Just in case.
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if addr is NULL and nread == 0:
 */
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":329
 *     __loop_free_buffer(loop)
 * 
 *     if udp._closed:             # <<<<<<<<<<<<<<
 *         # The handle was closed, there is no reason to
 *         # do any work now.
 */
  }

  /* "uvloop/handles/udp.pyx":335
 *         return
 * 
 *     if addr is NULL and nread == 0:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #      addr: struct sockaddr* containing the address
 */
  __pyx_t_1 = (__pyx_v_addr == NULL);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = (__pyx_v_nread == 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":346
 *         #      addr != NULL when an empty UDP packet is
 *         #      received.
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if addr is NULL:
 */
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":335
 *         return
 * 
 *     if addr is NULL and nread == 0:             # <<<<<<<<<<<<<<
 *         # From libuv docs:
 *         #      addr: struct sockaddr* containing the address
 */
  }

  /* "uvloop/handles/udp.pyx":348
 *         return
 * 
 *     if addr is NULL:             # <<<<<<<<<<<<<<
 *         pyaddr = None
 *     elif addr.sa_family == uv.AF_UNSPEC:
 */
  __pyx_t_2 = (__pyx_v_addr == NULL);
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":349
 * 
 *     if addr is NULL:
 *         pyaddr = None             # <<<<<<<<<<<<<<
 *     elif addr.sa_family == uv.AF_UNSPEC:
 *         # https://github.com/MagicStack/uvloop/issues/304
 */
    __Pyx_INCREF(Py_None);
    __pyx_v_pyaddr = Py_None;

    /* "uvloop/handles/udp.pyx":348
 *         return
 * 
 *     if addr is NULL:             # <<<<<<<<<<<<<<
 *         pyaddr = None
 *     elif addr.sa_family == uv.AF_UNSPEC:
 */
    goto __pyx_L8;
  }

  /* "uvloop/handles/udp.pyx":350
 *     if addr is NULL:
 *         pyaddr = None
 *     elif addr.sa_family == uv.AF_UNSPEC:             # <<<<<<<<<<<<<<
 *         # https://github.com/MagicStack/uvloop/issues/304
 *         if system.PLATFORM_IS_LINUX:
 */
  __pyx_t_2 = (__pyx_v_addr->sa_family == AF_UNSPEC);
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":352
 *     elif addr.sa_family == uv.AF_UNSPEC:
 *         # https://github.com/MagicStack/uvloop/issues/304
 *         if system.PLATFORM_IS_LINUX:             # <<<<<<<<<<<<<<
 *             pyaddr = None
 *         else:
 */
    __pyx_t_2 = (PLATFORM_IS_LINUX != 0);
    if (__pyx_t_2) {

      /* "uvloop/handles/udp.pyx":353
 *         # https://github.com/MagicStack/uvloop/issues/304
 *         if system.PLATFORM_IS_LINUX:
 *             pyaddr = None             # <<<<<<<<<<<<<<
 *         else:
 *             pyaddr = ''
 */
      __Pyx_INCREF(Py_None);
      __pyx_v_pyaddr = Py_None;

      /* "uvloop/handles/udp.pyx":352
 *     elif addr.sa_family == uv.AF_UNSPEC:
 *         # https://github.com/MagicStack/uvloop/issues/304
 *         if system.PLATFORM_IS_LINUX:             # <<<<<<<<<<<<<<
 *             pyaddr = None
 *         else:
 */
      goto __pyx_L9;
    }

    /* "uvloop/handles/udp.pyx":355
 *             pyaddr = None
 *         else:
 *             pyaddr = ''             # <<<<<<<<<<<<<<
 *     else:
 *         try:
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_kp_u__10);
      __pyx_v_pyaddr = __pyx_kp_u__10;
    }
    __pyx_L9:;

    /* "uvloop/handles/udp.pyx":350
 *     if addr is NULL:
 *         pyaddr = None
 *     elif addr.sa_family == uv.AF_UNSPEC:             # <<<<<<<<<<<<<<
 *         # https://github.com/MagicStack/uvloop/issues/304
 *         if system.PLATFORM_IS_LINUX:
 */
    goto __pyx_L8;
  }

  /* "uvloop/handles/udp.pyx":357
 *             pyaddr = ''
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             pyaddr = __convert_sockaddr_to_pyaddr(addr)
 *         except BaseException as exc:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "uvloop/handles/udp.pyx":358
 *     else:
 *         try:
 *             pyaddr = __convert_sockaddr_to_pyaddr(addr)             # <<<<<<<<<<<<<<
 *         except BaseException as exc:
 *             udp._error(exc, False)
 */
        __pyx_t_3 = __pyx_f_6uvloop_4loop___convert_sockaddr_to_pyaddr(__pyx_v_addr); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 358, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_v_pyaddr = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "uvloop/handles/udp.pyx":357
 *             pyaddr = ''
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             pyaddr = __convert_sockaddr_to_pyaddr(addr)
 *         except BaseException as exc:
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L15_try_end;
      __pyx_L10_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/handles/udp.pyx":359
 *         try:
 *             pyaddr = __convert_sockaddr_to_pyaddr(addr)
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             udp._error(exc, False)
 *             return
 */
      __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_7) {
        __Pyx_AddTraceback("uvloop.loop.__uv_udp_on_receive", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_8, &__pyx_t_9) < 0) __PYX_ERR(26, 359, __pyx_L12_except_error)
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __pyx_v_exc = __pyx_t_8;
        /*try:*/ {

          /* "uvloop/handles/udp.pyx":360
 *             pyaddr = __convert_sockaddr_to_pyaddr(addr)
 *         except BaseException as exc:
 *             udp._error(exc, False)             # <<<<<<<<<<<<<<
 *             return
 * 
 */
          __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_udp), __pyx_v_exc, Py_False); if (unlikely(!__pyx_t_10)) __PYX_ERR(26, 360, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "uvloop/handles/udp.pyx":361
 *         except BaseException as exc:
 *             udp._error(exc, False)
 *             return             # <<<<<<<<<<<<<<
 * 
 *     if nread < 0:
 */
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L20_return;
        }

        /* "uvloop/handles/udp.pyx":359
 *         try:
 *             pyaddr = __convert_sockaddr_to_pyaddr(addr)
 *         except BaseException as exc:             # <<<<<<<<<<<<<<
 *             udp._error(exc, False)
 *             return
 */
        /*finally:*/ {
          __pyx_L21_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __pyx_t_7 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
            }
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
            __pyx_lineno = __pyx_t_7; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
            goto __pyx_L12_except_error;
          }
          __pyx_L20_return: {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
            goto __pyx_L13_except_return;
          }
        }
      }
      goto __pyx_L12_except_error;

      /* "uvloop/handles/udp.pyx":357
 *             pyaddr = ''
 *     else:
 *         try:             # <<<<<<<<<<<<<<
 *             pyaddr = __convert_sockaddr_to_pyaddr(addr)
 *         except BaseException as exc:
 */
      __pyx_L12_except_error:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L13_except_return:;
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L0;
      __pyx_L15_try_end:;
    }
  }
  __pyx_L8:;

  /* "uvloop/handles/udp.pyx":363
 *             return
 * 
 *     if nread < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(nread)
 *         udp._on_receive(None, exc, pyaddr)
 */
  __pyx_t_2 = (__pyx_v_nread < 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":364
 * 
 *     if nread < 0:
 *         exc = convert_error(nread)             # <<<<<<<<<<<<<<
 *         udp._on_receive(None, exc, pyaddr)
 *         return
 */
    __pyx_t_9 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_nread); if (unlikely(!__pyx_t_9)) __PYX_ERR(26, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_v_exc = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "uvloop/handles/udp.pyx":365
 *     if nread < 0:
 *         exc = convert_error(nread)
 *         udp._on_receive(None, exc, pyaddr)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_on_receive(__pyx_v_udp, ((PyObject*)Py_None), __pyx_v_exc, __pyx_v_pyaddr); if (unlikely(!__pyx_t_9)) __PYX_ERR(26, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "uvloop/handles/udp.pyx":366
 *         exc = convert_error(nread)
 *         udp._on_receive(None, exc, pyaddr)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     if nread == 0:
 */
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":363
 *             return
 * 
 *     if nread < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(nread)
 *         udp._on_receive(None, exc, pyaddr)
 */
  }

  /* "uvloop/handles/udp.pyx":368
 *         return
 * 
 *     if nread == 0:             # <<<<<<<<<<<<<<
 *         data = b''
 *     else:
 */
  __pyx_t_2 = (__pyx_v_nread == 0);
  if (__pyx_t_2) {

    /* "uvloop/handles/udp.pyx":369
 * 
 *     if nread == 0:
 *         data = b''             # <<<<<<<<<<<<<<
 *     else:
 *         data = loop._recv_buffer[:nread]
 */
    __Pyx_INCREF(__pyx_kp_b__10);
    __pyx_v_data = __pyx_kp_b__10;

    /* "uvloop/handles/udp.pyx":368
 *         return
 * 
 *     if nread == 0:             # <<<<<<<<<<<<<<
 *         data = b''
 *     else:
 */
    goto __pyx_L28;
  }

  /* "uvloop/handles/udp.pyx":371
 *         data = b''
 *     else:
 *         data = loop._recv_buffer[:nread]             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  /*else*/ {
    __pyx_t_9 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_loop->_recv_buffer) + 0, __pyx_v_nread - 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(26, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_v_data = ((PyObject*)__pyx_t_9);
    __pyx_t_9 = 0;
  }
  __pyx_L28:;

  /* "uvloop/handles/udp.pyx":373
 *         data = loop._recv_buffer[:nread]
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         udp._on_receive(data, None, pyaddr)
 *     except BaseException as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "uvloop/handles/udp.pyx":374
 * 
 *     try:
 *         udp._on_receive(data, None, pyaddr)             # <<<<<<<<<<<<<<
 *     except BaseException as exc:
 *         udp._error(exc, False)
 */
      __pyx_t_9 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_on_receive(__pyx_v_udp, __pyx_v_data, Py_None, __pyx_v_pyaddr); if (unlikely(!__pyx_t_9)) __PYX_ERR(26, 374, __pyx_L29_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "uvloop/handles/udp.pyx":373
 *         data = loop._recv_buffer[:nread]
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         udp._on_receive(data, None, pyaddr)
 *     except BaseException as exc:
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L34_try_end;
    __pyx_L29_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "uvloop/handles/udp.pyx":375
 *     try:
 *         udp._on_receive(data, None, pyaddr)
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         udp._error(exc, False)
 * 
 */
    __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_11) {
      __Pyx_AddTraceback("uvloop.loop.__uv_udp_on_receive", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_8, &__pyx_t_3) < 0) __PYX_ERR(26, 375, __pyx_L31_except_error)
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __pyx_v_exc = __pyx_t_8;
      /*try:*/ {

        /* "uvloop/handles/udp.pyx":376
 *         udp._on_receive(data, None, pyaddr)
 *     except BaseException as exc:
 *         udp._error(exc, False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_udp), __pyx_v_exc, Py_False); if (unlikely(!__pyx_t_10)) __PYX_ERR(26, 376, __pyx_L40_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }

      /* "uvloop/handles/udp.pyx":375
 *     try:
 *         udp._on_receive(data, None, pyaddr)
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         udp._error(exc, False)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L41;
        }
        __pyx_L40_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_17, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_13);
          __pyx_t_11 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
          }
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_18, __pyx_t_17, __pyx_t_16);
          __pyx_t_18 = 0; __pyx_t_17 = 0; __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0;
          __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_19;
          goto __pyx_L31_except_error;
        }
        __pyx_L41:;
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L30_exception_handled;
    }
    goto __pyx_L31_except_error;

    /* "uvloop/handles/udp.pyx":373
 *         data = loop._recv_buffer[:nread]
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         udp._on_receive(data, None, pyaddr)
 *     except BaseException as exc:
 */
    __pyx_L31_except_error:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L30_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_5, __pyx_t_4);
    __pyx_L34_try_end:;
  }

  /* "uvloop/handles/udp.pyx":307
 * 
 * 
 * cdef void __uv_udp_on_receive(             # <<<<<<<<<<<<<<
 *     uv.uv_udp_t* handle,
 *     ssize_t nread,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_udp_on_receive", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_udp);
  __Pyx_XDECREF((PyObject *)__pyx_v_loop);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_pyaddr);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/handles/udp.pyx":379
 * 
 * 
 * cdef void __uv_udp_on_send(             # <<<<<<<<<<<<<<
 *     uv.uv_udp_send_t* req,
 *     int status,
 */

static void __pyx_f_6uvloop_4loop___uv_udp_on_send(uv_udp_send_t *__pyx_v_req, int __pyx_v_status) {
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_v_ctx = 0;
  struct __pyx_obj_6uvloop_4loop_UDPTransport *__pyx_v_udp = 0;
  PyObject *__pyx_v_exc = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("__uv_udp_on_send", 0);

  /* "uvloop/handles/udp.pyx":384
 * ) noexcept with gil:
 * 
 *     if req.data is NULL:             # <<<<<<<<<<<<<<
 *         # Shouldn't happen as:
 *         #    - _UDPSendContext does an extra INCREF in its 'init()'
 */
  __pyx_t_1 = (__pyx_v_req->data == NULL);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":388
 *         #    - _UDPSendContext does an extra INCREF in its 'init()'
 *         #    - _UDPSendContext holds a ref to the relevant UDPTransport
 *         aio_logger.error(             # <<<<<<<<<<<<<<
 *             'UVStream.write callback called with NULL req.data, status=%r',
 *             status)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_6uvloop_4loop_aio_logger, __pyx_n_s_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "uvloop/handles/udp.pyx":390
 *         aio_logger.error(
 *             'UVStream.write callback called with NULL req.data, status=%r',
 *             status)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_4)) __PYX_ERR(26, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_kp_u_UVStream_write_callback_called_w, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/udp.pyx":391
 *             'UVStream.write callback called with NULL req.data, status=%r',
 *             status)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
    goto __pyx_L0;

    /* "uvloop/handles/udp.pyx":384
 * ) noexcept with gil:
 * 
 *     if req.data is NULL:             # <<<<<<<<<<<<<<
 *         # Shouldn't happen as:
 *         #    - _UDPSendContext does an extra INCREF in its 'init()'
 */
  }

  /* "uvloop/handles/udp.pyx":394
 * 
 *     cdef:
 *         _UDPSendContext ctx = <_UDPSendContext> req.data             # <<<<<<<<<<<<<<
 *         UDPTransport udp = <UDPTransport>ctx.udp
 * 
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_req->data);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ctx = ((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/handles/udp.pyx":395
 *     cdef:
 *         _UDPSendContext ctx = <_UDPSendContext> req.data
 *         UDPTransport udp = <UDPTransport>ctx.udp             # <<<<<<<<<<<<<<
 * 
 *     ctx.close()
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_ctx->udp);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_udp = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/handles/udp.pyx":397
 *         UDPTransport udp = <UDPTransport>ctx.udp
 * 
 *     ctx.close()             # <<<<<<<<<<<<<<
 * 
 *     if status < 0:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop__UDPSendContext *)__pyx_v_ctx->__pyx_vtab)->close(__pyx_v_ctx); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/udp.pyx":399
 *     ctx.close()
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *         print(exc)
 */
  __pyx_t_1 = (__pyx_v_status < 0);
  if (__pyx_t_1) {

    /* "uvloop/handles/udp.pyx":400
 * 
 *     if status < 0:
 *         exc = convert_error(status)             # <<<<<<<<<<<<<<
 *         print(exc)
 *     else:
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_status); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "uvloop/handles/udp.pyx":401
 *     if status < 0:
 *         exc = convert_error(status)
 *         print(exc)             # <<<<<<<<<<<<<<
 *     else:
 *         exc = None
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_v_exc); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/handles/udp.pyx":399
 *     ctx.close()
 * 
 *     if status < 0:             # <<<<<<<<<<<<<<
 *         exc = convert_error(status)
 *         print(exc)
 */
    goto __pyx_L4;
  }

  /* "uvloop/handles/udp.pyx":403
 *         print(exc)
 *     else:
 *         exc = None             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_exc = Py_None;
  }
  __pyx_L4:;

  /* "uvloop/handles/udp.pyx":405
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         udp._on_sent(exc)
 *     except BaseException as exc:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    /*try:*/ {

      /* "uvloop/handles/udp.pyx":406
 * 
 *     try:
 *         udp._on_sent(exc)             # <<<<<<<<<<<<<<
 *     except BaseException as exc:
 *         udp._error(exc, False)
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_on_sent(__pyx_v_udp, __pyx_v_exc, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(26, 406, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/handles/udp.pyx":405
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         udp._on_sent(exc)
 *     except BaseException as exc:
 */
    }
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L10_try_end;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "uvloop/handles/udp.pyx":407
 *     try:
 *         udp._on_sent(exc)
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         udp._error(exc, False)
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
    if (__pyx_t_10) {
      __Pyx_AddTraceback("uvloop.loop.__uv_udp_on_send", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(26, 407, __pyx_L7_except_error)
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_exc, __pyx_t_3);
      /*try:*/ {

        /* "uvloop/handles/udp.pyx":408
 *         udp._on_sent(exc)
 *     except BaseException as exc:
 *         udp._error(exc, False)             # <<<<<<<<<<<<<<
 */
        __pyx_t_5 = ((struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport *)__pyx_v_udp->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base._error(((struct __pyx_obj_6uvloop_4loop_UVHandle *)__pyx_v_udp), __pyx_v_exc, Py_False); if (unlikely(!__pyx_t_5)) __PYX_ERR(26, 408, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }

      /* "uvloop/handles/udp.pyx":407
 *     try:
 *         udp._on_sent(exc)
 *     except BaseException as exc:             # <<<<<<<<<<<<<<
 *         udp._error(exc, False)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          goto __pyx_L17;
        }
        __pyx_L16_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __pyx_t_10 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc); __pyx_v_exc = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
          }
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
          __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
          goto __pyx_L7_except_error;
        }
        __pyx_L17:;
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L6_exception_handled;
    }
    goto __pyx_L7_except_error;

    /* "uvloop/handles/udp.pyx":405
 *         exc = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         udp._on_sent(exc)
 *     except BaseException as exc:
 */
    __pyx_L7_except_error:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    goto __pyx_L1_error;
    __pyx_L6_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
    __pyx_L10_try_end:;
  }

  /* "uvloop/handles/udp.pyx":379
 * 
 * 
 * cdef void __uv_udp_on_send(             # <<<<<<<<<<<<<<
 *     uv.uv_udp_send_t* req,
 *     int status,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uvloop.loop.__uv_udp_on_send", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XDECREF((PyObject *)__pyx_v_udp);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "uvloop/server.pyx":5
 * 
 * cdef class Server:
 *     def __cinit__(self, Loop loop):             # <<<<<<<<<<<<<<
 *         self._loop = loop
 *         self._servers = []
 */

/* Python wrapper */
static int __pyx_pw_6uvloop_4loop_6Server_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6uvloop_4loop_6Server_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_loop,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_loop)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(27, 5, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(27, 5, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, __pyx_nargs); __PYX_ERR(27, 5, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Server.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_loop), __pyx_ptype_6uvloop_4loop_Loop, 1, "loop", 0))) __PYX_ERR(27, 5, __pyx_L1_error)
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server___cinit__(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self), __pyx_v_loop);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6uvloop_4loop_6Server___cinit__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_Loop *__pyx_v_loop) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uvloop/server.pyx":6
 * cdef class Server:
 *     def __cinit__(self, Loop loop):
 *         self._loop = loop             # <<<<<<<<<<<<<<
 *         self._servers = []
 *         self._waiters = []
 */
  __Pyx_INCREF((PyObject *)__pyx_v_loop);
  __Pyx_GIVEREF((PyObject *)__pyx_v_loop);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_loop);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_loop);
  __pyx_v_self->_loop = __pyx_v_loop;

  /* "uvloop/server.pyx":7
 *     def __cinit__(self, Loop loop):
 *         self._loop = loop
 *         self._servers = []             # <<<<<<<<<<<<<<
 *         self._waiters = []
 *         self._active_count = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_servers);
  __Pyx_DECREF(__pyx_v_self->_servers);
  __pyx_v_self->_servers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/server.pyx":8
 *         self._loop = loop
 *         self._servers = []
 *         self._waiters = []             # <<<<<<<<<<<<<<
 *         self._active_count = 0
 *         self._serving_forever_fut = None
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_waiters);
  __Pyx_DECREF(__pyx_v_self->_waiters);
  __pyx_v_self->_waiters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/server.pyx":9
 *         self._servers = []
 *         self._waiters = []
 *         self._active_count = 0             # <<<<<<<<<<<<<<
 *         self._serving_forever_fut = None
 * 
 */
  __pyx_v_self->_active_count = 0;

  /* "uvloop/server.pyx":10
 *         self._waiters = []
 *         self._active_count = 0
 *         self._serving_forever_fut = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _add_server(self, UVStreamServer srv):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_serving_forever_fut);
  __Pyx_DECREF(__pyx_v_self->_serving_forever_fut);
  __pyx_v_self->_serving_forever_fut = Py_None;

  /* "uvloop/server.pyx":5
 * 
 * cdef class Server:
 *     def __cinit__(self, Loop loop):             # <<<<<<<<<<<<<<
 *         self._loop = loop
 *         self._servers = []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Server.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":12
 *         self._serving_forever_fut = None
 * 
 *     cdef _add_server(self, UVStreamServer srv):             # <<<<<<<<<<<<<<
 *         self._servers.append(srv)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_6Server__add_server(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, struct __pyx_obj_6uvloop_4loop_UVStreamServer *__pyx_v_srv) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_add_server", 1);

  /* "uvloop/server.pyx":13
 * 
 *     cdef _add_server(self, UVStreamServer srv):
 *         self._servers.append(srv)             # <<<<<<<<<<<<<<
 * 
 *     cdef _start_serving(self):
 */
  if (unlikely(__pyx_v_self->_servers == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(27, 13, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_Append(__pyx_v_self->_servers, ((PyObject *)__pyx_v_srv)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(27, 13, __pyx_L1_error)

  /* "uvloop/server.pyx":12
 *         self._serving_forever_fut = None
 * 
 *     cdef _add_server(self, UVStreamServer srv):             # <<<<<<<<<<<<<<
 *         self._servers.append(srv)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server._add_server", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":15
 *         self._servers.append(srv)
 * 
 *     cdef _start_serving(self):             # <<<<<<<<<<<<<<
 *         if self._serving:
 *             return
 */

static PyObject *__pyx_f_6uvloop_4loop_6Server__start_serving(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_v_server = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_start_serving", 1);

  /* "uvloop/server.pyx":16
 * 
 *     cdef _start_serving(self):
 *         if self._serving:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  if (__pyx_v_self->_serving) {

    /* "uvloop/server.pyx":17
 *     cdef _start_serving(self):
 *         if self._serving:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         self._serving = 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/server.pyx":16
 * 
 *     cdef _start_serving(self):
 *         if self._serving:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/server.pyx":19
 *             return
 * 
 *         self._serving = 1             # <<<<<<<<<<<<<<
 *         for server in self._servers:
 *             (<UVStreamServer>server).listen()
 */
  __pyx_v_self->_serving = 1;

  /* "uvloop/server.pyx":20
 * 
 *         self._serving = 1
 *         for server in self._servers:             # <<<<<<<<<<<<<<
 *             (<UVStreamServer>server).listen()
 * 
 */
  if (unlikely(__pyx_v_self->_servers == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(27, 20, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->_servers; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(27, 20, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(27, 20, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 20, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_server, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/server.pyx":21
 *         self._serving = 1
 *         for server in self._servers:
 *             (<UVStreamServer>server).listen()             # <<<<<<<<<<<<<<
 * 
 *     cdef _wakeup(self):
 */
    __pyx_t_3 = __pyx_f_6uvloop_4loop_14UVStreamServer_listen(((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_server)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 21, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/server.pyx":20
 * 
 *         self._serving = 1
 *         for server in self._servers:             # <<<<<<<<<<<<<<
 *             (<UVStreamServer>server).listen()
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/server.pyx":15
 *         self._servers.append(srv)
 * 
 *     cdef _start_serving(self):             # <<<<<<<<<<<<<<
 *         if self._serving:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Server._start_serving", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_server);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":23
 *             (<UVStreamServer>server).listen()
 * 
 *     cdef _wakeup(self):             # <<<<<<<<<<<<<<
 *         cdef list waiters
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_6Server__wakeup(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_v_waiters = 0;
  PyObject *__pyx_v_waiter = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_wakeup", 1);

  /* "uvloop/server.pyx":26
 *         cdef list waiters
 * 
 *         waiters = self._waiters             # <<<<<<<<<<<<<<
 *         self._waiters = None
 *         for waiter in waiters:
 */
  __pyx_t_1 = __pyx_v_self->_waiters;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_waiters = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/server.pyx":27
 * 
 *         waiters = self._waiters
 *         self._waiters = None             # <<<<<<<<<<<<<<
 *         for waiter in waiters:
 *             if not waiter.done():
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_waiters);
  __Pyx_DECREF(__pyx_v_self->_waiters);
  __pyx_v_self->_waiters = ((PyObject*)Py_None);

  /* "uvloop/server.pyx":28
 *         waiters = self._waiters
 *         self._waiters = None
 *         for waiter in waiters:             # <<<<<<<<<<<<<<
 *             if not waiter.done():
 *                 waiter.set_result(waiter)
 */
  if (unlikely(__pyx_v_waiters == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(27, 28, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_waiters; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(27, 28, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(27, 28, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 28, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_waiter, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "uvloop/server.pyx":29
 *         self._waiters = None
 *         for waiter in waiters:
 *             if not waiter.done():             # <<<<<<<<<<<<<<
 *                 waiter.set_result(waiter)
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_done); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 29, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(27, 29, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = (!__pyx_t_7);
    if (__pyx_t_8) {

      /* "uvloop/server.pyx":30
 *         for waiter in waiters:
 *             if not waiter.done():
 *                 waiter.set_result(waiter)             # <<<<<<<<<<<<<<
 * 
 *     cdef _attach(self):
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_waiter, __pyx_n_s_set_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 30, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_waiter};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 30, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "uvloop/server.pyx":29
 *         self._waiters = None
 *         for waiter in waiters:
 *             if not waiter.done():             # <<<<<<<<<<<<<<
 *                 waiter.set_result(waiter)
 * 
 */
    }

    /* "uvloop/server.pyx":28
 *         waiters = self._waiters
 *         self._waiters = None
 *         for waiter in waiters:             # <<<<<<<<<<<<<<
 *             if not waiter.done():
 *                 waiter.set_result(waiter)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/server.pyx":23
 *             (<UVStreamServer>server).listen()
 * 
 *     cdef _wakeup(self):             # <<<<<<<<<<<<<<
 *         cdef list waiters
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.Server._wakeup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_waiters);
  __Pyx_XDECREF(__pyx_v_waiter);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":32
 *                 waiter.set_result(waiter)
 * 
 *     cdef _attach(self):             # <<<<<<<<<<<<<<
 *         assert self._servers is not None
 *         self._active_count += 1
 */

static PyObject *__pyx_f_6uvloop_4loop_6Server__attach(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_attach", 1);

  /* "uvloop/server.pyx":33
 * 
 *     cdef _attach(self):
 *         assert self._servers is not None             # <<<<<<<<<<<<<<
 *         self._active_count += 1
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_self->_servers != ((PyObject*)Py_None));
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(27, 33, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(27, 33, __pyx_L1_error)
  #endif

  /* "uvloop/server.pyx":34
 *     cdef _attach(self):
 *         assert self._servers is not None
 *         self._active_count += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef _detach(self):
 */
  __pyx_v_self->_active_count = (__pyx_v_self->_active_count + 1);

  /* "uvloop/server.pyx":32
 *                 waiter.set_result(waiter)
 * 
 *     cdef _attach(self):             # <<<<<<<<<<<<<<
 *         assert self._servers is not None
 *         self._active_count += 1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server._attach", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":36
 *         self._active_count += 1
 * 
 *     cdef _detach(self):             # <<<<<<<<<<<<<<
 *         assert self._active_count > 0
 *         self._active_count -= 1
 */

static PyObject *__pyx_f_6uvloop_4loop_6Server__detach(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_detach", 1);

  /* "uvloop/server.pyx":37
 * 
 *     cdef _detach(self):
 *         assert self._active_count > 0             # <<<<<<<<<<<<<<
 *         self._active_count -= 1
 *         if self._active_count == 0 and self._servers is None:
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = (__pyx_v_self->_active_count > 0);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(27, 37, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(27, 37, __pyx_L1_error)
  #endif

  /* "uvloop/server.pyx":38
 *     cdef _detach(self):
 *         assert self._active_count > 0
 *         self._active_count -= 1             # <<<<<<<<<<<<<<
 *         if self._active_count == 0 and self._servers is None:
 *             self._wakeup()
 */
  __pyx_v_self->_active_count = (__pyx_v_self->_active_count - 1);

  /* "uvloop/server.pyx":39
 *         assert self._active_count > 0
 *         self._active_count -= 1
 *         if self._active_count == 0 and self._servers is None:             # <<<<<<<<<<<<<<
 *             self._wakeup()
 * 
 */
  __pyx_t_2 = (__pyx_v_self->_active_count == 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->_servers == ((PyObject*)Py_None));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/server.pyx":40
 *         self._active_count -= 1
 *         if self._active_count == 0 and self._servers is None:
 *             self._wakeup()             # <<<<<<<<<<<<<<
 * 
 *     cdef _ref(self):
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_v_self->__pyx_vtab)->_wakeup(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uvloop/server.pyx":39
 *         assert self._active_count > 0
 *         self._active_count -= 1
 *         if self._active_count == 0 and self._servers is None:             # <<<<<<<<<<<<<<
 *             self._wakeup()
 * 
 */
  }

  /* "uvloop/server.pyx":36
 *         self._active_count += 1
 * 
 *     cdef _detach(self):             # <<<<<<<<<<<<<<
 *         assert self._active_count > 0
 *         self._active_count -= 1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uvloop.loop.Server._detach", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":42
 *             self._wakeup()
 * 
 *     cdef _ref(self):             # <<<<<<<<<<<<<<
 *         # Keep the server object alive while it's not explicitly closed.
 *         self._loop._servers.add(self)
 */

static PyObject *__pyx_f_6uvloop_4loop_6Server__ref(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ref", 1);

  /* "uvloop/server.pyx":44
 *     cdef _ref(self):
 *         # Keep the server object alive while it's not explicitly closed.
 *         self._loop._servers.add(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef _unref(self):
 */
  if (unlikely(__pyx_v_self->_loop->_servers == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
    __PYX_ERR(27, 44, __pyx_L1_error)
  }
  __pyx_t_1 = PySet_Add(__pyx_v_self->_loop->_servers, ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(27, 44, __pyx_L1_error)

  /* "uvloop/server.pyx":42
 *             self._wakeup()
 * 
 *     cdef _ref(self):             # <<<<<<<<<<<<<<
 *         # Keep the server object alive while it's not explicitly closed.
 *         self._loop._servers.add(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server._ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":46
 *         self._loop._servers.add(self)
 * 
 *     cdef _unref(self):             # <<<<<<<<<<<<<<
 *         self._loop._servers.discard(self)
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop_6Server__unref(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unref", 1);

  /* "uvloop/server.pyx":47
 * 
 *     cdef _unref(self):
 *         self._loop._servers.discard(self)             # <<<<<<<<<<<<<<
 * 
 *     # Public API
 */
  if (unlikely(__pyx_v_self->_loop->_servers == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "discard");
    __PYX_ERR(27, 47, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PySet_Discard(__pyx_v_self->_loop->_servers, ((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(27, 47, __pyx_L1_error)

  /* "uvloop/server.pyx":46
 *         self._loop._servers.add(self)
 * 
 *     cdef _unref(self):             # <<<<<<<<<<<<<<
 *         self._loop._servers.discard(self)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server._unref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_6Server_4generator24(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/server.pyx":51
 *     # Public API
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aenter__(self):
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_3__aenter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_2__aenter__, "Server.__aenter__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_3__aenter__ = {"__aenter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_3__aenter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_2__aenter__};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_3__aenter__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aenter__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__aenter__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__aenter__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_2__aenter__(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_2__aenter__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__aenter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_33___aenter__(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(27, 51, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_6Server_4generator24, __pyx_codeobj__189, (PyObject *) __pyx_cur_scope, __pyx_n_s_aenter, __pyx_n_s_Server___aenter, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(27, 51, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server.__aenter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_6Server_4generator24(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aenter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 51, __pyx_L1_error)

  /* "uvloop/server.pyx":53
 *     @cython.iterable_coroutine
 *     async def __aenter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(((PyObject *)__pyx_cur_scope->__pyx_v_self));
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/server.pyx":51
 *     # Public API
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aenter__(self):
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_AddTraceback("__aenter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_6Server_7generator25(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/server.pyx":55
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aexit__(self, *exc):
 *         self.close()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_6__aexit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_5__aexit__, "Server.__aexit__(self, *exc)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_6__aexit__ = {"__aexit__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_6Server_6__aexit__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_5__aexit__};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_6__aexit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_exc = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aexit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__aexit__", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_exc = __pyx_args;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_5__aexit__(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self), __pyx_v_exc);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_5__aexit__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_exc) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__aexit__", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_34___aexit__(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(27, 55, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_exc = __pyx_v_exc;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_exc);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_exc);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_6Server_7generator25, __pyx_codeobj__190, (PyObject *) __pyx_cur_scope, __pyx_n_s_aexit, __pyx_n_s_Server___aexit, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(27, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server.__aexit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_6Server_7generator25(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__aexit__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 55, __pyx_L1_error)

  /* "uvloop/server.pyx":57
 *     @cython.iterable_coroutine
 *     async def __aexit__(self, *exc):
 *         self.close()             # <<<<<<<<<<<<<<
 *         await self.wait_closed()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uvloop/server.pyx":58
 *     async def __aexit__(self, *exc):
 *         self.close()
 *         await self.wait_closed()             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_wait_closed); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L4_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 58, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(27, 58, __pyx_L1_error)
    }
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/server.pyx":55
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aexit__(self, *exc):
 *         self.close()
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("__aexit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":60
 *         await self.wait_closed()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<%s sockets=%r>' % (self.__class__.__name__, self.sockets)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_9__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_6Server_9__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_8__repr__(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_8__repr__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uvloop/server.pyx":61
 * 
 *     def __repr__(self):
 *         return '<%s sockets=%r>' % (self.__class__.__name__, self.sockets)             # <<<<<<<<<<<<<<
 * 
 *     def get_loop(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u__128);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__128);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__128);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_5), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u_sockets);
  __pyx_t_2 += 9;
  __Pyx_GIVEREF(__pyx_kp_u_sockets);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_sockets);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sockets_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__130);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__130);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__130);
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "uvloop/server.pyx":60
 *         await self.wait_closed()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return '<%s sockets=%r>' % (self.__class__.__name__, self.sockets)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uvloop.loop.Server.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":63
 *         return '<%s sockets=%r>' % (self.__class__.__name__, self.sockets)
 * 
 *     def get_loop(self):             # <<<<<<<<<<<<<<
 *         return self._loop
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_11get_loop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_10get_loop, "Server.get_loop(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_11get_loop = {"get_loop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_11get_loop, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_10get_loop};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_11get_loop(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loop (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_loop", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_loop", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_10get_loop(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_10get_loop(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loop", 1);

  /* "uvloop/server.pyx":64
 * 
 *     def get_loop(self):
 *         return self._loop             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_loop);
  __pyx_r = ((PyObject *)__pyx_v_self->_loop);
  goto __pyx_L0;

  /* "uvloop/server.pyx":63
 *         return '<%s sockets=%r>' % (self.__class__.__name__, self.sockets)
 * 
 *     def get_loop(self):             # <<<<<<<<<<<<<<
 *         return self._loop
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_6Server_14generator26(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/server.pyx":66
 *         return self._loop
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def wait_closed(self):
 *         # Do not remove `self._servers is None` below
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_13wait_closed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_12wait_closed, "Server.wait_closed(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_13wait_closed = {"wait_closed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_13wait_closed, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_12wait_closed};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_13wait_closed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wait_closed (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("wait_closed", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "wait_closed", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_12wait_closed(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_12wait_closed(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wait_closed", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_35_wait_closed(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(27, 66, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_6Server_14generator26, __pyx_codeobj__191, (PyObject *) __pyx_cur_scope, __pyx_n_s_wait_closed, __pyx_n_s_Server_wait_closed, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(27, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server.wait_closed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_6Server_14generator26(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wait_closed", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 66, __pyx_L1_error)

  /* "uvloop/server.pyx":71
 *         # because close() method only closes server sockets
 *         # and existing client connections are left open.
 *         if self._servers is None or self._waiters is None:             # <<<<<<<<<<<<<<
 *             return
 *         waiter = self._loop._new_future()
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_self->_servers == ((PyObject*)Py_None));
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_self->_waiters == ((PyObject*)Py_None));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uvloop/server.pyx":72
 *         # and existing client connections are left open.
 *         if self._servers is None or self._waiters is None:
 *             return             # <<<<<<<<<<<<<<
 *         waiter = self._loop._new_future()
 *         self._waiters.append(waiter)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "uvloop/server.pyx":71
 *         # because close() method only closes server sockets
 *         # and existing client connections are left open.
 *         if self._servers is None or self._waiters is None:             # <<<<<<<<<<<<<<
 *             return
 *         waiter = self._loop._new_future()
 */
  }

  /* "uvloop/server.pyx":73
 *         if self._servers is None or self._waiters is None:
 *             return
 *         waiter = self._loop._new_future()             # <<<<<<<<<<<<<<
 *         self._waiters.append(waiter)
 *         await waiter
 */
  __pyx_t_3 = __pyx_f_6uvloop_4loop_4Loop__new_future(__pyx_cur_scope->__pyx_v_self->_loop); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_waiter = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "uvloop/server.pyx":74
 *             return
 *         waiter = self._loop._new_future()
 *         self._waiters.append(waiter)             # <<<<<<<<<<<<<<
 *         await waiter
 * 
 */
  if (unlikely(__pyx_cur_scope->__pyx_v_self->_waiters == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(27, 74, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_self->_waiters, __pyx_cur_scope->__pyx_v_waiter); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(27, 74, __pyx_L1_error)

  /* "uvloop/server.pyx":75
 *         waiter = self._loop._new_future()
 *         self._waiters.append(waiter)
 *         await waiter             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_waiter);
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 75, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(27, 75, __pyx_L1_error)
    }
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/server.pyx":66
 *         return self._loop
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def wait_closed(self):
 *         # Do not remove `self._servers is None` below
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("wait_closed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":77
 *         await waiter
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         cdef list servers
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_16close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_15close, "Server.close(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_16close = {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_16close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_15close};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_16close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "close", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_15close(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_15close(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_v_servers = 0;
  PyObject *__pyx_v_server = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 1);

  /* "uvloop/server.pyx":80
 *         cdef list servers
 * 
 *         if self._servers is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_servers == ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "uvloop/server.pyx":81
 * 
 *         if self._servers is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/server.pyx":80
 *         cdef list servers
 * 
 *         if self._servers is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "uvloop/server.pyx":83
 *             return
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             servers = self._servers
 *             self._servers = None
 */
  /*try:*/ {

    /* "uvloop/server.pyx":84
 * 
 *         try:
 *             servers = self._servers             # <<<<<<<<<<<<<<
 *             self._servers = None
 *             self._serving = 0
 */
    __pyx_t_2 = __pyx_v_self->_servers;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_servers = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "uvloop/server.pyx":85
 *         try:
 *             servers = self._servers
 *             self._servers = None             # <<<<<<<<<<<<<<
 *             self._serving = 0
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_servers);
    __Pyx_DECREF(__pyx_v_self->_servers);
    __pyx_v_self->_servers = ((PyObject*)Py_None);

    /* "uvloop/server.pyx":86
 *             servers = self._servers
 *             self._servers = None
 *             self._serving = 0             # <<<<<<<<<<<<<<
 * 
 *             for server in servers:
 */
    __pyx_v_self->_serving = 0;

    /* "uvloop/server.pyx":88
 *             self._serving = 0
 * 
 *             for server in servers:             # <<<<<<<<<<<<<<
 *                 (<UVStreamServer>server)._close()
 * 
 */
    if (unlikely(__pyx_v_servers == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(27, 88, __pyx_L5_error)
    }
    __pyx_t_2 = __pyx_v_servers; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(27, 88, __pyx_L5_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(27, 88, __pyx_L5_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 88, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_server, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/server.pyx":89
 * 
 *             for server in servers:
 *                 (<UVStreamServer>server)._close()             # <<<<<<<<<<<<<<
 * 
 *             if self._active_count == 0:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_server)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base._close(((struct __pyx_obj_6uvloop_4loop_UVHandle *)((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_server))); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 89, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/server.pyx":88
 *             self._serving = 0
 * 
 *             for server in servers:             # <<<<<<<<<<<<<<
 *                 (<UVStreamServer>server)._close()
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/server.pyx":91
 *                 (<UVStreamServer>server)._close()
 * 
 *             if self._active_count == 0:             # <<<<<<<<<<<<<<
 *                 self._wakeup()
 *         finally:
 */
    __pyx_t_1 = (__pyx_v_self->_active_count == 0);
    if (__pyx_t_1) {

      /* "uvloop/server.pyx":92
 * 
 *             if self._active_count == 0:
 *                 self._wakeup()             # <<<<<<<<<<<<<<
 *         finally:
 *             self._unref()
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_v_self->__pyx_vtab)->_wakeup(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 92, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uvloop/server.pyx":91
 *                 (<UVStreamServer>server)._close()
 * 
 *             if self._active_count == 0:             # <<<<<<<<<<<<<<
 *                 self._wakeup()
 *         finally:
 */
    }
  }

  /* "uvloop/server.pyx":94
 *                 self._wakeup()
 *         finally:
 *             self._unref()             # <<<<<<<<<<<<<<
 * 
 *     def is_serving(self):
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_v_self->__pyx_vtab)->_unref(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 94, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_v_self->__pyx_vtab)->_unref(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 94, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
      __pyx_L12_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uvloop/server.pyx":77
 *         await waiter
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         cdef list servers
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Server.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_servers);
  __Pyx_XDECREF(__pyx_v_server);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":96
 *             self._unref()
 * 
 *     def is_serving(self):             # <<<<<<<<<<<<<<
 *         return self._serving
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_18is_serving(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_17is_serving, "Server.is_serving(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_18is_serving = {"is_serving", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_18is_serving, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_17is_serving};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_18is_serving(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_serving (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_serving", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_serving", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_17is_serving(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_17is_serving(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_serving", 1);

  /* "uvloop/server.pyx":97
 * 
 *     def is_serving(self):
 *         return self._serving             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_serving); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uvloop/server.pyx":96
 *             self._unref()
 * 
 *     def is_serving(self):             # <<<<<<<<<<<<<<
 *         return self._serving
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uvloop.loop.Server.is_serving", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_6Server_21generator27(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/server.pyx":99
 *         return self._serving
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_serving(self):
 *         self._start_serving()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_20start_serving(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_19start_serving, "Server.start_serving(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_20start_serving = {"start_serving", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_20start_serving, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_19start_serving};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_20start_serving(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_serving (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("start_serving", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "start_serving", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_19start_serving(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_19start_serving(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start_serving", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_36_start_serving(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(27, 99, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_6Server_21generator27, __pyx_codeobj__192, (PyObject *) __pyx_cur_scope, __pyx_n_s_start_serving, __pyx_n_s_Server_start_serving, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(27, 99, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server.start_serving", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_6Server_21generator27(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_serving", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 99, __pyx_L1_error)

  /* "uvloop/server.pyx":101
 *     @cython.iterable_coroutine
 *     async def start_serving(self):
 *         self._start_serving()             # <<<<<<<<<<<<<<
 * 
 *     @cython.iterable_coroutine
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_start_serving(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/server.pyx":99
 *         return self._serving
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_serving(self):
 *         self._start_serving()
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("start_serving", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_6Server_24generator28(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/server.pyx":103
 *         self._start_serving()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_23serve_forever(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_22serve_forever, "Server.serve_forever(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_23serve_forever = {"serve_forever", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_23serve_forever, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_22serve_forever};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_23serve_forever(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("serve_forever (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("serve_forever", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "serve_forever", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_22serve_forever(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_22serve_forever(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("serve_forever", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_37_serve_forever(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(27, 103, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_6Server_24generator28, __pyx_codeobj__193, (PyObject *) __pyx_cur_scope, __pyx_n_s_serve_forever, __pyx_n_s_Server_serve_forever, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(27, 103, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server.serve_forever", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_6Server_24generator28(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *__pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  char const *__pyx_t_22;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("serve_forever", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L15_resume_from_await;
    case 2: goto __pyx_L23_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 103, __pyx_L1_error)

  /* "uvloop/server.pyx":105
 *     @cython.iterable_coroutine
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f'server {self!r} is already being awaited on serve_forever()')
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_self->_serving_forever_fut != Py_None);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/server.pyx":107
 *         if self._serving_forever_fut is not None:
 *             raise RuntimeError(
 *                 f'server {self!r} is already being awaited on serve_forever()')             # <<<<<<<<<<<<<<
 *         if self._servers is None:
 *             raise RuntimeError(f'server {self!r} is closed')
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_server);
    __pyx_t_3 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_server);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_server);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_cur_scope->__pyx_v_self)), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_already_being_awaited_on_ser);
    __pyx_t_3 += 44;
    __Pyx_GIVEREF(__pyx_kp_u_is_already_being_awaited_on_ser);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_is_already_being_awaited_on_ser);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/server.pyx":106
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 f'server {self!r} is already being awaited on serve_forever()')
 *         if self._servers is None:
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(27, 106, __pyx_L1_error)

    /* "uvloop/server.pyx":105
 *     @cython.iterable_coroutine
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(
 *                 f'server {self!r} is already being awaited on serve_forever()')
 */
  }

  /* "uvloop/server.pyx":108
 *             raise RuntimeError(
 *                 f'server {self!r} is already being awaited on serve_forever()')
 *         if self._servers is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(f'server {self!r} is closed')
 * 
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_self->_servers == ((PyObject*)Py_None));
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/server.pyx":109
 *                 f'server {self!r} is already being awaited on serve_forever()')
 *         if self._servers is None:
 *             raise RuntimeError(f'server {self!r} is closed')             # <<<<<<<<<<<<<<
 * 
 *         self._start_serving()
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_server);
    __pyx_t_3 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_server);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_server);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(((PyObject *)__pyx_cur_scope->__pyx_v_self)), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_closed_2);
    __pyx_t_3 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_is_closed_2);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_is_closed_2);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(27, 109, __pyx_L1_error)

    /* "uvloop/server.pyx":108
 *             raise RuntimeError(
 *                 f'server {self!r} is already being awaited on serve_forever()')
 *         if self._servers is None:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(f'server {self!r} is closed')
 * 
 */
  }

  /* "uvloop/server.pyx":111
 *             raise RuntimeError(f'server {self!r} is closed')
 * 
 *         self._start_serving()             # <<<<<<<<<<<<<<
 *         self._serving_forever_fut = self._loop.create_future()
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6uvloop_4loop_Server *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_start_serving(__pyx_cur_scope->__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/server.pyx":112
 * 
 *         self._start_serving()
 *         self._serving_forever_fut = self._loop.create_future()             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self->_loop), __pyx_n_s_create_future); if (unlikely(!__pyx_t_5)) __PYX_ERR(27, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(27, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_serving_forever_fut);
  __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_serving_forever_fut);
  __pyx_cur_scope->__pyx_v_self->_serving_forever_fut = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uvloop/server.pyx":114
 *         self._serving_forever_fut = self._loop.create_future()
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await self._serving_forever_fut
 *         except asyncio.CancelledError:
 */
  /*try:*/ {
    {
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      /*try:*/ {

        /* "uvloop/server.pyx":115
 * 
 *         try:
 *             await self._serving_forever_fut             # <<<<<<<<<<<<<<
 *         except asyncio.CancelledError:
 *             try:
 */
        __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_cur_scope->__pyx_v_self->_serving_forever_fut);
        __Pyx_XGOTREF(__pyx_r);
        if (likely(__pyx_r)) {
          __Pyx_XGIVEREF(__pyx_t_8);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_8;
          __Pyx_XGIVEREF(__pyx_t_9);
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_9;
          __Pyx_XGIVEREF(__pyx_t_10);
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_10;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, awaiting value */
          __pyx_generator->resume_label = 1;
          return __pyx_r;
          __pyx_L15_resume_from_await:;
          __pyx_t_8 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_8);
          __pyx_t_9 = __pyx_cur_scope->__pyx_t_1;
          __pyx_cur_scope->__pyx_t_1 = 0;
          __Pyx_XGOTREF(__pyx_t_9);
          __pyx_t_10 = __pyx_cur_scope->__pyx_t_2;
          __pyx_cur_scope->__pyx_t_2 = 0;
          __Pyx_XGOTREF(__pyx_t_10);
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 115, __pyx_L9_error)
        } else {
          PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
            else __PYX_ERR(27, 115, __pyx_L9_error)
          }
        }

        /* "uvloop/server.pyx":114
 *         self._serving_forever_fut = self._loop.create_future()
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await self._serving_forever_fut
 *         except asyncio.CancelledError:
 */
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L14_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uvloop/server.pyx":116
 *         try:
 *             await self._serving_forever_fut
 *         except asyncio.CancelledError:             # <<<<<<<<<<<<<<
 *             try:
 *                 self.close()
 */
      __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_5, &__pyx_t_6);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_11)) __PYX_ERR(27, 116, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_CancelledError); if (unlikely(!__pyx_t_12)) __PYX_ERR(27, 116, __pyx_L11_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_12);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_5, __pyx_t_6);
      __pyx_t_2 = 0; __pyx_t_5 = 0; __pyx_t_6 = 0;
      if (__pyx_t_13) {
        __Pyx_AddTraceback("uvloop.loop.Server.serve_forever", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_5, &__pyx_t_2) < 0) __PYX_ERR(27, 116, __pyx_L11_except_error)
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_2);

        /* "uvloop/server.pyx":117
 *             await self._serving_forever_fut
 *         except asyncio.CancelledError:
 *             try:             # <<<<<<<<<<<<<<
 *                 self.close()
 *                 await self.wait_closed()
 */
        /*try:*/ {

          /* "uvloop/server.pyx":118
 *         except asyncio.CancelledError:
 *             try:
 *                 self.close()             # <<<<<<<<<<<<<<
 *                 await self.wait_closed()
 *             finally:
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_close); if (unlikely(!__pyx_t_11)) __PYX_ERR(27, 118, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_14 = NULL;
          __pyx_t_7 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_7 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(27, 118, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "uvloop/server.pyx":119
 *             try:
 *                 self.close()
 *                 await self.wait_closed()             # <<<<<<<<<<<<<<
 *             finally:
 *                 raise
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_wait_closed); if (unlikely(!__pyx_t_11)) __PYX_ERR(27, 119, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_14 = NULL;
          __pyx_t_7 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_7 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(27, 119, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XGOTREF(__pyx_r);
          if (likely(__pyx_r)) {
            __Pyx_XGIVEREF(__pyx_t_2);
            __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
            __Pyx_XGIVEREF(__pyx_t_5);
            __pyx_cur_scope->__pyx_t_1 = __pyx_t_5;
            __Pyx_XGIVEREF(__pyx_t_6);
            __pyx_cur_scope->__pyx_t_2 = __pyx_t_6;
            __Pyx_XGIVEREF(__pyx_t_8);
            __pyx_cur_scope->__pyx_t_3 = __pyx_t_8;
            __Pyx_XGIVEREF(__pyx_t_9);
            __pyx_cur_scope->__pyx_t_4 = __pyx_t_9;
            __Pyx_XGIVEREF(__pyx_t_10);
            __pyx_cur_scope->__pyx_t_5 = __pyx_t_10;
            __Pyx_XGIVEREF(__pyx_r);
            __Pyx_RefNannyFinishContext();
            __Pyx_Coroutine_SwapException(__pyx_generator);
            /* return from generator, awaiting value */
            __pyx_generator->resume_label = 2;
            return __pyx_r;
            __pyx_L23_resume_from_await:;
            __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
            __pyx_cur_scope->__pyx_t_0 = 0;
            __Pyx_XGOTREF(__pyx_t_2);
            __pyx_t_5 = __pyx_cur_scope->__pyx_t_1;
            __pyx_cur_scope->__pyx_t_1 = 0;
            __Pyx_XGOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_cur_scope->__pyx_t_2;
            __pyx_cur_scope->__pyx_t_2 = 0;
            __Pyx_XGOTREF(__pyx_t_6);
            __pyx_t_8 = __pyx_cur_scope->__pyx_t_3;
            __pyx_cur_scope->__pyx_t_3 = 0;
            __Pyx_XGOTREF(__pyx_t_8);
            __pyx_t_9 = __pyx_cur_scope->__pyx_t_4;
            __pyx_cur_scope->__pyx_t_4 = 0;
            __Pyx_XGOTREF(__pyx_t_9);
            __pyx_t_10 = __pyx_cur_scope->__pyx_t_5;
            __pyx_cur_scope->__pyx_t_5 = 0;
            __Pyx_XGOTREF(__pyx_t_10);
            if (unlikely(!__pyx_sent_value)) __PYX_ERR(27, 119, __pyx_L21_error)
          } else {
            PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
              else __PYX_ERR(27, 119, __pyx_L21_error)
            }
          }
        }

        /* "uvloop/server.pyx":121
 *                 await self.wait_closed()
 *             finally:
 *                 raise             # <<<<<<<<<<<<<<
 *         finally:
 *             self._serving_forever_fut = None
 */
        /*finally:*/ {
          /*normal exit:*/{
            __Pyx_GIVEREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_XGIVEREF(__pyx_t_2);
            __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_5, __pyx_t_2);
            __pyx_t_6 = 0; __pyx_t_5 = 0; __pyx_t_2 = 0; 
            __PYX_ERR(27, 121, __pyx_L11_except_error)
          }
          __pyx_L21_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_assign
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            {
              __Pyx_GIVEREF(__pyx_t_15);
              __Pyx_GIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_ErrRestoreWithState(__pyx_t_15, __pyx_t_16, __pyx_t_17);
              __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; 
              __PYX_ERR(27, 121, __pyx_L27_error)
            }
            __pyx_L27_error:;
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            goto __pyx_L11_except_error;
          }
        }
      }
      goto __pyx_L11_except_error;

      /* "uvloop/server.pyx":114
 *         self._serving_forever_fut = self._loop.create_future()
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             await self._serving_forever_fut
 *         except asyncio.CancelledError:
 */
      __pyx_L11_except_error:;
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      goto __pyx_L7_error;
      __pyx_L14_try_end:;
    }
  }

  /* "uvloop/server.pyx":123
 *                 raise
 *         finally:
 *             self._serving_forever_fut = None             # <<<<<<<<<<<<<<
 * 
 *     property sockets:
 */
  /*finally:*/ {
    /*normal exit:*/{
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_serving_forever_fut);
      __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_serving_forever_fut);
      __pyx_cur_scope->__pyx_v_self->_serving_forever_fut = Py_None;
      goto __pyx_L8;
    }
    __pyx_L7_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_assign
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_13 = __pyx_lineno; __pyx_t_21 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->_serving_forever_fut);
        __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->_serving_forever_fut);
        __pyx_cur_scope->__pyx_v_self->_serving_forever_fut = Py_None;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_21; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
    }
    __pyx_L8:;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "uvloop/server.pyx":103
 *         self._start_serving()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("serve_forever", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/server.pyx":126
 * 
 *     property sockets:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef list sockets = []
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_7sockets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6uvloop_4loop_6Server_7sockets_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_7sockets___get__(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_7sockets___get__(struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_v_sockets = 0;
  PyObject *__pyx_v_server = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uvloop/server.pyx":127
 *     property sockets:
 *         def __get__(self):
 *             cdef list sockets = []             # <<<<<<<<<<<<<<
 * 
 *             # Guard against `self._servers is None`
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(27, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sockets = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uvloop/server.pyx":130
 * 
 *             # Guard against `self._servers is None`
 *             if self._servers:             # <<<<<<<<<<<<<<
 *                 for server in self._servers:
 *                     sockets.append(
 */
  __pyx_t_2 = (__pyx_v_self->_servers != Py_None)&&(PyList_GET_SIZE(__pyx_v_self->_servers) != 0);
  if (__pyx_t_2) {

    /* "uvloop/server.pyx":131
 *             # Guard against `self._servers is None`
 *             if self._servers:
 *                 for server in self._servers:             # <<<<<<<<<<<<<<
 *                     sockets.append(
 *                         (<UVStreamServer>server)._get_socket()
 */
    if (unlikely(__pyx_v_self->_servers == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(27, 131, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->_servers; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(27, 131, __pyx_L1_error)
        #endif
        if (__pyx_t_3 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(27, 131, __pyx_L1_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_server, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "uvloop/server.pyx":133
 *                 for server in self._servers:
 *                     sockets.append(
 *                         (<UVStreamServer>server)._get_socket()             # <<<<<<<<<<<<<<
 *                     )
 * 
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer *)((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_server)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._get_socket(((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)__pyx_v_server))); if (unlikely(!__pyx_t_4)) __PYX_ERR(27, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "uvloop/server.pyx":132
 *             if self._servers:
 *                 for server in self._servers:
 *                     sockets.append(             # <<<<<<<<<<<<<<
 *                         (<UVStreamServer>server)._get_socket()
 *                     )
 */
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_sockets, __pyx_t_4); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(27, 132, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uvloop/server.pyx":131
 *             # Guard against `self._servers is None`
 *             if self._servers:
 *                 for server in self._servers:             # <<<<<<<<<<<<<<
 *                     sockets.append(
 *                         (<UVStreamServer>server)._get_socket()
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uvloop/server.pyx":130
 * 
 *             # Guard against `self._servers is None`
 *             if self._servers:             # <<<<<<<<<<<<<<
 *                 for server in self._servers:
 *                     sockets.append(
 */
  }

  /* "uvloop/server.pyx":136
 *                     )
 * 
 *             return sockets             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_sockets);
  __pyx_r = __pyx_v_sockets;
  goto __pyx_L0;

  /* "uvloop/server.pyx":126
 * 
 *     property sockets:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef list sockets = []
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.Server.sockets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sockets);
  __Pyx_XDECREF(__pyx_v_server);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_26__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_25__reduce_cython__, "Server.__reduce_cython__(self)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_26__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_26__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_25__reduce_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_26__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_25__reduce_cython__(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_25__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_6Server_28__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6Server_27__setstate_cython__, "Server.__setstate_cython__(self, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_6Server_28__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_28__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_27__setstate_cython__};
static PyObject *__pyx_pw_6uvloop_4loop_6Server_28__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(6, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(6, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.Server.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_6Server_27__setstate_cython__(((struct __pyx_obj_6uvloop_4loop_Server *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6Server_27__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6uvloop_4loop_Server *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(6, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop.Server.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3374
 * 
 * 
 * cdef void __get_fork_handler() noexcept nogil:             # <<<<<<<<<<<<<<
 *     with gil:
 *         if (__forking and __forking_loop is not None and
 */

static void __pyx_f_6uvloop_4loop___get_fork_handler(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__get_fork_handler", 1);

  /* "uvloop/loop.pyx":3375
 * 
 * cdef void __get_fork_handler() noexcept nogil:
 *     with gil:             # <<<<<<<<<<<<<<
 *         if (__forking and __forking_loop is not None and
 *                 __forking_loop.active_process_handler is not None):
 */
  {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      /*try:*/ {

        /* "uvloop/loop.pyx":3376
 * cdef void __get_fork_handler() noexcept nogil:
 *     with gil:
 *         if (__forking and __forking_loop is not None and             # <<<<<<<<<<<<<<
 *                 __forking_loop.active_process_handler is not None):
 *             __forking_loop.active_process_handler._after_fork()
 */
        __pyx_t_2 = (__pyx_v_6uvloop_4loop___forking != 0);
        if (__pyx_t_2) {
        } else {
          __pyx_t_1 = __pyx_t_2;
          goto __pyx_L7_bool_binop_done;
        }
        __pyx_t_2 = (((PyObject *)__pyx_v_6uvloop_4loop___forking_loop) != Py_None);
        if (__pyx_t_2) {
        } else {
          __pyx_t_1 = __pyx_t_2;
          goto __pyx_L7_bool_binop_done;
        }

        /* "uvloop/loop.pyx":3377
 *     with gil:
 *         if (__forking and __forking_loop is not None and
 *                 __forking_loop.active_process_handler is not None):             # <<<<<<<<<<<<<<
 *             __forking_loop.active_process_handler._after_fork()
 * 
 */
        __pyx_t_2 = (((PyObject *)__pyx_v_6uvloop_4loop___forking_loop->active_process_handler) != Py_None);
        __pyx_t_1 = __pyx_t_2;
        __pyx_L7_bool_binop_done:;

        /* "uvloop/loop.pyx":3376
 * cdef void __get_fork_handler() noexcept nogil:
 *     with gil:
 *         if (__forking and __forking_loop is not None and             # <<<<<<<<<<<<<<
 *                 __forking_loop.active_process_handler is not None):
 *             __forking_loop.active_process_handler._after_fork()
 */
        if (__pyx_t_1) {

          /* "uvloop/loop.pyx":3378
 *         if (__forking and __forking_loop is not None and
 *                 __forking_loop.active_process_handler is not None):
 *             __forking_loop.active_process_handler._after_fork()             # <<<<<<<<<<<<<<
 * 
 * cdef __install_atfork():
 */
          __pyx_t_3 = ((struct __pyx_vtabstruct_6uvloop_4loop_UVProcess *)__pyx_v_6uvloop_4loop___forking_loop->active_process_handler->__pyx_base.__pyx_vtab)->_after_fork(__pyx_v_6uvloop_4loop___forking_loop->active_process_handler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3378, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "uvloop/loop.pyx":3376
 * cdef void __get_fork_handler() noexcept nogil:
 *     with gil:
 *         if (__forking and __forking_loop is not None and             # <<<<<<<<<<<<<<
 *                 __forking_loop.active_process_handler is not None):
 *             __forking_loop.active_process_handler._after_fork()
 */
        }
      }

      /* "uvloop/loop.pyx":3375
 * 
 * cdef void __get_fork_handler() noexcept nogil:
 *     with gil:             # <<<<<<<<<<<<<<
 *         if (__forking and __forking_loop is not None and
 *                 __forking_loop.active_process_handler is not None):
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_PyGILState_Release(__pyx_gilstate_save);
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_PyGILState_Release(__pyx_gilstate_save);
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "uvloop/loop.pyx":3374
 * 
 * 
 * cdef void __get_fork_handler() noexcept nogil:             # <<<<<<<<<<<<<<
 *     with gil:
 *         if (__forking and __forking_loop is not None and
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uvloop.loop.__get_fork_handler", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __pyx_L0:;
  __Pyx_RefNannyFinishContextNogil()
}

/* "uvloop/loop.pyx":3380
 *             __forking_loop.active_process_handler._after_fork()
 * 
 * cdef __install_atfork():             # <<<<<<<<<<<<<<
 *     global __atfork_installed
 * 
 */

static PyObject *__pyx_f_6uvloop_4loop___install_atfork(void) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__install_atfork", 1);

  /* "uvloop/loop.pyx":3383
 *     global __atfork_installed
 * 
 *     if __atfork_installed:             # <<<<<<<<<<<<<<
 *         return
 *     __atfork_installed = 1
 */
  __pyx_t_1 = (__pyx_v_6uvloop_4loop___atfork_installed != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3384
 * 
 *     if __atfork_installed:
 *         return             # <<<<<<<<<<<<<<
 *     __atfork_installed = 1
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":3383
 *     global __atfork_installed
 * 
 *     if __atfork_installed:             # <<<<<<<<<<<<<<
 *         return
 *     __atfork_installed = 1
 */
  }

  /* "uvloop/loop.pyx":3385
 *     if __atfork_installed:
 *         return
 *     __atfork_installed = 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int err
 */
  __pyx_v_6uvloop_4loop___atfork_installed = 1;

  /* "uvloop/loop.pyx":3389
 *     cdef int err
 * 
 *     err = system.pthread_atfork(NULL, NULL, &system.handleAtFork)             # <<<<<<<<<<<<<<
 *     if err:
 *         __atfork_installed = 0
 */
  __pyx_v_err = pthread_atfork(NULL, NULL, (&handleAtFork));

  /* "uvloop/loop.pyx":3390
 * 
 *     err = system.pthread_atfork(NULL, NULL, &system.handleAtFork)
 *     if err:             # <<<<<<<<<<<<<<
 *         __atfork_installed = 0
 *         raise convert_error(-err)
 */
  __pyx_t_1 = (__pyx_v_err != 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":3391
 *     err = system.pthread_atfork(NULL, NULL, &system.handleAtFork)
 *     if err:
 *         __atfork_installed = 0             # <<<<<<<<<<<<<<
 *         raise convert_error(-err)
 * 
 */
    __pyx_v_6uvloop_4loop___atfork_installed = 0;

    /* "uvloop/loop.pyx":3392
 *     if err:
 *         __atfork_installed = 0
 *         raise convert_error(-err)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error((-__pyx_v_err)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3392, __pyx_L1_error)

    /* "uvloop/loop.pyx":3390
 * 
 *     err = system.pthread_atfork(NULL, NULL, &system.handleAtFork)
 *     if err:             # <<<<<<<<<<<<<<
 *         __atfork_installed = 0
 *         raise convert_error(-err)
 */
  }

  /* "uvloop/loop.pyx":3380
 *             __forking_loop.active_process_handler._after_fork()
 * 
 * cdef __install_atfork():             # <<<<<<<<<<<<<<
 *     global __atfork_installed
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.__install_atfork", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3397
 * # Install PyMem* memory allocators
 * cdef vint __mem_installed = 0
 * cdef __install_pymem():             # <<<<<<<<<<<<<<
 *     global __mem_installed
 *     if __mem_installed:
 */

static PyObject *__pyx_f_6uvloop_4loop___install_pymem(void) {
  int __pyx_v_err;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__install_pymem", 1);

  /* "uvloop/loop.pyx":3399
 * cdef __install_pymem():
 *     global __mem_installed
 *     if __mem_installed:             # <<<<<<<<<<<<<<
 *         return
 *     __mem_installed = 1
 */
  __pyx_t_1 = (__pyx_v_6uvloop_4loop___mem_installed != 0);
  if (__pyx_t_1) {

    /* "uvloop/loop.pyx":3400
 *     global __mem_installed
 *     if __mem_installed:
 *         return             # <<<<<<<<<<<<<<
 *     __mem_installed = 1
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uvloop/loop.pyx":3399
 * cdef __install_pymem():
 *     global __mem_installed
 *     if __mem_installed:             # <<<<<<<<<<<<<<
 *         return
 *     __mem_installed = 1
 */
  }

  /* "uvloop/loop.pyx":3401
 *     if __mem_installed:
 *         return
 *     __mem_installed = 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int err
 */
  __pyx_v_6uvloop_4loop___mem_installed = 1;

  /* "uvloop/loop.pyx":3404
 * 
 *     cdef int err
 *     err = uv.uv_replace_allocator(<uv.uv_malloc_func>PyMem_RawMalloc,             # <<<<<<<<<<<<<<
 *                                   <uv.uv_realloc_func>PyMem_RawRealloc,
 *                                   <uv.uv_calloc_func>PyMem_RawCalloc,
 */
  __pyx_v_err = uv_replace_allocator(((uv_malloc_func)PyMem_RawMalloc), ((uv_realloc_func)PyMem_RawRealloc), ((uv_calloc_func)PyMem_RawCalloc), ((uv_free_func)PyMem_RawFree));

  /* "uvloop/loop.pyx":3408
 *                                   <uv.uv_calloc_func>PyMem_RawCalloc,
 *                                   <uv.uv_free_func>PyMem_RawFree)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         __mem_installed = 0
 *         raise convert_error(err)
 */
  __pyx_t_1 = (__pyx_v_err < 0);
  if (unlikely(__pyx_t_1)) {

    /* "uvloop/loop.pyx":3409
 *                                   <uv.uv_free_func>PyMem_RawFree)
 *     if err < 0:
 *         __mem_installed = 0             # <<<<<<<<<<<<<<
 *         raise convert_error(err)
 * 
 */
    __pyx_v_6uvloop_4loop___mem_installed = 0;

    /* "uvloop/loop.pyx":3410
 *     if err < 0:
 *         __mem_installed = 0
 *         raise convert_error(err)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __pyx_f_6uvloop_4loop_convert_error(__pyx_v_err); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 3410, __pyx_L1_error)

    /* "uvloop/loop.pyx":3408
 *                                   <uv.uv_calloc_func>PyMem_RawCalloc,
 *                                   <uv.uv_free_func>PyMem_RawFree)
 *     if err < 0:             # <<<<<<<<<<<<<<
 *         __mem_installed = 0
 *         raise convert_error(err)
 */
  }

  /* "uvloop/loop.pyx":3397
 * # Install PyMem* memory allocators
 * cdef vint __mem_installed = 0
 * cdef __install_pymem():             # <<<<<<<<<<<<<<
 *     global __mem_installed
 *     if __mem_installed:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uvloop.loop.__install_pymem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uvloop/loop.pyx":3413
 * 
 * 
 * cdef _set_signal_wakeup_fd(fd):             # <<<<<<<<<<<<<<
 *     if fd >= 0:
 *         return signal_set_wakeup_fd(fd, warn_on_full_buffer=False)
 */

static PyObject *__pyx_f_6uvloop_4loop__set_signal_wakeup_fd(PyObject *__pyx_v_fd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_signal_wakeup_fd", 1);

  /* "uvloop/loop.pyx":3414
 * 
 * cdef _set_signal_wakeup_fd(fd):
 *     if fd >= 0:             # <<<<<<<<<<<<<<
 *         return signal_set_wakeup_fd(fd, warn_on_full_buffer=False)
 *     else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_fd, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3414, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 3414, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uvloop/loop.pyx":3415
 * cdef _set_signal_wakeup_fd(fd):
 *     if fd >= 0:
 *         return signal_set_wakeup_fd(fd, warn_on_full_buffer=False)             # <<<<<<<<<<<<<<
 *     else:
 *         return signal_set_wakeup_fd(fd)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_fd);
    __Pyx_GIVEREF(__pyx_v_fd);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_fd)) __PYX_ERR(0, 3415, __pyx_L1_error);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_warn_on_full_buffer, Py_False) < 0) __PYX_ERR(0, 3415, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_6uvloop_4loop_signal_set_wakeup_fd, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "uvloop/loop.pyx":3414
 * 
 * cdef _set_signal_wakeup_fd(fd):
 *     if fd >= 0:             # <<<<<<<<<<<<<<
 *         return signal_set_wakeup_fd(fd, warn_on_full_buffer=False)
 *     else:
 */
  }

  /* "uvloop/loop.pyx":3417
 *         return signal_set_wakeup_fd(fd, warn_on_full_buffer=False)
 *     else:
 *         return signal_set_wakeup_fd(fd)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6uvloop_4loop_signal_set_wakeup_fd);
    __pyx_t_3 = __pyx_v_6uvloop_4loop_signal_set_wakeup_fd; __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_fd};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "uvloop/loop.pyx":3413
 * 
 * 
 * cdef _set_signal_wakeup_fd(fd):             # <<<<<<<<<<<<<<
 *     if fd >= 0:
 *         return signal_set_wakeup_fd(fd, warn_on_full_buffer=False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop._set_signal_wakeup_fd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6uvloop_4loop_8generator29(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uvloop/loop.pyx":3422
 * # Helpers for tests
 * 
 * @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 * async def _test_coroutine_1():
 *     return 42
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_7_test_coroutine_1(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_6_test_coroutine_1, "_test_coroutine_1()");
static PyMethodDef __pyx_mdef_6uvloop_4loop_7_test_coroutine_1 = {"_test_coroutine_1", (PyCFunction)__pyx_pw_6uvloop_4loop_7_test_coroutine_1, METH_NOARGS, __pyx_doc_6uvloop_4loop_6_test_coroutine_1};
static PyObject *__pyx_pw_6uvloop_4loop_7_test_coroutine_1(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_coroutine_1 (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_6uvloop_4loop_6_test_coroutine_1(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_6_test_coroutine_1(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_coroutine_1", 0);
  __pyx_cur_scope = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 3422, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  {
    __pyx_CoroutineObject *gen = __Pyx_IterableCoroutine_New((__pyx_coroutine_body_t) __pyx_gb_6uvloop_4loop_8generator29, __pyx_codeobj__194, (PyObject *) __pyx_cur_scope, __pyx_n_s_test_coroutine_1, __pyx_n_s_test_coroutine_1, __pyx_n_s_uvloop_loop); if (unlikely(!gen)) __PYX_ERR(0, 3422, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uvloop.loop._test_coroutine_1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6uvloop_4loop_8generator29(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_coroutine_1", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 3422, __pyx_L1_error)

  /* "uvloop/loop.pyx":3424
 * @cython.iterable_coroutine
 * async def _test_coroutine_1():
 *     return 42             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_int_42);
  goto __pyx_L0;

  /* "uvloop/loop.pyx":3422
 * # Helpers for tests
 * 
 * @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 * async def _test_coroutine_1():
 *     return 42
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_AddTraceback("_test_coroutine_1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PseudoSocket(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_10__pyx_unpickle_PseudoSocket(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_9__pyx_unpickle_PseudoSocket, "__pyx_unpickle_PseudoSocket(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_10__pyx_unpickle_PseudoSocket = {"__pyx_unpickle_PseudoSocket", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_10__pyx_unpickle_PseudoSocket, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9__pyx_unpickle_PseudoSocket};
static PyObject *__pyx_pw_6uvloop_4loop_10__pyx_unpickle_PseudoSocket(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PseudoSocket (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PseudoSocket", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PseudoSocket", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PseudoSocket") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PseudoSocket", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.__pyx_unpickle_PseudoSocket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_9__pyx_unpickle_PseudoSocket(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_9__pyx_unpickle_PseudoSocket(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PseudoSocket", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xbfaab04, 0xa551a14, 0x88c0b03):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__195, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xbfaab04, 0xa551a14, 0x88c0b03):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum
 *     __pyx_result = PseudoSocket.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(6, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xbfaab04, 0xa551a14, 0x88c0b03):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PseudoSocket.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xbfaab04, 0xa551a14, 0x88c0b03):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum
 *     __pyx_result = PseudoSocket.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum
 *     __pyx_result = PseudoSocket.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PseudoSocket.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_6uvloop_4loop___pyx_unpickle_PseudoSocket__set_state(((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum
 *     __pyx_result = PseudoSocket.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PseudoSocket(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.__pyx_unpickle_PseudoSocket", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6uvloop_4loop___pyx_unpickle_PseudoSocket__set_state(struct __pyx_obj_6uvloop_4loop_PseudoSocket *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PseudoSocket__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_family = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_fd = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_peername);
  __Pyx_DECREF(__pyx_v___pyx_result->_peername);
  __pyx_v___pyx_result->_peername = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_proto = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_sockname);
  __Pyx_DECREF(__pyx_v___pyx_result->_sockname);
  __pyx_v___pyx_result->_sockname = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_type = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 6);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.__pyx_unpickle_PseudoSocket__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_LruCache(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6uvloop_4loop_12__pyx_unpickle_LruCache(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_6uvloop_4loop_11__pyx_unpickle_LruCache, "__pyx_unpickle_LruCache(__pyx_type, long __pyx_checksum, __pyx_state)");
static PyMethodDef __pyx_mdef_6uvloop_4loop_12__pyx_unpickle_LruCache = {"__pyx_unpickle_LruCache", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12__pyx_unpickle_LruCache, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11__pyx_unpickle_LruCache};
static PyObject *__pyx_pw_6uvloop_4loop_12__pyx_unpickle_LruCache(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_LruCache (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LruCache", 1, 3, 3, 1); __PYX_ERR(6, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LruCache", 1, 3, 3, 2); __PYX_ERR(6, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_LruCache") < 0)) __PYX_ERR(6, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(6, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_LruCache", 1, 3, 3, __pyx_nargs); __PYX_ERR(6, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uvloop.loop.__pyx_unpickle_LruCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6uvloop_4loop_11__pyx_unpickle_LruCache(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6uvloop_4loop_11__pyx_unpickle_LruCache(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LruCache", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0e1f4d2, 0x737459e, 0x53368d7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__197, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0e1f4d2, 0x737459e, 0x53368d7):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))" % __pyx_checksum
 *     __pyx_result = LruCache.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(6, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x0e1f4d2, 0x737459e, 0x53368d7):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = LruCache.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(6, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x0e1f4d2, 0x737459e, 0x53368d7):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))" % __pyx_checksum
 *     __pyx_result = LruCache.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LruCache__set_state(<LruCache> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6uvloop_4loop_LruCache), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))" % __pyx_checksum
 *     __pyx_result = LruCache.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LruCache__set_state(<LruCache> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = LruCache.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LruCache__set_state(<LruCache> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_LruCache__set_state(LruCache __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(6, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_6uvloop_4loop___pyx_unpickle_LruCache__set_state(((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x0e1f4d2, 0x737459e, 0x53368d7) = (_dict, _dict_get, _dict_move_to_end, _maxsize))" % __pyx_checksum
 *     __pyx_result = LruCache.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_LruCache__set_state(<LruCache> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_LruCache__set_state(<LruCache> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_LruCache__set_state(LruCache __pyx_result, tuple __pyx_state):
 *     __pyx_result._dict = __pyx_state[0]; __pyx_result._dict_get = __pyx_state[1]; __pyx_result._dict_move_to_end = __pyx_state[2]; __pyx_result._maxsize = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_LruCache(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uvloop.loop.__pyx_unpickle_LruCache", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_LruCache__set_state(<LruCache> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LruCache__set_state(LruCache __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._dict = __pyx_state[0]; __pyx_result._dict_get = __pyx_state[1]; __pyx_result._dict_move_to_end = __pyx_state[2]; __pyx_result._maxsize = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6uvloop_4loop___pyx_unpickle_LruCache__set_state(struct __pyx_obj_6uvloop_4loop_LruCache *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_LruCache__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_LruCache__set_state(LruCache __pyx_result, tuple __pyx_state):
 *     __pyx_result._dict = __pyx_state[0]; __pyx_result._dict_get = __pyx_state[1]; __pyx_result._dict_move_to_end = __pyx_state[2]; __pyx_result._maxsize = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_dict);
  __Pyx_DECREF(__pyx_v___pyx_result->_dict);
  __pyx_v___pyx_result->_dict = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_dict_get);
  __Pyx_DECREF(__pyx_v___pyx_result->_dict_get);
  __pyx_v___pyx_result->_dict_get = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_dict_move_to_end);
  __Pyx_DECREF(__pyx_v___pyx_result->_dict_move_to_end);
  __pyx_v___pyx_result->_dict_move_to_end = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(6, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_maxsize = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_LruCache__set_state(LruCache __pyx_result, tuple __pyx_state):
 *     __pyx_result._dict = __pyx_state[0]; __pyx_result._dict_get = __pyx_state[1]; __pyx_result._dict_move_to_end = __pyx_state[2]; __pyx_result._maxsize = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(6, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 4);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._dict = __pyx_state[0]; __pyx_result._dict_get = __pyx_state[1]; __pyx_result._dict_move_to_end = __pyx_state[2]; __pyx_result._maxsize = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(6, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_LruCache__set_state(LruCache __pyx_result, tuple __pyx_state):
 *     __pyx_result._dict = __pyx_state[0]; __pyx_result._dict_get = __pyx_state[1]; __pyx_result._dict_move_to_end = __pyx_state[2]; __pyx_result._maxsize = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_LruCache__set_state(<LruCache> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_LruCache__set_state(LruCache __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._dict = __pyx_state[0]; __pyx_result._dict_get = __pyx_state[1]; __pyx_result._dict_move_to_end = __pyx_state[2]; __pyx_result._maxsize = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uvloop.loop.__pyx_unpickle_LruCache__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6uvloop_4loop_UVHandle __pyx_vtable_6uvloop_4loop_UVHandle;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVHandle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVHandle *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_UVHandle *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_UVHandle;
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  p->_source_traceback = Py_None; Py_INCREF(Py_None);
  p->context = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_8UVHandle_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVHandle(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVHandle *p = (struct __pyx_obj_6uvloop_4loop_UVHandle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVHandle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_8UVHandle_5__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_loop);
  Py_CLEAR(p->_source_traceback);
  Py_CLEAR(p->context);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop_UVHandle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVHandle *p = (struct __pyx_obj_6uvloop_4loop_UVHandle *)o;
  if (p->_loop) {
    e = (*v)(((PyObject *)p->_loop), a); if (e) return e;
  }
  if (p->_source_traceback) {
    e = (*v)(p->_source_traceback, a); if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_UVHandle(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_UVHandle *p = (struct __pyx_obj_6uvloop_4loop_UVHandle *)o;
  tmp = ((PyObject*)p->_loop);
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_source_traceback);
  p->_source_traceback = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->context);
  p->context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_8UVHandle__source_traceback(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_8UVHandle_17_source_traceback_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6uvloop_4loop_8UVHandle_7__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__(self);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVHandle[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVHandle_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVHandle_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVHandle_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVHandle_10__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop_UVHandle[] = {
  {(char *)"_source_traceback", __pyx_getprop_6uvloop_4loop_8UVHandle__source_traceback, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVHandle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVHandle},
  {Py_tp_repr, (void *)__pyx_pw_6uvloop_4loop_8UVHandle_7__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("UVHandle()\nA base class for all libuv handles.\n\n    Automatically manages memory deallocation and closing.\n\n    Important:\n\n       1. call \"_ensure_alive()\" before calling any libuv functions on\n          your handles.\n\n       2. call \"__ensure_handle_data\" in *all* libuv handle callbacks.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVHandle},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_UVHandle},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVHandle},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop_UVHandle},
  {Py_tp_init, (void *)__pyx_pw_6uvloop_4loop_8UVHandle_3__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVHandle},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVHandle_spec = {
  "uvloop.loop.UVHandle",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVHandle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVHandle_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVHandle = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVHandle", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVHandle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("UVHandle()\nA base class for all libuv handles.\n\n    Automatically manages memory deallocation and closing.\n\n    Important:\n\n       1. call \"_ensure_alive()\" before calling any libuv functions on\n          your handles.\n\n       2. call \"__ensure_handle_data\" in *all* libuv handle callbacks.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVHandle, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_UVHandle, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVHandle, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop_UVHandle, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVHandle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVSocketHandle __pyx_vtable_6uvloop_4loop_UVSocketHandle;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVSocketHandle(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVSocketHandle *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVSocketHandle;
  p->_fileobj = Py_None; Py_INCREF(Py_None);
  p->_UVSocketHandle__cached_socket = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_14UVSocketHandle_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVSocketHandle(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVSocketHandle *p = (struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVSocketHandle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_fileobj);
  Py_CLEAR(p->_UVSocketHandle__cached_socket);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVSocketHandle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVSocketHandle *p = (struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->_fileobj) {
    e = (*v)(p->_fileobj, a); if (e) return e;
  }
  if (p->_UVSocketHandle__cached_socket) {
    e = (*v)(p->_UVSocketHandle__cached_socket, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_UVSocketHandle(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_UVSocketHandle *p = (struct __pyx_obj_6uvloop_4loop_UVSocketHandle *)o;
  __pyx_tp_clear_6uvloop_4loop_UVHandle(o);
  tmp = ((PyObject*)p->_fileobj);
  p->_fileobj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_UVSocketHandle__cached_socket);
  p->_UVSocketHandle__cached_socket = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVSocketHandle[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVSocketHandle_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVSocketHandle_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVSocketHandle_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVSocketHandle_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVSocketHandle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVSocketHandle},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVSocketHandle},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_UVSocketHandle},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVSocketHandle},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVSocketHandle},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVSocketHandle_spec = {
  "uvloop.loop.UVSocketHandle",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVSocketHandle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVSocketHandle_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVSocketHandle = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVSocketHandle", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVSocketHandle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVSocketHandle, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVSocketHandle, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_UVSocketHandle, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVSocketHandle, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVSocketHandle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVAsync __pyx_vtable_6uvloop_4loop_UVAsync;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVAsync(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVAsync *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVAsync *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVAsync;
  p->ctx = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVAsync(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVAsync *p = (struct __pyx_obj_6uvloop_4loop_UVAsync *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVAsync) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->ctx);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVAsync(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVAsync *p = (struct __pyx_obj_6uvloop_4loop_UVAsync *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->ctx) {
    e = (*v)(p->ctx, a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVAsync[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVAsync_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVAsync___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVAsync_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVAsync_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVAsync_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVAsync},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVAsync},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVAsync},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVAsync},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVAsync_spec = {
  "uvloop.loop.UVAsync",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVAsync),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVAsync_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVAsync = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVAsync", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVAsync), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVAsync, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVAsync, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVAsync, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVAsync, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVTimer __pyx_vtable_6uvloop_4loop_UVTimer;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVTimer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVTimer *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVTimer *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVTimer;
  p->ctx = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVTimer(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVTimer *p = (struct __pyx_obj_6uvloop_4loop_UVTimer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVTimer) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->ctx);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVTimer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVTimer *p = (struct __pyx_obj_6uvloop_4loop_UVTimer *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->ctx) {
    e = (*v)(p->ctx, a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVTimer[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVTimer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVTimer___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVTimer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVTimer_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVTimer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVTimer},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVTimer},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVTimer},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVTimer},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVTimer_spec = {
  "uvloop.loop.UVTimer",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVTimer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVTimer_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVTimer = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVTimer", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVTimer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVTimer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVTimer, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVTimer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVTimer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVIdle __pyx_vtable_6uvloop_4loop_UVIdle;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVIdle(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVIdle *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVIdle *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVIdle;
  p->h = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVIdle(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVIdle *p = (struct __pyx_obj_6uvloop_4loop_UVIdle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVIdle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->h);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVIdle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVIdle *p = (struct __pyx_obj_6uvloop_4loop_UVIdle *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->h) {
    e = (*v)(((PyObject *)p->h), a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVIdle[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVIdle_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVIdle___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVIdle_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVIdle_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVIdle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVIdle},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVIdle},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVIdle},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVIdle},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVIdle_spec = {
  "uvloop.loop.UVIdle",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVIdle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVIdle_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVIdle = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVIdle", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVIdle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVIdle, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVIdle, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVIdle, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVIdle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVBaseTransport __pyx_vtable_6uvloop_4loop_UVBaseTransport;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVBaseTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVBaseTransport *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVSocketHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVBaseTransport;
  p->_protocol_data_received = Py_None; Py_INCREF(Py_None);
  p->_protocol = Py_None; Py_INCREF(Py_None);
  p->_server = ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None); Py_INCREF(Py_None);
  p->_waiter = Py_None; Py_INCREF(Py_None);
  p->_extra_info = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_15UVBaseTransport_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVBaseTransport(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVBaseTransport *p = (struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVBaseTransport) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (p->__weakref__) PyObject_ClearWeakRefs(o);
  Py_CLEAR(p->_protocol_data_received);
  Py_CLEAR(p->_protocol);
  Py_CLEAR(p->_server);
  Py_CLEAR(p->_waiter);
  Py_CLEAR(p->_extra_info);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVSocketHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVBaseTransport(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVBaseTransport *p = (struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVSocketHandle(o, v, a); if (e) return e;
  if (p->_protocol_data_received) {
    e = (*v)(p->_protocol_data_received, a); if (e) return e;
  }
  if (p->_protocol) {
    e = (*v)(p->_protocol, a); if (e) return e;
  }
  if (p->_server) {
    e = (*v)(((PyObject *)p->_server), a); if (e) return e;
  }
  if (p->_waiter) {
    e = (*v)(p->_waiter, a); if (e) return e;
  }
  if (p->_extra_info) {
    e = (*v)(p->_extra_info, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_UVBaseTransport(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_UVBaseTransport *p = (struct __pyx_obj_6uvloop_4loop_UVBaseTransport *)o;
  __pyx_tp_clear_6uvloop_4loop_UVSocketHandle(o);
  tmp = ((PyObject*)p->_protocol_data_received);
  p->_protocol_data_received = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_protocol);
  p->_protocol = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_server);
  p->_server = ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_waiter);
  p->_waiter = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_extra_info);
  p->_extra_info = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_15UVBaseTransport__paused(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_15UVBaseTransport_7_paused_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_15UVBaseTransport__closing(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_15UVBaseTransport_8_closing_1__get__(o);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVBaseTransport[] = {
  {"get_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_3get_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_2get_protocol},
  {"set_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_5set_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_4set_protocol},
  {"_force_close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_7_force_close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_6_force_close},
  {"abort", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_9abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_8abort},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_11close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_10close},
  {"is_closing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_13is_closing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_12is_closing},
  {"get_write_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_15get_write_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_14get_write_buffer_size},
  {"set_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_17set_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_16set_write_buffer_limits},
  {"get_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_19get_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_18get_write_buffer_limits},
  {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_21get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_20get_extra_info},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_22__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15UVBaseTransport_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15UVBaseTransport_24__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop_UVBaseTransport[] = {
  {(char *)"_paused", __pyx_getprop_6uvloop_4loop_15UVBaseTransport__paused, 0, (char *)0, 0},
  {(char *)"_closing", __pyx_getprop_6uvloop_4loop_15UVBaseTransport__closing, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVBaseTransport_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVBaseTransport},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVBaseTransport},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_UVBaseTransport},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVBaseTransport},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop_UVBaseTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVBaseTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVBaseTransport_spec = {
  "uvloop.loop.UVBaseTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVBaseTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVBaseTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVBaseTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVBaseTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVBaseTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVBaseTransport, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVBaseTransport, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_UVBaseTransport, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVBaseTransport, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop_UVBaseTransport, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVBaseTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_Loop __pyx_vtable_6uvloop_4loop_Loop;

static PyObject *__pyx_tp_new_6uvloop_4loop_Loop(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_Loop *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_Loop *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_Loop;
  p->slow_callback_duration = Py_None; Py_INCREF(Py_None);
  p->_task_factory = Py_None; Py_INCREF(Py_None);
  p->_exception_handler = Py_None; Py_INCREF(Py_None);
  p->_default_executor = Py_None; Py_INCREF(Py_None);
  p->_ready = Py_None; Py_INCREF(Py_None);
  p->_queued_streams = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_executing_streams = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_servers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_transports = Py_None; Py_INCREF(Py_None);
  p->_processes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_fd_to_reader_fileobj = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_fd_to_writer_fileobj = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_unix_server_sockets = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_signals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_signal_handlers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_ssock = Py_None; Py_INCREF(Py_None);
  p->_csock = Py_None; Py_INCREF(Py_None);
  p->_timers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_polls = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->active_process_handler = ((struct __pyx_obj_6uvloop_4loop_UVProcess *)Py_None); Py_INCREF(Py_None);
  p->handler_async = ((struct __pyx_obj_6uvloop_4loop_UVAsync *)Py_None); Py_INCREF(Py_None);
  p->handler_idle = ((struct __pyx_obj_6uvloop_4loop_UVIdle *)Py_None); Py_INCREF(Py_None);
  p->handler_check__exec_writes = ((struct __pyx_obj_6uvloop_4loop_UVCheck *)Py_None); Py_INCREF(Py_None);
  p->_last_error = Py_None; Py_INCREF(Py_None);
  p->_asyncgens = Py_None; Py_INCREF(Py_None);
  p->_debug_handles_total = Py_None; Py_INCREF(Py_None);
  p->_debug_handles_closed = Py_None; Py_INCREF(Py_None);
  p->_debug_handles_current = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_4Loop_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_Loop(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_Loop *p = (struct __pyx_obj_6uvloop_4loop_Loop *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_Loop) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (p->__weakref__) PyObject_ClearWeakRefs(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_4Loop_5__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->slow_callback_duration);
  Py_CLEAR(p->_task_factory);
  Py_CLEAR(p->_exception_handler);
  Py_CLEAR(p->_default_executor);
  Py_CLEAR(p->_ready);
  Py_CLEAR(p->_queued_streams);
  Py_CLEAR(p->_executing_streams);
  Py_CLEAR(p->_servers);
  Py_CLEAR(p->_transports);
  Py_CLEAR(p->_processes);
  Py_CLEAR(p->_fd_to_reader_fileobj);
  Py_CLEAR(p->_fd_to_writer_fileobj);
  Py_CLEAR(p->_unix_server_sockets);
  Py_CLEAR(p->_signals);
  Py_CLEAR(p->_signal_handlers);
  Py_CLEAR(p->_ssock);
  Py_CLEAR(p->_csock);
  Py_CLEAR(p->_timers);
  Py_CLEAR(p->_polls);
  Py_CLEAR(p->active_process_handler);
  Py_CLEAR(p->handler_async);
  Py_CLEAR(p->handler_idle);
  Py_CLEAR(p->handler_check__exec_writes);
  Py_CLEAR(p->_last_error);
  Py_CLEAR(p->_asyncgens);
  Py_CLEAR(p->_debug_handles_total);
  Py_CLEAR(p->_debug_handles_closed);
  Py_CLEAR(p->_debug_handles_current);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop_Loop(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_Loop *p = (struct __pyx_obj_6uvloop_4loop_Loop *)o;
  if (p->slow_callback_duration) {
    e = (*v)(p->slow_callback_duration, a); if (e) return e;
  }
  if (p->_task_factory) {
    e = (*v)(p->_task_factory, a); if (e) return e;
  }
  if (p->_exception_handler) {
    e = (*v)(p->_exception_handler, a); if (e) return e;
  }
  if (p->_default_executor) {
    e = (*v)(p->_default_executor, a); if (e) return e;
  }
  if (p->_ready) {
    e = (*v)(p->_ready, a); if (e) return e;
  }
  if (p->_queued_streams) {
    e = (*v)(p->_queued_streams, a); if (e) return e;
  }
  if (p->_executing_streams) {
    e = (*v)(p->_executing_streams, a); if (e) return e;
  }
  if (p->_servers) {
    e = (*v)(p->_servers, a); if (e) return e;
  }
  if (p->_transports) {
    e = (*v)(p->_transports, a); if (e) return e;
  }
  if (p->_processes) {
    e = (*v)(p->_processes, a); if (e) return e;
  }
  if (p->_fd_to_reader_fileobj) {
    e = (*v)(p->_fd_to_reader_fileobj, a); if (e) return e;
  }
  if (p->_fd_to_writer_fileobj) {
    e = (*v)(p->_fd_to_writer_fileobj, a); if (e) return e;
  }
  if (p->_unix_server_sockets) {
    e = (*v)(p->_unix_server_sockets, a); if (e) return e;
  }
  if (p->_signals) {
    e = (*v)(p->_signals, a); if (e) return e;
  }
  if (p->_signal_handlers) {
    e = (*v)(p->_signal_handlers, a); if (e) return e;
  }
  if (p->_ssock) {
    e = (*v)(p->_ssock, a); if (e) return e;
  }
  if (p->_csock) {
    e = (*v)(p->_csock, a); if (e) return e;
  }
  if (p->_timers) {
    e = (*v)(p->_timers, a); if (e) return e;
  }
  if (p->_polls) {
    e = (*v)(p->_polls, a); if (e) return e;
  }
  if (p->active_process_handler) {
    e = (*v)(((PyObject *)p->active_process_handler), a); if (e) return e;
  }
  if (p->handler_async) {
    e = (*v)(((PyObject *)p->handler_async), a); if (e) return e;
  }
  if (p->handler_idle) {
    e = (*v)(((PyObject *)p->handler_idle), a); if (e) return e;
  }
  if (p->handler_check__exec_writes) {
    e = (*v)(((PyObject *)p->handler_check__exec_writes), a); if (e) return e;
  }
  if (p->_last_error) {
    e = (*v)(p->_last_error, a); if (e) return e;
  }
  if (p->_asyncgens) {
    e = (*v)(p->_asyncgens, a); if (e) return e;
  }
  if (p->_debug_handles_total) {
    e = (*v)(p->_debug_handles_total, a); if (e) return e;
  }
  if (p->_debug_handles_closed) {
    e = (*v)(p->_debug_handles_closed, a); if (e) return e;
  }
  if (p->_debug_handles_current) {
    e = (*v)(p->_debug_handles_current, a); if (e) return e;
  }
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop_print_debug_info(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_16print_debug_info_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop_slow_callback_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_1__get__(o);
}

static int __pyx_setprop_6uvloop_4loop_4Loop_slow_callback_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_3__set__(o, v);
  }
  else {
    return __pyx_pw_6uvloop_4loop_4Loop_22slow_callback_duration_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__closed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_7_closed_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_cc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_9_debug_cc_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_handles_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_20_debug_handles_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_handles_closed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_21_debug_handles_closed_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_handles_current(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_22_debug_handles_current_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_uv_handles_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_23_debug_uv_handles_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_uv_handles_freed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_23_debug_uv_handles_freed_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_cb_handles_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_23_debug_cb_handles_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_cb_handles_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_23_debug_cb_handles_count_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_cb_timer_handles_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_cb_timer_handles_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_29_debug_cb_timer_handles_count_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_shutdown_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_35_debug_stream_shutdown_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_listen_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_33_debug_stream_listen_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_cb_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_27_debug_stream_read_cb_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_cb_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_34_debug_stream_read_cb_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_eof_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_28_debug_stream_read_eof_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_eof_cb_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_38_debug_stream_read_eof_cb_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_31_debug_stream_read_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_tries(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_25_debug_stream_write_tries_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_32_debug_stream_write_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_ctx_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_29_debug_stream_write_ctx_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_ctx_cnt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_27_debug_stream_write_ctx_cnt_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_cb_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_35_debug_stream_write_cb_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__poll_read_events_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_23_poll_read_events_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__poll_read_cb_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_26_poll_read_cb_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__poll_write_events_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_24_poll_write_events_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__poll_write_cb_errors_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_27_poll_write_cb_errors_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__sock_try_write_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_21_sock_try_write_total_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_4Loop__debug_exception_handler_cnt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_4Loop_28_debug_exception_handler_cnt_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6uvloop_4loop_4Loop_11__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6uvloop_4loop_4Loop_11__repr__(self);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_Loop[] = {
  {"_Loop__sighandler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_7_Loop__sighandler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_6__sighandler},
  {"_get_backend_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_9_get_backend_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_8_get_backend_id},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6uvloop_4loop_4Loop_11__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"call_soon", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_13call_soon, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_12call_soon},
  {"call_soon_threadsafe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_15call_soon_threadsafe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_14call_soon_threadsafe},
  {"call_later", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_17call_later, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_16call_later},
  {"call_at", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_19call_at, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_18call_at},
  {"time", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_21time, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_20time},
  {"stop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_23stop, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_22stop},
  {"run_forever", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_25run_forever, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_24run_forever},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_27close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_26close},
  {"get_debug", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_29get_debug, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_28get_debug},
  {"set_debug", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_31set_debug, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_30set_debug},
  {"is_running", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_33is_running, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_32is_running},
  {"is_closed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_35is_closed, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_34is_closed},
  {"create_future", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_37create_future, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_36create_future},
  {"create_task", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_39create_task, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_38create_task},
  {"set_task_factory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_41set_task_factory, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_40set_task_factory},
  {"get_task_factory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_43get_task_factory, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_42get_task_factory},
  {"run_until_complete", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_45run_until_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_44run_until_complete},
  {"getaddrinfo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_47getaddrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_46getaddrinfo},
  {"getnameinfo", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_50getnameinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_49getnameinfo},
  {"start_tls", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_53start_tls, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_52start_tls},
  {"create_server", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_56create_server, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_55create_server},
  {"create_connection", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_59create_connection, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_58create_connection},
  {"create_unix_server", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_62create_unix_server, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_61create_unix_server},
  {"create_unix_connection", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_65create_unix_connection, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_64create_unix_connection},
  {"default_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_68default_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_67default_exception_handler},
  {"get_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_70get_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_69get_exception_handler},
  {"set_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_72set_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_71set_exception_handler},
  {"call_exception_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_74call_exception_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_73call_exception_handler},
  {"add_reader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_76add_reader, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_75add_reader},
  {"remove_reader", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_78remove_reader, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_77remove_reader},
  {"add_writer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_80add_writer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_79add_writer},
  {"remove_writer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_82remove_writer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_81remove_writer},
  {"sock_recv", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_84sock_recv, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_83sock_recv},
  {"sock_recv_into", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_87sock_recv_into, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_86sock_recv_into},
  {"sock_sendall", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_90sock_sendall, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_89sock_sendall},
  {"sock_accept", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_93sock_accept, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_92sock_accept},
  {"sock_connect", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_96sock_connect, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_95sock_connect},
  {"sock_recvfrom", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_99sock_recvfrom, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_98sock_recvfrom},
  {"sock_recvfrom_into", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_102sock_recvfrom_into, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_101sock_recvfrom_into},
  {"sock_sendto", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_105sock_sendto, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_104sock_sendto},
  {"connect_accepted_socket", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_108connect_accepted_socket, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_107connect_accepted_socket},
  {"run_in_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_111run_in_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_110run_in_executor},
  {"set_default_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_113set_default_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_112set_default_executor},
  {"_Loop__subprocess_run", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_115_Loop__subprocess_run, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_114__subprocess_run},
  {"subprocess_shell", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_118subprocess_shell, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_117subprocess_shell},
  {"subprocess_exec", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_121subprocess_exec, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_120subprocess_exec},
  {"connect_read_pipe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_124connect_read_pipe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_123connect_read_pipe},
  {"connect_write_pipe", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_127connect_write_pipe, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_126connect_write_pipe},
  {"add_signal_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_130add_signal_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_129add_signal_handler},
  {"remove_signal_handler", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_132remove_signal_handler, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_131remove_signal_handler},
  {"create_datagram_endpoint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_134create_datagram_endpoint, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_133create_datagram_endpoint},
  {"_monitor_fs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_137_monitor_fs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_136_monitor_fs},
  {"_check_default_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_139_check_default_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_138_check_default_executor},
  {"_asyncgen_finalizer_hook", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_141_asyncgen_finalizer_hook, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_140_asyncgen_finalizer_hook},
  {"_asyncgen_firstiter_hook", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_143_asyncgen_firstiter_hook, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_142_asyncgen_firstiter_hook},
  {"shutdown_asyncgens", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_145shutdown_asyncgens, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_144shutdown_asyncgens},
  {"shutdown_default_executor", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_148shutdown_default_executor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_147shutdown_default_executor},
  {"_do_shutdown", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_151_do_shutdown, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_150_do_shutdown},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_153__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_152__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_4Loop_155__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_4Loop_154__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop_Loop[] = {
  {(char *)"print_debug_info", __pyx_getprop_6uvloop_4loop_4Loop_print_debug_info, 0, (char *)0, 0},
  {(char *)"slow_callback_duration", __pyx_getprop_6uvloop_4loop_4Loop_slow_callback_duration, __pyx_setprop_6uvloop_4loop_4Loop_slow_callback_duration, (char *)PyDoc_STR("slow_callback_duration: object"), 0},
  {(char *)"_closed", __pyx_getprop_6uvloop_4loop_4Loop__closed, 0, (char *)0, 0},
  {(char *)"_debug_cc", __pyx_getprop_6uvloop_4loop_4Loop__debug_cc, 0, (char *)0, 0},
  {(char *)"_debug_handles_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_handles_total, 0, (char *)0, 0},
  {(char *)"_debug_handles_closed", __pyx_getprop_6uvloop_4loop_4Loop__debug_handles_closed, 0, (char *)0, 0},
  {(char *)"_debug_handles_current", __pyx_getprop_6uvloop_4loop_4Loop__debug_handles_current, 0, (char *)0, 0},
  {(char *)"_debug_uv_handles_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_uv_handles_total, 0, (char *)0, 0},
  {(char *)"_debug_uv_handles_freed", __pyx_getprop_6uvloop_4loop_4Loop__debug_uv_handles_freed, 0, (char *)0, 0},
  {(char *)"_debug_cb_handles_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_cb_handles_total, 0, (char *)0, 0},
  {(char *)"_debug_cb_handles_count", __pyx_getprop_6uvloop_4loop_4Loop__debug_cb_handles_count, 0, (char *)0, 0},
  {(char *)"_debug_cb_timer_handles_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_cb_timer_handles_total, 0, (char *)0, 0},
  {(char *)"_debug_cb_timer_handles_count", __pyx_getprop_6uvloop_4loop_4Loop__debug_cb_timer_handles_count, 0, (char *)0, 0},
  {(char *)"_debug_stream_shutdown_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_shutdown_errors_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_listen_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_listen_errors_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_read_cb_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_cb_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_read_cb_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_cb_errors_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_read_eof_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_eof_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_read_eof_cb_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_eof_cb_errors_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_read_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_read_errors_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_write_tries", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_tries, 0, (char *)0, 0},
  {(char *)"_debug_stream_write_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_errors_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_write_ctx_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_ctx_total, 0, (char *)0, 0},
  {(char *)"_debug_stream_write_ctx_cnt", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_ctx_cnt, 0, (char *)0, 0},
  {(char *)"_debug_stream_write_cb_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__debug_stream_write_cb_errors_total, 0, (char *)0, 0},
  {(char *)"_poll_read_events_total", __pyx_getprop_6uvloop_4loop_4Loop__poll_read_events_total, 0, (char *)0, 0},
  {(char *)"_poll_read_cb_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__poll_read_cb_errors_total, 0, (char *)0, 0},
  {(char *)"_poll_write_events_total", __pyx_getprop_6uvloop_4loop_4Loop__poll_write_events_total, 0, (char *)0, 0},
  {(char *)"_poll_write_cb_errors_total", __pyx_getprop_6uvloop_4loop_4Loop__poll_write_cb_errors_total, 0, (char *)0, 0},
  {(char *)"_sock_try_write_total", __pyx_getprop_6uvloop_4loop_4Loop__sock_try_write_total, 0, (char *)0, 0},
  {(char *)"_debug_exception_handler_cnt", __pyx_getprop_6uvloop_4loop_4Loop__debug_exception_handler_cnt, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_Loop_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_Loop},
  {Py_tp_repr, (void *)__pyx_pw_6uvloop_4loop_4Loop_11__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Loop()")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_Loop},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_Loop},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop_Loop},
  {Py_tp_init, (void *)__pyx_pw_6uvloop_4loop_4Loop_3__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_Loop},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_Loop_spec = {
  "uvloop.loop.Loop",
  sizeof(struct __pyx_obj_6uvloop_4loop_Loop),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_Loop_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_Loop = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""Loop", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_Loop), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_Loop, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6uvloop_4loop_4Loop_11__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Loop()"), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_Loop, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_Loop, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop_Loop, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6uvloop_4loop_4Loop_3__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_Loop, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_Handle __pyx_vtable_6uvloop_4loop_Handle;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop_Handle *__pyx_freelist_6uvloop_4loop_Handle[250];
static int __pyx_freecount_6uvloop_4loop_Handle = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop_Handle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_Handle *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop_Handle > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_Handle)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop_Handle[--__pyx_freecount_6uvloop_4loop_Handle];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop_Handle));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_Handle *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_Handle;
  p->loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  p->context = Py_None; Py_INCREF(Py_None);
  p->meth_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->arg1 = Py_None; Py_INCREF(Py_None);
  p->arg2 = Py_None; Py_INCREF(Py_None);
  p->arg3 = Py_None; Py_INCREF(Py_None);
  p->arg4 = Py_None; Py_INCREF(Py_None);
  p->_source_traceback = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_6Handle_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_Handle(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_Handle *p = (struct __pyx_obj_6uvloop_4loop_Handle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_Handle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (p->__weakref__) PyObject_ClearWeakRefs(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_6Handle_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->loop);
  Py_CLEAR(p->context);
  Py_CLEAR(p->meth_name);
  Py_CLEAR(p->arg1);
  Py_CLEAR(p->arg2);
  Py_CLEAR(p->arg3);
  Py_CLEAR(p->arg4);
  Py_CLEAR(p->_source_traceback);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop_Handle < 250) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_Handle)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_6uvloop_4loop_Handle[__pyx_freecount_6uvloop_4loop_Handle++] = ((struct __pyx_obj_6uvloop_4loop_Handle *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop_Handle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_Handle *p = (struct __pyx_obj_6uvloop_4loop_Handle *)o;
  if (p->loop) {
    e = (*v)(((PyObject *)p->loop), a); if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  if (p->arg1) {
    e = (*v)(p->arg1, a); if (e) return e;
  }
  if (p->arg2) {
    e = (*v)(p->arg2, a); if (e) return e;
  }
  if (p->arg3) {
    e = (*v)(p->arg3, a); if (e) return e;
  }
  if (p->arg4) {
    e = (*v)(p->arg4, a); if (e) return e;
  }
  if (p->_source_traceback) {
    e = (*v)(p->_source_traceback, a); if (e) return e;
  }
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_6Handle__source_traceback(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_6Handle_17_source_traceback_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6uvloop_4loop_6Handle_7__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6uvloop_4loop_6Handle_7__repr__(self);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_Handle[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6uvloop_4loop_6Handle_7__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_9cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_8cancel},
  {"cancelled", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_11cancelled, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_10cancelled},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_12__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Handle_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Handle_14__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop_Handle[] = {
  {(char *)"_source_traceback", __pyx_getprop_6uvloop_4loop_6Handle__source_traceback, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_Handle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_Handle},
  {Py_tp_repr, (void *)__pyx_pw_6uvloop_4loop_6Handle_7__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("Handle()")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_Handle},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_Handle},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop_Handle},
  {Py_tp_init, (void *)__pyx_pw_6uvloop_4loop_6Handle_5__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_Handle},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_Handle_spec = {
  "uvloop.loop.Handle",
  sizeof(struct __pyx_obj_6uvloop_4loop_Handle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_Handle_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_Handle = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""Handle", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_Handle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_Handle, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6uvloop_4loop_6Handle_7__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Handle()"), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_Handle, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_Handle, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop_Handle, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6uvloop_4loop_6Handle_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_Handle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_TimerHandle __pyx_vtable_6uvloop_4loop_TimerHandle;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop_TimerHandle *__pyx_freelist_6uvloop_4loop_TimerHandle[250];
static int __pyx_freecount_6uvloop_4loop_TimerHandle = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop_TimerHandle(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_TimerHandle *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop_TimerHandle > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_TimerHandle)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop_TimerHandle[--__pyx_freecount_6uvloop_4loop_TimerHandle];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop_TimerHandle));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_TimerHandle *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_TimerHandle;
  p->callback = Py_None; Py_INCREF(Py_None);
  p->args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->timer = ((struct __pyx_obj_6uvloop_4loop_UVTimer *)Py_None); Py_INCREF(Py_None);
  p->loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  p->context = Py_None; Py_INCREF(Py_None);
  p->_debug_info = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_when = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_11TimerHandle_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_TimerHandle(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_TimerHandle *p = (struct __pyx_obj_6uvloop_4loop_TimerHandle *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_TimerHandle) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (p->__weakref__) PyObject_ClearWeakRefs(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_11TimerHandle_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->callback);
  Py_CLEAR(p->args);
  Py_CLEAR(p->timer);
  Py_CLEAR(p->loop);
  Py_CLEAR(p->context);
  Py_CLEAR(p->_debug_info);
  Py_CLEAR(p->_when);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop_TimerHandle < 250) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_TimerHandle)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_6uvloop_4loop_TimerHandle[__pyx_freecount_6uvloop_4loop_TimerHandle++] = ((struct __pyx_obj_6uvloop_4loop_TimerHandle *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop_TimerHandle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_TimerHandle *p = (struct __pyx_obj_6uvloop_4loop_TimerHandle *)o;
  if (p->callback) {
    e = (*v)(p->callback, a); if (e) return e;
  }
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  if (p->timer) {
    e = (*v)(((PyObject *)p->timer), a); if (e) return e;
  }
  if (p->loop) {
    e = (*v)(((PyObject *)p->loop), a); if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  if (p->_debug_info) {
    e = (*v)(p->_debug_info, a); if (e) return e;
  }
  if (p->_when) {
    e = (*v)(p->_when, a); if (e) return e;
  }
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_11TimerHandle__source_traceback(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_11TimerHandle_17_source_traceback_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6uvloop_4loop_11TimerHandle_5__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6uvloop_4loop_11TimerHandle_5__repr__(self);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_TimerHandle[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6uvloop_4loop_11TimerHandle_5__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"cancelled", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_7cancelled, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_6cancelled},
  {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_9cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_8cancel},
  {"when", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_11when, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_10when},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_13__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_12__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11TimerHandle_15__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11TimerHandle_14__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop_TimerHandle[] = {
  {(char *)"_source_traceback", __pyx_getprop_6uvloop_4loop_11TimerHandle__source_traceback, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_TimerHandle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_TimerHandle},
  {Py_tp_repr, (void *)__pyx_pw_6uvloop_4loop_11TimerHandle_5__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_TimerHandle},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_TimerHandle},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop_TimerHandle},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_TimerHandle},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_TimerHandle_spec = {
  "uvloop.loop.TimerHandle",
  sizeof(struct __pyx_obj_6uvloop_4loop_TimerHandle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_TimerHandle_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_TimerHandle = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""TimerHandle", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_TimerHandle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_TimerHandle, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6uvloop_4loop_11TimerHandle_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_TimerHandle, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_TimerHandle, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop_TimerHandle, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_TimerHandle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVCheck __pyx_vtable_6uvloop_4loop_UVCheck;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVCheck(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVCheck *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVCheck *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVCheck;
  p->h = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVCheck(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVCheck *p = (struct __pyx_obj_6uvloop_4loop_UVCheck *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVCheck) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->h);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVCheck(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVCheck *p = (struct __pyx_obj_6uvloop_4loop_UVCheck *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->h) {
    e = (*v)(((PyObject *)p->h), a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVCheck[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVCheck_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVCheck___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_7UVCheck_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_7UVCheck_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVCheck_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVCheck},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVCheck},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVCheck},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVCheck},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVCheck_spec = {
  "uvloop.loop.UVCheck",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVCheck),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVCheck_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVCheck = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVCheck", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVCheck), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVCheck, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVCheck, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVCheck, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVCheck, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVPoll __pyx_vtable_6uvloop_4loop_UVPoll;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVPoll(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVPoll *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVPoll;
  p->reading_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None); Py_INCREF(Py_None);
  p->writing_handle = ((struct __pyx_obj_6uvloop_4loop_Handle *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVPoll(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVPoll *p = (struct __pyx_obj_6uvloop_4loop_UVPoll *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVPoll) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->reading_handle);
  Py_CLEAR(p->writing_handle);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVPoll(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVPoll *p = (struct __pyx_obj_6uvloop_4loop_UVPoll *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->reading_handle) {
    e = (*v)(((PyObject *)p->reading_handle), a); if (e) return e;
  }
  if (p->writing_handle) {
    e = (*v)(((PyObject *)p->writing_handle), a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVPoll[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVPoll_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVPoll___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6UVPoll_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6UVPoll_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVPoll_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVPoll},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVPoll},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVPoll},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVPoll},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVPoll_spec = {
  "uvloop.loop.UVPoll",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVPoll),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVPoll_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVPoll = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVPoll", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVPoll), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVPoll, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVPoll, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVPoll, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVPoll, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVStream __pyx_vtable_6uvloop_4loop_UVStream;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVStream(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVStream *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVBaseTransport(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVStream *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVStream;
  p->_protocol_get_buffer = Py_None; Py_INCREF(Py_None);
  p->_protocol_buffer_updated = Py_None; Py_INCREF(Py_None);
  p->_buffer = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_read_pybuf.obj = NULL;
  if (unlikely(__pyx_pw_6uvloop_4loop_8UVStream_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVStream(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVStream *p = (struct __pyx_obj_6uvloop_4loop_UVStream *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVStream) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_protocol_get_buffer);
  Py_CLEAR(p->_protocol_buffer_updated);
  Py_CLEAR(p->_buffer);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVBaseTransport(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVStream(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVStream *p = (struct __pyx_obj_6uvloop_4loop_UVStream *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVBaseTransport(o, v, a); if (e) return e;
  if (p->_protocol_get_buffer) {
    e = (*v)(p->_protocol_get_buffer, a); if (e) return e;
  }
  if (p->_protocol_buffer_updated) {
    e = (*v)(p->_protocol_buffer_updated, a); if (e) return e;
  }
  if (p->_buffer) {
    e = (*v)(p->_buffer, a); if (e) return e;
  }
  if (p->_read_pybuf.obj) {
    e = (*v)(p->_read_pybuf.obj, a); if (e) return e;
  }
  return 0;
}

static PyObject *__pyx_specialmethod___pyx_pw_6uvloop_4loop_8UVStream_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6uvloop_4loop_8UVStream_3__repr__(self);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVStream[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6uvloop_4loop_8UVStream_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"write", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_5write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_4write},
  {"writelines", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_7writelines, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_6writelines},
  {"write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_9write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_8write_eof},
  {"can_write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_11can_write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_10can_write_eof},
  {"is_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_13is_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_12is_reading},
  {"pause_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_15pause_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_14pause_reading},
  {"resume_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_17resume_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_16resume_reading},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_18__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8UVStream_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8UVStream_20__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVStream_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVStream},
  {Py_tp_repr, (void *)__pyx_pw_6uvloop_4loop_8UVStream_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStream},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVStream},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVStream},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVStream_spec = {
  "uvloop.loop.UVStream",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVStream),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVStream_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVStream = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVStream", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVStream), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStream, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6uvloop_4loop_8UVStream_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStream, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVStream, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVStream, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVStreamServer __pyx_vtable_6uvloop_4loop_UVStreamServer;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVStreamServer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVStreamServer *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVSocketHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVStreamServer *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVStreamServer;
  p->ssl = Py_None; Py_INCREF(Py_None);
  p->ssl_handshake_timeout = Py_None; Py_INCREF(Py_None);
  p->ssl_shutdown_timeout = Py_None; Py_INCREF(Py_None);
  p->protocol_factory = Py_None; Py_INCREF(Py_None);
  p->_server = ((struct __pyx_obj_6uvloop_4loop_Server *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_14UVStreamServer_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVStreamServer(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVStreamServer *p = (struct __pyx_obj_6uvloop_4loop_UVStreamServer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVStreamServer) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->ssl);
  Py_CLEAR(p->ssl_handshake_timeout);
  Py_CLEAR(p->ssl_shutdown_timeout);
  Py_CLEAR(p->protocol_factory);
  Py_CLEAR(p->_server);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVSocketHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVStreamServer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVStreamServer *p = (struct __pyx_obj_6uvloop_4loop_UVStreamServer *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVSocketHandle(o, v, a); if (e) return e;
  if (p->ssl) {
    e = (*v)(p->ssl, a); if (e) return e;
  }
  if (p->ssl_handshake_timeout) {
    e = (*v)(p->ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->ssl_shutdown_timeout) {
    e = (*v)(p->ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->protocol_factory) {
    e = (*v)(p->protocol_factory, a); if (e) return e;
  }
  if (p->_server) {
    e = (*v)(((PyObject *)p->_server), a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVStreamServer[] = {
  {"_UVStreamServer__on_ssl_connected", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVStreamServer_3_UVStreamServer__on_ssl_connected, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVStreamServer_2__on_ssl_connected},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVStreamServer_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVStreamServer_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14UVStreamServer_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14UVStreamServer_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVStreamServer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVStreamServer},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStreamServer},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVStreamServer},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVStreamServer},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVStreamServer_spec = {
  "uvloop.loop.UVStreamServer",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVStreamServer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVStreamServer_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVStreamServer = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVStreamServer", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVStreamServer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStreamServer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStreamServer, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVStreamServer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVStreamServer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_TCPServer __pyx_vtable_6uvloop_4loop_TCPServer;

static PyObject *__pyx_tp_new_6uvloop_4loop_TCPServer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_TCPServer *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVStreamServer(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_TCPServer *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_TCPServer;
  return o;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_TCPServer[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9TCPServer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9TCPServer___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9TCPServer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9TCPServer_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_TCPServer_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStreamServer},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_TCPServer},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_TCPServer},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_TCPServer_spec = {
  "uvloop.loop.TCPServer",
  sizeof(struct __pyx_obj_6uvloop_4loop_TCPServer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_TCPServer_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_TCPServer = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""TCPServer", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_TCPServer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStreamServer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStreamServer, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_TCPServer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_TCPServer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_TCPTransport __pyx_vtable_6uvloop_4loop_TCPTransport;

static PyObject *__pyx_tp_new_6uvloop_4loop_TCPTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_TCPTransport *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVStream(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_TCPTransport;
  return o;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_TCPTransport[] = {
  {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12TCPTransport_1get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12TCPTransport_get_extra_info},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12TCPTransport_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12TCPTransport_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12TCPTransport_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12TCPTransport_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_TCPTransport_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStream},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_TCPTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_TCPTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_TCPTransport_spec = {
  "uvloop.loop.TCPTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop_TCPTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_TCPTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_TCPTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""TCPTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_TCPTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStream, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVStream_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStream, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_TCPTransport, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_TCPTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UnixServer __pyx_vtable_6uvloop_4loop_UnixServer;

static PyObject *__pyx_tp_new_6uvloop_4loop_UnixServer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UnixServer *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVStreamServer(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UnixServer *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UnixServer;
  return o;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UnixServer[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_10UnixServer_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_10UnixServer___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_10UnixServer_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_10UnixServer_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UnixServer_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStreamServer},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UnixServer},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UnixServer},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UnixServer_spec = {
  "uvloop.loop.UnixServer",
  sizeof(struct __pyx_obj_6uvloop_4loop_UnixServer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UnixServer_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UnixServer = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UnixServer", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UnixServer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStreamServer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStreamServer, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UnixServer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UnixServer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UnixTransport __pyx_vtable_6uvloop_4loop_UnixTransport;

static PyObject *__pyx_tp_new_6uvloop_4loop_UnixTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UnixTransport *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVStream(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UnixTransport;
  return o;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UnixTransport[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_13UnixTransport_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_13UnixTransport___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_13UnixTransport_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_13UnixTransport_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UnixTransport_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStream},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UnixTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UnixTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UnixTransport_spec = {
  "uvloop.loop.UnixTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop_UnixTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UnixTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UnixTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UnixTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UnixTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStream, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVStream_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStream, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UnixTransport, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UnixTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_ReadUnixTransport __pyx_vtable_6uvloop_4loop_ReadUnixTransport;

static PyObject *__pyx_tp_new_6uvloop_4loop_ReadUnixTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVStream(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_ReadUnixTransport;
  return o;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_ReadUnixTransport[] = {
  {"get_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_1get_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_get_write_buffer_limits},
  {"set_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_3set_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_2set_write_buffer_limits},
  {"get_write_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_5get_write_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_4get_write_buffer_size},
  {"write", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_7write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_6write},
  {"writelines", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_9writelines, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_8writelines},
  {"write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_11write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_10write_eof},
  {"can_write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_13can_write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_12can_write_eof},
  {"abort", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_15abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_14abort},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_17ReadUnixTransport_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_17ReadUnixTransport_18__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_ReadUnixTransport_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStream},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_ReadUnixTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_ReadUnixTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_ReadUnixTransport_spec = {
  "uvloop.loop.ReadUnixTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop_ReadUnixTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_ReadUnixTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_ReadUnixTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""ReadUnixTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_ReadUnixTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStream, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVStream_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStream, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_ReadUnixTransport, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_ReadUnixTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_WriteUnixTransport __pyx_vtable_6uvloop_4loop_WriteUnixTransport;

static PyObject *__pyx_tp_new_6uvloop_4loop_WriteUnixTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVStream(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_WriteUnixTransport;
  return o;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_WriteUnixTransport[] = {
  {"pause_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_1pause_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_pause_reading},
  {"resume_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_3resume_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_2resume_reading},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18WriteUnixTransport_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18WriteUnixTransport_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_WriteUnixTransport_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVStream},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_WriteUnixTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_WriteUnixTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_WriteUnixTransport_spec = {
  "uvloop.loop.WriteUnixTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop_WriteUnixTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_WriteUnixTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_WriteUnixTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""WriteUnixTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_WriteUnixTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVStream, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVStream_3__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVStream, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_WriteUnixTransport, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_WriteUnixTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVProcess __pyx_vtable_6uvloop_4loop_UVProcess;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVProcess(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVProcess *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVProcess *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVProcess;
  p->_returncode = Py_None; Py_INCREF(Py_None);
  p->_pid = Py_None; Py_INCREF(Py_None);
  p->_errpipe_read = Py_None; Py_INCREF(Py_None);
  p->_errpipe_write = Py_None; Py_INCREF(Py_None);
  p->_preexec_fn = Py_None; Py_INCREF(Py_None);
  p->_fds_to_close = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_UVProcess__env = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_UVProcess__args = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_UVProcess__cwd = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_9UVProcess_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVProcess(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVProcess *p = (struct __pyx_obj_6uvloop_4loop_UVProcess *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVProcess) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_9UVProcess_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_returncode);
  Py_CLEAR(p->_pid);
  Py_CLEAR(p->_errpipe_read);
  Py_CLEAR(p->_errpipe_write);
  Py_CLEAR(p->_preexec_fn);
  Py_CLEAR(p->_fds_to_close);
  Py_CLEAR(p->_UVProcess__env);
  Py_CLEAR(p->_UVProcess__args);
  Py_CLEAR(p->_UVProcess__cwd);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVProcess(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVProcess *p = (struct __pyx_obj_6uvloop_4loop_UVProcess *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->_returncode) {
    e = (*v)(p->_returncode, a); if (e) return e;
  }
  if (p->_pid) {
    e = (*v)(p->_pid, a); if (e) return e;
  }
  if (p->_errpipe_read) {
    e = (*v)(p->_errpipe_read, a); if (e) return e;
  }
  if (p->_errpipe_write) {
    e = (*v)(p->_errpipe_write, a); if (e) return e;
  }
  if (p->_preexec_fn) {
    e = (*v)(p->_preexec_fn, a); if (e) return e;
  }
  if (p->_fds_to_close) {
    e = (*v)(p->_fds_to_close, a); if (e) return e;
  }
  if (p->_UVProcess__env) {
    e = (*v)(p->_UVProcess__env, a); if (e) return e;
  }
  if (p->_UVProcess__args) {
    e = (*v)(p->_UVProcess__args, a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVProcess[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVProcess_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVProcess_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVProcess_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVProcess_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVProcess_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVProcess},
  {Py_tp_doc, (void *)PyDoc_STR("Abstract class; wrapper over uv_process_t handle.")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVProcess},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVProcess},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVProcess},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVProcess_spec = {
  "uvloop.loop.UVProcess",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVProcess),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVProcess_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVProcess = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVProcess", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVProcess), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVProcess, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Abstract class; wrapper over uv_process_t handle."), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVProcess, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVProcess, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVProcess, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVProcessTransport __pyx_vtable_6uvloop_4loop_UVProcessTransport;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVProcessTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVProcessTransport *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVProcess(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVProcessTransport;
  p->_exit_waiters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_init_futs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_pending_calls = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_protocol = Py_None; Py_INCREF(Py_None);
  p->_stdin = ((struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *)Py_None); Py_INCREF(Py_None);
  p->_stdout = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)Py_None); Py_INCREF(Py_None);
  p->_stderr = ((struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *)Py_None); Py_INCREF(Py_None);
  p->stdin_proto = Py_None; Py_INCREF(Py_None);
  p->stdout_proto = Py_None; Py_INCREF(Py_None);
  p->stderr_proto = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_18UVProcessTransport_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVProcessTransport(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVProcessTransport *p = (struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVProcessTransport) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_exit_waiters);
  Py_CLEAR(p->_init_futs);
  Py_CLEAR(p->_pending_calls);
  Py_CLEAR(p->_protocol);
  Py_CLEAR(p->_stdin);
  Py_CLEAR(p->_stdout);
  Py_CLEAR(p->_stderr);
  Py_CLEAR(p->stdin_proto);
  Py_CLEAR(p->stdout_proto);
  Py_CLEAR(p->stderr_proto);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVProcess(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVProcessTransport(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVProcessTransport *p = (struct __pyx_obj_6uvloop_4loop_UVProcessTransport *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVProcess(o, v, a); if (e) return e;
  if (p->_exit_waiters) {
    e = (*v)(p->_exit_waiters, a); if (e) return e;
  }
  if (p->_init_futs) {
    e = (*v)(p->_init_futs, a); if (e) return e;
  }
  if (p->_pending_calls) {
    e = (*v)(p->_pending_calls, a); if (e) return e;
  }
  if (p->_protocol) {
    e = (*v)(p->_protocol, a); if (e) return e;
  }
  if (p->_stdin) {
    e = (*v)(((PyObject *)p->_stdin), a); if (e) return e;
  }
  if (p->_stdout) {
    e = (*v)(((PyObject *)p->_stdout), a); if (e) return e;
  }
  if (p->_stderr) {
    e = (*v)(((PyObject *)p->_stderr), a); if (e) return e;
  }
  if (p->stdin_proto) {
    e = (*v)(p->stdin_proto, a); if (e) return e;
  }
  if (p->stdout_proto) {
    e = (*v)(p->stdout_proto, a); if (e) return e;
  }
  if (p->stderr_proto) {
    e = (*v)(p->stderr_proto, a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVProcessTransport[] = {
  {"_UVProcessTransport__stdio_inited", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_3_UVProcessTransport__stdio_inited, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_2__stdio_inited},
  {"get_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_5get_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_4get_protocol},
  {"set_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_7set_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_6set_protocol},
  {"get_pid", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_9get_pid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_8get_pid},
  {"get_returncode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_11get_returncode, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_10get_returncode},
  {"get_pipe_transport", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_13get_pipe_transport, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_12get_pipe_transport},
  {"terminate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_15terminate, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_14terminate},
  {"kill", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_17kill, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_16kill},
  {"send_signal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_19send_signal, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_18send_signal},
  {"is_closing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_21is_closing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_20is_closing},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_23close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_22close},
  {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_25get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_24get_extra_info},
  {"_wait", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_27_wait, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_26_wait},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_29__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_28__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18UVProcessTransport_31__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18UVProcessTransport_30__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVProcessTransport_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVProcessTransport},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVProcessTransport},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVProcessTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVProcessTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVProcessTransport_spec = {
  "uvloop.loop.UVProcessTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVProcessTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVProcessTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVProcessTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVProcessTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVProcessTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVProcessTransport, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVProcessTransport, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVProcessTransport, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVProcessTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVFSEvent __pyx_vtable_6uvloop_4loop_UVFSEvent;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVFSEvent(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVHandle(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UVFSEvent *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UVFSEvent;
  p->callback = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVFSEvent(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *p = (struct __pyx_obj_6uvloop_4loop_UVFSEvent *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVFSEvent) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->callback);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVHandle(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UVFSEvent(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVFSEvent *p = (struct __pyx_obj_6uvloop_4loop_UVFSEvent *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVHandle(o, v, a); if (e) return e;
  if (p->callback) {
    e = (*v)(p->callback, a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVFSEvent[] = {
  {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_1cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_cancel},
  {"cancelled", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_3cancelled, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_2cancelled},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVFSEvent_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVFSEvent_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVFSEvent_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVFSEvent},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVFSEvent},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVFSEvent},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVFSEvent},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVFSEvent_spec = {
  "uvloop.loop.UVFSEvent",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVFSEvent),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVFSEvent_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVFSEvent = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVFSEvent", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVFSEvent), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVFSEvent, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVFSEvent, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVFSEvent, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVFSEvent, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UVRequest __pyx_vtable_6uvloop_4loop_UVRequest;

static PyObject *__pyx_tp_new_6uvloop_4loop_UVRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UVRequest *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_UVRequest *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_UVRequest;
  p->loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_9UVRequest_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UVRequest(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UVRequest *p = (struct __pyx_obj_6uvloop_4loop_UVRequest *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UVRequest) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->loop);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop_UVRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UVRequest *p = (struct __pyx_obj_6uvloop_4loop_UVRequest *)o;
  if (p->loop) {
    e = (*v)(((PyObject *)p->loop), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_UVRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_UVRequest *p = (struct __pyx_obj_6uvloop_4loop_UVRequest *)o;
  tmp = ((PyObject*)p->loop);
  p->loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UVRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVRequest_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_9UVRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_9UVRequest_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UVRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UVRequest},
  {Py_tp_doc, (void *)PyDoc_STR("A base class for all libuv requests (uv_getaddrinfo_t, etc).\n\n    Important: it's a responsibility of the subclass to call the\n    \"on_done\" method in the request's callback.\n\n    If \"on_done\" isn't called, the request object will never die.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UVRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_UVRequest},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UVRequest},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UVRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UVRequest_spec = {
  "uvloop.loop.UVRequest",
  sizeof(struct __pyx_obj_6uvloop_4loop_UVRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UVRequest_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UVRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UVRequest", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UVRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UVRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("A base class for all libuv requests (uv_getaddrinfo_t, etc).\n\n    Important: it's a responsibility of the subclass to call the\n    \"on_done\" method in the request's callback.\n\n    If \"on_done\" isn't called, the request object will never die.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UVRequest, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_UVRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UVRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UVRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop__SSLProtocolTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)o);
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  p->_ssl_protocol = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)Py_None); Py_INCREF(Py_None);
  p->context = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop__SSLProtocolTransport(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *p = (struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop__SSLProtocolTransport) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_13__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_loop);
  Py_CLEAR(p->_ssl_protocol);
  Py_CLEAR(p->context);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop__SSLProtocolTransport(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *p = (struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)o;
  if (p->_loop) {
    e = (*v)(((PyObject *)p->_loop), a); if (e) return e;
  }
  if (p->_ssl_protocol) {
    e = (*v)(((PyObject *)p->_ssl_protocol), a); if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop__SSLProtocolTransport(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *p = (struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)o;
  tmp = ((PyObject*)p->_loop);
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ssl_protocol);
  p->_ssl_protocol = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->context);
  p->context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_21_SSLProtocolTransport__protocol_paused(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_16_protocol_paused_1__get__(o);
}

static PyMethodDef __pyx_methods_6uvloop_4loop__SSLProtocolTransport[] = {
  {"get_extra_info", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_3get_extra_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_2get_extra_info},
  {"set_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_5set_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_4set_protocol},
  {"get_protocol", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_7get_protocol, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_6get_protocol},
  {"is_closing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_9is_closing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_8is_closing},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_11close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_10close},
  {"is_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_15is_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_14is_reading},
  {"pause_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_17pause_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_16pause_reading},
  {"resume_reading", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_19resume_reading, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_18resume_reading},
  {"set_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_21set_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_20set_write_buffer_limits},
  {"get_write_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_23get_write_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_22get_write_buffer_limits},
  {"get_write_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_25get_write_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_24get_write_buffer_size},
  {"set_read_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_27set_read_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_26set_read_buffer_limits},
  {"get_read_buffer_limits", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_29get_read_buffer_limits, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_28get_read_buffer_limits},
  {"get_read_buffer_size", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_31get_read_buffer_size, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_30get_read_buffer_size},
  {"write", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_33write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_32write},
  {"writelines", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_35writelines, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_34writelines},
  {"write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_37write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_36write_eof},
  {"can_write_eof", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_39can_write_eof, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_38can_write_eof},
  {"abort", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_41abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_40abort},
  {"_force_close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_43_force_close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_42_force_close},
  {"_test__append_write_backlog", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_45_test__append_write_backlog, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_44_test__append_write_backlog},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_47__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_46__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_21_SSLProtocolTransport_49__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_21_SSLProtocolTransport_48__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop__SSLProtocolTransport[] = {
  {(char *)"_protocol_paused", __pyx_getprop_6uvloop_4loop_21_SSLProtocolTransport__protocol_paused, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop__SSLProtocolTransport_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop__SSLProtocolTransport},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop__SSLProtocolTransport},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop__SSLProtocolTransport},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop__SSLProtocolTransport},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop__SSLProtocolTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop__SSLProtocolTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop__SSLProtocolTransport_spec = {
  "uvloop.loop._SSLProtocolTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop__SSLProtocolTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop__SSLProtocolTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""_SSLProtocolTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop__SSLProtocolTransport, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop__SSLProtocolTransport, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop__SSLProtocolTransport, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop__SSLProtocolTransport, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop__SSLProtocolTransport, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop__SSLProtocolTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_SSLProtocol __pyx_vtable_6uvloop_4loop_SSLProtocol;

static PyObject *__pyx_tp_new_6uvloop_4loop_SSLProtocol(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_SSLProtocol;
  p->_server_hostname = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_sslcontext = Py_None; Py_INCREF(Py_None);
  p->_extra = Py_None; Py_INCREF(Py_None);
  p->_write_backlog = Py_None; Py_INCREF(Py_None);
  p->_waiter = Py_None; Py_INCREF(Py_None);
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  p->_app_transport = ((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)Py_None); Py_INCREF(Py_None);
  p->_transport = Py_None; Py_INCREF(Py_None);
  p->_ssl_handshake_timeout = Py_None; Py_INCREF(Py_None);
  p->_ssl_shutdown_timeout = Py_None; Py_INCREF(Py_None);
  p->_sslobj = Py_None; Py_INCREF(Py_None);
  p->_sslobj_read = Py_None; Py_INCREF(Py_None);
  p->_sslobj_write = Py_None; Py_INCREF(Py_None);
  p->_incoming = Py_None; Py_INCREF(Py_None);
  p->_incoming_write = Py_None; Py_INCREF(Py_None);
  p->_outgoing = Py_None; Py_INCREF(Py_None);
  p->_outgoing_read = Py_None; Py_INCREF(Py_None);
  p->_ssl_buffer_view = Py_None; Py_INCREF(Py_None);
  p->_app_protocol = Py_None; Py_INCREF(Py_None);
  p->_app_protocol_get_buffer = Py_None; Py_INCREF(Py_None);
  p->_app_protocol_buffer_updated = Py_None; Py_INCREF(Py_None);
  p->_handshake_start_time = Py_None; Py_INCREF(Py_None);
  p->_handshake_timeout_handle = Py_None; Py_INCREF(Py_None);
  p->_shutdown_timeout_handle = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_11SSLProtocol_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_SSLProtocol(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *p = (struct __pyx_obj_6uvloop_4loop_SSLProtocol *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_SSLProtocol) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_11SSLProtocol_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_server_hostname);
  Py_CLEAR(p->_sslcontext);
  Py_CLEAR(p->_extra);
  Py_CLEAR(p->_write_backlog);
  Py_CLEAR(p->_waiter);
  Py_CLEAR(p->_loop);
  Py_CLEAR(p->_app_transport);
  Py_CLEAR(p->_transport);
  Py_CLEAR(p->_ssl_handshake_timeout);
  Py_CLEAR(p->_ssl_shutdown_timeout);
  Py_CLEAR(p->_sslobj);
  Py_CLEAR(p->_sslobj_read);
  Py_CLEAR(p->_sslobj_write);
  Py_CLEAR(p->_incoming);
  Py_CLEAR(p->_incoming_write);
  Py_CLEAR(p->_outgoing);
  Py_CLEAR(p->_outgoing_read);
  Py_CLEAR(p->_ssl_buffer_view);
  Py_CLEAR(p->_app_protocol);
  Py_CLEAR(p->_app_protocol_get_buffer);
  Py_CLEAR(p->_app_protocol_buffer_updated);
  Py_CLEAR(p->_handshake_start_time);
  Py_CLEAR(p->_handshake_timeout_handle);
  Py_CLEAR(p->_shutdown_timeout_handle);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop_SSLProtocol(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *p = (struct __pyx_obj_6uvloop_4loop_SSLProtocol *)o;
  if (p->_sslcontext) {
    e = (*v)(p->_sslcontext, a); if (e) return e;
  }
  if (p->_extra) {
    e = (*v)(p->_extra, a); if (e) return e;
  }
  if (p->_write_backlog) {
    e = (*v)(p->_write_backlog, a); if (e) return e;
  }
  if (p->_waiter) {
    e = (*v)(p->_waiter, a); if (e) return e;
  }
  if (p->_loop) {
    e = (*v)(((PyObject *)p->_loop), a); if (e) return e;
  }
  if (p->_app_transport) {
    e = (*v)(((PyObject *)p->_app_transport), a); if (e) return e;
  }
  if (p->_transport) {
    e = (*v)(p->_transport, a); if (e) return e;
  }
  if (p->_ssl_handshake_timeout) {
    e = (*v)(p->_ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->_ssl_shutdown_timeout) {
    e = (*v)(p->_ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->_sslobj) {
    e = (*v)(p->_sslobj, a); if (e) return e;
  }
  if (p->_sslobj_read) {
    e = (*v)(p->_sslobj_read, a); if (e) return e;
  }
  if (p->_sslobj_write) {
    e = (*v)(p->_sslobj_write, a); if (e) return e;
  }
  if (p->_incoming) {
    e = (*v)(p->_incoming, a); if (e) return e;
  }
  if (p->_incoming_write) {
    e = (*v)(p->_incoming_write, a); if (e) return e;
  }
  if (p->_outgoing) {
    e = (*v)(p->_outgoing, a); if (e) return e;
  }
  if (p->_outgoing_read) {
    e = (*v)(p->_outgoing_read, a); if (e) return e;
  }
  if (p->_ssl_buffer_view) {
    e = (*v)(p->_ssl_buffer_view, a); if (e) return e;
  }
  if (p->_app_protocol) {
    e = (*v)(p->_app_protocol, a); if (e) return e;
  }
  if (p->_app_protocol_get_buffer) {
    e = (*v)(p->_app_protocol_get_buffer, a); if (e) return e;
  }
  if (p->_app_protocol_buffer_updated) {
    e = (*v)(p->_app_protocol_buffer_updated, a); if (e) return e;
  }
  if (p->_handshake_start_time) {
    e = (*v)(p->_handshake_start_time, a); if (e) return e;
  }
  if (p->_handshake_timeout_handle) {
    e = (*v)(p->_handshake_timeout_handle, a); if (e) return e;
  }
  if (p->_shutdown_timeout_handle) {
    e = (*v)(p->_shutdown_timeout_handle, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_SSLProtocol(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_SSLProtocol *p = (struct __pyx_obj_6uvloop_4loop_SSLProtocol *)o;
  tmp = ((PyObject*)p->_sslcontext);
  p->_sslcontext = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_extra);
  p->_extra = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_write_backlog);
  p->_write_backlog = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_waiter);
  p->_waiter = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_loop);
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_app_transport);
  p->_app_transport = ((struct __pyx_obj_6uvloop_4loop__SSLProtocolTransport *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_transport);
  p->_transport = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ssl_handshake_timeout);
  p->_ssl_handshake_timeout = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ssl_shutdown_timeout);
  p->_ssl_shutdown_timeout = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_sslobj);
  p->_sslobj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_sslobj_read);
  p->_sslobj_read = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_sslobj_write);
  p->_sslobj_write = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_incoming);
  p->_incoming = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_incoming_write);
  p->_incoming_write = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_outgoing);
  p->_outgoing = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_outgoing_read);
  p->_outgoing_read = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ssl_buffer_view);
  p->_ssl_buffer_view = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_app_protocol);
  p->_app_protocol = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_app_protocol_get_buffer);
  p->_app_protocol_get_buffer = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_app_protocol_buffer_updated);
  p->_app_protocol_buffer_updated = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_handshake_start_time);
  p->_handshake_start_time = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_handshake_timeout_handle);
  p->_handshake_timeout_handle = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_shutdown_timeout_handle);
  p->_shutdown_timeout_handle = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_SSLProtocol[] = {
  {"_get_app_transport", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_7_get_app_transport, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_6_get_app_transport},
  {"connection_made", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_9connection_made, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_8connection_made},
  {"connection_lost", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_11connection_lost, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_10connection_lost},
  {"get_buffer", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_13get_buffer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_12get_buffer},
  {"buffer_updated", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_15buffer_updated, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_14buffer_updated},
  {"eof_received", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_17eof_received, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_16eof_received},
  {"pause_writing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_19pause_writing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_18pause_writing},
  {"resume_writing", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_21resume_writing, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_20resume_writing},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_22__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_11SSLProtocol_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_11SSLProtocol_24__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_SSLProtocol_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_SSLProtocol},
  {Py_tp_doc, (void *)PyDoc_STR("SSLProtocol(loop, app_protocol, sslcontext, waiter, server_side=False, server_hostname=None, call_connection_made=True, ssl_handshake_timeout=None, ssl_shutdown_timeout=None)\nSSL protocol.\n\n    Implementation of SSL on top of a socket using incoming and outgoing\n    buffers which are ssl.MemoryBIO objects.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_SSLProtocol},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_SSLProtocol},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_SSLProtocol},
  {Py_tp_init, (void *)__pyx_pw_6uvloop_4loop_11SSLProtocol_5__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_SSLProtocol},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_SSLProtocol_spec = {
  "uvloop.loop.SSLProtocol",
  sizeof(struct __pyx_obj_6uvloop_4loop_SSLProtocol),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_SSLProtocol_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_SSLProtocol = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""SSLProtocol", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_SSLProtocol), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_SSLProtocol, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("SSLProtocol(loop, app_protocol, sslcontext, waiter, server_side=False, server_hostname=None, call_connection_made=True, ssl_handshake_timeout=None, ssl_shutdown_timeout=None)\nSSL protocol.\n\n    Implementation of SSL on top of a socket using incoming and outgoing\n    buffers which are ssl.MemoryBIO objects.\n    "), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_SSLProtocol, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_SSLProtocol, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_SSLProtocol, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6uvloop_4loop_11SSLProtocol_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_SSLProtocol, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_UDPTransport __pyx_vtable_6uvloop_4loop_UDPTransport;

static PyObject *__pyx_tp_new_6uvloop_4loop_UDPTransport(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_UDPTransport *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVBaseTransport(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVHandle*)__pyx_vtabptr_6uvloop_4loop_UDPTransport;
  p->_address = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_12UDPTransport_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_UDPTransport(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_UDPTransport *p = (struct __pyx_obj_6uvloop_4loop_UDPTransport *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_UDPTransport) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_address);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVBaseTransport(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_UDPTransport(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_UDPTransport *p = (struct __pyx_obj_6uvloop_4loop_UDPTransport *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVBaseTransport(o, v, a); if (e) return e;
  if (p->_address) {
    e = (*v)(p->_address, a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_UDPTransport[] = {
  {"sendto", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12UDPTransport_3sendto, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12UDPTransport_2sendto},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12UDPTransport_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12UDPTransport_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12UDPTransport_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12UDPTransport_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_UDPTransport_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_UDPTransport},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_UDPTransport},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_UDPTransport},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_UDPTransport},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_UDPTransport_spec = {
  "uvloop.loop.UDPTransport",
  sizeof(struct __pyx_obj_6uvloop_4loop_UDPTransport),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_UDPTransport_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_UDPTransport = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""UDPTransport", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_UDPTransport), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_UDPTransport, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_7__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_UDPTransport, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_UDPTransport, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_6uvloop_4loop_8UVHandle_3__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_UDPTransport, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_Server __pyx_vtable_6uvloop_4loop_Server;

static PyObject *__pyx_tp_new_6uvloop_4loop_Server(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_Server *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_Server *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_Server;
  p->_servers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_waiters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  p->_serving_forever_fut = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_6Server_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_Server(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_Server *p = (struct __pyx_obj_6uvloop_4loop_Server *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_Server) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (p->__weakref__) PyObject_ClearWeakRefs(o);
  Py_CLEAR(p->_servers);
  Py_CLEAR(p->_waiters);
  Py_CLEAR(p->_loop);
  Py_CLEAR(p->_serving_forever_fut);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop_Server(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_Server *p = (struct __pyx_obj_6uvloop_4loop_Server *)o;
  if (p->_servers) {
    e = (*v)(p->_servers, a); if (e) return e;
  }
  if (p->_waiters) {
    e = (*v)(p->_waiters, a); if (e) return e;
  }
  if (p->_loop) {
    e = (*v)(((PyObject *)p->_loop), a); if (e) return e;
  }
  if (p->_serving_forever_fut) {
    e = (*v)(p->_serving_forever_fut, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_Server(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_Server *p = (struct __pyx_obj_6uvloop_4loop_Server *)o;
  tmp = ((PyObject*)p->_servers);
  p->_servers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_waiters);
  p->_waiters = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_loop);
  p->_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_serving_forever_fut);
  p->_serving_forever_fut = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_6Server_sockets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_6Server_7sockets_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6uvloop_4loop_6Server_9__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6uvloop_4loop_6Server_9__repr__(self);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_Server[] = {
  {"__aenter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_3__aenter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_2__aenter__},
  {"__aexit__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_6Server_6__aexit__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_5__aexit__},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6uvloop_4loop_6Server_9__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"get_loop", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_11get_loop, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_10get_loop},
  {"wait_closed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_13wait_closed, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_12wait_closed},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_16close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_15close},
  {"is_serving", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_18is_serving, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_17is_serving},
  {"start_serving", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_20start_serving, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_19start_serving},
  {"serve_forever", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_23serve_forever, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_22serve_forever},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_26__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_25__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_6Server_28__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_6Server_27__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop_Server[] = {
  {(char *)"sockets", __pyx_getprop_6uvloop_4loop_6Server_sockets, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_Server_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_Server},
  {Py_tp_repr, (void *)__pyx_pw_6uvloop_4loop_6Server_9__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_Server},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_Server},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_Server},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop_Server},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_Server},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_Server_spec = {
  "uvloop.loop.Server",
  sizeof(struct __pyx_obj_6uvloop_4loop_Server),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_Server_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_Server = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""Server", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_Server), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_Server, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6uvloop_4loop_6Server_9__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_Server, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_Server, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_Server, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop_Server, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_Server, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_PseudoSocket __pyx_vtable_6uvloop_4loop_PseudoSocket;

static PyObject *__pyx_tp_new_6uvloop_4loop_PseudoSocket(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_PseudoSocket *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_PseudoSocket *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_PseudoSocket;
  p->_peername = Py_None; Py_INCREF(Py_None);
  p->_sockname = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_PseudoSocket(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_PseudoSocket *p = (struct __pyx_obj_6uvloop_4loop_PseudoSocket *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_PseudoSocket) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_peername);
  Py_CLEAR(p->_sockname);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop_PseudoSocket(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_PseudoSocket *p = (struct __pyx_obj_6uvloop_4loop_PseudoSocket *)o;
  if (p->_peername) {
    e = (*v)(p->_peername, a); if (e) return e;
  }
  if (p->_sockname) {
    e = (*v)(p->_sockname, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_PseudoSocket(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_PseudoSocket *p = (struct __pyx_obj_6uvloop_4loop_PseudoSocket *)o;
  tmp = ((PyObject*)p->_peername);
  p->_peername = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_sockname);
  p->_sockname = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6uvloop_4loop_12PseudoSocket_family(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_12PseudoSocket_6family_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_12PseudoSocket_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_12PseudoSocket_4type_1__get__(o);
}

static PyObject *__pyx_getprop_6uvloop_4loop_12PseudoSocket_proto(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6uvloop_4loop_12PseudoSocket_5proto_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6uvloop_4loop_12PseudoSocket_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_6uvloop_4loop_12PseudoSocket_3__repr__(self);
}

static PyMethodDef __pyx_methods_6uvloop_4loop_PseudoSocket[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_6uvloop_4loop_12PseudoSocket_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__getstate__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_5__getstate__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_4__getstate__},
  {"fileno", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_7fileno, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_6fileno},
  {"dup", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_9dup, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_8dup},
  {"get_inheritable", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_11get_inheritable, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_10get_inheritable},
  {"set_inheritable", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_13set_inheritable, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_12set_inheritable},
  {"ioctl", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_15ioctl, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_14ioctl},
  {"getsockopt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_17getsockopt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_16getsockopt},
  {"setsockopt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_19setsockopt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_18setsockopt},
  {"getpeername", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_21getpeername, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_20getpeername},
  {"getsockname", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_23getsockname, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_22getsockname},
  {"share", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_25share, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_24share},
  {"accept", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_27accept, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_26accept},
  {"connect", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_29connect, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_28connect},
  {"connect_ex", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_31connect_ex, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_30connect_ex},
  {"bind", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_33bind, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_32bind},
  {"listen", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_35listen, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_34listen},
  {"makefile", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_37makefile, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_36makefile},
  {"sendfile", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_39sendfile, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_38sendfile},
  {"close", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_41close, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_40close},
  {"detach", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_43detach, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_42detach},
  {"shutdown", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_45shutdown, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_44shutdown},
  {"sendmsg_afalg", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_47sendmsg_afalg, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_46sendmsg_afalg},
  {"sendmsg", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_49sendmsg, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_48sendmsg},
  {"sendto", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_51sendto, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_50sendto},
  {"send", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_53send, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_52send},
  {"sendall", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_55sendall, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_54sendall},
  {"recv_into", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_57recv_into, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_56recv_into},
  {"recvfrom_into", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_59recvfrom_into, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_58recvfrom_into},
  {"recvmsg_into", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_61recvmsg_into, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_60recvmsg_into},
  {"recvmsg", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_63recvmsg, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_62recvmsg},
  {"recvfrom", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_65recvfrom, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_64recvfrom},
  {"recv", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_67recv, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_66recv},
  {"settimeout", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_69settimeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_68settimeout},
  {"gettimeout", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_71gettimeout, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_70gettimeout},
  {"setblocking", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_73setblocking, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_72setblocking},
  {"__enter__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_75__enter__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_74__enter__},
  {"__exit__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_77__exit__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_76__exit__},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_79__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_78__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_12PseudoSocket_81__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_12PseudoSocket_80__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6uvloop_4loop_PseudoSocket[] = {
  {(char *)"family", __pyx_getprop_6uvloop_4loop_12PseudoSocket_family, 0, (char *)0, 0},
  {(char *)"type", __pyx_getprop_6uvloop_4loop_12PseudoSocket_type, 0, (char *)0, 0},
  {(char *)"proto", __pyx_getprop_6uvloop_4loop_12PseudoSocket_proto, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_PseudoSocket_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_PseudoSocket},
  {Py_tp_repr, (void *)__pyx_pw_6uvloop_4loop_12PseudoSocket_3__repr__},
  {Py_tp_doc, (void *)PyDoc_STR("PseudoSocket(int family, int type, int proto, int fd)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_PseudoSocket},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_PseudoSocket},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_PseudoSocket},
  {Py_tp_getset, (void *)__pyx_getsets_6uvloop_4loop_PseudoSocket},
  {Py_tp_init, (void *)__pyx_pw_6uvloop_4loop_12PseudoSocket_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_PseudoSocket},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_PseudoSocket_spec = {
  "uvloop.loop.PseudoSocket",
  sizeof(struct __pyx_obj_6uvloop_4loop_PseudoSocket),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_PseudoSocket_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_PseudoSocket = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""PseudoSocket", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_PseudoSocket), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_PseudoSocket, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6uvloop_4loop_12PseudoSocket_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("PseudoSocket(int family, int type, int proto, int fd)"), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_PseudoSocket, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_PseudoSocket, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_PseudoSocket, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6uvloop_4loop_PseudoSocket, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6uvloop_4loop_12PseudoSocket_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_PseudoSocket, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_LruCache __pyx_vtable_6uvloop_4loop_LruCache;

static PyObject *__pyx_tp_new_6uvloop_4loop_LruCache(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_LruCache *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_LruCache *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_LruCache;
  p->_dict = Py_None; Py_INCREF(Py_None);
  p->_dict_move_to_end = Py_None; Py_INCREF(Py_None);
  p->_dict_get = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_LruCache(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_LruCache *p = (struct __pyx_obj_6uvloop_4loop_LruCache *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_dict);
  Py_CLEAR(p->_dict_move_to_end);
  Py_CLEAR(p->_dict_get);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_6uvloop_4loop_LruCache(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_LruCache *p = (struct __pyx_obj_6uvloop_4loop_LruCache *)o;
  if (p->_dict) {
    e = (*v)(p->_dict, a); if (e) return e;
  }
  if (p->_dict_move_to_end) {
    e = (*v)(p->_dict_move_to_end, a); if (e) return e;
  }
  if (p->_dict_get) {
    e = (*v)(p->_dict_get, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_LruCache(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_LruCache *p = (struct __pyx_obj_6uvloop_4loop_LruCache *)o;
  tmp = ((PyObject*)p->_dict);
  p->_dict = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_dict_move_to_end);
  p->_dict_move_to_end = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_dict_get);
  p->_dict_get = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6uvloop_4loop_LruCache(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_6uvloop_4loop_LruCache(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_6uvloop_4loop_8LruCache_5__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_6uvloop_4loop_8LruCache_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_6uvloop_4loop_LruCache[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8LruCache_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8LruCache_14__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8LruCache_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8LruCache_16__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_LruCache_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_LruCache},
  {Py_sq_length, (void *)__pyx_pw_6uvloop_4loop_8LruCache_11__len__},
  {Py_sq_item, (void *)__pyx_sq_item_6uvloop_4loop_LruCache},
  {Py_sq_contains, (void *)__pyx_pw_6uvloop_4loop_8LruCache_9__contains__},
  {Py_mp_length, (void *)__pyx_pw_6uvloop_4loop_8LruCache_11__len__},
  {Py_mp_subscript, (void *)__pyx_pw_6uvloop_4loop_8LruCache_3__getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_6uvloop_4loop_LruCache},
  {Py_tp_doc, (void *)PyDoc_STR("LruCache(maxsize, *)")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_LruCache},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_LruCache},
  {Py_tp_iter, (void *)__pyx_pw_6uvloop_4loop_8LruCache_13__iter__},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_LruCache},
  {Py_tp_init, (void *)__pyx_pw_6uvloop_4loop_8LruCache_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_LruCache},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_LruCache_spec = {
  "uvloop.loop.LruCache",
  sizeof(struct __pyx_obj_6uvloop_4loop_LruCache),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_LruCache_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_LruCache = {
  __pyx_pw_6uvloop_4loop_8LruCache_11__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6uvloop_4loop_LruCache, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_6uvloop_4loop_8LruCache_9__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_LruCache = {
  __pyx_pw_6uvloop_4loop_8LruCache_11__len__, /*mp_length*/
  __pyx_pw_6uvloop_4loop_8LruCache_3__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_6uvloop_4loop_LruCache, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6uvloop_4loop_LruCache = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""LruCache", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_LruCache), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_LruCache, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_LruCache, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_LruCache, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("LruCache(maxsize, *)"), /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_LruCache, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_LruCache, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_6uvloop_4loop_8LruCache_13__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_LruCache, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_6uvloop_4loop_8LruCache_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_LruCache, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop__StreamWriteContext __pyx_vtable_6uvloop_4loop__StreamWriteContext;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop__StreamWriteContext *__pyx_freelist_6uvloop_4loop__StreamWriteContext[250];
static int __pyx_freecount_6uvloop_4loop__StreamWriteContext = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop__StreamWriteContext(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop__StreamWriteContext > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop__StreamWriteContext)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop__StreamWriteContext[--__pyx_freecount_6uvloop_4loop__StreamWriteContext];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop__StreamWriteContext));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop__StreamWriteContext;
  p->buffers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->stream = ((struct __pyx_obj_6uvloop_4loop_UVStream *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop__StreamWriteContext(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *p = (struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop__StreamWriteContext) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_19_StreamWriteContext_1__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->buffers);
  Py_CLEAR(p->stream);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop__StreamWriteContext < 250) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop__StreamWriteContext)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_6uvloop_4loop__StreamWriteContext[__pyx_freecount_6uvloop_4loop__StreamWriteContext++] = ((struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop__StreamWriteContext(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop__StreamWriteContext *p = (struct __pyx_obj_6uvloop_4loop__StreamWriteContext *)o;
  if (p->buffers) {
    e = (*v)(p->buffers, a); if (e) return e;
  }
  if (p->stream) {
    e = (*v)(((PyObject *)p->stream), a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop__StreamWriteContext[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_StreamWriteContext_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_StreamWriteContext_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_StreamWriteContext_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_StreamWriteContext_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop__StreamWriteContext_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop__StreamWriteContext},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop__StreamWriteContext},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop__StreamWriteContext},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop__StreamWriteContext},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop__StreamWriteContext_spec = {
  "uvloop.loop._StreamWriteContext",
  sizeof(struct __pyx_obj_6uvloop_4loop__StreamWriteContext),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop__StreamWriteContext_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop__StreamWriteContext = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""_StreamWriteContext", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop__StreamWriteContext), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop__StreamWriteContext, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop__StreamWriteContext, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop__StreamWriteContext, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop__StreamWriteContext, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop__TCPConnectRequest __pyx_vtable_6uvloop_4loop__TCPConnectRequest;

static PyObject *__pyx_tp_new_6uvloop_4loop__TCPConnectRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVRequest*)__pyx_vtabptr_6uvloop_4loop__TCPConnectRequest;
  p->transport = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop__TCPConnectRequest(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *p = (struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop__TCPConnectRequest) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->transport);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVRequest(o);
}

static int __pyx_tp_traverse_6uvloop_4loop__TCPConnectRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *p = (struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVRequest(o, v, a); if (e) return e;
  if (p->transport) {
    e = (*v)(((PyObject *)p->transport), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop__TCPConnectRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *p = (struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *)o;
  __pyx_tp_clear_6uvloop_4loop_UVRequest(o);
  tmp = ((PyObject*)p->transport);
  p->transport = ((struct __pyx_obj_6uvloop_4loop_TCPTransport *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop__TCPConnectRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18_TCPConnectRequest_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_18_TCPConnectRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_18_TCPConnectRequest_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop__TCPConnectRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop__TCPConnectRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop__TCPConnectRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop__TCPConnectRequest},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop__TCPConnectRequest},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop__TCPConnectRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop__TCPConnectRequest_spec = {
  "uvloop.loop._TCPConnectRequest",
  sizeof(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop__TCPConnectRequest_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop__TCPConnectRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""_TCPConnectRequest", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop__TCPConnectRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop__TCPConnectRequest, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop__TCPConnectRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop__TCPConnectRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop__TCPConnectRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop__PipeConnectRequest __pyx_vtable_6uvloop_4loop__PipeConnectRequest;

static PyObject *__pyx_tp_new_6uvloop_4loop__PipeConnectRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVRequest*)__pyx_vtabptr_6uvloop_4loop__PipeConnectRequest;
  p->transport = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop__PipeConnectRequest(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *p = (struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop__PipeConnectRequest) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->transport);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVRequest(o);
}

static int __pyx_tp_traverse_6uvloop_4loop__PipeConnectRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *p = (struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVRequest(o, v, a); if (e) return e;
  if (p->transport) {
    e = (*v)(((PyObject *)p->transport), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop__PipeConnectRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *p = (struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *)o;
  __pyx_tp_clear_6uvloop_4loop_UVRequest(o);
  tmp = ((PyObject*)p->transport);
  p->transport = ((struct __pyx_obj_6uvloop_4loop_UnixTransport *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop__PipeConnectRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_PipeConnectRequest_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_19_PipeConnectRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_19_PipeConnectRequest_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop__PipeConnectRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop__PipeConnectRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop__PipeConnectRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop__PipeConnectRequest},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop__PipeConnectRequest},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop__PipeConnectRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop__PipeConnectRequest_spec = {
  "uvloop.loop._PipeConnectRequest",
  sizeof(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop__PipeConnectRequest_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop__PipeConnectRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""_PipeConnectRequest", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop__PipeConnectRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop__PipeConnectRequest, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop__PipeConnectRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop__PipeConnectRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop__PipeConnectRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop_SockAddrHolder *__pyx_freelist_6uvloop_4loop_SockAddrHolder[250];
static int __pyx_freecount_6uvloop_4loop_SockAddrHolder = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop_SockAddrHolder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop_SockAddrHolder > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_SockAddrHolder)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop_SockAddrHolder[--__pyx_freecount_6uvloop_4loop_SockAddrHolder];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop_SockAddrHolder));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop_SockAddrHolder(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_SockAddrHolder) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop_SockAddrHolder < 250) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_SockAddrHolder)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_6uvloop_4loop_SockAddrHolder[__pyx_freecount_6uvloop_4loop_SockAddrHolder++] = ((struct __pyx_obj_6uvloop_4loop_SockAddrHolder *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static PyMethodDef __pyx_methods_6uvloop_4loop_SockAddrHolder[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14SockAddrHolder_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14SockAddrHolder___reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_14SockAddrHolder_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_14SockAddrHolder_2__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_SockAddrHolder_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_SockAddrHolder},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_SockAddrHolder},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_SockAddrHolder},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_SockAddrHolder_spec = {
  "uvloop.loop.SockAddrHolder",
  sizeof(struct __pyx_obj_6uvloop_4loop_SockAddrHolder),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_6uvloop_4loop_SockAddrHolder_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_SockAddrHolder = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""SockAddrHolder", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_SockAddrHolder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_SockAddrHolder, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_SockAddrHolder, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_SockAddrHolder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_AddrInfo __pyx_vtable_6uvloop_4loop_AddrInfo;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop_AddrInfo *__pyx_freelist_6uvloop_4loop_AddrInfo[250];
static int __pyx_freecount_6uvloop_4loop_AddrInfo = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop_AddrInfo(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_AddrInfo *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop_AddrInfo > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_AddrInfo)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop_AddrInfo[--__pyx_freecount_6uvloop_4loop_AddrInfo];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop_AddrInfo));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop_AddrInfo;
  if (unlikely(__pyx_pw_6uvloop_4loop_8AddrInfo_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_AddrInfo(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_AddrInfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_8AddrInfo_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop_AddrInfo < 250) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop_AddrInfo)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_6uvloop_4loop_AddrInfo[__pyx_freecount_6uvloop_4loop_AddrInfo++] = ((struct __pyx_obj_6uvloop_4loop_AddrInfo *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static PyMethodDef __pyx_methods_6uvloop_4loop_AddrInfo[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8AddrInfo_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8AddrInfo_4__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_8AddrInfo_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_8AddrInfo_6__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_AddrInfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_AddrInfo},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_AddrInfo},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_AddrInfo},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_AddrInfo_spec = {
  "uvloop.loop.AddrInfo",
  sizeof(struct __pyx_obj_6uvloop_4loop_AddrInfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_6uvloop_4loop_AddrInfo_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_AddrInfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""AddrInfo", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_AddrInfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_AddrInfo, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_AddrInfo, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_AddrInfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_AddrInfoRequest __pyx_vtable_6uvloop_4loop_AddrInfoRequest;

static PyObject *__pyx_tp_new_6uvloop_4loop_AddrInfoRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVRequest*)__pyx_vtabptr_6uvloop_4loop_AddrInfoRequest;
  p->callback = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_15AddrInfoRequest_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_AddrInfoRequest(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *p = (struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_AddrInfoRequest) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->callback);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVRequest(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_AddrInfoRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *p = (struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVRequest(o, v, a); if (e) return e;
  if (p->callback) {
    e = (*v)(p->callback, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_AddrInfoRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *p = (struct __pyx_obj_6uvloop_4loop_AddrInfoRequest *)o;
  __pyx_tp_clear_6uvloop_4loop_UVRequest(o);
  tmp = ((PyObject*)p->callback);
  p->callback = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_AddrInfoRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15AddrInfoRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15AddrInfoRequest_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15AddrInfoRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15AddrInfoRequest_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_AddrInfoRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_AddrInfoRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_AddrInfoRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_AddrInfoRequest},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_AddrInfoRequest},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_AddrInfoRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_AddrInfoRequest_spec = {
  "uvloop.loop.AddrInfoRequest",
  sizeof(struct __pyx_obj_6uvloop_4loop_AddrInfoRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_AddrInfoRequest_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_AddrInfoRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""AddrInfoRequest", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_AddrInfoRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_AddrInfoRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_AddrInfoRequest, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_AddrInfoRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_AddrInfoRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_AddrInfoRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop_NameInfoRequest __pyx_vtable_6uvloop_4loop_NameInfoRequest;

static PyObject *__pyx_tp_new_6uvloop_4loop_NameInfoRequest(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6uvloop_4loop_NameInfoRequest *p;
  PyObject *o = __pyx_tp_new_6uvloop_4loop_UVRequest(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_6uvloop_4loop_UVRequest*)__pyx_vtabptr_6uvloop_4loop_NameInfoRequest;
  p->callback = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_6uvloop_4loop_15NameInfoRequest_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6uvloop_4loop_NameInfoRequest(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop_NameInfoRequest *p = (struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop_NameInfoRequest) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->callback);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6uvloop_4loop_UVRequest(o);
}

static int __pyx_tp_traverse_6uvloop_4loop_NameInfoRequest(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop_NameInfoRequest *p = (struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)o;
  e = __pyx_tp_traverse_6uvloop_4loop_UVRequest(o, v, a); if (e) return e;
  if (p->callback) {
    e = (*v)(p->callback, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop_NameInfoRequest(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop_NameInfoRequest *p = (struct __pyx_obj_6uvloop_4loop_NameInfoRequest *)o;
  __pyx_tp_clear_6uvloop_4loop_UVRequest(o);
  tmp = ((PyObject*)p->callback);
  p->callback = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop_NameInfoRequest[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15NameInfoRequest_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15NameInfoRequest_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15NameInfoRequest_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15NameInfoRequest_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop_NameInfoRequest_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop_NameInfoRequest},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop_NameInfoRequest},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop_NameInfoRequest},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop_NameInfoRequest},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop_NameInfoRequest},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop_NameInfoRequest_spec = {
  "uvloop.loop.NameInfoRequest",
  sizeof(struct __pyx_obj_6uvloop_4loop_NameInfoRequest),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop_NameInfoRequest_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop_NameInfoRequest = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""NameInfoRequest", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop_NameInfoRequest), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop_NameInfoRequest, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop_NameInfoRequest, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop_NameInfoRequest, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop_NameInfoRequest, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop_NameInfoRequest, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_6uvloop_4loop__UDPSendContext __pyx_vtable_6uvloop_4loop__UDPSendContext;

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop__UDPSendContext *__pyx_freelist_6uvloop_4loop__UDPSendContext[250];
static int __pyx_freecount_6uvloop_4loop__UDPSendContext = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop__UDPSendContext(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop__UDPSendContext > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop__UDPSendContext)) & (int)(!__Pyx_PyType_HasFeature(t, (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop__UDPSendContext[--__pyx_freecount_6uvloop_4loop__UDPSendContext];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop__UDPSendContext));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)o);
  p->__pyx_vtab = __pyx_vtabptr_6uvloop_4loop__UDPSendContext;
  p->udp = ((struct __pyx_obj_6uvloop_4loop_UDPTransport *)Py_None); Py_INCREF(Py_None);
  p->py_buf.obj = NULL;
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop__UDPSendContext(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *p = (struct __pyx_obj_6uvloop_4loop__UDPSendContext *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop__UDPSendContext) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_6uvloop_4loop_15_UDPSendContext_1__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->udp);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop__UDPSendContext < 250) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop__UDPSendContext)) & (int)(!__Pyx_PyType_HasFeature(Py_TYPE(o), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE))))) {
    __pyx_freelist_6uvloop_4loop__UDPSendContext[__pyx_freecount_6uvloop_4loop__UDPSendContext++] = ((struct __pyx_obj_6uvloop_4loop__UDPSendContext *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop__UDPSendContext(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop__UDPSendContext *p = (struct __pyx_obj_6uvloop_4loop__UDPSendContext *)o;
  if (p->udp) {
    e = (*v)(((PyObject *)p->udp), a); if (e) return e;
  }
  if (p->py_buf.obj) {
    e = (*v)(p->py_buf.obj, a); if (e) return e;
  }
  return 0;
}

static PyMethodDef __pyx_methods_6uvloop_4loop__UDPSendContext[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15_UDPSendContext_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15_UDPSendContext_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_6uvloop_4loop_15_UDPSendContext_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_6uvloop_4loop_15_UDPSendContext_4__setstate_cython__},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop__UDPSendContext_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop__UDPSendContext},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop__UDPSendContext},
  {Py_tp_methods, (void *)__pyx_methods_6uvloop_4loop__UDPSendContext},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop__UDPSendContext},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop__UDPSendContext_spec = {
  "uvloop.loop._UDPSendContext",
  sizeof(struct __pyx_obj_6uvloop_4loop__UDPSendContext),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_6uvloop_4loop__UDPSendContext_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop__UDPSendContext = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""_UDPSendContext", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop__UDPSendContext), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop__UDPSendContext, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop__UDPSendContext, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6uvloop_4loop__UDPSendContext, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop__UDPSendContext, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *__pyx_freelist_6uvloop_4loop___pyx_scope_struct___getaddrinfo[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct___getaddrinfo = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct___getaddrinfo(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct___getaddrinfo > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct___getaddrinfo[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct___getaddrinfo];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct___getaddrinfo(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct___getaddrinfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_fut);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct___getaddrinfo < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct___getaddrinfo[__pyx_freecount_6uvloop_4loop___pyx_scope_struct___getaddrinfo++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct___getaddrinfo(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *)o;
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct___getaddrinfo(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo *)o;
  tmp = ((PyObject*)p->__pyx_v_fut);
  p->__pyx_v_fut = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct___getaddrinfo},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct___getaddrinfo},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop___pyx_scope_struct___getaddrinfo},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct___getaddrinfo},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo_spec = {
  "uvloop.loop.__pyx_scope_struct___getaddrinfo",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct___getaddrinfo", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct___getaddrinfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct___getaddrinfo, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct___getaddrinfo, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct___getaddrinfo, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct___getaddrinfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_1__getnameinfo[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_1__getnameinfo = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_1__getnameinfo(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_1__getnameinfo > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_1__getnameinfo[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_1__getnameinfo];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_1__getnameinfo(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_1__getnameinfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_fut);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_1__getnameinfo < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_1__getnameinfo[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_1__getnameinfo++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_1__getnameinfo(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *)o;
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_1__getnameinfo(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo *)o;
  tmp = ((PyObject*)p->__pyx_v_fut);
  p->__pyx_v_fut = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_1__getnameinfo},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_1__getnameinfo},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_1__getnameinfo},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_1__getnameinfo},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo_spec = {
  "uvloop.loop.__pyx_scope_struct_1__getnameinfo",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_1__getnameinfo", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_1__getnameinfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_1__getnameinfo, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_1__getnameinfo, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_1__getnameinfo, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_1__getnameinfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_2___get__[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_2___get__ = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_2___get__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_2___get__ > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_2___get__[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_2___get__];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_2___get__(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_2___get__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_2___get__ < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_2___get__[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_2___get__++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_2___get__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_2___get__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_2___get___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_2___get__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_2___get__},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_2___get__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_2___get__},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_2___get___spec = {
  "uvloop.loop.__pyx_scope_struct_2___get__",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_2___get___slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_2___get__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_2___get__", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_2___get__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_2___get__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_2___get__, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_2___get__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_2___get__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_3_run_until_complete[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_3_run_until_complete = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_3_run_until_complete(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_3_run_until_complete > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_3_run_until_complete[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_3_run_until_complete];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_3_run_until_complete(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_3_run_until_complete) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_3_run_until_complete < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_3_run_until_complete[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_3_run_until_complete++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_3_run_until_complete(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_3_run_until_complete(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_3_run_until_complete},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_3_run_until_complete},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_3_run_until_complete},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_3_run_until_complete},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete_spec = {
  "uvloop.loop.__pyx_scope_struct_3_run_until_complete",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_3_run_until_complete", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_3_run_until_complete), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_3_run_until_complete, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_3_run_until_complete, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_3_run_until_complete, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_3_run_until_complete, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_addr);
  Py_CLEAR(p->__pyx_v_host);
  Py_CLEAR(p->__pyx_v_port);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo *)o;
  if (p->__pyx_v_addr) {
    e = (*v)(p->__pyx_v_addr, a); if (e) return e;
  }
  if (p->__pyx_v_host) {
    e = (*v)(p->__pyx_v_host, a); if (e) return e;
  }
  if (p->__pyx_v_port) {
    e = (*v)(p->__pyx_v_port, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo_spec = {
  "uvloop.loop.__pyx_scope_struct_4_getaddrinfo",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_4_getaddrinfo", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_5_getnameinfo[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_5_getnameinfo = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_5_getnameinfo(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_5_getnameinfo > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_5_getnameinfo[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_5_getnameinfo];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_5_getnameinfo(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_5_getnameinfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ai_cnt);
  Py_CLEAR(p->__pyx_v_flowinfo);
  Py_CLEAR(p->__pyx_v_scope_id);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sockaddr);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_5_getnameinfo < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_5_getnameinfo[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_5_getnameinfo++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_5_getnameinfo(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo *)o;
  if (p->__pyx_v_ai_cnt) {
    e = (*v)(((PyObject *)p->__pyx_v_ai_cnt), a); if (e) return e;
  }
  if (p->__pyx_v_flowinfo) {
    e = (*v)(p->__pyx_v_flowinfo, a); if (e) return e;
  }
  if (p->__pyx_v_scope_id) {
    e = (*v)(p->__pyx_v_scope_id, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sockaddr) {
    e = (*v)(p->__pyx_v_sockaddr, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_5_getnameinfo},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_5_getnameinfo},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_5_getnameinfo},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo_spec = {
  "uvloop.loop.__pyx_scope_struct_5_getnameinfo",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_5_getnameinfo", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_5_getnameinfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_5_getnameinfo, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_5_getnameinfo, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_5_getnameinfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_6_start_tls[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_6_start_tls = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_6_start_tls(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_6_start_tls > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_6_start_tls[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_6_start_tls];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_6_start_tls(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_6_start_tls) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_app_transport);
  Py_CLEAR(p->__pyx_v_conmade_cb);
  Py_CLEAR(p->__pyx_v_context);
  Py_CLEAR(p->__pyx_v_protocol);
  Py_CLEAR(p->__pyx_v_resume_cb);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_server_hostname);
  Py_CLEAR(p->__pyx_v_server_side);
  Py_CLEAR(p->__pyx_v_ssl_handshake_timeout);
  Py_CLEAR(p->__pyx_v_ssl_protocol);
  Py_CLEAR(p->__pyx_v_ssl_shutdown_timeout);
  Py_CLEAR(p->__pyx_v_sslcontext);
  Py_CLEAR(p->__pyx_v_transport);
  Py_CLEAR(p->__pyx_v_waiter);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_6_start_tls < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_6_start_tls[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_6_start_tls++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_6_start_tls(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls *)o;
  if (p->__pyx_v_app_transport) {
    e = (*v)(p->__pyx_v_app_transport, a); if (e) return e;
  }
  if (p->__pyx_v_conmade_cb) {
    e = (*v)(p->__pyx_v_conmade_cb, a); if (e) return e;
  }
  if (p->__pyx_v_context) {
    e = (*v)(p->__pyx_v_context, a); if (e) return e;
  }
  if (p->__pyx_v_protocol) {
    e = (*v)(p->__pyx_v_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_resume_cb) {
    e = (*v)(p->__pyx_v_resume_cb, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_server_hostname) {
    e = (*v)(p->__pyx_v_server_hostname, a); if (e) return e;
  }
  if (p->__pyx_v_server_side) {
    e = (*v)(p->__pyx_v_server_side, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_handshake_timeout) {
    e = (*v)(p->__pyx_v_ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_protocol) {
    e = (*v)(((PyObject *)p->__pyx_v_ssl_protocol), a); if (e) return e;
  }
  if (p->__pyx_v_ssl_shutdown_timeout) {
    e = (*v)(p->__pyx_v_ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_sslcontext) {
    e = (*v)(p->__pyx_v_sslcontext, a); if (e) return e;
  }
  if (p->__pyx_v_transport) {
    e = (*v)(p->__pyx_v_transport, a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_6_start_tls},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_6_start_tls},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_6_start_tls},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls_spec = {
  "uvloop.loop.__pyx_scope_struct_6_start_tls",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_6_start_tls", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_6_start_tls), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_6_start_tls, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_6_start_tls, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_6_start_tls, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_7_create_server[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_7_create_server = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_7_create_server(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_7_create_server > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_7_create_server[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_7_create_server];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_7_create_server(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_7_create_server) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_backlog);
  Py_CLEAR(p->__pyx_v_err);
  Py_CLEAR(p->__pyx_v_fs);
  Py_CLEAR(p->__pyx_7genexpr__pyx_v_host);
  Py_CLEAR(p->__pyx_v_host);
  Py_CLEAR(p->__pyx_v_hosts);
  Py_CLEAR(p->__pyx_v_info);
  Py_CLEAR(p->__pyx_v_infos);
  Py_CLEAR(p->__pyx_v_port);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_pyaddr);
  Py_CLEAR(p->__pyx_v_reuse_address);
  Py_CLEAR(p->__pyx_v_reuse_port);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_server);
  Py_CLEAR(p->__pyx_v_sock);
  Py_CLEAR(p->__pyx_v_ssl);
  Py_CLEAR(p->__pyx_v_ssl_handshake_timeout);
  Py_CLEAR(p->__pyx_v_ssl_shutdown_timeout);
  Py_CLEAR(p->__pyx_v_start_serving);
  Py_CLEAR(p->__pyx_v_tcp);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_7_create_server < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_7_create_server[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_7_create_server++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_7_create_server(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server *)o;
  if (p->__pyx_v_backlog) {
    e = (*v)(p->__pyx_v_backlog, a); if (e) return e;
  }
  if (p->__pyx_v_err) {
    e = (*v)(p->__pyx_v_err, a); if (e) return e;
  }
  if (p->__pyx_v_fs) {
    e = (*v)(p->__pyx_v_fs, a); if (e) return e;
  }
  if (p->__pyx_7genexpr__pyx_v_host) {
    e = (*v)(p->__pyx_7genexpr__pyx_v_host, a); if (e) return e;
  }
  if (p->__pyx_v_host) {
    e = (*v)(p->__pyx_v_host, a); if (e) return e;
  }
  if (p->__pyx_v_hosts) {
    e = (*v)(p->__pyx_v_hosts, a); if (e) return e;
  }
  if (p->__pyx_v_info) {
    e = (*v)(p->__pyx_v_info, a); if (e) return e;
  }
  if (p->__pyx_v_infos) {
    e = (*v)(p->__pyx_v_infos, a); if (e) return e;
  }
  if (p->__pyx_v_port) {
    e = (*v)(p->__pyx_v_port, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_pyaddr) {
    e = (*v)(p->__pyx_v_pyaddr, a); if (e) return e;
  }
  if (p->__pyx_v_reuse_address) {
    e = (*v)(p->__pyx_v_reuse_address, a); if (e) return e;
  }
  if (p->__pyx_v_reuse_port) {
    e = (*v)(p->__pyx_v_reuse_port, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_server) {
    e = (*v)(((PyObject *)p->__pyx_v_server), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  if (p->__pyx_v_ssl) {
    e = (*v)(p->__pyx_v_ssl, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_handshake_timeout) {
    e = (*v)(p->__pyx_v_ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_shutdown_timeout) {
    e = (*v)(p->__pyx_v_ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_start_serving) {
    e = (*v)(p->__pyx_v_start_serving, a); if (e) return e;
  }
  if (p->__pyx_v_tcp) {
    e = (*v)(((PyObject *)p->__pyx_v_tcp), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_7_create_server},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_7_create_server},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_7_create_server},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server_spec = {
  "uvloop.loop.__pyx_scope_struct_7_create_server",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_7_create_server", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_7_create_server), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_7_create_server, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_7_create_server, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_7_create_server, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_8_create_connection[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_8_create_connection = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_8_create_connection(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_8_create_connection > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_8_create_connection[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_8_create_connection];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_8_create_connection(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_8_create_connection) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_addr);
  Py_CLEAR(p->__pyx_v_ai_local);
  Py_CLEAR(p->__pyx_v_ai_remote);
  Py_CLEAR(p->__pyx_v_app_protocol);
  Py_CLEAR(p->__pyx_v_app_transport);
  Py_CLEAR(p->__pyx_v_context);
  Py_CLEAR(p->__pyx_v_exc);
  Py_CLEAR(p->__pyx_v_exceptions);
  Py_CLEAR(p->__pyx_v_f1);
  Py_CLEAR(p->__pyx_v_f2);
  Py_CLEAR(p->__pyx_v_family);
  Py_CLEAR(p->__pyx_v_flags);
  Py_CLEAR(p->__pyx_v_fs);
  Py_CLEAR(p->__pyx_v_genexpr);
  Py_CLEAR(p->__pyx_v_host);
  Py_CLEAR(p->__pyx_v_local_addr);
  Py_CLEAR(p->__pyx_v_model);
  Py_CLEAR(p->__pyx_v_port);
  Py_CLEAR(p->__pyx_v_proto);
  Py_CLEAR(p->__pyx_v_protocol);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_server_hostname);
  Py_CLEAR(p->__pyx_v_sock);
  Py_CLEAR(p->__pyx_v_ssl);
  Py_CLEAR(p->__pyx_v_ssl_handshake_timeout);
  Py_CLEAR(p->__pyx_v_ssl_shutdown_timeout);
  Py_CLEAR(p->__pyx_v_ssl_waiter);
  Py_CLEAR(p->__pyx_v_sslcontext);
  Py_CLEAR(p->__pyx_v_tr);
  Py_CLEAR(p->__pyx_v_waiter);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_8_create_connection < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_8_create_connection[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_8_create_connection++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_8_create_connection(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection *)o;
  if (p->__pyx_v_addr) {
    e = (*v)(p->__pyx_v_addr, a); if (e) return e;
  }
  if (p->__pyx_v_ai_local) {
    e = (*v)(((PyObject *)p->__pyx_v_ai_local), a); if (e) return e;
  }
  if (p->__pyx_v_ai_remote) {
    e = (*v)(((PyObject *)p->__pyx_v_ai_remote), a); if (e) return e;
  }
  if (p->__pyx_v_app_protocol) {
    e = (*v)(p->__pyx_v_app_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_app_transport) {
    e = (*v)(p->__pyx_v_app_transport, a); if (e) return e;
  }
  if (p->__pyx_v_context) {
    e = (*v)(p->__pyx_v_context, a); if (e) return e;
  }
  if (p->__pyx_v_exc) {
    e = (*v)(p->__pyx_v_exc, a); if (e) return e;
  }
  if (p->__pyx_v_exceptions) {
    e = (*v)(p->__pyx_v_exceptions, a); if (e) return e;
  }
  if (p->__pyx_v_f1) {
    e = (*v)(p->__pyx_v_f1, a); if (e) return e;
  }
  if (p->__pyx_v_f2) {
    e = (*v)(p->__pyx_v_f2, a); if (e) return e;
  }
  if (p->__pyx_v_family) {
    e = (*v)(p->__pyx_v_family, a); if (e) return e;
  }
  if (p->__pyx_v_flags) {
    e = (*v)(p->__pyx_v_flags, a); if (e) return e;
  }
  if (p->__pyx_v_fs) {
    e = (*v)(p->__pyx_v_fs, a); if (e) return e;
  }
  if (p->__pyx_v_genexpr) {
    e = (*v)(p->__pyx_v_genexpr, a); if (e) return e;
  }
  if (p->__pyx_v_host) {
    e = (*v)(p->__pyx_v_host, a); if (e) return e;
  }
  if (p->__pyx_v_local_addr) {
    e = (*v)(p->__pyx_v_local_addr, a); if (e) return e;
  }
  if (p->__pyx_v_model) {
    e = (*v)(p->__pyx_v_model, a); if (e) return e;
  }
  if (p->__pyx_v_port) {
    e = (*v)(p->__pyx_v_port, a); if (e) return e;
  }
  if (p->__pyx_v_proto) {
    e = (*v)(p->__pyx_v_proto, a); if (e) return e;
  }
  if (p->__pyx_v_protocol) {
    e = (*v)(p->__pyx_v_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_server_hostname) {
    e = (*v)(p->__pyx_v_server_hostname, a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  if (p->__pyx_v_ssl) {
    e = (*v)(p->__pyx_v_ssl, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_handshake_timeout) {
    e = (*v)(p->__pyx_v_ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_shutdown_timeout) {
    e = (*v)(p->__pyx_v_ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_waiter) {
    e = (*v)(p->__pyx_v_ssl_waiter, a); if (e) return e;
  }
  if (p->__pyx_v_sslcontext) {
    e = (*v)(p->__pyx_v_sslcontext, a); if (e) return e;
  }
  if (p->__pyx_v_tr) {
    e = (*v)(((PyObject *)p->__pyx_v_tr), a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_8_create_connection},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_8_create_connection},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_8_create_connection},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection_spec = {
  "uvloop.loop.__pyx_scope_struct_8_create_connection",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_8_create_connection", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_8_create_connection), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_8_create_connection, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_8_create_connection, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_8_create_connection, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_9_genexpr[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_9_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_9_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_9_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_9_genexpr[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_9_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_9_genexpr(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_9_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_exc);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_9_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_9_genexpr[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_9_genexpr++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_9_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_exc) {
    e = (*v)(p->__pyx_v_exc, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_9_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_9_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_9_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr_spec = {
  "uvloop.loop.__pyx_scope_struct_9_genexpr",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_9_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_9_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_9_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_9_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_9_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_10_genexpr[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_10_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_10_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_10_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_10_genexpr[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_10_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_10_genexpr(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_10_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_exc);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_10_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_10_genexpr[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_10_genexpr++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_10_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_exc) {
    e = (*v)(p->__pyx_v_exc, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_10_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_10_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_10_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr_spec = {
  "uvloop.loop.__pyx_scope_struct_10_genexpr",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_10_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_10_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_10_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_10_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_10_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_11_create_unix_server[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_11_create_unix_server = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_11_create_unix_server(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_11_create_unix_server > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_11_create_unix_server[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_11_create_unix_server];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_11_create_unix_server(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_11_create_unix_server) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_backlog);
  Py_CLEAR(p->__pyx_v_cleanup_socket);
  Py_CLEAR(p->__pyx_v_err);
  Py_CLEAR(p->__pyx_v_exc);
  Py_CLEAR(p->__pyx_v_msg);
  Py_CLEAR(p->__pyx_v_orig_path);
  Py_CLEAR(p->__pyx_v_path);
  Py_CLEAR(p->__pyx_v_pipe);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_server);
  Py_CLEAR(p->__pyx_v_sock);
  Py_CLEAR(p->__pyx_v_ssl);
  Py_CLEAR(p->__pyx_v_ssl_handshake_timeout);
  Py_CLEAR(p->__pyx_v_ssl_shutdown_timeout);
  Py_CLEAR(p->__pyx_v_start_serving);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_11_create_unix_server < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_11_create_unix_server[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_11_create_unix_server++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_11_create_unix_server(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server *)o;
  if (p->__pyx_v_backlog) {
    e = (*v)(p->__pyx_v_backlog, a); if (e) return e;
  }
  if (p->__pyx_v_cleanup_socket) {
    e = (*v)(p->__pyx_v_cleanup_socket, a); if (e) return e;
  }
  if (p->__pyx_v_err) {
    e = (*v)(p->__pyx_v_err, a); if (e) return e;
  }
  if (p->__pyx_v_exc) {
    e = (*v)(p->__pyx_v_exc, a); if (e) return e;
  }
  if (p->__pyx_v_msg) {
    e = (*v)(p->__pyx_v_msg, a); if (e) return e;
  }
  if (p->__pyx_v_orig_path) {
    e = (*v)(p->__pyx_v_orig_path, a); if (e) return e;
  }
  if (p->__pyx_v_path) {
    e = (*v)(p->__pyx_v_path, a); if (e) return e;
  }
  if (p->__pyx_v_pipe) {
    e = (*v)(((PyObject *)p->__pyx_v_pipe), a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_server) {
    e = (*v)(((PyObject *)p->__pyx_v_server), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  if (p->__pyx_v_ssl) {
    e = (*v)(p->__pyx_v_ssl, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_handshake_timeout) {
    e = (*v)(p->__pyx_v_ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_shutdown_timeout) {
    e = (*v)(p->__pyx_v_ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_start_serving) {
    e = (*v)(p->__pyx_v_start_serving, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_11_create_unix_server},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_11_create_unix_server},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_11_create_unix_server},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server_spec = {
  "uvloop.loop.__pyx_scope_struct_11_create_unix_server",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_11_create_unix_server", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_11_create_unix_server), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_11_create_unix_server, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_11_create_unix_server, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_11_create_unix_server, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_app_protocol);
  Py_CLEAR(p->__pyx_v_app_transport);
  Py_CLEAR(p->__pyx_v_context);
  Py_CLEAR(p->__pyx_v_path);
  Py_CLEAR(p->__pyx_v_protocol);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_server_hostname);
  Py_CLEAR(p->__pyx_v_sock);
  Py_CLEAR(p->__pyx_v_ssl);
  Py_CLEAR(p->__pyx_v_ssl_handshake_timeout);
  Py_CLEAR(p->__pyx_v_ssl_shutdown_timeout);
  Py_CLEAR(p->__pyx_v_ssl_waiter);
  Py_CLEAR(p->__pyx_v_sslcontext);
  Py_CLEAR(p->__pyx_v_tr);
  Py_CLEAR(p->__pyx_v_waiter);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection *)o;
  if (p->__pyx_v_app_protocol) {
    e = (*v)(p->__pyx_v_app_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_app_transport) {
    e = (*v)(p->__pyx_v_app_transport, a); if (e) return e;
  }
  if (p->__pyx_v_context) {
    e = (*v)(p->__pyx_v_context, a); if (e) return e;
  }
  if (p->__pyx_v_path) {
    e = (*v)(p->__pyx_v_path, a); if (e) return e;
  }
  if (p->__pyx_v_protocol) {
    e = (*v)(p->__pyx_v_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_server_hostname) {
    e = (*v)(p->__pyx_v_server_hostname, a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  if (p->__pyx_v_ssl) {
    e = (*v)(p->__pyx_v_ssl, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_handshake_timeout) {
    e = (*v)(p->__pyx_v_ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_shutdown_timeout) {
    e = (*v)(p->__pyx_v_ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_waiter) {
    e = (*v)(p->__pyx_v_ssl_waiter, a); if (e) return e;
  }
  if (p->__pyx_v_sslcontext) {
    e = (*v)(p->__pyx_v_sslcontext, a); if (e) return e;
  }
  if (p->__pyx_v_tr) {
    e = (*v)(((PyObject *)p->__pyx_v_tr), a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection_spec = {
  "uvloop.loop.__pyx_scope_struct_12_create_unix_connection",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_12_create_unix_connection", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_13_sock_recv[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_13_sock_recv = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_13_sock_recv(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_13_sock_recv > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_13_sock_recv[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_13_sock_recv];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_13_sock_recv(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_13_sock_recv) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_fut);
  Py_CLEAR(p->__pyx_v_handle);
  Py_CLEAR(p->__pyx_v_n);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_13_sock_recv < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_13_sock_recv[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_13_sock_recv++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_13_sock_recv(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv *)o;
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  if (p->__pyx_v_handle) {
    e = (*v)(((PyObject *)p->__pyx_v_handle), a); if (e) return e;
  }
  if (p->__pyx_v_n) {
    e = (*v)(p->__pyx_v_n, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_13_sock_recv},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_13_sock_recv},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_13_sock_recv},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv_spec = {
  "uvloop.loop.__pyx_scope_struct_13_sock_recv",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_13_sock_recv", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_13_sock_recv), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_13_sock_recv, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_13_sock_recv, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_13_sock_recv, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_buf);
  Py_CLEAR(p->__pyx_v_fut);
  Py_CLEAR(p->__pyx_v_handle);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into *)o;
  if (p->__pyx_v_buf) {
    e = (*v)(p->__pyx_v_buf, a); if (e) return e;
  }
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  if (p->__pyx_v_handle) {
    e = (*v)(((PyObject *)p->__pyx_v_handle), a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into_spec = {
  "uvloop.loop.__pyx_scope_struct_14_sock_recv_into",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_14_sock_recv_into", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_15_sock_sendall[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_15_sock_sendall = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_15_sock_sendall(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_15_sock_sendall > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_15_sock_sendall[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_15_sock_sendall];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_15_sock_sendall(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_15_sock_sendall) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_data);
  Py_CLEAR(p->__pyx_v_fut);
  Py_CLEAR(p->__pyx_v_handle);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_15_sock_sendall < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_15_sock_sendall[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_15_sock_sendall++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_15_sock_sendall(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall *)o;
  if (p->__pyx_v_data) {
    e = (*v)(p->__pyx_v_data, a); if (e) return e;
  }
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  if (p->__pyx_v_handle) {
    e = (*v)(((PyObject *)p->__pyx_v_handle), a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_15_sock_sendall},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_15_sock_sendall},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_15_sock_sendall},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall_spec = {
  "uvloop.loop.__pyx_scope_struct_15_sock_sendall",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_15_sock_sendall", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_15_sock_sendall), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_15_sock_sendall, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_15_sock_sendall, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_15_sock_sendall, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_16_sock_accept[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_16_sock_accept = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_16_sock_accept(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_16_sock_accept > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_16_sock_accept[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_16_sock_accept];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_16_sock_accept(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_16_sock_accept) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_fut);
  Py_CLEAR(p->__pyx_v_handle);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_16_sock_accept < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_16_sock_accept[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_16_sock_accept++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_16_sock_accept(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept *)o;
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  if (p->__pyx_v_handle) {
    e = (*v)(((PyObject *)p->__pyx_v_handle), a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_16_sock_accept},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_16_sock_accept},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_16_sock_accept},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept_spec = {
  "uvloop.loop.__pyx_scope_struct_16_sock_accept",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_16_sock_accept", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_16_sock_accept), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_16_sock_accept, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_16_sock_accept, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_16_sock_accept, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_17_sock_connect[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_17_sock_connect = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_17_sock_connect(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_17_sock_connect > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_17_sock_connect[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_17_sock_connect];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_17_sock_connect(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_17_sock_connect) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v__);
  Py_CLEAR(p->__pyx_v_address);
  Py_CLEAR(p->__pyx_v_addrs);
  Py_CLEAR(p->__pyx_v_fut);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_17_sock_connect < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_17_sock_connect[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_17_sock_connect++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_17_sock_connect(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect *)o;
  if (p->__pyx_v__) {
    e = (*v)(p->__pyx_v__, a); if (e) return e;
  }
  if (p->__pyx_v_address) {
    e = (*v)(p->__pyx_v_address, a); if (e) return e;
  }
  if (p->__pyx_v_addrs) {
    e = (*v)(p->__pyx_v_addrs, a); if (e) return e;
  }
  if (p->__pyx_v_fut) {
    e = (*v)(p->__pyx_v_fut, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_17_sock_connect},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_17_sock_connect},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_17_sock_connect},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect_spec = {
  "uvloop.loop.__pyx_scope_struct_17_sock_connect",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_17_sock_connect", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_17_sock_connect), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_17_sock_connect, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_17_sock_connect, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_17_sock_connect, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_bufsize);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom *)o;
  if (p->__pyx_v_bufsize) {
    e = (*v)(p->__pyx_v_bufsize, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom_spec = {
  "uvloop.loop.__pyx_scope_struct_18_sock_recvfrom",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_18_sock_recvfrom", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_buf);
  Py_CLEAR(p->__pyx_v_nbytes);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into *)o;
  if (p->__pyx_v_buf) {
    e = (*v)(p->__pyx_v_buf, a); if (e) return e;
  }
  if (p->__pyx_v_nbytes) {
    e = (*v)(p->__pyx_v_nbytes, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into_spec = {
  "uvloop.loop.__pyx_scope_struct_19_sock_recvfrom_into",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_19_sock_recvfrom_into", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_20_sock_sendto[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_20_sock_sendto = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_20_sock_sendto(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_20_sock_sendto > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_20_sock_sendto[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_20_sock_sendto];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_20_sock_sendto(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_20_sock_sendto) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_address);
  Py_CLEAR(p->__pyx_v_data);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_20_sock_sendto < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_20_sock_sendto[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_20_sock_sendto++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_20_sock_sendto(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto *)o;
  if (p->__pyx_v_address) {
    e = (*v)(p->__pyx_v_address, a); if (e) return e;
  }
  if (p->__pyx_v_data) {
    e = (*v)(p->__pyx_v_data, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_20_sock_sendto},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_20_sock_sendto},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_20_sock_sendto},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto_spec = {
  "uvloop.loop.__pyx_scope_struct_20_sock_sendto",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_20_sock_sendto", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_20_sock_sendto), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_20_sock_sendto, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_20_sock_sendto, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_20_sock_sendto, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_app_protocol);
  Py_CLEAR(p->__pyx_v_app_transport);
  Py_CLEAR(p->__pyx_v_context);
  Py_CLEAR(p->__pyx_v_protocol);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  Py_CLEAR(p->__pyx_v_ssl);
  Py_CLEAR(p->__pyx_v_ssl_handshake_timeout);
  Py_CLEAR(p->__pyx_v_ssl_shutdown_timeout);
  Py_CLEAR(p->__pyx_v_transport);
  Py_CLEAR(p->__pyx_v_transport_waiter);
  Py_CLEAR(p->__pyx_v_waiter);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket *)o;
  if (p->__pyx_v_app_protocol) {
    e = (*v)(p->__pyx_v_app_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_app_transport) {
    e = (*v)(p->__pyx_v_app_transport, a); if (e) return e;
  }
  if (p->__pyx_v_context) {
    e = (*v)(p->__pyx_v_context, a); if (e) return e;
  }
  if (p->__pyx_v_protocol) {
    e = (*v)(p->__pyx_v_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  if (p->__pyx_v_ssl) {
    e = (*v)(p->__pyx_v_ssl, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_handshake_timeout) {
    e = (*v)(p->__pyx_v_ssl_handshake_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_ssl_shutdown_timeout) {
    e = (*v)(p->__pyx_v_ssl_shutdown_timeout, a); if (e) return e;
  }
  if (p->__pyx_v_transport) {
    e = (*v)(((PyObject *)p->__pyx_v_transport), a); if (e) return e;
  }
  if (p->__pyx_v_transport_waiter) {
    e = (*v)(p->__pyx_v_transport_waiter, a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket_spec = {
  "uvloop.loop.__pyx_scope_struct_21_connect_accepted_socket",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_21_connect_accepted_socket", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_args);
  Py_CLEAR(p->__pyx_v_bufsize);
  Py_CLEAR(p->__pyx_v_close_fds);
  Py_CLEAR(p->__pyx_v_creationflags);
  Py_CLEAR(p->__pyx_v_cwd);
  Py_CLEAR(p->__pyx_v_env);
  Py_CLEAR(p->__pyx_v_executable);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_pass_fds);
  Py_CLEAR(p->__pyx_v_preexec_fn);
  Py_CLEAR(p->__pyx_v_proc);
  Py_CLEAR(p->__pyx_v_protocol);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_restore_signals);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_shell);
  Py_CLEAR(p->__pyx_v_start_new_session);
  Py_CLEAR(p->__pyx_v_startupinfo);
  Py_CLEAR(p->__pyx_v_stderr);
  Py_CLEAR(p->__pyx_v_stdin);
  Py_CLEAR(p->__pyx_v_stdout);
  Py_CLEAR(p->__pyx_v_universal_newlines);
  Py_CLEAR(p->__pyx_v_waiter);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run *)o;
  if (p->__pyx_v_args) {
    e = (*v)(p->__pyx_v_args, a); if (e) return e;
  }
  if (p->__pyx_v_bufsize) {
    e = (*v)(p->__pyx_v_bufsize, a); if (e) return e;
  }
  if (p->__pyx_v_close_fds) {
    e = (*v)(p->__pyx_v_close_fds, a); if (e) return e;
  }
  if (p->__pyx_v_creationflags) {
    e = (*v)(p->__pyx_v_creationflags, a); if (e) return e;
  }
  if (p->__pyx_v_cwd) {
    e = (*v)(p->__pyx_v_cwd, a); if (e) return e;
  }
  if (p->__pyx_v_env) {
    e = (*v)(p->__pyx_v_env, a); if (e) return e;
  }
  if (p->__pyx_v_executable) {
    e = (*v)(p->__pyx_v_executable, a); if (e) return e;
  }
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_pass_fds) {
    e = (*v)(p->__pyx_v_pass_fds, a); if (e) return e;
  }
  if (p->__pyx_v_preexec_fn) {
    e = (*v)(p->__pyx_v_preexec_fn, a); if (e) return e;
  }
  if (p->__pyx_v_proc) {
    e = (*v)(((PyObject *)p->__pyx_v_proc), a); if (e) return e;
  }
  if (p->__pyx_v_protocol) {
    e = (*v)(p->__pyx_v_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_restore_signals) {
    e = (*v)(p->__pyx_v_restore_signals, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_shell) {
    e = (*v)(p->__pyx_v_shell, a); if (e) return e;
  }
  if (p->__pyx_v_start_new_session) {
    e = (*v)(p->__pyx_v_start_new_session, a); if (e) return e;
  }
  if (p->__pyx_v_startupinfo) {
    e = (*v)(p->__pyx_v_startupinfo, a); if (e) return e;
  }
  if (p->__pyx_v_stderr) {
    e = (*v)(p->__pyx_v_stderr, a); if (e) return e;
  }
  if (p->__pyx_v_stdin) {
    e = (*v)(p->__pyx_v_stdin, a); if (e) return e;
  }
  if (p->__pyx_v_stdout) {
    e = (*v)(p->__pyx_v_stdout, a); if (e) return e;
  }
  if (p->__pyx_v_universal_newlines) {
    e = (*v)(p->__pyx_v_universal_newlines, a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run_spec = {
  "uvloop.loop.__pyx_scope_struct_22__Loop__subprocess_run",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_22__Loop__subprocess_run", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_args);
  Py_CLEAR(p->__pyx_v_cmd);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_shell);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell *)o;
  if (p->__pyx_v_args) {
    e = (*v)(p->__pyx_v_args, a); if (e) return e;
  }
  if (p->__pyx_v_cmd) {
    e = (*v)(p->__pyx_v_cmd, a); if (e) return e;
  }
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_shell) {
    e = (*v)(p->__pyx_v_shell, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell_spec = {
  "uvloop.loop.__pyx_scope_struct_23_subprocess_shell",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_23_subprocess_shell", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_args);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_program);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_shell);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec *)o;
  if (p->__pyx_v_args) {
    e = (*v)(p->__pyx_v_args, a); if (e) return e;
  }
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_program) {
    e = (*v)(p->__pyx_v_program, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_shell) {
    e = (*v)(p->__pyx_v_shell, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec_spec = {
  "uvloop.loop.__pyx_scope_struct_24_subprocess_exec",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_24_subprocess_exec", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_pipe);
  Py_CLEAR(p->__pyx_v_proto);
  Py_CLEAR(p->__pyx_v_proto_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_transp);
  Py_CLEAR(p->__pyx_v_waiter);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe *)o;
  if (p->__pyx_v_pipe) {
    e = (*v)(p->__pyx_v_pipe, a); if (e) return e;
  }
  if (p->__pyx_v_proto) {
    e = (*v)(p->__pyx_v_proto, a); if (e) return e;
  }
  if (p->__pyx_v_proto_factory) {
    e = (*v)(p->__pyx_v_proto_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_transp) {
    e = (*v)(((PyObject *)p->__pyx_v_transp), a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe_spec = {
  "uvloop.loop.__pyx_scope_struct_25_connect_read_pipe",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_25_connect_read_pipe", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_pipe);
  Py_CLEAR(p->__pyx_v_proto);
  Py_CLEAR(p->__pyx_v_proto_factory);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_transp);
  Py_CLEAR(p->__pyx_v_waiter);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe *)o;
  if (p->__pyx_v_pipe) {
    e = (*v)(p->__pyx_v_pipe, a); if (e) return e;
  }
  if (p->__pyx_v_proto) {
    e = (*v)(p->__pyx_v_proto, a); if (e) return e;
  }
  if (p->__pyx_v_proto_factory) {
    e = (*v)(p->__pyx_v_proto_factory, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_transp) {
    e = (*v)(((PyObject *)p->__pyx_v_transp), a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe_spec = {
  "uvloop.loop.__pyx_scope_struct_26_connect_write_pipe",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_26_connect_write_pipe", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_allow_broadcast);
  Py_CLEAR(p->__pyx_v_ctx);
  Py_CLEAR(p->__pyx_v_ex);
  Py_CLEAR(p->__pyx_v_excs);
  Py_CLEAR(p->__pyx_v_family);
  Py_CLEAR(p->__pyx_v_flags);
  Py_CLEAR(p->__pyx_v_genexpr);
  Py_CLEAR(p->__pyx_v_lads);
  Py_CLEAR(p->__pyx_v_local_addr);
  Py_CLEAR(p->__pyx_v_opts);
  Py_CLEAR(p->__pyx_v_problems);
  Py_CLEAR(p->__pyx_v_proto);
  Py_CLEAR(p->__pyx_v_protocol);
  Py_CLEAR(p->__pyx_v_protocol_factory);
  Py_CLEAR(p->__pyx_v_rads);
  Py_CLEAR(p->__pyx_v_remote_addr);
  Py_CLEAR(p->__pyx_v_reuse_address);
  Py_CLEAR(p->__pyx_v_reuse_port);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_sock);
  Py_CLEAR(p->__pyx_v_udp);
  Py_CLEAR(p->__pyx_v_waiter);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint *)o;
  if (p->__pyx_v_allow_broadcast) {
    e = (*v)(p->__pyx_v_allow_broadcast, a); if (e) return e;
  }
  if (p->__pyx_v_ctx) {
    e = (*v)(p->__pyx_v_ctx, a); if (e) return e;
  }
  if (p->__pyx_v_ex) {
    e = (*v)(p->__pyx_v_ex, a); if (e) return e;
  }
  if (p->__pyx_v_excs) {
    e = (*v)(p->__pyx_v_excs, a); if (e) return e;
  }
  if (p->__pyx_v_family) {
    e = (*v)(p->__pyx_v_family, a); if (e) return e;
  }
  if (p->__pyx_v_flags) {
    e = (*v)(p->__pyx_v_flags, a); if (e) return e;
  }
  if (p->__pyx_v_genexpr) {
    e = (*v)(p->__pyx_v_genexpr, a); if (e) return e;
  }
  if (p->__pyx_v_lads) {
    e = (*v)(p->__pyx_v_lads, a); if (e) return e;
  }
  if (p->__pyx_v_local_addr) {
    e = (*v)(p->__pyx_v_local_addr, a); if (e) return e;
  }
  if (p->__pyx_v_opts) {
    e = (*v)(p->__pyx_v_opts, a); if (e) return e;
  }
  if (p->__pyx_v_proto) {
    e = (*v)(p->__pyx_v_proto, a); if (e) return e;
  }
  if (p->__pyx_v_protocol) {
    e = (*v)(p->__pyx_v_protocol, a); if (e) return e;
  }
  if (p->__pyx_v_protocol_factory) {
    e = (*v)(p->__pyx_v_protocol_factory, a); if (e) return e;
  }
  if (p->__pyx_v_rads) {
    e = (*v)(p->__pyx_v_rads, a); if (e) return e;
  }
  if (p->__pyx_v_remote_addr) {
    e = (*v)(p->__pyx_v_remote_addr, a); if (e) return e;
  }
  if (p->__pyx_v_reuse_address) {
    e = (*v)(p->__pyx_v_reuse_address, a); if (e) return e;
  }
  if (p->__pyx_v_reuse_port) {
    e = (*v)(p->__pyx_v_reuse_port, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_sock) {
    e = (*v)(p->__pyx_v_sock, a); if (e) return e;
  }
  if (p->__pyx_v_udp) {
    e = (*v)(((PyObject *)p->__pyx_v_udp), a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint_spec = {
  "uvloop.loop.__pyx_scope_struct_27_create_datagram_endpoint",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_27_create_datagram_endpoint", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_28_genexpr[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_28_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_28_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_28_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_28_genexpr[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_28_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_28_genexpr(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_28_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_28_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_28_genexpr[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_28_genexpr++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_28_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_28_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_28_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_28_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr_spec = {
  "uvloop.loop.__pyx_scope_struct_28_genexpr",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_28_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_28_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_28_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_28_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_28_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_8genexpr4__pyx_v_ag);
  Py_CLEAR(p->__pyx_v_agen);
  Py_CLEAR(p->__pyx_v_closing_agens);
  Py_CLEAR(p->__pyx_v_result);
  Py_CLEAR(p->__pyx_v_results);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_shutdown_coro);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens *)o;
  if (p->__pyx_8genexpr4__pyx_v_ag) {
    e = (*v)(p->__pyx_8genexpr4__pyx_v_ag, a); if (e) return e;
  }
  if (p->__pyx_v_agen) {
    e = (*v)(p->__pyx_v_agen, a); if (e) return e;
  }
  if (p->__pyx_v_closing_agens) {
    e = (*v)(p->__pyx_v_closing_agens, a); if (e) return e;
  }
  if (p->__pyx_v_result) {
    e = (*v)(p->__pyx_v_result, a); if (e) return e;
  }
  if (p->__pyx_v_results) {
    e = (*v)(p->__pyx_v_results, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_shutdown_coro) {
    e = (*v)(p->__pyx_v_shutdown_coro, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens_spec = {
  "uvloop.loop.__pyx_scope_struct_29_shutdown_asyncgens",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_29_shutdown_asyncgens", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_future);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_thread);
  Py_CLEAR(p->__pyx_v_timeout);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor *)o;
  if (p->__pyx_v_future) {
    e = (*v)(p->__pyx_v_future, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_thread) {
    e = (*v)(p->__pyx_v_thread, a); if (e) return e;
  }
  if (p->__pyx_v_timeout) {
    e = (*v)(p->__pyx_v_timeout, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor_spec = {
  "uvloop.loop.__pyx_scope_struct_30_shutdown_default_executor",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_30_shutdown_default_executor", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_31__start_handshake[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_31__start_handshake = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_31__start_handshake(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_31__start_handshake > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_31__start_handshake[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_31__start_handshake];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_31__start_handshake(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_31__start_handshake) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_31__start_handshake < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_31__start_handshake[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_31__start_handshake++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_31__start_handshake(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_31__start_handshake(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_31__start_handshake},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_31__start_handshake},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_31__start_handshake},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_31__start_handshake},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake_spec = {
  "uvloop.loop.__pyx_scope_struct_31__start_handshake",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_31__start_handshake", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_31__start_handshake), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_31__start_handshake, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_31__start_handshake, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_31__start_handshake, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_31__start_handshake, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_32__start_shutdown[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_32__start_shutdown = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_32__start_shutdown(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_32__start_shutdown > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_32__start_shutdown[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_32__start_shutdown];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_32__start_shutdown(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_32__start_shutdown) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_32__start_shutdown < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_32__start_shutdown[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_32__start_shutdown++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_32__start_shutdown(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_32__start_shutdown(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_6uvloop_4loop_SSLProtocol *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_32__start_shutdown},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_32__start_shutdown},
  {Py_tp_clear, (void *)__pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_32__start_shutdown},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_32__start_shutdown},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown_spec = {
  "uvloop.loop.__pyx_scope_struct_32__start_shutdown",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_32__start_shutdown", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_32__start_shutdown), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_32__start_shutdown, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_32__start_shutdown, /*tp_traverse*/
  __pyx_tp_clear_6uvloop_4loop___pyx_scope_struct_32__start_shutdown, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_32__start_shutdown, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_33___aenter__[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_33___aenter__ = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_33___aenter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_33___aenter__ > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_33___aenter__[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_33___aenter__];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_33___aenter__(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_33___aenter__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_33___aenter__ < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_33___aenter__[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_33___aenter__++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_33___aenter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_33___aenter__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_33___aenter__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_33___aenter__},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter___spec = {
  "uvloop.loop.__pyx_scope_struct_33___aenter__",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter___slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_33___aenter__", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_33___aenter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_33___aenter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_33___aenter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_33___aenter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_34___aexit__[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_34___aexit__ = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_34___aexit__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_34___aexit__ > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_34___aexit__[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_34___aexit__];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_34___aexit__(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_34___aexit__) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_exc);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_34___aexit__ < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_34___aexit__[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_34___aexit__++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_34___aexit__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__ *)o;
  if (p->__pyx_v_exc) {
    e = (*v)(p->__pyx_v_exc, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit___slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_34___aexit__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_34___aexit__},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_34___aexit__},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit___spec = {
  "uvloop.loop.__pyx_scope_struct_34___aexit__",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit___slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_34___aexit__", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_34___aexit__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_34___aexit__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_34___aexit__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_34___aexit__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_35_wait_closed[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_35_wait_closed = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_35_wait_closed(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_35_wait_closed > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_35_wait_closed[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_35_wait_closed];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_35_wait_closed(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_35_wait_closed) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_waiter);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_35_wait_closed < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_35_wait_closed[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_35_wait_closed++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_35_wait_closed(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_waiter) {
    e = (*v)(p->__pyx_v_waiter, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_35_wait_closed},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_35_wait_closed},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_35_wait_closed},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed_spec = {
  "uvloop.loop.__pyx_scope_struct_35_wait_closed",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_35_wait_closed", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_35_wait_closed), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_35_wait_closed, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_35_wait_closed, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_35_wait_closed, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_36_start_serving[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_36_start_serving = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_36_start_serving(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_36_start_serving > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_36_start_serving[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_36_start_serving];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_36_start_serving(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_36_start_serving) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_36_start_serving < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_36_start_serving[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_36_start_serving++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_36_start_serving(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_36_start_serving},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_36_start_serving},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_36_start_serving},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving_spec = {
  "uvloop.loop.__pyx_scope_struct_36_start_serving",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_36_start_serving", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_36_start_serving), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_36_start_serving, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_36_start_serving, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_36_start_serving, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_37_serve_forever[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_37_serve_forever = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_37_serve_forever(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_37_serve_forever > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_37_serve_forever[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_37_serve_forever];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_37_serve_forever(PyObject *o) {
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_37_serve_forever) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  Py_CLEAR(p->__pyx_t_3);
  Py_CLEAR(p->__pyx_t_4);
  Py_CLEAR(p->__pyx_t_5);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_37_serve_forever < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_37_serve_forever[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_37_serve_forever++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_37_serve_forever(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *p = (struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  if (p->__pyx_t_3) {
    e = (*v)(p->__pyx_t_3, a); if (e) return e;
  }
  if (p->__pyx_t_4) {
    e = (*v)(p->__pyx_t_4, a); if (e) return e;
  }
  if (p->__pyx_t_5) {
    e = (*v)(p->__pyx_t_5, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_37_serve_forever},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_37_serve_forever},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_37_serve_forever},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever_spec = {
  "uvloop.loop.__pyx_scope_struct_37_serve_forever",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_37_serve_forever", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_37_serve_forever), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_37_serve_forever, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6uvloop_4loop___pyx_scope_struct_37_serve_forever, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_37_serve_forever, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 *__pyx_freelist_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1[8];
static int __pyx_freecount_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 = 0;
#endif

static PyObject *__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1)))) {
    o = (PyObject*)__pyx_freelist_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1[--__pyx_freecount_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1];
    memset(o, 0, sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1)))) {
    __pyx_freelist_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1[__pyx_freecount_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1++] = ((struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1},
  {Py_tp_new, (void *)__pyx_tp_new_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1},
  {0, 0},
};
static PyType_Spec __pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1_spec = {
  "uvloop.loop.__pyx_scope_struct_38__test_coroutine_1",
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1_slots,
};
#else

static PyTypeObject __pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct_38__test_coroutine_1", /*tp_name*/
  sizeof(struct __pyx_obj_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled = 0;
#endif

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled},
  {Py_tp_new, (void *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled},
  {0, 0},
};
static PyType_Spec __pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_spec = {
  "uvloop.loop.__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled",
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_slots,
};
#else

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled = {
  PyVarObject_HEAD_INIT(0, 0)
  "uvloop.loop.""__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_b_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 0, 0, 0},
    {&__pyx_kp_u_18_7_7_7, __pyx_k_18_7_7_7, sizeof(__pyx_k_18_7_7_7), 0, 1, 0, 0},
    {&__pyx_kp_u_AF_INET6_address_must_be_tuple, __pyx_k_AF_INET6_address_must_be_tuple, sizeof(__pyx_k_AF_INET6_address_must_be_tuple), 0, 1, 0, 0},
    {&__pyx_kp_u_AF_INET6_must_be_a_tuple_of_2_4, __pyx_k_AF_INET6_must_be_a_tuple_of_2_4, sizeof(__pyx_k_AF_INET6_must_be_a_tuple_of_2_4), 0, 1, 0, 0},
    {&__pyx_kp_u_AF_INET_address_must_be_tuple, __pyx_k_AF_INET_address_must_be_tuple, sizeof(__pyx_k_AF_INET_address_must_be_tuple), 0, 1, 0, 0},
    {&__pyx_kp_u_AF_INET_address_must_be_tuple_of, __pyx_k_AF_INET_address_must_be_tuple_of, sizeof(__pyx_k_AF_INET_address_must_be_tuple_of), 0, 1, 0, 0},
    {&__pyx_kp_u_AF_UNIX_address_must_be_a_str_or, __pyx_k_AF_UNIX_address_must_be_a_str_or, sizeof(__pyx_k_AF_UNIX_address_must_be_a_str_or), 0, 1, 0, 0},
    {&__pyx_kp_u_AF_UNSPEC_in_DNS_results, __pyx_k_AF_UNSPEC_in_DNS_results, sizeof(__pyx_k_AF_UNSPEC_in_DNS_results), 0, 1, 0, 0},
    {&__pyx_n_u_AI_CANONNAME, __pyx_k_AI_CANONNAME, sizeof(__pyx_k_AI_CANONNAME), 0, 1, 0, 1},
    {&__pyx_kp_u_A_Stream_Socket_was_expected_got, __pyx_k_A_Stream_Socket_was_expected_got, sizeof(__pyx_k_A_Stream_Socket_was_expected_got), 0, 1, 0, 0},
    {&__pyx_kp_u_A_UDP_Socket_was_expected_got_r, __pyx_k_A_UDP_Socket_was_expected_got_r, sizeof(__pyx_k_A_UDP_Socket_was_expected_got_r), 0, 1, 0, 0},
    {&__pyx_kp_u_A_UNIX_Domain_Stream_Socket_was, __pyx_k_A_UNIX_Domain_Stream_Socket_was, sizeof(__pyx_k_A_UNIX_Domain_Stream_Socket_was), 0, 1, 0, 0},
    {&__pyx_kp_u_A_callable_object_or_None_is_exp, __pyx_k_A_callable_object_or_None_is_exp, sizeof(__pyx_k_A_callable_object_or_None_is_exp), 0, 1, 0, 0},
    {&__pyx_n_s_AbstractChildWatcher, __pyx_k_AbstractChildWatcher, sizeof(__pyx_k_AbstractChildWatcher), 0, 0, 1, 1},
    {&__pyx_n_s_AddrInfo, __pyx_k_AddrInfo, sizeof(__pyx_k_AddrInfo), 0, 0, 1, 1},
    {&__pyx_n_s_AddrInfoRequest, __pyx_k_AddrInfoRequest, sizeof(__pyx_k_AddrInfoRequest), 0, 0, 1, 1},
    {&__pyx_n_s_AddrInfoRequest___reduce_cython, __pyx_k_AddrInfoRequest___reduce_cython, sizeof(__pyx_k_AddrInfoRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_AddrInfoRequest___setstate_cytho, __pyx_k_AddrInfoRequest___setstate_cytho, sizeof(__pyx_k_AddrInfoRequest___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_kp_u_AddrInfoRequest_callback_called, __pyx_k_AddrInfoRequest_callback_called, sizeof(__pyx_k_AddrInfoRequest_callback_called), 0, 1, 0, 0},
    {&__pyx_n_s_AddrInfo___reduce_cython, __pyx_k_AddrInfo___reduce_cython, sizeof(__pyx_k_AddrInfo___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_AddrInfo___setstate_cython, __pyx_k_AddrInfo___setstate_cython, sizeof(__pyx_k_AddrInfo___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_AddrInfo_data_is_NULL, __pyx_k_AddrInfo_data_is_NULL, sizeof(__pyx_k_AddrInfo_data_is_NULL), 0, 1, 0, 0},
    {&__pyx_n_s_AddressFamily, __pyx_k_AddressFamily, sizeof(__pyx_k_AddressFamily), 0, 0, 1, 1},
    {&__pyx_kp_u_Address_r_is_already_in_use, __pyx_k_Address_r_is_already_in_use, sizeof(__pyx_k_Address_r_is_already_in_use), 0, 1, 0, 0},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
    {&__pyx_kp_u_Bad_exception_data_from_child_r, __pyx_k_Bad_exception_data_from_child_r, sizeof(__pyx_k_Bad_exception_data_from_child_r), 0, 1, 0, 0},
    {&__pyx_n_s_BaseException, __pyx_k_BaseException, sizeof(__pyx_k_BaseException), 0, 0, 1, 1},
    {&__pyx_n_s_BaseProtocol, __pyx_k_BaseProtocol, sizeof(__pyx_k_BaseProtocol), 0, 0, 1, 1},
    {&__pyx_n_s_BlockingIOError, __pyx_k_BlockingIOError, sizeof(__pyx_k_BlockingIOError), 0, 0, 1, 1},
    {&__pyx_n_s_BrokenPipeError, __pyx_k_BrokenPipeError, sizeof(__pyx_k_BrokenPipeError), 0, 0, 1, 1},
    {&__pyx_n_s_CHANGE, __pyx_k_CHANGE, sizeof(__pyx_k_CHANGE), 0, 0, 1, 1},
    {&__pyx_kp_u_Callback_handles_8, __pyx_k_Callback_handles_8, sizeof(__pyx_k_Callback_handles_8), 0, 1, 0, 0},
    {&__pyx_n_s_CancelledError, __pyx_k_CancelledError, sizeof(__pyx_k_CancelledError), 0, 0, 1, 1},
    {&__pyx_kp_u_Cannot_call_write_after_write_eo, __pyx_k_Cannot_call_write_after_write_eo, sizeof(__pyx_k_Cannot_call_write_after_write_eo), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_call_writelines_after_wri, __pyx_k_Cannot_call_writelines_after_wri, sizeof(__pyx_k_Cannot_call_writelines_after_wri), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_close_a_running_event_loo, __pyx_k_Cannot_close_a_running_event_loo, sizeof(__pyx_k_Cannot_close_a_running_event_loo), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_run_the_event_loop_while, __pyx_k_Cannot_run_the_event_loop_while, sizeof(__pyx_k_Cannot_run_the_event_loop_while), 0, 1, 0, 0},
    {&__pyx_kp_u_Cannot_serialize_socket_object, __pyx_k_Cannot_serialize_socket_object, sizeof(__pyx_k_Cannot_serialize_socket_object), 0, 1, 0, 0},
    {&__pyx_n_s_CertificateError, __pyx_k_CertificateError, sizeof(__pyx_k_CertificateError), 0, 0, 1, 1},
    {&__pyx_kp_u_Connect_call_failed, __pyx_k_Connect_call_failed, sizeof(__pyx_k_Connect_call_failed), 0, 1, 0, 0},
    {&__pyx_n_s_ConnectionAbortedError, __pyx_k_ConnectionAbortedError, sizeof(__pyx_k_ConnectionAbortedError), 0, 0, 1, 1},
    {&__pyx_n_s_ConnectionRefusedError, __pyx_k_ConnectionRefusedError, sizeof(__pyx_k_ConnectionRefusedError), 0, 0, 1, 1},
    {&__pyx_n_s_ConnectionResetError, __pyx_k_ConnectionResetError, sizeof(__pyx_k_ConnectionResetError), 0, 0, 1, 1},
    {&__pyx_n_s_Counter, __pyx_k_Counter, sizeof(__pyx_k_Counter), 0, 0, 1, 1},
    {&__pyx_kp_u_Creating__SSLProtocolTransport_t, __pyx_k_Creating__SSLProtocolTransport_t, sizeof(__pyx_k_Creating__SSLProtocolTransport_t), 0, 1, 0, 0},
    {&__pyx_n_s_DEVNULL, __pyx_k_DEVNULL, sizeof(__pyx_k_DEVNULL), 0, 0, 1, 1},
    {&__pyx_n_s_DeprecationWarning, __pyx_k_DeprecationWarning, sizeof(__pyx_k_DeprecationWarning), 0, 0, 1, 1},
    {&__pyx_n_u_EAI_ADDRFAMILY, __pyx_k_EAI_ADDRFAMILY, sizeof(__pyx_k_EAI_ADDRFAMILY), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_AGAIN, __pyx_k_EAI_AGAIN, sizeof(__pyx_k_EAI_AGAIN), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_BADFLAGS, __pyx_k_EAI_BADFLAGS, sizeof(__pyx_k_EAI_BADFLAGS), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_BADHINTS, __pyx_k_EAI_BADHINTS, sizeof(__pyx_k_EAI_BADHINTS), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_CANCELED, __pyx_k_EAI_CANCELED, sizeof(__pyx_k_EAI_CANCELED), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_FAIL, __pyx_k_EAI_FAIL, sizeof(__pyx_k_EAI_FAIL), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_FAMILY, __pyx_k_EAI_FAMILY, sizeof(__pyx_k_EAI_FAMILY), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_MEMORY, __pyx_k_EAI_MEMORY, sizeof(__pyx_k_EAI_MEMORY), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_NODATA, __pyx_k_EAI_NODATA, sizeof(__pyx_k_EAI_NODATA), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_NONAME, __pyx_k_EAI_NONAME, sizeof(__pyx_k_EAI_NONAME), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_OVERFLOW, __pyx_k_EAI_OVERFLOW, sizeof(__pyx_k_EAI_OVERFLOW), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_PROTOCOL, __pyx_k_EAI_PROTOCOL, sizeof(__pyx_k_EAI_PROTOCOL), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_SERVICE, __pyx_k_EAI_SERVICE, sizeof(__pyx_k_EAI_SERVICE), 0, 1, 0, 1},
    {&__pyx_n_u_EAI_SOCKTYPE, __pyx_k_EAI_SOCKTYPE, sizeof(__pyx_k_EAI_SOCKTYPE), 0, 1, 0, 1},
    {&__pyx_kp_u_Error_calling_eof_received, __pyx_k_Error_calling_eof_received, sizeof(__pyx_k_Error_calling_eof_received), 0, 1, 0, 0},
    {&__pyx_kp_u_Error_occurred_during_shutdown, __pyx_k_Error_occurred_during_shutdown, sizeof(__pyx_k_Error_occurred_during_shutdown), 0, 1, 0, 0},
    {&__pyx_kp_u_Errors_logged, __pyx_k_Errors_logged, sizeof(__pyx_k_Errors_logged), 0, 1, 0, 0},
    {&__pyx_kp_u_Event_loop_is_closed, __pyx_k_Event_loop_is_closed, sizeof(__pyx_k_Event_loop_is_closed), 0, 1, 0, 0},
    {&__pyx_kp_u_Event_loop_stopped_before_Future, __pyx_k_Event_loop_stopped_before_Future, sizeof(__pyx_k_Event_loop_stopped_before_Future), 0, 1, 0, 0},
    {&__pyx_kp_u_Exception_in___repr___r_value_ty, __pyx_k_Exception_in___repr___r_value_ty, sizeof(__pyx_k_Exception_in___repr___r_value_ty), 0, 1, 0, 0},
    {&__pyx_kp_u_Exception_in_callback, __pyx_k_Exception_in_callback, sizeof(__pyx_k_Exception_in_callback), 0, 1, 0, 0},
    {&__pyx_kp_u_Exception_in_default_exception_h, __pyx_k_Exception_in_default_exception_h, sizeof(__pyx_k_Exception_in_default_exception_h), 0, 1, 0, 0},
    {&__pyx_kp_u_Exception_in_default_exception_h_2, __pyx_k_Exception_in_default_exception_h_2, sizeof(__pyx_k_Exception_in_default_exception_h_2), 0, 1, 0, 0},
    {&__pyx_kp_u_Exception_occurred_in_preexec_fn, __pyx_k_Exception_occurred_in_preexec_fn, sizeof(__pyx_k_Exception_occurred_in_preexec_fn), 0, 1, 0, 0},
    {&__pyx_kp_u_Executing_r_took_3f_seconds, __pyx_k_Executing_r_took_3f_seconds, sizeof(__pyx_k_Executing_r_took_3f_seconds), 0, 1, 0, 0},
    {&__pyx_kp_u_Executing_s_took_3f_seconds, __pyx_k_Executing_s_took_3f_seconds, sizeof(__pyx_k_Executing_s_took_3f_seconds), 0, 1, 0, 0},
    {&__pyx_kp_u_Executor_shutdown_has_been_calle, __pyx_k_Executor_shutdown_has_been_calle, sizeof(__pyx_k_Executor_shutdown_has_been_calle), 0, 1, 0, 0},
    {&__pyx_kp_u_Fatal_error_on_SSL_protocol, __pyx_k_Fatal_error_on_SSL_protocol, sizeof(__pyx_k_Fatal_error_on_SSL_protocol), 0, 1, 0, 0},
    {&__pyx_kp_u_Fatal_error_on_server, __pyx_k_Fatal_error_on_server, sizeof(__pyx_k_Fatal_error_on_server), 0, 1, 0, 0},
    {&__pyx_kp_u_Fatal_error_on_transport, __pyx_k_Fatal_error_on_transport, sizeof(__pyx_k_Fatal_error_on_transport), 0, 1, 0, 0},
    {&__pyx_kp_u_Fatal_error_on_transport_2, __pyx_k_Fatal_error_on_transport_2, sizeof(__pyx_k_Fatal_error_on_transport_2), 0, 1, 0, 0},
    {&__pyx_kp_u_Fatal_write_error_on_datagram_tr, __pyx_k_Fatal_write_error_on_datagram_tr, sizeof(__pyx_k_Fatal_write_error_on_datagram_tr), 0, 1, 0, 0},
    {&__pyx_n_s_FileExistsError, __pyx_k_FileExistsError, sizeof(__pyx_k_FileExistsError), 0, 0, 1, 1},
    {&__pyx_n_s_FileNotFoundError, __pyx_k_FileNotFoundError, sizeof(__pyx_k_FileNotFoundError), 0, 0, 1, 1},
    {&__pyx_n_s_FileSystemEvent, __pyx_k_FileSystemEvent, sizeof(__pyx_k_FileSystemEvent), 0, 0, 1, 1},
    {&__pyx_kp_u_File_descriptor_r_is_used_by_tra, __pyx_k_File_descriptor_r_is_used_by_tra, sizeof(__pyx_k_File_descriptor_r_is_used_by_tra), 0, 1, 0, 0},
    {&__pyx_n_s_FlowControlMixin, __pyx_k_FlowControlMixin, sizeof(__pyx_k_FlowControlMixin), 0, 0, 1, 1},
    {&__pyx_n_s_Future, __pyx_k_Future, sizeof(__pyx_k_Future), 0, 0, 1, 1},
    {&__pyx_n_s_Handle, __pyx_k_Handle, sizeof(__pyx_k_Handle), 0, 0, 1, 1},
    {&__pyx_n_s_Handle___reduce_cython, __pyx_k_Handle___reduce_cython, sizeof(__pyx_k_Handle___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Handle___setstate_cython, __pyx_k_Handle___setstate_cython, sizeof(__pyx_k_Handle___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Handle_cancel, __pyx_k_Handle_cancel, sizeof(__pyx_k_Handle_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_Handle_cancelled, __pyx_k_Handle_cancelled, sizeof(__pyx_k_Handle_cancelled), 0, 0, 1, 1},
    {&__pyx_kp_u_Handle_loop_is_None_in_Handle, __pyx_k_Handle_loop_is_None_in_Handle, sizeof(__pyx_k_Handle_loop_is_None_in_Handle), 0, 1, 0, 0},
    {&__pyx_n_u_IPV6_V6ONLY, __pyx_k_IPV6_V6ONLY, sizeof(__pyx_k_IPV6_V6ONLY), 0, 1, 0, 1},
    {&__pyx_kp_u_IPv4_sockaddr_must_be_2_tuple, __pyx_k_IPv4_sockaddr_must_be_2_tuple, sizeof(__pyx_k_IPv4_sockaddr_must_be_2_tuple), 0, 1, 0, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
    {&__pyx_n_s_InterruptedError, __pyx_k_InterruptedError, sizeof(__pyx_k_InterruptedError), 0, 0, 1, 1},
    {&__pyx_n_s_InvalidStateError, __pyx_k_InvalidStateError, sizeof(__pyx_k_InvalidStateError), 0, 0, 1, 1},
    {&__pyx_kp_u_Invalid_address_must_be_None_or, __pyx_k_Invalid_address_must_be_None_or, sizeof(__pyx_k_Invalid_address_must_be_None_or), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_file_descriptor, __pyx_k_Invalid_file_descriptor, sizeof(__pyx_k_Invalid_file_descriptor), 0, 1, 0, 0},
    {&__pyx_kp_u_Invalid_file_object_r, __pyx_k_Invalid_file_object_r, sizeof(__pyx_k_Invalid_file_object_r), 0, 1, 0, 0},
    {&__pyx_n_s_IsADirectoryError, __pyx_k_IsADirectoryError, sizeof(__pyx_k_IsADirectoryError), 0, 0, 1, 1},
    {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
    {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
    {&__pyx_n_s_KeyboardInterrupt, __pyx_k_KeyboardInterrupt, sizeof(__pyx_k_KeyboardInterrupt), 0, 0, 1, 1},
    {&__pyx_kp_u_Listen_errors, __pyx_k_Listen_errors, sizeof(__pyx_k_Listen_errors), 0, 1, 0, 0},
    {&__pyx_n_s_Loop, __pyx_k_Loop, sizeof(__pyx_k_Loop), 0, 0, 1, 1},
    {&__pyx_n_s_Loop___get___locals_lambda, __pyx_k_Loop___get___locals_lambda, sizeof(__pyx_k_Loop___get___locals_lambda), 0, 0, 1, 1},
    {&__pyx_n_s_Loop___reduce_cython, __pyx_k_Loop___reduce_cython, sizeof(__pyx_k_Loop___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Loop___setstate_cython, __pyx_k_Loop___setstate_cython, sizeof(__pyx_k_Loop___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Loop___sighandler, __pyx_k_Loop___sighandler, sizeof(__pyx_k_Loop___sighandler), 0, 0, 1, 1},
    {&__pyx_n_s_Loop___subprocess_run, __pyx_k_Loop___subprocess_run, sizeof(__pyx_k_Loop___subprocess_run), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__asyncgen_finalizer_hook, __pyx_k_Loop__asyncgen_finalizer_hook, sizeof(__pyx_k_Loop__asyncgen_finalizer_hook), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__asyncgen_firstiter_hook, __pyx_k_Loop__asyncgen_firstiter_hook, sizeof(__pyx_k_Loop__asyncgen_firstiter_hook), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__check_default_executor, __pyx_k_Loop__check_default_executor, sizeof(__pyx_k_Loop__check_default_executor), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__do_shutdown, __pyx_k_Loop__do_shutdown, sizeof(__pyx_k_Loop__do_shutdown), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__get_backend_id, __pyx_k_Loop__get_backend_id, sizeof(__pyx_k_Loop__get_backend_id), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__getaddrinfo_locals_callbac, __pyx_k_Loop__getaddrinfo_locals_callbac, sizeof(__pyx_k_Loop__getaddrinfo_locals_callbac), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__getnameinfo_locals_callbac, __pyx_k_Loop__getnameinfo_locals_callbac, sizeof(__pyx_k_Loop__getnameinfo_locals_callbac), 0, 0, 1, 1},
    {&__pyx_n_s_Loop__monitor_fs, __pyx_k_Loop__monitor_fs, sizeof(__pyx_k_Loop__monitor_fs), 0, 0, 1, 1},
    {&__pyx_kp_u_Loop__read_from_self, __pyx_k_Loop__read_from_self, sizeof(__pyx_k_Loop__read_from_self), 0, 1, 0, 0},
    {&__pyx_n_s_Loop__sighandler, __pyx_k_Loop__sighandler, sizeof(__pyx_k_Loop__sighandler), 0, 0, 1, 1},
    {&__pyx_kp_u_Loop__sock_accept, __pyx_k_Loop__sock_accept, sizeof(__pyx_k_Loop__sock_accept), 0, 1, 0, 0},
    {&__pyx_kp_u_Loop__sock_connect, __pyx_k_Loop__sock_connect, sizeof(__pyx_k_Loop__sock_connect), 0, 1, 0, 0},
    {&__pyx_kp_u_Loop__sock_recv, __pyx_k_Loop__sock_recv, sizeof(__pyx_k_Loop__sock_recv), 0, 1, 0, 0},
    {&__pyx_kp_u_Loop__sock_recv_into, __pyx_k_Loop__sock_recv_into, sizeof(__pyx_k_Loop__sock_recv_into), 0, 1, 0, 0},
    {&__pyx_kp_u_Loop__sock_sendall, __pyx_k_Loop__sock_sendall, sizeof(__pyx_k_Loop__sock_sendall), 0, 1, 0, 0},
    {&__pyx_kp_u_Loop__stop, __pyx_k_Loop__stop, sizeof(__pyx_k_Loop__stop), 0, 1, 0, 0},
    {&__pyx_n_s_Loop__subprocess_run, __pyx_k_Loop__subprocess_run, sizeof(__pyx_k_Loop__subprocess_run), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_add_reader, __pyx_k_Loop_add_reader, sizeof(__pyx_k_Loop_add_reader), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_add_signal_handler, __pyx_k_Loop_add_signal_handler, sizeof(__pyx_k_Loop_add_signal_handler), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_add_writer, __pyx_k_Loop_add_writer, sizeof(__pyx_k_Loop_add_writer), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_call_at, __pyx_k_Loop_call_at, sizeof(__pyx_k_Loop_call_at), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_call_exception_handler, __pyx_k_Loop_call_exception_handler, sizeof(__pyx_k_Loop_call_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_call_later, __pyx_k_Loop_call_later, sizeof(__pyx_k_Loop_call_later), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_call_soon, __pyx_k_Loop_call_soon, sizeof(__pyx_k_Loop_call_soon), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_call_soon_threadsafe, __pyx_k_Loop_call_soon_threadsafe, sizeof(__pyx_k_Loop_call_soon_threadsafe), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_close, __pyx_k_Loop_close, sizeof(__pyx_k_Loop_close), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_connect_accepted_socket, __pyx_k_Loop_connect_accepted_socket, sizeof(__pyx_k_Loop_connect_accepted_socket), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_connect_read_pipe, __pyx_k_Loop_connect_read_pipe, sizeof(__pyx_k_Loop_connect_read_pipe), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_connect_write_pipe, __pyx_k_Loop_connect_write_pipe, sizeof(__pyx_k_Loop_connect_write_pipe), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_create_connection, __pyx_k_Loop_create_connection, sizeof(__pyx_k_Loop_create_connection), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_create_datagram_endpoint, __pyx_k_Loop_create_datagram_endpoint, sizeof(__pyx_k_Loop_create_datagram_endpoint), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_create_future, __pyx_k_Loop_create_future, sizeof(__pyx_k_Loop_create_future), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_create_server, __pyx_k_Loop_create_server, sizeof(__pyx_k_Loop_create_server), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_create_task, __pyx_k_Loop_create_task, sizeof(__pyx_k_Loop_create_task), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_create_unix_connection, __pyx_k_Loop_create_unix_connection, sizeof(__pyx_k_Loop_create_unix_connection), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_create_unix_server, __pyx_k_Loop_create_unix_server, sizeof(__pyx_k_Loop_create_unix_server), 0, 0, 1, 1},
    {&__pyx_kp_u_Loop_debug_info, __pyx_k_Loop_debug_info, sizeof(__pyx_k_Loop_debug_info), 0, 1, 0, 0},
    {&__pyx_n_s_Loop_default_exception_handler, __pyx_k_Loop_default_exception_handler, sizeof(__pyx_k_Loop_default_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_get_debug, __pyx_k_Loop_get_debug, sizeof(__pyx_k_Loop_get_debug), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_get_exception_handler, __pyx_k_Loop_get_exception_handler, sizeof(__pyx_k_Loop_get_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_get_task_factory, __pyx_k_Loop_get_task_factory, sizeof(__pyx_k_Loop_get_task_factory), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_getaddrinfo, __pyx_k_Loop_getaddrinfo, sizeof(__pyx_k_Loop_getaddrinfo), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_getnameinfo, __pyx_k_Loop_getnameinfo, sizeof(__pyx_k_Loop_getnameinfo), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_is_closed, __pyx_k_Loop_is_closed, sizeof(__pyx_k_Loop_is_closed), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_is_running, __pyx_k_Loop_is_running, sizeof(__pyx_k_Loop_is_running), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_remove_reader, __pyx_k_Loop_remove_reader, sizeof(__pyx_k_Loop_remove_reader), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_remove_signal_handler, __pyx_k_Loop_remove_signal_handler, sizeof(__pyx_k_Loop_remove_signal_handler), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_remove_writer, __pyx_k_Loop_remove_writer, sizeof(__pyx_k_Loop_remove_writer), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_run_forever, __pyx_k_Loop_run_forever, sizeof(__pyx_k_Loop_run_forever), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_run_in_executor, __pyx_k_Loop_run_in_executor, sizeof(__pyx_k_Loop_run_in_executor), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_run_until_complete, __pyx_k_Loop_run_until_complete, sizeof(__pyx_k_Loop_run_until_complete), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_set_debug, __pyx_k_Loop_set_debug, sizeof(__pyx_k_Loop_set_debug), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_set_default_executor, __pyx_k_Loop_set_default_executor, sizeof(__pyx_k_Loop_set_default_executor), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_set_exception_handler, __pyx_k_Loop_set_exception_handler, sizeof(__pyx_k_Loop_set_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_set_task_factory, __pyx_k_Loop_set_task_factory, sizeof(__pyx_k_Loop_set_task_factory), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_shutdown_asyncgens, __pyx_k_Loop_shutdown_asyncgens, sizeof(__pyx_k_Loop_shutdown_asyncgens), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_shutdown_default_executor, __pyx_k_Loop_shutdown_default_executor, sizeof(__pyx_k_Loop_shutdown_default_executor), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_accept, __pyx_k_Loop_sock_accept, sizeof(__pyx_k_Loop_sock_accept), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_connect, __pyx_k_Loop_sock_connect, sizeof(__pyx_k_Loop_sock_connect), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_recv, __pyx_k_Loop_sock_recv, sizeof(__pyx_k_Loop_sock_recv), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_recv_into, __pyx_k_Loop_sock_recv_into, sizeof(__pyx_k_Loop_sock_recv_into), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_recvfrom, __pyx_k_Loop_sock_recvfrom, sizeof(__pyx_k_Loop_sock_recvfrom), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_recvfrom_into, __pyx_k_Loop_sock_recvfrom_into, sizeof(__pyx_k_Loop_sock_recvfrom_into), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_sendall, __pyx_k_Loop_sock_sendall, sizeof(__pyx_k_Loop_sock_sendall), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_sock_sendto, __pyx_k_Loop_sock_sendto, sizeof(__pyx_k_Loop_sock_sendto), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_start_tls, __pyx_k_Loop_start_tls, sizeof(__pyx_k_Loop_start_tls), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_stop, __pyx_k_Loop_stop, sizeof(__pyx_k_Loop_stop), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_subprocess_exec, __pyx_k_Loop_subprocess_exec, sizeof(__pyx_k_Loop_subprocess_exec), 0, 0, 1, 1},
    {&__pyx_n_s_Loop_subprocess_shell, __pyx_k_Loop_subprocess_shell, sizeof(__pyx_k_Loop_subprocess_shell), 0, 0, 1, 1},
    {&__pyx_kp_u_Loop_time, __pyx_k_Loop_time, sizeof(__pyx_k_Loop_time), 0, 1, 0, 0},
    {&__pyx_n_s_Loop_time_2, __pyx_k_Loop_time_2, sizeof(__pyx_k_Loop_time_2), 0, 0, 1, 1},
    {&__pyx_n_s_LruCache, __pyx_k_LruCache, sizeof(__pyx_k_LruCache), 0, 0, 1, 1},
    {&__pyx_n_s_LruCache___reduce_cython, __pyx_k_LruCache___reduce_cython, sizeof(__pyx_k_LruCache___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_LruCache___setstate_cython, __pyx_k_LruCache___setstate_cython, sizeof(__pyx_k_LruCache___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryBIO, __pyx_k_MemoryBIO, sizeof(__pyx_k_MemoryBIO), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_kp_u_Multiple_exceptions, __pyx_k_Multiple_exceptions, sizeof(__pyx_k_Multiple_exceptions), 0, 1, 0, 0},
    {&__pyx_n_s_NSIG, __pyx_k_NSIG, sizeof(__pyx_k_NSIG), 0, 0, 1, 1},
    {&__pyx_n_s_NameInfoRequest, __pyx_k_NameInfoRequest, sizeof(__pyx_k_NameInfoRequest), 0, 0, 1, 1},
    {&__pyx_n_s_NameInfoRequest___reduce_cython, __pyx_k_NameInfoRequest___reduce_cython, sizeof(__pyx_k_NameInfoRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_NameInfoRequest___setstate_cytho, __pyx_k_NameInfoRequest___setstate_cytho, sizeof(__pyx_k_NameInfoRequest___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_kp_u_Neither_host_port_nor_sock_were, __pyx_k_Neither_host_port_nor_sock_were, sizeof(__pyx_k_Neither_host_port_nor_sock_were), 0, 1, 0, 0},
    {&__pyx_kp_u_Non_thread_safe_operation_invoke, __pyx_k_Non_thread_safe_operation_invoke, sizeof(__pyx_k_Non_thread_safe_operation_invoke), 0, 1, 0, 0},
    {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
    {&__pyx_kp_u_Number_of_signals, __pyx_k_Number_of_signals, sizeof(__pyx_k_Number_of_signals), 0, 1, 0, 0},
    {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
    {&__pyx_n_s_O_RDWR, __pyx_k_O_RDWR, sizeof(__pyx_k_O_RDWR), 0, 0, 1, 1},
    {&__pyx_kp_u_Object_created_at_most_recent_ca, __pyx_k_Object_created_at_most_recent_ca, sizeof(__pyx_k_Object_created_at_most_recent_ca), 0, 1, 0, 0},
    {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
    {&__pyx_n_s_OverflowError, __pyx_k_OverflowError, sizeof(__pyx_k_OverflowError), 0, 0, 1, 1},
    {&__pyx_n_s_PIPE, __pyx_k_PIPE, sizeof(__pyx_k_PIPE), 0, 0, 1, 1},
    {&__pyx_n_u_PYTHONASYNCIODEBUG, __pyx_k_PYTHONASYNCIODEBUG, sizeof(__pyx_k_PYTHONASYNCIODEBUG), 0, 1, 0, 1},
    {&__pyx_kp_u_Passing_reuse_address_True_is_no, __pyx_k_Passing_reuse_address_True_is_no, sizeof(__pyx_k_Passing_reuse_address_True_is_no), 0, 1, 0, 0},
    {&__pyx_n_s_PermissionError, __pyx_k_PermissionError, sizeof(__pyx_k_PermissionError), 0, 0, 1, 1},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_kp_s_Pickling_of_struct_members_such, __pyx_k_Pickling_of_struct_members_such, sizeof(__pyx_k_Pickling_of_struct_members_such), 0, 0, 1, 0},
    {&__pyx_n_s_PipeConnectRequest, __pyx_k_PipeConnectRequest, sizeof(__pyx_k_PipeConnectRequest), 0, 0, 1, 1},
    {&__pyx_n_s_PipeConnectRequest___reduce_cyt, __pyx_k_PipeConnectRequest___reduce_cyt, sizeof(__pyx_k_PipeConnectRequest___reduce_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_PipeConnectRequest___setstate_c, __pyx_k_PipeConnectRequest___setstate_c, sizeof(__pyx_k_PipeConnectRequest___setstate_c), 0, 0, 1, 1},
    {&__pyx_kp_u_Polls_debug_info, __pyx_k_Polls_debug_info, sizeof(__pyx_k_Polls_debug_info), 0, 1, 0, 0},
    {&__pyx_n_s_ProcessLookupError, __pyx_k_ProcessLookupError, sizeof(__pyx_k_ProcessLookupError), 0, 0, 1, 1},
    {&__pyx_kp_u_Process_info, __pyx_k_Process_info, sizeof(__pyx_k_Process_info), 0, 1, 0, 0},
    {&__pyx_kp_u_Process_memory, __pyx_k_Process_memory, sizeof(__pyx_k_Process_memory), 0, 1, 0, 0},
    {&__pyx_n_s_Protocol, __pyx_k_Protocol, sizeof(__pyx_k_Protocol), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket, __pyx_k_PseudoSocket, sizeof(__pyx_k_PseudoSocket), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket___enter, __pyx_k_PseudoSocket___enter, sizeof(__pyx_k_PseudoSocket___enter), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket___exit, __pyx_k_PseudoSocket___exit, sizeof(__pyx_k_PseudoSocket___exit), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket___getstate, __pyx_k_PseudoSocket___getstate, sizeof(__pyx_k_PseudoSocket___getstate), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket___reduce_cython, __pyx_k_PseudoSocket___reduce_cython, sizeof(__pyx_k_PseudoSocket___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket___setstate_cython, __pyx_k_PseudoSocket___setstate_cython, sizeof(__pyx_k_PseudoSocket___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_accept, __pyx_k_PseudoSocket_accept, sizeof(__pyx_k_PseudoSocket_accept), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_bind, __pyx_k_PseudoSocket_bind, sizeof(__pyx_k_PseudoSocket_bind), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_close, __pyx_k_PseudoSocket_close, sizeof(__pyx_k_PseudoSocket_close), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_connect, __pyx_k_PseudoSocket_connect, sizeof(__pyx_k_PseudoSocket_connect), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_connect_ex, __pyx_k_PseudoSocket_connect_ex, sizeof(__pyx_k_PseudoSocket_connect_ex), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_detach, __pyx_k_PseudoSocket_detach, sizeof(__pyx_k_PseudoSocket_detach), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_dup, __pyx_k_PseudoSocket_dup, sizeof(__pyx_k_PseudoSocket_dup), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_fileno, __pyx_k_PseudoSocket_fileno, sizeof(__pyx_k_PseudoSocket_fileno), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_get_inheritable, __pyx_k_PseudoSocket_get_inheritable, sizeof(__pyx_k_PseudoSocket_get_inheritable), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_getpeername, __pyx_k_PseudoSocket_getpeername, sizeof(__pyx_k_PseudoSocket_getpeername), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_getsockname, __pyx_k_PseudoSocket_getsockname, sizeof(__pyx_k_PseudoSocket_getsockname), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_getsockopt, __pyx_k_PseudoSocket_getsockopt, sizeof(__pyx_k_PseudoSocket_getsockopt), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_gettimeout, __pyx_k_PseudoSocket_gettimeout, sizeof(__pyx_k_PseudoSocket_gettimeout), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_ioctl, __pyx_k_PseudoSocket_ioctl, sizeof(__pyx_k_PseudoSocket_ioctl), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_listen, __pyx_k_PseudoSocket_listen, sizeof(__pyx_k_PseudoSocket_listen), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_makefile, __pyx_k_PseudoSocket_makefile, sizeof(__pyx_k_PseudoSocket_makefile), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_recv, __pyx_k_PseudoSocket_recv, sizeof(__pyx_k_PseudoSocket_recv), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_recv_into, __pyx_k_PseudoSocket_recv_into, sizeof(__pyx_k_PseudoSocket_recv_into), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_recvfrom, __pyx_k_PseudoSocket_recvfrom, sizeof(__pyx_k_PseudoSocket_recvfrom), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_recvfrom_into, __pyx_k_PseudoSocket_recvfrom_into, sizeof(__pyx_k_PseudoSocket_recvfrom_into), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_recvmsg, __pyx_k_PseudoSocket_recvmsg, sizeof(__pyx_k_PseudoSocket_recvmsg), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_recvmsg_into, __pyx_k_PseudoSocket_recvmsg_into, sizeof(__pyx_k_PseudoSocket_recvmsg_into), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_send, __pyx_k_PseudoSocket_send, sizeof(__pyx_k_PseudoSocket_send), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_sendall, __pyx_k_PseudoSocket_sendall, sizeof(__pyx_k_PseudoSocket_sendall), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_sendfile, __pyx_k_PseudoSocket_sendfile, sizeof(__pyx_k_PseudoSocket_sendfile), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_sendmsg, __pyx_k_PseudoSocket_sendmsg, sizeof(__pyx_k_PseudoSocket_sendmsg), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_sendmsg_afalg, __pyx_k_PseudoSocket_sendmsg_afalg, sizeof(__pyx_k_PseudoSocket_sendmsg_afalg), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_sendto, __pyx_k_PseudoSocket_sendto, sizeof(__pyx_k_PseudoSocket_sendto), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_set_inheritable, __pyx_k_PseudoSocket_set_inheritable, sizeof(__pyx_k_PseudoSocket_set_inheritable), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_setblocking, __pyx_k_PseudoSocket_setblocking, sizeof(__pyx_k_PseudoSocket_setblocking), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_setsockopt, __pyx_k_PseudoSocket_setsockopt, sizeof(__pyx_k_PseudoSocket_setsockopt), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_settimeout, __pyx_k_PseudoSocket_settimeout, sizeof(__pyx_k_PseudoSocket_settimeout), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_share, __pyx_k_PseudoSocket_share, sizeof(__pyx_k_PseudoSocket_share), 0, 0, 1, 1},
    {&__pyx_n_s_PseudoSocket_shutdown, __pyx_k_PseudoSocket_shutdown, sizeof(__pyx_k_PseudoSocket_shutdown), 0, 0, 1, 1},
    {&__pyx_n_s_Pyx_CFunc_6uvloop_4loop_4Loop, __pyx_k_Pyx_CFunc_6uvloop_4loop_4Loop, sizeof(__pyx_k_Pyx_CFunc_6uvloop_4loop_4Loop), 0, 0, 1, 1},
    {&__pyx_n_s_RENAME, __pyx_k_RENAME, sizeof(__pyx_k_RENAME), 0, 0, 1, 1},
    {&__pyx_n_s_RENAME_CHANGE, __pyx_k_RENAME_CHANGE, sizeof(__pyx_k_RENAME_CHANGE), 0, 0, 1, 1},
    {&__pyx_kp_u_Racing_with_another_loop_to_spaw, __pyx_k_Racing_with_another_loop_to_spaw, sizeof(__pyx_k_Racing_with_another_loop_to_spaw), 0, 1, 0, 0},
    {&__pyx_n_s_ReadSubprocessPipeProto, __pyx_k_ReadSubprocessPipeProto, sizeof(__pyx_k_ReadSubprocessPipeProto), 0, 0, 1, 1},
    {&__pyx_n_s_ReadSubprocessPipeProto_data_rec, __pyx_k_ReadSubprocessPipeProto_data_rec, sizeof(__pyx_k_ReadSubprocessPipeProto_data_rec), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport, __pyx_k_ReadUnixTransport, sizeof(__pyx_k_ReadUnixTransport), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport___reduce_cytho, __pyx_k_ReadUnixTransport___reduce_cytho, sizeof(__pyx_k_ReadUnixTransport___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport___setstate_cyt, __pyx_k_ReadUnixTransport___setstate_cyt, sizeof(__pyx_k_ReadUnixTransport___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_abort, __pyx_k_ReadUnixTransport_abort, sizeof(__pyx_k_ReadUnixTransport_abort), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_can_write_eof, __pyx_k_ReadUnixTransport_can_write_eof, sizeof(__pyx_k_ReadUnixTransport_can_write_eof), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_get_write_buff, __pyx_k_ReadUnixTransport_get_write_buff, sizeof(__pyx_k_ReadUnixTransport_get_write_buff), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_get_write_buff_2, __pyx_k_ReadUnixTransport_get_write_buff_2, sizeof(__pyx_k_ReadUnixTransport_get_write_buff_2), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_set_write_buff, __pyx_k_ReadUnixTransport_set_write_buff, sizeof(__pyx_k_ReadUnixTransport_set_write_buff), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_write, __pyx_k_ReadUnixTransport_write, sizeof(__pyx_k_ReadUnixTransport_write), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_write_eof, __pyx_k_ReadUnixTransport_write_eof, sizeof(__pyx_k_ReadUnixTransport_write_eof), 0, 0, 1, 1},
    {&__pyx_n_s_ReadUnixTransport_writelines, __pyx_k_ReadUnixTransport_writelines, sizeof(__pyx_k_ReadUnixTransport_writelines), 0, 0, 1, 1},
    {&__pyx_kp_u_Read_EOF_failed_callbacks, __pyx_k_Read_EOF_failed_callbacks, sizeof(__pyx_k_Read_EOF_failed_callbacks), 0, 1, 0, 0},
    {&__pyx_kp_u_Read_EOFs, __pyx_k_Read_EOFs, sizeof(__pyx_k_Read_EOFs), 0, 1, 0, 0},
    {&__pyx_kp_u_Read_callbacks, __pyx_k_Read_callbacks, sizeof(__pyx_k_Read_callbacks), 0, 1, 0, 0},
    {&__pyx_kp_u_Read_callbacks_failed, __pyx_k_Read_callbacks_failed, sizeof(__pyx_k_Read_callbacks_failed), 0, 1, 0, 0},
    {&__pyx_kp_u_Read_errors, __pyx_k_Read_errors, sizeof(__pyx_k_Read_errors), 0, 1, 0, 0},
    {&__pyx_kp_u_Read_events, __pyx_k_Read_events, sizeof(__pyx_k_Read_events), 0, 1, 0, 0},
    {&__pyx_kp_u_Read_failed_callbacks, __pyx_k_Read_failed_callbacks, sizeof(__pyx_k_Read_failed_callbacks), 0, 1, 0, 0},
    {&__pyx_n_s_ResourceWarning, __pyx_k_ResourceWarning, sizeof(__pyx_k_ResourceWarning), 0, 0, 1, 1},
    {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
    {&__pyx_n_s_RuntimeWarning, __pyx_k_RuntimeWarning, sizeof(__pyx_k_RuntimeWarning), 0, 0, 1, 1},
    {&__pyx_n_s_SIG_DFL, __pyx_k_SIG_DFL, sizeof(__pyx_k_SIG_DFL), 0, 0, 1, 1},
    {&__pyx_n_u_SOCK_NONBLOCK, __pyx_k_SOCK_NONBLOCK, sizeof(__pyx_k_SOCK_NONBLOCK), 0, 1, 0, 1},
    {&__pyx_n_u_SO_BROADCAST, __pyx_k_SO_BROADCAST, sizeof(__pyx_k_SO_BROADCAST), 0, 1, 0, 1},
    {&__pyx_n_u_SO_REUSEPORT, __pyx_k_SO_REUSEPORT, sizeof(__pyx_k_SO_REUSEPORT), 0, 1, 0, 1},
    {&__pyx_n_s_SSLContext, __pyx_k_SSLContext, sizeof(__pyx_k_SSLContext), 0, 0, 1, 1},
    {&__pyx_n_s_SSLError, __pyx_k_SSLError, sizeof(__pyx_k_SSLError), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol, __pyx_k_SSLProtocol, sizeof(__pyx_k_SSLProtocol), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport, __pyx_k_SSLProtocolTransport, sizeof(__pyx_k_SSLProtocolTransport), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport___reduce_c, __pyx_k_SSLProtocolTransport___reduce_c, sizeof(__pyx_k_SSLProtocolTransport___reduce_c), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport___setstate, __pyx_k_SSLProtocolTransport___setstate, sizeof(__pyx_k_SSLProtocolTransport___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport__force_clo, __pyx_k_SSLProtocolTransport__force_clo, sizeof(__pyx_k_SSLProtocolTransport__force_clo), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport__test__app, __pyx_k_SSLProtocolTransport__test__app, sizeof(__pyx_k_SSLProtocolTransport__test__app), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_abort, __pyx_k_SSLProtocolTransport_abort, sizeof(__pyx_k_SSLProtocolTransport_abort), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_can_write, __pyx_k_SSLProtocolTransport_can_write, sizeof(__pyx_k_SSLProtocolTransport_can_write), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_close, __pyx_k_SSLProtocolTransport_close, sizeof(__pyx_k_SSLProtocolTransport_close), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_get_extra, __pyx_k_SSLProtocolTransport_get_extra, sizeof(__pyx_k_SSLProtocolTransport_get_extra), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_get_protoc, __pyx_k_SSLProtocolTransport_get_protoc, sizeof(__pyx_k_SSLProtocolTransport_get_protoc), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_get_read_b, __pyx_k_SSLProtocolTransport_get_read_b, sizeof(__pyx_k_SSLProtocolTransport_get_read_b), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_get_read_b_2, __pyx_k_SSLProtocolTransport_get_read_b_2, sizeof(__pyx_k_SSLProtocolTransport_get_read_b_2), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_get_write, __pyx_k_SSLProtocolTransport_get_write, sizeof(__pyx_k_SSLProtocolTransport_get_write), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_get_write_2, __pyx_k_SSLProtocolTransport_get_write_2, sizeof(__pyx_k_SSLProtocolTransport_get_write_2), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_is_closing, __pyx_k_SSLProtocolTransport_is_closing, sizeof(__pyx_k_SSLProtocolTransport_is_closing), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_is_reading, __pyx_k_SSLProtocolTransport_is_reading, sizeof(__pyx_k_SSLProtocolTransport_is_reading), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_pause_read, __pyx_k_SSLProtocolTransport_pause_read, sizeof(__pyx_k_SSLProtocolTransport_pause_read), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_resume_rea, __pyx_k_SSLProtocolTransport_resume_rea, sizeof(__pyx_k_SSLProtocolTransport_resume_rea), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_set_protoc, __pyx_k_SSLProtocolTransport_set_protoc, sizeof(__pyx_k_SSLProtocolTransport_set_protoc), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_set_read_b, __pyx_k_SSLProtocolTransport_set_read_b, sizeof(__pyx_k_SSLProtocolTransport_set_read_b), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_set_write, __pyx_k_SSLProtocolTransport_set_write, sizeof(__pyx_k_SSLProtocolTransport_set_write), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_write, __pyx_k_SSLProtocolTransport_write, sizeof(__pyx_k_SSLProtocolTransport_write), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_write_eof, __pyx_k_SSLProtocolTransport_write_eof, sizeof(__pyx_k_SSLProtocolTransport_write_eof), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocolTransport_writelines, __pyx_k_SSLProtocolTransport_writelines, sizeof(__pyx_k_SSLProtocolTransport_writelines), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol___reduce_cython, __pyx_k_SSLProtocol___reduce_cython, sizeof(__pyx_k_SSLProtocol___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol___setstate_cython, __pyx_k_SSLProtocol___setstate_cython, sizeof(__pyx_k_SSLProtocol___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_SSLProtocol__do_read, __pyx_k_SSLProtocol__do_read, sizeof(__pyx_k_SSLProtocol__do_read), 0, 1, 0, 0},
    {&__pyx_n_s_SSLProtocol__get_app_transport, __pyx_k_SSLProtocol__get_app_transport, sizeof(__pyx_k_SSLProtocol__get_app_transport), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol__start_handshake_loc, __pyx_k_SSLProtocol__start_handshake_loc, sizeof(__pyx_k_SSLProtocol__start_handshake_loc), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol__start_shutdown_loca, __pyx_k_SSLProtocol__start_shutdown_loca, sizeof(__pyx_k_SSLProtocol__start_shutdown_loca), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol_buffer_updated, __pyx_k_SSLProtocol_buffer_updated, sizeof(__pyx_k_SSLProtocol_buffer_updated), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol_connection_lost, __pyx_k_SSLProtocol_connection_lost, sizeof(__pyx_k_SSLProtocol_connection_lost), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol_connection_made, __pyx_k_SSLProtocol_connection_made, sizeof(__pyx_k_SSLProtocol_connection_made), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol_eof_received, __pyx_k_SSLProtocol_eof_received, sizeof(__pyx_k_SSLProtocol_eof_received), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol_get_buffer, __pyx_k_SSLProtocol_get_buffer, sizeof(__pyx_k_SSLProtocol_get_buffer), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol_pause_writing, __pyx_k_SSLProtocol_pause_writing, sizeof(__pyx_k_SSLProtocol_pause_writing), 0, 0, 1, 1},
    {&__pyx_n_s_SSLProtocol_resume_writing, __pyx_k_SSLProtocol_resume_writing, sizeof(__pyx_k_SSLProtocol_resume_writing), 0, 0, 1, 1},
    {&__pyx_n_s_SSLSyscallError, __pyx_k_SSLSyscallError, sizeof(__pyx_k_SSLSyscallError), 0, 0, 1, 1},
    {&__pyx_n_s_SSLWantReadError, __pyx_k_SSLWantReadError, sizeof(__pyx_k_SSLWantReadError), 0, 0, 1, 1},
    {&__pyx_n_s_SSLZeroReturnError, __pyx_k_SSLZeroReturnError, sizeof(__pyx_k_SSLZeroReturnError), 0, 0, 1, 1},
    {&__pyx_n_s_SSL_ERROR_SYSCALL, __pyx_k_SSL_ERROR_SYSCALL, sizeof(__pyx_k_SSL_ERROR_SYSCALL), 0, 0, 1, 1},
    {&__pyx_n_s_SSL_ERROR_WANT_READ, __pyx_k_SSL_ERROR_WANT_READ, sizeof(__pyx_k_SSL_ERROR_WANT_READ), 0, 0, 1, 1},
    {&__pyx_n_s_SSL_ERROR_WANT_WRITE, __pyx_k_SSL_ERROR_WANT_WRITE, sizeof(__pyx_k_SSL_ERROR_WANT_WRITE), 0, 0, 1, 1},
    {&__pyx_kp_u_SSL_connection_is_closed, __pyx_k_SSL_connection_is_closed, sizeof(__pyx_k_SSL_connection_is_closed), 0, 1, 0, 0},
    {&__pyx_kp_u_SSL_handshake_failed, __pyx_k_SSL_handshake_failed, sizeof(__pyx_k_SSL_handshake_failed), 0, 1, 0, 0},
    {&__pyx_kp_u_SSL_handshake_failed_on_verifyin, __pyx_k_SSL_handshake_failed_on_verifyin, sizeof(__pyx_k_SSL_handshake_failed_on_verifyin), 0, 1, 0, 0},
    {&__pyx_kp_u_SSL_handshake_is_taking_longer_t, __pyx_k_SSL_handshake_is_taking_longer_t, sizeof(__pyx_k_SSL_handshake_is_taking_longer_t), 0, 1, 0, 0},
    {&__pyx_kp_u_SSL_shutdown_timed_out, __pyx_k_SSL_shutdown_timed_out, sizeof(__pyx_k_SSL_shutdown_timed_out), 0, 1, 0, 0},
    {&__pyx_n_s_STDOUT, __pyx_k_STDOUT, sizeof(__pyx_k_STDOUT), 0, 0, 1, 1},
    {&__pyx_n_s_S_ISSOCK, __pyx_k_S_ISSOCK, sizeof(__pyx_k_S_ISSOCK), 0, 0, 1, 1},
    {&__pyx_n_s_Server, __pyx_k_Server, sizeof(__pyx_k_Server), 0, 0, 1, 1},
    {&__pyx_n_s_Server___aenter, __pyx_k_Server___aenter, sizeof(__pyx_k_Server___aenter), 0, 0, 1, 1},
    {&__pyx_n_s_Server___aexit, __pyx_k_Server___aexit, sizeof(__pyx_k_Server___aexit), 0, 0, 1, 1},
    {&__pyx_n_s_Server___reduce_cython, __pyx_k_Server___reduce_cython, sizeof(__pyx_k_Server___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Server___setstate_cython, __pyx_k_Server___setstate_cython, sizeof(__pyx_k_Server___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Server_close, __pyx_k_Server_close, sizeof(__pyx_k_Server_close), 0, 0, 1, 1},
    {&__pyx_n_s_Server_get_loop, __pyx_k_Server_get_loop, sizeof(__pyx_k_Server_get_loop), 0, 0, 1, 1},
    {&__pyx_n_s_Server_is_serving, __pyx_k_Server_is_serving, sizeof(__pyx_k_Server_is_serving), 0, 0, 1, 1},
    {&__pyx_n_s_Server_serve_forever, __pyx_k_Server_serve_forever, sizeof(__pyx_k_Server_serve_forever), 0, 0, 1, 1},
    {&__pyx_kp_u_Server_side_SSL_needs_a_valid_SS, __pyx_k_Server_side_SSL_needs_a_valid_SS, sizeof(__pyx_k_Server_side_SSL_needs_a_valid_SS), 0, 1, 0, 0},
    {&__pyx_n_s_Server_start_serving, __pyx_k_Server_start_serving, sizeof(__pyx_k_Server_start_serving), 0, 0, 1, 1},
    {&__pyx_n_s_Server_wait_closed, __pyx_k_Server_wait_closed, sizeof(__pyx_k_Server_wait_closed), 0, 0, 1, 1},
    {&__pyx_kp_u_Shutdown_errors, __pyx_k_Shutdown_errors, sizeof(__pyx_k_Shutdown_errors), 0, 1, 0, 0},
    {&__pyx_n_s_SockAddrHolder, __pyx_k_SockAddrHolder, sizeof(__pyx_k_SockAddrHolder), 0, 0, 1, 1},
    {&__pyx_n_s_SockAddrHolder___reduce_cython, __pyx_k_SockAddrHolder___reduce_cython, sizeof(__pyx_k_SockAddrHolder___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SockAddrHolder___setstate_cython, __pyx_k_SockAddrHolder___setstate_cython, sizeof(__pyx_k_SockAddrHolder___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_Sock_ops_successful_on_1st_try, __pyx_k_Sock_ops_successful_on_1st_try, sizeof(__pyx_k_Sock_ops_successful_on_1st_try), 0, 1, 0, 0},
    {&__pyx_n_s_SocketKind, __pyx_k_SocketKind, sizeof(__pyx_k_SocketKind), 0, 0, 1, 1},
    {&__pyx_kp_u_Socket_try_writes, __pyx_k_Socket_try_writes, sizeof(__pyx_k_Socket_try_writes), 0, 1, 0, 0},
    {&__pyx_n_s_StackSummary, __pyx_k_StackSummary, sizeof(__pyx_k_StackSummary), 0, 0, 1, 1},
    {&__pyx_n_s_StreamWriteContext, __pyx_k_StreamWriteContext, sizeof(__pyx_k_StreamWriteContext), 0, 0, 1, 1},
    {&__pyx_n_s_StreamWriteContext___reduce_cyt, __pyx_k_StreamWriteContext___reduce_cyt, sizeof(__pyx_k_StreamWriteContext___reduce_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_StreamWriteContext___setstate_c, __pyx_k_StreamWriteContext___setstate_c, sizeof(__pyx_k_StreamWriteContext___setstate_c), 0, 0, 1, 1},
    {&__pyx_kp_u_StreamWriteContext_close_py_buf, __pyx_k_StreamWriteContext_close_py_buf, sizeof(__pyx_k_StreamWriteContext_close_py_buf), 0, 1, 0, 0},
    {&__pyx_kp_u_StreamWriteContext_close_uv_buf, __pyx_k_StreamWriteContext_close_uv_buf, sizeof(__pyx_k_StreamWriteContext_close_uv_buf), 0, 1, 0, 0},
    {&__pyx_kp_u_Streams_debug_info, __pyx_k_Streams_debug_info, sizeof(__pyx_k_Streams_debug_info), 0, 1, 0, 0},
    {&__pyx_n_s_SubprocessError, __pyx_k_SubprocessError, sizeof(__pyx_k_SubprocessError), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketReaderFuture, __pyx_k_SyncSocketReaderFuture, sizeof(__pyx_k_SyncSocketReaderFuture), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketReaderFuture___init, __pyx_k_SyncSocketReaderFuture___init, sizeof(__pyx_k_SyncSocketReaderFuture___init), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketReaderFuture___remove, __pyx_k_SyncSocketReaderFuture___remove, sizeof(__pyx_k_SyncSocketReaderFuture___remove), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketReaderFuture__loop, __pyx_k_SyncSocketReaderFuture__loop, sizeof(__pyx_k_SyncSocketReaderFuture__loop), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketReaderFuture__remove, __pyx_k_SyncSocketReaderFuture__remove, sizeof(__pyx_k_SyncSocketReaderFuture__remove), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketReaderFuture__sock, __pyx_k_SyncSocketReaderFuture__sock, sizeof(__pyx_k_SyncSocketReaderFuture__sock), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketReaderFuture_cancel, __pyx_k_SyncSocketReaderFuture_cancel, sizeof(__pyx_k_SyncSocketReaderFuture_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketWriterFuture, __pyx_k_SyncSocketWriterFuture, sizeof(__pyx_k_SyncSocketWriterFuture), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketWriterFuture___init, __pyx_k_SyncSocketWriterFuture___init, sizeof(__pyx_k_SyncSocketWriterFuture___init), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketWriterFuture___remove, __pyx_k_SyncSocketWriterFuture___remove, sizeof(__pyx_k_SyncSocketWriterFuture___remove), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketWriterFuture__loop, __pyx_k_SyncSocketWriterFuture__loop, sizeof(__pyx_k_SyncSocketWriterFuture__loop), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketWriterFuture__remove, __pyx_k_SyncSocketWriterFuture__remove, sizeof(__pyx_k_SyncSocketWriterFuture__remove), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketWriterFuture__sock, __pyx_k_SyncSocketWriterFuture__sock, sizeof(__pyx_k_SyncSocketWriterFuture__sock), 0, 0, 1, 1},
    {&__pyx_n_s_SyncSocketWriterFuture_cancel, __pyx_k_SyncSocketWriterFuture_cancel, sizeof(__pyx_k_SyncSocketWriterFuture_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_SystemExit, __pyx_k_SystemExit, sizeof(__pyx_k_SystemExit), 0, 0, 1, 1},
    {&__pyx_n_s_TCPConnectRequest, __pyx_k_TCPConnectRequest, sizeof(__pyx_k_TCPConnectRequest), 0, 0, 1, 1},
    {&__pyx_n_s_TCPConnectRequest___reduce_cyth, __pyx_k_TCPConnectRequest___reduce_cyth, sizeof(__pyx_k_TCPConnectRequest___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_TCPConnectRequest___setstate_cy, __pyx_k_TCPConnectRequest___setstate_cy, sizeof(__pyx_k_TCPConnectRequest___setstate_cy), 0, 0, 1, 1},
    {&__pyx_n_s_TCPServer, __pyx_k_TCPServer, sizeof(__pyx_k_TCPServer), 0, 0, 1, 1},
    {&__pyx_n_s_TCPServer___reduce_cython, __pyx_k_TCPServer___reduce_cython, sizeof(__pyx_k_TCPServer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TCPServer___setstate_cython, __pyx_k_TCPServer___setstate_cython, sizeof(__pyx_k_TCPServer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TCPTransport, __pyx_k_TCPTransport, sizeof(__pyx_k_TCPTransport), 0, 0, 1, 1},
    {&__pyx_n_s_TCPTransport___reduce_cython, __pyx_k_TCPTransport___reduce_cython, sizeof(__pyx_k_TCPTransport___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TCPTransport___setstate_cython, __pyx_k_TCPTransport___setstate_cython, sizeof(__pyx_k_TCPTransport___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TCPTransport_get_extra_info, __pyx_k_TCPTransport_get_extra_info, sizeof(__pyx_k_TCPTransport_get_extra_info), 0, 0, 1, 1},
    {&__pyx_n_s_Task, __pyx_k_Task, sizeof(__pyx_k_Task), 0, 0, 1, 1},
    {&__pyx_kp_u_The_executor_did_not_finishing_j, __pyx_k_The_executor_did_not_finishing_j, sizeof(__pyx_k_The_executor_did_not_finishing_j), 0, 1, 0, 0},
    {&__pyx_kp_u_The_reuse_address_parameter_has, __pyx_k_The_reuse_address_parameter_has, sizeof(__pyx_k_The_reuse_address_parameter_has), 0, 1, 0, 0},
    {&__pyx_n_s_Thread, __pyx_k_Thread, sizeof(__pyx_k_Thread), 0, 0, 1, 1},
    {&__pyx_n_s_ThreadPoolExecutor, __pyx_k_ThreadPoolExecutor, sizeof(__pyx_k_ThreadPoolExecutor), 0, 0, 1, 1},
    {&__pyx_n_s_TimeoutError, __pyx_k_TimeoutError, sizeof(__pyx_k_TimeoutError), 0, 0, 1, 1},
    {&__pyx_n_s_TimerHandle, __pyx_k_TimerHandle, sizeof(__pyx_k_TimerHandle), 0, 0, 1, 1},
    {&__pyx_n_s_TimerHandle___reduce_cython, __pyx_k_TimerHandle___reduce_cython, sizeof(__pyx_k_TimerHandle___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimerHandle___setstate_cython, __pyx_k_TimerHandle___setstate_cython, sizeof(__pyx_k_TimerHandle___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_TimerHandle_cancel, __pyx_k_TimerHandle_cancel, sizeof(__pyx_k_TimerHandle_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_TimerHandle_cancelled, __pyx_k_TimerHandle_cancelled, sizeof(__pyx_k_TimerHandle_cancelled), 0, 0, 1, 1},
    {&__pyx_n_s_TimerHandle_when, __pyx_k_TimerHandle_when, sizeof(__pyx_k_TimerHandle_when), 0, 0, 1, 1},
    {&__pyx_kp_u_Timer_handles_8, __pyx_k_Timer_handles_8, sizeof(__pyx_k_Timer_handles_8), 0, 1, 0, 0},
    {&__pyx_n_s_Transport, __pyx_k_Transport, sizeof(__pyx_k_Transport), 0, 0, 1, 1},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_UDPSendContext, __pyx_k_UDPSendContext, sizeof(__pyx_k_UDPSendContext), 0, 0, 1, 1},
    {&__pyx_n_s_UDPSendContext___reduce_cython, __pyx_k_UDPSendContext___reduce_cython, sizeof(__pyx_k_UDPSendContext___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UDPSendContext___setstate_cytho, __pyx_k_UDPSendContext___setstate_cytho, sizeof(__pyx_k_UDPSendContext___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_UDPTransport, __pyx_k_UDPTransport, sizeof(__pyx_k_UDPTransport), 0, 0, 1, 1},
    {&__pyx_n_s_UDPTransport___reduce_cython, __pyx_k_UDPTransport___reduce_cython, sizeof(__pyx_k_UDPTransport___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UDPTransport___setstate_cython, __pyx_k_UDPTransport___setstate_cython, sizeof(__pyx_k_UDPTransport___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_UDPTransport_family_is_undefined, __pyx_k_UDPTransport_family_is_undefined, sizeof(__pyx_k_UDPTransport_family_is_undefined), 0, 1, 0, 0},
    {&__pyx_kp_u_UDPTransport_family_is_undefined_2, __pyx_k_UDPTransport_family_is_undefined_2, sizeof(__pyx_k_UDPTransport_family_is_undefined_2), 0, 1, 0, 0},
    {&__pyx_n_s_UDPTransport_sendto, __pyx_k_UDPTransport_sendto, sizeof(__pyx_k_UDPTransport_sendto), 0, 0, 1, 1},
    {&__pyx_n_s_UVAsync, __pyx_k_UVAsync, sizeof(__pyx_k_UVAsync), 0, 0, 1, 1},
    {&__pyx_n_s_UVAsync___reduce_cython, __pyx_k_UVAsync___reduce_cython, sizeof(__pyx_k_UVAsync___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVAsync___setstate_cython, __pyx_k_UVAsync___setstate_cython, sizeof(__pyx_k_UVAsync___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport, __pyx_k_UVBaseTransport, sizeof(__pyx_k_UVBaseTransport), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport___reduce_cython, __pyx_k_UVBaseTransport___reduce_cython, sizeof(__pyx_k_UVBaseTransport___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport___setstate_cytho, __pyx_k_UVBaseTransport___setstate_cytho, sizeof(__pyx_k_UVBaseTransport___setstate_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport__force_close, __pyx_k_UVBaseTransport__force_close, sizeof(__pyx_k_UVBaseTransport__force_close), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_abort, __pyx_k_UVBaseTransport_abort, sizeof(__pyx_k_UVBaseTransport_abort), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_close, __pyx_k_UVBaseTransport_close, sizeof(__pyx_k_UVBaseTransport_close), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_get_extra_info, __pyx_k_UVBaseTransport_get_extra_info, sizeof(__pyx_k_UVBaseTransport_get_extra_info), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_get_protocol, __pyx_k_UVBaseTransport_get_protocol, sizeof(__pyx_k_UVBaseTransport_get_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_get_write_buffer, __pyx_k_UVBaseTransport_get_write_buffer, sizeof(__pyx_k_UVBaseTransport_get_write_buffer), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_get_write_buffer_2, __pyx_k_UVBaseTransport_get_write_buffer_2, sizeof(__pyx_k_UVBaseTransport_get_write_buffer_2), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_is_closing, __pyx_k_UVBaseTransport_is_closing, sizeof(__pyx_k_UVBaseTransport_is_closing), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_set_protocol, __pyx_k_UVBaseTransport_set_protocol, sizeof(__pyx_k_UVBaseTransport_set_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_UVBaseTransport_set_write_buffer, __pyx_k_UVBaseTransport_set_write_buffer, sizeof(__pyx_k_UVBaseTransport_set_write_buffer), 0, 0, 1, 1},
    {&__pyx_n_s_UVCheck, __pyx_k_UVCheck, sizeof(__pyx_k_UVCheck), 0, 0, 1, 1},
    {&__pyx_n_s_UVCheck___reduce_cython, __pyx_k_UVCheck___reduce_cython, sizeof(__pyx_k_UVCheck___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVCheck___setstate_cython, __pyx_k_UVCheck___setstate_cython, sizeof(__pyx_k_UVCheck___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVFSEvent, __pyx_k_UVFSEvent, sizeof(__pyx_k_UVFSEvent), 0, 0, 1, 1},
    {&__pyx_n_s_UVFSEvent___reduce_cython, __pyx_k_UVFSEvent___reduce_cython, sizeof(__pyx_k_UVFSEvent___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVFSEvent___setstate_cython, __pyx_k_UVFSEvent___setstate_cython, sizeof(__pyx_k_UVFSEvent___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVFSEvent_cancel, __pyx_k_UVFSEvent_cancel, sizeof(__pyx_k_UVFSEvent_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_UVFSEvent_cancelled, __pyx_k_UVFSEvent_cancelled, sizeof(__pyx_k_UVFSEvent_cancelled), 0, 0, 1, 1},
    {&__pyx_n_s_UVHandle, __pyx_k_UVHandle, sizeof(__pyx_k_UVHandle), 0, 0, 1, 1},
    {&__pyx_n_s_UVHandle___reduce_cython, __pyx_k_UVHandle___reduce_cython, sizeof(__pyx_k_UVHandle___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVHandle___setstate_cython, __pyx_k_UVHandle___setstate_cython, sizeof(__pyx_k_UVHandle___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_UVHandles_python_libuv_total, __pyx_k_UVHandles_python_libuv_total, sizeof(__pyx_k_UVHandles_python_libuv_total), 0, 1, 0, 0},
    {&__pyx_n_s_UVIdle, __pyx_k_UVIdle, sizeof(__pyx_k_UVIdle), 0, 0, 1, 1},
    {&__pyx_n_s_UVIdle___reduce_cython, __pyx_k_UVIdle___reduce_cython, sizeof(__pyx_k_UVIdle___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVIdle___setstate_cython, __pyx_k_UVIdle___setstate_cython, sizeof(__pyx_k_UVIdle___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVPoll, __pyx_k_UVPoll, sizeof(__pyx_k_UVPoll), 0, 0, 1, 1},
    {&__pyx_n_s_UVPoll___reduce_cython, __pyx_k_UVPoll___reduce_cython, sizeof(__pyx_k_UVPoll___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVPoll___setstate_cython, __pyx_k_UVPoll___setstate_cython, sizeof(__pyx_k_UVPoll___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcess, __pyx_k_UVProcess, sizeof(__pyx_k_UVProcess), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport, __pyx_k_UVProcessTransport, sizeof(__pyx_k_UVProcessTransport), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport___reduce_cyth, __pyx_k_UVProcessTransport___reduce_cyth, sizeof(__pyx_k_UVProcessTransport___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport___setstate_cy, __pyx_k_UVProcessTransport___setstate_cy, sizeof(__pyx_k_UVProcessTransport___setstate_cy), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport___stdio_inite, __pyx_k_UVProcessTransport___stdio_inite, sizeof(__pyx_k_UVProcessTransport___stdio_inite), 0, 0, 1, 1},
    {&__pyx_kp_u_UVProcessTransport__call_connect, __pyx_k_UVProcessTransport__call_connect, sizeof(__pyx_k_UVProcessTransport__call_connect), 0, 1, 0, 0},
    {&__pyx_n_s_UVProcessTransport__stdio_inite, __pyx_k_UVProcessTransport__stdio_inite, sizeof(__pyx_k_UVProcessTransport__stdio_inite), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport__wait, __pyx_k_UVProcessTransport__wait, sizeof(__pyx_k_UVProcessTransport__wait), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_close, __pyx_k_UVProcessTransport_close, sizeof(__pyx_k_UVProcessTransport_close), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_get_extra_inf, __pyx_k_UVProcessTransport_get_extra_inf, sizeof(__pyx_k_UVProcessTransport_get_extra_inf), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_get_pid, __pyx_k_UVProcessTransport_get_pid, sizeof(__pyx_k_UVProcessTransport_get_pid), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_get_pipe_tran, __pyx_k_UVProcessTransport_get_pipe_tran, sizeof(__pyx_k_UVProcessTransport_get_pipe_tran), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_get_protocol, __pyx_k_UVProcessTransport_get_protocol, sizeof(__pyx_k_UVProcessTransport_get_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_get_returncod, __pyx_k_UVProcessTransport_get_returncod, sizeof(__pyx_k_UVProcessTransport_get_returncod), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_is_closing, __pyx_k_UVProcessTransport_is_closing, sizeof(__pyx_k_UVProcessTransport_is_closing), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_kill, __pyx_k_UVProcessTransport_kill, sizeof(__pyx_k_UVProcessTransport_kill), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_send_signal, __pyx_k_UVProcessTransport_send_signal, sizeof(__pyx_k_UVProcessTransport_send_signal), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_set_protocol, __pyx_k_UVProcessTransport_set_protocol, sizeof(__pyx_k_UVProcessTransport_set_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcessTransport_terminate, __pyx_k_UVProcessTransport_terminate, sizeof(__pyx_k_UVProcessTransport_terminate), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcess___reduce_cython, __pyx_k_UVProcess___reduce_cython, sizeof(__pyx_k_UVProcess___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVProcess___setstate_cython, __pyx_k_UVProcess___setstate_cython, sizeof(__pyx_k_UVProcess___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_UVProcess__close_after_spawn_cal, __pyx_k_UVProcess__close_after_spawn_cal, sizeof(__pyx_k_UVProcess__close_after_spawn_cal), 0, 1, 0, 0},
    {&__pyx_n_s_UVRequest, __pyx_k_UVRequest, sizeof(__pyx_k_UVRequest), 0, 0, 1, 1},
    {&__pyx_n_s_UVRequest___reduce_cython, __pyx_k_UVRequest___reduce_cython, sizeof(__pyx_k_UVRequest___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVRequest___setstate_cython, __pyx_k_UVRequest___setstate_cython, sizeof(__pyx_k_UVRequest___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVSocketHandle, __pyx_k_UVSocketHandle, sizeof(__pyx_k_UVSocketHandle), 0, 0, 1, 1},
    {&__pyx_n_s_UVSocketHandle___reduce_cython, __pyx_k_UVSocketHandle___reduce_cython, sizeof(__pyx_k_UVSocketHandle___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVSocketHandle___setstate_cython, __pyx_k_UVSocketHandle___setstate_cython, sizeof(__pyx_k_UVSocketHandle___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVStream, __pyx_k_UVStream, sizeof(__pyx_k_UVStream), 0, 0, 1, 1},
    {&__pyx_n_s_UVStreamServer, __pyx_k_UVStreamServer, sizeof(__pyx_k_UVStreamServer), 0, 0, 1, 1},
    {&__pyx_n_s_UVStreamServer___on_ssl_connecte, __pyx_k_UVStreamServer___on_ssl_connecte, sizeof(__pyx_k_UVStreamServer___on_ssl_connecte), 0, 0, 1, 1},
    {&__pyx_n_s_UVStreamServer___reduce_cython, __pyx_k_UVStreamServer___reduce_cython, sizeof(__pyx_k_UVStreamServer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVStreamServer___setstate_cython, __pyx_k_UVStreamServer___setstate_cython, sizeof(__pyx_k_UVStreamServer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVStreamServer__on_ssl_connecte, __pyx_k_UVStreamServer__on_ssl_connecte, sizeof(__pyx_k_UVStreamServer__on_ssl_connecte), 0, 0, 1, 1},
    {&__pyx_n_s_UVStream___reduce_cython, __pyx_k_UVStream___reduce_cython, sizeof(__pyx_k_UVStream___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVStream___setstate_cython, __pyx_k_UVStream___setstate_cython, sizeof(__pyx_k_UVStream___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_UVStream__try_write_called_with, __pyx_k_UVStream__try_write_called_with, sizeof(__pyx_k_UVStream__try_write_called_with), 0, 1, 0, 0},
    {&__pyx_n_s_UVStream_can_write_eof, __pyx_k_UVStream_can_write_eof, sizeof(__pyx_k_UVStream_can_write_eof), 0, 0, 1, 1},
    {&__pyx_n_s_UVStream_is_reading, __pyx_k_UVStream_is_reading, sizeof(__pyx_k_UVStream_is_reading), 0, 0, 1, 1},
    {&__pyx_n_s_UVStream_pause_reading, __pyx_k_UVStream_pause_reading, sizeof(__pyx_k_UVStream_pause_reading), 0, 0, 1, 1},
    {&__pyx_n_s_UVStream_resume_reading, __pyx_k_UVStream_resume_reading, sizeof(__pyx_k_UVStream_resume_reading), 0, 0, 1, 1},
    {&__pyx_kp_u_UVStream_shutdown_callback_calle, __pyx_k_UVStream_shutdown_callback_calle, sizeof(__pyx_k_UVStream_shutdown_callback_calle), 0, 1, 0, 0},
    {&__pyx_n_s_UVStream_write, __pyx_k_UVStream_write, sizeof(__pyx_k_UVStream_write), 0, 0, 1, 1},
    {&__pyx_kp_u_UVStream_write_callback_called_w, __pyx_k_UVStream_write_callback_called_w, sizeof(__pyx_k_UVStream_write_callback_called_w), 0, 1, 0, 0},
    {&__pyx_n_s_UVStream_write_eof, __pyx_k_UVStream_write_eof, sizeof(__pyx_k_UVStream_write_eof), 0, 0, 1, 1},
    {&__pyx_n_s_UVStream_writelines, __pyx_k_UVStream_writelines, sizeof(__pyx_k_UVStream_writelines), 0, 0, 1, 1},
    {&__pyx_n_s_UVTimer, __pyx_k_UVTimer, sizeof(__pyx_k_UVTimer), 0, 0, 1, 1},
    {&__pyx_n_s_UVTimer___reduce_cython, __pyx_k_UVTimer___reduce_cython, sizeof(__pyx_k_UVTimer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UVTimer___setstate_cython, __pyx_k_UVTimer___setstate_cython, sizeof(__pyx_k_UVTimer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_kp_u_UVTransport__call_connection_los, __pyx_k_UVTransport__call_connection_los, sizeof(__pyx_k_UVTransport__call_connection_los), 0, 1, 0, 0},
    {&__pyx_kp_u_UVTransport__call_connection_mad, __pyx_k_UVTransport__call_connection_mad, sizeof(__pyx_k_UVTransport__call_connection_mad), 0, 1, 0, 0},
    {&__pyx_kp_u_Unable_to_check_or_remove_stale, __pyx_k_Unable_to_check_or_remove_stale, sizeof(__pyx_k_Unable_to_check_or_remove_stale), 0, 1, 0, 0},
    {&__pyx_kp_u_Unable_to_clean_up_listening_UNI, __pyx_k_Unable_to_clean_up_listening_UNI, sizeof(__pyx_k_Unable_to_clean_up_listening_UNI), 0, 1, 0, 0},
    {&__pyx_kp_u_Unhandled_error_in_exception_han, __pyx_k_Unhandled_error_in_exception_han, sizeof(__pyx_k_Unhandled_error_in_exception_han), 0, 1, 0, 0},
    {&__pyx_kp_u_Unhandled_exception_in_event_loo, __pyx_k_Unhandled_exception_in_event_loo, sizeof(__pyx_k_Unhandled_exception_in_event_loo), 0, 1, 0, 0},
    {&__pyx_n_s_UnicodeEncodeError, __pyx_k_UnicodeEncodeError, sizeof(__pyx_k_UnicodeEncodeError), 0, 0, 1, 1},
    {&__pyx_n_s_UnixServer, __pyx_k_UnixServer, sizeof(__pyx_k_UnixServer), 0, 0, 1, 1},
    {&__pyx_n_s_UnixServer___reduce_cython, __pyx_k_UnixServer___reduce_cython, sizeof(__pyx_k_UnixServer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UnixServer___setstate_cython, __pyx_k_UnixServer___setstate_cython, sizeof(__pyx_k_UnixServer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UnixTransport, __pyx_k_UnixTransport, sizeof(__pyx_k_UnixTransport), 0, 0, 1, 1},
    {&__pyx_n_s_UnixTransport___reduce_cython, __pyx_k_UnixTransport___reduce_cython, sizeof(__pyx_k_UnixTransport___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_UnixTransport___setstate_cython, __pyx_k_UnixTransport___setstate_cython, sizeof(__pyx_k_UnixTransport___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s_WeakSet, __pyx_k_WeakSet, sizeof(__pyx_k_WeakSet), 0, 0, 1, 1},
    {&__pyx_n_s_WeakValueDictionary, __pyx_k_WeakValueDictionary, sizeof(__pyx_k_WeakValueDictionary), 0, 0, 1, 1},
    {&__pyx_n_s_WriteSubprocessPipeProto, __pyx_k_WriteSubprocessPipeProto, sizeof(__pyx_k_WriteSubprocessPipeProto), 0, 0, 1, 1},
    {&__pyx_n_s_WriteSubprocessPipeProto___init, __pyx_k_WriteSubprocessPipeProto___init, sizeof(__pyx_k_WriteSubprocessPipeProto___init), 0, 0, 1, 1},
    {&__pyx_n_s_WriteSubprocessPipeProto___repr, __pyx_k_WriteSubprocessPipeProto___repr, sizeof(__pyx_k_WriteSubprocessPipeProto___repr), 0, 0, 1, 1},
    {&__pyx_n_s_WriteSubprocessPipeProto_connect, __pyx_k_WriteSubprocessPipeProto_connect, sizeof(__pyx_k_WriteSubprocessPipeProto_connect), 0, 0, 1, 1},
    {&__pyx_n_s_WriteSubprocessPipeProto_connect_2, __pyx_k_WriteSubprocessPipeProto_connect_2, sizeof(__pyx_k_WriteSubprocessPipeProto_connect_2), 0, 0, 1, 1},
    {&__pyx_n_s_WriteSubprocessPipeProto_pause_w, __pyx_k_WriteSubprocessPipeProto_pause_w, sizeof(__pyx_k_WriteSubprocessPipeProto_pause_w), 0, 0, 1, 1},
    {&__pyx_n_s_WriteSubprocessPipeProto_resume, __pyx_k_WriteSubprocessPipeProto_resume, sizeof(__pyx_k_WriteSubprocessPipeProto_resume), 0, 0, 1, 1},
    {&__pyx_n_s_WriteUnixTransport, __pyx_k_WriteUnixTransport, sizeof(__pyx_k_WriteUnixTransport), 0, 0, 1, 1},
    {&__pyx_n_s_WriteUnixTransport___reduce_cyth, __pyx_k_WriteUnixTransport___reduce_cyth, sizeof(__pyx_k_WriteUnixTransport___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_WriteUnixTransport___setstate_cy, __pyx_k_WriteUnixTransport___setstate_cy, sizeof(__pyx_k_WriteUnixTransport___setstate_cy), 0, 0, 1, 1},
    {&__pyx_n_s_WriteUnixTransport_pause_reading, __pyx_k_WriteUnixTransport_pause_reading, sizeof(__pyx_k_WriteUnixTransport_pause_reading), 0, 0, 1, 1},
    {&__pyx_n_s_WriteUnixTransport_resume_readin, __pyx_k_WriteUnixTransport_resume_readin, sizeof(__pyx_k_WriteUnixTransport_resume_readin), 0, 0, 1, 1},
    {&__pyx_kp_u_Write_callbacks_failed, __pyx_k_Write_callbacks_failed, sizeof(__pyx_k_Write_callbacks_failed), 0, 1, 0, 0},
    {&__pyx_kp_u_Write_contexts_8, __pyx_k_Write_contexts_8, sizeof(__pyx_k_Write_contexts_8), 0, 1, 0, 0},
    {&__pyx_kp_u_Write_errors, __pyx_k_Write_errors, sizeof(__pyx_k_Write_errors), 0, 1, 0, 0},
    {&__pyx_kp_u_Write_events, __pyx_k_Write_events, sizeof(__pyx_k_Write_events), 0, 1, 0, 0},
    {&__pyx_kp_u_Write_failed_callbacks, __pyx_k_Write_failed_callbacks, sizeof(__pyx_k_Write_failed_callbacks), 0, 1, 0, 0},
    {&__pyx_kp_u_Write_without_poll, __pyx_k_Write_without_poll, sizeof(__pyx_k_Write_without_poll), 0, 1, 0, 0},
    {&__pyx_kp_u_XXX_an_allocated_buffer_in_trans, __pyx_k_XXX_an_allocated_buffer_in_trans, sizeof(__pyx_k_XXX_an_allocated_buffer_in_trans), 0, 1, 0, 0},
    {&__pyx_kp_u_You_must_set_server_hostname_whe, __pyx_k_You_must_set_server_hostname_whe, sizeof(__pyx_k_You_must_set_server_hostname_whe), 0, 1, 0, 0},
    {&__pyx_n_s__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 0, 1, 1},
    {&__pyx_kp_b__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 0, 0, 0},
    {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
    {&__pyx_kp_u__115, __pyx_k__115, sizeof(__pyx_k__115), 0, 1, 0, 0},
    {&__pyx_kp_u__128, __pyx_k__128, sizeof(__pyx_k__128), 0, 1, 0, 0},
    {&__pyx_kp_u__129, __pyx_k__129, sizeof(__pyx_k__129), 0, 1, 0, 0},
    {&__pyx_kp_u__130, __pyx_k__130, sizeof(__pyx_k__130), 0, 1, 0, 0},
    {&__pyx_kp_u__140, __pyx_k__140, sizeof(__pyx_k__140), 0, 1, 0, 0},
    {&__pyx_kp_u__141, __pyx_k__141, sizeof(__pyx_k__141), 0, 1, 0, 0},
    {&__pyx_kp_b__161, __pyx_k__161, sizeof(__pyx_k__161), 0, 0, 0, 0},
    {&__pyx_kp_b__170, __pyx_k__170, sizeof(__pyx_k__170), 0, 0, 0, 0},
    {&__pyx_kp_u__196, __pyx_k__196, sizeof(__pyx_k__196), 0, 1, 0, 0},
    {&__pyx_n_s__198, __pyx_k__198, sizeof(__pyx_k__198), 0, 0, 1, 1},
    {&__pyx_n_s__263, __pyx_k__263, sizeof(__pyx_k__263), 0, 0, 1, 1},
    {&__pyx_kp_u__33, __pyx_k__33, sizeof(__pyx_k__33), 0, 1, 0, 0},
    {&__pyx_n_s__542, __pyx_k__542, sizeof(__pyx_k__542), 0, 0, 1, 1},
    {&__pyx_kp_u__61, __pyx_k__61, sizeof(__pyx_k__61), 0, 1, 0, 0},
    {&__pyx_kp_u__69, __pyx_k__69, sizeof(__pyx_k__69), 0, 1, 0, 0},
    {&__pyx_kp_u__75, __pyx_k__75, sizeof(__pyx_k__75), 0, 1, 0, 0},
    {&__pyx_kp_u__78, __pyx_k__78, sizeof(__pyx_k__78), 0, 1, 0, 0},
    {&__pyx_kp_u__79, __pyx_k__79, sizeof(__pyx_k__79), 0, 1, 0, 0},
    {&__pyx_n_s_abc, __pyx_k_abc, sizeof(__pyx_k_abc), 0, 0, 1, 1},
    {&__pyx_n_s_abort, __pyx_k_abort, sizeof(__pyx_k_abort), 0, 0, 1, 1},
    {&__pyx_kp_u_abort_init__closed_is_set, __pyx_k_abort_init__closed_is_set, sizeof(__pyx_k_abort_init__closed_is_set), 0, 1, 0, 0},
    {&__pyx_kp_u_abort_init__inited_is_set, __pyx_k_abort_init__inited_is_set, sizeof(__pyx_k_abort_init__inited_is_set), 0, 1, 0, 0},
    {&__pyx_n_s_accept, __pyx_k_accept, sizeof(__pyx_k_accept), 0, 0, 1, 1},
    {&__pyx_kp_u_accept_method, __pyx_k_accept_method, sizeof(__pyx_k_accept_method), 0, 1, 0, 0},
    {&__pyx_n_s_aclose, __pyx_k_aclose, sizeof(__pyx_k_aclose), 0, 0, 1, 1},
    {&__pyx_kp_u_active_TimerHandle_is_deallacati, __pyx_k_active_TimerHandle_is_deallacati, sizeof(__pyx_k_active_TimerHandle_is_deallacati), 0, 1, 0, 0},
    {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
    {&__pyx_n_s_add_done_callback, __pyx_k_add_done_callback, sizeof(__pyx_k_add_done_callback), 0, 0, 1, 1},
    {&__pyx_n_s_add_reader, __pyx_k_add_reader, sizeof(__pyx_k_add_reader), 0, 0, 1, 1},
    {&__pyx_n_s_add_signal_handler, __pyx_k_add_signal_handler, sizeof(__pyx_k_add_signal_handler), 0, 0, 1, 1},
    {&__pyx_kp_u_add_signal_handler_can_only_be_c, __pyx_k_add_signal_handler_can_only_be_c, sizeof(__pyx_k_add_signal_handler_can_only_be_c), 0, 1, 0, 0},
    {&__pyx_n_s_add_writer, __pyx_k_add_writer, sizeof(__pyx_k_add_writer), 0, 0, 1, 1},
    {&__pyx_n_s_addr, __pyx_k_addr, sizeof(__pyx_k_addr), 0, 0, 1, 1},
    {&__pyx_n_s_address, __pyx_k_address, sizeof(__pyx_k_address), 0, 0, 1, 1},
    {&__pyx_n_s_addrinfo, __pyx_k_addrinfo, sizeof(__pyx_k_addrinfo), 0, 0, 1, 1},
    {&__pyx_n_s_addrs, __pyx_k_addrs, sizeof(__pyx_k_addrs), 0, 0, 1, 1},
    {&__pyx_n_s_aenter, __pyx_k_aenter, sizeof(__pyx_k_aenter), 0, 0, 1, 1},
    {&__pyx_n_s_aexit, __pyx_k_aexit, sizeof(__pyx_k_aexit), 0, 0, 1, 1},
    {&__pyx_n_s_ag, __pyx_k_ag, sizeof(__pyx_k_ag), 0, 0, 1, 1},
    {&__pyx_n_s_agen, __pyx_k_agen, sizeof(__pyx_k_agen), 0, 0, 1, 1},
    {&__pyx_n_s_ai, __pyx_k_ai, sizeof(__pyx_k_ai), 0, 0, 1, 1},
    {&__pyx_n_s_ai_cnt, __pyx_k_ai_cnt, sizeof(__pyx_k_ai_cnt), 0, 0, 1, 1},
    {&__pyx_n_s_ai_local, __pyx_k_ai_local, sizeof(__pyx_k_ai_local), 0, 0, 1, 1},
    {&__pyx_n_s_ai_remote, __pyx_k_ai_remote, sizeof(__pyx_k_ai_remote), 0, 0, 1, 1},
    {&__pyx_kp_u_alive_closed, __pyx_k_alive_closed, sizeof(__pyx_k_alive_closed), 0, 1, 0, 0},
    {&__pyx_kp_u_all_args_must_be_str_or_bytes, __pyx_k_all_args_must_be_str_or_bytes, sizeof(__pyx_k_all_args_must_be_str_or_bytes), 0, 1, 0, 0},
    {&__pyx_kp_u_all_environment_values_must_be_b, __pyx_k_all_environment_values_must_be_b, sizeof(__pyx_k_all_environment_values_must_be_b), 0, 1, 0, 0},
    {&__pyx_kp_u_all_environment_vars_must_be_byt, __pyx_k_all_environment_vars_must_be_byt, sizeof(__pyx_k_all_environment_vars_must_be_byt), 0, 1, 0, 0},
    {&__pyx_n_s_allow_broadcast, __pyx_k_allow_broadcast, sizeof(__pyx_k_allow_broadcast), 0, 0, 1, 1},
    {&__pyx_kp_u_an_error_occurred_during_closing, __pyx_k_an_error_occurred_during_closing, sizeof(__pyx_k_an_error_occurred_during_closing), 0, 1, 0, 0},
    {&__pyx_n_s_app_protocol, __pyx_k_app_protocol, sizeof(__pyx_k_app_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_app_transport, __pyx_k_app_transport, sizeof(__pyx_k_app_transport), 0, 0, 1, 1},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_u_ascii, __pyx_k_ascii, sizeof(__pyx_k_ascii), 0, 1, 0, 1},
    {&__pyx_n_u_asyncgen, __pyx_k_asyncgen, sizeof(__pyx_k_asyncgen), 0, 1, 0, 1},
    {&__pyx_n_s_asyncgen_finalizer_hook, __pyx_k_asyncgen_finalizer_hook, sizeof(__pyx_k_asyncgen_finalizer_hook), 0, 0, 1, 1},
    {&__pyx_n_s_asyncgen_firstiter_hook, __pyx_k_asyncgen_firstiter_hook, sizeof(__pyx_k_asyncgen_firstiter_hook), 0, 0, 1, 1},
    {&__pyx_kp_u_asynchronous_generator_r_was_sch, __pyx_k_asynchronous_generator_r_was_sch, sizeof(__pyx_k_asynchronous_generator_r_was_sch), 0, 1, 0, 0},
    {&__pyx_n_s_asyncio, __pyx_k_asyncio, sizeof(__pyx_k_asyncio), 0, 0, 1, 1},
    {&__pyx_kp_s_asyncio_Handle, __pyx_k_asyncio_Handle, sizeof(__pyx_k_asyncio_Handle), 0, 0, 1, 0},
    {&__pyx_n_s_asyncio_base_events, __pyx_k_asyncio_base_events, sizeof(__pyx_k_asyncio_base_events), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_futures, __pyx_k_asyncio_futures, sizeof(__pyx_k_asyncio_futures), 0, 0, 1, 1},
    {&__pyx_kp_u_asyncio_is_trying_to_install_it, __pyx_k_asyncio_is_trying_to_install_it, sizeof(__pyx_k_asyncio_is_trying_to_install_it), 0, 1, 0, 0},
    {&__pyx_n_s_asyncio_log, __pyx_k_asyncio_log, sizeof(__pyx_k_asyncio_log), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_sslproto, __pyx_k_asyncio_sslproto, sizeof(__pyx_k_asyncio_sslproto), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_tasks, __pyx_k_asyncio_tasks, sizeof(__pyx_k_asyncio_tasks), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_transports, __pyx_k_asyncio_transports, sizeof(__pyx_k_asyncio_transports), 0, 0, 1, 1},
    {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
    {&__pyx_n_u_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 1, 0, 1},
    {&__pyx_n_s_backlog, __pyx_k_backlog, sizeof(__pyx_k_backlog), 0, 0, 1, 1},
    {&__pyx_n_s_base_events, __pyx_k_base_events, sizeof(__pyx_k_base_events), 0, 0, 1, 1},
    {&__pyx_kp_b_bin_sh, __pyx_k_bin_sh, sizeof(__pyx_k_bin_sh), 0, 0, 0, 0},
    {&__pyx_n_s_bind, __pyx_k_bind, sizeof(__pyx_k_bind), 0, 0, 1, 1},
    {&__pyx_kp_u_bind_method, __pyx_k_bind_method, sizeof(__pyx_k_bind_method), 0, 1, 0, 0},
    {&__pyx_n_s_buf, __pyx_k_buf, sizeof(__pyx_k_buf), 0, 0, 1, 1},
    {&__pyx_kp_u_buffer_size_is_not_0_after_a_su, __pyx_k_buffer_size_is_not_0_after_a_su, sizeof(__pyx_k_buffer_size_is_not_0_after_a_su), 0, 1, 0, 0},
    {&__pyx_n_s_buffer_updated, __pyx_k_buffer_updated, sizeof(__pyx_k_buffer_updated), 0, 0, 1, 1},
    {&__pyx_n_s_bufs, __pyx_k_bufs, sizeof(__pyx_k_bufs), 0, 0, 1, 1},
    {&__pyx_n_s_bufsize, __pyx_k_bufsize, sizeof(__pyx_k_bufsize), 0, 0, 1, 1},
    {&__pyx_kp_u_bufsize_must_be_0, __pyx_k_bufsize_must_be_0, sizeof(__pyx_k_bufsize_must_be_0), 0, 1, 0, 0},
    {&__pyx_n_s_builtins, __pyx_k_builtins, sizeof(__pyx_k_builtins), 0, 0, 1, 1},
    {&__pyx_kp_b_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 0, 0},
    {&__pyx_n_s_c_str_path, __pyx_k_c_str_path, sizeof(__pyx_k_c_str_path), 0, 0, 1, 1},
    {&__pyx_n_s_call_at, __pyx_k_call_at, sizeof(__pyx_k_call_at), 0, 0, 1, 1},
    {&__pyx_n_s_call_connection_made, __pyx_k_call_connection_made, sizeof(__pyx_k_call_connection_made), 0, 0, 1, 1},
    {&__pyx_n_s_call_exception_handler, __pyx_k_call_exception_handler, sizeof(__pyx_k_call_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_call_later, __pyx_k_call_later, sizeof(__pyx_k_call_later), 0, 0, 1, 1},
    {&__pyx_n_s_call_soon, __pyx_k_call_soon, sizeof(__pyx_k_call_soon), 0, 0, 1, 1},
    {&__pyx_n_s_call_soon_threadsafe, __pyx_k_call_soon_threadsafe, sizeof(__pyx_k_call_soon_threadsafe), 0, 0, 1, 1},
    {&__pyx_n_s_callback, __pyx_k_callback, sizeof(__pyx_k_callback), 0, 0, 1, 1},
    {&__pyx_kp_u_called_after_destroying_the_UVH, __pyx_k_called_after_destroying_the_UVH, sizeof(__pyx_k_called_after_destroying_the_UVH), 0, 1, 0, 0},
    {&__pyx_kp_u_called_with_handle_data_NULL, __pyx_k_called_with_handle_data_NULL, sizeof(__pyx_k_called_with_handle_data_NULL), 0, 1, 0, 0},
    {&__pyx_n_s_can_write_eof, __pyx_k_can_write_eof, sizeof(__pyx_k_can_write_eof), 0, 0, 1, 1},
    {&__pyx_n_s_cancel, __pyx_k_cancel, sizeof(__pyx_k_cancel), 0, 0, 1, 1},
    {&__pyx_kp_u_cancel_request_data_is_NULL, __pyx_k_cancel_request_data_is_NULL, sizeof(__pyx_k_cancel_request_data_is_NULL), 0, 1, 0, 0},
    {&__pyx_kp_u_cancel_request_data_is_not_UVRe, __pyx_k_cancel_request_data_is_not_UVRe, sizeof(__pyx_k_cancel_request_data_is_not_UVRe), 0, 1, 0, 0},
    {&__pyx_kp_u_cancel_request_is_NULL, __pyx_k_cancel_request_is_NULL, sizeof(__pyx_k_cancel_request_is_NULL), 0, 1, 0, 0},
    {&__pyx_n_s_cancelled, __pyx_k_cancelled, sizeof(__pyx_k_cancelled), 0, 0, 1, 1},
    {&__pyx_n_u_cancelled, __pyx_k_cancelled, sizeof(__pyx_k_cancelled), 0, 1, 0, 1},
    {&__pyx_kp_u_cannot_add_a_signal_handler_for, __pyx_k_cannot_add_a_signal_handler_for, sizeof(__pyx_k_cannot_add_a_signal_handler_for), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_apply_subprocess_STDOUT, __pyx_k_cannot_apply_subprocess_STDOUT, sizeof(__pyx_k_cannot_apply_subprocess_STDOUT), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_cleanup_signal_handlers_c, __pyx_k_cannot_cleanup_signal_handlers_c, sizeof(__pyx_k_cannot_cleanup_signal_handlers_c), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_convert_sockaddr_into_Pyt, __pyx_k_cannot_convert_sockaddr_into_Pyt, sizeof(__pyx_k_cannot_convert_sockaddr_into_Pyt), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_open_a_UDP_handle_invalid, __pyx_k_cannot_open_a_UDP_handle_invalid, sizeof(__pyx_k_cannot_open_a_UDP_handle_invalid), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_pause_signals_handling_no, __pyx_k_cannot_pause_signals_handling_no, sizeof(__pyx_k_cannot_pause_signals_handling_no), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_run_Handle_callback_is_no, __pyx_k_cannot_run_Handle_callback_is_no, sizeof(__pyx_k_cannot_run_Handle_callback_is_no), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_run_TimerHandle_callback, __pyx_k_cannot_run_TimerHandle_callback, sizeof(__pyx_k_cannot_run_TimerHandle_callback), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_shutdown_signals_handling, __pyx_k_cannot_shutdown_signals_handling, sizeof(__pyx_k_cannot_shutdown_signals_handling), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_spawn_a_process_args_are, __pyx_k_cannot_spawn_a_process_args_are, sizeof(__pyx_k_cannot_spawn_a_process_args_are), 0, 1, 0, 0},
    {&__pyx_kp_u_cannot_switch_state_from_to, __pyx_k_cannot_switch_state_from_to, sizeof(__pyx_k_cannot_switch_state_from_to), 0, 1, 0, 0},
    {&__pyx_n_s_cast, __pyx_k_cast, sizeof(__pyx_k_cast), 0, 0, 1, 1},
    {&__pyx_n_s_cause, __pyx_k_cause, sizeof(__pyx_k_cause), 0, 0, 1, 1},
    {&__pyx_n_s_cfunc_to_py, __pyx_k_cfunc_to_py, sizeof(__pyx_k_cfunc_to_py), 0, 0, 1, 1},
    {&__pyx_n_s_chain, __pyx_k_chain, sizeof(__pyx_k_chain), 0, 0, 1, 1},
    {&__pyx_n_s_check_default_executor, __pyx_k_check_default_executor, sizeof(__pyx_k_check_default_executor), 0, 0, 1, 1},
    {&__pyx_n_s_check_hostname, __pyx_k_check_hostname, sizeof(__pyx_k_check_hostname), 0, 0, 1, 1},
    {&__pyx_n_s_cipher, __pyx_k_cipher, sizeof(__pyx_k_cipher), 0, 0, 1, 1},
    {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_cleanup_socket, __pyx_k_cleanup_socket, sizeof(__pyx_k_cleanup_socket), 0, 0, 1, 1},
    {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_kp_u_close__handle_data_is_NULL, __pyx_k_close__handle_data_is_NULL, sizeof(__pyx_k_close__handle_data_is_NULL), 0, 1, 0, 0},
    {&__pyx_kp_u_close__handle_data_is_not_UVHa, __pyx_k_close__handle_data_is_not_UVHa, sizeof(__pyx_k_close__handle_data_is_not_UVHa), 0, 1, 0, 0},
    {&__pyx_n_s_close_fds, __pyx_k_close_fds, sizeof(__pyx_k_close_fds), 0, 0, 1, 1},
    {&__pyx_kp_u_close_method, __pyx_k_close_method, sizeof(__pyx_k_close_method), 0, 1, 0, 0},
    {&__pyx_kp_u_close_uv_is_closing_is_true, __pyx_k_close_uv_is_closing_is_true, sizeof(__pyx_k_close_uv_is_closing_is_true), 0, 1, 0, 0},
    {&__pyx_kp_u_closed_Transport_handle_and_unse, __pyx_k_closed_Transport_handle_and_unse, sizeof(__pyx_k_closed_Transport_handle_and_unse), 0, 1, 0, 0},
    {&__pyx_kp_u_closed_reading_x, __pyx_k_closed_reading_x, sizeof(__pyx_k_closed_reading_x), 0, 1, 0, 0},
    {&__pyx_kp_u_closed_x, __pyx_k_closed_x, sizeof(__pyx_k_closed_x), 0, 1, 0, 0},
    {&__pyx_n_s_closing_agens, __pyx_k_closing_agens, sizeof(__pyx_k_closing_agens), 0, 0, 1, 1},
    {&__pyx_n_s_cmd, __pyx_k_cmd, sizeof(__pyx_k_cmd), 0, 0, 1, 1},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_n_s_collections_abc, __pyx_k_collections_abc, sizeof(__pyx_k_collections_abc), 0, 0, 1, 1},
    {&__pyx_n_s_completed, __pyx_k_completed, sizeof(__pyx_k_completed), 0, 0, 1, 1},
    {&__pyx_n_s_compression, __pyx_k_compression, sizeof(__pyx_k_compression), 0, 0, 1, 1},
    {&__pyx_n_s_concurrent, __pyx_k_concurrent, sizeof(__pyx_k_concurrent), 0, 0, 1, 1},
    {&__pyx_kp_u_concurrent_allocations, __pyx_k_concurrent_allocations, sizeof(__pyx_k_concurrent_allocations), 0, 1, 0, 0},
    {&__pyx_n_s_concurrent_futures, __pyx_k_concurrent_futures, sizeof(__pyx_k_concurrent_futures), 0, 0, 1, 1},
    {&__pyx_n_s_conmade_cb, __pyx_k_conmade_cb, sizeof(__pyx_k_conmade_cb), 0, 0, 1, 1},
    {&__pyx_n_s_connect, __pyx_k_connect, sizeof(__pyx_k_connect), 0, 0, 1, 1},
    {&__pyx_n_s_connect_accepted_socket, __pyx_k_connect_accepted_socket, sizeof(__pyx_k_connect_accepted_socket), 0, 0, 1, 1},
    {&__pyx_n_s_connect_ex, __pyx_k_connect_ex, sizeof(__pyx_k_connect_ex), 0, 0, 1, 1},
    {&__pyx_kp_u_connect_ex_method, __pyx_k_connect_ex_method, sizeof(__pyx_k_connect_ex_method), 0, 1, 0, 0},
    {&__pyx_kp_u_connect_failed, __pyx_k_connect_failed, sizeof(__pyx_k_connect_failed), 0, 1, 0, 0},
    {&__pyx_kp_u_connect_method, __pyx_k_connect_method, sizeof(__pyx_k_connect_method), 0, 1, 0, 0},
    {&__pyx_n_s_connect_read_pipe, __pyx_k_connect_read_pipe, sizeof(__pyx_k_connect_read_pipe), 0, 0, 1, 1},
    {&__pyx_n_s_connect_write_pipe, __pyx_k_connect_write_pipe, sizeof(__pyx_k_connect_write_pipe), 0, 0, 1, 1},
    {&__pyx_n_s_connection_lost, __pyx_k_connection_lost, sizeof(__pyx_k_connection_lost), 0, 0, 1, 1},
    {&__pyx_n_s_connection_made, __pyx_k_connection_made, sizeof(__pyx_k_connection_made), 0, 0, 1, 1},
    {&__pyx_n_s_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 0, 1, 1},
    {&__pyx_n_u_context, __pyx_k_context, sizeof(__pyx_k_context), 0, 1, 0, 1},
    {&__pyx_kp_u_context_manager_protocol, __pyx_k_context_manager_protocol, sizeof(__pyx_k_context_manager_protocol), 0, 1, 0, 0},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_s_coro, __pyx_k_coro, sizeof(__pyx_k_coro), 0, 0, 1, 1},
    {&__pyx_n_s_coroutines, __pyx_k_coroutines, sizeof(__pyx_k_coroutines), 0, 0, 1, 1},
    {&__pyx_kp_u_coroutines_cannot_be_used_with_a, __pyx_k_coroutines_cannot_be_used_with_a, sizeof(__pyx_k_coroutines_cannot_be_used_with_a), 0, 1, 0, 0},
    {&__pyx_kp_u_coroutines_cannot_be_used_with_r, __pyx_k_coroutines_cannot_be_used_with_r, sizeof(__pyx_k_coroutines_cannot_be_used_with_r), 0, 1, 0, 0},
    {&__pyx_kp_u_could_not_bind_to_local_addr, __pyx_k_could_not_bind_to_local_addr, sizeof(__pyx_k_could_not_bind_to_local_addr), 0, 1, 0, 0},
    {&__pyx_kp_u_could_not_bind_to_remote_addr, __pyx_k_could_not_bind_to_remote_addr, sizeof(__pyx_k_could_not_bind_to_remote_addr), 0, 1, 0, 0},
    {&__pyx_kp_u_could_not_close_attached_file_ob, __pyx_k_could_not_close_attached_file_ob, sizeof(__pyx_k_could_not_close_attached_file_ob), 0, 1, 0, 0},
    {&__pyx_n_s_create_connection, __pyx_k_create_connection, sizeof(__pyx_k_create_connection), 0, 0, 1, 1},
    {&__pyx_n_s_create_connection_locals_genexpr, __pyx_k_create_connection_locals_genexpr, sizeof(__pyx_k_create_connection_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_create_datagram_endpoint, __pyx_k_create_datagram_endpoint, sizeof(__pyx_k_create_datagram_endpoint), 0, 0, 1, 1},
    {&__pyx_n_s_create_datagram_endpoint_locals, __pyx_k_create_datagram_endpoint_locals, sizeof(__pyx_k_create_datagram_endpoint_locals), 0, 0, 1, 1},
    {&__pyx_n_s_create_default_context, __pyx_k_create_default_context, sizeof(__pyx_k_create_default_context), 0, 0, 1, 1},
    {&__pyx_n_s_create_future, __pyx_k_create_future, sizeof(__pyx_k_create_future), 0, 0, 1, 1},
    {&__pyx_n_s_create_server, __pyx_k_create_server, sizeof(__pyx_k_create_server), 0, 0, 1, 1},
    {&__pyx_kp_u_create_server_failed_to_create_s, __pyx_k_create_server_failed_to_create_s, sizeof(__pyx_k_create_server_failed_to_create_s), 0, 1, 0, 0},
    {&__pyx_n_s_create_task, __pyx_k_create_task, sizeof(__pyx_k_create_task), 0, 0, 1, 1},
    {&__pyx_n_s_create_unix_connection, __pyx_k_create_unix_connection, sizeof(__pyx_k_create_unix_connection), 0, 0, 1, 1},
    {&__pyx_n_s_create_unix_server, __pyx_k_create_unix_server, sizeof(__pyx_k_create_unix_server), 0, 0, 1, 1},
    {&__pyx_kp_u_created_at, __pyx_k_created_at, sizeof(__pyx_k_created_at), 0, 1, 0, 0},
    {&__pyx_n_s_creationflags, __pyx_k_creationflags, sizeof(__pyx_k_creationflags), 0, 0, 1, 1},
    {&__pyx_kp_u_creationflags_is_not_supported, __pyx_k_creationflags_is_not_supported, sizeof(__pyx_k_creationflags_is_not_supported), 0, 1, 0, 0},
    {&__pyx_n_s_ctx, __pyx_k_ctx, sizeof(__pyx_k_ctx), 0, 0, 1, 1},
    {&__pyx_n_s_cwd, __pyx_k_cwd, sizeof(__pyx_k_cwd), 0, 0, 1, 1},
    {&__pyx_kp_u_cwd_must_be_a_str_or_bytes_objec, __pyx_k_cwd_must_be_a_str_or_bytes_objec, sizeof(__pyx_k_cwd_must_be_a_str_or_bytes_objec), 0, 1, 0, 0},
    {&__pyx_n_u_cygwin, __pyx_k_cygwin, sizeof(__pyx_k_cygwin), 0, 1, 0, 1},
    {&__pyx_n_u_darwin, __pyx_k_darwin, sizeof(__pyx_k_darwin), 0, 1, 0, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_kp_u_data_expecting_a_bytes_like_inst, __pyx_k_data_expecting_a_bytes_like_inst, sizeof(__pyx_k_data_expecting_a_bytes_like_inst), 0, 1, 0, 0},
    {&__pyx_n_s_data_received, __pyx_k_data_received, sizeof(__pyx_k_data_received), 0, 0, 1, 1},
    {&__pyx_n_s_datagram_received, __pyx_k_datagram_received, sizeof(__pyx_k_datagram_received), 0, 0, 1, 1},
    {&__pyx_kp_u_dealloc____handle_is_NULL__cl, __pyx_k_dealloc____handle_is_NULL__cl, sizeof(__pyx_k_dealloc____handle_is_NULL__cl), 0, 1, 0, 0},
    {&__pyx_kp_u_deallocating_a_running_event_loo, __pyx_k_deallocating_a_running_event_loo, sizeof(__pyx_k_deallocating_a_running_event_loo), 0, 1, 0, 0},
    {&__pyx_kp_u_deallocating_an_open_event_loop, __pyx_k_deallocating_an_open_event_loop, sizeof(__pyx_k_deallocating_an_open_event_loop), 0, 1, 0, 0},
    {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
    {&__pyx_n_s_debug_flags, __pyx_k_debug_flags, sizeof(__pyx_k_debug_flags), 0, 0, 1, 1},
    {&__pyx_n_u_debug_wrapper, __pyx_k_debug_wrapper, sizeof(__pyx_k_debug_wrapper), 0, 1, 0, 1},
    {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
    {&__pyx_n_s_decref_socketios, __pyx_k_decref_socketios, sizeof(__pyx_k_decref_socketios), 0, 0, 1, 1},
    {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
    {&__pyx_n_s_default_exception_handler, __pyx_k_default_exception_handler, sizeof(__pyx_k_default_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_default_int_handler, __pyx_k_default_int_handler, sizeof(__pyx_k_default_int_handler), 0, 0, 1, 1},
    {&__pyx_n_s_delay, __pyx_k_delay, sizeof(__pyx_k_delay), 0, 0, 1, 1},
    {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
    {&__pyx_n_s_detach, __pyx_k_detach, sizeof(__pyx_k_detach), 0, 0, 1, 1},
    {&__pyx_kp_u_detach_method, __pyx_k_detach_method, sizeof(__pyx_k_detach_method), 0, 1, 0, 0},
    {&__pyx_n_s_dev_mode, __pyx_k_dev_mode, sizeof(__pyx_k_dev_mode), 0, 0, 1, 1},
    {&__pyx_n_s_devnull, __pyx_k_devnull, sizeof(__pyx_k_devnull), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_n_s_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_discard, __pyx_k_discard, sizeof(__pyx_k_discard), 0, 0, 1, 1},
    {&__pyx_n_s_disconnected, __pyx_k_disconnected, sizeof(__pyx_k_disconnected), 0, 0, 1, 1},
    {&__pyx_n_s_do_handshake, __pyx_k_do_handshake, sizeof(__pyx_k_do_handshake), 0, 0, 1, 1},
    {&__pyx_n_s_do_shutdown, __pyx_k_do_shutdown, sizeof(__pyx_k_do_shutdown), 0, 0, 1, 1},
    {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
    {&__pyx_kp_u_does_not_have_a_reader_in_the, __pyx_k_does_not_have_a_reader_in_the, sizeof(__pyx_k_does_not_have_a_reader_in_the), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_have_a_reader_in_the_2, __pyx_k_does_not_have_a_reader_in_the_2, sizeof(__pyx_k_does_not_have_a_reader_in_the_2), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_have_a_writer_in_the, __pyx_k_does_not_have_a_writer_in_the, sizeof(__pyx_k_does_not_have_a_writer_in_the), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_have_a_writer_in_the_2, __pyx_k_does_not_have_a_writer_in_the_2, sizeof(__pyx_k_does_not_have_a_writer_in_the_2), 0, 1, 0, 0},
    {&__pyx_n_s_done, __pyx_k_done, sizeof(__pyx_k_done), 0, 0, 1, 1},
    {&__pyx_n_s_done_cb, __pyx_k_done_cb, sizeof(__pyx_k_done_cb), 0, 0, 1, 1},
    {&__pyx_n_s_dup, __pyx_k_dup, sizeof(__pyx_k_dup), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_enabled, __pyx_k_enabled, sizeof(__pyx_k_enabled), 0, 0, 1, 1},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_ensure_future, __pyx_k_ensure_future, sizeof(__pyx_k_ensure_future), 0, 0, 1, 1},
    {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
    {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
    {&__pyx_n_s_env, __pyx_k_env, sizeof(__pyx_k_env), 0, 0, 1, 1},
    {&__pyx_n_s_environ, __pyx_k_environ, sizeof(__pyx_k_environ), 0, 0, 1, 1},
    {&__pyx_n_s_eof_received, __pyx_k_eof_received, sizeof(__pyx_k_eof_received), 0, 0, 1, 1},
    {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
    {&__pyx_n_s_errno, __pyx_k_errno, sizeof(__pyx_k_errno), 0, 0, 1, 1},
    {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
    {&__pyx_n_s_error_received, __pyx_k_error_received, sizeof(__pyx_k_error_received), 0, 0, 1, 1},
    {&__pyx_kp_u_error_status_in_uv_stream_t_list, __pyx_k_error_status_in_uv_stream_t_list, sizeof(__pyx_k_error_status_in_uv_stream_t_list), 0, 1, 0, 0},
    {&__pyx_kp_u_error_status_in_uv_stream_t_read, __pyx_k_error_status_in_uv_stream_t_read, sizeof(__pyx_k_error_status_in_uv_stream_t_read), 0, 1, 0, 0},
    {&__pyx_kp_u_error_status_in_uv_stream_t_shut, __pyx_k_error_status_in_uv_stream_t_shut, sizeof(__pyx_k_error_status_in_uv_stream_t_shut), 0, 1, 0, 0},
    {&__pyx_kp_u_error_status_in_uv_stream_t_writ, __pyx_k_error_status_in_uv_stream_t_writ, sizeof(__pyx_k_error_status_in_uv_stream_t_writ), 0, 1, 0, 0},
    {&__pyx_kp_u_error_while_attempting_to_bind_o, __pyx_k_error_while_attempting_to_bind_o, sizeof(__pyx_k_error_while_attempting_to_bind_o), 0, 1, 0, 0},
    {&__pyx_n_s_ex, __pyx_k_ex, sizeof(__pyx_k_ex), 0, 0, 1, 1},
    {&__pyx_n_s_exc, __pyx_k_exc, sizeof(__pyx_k_exc), 0, 0, 1, 1},
    {&__pyx_n_s_exc_info, __pyx_k_exc_info, sizeof(__pyx_k_exc_info), 0, 0, 1, 1},
    {&__pyx_n_s_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 0, 1, 1},
    {&__pyx_n_u_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 1, 0, 1},
    {&__pyx_kp_u_exception_in_Task___repr, __pyx_k_exception_in_Task___repr, sizeof(__pyx_k_exception_in_Task___repr), 0, 1, 0, 0},
    {&__pyx_n_s_exceptions, __pyx_k_exceptions, sizeof(__pyx_k_exceptions), 0, 0, 1, 1},
    {&__pyx_n_s_excs, __pyx_k_excs, sizeof(__pyx_k_excs), 0, 0, 1, 1},
    {&__pyx_n_s_executable, __pyx_k_executable, sizeof(__pyx_k_executable), 0, 0, 1, 1},
    {&__pyx_n_s_executor, __pyx_k_executor, sizeof(__pyx_k_executor), 0, 0, 1, 1},
    {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
    {&__pyx_kp_u_expected_AF_INET_AF_INET6_or_AF, __pyx_k_expected_AF_INET_AF_INET6_or_AF, sizeof(__pyx_k_expected_AF_INET_AF_INET6_or_AF), 0, 1, 0, 0},
    {&__pyx_kp_u_expected_asyncio_Future, __pyx_k_expected_asyncio_Future, sizeof(__pyx_k_expected_asyncio_Future), 0, 1, 0, 0},
    {&__pyx_n_s_extract, __pyx_k_extract, sizeof(__pyx_k_extract), 0, 0, 1, 1},
    {&__pyx_n_s_f1, __pyx_k_f1, sizeof(__pyx_k_f1), 0, 0, 1, 1},
    {&__pyx_n_s_f2, __pyx_k_f2, sizeof(__pyx_k_f2), 0, 0, 1, 1},
    {&__pyx_n_s_factory, __pyx_k_factory, sizeof(__pyx_k_factory), 0, 0, 1, 1},
    {&__pyx_n_s_family, __pyx_k_family, sizeof(__pyx_k_family), 0, 0, 1, 1},
    {&__pyx_kp_u_fatal_Could_not_advance__StreamW, __pyx_k_fatal_Could_not_advance__StreamW, sizeof(__pyx_k_fatal_Could_not_advance__StreamW), 0, 1, 0, 0},
    {&__pyx_kp_u_fatal_sent_0_in_advance_uv_buf, __pyx_k_fatal_sent_0_in_advance_uv_buf, sizeof(__pyx_k_fatal_sent_0_in_advance_uv_buf), 0, 1, 0, 0},
    {&__pyx_n_s_fd, __pyx_k_fd, sizeof(__pyx_k_fd), 0, 0, 1, 1},
    {&__pyx_kp_u_fd_2, __pyx_k_fd_2, sizeof(__pyx_k_fd_2), 0, 1, 0, 0},
    {&__pyx_n_s_fileno, __pyx_k_fileno, sizeof(__pyx_k_fileno), 0, 0, 1, 1},
    {&__pyx_n_s_fileobj, __pyx_k_fileobj, sizeof(__pyx_k_fileobj), 0, 0, 1, 1},
    {&__pyx_n_s_finalizer, __pyx_k_finalizer, sizeof(__pyx_k_finalizer), 0, 0, 1, 1},
    {&__pyx_n_s_firstiter, __pyx_k_firstiter, sizeof(__pyx_k_firstiter), 0, 0, 1, 1},
    {&__pyx_n_s_flag, __pyx_k_flag, sizeof(__pyx_k_flag), 0, 0, 1, 1},
    {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
    {&__pyx_n_s_flowinfo, __pyx_k_flowinfo, sizeof(__pyx_k_flowinfo), 0, 0, 1, 1},
    {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
    {&__pyx_n_s_force_close, __pyx_k_force_close, sizeof(__pyx_k_force_close), 0, 0, 1, 1},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_s_format_list, __pyx_k_format_list, sizeof(__pyx_k_format_list), 0, 0, 1, 1},
    {&__pyx_n_s_frame, __pyx_k_frame, sizeof(__pyx_k_frame), 0, 0, 1, 1},
    {&__pyx_n_s_fs, __pyx_k_fs, sizeof(__pyx_k_fs), 0, 0, 1, 1},
    {&__pyx_n_s_fs_handle, __pyx_k_fs_handle, sizeof(__pyx_k_fs_handle), 0, 0, 1, 1},
    {&__pyx_n_s_fspath, __pyx_k_fspath, sizeof(__pyx_k_fspath), 0, 0, 1, 1},
    {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
    {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
    {&__pyx_n_s_fut, __pyx_k_fut, sizeof(__pyx_k_fut), 0, 0, 1, 1},
    {&__pyx_n_s_future, __pyx_k_future, sizeof(__pyx_k_future), 0, 0, 1, 1},
    {&__pyx_n_s_futures, __pyx_k_futures, sizeof(__pyx_k_futures), 0, 0, 1, 1},
    {&__pyx_n_s_gaierror, __pyx_k_gaierror, sizeof(__pyx_k_gaierror), 0, 0, 1, 1},
    {&__pyx_n_s_gather, __pyx_k_gather, sizeof(__pyx_k_gather), 0, 0, 1, 1},
    {&__pyx_n_s_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_get_app_transport, __pyx_k_get_app_transport, sizeof(__pyx_k_get_app_transport), 0, 0, 1, 1},
    {&__pyx_n_s_get_asyncgen_hooks, __pyx_k_get_asyncgen_hooks, sizeof(__pyx_k_get_asyncgen_hooks), 0, 0, 1, 1},
    {&__pyx_n_s_get_backend_id, __pyx_k_get_backend_id, sizeof(__pyx_k_get_backend_id), 0, 0, 1, 1},
    {&__pyx_n_s_get_buffer, __pyx_k_get_buffer, sizeof(__pyx_k_get_buffer), 0, 0, 1, 1},
    {&__pyx_n_u_get_buffer, __pyx_k_get_buffer, sizeof(__pyx_k_get_buffer), 0, 1, 0, 1},
    {&__pyx_n_s_get_coroutine_origin_tracking_de, __pyx_k_get_coroutine_origin_tracking_de, sizeof(__pyx_k_get_coroutine_origin_tracking_de), 0, 0, 1, 1},
    {&__pyx_n_u_get_coroutine_wrapper, __pyx_k_get_coroutine_wrapper, sizeof(__pyx_k_get_coroutine_wrapper), 0, 1, 0, 1},
    {&__pyx_n_s_get_debug, __pyx_k_get_debug, sizeof(__pyx_k_get_debug), 0, 0, 1, 1},
    {&__pyx_n_s_get_event_loop, __pyx_k_get_event_loop, sizeof(__pyx_k_get_event_loop), 0, 0, 1, 1},
    {&__pyx_n_s_get_exception_handler, __pyx_k_get_exception_handler, sizeof(__pyx_k_get_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_get_extra_info, __pyx_k_get_extra_info, sizeof(__pyx_k_get_extra_info), 0, 0, 1, 1},
    {&__pyx_n_s_get_inheritable, __pyx_k_get_inheritable, sizeof(__pyx_k_get_inheritable), 0, 0, 1, 1},
    {&__pyx_n_s_get_loop, __pyx_k_get_loop, sizeof(__pyx_k_get_loop), 0, 0, 1, 1},
    {&__pyx_n_s_get_pid, __pyx_k_get_pid, sizeof(__pyx_k_get_pid), 0, 0, 1, 1},
    {&__pyx_n_s_get_pipe_transport, __pyx_k_get_pipe_transport, sizeof(__pyx_k_get_pipe_transport), 0, 0, 1, 1},
    {&__pyx_n_s_get_protocol, __pyx_k_get_protocol, sizeof(__pyx_k_get_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_get_read_buffer_limits, __pyx_k_get_read_buffer_limits, sizeof(__pyx_k_get_read_buffer_limits), 0, 0, 1, 1},
    {&__pyx_n_s_get_read_buffer_size, __pyx_k_get_read_buffer_size, sizeof(__pyx_k_get_read_buffer_size), 0, 0, 1, 1},
    {&__pyx_n_s_get_returncode, __pyx_k_get_returncode, sizeof(__pyx_k_get_returncode), 0, 0, 1, 1},
    {&__pyx_n_u_get_running_loop, __pyx_k_get_running_loop, sizeof(__pyx_k_get_running_loop), 0, 1, 0, 1},
    {&__pyx_n_s_get_task_factory, __pyx_k_get_task_factory, sizeof(__pyx_k_get_task_factory), 0, 0, 1, 1},
    {&__pyx_n_s_get_write_buffer_limits, __pyx_k_get_write_buffer_limits, sizeof(__pyx_k_get_write_buffer_limits), 0, 0, 1, 1},
    {&__pyx_n_s_get_write_buffer_size, __pyx_k_get_write_buffer_size, sizeof(__pyx_k_get_write_buffer_size), 0, 0, 1, 1},
    {&__pyx_n_s_getaddrinfo, __pyx_k_getaddrinfo, sizeof(__pyx_k_getaddrinfo), 0, 0, 1, 1},
    {&__pyx_kp_u_getaddrinfo_returned_empty_list, __pyx_k_getaddrinfo_returned_empty_list, sizeof(__pyx_k_getaddrinfo_returned_empty_list), 0, 1, 0, 0},
    {&__pyx_kp_u_getaddrinfo_returned_empty_list_2, __pyx_k_getaddrinfo_returned_empty_list_2, sizeof(__pyx_k_getaddrinfo_returned_empty_list_2), 0, 1, 0, 0},
    {&__pyx_n_s_getfilesystemencoding, __pyx_k_getfilesystemencoding, sizeof(__pyx_k_getfilesystemencoding), 0, 0, 1, 1},
    {&__pyx_n_s_getframe, __pyx_k_getframe, sizeof(__pyx_k_getframe), 0, 0, 1, 1},
    {&__pyx_n_s_getnameinfo, __pyx_k_getnameinfo, sizeof(__pyx_k_getnameinfo), 0, 0, 1, 1},
    {&__pyx_kp_u_getnameinfo_argument_1_must_be_a, __pyx_k_getnameinfo_argument_1_must_be_a, sizeof(__pyx_k_getnameinfo_argument_1_must_be_a), 0, 1, 0, 0},
    {&__pyx_kp_u_getnameinfo_flowinfo_must_be_0_1, __pyx_k_getnameinfo_flowinfo_must_be_0_1, sizeof(__pyx_k_getnameinfo_flowinfo_must_be_0_1), 0, 1, 0, 0},
    {&__pyx_n_s_getpeercert, __pyx_k_getpeercert, sizeof(__pyx_k_getpeercert), 0, 0, 1, 1},
    {&__pyx_n_s_getpeername, __pyx_k_getpeername, sizeof(__pyx_k_getpeername), 0, 0, 1, 1},
    {&__pyx_n_s_getservbyname, __pyx_k_getservbyname, sizeof(__pyx_k_getservbyname), 0, 0, 1, 1},
    {&__pyx_kp_u_getsockaddrarg_scope_id_must_be, __pyx_k_getsockaddrarg_scope_id_must_be, sizeof(__pyx_k_getsockaddrarg_scope_id_must_be), 0, 1, 0, 0},
    {&__pyx_n_s_getsockname, __pyx_k_getsockname, sizeof(__pyx_k_getsockname), 0, 0, 1, 1},
    {&__pyx_n_s_getsockopt, __pyx_k_getsockopt, sizeof(__pyx_k_getsockopt), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_gettimeout, __pyx_k_gettimeout, sizeof(__pyx_k_gettimeout), 0, 0, 1, 1},
    {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
    {&__pyx_n_s_handle, __pyx_k_handle, sizeof(__pyx_k_handle), 0, 0, 1, 1},
    {&__pyx_n_u_handle, __pyx_k_handle, sizeof(__pyx_k_handle), 0, 1, 0, 1},
    {&__pyx_kp_u_handle_data_is_NULL_in___close_a, __pyx_k_handle_data_is_NULL_in___close_a, sizeof(__pyx_k_handle_data_is_NULL_in___close_a), 0, 1, 0, 0},
    {&__pyx_kp_u_handle_loop_data_is_NULL_in___en, __pyx_k_handle_loop_data_is_NULL_in___en, sizeof(__pyx_k_handle_loop_data_is_NULL_in___en), 0, 1, 0, 0},
    {&__pyx_kp_u_handle_loop_is_NULL_in___ensure, __pyx_k_handle_loop_is_NULL_in___ensure, sizeof(__pyx_k_handle_loop_is_NULL_in___ensure), 0, 1, 0, 0},
    {&__pyx_n_s_handler, __pyx_k_handler, sizeof(__pyx_k_handler), 0, 0, 1, 1},
    {&__pyx_kp_u_has_handle_0_in___uv_close_handl, __pyx_k_has_handle_0_in___uv_close_handl, sizeof(__pyx_k_has_handle_0_in___uv_close_handl), 0, 1, 0, 0},
    {&__pyx_kp_u_high, __pyx_k_high, sizeof(__pyx_k_high), 0, 1, 0, 0},
    {&__pyx_n_s_high_2, __pyx_k_high_2, sizeof(__pyx_k_high_2), 0, 0, 1, 1},
    {&__pyx_n_s_host, __pyx_k_host, sizeof(__pyx_k_host), 0, 0, 1, 1},
    {&__pyx_kp_u_host_and_port_was_not_specified, __pyx_k_host_and_port_was_not_specified, sizeof(__pyx_k_host_and_port_was_not_specified), 0, 1, 0, 0},
    {&__pyx_kp_u_host_must_be_a_str_or_bytes, __pyx_k_host_must_be_a_str_or_bytes, sizeof(__pyx_k_host_must_be_a_str_or_bytes), 0, 1, 0, 0},
    {&__pyx_kp_u_host_must_be_a_string_or_bytes_o, __pyx_k_host_must_be_a_string_or_bytes_o, sizeof(__pyx_k_host_must_be_a_string_or_bytes_o), 0, 1, 0, 0},
    {&__pyx_kp_u_host_port_and_sock_can_not_be_sp, __pyx_k_host_port_and_sock_can_not_be_sp, sizeof(__pyx_k_host_port_and_sock_can_not_be_sp), 0, 1, 0, 0},
    {&__pyx_n_s_hosts, __pyx_k_hosts, sizeof(__pyx_k_hosts), 0, 0, 1, 1},
    {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
    {&__pyx_n_s_ident, __pyx_k_ident, sizeof(__pyx_k_ident), 0, 0, 1, 1},
    {&__pyx_n_u_idna, __pyx_k_idna, sizeof(__pyx_k_idna), 0, 1, 0, 1},
    {&__pyx_n_s_ignore_environment, __pyx_k_ignore_environment, sizeof(__pyx_k_ignore_environment), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_u_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 1, 0, 1},
    {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
    {&__pyx_n_s_infos, __pyx_k_infos, sizeof(__pyx_k_infos), 0, 0, 1, 1},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
    {&__pyx_kp_u_integer_argument_expected_got, __pyx_k_integer_argument_expected_got, sizeof(__pyx_k_integer_argument_expected_got), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_Handle_cb_type, __pyx_k_invalid_Handle_cb_type, sizeof(__pyx_k_invalid_Handle_cb_type), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid__init_protocol_call, __pyx_k_invalid__init_protocol_call, sizeof(__pyx_k_invalid__init_protocol_call), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_data_in_writebuf_an_inst, __pyx_k_invalid_data_in_writebuf_an_inst, sizeof(__pyx_k_invalid_data_in_writebuf_an_inst), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_socket_family_expected_A, __pyx_k_invalid_socket_family_expected_A, sizeof(__pyx_k_invalid_socket_family_expected_A), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_waiter_object, __pyx_k_invalid_waiter_object, sizeof(__pyx_k_invalid_waiter_object), 0, 1, 0, 0},
    {&__pyx_n_s_io_refs, __pyx_k_io_refs, sizeof(__pyx_k_io_refs), 0, 0, 1, 1},
    {&__pyx_n_s_ioctl, __pyx_k_ioctl, sizeof(__pyx_k_ioctl), 0, 0, 1, 1},
    {&__pyx_n_s_is_alive, __pyx_k_is_alive, sizeof(__pyx_k_is_alive), 0, 0, 1, 1},
    {&__pyx_kp_u_is_alive_but__handle_data_is_no, __pyx_k_is_alive_but__handle_data_is_no, sizeof(__pyx_k_is_alive_but__handle_data_is_no), 0, 1, 0, 0},
    {&__pyx_kp_u_is_alive_but__handle_is_NULL, __pyx_k_is_alive_but__handle_is_NULL, sizeof(__pyx_k_is_alive_but__handle_is_NULL), 0, 1, 0, 0},
    {&__pyx_kp_u_is_alive_but__handle_loop_is_no, __pyx_k_is_alive_but__handle_loop_is_no, sizeof(__pyx_k_is_alive_but__handle_loop_is_no), 0, 1, 0, 0},
    {&__pyx_kp_u_is_alive_but__loop_is_None, __pyx_k_is_alive_but__loop_is_None, sizeof(__pyx_k_is_alive_but__loop_is_None), 0, 1, 0, 0},
    {&__pyx_kp_u_is_already_being_awaited_on_ser, __pyx_k_is_already_being_awaited_on_ser, sizeof(__pyx_k_is_already_being_awaited_on_ser), 0, 1, 0, 0},
    {&__pyx_n_s_is_closed, __pyx_k_is_closed, sizeof(__pyx_k_is_closed), 0, 0, 1, 1},
    {&__pyx_kp_u_is_closed_2, __pyx_k_is_closed_2, sizeof(__pyx_k_is_closed_2), 0, 1, 0, 0},
    {&__pyx_n_s_is_closing, __pyx_k_is_closing, sizeof(__pyx_k_is_closing), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_is_longer_than_107_characters, __pyx_k_is_longer_than_107_characters, sizeof(__pyx_k_is_longer_than_107_characters), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_supported_by_start_tls, __pyx_k_is_not_supported_by_start_tls, sizeof(__pyx_k_is_not_supported_by_start_tls), 0, 1, 0, 0},
    {&__pyx_kp_u_is_not_supposed_to_be_instantia, __pyx_k_is_not_supposed_to_be_instantia, sizeof(__pyx_k_is_not_supposed_to_be_instantia), 0, 1, 0, 0},
    {&__pyx_kp_u_is_open_in___dealloc___with_loo, __pyx_k_is_open_in___dealloc___with_loo, sizeof(__pyx_k_is_open_in___dealloc___with_loo), 0, 1, 0, 0},
    {&__pyx_n_s_is_reading, __pyx_k_is_reading, sizeof(__pyx_k_is_reading), 0, 0, 1, 1},
    {&__pyx_n_s_is_running, __pyx_k_is_running, sizeof(__pyx_k_is_running), 0, 0, 1, 1},
    {&__pyx_n_s_is_serving, __pyx_k_is_serving, sizeof(__pyx_k_is_serving), 0, 0, 1, 1},
    {&__pyx_n_s_iscoroutine, __pyx_k_iscoroutine, sizeof(__pyx_k_iscoroutine), 0, 0, 1, 1},
    {&__pyx_n_s_iscoroutinefunction, __pyx_k_iscoroutinefunction, sizeof(__pyx_k_iscoroutinefunction), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_u_isfuture, __pyx_k_isfuture, sizeof(__pyx_k_isfuture), 0, 1, 0, 1},
    {&__pyx_n_s_isgenerator, __pyx_k_isgenerator, sizeof(__pyx_k_isgenerator), 0, 0, 1, 1},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
    {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_s_kill, __pyx_k_kill, sizeof(__pyx_k_kill), 0, 0, 1, 1},
    {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
    {&__pyx_kp_u_laddr_s, __pyx_k_laddr_s, sizeof(__pyx_k_laddr_s), 0, 1, 0, 0},
    {&__pyx_n_s_lads, __pyx_k_lads, sizeof(__pyx_k_lads), 0, 0, 1, 1},
    {&__pyx_n_s_lai, __pyx_k_lai, sizeof(__pyx_k_lai), 0, 0, 1, 1},
    {&__pyx_n_s_lai_addr_static, __pyx_k_lai_addr_static, sizeof(__pyx_k_lai_addr_static), 0, 0, 1, 1},
    {&__pyx_n_s_lai_iter, __pyx_k_lai_iter, sizeof(__pyx_k_lai_iter), 0, 0, 1, 1},
    {&__pyx_n_s_lai_static, __pyx_k_lai_static, sizeof(__pyx_k_lai_static), 0, 0, 1, 1},
    {&__pyx_n_s_last, __pyx_k_last, sizeof(__pyx_k_last), 0, 0, 1, 1},
    {&__pyx_n_s_libuv_get_loop_t_ptr, __pyx_k_libuv_get_loop_t_ptr, sizeof(__pyx_k_libuv_get_loop_t_ptr), 0, 0, 1, 1},
    {&__pyx_n_s_libuv_get_version, __pyx_k_libuv_get_version, sizeof(__pyx_k_libuv_get_version), 0, 0, 1, 1},
    {&__pyx_n_s_limit, __pyx_k_limit, sizeof(__pyx_k_limit), 0, 0, 1, 1},
    {&__pyx_n_s_list_of_data, __pyx_k_list_of_data, sizeof(__pyx_k_list_of_data), 0, 0, 1, 1},
    {&__pyx_n_s_listen, __pyx_k_listen, sizeof(__pyx_k_listen), 0, 0, 1, 1},
    {&__pyx_kp_u_listen_method, __pyx_k_listen_method, sizeof(__pyx_k_listen_method), 0, 1, 0, 0},
    {&__pyx_n_s_local_addr, __pyx_k_local_addr, sizeof(__pyx_k_local_addr), 0, 0, 1, 1},
    {&__pyx_kp_u_local_addr_must_be_a_tuple_of_ho, __pyx_k_local_addr_must_be_a_tuple_of_ho, sizeof(__pyx_k_local_addr_must_be_a_tuple_of_ho), 0, 1, 0, 0},
    {&__pyx_kp_u_local_addr_must_be_a_tuple_of_ho_2, __pyx_k_local_addr_must_be_a_tuple_of_ho_2, sizeof(__pyx_k_local_addr_must_be_a_tuple_of_ho_2), 0, 1, 0, 0},
    {&__pyx_n_s_log, __pyx_k_log, sizeof(__pyx_k_log), 0, 0, 1, 1},
    {&__pyx_n_s_log_destroy_pending, __pyx_k_log_destroy_pending, sizeof(__pyx_k_log_destroy_pending), 0, 0, 1, 1},
    {&__pyx_n_s_log_lines, __pyx_k_log_lines, sizeof(__pyx_k_log_lines), 0, 0, 1, 1},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_lookup_lines, __pyx_k_lookup_lines, sizeof(__pyx_k_lookup_lines), 0, 0, 1, 1},
    {&__pyx_n_s_loop, __pyx_k_loop, sizeof(__pyx_k_loop), 0, 0, 1, 1},
    {&__pyx_kp_u_loop__exec_queued_writes, __pyx_k_loop__exec_queued_writes, sizeof(__pyx_k_loop__exec_queued_writes), 0, 1, 0, 0},
    {&__pyx_kp_u_loop__on_idle, __pyx_k_loop__on_idle, sizeof(__pyx_k_loop__on_idle), 0, 1, 0, 0},
    {&__pyx_n_s_low, __pyx_k_low, sizeof(__pyx_k_low), 0, 0, 1, 1},
    {&__pyx_n_s_lower, __pyx_k_lower, sizeof(__pyx_k_lower), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_main_thread, __pyx_k_main_thread, sizeof(__pyx_k_main_thread), 0, 0, 1, 1},
    {&__pyx_n_s_makefile, __pyx_k_makefile, sizeof(__pyx_k_makefile), 0, 0, 1, 1},
    {&__pyx_kp_u_makefile_method, __pyx_k_makefile_method, sizeof(__pyx_k_makefile_method), 0, 1, 0, 0},
    {&__pyx_n_s_maxsize, __pyx_k_maxsize, sizeof(__pyx_k_maxsize), 0, 0, 1, 1},
    {&__pyx_kp_u_maxsize_is_expected_to_be_greate, __pyx_k_maxsize_is_expected_to_be_greate, sizeof(__pyx_k_maxsize_is_expected_to_be_greate), 0, 1, 0, 0},
    {&__pyx_n_s_message, __pyx_k_message, sizeof(__pyx_k_message), 0, 0, 1, 1},
    {&__pyx_n_u_message, __pyx_k_message, sizeof(__pyx_k_message), 0, 1, 0, 1},
    {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
    {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
    {&__pyx_n_s_model, __pyx_k_model, sizeof(__pyx_k_model), 0, 0, 1, 1},
    {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
    {&__pyx_n_s_monitor_fs, __pyx_k_monitor_fs, sizeof(__pyx_k_monitor_fs), 0, 0, 1, 1},
    {&__pyx_n_s_monotonic, __pyx_k_monotonic, sizeof(__pyx_k_monotonic), 0, 0, 1, 1},
    {&__pyx_n_s_move_to_end, __pyx_k_move_to_end, sizeof(__pyx_k_move_to_end), 0, 0, 1, 1},
    {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
    {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
    {&__pyx_kp_u_must_be_0, __pyx_k_must_be_0, sizeof(__pyx_k_must_be_0), 0, 1, 0, 0},
    {&__pyx_kp_u_must_be_low, __pyx_k_must_be_low, sizeof(__pyx_k_must_be_low), 0, 1, 0, 0},
    {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_u_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 1, 0, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_nbytes, __pyx_k_nbytes, sizeof(__pyx_k_nbytes), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_kp_u_new_callbacks_were_queued_during, __pyx_k_new_callbacks_were_queued_during, sizeof(__pyx_k_new_callbacks_were_queued_during), 0, 1, 0, 0},
    {&__pyx_kp_u_new_poll_handles_were_queued_dur, __pyx_k_new_poll_handles_were_queued_dur, sizeof(__pyx_k_new_poll_handles_were_queued_dur), 0, 1, 0, 0},
    {&__pyx_n_s_new_task, __pyx_k_new_task, sizeof(__pyx_k_new_task), 0, 0, 1, 1},
    {&__pyx_kp_u_new_timers_were_queued_during_lo, __pyx_k_new_timers_were_queued_during_lo, sizeof(__pyx_k_new_timers_were_queued_during_lo), 0, 1, 0, 0},
    {&__pyx_n_s_nexc, __pyx_k_nexc, sizeof(__pyx_k_nexc), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_kp_u_no_path_and_sock_were_specified, __pyx_k_no_path_and_sock_were_specified, sizeof(__pyx_k_no_path_and_sock_were_specified), 0, 1, 0, 0},
    {&__pyx_kp_u_no_python_buffer_is_allocated_in, __pyx_k_no_python_buffer_is_allocated_in, sizeof(__pyx_k_no_python_buffer_is_allocated_in), 0, 1, 0, 0},
    {&__pyx_n_s_noop, __pyx_k_noop, sizeof(__pyx_k_noop), 0, 0, 1, 1},
    {&__pyx_n_s_noop_2, __pyx_k_noop_2, sizeof(__pyx_k_noop_2), 0, 0, 1, 1},
    {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
    {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
    {&__pyx_kp_u_object_created_at_most_recent_ca, __pyx_k_object_created_at_most_recent_ca, sizeof(__pyx_k_object_created_at_most_recent_ca), 0, 1, 0, 0},
    {&__pyx_kp_u_objs_handles, __pyx_k_objs_handles, sizeof(__pyx_k_objs_handles), 0, 1, 0, 0},
    {&__pyx_n_s_old_agen_hooks, __pyx_k_old_agen_hooks, sizeof(__pyx_k_old_agen_hooks), 0, 0, 1, 1},
    {&__pyx_n_s_on_ssl_connected, __pyx_k_on_ssl_connected, sizeof(__pyx_k_on_ssl_connected), 0, 0, 1, 1},
    {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
    {&__pyx_kp_u_open__StreamWriteContext_is_bein, __pyx_k_open__StreamWriteContext_is_bein, sizeof(__pyx_k_open__StreamWriteContext_is_bein), 0, 1, 0, 0},
    {&__pyx_kp_u_open__UDPSendContext_is_being_de, __pyx_k_open__UDPSendContext_is_being_de, sizeof(__pyx_k_open__UDPSendContext_is_being_de), 0, 1, 0, 0},
    {&__pyx_n_s_opts, __pyx_k_opts, sizeof(__pyx_k_opts), 0, 0, 1, 1},
    {&__pyx_n_s_orig_path, __pyx_k_orig_path, sizeof(__pyx_k_orig_path), 0, 0, 1, 1},
    {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
    {&__pyx_n_s_p_bytes, __pyx_k_p_bytes, sizeof(__pyx_k_p_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
    {&__pyx_n_s_pass_fds, __pyx_k_pass_fds, sizeof(__pyx_k_pass_fds), 0, 0, 1, 1},
    {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
    {&__pyx_kp_u_path_and_sock_can_not_be_specifi, __pyx_k_path_and_sock_can_not_be_specifi, sizeof(__pyx_k_path_and_sock_can_not_be_specifi), 0, 1, 0, 0},
    {&__pyx_kp_u_path_was_not_specified_and_no_so, __pyx_k_path_was_not_specified_and_no_so, sizeof(__pyx_k_path_was_not_specified_and_no_so), 0, 1, 0, 0},
    {&__pyx_n_s_pause_reading, __pyx_k_pause_reading, sizeof(__pyx_k_pause_reading), 0, 0, 1, 1},
    {&__pyx_n_s_pause_writing, __pyx_k_pause_writing, sizeof(__pyx_k_pause_writing), 0, 0, 1, 1},
    {&__pyx_n_s_peercert, __pyx_k_peercert, sizeof(__pyx_k_peercert), 0, 0, 1, 1},
    {&__pyx_n_u_peername, __pyx_k_peername, sizeof(__pyx_k_peername), 0, 1, 0, 1},
    {&__pyx_n_s_pending, __pyx_k_pending, sizeof(__pyx_k_pending), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_pipe, __pyx_k_pipe, sizeof(__pyx_k_pipe), 0, 0, 1, 1},
    {&__pyx_n_u_pipe, __pyx_k_pipe, sizeof(__pyx_k_pipe), 0, 1, 0, 1},
    {&__pyx_kp_u_pipe_2, __pyx_k_pipe_2, sizeof(__pyx_k_pipe_2), 0, 1, 0, 0},
    {&__pyx_n_s_pipe_connection_lost, __pyx_k_pipe_connection_lost, sizeof(__pyx_k_pipe_connection_lost), 0, 0, 1, 1},
    {&__pyx_n_s_pipe_data_received, __pyx_k_pipe_data_received, sizeof(__pyx_k_pipe_data_received), 0, 0, 1, 1},
    {&__pyx_n_s_platform, __pyx_k_platform, sizeof(__pyx_k_platform), 0, 0, 1, 1},
    {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
    {&__pyx_n_s_popitem, __pyx_k_popitem, sizeof(__pyx_k_popitem), 0, 0, 1, 1},
    {&__pyx_n_s_popleft, __pyx_k_popleft, sizeof(__pyx_k_popleft), 0, 0, 1, 1},
    {&__pyx_n_s_port, __pyx_k_port, sizeof(__pyx_k_port), 0, 0, 1, 1},
    {&__pyx_kp_u_port_must_be_a_str_bytes_or_int, __pyx_k_port_must_be_a_str_bytes_or_int, sizeof(__pyx_k_port_must_be_a_str_bytes_or_int), 0, 1, 0, 0},
    {&__pyx_n_u_posix, __pyx_k_posix, sizeof(__pyx_k_posix), 0, 1, 0, 1},
    {&__pyx_n_s_preexec_fn, __pyx_k_preexec_fn, sizeof(__pyx_k_preexec_fn), 0, 0, 1, 1},
    {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
    {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
    {&__pyx_n_u_print_debug_info, __pyx_k_print_debug_info, sizeof(__pyx_k_print_debug_info), 0, 1, 0, 1},
    {&__pyx_n_s_problems, __pyx_k_problems, sizeof(__pyx_k_problems), 0, 0, 1, 1},
    {&__pyx_n_s_proc, __pyx_k_proc, sizeof(__pyx_k_proc), 0, 0, 1, 1},
    {&__pyx_n_s_process_exited, __pyx_k_process_exited, sizeof(__pyx_k_process_exited), 0, 0, 1, 1},
    {&__pyx_n_s_process_id, __pyx_k_process_id, sizeof(__pyx_k_process_id), 0, 0, 1, 1},
    {&__pyx_n_s_program, __pyx_k_program, sizeof(__pyx_k_program), 0, 0, 1, 1},
    {&__pyx_n_s_proto, __pyx_k_proto, sizeof(__pyx_k_proto), 0, 0, 1, 1},
    {&__pyx_n_s_proto_factory, __pyx_k_proto_factory, sizeof(__pyx_k_proto_factory), 0, 0, 1, 1},
    {&__pyx_n_s_protocol, __pyx_k_protocol, sizeof(__pyx_k_protocol), 0, 0, 1, 1},
    {&__pyx_n_u_protocol, __pyx_k_protocol, sizeof(__pyx_k_protocol), 0, 1, 0, 1},
    {&__pyx_n_s_protocol_factory, __pyx_k_protocol_factory, sizeof(__pyx_k_protocol_factory), 0, 0, 1, 1},
    {&__pyx_kp_u_protocol_is_not_set_cannot_call, __pyx_k_protocol_is_not_set_cannot_call, sizeof(__pyx_k_protocol_is_not_set_cannot_call), 0, 1, 0, 0},
    {&__pyx_kp_u_protocol_is_required, __pyx_k_protocol_is_required, sizeof(__pyx_k_protocol_is_required), 0, 1, 0, 0},
    {&__pyx_kp_u_protocol_pause_writing_failed, __pyx_k_protocol_pause_writing_failed, sizeof(__pyx_k_protocol_pause_writing_failed), 0, 1, 0, 0},
    {&__pyx_kp_u_protocol_resume_writing_failed, __pyx_k_protocol_resume_writing_failed, sizeof(__pyx_k_protocol_resume_writing_failed), 0, 1, 0, 0},
    {&__pyx_n_s_pyaddr, __pyx_k_pyaddr, sizeof(__pyx_k_pyaddr), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_LruCache, __pyx_k_pyx_unpickle_LruCache, sizeof(__pyx_k_pyx_unpickle_LruCache), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PseudoSocket, __pyx_k_pyx_unpickle_PseudoSocket, sizeof(__pyx_k_pyx_unpickle_PseudoSocket), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
    {&__pyx_n_u_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 1, 0, 1},
    {&__pyx_kp_u_r_SSL_handshake_took_1f_ms, __pyx_k_r_SSL_handshake_took_1f_ms, sizeof(__pyx_k_r_SSL_handshake_took_1f_ms), 0, 1, 0, 0},
    {&__pyx_kp_u_r_fileno_returned_non_integer, __pyx_k_r_fileno_returned_non_integer, sizeof(__pyx_k_r_fileno_returned_non_integer), 0, 1, 0, 0},
    {&__pyx_kp_u_r_received_EOF, __pyx_k_r_received_EOF, sizeof(__pyx_k_r_received_EOF), 0, 1, 0, 0},
    {&__pyx_kp_u_r_s, __pyx_k_r_s, sizeof(__pyx_k_r_s), 0, 1, 0, 0},
    {&__pyx_kp_u_r_starts_SSL_handshake, __pyx_k_r_starts_SSL_handshake, sizeof(__pyx_k_r_starts_SSL_handshake), 0, 1, 0, 0},
    {&__pyx_kp_u_raddr_s, __pyx_k_raddr_s, sizeof(__pyx_k_raddr_s), 0, 1, 0, 0},
    {&__pyx_n_s_rads, __pyx_k_rads, sizeof(__pyx_k_rads), 0, 0, 1, 1},
    {&__pyx_n_s_rai, __pyx_k_rai, sizeof(__pyx_k_rai), 0, 0, 1, 1},
    {&__pyx_n_s_rai_addr_static, __pyx_k_rai_addr_static, sizeof(__pyx_k_rai_addr_static), 0, 0, 1, 1},
    {&__pyx_n_s_rai_iter, __pyx_k_rai_iter, sizeof(__pyx_k_rai_iter), 0, 0, 1, 1},
    {&__pyx_n_s_rai_static, __pyx_k_rai_static, sizeof(__pyx_k_rai_static), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
    {&__pyx_n_s_recv, __pyx_k_recv, sizeof(__pyx_k_recv), 0, 0, 1, 1},
    {&__pyx_n_s_recv_into, __pyx_k_recv_into, sizeof(__pyx_k_recv_into), 0, 0, 1, 1},
    {&__pyx_kp_u_recv_into_method, __pyx_k_recv_into_method, sizeof(__pyx_k_recv_into_method), 0, 1, 0, 0},
    {&__pyx_kp_u_recv_method, __pyx_k_recv_method, sizeof(__pyx_k_recv_method), 0, 1, 0, 0},
    {&__pyx_n_s_recvfrom, __pyx_k_recvfrom, sizeof(__pyx_k_recvfrom), 0, 0, 1, 1},
    {&__pyx_n_s_recvfrom_into, __pyx_k_recvfrom_into, sizeof(__pyx_k_recvfrom_into), 0, 0, 1, 1},
    {&__pyx_kp_u_recvfrom_into_method, __pyx_k_recvfrom_into_method, sizeof(__pyx_k_recvfrom_into_method), 0, 1, 0, 0},
    {&__pyx_kp_u_recvfrom_method, __pyx_k_recvfrom_method, sizeof(__pyx_k_recvfrom_method), 0, 1, 0, 0},
    {&__pyx_n_s_recvmsg, __pyx_k_recvmsg, sizeof(__pyx_k_recvmsg), 0, 0, 1, 1},
    {&__pyx_n_s_recvmsg_into, __pyx_k_recvmsg_into, sizeof(__pyx_k_recvmsg_into), 0, 0, 1, 1},
    {&__pyx_kp_u_recvmsg_into_method, __pyx_k_recvmsg_into_method, sizeof(__pyx_k_recvmsg_into_method), 0, 1, 0, 0},
    {&__pyx_kp_u_recvmsg_method, __pyx_k_recvmsg_method, sizeof(__pyx_k_recvmsg_method), 0, 1, 0, 0},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_remote_addr, __pyx_k_remote_addr, sizeof(__pyx_k_remote_addr), 0, 0, 1, 1},
    {&__pyx_kp_u_remote_addr_must_be_a_tuple_of_h, __pyx_k_remote_addr_must_be_a_tuple_of_h, sizeof(__pyx_k_remote_addr_must_be_a_tuple_of_h), 0, 1, 0, 0},
    {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
    {&__pyx_n_s_remove_done_callback, __pyx_k_remove_done_callback, sizeof(__pyx_k_remove_done_callback), 0, 0, 1, 1},
    {&__pyx_n_s_remove_reader, __pyx_k_remove_reader, sizeof(__pyx_k_remove_reader), 0, 0, 1, 1},
    {&__pyx_n_s_remove_reader_2, __pyx_k_remove_reader_2, sizeof(__pyx_k_remove_reader_2), 0, 0, 1, 1},
    {&__pyx_n_s_remove_signal_handler, __pyx_k_remove_signal_handler, sizeof(__pyx_k_remove_signal_handler), 0, 0, 1, 1},
    {&__pyx_kp_u_remove_signal_handler_can_only_b, __pyx_k_remove_signal_handler_can_only_b, sizeof(__pyx_k_remove_signal_handler_can_only_b), 0, 1, 0, 0},
    {&__pyx_n_s_remove_writer, __pyx_k_remove_writer, sizeof(__pyx_k_remove_writer), 0, 0, 1, 1},
    {&__pyx_n_s_remove_writer_2, __pyx_k_remove_writer_2, sizeof(__pyx_k_remove_writer_2), 0, 0, 1, 1},
    {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
    {&__pyx_n_s_restore_signals, __pyx_k_restore_signals, sizeof(__pyx_k_restore_signals), 0, 0, 1, 1},
    {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
    {&__pyx_n_s_results, __pyx_k_results, sizeof(__pyx_k_results), 0, 0, 1, 1},
    {&__pyx_n_s_resume_cb, __pyx_k_resume_cb, sizeof(__pyx_k_resume_cb), 0, 0, 1, 1},
    {&__pyx_n_s_resume_reading, __pyx_k_resume_reading, sizeof(__pyx_k_resume_reading), 0, 0, 1, 1},
    {&__pyx_n_s_resume_writing, __pyx_k_resume_writing, sizeof(__pyx_k_resume_writing), 0, 0, 1, 1},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_s_return_exceptions, __pyx_k_return_exceptions, sizeof(__pyx_k_return_exceptions), 0, 0, 1, 1},
    {&__pyx_kp_u_returning_true_from_eof_received, __pyx_k_returning_true_from_eof_received, sizeof(__pyx_k_returning_true_from_eof_received), 0, 1, 0, 0},
    {&__pyx_n_s_reuse_address, __pyx_k_reuse_address, sizeof(__pyx_k_reuse_address), 0, 0, 1, 1},
    {&__pyx_n_s_reuse_port, __pyx_k_reuse_port, sizeof(__pyx_k_reuse_port), 0, 0, 1, 1},
    {&__pyx_kp_u_reuse_port_not_supported_by_sock, __pyx_k_reuse_port_not_supported_by_sock, sizeof(__pyx_k_reuse_port_not_supported_by_sock), 0, 1, 0, 0},
    {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
    {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
    {&__pyx_n_s_rstrip, __pyx_k_rstrip, sizeof(__pyx_k_rstrip), 0, 0, 1, 1},
    {&__pyx_n_s_run, __pyx_k_run, sizeof(__pyx_k_run), 0, 0, 1, 1},
    {&__pyx_n_s_run_forever, __pyx_k_run_forever, sizeof(__pyx_k_run_forever), 0, 0, 1, 1},
    {&__pyx_n_s_run_in_executor, __pyx_k_run_in_executor, sizeof(__pyx_k_run_in_executor), 0, 0, 1, 1},
    {&__pyx_n_s_run_until_complete, __pyx_k_run_until_complete, sizeof(__pyx_k_run_until_complete), 0, 0, 1, 1},
    {&__pyx_n_s_run_until_complete_locals_done_c, __pyx_k_run_until_complete_locals_done_c, sizeof(__pyx_k_run_until_complete_locals_done_c), 0, 0, 1, 1},
    {&__pyx_kp_u_running_closed_debug, __pyx_k_running_closed_debug, sizeof(__pyx_k_running_closed_debug), 0, 1, 0, 0},
    {&__pyx_n_s_scope_id, __pyx_k_scope_id, sizeof(__pyx_k_scope_id), 0, 0, 1, 1},
    {&__pyx_kp_u_seconds, __pyx_k_seconds, sizeof(__pyx_k_seconds), 0, 1, 0, 0},
    {&__pyx_kp_u_seconds_aborting_the_connection, __pyx_k_seconds_aborting_the_connection, sizeof(__pyx_k_seconds_aborting_the_connection), 0, 1, 0, 0},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_self_2, __pyx_k_self_2, sizeof(__pyx_k_self_2), 0, 0, 1, 1},
    {&__pyx_n_u_self_2, __pyx_k_self_2, sizeof(__pyx_k_self_2), 0, 1, 0, 1},
    {&__pyx_kp_u_self_pipe_exists_before_loop_run, __pyx_k_self_pipe_exists_before_loop_run, sizeof(__pyx_k_self_pipe_exists_before_loop_run), 0, 1, 0, 0},
    {&__pyx_kp_u_self_pipe_was_not_cleaned_up_aft, __pyx_k_self_pipe_was_not_cleaned_up_aft, sizeof(__pyx_k_self_pipe_was_not_cleaned_up_aft), 0, 1, 0, 0},
    {&__pyx_kp_s_self_py_buf_self_req_cannot_be_c, __pyx_k_self_py_buf_self_req_cannot_be_c, sizeof(__pyx_k_self_py_buf_self_req_cannot_be_c), 0, 0, 1, 0},
    {&__pyx_kp_s_self_py_bufs_self_py_bufs_sml_se, __pyx_k_self_py_bufs_self_py_bufs_sml_se, sizeof(__pyx_k_self_py_bufs_self_py_bufs_sml_se), 0, 0, 1, 0},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_kp_u_send_method, __pyx_k_send_method, sizeof(__pyx_k_send_method), 0, 1, 0, 0},
    {&__pyx_n_s_send_signal, __pyx_k_send_signal, sizeof(__pyx_k_send_signal), 0, 0, 1, 1},
    {&__pyx_n_s_sendall, __pyx_k_sendall, sizeof(__pyx_k_sendall), 0, 0, 1, 1},
    {&__pyx_kp_u_sendall_method, __pyx_k_sendall_method, sizeof(__pyx_k_sendall_method), 0, 1, 0, 0},
    {&__pyx_n_s_sendfile, __pyx_k_sendfile, sizeof(__pyx_k_sendfile), 0, 0, 1, 1},
    {&__pyx_kp_u_sendfile_method, __pyx_k_sendfile_method, sizeof(__pyx_k_sendfile_method), 0, 1, 0, 0},
    {&__pyx_n_s_sendmsg, __pyx_k_sendmsg, sizeof(__pyx_k_sendmsg), 0, 0, 1, 1},
    {&__pyx_n_s_sendmsg_afalg, __pyx_k_sendmsg_afalg, sizeof(__pyx_k_sendmsg_afalg), 0, 0, 1, 1},
    {&__pyx_kp_u_sendmsg_afalg_method, __pyx_k_sendmsg_afalg_method, sizeof(__pyx_k_sendmsg_afalg_method), 0, 1, 0, 0},
    {&__pyx_kp_u_sendmsg_method, __pyx_k_sendmsg_method, sizeof(__pyx_k_sendmsg_method), 0, 1, 0, 0},
    {&__pyx_n_s_sendto, __pyx_k_sendto, sizeof(__pyx_k_sendto), 0, 0, 1, 1},
    {&__pyx_kp_u_sendto_method, __pyx_k_sendto_method, sizeof(__pyx_k_sendto_method), 0, 1, 0, 0},
    {&__pyx_n_s_serve_forever, __pyx_k_serve_forever, sizeof(__pyx_k_serve_forever), 0, 0, 1, 1},
    {&__pyx_kp_u_server, __pyx_k_server, sizeof(__pyx_k_server), 0, 1, 0, 0},
    {&__pyx_n_s_server_2, __pyx_k_server_2, sizeof(__pyx_k_server_2), 0, 0, 1, 1},
    {&__pyx_n_s_server_hostname, __pyx_k_server_hostname, sizeof(__pyx_k_server_hostname), 0, 0, 1, 1},
    {&__pyx_kp_u_server_hostname_is_only_meaningf, __pyx_k_server_hostname_is_only_meaningf, sizeof(__pyx_k_server_hostname_is_only_meaningf), 0, 1, 0, 0},
    {&__pyx_n_s_server_side, __pyx_k_server_side, sizeof(__pyx_k_server_side), 0, 0, 1, 1},
    {&__pyx_n_s_servers, __pyx_k_servers, sizeof(__pyx_k_servers), 0, 0, 1, 1},
    {&__pyx_kp_u_service_proto_not_found, __pyx_k_service_proto_not_found, sizeof(__pyx_k_service_proto_not_found), 0, 1, 0, 0},
    {&__pyx_n_s_set_asyncgen_hooks, __pyx_k_set_asyncgen_hooks, sizeof(__pyx_k_set_asyncgen_hooks), 0, 0, 1, 1},
    {&__pyx_n_s_set_coroutine_origin_tracking_de, __pyx_k_set_coroutine_origin_tracking_de, sizeof(__pyx_k_set_coroutine_origin_tracking_de), 0, 0, 1, 1},
    {&__pyx_n_u_set_coroutine_wrapper, __pyx_k_set_coroutine_wrapper, sizeof(__pyx_k_set_coroutine_wrapper), 0, 1, 0, 1},
    {&__pyx_n_s_set_debug, __pyx_k_set_debug, sizeof(__pyx_k_set_debug), 0, 0, 1, 1},
    {&__pyx_n_s_set_default_executor, __pyx_k_set_default_executor, sizeof(__pyx_k_set_default_executor), 0, 0, 1, 1},
    {&__pyx_n_s_set_exception, __pyx_k_set_exception, sizeof(__pyx_k_set_exception), 0, 0, 1, 1},
    {&__pyx_n_s_set_exception_handler, __pyx_k_set_exception_handler, sizeof(__pyx_k_set_exception_handler), 0, 0, 1, 1},
    {&__pyx_n_s_set_inheritable, __pyx_k_set_inheritable, sizeof(__pyx_k_set_inheritable), 0, 0, 1, 1},
    {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
    {&__pyx_n_s_set_name_2, __pyx_k_set_name_2, sizeof(__pyx_k_set_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_set_protocol, __pyx_k_set_protocol, sizeof(__pyx_k_set_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_set_read_buffer_limits, __pyx_k_set_read_buffer_limits, sizeof(__pyx_k_set_read_buffer_limits), 0, 0, 1, 1},
    {&__pyx_n_s_set_result, __pyx_k_set_result, sizeof(__pyx_k_set_result), 0, 0, 1, 1},
    {&__pyx_n_u_set_running_loop, __pyx_k_set_running_loop, sizeof(__pyx_k_set_running_loop), 0, 1, 0, 1},
    {&__pyx_n_s_set_task_factory, __pyx_k_set_task_factory, sizeof(__pyx_k_set_task_factory), 0, 0, 1, 1},
    {&__pyx_n_s_set_wakeup_fd, __pyx_k_set_wakeup_fd, sizeof(__pyx_k_set_wakeup_fd), 0, 0, 1, 1},
    {&__pyx_kp_u_set_wakeup_fd_1_failed_s, __pyx_k_set_wakeup_fd_1_failed_s, sizeof(__pyx_k_set_wakeup_fd_1_failed_s), 0, 1, 0, 0},
    {&__pyx_n_s_set_write_buffer_limits, __pyx_k_set_write_buffer_limits, sizeof(__pyx_k_set_write_buffer_limits), 0, 0, 1, 1},
    {&__pyx_n_s_setblocking, __pyx_k_setblocking, sizeof(__pyx_k_setblocking), 0, 0, 1, 1},
    {&__pyx_kp_u_setblocking_transport_sockets_ca, __pyx_k_setblocking_transport_sockets_ca, sizeof(__pyx_k_setblocking_transport_sockets_ca), 0, 1, 0, 0},
    {&__pyx_n_s_setsockopt, __pyx_k_setsockopt, sizeof(__pyx_k_setsockopt), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_settimeout, __pyx_k_settimeout, sizeof(__pyx_k_settimeout), 0, 0, 1, 1},
    {&__pyx_kp_u_settimeout_only_0_timeout_is_all, __pyx_k_settimeout_only_0_timeout_is_all, sizeof(__pyx_k_settimeout_only_0_timeout_is_all), 0, 1, 0, 0},
    {&__pyx_n_s_share, __pyx_k_share, sizeof(__pyx_k_share), 0, 0, 1, 1},
    {&__pyx_n_s_shell, __pyx_k_shell, sizeof(__pyx_k_shell), 0, 0, 1, 1},
    {&__pyx_kp_u_shell_must_be_False, __pyx_k_shell_must_be_False, sizeof(__pyx_k_shell_must_be_False), 0, 1, 0, 0},
    {&__pyx_kp_u_shell_must_be_True, __pyx_k_shell_must_be_True, sizeof(__pyx_k_shell_must_be_True), 0, 1, 0, 0},
    {&__pyx_n_s_shutdown, __pyx_k_shutdown, sizeof(__pyx_k_shutdown), 0, 0, 1, 1},
    {&__pyx_n_s_shutdown_asyncgens, __pyx_k_shutdown_asyncgens, sizeof(__pyx_k_shutdown_asyncgens), 0, 0, 1, 1},
    {&__pyx_n_s_shutdown_coro, __pyx_k_shutdown_coro, sizeof(__pyx_k_shutdown_coro), 0, 0, 1, 1},
    {&__pyx_n_s_shutdown_default_executor, __pyx_k_shutdown_default_executor, sizeof(__pyx_k_shutdown_default_executor), 0, 0, 1, 1},
    {&__pyx_kp_u_shutdown_method, __pyx_k_shutdown_method, sizeof(__pyx_k_shutdown_method), 0, 1, 0, 0},
    {&__pyx_n_s_sig, __pyx_k_sig, sizeof(__pyx_k_sig), 0, 0, 1, 1},
    {&__pyx_kp_u_sig_cannot_be_caught, __pyx_k_sig_cannot_be_caught, sizeof(__pyx_k_sig_cannot_be_caught), 0, 1, 0, 0},
    {&__pyx_kp_u_sig_must_be_an_int_not_r, __pyx_k_sig_must_be_an_int_not_r, sizeof(__pyx_k_sig_must_be_an_int_not_r), 0, 1, 0, 0},
    {&__pyx_kp_u_sig_out_of_range_1, __pyx_k_sig_out_of_range_1, sizeof(__pyx_k_sig_out_of_range_1), 0, 1, 0, 0},
    {&__pyx_n_s_sighandler, __pyx_k_sighandler, sizeof(__pyx_k_sighandler), 0, 0, 1, 1},
    {&__pyx_n_s_siginterrupt, __pyx_k_siginterrupt, sizeof(__pyx_k_siginterrupt), 0, 0, 1, 1},
    {&__pyx_n_s_signal, __pyx_k_signal, sizeof(__pyx_k_signal), 0, 0, 1, 1},
    {&__pyx_kp_u_signals_handling_has_been_alread, __pyx_k_signals_handling_has_been_alread, sizeof(__pyx_k_signals_handling_has_been_alread), 0, 1, 0, 0},
    {&__pyx_kp_u_signals_handling_has_not_been_se, __pyx_k_signals_handling_has_not_been_se, sizeof(__pyx_k_signals_handling_has_not_been_se), 0, 1, 0, 0},
    {&__pyx_n_s_signum, __pyx_k_signum, sizeof(__pyx_k_signum), 0, 0, 1, 1},
    {&__pyx_n_s_sin6, __pyx_k_sin6, sizeof(__pyx_k_sin6), 0, 0, 1, 1},
    {&__pyx_n_s_sl, __pyx_k_sl, sizeof(__pyx_k_sl), 0, 0, 1, 1},
    {&__pyx_n_s_sleep, __pyx_k_sleep, sizeof(__pyx_k_sleep), 0, 0, 1, 1},
    {&__pyx_n_s_sock, __pyx_k_sock, sizeof(__pyx_k_sock), 0, 0, 1, 1},
    {&__pyx_n_s_sock_accept, __pyx_k_sock_accept, sizeof(__pyx_k_sock_accept), 0, 0, 1, 1},
    {&__pyx_n_s_sock_connect, __pyx_k_sock_connect, sizeof(__pyx_k_sock_connect), 0, 0, 1, 1},
    {&__pyx_kp_u_sock_connect_cb_is_called_on_a, __pyx_k_sock_connect_cb_is_called_on_a, sizeof(__pyx_k_sock_connect_cb_is_called_on_a), 0, 1, 0, 0},
    {&__pyx_n_s_sock_recv, __pyx_k_sock_recv, sizeof(__pyx_k_sock_recv), 0, 0, 1, 1},
    {&__pyx_n_s_sock_recv_into, __pyx_k_sock_recv_into, sizeof(__pyx_k_sock_recv_into), 0, 0, 1, 1},
    {&__pyx_kp_u_sock_recv_into_is_called_on_a_c, __pyx_k_sock_recv_into_is_called_on_a_c, sizeof(__pyx_k_sock_recv_into_is_called_on_a_c), 0, 1, 0, 0},
    {&__pyx_kp_u_sock_recv_is_called_on_a_cancel, __pyx_k_sock_recv_is_called_on_a_cancel, sizeof(__pyx_k_sock_recv_is_called_on_a_cancel), 0, 1, 0, 0},
    {&__pyx_n_s_sock_recvfrom, __pyx_k_sock_recvfrom, sizeof(__pyx_k_sock_recvfrom), 0, 0, 1, 1},
    {&__pyx_n_s_sock_recvfrom_into, __pyx_k_sock_recvfrom_into, sizeof(__pyx_k_sock_recvfrom_into), 0, 0, 1, 1},
    {&__pyx_n_s_sock_sendall, __pyx_k_sock_sendall, sizeof(__pyx_k_sock_sendall), 0, 0, 1, 1},
    {&__pyx_kp_u_sock_sendall_is_called_on_a_can, __pyx_k_sock_sendall_is_called_on_a_can, sizeof(__pyx_k_sock_sendall_is_called_on_a_can), 0, 1, 0, 0},
    {&__pyx_n_s_sock_sendto, __pyx_k_sock_sendto, sizeof(__pyx_k_sock_sendto), 0, 0, 1, 1},
    {&__pyx_n_s_sockaddr, __pyx_k_sockaddr, sizeof(__pyx_k_sockaddr), 0, 0, 1, 1},
    {&__pyx_kp_u_sockaddr_must_be_a_tuple_of_2_3, __pyx_k_sockaddr_must_be_a_tuple_of_2_3, sizeof(__pyx_k_sockaddr_must_be_a_tuple_of_2_3), 0, 1, 0, 0},
    {&__pyx_kp_u_sockaddr_resolved_to_multiple_ad, __pyx_k_sockaddr_resolved_to_multiple_ad, sizeof(__pyx_k_sockaddr_resolved_to_multiple_ad), 0, 1, 0, 0},
    {&__pyx_kp_u_socket, __pyx_k_socket, sizeof(__pyx_k_socket), 0, 1, 0, 0},
    {&__pyx_n_s_socket_2, __pyx_k_socket_2, sizeof(__pyx_k_socket_2), 0, 0, 1, 1},
    {&__pyx_n_u_socket_2, __pyx_k_socket_2, sizeof(__pyx_k_socket_2), 0, 1, 0, 1},
    {&__pyx_kp_u_socket_family_mismatch_or_a_DNS, __pyx_k_socket_family_mismatch_or_a_DNS, sizeof(__pyx_k_socket_family_mismatch_or_a_DNS), 0, 1, 0, 0},
    {&__pyx_kp_u_socket_modifier_keyword_argument, __pyx_k_socket_modifier_keyword_argument, sizeof(__pyx_k_socket_modifier_keyword_argument), 0, 1, 0, 0},
    {&__pyx_kp_u_socket_send_raised_exception, __pyx_k_socket_send_raised_exception, sizeof(__pyx_k_socket_send_raised_exception), 0, 1, 0, 0},
    {&__pyx_n_s_socketpair, __pyx_k_socketpair, sizeof(__pyx_k_socketpair), 0, 0, 1, 1},
    {&__pyx_kp_u_sockets, __pyx_k_sockets, sizeof(__pyx_k_sockets), 0, 1, 0, 0},
    {&__pyx_n_s_sockets_2, __pyx_k_sockets_2, sizeof(__pyx_k_sockets_2), 0, 0, 1, 1},
    {&__pyx_n_u_sockname, __pyx_k_sockname, sizeof(__pyx_k_sockname), 0, 1, 0, 1},
    {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
    {&__pyx_n_u_source_traceback, __pyx_k_source_traceback, sizeof(__pyx_k_source_traceback), 0, 1, 0, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
    {&__pyx_n_s_ssl, __pyx_k_ssl, sizeof(__pyx_k_ssl), 0, 0, 1, 1},
    {&__pyx_kp_u_ssl_argument_must_be_an_SSLConte, __pyx_k_ssl_argument_must_be_an_SSLConte, sizeof(__pyx_k_ssl_argument_must_be_an_SSLConte), 0, 1, 0, 0},
    {&__pyx_n_s_ssl_handshake_timeout, __pyx_k_ssl_handshake_timeout, sizeof(__pyx_k_ssl_handshake_timeout), 0, 0, 1, 1},
    {&__pyx_kp_u_ssl_handshake_timeout_is_only_me, __pyx_k_ssl_handshake_timeout_is_only_me, sizeof(__pyx_k_ssl_handshake_timeout_is_only_me), 0, 1, 0, 0},
    {&__pyx_kp_u_ssl_handshake_timeout_should_be, __pyx_k_ssl_handshake_timeout_should_be, sizeof(__pyx_k_ssl_handshake_timeout_should_be), 0, 1, 0, 0},
    {&__pyx_kp_u_ssl_is_expected_to_be_None_or_an, __pyx_k_ssl_is_expected_to_be_None_or_an, sizeof(__pyx_k_ssl_is_expected_to_be_None_or_an), 0, 1, 0, 0},
    {&__pyx_n_s_ssl_object, __pyx_k_ssl_object, sizeof(__pyx_k_ssl_object), 0, 0, 1, 1},
    {&__pyx_n_s_ssl_protocol, __pyx_k_ssl_protocol, sizeof(__pyx_k_ssl_protocol), 0, 0, 1, 1},
    {&__pyx_n_s_ssl_shutdown_timeout, __pyx_k_ssl_shutdown_timeout, sizeof(__pyx_k_ssl_shutdown_timeout), 0, 0, 1, 1},
    {&__pyx_kp_u_ssl_shutdown_timeout_is_only_mea, __pyx_k_ssl_shutdown_timeout_is_only_mea, sizeof(__pyx_k_ssl_shutdown_timeout_is_only_mea), 0, 1, 0, 0},
    {&__pyx_kp_u_ssl_shutdown_timeout_should_be_a, __pyx_k_ssl_shutdown_timeout_should_be_a, sizeof(__pyx_k_ssl_shutdown_timeout_should_be_a), 0, 1, 0, 0},
    {&__pyx_n_s_ssl_waiter, __pyx_k_ssl_waiter, sizeof(__pyx_k_ssl_waiter), 0, 0, 1, 1},
    {&__pyx_n_s_sslcontext, __pyx_k_sslcontext, sizeof(__pyx_k_sslcontext), 0, 0, 1, 1},
    {&__pyx_kp_u_sslcontext_is_expected_to_be_an, __pyx_k_sslcontext_is_expected_to_be_an, sizeof(__pyx_k_sslcontext_is_expected_to_be_an), 0, 1, 0, 0},
    {&__pyx_n_s_sslproto, __pyx_k_sslproto, sizeof(__pyx_k_sslproto), 0, 0, 1, 1},
    {&__pyx_n_s_st_ino, __pyx_k_st_ino, sizeof(__pyx_k_st_ino), 0, 0, 1, 1},
    {&__pyx_n_s_st_mode, __pyx_k_st_mode, sizeof(__pyx_k_st_mode), 0, 0, 1, 1},
    {&__pyx_n_s_stacklevel, __pyx_k_stacklevel, sizeof(__pyx_k_stacklevel), 0, 0, 1, 1},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_kp_u_start_init_can_only_be_called, __pyx_k_start_init_can_only_be_called, sizeof(__pyx_k_start_init_can_only_be_called), 0, 1, 0, 0},
    {&__pyx_n_s_start_new_session, __pyx_k_start_new_session, sizeof(__pyx_k_start_new_session), 0, 0, 1, 1},
    {&__pyx_n_s_start_serving, __pyx_k_start_serving, sizeof(__pyx_k_start_serving), 0, 0, 1, 1},
    {&__pyx_n_s_start_tls, __pyx_k_start_tls, sizeof(__pyx_k_start_tls), 0, 0, 1, 1},
    {&__pyx_n_s_startupinfo, __pyx_k_startupinfo, sizeof(__pyx_k_startupinfo), 0, 0, 1, 1},
    {&__pyx_kp_u_startupinfo_is_not_supported, __pyx_k_startupinfo_is_not_supported, sizeof(__pyx_k_startupinfo_is_not_supported), 0, 1, 0, 0},
    {&__pyx_n_s_stat, __pyx_k_stat, sizeof(__pyx_k_stat), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_stderr, __pyx_k_stderr, sizeof(__pyx_k_stderr), 0, 0, 1, 1},
    {&__pyx_n_s_stdin, __pyx_k_stdin, sizeof(__pyx_k_stdin), 0, 0, 1, 1},
    {&__pyx_n_s_stdio_fut, __pyx_k_stdio_fut, sizeof(__pyx_k_stdio_fut), 0, 0, 1, 1},
    {&__pyx_n_s_stdio_inited, __pyx_k_stdio_inited, sizeof(__pyx_k_stdio_inited), 0, 0, 1, 1},
    {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
    {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_kp_u_stream_is_open_after_UVStream__t, __pyx_k_stream_is_open_after_UVStream__t, sizeof(__pyx_k_stream_is_open_after_UVStream__t), 0, 1, 0, 0},
    {&__pyx_n_s_strerror, __pyx_k_strerror, sizeof(__pyx_k_strerror), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_submit, __pyx_k_submit, sizeof(__pyx_k_submit), 0, 0, 1, 1},
    {&__pyx_n_s_subprocess, __pyx_k_subprocess, sizeof(__pyx_k_subprocess), 0, 0, 1, 1},
    {&__pyx_kp_u_subprocess_STDOUT_is_supported_o, __pyx_k_subprocess_STDOUT_is_supported_o, sizeof(__pyx_k_subprocess_STDOUT_is_supported_o), 0, 1, 0, 0},
    {&__pyx_n_s_subprocess_exec, __pyx_k_subprocess_exec, sizeof(__pyx_k_subprocess_exec), 0, 0, 1, 1},
    {&__pyx_n_s_subprocess_run, __pyx_k_subprocess_run, sizeof(__pyx_k_subprocess_run), 0, 0, 1, 1},
    {&__pyx_n_s_subprocess_shell, __pyx_k_subprocess_shell, sizeof(__pyx_k_subprocess_shell), 0, 0, 1, 1},
    {&__pyx_n_s_subtract, __pyx_k_subtract, sizeof(__pyx_k_subtract), 0, 0, 1, 1},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
    {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
    {&__pyx_n_s_task, __pyx_k_task, sizeof(__pyx_k_task), 0, 0, 1, 1},
    {&__pyx_n_u_task, __pyx_k_task, sizeof(__pyx_k_task), 0, 1, 0, 1},
    {&__pyx_kp_u_task_factory_must_be_a_callable, __pyx_k_task_factory_must_be_a_callable, sizeof(__pyx_k_task_factory_must_be_a_callable), 0, 1, 0, 0},
    {&__pyx_n_s_tb, __pyx_k_tb, sizeof(__pyx_k_tb), 0, 0, 1, 1},
    {&__pyx_n_s_tcp, __pyx_k_tcp, sizeof(__pyx_k_tcp), 0, 0, 1, 1},
    {&__pyx_n_u_tcp, __pyx_k_tcp, sizeof(__pyx_k_tcp), 0, 1, 0, 1},
    {&__pyx_n_s_terminate, __pyx_k_terminate, sizeof(__pyx_k_terminate), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_test__append_write_backlog, __pyx_k_test__append_write_backlog, sizeof(__pyx_k_test__append_write_backlog), 0, 0, 1, 1},
    {&__pyx_n_s_test_coroutine_1, __pyx_k_test_coroutine_1, sizeof(__pyx_k_test_coroutine_1), 0, 0, 1, 1},
    {&__pyx_n_s_testhelper_unwrap_capsuled_poin, __pyx_k_testhelper_unwrap_capsuled_poin, sizeof(__pyx_k_testhelper_unwrap_capsuled_poin), 0, 0, 1, 1},
    {&__pyx_kp_u_the_socket_must_be_non_blocking, __pyx_k_the_socket_must_be_non_blocking, sizeof(__pyx_k_the_socket_must_be_non_blocking), 0, 1, 0, 0},
    {&__pyx_kp_u_this_event_loop_is_already_runni, __pyx_k_this_event_loop_is_already_runni, sizeof(__pyx_k_this_event_loop_is_already_runni), 0, 1, 0, 0},
    {&__pyx_n_s_thread, __pyx_k_thread, sizeof(__pyx_k_thread), 0, 0, 1, 1},
    {&__pyx_n_s_threading, __pyx_k_threading, sizeof(__pyx_k_threading), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_s_timeout, __pyx_k_timeout, sizeof(__pyx_k_timeout), 0, 0, 1, 1},
    {&__pyx_n_s_tr, __pyx_k_tr, sizeof(__pyx_k_tr), 0, 0, 1, 1},
    {&__pyx_n_s_traceback, __pyx_k_traceback, sizeof(__pyx_k_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_traceback_2, __pyx_k_traceback_2, sizeof(__pyx_k_traceback_2), 0, 0, 1, 1},
    {&__pyx_n_s_transp, __pyx_k_transp, sizeof(__pyx_k_transp), 0, 0, 1, 1},
    {&__pyx_n_s_transport, __pyx_k_transport, sizeof(__pyx_k_transport), 0, 0, 1, 1},
    {&__pyx_n_u_transport, __pyx_k_transport, sizeof(__pyx_k_transport), 0, 1, 0, 1},
    {&__pyx_kp_u_transport_2, __pyx_k_transport_2, sizeof(__pyx_k_transport_2), 0, 1, 0, 0},
    {&__pyx_kp_u_transport_sockets_do_not_support, __pyx_k_transport_sockets_do_not_support, sizeof(__pyx_k_transport_sockets_do_not_support), 0, 1, 0, 0},
    {&__pyx_n_s_transport_waiter, __pyx_k_transport_waiter, sizeof(__pyx_k_transport_waiter), 0, 0, 1, 1},
    {&__pyx_n_s_transports, __pyx_k_transports, sizeof(__pyx_k_transports), 0, 0, 1, 1},
    {&__pyx_kp_u_try_write_sent_all_data_and_ret, __pyx_k_try_write_sent_all_data_and_ret, sizeof(__pyx_k_try_write_sent_all_data_and_ret), 0, 1, 0, 0},
    {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
    {&__pyx_n_s_udp, __pyx_k_udp, sizeof(__pyx_k_udp), 0, 0, 1, 1},
    {&__pyx_n_u_udp, __pyx_k_udp, sizeof(__pyx_k_udp), 0, 1, 0, 1},
    {&__pyx_kp_u_unable_to_listen_no_protocol_fac, __pyx_k_unable_to_listen_no_protocol_fac, sizeof(__pyx_k_unable_to_listen_no_protocol_fac), 0, 1, 0, 0},
    {&__pyx_kp_u_unable_to_perform_operation_on_r, __pyx_k_unable_to_perform_operation_on_r, sizeof(__pyx_k_unable_to_perform_operation_on_r), 0, 1, 0, 0},
    {&__pyx_kp_u_unable_to_start_the_loop_it_was, __pyx_k_unable_to_start_the_loop_it_was, sizeof(__pyx_k_unable_to_start_the_loop_it_was), 0, 1, 0, 0},
    {&__pyx_kp_u_unclosed_resource_r, __pyx_k_unclosed_resource_r, sizeof(__pyx_k_unclosed_resource_r), 0, 1, 0, 0},
    {&__pyx_kp_u_unclosed_resource_r_2, __pyx_k_unclosed_resource_r_2, sizeof(__pyx_k_unclosed_resource_r_2), 0, 1, 0, 0},
    {&__pyx_kp_u_unclosed_resource_r_could_not_se, __pyx_k_unclosed_resource_r_could_not_se, sizeof(__pyx_k_unclosed_resource_r_could_not_se), 0, 1, 0, 0},
    {&__pyx_kp_u_unclosed_transport_uvloop_loop, __pyx_k_unclosed_transport_uvloop_loop, sizeof(__pyx_k_unclosed_transport_uvloop_loop), 0, 1, 0, 0},
    {&__pyx_kp_u_unexpected_address_family, __pyx_k_unexpected_address_family, sizeof(__pyx_k_unexpected_address_family), 0, 1, 0, 0},
    {&__pyx_kp_u_unexpected_kwargs, __pyx_k_unexpected_kwargs, sizeof(__pyx_k_unexpected_kwargs), 0, 1, 0, 0},
    {&__pyx_kp_u_unhandled_error_or_an_empty_buff, __pyx_k_unhandled_error_or_an_empty_buff, sizeof(__pyx_k_unhandled_error_or_an_empty_buff), 0, 1, 0, 0},
    {&__pyx_n_s_universal_newlines, __pyx_k_universal_newlines, sizeof(__pyx_k_universal_newlines), 0, 0, 1, 1},
    {&__pyx_kp_u_universal_newlines_must_be_False, __pyx_k_universal_newlines_must_be_False, sizeof(__pyx_k_universal_newlines_must_be_False), 0, 1, 0, 0},
    {&__pyx_kp_u_unix_socket_path, __pyx_k_unix_socket_path, sizeof(__pyx_k_unix_socket_path), 0, 1, 0, 0},
    {&__pyx_n_s_unlink, __pyx_k_unlink, sizeof(__pyx_k_unlink), 0, 0, 1, 1},
    {&__pyx_kp_u_unopened_TCPServer, __pyx_k_unopened_TCPServer, sizeof(__pyx_k_unopened_TCPServer), 0, 1, 0, 0},
    {&__pyx_n_s_unset, __pyx_k_unset, sizeof(__pyx_k_unset), 0, 0, 1, 1},
    {&__pyx_n_s_unwrap, __pyx_k_unwrap, sizeof(__pyx_k_unwrap), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_kp_u_uv_close_handle_cb_handle_loop, __pyx_k_uv_close_handle_cb_handle_loop, sizeof(__pyx_k_uv_close_handle_cb_handle_loop), 0, 1, 0, 0},
    {&__pyx_kp_u_uv_handle_t_current_freed_total, __pyx_k_uv_handle_t_current_freed_total, sizeof(__pyx_k_uv_handle_t_current_freed_total), 0, 1, 0, 0},
    {&__pyx_kp_u_uvloop_PseudoSocket_fd_family_s, __pyx_k_uvloop_PseudoSocket_fd_family_s, sizeof(__pyx_k_uvloop_PseudoSocket_fd_family_s), 0, 1, 0, 0},
    {&__pyx_kp_s_uvloop_cbhandles_pyx, __pyx_k_uvloop_cbhandles_pyx, sizeof(__pyx_k_uvloop_cbhandles_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_k_uvloop_handles_basetransport_pyx, sizeof(__pyx_k_uvloop_handles_basetransport_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_fsevent_pyx, __pyx_k_uvloop_handles_fsevent_pyx, sizeof(__pyx_k_uvloop_handles_fsevent_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_k_uvloop_handles_pipe_pyx, sizeof(__pyx_k_uvloop_handles_pipe_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_process_pyx, __pyx_k_uvloop_handles_process_pyx, sizeof(__pyx_k_uvloop_handles_process_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_stream_pyx, __pyx_k_uvloop_handles_stream_pyx, sizeof(__pyx_k_uvloop_handles_stream_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_streamserver_pyx, __pyx_k_uvloop_handles_streamserver_pyx, sizeof(__pyx_k_uvloop_handles_streamserver_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_tcp_pyx, __pyx_k_uvloop_handles_tcp_pyx, sizeof(__pyx_k_uvloop_handles_tcp_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_handles_udp_pyx, __pyx_k_uvloop_handles_udp_pyx, sizeof(__pyx_k_uvloop_handles_udp_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_uvloop_loop, __pyx_k_uvloop_loop, sizeof(__pyx_k_uvloop_loop), 0, 0, 1, 1},
    {&__pyx_kp_s_uvloop_loop_pyx, __pyx_k_uvloop_loop_pyx, sizeof(__pyx_k_uvloop_loop_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_pseudosock_pyx, __pyx_k_uvloop_pseudosock_pyx, sizeof(__pyx_k_uvloop_pseudosock_pyx), 0, 0, 1, 0},
    {&__pyx_kp_s_uvloop_server_pyx, __pyx_k_uvloop_server_pyx, sizeof(__pyx_k_uvloop_server_pyx), 0, 0, 1, 0},
    {&__pyx_n_u_uvloop_sleep_after_fork, __pyx_k_uvloop_sleep_after_fork, sizeof(__pyx_k_uvloop_sleep_after_fork), 0, 1, 0, 1},
    {&__pyx_kp_u_uvloop_sslproto, __pyx_k_uvloop_sslproto, sizeof(__pyx_k_uvloop_sslproto), 0, 1, 0, 0},
    {&__pyx_kp_s_uvloop_sslproto_pyx, __pyx_k_uvloop_sslproto_pyx, sizeof(__pyx_k_uvloop_sslproto_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
    {&__pyx_n_s_version_info, __pyx_k_version_info, sizeof(__pyx_k_version_info), 0, 0, 1, 1},
    {&__pyx_n_s_wait, __pyx_k_wait, sizeof(__pyx_k_wait), 0, 0, 1, 1},
    {&__pyx_n_s_wait_2, __pyx_k_wait_2, sizeof(__pyx_k_wait_2), 0, 0, 1, 1},
    {&__pyx_n_s_wait_closed, __pyx_k_wait_closed, sizeof(__pyx_k_wait_closed), 0, 0, 1, 1},
    {&__pyx_n_s_waiter, __pyx_k_waiter, sizeof(__pyx_k_waiter), 0, 0, 1, 1},
    {&__pyx_n_s_walk_stack, __pyx_k_walk_stack, sizeof(__pyx_k_walk_stack), 0, 0, 1, 1},
    {&__pyx_n_s_want, __pyx_k_want, sizeof(__pyx_k_want), 0, 0, 1, 1},
    {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
    {&__pyx_n_s_warn_on_full_buffer, __pyx_k_warn_on_full_buffer, sizeof(__pyx_k_warn_on_full_buffer), 0, 0, 1, 1},
    {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
    {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
    {&__pyx_n_u_wb, __pyx_k_wb, sizeof(__pyx_k_wb), 0, 1, 0, 1},
    {&__pyx_n_s_weakref, __pyx_k_weakref, sizeof(__pyx_k_weakref), 0, 0, 1, 1},
    {&__pyx_n_s_when, __pyx_k_when, sizeof(__pyx_k_when), 0, 0, 1, 1},
    {&__pyx_kp_u_without_no_gc_clear_loop_was_se, __pyx_k_without_no_gc_clear_loop_was_se, sizeof(__pyx_k_without_no_gc_clear_loop_was_se), 0, 1, 0, 0},
    {&__pyx_n_s_wrap, __pyx_k_wrap, sizeof(__pyx_k_wrap), 0, 0, 1, 1},
    {&__pyx_n_s_wrap_bio, __pyx_k_wrap_bio, sizeof(__pyx_k_wrap_bio), 0, 0, 1, 1},
    {&__pyx_n_s_wrap_future, __pyx_k_wrap_future, sizeof(__pyx_k_wrap_future), 0, 0, 1, 1},
    {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
    {&__pyx_n_s_write_eof, __pyx_k_write_eof, sizeof(__pyx_k_write_eof), 0, 0, 1, 1},
    {&__pyx_n_s_writelines, __pyx_k_writelines, sizeof(__pyx_k_writelines), 0, 0, 1, 1},
    {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 235, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 429, __pyx_L1_error)
  __pyx_builtin_BaseException = __Pyx_GetBuiltinName(__pyx_n_s_BaseException); if (!__pyx_builtin_BaseException) __PYX_ERR(0, 461, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 697, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 700, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 749, __pyx_L1_error)
  __pyx_builtin_KeyboardInterrupt = __Pyx_GetBuiltinName(__pyx_n_s_KeyboardInterrupt); if (!__pyx_builtin_KeyboardInterrupt) __PYX_ERR(0, 921, __pyx_L1_error)
  __pyx_builtin_SystemExit = __Pyx_GetBuiltinName(__pyx_n_s_SystemExit); if (!__pyx_builtin_SystemExit) __PYX_ERR(0, 921, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 1167, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 1326, __pyx_L1_error)
  __pyx_builtin_OverflowError = __Pyx_GetBuiltinName(__pyx_n_s_OverflowError); if (!__pyx_builtin_OverflowError) __PYX_ERR(0, 1550, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 2639, __pyx_L1_error)
  __pyx_builtin_RuntimeWarning = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeWarning); if (!__pyx_builtin_RuntimeWarning) __PYX_ERR(0, 2927, __pyx_L1_error)
  __pyx_builtin_DeprecationWarning = __Pyx_GetBuiltinName(__pyx_n_s_DeprecationWarning); if (!__pyx_builtin_DeprecationWarning) __PYX_ERR(0, 3067, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 3159, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 3214, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(2, 218, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(3, 166, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(4, 193, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(4, 232, __pyx_L1_error)
  __pyx_builtin_UnicodeEncodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeEncodeError); if (!__pyx_builtin_UnicodeEncodeError) __PYX_ERR(5, 109, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "cfunc.to_py":67
 * @cname("__Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled")
 * cdef object __Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled(object (*f)(Loop, bint) ):
 *     def wrap(Loop self, bint enabled):             # <<<<<<<<<<<<<<
 *         """wrap(self: 'Loop', enabled: bool)"""
 *         return f(self, enabled)
 */
  __pyx_tuple_ = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_enabled); if (unlikely(!__pyx_tuple_)) __PYX_ERR(6, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple_, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 67, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(6, 67, __pyx_L1_error)

  /* "uvloop/loop.pyx":235
 *     def __dealloc__(self):
 *         if self._running == 1:
 *             raise RuntimeError('deallocating a running event loop!')             # <<<<<<<<<<<<<<
 *         if self._closed == 0:
 *             aio_logger.error("deallocating an open event loop")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_deallocating_a_running_event_loo); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "uvloop/loop.pyx":292
 * 
 *         if self._listening_signals:
 *             raise RuntimeError('signals handling has been already setup')             # <<<<<<<<<<<<<<
 * 
 *         if self._ssock is not None:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_signals_handling_has_been_alread); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "uvloop/loop.pyx":295
 * 
 *         if self._ssock is not None:
 *             raise RuntimeError('self-pipe exists before loop run')             # <<<<<<<<<<<<<<
 * 
 *         # Create a self-pipe and call set_signal_wakeup_fd() with one
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_self_pipe_exists_before_loop_run); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "uvloop/loop.pyx":336
 *         if not self._is_main_thread():
 *             if self._listening_signals:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     'cannot pause signals handling; no longer running in '
 *                     'the main thread')
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_cannot_pause_signals_handling_no); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "uvloop/loop.pyx":343
 * 
 *         if not self._listening_signals:
 *             raise RuntimeError('signals handling has not been setup')             # <<<<<<<<<<<<<<
 * 
 *         self._listening_signals = False
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_signals_handling_has_not_been_se); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "uvloop/loop.pyx":364
 * 
 *         if self._listening_signals:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'cannot shutdown signals handling as it has not been paused')
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_cannot_shutdown_signals_handling); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "uvloop/loop.pyx":368
 * 
 *         if self._ssock:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'self-pipe was not cleaned up after loop was run')
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_self_pipe_was_not_cleaned_up_aft); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "uvloop/loop.pyx":517
 * 
 *         if self._closed == 1:
 *             raise RuntimeError('unable to start the loop; it was closed')             # <<<<<<<<<<<<<<
 * 
 *         if self._running == 1:
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_unable_to_start_the_loop_it_was); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "uvloop/loop.pyx":520
 * 
 *         if self._running == 1:
 *             raise RuntimeError('this event loop is already running.')             # <<<<<<<<<<<<<<
 * 
 *         if (aio_get_running_loop is not None and
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_this_event_loop_is_already_runni); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "uvloop/loop.pyx":524
 *         if (aio_get_running_loop is not None and
 *                 aio_get_running_loop() is not None):
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'Cannot run the event loop while another loop is running')
 * 
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Cannot_run_the_event_loop_while); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "uvloop/loop.pyx":563
 * 
 *         if self._running == 1:
 *             raise RuntimeError("Cannot close a running event loop")             # <<<<<<<<<<<<<<
 * 
 *         if self._closed == 1:
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_u_Cannot_close_a_running_event_loo); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "uvloop/loop.pyx":705
 *     cdef inline _check_closed(self):
 *         if self._closed == 1:
 *             raise RuntimeError('Event loop is closed')             # <<<<<<<<<<<<<<
 * 
 *     cdef inline _check_thread(self):
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_u_Event_loop_is_closed); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "uvloop/loop.pyx":715
 * 
 *         if thread_id != self._thread_id:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 "Non-thread-safe operation invoked on an event loop other "
 *                 "than the current one")
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_u_Non_thread_safe_operation_invoke); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "uvloop/loop.pyx":904
 *             port = str(port).encode()
 *         if port is not None and not isinstance(port, bytes):
 *             raise TypeError('port must be a str, bytes or int')             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(host, str):
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_port_must_be_a_str_bytes_or_int); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "uvloop/loop.pyx":910
 *         if host is not None:
 *             if not isinstance(host, bytes):
 *                 raise TypeError('host must be a str or bytes')             # <<<<<<<<<<<<<<
 * 
 *         fut = self._new_future()
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_host_must_be_a_str_or_bytes); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "uvloop/loop.pyx":914
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if AddrInfo.isinstance(result):
 *                 try:
 */
  __pyx_tuple__19 = PyTuple_Pack(3, __pyx_n_s_result, __pyx_n_s_data, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_callback, 914, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 914, __pyx_L1_error)

  /* "uvloop/loop.pyx":940
 *         fut = self._new_future()
 * 
 *         def callback(result):             # <<<<<<<<<<<<<<
 *             if isinstance(result, tuple):
 *                 fut.set_result(result)
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_n_s_result); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_callback, 940, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 940, __pyx_L1_error)

  /* "uvloop/loop.pyx":954
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_recv is called on a cancelled Future')
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_u_sock_recv_is_called_on_a_cancel); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "uvloop/loop.pyx":981
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketReaderFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_recv_into is called on a cancelled Future')
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_u_sock_recv_into_is_called_on_a_c); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "uvloop/loop.pyx":1012
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_sendall is called on a cancelled Future')
 * 
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_u_sock_sendall_is_called_on_a_can); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "uvloop/loop.pyx":1096
 *             if fut.cancelled():
 *                 # Shouldn't happen with _SyncSocketWriterFuture.
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f'_sock_connect_cb is called on a cancelled Future')
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_sock_connect_cb_is_called_on_a); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "uvloop/loop.pyx":1167
 *         # OS
 * 
 *         print('---- Process info: -----')             # <<<<<<<<<<<<<<
 *         print('Process memory:            {}'.format(rusage.ru_maxrss))
 *         print('Number of signals:         {}'.format(rusage.ru_nsignals))
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_u_Process_info); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 1167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "uvloop/loop.pyx":1170
 *         print('Process memory:            {}'.format(rusage.ru_maxrss))
 *         print('Number of signals:         {}'.format(rusage.ru_nsignals))
 *         print('')             # <<<<<<<<<<<<<<
 * 
 *         # Loop
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u__10); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "uvloop/loop.pyx":1174
 *         # Loop
 * 
 *         print('--- Loop debug info: ---')             # <<<<<<<<<<<<<<
 *         print('Loop time:                 {}'.format(self.time()))
 *         print('Errors logged:             {}'.format(
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_u_Loop_debug_info); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "uvloop/loop.pyx":1187
 *         print()
 * 
 *         print('                        alive  | closed  |')             # <<<<<<<<<<<<<<
 *         print('UVHandles               python | libuv   | total')
 *         print('                        objs   | handles |')
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_u_alive_closed); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 1187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "uvloop/loop.pyx":1188
 * 
 *         print('                        alive  | closed  |')
 *         print('UVHandles               python | libuv   | total')             # <<<<<<<<<<<<<<
 *         print('                        objs   | handles |')
 *         print('-------------------------------+---------+---------')
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_u_UVHandles_python_libuv_total); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "uvloop/loop.pyx":1189
 *         print('                        alive  | closed  |')
 *         print('UVHandles               python | libuv   | total')
 *         print('                        objs   | handles |')             # <<<<<<<<<<<<<<
 *         print('-------------------------------+---------+---------')
 *         for name in sorted(self._debug_handles_total):
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_u_objs_handles); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "uvloop/loop.pyx":1190
 *         print('UVHandles               python | libuv   | total')
 *         print('                        objs   | handles |')
 *         print('-------------------------------+---------+---------')             # <<<<<<<<<<<<<<
 *         for name in sorted(self._debug_handles_total):
 *             print('    {: <18} {: >7} | {: >7} | {: >7}'.format(
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_u__33); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "uvloop/loop.pyx":1205
 *         print()
 * 
 *         print('--- Streams debug info: ---')             # <<<<<<<<<<<<<<
 *         print('Write errors:              {}'.format(
 *             self._debug_stream_write_errors_total))
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_u_Streams_debug_info); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "uvloop/loop.pyx":1233
 *         print()
 * 
 *         print('--- Polls debug info: ---')             # <<<<<<<<<<<<<<
 *         print('Read events:               {}'.format(
 *             self._poll_read_events_total))
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_u_Polls_debug_info); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "uvloop/loop.pyx":1244
 *         print()
 * 
 *         print('--- Sock ops successful on 1st try: ---')             # <<<<<<<<<<<<<<
 *         print('Socket try-writes:         {}'.format(
 *             self._sock_try_write_total))
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_u_Sock_ops_successful_on_1st_try); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "uvloop/loop.pyx":1255
 *                 return lambda: self._print_debug_info()
 *             else:
 *                 raise AttributeError('print_debug_info')             # <<<<<<<<<<<<<<
 * 
 *     # Public API
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_n_u_print_debug_info); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "uvloop/loop.pyx":1467
 *         """
 *         if factory is not None and not callable(factory):
 *             raise TypeError('task factory must be a callable or None')             # <<<<<<<<<<<<<<
 *         self._task_factory = factory
 * 
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_u_task_factory_must_be_a_callable); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "uvloop/loop.pyx":1494
 *             future._log_destroy_pending = False
 * 
 *         def done_cb(fut):             # <<<<<<<<<<<<<<
 *             if not fut.cancelled():
 *                 exc = fut.exception()
 */
  __pyx_tuple__40 = PyTuple_Pack(2, __pyx_n_s_fut, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_done_cb, 1494, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 1494, __pyx_L1_error)

  /* "uvloop/loop.pyx":1516
 *             future.remove_done_callback(done_cb)
 *         if not future.done():
 *             raise RuntimeError('Event loop stopped before Future completed.')             # <<<<<<<<<<<<<<
 * 
 *         return future.result()
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_u_Event_loop_stopped_before_Future); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "uvloop/loop.pyx":1540
 * 
 *         if not isinstance(sockaddr, tuple):
 *             raise TypeError('getnameinfo() argument 1 must be a tuple')             # <<<<<<<<<<<<<<
 * 
 *         sl = len(sockaddr)
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_u_getnameinfo_argument_1_must_be_a); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "uvloop/loop.pyx":1545
 * 
 *         if sl < 2 or sl > 4:
 *             raise ValueError('sockaddr must be a tuple of 2, 3 or 4 values')             # <<<<<<<<<<<<<<
 * 
 *         if sl > 2:
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_u_sockaddr_must_be_a_tuple_of_2_3); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "uvloop/loop.pyx":1550
 *             flowinfo = sockaddr[2]
 *             if flowinfo < 0 or flowinfo > 0xfffff:
 *                 raise OverflowError(             # <<<<<<<<<<<<<<
 *                     'getnameinfo(): flowinfo must be 0-1048575.')
 *         else:
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_u_getnameinfo_flowinfo_must_be_0_1); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "uvloop/loop.pyx":1558
 *             scope_id = sockaddr[3]
 *             if scope_id < 0 or scope_id > 2 ** 32:
 *                 raise OverflowError(             # <<<<<<<<<<<<<<
 *                     'getsockaddrarg: scope_id must be unsigned 32 bit integer')
 *         else:
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_u_getsockaddrarg_scope_id_must_be); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 1558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "uvloop/loop.pyx":1574
 * 
 *         if ai.ai_next:
 *             raise OSError("sockaddr resolved to multiple addresses")             # <<<<<<<<<<<<<<
 * 
 *         if ai.ai_family == uv.AF_INET:
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_kp_u_sockaddr_resolved_to_multiple_ad); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "uvloop/loop.pyx":1578
 *         if ai.ai_family == uv.AF_INET:
 *             if sl > 2:
 *                 raise OSError("IPv4 sockaddr must be 2 tuple")             # <<<<<<<<<<<<<<
 *         elif ai.ai_family == uv.AF_INET6:
 *             # Modify some fields in `ai`
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_u_IPv4_sockaddr_must_be_2_tuple); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "uvloop/loop.pyx":1706
 *         if sock is not None and sock.family == uv.AF_UNIX:
 *             if host is not None or port is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'host/port and sock can not be specified at the same time')
 *             return await self.create_unix_server(
 */
  __pyx_tuple__55 = PyTuple_Pack(1, __pyx_kp_u_host_port_and_sock_can_not_be_sp); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "uvloop/loop.pyx":1719
 *         if ssl is not None:
 *             if not isinstance(ssl, ssl_SSLContext):
 *                 raise TypeError('ssl argument must be an SSLContext or None')             # <<<<<<<<<<<<<<
 *         else:
 *             if ssl_handshake_timeout is not None:
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_u_ssl_argument_must_be_an_SSLConte); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 1719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "uvloop/loop.pyx":1722
 *         else:
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_u_ssl_handshake_timeout_is_only_me); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 1722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "uvloop/loop.pyx":1725
 *                     'ssl_handshake_timeout is only meaningful with ssl')
 *             if ssl_shutdown_timeout is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'ssl_shutdown_timeout is only meaningful with ssl')
 * 
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_u_ssl_shutdown_timeout_is_only_mea); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 1725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "uvloop/loop.pyx":1737
 *             reuse_port = bool(reuse_port)
 *             if reuse_port and not has_SO_REUSEPORT:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'reuse_port not supported by socket module')
 * 
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_u_reuse_port_not_supported_by_sock); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 1737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "uvloop/loop.pyx":1760
 *                     while addrinfo != NULL:
 *                         if addrinfo.ai_family == uv.AF_UNSPEC:
 *                             raise RuntimeError('AF_UNSPEC in DNS results')             # <<<<<<<<<<<<<<
 * 
 *                         try:
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_u_AF_UNSPEC_in_DNS_results); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 1760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "uvloop/loop.pyx":1826
 *         else:
 *             if sock is None:
 *                 raise ValueError('Neither host/port nor sock were specified')             # <<<<<<<<<<<<<<
 *             if not _is_sock_stream(sock.type):
 *                 raise ValueError(
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_kp_u_Neither_host_port_nor_sock_were); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "uvloop/loop.pyx":1911
 *             if server_hostname is None:
 *                 if not host:
 *                     raise ValueError('You must set server_hostname '             # <<<<<<<<<<<<<<
 *                                      'when using ssl without a host')
 *                 server_hostname = host
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_kp_u_You_must_set_server_hostname_whe); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 1911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "uvloop/loop.pyx":1924
 *         else:
 *             if server_hostname is not None:
 *                 raise ValueError('server_hostname is only meaningful with ssl')             # <<<<<<<<<<<<<<
 *             if ssl_handshake_timeout is not None:
 *                 raise ValueError(
 */
  __pyx_tuple__65 = PyTuple_Pack(1, __pyx_kp_u_server_hostname_is_only_meaningf); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 1924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "uvloop/loop.pyx":1959
 *                 if not isinstance(local_addr, (tuple, list)) or \
 *                         len(local_addr) != 2:
 *                     raise ValueError(             # <<<<<<<<<<<<<<
 *                         'local_addr must be a tuple of host and port')
 * 
 */
  __pyx_tuple__66 = PyTuple_Pack(1, __pyx_kp_u_local_addr_must_be_a_tuple_of_ho); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 1959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "uvloop/loop.pyx":1984
 *                 ai_remote = f1.result()
 *                 if ai_remote.data is NULL:
 *                     raise OSError('getaddrinfo() returned empty list')             # <<<<<<<<<<<<<<
 *                 rai = ai_remote.data
 * 
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_kp_u_getaddrinfo_returned_empty_list); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 1984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "uvloop/loop.pyx":1990
 *                 ai_local = f2.result()
 *                 if ai_local.data is NULL:
 *                     raise OSError(             # <<<<<<<<<<<<<<
 *                         'getaddrinfo() returned empty list for local_addr')
 *                 lai = ai_local.data
 */
  __pyx_tuple__68 = PyTuple_Pack(1, __pyx_kp_u_getaddrinfo_returned_empty_list_2); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "uvloop/loop.pyx":2050
 *         else:
 *             if sock is None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'host and port was not specified and no sock specified')
 *             if not _is_sock_stream(sock.type):
 */
  __pyx_tuple__70 = PyTuple_Pack(1, __pyx_kp_u_host_and_port_was_not_specified); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 2050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "uvloop/loop.pyx":2146
 *         if path is not None:
 *             if sock is not None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'path and sock can not be specified at the same time')
 *             orig_path = path
 */
  __pyx_tuple__73 = PyTuple_Pack(1, __pyx_kp_u_path_and_sock_can_not_be_specifi); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 2146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "uvloop/loop.pyx":2193
 *         else:
 *             if sock is None:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'path was not specified, and no sock specified')
 * 
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_kp_u_path_was_not_specified_and_no_so); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 2193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "uvloop/loop.pyx":2297
 *         else:
 *             if sock is None:
 *                 raise ValueError('no path and sock were specified')             # <<<<<<<<<<<<<<
 * 
 *             if sock.family != uv.AF_UNIX or not _is_sock_stream(sock.type):
 */
  __pyx_tuple__77 = PyTuple_Pack(1, __pyx_kp_u_no_path_and_sock_were_specified); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 2297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "uvloop/loop.pyx":2429
 *                 # in the default implementation, as well as for subclassed
 *                 # event loops with overloaded "default_exception_handler".
 *                 aio_logger.error('Exception in default exception handler',             # <<<<<<<<<<<<<<
 *                                  exc_info=True)
 *         else:
 */
  __pyx_tuple__80 = PyTuple_Pack(1, __pyx_kp_u_Exception_in_default_exception_h); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 2429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "uvloop/loop.pyx":2450
 *                     # Guard 'default_exception_handler' in case it is
 *                     # overloaded.
 *                     aio_logger.error('Exception in default exception handler '             # <<<<<<<<<<<<<<
 *                                      'while handling an unexpected error '
 *                                      'in custom exception handler',
 */
  __pyx_tuple__81 = PyTuple_Pack(1, __pyx_kp_u_Exception_in_default_exception_h_2); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 2450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "uvloop/loop.pyx":2489
 * 
 *         if self._debug and sock.gettimeout() != 0:
 *             raise ValueError("the socket must be non-blocking")             # <<<<<<<<<<<<<<
 * 
 *         fut = _SyncSocketReaderFuture(sock, self)
 */
  __pyx_tuple__83 = PyTuple_Pack(1, __pyx_kp_u_the_socket_must_be_non_blocking); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "uvloop/loop.pyx":2628
 *             else:
 *                 addrs = await self.getaddrinfo(
 *                     *address[:2], family=sock.family)             # <<<<<<<<<<<<<<
 * 
 *                 _, _, _, _, address = addrs[0]
 */
  __pyx_slice__88 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__88)) __PYX_ERR(0, 2628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__88);
  __Pyx_GIVEREF(__pyx_slice__88);

  /* "uvloop/loop.pyx":2706
 * 
 *         if transport is None:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'invalid socket family, expected AF_UNIX, AF_INET or AF_INET6')
 * 
 */
  __pyx_tuple__93 = PyTuple_Pack(1, __pyx_kp_u_invalid_socket_family_expected_A); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 2706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "uvloop/loop.pyx":2735
 *     def run_in_executor(self, executor, func, *args):
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):
 *             raise TypeError("coroutines cannot be used with run_in_executor()")             # <<<<<<<<<<<<<<
 * 
 *         self._check_closed()
 */
  __pyx_tuple__94 = PyTuple_Pack(1, __pyx_kp_u_coroutines_cannot_be_used_with_r); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 2735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);

  /* "uvloop/loop.pyx":2779
 * 
 *         if universal_newlines:
 *             raise ValueError("universal_newlines must be False")             # <<<<<<<<<<<<<<
 *         if bufsize != 0:
 *             raise ValueError("bufsize must be 0")
 */
  __pyx_tuple__99 = PyTuple_Pack(1, __pyx_kp_u_universal_newlines_must_be_False); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 2779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);

  /* "uvloop/loop.pyx":2781
 *             raise ValueError("universal_newlines must be False")
 *         if bufsize != 0:
 *             raise ValueError("bufsize must be 0")             # <<<<<<<<<<<<<<
 *         if startupinfo is not None:
 *             raise ValueError('startupinfo is not supported')
 */
  __pyx_tuple__100 = PyTuple_Pack(1, __pyx_kp_u_bufsize_must_be_0); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 2781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);

  /* "uvloop/loop.pyx":2783
 *             raise ValueError("bufsize must be 0")
 *         if startupinfo is not None:
 *             raise ValueError('startupinfo is not supported')             # <<<<<<<<<<<<<<
 *         if creationflags != 0:
 *             raise ValueError('creationflags is not supported')
 */
  __pyx_tuple__101 = PyTuple_Pack(1, __pyx_kp_u_startupinfo_is_not_supported); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 2783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);

  /* "uvloop/loop.pyx":2785
 *             raise ValueError('startupinfo is not supported')
 *         if creationflags != 0:
 *             raise ValueError('creationflags is not supported')             # <<<<<<<<<<<<<<
 * 
 *         if executable is not None:
 */
  __pyx_tuple__102 = PyTuple_Pack(1, __pyx_kp_u_creationflags_is_not_supported); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(0, 2785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);

  /* "uvloop/loop.pyx":2823
 * 
 *         if not shell:
 *             raise ValueError("shell must be True")             # <<<<<<<<<<<<<<
 * 
 *         args = [cmd]
 */
  __pyx_tuple__104 = PyTuple_Pack(1, __pyx_kp_u_shell_must_be_True); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(0, 2823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "uvloop/loop.pyx":2837
 * 
 *         if shell:
 *             raise ValueError("shell must be False")             # <<<<<<<<<<<<<<
 * 
 *         args = list((program,) + args)
 */
  __pyx_tuple__106 = PyTuple_Pack(1, __pyx_kp_u_shell_must_be_False); if (unlikely(!__pyx_tuple__106)) __PYX_ERR(0, 2837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "uvloop/loop.pyx":2908
 * 
 *         if not self._is_main_thread():
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'add_signal_handler() can only be called from '
 *                 'the main thread')
 */
  __pyx_tuple__109 = PyTuple_Pack(1, __pyx_kp_u_add_signal_handler_can_only_be_c); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(0, 2908, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);

  /* "uvloop/loop.pyx":2914
 *         if (aio_iscoroutine(callback)
 *                 or aio_iscoroutinefunction(callback)):
 *             raise TypeError(             # <<<<<<<<<<<<<<
 *                 "coroutines cannot be used with add_signal_handler()")
 * 
 */
  __pyx_tuple__110 = PyTuple_Pack(1, __pyx_kp_u_coroutines_cannot_be_used_with_a); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 2914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);

  /* "uvloop/loop.pyx":2921
 *                     isinstance(callback.__self__, aio_AbstractChildWatcher)):
 * 
 *                 warnings_warn(             # <<<<<<<<<<<<<<
 *                     "!!! asyncio is trying to install its ChildWatcher for "
 *                     "SIGCHLD signal !!!\n\nThis is probably because a uvloop "
 */
  __pyx_tuple__111 = PyTuple_Pack(2, __pyx_kp_u_asyncio_is_trying_to_install_it, __pyx_builtin_RuntimeWarning); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 2921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);

  /* "uvloop/loop.pyx":2936
 *                 return
 * 
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'cannot add a signal handler for SIGCHLD: it is used '
 *                 'by the event loop to track subprocesses')
 */
  __pyx_tuple__112 = PyTuple_Pack(1, __pyx_kp_u_cannot_add_a_signal_handler_for); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 2936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);

  /* "uvloop/loop.pyx":2973
 * 
 *         if not self._is_main_thread():
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'remove_signal_handler() can only be called from '
 *                 'the main thread')
 */
  __pyx_tuple__113 = PyTuple_Pack(1, __pyx_kp_u_remove_signal_handler_can_only_b); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(0, 2973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);

  /* "uvloop/loop.pyx":3061
 *             if reuse_address is not _unset:
 *                 if reuse_address:
 *                     raise ValueError("Passing `reuse_address=True` is no "             # <<<<<<<<<<<<<<
 *                                      "longer supported, as the usage of "
 *                                      "SO_REUSEPORT in UDP poses a significant "
 */
  __pyx_tuple__117 = PyTuple_Pack(1, __pyx_kp_u_Passing_reuse_address_True_is_no); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 3061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);

  /* "uvloop/loop.pyx":3066
 *                                      "security concern.")
 *                 else:
 *                     warnings_warn("The *reuse_address* parameter has been "             # <<<<<<<<<<<<<<
 *                                   "deprecated as of 0.15.", DeprecationWarning,
 *                                   stacklevel=2)
 */
  __pyx_tuple__118 = PyTuple_Pack(2, __pyx_kp_u_The_reuse_address_parameter_has, __pyx_builtin_DeprecationWarning); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 3066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);

  /* "uvloop/loop.pyx":3078
 *                 if (not isinstance(local_addr, (tuple, list)) or
 *                         len(local_addr) != 2):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'local_addr must be a tuple of (host, port)')
 *                 lads = await self._getaddrinfo(
 */
  __pyx_tuple__119 = PyTuple_Pack(1, __pyx_kp_u_local_addr_must_be_a_tuple_of_ho_2); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(0, 3078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);

  /* "uvloop/loop.pyx":3089
 *                 if (not isinstance(remote_addr, (tuple, list)) or
 *                         len(remote_addr) != 2):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'remote_addr must be a tuple of (host, port)')
 *                 rads = await self._getaddrinfo(
 */
  __pyx_tuple__120 = PyTuple_Pack(1, __pyx_kp_u_remote_addr_must_be_a_tuple_of_h); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 3089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);

  /* "uvloop/loop.pyx":3106
 *                 else:
 *                     if family not in (uv.AF_INET, uv.AF_INET6):
 *                         raise ValueError('unexpected address family')             # <<<<<<<<<<<<<<
 *                     udp = UDPTransport.__new__(UDPTransport)
 *                     udp._init(self, family)
 */
  __pyx_tuple__121 = PyTuple_Pack(1, __pyx_kp_u_unexpected_address_family); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(0, 3106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);

  /* "uvloop/loop.pyx":3182
 *     def _check_default_executor(self):
 *         if self._executor_shutdown_called:
 *             raise RuntimeError('Executor shutdown has been called')             # <<<<<<<<<<<<<<
 * 
 *     def _asyncgen_finalizer_hook(self, agen):
 */
  __pyx_tuple__122 = PyTuple_Pack(1, __pyx_kp_u_Executor_shutdown_has_been_calle); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(0, 3182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);

  /* "uvloop/loop.pyx":3282
 *     if loop._recv_buffer_in_use == 1:
 *         buf.len = 0
 *         exc = RuntimeError('concurrent allocations')             # <<<<<<<<<<<<<<
 *         loop._handle_exception(exc)
 *         return
 */
  __pyx_tuple__125 = PyTuple_Pack(1, __pyx_kp_u_concurrent_allocations); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 3282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);

  /* "uvloop/cbhandles.pyx":26
 *             self.loop._debug_cb_handles_count -= 1
 *         if self.loop is None:
 *             raise RuntimeError('Handle.loop is None in Handle.__dealloc__')             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
 */
  __pyx_tuple__126 = PyTuple_Pack(1, __pyx_kp_u_Handle_loop_is_None_in_Handle); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(10, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);

  /* "uvloop/cbhandles.pyx":55
 *                 callback = self.arg1
 *                 if callback is None:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         'cannot run Handle; callback is not set')
 * 
 */
  __pyx_tuple__127 = PyTuple_Pack(1, __pyx_kp_u_cannot_run_Handle_callback_is_no); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(10, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);

  /* "uvloop/cbhandles.pyx":210
 *             self.loop._debug_cb_timer_handles_count -= 1
 *         if self.timer is not None:
 *             raise RuntimeError('active TimerHandle is deallacating')             # <<<<<<<<<<<<<<
 * 
 *     cdef _cancel(self):
 */
  __pyx_tuple__131 = PyTuple_Pack(1, __pyx_kp_u_active_TimerHandle_is_deallacati); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(10, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "uvloop/cbhandles.pyx":235
 *             return
 *         if self.callback is None:
 *             raise RuntimeError('cannot run TimerHandle; callback is not set')             # <<<<<<<<<<<<<<
 * 
 *         callback = self.callback
 */
  __pyx_tuple__132 = PyTuple_Pack(1, __pyx_kp_u_cannot_run_TimerHandle_callback); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(10, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);

  /* "uvloop/pseudosock.pyx":64
 * 
 *     def __getstate__(self):
 *         raise TypeError("Cannot serialize socket object")             # <<<<<<<<<<<<<<
 * 
 *     def fileno(self):
 */
  __pyx_tuple__133 = PyTuple_Pack(1, __pyx_kp_u_Cannot_serialize_socket_object); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(11, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);

  /* "uvloop/pseudosock.pyx":193
 *         if value == 0:
 *             return
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             'settimeout(): only 0 timeout is allowed on transport sockets')
 * 
 */
  __pyx_tuple__134 = PyTuple_Pack(1, __pyx_kp_u_settimeout_only_0_timeout_is_all); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(11, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);

  /* "uvloop/pseudosock.pyx":202
 *         if not flag:
 *             return
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             'setblocking(): transport sockets cannot be blocking')
 * 
 */
  __pyx_tuple__135 = PyTuple_Pack(1, __pyx_kp_u_setblocking_transport_sockets_ca); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(11, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);

  /* "uvloop/handles/handle.pyx":311
 *     if UVLOOP_DEBUG:
 *         if handle.loop is NULL:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'handle.loop is NULL in __ensure_handle_data')
 * 
 */
  __pyx_tuple__136 = PyTuple_Pack(1, __pyx_kp_u_handle_loop_is_NULL_in___ensure); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(2, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);

  /* "uvloop/handles/handle.pyx":315
 * 
 *         if handle.loop.data is NULL:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'handle.loop.data is NULL in __ensure_handle_data')
 * 
 */
  __pyx_tuple__137 = PyTuple_Pack(1, __pyx_kp_u_handle_loop_data_is_NULL_in___en); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(2, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "uvloop/handles/handle.pyx":347
 *         if UVLOOP_DEBUG:
 *             if handle.loop == NULL or handle.loop.data == NULL:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     '__uv_close_handle_cb: handle.loop is invalid')
 *             (<Loop>handle.loop.data)._debug_uv_handles_freed += 1
 */
  __pyx_tuple__138 = PyTuple_Pack(1, __pyx_kp_u_uv_close_handle_cb_handle_loop); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(2, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);

  /* "uvloop/handles/handle.pyx":357
 *             if UVLOOP_DEBUG:
 *                 if not h._has_handle:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         'has_handle=0 in __uv_close_handle_cb')
 *                 h._loop._debug_handles_closed.update([
 */
  __pyx_tuple__139 = PyTuple_Pack(1, __pyx_kp_u_has_handle_0_in___uv_close_handl); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(2, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);

  /* "uvloop/handles/basetransport.pyx":114
 *                 if not self._is_alive():
 *                     self._waiter.set_exception(
 *                         RuntimeError(             # <<<<<<<<<<<<<<
 *                             'closed Transport handle and unset waiter'))
 *                 else:
 */
  __pyx_tuple__142 = PyTuple_Pack(1, __pyx_kp_u_closed_Transport_handle_and_unse); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(18, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);

  /* "uvloop/handles/basetransport.pyx":122
 *     cdef _call_connection_made(self):
 *         if self._protocol is None:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'protocol is not set, cannot call connection_made()')
 * 
 */
  __pyx_tuple__143 = PyTuple_Pack(1, __pyx_kp_u_protocol_is_not_set_cannot_call); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(18, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);

  /* "uvloop/handles/basetransport.pyx":206
 *         self._loop._track_transport(self)
 *         if self._protocol is None:
 *             raise RuntimeError('invalid _init_protocol call')             # <<<<<<<<<<<<<<
 *         self._schedule_call_connection_made()
 * 
 */
  __pyx_tuple__144 = PyTuple_Pack(1, __pyx_kp_u_invalid__init_protocol_call); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(18, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);

  /* "uvloop/handles/stream.pyx":41
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_StreamWriteContext.close: uv_bufs != NULL and '
 *                         'py_bufs_sml_inuse is True')
 */
  __pyx_tuple__145 = PyTuple_Pack(1, __pyx_kp_u_StreamWriteContext_close_uv_buf); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(20, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "uvloop/handles/stream.pyx":52
 *             if UVLOOP_DEBUG:
 *                 if self.py_bufs_sml_inuse:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_StreamWriteContext.close: py_bufs != NULL and '
 *                         'py_bufs_sml_inuse is True')
 */
  __pyx_tuple__146 = PyTuple_Pack(1, __pyx_kp_u_StreamWriteContext_close_py_buf); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(20, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);

  /* "uvloop/handles/stream.pyx":97
 *             if UVLOOP_DEBUG:
 *                 if sent < 0:
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')             # <<<<<<<<<<<<<<
 * 
 *         raise RuntimeError('fatal: Could not advance _StreamWriteContext')
 */
  __pyx_tuple__147 = PyTuple_Pack(1, __pyx_kp_u_fatal_sent_0_in_advance_uv_buf); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(20, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "uvloop/handles/stream.pyx":99
 *                     raise RuntimeError('fatal: sent < 0 in advance_uv_buf')
 * 
 *         raise RuntimeError('fatal: Could not advance _StreamWriteContext')             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_tuple__148 = PyTuple_Pack(1, __pyx_kp_u_fatal_Could_not_advance__StreamW); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(20, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);

  /* "uvloop/handles/stream.pyx":203
 *             # Because we do an INCREF in _StreamWriteContext.new,
 *             # __dealloc__ shouldn't ever happen with `self.closed == 1`
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'open _StreamWriteContext is being deallocated')
 * 
 */
  __pyx_tuple__149 = PyTuple_Pack(1, __pyx_kp_u_open__StreamWriteContext_is_bein); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(20, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);

  /* "uvloop/handles/stream.pyx":231
 *     cdef _set_protocol(self, object protocol):
 *         if protocol is None:
 *             raise TypeError('protocol is required')             # <<<<<<<<<<<<<<
 * 
 *         UVBaseTransport._set_protocol(self, protocol)
 */
  __pyx_tuple__150 = PyTuple_Pack(1, __pyx_kp_u_protocol_is_required); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(20, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);

  /* "uvloop/handles/stream.pyx":358
 * 
 *         if (<uv.uv_stream_t*>self._handle).write_queue_size != 0:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'UVStream._try_write called with data in uv buffers')
 * 
 */
  __pyx_tuple__151 = PyTuple_Pack(1, __pyx_kp_u_UVStream__try_write_called_with); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(20, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);

  /* "uvloop/handles/stream.pyx":441
 *             if UVLOOP_DEBUG:
 *                 if self._buffer_size != 0 or self._buffer != []:
 *                     raise RuntimeError(             # <<<<<<<<<<<<<<
 *                         '_buffer_size is not 0 after a successful _exec_write')
 * 
 */
  __pyx_tuple__152 = PyTuple_Pack(1, __pyx_kp_u_buffer_size_is_not_0_after_a_su); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(20, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);

  /* "uvloop/handles/stream.pyx":487
 *                     if not self._closing:
 *                         # This should never happen.
 *                         raise RuntimeError(             # <<<<<<<<<<<<<<
 *                             'stream is open after UVStream._try_write '
 *                             'returned None')
 */
  __pyx_tuple__153 = PyTuple_Pack(1, __pyx_kp_u_stream_is_open_after_UVStream__t); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(20, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);

  /* "uvloop/handles/stream.pyx":503
 *                     if UVLOOP_DEBUG:
 *                         if sent == len(data):
 *                             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                                 '_try_write sent all data and returned '
 *                                 'non-zero')
 */
  __pyx_tuple__154 = PyTuple_Pack(1, __pyx_kp_u_try_write_sent_all_data_and_ret); if (unlikely(!__pyx_tuple__154)) __PYX_ERR(20, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);

  /* "uvloop/handles/stream.pyx":681
 * 
 *         if self._eof:
 *             raise RuntimeError('Cannot call write() after write_eof()')             # <<<<<<<<<<<<<<
 *         if not buf:
 *             return
 */
  __pyx_tuple__155 = PyTuple_Pack(1, __pyx_kp_u_Cannot_call_write_after_write_eo); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(20, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);

  /* "uvloop/handles/stream.pyx":694
 * 
 *         if self._eof:
 *             raise RuntimeError('Cannot call writelines() after write_eof()')             # <<<<<<<<<<<<<<
 *         if self._conn_lost:
 *             self._conn_lost += 1
 */
  __pyx_tuple__156 = PyTuple_Pack(1, __pyx_kp_u_Cannot_call_writelines_after_wri); if (unlikely(!__pyx_tuple__156)) __PYX_ERR(20, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__156);
  __Pyx_GIVEREF(__pyx_tuple__156);

  /* "uvloop/handles/stream.pyx":984
 *     if nread == uv.UV_ENOBUFS:
 *         sc._fatal_error(
 *             RuntimeError(             # <<<<<<<<<<<<<<
 *                 'unhandled error (or an empty buffer) in get_buffer()'),
 *             False)
 */
  __pyx_tuple__157 = PyTuple_Pack(1, __pyx_kp_u_unhandled_error_or_an_empty_buff); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(20, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);

  /* "uvloop/handles/streamserver.pyx":51
 * 
 *         if self.protocol_factory is None:
 *             raise RuntimeError('unable to listen(); no protocol_factory')             # <<<<<<<<<<<<<<
 * 
 *         if self.opened != 1:
 */
  __pyx_tuple__158 = PyTuple_Pack(1, __pyx_kp_u_unable_to_listen_no_protocol_fac); if (unlikely(!__pyx_tuple__158)) __PYX_ERR(21, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);

  /* "uvloop/handles/streamserver.pyx":54
 * 
 *         if self.opened != 1:
 *             raise RuntimeError('unopened TCPServer')             # <<<<<<<<<<<<<<
 * 
 *         self.context = Context_CopyCurrent()
 */
  __pyx_tuple__159 = PyTuple_Pack(1, __pyx_kp_u_unopened_TCPServer); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(21, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);

  /* "uvloop/handles/process.pyx":76
 *             # GIL should help us to avoid that.)
 *             self._abort_init()
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'Racing with another loop to spawn a process.')
 * 
 */
  __pyx_tuple__160 = PyTuple_Pack(1, __pyx_kp_u_Racing_with_another_loop_to_spaw); if (unlikely(!__pyx_tuple__160)) __PYX_ERR(4, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);

  /* "uvloop/handles/process.pyx":157
 *             # process must be dead now.
 *             try:
 *                 exc_name, exc_msg = errpipe_data.split(b':', 1)             # <<<<<<<<<<<<<<
 *                 exc_name = exc_name.decode()
 *                 exc_msg = exc_msg.decode()
 */
  __pyx_tuple__162 = PyTuple_Pack(2, __pyx_kp_b__161, __pyx_int_1); if (unlikely(!__pyx_tuple__162)) __PYX_ERR(4, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);

  /* "uvloop/handles/process.pyx":193
 *             except BaseException as ex:
 *                 try:
 *                     with open(self._errpipe_write, 'wb') as f:             # <<<<<<<<<<<<<<
 *                         f.write(str(ex.__class__.__name__).encode())
 *                         f.write(b':')
 */
  __pyx_tuple__163 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(4, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);

  /* "uvloop/handles/process.pyx":207
 *     cdef _close_after_spawn(self, int fd):
 *         if self._fds_to_close is None:
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'UVProcess._close_after_spawn called after uv_spawn')
 *         self._fds_to_close.append(fd)
 */
  __pyx_tuple__164 = PyTuple_Pack(1, __pyx_kp_u_UVProcess__close_after_spawn_cal); if (unlikely(!__pyx_tuple__164)) __PYX_ERR(4, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);

  /* "uvloop/handles/process.pyx":278
 *                 cwd = PyUnicode_EncodeFSDefault(cwd)
 *             if not isinstance(cwd, bytes):
 *                 raise ValueError('cwd must be a str or bytes object')             # <<<<<<<<<<<<<<
 * 
 *             self.__cwd = cwd
 */
  __pyx_tuple__165 = PyTuple_Pack(1, __pyx_kp_u_cwd_must_be_a_str_or_bytes_objec); if (unlikely(!__pyx_tuple__165)) __PYX_ERR(4, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);

  /* "uvloop/handles/process.pyx":293
 * 
 *         if an < 1:
 *             raise ValueError('cannot spawn a process: args are empty')             # <<<<<<<<<<<<<<
 * 
 *         self.__args = args.copy()
 */
  __pyx_tuple__166 = PyTuple_Pack(1, __pyx_kp_u_cannot_spawn_a_process_args_are); if (unlikely(!__pyx_tuple__166)) __PYX_ERR(4, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);

  /* "uvloop/handles/process.pyx":301
 *                 self.__args[i] = PyUnicode_EncodeFSDefault(arg)
 *             elif not isinstance(arg, bytes):
 *                 raise TypeError('all args must be str or bytes')             # <<<<<<<<<<<<<<
 * 
 *         path = self.__args[0]
 */
  __pyx_tuple__167 = PyTuple_Pack(1, __pyx_kp_u_all_args_must_be_str_or_bytes); if (unlikely(!__pyx_tuple__167)) __PYX_ERR(4, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);

  /* "uvloop/handles/process.pyx":316
 *                     key = PyUnicode_EncodeFSDefault(key)
 *                 elif not isinstance(key, bytes):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'all environment vars must be bytes or str')
 * 
 */
  __pyx_tuple__168 = PyTuple_Pack(1, __pyx_kp_u_all_environment_vars_must_be_byt); if (unlikely(!__pyx_tuple__168)) __PYX_ERR(4, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);

  /* "uvloop/handles/process.pyx":322
 *                     val = PyUnicode_EncodeFSDefault(val)
 *                 elif not isinstance(val, bytes):
 *                     raise TypeError(             # <<<<<<<<<<<<<<
 *                         'all environment values must be bytes or str')
 * 
 */
  __pyx_tuple__169 = PyTuple_Pack(1, __pyx_kp_u_all_environment_values_must_be_b); if (unlikely(!__pyx_tuple__169)) __PYX_ERR(4, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);

  /* "uvloop/handles/process.pyx":388
 *                                  context=self.context)
 *         else:
 *             self._pending_calls.append((_CALL_PROCESS_EXITED, None, None))             # <<<<<<<<<<<<<<
 * 
 *         self._try_finish()
 */
  __pyx_tuple__171 = PyTuple_Pack(3, __pyx_int_2, Py_None, Py_None); if (unlikely(!__pyx_tuple__171)) __PYX_ERR(4, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);

  /* "uvloop/handles/process.pyx":467
 *                 io[0] = self._file_devnull()
 *             elif _stdout == subprocess_STDOUT:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     'subprocess.STDOUT is supported only by stderr parameter')
 *             else:
 */
  __pyx_tuple__172 = PyTuple_Pack(1, __pyx_kp_u_subprocess_STDOUT_is_supported_o); if (unlikely(!__pyx_tuple__172)) __PYX_ERR(4, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__172);
  __Pyx_GIVEREF(__pyx_tuple__172);

  /* "uvloop/handles/process.pyx":517
 *                 if io[1] is None:
 *                     # shouldn't ever happen
 *                     raise RuntimeError('cannot apply subprocess.STDOUT')             # <<<<<<<<<<<<<<
 * 
 *                 io[2] = self._file_redirect_stdio(io[1])
 */
  __pyx_tuple__173 = PyTuple_Pack(1, __pyx_kp_u_cannot_apply_subprocess_STDOUT); if (unlikely(!__pyx_tuple__173)) __PYX_ERR(4, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);

  /* "uvloop/handles/process.pyx":582
 *                                      context=self.context)
 *             else:
 *                 self._pending_calls.append((_CALL_CONNECTION_LOST, None, None))             # <<<<<<<<<<<<<<
 * 
 *     def __stdio_inited(self, waiter, stdio_fut):
 */
  __pyx_tuple__174 = PyTuple_Pack(3, __pyx_int_3, Py_None, Py_None); if (unlikely(!__pyx_tuple__174)) __PYX_ERR(4, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__174);
  __Pyx_GIVEREF(__pyx_tuple__174);

  /* "uvloop/dns.pyx":22
 *             return socket_getservbyname(port, 'udp')
 * 
 *     raise OSError('service/proto not found')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__175 = PyTuple_Pack(1, __pyx_kp_u_service_proto_not_found); if (unlikely(!__pyx_tuple__175)) __PYX_ERR(5, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);

  /* "uvloop/dns.pyx":68
 *         return system.MakeUnixSockPyAddr(addr_un)
 * 
 *     raise RuntimeError("cannot convert sockaddr into Python object")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__176 = PyTuple_Pack(1, __pyx_kp_u_cannot_convert_sockaddr_into_Pyt); if (unlikely(!__pyx_tuple__176)) __PYX_ERR(5, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__176);
  __Pyx_GIVEREF(__pyx_tuple__176);

  /* "uvloop/dns.pyx":101
 *     if family == uv.AF_INET:
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET address must be tuple')             # <<<<<<<<<<<<<<
 *         if len(addr) != 2:
 *             raise ValueError('AF_INET address must be tuple of (host, port)')
 */
  __pyx_tuple__177 = PyTuple_Pack(1, __pyx_kp_u_AF_INET_address_must_be_tuple); if (unlikely(!__pyx_tuple__177)) __PYX_ERR(5, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);

  /* "uvloop/dns.pyx":103
 *             raise TypeError('AF_INET address must be tuple')
 *         if len(addr) != 2:
 *             raise ValueError('AF_INET address must be tuple of (host, port)')             # <<<<<<<<<<<<<<
 *         host, port = addr
 *         if isinstance(host, str):
 */
  __pyx_tuple__178 = PyTuple_Pack(1, __pyx_kp_u_AF_INET_address_must_be_tuple_of); if (unlikely(!__pyx_tuple__178)) __PYX_ERR(5, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__178);
  __Pyx_GIVEREF(__pyx_tuple__178);

  /* "uvloop/dns.pyx":112
 *                 host = host.encode('idna')
 *         if not isinstance(host, (bytes, bytearray)):
 *             raise TypeError('host must be a string or bytes object')             # <<<<<<<<<<<<<<
 * 
 *         port = __port_to_int(port, None)
 */
  __pyx_tuple__179 = PyTuple_Pack(1, __pyx_kp_u_host_must_be_a_string_or_bytes_o); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(5, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);

  /* "uvloop/dns.pyx":123
 *     elif family == uv.AF_INET6:
 *         if not isinstance(addr, tuple):
 *             raise TypeError('AF_INET6 address must be tuple')             # <<<<<<<<<<<<<<
 * 
 *         addr_len = len(addr)
 */
  __pyx_tuple__180 = PyTuple_Pack(1, __pyx_kp_u_AF_INET6_address_must_be_tuple); if (unlikely(!__pyx_tuple__180)) __PYX_ERR(5, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__180);
  __Pyx_GIVEREF(__pyx_tuple__180);

  /* "uvloop/dns.pyx":127
 *         addr_len = len(addr)
 *         if addr_len < 2 or addr_len > 4:
 *             raise ValueError(             # <<<<<<<<<<<<<<
 *                 'AF_INET6 must be a tuple of 2-4 parameters: '
 *                 '(host, port, flowinfo?, scope_id?)')
 */
  __pyx_tuple__181 = PyTuple_Pack(1, __pyx_kp_u_AF_INET6_must_be_a_tuple_of_2_4); if (unlikely(!__pyx_tuple__181)) __PYX_ERR(5, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);

  /* "uvloop/dns.pyx":161
 *             addr = addr.encode(sys_getfilesystemencoding())
 *         elif not isinstance(addr, bytes):
 *             raise TypeError('AF_UNIX address must be a str or a bytes object')             # <<<<<<<<<<<<<<
 * 
 *         PyBytes_AsStringAndSize(addr, &buf, &buflen)
 */
  __pyx_tuple__182 = PyTuple_Pack(1, __pyx_kp_u_AF_UNIX_address_must_be_a_str_or); if (unlikely(!__pyx_tuple__182)) __PYX_ERR(5, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__182);
  __Pyx_GIVEREF(__pyx_tuple__182);

  /* "uvloop/dns.pyx":310
 * 
 *         if self.data is NULL:
 *             raise RuntimeError('AddrInfo.data is NULL')             # <<<<<<<<<<<<<<
 * 
 *         ptr = self.data
 */
  __pyx_tuple__183 = PyTuple_Pack(1, __pyx_kp_u_AddrInfo_data_is_NULL); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(5, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);

  /* "uvloop/sslproto.pyx":3
 * cdef _create_transport_context(server_side, server_hostname):
 *     if server_side:
 *         raise ValueError('Server side SSL needs a valid SSLContext')             # <<<<<<<<<<<<<<
 * 
 *     # Client side may pass ssl=True to use a default
 */
  __pyx_tuple__184 = PyTuple_Pack(1, __pyx_kp_u_Server_side_SSL_needs_a_valid_SS); if (unlikely(!__pyx_tuple__184)) __PYX_ERR(25, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__184);
  __Pyx_GIVEREF(__pyx_tuple__184);

  /* "uvloop/sslproto.pyx":313
 *         if self._app_transport is None:
 *             if self._app_transport_created:
 *                 raise RuntimeError('Creating _SSLProtocolTransport twice')             # <<<<<<<<<<<<<<
 *             self._app_transport = _SSLProtocolTransport(self._loop, self,
 *                                                         context)
 */
  __pyx_tuple__185 = PyTuple_Pack(1, __pyx_kp_u_Creating__SSLProtocolTransport_t); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(25, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);

  /* "uvloop/handles/udp.pyx":48
 *         if UVLOOP_DEBUG:
 *             if not self.closed:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     'open _UDPSendContext is being deallocated')
 *         self.udp = None
 */
  __pyx_tuple__186 = PyTuple_Pack(1, __pyx_kp_u_open__UDPSendContext_is_being_de); if (unlikely(!__pyx_tuple__186)) __PYX_ERR(26, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__186);
  __Pyx_GIVEREF(__pyx_tuple__186);

  /* "uvloop/handles/udp.pyx":188
 *     cdef _new_socket(self):
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             raise RuntimeError(             # <<<<<<<<<<<<<<
 *                 'UDPTransport.family is undefined; '
 *                 'cannot create python socket')
 */
  __pyx_tuple__187 = PyTuple_Pack(1, __pyx_kp_u_UDPTransport_family_is_undefined); if (unlikely(!__pyx_tuple__187)) __PYX_ERR(26, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__187);
  __Pyx_GIVEREF(__pyx_tuple__187);

  /* "uvloop/handles/udp.pyx":206
 * 
 *         if self._family not in (uv.AF_INET, uv.AF_INET6, uv.AF_UNIX):
 *             raise RuntimeError('UDPTransport.family is undefined; cannot send')             # <<<<<<<<<<<<<<
 * 
 *         if addr is None:
 */
  __pyx_tuple__188 = PyTuple_Pack(1, __pyx_kp_u_UDPTransport_family_is_undefined_2); if (unlikely(!__pyx_tuple__188)) __PYX_ERR(26, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__188);
  __Pyx_GIVEREF(__pyx_tuple__188);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xbfaab04, 0xa551a14, 0x88c0b03):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xbfaab04, 0xa551a14, 0x88c0b03) = (_family, _fd, _peername, _proto, _sockname, _type))" % __pyx_checksum
 */
  __pyx_tuple__195 = PyTuple_Pack(3, __pyx_int_200977156, __pyx_int_173349396, __pyx_int_143395587); if (unlikely(!__pyx_tuple__195)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__195);
  __Pyx_GIVEREF(__pyx_tuple__195);
  __pyx_tuple__197 = PyTuple_Pack(3, __pyx_int_14808274, __pyx_int_121062814, __pyx_int_87255255); if (unlikely(!__pyx_tuple__197)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__197);
  __Pyx_GIVEREF(__pyx_tuple__197);

  /* "uvloop/includes/stdlib.pxi":4
 * 
 * 
 * import asyncio, asyncio.log, asyncio.base_events, \             # <<<<<<<<<<<<<<
 *        asyncio.sslproto, asyncio.coroutines, \
 *        asyncio.futures, asyncio.transports
 */
  __pyx_tuple__199 = PyTuple_Pack(2, __pyx_n_s_asyncio, __pyx_n_s_log); if (unlikely(!__pyx_tuple__199)) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__199);
  __Pyx_GIVEREF(__pyx_tuple__199);
  __pyx_tuple__200 = PyTuple_Pack(2, __pyx_n_s_asyncio, __pyx_n_s_base_events); if (unlikely(!__pyx_tuple__200)) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__200);
  __Pyx_GIVEREF(__pyx_tuple__200);

  /* "uvloop/includes/stdlib.pxi":5
 * 
 * import asyncio, asyncio.log, asyncio.base_events, \
 *        asyncio.sslproto, asyncio.coroutines, \             # <<<<<<<<<<<<<<
 *        asyncio.futures, asyncio.transports
 * import collections.abc
 */
  __pyx_tuple__201 = PyTuple_Pack(2, __pyx_n_s_asyncio, __pyx_n_s_sslproto); if (unlikely(!__pyx_tuple__201)) __PYX_ERR(31, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__201);
  __Pyx_GIVEREF(__pyx_tuple__201);
  __pyx_tuple__202 = PyTuple_Pack(2, __pyx_n_s_asyncio, __pyx_n_s_coroutines); if (unlikely(!__pyx_tuple__202)) __PYX_ERR(31, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__202);
  __Pyx_GIVEREF(__pyx_tuple__202);

  /* "uvloop/includes/stdlib.pxi":6
 * import asyncio, asyncio.log, asyncio.base_events, \
 *        asyncio.sslproto, asyncio.coroutines, \
 *        asyncio.futures, asyncio.transports             # <<<<<<<<<<<<<<
 * import collections.abc
 * import concurrent.futures
 */
  __pyx_tuple__203 = PyTuple_Pack(2, __pyx_n_s_asyncio, __pyx_n_s_futures); if (unlikely(!__pyx_tuple__203)) __PYX_ERR(31, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__203);
  __Pyx_GIVEREF(__pyx_tuple__203);
  __pyx_tuple__204 = PyTuple_Pack(2, __pyx_n_s_asyncio, __pyx_n_s_transports); if (unlikely(!__pyx_tuple__204)) __PYX_ERR(31, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__204);
  __Pyx_GIVEREF(__pyx_tuple__204);

  /* "uvloop/includes/stdlib.pxi":7
 *        asyncio.sslproto, asyncio.coroutines, \
 *        asyncio.futures, asyncio.transports
 * import collections.abc             # <<<<<<<<<<<<<<
 * import concurrent.futures
 * import errno
 */
  __pyx_tuple__205 = PyTuple_Pack(2, __pyx_n_s_collections, __pyx_n_s_abc); if (unlikely(!__pyx_tuple__205)) __PYX_ERR(31, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__205);
  __Pyx_GIVEREF(__pyx_tuple__205);

  /* "uvloop/includes/stdlib.pxi":8
 *        asyncio.futures, asyncio.transports
 * import collections.abc
 * import concurrent.futures             # <<<<<<<<<<<<<<
 * import errno
 * import functools
 */
  __pyx_tuple__206 = PyTuple_Pack(2, __pyx_n_s_concurrent, __pyx_n_s_futures); if (unlikely(!__pyx_tuple__206)) __PYX_ERR(31, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__206);
  __Pyx_GIVEREF(__pyx_tuple__206);

  /* "uvloop/loop.pyx":374
 *             self.remove_signal_handler(sig)
 * 
 *     def __sighandler(self, signum, frame):             # <<<<<<<<<<<<<<
 *         self._signals.add(signum)
 * 
 */
  __pyx_tuple__207 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_signum, __pyx_n_s_frame); if (unlikely(!__pyx_tuple__207)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__207);
  __Pyx_GIVEREF(__pyx_tuple__207);
  __pyx_codeobj__208 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sighandler, 374, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__208)) __PYX_ERR(0, 374, __pyx_L1_error)

  /* "uvloop/loop.pyx":1152
 *         self._coroutine_debug_set = enabled
 * 
 *     def _get_backend_id(self):             # <<<<<<<<<<<<<<
 *         """This method is used by uvloop tests and is not part of the API."""
 *         return uv.uv_backend_fd(self.uvloop)
 */
  __pyx_tuple__209 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__209)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__209);
  __Pyx_GIVEREF(__pyx_tuple__209);
  __pyx_codeobj__210 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_get_backend_id, 1152, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__210)) __PYX_ERR(0, 1152, __pyx_L1_error)

  /* "uvloop/loop.pyx":1268
 *         )
 * 
 *     def call_soon(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called as soon as possible.
 * 
 */
  __pyx_tuple__211 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_callback, __pyx_n_s_context, __pyx_n_s_args); if (unlikely(!__pyx_tuple__211)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__211);
  __Pyx_GIVEREF(__pyx_tuple__211);
  __pyx_codeobj__212 = (PyObject*)__Pyx_PyCode_New(2, 0, 1, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__211, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_call_soon, 1268, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__212)) __PYX_ERR(0, 1268, __pyx_L1_error)

  /* "uvloop/loop.pyx":1285
 *             return self._call_soon(callback, None, context)
 * 
 *     def call_soon_threadsafe(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_soon(), but thread-safe."""
 *         if not args:
 */
  __pyx_tuple__213 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_callback, __pyx_n_s_context, __pyx_n_s_args, __pyx_n_s_handle); if (unlikely(!__pyx_tuple__213)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__213);
  __Pyx_GIVEREF(__pyx_tuple__213);
  __pyx_codeobj__214 = (PyObject*)__Pyx_PyCode_New(2, 0, 1, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__213, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_call_soon_threadsafe, 1285, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__214)) __PYX_ERR(0, 1285, __pyx_L1_error)

  /* "uvloop/loop.pyx":1297
 *         return handle
 * 
 *     def call_later(self, delay, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called at a given time.
 * 
 */
  __pyx_tuple__215 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_delay, __pyx_n_s_callback, __pyx_n_s_context, __pyx_n_s_args, __pyx_n_s_when); if (unlikely(!__pyx_tuple__215)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__215);
  __Pyx_GIVEREF(__pyx_tuple__215);
  __pyx_codeobj__216 = (PyObject*)__Pyx_PyCode_New(3, 0, 1, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__215, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_call_later, 1297, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__216)) __PYX_ERR(0, 1297, __pyx_L1_error)

  /* "uvloop/loop.pyx":1334
 *             return self._call_later(when, callback, args, context)
 * 
 *     def call_at(self, when, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_later(), but uses an absolute time.
 * 
 */
  __pyx_tuple__217 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_when, __pyx_n_s_callback, __pyx_n_s_context, __pyx_n_s_args); if (unlikely(!__pyx_tuple__217)) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__217);
  __Pyx_GIVEREF(__pyx_tuple__217);
  __pyx_codeobj__218 = (PyObject*)__Pyx_PyCode_New(3, 0, 1, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__217, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_call_at, 1334, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__218)) __PYX_ERR(0, 1334, __pyx_L1_error)

  /* "uvloop/loop.pyx":1342
 *             when - self.time(), callback, *args, context=context)
 * 
 *     def time(self):             # <<<<<<<<<<<<<<
 *         """Return the time according to the event loop's clock.
 * 
 */
  __pyx_codeobj__219 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_time, 1342, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__219)) __PYX_ERR(0, 1342, __pyx_L1_error)

  /* "uvloop/loop.pyx":1351
 *         return self._time() / 1000
 * 
 *     def stop(self):             # <<<<<<<<<<<<<<
 *         """Stop running the event loop.
 * 
 */
  __pyx_codeobj__220 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_stop, 1351, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__220)) __PYX_ERR(0, 1351, __pyx_L1_error)

  /* "uvloop/loop.pyx":1366
 *                 None))
 * 
 *     def run_forever(self):             # <<<<<<<<<<<<<<
 *         """Run the event loop until stop() is called."""
 *         self._check_closed()
 */
  __pyx_tuple__221 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_mode, __pyx_n_s_old_agen_hooks); if (unlikely(!__pyx_tuple__221)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__221);
  __Pyx_GIVEREF(__pyx_tuple__221);
  __pyx_codeobj__222 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__221, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_run_forever, 1366, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__222)) __PYX_ERR(0, 1366, __pyx_L1_error)

  /* "uvloop/loop.pyx":1384
 *             sys.set_asyncgen_hooks(*old_agen_hooks)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the event loop.
 * 
 */
  __pyx_codeobj__223 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_close, 1384, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__223)) __PYX_ERR(0, 1384, __pyx_L1_error)

  /* "uvloop/loop.pyx":1395
 *         self._close()
 * 
 *     def get_debug(self):             # <<<<<<<<<<<<<<
 *         return bool(self._debug)
 * 
 */
  __pyx_codeobj__224 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_get_debug, 1395, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__224)) __PYX_ERR(0, 1395, __pyx_L1_error)

  /* "uvloop/loop.pyx":1398
 *         return bool(self._debug)
 * 
 *     def set_debug(self, enabled):             # <<<<<<<<<<<<<<
 *         self._debug = bool(enabled)
 *         if self.is_running():
 */
  __pyx_codeobj__225 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple_, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_set_debug, 1398, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__225)) __PYX_ERR(0, 1398, __pyx_L1_error)

  /* "uvloop/loop.pyx":1403
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)
 * 
 *     def is_running(self):             # <<<<<<<<<<<<<<
 *         """Return whether the event loop is currently running."""
 *         return bool(self._running)
 */
  __pyx_codeobj__226 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_is_running, 1403, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__226)) __PYX_ERR(0, 1403, __pyx_L1_error)

  /* "uvloop/loop.pyx":1407
 *         return bool(self._running)
 * 
 *     def is_closed(self):             # <<<<<<<<<<<<<<
 *         """Returns True if the event loop was closed."""
 *         return bool(self._closed)
 */
  __pyx_codeobj__227 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_is_closed, 1407, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__227)) __PYX_ERR(0, 1407, __pyx_L1_error)

  /* "uvloop/loop.pyx":1411
 *         return bool(self._closed)
 * 
 *     def create_future(self):             # <<<<<<<<<<<<<<
 *         """Create a Future object attached to the loop."""
 *         return self._new_future()
 */
  __pyx_codeobj__228 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_create_future, 1411, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__228)) __PYX_ERR(0, 1411, __pyx_L1_error)

  /* "uvloop/loop.pyx":1415
 *         return self._new_future()
 * 
 *     def create_task(self, coro, *, name=None, context=None):             # <<<<<<<<<<<<<<
 *         """Schedule a coroutine object.
 * 
 */
  __pyx_tuple__229 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_coro, __pyx_n_s_name_2, __pyx_n_s_context, __pyx_n_s_task, __pyx_n_s_set_name); if (unlikely(!__pyx_tuple__229)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__229);
  __Pyx_GIVEREF(__pyx_tuple__229);
  __pyx_codeobj__230 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__229, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_create_task, 1415, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__230)) __PYX_ERR(0, 1415, __pyx_L1_error)

  /* "uvloop/loop.pyx":1456
 *         return task
 * 
 *     def set_task_factory(self, factory):             # <<<<<<<<<<<<<<
 *         """Set a task factory that will be used by loop.create_task().
 * 
 */
  __pyx_tuple__231 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_factory); if (unlikely(!__pyx_tuple__231)) __PYX_ERR(0, 1456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__231);
  __Pyx_GIVEREF(__pyx_tuple__231);
  __pyx_codeobj__232 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__231, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_set_task_factory, 1456, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__232)) __PYX_ERR(0, 1456, __pyx_L1_error)

  /* "uvloop/loop.pyx":1470
 *         self._task_factory = factory
 * 
 *     def get_task_factory(self):             # <<<<<<<<<<<<<<
 *         """Return a task factory, or None if the default one is in use."""
 *         return self._task_factory
 */
  __pyx_codeobj__233 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_get_task_factory, 1470, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__233)) __PYX_ERR(0, 1470, __pyx_L1_error)

  /* "uvloop/loop.pyx":1474
 *         return self._task_factory
 * 
 *     def run_until_complete(self, future):             # <<<<<<<<<<<<<<
 *         """Run until the Future is done.
 * 
 */
  __pyx_tuple__234 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_future, __pyx_n_s_new_task, __pyx_n_s_done_cb, __pyx_n_s_done_cb); if (unlikely(!__pyx_tuple__234)) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__234);
  __Pyx_GIVEREF(__pyx_tuple__234);
  __pyx_codeobj__235 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__234, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_run_until_complete, 1474, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__235)) __PYX_ERR(0, 1474, __pyx_L1_error)

  /* "uvloop/loop.pyx":1520
 *         return future.result()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getaddrinfo(self, object host, object port, *,
 *                           int family=0, int type=0, int proto=0, int flags=0):
 */
  __pyx_tuple__236 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_host, __pyx_n_s_port, __pyx_n_s_family, __pyx_n_s_type, __pyx_n_s_proto, __pyx_n_s_flags, __pyx_n_s_addr); if (unlikely(!__pyx_tuple__236)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__236);
  __Pyx_GIVEREF(__pyx_tuple__236);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__236, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_getaddrinfo, 1520, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(0, 1520, __pyx_L1_error)

  /* "uvloop/loop.pyx":1532
 *             host, port, family, type, proto, flags, 1)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getnameinfo(self, sockaddr, int flags=0):
 *         cdef:
 */
  __pyx_tuple__237 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_sockaddr, __pyx_n_s_flags, __pyx_n_s_ai_cnt, __pyx_n_s_ai, __pyx_n_s_sin6, __pyx_n_s_sl, __pyx_n_s_flowinfo, __pyx_n_s_scope_id); if (unlikely(!__pyx_tuple__237)) __PYX_ERR(0, 1532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__237);
  __Pyx_GIVEREF(__pyx_tuple__237);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__237, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_getnameinfo, 1532, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 1532, __pyx_L1_error)
  __pyx_tuple__238 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__238)) __PYX_ERR(0, 1532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__238);
  __Pyx_GIVEREF(__pyx_tuple__238);

  /* "uvloop/loop.pyx":1587
 *         return await self._getnameinfo(ai.ai_addr, flags)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 */
  __pyx_tuple__239 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_transport, __pyx_n_s_protocol, __pyx_n_s_sslcontext, __pyx_n_s_server_side, __pyx_n_s_server_hostname, __pyx_n_s_ssl_handshake_timeout, __pyx_n_s_ssl_shutdown_timeout, __pyx_n_s_context, __pyx_n_s_waiter, __pyx_n_s_ssl_protocol, __pyx_n_s_conmade_cb, __pyx_n_s_resume_cb, __pyx_n_s_app_transport); if (unlikely(!__pyx_tuple__239)) __PYX_ERR(0, 1587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__239);
  __Pyx_GIVEREF(__pyx_tuple__239);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__239, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_start_tls, 1587, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 1587, __pyx_L1_error)

  /* "uvloop/loop.pyx":1643
 *         return app_transport
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 */
  __pyx_tuple__240 = PyTuple_Pack(25, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_host, __pyx_n_s_port, __pyx_n_s_family, __pyx_n_s_flags, __pyx_n_s_sock, __pyx_n_s_backlog, __pyx_n_s_ssl, __pyx_n_s_reuse_address, __pyx_n_s_reuse_port, __pyx_n_s_ssl_handshake_timeout, __pyx_n_s_ssl_shutdown_timeout, __pyx_n_s_start_serving, __pyx_n_s_tcp, __pyx_n_s_addrinfo, __pyx_n_s_server_2, __pyx_n_s_hosts, __pyx_n_s_fs, __pyx_n_s_infos, __pyx_n_s_completed, __pyx_n_s_info, __pyx_n_s_pyaddr, __pyx_n_s_err, __pyx_n_s_host); if (unlikely(!__pyx_tuple__240)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__240);
  __Pyx_GIVEREF(__pyx_tuple__240);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(4, 0, 10, 25, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_create_server, 1643, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __pyx_tuple__241 = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_tuple__241)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__241);
  __Pyx_GIVEREF(__pyx_tuple__241);

  /* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */
  __pyx_tuple__242 = PyTuple_Pack(41, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_host, __pyx_n_s_port, __pyx_n_s_ssl, __pyx_n_s_family, __pyx_n_s_proto, __pyx_n_s_flags, __pyx_n_s_sock, __pyx_n_s_local_addr, __pyx_n_s_server_hostname, __pyx_n_s_ssl_handshake_timeout, __pyx_n_s_ssl_shutdown_timeout, __pyx_n_s_ai_local, __pyx_n_s_ai_remote, __pyx_n_s_tr, __pyx_n_s_rai, __pyx_n_s_lai, __pyx_n_s_rai_iter, __pyx_n_s_lai_iter, __pyx_n_s_rai_static, __pyx_n_s_rai_addr_static, __pyx_n_s_lai_static, __pyx_n_s_lai_addr_static, __pyx_n_s_app_protocol, __pyx_n_s_app_transport, __pyx_n_s_protocol, __pyx_n_s_ssl_waiter, __pyx_n_s_context, __pyx_n_s_sslcontext, __pyx_n_s_fs, __pyx_n_s_f1, __pyx_n_s_f2, __pyx_n_s_addr, __pyx_n_s_exceptions, __pyx_n_s_waiter, __pyx_n_s_exc, __pyx_n_s_model, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__242)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__242);
  __Pyx_GIVEREF(__pyx_tuple__242);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 41, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__242, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_create_connection, 1857, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 1857, __pyx_L1_error)

  /* "uvloop/loop.pyx":2093
 *             return tr, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 */
  __pyx_tuple__243 = PyTuple_Pack(16, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_path, __pyx_n_s_backlog, __pyx_n_s_sock, __pyx_n_s_ssl, __pyx_n_s_ssl_handshake_timeout, __pyx_n_s_ssl_shutdown_timeout, __pyx_n_s_start_serving, __pyx_n_s_cleanup_socket, __pyx_n_s_pipe, __pyx_n_s_server_2, __pyx_n_s_orig_path, __pyx_n_s_err, __pyx_n_s_exc, __pyx_n_s_msg); if (unlikely(!__pyx_tuple__243)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__243);
  __Pyx_GIVEREF(__pyx_tuple__243);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(3, 0, 7, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__243, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_create_unix_server, 2093, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __pyx_tuple__244 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__244)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__244);
  __Pyx_GIVEREF(__pyx_tuple__244);

  /* "uvloop/loop.pyx":2235
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 */
  __pyx_tuple__245 = PyTuple_Pack(16, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_path, __pyx_n_s_ssl, __pyx_n_s_sock, __pyx_n_s_server_hostname, __pyx_n_s_ssl_handshake_timeout, __pyx_n_s_ssl_shutdown_timeout, __pyx_n_s_tr, __pyx_n_s_app_protocol, __pyx_n_s_app_transport, __pyx_n_s_protocol, __pyx_n_s_ssl_waiter, __pyx_n_s_context, __pyx_n_s_sslcontext, __pyx_n_s_waiter); if (unlikely(!__pyx_tuple__245)) __PYX_ERR(0, 2235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__245);
  __Pyx_GIVEREF(__pyx_tuple__245);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__245, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_create_unix_connection, 2235, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 2235, __pyx_L1_error)

  /* "uvloop/loop.pyx":2335
 *             return tr, protocol
 * 
 *     def default_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Default exception handler.
 * 
 */
  __pyx_tuple__246 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_context, __pyx_n_s_message, __pyx_n_s_exception, __pyx_n_s_exc_info, __pyx_n_s_log_lines, __pyx_n_s_key, __pyx_n_s_value, __pyx_n_s_tb, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__246)) __PYX_ERR(0, 2335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__246);
  __Pyx_GIVEREF(__pyx_tuple__246);
  __pyx_codeobj__247 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__246, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_default_exception_handler, 2335, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__247)) __PYX_ERR(0, 2335, __pyx_L1_error)

  /* "uvloop/loop.pyx":2376
 *         aio_logger.error('\n'.join(log_lines), exc_info=exc_info)
 * 
 *     def get_exception_handler(self):             # <<<<<<<<<<<<<<
 *         """Return an exception handler, or None if the default one is in use.
 *         """
 */
  __pyx_codeobj__248 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_get_exception_handler, 2376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__248)) __PYX_ERR(0, 2376, __pyx_L1_error)

  /* "uvloop/loop.pyx":2381
 *         return self._exception_handler
 * 
 *     def set_exception_handler(self, handler):             # <<<<<<<<<<<<<<
 *         """Set handler as the new event loop exception handler.
 * 
 */
  __pyx_tuple__249 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_handler); if (unlikely(!__pyx_tuple__249)) __PYX_ERR(0, 2381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__249);
  __Pyx_GIVEREF(__pyx_tuple__249);
  __pyx_codeobj__250 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__249, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_set_exception_handler, 2381, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__250)) __PYX_ERR(0, 2381, __pyx_L1_error)

  /* "uvloop/loop.pyx":2398
 *         self._exception_handler = handler
 * 
 *     def call_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Call the current event loop's exception handler.
 * 
 */
  __pyx_tuple__251 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_context, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__251)) __PYX_ERR(0, 2398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__251);
  __Pyx_GIVEREF(__pyx_tuple__251);
  __pyx_codeobj__252 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__251, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_call_exception_handler, 2398, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__252)) __PYX_ERR(0, 2398, __pyx_L1_error)

  /* "uvloop/loop.pyx":2455
 *                                      exc_info=True)
 * 
 *     def add_reader(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a reader callback."""
 *         if len(args) == 0:
 */
  __pyx_tuple__253 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_fileobj, __pyx_n_s_callback, __pyx_n_s_args); if (unlikely(!__pyx_tuple__253)) __PYX_ERR(0, 2455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__253);
  __Pyx_GIVEREF(__pyx_tuple__253);
  __pyx_codeobj__254 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__253, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_add_reader, 2455, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__254)) __PYX_ERR(0, 2455, __pyx_L1_error)

  /* "uvloop/loop.pyx":2461
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a reader callback."""
 *         self._remove_reader(fileobj)
 */
  __pyx_tuple__255 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_fileobj); if (unlikely(!__pyx_tuple__255)) __PYX_ERR(0, 2461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__255);
  __Pyx_GIVEREF(__pyx_tuple__255);
  __pyx_codeobj__256 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__255, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_remove_reader, 2461, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__256)) __PYX_ERR(0, 2461, __pyx_L1_error)

  /* "uvloop/loop.pyx":2465
 *         self._remove_reader(fileobj)
 * 
 *     def add_writer(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a writer callback.."""
 *         if len(args) == 0:
 */
  __pyx_codeobj__257 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__253, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_add_writer, 2465, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__257)) __PYX_ERR(0, 2465, __pyx_L1_error)

  /* "uvloop/loop.pyx":2471
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a writer callback."""
 *         self._remove_writer(fileobj)
 */
  __pyx_codeobj__258 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__255, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_remove_writer, 2471, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__258)) __PYX_ERR(0, 2471, __pyx_L1_error)

  /* "uvloop/loop.pyx":2475
 *         self._remove_writer(fileobj)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv(self, sock, n):
 *         """Receive data from the socket.
 */
  __pyx_tuple__259 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_n, __pyx_n_s_handle, __pyx_n_s_fut); if (unlikely(!__pyx_tuple__259)) __PYX_ERR(0, 2475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__259);
  __Pyx_GIVEREF(__pyx_tuple__259);
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__259, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_recv, 2475, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(0, 2475, __pyx_L1_error)

  /* "uvloop/loop.pyx":2503
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv_into(self, sock, buf):
 *         """Receive data from the socket.
 */
  __pyx_tuple__260 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_buf, __pyx_n_s_handle, __pyx_n_s_fut); if (unlikely(!__pyx_tuple__260)) __PYX_ERR(0, 2503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__260);
  __Pyx_GIVEREF(__pyx_tuple__260);
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__260, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_recv_into, 2503, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(0, 2503, __pyx_L1_error)

  /* "uvloop/loop.pyx":2530
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendall(self, sock, data):
 *         """Send data to the socket.
 */
  __pyx_tuple__261 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_data, __pyx_n_s_handle, __pyx_n_s_n, __pyx_n_s_fut); if (unlikely(!__pyx_tuple__261)) __PYX_ERR(0, 2530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__261);
  __Pyx_GIVEREF(__pyx_tuple__261);
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__261, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_sendall, 2530, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(0, 2530, __pyx_L1_error)

  /* "uvloop/loop.pyx":2584
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_accept(self, sock):
 *         """Accept a connection.
 */
  __pyx_tuple__262 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_handle, __pyx_n_s_fut); if (unlikely(!__pyx_tuple__262)) __PYX_ERR(0, 2584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__262);
  __Pyx_GIVEREF(__pyx_tuple__262);
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__262, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_accept, 2584, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(0, 2584, __pyx_L1_error)

  /* "uvloop/loop.pyx":2613
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_connect(self, sock, address):
 *         """Connect to a remote socket at address.
 */
  __pyx_tuple__264 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_address, __pyx_n_s_fut, __pyx_n_s_addrs, __pyx_n_s__263); if (unlikely(!__pyx_tuple__264)) __PYX_ERR(0, 2613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__264);
  __Pyx_GIVEREF(__pyx_tuple__264);
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__264, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_connect, 2613, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 2613, __pyx_L1_error)

  /* "uvloop/loop.pyx":2637
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom(self, sock, bufsize):
 *         raise NotImplementedError
 */
  __pyx_tuple__265 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_bufsize); if (unlikely(!__pyx_tuple__265)) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__265);
  __Pyx_GIVEREF(__pyx_tuple__265);
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__265, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_recvfrom, 2637, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 2637, __pyx_L1_error)

  /* "uvloop/loop.pyx":2641
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom_into(self, sock, buf, nbytes=0):
 *         raise NotImplementedError
 */
  __pyx_tuple__266 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_buf, __pyx_n_s_nbytes); if (unlikely(!__pyx_tuple__266)) __PYX_ERR(0, 2641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__266);
  __Pyx_GIVEREF(__pyx_tuple__266);
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__266, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_recvfrom_into, 2641, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 2641, __pyx_L1_error)

  /* "uvloop/loop.pyx":2645
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendto(self, sock, data, address):
 *         raise NotImplementedError
 */
  __pyx_tuple__267 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_data, __pyx_n_s_address); if (unlikely(!__pyx_tuple__267)) __PYX_ERR(0, 2645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__267);
  __Pyx_GIVEREF(__pyx_tuple__267);
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__267, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_sock_sendto, 2645, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 2645, __pyx_L1_error)

  /* "uvloop/loop.pyx":2649
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 */
  __pyx_tuple__268 = PyTuple_Pack(13, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_sock, __pyx_n_s_ssl, __pyx_n_s_ssl_handshake_timeout, __pyx_n_s_ssl_shutdown_timeout, __pyx_n_s_transport, __pyx_n_s_app_protocol, __pyx_n_s_waiter, __pyx_n_s_transport_waiter, __pyx_n_s_context, __pyx_n_s_protocol, __pyx_n_s_app_transport); if (unlikely(!__pyx_tuple__268)) __PYX_ERR(0, 2649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__268);
  __Pyx_GIVEREF(__pyx_tuple__268);
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__268, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_connect_accepted_socket, 2649, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 2649, __pyx_L1_error)

  /* "uvloop/loop.pyx":2733
 *             return transport, protocol
 * 
 *     def run_in_executor(self, executor, func, *args):             # <<<<<<<<<<<<<<
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):
 *             raise TypeError("coroutines cannot be used with run_in_executor()")
 */
  __pyx_tuple__269 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_executor, __pyx_n_s_func, __pyx_n_s_args); if (unlikely(!__pyx_tuple__269)) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__269);
  __Pyx_GIVEREF(__pyx_tuple__269);
  __pyx_codeobj__270 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__269, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_run_in_executor, 2733, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__270)) __PYX_ERR(0, 2733, __pyx_L1_error)

  /* "uvloop/loop.pyx":2749
 *         return aio_wrap_future(executor.submit(func, *args), loop=self)
 * 
 *     def set_default_executor(self, executor):             # <<<<<<<<<<<<<<
 *         self._default_executor = executor
 * 
 */
  __pyx_tuple__271 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_executor); if (unlikely(!__pyx_tuple__271)) __PYX_ERR(0, 2749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__271);
  __Pyx_GIVEREF(__pyx_tuple__271);
  __pyx_codeobj__272 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__271, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_set_default_executor, 2749, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__272)) __PYX_ERR(0, 2749, __pyx_L1_error)

  /* "uvloop/loop.pyx":2752
 *         self._default_executor = executor
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 */
  __pyx_tuple__273 = PyTuple_Pack(24, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_args, __pyx_n_s_stdin, __pyx_n_s_stdout, __pyx_n_s_stderr, __pyx_n_s_universal_newlines, __pyx_n_s_shell, __pyx_n_s_bufsize, __pyx_n_s_preexec_fn, __pyx_n_s_close_fds, __pyx_n_s_cwd, __pyx_n_s_env, __pyx_n_s_startupinfo, __pyx_n_s_creationflags, __pyx_n_s_restore_signals, __pyx_n_s_start_new_session, __pyx_n_s_executable, __pyx_n_s_pass_fds, __pyx_n_s_kwargs, __pyx_n_s_debug_flags, __pyx_n_s_waiter, __pyx_n_s_protocol, __pyx_n_s_proc); if (unlikely(!__pyx_tuple__273)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__273);
  __Pyx_GIVEREF(__pyx_tuple__273);
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(19, 0, 0, 24, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__273, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_subprocess_run, 2752, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(0, 2752, __pyx_L1_error)

  /* "uvloop/loop.pyx":2817
 *         return proc, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,
 */
  __pyx_tuple__274 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_cmd, __pyx_n_s_shell, __pyx_n_s_kwargs, __pyx_n_s_args); if (unlikely(!__pyx_tuple__274)) __PYX_ERR(0, 2817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__274);
  __Pyx_GIVEREF(__pyx_tuple__274);
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(3, 0, 1, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__274, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_subprocess_shell, 2817, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 2817, __pyx_L1_error)

  /* "uvloop/loop.pyx":2832
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):
 */
  __pyx_tuple__275 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_program, __pyx_n_s_shell, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__275)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__275);
  __Pyx_GIVEREF(__pyx_tuple__275);
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(3, 0, 1, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__275, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_subprocess_exec, 2832, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(0, 2832, __pyx_L1_error)

  /* "uvloop/loop.pyx":2844
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_read_pipe(self, proto_factory, pipe):
 *         """Register read pipe in event loop. Set the pipe to non-blocking mode.
 */
  __pyx_tuple__276 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_proto_factory, __pyx_n_s_pipe, __pyx_n_s_transp, __pyx_n_s_waiter, __pyx_n_s_proto); if (unlikely(!__pyx_tuple__276)) __PYX_ERR(0, 2844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__276);
  __Pyx_GIVEREF(__pyx_tuple__276);
  __pyx_codeobj__107 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__276, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_connect_read_pipe, 2844, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__107)) __PYX_ERR(0, 2844, __pyx_L1_error)

  /* "uvloop/loop.pyx":2871
 *         return transp, proto
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_write_pipe(self, proto_factory, pipe):
 *         """Register write pipe in event loop.
 */
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__276, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_connect_write_pipe, 2871, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 2871, __pyx_L1_error)

  /* "uvloop/loop.pyx":2898
 *         return transp, proto
 * 
 *     def add_signal_handler(self, sig, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a handler for a signal.  UNIX only.
 * 
 */
  __pyx_tuple__277 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_sig, __pyx_n_s_callback, __pyx_n_s_args, __pyx_n_s_h, __pyx_n_s_exc, __pyx_n_s_nexc); if (unlikely(!__pyx_tuple__277)) __PYX_ERR(0, 2898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__277);
  __Pyx_GIVEREF(__pyx_tuple__277);
  __pyx_codeobj__278 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__277, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_add_signal_handler, 2898, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__278)) __PYX_ERR(0, 2898, __pyx_L1_error)

  /* "uvloop/loop.pyx":2966
 *                 raise
 * 
 *     def remove_signal_handler(self, sig):             # <<<<<<<<<<<<<<
 *         """Remove a handler for a signal.  UNIX only.
 * 
 */
  __pyx_tuple__279 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_sig, __pyx_n_s_handler, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__279)) __PYX_ERR(0, 2966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__279);
  __Pyx_GIVEREF(__pyx_tuple__279);
  __pyx_codeobj__280 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__279, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_remove_signal_handler, 2966, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__280)) __PYX_ERR(0, 2966, __pyx_L1_error)

  /* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */
  __pyx_tuple__281 = PyTuple_Pack(25, __pyx_n_s_self, __pyx_n_s_protocol_factory, __pyx_n_s_local_addr, __pyx_n_s_remote_addr, __pyx_n_s_family, __pyx_n_s_proto, __pyx_n_s_flags, __pyx_n_s_reuse_address, __pyx_n_s_reuse_port, __pyx_n_s_allow_broadcast, __pyx_n_s_sock, __pyx_n_s_udp, __pyx_n_s_lai, __pyx_n_s_rai, __pyx_n_s_opts, __pyx_n_s_problems, __pyx_n_s_lads, __pyx_n_s_rads, __pyx_n_s_excs, __pyx_n_s_ex, __pyx_n_s_ctx, __pyx_n_s_protocol, __pyx_n_s_waiter, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__281)) __PYX_ERR(0, 3002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__281);
  __Pyx_GIVEREF(__pyx_tuple__281);
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(4, 0, 7, 25, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__281, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_create_datagram_endpoint, 3002, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 3002, __pyx_L1_error)

  /* "uvloop/loop.pyx":3167
 *         return udp, protocol
 * 
 *     def _monitor_fs(self, path: str, callback) -> asyncio.Handle:             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVFSEvent fs_handle
 */
  __pyx_tuple__282 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_path, __pyx_n_s_callback, __pyx_n_s_fs_handle, __pyx_n_s_c_str_path, __pyx_n_s_p_bytes, __pyx_n_s_flags); if (unlikely(!__pyx_tuple__282)) __PYX_ERR(0, 3167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__282);
  __Pyx_GIVEREF(__pyx_tuple__282);
  __pyx_codeobj__283 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__282, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_monitor_fs, 3167, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__283)) __PYX_ERR(0, 3167, __pyx_L1_error)

  /* "uvloop/loop.pyx":3180
 *         return fs_handle
 * 
 *     def _check_default_executor(self):             # <<<<<<<<<<<<<<
 *         if self._executor_shutdown_called:
 *             raise RuntimeError('Executor shutdown has been called')
 */
  __pyx_codeobj__284 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_check_default_executor, 3180, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__284)) __PYX_ERR(0, 3180, __pyx_L1_error)

  /* "uvloop/loop.pyx":3184
 *             raise RuntimeError('Executor shutdown has been called')
 * 
 *     def _asyncgen_finalizer_hook(self, agen):             # <<<<<<<<<<<<<<
 *         self._asyncgens.discard(agen)
 *         if not self.is_closed():
 */
  __pyx_tuple__285 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_agen); if (unlikely(!__pyx_tuple__285)) __PYX_ERR(0, 3184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__285);
  __Pyx_GIVEREF(__pyx_tuple__285);
  __pyx_codeobj__286 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__285, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_asyncgen_finalizer_hook, 3184, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__286)) __PYX_ERR(0, 3184, __pyx_L1_error)

  /* "uvloop/loop.pyx":3189
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())
 * 
 *     def _asyncgen_firstiter_hook(self, agen):             # <<<<<<<<<<<<<<
 *         if self._asyncgens_shutdown_called:
 *             warnings_warn(
 */
  __pyx_codeobj__287 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__285, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_asyncgen_firstiter_hook, 3189, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__287)) __PYX_ERR(0, 3189, __pyx_L1_error)

  /* "uvloop/loop.pyx":3198
 *         self._asyncgens.add(agen)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_asyncgens(self):
 *         """Shutdown all active asynchronous generators."""
 */
  __pyx_tuple__288 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_closing_agens, __pyx_n_s_shutdown_coro, __pyx_n_s_results, __pyx_n_s_result, __pyx_n_s_agen, __pyx_n_s_ag); if (unlikely(!__pyx_tuple__288)) __PYX_ERR(0, 3198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__288);
  __Pyx_GIVEREF(__pyx_tuple__288);
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__288, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_shutdown_asyncgens, 3198, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(0, 3198, __pyx_L1_error)

  /* "uvloop/loop.pyx":3223
 *                 })
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_default_executor(self, timeout=None):
 *         """Schedule the shutdown of the default executor.
 */
  __pyx_tuple__289 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_timeout, __pyx_n_s_future, __pyx_n_s_thread); if (unlikely(!__pyx_tuple__289)) __PYX_ERR(0, 3223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__289);
  __Pyx_GIVEREF(__pyx_tuple__289);
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__289, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_shutdown_default_executor, 3223, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(0, 3223, __pyx_L1_error)

  /* "uvloop/loop.pyx":3251
 *             self._default_executor.shutdown(wait=False)
 * 
 *     def _do_shutdown(self, future):             # <<<<<<<<<<<<<<
 *         try:
 *             self._default_executor.shutdown(wait=True)
 */
  __pyx_tuple__290 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_future, __pyx_n_s_ex); if (unlikely(!__pyx_tuple__290)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__290);
  __Pyx_GIVEREF(__pyx_tuple__290);
  __pyx_codeobj__291 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__290, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_do_shutdown, 3251, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__291)) __PYX_ERR(0, 3251, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__292 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__292)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_tuple__293 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__293)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__293);
  __Pyx_GIVEREF(__pyx_tuple__293);
  __pyx_codeobj__294 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__294)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/loop.pyx":3260
 * 
 * # Expose pointer for integration with other C-extensions
 * def libuv_get_loop_t_ptr(loop):             # <<<<<<<<<<<<<<
 *     return PyCapsule_New(<void *>(<Loop>loop).uvloop, NULL, NULL)
 * 
 */
  __pyx_tuple__295 = PyTuple_Pack(1, __pyx_n_s_loop); if (unlikely(!__pyx_tuple__295)) __PYX_ERR(0, 3260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__295);
  __Pyx_GIVEREF(__pyx_tuple__295);
  __pyx_codeobj__296 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__295, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_libuv_get_loop_t_ptr, 3260, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__296)) __PYX_ERR(0, 3260, __pyx_L1_error)

  /* "uvloop/loop.pyx":3264
 * 
 * 
 * def libuv_get_version():             # <<<<<<<<<<<<<<
 *     return uv.uv_version()
 * 
 */
  __pyx_codeobj__297 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_libuv_get_version, 3264, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__297)) __PYX_ERR(0, 3264, __pyx_L1_error)

  /* "uvloop/loop.pyx":3268
 * 
 * 
 * def _testhelper_unwrap_capsuled_pointer(obj):             # <<<<<<<<<<<<<<
 *     return <uint64_t>PyCapsule_GetPointer(obj, NULL)
 * 
 */
  __pyx_tuple__298 = PyTuple_Pack(1, __pyx_n_s_obj); if (unlikely(!__pyx_tuple__298)) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__298);
  __Pyx_GIVEREF(__pyx_tuple__298);
  __pyx_codeobj__299 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__298, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_testhelper_unwrap_capsuled_poin, 3268, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__299)) __PYX_ERR(0, 3268, __pyx_L1_error)

  /* "uvloop/loop.pyx":3297
 * class _SyncSocketReaderFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */
  __pyx_tuple__300 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_sock, __pyx_n_s_loop); if (unlikely(!__pyx_tuple__300)) __PYX_ERR(0, 3297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__300);
  __Pyx_GIVEREF(__pyx_tuple__300);
  __pyx_codeobj__301 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__300, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_init, 3297, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__301)) __PYX_ERR(0, 3297, __pyx_L1_error)

  /* "uvloop/loop.pyx":3302
 *         self.__loop = loop
 * 
 *     def __remove_reader(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_reader(self.__sock)
 */
  __pyx_codeobj__302 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_remove_reader_2, 3302, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__302)) __PYX_ERR(0, 3302, __pyx_L1_error)

  /* "uvloop/loop.pyx":3308
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self, msg=msg)
 */
  __pyx_tuple__303 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_msg); if (unlikely(!__pyx_tuple__303)) __PYX_ERR(0, 3308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__303);
  __Pyx_GIVEREF(__pyx_tuple__303);
  __pyx_codeobj__304 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__303, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_cancel, 3308, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__304)) __PYX_ERR(0, 3308, __pyx_L1_error)

  /* "uvloop/loop.pyx":3313
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self)
 */
  __pyx_codeobj__305 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_cancel, 3313, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__305)) __PYX_ERR(0, 3313, __pyx_L1_error)

  /* "uvloop/loop.pyx":3320
 * class _SyncSocketWriterFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */
  __pyx_codeobj__306 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__300, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_init, 3320, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__306)) __PYX_ERR(0, 3320, __pyx_L1_error)

  /* "uvloop/loop.pyx":3325
 *         self.__loop = loop
 * 
 *     def __remove_writer(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_writer(self.__sock)
 */
  __pyx_codeobj__307 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_remove_writer_2, 3325, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__307)) __PYX_ERR(0, 3325, __pyx_L1_error)

  /* "uvloop/loop.pyx":3331
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self, msg=msg)
 */
  __pyx_codeobj__308 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__303, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_cancel, 3331, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__308)) __PYX_ERR(0, 3331, __pyx_L1_error)

  /* "uvloop/loop.pyx":3336
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self)
 */
  __pyx_codeobj__309 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_cancel, 3336, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__309)) __PYX_ERR(0, 3336, __pyx_L1_error)

  /* "uvloop/cbhandles.pyx":158
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */
  __pyx_codeobj__310 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_cbhandles_pyx, __pyx_n_s_cancel, 158, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__310)) __PYX_ERR(10, 158, __pyx_L1_error)

  /* "uvloop/cbhandles.pyx":161
 *         self._cancel()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */
  __pyx_codeobj__311 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_cbhandles_pyx, __pyx_n_s_cancelled, 161, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__311)) __PYX_ERR(10, 161, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__312 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__312)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__313 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__313)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/cbhandles.pyx":306
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */
  __pyx_codeobj__314 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_cbhandles_pyx, __pyx_n_s_cancelled, 306, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__314)) __PYX_ERR(10, 306, __pyx_L1_error)

  /* "uvloop/cbhandles.pyx":309
 *         return self._cancelled
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */
  __pyx_codeobj__315 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_cbhandles_pyx, __pyx_n_s_cancel, 309, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__315)) __PYX_ERR(10, 309, __pyx_L1_error)

  /* "uvloop/cbhandles.pyx":312
 *         self._cancel()
 * 
 *     def when(self):             # <<<<<<<<<<<<<<
 *         return self._when
 * 
 */
  __pyx_codeobj__316 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_cbhandles_pyx, __pyx_n_s_when, 312, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__316)) __PYX_ERR(10, 312, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__317 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__317)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__318 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__318)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":63
 *         return s
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError("Cannot serialize socket object")
 * 
 */
  __pyx_codeobj__319 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_getstate, 63, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__319)) __PYX_ERR(11, 63, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":66
 *         raise TypeError("Cannot serialize socket object")
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         return self._fd
 * 
 */
  __pyx_codeobj__320 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_fileno, 66, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__320)) __PYX_ERR(11, 66, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":69
 *         return self._fd
 * 
 *     def dup(self):             # <<<<<<<<<<<<<<
 *         fd = os_dup(self._fd)
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)
 */
  __pyx_tuple__321 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_fd, __pyx_n_s_sock); if (unlikely(!__pyx_tuple__321)) __PYX_ERR(11, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__321);
  __Pyx_GIVEREF(__pyx_tuple__321);
  __pyx_codeobj__322 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__321, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_dup, 69, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__322)) __PYX_ERR(11, 69, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":75
 *         return sock
 * 
 *     def get_inheritable(self):             # <<<<<<<<<<<<<<
 *         return os_get_inheritable(self._fd)
 * 
 */
  __pyx_codeobj__323 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_get_inheritable, 75, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__323)) __PYX_ERR(11, 75, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":78
 *         return os_get_inheritable(self._fd)
 * 
 *     def set_inheritable(self):             # <<<<<<<<<<<<<<
 *         os_set_inheritable(self._fd)
 * 
 */
  __pyx_codeobj__324 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_set_inheritable, 78, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__324)) __PYX_ERR(11, 78, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":81
 *         os_set_inheritable(self._fd)
 * 
 *     def ioctl(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__325 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__325)) __PYX_ERR(11, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__325);
  __Pyx_GIVEREF(__pyx_tuple__325);
  __pyx_codeobj__326 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_ioctl, 81, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__326)) __PYX_ERR(11, 81, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":84
 *         pass
 * 
 *     def getsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */
  __pyx_tuple__327 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_args, __pyx_n_s_kwargs, __pyx_n_s_sock); if (unlikely(!__pyx_tuple__327)) __PYX_ERR(11, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__327);
  __Pyx_GIVEREF(__pyx_tuple__327);
  __pyx_codeobj__328 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__327, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_getsockopt, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__328)) __PYX_ERR(11, 84, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":91
 *             sock.detach()
 * 
 *     def setsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */
  __pyx_codeobj__329 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__327, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_setsockopt, 91, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__329)) __PYX_ERR(11, 91, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":98
 *             sock.detach()
 * 
 *     def getpeername(self):             # <<<<<<<<<<<<<<
 *         if self._peername is not None:
 *             return self._peername
 */
  __pyx_tuple__330 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_sock); if (unlikely(!__pyx_tuple__330)) __PYX_ERR(11, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__330);
  __Pyx_GIVEREF(__pyx_tuple__330);
  __pyx_codeobj__331 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__330, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_getpeername, 98, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__331)) __PYX_ERR(11, 98, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":109
 *             sock.detach()
 * 
 *     def getsockname(self):             # <<<<<<<<<<<<<<
 *         if self._sockname is not None:
 *             return self._sockname
 */
  __pyx_codeobj__332 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__330, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_getsockname, 109, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__332)) __PYX_ERR(11, 109, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":120
 *             sock.detach()
 * 
 *     def share(self, process_id):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */
  __pyx_tuple__333 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_process_id, __pyx_n_s_sock); if (unlikely(!__pyx_tuple__333)) __PYX_ERR(11, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__333);
  __Pyx_GIVEREF(__pyx_tuple__333);
  __pyx_codeobj__334 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__333, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_share, 120, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__334)) __PYX_ERR(11, 120, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":127
 *             sock.detach()
 * 
 *     def accept(self):             # <<<<<<<<<<<<<<
 *         self._na('accept() method')
 * 
 */
  __pyx_codeobj__335 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_accept, 127, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__335)) __PYX_ERR(11, 127, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":130
 *         self._na('accept() method')
 * 
 *     def connect(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect() method')
 * 
 */
  __pyx_tuple__336 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_args); if (unlikely(!__pyx_tuple__336)) __PYX_ERR(11, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__336);
  __Pyx_GIVEREF(__pyx_tuple__336);
  __pyx_codeobj__337 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__336, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_connect, 130, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__337)) __PYX_ERR(11, 130, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":133
 *         self._na('connect() method')
 * 
 *     def connect_ex(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect_ex() method')
 * 
 */
  __pyx_codeobj__338 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__336, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_connect_ex, 133, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__338)) __PYX_ERR(11, 133, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":136
 *         self._na('connect_ex() method')
 * 
 *     def bind(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('bind() method')
 * 
 */
  __pyx_codeobj__339 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__336, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_bind, 136, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__339)) __PYX_ERR(11, 136, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":139
 *         self._na('bind() method')
 * 
 *     def listen(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('listen() method')
 * 
 */
  __pyx_codeobj__340 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_listen, 139, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__340)) __PYX_ERR(11, 139, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":142
 *         self._na('listen() method')
 * 
 *     def makefile(self):             # <<<<<<<<<<<<<<
 *         self._na('makefile() method')
 * 
 */
  __pyx_codeobj__341 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_makefile, 142, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__341)) __PYX_ERR(11, 142, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":145
 *         self._na('makefile() method')
 * 
 *     def sendfile(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendfile() method')
 * 
 */
  __pyx_codeobj__342 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_sendfile, 145, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__342)) __PYX_ERR(11, 145, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":148
 *         self._na('sendfile() method')
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._na('close() method')
 * 
 */
  __pyx_codeobj__343 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_close, 148, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__343)) __PYX_ERR(11, 148, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":151
 *         self._na('close() method')
 * 
 *     def detach(self):             # <<<<<<<<<<<<<<
 *         self._na('detach() method')
 * 
 */
  __pyx_codeobj__344 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_detach, 151, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__344)) __PYX_ERR(11, 151, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":154
 *         self._na('detach() method')
 * 
 *     def shutdown(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('shutdown() method')
 * 
 */
  __pyx_codeobj__345 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__336, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_shutdown, 154, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__345)) __PYX_ERR(11, 154, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":157
 *         self._na('shutdown() method')
 * 
 *     def sendmsg_afalg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg_afalg() method')
 * 
 */
  __pyx_codeobj__346 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_sendmsg_afalg, 157, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__346)) __PYX_ERR(11, 157, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":160
 *         self._na('sendmsg_afalg() method')
 * 
 *     def sendmsg(self):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg() method')
 * 
 */
  __pyx_codeobj__347 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_sendmsg, 160, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__347)) __PYX_ERR(11, 160, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":163
 *         self._na('sendmsg() method')
 * 
 *     def sendto(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendto() method')
 * 
 */
  __pyx_codeobj__348 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_sendto, 163, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__348)) __PYX_ERR(11, 163, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":166
 *         self._na('sendto() method')
 * 
 *     def send(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('send() method')
 * 
 */
  __pyx_codeobj__349 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_send, 166, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__349)) __PYX_ERR(11, 166, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":169
 *         self._na('send() method')
 * 
 *     def sendall(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendall() method')
 * 
 */
  __pyx_codeobj__350 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_sendall, 169, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__350)) __PYX_ERR(11, 169, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":172
 *         self._na('sendall() method')
 * 
 *     def recv_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv_into() method')
 * 
 */
  __pyx_codeobj__351 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_recv_into, 172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__351)) __PYX_ERR(11, 172, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":175
 *         self._na('recv_into() method')
 * 
 *     def recvfrom_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom_into() method')
 * 
 */
  __pyx_codeobj__352 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_recvfrom_into, 175, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__352)) __PYX_ERR(11, 175, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":178
 *         self._na('recvfrom_into() method')
 * 
 *     def recvmsg_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg_into() method')
 * 
 */
  __pyx_codeobj__353 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_recvmsg_into, 178, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__353)) __PYX_ERR(11, 178, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":181
 *         self._na('recvmsg_into() method')
 * 
 *     def recvmsg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg() method')
 * 
 */
  __pyx_codeobj__354 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_recvmsg, 181, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__354)) __PYX_ERR(11, 181, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":184
 *         self._na('recvmsg() method')
 * 
 *     def recvfrom(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom() method')
 * 
 */
  __pyx_codeobj__355 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_recvfrom, 184, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__355)) __PYX_ERR(11, 184, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":187
 *         self._na('recvfrom() method')
 * 
 *     def recv(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv() method')
 * 
 */
  __pyx_codeobj__356 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_recv, 187, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__356)) __PYX_ERR(11, 187, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":190
 *         self._na('recv() method')
 * 
 *     def settimeout(self, value):             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             return
 */
  __pyx_tuple__357 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_value); if (unlikely(!__pyx_tuple__357)) __PYX_ERR(11, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__357);
  __Pyx_GIVEREF(__pyx_tuple__357);
  __pyx_codeobj__358 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__357, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_settimeout, 190, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__358)) __PYX_ERR(11, 190, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":196
 *             'settimeout(): only 0 timeout is allowed on transport sockets')
 * 
 *     def gettimeout(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_codeobj__359 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_gettimeout, 196, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__359)) __PYX_ERR(11, 196, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":199
 *         return 0
 * 
 *     def setblocking(self, flag):             # <<<<<<<<<<<<<<
 *         if not flag:
 *             return
 */
  __pyx_tuple__360 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_flag); if (unlikely(!__pyx_tuple__360)) __PYX_ERR(11, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__360);
  __Pyx_GIVEREF(__pyx_tuple__360);
  __pyx_codeobj__361 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__360, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_setblocking, 199, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__361)) __PYX_ERR(11, 199, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":205
 *             'setblocking(): transport sockets cannot be blocking')
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 * 
 */
  __pyx_codeobj__362 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_enter, 205, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__362)) __PYX_ERR(11, 205, __pyx_L1_error)

  /* "uvloop/pseudosock.pyx":208
 *         self._na('context manager protocol')
 * 
 *     def __exit__(self, *err):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 */
  __pyx_tuple__363 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_err); if (unlikely(!__pyx_tuple__363)) __PYX_ERR(11, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__363);
  __Pyx_GIVEREF(__pyx_tuple__363);
  __pyx_codeobj__364 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__363, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_pseudosock_pyx, __pyx_n_s_exit, 208, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__364)) __PYX_ERR(11, 208, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__365 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__365)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__365);
  __Pyx_GIVEREF(__pyx_tuple__365);
  __pyx_codeobj__366 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__365, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__366)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PseudoSocket__set_state(self, __pyx_state)
 */
  __pyx_codeobj__367 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__367)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__368 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__365, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__368)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LruCache__set_state(self, __pyx_state)
 */
  __pyx_codeobj__369 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__369)) __PYX_ERR(6, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__370 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__370)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__371 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__371)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__372 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__372)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__373 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__373)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__374 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__374)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__375 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__375)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__376 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__376)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__377 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__377)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__378 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__378)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__379 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__379)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__380 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__380)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__381 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__381)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__382 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__382)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__383 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__383)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":230
 *             return bool(not self._is_reading())
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */
  __pyx_codeobj__384 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_get_protocol, 230, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__384)) __PYX_ERR(18, 230, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":233
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._set_protocol(protocol)
 *         if self._is_reading():
 */
  __pyx_tuple__385 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_protocol); if (unlikely(!__pyx_tuple__385)) __PYX_ERR(18, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__385);
  __Pyx_GIVEREF(__pyx_tuple__385);
  __pyx_codeobj__386 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__385, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_set_protocol, 233, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__386)) __PYX_ERR(18, 233, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":239
 *             self._start_reading()
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         # Used by SSLProto.  Might be removed in the future.
 *         if self._conn_lost or self._closed:
 */
  __pyx_tuple__387 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__387)) __PYX_ERR(18, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__387);
  __Pyx_GIVEREF(__pyx_tuple__387);
  __pyx_codeobj__388 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__387, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_force_close, 239, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__388)) __PYX_ERR(18, 239, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":249
 *         self._schedule_call_connection_lost(exc)
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         self._force_close(None)
 * 
 */
  __pyx_codeobj__389 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_abort, 249, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__389)) __PYX_ERR(18, 249, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":252
 *         self._force_close(None)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._closing or self._closed:
 *             return
 */
  __pyx_codeobj__390 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_close, 252, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__390)) __PYX_ERR(18, 252, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":264
 *             self._schedule_call_connection_lost(None)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closing
 * 
 */
  __pyx_codeobj__391 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_is_closing, 264, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__391)) __PYX_ERR(18, 264, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":267
 *         return self._closing
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._get_write_buffer_size()
 * 
 */
  __pyx_codeobj__392 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_get_write_buffer_size, 267, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__392)) __PYX_ERR(18, 267, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":270
 *         return self._get_write_buffer_size()
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_tuple__393 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_high_2, __pyx_n_s_low); if (unlikely(!__pyx_tuple__393)) __PYX_ERR(18, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__393);
  __Pyx_GIVEREF(__pyx_tuple__393);
  __pyx_codeobj__394 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__393, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_set_write_buffer_limits, 270, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__394)) __PYX_ERR(18, 270, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":278
 *         self._maybe_pause_protocol()
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._low_water, self._high_water)
 * 
 */
  __pyx_codeobj__395 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_get_write_buffer_limits, 278, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__395)) __PYX_ERR(18, 278, __pyx_L1_error)

  /* "uvloop/handles/basetransport.pyx":281
 *         return (self._low_water, self._high_water)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if self._extra_info is not None and name in self._extra_info:
 *             return self._extra_info[name]
 */
  __pyx_tuple__396 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_name_2, __pyx_n_s_default); if (unlikely(!__pyx_tuple__396)) __PYX_ERR(18, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__396);
  __Pyx_GIVEREF(__pyx_tuple__396);
  __pyx_codeobj__397 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__396, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_basetransport_pyx, __pyx_n_s_get_extra_info, 281, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__397)) __PYX_ERR(18, 281, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__398 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__398)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__399 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__399)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__400 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__400)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__401 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__401)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":677
 *             id(self))
 * 
 *     def write(self, object buf):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_tuple__402 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_buf); if (unlikely(!__pyx_tuple__402)) __PYX_ERR(20, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__402);
  __Pyx_GIVEREF(__pyx_tuple__402);
  __pyx_codeobj__403 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__402, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_stream_pyx, __pyx_n_s_write, 677, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__403)) __PYX_ERR(20, 677, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":690
 *         self._initiate_write()
 * 
 *     def writelines(self, bufs):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_tuple__404 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_bufs, __pyx_n_s_buf); if (unlikely(!__pyx_tuple__404)) __PYX_ERR(20, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__404);
  __Pyx_GIVEREF(__pyx_tuple__404);
  __pyx_codeobj__405 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__404, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_stream_pyx, __pyx_n_s_writelines, 690, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__405)) __PYX_ERR(20, 690, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":702
 *         self._initiate_write()
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_codeobj__406 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_stream_pyx, __pyx_n_s_write_eof, 702, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__406)) __PYX_ERR(20, 702, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":712
 *             self._shutdown()
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_codeobj__407 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_stream_pyx, __pyx_n_s_can_write_eof, 712, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__407)) __PYX_ERR(20, 712, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":715
 *         return True
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return self._is_reading()
 * 
 */
  __pyx_codeobj__408 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_stream_pyx, __pyx_n_s_is_reading, 715, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__408)) __PYX_ERR(20, 715, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":718
 *         return self._is_reading()
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         if self._closing or not self._is_reading():
 *             return
 */
  __pyx_codeobj__409 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_stream_pyx, __pyx_n_s_pause_reading, 718, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__409)) __PYX_ERR(20, 718, __pyx_L1_error)

  /* "uvloop/handles/stream.pyx":723
 *         self._stop_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         if self._is_reading() or self._closing:
 *             return
 */
  __pyx_codeobj__410 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_stream_pyx, __pyx_n_s_resume_reading, 723, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__410)) __PYX_ERR(20, 723, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__411 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__411)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__412 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__412)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/streamserver.pyx":118
 *         raise NotImplementedError
 * 
 *     def __on_ssl_connected(self, transport, fut):             # <<<<<<<<<<<<<<
 *         exc = fut.exception()
 *         if exc is not None:
 */
  __pyx_tuple__413 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_transport, __pyx_n_s_fut, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__413)) __PYX_ERR(21, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__413);
  __Pyx_GIVEREF(__pyx_tuple__413);
  __pyx_codeobj__414 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__413, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_streamserver_pyx, __pyx_n_s_on_ssl_connected, 118, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__414)) __PYX_ERR(21, 118, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__415 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__415)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__416 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__416)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__417 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__417)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__418 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__418)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/tcp.pyx":157
 *         UVBaseTransport._call_connection_made(self)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if name == 'sockname':
 *             if self.__sockname_set:
 */
  __pyx_codeobj__419 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__396, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_tcp_pyx, __pyx_n_s_get_extra_info, 157, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__419)) __PYX_ERR(3, 157, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__420 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__420)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__421 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__421)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__422 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__422)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__423 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__423)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__424 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__424)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__425 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__425)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__426 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__426)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__427 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__427)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":150
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__428 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_get_write_buffer_limits, 150, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__428)) __PYX_ERR(22, 150, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":153
 *         raise NotImplementedError
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__429 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__393, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_set_write_buffer_limits, 153, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__429)) __PYX_ERR(22, 153, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":156
 *         raise NotImplementedError
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__430 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_get_write_buffer_size, 156, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__430)) __PYX_ERR(22, 156, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":159
 *         raise NotImplementedError
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_tuple__431 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_data); if (unlikely(!__pyx_tuple__431)) __PYX_ERR(22, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__431);
  __Pyx_GIVEREF(__pyx_tuple__431);
  __pyx_codeobj__432 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__431, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_write, 159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__432)) __PYX_ERR(22, 159, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":162
 *         raise NotImplementedError
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_tuple__433 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_list_of_data); if (unlikely(!__pyx_tuple__433)) __PYX_ERR(22, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__433);
  __Pyx_GIVEREF(__pyx_tuple__433);
  __pyx_codeobj__434 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__433, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_writelines, 162, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__434)) __PYX_ERR(22, 162, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":165
 *         raise NotImplementedError
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__435 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_write_eof, 165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__435)) __PYX_ERR(22, 165, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":168
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__436 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_can_write_eof, 168, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__436)) __PYX_ERR(22, 168, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":171
 *         raise NotImplementedError
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__437 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_abort, 171, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__437)) __PYX_ERR(22, 171, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__438 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__438)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__439 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__439)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":202
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__440 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_pause_reading, 202, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__440)) __PYX_ERR(22, 202, __pyx_L1_error)

  /* "uvloop/handles/pipe.pyx":205
 *         raise NotImplementedError
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_codeobj__441 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_pipe_pyx, __pyx_n_s_resume_reading, 205, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__441)) __PYX_ERR(22, 205, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__442 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__442)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__443 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__443)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__444 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__444)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__445 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__445)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__446 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__446)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__447 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__447)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":584
 *                 self._pending_calls.append((_CALL_CONNECTION_LOST, None, None))
 * 
 *     def __stdio_inited(self, waiter, stdio_fut):             # <<<<<<<<<<<<<<
 *         exc = stdio_fut.exception()
 *         if exc is not None:
 */
  __pyx_tuple__448 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_waiter, __pyx_n_s_stdio_fut, __pyx_n_s_exc); if (unlikely(!__pyx_tuple__448)) __PYX_ERR(4, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__448);
  __Pyx_GIVEREF(__pyx_tuple__448);
  __pyx_codeobj__449 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__448, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_stdio_inited, 584, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__449)) __PYX_ERR(4, 584, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":638
 *         return handle
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */
  __pyx_codeobj__450 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_get_protocol, 638, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__450)) __PYX_ERR(4, 638, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":641
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._protocol = protocol
 * 
 */
  __pyx_codeobj__451 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__385, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_set_protocol, 641, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__451)) __PYX_ERR(4, 641, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":644
 *         self._protocol = protocol
 * 
 *     def get_pid(self):             # <<<<<<<<<<<<<<
 *         return self._pid
 * 
 */
  __pyx_codeobj__452 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_get_pid, 644, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__452)) __PYX_ERR(4, 644, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":647
 *         return self._pid
 * 
 *     def get_returncode(self):             # <<<<<<<<<<<<<<
 *         return self._returncode
 * 
 */
  __pyx_codeobj__453 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_get_returncode, 647, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__453)) __PYX_ERR(4, 647, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":650
 *         return self._returncode
 * 
 *     def get_pipe_transport(self, fd):             # <<<<<<<<<<<<<<
 *         if fd == 0:
 *             return self._stdin
 */
  __pyx_tuple__454 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_fd); if (unlikely(!__pyx_tuple__454)) __PYX_ERR(4, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__454);
  __Pyx_GIVEREF(__pyx_tuple__454);
  __pyx_codeobj__455 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__454, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_get_pipe_transport, 650, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__455)) __PYX_ERR(4, 650, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":658
 *             return self._stderr
 * 
 *     def terminate(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGTERM)
 */
  __pyx_codeobj__456 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_terminate, 658, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__456)) __PYX_ERR(4, 658, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":662
 *         self._kill(uv.SIGTERM)
 * 
 *     def kill(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGKILL)
 */
  __pyx_codeobj__457 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_kill, 662, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__457)) __PYX_ERR(4, 662, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":666
 *         self._kill(uv.SIGKILL)
 * 
 *     def send_signal(self, int signal):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(signal)
 */
  __pyx_tuple__458 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_signal); if (unlikely(!__pyx_tuple__458)) __PYX_ERR(4, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__458);
  __Pyx_GIVEREF(__pyx_tuple__458);
  __pyx_codeobj__459 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__458, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_send_signal, 666, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__459)) __PYX_ERR(4, 666, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":670
 *         self._kill(signal)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */
  __pyx_codeobj__460 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_is_closing, 670, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__460)) __PYX_ERR(4, 670, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":673
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._returncode is None:
 *             self._kill(uv.SIGKILL)
 */
  __pyx_codeobj__461 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_close, 673, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__461)) __PYX_ERR(4, 673, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":693
 *             self._close()
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         return default
 * 
 */
  __pyx_codeobj__462 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__396, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_get_extra_info, 693, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__462)) __PYX_ERR(4, 693, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":696
 *         return default
 * 
 *     def _wait(self):             # <<<<<<<<<<<<<<
 *         fut = self._loop._new_future()
 *         if self._returncode is not None:
 */
  __pyx_tuple__463 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_fut); if (unlikely(!__pyx_tuple__463)) __PYX_ERR(4, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__463);
  __Pyx_GIVEREF(__pyx_tuple__463);
  __pyx_codeobj__464 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__463, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_wait_2, 696, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__464)) __PYX_ERR(4, 696, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__465 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__465)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__466 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__466)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":708
 * class WriteSubprocessPipeProto(aio_BaseProtocol):
 * 
 *     def __init__(self, proc, fd):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if type(proc) is not UVProcessTransport:
 */
  __pyx_tuple__467 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_proc, __pyx_n_s_fd); if (unlikely(!__pyx_tuple__467)) __PYX_ERR(4, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__467);
  __Pyx_GIVEREF(__pyx_tuple__467);
  __pyx_codeobj__468 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__467, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_init, 708, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__468)) __PYX_ERR(4, 708, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":719
 *         self.disconnected = False
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.pipe = transport
 * 
 */
  __pyx_tuple__469 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_transport); if (unlikely(!__pyx_tuple__469)) __PYX_ERR(4, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__469);
  __Pyx_GIVEREF(__pyx_tuple__469);
  __pyx_codeobj__470 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__469, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_connection_made, 719, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__470)) __PYX_ERR(4, 719, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":722
 *         self.pipe = transport
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ('<%s fd=%s pipe=%r>'
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 */
  __pyx_codeobj__471 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_repr, 722, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__471)) __PYX_ERR(4, 722, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":726
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.disconnected = True
 *         (<UVProcessTransport>self.proc)._pipe_connection_lost(self.fd, exc)
 */
  __pyx_codeobj__472 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__387, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_connection_lost, 726, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__472)) __PYX_ERR(4, 726, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":731
 *         self.proc = None
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 */
  __pyx_codeobj__473 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_pause_writing, 731, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__473)) __PYX_ERR(4, 731, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":734
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.resume_writing()
 * 
 */
  __pyx_codeobj__474 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_resume_writing, 734, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__474)) __PYX_ERR(4, 734, __pyx_L1_error)

  /* "uvloop/handles/process.pyx":741
 *                               aio_Protocol):
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._pipe_data_received(self.fd, data)
 * 
 */
  __pyx_codeobj__475 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__431, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_process_pyx, __pyx_n_s_data_received, 741, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__475)) __PYX_ERR(4, 741, __pyx_L1_error)

  /* "uvloop/handles/fsevent.pyx":78
 *             UVHandle._close(<UVHandle>self)
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._close()
 * 
 */
  __pyx_codeobj__476 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_fsevent_pyx, __pyx_n_s_cancel, 78, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__476)) __PYX_ERR(23, 78, __pyx_L1_error)

  /* "uvloop/handles/fsevent.pyx":81
 *         self._close()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self.running == 0
 * 
 */
  __pyx_codeobj__477 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_fsevent_pyx, __pyx_n_s_cancelled, 81, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__477)) __PYX_ERR(23, 81, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__478 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__478)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__479 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__479)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__480 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__480)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__481 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__481)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__482 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__482)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 */
  __pyx_codeobj__483 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__483)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__484 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__484)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__485 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__485)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__486 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__486)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__487 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__487)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__488 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__488)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__489 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__489)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":29
 *         self.context = context
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         """Get optional transport information."""
 *         return self._ssl_protocol._get_extra_info(name, default)
 */
  __pyx_codeobj__490 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__396, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_extra_info, 29, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__490)) __PYX_ERR(25, 29, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":33
 *         return self._ssl_protocol._get_extra_info(name, default)
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 */
  __pyx_codeobj__491 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__385, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_set_protocol, 33, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__491)) __PYX_ERR(25, 33, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":36
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._ssl_protocol._app_protocol
 * 
 */
  __pyx_codeobj__492 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_protocol, 36, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__492)) __PYX_ERR(25, 36, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":39
 *         return self._ssl_protocol._app_protocol
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */
  __pyx_codeobj__493 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_is_closing, 39, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__493)) __PYX_ERR(25, 39, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":42
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the transport.
 * 
 */
  __pyx_codeobj__494 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_close, 42, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__494)) __PYX_ERR(25, 42, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":60
 *                 "object>", ResourceWarning)
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return not self._ssl_protocol._app_reading_paused
 * 
 */
  __pyx_codeobj__495 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_is_reading, 60, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__495)) __PYX_ERR(25, 60, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":63
 *         return not self._ssl_protocol._app_reading_paused
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         """Pause the receiving end.
 * 
 */
  __pyx_codeobj__496 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_pause_reading, 63, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__496)) __PYX_ERR(25, 63, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":71
 *         self._ssl_protocol._pause_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         """Resume the receiving end.
 * 
 */
  __pyx_codeobj__497 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_resume_reading, 71, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__497)) __PYX_ERR(25, 71, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":79
 *         self._ssl_protocol._resume_reading(self.context.copy())
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for write flow control.
 * 
 */
  __pyx_codeobj__498 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__393, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_set_write_buffer_limits, 79, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__498)) __PYX_ERR(25, 79, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":101
 *         self._ssl_protocol._control_app_writing(self.context.copy())
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._outgoing_low_water,
 *                 self._ssl_protocol._outgoing_high_water)
 */
  __pyx_codeobj__499 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_write_buffer_limits, 101, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__499)) __PYX_ERR(25, 101, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":105
 *                 self._ssl_protocol._outgoing_high_water)
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the write buffers."""
 *         return self._ssl_protocol._get_write_buffer_size()
 */
  __pyx_codeobj__500 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_write_buffer_size, 105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__500)) __PYX_ERR(25, 105, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":109
 *         return self._ssl_protocol._get_write_buffer_size()
 * 
 *     def set_read_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for read flow control.
 * 
 */
  __pyx_codeobj__501 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__393, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_set_read_buffer_limits, 109, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__501)) __PYX_ERR(25, 109, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":131
 *         self._ssl_protocol._control_ssl_reading()
 * 
 *     def get_read_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._incoming_low_water,
 *                 self._ssl_protocol._incoming_high_water)
 */
  __pyx_codeobj__502 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_read_buffer_limits, 131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__502)) __PYX_ERR(25, 131, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":135
 *                 self._ssl_protocol._incoming_high_water)
 * 
 *     def get_read_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the read buffer."""
 *         return self._ssl_protocol._get_read_buffer_size()
 */
  __pyx_codeobj__503 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_read_buffer_size, 135, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__503)) __PYX_ERR(25, 135, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":144
 *         return self._ssl_protocol._app_writing_paused
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         """Write some data bytes to the transport.
 * 
 */
  __pyx_codeobj__504 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__431, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_write, 144, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__504)) __PYX_ERR(25, 144, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":157
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         """Write a list (or any iterable) of data bytes to the transport.
 * 
 */
  __pyx_codeobj__505 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__433, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_writelines, 157, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__505)) __PYX_ERR(25, 157, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":165
 *         self._ssl_protocol._write_appdata(list_of_data, self.context.copy())
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         """Close the write end after flushing buffered data.
 * 
 */
  __pyx_codeobj__506 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_write_eof, 165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__506)) __PYX_ERR(25, 165, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":172
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         """Return True if this transport supports write_eof(), False if not."""
 *         return False
 */
  __pyx_codeobj__507 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_can_write_eof, 172, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__507)) __PYX_ERR(25, 172, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":176
 *         return False
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         """Close the transport immediately.
 * 
 */
  __pyx_codeobj__508 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_abort, 176, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__508)) __PYX_ERR(25, 176, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":185
 *         self._force_close(None)
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         self._closed = True
 *         self._ssl_protocol._abort(exc)
 */
  __pyx_codeobj__509 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__387, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_force_close, 185, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__509)) __PYX_ERR(25, 185, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":189
 *         self._ssl_protocol._abort(exc)
 * 
 *     def _test__append_write_backlog(self, data):             # <<<<<<<<<<<<<<
 *         # for test only
 *         self._ssl_protocol._write_backlog.append(data)
 */
  __pyx_codeobj__510 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__431, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_test__append_write_backlog, 189, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__510)) __PYX_ERR(25, 189, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__511 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__511)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__512 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__512)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":310
 *         self._waiter = None
 * 
 *     def _get_app_transport(self, context=None):             # <<<<<<<<<<<<<<
 *         if self._app_transport is None:
 *             if self._app_transport_created:
 */
  __pyx_tuple__513 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_context); if (unlikely(!__pyx_tuple__513)) __PYX_ERR(25, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__513);
  __Pyx_GIVEREF(__pyx_tuple__513);
  __pyx_codeobj__514 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__513, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_app_transport, 310, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__514)) __PYX_ERR(25, 310, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":319
 *         return self._app_transport
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is made.
 * 
 */
  __pyx_codeobj__515 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__469, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_connection_made, 319, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__515)) __PYX_ERR(25, 319, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":327
 *         self._start_handshake()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is lost or closed.
 * 
 */
  __pyx_codeobj__516 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__387, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_connection_lost, 327, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__516)) __PYX_ERR(25, 327, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":361
 *             self._handshake_timeout_handle = None
 * 
 *     def get_buffer(self, n):             # <<<<<<<<<<<<<<
 *         cdef size_t want = n
 *         if want > SSL_READ_MAX_SIZE:
 */
  __pyx_tuple__517 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_n, __pyx_n_s_want); if (unlikely(!__pyx_tuple__517)) __PYX_ERR(25, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__517);
  __Pyx_GIVEREF(__pyx_tuple__517);
  __pyx_codeobj__518 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__517, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_get_buffer, 361, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__518)) __PYX_ERR(25, 361, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":374
 *         return self._ssl_buffer_view
 * 
 *     def buffer_updated(self, nbytes):             # <<<<<<<<<<<<<<
 *         self._incoming_write(PyMemoryView_FromMemory(
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))
 */
  __pyx_tuple__519 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_nbytes); if (unlikely(!__pyx_tuple__519)) __PYX_ERR(25, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__519);
  __Pyx_GIVEREF(__pyx_tuple__519);
  __pyx_codeobj__520 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__519, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_buffer_updated, 374, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__520)) __PYX_ERR(25, 374, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":390
 *             self._do_shutdown()
 * 
 *     def eof_received(self):             # <<<<<<<<<<<<<<
 *         """Called when the other end of the low-level stream
 *         is half-closed.
 */
  __pyx_codeobj__521 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_eof_received, 390, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__521)) __PYX_ERR(25, 390, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":911
 *     # Flow control for writes to SSL socket
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer goes over
 *         the high-water mark.
 */
  __pyx_codeobj__522 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_pause_writing, 911, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__522)) __PYX_ERR(25, 911, __pyx_L1_error)

  /* "uvloop/sslproto.pyx":918
 *         self._ssl_writing_paused = True
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer drains below
 *         the low-water mark.
 */
  __pyx_codeobj__523 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_sslproto_pyx, __pyx_n_s_resume_writing, 918, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__523)) __PYX_ERR(25, 918, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__524 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__524)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__525 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__525)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__526 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__526)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__527 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__527)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/handles/udp.pyx":282
 *     # === Public API ===
 * 
 *     def sendto(self, data, addr=None):             # <<<<<<<<<<<<<<
 *         if not data:
 *             # Replicating asyncio logic here.
 */
  __pyx_tuple__528 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_data, __pyx_n_s_addr); if (unlikely(!__pyx_tuple__528)) __PYX_ERR(26, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__528);
  __Pyx_GIVEREF(__pyx_tuple__528);
  __pyx_codeobj__529 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__528, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_handles_udp_pyx, __pyx_n_s_sendto, 282, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__529)) __PYX_ERR(26, 282, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__530 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__530)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__531 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__531)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/server.pyx":51
 *     # Public API
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aenter__(self):
 *         return self
 */
  __pyx_codeobj__189 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_aenter, 51, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__189)) __PYX_ERR(27, 51, __pyx_L1_error)

  /* "uvloop/server.pyx":55
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aexit__(self, *exc):
 *         self.close()
 */
  __pyx_codeobj__190 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__387, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_aexit, 55, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__190)) __PYX_ERR(27, 55, __pyx_L1_error)

  /* "uvloop/server.pyx":63
 *         return '<%s sockets=%r>' % (self.__class__.__name__, self.sockets)
 * 
 *     def get_loop(self):             # <<<<<<<<<<<<<<
 *         return self._loop
 * 
 */
  __pyx_codeobj__532 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_get_loop, 63, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__532)) __PYX_ERR(27, 63, __pyx_L1_error)

  /* "uvloop/server.pyx":66
 *         return self._loop
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def wait_closed(self):
 *         # Do not remove `self._servers is None` below
 */
  __pyx_tuple__533 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_waiter); if (unlikely(!__pyx_tuple__533)) __PYX_ERR(27, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__533);
  __Pyx_GIVEREF(__pyx_tuple__533);
  __pyx_codeobj__191 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__533, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_wait_closed, 66, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__191)) __PYX_ERR(27, 66, __pyx_L1_error)

  /* "uvloop/server.pyx":77
 *         await waiter
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         cdef list servers
 * 
 */
  __pyx_tuple__534 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_servers, __pyx_n_s_server_2); if (unlikely(!__pyx_tuple__534)) __PYX_ERR(27, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__534);
  __Pyx_GIVEREF(__pyx_tuple__534);
  __pyx_codeobj__535 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__534, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_close, 77, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__535)) __PYX_ERR(27, 77, __pyx_L1_error)

  /* "uvloop/server.pyx":96
 *             self._unref()
 * 
 *     def is_serving(self):             # <<<<<<<<<<<<<<
 *         return self._serving
 * 
 */
  __pyx_codeobj__536 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_is_serving, 96, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__536)) __PYX_ERR(27, 96, __pyx_L1_error)

  /* "uvloop/server.pyx":99
 *         return self._serving
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_serving(self):
 *         self._start_serving()
 */
  __pyx_codeobj__192 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_start_serving, 99, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__192)) __PYX_ERR(27, 99, __pyx_L1_error)

  /* "uvloop/server.pyx":103
 *         self._start_serving()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:
 */
  __pyx_codeobj__193 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_server_pyx, __pyx_n_s_serve_forever, 103, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__193)) __PYX_ERR(27, 103, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__537 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__537)) __PYX_ERR(6, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__538 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__538)) __PYX_ERR(6, 3, __pyx_L1_error)

  /* "uvloop/loop.pyx":3422
 * # Helpers for tests
 * 
 * @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 * async def _test_coroutine_1():
 *     return 42
 */
  __pyx_codeobj__194 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_uvloop_loop_pyx, __pyx_n_s_test_coroutine_1, 3422, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__194)) __PYX_ERR(0, 3422, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_PseudoSocket(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__539 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__539)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__539);
  __Pyx_GIVEREF(__pyx_tuple__539);
  __pyx_codeobj__540 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__539, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PseudoSocket, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__540)) __PYX_ERR(6, 1, __pyx_L1_error)
  __pyx_codeobj__541 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__539, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_LruCache, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__541)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_keys.method_name = &__pyx_n_s_keys;
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.method_name = &__pyx_n_s_pop;
  __pyx_umethod_PyList_Type_copy.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PyList_Type_copy.method_name = &__pyx_n_s_copy;
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PyList_Type_pop.method_name = &__pyx_n_s_pop;
  __pyx_umethod_PyUnicode_Type_rstrip.type = (PyObject*)&PyUnicode_Type;
  __pyx_umethod_PyUnicode_Type_rstrip.method_name = &__pyx_n_s_rstrip;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_1 = PyFloat_FromDouble(0.1); if (unlikely(!__pyx_float_0_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1e3 = PyFloat_FromDouble(1e3); if (unlikely(!__pyx_float_1e3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_3 = PyFloat_FromDouble(1e-3); if (unlikely(!__pyx_float_1eneg_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_42 = PyInt_FromLong(42); if (unlikely(!__pyx_int_42)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1000 = PyInt_FromLong(1000); if (unlikely(!__pyx_int_1000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_50000 = PyInt_FromLong(50000L); if (unlikely(!__pyx_int_50000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65536 = PyInt_FromLong(65536L); if (unlikely(!__pyx_int_65536)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1048575 = PyInt_FromLong(1048575L); if (unlikely(!__pyx_int_1048575)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_14808274 = PyInt_FromLong(14808274L); if (unlikely(!__pyx_int_14808274)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_87255255 = PyInt_FromLong(87255255L); if (unlikely(!__pyx_int_87255255)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_121062814 = PyInt_FromLong(121062814L); if (unlikely(!__pyx_int_121062814)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_143395587 = PyInt_FromLong(143395587L); if (unlikely(!__pyx_int_143395587)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_173349396 = PyInt_FromLong(173349396L); if (unlikely(!__pyx_int_173349396)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_200977156 = PyInt_FromLong(200977156L); if (unlikely(!__pyx_int_200977156)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967296 = PyInt_FromString((char *)"4294967296", 0, 0); if (unlikely(!__pyx_int_4294967296)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_6uvloop_4loop_aio_get_event_loop = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_CancelledError = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_InvalidStateError = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_TimeoutError = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_Future = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_Task = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_ensure_future = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_gather = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_wait = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_wrap_future = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_logger = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_iscoroutine = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_iscoroutinefunction = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_BaseProtocol = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_Protocol = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_isfuture = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_get_running_loop = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_set_running_loop = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_debug_wrapper = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_AbstractChildWatcher = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_Transport = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_aio_FlowControlMixin = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_col_deque = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_col_Iterable = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_col_Counter = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_col_OrderedDict = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_cc_ThreadPoolExecutor = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_cc_Future = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_errno_EBADF = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_errno_EINVAL = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ft_partial = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_gc_disable = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_iter_chain = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_inspect_isgenerator = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_gaierror = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_error = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_timeout = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_socket = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_socketpair = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_getservbyname = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_AddressFamily = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_socket_SocketKind = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_environ = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_dup = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_set_inheritable = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_get_inheritable = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_close = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_open = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_devnull = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_O_RDWR = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_pipe = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_read = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_remove = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_stat = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_unlink = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_os_fspath = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_stat_S_ISSOCK = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_ignore_environment = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_dev_mode = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_exc_info = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_set_coroutine_wrapper = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_get_coroutine_wrapper = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_getframe = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_version_info = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_getfilesystemencoding = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sys_platform = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ssl_SSLContext = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ssl_MemoryBIO = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ssl_create_default_context = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ssl_SSLError = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ssl_SSLAgainErrors = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ssl_SSLZeroReturnError = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_ssl_CertificateError = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_threading_Thread = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_threading_main_thread = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_subprocess_SubprocessError = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_signal_signal = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_signal_siginterrupt = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_signal_set_wakeup_fd = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_signal_default_int_handler = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_signal_SIG_DFL = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_time_sleep = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_time_monotonic = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_tb_StackSummary = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_tb_walk_stack = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_tb_format_list = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_warnings_warn = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_weakref_WeakValueDictionary = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_weakref_WeakSet = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_py_inf = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop__LRU_MARKER = Py_None; Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop_sockaddrs = ((struct __pyx_obj_6uvloop_4loop_LruCache *)Py_None); Py_INCREF(Py_None);
  __pyx_v_6uvloop_4loop___forking_loop = ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6uvloop_4loop_UVHandle = &__pyx_vtable_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVHandle._start_init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *, struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_8UVHandle__start_init;
  __pyx_vtable_6uvloop_4loop_UVHandle._abort_init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVHandle__abort_init;
  __pyx_vtable_6uvloop_4loop_UVHandle._finish_init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVHandle__finish_init;
  __pyx_vtable_6uvloop_4loop_UVHandle._is_alive = (int (*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVHandle__is_alive;
  __pyx_vtable_6uvloop_4loop_UVHandle._ensure_alive = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVHandle__ensure_alive;
  __pyx_vtable_6uvloop_4loop_UVHandle._error = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_8UVHandle__error;
  __pyx_vtable_6uvloop_4loop_UVHandle._fatal_error = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *, PyObject *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error *__pyx_optional_args))__pyx_f_6uvloop_4loop_8UVHandle__fatal_error;
  __pyx_vtable_6uvloop_4loop_UVHandle._warn_unclosed = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVHandle__warn_unclosed;
  __pyx_vtable_6uvloop_4loop_UVHandle._free = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVHandle__free;
  __pyx_vtable_6uvloop_4loop_UVHandle._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVHandle__close;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_UVHandle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVHandle_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_UVHandle)) __PYX_ERR(2, 1, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVHandle_spec, __pyx_ptype_6uvloop_4loop_UVHandle) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVHandle = &__pyx_type_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVHandle) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVHandle->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVHandle->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVHandle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVHandle->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVHandle, __pyx_vtabptr_6uvloop_4loop_UVHandle) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVHandle) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVHandle, (PyObject *) __pyx_ptype_6uvloop_4loop_UVHandle) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVHandle) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVSocketHandle = &__pyx_vtable_6uvloop_4loop_UVSocketHandle;
  __pyx_vtable_6uvloop_4loop_UVSocketHandle.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVSocketHandle.__pyx_base._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_14UVSocketHandle__close;
  __pyx_vtable_6uvloop_4loop_UVSocketHandle._fileno = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_14UVSocketHandle__fileno;
  __pyx_vtable_6uvloop_4loop_UVSocketHandle._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_14UVSocketHandle__new_socket;
  __pyx_vtable_6uvloop_4loop_UVSocketHandle._get_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_14UVSocketHandle__get_socket;
  __pyx_vtable_6uvloop_4loop_UVSocketHandle._attach_fileobj = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, PyObject *))__pyx_f_6uvloop_4loop_14UVSocketHandle__attach_fileobj;
  __pyx_vtable_6uvloop_4loop_UVSocketHandle._open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int))__pyx_f_6uvloop_4loop_14UVSocketHandle__open;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVSocketHandle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVSocketHandle_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVSocketHandle)) __PYX_ERR(2, 221, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVSocketHandle_spec, __pyx_ptype_6uvloop_4loop_UVSocketHandle) < 0) __PYX_ERR(2, 221, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVSocketHandle = &__pyx_type_6uvloop_4loop_UVSocketHandle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVSocketHandle->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVSocketHandle) < 0) __PYX_ERR(2, 221, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVSocketHandle->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVSocketHandle->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVSocketHandle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVSocketHandle->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVSocketHandle, __pyx_vtabptr_6uvloop_4loop_UVSocketHandle) < 0) __PYX_ERR(2, 221, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVSocketHandle) < 0) __PYX_ERR(2, 221, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVSocketHandle, (PyObject *) __pyx_ptype_6uvloop_4loop_UVSocketHandle) < 0) __PYX_ERR(2, 221, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVSocketHandle) < 0) __PYX_ERR(2, 221, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVAsync = &__pyx_vtable_6uvloop_4loop_UVAsync;
  __pyx_vtable_6uvloop_4loop_UVAsync.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVAsync._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVAsync *, struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *))__pyx_f_6uvloop_4loop_7UVAsync__init;
  __pyx_vtable_6uvloop_4loop_UVAsync.send = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVAsync *))__pyx_f_6uvloop_4loop_7UVAsync_send;
  __pyx_vtable_6uvloop_4loop_UVAsync.new = (struct __pyx_obj_6uvloop_4loop_UVAsync *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *))__pyx_f_6uvloop_4loop_7UVAsync_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVAsync = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVAsync_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVAsync)) __PYX_ERR(13, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVAsync_spec, __pyx_ptype_6uvloop_4loop_UVAsync) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVAsync = &__pyx_type_6uvloop_4loop_UVAsync;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVAsync->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVAsync) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVAsync->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVAsync->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVAsync->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVAsync->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVAsync, __pyx_vtabptr_6uvloop_4loop_UVAsync) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVAsync) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVAsync, (PyObject *) __pyx_ptype_6uvloop_4loop_UVAsync) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVAsync) < 0) __PYX_ERR(13, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVTimer = &__pyx_vtable_6uvloop_4loop_UVTimer;
  __pyx_vtable_6uvloop_4loop_UVTimer.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVTimer._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVTimer *, struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *, uint64_t))__pyx_f_6uvloop_4loop_7UVTimer__init;
  __pyx_vtable_6uvloop_4loop_UVTimer.stop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVTimer *))__pyx_f_6uvloop_4loop_7UVTimer_stop;
  __pyx_vtable_6uvloop_4loop_UVTimer.start = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVTimer *))__pyx_f_6uvloop_4loop_7UVTimer_start;
  __pyx_vtable_6uvloop_4loop_UVTimer.get_when = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVTimer *))__pyx_f_6uvloop_4loop_7UVTimer_get_when;
  __pyx_vtable_6uvloop_4loop_UVTimer.new = (struct __pyx_obj_6uvloop_4loop_UVTimer *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, __pyx_t_6uvloop_4loop_method_t, PyObject *, uint64_t))__pyx_f_6uvloop_4loop_7UVTimer_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVTimer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVTimer_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVTimer)) __PYX_ERR(16, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVTimer_spec, __pyx_ptype_6uvloop_4loop_UVTimer) < 0) __PYX_ERR(16, 2, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVTimer = &__pyx_type_6uvloop_4loop_UVTimer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVTimer->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVTimer) < 0) __PYX_ERR(16, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVTimer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVTimer->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVTimer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVTimer->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVTimer, __pyx_vtabptr_6uvloop_4loop_UVTimer) < 0) __PYX_ERR(16, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVTimer) < 0) __PYX_ERR(16, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVTimer, (PyObject *) __pyx_ptype_6uvloop_4loop_UVTimer) < 0) __PYX_ERR(16, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVTimer) < 0) __PYX_ERR(16, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVIdle = &__pyx_vtable_6uvloop_4loop_UVIdle;
  __pyx_vtable_6uvloop_4loop_UVIdle.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVIdle._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVIdle *, struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_6UVIdle__init;
  __pyx_vtable_6uvloop_4loop_UVIdle.stop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVIdle *))__pyx_f_6uvloop_4loop_6UVIdle_stop;
  __pyx_vtable_6uvloop_4loop_UVIdle.start = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVIdle *))__pyx_f_6uvloop_4loop_6UVIdle_start;
  __pyx_vtable_6uvloop_4loop_UVIdle.new = (struct __pyx_obj_6uvloop_4loop_UVIdle *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_6UVIdle_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVIdle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVIdle_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVIdle)) __PYX_ERR(14, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVIdle_spec, __pyx_ptype_6uvloop_4loop_UVIdle) < 0) __PYX_ERR(14, 2, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVIdle = &__pyx_type_6uvloop_4loop_UVIdle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVIdle->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVIdle) < 0) __PYX_ERR(14, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVIdle->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVIdle->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVIdle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVIdle->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVIdle, __pyx_vtabptr_6uvloop_4loop_UVIdle) < 0) __PYX_ERR(14, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVIdle) < 0) __PYX_ERR(14, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVIdle, (PyObject *) __pyx_ptype_6uvloop_4loop_UVIdle) < 0) __PYX_ERR(14, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVIdle) < 0) __PYX_ERR(14, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVBaseTransport = &__pyx_vtable_6uvloop_4loop_UVBaseTransport;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVSocketHandle;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport.__pyx_base.__pyx_base._fatal_error = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *, PyObject *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error *__pyx_optional_args))__pyx_f_6uvloop_4loop_15UVBaseTransport__fatal_error;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._maybe_pause_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_pause_protocol;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._maybe_resume_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__maybe_resume_protocol;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._schedule_call_connection_made = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_made;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._schedule_call_connection_lost = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *))__pyx_f_6uvloop_4loop_15UVBaseTransport__schedule_call_connection_lost;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._wakeup_waiter = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__wakeup_waiter;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._call_connection_made = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__call_connection_made;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._call_connection_lost = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *))__pyx_f_6uvloop_4loop_15UVBaseTransport__call_connection_lost;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._set_server = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, struct __pyx_obj_6uvloop_4loop_Server *))__pyx_f_6uvloop_4loop_15UVBaseTransport__set_server;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._set_waiter = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *))__pyx_f_6uvloop_4loop_15UVBaseTransport__set_waiter;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._set_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *))__pyx_f_6uvloop_4loop_15UVBaseTransport__set_protocol;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._clear_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__clear_protocol;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._init_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__init_protocol;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._add_extra_info = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_15UVBaseTransport__add_extra_info;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._get_write_buffer_size = (size_t (*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__get_write_buffer_size;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._is_reading = (int (*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__is_reading;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._start_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__start_reading;
  __pyx_vtable_6uvloop_4loop_UVBaseTransport._stop_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_15UVBaseTransport__stop_reading;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVSocketHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(18, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVBaseTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVBaseTransport_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVBaseTransport)) __PYX_ERR(18, 1, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVBaseTransport_spec, __pyx_ptype_6uvloop_4loop_UVBaseTransport) < 0) __PYX_ERR(18, 1, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVBaseTransport = &__pyx_type_6uvloop_4loop_UVBaseTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVBaseTransport->tp_base = __pyx_ptype_6uvloop_4loop_UVSocketHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVBaseTransport) < 0) __PYX_ERR(18, 1, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVBaseTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVBaseTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVBaseTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVBaseTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_vtabptr_6uvloop_4loop_UVBaseTransport) < 0) __PYX_ERR(18, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVBaseTransport) < 0) __PYX_ERR(18, 1, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVBaseTransport, (PyObject *) __pyx_ptype_6uvloop_4loop_UVBaseTransport) < 0) __PYX_ERR(18, 1, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_UVBaseTransport->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_UVBaseTransport->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_UVBaseTransport, __weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVBaseTransport) < 0) __PYX_ERR(18, 1, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_Loop = &__pyx_vtable_6uvloop_4loop_Loop;
  __pyx_vtable_6uvloop_4loop_Loop._init_debug_fields = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__init_debug_fields;
  __pyx_vtable_6uvloop_4loop_Loop._on_wake = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__on_wake;
  __pyx_vtable_6uvloop_4loop_Loop._on_idle = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__on_idle;
  __pyx_vtable_6uvloop_4loop_Loop._Loop__run = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, uv_run_mode))__pyx_f_6uvloop_4loop_4Loop__Loop__run;
  __pyx_vtable_6uvloop_4loop_Loop._run = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, uv_run_mode))__pyx_f_6uvloop_4loop_4Loop__run;
  __pyx_vtable_6uvloop_4loop_Loop._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__close;
  __pyx_vtable_6uvloop_4loop_Loop._stop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__stop;
  __pyx_vtable_6uvloop_4loop_Loop._time = (uint64_t (*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__time;
  __pyx_vtable_6uvloop_4loop_Loop._queue_write = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_4Loop__queue_write;
  __pyx_vtable_6uvloop_4loop_Loop._exec_queued_writes = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__exec_queued_writes;
  __pyx_vtable_6uvloop_4loop_Loop._call_soon = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__call_soon;
  __pyx_vtable_6uvloop_4loop_Loop._append_ready_handle = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_4Loop__append_ready_handle;
  __pyx_vtable_6uvloop_4loop_Loop._call_soon_handle = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_4Loop__call_soon_handle;
  __pyx_vtable_6uvloop_4loop_Loop._call_later = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, uint64_t, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__call_later;
  __pyx_vtable_6uvloop_4loop_Loop._handle_exception = (void (*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__handle_exception;
  __pyx_vtable_6uvloop_4loop_Loop._is_main_thread = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__is_main_thread;
  __pyx_vtable_6uvloop_4loop_Loop._new_future = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__new_future;
  __pyx_vtable_6uvloop_4loop_Loop._check_signal = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__check_signal;
  __pyx_vtable_6uvloop_4loop_Loop._check_closed = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__check_closed;
  __pyx_vtable_6uvloop_4loop_Loop._check_thread = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__check_thread;
  __pyx_vtable_6uvloop_4loop_Loop._getaddrinfo = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, int, int, int, int, int))__pyx_f_6uvloop_4loop_4Loop__getaddrinfo;
  __pyx_vtable_6uvloop_4loop_Loop._getnameinfo = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct sockaddr *, int))__pyx_f_6uvloop_4loop_4Loop__getnameinfo;
  __pyx_vtable_6uvloop_4loop_Loop._track_transport = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_4Loop__track_transport;
  __pyx_vtable_6uvloop_4loop_Loop._fileobj_to_fd = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__fileobj_to_fd;
  __pyx_vtable_6uvloop_4loop_Loop._ensure_fd_no_transport = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__ensure_fd_no_transport;
  __pyx_vtable_6uvloop_4loop_Loop._track_process = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVProcess *))__pyx_f_6uvloop_4loop_4Loop__track_process;
  __pyx_vtable_6uvloop_4loop_Loop._untrack_process = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_UVProcess *))__pyx_f_6uvloop_4loop_4Loop__untrack_process;
  __pyx_vtable_6uvloop_4loop_Loop._add_reader = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_4Loop__add_reader;
  __pyx_vtable_6uvloop_4loop_Loop._has_reader = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__has_reader;
  __pyx_vtable_6uvloop_4loop_Loop._remove_reader = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__remove_reader;
  __pyx_vtable_6uvloop_4loop_Loop._add_writer = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_4Loop__add_writer;
  __pyx_vtable_6uvloop_4loop_Loop._has_writer = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__has_writer;
  __pyx_vtable_6uvloop_4loop_Loop._remove_writer = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__remove_writer;
  __pyx_vtable_6uvloop_4loop_Loop._sock_recv = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__sock_recv;
  __pyx_vtable_6uvloop_4loop_Loop._sock_recv_into = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__sock_recv_into;
  __pyx_vtable_6uvloop_4loop_Loop._sock_sendall = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__sock_sendall;
  __pyx_vtable_6uvloop_4loop_Loop._sock_accept = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__sock_accept;
  __pyx_vtable_6uvloop_4loop_Loop._sock_connect = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__sock_connect;
  __pyx_vtable_6uvloop_4loop_Loop._sock_connect_cb = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__sock_connect_cb;
  __pyx_vtable_6uvloop_4loop_Loop._sock_set_reuseport = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, int))__pyx_f_6uvloop_4loop_4Loop__sock_set_reuseport;
  __pyx_vtable_6uvloop_4loop_Loop._setup_or_resume_signals = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__setup_or_resume_signals;
  __pyx_vtable_6uvloop_4loop_Loop._shutdown_signals = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__shutdown_signals;
  __pyx_vtable_6uvloop_4loop_Loop._pause_signals = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__pause_signals;
  __pyx_vtable_6uvloop_4loop_Loop._handle_signal = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__handle_signal;
  __pyx_vtable_6uvloop_4loop_Loop._read_from_self = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__read_from_self;
  __pyx_vtable_6uvloop_4loop_Loop._ceval_process_signals = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__ceval_process_signals;
  __pyx_vtable_6uvloop_4loop_Loop._invoke_signals = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *))__pyx_f_6uvloop_4loop_4Loop__invoke_signals;
  __pyx_vtable_6uvloop_4loop_Loop._set_coroutine_debug = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, int))__pyx_f_6uvloop_4loop_4Loop__set_coroutine_debug;
  __pyx_vtable_6uvloop_4loop_Loop._print_debug_info = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_4Loop__print_debug_info;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_Loop = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_Loop_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_Loop)) __PYX_ERR(0, 128, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_Loop_spec, __pyx_ptype_6uvloop_4loop_Loop) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_Loop = &__pyx_type_6uvloop_4loop_Loop;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_Loop) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_Loop->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_Loop->tp_dictoffset && __pyx_ptype_6uvloop_4loop_Loop->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_Loop->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_Loop, __pyx_vtabptr_6uvloop_4loop_Loop) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_Loop) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Loop, (PyObject *) __pyx_ptype_6uvloop_4loop_Loop) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_Loop->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_Loop->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_Loop, __weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_Loop) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_Handle = &__pyx_vtable_6uvloop_4loop_Handle;
  __pyx_vtable_6uvloop_4loop_Handle._set_loop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Handle *, struct __pyx_obj_6uvloop_4loop_Loop *))__pyx_f_6uvloop_4loop_6Handle__set_loop;
  __pyx_vtable_6uvloop_4loop_Handle._set_context = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Handle *, PyObject *))__pyx_f_6uvloop_4loop_6Handle__set_context;
  __pyx_vtable_6uvloop_4loop_Handle._run = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_6Handle__run;
  __pyx_vtable_6uvloop_4loop_Handle._cancel = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_6Handle__cancel;
  __pyx_vtable_6uvloop_4loop_Handle._format_handle = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_6Handle__format_handle;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_Handle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_Handle_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_Handle)) __PYX_ERR(10, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_Handle_spec, __pyx_ptype_6uvloop_4loop_Handle) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_Handle = &__pyx_type_6uvloop_4loop_Handle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_Handle) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_Handle->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_Handle->tp_dictoffset && __pyx_ptype_6uvloop_4loop_Handle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_Handle->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_Handle, __pyx_vtabptr_6uvloop_4loop_Handle) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_Handle) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Handle, (PyObject *) __pyx_ptype_6uvloop_4loop_Handle) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_Handle->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_Handle->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_Handle, __weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_Handle) < 0) __PYX_ERR(10, 3, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_TimerHandle = &__pyx_vtable_6uvloop_4loop_TimerHandle;
  __pyx_vtable_6uvloop_4loop_TimerHandle._run = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_TimerHandle *))__pyx_f_6uvloop_4loop_11TimerHandle__run;
  __pyx_vtable_6uvloop_4loop_TimerHandle._cancel = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_TimerHandle *))__pyx_f_6uvloop_4loop_11TimerHandle__cancel;
  __pyx_vtable_6uvloop_4loop_TimerHandle._clear = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_TimerHandle *))__pyx_f_6uvloop_4loop_11TimerHandle__clear;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_TimerHandle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_TimerHandle_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_TimerHandle)) __PYX_ERR(10, 167, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_TimerHandle_spec, __pyx_ptype_6uvloop_4loop_TimerHandle) < 0) __PYX_ERR(10, 167, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_TimerHandle = &__pyx_type_6uvloop_4loop_TimerHandle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_TimerHandle) < 0) __PYX_ERR(10, 167, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_TimerHandle->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_TimerHandle->tp_dictoffset && __pyx_ptype_6uvloop_4loop_TimerHandle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_TimerHandle->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_TimerHandle, __pyx_vtabptr_6uvloop_4loop_TimerHandle) < 0) __PYX_ERR(10, 167, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_TimerHandle) < 0) __PYX_ERR(10, 167, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TimerHandle, (PyObject *) __pyx_ptype_6uvloop_4loop_TimerHandle) < 0) __PYX_ERR(10, 167, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_TimerHandle->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_TimerHandle->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_TimerHandle, __weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_TimerHandle) < 0) __PYX_ERR(10, 167, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVCheck = &__pyx_vtable_6uvloop_4loop_UVCheck;
  __pyx_vtable_6uvloop_4loop_UVCheck.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVCheck._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVCheck *, struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_7UVCheck__init;
  __pyx_vtable_6uvloop_4loop_UVCheck.stop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVCheck *))__pyx_f_6uvloop_4loop_7UVCheck_stop;
  __pyx_vtable_6uvloop_4loop_UVCheck.start = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVCheck *))__pyx_f_6uvloop_4loop_7UVCheck_start;
  __pyx_vtable_6uvloop_4loop_UVCheck.new = (struct __pyx_obj_6uvloop_4loop_UVCheck *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_7UVCheck_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVCheck = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVCheck_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVCheck)) __PYX_ERR(15, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVCheck_spec, __pyx_ptype_6uvloop_4loop_UVCheck) < 0) __PYX_ERR(15, 2, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVCheck = &__pyx_type_6uvloop_4loop_UVCheck;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVCheck->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVCheck) < 0) __PYX_ERR(15, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVCheck->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVCheck->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVCheck->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVCheck->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVCheck, __pyx_vtabptr_6uvloop_4loop_UVCheck) < 0) __PYX_ERR(15, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVCheck) < 0) __PYX_ERR(15, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVCheck, (PyObject *) __pyx_ptype_6uvloop_4loop_UVCheck) < 0) __PYX_ERR(15, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVCheck) < 0) __PYX_ERR(15, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVPoll = &__pyx_vtable_6uvloop_4loop_UVPoll;
  __pyx_vtable_6uvloop_4loop_UVPoll.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVPoll.__pyx_base._fatal_error = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *, PyObject *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error *__pyx_optional_args))__pyx_f_6uvloop_4loop_6UVPoll__fatal_error;
  __pyx_vtable_6uvloop_4loop_UVPoll.__pyx_base._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_6UVPoll__close;
  __pyx_vtable_6uvloop_4loop_UVPoll._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *, struct __pyx_obj_6uvloop_4loop_Loop *, int))__pyx_f_6uvloop_4loop_6UVPoll__init;
  __pyx_vtable_6uvloop_4loop_UVPoll._poll_start = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *, int))__pyx_f_6uvloop_4loop_6UVPoll__poll_start;
  __pyx_vtable_6uvloop_4loop_UVPoll._poll_stop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *))__pyx_f_6uvloop_4loop_6UVPoll__poll_stop;
  __pyx_vtable_6uvloop_4loop_UVPoll.is_active = (int (*)(struct __pyx_obj_6uvloop_4loop_UVPoll *))__pyx_f_6uvloop_4loop_6UVPoll_is_active;
  __pyx_vtable_6uvloop_4loop_UVPoll.is_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *))__pyx_f_6uvloop_4loop_6UVPoll_is_reading;
  __pyx_vtable_6uvloop_4loop_UVPoll.is_writing = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *))__pyx_f_6uvloop_4loop_6UVPoll_is_writing;
  __pyx_vtable_6uvloop_4loop_UVPoll.start_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_6UVPoll_start_reading;
  __pyx_vtable_6uvloop_4loop_UVPoll.start_writing = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *, struct __pyx_obj_6uvloop_4loop_Handle *))__pyx_f_6uvloop_4loop_6UVPoll_start_writing;
  __pyx_vtable_6uvloop_4loop_UVPoll.stop_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *))__pyx_f_6uvloop_4loop_6UVPoll_stop_reading;
  __pyx_vtable_6uvloop_4loop_UVPoll.stop_writing = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *))__pyx_f_6uvloop_4loop_6UVPoll_stop_writing;
  __pyx_vtable_6uvloop_4loop_UVPoll.stop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVPoll *))__pyx_f_6uvloop_4loop_6UVPoll_stop;
  __pyx_vtable_6uvloop_4loop_UVPoll.new = (struct __pyx_obj_6uvloop_4loop_UVPoll *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, int))__pyx_f_6uvloop_4loop_6UVPoll_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVPoll = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVPoll_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVPoll)) __PYX_ERR(17, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVPoll_spec, __pyx_ptype_6uvloop_4loop_UVPoll) < 0) __PYX_ERR(17, 2, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVPoll = &__pyx_type_6uvloop_4loop_UVPoll;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVPoll->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVPoll) < 0) __PYX_ERR(17, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVPoll->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVPoll->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVPoll->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVPoll->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVPoll, __pyx_vtabptr_6uvloop_4loop_UVPoll) < 0) __PYX_ERR(17, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVPoll) < 0) __PYX_ERR(17, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVPoll, (PyObject *) __pyx_ptype_6uvloop_4loop_UVPoll) < 0) __PYX_ERR(17, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVPoll) < 0) __PYX_ERR(17, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVStream = &__pyx_vtable_6uvloop_4loop_UVStream;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVBaseTransport;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base.__pyx_base.__pyx_base._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_8UVStream__close;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base._set_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *, PyObject *))__pyx_f_6uvloop_4loop_8UVStream__set_protocol;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base._clear_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_8UVStream__clear_protocol;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base._get_write_buffer_size = (size_t (*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_8UVStream__get_write_buffer_size;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base._is_reading = (int (*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_8UVStream__is_reading;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base._start_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_8UVStream__start_reading;
  __pyx_vtable_6uvloop_4loop_UVStream.__pyx_base._stop_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_8UVStream__stop_reading;
  __pyx_vtable_6uvloop_4loop_UVStream._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_8UVStream__init;
  __pyx_vtable_6uvloop_4loop_UVStream._shutdown = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__shutdown;
  __pyx_vtable_6uvloop_4loop_UVStream._accept = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *, struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__accept;
  __pyx_vtable_6uvloop_4loop_UVStream._close_on_read_error = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__close_on_read_error;
  __pyx_vtable_6uvloop_4loop_UVStream._UVStream__reading_started = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_started;
  __pyx_vtable_6uvloop_4loop_UVStream._UVStream__reading_stopped = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__UVStream__reading_stopped;
  __pyx_vtable_6uvloop_4loop_UVStream._buffer_write = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *))__pyx_f_6uvloop_4loop_8UVStream__buffer_write;
  __pyx_vtable_6uvloop_4loop_UVStream._initiate_write = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__initiate_write;
  __pyx_vtable_6uvloop_4loop_UVStream._exec_write = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__exec_write;
  __pyx_vtable_6uvloop_4loop_UVStream._try_write = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *))__pyx_f_6uvloop_4loop_8UVStream__try_write;
  __pyx_vtable_6uvloop_4loop_UVStream._on_accept = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__on_accept;
  __pyx_vtable_6uvloop_4loop_UVStream._on_eof = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__on_eof;
  __pyx_vtable_6uvloop_4loop_UVStream._on_write = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *))__pyx_f_6uvloop_4loop_8UVStream__on_write;
  __pyx_vtable_6uvloop_4loop_UVStream._on_connect = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *))__pyx_f_6uvloop_4loop_8UVStream__on_connect;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport); if (unlikely(!__pyx_t_1)) __PYX_ERR(20, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVStream = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVStream_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVStream)) __PYX_ERR(20, 213, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVStream_spec, __pyx_ptype_6uvloop_4loop_UVStream) < 0) __PYX_ERR(20, 213, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVStream = &__pyx_type_6uvloop_4loop_UVStream;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVStream->tp_base = __pyx_ptype_6uvloop_4loop_UVBaseTransport;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVStream) < 0) __PYX_ERR(20, 213, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVStream->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVStream->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVStream->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVStream->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVStream, __pyx_vtabptr_6uvloop_4loop_UVStream) < 0) __PYX_ERR(20, 213, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVStream) < 0) __PYX_ERR(20, 213, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVStream, (PyObject *) __pyx_ptype_6uvloop_4loop_UVStream) < 0) __PYX_ERR(20, 213, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_UVStream->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_UVStream->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_UVStream, __pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVStream) < 0) __PYX_ERR(20, 213, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVStreamServer = &__pyx_vtable_6uvloop_4loop_UVStreamServer;
  __pyx_vtable_6uvloop_4loop_UVStreamServer.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVSocketHandle;
  __pyx_vtable_6uvloop_4loop_UVStreamServer.__pyx_base.__pyx_base._fatal_error = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *, PyObject *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_8UVHandle__fatal_error *__pyx_optional_args))__pyx_f_6uvloop_4loop_14UVStreamServer__fatal_error;
  __pyx_vtable_6uvloop_4loop_UVStreamServer._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_14UVStreamServer__init;
  __pyx_vtable_6uvloop_4loop_UVStreamServer._mark_as_open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *))__pyx_f_6uvloop_4loop_14UVStreamServer__mark_as_open;
  __pyx_vtable_6uvloop_4loop_UVStreamServer.listen = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *))__pyx_f_6uvloop_4loop_14UVStreamServer_listen;
  __pyx_vtable_6uvloop_4loop_UVStreamServer._on_listen = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *))__pyx_f_6uvloop_4loop_14UVStreamServer__on_listen;
  __pyx_vtable_6uvloop_4loop_UVStreamServer._make_new_transport = (struct __pyx_obj_6uvloop_4loop_UVStream *(*)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_14UVStreamServer__make_new_transport;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVSocketHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(21, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVStreamServer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVStreamServer_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVStreamServer)) __PYX_ERR(21, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVStreamServer_spec, __pyx_ptype_6uvloop_4loop_UVStreamServer) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVStreamServer = &__pyx_type_6uvloop_4loop_UVStreamServer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVStreamServer->tp_base = __pyx_ptype_6uvloop_4loop_UVSocketHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVStreamServer) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVStreamServer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVStreamServer->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVStreamServer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVStreamServer->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVStreamServer, __pyx_vtabptr_6uvloop_4loop_UVStreamServer) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVStreamServer) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVStreamServer, (PyObject *) __pyx_ptype_6uvloop_4loop_UVStreamServer) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVStreamServer) < 0) __PYX_ERR(21, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_TCPServer = &__pyx_vtable_6uvloop_4loop_TCPServer;
  __pyx_vtable_6uvloop_4loop_TCPServer.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVStreamServer;
  __pyx_vtable_6uvloop_4loop_TCPServer.__pyx_base.__pyx_base._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_9TCPServer__new_socket;
  __pyx_vtable_6uvloop_4loop_TCPServer.__pyx_base.__pyx_base._open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int))__pyx_f_6uvloop_4loop_9TCPServer__open;
  __pyx_vtable_6uvloop_4loop_TCPServer.__pyx_base._make_new_transport = (struct __pyx_obj_6uvloop_4loop_UVStream *(*)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_9TCPServer__make_new_transport;
  __pyx_vtable_6uvloop_4loop_TCPServer.bind = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_TCPServer *, struct sockaddr *, struct __pyx_opt_args_6uvloop_4loop_9TCPServer_bind *__pyx_optional_args))__pyx_f_6uvloop_4loop_9TCPServer_bind;
  __pyx_vtable_6uvloop_4loop_TCPServer.new = (struct __pyx_obj_6uvloop_4loop_TCPServer *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, unsigned int, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_9TCPServer_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVStreamServer); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_TCPServer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_TCPServer_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_TCPServer)) __PYX_ERR(3, 56, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_TCPServer_spec, __pyx_ptype_6uvloop_4loop_TCPServer) < 0) __PYX_ERR(3, 56, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_TCPServer = &__pyx_type_6uvloop_4loop_TCPServer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_TCPServer->tp_base = __pyx_ptype_6uvloop_4loop_UVStreamServer;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_TCPServer) < 0) __PYX_ERR(3, 56, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_TCPServer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_TCPServer->tp_dictoffset && __pyx_ptype_6uvloop_4loop_TCPServer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_TCPServer->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_TCPServer, __pyx_vtabptr_6uvloop_4loop_TCPServer) < 0) __PYX_ERR(3, 56, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_TCPServer) < 0) __PYX_ERR(3, 56, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TCPServer, (PyObject *) __pyx_ptype_6uvloop_4loop_TCPServer) < 0) __PYX_ERR(3, 56, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_TCPServer) < 0) __PYX_ERR(3, 56, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_TCPTransport = &__pyx_vtable_6uvloop_4loop_TCPTransport;
  __pyx_vtable_6uvloop_4loop_TCPTransport.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVStream;
  __pyx_vtable_6uvloop_4loop_TCPTransport.__pyx_base.__pyx_base.__pyx_base._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_12TCPTransport__new_socket;
  __pyx_vtable_6uvloop_4loop_TCPTransport.__pyx_base.__pyx_base.__pyx_base._open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int))__pyx_f_6uvloop_4loop_12TCPTransport__open;
  __pyx_vtable_6uvloop_4loop_TCPTransport.__pyx_base.__pyx_base._call_connection_made = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_12TCPTransport__call_connection_made;
  __pyx_vtable_6uvloop_4loop_TCPTransport.bind = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_TCPTransport *, struct sockaddr *, struct __pyx_opt_args_6uvloop_4loop_12TCPTransport_bind *__pyx_optional_args))__pyx_f_6uvloop_4loop_12TCPTransport_bind;
  __pyx_vtable_6uvloop_4loop_TCPTransport.connect = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_TCPTransport *, struct sockaddr *))__pyx_f_6uvloop_4loop_12TCPTransport_connect;
  __pyx_vtable_6uvloop_4loop_TCPTransport._set_nodelay = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_TCPTransport *))__pyx_f_6uvloop_4loop_12TCPTransport__set_nodelay;
  __pyx_vtable_6uvloop_4loop_TCPTransport.new = (struct __pyx_obj_6uvloop_4loop_TCPTransport *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_12TCPTransport_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVStream); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_TCPTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_TCPTransport_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_TCPTransport)) __PYX_ERR(3, 103, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_TCPTransport_spec, __pyx_ptype_6uvloop_4loop_TCPTransport) < 0) __PYX_ERR(3, 103, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_TCPTransport = &__pyx_type_6uvloop_4loop_TCPTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_TCPTransport->tp_base = __pyx_ptype_6uvloop_4loop_UVStream;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_TCPTransport) < 0) __PYX_ERR(3, 103, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_TCPTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_TCPTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop_TCPTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_TCPTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_TCPTransport, __pyx_vtabptr_6uvloop_4loop_TCPTransport) < 0) __PYX_ERR(3, 103, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_TCPTransport) < 0) __PYX_ERR(3, 103, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TCPTransport, (PyObject *) __pyx_ptype_6uvloop_4loop_TCPTransport) < 0) __PYX_ERR(3, 103, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_TCPTransport->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_TCPTransport->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_TCPTransport, __pyx_base.__pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_TCPTransport) < 0) __PYX_ERR(3, 103, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UnixServer = &__pyx_vtable_6uvloop_4loop_UnixServer;
  __pyx_vtable_6uvloop_4loop_UnixServer.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVStreamServer;
  __pyx_vtable_6uvloop_4loop_UnixServer.__pyx_base.__pyx_base.__pyx_base._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_10UnixServer__close;
  __pyx_vtable_6uvloop_4loop_UnixServer.__pyx_base.__pyx_base._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_10UnixServer__new_socket;
  __pyx_vtable_6uvloop_4loop_UnixServer.__pyx_base.__pyx_base._open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int))__pyx_f_6uvloop_4loop_10UnixServer__open;
  __pyx_vtable_6uvloop_4loop_UnixServer.__pyx_base._make_new_transport = (struct __pyx_obj_6uvloop_4loop_UVStream *(*)(struct __pyx_obj_6uvloop_4loop_UVStreamServer *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_10UnixServer__make_new_transport;
  __pyx_vtable_6uvloop_4loop_UnixServer.bind = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UnixServer *, PyObject *))__pyx_f_6uvloop_4loop_10UnixServer_bind;
  __pyx_vtable_6uvloop_4loop_UnixServer.new = (struct __pyx_obj_6uvloop_4loop_UnixServer *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_10UnixServer_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVStreamServer); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UnixServer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UnixServer_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UnixServer)) __PYX_ERR(22, 40, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UnixServer_spec, __pyx_ptype_6uvloop_4loop_UnixServer) < 0) __PYX_ERR(22, 40, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UnixServer = &__pyx_type_6uvloop_4loop_UnixServer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UnixServer->tp_base = __pyx_ptype_6uvloop_4loop_UVStreamServer;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UnixServer) < 0) __PYX_ERR(22, 40, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UnixServer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UnixServer->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UnixServer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UnixServer->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UnixServer, __pyx_vtabptr_6uvloop_4loop_UnixServer) < 0) __PYX_ERR(22, 40, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UnixServer) < 0) __PYX_ERR(22, 40, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UnixServer, (PyObject *) __pyx_ptype_6uvloop_4loop_UnixServer) < 0) __PYX_ERR(22, 40, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UnixServer) < 0) __PYX_ERR(22, 40, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UnixTransport = &__pyx_vtable_6uvloop_4loop_UnixTransport;
  __pyx_vtable_6uvloop_4loop_UnixTransport.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVStream;
  __pyx_vtable_6uvloop_4loop_UnixTransport.__pyx_base.__pyx_base.__pyx_base._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_13UnixTransport__new_socket;
  __pyx_vtable_6uvloop_4loop_UnixTransport.__pyx_base.__pyx_base.__pyx_base._open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int))__pyx_f_6uvloop_4loop_13UnixTransport__open;
  __pyx_vtable_6uvloop_4loop_UnixTransport.new = (struct __pyx_obj_6uvloop_4loop_UnixTransport *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_13UnixTransport_new;
  __pyx_vtable_6uvloop_4loop_UnixTransport.connect = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UnixTransport *, char *))__pyx_f_6uvloop_4loop_13UnixTransport_connect;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVStream); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UnixTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UnixTransport_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UnixTransport)) __PYX_ERR(22, 106, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UnixTransport_spec, __pyx_ptype_6uvloop_4loop_UnixTransport) < 0) __PYX_ERR(22, 106, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UnixTransport = &__pyx_type_6uvloop_4loop_UnixTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UnixTransport->tp_base = __pyx_ptype_6uvloop_4loop_UVStream;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UnixTransport) < 0) __PYX_ERR(22, 106, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UnixTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UnixTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UnixTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UnixTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UnixTransport, __pyx_vtabptr_6uvloop_4loop_UnixTransport) < 0) __PYX_ERR(22, 106, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UnixTransport) < 0) __PYX_ERR(22, 106, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UnixTransport, (PyObject *) __pyx_ptype_6uvloop_4loop_UnixTransport) < 0) __PYX_ERR(22, 106, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_UnixTransport->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_UnixTransport->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_UnixTransport, __pyx_base.__pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UnixTransport) < 0) __PYX_ERR(22, 106, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_ReadUnixTransport = &__pyx_vtable_6uvloop_4loop_ReadUnixTransport;
  __pyx_vtable_6uvloop_4loop_ReadUnixTransport.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVStream;
  __pyx_vtable_6uvloop_4loop_ReadUnixTransport.__pyx_base.__pyx_base.__pyx_base._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_17ReadUnixTransport__new_socket;
  __pyx_vtable_6uvloop_4loop_ReadUnixTransport.__pyx_base.__pyx_base.__pyx_base._open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int))__pyx_f_6uvloop_4loop_17ReadUnixTransport__open;
  __pyx_vtable_6uvloop_4loop_ReadUnixTransport.new = (struct __pyx_obj_6uvloop_4loop_ReadUnixTransport *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *))__pyx_f_6uvloop_4loop_17ReadUnixTransport_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVStream); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_ReadUnixTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_ReadUnixTransport_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_ReadUnixTransport)) __PYX_ERR(22, 131, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_ReadUnixTransport_spec, __pyx_ptype_6uvloop_4loop_ReadUnixTransport) < 0) __PYX_ERR(22, 131, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_ReadUnixTransport = &__pyx_type_6uvloop_4loop_ReadUnixTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_ReadUnixTransport->tp_base = __pyx_ptype_6uvloop_4loop_UVStream;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_ReadUnixTransport) < 0) __PYX_ERR(22, 131, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_ReadUnixTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_ReadUnixTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop_ReadUnixTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_ReadUnixTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_vtabptr_6uvloop_4loop_ReadUnixTransport) < 0) __PYX_ERR(22, 131, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_ReadUnixTransport) < 0) __PYX_ERR(22, 131, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ReadUnixTransport, (PyObject *) __pyx_ptype_6uvloop_4loop_ReadUnixTransport) < 0) __PYX_ERR(22, 131, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_ReadUnixTransport->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_ReadUnixTransport->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_ReadUnixTransport, __pyx_base.__pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_ReadUnixTransport) < 0) __PYX_ERR(22, 131, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_WriteUnixTransport = &__pyx_vtable_6uvloop_4loop_WriteUnixTransport;
  __pyx_vtable_6uvloop_4loop_WriteUnixTransport.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVStream;
  __pyx_vtable_6uvloop_4loop_WriteUnixTransport.__pyx_base.__pyx_base.__pyx_base._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_18WriteUnixTransport__new_socket;
  __pyx_vtable_6uvloop_4loop_WriteUnixTransport.__pyx_base.__pyx_base.__pyx_base._open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *, int))__pyx_f_6uvloop_4loop_18WriteUnixTransport__open;
  __pyx_vtable_6uvloop_4loop_WriteUnixTransport.new = (struct __pyx_obj_6uvloop_4loop_WriteUnixTransport *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, struct __pyx_obj_6uvloop_4loop_Server *, PyObject *))__pyx_f_6uvloop_4loop_18WriteUnixTransport_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVStream); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_WriteUnixTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_WriteUnixTransport_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_WriteUnixTransport)) __PYX_ERR(22, 176, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_WriteUnixTransport_spec, __pyx_ptype_6uvloop_4loop_WriteUnixTransport) < 0) __PYX_ERR(22, 176, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_WriteUnixTransport = &__pyx_type_6uvloop_4loop_WriteUnixTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_WriteUnixTransport->tp_base = __pyx_ptype_6uvloop_4loop_UVStream;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_WriteUnixTransport) < 0) __PYX_ERR(22, 176, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_WriteUnixTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_WriteUnixTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop_WriteUnixTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_WriteUnixTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_WriteUnixTransport, __pyx_vtabptr_6uvloop_4loop_WriteUnixTransport) < 0) __PYX_ERR(22, 176, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_WriteUnixTransport) < 0) __PYX_ERR(22, 176, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_WriteUnixTransport, (PyObject *) __pyx_ptype_6uvloop_4loop_WriteUnixTransport) < 0) __PYX_ERR(22, 176, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_WriteUnixTransport->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_WriteUnixTransport->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_WriteUnixTransport, __pyx_base.__pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_WriteUnixTransport) < 0) __PYX_ERR(22, 176, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVProcess = &__pyx_vtable_6uvloop_4loop_UVProcess;
  __pyx_vtable_6uvloop_4loop_UVProcess.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVProcess.__pyx_base._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_9UVProcess__close;
  __pyx_vtable_6uvloop_4loop_UVProcess._close_process_handle = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *))__pyx_f_6uvloop_4loop_9UVProcess__close_process_handle;
  __pyx_vtable_6uvloop_4loop_UVProcess._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_9UVProcess__init;
  __pyx_vtable_6uvloop_4loop_UVProcess._after_fork = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *))__pyx_f_6uvloop_4loop_9UVProcess__after_fork;
  __pyx_vtable_6uvloop_4loop_UVProcess._UVProcess__to_cstring_array = (char **(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *))__pyx_f_6uvloop_4loop_9UVProcess__UVProcess__to_cstring_array;
  __pyx_vtable_6uvloop_4loop_UVProcess._init_args = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *))__pyx_f_6uvloop_4loop_9UVProcess__init_args;
  __pyx_vtable_6uvloop_4loop_UVProcess._init_env = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *))__pyx_f_6uvloop_4loop_9UVProcess__init_env;
  __pyx_vtable_6uvloop_4loop_UVProcess._init_files = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_9UVProcess__init_files;
  __pyx_vtable_6uvloop_4loop_UVProcess._init_options = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, int))__pyx_f_6uvloop_4loop_9UVProcess__init_options;
  __pyx_vtable_6uvloop_4loop_UVProcess._close_after_spawn = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, int))__pyx_f_6uvloop_4loop_9UVProcess__close_after_spawn;
  __pyx_vtable_6uvloop_4loop_UVProcess._on_exit = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, int64_t, int))__pyx_f_6uvloop_4loop_9UVProcess__on_exit;
  __pyx_vtable_6uvloop_4loop_UVProcess._kill = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, int))__pyx_f_6uvloop_4loop_9UVProcess__kill;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVProcess = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVProcess_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVProcess)) __PYX_ERR(4, 2, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVProcess_spec, __pyx_ptype_6uvloop_4loop_UVProcess) < 0) __PYX_ERR(4, 2, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVProcess = &__pyx_type_6uvloop_4loop_UVProcess;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVProcess->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVProcess) < 0) __PYX_ERR(4, 2, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVProcess->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVProcess->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVProcess->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVProcess->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVProcess, __pyx_vtabptr_6uvloop_4loop_UVProcess) < 0) __PYX_ERR(4, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVProcess) < 0) __PYX_ERR(4, 2, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVProcess, (PyObject *) __pyx_ptype_6uvloop_4loop_UVProcess) < 0) __PYX_ERR(4, 2, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVProcess) < 0) __PYX_ERR(4, 2, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVProcessTransport = &__pyx_vtable_6uvloop_4loop_UVProcessTransport;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVProcess;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport.__pyx_base._init_files = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_18UVProcessTransport__init_files;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport.__pyx_base._on_exit = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcess *, int64_t, int))__pyx_f_6uvloop_4loop_18UVProcessTransport__on_exit;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._file_redirect_stdio = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, int))__pyx_f_6uvloop_4loop_18UVProcessTransport__file_redirect_stdio;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._file_devnull = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *))__pyx_f_6uvloop_4loop_18UVProcessTransport__file_devnull;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._file_inpipe = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *))__pyx_f_6uvloop_4loop_18UVProcessTransport__file_inpipe;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._file_outpipe = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *))__pyx_f_6uvloop_4loop_18UVProcessTransport__file_outpipe;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._check_proc = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *))__pyx_f_6uvloop_4loop_18UVProcessTransport__check_proc;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._pipe_connection_lost = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, int, PyObject *))__pyx_f_6uvloop_4loop_18UVProcessTransport__pipe_connection_lost;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._pipe_data_received = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, int, PyObject *))__pyx_f_6uvloop_4loop_18UVProcessTransport__pipe_data_received;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._call_connection_made = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *, PyObject *))__pyx_f_6uvloop_4loop_18UVProcessTransport__call_connection_made;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport._try_finish = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVProcessTransport *))__pyx_f_6uvloop_4loop_18UVProcessTransport__try_finish;
  __pyx_vtable_6uvloop_4loop_UVProcessTransport.new = (struct __pyx_obj_6uvloop_4loop_UVProcessTransport *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_18UVProcessTransport_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVProcess); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVProcessTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVProcessTransport_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVProcessTransport)) __PYX_ERR(4, 367, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVProcessTransport_spec, __pyx_ptype_6uvloop_4loop_UVProcessTransport) < 0) __PYX_ERR(4, 367, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVProcessTransport = &__pyx_type_6uvloop_4loop_UVProcessTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVProcessTransport->tp_base = __pyx_ptype_6uvloop_4loop_UVProcess;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVProcessTransport) < 0) __PYX_ERR(4, 367, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVProcessTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVProcessTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVProcessTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVProcessTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_vtabptr_6uvloop_4loop_UVProcessTransport) < 0) __PYX_ERR(4, 367, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVProcessTransport) < 0) __PYX_ERR(4, 367, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVProcessTransport, (PyObject *) __pyx_ptype_6uvloop_4loop_UVProcessTransport) < 0) __PYX_ERR(4, 367, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVProcessTransport) < 0) __PYX_ERR(4, 367, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVFSEvent = &__pyx_vtable_6uvloop_4loop_UVFSEvent;
  __pyx_vtable_6uvloop_4loop_UVFSEvent.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVHandle;
  __pyx_vtable_6uvloop_4loop_UVFSEvent.__pyx_base._close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVHandle *))__pyx_f_6uvloop_4loop_9UVFSEvent__close;
  __pyx_vtable_6uvloop_4loop_UVFSEvent._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVFSEvent *, struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_9UVFSEvent__init;
  __pyx_vtable_6uvloop_4loop_UVFSEvent.start = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVFSEvent *, char *, int))__pyx_f_6uvloop_4loop_9UVFSEvent_start;
  __pyx_vtable_6uvloop_4loop_UVFSEvent.stop = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVFSEvent *))__pyx_f_6uvloop_4loop_9UVFSEvent_stop;
  __pyx_vtable_6uvloop_4loop_UVFSEvent.new = (struct __pyx_obj_6uvloop_4loop_UVFSEvent *(*)(struct __pyx_obj_6uvloop_4loop_Loop *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_9UVFSEvent_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVHandle); if (unlikely(!__pyx_t_1)) __PYX_ERR(23, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UVFSEvent = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVFSEvent_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UVFSEvent)) __PYX_ERR(23, 11, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVFSEvent_spec, __pyx_ptype_6uvloop_4loop_UVFSEvent) < 0) __PYX_ERR(23, 11, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVFSEvent = &__pyx_type_6uvloop_4loop_UVFSEvent;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UVFSEvent->tp_base = __pyx_ptype_6uvloop_4loop_UVHandle;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVFSEvent) < 0) __PYX_ERR(23, 11, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVFSEvent->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVFSEvent->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVFSEvent->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVFSEvent->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVFSEvent, __pyx_vtabptr_6uvloop_4loop_UVFSEvent) < 0) __PYX_ERR(23, 11, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVFSEvent) < 0) __PYX_ERR(23, 11, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVFSEvent, (PyObject *) __pyx_ptype_6uvloop_4loop_UVFSEvent) < 0) __PYX_ERR(23, 11, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVFSEvent) < 0) __PYX_ERR(23, 11, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UVRequest = &__pyx_vtable_6uvloop_4loop_UVRequest;
  __pyx_vtable_6uvloop_4loop_UVRequest.on_done = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVRequest *))__pyx_f_6uvloop_4loop_9UVRequest_on_done;
  __pyx_vtable_6uvloop_4loop_UVRequest.cancel = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVRequest *))__pyx_f_6uvloop_4loop_9UVRequest_cancel;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_UVRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UVRequest_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_UVRequest)) __PYX_ERR(24, 1, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UVRequest_spec, __pyx_ptype_6uvloop_4loop_UVRequest) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UVRequest = &__pyx_type_6uvloop_4loop_UVRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UVRequest) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UVRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UVRequest->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UVRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UVRequest->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UVRequest, __pyx_vtabptr_6uvloop_4loop_UVRequest) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UVRequest) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UVRequest, (PyObject *) __pyx_ptype_6uvloop_4loop_UVRequest) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UVRequest) < 0) __PYX_ERR(24, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop__SSLProtocolTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop__SSLProtocolTransport_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop__SSLProtocolTransport)) __PYX_ERR(25, 15, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop__SSLProtocolTransport_spec, __pyx_ptype_6uvloop_4loop__SSLProtocolTransport) < 0) __PYX_ERR(25, 15, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop__SSLProtocolTransport = &__pyx_type_6uvloop_4loop__SSLProtocolTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport) < 0) __PYX_ERR(25, 15, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop__SSLProtocolTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop__SSLProtocolTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop__SSLProtocolTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop__SSLProtocolTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SSLProtocolTransport, (PyObject *) __pyx_ptype_6uvloop_4loop__SSLProtocolTransport) < 0) __PYX_ERR(25, 15, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop__SSLProtocolTransport) < 0) __PYX_ERR(25, 15, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_SSLProtocol = &__pyx_vtable_6uvloop_4loop_SSLProtocol;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._set_app_protocol = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *))__pyx_f_6uvloop_4loop_11SSLProtocol__set_app_protocol;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._wakeup_waiter = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__wakeup_waiter *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__wakeup_waiter;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._get_extra_info = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__get_extra_info *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__get_extra_info;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._set_state = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, enum __pyx_t_6uvloop_4loop_SSLProtocolState))__pyx_f_6uvloop_4loop_11SSLProtocol__set_state;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._start_handshake = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__start_handshake;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._check_handshake_timeout = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__check_handshake_timeout;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_handshake = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__do_handshake;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._on_handshake_complete = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *))__pyx_f_6uvloop_4loop_11SSLProtocol__on_handshake_complete;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._start_shutdown = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__start_shutdown *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__start_shutdown;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._check_shutdown_timeout = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__check_shutdown_timeout;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_read_into_void = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *))__pyx_f_6uvloop_4loop_11SSLProtocol__do_read_into_void;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_flush = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_flush *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__do_flush;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_shutdown = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__do_shutdown *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__do_shutdown;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._on_shutdown_complete = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *))__pyx_f_6uvloop_4loop_11SSLProtocol__on_shutdown_complete;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._abort = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *))__pyx_f_6uvloop_4loop_11SSLProtocol__abort;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._write_appdata = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_11SSLProtocol__write_appdata;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_write = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__do_write;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._process_outgoing = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__process_outgoing;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_read = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__do_read;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_read__buffered = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__do_read__buffered;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._do_read__copied = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__do_read__copied;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._call_eof_received = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__call_eof_received *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__call_eof_received;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._control_app_writing = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__control_app_writing *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__control_app_writing;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._get_write_buffer_size = (size_t (*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__get_write_buffer_size;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._set_write_buffer_limits = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__set_write_buffer_limits;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._pause_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__pause_reading;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._resume_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *))__pyx_f_6uvloop_4loop_11SSLProtocol__resume_reading;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._control_ssl_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__control_ssl_reading;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._set_read_buffer_limits = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__set_read_buffer_limits;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._get_read_buffer_size = (size_t (*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *))__pyx_f_6uvloop_4loop_11SSLProtocol__get_read_buffer_size;
  __pyx_vtable_6uvloop_4loop_SSLProtocol._fatal_error = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_SSLProtocol *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_11SSLProtocol__fatal_error *__pyx_optional_args))__pyx_f_6uvloop_4loop_11SSLProtocol__fatal_error;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_SSLProtocol = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_SSLProtocol_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_SSLProtocol)) __PYX_ERR(25, 195, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_SSLProtocol_spec, __pyx_ptype_6uvloop_4loop_SSLProtocol) < 0) __PYX_ERR(25, 195, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_SSLProtocol = &__pyx_type_6uvloop_4loop_SSLProtocol;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_SSLProtocol) < 0) __PYX_ERR(25, 195, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_SSLProtocol->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_SSLProtocol->tp_dictoffset && __pyx_ptype_6uvloop_4loop_SSLProtocol->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_SSLProtocol->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_vtabptr_6uvloop_4loop_SSLProtocol) < 0) __PYX_ERR(25, 195, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_SSLProtocol) < 0) __PYX_ERR(25, 195, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SSLProtocol, (PyObject *) __pyx_ptype_6uvloop_4loop_SSLProtocol) < 0) __PYX_ERR(25, 195, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_SSLProtocol) < 0) __PYX_ERR(25, 195, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_UDPTransport = &__pyx_vtable_6uvloop_4loop_UDPTransport;
  __pyx_vtable_6uvloop_4loop_UDPTransport.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVBaseTransport;
  __pyx_vtable_6uvloop_4loop_UDPTransport.__pyx_base.__pyx_base._new_socket = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVSocketHandle *))__pyx_f_6uvloop_4loop_12UDPTransport__new_socket;
  __pyx_vtable_6uvloop_4loop_UDPTransport.__pyx_base._get_write_buffer_size = (size_t (*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_12UDPTransport__get_write_buffer_size;
  __pyx_vtable_6uvloop_4loop_UDPTransport.__pyx_base._is_reading = (int (*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_12UDPTransport__is_reading;
  __pyx_vtable_6uvloop_4loop_UDPTransport.__pyx_base._start_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_12UDPTransport__start_reading;
  __pyx_vtable_6uvloop_4loop_UDPTransport.__pyx_base._stop_reading = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UVBaseTransport *))__pyx_f_6uvloop_4loop_12UDPTransport__stop_reading;
  __pyx_vtable_6uvloop_4loop_UDPTransport._init = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct __pyx_obj_6uvloop_4loop_Loop *, unsigned int))__pyx_f_6uvloop_4loop_12UDPTransport__init;
  __pyx_vtable_6uvloop_4loop_UDPTransport._set_address = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct addrinfo *))__pyx_f_6uvloop_4loop_12UDPTransport__set_address;
  __pyx_vtable_6uvloop_4loop_UDPTransport._connect = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct sockaddr *, size_t))__pyx_f_6uvloop_4loop_12UDPTransport__connect;
  __pyx_vtable_6uvloop_4loop_UDPTransport._bind = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, struct sockaddr *))__pyx_f_6uvloop_4loop_12UDPTransport__bind;
  __pyx_vtable_6uvloop_4loop_UDPTransport.open = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, int, int))__pyx_f_6uvloop_4loop_12UDPTransport_open;
  __pyx_vtable_6uvloop_4loop_UDPTransport._set_broadcast = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, int))__pyx_f_6uvloop_4loop_12UDPTransport__set_broadcast;
  __pyx_vtable_6uvloop_4loop_UDPTransport._UDPTransport__receiving_started = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *))__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_started;
  __pyx_vtable_6uvloop_4loop_UDPTransport._UDPTransport__receiving_stopped = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *))__pyx_f_6uvloop_4loop_12UDPTransport__UDPTransport__receiving_stopped;
  __pyx_vtable_6uvloop_4loop_UDPTransport._send = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_12UDPTransport__send;
  __pyx_vtable_6uvloop_4loop_UDPTransport._on_receive = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_12UDPTransport__on_receive;
  __pyx_vtable_6uvloop_4loop_UDPTransport._on_sent = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *, struct __pyx_opt_args_6uvloop_4loop_12UDPTransport__on_sent *__pyx_optional_args))__pyx_f_6uvloop_4loop_12UDPTransport__on_sent;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport); if (unlikely(!__pyx_t_1)) __PYX_ERR(26, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_UDPTransport = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_UDPTransport_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_UDPTransport)) __PYX_ERR(26, 54, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_UDPTransport_spec, __pyx_ptype_6uvloop_4loop_UDPTransport) < 0) __PYX_ERR(26, 54, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_UDPTransport = &__pyx_type_6uvloop_4loop_UDPTransport;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_UDPTransport->tp_base = __pyx_ptype_6uvloop_4loop_UVBaseTransport;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_UDPTransport) < 0) __PYX_ERR(26, 54, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_UDPTransport->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_UDPTransport->tp_dictoffset && __pyx_ptype_6uvloop_4loop_UDPTransport->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_UDPTransport->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_UDPTransport, __pyx_vtabptr_6uvloop_4loop_UDPTransport) < 0) __PYX_ERR(26, 54, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_UDPTransport) < 0) __PYX_ERR(26, 54, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UDPTransport, (PyObject *) __pyx_ptype_6uvloop_4loop_UDPTransport) < 0) __PYX_ERR(26, 54, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_UDPTransport->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_UDPTransport->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_UDPTransport, __pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_UDPTransport) < 0) __PYX_ERR(26, 54, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_Server = &__pyx_vtable_6uvloop_4loop_Server;
  __pyx_vtable_6uvloop_4loop_Server._add_server = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Server *, struct __pyx_obj_6uvloop_4loop_UVStreamServer *))__pyx_f_6uvloop_4loop_6Server__add_server;
  __pyx_vtable_6uvloop_4loop_Server._start_serving = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Server *))__pyx_f_6uvloop_4loop_6Server__start_serving;
  __pyx_vtable_6uvloop_4loop_Server._wakeup = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Server *))__pyx_f_6uvloop_4loop_6Server__wakeup;
  __pyx_vtable_6uvloop_4loop_Server._attach = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Server *))__pyx_f_6uvloop_4loop_6Server__attach;
  __pyx_vtable_6uvloop_4loop_Server._detach = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Server *))__pyx_f_6uvloop_4loop_6Server__detach;
  __pyx_vtable_6uvloop_4loop_Server._ref = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Server *))__pyx_f_6uvloop_4loop_6Server__ref;
  __pyx_vtable_6uvloop_4loop_Server._unref = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_Server *))__pyx_f_6uvloop_4loop_6Server__unref;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_Server = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_Server_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_Server)) __PYX_ERR(27, 4, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_Server_spec, __pyx_ptype_6uvloop_4loop_Server) < 0) __PYX_ERR(27, 4, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_Server = &__pyx_type_6uvloop_4loop_Server;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_Server) < 0) __PYX_ERR(27, 4, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_Server->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_Server->tp_dictoffset && __pyx_ptype_6uvloop_4loop_Server->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_Server->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_Server, __pyx_vtabptr_6uvloop_4loop_Server) < 0) __PYX_ERR(27, 4, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_Server) < 0) __PYX_ERR(27, 4, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Server, (PyObject *) __pyx_ptype_6uvloop_4loop_Server) < 0) __PYX_ERR(27, 4, __pyx_L1_error)
  if (__pyx_ptype_6uvloop_4loop_Server->tp_weaklistoffset == 0) __pyx_ptype_6uvloop_4loop_Server->tp_weaklistoffset = offsetof(struct __pyx_obj_6uvloop_4loop_Server, __weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_Server) < 0) __PYX_ERR(27, 4, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_PseudoSocket = &__pyx_vtable_6uvloop_4loop_PseudoSocket;
  __pyx_vtable_6uvloop_4loop_PseudoSocket._na = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_PseudoSocket *, PyObject *))__pyx_f_6uvloop_4loop_12PseudoSocket__na;
  __pyx_vtable_6uvloop_4loop_PseudoSocket._make_sock = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_PseudoSocket *))__pyx_f_6uvloop_4loop_12PseudoSocket__make_sock;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_PseudoSocket = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_PseudoSocket_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_PseudoSocket)) __PYX_ERR(11, 1, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_PseudoSocket_spec, __pyx_ptype_6uvloop_4loop_PseudoSocket) < 0) __PYX_ERR(11, 1, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_PseudoSocket = &__pyx_type_6uvloop_4loop_PseudoSocket;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_PseudoSocket) < 0) __PYX_ERR(11, 1, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_PseudoSocket->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_PseudoSocket->tp_dictoffset && __pyx_ptype_6uvloop_4loop_PseudoSocket->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_PseudoSocket->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_vtabptr_6uvloop_4loop_PseudoSocket) < 0) __PYX_ERR(11, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_PseudoSocket) < 0) __PYX_ERR(11, 1, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PseudoSocket, (PyObject *) __pyx_ptype_6uvloop_4loop_PseudoSocket) < 0) __PYX_ERR(11, 1, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_PseudoSocket) < 0) __PYX_ERR(11, 1, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_LruCache = &__pyx_vtable_6uvloop_4loop_LruCache;
  __pyx_vtable_6uvloop_4loop_LruCache.get = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_LruCache *, PyObject *, PyObject *))__pyx_f_6uvloop_4loop_8LruCache_get;
  __pyx_vtable_6uvloop_4loop_LruCache.needs_cleanup = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_LruCache *))__pyx_f_6uvloop_4loop_8LruCache_needs_cleanup;
  __pyx_vtable_6uvloop_4loop_LruCache.cleanup_one = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_LruCache *))__pyx_f_6uvloop_4loop_8LruCache_cleanup_one;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_LruCache = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_LruCache_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_LruCache)) __PYX_ERR(12, 5, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_LruCache_spec, __pyx_ptype_6uvloop_4loop_LruCache) < 0) __PYX_ERR(12, 5, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_LruCache = &__pyx_type_6uvloop_4loop_LruCache;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_LruCache) < 0) __PYX_ERR(12, 5, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_LruCache->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_LruCache->tp_dictoffset && __pyx_ptype_6uvloop_4loop_LruCache->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_LruCache->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_LruCache, __pyx_vtabptr_6uvloop_4loop_LruCache) < 0) __PYX_ERR(12, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_LruCache) < 0) __PYX_ERR(12, 5, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LruCache, (PyObject *) __pyx_ptype_6uvloop_4loop_LruCache) < 0) __PYX_ERR(12, 5, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_LruCache) < 0) __PYX_ERR(12, 5, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop__StreamWriteContext = &__pyx_vtable_6uvloop_4loop__StreamWriteContext;
  __pyx_vtable_6uvloop_4loop__StreamWriteContext.free_bufs = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *))__pyx_f_6uvloop_4loop_19_StreamWriteContext_free_bufs;
  __pyx_vtable_6uvloop_4loop__StreamWriteContext.close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *))__pyx_f_6uvloop_4loop_19_StreamWriteContext_close;
  __pyx_vtable_6uvloop_4loop__StreamWriteContext.advance_uv_buf = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop__StreamWriteContext *, size_t))__pyx_f_6uvloop_4loop_19_StreamWriteContext_advance_uv_buf;
  __pyx_vtable_6uvloop_4loop__StreamWriteContext.new = (struct __pyx_obj_6uvloop_4loop__StreamWriteContext *(*)(struct __pyx_obj_6uvloop_4loop_UVStream *, PyObject *))__pyx_f_6uvloop_4loop_19_StreamWriteContext_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop__StreamWriteContext = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop__StreamWriteContext_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop__StreamWriteContext)) __PYX_ERR(20, 10, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop__StreamWriteContext_spec, __pyx_ptype_6uvloop_4loop__StreamWriteContext) < 0) __PYX_ERR(20, 10, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop__StreamWriteContext = &__pyx_type_6uvloop_4loop__StreamWriteContext;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop__StreamWriteContext) < 0) __PYX_ERR(20, 10, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop__StreamWriteContext->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop__StreamWriteContext->tp_dictoffset && __pyx_ptype_6uvloop_4loop__StreamWriteContext->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop__StreamWriteContext->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop__StreamWriteContext, __pyx_vtabptr_6uvloop_4loop__StreamWriteContext) < 0) __PYX_ERR(20, 10, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop__StreamWriteContext) < 0) __PYX_ERR(20, 10, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_StreamWriteContext, (PyObject *) __pyx_ptype_6uvloop_4loop__StreamWriteContext) < 0) __PYX_ERR(20, 10, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop__StreamWriteContext) < 0) __PYX_ERR(20, 10, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop__TCPConnectRequest = &__pyx_vtable_6uvloop_4loop__TCPConnectRequest;
  __pyx_vtable_6uvloop_4loop__TCPConnectRequest.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVRequest;
  __pyx_vtable_6uvloop_4loop__TCPConnectRequest.connect = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop__TCPConnectRequest *, struct sockaddr *))__pyx_f_6uvloop_4loop_18_TCPConnectRequest_connect;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop__TCPConnectRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop__TCPConnectRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop__TCPConnectRequest)) __PYX_ERR(3, 185, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop__TCPConnectRequest_spec, __pyx_ptype_6uvloop_4loop__TCPConnectRequest) < 0) __PYX_ERR(3, 185, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop__TCPConnectRequest = &__pyx_type_6uvloop_4loop__TCPConnectRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop__TCPConnectRequest->tp_base = __pyx_ptype_6uvloop_4loop_UVRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop__TCPConnectRequest) < 0) __PYX_ERR(3, 185, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop__TCPConnectRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop__TCPConnectRequest->tp_dictoffset && __pyx_ptype_6uvloop_4loop__TCPConnectRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop__TCPConnectRequest->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop__TCPConnectRequest, __pyx_vtabptr_6uvloop_4loop__TCPConnectRequest) < 0) __PYX_ERR(3, 185, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop__TCPConnectRequest) < 0) __PYX_ERR(3, 185, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TCPConnectRequest, (PyObject *) __pyx_ptype_6uvloop_4loop__TCPConnectRequest) < 0) __PYX_ERR(3, 185, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop__TCPConnectRequest) < 0) __PYX_ERR(3, 185, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop__PipeConnectRequest = &__pyx_vtable_6uvloop_4loop__PipeConnectRequest;
  __pyx_vtable_6uvloop_4loop__PipeConnectRequest.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVRequest;
  __pyx_vtable_6uvloop_4loop__PipeConnectRequest.connect = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop__PipeConnectRequest *, char *))__pyx_f_6uvloop_4loop_19_PipeConnectRequest_connect;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(22, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop__PipeConnectRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop__PipeConnectRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop__PipeConnectRequest)) __PYX_ERR(22, 209, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop__PipeConnectRequest_spec, __pyx_ptype_6uvloop_4loop__PipeConnectRequest) < 0) __PYX_ERR(22, 209, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop__PipeConnectRequest = &__pyx_type_6uvloop_4loop__PipeConnectRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop__PipeConnectRequest->tp_base = __pyx_ptype_6uvloop_4loop_UVRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop__PipeConnectRequest) < 0) __PYX_ERR(22, 209, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop__PipeConnectRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop__PipeConnectRequest->tp_dictoffset && __pyx_ptype_6uvloop_4loop__PipeConnectRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop__PipeConnectRequest->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop__PipeConnectRequest, __pyx_vtabptr_6uvloop_4loop__PipeConnectRequest) < 0) __PYX_ERR(22, 209, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop__PipeConnectRequest) < 0) __PYX_ERR(22, 209, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PipeConnectRequest, (PyObject *) __pyx_ptype_6uvloop_4loop__PipeConnectRequest) < 0) __PYX_ERR(22, 209, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop__PipeConnectRequest) < 0) __PYX_ERR(22, 209, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_SockAddrHolder = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_SockAddrHolder_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_SockAddrHolder)) __PYX_ERR(5, 72, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_SockAddrHolder_spec, __pyx_ptype_6uvloop_4loop_SockAddrHolder) < 0) __PYX_ERR(5, 72, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_SockAddrHolder = &__pyx_type_6uvloop_4loop_SockAddrHolder;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_SockAddrHolder) < 0) __PYX_ERR(5, 72, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_SockAddrHolder->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_SockAddrHolder->tp_dictoffset && __pyx_ptype_6uvloop_4loop_SockAddrHolder->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_SockAddrHolder->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SockAddrHolder, (PyObject *) __pyx_ptype_6uvloop_4loop_SockAddrHolder) < 0) __PYX_ERR(5, 72, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_SockAddrHolder) < 0) __PYX_ERR(5, 72, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_AddrInfo = &__pyx_vtable_6uvloop_4loop_AddrInfo;
  __pyx_vtable_6uvloop_4loop_AddrInfo.set_data = (void (*)(struct __pyx_obj_6uvloop_4loop_AddrInfo *, struct addrinfo *))__pyx_f_6uvloop_4loop_8AddrInfo_set_data;
  __pyx_vtable_6uvloop_4loop_AddrInfo.unpack = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_AddrInfo *))__pyx_f_6uvloop_4loop_8AddrInfo_unpack;
  __pyx_vtable_6uvloop_4loop_AddrInfo.isinstance = (int (*)(PyObject *))__pyx_f_6uvloop_4loop_8AddrInfo_isinstance;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop_AddrInfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_AddrInfo_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop_AddrInfo)) __PYX_ERR(5, 289, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_AddrInfo_spec, __pyx_ptype_6uvloop_4loop_AddrInfo) < 0) __PYX_ERR(5, 289, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_AddrInfo = &__pyx_type_6uvloop_4loop_AddrInfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_AddrInfo) < 0) __PYX_ERR(5, 289, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_AddrInfo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_AddrInfo->tp_dictoffset && __pyx_ptype_6uvloop_4loop_AddrInfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_AddrInfo->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_AddrInfo, __pyx_vtabptr_6uvloop_4loop_AddrInfo) < 0) __PYX_ERR(5, 289, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_AddrInfo) < 0) __PYX_ERR(5, 289, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AddrInfo, (PyObject *) __pyx_ptype_6uvloop_4loop_AddrInfo) < 0) __PYX_ERR(5, 289, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_AddrInfo) < 0) __PYX_ERR(5, 289, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_AddrInfoRequest = &__pyx_vtable_6uvloop_4loop_AddrInfoRequest;
  __pyx_vtable_6uvloop_4loop_AddrInfoRequest.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVRequest;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_AddrInfoRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_AddrInfoRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_AddrInfoRequest)) __PYX_ERR(5, 333, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_AddrInfoRequest_spec, __pyx_ptype_6uvloop_4loop_AddrInfoRequest) < 0) __PYX_ERR(5, 333, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_AddrInfoRequest = &__pyx_type_6uvloop_4loop_AddrInfoRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_AddrInfoRequest->tp_base = __pyx_ptype_6uvloop_4loop_UVRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_AddrInfoRequest) < 0) __PYX_ERR(5, 333, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_AddrInfoRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_AddrInfoRequest->tp_dictoffset && __pyx_ptype_6uvloop_4loop_AddrInfoRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_AddrInfoRequest->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_AddrInfoRequest, __pyx_vtabptr_6uvloop_4loop_AddrInfoRequest) < 0) __PYX_ERR(5, 333, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_AddrInfoRequest) < 0) __PYX_ERR(5, 333, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AddrInfoRequest, (PyObject *) __pyx_ptype_6uvloop_4loop_AddrInfoRequest) < 0) __PYX_ERR(5, 333, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_AddrInfoRequest) < 0) __PYX_ERR(5, 333, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop_NameInfoRequest = &__pyx_vtable_6uvloop_4loop_NameInfoRequest;
  __pyx_vtable_6uvloop_4loop_NameInfoRequest.__pyx_base = *__pyx_vtabptr_6uvloop_4loop_UVRequest;
  __pyx_vtable_6uvloop_4loop_NameInfoRequest.query = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop_NameInfoRequest *, struct sockaddr *, int))__pyx_f_6uvloop_4loop_15NameInfoRequest_query;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_6uvloop_4loop_UVRequest); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6uvloop_4loop_NameInfoRequest = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop_NameInfoRequest_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_6uvloop_4loop_NameInfoRequest)) __PYX_ERR(5, 396, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop_NameInfoRequest_spec, __pyx_ptype_6uvloop_4loop_NameInfoRequest) < 0) __PYX_ERR(5, 396, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop_NameInfoRequest = &__pyx_type_6uvloop_4loop_NameInfoRequest;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_6uvloop_4loop_NameInfoRequest->tp_base = __pyx_ptype_6uvloop_4loop_UVRequest;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop_NameInfoRequest) < 0) __PYX_ERR(5, 396, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop_NameInfoRequest->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop_NameInfoRequest->tp_dictoffset && __pyx_ptype_6uvloop_4loop_NameInfoRequest->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop_NameInfoRequest->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop_NameInfoRequest, __pyx_vtabptr_6uvloop_4loop_NameInfoRequest) < 0) __PYX_ERR(5, 396, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop_NameInfoRequest) < 0) __PYX_ERR(5, 396, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NameInfoRequest, (PyObject *) __pyx_ptype_6uvloop_4loop_NameInfoRequest) < 0) __PYX_ERR(5, 396, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop_NameInfoRequest) < 0) __PYX_ERR(5, 396, __pyx_L1_error)
  #endif
  __pyx_vtabptr_6uvloop_4loop__UDPSendContext = &__pyx_vtable_6uvloop_4loop__UDPSendContext;
  __pyx_vtable_6uvloop_4loop__UDPSendContext.close = (PyObject *(*)(struct __pyx_obj_6uvloop_4loop__UDPSendContext *))__pyx_f_6uvloop_4loop_15_UDPSendContext_close;
  __pyx_vtable_6uvloop_4loop__UDPSendContext.new = (struct __pyx_obj_6uvloop_4loop__UDPSendContext *(*)(struct __pyx_obj_6uvloop_4loop_UDPTransport *, PyObject *))__pyx_f_6uvloop_4loop_15_UDPSendContext_new;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop__UDPSendContext = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop__UDPSendContext_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop__UDPSendContext)) __PYX_ERR(26, 3, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop__UDPSendContext_spec, __pyx_ptype_6uvloop_4loop__UDPSendContext) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop__UDPSendContext = &__pyx_type_6uvloop_4loop__UDPSendContext;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop__UDPSendContext) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop__UDPSendContext->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop__UDPSendContext->tp_dictoffset && __pyx_ptype_6uvloop_4loop__UDPSendContext->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop__UDPSendContext->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_6uvloop_4loop__UDPSendContext, __pyx_vtabptr_6uvloop_4loop__UDPSendContext) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_6uvloop_4loop__UDPSendContext) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UDPSendContext, (PyObject *) __pyx_ptype_6uvloop_4loop__UDPSendContext) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_6uvloop_4loop__UDPSendContext) < 0) __PYX_ERR(26, 3, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo)) __PYX_ERR(0, 894, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo) < 0) __PYX_ERR(0, 894, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo = &__pyx_type_6uvloop_4loop___pyx_scope_struct___getaddrinfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo) < 0) __PYX_ERR(0, 894, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct___getaddrinfo->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo)) __PYX_ERR(0, 936, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo) < 0) __PYX_ERR(0, 936, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo = &__pyx_type_6uvloop_4loop___pyx_scope_struct_1__getnameinfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo) < 0) __PYX_ERR(0, 936, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_1__getnameinfo->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_2___get___spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__)) __PYX_ERR(0, 1251, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_2___get___spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__) < 0) __PYX_ERR(0, 1251, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__ = &__pyx_type_6uvloop_4loop___pyx_scope_struct_2___get__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__) < 0) __PYX_ERR(0, 1251, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_2___get__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete)) __PYX_ERR(0, 1474, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete) < 0) __PYX_ERR(0, 1474, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete = &__pyx_type_6uvloop_4loop___pyx_scope_struct_3_run_until_complete;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete) < 0) __PYX_ERR(0, 1474, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_3_run_until_complete->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo)) __PYX_ERR(0, 1520, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo = &__pyx_type_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_4_getaddrinfo->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo)) __PYX_ERR(0, 1532, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo) < 0) __PYX_ERR(0, 1532, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo = &__pyx_type_6uvloop_4loop___pyx_scope_struct_5_getnameinfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo) < 0) __PYX_ERR(0, 1532, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_5_getnameinfo->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls)) __PYX_ERR(0, 1587, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls) < 0) __PYX_ERR(0, 1587, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls = &__pyx_type_6uvloop_4loop___pyx_scope_struct_6_start_tls;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls) < 0) __PYX_ERR(0, 1587, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_6_start_tls->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server)) __PYX_ERR(0, 1643, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server = &__pyx_type_6uvloop_4loop___pyx_scope_struct_7_create_server;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_7_create_server->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection)) __PYX_ERR(0, 1857, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection = &__pyx_type_6uvloop_4loop___pyx_scope_struct_8_create_connection;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_8_create_connection->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr)) __PYX_ERR(0, 2042, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr) < 0) __PYX_ERR(0, 2042, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr = &__pyx_type_6uvloop_4loop___pyx_scope_struct_9_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr) < 0) __PYX_ERR(0, 2042, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_9_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr)) __PYX_ERR(0, 2047, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr) < 0) __PYX_ERR(0, 2047, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr = &__pyx_type_6uvloop_4loop___pyx_scope_struct_10_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr) < 0) __PYX_ERR(0, 2047, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_10_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server)) __PYX_ERR(0, 2093, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server = &__pyx_type_6uvloop_4loop___pyx_scope_struct_11_create_unix_server;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_11_create_unix_server->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection)) __PYX_ERR(0, 2235, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection = &__pyx_type_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_12_create_unix_connection->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv)) __PYX_ERR(0, 2475, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv) < 0) __PYX_ERR(0, 2475, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv = &__pyx_type_6uvloop_4loop___pyx_scope_struct_13_sock_recv;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv) < 0) __PYX_ERR(0, 2475, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_13_sock_recv->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into)) __PYX_ERR(0, 2503, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into) < 0) __PYX_ERR(0, 2503, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into = &__pyx_type_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into) < 0) __PYX_ERR(0, 2503, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_14_sock_recv_into->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall)) __PYX_ERR(0, 2530, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall) < 0) __PYX_ERR(0, 2530, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall = &__pyx_type_6uvloop_4loop___pyx_scope_struct_15_sock_sendall;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall) < 0) __PYX_ERR(0, 2530, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_15_sock_sendall->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept)) __PYX_ERR(0, 2584, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept) < 0) __PYX_ERR(0, 2584, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept = &__pyx_type_6uvloop_4loop___pyx_scope_struct_16_sock_accept;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept) < 0) __PYX_ERR(0, 2584, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_16_sock_accept->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect)) __PYX_ERR(0, 2613, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect) < 0) __PYX_ERR(0, 2613, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect = &__pyx_type_6uvloop_4loop___pyx_scope_struct_17_sock_connect;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect) < 0) __PYX_ERR(0, 2613, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_17_sock_connect->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom)) __PYX_ERR(0, 2637, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom) < 0) __PYX_ERR(0, 2637, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom = &__pyx_type_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom) < 0) __PYX_ERR(0, 2637, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_18_sock_recvfrom->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into)) __PYX_ERR(0, 2641, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into) < 0) __PYX_ERR(0, 2641, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into = &__pyx_type_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into) < 0) __PYX_ERR(0, 2641, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_19_sock_recvfrom_into->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto)) __PYX_ERR(0, 2645, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto) < 0) __PYX_ERR(0, 2645, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto = &__pyx_type_6uvloop_4loop___pyx_scope_struct_20_sock_sendto;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto) < 0) __PYX_ERR(0, 2645, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_20_sock_sendto->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket)) __PYX_ERR(0, 2649, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket) < 0) __PYX_ERR(0, 2649, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket = &__pyx_type_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket) < 0) __PYX_ERR(0, 2649, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_21_connect_accepted_socket->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run)) __PYX_ERR(0, 2752, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run) < 0) __PYX_ERR(0, 2752, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run = &__pyx_type_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run) < 0) __PYX_ERR(0, 2752, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_22__Loop__subprocess_run->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell)) __PYX_ERR(0, 2817, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell = &__pyx_type_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_23_subprocess_shell->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec)) __PYX_ERR(0, 2832, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec) < 0) __PYX_ERR(0, 2832, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec = &__pyx_type_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec) < 0) __PYX_ERR(0, 2832, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_24_subprocess_exec->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe)) __PYX_ERR(0, 2844, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe) < 0) __PYX_ERR(0, 2844, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe = &__pyx_type_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe) < 0) __PYX_ERR(0, 2844, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_25_connect_read_pipe->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe)) __PYX_ERR(0, 2871, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe) < 0) __PYX_ERR(0, 2871, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe = &__pyx_type_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe) < 0) __PYX_ERR(0, 2871, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_26_connect_write_pipe->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint)) __PYX_ERR(0, 3002, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint = &__pyx_type_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_27_create_datagram_endpoint->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr)) __PYX_ERR(0, 3049, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr) < 0) __PYX_ERR(0, 3049, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr = &__pyx_type_6uvloop_4loop___pyx_scope_struct_28_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr) < 0) __PYX_ERR(0, 3049, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_28_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens)) __PYX_ERR(0, 3198, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens) < 0) __PYX_ERR(0, 3198, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens = &__pyx_type_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens) < 0) __PYX_ERR(0, 3198, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_29_shutdown_asyncgens->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor)) __PYX_ERR(0, 3223, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor) < 0) __PYX_ERR(0, 3223, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor = &__pyx_type_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor) < 0) __PYX_ERR(0, 3223, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_30_shutdown_default_executor->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake)) __PYX_ERR(25, 462, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake) < 0) __PYX_ERR(25, 462, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake = &__pyx_type_6uvloop_4loop___pyx_scope_struct_31__start_handshake;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake) < 0) __PYX_ERR(25, 462, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_31__start_handshake->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown)) __PYX_ERR(25, 559, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown) < 0) __PYX_ERR(25, 559, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown = &__pyx_type_6uvloop_4loop___pyx_scope_struct_32__start_shutdown;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown) < 0) __PYX_ERR(25, 559, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_32__start_shutdown->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter___spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__)) __PYX_ERR(27, 51, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter___spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__) < 0) __PYX_ERR(27, 51, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__ = &__pyx_type_6uvloop_4loop___pyx_scope_struct_33___aenter__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__) < 0) __PYX_ERR(27, 51, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_33___aenter__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit___spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__)) __PYX_ERR(27, 55, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit___spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__) < 0) __PYX_ERR(27, 55, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__ = &__pyx_type_6uvloop_4loop___pyx_scope_struct_34___aexit__;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__) < 0) __PYX_ERR(27, 55, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_34___aexit__->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed)) __PYX_ERR(27, 66, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed) < 0) __PYX_ERR(27, 66, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed = &__pyx_type_6uvloop_4loop___pyx_scope_struct_35_wait_closed;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed) < 0) __PYX_ERR(27, 66, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_35_wait_closed->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving)) __PYX_ERR(27, 99, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving) < 0) __PYX_ERR(27, 99, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving = &__pyx_type_6uvloop_4loop___pyx_scope_struct_36_start_serving;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving) < 0) __PYX_ERR(27, 99, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_36_start_serving->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever)) __PYX_ERR(27, 103, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever) < 0) __PYX_ERR(27, 103, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever = &__pyx_type_6uvloop_4loop___pyx_scope_struct_37_serve_forever;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever) < 0) __PYX_ERR(27, 103, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_37_serve_forever->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1_spec, NULL); if (unlikely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1)) __PYX_ERR(0, 3422, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1_spec, __pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1) < 0) __PYX_ERR(0, 3422, __pyx_L1_error)
  #else
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1 = &__pyx_type_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1) < 0) __PYX_ERR(0, 3422, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1->tp_dictoffset && __pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_6uvloop_4loop___pyx_scope_struct_38__test_coroutine_1->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_spec, NULL); if (unlikely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled)) __PYX_ERR(6, 66, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled_spec, __pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #else
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled = &__pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled) < 0) __PYX_ERR(6, 66, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled->tp_dictoffset && __pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype___pyx_scope_struct____Pyx_CFunc_6uvloop_4loop_4Loop_object__lParenLoop__comma_bint__rParen_to_py_4self_7enabled->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(28, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(28, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(29, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(29, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(30, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(30, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_loop(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_loop},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "loop",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initloop(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initloop(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_loop(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_loop(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_loop(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  double __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'loop' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("loop", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "loop" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_loop(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_uvloop__loop) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "uvloop.loop")) {
      if (unlikely((PyDict_SetItemString(modules, "uvloop.loop", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "uvloop/loop.pyx":3
 * # cython: language_level=3, embedsignature=True
 * 
 * import asyncio             # <<<<<<<<<<<<<<
 * cimport cython
 * 
 */
  __pyx_t_2 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":43
 * from cpython.pycapsule cimport PyCapsule_New, PyCapsule_GetPointer
 * 
 * from . import _noop             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_noop);
  __Pyx_GIVEREF(__pyx_n_s_noop);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_noop)) __PYX_ERR(0, 43, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s__10, __pyx_t_2, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_noop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_noop, __pyx_t_2) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":4
 * 
 * 
 * import asyncio, asyncio.log, asyncio.base_events, \             # <<<<<<<<<<<<<<
 *        asyncio.sslproto, asyncio.coroutines, \
 *        asyncio.futures, asyncio.transports
 */
  __pyx_t_3 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asyncio_log, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asyncio_base_events, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(31, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":5
 * 
 * import asyncio, asyncio.log, asyncio.base_events, \
 *        asyncio.sslproto, asyncio.coroutines, \             # <<<<<<<<<<<<<<
 *        asyncio.futures, asyncio.transports
 * import collections.abc
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asyncio_sslproto, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(31, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asyncio_coroutines, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(31, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":6
 * import asyncio, asyncio.log, asyncio.base_events, \
 *        asyncio.sslproto, asyncio.coroutines, \
 *        asyncio.futures, asyncio.transports             # <<<<<<<<<<<<<<
 * import collections.abc
 * import concurrent.futures
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asyncio_futures, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(31, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_asyncio_transports, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(31, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":7
 *        asyncio.sslproto, asyncio.coroutines, \
 *        asyncio.futures, asyncio.transports
 * import collections.abc             # <<<<<<<<<<<<<<
 * import concurrent.futures
 * import errno
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_collections_abc, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_collections, __pyx_t_3) < 0) __PYX_ERR(31, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":8
 *        asyncio.futures, asyncio.transports
 * import collections.abc
 * import concurrent.futures             # <<<<<<<<<<<<<<
 * import errno
 * import functools
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_concurrent_futures, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_concurrent, __pyx_t_3) < 0) __PYX_ERR(31, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":9
 * import collections.abc
 * import concurrent.futures
 * import errno             # <<<<<<<<<<<<<<
 * import functools
 * import gc
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_errno, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_errno, __pyx_t_3) < 0) __PYX_ERR(31, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":10
 * import concurrent.futures
 * import errno
 * import functools             # <<<<<<<<<<<<<<
 * import gc
 * import inspect
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_functools, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_functools, __pyx_t_3) < 0) __PYX_ERR(31, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":11
 * import errno
 * import functools
 * import gc             # <<<<<<<<<<<<<<
 * import inspect
 * import itertools
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_gc, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_gc, __pyx_t_3) < 0) __PYX_ERR(31, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":12
 * import functools
 * import gc
 * import inspect             # <<<<<<<<<<<<<<
 * import itertools
 * import os
 */
  __pyx_t_3 = __Pyx_patch_inspect(__Pyx_ImportDottedModule(__pyx_n_s_inspect, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inspect, __pyx_t_3) < 0) __PYX_ERR(31, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":13
 * import gc
 * import inspect
 * import itertools             # <<<<<<<<<<<<<<
 * import os
 * import signal
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_itertools, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itertools, __pyx_t_3) < 0) __PYX_ERR(31, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":14
 * import inspect
 * import itertools
 * import os             # <<<<<<<<<<<<<<
 * import signal
 * import socket
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_os, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_3) < 0) __PYX_ERR(31, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":15
 * import itertools
 * import os
 * import signal             # <<<<<<<<<<<<<<
 * import socket
 * import subprocess
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_signal, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_signal, __pyx_t_3) < 0) __PYX_ERR(31, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":16
 * import os
 * import signal
 * import socket             # <<<<<<<<<<<<<<
 * import subprocess
 * import ssl
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_socket_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_socket_2, __pyx_t_3) < 0) __PYX_ERR(31, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":17
 * import signal
 * import socket
 * import subprocess             # <<<<<<<<<<<<<<
 * import ssl
 * import stat
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_subprocess, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_subprocess, __pyx_t_3) < 0) __PYX_ERR(31, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":18
 * import socket
 * import subprocess
 * import ssl             # <<<<<<<<<<<<<<
 * import stat
 * import sys
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_ssl, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ssl, __pyx_t_3) < 0) __PYX_ERR(31, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":19
 * import subprocess
 * import ssl
 * import stat             # <<<<<<<<<<<<<<
 * import sys
 * import threading
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_stat, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stat, __pyx_t_3) < 0) __PYX_ERR(31, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":20
 * import ssl
 * import stat
 * import sys             # <<<<<<<<<<<<<<
 * import threading
 * import traceback
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_sys, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_3) < 0) __PYX_ERR(31, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":21
 * import stat
 * import sys
 * import threading             # <<<<<<<<<<<<<<
 * import traceback
 * import time
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_threading, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_threading, __pyx_t_3) < 0) __PYX_ERR(31, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":22
 * import sys
 * import threading
 * import traceback             # <<<<<<<<<<<<<<
 * import time
 * import warnings
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_traceback_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_traceback_2, __pyx_t_3) < 0) __PYX_ERR(31, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":23
 * import threading
 * import traceback
 * import time             # <<<<<<<<<<<<<<
 * import warnings
 * import weakref
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_time, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_3) < 0) __PYX_ERR(31, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":24
 * import traceback
 * import time
 * import warnings             # <<<<<<<<<<<<<<
 * import weakref
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_warnings, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_3) < 0) __PYX_ERR(31, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":25
 * import time
 * import warnings
 * import weakref             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_weakref, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_weakref, __pyx_t_3) < 0) __PYX_ERR(31, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":28
 * 
 * 
 * cdef aio_get_event_loop = asyncio.get_event_loop             # <<<<<<<<<<<<<<
 * cdef aio_CancelledError = asyncio.CancelledError
 * cdef aio_InvalidStateError = asyncio.InvalidStateError
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_event_loop); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_get_event_loop);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_get_event_loop, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":29
 * 
 * cdef aio_get_event_loop = asyncio.get_event_loop
 * cdef aio_CancelledError = asyncio.CancelledError             # <<<<<<<<<<<<<<
 * cdef aio_InvalidStateError = asyncio.InvalidStateError
 * cdef aio_TimeoutError = asyncio.TimeoutError
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_CancelledError); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_CancelledError);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_CancelledError, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":30
 * cdef aio_get_event_loop = asyncio.get_event_loop
 * cdef aio_CancelledError = asyncio.CancelledError
 * cdef aio_InvalidStateError = asyncio.InvalidStateError             # <<<<<<<<<<<<<<
 * cdef aio_TimeoutError = asyncio.TimeoutError
 * cdef aio_Future = asyncio.Future
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_InvalidStateError); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_InvalidStateError);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_InvalidStateError, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":31
 * cdef aio_CancelledError = asyncio.CancelledError
 * cdef aio_InvalidStateError = asyncio.InvalidStateError
 * cdef aio_TimeoutError = asyncio.TimeoutError             # <<<<<<<<<<<<<<
 * cdef aio_Future = asyncio.Future
 * cdef aio_Task = asyncio.Task
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TimeoutError); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_TimeoutError);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_TimeoutError, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":32
 * cdef aio_InvalidStateError = asyncio.InvalidStateError
 * cdef aio_TimeoutError = asyncio.TimeoutError
 * cdef aio_Future = asyncio.Future             # <<<<<<<<<<<<<<
 * cdef aio_Task = asyncio.Task
 * cdef aio_ensure_future = asyncio.ensure_future
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Future); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_Future);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_Future, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":33
 * cdef aio_TimeoutError = asyncio.TimeoutError
 * cdef aio_Future = asyncio.Future
 * cdef aio_Task = asyncio.Task             # <<<<<<<<<<<<<<
 * cdef aio_ensure_future = asyncio.ensure_future
 * cdef aio_gather = asyncio.gather
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Task); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_Task);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_Task, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":34
 * cdef aio_Future = asyncio.Future
 * cdef aio_Task = asyncio.Task
 * cdef aio_ensure_future = asyncio.ensure_future             # <<<<<<<<<<<<<<
 * cdef aio_gather = asyncio.gather
 * cdef aio_wait = asyncio.wait
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ensure_future); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_ensure_future);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_ensure_future, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":35
 * cdef aio_Task = asyncio.Task
 * cdef aio_ensure_future = asyncio.ensure_future
 * cdef aio_gather = asyncio.gather             # <<<<<<<<<<<<<<
 * cdef aio_wait = asyncio.wait
 * cdef aio_wrap_future = asyncio.wrap_future
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_gather); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_gather);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_gather, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":36
 * cdef aio_ensure_future = asyncio.ensure_future
 * cdef aio_gather = asyncio.gather
 * cdef aio_wait = asyncio.wait             # <<<<<<<<<<<<<<
 * cdef aio_wrap_future = asyncio.wrap_future
 * cdef aio_logger = asyncio.log.logger
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_wait); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_wait);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_wait, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":37
 * cdef aio_gather = asyncio.gather
 * cdef aio_wait = asyncio.wait
 * cdef aio_wrap_future = asyncio.wrap_future             # <<<<<<<<<<<<<<
 * cdef aio_logger = asyncio.log.logger
 * cdef aio_iscoroutine = asyncio.iscoroutine
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_wrap_future); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_wrap_future);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_wrap_future, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":38
 * cdef aio_wait = asyncio.wait
 * cdef aio_wrap_future = asyncio.wrap_future
 * cdef aio_logger = asyncio.log.logger             # <<<<<<<<<<<<<<
 * cdef aio_iscoroutine = asyncio.iscoroutine
 * cdef aio_iscoroutinefunction = asyncio.iscoroutinefunction
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_logger);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_logger, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":39
 * cdef aio_wrap_future = asyncio.wrap_future
 * cdef aio_logger = asyncio.log.logger
 * cdef aio_iscoroutine = asyncio.iscoroutine             # <<<<<<<<<<<<<<
 * cdef aio_iscoroutinefunction = asyncio.iscoroutinefunction
 * cdef aio_BaseProtocol = asyncio.BaseProtocol
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_iscoroutine); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_iscoroutine);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_iscoroutine, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":40
 * cdef aio_logger = asyncio.log.logger
 * cdef aio_iscoroutine = asyncio.iscoroutine
 * cdef aio_iscoroutinefunction = asyncio.iscoroutinefunction             # <<<<<<<<<<<<<<
 * cdef aio_BaseProtocol = asyncio.BaseProtocol
 * cdef aio_Protocol = asyncio.Protocol
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_iscoroutinefunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_iscoroutinefunction);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_iscoroutinefunction, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":41
 * cdef aio_iscoroutine = asyncio.iscoroutine
 * cdef aio_iscoroutinefunction = asyncio.iscoroutinefunction
 * cdef aio_BaseProtocol = asyncio.BaseProtocol             # <<<<<<<<<<<<<<
 * cdef aio_Protocol = asyncio.Protocol
 * cdef aio_isfuture = getattr(asyncio, 'isfuture', None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_BaseProtocol); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_BaseProtocol);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_BaseProtocol, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":42
 * cdef aio_iscoroutinefunction = asyncio.iscoroutinefunction
 * cdef aio_BaseProtocol = asyncio.BaseProtocol
 * cdef aio_Protocol = asyncio.Protocol             # <<<<<<<<<<<<<<
 * cdef aio_isfuture = getattr(asyncio, 'isfuture', None)
 * cdef aio_get_running_loop = getattr(asyncio, '_get_running_loop', None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Protocol); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_Protocol);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_Protocol, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":43
 * cdef aio_BaseProtocol = asyncio.BaseProtocol
 * cdef aio_Protocol = asyncio.Protocol
 * cdef aio_isfuture = getattr(asyncio, 'isfuture', None)             # <<<<<<<<<<<<<<
 * cdef aio_get_running_loop = getattr(asyncio, '_get_running_loop', None)
 * cdef aio_set_running_loop = getattr(asyncio, '_set_running_loop', None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_isfuture, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_isfuture);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_isfuture, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":44
 * cdef aio_Protocol = asyncio.Protocol
 * cdef aio_isfuture = getattr(asyncio, 'isfuture', None)
 * cdef aio_get_running_loop = getattr(asyncio, '_get_running_loop', None)             # <<<<<<<<<<<<<<
 * cdef aio_set_running_loop = getattr(asyncio, '_set_running_loop', None)
 * cdef aio_debug_wrapper = getattr(asyncio.coroutines, 'debug_wrapper', None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_get_running_loop, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_get_running_loop);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_get_running_loop, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":45
 * cdef aio_isfuture = getattr(asyncio, 'isfuture', None)
 * cdef aio_get_running_loop = getattr(asyncio, '_get_running_loop', None)
 * cdef aio_set_running_loop = getattr(asyncio, '_set_running_loop', None)             # <<<<<<<<<<<<<<
 * cdef aio_debug_wrapper = getattr(asyncio.coroutines, 'debug_wrapper', None)
 * cdef aio_AbstractChildWatcher = asyncio.AbstractChildWatcher
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_set_running_loop, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_set_running_loop);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_set_running_loop, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":46
 * cdef aio_get_running_loop = getattr(asyncio, '_get_running_loop', None)
 * cdef aio_set_running_loop = getattr(asyncio, '_set_running_loop', None)
 * cdef aio_debug_wrapper = getattr(asyncio.coroutines, 'debug_wrapper', None)             # <<<<<<<<<<<<<<
 * cdef aio_AbstractChildWatcher = asyncio.AbstractChildWatcher
 * cdef aio_Transport = asyncio.Transport
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_coroutines); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_debug_wrapper, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_debug_wrapper);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_debug_wrapper, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":47
 * cdef aio_set_running_loop = getattr(asyncio, '_set_running_loop', None)
 * cdef aio_debug_wrapper = getattr(asyncio.coroutines, 'debug_wrapper', None)
 * cdef aio_AbstractChildWatcher = asyncio.AbstractChildWatcher             # <<<<<<<<<<<<<<
 * cdef aio_Transport = asyncio.Transport
 * cdef aio_FlowControlMixin = asyncio.transports._FlowControlMixin
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_AbstractChildWatcher); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_AbstractChildWatcher);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_AbstractChildWatcher, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":48
 * cdef aio_debug_wrapper = getattr(asyncio.coroutines, 'debug_wrapper', None)
 * cdef aio_AbstractChildWatcher = asyncio.AbstractChildWatcher
 * cdef aio_Transport = asyncio.Transport             # <<<<<<<<<<<<<<
 * cdef aio_FlowControlMixin = asyncio.transports._FlowControlMixin
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Transport); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_Transport);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_Transport, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":49
 * cdef aio_AbstractChildWatcher = asyncio.AbstractChildWatcher
 * cdef aio_Transport = asyncio.Transport
 * cdef aio_FlowControlMixin = asyncio.transports._FlowControlMixin             # <<<<<<<<<<<<<<
 * 
 * cdef col_deque = collections.deque
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_transports); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_FlowControlMixin); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_aio_FlowControlMixin);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_aio_FlowControlMixin, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":51
 * cdef aio_FlowControlMixin = asyncio.transports._FlowControlMixin
 * 
 * cdef col_deque = collections.deque             # <<<<<<<<<<<<<<
 * cdef col_Iterable = collections.abc.Iterable
 * cdef col_Counter = collections.Counter
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_collections); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_col_deque);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_col_deque, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":52
 * 
 * cdef col_deque = collections.deque
 * cdef col_Iterable = collections.abc.Iterable             # <<<<<<<<<<<<<<
 * cdef col_Counter = collections.Counter
 * cdef col_OrderedDict = collections.OrderedDict
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_collections); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_abc); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_col_Iterable);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_col_Iterable, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":53
 * cdef col_deque = collections.deque
 * cdef col_Iterable = collections.abc.Iterable
 * cdef col_Counter = collections.Counter             # <<<<<<<<<<<<<<
 * cdef col_OrderedDict = collections.OrderedDict
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_collections); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_Counter); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_col_Counter);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_col_Counter, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":54
 * cdef col_Iterable = collections.abc.Iterable
 * cdef col_Counter = collections.Counter
 * cdef col_OrderedDict = collections.OrderedDict             # <<<<<<<<<<<<<<
 * 
 * cdef cc_ThreadPoolExecutor = concurrent.futures.ThreadPoolExecutor
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_collections); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_col_OrderedDict);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_col_OrderedDict, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":56
 * cdef col_OrderedDict = collections.OrderedDict
 * 
 * cdef cc_ThreadPoolExecutor = concurrent.futures.ThreadPoolExecutor             # <<<<<<<<<<<<<<
 * cdef cc_Future = concurrent.futures.Future
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_concurrent); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_futures); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ThreadPoolExecutor); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_cc_ThreadPoolExecutor);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_cc_ThreadPoolExecutor, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":57
 * 
 * cdef cc_ThreadPoolExecutor = concurrent.futures.ThreadPoolExecutor
 * cdef cc_Future = concurrent.futures.Future             # <<<<<<<<<<<<<<
 * 
 * cdef errno_EBADF = errno.EBADF
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_concurrent); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_futures); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Future); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_cc_Future);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_cc_Future, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":59
 * cdef cc_Future = concurrent.futures.Future
 * 
 * cdef errno_EBADF = errno.EBADF             # <<<<<<<<<<<<<<
 * cdef errno_EINVAL = errno.EINVAL
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From___pyx_anon_enum(EBADF); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_errno_EBADF);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_errno_EBADF, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":60
 * 
 * cdef errno_EBADF = errno.EBADF
 * cdef errno_EINVAL = errno.EINVAL             # <<<<<<<<<<<<<<
 * 
 * cdef ft_partial = functools.partial
 */
  __pyx_t_3 = __Pyx_PyInt_From___pyx_anon_enum(EINVAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_errno_EINVAL);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_errno_EINVAL, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":62
 * cdef errno_EINVAL = errno.EINVAL
 * 
 * cdef ft_partial = functools.partial             # <<<<<<<<<<<<<<
 * 
 * cdef gc_disable = gc.disable
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_functools); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ft_partial);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ft_partial, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":64
 * cdef ft_partial = functools.partial
 * 
 * cdef gc_disable = gc.disable             # <<<<<<<<<<<<<<
 * 
 * cdef iter_chain = itertools.chain
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_gc); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_disable); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_gc_disable);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_gc_disable, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":66
 * cdef gc_disable = gc.disable
 * 
 * cdef iter_chain = itertools.chain             # <<<<<<<<<<<<<<
 * cdef inspect_isgenerator = inspect.isgenerator
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_itertools); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_chain); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_iter_chain);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_iter_chain, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":67
 * 
 * cdef iter_chain = itertools.chain
 * cdef inspect_isgenerator = inspect.isgenerator             # <<<<<<<<<<<<<<
 * 
 * cdef int has_IPV6_V6ONLY = hasattr(socket, 'IPV6_V6ONLY')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_inspect); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_isgenerator); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_inspect_isgenerator);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_inspect_isgenerator, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":69
 * cdef inspect_isgenerator = inspect.isgenerator
 * 
 * cdef int has_IPV6_V6ONLY = hasattr(socket, 'IPV6_V6ONLY')             # <<<<<<<<<<<<<<
 * cdef int IPV6_V6ONLY = getattr(socket, 'IPV6_V6ONLY', -1)
 * cdef int has_SO_REUSEPORT = hasattr(socket, 'SO_REUSEPORT')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_HasAttr(__pyx_t_3, __pyx_n_u_IPV6_V6ONLY); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(31, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_has_IPV6_V6ONLY = __pyx_t_4;

  /* "uvloop/includes/stdlib.pxi":70
 * 
 * cdef int has_IPV6_V6ONLY = hasattr(socket, 'IPV6_V6ONLY')
 * cdef int IPV6_V6ONLY = getattr(socket, 'IPV6_V6ONLY', -1)             # <<<<<<<<<<<<<<
 * cdef int has_SO_REUSEPORT = hasattr(socket, 'SO_REUSEPORT')
 * cdef int SO_REUSEPORT = getattr(socket, 'SO_REUSEPORT', 0)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_IPV6_V6ONLY, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_IPV6_V6ONLY = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":71
 * cdef int has_IPV6_V6ONLY = hasattr(socket, 'IPV6_V6ONLY')
 * cdef int IPV6_V6ONLY = getattr(socket, 'IPV6_V6ONLY', -1)
 * cdef int has_SO_REUSEPORT = hasattr(socket, 'SO_REUSEPORT')             # <<<<<<<<<<<<<<
 * cdef int SO_REUSEPORT = getattr(socket, 'SO_REUSEPORT', 0)
 * cdef int SO_BROADCAST = getattr(socket, 'SO_BROADCAST')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_HasAttr(__pyx_t_2, __pyx_n_u_SO_REUSEPORT); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(31, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_has_SO_REUSEPORT = __pyx_t_4;

  /* "uvloop/includes/stdlib.pxi":72
 * cdef int IPV6_V6ONLY = getattr(socket, 'IPV6_V6ONLY', -1)
 * cdef int has_SO_REUSEPORT = hasattr(socket, 'SO_REUSEPORT')
 * cdef int SO_REUSEPORT = getattr(socket, 'SO_REUSEPORT', 0)             # <<<<<<<<<<<<<<
 * cdef int SO_BROADCAST = getattr(socket, 'SO_BROADCAST')
 * cdef int SOCK_NONBLOCK = getattr(socket, 'SOCK_NONBLOCK', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_SO_REUSEPORT, __pyx_int_0); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_SO_REUSEPORT = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":73
 * cdef int has_SO_REUSEPORT = hasattr(socket, 'SO_REUSEPORT')
 * cdef int SO_REUSEPORT = getattr(socket, 'SO_REUSEPORT', 0)
 * cdef int SO_BROADCAST = getattr(socket, 'SO_BROADCAST')             # <<<<<<<<<<<<<<
 * cdef int SOCK_NONBLOCK = getattr(socket, 'SOCK_NONBLOCK', -1)
 * cdef int socket_AI_CANONNAME = getattr(socket, 'AI_CANONNAME')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_3, __pyx_n_u_SO_BROADCAST); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_SO_BROADCAST = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":74
 * cdef int SO_REUSEPORT = getattr(socket, 'SO_REUSEPORT', 0)
 * cdef int SO_BROADCAST = getattr(socket, 'SO_BROADCAST')
 * cdef int SOCK_NONBLOCK = getattr(socket, 'SOCK_NONBLOCK', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_AI_CANONNAME = getattr(socket, 'AI_CANONNAME')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_SOCK_NONBLOCK, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_SOCK_NONBLOCK = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":75
 * cdef int SO_BROADCAST = getattr(socket, 'SO_BROADCAST')
 * cdef int SOCK_NONBLOCK = getattr(socket, 'SOCK_NONBLOCK', -1)
 * cdef int socket_AI_CANONNAME = getattr(socket, 'AI_CANONNAME')             # <<<<<<<<<<<<<<
 * 
 * cdef socket_gaierror = socket.gaierror
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_3, __pyx_n_u_AI_CANONNAME); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_AI_CANONNAME = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":77
 * cdef int socket_AI_CANONNAME = getattr(socket, 'AI_CANONNAME')
 * 
 * cdef socket_gaierror = socket.gaierror             # <<<<<<<<<<<<<<
 * cdef socket_error = socket.error
 * cdef socket_timeout = socket.timeout
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_gaierror); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_gaierror);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_gaierror, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":78
 * 
 * cdef socket_gaierror = socket.gaierror
 * cdef socket_error = socket.error             # <<<<<<<<<<<<<<
 * cdef socket_timeout = socket.timeout
 * cdef socket_socket = socket.socket
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_error);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_error, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":79
 * cdef socket_gaierror = socket.gaierror
 * cdef socket_error = socket.error
 * cdef socket_timeout = socket.timeout             # <<<<<<<<<<<<<<
 * cdef socket_socket = socket.socket
 * cdef socket_socketpair = socket.socketpair
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_timeout);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_timeout, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":80
 * cdef socket_error = socket.error
 * cdef socket_timeout = socket.timeout
 * cdef socket_socket = socket.socket             # <<<<<<<<<<<<<<
 * cdef socket_socketpair = socket.socketpair
 * cdef socket_getservbyname = socket.getservbyname
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_socket);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_socket, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":81
 * cdef socket_timeout = socket.timeout
 * cdef socket_socket = socket.socket
 * cdef socket_socketpair = socket.socketpair             # <<<<<<<<<<<<<<
 * cdef socket_getservbyname = socket.getservbyname
 * cdef socket_AddressFamily = socket.AddressFamily
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_socketpair); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_socketpair);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_socketpair, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":82
 * cdef socket_socket = socket.socket
 * cdef socket_socketpair = socket.socketpair
 * cdef socket_getservbyname = socket.getservbyname             # <<<<<<<<<<<<<<
 * cdef socket_AddressFamily = socket.AddressFamily
 * cdef socket_SocketKind = socket.SocketKind
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_getservbyname); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_getservbyname);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_getservbyname, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":83
 * cdef socket_socketpair = socket.socketpair
 * cdef socket_getservbyname = socket.getservbyname
 * cdef socket_AddressFamily = socket.AddressFamily             # <<<<<<<<<<<<<<
 * cdef socket_SocketKind = socket.SocketKind
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_AddressFamily); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_AddressFamily);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_AddressFamily, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":84
 * cdef socket_getservbyname = socket.getservbyname
 * cdef socket_AddressFamily = socket.AddressFamily
 * cdef socket_SocketKind = socket.SocketKind             # <<<<<<<<<<<<<<
 * 
 * cdef int socket_EAI_ADDRFAMILY = getattr(socket, 'EAI_ADDRFAMILY', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SocketKind); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_socket_SocketKind);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_socket_SocketKind, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":86
 * cdef socket_SocketKind = socket.SocketKind
 * 
 * cdef int socket_EAI_ADDRFAMILY = getattr(socket, 'EAI_ADDRFAMILY', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_AGAIN      = getattr(socket, 'EAI_AGAIN', -1)
 * cdef int socket_EAI_BADFLAGS   = getattr(socket, 'EAI_BADFLAGS', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_EAI_ADDRFAMILY, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_ADDRFAMILY = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":87
 * 
 * cdef int socket_EAI_ADDRFAMILY = getattr(socket, 'EAI_ADDRFAMILY', -1)
 * cdef int socket_EAI_AGAIN      = getattr(socket, 'EAI_AGAIN', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_BADFLAGS   = getattr(socket, 'EAI_BADFLAGS', -1)
 * cdef int socket_EAI_BADHINTS   = getattr(socket, 'EAI_BADHINTS', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_EAI_AGAIN, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_AGAIN = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":88
 * cdef int socket_EAI_ADDRFAMILY = getattr(socket, 'EAI_ADDRFAMILY', -1)
 * cdef int socket_EAI_AGAIN      = getattr(socket, 'EAI_AGAIN', -1)
 * cdef int socket_EAI_BADFLAGS   = getattr(socket, 'EAI_BADFLAGS', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_BADHINTS   = getattr(socket, 'EAI_BADHINTS', -1)
 * cdef int socket_EAI_CANCELED   = getattr(socket, 'EAI_CANCELED', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_EAI_BADFLAGS, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_BADFLAGS = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":89
 * cdef int socket_EAI_AGAIN      = getattr(socket, 'EAI_AGAIN', -1)
 * cdef int socket_EAI_BADFLAGS   = getattr(socket, 'EAI_BADFLAGS', -1)
 * cdef int socket_EAI_BADHINTS   = getattr(socket, 'EAI_BADHINTS', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_CANCELED   = getattr(socket, 'EAI_CANCELED', -1)
 * cdef int socket_EAI_FAIL       = getattr(socket, 'EAI_FAIL', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_EAI_BADHINTS, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_BADHINTS = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":90
 * cdef int socket_EAI_BADFLAGS   = getattr(socket, 'EAI_BADFLAGS', -1)
 * cdef int socket_EAI_BADHINTS   = getattr(socket, 'EAI_BADHINTS', -1)
 * cdef int socket_EAI_CANCELED   = getattr(socket, 'EAI_CANCELED', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_FAIL       = getattr(socket, 'EAI_FAIL', -1)
 * cdef int socket_EAI_FAMILY     = getattr(socket, 'EAI_FAMILY', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_EAI_CANCELED, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_CANCELED = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":91
 * cdef int socket_EAI_BADHINTS   = getattr(socket, 'EAI_BADHINTS', -1)
 * cdef int socket_EAI_CANCELED   = getattr(socket, 'EAI_CANCELED', -1)
 * cdef int socket_EAI_FAIL       = getattr(socket, 'EAI_FAIL', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_FAMILY     = getattr(socket, 'EAI_FAMILY', -1)
 * cdef int socket_EAI_MEMORY     = getattr(socket, 'EAI_MEMORY', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_EAI_FAIL, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_FAIL = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":92
 * cdef int socket_EAI_CANCELED   = getattr(socket, 'EAI_CANCELED', -1)
 * cdef int socket_EAI_FAIL       = getattr(socket, 'EAI_FAIL', -1)
 * cdef int socket_EAI_FAMILY     = getattr(socket, 'EAI_FAMILY', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_MEMORY     = getattr(socket, 'EAI_MEMORY', -1)
 * cdef int socket_EAI_NODATA     = getattr(socket, 'EAI_NODATA', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_EAI_FAMILY, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_FAMILY = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":93
 * cdef int socket_EAI_FAIL       = getattr(socket, 'EAI_FAIL', -1)
 * cdef int socket_EAI_FAMILY     = getattr(socket, 'EAI_FAMILY', -1)
 * cdef int socket_EAI_MEMORY     = getattr(socket, 'EAI_MEMORY', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_NODATA     = getattr(socket, 'EAI_NODATA', -1)
 * cdef int socket_EAI_NONAME     = getattr(socket, 'EAI_NONAME', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_EAI_MEMORY, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_MEMORY = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":94
 * cdef int socket_EAI_FAMILY     = getattr(socket, 'EAI_FAMILY', -1)
 * cdef int socket_EAI_MEMORY     = getattr(socket, 'EAI_MEMORY', -1)
 * cdef int socket_EAI_NODATA     = getattr(socket, 'EAI_NODATA', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_NONAME     = getattr(socket, 'EAI_NONAME', -1)
 * cdef int socket_EAI_OVERFLOW   = getattr(socket, 'EAI_OVERFLOW', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_EAI_NODATA, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_NODATA = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":95
 * cdef int socket_EAI_MEMORY     = getattr(socket, 'EAI_MEMORY', -1)
 * cdef int socket_EAI_NODATA     = getattr(socket, 'EAI_NODATA', -1)
 * cdef int socket_EAI_NONAME     = getattr(socket, 'EAI_NONAME', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_OVERFLOW   = getattr(socket, 'EAI_OVERFLOW', -1)
 * cdef int socket_EAI_PROTOCOL   = getattr(socket, 'EAI_PROTOCOL', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_EAI_NONAME, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_NONAME = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":96
 * cdef int socket_EAI_NODATA     = getattr(socket, 'EAI_NODATA', -1)
 * cdef int socket_EAI_NONAME     = getattr(socket, 'EAI_NONAME', -1)
 * cdef int socket_EAI_OVERFLOW   = getattr(socket, 'EAI_OVERFLOW', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_PROTOCOL   = getattr(socket, 'EAI_PROTOCOL', -1)
 * cdef int socket_EAI_SERVICE    = getattr(socket, 'EAI_SERVICE', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_EAI_OVERFLOW, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_OVERFLOW = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":97
 * cdef int socket_EAI_NONAME     = getattr(socket, 'EAI_NONAME', -1)
 * cdef int socket_EAI_OVERFLOW   = getattr(socket, 'EAI_OVERFLOW', -1)
 * cdef int socket_EAI_PROTOCOL   = getattr(socket, 'EAI_PROTOCOL', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_SERVICE    = getattr(socket, 'EAI_SERVICE', -1)
 * cdef int socket_EAI_SOCKTYPE   = getattr(socket, 'EAI_SOCKTYPE', -1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_EAI_PROTOCOL, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_PROTOCOL = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":98
 * cdef int socket_EAI_OVERFLOW   = getattr(socket, 'EAI_OVERFLOW', -1)
 * cdef int socket_EAI_PROTOCOL   = getattr(socket, 'EAI_PROTOCOL', -1)
 * cdef int socket_EAI_SERVICE    = getattr(socket, 'EAI_SERVICE', -1)             # <<<<<<<<<<<<<<
 * cdef int socket_EAI_SOCKTYPE   = getattr(socket, 'EAI_SOCKTYPE', -1)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_EAI_SERVICE, __pyx_int_neg_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 98, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_SERVICE = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":99
 * cdef int socket_EAI_PROTOCOL   = getattr(socket, 'EAI_PROTOCOL', -1)
 * cdef int socket_EAI_SERVICE    = getattr(socket, 'EAI_SERVICE', -1)
 * cdef int socket_EAI_SOCKTYPE   = getattr(socket, 'EAI_SOCKTYPE', -1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_socket_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_EAI_SOCKTYPE, __pyx_int_neg_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_6uvloop_4loop_socket_EAI_SOCKTYPE = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":102
 * 
 * 
 * cdef str os_name = os.name             # <<<<<<<<<<<<<<
 * cdef os_environ = os.environ
 * cdef os_dup = os.dup
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(31, 102, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_name);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_name, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":103
 * 
 * cdef str os_name = os.name
 * cdef os_environ = os.environ             # <<<<<<<<<<<<<<
 * cdef os_dup = os.dup
 * cdef os_set_inheritable = os.set_inheritable
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_environ); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_environ);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_environ, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":104
 * cdef str os_name = os.name
 * cdef os_environ = os.environ
 * cdef os_dup = os.dup             # <<<<<<<<<<<<<<
 * cdef os_set_inheritable = os.set_inheritable
 * cdef os_get_inheritable = os.get_inheritable
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_dup); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_dup);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_dup, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":105
 * cdef os_environ = os.environ
 * cdef os_dup = os.dup
 * cdef os_set_inheritable = os.set_inheritable             # <<<<<<<<<<<<<<
 * cdef os_get_inheritable = os.get_inheritable
 * cdef os_close = os.close
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_set_inheritable); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_set_inheritable);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_set_inheritable, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":106
 * cdef os_dup = os.dup
 * cdef os_set_inheritable = os.set_inheritable
 * cdef os_get_inheritable = os.get_inheritable             # <<<<<<<<<<<<<<
 * cdef os_close = os.close
 * cdef os_open = os.open
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get_inheritable); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_get_inheritable);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_get_inheritable, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":107
 * cdef os_set_inheritable = os.set_inheritable
 * cdef os_get_inheritable = os.get_inheritable
 * cdef os_close = os.close             # <<<<<<<<<<<<<<
 * cdef os_open = os.open
 * cdef os_devnull = os.devnull
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_close);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_close, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":108
 * cdef os_get_inheritable = os.get_inheritable
 * cdef os_close = os.close
 * cdef os_open = os.open             # <<<<<<<<<<<<<<
 * cdef os_devnull = os.devnull
 * cdef os_O_RDWR = os.O_RDWR
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_open); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_open);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_open, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":109
 * cdef os_close = os.close
 * cdef os_open = os.open
 * cdef os_devnull = os.devnull             # <<<<<<<<<<<<<<
 * cdef os_O_RDWR = os.O_RDWR
 * cdef os_pipe = os.pipe
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_devnull); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_devnull);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_devnull, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":110
 * cdef os_open = os.open
 * cdef os_devnull = os.devnull
 * cdef os_O_RDWR = os.O_RDWR             # <<<<<<<<<<<<<<
 * cdef os_pipe = os.pipe
 * cdef os_read = os.read
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_O_RDWR); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_O_RDWR);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_O_RDWR, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":111
 * cdef os_devnull = os.devnull
 * cdef os_O_RDWR = os.O_RDWR
 * cdef os_pipe = os.pipe             # <<<<<<<<<<<<<<
 * cdef os_read = os.read
 * cdef os_remove = os.remove
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_pipe); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_pipe);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_pipe, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":112
 * cdef os_O_RDWR = os.O_RDWR
 * cdef os_pipe = os.pipe
 * cdef os_read = os.read             # <<<<<<<<<<<<<<
 * cdef os_remove = os.remove
 * cdef os_stat = os.stat
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_read); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_read);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_read, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":113
 * cdef os_pipe = os.pipe
 * cdef os_read = os.read
 * cdef os_remove = os.remove             # <<<<<<<<<<<<<<
 * cdef os_stat = os.stat
 * cdef os_unlink = os.unlink
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_remove);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_remove, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":114
 * cdef os_read = os.read
 * cdef os_remove = os.remove
 * cdef os_stat = os.stat             # <<<<<<<<<<<<<<
 * cdef os_unlink = os.unlink
 * cdef os_fspath = os.fspath
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_stat); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_stat);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_stat, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":115
 * cdef os_remove = os.remove
 * cdef os_stat = os.stat
 * cdef os_unlink = os.unlink             # <<<<<<<<<<<<<<
 * cdef os_fspath = os.fspath
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_unlink); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_unlink);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_unlink, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":116
 * cdef os_stat = os.stat
 * cdef os_unlink = os.unlink
 * cdef os_fspath = os.fspath             # <<<<<<<<<<<<<<
 * 
 * cdef stat_S_ISSOCK = stat.S_ISSOCK
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_fspath); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_os_fspath);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_os_fspath, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":118
 * cdef os_fspath = os.fspath
 * 
 * cdef stat_S_ISSOCK = stat.S_ISSOCK             # <<<<<<<<<<<<<<
 * 
 * cdef sys_ignore_environment = sys.flags.ignore_environment
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_stat); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_S_ISSOCK); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_stat_S_ISSOCK);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_stat_S_ISSOCK, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":120
 * cdef stat_S_ISSOCK = stat.S_ISSOCK
 * 
 * cdef sys_ignore_environment = sys.flags.ignore_environment             # <<<<<<<<<<<<<<
 * cdef sys_dev_mode = sys.flags.dev_mode
 * cdef sys_exc_info = sys.exc_info
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ignore_environment); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_ignore_environment);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_ignore_environment, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":121
 * 
 * cdef sys_ignore_environment = sys.flags.ignore_environment
 * cdef sys_dev_mode = sys.flags.dev_mode             # <<<<<<<<<<<<<<
 * cdef sys_exc_info = sys.exc_info
 * cdef sys_set_coroutine_wrapper = getattr(sys, 'set_coroutine_wrapper', None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dev_mode); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_dev_mode);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_dev_mode, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":122
 * cdef sys_ignore_environment = sys.flags.ignore_environment
 * cdef sys_dev_mode = sys.flags.dev_mode
 * cdef sys_exc_info = sys.exc_info             # <<<<<<<<<<<<<<
 * cdef sys_set_coroutine_wrapper = getattr(sys, 'set_coroutine_wrapper', None)
 * cdef sys_get_coroutine_wrapper = getattr(sys, 'get_coroutine_wrapper', None)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_exc_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_exc_info);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_exc_info, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":123
 * cdef sys_dev_mode = sys.flags.dev_mode
 * cdef sys_exc_info = sys.exc_info
 * cdef sys_set_coroutine_wrapper = getattr(sys, 'set_coroutine_wrapper', None)             # <<<<<<<<<<<<<<
 * cdef sys_get_coroutine_wrapper = getattr(sys, 'get_coroutine_wrapper', None)
 * cdef sys_getframe = sys._getframe
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_GetAttr3(__pyx_t_3, __pyx_n_u_set_coroutine_wrapper, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_set_coroutine_wrapper);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_set_coroutine_wrapper, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":124
 * cdef sys_exc_info = sys.exc_info
 * cdef sys_set_coroutine_wrapper = getattr(sys, 'set_coroutine_wrapper', None)
 * cdef sys_get_coroutine_wrapper = getattr(sys, 'get_coroutine_wrapper', None)             # <<<<<<<<<<<<<<
 * cdef sys_getframe = sys._getframe
 * cdef sys_version_info = sys.version_info
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetAttr3(__pyx_t_2, __pyx_n_u_get_coroutine_wrapper, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_get_coroutine_wrapper);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_get_coroutine_wrapper, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":125
 * cdef sys_set_coroutine_wrapper = getattr(sys, 'set_coroutine_wrapper', None)
 * cdef sys_get_coroutine_wrapper = getattr(sys, 'get_coroutine_wrapper', None)
 * cdef sys_getframe = sys._getframe             # <<<<<<<<<<<<<<
 * cdef sys_version_info = sys.version_info
 * cdef sys_getfilesystemencoding = sys.getfilesystemencoding
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_getframe); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_getframe);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_getframe, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":126
 * cdef sys_get_coroutine_wrapper = getattr(sys, 'get_coroutine_wrapper', None)
 * cdef sys_getframe = sys._getframe
 * cdef sys_version_info = sys.version_info             # <<<<<<<<<<<<<<
 * cdef sys_getfilesystemencoding = sys.getfilesystemencoding
 * cdef str sys_platform = sys.platform
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_version_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_version_info);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_version_info, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":127
 * cdef sys_getframe = sys._getframe
 * cdef sys_version_info = sys.version_info
 * cdef sys_getfilesystemencoding = sys.getfilesystemencoding             # <<<<<<<<<<<<<<
 * cdef str sys_platform = sys.platform
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_getfilesystemencoding); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_getfilesystemencoding);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_getfilesystemencoding, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":128
 * cdef sys_version_info = sys.version_info
 * cdef sys_getfilesystemencoding = sys.getfilesystemencoding
 * cdef str sys_platform = sys.platform             # <<<<<<<<<<<<<<
 * 
 * cdef ssl_SSLContext = ssl.SSLContext
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_platform); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(31, 128, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_sys_platform);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sys_platform, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":130
 * cdef str sys_platform = sys.platform
 * 
 * cdef ssl_SSLContext = ssl.SSLContext             # <<<<<<<<<<<<<<
 * cdef ssl_MemoryBIO = ssl.MemoryBIO
 * cdef ssl_create_default_context = ssl.create_default_context
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ssl); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SSLContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ssl_SSLContext);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ssl_SSLContext, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":131
 * 
 * cdef ssl_SSLContext = ssl.SSLContext
 * cdef ssl_MemoryBIO = ssl.MemoryBIO             # <<<<<<<<<<<<<<
 * cdef ssl_create_default_context = ssl.create_default_context
 * cdef ssl_SSLError = ssl.SSLError
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ssl); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MemoryBIO); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ssl_MemoryBIO);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ssl_MemoryBIO, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":132
 * cdef ssl_SSLContext = ssl.SSLContext
 * cdef ssl_MemoryBIO = ssl.MemoryBIO
 * cdef ssl_create_default_context = ssl.create_default_context             # <<<<<<<<<<<<<<
 * cdef ssl_SSLError = ssl.SSLError
 * cdef ssl_SSLAgainErrors = (ssl.SSLWantReadError, ssl.SSLSyscallError)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ssl); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_create_default_context); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ssl_create_default_context);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ssl_create_default_context, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/includes/stdlib.pxi":133
 * cdef ssl_MemoryBIO = ssl.MemoryBIO
 * cdef ssl_create_default_context = ssl.create_default_context
 * cdef ssl_SSLError = ssl.SSLError             # <<<<<<<<<<<<<<
 * cdef ssl_SSLAgainErrors = (ssl.SSLWantReadError, ssl.SSLSyscallError)
 * cdef ssl_SSLZeroReturnError = ssl.SSLZeroReturnError
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ssl); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SSLError); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ssl_SSLError);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ssl_SSLError, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":134
 * cdef ssl_create_default_context = ssl.create_default_context
 * cdef ssl_SSLError = ssl.SSLError
 * cdef ssl_SSLAgainErrors = (ssl.SSLWantReadError, ssl.SSLSyscallError)             # <<<<<<<<<<<<<<
 * cdef ssl_SSLZeroReturnError = ssl.SSLZeroReturnError
 * cdef ssl_CertificateError = ssl.CertificateError
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ssl); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SSLWantReadError); if (unlikely(!__pyx_t_2)) __PYX_ERR(31, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ssl); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SSLSyscallError); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(31, 134, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6)) __PYX_ERR(31, 134, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ssl_SSLAgainErrors, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":135
 * cdef ssl_SSLError = ssl.SSLError
 * cdef ssl_SSLAgainErrors = (ssl.SSLWantReadError, ssl.SSLSyscallError)
 * cdef ssl_SSLZeroReturnError = ssl.SSLZeroReturnError             # <<<<<<<<<<<<<<
 * cdef ssl_CertificateError = ssl.CertificateError
 * cdef int ssl_SSL_ERROR_WANT_READ = ssl.SSL_ERROR_WANT_READ
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ssl); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SSLZeroReturnError); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ssl_SSLZeroReturnError);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ssl_SSLZeroReturnError, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":136
 * cdef ssl_SSLAgainErrors = (ssl.SSLWantReadError, ssl.SSLSyscallError)
 * cdef ssl_SSLZeroReturnError = ssl.SSLZeroReturnError
 * cdef ssl_CertificateError = ssl.CertificateError             # <<<<<<<<<<<<<<
 * cdef int ssl_SSL_ERROR_WANT_READ = ssl.SSL_ERROR_WANT_READ
 * cdef int ssl_SSL_ERROR_WANT_WRITE = ssl.SSL_ERROR_WANT_WRITE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ssl); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_CertificateError); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_ssl_CertificateError);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_ssl_CertificateError, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":137
 * cdef ssl_SSLZeroReturnError = ssl.SSLZeroReturnError
 * cdef ssl_CertificateError = ssl.CertificateError
 * cdef int ssl_SSL_ERROR_WANT_READ = ssl.SSL_ERROR_WANT_READ             # <<<<<<<<<<<<<<
 * cdef int ssl_SSL_ERROR_WANT_WRITE = ssl.SSL_ERROR_WANT_WRITE
 * cdef int ssl_SSL_ERROR_SYSCALL = ssl.SSL_ERROR_SYSCALL
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ssl); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SSL_ERROR_WANT_READ); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_6uvloop_4loop_ssl_SSL_ERROR_WANT_READ = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":138
 * cdef ssl_CertificateError = ssl.CertificateError
 * cdef int ssl_SSL_ERROR_WANT_READ = ssl.SSL_ERROR_WANT_READ
 * cdef int ssl_SSL_ERROR_WANT_WRITE = ssl.SSL_ERROR_WANT_WRITE             # <<<<<<<<<<<<<<
 * cdef int ssl_SSL_ERROR_SYSCALL = ssl.SSL_ERROR_SYSCALL
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ssl); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_SSL_ERROR_WANT_WRITE); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_ssl_SSL_ERROR_WANT_WRITE = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":139
 * cdef int ssl_SSL_ERROR_WANT_READ = ssl.SSL_ERROR_WANT_READ
 * cdef int ssl_SSL_ERROR_WANT_WRITE = ssl.SSL_ERROR_WANT_WRITE
 * cdef int ssl_SSL_ERROR_SYSCALL = ssl.SSL_ERROR_SYSCALL             # <<<<<<<<<<<<<<
 * 
 * cdef threading_Thread = threading.Thread
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_ssl); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SSL_ERROR_SYSCALL); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_6uvloop_4loop_ssl_SSL_ERROR_SYSCALL = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":141
 * cdef int ssl_SSL_ERROR_SYSCALL = ssl.SSL_ERROR_SYSCALL
 * 
 * cdef threading_Thread = threading.Thread             # <<<<<<<<<<<<<<
 * cdef threading_main_thread = threading.main_thread
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_threading); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_Thread); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_threading_Thread);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_threading_Thread, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":142
 * 
 * cdef threading_Thread = threading.Thread
 * cdef threading_main_thread = threading.main_thread             # <<<<<<<<<<<<<<
 * 
 * cdef int subprocess_PIPE = subprocess.PIPE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_threading); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_main_thread); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_threading_main_thread);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_threading_main_thread, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":144
 * cdef threading_main_thread = threading.main_thread
 * 
 * cdef int subprocess_PIPE = subprocess.PIPE             # <<<<<<<<<<<<<<
 * cdef int subprocess_STDOUT = subprocess.STDOUT
 * cdef int subprocess_DEVNULL = subprocess.DEVNULL
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_PIPE); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_subprocess_PIPE = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":145
 * 
 * cdef int subprocess_PIPE = subprocess.PIPE
 * cdef int subprocess_STDOUT = subprocess.STDOUT             # <<<<<<<<<<<<<<
 * cdef int subprocess_DEVNULL = subprocess.DEVNULL
 * cdef subprocess_SubprocessError = subprocess.SubprocessError
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_STDOUT); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_6uvloop_4loop_subprocess_STDOUT = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":146
 * cdef int subprocess_PIPE = subprocess.PIPE
 * cdef int subprocess_STDOUT = subprocess.STDOUT
 * cdef int subprocess_DEVNULL = subprocess.DEVNULL             # <<<<<<<<<<<<<<
 * cdef subprocess_SubprocessError = subprocess.SubprocessError
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DEVNULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_subprocess_DEVNULL = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":147
 * cdef int subprocess_STDOUT = subprocess.STDOUT
 * cdef int subprocess_DEVNULL = subprocess.DEVNULL
 * cdef subprocess_SubprocessError = subprocess.SubprocessError             # <<<<<<<<<<<<<<
 * 
 * cdef int signal_NSIG = signal.NSIG
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_subprocess); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SubprocessError); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_subprocess_SubprocessError);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_subprocess_SubprocessError, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":149
 * cdef subprocess_SubprocessError = subprocess.SubprocessError
 * 
 * cdef int signal_NSIG = signal.NSIG             # <<<<<<<<<<<<<<
 * cdef signal_signal = signal.signal
 * cdef signal_siginterrupt = signal.siginterrupt
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signal); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_NSIG); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(31, 149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_6uvloop_4loop_signal_NSIG = __pyx_t_5;

  /* "uvloop/includes/stdlib.pxi":150
 * 
 * cdef int signal_NSIG = signal.NSIG
 * cdef signal_signal = signal.signal             # <<<<<<<<<<<<<<
 * cdef signal_siginterrupt = signal.siginterrupt
 * cdef signal_set_wakeup_fd = signal.set_wakeup_fd
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signal); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_signal); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_signal_signal);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_signal_signal, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":151
 * cdef int signal_NSIG = signal.NSIG
 * cdef signal_signal = signal.signal
 * cdef signal_siginterrupt = signal.siginterrupt             # <<<<<<<<<<<<<<
 * cdef signal_set_wakeup_fd = signal.set_wakeup_fd
 * cdef signal_default_int_handler = signal.default_int_handler
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signal); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_siginterrupt); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_signal_siginterrupt);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_signal_siginterrupt, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":152
 * cdef signal_signal = signal.signal
 * cdef signal_siginterrupt = signal.siginterrupt
 * cdef signal_set_wakeup_fd = signal.set_wakeup_fd             # <<<<<<<<<<<<<<
 * cdef signal_default_int_handler = signal.default_int_handler
 * cdef signal_SIG_DFL = signal.SIG_DFL
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signal); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_set_wakeup_fd); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_signal_set_wakeup_fd);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_signal_set_wakeup_fd, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":153
 * cdef signal_siginterrupt = signal.siginterrupt
 * cdef signal_set_wakeup_fd = signal.set_wakeup_fd
 * cdef signal_default_int_handler = signal.default_int_handler             # <<<<<<<<<<<<<<
 * cdef signal_SIG_DFL = signal.SIG_DFL
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_signal); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_default_int_handler); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_signal_default_int_handler);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_signal_default_int_handler, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":154
 * cdef signal_set_wakeup_fd = signal.set_wakeup_fd
 * cdef signal_default_int_handler = signal.default_int_handler
 * cdef signal_SIG_DFL = signal.SIG_DFL             # <<<<<<<<<<<<<<
 * 
 * cdef time_sleep = time.sleep
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_signal); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SIG_DFL); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_signal_SIG_DFL);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_signal_SIG_DFL, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":156
 * cdef signal_SIG_DFL = signal.SIG_DFL
 * 
 * cdef time_sleep = time.sleep             # <<<<<<<<<<<<<<
 * cdef time_monotonic = time.monotonic
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_sleep); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_time_sleep);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_time_sleep, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":157
 * 
 * cdef time_sleep = time.sleep
 * cdef time_monotonic = time.monotonic             # <<<<<<<<<<<<<<
 * 
 * cdef tb_StackSummary = traceback.StackSummary
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_monotonic); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_time_monotonic);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_time_monotonic, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":159
 * cdef time_monotonic = time.monotonic
 * 
 * cdef tb_StackSummary = traceback.StackSummary             # <<<<<<<<<<<<<<
 * cdef tb_walk_stack = traceback.walk_stack
 * cdef tb_format_list = traceback.format_list
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_traceback_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_StackSummary); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_tb_StackSummary);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_tb_StackSummary, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":160
 * 
 * cdef tb_StackSummary = traceback.StackSummary
 * cdef tb_walk_stack = traceback.walk_stack             # <<<<<<<<<<<<<<
 * cdef tb_format_list = traceback.format_list
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_traceback_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_walk_stack); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_tb_walk_stack);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_tb_walk_stack, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":161
 * cdef tb_StackSummary = traceback.StackSummary
 * cdef tb_walk_stack = traceback.walk_stack
 * cdef tb_format_list = traceback.format_list             # <<<<<<<<<<<<<<
 * 
 * cdef warnings_warn = warnings.warn
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_traceback_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_format_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_tb_format_list);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_tb_format_list, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":163
 * cdef tb_format_list = traceback.format_list
 * 
 * cdef warnings_warn = warnings.warn             # <<<<<<<<<<<<<<
 * 
 * cdef weakref_WeakValueDictionary = weakref.WeakValueDictionary
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_warnings); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warn); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_warnings_warn);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_warnings_warn, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":165
 * cdef warnings_warn = warnings.warn
 * 
 * cdef weakref_WeakValueDictionary = weakref.WeakValueDictionary             # <<<<<<<<<<<<<<
 * cdef weakref_WeakSet = weakref.WeakSet
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_weakref); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_WeakValueDictionary); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_weakref_WeakValueDictionary);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_weakref_WeakValueDictionary, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/includes/stdlib.pxi":166
 * 
 * cdef weakref_WeakValueDictionary = weakref.WeakValueDictionary
 * cdef weakref_WeakSet = weakref.WeakSet             # <<<<<<<<<<<<<<
 * 
 * cdef py_inf = float('inf')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(31, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_WeakSet); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_weakref_WeakSet);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_weakref_WeakSet, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":168
 * cdef weakref_WeakSet = weakref.WeakSet
 * 
 * cdef py_inf = float('inf')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyUnicode_AsDouble(__pyx_n_u_inf); if (unlikely(__pyx_t_7 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(31, 168, __pyx_L1_error)
  __pyx_t_6 = PyFloat_FromDouble(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(31, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop_py_inf);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_py_inf, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/includes/stdlib.pxi":173
 * # Cython doesn't clean-up imported objects properly in Py3 mode,
 * # so we delete refs to all modules manually (except sys)
 * del asyncio, concurrent, collections, errno             # <<<<<<<<<<<<<<
 * del functools, inspect, itertools, socket, os, threading
 * del signal, subprocess, ssl
 */
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_asyncio) < 0) __PYX_ERR(31, 173, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_concurrent) < 0) __PYX_ERR(31, 173, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_collections) < 0) __PYX_ERR(31, 173, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_errno) < 0) __PYX_ERR(31, 173, __pyx_L1_error)

  /* "uvloop/includes/stdlib.pxi":174
 * # so we delete refs to all modules manually (except sys)
 * del asyncio, concurrent, collections, errno
 * del functools, inspect, itertools, socket, os, threading             # <<<<<<<<<<<<<<
 * del signal, subprocess, ssl
 * del time, traceback, warnings, weakref
 */
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_functools) < 0) __PYX_ERR(31, 174, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_inspect) < 0) __PYX_ERR(31, 174, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_itertools) < 0) __PYX_ERR(31, 174, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_socket_2) < 0) __PYX_ERR(31, 174, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_os) < 0) __PYX_ERR(31, 174, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_threading) < 0) __PYX_ERR(31, 174, __pyx_L1_error)

  /* "uvloop/includes/stdlib.pxi":175
 * del asyncio, concurrent, collections, errno
 * del functools, inspect, itertools, socket, os, threading
 * del signal, subprocess, ssl             # <<<<<<<<<<<<<<
 * del time, traceback, warnings, weakref
 */
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_signal) < 0) __PYX_ERR(31, 175, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_subprocess) < 0) __PYX_ERR(31, 175, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_ssl) < 0) __PYX_ERR(31, 175, __pyx_L1_error)

  /* "uvloop/includes/stdlib.pxi":176
 * del functools, inspect, itertools, socket, os, threading
 * del signal, subprocess, ssl
 * del time, traceback, warnings, weakref             # <<<<<<<<<<<<<<
 */
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_time) < 0) __PYX_ERR(31, 176, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_traceback_2) < 0) __PYX_ERR(31, 176, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_warnings) < 0) __PYX_ERR(31, 176, __pyx_L1_error)
  if (__Pyx_PyObject_DelAttrStr(__pyx_m, __pyx_n_s_weakref) < 0) __PYX_ERR(31, 176, __pyx_L1_error)

  /* "uvloop/loop.pyx":51
 * 
 * cdef:
 *     int PY39 = PY_VERSION_HEX >= 0x03090000             # <<<<<<<<<<<<<<
 *     int PY311 = PY_VERSION_HEX >= 0x030b0000
 *     int PY313 = PY_VERSION_HEX >= 0x030d0000
 */
  __pyx_v_6uvloop_4loop_PY39 = (PY_VERSION_HEX >= 0x03090000);

  /* "uvloop/loop.pyx":52
 * cdef:
 *     int PY39 = PY_VERSION_HEX >= 0x03090000
 *     int PY311 = PY_VERSION_HEX >= 0x030b0000             # <<<<<<<<<<<<<<
 *     int PY313 = PY_VERSION_HEX >= 0x030d0000
 *     uint64_t MAX_SLEEP = 3600 * 24 * 365 * 100
 */
  __pyx_v_6uvloop_4loop_PY311 = (PY_VERSION_HEX >= 0x030b0000);

  /* "uvloop/loop.pyx":53
 *     int PY39 = PY_VERSION_HEX >= 0x03090000
 *     int PY311 = PY_VERSION_HEX >= 0x030b0000
 *     int PY313 = PY_VERSION_HEX >= 0x030d0000             # <<<<<<<<<<<<<<
 *     uint64_t MAX_SLEEP = 3600 * 24 * 365 * 100
 * 
 */
  __pyx_v_6uvloop_4loop_PY313 = (PY_VERSION_HEX >= 0x030d0000);

  /* "uvloop/loop.pyx":54
 *     int PY311 = PY_VERSION_HEX >= 0x030b0000
 *     int PY313 = PY_VERSION_HEX >= 0x030d0000
 *     uint64_t MAX_SLEEP = 3600 * 24 * 365 * 100             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_6uvloop_4loop_MAX_SLEEP = 0xbbf81e00;

  /* "uvloop/loop.pyx":124
 * # Used for deprecation and removal of `loop.create_datagram_endpoint()`'s
 * # *reuse_address* parameter
 * _unset = object()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unset, __pyx_t_6) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":374
 *             self.remove_signal_handler(sig)
 * 
 *     def __sighandler(self, signum, frame):             # <<<<<<<<<<<<<<
 *         self._signals.add(signum)
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_7_Loop__sighandler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop___sighandler, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__208)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_Loop__sighandler, __pyx_t_6) < 0) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1152
 *         self._coroutine_debug_set = enabled
 * 
 *     def _get_backend_id(self):             # <<<<<<<<<<<<<<
 *         """This method is used by uvloop tests and is not part of the API."""
 *         return uv.uv_backend_fd(self.uvloop)
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_9_get_backend_id, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop__get_backend_id, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__210)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_get_backend_id, __pyx_t_6) < 0) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1268
 *         )
 * 
 *     def call_soon(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called as soon as possible.
 * 
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_context, Py_None) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_13call_soon, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_call_soon, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__212)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_3, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_call_soon, __pyx_t_3) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1285
 *             return self._call_soon(callback, None, context)
 * 
 *     def call_soon_threadsafe(self, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_soon(), but thread-safe."""
 *         if not args:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_context, Py_None) < 0) __PYX_ERR(0, 1285, __pyx_L1_error)
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_15call_soon_threadsafe, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_call_soon_threadsafe, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__214)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_6, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_call_soon_threadsafe, __pyx_t_6) < 0) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1297
 *         return handle
 * 
 *     def call_later(self, delay, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Arrange for a callback to be called at a given time.
 * 
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_context, Py_None) < 0) __PYX_ERR(0, 1297, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_17call_later, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_call_later, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__216)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_3, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_call_later, __pyx_t_3) < 0) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1334
 *             return self._call_later(when, callback, args, context)
 * 
 *     def call_at(self, when, callback, *args, context=None):             # <<<<<<<<<<<<<<
 *         """Like call_later(), but uses an absolute time.
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_context, Py_None) < 0) __PYX_ERR(0, 1334, __pyx_L1_error)
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_19call_at, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_call_at, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__218)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_6, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_call_at, __pyx_t_6) < 0) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1342
 *             when - self.time(), callback, *args, context=context)
 * 
 *     def time(self):             # <<<<<<<<<<<<<<
 *         """Return the time according to the event loop's clock.
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_21time, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_time_2, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__219)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_time, __pyx_t_6) < 0) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1351
 *         return self._time() / 1000
 * 
 *     def stop(self):             # <<<<<<<<<<<<<<
 *         """Stop running the event loop.
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_23stop, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_stop, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__220)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_stop, __pyx_t_6) < 0) __PYX_ERR(0, 1351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1366
 *                 None))
 * 
 *     def run_forever(self):             # <<<<<<<<<<<<<<
 *         """Run the event loop until stop() is called."""
 *         self._check_closed()
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_25run_forever, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_run_forever, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__222)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_run_forever, __pyx_t_6) < 0) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1384
 *             sys.set_asyncgen_hooks(*old_agen_hooks)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the event loop.
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_27close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_close, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__223)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_close, __pyx_t_6) < 0) __PYX_ERR(0, 1384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1395
 *         self._close()
 * 
 *     def get_debug(self):             # <<<<<<<<<<<<<<
 *         return bool(self._debug)
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_29get_debug, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_get_debug, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__224)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_get_debug, __pyx_t_6) < 0) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1398
 *         return bool(self._debug)
 * 
 *     def set_debug(self, enabled):             # <<<<<<<<<<<<<<
 *         self._debug = bool(enabled)
 *         if self.is_running():
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_31set_debug, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_set_debug, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__225)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_set_debug, __pyx_t_6) < 0) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1403
 *              self.call_soon_threadsafe(self._set_coroutine_debug, self._debug)
 * 
 *     def is_running(self):             # <<<<<<<<<<<<<<
 *         """Return whether the event loop is currently running."""
 *         return bool(self._running)
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_33is_running, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_is_running, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__226)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_is_running, __pyx_t_6) < 0) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1407
 *         return bool(self._running)
 * 
 *     def is_closed(self):             # <<<<<<<<<<<<<<
 *         """Returns True if the event loop was closed."""
 *         return bool(self._closed)
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_35is_closed, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_is_closed, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__227)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_is_closed, __pyx_t_6) < 0) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1411
 *         return bool(self._closed)
 * 
 *     def create_future(self):             # <<<<<<<<<<<<<<
 *         """Create a Future object attached to the loop."""
 *         return self._new_future()
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_37create_future, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_create_future, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__228)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_create_future, __pyx_t_6) < 0) __PYX_ERR(0, 1411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1415
 *         return self._new_future()
 * 
 *     def create_task(self, coro, *, name=None, context=None):             # <<<<<<<<<<<<<<
 *         """Schedule a coroutine object.
 * 
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_name_2, Py_None) < 0) __PYX_ERR(0, 1415, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_context, Py_None) < 0) __PYX_ERR(0, 1415, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_39create_task, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_create_task, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__230)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_3, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_create_task, __pyx_t_3) < 0) __PYX_ERR(0, 1415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1456
 *         return task
 * 
 *     def set_task_factory(self, factory):             # <<<<<<<<<<<<<<
 *         """Set a task factory that will be used by loop.create_task().
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_41set_task_factory, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_set_task_factory, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__232)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_set_task_factory, __pyx_t_3) < 0) __PYX_ERR(0, 1456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1470
 *         self._task_factory = factory
 * 
 *     def get_task_factory(self):             # <<<<<<<<<<<<<<
 *         """Return a task factory, or None if the default one is in use."""
 *         return self._task_factory
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_43get_task_factory, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_get_task_factory, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__233)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_get_task_factory, __pyx_t_3) < 0) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1474
 *         return self._task_factory
 * 
 *     def run_until_complete(self, future):             # <<<<<<<<<<<<<<
 *         """Run until the Future is done.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_45run_until_complete, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_run_until_complete, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__235)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_run_until_complete, __pyx_t_3) < 0) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1520
 *         return future.result()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getaddrinfo(self, object host, object port, *,
 *                           int family=0, int type=0, int proto=0, int flags=0):
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_family, __pyx_int_0) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_type, __pyx_int_0) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_proto, __pyx_int_0) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_flags, __pyx_int_0) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_47getaddrinfo, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_getaddrinfo, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_6, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_getaddrinfo, __pyx_t_6) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1532
 *             host, port, family, type, proto, flags, 1)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def getnameinfo(self, sockaddr, int flags=0):
 *         cdef:
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_50getnameinfo, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_getnameinfo, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_6, __pyx_tuple__238);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_getnameinfo, __pyx_t_6) < 0) __PYX_ERR(0, 1532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1587
 *         return await self._getnameinfo(ai.ai_addr, flags)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "uvloop/loop.pyx":1589
 *     @cython.iterable_coroutine
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,             # <<<<<<<<<<<<<<
 *                         server_hostname=None,
 *                         ssl_handshake_timeout=None,
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_server_side, Py_False) < 0) __PYX_ERR(0, 1587, __pyx_L1_error)

  /* "uvloop/loop.pyx":1590
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 *                         server_hostname=None,             # <<<<<<<<<<<<<<
 *                         ssl_handshake_timeout=None,
 *                         ssl_shutdown_timeout=None):
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_server_hostname, Py_None) < 0) __PYX_ERR(0, 1587, __pyx_L1_error)

  /* "uvloop/loop.pyx":1591
 *                         server_side=False,
 *                         server_hostname=None,
 *                         ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                         ssl_shutdown_timeout=None):
 *         """Upgrade transport to TLS.
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl_handshake_timeout, Py_None) < 0) __PYX_ERR(0, 1587, __pyx_L1_error)

  /* "uvloop/loop.pyx":1592
 *                         server_hostname=None,
 *                         ssl_handshake_timeout=None,
 *                         ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 *         """Upgrade transport to TLS.
 * 
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl_shutdown_timeout, Py_None) < 0) __PYX_ERR(0, 1587, __pyx_L1_error)

  /* "uvloop/loop.pyx":1587
 *         return await self._getnameinfo(ai.ai_addr, flags)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_tls(self, transport, protocol, sslcontext, *,
 *                         server_side=False,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_53start_tls, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_start_tls, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_3, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_start_tls, __pyx_t_3) < 0) __PYX_ERR(0, 1587, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1646
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 *                             int family=uv.AF_UNSPEC,             # <<<<<<<<<<<<<<
 *                             int flags=uv.AI_PASSIVE,
 *                             sock=None,
 */
  __pyx_k__52 = AF_UNSPEC;

  /* "uvloop/loop.pyx":1647
 *                             *,
 *                             int family=uv.AF_UNSPEC,
 *                             int flags=uv.AI_PASSIVE,             # <<<<<<<<<<<<<<
 *                             sock=None,
 *                             backlog=100,
 */
  __pyx_k__53 = AI_PASSIVE;

  /* "uvloop/loop.pyx":1643
 *         return app_transport
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":1646
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 *                             int family=uv.AF_UNSPEC,             # <<<<<<<<<<<<<<
 *                             int flags=uv.AI_PASSIVE,
 *                             sock=None,
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(AF_UNSPEC); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_family, __pyx_t_6) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1647
 *                             *,
 *                             int family=uv.AF_UNSPEC,
 *                             int flags=uv.AI_PASSIVE,             # <<<<<<<<<<<<<<
 *                             sock=None,
 *                             backlog=100,
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(AI_PASSIVE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_flags, __pyx_t_6) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":1648
 *                             int family=uv.AF_UNSPEC,
 *                             int flags=uv.AI_PASSIVE,
 *                             sock=None,             # <<<<<<<<<<<<<<
 *                             backlog=100,
 *                             ssl=None,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sock, Py_None) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_backlog, __pyx_int_100) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1650
 *                             sock=None,
 *                             backlog=100,
 *                             ssl=None,             # <<<<<<<<<<<<<<
 *                             reuse_address=None,
 *                             reuse_port=None,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl, Py_None) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1651
 *                             backlog=100,
 *                             ssl=None,
 *                             reuse_address=None,             # <<<<<<<<<<<<<<
 *                             reuse_port=None,
 *                             ssl_handshake_timeout=None,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_reuse_address, Py_None) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1652
 *                             ssl=None,
 *                             reuse_address=None,
 *                             reuse_port=None,             # <<<<<<<<<<<<<<
 *                             ssl_handshake_timeout=None,
 *                             ssl_shutdown_timeout=None,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_reuse_port, Py_None) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1653
 *                             reuse_address=None,
 *                             reuse_port=None,
 *                             ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                             ssl_shutdown_timeout=None,
 *                             start_serving=True):
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_handshake_timeout, Py_None) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1654
 *                             reuse_port=None,
 *                             ssl_handshake_timeout=None,
 *                             ssl_shutdown_timeout=None,             # <<<<<<<<<<<<<<
 *                             start_serving=True):
 *         """A coroutine which creates a TCP server bound to host and port.
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_shutdown_timeout, Py_None) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1655
 *                             ssl_handshake_timeout=None,
 *                             ssl_shutdown_timeout=None,
 *                             start_serving=True):             # <<<<<<<<<<<<<<
 *         """A coroutine which creates a TCP server bound to host and port.
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_start_serving, Py_True) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)

  /* "uvloop/loop.pyx":1643
 *         return app_transport
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_server(self, protocol_factory, host=None, port=None,
 *                             *,
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_56create_server, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_create_server, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_6, __pyx_tuple__241);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_6, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_create_server, __pyx_t_6) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "uvloop/loop.pyx":1860
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 *                                 ssl=None,             # <<<<<<<<<<<<<<
 *                                 family=0, proto=0, flags=0, sock=None,
 *                                 local_addr=None, server_hostname=None,
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl, Py_None) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_family, __pyx_int_0) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_proto, __pyx_int_0) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_flags, __pyx_int_0) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)

  /* "uvloop/loop.pyx":1861
 *                                 *,
 *                                 ssl=None,
 *                                 family=0, proto=0, flags=0, sock=None,             # <<<<<<<<<<<<<<
 *                                 local_addr=None, server_hostname=None,
 *                                 ssl_handshake_timeout=None,
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_sock, Py_None) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)

  /* "uvloop/loop.pyx":1862
 *                                 ssl=None,
 *                                 family=0, proto=0, flags=0, sock=None,
 *                                 local_addr=None, server_hostname=None,             # <<<<<<<<<<<<<<
 *                                 ssl_handshake_timeout=None,
 *                                 ssl_shutdown_timeout=None):
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_local_addr, Py_None) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_server_hostname, Py_None) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)

  /* "uvloop/loop.pyx":1863
 *                                 family=0, proto=0, flags=0, sock=None,
 *                                 local_addr=None, server_hostname=None,
 *                                 ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                 ssl_shutdown_timeout=None):
 *         """Connect to a TCP server.
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl_handshake_timeout, Py_None) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)

  /* "uvloop/loop.pyx":1864
 *                                 local_addr=None, server_hostname=None,
 *                                 ssl_handshake_timeout=None,
 *                                 ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 *         """Connect to a TCP server.
 * 
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl_shutdown_timeout, Py_None) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)

  /* "uvloop/loop.pyx":1857
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_connection(self, protocol_factory, host=None, port=None,
 *                                 *,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_59create_connection, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_create_connection, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__241);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_3, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_create_connection, __pyx_t_3) < 0) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2098
 *                                  ssl_handshake_timeout=None,
 *                                  ssl_shutdown_timeout=None,
 *                                  start_serving=True, cleanup_socket=PY313):             # <<<<<<<<<<<<<<
 *         """A coroutine which creates a UNIX Domain Socket server.
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_PY313); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_k__71 = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":2093
 *             return tr, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_backlog, __pyx_int_100) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)

  /* "uvloop/loop.pyx":2095
 *     @cython.iterable_coroutine
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,             # <<<<<<<<<<<<<<
 *                                  ssl_handshake_timeout=None,
 *                                  ssl_shutdown_timeout=None,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_sock, Py_None) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl, Py_None) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)

  /* "uvloop/loop.pyx":2096
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 *                                  ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                  ssl_shutdown_timeout=None,
 *                                  start_serving=True, cleanup_socket=PY313):
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_handshake_timeout, Py_None) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)

  /* "uvloop/loop.pyx":2097
 *                                  *, backlog=100, sock=None, ssl=None,
 *                                  ssl_handshake_timeout=None,
 *                                  ssl_shutdown_timeout=None,             # <<<<<<<<<<<<<<
 *                                  start_serving=True, cleanup_socket=PY313):
 *         """A coroutine which creates a UNIX Domain Socket server.
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_shutdown_timeout, Py_None) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)

  /* "uvloop/loop.pyx":2098
 *                                  ssl_handshake_timeout=None,
 *                                  ssl_shutdown_timeout=None,
 *                                  start_serving=True, cleanup_socket=PY313):             # <<<<<<<<<<<<<<
 *         """A coroutine which creates a UNIX Domain Socket server.
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_start_serving, Py_True) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_PY313); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_cleanup_socket, __pyx_t_6) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":2093
 *             return tr, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_server(self, protocol_factory, path=None,
 *                                  *, backlog=100, sock=None, ssl=None,
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_62create_unix_server, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_create_unix_server, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_6, __pyx_tuple__244);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_6, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_create_unix_server, __pyx_t_6) < 0) __PYX_ERR(0, 2093, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2235
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "uvloop/loop.pyx":2237
 *     @cython.iterable_coroutine
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,             # <<<<<<<<<<<<<<
 *                                      server_hostname=None,
 *                                      ssl_handshake_timeout=None,
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl, Py_None) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_sock, Py_None) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)

  /* "uvloop/loop.pyx":2238
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 *                                      server_hostname=None,             # <<<<<<<<<<<<<<
 *                                      ssl_handshake_timeout=None,
 *                                      ssl_shutdown_timeout=None):
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_server_hostname, Py_None) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)

  /* "uvloop/loop.pyx":2239
 *                                      ssl=None, sock=None,
 *                                      server_hostname=None,
 *                                      ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                      ssl_shutdown_timeout=None):
 * 
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl_handshake_timeout, Py_None) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)

  /* "uvloop/loop.pyx":2240
 *                                      server_hostname=None,
 *                                      ssl_handshake_timeout=None,
 *                                      ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_ssl_shutdown_timeout, Py_None) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)

  /* "uvloop/loop.pyx":2235
 *         return server
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_unix_connection(self, protocol_factory, path=None, *,
 *                                      ssl=None, sock=None,
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_65create_unix_connection, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_create_unix_connection, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__244);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_3, __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_create_unix_connection, __pyx_t_3) < 0) __PYX_ERR(0, 2235, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2335
 *             return tr, protocol
 * 
 *     def default_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Default exception handler.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_68default_exception_handler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_default_exception_handler, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__247)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_default_exception_handler, __pyx_t_3) < 0) __PYX_ERR(0, 2335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2376
 *         aio_logger.error('\n'.join(log_lines), exc_info=exc_info)
 * 
 *     def get_exception_handler(self):             # <<<<<<<<<<<<<<
 *         """Return an exception handler, or None if the default one is in use.
 *         """
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_70get_exception_handler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_get_exception_handler, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__248)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_get_exception_handler, __pyx_t_3) < 0) __PYX_ERR(0, 2376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2381
 *         return self._exception_handler
 * 
 *     def set_exception_handler(self, handler):             # <<<<<<<<<<<<<<
 *         """Set handler as the new event loop exception handler.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_72set_exception_handler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_set_exception_handler, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__250)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_set_exception_handler, __pyx_t_3) < 0) __PYX_ERR(0, 2381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2398
 *         self._exception_handler = handler
 * 
 *     def call_exception_handler(self, context):             # <<<<<<<<<<<<<<
 *         """Call the current event loop's exception handler.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_74call_exception_handler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_call_exception_handler, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__252)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_call_exception_handler, __pyx_t_3) < 0) __PYX_ERR(0, 2398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2455
 *                                      exc_info=True)
 * 
 *     def add_reader(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a reader callback."""
 *         if len(args) == 0:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_76add_reader, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_add_reader, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__254)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_add_reader, __pyx_t_3) < 0) __PYX_ERR(0, 2455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2461
 *         self._add_reader(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_reader(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a reader callback."""
 *         self._remove_reader(fileobj)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_78remove_reader, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_remove_reader, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__256)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_remove_reader, __pyx_t_3) < 0) __PYX_ERR(0, 2461, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2465
 *         self._remove_reader(fileobj)
 * 
 *     def add_writer(self, fileobj, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a writer callback.."""
 *         if len(args) == 0:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_80add_writer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_add_writer, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__257)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_add_writer, __pyx_t_3) < 0) __PYX_ERR(0, 2465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2471
 *         self._add_writer(fileobj, new_Handle(self, callback, args, None))
 * 
 *     def remove_writer(self, fileobj):             # <<<<<<<<<<<<<<
 *         """Remove a writer callback."""
 *         self._remove_writer(fileobj)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_82remove_writer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_remove_writer, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__258)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_remove_writer, __pyx_t_3) < 0) __PYX_ERR(0, 2471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2475
 *         self._remove_writer(fileobj)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv(self, sock, n):
 *         """Receive data from the socket.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_84sock_recv, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_recv, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_recv, __pyx_t_3) < 0) __PYX_ERR(0, 2475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2503
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recv_into(self, sock, buf):
 *         """Receive data from the socket.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_87sock_recv_into, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_recv_into, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_recv_into, __pyx_t_3) < 0) __PYX_ERR(0, 2503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2530
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendall(self, sock, data):
 *         """Send data to the socket.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_90sock_sendall, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_sendall, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_sendall, __pyx_t_3) < 0) __PYX_ERR(0, 2530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2584
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_accept(self, sock):
 *         """Accept a connection.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_93sock_accept, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_accept, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_accept, __pyx_t_3) < 0) __PYX_ERR(0, 2584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2613
 *         return await fut
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_connect(self, sock, address):
 *         """Connect to a remote socket at address.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_96sock_connect, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_connect, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_connect, __pyx_t_3) < 0) __PYX_ERR(0, 2613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2637
 *             socket_dec_io_ref(sock)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom(self, sock, bufsize):
 *         raise NotImplementedError
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_99sock_recvfrom, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_recvfrom, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__89)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_recvfrom, __pyx_t_3) < 0) __PYX_ERR(0, 2637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2641
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_recvfrom_into(self, sock, buf, nbytes=0):
 *         raise NotImplementedError
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_102sock_recvfrom_into, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_recvfrom_into, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__238);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_recvfrom_into, __pyx_t_3) < 0) __PYX_ERR(0, 2641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2645
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def sock_sendto(self, sock, data, address):
 *         raise NotImplementedError
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_105sock_sendto, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_sock_sendto, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_sock_sendto, __pyx_t_3) < 0) __PYX_ERR(0, 2645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2649
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":2651
 *     @cython.iterable_coroutine
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,             # <<<<<<<<<<<<<<
 *                                       ssl_handshake_timeout=None,
 *                                       ssl_shutdown_timeout=None):
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl, Py_None) < 0) __PYX_ERR(0, 2649, __pyx_L1_error)

  /* "uvloop/loop.pyx":2652
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 *                                       ssl_handshake_timeout=None,             # <<<<<<<<<<<<<<
 *                                       ssl_shutdown_timeout=None):
 *         """Handle an accepted connection.
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_handshake_timeout, Py_None) < 0) __PYX_ERR(0, 2649, __pyx_L1_error)

  /* "uvloop/loop.pyx":2653
 *                                       ssl=None,
 *                                       ssl_handshake_timeout=None,
 *                                       ssl_shutdown_timeout=None):             # <<<<<<<<<<<<<<
 *         """Handle an accepted connection.
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_ssl_shutdown_timeout, Py_None) < 0) __PYX_ERR(0, 2649, __pyx_L1_error)

  /* "uvloop/loop.pyx":2649
 *         raise NotImplementedError
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_accepted_socket(self, protocol_factory, sock, *,
 *                                       ssl=None,
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_108connect_accepted_socket, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_connect_accepted_socket, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_6, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_connect_accepted_socket, __pyx_t_6) < 0) __PYX_ERR(0, 2649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2733
 *             return transport, protocol
 * 
 *     def run_in_executor(self, executor, func, *args):             # <<<<<<<<<<<<<<
 *         if aio_iscoroutine(func) or aio_iscoroutinefunction(func):
 *             raise TypeError("coroutines cannot be used with run_in_executor()")
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_111run_in_executor, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_run_in_executor, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__270)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_run_in_executor, __pyx_t_6) < 0) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2749
 *         return aio_wrap_future(executor.submit(func, *args), loop=self)
 * 
 *     def set_default_executor(self, executor):             # <<<<<<<<<<<<<<
 *         self._default_executor = executor
 * 
 */
  __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_113set_default_executor, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_set_default_executor, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__272)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_set_default_executor, __pyx_t_6) < 0) __PYX_ERR(0, 2749, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2754
 *     @cython.iterable_coroutine
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,             # <<<<<<<<<<<<<<
 *                                stdout=subprocess_PIPE,
 *                                stderr=subprocess_PIPE,
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_k__95 = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":2755
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 *                                stdout=subprocess_PIPE,             # <<<<<<<<<<<<<<
 *                                stderr=subprocess_PIPE,
 *                                universal_newlines=False,
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_k__96 = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":2756
 *                                stdin=subprocess_PIPE,
 *                                stdout=subprocess_PIPE,
 *                                stderr=subprocess_PIPE,             # <<<<<<<<<<<<<<
 *                                universal_newlines=False,
 *                                shell=True,
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_k__97 = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uvloop/loop.pyx":2754
 *     @cython.iterable_coroutine
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,             # <<<<<<<<<<<<<<
 *                                stdout=subprocess_PIPE,
 *                                stderr=subprocess_PIPE,
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "uvloop/loop.pyx":2755
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 *                                stdout=subprocess_PIPE,             # <<<<<<<<<<<<<<
 *                                stderr=subprocess_PIPE,
 *                                universal_newlines=False,
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uvloop/loop.pyx":2756
 *                                stdin=subprocess_PIPE,
 *                                stdout=subprocess_PIPE,
 *                                stderr=subprocess_PIPE,             # <<<<<<<<<<<<<<
 *                                universal_newlines=False,
 *                                shell=True,
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_6uvloop_4loop_subprocess_PIPE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/loop.pyx":2752
 *         self._default_executor = executor
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __subprocess_run(self, protocol_factory, args,
 *                                stdin=subprocess_PIPE,
 */
  __pyx_t_8 = PyTuple_New(16); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_2)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 3, Py_False)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 4, Py_True)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_int_0)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 6, Py_None)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 7, Py_None)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 8, Py_None)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 9, Py_None)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 10, Py_None)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 11, __pyx_int_0)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 12, Py_True)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 13, Py_False)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 14, Py_None)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __Pyx_INCREF(__pyx_empty_tuple);
  __Pyx_GIVEREF(__pyx_empty_tuple);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 15, __pyx_empty_tuple)) __PYX_ERR(0, 2752, __pyx_L1_error);
  __pyx_t_6 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_115_Loop__subprocess_run, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop___subprocess_run, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__98)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_Loop__subprocess_run, __pyx_t_2) < 0) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2817
 *         return proc, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uvloop/loop.pyx":2819
 *     @cython.iterable_coroutine
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,             # <<<<<<<<<<<<<<
 *                                **kwargs):
 * 
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shell, Py_True) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)

  /* "uvloop/loop.pyx":2817
 *         return proc, protocol
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_shell(self, protocol_factory, cmd, *,
 *                                shell=True,
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_118subprocess_shell, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_subprocess_shell, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_8, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_subprocess_shell, __pyx_t_8) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2832
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "uvloop/loop.pyx":2834
 *     @cython.iterable_coroutine
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):             # <<<<<<<<<<<<<<
 * 
 *         if shell:
 */
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_shell, Py_False) < 0) __PYX_ERR(0, 2832, __pyx_L1_error)

  /* "uvloop/loop.pyx":2832
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def subprocess_exec(self, protocol_factory, program, *args,
 *                               shell=False, **kwargs):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_121subprocess_exec, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_subprocess_exec, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_2, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_subprocess_exec, __pyx_t_2) < 0) __PYX_ERR(0, 2832, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2844
 *                                            **kwargs)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_read_pipe(self, proto_factory, pipe):
 *         """Register read pipe in event loop. Set the pipe to non-blocking mode.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_124connect_read_pipe, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_connect_read_pipe, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__107)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_connect_read_pipe, __pyx_t_2) < 0) __PYX_ERR(0, 2844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2871
 *         return transp, proto
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def connect_write_pipe(self, proto_factory, pipe):
 *         """Register write pipe in event loop.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_127connect_write_pipe, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_connect_write_pipe, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_connect_write_pipe, __pyx_t_2) < 0) __PYX_ERR(0, 2871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2898
 *         return transp, proto
 * 
 *     def add_signal_handler(self, sig, callback, *args):             # <<<<<<<<<<<<<<
 *         """Add a handler for a signal.  UNIX only.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_130add_signal_handler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_add_signal_handler, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__278)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_add_signal_handler, __pyx_t_2) < 0) __PYX_ERR(0, 2898, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":2966
 *                 raise
 * 
 *     def remove_signal_handler(self, sig):             # <<<<<<<<<<<<<<
 *         """Remove a handler for a signal.  UNIX only.
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_132remove_signal_handler, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop_remove_signal_handler, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__280)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_remove_signal_handler, __pyx_t_2) < 0) __PYX_ERR(0, 2966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3006
 *                                        local_addr=None, remote_addr=None, *,
 *                                        family=0, proto=0, flags=0,
 *                                        reuse_address=_unset, reuse_port=None,             # <<<<<<<<<<<<<<
 *                                        allow_broadcast=None, sock=None):
 *         """A coroutine which creates a datagram endpoint.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_unset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__114 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_family, __pyx_int_0) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_proto, __pyx_int_0) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_flags, __pyx_int_0) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)

  /* "uvloop/loop.pyx":3006
 *                                        local_addr=None, remote_addr=None, *,
 *                                        family=0, proto=0, flags=0,
 *                                        reuse_address=_unset, reuse_port=None,             # <<<<<<<<<<<<<<
 *                                        allow_broadcast=None, sock=None):
 *         """A coroutine which creates a datagram endpoint.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_unset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_reuse_address, __pyx_t_8) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_reuse_port, Py_None) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)

  /* "uvloop/loop.pyx":3007
 *                                        family=0, proto=0, flags=0,
 *                                        reuse_address=_unset, reuse_port=None,
 *                                        allow_broadcast=None, sock=None):             # <<<<<<<<<<<<<<
 *         """A coroutine which creates a datagram endpoint.
 * 
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_allow_broadcast, Py_None) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_sock, Py_None) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)

  /* "uvloop/loop.pyx":3002
 *         return True
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def create_datagram_endpoint(self, protocol_factory,
 *                                        local_addr=None, remote_addr=None, *,
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_134create_datagram_endpoint, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_create_datagram_endpoint, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_8, __pyx_tuple__241);
  __Pyx_CyFunction_SetDefaultsKwDict(__pyx_t_8, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_create_datagram_endpoint, __pyx_t_8) < 0) __PYX_ERR(0, 3002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3167
 *         return udp, protocol
 * 
 *     def _monitor_fs(self, path: str, callback) -> asyncio.Handle:             # <<<<<<<<<<<<<<
 *         cdef:
 *             UVFSEvent fs_handle
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_path, __pyx_n_s_str) < 0) __PYX_ERR(0, 3167, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_return, __pyx_kp_s_asyncio_Handle) < 0) __PYX_ERR(0, 3167, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_137_monitor_fs, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop__monitor_fs, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__283)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_monitor_fs, __pyx_t_2) < 0) __PYX_ERR(0, 3167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3180
 *         return fs_handle
 * 
 *     def _check_default_executor(self):             # <<<<<<<<<<<<<<
 *         if self._executor_shutdown_called:
 *             raise RuntimeError('Executor shutdown has been called')
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_139_check_default_executor, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop__check_default_executor, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__284)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_check_default_executor, __pyx_t_2) < 0) __PYX_ERR(0, 3180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3184
 *             raise RuntimeError('Executor shutdown has been called')
 * 
 *     def _asyncgen_finalizer_hook(self, agen):             # <<<<<<<<<<<<<<
 *         self._asyncgens.discard(agen)
 *         if not self.is_closed():
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_141_asyncgen_finalizer_hook, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop__asyncgen_finalizer_hook, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__286)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_asyncgen_finalizer_hook, __pyx_t_2) < 0) __PYX_ERR(0, 3184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3189
 *             self.call_soon_threadsafe(self.create_task, agen.aclose())
 * 
 *     def _asyncgen_firstiter_hook(self, agen):             # <<<<<<<<<<<<<<
 *         if self._asyncgens_shutdown_called:
 *             warnings_warn(
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_143_asyncgen_firstiter_hook, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop__asyncgen_firstiter_hook, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__287)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_asyncgen_firstiter_hook, __pyx_t_2) < 0) __PYX_ERR(0, 3189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3198
 *         self._asyncgens.add(agen)
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_asyncgens(self):
 *         """Shutdown all active asynchronous generators."""
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_145shutdown_asyncgens, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_shutdown_asyncgens, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__123)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_shutdown_asyncgens, __pyx_t_2) < 0) __PYX_ERR(0, 3198, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3223
 *                 })
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def shutdown_default_executor(self, timeout=None):
 *         """Schedule the shutdown of the default executor.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_148shutdown_default_executor, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Loop_shutdown_default_executor, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__124)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__244);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_shutdown_default_executor, __pyx_t_2) < 0) __PYX_ERR(0, 3223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "uvloop/loop.pyx":3251
 *             self._default_executor.shutdown(wait=False)
 * 
 *     def _do_shutdown(self, future):             # <<<<<<<<<<<<<<
 *         try:
 *             self._default_executor.shutdown(wait=True)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_151_do_shutdown, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop__do_shutdown, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__291)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Loop, __pyx_n_s_do_shutdown, __pyx_t_2) < 0) __PYX_ERR(0, 3251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Loop);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_153__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__292)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_4Loop_155__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Loop___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__294)); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3260
 * 
 * # Expose pointer for integration with other C-extensions
 * def libuv_get_loop_t_ptr(loop):             # <<<<<<<<<<<<<<
 *     return PyCapsule_New(<void *>(<Loop>loop).uvloop, NULL, NULL)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_1libuv_get_loop_t_ptr, 0, __pyx_n_s_libuv_get_loop_t_ptr, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__296)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_libuv_get_loop_t_ptr, __pyx_t_2) < 0) __PYX_ERR(0, 3260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3264
 * 
 * 
 * def libuv_get_version():             # <<<<<<<<<<<<<<
 *     return uv.uv_version()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_3libuv_get_version, 0, __pyx_n_s_libuv_get_version, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__297)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_libuv_get_version, __pyx_t_2) < 0) __PYX_ERR(0, 3264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3268
 * 
 * 
 * def _testhelper_unwrap_capsuled_pointer(obj):             # <<<<<<<<<<<<<<
 *     return <uint64_t>PyCapsule_GetPointer(obj, NULL)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_5_testhelper_unwrap_capsuled_pointer, 0, __pyx_n_s_testhelper_unwrap_capsuled_poin, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__299)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_testhelper_unwrap_capsuled_poin, __pyx_t_2) < 0) __PYX_ERR(0, 3268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3295
 * 
 * 
 * class _SyncSocketReaderFuture(aio_Future):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, sock, loop):
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_Future);
  __Pyx_GIVEREF(__pyx_v_6uvloop_4loop_aio_Future);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_6uvloop_4loop_aio_Future)) __PYX_ERR(0, 3295, __pyx_L1_error);
  __pyx_t_8 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_8, __pyx_n_s_SyncSocketReaderFuture, __pyx_n_s_SyncSocketReaderFuture, (PyObject *) NULL, __pyx_n_s_uvloop_loop, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_8 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_6, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 3295, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3297
 * class _SyncSocketReaderFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_1__init__, 0, __pyx_n_s_SyncSocketReaderFuture___init, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__301)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_init, __pyx_t_2) < 0) __PYX_ERR(0, 3297, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3302
 *         self.__loop = loop
 * 
 *     def __remove_reader(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_reader(self.__sock)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_3__remove_reader, 0, __pyx_n_s_SyncSocketReaderFuture___remove, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__302)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_SyncSocketReaderFuture__remove, __pyx_t_2) < 0) __PYX_ERR(0, 3302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/loop.pyx":3307
 *             self.__sock = None
 * 
 *     if PY39:             # <<<<<<<<<<<<<<
 *         def cancel(self, msg=None):
 *             self.__remove_reader()
 */
  __pyx_t_4 = (__pyx_v_6uvloop_4loop_PY39 != 0);
  if (__pyx_t_4) {

    /* "uvloop/loop.pyx":3308
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self, msg=msg)
 */
    __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_5cancel, 0, __pyx_n_s_SyncSocketReaderFuture_cancel, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__304)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__244);
    if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_cancel, __pyx_t_2) < 0) __PYX_ERR(0, 3308, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uvloop/loop.pyx":3307
 *             self.__sock = None
 * 
 *     if PY39:             # <<<<<<<<<<<<<<
 *         def cancel(self, msg=None):
 *             self.__remove_reader()
 */
    goto __pyx_L2;
  }

  /* "uvloop/loop.pyx":3313
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_reader()
 *             aio_Future.cancel(self)
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketReaderFuture_7cancel, 0, __pyx_n_s_SyncSocketReaderFuture_cancel, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__305)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_SetNameInClass(__pyx_t_6, __pyx_n_s_cancel, __pyx_t_2) < 0) __PYX_ERR(0, 3313, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L2:;

  /* "uvloop/loop.pyx":3295
 * 
 * 
 * class _SyncSocketReaderFuture(aio_Future):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, sock, loop):
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_SyncSocketReaderFuture, __pyx_t_8, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SyncSocketReaderFuture, __pyx_t_2) < 0) __PYX_ERR(0, 3295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/loop.pyx":3318
 * 
 * 
 * class _SyncSocketWriterFuture(aio_Future):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, sock, loop):
 */
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_Future);
  __Pyx_GIVEREF(__pyx_v_6uvloop_4loop_aio_Future);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_6uvloop_4loop_aio_Future)) __PYX_ERR(0, 3318, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PEP560_update_bases(__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_CalculateMetaclass(NULL, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_6, __pyx_t_3, __pyx_n_s_SyncSocketWriterFuture, __pyx_n_s_SyncSocketWriterFuture, (PyObject *) NULL, __pyx_n_s_uvloop_loop, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_3 != __pyx_t_8) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_8) < 0))) __PYX_ERR(0, 3318, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/loop.pyx":3320
 * class _SyncSocketWriterFuture(aio_Future):
 * 
 *     def __init__(self, sock, loop):             # <<<<<<<<<<<<<<
 *         aio_Future.__init__(self, loop=loop)
 *         self.__sock = sock
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_1__init__, 0, __pyx_n_s_SyncSocketWriterFuture___init, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__306)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_8) < 0) __PYX_ERR(0, 3320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/loop.pyx":3325
 *         self.__loop = loop
 * 
 *     def __remove_writer(self):             # <<<<<<<<<<<<<<
 *         if self.__sock is not None and self.__sock.fileno() != -1:
 *             self.__loop.remove_writer(self.__sock)
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_3__remove_writer, 0, __pyx_n_s_SyncSocketWriterFuture___remove, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__307)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_SyncSocketWriterFuture__remove, __pyx_t_8) < 0) __PYX_ERR(0, 3325, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/loop.pyx":3330
 *             self.__sock = None
 * 
 *     if PY39:             # <<<<<<<<<<<<<<
 *         def cancel(self, msg=None):
 *             self.__remove_writer()
 */
  __pyx_t_4 = (__pyx_v_6uvloop_4loop_PY39 != 0);
  if (__pyx_t_4) {

    /* "uvloop/loop.pyx":3331
 * 
 *     if PY39:
 *         def cancel(self, msg=None):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self, msg=msg)
 */
    __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_5cancel, 0, __pyx_n_s_SyncSocketWriterFuture_cancel, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__308)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_8, __pyx_tuple__244);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_cancel, __pyx_t_8) < 0) __PYX_ERR(0, 3331, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "uvloop/loop.pyx":3330
 *             self.__sock = None
 * 
 *     if PY39:             # <<<<<<<<<<<<<<
 *         def cancel(self, msg=None):
 *             self.__remove_writer()
 */
    goto __pyx_L3;
  }

  /* "uvloop/loop.pyx":3336
 * 
 *     else:
 *         def cancel(self):             # <<<<<<<<<<<<<<
 *             self.__remove_writer()
 *             aio_Future.cancel(self)
 */
  /*else*/ {
    __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23_SyncSocketWriterFuture_7cancel, 0, __pyx_n_s_SyncSocketWriterFuture_cancel, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__309)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_cancel, __pyx_t_8) < 0) __PYX_ERR(0, 3336, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __pyx_L3:;

  /* "uvloop/loop.pyx":3318
 * 
 * 
 * class _SyncSocketWriterFuture(aio_Future):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, sock, loop):
 */
  __pyx_t_8 = __Pyx_Py3ClassCreate(__pyx_t_6, __pyx_n_s_SyncSocketWriterFuture, __pyx_t_3, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SyncSocketWriterFuture, __pyx_t_8) < 0) __PYX_ERR(0, 3318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/cbhandles.pyx":158
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Handle_9cancel, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Handle_cancel, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__310)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Handle, __pyx_n_s_cancel, __pyx_t_3) < 0) __PYX_ERR(10, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Handle);

  /* "uvloop/cbhandles.pyx":161
 *         self._cancel()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Handle_11cancelled, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Handle_cancelled, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__311)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Handle, __pyx_n_s_cancelled, __pyx_t_3) < 0) __PYX_ERR(10, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Handle);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Handle_13__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Handle___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__312)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Handle_15__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Handle___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__313)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/cbhandles.pyx":306
 *         return '<' + ' '.join(info) + '>'
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self._cancelled
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11TimerHandle_7cancelled, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimerHandle_cancelled, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__314)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_TimerHandle, __pyx_n_s_cancelled, __pyx_t_3) < 0) __PYX_ERR(10, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_TimerHandle);

  /* "uvloop/cbhandles.pyx":309
 *         return self._cancelled
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._cancel()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11TimerHandle_9cancel, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimerHandle_cancel, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__315)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_TimerHandle, __pyx_n_s_cancel, __pyx_t_3) < 0) __PYX_ERR(10, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_TimerHandle);

  /* "uvloop/cbhandles.pyx":312
 *         self._cancel()
 * 
 *     def when(self):             # <<<<<<<<<<<<<<
 *         return self._when
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11TimerHandle_11when, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimerHandle_when, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__316)); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_TimerHandle, __pyx_n_s_when, __pyx_t_3) < 0) __PYX_ERR(10, 312, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_TimerHandle);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11TimerHandle_13__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimerHandle___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__317)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11TimerHandle_15__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TimerHandle___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__318)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/pseudosock.pyx":63
 *         return s
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         raise TypeError("Cannot serialize socket object")
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_5__getstate__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket___getstate, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__319)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_getstate, __pyx_t_3) < 0) __PYX_ERR(11, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":66
 *         raise TypeError("Cannot serialize socket object")
 * 
 *     def fileno(self):             # <<<<<<<<<<<<<<
 *         return self._fd
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_7fileno, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_fileno, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__320)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_fileno, __pyx_t_3) < 0) __PYX_ERR(11, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":69
 *         return self._fd
 * 
 *     def dup(self):             # <<<<<<<<<<<<<<
 *         fd = os_dup(self._fd)
 *         sock = socket_socket(self._family, self._type, self._proto, fileno=fd)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_9dup, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_dup, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__322)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_dup, __pyx_t_3) < 0) __PYX_ERR(11, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":75
 *         return sock
 * 
 *     def get_inheritable(self):             # <<<<<<<<<<<<<<
 *         return os_get_inheritable(self._fd)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_11get_inheritable, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_get_inheritable, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__323)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_get_inheritable, __pyx_t_3) < 0) __PYX_ERR(11, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":78
 *         return os_get_inheritable(self._fd)
 * 
 *     def set_inheritable(self):             # <<<<<<<<<<<<<<
 *         os_set_inheritable(self._fd)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_13set_inheritable, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_set_inheritable, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__324)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_set_inheritable, __pyx_t_3) < 0) __PYX_ERR(11, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":81
 *         os_set_inheritable(self._fd)
 * 
 *     def ioctl(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_15ioctl, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_ioctl, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__326)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_ioctl, __pyx_t_3) < 0) __PYX_ERR(11, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":84
 *         pass
 * 
 *     def getsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_17getsockopt, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_getsockopt, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__328)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_getsockopt, __pyx_t_3) < 0) __PYX_ERR(11, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":91
 *             sock.detach()
 * 
 *     def setsockopt(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_19setsockopt, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_setsockopt, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__329)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_setsockopt, __pyx_t_3) < 0) __PYX_ERR(11, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":98
 *             sock.detach()
 * 
 *     def getpeername(self):             # <<<<<<<<<<<<<<
 *         if self._peername is not None:
 *             return self._peername
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_21getpeername, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_getpeername, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__331)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_getpeername, __pyx_t_3) < 0) __PYX_ERR(11, 98, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":109
 *             sock.detach()
 * 
 *     def getsockname(self):             # <<<<<<<<<<<<<<
 *         if self._sockname is not None:
 *             return self._sockname
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_23getsockname, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_getsockname, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__332)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_getsockname, __pyx_t_3) < 0) __PYX_ERR(11, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":120
 *             sock.detach()
 * 
 *     def share(self, process_id):             # <<<<<<<<<<<<<<
 *         sock = self._make_sock()
 *         try:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_25share, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_share, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__334)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_share, __pyx_t_3) < 0) __PYX_ERR(11, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":127
 *             sock.detach()
 * 
 *     def accept(self):             # <<<<<<<<<<<<<<
 *         self._na('accept() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_27accept, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_accept, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__335)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_accept, __pyx_t_3) < 0) __PYX_ERR(11, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":130
 *         self._na('accept() method')
 * 
 *     def connect(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_29connect, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_connect, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__337)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_connect, __pyx_t_3) < 0) __PYX_ERR(11, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":133
 *         self._na('connect() method')
 * 
 *     def connect_ex(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('connect_ex() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_31connect_ex, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_connect_ex, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__338)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_connect_ex, __pyx_t_3) < 0) __PYX_ERR(11, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":136
 *         self._na('connect_ex() method')
 * 
 *     def bind(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('bind() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_33bind, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_bind, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__339)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_bind, __pyx_t_3) < 0) __PYX_ERR(11, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":139
 *         self._na('bind() method')
 * 
 *     def listen(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('listen() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_35listen, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_listen, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__340)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_listen, __pyx_t_3) < 0) __PYX_ERR(11, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":142
 *         self._na('listen() method')
 * 
 *     def makefile(self):             # <<<<<<<<<<<<<<
 *         self._na('makefile() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_37makefile, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_makefile, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__341)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_makefile, __pyx_t_3) < 0) __PYX_ERR(11, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":145
 *         self._na('makefile() method')
 * 
 *     def sendfile(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendfile() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_39sendfile, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_sendfile, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__342)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_sendfile, __pyx_t_3) < 0) __PYX_ERR(11, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":148
 *         self._na('sendfile() method')
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self._na('close() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_41close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_close, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__343)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_close, __pyx_t_3) < 0) __PYX_ERR(11, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":151
 *         self._na('close() method')
 * 
 *     def detach(self):             # <<<<<<<<<<<<<<
 *         self._na('detach() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_43detach, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_detach, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__344)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_detach, __pyx_t_3) < 0) __PYX_ERR(11, 151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":154
 *         self._na('detach() method')
 * 
 *     def shutdown(self, *args):             # <<<<<<<<<<<<<<
 *         self._na('shutdown() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_45shutdown, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_shutdown, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__345)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_shutdown, __pyx_t_3) < 0) __PYX_ERR(11, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":157
 *         self._na('shutdown() method')
 * 
 *     def sendmsg_afalg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg_afalg() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_47sendmsg_afalg, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_sendmsg_afalg, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__346)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_sendmsg_afalg, __pyx_t_3) < 0) __PYX_ERR(11, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":160
 *         self._na('sendmsg_afalg() method')
 * 
 *     def sendmsg(self):             # <<<<<<<<<<<<<<
 *         self._na('sendmsg() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_49sendmsg, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_sendmsg, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__347)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_sendmsg, __pyx_t_3) < 0) __PYX_ERR(11, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":163
 *         self._na('sendmsg() method')
 * 
 *     def sendto(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendto() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_51sendto, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_sendto, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__348)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_sendto, __pyx_t_3) < 0) __PYX_ERR(11, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":166
 *         self._na('sendto() method')
 * 
 *     def send(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('send() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_53send, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_send, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__349)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_send, __pyx_t_3) < 0) __PYX_ERR(11, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":169
 *         self._na('send() method')
 * 
 *     def sendall(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('sendall() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_55sendall, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_sendall, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__350)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_sendall, __pyx_t_3) < 0) __PYX_ERR(11, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":172
 *         self._na('sendall() method')
 * 
 *     def recv_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv_into() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_57recv_into, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_recv_into, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__351)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_recv_into, __pyx_t_3) < 0) __PYX_ERR(11, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":175
 *         self._na('recv_into() method')
 * 
 *     def recvfrom_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom_into() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_59recvfrom_into, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_recvfrom_into, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__352)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_recvfrom_into, __pyx_t_3) < 0) __PYX_ERR(11, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":178
 *         self._na('recvfrom_into() method')
 * 
 *     def recvmsg_into(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg_into() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_61recvmsg_into, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_recvmsg_into, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__353)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_recvmsg_into, __pyx_t_3) < 0) __PYX_ERR(11, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":181
 *         self._na('recvmsg_into() method')
 * 
 *     def recvmsg(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvmsg() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_63recvmsg, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_recvmsg, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__354)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_recvmsg, __pyx_t_3) < 0) __PYX_ERR(11, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":184
 *         self._na('recvmsg() method')
 * 
 *     def recvfrom(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recvfrom() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_65recvfrom, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_recvfrom, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__355)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_recvfrom, __pyx_t_3) < 0) __PYX_ERR(11, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":187
 *         self._na('recvfrom() method')
 * 
 *     def recv(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         self._na('recv() method')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_67recv, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_recv, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__356)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_recv, __pyx_t_3) < 0) __PYX_ERR(11, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":190
 *         self._na('recv() method')
 * 
 *     def settimeout(self, value):             # <<<<<<<<<<<<<<
 *         if value == 0:
 *             return
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_69settimeout, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_settimeout, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__358)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_settimeout, __pyx_t_3) < 0) __PYX_ERR(11, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":196
 *             'settimeout(): only 0 timeout is allowed on transport sockets')
 * 
 *     def gettimeout(self):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_71gettimeout, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_gettimeout, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__359)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_gettimeout, __pyx_t_3) < 0) __PYX_ERR(11, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":199
 *         return 0
 * 
 *     def setblocking(self, flag):             # <<<<<<<<<<<<<<
 *         if not flag:
 *             return
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_73setblocking, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket_setblocking, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__361)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_setblocking, __pyx_t_3) < 0) __PYX_ERR(11, 199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":205
 *             'setblocking(): transport sockets cannot be blocking')
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_75__enter__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket___enter, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__362)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_enter, __pyx_t_3) < 0) __PYX_ERR(11, 205, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/pseudosock.pyx":208
 *         self._na('context manager protocol')
 * 
 *     def __exit__(self, *err):             # <<<<<<<<<<<<<<
 *         self._na('context manager protocol')
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_77__exit__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket___exit, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__364)); if (unlikely(!__pyx_t_3)) __PYX_ERR(11, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_exit, __pyx_t_3) < 0) __PYX_ERR(11, 208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_79__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__366)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PseudoSocket, (type(self), 0xbfaab04, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PseudoSocket__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12PseudoSocket_81__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PseudoSocket___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__367)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_PseudoSocket, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_PseudoSocket);

  /* "uvloop/lru.pyx":1
 * cdef object _LRU_MARKER = object()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(12, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_6uvloop_4loop__LRU_MARKER);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop__LRU_MARKER, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8LruCache_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LruCache___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__368)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_LruCache, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_LruCache);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_LruCache, (type(self), 0x0e1f4d2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_LruCache__set_state(self, __pyx_state)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8LruCache_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_LruCache___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__369)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_LruCache, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_LruCache);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVHandle_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVHandle___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__370)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVHandle_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVHandle___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__371)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_14UVSocketHandle_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVSocketHandle___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__372)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_14UVSocketHandle_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVSocketHandle___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__373)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_7UVAsync_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVAsync___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__374)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_7UVAsync_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVAsync___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__375)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6UVIdle_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVIdle___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__376)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6UVIdle_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVIdle___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__377)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_7UVCheck_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVCheck___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__378)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_7UVCheck_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVCheck___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__379)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_7UVTimer_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVTimer___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__380)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_7UVTimer_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVTimer___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__381)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6UVPoll_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVPoll___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__382)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6UVPoll_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVPoll___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__383)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/basetransport.pyx":230
 *             return bool(not self._is_reading())
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_3get_protocol, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_get_protocol, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__384)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_get_protocol, __pyx_t_3) < 0) __PYX_ERR(18, 230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":233
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._set_protocol(protocol)
 *         if self._is_reading():
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_5set_protocol, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_set_protocol, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__386)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_set_protocol, __pyx_t_3) < 0) __PYX_ERR(18, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":239
 *             self._start_reading()
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         # Used by SSLProto.  Might be removed in the future.
 *         if self._conn_lost or self._closed:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_7_force_close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport__force_close, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__388)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_force_close, __pyx_t_3) < 0) __PYX_ERR(18, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":249
 *         self._schedule_call_connection_lost(exc)
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         self._force_close(None)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_9abort, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_abort, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__389)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_abort, __pyx_t_3) < 0) __PYX_ERR(18, 249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":252
 *         self._force_close(None)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._closing or self._closed:
 *             return
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_11close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_close, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__390)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_close, __pyx_t_3) < 0) __PYX_ERR(18, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":264
 *             self._schedule_call_connection_lost(None)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closing
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_13is_closing, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_is_closing, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__391)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_is_closing, __pyx_t_3) < 0) __PYX_ERR(18, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":267
 *         return self._closing
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         return self._get_write_buffer_size()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_15get_write_buffer_size, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_get_write_buffer, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__392)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_get_write_buffer_size, __pyx_t_3) < 0) __PYX_ERR(18, 267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":270
 *         return self._get_write_buffer_size()
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_17set_write_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_set_write_buffer, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__394)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__241);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_set_write_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(18, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":278
 *         self._maybe_pause_protocol()
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._low_water, self._high_water)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_19get_write_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_get_write_buffer_2, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__395)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_get_write_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(18, 278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "uvloop/handles/basetransport.pyx":281
 *         return (self._low_water, self._high_water)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if self._extra_info is not None and name in self._extra_info:
 *             return self._extra_info[name]
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_21get_extra_info, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport_get_extra_info, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__397)); if (unlikely(!__pyx_t_3)) __PYX_ERR(18, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__244);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVBaseTransport, __pyx_n_s_get_extra_info, __pyx_t_3) < 0) __PYX_ERR(18, 281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVBaseTransport);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_23__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__398)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15UVBaseTransport_25__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVBaseTransport___setstate_cytho, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__399)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_19_StreamWriteContext_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_StreamWriteContext___reduce_cyt, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__400)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_bufs,self.py_bufs_sml,self.req,self.uv_bufs,self.uv_bufs_start cannot be converted to a Python object for pickling"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_19_StreamWriteContext_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_StreamWriteContext___setstate_c, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__401)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/stream.pyx":677
 *             id(self))
 * 
 *     def write(self, object buf):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_5write, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream_write, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__403)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStream, __pyx_n_s_write, __pyx_t_3) < 0) __PYX_ERR(20, 677, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStream);

  /* "uvloop/handles/stream.pyx":690
 *         self._initiate_write()
 * 
 *     def writelines(self, bufs):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_7writelines, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream_writelines, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__405)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStream, __pyx_n_s_writelines, __pyx_t_3) < 0) __PYX_ERR(20, 690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStream);

  /* "uvloop/handles/stream.pyx":702
 *         self._initiate_write()
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         self._ensure_alive()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_9write_eof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream_write_eof, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__406)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStream, __pyx_n_s_write_eof, __pyx_t_3) < 0) __PYX_ERR(20, 702, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStream);

  /* "uvloop/handles/stream.pyx":712
 *             self._shutdown()
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_11can_write_eof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream_can_write_eof, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__407)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStream, __pyx_n_s_can_write_eof, __pyx_t_3) < 0) __PYX_ERR(20, 712, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStream);

  /* "uvloop/handles/stream.pyx":715
 *         return True
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return self._is_reading()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_13is_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream_is_reading, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__408)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStream, __pyx_n_s_is_reading, __pyx_t_3) < 0) __PYX_ERR(20, 715, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStream);

  /* "uvloop/handles/stream.pyx":718
 *         return self._is_reading()
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         if self._closing or not self._is_reading():
 *             return
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_15pause_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream_pause_reading, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__409)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStream, __pyx_n_s_pause_reading, __pyx_t_3) < 0) __PYX_ERR(20, 718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStream);

  /* "uvloop/handles/stream.pyx":723
 *         self._stop_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         if self._is_reading() or self._closing:
 *             return
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_17resume_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream_resume_reading, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__410)); if (unlikely(!__pyx_t_3)) __PYX_ERR(20, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStream, __pyx_n_s_resume_reading, __pyx_t_3) < 0) __PYX_ERR(20, 723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStream);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_19__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__411)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8UVStream_21__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStream___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__412)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/streamserver.pyx":118
 *         raise NotImplementedError
 * 
 *     def __on_ssl_connected(self, transport, fut):             # <<<<<<<<<<<<<<
 *         exc = fut.exception()
 *         if exc is not None:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_14UVStreamServer_3_UVStreamServer__on_ssl_connected, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStreamServer___on_ssl_connecte, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__414)); if (unlikely(!__pyx_t_3)) __PYX_ERR(21, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVStreamServer, __pyx_n_s_UVStreamServer__on_ssl_connecte, __pyx_t_3) < 0) __PYX_ERR(21, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVStreamServer);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_14UVStreamServer_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStreamServer___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__415)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_14UVStreamServer_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVStreamServer___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__416)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9TCPServer_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TCPServer___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__417)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9TCPServer_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TCPServer___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__418)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/tcp.pyx":157
 *         UVBaseTransport._call_connection_made(self)
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         if name == 'sockname':
 *             if self.__sockname_set:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12TCPTransport_1get_extra_info, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TCPTransport_get_extra_info, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__419)); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__244);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_TCPTransport, __pyx_n_s_get_extra_info, __pyx_t_3) < 0) __PYX_ERR(3, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_TCPTransport);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12TCPTransport_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TCPTransport___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__420)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12TCPTransport_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TCPTransport___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__421)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18_TCPConnectRequest_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TCPConnectRequest___reduce_cyth, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__422)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18_TCPConnectRequest_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_TCPConnectRequest___setstate_cy, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__423)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_10UnixServer_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnixServer___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__424)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_10UnixServer_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnixServer___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__425)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_13UnixTransport_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnixTransport___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__426)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_13UnixTransport_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UnixTransport___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__427)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/pipe.pyx":150
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_1get_write_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_get_write_buff, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__428)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_get_write_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(22, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "uvloop/handles/pipe.pyx":153
 *         raise NotImplementedError
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_3set_write_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_set_write_buff, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__429)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__241);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_set_write_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(22, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "uvloop/handles/pipe.pyx":156
 *         raise NotImplementedError
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_5get_write_buffer_size, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_get_write_buff_2, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__430)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_get_write_buffer_size, __pyx_t_3) < 0) __PYX_ERR(22, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "uvloop/handles/pipe.pyx":159
 *         raise NotImplementedError
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_7write, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_write, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__432)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_write, __pyx_t_3) < 0) __PYX_ERR(22, 159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "uvloop/handles/pipe.pyx":162
 *         raise NotImplementedError
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_9writelines, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_writelines, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__434)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_writelines, __pyx_t_3) < 0) __PYX_ERR(22, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "uvloop/handles/pipe.pyx":165
 *         raise NotImplementedError
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_11write_eof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_write_eof, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__435)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_write_eof, __pyx_t_3) < 0) __PYX_ERR(22, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "uvloop/handles/pipe.pyx":168
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_13can_write_eof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_can_write_eof, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__436)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_can_write_eof, __pyx_t_3) < 0) __PYX_ERR(22, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "uvloop/handles/pipe.pyx":171
 *         raise NotImplementedError
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_15abort, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport_abort, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__437)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_ReadUnixTransport, __pyx_n_s_abort, __pyx_t_3) < 0) __PYX_ERR(22, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_ReadUnixTransport);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_17__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport___reduce_cytho, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__438)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_17ReadUnixTransport_19__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ReadUnixTransport___setstate_cyt, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__439)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/pipe.pyx":202
 *         __pipe_open(<UVStream>self, sockfd)
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18WriteUnixTransport_1pause_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_WriteUnixTransport_pause_reading, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__440)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_WriteUnixTransport, __pyx_n_s_pause_reading, __pyx_t_3) < 0) __PYX_ERR(22, 202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_WriteUnixTransport);

  /* "uvloop/handles/pipe.pyx":205
 *         raise NotImplementedError
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18WriteUnixTransport_3resume_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_WriteUnixTransport_resume_readin, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__441)); if (unlikely(!__pyx_t_3)) __PYX_ERR(22, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_WriteUnixTransport, __pyx_n_s_resume_reading, __pyx_t_3) < 0) __PYX_ERR(22, 205, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_WriteUnixTransport);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18WriteUnixTransport_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_WriteUnixTransport___reduce_cyth, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__442)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18WriteUnixTransport_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_WriteUnixTransport___setstate_cy, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__443)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_19_PipeConnectRequest_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PipeConnectRequest___reduce_cyt, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__444)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_19_PipeConnectRequest_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PipeConnectRequest___setstate_c, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__445)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVProcess_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcess___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__446)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVProcess_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcess___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__447)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":584
 *                 self._pending_calls.append((_CALL_CONNECTION_LOST, None, None))
 * 
 *     def __stdio_inited(self, waiter, stdio_fut):             # <<<<<<<<<<<<<<
 *         exc = stdio_fut.exception()
 *         if exc is not None:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_3_UVProcessTransport__stdio_inited, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport___stdio_inite, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__449)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_UVProcessTransport__stdio_inite, __pyx_t_3) < 0) __PYX_ERR(4, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":638
 *         return handle
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._protocol
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_5get_protocol, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_get_protocol, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__450)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_get_protocol, __pyx_t_3) < 0) __PYX_ERR(4, 638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":641
 *         return self._protocol
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._protocol = protocol
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_7set_protocol, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_set_protocol, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__451)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_set_protocol, __pyx_t_3) < 0) __PYX_ERR(4, 641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":644
 *         self._protocol = protocol
 * 
 *     def get_pid(self):             # <<<<<<<<<<<<<<
 *         return self._pid
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_9get_pid, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_get_pid, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__452)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_get_pid, __pyx_t_3) < 0) __PYX_ERR(4, 644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":647
 *         return self._pid
 * 
 *     def get_returncode(self):             # <<<<<<<<<<<<<<
 *         return self._returncode
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_11get_returncode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_get_returncod, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__453)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_get_returncode, __pyx_t_3) < 0) __PYX_ERR(4, 647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":650
 *         return self._returncode
 * 
 *     def get_pipe_transport(self, fd):             # <<<<<<<<<<<<<<
 *         if fd == 0:
 *             return self._stdin
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_13get_pipe_transport, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_get_pipe_tran, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__455)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_get_pipe_transport, __pyx_t_3) < 0) __PYX_ERR(4, 650, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":658
 *             return self._stderr
 * 
 *     def terminate(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGTERM)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_15terminate, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_terminate, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__456)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_terminate, __pyx_t_3) < 0) __PYX_ERR(4, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":662
 *         self._kill(uv.SIGTERM)
 * 
 *     def kill(self):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(uv.SIGKILL)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_17kill, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_kill, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__457)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_kill, __pyx_t_3) < 0) __PYX_ERR(4, 662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":666
 *         self._kill(uv.SIGKILL)
 * 
 *     def send_signal(self, int signal):             # <<<<<<<<<<<<<<
 *         self._check_proc()
 *         self._kill(signal)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_19send_signal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_send_signal, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__459)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_send_signal, __pyx_t_3) < 0) __PYX_ERR(4, 666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":670
 *         self._kill(signal)
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_21is_closing, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_is_closing, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__460)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_is_closing, __pyx_t_3) < 0) __PYX_ERR(4, 670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":673
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         if self._returncode is None:
 *             self._kill(uv.SIGKILL)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_23close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_close, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__461)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_close, __pyx_t_3) < 0) __PYX_ERR(4, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":693
 *             self._close()
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         return default
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_25get_extra_info, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport_get_extra_inf, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__462)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__244);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_get_extra_info, __pyx_t_3) < 0) __PYX_ERR(4, 693, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "uvloop/handles/process.pyx":696
 *         return default
 * 
 *     def _wait(self):             # <<<<<<<<<<<<<<
 *         fut = self._loop._new_future()
 *         if self._returncode is not None:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_27_wait, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport__wait, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__464)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVProcessTransport, __pyx_n_s_wait_2, __pyx_t_3) < 0) __PYX_ERR(4, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVProcessTransport);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_29__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport___reduce_cyth, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__465)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_18UVProcessTransport_31__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVProcessTransport___setstate_cy, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__466)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":706
 * 
 * 
 * class WriteSubprocessPipeProto(aio_BaseProtocol):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, proc, fd):
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_BaseProtocol);
  __Pyx_GIVEREF(__pyx_v_6uvloop_4loop_aio_BaseProtocol);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_6uvloop_4loop_aio_BaseProtocol)) __PYX_ERR(4, 706, __pyx_L1_error);
  __pyx_t_6 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_6, __pyx_n_s_WriteSubprocessPipeProto, __pyx_n_s_WriteSubprocessPipeProto, (PyObject *) NULL, __pyx_n_s_uvloop_loop, (PyObject *) NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__pyx_t_6 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_8, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(4, 706, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":708
 * class WriteSubprocessPipeProto(aio_BaseProtocol):
 * 
 *     def __init__(self, proc, fd):             # <<<<<<<<<<<<<<
 *         if UVLOOP_DEBUG:
 *             if type(proc) is not UVProcessTransport:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_1__init__, 0, __pyx_n_s_WriteSubprocessPipeProto___init, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__468)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_n_s_init, __pyx_t_3) < 0) __PYX_ERR(4, 708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":719
 *         self.disconnected = False
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         self.pipe = transport
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_3connection_made, 0, __pyx_n_s_WriteSubprocessPipeProto_connect, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__470)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_n_s_connection_made, __pyx_t_3) < 0) __PYX_ERR(4, 719, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":722
 *         self.pipe = transport
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ('<%s fd=%s pipe=%r>'
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_5__repr__, 0, __pyx_n_s_WriteSubprocessPipeProto___repr, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__471)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_n_s_repr, __pyx_t_3) < 0) __PYX_ERR(4, 722, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":726
 *                 % (self.__class__.__name__, self.fd, self.pipe))
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         self.disconnected = True
 *         (<UVProcessTransport>self.proc)._pipe_connection_lost(self.fd, exc)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_7connection_lost, 0, __pyx_n_s_WriteSubprocessPipeProto_connect_2, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__472)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_n_s_connection_lost, __pyx_t_3) < 0) __PYX_ERR(4, 726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":731
 *         self.proc = None
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_9pause_writing, 0, __pyx_n_s_WriteSubprocessPipeProto_pause_w, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__473)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_n_s_pause_writing, __pyx_t_3) < 0) __PYX_ERR(4, 731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":734
 *         (<UVProcessTransport>self.proc)._protocol.pause_writing()
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._protocol.resume_writing()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_24WriteSubprocessPipeProto_11resume_writing, 0, __pyx_n_s_WriteSubprocessPipeProto_resume, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__474)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_8, __pyx_n_s_resume_writing, __pyx_t_3) < 0) __PYX_ERR(4, 734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/process.pyx":706
 * 
 * 
 * class WriteSubprocessPipeProto(aio_BaseProtocol):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, proc, fd):
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_WriteSubprocessPipeProto, __pyx_t_6, __pyx_t_8, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_WriteSubprocessPipeProto, __pyx_t_3) < 0) __PYX_ERR(4, 706, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/handles/process.pyx":738
 * 
 * 
 * class ReadSubprocessPipeProto(WriteSubprocessPipeProto,             # <<<<<<<<<<<<<<
 *                               aio_Protocol):
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_WriteSubprocessPipeProto); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "uvloop/handles/process.pyx":739
 * 
 * class ReadSubprocessPipeProto(WriteSubprocessPipeProto,
 *                               aio_Protocol):             # <<<<<<<<<<<<<<
 * 
 *     def data_received(self, data):
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6)) __PYX_ERR(4, 738, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_6uvloop_4loop_aio_Protocol);
  __Pyx_GIVEREF(__pyx_v_6uvloop_4loop_aio_Protocol);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_6uvloop_4loop_aio_Protocol)) __PYX_ERR(4, 738, __pyx_L1_error);
  __pyx_t_6 = 0;

  /* "uvloop/handles/process.pyx":738
 * 
 * 
 * class ReadSubprocessPipeProto(WriteSubprocessPipeProto,             # <<<<<<<<<<<<<<
 *                               aio_Protocol):
 * 
 */
  __pyx_t_6 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_CalculateMetaclass(NULL, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_8, __pyx_t_6, __pyx_n_s_ReadSubprocessPipeProto, __pyx_n_s_ReadSubprocessPipeProto, (PyObject *) NULL, __pyx_n_s_uvloop_loop, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_6 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(4, 738, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":741
 *                               aio_Protocol):
 * 
 *     def data_received(self, data):             # <<<<<<<<<<<<<<
 *         (<UVProcessTransport>self.proc)._pipe_data_received(self.fd, data)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_23ReadSubprocessPipeProto_1data_received, 0, __pyx_n_s_ReadSubprocessPipeProto_data_rec, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__475)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_data_received, __pyx_t_2) < 0) __PYX_ERR(4, 741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uvloop/handles/process.pyx":738
 * 
 * 
 * class ReadSubprocessPipeProto(WriteSubprocessPipeProto,             # <<<<<<<<<<<<<<
 *                               aio_Protocol):
 * 
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_8, __pyx_n_s_ReadSubprocessPipeProto, __pyx_t_6, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ReadSubprocessPipeProto, __pyx_t_2) < 0) __PYX_ERR(4, 738, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/handles/fsevent.pyx":1
 * import enum             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = __Pyx_ImportDottedModule(__pyx_n_s_enum, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_enum, __pyx_t_6) < 0) __PYX_ERR(23, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/handles/fsevent.pyx":4
 * 
 * 
 * class FileSystemEvent(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     RENAME = uv.UV_RENAME
 *     CHANGE = uv.UV_CHANGE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_enum); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8)) __PYX_ERR(23, 4, __pyx_L1_error);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PEP560_update_bases(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_8, __pyx_n_s_FileSystemEvent, __pyx_n_s_FileSystemEvent, (PyObject *) NULL, __pyx_n_s_uvloop_loop, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_8 != __pyx_t_6) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_6) < 0))) __PYX_ERR(23, 4, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/handles/fsevent.pyx":5
 * 
 * class FileSystemEvent(enum.IntEnum):
 *     RENAME = uv.UV_RENAME             # <<<<<<<<<<<<<<
 *     CHANGE = uv.UV_CHANGE
 *     RENAME_CHANGE = RENAME | CHANGE
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__uv_fs_event(UV_RENAME); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_RENAME, __pyx_t_6) < 0) __PYX_ERR(23, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/handles/fsevent.pyx":6
 * class FileSystemEvent(enum.IntEnum):
 *     RENAME = uv.UV_RENAME
 *     CHANGE = uv.UV_CHANGE             # <<<<<<<<<<<<<<
 *     RENAME_CHANGE = RENAME | CHANGE
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__uv_fs_event(UV_CHANGE); if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_CHANGE, __pyx_t_6) < 0) __PYX_ERR(23, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "uvloop/handles/fsevent.pyx":7
 *     RENAME = uv.UV_RENAME
 *     CHANGE = uv.UV_CHANGE
 *     RENAME_CHANGE = RENAME | CHANGE             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_6 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_RENAME);
  if (unlikely(!__pyx_t_6)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_RENAME);
  }
  if (unlikely(!__pyx_t_6)) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_CHANGE);
  if (unlikely(!__pyx_t_9)) {
    PyErr_Clear();
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_CHANGE);
  }
  if (unlikely(!__pyx_t_9)) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyNumber_Or(__pyx_t_6, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_RENAME_CHANGE, __pyx_t_10) < 0) __PYX_ERR(23, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "uvloop/handles/fsevent.pyx":4
 * 
 * 
 * class FileSystemEvent(enum.IntEnum):             # <<<<<<<<<<<<<<
 *     RENAME = uv.UV_RENAME
 *     CHANGE = uv.UV_CHANGE
 */
  __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_FileSystemEvent, __pyx_t_8, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FileSystemEvent, __pyx_t_10) < 0) __PYX_ERR(23, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/handles/fsevent.pyx":78
 *             UVHandle._close(<UVHandle>self)
 * 
 *     def cancel(self):             # <<<<<<<<<<<<<<
 *         self._close()
 * 
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVFSEvent_1cancel, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVFSEvent_cancel, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__476)); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVFSEvent, __pyx_n_s_cancel, __pyx_t_8) < 0) __PYX_ERR(23, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVFSEvent);

  /* "uvloop/handles/fsevent.pyx":81
 *         self._close()
 * 
 *     def cancelled(self):             # <<<<<<<<<<<<<<
 *         return self.running == 0
 * 
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVFSEvent_3cancelled, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVFSEvent_cancelled, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__477)); if (unlikely(!__pyx_t_8)) __PYX_ERR(23, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UVFSEvent, __pyx_n_s_cancelled, __pyx_t_8) < 0) __PYX_ERR(23, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UVFSEvent);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVFSEvent_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVFSEvent___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__478)); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_8) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVFSEvent_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVFSEvent___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__479)); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_8) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVRequest_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVRequest___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__480)); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_8) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_9UVRequest_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UVRequest___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__481)); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_8) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_14SockAddrHolder_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SockAddrHolder___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__482)); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_8) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.addr must be explicitly requested with @auto_pickle(True)"
 */
  __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_14SockAddrHolder_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SockAddrHolder___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__483)); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_8) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "uvloop/dns.pyx":79
 * 
 * 
 * cdef LruCache sockaddrs = LruCache(maxsize=DNS_PYADDR_TO_SOCKADDR_CACHE_SIZE)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyInt_From___pyx_anon_enum(__pyx_e_6uvloop_4loop_DNS_PYADDR_TO_SOCKADDR_CACHE_SIZE); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_maxsize, __pyx_t_3) < 0) __PYX_ERR(5, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6uvloop_4loop_LruCache), __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_XGOTREF((PyObject *)__pyx_v_6uvloop_4loop_sockaddrs);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop_sockaddrs, ((struct __pyx_obj_6uvloop_4loop_LruCache *)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8AddrInfo_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AddrInfo___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__484)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_8AddrInfo_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AddrInfo___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__485)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15AddrInfoRequest_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AddrInfoRequest___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__486)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15AddrInfoRequest_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AddrInfoRequest___setstate_cytho, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__487)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15NameInfoRequest_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NameInfoRequest___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__488)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15NameInfoRequest_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NameInfoRequest___setstate_cytho, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__489)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/sslproto.pyx":29
 *         self.context = context
 * 
 *     def get_extra_info(self, name, default=None):             # <<<<<<<<<<<<<<
 *         """Get optional transport information."""
 *         return self._ssl_protocol._get_extra_info(name, default)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_3get_extra_info, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_get_extra, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__490)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__244);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_get_extra_info, __pyx_t_3) < 0) __PYX_ERR(25, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":33
 *         return self._ssl_protocol._get_extra_info(name, default)
 * 
 *     def set_protocol(self, protocol):             # <<<<<<<<<<<<<<
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_5set_protocol, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_set_protoc, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__491)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_set_protocol, __pyx_t_3) < 0) __PYX_ERR(25, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":36
 *         self._ssl_protocol._set_app_protocol(protocol)
 * 
 *     def get_protocol(self):             # <<<<<<<<<<<<<<
 *         return self._ssl_protocol._app_protocol
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_7get_protocol, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_get_protoc, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__492)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_get_protocol, __pyx_t_3) < 0) __PYX_ERR(25, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":39
 *         return self._ssl_protocol._app_protocol
 * 
 *     def is_closing(self):             # <<<<<<<<<<<<<<
 *         return self._closed
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_9is_closing, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_is_closing, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__493)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_is_closing, __pyx_t_3) < 0) __PYX_ERR(25, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":42
 *         return self._closed
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close the transport.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_11close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_close, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__494)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_close, __pyx_t_3) < 0) __PYX_ERR(25, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":60
 *                 "object>", ResourceWarning)
 * 
 *     def is_reading(self):             # <<<<<<<<<<<<<<
 *         return not self._ssl_protocol._app_reading_paused
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_15is_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_is_reading, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__495)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_is_reading, __pyx_t_3) < 0) __PYX_ERR(25, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":63
 *         return not self._ssl_protocol._app_reading_paused
 * 
 *     def pause_reading(self):             # <<<<<<<<<<<<<<
 *         """Pause the receiving end.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_17pause_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_pause_read, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__496)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_pause_reading, __pyx_t_3) < 0) __PYX_ERR(25, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":71
 *         self._ssl_protocol._pause_reading()
 * 
 *     def resume_reading(self):             # <<<<<<<<<<<<<<
 *         """Resume the receiving end.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_19resume_reading, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_resume_rea, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__497)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_resume_reading, __pyx_t_3) < 0) __PYX_ERR(25, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":79
 *         self._ssl_protocol._resume_reading(self.context.copy())
 * 
 *     def set_write_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for write flow control.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_21set_write_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_set_write, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__498)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__241);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_set_write_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(25, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":101
 *         self._ssl_protocol._control_app_writing(self.context.copy())
 * 
 *     def get_write_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._outgoing_low_water,
 *                 self._ssl_protocol._outgoing_high_water)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_23get_write_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_get_write, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__499)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_get_write_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(25, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":105
 *                 self._ssl_protocol._outgoing_high_water)
 * 
 *     def get_write_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the write buffers."""
 *         return self._ssl_protocol._get_write_buffer_size()
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_25get_write_buffer_size, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_get_write_2, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__500)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_get_write_buffer_size, __pyx_t_3) < 0) __PYX_ERR(25, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":109
 *         return self._ssl_protocol._get_write_buffer_size()
 * 
 *     def set_read_buffer_limits(self, high=None, low=None):             # <<<<<<<<<<<<<<
 *         """Set the high- and low-water limits for read flow control.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_27set_read_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_set_read_b, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__501)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__241);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_set_read_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(25, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":131
 *         self._ssl_protocol._control_ssl_reading()
 * 
 *     def get_read_buffer_limits(self):             # <<<<<<<<<<<<<<
 *         return (self._ssl_protocol._incoming_low_water,
 *                 self._ssl_protocol._incoming_high_water)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_29get_read_buffer_limits, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_get_read_b, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__502)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_get_read_buffer_limits, __pyx_t_3) < 0) __PYX_ERR(25, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":135
 *                 self._ssl_protocol._incoming_high_water)
 * 
 *     def get_read_buffer_size(self):             # <<<<<<<<<<<<<<
 *         """Return the current size of the read buffer."""
 *         return self._ssl_protocol._get_read_buffer_size()
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_31get_read_buffer_size, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_get_read_b_2, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__503)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_get_read_buffer_size, __pyx_t_3) < 0) __PYX_ERR(25, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":144
 *         return self._ssl_protocol._app_writing_paused
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         """Write some data bytes to the transport.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_33write, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_write, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__504)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_write, __pyx_t_3) < 0) __PYX_ERR(25, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":157
 *         self._ssl_protocol._write_appdata((data,), self.context.copy())
 * 
 *     def writelines(self, list_of_data):             # <<<<<<<<<<<<<<
 *         """Write a list (or any iterable) of data bytes to the transport.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_35writelines, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_writelines, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__505)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_writelines, __pyx_t_3) < 0) __PYX_ERR(25, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":165
 *         self._ssl_protocol._write_appdata(list_of_data, self.context.copy())
 * 
 *     def write_eof(self):             # <<<<<<<<<<<<<<
 *         """Close the write end after flushing buffered data.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_37write_eof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_write_eof, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__506)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_write_eof, __pyx_t_3) < 0) __PYX_ERR(25, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":172
 *         raise NotImplementedError
 * 
 *     def can_write_eof(self):             # <<<<<<<<<<<<<<
 *         """Return True if this transport supports write_eof(), False if not."""
 *         return False
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_39can_write_eof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_can_write, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__507)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_can_write_eof, __pyx_t_3) < 0) __PYX_ERR(25, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":176
 *         return False
 * 
 *     def abort(self):             # <<<<<<<<<<<<<<
 *         """Close the transport immediately.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_41abort, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport_abort, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__508)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_abort, __pyx_t_3) < 0) __PYX_ERR(25, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":185
 *         self._force_close(None)
 * 
 *     def _force_close(self, exc):             # <<<<<<<<<<<<<<
 *         self._closed = True
 *         self._ssl_protocol._abort(exc)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_43_force_close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport__force_clo, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__509)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_force_close, __pyx_t_3) < 0) __PYX_ERR(25, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "uvloop/sslproto.pyx":189
 *         self._ssl_protocol._abort(exc)
 * 
 *     def _test__append_write_backlog(self, data):             # <<<<<<<<<<<<<<
 *         # for test only
 *         self._ssl_protocol._write_backlog.append(data)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_45_test__append_write_backlog, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport__test__app, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__510)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop__SSLProtocolTransport, __pyx_n_s_test__append_write_backlog, __pyx_t_3) < 0) __PYX_ERR(25, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop__SSLProtocolTransport);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_47__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport___reduce_c, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__511)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_21_SSLProtocolTransport_49__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocolTransport___setstate, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__512)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/sslproto.pyx":310
 *         self._waiter = None
 * 
 *     def _get_app_transport(self, context=None):             # <<<<<<<<<<<<<<
 *         if self._app_transport is None:
 *             if self._app_transport_created:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_7_get_app_transport, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol__get_app_transport, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__514)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__244);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_get_app_transport, __pyx_t_3) < 0) __PYX_ERR(25, 310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "uvloop/sslproto.pyx":319
 *         return self._app_transport
 * 
 *     def connection_made(self, transport):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is made.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_9connection_made, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol_connection_made, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__515)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_connection_made, __pyx_t_3) < 0) __PYX_ERR(25, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "uvloop/sslproto.pyx":327
 *         self._start_handshake()
 * 
 *     def connection_lost(self, exc):             # <<<<<<<<<<<<<<
 *         """Called when the low-level connection is lost or closed.
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_11connection_lost, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol_connection_lost, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__516)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_connection_lost, __pyx_t_3) < 0) __PYX_ERR(25, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "uvloop/sslproto.pyx":361
 *             self._handshake_timeout_handle = None
 * 
 *     def get_buffer(self, n):             # <<<<<<<<<<<<<<
 *         cdef size_t want = n
 *         if want > SSL_READ_MAX_SIZE:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_13get_buffer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol_get_buffer, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__518)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_get_buffer, __pyx_t_3) < 0) __PYX_ERR(25, 361, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "uvloop/sslproto.pyx":374
 *         return self._ssl_buffer_view
 * 
 *     def buffer_updated(self, nbytes):             # <<<<<<<<<<<<<<
 *         self._incoming_write(PyMemoryView_FromMemory(
 *             self._ssl_buffer, nbytes, PyBUF_WRITE))
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_15buffer_updated, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol_buffer_updated, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__520)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_buffer_updated, __pyx_t_3) < 0) __PYX_ERR(25, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "uvloop/sslproto.pyx":390
 *             self._do_shutdown()
 * 
 *     def eof_received(self):             # <<<<<<<<<<<<<<
 *         """Called when the other end of the low-level stream
 *         is half-closed.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_17eof_received, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol_eof_received, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__521)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_eof_received, __pyx_t_3) < 0) __PYX_ERR(25, 390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "uvloop/sslproto.pyx":911
 *     # Flow control for writes to SSL socket
 * 
 *     def pause_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer goes over
 *         the high-water mark.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_19pause_writing, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol_pause_writing, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__522)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 911, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_pause_writing, __pyx_t_3) < 0) __PYX_ERR(25, 911, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "uvloop/sslproto.pyx":918
 *         self._ssl_writing_paused = True
 * 
 *     def resume_writing(self):             # <<<<<<<<<<<<<<
 *         """Called when the low-level transport's buffer drains below
 *         the low-water mark.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_21resume_writing, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol_resume_writing, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__523)); if (unlikely(!__pyx_t_3)) __PYX_ERR(25, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_SSLProtocol, __pyx_n_s_resume_writing, __pyx_t_3) < 0) __PYX_ERR(25, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_SSLProtocol);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_23__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__524)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_11SSLProtocol_25__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SSLProtocol___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__525)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15_UDPSendContext_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPSendContext___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__526)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.py_buf,self.req cannot be converted to a Python object for pickling"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_15_UDPSendContext_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPSendContext___setstate_cytho, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__527)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/handles/udp.pyx":282
 *     # === Public API ===
 * 
 *     def sendto(self, data, addr=None):             # <<<<<<<<<<<<<<
 *         if not data:
 *             # Replicating asyncio logic here.
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12UDPTransport_3sendto, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPTransport_sendto, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__529)); if (unlikely(!__pyx_t_3)) __PYX_ERR(26, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__244);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_UDPTransport, __pyx_n_s_sendto, __pyx_t_3) < 0) __PYX_ERR(26, 282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_UDPTransport);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12UDPTransport_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPTransport___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__530)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12UDPTransport_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPTransport___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__531)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/server.pyx":1
 * import asyncio             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_3) < 0) __PYX_ERR(27, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/server.pyx":51
 *     # Public API
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aenter__(self):
 *         return self
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_3__aenter__, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Server___aenter, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__189)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_aenter, __pyx_t_3) < 0) __PYX_ERR(27, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "uvloop/server.pyx":55
 *         return self
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def __aexit__(self, *exc):
 *         self.close()
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_6__aexit__, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Server___aexit, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__190)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_aexit, __pyx_t_3) < 0) __PYX_ERR(27, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "uvloop/server.pyx":63
 *         return '<%s sockets=%r>' % (self.__class__.__name__, self.sockets)
 * 
 *     def get_loop(self):             # <<<<<<<<<<<<<<
 *         return self._loop
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_11get_loop, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Server_get_loop, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__532)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_get_loop, __pyx_t_3) < 0) __PYX_ERR(27, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "uvloop/server.pyx":66
 *         return self._loop
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def wait_closed(self):
 *         # Do not remove `self._servers is None` below
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_13wait_closed, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Server_wait_closed, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__191)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_wait_closed, __pyx_t_3) < 0) __PYX_ERR(27, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "uvloop/server.pyx":77
 *         await waiter
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         cdef list servers
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_16close, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Server_close, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__535)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_close, __pyx_t_3) < 0) __PYX_ERR(27, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "uvloop/server.pyx":96
 *             self._unref()
 * 
 *     def is_serving(self):             # <<<<<<<<<<<<<<
 *         return self._serving
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_18is_serving, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Server_is_serving, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__536)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_is_serving, __pyx_t_3) < 0) __PYX_ERR(27, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "uvloop/server.pyx":99
 *         return self._serving
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def start_serving(self):
 *         self._start_serving()
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_20start_serving, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Server_start_serving, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__192)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_start_serving, __pyx_t_3) < 0) __PYX_ERR(27, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "uvloop/server.pyx":103
 *         self._start_serving()
 * 
 *     @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 *     async def serve_forever(self):
 *         if self._serving_forever_fut is not None:
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_23serve_forever, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_Server_serve_forever, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__193)); if (unlikely(!__pyx_t_3)) __PYX_ERR(27, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_6uvloop_4loop_Server, __pyx_n_s_serve_forever, __pyx_t_3) < 0) __PYX_ERR(27, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_6uvloop_4loop_Server);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_26__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Server___reduce_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__537)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_6Server_28__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Server___setstate_cython, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__538)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":3369
 * 
 * # Used in UVProcess
 * cdef vint __atfork_installed = 0             # <<<<<<<<<<<<<<
 * cdef vint __forking = 0
 * cdef Loop __forking_loop = None
 */
  __pyx_v_6uvloop_4loop___atfork_installed = 0;

  /* "uvloop/loop.pyx":3370
 * # Used in UVProcess
 * cdef vint __atfork_installed = 0
 * cdef vint __forking = 0             # <<<<<<<<<<<<<<
 * cdef Loop __forking_loop = None
 * 
 */
  __pyx_v_6uvloop_4loop___forking = 0;

  /* "uvloop/loop.pyx":3371
 * cdef vint __atfork_installed = 0
 * cdef vint __forking = 0
 * cdef Loop __forking_loop = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF((PyObject *)__pyx_v_6uvloop_4loop___forking_loop);
  __Pyx_DECREF_SET(__pyx_v_6uvloop_4loop___forking_loop, ((struct __pyx_obj_6uvloop_4loop_Loop *)Py_None));
  __Pyx_GIVEREF(Py_None);

  /* "uvloop/loop.pyx":3396
 * 
 * # Install PyMem* memory allocators
 * cdef vint __mem_installed = 0             # <<<<<<<<<<<<<<
 * cdef __install_pymem():
 *     global __mem_installed
 */
  __pyx_v_6uvloop_4loop___mem_installed = 0;

  /* "uvloop/loop.pyx":3422
 * # Helpers for tests
 * 
 * @cython.iterable_coroutine             # <<<<<<<<<<<<<<
 * async def _test_coroutine_1():
 *     return 42
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_7_test_coroutine_1, __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_test_coroutine_1, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__194)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 3422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test_coroutine_1, __pyx_t_3) < 0) __PYX_ERR(0, 3422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PseudoSocket(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_10__pyx_unpickle_PseudoSocket, 0, __pyx_n_s_pyx_unpickle_PseudoSocket, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__540)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PseudoSocket, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_PseudoSocket__set_state(<PseudoSocket> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PseudoSocket__set_state(PseudoSocket __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._family = __pyx_state[0]; __pyx_result._fd = __pyx_state[1]; __pyx_result._peername = __pyx_state[2]; __pyx_result._proto = __pyx_state[3]; __pyx_result._sockname = __pyx_state[4]; __pyx_result._type = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_6uvloop_4loop_12__pyx_unpickle_LruCache, 0, __pyx_n_s_pyx_unpickle_LruCache, NULL, __pyx_n_s_uvloop_loop, __pyx_d, ((PyObject *)__pyx_codeobj__541)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_LruCache, __pyx_t_3) < 0) __PYX_ERR(6, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uvloop/loop.pyx":1
 * # cython: language_level=3, embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * import asyncio
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init uvloop.loop", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init uvloop.loop");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* DivInt[long] */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
            return PyInt_FromLong(a & b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if ((intval & PyLong_MASK) == intval) {
            long last_digit = (long) __Pyx_PyLong_Digits(op1)[0];
            long result = intval & (likely(__Pyx_PyLong_IsPos(op1)) ? last_digit : (PyLong_MASK - last_digit + 1));
            return PyLong_FromLong(result);
        }
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_and(op1, op2);
            }
        }
                x = a & b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla & llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceAnd : PyNumber_And)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* py_set_discard_unhashable */
static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_discard */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    return found;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || PY_VERSION_HEX >= 0x030d0000 || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3 & PY_VERSION_HEX < 0x030d0000
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* RaiseClosureNameError */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a != b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
#ifdef HAVE_LONG_LONG
            if (sizeof(PY_LONG_LONG) > sizeof(long)) {
                PY_LONG_LONG result = (PY_LONG_LONG)a * (PY_LONG_LONG)b;
                return (result >= LONG_MIN && result <= LONG_MAX) ?
                    PyInt_FromLong((long)result) : PyLong_FromLongLong(result);
            }
#endif
#if CYTHON_USE_TYPE_SLOTS
            return PyInt_Type.tp_as_number->nb_multiply(op1, op2);
#else
            return PyNumber_Multiply(op1, op2);
#endif
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            }
        }
                CYTHON_UNUSED_VAR(a);
                CYTHON_UNUSED_VAR(b);
                #ifdef HAVE_LONG_LONG
                lla = a;
                goto long_long;
                #else
                return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
                #endif
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("multiply", return NULL)
            result = ((double)a) * (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#endif

/* ReturnWithStopIteration */
static void __Pyx__ReturnWithStopIteration(PyObject* value) {
    PyObject *exc, *args;
#if CYTHON_COMPILING_IN_CPYTHON
    __Pyx_PyThreadState_declare
    if (PY_VERSION_HEX >= 0x030C00A6
            || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        args = PyTuple_New(1);
        if (unlikely(!args)) return;
        Py_INCREF(value);
        PyTuple_SET_ITEM(args, 0, value);
        exc = PyType_Type.tp_call(PyExc_StopIteration, args, NULL);
        Py_DECREF(args);
        if (!exc) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_value)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(PyExc_StopIteration);
        __Pyx_ErrRestore(PyExc_StopIteration, exc, NULL);
        return;
    }
    #endif
#else
    args = PyTuple_Pack(1, value);
    if (unlikely(!args)) return;
    exc = PyObject_Call(PyExc_StopIteration, args, NULL);
    Py_DECREF(args);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(PyExc_StopIteration, exc);
    Py_DECREF(exc);
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Coroutine */
static void __Pyx_CoroutineAwait_dealloc(PyObject *self) {
    PyObject_GC_UnTrack(self);
    Py_CLEAR(((__pyx_CoroutineAwaitObject*)self)->coroutine);
    __Pyx_PyHeapTypeObject_GC_Del(self);
}
static int __Pyx_CoroutineAwait_traverse(__pyx_CoroutineAwaitObject *self, visitproc visit, void *arg) {
    Py_VISIT(self->coroutine);
    return 0;
}
static int __Pyx_CoroutineAwait_clear(__pyx_CoroutineAwaitObject *self) {
    Py_CLEAR(self->coroutine);
    return 0;
}
static PyObject *__Pyx_CoroutineAwait_Next(__pyx_CoroutineAwaitObject *self) {
    return __Pyx_Generator_Next(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_Send(__pyx_CoroutineAwaitObject *self, PyObject *value) {
    return __Pyx_Coroutine_Send(self->coroutine, value);
}
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args) {
    return __Pyx_Coroutine_Throw(self->coroutine, args);
}
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_self(PyObject *self) {
    Py_INCREF(self);
    return self;
}
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_CoroutineAwait_no_new(PyTypeObject *type, PyObject *args, PyObject *kwargs) {
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(args);
    CYTHON_UNUSED_VAR(kwargs);
    PyErr_SetString(PyExc_TypeError, "cannot instantiate type, use 'await coroutine' instead");
    return NULL;
}
#endif
static PyObject *__Pyx_CoroutineAwait_reduce_ex(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    PyErr_Format(PyExc_TypeError, "cannot pickle '%.200s' object",
                         Py_TYPE(self)->tp_name);
    return NULL;
}
static PyMethodDef __pyx_CoroutineAwait_methods[] = {
    {"send", (PyCFunction) __Pyx_CoroutineAwait_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_CoroutineAwait_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_CoroutineAwait_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineAwaitType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CoroutineAwait_dealloc},
    {Py_tp_traverse, (void *)__Pyx_CoroutineAwait_traverse},
    {Py_tp_clear, (void *)__Pyx_CoroutineAwait_clear},
#if !CYTHON_COMPILING_IN_PYPY
    {Py_tp_new, (void *)__Pyx_CoroutineAwait_no_new},
#endif
    {Py_tp_methods, (void *)__pyx_CoroutineAwait_methods},
    {Py_tp_iter, (void *)__Pyx_CoroutineAwait_self},
    {Py_tp_iternext, (void *)__Pyx_CoroutineAwait_Next},
    {0, 0},
};
static PyType_Spec __pyx_CoroutineAwaitType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    __pyx_CoroutineAwaitType_slots
};
#else
static PyTypeObject __pyx_CoroutineAwaitType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    (destructor) __Pyx_CoroutineAwait_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    PyDoc_STR("A wrapper object implementing __await__ for coroutines."),
    (traverseproc) __Pyx_CoroutineAwait_traverse,
    (inquiry) __Pyx_CoroutineAwait_clear,
    0,
    0,
    __Pyx_CoroutineAwait_self,
    (iternextfunc) __Pyx_CoroutineAwait_Next,
    __pyx_CoroutineAwait_methods,
    0                         ,
    0                      ,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if !CYTHON_COMPILING_IN_PYPY
    __Pyx_CoroutineAwait_no_new,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
#if PY_VERSION_HEX < 0x030500B1 || defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static CYTHON_INLINE PyObject *__Pyx__Coroutine_await(PyObject *coroutine) {
    __pyx_CoroutineAwaitObject *await = PyObject_GC_New(__pyx_CoroutineAwaitObject, __pyx_CoroutineAwaitType);
    if (unlikely(!await)) return NULL;
    Py_INCREF(coroutine);
    await->coroutine = coroutine;
    PyObject_GC_Track(await);
    return (PyObject*)await;
}
#endif
#if PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_await_method(PyObject *coroutine, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static PyObject *__Pyx_Coroutine_await(PyObject *coroutine) {
    if (unlikely(!coroutine || !__Pyx_Coroutine_Check(coroutine))) {
        PyErr_SetString(PyExc_TypeError, "invalid input, expected coroutine");
        return NULL;
    }
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_compare(PyObject *obj, PyObject *other, int op) {
    PyObject* result;
    switch (op) {
        case Py_EQ: result = (other == obj) ? Py_True : Py_False; break;
        case Py_NE: result = (other != obj) ? Py_True : Py_False; break;
        default:
            result = Py_NotImplemented;
    }
    Py_INCREF(result);
    return result;
}
#endif
static PyMethodDef __pyx_Coroutine_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
#if PY_VERSION_HEX < 0x030500B1
    {"__await__", (PyCFunction) __Pyx_Coroutine_await_method, METH_NOARGS,
     (char*) PyDoc_STR("__await__() -> return an iterator to be used in await expression.")},
#endif
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Coroutine_memberlist[] = {
    {(char *) "cr_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "cr_await", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being awaited, or None")},
    {(char*) "cr_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Coroutine_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the coroutine"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the coroutine"), 0},
    {(char *) "cr_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the coroutine"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_am_await, (void *)&__Pyx_Coroutine_await},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_methods, (void *)__pyx_Coroutine_methods},
    {Py_tp_members, (void *)__pyx_Coroutine_memberlist},
    {Py_tp_getset, (void *)__pyx_Coroutine_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_CoroutineType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_CoroutineType_slots
};
#else
#if CYTHON_USE_ASYNC_SLOTS
static __Pyx_PyAsyncMethodsStruct __pyx_Coroutine_as_async = {
    __Pyx_Coroutine_await,
    0,
    0,
#if PY_VERSION_HEX >= 0x030A00A3
    0,
#endif
};
#endif
static PyTypeObject __pyx_CoroutineType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
#if CYTHON_USE_ASYNC_SLOTS
    &__pyx_Coroutine_as_async,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
#if CYTHON_USE_ASYNC_SLOTS && CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
    __Pyx_Coroutine_compare,
#else
    0,
#endif
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    0,
    __pyx_Coroutine_methods,
    __pyx_Coroutine_memberlist,
    __pyx_Coroutine_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Coroutine_init(PyObject *module) {
    CYTHON_MAYBE_UNUSED_VAR(module);
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineType_spec, NULL);
#else
    __pyx_CoroutineType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_CoroutineType = __Pyx_FetchCommonType(&__pyx_CoroutineType_type);
#endif
    if (unlikely(!__pyx_CoroutineType))
        return -1;
#ifdef __Pyx_IterableCoroutine_USED
    if (unlikely(__pyx_IterableCoroutine_init(module) == -1))
        return -1;
#endif
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineAwaitType_spec, NULL);
#else
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonType(&__pyx_CoroutineAwaitType_type);
#endif
    if (unlikely(!__pyx_CoroutineAwaitType))
        return -1;
    return 0;
}

/* GetAwaitIter */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o) {
#ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(o)) {
        return __Pyx_NewRef(o);
    }
#endif
    return __Pyx__Coroutine_GetAwaitableIter(o);
}
static void __Pyx_Coroutine_AwaitableIterError(PyObject *source) {
#if PY_VERSION_HEX >= 0x030600B3 && PY_VERSION_HEX < 0x030d0000 || defined(_PyErr_FormatFromCause)
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    _PyErr_FormatFromCause(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
#elif PY_MAJOR_VERSION >= 3
    PyObject *exc, *val, *val2, *tb;
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    assert(PyErr_Occurred());
    PyErr_Fetch(&exc, &val, &tb);
    PyErr_NormalizeException(&exc, &val, &tb);
    if (tb != NULL) {
        PyException_SetTraceback(val, tb);
        Py_DECREF(tb);
    }
    Py_DECREF(exc);
    assert(!PyErr_Occurred());
    PyErr_Format(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
    PyErr_Fetch(&exc, &val2, &tb);
    PyErr_NormalizeException(&exc, &val2, &tb);
    Py_INCREF(val);
    PyException_SetCause(val2, val);
    PyException_SetContext(val2, val);
    PyErr_Restore(exc, val2, tb);
#else
    CYTHON_UNUSED_VAR(source);
#endif
}
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *obj) {
    PyObject *res;
#if CYTHON_USE_ASYNC_SLOTS
    __Pyx_PyAsyncMethodsStruct* am = __Pyx_PyType_AsAsync(obj);
    if (likely(am && am->am_await)) {
        res = (*am->am_await)(obj);
    } else
#endif
#if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
    if (PyCoro_CheckExact(obj)) {
        return __Pyx_NewRef(obj);
    } else
#endif
#if CYTHON_COMPILING_IN_CPYTHON && defined(CO_ITERABLE_COROUTINE)
#if PY_VERSION_HEX >= 0x030C00A6
    if (PyGen_CheckExact(obj) && (PyGen_GetCode((PyGenObject*)obj)->co_flags & CO_ITERABLE_COROUTINE)) {
#else
    if (PyGen_CheckExact(obj) && ((PyGenObject*)obj)->gi_code && ((PyCodeObject *)((PyGenObject*)obj)->gi_code)->co_flags & CO_ITERABLE_COROUTINE) {
#endif
        return __Pyx_NewRef(obj);
    } else
#endif
    {
        PyObject *method = NULL;
        int is_method = __Pyx_PyObject_GetMethod(obj, __pyx_n_s_await, &method);
        if (likely(is_method)) {
            res = __Pyx_PyObject_CallOneArg(method, obj);
        } else if (likely(method)) {
            res = __Pyx_PyObject_CallNoArg(method);
        } else
            goto slot_error;
        Py_DECREF(method);
    }
    if (unlikely(!res)) {
        __Pyx_Coroutine_AwaitableIterError(obj);
        goto bad;
    }
    if (unlikely(!PyIter_Check(res))) {
        __Pyx_TypeName res_type_name = __Pyx_PyType_GetName(Py_TYPE(res));
        PyErr_Format(PyExc_TypeError,
            "__await__() returned non-iterator of type '" __Pyx_FMT_TYPENAME "'", res_type_name);
        __Pyx_DECREF_TypeName(res_type_name);
        Py_CLEAR(res);
    } else {
        int is_coroutine = 0;
        #ifdef __Pyx_Coroutine_USED
        is_coroutine |= __Pyx_Coroutine_Check(res);
        #endif
        #if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
        is_coroutine |= PyCoro_CheckExact(res);
        #endif
        if (unlikely(is_coroutine)) {
            /* __await__ must return an *iterator*, not
               a coroutine or another awaitable (see PEP 492) */
            PyErr_SetString(PyExc_TypeError,
                            "__await__() returned a coroutine");
            Py_CLEAR(res);
        }
    }
    return res;
slot_error:
    {
        __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
        PyErr_Format(PyExc_TypeError,
            "object " __Pyx_FMT_TYPENAME " can't be used in 'await' expression", obj_type_name);
        __Pyx_DECREF_TypeName(obj_type_name);
    }
bad:
    return NULL;
}

/* CoroutineYieldFrom */
  static PyObject* __Pyx__Coroutine_Yield_From_Generic(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    PyObject *source_gen = __Pyx__Coroutine_GetAwaitableIter(source);
    if (unlikely(!source_gen)) {
        return NULL;
    }
    if (__Pyx_Coroutine_Check(source_gen)) {
        retval = __Pyx_Generator_Next(source_gen);
    } else {
        retval = __Pyx_PyObject_GetIterNextFunc(source_gen)(source_gen);
    }
    if (retval) {
        gen->yieldfrom = source_gen;
        return retval;
    }
    Py_DECREF(source_gen);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    if (__Pyx_Coroutine_Check(source)) {
        if (unlikely(((__pyx_CoroutineObject*)source)->yieldfrom)) {
            PyErr_SetString(
                PyExc_RuntimeError,
                "coroutine is being awaited already");
            return NULL;
        }
        retval = __Pyx_Generator_Next(source);
#ifdef __Pyx_AsyncGen_USED
    } else if (__pyx_PyAsyncGenASend_CheckExact(source)) {
        retval = __Pyx_async_gen_asend_iternext(source);
#endif
    } else {
        return __Pyx__Coroutine_Yield_From_Generic(gen, source);
    }
    if (retval) {
        Py_INCREF(source);
        gen->yieldfrom = source;
    }
    return retval;
}

/* pep479 */
  static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* RaiseUnboundLocalError */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* CallUnboundCMethod0 */
  static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
    if (is_list || !PyMapping_Check(o))
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* py_dict_keys */
  static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_keys, d);
    else
        return PyDict_Keys(d);
}

/* MergeKeywords */
  static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping) {
    PyObject *iter, *key = NULL, *value = NULL;
    int source_is_dict, result;
    Py_ssize_t orig_length, ppos = 0;
    iter = __Pyx_dict_iterator(source_mapping, 0, __pyx_n_s_items, &orig_length, &source_is_dict);
    if (unlikely(!iter)) {
        PyObject *args;
        if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError))) goto bad;
        PyErr_Clear();
        args = PyTuple_Pack(1, source_mapping);
        if (likely(args)) {
            PyObject *fallback = PyObject_Call((PyObject*)&PyDict_Type, args, NULL);
            Py_DECREF(args);
            if (likely(fallback)) {
                iter = __Pyx_dict_iterator(fallback, 1, __pyx_n_s_items, &orig_length, &source_is_dict);
                Py_DECREF(fallback);
            }
        }
        if (unlikely(!iter)) goto bad;
    }
    while (1) {
        result = __Pyx_dict_iter_next(iter, orig_length, &ppos, &key, &value, NULL, source_is_dict);
        if (unlikely(result < 0)) goto bad;
        if (!result) break;
        if (unlikely(PyDict_Contains(kwdict, key))) {
            __Pyx_RaiseDoubleKeywordsError("function", key);
            result = -1;
        } else {
            result = PyDict_SetItem(kwdict, key, value);
        }
        Py_DECREF(key);
        Py_DECREF(value);
        if (unlikely(result < 0)) goto bad;
    }
    Py_XDECREF(iter);
    return 0;
bad:
    Py_XDECREF(iter);
    return -1;
}

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetAttr3 */
  #if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* UnicodeConcatInPlace */
  # if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
static int
__Pyx_unicode_modifiable(PyObject *unicode)
{
    if (Py_REFCNT(unicode) != 1)
        return 0;
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* py_set_remove */
  static int __Pyx_PySet_RemoveNotFound(PyObject *set, PyObject *key, int found) {
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    if (likely(found == 0)) {
        PyObject *tup;
        tup = PyTuple_Pack(1, key);
        if (!tup)
            return -1;
        PyErr_SetObject(PyExc_KeyError, tup);
        Py_DECREF(tup);
        return -1;
    }
    return found;
}
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found != 1)) {
        return __Pyx_PySet_RemoveNotFound(set, key, found);
    }
    return 0;
}

/* PyIntCompare */
  static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a == b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* RaiseKeywordRequired */
  static void __Pyx_RaiseKeywordRequired(const char* func_name, PyObject* kw_name) {
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() needs keyword-only argument %U", func_name, kw_name);
        #else
        "%s() needs keyword-only argument %s", func_name,
        PyString_AS_STRING(kw_name));
        #endif
}

/* CIntToPyUnicode */
  static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* pop */
  static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (__Pyx_IS_TYPE(L, &PySet_Type)) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* PyObjectLookupSpecial */
  #if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* decode_c_bytes */
  static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* DelItemInt */
  static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_DelItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               int is_list, CYTHON_NCP_UNUSED int wraparound) {
#if !CYTHON_USE_TYPE_SLOTS
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_DelItem(o, i);
    }
#else
    PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
    PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
    if ((!is_list) && mm && mm->mp_ass_subscript) {
        PyObject *key = PyInt_FromSsize_t(i);
        return likely(key) ? mm->mp_ass_subscript(o, key, (PyObject *)NULL) : -1;
    }
    if (likely(sm && sm->sq_ass_item)) {
        if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
            Py_ssize_t l = sm->sq_length(o);
            if (likely(l >= 0)) {
                i += l;
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    return -1;
                PyErr_Clear();
            }
        }
        return sm->sq_ass_item(o, i, (PyObject *)NULL);
    }
#endif
    return __Pyx_DelItem_Generic(o, PyInt_FromSsize_t(i));
}

/* StringJoin */
  static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    (void) __Pyx_PyObject_CallMethod1;
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION < 3
    return _PyString_Join(sep, values);
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    return _PyBytes_Join(sep, values);
#else
    return __Pyx_PyObject_CallMethod1(sep, __pyx_n_s_join, values);
#endif
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__196);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* ValidateBasesTuple */
  #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
  static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
  static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType_3_0_11
#define __PYX_HAVE_RT_ImportType_3_0_11
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_11 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_11 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportDottedModule */
  #if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__198;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* PatchInspect */
  static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* PatchAsyncIO */
  static PyObject* __Pyx_patch_asyncio(PyObject* module) {
#if PY_VERSION_HEX < 0x030500B2 &&\
        (defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED)) &&\
        (!defined(CYTHON_PATCH_ASYNCIO) || CYTHON_PATCH_ASYNCIO)
    PyObject *patch_module = NULL;
    static int asyncio_patched = 0;
    if (unlikely((!asyncio_patched) && module)) {
        PyObject *package;
        package = __Pyx_Import(__pyx_n_s_asyncio_coroutines, NULL, 0);
        if (package) {
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "coroutines"), ""
"try:\n"
"    coro_types = _module._COROUTINE_TYPES\n"
"except AttributeError: pass\n"
"else:\n"
"    if _cython_coroutine_type is not None and _cython_coroutine_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_coroutine_type,)\n"
"    if _cython_generator_type is not None and _cython_generator_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_generator_type,)\n"
"_module._COROUTINE_TYPES = coro_types\n"
            );
        } else {
            PyErr_Clear();
            package = __Pyx_Import(__pyx_n_s_asyncio_tasks, NULL, 0);
            if (unlikely(!package)) goto asyncio_done;
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "tasks"), ""
"if hasattr(_module, 'iscoroutine'):\n"
"    old_types = getattr(_module.iscoroutine, '_cython_coroutine_types', None)\n"
"    if old_types is None or not isinstance(old_types, set):\n"
"        old_types = set()\n"
"        def cy_wrap(orig_func, type=type, cython_coroutine_types=old_types):\n"
"            def cy_iscoroutine(obj): return type(obj) in cython_coroutine_types or orig_func(obj)\n"
"            cy_iscoroutine._cython_coroutine_types = cython_coroutine_types\n"
"            return cy_iscoroutine\n"
"        _module.iscoroutine = cy_wrap(_module.iscoroutine)\n"
"    if _cython_coroutine_type is not None:\n"
"        old_types.add(_cython_coroutine_type)\n"
"    if _cython_generator_type is not None:\n"
"        old_types.add(_cython_generator_type)\n"
            );
        }
        Py_DECREF(package);
        if (unlikely(!patch_module)) goto ignore;
asyncio_done:
        PyErr_Clear();
        asyncio_patched = 1;
#ifdef __Pyx_Generator_USED
        {
            PyObject *inspect_module;
            if (patch_module) {
                inspect_module = PyObject_GetAttr(patch_module, __pyx_n_s_inspect);
                Py_DECREF(patch_module);
            } else {
                inspect_module = __Pyx_Import(__pyx_n_s_inspect, NULL, 0);
            }
            if (unlikely(!inspect_module)) goto ignore;
            inspect_module = __Pyx_patch_inspect(inspect_module);
            if (unlikely(!inspect_module)) {
                Py_DECREF(module);
                module = NULL;
            }
            Py_XDECREF(inspect_module);
        }
#else
        if ((0)) return __Pyx_patch_inspect(module);
#endif
    }
    return module;
ignore:
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch asyncio package with custom generator type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    if ((0)) return __Pyx_patch_inspect(__Pyx_Coroutine_patch_module(module, NULL));
#endif
    return module;
}

/* pybytes_as_double */
  static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value;
#if PY_MAJOR_VERSION >= 3
    float_value = PyFloat_FromString(obj);
#else
    float_value = PyFloat_FromString(obj, 0);
#endif
    if (likely(float_value)) {
#if CYTHON_ASSUME_SAFE_MACROS
        double value = PyFloat_AS_DOUBLE(float_value);
#else
        double value = PyFloat_AsDouble(float_value);
#endif
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* Py3UpdateBases */
  static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* CalculateMetaclass */
  static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_MACROS
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
  static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name_2);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From___pyx_anon_enum(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        uint64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__uv_fs_event(enum uv_fs_event value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum uv_fs_event neg_one = (enum uv_fs_event) -1, const_zero = (enum uv_fs_event) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum uv_fs_event) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum uv_fs_event) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum uv_fs_event) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum uv_fs_event) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum uv_fs_event) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum uv_fs_event),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum uv_fs_event));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uv_os_fd_t(uv_os_fd_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uv_os_fd_t neg_one = (uv_os_fd_t) -1, const_zero = (uv_os_fd_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uv_os_fd_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uv_os_fd_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uv_os_fd_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uv_os_fd_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uv_os_fd_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uv_os_fd_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uv_os_fd_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyInt_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned long) 1) << (sizeof(unsigned long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned int) 1) << (sizeof(unsigned int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(size_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_short(unsigned short value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned short neg_one = (unsigned short) -1, const_zero = (unsigned short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned short) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned short) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned short) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned short) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned short),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned short));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_volatile__int(volatile int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const volatile int neg_one = (volatile int) -1, const_zero = (volatile int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(volatile int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(volatile int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(volatile int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(volatile int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(volatile int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(volatile int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(volatile int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_6uvloop_4loop_SSLProtocolState(enum __pyx_t_6uvloop_4loop_SSLProtocolState value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_6uvloop_4loop_SSLProtocolState neg_one = (enum __pyx_t_6uvloop_4loop_SSLProtocolState) -1, const_zero = (enum __pyx_t_6uvloop_4loop_SSLProtocolState) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_6uvloop_4loop_SSLProtocolState) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_6uvloop_4loop_SSLProtocolState) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_6uvloop_4loop_SSLProtocolState) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_6uvloop_4loop_SSLProtocolState) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_6uvloop_4loop_SSLProtocolState) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_6uvloop_4loop_SSLProtocolState),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_6uvloop_4loop_SSLProtocolState));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
  #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__542);
    }
    return name;
}
#endif

/* IterableCoroutine */
  #if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_IterableCoroutineType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_am_await, (void *)&__Pyx_Coroutine_await},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)__Pyx_Coroutine_await},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Coroutine_methods},
    {Py_tp_members, (void *)__pyx_Coroutine_memberlist},
    {Py_tp_getset, (void *)__pyx_Coroutine_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_IterableCoroutineType_spec = {
    __PYX_TYPE_MODULE_PREFIX "iterable_coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_IterableCoroutineType_slots
};
#else
static PyTypeObject __pyx_IterableCoroutineType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "iterable_coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
#if CYTHON_USE_ASYNC_SLOTS
    &__pyx_Coroutine_as_async,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
#if CYTHON_USE_ASYNC_SLOTS && CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
    __Pyx_Coroutine_compare,
#else
    0,
#endif
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    __Pyx_Coroutine_await,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Coroutine_methods,
    __pyx_Coroutine_memberlist,
    __pyx_Coroutine_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if PY_VERSION_HEX >= 0x030400a1 && !CYTHON_COMPILING_IN_PYPY
    __Pyx_Coroutine_del,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_IterableCoroutine_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_IterableCoroutineType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_IterableCoroutineType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_IterableCoroutineType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_IterableCoroutineType = __Pyx_FetchCommonType(&__pyx_IterableCoroutineType_type);
#endif
    if (unlikely(!__pyx_IterableCoroutineType))
        return -1;
    return 0;
}

/* Generator */
  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
  static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* InitStrings */
  #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
